import numpy as np
from scipy.interpolate import interp1d



def SRD(IFO, f): 
    import numpy as np
    from scipy.interpolate import interp1d

    #wswfunction [S, fstop] = SRD(IFO,f)

    # % SRD - Design sensitivity curve for the specified GW detector.
    # %
    # %  [S, fstop] = SRD(DET,f)
    # %
    # %  DET    String designating desired detector. Must be one of 'aLIGO',
    # %         'aLIGO-*' (see below), aVirgo, aVirgo-*' (see below), 'ETB', 'ETC', 
    # %         'ETD', 'GEO', 'GEO-HF', 'KAGRA*' (see below), 'LIGO', 'LIGO-LV',
    # %         'LHO', 'LLO', 'LISA', 'LISA-LIGHT1', 'LISA-LIGHT2', 'LISA-WD', 'TAMA',
    # %         'Virgo', 'Virgo-LV', 'LIGO3-*' (see below), 'LIGO-A', 'LIGO-Voyager',
    # %         'LIGO-CosmicExplorer', or 'white* (see below)'.  
    # %         Not case sensitive.  
    # %  f      Vector of positive frequencies (Hz) at which design sensitivity 
    # %         is desired. 
    # %
    # %  S      One-sided power spectral density (1/Hz) for requested detector
    # %         and frequencies.  S has the same size as f.
    # %  fstop  Minimum and maximum frequency (Hz) at which fit is "guaranteed"
    # %         to be accurate.  A warning message is printed if min(f)<fstop(1)
    # %         or max(f)>fstop(2). 
    # %
    # % For full details of the source of each noise spectrum, see the code.
    # %
    # % Several different types of aLIGO spectra are available:
    # %
    # %  aLIGO        Defaults to aLIGO-ZDHP.
    # %  aLIGO-BBH    Optimized for 30-30 solar mass black hole binary inspirals.
    # %               "This is slightly more difficult from a control system 
    # %               point of view." 
    # %  aLIGO-BNS    Optimized for 1.4 solar mass neutron star binary. This is
    # %               slightly more difficult from a control system point of
    # %               view. Gives ~200 Mpc reach for a single interferometer, SNR
    # %               8, averaged over directions and polarizations.    
    # %  aLIGOEARLY-50    Possible early configuration, c. 2015, with excess
    # %                   low-frequency noise and ~50 Mpc BNS range.
    # %  aLIGOEARLY-100   Possible early configuration, c. 2016-17, with excess
    # %                   low-frequency noise and ~100 Mpc BNS range.
    # %  aLIGOEARLY-150   Possible early configuration, c. 2017-18, with ~150 Mpc
    # %                   BNS range. 
    # %  aLIGO-HF     High Freq: an example of a narrowband tuning at 1kHz.
    # %               Unlike all the other tunings, it needs a recycling mirror
    # %               with a higher reflectivity.   
    # %  aLIGO-NOSRM	No SRM: a possible early configuration, where no signal
    # %               recycling mirror is used.  
    # %  aLIGO-ZDHP	Zero Det, High Power: zero-detuning of the signal
    # %               recycling mirror, with high laser power. "This gives most
    # %               of the sensitivity with the fewest technical difficulties." 
    # %               This spectrum is only available to a maximum frequency of
    # %               5000 Hz.
    # %  aLIGO-ZDLP   Zero Det, Low Power: a possible early configuration with
    # %               zero-detuning of the signal recycling mirror, with low
    # %               laser power. 
    # %  aLIGO-Hooper Analytic approximation from Hooper et al., 1108.3186 / Phys
    # %               Rev D.86.024012 (SPIIR paper). WARNING: This does not match
    # %               any of the official aLIGO spectra.
    # %
    # % The available aVirgo spectra are:
    # %
    # %  aVIRGO       Defaults to aVIRGO-ZDHP.
    # %  aVIRGO-BNS   Optimized for 1.4 solar mass neutron star binary (as of May 2010). 
    # %  aVIRGO-NOSRM	No SRM: a possible early configuration, where no signal
    # %               recycling mirror is used.  
    # %  aVIRGO-ZDHP  Optimized wide-band sensitivity. 
    # %  aVIRGO-ZDLP  Zero Det, Low Power: a possible early configuration with
    # %               zero-detuning of the signal recycling mirror, with low
    # %               laser power. 
    # %  
    # % The available KAGRA spectra are:
    # %
    # %  KAGRA_Early  Anticipated sensitivity c.2019-20.
    # %  KAGRA_Mid    Anticipated sensitivity c.2020-21.
    # %  KAGRA_Late   Anticipated sensitivity c.2021-22.
    # %  KAGRA        Design sensitivity c.2022+.
    # %
    # % The available LIGO3 spectra are:
    # %
    # %  LIGO3-default Standard 3G: 100kg masses, 10dB squeezing, 2.5x thermal noise reduction
    # %  LIGO3-mid     mid-freq  3G: 100kg masses, 10dB squeezing, 5x thermal noise reduction
    # %  LIGO3-high    high-freq  3G: 100kg masses, 15dB squeezing, 2.5x thermal noise reduction
    # %  LIGO3         Defaults to LIGO3-default
    # %
    # % The 'LIGO-A' (A+), 'LIGO-Voyager', and 'LIGO-CosmicExplorer' options yield the 
    # % strawman sensitivity curves described in T1400316-v5. The implementation here
    # % is based on interpolation of sampled points extracted from Figure 3 of
    # % T1400316-v5 and should be accurate to a few percent in amplitude. These
    # % spectra are available from 10Hz to 2000Hz only.
    # %
    # % The choice IFO = 'whiteFS' where FS is a number returns a constant S =
    # % 2/FS Hz^{-1}, corresponding to unit-variance discrete white noise.  The
    # % choice 'white' uses FS='1'.  This may be handy for testing codes which
    # % use noise weighting.
    # %
    # % For information on where each spectrum is taken from, please read the
    # % relevant portion of the code.
    # %
    # % $Id: SRD.m 5600 2018-04-06 09:03:21Z patrick.sutton@LIGO.ORG $ 

    #% ---- Argument checks.
    ####narginchk(2,2);
    
    if type(IFO)!=str:
        raise TypeError('Detector argument must be a string.')

    if type(f).__module__!=np.__name__:
        raise TypeError('Input frequencies must be a numpy array')

    if np.size(f)!=len(f):
        raise TypeError('Input frequencies must be a vector.')

    if min(f)<=0:
        raise ValueError('Input frequencies must be positive.')

    #% ---- Recognize detector and compute noise spectrum.

    #if upper(IFO)==
            #  case 'ADVLIGO'
            # % ---- Deprecated.  The current aLIGO curves do not include a 
            # %      spectrum that looks like this one.  Select one of the 
            # %      aLIGO-* curves instead.
            # fstop = [0,Inf];
            # f0 = 215;
            # x = f/f0;
            # S = (1e-49)*( x.^(-4.14) -5*x.^(-2) + 111*(1-x.^2+x.^4/2)./(1+x.^2/2) );
    if IFO.upper()=='ALIGO':
        [S, fstop] = aligospectrum('aLIGO-ZDHP',f)
                # % ---- Analytic approximation to ALIGO-ZDHP spectrum data, by P Ajith,
                # %      https://dcc.ligo.org/cgi-bin/private/DocDB/ShowDocument?docid=63161 
                # %      The fit is accurate to withn 2% everywhere above 20 Hz band except
                # %      that it does not include lines.  
                # % x = f/245.4;
                # % S = 1e-48 * (0.0152*x.^(-4) + 0.2935*x.^(9/4) + 2.7951*x.^(3/2) - 6.5080*x.^(3/4) + 17.7622); 
                # % fstop = [20 1e4];
                # % ---- Analytic approximation to the ALIGO-ZDHP spectra, taken
                # %      from Del Pozzo by email, 2012.06.03.  Based on 1108.1317,
                # %      but this paper contains an error.  Corrected formula below
                # %      is accurate to within 10% in amplitude everywhere in the
                # %      range [11,3539] Hz.
                # % S = 1.35e-50 * f .* ( 60000*(f/10).^(-30) + 5*(f/50).^(-6) ...
                # %     + 1.07*(f/100).^(-3.25) + 3.7*(f/200).^(-1.25) ...
                # %     + 0.9*(f/300).^(-0.08) + 0.85*(f/1000).^(0.8) ...
                # %     + 0.53*(f/2000).^(3) );
                # % fstop = [11,3539];
    elif IFO.upper()=='AVIRGO':
        [S, fstop] = avirgospectrum('aVIRGO-ZDHP',f)
                # % case 'ET'
                # %     % ---- Deprecated pre-design version of the ET spectrum, by Leone Bosi. 
                # %     fstop = [1,Inf]; 
                # %     f0 = 200;
                # %     x = f/f0;
                # %     %
                # %     p1 = -4.05049;
                # %     p2 = -0.687263;
                # %     a0 = 185.623;
                # %     b0 = 232.565;
                # %     b1 = 31.1777;
                # %     b2 = -64.7178;
                # %     b3 = 52.242;
                # %     b4 = -42.1643;
                # %     b5 = 10.1708;
                # %     b6 = 11.5317;
                # %     c1 = 13.5793;
                # %     c2 = -36.4586;
                # %     c3 = 18.5625;
                # %     c4 = 27.4297;
                # %     S0 = 1.449936*10^-52;
                # %     %
                # %     S  = S0*(x.^p1 + a0.*x.^p2 + b0 * ...
                # %         ( 1. + b1*x + b2*x.^2 + b3*x.^3 + b4*x.^4 + b5*x.^5 + b6*x.^6 ) ...
                # %         ./ (1. + c1*x + c2*x.^2 + c3*x.^3 + c4*x.^4) );
    elif IFO.upper()=='ETB':
        [S, fstop] = etb(f)
                # % ---- This following analytic approximation is good to within 10%
                # %      across the [1,10000] Hz band.
                # % [S, fstop] = ETB_fit();
                # % S = S.^2;
                # % ---- Read sampled spectrum supplied by B. S. Sathyaprakash.

    elif IFO.upper()=='ETC':
        [S, fstop] = ETC_fit(f)
        S = S**2
    elif IFO.upper()=='ETD':
        [S, fstop] = etd(f)
                #% ---- Read sampled spectrum supplied by B. S. Sathyaprakash.
    elif IFO.upper()=='GEO':
        fstop = [40,float('inf')]
        f0 = 150
        x = f/f0
        S = (1e-46)*( (3.4*x)**(-30) + 34*x**(-1) + 20*(1-x**2+x**4/2)/(1+x**2/2) )
                #% ---- From L. P. Grishchuk, V. M. Lipunov, K. A. Postnov, M. E.
                #%      Prokhorov, and B. S. Sathyaprakash, Physics-Uspekhi 71, 3
                #%      (2000). 
    elif IFO.upper()=='GEO-HF':
        fstop = [40,float('inf')];
        S = 7.18e-46*(1+(f/1059)**2) + 4.90e-41/f**2 + 8.91e-43/f + 1.6e-17/f**16;
                #% ---- From LIGO-P0900199, analytical fit with sub percent precision
                #%      across the 50-8000 Hz band (GEO-HF logbook p.60)
                #%      https://intranet.aei.uni-hannover.de/geo600/geohflogbook.nsf/7e8722dffa24dea0c1256de900406c84/4837a612ac990060c12575ce004e70fd?OpenDocument
                #%      fstop not well defined, there is not design for the control
                #%      noise which currently dominates the spectrum below 100Hz

    elif IFO.upper() in ['KAGRA','KAGRA_EARLY','KAGRA_MID','KAGRA_LATE']:
        [S, fstop] = kagraspectrum(IFO,f)
    elif IFO.upper() in ['LIGO3', 'LIGO3-DEFAULT', 'LIGO3-MID', 'LIGO3-HIGH']:
        [S, fstop] = ligo3spectrumold(IFO,f)
    elif IFO.upper() in ['LIGO-A', 'LIGO-VOYAGER', 'LIGO-COSMICEXPLORER']:
        [S, fstop] = ligo3spectrum(IFO,f)
    elif IFO.upper() in ['LIGO', 'LHO', 'LLO']:
        fstop = [40,float('inf')]
        f0 = 150
        x = f/f0
        S = 9*(1e-46)*( (4.49*x)**(-56) + 0.16*x**(-4.52) + 0.52 + 0.32*x**2 )
                #% ---- From L. P. Grishchuk, V. M. Lipunov, K. A. Postnov, M. E.
                #%      Prokhorov, and B. S. Sathyaprakash, Physics-Uspekhi 71, 3
                #%      (2000). 
    elif IFO.upper()=='LIGO-LV':
        fstop = [40,float('inf')]
        seismicReference = 3.3e-21 * (30/f)**14
        thermalReference = 3.8e-23 * (100/f)**2
        shotReference = 1.13e-23 * np.sqrt(1 + (f / 90)**2)
        S = shotReference**2 + thermalReference**2 + seismicReference**2
                #%----- Design used by Chatterji in generating simulated data for 
                #%      the LIGO-Virgo project Ia,b.  This design does not include
                #%      the lines that were added to the simulated data.  fstop is 
                #%      guessed.
    elif IFO.upper()=='LISA':
                #% % ---- From L. P. Grishchuk, V. M. Lipunov, K. A. Postnov, M. E.
                #% %      Prokhorov, and B. S. Sathyaprakash, Physics-Uspekhi 71, 3
                #% %      (2000).  
                #% fstop = [1e-5,Inf];
                #% f0 = 1e-3;
                #% x = f/f0;
                #% S = 420*(1e-46)*( (x/5.62).^(-14/3) + 1e3 + x.^2 );
                #% ---- Sky-averaged instrumental noise for LISA, from 
                #%      1001.3099v2.  Code by Ioannis Kamaretsos.
        L = 5*10**9        #%-- arm length of LISA [m]
        Sp = 4*10**-22      #%-- noise scale constant [m^2 Hz^-1]
        Sa = 9*10**-30      #%-- noise scale constant [m^2 Hz^-1]
        c = 299792458      #%-- speed of light [m/s]
        fo = c/(2*pi*L)    #%-- LISA arm transfer frequency [Hz]
            #% ---- Instrumental noise.
        S = (L**-2)*((1+0.5*(f**2/fo**2))*Sp + (1+10**-8*(f**-2))*(4*Sa*((2*pi)**-4*f**-4)))
        fstop = [1e-5,1e0]  #%-- made up

    elif IFO.upper()=='LISA-LIGHT1':
                #% ---- CLISA1_P005_LPF configuration from
                #%      https://lisa-light.aei.mpg.de/lisa-light/pub/DetectorConfigurations/NewConfigurations/NewLISAConfig_v1.pdf 
                #%      This does not include the WD-WD confusion background.
        c = 299792458     #%-- speed of light [m/s]
                #% ---- Acceleration noise [m^2/Hz].
        Sacc = (28.99e-30*(1 + 1.8e-4/f)**2)*(2*pi*f)**-4
                #% ---- Shot noise [m^2/Hz].
        Ssn = 2.25e-21
                #% ---- "Other measurement noise" [m^2/Hz].
        Somn = 6.4e-23
                #% ---- Arm length [m].
        L = 1e9
            #% ---- "Very approximate" sensitivity.
        S = 20/3 * ( 1 + (2*L*f/(0.41*c))**2 ) * (Sacc + Ssn + Somn) / L**2        
        fstop = [1e-5,1e0]  #%-- made up
    elif IFO.upper()=='LISA-LIGHT2':
                #% ---- 10LISA1_P2_DRS configuration from
                #%      https://lisa-light.aei.mpg.de/lisa-light/pub/DetectorConfigurations/NewConfigurations/NewLISAConfig_v1.pdf 
                #%      This does not include the WD-WD confusion background.
        c = 299792458     #%-- speed of light [m/s]
                #% ---- Acceleration noise [m^2/Hz].
        Sacc = 21.30e-30/(2*pi*f)**4
                #% ---- Shot noise [m^2/Hz].
        Ssn = 1.4e-24
                #% ---- "Other measurement noise" [m^2/Hz].
        Somn = 6.4e-23
                #% ---- Arm length [m].
        L = 1e9
                #% ---- "Very approximate" sensitivity.
        S = 20/3 * ( 1 + (2*L*f/(0.41*c))**2 ) * (Sacc + Ssn + Somn) / L**2        
                #% S = 20/3 * ( 1 + (2*L*f/(0.41*c)).^2 ) .* (4*Sacc + Ssn + Somn) / L^2;        
        fstop = [1e-5,1e0]  #%-- made up
    # elif IFO.upper()=='LISA-WD':
    #                 #% ---- Sky-averaged instrumental noise for LISA, including WD-WD
    #                 #%      confusion background.  From 1001.3099v2.  Code by Ioannis
    #                 #%      Kamaretsos. 
    #     L = 5*10**9        #%-- arm length of LISA [m]
    #     Sp = 4*10**-22     #%-- noise scale constant [m^2 Hz^-1]
    #     Sa = 9*10**-30     #%-- noise scale constant [m^2 Hz^3]
    #     c = 299792458      #%-- speed of light [m/s]
    #     fo = c/(2*pi*L)    #%-- LISA arm transfer frequency [Hz]
    #             #% ---- Instrumental noise.
    #     S = (L**-2)*((1+0.5*(f**2/fo**2))*Sp + (1+10**-8*(f**-2))*(4*Sa*((2*pi)**-4*f**-4)));
    #             #% ---- WD-WD confusion noise.  Note typo in f<1e-3 term has been
    #             #%      corrected: f^(-2/3) -> f^(-2.3).
    #     S_conf = zeros(size(f));
    #     ind = find(f <= 1e-3);
    #     S_conf(ind) = (10^(-44.62))*f(ind)**(-2.3);

    #     ind = find(f > 1e-3 & f <= 10^(-2.7));
    #     S_conf(ind) = (10^(-50.92))*f(ind)**(-4.4);

    #     ind = find(f > 10^(-2.7) & f <= 10^(-2.4));
    #     S_conf(ind) = (10^(-62.8))*f(ind).^(-8.8);

    #     ind = find(f > 10^(-2.4) & f <= 1e-2);
    #     S_conf(ind) = (10^(-89.68))*f(ind).^(-20);

    #     S = S + S_conf;
    #         fstop = [1e-5,1e0];  %-- made up
    elif IFO.upper()=='TAMA':
            #% ---- From L. P. Grishchuk, V. M. Lipunov, K. A. Postnov, M. E.
            #%      Prokhorov, and B. S. Sathyaprakash, Physics-Uspekhi 71, 3
            #%      (2000).  
        fstop = [75,float('inf')]
        f0 = 400
        x = f/f0
        S = 75*(1e-46)*( x**(-5) + 13*x**(-1) + 9 + 9*x**2 )
    elif IFO.upper()=='VIRGO':
            #% ---- From L. P. Grishchuk, V. M. Lipunov, K. A. Postnov, M. E.
            #%      Prokhorov, and B. S. Sathyaprakash, Physics-Uspekhi 71, 3
            #%      (2000).  
        fstop = [20,float('inf')];
        f0 = 500
        x = f/f0
        S = 3.24*(1e-46)*( (6.23*x)**(-5) + 2*x**(-1) + 1 + x**2 )
    elif IFO.upper()=='VIRGO-LV':
            #%----- Design used in generating simulated data for the LIGO-Virgo 
            #%      project Ia,b.  This design does not include the lines that
            #%      were added to the simulated data.  Note that f0 is reset
            #%      from 500 in 'VIRGO' to 300 in 'VIRGO-LV'. 
        fstop = [20,float('inf')]
        f0 = 500*0.6
        x = f/f0
        S = 3.24*(1e-46)*( (6.23*x)**(-5) + 2*x**(-1) + 2 + 1.2*x**2 )
    elif IFO.upper()=='WHITE':
        fstop = [f[0],f[-1]]
        S = 2 * np.ones(len(f))
    #     otherwise
    #         % ---- Check for one of the various aLIGO or aVirgo configurations,
    #         %      or white noise. 
    #         if (length(IFO)==12) && (strcmpi(IFO,'ALIGO-HOOPER'))
    #             % ---- Analytic approximation from Hooper et al., 1108.3186 / Phys
    #             %      Rev D.86.024012 (SPIIR paper). WARNING: This does not match
    #             %      any of the official aLIGO spectra.
    #             x = f/215;
    #             S = 1e-49 * (x.^(-4.14) - 5*x.^(-2) + 111*(1 - x.^2 + 0.5*x.^4)./(1+0.5*x.^2)); 
    #             fstop = [9 Inf];        
    #         elseif (length(IFO)>6) && (strcmpi(IFO(1:6),'ALIGO-'))
    #             [S, fstop] = aligospectrum(IFO,f);
    #         elseif (length(IFO)>11) && (strcmpi(IFO(1:11),'ALIGOEARLY-'))
    #             [S, fstop] = aligoearlyspectrum(IFO,f);
    #         elseif (length(IFO)>7) && (strcmpi(IFO(1:7),'AVIRGO-'))
    #             [S, fstop] = avirgospectrum(IFO,f);
    #         elseif (length(IFO)>=5) && (strcmpi(IFO(1:5),'WHITE'))
    #             fs = str2num(IFO(6:end));
    #             fstop = [f(1),f(end)];
    #             S = 2 * ones(size(f)) / fs;
    #         else
    #             error(['Detector ' IFO ' not recognized.']);
    #         end
    elif IFO.upper() == 'LIGO2015':
        [S, fstop] = ligo2015()
    elif IFO.upper() == 'LIGO2016':
        [S, fstop] = ligo2016()
    elif IFO.upper() == 'VIRGO2016':
        [S, fstop] = virgo2016()
    #% ---- Display warning message if asking for spectrum at frequencies below
    #%      or above the recommended cut-offs.
    if min(f)<fstop[0]:
        pass
        #print('Warning: Requested frequency range extends below minimum frequency of '+str(fstop[0])+' Hz for SRD fit for '+IFO+' detector.')

    if max(f)>fstop[1]:
        pass
        #print('Warning: Requested frequency range extends above maximum frequency of '+str(fstop[1])+' Hz for SRD fit for '+IFO+' detector.')



    #% ---- Done.
    return S, fstop

'''

# % -------------------------------------------------------------------------
# %                              ET-B SRD
# % -------------------------------------------------------------------------
# function[sqrtS, fstop] = ETB_fit(f)
# % ETB_fit - simple fit to the "ET-B" sensitivity curve of arXiv:0810.0604v2 
# % 
# % [sqrtS, fstop] = ETB_fit(f)
# %
# %   f       Vector of sampled frequencies (Hz).
# %
# %   S       Vector of the same shape as f, containing amplitude spectral 
# %           densities of ET_B (1/sqrt(Hz)).
# %   fstop   Minimum frequency (Hz) at which fit is "guaranteed" to be
# %           accurate.
# %
# % ETB_fit provides a simple fit to the "ET-B" design noise spectrum curve
# % described in arXiv:0810.0604v2.  For data see 
# %   https://workarea.et-gw.eu/et/WG3-Topology 
# % The following analytic approximation is good to within 10% across the
# % [1,10000] Hz band.  Compared to the sampled spectrum, the analytic
# % approximation Sanal has the properties 
# %        min((Sanal/S).^0.5)  = 0.90
# %        mean((Sanal/S).^0.5) = 0.95
# %        min((Sanal/S).^0.5)  = 1.03
# %
# % Created by Thomas Dent (thomas.dent@astro.cf.ac.uk) on 14/05/09

xt = f/100.
fit = 2.39*10**(-27)*xt**(-15.64) +0.349*xt**(-2.145) +1.76*xt**(-0.12) +0.409*xt**(1.10)
sqrtS = 10**(-25)*fit
fstop = [1,float('inf')]  #%-- chosen arbitrarily -- looks okay.


# % -------------------------------------------------------------------------
# %                              ET-C SRD
# % -------------------------------------------------------------------------

def ETC(f):   #function [ET_C, fstop] = ETC_fit(f)
# % ETC - SRD noise curve for the ET detector, version "C".
# %
# % This function produces for a given frequency vector the corresponding
# % amplitude spectral density of the ET-C sensitivity curve.  It is a fit to
# % the ET-C spectrum found in S. Hild et al., CQG 27 015003 (2010), with fit
# % written by S. Hild. The fit is only valid for frequencies between 1 Hz
# % and 10kHz and the the accuracy is to within 1 percent. Please also note
# % that the fit is not given by a single function, but is split into several
# % individual frequency bands.
# %
# % usage:
# %
# %   [S, fstop] = ETC_fit(f)
# %
# %   f       Vector of sampled frequencies (Hz).
# %
# %   S       Vector of the same shape as f, containing amplitude spectral 
# %           densities of ET_C (1/sqrt(Hz)).
# %   fstop   Minimum and maximum frequency (Hz) at which fit is "guaranteed"
# %           to be accurate. 
# %
# % A detailed description of ET-C can be found in:
# % S. Hild et al.: 'A xylophone configuration for a third-generation 
# % gravitational wave detector', CQG, 2010, 27, 015003.
# %
# % Stefan Hild, 4/1/2010 (stefan.hild@ligo.org)

#% ---- Check range of f.
fstop = [1,1e4];
# % if min(f)<1
# %     error('Minimum frequency 1 Hz.');
# % end
# % if max(f)>1e4
# %     warning('Maximum frequency 1e4 Hz.');
# % end

# % ---- Force f to be a row vector.
####isColumnVector = 0;
####if size(f,1)>size(f,2)
####    isColumnVector = 1;
####    f = f.';
####end

#% ---- Append extra element 1 Hz above highest f value.
f = list(f)
f.append(f[-1]+1)
f=np.array(f)

ET_C = f;    #% initializing vector for ET_C

# % -------------------------------------------------------------------------
# %                  Piecewise computation of ET_C SRD 
# % -------------------------------------------------------------------------

# % ---- Below here the functions for the individual frequency bands are given.

f1 = 4600        #% This is the frequency range from 4.6 to 10 kHz.
[y1-1, x1-1] = np.min(np.abs(f-f1))
ET_C(x1:len(f)) = 6.405e-28.*f(x1:length(f)).^0.9969-5.092e-27;

f2 = 1900;        % This is the frequency range from 1.9 to 4.6 kHz.
[y2, x2] = min(abs(f-f2));
ET_C(x2:x1-1) = 7.458e-28.*f(x2:x1-1).^0.9794-1.793e-26;

f3 = 210;
[y3, x3] = min(abs(f-f3));
ET_C(x3:x2-1) = -7.646e-55.*f(x3:x2-1).^9 +  1.54e-50.*f(x3:x2-1).^8 -1.331e-46.*f(x3:x2-1).^7 + 6.458e-43.*f(x3:x2-1).^6 -1.936e-39.*f(x3:x2-1).^5 +  3.719e-36.*f(x3:x2-1).^4 -4.606e-33.*f(x3:x2-1).^3 + 3.614e-30.*f(x3:x2-1).^2 -1.095e-27.*f(x3:x2-1) + 4.397e-25;

f4 = 110;
[y4, x4] = min(abs(f-f4));
ET_C(x4:x3-1)=-3.317e-32.*f(x4:x3-1).^3+2.027e-29.*f(x4:x3-1).^2+-4.354e-27.*f(x4:x3-1)+ 6.601e-25;

f5 = 80;
[y5, x5] = min(abs(f-f5));
ET_C(x5:x4-1) =   -2.935e-46.*f(x5:x4-1).^9 +  6.757e-43.*f(x5:x4-1).^8 -6.794e-40.*f(x5:x4-1).^7 + 3.916e-37.*f(x5:x4-1).^6 -1.426e-34.*f(x5:x4-1).^5 +   3.414e-32.*f(x5:x4-1).^4-5.399e-30.*f(x5:x4-1).^3 + 5.52e-28.*f(x5:x4-1).^2 -3.433e-26.*f(x5:x4-1) + 1.389e-24;

f6 = 51;
[y6, x6] = min(abs(f-f6));
ET_C(x6:x5-1) =  -2.998e-30.*f(x6:x5-1).^3 +7.391e-28.*f(x6:x5-1).^2 -6.303e-26.*f(x6:x5-1) +2.279e-24;

f7 = 38;
[y7, x7] = min(abs(f-f7));
ET_C(x7:x6-1) =  1.27e-29.*f(x7:x6-1).^3 -1.118e-27.*f(x7:x6-1).^2 +2.716e-27.*f(x7:x6-1) +1.674e-24;

f8 = 29.8;
[y8, x8] = min(abs(f-f8));
ET_C(x8:x7-1) =  3.151e-28.*f(x8:x7-1).^3  -3.676e-26.*f(x8:x7-1).^2 + 1.4e-24.*f(x8:x7-1) -1.655e-23;

f9 = 22;
[y9, x9] = min(abs(f-f9));
ET_C(x9:x8-1) =  -3.173e-28.*f(x9:x8-1).^3  +   3.028e-26.*f(x9:x8-1).^2 -9.138e-25.*f(x9:x8-1) + 9.6e-24;

f10 = 18.5;
[y10, x10] = min(abs(f-f10));
ET_C(x10:x9-1) =  -5.873e-29.*f(x10:x9-1).^4+ 5.781e-27.*f(x10:x9-1).^3   -2.06e-25.*f(x10:x9-1).^2 +  3.134e-24.*f(x10:x9-1)  -1.627e-23;

f11 = 9;
[y11, x11] = min(abs(f-f11));
ET_C(x11:x10-1) =  5.104e-28.*f(x11:x10-1).^3   -2.817e-26.*f(x11:x10-1).^2 +  4.761e-25.*f(x11:x10-1)  -1.466e-24;

f12 = 7.5;
[y12, x12] = min(abs(f-f12));
ET_C(x12:x11-1) =  -1.368e-26.*f(x12:x11-1).^3  + 3.581e-25.*f(x12:x11-1).^2  -3.031e-24.*f(x12:x11-1) +  9.151e-24;

f13 = 5.7;
[y13, x13] = min(abs(f-f13));
ET_C(x13:x12-1) =   -6.333e-26.*f(x13:x12-1).^3  + 1.509e-24.*f(x13:x12-1).^2 -1.193e-23.*f(x13:x12-1) +  3.210e-23;

f14 = 4.0;
[y14, x14] = min(abs(f-f14));
ET_C(x14:x13-1) =   1.362e-21.*f(x14:x13-1).^(-3.952)  -3.004e-27;

f15 = 3.0;
[y15, x15] = min(abs(f-f15));
ET_C(x15:x14-1) =  1.135e-21.*f(x15:x14-1).^( -3.806)   -1.16e-25;

f16 = 2.4;
[y16, x16] = min(abs(f-f16));
ET_C(x16:x15-1) =  8.52e-22.*f(x16:x15-1).^(-3.53)   -4.264e-25;

f17 = 2;
[y17, x17] = min(abs(f-f17));
ET_C(x17:x16-1) =   6.899e-22.*f(x17:x16-1).^(-3.276)   -9.181e-25;

f18 = 1.65;
[y18, x18] = min(abs(f-f18));
ET_C(x18:x17-1) =   7.056e-22.*f(x18:x17-1).^(-3.521)  + 9.042e-24;

f19 = 1.4;
[y19, x19] = min(abs(f-f19));
ET_C(x19:x18-1) = 2.717e-21.*f(x19:x18-1).^4 -1.842e-20.*f(x19:x18-1).^3  +  4.709e-20.*f(x19:x18-1).^2 -5.402e-20.*f(x19:x18-1) +2.3667e-20;

f20 = 1.26;
[y20, x20] = min(abs(f-f20));
ET_C(x20:x19-1) =  -1.98e-20.*f(x20:x19-1).^3  +  8.486e-20.*f(x20:x19-1).^2  -1.219e-19.*f(x20:x19-1) + 5.8894e-20;
  
f21 = 1.12;
[y21, x21] = min(abs(f-f21));
ET_C(x21:x20-1) =  -1.042e-19.*f(x21:x20-1).^3  +  4.045e-19.*f(x21:x20-1).^2  -5.257e-19.*f(x21:x20-1) +  2.29051e-19;

f22 = 1;
[y22, x22] = min(abs(f-f22));
ET_C(x22:x21-1) =  4.741e-21.*f(x22:x21-1).^(-11.72)  + 1.906e-23;

ET_C = ET_C(1:length(ET_C)-1);

% -------------------------------------------------------------------------

% ---- Return spectrum in the format (row/column vector) matching input.
if isColumnVector
    ET_C = ET_C.';
end
 
return
'''

# % -------------------------------------------------------------------------
# %                              aLIGO spectra
# % -------------------------------------------------------------------------

def aligoearlyspectrum(IFO,f):
    # % ALIGOEARLYSPECTRUM - Possible noise curves for early aLIGO.
    # %
    # % Arguments are the same as for the parent function.
    # %
    # % The aLIGO spectra are taken from LIGO Document T1200307_v3, 
    # % https://dcc.ligo.org/cgi-bin/private/DocDB/ShowDocument?docid=93068

    # % ---- Recognize detector and compute noise spectrum.
    if IFO.upper=='ALIGOEARLY-50':
        col = 1
    elif IFO.upper=='ALIGOEARLY-100':
        col = 2
    elif IFO.upper=='ALIGOEARLY-150':
        col = 3
    else:
        raise TypeError('Detector '+IFO+' not recognized.')


    #% ---- Extract pre-sampled noise spectrum from data.  First column is
    #%      frequency, other columns map to configurations as above.


    data = np.array([[9.0, 2.0315133e-21, 1.7564389e-21, 1.7360779e-21], [9.0203998, 2.4115253e-21, 2.1869794e-21, 2.1708083e-21], [9.0408458, 3.09785e-21, 2.9280383e-21, 2.9160887e-21], [9.0613382, 4.6064717e-21, 4.4950573e-21, 4.4873528e-21], [9.081877, 1.0007255e-20, 9.9569194e-21, 9.9534749e-21], [9.1024624, 4.1375496e-20, 4.136346e-20, 4.1362638e-20], [9.1230944, 6.7000822e-21, 6.6260274e-21, 6.6209432e-21], [9.1437733, 3.7140344e-21, 3.5799561e-21, 3.5706222e-21], [9.1644989, 2.6361497e-21, 2.4454318e-21, 2.4318707e-21], [9.1852716, 2.097615e-21, 1.8546095e-21, 1.8368537e-21], [9.2060913, 1.7837581e-21, 1.4933317e-21, 1.4714229e-21], [9.2269583, 1.5832162e-21, 1.2503601e-21, 1.2243496e-21], [9.2478725, 1.4467633e-21, 1.0763148e-21, 1.0462626e-21], [9.2688341, 1.3494487e-21, 9.4592165e-22, 9.1189619e-22], [9.2898433, 1.2774003e-21, 8.4490099e-22, 8.0697861e-22], [9.3109, 1.222372e-21, 7.6457475e-22, 7.2283892e-22], [9.3320045, 1.1792069e-21, 6.9936437e-22, 6.5390499e-22], [9.3531568, 1.1445471e-21, 6.4552008e-22, 5.9643274e-22], [9.3743571, 1.1161329e-21, 6.0042814e-22, 5.4781329e-22], [9.3956054, 1.0924006e-21, 5.6221082e-22, 5.0617302e-22], [9.4169019, 1.0722422e-21, 5.2948451e-22, 4.7013159e-22], [9.4382467, 1.0548558e-21, 5.0120728e-22, 4.3864963e-22], [9.4596398, 1.0396507e-21, 4.7657977e-22, 4.1092956e-22], [9.4810815, 1.0261843e-21, 4.5497866e-22, 3.8634918e-22], [9.5025717, 1.0141201e-21, 4.3591091e-22, 3.6441594e-22], [9.5241107, 1.0031991e-21, 4.1898162e-22, 3.4473484e-22], [9.5456984, 9.9321909e-22, 4.0387096e-22, 3.2698543e-22], [9.5673351, 9.840209e-22, 3.9031731e-22, 3.1090499e-22], [9.5890209, 9.7547772e-22, 3.7810474e-22, 2.9627613e-22], [9.6107557, 9.6748765e-22, 3.6705367e-22, 2.8291738e-22], [9.6325399, 9.5996809e-22, 3.5701355e-22, 2.7067609e-22], [9.6543734, 9.5285155e-22, 3.4785743e-22, 2.594229e-22], [9.6762565, 9.4608248e-22, 3.3947757e-22, 2.4904746e-22], [9.6981891, 9.3961485e-22, 3.3178204e-22, 2.3945509e-22], [9.7201714, 9.3341028e-22, 3.2469202e-22, 2.3056411e-22], [9.7422036, 9.2743655e-22, 3.1813961e-22, 2.2230363e-22], [9.7642857, 9.2166648e-22, 3.1206603e-22, 2.1461184e-22], [9.7864179, 9.1607701e-22, 3.0642024e-22, 2.0743464e-22], [9.8086002, 9.1064848e-22, 3.011577e-22, 2.0072442e-22], [9.8308328, 9.0536404e-22, 2.9623948e-22, 1.9443917e-22], [9.8531158, 9.0020922e-22, 2.9163138e-22, 1.8854165e-22], [9.8754493, 8.951715e-22, 2.8730329e-22, 1.8299875e-22], [9.8978334, 8.9024001e-22, 2.8322861e-22, 1.7778092e-22], [9.9202683, 8.854053e-22, 2.7938378e-22, 1.7286171e-22], [9.942754, 8.8065911e-22, 2.7574789e-22, 1.6821737e-22], [9.9652907, 8.7599418e-22, 2.7230231e-22, 1.6382652e-22], [9.9878784, 8.7140414e-22, 2.6903044e-22, 1.5966988e-22], [10.010517, 8.6688336e-22, 2.6591743e-22, 1.5573003e-22], [10.033208, 8.6242686e-22, 2.6294998e-22, 1.5199117e-22], [10.055949, 8.5803021e-22, 2.6011614e-22, 1.4843892e-22], [10.078743, 8.5368948e-22, 2.5740515e-22, 1.450602e-22], [10.101588, 8.4940114e-22, 2.5480731e-22, 1.4184304e-22], [10.124484, 8.4516204e-22, 2.5231385e-22, 1.387765e-22], [10.147433, 8.4096935e-22, 2.4991684e-22, 1.3585058e-22], [10.170433, 8.3682052e-22, 2.476091e-22, 1.330561e-22], [10.193486, 8.3271326e-22, 2.453841e-22, 1.303846e-22], [10.216591, 8.286455e-22, 2.4323591e-22, 1.2782832e-22], [10.239749, 8.2461536e-22, 2.4115912e-22, 1.2538011e-22], [10.262958, 8.2062112e-22, 2.3914882e-22, 1.2303333e-22], [10.286221, 8.1666124e-22, 2.372005e-22, 1.2078187e-22], [10.309536, 8.127343e-22, 2.3531004e-22, 1.1862008e-22], [10.332904, 8.0883901e-22, 2.3347367e-22, 1.1654271e-22], [10.356325, 8.0497418e-22, 2.3168795e-22, 1.145449e-22], [10.379799, 8.0113874e-22, 2.2994969e-22, 1.1262213e-22], [10.403327, 7.9733167e-22, 2.2825598e-22, 1.1077018e-22], [10.426907, 7.9355207e-22, 2.2660413e-22, 1.0898516e-22], [10.450541, 7.8979908e-22, 2.2499167e-22, 1.072634e-22], [10.474229, 7.8607193e-22, 2.2341631e-22, 1.0560152e-22], [10.49797, 7.823699e-22, 2.2187595e-22, 1.0399634e-22], [10.521765, 7.786923e-22, 2.2036864e-22, 1.0244489e-22], [10.545615, 7.7503852e-22, 2.1889258e-22, 1.009444e-22], [10.569518, 7.7140798e-22, 2.1744609e-22, 9.9492265e-23], [10.593475, 7.6780014e-22, 2.1602762e-22, 9.8086061e-23], [10.617487, 7.642145e-22, 2.1463573e-22, 9.6723508e-23], [10.641553, 7.6065058e-22, 2.132691e-22, 9.5402468e-23], [10.665673, 7.5710794e-22, 2.1192647e-22, 9.4120931e-23], [10.689849, 7.5358617e-22, 2.1060668e-22, 9.2877013e-23], [10.714079, 7.5008489e-22, 2.0930866e-22, 9.1668939e-23], [10.738364, 7.4660373e-22, 2.0803141e-22, 9.0495044e-23], [10.762704, 7.4314234e-22, 2.0677397e-22, 8.9353759e-23], [10.787099, 7.397004e-22, 2.0553549e-22, 8.8243604e-23], [10.811549, 7.362776e-22, 2.0431513e-22, 8.7163188e-23], [10.836055, 7.3287366e-22, 2.0311213e-22, 8.6111195e-23], [10.860617, 7.294883e-22, 2.0192577e-22, 8.5086388e-23], [10.885234, 7.2612126e-22, 2.0075537e-22, 8.4087598e-23], [10.909907, 7.2277229e-22, 1.9960031e-22, 8.311372e-23], [10.934636, 7.1944115e-22, 1.9846e-22, 8.2163709e-23], [10.959421, 7.1612763e-22, 1.9733386e-22, 8.123658e-23], [10.984262, 7.1283151e-22, 1.9622139e-22, 8.0331397e-23], [11.009159, 7.0955258e-22, 1.9512209e-22, 7.9447278e-23], [11.034113, 7.0629066e-22, 1.9403549e-22, 7.8583388e-23], [11.059124, 7.0304554e-22, 1.9296116e-22, 7.7738933e-23], [11.084191, 6.9981706e-22, 1.9189868e-22, 7.6913161e-23], [11.109315, 6.9660504e-22, 1.9084767e-22, 7.6105362e-23], [11.134495, 6.9340931e-22, 1.8980776e-22, 7.5314858e-23], [11.159733, 6.9022972e-22, 1.887786e-22, 7.454101e-23], [11.185028, 6.8706612e-22, 1.8775986e-22, 7.3783208e-23], [11.210381, 6.8391835e-22, 1.8675124e-22, 7.3040875e-23], [11.235791, 6.8078627e-22, 1.8575244e-22, 7.2313459e-23], [11.261258, 6.7766975e-22, 1.8476317e-22, 7.1600438e-23], [11.286784, 6.7456864e-22, 1.8378318e-22, 7.0901315e-23], [11.312367, 6.7148283e-22, 1.8281222e-22, 7.0215617e-23], [11.338008, 6.6841219e-22, 1.8185004e-22, 6.9542893e-23], [11.363707, 6.6535659e-22, 1.8089643e-22, 6.8882715e-23], [11.389465, 6.6231592e-22, 1.7995116e-22, 6.8234672e-23], [11.41528, 6.5929006e-22, 1.7901403e-22, 6.7598374e-23], [11.441155, 6.562789e-22, 1.7808485e-22, 6.6973451e-23], [11.467088, 6.5328234e-22, 1.7716343e-22, 6.6359547e-23], [11.49308, 6.5030025e-22, 1.762496e-22, 6.5756323e-23], [11.51913, 6.4733255e-22, 1.7534318e-22, 6.5163455e-23], [11.54524, 6.4437914e-22, 1.7444401e-22, 6.4580633e-23], [11.571409, 6.414399e-22, 1.7355194e-22, 6.4007563e-23], [11.597637, 6.3851475e-22, 1.7266683e-22, 6.344396e-23], [11.623925, 6.3560359e-22, 1.7178853e-22, 6.2889556e-23], [11.650272, 6.3270633e-22, 1.709169e-22, 6.2344089e-23], [11.676679, 6.2982288e-22, 1.7005182e-22, 6.1807312e-23], [11.703146, 6.2695314e-22, 1.6919317e-22, 6.1278987e-23], [11.729673, 6.2409704e-22, 1.6834082e-22, 6.0758886e-23], [11.75626, 6.2125449e-22, 1.6749466e-22, 6.0246791e-23], [11.782907, 6.184254e-22, 1.6665459e-22, 5.974249e-23], [11.809615, 6.1560969e-22, 1.6582049e-22, 5.9245784e-23], [11.836383, 6.1280728e-22, 1.6499226e-22, 5.8756477e-23], [11.863212, 6.1001809e-22, 1.6416982e-22, 5.8274384e-23], [11.890102, 6.0724204e-22, 1.6335306e-22, 5.7799326e-23], [11.917052, 6.0447905e-22, 1.6254189e-22, 5.7331131e-23], [11.944064, 6.0172905e-22, 1.6173623e-22, 5.6869634e-23], [11.971137, 5.9899196e-22, 1.60936e-22, 5.6414674e-23], [11.998271, 5.9626771e-22, 1.6014111e-22, 5.5966098e-23], [12.025467, 5.9355623e-22, 1.5935148e-22, 5.5523759e-23], [12.052725, 5.9085744e-22, 1.5856704e-22, 5.5087513e-23], [12.080044, 5.8817127e-22, 1.5778771e-22, 5.4657223e-23], [12.107425, 5.8549765e-22, 1.5701343e-22, 5.4232755e-23], [12.134868, 5.8283651e-22, 1.5624413e-22, 5.3813982e-23], [12.162374, 5.8018779e-22, 1.5547973e-22, 5.3400779e-23], [12.189941, 5.7755141e-22, 1.5472019e-22, 5.2993026e-23], [12.217572, 5.7492731e-22, 1.5396542e-22, 5.2590608e-23], [12.245264, 5.7231542e-22, 1.5321538e-22, 5.2193413e-23], [12.27302, 5.6971568e-22, 1.5247e-22, 5.1801331e-23], [12.300839, 5.6712803e-22, 1.5172922e-22, 5.1414258e-23], [12.32872, 5.6455239e-22, 1.50993e-22, 5.1032093e-23], [12.356665, 5.6198871e-22, 1.5026127e-22, 5.0654736e-23], [12.384673, 5.5943691e-22, 1.4953399e-22, 5.0282092e-23], [12.412745, 5.5689695e-22, 1.4881111e-22, 4.9914067e-23], [12.44088, 5.5436876e-22, 1.4809256e-22, 4.9550573e-23], [12.469079, 5.5185227e-22, 1.4737832e-22, 4.9191522e-23], [12.497342, 5.4934743e-22, 1.4666832e-22, 4.8836829e-23], [12.525669, 5.4685417e-22, 1.4596252e-22, 4.8486411e-23], [12.55406, 5.4437244e-22, 1.4526088e-22, 4.8140188e-23], [12.582516, 5.4190218e-22, 1.4456335e-22, 4.7798083e-23], [12.611036, 5.3944333e-22, 1.438699e-22, 4.7460019e-23], [12.639621, 5.3699583e-22, 1.4318047e-22, 4.7125923e-23], [12.66827, 5.3455962e-22, 1.4249503e-22, 4.6795724e-23], [12.696985, 5.3213465e-22, 1.4181354e-22, 4.6469351e-23], [12.725764, 5.2972086e-22, 1.4113596e-22, 4.6146736e-23], [12.754609, 5.273182e-22, 1.4046225e-22, 4.5827813e-23], [12.783519, 5.249266e-22, 1.3979237e-22, 4.5512518e-23], [12.812495, 5.2254602e-22, 1.3912629e-22, 4.5200787e-23], [12.841536, 5.2017639e-22, 1.3846397e-22, 4.4892559e-23], [12.870644, 5.1781767e-22, 1.3780538e-22, 4.4587775e-23], [12.899817, 5.154698e-22, 1.3715047e-22, 4.4286375e-23], [12.929056, 5.1313272e-22, 1.3649923e-22, 4.3988302e-23], [12.958362, 5.1080639e-22, 1.3585161e-22, 4.3693502e-23], [12.987734, 5.0849075e-22, 1.3520758e-22, 4.3401918e-23], [13.017172, 5.0618574e-22, 1.3456711e-22, 4.31135e-23], [13.046677, 5.0389132e-22, 1.3393017e-22, 4.2828193e-23], [13.07625, 5.0160744e-22, 1.3329673e-22, 4.2545948e-23], [13.105889, 4.9933403e-22, 1.3266676e-22, 4.2266715e-23], [13.135595, 4.9707106e-22, 1.3204023e-22, 4.1990446e-23], [13.165369, 4.9481846e-22, 1.3141711e-22, 4.1717093e-23], [13.19521, 4.925762e-22, 1.3079737e-22, 4.144661e-23], [13.225119, 4.9034422e-22, 1.3018098e-22, 4.1178951e-23], [13.255096, 4.8812246e-22, 1.2956791e-22, 4.0914072e-23], [13.28514, 4.8591088e-22, 1.2895815e-22, 4.0651929e-23], [13.315253, 4.8370944e-22, 1.2835166e-22, 4.0392481e-23], [13.345434, 4.8151808e-22, 1.2774841e-22, 4.0135685e-23], [13.375683, 4.7933675e-22, 1.2714838e-22, 3.98815e-23], [13.406001, 4.7716541e-22, 1.2655155e-22, 3.9629887e-23], [13.436388, 4.75004e-22, 1.2595788e-22, 3.9380807e-23], [13.466843, 4.7285248e-22, 1.2536736e-22, 3.9134221e-23], [13.497368, 4.7071081e-22, 1.2477996e-22, 3.8890091e-23], [13.527961, 4.6857893e-22, 1.2419566e-22, 3.8648382e-23], [13.558624, 4.664568e-22, 1.2361443e-22, 3.8409056e-23], [13.589357, 4.6434437e-22, 1.2303625e-22, 3.8172079e-23], [13.620159, 4.6224159e-22, 1.224611e-22, 3.7937416e-23], [13.651031, 4.6014843e-22, 1.2188895e-22, 3.7705033e-23], [13.681973, 4.5806483e-22, 1.2131979e-22, 3.7474896e-23], [13.712985, 4.5599074e-22, 1.2075358e-22, 3.7246974e-23], [13.744068, 4.5392613e-22, 1.2019031e-22, 3.7021234e-23], [13.775221, 4.5187095e-22, 1.1962996e-22, 3.6797644e-23], [13.806444, 4.4982515e-22, 1.1907251e-22, 3.6576174e-23], [13.837739, 4.4778869e-22, 1.1851793e-22, 3.6356794e-23], [13.869104, 4.4576153e-22, 1.1796621e-22, 3.6139475e-23], [13.90054, 4.4374361e-22, 1.1741733e-22, 3.5924186e-23], [13.932048, 4.417349e-22, 1.1687125e-22, 3.5710899e-23], [13.963627, 4.3973536e-22, 1.1632798e-22, 3.5499588e-23], [13.995277, 4.3774494e-22, 1.1578748e-22, 3.5290223e-23], [14.027, 4.3576359e-22, 1.1524974e-22, 3.5082779e-23], [14.058794, 4.3379128e-22, 1.1471474e-22, 3.4877228e-23], [14.09066, 4.3182796e-22, 1.1418246e-22, 3.4673545e-23], [14.122599, 4.2987359e-22, 1.1365287e-22, 3.4471704e-23], [14.15461, 4.2792813e-22, 1.1312598e-22, 3.4271681e-23], [14.186693, 4.2599154e-22, 1.1260174e-22, 3.407345e-23], [14.218849, 4.2406377e-22, 1.1208016e-22, 3.3876988e-23], [14.251078, 4.2214478e-22, 1.115612e-22, 3.368227e-23], [14.28338, 4.2023453e-22, 1.1104486e-22, 3.3489275e-23], [14.315756, 4.1833299e-22, 1.1053111e-22, 3.3297978e-23], [14.348204, 4.1644011e-22, 1.1001994e-22, 3.3108357e-23], [14.380727, 4.1455584e-22, 1.0951134e-22, 3.2920392e-23], [14.413323, 4.1268016e-22, 1.0900528e-22, 3.2734058e-23], [14.445993, 4.1081301e-22, 1.0850174e-22, 3.2549337e-23], [14.478736, 4.0895437e-22, 1.0800072e-22, 3.2366206e-23], [14.511555, 4.0710419e-22, 1.075022e-22, 3.2184645e-23], [14.544447, 4.0526242e-22, 1.0700616e-22, 3.2004635e-23], [14.577414, 4.0342904e-22, 1.0651259e-22, 3.1826154e-23], [14.610456, 4.01604e-22, 1.0602146e-22, 3.1649185e-23], [14.643573, 3.9978726e-22, 1.0553277e-22, 3.1473707e-23], [14.676764, 3.9797879e-22, 1.050465e-22, 3.1299703e-23], [14.710031, 3.9617855e-22, 1.0456263e-22, 3.1127153e-23], [14.743374, 3.9438649e-22, 1.0408116e-22, 3.0956039e-23], [14.776792, 3.9260259e-22, 1.0360206e-22, 3.0786345e-23], [14.810286, 3.908268e-22, 1.0312532e-22, 3.0618051e-23], [14.843855, 3.8905908e-22, 1.0265092e-22, 3.0451142e-23], [14.877501, 3.872994e-22, 1.0217886e-22, 3.0285599e-23], [14.911223, 3.8554772e-22, 1.0170912e-22, 3.0121408e-23], [14.945021, 3.83804e-22, 1.0124168e-22, 2.995855e-23], [14.978896, 3.8206821e-22, 1.0077653e-22, 2.9797011e-23], [15.012848, 3.8034031e-22, 1.0031366e-22, 2.9636774e-23], [15.046877, 3.7862026e-22, 9.9853051e-23, 2.9477824e-23], [15.080983, 3.7690803e-22, 9.9394693e-23, 2.9320145e-23], [15.115166, 3.7520358e-22, 9.8938571e-23, 2.9163723e-23], [15.149427, 3.7350688e-22, 9.8484674e-23, 2.9008544e-23], [15.183765, 3.7181788e-22, 9.8032987e-23, 2.8854591e-23], [15.218181, 3.7013656e-22, 9.7583498e-23, 2.8701852e-23], [15.252676, 3.6846288e-22, 9.7136194e-23, 2.8550312e-23], [15.287248, 3.667968e-22, 9.6691062e-23, 2.8399957e-23], [15.321899, 3.6513828e-22, 9.624809e-23, 2.8250774e-23], [15.356628, 3.634873e-22, 9.5807266e-23, 2.810275e-23], [15.391436, 3.6184382e-22, 9.5368578e-23, 2.7955871e-23], [15.426323, 3.602078e-22, 9.4932012e-23, 2.7810125e-23], [15.461289, 3.5857921e-22, 9.4497557e-23, 2.7665498e-23], [15.496334, 3.5695801e-22, 9.4065202e-23, 2.7521978e-23], [15.531459, 3.5534418e-22, 9.3634934e-23, 2.7379554e-23], [15.566663, 3.5373767e-22, 9.3206741e-23, 2.7238213e-23], [15.601947, 3.5213846e-22, 9.2780613e-23, 2.7097942e-23], [15.637311, 3.5054651e-22, 9.2356536e-23, 2.6958731e-23], [15.672755, 3.4896178e-22, 9.1934501e-23, 2.6820568e-23], [15.70828, 3.4738425e-22, 9.1514495e-23, 2.6683442e-23], [15.743885, 3.4581387e-22, 9.1096508e-23, 2.654734e-23], [15.779571, 3.4425063e-22, 9.0680528e-23, 2.6412253e-23], [15.815337, 3.4269448e-22, 9.0266544e-23, 2.627817e-23], [15.851185, 3.4114539e-22, 8.9854545e-23, 2.614508e-23], [15.887114, 3.3960333e-22, 8.9444521e-23, 2.6012972e-23], [15.923125, 3.3806827e-22, 8.903646e-23, 2.5881836e-23], [15.959217, 3.3654017e-22, 8.8630352e-23, 2.5751662e-23], [15.99539, 3.3501901e-22, 8.8226186e-23, 2.5622441e-23], [16.031646, 3.3350475e-22, 8.7823952e-23, 2.5494161e-23], [16.067984, 3.3199736e-22, 8.742364e-23, 2.5366814e-23], [16.104405, 3.304968e-22, 8.7025238e-23, 2.524039e-23], [16.140908, 3.2900306e-22, 8.6628737e-23, 2.511488e-23], [16.177493, 3.2751609e-22, 8.6234127e-23, 2.4990274e-23], [16.214162, 3.2603587e-22, 8.5841397e-23, 2.4866563e-23], [16.250914, 3.2456236e-22, 8.5450537e-23, 2.4743738e-23], [16.287749, 3.2309553e-22, 8.5061537e-23, 2.4621791e-23], [16.324667, 3.2163536e-22, 8.4674389e-23, 2.4500713e-23], [16.36167, 3.201818e-22, 8.4289081e-23, 2.4380495e-23], [16.398756, 3.1873484e-22, 8.3905604e-23, 2.4261128e-23], [16.435926, 3.1729445e-22, 8.3523948e-23, 2.4142605e-23], [16.47318, 3.1586058e-22, 8.3144104e-23, 2.4024916e-23], [16.510519, 3.1443322e-22, 8.2766062e-23, 2.3908055e-23], [16.547942, 3.1301233e-22, 8.2389813e-23, 2.3792013e-23], [16.585451, 3.1159788e-22, 8.2015348e-23, 2.3676783e-23], [16.623044, 3.1018985e-22, 8.1642656e-23, 2.3562355e-23], [16.660723, 3.087882e-22, 8.127173e-23, 2.3448724e-23], [16.698486, 3.073929e-22, 8.0902559e-23, 2.3335881e-23], [16.736336, 3.0600393e-22, 8.0535135e-23, 2.3223819e-23], [16.774271, 3.0462126e-22, 8.0169448e-23, 2.311253e-23], [16.812293, 3.0324486e-22, 7.980549e-23, 2.3002008e-23], [16.8504, 3.0187469e-22, 7.9443252e-23, 2.2892245e-23], [16.888594, 3.0051074e-22, 7.9082724e-23, 2.2783235e-23], [16.926874, 2.9915297e-22, 7.8723899e-23, 2.267497e-23], [16.965242, 2.9780136e-22, 7.8366766e-23, 2.2567443e-23], [17.003696, 2.9645587e-22, 7.8011319e-23, 2.2460649e-23], [17.042237, 2.9511648e-22, 7.7657547e-23, 2.235458e-23], [17.080866, 2.9378316e-22, 7.7305443e-23, 2.2249229e-23], [17.119582, 2.9245588e-22, 7.6954998e-23, 2.214459e-23], [17.158386, 2.9113462e-22, 7.6606204e-23, 2.2040658e-23], [17.197278, 2.8981935e-22, 7.6259052e-23, 2.1937425e-23], [17.236258, 2.8851004e-22, 7.5913533e-23, 2.1834886e-23], [17.275326, 2.8720667e-22, 7.556964e-23, 2.1733034e-23], [17.314483, 2.859092e-22, 7.5227365e-23, 2.1631863e-23], [17.353729, 2.8461761e-22, 7.4886699e-23, 2.1531367e-23], [17.393064, 2.8333187e-22, 7.4547634e-23, 2.1431541e-23], [17.432488, 2.8205196e-22, 7.4210162e-23, 2.1332379e-23], [17.472001, 2.8077785e-22, 7.3874275e-23, 2.1233875e-23], [17.511604, 2.7950951e-22, 7.3539965e-23, 2.1136023e-23], [17.551296, 2.7824692e-22, 7.3207225e-23, 2.1038817e-23], [17.591079, 2.7699005e-22, 7.2876046e-23, 2.0942253e-23], [17.630952, 2.7573888e-22, 7.254642e-23, 2.0846325e-23], [17.670915, 2.7449338e-22, 7.2218341e-23, 2.0751027e-23], [17.710968, 2.7325352e-22, 7.1891799e-23, 2.0656355e-23], [17.751113, 2.7201927e-22, 7.1566789e-23, 2.0562302e-23], [17.791348, 2.7079062e-22, 7.1243301e-23, 2.0468865e-23], [17.831675, 2.6956754e-22, 7.0921328e-23, 2.0376037e-23], [17.872093, 2.6835e-22, 7.0600863e-23, 2.0283813e-23], [17.912603, 2.6713797e-22, 7.0281899e-23, 2.019219e-23], [17.953204, 2.6593143e-22, 6.9964428e-23, 2.0101161e-23], [17.993898, 2.6473037e-22, 6.9648442e-23, 2.0010722e-23], [18.034683, 2.6353474e-22, 6.9333935e-23, 1.9920868e-23], [18.075562, 2.6234453e-22, 6.9020899e-23, 1.9831595e-23], [18.116532, 2.6115971e-22, 6.8709327e-23, 1.9742897e-23], [18.157596, 2.5998025e-22, 6.8399211e-23, 1.9654771e-23], [18.198753, 2.5880615e-22, 6.8090545e-23, 1.9567211e-23], [18.240003, 2.5763736e-22, 6.7783321e-23, 1.9480213e-23], [18.281347, 2.5647386e-22, 6.7477532e-23, 1.9393773e-23], [18.322784, 2.5531564e-22, 6.7173172e-23, 1.9307886e-23], [18.364315, 2.5416266e-22, 6.6870233e-23, 1.9222547e-23], [18.405941, 2.530149e-22, 6.6568709e-23, 1.9137754e-23], [18.44766, 2.5187235e-22, 6.6268592e-23, 1.90535e-23], [18.489475, 2.5073497e-22, 6.5969876e-23, 1.8969783e-23], [18.531384, 2.4960274e-22, 6.5672553e-23, 1.8886598e-23], [18.573388, 2.4847564e-22, 6.5376618e-23, 1.880394e-23], [18.615487, 2.4735364e-22, 6.5082063e-23, 1.8721807e-23], [18.657682, 2.4623673e-22, 6.4788882e-23, 1.8640193e-23], [18.699972, 2.4512487e-22, 6.4497068e-23, 1.8559095e-23], [18.742358, 2.4401806e-22, 6.4206614e-23, 1.8478509e-23], [18.78484, 2.4291625e-22, 6.3917514e-23, 1.8398431e-23], [18.827419, 2.4181944e-22, 6.3629761e-23, 1.8318857e-23], [18.870094, 2.4072759e-22, 6.3343349e-23, 1.8239783e-23], [18.912866, 2.3964069e-22, 6.3058271e-23, 1.8161206e-23], [18.955734, 2.3855871e-22, 6.2774522e-23, 1.8083122e-23], [18.9987, 2.3748164e-22, 6.2492094e-23, 1.8005527e-23], [19.041764, 2.3640944e-22, 6.2210981e-23, 1.7928418e-23], [19.084925, 2.3534209e-22, 6.1931176e-23, 1.785179e-23], [19.128183, 2.3427958e-22, 6.1652675e-23, 1.7775642e-23], [19.17154, 2.3322189e-22, 6.1375469e-23, 1.7699968e-23], [19.214995, 2.3216898e-22, 6.1099554e-23, 1.7624765e-23], [19.258549, 2.3112084e-22, 6.0824923e-23, 1.7550031e-23], [19.302201, 2.3007745e-22, 6.0551569e-23, 1.7475761e-23], [19.345952, 2.2903879e-22, 6.0279487e-23, 1.7401953e-23], [19.389802, 2.2800483e-22, 6.0008671e-23, 1.7328603e-23], [19.433752, 2.2697555e-22, 5.9739114e-23, 1.7255707e-23], [19.477802, 2.2595093e-22, 5.9470811e-23, 1.7183262e-23], [19.521951, 2.2493095e-22, 5.9203755e-23, 1.7111266e-23], [19.5662, 2.239156e-22, 5.8937941e-23, 1.7039715e-23], [19.61055, 2.2290484e-22, 5.8673363e-23, 1.6968605e-23], [19.655, 2.2189866e-22, 5.8410015e-23, 1.6897934e-23], [19.699551, 2.2089703e-22, 5.814789e-23, 1.6827699e-23], [19.744203, 2.1989994e-22, 5.7886984e-23, 1.6757896e-23], [19.788956, 2.1890737e-22, 5.762729e-23, 1.6688522e-23], [19.83381, 2.1791929e-22, 5.7368803e-23, 1.6619575e-23], [19.878766, 2.1693569e-22, 5.7111516e-23, 1.6551052e-23], [19.923825, 2.1595654e-22, 5.6855425e-23, 1.6482949e-23], [19.968985, 2.1498183e-22, 5.6600524e-23, 1.6415263e-23], [20.014247, 2.1401153e-22, 5.6346807e-23, 1.6347992e-23], [20.059612, 2.1304562e-22, 5.6094267e-23, 1.6281133e-23], [20.10508, 2.1208409e-22, 5.5842901e-23, 1.6214683e-23], [20.150652, 2.1112692e-22, 5.5592702e-23, 1.614864e-23], [20.196326, 2.1017407e-22, 5.5343665e-23, 1.6082999e-23], [20.242104, 2.0922555e-22, 5.5095784e-23, 1.601776e-23], [20.287985, 2.0828132e-22, 5.4849054e-23, 1.5952918e-23], [20.333971, 2.0734136e-22, 5.4603469e-23, 1.5888471e-23], [20.380061, 2.0640566e-22, 5.4359024e-23, 1.5824417e-23], [20.426255, 2.054742e-22, 5.4115714e-23, 1.5760753e-23], [20.472554, 2.0454696e-22, 5.3873534e-23, 1.5697477e-23], [20.518958, 2.0362392e-22, 5.3632477e-23, 1.5634584e-23], [20.565467, 2.0270505e-22, 5.3392539e-23, 1.5572074e-23], [20.612082, 2.0179035e-22, 5.3153715e-23, 1.5509944e-23], [20.658802, 2.0087979e-22, 5.2915998e-23, 1.5448191e-23], [20.705628, 1.9997336e-22, 5.2679385e-23, 1.5386812e-23], [20.752561, 1.9907103e-22, 5.244387e-23, 1.5325805e-23], [20.799599, 1.9817278e-22, 5.2209448e-23, 1.5265168e-23], [20.846745, 1.9727861e-22, 5.1976113e-23, 1.5204897e-23], [20.893997, 1.9638848e-22, 5.1743861e-23, 1.5144992e-23], [20.941356, 1.9550238e-22, 5.1512686e-23, 1.5085449e-23], [20.988823, 1.946203e-22, 5.1282584e-23, 1.5026266e-23], [21.036397, 1.9374221e-22, 5.1053549e-23, 1.496744e-23], [21.084079, 1.928681e-22, 5.0825577e-23, 1.490897e-23], [21.131869, 1.9199795e-22, 5.0598662e-23, 1.4850853e-23], [21.179767, 1.9113174e-22, 5.03728e-23, 1.4793086e-23], [21.227774, 1.9026945e-22, 5.0147986e-23, 1.4735668e-23], [21.27589, 1.8941106e-22, 4.9924214e-23, 1.4678596e-23], [21.324115, 1.8855657e-22, 4.9701481e-23, 1.4621868e-23], [21.372449, 1.8770594e-22, 4.947978e-23, 1.4565481e-23], [21.420893, 1.8685917e-22, 4.9259108e-23, 1.4509434e-23], [21.469446, 1.8601623e-22, 4.903946e-23, 1.4453725e-23], [21.51811, 1.8517711e-22, 4.882083e-23, 1.439835e-23], [21.566884, 1.8434179e-22, 4.8603214e-23, 1.4343309e-23], [21.615768, 1.8351025e-22, 4.8386608e-23, 1.4288598e-23], [21.664764, 1.8268248e-22, 4.8171006e-23, 1.4234217e-23], [21.71387, 1.8185846e-22, 4.7956404e-23, 1.4180162e-23], [21.763087, 1.8103817e-22, 4.7742798e-23, 1.4126432e-23], [21.812417, 1.8022159e-22, 4.7530182e-23, 1.4073024e-23], [21.861858, 1.7940872e-22, 4.7318553e-23, 1.4019937e-23], [21.911411, 1.7859952e-22, 4.7107905e-23, 1.3967169e-23], [21.961076, 1.7779399e-22, 4.6898234e-23, 1.3914717e-23], [22.010854, 1.7699211e-22, 4.6689536e-23, 1.386258e-23], [22.060745, 1.7619386e-22, 4.6481806e-23, 1.3810756e-23], [22.110749, 1.7539923e-22, 4.6275039e-23, 1.3759242e-23], [22.160866, 1.7460819e-22, 4.6069231e-23, 1.3708037e-23], [22.211097, 1.7382074e-22, 4.5864379e-23, 1.3657139e-23], [22.261441, 1.7303686e-22, 4.5660476e-23, 1.3606545e-23], [22.3119, 1.7225652e-22, 4.5457519e-23, 1.3556255e-23], [22.362473, 1.7147972e-22, 4.5255504e-23, 1.3506266e-23], [22.413161, 1.7070644e-22, 4.5054426e-23, 1.3456577e-23], [22.463963, 1.6993666e-22, 4.4854281e-23, 1.3407184e-23], [22.514881, 1.6917037e-22, 4.4655065e-23, 1.3358088e-23], [22.565914, 1.6840755e-22, 4.4456773e-23, 1.3309285e-23], [22.617063, 1.6764819e-22, 4.42594e-23, 1.3260775e-23], [22.668328, 1.6689226e-22, 4.4062944e-23, 1.3212554e-23], [22.719709, 1.6613976e-22, 4.3867399e-23, 1.3164622e-23], [22.771207, 1.6539067e-22, 4.3672762e-23, 1.3116977e-23], [22.822821, 1.6464497e-22, 4.3479028e-23, 1.3069617e-23], [22.874552, 1.6390265e-22, 4.3286193e-23, 1.302254e-23], [22.926401, 1.6316369e-22, 4.3094253e-23, 1.2975745e-23], [22.978367, 1.6242808e-22, 4.2903204e-23, 1.2929229e-23], [23.03045, 1.616958e-22, 4.2713041e-23, 1.2882992e-23], [23.082652, 1.6096684e-22, 4.2523762e-23, 1.2837031e-23], [23.134972, 1.6024118e-22, 4.2335361e-23, 1.2791345e-23], [23.187411, 1.5951881e-22, 4.2147834e-23, 1.2745932e-23], [23.239969, 1.5879971e-22, 4.1961178e-23, 1.270079e-23], [23.292645, 1.5808387e-22, 4.1775389e-23, 1.2655918e-23], [23.345442, 1.5737127e-22, 4.1590462e-23, 1.2611315e-23], [23.398357, 1.566619e-22, 4.1406395e-23, 1.2566978e-23], [23.451393, 1.5595574e-22, 4.1223181e-23, 1.2522906e-23], [23.504549, 1.5525279e-22, 4.1040819e-23, 1.2479098e-23], [23.557825, 1.5455302e-22, 4.0859304e-23, 1.2435551e-23], [23.611223, 1.5385642e-22, 4.0678632e-23, 1.2392265e-23], [23.664741, 1.5316297e-22, 4.0498799e-23, 1.2349238e-23], [23.71838, 1.5247267e-22, 4.0319802e-23, 1.2306468e-23], [23.772142, 1.5178549e-22, 4.0141636e-23, 1.2263953e-23], [23.826025, 1.5110143e-22, 3.9964298e-23, 1.2221693e-23], [23.88003, 1.5042047e-22, 3.9787784e-23, 1.2179686e-23], [23.934157, 1.4974259e-22, 3.961209e-23, 1.213793e-23], [23.988407, 1.4906779e-22, 3.9437213e-23, 1.2096423e-23], [24.042781, 1.4839604e-22, 3.9263149e-23, 1.2055165e-23], [24.097277, 1.4772734e-22, 3.9089893e-23, 1.2014153e-23], [24.151897, 1.4706166e-22, 3.8917443e-23, 1.1973387e-23], [24.206641, 1.4639901e-22, 3.8745795e-23, 1.1932865e-23], [24.261508, 1.4573935e-22, 3.8574945e-23, 1.1892585e-23], [24.316501, 1.4508269e-22, 3.840489e-23, 1.1852546e-23], [24.371617, 1.44429e-22, 3.8235625e-23, 1.1812747e-23], [24.426859, 1.4377828e-22, 3.8067148e-23, 1.1773185e-23], [24.482226, 1.431305e-22, 3.7899454e-23, 1.1733861e-23], [24.537719, 1.4248566e-22, 3.7732541e-23, 1.1694772e-23], [24.593337, 1.4184374e-22, 3.7566404e-23, 1.1655918e-23], [24.649081, 1.4120473e-22, 3.740104e-23, 1.1617296e-23], [24.704952, 1.4056862e-22, 3.7236446e-23, 1.1578905e-23], [24.760949, 1.3993538e-22, 3.7072617e-23, 1.1540744e-23], [24.817074, 1.3930502e-22, 3.6909552e-23, 1.1502812e-23], [24.873325, 1.3867752e-22, 3.6747245e-23, 1.1465107e-23], [24.929704, 1.3805286e-22, 3.6585694e-23, 1.1427628e-23], [24.986211, 1.3743103e-22, 3.6424896e-23, 1.1390374e-23], [25.042846, 1.3681202e-22, 3.6264846e-23, 1.1353344e-23], [25.099609, 1.3619582e-22, 3.6105541e-23, 1.1316535e-23], [25.156501, 1.3558241e-22, 3.5946979e-23, 1.1279947e-23], [25.213522, 1.3497178e-22, 3.5789156e-23, 1.1243579e-23], [25.270672, 1.3436392e-22, 3.5632067e-23, 1.1207429e-23], [25.327951, 1.3375881e-22, 3.5475711e-23, 1.1171495e-23], [25.385361, 1.3315645e-22, 3.5320084e-23, 1.1135778e-23], [25.4429, 1.3255682e-22, 3.5165182e-23, 1.1100275e-23], [25.50057, 1.3195991e-22, 3.5011002e-23, 1.1064985e-23], [25.558371, 1.313657e-22, 3.4857541e-23, 1.1029907e-23], [25.616303, 1.3077418e-22, 3.4704796e-23, 1.0995041e-23], [25.674366, 1.3018535e-22, 3.4552763e-23, 1.0960383e-23], [25.73256, 1.2959919e-22, 3.4401439e-23, 1.0925934e-23], [25.790887, 1.2901569e-22, 3.4250822e-23, 1.0891692e-23], [25.849346, 1.2843483e-22, 3.4100907e-23, 1.0857656e-23], [25.907937, 1.278566e-22, 3.3951692e-23, 1.0823825e-23], [25.966661, 1.27281e-22, 3.3803173e-23, 1.0790198e-23], [26.025518, 1.2670801e-22, 3.3655348e-23, 1.0756773e-23], [26.084509, 1.2613761e-22, 3.3508213e-23, 1.0723549e-23], [26.143633, 1.255698e-22, 3.3361765e-23, 1.0690525e-23], [26.202891, 1.2500457e-22, 3.3216001e-23, 1.0657701e-23], [26.262284, 1.244419e-22, 3.3070918e-23, 1.0625074e-23], [26.321811, 1.2388178e-22, 3.2926512e-23, 1.0592643e-23], [26.381473, 1.233242e-22, 3.2782782e-23, 1.0560409e-23], [26.441271, 1.2276915e-22, 3.2639724e-23, 1.0528368e-23], [26.501204, 1.2221662e-22, 3.2497334e-23, 1.0496521e-23], [26.561272, 1.216666e-22, 3.235561e-23, 1.0464867e-23], [26.621477, 1.2111906e-22, 3.2214548e-23, 1.0433403e-23], [26.681819, 1.2057402e-22, 3.2074147e-23, 1.040213e-23], [26.742297, 1.2003144e-22, 3.1934402e-23, 1.0371045e-23], [26.802912, 1.1949133e-22, 3.1795312e-23, 1.0340148e-23], [26.863665, 1.1895366e-22, 3.1656872e-23, 1.0309438e-23], [26.924555, 1.1841844e-22, 3.151908e-23, 1.0278913e-23], [26.985584, 1.1788564e-22, 3.1381933e-23, 1.0248574e-23], [27.04675, 1.1735526e-22, 3.1245429e-23, 1.0218417e-23], [27.108056, 1.1682728e-22, 3.1109564e-23, 1.0188444e-23], [27.1695, 1.1630171e-22, 3.0974335e-23, 1.0158651e-23], [27.231084, 1.1577851e-22, 3.083974e-23, 1.012904e-23], [27.292807, 1.1525769e-22, 3.0705776e-23, 1.0099607e-23], [27.35467, 1.1473923e-22, 3.0572439e-23, 1.0070353e-23], [27.416673, 1.1422313e-22, 3.0439728e-23, 1.0041277e-23], [27.478817, 1.1370936e-22, 3.0307639e-23, 1.0012376e-23], [27.541102, 1.1319793e-22, 3.017617e-23, 9.9836511e-24], [27.603527, 1.1268882e-22, 3.0045318e-23, 9.9551004e-24], [27.666095, 1.1218202e-22, 2.9915079e-23, 9.926723e-24], [27.728804, 1.1167752e-22, 2.9785452e-23, 9.8985179e-24], [27.791655, 1.1117531e-22, 2.9656434e-23, 9.8704842e-24], [27.854649, 1.1067537e-22, 2.9528021e-23, 9.8426208e-24], [27.917786, 1.1017771e-22, 2.9400211e-23, 9.8149268e-24], [27.981065, 1.096823e-22, 2.9273002e-23, 9.7874012e-24], [28.044488, 1.0918915e-22, 2.914639e-23, 9.760043e-24], [28.108055, 1.0869823e-22, 2.9020374e-23, 9.7328513e-24], [28.171766, 1.0820954e-22, 2.889495e-23, 9.7058252e-24], [28.235622, 1.0772307e-22, 2.8770116e-23, 9.6789636e-24], [28.299622, 1.0723881e-22, 2.8645869e-23, 9.6522656e-24], [28.363767, 1.0675674e-22, 2.8522206e-23, 9.6257303e-24], [28.428057, 1.0627687e-22, 2.8399126e-23, 9.5993568e-24], [28.492494, 1.0579917e-22, 2.8276625e-23, 9.5731441e-24], [28.557076, 1.0532364e-22, 2.81547e-23, 9.5470914e-24], [28.621805, 1.0485027e-22, 2.803335e-23, 9.5211976e-24], [28.68668, 1.0437905e-22, 2.7912571e-23, 9.4954618e-24], [28.751703, 1.0390997e-22, 2.7792361e-23, 9.4698833e-24], [28.816873, 1.0344302e-22, 2.7672718e-23, 9.4444609e-24], [28.88219, 1.0297819e-22, 2.7553639e-23, 9.419194e-24], [28.947656, 1.0251547e-22, 2.7435122e-23, 9.3940815e-24], [29.01327, 1.0205485e-22, 2.7317163e-23, 9.3691225e-24], [29.079033, 1.0159633e-22, 2.7199761e-23, 9.3443163e-24], [29.144944, 1.0113988e-22, 2.7082914e-23, 9.3196618e-24], [29.211006, 1.0068551e-22, 2.6966617e-23, 9.2951583e-24], [29.277216, 1.0023321e-22, 2.685087e-23, 9.2708049e-24], [29.343577, 9.9782954e-23, 2.673567e-23, 9.2466006e-24], [29.410089, 9.9334747e-23, 2.6621014e-23, 9.2225446e-24], [29.476751, 9.8888576e-23, 2.6506899e-23, 9.1986362e-24], [29.543564, 9.8444432e-23, 2.6393325e-23, 9.1748743e-24], [29.610529, 9.8002306e-23, 2.6280287e-23, 9.1512582e-24], [29.677646, 9.7562189e-23, 2.6167784e-23, 9.1277871e-24], [29.744914, 9.7124072e-23, 2.6055813e-23, 9.10446e-24], [29.812335, 9.6687945e-23, 2.5944372e-23, 9.0812762e-24], [29.879909, 9.6253799e-23, 2.5833459e-23, 9.0582348e-24], [29.947636, 9.5821627e-23, 2.5723071e-23, 9.035335e-24], [30.015517, 9.5391418e-23, 2.5613207e-23, 9.0125759e-24], [30.083552, 9.4963165e-23, 2.5503862e-23, 8.9899568e-24], [30.15174, 9.4536858e-23, 2.5395036e-23, 8.9674769e-24], [30.220084, 9.4112488e-23, 2.5286726e-23, 8.9451353e-24], [30.288582, 9.3690048e-23, 2.517893e-23, 8.9229312e-24], [30.357235, 9.3269527e-23, 2.5071646e-23, 8.9008638e-24], [30.426044, 9.2850918e-23, 2.496487e-23, 8.8789324e-24], [30.495009, 9.2434212e-23, 2.4858602e-23, 8.8571361e-24], [30.56413, 9.20194e-23, 2.4752838e-23, 8.8354741e-24], [30.633408, 9.1606474e-23, 2.4647577e-23, 8.8139457e-24], [30.702844, 9.1195425e-23, 2.4542816e-23, 8.7925501e-24], [30.772436, 9.0786246e-23, 2.4438553e-23, 8.7712865e-24], [30.842186, 9.0378927e-23, 2.4334786e-23, 8.7501541e-24], [30.912094, 8.997346e-23, 2.4231512e-23, 8.7291521e-24], [30.982161, 8.9569837e-23, 2.4128731e-23, 8.7082799e-24], [31.052387, 8.9168049e-23, 2.4026438e-23, 8.6875366e-24], [31.122771, 8.8768089e-23, 2.3924633e-23, 8.6669215e-24], [31.193315, 8.8369948e-23, 2.3823312e-23, 8.6464338e-24], [31.26402, 8.7973618e-23, 2.3722475e-23, 8.6260728e-24], [31.334884, 8.7579091e-23, 2.3622118e-23, 8.6058377e-24], [31.405909, 8.7186358e-23, 2.352224e-23, 8.5857278e-24], [31.477095, 8.6795412e-23, 2.3422839e-23, 8.5657424e-24], [31.548442, 8.6406244e-23, 2.3323912e-23, 8.5458807e-24], [31.619951, 8.6018847e-23, 2.3225457e-23, 8.5261419e-24], [31.691623, 8.5633213e-23, 2.3127473e-23, 8.5065255e-24], [31.763456, 8.5249334e-23, 2.3029957e-23, 8.4870305e-24], [31.835453, 8.4867201e-23, 2.2932907e-23, 8.4676564e-24], [31.907612, 8.4486807e-23, 2.2836321e-23, 8.4484024e-24], [31.979936, 8.4108145e-23, 2.2740197e-23, 8.4292678e-24], [32.052423, 8.3731206e-23, 2.2644533e-23, 8.4102519e-24], [32.125074, 8.3355982e-23, 2.2549327e-23, 8.3913539e-24], [32.19789, 8.2982467e-23, 2.2454577e-23, 8.3725733e-24], [32.270871, 8.2610652e-23, 2.2360281e-23, 8.3539092e-24], [32.344018, 8.224053e-23, 2.2266438e-23, 8.3353609e-24], [32.41733, 8.1872092e-23, 2.2173044e-23, 8.3169279e-24], [32.490809, 8.1505332e-23, 2.2080098e-23, 8.2986094e-24], [32.564454, 8.1140243e-23, 2.1987598e-23, 8.2804047e-24], [32.638266, 8.0776815e-23, 2.1895543e-23, 8.2623132e-24], [32.712245, 8.0415043e-23, 2.1803929e-23, 8.2443341e-24], [32.786392, 8.0054918e-23, 2.1712756e-23, 8.2264669e-24], [32.860707, 7.9696433e-23, 2.1622021e-23, 8.2087107e-24], [32.935191, 7.9339582e-23, 2.1531722e-23, 8.191065e-24], [33.009843, 7.8984355e-23, 2.1441858e-23, 8.1735291e-24], [33.084665, 7.8630747e-23, 2.1352427e-23, 8.1561024e-24], [33.159656, 7.827875e-23, 2.1263426e-23, 8.1387841e-24], [33.234817, 7.7928357e-23, 2.1174854e-23, 8.1215737e-24], [33.310148, 7.757956e-23, 2.1086709e-23, 8.1044704e-24], [33.385651, 7.7232353e-23, 2.0998989e-23, 8.0874736e-24], [33.461324, 7.6886728e-23, 2.0911692e-23, 8.0705828e-24], [33.537169, 7.6542678e-23, 2.0824817e-23, 8.0537972e-24], [33.613186, 7.6200196e-23, 2.0738361e-23, 8.0371162e-24], [33.689375, 7.5859275e-23, 2.0652322e-23, 8.0205392e-24], [33.765737, 7.5519908e-23, 2.05667e-23, 8.0040655e-24], [33.842272, 7.5182088e-23, 2.0481491e-23, 7.9876946e-24], [33.91898, 7.4845808e-23, 2.0396695e-23, 7.9714257e-24], [33.995862, 7.4511062e-23, 2.0312309e-23, 7.9552584e-24], [34.072919, 7.4177842e-23, 2.0228332e-23, 7.9391919e-24], [34.15015, 7.3846141e-23, 2.0144761e-23, 7.9232256e-24], [34.227556, 7.3515954e-23, 2.0061596e-23, 7.907359e-24], [34.305138, 7.3187272e-23, 1.9978834e-23, 7.8915914e-24], [34.382895, 7.2860089e-23, 1.9896473e-23, 7.8759222e-24], [34.460829, 7.2534399e-23, 1.9814513e-23, 7.8603509e-24], [34.53894, 7.2210195e-23, 1.973295e-23, 7.8448768e-24], [34.617227, 7.1887469e-23, 1.9651784e-23, 7.8294993e-24], [34.695692, 7.1566217e-23, 1.9571012e-23, 7.8142178e-24], [34.774335, 7.124643e-23, 1.9490633e-23, 7.7990318e-24], [34.853156, 7.0928102e-23, 1.9410645e-23, 7.7839406e-24], [34.932155, 7.0611227e-23, 1.9331047e-23, 7.7689438e-24], [35.011334, 7.0295799e-23, 1.9251836e-23, 7.7540406e-24], [35.090692, 6.998181e-23, 1.9173012e-23, 7.7392305e-24], [35.17023, 6.9669255e-23, 1.9094571e-23, 7.724513e-24], [35.249949, 6.9358127e-23, 1.9016514e-23, 7.7098875e-24], [35.329848, 6.9048419e-23, 1.8938838e-23, 7.6953533e-24], [35.409928, 6.8740126e-23, 1.8861541e-23, 7.6809101e-24], [35.49019, 6.843324e-23, 1.8784622e-23, 7.666557e-24], [35.570633, 6.8127756e-23, 1.8708078e-23, 7.6522938e-24], [35.651259, 6.7823667e-23, 1.863191e-23, 7.6381196e-24], [35.732068, 6.7520967e-23, 1.8556114e-23, 7.6240341e-24], [35.81306, 6.721965e-23, 1.8480689e-23, 7.6100366e-24], [35.894235, 6.691971e-23, 1.8405634e-23, 7.5961267e-24], [35.975595, 6.6621141e-23, 1.8330947e-23, 7.5823037e-24], [36.057139, 6.6323935e-23, 1.8256627e-23, 7.5685671e-24], [36.138867, 6.6028088e-23, 1.8182671e-23, 7.5549165e-24], [36.220781, 6.5733593e-23, 1.8109079e-23, 7.5413511e-24], [36.302881, 6.5440444e-23, 1.8035848e-23, 7.5278706e-24], [36.385167, 6.5148636e-23, 1.7962977e-23, 7.5144744e-24], [36.467639, 6.4858161e-23, 1.7890465e-23, 7.5011619e-24], [36.550298, 6.4569015e-23, 1.7818309e-23, 7.4879327e-24], [36.633144, 6.428119e-23, 1.774651e-23, 7.4747861e-24], [36.716179, 6.3994683e-23, 1.7675064e-23, 7.4617218e-24], [36.799401, 6.3709485e-23, 1.760397e-23, 7.4487391e-24], [36.882812, 6.3425592e-23, 1.7533227e-23, 7.4358376e-24], [36.966413, 6.3142998e-23, 1.7462834e-23, 7.4230167e-24], [37.050202, 6.2861697e-23, 1.7392788e-23, 7.410276e-24], [37.134182, 6.2581683e-23, 1.7323088e-23, 7.3976149e-24], [37.218352, 6.230295e-23, 1.7253734e-23, 7.385033e-24], [37.302713, 6.2025493e-23, 1.7184723e-23, 7.3725296e-24], [37.387265, 6.1749307e-23, 1.7116053e-23, 7.3601044e-24], [37.472008, 6.1474384e-23, 1.7047724e-23, 7.3477569e-24], [37.556944, 6.120072e-23, 1.6979733e-23, 7.3354864e-24], [37.642072, 6.0928309e-23, 1.691208e-23, 7.3232927e-24], [37.727393, 6.0657146e-23, 1.6844763e-23, 7.3111751e-24], [37.812908, 6.0387224e-23, 1.6777781e-23, 7.2991332e-24], [37.898616, 6.0118539e-23, 1.6711131e-23, 7.2871665e-24], [37.984519, 5.9851084e-23, 1.6644813e-23, 7.2752745e-24], [38.070616, 5.9584855e-23, 1.6578826e-23, 7.2634567e-24], [38.156909, 5.9319846e-23, 1.6513166e-23, 7.2517128e-24], [38.243397, 5.9056051e-23, 1.6447835e-23, 7.2400421e-24], [38.330081, 5.8793466e-23, 1.6382829e-23, 7.2284443e-24], [38.416962, 5.8532084e-23, 1.6318147e-23, 7.2169188e-24], [38.504039, 5.82719e-23, 1.6253789e-23, 7.2054652e-24], [38.591314, 5.8012909e-23, 1.6189752e-23, 7.1940831e-24], [38.678787, 5.7755105e-23, 1.6126036e-23, 7.182772e-24], [38.766458, 5.7498484e-23, 1.6062639e-23, 7.1715314e-24], [38.854328, 5.724304e-23, 1.5999559e-23, 7.1603609e-24], [38.942397, 5.6988767e-23, 1.5936795e-23, 7.14926e-24], [39.030665, 5.6735662e-23, 1.5874346e-23, 7.1382282e-24], [39.119134, 5.6483717e-23, 1.5812211e-23, 7.1272652e-24], [39.207803, 5.6232929e-23, 1.5750387e-23, 7.1163705e-24], [39.296673, 5.5983291e-23, 1.5688875e-23, 7.1055437e-24], [39.385745, 5.57348e-23, 1.5627671e-23, 7.0947842e-24], [39.475018, 5.5487449e-23, 1.5566776e-23, 7.0840917e-24], [39.564494, 5.5241234e-23, 1.5506188e-23, 7.0734658e-24], [39.654172, 5.499615e-23, 1.5445905e-23, 7.062906e-24], [39.744054, 5.4752191e-23, 1.5385926e-23, 7.0524118e-24], [39.83414, 5.4509353e-23, 1.5326249e-23, 7.0419829e-24], [39.92443, 5.4267631e-23, 1.5266875e-23, 7.0316188e-24], [40.014924, 5.4027019e-23, 1.52078e-23, 7.0213192e-24], [40.105624, 5.3787513e-23, 1.5149024e-23, 7.0110835e-24], [40.196529, 5.3549107e-23, 1.5090546e-23, 7.0009114e-24], [40.28764, 5.3311798e-23, 1.5032364e-23, 6.9908025e-24], [40.378958, 5.3075579e-23, 1.4974478e-23, 6.9807563e-24], [40.470483, 5.2840447e-23, 1.4916885e-23, 6.9707725e-24], [40.562215, 5.2606396e-23, 1.4859584e-23, 6.9608506e-24], [40.654155, 5.2373422e-23, 1.4802575e-23, 6.9509902e-24], [40.746303, 5.2141519e-23, 1.4745855e-23, 6.9411909e-24], [40.838661, 5.1910683e-23, 1.4689425e-23, 6.9314524e-24], [40.931227, 5.1680909e-23, 1.4633282e-23, 6.9217742e-24], [41.024004, 5.1452193e-23, 1.4577425e-23, 6.912156e-24], [41.116991, 5.122453e-23, 1.4521853e-23, 6.9025972e-24], [41.210188, 5.0997915e-23, 1.4466565e-23, 6.8930977e-24], [41.303597, 5.0772343e-23, 1.4411559e-23, 6.8836569e-24], [41.397218, 5.054781e-23, 1.4356835e-23, 6.8742744e-24], [41.49105, 5.0324312e-23, 1.430239e-23, 6.86495e-24], [41.585096, 5.0101843e-23, 1.4248225e-23, 6.8556832e-24], [41.679354, 4.9880399e-23, 1.4194337e-23, 6.8464736e-24], [41.773827, 4.9659975e-23, 1.4140726e-23, 6.8373208e-24], [41.868513, 4.9440568e-23, 1.408739e-23, 6.8282246e-24], [41.963414, 4.9222171e-23, 1.4034329e-23, 6.8191844e-24], [42.05853, 4.9004782e-23, 1.398154e-23, 6.8102e-24], [42.153862, 4.8788396e-23, 1.3929023e-23, 6.801271e-24], [42.24941, 4.8573007e-23, 1.3876776e-23, 6.792397e-24], [42.345174, 4.8358612e-23, 1.3824799e-23, 6.7835776e-24], [42.441155, 4.8145207e-23, 1.3773091e-23, 6.7748125e-24], [42.537354, 4.7932786e-23, 1.3721649e-23, 6.7661013e-24], [42.633771, 4.7721345e-23, 1.3670473e-23, 6.7574436e-24], [42.730407, 4.7510881e-23, 1.3619562e-23, 6.7488392e-24], [42.827262, 4.7301389e-23, 1.3568915e-23, 6.7402877e-24], [42.924336, 4.7092864e-23, 1.351853e-23, 6.7317886e-24], [43.02163, 4.6885302e-23, 1.3468407e-23, 6.7233417e-24], [43.119145, 4.6678699e-23, 1.3418543e-23, 6.7149466e-24], [43.21688, 4.6473051e-23, 1.3368939e-23, 6.706603e-24], [43.314838, 4.6268353e-23, 1.3319593e-23, 6.6983105e-24], [43.413017, 4.6064602e-23, 1.3270504e-23, 6.6900688e-24], [43.511419, 4.5861792e-23, 1.3221671e-23, 6.6818775e-24], [43.610044, 4.5659921e-23, 1.3173092e-23, 6.6737363e-24], [43.708892, 4.5458983e-23, 1.3124767e-23, 6.665645e-24], [43.807965, 4.5258975e-23, 1.3076694e-23, 6.657603e-24], [43.907262, 4.5059892e-23, 1.3028873e-23, 6.6496102e-24], [44.006784, 4.4861731e-23, 1.2981302e-23, 6.6416662e-24], [44.106531, 4.4664487e-23, 1.293398e-23, 6.6337706e-24], [44.206505, 4.4468157e-23, 1.2886906e-23, 6.6259232e-24], [44.306706, 4.4272735e-23, 1.284008e-23, 6.6181236e-24], [44.407133, 4.4078219e-23, 1.2793499e-23, 6.6103715e-24], [44.507788, 4.3884604e-23, 1.2747163e-23, 6.6026665e-24], [44.608672, 4.3691887e-23, 1.2701071e-23, 6.5950085e-24], [44.709784, 4.3500062e-23, 1.2655222e-23, 6.5873969e-24], [44.811125, 4.3309127e-23, 1.2609615e-23, 6.5798317e-24], [44.912696, 4.3119077e-23, 1.2564248e-23, 6.5723123e-24], [45.014497, 4.2929909e-23, 1.2519121e-23, 6.5648386e-24], [45.116528, 4.2741619e-23, 1.2474232e-23, 6.5574102e-24], [45.218792, 4.2554202e-23, 1.2429582e-23, 6.5500268e-24], [45.321286, 4.2367655e-23, 1.2385167e-23, 6.5426881e-24], [45.424014, 4.2181974e-23, 1.2340989e-23, 6.5353939e-24], [45.526974, 4.1997155e-23, 1.2297044e-23, 6.5281437e-24], [45.630167, 4.1813195e-23, 1.2253333e-23, 6.5209374e-24], [45.733594, 4.1630089e-23, 1.2209855e-23, 6.5137746e-24], [45.837256, 4.1447834e-23, 1.2166608e-23, 6.506655e-24], [45.941153, 4.1266427e-23, 1.2123592e-23, 6.4995784e-24], [46.045285, 4.1085862e-23, 1.2080805e-23, 6.4925444e-24], [46.149653, 4.0906137e-23, 1.2038246e-23, 6.4855528e-24], [46.254258, 4.0727249e-23, 1.1995915e-23, 6.4786033e-24], [46.3591, 4.0549192e-23, 1.195381e-23, 6.4716956e-24], [46.46418, 4.0371964e-23, 1.1911931e-23, 6.4648293e-24], [46.569498, 4.0195562e-23, 1.1870276e-23, 6.4580044e-24], [46.675054, 4.001998e-23, 1.1828845e-23, 6.4512204e-24], [46.78085, 3.9845216e-23, 1.1787636e-23, 6.4444771e-24], [46.886885, 3.9671267e-23, 1.1746649e-23, 6.4377742e-24], [46.993161, 3.9498128e-23, 1.1705882e-23, 6.4311114e-24], [47.099678, 3.9325795e-23, 1.1665335e-23, 6.4244885e-24], [47.206436, 3.9154267e-23, 1.1625006e-23, 6.4179052e-24], [47.313436, 3.8983538e-23, 1.1584896e-23, 6.4113613e-24], [47.420679, 3.8813605e-23, 1.1545002e-23, 6.4048564e-24], [47.528165, 3.8644466e-23, 1.1505324e-23, 6.3983903e-24], [47.635894, 3.8476116e-23, 1.146586e-23, 6.3919628e-24], [47.743868, 3.8308552e-23, 1.1426611e-23, 6.3855736e-24], [47.852086, 3.814177e-23, 1.1387575e-23, 6.3792224e-24], [47.96055, 3.7975767e-23, 1.134875e-23, 6.3729089e-24], [48.069259, 3.781054e-23, 1.1310137e-23, 6.366633e-24], [48.178215, 3.7646085e-23, 1.1271734e-23, 6.3603944e-24], [48.287418, 3.74824e-23, 1.123354e-23, 6.3541928e-24], [48.396868, 3.7319479e-23, 1.1195554e-23, 6.3480279e-24], [48.506567, 3.7157321e-23, 1.1157776e-23, 6.3418996e-24], [48.616514, 3.6995921e-23, 1.1120205e-23, 6.3358075e-24], [48.72671, 3.6835277e-23, 1.1082839e-23, 6.3297515e-24], [48.837156, 3.6675385e-23, 1.1045677e-23, 6.3237312e-24], [48.947853, 3.6516242e-23, 1.100872e-23, 6.3177465e-24], [49.0588, 3.6357844e-23, 1.0971965e-23, 6.3117971e-24], [49.169999, 3.6200189e-23, 1.0935412e-23, 6.3058828e-24], [49.28145, 3.6043273e-23, 1.089906e-23, 6.3000033e-24], [49.393153, 3.5887092e-23, 1.0862909e-23, 6.2941584e-24], [49.50511, 3.5731645e-23, 1.0826957e-23, 6.2883479e-24], [49.61732, 3.5576926e-23, 1.0791203e-23, 6.2825715e-24], [49.729785, 3.5422934e-23, 1.0755647e-23, 6.276829e-24], [49.842505, 3.5269665e-23, 1.0720287e-23, 6.2711202e-24], [49.95548, 3.5117115e-23, 1.0685124e-23, 6.2654449e-24], [50.068711, 3.4965283e-23, 1.0650155e-23, 6.2598028e-24], [50.182199, 3.4814164e-23, 1.061538e-23, 6.2541936e-24], [50.295944, 3.4663755e-23, 1.0580798e-23, 6.2486173e-24], [50.409947, 3.4514054e-23, 1.0546409e-23, 6.2430735e-24], [50.524209, 3.4365057e-23, 1.0512211e-23, 6.2375621e-24], [50.638729, 3.4216762e-23, 1.0478204e-23, 6.2320827e-24], [50.753509, 3.4069164e-23, 1.0444387e-23, 6.2266353e-24], [50.868549, 3.3922262e-23, 1.0410758e-23, 6.2212196e-24], [50.98385, 3.3776051e-23, 1.0377318e-23, 6.2158354e-24], [51.099412, 3.363053e-23, 1.0344065e-23, 6.2104824e-24], [51.215237, 3.3485695e-23, 1.0310998e-23, 6.2051605e-24], [51.331323, 3.3341543e-23, 1.0278117e-23, 6.1998695e-24], [51.447673, 3.3198071e-23, 1.024542e-23, 6.1946091e-24], [51.564287, 3.3055276e-23, 1.0212907e-23, 6.1893791e-24], [51.681164, 3.2913155e-23, 1.0180578e-23, 6.1841794e-24], [51.798307, 3.2771705e-23, 1.014843e-23, 6.1790097e-24], [51.915716, 3.2630924e-23, 1.0116465e-23, 6.1738699e-24], [52.03339, 3.2490808e-23, 1.0084679e-23, 6.1687596e-24], [52.151331, 3.2351355e-23, 1.0053074e-23, 6.1636789e-24], [52.26954, 3.2212562e-23, 1.0021647e-23, 6.1586273e-24], [52.388016, 3.2074425e-23, 9.9903986e-24, 6.1536048e-24], [52.506761, 3.1936942e-23, 9.9593275e-24, 6.1486112e-24], [52.625775, 3.180011e-23, 9.928433e-24, 6.1436462e-24], [52.745059, 3.1663926e-23, 9.8977142e-24, 6.1387097e-24], [52.864613, 3.1528388e-23, 9.8671703e-24, 6.1338015e-24], [52.984439, 3.1393493e-23, 9.8368007e-24, 6.1289213e-24], [53.104535, 3.1259237e-23, 9.8066043e-24, 6.1240691e-24], [53.224904, 3.1125619e-23, 9.7765806e-24, 6.1192445e-24], [53.345546, 3.0992634e-23, 9.7467286e-24, 6.1144475e-24], [53.466462, 3.0860282e-23, 9.7170476e-24, 6.1096779e-24], [53.587651, 3.0728558e-23, 9.6875368e-24, 6.1049354e-24], [53.709115, 3.059746e-23, 9.6581955e-24, 6.1002199e-24], [53.830855, 3.0466986e-23, 9.6290228e-24, 6.0955312e-24], [53.95287, 3.0337132e-23, 9.6000181e-24, 6.0908692e-24], [54.075162, 3.0207896e-23, 9.5711804e-24, 6.0862336e-24], [54.197731, 3.0079276e-23, 9.5425091e-24, 6.0816243e-24], [54.320578, 2.9951268e-23, 9.5140033e-24, 6.0770411e-24], [54.443703, 2.9823871e-23, 9.4856624e-24, 6.0724838e-24], [54.567108, 2.969708e-23, 9.4574855e-24, 6.0679523e-24], [54.690792, 2.9570895e-23, 9.429472e-24, 6.0634464e-24], [54.814757, 2.9445311e-23, 9.401621e-24, 6.0589659e-24], [54.939002, 2.9320327e-23, 9.3739317e-24, 6.0545106e-24], [55.063529, 2.919594e-23, 9.3464035e-24, 6.0500805e-24], [55.188339, 2.9072148e-23, 9.3190357e-24, 6.0456752e-24], [55.313431, 2.8948947e-23, 9.2918273e-24, 6.0412948e-24], [55.438807, 2.8826336e-23, 9.2647778e-24, 6.0369389e-24], [55.564467, 2.8704311e-23, 9.2378863e-24, 6.0326075e-24], [55.690412, 2.858287e-23, 9.2111522e-24, 6.0283003e-24], [55.816642, 2.8462011e-23, 9.1845746e-24, 6.0240173e-24], [55.943158, 2.8341732e-23, 9.1581529e-24, 6.0197582e-24], [56.069962, 2.8222029e-23, 9.1318864e-24, 6.0155229e-24], [56.197052, 2.81029e-23, 9.1057743e-24, 6.0113113e-24], [56.324431, 2.7984342e-23, 9.0798158e-24, 6.0071232e-24], [56.452098, 2.7866355e-23, 9.0540103e-24, 6.0029584e-24], [56.580055, 2.7748933e-23, 9.028357e-24, 5.9988168e-24], [56.708302, 2.7632077e-23, 9.0028553e-24, 5.9946982e-24], [56.83684, 2.7515782e-23, 8.9775044e-24, 5.9906026e-24], [56.965668, 2.7400046e-23, 8.9523035e-24, 5.9865296e-24], [57.094789, 2.7284868e-23, 8.927252e-24, 5.9824793e-24], [57.224203, 2.7170245e-23, 8.9023493e-24, 5.9784514e-24], [57.35391, 2.7056174e-23, 8.8775944e-24, 5.9744458e-24], [57.483911, 2.6942652e-23, 8.8529869e-24, 5.9704624e-24], [57.614206, 2.6829679e-23, 8.8285259e-24, 5.966501e-24], [57.744797, 2.6717251e-23, 8.8042108e-24, 5.9625614e-24], [57.875684, 2.6605365e-23, 8.7800408e-24, 5.9586436e-24], [58.006867, 2.649402e-23, 8.7560154e-24, 5.9547474e-24], [58.138348, 2.6383214e-23, 8.7321337e-24, 5.9508726e-24], [58.270127, 2.6272943e-23, 8.7083951e-24, 5.9470191e-24], [58.402205, 2.6163207e-23, 8.6847989e-24, 5.9431868e-24], [58.534582, 2.6054001e-23, 8.6613444e-24, 5.9393756e-24], [58.667259, 2.5945325e-23, 8.638031e-24, 5.9355853e-24], [58.800237, 2.5837175e-23, 8.614858e-24, 5.9318157e-24], [58.933516, 2.572955e-23, 8.5918246e-24, 5.9280668e-24], [59.067097, 2.5622448e-23, 8.5689302e-24, 5.9243384e-24], [59.200981, 2.5515866e-23, 8.5461742e-24, 5.9206304e-24], [59.335169, 2.5409801e-23, 8.5235558e-24, 5.9169427e-24], [59.46966, 2.5304252e-23, 8.5010745e-24, 5.913275e-24], [59.604457, 2.5199217e-23, 8.4787294e-24, 5.9096274e-24], [59.739559, 2.5094693e-23, 8.4565201e-24, 5.9059996e-24], [59.874967, 2.4990678e-23, 8.4344457e-24, 5.9023916e-24], [60.010683, 2.4887171e-23, 8.4125057e-24, 5.8988032e-24], [60.146705, 2.4784167e-23, 8.3906994e-24, 5.8952343e-24], [60.283037, 2.4681667e-23, 8.3690261e-24, 5.8916848e-24], [60.419677, 2.4579667e-23, 8.3474852e-24, 5.8881545e-24], [60.556627, 2.4478166e-23, 8.326076e-24, 5.8846433e-24], [60.693887, 2.437716e-23, 8.3047978e-24, 5.8811512e-24], [60.831458, 2.4276649e-23, 8.2836502e-24, 5.877678e-24], [60.969342, 2.417663e-23, 8.2626323e-24, 5.8742235e-24], [61.107537, 2.4077101e-23, 8.2417435e-24, 5.8707877e-24], [61.246046, 2.3978059e-23, 8.2209833e-24, 5.8673705e-24], [61.384869, 2.3879504e-23, 8.2003509e-24, 5.8639717e-24], [61.524007, 2.3781432e-23, 8.1798458e-24, 5.8605912e-24], [61.66346, 2.3683841e-23, 8.1594672e-24, 5.8572289e-24], [61.803229, 2.3586731e-23, 8.1392146e-24, 5.8538847e-24], [61.943315, 2.3490098e-23, 8.1190874e-24, 5.8505584e-24], [62.083718, 2.339394e-23, 8.0990849e-24, 5.8472501e-24], [62.22444, 2.3298256e-23, 8.0792064e-24, 5.8439595e-24], [62.365481, 2.3203044e-23, 8.0594515e-24, 5.8406865e-24], [62.506841, 2.3108301e-23, 8.0398193e-24, 5.8374311e-24], [62.648522, 2.3014026e-23, 8.0203094e-24, 5.8341932e-24], [62.790524, 2.2920216e-23, 8.0009211e-24, 5.8309725e-24], [62.932847, 2.282687e-23, 7.9816538e-24, 5.8277691e-24], [63.075494, 2.2733985e-23, 7.9625068e-24, 5.8245828e-24], [63.218463, 2.264156e-23, 7.9434797e-24, 5.8214136e-24], [63.361757, 2.2549593e-23, 7.9245717e-24, 5.8182612e-24], [63.505376, 2.2458082e-23, 7.9057822e-24, 5.8151257e-24], [63.64932, 2.2367024e-23, 7.8871107e-24, 5.8120068e-24], [63.79359, 2.2276419e-23, 7.8685566e-24, 5.8089046e-24], [63.938187, 2.2186263e-23, 7.8501192e-24, 5.8058189e-24], [64.083113, 2.2096556e-23, 7.8317979e-24, 5.8027496e-24], [64.228366, 2.2007294e-23, 7.8135922e-24, 5.7996967e-24], [64.373949, 2.1918478e-23, 7.7955015e-24, 5.7966599e-24], [64.519862, 2.1830103e-23, 7.7775251e-24, 5.7936393e-24], [64.666105, 2.174217e-23, 7.7596625e-24, 5.7906347e-24], [64.81268, 2.1654675e-23, 7.7419132e-24, 5.787646e-24], [64.959588, 2.1567617e-23, 7.7242764e-24, 5.7846732e-24], [65.106828, 2.1480994e-23, 7.7067516e-24, 5.7817161e-24], [65.254402, 2.1394804e-23, 7.6893383e-24, 5.7787747e-24], [65.40231, 2.1309046e-23, 7.6720359e-24, 5.7758489e-24], [65.550554, 2.1223717e-23, 7.6548437e-24, 5.7729385e-24], [65.699134, 2.1138816e-23, 7.6377613e-24, 5.7700435e-24], [65.84805, 2.1054341e-23, 7.620788e-24, 5.7671638e-24], [65.997304, 2.097029e-23, 7.6039232e-24, 5.7642993e-24], [66.146897, 2.0886662e-23, 7.5871665e-24, 5.7614499e-24], [66.296828, 2.0803454e-23, 7.5705172e-24, 5.7586156e-24], [66.447099, 2.0720666e-23, 7.5539747e-24, 5.7557962e-24], [66.597711, 2.0638294e-23, 7.5375386e-24, 5.7529916e-24], [66.748665, 2.0556338e-23, 7.5212082e-24, 5.7502019e-24], [66.89996, 2.0474795e-23, 7.5049829e-24, 5.7474268e-24], [67.051599, 2.0393664e-23, 7.4888624e-24, 5.7446663e-24], [67.203581, 2.0312943e-23, 7.4728458e-24, 5.7419204e-24], [67.355907, 2.0232631e-23, 7.4569328e-24, 5.7391889e-24], [67.508579, 2.0152726e-23, 7.4411228e-24, 5.7364717e-24], [67.661597, 2.0073225e-23, 7.4254152e-24, 5.7337689e-24], [67.814962, 1.9994128e-23, 7.4098094e-24, 5.7310802e-24], [67.968674, 1.9915432e-23, 7.394305e-24, 5.7284057e-24], [68.122735, 1.9837137e-23, 7.3789014e-24, 5.7257452e-24], [68.277145, 1.9759239e-23, 7.363598e-24, 5.7230987e-24], [68.431905, 1.9681739e-23, 7.3483943e-24, 5.720466e-24], [68.587015, 1.9604633e-23, 7.3332898e-24, 5.7178472e-24], [68.742478, 1.952792e-23, 7.3182839e-24, 5.7152421e-24], [68.898292, 1.9451599e-23, 7.3033761e-24, 5.7126507e-24], [69.05446, 1.9375669e-23, 7.2885659e-24, 5.7100728e-24], [69.210982, 1.9300126e-23, 7.2738527e-24, 5.7075085e-24], [69.367859, 1.922497e-23, 7.2592361e-24, 5.7049576e-24], [69.525091, 1.91502e-23, 7.2447154e-24, 5.70242e-24], [69.68268, 1.9075813e-23, 7.2302902e-24, 5.6998958e-24], [69.840626, 1.9001807e-23, 7.2159599e-24, 5.6973848e-24], [69.998929, 1.8928183e-23, 7.2017241e-24, 5.6948869e-24], [70.157592, 1.8854937e-23, 7.1875821e-24, 5.6924022e-24], [70.316614, 1.8782068e-23, 7.1735336e-24, 5.6899304e-24], [70.475997, 1.8709574e-23, 7.1595779e-24, 5.6874716e-24], [70.635741, 1.8637455e-23, 7.1457146e-24, 5.6850256e-24], [70.795847, 1.8565708e-23, 7.1319432e-24, 5.6825925e-24], [70.956316, 1.8494332e-23, 7.1182631e-24, 5.6801721e-24], [71.117149, 1.8423326e-23, 7.1046739e-24, 5.6777643e-24], [71.278346, 1.8352687e-23, 7.091175e-24, 5.6753692e-24], [71.439908, 1.8282414e-23, 7.077766e-24, 5.6729866e-24], [71.601837, 1.8212507e-23, 7.0644463e-24, 5.6706165e-24], [71.764133, 1.8142962e-23, 7.0512155e-24, 5.6682588e-24], [71.926797, 1.8073779e-23, 7.038073e-24, 5.6659135e-24], [72.089829, 1.8004957e-23, 7.0250184e-24, 5.6635804e-24], [72.253231, 1.7936493e-23, 7.0120511e-24, 5.6612596e-24], [72.417004, 1.7868386e-23, 6.9991707e-24, 5.6589509e-24], [72.581147, 1.7800636e-23, 6.9863768e-24, 5.6566544e-24], [72.745663, 1.7733239e-23, 6.9736687e-24, 5.6543698e-24], [72.910551, 1.7666195e-23, 6.961046e-24, 5.6520973e-24], [73.075813, 1.7599503e-23, 6.9485083e-24, 5.6498367e-24], [73.24145, 1.753316e-23, 6.9360551e-24, 5.6475879e-24], [73.407463, 1.7467166e-23, 6.9236859e-24, 5.6453509e-24], [73.573851, 1.7401518e-23, 6.9114001e-24, 5.6431257e-24], [73.740617, 1.7336217e-23, 6.8991974e-24, 5.6409122e-24], [73.907761, 1.7271259e-23, 6.8870773e-24, 5.6387103e-24], [74.075283, 1.7206644e-23, 6.8750392e-24, 5.6365199e-24], [74.243185, 1.714237e-23, 6.8630828e-24, 5.6343411e-24], [74.411468, 1.7078436e-23, 6.8512076e-24, 5.6321737e-24], [74.580133, 1.701484e-23, 6.8394131e-24, 5.6300178e-24], [74.749179, 1.6951581e-23, 6.8276988e-24, 5.6278732e-24], [74.918609, 1.6888658e-23, 6.8160643e-24, 5.6257399e-24], [75.088423, 1.6826069e-23, 6.8045091e-24, 5.6236178e-24], [75.258621, 1.6763813e-23, 6.7930327e-24, 5.621507e-24], [75.429206, 1.6701888e-23, 6.7816348e-24, 5.6194072e-24], [75.600177, 1.6640293e-23, 6.7703149e-24, 5.6173186e-24], [75.771536, 1.6579027e-23, 6.7590724e-24, 5.615241e-24], [75.943283, 1.6518088e-23, 6.7479071e-24, 5.6131744e-24], [76.115419, 1.6457476e-23, 6.7368183e-24, 5.6111187e-24], [76.287946, 1.6397187e-23, 6.7258057e-24, 5.6090739e-24], [76.460863, 1.6337222e-23, 6.7148689e-24, 5.6070399e-24], [76.634173, 1.6277579e-23, 6.7040073e-24, 5.6050167e-24], [76.807875, 1.6218256e-23, 6.6932206e-24, 5.6030043e-24], [76.981971, 1.6159253e-23, 6.6825084e-24, 5.6010025e-24], [77.156462, 1.6100567e-23, 6.6718701e-24, 5.5990114e-24], [77.331348, 1.6042198e-23, 6.6613053e-24, 5.5970309e-24], [77.506631, 1.5984145e-23, 6.6508137e-24, 5.5950609e-24], [77.682311, 1.5926405e-23, 6.6403948e-24, 5.5931014e-24], [77.858389, 1.5868978e-23, 6.6300482e-24, 5.5911524e-24], [78.034866, 1.5811862e-23, 6.6197733e-24, 5.5892138e-24], [78.211743, 1.5755056e-23, 6.60957e-24, 5.5872855e-24], [78.389021, 1.5698559e-23, 6.5994376e-24, 5.5853676e-24], [78.566701, 1.5642369e-23, 6.5893758e-24, 5.5834599e-24], [78.744784, 1.5586486e-23, 6.5793842e-24, 5.5815625e-24], [78.923271, 1.5530907e-23, 6.5694623e-24, 5.5796752e-24], [79.102161, 1.5475632e-23, 6.5596097e-24, 5.5777981e-24], [79.281458, 1.542066e-23, 6.5498261e-24, 5.5759312e-24], [79.461161, 1.5365988e-23, 6.540111e-24, 5.5740742e-24], [79.641271, 1.5311617e-23, 6.5304641e-24, 5.5722273e-24], [79.821789, 1.5257544e-23, 6.5208848e-24, 5.5703904e-24], [80.002717, 1.5203768e-23, 6.5113728e-24, 5.5685634e-24], [80.184054, 1.5150289e-23, 6.5019278e-24, 5.5667463e-24], [80.365803, 1.5097104e-23, 6.4925492e-24, 5.564939e-24], [80.547964, 1.5044214e-23, 6.4832368e-24, 5.5631416e-24], [80.730537, 1.4991615e-23, 6.47399e-24, 5.561354e-24], [80.913525, 1.4939308e-23, 6.4648086e-24, 5.5595761e-24], [81.096927, 1.4887291e-23, 6.4556921e-24, 5.5578079e-24], [81.280745, 1.4835562e-23, 6.4466402e-24, 5.5560493e-24], [81.464979, 1.4784122e-23, 6.4376524e-24, 5.5543004e-24], [81.649631, 1.4732967e-23, 6.4287284e-24, 5.5525611e-24], [81.834702, 1.4682098e-23, 6.4198677e-24, 5.5508313e-24], [82.020192, 1.4631513e-23, 6.4110701e-24, 5.5491111e-24], [82.206103, 1.4581211e-23, 6.4023351e-24, 5.5474003e-24], [82.392434, 1.453119e-23, 6.3936623e-24, 5.545699e-24], [82.579189, 1.448145e-23, 6.3850514e-24, 5.544007e-24], [82.766366, 1.4431989e-23, 6.376502e-24, 5.5423245e-24], [82.953968, 1.4382806e-23, 6.3680137e-24, 5.5406513e-24], [83.141995, 1.43339e-23, 6.3595861e-24, 5.5389874e-24], [83.330449, 1.428527e-23, 6.3512189e-24, 5.5373328e-24], [83.519329, 1.4236914e-23, 6.3429118e-24, 5.5356874e-24], [83.708638, 1.4188832e-23, 6.3346643e-24, 5.5340512e-24], [83.898375, 1.4141023e-23, 6.3264761e-24, 5.5324242e-24], [84.088543, 1.4093484e-23, 6.3183468e-24, 5.5308063e-24], [84.279142, 1.4046216e-23, 6.3102761e-24, 5.5291976e-24], [84.470173, 1.3999216e-23, 6.3022636e-24, 5.5275979e-24], [84.661636, 1.3952485e-23, 6.2943089e-24, 5.5260073e-24], [84.853534, 1.390602e-23, 6.2864118e-24, 5.5244256e-24], [85.045867, 1.385982e-23, 6.2785718e-24, 5.522853e-24], [85.238636, 1.3813885e-23, 6.2707886e-24, 5.5212893e-24], [85.431841, 1.3768213e-23, 6.2630618e-24, 5.5197346e-24], [85.625485, 1.3722804e-23, 6.2553912e-24, 5.5181887e-24], [85.819567, 1.3677655e-23, 6.2477763e-24, 5.5166517e-24], [86.01409, 1.3632767e-23, 6.2402169e-24, 5.5151236e-24], [86.209053, 1.3588138e-23, 6.2327125e-24, 5.5136042e-24], [86.404458, 1.3543766e-23, 6.2252629e-24, 5.5120937e-24], [86.600306, 1.3499651e-23, 6.2178677e-24, 5.5105918e-24], [86.796598, 1.3455791e-23, 6.2105265e-24, 5.5090988e-24], [86.993335, 1.3412187e-23, 6.2032391e-24, 5.5076144e-24], [87.190518, 1.3368835e-23, 6.1960051e-24, 5.5061386e-24], [87.388148, 1.3325736e-23, 6.1888242e-24, 5.5046716e-24], [87.586226, 1.3282889e-23, 6.181696e-24, 5.5032131e-24], [87.784753, 1.3240291e-23, 6.1746203e-24, 5.5017633e-24], [87.983729, 1.3197943e-23, 6.1675966e-24, 5.500322e-24], [88.183157, 1.3155843e-23, 6.1606248e-24, 5.4988892e-24], [88.383037, 1.311399e-23, 6.1537044e-24, 5.497465e-24], [88.58337, 1.3072383e-23, 6.1468351e-24, 5.4960492e-24], [88.784157, 1.3031022e-23, 6.1400167e-24, 5.4946419e-24], [88.985399, 1.2989904e-23, 6.1332488e-24, 5.4932431e-24], [89.187097, 1.2949029e-23, 6.126531e-24, 5.4918527e-24], [89.389252, 1.2908395e-23, 6.1198632e-24, 5.4904707e-24], [89.591866, 1.2868003e-23, 6.1132449e-24, 5.489097e-24], [89.794939, 1.282785e-23, 6.1066759e-24, 5.4877317e-24], [89.998472, 1.2787937e-23, 6.1001559e-24, 5.4863748e-24], [90.202466, 1.2748261e-23, 6.0936846e-24, 5.4850261e-24], [90.406923, 1.2708821e-23, 6.0872615e-24, 5.4836857e-24], [90.611843, 1.2669618e-23, 6.0808866e-24, 5.4823536e-24], [90.817228, 1.2630649e-23, 6.0745594e-24, 5.4810297e-24], [91.023079, 1.2591914e-23, 6.0682797e-24, 5.4797141e-24], [91.229395, 1.2553411e-23, 6.0620471e-24, 5.4784066e-24], [91.43618, 1.251514e-23, 6.0558614e-24, 5.4771073e-24], [91.643433, 1.24771e-23, 6.0497223e-24, 5.4758162e-24], [91.851156, 1.243929e-23, 6.0436295e-24, 5.4745332e-24], [92.05935, 1.2401708e-23, 6.0375827e-24, 5.4732583e-24], [92.268016, 1.2364354e-23, 6.0315817e-24, 5.4719916e-24], [92.477155, 1.2327227e-23, 6.025626e-24, 5.4707329e-24], [92.686767, 1.2290325e-23, 6.0197155e-24, 5.4694822e-24], [92.896855, 1.2253648e-23, 6.0138499e-24, 5.4682396e-24], [93.107419, 1.2217195e-23, 6.0080289e-24, 5.4670051e-24], [93.318461, 1.2180965e-23, 6.0022521e-24, 5.4657785e-24], [93.52998, 1.2144957e-23, 5.9965195e-24, 5.4645599e-24], [93.741979, 1.2109169e-23, 5.9908306e-24, 5.4633493e-24], [93.954459, 1.2073602e-23, 5.9851851e-24, 5.4621466e-24], [94.16742, 1.2038253e-23, 5.9795829e-24, 5.4609519e-24], [94.380864, 1.2003123e-23, 5.9740237e-24, 5.4597651e-24], [94.594792, 1.1968209e-23, 5.9685071e-24, 5.4585862e-24], [94.809205, 1.1933511e-23, 5.963033e-24, 5.4574152e-24], [95.024104, 1.1899029e-23, 5.957601e-24, 5.456252e-24], [95.239489, 1.1864761e-23, 5.9522109e-24, 5.4550967e-24], [95.455363, 1.1830706e-23, 5.9468624e-24, 5.4539493e-24], [95.671727, 1.1796864e-23, 5.9415553e-24, 5.4528096e-24], [95.88858, 1.1763232e-23, 5.9362893e-24, 5.4516778e-24], [96.105926, 1.1729812e-23, 5.9310642e-24, 5.4505538e-24], [96.323764, 1.16966e-23, 5.9258797e-24, 5.4494375e-24], [96.542095, 1.1663598e-23, 5.9207355e-24, 5.448329e-24], [96.760922, 1.1630803e-23, 5.9156315e-24, 5.4472282e-24], [96.980244, 1.1598214e-23, 5.9105672e-24, 5.4461352e-24], [97.200064, 1.1565832e-23, 5.9055426e-24, 5.4450499e-24], [97.420382, 1.1533654e-23, 5.9005574e-24, 5.4439723e-24], [97.641199, 1.1501681e-23, 5.8956113e-24, 5.4429024e-24], [97.862517, 1.146991e-23, 5.890704e-24, 5.4418402e-24], [98.084336, 1.1438342e-23, 5.8858354e-24, 5.4407857e-24], [98.306659, 1.1406975e-23, 5.8810051e-24, 5.4397388e-24], [98.529485, 1.1375808e-23, 5.8762131e-24, 5.4386996e-24], [98.752816, 1.1344841e-23, 5.8714589e-24, 5.437668e-24], [98.976653, 1.1314073e-23, 5.8667424e-24, 5.436644e-24], [99.200998, 1.1283502e-23, 5.8620634e-24, 5.4356276e-24], [99.425852, 1.1253128e-23, 5.8574216e-24, 5.4346188e-24], [99.651215, 1.1222949e-23, 5.8528168e-24, 5.4336177e-24], [99.877088, 1.1192966e-23, 5.8482487e-24, 5.432624e-24], [100.10347, 1.1163177e-23, 5.8437172e-24, 5.431638e-24], [100.33037, 1.1133581e-23, 5.8392219e-24, 5.4306595e-24], [100.55779, 1.1104178e-23, 5.8347628e-24, 5.4296886e-24], [100.78571, 1.1074966e-23, 5.8303395e-24, 5.4287251e-24], [101.01416, 1.1045945e-23, 5.8259519e-24, 5.4277693e-24], [101.24312, 1.1017113e-23, 5.8215997e-24, 5.4268209e-24], [101.47261, 1.098847e-23, 5.8172827e-24, 5.42588e-24], [101.70261, 1.0960016e-23, 5.8130007e-24, 5.4249467e-24], [101.93313, 1.0931748e-23, 5.8087534e-24, 5.4240208e-24], [102.16418, 1.0903667e-23, 5.8045407e-24, 5.4231024e-24], [102.39575, 1.0875771e-23, 5.8003624e-24, 5.4221915e-24], [102.62784, 1.084806e-23, 5.7962182e-24, 5.421288e-24], [102.86046, 1.0820533e-23, 5.7921079e-24, 5.420392e-24], [103.09361, 1.0793188e-23, 5.7880313e-24, 5.4195034e-24], [103.32729, 1.0766026e-23, 5.7839883e-24, 5.4186223e-24], [103.56149, 1.0739045e-23, 5.7799786e-24, 5.4177486e-24], [103.79623, 1.0712244e-23, 5.776002e-24, 5.4168823e-24], [104.0315, 1.0685623e-23, 5.7720583e-24, 5.4160235e-24], [104.2673, 1.065918e-23, 5.7681473e-24, 5.4151721e-24], [104.50364, 1.0632915e-23, 5.7642688e-24, 5.414328e-24], [104.74051, 1.0606828e-23, 5.7604226e-24, 5.4134914e-24], [104.97792, 1.0580916e-23, 5.7566086e-24, 5.4126621e-24], [105.21587, 1.055518e-23, 5.7528265e-24, 5.4118403e-24], [105.45435, 1.0529618e-23, 5.7490761e-24, 5.4110258e-24], [105.69338, 1.0504231e-23, 5.7453572e-24, 5.4102187e-24], [105.93295, 1.0479016e-23, 5.7416697e-24, 5.409419e-24], [106.17306, 1.0453973e-23, 5.7380133e-24, 5.4086266e-24], [106.41372, 1.0429101e-23, 5.734388e-24, 5.4078416e-24], [106.65492, 1.04044e-23, 5.7307934e-24, 5.4070639e-24], [106.89667, 1.0379869e-23, 5.7272294e-24, 5.4062936e-24], [107.13897, 1.0355506e-23, 5.7236958e-24, 5.4055306e-24], [107.38181, 1.0331312e-23, 5.7201925e-24, 5.404775e-24], [107.62521, 1.0307285e-23, 5.7167192e-24, 5.4040267e-24], [107.86916, 1.0283424e-23, 5.7132759e-24, 5.4032857e-24], [108.11366, 1.0259729e-23, 5.7098622e-24, 5.4025521e-24], [108.35871, 1.0236199e-23, 5.706478e-24, 5.4018258e-24], [108.60432, 1.0212833e-23, 5.7031232e-24, 5.4011068e-24], [108.85049, 1.018963e-23, 5.6997976e-24, 5.4003951e-24], [109.09722, 1.0166589e-23, 5.696501e-24, 5.3996907e-24], [109.3445, 1.0143711e-23, 5.6932332e-24, 5.3989937e-24], [109.59235, 1.0120993e-23, 5.6899941e-24, 5.398304e-24], [109.84075, 1.0098435e-23, 5.6867834e-24, 5.3976215e-24], [110.08972, 1.0076037e-23, 5.6836012e-24, 5.3969464e-24], [110.33926, 1.0053798e-23, 5.680447e-24, 5.3962786e-24], [110.58936, 1.0031716e-23, 5.6773209e-24, 5.395618e-24], [110.84002, 1.0009791e-23, 5.6742226e-24, 5.3949648e-24], [111.09126, 9.988023e-24, 5.671152e-24, 5.3943189e-24], [111.34306, 9.9664102e-24, 5.6681089e-24, 5.3936803e-24], [111.59544, 9.9449521e-24, 5.6650932e-24, 5.3930489e-24], [111.84839, 9.9236481e-24, 5.6621046e-24, 5.3924249e-24], [112.10191, 9.9024973e-24, 5.6591431e-24, 5.3918081e-24], [112.356, 9.881499e-24, 5.6562085e-24, 5.3911986e-24], [112.61067, 9.8606525e-24, 5.6533006e-24, 5.3905965e-24], [112.86592, 9.8399569e-24, 5.6504193e-24, 5.3900016e-24], [113.12175, 9.8194116e-24, 5.6475644e-24, 5.389414e-24], [113.37815, 9.7990158e-24, 5.6447358e-24, 5.3888337e-24], [113.63514, 9.7787687e-24, 5.6419334e-24, 5.3882607e-24], [113.89271, 9.7586697e-24, 5.6391569e-24, 5.387695e-24], [114.15087, 9.7387179e-24, 5.6364062e-24, 5.3871365e-24], [114.40961, 9.7189127e-24, 5.6336812e-24, 5.3865854e-24], [114.66893, 9.6992533e-24, 5.6309818e-24, 5.3860415e-24], [114.92885, 9.6797389e-24, 5.6283078e-24, 5.385505e-24], [115.18935, 9.6603689e-24, 5.625659e-24, 5.3849757e-24], [115.45044, 9.6411426e-24, 5.6230353e-24, 5.3844538e-24], [115.71213, 9.6220591e-24, 5.6204366e-24, 5.3839391e-24], [115.9744, 9.6031178e-24, 5.6178628e-24, 5.3834317e-24], [116.23728, 9.5843179e-24, 5.6153136e-24, 5.3829317e-24], [116.50075, 9.5656588e-24, 5.612789e-24, 5.3824389e-24], [116.76481, 9.5471398e-24, 5.6102889e-24, 5.3819535e-24], [117.02948, 9.52876e-24, 5.607813e-24, 5.3814753e-24], [117.29474, 9.5105188e-24, 5.6053613e-24, 5.3810045e-24], [117.5606, 9.4924155e-24, 5.6029336e-24, 5.3805409e-24], [117.82707, 9.4744494e-24, 5.6005298e-24, 5.3800847e-24], [118.09414, 9.4566198e-24, 5.5981498e-24, 5.3796358e-24], [118.36182, 9.438926e-24, 5.5957935e-24, 5.3791942e-24], [118.63011, 9.4213673e-24, 5.5934606e-24, 5.37876e-24], [118.899, 9.4039429e-24, 5.5911511e-24, 5.3783331e-24], [119.1685, 9.3866523e-24, 5.5888649e-24, 5.3779135e-24], [119.43861, 9.3694947e-24, 5.5866019e-24, 5.3775012e-24], [119.70934, 9.3524694e-24, 5.5843618e-24, 5.3770963e-24], [119.98068, 9.3355757e-24, 5.5821447e-24, 5.3766987e-24], [120.25263, 9.318813e-24, 5.5799503e-24, 5.3763085e-24], [120.5252, 9.3021805e-24, 5.5777785e-24, 5.3759256e-24], [120.79839, 9.2856776e-24, 5.5756293e-24, 5.3755501e-24], [121.07219, 9.2693037e-24, 5.5735026e-24, 5.3751819e-24], [121.34662, 9.253058e-24, 5.5713981e-24, 5.3748211e-24], [121.62167, 9.2369399e-24, 5.5693158e-24, 5.3744677e-24], [121.89734, 9.2209486e-24, 5.5672555e-24, 5.3741216e-24], [122.17364, 9.2050836e-24, 5.5652173e-24, 5.3737829e-24], [122.45057, 9.1893442e-24, 5.5632008e-24, 5.3734516e-24], [122.72812, 9.1737297e-24, 5.5612061e-24, 5.3731277e-24], [123.0063, 9.1582394e-24, 5.559233e-24, 5.3728112e-24], [123.28511, 9.1428727e-24, 5.5572815e-24, 5.3725021e-24], [123.56455, 9.127629e-24, 5.5553513e-24, 5.3722004e-24], [123.84463, 9.1125076e-24, 5.5534425e-24, 5.3719062e-24], [124.12534, 9.0975077e-24, 5.5515548e-24, 5.3716193e-24], [124.40669, 9.0826289e-24, 5.5496882e-24, 5.3713399e-24], [124.68868, 9.0678705e-24, 5.5478426e-24, 5.3710679e-24], [124.9713, 9.0532317e-24, 5.5460178e-24, 5.3708033e-24], [125.25457, 9.038712e-24, 5.5442139e-24, 5.3705462e-24], [125.53847, 9.0243107e-24, 5.5424306e-24, 5.3702965e-24], [125.82302, 9.0100273e-24, 5.5406679e-24, 5.3700543e-24], [126.10822, 8.9958609e-24, 5.5389256e-24, 5.3698196e-24], [126.39406, 8.9818112e-24, 5.5372037e-24, 5.3695923e-24], [126.68055, 8.9678773e-24, 5.5355021e-24, 5.3693725e-24], [126.96769, 8.9540587e-24, 5.5338207e-24, 5.3691602e-24], [127.25548, 8.9403548e-24, 5.5321594e-24, 5.3689555e-24], [127.54393, 8.9267649e-24, 5.530518e-24, 5.3687582e-24], [127.83302, 8.9132884e-24, 5.5288966e-24, 5.3685684e-24], [128.12277, 8.8999247e-24, 5.5272949e-24, 5.3683862e-24], [128.41318, 8.8866733e-24, 5.5257129e-24, 5.3682114e-24], [128.70425, 8.8735334e-24, 5.5241506e-24, 5.3680443e-24], [128.99598, 8.8605045e-24, 5.5226078e-24, 5.3678846e-24], [129.28836, 8.8475859e-24, 5.5210844e-24, 5.3677326e-24], [129.58142, 8.8347772e-24, 5.5195804e-24, 5.3675881e-24], [129.87513, 8.8220776e-24, 5.5180956e-24, 5.3674511e-24], [130.16951, 8.8094866e-24, 5.51663e-24, 5.3673218e-24], [130.46456, 8.7970035e-24, 5.5151835e-24, 5.3672e-24], [130.76028, 8.7846279e-24, 5.513756e-24, 5.3670859e-24], [131.05666, 8.772359e-24, 5.5123474e-24, 5.3669794e-24], [131.35372, 8.7601964e-24, 5.5109577e-24, 5.3668804e-24], [131.65145, 8.7481394e-24, 5.5095867e-24, 5.3667892e-24], [131.94986, 8.7361874e-24, 5.5082343e-24, 5.3667055e-24], [132.24894, 8.7243399e-24, 5.5069006e-24, 5.3666295e-24], [132.5487, 8.7125963e-24, 5.5055854e-24, 5.3665612e-24], [132.84915, 8.700956e-24, 5.5042886e-24, 5.3665006e-24], [133.15027, 8.6894185e-24, 5.5030101e-24, 5.3664476e-24], [133.45207, 8.6779831e-24, 5.5017499e-24, 5.3664023e-24], [133.75456, 8.6666493e-24, 5.5005079e-24, 5.3663648e-24], [134.05773, 8.6554166e-24, 5.499284e-24, 5.3663349e-24], [134.36159, 8.6442844e-24, 5.4980782e-24, 5.3663128e-24], [134.66614, 8.6332521e-24, 5.4968904e-24, 5.3662984e-24], [134.97138, 8.6223191e-24, 5.4957204e-24, 5.3662918e-24], [135.27732, 8.611485e-24, 5.4945683e-24, 5.3662929e-24], [135.58394, 8.6007491e-24, 5.4934339e-24, 5.3663018e-24], [135.89126, 8.5901109e-24, 5.4923172e-24, 5.3663185e-24], [136.19928, 8.5795699e-24, 5.4912181e-24, 5.366343e-24], [136.508, 8.5691255e-24, 5.4901366e-24, 5.3663753e-24], [136.81741, 8.5587772e-24, 5.4890725e-24, 5.3664154e-24], [137.12753, 8.5485245e-24, 5.4880258e-24, 5.3664633e-24], [137.43835, 8.5383667e-24, 5.4869965e-24, 5.3665191e-24], [137.74987, 8.5283034e-24, 5.4859844e-24, 5.3665828e-24], [138.0621, 8.5183341e-24, 5.4849895e-24, 5.3666543e-24], [138.37504, 8.5084582e-24, 5.4840118e-24, 5.3667337e-24], [138.68868, 8.4986751e-24, 5.4830511e-24, 5.366821e-24], [139.00304, 8.4889845e-24, 5.4821075e-24, 5.3669162e-24], [139.31811, 8.4793857e-24, 5.4811808e-24, 5.3670194e-24], [139.6339, 8.4698782e-24, 5.480271e-24, 5.3671304e-24], [139.9504, 8.4604615e-24, 5.479378e-24, 5.3672495e-24], [140.26761, 8.4511351e-24, 5.4785018e-24, 5.3673764e-24], [140.58555, 8.4418985e-24, 5.4776423e-24, 5.3675114e-24], [140.90421, 8.4327512e-24, 5.4767994e-24, 5.3676544e-24], [141.22359, 8.4236927e-24, 5.4759731e-24, 5.3678053e-24], [141.54369, 8.4147224e-24, 5.4751634e-24, 5.3679643e-24], [141.86452, 8.4058399e-24, 5.4743701e-24, 5.3681313e-24], [142.18608, 8.3970447e-24, 5.4735932e-24, 5.3683064e-24], [142.50836, 8.3883363e-24, 5.4728327e-24, 5.3684895e-24], [142.83138, 8.3797142e-24, 5.4720885e-24, 5.3686808e-24], [143.15513, 8.3711779e-24, 5.4713605e-24, 5.3688801e-24], [143.47961, 8.3627269e-24, 5.4706487e-24, 5.3690875e-24], [143.80482, 8.3543607e-24, 5.4699531e-24, 5.369303e-24], [144.13078, 8.3460789e-24, 5.4692735e-24, 5.3695267e-24], [144.45747, 8.3378809e-24, 5.46861e-24, 5.3697585e-24], [144.78491, 8.3297663e-24, 5.4679625e-24, 5.3699985e-24], [145.11308, 8.3217347e-24, 5.4673309e-24, 5.3702466e-24], [145.442, 8.3137855e-24, 5.4667152e-24, 5.370503e-24], [145.77167, 8.3059183e-24, 5.4661154e-24, 5.3707676e-24], [146.10208, 8.2981326e-24, 5.4655313e-24, 5.3710404e-24], [146.43324, 8.2904279e-24, 5.464963e-24, 5.3713215e-24], [146.76515, 8.2828039e-24, 5.4644104e-24, 5.3716108e-24], [147.09782, 8.2752599e-24, 5.4638734e-24, 5.3719084e-24], [147.43123, 8.2677957e-24, 5.463352e-24, 5.3722143e-24], [147.76541, 8.2604106e-24, 5.4628462e-24, 5.3725285e-24], [148.10034, 8.2531044e-24, 5.4623558e-24, 5.3728511e-24], [148.43603, 8.2458764e-24, 5.461881e-24, 5.373182e-24], [148.77248, 8.2387264e-24, 5.4614216e-24, 5.3735212e-24], [149.1097, 8.2316537e-24, 5.4609775e-24, 5.3738689e-24], [149.44768, 8.2246581e-24, 5.4605488e-24, 5.3742249e-24], [149.78642, 8.217739e-24, 5.4601354e-24, 5.3745894e-24], [150.12593, 8.2108961e-24, 5.4597372e-24, 5.3749622e-24], [150.46621, 8.2041288e-24, 5.4593543e-24, 5.3753436e-24], [150.80727, 8.1974368e-24, 5.4589865e-24, 5.3757334e-24], [151.14909, 8.1908196e-24, 5.4586339e-24, 5.3761316e-24], [151.4917, 8.1842768e-24, 5.4582964e-24, 5.3765384e-24], [151.83507, 8.177808e-24, 5.4579739e-24, 5.3769537e-24], [152.17923, 8.1714128e-24, 5.4576664e-24, 5.3773776e-24], [152.52417, 8.1650907e-24, 5.457374e-24, 5.37781e-24], [152.86988, 8.1588414e-24, 5.4570965e-24, 5.378251e-24], [153.21638, 8.1526643e-24, 5.4568339e-24, 5.3787005e-24], [153.56367, 8.1465592e-24, 5.4565861e-24, 5.3791587e-24], [153.91175, 8.1405256e-24, 5.4563533e-24, 5.3796255e-24], [154.26061, 8.1345631e-24, 5.4561352e-24, 5.380101e-24], [154.61026, 8.1286712e-24, 5.4559319e-24, 5.3805851e-24], [154.96071, 8.1228497e-24, 5.4557433e-24, 5.381078e-24], [155.31195, 8.1170981e-24, 5.4555695e-24, 5.3815795e-24], [155.66399, 8.1114159e-24, 5.4554103e-24, 5.3820898e-24], [156.01682, 8.1058029e-24, 5.4552658e-24, 5.3826088e-24], [156.37046, 8.1002586e-24, 5.4551359e-24, 5.3831366e-24], [156.72489, 8.0947826e-24, 5.4550206e-24, 5.3836731e-24], [157.08013, 8.0893746e-24, 5.4549198e-24, 5.3842185e-24], [157.43618, 8.0840342e-24, 5.4548336e-24, 5.3847727e-24], [157.79303, 8.0787609e-24, 5.4547619e-24, 5.3853357e-24], [158.15069, 8.0735544e-24, 5.4547046e-24, 5.3859076e-24], [158.50916, 8.0684144e-24, 5.4546618e-24, 5.3864884e-24], [158.86844, 8.0633405e-24, 5.4546334e-24, 5.3870782e-24], [159.22854, 8.0583322e-24, 5.4546194e-24, 5.3876768e-24], [159.58946, 8.0533893e-24, 5.4546197e-24, 5.3882844e-24], [159.95119, 8.0485113e-24, 5.4546344e-24, 5.3889009e-24], [160.31374, 8.0436979e-24, 5.4546634e-24, 5.3895265e-24], [160.67712, 8.0389488e-24, 5.4547067e-24, 5.390161e-24], [161.04131, 8.0342636e-24, 5.4547643e-24, 5.3908046e-24], [161.40634, 8.0296419e-24, 5.454836e-24, 5.3914573e-24], [161.77219, 8.0250833e-24, 5.4549221e-24, 5.392119e-24], [162.13887, 8.0205876e-24, 5.4550223e-24, 5.3927898e-24], [162.50638, 8.0161543e-24, 5.4551366e-24, 5.3934698e-24], [162.87472, 8.0117832e-24, 5.4552652e-24, 5.3941588e-24], [163.2439, 8.0074739e-24, 5.4554078e-24, 5.3948571e-24], [163.61392, 8.003226e-24, 5.4555646e-24, 5.3955645e-24], [163.98477, 7.9990393e-24, 5.4557354e-24, 5.3962812e-24], [164.35647, 7.9949133e-24, 5.4559203e-24, 5.397007e-24], [164.729, 7.9908477e-24, 5.4561193e-24, 5.3977422e-24], [165.10239, 7.9868423e-24, 5.4563323e-24, 5.3984866e-24], [165.47661, 7.9828966e-24, 5.4565593e-24, 5.3992403e-24], [165.85169, 7.9790104e-24, 5.4568003e-24, 5.4000033e-24], [166.22762, 7.9751834e-24, 5.4570553e-24, 5.4007756e-24], [166.6044, 7.9714151e-24, 5.4573243e-24, 5.4015574e-24], [166.98203, 7.9677053e-24, 5.4576072e-24, 5.4023485e-24], [167.36052, 7.9640537e-24, 5.457904e-24, 5.403149e-24], [167.73986, 7.96046e-24, 5.4582147e-24, 5.403959e-24], [168.12007, 7.9569238e-24, 5.4585393e-24, 5.4047784e-24], [168.50114, 7.9534448e-24, 5.4588779e-24, 5.4056073e-24], [168.88307, 7.9500228e-24, 5.4592302e-24, 5.4064457e-24], [169.26587, 7.9466574e-24, 5.4595965e-24, 5.4072937e-24], [169.64953, 7.9433484e-24, 5.4599765e-24, 5.4081512e-24], [170.03407, 7.9400953e-24, 5.4603704e-24, 5.4090183e-24], [170.41948, 7.936898e-24, 5.4607782e-24, 5.409895e-24], [170.80576, 7.9337561e-24, 5.4611997e-24, 5.4107813e-24], [171.19291, 7.9306693e-24, 5.461635e-24, 5.4116773e-24], [171.58095, 7.9276374e-24, 5.4620841e-24, 5.4125829e-24], [171.96986, 7.92466e-24, 5.462547e-24, 5.4134983e-24], [172.35965, 7.9217369e-24, 5.4630236e-24, 5.4144234e-24], [172.75033, 7.9188678e-24, 5.463514e-24, 5.4153582e-24], [173.14189, 7.9160523e-24, 5.4640181e-24, 5.4163028e-24], [173.53435, 7.9132903e-24, 5.4645359e-24, 5.4172572e-24], [173.92769, 7.9105815e-24, 5.4650675e-24, 5.4182215e-24], [174.32192, 7.9079255e-24, 5.4656128e-24, 5.4191956e-24], [174.71704, 7.905322e-24, 5.4661717e-24, 5.4201796e-24], [175.11306, 7.9027709e-24, 5.4667444e-24, 5.4211734e-24], [175.50998, 7.9002719e-24, 5.4673308e-24, 5.4221772e-24], [175.9078, 7.8978246e-24, 5.4679309e-24, 5.423191e-24], [176.30652, 7.8954289e-24, 5.4685446e-24, 5.4242148e-24], [176.70615, 7.8930844e-24, 5.469172e-24, 5.4252485e-24], [177.10668, 7.8907909e-24, 5.4698131e-24, 5.4262923e-24], [177.50811, 7.8885482e-24, 5.4704678e-24, 5.4273462e-24], [177.91046, 7.8863559e-24, 5.4711362e-24, 5.4284101e-24], [178.31372, 7.8842139e-24, 5.4718183e-24, 5.4294842e-24], [178.7179, 7.8821218e-24, 5.472514e-24, 5.4305684e-24], [179.12299, 7.8800794e-24, 5.4732233e-24, 5.4316628e-24], [179.52899, 7.8780866e-24, 5.4739463e-24, 5.4327673e-24], [179.93592, 7.8761429e-24, 5.4746829e-24, 5.4338821e-24], [180.34377, 7.8742483e-24, 5.4754331e-24, 5.4350072e-24], [180.75255, 7.8724024e-24, 5.476197e-24, 5.4361425e-24], [181.16225, 7.870605e-24, 5.4769746e-24, 5.4372881e-24], [181.57288, 7.8688558e-24, 5.4777657e-24, 5.4384441e-24], [181.98444, 7.8671547e-24, 5.4785705e-24, 5.4396104e-24], [182.39693, 7.8655015e-24, 5.4793889e-24, 5.4407871e-24], [182.81036, 7.8638957e-24, 5.4802209e-24, 5.4419742e-24], [183.22473, 7.8623374e-24, 5.4810666e-24, 5.4431718e-24], [183.64003, 7.8608261e-24, 5.4819259e-24, 5.4443798e-24], [184.05628, 7.8593617e-24, 5.4827988e-24, 5.4455984e-24], [184.47347, 7.857944e-24, 5.4836854e-24, 5.4468274e-24], [184.89161, 7.8565727e-24, 5.4845856e-24, 5.4480671e-24], [185.31069, 7.8552477e-24, 5.4854994e-24, 5.4493173e-24], [185.73072, 7.8539686e-24, 5.4864269e-24, 5.4505781e-24], [186.15171, 7.8527354e-24, 5.487368e-24, 5.4518496e-24], [186.57365, 7.8515477e-24, 5.4883227e-24, 5.4531317e-24], [186.99654, 7.8504054e-24, 5.4892911e-24, 5.4544246e-24], [187.4204, 7.8493083e-24, 5.4902731e-24, 5.4557282e-24], [187.84521, 7.8482561e-24, 5.4912688e-24, 5.4570425e-24], [188.27099, 7.8472487e-24, 5.4922782e-24, 5.4583677e-24], [188.69774, 7.8462858e-24, 5.4933012e-24, 5.4597036e-24], [189.12545, 7.8453672e-24, 5.4943379e-24, 5.4610505e-24], [189.55413, 7.8444928e-24, 5.4953882e-24, 5.4624082e-24], [189.98378, 7.8436624e-24, 5.4964522e-24, 5.4637768e-24], [190.4144, 7.8428756e-24, 5.4975299e-24, 5.4651563e-24], [190.846, 7.8421325e-24, 5.4986213e-24, 5.4665468e-24], [191.27858, 7.8414327e-24, 5.4997264e-24, 5.4679484e-24], [191.71214, 7.840776e-24, 5.5008452e-24, 5.4693609e-24], [192.14669, 7.8401624e-24, 5.5019777e-24, 5.4707845e-24], [192.58222, 7.8395916e-24, 5.503124e-24, 5.4722193e-24], [193.01873, 7.8390633e-24, 5.5042839e-24, 5.4736651e-24], [193.45624, 7.8385775e-24, 5.5054576e-24, 5.4751221e-24], [193.89473, 7.838134e-24, 5.506645e-24, 5.4765903e-24], [194.33422, 7.8377325e-24, 5.5078462e-24, 5.4780697e-24], [194.77471, 7.8373729e-24, 5.5090611e-24, 5.4795603e-24], [195.21619, 7.837055e-24, 5.5102898e-24, 5.4810622e-24], [195.65868, 7.8367786e-24, 5.5115323e-24, 5.4825755e-24], [196.10217, 7.8365437e-24, 5.5127886e-24, 5.4841001e-24], [196.54666, 7.8363499e-24, 5.5140587e-24, 5.485636e-24], [196.99216, 7.8361972e-24, 5.5153426e-24, 5.4871834e-24], [197.43867, 7.8360853e-24, 5.5166403e-24, 5.4887422e-24], [197.8862, 7.8360141e-24, 5.5179518e-24, 5.4903125e-24], [198.33473, 7.8359835e-24, 5.5192772e-24, 5.4918943e-24], [198.78429, 7.8359932e-24, 5.5206164e-24, 5.4934876e-24], [199.23486, 7.8360431e-24, 5.5219695e-24, 5.4950925e-24], [199.68646, 7.8361332e-24, 5.5233365e-24, 5.496709e-24], [200.13907, 7.8362631e-24, 5.5247174e-24, 5.4983372e-24], [200.59272, 7.8364327e-24, 5.5261122e-24, 5.499977e-24], [201.04739, 7.8366419e-24, 5.5275209e-24, 5.5016285e-24], [201.50309, 7.8368906e-24, 5.5289435e-24, 5.5032917e-24], [201.95983, 7.8371786e-24, 5.5303801e-24, 5.5049667e-24], [202.4176, 7.8375057e-24, 5.5318306e-24, 5.5066536e-24], [202.87641, 7.8378718e-24, 5.5332951e-24, 5.5083522e-24], [203.33626, 7.8382767e-24, 5.5347736e-24, 5.5100627e-24], [203.79715, 7.8387204e-24, 5.5362662e-24, 5.5117851e-24], [204.25908, 7.8392026e-24, 5.5377727e-24, 5.5135195e-24], [204.72207, 7.8397232e-24, 5.5392933e-24, 5.5152658e-24], [205.1861, 7.8402822e-24, 5.5408279e-24, 5.5170242e-24], [205.65118, 7.8408792e-24, 5.5423766e-24, 5.5187945e-24], [206.11732, 7.8415143e-24, 5.5439393e-24, 5.520577e-24], [206.58451, 7.8421873e-24, 5.5455162e-24, 5.5223715e-24], [207.05277, 7.842898e-24, 5.5471072e-24, 5.5241782e-24], [207.52208, 7.8436464e-24, 5.5487123e-24, 5.5259971e-24], [207.99246, 7.8444322e-24, 5.5503316e-24, 5.5278282e-24], [208.46391, 7.8452554e-24, 5.5519651e-24, 5.5296715e-24], [208.93642, 7.8461158e-24, 5.5536127e-24, 5.5315271e-24], [209.41, 7.8470133e-24, 5.5552746e-24, 5.533395e-24], [209.88466, 7.8479478e-24, 5.5569506e-24, 5.5352753e-24], [210.3604, 7.8489192e-24, 5.558641e-24, 5.537168e-24], [210.83721, 7.8499273e-24, 5.5603455e-24, 5.5390731e-24], [211.3151, 7.8509721e-24, 5.5620644e-24, 5.5409906e-24], [211.79408, 7.8520534e-24, 5.5637976e-24, 5.5429206e-24], [212.27414, 7.853171e-24, 5.5655451e-24, 5.5448632e-24], [212.75529, 7.8543249e-24, 5.5673069e-24, 5.5468184e-24], [213.23753, 7.855515e-24, 5.5690831e-24, 5.5487861e-24], [213.72086, 7.8567412e-24, 5.5708737e-24, 5.5507665e-24], [214.20529, 7.8580033e-24, 5.5726786e-24, 5.5527596e-24], [214.69082, 7.8593013e-24, 5.5744981e-24, 5.5547653e-24], [215.17745, 7.8606349e-24, 5.5763319e-24, 5.5567838e-24], [215.66518, 7.8620042e-24, 5.5781802e-24, 5.5588152e-24], [216.15401, 7.863409e-24, 5.5800431e-24, 5.5608593e-24], [216.64396, 7.8648493e-24, 5.5819204e-24, 5.5629163e-24], [217.13501, 7.8663248e-24, 5.5838122e-24, 5.5649862e-24], [217.62718, 7.8678356e-24, 5.5857187e-24, 5.567069e-24], [218.12046, 7.8693815e-24, 5.5876397e-24, 5.5691648e-24], [218.61486, 7.8709624e-24, 5.5895753e-24, 5.5712736e-24], [219.11039, 7.8725782e-24, 5.5915255e-24, 5.5733954e-24], [219.60703, 7.8742289e-24, 5.5934904e-24, 5.5755303e-24], [220.1048, 7.8759143e-24, 5.59547e-24, 5.5776784e-24], [220.6037, 7.8776344e-24, 5.5974642e-24, 5.5798396e-24], [221.10373, 7.879389e-24, 5.5994732e-24, 5.582014e-24], [221.60489, 7.8811781e-24, 5.6014969e-24, 5.5842017e-24], [222.10719, 7.8830015e-24, 5.6035354e-24, 5.5864026e-24], [222.61063, 7.8848593e-24, 5.6055887e-24, 5.5886168e-24], [223.11521, 7.8867512e-24, 5.6076569e-24, 5.5908444e-24], [223.62093, 7.8886773e-24, 5.6097398e-24, 5.5930854e-24], [224.1278, 7.8906374e-24, 5.6118377e-24, 5.5953398e-24], [224.63582, 7.8926315e-24, 5.6139504e-24, 5.5976077e-24], [225.14499, 7.8946594e-24, 5.6160781e-24, 5.599889e-24], [225.65531, 7.8967212e-24, 5.6182207e-24, 5.602184e-24], [226.16679, 7.8988166e-24, 5.6203783e-24, 5.6044925e-24], [226.67943, 7.9009458e-24, 5.6225509e-24, 5.6068146e-24], [227.19323, 7.9031084e-24, 5.6247386e-24, 5.6091504e-24], [227.7082, 7.9053046e-24, 5.6269413e-24, 5.6114999e-24], [228.22433, 7.9075342e-24, 5.6291591e-24, 5.6138631e-24], [228.74164, 7.9097972e-24, 5.631392e-24, 5.6162402e-24], [229.26011, 7.9120934e-24, 5.6336401e-24, 5.618631e-24], [229.77976, 7.9144229e-24, 5.6359033e-24, 5.6210357e-24], [230.30059, 7.9167855e-24, 5.6381817e-24, 5.6234544e-24], [230.8226, 7.9191811e-24, 5.6404754e-24, 5.6258869e-24], [231.3458, 7.9216098e-24, 5.6427843e-24, 5.6283335e-24], [231.87017, 7.9240715e-24, 5.6451086e-24, 5.630794e-24], [232.39574, 7.926566e-24, 5.6474481e-24, 5.6332687e-24], [232.9225, 7.9290933e-24, 5.649803e-24, 5.6357574e-24], [233.45045, 7.9316534e-24, 5.6521732e-24, 5.6382603e-24], [233.9796, 7.9342462e-24, 5.6545589e-24, 5.6407774e-24], [234.50995, 7.9368716e-24, 5.65696e-24, 5.6433087e-24], [235.0415, 7.9395297e-24, 5.6593766e-24, 5.6458543e-24], [235.57425, 7.9422202e-24, 5.6618086e-24, 5.6484142e-24], [236.10822, 7.9449433e-24, 5.6642562e-24, 5.6509884e-24], [236.64339, 7.9476987e-24, 5.6667194e-24, 5.653577e-24], [237.17978, 7.9504865e-24, 5.6691981e-24, 5.6561801e-24], [237.71738, 7.9533066e-24, 5.6716925e-24, 5.6587977e-24], [238.2562, 7.956159e-24, 5.6742025e-24, 5.6614298e-24], [238.79624, 7.9590436e-24, 5.6767282e-24, 5.6640764e-24], [239.33751, 7.9619604e-24, 5.6792697e-24, 5.6667376e-24], [239.88, 7.9649093e-24, 5.6818268e-24, 5.6694135e-24], [240.42372, 7.9678902e-24, 5.6843998e-24, 5.6721041e-24], [240.96868, 7.9709032e-24, 5.6869886e-24, 5.6748094e-24], [241.51487, 7.9739481e-24, 5.6895932e-24, 5.6775295e-24], [242.0623, 7.977025e-24, 5.6922137e-24, 5.6802644e-24], [242.61097, 7.9801338e-24, 5.6948502e-24, 5.6830142e-24], [243.16088, 7.9832744e-24, 5.6975026e-24, 5.6857788e-24], [243.71204, 7.9864468e-24, 5.7001709e-24, 5.6885585e-24], [244.26445, 7.989651e-24, 5.7028553e-24, 5.6913531e-24], [244.81811, 7.992887e-24, 5.7055557e-24, 5.6941627e-24], [245.37302, 7.9961546e-24, 5.7082723e-24, 5.6969874e-24], [245.92919, 7.9994539e-24, 5.7110049e-24, 5.6998273e-24], [246.48663, 8.0027848e-24, 5.7137537e-24, 5.7026823e-24], [247.04533, 8.0061473e-24, 5.7165187e-24, 5.7055525e-24], [247.60529, 8.0095414e-24, 5.7192999e-24, 5.7084379e-24], [248.16652, 8.012967e-24, 5.7220973e-24, 5.7113387e-24], [248.72903, 8.016424e-24, 5.7249111e-24, 5.7142548e-24], [249.29281, 8.0199126e-24, 5.7277412e-24, 5.7171863e-24], [249.85787, 8.0234325e-24, 5.7305877e-24, 5.7201332e-24], [250.4242, 8.0269839e-24, 5.7334505e-24, 5.7230956e-24], [250.99183, 8.0305666e-24, 5.7363298e-24, 5.7260735e-24], [251.56074, 8.0341807e-24, 5.7392256e-24, 5.7290669e-24], [252.13093, 8.0378261e-24, 5.7421379e-24, 5.732076e-24], [252.70243, 8.0415028e-24, 5.7450668e-24, 5.7351007e-24], [253.27521, 8.0452107e-24, 5.7480122e-24, 5.7381412e-24], [253.8493, 8.0489499e-24, 5.7509742e-24, 5.7411973e-24], [254.42468, 8.0527204e-24, 5.753953e-24, 5.7442693e-24], [255.00137, 8.056522e-24, 5.7569484e-24, 5.7473571e-24], [255.57937, 8.0603548e-24, 5.7599605e-24, 5.7504607e-24], [256.15868, 8.0642188e-24, 5.7629894e-24, 5.7535803e-24], [256.7393, 8.0681139e-24, 5.7660352e-24, 5.7567159e-24], [257.32123, 8.0720402e-24, 5.7690978e-24, 5.7598675e-24], [257.90449, 8.0759975e-24, 5.7721772e-24, 5.7630351e-24], [258.48907, 8.079986e-24, 5.7752737e-24, 5.7662189e-24], [259.07497, 8.0840055e-24, 5.778387e-24, 5.7694188e-24], [259.6622, 8.0880561e-24, 5.7815174e-24, 5.7726349e-24], [260.25076, 8.0921377e-24, 5.7846648e-24, 5.7758673e-24], [260.84066, 8.0962504e-24, 5.7878293e-24, 5.7791159e-24], [261.43189, 8.1003941e-24, 5.791011e-24, 5.7823809e-24], [262.02446, 8.1045688e-24, 5.7942098e-24, 5.7856623e-24], [262.61838, 8.1087745e-24, 5.7974258e-24, 5.7889601e-24], [263.21364, 8.1130112e-24, 5.800659e-24, 5.7922744e-24], [263.81025, 8.1172789e-24, 5.8039095e-24, 5.7956053e-24], [264.40822, 8.1215775e-24, 5.8071774e-24, 5.7989527e-24], [265.00754, 8.1259072e-24, 5.8104626e-24, 5.8023167e-24], [265.60821, 8.1302678e-24, 5.8137652e-24, 5.8056974e-24], [266.21025, 8.1346593e-24, 5.8170853e-24, 5.8090948e-24], [266.81366, 8.1390818e-24, 5.8204228e-24, 5.812509e-24], [267.41843, 8.1435352e-24, 5.8237779e-24, 5.81594e-24], [268.02457, 8.1480196e-24, 5.8271506e-24, 5.8193879e-24], [268.63209, 8.152535e-24, 5.8305408e-24, 5.8228527e-24], [269.24098, 8.1570812e-24, 5.8339488e-24, 5.8263344e-24], [269.85125, 8.1616585e-24, 5.8373744e-24, 5.8298331e-24], [270.46291, 8.1662666e-24, 5.8408177e-24, 5.8333489e-24], [271.07595, 8.1709057e-24, 5.8442789e-24, 5.8368819e-24], [271.69039, 8.1755757e-24, 5.8477578e-24, 5.8404319e-24], [272.30621, 8.1802767e-24, 5.8512547e-24, 5.8439992e-24], [272.92343, 8.1850086e-24, 5.8547694e-24, 5.8475837e-24], [273.54205, 8.1897715e-24, 5.8583021e-24, 5.8511855e-24], [274.16208, 8.1945653e-24, 5.8618528e-24, 5.8548046e-24], [274.7835, 8.1993901e-24, 5.8654215e-24, 5.8584412e-24], [275.40634, 8.2042458e-24, 5.8690083e-24, 5.8620952e-24], [276.03059, 8.2091325e-24, 5.8726132e-24, 5.8657667e-24], [276.65625, 8.2140502e-24, 5.8762363e-24, 5.8694557e-24], [277.28333, 8.2189989e-24, 5.8798776e-24, 5.8731624e-24], [277.91183, 8.2239786e-24, 5.8835372e-24, 5.8768867e-24], [278.54176, 8.2289892e-24, 5.8872151e-24, 5.8806286e-24], [279.17312, 8.2340309e-24, 5.8909113e-24, 5.8843884e-24], [279.8059, 8.2391036e-24, 5.8946259e-24, 5.8881659e-24], [280.44012, 8.2442074e-24, 5.8983589e-24, 5.8919612e-24], [281.07578, 8.2493422e-24, 5.9021104e-24, 5.8957745e-24], [281.71288, 8.254508e-24, 5.9058805e-24, 5.8996057e-24], [282.35142, 8.2597049e-24, 5.9096691e-24, 5.9034549e-24], [282.99141, 8.2649329e-24, 5.9134763e-24, 5.9073222e-24], [283.63285, 8.270192e-24, 5.9173022e-24, 5.9112075e-24], [284.27575, 8.2754823e-24, 5.9211468e-24, 5.915111e-24], [284.9201, 8.2808036e-24, 5.9250101e-24, 5.9190327e-24], [285.56591, 8.2861561e-24, 5.9288923e-24, 5.9229727e-24], [286.21319, 8.2915398e-24, 5.9327933e-24, 5.9269309e-24], [286.86193, 8.2969546e-24, 5.9367132e-24, 5.9309075e-24], [287.51214, 8.3024007e-24, 5.940652e-24, 5.9349025e-24], [288.16383, 8.3078779e-24, 5.9446098e-24, 5.938916e-24], [288.817, 8.3133865e-24, 5.9485866e-24, 5.9429479e-24], [289.47164, 8.3189262e-24, 5.9525826e-24, 5.9469985e-24], [290.12777, 8.3244973e-24, 5.9565976e-24, 5.9510676e-24], [290.78539, 8.3300996e-24, 5.9606318e-24, 5.9551554e-24], [291.44449, 8.3357333e-24, 5.9646853e-24, 5.9592619e-24], [292.1051, 8.3413984e-24, 5.968758e-24, 5.9633872e-24], [292.76719, 8.3470948e-24, 5.97285e-24, 5.9675313e-24], [293.43079, 8.3528226e-24, 5.9769614e-24, 5.9716943e-24], [294.0959, 8.3585818e-24, 5.9810921e-24, 5.9758762e-24], [294.76251, 8.3643724e-24, 5.9852424e-24, 5.9800771e-24], [295.43063, 8.3701946e-24, 5.9894121e-24, 5.984297e-24], [296.10026, 8.3760482e-24, 5.9936015e-24, 5.988536e-24], [296.77142, 8.3819334e-24, 5.9978104e-24, 5.9927941e-24], [297.44409, 8.3878501e-24, 6.0020389e-24, 5.9970715e-24], [298.11829, 8.3937983e-24, 6.0062872e-24, 6.001368e-24], [298.79402, 8.3997782e-24, 6.0105552e-24, 6.0056839e-24], [299.47128, 8.4057898e-24, 6.0148431e-24, 6.0100191e-24], [300.15007, 8.411833e-24, 6.0191507e-24, 6.0143737e-24], [300.83041, 8.4179079e-24, 6.0234783e-24, 6.0187477e-24], [301.51228, 8.4240145e-24, 6.0278258e-24, 6.0231413e-24], [302.1957, 8.4301529e-24, 6.0321934e-24, 6.0275544e-24], [302.88067, 8.4363231e-24, 6.0365809e-24, 6.0319872e-24], [303.5672, 8.4425251e-24, 6.0409886e-24, 6.0364396e-24], [304.25528, 8.4487589e-24, 6.0454164e-24, 6.0409117e-24], [304.94491, 8.4550247e-24, 6.0498644e-24, 6.0454037e-24], [305.63612, 8.4613223e-24, 6.0543327e-24, 6.0499154e-24], [306.32888, 8.467652e-24, 6.0588212e-24, 6.0544471e-24], [307.02322, 8.4740136e-24, 6.0633301e-24, 6.0589987e-24], [307.71913, 8.4804072e-24, 6.0678594e-24, 6.0635702e-24], [308.41662, 8.486833e-24, 6.0724092e-24, 6.0681619e-24], [309.11569, 8.4932908e-24, 6.0769795e-24, 6.0727736e-24], [309.81635, 8.4997808e-24, 6.0815703e-24, 6.0774055e-24], [310.51859, 8.5063029e-24, 6.0861817e-24, 6.0820576e-24], [311.22243, 8.5128573e-24, 6.0908138e-24, 6.08673e-24], [311.92786, 8.5194439e-24, 6.0954666e-24, 6.0914227e-24], [312.63489, 8.5260628e-24, 6.1001402e-24, 6.0961358e-24], [313.34352, 8.532714e-24, 6.1048345e-24, 6.1008693e-24], [314.05376, 8.5393977e-24, 6.1095497e-24, 6.1056234e-24], [314.76561, 8.5461137e-24, 6.1142859e-24, 6.1103979e-24], [315.47907, 8.5528622e-24, 6.119043e-24, 6.1151931e-24], [316.19415, 8.5596432e-24, 6.1238212e-24, 6.120009e-24], [316.91085, 8.5664567e-24, 6.1286204e-24, 6.1248455e-24], [317.62917, 8.5733028e-24, 6.1334407e-24, 6.1297028e-24], [318.34912, 8.5801815e-24, 6.1382822e-24, 6.134581e-24], [319.07071, 8.5870929e-24, 6.143145e-24, 6.13948e-24], [319.79393, 8.5940371e-24, 6.148029e-24, 6.1444e-24], [320.51878, 8.6010139e-24, 6.1529344e-24, 6.149341e-24], [321.24529, 8.6080236e-24, 6.1578612e-24, 6.154303e-24], [321.97343, 8.6150661e-24, 6.1628094e-24, 6.1592862e-24], [322.70323, 8.6221416e-24, 6.1677791e-24, 6.1642905e-24], [323.43469, 8.6292499e-24, 6.1727704e-24, 6.169316e-24], [324.1678, 8.6363913e-24, 6.1777833e-24, 6.1743628e-24], [324.90257, 8.6435657e-24, 6.1828179e-24, 6.179431e-24], [325.63901, 8.6507731e-24, 6.1878741e-24, 6.1845205e-24], [326.37712, 8.6580137e-24, 6.1929522e-24, 6.1896316e-24], [327.1169, 8.6652875e-24, 6.198052e-24, 6.1947641e-24], [327.85835, 8.6725945e-24, 6.2031738e-24, 6.1999182e-24], [328.60149, 8.6799348e-24, 6.2083174e-24, 6.2050939e-24], [329.34632, 8.6873084e-24, 6.2134831e-24, 6.2102913e-24], [330.09283, 8.6947154e-24, 6.2186708e-24, 6.2155104e-24], [330.84103, 8.7021558e-24, 6.2238806e-24, 6.2207513e-24], [331.59093, 8.7096297e-24, 6.2291126e-24, 6.2260141e-24], [332.34253, 8.7171372e-24, 6.2343668e-24, 6.2312988e-24], [333.09583, 8.7246782e-24, 6.2396432e-24, 6.2366055e-24], [333.85084, 8.7322528e-24, 6.244942e-24, 6.2419342e-24], [334.60756, 8.7398612e-24, 6.2502631e-24, 6.247285e-24], [335.366, 8.7475033e-24, 6.2556067e-24, 6.252658e-24], [336.12615, 8.7551792e-24, 6.2609728e-24, 6.2580531e-24], [336.88803, 8.7628889e-24, 6.2663614e-24, 6.2634706e-24], [337.65163, 8.7706325e-24, 6.2717726e-24, 6.2689103e-24], [338.41697, 8.7784101e-24, 6.2772064e-24, 6.2743724e-24], [339.18404, 8.7862217e-24, 6.282663e-24, 6.279857e-24], [339.95285, 8.7940674e-24, 6.2881424e-24, 6.2853641e-24], [340.7234, 8.8019472e-24, 6.2936446e-24, 6.2908937e-24], [341.4957, 8.8098612e-24, 6.2991696e-24, 6.296446e-24], [342.26975, 8.8178095e-24, 6.3047176e-24, 6.302021e-24], [343.04555, 8.825792e-24, 6.3102886e-24, 6.3076186e-24], [343.82312, 8.8338089e-24, 6.3158827e-24, 6.3132391e-24], [344.60244, 8.8418602e-24, 6.3214999e-24, 6.3188825e-24], [345.38353, 8.849946e-24, 6.3271402e-24, 6.3245487e-24], [346.16639, 8.8580663e-24, 6.3328038e-24, 6.330238e-24], [346.95103, 8.8662212e-24, 6.3384907e-24, 6.3359503e-24], [347.73744, 8.8744107e-24, 6.3442009e-24, 6.3416856e-24], [348.52564, 8.882635e-24, 6.3499345e-24, 6.3474442e-24], [349.31562, 8.890894e-24, 6.3556916e-24, 6.3532259e-24], [350.1074, 8.8991879e-24, 6.3614722e-24, 6.359031e-24], [350.90097, 8.9075167e-24, 6.3672763e-24, 6.3648594e-24], [351.69633, 8.9158804e-24, 6.3731041e-24, 6.3707112e-24], [352.4935, 8.9242791e-24, 6.3789556e-24, 6.3765864e-24], [353.29248, 8.9327129e-24, 6.3848309e-24, 6.3824852e-24], [354.09327, 8.9411818e-24, 6.3907299e-24, 6.3884075e-24], [354.89587, 8.949686e-24, 6.3966528e-24, 6.3943535e-24], [355.70029, 8.9582254e-24, 6.4025997e-24, 6.4003232e-24], [356.50654, 8.9668002e-24, 6.4085706e-24, 6.4063167e-24], [357.31461, 8.9754103e-24, 6.4145655e-24, 6.412334e-24], [358.12452, 8.9840559e-24, 6.4205845e-24, 6.4183752e-24], [358.93626, 8.992737e-24, 6.4266276e-24, 6.4244404e-24], [359.74984, 9.0014537e-24, 6.432695e-24, 6.4305295e-24], [360.56526, 9.0102061e-24, 6.4387867e-24, 6.4366428e-24], [361.38254, 9.0189942e-24, 6.4449027e-24, 6.4427802e-24], [362.20166, 9.0278181e-24, 6.4510432e-24, 6.4489417e-24], [363.02265, 9.0366778e-24, 6.4572081e-24, 6.4551276e-24], [363.84549, 9.0455734e-24, 6.4633975e-24, 6.4613377e-24], [364.6702, 9.0545051e-24, 6.4696115e-24, 6.4675723e-24], [365.49677, 9.0634727e-24, 6.4758502e-24, 6.4738313e-24], [366.32522, 9.0724765e-24, 6.4821136e-24, 6.4801148e-24], [367.15555, 9.0815165e-24, 6.4884017e-24, 6.4864229e-24], [367.98776, 9.0905928e-24, 6.4947147e-24, 6.4927556e-24], [368.82186, 9.0997054e-24, 6.5010526e-24, 6.4991131e-24], [369.65785, 9.1088543e-24, 6.5074154e-24, 6.5054952e-24], [370.49573, 9.1180398e-24, 6.5138032e-24, 6.5119023e-24], [371.33551, 9.1272618e-24, 6.5202161e-24, 6.5183342e-24], [372.1772, 9.1365203e-24, 6.5266542e-24, 6.524791e-24], [373.02079, 9.1458156e-24, 6.5331174e-24, 6.5312729e-24], [373.8663, 9.1551476e-24, 6.5396059e-24, 6.5377798e-24], [374.71372, 9.1645164e-24, 6.5461198e-24, 6.5443119e-24], [375.56306, 9.1739221e-24, 6.552659e-24, 6.5508692e-24], [376.41433, 9.1833648e-24, 6.5592236e-24, 6.5574518e-24], [377.26753, 9.1928445e-24, 6.5658138e-24, 6.5640597e-24], [378.12266, 9.2023613e-24, 6.5724295e-24, 6.570693e-24], [378.97973, 9.2119153e-24, 6.5790709e-24, 6.5773518e-24], [379.83874, 9.2215066e-24, 6.585738e-24, 6.5840361e-24], [380.6997, 9.2311351e-24, 6.5924308e-24, 6.590746e-24], [381.56261, 9.2408011e-24, 6.5991494e-24, 6.5974815e-24], [382.42747, 9.2505045e-24, 6.605894e-24, 6.6042428e-24], [383.2943, 9.2602455e-24, 6.6126645e-24, 6.6110298e-24], [384.16309, 9.2700241e-24, 6.619461e-24, 6.6178427e-24], [385.03385, 9.2798404e-24, 6.6262835e-24, 6.6246815e-24], [385.90659, 9.2896945e-24, 6.6331322e-24, 6.6315463e-24], [386.7813, 9.2995864e-24, 6.6400071e-24, 6.6384371e-24], [387.658, 9.3095162e-24, 6.6469083e-24, 6.6453541e-24], [388.53668, 9.3194841e-24, 6.6538358e-24, 6.6522972e-24], [389.41735, 9.32949e-24, 6.6607897e-24, 6.6592666e-24], [390.30002, 9.339534e-24, 6.6677701e-24, 6.6662622e-24], [391.1847, 9.3496163e-24, 6.674777e-24, 6.6732843e-24], [392.07137, 9.3597369e-24, 6.6818104e-24, 6.6803328e-24], [392.96006, 9.3698959e-24, 6.6888705e-24, 6.6874078e-24], [393.85076, 9.3800933e-24, 6.6959574e-24, 6.6945093e-24], [394.74348, 9.3903293e-24, 6.703071e-24, 6.7016375e-24], [395.63822, 9.4006039e-24, 6.7102115e-24, 6.7087924e-24], [396.53499, 9.4109172e-24, 6.7173788e-24, 6.7159741e-24], [397.43379, 9.4212693e-24, 6.7245732e-24, 6.7231826e-24], [398.33464, 9.4316602e-24, 6.7317946e-24, 6.730418e-24], [399.23752, 9.4420901e-24, 6.7390431e-24, 6.7376804e-24], [400.14245, 9.452559e-24, 6.7463188e-24, 6.7449699e-24], [401.04943, 9.463067e-24, 6.7536217e-24, 6.7522864e-24], [401.95846, 9.4736141e-24, 6.760952e-24, 6.7596301e-24], [402.86956, 9.4842005e-24, 6.7683096e-24, 6.7670011e-24], [403.78272, 9.4948263e-24, 6.7756947e-24, 6.7743993e-24], [404.69795, 9.5054915e-24, 6.7831072e-24, 6.781825e-24], [405.61526, 9.5161962e-24, 6.7905474e-24, 6.7892781e-24], [406.53465, 9.5269404e-24, 6.7980152e-24, 6.7967587e-24], [407.45611, 9.5377244e-24, 6.8055107e-24, 6.8042669e-24], [408.37967, 9.5485481e-24, 6.8130339e-24, 6.8118028e-24], [409.30532, 9.5594116e-24, 6.8205851e-24, 6.8193663e-24], [410.23307, 9.5703151e-24, 6.8281641e-24, 6.8269577e-24], [411.16293, 9.5812586e-24, 6.8357711e-24, 6.834577e-24], [412.09489, 9.5922422e-24, 6.8434062e-24, 6.8422241e-24], [413.02896, 9.6032659e-24, 6.8510694e-24, 6.8498993e-24], [413.96515, 9.6143299e-24, 6.8587608e-24, 6.8576026e-24], [414.90346, 9.6254343e-24, 6.8664805e-24, 6.865334e-24], [415.8439, 9.636579e-24, 6.8742285e-24, 6.8730936e-24], [416.78647, 9.6477644e-24, 6.8820049e-24, 6.8808815e-24], [417.73117, 9.6589903e-24, 6.8898098e-24, 6.8886978e-24], [418.67802, 9.6702569e-24, 6.8976433e-24, 6.8965425e-24], [419.62701, 9.6815643e-24, 6.9055053e-24, 6.9044158e-24], [420.57816, 9.6929126e-24, 6.9133961e-24, 6.9123176e-24], [421.53146, 9.7043019e-24, 6.9213156e-24, 6.920248e-24], [422.48692, 9.7157322e-24, 6.929264e-24, 6.9282073e-24], [423.44455, 9.7272037e-24, 6.9372413e-24, 6.9361953e-24], [424.40435, 9.7387164e-24, 6.9452476e-24, 6.9442122e-24], [425.36632, 9.7502704e-24, 6.953283e-24, 6.9522581e-24], [426.33047, 9.7618659e-24, 6.9613475e-24, 6.960333e-24], [427.29681, 9.7735029e-24, 6.9694412e-24, 6.9684371e-24], [428.26534, 9.7851815e-24, 6.9775642e-24, 6.9765703e-24], [429.23607, 9.7969018e-24, 6.9857167e-24, 6.9847329e-24], [430.20899, 9.8086639e-24, 6.9938986e-24, 6.9929248e-24], [431.18412, 9.820468e-24, 7.0021101e-24, 7.0011462e-24], [432.16147, 9.832314e-24, 7.0103512e-24, 7.0093971e-24], [433.14102, 9.8442022e-24, 7.018622e-24, 7.0176776e-24], [434.1228, 9.8561326e-24, 7.0269227e-24, 7.0259879e-24], [435.1068, 9.8681053e-24, 7.0352532e-24, 7.034328e-24], [436.09303, 9.8801205e-24, 7.0436138e-24, 7.042698e-24], [437.0815, 9.8921782e-24, 7.0520045e-24, 7.0510981e-24], [438.07221, 9.9042785e-24, 7.0604254e-24, 7.0595282e-24], [439.06516, 9.9164217e-24, 7.0688766e-24, 7.0679886e-24], [440.06037, 9.9286077e-24, 7.0773583e-24, 7.0764793e-24], [441.05783, 9.9408367e-24, 7.0858705e-24, 7.0850005e-24], [442.05755, 9.9531089e-24, 7.0944134e-24, 7.0935522e-24], [443.05953, 9.9654244e-24, 7.1029871e-24, 7.1021347e-24], [444.06379, 9.9777833e-24, 7.1115917e-24, 7.1107481e-24], [445.07033, 9.9901857e-24, 7.1202274e-24, 7.1193924e-24], [446.07914, 1.0002632e-23, 7.1288944e-24, 7.128068e-24], [447.09024, 1.0015122e-23, 7.1375929e-24, 7.1367748e-24], [448.10364, 1.0027656e-23, 7.146323e-24, 7.1455133e-24], [449.11933, 1.0040235e-23, 7.1550849e-24, 7.1542835e-24], [450.13732, 1.0052858e-23, 7.163879e-24, 7.1630858e-24], [451.15762, 1.0065525e-23, 7.1727054e-24, 7.1719204e-24], [452.18024, 1.0078238e-23, 7.1815646e-24, 7.1807875e-24], [453.20517, 1.0090996e-23, 7.1904568e-24, 7.1896877e-24], [454.23242, 1.01038e-23, 7.1993825e-24, 7.1986213e-24], [455.26201, 1.011665e-23, 7.2083421e-24, 7.2075887e-24], [456.29392, 1.0129546e-23, 7.2173364e-24, 7.2165907e-24], [457.32818, 1.0142489e-23, 7.2263659e-24, 7.2256279e-24], [458.36478, 1.015548e-23, 7.2354317e-24, 7.2347012e-24], [459.40373, 1.016852e-23, 7.2445346e-24, 7.2438116e-24], [460.44503, 1.0181609e-23, 7.2536761e-24, 7.2529606e-24], [461.4887, 1.0194749e-23, 7.2628579e-24, 7.2621498e-24], [462.53473, 1.0207942e-23, 7.2720823e-24, 7.2713813e-24], [463.58313, 1.0221188e-23, 7.281352e-24, 7.2806583e-24], [464.63391, 1.0234493e-23, 7.290671e-24, 7.2899844e-24], [465.68707, 1.0247858e-23, 7.3000446e-24, 7.2993651e-24], [466.74261, 1.026129e-23, 7.3094802e-24, 7.3088076e-24], [467.80055, 1.0274796e-23, 7.3189884e-24, 7.3183228e-24], [468.86089, 1.0288387e-23, 7.328585e-24, 7.3279263e-24], [469.92363, 1.0302081e-23, 7.3382941e-24, 7.3376422e-24], [470.98878, 1.0315905e-23, 7.3481543e-24, 7.3475091e-24], [472.05634, 1.0329906e-23, 7.3582302e-24, 7.3575917e-24], [473.12633, 1.0344165e-23, 7.3686376e-24, 7.3680058e-24], [474.19874, 1.0358842e-23, 7.379601e-24, 7.3789758e-24], [475.27358, 1.0374283e-23, 7.3916033e-24, 7.3909847e-24], [476.35085, 1.0391347e-23, 7.4058505e-24, 7.4052387e-24], [477.43057, 1.0412718e-23, 7.4261062e-24, 7.4255016e-24], [478.51274, 1.0450754e-23, 7.4696566e-24, 7.4690609e-24], [479.59735, 1.0648976e-23, 7.7351589e-24, 7.7345888e-24], [480.68443, 1.1939038e-23, 9.4247775e-24, 9.4243138e-24], [481.77397, 1.0524153e-23, 7.5429907e-24, 7.5424166e-24], [482.86598, 1.0488183e-23, 7.4828695e-24, 7.482296e-24], [483.96047, 1.0490462e-23, 7.4761559e-24, 7.4755871e-24], [485.05743, 1.0500055e-23, 7.4796619e-24, 7.4790985e-24], [486.15688, 1.0512019e-23, 7.4864609e-24, 7.4859031e-24], [487.25883, 1.0524994e-23, 7.4946482e-24, 7.494096e-24], [488.36327, 1.0538488e-23, 7.5035305e-24, 7.5029839e-24], [489.47022, 1.0552286e-23, 7.5128068e-24, 7.5122658e-24], [490.57967, 1.0566283e-23, 7.5223294e-24, 7.521794e-24], [491.69164, 1.0580422e-23, 7.5320192e-24, 7.5314893e-24], [492.80613, 1.0594672e-23, 7.5418307e-24, 7.5413062e-24], [493.92315, 1.0609011e-23, 7.551736e-24, 7.551217e-24], [495.04269, 1.0623429e-23, 7.5617178e-24, 7.5612041e-24], [496.16478, 1.0637916e-23, 7.5717643e-24, 7.5712559e-24], [497.28941, 1.0652468e-23, 7.5818677e-24, 7.5813646e-24], [498.41659, 1.0667079e-23, 7.5920226e-24, 7.5915246e-24], [499.54632, 1.0681749e-23, 7.6022249e-24, 7.6017321e-24], [500.67861, 1.0696473e-23, 7.6124719e-24, 7.6119842e-24], [501.81347, 1.0711252e-23, 7.6227615e-24, 7.6222788e-24], [502.9509, 1.0726085e-23, 7.633092e-24, 7.6326144e-24], [504.09091, 1.0740969e-23, 7.6434624e-24, 7.6429897e-24], [505.23351, 1.0755905e-23, 7.6538716e-24, 7.6534039e-24], [506.37869, 1.0770892e-23, 7.6643191e-24, 7.6638562e-24], [507.52647, 1.078593e-23, 7.6748042e-24, 7.6743461e-24], [508.67686, 1.0801019e-23, 7.6853265e-24, 7.6848732e-24], [509.82984, 1.0816158e-23, 7.6958858e-24, 7.6954371e-24], [510.98545, 1.0831348e-23, 7.7064817e-24, 7.7060377e-24], [512.14367, 1.0846588e-23, 7.7171141e-24, 7.7166747e-24], [513.30452, 1.0861878e-23, 7.7277828e-24, 7.7273479e-24], [514.46799, 1.0877218e-23, 7.7384877e-24, 7.7380573e-24], [515.63411, 1.0892608e-23, 7.7492287e-24, 7.7488028e-24], [516.80287, 1.0908048e-23, 7.7600058e-24, 7.7595843e-24], [517.97428, 1.0923539e-23, 7.7708189e-24, 7.7704019e-24], [519.14834, 1.093908e-23, 7.781668e-24, 7.7812553e-24], [520.32506, 1.0954671e-23, 7.7925532e-24, 7.7921448e-24], [521.50446, 1.0970313e-23, 7.8034744e-24, 7.8030702e-24], [522.68652, 1.0986004e-23, 7.8144316e-24, 7.8140316e-24], [523.87127, 1.1001747e-23, 7.8254248e-24, 7.825029e-24], [525.05869, 1.101754e-23, 7.8364542e-24, 7.8360625e-24], [526.24882, 1.1033383e-23, 7.8475197e-24, 7.8471321e-24], [527.44163, 1.1049277e-23, 7.8586213e-24, 7.8582377e-24], [528.63716, 1.1065222e-23, 7.8697592e-24, 7.8693796e-24], [529.83539, 1.1081218e-23, 7.8809333e-24, 7.8805577e-24], [531.03634, 1.1097265e-23, 7.8921437e-24, 7.891772e-24], [532.24001, 1.1113362e-23, 7.9033906e-24, 7.9030227e-24], [533.44641, 1.1129511e-23, 7.9146738e-24, 7.9143098e-24], [534.65554, 1.1145711e-23, 7.9259936e-24, 7.9256334e-24], [535.86741, 1.1161963e-23, 7.9373499e-24, 7.9369934e-24], [537.08203, 1.1178266e-23, 7.9487428e-24, 7.9483901e-24], [538.29941, 1.119462e-23, 7.9601725e-24, 7.9598234e-24], [539.51954, 1.1211026e-23, 7.9716388e-24, 7.9712935e-24], [540.74244, 1.1227484e-23, 7.9831421e-24, 7.9828003e-24], [541.96811, 1.1243994e-23, 7.9946822e-24, 7.994344e-24], [543.19656, 1.1260556e-23, 8.0062593e-24, 8.0059246e-24], [544.42779, 1.1277169e-23, 8.0178734e-24, 8.0175422e-24], [545.66181, 1.1293835e-23, 8.0295247e-24, 8.029197e-24], [546.89863, 1.1310553e-23, 8.0412131e-24, 8.0408888e-24], [548.13826, 1.1327324e-23, 8.0529388e-24, 8.0526179e-24], [549.38069, 1.1344147e-23, 8.0647019e-24, 8.0643843e-24], [550.62594, 1.1361022e-23, 8.0765024e-24, 8.0761881e-24], [551.87402, 1.137795e-23, 8.0883403e-24, 8.0880294e-24], [553.12492, 1.1394932e-23, 8.1002158e-24, 8.0999082e-24], [554.37866, 1.1411965e-23, 8.112129e-24, 8.1118246e-24], [555.63523, 1.1429052e-23, 8.1240799e-24, 8.1237786e-24], [556.89466, 1.1446193e-23, 8.1360686e-24, 8.1357705e-24], [558.15694, 1.1463386e-23, 8.1480952e-24, 8.1478002e-24], [559.42209, 1.1480633e-23, 8.1601597e-24, 8.1598678e-24], [560.6901, 1.1497933e-23, 8.1722623e-24, 8.1719735e-24], [561.96098, 1.1515287e-23, 8.184403e-24, 8.1841172e-24], [563.23475, 1.1532694e-23, 8.1965819e-24, 8.1962991e-24], [564.5114, 1.1550155e-23, 8.208799e-24, 8.2085192e-24], [565.79095, 1.156767e-23, 8.2210545e-24, 8.2207776e-24], [567.07339, 1.1585239e-23, 8.2333485e-24, 8.2330745e-24], [568.35874, 1.1602863e-23, 8.245681e-24, 8.2454098e-24], [569.64701, 1.162054e-23, 8.258052e-24, 8.2577837e-24], [570.9382, 1.1638272e-23, 8.2704618e-24, 8.2701963e-24], [572.23231, 1.1656059e-23, 8.2829103e-24, 8.2826476e-24], [573.52936, 1.16739e-23, 8.2953976e-24, 8.2951377e-24], [574.82935, 1.1691795e-23, 8.3079239e-24, 8.3076667e-24], [576.13228, 1.1709746e-23, 8.3204892e-24, 8.3202347e-24], [577.43817, 1.1727751e-23, 8.3330935e-24, 8.3328417e-24], [578.74701, 1.1745812e-23, 8.3457371e-24, 8.3454879e-24], [580.05883, 1.1763928e-23, 8.3584199e-24, 8.3581733e-24], [581.37361, 1.1782099e-23, 8.371142e-24, 8.370898e-24], [582.69138, 1.1800325e-23, 8.3839035e-24, 8.3836621e-24], [584.01213, 1.1818607e-23, 8.3967046e-24, 8.3964657e-24], [585.33588, 1.1836944e-23, 8.4095452e-24, 8.4093089e-24], [586.66263, 1.1855338e-23, 8.4224255e-24, 8.4221916e-24], [587.99238, 1.1873787e-23, 8.4353455e-24, 8.4351142e-24], [589.32515, 1.1892292e-23, 8.4483054e-24, 8.4480765e-24], [590.66094, 1.1910853e-23, 8.4613052e-24, 8.4610787e-24], [591.99976, 1.1929471e-23, 8.474345e-24, 8.4741209e-24], [593.34161, 1.1948144e-23, 8.4874249e-24, 8.4872031e-24], [594.68651, 1.1966875e-23, 8.5005449e-24, 8.5003255e-24], [596.03445, 1.1985661e-23, 8.5137053e-24, 8.5134881e-24], [597.38545, 1.2004505e-23, 8.5269059e-24, 8.5266911e-24], [598.73951, 1.2023405e-23, 8.540147e-24, 8.5399344e-24], [600.09664, 1.2042363e-23, 8.5534286e-24, 8.5532183e-24], [601.45684, 1.2061377e-23, 8.5667508e-24, 8.5665427e-24], [602.82013, 1.2080448e-23, 8.5801137e-24, 8.5799078e-24], [604.18651, 1.2099577e-23, 8.5935173e-24, 8.5933136e-24], [605.55598, 1.2118763e-23, 8.6069618e-24, 8.6067602e-24], [606.92856, 1.2138007e-23, 8.6204472e-24, 8.6202478e-24], [608.30425, 1.2157309e-23, 8.6339737e-24, 8.6337764e-24], [609.68306, 1.2176668e-23, 8.6475412e-24, 8.647346e-24], [611.065, 1.2196085e-23, 8.66115e-24, 8.6609568e-24], [612.45006, 1.221556e-23, 8.6748001e-24, 8.6746089e-24], [613.83827, 1.2235093e-23, 8.6884915e-24, 8.6883024e-24], [615.22962, 1.2254684e-23, 8.7022244e-24, 8.7020373e-24], [616.62413, 1.2274334e-23, 8.7159988e-24, 8.7158137e-24], [618.0218, 1.2294043e-23, 8.7298148e-24, 8.7296317e-24], [619.42263, 1.231381e-23, 8.7436726e-24, 8.7434914e-24], [620.82664, 1.2333635e-23, 8.7575722e-24, 8.7573929e-24], [622.23384, 1.235352e-23, 8.7715137e-24, 8.7713363e-24], [623.64422, 1.2373464e-23, 8.7854971e-24, 8.7853216e-24], [625.0578, 1.2393466e-23, 8.7995227e-24, 8.799349e-24], [626.47458, 1.2413528e-23, 8.8135903e-24, 8.8134185e-24], [627.89458, 1.243365e-23, 8.8277003e-24, 8.8275302e-24], [629.31779, 1.2453831e-23, 8.8418525e-24, 8.8416843e-24], [630.74423, 1.2474071e-23, 8.8560472e-24, 8.8558808e-24], [632.1739, 1.2494371e-23, 8.8702844e-24, 8.8701197e-24], [633.60681, 1.2514731e-23, 8.8845642e-24, 8.8844013e-24], [635.04297, 1.2535151e-23, 8.8988867e-24, 8.8987255e-24], [636.48239, 1.2555631e-23, 8.913252e-24, 8.9130925e-24], [637.92507, 1.2576172e-23, 8.9276601e-24, 8.9275023e-24], [639.37102, 1.2596773e-23, 8.9421112e-24, 8.9419551e-24], [640.82024, 1.2617434e-23, 8.9566053e-24, 8.9564509e-24], [642.27276, 1.2638156e-23, 8.9711426e-24, 8.9709898e-24], [643.72856, 1.2658938e-23, 8.9857231e-24, 8.9855719e-24], [645.18766, 1.2679782e-23, 9.0003469e-24, 9.0001973e-24], [646.65007, 1.2700686e-23, 9.0150141e-24, 9.0148662e-24], [648.1158, 1.2721652e-23, 9.0297249e-24, 9.0295784e-24], [649.58485, 1.2742679e-23, 9.0444792e-24, 9.0443343e-24], [651.05722, 1.2763767e-23, 9.0592772e-24, 9.0591338e-24], [652.53294, 1.2784917e-23, 9.0741189e-24, 9.0739771e-24], [654.012, 1.2806129e-23, 9.0890045e-24, 9.0888643e-24], [655.49441, 1.2827402e-23, 9.1039341e-24, 9.1037953e-24], [656.98018, 1.2848737e-23, 9.1189077e-24, 9.1187704e-24], [658.46932, 1.2870134e-23, 9.1339255e-24, 9.1337896e-24], [659.96184, 1.2891593e-23, 9.1489875e-24, 9.1488531e-24], [661.45774, 1.2913115e-23, 9.1640938e-24, 9.1639608e-24], [662.95703, 1.2934699e-23, 9.1792445e-24, 9.1791129e-24], [664.45971, 1.2956345e-23, 9.1944397e-24, 9.1943095e-24], [665.96581, 1.2978055e-23, 9.2096795e-24, 9.2095507e-24], [667.47531, 1.2999827e-23, 9.224964e-24, 9.2248366e-24], [668.98824, 1.3021662e-23, 9.2402933e-24, 9.2401672e-24], [670.5046, 1.304356e-23, 9.2556674e-24, 9.2555427e-24], [672.02439, 1.3065521e-23, 9.2710865e-24, 9.2709632e-24], [673.54763, 1.3087546e-23, 9.2865507e-24, 9.2864287e-24], [675.07433, 1.3109634e-23, 9.30206e-24, 9.3019393e-24], [676.60448, 1.3131785e-23, 9.3176146e-24, 9.3174952e-24], [678.1381, 1.3154001e-23, 9.3332145e-24, 9.3330964e-24], [679.6752, 1.317628e-23, 9.3488599e-24, 9.348743e-24], [681.21578, 1.3198623e-23, 9.3645508e-24, 9.3644351e-24], [682.75985, 1.322103e-23, 9.3802873e-24, 9.3801728e-24], [684.30743, 1.3243502e-23, 9.3960695e-24, 9.3959563e-24], [685.85851, 1.3266038e-23, 9.4118976e-24, 9.4117855e-24], [687.41311, 1.3288639e-23, 9.4277715e-24, 9.4276607e-24], [688.97123, 1.3311304e-23, 9.4436914e-24, 9.4435818e-24], [690.53288, 1.3334034e-23, 9.4596575e-24, 9.459549e-24], [692.09807, 1.3356829e-23, 9.4756697e-24, 9.4755624e-24], [693.66681, 1.3379689e-23, 9.4917283e-24, 9.4916221e-24], [695.23911, 1.3402614e-23, 9.5078332e-24, 9.5077281e-24], [696.81497, 1.3425604e-23, 9.5239846e-24, 9.5238807e-24], [698.3944, 1.344866e-23, 9.5401825e-24, 9.5400797e-24], [699.97741, 1.3471782e-23, 9.5564272e-24, 9.5563255e-24], [701.56401, 1.3494969e-23, 9.5727186e-24, 9.572618e-24], [703.1542, 1.3518223e-23, 9.5890568e-24, 9.5889573e-24], [704.748, 1.3541542e-23, 9.6054421e-24, 9.6053436e-24], [706.34542, 1.3564927e-23, 9.6218744e-24, 9.621777e-24], [707.94645, 1.3588379e-23, 9.6383538e-24, 9.6382574e-24], [709.55111, 1.3611897e-23, 9.6548805e-24, 9.6547852e-24], [711.15941, 1.3635482e-23, 9.6714545e-24, 9.6713602e-24], [712.77136, 1.3659133e-23, 9.688076e-24, 9.6879828e-24], [714.38695, 1.3682852e-23, 9.7047451e-24, 9.7046528e-24], [716.00622, 1.3706637e-23, 9.7214618e-24, 9.7213705e-24], [717.62915, 1.3730489e-23, 9.7382262e-24, 9.7381359e-24], [719.25576, 1.3754409e-23, 9.7550385e-24, 9.7549491e-24], [720.88605, 1.3778396e-23, 9.7718987e-24, 9.7718103e-24], [722.52004, 1.3802451e-23, 9.7888069e-24, 9.7887195e-24], [724.15774, 1.3826573e-23, 9.8057633e-24, 9.8056768e-24], [725.79915, 1.3850763e-23, 9.822768e-24, 9.8226824e-24], [727.44428, 1.3875021e-23, 9.8398209e-24, 9.8397362e-24], [729.09313, 1.3899347e-23, 9.8569223e-24, 9.8568386e-24], [730.74573, 1.3923742e-23, 9.8740722e-24, 9.8739894e-24], [732.40207, 1.3948204e-23, 9.8912708e-24, 9.8911888e-24], [734.06216, 1.3972736e-23, 9.9085181e-24, 9.908437e-24], [735.72602, 1.3997336e-23, 9.9258142e-24, 9.925734e-24], [737.39365, 1.4022005e-23, 9.9431593e-24, 9.94308e-24], [739.06506, 1.4046742e-23, 9.9605534e-24, 9.9604749e-24], [740.74025, 1.4071549e-23, 9.9779967e-24, 9.977919e-24], [742.41925, 1.4096425e-23, 9.9954891e-24, 9.9954123e-24], [744.10205, 1.4121371e-23, 1.0013031e-23, 1.0012955e-23], [745.78866, 1.4146386e-23, 1.0030622e-23, 1.0030547e-23], [747.4791, 1.417147e-23, 1.0048263e-23, 1.0048189e-23], [749.17337, 1.4196625e-23, 1.0065953e-23, 1.006588e-23], [750.87148, 1.4221849e-23, 1.0083694e-23, 1.0083621e-23], [752.57344, 1.4247144e-23, 1.0101484e-23, 1.0101412e-23], [754.27925, 1.4272509e-23, 1.0119324e-23, 1.0119253e-23], [755.98893, 1.4297944e-23, 1.0137214e-23, 1.0137143e-23], [757.70249, 1.432345e-23, 1.0155154e-23, 1.0155084e-23], [759.41993, 1.4349026e-23, 1.0173144e-23, 1.0173076e-23], [761.14127, 1.4374674e-23, 1.0191185e-23, 1.0191117e-23], [762.8665, 1.4400392e-23, 1.0209277e-23, 1.0209209e-23], [764.59565, 1.4426181e-23, 1.0227418e-23, 1.0227352e-23], [766.32872, 1.4452042e-23, 1.0245611e-23, 1.0245545e-23], [768.06571, 1.4477974e-23, 1.0263854e-23, 1.0263789e-23], [769.80664, 1.4503978e-23, 1.0282149e-23, 1.0282084e-23], [771.55152, 1.4530053e-23, 1.0300494e-23, 1.030043e-23], [773.30035, 1.4556201e-23, 1.031889e-23, 1.0318827e-23], [775.05315, 1.458242e-23, 1.0337338e-23, 1.0337275e-23], [776.80992, 1.4608711e-23, 1.0355836e-23, 1.0355774e-23], [778.57067, 1.4635075e-23, 1.0374387e-23, 1.0374325e-23], [780.33541, 1.4661511e-23, 1.0392988e-23, 1.0392928e-23], [782.10415, 1.468802e-23, 1.0411642e-23, 1.0411582e-23], [783.8769, 1.4714602e-23, 1.0430347e-23, 1.0430287e-23], [785.65367, 1.4741256e-23, 1.0449104e-23, 1.0449045e-23], [787.43447, 1.4767984e-23, 1.0467912e-23, 1.0467854e-23], [789.2193, 1.4794785e-23, 1.0486773e-23, 1.0486716e-23], [791.00818, 1.4821659e-23, 1.0505686e-23, 1.050563e-23], [792.80112, 1.4848607e-23, 1.0524652e-23, 1.0524595e-23], [794.59812, 1.4875628e-23, 1.0543669e-23, 1.0543614e-23], [796.39919, 1.4902723e-23, 1.0562739e-23, 1.0562684e-23], [798.20434, 1.4929893e-23, 1.0581862e-23, 1.0581808e-23], [800.01358, 1.4957136e-23, 1.0601037e-23, 1.0600984e-23], [801.82693, 1.4984454e-23, 1.0620266e-23, 1.0620212e-23], [803.64439, 1.5011846e-23, 1.0639547e-23, 1.0639494e-23], [805.46596, 1.5039312e-23, 1.0658881e-23, 1.0658829e-23], [807.29167, 1.5066854e-23, 1.0678268e-23, 1.0678217e-23], [809.12151, 1.509447e-23, 1.0697708e-23, 1.0697658e-23], [810.9555, 1.5122162e-23, 1.0717202e-23, 1.0717152e-23], [812.79365, 1.5149928e-23, 1.073675e-23, 1.07367e-23], [814.63596, 1.517777e-23, 1.075635e-23, 1.0756301e-23], [816.48245, 1.5205688e-23, 1.0776005e-23, 1.0775956e-23], [818.33312, 1.5233681e-23, 1.0795713e-23, 1.0795665e-23], [820.18799, 1.526175e-23, 1.0815475e-23, 1.0815428e-23], [822.04707, 1.5289895e-23, 1.0835291e-23, 1.0835244e-23], [823.91036, 1.5318116e-23, 1.0855162e-23, 1.0855115e-23], [825.77787, 1.5346413e-23, 1.0875086e-23, 1.087504e-23], [827.64961, 1.5374787e-23, 1.0895065e-23, 1.0895019e-23], [829.5256, 1.5403238e-23, 1.0915098e-23, 1.0915053e-23], [831.40584, 1.5431765e-23, 1.0935186e-23, 1.0935141e-23], [833.29034, 1.546037e-23, 1.0955329e-23, 1.0955284e-23], [835.17911, 1.5489051e-23, 1.0975526e-23, 1.0975482e-23], [837.07216, 1.551781e-23, 1.0995778e-23, 1.0995735e-23], [838.96951, 1.5546646e-23, 1.1016085e-23, 1.1016043e-23], [840.87115, 1.5575559e-23, 1.1036448e-23, 1.1036405e-23], [842.77711, 1.5604551e-23, 1.1056865e-23, 1.1056823e-23], [844.68738, 1.563362e-23, 1.1077338e-23, 1.1077297e-23], [846.60199, 1.5662768e-23, 1.1097866e-23, 1.1097825e-23], [848.52093, 1.5691993e-23, 1.111845e-23, 1.111841e-23], [850.44423, 1.5721297e-23, 1.113909e-23, 1.113905e-23], [852.37188, 1.575068e-23, 1.1159785e-23, 1.1159746e-23], [854.30391, 1.5780141e-23, 1.1180537e-23, 1.1180497e-23], [856.24031, 1.5809681e-23, 1.1201344e-23, 1.1201305e-23], [858.1811, 1.5839301e-23, 1.1222208e-23, 1.1222169e-23], [860.12629, 1.5868999e-23, 1.1243127e-23, 1.1243089e-23], [862.07589, 1.5898777e-23, 1.1264103e-23, 1.1264066e-23], [864.02991, 1.5928634e-23, 1.1285136e-23, 1.1285099e-23], [865.98836, 1.5958571e-23, 1.1306225e-23, 1.1306189e-23], [867.95124, 1.5988588e-23, 1.1327371e-23, 1.1327335e-23], [869.91858, 1.6018685e-23, 1.1348574e-23, 1.1348538e-23], [871.89038, 1.6048862e-23, 1.1369834e-23, 1.1369799e-23], [873.86664, 1.607912e-23, 1.1391151e-23, 1.1391116e-23], [875.84739, 1.6109458e-23, 1.1412525e-23, 1.141249e-23], [877.83262, 1.6139876e-23, 1.1433956e-23, 1.1433922e-23], [879.82235, 1.6170376e-23, 1.1455445e-23, 1.1455411e-23], [881.8166, 1.6200956e-23, 1.1476992e-23, 1.1476958e-23], [883.81536, 1.6231618e-23, 1.1498596e-23, 1.1498562e-23], [885.81866, 1.6262361e-23, 1.1520258e-23, 1.1520225e-23], [887.82649, 1.6293186e-23, 1.1541977e-23, 1.1541945e-23], [889.83888, 1.6324093e-23, 1.1563755e-23, 1.1563723e-23], [891.85583, 1.6355081e-23, 1.1585591e-23, 1.1585559e-23], [893.87734, 1.6386151e-23, 1.1607485e-23, 1.1607454e-23], [895.90345, 1.6417304e-23, 1.1629438e-23, 1.1629407e-23], [897.93414, 1.6448539e-23, 1.1651449e-23, 1.1651418e-23], [899.96944, 1.6479856e-23, 1.1673519e-23, 1.1673488e-23], [902.00935, 1.6511257e-23, 1.1695647e-23, 1.1695617e-23], [904.05388, 1.654274e-23, 1.1717835e-23, 1.1717805e-23], [906.10305, 1.6574306e-23, 1.1740081e-23, 1.1740051e-23], [908.15686, 1.6605956e-23, 1.1762386e-23, 1.1762357e-23], [910.21533, 1.6637688e-23, 1.1784751e-23, 1.1784722e-23], [912.27846, 1.6669505e-23, 1.1807175e-23, 1.1807147e-23], [914.34627, 1.6701405e-23, 1.1829659e-23, 1.1829631e-23], [916.41877, 1.673339e-23, 1.1852202e-23, 1.1852174e-23], [918.49597, 1.6765458e-23, 1.1874805e-23, 1.1874778e-23], [920.57787, 1.6797611e-23, 1.1897468e-23, 1.1897441e-23], [922.66449, 1.6829848e-23, 1.1920191e-23, 1.1920164e-23], [924.75584, 1.686217e-23, 1.1942975e-23, 1.1942948e-23], [926.85193, 1.6894577e-23, 1.1965818e-23, 1.1965792e-23], [928.95278, 1.6927069e-23, 1.1988722e-23, 1.1988696e-23], [931.05838, 1.6959647e-23, 1.2011687e-23, 1.2011661e-23], [933.16876, 1.699231e-23, 1.2034713e-23, 1.2034687e-23], [935.28392, 1.7025059e-23, 1.20578e-23, 1.2057774e-23], [937.40388, 1.7057893e-23, 1.2080948e-23, 1.2080923e-23], [939.52864, 1.7090815e-23, 1.2104158e-23, 1.2104133e-23], [941.65821, 1.7123823e-23, 1.212743e-23, 1.2127406e-23], [943.79262, 1.7156919e-23, 1.2150765e-23, 1.2150741e-23], [945.93186, 1.7190103e-23, 1.2174165e-23, 1.2174141e-23], [948.07595, 1.7223377e-23, 1.2197631e-23, 1.2197607e-23], [950.2249, 1.7256745e-23, 1.2221168e-23, 1.2221145e-23], [952.37872, 1.7290215e-23, 1.2244788e-23, 1.2244765e-23], [954.53743, 1.732381e-23, 1.2268523e-23, 1.22685e-23], [956.70102, 1.735763e-23, 1.2292516e-23, 1.2292494e-23], [958.86952, 1.7392621e-23, 1.2318099e-23, 1.2318077e-23], [961.04294, 1.7506943e-23, 1.2455377e-23, 1.2455355e-23], [963.22128, 1.7459609e-23, 1.2364856e-23, 1.2364834e-23], [965.40456, 1.7492862e-23, 1.23878e-23, 1.2387778e-23], [967.59279, 1.7526766e-23, 1.2411602e-23, 1.2411581e-23], [969.78598, 1.7560852e-23, 1.2435599e-23, 1.2435578e-23], [971.98413, 1.7595055e-23, 1.24597e-23, 1.2459679e-23], [974.18728, 1.7629359e-23, 1.2483879e-23, 1.2483859e-23], [976.39541, 1.7663757e-23, 1.250813e-23, 1.2508109e-23], [978.60855, 1.7698247e-23, 1.2532448e-23, 1.2532427e-23], [980.82671, 1.7732828e-23, 1.2556831e-23, 1.2556811e-23], [983.04989, 1.7767499e-23, 1.258128e-23, 1.258126e-23], [985.27812, 1.7802261e-23, 1.2605793e-23, 1.2605773e-23], [987.51139, 1.7837114e-23, 1.2630371e-23, 1.2630351e-23], [989.74973, 1.7872057e-23, 1.2655013e-23, 1.2654994e-23], [991.99314, 1.790709e-23, 1.267972e-23, 1.2679701e-23], [994.24163, 1.7942214e-23, 1.2704491e-23, 1.2704472e-23], [996.49523, 1.7977429e-23, 1.2729326e-23, 1.2729307e-23], [998.75393, 1.8012734e-23, 1.2754226e-23, 1.2754208e-23], [1001.0177, 1.8048131e-23, 1.2779191e-23, 1.2779173e-23], [1003.2867, 1.8083619e-23, 1.280422e-23, 1.2804202e-23], [1005.5608, 1.8119198e-23, 1.2829315e-23, 1.2829297e-23], [1007.84, 1.815487e-23, 1.2854474e-23, 1.2854457e-23], [1010.1245, 1.8190632e-23, 1.2879699e-23, 1.2879682e-23], [1012.414, 1.8226487e-23, 1.2904989e-23, 1.2904972e-23], [1014.7088, 1.8262434e-23, 1.2930345e-23, 1.2930328e-23], [1017.0088, 1.8298474e-23, 1.2955766e-23, 1.2955749e-23], [1019.314, 1.8334606e-23, 1.2981252e-23, 1.2981236e-23], [1021.6244, 1.8370831e-23, 1.3006805e-23, 1.3006788e-23], [1023.9401, 1.8407149e-23, 1.3032423e-23, 1.3032407e-23], [1026.261, 1.8443559e-23, 1.3058108e-23, 1.3058091e-23], [1028.5872, 1.8480064e-23, 1.3083858e-23, 1.3083842e-23], [1030.9186, 1.8516661e-23, 1.3109675e-23, 1.310966e-23], [1033.2553, 1.8553353e-23, 1.3135559e-23, 1.3135543e-23], [1035.5973, 1.8590138e-23, 1.3161509e-23, 1.3161494e-23], [1037.9447, 1.8627018e-23, 1.3187526e-23, 1.3187511e-23], [1040.2973, 1.8663992e-23, 1.321361e-23, 1.3213595e-23], [1042.6553, 1.870106e-23, 1.3239761e-23, 1.3239746e-23], [1045.0186, 1.8738223e-23, 1.3265979e-23, 1.3265964e-23], [1047.3873, 1.8775481e-23, 1.3292264e-23, 1.3292249e-23], [1049.7614, 1.8812834e-23, 1.3318617e-23, 1.3318602e-23], [1052.1408, 1.8850282e-23, 1.3345037e-23, 1.3345023e-23], [1054.5256, 1.8887826e-23, 1.3371525e-23, 1.3371511e-23], [1056.9159, 1.8925465e-23, 1.3398081e-23, 1.3398067e-23], [1059.3115, 1.8963201e-23, 1.3424705e-23, 1.3424692e-23], [1061.7126, 1.9001032e-23, 1.3451397e-23, 1.3451384e-23], [1064.1191, 1.903896e-23, 1.3478158e-23, 1.3478144e-23], [1066.5311, 1.9076984e-23, 1.3504987e-23, 1.3504974e-23], [1068.9486, 1.9115105e-23, 1.3531884e-23, 1.3531871e-23], [1071.3715, 1.9153323e-23, 1.3558851e-23, 1.3558838e-23], [1073.7999, 1.9191638e-23, 1.3585886e-23, 1.3585873e-23], [1076.2338, 1.923005e-23, 1.361299e-23, 1.3612977e-23], [1078.6733, 1.926856e-23, 1.3640163e-23, 1.364015e-23], [1081.1182, 1.9307167e-23, 1.3667406e-23, 1.3667393e-23], [1083.5687, 1.9345873e-23, 1.3694718e-23, 1.3694705e-23], [1086.0248, 1.9384676e-23, 1.3722099e-23, 1.3722087e-23], [1088.4864, 1.9423578e-23, 1.374955e-23, 1.3749538e-23], [1090.9536, 1.9462579e-23, 1.3777072e-23, 1.377706e-23], [1093.4265, 1.9501678e-23, 1.3804663e-23, 1.3804651e-23], [1095.9049, 1.9540876e-23, 1.3832324e-23, 1.3832312e-23], [1098.3889, 1.9580173e-23, 1.3860056e-23, 1.3860044e-23], [1100.8785, 1.961957e-23, 1.3887858e-23, 1.3887846e-23], [1103.3738, 1.9659066e-23, 1.3915731e-23, 1.3915719e-23], [1105.8748, 1.9698662e-23, 1.3943674e-23, 1.3943663e-23], [1108.3814, 1.9738358e-23, 1.3971688e-23, 1.3971677e-23], [1110.8937, 1.9778154e-23, 1.3999774e-23, 1.3999763e-23], [1113.4117, 1.9818051e-23, 1.4027931e-23, 1.402792e-23], [1115.9354, 1.9858048e-23, 1.4056158e-23, 1.4056148e-23], [1118.4649, 1.9898146e-23, 1.4084458e-23, 1.4084447e-23], [1121.0, 1.9938345e-23, 1.4112829e-23, 1.4112819e-23], [1123.5409, 1.9978645e-23, 1.4141272e-23, 1.4141262e-23], [1126.0876, 2.0019047e-23, 1.4169787e-23, 1.4169777e-23], [1128.64, 2.005955e-23, 1.4198374e-23, 1.4198364e-23], [1131.1983, 2.0100156e-23, 1.4227033e-23, 1.4227023e-23], [1133.7623, 2.0140863e-23, 1.4255764e-23, 1.4255755e-23], [1136.3321, 2.0181673e-23, 1.4284569e-23, 1.4284559e-23], [1138.9078, 2.0222586e-23, 1.4313445e-23, 1.4313436e-23], [1141.4893, 2.0263601e-23, 1.4342395e-23, 1.4342386e-23], [1144.0766, 2.0304719e-23, 1.4371418e-23, 1.4371408e-23], [1146.6698, 2.034594e-23, 1.4400514e-23, 1.4400504e-23], [1149.2689, 2.0387265e-23, 1.4429683e-23, 1.4429673e-23], [1151.8739, 2.0428693e-23, 1.4458925e-23, 1.4458916e-23], [1154.4848, 2.0470225e-23, 1.4488241e-23, 1.4488232e-23], [1157.1016, 2.0511862e-23, 1.4517631e-23, 1.4517623e-23], [1159.7243, 2.0553602e-23, 1.4547095e-23, 1.4547086e-23], [1162.353, 2.0595447e-23, 1.4576633e-23, 1.4576625e-23], [1164.9877, 2.0637397e-23, 1.4606246e-23, 1.4606237e-23], [1167.6283, 2.0679451e-23, 1.4635932e-23, 1.4635924e-23], [1170.2749, 2.0721611e-23, 1.4665693e-23, 1.4665685e-23], [1172.9275, 2.0763877e-23, 1.4695529e-23, 1.4695521e-23], [1175.5861, 2.0806247e-23, 1.472544e-23, 1.4725431e-23], [1178.2507, 2.0848724e-23, 1.4755425e-23, 1.4755417e-23], [1180.9214, 2.0891307e-23, 1.4785486e-23, 1.4785478e-23], [1183.5981, 2.0933996e-23, 1.4815623e-23, 1.4815615e-23], [1186.2809, 2.0976791e-23, 1.4845834e-23, 1.4845826e-23], [1188.9698, 2.1019694e-23, 1.4876122e-23, 1.4876114e-23], [1191.6648, 2.1062703e-23, 1.4906485e-23, 1.4906477e-23], [1194.3658, 2.1105819e-23, 1.4936924e-23, 1.4936916e-23], [1197.073, 2.1149043e-23, 1.4967439e-23, 1.4967431e-23], [1199.7864, 2.1192374e-23, 1.4998031e-23, 1.4998023e-23], [1202.5059, 2.1235813e-23, 1.5028698e-23, 1.5028691e-23], [1205.2315, 2.1279361e-23, 1.5059443e-23, 1.5059436e-23], [1207.9634, 2.1323016e-23, 1.5090264e-23, 1.5090257e-23], [1210.7014, 2.1366781e-23, 1.5121163e-23, 1.5121155e-23], [1213.4456, 2.1410654e-23, 1.5152138e-23, 1.5152131e-23], [1216.1961, 2.1454636e-23, 1.518319e-23, 1.5183183e-23], [1218.9527, 2.1498727e-23, 1.521432e-23, 1.5214314e-23], [1221.7157, 2.1542927e-23, 1.5245528e-23, 1.5245521e-23], [1224.4849, 2.1587238e-23, 1.5276813e-23, 1.5276807e-23], [1227.2603, 2.1631658e-23, 1.5308177e-23, 1.530817e-23], [1230.0421, 2.1676189e-23, 1.5339618e-23, 1.5339611e-23], [1232.8302, 2.1720829e-23, 1.5371137e-23, 1.5371131e-23], [1235.6246, 2.1765581e-23, 1.5402735e-23, 1.5402729e-23], [1238.4253, 2.1810443e-23, 1.5434412e-23, 1.5434405e-23], [1241.2323, 2.1855417e-23, 1.5466167e-23, 1.546616e-23], [1244.0458, 2.1900501e-23, 1.5498001e-23, 1.5497995e-23], [1246.8656, 2.1945698e-23, 1.5529914e-23, 1.5529908e-23], [1249.6918, 2.1991006e-23, 1.5561906e-23, 1.55619e-23], [1252.5244, 2.2036426e-23, 1.5593978e-23, 1.5593972e-23], [1255.3634, 2.2081958e-23, 1.5626129e-23, 1.5626123e-23], [1258.2089, 2.2127603e-23, 1.565836e-23, 1.5658354e-23], [1261.0608, 2.2173361e-23, 1.5690671e-23, 1.5690665e-23], [1263.9192, 2.2219232e-23, 1.5723062e-23, 1.5723056e-23], [1266.784, 2.2265216e-23, 1.5755533e-23, 1.5755528e-23], [1269.6554, 2.2311313e-23, 1.5788085e-23, 1.5788079e-23], [1272.5332, 2.2357524e-23, 1.5820717e-23, 1.5820712e-23], [1275.4176, 2.2403849e-23, 1.585343e-23, 1.5853425e-23], [1278.3085, 2.2450289e-23, 1.5886224e-23, 1.5886218e-23], [1281.206, 2.2496843e-23, 1.5919099e-23, 1.5919093e-23], [1284.11, 2.2543511e-23, 1.5952055e-23, 1.595205e-23], [1287.0206, 2.2590294e-23, 1.5985093e-23, 1.5985087e-23], [1289.9379, 2.2637193e-23, 1.6018212e-23, 1.6018207e-23], [1292.8617, 2.2684207e-23, 1.6051413e-23, 1.6051408e-23], [1295.7922, 2.2731336e-23, 1.6084696e-23, 1.6084691e-23], [1298.7292, 2.2778582e-23, 1.6118061e-23, 1.6118056e-23], [1301.673, 2.2825943e-23, 1.6151508e-23, 1.6151503e-23], [1304.6234, 2.2873422e-23, 1.6185038e-23, 1.6185033e-23], [1307.5806, 2.2921016e-23, 1.621865e-23, 1.6218645e-23], [1310.5444, 2.2968728e-23, 1.6252345e-23, 1.625234e-23], [1313.5149, 2.3016557e-23, 1.6286124e-23, 1.6286119e-23], [1316.4922, 2.3064503e-23, 1.6319985e-23, 1.631998e-23], [1319.4762, 2.3112566e-23, 1.635393e-23, 1.6353925e-23], [1322.467, 2.3160748e-23, 1.6387958e-23, 1.6387953e-23], [1325.4645, 2.3209048e-23, 1.642207e-23, 1.6422065e-23], [1328.4689, 2.3257466e-23, 1.6456266e-23, 1.6456261e-23], [1331.4801, 2.3306003e-23, 1.6490545e-23, 1.6490541e-23], [1334.4981, 2.3354659e-23, 1.6524909e-23, 1.6524905e-23], [1337.5229, 2.3403434e-23, 1.6559358e-23, 1.6559354e-23], [1340.5546, 2.3452328e-23, 1.6593891e-23, 1.6593887e-23], [1343.5931, 2.3501342e-23, 1.6628509e-23, 1.6628505e-23], [1346.6386, 2.3550476e-23, 1.6663212e-23, 1.6663207e-23], [1349.6909, 2.359973e-23, 1.6698e-23, 1.6697995e-23], [1352.7502, 2.3649105e-23, 1.6732873e-23, 1.6732869e-23], [1355.8164, 2.36986e-23, 1.6767831e-23, 1.6767827e-23], [1358.8896, 2.3748216e-23, 1.6802876e-23, 1.6802872e-23], [1361.9697, 2.3797953e-23, 1.6838006e-23, 1.6838002e-23], [1365.0568, 2.3847812e-23, 1.6873222e-23, 1.6873218e-23], [1368.1509, 2.3897793e-23, 1.6908525e-23, 1.6908521e-23], [1371.252, 2.3947895e-23, 1.6943914e-23, 1.694391e-23], [1374.3601, 2.399812e-23, 1.6979389e-23, 1.6979385e-23], [1377.4753, 2.4048467e-23, 1.7014951e-23, 1.7014947e-23], [1380.5976, 2.4098937e-23, 1.70506e-23, 1.7050597e-23], [1383.7269, 2.4149531e-23, 1.7086337e-23, 1.7086333e-23], [1386.8633, 2.4200247e-23, 1.712216e-23, 1.7122157e-23], [1390.0068, 2.4251087e-23, 1.7158072e-23, 1.7158068e-23], [1393.1575, 2.4302051e-23, 1.719407e-23, 1.7194067e-23], [1396.3153, 2.4353138e-23, 1.7230157e-23, 1.7230154e-23], [1399.4802, 2.4404351e-23, 1.7266332e-23, 1.7266329e-23], [1402.6523, 2.4455687e-23, 1.7302595e-23, 1.7302592e-23], [1405.8316, 2.4507149e-23, 1.7338947e-23, 1.7338943e-23], [1409.0182, 2.4558736e-23, 1.7375387e-23, 1.7375384e-23], [1412.2119, 2.4610449e-23, 1.7411916e-23, 1.7411913e-23], [1415.4129, 2.4662287e-23, 1.7448535e-23, 1.7448531e-23], [1418.6211, 2.4714251e-23, 1.7485242e-23, 1.7485239e-23], [1421.8366, 2.4766342e-23, 1.752204e-23, 1.7522037e-23], [1425.0594, 2.4818559e-23, 1.7558927e-23, 1.7558924e-23], [1428.2895, 2.4870905e-23, 1.7595906e-23, 1.7595903e-23], [1431.527, 2.4923379e-23, 1.7632978e-23, 1.7632975e-23], [1434.7717, 2.4975989e-23, 1.7670151e-23, 1.7670148e-23], [1438.0238, 2.5028784e-23, 1.7707496e-23, 1.7707493e-23], [1441.2833, 2.5140746e-23, 1.7828283e-23, 1.782828e-23], [1444.5502, 2.5134603e-23, 1.7782239e-23, 1.7782236e-23], [1447.8245, 2.5187656e-23, 1.7819677e-23, 1.7819674e-23], [1451.1062, 2.5240883e-23, 1.7857271e-23, 1.7857268e-23], [1454.3953, 2.5294246e-23, 1.7894967e-23, 1.7894964e-23], [1457.6919, 2.5347741e-23, 1.7932757e-23, 1.7932754e-23], [1460.996, 2.5401366e-23, 1.7970639e-23, 1.7970637e-23], [1464.3075, 2.5455121e-23, 1.8008615e-23, 1.8008612e-23], [1467.6266, 2.5509007e-23, 1.8046682e-23, 1.8046679e-23], [1470.9532, 2.5563023e-23, 1.8084842e-23, 1.808484e-23], [1474.2873, 2.5617169e-23, 1.8123095e-23, 1.8123092e-23], [1477.629, 2.5671447e-23, 1.8161441e-23, 1.8161438e-23], [1480.9783, 2.5725856e-23, 1.8199879e-23, 1.8199876e-23], [1484.3351, 2.5780396e-23, 1.823841e-23, 1.8238408e-23], [1487.6996, 2.5835068e-23, 1.8277035e-23, 1.8277033e-23], [1491.0717, 2.5889872e-23, 1.8315754e-23, 1.8315751e-23], [1494.4514, 2.5944808e-23, 1.8354566e-23, 1.8354563e-23], [1497.8388, 2.5999877e-23, 1.8393472e-23, 1.8393469e-23], [1501.2338, 2.6055079e-23, 1.8432472e-23, 1.8432469e-23], [1504.6366, 2.6110414e-23, 1.8471566e-23, 1.8471563e-23], [1508.0471, 2.6165883e-23, 1.8510755e-23, 1.8510752e-23], [1511.4653, 2.6221485e-23, 1.8550038e-23, 1.8550036e-23], [1514.8912, 2.6277221e-23, 1.8589417e-23, 1.8589414e-23], [1518.325, 2.6333092e-23, 1.862889e-23, 1.8628888e-23], [1521.7665, 2.6389097e-23, 1.8668459e-23, 1.8668457e-23], [1525.2158, 2.6445237e-23, 1.8708124e-23, 1.8708121e-23], [1528.6729, 2.6501513e-23, 1.8747884e-23, 1.8747881e-23], [1532.1378, 2.6557923e-23, 1.878774e-23, 1.8787737e-23], [1535.6106, 2.661447e-23, 1.8827692e-23, 1.882769e-23], [1539.0913, 2.6671153e-23, 1.886774e-23, 1.8867738e-23], [1542.5799, 2.6727972e-23, 1.8907885e-23, 1.8907883e-23], [1546.0764, 2.6784927e-23, 1.8948127e-23, 1.8948125e-23], [1549.5808, 2.684202e-23, 1.8988466e-23, 1.8988464e-23], [1553.0931, 2.689925e-23, 1.9028902e-23, 1.9028899e-23], [1556.6134, 2.6956617e-23, 1.9069435e-23, 1.9069433e-23], [1560.1417, 2.7014122e-23, 1.9110066e-23, 1.9110064e-23], [1563.678, 2.7071766e-23, 1.9150794e-23, 1.9150792e-23], [1567.2223, 2.7129547e-23, 1.9191621e-23, 1.9191619e-23], [1570.7746, 2.7187468e-23, 1.9232546e-23, 1.9232544e-23], [1574.335, 2.7245527e-23, 1.9273569e-23, 1.9273567e-23], [1577.9035, 2.7303726e-23, 1.9314691e-23, 1.9314689e-23], [1581.48, 2.7362065e-23, 1.9355912e-23, 1.935591e-23], [1585.0647, 2.7420543e-23, 1.9397232e-23, 1.939723e-23], [1588.6575, 2.7479162e-23, 1.9438651e-23, 1.9438649e-23], [1592.2584, 2.7537921e-23, 1.9480169e-23, 1.9480167e-23], [1595.8675, 2.7596821e-23, 1.9521788e-23, 1.9521786e-23], [1599.4847, 2.7655862e-23, 1.9563506e-23, 1.9563504e-23], [1603.1102, 2.7715045e-23, 1.9605324e-23, 1.9605322e-23], [1606.7439, 2.7774369e-23, 1.9647243e-23, 1.9647241e-23], [1610.3858, 2.7833836e-23, 1.9689262e-23, 1.968926e-23], [1614.036, 2.7893445e-23, 1.9731382e-23, 1.973138e-23], [1617.6944, 2.7953196e-23, 1.9773603e-23, 1.9773601e-23], [1621.3611, 2.801309e-23, 1.9815925e-23, 1.9815924e-23], [1625.0362, 2.8073128e-23, 1.9858349e-23, 1.9858347e-23], [1628.7196, 2.813331e-23, 1.9900874e-23, 1.9900873e-23], [1632.4113, 2.8193635e-23, 1.9943502e-23, 1.99435e-23], [1636.1114, 2.8254104e-23, 1.9986231e-23, 1.9986229e-23], [1639.8199, 2.8314718e-23, 2.0029063e-23, 2.0029061e-23], [1643.5368, 2.8375477e-23, 2.0071997e-23, 2.0071995e-23], [1647.2621, 2.8436381e-23, 2.0115034e-23, 2.0115032e-23], [1650.9958, 2.8497431e-23, 2.0158173e-23, 2.0158172e-23], [1654.738, 2.8558626e-23, 2.0201417e-23, 2.0201415e-23], [1658.4887, 2.8619968e-23, 2.0244763e-23, 2.0244762e-23], [1662.2479, 2.8681455e-23, 2.0288213e-23, 2.0288212e-23], [1666.0157, 2.874309e-23, 2.0331767e-23, 2.0331766e-23], [1669.7919, 2.8804872e-23, 2.0375425e-23, 2.0375424e-23], [1673.5768, 2.8866801e-23, 2.0419188e-23, 2.0419186e-23], [1677.3702, 2.8928877e-23, 2.0463055e-23, 2.0463053e-23], [1681.1722, 2.8991102e-23, 2.0507027e-23, 2.0507025e-23], [1684.9828, 2.9053475e-23, 2.0551103e-23, 2.0551102e-23], [1688.802, 2.9115997e-23, 2.0595285e-23, 2.0595284e-23], [1692.6299, 2.9178668e-23, 2.0639573e-23, 2.0639571e-23], [1696.4665, 2.9241488e-23, 2.0683966e-23, 2.0683964e-23], [1700.3118, 2.9304457e-23, 2.0728465e-23, 2.0728463e-23], [1704.1658, 2.9367577e-23, 2.077307e-23, 2.0773069e-23], [1708.0286, 2.9430847e-23, 2.0817781e-23, 2.081778e-23], [1711.9001, 2.9494267e-23, 2.08626e-23, 2.0862598e-23], [1715.7803, 2.9557839e-23, 2.0907525e-23, 2.0907523e-23], [1719.6694, 2.9621562e-23, 2.0952557e-23, 2.0952555e-23], [1723.5673, 2.9685436e-23, 2.0997696e-23, 2.0997695e-23], [1727.474, 2.9749462e-23, 2.1042943e-23, 2.1042941e-23], [1731.3895, 2.9813641e-23, 2.1088297e-23, 2.1088296e-23], [1735.314, 2.9877972e-23, 2.113376e-23, 2.1133759e-23], [1739.2473, 2.9942456e-23, 2.1179331e-23, 2.1179329e-23], [1743.1896, 3.0007093e-23, 2.122501e-23, 2.1225009e-23], [1747.1408, 3.0071884e-23, 2.1270798e-23, 2.1270797e-23], [1751.1009, 3.0136828e-23, 2.1316695e-23, 2.1316694e-23], [1755.07, 3.0201927e-23, 2.1362701e-23, 2.13627e-23], [1759.0482, 3.0267181e-23, 2.1408816e-23, 2.1408815e-23], [1763.0353, 3.0332589e-23, 2.1455041e-23, 2.145504e-23], [1767.0315, 3.0398153e-23, 2.1501376e-23, 2.1501375e-23], [1771.0367, 3.0463872e-23, 2.1547822e-23, 2.154782e-23], [1775.051, 3.0529748e-23, 2.1594377e-23, 2.1594376e-23], [1779.0744, 3.0595779e-23, 2.1641043e-23, 2.1641042e-23], [1783.1069, 3.0661967e-23, 2.168782e-23, 2.1687819e-23], [1787.1486, 3.0728312e-23, 2.1734708e-23, 2.1734707e-23], [1791.1994, 3.0794815e-23, 2.1781707e-23, 2.1781706e-23], [1795.2594, 3.0861475e-23, 2.1828818e-23, 2.1828817e-23], [1799.3287, 3.0928293e-23, 2.1876041e-23, 2.187604e-23], [1803.4071, 3.0995269e-23, 2.1923375e-23, 2.1923374e-23], [1807.4948, 3.1062404e-23, 2.1970822e-23, 2.1970821e-23], [1811.5917, 3.1129698e-23, 2.2018382e-23, 2.2018381e-23], [1815.698, 3.1197151e-23, 2.2066054e-23, 2.2066053e-23], [1819.8135, 3.1264764e-23, 2.211384e-23, 2.2113839e-23], [1823.9384, 3.1332537e-23, 2.2161738e-23, 2.2161737e-23], [1828.0726, 3.1400471e-23, 2.220975e-23, 2.2209749e-23], [1832.2162, 3.1468565e-23, 2.2257876e-23, 2.2257875e-23], [1836.3692, 3.1536821e-23, 2.2306116e-23, 2.2306115e-23], [1840.5316, 3.1605237e-23, 2.235447e-23, 2.2354469e-23], [1844.7034, 3.1673816e-23, 2.2402939e-23, 2.2402938e-23], [1848.8847, 3.1742557e-23, 2.2451522e-23, 2.2451521e-23], [1853.0754, 3.181146e-23, 2.2500221e-23, 2.250022e-23], [1857.2757, 3.1880527e-23, 2.2549034e-23, 2.2549034e-23], [1861.4855, 3.1949756e-23, 2.2597964e-23, 2.2597963e-23], [1865.7048, 3.2019149e-23, 2.2647009e-23, 2.2647008e-23], [1869.9337, 3.2088706e-23, 2.269617e-23, 2.2696169e-23], [1874.1722, 3.2158427e-23, 2.2745447e-23, 2.2745446e-23], [1878.4202, 3.2228313e-23, 2.2794841e-23, 2.279484e-23], [1882.678, 3.2298364e-23, 2.2844351e-23, 2.284435e-23], [1886.9453, 3.2368581e-23, 2.2893979e-23, 2.2893978e-23], [1891.2223, 3.2438963e-23, 2.2943724e-23, 2.2943723e-23], [1895.5091, 3.2509511e-23, 2.2993586e-23, 2.2993585e-23], [1899.8055, 3.2580226e-23, 2.3043567e-23, 2.3043566e-23], [1904.1117, 3.2651107e-23, 2.3093665e-23, 2.3093665e-23], [1908.4276, 3.2722157e-23, 2.3143883e-23, 2.3143882e-23], [1912.7534, 3.2793374e-23, 2.319422e-23, 2.3194219e-23], [1917.0889, 3.2864767e-23, 2.3244686e-23, 2.3244686e-23], [1921.4343, 3.2954526e-23, 2.3320994e-23, 2.3320993e-23], [1925.7895, 3.3008042e-23, 2.3345953e-23, 2.3345952e-23], [1930.1545, 3.3079922e-23, 2.3396751e-23, 2.3396751e-23], [1934.5295, 3.315198e-23, 2.3447681e-23, 2.344768e-23], [1938.9144, 3.3224209e-23, 2.3498732e-23, 2.3498731e-23], [1943.3092, 3.3296608e-23, 2.3549904e-23, 2.3549903e-23], [1947.714, 3.3369178e-23, 2.3601197e-23, 2.3601196e-23], [1952.1288, 3.3441918e-23, 2.365261e-23, 2.365261e-23], [1956.5536, 3.351483e-23, 2.3704145e-23, 2.3704145e-23], [1960.9884, 3.3587914e-23, 2.3755802e-23, 2.3755801e-23], [1965.4332, 3.366117e-23, 2.380758e-23, 2.3807579e-23], [1969.8882, 3.3734598e-23, 2.385948e-23, 2.385948e-23], [1974.3532, 3.3808199e-23, 2.3911503e-23, 2.3911502e-23], [1978.8284, 3.3881973e-23, 2.3963648e-23, 2.3963647e-23], [1983.3137, 3.395592e-23, 2.4015916e-23, 2.4015915e-23], [1987.8091, 3.4030042e-23, 2.4068306e-23, 2.4068306e-23], [1992.3148, 3.4104338e-23, 2.4120821e-23, 2.412082e-23], [1996.8307, 3.4178808e-23, 2.4173458e-23, 2.4173458e-23], [2001.3568, 3.4253453e-23, 2.422622e-23, 2.4226219e-23], [2005.8931, 3.4328274e-23, 2.4279106e-23, 2.4279105e-23], [2010.4398, 3.4403271e-23, 2.4332116e-23, 2.4332115e-23], [2014.9967, 3.4478444e-23, 2.4385251e-23, 2.438525e-23], [2019.564, 3.4553793e-23, 2.443851e-23, 2.443851e-23], [2024.1416, 3.462932e-23, 2.4491895e-23, 2.4491895e-23], [2028.7296, 3.4705024e-23, 2.4545406e-23, 2.4545405e-23], [2033.328, 3.4780905e-23, 2.4599042e-23, 2.4599041e-23], [2037.9369, 3.4856965e-23, 2.4652804e-23, 2.4652804e-23], [2042.5562, 3.4933204e-23, 2.4706693e-23, 2.4706692e-23], [2047.1859, 3.5009621e-23, 2.4760708e-23, 2.4760708e-23], [2051.8261, 3.5086217e-23, 2.481485e-23, 2.481485e-23], [2056.4769, 3.5162994e-23, 2.4869119e-23, 2.4869119e-23], [2061.1382, 3.523995e-23, 2.4923516e-23, 2.4923516e-23], [2065.8101, 3.5317087e-23, 2.497804e-23, 2.497804e-23], [2070.4925, 3.5394405e-23, 2.5032693e-23, 2.5032692e-23], [2075.1856, 3.5471904e-23, 2.5087473e-23, 2.5087473e-23], [2079.8893, 3.5549585e-23, 2.5142383e-23, 2.5142382e-23], [2084.6037, 3.5627447e-23, 2.5197421e-23, 2.519742e-23], [2089.3287, 3.5705493e-23, 2.5252588e-23, 2.5252587e-23], [2094.0645, 3.5783721e-23, 2.5307884e-23, 2.5307884e-23], [2098.811, 3.5862133e-23, 2.5363311e-23, 2.536331e-23], [2103.5682, 3.5940728e-23, 2.5418867e-23, 2.5418866e-23], [2108.3363, 3.6019507e-23, 2.5474553e-23, 2.5474553e-23], [2113.1151, 3.6098471e-23, 2.553037e-23, 2.553037e-23], [2117.9048, 3.617762e-23, 2.5586318e-23, 2.5586318e-23], [2122.7053, 3.6256954e-23, 2.5642397e-23, 2.5642397e-23], [2127.5168, 3.6336474e-23, 2.5698608e-23, 2.5698607e-23], [2132.3391, 3.641618e-23, 2.575495e-23, 2.5754949e-23], [2137.1723, 3.6496073e-23, 2.5811424e-23, 2.5811423e-23], [2142.0165, 3.6576152e-23, 2.586803e-23, 2.586803e-23], [2146.8717, 3.6656419e-23, 2.5924769e-23, 2.5924769e-23], [2151.7379, 3.6736874e-23, 2.5981641e-23, 2.598164e-23], [2156.6151, 3.6817517e-23, 2.6038646e-23, 2.6038645e-23], [2161.5034, 3.6898349e-23, 2.6095784e-23, 2.6095784e-23], [2166.4028, 3.6979369e-23, 2.6153056e-23, 2.6153056e-23], [2171.3132, 3.7060579e-23, 2.6210462e-23, 2.6210462e-23], [2176.2348, 3.7141979e-23, 2.6268003e-23, 2.6268002e-23], [2181.1676, 3.7223569e-23, 2.6325678e-23, 2.6325677e-23], [2186.1115, 3.730535e-23, 2.6383487e-23, 2.6383487e-23], [2191.0667, 3.7387322e-23, 2.6441433e-23, 2.6441432e-23], [2196.033, 3.7469485e-23, 2.6499513e-23, 2.6499513e-23], [2201.0106, 3.7551841e-23, 2.655773e-23, 2.6557729e-23], [2205.9996, 3.7634389e-23, 2.6616082e-23, 2.6616082e-23], [2210.9998, 3.7717129e-23, 2.6674571e-23, 2.6674571e-23], [2216.0113, 3.7800063e-23, 2.6733197e-23, 2.6733197e-23], [2221.0342, 3.7883191e-23, 2.679196e-23, 2.6791959e-23], [2226.0685, 3.7966512e-23, 2.685086e-23, 2.6850859e-23], [2231.1142, 3.8050029e-23, 2.6909897e-23, 2.6909897e-23], [2236.1714, 3.813374e-23, 2.6969073e-23, 2.6969072e-23], [2241.24, 3.8217646e-23, 2.7028386e-23, 2.7028386e-23], [2246.3201, 3.8301748e-23, 2.7087839e-23, 2.7087838e-23], [2251.4117, 3.8386047e-23, 2.714743e-23, 2.7147429e-23], [2256.5148, 3.8470542e-23, 2.720716e-23, 2.720716e-23], [2261.6295, 3.8555234e-23, 2.726703e-23, 2.7267029e-23], [2266.7558, 3.8640124e-23, 2.7327039e-23, 2.7327039e-23], [2271.8938, 3.8725212e-23, 2.7387188e-23, 2.7387188e-23], [2277.0433, 3.8810498e-23, 2.7447478e-23, 2.7447478e-23], [2282.2046, 3.8895983e-23, 2.7507909e-23, 2.7507908e-23], [2287.3775, 3.8981668e-23, 2.756848e-23, 2.756848e-23], [2292.5622, 3.9067552e-23, 2.7629193e-23, 2.7629193e-23], [2297.7586, 3.9153636e-23, 2.7690047e-23, 2.7690047e-23], [2302.9668, 3.9239921e-23, 2.7751044e-23, 2.7751043e-23], [2308.1868, 3.9326407e-23, 2.7812182e-23, 2.7812182e-23], [2313.4187, 3.9413094e-23, 2.7873463e-23, 2.7873463e-23], [2318.6624, 3.9499984e-23, 2.7934887e-23, 2.7934887e-23], [2323.9179, 3.9587075e-23, 2.7996454e-23, 2.7996454e-23], [2329.1854, 3.967437e-23, 2.8058165e-23, 2.8058164e-23], [2334.4649, 3.9761868e-23, 2.8120019e-23, 2.8120019e-23], [2339.7563, 3.9849569e-23, 2.8182018e-23, 2.8182017e-23], [2345.0597, 3.9937475e-23, 2.8244161e-23, 2.824416e-23], [2350.3751, 4.0025586e-23, 2.8306448e-23, 2.8306448e-23], [2355.7025, 4.0113901e-23, 2.8368881e-23, 2.8368881e-23], [2361.0421, 4.0202422e-23, 2.8431459e-23, 2.8431459e-23], [2366.3937, 4.0291149e-23, 2.8494183e-23, 2.8494183e-23], [2371.7575, 4.0380083e-23, 2.8557053e-23, 2.8557053e-23], [2377.1334, 4.0469223e-23, 2.8620069e-23, 2.8620069e-23], [2382.5215, 4.0558571e-23, 2.8683233e-23, 2.8683232e-23], [2387.9219, 4.0648127e-23, 2.8746543e-23, 2.8746543e-23], [2393.3344, 4.0737892e-23, 2.8810001e-23, 2.8810001e-23], [2398.7593, 4.0827871e-23, 2.8873616e-23, 2.8873616e-23], [2404.1964, 4.0918061e-23, 2.893738e-23, 2.893738e-23], [2409.6459, 4.1008438e-23, 2.9001261e-23, 2.9001261e-23], [2415.1077, 4.1099039e-23, 2.906531e-23, 2.9065309e-23], [2420.5819, 4.1189851e-23, 2.9129508e-23, 2.9129508e-23], [2426.0685, 4.1280874e-23, 2.9193856e-23, 2.9193856e-23], [2431.5675, 4.1372109e-23, 2.9258354e-23, 2.9258354e-23], [2437.079, 4.1463556e-23, 2.9323002e-23, 2.9323002e-23], [2442.603, 4.1555215e-23, 2.93878e-23, 2.93878e-23], [2448.1395, 4.1647087e-23, 2.9452749e-23, 2.9452749e-23], [2453.6885, 4.1739173e-23, 2.9517849e-23, 2.9517848e-23], [2459.2502, 4.1831472e-23, 2.95831e-23, 2.9583099e-23], [2464.8244, 4.1923986e-23, 2.9648502e-23, 2.9648502e-23], [2470.4113, 4.2016715e-23, 2.9714057e-23, 2.9714057e-23], [2476.0108, 4.2109659e-23, 2.9779764e-23, 2.9779763e-23], [2481.6231, 4.2202819e-23, 2.9845623e-23, 2.9845623e-23], [2487.248, 4.2296195e-23, 2.9911636e-23, 2.9911636e-23], [2492.8857, 4.2389788e-23, 2.9977802e-23, 2.9977801e-23], [2498.5362, 4.2483598e-23, 3.0044121e-23, 3.0044121e-23], [2504.1995, 4.2577626e-23, 3.0110595e-23, 3.0110594e-23], [2509.8756, 4.2671872e-23, 3.0177223e-23, 3.0177222e-23], [2515.5646, 4.2766337e-23, 3.0244005e-23, 3.0244005e-23], [2521.2665, 4.286102e-23, 3.0310943e-23, 3.0310942e-23], [2526.9813, 4.2955924e-23, 3.0378035e-23, 3.0378035e-23], [2532.7091, 4.3051047e-23, 3.0445284e-23, 3.0445284e-23], [2538.4499, 4.3146392e-23, 3.0512689e-23, 3.0512689e-23], [2544.2036, 4.3241957e-23, 3.058025e-23, 3.058025e-23], [2549.9704, 4.3337744e-23, 3.0647968e-23, 3.0647967e-23], [2555.7503, 4.3433753e-23, 3.0715843e-23, 3.0715842e-23], [2561.5433, 4.3529984e-23, 3.0783875e-23, 3.0783875e-23], [2567.3494, 4.3626439e-23, 3.0852065e-23, 3.0852065e-23], [2573.1686, 4.3723117e-23, 3.0920413e-23, 3.0920413e-23], [2579.0011, 4.382002e-23, 3.098892e-23, 3.098892e-23], [2584.8468, 4.3917147e-23, 3.1057586e-23, 3.1057586e-23], [2590.7057, 4.4014498e-23, 3.1126411e-23, 3.112641e-23], [2596.5779, 4.4112076e-23, 3.1195395e-23, 3.1195395e-23], [2602.4634, 4.420988e-23, 3.1264539e-23, 3.1264539e-23], [2608.3623, 4.430791e-23, 3.1333844e-23, 3.1333844e-23], [2614.2745, 4.4406167e-23, 3.1403309e-23, 3.1403309e-23], [2620.2001, 4.4504652e-23, 3.1472935e-23, 3.1472935e-23], [2626.1392, 4.4603364e-23, 3.1542722e-23, 3.1542722e-23], [2632.0917, 4.4702306e-23, 3.1612671e-23, 3.1612671e-23], [2638.0577, 4.4801476e-23, 3.1682783e-23, 3.1682782e-23], [2644.0373, 4.4900876e-23, 3.1753056e-23, 3.1753056e-23], [2650.0304, 4.5000507e-23, 3.1823492e-23, 3.1823492e-23], [2656.037, 4.5100368e-23, 3.1894092e-23, 3.1894092e-23], [2662.0573, 4.520046e-23, 3.1964855e-23, 3.1964855e-23], [2668.0913, 4.5300783e-23, 3.2035782e-23, 3.2035781e-23], [2674.1389, 4.5401339e-23, 3.2106873e-23, 3.2106872e-23], [2680.2002, 4.5502127e-23, 3.2178128e-23, 3.2178128e-23], [2686.2753, 4.5603149e-23, 3.2249549e-23, 3.2249548e-23], [2692.3641, 4.5704404e-23, 3.2321134e-23, 3.2321134e-23], [2698.4667, 4.5805894e-23, 3.2392886e-23, 3.2392886e-23], [2704.5832, 4.5907618e-23, 3.2464803e-23, 3.2464803e-23], [2710.7135, 4.6009578e-23, 3.2536887e-23, 3.2536887e-23], [2716.8577, 4.6111773e-23, 3.2609138e-23, 3.2609138e-23], [2723.0159, 4.6214205e-23, 3.2681556e-23, 3.2681556e-23], [2729.188, 4.6316874e-23, 3.2754141e-23, 3.2754141e-23], [2735.3741, 4.641978e-23, 3.2826895e-23, 3.2826895e-23], [2741.5742, 4.6522924e-23, 3.2899816e-23, 3.2899816e-23], [2747.7884, 4.6626306e-23, 3.2972906e-23, 3.2972906e-23], [2754.0166, 4.6729927e-23, 3.3046166e-23, 3.3046166e-23], [2760.259, 4.6833788e-23, 3.3119594e-23, 3.3119594e-23], [2766.5155, 4.6937889e-23, 3.3193193e-23, 3.3193193e-23], [2772.7862, 4.704223e-23, 3.3266961e-23, 3.3266961e-23], [2779.0712, 4.7146813e-23, 3.33409e-23, 3.33409e-23], [2785.3703, 4.7251637e-23, 3.341501e-23, 3.341501e-23], [2791.6838, 4.7356703e-23, 3.3489291e-23, 3.3489291e-23], [2798.0115, 4.7462012e-23, 3.3563744e-23, 3.3563744e-23], [2804.3536, 4.7567564e-23, 3.3638369e-23, 3.3638369e-23], [2810.7101, 4.767336e-23, 3.3713167e-23, 3.3713167e-23], [2817.081, 4.7779401e-23, 3.3788137e-23, 3.3788137e-23], [2823.4663, 4.7885686e-23, 3.386328e-23, 3.386328e-23], [2829.8661, 4.7992217e-23, 3.3938597e-23, 3.3938597e-23], [2836.2804, 4.8098993e-23, 3.4014088e-23, 3.4014088e-23], [2842.7092, 4.8206016e-23, 3.4089754e-23, 3.4089754e-23], [2849.1526, 4.8313287e-23, 3.4165594e-23, 3.4165594e-23], [2855.6106, 4.8420804e-23, 3.4241609e-23, 3.4241609e-23], [2862.0833, 4.852857e-23, 3.43178e-23, 3.43178e-23], [2868.5706, 4.8636585e-23, 3.4394167e-23, 3.4394167e-23], [2875.0726, 4.874485e-23, 3.447071e-23, 3.447071e-23], [2881.5894, 4.885343e-23, 3.4547523e-23, 3.4547523e-23], [2888.1209, 4.8962128e-23, 3.4624327e-23, 3.4624327e-23], [2894.6673, 4.9071143e-23, 3.47014e-23, 3.47014e-23], [2901.2285, 4.9180409e-23, 3.4778652e-23, 3.4778652e-23], [2907.8045, 4.9289928e-23, 3.4856083e-23, 3.4856082e-23], [2914.3955, 4.9399699e-23, 3.4933692e-23, 3.4933692e-23], [2921.0014, 4.9509724e-23, 3.501148e-23, 3.501148e-23], [2927.6222, 4.9620002e-23, 3.5089447e-23, 3.5089447e-23], [2934.2581, 4.9730534e-23, 3.5167595e-23, 3.5167595e-23], [2940.909, 4.9841322e-23, 3.5245923e-23, 3.5245922e-23], [2947.575, 4.9952365e-23, 3.5324431e-23, 3.5324431e-23], [2954.2561, 5.0063664e-23, 3.5403121e-23, 3.540312e-23], [2960.9524, 5.0175219e-23, 3.5481992e-23, 3.5481991e-23], [2967.6638, 5.0287032e-23, 3.5561044e-23, 3.5561044e-23], [2974.3904, 5.0399103e-23, 3.564028e-23, 3.5640279e-23], [2981.1323, 5.0511431e-23, 3.5719697e-23, 3.5719697e-23], [2987.8895, 5.0624019e-23, 3.5799298e-23, 3.5799298e-23], [2994.662, 5.0736866e-23, 3.5879083e-23, 3.5879083e-23], [3001.4498, 5.0849973e-23, 3.5959051e-23, 3.5959051e-23], [3008.253, 5.0963341e-23, 3.6039204e-23, 3.6039204e-23], [3015.0716, 5.107697e-23, 3.6119541e-23, 3.6119541e-23], [3021.9057, 5.1190861e-23, 3.6200064e-23, 3.6200064e-23], [3028.7553, 5.1305014e-23, 3.6280772e-23, 3.6280772e-23], [3035.6204, 5.1419431e-23, 3.6361667e-23, 3.6361666e-23], [3042.5011, 5.153411e-23, 3.6442747e-23, 3.6442747e-23], [3049.3974, 5.1649054e-23, 3.6524014e-23, 3.6524014e-23], [3056.3093, 5.1764263e-23, 3.6605469e-23, 3.6605469e-23], [3063.2368, 5.1879737e-23, 3.6687111e-23, 3.6687111e-23], [3070.1801, 5.1995476e-23, 3.6768942e-23, 3.6768942e-23], [3077.1391, 5.2111483e-23, 3.685096e-23, 3.685096e-23], [3084.1139, 5.2227756e-23, 3.6933168e-23, 3.6933168e-23], [3091.1045, 5.2344297e-23, 3.7015565e-23, 3.7015565e-23], [3098.1109, 5.2461107e-23, 3.7098152e-23, 3.7098152e-23], [3105.1332, 5.2578185e-23, 3.7180929e-23, 3.7180929e-23], [3112.1714, 5.2695533e-23, 3.7263897e-23, 3.7263896e-23], [3119.2256, 5.281315e-23, 3.7347055e-23, 3.7347055e-23], [3126.2958, 5.2931039e-23, 3.7430405e-23, 3.7430405e-23], [3133.382, 5.3049199e-23, 3.7513947e-23, 3.7513947e-23], [3140.4843, 5.3167631e-23, 3.7597681e-23, 3.7597681e-23], [3147.6026, 5.3286335e-23, 3.7681608e-23, 3.7681608e-23], [3154.7371, 5.3405312e-23, 3.7765728e-23, 3.7765728e-23], [3161.8878, 5.3524564e-23, 3.7850042e-23, 3.7850042e-23], [3169.0546, 5.3644089e-23, 3.793455e-23, 3.793455e-23], [3176.2378, 5.3763889e-23, 3.8019252e-23, 3.8019252e-23], [3183.4372, 5.3883965e-23, 3.8104149e-23, 3.8104149e-23], [3190.6529, 5.4004318e-23, 3.8189241e-23, 3.8189241e-23], [3197.885, 5.4124947e-23, 3.827453e-23, 3.8274529e-23], [3205.1334, 5.4245853e-23, 3.8360014e-23, 3.8360014e-23], [3212.3983, 5.4367038e-23, 3.8445695e-23, 3.8445695e-23], [3219.6797, 5.4488501e-23, 3.8531573e-23, 3.8531573e-23], [3226.9775, 5.4610243e-23, 3.8617649e-23, 3.8617649e-23], [3234.2919, 5.4732265e-23, 3.8703922e-23, 3.8703922e-23], [3241.6229, 5.4854568e-23, 3.8790394e-23, 3.8790394e-23], [3248.9705, 5.4977152e-23, 3.8877065e-23, 3.8877065e-23], [3256.3348, 5.5100018e-23, 3.8963935e-23, 3.8963935e-23], [3263.7158, 5.5223166e-23, 3.9051005e-23, 3.9051005e-23], [3271.1134, 5.5346598e-23, 3.9138275e-23, 3.9138275e-23], [3278.5279, 5.5470313e-23, 3.9225746e-23, 3.9225746e-23], [3285.9591, 5.5594312e-23, 3.9313418e-23, 3.9313417e-23], [3293.4072, 5.5718596e-23, 3.9401291e-23, 3.9401291e-23], [3300.8722, 5.5843166e-23, 3.9489366e-23, 3.9489366e-23], [3308.3541, 5.5968022e-23, 3.9577644e-23, 3.9577644e-23], [3315.853, 5.6093165e-23, 3.9666124e-23, 3.9666124e-23], [3323.3688, 5.6218596e-23, 3.9754808e-23, 3.9754808e-23], [3330.9017, 5.6344314e-23, 3.9843696e-23, 3.9843696e-23], [3338.4517, 5.6470322e-23, 3.9932788e-23, 3.9932788e-23], [3346.0188, 5.6596619e-23, 4.0022085e-23, 4.0022085e-23], [3353.603, 5.6723206e-23, 4.0111587e-23, 4.0111587e-23], [3361.2044, 5.6850091e-23, 4.0201305e-23, 4.0201305e-23], [3368.8231, 5.6977253e-23, 4.0291208e-23, 4.0291208e-23], [3376.459, 5.7104714e-23, 4.0381328e-23, 4.0381328e-23], [3384.1122, 5.7232468e-23, 4.0471655e-23, 4.0471655e-23], [3391.7828, 5.7360515e-23, 4.0562189e-23, 4.0562189e-23], [3399.4708, 5.7488856e-23, 4.0652932e-23, 4.0652932e-23], [3407.1762, 5.7617492e-23, 4.0743883e-23, 4.0743883e-23], [3414.899, 5.7746424e-23, 4.0835042e-23, 4.0835042e-23], [3422.6394, 5.7875651e-23, 4.0926411e-23, 4.0926411e-23], [3430.3973, 5.8005175e-23, 4.101799e-23, 4.101799e-23], [3438.1728, 5.8134996e-23, 4.1109779e-23, 4.1109779e-23], [3445.9659, 5.8265115e-23, 4.1201779e-23, 4.1201779e-23], [3453.7767, 5.8395533e-23, 4.129399e-23, 4.129399e-23], [3461.6051, 5.8526251e-23, 4.1386413e-23, 4.1386413e-23], [3469.4514, 5.8657268e-23, 4.1479048e-23, 4.1479048e-23], [3477.3154, 5.8788586e-23, 4.1571896e-23, 4.1571896e-23], [3485.1972, 5.8920205e-23, 4.1664956e-23, 4.1664956e-23], [3493.0969, 5.9052126e-23, 4.1758231e-23, 4.1758231e-23], [3501.0145, 5.918435e-23, 4.1851719e-23, 4.1851719e-23], [3508.9501, 5.9316877e-23, 4.1945422e-23, 4.1945422e-23], [3516.9036, 5.9449709e-23, 4.203934e-23, 4.203934e-23], [3524.8752, 5.9582845e-23, 4.2133473e-23, 4.2133473e-23], [3532.8648, 5.9716286e-23, 4.2227823e-23, 4.2227823e-23], [3540.8726, 5.9850034e-23, 4.2322388e-23, 4.2322388e-23], [3548.8985, 5.9984088e-23, 4.2417171e-23, 4.2417171e-23], [3556.9425, 6.011845e-23, 4.2512172e-23, 4.2512171e-23], [3565.0049, 6.025312e-23, 4.260739e-23, 4.260739e-23], [3573.0855, 6.0388099e-23, 4.2702826e-23, 4.2702826e-23], [3581.1844, 6.0523387e-23, 4.2798482e-23, 4.2798482e-23], [3589.3016, 6.0658985e-23, 4.2894357e-23, 4.2894357e-23], [3597.4373, 6.0794895e-23, 4.2990451e-23, 4.2990451e-23], [3605.5914, 6.0931116e-23, 4.3086767e-23, 4.3086767e-23], [3613.764, 6.106765e-23, 4.3183303e-23, 4.3183303e-23], [3621.9551, 6.1204496e-23, 4.328006e-23, 4.328006e-23], [3630.1648, 6.1341656e-23, 4.3377039e-23, 4.3377039e-23], [3638.3931, 6.1479131e-23, 4.3474241e-23, 4.3474241e-23], [3646.64, 6.1616921e-23, 4.3571666e-23, 4.3571666e-23], [3654.9057, 6.1755026e-23, 4.3669313e-23, 4.3669313e-23], [3663.19, 6.1893448e-23, 4.3767185e-23, 4.3767185e-23], [3671.4932, 6.2032188e-23, 4.3865281e-23, 4.3865281e-23], [3679.8152, 6.2171245e-23, 4.3963602e-23, 4.3963602e-23], [3688.156, 6.2310621e-23, 4.4062149e-23, 4.4062148e-23], [3696.5157, 6.2450317e-23, 4.4160921e-23, 4.4160921e-23], [3704.8944, 6.2590332e-23, 4.4259919e-23, 4.4259919e-23], [3713.2921, 6.2730669e-23, 4.4359145e-23, 4.4359145e-23], [3721.7088, 6.2871327e-23, 4.4458598e-23, 4.4458598e-23], [3730.1446, 6.3012307e-23, 4.4558278e-23, 4.4558278e-23], [3738.5995, 6.315361e-23, 4.4658188e-23, 4.4658188e-23], [3747.0736, 6.3295237e-23, 4.4758326e-23, 4.4758326e-23], [3755.5668, 6.3437188e-23, 4.4858693e-23, 4.4858693e-23], [3764.0794, 6.3579465e-23, 4.4959291e-23, 4.4959291e-23], [3772.6112, 6.3722067e-23, 4.5060119e-23, 4.5060119e-23], [3781.1624, 6.3864996e-23, 4.5161178e-23, 4.5161178e-23], [3789.7329, 6.4008253e-23, 4.5262468e-23, 4.5262468e-23], [3798.3229, 6.4151837e-23, 4.5363991e-23, 4.5363991e-23], [3806.9323, 6.4295751e-23, 4.5465746e-23, 4.5465746e-23], [3815.5613, 6.4439994e-23, 4.5567734e-23, 4.5567734e-23], [3824.2098, 6.4584567e-23, 4.5669956e-23, 4.5669956e-23], [3832.8779, 6.4729471e-23, 4.5772412e-23, 4.5772412e-23], [3841.5657, 6.4874712e-23, 4.5875108e-23, 4.5875108e-23], [3850.2732, 6.5020276e-23, 4.5978028e-23, 4.5978028e-23], [3859.0004, 6.5166178e-23, 4.6081189e-23, 4.6081189e-23], [3867.7473, 6.5312414e-23, 4.6184587e-23, 4.6184587e-23], [3876.5141, 6.5458985e-23, 4.6288221e-23, 4.6288221e-23], [3885.3008, 6.5605891e-23, 4.6392093e-23, 4.6392093e-23], [3894.1074, 6.5753134e-23, 4.6496202e-23, 4.6496202e-23], [3902.934, 6.5900713e-23, 4.660055e-23, 4.660055e-23], [3911.7805, 6.6048631e-23, 4.6705137e-23, 4.6705137e-23], [3920.6471, 6.6196887e-23, 4.6809963e-23, 4.6809963e-23], [3929.5339, 6.6345482e-23, 4.6915029e-23, 4.6915029e-23], [3938.4407, 6.6494418e-23, 4.7020335e-23, 4.7020335e-23], [3947.3677, 6.6643694e-23, 4.7125883e-23, 4.7125883e-23], [3956.315, 6.6793312e-23, 4.7231672e-23, 4.7231672e-23], [3965.2826, 6.6943272e-23, 4.7337703e-23, 4.7337703e-23], [3974.2705, 6.7093575e-23, 4.7443977e-23, 4.7443977e-23], [3983.2787, 6.7244223e-23, 4.7550495e-23, 4.7550495e-23], [3992.3074, 6.7395215e-23, 4.7657256e-23, 4.7657256e-23], [4001.3565, 6.7546552e-23, 4.7764261e-23, 4.7764261e-23], [4010.4262, 6.7698236e-23, 4.7871511e-23, 4.7871511e-23], [4019.5164, 6.7850266e-23, 4.7979006e-23, 4.7979006e-23], [4028.6272, 6.8002645e-23, 4.8086748e-23, 4.8086748e-23], [4037.7587, 6.8155372e-23, 4.8194736e-23, 4.8194736e-23], [4046.9108, 6.8308449e-23, 4.8302971e-23, 4.8302971e-23], [4056.0837, 6.8461875e-23, 4.8411454e-23, 4.8411453e-23], [4065.2774, 6.8615653e-23, 4.8520184e-23, 4.8520184e-23], [4074.492, 6.8769782e-23, 4.8629164e-23, 4.8629164e-23], [4083.7274, 6.8924264e-23, 4.8738393e-23, 4.8738393e-23], [4092.9837, 6.9079099e-23, 4.8847871e-23, 4.8847871e-23], [4102.2611, 6.9234288e-23, 4.89576e-23, 4.89576e-23], [4111.5594, 6.9389832e-23, 4.906758e-23, 4.906758e-23], [4120.8789, 6.9545732e-23, 4.9177812e-23, 4.9177812e-23], [4130.2194, 6.9701988e-23, 4.9288296e-23, 4.9288296e-23], [4139.5812, 6.9858602e-23, 4.9399032e-23, 4.9399032e-23], [4148.9641, 7.0015573e-23, 4.9510022e-23, 4.9510022e-23], [4158.3684, 7.0172904e-23, 4.9621265e-23, 4.9621265e-23], [4167.7939, 7.0330594e-23, 4.9732763e-23, 4.9732763e-23], [4177.2408, 7.0488645e-23, 4.9844516e-23, 4.9844515e-23], [4186.7091, 7.0647057e-23, 4.9956524e-23, 4.9956524e-23], [4196.1989, 7.0805831e-23, 5.0068788e-23, 5.0068788e-23], [4205.7102, 7.0964968e-23, 5.0181309e-23, 5.0181309e-23], [4215.243, 7.1124469e-23, 5.0294087e-23, 5.0294087e-23], [4224.7975, 7.1284335e-23, 5.0407123e-23, 5.0407123e-23], [4234.3736, 7.1444566e-23, 5.0520418e-23, 5.0520418e-23], [4243.9714, 7.1605163e-23, 5.0633971e-23, 5.0633971e-23], [4253.591, 7.1766127e-23, 5.0747784e-23, 5.0747784e-23], [4263.2324, 7.1927459e-23, 5.0861857e-23, 5.0861857e-23], [4272.8956, 7.208916e-23, 5.0976191e-23, 5.0976191e-23], [4282.5807, 7.225123e-23, 5.1090786e-23, 5.1090786e-23], [4292.2878, 7.2413671e-23, 5.1205643e-23, 5.1205643e-23], [4302.0169, 7.2576482e-23, 5.1320763e-23, 5.1320763e-23], [4311.768, 7.2739666e-23, 5.1436146e-23, 5.1436146e-23], [4321.5413, 7.2903225e-23, 5.1551794e-23, 5.1551794e-23], [4331.3367, 7.3067154e-23, 5.1667703e-23, 5.1667703e-23], [4341.1543, 7.3231458e-23, 5.1783878e-23, 5.1783878e-23], [4350.9941, 7.3396139e-23, 5.1900319e-23, 5.1900319e-23], [4360.8563, 7.3561195e-23, 5.2017026e-23, 5.2017026e-23], [4370.7408, 7.3726629e-23, 5.2134e-23, 5.2134e-23], [4380.6477, 7.3892441e-23, 5.225124e-23, 5.225124e-23], [4390.5771, 7.4058631e-23, 5.2368749e-23, 5.2368749e-23], [4400.5289, 7.4225201e-23, 5.2486526e-23, 5.2486526e-23], [4410.5034, 7.4392151e-23, 5.2604573e-23, 5.2604573e-23], [4420.5004, 7.4559483e-23, 5.2722889e-23, 5.2722889e-23], [4430.5201, 7.4727197e-23, 5.2841475e-23, 5.2841475e-23], [4440.5625, 7.4895294e-23, 5.2960332e-23, 5.2960332e-23], [4450.6277, 7.5063775e-23, 5.307946e-23, 5.307946e-23], [4460.7157, 7.5232641e-23, 5.3198861e-23, 5.3198861e-23], [4470.8265, 7.5401892e-23, 5.3318534e-23, 5.3318534e-23], [4480.9603, 7.557153e-23, 5.3438481e-23, 5.3438481e-23], [4491.117, 7.5741555e-23, 5.3558702e-23, 5.3558702e-23], [4501.2968, 7.5911969e-23, 5.3679197e-23, 5.3679197e-23], [4511.4996, 7.6082772e-23, 5.3799967e-23, 5.3799967e-23], [4521.7256, 7.6253964e-23, 5.3921013e-23, 5.3921013e-23], [4531.9747, 7.6425548e-23, 5.4042336e-23, 5.4042336e-23], [4542.2471, 7.6597523e-23, 5.4163936e-23, 5.4163936e-23], [4552.5428, 7.6769891e-23, 5.4285813e-23, 5.4285813e-23], [4562.8618, 7.6942652e-23, 5.4407968e-23, 5.4407968e-23], [4573.2041, 7.7115808e-23, 5.4530403e-23, 5.4530403e-23], [4583.57, 7.7289359e-23, 5.4653117e-23, 5.4653117e-23], [4593.9593, 7.7463307e-23, 5.4776111e-23, 5.4776111e-23], [4604.3722, 7.7637651e-23, 5.4899386e-23, 5.4899386e-23], [4614.8086, 7.7812394e-23, 5.5022942e-23, 5.5022942e-23], [4625.2688, 7.7987535e-23, 5.5146781e-23, 5.5146781e-23], [4635.7526, 7.8163076e-23, 5.5270902e-23, 5.5270902e-23], [4646.2602, 7.8339018e-23, 5.5395306e-23, 5.5395306e-23], [4656.7916, 7.8515361e-23, 5.5519995e-23, 5.5519995e-23], [4667.3469, 7.8692107e-23, 5.5644968e-23, 5.5644968e-23], [4677.9261, 7.8869256e-23, 5.5770226e-23, 5.5770226e-23], [4688.5293, 7.9046809e-23, 5.589577e-23, 5.589577e-23], [4699.1565, 7.9224768e-23, 5.6021601e-23, 5.6021601e-23], [4709.8078, 7.9403133e-23, 5.6147719e-23, 5.6147719e-23], [4720.4833, 7.9581904e-23, 5.6274125e-23, 5.6274125e-23], [4731.1829, 7.9761084e-23, 5.6400819e-23, 5.6400819e-23], [4741.9069, 7.9940673e-23, 5.6527802e-23, 5.6527802e-23], [4752.6551, 8.0120671e-23, 5.6655076e-23, 5.6655076e-23], [4763.4276, 8.030108e-23, 5.6782639e-23, 5.6782639e-23], [4774.2246, 8.04819e-23, 5.6910494e-23, 5.6910494e-23], [4785.0461, 8.0663134e-23, 5.703864e-23, 5.703864e-23], [4795.8921, 8.084478e-23, 5.7167079e-23, 5.7167079e-23], [4806.7627, 8.1026842e-23, 5.7295811e-23, 5.7295811e-23], [4817.6579, 8.1209318e-23, 5.7424836e-23, 5.7424836e-23], [4828.5778, 8.1392211e-23, 5.7554156e-23, 5.7554156e-23], [4839.5225, 8.1575521e-23, 5.7683771e-23, 5.7683771e-23], [4850.4919, 8.1759249e-23, 5.7813682e-23, 5.7813682e-23], [4861.4863, 8.1943396e-23, 5.7943889e-23, 5.7943889e-23], [4872.5055, 8.2127963e-23, 5.8074393e-23, 5.8074393e-23], [4883.5498, 8.2312951e-23, 5.8205194e-23, 5.8205194e-23], [4894.619, 8.2498362e-23, 5.8336294e-23, 5.8336294e-23], [4905.7134, 8.2684195e-23, 5.8467694e-23, 5.8467694e-23], [4916.8329, 8.2870452e-23, 5.8599393e-23, 5.8599392e-23], [4927.9776, 8.3057133e-23, 5.8731392e-23, 5.8731392e-23], [4939.1476, 8.3244241e-23, 5.8863692e-23, 5.8863692e-23], [4950.3429, 8.3431775e-23, 5.8996294e-23, 5.8996294e-23], [4961.5635, 8.3619737e-23, 5.9129199e-23, 5.9129199e-23], [4972.8096, 8.3808127e-23, 5.9262406e-23, 5.9262406e-23], [4984.0812, 8.3996947e-23, 5.9395918e-23, 5.9395918e-23], [4995.3784, 8.4186198e-23, 5.9529734e-23, 5.9529734e-23], [5006.7011, 8.437588e-23, 5.9663855e-23, 5.9663855e-23], [5018.0495, 8.4565995e-23, 5.9798282e-23, 5.9798282e-23], [5029.4236, 8.4756543e-23, 5.9933015e-23, 5.9933015e-23], [5040.8236, 8.4947526e-23, 6.0068056e-23, 6.0068056e-23], [5052.2493, 8.5138944e-23, 6.0203404e-23, 6.0203404e-23], [5063.701, 8.5330799e-23, 6.0339062e-23, 6.0339062e-23], [5075.1786, 8.5523091e-23, 6.0475028e-23, 6.0475028e-23], [5086.6822, 8.5715821e-23, 6.0611305e-23, 6.0611305e-23], [5098.2119, 8.5908991e-23, 6.0747892e-23, 6.0747892e-23], [5109.7677, 8.6102601e-23, 6.0884791e-23, 6.0884791e-23], [5121.3497, 8.6296653e-23, 6.1022001e-23, 6.1022001e-23], [5132.958, 8.6491146e-23, 6.1159525e-23, 6.1159525e-23], [5144.5926, 8.6686084e-23, 6.1297362e-23, 6.1297362e-23], [5156.2535, 8.6881465e-23, 6.1435514e-23, 6.1435514e-23], [5167.9409, 8.7077292e-23, 6.157398e-23, 6.157398e-23], [5179.6548, 8.7273566e-23, 6.1712762e-23, 6.1712762e-23], [5191.3953, 8.7470286e-23, 6.185186e-23, 6.185186e-23], [5203.1623, 8.7667455e-23, 6.1991275e-23, 6.1991275e-23], [5214.956, 8.7865074e-23, 6.2131008e-23, 6.2131008e-23], [5226.7765, 8.8063143e-23, 6.227106e-23, 6.227106e-23], [5238.6237, 8.8261663e-23, 6.2411431e-23, 6.2411431e-23], [5250.4978, 8.8460636e-23, 6.2552122e-23, 6.2552122e-23], [5262.3988, 8.8660062e-23, 6.2693133e-23, 6.2693133e-23], [5274.3268, 8.8859942e-23, 6.2834466e-23, 6.2834466e-23], [5286.2818, 8.9060279e-23, 6.2976121e-23, 6.2976121e-23], [5298.2639, 8.9261071e-23, 6.3118098e-23, 6.3118098e-23], [5310.2732, 8.9462321e-23, 6.32604e-23, 6.32604e-23], [5322.3097, 8.966403e-23, 6.3403025e-23, 6.3403025e-23], [5334.3735, 8.9866199e-23, 6.3545976e-23, 6.3545976e-23], [5346.4646, 9.0068828e-23, 6.3689252e-23, 6.3689252e-23], [5358.5831, 9.0271919e-23, 6.3832855e-23, 6.3832855e-23], [5370.7291, 9.0475472e-23, 6.3976785e-23, 6.3976785e-23], [5382.9027, 9.0679489e-23, 6.4121043e-23, 6.4121043e-23], [5395.1038, 9.0883972e-23, 6.426563e-23, 6.426563e-23], [5407.3326, 9.108892e-23, 6.4410546e-23, 6.4410546e-23], [5419.5891, 9.1294335e-23, 6.4555792e-23, 6.4555792e-23], [5431.8733, 9.1500217e-23, 6.4701369e-23, 6.4701369e-23], [5444.1855, 9.1706569e-23, 6.4847278e-23, 6.4847278e-23], [5456.5255, 9.1913391e-23, 6.499352e-23, 6.499352e-23], [5468.8935, 9.2120685e-23, 6.5140094e-23, 6.5140094e-23], [5481.2895, 9.232845e-23, 6.5287003e-23, 6.5287003e-23], [5493.7136, 9.2536689e-23, 6.5434246e-23, 6.5434246e-23], [5506.1659, 9.2745402e-23, 6.5581824e-23, 6.5581824e-23], [5518.6465, 9.295459e-23, 6.5729739e-23, 6.5729739e-23], [5531.1553, 9.3164255e-23, 6.5877991e-23, 6.5877991e-23], [5543.6924, 9.3374398e-23, 6.602658e-23, 6.602658e-23], [5556.258, 9.3585019e-23, 6.6175508e-23, 6.6175508e-23], [5568.8521, 9.379612e-23, 6.6324775e-23, 6.6324775e-23], [5581.4747, 9.4007702e-23, 6.6474382e-23, 6.6474382e-23], [5594.1259, 9.4219765e-23, 6.662433e-23, 6.662433e-23], [5606.8058, 9.4432312e-23, 6.6774619e-23, 6.6774619e-23], [5619.5144, 9.4645343e-23, 6.6925251e-23, 6.6925251e-23], [5632.2518, 9.4858858e-23, 6.7076226e-23, 6.7076226e-23], [5645.0181, 9.5072861e-23, 6.7227544e-23, 6.7227544e-23], [5657.8134, 9.528735e-23, 6.7379207e-23, 6.7379207e-23], [5670.6376, 9.5502328e-23, 6.7531216e-23, 6.7531216e-23], [5683.4909, 9.5717795e-23, 6.7683571e-23, 6.7683571e-23], [5696.3734, 9.5933753e-23, 6.7836273e-23, 6.7836273e-23], [5709.285, 9.6150203e-23, 6.7989322e-23, 6.7989322e-23], [5722.226, 9.6367146e-23, 6.814272e-23, 6.814272e-23], [5735.1962, 9.6584583e-23, 6.8296467e-23, 6.8296467e-23], [5748.1959, 9.6802514e-23, 6.8450565e-23, 6.8450565e-23], [5761.225, 9.7020943e-23, 6.8605013e-23, 6.8605013e-23], [5774.2836, 9.7239868e-23, 6.8759813e-23, 6.8759813e-23], [5787.3719, 9.7459292e-23, 6.8914966e-23, 6.8914966e-23], [5800.4898, 9.7679215e-23, 6.9070471e-23, 6.9070471e-23], [5813.6374, 9.7899639e-23, 6.9226331e-23, 6.9226331e-23], [5826.8149, 9.8120565e-23, 6.9382546e-23, 6.9382546e-23], [5840.0222, 9.8341994e-23, 6.9539116e-23, 6.9539116e-23], [5853.2594, 9.8563927e-23, 6.9696043e-23, 6.9696043e-23], [5866.5267, 9.8786365e-23, 6.9853327e-23, 6.9853327e-23], [5879.824, 9.900931e-23, 7.0010969e-23, 7.0010969e-23], [5893.1515, 9.9232762e-23, 7.016897e-23, 7.016897e-23], [5906.5091, 9.9456723e-23, 7.0327331e-23, 7.0327331e-23], [5919.8971, 9.9681194e-23, 7.0486053e-23, 7.0486053e-23], [5933.3154, 9.9906175e-23, 7.0645135e-23, 7.0645135e-23], [5946.7641, 1.0013167e-22, 7.080458e-23, 7.080458e-23], [5960.2433, 1.0035768e-22, 7.0964388e-23, 7.0964388e-23], [5973.753, 1.005842e-22, 7.1124559e-23, 7.1124559e-23], [5987.2934, 1.0081124e-22, 7.1285095e-23, 7.1285095e-23], [6000.8645, 1.0103879e-22, 7.1445997e-23, 7.1445997e-23], [6014.4663, 1.0126686e-22, 7.1607264e-23, 7.1607264e-23], [6028.0989, 1.0149545e-22, 7.1768899e-23, 7.1768899e-23], [6041.7625, 1.0172456e-22, 7.1930902e-23, 7.1930902e-23], [6055.457, 1.019542e-22, 7.2093273e-23, 7.2093273e-23], [6069.1826, 1.0218435e-22, 7.2256014e-23, 7.2256014e-23], [6082.9393, 1.0241503e-22, 7.2419126e-23, 7.2419126e-23], [6096.7271, 1.0264623e-22, 7.2582609e-23, 7.2582609e-23], [6110.5462, 1.0287796e-22, 7.2746463e-23, 7.2746463e-23], [6124.3967, 1.0311022e-22, 7.2910691e-23, 7.2910691e-23], [6138.2785, 1.0334301e-22, 7.3075292e-23, 7.3075292e-23], [6152.1918, 1.0357632e-22, 7.3240268e-23, 7.3240268e-23], [6166.1366, 1.0381017e-22, 7.340562e-23, 7.340562e-23], [6180.113, 1.0404455e-22, 7.3571347e-23, 7.3571347e-23], [6194.1211, 1.0427946e-22, 7.3737452e-23, 7.3737452e-23], [6208.161, 1.0451491e-22, 7.3903935e-23, 7.3903935e-23], [6222.2327, 1.0475089e-22, 7.4070797e-23, 7.4070797e-23], [6236.3363, 1.0498741e-22, 7.4238038e-23, 7.4238038e-23], [6250.4718, 1.0522447e-22, 7.440566e-23, 7.440566e-23], [6264.6394, 1.0546207e-22, 7.4573664e-23, 7.4573664e-23], [6278.8391, 1.057002e-22, 7.4742049e-23, 7.4742049e-23], [6293.071, 1.0593888e-22, 7.4910818e-23, 7.4910818e-23], [6307.3352, 1.0617811e-22, 7.5079971e-23, 7.5079971e-23], [6321.6317, 1.0641787e-22, 7.5249509e-23, 7.5249509e-23], [6335.9605, 1.0665819e-22, 7.5419432e-23, 7.5419432e-23], [6350.3219, 1.0689904e-22, 7.5589742e-23, 7.5589742e-23], [6364.7158, 1.0714045e-22, 7.576044e-23, 7.576044e-23], [6379.1424, 1.0738241e-22, 7.5931525e-23, 7.5931525e-23], [6393.6016, 1.0762491e-22, 7.6103001e-23, 7.6103001e-23], [6408.0936, 1.0786797e-22, 7.6274866e-23, 7.6274866e-23], [6422.6185, 1.0811158e-22, 7.6447122e-23, 7.6447122e-23], [6437.1763, 1.0835575e-22, 7.661977e-23, 7.661977e-23], [6451.7671, 1.0860047e-22, 7.6792811e-23, 7.6792811e-23], [6466.3909, 1.0884575e-22, 7.6966246e-23, 7.6966246e-23], [6481.0479, 1.0909158e-22, 7.7140075e-23, 7.7140075e-23], [6495.7382, 1.0933798e-22, 7.7314299e-23, 7.7314299e-23], [6510.4617, 1.0958493e-22, 7.748892e-23, 7.748892e-23], [6525.2186, 1.0983245e-22, 7.7663938e-23, 7.7663938e-23], [6540.0089, 1.1008053e-22, 7.7839354e-23, 7.7839354e-23], [6554.8328, 1.1032918e-22, 7.8015169e-23, 7.8015169e-23], [6569.6903, 1.1057839e-22, 7.8191385e-23, 7.8191385e-23], [6584.5814, 1.1082816e-22, 7.8368e-23, 7.8368e-23], [6599.5063, 1.1107851e-22, 7.8545018e-23, 7.8545018e-23], [6614.465, 1.1132942e-22, 7.8722438e-23, 7.8722438e-23], [6629.4577, 1.1158091e-22, 7.8900262e-23, 7.8900262e-23], [6644.4843, 1.1183296e-22, 7.9078491e-23, 7.9078491e-23], [6659.545, 1.1208559e-22, 7.9257124e-23, 7.9257124e-23], [6674.6398, 1.123388e-22, 7.9436164e-23, 7.9436164e-23], [6689.7688, 1.1259258e-22, 7.9615611e-23, 7.9615611e-23], [6704.9321, 1.1284694e-22, 7.9795467e-23, 7.9795467e-23], [6720.1298, 1.1310187e-22, 7.9975731e-23, 7.9975731e-23], [6735.362, 1.1335739e-22, 8.0156405e-23, 8.0156405e-23], [6750.6286, 1.1361349e-22, 8.033749e-23, 8.033749e-23], [6765.9299, 1.1387017e-22, 8.0518988e-23, 8.0518988e-23], [6781.2659, 1.1412743e-22, 8.0700897e-23, 8.0700897e-23], [6796.6366, 1.1438528e-22, 8.0883221e-23, 8.0883221e-23], [6812.0421, 1.1464371e-22, 8.1065959e-23, 8.1065959e-23], [6827.4826, 1.1490274e-22, 8.1249112e-23, 8.1249112e-23], [6842.9581, 1.1516235e-22, 8.1432683e-23, 8.1432683e-23], [6858.4686, 1.1542255e-22, 8.161667e-23, 8.161667e-23], [6874.0143, 1.1568334e-22, 8.1801076e-23, 8.1801076e-23], [6889.5953, 1.1594473e-22, 8.1985901e-23, 8.1985901e-23], [6905.2115, 1.1620671e-22, 8.2171147e-23, 8.2171147e-23], [6920.8632, 1.1646928e-22, 8.2356814e-23, 8.2356814e-23], [6936.5503, 1.1673246e-22, 8.2542903e-23, 8.2542903e-23], [6952.273, 1.1699623e-22, 8.2729415e-23, 8.2729415e-23], [6968.0313, 1.172606e-22, 8.2916351e-23, 8.2916351e-23], [6983.8254, 1.1752557e-22, 8.3103712e-23, 8.3103712e-23], [6999.6552, 1.1779115e-22, 8.3291499e-23, 8.3291499e-23], [7015.5209, 1.1805733e-22, 8.3479713e-23, 8.3479713e-23], [7031.4226, 1.1832411e-22, 8.3668356e-23, 8.3668356e-23], [7047.3604, 1.185915e-22, 8.3857427e-23, 8.3857427e-23], [7063.3342, 1.188595e-22, 8.4046927e-23, 8.4046927e-23], [7079.3443, 1.1912811e-22, 8.4236859e-23, 8.4236859e-23], [7095.3906, 1.1939733e-22, 8.4427223e-23, 8.4427223e-23], [7111.4734, 1.1966716e-22, 8.4618019e-23, 8.4618019e-23], [7127.5925, 1.199376e-22, 8.4809249e-23, 8.4809249e-23], [7143.7482, 1.2020866e-22, 8.5000914e-23, 8.5000914e-23], [7159.9406, 1.2048033e-22, 8.5193014e-23, 8.5193014e-23], [7176.1696, 1.2075263e-22, 8.5385552e-23, 8.5385552e-23], [7192.4354, 1.2102554e-22, 8.5578527e-23, 8.5578527e-23], [7208.7381, 1.2129907e-22, 8.577194e-23, 8.577194e-23], [7225.0778, 1.2157322e-22, 8.5965793e-23, 8.5965793e-23], [7241.4544, 1.21848e-22, 8.6160087e-23, 8.6160087e-23], [7257.8682, 1.221234e-22, 8.6354823e-23, 8.6354823e-23], [7274.3192, 1.2239943e-22, 8.6550001e-23, 8.6550001e-23], [7290.8075, 1.2267608e-22, 8.6745623e-23, 8.6745623e-23], [7307.3332, 1.2295337e-22, 8.694169e-23, 8.694169e-23], [7323.8963, 1.2323128e-22, 8.7138202e-23, 8.7138202e-23], [7340.497, 1.2350983e-22, 8.7335162e-23, 8.7335162e-23], [7357.1353, 1.2378901e-22, 8.7532568e-23, 8.7532568e-23], [7373.8113, 1.2406882e-22, 8.7730424e-23, 8.7730424e-23], [7390.5251, 1.2434927e-22, 8.7928729e-23, 8.7928729e-23], [7407.2768, 1.2463036e-22, 8.8127485e-23, 8.8127485e-23], [7424.0664, 1.2491208e-22, 8.8326692e-23, 8.8326692e-23], [7440.8941, 1.2519445e-22, 8.8526353e-23, 8.8526353e-23], [7457.76, 1.2547745e-22, 8.8726467e-23, 8.8726467e-23], [7474.6641, 1.2576111e-22, 8.8927036e-23, 8.8927036e-23], [7491.6065, 1.260454e-22, 8.9128061e-23, 8.9128061e-23], [7508.5873, 1.2633034e-22, 8.9329543e-23, 8.9329543e-23], [7525.6066, 1.2661593e-22, 8.9531482e-23, 8.9531482e-23], [7542.6644, 1.2690217e-22, 8.9733881e-23, 8.9733881e-23], [7559.761, 1.2718906e-22, 8.993674e-23, 8.993674e-23], [7576.8963, 1.274766e-22, 9.0140059e-23, 9.0140059e-23], [7594.0704, 1.2776479e-22, 9.0343841e-23, 9.0343841e-23], [7611.2834, 1.2805364e-22, 9.0548086e-23, 9.0548086e-23], [7628.5355, 1.2834315e-22, 9.0752795e-23, 9.0752795e-23], [7645.8267, 1.2863331e-22, 9.0957969e-23, 9.0957969e-23], [7663.157, 1.2892414e-22, 9.116361e-23, 9.116361e-23], [7680.5267, 1.2921562e-22, 9.1369717e-23, 9.1369717e-23], [7697.9357, 1.2950776e-22, 9.1576294e-23, 9.1576294e-23], [7715.3842, 1.2980057e-22, 9.1783339e-23, 9.1783339e-23], [7732.8722, 1.3009405e-22, 9.1990855e-23, 9.1990855e-23], [7750.3999, 1.3038819e-22, 9.2198843e-23, 9.2198843e-23], [7767.9673, 1.30683e-22, 9.2407303e-23, 9.2407303e-23], [7785.5745, 1.3097848e-22, 9.2616237e-23, 9.2616237e-23], [7803.2216, 1.3127464e-22, 9.2825646e-23, 9.2825646e-23], [7820.9087, 1.3157146e-22, 9.3035531e-23, 9.3035531e-23], [7838.6359, 1.3186896e-22, 9.3245892e-23, 9.3245892e-23], [7856.4033, 1.3216713e-22, 9.3456732e-23, 9.3456732e-23], [7874.211, 1.3246599e-22, 9.366805e-23, 9.366805e-23], [7892.059, 1.3276552e-22, 9.3879849e-23, 9.3879849e-23], [7909.9475, 1.3306573e-22, 9.4092129e-23, 9.4092129e-23], [7927.8766, 1.3336663e-22, 9.4304891e-23, 9.4304891e-23], [7945.8462, 1.336682e-22, 9.4518137e-23, 9.4518137e-23], [7963.8566, 1.3397047e-22, 9.4731867e-23, 9.4731867e-23], [7981.9079, 1.3427342e-22, 9.4946083e-23, 9.4946083e-23], [8000.0, 1.3457706e-22, 9.5160785e-23, 9.5160785e-23]])


    f1 = data[:,0]
    S1 = data[:,col]**2
    fstop = [f1[0],f1[-1]]

    #% ---- Interpolate to desired frequencies.  Set noise to infinity outside
    #%      band of frequencies covered by f1.
    ##S = inf(size(f)); 
    ##k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop


# % -------------------------------------------------------------------------
# %                              aLIGO spectra
# % -------------------------------------------------------------------------

def aligospectrum(IFO,f):
# % ALIGOSPECTRUM - Design noise curve for different configurations of aLIGO.
# %
# % Arguments are the same as for the parent function.
# %
# % The aLIGO spectra are taken from LIGO Document T0900288-v3, 
# % https://dcc.ligo.org/cgi-bin/private/DocDB/ShowDocument?docid=2974

    source = 'T0900288-v3'
# % ---- Recognize detector and compute noise spectrum.
    if IFO.upper()=='ALIGO-NOSRM':
        col = 1
    elif IFO.upper()=='ALIGO-ZDLP':
        col = 3
    elif IFO.upper()=='ALIGO-ZDHP':
        #% col = 4;
        #% ZDHP baseline has been updated in
        #% https://dcc.ligo.org/LIGO-T1800044 - v4 with slightly lower
        #% range.
        source = 'T1800044-v4'
        col = 1       
    elif IFO.upper()=='ALIGO-BNS':
        col = 5
    elif IFO.upper()=='ALIGO-BBH':
        col = 6
    elif IFO.upper()=='ALIGO-HF':
        col = 7
    else:
        raise ValueError('Detector '+IFO+' not recognized.')


# % ---- Extract pre-sampled noise spectrum from data.  First column is
# %      frequency, other columns map to configurations as above.
    if source=='T0900288-v3':
        data = np.array([[9.0, 1.74133e-21, 1.73571e-21, 1.73707e-21, 1.7391e-21, 1.73646e-21, 1.74313e-21], [9.02047, 2.17689e-21, 2.17244e-21, 2.17352e-21, 2.17512e-21, 2.17303e-21, 2.17831e-21], [9.04099, 2.92623e-21, 2.92296e-21, 2.92375e-21, 2.92493e-21, 2.92339e-21, 2.92728e-21], [9.06155, 4.51505e-21, 4.51295e-21, 4.51345e-21, 4.51422e-21, 4.51322e-21, 4.51572e-21], [9.08216, 1.01283e-20, 1.01274e-20, 1.01276e-20, 1.0128e-20, 1.01275e-20, 1.01286e-20], [9.10282, 3.79178e-20, 3.79175e-20, 3.79176e-20, 3.79177e-20, 3.79176e-20, 3.79178e-20], [9.12353, 6.50675e-21, 6.50533e-21, 6.50567e-21, 6.50619e-21, 6.50552e-21, 6.5072e-21], [9.14428, 3.53294e-21, 3.53035e-21, 3.53097e-21, 3.53191e-21, 3.53069e-21, 3.53375e-21], [9.16508, 2.41373e-21, 2.40997e-21, 2.41088e-21, 2.41224e-21, 2.41047e-21, 2.41491e-21], [9.18593, 1.82732e-21, 1.8224e-21, 1.82359e-21, 1.82537e-21, 1.82305e-21, 1.82886e-21], [9.20682, 1.46687e-21, 1.4608e-21, 1.46227e-21, 1.46447e-21, 1.4616e-21, 1.46876e-21], [9.22776, 1.2232e-21, 1.21597e-21, 1.21772e-21, 1.22034e-21, 1.21692e-21, 1.22544e-21], [9.24875, 1.0477e-21, 1.03933e-21, 1.04136e-21, 1.04439e-21, 1.04043e-21, 1.05028e-21], [9.26979, 9.15454e-22, 9.05952e-22, 9.08259e-22, 9.11697e-22, 9.072e-22, 9.18367e-22], [9.29087, 8.12362e-22, 8.01738e-22, 8.0432e-22, 8.08165e-22, 8.03134e-22, 8.15602e-22], [9.312, 7.2985e-22, 7.18117e-22, 7.20972e-22, 7.25218e-22, 7.19658e-22, 7.33409e-22], [9.33319, 6.62403e-22, 6.49573e-22, 6.527e-22, 6.57344e-22, 6.51259e-22, 6.66273e-22], [9.35441, 6.06313e-22, 5.924e-22, 5.95796e-22, 6.00833e-22, 5.9423e-22, 6.10485e-22], [9.37569, 5.58993e-22, 5.44013e-22, 5.47675e-22, 5.53101e-22, 5.45984e-22, 5.63459e-22], [9.39702, 5.18587e-22, 5.02555e-22, 5.06481e-22, 5.12289e-22, 5.04666e-22, 5.23337e-22], [9.41839, 4.83723e-22, 4.66657e-22, 4.70844e-22, 4.77028e-22, 4.68907e-22, 4.88748e-22], [9.43981, 4.53371e-22, 4.35288e-22, 4.39732e-22, 4.46287e-22, 4.37674e-22, 4.5866e-22], [9.46129, 4.26738e-22, 4.07657e-22, 4.12356e-22, 4.19275e-22, 4.10177e-22, 4.32282e-22], [9.48281, 4.03206e-22, 3.83146e-22, 3.88096e-22, 3.95372e-22, 3.85799e-22, 4.08994e-22], [9.50438, 3.82285e-22, 3.61266e-22, 3.66464e-22, 3.74091e-22, 3.64049e-22, 3.88307e-22], [9.52599, 3.63582e-22, 3.41625e-22, 3.47068e-22, 3.55037e-22, 3.44537e-22, 3.69828e-22], [9.54766, 3.46779e-22, 3.23905e-22, 3.29589e-22, 3.37894e-22, 3.26943e-22, 3.53238e-22], [9.56938, 3.31614e-22, 3.07845e-22, 3.13765e-22, 3.22399e-22, 3.11007e-22, 3.38277e-22], [9.59114, 3.17871e-22, 2.93229e-22, 2.99382e-22, 3.08336e-22, 2.96512e-22, 3.24727e-22], [9.61296, 3.0537e-22, 2.79876e-22, 2.86259e-22, 2.95525e-22, 2.83279e-22, 3.12408e-22], [9.63483, 2.93958e-22, 2.67636e-22, 2.74243e-22, 2.83813e-22, 2.71156e-22, 3.01169e-22], [9.65674, 2.83506e-22, 2.56379e-22, 2.63206e-22, 2.73073e-22, 2.60014e-22, 2.9088e-22], [9.67871, 2.73905e-22, 2.45996e-22, 2.5304e-22, 2.63195e-22, 2.49743e-22, 2.81432e-22], [9.70072, 2.65061e-22, 2.36393e-22, 2.43648e-22, 2.54083e-22, 2.40249e-22, 2.72732e-22], [9.72279, 2.56893e-22, 2.27488e-22, 2.34951e-22, 2.45657e-22, 2.31451e-22, 2.64698e-22], [9.7449, 2.49329e-22, 2.19212e-22, 2.26877e-22, 2.37846e-22, 2.23279e-22, 2.57259e-22], [9.76707, 2.42309e-22, 2.11502e-22, 2.19365e-22, 2.3059e-22, 2.15671e-22, 2.50356e-22], [9.78928, 2.35779e-22, 2.04305e-22, 2.12362e-22, 2.23833e-22, 2.08574e-22, 2.43934e-22], [9.81155, 2.29692e-22, 1.97574e-22, 2.05819e-22, 2.17529e-22, 2.01939e-22, 2.37946e-22], [9.83387, 2.24007e-22, 1.91266e-22, 1.99696e-22, 2.11636e-22, 1.95726e-22, 2.32352e-22], [9.85623, 2.18686e-22, 1.85345e-22, 1.93955e-22, 2.06117e-22, 1.89897e-22, 2.27114e-22], [9.87865, 2.13696e-22, 1.79779e-22, 1.88563e-22, 2.00939e-22, 1.84419e-22, 2.22201e-22], [9.90112, 2.0901e-22, 1.74536e-22, 1.83491e-22, 1.96073e-22, 1.79263e-22, 2.17583e-22], [9.92364, 2.046e-22, 1.69592e-22, 1.78712e-22, 1.91493e-22, 1.74402e-22, 2.13235e-22], [9.94621, 2.00444e-22, 1.64922e-22, 1.74204e-22, 1.87175e-22, 1.69814e-22, 2.09134e-22], [9.96884, 1.96521e-22, 1.60506e-22, 1.69944e-22, 1.83097e-22, 1.65476e-22, 2.0526e-22], [9.99151, 1.92812e-22, 1.56324e-22, 1.65913e-22, 1.79243e-22, 1.6137e-22, 2.01593e-22], [10.0142, 1.89299e-22, 1.52358e-22, 1.62095e-22, 1.75593e-22, 1.57478e-22, 1.98118e-22], [10.037, 1.85969e-22, 1.48594e-22, 1.58473e-22, 1.72132e-22, 1.53784e-22, 1.94818e-22], [10.0598, 1.82806e-22, 1.45016e-22, 1.55034e-22, 1.68846e-22, 1.50275e-22, 1.91681e-22], [10.0827, 1.79799e-22, 1.41612e-22, 1.51764e-22, 1.65724e-22, 1.46937e-22, 1.88694e-22], [10.1057, 1.76935e-22, 1.38369e-22, 1.48651e-22, 1.62751e-22, 1.43759e-22, 1.85846e-22], [10.1286, 1.74205e-22, 1.35278e-22, 1.45685e-22, 1.59919e-22, 1.40729e-22, 1.83127e-22], [10.1517, 1.71599e-22, 1.32328e-22, 1.42856e-22, 1.57218e-22, 1.37837e-22, 1.80527e-22], [10.1748, 1.69108e-22, 1.29509e-22, 1.40154e-22, 1.54637e-22, 1.35075e-22, 1.78039e-22], [10.1979, 1.66724e-22, 1.26814e-22, 1.37572e-22, 1.5217e-22, 1.32434e-22, 1.75653e-22], [10.2211, 1.6444e-22, 1.24234e-22, 1.35101e-22, 1.49809e-22, 1.29907e-22, 1.73364e-22], [10.2444, 1.6225e-22, 1.21763e-22, 1.32735e-22, 1.47547e-22, 1.27486e-22, 1.71165e-22], [10.2677, 1.60147e-22, 1.19394e-22, 1.30467e-22, 1.45377e-22, 1.25165e-22, 1.69049e-22], [10.291, 1.58125e-22, 1.17121e-22, 1.28291e-22, 1.43294e-22, 1.22937e-22, 1.67011e-22], [10.3144, 1.5618e-22, 1.14937e-22, 1.26201e-22, 1.41292e-22, 1.20797e-22, 1.65047e-22], [10.3379, 1.54306e-22, 1.12839e-22, 1.24193e-22, 1.39366e-22, 1.18741e-22, 1.63151e-22], [10.3614, 1.525e-22, 1.10821e-22, 1.22261e-22, 1.37512e-22, 1.16762e-22, 1.6132e-22], [10.385, 1.50756e-22, 1.08878e-22, 1.20401e-22, 1.35725e-22, 1.14857e-22, 1.59549e-22], [10.4086, 1.49072e-22, 1.07007e-22, 1.18609e-22, 1.34002e-22, 1.13021e-22, 1.57834e-22], [10.4323, 1.47443e-22, 1.05203e-22, 1.16882e-22, 1.32338e-22, 1.11251e-22, 1.56173e-22], [10.456, 1.45867e-22, 1.03462e-22, 1.15215e-22, 1.30731e-22, 1.09543e-22, 1.54563e-22], [10.4798, 1.44341e-22, 1.01783e-22, 1.13605e-22, 1.29177e-22, 1.07894e-22, 1.52999e-22], [10.5036, 1.42861e-22, 1.0016e-22, 1.12049e-22, 1.27673e-22, 1.063e-22, 1.51481e-22], [10.5275, 1.41425e-22, 9.85919e-23, 1.10544e-22, 1.26216e-22, 1.04759e-22, 1.50004e-22], [10.5515, 1.40031e-22, 9.7075e-23, 1.09088e-22, 1.24805e-22, 1.03267e-22, 1.48567e-22], [10.5755, 1.38676e-22, 9.56071e-23, 1.07679e-22, 1.23436e-22, 1.01823e-22, 1.47168e-22], [10.5995, 1.37359e-22, 9.41855e-23, 1.06312e-22, 1.22107e-22, 1.00424e-22, 1.45805e-22], [10.6236, 1.36077e-22, 9.28082e-23, 1.04988e-22, 1.20816e-22, 9.90673e-23, 1.44476e-22], [10.6478, 1.34829e-22, 9.14728e-23, 1.03702e-22, 1.19561e-22, 9.77512e-23, 1.43178e-22], [10.672, 1.33612e-22, 9.01774e-23, 1.02454e-22, 1.18341e-22, 9.64736e-23, 1.41911e-22], [10.6963, 1.32426e-22, 8.89201e-23, 1.01242e-22, 1.17154e-22, 9.52328e-23, 1.40673e-22], [10.7206, 1.31268e-22, 8.76992e-23, 1.00064e-22, 1.15997e-22, 9.4027e-23, 1.39463e-22], [10.745, 1.30137e-22, 8.65129e-23, 9.89177e-23, 1.1487e-22, 9.28545e-23, 1.38279e-22], [10.7694, 1.29033e-22, 8.53596e-23, 9.78025e-23, 1.13771e-22, 9.17138e-23, 1.37119e-22], [10.7939, 1.27954e-22, 8.4238e-23, 9.67166e-23, 1.127e-22, 9.06034e-23, 1.35984e-22], [10.8185, 1.26898e-22, 8.31465e-23, 9.56589e-23, 1.11653e-22, 8.95221e-23, 1.34871e-22], [10.8431, 1.25864e-22, 8.20839e-23, 9.46279e-23, 1.10631e-22, 8.84684e-23, 1.3378e-22], [10.8677, 1.24852e-22, 8.10489e-23, 9.36225e-23, 1.09633e-22, 8.74412e-23, 1.32709e-22], [10.8925, 1.23861e-22, 8.00403e-23, 9.26416e-23, 1.08657e-22, 8.64393e-23, 1.31659e-22], [10.9172, 1.2289e-22, 7.9057e-23, 9.16842e-23, 1.07702e-22, 8.54617e-23, 1.30627e-22], [10.9421, 1.21937e-22, 7.80981e-23, 9.07492e-23, 1.06767e-22, 8.45074e-23, 1.29613e-22], [10.967, 1.21003e-22, 7.71624e-23, 8.98357e-23, 1.05852e-22, 8.35753e-23, 1.28617e-22], [10.9919, 1.20085e-22, 7.6249e-23, 8.89428e-23, 1.04956e-22, 8.26646e-23, 1.27638e-22], [11.0169, 1.19185e-22, 7.53571e-23, 8.80696e-23, 1.04078e-22, 8.17743e-23, 1.26674e-22], [11.042, 1.183e-22, 7.44858e-23, 8.72155e-23, 1.03217e-22, 8.09038e-23, 1.25726e-22], [11.0671, 1.1743e-22, 7.36342e-23, 8.63795e-23, 1.02373e-22, 8.00521e-23, 1.24793e-22], [11.0923, 1.16576e-22, 7.28018e-23, 8.55611e-23, 1.01544e-22, 7.92187e-23, 1.23874e-22], [11.1175, 1.15735e-22, 7.19876e-23, 8.47596e-23, 1.00731e-22, 7.84027e-23, 1.22969e-22], [11.1428, 1.14908e-22, 7.11911e-23, 8.39742e-23, 9.99324e-23, 7.76036e-23, 1.22077e-22], [11.1681, 1.14094e-22, 7.04116e-23, 8.32044e-23, 9.9148e-23, 7.68206e-23, 1.21197e-22], [11.1935, 1.13293e-22, 6.96484e-23, 8.24496e-23, 9.83773e-23, 7.60533e-23, 1.2033e-22], [11.219, 1.12504e-22, 6.8901e-23, 8.17093e-23, 9.76197e-23, 7.5301e-23, 1.19475e-22], [11.2445, 1.11727e-22, 6.81689e-23, 8.0983e-23, 9.68749e-23, 7.45632e-23, 1.18632e-22], [11.2701, 1.10961e-22, 6.74515e-23, 8.02701e-23, 9.61422e-23, 7.38394e-23, 1.17799e-22], [11.2957, 1.10206e-22, 6.67483e-23, 7.95703e-23, 9.54214e-23, 7.31291e-23, 1.16977e-22], [11.3214, 1.09462e-22, 6.60587e-23, 7.88829e-23, 9.47119e-23, 7.24319e-23, 1.16166e-22], [11.3472, 1.08728e-22, 6.53825e-23, 7.82077e-23, 9.40135e-23, 7.17473e-23, 1.15365e-22], [11.373, 1.08004e-22, 6.47191e-23, 7.75443e-23, 9.33258e-23, 7.1075e-23, 1.14573e-22], [11.3988, 1.0729e-22, 6.40681e-23, 7.68922e-23, 9.26484e-23, 7.04144e-23, 1.13791e-22], [11.4248, 1.06585e-22, 6.34291e-23, 7.6251e-23, 9.1981e-23, 6.97653e-23, 1.13018e-22], [11.4507, 1.05889e-22, 6.28017e-23, 7.56205e-23, 9.13233e-23, 6.91272e-23, 1.12254e-22], [11.4768, 1.05202e-22, 6.21857e-23, 7.50003e-23, 9.0675e-23, 6.84998e-23, 1.11499e-22], [11.5029, 1.04523e-22, 6.15805e-23, 7.43901e-23, 9.00358e-23, 6.78829e-23, 1.10752e-22], [11.5291, 1.03853e-22, 6.0986e-23, 7.37895e-23, 8.94054e-23, 6.7276e-23, 1.10014e-22], [11.5553, 1.03191e-22, 6.04018e-23, 7.31983e-23, 8.87836e-23, 6.6679e-23, 1.09283e-22], [11.5816, 1.02536e-22, 5.98275e-23, 7.26163e-23, 8.81701e-23, 6.60914e-23, 1.0856e-22], [11.6079, 1.0189e-22, 5.9263e-23, 7.20431e-23, 8.75648e-23, 6.5513e-23, 1.07845e-22], [11.6343, 1.0125e-22, 5.87079e-23, 7.14785e-23, 8.69673e-23, 6.49436e-23, 1.07137e-22], [11.6608, 1.00618e-22, 5.8162e-23, 7.09223e-23, 8.63774e-23, 6.43829e-23, 1.06436e-22], [11.6873, 9.99932e-23, 5.76249e-23, 7.03742e-23, 8.5795e-23, 6.38307e-23, 1.05742e-22], [11.7139, 9.93749e-23, 5.70966e-23, 6.98339e-23, 8.52199e-23, 6.32867e-23, 1.05056e-22], [11.7405, 9.87634e-23, 5.65767e-23, 6.93014e-23, 8.46519e-23, 6.27508e-23, 1.04375e-22], [11.7672, 9.81584e-23, 5.60649e-23, 6.87764e-23, 8.40907e-23, 6.22226e-23, 1.03702e-22], [11.794, 9.75598e-23, 5.55612e-23, 6.82587e-23, 8.35363e-23, 6.17021e-23, 1.03035e-22], [11.8208, 9.69674e-23, 5.50653e-23, 6.77481e-23, 8.29884e-23, 6.11889e-23, 1.02374e-22], [11.8477, 9.63811e-23, 5.4577e-23, 6.72444e-23, 8.24468e-23, 6.0683e-23, 1.01719e-22], [11.8747, 9.58008e-23, 5.40961e-23, 6.67475e-23, 8.19116e-23, 6.01842e-23, 1.0107e-22], [11.9017, 9.52262e-23, 5.36225e-23, 6.62571e-23, 8.13824e-23, 5.96922e-23, 1.00427e-22], [11.9287, 9.46574e-23, 5.31559e-23, 6.57732e-23, 8.08592e-23, 5.92069e-23, 9.97901e-23], [11.9559, 9.40941e-23, 5.26961e-23, 6.52956e-23, 8.03417e-23, 5.87281e-23, 9.91586e-23], [11.9831, 9.35363e-23, 5.22431e-23, 6.48241e-23, 7.983e-23, 5.82558e-23, 9.85326e-23], [12.0103, 9.29838e-23, 5.17967e-23, 6.43586e-23, 7.93238e-23, 5.77897e-23, 9.7912e-23], [12.0376, 9.24365e-23, 5.13566e-23, 6.38989e-23, 7.88231e-23, 5.73296e-23, 9.72967e-23], [12.065, 9.18943e-23, 5.09229e-23, 6.3445e-23, 7.83277e-23, 5.68756e-23, 9.66867e-23], [12.0925, 9.13572e-23, 5.04952e-23, 6.29966e-23, 7.78374e-23, 5.64274e-23, 9.60818e-23], [12.12, 9.08249e-23, 5.00736e-23, 6.25537e-23, 7.73523e-23, 5.59849e-23, 9.54818e-23], [12.1475, 9.02975e-23, 4.96578e-23, 6.21162e-23, 7.68722e-23, 5.55479e-23, 9.48869e-23], [12.1752, 8.97749e-23, 4.92477e-23, 6.16839e-23, 7.6397e-23, 5.51164e-23, 9.42968e-23], [12.2029, 8.92568e-23, 4.88433e-23, 6.12567e-23, 7.59265e-23, 5.46903e-23, 9.37114e-23], [12.2306, 8.87434e-23, 4.84444e-23, 6.08346e-23, 7.54608e-23, 5.42694e-23, 9.31307e-23], [12.2584, 8.82344e-23, 4.80508e-23, 6.04173e-23, 7.49997e-23, 5.38536e-23, 9.25547e-23], [12.2863, 8.77298e-23, 4.76625e-23, 6.00049e-23, 7.45431e-23, 5.34429e-23, 9.19832e-23], [12.3143, 8.72296e-23, 4.72793e-23, 5.95972e-23, 7.40909e-23, 5.3037e-23, 9.14161e-23], [12.3423, 8.67336e-23, 4.69012e-23, 5.91941e-23, 7.36431e-23, 5.2636e-23, 9.08535e-23], [12.3704, 8.62417e-23, 4.65281e-23, 5.87956e-23, 7.31996e-23, 5.22397e-23, 9.02951e-23], [12.3985, 8.5754e-23, 4.61598e-23, 5.84015e-23, 7.27602e-23, 5.1848e-23, 8.9741e-23], [12.4267, 8.52703e-23, 4.57963e-23, 5.80117e-23, 7.2325e-23, 5.14609e-23, 8.91911e-23], [12.455, 8.47906e-23, 4.54374e-23, 5.76263e-23, 7.18938e-23, 5.10782e-23, 8.86454e-23], [12.4833, 8.43147e-23, 4.50832e-23, 5.7245e-23, 7.14666e-23, 5.06999e-23, 8.81037e-23], [12.5117, 8.38428e-23, 4.47334e-23, 5.68678e-23, 7.10433e-23, 5.03259e-23, 8.7566e-23], [12.5401, 8.33746e-23, 4.4388e-23, 5.64947e-23, 7.06239e-23, 4.99562e-23, 8.70322e-23], [12.5687, 8.29101e-23, 4.4047e-23, 5.61256e-23, 7.02082e-23, 4.95905e-23, 8.65024e-23], [12.5973, 8.24493e-23, 4.37102e-23, 5.57603e-23, 6.97962e-23, 4.92289e-23, 8.59764e-23], [12.6259, 8.19921e-23, 4.33776e-23, 5.53989e-23, 6.93879e-23, 4.88713e-23, 8.54542e-23], [12.6546, 8.15385e-23, 4.3049e-23, 5.50413e-23, 6.89832e-23, 4.85176e-23, 8.49357e-23], [12.6834, 8.10884e-23, 4.27245e-23, 5.46873e-23, 6.8582e-23, 4.81678e-23, 8.44209e-23], [12.7123, 8.06417e-23, 4.24039e-23, 5.4337e-23, 6.81843e-23, 4.78217e-23, 8.39097e-23], [12.7412, 8.01984e-23, 4.20873e-23, 5.39902e-23, 6.779e-23, 4.74794e-23, 8.34022e-23], [12.7702, 7.97585e-23, 4.17744e-23, 5.36469e-23, 6.73991e-23, 4.71407e-23, 8.28981e-23], [12.7992, 7.93219e-23, 4.14652e-23, 5.33071e-23, 6.70115e-23, 4.68055e-23, 8.23976e-23], [12.8283, 7.88886e-23, 4.11598e-23, 5.29707e-23, 6.66271e-23, 4.6474e-23, 8.19005e-23], [12.8575, 7.84585e-23, 4.08579e-23, 5.26376e-23, 6.6246e-23, 4.61459e-23, 8.14068e-23], [12.8867, 7.80315e-23, 4.05597e-23, 5.23078e-23, 6.5868e-23, 4.58212e-23, 8.09165e-23], [12.916, 7.76077e-23, 4.02649e-23, 5.19812e-23, 6.54932e-23, 4.54998e-23, 8.04295e-23], [12.9454, 7.7187e-23, 3.99735e-23, 5.16577e-23, 6.51214e-23, 4.51818e-23, 7.99458e-23], [12.9749, 7.67693e-23, 3.96855e-23, 5.13374e-23, 6.47527e-23, 4.4867e-23, 7.94654e-23], [13.0044, 7.63546e-23, 3.94009e-23, 5.10202e-23, 6.43869e-23, 4.45554e-23, 7.89881e-23], [13.034, 7.59429e-23, 3.91195e-23, 5.0706e-23, 6.40241e-23, 4.4247e-23, 7.8514e-23], [13.0636, 7.55341e-23, 3.88413e-23, 5.03948e-23, 6.36642e-23, 4.39416e-23, 7.80431e-23], [13.0933, 7.51283e-23, 3.85663e-23, 5.00865e-23, 6.33072e-23, 4.36393e-23, 7.75752e-23], [13.1231, 7.47252e-23, 3.82944e-23, 4.9781e-23, 6.2953e-23, 4.334e-23, 7.71105e-23], [13.153, 7.4325e-23, 3.80255e-23, 4.94785e-23, 6.26015e-23, 4.30436e-23, 7.66487e-23], [13.1829, 7.39276e-23, 3.77597e-23, 4.91787e-23, 6.22528e-23, 4.27502e-23, 7.61899e-23], [13.2129, 7.35329e-23, 3.74968e-23, 4.88816e-23, 6.19068e-23, 4.24596e-23, 7.57341e-23], [13.2429, 7.3141e-23, 3.72369e-23, 4.85873e-23, 6.15635e-23, 4.21719e-23, 7.52812e-23], [13.273, 7.27517e-23, 3.69798e-23, 4.82956e-23, 6.12228e-23, 4.18869e-23, 7.48312e-23], [13.3032, 7.23651e-23, 3.67255e-23, 4.80066e-23, 6.08847e-23, 4.16046e-23, 7.43841e-23], [13.3335, 7.19811e-23, 3.6474e-23, 4.77202e-23, 6.05491e-23, 4.1325e-23, 7.39398e-23], [13.3638, 7.15997e-23, 3.62253e-23, 4.74363e-23, 6.02161e-23, 4.10481e-23, 7.34983e-23], [13.3942, 7.12209e-23, 3.59793e-23, 4.7155e-23, 5.98856e-23, 4.07739e-23, 7.30596e-23], [13.4247, 7.08446e-23, 3.57359e-23, 4.68761e-23, 5.95576e-23, 4.05021e-23, 7.26237e-23], [13.4552, 7.04708e-23, 3.54951e-23, 4.65997e-23, 5.9232e-23, 4.0233e-23, 7.21904e-23], [13.4858, 7.00994e-23, 3.52569e-23, 4.63256e-23, 5.89088e-23, 3.99663e-23, 7.17599e-23], [13.5165, 6.97306e-23, 3.50212e-23, 4.6054e-23, 5.85879e-23, 3.97021e-23, 7.1332e-23], [13.5472, 6.93641e-23, 3.47881e-23, 4.57847e-23, 5.82694e-23, 3.94403e-23, 7.09067e-23], [13.578, 6.9e-23, 3.45574e-23, 4.55177e-23, 5.79533e-23, 3.91809e-23, 7.0484e-23], [13.6089, 6.86383e-23, 3.43291e-23, 4.5253e-23, 5.76394e-23, 3.89239e-23, 7.0064e-23], [13.6399, 6.82789e-23, 3.41032e-23, 4.49905e-23, 5.73277e-23, 3.86693e-23, 6.96465e-23], [13.6709, 6.79219e-23, 3.38797e-23, 4.47303e-23, 5.70183e-23, 3.84169e-23, 6.92315e-23], [13.702, 6.75671e-23, 3.36585e-23, 4.44722e-23, 5.67111e-23, 3.81668e-23, 6.8819e-23], [13.7332, 6.72146e-23, 3.34395e-23, 4.42163e-23, 5.64061e-23, 3.79189e-23, 6.8409e-23], [13.7644, 6.68644e-23, 3.32229e-23, 4.39625e-23, 5.61032e-23, 3.76732e-23, 6.80015e-23], [13.7957, 6.65163e-23, 3.30084e-23, 4.37108e-23, 5.58025e-23, 3.74297e-23, 6.75964e-23], [13.8271, 6.61705e-23, 3.27961e-23, 4.34612e-23, 5.55038e-23, 3.71883e-23, 6.71938e-23], [13.8585, 6.58268e-23, 3.2586e-23, 4.32136e-23, 5.52072e-23, 3.69491e-23, 6.67935e-23], [13.8901, 6.54853e-23, 3.2378e-23, 4.2968e-23, 5.49127e-23, 3.67119e-23, 6.63956e-23], [13.9217, 6.51459e-23, 3.21721e-23, 4.27245e-23, 5.46202e-23, 3.64768e-23, 6.6e-23], [13.9533, 6.48086e-23, 3.19683e-23, 4.24829e-23, 5.43297e-23, 3.62438e-23, 6.56068e-23], [13.9851, 6.44734e-23, 3.17665e-23, 4.22432e-23, 5.40412e-23, 3.60127e-23, 6.52159e-23], [14.0169, 6.41403e-23, 3.15667e-23, 4.20054e-23, 5.37547e-23, 3.57836e-23, 6.48273e-23], [14.0488, 6.38092e-23, 3.13689e-23, 4.17696e-23, 5.34701e-23, 3.55565e-23, 6.44409e-23], [14.0807, 6.34801e-23, 3.11731e-23, 4.15355e-23, 5.31874e-23, 3.53313e-23, 6.40568e-23], [14.1127, 6.31531e-23, 3.09791e-23, 4.13034e-23, 5.29066e-23, 3.5108e-23, 6.36749e-23], [14.1448, 6.2828e-23, 3.07871e-23, 4.1073e-23, 5.26277e-23, 3.48866e-23, 6.32952e-23], [14.177, 6.25049e-23, 3.0597e-23, 4.08445e-23, 5.23506e-23, 3.4667e-23, 6.29177e-23], [14.2093, 6.21837e-23, 3.04086e-23, 4.06177e-23, 5.20754e-23, 3.44493e-23, 6.25424e-23], [14.2416, 6.18645e-23, 3.02222e-23, 4.03927e-23, 5.1802e-23, 3.42334e-23, 6.21692e-23], [14.274, 6.15472e-23, 3.00375e-23, 4.01694e-23, 5.15304e-23, 3.40192e-23, 6.17982e-23], [14.3064, 6.12317e-23, 2.98546e-23, 3.99478e-23, 5.12606e-23, 3.38069e-23, 6.14292e-23], [14.339, 6.09182e-23, 2.96734e-23, 3.97278e-23, 5.09925e-23, 3.35962e-23, 6.10624e-23], [14.3716, 6.06065e-23, 2.94939e-23, 3.95096e-23, 5.07262e-23, 3.33873e-23, 6.06977e-23], [14.4043, 6.02966e-23, 2.93162e-23, 3.9293e-23, 5.04616e-23, 3.31801e-23, 6.0335e-23], [14.4371, 5.99886e-23, 2.91401e-23, 3.9078e-23, 5.01987e-23, 3.29746e-23, 5.99743e-23], [14.4699, 5.96824e-23, 2.89657e-23, 3.88647e-23, 4.99375e-23, 3.27707e-23, 5.96157e-23], [14.5028, 5.9378e-23, 2.8793e-23, 3.86529e-23, 4.96779e-23, 3.25685e-23, 5.92591e-23], [14.5358, 5.90753e-23, 2.86218e-23, 3.84427e-23, 4.942e-23, 3.23679e-23, 5.89045e-23], [14.5689, 5.87744e-23, 2.84523e-23, 3.8234e-23, 4.91638e-23, 3.21688e-23, 5.85519e-23], [14.602, 5.84753e-23, 2.82843e-23, 3.80269e-23, 4.89091e-23, 3.19714e-23, 5.82013e-23], [14.6352, 5.81779e-23, 2.81179e-23, 3.78213e-23, 4.86561e-23, 3.17755e-23, 5.78526e-23], [14.6685, 5.78822e-23, 2.7953e-23, 3.76172e-23, 4.84047e-23, 3.15812e-23, 5.75058e-23], [14.7019, 5.75882e-23, 2.77896e-23, 3.74146e-23, 4.81548e-23, 3.13884e-23, 5.7161e-23], [14.7353, 5.72959e-23, 2.76277e-23, 3.72134e-23, 4.79065e-23, 3.11971e-23, 5.6818e-23], [14.7688, 5.70052e-23, 2.74673e-23, 3.70137e-23, 4.76598e-23, 3.10073e-23, 5.6477e-23], [14.8024, 5.67163e-23, 2.73084e-23, 3.68155e-23, 4.74145e-23, 3.08189e-23, 5.61378e-23], [14.8361, 5.64289e-23, 2.71508e-23, 3.66186e-23, 4.71708e-23, 3.06321e-23, 5.58005e-23], [14.8698, 5.61432e-23, 2.69948e-23, 3.64231e-23, 4.69286e-23, 3.04466e-23, 5.5465e-23], [14.9036, 5.58591e-23, 2.68401e-23, 3.62291e-23, 4.66879e-23, 3.02626e-23, 5.51314e-23], [14.9375, 5.55766e-23, 2.66868e-23, 3.60363e-23, 4.64486e-23, 3.008e-23, 5.47996e-23], [14.9715, 5.52957e-23, 2.65348e-23, 3.5845e-23, 4.62108e-23, 2.98988e-23, 5.44696e-23], [15.0056, 5.50164e-23, 2.63843e-23, 3.5655e-23, 4.59745e-23, 2.9719e-23, 5.41414e-23], [15.0397, 5.47387e-23, 2.6235e-23, 3.54663e-23, 4.57395e-23, 2.95405e-23, 5.3815e-23], [15.0739, 5.44625e-23, 2.60871e-23, 3.52789e-23, 4.5506e-23, 2.93634e-23, 5.34904e-23], [15.1082, 5.41878e-23, 2.59405e-23, 3.50929e-23, 4.52739e-23, 2.91876e-23, 5.31675e-23], [15.1426, 5.39147e-23, 2.57951e-23, 3.49081e-23, 4.50432e-23, 2.90131e-23, 5.28463e-23], [15.177, 5.3643e-23, 2.56511e-23, 3.47245e-23, 4.48139e-23, 2.88399e-23, 5.25269e-23], [15.2115, 5.33729e-23, 2.55083e-23, 3.45423e-23, 4.4586e-23, 2.8668e-23, 5.22092e-23], [15.2461, 5.31043e-23, 2.53667e-23, 3.43613e-23, 4.43594e-23, 2.84974e-23, 5.18932e-23], [15.2808, 5.28371e-23, 2.52264e-23, 3.41815e-23, 4.41341e-23, 2.83281e-23, 5.15789e-23], [15.3156, 5.25715e-23, 2.50873e-23, 3.40029e-23, 4.39102e-23, 2.816e-23, 5.12663e-23], [15.3504, 5.23072e-23, 2.49493e-23, 3.38256e-23, 4.36876e-23, 2.79931e-23, 5.09553e-23], [15.3853, 5.20445e-23, 2.48126e-23, 3.36494e-23, 4.34663e-23, 2.78275e-23, 5.06461e-23], [15.4203, 5.17831e-23, 2.46771e-23, 3.34744e-23, 4.32463e-23, 2.76631e-23, 5.03384e-23], [15.4554, 5.15232e-23, 2.45426e-23, 3.33006e-23, 4.30276e-23, 2.74999e-23, 5.00324e-23], [15.4905, 5.12647e-23, 2.44094e-23, 3.31279e-23, 4.28102e-23, 2.73378e-23, 4.97281e-23], [15.5258, 5.10076e-23, 2.42773e-23, 3.29564e-23, 4.2594e-23, 2.7177e-23, 4.94253e-23], [15.5611, 5.07519e-23, 2.41463e-23, 3.2786e-23, 4.23791e-23, 2.70173e-23, 4.91241e-23], [15.5965, 5.04976e-23, 2.40164e-23, 3.26168e-23, 4.21655e-23, 2.68587e-23, 4.88246e-23], [15.632, 5.02446e-23, 2.38876e-23, 3.24487e-23, 4.19531e-23, 2.67013e-23, 4.85266e-23], [15.6675, 4.9993e-23, 2.37598e-23, 3.22816e-23, 4.17419e-23, 2.6545e-23, 4.82302e-23], [15.7032, 4.97428e-23, 2.36332e-23, 3.21157e-23, 4.15319e-23, 2.63899e-23, 4.79354e-23], [15.7389, 4.94939e-23, 2.35076e-23, 3.19508e-23, 4.13231e-23, 2.62358e-23, 4.76421e-23], [15.7747, 4.92463e-23, 2.3383e-23, 3.17871e-23, 4.11155e-23, 2.60829e-23, 4.73503e-23], [15.8105, 4.90001e-23, 2.32595e-23, 3.16243e-23, 4.09092e-23, 2.5931e-23, 4.70601e-23], [15.8465, 4.87552e-23, 2.3137e-23, 3.14627e-23, 4.07039e-23, 2.57802e-23, 4.67714e-23], [15.8826, 4.85116e-23, 2.30155e-23, 3.1302e-23, 4.04999e-23, 2.56305e-23, 4.64843e-23], [15.9187, 4.82693e-23, 2.28951e-23, 3.11424e-23, 4.0297e-23, 2.54818e-23, 4.61986e-23], [15.9549, 4.80282e-23, 2.27756e-23, 3.09838e-23, 4.00952e-23, 2.53341e-23, 4.59144e-23], [15.9912, 4.77885e-23, 2.26571e-23, 3.08263e-23, 3.98946e-23, 2.51875e-23, 4.56317e-23], [16.0276, 4.755e-23, 2.25395e-23, 3.06697e-23, 3.96952e-23, 2.5042e-23, 4.53505e-23], [16.064, 4.73128e-23, 2.2423e-23, 3.05142e-23, 3.94968e-23, 2.48974e-23, 4.50707e-23], [16.1005, 4.70768e-23, 2.23073e-23, 3.03596e-23, 3.92996e-23, 2.47539e-23, 4.47924e-23], [16.1372, 4.68421e-23, 2.21927e-23, 3.0206e-23, 3.91034e-23, 2.46113e-23, 4.45156e-23], [16.1739, 4.66086e-23, 2.20789e-23, 3.00533e-23, 3.89084e-23, 2.44697e-23, 4.42402e-23], [16.2107, 4.63763e-23, 2.19661e-23, 2.99017e-23, 3.87145e-23, 2.43291e-23, 4.39662e-23], [16.2475, 4.61452e-23, 2.18541e-23, 2.97509e-23, 3.85216e-23, 2.41895e-23, 4.36936e-23], [16.2845, 4.59154e-23, 2.17431e-23, 2.96011e-23, 3.83298e-23, 2.40509e-23, 4.34225e-23], [16.3215, 4.56868e-23, 2.1633e-23, 2.94523e-23, 3.81391e-23, 2.39132e-23, 4.31528e-23], [16.3587, 4.54593e-23, 2.15237e-23, 2.93044e-23, 3.79494e-23, 2.37764e-23, 4.28844e-23], [16.3959, 4.52331e-23, 2.14154e-23, 2.91574e-23, 3.77608e-23, 2.36406e-23, 4.26175e-23], [16.4332, 4.5008e-23, 2.13079e-23, 2.90113e-23, 3.75732e-23, 2.35057e-23, 4.23519e-23], [16.4705, 4.47841e-23, 2.12012e-23, 2.88661e-23, 3.73866e-23, 2.33717e-23, 4.20877e-23], [16.508, 4.45613e-23, 2.10954e-23, 2.87218e-23, 3.72011e-23, 2.32387e-23, 4.18248e-23], [16.5455, 4.43397e-23, 2.09904e-23, 2.85784e-23, 3.70166e-23, 2.31065e-23, 4.15633e-23], [16.5832, 4.41193e-23, 2.08863e-23, 2.84358e-23, 3.68331e-23, 2.29753e-23, 4.13032e-23], [16.6209, 4.39e-23, 2.0783e-23, 2.82942e-23, 3.66506e-23, 2.28449e-23, 4.10444e-23], [16.6587, 4.36818e-23, 2.06805e-23, 2.81534e-23, 3.64691e-23, 2.27154e-23, 4.07869e-23], [16.6966, 4.34648e-23, 2.05788e-23, 2.80134e-23, 3.62886e-23, 2.25868e-23, 4.05308e-23], [16.7346, 4.32489e-23, 2.04779e-23, 2.78743e-23, 3.6109e-23, 2.24591e-23, 4.0276e-23], [16.7726, 4.3034e-23, 2.03778e-23, 2.77361e-23, 3.59305e-23, 2.23322e-23, 4.00224e-23], [16.8108, 4.28203e-23, 2.02785e-23, 2.75987e-23, 3.57529e-23, 2.22061e-23, 3.97702e-23], [16.849, 4.26077e-23, 2.018e-23, 2.74621e-23, 3.55763e-23, 2.2081e-23, 3.95193e-23], [16.8874, 4.23962e-23, 2.00822e-23, 2.73263e-23, 3.54006e-23, 2.19566e-23, 3.92697e-23], [16.9258, 4.21858e-23, 1.99852e-23, 2.71914e-23, 3.52259e-23, 2.18331e-23, 3.90213e-23], [16.9643, 4.19764e-23, 1.98889e-23, 2.70573e-23, 3.50521e-23, 2.17104e-23, 3.87742e-23], [17.0028, 4.17681e-23, 1.97934e-23, 2.69239e-23, 3.48793e-23, 2.15885e-23, 3.85284e-23], [17.0415, 4.15609e-23, 1.96987e-23, 2.67914e-23, 3.47073e-23, 2.14675e-23, 3.82839e-23], [17.0803, 4.13548e-23, 1.96046e-23, 2.66597e-23, 3.45364e-23, 2.13472e-23, 3.80406e-23], [17.1191, 4.11496e-23, 1.95113e-23, 2.65287e-23, 3.43663e-23, 2.12278e-23, 3.77985e-23], [17.1581, 4.09456e-23, 1.94187e-23, 2.63985e-23, 3.41971e-23, 2.11091e-23, 3.75577e-23], [17.1971, 4.07425e-23, 1.93269e-23, 2.62691e-23, 3.40288e-23, 2.09913e-23, 3.73181e-23], [17.2362, 4.05406e-23, 1.92357e-23, 2.61405e-23, 3.38615e-23, 2.08742e-23, 3.70797e-23], [17.2754, 4.03396e-23, 1.91452e-23, 2.60126e-23, 3.3695e-23, 2.07579e-23, 3.68426e-23], [17.3147, 4.01396e-23, 1.90555e-23, 2.58855e-23, 3.35294e-23, 2.06423e-23, 3.66066e-23], [17.3541, 3.99407e-23, 1.89664e-23, 2.57591e-23, 3.33647e-23, 2.05275e-23, 3.63719e-23], [17.3936, 3.97428e-23, 1.8878e-23, 2.56334e-23, 3.32009e-23, 2.04135e-23, 3.61384e-23], [17.4331, 3.95458e-23, 1.87902e-23, 2.55085e-23, 3.30379e-23, 2.03002e-23, 3.5906e-23], [17.4728, 3.93499e-23, 1.87032e-23, 2.53844e-23, 3.28758e-23, 2.01877e-23, 3.56749e-23], [17.5125, 3.9155e-23, 1.86168e-23, 2.52609e-23, 3.27146e-23, 2.00759e-23, 3.54449e-23], [17.5524, 3.8961e-23, 1.85311e-23, 2.51382e-23, 3.25542e-23, 1.99649e-23, 3.52161e-23], [17.5923, 3.8768e-23, 1.8446e-23, 2.50162e-23, 3.23947e-23, 1.98546e-23, 3.49885e-23], [17.6323, 3.8576e-23, 1.83615e-23, 2.48949e-23, 3.2236e-23, 1.9745e-23, 3.4762e-23], [17.6724, 3.8385e-23, 1.82777e-23, 2.47743e-23, 3.20781e-23, 1.96361e-23, 3.45367e-23], [17.7126, 3.81949e-23, 1.81946e-23, 2.46544e-23, 3.19211e-23, 1.9528e-23, 3.43125e-23], [17.7529, 3.80058e-23, 1.81121e-23, 2.45352e-23, 3.17648e-23, 1.94205e-23, 3.40895e-23], [17.7933, 3.78176e-23, 1.80301e-23, 2.44167e-23, 3.16095e-23, 1.93138e-23, 3.38676e-23], [17.8337, 3.76304e-23, 1.79489e-23, 2.42989e-23, 3.14549e-23, 1.92077e-23, 3.36469e-23], [17.8743, 3.74441e-23, 1.78682e-23, 2.41817e-23, 3.13011e-23, 1.91024e-23, 3.34273e-23], [17.915, 3.72588e-23, 1.77881e-23, 2.40653e-23, 3.11482e-23, 1.89977e-23, 3.32088e-23], [17.9557, 3.70743e-23, 1.77086e-23, 2.39495e-23, 3.0996e-23, 1.88937e-23, 3.29914e-23], [17.9966, 3.68908e-23, 1.76298e-23, 2.38343e-23, 3.08446e-23, 1.87904e-23, 3.27751e-23], [18.0375, 3.67083e-23, 1.75515e-23, 2.37198e-23, 3.0694e-23, 1.86878e-23, 3.25599e-23], [18.0785, 3.65266e-23, 1.74738e-23, 2.3606e-23, 3.05442e-23, 1.85859e-23, 3.23459e-23], [18.1196, 3.63458e-23, 1.73967e-23, 2.34928e-23, 3.03952e-23, 1.84846e-23, 3.21329e-23], [18.1609, 3.6166e-23, 1.73202e-23, 2.33803e-23, 3.0247e-23, 1.8384e-23, 3.1921e-23], [18.2022, 3.5987e-23, 1.72442e-23, 2.32684e-23, 3.00995e-23, 1.8284e-23, 3.17102e-23], [18.2436, 3.5809e-23, 1.71689e-23, 2.31572e-23, 2.99528e-23, 1.81847e-23, 3.15004e-23], [18.2851, 3.56318e-23, 1.7094e-23, 2.30465e-23, 2.98069e-23, 1.8086e-23, 3.12918e-23], [18.3267, 3.54555e-23, 1.70198e-23, 2.29366e-23, 2.96617e-23, 1.7988e-23, 3.10842e-23], [18.3683, 3.52801e-23, 1.69461e-23, 2.28272e-23, 2.95173e-23, 1.78906e-23, 3.08776e-23], [18.4101, 3.51055e-23, 1.68729e-23, 2.27184e-23, 2.93736e-23, 1.77939e-23, 3.06722e-23], [18.452, 3.49318e-23, 1.68003e-23, 2.26103e-23, 2.92306e-23, 1.76977e-23, 3.04677e-23], [18.494, 3.4759e-23, 1.67282e-23, 2.25028e-23, 2.90884e-23, 1.76022e-23, 3.02644e-23], [18.536, 3.45871e-23, 1.66567e-23, 2.23959e-23, 2.89469e-23, 1.75074e-23, 3.0062e-23], [18.5782, 3.4416e-23, 1.65857e-23, 2.22896e-23, 2.88062e-23, 1.74131e-23, 2.98607e-23], [18.6204, 3.42457e-23, 1.65152e-23, 2.21838e-23, 2.86662e-23, 1.73195e-23, 2.96605e-23], [18.6628, 3.40763e-23, 1.64453e-23, 2.20787e-23, 2.85269e-23, 1.72265e-23, 2.94612e-23], [18.7053, 3.39077e-23, 1.63759e-23, 2.19742e-23, 2.83883e-23, 1.71341e-23, 2.9263e-23], [18.7478, 3.374e-23, 1.6307e-23, 2.18703e-23, 2.82504e-23, 1.70423e-23, 2.90658e-23], [18.7904, 3.35731e-23, 1.62386e-23, 2.17669e-23, 2.81133e-23, 1.6951e-23, 2.88696e-23], [18.8332, 3.3407e-23, 1.61707e-23, 2.16641e-23, 2.79768e-23, 1.68604e-23, 2.86745e-23], [18.876, 3.32418e-23, 1.61033e-23, 2.15619e-23, 2.78411e-23, 1.67704e-23, 2.84803e-23], [18.919, 3.30774e-23, 1.60364e-23, 2.14603e-23, 2.7706e-23, 1.6681e-23, 2.82871e-23], [18.962, 3.29137e-23, 1.597e-23, 2.13592e-23, 2.75716e-23, 1.65921e-23, 2.80949e-23], [19.0051, 3.27509e-23, 1.59041e-23, 2.12587e-23, 2.74379e-23, 1.65039e-23, 2.79037e-23], [19.0483, 3.25889e-23, 1.58387e-23, 2.11588e-23, 2.7305e-23, 1.64162e-23, 2.77135e-23], [19.0917, 3.24277e-23, 1.57738e-23, 2.10594e-23, 2.71726e-23, 1.6329e-23, 2.75243e-23], [19.1351, 3.22673e-23, 1.57093e-23, 2.09605e-23, 2.7041e-23, 1.62425e-23, 2.73361e-23], [19.1786, 3.21077e-23, 1.56454e-23, 2.08623e-23, 2.691e-23, 1.61565e-23, 2.71488e-23], [19.2222, 3.19489e-23, 1.55819e-23, 2.07645e-23, 2.67797e-23, 1.60711e-23, 2.69625e-23], [19.266, 3.17908e-23, 1.55188e-23, 2.06673e-23, 2.66501e-23, 1.59862e-23, 2.67772e-23], [19.3098, 3.16336e-23, 1.54563e-23, 2.05707e-23, 2.65211e-23, 1.59019e-23, 2.65928e-23], [19.3537, 3.14771e-23, 1.53942e-23, 2.04745e-23, 2.63928e-23, 1.58182e-23, 2.64094e-23], [19.3977, 3.13213e-23, 1.53325e-23, 2.03789e-23, 2.62651e-23, 1.5735e-23, 2.62269e-23], [19.4419, 3.11664e-23, 1.52714e-23, 2.02839e-23, 2.61381e-23, 1.56524e-23, 2.60454e-23], [19.4861, 3.10122e-23, 1.52106e-23, 2.01893e-23, 2.60117e-23, 1.55703e-23, 2.58648e-23], [19.5304, 3.08588e-23, 1.51503e-23, 2.00953e-23, 2.5886e-23, 1.54887e-23, 2.56852e-23], [19.5748, 3.07061e-23, 1.50905e-23, 2.00018e-23, 2.57609e-23, 1.54077e-23, 2.55065e-23], [19.6193, 3.05542e-23, 1.50311e-23, 1.99089e-23, 2.56365e-23, 1.53272e-23, 2.53288e-23], [19.664, 3.0403e-23, 1.49722e-23, 1.98164e-23, 2.55126e-23, 1.52472e-23, 2.51519e-23], [19.7087, 3.02526e-23, 1.49136e-23, 1.97244e-23, 2.53894e-23, 1.51678e-23, 2.4976e-23], [19.7535, 3.01029e-23, 1.48556e-23, 1.9633e-23, 2.52669e-23, 1.50889e-23, 2.4801e-23], [19.7985, 2.99539e-23, 1.47979e-23, 1.9542e-23, 2.51449e-23, 1.50105e-23, 2.4627e-23], [19.8435, 2.98057e-23, 1.47407e-23, 1.94516e-23, 2.50236e-23, 1.49326e-23, 2.44538e-23], [19.8886, 2.96582e-23, 1.46839e-23, 1.93616e-23, 2.49029e-23, 1.48553e-23, 2.42816e-23], [19.9339, 2.95114e-23, 1.46275e-23, 1.92721e-23, 2.47828e-23, 1.47785e-23, 2.41102e-23], [19.9792, 2.93654e-23, 1.45715e-23, 1.91832e-23, 2.46633e-23, 1.47022e-23, 2.39398e-23], [20.0246, 2.922e-23, 1.45159e-23, 1.90947e-23, 2.45444e-23, 1.46263e-23, 2.37703e-23], [20.0702, 2.90754e-23, 1.44608e-23, 1.90067e-23, 2.44261e-23, 1.4551e-23, 2.36016e-23], [20.1158, 2.89315e-23, 1.44061e-23, 1.89192e-23, 2.43084e-23, 1.44762e-23, 2.34339e-23], [20.1616, 2.87883e-23, 1.43517e-23, 1.88321e-23, 2.41913e-23, 1.44019e-23, 2.32671e-23], [20.2075, 2.86458e-23, 1.42978e-23, 1.87456e-23, 2.40747e-23, 1.43281e-23, 2.31011e-23], [20.2534, 2.8504e-23, 1.42443e-23, 1.86595e-23, 2.39588e-23, 1.42548e-23, 2.2936e-23], [20.2995, 2.83629e-23, 1.41911e-23, 1.85738e-23, 2.38435e-23, 1.4182e-23, 2.27718e-23], [20.3457, 2.82225e-23, 1.41384e-23, 1.84887e-23, 2.37287e-23, 1.41097e-23, 2.26085e-23], [20.3919, 2.80828e-23, 1.4086e-23, 1.8404e-23, 2.36145e-23, 1.40378e-23, 2.2446e-23], [20.4383, 2.79437e-23, 1.4034e-23, 1.83198e-23, 2.35009e-23, 1.39665e-23, 2.22845e-23], [20.4848, 2.78054e-23, 1.39825e-23, 1.8236e-23, 2.33879e-23, 1.38956e-23, 2.21237e-23], [20.5314, 2.76677e-23, 1.39313e-23, 1.81527e-23, 2.32754e-23, 1.38252e-23, 2.19639e-23], [20.5781, 2.75307e-23, 1.38804e-23, 1.80698e-23, 2.31635e-23, 1.37553e-23, 2.18049e-23], [20.6249, 2.73944e-23, 1.383e-23, 1.79874e-23, 2.30522e-23, 1.36858e-23, 2.16468e-23], [20.6718, 2.72587e-23, 1.37799e-23, 1.79054e-23, 2.29414e-23, 1.36169e-23, 2.14895e-23], [20.7188, 2.71237e-23, 1.37302e-23, 1.78239e-23, 2.28312e-23, 1.35484e-23, 2.13331e-23], [20.766, 2.69894e-23, 1.36809e-23, 1.77428e-23, 2.27216e-23, 1.34803e-23, 2.11775e-23], [20.8132, 2.68557e-23, 1.36319e-23, 1.76622e-23, 2.26124e-23, 1.34128e-23, 2.10228e-23], [20.8605, 2.67227e-23, 1.35833e-23, 1.7582e-23, 2.25039e-23, 1.33456e-23, 2.08689e-23], [20.908, 2.65903e-23, 1.35351e-23, 1.75023e-23, 2.23959e-23, 1.3279e-23, 2.07159e-23], [20.9556, 2.64586e-23, 1.34872e-23, 1.74229e-23, 2.22884e-23, 1.32128e-23, 2.05637e-23], [21.0032, 2.63275e-23, 1.34397e-23, 1.7344e-23, 2.21814e-23, 1.31471e-23, 2.04123e-23], [21.051, 2.61971e-23, 1.33925e-23, 1.72656e-23, 2.2075e-23, 1.30818e-23, 2.02618e-23], [21.0989, 2.60673e-23, 1.33457e-23, 1.71875e-23, 2.19692e-23, 1.30169e-23, 2.01121e-23], [21.1469, 2.59381e-23, 1.32992e-23, 1.71099e-23, 2.18638e-23, 1.29526e-23, 1.99632e-23], [21.195, 2.58096e-23, 1.32531e-23, 1.70327e-23, 2.1759e-23, 1.28886e-23, 1.98152e-23], [21.2432, 2.56817e-23, 1.32073e-23, 1.69559e-23, 2.16548e-23, 1.28251e-23, 1.9668e-23], [21.2915, 2.55544e-23, 1.31618e-23, 1.68795e-23, 2.1551e-23, 1.27621e-23, 1.95216e-23], [21.3399, 2.54278e-23, 1.31167e-23, 1.68036e-23, 2.14478e-23, 1.26994e-23, 1.9376e-23], [21.3885, 2.53018e-23, 1.30719e-23, 1.6728e-23, 2.1345e-23, 1.26373e-23, 1.92312e-23], [21.4371, 2.51764e-23, 1.30275e-23, 1.66529e-23, 2.12428e-23, 1.25755e-23, 1.90873e-23], [21.4859, 2.50516e-23, 1.29834e-23, 1.65782e-23, 2.11411e-23, 1.25142e-23, 1.89441e-23], [21.5347, 2.49274e-23, 1.29396e-23, 1.65038e-23, 2.104e-23, 1.24533e-23, 1.88018e-23], [21.5837, 2.48038e-23, 1.28962e-23, 1.64299e-23, 2.09393e-23, 1.23929e-23, 1.86603e-23], [21.6328, 2.46809e-23, 1.2853e-23, 1.63564e-23, 2.08391e-23, 1.23329e-23, 1.85196e-23], [21.682, 2.45585e-23, 1.28102e-23, 1.62833e-23, 2.07394e-23, 1.22733e-23, 1.83797e-23], [21.7313, 2.44367e-23, 1.27678e-23, 1.62105e-23, 2.06403e-23, 1.22141e-23, 1.82405e-23], [21.7808, 2.43156e-23, 1.27256e-23, 1.61382e-23, 2.05416e-23, 1.21553e-23, 1.81022e-23], [21.8303, 2.4195e-23, 1.26838e-23, 1.60662e-23, 2.04434e-23, 1.2097e-23, 1.79647e-23], [21.88, 2.4075e-23, 1.26422e-23, 1.59947e-23, 2.03457e-23, 1.20391e-23, 1.7828e-23], [21.9297, 2.39556e-23, 1.2601e-23, 1.59235e-23, 2.02485e-23, 1.19816e-23, 1.76921e-23], [21.9796, 2.38368e-23, 1.25601e-23, 1.58527e-23, 2.01518e-23, 1.19245e-23, 1.7557e-23], [22.0296, 2.37186e-23, 1.25195e-23, 1.57823e-23, 2.00556e-23, 1.18678e-23, 1.74226e-23], [22.0797, 2.3601e-23, 1.24792e-23, 1.57123e-23, 1.99598e-23, 1.18115e-23, 1.72891e-23], [22.1299, 2.34839e-23, 1.24392e-23, 1.56426e-23, 1.98646e-23, 1.17556e-23, 1.71563e-23], [22.1803, 2.33674e-23, 1.23995e-23, 1.55733e-23, 1.97698e-23, 1.17002e-23, 1.70243e-23], [22.2307, 2.32515e-23, 1.23601e-23, 1.55044e-23, 1.96755e-23, 1.16451e-23, 1.68931e-23], [22.2813, 2.31362e-23, 1.2321e-23, 1.54359e-23, 1.95816e-23, 1.15905e-23, 1.67627e-23], [22.332, 2.30214e-23, 1.22822e-23, 1.53677e-23, 1.94882e-23, 1.15362e-23, 1.66331e-23], [22.3828, 2.29072e-23, 1.22437e-23, 1.52999e-23, 1.93953e-23, 1.14823e-23, 1.65043e-23], [22.4337, 2.27935e-23, 1.22054e-23, 1.52325e-23, 1.93029e-23, 1.14289e-23, 1.63762e-23], [22.4847, 2.26804e-23, 1.21675e-23, 1.51655e-23, 1.92109e-23, 1.13758e-23, 1.62489e-23], [22.5358, 2.25679e-23, 1.21299e-23, 1.50988e-23, 1.91194e-23, 1.13231e-23, 1.61224e-23], [22.5871, 2.24559e-23, 1.20925e-23, 1.50324e-23, 1.90283e-23, 1.12708e-23, 1.59967e-23], [22.6385, 2.23444e-23, 1.20554e-23, 1.49664e-23, 1.89377e-23, 1.12189e-23, 1.58717e-23], [22.69, 2.22335e-23, 1.20186e-23, 1.49008e-23, 1.88476e-23, 1.11674e-23, 1.57475e-23], [22.7416, 2.21232e-23, 1.19821e-23, 1.48355e-23, 1.87579e-23, 1.11163e-23, 1.56241e-23], [22.7933, 2.20134e-23, 1.19459e-23, 1.47706e-23, 1.86686e-23, 1.10656e-23, 1.55014e-23], [22.8452, 2.19041e-23, 1.19099e-23, 1.4706e-23, 1.85798e-23, 1.10152e-23, 1.53796e-23], [22.8971, 2.17954e-23, 1.18742e-23, 1.46418e-23, 1.84915e-23, 1.09652e-23, 1.52584e-23], [22.9492, 2.16872e-23, 1.18388e-23, 1.45779e-23, 1.84035e-23, 1.09156e-23, 1.51381e-23], [23.0014, 2.15796e-23, 1.18036e-23, 1.45144e-23, 1.83161e-23, 1.08664e-23, 1.50185e-23], [23.0537, 2.14724e-23, 1.17688e-23, 1.44512e-23, 1.8229e-23, 1.08175e-23, 1.48997e-23], [23.1062, 2.13658e-23, 1.17341e-23, 1.43883e-23, 1.81424e-23, 1.0769e-23, 1.47817e-23], [23.1587, 2.12598e-23, 1.16998e-23, 1.43258e-23, 1.80562e-23, 1.07209e-23, 1.46644e-23], [23.2114, 2.11542e-23, 1.16657e-23, 1.42637e-23, 1.79705e-23, 1.06732e-23, 1.45479e-23], [23.2642, 2.10492e-23, 1.16319e-23, 1.42018e-23, 1.78852e-23, 1.06258e-23, 1.44322e-23], [23.3171, 2.09447e-23, 1.15983e-23, 1.41403e-23, 1.78003e-23, 1.05788e-23, 1.43172e-23], [23.3701, 2.08407e-23, 1.1565e-23, 1.40791e-23, 1.77159e-23, 1.05322e-23, 1.4203e-23], [23.4233, 2.07372e-23, 1.15319e-23, 1.40183e-23, 1.76318e-23, 1.04859e-23, 1.40895e-23], [23.4766, 2.06342e-23, 1.14991e-23, 1.39578e-23, 1.75482e-23, 1.044e-23, 1.39768e-23], [23.53, 2.05318e-23, 1.14666e-23, 1.38976e-23, 1.7465e-23, 1.03944e-23, 1.38649e-23], [23.5835, 2.04298e-23, 1.14343e-23, 1.38377e-23, 1.73822e-23, 1.03492e-23, 1.37538e-23], [23.6371, 2.03284e-23, 1.14022e-23, 1.37782e-23, 1.72999e-23, 1.03044e-23, 1.36434e-23], [23.6909, 2.02274e-23, 1.13704e-23, 1.3719e-23, 1.72179e-23, 1.02599e-23, 1.35337e-23], [23.7448, 2.0127e-23, 1.13388e-23, 1.36601e-23, 1.71364e-23, 1.02157e-23, 1.34249e-23], [23.7988, 2.0027e-23, 1.13075e-23, 1.36015e-23, 1.70553e-23, 1.01719e-23, 1.33168e-23], [23.8529, 1.99276e-23, 1.12764e-23, 1.35432e-23, 1.69745e-23, 1.01285e-23, 1.32094e-23], [23.9072, 1.98286e-23, 1.12456e-23, 1.34853e-23, 1.68942e-23, 1.00854e-23, 1.31028e-23], [23.9616, 1.97301e-23, 1.1215e-23, 1.34277e-23, 1.68143e-23, 1.00427e-23, 1.2997e-23], [24.0161, 1.96321e-23, 1.11846e-23, 1.33703e-23, 1.67348e-23, 1.00003e-23, 1.2892e-23], [24.0707, 1.95346e-23, 1.11545e-23, 1.33133e-23, 1.66557e-23, 9.95823e-24, 1.27877e-23], [24.1254, 1.94376e-23, 1.11246e-23, 1.32566e-23, 1.6577e-23, 9.91652e-24, 1.26842e-23], [24.1803, 1.93411e-23, 1.10949e-23, 1.32002e-23, 1.64987e-23, 9.87515e-24, 1.25815e-23], [24.2353, 1.9245e-23, 1.10655e-23, 1.31442e-23, 1.64207e-23, 9.83412e-24, 1.24795e-23], [24.2904, 1.91495e-23, 1.10363e-23, 1.30884e-23, 1.63432e-23, 9.79343e-24, 1.23783e-23], [24.3457, 1.90544e-23, 1.10073e-23, 1.30329e-23, 1.62661e-23, 9.75307e-24, 1.22778e-23], [24.4011, 1.89597e-23, 1.09786e-23, 1.29777e-23, 1.61893e-23, 9.71305e-24, 1.21782e-23], [24.4566, 1.88656e-23, 1.09501e-23, 1.29228e-23, 1.61129e-23, 9.67336e-24, 1.20793e-23], [24.5122, 1.87719e-23, 1.09217e-23, 1.28682e-23, 1.60369e-23, 9.634e-24, 1.19811e-23], [24.568, 1.86786e-23, 1.08937e-23, 1.28139e-23, 1.59613e-23, 9.59497e-24, 1.18838e-23], [24.6238, 1.85859e-23, 1.08658e-23, 1.276e-23, 1.58861e-23, 9.55627e-24, 1.17872e-23], [24.6798, 1.84936e-23, 1.08382e-23, 1.27063e-23, 1.58113e-23, 9.5179e-24, 1.16914e-23], [24.736, 1.84018e-23, 1.08107e-23, 1.26528e-23, 1.57368e-23, 9.47985e-24, 1.15964e-23], [24.7922, 1.83104e-23, 1.07835e-23, 1.25997e-23, 1.56627e-23, 9.44213e-24, 1.15022e-23], [24.8486, 1.82195e-23, 1.07565e-23, 1.25469e-23, 1.5589e-23, 9.40473e-24, 1.14087e-23], [24.9052, 1.8129e-23, 1.07297e-23, 1.24944e-23, 1.55156e-23, 9.36765e-24, 1.1316e-23], [24.9618, 1.8039e-23, 1.07031e-23, 1.24421e-23, 1.54427e-23, 9.33089e-24, 1.12241e-23], [25.0186, 1.79494e-23, 1.06768e-23, 1.23901e-23, 1.53701e-23, 9.29444e-24, 1.1133e-23], [25.0755, 1.78603e-23, 1.06506e-23, 1.23384e-23, 1.52978e-23, 9.25832e-24, 1.10427e-23], [25.1325, 1.77716e-23, 1.06247e-23, 1.2287e-23, 1.52259e-23, 9.22251e-24, 1.09531e-23], [25.1897, 1.76834e-23, 1.05989e-23, 1.22359e-23, 1.51544e-23, 9.18701e-24, 1.08644e-23], [25.247, 1.75956e-23, 1.05734e-23, 1.21851e-23, 1.50833e-23, 9.15182e-24, 1.07765e-23], [25.3044, 1.75083e-23, 1.0548e-23, 1.21345e-23, 1.50125e-23, 9.11695e-24, 1.06893e-23], [25.362, 1.74213e-23, 1.05229e-23, 1.20842e-23, 1.49421e-23, 9.08238e-24, 1.06029e-23], [25.4197, 1.73349e-23, 1.04979e-23, 1.20342e-23, 1.4872e-23, 9.04812e-24, 1.05174e-23], [25.4775, 1.72488e-23, 1.04732e-23, 1.19844e-23, 1.48023e-23, 9.01417e-24, 1.04326e-23], [25.5354, 1.71632e-23, 1.04486e-23, 1.1935e-23, 1.47329e-23, 8.98052e-24, 1.03487e-23], [25.5935, 1.70781e-23, 1.04243e-23, 1.18857e-23, 1.46639e-23, 8.94717e-24, 1.02655e-23], [25.6517, 1.69933e-23, 1.04001e-23, 1.18368e-23, 1.45952e-23, 8.91413e-24, 1.01832e-23], [25.7101, 1.6909e-23, 1.03761e-23, 1.17881e-23, 1.45269e-23, 8.88139e-24, 1.01017e-23], [25.7685, 1.68251e-23, 1.03523e-23, 1.17397e-23, 1.44589e-23, 8.84894e-24, 1.00209e-23], [25.8272, 1.67416e-23, 1.03288e-23, 1.16916e-23, 1.43913e-23, 8.81679e-24, 9.94105e-24], [25.8859, 1.66586e-23, 1.03054e-23, 1.16437e-23, 1.4324e-23, 8.78494e-24, 9.86198e-24], [25.9448, 1.6576e-23, 1.02821e-23, 1.15961e-23, 1.42571e-23, 8.75339e-24, 9.78373e-24], [26.0038, 1.64938e-23, 1.02591e-23, 1.15488e-23, 1.41905e-23, 8.72212e-24, 9.7063e-24], [26.0629, 1.6412e-23, 1.02363e-23, 1.15017e-23, 1.41242e-23, 8.69115e-24, 9.6297e-24], [26.1222, 1.63306e-23, 1.02136e-23, 1.14549e-23, 1.40583e-23, 8.66047e-24, 9.55394e-24], [26.1816, 1.62496e-23, 1.01911e-23, 1.14083e-23, 1.39927e-23, 8.63007e-24, 9.479e-24], [26.2412, 1.6169e-23, 1.01688e-23, 1.1362e-23, 1.39274e-23, 8.59997e-24, 9.40491e-24], [26.3009, 1.60889e-23, 1.01467e-23, 1.13159e-23, 1.38625e-23, 8.57015e-24, 9.33166e-24], [26.3607, 1.60092e-23, 1.01247e-23, 1.12701e-23, 1.37979e-23, 8.54061e-24, 9.25924e-24], [26.4207, 1.59298e-23, 1.0103e-23, 1.12246e-23, 1.37336e-23, 8.51135e-24, 9.18768e-24], [26.4808, 1.58509e-23, 1.00814e-23, 1.11793e-23, 1.36697e-23, 8.48238e-24, 9.11697e-24], [26.541, 1.57723e-23, 1.006e-23, 1.11342e-23, 1.36061e-23, 8.45369e-24, 9.04711e-24], [26.6014, 1.56942e-23, 1.00387e-23, 1.10894e-23, 1.35428e-23, 8.42527e-24, 8.97811e-24], [26.6619, 1.56165e-23, 1.00176e-23, 1.10449e-23, 1.34798e-23, 8.39714e-24, 8.90997e-24], [26.7225, 1.55391e-23, 9.99674e-24, 1.10005e-23, 1.34172e-23, 8.36928e-24, 8.84269e-24], [26.7833, 1.54622e-23, 9.97601e-24, 1.09565e-23, 1.33549e-23, 8.34169e-24, 8.77628e-24], [26.8442, 1.53856e-23, 9.95544e-24, 1.09127e-23, 1.32929e-23, 8.31437e-24, 8.71074e-24], [26.9053, 1.53095e-23, 9.93504e-24, 1.08691e-23, 1.32312e-23, 8.28733e-24, 8.64607e-24], [26.9665, 1.52337e-23, 9.91481e-24, 1.08258e-23, 1.31698e-23, 8.26056e-24, 8.58228e-24], [27.0278, 1.51583e-23, 9.89474e-24, 1.07827e-23, 1.31088e-23, 8.23405e-24, 8.51937e-24], [27.0893, 1.50833e-23, 9.87483e-24, 1.07398e-23, 1.3048e-23, 8.20781e-24, 8.45735e-24], [27.1509, 1.50087e-23, 9.85508e-24, 1.06972e-23, 1.29876e-23, 8.18184e-24, 8.39622e-24], [27.2127, 1.49345e-23, 9.8355e-24, 1.06548e-23, 1.29275e-23, 8.15613e-24, 8.33597e-24], [27.2746, 1.48607e-23, 9.81607e-24, 1.06127e-23, 1.28677e-23, 8.13068e-24, 8.27662e-24], [27.3366, 1.47872e-23, 9.79681e-24, 1.05707e-23, 1.28082e-23, 8.10549e-24, 8.21817e-24], [27.3988, 1.47141e-23, 9.7777e-24, 1.05291e-23, 1.2749e-23, 8.08057e-24, 8.16063e-24], [27.4611, 1.46414e-23, 9.75874e-24, 1.04876e-23, 1.26901e-23, 8.0559e-24, 8.10398e-24], [27.5236, 1.4569e-23, 9.73995e-24, 1.04464e-23, 1.26315e-23, 8.03149e-24, 8.04824e-24], [27.5862, 1.44971e-23, 9.7213e-24, 1.04054e-23, 1.25733e-23, 8.00733e-24, 7.99342e-24], [27.6489, 1.44255e-23, 9.70281e-24, 1.03647e-23, 1.25153e-23, 7.98343e-24, 7.93951e-24], [27.7118, 1.43543e-23, 9.68447e-24, 1.03241e-23, 1.24576e-23, 7.95978e-24, 7.88652e-24], [27.7748, 1.42834e-23, 9.66628e-24, 1.02838e-23, 1.24002e-23, 7.93638e-24, 7.83445e-24], [27.838, 1.42129e-23, 9.64824e-24, 1.02438e-23, 1.23431e-23, 7.91323e-24, 7.7833e-24], [27.9013, 1.41428e-23, 9.63035e-24, 1.02039e-23, 1.22863e-23, 7.89032e-24, 7.73308e-24], [27.9648, 1.40731e-23, 9.6126e-24, 1.01643e-23, 1.22298e-23, 7.86767e-24, 7.68378e-24], [28.0284, 1.40037e-23, 9.595e-24, 1.01249e-23, 1.21736e-23, 7.84526e-24, 7.63542e-24], [28.0921, 1.39346e-23, 9.57755e-24, 1.00857e-23, 1.21177e-23, 7.82309e-24, 7.588e-24], [28.156, 1.3866e-23, 9.56024e-24, 1.00467e-23, 1.20621e-23, 7.80116e-24, 7.54151e-24], [28.2201, 1.37977e-23, 9.54307e-24, 1.0008e-23, 1.20068e-23, 7.77948e-24, 7.49596e-24], [28.2843, 1.37297e-23, 9.52605e-24, 9.96948e-24, 1.19517e-23, 7.75803e-24, 7.45135e-24], [28.3486, 1.36621e-23, 9.50916e-24, 9.93117e-24, 1.1897e-23, 7.73682e-24, 7.40768e-24], [28.4131, 1.35949e-23, 9.49242e-24, 9.89308e-24, 1.18425e-23, 7.71585e-24, 7.36496e-24], [28.4777, 1.3528e-23, 9.47581e-24, 9.8552e-24, 1.17883e-23, 7.69512e-24, 7.32319e-24], [28.5425, 1.34614e-23, 9.45934e-24, 9.81754e-24, 1.17344e-23, 7.67461e-24, 7.28236e-24], [28.6074, 1.33952e-23, 9.44301e-24, 9.78009e-24, 1.16808e-23, 7.65434e-24, 7.24249e-24], [28.6725, 1.33294e-23, 9.42682e-24, 9.74285e-24, 1.16275e-23, 7.6343e-24, 7.20356e-24], [28.7377, 1.32639e-23, 9.41075e-24, 9.70583e-24, 1.15744e-23, 7.61448e-24, 7.16559e-24], [28.8031, 1.31987e-23, 9.39483e-24, 9.66901e-24, 1.15216e-23, 7.5949e-24, 7.12857e-24], [28.8686, 1.31339e-23, 9.37903e-24, 9.6324e-24, 1.14691e-23, 7.57554e-24, 7.0925e-24], [28.9342, 1.30694e-23, 9.36337e-24, 9.596e-24, 1.14169e-23, 7.5564e-24, 7.05738e-24], [29.0001, 1.30053e-23, 9.34783e-24, 9.55981e-24, 1.13649e-23, 7.53749e-24, 7.02322e-24], [29.066, 1.29415e-23, 9.33243e-24, 9.52382e-24, 1.13132e-23, 7.5188e-24, 6.99001e-24], [29.1321, 1.28781e-23, 9.31715e-24, 9.48804e-24, 1.12618e-23, 7.50033e-24, 6.95775e-24], [29.1984, 1.28149e-23, 9.30201e-24, 9.45246e-24, 1.12107e-23, 7.48207e-24, 6.92645e-24], [29.2648, 1.27522e-23, 9.28699e-24, 9.41709e-24, 1.11598e-23, 7.46404e-24, 6.89609e-24], [29.3314, 1.26897e-23, 9.27209e-24, 9.38191e-24, 1.11092e-23, 7.44622e-24, 6.86669e-24], [29.3981, 1.26276e-23, 9.25732e-24, 9.34694e-24, 1.10589e-23, 7.42861e-24, 6.83823e-24], [29.465, 1.25658e-23, 9.24267e-24, 9.31217e-24, 1.10088e-23, 7.41122e-24, 6.81072e-24], [29.532, 1.25044e-23, 9.22815e-24, 9.27759e-24, 1.0959e-23, 7.39403e-24, 6.78416e-24], [29.5992, 1.24432e-23, 9.21375e-24, 9.24322e-24, 1.09094e-23, 7.37706e-24, 6.75853e-24], [29.6665, 1.23824e-23, 9.19947e-24, 9.20904e-24, 1.08602e-23, 7.3603e-24, 6.73385e-24], [29.734, 1.2322e-23, 9.1853e-24, 9.17506e-24, 1.08111e-23, 7.34374e-24, 6.7101e-24], [29.8016, 1.22618e-23, 9.17126e-24, 9.14127e-24, 1.07624e-23, 7.32739e-24, 6.68729e-24], [29.8694, 1.2202e-23, 9.15734e-24, 9.10768e-24, 1.07139e-23, 7.31124e-24, 6.6654e-24], [29.9373, 1.21425e-23, 9.14353e-24, 9.07428e-24, 1.06656e-23, 7.29529e-24, 6.64444e-24], [30.0054, 1.20833e-23, 9.12984e-24, 9.04108e-24, 1.06176e-23, 7.27955e-24, 6.6244e-24], [30.0737, 1.20244e-23, 9.11627e-24, 9.00806e-24, 1.05699e-23, 7.264e-24, 6.60528e-24], [30.1421, 1.19659e-23, 9.10281e-24, 8.97524e-24, 1.05224e-23, 7.24865e-24, 6.58708e-24], [30.2106, 1.19077e-23, 9.08946e-24, 8.9426e-24, 1.04752e-23, 7.2335e-24, 6.56978e-24], [30.2793, 1.18498e-23, 9.07623e-24, 8.91016e-24, 1.04282e-23, 7.21854e-24, 6.55338e-24], [30.3482, 1.17922e-23, 9.0631e-24, 8.8779e-24, 1.03815e-23, 7.20378e-24, 6.53788e-24], [30.4172, 1.17349e-23, 9.05009e-24, 8.84583e-24, 1.03351e-23, 7.18921e-24, 6.52327e-24], [30.4864, 1.16779e-23, 9.03719e-24, 8.81394e-24, 1.02888e-23, 7.17482e-24, 6.50954e-24], [30.5558, 1.16213e-23, 9.0244e-24, 8.78225e-24, 1.02429e-23, 7.16063e-24, 6.4967e-24], [30.6253, 1.15649e-23, 9.01172e-24, 8.75073e-24, 1.01971e-23, 7.14663e-24, 6.48472e-24], [30.6949, 1.15089e-23, 8.99914e-24, 8.7194e-24, 1.01517e-23, 7.13281e-24, 6.47361e-24], [30.7647, 1.14531e-23, 8.98667e-24, 8.68825e-24, 1.01064e-23, 7.11918e-24, 6.46336e-24], [30.8347, 1.13977e-23, 8.97431e-24, 8.65728e-24, 1.00615e-23, 7.10573e-24, 6.45395e-24], [30.9049, 1.13426e-23, 8.96205e-24, 8.6265e-24, 1.00167e-23, 7.09246e-24, 6.44539e-24], [30.9752, 1.12877e-23, 8.9499e-24, 8.59589e-24, 9.9722e-24, 7.07937e-24, 6.43766e-24], [31.0456, 1.12332e-23, 8.93785e-24, 8.56546e-24, 9.92794e-24, 7.06646e-24, 6.43076e-24], [31.1162, 1.1179e-23, 8.9259e-24, 8.53521e-24, 9.88392e-24, 7.05373e-24, 6.42467e-24], [31.187, 1.11251e-23, 8.91406e-24, 8.50514e-24, 9.84013e-24, 7.04117e-24, 6.4194e-24], [31.2579, 1.10714e-23, 8.90231e-24, 8.47524e-24, 9.79659e-24, 7.02879e-24, 6.41492e-24], [31.329, 1.10181e-23, 8.89067e-24, 8.44552e-24, 9.75328e-24, 7.01658e-24, 6.41123e-24], [31.4003, 1.09651e-23, 8.87912e-24, 8.41598e-24, 9.7102e-24, 7.00454e-24, 6.40832e-24], [31.4717, 1.09124e-23, 8.86767e-24, 8.3866e-24, 9.66736e-24, 6.99267e-24, 6.40618e-24], [31.5433, 1.08599e-23, 8.85633e-24, 8.35741e-24, 9.62476e-24, 6.98098e-24, 6.4048e-24], [31.6151, 1.08078e-23, 8.84507e-24, 8.32838e-24, 9.58239e-24, 6.96944e-24, 6.40418e-24], [31.687, 1.07559e-23, 8.83392e-24, 8.29952e-24, 9.54024e-24, 6.95808e-24, 6.40429e-24], [31.759, 1.07043e-23, 8.82286e-24, 8.27084e-24, 9.49833e-24, 6.94688e-24, 6.40514e-24], [31.8313, 1.06531e-23, 8.81189e-24, 8.24233e-24, 9.45665e-24, 6.93584e-24, 6.40671e-24], [31.9037, 1.06021e-23, 8.80102e-24, 8.21398e-24, 9.4152e-24, 6.92497e-24, 6.40898e-24], [31.9762, 1.05514e-23, 8.79024e-24, 8.1858e-24, 9.37398e-24, 6.91425e-24, 6.41196e-24], [32.049, 1.0501e-23, 8.77956e-24, 8.15779e-24, 9.33298e-24, 6.90369e-24, 6.41562e-24], [32.1219, 1.04508e-23, 8.76896e-24, 8.12995e-24, 9.29221e-24, 6.89329e-24, 6.41997e-24], [32.1949, 1.0401e-23, 8.75846e-24, 8.10227e-24, 9.25166e-24, 6.88305e-24, 6.42498e-24], [32.2682, 1.03514e-23, 8.74804e-24, 8.07476e-24, 9.21134e-24, 6.87296e-24, 6.43064e-24], [32.3416, 1.03021e-23, 8.73772e-24, 8.04741e-24, 9.17124e-24, 6.86302e-24, 6.43695e-24], [32.4151, 1.02531e-23, 8.72748e-24, 8.02022e-24, 9.13137e-24, 6.85324e-24, 6.44389e-24], [32.4889, 1.02044e-23, 8.71734e-24, 7.9932e-24, 9.09171e-24, 6.8436e-24, 6.45146e-24], [32.5628, 1.0156e-23, 8.70728e-24, 7.96633e-24, 9.05228e-24, 6.83412e-24, 6.45964e-24], [32.6368, 1.01078e-23, 8.6973e-24, 7.93963e-24, 9.01306e-24, 6.82478e-24, 6.46841e-24], [32.7111, 1.00599e-23, 8.68741e-24, 7.91309e-24, 8.97406e-24, 6.81559e-24, 6.47778e-24], [32.7855, 1.00123e-23, 8.67761e-24, 7.88671e-24, 8.93528e-24, 6.80654e-24, 6.48772e-24], [32.86, 9.96495e-24, 8.66789e-24, 7.86048e-24, 8.89672e-24, 6.79764e-24, 6.49823e-24], [32.9348, 9.91788e-24, 8.65826e-24, 7.83442e-24, 8.85837e-24, 6.78887e-24, 6.5093e-24], [33.0097, 9.87108e-24, 8.64871e-24, 7.80851e-24, 8.82024e-24, 6.78025e-24, 6.52091e-24], [33.0848, 9.82456e-24, 8.63924e-24, 7.78276e-24, 8.78232e-24, 6.77177e-24, 6.53305e-24], [33.16, 9.7783e-24, 8.62985e-24, 7.75716e-24, 8.74461e-24, 6.76343e-24, 6.54572e-24], [33.2354, 9.73232e-24, 8.62054e-24, 7.73172e-24, 8.70712e-24, 6.75522e-24, 6.5589e-24], [33.311, 9.6866e-24, 8.61132e-24, 7.70643e-24, 8.66983e-24, 6.74715e-24, 6.57257e-24], [33.3868, 9.64115e-24, 8.60217e-24, 7.68129e-24, 8.63276e-24, 6.73921e-24, 6.58674e-24], [33.4628, 9.59596e-24, 8.5931e-24, 7.65631e-24, 8.5959e-24, 6.7314e-24, 6.60138e-24], [33.5389, 9.55104e-24, 8.58412e-24, 7.63147e-24, 8.55924e-24, 6.72373e-24, 6.61649e-24], [33.6152, 9.50639e-24, 8.5752e-24, 7.60679e-24, 8.5228e-24, 6.71619e-24, 6.63206e-24], [33.6916, 9.46199e-24, 8.56637e-24, 7.58226e-24, 8.48655e-24, 6.70877e-24, 6.64807e-24], [33.7682, 9.41786e-24, 8.55761e-24, 7.55788e-24, 8.45052e-24, 6.70148e-24, 6.66452e-24], [33.8451, 9.37399e-24, 8.54893e-24, 7.53365e-24, 8.41469e-24, 6.69432e-24, 6.6814e-24], [33.922, 9.33038e-24, 8.54032e-24, 7.50956e-24, 8.37906e-24, 6.68728e-24, 6.69869e-24], [33.9992, 9.28703e-24, 8.53179e-24, 7.48562e-24, 8.34364e-24, 6.68037e-24, 6.71638e-24], [34.0765, 9.24394e-24, 8.52333e-24, 7.46183e-24, 8.30842e-24, 6.67357e-24, 6.73447e-24], [34.154, 9.2011e-24, 8.51494e-24, 7.43818e-24, 8.27341e-24, 6.6669e-24, 6.75294e-24], [34.2317, 9.15852e-24, 8.50663e-24, 7.41468e-24, 8.23859e-24, 6.66035e-24, 6.77179e-24], [34.3096, 9.1162e-24, 8.49839e-24, 7.39132e-24, 8.20397e-24, 6.65392e-24, 6.791e-24], [34.3876, 9.07413e-24, 8.49022e-24, 7.36811e-24, 8.16955e-24, 6.6476e-24, 6.81057e-24], [34.4658, 9.03231e-24, 8.48212e-24, 7.34504e-24, 8.13533e-24, 6.6414e-24, 6.83049e-24], [34.5442, 8.99075e-24, 8.47409e-24, 7.32211e-24, 8.10131e-24, 6.63531e-24, 6.85074e-24], [34.6228, 8.94943e-24, 8.46613e-24, 7.29932e-24, 8.06748e-24, 6.62934e-24, 6.87131e-24], [34.7016, 8.90837e-24, 8.45824e-24, 7.27668e-24, 8.03385e-24, 6.62348e-24, 6.89221e-24], [34.7805, 8.86756e-24, 8.45042e-24, 7.25417e-24, 8.00042e-24, 6.61773e-24, 6.91341e-24], [34.8596, 8.827e-24, 8.44267e-24, 7.2318e-24, 7.96718e-24, 6.61209e-24, 6.93492e-24], [34.9389, 8.78668e-24, 8.43498e-24, 7.20957e-24, 7.93413e-24, 6.60655e-24, 6.95671e-24], [35.0184, 8.74661e-24, 8.42736e-24, 7.18748e-24, 7.90127e-24, 6.60113e-24, 6.97879e-24], [35.098, 8.70679e-24, 8.41981e-24, 7.16552e-24, 7.86861e-24, 6.59581e-24, 7.00114e-24], [35.1779, 8.66721e-24, 8.41232e-24, 7.1437e-24, 7.83614e-24, 6.59059e-24, 7.02376e-24], [35.2579, 8.62787e-24, 8.4049e-24, 7.12202e-24, 7.80386e-24, 6.58548e-24, 7.04664e-24], [35.3381, 8.58878e-24, 8.39754e-24, 7.10047e-24, 7.77176e-24, 6.58047e-24, 7.06977e-24], [35.4184, 8.54993e-24, 8.39024e-24, 7.07906e-24, 7.73986e-24, 6.57556e-24, 7.09314e-24], [35.499, 8.51133e-24, 8.38301e-24, 7.05778e-24, 7.70814e-24, 6.57075e-24, 7.11674e-24], [35.5798, 8.47296e-24, 8.37584e-24, 7.03663e-24, 7.67661e-24, 6.56604e-24, 7.14057e-24], [35.6607, 8.43483e-24, 8.36874e-24, 7.01562e-24, 7.64527e-24, 6.56143e-24, 7.16462e-24], [35.7418, 8.39695e-24, 8.36169e-24, 6.99473e-24, 7.61411e-24, 6.55691e-24, 7.18888e-24], [35.8231, 8.35929e-24, 8.35471e-24, 6.97398e-24, 7.58314e-24, 6.55249e-24, 7.21335e-24], [35.9046, 8.32188e-24, 8.34779e-24, 6.95336e-24, 7.55235e-24, 6.54817e-24, 7.23802e-24], [35.9862, 8.2847e-24, 8.34092e-24, 6.93287e-24, 7.52174e-24, 6.54393e-24, 7.26287e-24], [36.0681, 8.24776e-24, 8.33412e-24, 6.9125e-24, 7.49131e-24, 6.53979e-24, 7.28791e-24], [36.1501, 8.21106e-24, 8.32737e-24, 6.89227e-24, 7.46107e-24, 6.53574e-24, 7.31313e-24], [36.2324, 8.17458e-24, 8.32069e-24, 6.87216e-24, 7.43101e-24, 6.53178e-24, 7.33852e-24], [36.3148, 8.13834e-24, 8.31406e-24, 6.85218e-24, 7.40113e-24, 6.52791e-24, 7.36407e-24], [36.3974, 8.10233e-24, 8.30749e-24, 6.83232e-24, 7.37142e-24, 6.52412e-24, 7.38978e-24], [36.4802, 8.06655e-24, 8.30098e-24, 6.81259e-24, 7.3419e-24, 6.52042e-24, 7.41564e-24], [36.5631, 8.03101e-24, 8.29452e-24, 6.79299e-24, 7.31255e-24, 6.51681e-24, 7.44165e-24], [36.6463, 7.99569e-24, 8.28812e-24, 6.77351e-24, 7.28338e-24, 6.51328e-24, 7.4678e-24], [36.7297, 7.9606e-24, 8.28178e-24, 6.75415e-24, 7.25439e-24, 6.50983e-24, 7.49409e-24], [36.8132, 7.92574e-24, 8.27549e-24, 6.73491e-24, 7.22557e-24, 6.50647e-24, 7.5205e-24], [36.8969, 7.8911e-24, 8.26925e-24, 6.7158e-24, 7.19693e-24, 6.50319e-24, 7.54704e-24], [36.9809, 7.85669e-24, 8.26307e-24, 6.69681e-24, 7.16846e-24, 6.49998e-24, 7.57369e-24], [37.065, 7.82251e-24, 8.25694e-24, 6.67794e-24, 7.14016e-24, 6.49686e-24, 7.60046e-24], [37.1493, 7.78855e-24, 8.25087e-24, 6.65919e-24, 7.11204e-24, 6.49381e-24, 7.62733e-24], [37.2338, 7.75481e-24, 8.24485e-24, 6.64056e-24, 7.08408e-24, 6.49085e-24, 7.65431e-24], [37.3185, 7.7213e-24, 8.23888e-24, 6.62205e-24, 7.0563e-24, 6.48795e-24, 7.68139e-24], [37.4034, 7.68801e-24, 8.23296e-24, 6.60366e-24, 7.02869e-24, 6.48513e-24, 7.70855e-24], [37.4884, 7.65494e-24, 8.22709e-24, 6.58539e-24, 7.00125e-24, 6.48239e-24, 7.73581e-24], [37.5737, 7.62209e-24, 8.22128e-24, 6.56723e-24, 6.97398e-24, 6.47972e-24, 7.76315e-24], [37.6592, 7.58946e-24, 8.21551e-24, 6.54919e-24, 6.94688e-24, 6.47712e-24, 7.79057e-24], [37.7448, 7.55705e-24, 8.20979e-24, 6.53126e-24, 6.91994e-24, 6.47459e-24, 7.81806e-24], [37.8307, 7.52486e-24, 8.20413e-24, 6.51345e-24, 6.89317e-24, 6.47213e-24, 7.84562e-24], [37.9167, 7.49288e-24, 8.19851e-24, 6.49576e-24, 6.86657e-24, 6.46974e-24, 7.87325e-24], [38.003, 7.46112e-24, 8.19294e-24, 6.47818e-24, 6.84013e-24, 6.46742e-24, 7.90094e-24], [38.0894, 7.42958e-24, 8.18742e-24, 6.46071e-24, 6.81385e-24, 6.46516e-24, 7.92869e-24], [38.176, 7.39825e-24, 8.18195e-24, 6.44336e-24, 6.78774e-24, 6.46298e-24, 7.95649e-24], [38.2629, 7.36713e-24, 8.17653e-24, 6.42611e-24, 6.7618e-24, 6.46085e-24, 7.98435e-24], [38.3499, 7.33623e-24, 8.17115e-24, 6.40898e-24, 6.73601e-24, 6.45879e-24, 8.01225e-24], [38.4371, 7.30554e-24, 8.16582e-24, 6.39196e-24, 6.71039e-24, 6.4568e-24, 8.04019e-24], [38.5246, 7.27507e-24, 8.16053e-24, 6.37505e-24, 6.68493e-24, 6.45486e-24, 8.06817e-24], [38.6122, 7.2448e-24, 8.15529e-24, 6.35826e-24, 6.65963e-24, 6.45299e-24, 8.09619e-24], [38.7, 7.21474e-24, 8.1501e-24, 6.34156e-24, 6.63449e-24, 6.45118e-24, 8.12424e-24], [38.788, 7.1849e-24, 8.14495e-24, 6.32498e-24, 6.6095e-24, 6.44943e-24, 8.15232e-24], [38.8763, 7.15526e-24, 8.13984e-24, 6.30851e-24, 6.58468e-24, 6.44774e-24, 8.18042e-24], [38.9647, 7.12583e-24, 8.13478e-24, 6.29214e-24, 6.56001e-24, 6.44611e-24, 8.20855e-24], [39.0533, 7.0966e-24, 8.12976e-24, 6.27588e-24, 6.5355e-24, 6.44453e-24, 8.2367e-24], [39.1422, 7.06759e-24, 8.12479e-24, 6.25973e-24, 6.51115e-24, 6.44301e-24, 8.26486e-24], [39.2312, 7.03877e-24, 8.11986e-24, 6.24368e-24, 6.48695e-24, 6.44154e-24, 8.29304e-24], [39.3204, 7.01017e-24, 8.11497e-24, 6.22774e-24, 6.46291e-24, 6.44013e-24, 8.32123e-24], [39.4099, 6.98176e-24, 8.11012e-24, 6.2119e-24, 6.43902e-24, 6.43878e-24, 8.34942e-24], [39.4995, 6.95357e-24, 8.10532e-24, 6.19616e-24, 6.41529e-24, 6.43747e-24, 8.37763e-24], [39.5893, 6.92557e-24, 8.10056e-24, 6.18053e-24, 6.3917e-24, 6.43622e-24, 8.40583e-24], [39.6794, 6.89777e-24, 8.09583e-24, 6.165e-24, 6.36827e-24, 6.43502e-24, 8.43403e-24], [39.7696, 6.87018e-24, 8.09115e-24, 6.14958e-24, 6.345e-24, 6.43387e-24, 8.46224e-24], [39.8601, 6.84279e-24, 8.08651e-24, 6.13425e-24, 6.32187e-24, 6.43278e-24, 8.49043e-24], [39.9508, 6.81559e-24, 8.08191e-24, 6.11903e-24, 6.29889e-24, 6.43173e-24, 8.51862e-24], [40.0416, 6.7886e-24, 8.07734e-24, 6.1039e-24, 6.27606e-24, 6.43072e-24, 8.54681e-24], [40.1327, 6.7618e-24, 8.07282e-24, 6.08888e-24, 6.25338e-24, 6.42977e-24, 8.57497e-24], [40.224, 6.7352e-24, 8.06834e-24, 6.07395e-24, 6.23085e-24, 6.42886e-24, 8.60313e-24], [40.3155, 6.7088e-24, 8.06389e-24, 6.05913e-24, 6.20847e-24, 6.428e-24, 8.63127e-24], [40.4072, 6.68259e-24, 8.05948e-24, 6.0444e-24, 6.18623e-24, 6.42718e-24, 8.65939e-24], [40.4991, 6.65658e-24, 8.05511e-24, 6.02977e-24, 6.16414e-24, 6.42641e-24, 8.68749e-24], [40.5912, 6.63076e-24, 8.05078e-24, 6.01523e-24, 6.1422e-24, 6.42568e-24, 8.71557e-24], [40.6836, 6.60514e-24, 8.04648e-24, 6.0008e-24, 6.1204e-24, 6.42499e-24, 8.74362e-24], [40.7761, 6.57971e-24, 8.04222e-24, 5.98645e-24, 6.09874e-24, 6.42435e-24, 8.77165e-24], [40.8688, 6.55447e-24, 8.038e-24, 5.97221e-24, 6.07723e-24, 6.42375e-24, 8.79965e-24], [40.9618, 6.52942e-24, 8.03381e-24, 5.95806e-24, 6.05586e-24, 6.42319e-24, 8.82762e-24], [41.055, 6.50456e-24, 8.02966e-24, 5.944e-24, 6.03464e-24, 6.42266e-24, 8.85556e-24], [41.1483, 6.4799e-24, 8.02555e-24, 5.93004e-24, 6.01355e-24, 6.42218e-24, 8.88347e-24], [41.2419, 6.45542e-24, 8.02146e-24, 5.91617e-24, 5.99261e-24, 6.42174e-24, 8.91134e-24], [41.3358, 6.43113e-24, 8.01742e-24, 5.90239e-24, 5.9718e-24, 6.42133e-24, 8.93917e-24], [41.4298, 6.40703e-24, 8.01341e-24, 5.8887e-24, 5.95114e-24, 6.42097e-24, 8.96697e-24], [41.524, 6.38312e-24, 8.00943e-24, 5.87511e-24, 5.93062e-24, 6.42063e-24, 8.99473e-24], [41.6185, 6.35939e-24, 8.00549e-24, 5.8616e-24, 5.91023e-24, 6.42034e-24, 9.02245e-24], [41.7131, 6.33585e-24, 8.00158e-24, 5.84819e-24, 5.88998e-24, 6.42008e-24, 9.05012e-24], [41.808, 6.3125e-24, 7.9977e-24, 5.83487e-24, 5.86987e-24, 6.41985e-24, 9.07775e-24], [41.9031, 6.28932e-24, 7.99385e-24, 5.82163e-24, 5.8499e-24, 6.41966e-24, 9.10534e-24], [41.9984, 6.26634e-24, 7.99004e-24, 5.80849e-24, 5.83006e-24, 6.4195e-24, 9.13288e-24], [42.0939, 6.24353e-24, 7.98626e-24, 5.79543e-24, 5.81035e-24, 6.41938e-24, 9.16037e-24], [42.1897, 6.22091e-24, 7.98251e-24, 5.78246e-24, 5.79078e-24, 6.41929e-24, 9.18781e-24], [42.2856, 6.19847e-24, 7.9788e-24, 5.76958e-24, 5.77135e-24, 6.41922e-24, 9.2152e-24], [42.3818, 6.17621e-24, 7.97511e-24, 5.75679e-24, 5.75205e-24, 6.41919e-24, 9.24254e-24], [42.4782, 6.15413e-24, 7.97146e-24, 5.74408e-24, 5.73288e-24, 6.41919e-24, 9.26983e-24], [42.5748, 6.13222e-24, 7.96784e-24, 5.73146e-24, 5.71384e-24, 6.41922e-24, 9.29707e-24], [42.6717, 6.1105e-24, 7.96425e-24, 5.71892e-24, 5.69494e-24, 6.41928e-24, 9.32425e-24], [42.7687, 6.08896e-24, 7.96068e-24, 5.70646e-24, 5.67616e-24, 6.41937e-24, 9.35138e-24], [42.866, 6.06759e-24, 7.95715e-24, 5.6941e-24, 5.65752e-24, 6.41948e-24, 9.37844e-24], [42.9635, 6.0464e-24, 7.95365e-24, 5.68181e-24, 5.639e-24, 6.41962e-24, 9.40545e-24], [43.0613, 6.02538e-24, 7.95018e-24, 5.66961e-24, 5.62062e-24, 6.41979e-24, 9.43241e-24], [43.1592, 6.00454e-24, 7.94673e-24, 5.65749e-24, 5.60236e-24, 6.41999e-24, 9.4593e-24], [43.2574, 5.98388e-24, 7.94332e-24, 5.64545e-24, 5.58423e-24, 6.42021e-24, 9.48613e-24], [43.3558, 5.96339e-24, 7.93993e-24, 5.63349e-24, 5.56623e-24, 6.42046e-24, 9.5129e-24], [43.4544, 5.94307e-24, 7.93657e-24, 5.62162e-24, 5.54835e-24, 6.42073e-24, 9.53961e-24], [43.5532, 5.92292e-24, 7.93324e-24, 5.60983e-24, 5.5306e-24, 6.42102e-24, 9.56626e-24], [43.6523, 5.90294e-24, 7.92994e-24, 5.59811e-24, 5.51298e-24, 6.42134e-24, 9.59284e-24], [43.7516, 5.88314e-24, 7.92666e-24, 5.58648e-24, 5.49548e-24, 6.42168e-24, 9.61936e-24], [43.8511, 5.86351e-24, 7.92342e-24, 5.57492e-24, 5.4781e-24, 6.42204e-24, 9.64581e-24], [43.9508, 5.84404e-24, 7.9202e-24, 5.56345e-24, 5.46085e-24, 6.42243e-24, 9.6722e-24], [44.0508, 5.82474e-24, 7.917e-24, 5.55205e-24, 5.44372e-24, 6.42284e-24, 9.69852e-24], [44.151, 5.80562e-24, 7.91384e-24, 5.54073e-24, 5.42672e-24, 6.42326e-24, 9.72477e-24], [44.2514, 5.78666e-24, 7.91069e-24, 5.52948e-24, 5.40983e-24, 6.42371e-24, 9.75095e-24], [44.3521, 5.76786e-24, 7.90758e-24, 5.51832e-24, 5.39307e-24, 6.42418e-24, 9.77707e-24], [44.453, 5.74923e-24, 7.90449e-24, 5.50722e-24, 5.37642e-24, 6.42467e-24, 9.80311e-24], [44.5541, 5.73077e-24, 7.90143e-24, 5.49621e-24, 5.3599e-24, 6.42517e-24, 9.82909e-24], [44.6554, 5.71247e-24, 7.89839e-24, 5.48527e-24, 5.34349e-24, 6.4257e-24, 9.855e-24], [44.757, 5.69434e-24, 7.89537e-24, 5.47441e-24, 5.32721e-24, 6.42624e-24, 9.88083e-24], [44.8588, 5.67637e-24, 7.89238e-24, 5.46361e-24, 5.31104e-24, 6.4268e-24, 9.9066e-24], [44.9608, 5.65856e-24, 7.88942e-24, 5.4529e-24, 5.29499e-24, 6.42738e-24, 9.93229e-24], [45.0631, 5.64091e-24, 7.88648e-24, 5.44225e-24, 5.27905e-24, 6.42798e-24, 9.95791e-24], [45.1656, 5.62342e-24, 7.88356e-24, 5.43168e-24, 5.26324e-24, 6.42859e-24, 9.98345e-24], [45.2683, 5.6061e-24, 7.88067e-24, 5.42119e-24, 5.24753e-24, 6.42921e-24, 1.00089e-23], [45.3713, 5.58893e-24, 7.8778e-24, 5.41076e-24, 5.23195e-24, 6.42985e-24, 1.00343e-23], [45.4745, 5.57192e-24, 7.87496e-24, 5.40041e-24, 5.21647e-24, 6.43051e-24, 1.00597e-23], [45.5779, 5.55508e-24, 7.87214e-24, 5.39012e-24, 5.20112e-24, 6.43118e-24, 1.00849e-23], [45.6816, 5.53838e-24, 7.86934e-24, 5.37991e-24, 5.18587e-24, 6.43186e-24, 1.01101e-23], [45.7855, 5.52185e-24, 7.86656e-24, 5.36977e-24, 5.17074e-24, 6.43256e-24, 1.01352e-23], [45.8896, 5.50547e-24, 7.86381e-24, 5.35969e-24, 5.15572e-24, 6.43327e-24, 1.01602e-23], [45.994, 5.48925e-24, 7.86108e-24, 5.34969e-24, 5.14081e-24, 6.434e-24, 1.01851e-23], [46.0986, 5.47318e-24, 7.85837e-24, 5.33976e-24, 5.12601e-24, 6.43473e-24, 1.021e-23], [46.2035, 5.45726e-24, 7.85568e-24, 5.32989e-24, 5.11132e-24, 6.43548e-24, 1.02348e-23], [46.3086, 5.4415e-24, 7.85302e-24, 5.32009e-24, 5.09674e-24, 6.43624e-24, 1.02595e-23], [46.4139, 5.42589e-24, 7.85038e-24, 5.31036e-24, 5.08227e-24, 6.43701e-24, 1.02842e-23], [46.5195, 5.41044e-24, 7.84775e-24, 5.3007e-24, 5.06791e-24, 6.43779e-24, 1.03087e-23], [46.6253, 5.39513e-24, 7.84515e-24, 5.2911e-24, 5.05366e-24, 6.43859e-24, 1.03332e-23], [46.7313, 5.37998e-24, 7.84257e-24, 5.28157e-24, 5.03951e-24, 6.43939e-24, 1.03576e-23], [46.8376, 5.36497e-24, 7.84001e-24, 5.2721e-24, 5.02547e-24, 6.4402e-24, 1.03819e-23], [46.9442, 5.35012e-24, 7.83747e-24, 5.2627e-24, 5.01154e-24, 6.44102e-24, 1.04062e-23], [47.0509, 5.33541e-24, 7.83495e-24, 5.25337e-24, 4.99771e-24, 6.44185e-24, 1.04304e-23], [47.158, 5.32085e-24, 7.83246e-24, 5.24409e-24, 4.98399e-24, 6.44269e-24, 1.04545e-23], [47.2652, 5.30644e-24, 7.82998e-24, 5.23489e-24, 4.97038e-24, 6.44353e-24, 1.04785e-23], [47.3727, 5.29217e-24, 7.82752e-24, 5.22574e-24, 4.95686e-24, 6.44439e-24, 1.05024e-23], [47.4805, 5.27805e-24, 7.82508e-24, 5.21666e-24, 4.94345e-24, 6.44525e-24, 1.05263e-23], [47.5885, 5.26407e-24, 7.82266e-24, 5.20765e-24, 4.93014e-24, 6.44611e-24, 1.055e-23], [47.6967, 5.25024e-24, 7.82026e-24, 5.19869e-24, 4.91694e-24, 6.44699e-24, 1.05737e-23], [47.8052, 5.23655e-24, 7.81788e-24, 5.1898e-24, 4.90383e-24, 6.44787e-24, 1.05973e-23], [47.914, 5.22301e-24, 7.81552e-24, 5.18097e-24, 4.89083e-24, 6.44876e-24, 1.06209e-23], [48.0229, 5.20961e-24, 7.81317e-24, 5.17219e-24, 4.87793e-24, 6.44965e-24, 1.06443e-23], [48.1322, 5.19634e-24, 7.81085e-24, 5.16348e-24, 4.86513e-24, 6.45055e-24, 1.06677e-23], [48.2417, 5.18322e-24, 7.80854e-24, 5.15483e-24, 4.85242e-24, 6.45145e-24, 1.0691e-23], [48.3514, 5.17024e-24, 7.80625e-24, 5.14624e-24, 4.83982e-24, 6.45236e-24, 1.07142e-23], [48.4614, 5.1574e-24, 7.80398e-24, 5.13771e-24, 4.82731e-24, 6.45327e-24, 1.07373e-23], [48.5716, 5.1447e-24, 7.80172e-24, 5.12924e-24, 4.8149e-24, 6.45419e-24, 1.07604e-23], [48.6821, 5.13213e-24, 7.79949e-24, 5.12083e-24, 4.80259e-24, 6.45511e-24, 1.07833e-23], [48.7928, 5.1197e-24, 7.79727e-24, 5.11247e-24, 4.79038e-24, 6.45603e-24, 1.08062e-23], [48.9038, 5.10741e-24, 7.79507e-24, 5.10418e-24, 4.77826e-24, 6.45696e-24, 1.0829e-23], [49.015, 5.09526e-24, 7.79289e-24, 5.09594e-24, 4.76623e-24, 6.45789e-24, 1.08518e-23], [49.1265, 5.08324e-24, 7.79072e-24, 5.08775e-24, 4.75431e-24, 6.45882e-24, 1.08744e-23], [49.2383, 5.07135e-24, 7.78857e-24, 5.07963e-24, 4.74247e-24, 6.45975e-24, 1.0897e-23], [49.3502, 5.0596e-24, 7.78643e-24, 5.07156e-24, 4.73073e-24, 6.46069e-24, 1.09194e-23], [49.4625, 5.04798e-24, 7.78432e-24, 5.06355e-24, 4.71908e-24, 6.46163e-24, 1.09419e-23], [49.575, 5.03649e-24, 7.78222e-24, 5.05559e-24, 4.70753e-24, 6.46257e-24, 1.09642e-23], [49.6878, 5.02514e-24, 7.78013e-24, 5.04768e-24, 4.69606e-24, 6.46351e-24, 1.09864e-23], [49.8008, 5.01391e-24, 7.77806e-24, 5.03984e-24, 4.68469e-24, 6.46445e-24, 1.10086e-23], [49.9141, 5.00282e-24, 7.77601e-24, 5.03204e-24, 4.67341e-24, 6.46539e-24, 1.10306e-23], [50.0276, 4.99185e-24, 7.77397e-24, 5.0243e-24, 4.66222e-24, 6.46634e-24, 1.10526e-23], [50.1414, 4.98102e-24, 7.77195e-24, 5.01661e-24, 4.65112e-24, 6.46728e-24, 1.10745e-23], [50.2554, 4.97031e-24, 7.76995e-24, 5.00898e-24, 4.6401e-24, 6.46822e-24, 1.10964e-23], [50.3697, 4.95973e-24, 7.76796e-24, 5.0014e-24, 4.62918e-24, 6.46916e-24, 1.11181e-23], [50.4843, 4.94928e-24, 7.76598e-24, 4.99387e-24, 4.61835e-24, 6.4701e-24, 1.11398e-23], [50.5991, 4.93895e-24, 7.76402e-24, 4.9864e-24, 4.6076e-24, 6.47104e-24, 1.11614e-23], [50.7142, 4.92875e-24, 7.76207e-24, 4.97897e-24, 4.59694e-24, 6.47198e-24, 1.11829e-23], [50.8296, 4.91867e-24, 7.76014e-24, 4.9716e-24, 4.58636e-24, 6.47292e-24, 1.12043e-23], [50.9452, 4.90872e-24, 7.75823e-24, 4.96428e-24, 4.57588e-24, 6.47386e-24, 1.12257e-23], [51.0611, 4.89889e-24, 7.75632e-24, 4.957e-24, 4.56547e-24, 6.47479e-24, 1.12469e-23], [51.1772, 4.88918e-24, 7.75444e-24, 4.94978e-24, 4.55515e-24, 6.47572e-24, 1.12681e-23], [51.2936, 4.87959e-24, 7.75256e-24, 4.94261e-24, 4.54492e-24, 6.47665e-24, 1.12892e-23], [51.4103, 4.87013e-24, 7.7507e-24, 4.93549e-24, 4.53477e-24, 6.47758e-24, 1.13103e-23], [51.5272, 4.86079e-24, 7.74886e-24, 4.92842e-24, 4.52471e-24, 6.4785e-24, 1.13312e-23], [51.6444, 4.85156e-24, 7.74703e-24, 4.92139e-24, 4.51472e-24, 6.47943e-24, 1.13521e-23], [51.7619, 4.84246e-24, 7.74521e-24, 4.91442e-24, 4.50482e-24, 6.48034e-24, 1.13728e-23], [51.8796, 4.83347e-24, 7.7434e-24, 4.90749e-24, 4.495e-24, 6.48126e-24, 1.13935e-23], [51.9976, 4.8246e-24, 7.74161e-24, 4.90061e-24, 4.48526e-24, 6.48217e-24, 1.14142e-23], [52.1159, 4.81585e-24, 7.73983e-24, 4.89378e-24, 4.4756e-24, 6.48307e-24, 1.14347e-23], [52.2345, 4.80722e-24, 7.73807e-24, 4.887e-24, 4.46602e-24, 6.48398e-24, 1.14552e-23], [52.3533, 4.7987e-24, 7.73632e-24, 4.88026e-24, 4.45653e-24, 6.48488e-24, 1.14756e-23], [52.4724, 4.79029e-24, 7.73458e-24, 4.87357e-24, 4.44711e-24, 6.48577e-24, 1.14959e-23], [52.5917, 4.782e-24, 7.73285e-24, 4.86692e-24, 4.43777e-24, 6.48666e-24, 1.15161e-23], [52.7113, 4.77383e-24, 7.73114e-24, 4.86032e-24, 4.4285e-24, 6.48754e-24, 1.15362e-23], [52.8312, 4.76577e-24, 7.72943e-24, 4.85377e-24, 4.41932e-24, 6.48842e-24, 1.15563e-23], [52.9514, 4.75782e-24, 7.72775e-24, 4.84726e-24, 4.41021e-24, 6.48929e-24, 1.15763e-23], [53.0718, 4.74998e-24, 7.72607e-24, 4.8408e-24, 4.40118e-24, 6.49016e-24, 1.15962e-23], [53.1925, 4.74225e-24, 7.7244e-24, 4.83438e-24, 4.39222e-24, 6.49102e-24, 1.1616e-23], [53.3135, 4.73463e-24, 7.72275e-24, 4.828e-24, 4.38334e-24, 6.49188e-24, 1.16358e-23], [53.4348, 4.72713e-24, 7.72111e-24, 4.82167e-24, 4.37453e-24, 6.49273e-24, 1.16554e-23], [53.5563, 4.71973e-24, 7.71948e-24, 4.81538e-24, 4.3658e-24, 6.49357e-24, 1.1675e-23], [53.6782, 4.71244e-24, 7.71787e-24, 4.80914e-24, 4.35714e-24, 6.49441e-24, 1.16945e-23], [53.8003, 4.70525e-24, 7.71626e-24, 4.80293e-24, 4.34856e-24, 6.49524e-24, 1.1714e-23], [53.9226, 4.69818e-24, 7.71467e-24, 4.79677e-24, 4.34004e-24, 6.49606e-24, 1.17333e-23], [54.0453, 4.69121e-24, 7.71308e-24, 4.79066e-24, 4.3316e-24, 6.49688e-24, 1.17526e-23], [54.1682, 4.68435e-24, 7.71151e-24, 4.78458e-24, 4.32324e-24, 6.49768e-24, 1.17718e-23], [54.2914, 4.67759e-24, 7.70995e-24, 4.77855e-24, 4.31494e-24, 6.49849e-24, 1.17909e-23], [54.4149, 4.67093e-24, 7.7084e-24, 4.77255e-24, 4.30671e-24, 6.49928e-24, 1.181e-23], [54.5387, 4.66438e-24, 7.70686e-24, 4.7666e-24, 4.29856e-24, 6.50007e-24, 1.18289e-23], [54.6627, 4.65794e-24, 7.70534e-24, 4.76069e-24, 4.29047e-24, 6.50084e-24, 1.18478e-23], [54.7871, 4.65159e-24, 7.70382e-24, 4.75482e-24, 4.28245e-24, 6.50161e-24, 1.18666e-23], [54.9117, 4.64535e-24, 7.70231e-24, 4.74899e-24, 4.27451e-24, 6.50238e-24, 1.18854e-23], [55.0366, 4.63921e-24, 7.70082e-24, 4.7432e-24, 4.26663e-24, 6.50313e-24, 1.1904e-23], [55.1618, 4.63316e-24, 7.69933e-24, 4.73745e-24, 4.25882e-24, 6.50387e-24, 1.19226e-23], [55.2872, 4.62722e-24, 7.69786e-24, 4.73174e-24, 4.25107e-24, 6.50461e-24, 1.19411e-23], [55.413, 4.62138e-24, 7.69639e-24, 4.72606e-24, 4.24339e-24, 6.50534e-24, 1.19595e-23], [55.539, 4.61564e-24, 7.69494e-24, 4.72043e-24, 4.23578e-24, 6.50606e-24, 1.19779e-23], [55.6654, 4.60999e-24, 7.6935e-24, 4.71483e-24, 4.22824e-24, 6.50677e-24, 1.19962e-23], [55.792, 4.60444e-24, 7.69206e-24, 4.70927e-24, 4.22076e-24, 6.50747e-24, 1.20144e-23], [55.9189, 4.59899e-24, 7.69064e-24, 4.70375e-24, 4.21334e-24, 6.50816e-24, 1.20325e-23], [56.0461, 4.59363e-24, 7.68922e-24, 4.69827e-24, 4.20599e-24, 6.50884e-24, 1.20505e-23], [56.1736, 4.58837e-24, 7.68782e-24, 4.69282e-24, 4.1987e-24, 6.50951e-24, 1.20685e-23], [56.3013, 4.5832e-24, 7.68642e-24, 4.68742e-24, 4.19148e-24, 6.51017e-24, 1.20864e-23], [56.4294, 4.57813e-24, 7.68504e-24, 4.68204e-24, 4.18432e-24, 6.51082e-24, 1.21043e-23], [56.5577, 4.57315e-24, 7.68366e-24, 4.67671e-24, 4.17722e-24, 6.51146e-24, 1.2122e-23], [56.6864, 4.56827e-24, 7.68229e-24, 4.67141e-24, 4.17018e-24, 6.51209e-24, 1.21397e-23], [56.8153, 4.56348e-24, 7.68094e-24, 4.66614e-24, 4.16321e-24, 6.51271e-24, 1.21573e-23], [56.9445, 4.55877e-24, 7.67959e-24, 4.66091e-24, 4.1563e-24, 6.51332e-24, 1.21748e-23], [57.0741, 4.55416e-24, 7.67825e-24, 4.65572e-24, 4.14944e-24, 6.51392e-24, 1.21923e-23], [57.2039, 4.54964e-24, 7.67692e-24, 4.65056e-24, 4.14265e-24, 6.51451e-24, 1.22096e-23], [57.334, 4.54521e-24, 7.6756e-24, 4.64544e-24, 4.13592e-24, 6.51509e-24, 1.2227e-23], [57.4644, 4.54087e-24, 7.67428e-24, 4.64035e-24, 4.12924e-24, 6.51565e-24, 1.22442e-23], [57.5951, 4.53662e-24, 7.67298e-24, 4.63529e-24, 4.12263e-24, 6.51621e-24, 1.22614e-23], [57.7261, 4.53245e-24, 7.67168e-24, 4.63027e-24, 4.11607e-24, 6.51675e-24, 1.22784e-23], [57.8574, 4.52838e-24, 7.6704e-24, 4.62528e-24, 4.10957e-24, 6.51728e-24, 1.22955e-23], [57.989, 4.52438e-24, 7.66912e-24, 4.62033e-24, 4.10313e-24, 6.5178e-24, 1.23124e-23], [58.1209, 4.52048e-24, 7.66785e-24, 4.61541e-24, 4.09675e-24, 6.51831e-24, 1.23293e-23], [58.2531, 4.51666e-24, 7.66659e-24, 4.61052e-24, 4.09042e-24, 6.51881e-24, 1.23461e-23], [58.3856, 4.51293e-24, 7.66533e-24, 4.60567e-24, 4.08415e-24, 6.51929e-24, 1.23628e-23], [58.5184, 4.50928e-24, 7.66409e-24, 4.60084e-24, 4.07793e-24, 6.51977e-24, 1.23795e-23], [58.6515, 4.50571e-24, 7.66285e-24, 4.59605e-24, 4.07177e-24, 6.52023e-24, 1.2396e-23], [58.7849, 4.50223e-24, 7.66162e-24, 4.59129e-24, 4.06566e-24, 6.52068e-24, 1.24126e-23], [58.9187, 4.49883e-24, 7.6604e-24, 4.58656e-24, 4.05961e-24, 6.52111e-24, 1.2429e-23], [59.0527, 4.49551e-24, 7.65919e-24, 4.58187e-24, 4.05361e-24, 6.52153e-24, 1.24454e-23], [59.187, 4.49228e-24, 7.65798e-24, 4.5772e-24, 4.04766e-24, 6.52195e-24, 1.24617e-23], [59.3216, 4.48912e-24, 7.65679e-24, 4.57257e-24, 4.04177e-24, 6.52234e-24, 1.24779e-23], [59.4565, 4.48605e-24, 7.6556e-24, 4.56796e-24, 4.03593e-24, 6.52273e-24, 1.24941e-23], [59.5918, 4.48305e-24, 7.65441e-24, 4.56339e-24, 4.03014e-24, 6.5231e-24, 1.25101e-23], [59.7273, 4.48014e-24, 7.65324e-24, 4.55885e-24, 4.02441e-24, 6.52346e-24, 1.25262e-23], [59.8632, 4.4773e-24, 7.65207e-24, 4.55433e-24, 4.01872e-24, 6.52381e-24, 1.25421e-23], [59.9993, 4.47454e-24, 7.65091e-24, 4.54985e-24, 4.01309e-24, 6.52414e-24, 1.2558e-23], [60.1358, 4.47186e-24, 7.64976e-24, 4.5454e-24, 4.0075e-24, 6.52446e-24, 1.25738e-23], [60.2726, 4.46926e-24, 7.64862e-24, 4.54097e-24, 4.00197e-24, 6.52477e-24, 1.25895e-23], [60.4097, 4.46673e-24, 7.64748e-24, 4.53658e-24, 3.99648e-24, 6.52506e-24, 1.26052e-23], [60.5471, 4.46428e-24, 7.64635e-24, 4.53221e-24, 3.99105e-24, 6.52534e-24, 1.26208e-23], [60.6848, 4.46191e-24, 7.64522e-24, 4.52787e-24, 3.98566e-24, 6.5256e-24, 1.26364e-23], [60.8229, 4.45961e-24, 7.64411e-24, 4.52356e-24, 3.98032e-24, 6.52586e-24, 1.26518e-23], [60.9612, 4.45738e-24, 7.643e-24, 4.51928e-24, 3.97503e-24, 6.52609e-24, 1.26672e-23], [61.0999, 4.45523e-24, 7.6419e-24, 4.51502e-24, 3.96979e-24, 6.52632e-24, 1.26826e-23], [61.2388, 4.45315e-24, 7.6408e-24, 4.5108e-24, 3.9646e-24, 6.52653e-24, 1.26978e-23], [61.3781, 4.45114e-24, 7.63971e-24, 4.5066e-24, 3.95945e-24, 6.52672e-24, 1.2713e-23], [61.5177, 4.44921e-24, 7.63863e-24, 4.50243e-24, 3.95434e-24, 6.5269e-24, 1.27281e-23], [61.6577, 4.44735e-24, 7.63755e-24, 4.49828e-24, 3.94929e-24, 6.52707e-24, 1.27432e-23], [61.7979, 4.44556e-24, 7.63648e-24, 4.49416e-24, 3.94427e-24, 6.52722e-24, 1.27582e-23], [61.9385, 4.44384e-24, 7.63542e-24, 4.49007e-24, 3.93931e-24, 6.52736e-24, 1.27731e-23], [62.0794, 4.44219e-24, 7.63437e-24, 4.48601e-24, 3.93439e-24, 6.52748e-24, 1.2788e-23], [62.2206, 4.44061e-24, 7.63332e-24, 4.48197e-24, 3.92951e-24, 6.52759e-24, 1.28028e-23], [62.3621, 4.4391e-24, 7.63227e-24, 4.47796e-24, 3.92467e-24, 6.52768e-24, 1.28175e-23], [62.5039, 4.43766e-24, 7.63124e-24, 4.47397e-24, 3.91988e-24, 6.52776e-24, 1.28322e-23], [62.6461, 4.43629e-24, 7.63021e-24, 4.47001e-24, 3.91514e-24, 6.52783e-24, 1.28468e-23], [62.7886, 4.43498e-24, 7.62918e-24, 4.46607e-24, 3.91043e-24, 6.52788e-24, 1.28613e-23], [62.9314, 4.43375e-24, 7.62816e-24, 4.46216e-24, 3.90577e-24, 6.52791e-24, 1.28758e-23], [63.0746, 4.43258e-24, 7.62715e-24, 4.45828e-24, 3.90115e-24, 6.52793e-24, 1.28902e-23], [63.218, 4.43147e-24, 7.62615e-24, 4.45441e-24, 3.89657e-24, 6.52793e-24, 1.29045e-23], [63.3618, 4.43043e-24, 7.62515e-24, 4.45058e-24, 3.89203e-24, 6.52792e-24, 1.29188e-23], [63.5059, 4.42946e-24, 7.62415e-24, 4.44676e-24, 3.88754e-24, 6.52789e-24, 1.2933e-23], [63.6504, 4.42855e-24, 7.62316e-24, 4.44298e-24, 3.88308e-24, 6.52785e-24, 1.29472e-23], [63.7952, 4.42771e-24, 7.62218e-24, 4.43921e-24, 3.87866e-24, 6.52779e-24, 1.29613e-23], [63.9403, 4.42693e-24, 7.6212e-24, 4.43547e-24, 3.87429e-24, 6.52772e-24, 1.29753e-23], [64.0857, 4.42622e-24, 7.62023e-24, 4.43175e-24, 3.86995e-24, 6.52763e-24, 1.29892e-23], [64.2315, 4.42557e-24, 7.61927e-24, 4.42806e-24, 3.86565e-24, 6.52752e-24, 1.30031e-23], [64.3776, 4.42498e-24, 7.61831e-24, 4.42439e-24, 3.86139e-24, 6.5274e-24, 1.3017e-23], [64.524, 4.42445e-24, 7.61735e-24, 4.42074e-24, 3.85717e-24, 6.52726e-24, 1.30307e-23], [64.6708, 4.42399e-24, 7.61641e-24, 4.41712e-24, 3.85299e-24, 6.52711e-24, 1.30444e-23], [64.8179, 4.42358e-24, 7.61546e-24, 4.41352e-24, 3.84884e-24, 6.52694e-24, 1.30581e-23], [64.9653, 4.42324e-24, 7.61453e-24, 4.40994e-24, 3.84474e-24, 6.52675e-24, 1.30717e-23], [65.1131, 4.42296e-24, 7.61359e-24, 4.40638e-24, 3.84067e-24, 6.52655e-24, 1.30852e-23], [65.2612, 4.42274e-24, 7.61267e-24, 4.40285e-24, 3.83663e-24, 6.52633e-24, 1.30986e-23], [65.4096, 4.42257e-24, 7.61175e-24, 4.39933e-24, 3.83263e-24, 6.5261e-24, 1.3112e-23], [65.5584, 4.42247e-24, 7.61083e-24, 4.39584e-24, 3.82867e-24, 6.52585e-24, 1.31254e-23], [65.7075, 4.42243e-24, 7.60992e-24, 4.39237e-24, 3.82474e-24, 6.52558e-24, 1.31387e-23], [65.857, 4.42244e-24, 7.60901e-24, 4.38892e-24, 3.82085e-24, 6.5253e-24, 1.31519e-23], [66.0068, 4.42252e-24, 7.60811e-24, 4.3855e-24, 3.81699e-24, 6.52499e-24, 1.3165e-23], [66.1569, 4.42265e-24, 7.60722e-24, 4.38209e-24, 3.81317e-24, 6.52468e-24, 1.31781e-23], [66.3074, 4.42284e-24, 7.60633e-24, 4.37871e-24, 3.80938e-24, 6.52434e-24, 1.31911e-23], [66.4582, 4.42308e-24, 7.60544e-24, 4.37534e-24, 3.80563e-24, 6.52399e-24, 1.32041e-23], [66.6094, 4.42338e-24, 7.60456e-24, 4.372e-24, 3.8019e-24, 6.52363e-24, 1.3217e-23], [66.7609, 4.42374e-24, 7.60369e-24, 4.36867e-24, 3.79822e-24, 6.52324e-24, 1.32299e-23], [66.9127, 4.42415e-24, 7.60282e-24, 4.36537e-24, 3.79456e-24, 6.52284e-24, 1.32427e-23], [67.0649, 4.42462e-24, 7.60195e-24, 4.36209e-24, 3.79094e-24, 6.52242e-24, 1.32554e-23], [67.2175, 4.42515e-24, 7.60109e-24, 4.35883e-24, 3.78735e-24, 6.52198e-24, 1.32681e-23], [67.3703, 4.42572e-24, 7.60023e-24, 4.35558e-24, 3.78379e-24, 6.52153e-24, 1.32807e-23], [67.5236, 4.42636e-24, 7.59938e-24, 4.35236e-24, 3.78026e-24, 6.52106e-24, 1.32933e-23], [67.6772, 4.42704e-24, 7.59854e-24, 4.34915e-24, 3.77677e-24, 6.52058e-24, 1.33058e-23], [67.8311, 4.42778e-24, 7.5977e-24, 4.34597e-24, 3.7733e-24, 6.52007e-24, 1.33182e-23], [67.9854, 4.42858e-24, 7.59686e-24, 4.3428e-24, 3.76987e-24, 6.51955e-24, 1.33306e-23], [68.14, 4.42942e-24, 7.59603e-24, 4.33966e-24, 3.76646e-24, 6.51901e-24, 1.33429e-23], [68.295, 4.43032e-24, 7.5952e-24, 4.33653e-24, 3.76309e-24, 6.51845e-24, 1.33552e-23], [68.4504, 4.43127e-24, 7.59437e-24, 4.33342e-24, 3.75974e-24, 6.51788e-24, 1.33674e-23], [68.6061, 4.43227e-24, 7.59356e-24, 4.33033e-24, 3.75643e-24, 6.51729e-24, 1.33795e-23], [68.7621, 4.43333e-24, 7.59274e-24, 4.32725e-24, 3.75314e-24, 6.51668e-24, 1.33916e-23], [68.9185, 4.43443e-24, 7.59193e-24, 4.3242e-24, 3.74989e-24, 6.51605e-24, 1.34037e-23], [69.0753, 4.43559e-24, 7.59113e-24, 4.32116e-24, 3.74666e-24, 6.5154e-24, 1.34157e-23], [69.2324, 4.43679e-24, 7.59032e-24, 4.31814e-24, 3.74346e-24, 6.51474e-24, 1.34276e-23], [69.3899, 4.43805e-24, 7.58953e-24, 4.31514e-24, 3.74029e-24, 6.51406e-24, 1.34394e-23], [69.5477, 4.43935e-24, 7.58873e-24, 4.31216e-24, 3.73714e-24, 6.51336e-24, 1.34513e-23], [69.7059, 4.44071e-24, 7.58795e-24, 4.30919e-24, 3.73402e-24, 6.51265e-24, 1.3463e-23], [69.8644, 4.44211e-24, 7.58716e-24, 4.30625e-24, 3.73093e-24, 6.51191e-24, 1.34747e-23], [70.0234, 4.44357e-24, 7.58638e-24, 4.30331e-24, 3.72787e-24, 6.51116e-24, 1.34864e-23], [70.1826, 4.44507e-24, 7.58561e-24, 4.3004e-24, 3.72483e-24, 6.51039e-24, 1.3498e-23], [70.3423, 4.44662e-24, 7.58483e-24, 4.2975e-24, 3.72182e-24, 6.5096e-24, 1.35095e-23], [70.5023, 4.44821e-24, 7.58407e-24, 4.29462e-24, 3.71884e-24, 6.50879e-24, 1.3521e-23], [70.6626, 4.44986e-24, 7.5833e-24, 4.29176e-24, 3.71588e-24, 6.50797e-24, 1.35324e-23], [70.8233, 4.45155e-24, 7.58254e-24, 4.28891e-24, 3.71295e-24, 6.50712e-24, 1.35438e-23], [70.9844, 4.45329e-24, 7.58179e-24, 4.28608e-24, 3.71004e-24, 6.50626e-24, 1.35551e-23], [71.1459, 4.45507e-24, 7.58104e-24, 4.28327e-24, 3.70716e-24, 6.50538e-24, 1.35664e-23], [71.3077, 4.45691e-24, 7.58029e-24, 4.28047e-24, 3.7043e-24, 6.50448e-24, 1.35776e-23], [71.4699, 4.45878e-24, 7.57955e-24, 4.27768e-24, 3.70146e-24, 6.50356e-24, 1.35887e-23], [71.6325, 4.46071e-24, 7.57881e-24, 4.27492e-24, 3.69865e-24, 6.50263e-24, 1.35998e-23], [71.7954, 4.46268e-24, 7.57807e-24, 4.27217e-24, 3.69587e-24, 6.50167e-24, 1.36109e-23], [71.9587, 4.46469e-24, 7.57734e-24, 4.26943e-24, 3.6931e-24, 6.5007e-24, 1.36219e-23], [72.1224, 4.46675e-24, 7.57661e-24, 4.26671e-24, 3.69036e-24, 6.49971e-24, 1.36328e-23], [72.2864, 4.46886e-24, 7.57589e-24, 4.26401e-24, 3.68765e-24, 6.4987e-24, 1.36437e-23], [72.4509, 4.471e-24, 7.57517e-24, 4.26132e-24, 3.68495e-24, 6.49767e-24, 1.36546e-23], [72.6157, 4.4732e-24, 7.57445e-24, 4.25864e-24, 3.68228e-24, 6.49662e-24, 1.36654e-23], [72.7808, 4.47543e-24, 7.57374e-24, 4.25598e-24, 3.67963e-24, 6.49555e-24, 1.36761e-23], [72.9464, 4.47771e-24, 7.57303e-24, 4.25334e-24, 3.67701e-24, 6.49447e-24, 1.36868e-23], [73.1123, 4.48004e-24, 7.57232e-24, 4.25071e-24, 3.6744e-24, 6.49336e-24, 1.36974e-23], [73.2786, 4.4824e-24, 7.57162e-24, 4.24809e-24, 3.67182e-24, 6.49224e-24, 1.3708e-23], [73.4453, 4.48481e-24, 7.57092e-24, 4.24549e-24, 3.66926e-24, 6.4911e-24, 1.37185e-23], [73.6123, 4.48727e-24, 7.57023e-24, 4.24291e-24, 3.66672e-24, 6.48994e-24, 1.3729e-23], [73.7798, 4.48976e-24, 7.56954e-24, 4.24033e-24, 3.6642e-24, 6.48875e-24, 1.37394e-23], [73.9476, 4.4923e-24, 7.56885e-24, 4.23778e-24, 3.6617e-24, 6.48755e-24, 1.37498e-23], [74.1158, 4.49487e-24, 7.56816e-24, 4.23523e-24, 3.65922e-24, 6.48634e-24, 1.37601e-23], [74.2844, 4.49749e-24, 7.56748e-24, 4.2327e-24, 3.65676e-24, 6.4851e-24, 1.37704e-23], [74.4533, 4.50016e-24, 7.56681e-24, 4.23019e-24, 3.65432e-24, 6.48384e-24, 1.37806e-23], [74.6227, 4.50286e-24, 7.56613e-24, 4.22768e-24, 3.6519e-24, 6.48256e-24, 1.37908e-23], [74.7924, 4.5056e-24, 7.56546e-24, 4.22519e-24, 3.6495e-24, 6.48127e-24, 1.38009e-23], [74.9625, 4.50839e-24, 7.56479e-24, 4.22272e-24, 3.64712e-24, 6.47995e-24, 1.3811e-23], [75.133, 4.51121e-24, 7.56413e-24, 4.22026e-24, 3.64476e-24, 6.47861e-24, 1.3821e-23], [75.3039, 4.51408e-24, 7.56347e-24, 4.21781e-24, 3.64242e-24, 6.47726e-24, 1.3831e-23], [75.4752, 4.51698e-24, 7.56281e-24, 4.21537e-24, 3.64009e-24, 6.47589e-24, 1.38409e-23], [75.6469, 4.51993e-24, 7.56216e-24, 4.21295e-24, 3.63779e-24, 6.47449e-24, 1.38508e-23], [75.819, 4.52291e-24, 7.56151e-24, 4.21054e-24, 3.6355e-24, 6.47308e-24, 1.38606e-23], [75.9914, 4.52594e-24, 7.56086e-24, 4.20814e-24, 3.63323e-24, 6.47165e-24, 1.38704e-23], [76.1643, 4.529e-24, 7.56022e-24, 4.20576e-24, 3.63098e-24, 6.47019e-24, 1.38802e-23], [76.3375, 4.53211e-24, 7.55958e-24, 4.20339e-24, 3.62874e-24, 6.46872e-24, 1.38898e-23], [76.5111, 4.53525e-24, 7.55894e-24, 4.20103e-24, 3.62653e-24, 6.46723e-24, 1.38995e-23], [76.6852, 4.53843e-24, 7.55831e-24, 4.19868e-24, 3.62433e-24, 6.46572e-24, 1.39091e-23], [76.8596, 4.54165e-24, 7.55768e-24, 4.19635e-24, 3.62214e-24, 6.46418e-24, 1.39186e-23], [77.0344, 4.5449e-24, 7.55705e-24, 4.19402e-24, 3.61998e-24, 6.46263e-24, 1.39281e-23], [77.2096, 4.5482e-24, 7.55643e-24, 4.19171e-24, 3.61783e-24, 6.46106e-24, 1.39375e-23], [77.3852, 4.55153e-24, 7.55581e-24, 4.18942e-24, 3.61569e-24, 6.45947e-24, 1.39469e-23], [77.5613, 4.5549e-24, 7.55519e-24, 4.18713e-24, 3.61358e-24, 6.45786e-24, 1.39563e-23], [77.7377, 4.55831e-24, 7.55457e-24, 4.18486e-24, 3.61148e-24, 6.45623e-24, 1.39656e-23], [77.9145, 4.56176e-24, 7.55396e-24, 4.18259e-24, 3.60939e-24, 6.45458e-24, 1.39748e-23], [78.0917, 4.56524e-24, 7.55335e-24, 4.18034e-24, 3.60732e-24, 6.4529e-24, 1.39841e-23], [78.2694, 4.56876e-24, 7.55275e-24, 4.1781e-24, 3.60526e-24, 6.45121e-24, 1.39932e-23], [78.4474, 4.57232e-24, 7.55215e-24, 4.17588e-24, 3.60322e-24, 6.4495e-24, 1.40023e-23], [78.6258, 4.57591e-24, 7.55155e-24, 4.17366e-24, 3.6012e-24, 6.44777e-24, 1.40114e-23], [78.8047, 4.57954e-24, 7.55095e-24, 4.17145e-24, 3.59919e-24, 6.44602e-24, 1.40204e-23], [78.9839, 4.58321e-24, 7.55036e-24, 4.16926e-24, 3.59719e-24, 6.44424e-24, 1.40294e-23], [79.1636, 4.58691e-24, 7.54976e-24, 4.16708e-24, 3.59521e-24, 6.44245e-24, 1.40383e-23], [79.3436, 4.59065e-24, 7.54918e-24, 4.16491e-24, 3.59325e-24, 6.44064e-24, 1.40472e-23], [79.5241, 4.59443e-24, 7.54859e-24, 4.16275e-24, 3.59129e-24, 6.43881e-24, 1.40561e-23], [79.705, 4.59824e-24, 7.54801e-24, 4.1606e-24, 3.58935e-24, 6.43695e-24, 1.40649e-23], [79.8863, 4.60208e-24, 7.54743e-24, 4.15846e-24, 3.58743e-24, 6.43508e-24, 1.40736e-23], [80.068, 4.60596e-24, 7.54686e-24, 4.15633e-24, 3.58552e-24, 6.43318e-24, 1.40823e-23], [80.2501, 4.60988e-24, 7.54628e-24, 4.15421e-24, 3.58362e-24, 6.43127e-24, 1.4091e-23], [80.4326, 4.61383e-24, 7.54571e-24, 4.1521e-24, 3.58173e-24, 6.42933e-24, 1.40996e-23], [80.6156, 4.61782e-24, 7.54515e-24, 4.15e-24, 3.57986e-24, 6.42738e-24, 1.41082e-23], [80.7989, 4.62184e-24, 7.54458e-24, 4.14792e-24, 3.578e-24, 6.4254e-24, 1.41167e-23], [80.9827, 4.6259e-24, 7.54402e-24, 4.14584e-24, 3.57615e-24, 6.4234e-24, 1.41252e-23], [81.1669, 4.62999e-24, 7.54346e-24, 4.14377e-24, 3.57432e-24, 6.42138e-24, 1.41337e-23], [81.3516, 4.63412e-24, 7.54291e-24, 4.14172e-24, 3.5725e-24, 6.41934e-24, 1.41421e-23], [81.5366, 4.63828e-24, 7.54235e-24, 4.13967e-24, 3.57069e-24, 6.41728e-24, 1.41504e-23], [81.7221, 4.64247e-24, 7.5418e-24, 4.13764e-24, 3.56889e-24, 6.4152e-24, 1.41587e-23], [81.9079, 4.6467e-24, 7.54125e-24, 4.13561e-24, 3.5671e-24, 6.4131e-24, 1.4167e-23], [82.0942, 4.65096e-24, 7.54071e-24, 4.13359e-24, 3.56533e-24, 6.41098e-24, 1.41752e-23], [82.281, 4.65526e-24, 7.54017e-24, 4.13159e-24, 3.56356e-24, 6.40884e-24, 1.41834e-23], [82.4681, 4.65959e-24, 7.53963e-24, 4.12959e-24, 3.56181e-24, 6.40667e-24, 1.41916e-23], [82.6557, 4.66395e-24, 7.53909e-24, 4.1276e-24, 3.56007e-24, 6.40449e-24, 1.41997e-23], [82.8437, 4.66835e-24, 7.53856e-24, 4.12562e-24, 3.55834e-24, 6.40228e-24, 1.42077e-23], [83.0321, 4.67278e-24, 7.53803e-24, 4.12365e-24, 3.55662e-24, 6.40006e-24, 1.42157e-23], [83.221, 4.67724e-24, 7.5375e-24, 4.12169e-24, 3.55491e-24, 6.39781e-24, 1.42237e-23], [83.4103, 4.68174e-24, 7.53697e-24, 4.11974e-24, 3.55322e-24, 6.39554e-24, 1.42316e-23], [83.6, 4.68627e-24, 7.53645e-24, 4.1178e-24, 3.55153e-24, 6.39325e-24, 1.42395e-23], [83.7902, 4.69083e-24, 7.53593e-24, 4.11587e-24, 3.54985e-24, 6.39094e-24, 1.42474e-23], [83.9808, 4.69543e-24, 7.53541e-24, 4.11395e-24, 3.54818e-24, 6.38861e-24, 1.42552e-23], [84.1718, 4.70005e-24, 7.5349e-24, 4.11203e-24, 3.54653e-24, 6.38626e-24, 1.42629e-23], [84.3632, 4.70471e-24, 7.53438e-24, 4.11013e-24, 3.54488e-24, 6.38388e-24, 1.42707e-23], [84.5551, 4.70941e-24, 7.53387e-24, 4.10823e-24, 3.54324e-24, 6.38149e-24, 1.42783e-23], [84.7475, 4.71413e-24, 7.53337e-24, 4.10635e-24, 3.54162e-24, 6.37907e-24, 1.4286e-23], [84.9402, 4.71889e-24, 7.53286e-24, 4.10447e-24, 3.54e-24, 6.37663e-24, 1.42936e-23], [85.1334, 4.72368e-24, 7.53236e-24, 4.1026e-24, 3.53839e-24, 6.37417e-24, 1.43011e-23], [85.3271, 4.7285e-24, 7.53186e-24, 4.10073e-24, 3.53679e-24, 6.37169e-24, 1.43087e-23], [85.5211, 4.73336e-24, 7.53136e-24, 4.09888e-24, 3.5352e-24, 6.36919e-24, 1.43161e-23], [85.7157, 4.73825e-24, 7.53087e-24, 4.09704e-24, 3.53362e-24, 6.36666e-24, 1.43236e-23], [85.9106, 4.74316e-24, 7.53038e-24, 4.0952e-24, 3.53204e-24, 6.36412e-24, 1.4331e-23], [86.106, 4.74811e-24, 7.52989e-24, 4.09337e-24, 3.53048e-24, 6.36155e-24, 1.43383e-23], [86.3019, 4.7531e-24, 7.5294e-24, 4.09155e-24, 3.52892e-24, 6.35896e-24, 1.43457e-23], [86.4982, 4.75811e-24, 7.52892e-24, 4.08974e-24, 3.52738e-24, 6.35636e-24, 1.43529e-23], [86.6949, 4.76315e-24, 7.52844e-24, 4.08794e-24, 3.52584e-24, 6.35372e-24, 1.43602e-23], [86.8921, 4.76823e-24, 7.52796e-24, 4.08614e-24, 3.52431e-24, 6.35107e-24, 1.43674e-23], [87.0898, 4.77334e-24, 7.52748e-24, 4.08436e-24, 3.52278e-24, 6.3484e-24, 1.43745e-23], [87.2879, 4.77848e-24, 7.52701e-24, 4.08258e-24, 3.52127e-24, 6.3457e-24, 1.43817e-23], [87.4864, 4.78365e-24, 7.52654e-24, 4.08081e-24, 3.51976e-24, 6.34298e-24, 1.43888e-23], [87.6854, 4.78885e-24, 7.52607e-24, 4.07904e-24, 3.51826e-24, 6.34024e-24, 1.43958e-23], [87.8849, 4.79408e-24, 7.5256e-24, 4.07729e-24, 3.51677e-24, 6.33748e-24, 1.44028e-23], [88.0848, 4.79935e-24, 7.52514e-24, 4.07554e-24, 3.51528e-24, 6.3347e-24, 1.44098e-23], [88.2851, 4.80464e-24, 7.52468e-24, 4.0738e-24, 3.5138e-24, 6.3319e-24, 1.44167e-23], [88.4859, 4.80997e-24, 7.52422e-24, 4.07207e-24, 3.51233e-24, 6.32907e-24, 1.44236e-23], [88.6872, 4.81532e-24, 7.52376e-24, 4.07034e-24, 3.51087e-24, 6.32622e-24, 1.44304e-23], [88.8889, 4.82071e-24, 7.52331e-24, 4.06862e-24, 3.50941e-24, 6.32335e-24, 1.44373e-23], [89.0911, 4.82613e-24, 7.52286e-24, 4.06691e-24, 3.50796e-24, 6.32046e-24, 1.4444e-23], [89.2937, 4.83158e-24, 7.52241e-24, 4.06521e-24, 3.50652e-24, 6.31754e-24, 1.44508e-23], [89.4968, 4.83706e-24, 7.52196e-24, 4.06351e-24, 3.50508e-24, 6.31461e-24, 1.44575e-23], [89.7004, 4.84257e-24, 7.52152e-24, 4.06183e-24, 3.50365e-24, 6.31165e-24, 1.44641e-23], [89.9044, 4.84811e-24, 7.52107e-24, 4.06014e-24, 3.50223e-24, 6.30867e-24, 1.44708e-23], [90.1089, 4.85368e-24, 7.52063e-24, 4.05847e-24, 3.50081e-24, 6.30567e-24, 1.44774e-23], [90.3139, 4.85928e-24, 7.5202e-24, 4.0568e-24, 3.4994e-24, 6.30264e-24, 1.44839e-23], [90.5193, 4.86491e-24, 7.51976e-24, 4.05514e-24, 3.49799e-24, 6.2996e-24, 1.44904e-23], [90.7252, 4.87057e-24, 7.51933e-24, 4.05349e-24, 3.49659e-24, 6.29653e-24, 1.44969e-23], [90.9316, 4.87626e-24, 7.5189e-24, 4.05185e-24, 3.4952e-24, 6.29344e-24, 1.45034e-23], [91.1384, 4.88198e-24, 7.51847e-24, 4.05021e-24, 3.49381e-24, 6.29033e-24, 1.45098e-23], [91.3457, 4.88774e-24, 7.51805e-24, 4.04857e-24, 3.49243e-24, 6.28719e-24, 1.45161e-23], [91.5535, 4.89352e-24, 7.51763e-24, 4.04695e-24, 3.49105e-24, 6.28404e-24, 1.45225e-23], [91.7617, 4.89933e-24, 7.51721e-24, 4.04533e-24, 3.48968e-24, 6.28086e-24, 1.45288e-23], [91.9704, 4.90517e-24, 7.51679e-24, 4.04372e-24, 3.48831e-24, 6.27766e-24, 1.4535e-23], [92.1796, 4.91104e-24, 7.51637e-24, 4.04211e-24, 3.48695e-24, 6.27443e-24, 1.45412e-23], [92.3893, 4.91695e-24, 7.51596e-24, 4.04052e-24, 3.48559e-24, 6.27119e-24, 1.45474e-23], [92.5994, 4.92288e-24, 7.51555e-24, 4.03892e-24, 3.48424e-24, 6.26792e-24, 1.45536e-23], [92.8101, 4.92884e-24, 7.51514e-24, 4.03734e-24, 3.48289e-24, 6.26463e-24, 1.45597e-23], [93.0212, 4.93483e-24, 7.51474e-24, 4.03576e-24, 3.48155e-24, 6.26131e-24, 1.45658e-23], [93.2328, 4.94085e-24, 7.51433e-24, 4.03419e-24, 3.48022e-24, 6.25798e-24, 1.45718e-23], [93.4448, 4.9469e-24, 7.51393e-24, 4.03262e-24, 3.47888e-24, 6.25462e-24, 1.45779e-23], [93.6574, 4.95299e-24, 7.51353e-24, 4.03106e-24, 3.47756e-24, 6.25124e-24, 1.45838e-23], [93.8704, 4.9591e-24, 7.51314e-24, 4.02951e-24, 3.47623e-24, 6.24784e-24, 1.45898e-23], [94.0839, 4.96524e-24, 7.51274e-24, 4.02796e-24, 3.47491e-24, 6.24441e-24, 1.45957e-23], [94.2979, 4.97141e-24, 7.51235e-24, 4.02642e-24, 3.4736e-24, 6.24096e-24, 1.46016e-23], [94.5124, 4.9776e-24, 7.51196e-24, 4.02489e-24, 3.47229e-24, 6.23749e-24, 1.46074e-23], [94.7274, 4.98383e-24, 7.51157e-24, 4.02336e-24, 3.47098e-24, 6.234e-24, 1.46132e-23], [94.9428, 4.99009e-24, 7.51119e-24, 4.02184e-24, 3.46968e-24, 6.23049e-24, 1.4619e-23], [95.1588, 4.99638e-24, 7.51081e-24, 4.02032e-24, 3.46838e-24, 6.22695e-24, 1.46247e-23], [95.3752, 5.0027e-24, 7.51043e-24, 4.01881e-24, 3.46709e-24, 6.22339e-24, 1.46304e-23], [95.5922, 5.00904e-24, 7.51005e-24, 4.0173e-24, 3.4658e-24, 6.2198e-24, 1.46361e-23], [95.8096, 5.01542e-24, 7.50967e-24, 4.01581e-24, 3.46451e-24, 6.2162e-24, 1.46417e-23], [96.0275, 5.02183e-24, 7.5093e-24, 4.01431e-24, 3.46322e-24, 6.21257e-24, 1.46473e-23], [96.246, 5.02826e-24, 7.50893e-24, 4.01283e-24, 3.46194e-24, 6.20892e-24, 1.46529e-23], [96.4649, 5.03473e-24, 7.50856e-24, 4.01135e-24, 3.46067e-24, 6.20524e-24, 1.46584e-23], [96.6843, 5.04122e-24, 7.5082e-24, 4.00987e-24, 3.45939e-24, 6.20155e-24, 1.46639e-23], [96.9042, 5.04774e-24, 7.50783e-24, 4.0084e-24, 3.45812e-24, 6.19783e-24, 1.46694e-23], [97.1246, 5.0543e-24, 7.50747e-24, 4.00694e-24, 3.45686e-24, 6.19408e-24, 1.46748e-23], [97.3455, 5.06088e-24, 7.50711e-24, 4.00548e-24, 3.45559e-24, 6.19032e-24, 1.46802e-23], [97.567, 5.06749e-24, 7.50676e-24, 4.00403e-24, 3.45433e-24, 6.18653e-24, 1.46856e-23], [97.7889, 5.07413e-24, 7.5064e-24, 4.00258e-24, 3.45307e-24, 6.18272e-24, 1.46909e-23], [98.0113, 5.0808e-24, 7.50605e-24, 4.00114e-24, 3.45182e-24, 6.17889e-24, 1.46962e-23], [98.2342, 5.0875e-24, 7.5057e-24, 3.9997e-24, 3.45057e-24, 6.17503e-24, 1.47015e-23], [98.4577, 5.09423e-24, 7.50535e-24, 3.99827e-24, 3.44932e-24, 6.17115e-24, 1.47067e-23], [98.6816, 5.10098e-24, 7.50501e-24, 3.99685e-24, 3.44807e-24, 6.16725e-24, 1.47119e-23], [98.9061, 5.10777e-24, 7.50467e-24, 3.99543e-24, 3.44683e-24, 6.16332e-24, 1.47171e-23], [99.1311, 5.11459e-24, 7.50433e-24, 3.99402e-24, 3.44558e-24, 6.15937e-24, 1.47222e-23], [99.3565, 5.12143e-24, 7.50399e-24, 3.99261e-24, 3.44434e-24, 6.1554e-24, 1.47273e-23], [99.5825, 5.12831e-24, 7.50365e-24, 3.9912e-24, 3.44311e-24, 6.15141e-24, 1.47324e-23], [99.809, 5.13521e-24, 7.50332e-24, 3.98981e-24, 3.44187e-24, 6.14739e-24, 1.47374e-23], [100.036, 5.14214e-24, 7.50299e-24, 3.98841e-24, 3.44064e-24, 6.14335e-24, 1.47424e-23], [100.264, 5.1491e-24, 7.50266e-24, 3.98702e-24, 3.43941e-24, 6.13929e-24, 1.47474e-23], [100.492, 5.15609e-24, 7.50233e-24, 3.98564e-24, 3.43818e-24, 6.1352e-24, 1.47524e-23], [100.72, 5.16312e-24, 7.50201e-24, 3.98427e-24, 3.43696e-24, 6.13109e-24, 1.47573e-23], [100.949, 5.17016e-24, 7.50169e-24, 3.98289e-24, 3.43573e-24, 6.12696e-24, 1.47622e-23], [101.179, 5.17724e-24, 7.50137e-24, 3.98153e-24, 3.43451e-24, 6.1228e-24, 1.4767e-23], [101.409, 5.18435e-24, 7.50105e-24, 3.98016e-24, 3.43329e-24, 6.11863e-24, 1.47718e-23], [101.64, 5.19149e-24, 7.50074e-24, 3.97881e-24, 3.43207e-24, 6.11442e-24, 1.47766e-23], [101.871, 5.19866e-24, 7.50043e-24, 3.97746e-24, 3.43085e-24, 6.1102e-24, 1.47814e-23], [102.103, 5.20585e-24, 7.50012e-24, 3.97611e-24, 3.42964e-24, 6.10595e-24, 1.47861e-23], [102.335, 5.21308e-24, 7.49981e-24, 3.97477e-24, 3.42842e-24, 6.10168e-24, 1.47908e-23], [102.568, 5.22033e-24, 7.4995e-24, 3.97343e-24, 3.42721e-24, 6.09738e-24, 1.47955e-23], [102.801, 5.22761e-24, 7.4992e-24, 3.9721e-24, 3.426e-24, 6.09307e-24, 1.48001e-23], [103.035, 5.23493e-24, 7.4989e-24, 3.97077e-24, 3.42479e-24, 6.08873e-24, 1.48047e-23], [103.269, 5.24227e-24, 7.4986e-24, 3.96945e-24, 3.42358e-24, 6.08436e-24, 1.48093e-23], [103.504, 5.24964e-24, 7.49831e-24, 3.96813e-24, 3.42237e-24, 6.07998e-24, 1.48138e-23], [103.739, 5.25704e-24, 7.49801e-24, 3.96682e-24, 3.42117e-24, 6.07557e-24, 1.48183e-23], [103.975, 5.26447e-24, 7.49772e-24, 3.96551e-24, 3.41996e-24, 6.07113e-24, 1.48228e-23], [104.212, 5.27193e-24, 7.49743e-24, 3.9642e-24, 3.41876e-24, 6.06668e-24, 1.48273e-23], [104.449, 5.27942e-24, 7.49715e-24, 3.9629e-24, 3.41756e-24, 6.0622e-24, 1.48317e-23], [104.687, 5.28693e-24, 7.49686e-24, 3.96161e-24, 3.41636e-24, 6.05769e-24, 1.48361e-23], [104.925, 5.29448e-24, 7.49658e-24, 3.96032e-24, 3.41515e-24, 6.05317e-24, 1.48405e-23], [105.163, 5.30205e-24, 7.4963e-24, 3.95904e-24, 3.41396e-24, 6.04862e-24, 1.48448e-23], [105.403, 5.30966e-24, 7.49602e-24, 3.95776e-24, 3.41276e-24, 6.04404e-24, 1.48491e-23], [105.642, 5.31729e-24, 7.49575e-24, 3.95648e-24, 3.41156e-24, 6.03945e-24, 1.48534e-23], [105.883, 5.32496e-24, 7.49548e-24, 3.95521e-24, 3.41036e-24, 6.03483e-24, 1.48576e-23], [106.123, 5.33265e-24, 7.49521e-24, 3.95394e-24, 3.40916e-24, 6.03019e-24, 1.48618e-23], [106.365, 5.34037e-24, 7.49494e-24, 3.95268e-24, 3.40797e-24, 6.02552e-24, 1.4866e-23], [106.607, 5.34813e-24, 7.49467e-24, 3.95142e-24, 3.40677e-24, 6.02083e-24, 1.48702e-23], [106.849, 5.35591e-24, 7.49441e-24, 3.95017e-24, 3.40558e-24, 6.01612e-24, 1.48743e-23], [107.092, 5.36372e-24, 7.49415e-24, 3.94892e-24, 3.40438e-24, 6.01138e-24, 1.48784e-23], [107.336, 5.37156e-24, 7.49389e-24, 3.94767e-24, 3.40319e-24, 6.00662e-24, 1.48825e-23], [107.58, 5.37943e-24, 7.49364e-24, 3.94643e-24, 3.40199e-24, 6.00184e-24, 1.48865e-23], [107.825, 5.38733e-24, 7.49338e-24, 3.9452e-24, 3.4008e-24, 5.99704e-24, 1.48905e-23], [108.07, 5.39526e-24, 7.49313e-24, 3.94397e-24, 3.39961e-24, 5.99221e-24, 1.48945e-23], [108.316, 5.40322e-24, 7.49288e-24, 3.94274e-24, 3.39841e-24, 5.98736e-24, 1.48984e-23], [108.562, 5.4112e-24, 7.49264e-24, 3.94152e-24, 3.39722e-24, 5.98248e-24, 1.49024e-23], [108.809, 5.41922e-24, 7.49239e-24, 3.9403e-24, 3.39603e-24, 5.97758e-24, 1.49063e-23], [109.057, 5.42727e-24, 7.49215e-24, 3.93908e-24, 3.39483e-24, 5.97266e-24, 1.49101e-23], [109.305, 5.43535e-24, 7.49191e-24, 3.93787e-24, 3.39364e-24, 5.96772e-24, 1.4914e-23], [109.553, 5.44345e-24, 7.49168e-24, 3.93667e-24, 3.39245e-24, 5.96275e-24, 1.49178e-23], [109.802, 5.45159e-24, 7.49144e-24, 3.93546e-24, 3.39126e-24, 5.95776e-24, 1.49216e-23], [110.052, 5.45975e-24, 7.49121e-24, 3.93427e-24, 3.39006e-24, 5.95274e-24, 1.49253e-23], [110.302, 5.46795e-24, 7.49098e-24, 3.93307e-24, 3.38887e-24, 5.9477e-24, 1.49291e-23], [110.553, 5.47617e-24, 7.49076e-24, 3.93188e-24, 3.38768e-24, 5.94264e-24, 1.49328e-23], [110.805, 5.48443e-24, 7.49053e-24, 3.9307e-24, 3.38648e-24, 5.93756e-24, 1.49364e-23], [111.057, 5.49271e-24, 7.49031e-24, 3.92951e-24, 3.38529e-24, 5.93245e-24, 1.49401e-23], [111.309, 5.50103e-24, 7.49009e-24, 3.92834e-24, 3.38409e-24, 5.92732e-24, 1.49437e-23], [111.563, 5.50937e-24, 7.48987e-24, 3.92716e-24, 3.3829e-24, 5.92217e-24, 1.49473e-23], [111.816, 5.51775e-24, 7.48966e-24, 3.92599e-24, 3.3817e-24, 5.91699e-24, 1.49508e-23], [112.071, 5.52615e-24, 7.48945e-24, 3.92483e-24, 3.38051e-24, 5.91179e-24, 1.49544e-23], [112.326, 5.53459e-24, 7.48924e-24, 3.92367e-24, 3.37931e-24, 5.90656e-24, 1.49579e-23], [112.581, 5.54305e-24, 7.48903e-24, 3.92251e-24, 3.37812e-24, 5.90132e-24, 1.49614e-23], [112.837, 5.55154e-24, 7.48882e-24, 3.92136e-24, 3.37692e-24, 5.89605e-24, 1.49648e-23], [113.094, 5.56007e-24, 7.48862e-24, 3.92021e-24, 3.37572e-24, 5.89076e-24, 1.49682e-23], [113.351, 5.56862e-24, 7.48842e-24, 3.91906e-24, 3.37452e-24, 5.88544e-24, 1.49716e-23], [113.609, 5.57721e-24, 7.48822e-24, 3.91792e-24, 3.37332e-24, 5.8801e-24, 1.4975e-23], [113.867, 5.58582e-24, 7.48803e-24, 3.91678e-24, 3.37213e-24, 5.87474e-24, 1.49784e-23], [114.126, 5.59446e-24, 7.48784e-24, 3.91565e-24, 3.37092e-24, 5.86935e-24, 1.49817e-23], [114.386, 5.60314e-24, 7.48765e-24, 3.91452e-24, 3.36972e-24, 5.86395e-24, 1.4985e-23], [114.646, 5.61184e-24, 7.48746e-24, 3.91339e-24, 3.36852e-24, 5.85852e-24, 1.49882e-23], [114.907, 5.62058e-24, 7.48727e-24, 3.91227e-24, 3.36732e-24, 5.85306e-24, 1.49915e-23], [115.168, 5.62934e-24, 7.48709e-24, 3.91115e-24, 3.36611e-24, 5.84759e-24, 1.49947e-23], [115.43, 5.63814e-24, 7.48691e-24, 3.91003e-24, 3.36491e-24, 5.84209e-24, 1.49979e-23], [115.693, 5.64696e-24, 7.48673e-24, 3.90892e-24, 3.3637e-24, 5.83656e-24, 1.5001e-23], [115.956, 5.65582e-24, 7.48656e-24, 3.90781e-24, 3.3625e-24, 5.83102e-24, 1.50041e-23], [116.22, 5.66471e-24, 7.48639e-24, 3.90671e-24, 3.36129e-24, 5.82545e-24, 1.50072e-23], [116.484, 5.67362e-24, 7.48622e-24, 3.90561e-24, 3.36008e-24, 5.81986e-24, 1.50103e-23], [116.749, 5.68257e-24, 7.48605e-24, 3.90451e-24, 3.35887e-24, 5.81425e-24, 1.50134e-23], [117.014, 5.69155e-24, 7.48588e-24, 3.90342e-24, 3.35766e-24, 5.80861e-24, 1.50164e-23], [117.281, 5.70056e-24, 7.48572e-24, 3.90233e-24, 3.35645e-24, 5.80295e-24, 1.50194e-23], [117.547, 5.7096e-24, 7.48556e-24, 3.90125e-24, 3.35523e-24, 5.79727e-24, 1.50224e-23], [117.815, 5.71867e-24, 7.48541e-24, 3.90016e-24, 3.35402e-24, 5.79157e-24, 1.50253e-23], [118.083, 5.72777e-24, 7.48525e-24, 3.89909e-24, 3.3528e-24, 5.78584e-24, 1.50282e-23], [118.351, 5.7369e-24, 7.4851e-24, 3.89801e-24, 3.35158e-24, 5.78009e-24, 1.50311e-23], [118.621, 5.74606e-24, 7.48495e-24, 3.89694e-24, 3.35037e-24, 5.77432e-24, 1.5034e-23], [118.89, 5.75525e-24, 7.4848e-24, 3.89587e-24, 3.34915e-24, 5.76852e-24, 1.50369e-23], [119.161, 5.76448e-24, 7.48466e-24, 3.89481e-24, 3.34792e-24, 5.76271e-24, 1.50397e-23], [119.432, 5.77373e-24, 7.48452e-24, 3.89375e-24, 3.3467e-24, 5.75687e-24, 1.50425e-23], [119.703, 5.78301e-24, 7.48438e-24, 3.89269e-24, 3.34548e-24, 5.75101e-24, 1.50452e-23], [119.976, 5.79233e-24, 7.48424e-24, 3.89164e-24, 3.34425e-24, 5.74512e-24, 1.5048e-23], [120.249, 5.80168e-24, 7.48411e-24, 3.89059e-24, 3.34302e-24, 5.73922e-24, 1.50507e-23], [120.522, 5.81105e-24, 7.48398e-24, 3.88954e-24, 3.34179e-24, 5.73329e-24, 1.50534e-23], [120.796, 5.82046e-24, 7.48385e-24, 3.8885e-24, 3.34056e-24, 5.72734e-24, 1.50561e-23], [121.071, 5.8299e-24, 7.48372e-24, 3.88746e-24, 3.33933e-24, 5.72137e-24, 1.50587e-23], [121.346, 5.83937e-24, 7.4836e-24, 3.88642e-24, 3.3381e-24, 5.71537e-24, 1.50613e-23], [121.622, 5.84888e-24, 7.48348e-24, 3.88539e-24, 3.33686e-24, 5.70936e-24, 1.50639e-23], [121.899, 5.85841e-24, 7.48336e-24, 3.88436e-24, 3.33563e-24, 5.70332e-24, 1.50665e-23], [122.176, 5.86798e-24, 7.48325e-24, 3.88334e-24, 3.33439e-24, 5.69726e-24, 1.5069e-23], [122.454, 5.87757e-24, 7.48313e-24, 3.88231e-24, 3.33315e-24, 5.69118e-24, 1.50715e-23], [122.733, 5.8872e-24, 7.48302e-24, 3.88129e-24, 3.3319e-24, 5.68507e-24, 1.5074e-23], [123.012, 5.89686e-24, 7.48292e-24, 3.88028e-24, 3.33066e-24, 5.67895e-24, 1.50765e-23], [123.292, 5.90655e-24, 7.48281e-24, 3.87927e-24, 3.32941e-24, 5.6728e-24, 1.50789e-23], [123.572, 5.91627e-24, 7.48271e-24, 3.87826e-24, 3.32817e-24, 5.66663e-24, 1.50813e-23], [123.853, 5.92602e-24, 7.48261e-24, 3.87725e-24, 3.32692e-24, 5.66044e-24, 1.50837e-23], [124.135, 5.93581e-24, 7.48252e-24, 3.87625e-24, 3.32567e-24, 5.65423e-24, 1.50861e-23], [124.417, 5.94562e-24, 7.48242e-24, 3.87525e-24, 3.32441e-24, 5.648e-24, 1.50884e-23], [124.7, 5.95547e-24, 7.48233e-24, 3.87426e-24, 3.32316e-24, 5.64174e-24, 1.50908e-23], [124.984, 5.96535e-24, 7.48224e-24, 3.87327e-24, 3.3219e-24, 5.63547e-24, 1.50931e-23], [125.268, 5.97526e-24, 7.48216e-24, 3.87228e-24, 3.32064e-24, 5.62917e-24, 1.50953e-23], [125.553, 5.98521e-24, 7.48207e-24, 3.87129e-24, 3.31938e-24, 5.62286e-24, 1.50976e-23], [125.839, 5.99518e-24, 7.48199e-24, 3.87031e-24, 3.31812e-24, 5.61652e-24, 1.50998e-23], [126.125, 6.00519e-24, 7.48192e-24, 3.86933e-24, 3.31685e-24, 5.61016e-24, 1.5102e-23], [126.412, 6.01523e-24, 7.48184e-24, 3.86836e-24, 3.31558e-24, 5.60378e-24, 1.51042e-23], [126.699, 6.0253e-24, 7.48177e-24, 3.86739e-24, 3.31431e-24, 5.59738e-24, 1.51063e-23], [126.988, 6.0354e-24, 7.4817e-24, 3.86642e-24, 3.31304e-24, 5.59096e-24, 1.51084e-23], [127.276, 6.04554e-24, 7.48164e-24, 3.86545e-24, 3.31177e-24, 5.58452e-24, 1.51105e-23], [127.566, 6.0557e-24, 7.48157e-24, 3.86449e-24, 3.31049e-24, 5.57806e-24, 1.51126e-23], [127.856, 6.0659e-24, 7.48151e-24, 3.86353e-24, 3.30922e-24, 5.57157e-24, 1.51147e-23], [128.147, 6.07613e-24, 7.48146e-24, 3.86257e-24, 3.30794e-24, 5.56507e-24, 1.51167e-23], [128.438, 6.0864e-24, 7.4814e-24, 3.86162e-24, 3.30665e-24, 5.55855e-24, 1.51187e-23], [128.731, 6.09669e-24, 7.48135e-24, 3.86067e-24, 3.30537e-24, 5.55201e-24, 1.51207e-23], [129.023, 6.10702e-24, 7.4813e-24, 3.85973e-24, 3.30408e-24, 5.54544e-24, 1.51226e-23], [129.317, 6.11738e-24, 7.48125e-24, 3.85878e-24, 3.30279e-24, 5.53886e-24, 1.51246e-23], [129.611, 6.12778e-24, 7.48121e-24, 3.85784e-24, 3.3015e-24, 5.53226e-24, 1.51265e-23], [129.906, 6.1382e-24, 7.48117e-24, 3.85691e-24, 3.30021e-24, 5.52564e-24, 1.51283e-23], [130.201, 6.14866e-24, 7.48113e-24, 3.85597e-24, 3.29891e-24, 5.519e-24, 1.51302e-23], [130.497, 6.15915e-24, 7.4811e-24, 3.85504e-24, 3.29761e-24, 5.51234e-24, 1.5132e-23], [130.794, 6.16967e-24, 7.48107e-24, 3.85412e-24, 3.29631e-24, 5.50566e-24, 1.51339e-23], [131.092, 6.18023e-24, 7.48104e-24, 3.85319e-24, 3.29501e-24, 5.49896e-24, 1.51357e-23], [131.39, 6.19082e-24, 7.48101e-24, 3.85227e-24, 3.2937e-24, 5.49224e-24, 1.51374e-23], [131.689, 6.20144e-24, 7.48099e-24, 3.85135e-24, 3.2924e-24, 5.4855e-24, 1.51392e-23], [131.988, 6.2121e-24, 7.48097e-24, 3.85044e-24, 3.29109e-24, 5.47875e-24, 1.51409e-23], [132.288, 6.22279e-24, 7.48095e-24, 3.84953e-24, 3.28977e-24, 5.47197e-24, 1.51426e-23], [132.589, 6.23351e-24, 7.48094e-24, 3.84862e-24, 3.28846e-24, 5.46518e-24, 1.51443e-23], [132.891, 6.24426e-24, 7.48092e-24, 3.84772e-24, 3.28714e-24, 5.45837e-24, 1.51459e-23], [133.193, 6.25505e-24, 7.48092e-24, 3.84681e-24, 3.28582e-24, 5.45154e-24, 1.51475e-23], [133.496, 6.26587e-24, 7.48091e-24, 3.84591e-24, 3.2845e-24, 5.4447e-24, 1.51491e-23], [133.8, 6.27672e-24, 7.48091e-24, 3.84502e-24, 3.28317e-24, 5.43783e-24, 1.51507e-23], [134.104, 6.28761e-24, 7.48091e-24, 3.84413e-24, 3.28185e-24, 5.43095e-24, 1.51523e-23], [134.409, 6.29853e-24, 7.48091e-24, 3.84324e-24, 3.28051e-24, 5.42405e-24, 1.51538e-23], [134.715, 6.30948e-24, 7.48092e-24, 3.84235e-24, 3.27918e-24, 5.41713e-24, 1.51553e-23], [135.021, 6.32047e-24, 7.48093e-24, 3.84147e-24, 3.27785e-24, 5.4102e-24, 1.51568e-23], [135.328, 6.33149e-24, 7.48094e-24, 3.84059e-24, 3.27651e-24, 5.40324e-24, 1.51583e-23], [135.636, 6.34254e-24, 7.48096e-24, 3.83971e-24, 3.27517e-24, 5.39627e-24, 1.51597e-23], [135.945, 6.35363e-24, 7.48098e-24, 3.83884e-24, 3.27383e-24, 5.38929e-24, 1.51611e-23], [136.254, 6.36475e-24, 7.481e-24, 3.83796e-24, 3.27248e-24, 5.38229e-24, 1.51625e-23], [136.564, 6.3759e-24, 7.48102e-24, 3.8371e-24, 3.27113e-24, 5.37527e-24, 1.51639e-23], [136.875, 6.38709e-24, 7.48105e-24, 3.83623e-24, 3.26978e-24, 5.36823e-24, 1.51653e-23], [137.186, 6.39831e-24, 7.48108e-24, 3.83537e-24, 3.26843e-24, 5.36118e-24, 1.51666e-23], [137.498, 6.40957e-24, 7.48112e-24, 3.83451e-24, 3.26707e-24, 5.35411e-24, 1.51679e-23], [137.811, 6.42086e-24, 7.48115e-24, 3.83366e-24, 3.26571e-24, 5.34703e-24, 1.51692e-23], [138.124, 6.43218e-24, 7.48119e-24, 3.8328e-24, 3.26435e-24, 5.33993e-24, 1.51704e-23], [138.438, 6.44354e-24, 7.48124e-24, 3.83195e-24, 3.26298e-24, 5.33282e-24, 1.51717e-23], [138.753, 6.45493e-24, 7.48128e-24, 3.83111e-24, 3.26162e-24, 5.32569e-24, 1.51729e-23], [139.069, 6.46636e-24, 7.48133e-24, 3.83026e-24, 3.26025e-24, 5.31855e-24, 1.51741e-23], [139.385, 6.47782e-24, 7.48139e-24, 3.82942e-24, 3.25887e-24, 5.31139e-24, 1.51753e-23], [139.702, 6.48931e-24, 7.48144e-24, 3.82858e-24, 3.2575e-24, 5.30422e-24, 1.51764e-23], [140.02, 6.50084e-24, 7.4815e-24, 3.82775e-24, 3.25612e-24, 5.29703e-24, 1.51775e-23], [140.338, 6.5124e-24, 7.48157e-24, 3.82692e-24, 3.25474e-24, 5.28983e-24, 1.51786e-23], [140.658, 6.524e-24, 7.48163e-24, 3.82609e-24, 3.25335e-24, 5.28261e-24, 1.51797e-23], [140.978, 6.53563e-24, 7.4817e-24, 3.82526e-24, 3.25197e-24, 5.27538e-24, 1.51808e-23], [141.298, 6.5473e-24, 7.48177e-24, 3.82444e-24, 3.25058e-24, 5.26814e-24, 1.51818e-23], [141.62, 6.559e-24, 7.48185e-24, 3.82362e-24, 3.24919e-24, 5.26089e-24, 1.51828e-23], [141.942, 6.57074e-24, 7.48193e-24, 3.82281e-24, 3.24779e-24, 5.25362e-24, 1.51838e-23], [142.265, 6.58251e-24, 7.48201e-24, 3.82199e-24, 3.24639e-24, 5.24634e-24, 1.51848e-23], [142.588, 6.59431e-24, 7.48209e-24, 3.82118e-24, 3.24499e-24, 5.23904e-24, 1.51857e-23], [142.913, 6.60615e-24, 7.48218e-24, 3.82037e-24, 3.24359e-24, 5.23174e-24, 1.51867e-23], [143.238, 6.61803e-24, 7.48227e-24, 3.81957e-24, 3.24218e-24, 5.22442e-24, 1.51876e-23], [143.563, 6.62994e-24, 7.48237e-24, 3.81877e-24, 3.24077e-24, 5.21709e-24, 1.51884e-23], [143.89, 6.64188e-24, 7.48247e-24, 3.81797e-24, 3.23936e-24, 5.20975e-24, 1.51893e-23], [144.217, 6.65386e-24, 7.48257e-24, 3.81717e-24, 3.23795e-24, 5.2024e-24, 1.51901e-23], [144.545, 6.66588e-24, 7.48267e-24, 3.81638e-24, 3.23653e-24, 5.19503e-24, 1.51909e-23], [144.874, 6.67793e-24, 7.48278e-24, 3.81559e-24, 3.23511e-24, 5.18766e-24, 1.51917e-23], [145.204, 6.69001e-24, 7.48289e-24, 3.81481e-24, 3.23368e-24, 5.18028e-24, 1.51925e-23], [145.534, 6.70213e-24, 7.48301e-24, 3.81402e-24, 3.23226e-24, 5.17288e-24, 1.51933e-23], [145.865, 6.71429e-24, 7.48313e-24, 3.81324e-24, 3.23083e-24, 5.16548e-24, 1.5194e-23], [146.197, 6.72648e-24, 7.48325e-24, 3.81246e-24, 3.22939e-24, 5.15807e-24, 1.51947e-23], [146.529, 6.73871e-24, 7.48338e-24, 3.81169e-24, 3.22796e-24, 5.15064e-24, 1.51954e-23], [146.862, 6.75097e-24, 7.4835e-24, 3.81092e-24, 3.22652e-24, 5.14321e-24, 1.5196e-23], [147.197, 6.76327e-24, 7.48364e-24, 3.81015e-24, 3.22508e-24, 5.13577e-24, 1.51967e-23], [147.531, 6.77561e-24, 7.48377e-24, 3.80938e-24, 3.22363e-24, 5.12833e-24, 1.51973e-23], [147.867, 6.78798e-24, 7.48391e-24, 3.80862e-24, 3.22219e-24, 5.12087e-24, 1.51979e-23], [148.203, 6.80038e-24, 7.48405e-24, 3.80786e-24, 3.22074e-24, 5.11341e-24, 1.51984e-23], [148.54, 6.81283e-24, 7.4842e-24, 3.8071e-24, 3.21928e-24, 5.10594e-24, 1.5199e-23], [148.878, 6.8253e-24, 7.48435e-24, 3.80635e-24, 3.21783e-24, 5.09846e-24, 1.51995e-23], [149.217, 6.83782e-24, 7.4845e-24, 3.8056e-24, 3.21637e-24, 5.09098e-24, 1.52e-23], [149.556, 6.85037e-24, 7.48466e-24, 3.80485e-24, 3.2149e-24, 5.08349e-24, 1.52005e-23], [149.896, 6.86295e-24, 7.48482e-24, 3.8041e-24, 3.21344e-24, 5.07599e-24, 1.5201e-23], [150.237, 6.87558e-24, 7.48498e-24, 3.80336e-24, 3.21197e-24, 5.06849e-24, 1.52014e-23], [150.579, 6.88823e-24, 7.48515e-24, 3.80262e-24, 3.2105e-24, 5.06098e-24, 1.52018e-23], [150.922, 6.90093e-24, 7.48532e-24, 3.80189e-24, 3.20903e-24, 5.05347e-24, 1.52022e-23], [151.265, 6.91366e-24, 7.48549e-24, 3.80115e-24, 3.20755e-24, 5.04596e-24, 1.52026e-23], [151.609, 6.92643e-24, 7.48567e-24, 3.80042e-24, 3.20607e-24, 5.03844e-24, 1.52029e-23], [151.954, 6.93923e-24, 7.48585e-24, 3.7997e-24, 3.20459e-24, 5.03092e-24, 1.52033e-23], [152.299, 6.95207e-24, 7.48604e-24, 3.79897e-24, 3.2031e-24, 5.02339e-24, 1.52036e-23], [152.646, 6.96495e-24, 7.48622e-24, 3.79825e-24, 3.20161e-24, 5.01586e-24, 1.52039e-23], [152.993, 6.97787e-24, 7.48642e-24, 3.79753e-24, 3.20012e-24, 5.00833e-24, 1.52041e-23], [153.341, 6.99082e-24, 7.48661e-24, 3.79682e-24, 3.19862e-24, 5.0008e-24, 1.52044e-23], [153.69, 7.00381e-24, 7.48681e-24, 3.7961e-24, 3.19713e-24, 4.99327e-24, 1.52046e-23], [154.039, 7.01683e-24, 7.48701e-24, 3.79539e-24, 3.19562e-24, 4.98573e-24, 1.52048e-23], [154.39, 7.02989e-24, 7.48722e-24, 3.79469e-24, 3.19412e-24, 4.9782e-24, 1.5205e-23], [154.741, 7.04299e-24, 7.48743e-24, 3.79398e-24, 3.19261e-24, 4.97066e-24, 1.52051e-23], [155.093, 7.05613e-24, 7.48765e-24, 3.79328e-24, 3.1911e-24, 4.96312e-24, 1.52053e-23], [155.446, 7.0693e-24, 7.48786e-24, 3.79258e-24, 3.18959e-24, 4.95559e-24, 1.52054e-23], [155.799, 7.08251e-24, 7.48809e-24, 3.79189e-24, 3.18807e-24, 4.94806e-24, 1.52055e-23], [156.154, 7.09576e-24, 7.48831e-24, 3.7912e-24, 3.18655e-24, 4.94052e-24, 1.52055e-23], [156.509, 7.10905e-24, 7.48854e-24, 3.79051e-24, 3.18503e-24, 4.933e-24, 1.52056e-23], [156.865, 7.12237e-24, 7.48877e-24, 3.78982e-24, 3.18351e-24, 4.92547e-24, 1.52056e-23], [157.222, 7.13573e-24, 7.48901e-24, 3.78914e-24, 3.18198e-24, 4.91794e-24, 1.52056e-23], [157.579, 7.14913e-24, 7.48925e-24, 3.78846e-24, 3.18045e-24, 4.91042e-24, 1.52056e-23], [157.938, 7.16256e-24, 7.48949e-24, 3.78778e-24, 3.17891e-24, 4.90291e-24, 1.52056e-23], [158.297, 7.17604e-24, 7.48974e-24, 3.78711e-24, 3.17738e-24, 4.8954e-24, 1.52055e-23], [158.657, 7.18955e-24, 7.48999e-24, 3.78644e-24, 3.17584e-24, 4.88789e-24, 1.52054e-23], [159.018, 7.2031e-24, 7.49025e-24, 3.78577e-24, 3.17429e-24, 4.88039e-24, 1.52053e-23], [159.379, 7.21668e-24, 7.49051e-24, 3.78511e-24, 3.17275e-24, 4.8729e-24, 1.52052e-23], [159.742, 7.23031e-24, 7.49077e-24, 3.78444e-24, 3.1712e-24, 4.86541e-24, 1.52051e-23], [160.105, 7.24397e-24, 7.49104e-24, 3.78378e-24, 3.16965e-24, 4.85793e-24, 1.52049e-23], [160.47, 7.25767e-24, 7.49131e-24, 3.78313e-24, 3.16809e-24, 4.85046e-24, 1.52047e-23], [160.835, 7.27141e-24, 7.49159e-24, 3.78247e-24, 3.16653e-24, 4.843e-24, 1.52045e-23], [161.2, 7.28519e-24, 7.49187e-24, 3.78182e-24, 3.16497e-24, 4.83555e-24, 1.52043e-23], [161.567, 7.29901e-24, 7.49215e-24, 3.78118e-24, 3.16341e-24, 4.8281e-24, 1.5204e-23], [161.935, 7.31286e-24, 7.49244e-24, 3.78053e-24, 3.16184e-24, 4.82067e-24, 1.52037e-23], [162.303, 7.32676e-24, 7.49273e-24, 3.77989e-24, 3.16027e-24, 4.81325e-24, 1.52034e-23], [162.672, 7.34069e-24, 7.49302e-24, 3.77925e-24, 3.1587e-24, 4.80584e-24, 1.52031e-23], [163.042, 7.35466e-24, 7.49332e-24, 3.77862e-24, 3.15712e-24, 4.79844e-24, 1.52028e-23], [163.413, 7.36867e-24, 7.49362e-24, 3.77799e-24, 3.15555e-24, 4.79105e-24, 1.52024e-23], [163.785, 7.38272e-24, 7.49393e-24, 3.77736e-24, 3.15396e-24, 4.78368e-24, 1.5202e-23], [164.157, 7.3968e-24, 7.49424e-24, 3.77673e-24, 3.15238e-24, 4.77632e-24, 1.52016e-23], [164.53, 7.41093e-24, 7.49456e-24, 3.77611e-24, 3.15079e-24, 4.76898e-24, 1.52012e-23], [164.905, 7.42509e-24, 7.49488e-24, 3.77549e-24, 3.1492e-24, 4.76165e-24, 1.52007e-23], [165.28, 7.4393e-24, 7.4952e-24, 3.77487e-24, 3.14761e-24, 4.75434e-24, 1.52003e-23], [165.656, 7.45354e-24, 7.49553e-24, 3.77426e-24, 3.14601e-24, 4.74705e-24, 1.51998e-23], [166.033, 7.46782e-24, 7.49586e-24, 3.77364e-24, 3.14441e-24, 4.73977e-24, 1.51993e-23], [166.41, 7.48215e-24, 7.49619e-24, 3.77304e-24, 3.14281e-24, 4.73251e-24, 1.51987e-23], [166.789, 7.49651e-24, 7.49653e-24, 3.77243e-24, 3.14121e-24, 4.72527e-24, 1.51982e-23], [167.168, 7.51091e-24, 7.49688e-24, 3.77183e-24, 3.1396e-24, 4.71806e-24, 1.51976e-23], [167.548, 7.52535e-24, 7.49723e-24, 3.77123e-24, 3.13799e-24, 4.71086e-24, 1.5197e-23], [167.929, 7.53983e-24, 7.49758e-24, 3.77063e-24, 3.13638e-24, 4.70368e-24, 1.51964e-23], [168.311, 7.55435e-24, 7.49794e-24, 3.77004e-24, 3.13476e-24, 4.69652e-24, 1.51957e-23], [168.694, 7.56891e-24, 7.4983e-24, 3.76945e-24, 3.13314e-24, 4.68939e-24, 1.51951e-23], [169.078, 7.5835e-24, 7.49866e-24, 3.76886e-24, 3.13152e-24, 4.68228e-24, 1.51944e-23], [169.463, 7.59814e-24, 7.49903e-24, 3.76828e-24, 3.1299e-24, 4.6752e-24, 1.51937e-23], [169.848, 7.61282e-24, 7.4994e-24, 3.7677e-24, 3.12827e-24, 4.66814e-24, 1.5193e-23], [170.234, 7.62754e-24, 7.49978e-24, 3.76712e-24, 3.12664e-24, 4.66111e-24, 1.51922e-23], [170.621, 7.6423e-24, 7.50016e-24, 3.76655e-24, 3.12501e-24, 4.6541e-24, 1.51914e-23], [171.01, 7.6571e-24, 7.50055e-24, 3.76597e-24, 3.12337e-24, 4.64712e-24, 1.51906e-23], [171.399, 7.67194e-24, 7.50094e-24, 3.7654e-24, 3.12173e-24, 4.64017e-24, 1.51898e-23], [171.788, 7.68682e-24, 7.50134e-24, 3.76484e-24, 3.12009e-24, 4.63325e-24, 1.5189e-23], [172.179, 7.70174e-24, 7.50174e-24, 3.76428e-24, 3.11845e-24, 4.62636e-24, 1.51881e-23], [172.571, 7.7167e-24, 7.50214e-24, 3.76372e-24, 3.1168e-24, 4.6195e-24, 1.51873e-23], [172.963, 7.7317e-24, 7.50255e-24, 3.76316e-24, 3.11515e-24, 4.61267e-24, 1.51863e-23], [173.357, 7.74674e-24, 7.50296e-24, 3.76261e-24, 3.1135e-24, 4.60588e-24, 1.51854e-23], [173.751, 7.76183e-24, 7.50338e-24, 3.76206e-24, 3.11185e-24, 4.59912e-24, 1.51845e-23], [174.146, 7.77695e-24, 7.5038e-24, 3.76151e-24, 3.11019e-24, 4.59239e-24, 1.51835e-23], [174.542, 7.79211e-24, 7.50423e-24, 3.76097e-24, 3.10853e-24, 4.5857e-24, 1.51825e-23], [174.939, 7.80732e-24, 7.50466e-24, 3.76042e-24, 3.10687e-24, 4.57905e-24, 1.51815e-23], [175.337, 7.82257e-24, 7.50509e-24, 3.75989e-24, 3.1052e-24, 4.57244e-24, 1.51805e-23], [175.736, 7.83785e-24, 7.50553e-24, 3.75935e-24, 3.10353e-24, 4.56586e-24, 1.51794e-23], [176.136, 7.85318e-24, 7.50598e-24, 3.75882e-24, 3.10186e-24, 4.55932e-24, 1.51783e-23], [176.536, 7.86855e-24, 7.50642e-24, 3.75829e-24, 3.10019e-24, 4.55282e-24, 1.51772e-23], [176.938, 7.88396e-24, 7.50688e-24, 3.75777e-24, 3.09851e-24, 4.54637e-24, 1.51761e-23], [177.34, 7.89942e-24, 7.50734e-24, 3.75724e-24, 3.09684e-24, 4.53996e-24, 1.5175e-23], [177.744, 7.91491e-24, 7.5078e-24, 3.75672e-24, 3.09516e-24, 4.53359e-24, 1.51738e-23], [178.148, 7.93045e-24, 7.50826e-24, 3.75621e-24, 3.09347e-24, 4.52726e-24, 1.51726e-23], [178.553, 7.94603e-24, 7.50874e-24, 3.75569e-24, 3.09179e-24, 4.52099e-24, 1.51714e-23], [178.959, 7.96164e-24, 7.50921e-24, 3.75518e-24, 3.0901e-24, 4.51475e-24, 1.51702e-23], [179.367, 7.97731e-24, 7.50969e-24, 3.75468e-24, 3.08841e-24, 4.50857e-24, 1.51689e-23], [179.775, 7.99301e-24, 7.51018e-24, 3.75417e-24, 3.08672e-24, 4.50244e-24, 1.51677e-23], [180.183, 8.00875e-24, 7.51067e-24, 3.75367e-24, 3.08502e-24, 4.49635e-24, 1.51664e-23], [180.593, 8.02454e-24, 7.51116e-24, 3.75318e-24, 3.08332e-24, 4.49032e-24, 1.51651e-23], [181.004, 8.04037e-24, 7.51166e-24, 3.75268e-24, 3.08162e-24, 4.48434e-24, 1.51637e-23], [181.416, 8.05624e-24, 7.51217e-24, 3.75219e-24, 3.07992e-24, 4.47841e-24, 1.51624e-23], [181.828, 8.07216e-24, 7.51268e-24, 3.7517e-24, 3.07822e-24, 4.47253e-24, 1.5161e-23], [182.242, 8.08811e-24, 7.51319e-24, 3.75122e-24, 3.07651e-24, 4.46672e-24, 1.51596e-23], [182.656, 8.10411e-24, 7.51371e-24, 3.75074e-24, 3.0748e-24, 4.46096e-24, 1.51581e-23], [183.072, 8.12015e-24, 7.51423e-24, 3.75026e-24, 3.07309e-24, 4.45525e-24, 1.51567e-23], [183.488, 8.13624e-24, 7.51476e-24, 3.74978e-24, 3.07138e-24, 4.44961e-24, 1.51552e-23], [183.906, 8.15236e-24, 7.5153e-24, 3.74931e-24, 3.06966e-24, 4.44403e-24, 1.51537e-23], [184.324, 8.16853e-24, 7.51583e-24, 3.74884e-24, 3.06794e-24, 4.4385e-24, 1.51522e-23], [184.743, 8.18474e-24, 7.51638e-24, 3.74838e-24, 3.06622e-24, 4.43305e-24, 1.51507e-23], [185.163, 8.201e-24, 7.51693e-24, 3.74792e-24, 3.0645e-24, 4.42765e-24, 1.51491e-23], [185.585, 8.2173e-24, 7.51748e-24, 3.74746e-24, 3.06278e-24, 4.42232e-24, 1.51475e-23], [186.007, 8.23364e-24, 7.51804e-24, 3.747e-24, 3.06105e-24, 4.41706e-24, 1.51459e-23], [186.43, 8.25002e-24, 7.5186e-24, 3.74655e-24, 3.05932e-24, 4.41186e-24, 1.51443e-23], [186.854, 8.26645e-24, 7.51917e-24, 3.7461e-24, 3.05759e-24, 4.40674e-24, 1.51426e-23], [187.279, 8.28292e-24, 7.51974e-24, 3.74565e-24, 3.05586e-24, 4.40168e-24, 1.5141e-23], [187.705, 8.29944e-24, 7.52032e-24, 3.74521e-24, 3.05412e-24, 4.3967e-24, 1.51393e-23], [188.132, 8.316e-24, 7.5209e-24, 3.74477e-24, 3.05239e-24, 4.39179e-24, 1.51376e-23], [188.56, 8.3326e-24, 7.52149e-24, 3.74433e-24, 3.05065e-24, 4.38696e-24, 1.51358e-23], [188.989, 8.34924e-24, 7.52208e-24, 3.7439e-24, 3.04891e-24, 4.3822e-24, 1.5134e-23], [189.419, 8.36593e-24, 7.52268e-24, 3.74347e-24, 3.04716e-24, 4.37751e-24, 1.51323e-23], [189.849, 8.38267e-24, 7.52328e-24, 3.74304e-24, 3.04542e-24, 4.37291e-24, 1.51305e-23], [190.281, 8.39944e-24, 7.52389e-24, 3.74262e-24, 3.04367e-24, 4.36838e-24, 1.51286e-23], [190.714, 8.41627e-24, 7.52451e-24, 3.7422e-24, 3.04193e-24, 4.36394e-24, 1.51268e-23], [191.148, 8.43313e-24, 7.52512e-24, 3.74178e-24, 3.04018e-24, 4.35958e-24, 1.51249e-23], [191.583, 8.45004e-24, 7.52575e-24, 3.74137e-24, 3.03842e-24, 4.3553e-24, 1.5123e-23], [192.018, 8.46699e-24, 7.52638e-24, 3.74096e-24, 3.03667e-24, 4.35111e-24, 1.51211e-23], [192.455, 8.48399e-24, 7.52701e-24, 3.74055e-24, 3.03492e-24, 4.34701e-24, 1.51191e-23], [192.893, 8.50104e-24, 7.52765e-24, 3.74015e-24, 3.03316e-24, 4.34299e-24, 1.51172e-23], [193.332, 8.51812e-24, 7.5283e-24, 3.73975e-24, 3.0314e-24, 4.33906e-24, 1.51152e-23], [193.771, 8.53526e-24, 7.52895e-24, 3.73935e-24, 3.02964e-24, 4.33523e-24, 1.51132e-23], [194.212, 8.55243e-24, 7.5296e-24, 3.73896e-24, 3.02788e-24, 4.33148e-24, 1.51111e-23], [194.654, 8.56965e-24, 7.53026e-24, 3.73857e-24, 3.02612e-24, 4.32783e-24, 1.51091e-23], [195.097, 8.58692e-24, 7.53093e-24, 3.73818e-24, 3.02435e-24, 4.32428e-24, 1.5107e-23], [195.54, 8.60423e-24, 7.5316e-24, 3.7378e-24, 3.02259e-24, 4.32082e-24, 1.51049e-23], [195.985, 8.62159e-24, 7.53228e-24, 3.73742e-24, 3.02082e-24, 4.31747e-24, 1.51028e-23], [196.431, 8.63899e-24, 7.53296e-24, 3.73704e-24, 3.01905e-24, 4.31421e-24, 1.51006e-23], [196.878, 8.65644e-24, 7.53365e-24, 3.73667e-24, 3.01728e-24, 4.31105e-24, 1.50984e-23], [197.326, 8.67393e-24, 7.53434e-24, 3.7363e-24, 3.01551e-24, 4.308e-24, 1.50963e-23], [197.774, 8.69147e-24, 7.53504e-24, 3.73593e-24, 3.01374e-24, 4.30505e-24, 1.5094e-23], [198.224, 8.70905e-24, 7.53574e-24, 3.73557e-24, 3.01197e-24, 4.3022e-24, 1.50918e-23], [198.675, 8.72668e-24, 7.53645e-24, 3.73521e-24, 3.01019e-24, 4.29947e-24, 1.50895e-23], [199.127, 8.74435e-24, 7.53717e-24, 3.73485e-24, 3.00842e-24, 4.29684e-24, 1.50872e-23], [199.58, 8.76208e-24, 7.53789e-24, 3.7345e-24, 3.00664e-24, 4.29432e-24, 1.50849e-23], [200.034, 8.77984e-24, 7.53861e-24, 3.73415e-24, 3.00486e-24, 4.29192e-24, 1.50826e-23], [200.489, 8.79765e-24, 7.53935e-24, 3.73381e-24, 3.00308e-24, 4.28962e-24, 1.50802e-23], [200.945, 8.81551e-24, 7.54008e-24, 3.73346e-24, 3.0013e-24, 4.28745e-24, 1.50779e-23], [201.402, 8.83342e-24, 7.54083e-24, 3.73312e-24, 2.99952e-24, 4.28539e-24, 1.50755e-23], [201.86, 8.85137e-24, 7.54158e-24, 3.73279e-24, 2.99774e-24, 4.28345e-24, 1.5073e-23], [202.319, 8.86936e-24, 7.54233e-24, 3.73246e-24, 2.99596e-24, 4.28163e-24, 1.50706e-23], [202.779, 8.88741e-24, 7.54309e-24, 3.73213e-24, 2.99417e-24, 4.27993e-24, 1.50681e-23], [203.241, 8.9055e-24, 7.54386e-24, 3.7318e-24, 2.99239e-24, 4.27835e-24, 1.50656e-23], [203.703, 8.92364e-24, 7.54463e-24, 3.73148e-24, 2.9906e-24, 4.2769e-24, 1.50631e-23], [204.166, 8.94182e-24, 7.54541e-24, 3.73116e-24, 2.98882e-24, 4.27557e-24, 1.50605e-23], [204.631, 8.96005e-24, 7.54619e-24, 3.73085e-24, 2.98703e-24, 4.27437e-24, 1.5058e-23], [205.096, 8.97833e-24, 7.54698e-24, 3.73054e-24, 2.98525e-24, 4.2733e-24, 1.50554e-23], [205.563, 8.99665e-24, 7.54777e-24, 3.73023e-24, 2.98346e-24, 4.27236e-24, 1.50528e-23], [206.03, 9.01502e-24, 7.54857e-24, 3.72993e-24, 2.98167e-24, 4.27156e-24, 1.50501e-23], [206.499, 9.03344e-24, 7.54938e-24, 3.72962e-24, 2.97988e-24, 4.27088e-24, 1.50475e-23], [206.968, 9.05191e-24, 7.55019e-24, 3.72933e-24, 2.97809e-24, 4.27035e-24, 1.50448e-23], [207.439, 9.07042e-24, 7.55101e-24, 3.72903e-24, 2.9763e-24, 4.26995e-24, 1.50421e-23], [207.911, 9.08898e-24, 7.55183e-24, 3.72874e-24, 2.97451e-24, 4.26969e-24, 1.50393e-23], [208.384, 9.10759e-24, 7.55266e-24, 3.72846e-24, 2.97272e-24, 4.26957e-24, 1.50366e-23], [208.858, 9.12624e-24, 7.5535e-24, 3.72817e-24, 2.97093e-24, 4.26959e-24, 1.50338e-23], [209.333, 9.14495e-24, 7.55434e-24, 3.7279e-24, 2.96914e-24, 4.26975e-24, 1.5031e-23], [209.809, 9.1637e-24, 7.55519e-24, 3.72762e-24, 2.96735e-24, 4.27006e-24, 1.50281e-23], [210.286, 9.1825e-24, 7.55605e-24, 3.72735e-24, 2.96556e-24, 4.27051e-24, 1.50253e-23], [210.765, 9.20134e-24, 7.55691e-24, 3.72708e-24, 2.96377e-24, 4.27112e-24, 1.50224e-23], [211.244, 9.22024e-24, 7.55777e-24, 3.72682e-24, 2.96198e-24, 4.27187e-24, 1.50195e-23], [211.725, 9.23918e-24, 7.55864e-24, 3.72655e-24, 2.96019e-24, 4.27277e-24, 1.50166e-23], [212.206, 9.25817e-24, 7.55952e-24, 3.7263e-24, 2.9584e-24, 4.27383e-24, 1.50136e-23], [212.689, 9.27721e-24, 7.56041e-24, 3.72604e-24, 2.95661e-24, 4.27504e-24, 1.50106e-23], [213.173, 9.2963e-24, 7.5613e-24, 3.72579e-24, 2.95482e-24, 4.27641e-24, 1.50076e-23], [213.658, 9.31544e-24, 7.5622e-24, 3.72555e-24, 2.95303e-24, 4.27793e-24, 1.50046e-23], [214.143, 9.33462e-24, 7.5631e-24, 3.7253e-24, 2.95124e-24, 4.27961e-24, 1.50015e-23], [214.631, 9.35386e-24, 7.56401e-24, 3.72506e-24, 2.94946e-24, 4.28145e-24, 1.49985e-23], [215.119, 9.37314e-24, 7.56493e-24, 3.72483e-24, 2.94767e-24, 4.28346e-24, 1.49954e-23], [215.608, 9.39248e-24, 7.56585e-24, 3.7246e-24, 2.94588e-24, 4.28563e-24, 1.49922e-23], [216.098, 9.41186e-24, 7.56678e-24, 3.72437e-24, 2.9441e-24, 4.28796e-24, 1.49891e-23], [216.59, 9.43129e-24, 7.56771e-24, 3.72414e-24, 2.94231e-24, 4.29045e-24, 1.49859e-23], [217.083, 9.45077e-24, 7.56865e-24, 3.72392e-24, 2.94052e-24, 4.29312e-24, 1.49827e-23], [217.576, 9.4703e-24, 7.5696e-24, 3.72371e-24, 2.93874e-24, 4.29595e-24, 1.49795e-23], [218.071, 9.48988e-24, 7.57056e-24, 3.72349e-24, 2.93696e-24, 4.29896e-24, 1.49762e-23], [218.567, 9.5095e-24, 7.57152e-24, 3.72328e-24, 2.93517e-24, 4.30213e-24, 1.4973e-23], [219.064, 9.52918e-24, 7.57248e-24, 3.72308e-24, 2.93339e-24, 4.30548e-24, 1.49697e-23], [219.563, 9.54891e-24, 7.57346e-24, 3.72288e-24, 2.93161e-24, 4.30901e-24, 1.49663e-23], [220.062, 9.56869e-24, 7.57444e-24, 3.72268e-24, 2.92983e-24, 4.31271e-24, 1.4963e-23], [220.563, 9.58851e-24, 7.57542e-24, 3.72249e-24, 2.92806e-24, 4.31658e-24, 1.49596e-23], [221.064, 9.60839e-24, 7.57642e-24, 3.72229e-24, 2.92628e-24, 4.32064e-24, 1.49562e-23], [221.567, 9.62832e-24, 7.57742e-24, 3.72211e-24, 2.9245e-24, 4.32487e-24, 1.49528e-23], [222.071, 9.6483e-24, 7.57842e-24, 3.72193e-24, 2.92273e-24, 4.32929e-24, 1.49493e-23], [222.576, 9.66833e-24, 7.57944e-24, 3.72175e-24, 2.92096e-24, 4.33389e-24, 1.49459e-23], [223.083, 9.68841e-24, 7.58046e-24, 3.72157e-24, 2.91919e-24, 4.33867e-24, 1.49424e-23], [223.59, 9.70854e-24, 7.58148e-24, 3.7214e-24, 2.91742e-24, 4.34364e-24, 1.49388e-23], [224.099, 9.72872e-24, 7.58252e-24, 3.72123e-24, 2.91565e-24, 4.34879e-24, 1.49353e-23], [224.608, 9.74895e-24, 7.58356e-24, 3.72107e-24, 2.91388e-24, 4.35414e-24, 1.49317e-23], [225.119, 9.76923e-24, 7.5846e-24, 3.72091e-24, 2.91212e-24, 4.35967e-24, 1.49281e-23], [225.631, 9.78956e-24, 7.58566e-24, 3.72075e-24, 2.91036e-24, 4.36539e-24, 1.49244e-23], [226.144, 9.80994e-24, 7.58672e-24, 3.7206e-24, 2.9086e-24, 4.3713e-24, 1.49208e-23], [226.659, 9.83038e-24, 7.58778e-24, 3.72045e-24, 2.90684e-24, 4.3774e-24, 1.49171e-23], [227.174, 9.85087e-24, 7.58886e-24, 3.72031e-24, 2.90508e-24, 4.3837e-24, 1.49134e-23], [227.691, 9.8714e-24, 7.58994e-24, 3.72017e-24, 2.90333e-24, 4.39019e-24, 1.49097e-23], [228.209, 9.89199e-24, 7.59103e-24, 3.72003e-24, 2.90158e-24, 4.39688e-24, 1.49059e-23], [228.728, 9.91263e-24, 7.59212e-24, 3.7199e-24, 2.89983e-24, 4.40377e-24, 1.49021e-23], [229.248, 9.93333e-24, 7.59322e-24, 3.71977e-24, 2.89809e-24, 4.41085e-24, 1.48983e-23], [229.77, 9.95407e-24, 7.59433e-24, 3.71965e-24, 2.89634e-24, 4.41813e-24, 1.48945e-23], [230.292, 9.97487e-24, 7.59545e-24, 3.71953e-24, 2.8946e-24, 4.42561e-24, 1.48906e-23], [230.816, 9.99571e-24, 7.59657e-24, 3.71941e-24, 2.89286e-24, 4.43329e-24, 1.48867e-23], [231.341, 1.00166e-23, 7.5977e-24, 3.7193e-24, 2.89113e-24, 4.44118e-24, 1.48828e-23], [231.867, 1.00376e-23, 7.59884e-24, 3.71919e-24, 2.8894e-24, 4.44926e-24, 1.48788e-23], [232.395, 1.00586e-23, 7.59998e-24, 3.71909e-24, 2.88767e-24, 4.45755e-24, 1.48748e-23], [232.923, 1.00796e-23, 7.60113e-24, 3.71899e-24, 2.88594e-24, 4.46605e-24, 1.48708e-23], [233.453, 1.01007e-23, 7.60229e-24, 3.71889e-24, 2.88422e-24, 4.47475e-24, 1.48668e-23], [233.984, 1.01219e-23, 7.60346e-24, 3.7188e-24, 2.8825e-24, 4.48366e-24, 1.48628e-23], [234.516, 1.01431e-23, 7.60463e-24, 3.71871e-24, 2.88078e-24, 4.49277e-24, 1.48587e-23], [235.05, 1.01644e-23, 7.60581e-24, 3.71863e-24, 2.87907e-24, 4.50209e-24, 1.48546e-23], [235.585, 1.01857e-23, 7.607e-24, 3.71855e-24, 2.87736e-24, 4.51163e-24, 1.48504e-23], [236.12, 1.02071e-23, 7.60819e-24, 3.71847e-24, 2.87565e-24, 4.52137e-24, 1.48463e-23], [236.657, 1.02285e-23, 7.6094e-24, 3.7184e-24, 2.87395e-24, 4.53132e-24, 1.48421e-23], [237.196, 1.025e-23, 7.61061e-24, 3.71833e-24, 2.87225e-24, 4.54148e-24, 1.48379e-23], [237.735, 1.02715e-23, 7.61182e-24, 3.71827e-24, 2.87056e-24, 4.55186e-24, 1.48336e-23], [238.276, 1.02931e-23, 7.61305e-24, 3.71821e-24, 2.86887e-24, 4.56244e-24, 1.48294e-23], [238.818, 1.03147e-23, 7.61428e-24, 3.71816e-24, 2.86718e-24, 4.57324e-24, 1.48251e-23], [239.361, 1.03364e-23, 7.61552e-24, 3.7181e-24, 2.8655e-24, 4.58426e-24, 1.48207e-23], [239.906, 1.03582e-23, 7.61677e-24, 3.71806e-24, 2.86382e-24, 4.59548e-24, 1.48164e-23], [240.451, 1.038e-23, 7.61802e-24, 3.71802e-24, 2.86215e-24, 4.60693e-24, 1.4812e-23], [240.998, 1.04019e-23, 7.61928e-24, 3.71798e-24, 2.86048e-24, 4.61859e-24, 1.48076e-23], [241.546, 1.04238e-23, 7.62055e-24, 3.71794e-24, 2.85882e-24, 4.63046e-24, 1.48031e-23], [242.096, 1.04457e-23, 7.62183e-24, 3.71791e-24, 2.85716e-24, 4.64255e-24, 1.47987e-23], [242.646, 1.04677e-23, 7.62312e-24, 3.71789e-24, 2.8555e-24, 4.65486e-24, 1.47942e-23], [243.198, 1.04898e-23, 7.62441e-24, 3.71787e-24, 2.85385e-24, 4.66739e-24, 1.47897e-23], [243.752, 1.0512e-23, 7.62571e-24, 3.71785e-24, 2.85221e-24, 4.68014e-24, 1.47851e-23], [244.306, 1.05341e-23, 7.62702e-24, 3.71784e-24, 2.85057e-24, 4.6931e-24, 1.47805e-23], [244.862, 1.05564e-23, 7.62833e-24, 3.71783e-24, 2.84893e-24, 4.70628e-24, 1.47759e-23], [245.419, 1.05787e-23, 7.62966e-24, 3.71782e-24, 2.8473e-24, 4.71969e-24, 1.47713e-23], [245.977, 1.0601e-23, 7.63099e-24, 3.71782e-24, 2.84568e-24, 4.73331e-24, 1.47666e-23], [246.536, 1.06234e-23, 7.63233e-24, 3.71783e-24, 2.84406e-24, 4.74715e-24, 1.47619e-23], [247.097, 1.06459e-23, 7.63368e-24, 3.71784e-24, 2.84245e-24, 4.76121e-24, 1.47572e-23], [247.659, 1.06684e-23, 7.63503e-24, 3.71785e-24, 2.84085e-24, 4.7755e-24, 1.47525e-23], [248.222, 1.0691e-23, 7.6364e-24, 3.71787e-24, 2.83925e-24, 4.79e-24, 1.47477e-23], [248.787, 1.07136e-23, 7.63777e-24, 3.71789e-24, 2.83765e-24, 4.80473e-24, 1.47429e-23], [249.353, 1.07363e-23, 7.63915e-24, 3.71792e-24, 2.83606e-24, 4.81968e-24, 1.4738e-23], [249.92, 1.0759e-23, 7.64053e-24, 3.71795e-24, 2.83448e-24, 4.83485e-24, 1.47332e-23], [250.489, 1.07818e-23, 7.64193e-24, 3.71798e-24, 2.83291e-24, 4.85024e-24, 1.47283e-23], [251.058, 1.08046e-23, 7.64333e-24, 3.71802e-24, 2.83134e-24, 4.86585e-24, 1.47234e-23], [251.629, 1.08275e-23, 7.64474e-24, 3.71806e-24, 2.82978e-24, 4.88169e-24, 1.47184e-23], [252.202, 1.08505e-23, 7.64616e-24, 3.71811e-24, 2.82822e-24, 4.89775e-24, 1.47134e-23], [252.775, 1.08735e-23, 7.64759e-24, 3.71817e-24, 2.82667e-24, 4.91403e-24, 1.47084e-23], [253.35, 1.08966e-23, 7.64903e-24, 3.71822e-24, 2.82513e-24, 4.93053e-24, 1.47034e-23], [253.927, 1.09197e-23, 7.65047e-24, 3.71829e-24, 2.8236e-24, 4.94726e-24, 1.46983e-23], [254.504, 1.09429e-23, 7.65193e-24, 3.71835e-24, 2.82207e-24, 4.96421e-24, 1.46932e-23], [255.083, 1.09662e-23, 7.65339e-24, 3.71842e-24, 2.82055e-24, 4.98138e-24, 1.46881e-23], [255.663, 1.09895e-23, 7.65486e-24, 3.7185e-24, 2.81904e-24, 4.99878e-24, 1.46829e-23], [256.245, 1.10128e-23, 7.65634e-24, 3.71858e-24, 2.81753e-24, 5.0164e-24, 1.46777e-23], [256.828, 1.10362e-23, 7.65782e-24, 3.71866e-24, 2.81604e-24, 5.03424e-24, 1.46725e-23], [257.412, 1.10597e-23, 7.65932e-24, 3.71875e-24, 2.81455e-24, 5.05231e-24, 1.46673e-23], [257.997, 1.10832e-23, 7.66082e-24, 3.71884e-24, 2.81307e-24, 5.0706e-24, 1.4662e-23], [258.584, 1.11068e-23, 7.66233e-24, 3.71894e-24, 2.8116e-24, 5.08911e-24, 1.46567e-23], [259.172, 1.11305e-23, 7.66385e-24, 3.71905e-24, 2.81013e-24, 5.10785e-24, 1.46513e-23], [259.762, 1.11542e-23, 7.66538e-24, 3.71915e-24, 2.80868e-24, 5.12681e-24, 1.4646e-23], [260.353, 1.11779e-23, 7.66692e-24, 3.71926e-24, 2.80723e-24, 5.14599e-24, 1.46406e-23], [260.945, 1.12018e-23, 7.66847e-24, 3.71938e-24, 2.80579e-24, 5.16539e-24, 1.46351e-23], [261.538, 1.12256e-23, 7.67002e-24, 3.7195e-24, 2.80437e-24, 5.18502e-24, 1.46297e-23], [262.133, 1.12496e-23, 7.67159e-24, 3.71963e-24, 2.80295e-24, 5.20487e-24, 1.46242e-23], [262.73, 1.12736e-23, 7.67316e-24, 3.71976e-24, 2.80153e-24, 5.22494e-24, 1.46186e-23], [263.327, 1.12976e-23, 7.67474e-24, 3.71989e-24, 2.80013e-24, 5.24523e-24, 1.46131e-23], [263.926, 1.13217e-23, 7.67633e-24, 3.72003e-24, 2.79874e-24, 5.26575e-24, 1.46075e-23], [264.526, 1.13459e-23, 7.67793e-24, 3.72018e-24, 2.79736e-24, 5.28649e-24, 1.46019e-23], [265.128, 1.13701e-23, 7.67954e-24, 3.72033e-24, 2.79599e-24, 5.30745e-24, 1.45962e-23], [265.731, 1.13944e-23, 7.68115e-24, 3.72048e-24, 2.79462e-24, 5.32863e-24, 1.45906e-23], [266.336, 1.14188e-23, 7.68278e-24, 3.72064e-24, 2.79327e-24, 5.35003e-24, 1.45849e-23], [266.941, 1.14432e-23, 7.68441e-24, 3.72081e-24, 2.79193e-24, 5.37166e-24, 1.45791e-23], [267.549, 1.14677e-23, 7.68606e-24, 3.72097e-24, 2.7906e-24, 5.3935e-24, 1.45733e-23], [268.157, 1.14922e-23, 7.68771e-24, 3.72115e-24, 2.78928e-24, 5.41557e-24, 1.45675e-23], [268.767, 1.15168e-23, 7.68937e-24, 3.72133e-24, 2.78797e-24, 5.43785e-24, 1.45617e-23], [269.378, 1.15414e-23, 7.69104e-24, 3.72151e-24, 2.78667e-24, 5.46036e-24, 1.45558e-23], [269.991, 1.15661e-23, 7.69272e-24, 3.7217e-24, 2.78538e-24, 5.48309e-24, 1.45499e-23], [270.605, 1.15909e-23, 7.69441e-24, 3.72189e-24, 2.7841e-24, 5.50603e-24, 1.4544e-23], [271.221, 1.16157e-23, 7.69611e-24, 3.72209e-24, 2.78284e-24, 5.5292e-24, 1.4538e-23], [271.838, 1.16406e-23, 7.69782e-24, 3.72229e-24, 2.78158e-24, 5.55258e-24, 1.4532e-23], [272.456, 1.16655e-23, 7.69954e-24, 3.7225e-24, 2.78034e-24, 5.57618e-24, 1.4526e-23], [273.076, 1.16906e-23, 7.70126e-24, 3.72271e-24, 2.77911e-24, 5.60001e-24, 1.45199e-23], [273.697, 1.17156e-23, 7.703e-24, 3.72293e-24, 2.77789e-24, 5.62405e-24, 1.45138e-23], [274.319, 1.17407e-23, 7.70474e-24, 3.72315e-24, 2.77668e-24, 5.6483e-24, 1.45077e-23], [274.943, 1.17659e-23, 7.7065e-24, 3.72338e-24, 2.77549e-24, 5.67278e-24, 1.45015e-23], [275.569, 1.17912e-23, 7.70826e-24, 3.72361e-24, 2.77431e-24, 5.69747e-24, 1.44953e-23], [276.195, 1.18165e-23, 7.71003e-24, 3.72385e-24, 2.77314e-24, 5.72238e-24, 1.44891e-23], [276.824, 1.18419e-23, 7.71182e-24, 3.72409e-24, 2.77198e-24, 5.7475e-24, 1.44829e-23], [277.453, 1.18673e-23, 7.71361e-24, 3.72434e-24, 2.77084e-24, 5.77284e-24, 1.44766e-23], [278.084, 1.18928e-23, 7.71541e-24, 3.72459e-24, 2.76971e-24, 5.7984e-24, 1.44702e-23], [278.717, 1.19184e-23, 7.71722e-24, 3.72485e-24, 2.76859e-24, 5.82417e-24, 1.44639e-23], [279.351, 1.1944e-23, 7.71905e-24, 3.72511e-24, 2.76749e-24, 5.85016e-24, 1.44575e-23], [279.986, 1.19697e-23, 7.72088e-24, 3.72538e-24, 2.7664e-24, 5.87636e-24, 1.44511e-23], [280.623, 1.19954e-23, 7.72272e-24, 3.72565e-24, 2.76532e-24, 5.90278e-24, 1.44446e-23], [281.261, 1.20212e-23, 7.72457e-24, 3.72593e-24, 2.76426e-24, 5.92941e-24, 1.44381e-23], [281.901, 1.20471e-23, 7.72643e-24, 3.72621e-24, 2.76321e-24, 5.95625e-24, 1.44316e-23], [282.542, 1.2073e-23, 7.7283e-24, 3.7265e-24, 2.76218e-24, 5.98331e-24, 1.4425e-23], [283.185, 1.2099e-23, 7.73018e-24, 3.72679e-24, 2.76116e-24, 6.01058e-24, 1.44184e-23], [283.829, 1.21251e-23, 7.73207e-24, 3.72709e-24, 2.76016e-24, 6.03806e-24, 1.44118e-23], [284.475, 1.21512e-23, 7.73397e-24, 3.72739e-24, 2.75917e-24, 6.06576e-24, 1.44051e-23], [285.122, 1.21774e-23, 7.73588e-24, 3.7277e-24, 2.7582e-24, 6.09366e-24, 1.43984e-23], [285.77, 1.22036e-23, 7.7378e-24, 3.72802e-24, 2.75724e-24, 6.12178e-24, 1.43917e-23], [286.42, 1.22299e-23, 7.73973e-24, 3.72834e-24, 2.7563e-24, 6.15011e-24, 1.43849e-23], [287.072, 1.22563e-23, 7.74167e-24, 3.72866e-24, 2.75537e-24, 6.17865e-24, 1.43781e-23], [287.725, 1.22827e-23, 7.74362e-24, 3.72899e-24, 2.75446e-24, 6.20739e-24, 1.43713e-23], [288.379, 1.23092e-23, 7.74558e-24, 3.72933e-24, 2.75357e-24, 6.23635e-24, 1.43644e-23], [289.035, 1.23358e-23, 7.74755e-24, 3.72967e-24, 2.75269e-24, 6.26552e-24, 1.43575e-23], [289.693, 1.23624e-23, 7.74954e-24, 3.73001e-24, 2.75183e-24, 6.29489e-24, 1.43505e-23], [290.352, 1.23891e-23, 7.75153e-24, 3.73036e-24, 2.75099e-24, 6.32448e-24, 1.43436e-23], [291.012, 1.24159e-23, 7.75353e-24, 3.73072e-24, 2.75016e-24, 6.35427e-24, 1.43365e-23], [291.674, 1.24427e-23, 7.75554e-24, 3.73108e-24, 2.74935e-24, 6.38427e-24, 1.43295e-23], [292.337, 1.24696e-23, 7.75757e-24, 3.73145e-24, 2.74856e-24, 6.41448e-24, 1.43224e-23], [293.002, 1.24965e-23, 7.7596e-24, 3.73182e-24, 2.74778e-24, 6.44489e-24, 1.43153e-23], [293.669, 1.25235e-23, 7.76164e-24, 3.7322e-24, 2.74702e-24, 6.47551e-24, 1.43081e-23], [294.337, 1.25506e-23, 7.7637e-24, 3.73258e-24, 2.74628e-24, 6.50633e-24, 1.43009e-23], [295.006, 1.25778e-23, 7.76576e-24, 3.73297e-24, 2.74556e-24, 6.53737e-24, 1.42937e-23], [295.677, 1.2605e-23, 7.76784e-24, 3.73337e-24, 2.74486e-24, 6.5686e-24, 1.42864e-23], [296.35, 1.26322e-23, 7.76992e-24, 3.73377e-24, 2.74418e-24, 6.60004e-24, 1.42791e-23], [297.024, 1.26596e-23, 7.77202e-24, 3.73417e-24, 2.74351e-24, 6.63169e-24, 1.42718e-23], [297.699, 1.2687e-23, 7.77413e-24, 3.73458e-24, 2.74286e-24, 6.66354e-24, 1.42644e-23], [298.377, 1.27144e-23, 7.77625e-24, 3.735e-24, 2.74224e-24, 6.69559e-24, 1.4257e-23], [299.055, 1.2742e-23, 7.77838e-24, 3.73542e-24, 2.74163e-24, 6.72785e-24, 1.42496e-23], [299.735, 1.27696e-23, 7.78052e-24, 3.73585e-24, 2.74104e-24, 6.76031e-24, 1.42421e-23], [300.417, 1.27973e-23, 7.78267e-24, 3.73628e-24, 2.74047e-24, 6.79297e-24, 1.42346e-23], [301.101, 1.2825e-23, 7.78483e-24, 3.73672e-24, 2.73992e-24, 6.82583e-24, 1.4227e-23], [301.785, 1.28528e-23, 7.787e-24, 3.73717e-24, 2.7394e-24, 6.85889e-24, 1.42194e-23], [302.472, 1.28807e-23, 7.78919e-24, 3.73762e-24, 2.73889e-24, 6.89216e-24, 1.42118e-23], [303.16, 1.29086e-23, 7.79138e-24, 3.73807e-24, 2.7384e-24, 6.92563e-24, 1.42041e-23], [303.849, 1.29366e-23, 7.79359e-24, 3.73853e-24, 2.73794e-24, 6.95929e-24, 1.41964e-23], [304.541, 1.29647e-23, 7.7958e-24, 3.739e-24, 2.73749e-24, 6.99316e-24, 1.41886e-23], [305.233, 1.29928e-23, 7.79803e-24, 3.73947e-24, 2.73707e-24, 7.02723e-24, 1.41809e-23], [305.928, 1.3021e-23, 7.80027e-24, 3.73995e-24, 2.73667e-24, 7.06149e-24, 1.4173e-23], [306.623, 1.30493e-23, 7.80252e-24, 3.74044e-24, 2.73629e-24, 7.09596e-24, 1.41652e-23], [307.321, 1.30776e-23, 7.80479e-24, 3.74093e-24, 2.73593e-24, 7.13062e-24, 1.41573e-23], [308.02, 1.3106e-23, 7.80706e-24, 3.74142e-24, 2.73559e-24, 7.16548e-24, 1.41493e-23], [308.72, 1.31345e-23, 7.80934e-24, 3.74193e-24, 2.73528e-24, 7.20054e-24, 1.41414e-23], [309.423, 1.3163e-23, 7.81164e-24, 3.74243e-24, 2.73499e-24, 7.2358e-24, 1.41333e-23], [310.126, 1.31916e-23, 7.81395e-24, 3.74295e-24, 2.73472e-24, 7.27125e-24, 1.41253e-23], [310.832, 1.32203e-23, 7.81627e-24, 3.74347e-24, 2.73448e-24, 7.3069e-24, 1.41172e-23], [311.539, 1.32491e-23, 7.8186e-24, 3.74399e-24, 2.73426e-24, 7.34275e-24, 1.41091e-23], [312.248, 1.32779e-23, 7.82094e-24, 3.74452e-24, 2.73406e-24, 7.37879e-24, 1.41009e-23], [312.958, 1.33068e-23, 7.8233e-24, 3.74506e-24, 2.73389e-24, 7.41503e-24, 1.40927e-23], [313.67, 1.33357e-23, 7.82566e-24, 3.7456e-24, 2.73374e-24, 7.45146e-24, 1.40844e-23], [314.383, 1.33647e-23, 7.82804e-24, 3.74615e-24, 2.73362e-24, 7.48809e-24, 1.40761e-23], [315.098, 1.33938e-23, 7.83043e-24, 3.74671e-24, 2.73352e-24, 7.52491e-24, 1.40678e-23], [315.815, 1.3423e-23, 7.83283e-24, 3.74727e-24, 2.73344e-24, 7.56193e-24, 1.40594e-23], [316.533, 1.34522e-23, 7.83524e-24, 3.74784e-24, 2.73339e-24, 7.59914e-24, 1.4051e-23], [317.253, 1.34815e-23, 7.83767e-24, 3.74841e-24, 2.73337e-24, 7.63654e-24, 1.40426e-23], [317.975, 1.35109e-23, 7.84011e-24, 3.74899e-24, 2.73337e-24, 7.67414e-24, 1.40341e-23], [318.698, 1.35403e-23, 7.84256e-24, 3.74957e-24, 2.7334e-24, 7.71193e-24, 1.40256e-23], [319.423, 1.35698e-23, 7.84502e-24, 3.75016e-24, 2.73345e-24, 7.74992e-24, 1.4017e-23], [320.149, 1.35994e-23, 7.84749e-24, 3.75076e-24, 2.73353e-24, 7.78809e-24, 1.40084e-23], [320.878, 1.36291e-23, 7.84998e-24, 3.75136e-24, 2.73364e-24, 7.82646e-24, 1.39998e-23], [321.608, 1.36588e-23, 7.85247e-24, 3.75197e-24, 2.73377e-24, 7.86502e-24, 1.39911e-23], [322.339, 1.36886e-23, 7.85498e-24, 3.75259e-24, 2.73393e-24, 7.90377e-24, 1.39823e-23], [323.072, 1.37184e-23, 7.8575e-24, 3.75321e-24, 2.73412e-24, 7.94271e-24, 1.39736e-23], [323.807, 1.37484e-23, 7.86004e-24, 3.75384e-24, 2.73433e-24, 7.98184e-24, 1.39647e-23], [324.544, 1.37784e-23, 7.86259e-24, 3.75447e-24, 2.73457e-24, 8.02116e-24, 1.39559e-23], [325.282, 1.38084e-23, 7.86514e-24, 3.75512e-24, 2.73484e-24, 8.06068e-24, 1.3947e-23], [326.022, 1.38386e-23, 7.86772e-24, 3.75576e-24, 2.73514e-24, 8.10038e-24, 1.39381e-23], [326.763, 1.38688e-23, 7.8703e-24, 3.75642e-24, 2.73547e-24, 8.14027e-24, 1.39291e-23], [327.506, 1.38991e-23, 7.8729e-24, 3.75707e-24, 2.73583e-24, 8.18035e-24, 1.39201e-23], [328.251, 1.39295e-23, 7.87551e-24, 3.75774e-24, 2.73621e-24, 8.22062e-24, 1.3911e-23], [328.998, 1.39599e-23, 7.87813e-24, 3.75841e-24, 2.73663e-24, 8.26108e-24, 1.39019e-23], [329.746, 1.39904e-23, 7.88076e-24, 3.75909e-24, 2.73707e-24, 8.30173e-24, 1.38927e-23], [330.496, 1.4021e-23, 7.88341e-24, 3.75978e-24, 2.73754e-24, 8.34256e-24, 1.38835e-23], [331.248, 1.40516e-23, 7.88607e-24, 3.76047e-24, 2.73805e-24, 8.38358e-24, 1.38743e-23], [332.002, 1.40823e-23, 7.88874e-24, 3.76116e-24, 2.73858e-24, 8.4248e-24, 1.3865e-23], [332.757, 1.41131e-23, 7.89143e-24, 3.76187e-24, 2.73914e-24, 8.46619e-24, 1.38557e-23], [333.514, 1.4144e-23, 7.89413e-24, 3.76258e-24, 2.73974e-24, 8.50778e-24, 1.38464e-23], [334.272, 1.41749e-23, 7.89684e-24, 3.7633e-24, 2.74036e-24, 8.54955e-24, 1.3837e-23], [335.033, 1.4206e-23, 7.89956e-24, 3.76402e-24, 2.74102e-24, 8.59151e-24, 1.38275e-23], [335.795, 1.4237e-23, 7.9023e-24, 3.76475e-24, 2.74171e-24, 8.63365e-24, 1.3818e-23], [336.558, 1.42682e-23, 7.90505e-24, 3.76549e-24, 2.74243e-24, 8.67598e-24, 1.38085e-23], [337.324, 1.42994e-23, 7.90781e-24, 3.76623e-24, 2.74318e-24, 8.7185e-24, 1.37989e-23], [338.091, 1.43308e-23, 7.91059e-24, 3.76698e-24, 2.74397e-24, 8.7612e-24, 1.37893e-23], [338.86, 1.43621e-23, 7.91338e-24, 3.76774e-24, 2.74478e-24, 8.80409e-24, 1.37796e-23], [339.631, 1.43936e-23, 7.91618e-24, 3.7685e-24, 2.74563e-24, 8.84716e-24, 1.37699e-23], [340.403, 1.44251e-23, 7.919e-24, 3.76927e-24, 2.74652e-24, 8.89042e-24, 1.37601e-23], [341.178, 1.44567e-23, 7.92183e-24, 3.77005e-24, 2.74743e-24, 8.93386e-24, 1.37503e-23], [341.954, 1.44884e-23, 7.92467e-24, 3.77083e-24, 2.74838e-24, 8.97749e-24, 1.37405e-23], [342.732, 1.45202e-23, 7.92753e-24, 3.77162e-24, 2.74936e-24, 9.0213e-24, 1.37306e-23], [343.511, 1.4552e-23, 7.9304e-24, 3.77241e-24, 2.75038e-24, 9.0653e-24, 1.37207e-23], [344.293, 1.45839e-23, 7.93329e-24, 3.77322e-24, 2.75143e-24, 9.10948e-24, 1.37107e-23], [345.076, 1.46159e-23, 7.93618e-24, 3.77403e-24, 2.75252e-24, 9.15384e-24, 1.37007e-23], [345.861, 1.46479e-23, 7.93909e-24, 3.77484e-24, 2.75364e-24, 9.19838e-24, 1.36906e-23], [346.647, 1.46801e-23, 7.94202e-24, 3.77567e-24, 2.7548e-24, 9.24311e-24, 1.36805e-23], [347.436, 1.47123e-23, 7.94496e-24, 3.7765e-24, 2.75599e-24, 9.28803e-24, 1.36703e-23], [348.226, 1.47446e-23, 7.94791e-24, 3.77734e-24, 2.75721e-24, 9.33312e-24, 1.36601e-23], [349.018, 1.47769e-23, 7.95088e-24, 3.77818e-24, 2.75848e-24, 9.3784e-24, 1.36499e-23], [349.812, 1.48094e-23, 7.95386e-24, 3.77903e-24, 2.75978e-24, 9.42386e-24, 1.36396e-23], [350.608, 1.48419e-23, 7.95685e-24, 3.77989e-24, 2.76111e-24, 9.4695e-24, 1.36292e-23], [351.405, 1.48745e-23, 7.95986e-24, 3.78075e-24, 2.76248e-24, 9.51533e-24, 1.36189e-23], [352.204, 1.49071e-23, 7.96288e-24, 3.78163e-24, 2.76389e-24, 9.56133e-24, 1.36084e-23], [353.005, 1.49399e-23, 7.96592e-24, 3.7825e-24, 2.76534e-24, 9.60752e-24, 1.35979e-23], [353.808, 1.49727e-23, 7.96897e-24, 3.78339e-24, 2.76682e-24, 9.65389e-24, 1.35874e-23], [354.613, 1.50056e-23, 7.97204e-24, 3.78428e-24, 2.76834e-24, 9.70044e-24, 1.35768e-23], [355.42, 1.50386e-23, 7.97512e-24, 3.78518e-24, 2.7699e-24, 9.74717e-24, 1.35662e-23], [356.228, 1.50716e-23, 7.97821e-24, 3.78609e-24, 2.77149e-24, 9.79409e-24, 1.35555e-23], [357.038, 1.51048e-23, 7.98132e-24, 3.78701e-24, 2.77313e-24, 9.84118e-24, 1.35448e-23], [357.851, 1.5138e-23, 7.98444e-24, 3.78793e-24, 2.7748e-24, 9.88846e-24, 1.35341e-23], [358.664, 1.51713e-23, 7.98758e-24, 3.78886e-24, 2.77651e-24, 9.93591e-24, 1.35232e-23], [359.48, 1.52046e-23, 7.99073e-24, 3.78979e-24, 2.77826e-24, 9.98355e-24, 1.35124e-23], [360.298, 1.52381e-23, 7.9939e-24, 3.79073e-24, 2.78005e-24, 1.00314e-23, 1.35015e-23], [361.117, 1.52716e-23, 7.99708e-24, 3.79168e-24, 2.78188e-24, 1.00794e-23, 1.34905e-23], [361.939, 1.53052e-23, 8.00027e-24, 3.79264e-24, 2.78375e-24, 1.01275e-23, 1.34795e-23], [362.762, 1.53389e-23, 8.00348e-24, 3.79361e-24, 2.78566e-24, 1.01759e-23, 1.34684e-23], [363.587, 1.53727e-23, 8.00671e-24, 3.79458e-24, 2.78761e-24, 1.02244e-23, 1.34573e-23], [364.414, 1.54065e-23, 8.00995e-24, 3.79556e-24, 2.7896e-24, 1.02731e-23, 1.34462e-23], [365.243, 1.54404e-23, 8.0132e-24, 3.79654e-24, 2.79163e-24, 1.0322e-23, 1.3435e-23], [366.074, 1.54744e-23, 8.01647e-24, 3.79754e-24, 2.7937e-24, 1.03711e-23, 1.34237e-23], [366.907, 1.55085e-23, 8.01976e-24, 3.79854e-24, 2.79582e-24, 1.04204e-23, 1.34124e-23], [367.741, 1.55426e-23, 8.02306e-24, 3.79955e-24, 2.79797e-24, 1.04698e-23, 1.34011e-23], [368.578, 1.55769e-23, 8.02637e-24, 3.80056e-24, 2.80017e-24, 1.05194e-23, 1.33897e-23], [369.416, 1.56112e-23, 8.0297e-24, 3.80159e-24, 2.80241e-24, 1.05692e-23, 1.33783e-23], [370.256, 1.56456e-23, 8.03305e-24, 3.80262e-24, 2.80469e-24, 1.06192e-23, 1.33668e-23], [371.098, 1.56801e-23, 8.03641e-24, 3.80366e-24, 2.80701e-24, 1.06693e-23, 1.33552e-23], [371.942, 1.57147e-23, 8.03979e-24, 3.8047e-24, 2.80938e-24, 1.07196e-23, 1.33436e-23], [372.788, 1.57493e-23, 8.04318e-24, 3.80576e-24, 2.81179e-24, 1.07702e-23, 1.3332e-23], [373.636, 1.5784e-23, 8.04659e-24, 3.80682e-24, 2.81424e-24, 1.08208e-23, 1.33203e-23], [374.486, 1.58188e-23, 8.05001e-24, 3.80789e-24, 2.81674e-24, 1.08717e-23, 1.33085e-23], [375.338, 1.58537e-23, 8.05345e-24, 3.80896e-24, 2.81928e-24, 1.09227e-23, 1.32967e-23], [376.192, 1.58887e-23, 8.0569e-24, 3.81005e-24, 2.82186e-24, 1.0974e-23, 1.32849e-23], [377.047, 1.59237e-23, 8.06037e-24, 3.81114e-24, 2.82449e-24, 1.10254e-23, 1.3273e-23], [377.905, 1.59589e-23, 8.06385e-24, 3.81224e-24, 2.82716e-24, 1.10769e-23, 1.3261e-23], [378.765, 1.59941e-23, 8.06735e-24, 3.81335e-24, 2.82987e-24, 1.11287e-23, 1.3249e-23], [379.626, 1.60294e-23, 8.07087e-24, 3.81446e-24, 2.83263e-24, 1.11806e-23, 1.32369e-23], [380.49, 1.60648e-23, 8.0744e-24, 3.81558e-24, 2.83544e-24, 1.12327e-23, 1.32248e-23], [381.355, 1.61002e-23, 8.07795e-24, 3.81671e-24, 2.83829e-24, 1.1285e-23, 1.32127e-23], [382.223, 1.61358e-23, 8.08152e-24, 3.81785e-24, 2.84119e-24, 1.13375e-23, 1.32005e-23], [383.092, 1.61714e-23, 8.0851e-24, 3.819e-24, 2.84413e-24, 1.13901e-23, 1.31882e-23], [383.963, 1.62071e-23, 8.08869e-24, 3.82015e-24, 2.84712e-24, 1.14429e-23, 1.31759e-23], [384.837, 1.62429e-23, 8.09231e-24, 3.82132e-24, 2.85016e-24, 1.14959e-23, 1.31635e-23], [385.712, 1.62788e-23, 8.09594e-24, 3.82249e-24, 2.85324e-24, 1.15491e-23, 1.31511e-23], [386.589, 1.63148e-23, 8.09958e-24, 3.82366e-24, 2.85636e-24, 1.16024e-23, 1.31386e-23], [387.469, 1.63508e-23, 8.10324e-24, 3.82485e-24, 2.85954e-24, 1.1656e-23, 1.31261e-23], [388.35, 1.63869e-23, 8.10692e-24, 3.82604e-24, 2.86276e-24, 1.17097e-23, 1.31135e-23], [389.233, 1.64232e-23, 8.11062e-24, 3.82725e-24, 2.86603e-24, 1.17635e-23, 1.31008e-23], [390.119, 1.64595e-23, 8.11433e-24, 3.82846e-24, 2.86934e-24, 1.18176e-23, 1.30881e-23], [391.006, 1.64959e-23, 8.11805e-24, 3.82967e-24, 2.87271e-24, 1.18718e-23, 1.30754e-23], [391.895, 1.65323e-23, 8.1218e-24, 3.8309e-24, 2.87612e-24, 1.19262e-23, 1.30626e-23], [392.787, 1.65689e-23, 8.12556e-24, 3.83213e-24, 2.87958e-24, 1.19808e-23, 1.30498e-23], [393.68, 1.66055e-23, 8.12934e-24, 3.83338e-24, 2.88309e-24, 1.20356e-23, 1.30368e-23], [394.576, 1.66423e-23, 8.13313e-24, 3.83463e-24, 2.88665e-24, 1.20905e-23, 1.30239e-23], [395.473, 1.66791e-23, 8.13694e-24, 3.83589e-24, 2.89025e-24, 1.21456e-23, 1.30109e-23], [396.373, 1.6716e-23, 8.14077e-24, 3.83715e-24, 2.89391e-24, 1.22009e-23, 1.29978e-23], [397.274, 1.6753e-23, 8.14462e-24, 3.83843e-24, 2.89761e-24, 1.22564e-23, 1.29847e-23], [398.178, 1.679e-23, 8.14848e-24, 3.83971e-24, 2.90136e-24, 1.2312e-23, 1.29715e-23], [399.084, 1.68272e-23, 8.15236e-24, 3.841e-24, 2.90516e-24, 1.23679e-23, 1.29582e-23], [399.991, 1.68645e-23, 8.15626e-24, 3.84231e-24, 2.90902e-24, 1.24239e-23, 1.29449e-23], [400.901, 1.69018e-23, 8.16017e-24, 3.84361e-24, 2.91292e-24, 1.248e-23, 1.29316e-23], [401.813, 1.69392e-23, 8.1641e-24, 3.84493e-24, 2.91687e-24, 1.25364e-23, 1.29182e-23], [402.727, 1.69767e-23, 8.16805e-24, 3.84626e-24, 2.92087e-24, 1.25929e-23, 1.29047e-23], [403.643, 1.70143e-23, 8.17202e-24, 3.84759e-24, 2.92492e-24, 1.26496e-23, 1.28912e-23], [404.561, 1.7052e-23, 8.176e-24, 3.84893e-24, 2.92903e-24, 1.27065e-23, 1.28776e-23], [405.481, 1.70898e-23, 8.18e-24, 3.85029e-24, 2.93318e-24, 1.27635e-23, 1.2864e-23], [406.404, 1.71276e-23, 8.18402e-24, 3.85165e-24, 2.93739e-24, 1.28208e-23, 1.28503e-23], [407.328, 1.71656e-23, 8.18806e-24, 3.85301e-24, 2.94164e-24, 1.28782e-23, 1.28366e-23], [408.254, 1.72036e-23, 8.19211e-24, 3.85439e-24, 2.94595e-24, 1.29358e-23, 1.28228e-23], [409.183, 1.72417e-23, 8.19618e-24, 3.85578e-24, 2.95031e-24, 1.29935e-23, 1.28089e-23], [410.114, 1.728e-23, 8.20027e-24, 3.85717e-24, 2.95472e-24, 1.30515e-23, 1.2795e-23], [411.047, 1.73183e-23, 8.20438e-24, 3.85857e-24, 2.95919e-24, 1.31096e-23, 1.2781e-23], [411.982, 1.73567e-23, 8.20851e-24, 3.85999e-24, 2.9637e-24, 1.31679e-23, 1.2767e-23], [412.919, 1.73951e-23, 8.21265e-24, 3.86141e-24, 2.96827e-24, 1.32263e-23, 1.27529e-23], [413.858, 1.74337e-23, 8.21681e-24, 3.86284e-24, 2.97289e-24, 1.3285e-23, 1.27387e-23], [414.799, 1.74724e-23, 8.22099e-24, 3.86428e-24, 2.97756e-24, 1.33438e-23, 1.27245e-23], [415.743, 1.75111e-23, 8.22519e-24, 3.86572e-24, 2.98229e-24, 1.34028e-23, 1.27103e-23], [416.688, 1.755e-23, 8.22941e-24, 3.86718e-24, 2.98707e-24, 1.34619e-23, 1.26959e-23], [417.636, 1.75889e-23, 8.23364e-24, 3.86864e-24, 2.9919e-24, 1.35213e-23, 1.26815e-23], [418.586, 1.76279e-23, 8.23789e-24, 3.87012e-24, 2.99679e-24, 1.35808e-23, 1.26671e-23], [419.538, 1.76671e-23, 8.24217e-24, 3.8716e-24, 3.00172e-24, 1.36405e-23, 1.26526e-23], [420.492, 1.77063e-23, 8.24646e-24, 3.87309e-24, 3.00672e-24, 1.37004e-23, 1.2638e-23], [421.449, 1.77456e-23, 8.25077e-24, 3.8746e-24, 3.01176e-24, 1.37604e-23, 1.26234e-23], [422.407, 1.77849e-23, 8.25509e-24, 3.87611e-24, 3.01686e-24, 1.38207e-23, 1.26087e-23], [423.368, 1.78244e-23, 8.25944e-24, 3.87763e-24, 3.02202e-24, 1.38811e-23, 1.2594e-23], [424.331, 1.7864e-23, 8.26381e-24, 3.87916e-24, 3.02723e-24, 1.39416e-23, 1.25792e-23], [425.296, 1.79037e-23, 8.26819e-24, 3.88069e-24, 3.03249e-24, 1.40024e-23, 1.25643e-23], [426.264, 1.79434e-23, 8.27259e-24, 3.88224e-24, 3.03781e-24, 1.40633e-23, 1.25494e-23], [427.233, 1.79833e-23, 8.27702e-24, 3.8838e-24, 3.04318e-24, 1.41244e-23, 1.25344e-23], [428.205, 1.80232e-23, 8.28146e-24, 3.88537e-24, 3.04861e-24, 1.41857e-23, 1.25194e-23], [429.179, 1.80633e-23, 8.28592e-24, 3.88694e-24, 3.05409e-24, 1.42472e-23, 1.25042e-23], [430.155, 1.81034e-23, 8.2904e-24, 3.88853e-24, 3.05963e-24, 1.43088e-23, 1.24891e-23], [431.134, 1.81436e-23, 8.2949e-24, 3.89012e-24, 3.06523e-24, 1.43706e-23, 1.24738e-23], [432.114, 1.81839e-23, 8.29942e-24, 3.89173e-24, 3.07088e-24, 1.44326e-23, 1.24586e-23], [433.097, 1.82243e-23, 8.30396e-24, 3.89334e-24, 3.07658e-24, 1.44948e-23, 1.24432e-23], [434.082, 1.82648e-23, 8.30851e-24, 3.89497e-24, 3.08234e-24, 1.45571e-23, 1.24278e-23], [435.07, 1.83054e-23, 8.31309e-24, 3.8966e-24, 3.08816e-24, 1.46196e-23, 1.24123e-23], [436.059, 1.83461e-23, 8.31769e-24, 3.89824e-24, 3.09403e-24, 1.46823e-23, 1.23968e-23], [437.051, 1.83869e-23, 8.32231e-24, 3.8999e-24, 3.09997e-24, 1.47452e-23, 1.23812e-23], [438.045, 1.84278e-23, 8.32695e-24, 3.90156e-24, 3.10595e-24, 1.48083e-23, 1.23655e-23], [439.042, 1.84688e-23, 8.3316e-24, 3.90324e-24, 3.112e-24, 1.48715e-23, 1.23498e-23], [440.04, 1.85098e-23, 8.33628e-24, 3.90492e-24, 3.1181e-24, 1.49349e-23, 1.2334e-23], [441.041, 1.8551e-23, 8.34098e-24, 3.90662e-24, 3.12426e-24, 1.49985e-23, 1.23181e-23], [442.044, 1.85923e-23, 8.3457e-24, 3.90833e-24, 3.13047e-24, 1.50622e-23, 1.23022e-23], [443.05, 1.86336e-23, 8.35044e-24, 3.91004e-24, 3.13675e-24, 1.51261e-23, 1.22862e-23], [444.058, 1.86751e-23, 8.3552e-24, 3.91177e-24, 3.14308e-24, 1.51903e-23, 1.22702e-23], [445.068, 1.87166e-23, 8.35998e-24, 3.91351e-24, 3.14947e-24, 1.52545e-23, 1.22541e-23], [446.08, 1.87583e-23, 8.36478e-24, 3.91526e-24, 3.15592e-24, 1.5319e-23, 1.22379e-23], [447.095, 1.88e-23, 8.3696e-24, 3.91702e-24, 3.16243e-24, 1.53836e-23, 1.22216e-23], [448.111, 1.88419e-23, 8.37445e-24, 3.91879e-24, 3.169e-24, 1.54485e-23, 1.22053e-23], [449.131, 1.88838e-23, 8.37931e-24, 3.92058e-24, 3.17563e-24, 1.55135e-23, 1.2189e-23], [450.152, 1.89259e-23, 8.3842e-24, 3.92238e-24, 3.18232e-24, 1.55786e-23, 1.21725e-23], [451.176, 1.8968e-23, 8.38911e-24, 3.92419e-24, 3.18906e-24, 1.5644e-23, 1.2156e-23], [452.202, 1.90102e-23, 8.39404e-24, 3.92601e-24, 3.19588e-24, 1.57095e-23, 1.21395e-23], [453.231, 1.90526e-23, 8.399e-24, 3.92785e-24, 3.20275e-24, 1.57752e-23, 1.21229e-23], [454.262, 1.9095e-23, 8.40397e-24, 3.9297e-24, 3.20969e-24, 1.58411e-23, 1.21062e-23], [455.295, 1.91375e-23, 8.40897e-24, 3.93157e-24, 3.21669e-24, 1.59072e-23, 1.20894e-23], [456.331, 1.91802e-23, 8.414e-24, 3.93346e-24, 3.22375e-24, 1.59734e-23, 1.20726e-23], [457.369, 1.92229e-23, 8.41905e-24, 3.93536e-24, 3.23089e-24, 1.60398e-23, 1.20557e-23], [458.409, 1.92658e-23, 8.42412e-24, 3.93728e-24, 3.23809e-24, 1.61064e-23, 1.20388e-23], [459.452, 1.93087e-23, 8.42922e-24, 3.93923e-24, 3.24536e-24, 1.61732e-23, 1.20217e-23], [460.497, 1.93517e-23, 8.43435e-24, 3.94119e-24, 3.25271e-24, 1.62402e-23, 1.20047e-23], [461.544, 1.93949e-23, 8.43951e-24, 3.94319e-24, 3.26013e-24, 1.63073e-23, 1.19875e-23], [462.594, 1.94381e-23, 8.4447e-24, 3.94521e-24, 3.26764e-24, 1.63747e-23, 1.19703e-23], [463.646, 1.94815e-23, 8.44992e-24, 3.94727e-24, 3.27524e-24, 1.64422e-23, 1.19531e-23], [464.701, 1.95249e-23, 8.45518e-24, 3.94938e-24, 3.28293e-24, 1.65099e-23, 1.19358e-23], [465.758, 1.95685e-23, 8.46048e-24, 3.95154e-24, 3.29074e-24, 1.65777e-23, 1.19184e-23], [466.817, 1.96122e-23, 8.46583e-24, 3.95377e-24, 3.29867e-24, 1.66458e-23, 1.1901e-23], [467.879, 1.9656e-23, 8.47123e-24, 3.95608e-24, 3.30675e-24, 1.67141e-23, 1.18835e-23], [468.943, 1.96999e-23, 8.47671e-24, 3.95851e-24, 3.31502e-24, 1.67826e-23, 1.18661e-23], [470.01, 1.97439e-23, 8.48229e-24, 3.96111e-24, 3.32353e-24, 1.68513e-23, 1.18486e-23], [471.079, 1.97881e-23, 8.48799e-24, 3.96396e-24, 3.33237e-24, 1.69202e-23, 1.18311e-23], [472.15, 1.98325e-23, 8.49388e-24, 3.96716e-24, 3.34169e-24, 1.69894e-23, 1.18136e-23], [473.224, 1.98771e-23, 8.50007e-24, 3.97097e-24, 3.35175e-24, 1.70589e-23, 1.17963e-23], [474.301, 1.9922e-23, 8.50677e-24, 3.97583e-24, 3.3631e-24, 1.71288e-23, 1.17793e-23], [475.38, 1.99674e-23, 8.51443e-24, 3.98271e-24, 3.37687e-24, 1.71993e-23, 1.17629e-23], [476.461, 2.00138e-23, 8.52423e-24, 3.99412e-24, 3.39599e-24, 1.72711e-23, 1.17479e-23], [477.545, 2.00628e-23, 8.53996e-24, 4.01809e-24, 3.42982e-24, 1.73459e-23, 1.17372e-23], [478.631, 2.01223e-23, 8.58017e-24, 4.0935e-24, 3.52345e-24, 1.74329e-23, 1.17443e-23], [479.72, 2.0307e-23, 8.90924e-24, 4.73656e-24, 4.25826e-24, 1.7664e-23, 1.1965e-23], [480.811, 2.05924e-23, 9.45131e-24, 5.68421e-24, 5.29606e-24, 1.80094e-23, 1.23528e-23], [481.904, 2.02644e-23, 8.61397e-24, 4.13592e-24, 3.5895e-24, 1.76521e-23, 1.17021e-23], [483.0, 2.02887e-23, 8.57031e-24, 4.03453e-24, 3.47811e-24, 1.76988e-23, 1.16474e-23], [484.099, 2.03287e-23, 8.56328e-24, 4.00982e-24, 3.45544e-24, 1.77635e-23, 1.16195e-23], [485.2, 2.03719e-23, 8.56388e-24, 4.00126e-24, 3.45159e-24, 1.78322e-23, 1.15971e-23], [486.304, 2.04163e-23, 8.56702e-24, 3.99809e-24, 3.45408e-24, 1.79022e-23, 1.15766e-23], [487.41, 2.04613e-23, 8.57126e-24, 3.99724e-24, 3.45931e-24, 1.79729e-23, 1.15567e-23], [488.519, 2.05067e-23, 8.57606e-24, 3.99753e-24, 3.46592e-24, 1.80441e-23, 1.15371e-23], [489.63, 2.05522e-23, 8.58117e-24, 3.99846e-24, 3.47331e-24, 1.81156e-23, 1.15177e-23], [490.743, 2.05979e-23, 8.58648e-24, 3.99978e-24, 3.48121e-24, 1.81874e-23, 1.14983e-23], [491.86, 2.06438e-23, 8.59192e-24, 4.00134e-24, 3.48944e-24, 1.82594e-23, 1.1479e-23], [492.978, 2.06898e-23, 8.59746e-24, 4.00307e-24, 3.4979e-24, 1.83316e-23, 1.14596e-23], [494.1, 2.0736e-23, 8.60307e-24, 4.00492e-24, 3.50656e-24, 1.8404e-23, 1.14402e-23], [495.224, 2.07823e-23, 8.60874e-24, 4.00686e-24, 3.51536e-24, 1.84766e-23, 1.14208e-23], [496.35, 2.08286e-23, 8.61446e-24, 4.00886e-24, 3.52428e-24, 1.85494e-23, 1.14012e-23], [497.479, 2.08751e-23, 8.62022e-24, 4.01092e-24, 3.53332e-24, 1.86224e-23, 1.13817e-23], [498.611, 2.09218e-23, 8.62602e-24, 4.01302e-24, 3.54244e-24, 1.86955e-23, 1.1362e-23], [499.745, 2.09685e-23, 8.63186e-24, 4.01515e-24, 3.55166e-24, 1.87689e-23, 1.13423e-23], [500.881, 2.10153e-23, 8.63773e-24, 4.01732e-24, 3.56096e-24, 1.88425e-23, 1.13226e-23], [502.021, 2.10623e-23, 8.64363e-24, 4.01951e-24, 3.57033e-24, 1.89162e-23, 1.13027e-23], [503.163, 2.11093e-23, 8.64956e-24, 4.02172e-24, 3.57978e-24, 1.89901e-23, 1.12828e-23], [504.307, 2.11565e-23, 8.65552e-24, 4.02396e-24, 3.5893e-24, 1.90642e-23, 1.12628e-23], [505.454, 2.12038e-23, 8.66151e-24, 4.02622e-24, 3.59889e-24, 1.91385e-23, 1.12428e-23], [506.604, 2.12512e-23, 8.66752e-24, 4.02849e-24, 3.60855e-24, 1.9213e-23, 1.12227e-23], [507.756, 2.12987e-23, 8.67356e-24, 4.03078e-24, 3.61827e-24, 1.92876e-23, 1.12025e-23], [508.911, 2.13463e-23, 8.67963e-24, 4.03309e-24, 3.62806e-24, 1.93625e-23, 1.11822e-23], [510.069, 2.1394e-23, 8.68572e-24, 4.03542e-24, 3.63791e-24, 1.94375e-23, 1.11618e-23], [511.229, 2.14419e-23, 8.69184e-24, 4.03775e-24, 3.64783e-24, 1.95127e-23, 1.11414e-23], [512.392, 2.14898e-23, 8.69799e-24, 4.04011e-24, 3.65781e-24, 1.95881e-23, 1.11209e-23], [513.557, 2.15379e-23, 8.70416e-24, 4.04248e-24, 3.66785e-24, 1.96636e-23, 1.11003e-23], [514.725, 2.15861e-23, 8.71036e-24, 4.04486e-24, 3.67796e-24, 1.97394e-23, 1.10797e-23], [515.896, 2.16343e-23, 8.71658e-24, 4.04725e-24, 3.68812e-24, 1.98153e-23, 1.1059e-23], [517.069, 2.16827e-23, 8.72283e-24, 4.04966e-24, 3.69835e-24, 1.98914e-23, 1.10382e-23], [518.246, 2.17312e-23, 8.72911e-24, 4.05208e-24, 3.70864e-24, 1.99677e-23, 1.10173e-23], [519.424, 2.17799e-23, 8.73541e-24, 4.05452e-24, 3.71899e-24, 2.00442e-23, 1.09963e-23], [520.606, 2.18286e-23, 8.74173e-24, 4.05697e-24, 3.7294e-24, 2.01209e-23, 1.09753e-23], [521.79, 2.18774e-23, 8.74809e-24, 4.05943e-24, 3.73987e-24, 2.01977e-23, 1.09542e-23], [522.977, 2.19264e-23, 8.75446e-24, 4.0619e-24, 3.7504e-24, 2.02747e-23, 1.0933e-23], [524.166, 2.19754e-23, 8.76087e-24, 4.06439e-24, 3.76099e-24, 2.0352e-23, 1.09117e-23], [525.359, 2.20246e-23, 8.7673e-24, 4.06689e-24, 3.77165e-24, 2.04294e-23, 1.08904e-23], [526.554, 2.20739e-23, 8.77375e-24, 4.0694e-24, 3.78236e-24, 2.05069e-23, 1.08689e-23], [527.751, 2.21233e-23, 8.78023e-24, 4.07192e-24, 3.79313e-24, 2.05847e-23, 1.08474e-23], [528.952, 2.21729e-23, 8.78674e-24, 4.07446e-24, 3.80397e-24, 2.06627e-23, 1.08258e-23], [530.155, 2.22225e-23, 8.79327e-24, 4.07701e-24, 3.81486e-24, 2.07408e-23, 1.08042e-23], [531.361, 2.22722e-23, 8.79983e-24, 4.07957e-24, 3.82582e-24, 2.08191e-23, 1.07824e-23], [532.569, 2.23221e-23, 8.80642e-24, 4.08215e-24, 3.83683e-24, 2.08976e-23, 1.07606e-23], [533.781, 2.23721e-23, 8.81303e-24, 4.08474e-24, 3.8479e-24, 2.09763e-23, 1.07387e-23], [534.995, 2.24222e-23, 8.81967e-24, 4.08734e-24, 3.85904e-24, 2.10552e-23, 1.07167e-23], [536.212, 2.24724e-23, 8.82633e-24, 4.08995e-24, 3.87023e-24, 2.11343e-23, 1.06946e-23], [537.431, 2.25227e-23, 8.83302e-24, 4.09258e-24, 3.88148e-24, 2.12135e-23, 1.06725e-23], [538.654, 2.25732e-23, 8.83974e-24, 4.09521e-24, 3.8928e-24, 2.12929e-23, 1.06503e-23], [539.879, 2.26237e-23, 8.84648e-24, 4.09786e-24, 3.90417e-24, 2.13725e-23, 1.0628e-23], [541.107, 2.26744e-23, 8.85325e-24, 4.10053e-24, 3.9156e-24, 2.14523e-23, 1.06056e-23], [542.338, 2.27252e-23, 8.86005e-24, 4.1032e-24, 3.9271e-24, 2.15323e-23, 1.05831e-23], [543.571, 2.27761e-23, 8.86688e-24, 4.10589e-24, 3.93865e-24, 2.16125e-23, 1.05606e-23], [544.808, 2.28271e-23, 8.87373e-24, 4.1086e-24, 3.95026e-24, 2.16929e-23, 1.05379e-23], [546.047, 2.28783e-23, 8.8806e-24, 4.11131e-24, 3.96193e-24, 2.17734e-23, 1.05152e-23], [547.289, 2.29296e-23, 8.88751e-24, 4.11404e-24, 3.97366e-24, 2.18541e-23, 1.04924e-23], [548.534, 2.29809e-23, 8.89444e-24, 4.11678e-24, 3.98545e-24, 2.19351e-23, 1.04695e-23], [549.781, 2.30324e-23, 8.9014e-24, 4.11953e-24, 3.9973e-24, 2.20162e-23, 1.04466e-23], [551.032, 2.30841e-23, 8.90838e-24, 4.1223e-24, 4.00921e-24, 2.20975e-23, 1.04235e-23], [552.285, 2.31358e-23, 8.9154e-24, 4.12507e-24, 4.02117e-24, 2.2179e-23, 1.04004e-23], [553.542, 2.31877e-23, 8.92244e-24, 4.12787e-24, 4.0332e-24, 2.22606e-23, 1.03772e-23], [554.801, 2.32396e-23, 8.9295e-24, 4.13067e-24, 4.04529e-24, 2.23425e-23, 1.03539e-23], [556.063, 2.32917e-23, 8.9366e-24, 4.13349e-24, 4.05743e-24, 2.24245e-23, 1.03305e-23], [557.327, 2.33439e-23, 8.94372e-24, 4.13632e-24, 4.06964e-24, 2.25067e-23, 1.0307e-23], [558.595, 2.33963e-23, 8.95087e-24, 4.13916e-24, 4.0819e-24, 2.25892e-23, 1.02835e-23], [559.866, 2.34487e-23, 8.95805e-24, 4.14202e-24, 4.09423e-24, 2.26718e-23, 1.02599e-23], [561.139, 2.35013e-23, 8.96526e-24, 4.14489e-24, 4.10661e-24, 2.27546e-23, 1.02362e-23], [562.415, 2.3554e-23, 8.97249e-24, 4.14777e-24, 4.11905e-24, 2.28376e-23, 1.02124e-23], [563.695, 2.36068e-23, 8.97976e-24, 4.15067e-24, 4.13155e-24, 2.29207e-23, 1.01885e-23], [564.977, 2.36598e-23, 8.98705e-24, 4.15358e-24, 4.14411e-24, 2.30041e-23, 1.01645e-23], [566.262, 2.37128e-23, 8.99436e-24, 4.1565e-24, 4.15673e-24, 2.30876e-23, 1.01404e-23], [567.55, 2.3766e-23, 9.00171e-24, 4.15943e-24, 4.16941e-24, 2.31714e-23, 1.01163e-23], [568.841, 2.38193e-23, 9.00909e-24, 4.16238e-24, 4.18214e-24, 2.32553e-23, 1.00921e-23], [570.135, 2.38728e-23, 9.01649e-24, 4.16535e-24, 4.19494e-24, 2.33394e-23, 1.00678e-23], [571.432, 2.39263e-23, 9.02392e-24, 4.16832e-24, 4.20779e-24, 2.34237e-23, 1.00434e-23], [572.731, 2.398e-23, 9.03138e-24, 4.17131e-24, 4.22071e-24, 2.35082e-23, 1.00189e-23], [574.034, 2.40338e-23, 9.03887e-24, 4.17431e-24, 4.23368e-24, 2.35929e-23, 9.99433e-24], [575.34, 2.40877e-23, 9.04639e-24, 4.17733e-24, 4.24671e-24, 2.36778e-23, 9.96967e-24], [576.648, 2.41418e-23, 9.05394e-24, 4.18036e-24, 4.2598e-24, 2.37629e-23, 9.94493e-24], [577.96, 2.41959e-23, 9.06152e-24, 4.1834e-24, 4.27295e-24, 2.38481e-23, 9.92011e-24], [579.275, 2.42502e-23, 9.06912e-24, 4.18646e-24, 4.28615e-24, 2.39336e-23, 9.89519e-24], [580.592, 2.43046e-23, 9.07675e-24, 4.18953e-24, 4.29942e-24, 2.40192e-23, 9.87019e-24], [581.913, 2.43592e-23, 9.08442e-24, 4.19261e-24, 4.31274e-24, 2.41051e-23, 9.8451e-24], [583.236, 2.44139e-23, 9.09211e-24, 4.19571e-24, 4.32613e-24, 2.41911e-23, 9.81993e-24], [584.563, 2.44687e-23, 9.09983e-24, 4.19882e-24, 4.33957e-24, 2.42773e-23, 9.79467e-24], [585.893, 2.45236e-23, 9.10759e-24, 4.20195e-24, 4.35307e-24, 2.43637e-23, 9.76932e-24], [587.225, 2.45786e-23, 9.11537e-24, 4.20509e-24, 4.36663e-24, 2.44503e-23, 9.74388e-24], [588.561, 2.46338e-23, 9.12318e-24, 4.20824e-24, 4.38024e-24, 2.45371e-23, 9.71835e-24], [589.9, 2.46891e-23, 9.13102e-24, 4.21141e-24, 4.39392e-24, 2.46241e-23, 9.69274e-24], [591.241, 2.47445e-23, 9.13889e-24, 4.21459e-24, 4.40765e-24, 2.47113e-23, 9.66703e-24], [592.586, 2.48001e-23, 9.14679e-24, 4.21778e-24, 4.42145e-24, 2.47986e-23, 9.64124e-24], [593.934, 2.48558e-23, 9.15472e-24, 4.22099e-24, 4.4353e-24, 2.48862e-23, 9.61536e-24], [595.285, 2.49116e-23, 9.16268e-24, 4.22421e-24, 4.44921e-24, 2.4974e-23, 9.58939e-24], [596.639, 2.49675e-23, 9.17067e-24, 4.22745e-24, 4.46317e-24, 2.50619e-23, 9.56334e-24], [597.996, 2.50236e-23, 9.17869e-24, 4.2307e-24, 4.4772e-24, 2.51501e-23, 9.53719e-24], [599.356, 2.50798e-23, 9.18674e-24, 4.23397e-24, 4.49128e-24, 2.52384e-23, 9.51095e-24], [600.72, 2.51361e-23, 9.19482e-24, 4.23725e-24, 4.50542e-24, 2.53269e-23, 9.48463e-24], [602.086, 2.51926e-23, 9.20294e-24, 4.24054e-24, 4.51962e-24, 2.54157e-23, 9.45821e-24], [603.456, 2.52492e-23, 9.21108e-24, 4.24385e-24, 4.53388e-24, 2.55046e-23, 9.43171e-24], [604.828, 2.53059e-23, 9.21925e-24, 4.24717e-24, 4.5482e-24, 2.55937e-23, 9.40511e-24], [606.204, 2.53627e-23, 9.22746e-24, 4.25051e-24, 4.56257e-24, 2.5683e-23, 9.37843e-24], [607.583, 2.54197e-23, 9.23569e-24, 4.25386e-24, 4.577e-24, 2.57725e-23, 9.35165e-24], [608.965, 2.54768e-23, 9.24396e-24, 4.25723e-24, 4.5915e-24, 2.58622e-23, 9.32479e-24], [610.35, 2.55341e-23, 9.25225e-24, 4.26061e-24, 4.60604e-24, 2.59521e-23, 9.29783e-24], [611.738, 2.55914e-23, 9.26058e-24, 4.264e-24, 4.62065e-24, 2.60422e-23, 9.27078e-24], [613.13, 2.56489e-23, 9.26894e-24, 4.26741e-24, 4.63531e-24, 2.61325e-23, 9.24365e-24], [614.524, 2.57066e-23, 9.27733e-24, 4.27083e-24, 4.65004e-24, 2.6223e-23, 9.21642e-24], [615.922, 2.57643e-23, 9.28575e-24, 4.27427e-24, 4.66482e-24, 2.63137e-23, 9.1891e-24], [617.323, 2.58222e-23, 9.29421e-24, 4.27772e-24, 4.67965e-24, 2.64046e-23, 9.16169e-24], [618.727, 2.58803e-23, 9.30269e-24, 4.28119e-24, 4.69455e-24, 2.64957e-23, 9.13419e-24], [620.134, 2.59384e-23, 9.31121e-24, 4.28468e-24, 4.7095e-24, 2.65869e-23, 9.10659e-24], [621.545, 2.59967e-23, 9.31976e-24, 4.28817e-24, 4.72451e-24, 2.66784e-23, 9.07891e-24], [622.959, 2.60552e-23, 9.32834e-24, 4.29168e-24, 4.73958e-24, 2.67701e-23, 9.05113e-24], [624.376, 2.61137e-23, 9.33695e-24, 4.29521e-24, 4.75471e-24, 2.6862e-23, 9.02326e-24], [625.796, 2.61724e-23, 9.3456e-24, 4.29875e-24, 4.76989e-24, 2.6954e-23, 8.9953e-24], [627.219, 2.62313e-23, 9.35427e-24, 4.30231e-24, 4.78513e-24, 2.70463e-23, 8.96725e-24], [628.646, 2.62902e-23, 9.36298e-24, 4.30588e-24, 4.80043e-24, 2.71388e-23, 8.9391e-24], [630.076, 2.63493e-23, 9.37172e-24, 4.30947e-24, 4.81579e-24, 2.72314e-23, 8.91087e-24], [631.509, 2.64086e-23, 9.3805e-24, 4.31307e-24, 4.8312e-24, 2.73243e-23, 8.88254e-24], [632.945, 2.6468e-23, 9.3893e-24, 4.31669e-24, 4.84667e-24, 2.74174e-23, 8.85411e-24], [634.385, 2.65275e-23, 9.39814e-24, 4.32032e-24, 4.8622e-24, 2.75106e-23, 8.8256e-24], [635.828, 2.65871e-23, 9.40701e-24, 4.32397e-24, 4.87779e-24, 2.76041e-23, 8.79699e-24], [637.274, 2.66469e-23, 9.41592e-24, 4.32763e-24, 4.89343e-24, 2.76978e-23, 8.76829e-24], [638.724, 2.67069e-23, 9.42485e-24, 4.33131e-24, 4.90913e-24, 2.77916e-23, 8.7395e-24], [640.177, 2.67669e-23, 9.43382e-24, 4.335e-24, 4.92489e-24, 2.78857e-23, 8.71061e-24], [641.633, 2.68271e-23, 9.44283e-24, 4.33871e-24, 4.9407e-24, 2.798e-23, 8.68163e-24], [643.092, 2.68875e-23, 9.45186e-24, 4.34244e-24, 4.95658e-24, 2.80744e-23, 8.65256e-24], [644.555, 2.69479e-23, 9.46093e-24, 4.34618e-24, 4.97251e-24, 2.81691e-23, 8.62339e-24], [646.021, 2.70085e-23, 9.47003e-24, 4.34993e-24, 4.98849e-24, 2.8264e-23, 8.59413e-24], [647.49, 2.70693e-23, 9.47917e-24, 4.3537e-24, 5.00454e-24, 2.83591e-23, 8.56478e-24], [648.963, 2.71302e-23, 9.48834e-24, 4.35749e-24, 5.02064e-24, 2.84543e-23, 8.53533e-24], [650.439, 2.71912e-23, 9.49754e-24, 4.36129e-24, 5.0368e-24, 2.85498e-23, 8.50579e-24], [651.919, 2.72524e-23, 9.50678e-24, 4.36511e-24, 5.05301e-24, 2.86455e-23, 8.47615e-24], [653.402, 2.73137e-23, 9.51605e-24, 4.36894e-24, 5.06929e-24, 2.87414e-23, 8.44642e-24], [654.888, 2.73752e-23, 9.52536e-24, 4.37279e-24, 5.08562e-24, 2.88375e-23, 8.4166e-24], [656.377, 2.74368e-23, 9.53469e-24, 4.37666e-24, 5.102e-24, 2.89338e-23, 8.38668e-24], [657.87, 2.74985e-23, 9.54407e-24, 4.38054e-24, 5.11845e-24, 2.90303e-23, 8.35667e-24], [659.367, 2.75604e-23, 9.55347e-24, 4.38443e-24, 5.13495e-24, 2.9127e-23, 8.32657e-24], [660.866, 2.76224e-23, 9.56292e-24, 4.38835e-24, 5.15151e-24, 2.92239e-23, 8.29637e-24], [662.37, 2.76846e-23, 9.57239e-24, 4.39227e-24, 5.16812e-24, 2.9321e-23, 8.26607e-24], [663.876, 2.77469e-23, 9.5819e-24, 4.39622e-24, 5.18479e-24, 2.94183e-23, 8.23568e-24], [665.386, 2.78093e-23, 9.59145e-24, 4.40018e-24, 5.20152e-24, 2.95158e-23, 8.2052e-24], [666.9, 2.78719e-23, 9.60103e-24, 4.40416e-24, 5.21831e-24, 2.96136e-23, 8.17462e-24], [668.417, 2.79346e-23, 9.61064e-24, 4.40815e-24, 5.23515e-24, 2.97115e-23, 8.14395e-24], [669.937, 2.79975e-23, 9.62029e-24, 4.41216e-24, 5.25205e-24, 2.98096e-23, 8.11319e-24], [671.461, 2.80605e-23, 9.62997e-24, 4.41618e-24, 5.269e-24, 2.9908e-23, 8.08233e-24], [672.988, 2.81237e-23, 9.63969e-24, 4.42022e-24, 5.28602e-24, 3.00065e-23, 8.05137e-24], [674.519, 2.8187e-23, 9.64945e-24, 4.42428e-24, 5.30309e-24, 3.01053e-23, 8.02032e-24], [676.053, 2.82505e-23, 9.65923e-24, 4.42835e-24, 5.32021e-24, 3.02043e-23, 7.98918e-24], [677.591, 2.83141e-23, 9.66906e-24, 4.43245e-24, 5.33739e-24, 3.03035e-23, 7.95794e-24], [679.132, 2.83778e-23, 9.67892e-24, 4.43655e-24, 5.35463e-24, 3.04028e-23, 7.92661e-24], [680.677, 2.84417e-23, 9.68881e-24, 4.44068e-24, 5.37193e-24, 3.05024e-23, 7.89518e-24], [682.225, 2.85058e-23, 9.69875e-24, 4.44481e-24, 5.38928e-24, 3.06022e-23, 7.86365e-24], [683.777, 2.85699e-23, 9.70871e-24, 4.44897e-24, 5.40669e-24, 3.07022e-23, 7.83204e-24], [685.332, 2.86343e-23, 9.71872e-24, 4.45314e-24, 5.42416e-24, 3.08025e-23, 7.80033e-24], [686.891, 2.86988e-23, 9.72875e-24, 4.45733e-24, 5.44168e-24, 3.09029e-23, 7.76852e-24], [688.453, 2.87634e-23, 9.73883e-24, 4.46154e-24, 5.45926e-24, 3.10035e-23, 7.73662e-24], [690.019, 2.88282e-23, 9.74894e-24, 4.46576e-24, 5.4769e-24, 3.11044e-23, 7.70462e-24], [691.589, 2.88931e-23, 9.75909e-24, 4.47e-24, 5.49459e-24, 3.12055e-23, 7.67253e-24], [693.162, 2.89582e-23, 9.76927e-24, 4.47426e-24, 5.51234e-24, 3.13067e-23, 7.64035e-24], [694.739, 2.90234e-23, 9.77949e-24, 4.47853e-24, 5.53014e-24, 3.14082e-23, 7.60807e-24], [696.319, 2.90888e-23, 9.78975e-24, 4.48282e-24, 5.548e-24, 3.15099e-23, 7.5757e-24], [697.903, 2.91543e-23, 9.80004e-24, 4.48713e-24, 5.56592e-24, 3.16118e-23, 7.54323e-24], [699.49, 2.922e-23, 9.81037e-24, 4.49145e-24, 5.5839e-24, 3.17139e-23, 7.51067e-24], [701.081, 2.92858e-23, 9.82073e-24, 4.4958e-24, 5.60193e-24, 3.18163e-23, 7.47801e-24], [702.676, 2.93518e-23, 9.83114e-24, 4.50015e-24, 5.62002e-24, 3.19188e-23, 7.44526e-24], [704.274, 2.94179e-23, 9.84158e-24, 4.50453e-24, 5.63816e-24, 3.20216e-23, 7.41242e-24], [705.876, 2.94842e-23, 9.85206e-24, 4.50892e-24, 5.65636e-24, 3.21245e-23, 7.37948e-24], [707.481, 2.95506e-23, 9.86257e-24, 4.51333e-24, 5.67462e-24, 3.22277e-23, 7.34645e-24], [709.091, 2.96172e-23, 9.87312e-24, 4.51776e-24, 5.69293e-24, 3.23311e-23, 7.31333e-24], [710.704, 2.96839e-23, 9.88371e-24, 4.5222e-24, 5.7113e-24, 3.24347e-23, 7.28011e-24], [712.32, 2.97508e-23, 9.89434e-24, 4.52667e-24, 5.72973e-24, 3.25386e-23, 7.2468e-24], [713.94, 2.98178e-23, 9.905e-24, 4.53115e-24, 5.74821e-24, 3.26426e-23, 7.21339e-24], [715.564, 2.9885e-23, 9.9157e-24, 4.53564e-24, 5.76675e-24, 3.27469e-23, 7.1799e-24], [717.192, 2.99524e-23, 9.92644e-24, 4.54016e-24, 5.78534e-24, 3.28513e-23, 7.14631e-24], [718.823, 3.00199e-23, 9.93722e-24, 4.54469e-24, 5.80399e-24, 3.2956e-23, 7.11262e-24], [720.458, 3.00875e-23, 9.94804e-24, 4.54924e-24, 5.8227e-24, 3.30609e-23, 7.07885e-24], [722.097, 3.01553e-23, 9.95889e-24, 4.5538e-24, 5.84147e-24, 3.31661e-23, 7.04498e-24], [723.739, 3.02233e-23, 9.96978e-24, 4.55839e-24, 5.86029e-24, 3.32714e-23, 7.01102e-24], [725.386, 3.02914e-23, 9.98071e-24, 4.56299e-24, 5.87916e-24, 3.3377e-23, 6.97697e-24], [727.035, 3.03597e-23, 9.99168e-24, 4.56761e-24, 5.8981e-24, 3.34827e-23, 6.94282e-24], [728.689, 3.04281e-23, 1.00027e-23, 4.57225e-24, 5.91708e-24, 3.35887e-23, 6.90859e-24], [730.347, 3.04967e-23, 1.00137e-23, 4.5769e-24, 5.93613e-24, 3.36949e-23, 6.87426e-24], [732.008, 3.05655e-23, 1.00248e-23, 4.58158e-24, 5.95523e-24, 3.38014e-23, 6.83984e-24], [733.673, 3.06344e-23, 1.00359e-23, 4.58627e-24, 5.97439e-24, 3.3908e-23, 6.80534e-24], [735.342, 3.07034e-23, 1.00471e-23, 4.59098e-24, 5.9936e-24, 3.40149e-23, 6.77074e-24], [737.014, 3.07726e-23, 1.00583e-23, 4.59571e-24, 6.01287e-24, 3.4122e-23, 6.73605e-24], [738.691, 3.0842e-23, 1.00695e-23, 4.60045e-24, 6.0322e-24, 3.42293e-23, 6.70127e-24], [740.371, 3.09116e-23, 1.00808e-23, 4.60521e-24, 6.05158e-24, 3.43368e-23, 6.6664e-24], [742.055, 3.09813e-23, 1.00921e-23, 4.61e-24, 6.07102e-24, 3.44446e-23, 6.63145e-24], [743.743, 3.10511e-23, 1.01035e-23, 4.6148e-24, 6.09052e-24, 3.45525e-23, 6.5964e-24], [745.434, 3.11211e-23, 1.01149e-23, 4.61961e-24, 6.11007e-24, 3.46607e-23, 6.56127e-24], [747.13, 3.11913e-23, 1.01263e-23, 4.62445e-24, 6.12968e-24, 3.47691e-23, 6.52604e-24], [748.829, 3.12616e-23, 1.01378e-23, 4.62931e-24, 6.14934e-24, 3.48778e-23, 6.49073e-24], [750.533, 3.13321e-23, 1.01493e-23, 4.63418e-24, 6.16906e-24, 3.49866e-23, 6.45534e-24], [752.24, 3.14028e-23, 1.01609e-23, 4.63907e-24, 6.18884e-24, 3.50957e-23, 6.41985e-24], [753.951, 3.14736e-23, 1.01725e-23, 4.64398e-24, 6.20867e-24, 3.5205e-23, 6.38428e-24], [755.666, 3.15446e-23, 1.01841e-23, 4.64891e-24, 6.22856e-24, 3.53146e-23, 6.34863e-24], [757.385, 3.16157e-23, 1.01958e-23, 4.65386e-24, 6.24851e-24, 3.54243e-23, 6.31289e-24], [759.107, 3.1687e-23, 1.02075e-23, 4.65882e-24, 6.26851e-24, 3.55343e-23, 6.27706e-24], [760.834, 3.17585e-23, 1.02193e-23, 4.66381e-24, 6.28857e-24, 3.56445e-23, 6.24115e-24], [762.564, 3.18301e-23, 1.02311e-23, 4.66881e-24, 6.30868e-24, 3.57549e-23, 6.20516e-24], [764.299, 3.19019e-23, 1.02429e-23, 4.67383e-24, 6.32885e-24, 3.58656e-23, 6.16908e-24], [766.037, 3.19739e-23, 1.02548e-23, 4.67887e-24, 6.34908e-24, 3.59765e-23, 6.13292e-24], [767.78, 3.2046e-23, 1.02667e-23, 4.68393e-24, 6.36936e-24, 3.60876e-23, 6.09668e-24], [769.526, 3.21183e-23, 1.02787e-23, 4.68901e-24, 6.3897e-24, 3.61989e-23, 6.06036e-24], [771.277, 3.21908e-23, 1.02907e-23, 4.69411e-24, 6.41009e-24, 3.63105e-23, 6.02396e-24], [773.031, 3.22634e-23, 1.03027e-23, 4.69923e-24, 6.43055e-24, 3.64223e-23, 5.98747e-24], [774.789, 3.23362e-23, 1.03148e-23, 4.70436e-24, 6.45105e-24, 3.65343e-23, 5.95091e-24], [776.551, 3.24091e-23, 1.03269e-23, 4.70952e-24, 6.47162e-24, 3.66465e-23, 5.91427e-24], [778.318, 3.24823e-23, 1.03391e-23, 4.71469e-24, 6.49224e-24, 3.6759e-23, 5.87756e-24], [780.088, 3.25556e-23, 1.03513e-23, 4.71988e-24, 6.51292e-24, 3.68717e-23, 5.84076e-24], [781.862, 3.2629e-23, 1.03636e-23, 4.7251e-24, 6.53365e-24, 3.69847e-23, 5.80389e-24], [783.641, 3.27026e-23, 1.03759e-23, 4.73033e-24, 6.55444e-24, 3.70978e-23, 5.76695e-24], [785.423, 3.27764e-23, 1.03882e-23, 4.73558e-24, 6.57528e-24, 3.72112e-23, 5.72993e-24], [787.21, 3.28504e-23, 1.04006e-23, 4.74085e-24, 6.59619e-24, 3.73249e-23, 5.69284e-24], [789.0, 3.29245e-23, 1.0413e-23, 4.74614e-24, 6.61715e-24, 3.74387e-23, 5.65567e-24], [790.795, 3.29988e-23, 1.04255e-23, 4.75145e-24, 6.63816e-24, 3.75528e-23, 5.61844e-24], [792.594, 3.30733e-23, 1.0438e-23, 4.75678e-24, 6.65923e-24, 3.76671e-23, 5.58113e-24], [794.397, 3.31479e-23, 1.04505e-23, 4.76212e-24, 6.68036e-24, 3.77817e-23, 5.54376e-24], [796.203, 3.32227e-23, 1.04631e-23, 4.76749e-24, 6.70154e-24, 3.78965e-23, 5.50632e-24], [798.014, 3.32977e-23, 1.04757e-23, 4.77288e-24, 6.72279e-24, 3.80115e-23, 5.46881e-24], [799.83, 3.33729e-23, 1.04884e-23, 4.77829e-24, 6.74408e-24, 3.81267e-23, 5.43124e-24], [801.649, 3.34482e-23, 1.05011e-23, 4.78371e-24, 6.76544e-24, 3.82422e-23, 5.3936e-24], [803.472, 3.35237e-23, 1.05139e-23, 4.78916e-24, 6.78685e-24, 3.8358e-23, 5.3559e-24], [805.3, 3.35994e-23, 1.05267e-23, 4.79463e-24, 6.80831e-24, 3.84739e-23, 5.31814e-24], [807.132, 3.36752e-23, 1.05395e-23, 4.80011e-24, 6.82984e-24, 3.85901e-23, 5.28032e-24], [808.967, 3.37512e-23, 1.05524e-23, 4.80562e-24, 6.85142e-24, 3.87065e-23, 5.24244e-24], [810.807, 3.38274e-23, 1.05654e-23, 4.81114e-24, 6.87305e-24, 3.88232e-23, 5.2045e-24], [812.652, 3.39038e-23, 1.05783e-23, 4.81669e-24, 6.89475e-24, 3.89401e-23, 5.16651e-24], [814.5, 3.39803e-23, 1.05914e-23, 4.82226e-24, 6.91649e-24, 3.90572e-23, 5.12847e-24], [816.353, 3.4057e-23, 1.06044e-23, 4.82784e-24, 6.9383e-24, 3.91746e-23, 5.09037e-24], [818.21, 3.41339e-23, 1.06175e-23, 4.83345e-24, 6.96016e-24, 3.92922e-23, 5.05223e-24], [820.071, 3.4211e-23, 1.06307e-23, 4.83908e-24, 6.98208e-24, 3.94101e-23, 5.01404e-24], [821.936, 3.42882e-23, 1.06439e-23, 4.84473e-24, 7.00406e-24, 3.95282e-23, 4.9758e-24], [823.806, 3.43656e-23, 1.06571e-23, 4.85039e-24, 7.02609e-24, 3.96465e-23, 4.93751e-24], [825.679, 3.44432e-23, 1.06704e-23, 4.85608e-24, 7.04818e-24, 3.97651e-23, 4.89919e-24], [827.557, 3.4521e-23, 1.06837e-23, 4.86179e-24, 7.07033e-24, 3.98839e-23, 4.86082e-24], [829.44, 3.4599e-23, 1.06971e-23, 4.86752e-24, 7.09253e-24, 4.0003e-23, 4.82242e-24], [831.326, 3.46771e-23, 1.07105e-23, 4.87327e-24, 7.11479e-24, 4.01223e-23, 4.78399e-24], [833.217, 3.47554e-23, 1.0724e-23, 4.87904e-24, 7.13711e-24, 4.02418e-23, 4.74552e-24], [835.113, 3.48339e-23, 1.07375e-23, 4.88483e-24, 7.15948e-24, 4.03616e-23, 4.70702e-24], [837.012, 3.49125e-23, 1.07511e-23, 4.89064e-24, 7.18191e-24, 4.04816e-23, 4.66849e-24], [838.916, 3.49914e-23, 1.07647e-23, 4.89647e-24, 7.2044e-24, 4.06018e-23, 4.62994e-24], [840.824, 3.50704e-23, 1.07783e-23, 4.90232e-24, 7.22694e-24, 4.07224e-23, 4.59137e-24], [842.737, 3.51496e-23, 1.0792e-23, 4.9082e-24, 7.24954e-24, 4.08431e-23, 4.55278e-24], [844.653, 3.5229e-23, 1.08057e-23, 4.91409e-24, 7.2722e-24, 4.09641e-23, 4.51417e-24], [846.575, 3.53086e-23, 1.08195e-23, 4.92001e-24, 7.29492e-24, 4.10853e-23, 4.47555e-24], [848.5, 3.53883e-23, 1.08333e-23, 4.92595e-24, 7.31769e-24, 4.12068e-23, 4.43693e-24], [850.43, 3.54682e-23, 1.08472e-23, 4.9319e-24, 7.34052e-24, 4.13286e-23, 4.3983e-24], [852.365, 3.55484e-23, 1.08611e-23, 4.93788e-24, 7.36341e-24, 4.14505e-23, 4.35966e-24], [854.303, 3.56287e-23, 1.08751e-23, 4.94388e-24, 7.38635e-24, 4.15728e-23, 4.32103e-24], [856.247, 3.57091e-23, 1.08891e-23, 4.9499e-24, 7.40935e-24, 4.16952e-23, 4.28241e-24], [858.194, 3.57898e-23, 1.09032e-23, 4.95594e-24, 7.43241e-24, 4.18179e-23, 4.24379e-24], [860.146, 3.58706e-23, 1.09173e-23, 4.96201e-24, 7.45552e-24, 4.19409e-23, 4.20519e-24], [862.103, 3.59517e-23, 1.09314e-23, 4.96809e-24, 7.47869e-24, 4.20641e-23, 4.16661e-24], [864.064, 3.60329e-23, 1.09456e-23, 4.9742e-24, 7.50192e-24, 4.21876e-23, 4.12806e-24], [866.029, 3.61143e-23, 1.09598e-23, 4.98033e-24, 7.52521e-24, 4.23113e-23, 4.08953e-24], [867.999, 3.61959e-23, 1.09741e-23, 4.98647e-24, 7.54856e-24, 4.24353e-23, 4.05104e-24], [869.973, 3.62777e-23, 1.09885e-23, 4.99264e-24, 7.57196e-24, 4.25595e-23, 4.01258e-24], [871.952, 3.63596e-23, 1.10028e-23, 4.99884e-24, 7.59542e-24, 4.26839e-23, 3.97417e-24], [873.935, 3.64418e-23, 1.10173e-23, 5.00505e-24, 7.61893e-24, 4.28087e-23, 3.93581e-24], [875.923, 3.65241e-23, 1.10318e-23, 5.01129e-24, 7.64251e-24, 4.29336e-23, 3.89751e-24], [877.915, 3.66066e-23, 1.10463e-23, 5.01754e-24, 7.66614e-24, 4.30588e-23, 3.85927e-24], [879.912, 3.66893e-23, 1.10609e-23, 5.02382e-24, 7.68983e-24, 4.31843e-23, 3.8211e-24], [881.914, 3.67723e-23, 1.10755e-23, 5.03012e-24, 7.71358e-24, 4.331e-23, 3.78301e-24], [883.92, 3.68553e-23, 1.10901e-23, 5.03645e-24, 7.73738e-24, 4.3436e-23, 3.745e-24], [885.93, 3.69386e-23, 1.11049e-23, 5.04279e-24, 7.76124e-24, 4.35623e-23, 3.70708e-24], [887.945, 3.70221e-23, 1.11196e-23, 5.04916e-24, 7.78516e-24, 4.36887e-23, 3.66926e-24], [889.965, 3.71058e-23, 1.11344e-23, 5.05554e-24, 7.80914e-24, 4.38155e-23, 3.63155e-24], [891.989, 3.71896e-23, 1.11493e-23, 5.06196e-24, 7.83318e-24, 4.39425e-23, 3.59395e-24], [894.018, 3.72737e-23, 1.11642e-23, 5.06839e-24, 7.85727e-24, 4.40697e-23, 3.55648e-24], [896.052, 3.73579e-23, 1.11791e-23, 5.07484e-24, 7.88142e-24, 4.41973e-23, 3.51914e-24], [898.09, 3.74423e-23, 1.11941e-23, 5.08132e-24, 7.90563e-24, 4.4325e-23, 3.48194e-24], [900.133, 3.75269e-23, 1.12092e-23, 5.08782e-24, 7.9299e-24, 4.44531e-23, 3.4449e-24], [902.18, 3.76118e-23, 1.12243e-23, 5.09434e-24, 7.95422e-24, 4.45813e-23, 3.40802e-24], [904.232, 3.76968e-23, 1.12395e-23, 5.10089e-24, 7.97861e-24, 4.47099e-23, 3.37131e-24], [906.289, 3.7782e-23, 1.12547e-23, 5.10745e-24, 8.00305e-24, 4.48387e-23, 3.33479e-24], [908.35, 3.78674e-23, 1.12699e-23, 5.11404e-24, 8.02755e-24, 4.49678e-23, 3.29847e-24], [910.416, 3.7953e-23, 1.12852e-23, 5.12066e-24, 8.05211e-24, 4.50971e-23, 3.26236e-24], [912.487, 3.80387e-23, 1.13005e-23, 5.12729e-24, 8.07673e-24, 4.52267e-23, 3.22648e-24], [914.563, 3.81247e-23, 1.13159e-23, 5.13395e-24, 8.1014e-24, 4.53565e-23, 3.19083e-24], [916.643, 3.82109e-23, 1.13314e-23, 5.14063e-24, 8.12614e-24, 4.54866e-23, 3.15543e-24], [918.728, 3.82973e-23, 1.13469e-23, 5.14733e-24, 8.15093e-24, 4.5617e-23, 3.1203e-24], [920.818, 3.83839e-23, 1.13624e-23, 5.15406e-24, 8.17578e-24, 4.57476e-23, 3.08545e-24], [922.912, 3.84706e-23, 1.1378e-23, 5.16081e-24, 8.20069e-24, 4.58785e-23, 3.0509e-24], [925.011, 3.85576e-23, 1.13937e-23, 5.16758e-24, 8.22566e-24, 4.60097e-23, 3.01667e-24], [927.115, 3.86448e-23, 1.14094e-23, 5.17437e-24, 8.25068e-24, 4.61411e-23, 2.98276e-24], [929.224, 3.87321e-23, 1.14251e-23, 5.18119e-24, 8.27577e-24, 4.62728e-23, 2.94921e-24], [931.338, 3.88197e-23, 1.14409e-23, 5.18804e-24, 8.30091e-24, 4.64048e-23, 2.91603e-24], [933.456, 3.89075e-23, 1.14568e-23, 5.1949e-24, 8.32612e-24, 4.6537e-23, 2.88324e-24], [935.579, 3.89954e-23, 1.14727e-23, 5.20179e-24, 8.35138e-24, 4.66695e-23, 2.85087e-24], [937.707, 3.90836e-23, 1.14886e-23, 5.20871e-24, 8.37671e-24, 4.68023e-23, 2.81892e-24], [939.84, 3.9172e-23, 1.15046e-23, 5.21565e-24, 8.40209e-24, 4.69353e-23, 2.78743e-24], [941.978, 3.92606e-23, 1.15207e-23, 5.22262e-24, 8.42753e-24, 4.70686e-23, 2.75643e-24], [944.12, 3.93493e-23, 1.15368e-23, 5.22961e-24, 8.45304e-24, 4.72022e-23, 2.72593e-24], [946.268, 3.94383e-23, 1.15529e-23, 5.23664e-24, 8.47861e-24, 4.7336e-23, 2.69598e-24], [948.42, 3.95275e-23, 1.15691e-23, 5.2437e-24, 8.50425e-24, 4.74701e-23, 2.6666e-24], [950.578, 3.96169e-23, 1.15854e-23, 5.25081e-24, 8.52996e-24, 4.76045e-23, 2.63785e-24], [952.74, 3.97065e-23, 1.16018e-23, 5.258e-24, 8.55577e-24, 4.77392e-23, 2.60982e-24], [954.907, 3.97963e-23, 1.16183e-23, 5.26537e-24, 8.58172e-24, 4.78741e-23, 2.58274e-24], [957.079, 3.98864e-23, 1.16351e-23, 5.27337e-24, 8.60811e-24, 4.80094e-23, 2.55758e-24], [959.256, 3.99776e-23, 1.16548e-23, 5.28762e-24, 8.63837e-24, 4.81457e-23, 2.54612e-24], [961.438, 4.00727e-23, 1.16875e-23, 5.33051e-24, 8.68621e-24, 4.82853e-23, 2.59496e-24], [963.624, 4.01578e-23, 1.1685e-23, 5.29579e-24, 8.68658e-24, 4.84167e-23, 2.48298e-24], [965.816, 4.02484e-23, 1.17009e-23, 5.30138e-24, 8.71168e-24, 4.85528e-23, 2.45548e-24], [968.013, 4.03393e-23, 1.17174e-23, 5.30823e-24, 8.73759e-24, 4.86894e-23, 2.43143e-24], [970.215, 4.04305e-23, 1.17341e-23, 5.31534e-24, 8.76371e-24, 4.88262e-23, 2.40873e-24], [972.422, 4.0522e-23, 1.17508e-23, 5.32255e-24, 8.78994e-24, 4.89633e-23, 2.38708e-24], [974.634, 4.06136e-23, 1.17676e-23, 5.32982e-24, 8.81624e-24, 4.91007e-23, 2.36641e-24], [976.85, 4.07055e-23, 1.17845e-23, 5.33714e-24, 8.84261e-24, 4.92384e-23, 2.34673e-24], [979.072, 4.07976e-23, 1.18014e-23, 5.34448e-24, 8.86905e-24, 4.93763e-23, 2.32804e-24], [981.299, 4.08898e-23, 1.18184e-23, 5.35185e-24, 8.89556e-24, 4.95145e-23, 2.31038e-24], [983.531, 4.09823e-23, 1.18354e-23, 5.35925e-24, 8.92212e-24, 4.9653e-23, 2.29377e-24], [985.768, 4.1075e-23, 1.18525e-23, 5.36668e-24, 8.94875e-24, 4.97918e-23, 2.27825e-24], [988.011, 4.11679e-23, 1.18696e-23, 5.37413e-24, 8.97544e-24, 4.99309e-23, 2.26384e-24], [990.258, 4.12611e-23, 1.18868e-23, 5.38161e-24, 9.00219e-24, 5.00702e-23, 2.25058e-24], [992.51, 4.13544e-23, 1.19041e-23, 5.38912e-24, 9.029e-24, 5.02099e-23, 2.2385e-24], [994.768, 4.14479e-23, 1.19214e-23, 5.39665e-24, 9.05587e-24, 5.03498e-23, 2.22762e-24], [997.031, 4.15417e-23, 1.19388e-23, 5.4042e-24, 9.0828e-24, 5.049e-23, 2.21797e-24], [999.298, 4.16357e-23, 1.19562e-23, 5.41178e-24, 9.10979e-24, 5.06305e-23, 2.20958e-24], [1001.57, 4.17299e-23, 1.19737e-23, 5.41938e-24, 9.13684e-24, 5.07712e-23, 2.20247e-24], [1003.85, 4.18243e-23, 1.19912e-23, 5.42701e-24, 9.16396e-24, 5.09123e-23, 2.19667e-24], [1006.13, 4.19189e-23, 1.20088e-23, 5.43467e-24, 9.19113e-24, 5.10536e-23, 2.1922e-24], [1008.42, 4.20137e-23, 1.20264e-23, 5.44235e-24, 9.21837e-24, 5.11952e-23, 2.18907e-24], [1010.72, 4.21088e-23, 1.20441e-23, 5.45005e-24, 9.24567e-24, 5.13371e-23, 2.1873e-24], [1013.01, 4.22041e-23, 1.20618e-23, 5.45778e-24, 9.27303e-24, 5.14793e-23, 2.18691e-24], [1015.32, 4.22996e-23, 1.20796e-23, 5.46554e-24, 9.30045e-24, 5.16218e-23, 2.1879e-24], [1017.63, 4.23953e-23, 1.20975e-23, 5.47332e-24, 9.32793e-24, 5.17646e-23, 2.1903e-24], [1019.94, 4.24912e-23, 1.21154e-23, 5.48112e-24, 9.35547e-24, 5.19076e-23, 2.19409e-24], [1022.26, 4.25873e-23, 1.21333e-23, 5.48895e-24, 9.38307e-24, 5.2051e-23, 2.19929e-24], [1024.59, 4.26837e-23, 1.21514e-23, 5.49681e-24, 9.41074e-24, 5.21946e-23, 2.2059e-24], [1026.92, 4.27803e-23, 1.21694e-23, 5.50469e-24, 9.43846e-24, 5.23386e-23, 2.21391e-24], [1029.25, 4.28771e-23, 1.21876e-23, 5.5126e-24, 9.46625e-24, 5.24828e-23, 2.22332e-24], [1031.59, 4.29741e-23, 1.22057e-23, 5.52053e-24, 9.4941e-24, 5.26273e-23, 2.23413e-24], [1033.94, 4.30714e-23, 1.2224e-23, 5.52849e-24, 9.52201e-24, 5.27721e-23, 2.24632e-24], [1036.29, 4.31688e-23, 1.22423e-23, 5.53647e-24, 9.54999e-24, 5.29172e-23, 2.25989e-24], [1038.65, 4.32665e-23, 1.22606e-23, 5.54448e-24, 9.57802e-24, 5.30626e-23, 2.27481e-24], [1041.01, 4.33644e-23, 1.22791e-23, 5.55252e-24, 9.60612e-24, 5.32083e-23, 2.29108e-24], [1043.38, 4.34626e-23, 1.22975e-23, 5.56058e-24, 9.63428e-24, 5.33543e-23, 2.30867e-24], [1045.75, 4.35609e-23, 1.2316e-23, 5.56867e-24, 9.6625e-24, 5.35005e-23, 2.32757e-24], [1048.13, 4.36595e-23, 1.23346e-23, 5.57678e-24, 9.69079e-24, 5.36471e-23, 2.34776e-24], [1050.52, 4.37583e-23, 1.23533e-23, 5.58492e-24, 9.71914e-24, 5.3794e-23, 2.36921e-24], [1052.91, 4.38573e-23, 1.2372e-23, 5.59308e-24, 9.74755e-24, 5.39412e-23, 2.39189e-24], [1055.3, 4.39566e-23, 1.23907e-23, 5.60127e-24, 9.77602e-24, 5.40886e-23, 2.41579e-24], [1057.7, 4.40561e-23, 1.24095e-23, 5.60949e-24, 9.80455e-24, 5.42364e-23, 2.44089e-24], [1060.11, 4.41558e-23, 1.24284e-23, 5.61773e-24, 9.83315e-24, 5.43845e-23, 2.46714e-24], [1062.52, 4.42557e-23, 1.24473e-23, 5.626e-24, 9.86181e-24, 5.45329e-23, 2.49454e-24], [1064.94, 4.43559e-23, 1.24663e-23, 5.6343e-24, 9.89053e-24, 5.46815e-23, 2.52304e-24], [1067.36, 4.44563e-23, 1.24853e-23, 5.64262e-24, 9.91932e-24, 5.48305e-23, 2.55262e-24], [1069.79, 4.45569e-23, 1.25044e-23, 5.65097e-24, 9.94817e-24, 5.49798e-23, 2.58327e-24], [1072.22, 4.46578e-23, 1.25236e-23, 5.65934e-24, 9.97708e-24, 5.51293e-23, 2.61494e-24], [1074.66, 4.47589e-23, 1.25428e-23, 5.66775e-24, 1.00061e-23, 5.52792e-23, 2.64761e-24], [1077.1, 4.48602e-23, 1.25621e-23, 5.67617e-24, 1.00351e-23, 5.54294e-23, 2.68125e-24], [1079.55, 4.49617e-23, 1.25814e-23, 5.68463e-24, 1.00642e-23, 5.55799e-23, 2.71585e-24], [1082.01, 4.50635e-23, 1.26008e-23, 5.69311e-24, 1.00934e-23, 5.57307e-23, 2.75136e-24], [1084.47, 4.51655e-23, 1.26202e-23, 5.70162e-24, 1.01226e-23, 5.58818e-23, 2.78777e-24], [1086.93, 4.52678e-23, 1.26397e-23, 5.71015e-24, 1.01519e-23, 5.60332e-23, 2.82505e-24], [1089.41, 4.53702e-23, 1.26593e-23, 5.71871e-24, 1.01812e-23, 5.61849e-23, 2.86318e-24], [1091.89, 4.54729e-23, 1.26789e-23, 5.7273e-24, 1.02107e-23, 5.6337e-23, 2.90213e-24], [1094.37, 4.55759e-23, 1.26986e-23, 5.73592e-24, 1.02401e-23, 5.64893e-23, 2.94188e-24], [1096.86, 4.56791e-23, 1.27184e-23, 5.74456e-24, 1.02697e-23, 5.6642e-23, 2.9824e-24], [1099.35, 4.57825e-23, 1.27382e-23, 5.75323e-24, 1.02993e-23, 5.67949e-23, 3.02367e-24], [1101.85, 4.58861e-23, 1.2758e-23, 5.76193e-24, 1.0329e-23, 5.69482e-23, 3.06568e-24], [1104.36, 4.599e-23, 1.27779e-23, 5.77065e-24, 1.03587e-23, 5.71018e-23, 3.1084e-24], [1106.87, 4.60941e-23, 1.27979e-23, 5.7794e-24, 1.03885e-23, 5.72557e-23, 3.1518e-24], [1109.39, 4.61985e-23, 1.2818e-23, 5.78818e-24, 1.04184e-23, 5.74099e-23, 3.19588e-24], [1111.91, 4.63031e-23, 1.28381e-23, 5.79699e-24, 1.04483e-23, 5.75644e-23, 3.24061e-24], [1114.44, 4.64079e-23, 1.28582e-23, 5.80582e-24, 1.04783e-23, 5.77192e-23, 3.28597e-24], [1116.98, 4.6513e-23, 1.28784e-23, 5.81468e-24, 1.05084e-23, 5.78744e-23, 3.33195e-24], [1119.52, 4.66183e-23, 1.28987e-23, 5.82357e-24, 1.05385e-23, 5.80299e-23, 3.37852e-24], [1122.06, 4.67239e-23, 1.29191e-23, 5.83249e-24, 1.05687e-23, 5.81856e-23, 3.42568e-24], [1124.62, 4.68297e-23, 1.29395e-23, 5.84143e-24, 1.0599e-23, 5.83417e-23, 3.4734e-24], [1127.17, 4.69357e-23, 1.29599e-23, 5.8504e-24, 1.06293e-23, 5.84982e-23, 3.52168e-24], [1129.74, 4.7042e-23, 1.29805e-23, 5.8594e-24, 1.06597e-23, 5.86549e-23, 3.57049e-24], [1132.31, 4.71485e-23, 1.30011e-23, 5.86843e-24, 1.06901e-23, 5.8812e-23, 3.61983e-24], [1134.88, 4.72553e-23, 1.30217e-23, 5.87748e-24, 1.07207e-23, 5.89693e-23, 3.66967e-24], [1137.46, 4.73623e-23, 1.30424e-23, 5.88657e-24, 1.07513e-23, 5.9127e-23, 3.72001e-24], [1140.05, 4.74695e-23, 1.30632e-23, 5.89568e-24, 1.07819e-23, 5.92851e-23, 3.77084e-24], [1142.64, 4.7577e-23, 1.3084e-23, 5.90482e-24, 1.08126e-23, 5.94434e-23, 3.82214e-24], [1145.24, 4.76848e-23, 1.31049e-23, 5.91399e-24, 1.08434e-23, 5.96021e-23, 3.8739e-24], [1147.85, 4.77928e-23, 1.31259e-23, 5.92318e-24, 1.08743e-23, 5.97611e-23, 3.92611e-24], [1150.46, 4.7901e-23, 1.31469e-23, 5.93241e-24, 1.09052e-23, 5.99204e-23, 3.97876e-24], [1153.08, 4.80095e-23, 1.3168e-23, 5.94166e-24, 1.09362e-23, 6.008e-23, 4.03184e-24], [1155.7, 4.81182e-23, 1.31891e-23, 5.95094e-24, 1.09672e-23, 6.024e-23, 4.08533e-24], [1158.33, 4.82272e-23, 1.32103e-23, 5.96025e-24, 1.09983e-23, 6.04003e-23, 4.13924e-24], [1160.96, 4.83364e-23, 1.32316e-23, 5.96959e-24, 1.10295e-23, 6.05609e-23, 4.19355e-24], [1163.6, 4.84459e-23, 1.32529e-23, 5.97895e-24, 1.10608e-23, 6.07219e-23, 4.24826e-24], [1166.25, 4.85556e-23, 1.32743e-23, 5.98835e-24, 1.10921e-23, 6.08832e-23, 4.30335e-24], [1168.9, 4.86656e-23, 1.32958e-23, 5.99777e-24, 1.11234e-23, 6.10448e-23, 4.35881e-24], [1171.56, 4.87758e-23, 1.33173e-23, 6.00722e-24, 1.11549e-23, 6.12067e-23, 4.41465e-24], [1174.23, 4.88862e-23, 1.33389e-23, 6.0167e-24, 1.11864e-23, 6.1369e-23, 4.47085e-24], [1176.9, 4.8997e-23, 1.33605e-23, 6.02621e-24, 1.1218e-23, 6.15316e-23, 4.5274e-24], [1179.57, 4.9108e-23, 1.33822e-23, 6.03575e-24, 1.12496e-23, 6.16945e-23, 4.58431e-24], [1182.26, 4.92192e-23, 1.3404e-23, 6.04532e-24, 1.12814e-23, 6.18578e-23, 4.64155e-24], [1184.95, 4.93307e-23, 1.34259e-23, 6.05492e-24, 1.13132e-23, 6.20214e-23, 4.69913e-24], [1187.64, 4.94424e-23, 1.34478e-23, 6.06454e-24, 1.1345e-23, 6.21854e-23, 4.75705e-24], [1190.34, 4.95544e-23, 1.34697e-23, 6.0742e-24, 1.13769e-23, 6.23496e-23, 4.81528e-24], [1193.05, 4.96666e-23, 1.34918e-23, 6.08388e-24, 1.14089e-23, 6.25143e-23, 4.87384e-24], [1195.76, 4.97791e-23, 1.35139e-23, 6.0936e-24, 1.1441e-23, 6.26792e-23, 4.93271e-24], [1198.48, 4.98919e-23, 1.3536e-23, 6.10334e-24, 1.14731e-23, 6.28445e-23, 4.99189e-24], [1201.21, 5.00049e-23, 1.35583e-23, 6.11311e-24, 1.15053e-23, 6.30101e-23, 5.05137e-24], [1203.94, 5.01182e-23, 1.35805e-23, 6.12292e-24, 1.15376e-23, 6.31761e-23, 5.11116e-24], [1206.68, 5.02317e-23, 1.36029e-23, 6.13275e-24, 1.15699e-23, 6.33424e-23, 5.17123e-24], [1209.43, 5.03455e-23, 1.36253e-23, 6.14261e-24, 1.16023e-23, 6.35091e-23, 5.2316e-24], [1212.18, 5.04596e-23, 1.36478e-23, 6.1525e-24, 1.16348e-23, 6.36761e-23, 5.29226e-24], [1214.93, 5.05739e-23, 1.36704e-23, 6.16242e-24, 1.16673e-23, 6.38434e-23, 5.3532e-24], [1217.7, 5.06885e-23, 1.3693e-23, 6.17237e-24, 1.16999e-23, 6.40111e-23, 5.41442e-24], [1220.47, 5.08033e-23, 1.37157e-23, 6.18235e-24, 1.17326e-23, 6.41791e-23, 5.47591e-24], [1223.24, 5.09184e-23, 1.37384e-23, 6.19236e-24, 1.17653e-23, 6.43475e-23, 5.53768e-24], [1226.02, 5.10337e-23, 1.37612e-23, 6.20241e-24, 1.17981e-23, 6.45162e-23, 5.59972e-24], [1228.81, 5.11494e-23, 1.37841e-23, 6.21248e-24, 1.1831e-23, 6.46852e-23, 5.66202e-24], [1231.61, 5.12652e-23, 1.38071e-23, 6.22258e-24, 1.1864e-23, 6.48547e-23, 5.72458e-24], [1234.41, 5.13814e-23, 1.38301e-23, 6.23271e-24, 1.1897e-23, 6.50244e-23, 5.7874e-24], [1237.22, 5.14978e-23, 1.38532e-23, 6.24287e-24, 1.19301e-23, 6.51945e-23, 5.85048e-24], [1240.03, 5.16145e-23, 1.38763e-23, 6.25306e-24, 1.19633e-23, 6.5365e-23, 5.91381e-24], [1242.85, 5.17314e-23, 1.38995e-23, 6.26328e-24, 1.19965e-23, 6.55358e-23, 5.97739e-24], [1245.68, 5.18486e-23, 1.39228e-23, 6.27353e-24, 1.20298e-23, 6.5707e-23, 6.04122e-24], [1248.51, 5.19661e-23, 1.39461e-23, 6.28382e-24, 1.20632e-23, 6.58785e-23, 6.1053e-24], [1251.35, 5.20838e-23, 1.39696e-23, 6.29413e-24, 1.20966e-23, 6.60503e-23, 6.16962e-24], [1254.2, 5.22019e-23, 1.3993e-23, 6.30448e-24, 1.21301e-23, 6.62226e-23, 6.23418e-24], [1257.05, 5.23201e-23, 1.40166e-23, 6.31485e-24, 1.21637e-23, 6.63951e-23, 6.29898e-24], [1259.91, 5.24387e-23, 1.40402e-23, 6.32526e-24, 1.21974e-23, 6.65681e-23, 6.36402e-24], [1262.78, 5.25575e-23, 1.40639e-23, 6.33569e-24, 1.22311e-23, 6.67414e-23, 6.42929e-24], [1265.65, 5.26766e-23, 1.40877e-23, 6.34616e-24, 1.22649e-23, 6.6915e-23, 6.49479e-24], [1268.53, 5.2796e-23, 1.41115e-23, 6.35666e-24, 1.22988e-23, 6.7089e-23, 6.56052e-24], [1271.41, 5.29156e-23, 1.41354e-23, 6.36719e-24, 1.23327e-23, 6.72634e-23, 6.62649e-24], [1274.3, 5.30355e-23, 1.41593e-23, 6.37775e-24, 1.23667e-23, 6.74381e-23, 6.69268e-24], [1277.2, 5.31557e-23, 1.41834e-23, 6.38834e-24, 1.24008e-23, 6.76132e-23, 6.75909e-24], [1280.11, 5.32761e-23, 1.42075e-23, 6.39896e-24, 1.2435e-23, 6.77886e-23, 6.82573e-24], [1283.02, 5.33969e-23, 1.42316e-23, 6.40962e-24, 1.24692e-23, 6.79644e-23, 6.89259e-24], [1285.94, 5.35179e-23, 1.42559e-23, 6.4203e-24, 1.25035e-23, 6.81406e-23, 6.95967e-24], [1288.86, 5.36391e-23, 1.42802e-23, 6.43102e-24, 1.25379e-23, 6.83171e-23, 7.02697e-24], [1291.79, 5.37607e-23, 1.43045e-23, 6.44177e-24, 1.25723e-23, 6.8494e-23, 7.09449e-24], [1294.73, 5.38825e-23, 1.4329e-23, 6.45255e-24, 1.26068e-23, 6.86713e-23, 7.16222e-24], [1297.68, 5.40046e-23, 1.43535e-23, 6.46336e-24, 1.26414e-23, 6.88489e-23, 7.23017e-24], [1300.63, 5.4127e-23, 1.43781e-23, 6.4742e-24, 1.26761e-23, 6.90269e-23, 7.29833e-24], [1303.59, 5.42497e-23, 1.44027e-23, 6.48507e-24, 1.27108e-23, 6.92053e-23, 7.3667e-24], [1306.55, 5.43726e-23, 1.44275e-23, 6.49598e-24, 1.27456e-23, 6.9384e-23, 7.43528e-24], [1309.53, 5.44959e-23, 1.44523e-23, 6.50692e-24, 1.27805e-23, 6.95631e-23, 7.50407e-24], [1312.5, 5.46194e-23, 1.44771e-23, 6.51789e-24, 1.28154e-23, 6.97426e-23, 7.57307e-24], [1315.49, 5.47432e-23, 1.45021e-23, 6.52889e-24, 1.28505e-23, 6.99224e-23, 7.64228e-24], [1318.48, 5.48672e-23, 1.45271e-23, 6.53993e-24, 1.28856e-23, 7.01027e-23, 7.7117e-24], [1321.48, 5.49916e-23, 1.45521e-23, 6.55099e-24, 1.29208e-23, 7.02833e-23, 7.78132e-24], [1324.49, 5.51162e-23, 1.45773e-23, 6.56209e-24, 1.2956e-23, 7.04642e-23, 7.85114e-24], [1327.5, 5.52411e-23, 1.46025e-23, 6.57322e-24, 1.29913e-23, 7.06456e-23, 7.92117e-24], [1330.52, 5.53664e-23, 1.46278e-23, 6.58439e-24, 1.30267e-23, 7.08273e-23, 7.9914e-24], [1333.54, 5.54918e-23, 1.46532e-23, 6.59558e-24, 1.30622e-23, 7.10094e-23, 8.06183e-24], [1336.58, 5.56176e-23, 1.46786e-23, 6.60681e-24, 1.30977e-23, 7.11918e-23, 8.13246e-24], [1339.62, 5.57437e-23, 1.47041e-23, 6.61807e-24, 1.31334e-23, 7.13747e-23, 8.20329e-24], [1342.67, 5.587e-23, 1.47297e-23, 6.62937e-24, 1.31691e-23, 7.15579e-23, 8.27432e-24], [1345.72, 5.59967e-23, 1.47553e-23, 6.64069e-24, 1.32048e-23, 7.17415e-23, 8.34555e-24], [1348.78, 5.61236e-23, 1.4781e-23, 6.65205e-24, 1.32407e-23, 7.19255e-23, 8.41697e-24], [1351.85, 5.62508e-23, 1.48068e-23, 6.66344e-24, 1.32766e-23, 7.21099e-23, 8.48859e-24], [1354.92, 5.63783e-23, 1.48327e-23, 6.67487e-24, 1.33126e-23, 7.22947e-23, 8.56041e-24], [1358.0, 5.65061e-23, 1.48586e-23, 6.68633e-24, 1.33487e-23, 7.24798e-23, 8.63242e-24], [1361.09, 5.66342e-23, 1.48847e-23, 6.69782e-24, 1.33848e-23, 7.26653e-23, 8.70463e-24], [1364.19, 5.67626e-23, 1.49107e-23, 6.70934e-24, 1.3421e-23, 7.28512e-23, 8.77703e-24], [1367.29, 5.68912e-23, 1.49369e-23, 6.7209e-24, 1.34573e-23, 7.30375e-23, 8.84962e-24], [1370.4, 5.70202e-23, 1.49631e-23, 6.73249e-24, 1.34937e-23, 7.32242e-23, 8.92241e-24], [1373.52, 5.71495e-23, 1.49894e-23, 6.74411e-24, 1.35302e-23, 7.34113e-23, 8.99539e-24], [1376.64, 5.7279e-23, 1.50158e-23, 6.75577e-24, 1.35667e-23, 7.35988e-23, 9.06855e-24], [1379.77, 5.74089e-23, 1.50423e-23, 6.76746e-24, 1.36033e-23, 7.37866e-23, 9.14191e-24], [1382.91, 5.7539e-23, 1.50688e-23, 6.77918e-24, 1.364e-23, 7.39749e-23, 9.21546e-24], [1386.06, 5.76695e-23, 1.50954e-23, 6.79094e-24, 1.36767e-23, 7.41635e-23, 9.2892e-24], [1389.21, 5.78002e-23, 1.51221e-23, 6.80273e-24, 1.37136e-23, 7.43525e-23, 9.36313e-24], [1392.37, 5.79312e-23, 1.51488e-23, 6.81455e-24, 1.37505e-23, 7.45419e-23, 9.43725e-24], [1395.54, 5.80626e-23, 1.51756e-23, 6.82641e-24, 1.37875e-23, 7.47318e-23, 9.51156e-24], [1398.71, 5.81942e-23, 1.52025e-23, 6.83831e-24, 1.38245e-23, 7.4922e-23, 9.58605e-24], [1401.89, 5.83261e-23, 1.52295e-23, 6.85023e-24, 1.38617e-23, 7.51126e-23, 9.66073e-24], [1405.08, 5.84584e-23, 1.52566e-23, 6.86219e-24, 1.38989e-23, 7.53036e-23, 9.7356e-24], [1408.28, 5.85909e-23, 1.52837e-23, 6.87419e-24, 1.39362e-23, 7.5495e-23, 9.81066e-24], [1411.48, 5.87237e-23, 1.53109e-23, 6.88622e-24, 1.39736e-23, 7.56868e-23, 9.8859e-24], [1414.69, 5.88569e-23, 1.53382e-23, 6.89828e-24, 1.4011e-23, 7.5879e-23, 9.96132e-24], [1417.91, 5.89903e-23, 1.53655e-23, 6.91038e-24, 1.40486e-23, 7.60716e-23, 1.00369e-23], [1421.14, 5.9124e-23, 1.53929e-23, 6.92251e-24, 1.40862e-23, 7.62646e-23, 1.01127e-23], [1424.37, 5.92581e-23, 1.54204e-23, 6.93468e-24, 1.41239e-23, 7.6458e-23, 1.01887e-23], [1427.61, 5.93925e-23, 1.5448e-23, 6.94689e-24, 1.41616e-23, 7.66519e-23, 1.02649e-23], [1430.85, 5.95271e-23, 1.54757e-23, 6.95913e-24, 1.41995e-23, 7.68461e-23, 1.03412e-23], [1434.11, 5.96621e-23, 1.55034e-23, 6.97144e-24, 1.42374e-23, 7.70407e-23, 1.04178e-23], [1437.37, 5.97974e-23, 1.55313e-23, 6.9839e-24, 1.42755e-23, 7.72358e-23, 1.04946e-23], [1440.64, 5.99343e-23, 1.55644e-23, 7.00786e-24, 1.43193e-23, 7.74322e-23, 1.05792e-23], [1443.92, 6.00689e-23, 1.55872e-23, 7.00881e-24, 1.43519e-23, 7.76271e-23, 1.06487e-23], [1447.2, 6.0205e-23, 1.56151e-23, 7.02088e-24, 1.439e-23, 7.78233e-23, 1.07258e-23], [1450.49, 6.03415e-23, 1.56432e-23, 7.03328e-24, 1.44283e-23, 7.802e-23, 1.08032e-23], [1453.79, 6.04784e-23, 1.56714e-23, 7.04574e-24, 1.44667e-23, 7.8217e-23, 1.08808e-23], [1457.1, 6.06155e-23, 1.56997e-23, 7.05824e-24, 1.45051e-23, 7.84145e-23, 1.09586e-23], [1460.41, 6.07529e-23, 1.5728e-23, 7.07079e-24, 1.45437e-23, 7.86124e-23, 1.10366e-23], [1463.74, 6.08907e-23, 1.57565e-23, 7.08337e-24, 1.45824e-23, 7.88108e-23, 1.11148e-23], [1467.07, 6.10288e-23, 1.5785e-23, 7.09599e-24, 1.46211e-23, 7.90095e-23, 1.11931e-23], [1470.4, 6.11672e-23, 1.58135e-23, 7.10864e-24, 1.46599e-23, 7.92087e-23, 1.12717e-23], [1473.75, 6.13059e-23, 1.58422e-23, 7.12133e-24, 1.46988e-23, 7.94082e-23, 1.13504e-23], [1477.1, 6.14449e-23, 1.5871e-23, 7.13406e-24, 1.47378e-23, 7.96082e-23, 1.14293e-23], [1480.46, 6.15842e-23, 1.58998e-23, 7.14682e-24, 1.47768e-23, 7.98086e-23, 1.15084e-23], [1483.83, 6.17239e-23, 1.59287e-23, 7.15962e-24, 1.48159e-23, 8.00095e-23, 1.15876e-23], [1487.2, 6.18638e-23, 1.59576e-23, 7.17245e-24, 1.48552e-23, 8.02107e-23, 1.16671e-23], [1490.58, 6.20041e-23, 1.59867e-23, 7.18532e-24, 1.48945e-23, 8.04124e-23, 1.17467e-23], [1493.97, 6.21447e-23, 1.60158e-23, 7.19822e-24, 1.49338e-23, 8.06145e-23, 1.18265e-23], [1497.37, 6.22857e-23, 1.60451e-23, 7.21116e-24, 1.49733e-23, 8.0817e-23, 1.19064e-23], [1500.78, 6.24269e-23, 1.60744e-23, 7.22414e-24, 1.50129e-23, 8.10199e-23, 1.19866e-23], [1504.19, 6.25685e-23, 1.61037e-23, 7.23715e-24, 1.50525e-23, 8.12233e-23, 1.20669e-23], [1507.61, 6.27104e-23, 1.61332e-23, 7.2502e-24, 1.50922e-23, 8.14271e-23, 1.21475e-23], [1511.04, 6.28526e-23, 1.61627e-23, 7.26329e-24, 1.5132e-23, 8.16314e-23, 1.22281e-23], [1514.48, 6.29951e-23, 1.61923e-23, 7.27641e-24, 1.51719e-23, 8.1836e-23, 1.2309e-23], [1517.92, 6.3138e-23, 1.6222e-23, 7.28957e-24, 1.52118e-23, 8.20411e-23, 1.23901e-23], [1521.38, 6.32812e-23, 1.62518e-23, 7.30276e-24, 1.52519e-23, 8.22466e-23, 1.24713e-23], [1524.84, 6.34247e-23, 1.62817e-23, 7.31599e-24, 1.5292e-23, 8.24526e-23, 1.25527e-23], [1528.31, 6.35686e-23, 1.63116e-23, 7.32926e-24, 1.53322e-23, 8.2659e-23, 1.26343e-23], [1531.78, 6.37127e-23, 1.63416e-23, 7.34257e-24, 1.53725e-23, 8.28658e-23, 1.27161e-23], [1535.27, 6.38572e-23, 1.63717e-23, 7.35591e-24, 1.54129e-23, 8.3073e-23, 1.2798e-23], [1538.76, 6.40021e-23, 1.64019e-23, 7.36929e-24, 1.54534e-23, 8.32807e-23, 1.28801e-23], [1542.26, 6.41472e-23, 1.64322e-23, 7.3827e-24, 1.54939e-23, 8.34889e-23, 1.29624e-23], [1545.77, 6.42927e-23, 1.64625e-23, 7.39616e-24, 1.55345e-23, 8.36974e-23, 1.30449e-23], [1549.28, 6.44385e-23, 1.6493e-23, 7.40965e-24, 1.55753e-23, 8.39064e-23, 1.31275e-23], [1552.81, 6.45847e-23, 1.65235e-23, 7.42317e-24, 1.56161e-23, 8.41159e-23, 1.32104e-23], [1556.34, 6.47312e-23, 1.65541e-23, 7.43674e-24, 1.5657e-23, 8.43258e-23, 1.32934e-23], [1559.88, 6.4878e-23, 1.65848e-23, 7.45034e-24, 1.56979e-23, 8.45361e-23, 1.33766e-23], [1563.43, 6.50251e-23, 1.66155e-23, 7.46398e-24, 1.5739e-23, 8.47469e-23, 1.34599e-23], [1566.98, 6.51726e-23, 1.66464e-23, 7.47766e-24, 1.57802e-23, 8.49581e-23, 1.35435e-23], [1570.55, 6.53205e-23, 1.66773e-23, 7.49137e-24, 1.58214e-23, 8.51698e-23, 1.36272e-23], [1574.12, 6.54686e-23, 1.67083e-23, 7.50512e-24, 1.58627e-23, 8.53819e-23, 1.37111e-23], [1577.7, 6.56171e-23, 1.67394e-23, 7.51891e-24, 1.59041e-23, 8.55945e-23, 1.37952e-23], [1581.29, 6.57659e-23, 1.67706e-23, 7.53274e-24, 1.59456e-23, 8.58075e-23, 1.38794e-23], [1584.88, 6.59151e-23, 1.68018e-23, 7.54661e-24, 1.59872e-23, 8.60209e-23, 1.39639e-23], [1588.49, 6.60646e-23, 1.68332e-23, 7.56051e-24, 1.60289e-23, 8.62348e-23, 1.40485e-23], [1592.1, 6.62145e-23, 1.68646e-23, 7.57445e-24, 1.60706e-23, 8.64492e-23, 1.41333e-23], [1595.72, 6.63647e-23, 1.68961e-23, 7.58843e-24, 1.61125e-23, 8.6664e-23, 1.42182e-23], [1599.35, 6.65152e-23, 1.69277e-23, 7.60245e-24, 1.61544e-23, 8.68793e-23, 1.43034e-23], [1602.99, 6.66661e-23, 1.69594e-23, 7.6165e-24, 1.61964e-23, 8.7095e-23, 1.43887e-23], [1606.64, 6.68173e-23, 1.69912e-23, 7.6306e-24, 1.62385e-23, 8.73112e-23, 1.44742e-23], [1610.29, 6.69689e-23, 1.70231e-23, 7.64473e-24, 1.62807e-23, 8.75278e-23, 1.45599e-23], [1613.95, 6.71208e-23, 1.7055e-23, 7.6589e-24, 1.6323e-23, 8.77449e-23, 1.46457e-23], [1617.63, 6.72731e-23, 1.7087e-23, 7.67311e-24, 1.63654e-23, 8.79625e-23, 1.47317e-23], [1621.31, 6.74257e-23, 1.71191e-23, 7.68736e-24, 1.64078e-23, 8.81805e-23, 1.48179e-23], [1624.99, 6.75786e-23, 1.71513e-23, 7.70165e-24, 1.64504e-23, 8.8399e-23, 1.49043e-23], [1628.69, 6.77319e-23, 1.71836e-23, 7.71598e-24, 1.6493e-23, 8.86179e-23, 1.49909e-23], [1632.39, 6.78856e-23, 1.7216e-23, 7.73034e-24, 1.65357e-23, 8.88374e-23, 1.50776e-23], [1636.11, 6.80396e-23, 1.72484e-23, 7.74475e-24, 1.65786e-23, 8.90572e-23, 1.51645e-23], [1639.83, 6.81939e-23, 1.7281e-23, 7.75919e-24, 1.66215e-23, 8.92776e-23, 1.52516e-23], [1643.56, 6.83486e-23, 1.73136e-23, 7.77367e-24, 1.66645e-23, 8.94984e-23, 1.53389e-23], [1647.3, 6.85037e-23, 1.73463e-23, 7.7882e-24, 1.67075e-23, 8.97197e-23, 1.54264e-23], [1651.04, 6.86591e-23, 1.73791e-23, 7.80276e-24, 1.67507e-23, 8.99414e-23, 1.5514e-23], [1654.8, 6.88149e-23, 1.7412e-23, 7.81736e-24, 1.6794e-23, 9.01636e-23, 1.56018e-23], [1658.56, 6.8971e-23, 1.7445e-23, 7.832e-24, 1.68374e-23, 9.03863e-23, 1.56898e-23], [1662.34, 6.91275e-23, 1.74781e-23, 7.84668e-24, 1.68808e-23, 9.06095e-23, 1.57779e-23], [1666.12, 6.92843e-23, 1.75112e-23, 7.8614e-24, 1.69243e-23, 9.08331e-23, 1.58663e-23], [1669.91, 6.94415e-23, 1.75445e-23, 7.87616e-24, 1.6968e-23, 9.10572e-23, 1.59548e-23], [1673.7, 6.9599e-23, 1.75778e-23, 7.89096e-24, 1.70117e-23, 9.12818e-23, 1.60435e-23], [1677.51, 6.97569e-23, 1.76112e-23, 7.9058e-24, 1.70555e-23, 9.15068e-23, 1.61324e-23], [1681.33, 6.99152e-23, 1.76447e-23, 7.92067e-24, 1.70994e-23, 9.17324e-23, 1.62214e-23], [1685.15, 7.00738e-23, 1.76783e-23, 7.93559e-24, 1.71434e-23, 9.19584e-23, 1.63106e-23], [1688.98, 7.02328e-23, 1.7712e-23, 7.95055e-24, 1.71875e-23, 9.21849e-23, 1.64e-23], [1692.83, 7.03921e-23, 1.77458e-23, 7.96555e-24, 1.72317e-23, 9.24119e-23, 1.64896e-23], [1696.68, 7.05518e-23, 1.77797e-23, 7.98059e-24, 1.7276e-23, 9.26394e-23, 1.65794e-23], [1700.54, 7.07119e-23, 1.78136e-23, 7.99567e-24, 1.73203e-23, 9.28673e-23, 1.66693e-23], [1704.4, 7.08723e-23, 1.78476e-23, 8.01079e-24, 1.73648e-23, 9.30957e-23, 1.67595e-23], [1708.28, 7.10331e-23, 1.78818e-23, 8.02595e-24, 1.74094e-23, 9.33247e-23, 1.68498e-23], [1712.17, 7.11943e-23, 1.7916e-23, 8.04116e-24, 1.7454e-23, 9.35541e-23, 1.69402e-23], [1716.06, 7.13558e-23, 1.79503e-23, 8.0564e-24, 1.74988e-23, 9.3784e-23, 1.70309e-23], [1719.96, 7.15177e-23, 1.79847e-23, 8.07168e-24, 1.75436e-23, 9.40143e-23, 1.71217e-23], [1723.88, 7.168e-23, 1.80192e-23, 8.08701e-24, 1.75885e-23, 9.42452e-23, 1.72128e-23], [1727.8, 7.18427e-23, 1.80538e-23, 8.10237e-24, 1.76336e-23, 9.44766e-23, 1.73039e-23], [1731.73, 7.20057e-23, 1.80885e-23, 8.11778e-24, 1.76787e-23, 9.47085e-23, 1.73953e-23], [1735.67, 7.2169e-23, 1.81233e-23, 8.13323e-24, 1.77239e-23, 9.49408e-23, 1.74869e-23], [1739.61, 7.23328e-23, 1.81581e-23, 8.14872e-24, 1.77692e-23, 9.51737e-23, 1.75786e-23], [1743.57, 7.24969e-23, 1.81931e-23, 8.16425e-24, 1.78146e-23, 9.5407e-23, 1.76705e-23], [1747.54, 7.26614e-23, 1.82281e-23, 8.17982e-24, 1.78601e-23, 9.56409e-23, 1.77626e-23], [1751.51, 7.28263e-23, 1.82632e-23, 8.19543e-24, 1.79057e-23, 9.58752e-23, 1.78549e-23], [1755.5, 7.29915e-23, 1.82985e-23, 8.21109e-24, 1.79514e-23, 9.61101e-23, 1.79473e-23], [1759.49, 7.31572e-23, 1.83338e-23, 8.22678e-24, 1.79972e-23, 9.63454e-23, 1.804e-23], [1763.49, 7.33232e-23, 1.83692e-23, 8.24252e-24, 1.80431e-23, 9.65813e-23, 1.81328e-23], [1767.5, 7.34895e-23, 1.84047e-23, 8.2583e-24, 1.80891e-23, 9.68176e-23, 1.82258e-23], [1771.52, 7.36563e-23, 1.84403e-23, 8.27412e-24, 1.81352e-23, 9.70545e-23, 1.83189e-23], [1775.55, 7.38234e-23, 1.8476e-23, 8.28999e-24, 1.81813e-23, 9.72918e-23, 1.84123e-23], [1779.59, 7.3991e-23, 1.85118e-23, 8.30589e-24, 1.82276e-23, 9.75297e-23, 1.85058e-23], [1783.64, 7.41589e-23, 1.85477e-23, 8.32184e-24, 1.8274e-23, 9.77681e-23, 1.85996e-23], [1787.69, 7.43271e-23, 1.85837e-23, 8.33783e-24, 1.83204e-23, 9.8007e-23, 1.86934e-23], [1791.76, 7.44958e-23, 1.86197e-23, 8.35386e-24, 1.8367e-23, 9.82464e-23, 1.87875e-23], [1795.84, 7.46649e-23, 1.86559e-23, 8.36994e-24, 1.84137e-23, 9.84863e-23, 1.88818e-23], [1799.92, 7.48343e-23, 1.86921e-23, 8.38606e-24, 1.84604e-23, 9.87267e-23, 1.89762e-23], [1804.02, 7.50041e-23, 1.87285e-23, 8.40222e-24, 1.85073e-23, 9.89677e-23, 1.90708e-23], [1808.12, 7.51743e-23, 1.87649e-23, 8.41842e-24, 1.85543e-23, 9.92091e-23, 1.91656e-23], [1812.23, 7.53449e-23, 1.88015e-23, 8.43467e-24, 1.86013e-23, 9.94511e-23, 1.92606e-23], [1816.35, 7.55159e-23, 1.88381e-23, 8.45096e-24, 1.86485e-23, 9.96936e-23, 1.93558e-23], [1820.48, 7.56873e-23, 1.88748e-23, 8.46729e-24, 1.86957e-23, 9.99366e-23, 1.94511e-23], [1824.63, 7.5859e-23, 1.89117e-23, 8.48366e-24, 1.87431e-23, 1.0018e-22, 1.95467e-23], [1828.78, 7.60312e-23, 1.89486e-23, 8.50008e-24, 1.87906e-23, 1.00424e-22, 1.96424e-23], [1832.94, 7.62037e-23, 1.89856e-23, 8.51654e-24, 1.88381e-23, 1.00669e-22, 1.97383e-23], [1837.1, 7.63767e-23, 1.90227e-23, 8.53305e-24, 1.88858e-23, 1.00914e-22, 1.98344e-23], [1841.28, 7.655e-23, 1.90599e-23, 8.5496e-24, 1.89335e-23, 1.0116e-22, 1.99306e-23], [1845.47, 7.67237e-23, 1.90972e-23, 8.56619e-24, 1.89814e-23, 1.01406e-22, 2.00271e-23], [1849.67, 7.68978e-23, 1.91346e-23, 8.58282e-24, 1.90294e-23, 1.01652e-22, 2.01237e-23], [1853.88, 7.70724e-23, 1.91721e-23, 8.5995e-24, 1.90774e-23, 1.019e-22, 2.02205e-23], [1858.09, 7.72473e-23, 1.92097e-23, 8.61622e-24, 1.91256e-23, 1.02147e-22, 2.03175e-23], [1862.32, 7.74226e-23, 1.92474e-23, 8.63299e-24, 1.91738e-23, 1.02396e-22, 2.04147e-23], [1866.56, 7.75983e-23, 1.92852e-23, 8.6498e-24, 1.92222e-23, 1.02644e-22, 2.0512e-23], [1870.8, 7.77744e-23, 1.93231e-23, 8.66666e-24, 1.92707e-23, 1.02894e-22, 2.06096e-23], [1875.06, 7.79509e-23, 1.93611e-23, 8.68356e-24, 1.93193e-23, 1.03144e-22, 2.07073e-23], [1879.32, 7.81278e-23, 1.93992e-23, 8.7005e-24, 1.93679e-23, 1.03394e-22, 2.08052e-23], [1883.6, 7.83052e-23, 1.94374e-23, 8.71749e-24, 1.94167e-23, 1.03645e-22, 2.09033e-23], [1887.88, 7.84829e-23, 1.94757e-23, 8.73452e-24, 1.94656e-23, 1.03896e-22, 2.10016e-23], [1892.17, 7.8661e-23, 1.95141e-23, 8.7516e-24, 1.95146e-23, 1.04148e-22, 2.11e-23], [1896.48, 7.88395e-23, 1.95525e-23, 8.76872e-24, 1.95637e-23, 1.04401e-22, 2.11987e-23], [1900.79, 7.90185e-23, 1.95911e-23, 8.78588e-24, 1.96129e-23, 1.04654e-22, 2.12975e-23], [1905.12, 7.91978e-23, 1.96298e-23, 8.8031e-24, 1.96621e-23, 1.04908e-22, 2.13966e-23], [1909.45, 7.93776e-23, 1.96686e-23, 8.82035e-24, 1.97115e-23, 1.05162e-22, 2.14958e-23], [1913.79, 7.95577e-23, 1.97075e-23, 8.83766e-24, 1.97611e-23, 1.05417e-22, 2.15952e-23], [1918.14, 7.97383e-23, 1.97465e-23, 8.85507e-24, 1.98107e-23, 1.05672e-22, 2.16948e-23], [1922.51, 7.99194e-23, 1.9786e-23, 8.87332e-24, 1.98608e-23, 1.05928e-22, 2.17949e-23], [1926.88, 8.01007e-23, 1.98247e-23, 8.88985e-24, 1.99102e-23, 1.06184e-22, 2.18945e-23], [1931.26, 8.02825e-23, 1.9864e-23, 8.90731e-24, 1.99601e-23, 1.06441e-22, 2.19946e-23], [1935.66, 8.04647e-23, 1.99034e-23, 8.92483e-24, 2.00101e-23, 1.06699e-22, 2.20949e-23], [1940.06, 8.06473e-23, 1.99429e-23, 8.9424e-24, 2.00603e-23, 1.06957e-22, 2.21954e-23], [1944.47, 8.08304e-23, 1.99825e-23, 8.96002e-24, 2.01105e-23, 1.07216e-22, 2.22961e-23], [1948.89, 8.10139e-23, 2.00222e-23, 8.97768e-24, 2.01609e-23, 1.07475e-22, 2.2397e-23], [1953.33, 8.11977e-23, 2.00619e-23, 8.99539e-24, 2.02113e-23, 1.07734e-22, 2.24981e-23], [1957.77, 8.1382e-23, 2.01018e-23, 9.01315e-24, 2.02619e-23, 1.07995e-22, 2.25993e-23], [1962.22, 8.15668e-23, 2.01418e-23, 9.03095e-24, 2.03126e-23, 1.08256e-22, 2.27008e-23], [1966.69, 8.17519e-23, 2.01819e-23, 9.0488e-24, 2.03633e-23, 1.08517e-22, 2.28024e-23], [1971.16, 8.19375e-23, 2.02221e-23, 9.06669e-24, 2.04142e-23, 1.08779e-22, 2.29042e-23], [1975.64, 8.21235e-23, 2.02624e-23, 9.08463e-24, 2.04652e-23, 1.09042e-22, 2.30063e-23], [1980.14, 8.23099e-23, 2.03028e-23, 9.10262e-24, 2.05163e-23, 1.09305e-22, 2.31085e-23], [1984.64, 8.24967e-23, 2.03433e-23, 9.12065e-24, 2.05675e-23, 1.09568e-22, 2.32109e-23], [1989.16, 8.2684e-23, 2.0384e-23, 9.13873e-24, 2.06188e-23, 1.09833e-22, 2.33134e-23], [1993.68, 8.28717e-23, 2.04247e-23, 9.15686e-24, 2.06703e-23, 1.10097e-22, 2.34162e-23], [1998.21, 8.30598e-23, 2.04655e-23, 9.17503e-24, 2.07218e-23, 1.10363e-22, 2.35192e-23], [2002.76, 8.32483e-23, 2.05064e-23, 9.19325e-24, 2.07735e-23, 1.10629e-22, 2.36223e-23], [2007.32, 8.34373e-23, 2.05475e-23, 9.21152e-24, 2.08252e-23, 1.10895e-22, 2.37257e-23], [2011.88, 8.36267e-23, 2.05886e-23, 9.22983e-24, 2.08771e-23, 1.11162e-22, 2.38292e-23], [2016.46, 8.38165e-23, 2.06298e-23, 9.24819e-24, 2.09291e-23, 1.1143e-22, 2.39329e-23], [2021.04, 8.40068e-23, 2.06712e-23, 9.2666e-24, 2.09811e-23, 1.11698e-22, 2.40368e-23], [2025.64, 8.41975e-23, 2.07126e-23, 9.28506e-24, 2.10333e-23, 1.11967e-22, 2.41409e-23], [2030.25, 8.43886e-23, 2.07542e-23, 9.30356e-24, 2.10856e-23, 1.12237e-22, 2.42452e-23], [2034.87, 8.45802e-23, 2.07958e-23, 9.32211e-24, 2.11381e-23, 1.12507e-22, 2.43497e-23], [2039.49, 8.47722e-23, 2.08376e-23, 9.34071e-24, 2.11906e-23, 1.12777e-22, 2.44544e-23], [2044.13, 8.49646e-23, 2.08795e-23, 9.35936e-24, 2.12432e-23, 1.13049e-22, 2.45593e-23], [2048.78, 8.51575e-23, 2.09215e-23, 9.37805e-24, 2.1296e-23, 1.1332e-22, 2.46643e-23], [2053.44, 8.53508e-23, 2.09635e-23, 9.39679e-24, 2.13489e-23, 1.13593e-22, 2.47696e-23], [2058.11, 8.55446e-23, 2.10057e-23, 9.41558e-24, 2.14018e-23, 1.13866e-22, 2.4875e-23], [2062.8, 8.57388e-23, 2.1048e-23, 9.43442e-24, 2.14549e-23, 1.14139e-22, 2.49807e-23], [2067.49, 8.59334e-23, 2.10904e-23, 9.45331e-24, 2.15081e-23, 1.14413e-22, 2.50865e-23], [2072.19, 8.61285e-23, 2.1133e-23, 9.47224e-24, 2.15614e-23, 1.14688e-22, 2.51925e-23], [2076.9, 8.6324e-23, 2.11756e-23, 9.49123e-24, 2.16149e-23, 1.14963e-22, 2.52988e-23], [2081.63, 8.652e-23, 2.12183e-23, 9.51026e-24, 2.16684e-23, 1.15239e-22, 2.54052e-23], [2086.36, 8.67164e-23, 2.12612e-23, 9.52934e-24, 2.17221e-23, 1.15516e-22, 2.55118e-23], [2091.11, 8.69133e-23, 2.13041e-23, 9.54847e-24, 2.17759e-23, 1.15793e-22, 2.56186e-23], [2095.86, 8.71106e-23, 2.13472e-23, 9.56765e-24, 2.18297e-23, 1.16071e-22, 2.57256e-23], [2100.63, 8.73084e-23, 2.13903e-23, 9.58688e-24, 2.18837e-23, 1.16349e-22, 2.58328e-23], [2105.41, 8.75066e-23, 2.14336e-23, 9.60615e-24, 2.19379e-23, 1.16628e-22, 2.59402e-23], [2110.2, 8.77052e-23, 2.1477e-23, 9.62548e-24, 2.19921e-23, 1.16908e-22, 2.60478e-23], [2115.0, 8.79043e-23, 2.15205e-23, 9.64485e-24, 2.20464e-23, 1.17188e-22, 2.61556e-23], [2119.81, 8.81039e-23, 2.15641e-23, 9.66428e-24, 2.21009e-23, 1.17468e-22, 2.62635e-23], [2124.63, 8.83039e-23, 2.16078e-23, 9.68375e-24, 2.21555e-23, 1.1775e-22, 2.63717e-23], [2129.46, 8.85044e-23, 2.16516e-23, 9.70328e-24, 2.22102e-23, 1.18032e-22, 2.64801e-23], [2134.31, 8.87053e-23, 2.16955e-23, 9.72285e-24, 2.2265e-23, 1.18314e-22, 2.65887e-23], [2139.16, 8.89067e-23, 2.17396e-23, 9.74247e-24, 2.23199e-23, 1.18598e-22, 2.66974e-23], [2144.03, 8.91086e-23, 2.17838e-23, 9.76215e-24, 2.2375e-23, 1.18881e-22, 2.68064e-23], [2148.9, 8.93109e-23, 2.1828e-23, 9.78187e-24, 2.24301e-23, 1.19166e-22, 2.69156e-23], [2153.79, 8.95137e-23, 2.18724e-23, 9.80164e-24, 2.24854e-23, 1.19451e-22, 2.70249e-23], [2158.69, 8.97169e-23, 2.19169e-23, 9.82147e-24, 2.25408e-23, 1.19737e-22, 2.71345e-23], [2163.6, 8.99206e-23, 2.19615e-23, 9.84134e-24, 2.25964e-23, 1.20023e-22, 2.72443e-23], [2168.52, 9.01247e-23, 2.20062e-23, 9.86127e-24, 2.2652e-23, 1.2031e-22, 2.73542e-23], [2173.45, 9.03294e-23, 2.2051e-23, 9.88124e-24, 2.27077e-23, 1.20597e-22, 2.74644e-23], [2178.4, 9.05344e-23, 2.2096e-23, 9.90127e-24, 2.27636e-23, 1.20885e-22, 2.75747e-23], [2183.35, 9.074e-23, 2.2141e-23, 9.92134e-24, 2.28196e-23, 1.21174e-22, 2.76853e-23], [2188.32, 9.0946e-23, 2.21862e-23, 9.94147e-24, 2.28757e-23, 1.21464e-22, 2.77961e-23], [2193.3, 9.11525e-23, 2.22315e-23, 9.96165e-24, 2.2932e-23, 1.21754e-22, 2.7907e-23], [2198.29, 9.13595e-23, 2.22769e-23, 9.98188e-24, 2.29883e-23, 1.22044e-22, 2.80182e-23], [2203.29, 9.15669e-23, 2.23224e-23, 1.00022e-23, 2.30448e-23, 1.22336e-22, 2.81295e-23], [2208.3, 9.17748e-23, 2.2368e-23, 1.00225e-23, 2.31014e-23, 1.22628e-22, 2.82411e-23], [2213.32, 9.19832e-23, 2.24137e-23, 1.00429e-23, 2.31581e-23, 1.2292e-22, 2.83529e-23], [2218.35, 9.2192e-23, 2.24596e-23, 1.00633e-23, 2.3215e-23, 1.23213e-22, 2.84648e-23], [2223.4, 9.24014e-23, 2.25056e-23, 1.00838e-23, 2.32719e-23, 1.23507e-22, 2.8577e-23], [2228.46, 9.26112e-23, 2.25516e-23, 1.01043e-23, 2.3329e-23, 1.23802e-22, 2.86894e-23], [2233.53, 9.28214e-23, 2.25978e-23, 1.01249e-23, 2.33862e-23, 1.24097e-22, 2.8802e-23], [2238.61, 9.30322e-23, 2.26441e-23, 1.01456e-23, 2.34435e-23, 1.24393e-22, 2.89148e-23], [2243.7, 9.32434e-23, 2.26906e-23, 1.01663e-23, 2.3501e-23, 1.24689e-22, 2.90277e-23], [2248.8, 9.34551e-23, 2.27371e-23, 1.0187e-23, 2.35586e-23, 1.24986e-22, 2.91409e-23], [2253.92, 9.36673e-23, 2.27838e-23, 1.02078e-23, 2.36163e-23, 1.25284e-22, 2.92543e-23], [2259.04, 9.388e-23, 2.28305e-23, 1.02287e-23, 2.36741e-23, 1.25582e-22, 2.93679e-23], [2264.18, 9.40932e-23, 2.28774e-23, 1.02496e-23, 2.3732e-23, 1.25881e-22, 2.94817e-23], [2269.33, 9.43069e-23, 2.29245e-23, 1.02705e-23, 2.37901e-23, 1.26181e-22, 2.95957e-23], [2274.49, 9.4521e-23, 2.29716e-23, 1.02915e-23, 2.38483e-23, 1.26481e-22, 2.97099e-23], [2279.67, 9.47356e-23, 2.30188e-23, 1.03126e-23, 2.39066e-23, 1.26782e-22, 2.98244e-23], [2284.85, 9.49507e-23, 2.30662e-23, 1.03337e-23, 2.39651e-23, 1.27084e-22, 2.9939e-23], [2290.05, 9.51663e-23, 2.31137e-23, 1.03549e-23, 2.40236e-23, 1.27386e-22, 3.00538e-23], [2295.26, 9.53824e-23, 2.31613e-23, 1.03761e-23, 2.40823e-23, 1.27689e-22, 3.01689e-23], [2300.48, 9.5599e-23, 2.3209e-23, 1.03974e-23, 2.41412e-23, 1.27993e-22, 3.02841e-23], [2305.71, 9.58161e-23, 2.32568e-23, 1.04187e-23, 2.42001e-23, 1.28297e-22, 3.03996e-23], [2310.96, 9.60337e-23, 2.33048e-23, 1.04401e-23, 2.42592e-23, 1.28602e-22, 3.05153e-23], [2316.21, 9.62517e-23, 2.33529e-23, 1.04615e-23, 2.43184e-23, 1.28908e-22, 3.06311e-23], [2321.48, 9.64703e-23, 2.34011e-23, 1.0483e-23, 2.43777e-23, 1.29214e-22, 3.07472e-23], [2326.76, 9.66894e-23, 2.34494e-23, 1.05045e-23, 2.44372e-23, 1.29521e-22, 3.08635e-23], [2332.05, 9.69089e-23, 2.34978e-23, 1.05261e-23, 2.44967e-23, 1.29828e-22, 3.098e-23], [2337.36, 9.7129e-23, 2.35464e-23, 1.05478e-23, 2.45565e-23, 1.30137e-22, 3.10967e-23], [2342.67, 9.73495e-23, 2.3595e-23, 1.05695e-23, 2.46163e-23, 1.30446e-22, 3.12137e-23], [2348.0, 9.75706e-23, 2.36438e-23, 1.05912e-23, 2.46763e-23, 1.30755e-22, 3.13308e-23], [2353.34, 9.77922e-23, 2.36927e-23, 1.0613e-23, 2.47364e-23, 1.31066e-22, 3.14481e-23], [2358.7, 9.80142e-23, 2.37418e-23, 1.06349e-23, 2.47966e-23, 1.31377e-22, 3.15657e-23], [2364.06, 9.82368e-23, 2.37909e-23, 1.06568e-23, 2.48569e-23, 1.31688e-22, 3.16835e-23], [2369.44, 9.84599e-23, 2.38402e-23, 1.06788e-23, 2.49174e-23, 1.32001e-22, 3.18014e-23], [2374.83, 9.86835e-23, 2.38896e-23, 1.07008e-23, 2.4978e-23, 1.32314e-22, 3.19196e-23], [2380.23, 9.89076e-23, 2.39392e-23, 1.07229e-23, 2.50388e-23, 1.32628e-22, 3.2038e-23], [2385.64, 9.91322e-23, 2.39888e-23, 1.07451e-23, 2.50997e-23, 1.32942e-22, 3.21567e-23], [2391.07, 9.93573e-23, 2.40386e-23, 1.07673e-23, 2.51607e-23, 1.33257e-22, 3.22755e-23], [2396.51, 9.95829e-23, 2.40885e-23, 1.07895e-23, 2.52218e-23, 1.33573e-22, 3.23946e-23], [2401.96, 9.98576e-23, 2.43384e-23, 1.1251e-23, 2.5474e-23, 1.33926e-22, 3.26625e-23], [2407.42, 1.00036e-22, 2.41887e-23, 1.08342e-23, 2.53445e-23, 1.34207e-22, 3.26333e-23], [2412.9, 1.00263e-22, 2.42389e-23, 1.08566e-23, 2.5406e-23, 1.34524e-22, 3.2753e-23], [2418.39, 1.00491e-22, 2.42893e-23, 1.08791e-23, 2.54677e-23, 1.34843e-22, 3.28729e-23], [2423.89, 1.00719e-22, 2.43398e-23, 1.09016e-23, 2.55295e-23, 1.35162e-22, 3.2993e-23], [2429.4, 1.00948e-22, 2.43905e-23, 1.09242e-23, 2.55914e-23, 1.35482e-22, 3.31134e-23], [2434.93, 1.01177e-22, 2.44412e-23, 1.09468e-23, 2.56535e-23, 1.35803e-22, 3.32339e-23], [2440.47, 1.01407e-22, 2.44921e-23, 1.09695e-23, 2.57157e-23, 1.36124e-22, 3.33547e-23], [2446.02, 1.01637e-22, 2.45431e-23, 1.09923e-23, 2.5778e-23, 1.36446e-22, 3.34757e-23], [2451.58, 1.01868e-22, 2.45943e-23, 1.10151e-23, 2.58405e-23, 1.36769e-22, 3.35969e-23], [2457.16, 1.02099e-22, 2.46456e-23, 1.1038e-23, 2.59031e-23, 1.37093e-22, 3.37183e-23], [2462.75, 1.02331e-22, 2.4697e-23, 1.10609e-23, 2.59659e-23, 1.37417e-22, 3.384e-23], [2468.35, 1.02563e-22, 2.47485e-23, 1.10839e-23, 2.60287e-23, 1.37742e-22, 3.39619e-23], [2473.96, 1.02796e-22, 2.48001e-23, 1.11069e-23, 2.60918e-23, 1.38067e-22, 3.40839e-23], [2479.59, 1.0303e-22, 2.48519e-23, 1.113e-23, 2.61549e-23, 1.38394e-22, 3.42063e-23], [2485.23, 1.03264e-22, 2.49038e-23, 1.11532e-23, 2.62182e-23, 1.38721e-22, 3.43288e-23], [2490.88, 1.03498e-22, 2.49558e-23, 1.11764e-23, 2.62816e-23, 1.39049e-22, 3.44515e-23], [2496.55, 1.03733e-22, 2.5008e-23, 1.11997e-23, 2.63452e-23, 1.39377e-22, 3.45745e-23], [2502.23, 1.03969e-22, 2.50603e-23, 1.1223e-23, 2.64089e-23, 1.39706e-22, 3.46977e-23], [2507.92, 1.04205e-22, 2.51127e-23, 1.12464e-23, 2.64727e-23, 1.40036e-22, 3.48211e-23], [2513.62, 1.04442e-22, 2.51653e-23, 1.12698e-23, 2.65367e-23, 1.40367e-22, 3.49448e-23], [2519.34, 1.04679e-22, 2.52179e-23, 1.12933e-23, 2.66008e-23, 1.40698e-22, 3.50686e-23], [2525.07, 1.04917e-22, 2.52707e-23, 1.13169e-23, 2.6665e-23, 1.4103e-22, 3.51927e-23], [2530.81, 1.05155e-22, 2.53237e-23, 1.13405e-23, 2.67294e-23, 1.41363e-22, 3.5317e-23], [2536.57, 1.05394e-22, 2.53767e-23, 1.13641e-23, 2.67939e-23, 1.41697e-22, 3.54415e-23], [2542.34, 1.05633e-22, 2.54299e-23, 1.13879e-23, 2.68586e-23, 1.42031e-22, 3.55663e-23], [2548.12, 1.05873e-22, 2.54833e-23, 1.14117e-23, 2.69234e-23, 1.42366e-22, 3.56913e-23], [2553.92, 1.06113e-22, 2.55367e-23, 1.14355e-23, 2.69884e-23, 1.42702e-22, 3.58165e-23], [2559.73, 1.06354e-22, 2.55903e-23, 1.14594e-23, 2.70534e-23, 1.43039e-22, 3.59419e-23], [2565.55, 1.06596e-22, 2.5644e-23, 1.14834e-23, 2.71187e-23, 1.43376e-22, 3.60676e-23], [2571.39, 1.06838e-22, 2.56979e-23, 1.15074e-23, 2.7184e-23, 1.43714e-22, 3.61935e-23], [2577.23, 1.07081e-22, 2.57519e-23, 1.15315e-23, 2.72495e-23, 1.44053e-22, 3.63196e-23], [2583.1, 1.07324e-22, 2.5806e-23, 1.15557e-23, 2.73152e-23, 1.44392e-22, 3.64459e-23], [2588.97, 1.07568e-22, 2.58602e-23, 1.15799e-23, 2.7381e-23, 1.44732e-22, 3.65725e-23], [2594.86, 1.07812e-22, 2.59146e-23, 1.16041e-23, 2.74469e-23, 1.45073e-22, 3.66993e-23], [2600.76, 1.08057e-22, 2.59691e-23, 1.16284e-23, 2.7513e-23, 1.45415e-22, 3.68263e-23], [2606.68, 1.08302e-22, 2.60238e-23, 1.16528e-23, 2.75792e-23, 1.45758e-22, 3.69536e-23], [2612.61, 1.08548e-22, 2.60785e-23, 1.16773e-23, 2.76456e-23, 1.46101e-22, 3.70811e-23], [2618.55, 1.08795e-22, 2.61334e-23, 1.17018e-23, 2.77121e-23, 1.46445e-22, 3.72088e-23], [2624.51, 1.09042e-22, 2.61885e-23, 1.17263e-23, 2.77788e-23, 1.4679e-22, 3.73368e-23], [2630.48, 1.0929e-22, 2.62437e-23, 1.1751e-23, 2.78456e-23, 1.47135e-22, 3.74649e-23], [2636.46, 1.09538e-22, 2.6299e-23, 1.17757e-23, 2.79125e-23, 1.47481e-22, 3.75934e-23], [2642.46, 1.09787e-22, 2.63545e-23, 1.18004e-23, 2.79796e-23, 1.47828e-22, 3.7722e-23], [2648.47, 1.10036e-22, 2.641e-23, 1.18252e-23, 2.80468e-23, 1.48176e-22, 3.78509e-23], [2654.49, 1.10286e-22, 2.64658e-23, 1.18501e-23, 2.81142e-23, 1.48525e-22, 3.798e-23], [2660.53, 1.10536e-22, 2.65216e-23, 1.1875e-23, 2.81817e-23, 1.48874e-22, 3.81093e-23], [2666.58, 1.10787e-22, 2.65776e-23, 1.19e-23, 2.82494e-23, 1.49224e-22, 3.82389e-23], [2672.64, 1.11039e-22, 2.66338e-23, 1.1925e-23, 2.83172e-23, 1.49575e-22, 3.83687e-23], [2678.72, 1.11291e-22, 2.669e-23, 1.19501e-23, 2.83852e-23, 1.49927e-22, 3.84988e-23], [2684.82, 1.11544e-22, 2.67464e-23, 1.19753e-23, 2.84533e-23, 1.50279e-22, 3.86291e-23], [2690.92, 1.11797e-22, 2.6803e-23, 1.20006e-23, 2.85216e-23, 1.50633e-22, 3.87596e-23], [2697.04, 1.12051e-22, 2.68597e-23, 1.20259e-23, 2.859e-23, 1.50987e-22, 3.88904e-23], [2703.18, 1.12306e-22, 2.69165e-23, 1.20512e-23, 2.86586e-23, 1.51341e-22, 3.90214e-23], [2709.33, 1.12561e-22, 2.69734e-23, 1.20766e-23, 2.87273e-23, 1.51697e-22, 3.91526e-23], [2715.49, 1.12817e-22, 2.70305e-23, 1.21021e-23, 2.87961e-23, 1.52053e-22, 3.92841e-23], [2721.67, 1.13073e-22, 2.70878e-23, 1.21277e-23, 2.88651e-23, 1.5241e-22, 3.94158e-23], [2727.86, 1.1333e-22, 2.71451e-23, 1.21533e-23, 2.89343e-23, 1.52768e-22, 3.95477e-23], [2734.06, 1.13587e-22, 2.72027e-23, 1.21789e-23, 2.90036e-23, 1.53127e-22, 3.96799e-23], [2740.28, 1.13845e-22, 2.72603e-23, 1.22047e-23, 2.90731e-23, 1.53487e-22, 3.98123e-23], [2746.51, 1.14104e-22, 2.73181e-23, 1.22305e-23, 2.91427e-23, 1.53847e-22, 3.9945e-23], [2752.76, 1.14363e-22, 2.7376e-23, 1.22563e-23, 2.92124e-23, 1.54208e-22, 4.00779e-23], [2759.02, 1.14623e-22, 2.74341e-23, 1.22823e-23, 2.92823e-23, 1.5457e-22, 4.02111e-23], [2765.3, 1.14883e-22, 2.74923e-23, 1.23082e-23, 2.93524e-23, 1.54932e-22, 4.03445e-23], [2771.59, 1.15144e-22, 2.75507e-23, 1.23343e-23, 2.94226e-23, 1.55296e-22, 4.04781e-23], [2777.89, 1.15406e-22, 2.76092e-23, 1.23604e-23, 2.9493e-23, 1.5566e-22, 4.0612e-23], [2784.21, 1.15668e-22, 2.76678e-23, 1.23866e-23, 2.95635e-23, 1.56025e-22, 4.07461e-23], [2790.54, 1.1593e-22, 2.77266e-23, 1.24128e-23, 2.96342e-23, 1.56391e-22, 4.08805e-23], [2796.89, 1.16194e-22, 2.77855e-23, 1.24391e-23, 2.9705e-23, 1.56758e-22, 4.10151e-23], [2803.25, 1.16458e-22, 2.78446e-23, 1.24655e-23, 2.9776e-23, 1.57126e-22, 4.115e-23], [2809.63, 1.16722e-22, 2.79038e-23, 1.24919e-23, 2.98472e-23, 1.57494e-22, 4.12851e-23], [2816.02, 1.16987e-22, 2.79632e-23, 1.25184e-23, 2.99185e-23, 1.57863e-22, 4.14204e-23], [2822.42, 1.17253e-22, 2.80227e-23, 1.2545e-23, 2.99899e-23, 1.58233e-22, 4.1556e-23], [2828.84, 1.1752e-22, 2.80823e-23, 1.25716e-23, 3.00615e-23, 1.58604e-22, 4.16919e-23], [2835.28, 1.17787e-22, 2.81421e-23, 1.25983e-23, 3.01333e-23, 1.58975e-22, 4.1828e-23], [2841.73, 1.18054e-22, 2.8202e-23, 1.2625e-23, 3.02052e-23, 1.59348e-22, 4.19643e-23], [2848.19, 1.18322e-22, 2.82621e-23, 1.26519e-23, 3.02773e-23, 1.59721e-22, 4.21009e-23], [2854.67, 1.18591e-22, 2.83224e-23, 1.26787e-23, 3.03495e-23, 1.60095e-22, 4.22377e-23], [2861.16, 1.1886e-22, 2.83827e-23, 1.27057e-23, 3.04219e-23, 1.6047e-22, 4.23748e-23], [2867.67, 1.1913e-22, 2.84432e-23, 1.27327e-23, 3.04945e-23, 1.60846e-22, 4.25122e-23], [2874.19, 1.19401e-22, 2.85039e-23, 1.27598e-23, 3.05672e-23, 1.61222e-22, 4.26498e-23], [2880.73, 1.19672e-22, 2.85647e-23, 1.2787e-23, 3.06401e-23, 1.616e-22, 4.27876e-23], [2887.28, 1.19944e-22, 2.86257e-23, 1.28142e-23, 3.07131e-23, 1.61978e-22, 4.29257e-23], [2893.85, 1.20217e-22, 2.86868e-23, 1.28414e-23, 3.07863e-23, 1.62357e-22, 4.3064e-23], [2900.43, 1.2049e-22, 2.8748e-23, 1.28688e-23, 3.08596e-23, 1.62737e-22, 4.32026e-23], [2907.03, 1.20763e-22, 2.88094e-23, 1.28962e-23, 3.09331e-23, 1.63118e-22, 4.33415e-23], [2913.64, 1.21038e-22, 2.8871e-23, 1.29237e-23, 3.10068e-23, 1.63499e-22, 4.34806e-23], [2920.27, 1.21313e-22, 2.89327e-23, 1.29512e-23, 3.10806e-23, 1.63882e-22, 4.362e-23], [2926.91, 1.21588e-22, 2.89945e-23, 1.29788e-23, 3.11546e-23, 1.64265e-22, 4.37596e-23], [2933.57, 1.21865e-22, 2.90565e-23, 1.30065e-23, 3.12287e-23, 1.64649e-22, 4.38995e-23], [2940.24, 1.22141e-22, 2.91186e-23, 1.30343e-23, 3.1303e-23, 1.65034e-22, 4.40396e-23], [2946.93, 1.22419e-22, 2.91809e-23, 1.30621e-23, 3.13775e-23, 1.6542e-22, 4.418e-23], [2953.63, 1.22697e-22, 2.92434e-23, 1.30899e-23, 3.14522e-23, 1.65807e-22, 4.43206e-23], [2960.35, 1.22976e-22, 2.9306e-23, 1.31179e-23, 3.1527e-23, 1.66194e-22, 4.44615e-23], [2967.08, 1.23255e-22, 2.93687e-23, 1.31459e-23, 3.16019e-23, 1.66583e-22, 4.46027e-23], [2973.83, 1.23535e-22, 2.94316e-23, 1.3174e-23, 3.1677e-23, 1.66972e-22, 4.47441e-23], [2980.6, 1.23816e-22, 2.94946e-23, 1.32021e-23, 3.17523e-23, 1.67362e-22, 4.48858e-23], [2987.38, 1.24097e-22, 2.95578e-23, 1.32304e-23, 3.18278e-23, 1.67753e-22, 4.50278e-23], [2994.17, 1.24379e-22, 2.96212e-23, 1.32586e-23, 3.19034e-23, 1.68145e-22, 4.517e-23], [3000.98, 1.24661e-22, 2.96847e-23, 1.3287e-23, 3.19792e-23, 1.68538e-22, 4.53124e-23], [3007.81, 1.24945e-22, 2.97484e-23, 1.33154e-23, 3.20552e-23, 1.68931e-22, 4.54552e-23], [3014.65, 1.25229e-22, 2.98122e-23, 1.33439e-23, 3.21313e-23, 1.69326e-22, 4.55981e-23], [3021.51, 1.25513e-22, 2.98761e-23, 1.33725e-23, 3.22076e-23, 1.69721e-22, 4.57414e-23], [3028.38, 1.25798e-22, 2.99402e-23, 1.34011e-23, 3.2284e-23, 1.70117e-22, 4.58849e-23], [3035.27, 1.26084e-22, 3.00045e-23, 1.34298e-23, 3.23606e-23, 1.70514e-22, 4.60287e-23], [3042.17, 1.2637e-22, 3.00689e-23, 1.34586e-23, 3.24374e-23, 1.70912e-22, 4.61727e-23], [3049.09, 1.26657e-22, 3.01335e-23, 1.34874e-23, 3.25144e-23, 1.71311e-22, 4.63171e-23], [3056.03, 1.26945e-22, 3.01983e-23, 1.35163e-23, 3.25915e-23, 1.71711e-22, 4.64616e-23], [3062.98, 1.27234e-22, 3.02631e-23, 1.35453e-23, 3.26688e-23, 1.72112e-22, 4.66065e-23], [3069.95, 1.27523e-22, 3.03282e-23, 1.35743e-23, 3.27463e-23, 1.72513e-22, 4.67516e-23], [3076.93, 1.27812e-22, 3.03934e-23, 1.36035e-23, 3.28239e-23, 1.72916e-22, 4.6897e-23], [3083.93, 1.28103e-22, 3.04587e-23, 1.36327e-23, 3.29017e-23, 1.73319e-22, 4.70426e-23], [3090.94, 1.28394e-22, 3.05243e-23, 1.36619e-23, 3.29797e-23, 1.73723e-22, 4.71886e-23], [3097.97, 1.28685e-22, 3.05899e-23, 1.36912e-23, 3.30578e-23, 1.74128e-22, 4.73348e-23], [3105.02, 1.28978e-22, 3.06558e-23, 1.37206e-23, 3.31361e-23, 1.74534e-22, 4.74812e-23], [3112.08, 1.29271e-22, 3.07218e-23, 1.37501e-23, 3.32146e-23, 1.74941e-22, 4.76279e-23], [3119.16, 1.29565e-22, 3.07879e-23, 1.37796e-23, 3.32933e-23, 1.75349e-22, 4.7775e-23], [3126.25, 1.29859e-22, 3.08542e-23, 1.38093e-23, 3.33721e-23, 1.75758e-22, 4.79222e-23], [3133.37, 1.30154e-22, 3.09207e-23, 1.38389e-23, 3.34511e-23, 1.76168e-22, 4.80698e-23], [3140.49, 1.3045e-22, 3.09873e-23, 1.38687e-23, 3.35303e-23, 1.76578e-22, 4.82176e-23], [3147.64, 1.30746e-22, 3.10541e-23, 1.38985e-23, 3.36097e-23, 1.7699e-22, 4.83657e-23], [3154.8, 1.31043e-22, 3.1121e-23, 1.39284e-23, 3.36892e-23, 1.77402e-22, 4.85141e-23], [3161.97, 1.31341e-22, 3.11881e-23, 1.39584e-23, 3.37689e-23, 1.77815e-22, 4.86627e-23], [3169.16, 1.31639e-22, 3.12554e-23, 1.39884e-23, 3.38488e-23, 1.78229e-22, 4.88116e-23], [3176.37, 1.31938e-22, 3.13228e-23, 1.40185e-23, 3.39288e-23, 1.78645e-22, 4.89609e-23], [3183.6, 1.32238e-22, 3.13904e-23, 1.40487e-23, 3.40091e-23, 1.79061e-22, 4.91103e-23], [3190.84, 1.32539e-22, 3.14581e-23, 1.4079e-23, 3.40895e-23, 1.79478e-22, 4.92601e-23], [3198.1, 1.3284e-22, 3.1526e-23, 1.41093e-23, 3.417e-23, 1.79896e-22, 4.94101e-23], [3205.37, 1.33141e-22, 3.15941e-23, 1.41397e-23, 3.42508e-23, 1.80314e-22, 4.95604e-23], [3212.66, 1.33444e-22, 3.16623e-23, 1.41702e-23, 3.43317e-23, 1.80734e-22, 4.9711e-23], [3219.97, 1.33747e-22, 3.17307e-23, 1.42007e-23, 3.44128e-23, 1.81155e-22, 4.98619e-23], [3227.29, 1.34051e-22, 3.17993e-23, 1.42314e-23, 3.44941e-23, 1.81577e-22, 5.00131e-23], [3234.63, 1.34356e-22, 3.1868e-23, 1.42621e-23, 3.45756e-23, 1.81999e-22, 5.01645e-23], [3241.99, 1.34661e-22, 3.19369e-23, 1.42928e-23, 3.46573e-23, 1.82423e-22, 5.03162e-23], [3249.36, 1.34967e-22, 3.2006e-23, 1.43237e-23, 3.47391e-23, 1.82847e-22, 5.04682e-23], [3256.76, 1.35273e-22, 3.20752e-23, 1.43546e-23, 3.48211e-23, 1.83272e-22, 5.06205e-23], [3264.16, 1.35581e-22, 3.21446e-23, 1.43856e-23, 3.49033e-23, 1.83699e-22, 5.07731e-23], [3271.59, 1.35889e-22, 3.22141e-23, 1.44166e-23, 3.49857e-23, 1.84126e-22, 5.0926e-23], [3279.03, 1.36198e-22, 3.22838e-23, 1.44478e-23, 3.50682e-23, 1.84554e-22, 5.10791e-23], [3286.49, 1.36507e-22, 3.23537e-23, 1.4479e-23, 3.51509e-23, 1.84984e-22, 5.12325e-23], [3293.96, 1.36817e-22, 3.24238e-23, 1.45103e-23, 3.52339e-23, 1.85414e-22, 5.13862e-23], [3301.46, 1.37128e-22, 3.2494e-23, 1.45416e-23, 3.5317e-23, 1.85845e-22, 5.15402e-23], [3308.96, 1.3744e-22, 3.25644e-23, 1.45731e-23, 3.54002e-23, 1.86277e-22, 5.16945e-23], [3316.49, 1.37752e-22, 3.26349e-23, 1.46046e-23, 3.54837e-23, 1.8671e-22, 5.18491e-23], [3324.03, 1.38065e-22, 3.27056e-23, 1.46362e-23, 3.55673e-23, 1.87144e-22, 5.2004e-23], [3331.6, 1.38379e-22, 3.27765e-23, 1.46678e-23, 3.56512e-23, 1.87579e-22, 5.21592e-23], [3339.17, 1.38693e-22, 3.28476e-23, 1.46996e-23, 3.57352e-23, 1.88015e-22, 5.23146e-23], [3346.77, 1.39008e-22, 3.29188e-23, 1.47314e-23, 3.58194e-23, 1.88452e-22, 5.24704e-23], [3354.38, 1.39324e-22, 3.29902e-23, 1.47633e-23, 3.59038e-23, 1.8889e-22, 5.26264e-23], [3362.01, 1.39641e-22, 3.30618e-23, 1.47954e-23, 3.59884e-23, 1.89329e-22, 5.27828e-23], [3369.66, 1.39958e-22, 3.31335e-23, 1.48273e-23, 3.60731e-23, 1.89768e-22, 5.29394e-23], [3377.32, 1.40276e-22, 3.32054e-23, 1.48594e-23, 3.61581e-23, 1.90209e-22, 5.30963e-23], [3385.0, 1.40595e-22, 3.32775e-23, 1.48916e-23, 3.62432e-23, 1.90651e-22, 5.32535e-23], [3392.7, 1.40914e-22, 3.33497e-23, 1.49239e-23, 3.63285e-23, 1.91094e-22, 5.3411e-23], [3400.42, 1.41234e-22, 3.34221e-23, 1.49562e-23, 3.6414e-23, 1.91538e-22, 5.35688e-23], [3408.16, 1.41555e-22, 3.34947e-23, 1.49887e-23, 3.64997e-23, 1.91982e-22, 5.37269e-23], [3415.91, 1.41877e-22, 3.35675e-23, 1.50212e-23, 3.65856e-23, 1.92428e-22, 5.38853e-23], [3423.68, 1.42199e-22, 3.36404e-23, 1.50538e-23, 3.66717e-23, 1.92875e-22, 5.4044e-23], [3431.46, 1.42522e-22, 3.37135e-23, 1.50864e-23, 3.6758e-23, 1.93323e-22, 5.4203e-23], [3439.27, 1.42846e-22, 3.37868e-23, 1.51192e-23, 3.68444e-23, 1.93771e-22, 5.43623e-23], [3447.09, 1.43171e-22, 3.38603e-23, 1.5152e-23, 3.69311e-23, 1.94221e-22, 5.45219e-23], [3454.93, 1.43496e-22, 3.39339e-23, 1.51849e-23, 3.70179e-23, 1.94672e-22, 5.46818e-23], [3462.79, 1.43822e-22, 3.40077e-23, 1.52178e-23, 3.71049e-23, 1.95124e-22, 5.48421e-23], [3470.67, 1.44149e-22, 3.40817e-23, 1.52509e-23, 3.71921e-23, 1.95577e-22, 5.50026e-23], [3478.56, 1.44476e-22, 3.41559e-23, 1.5284e-23, 3.72796e-23, 1.9603e-22, 5.51634e-23], [3486.47, 1.44805e-22, 3.42302e-23, 1.53172e-23, 3.73672e-23, 1.96485e-22, 5.53245e-23], [3494.4, 1.45134e-22, 3.43047e-23, 1.53505e-23, 3.7455e-23, 1.96941e-22, 5.54859e-23], [3502.35, 1.45463e-22, 3.43794e-23, 1.53839e-23, 3.7543e-23, 1.97398e-22, 5.56477e-23], [3510.32, 1.45794e-22, 3.44543e-23, 1.54173e-23, 3.76312e-23, 1.97856e-22, 5.58097e-23], [3518.3, 1.46125e-22, 3.45293e-23, 1.54509e-23, 3.77195e-23, 1.98314e-22, 5.5972e-23], [3526.31, 1.46457e-22, 3.46045e-23, 1.54845e-23, 3.78081e-23, 1.98774e-22, 5.61347e-23], [3534.33, 1.4679e-22, 3.46799e-23, 1.55182e-23, 3.78969e-23, 1.99235e-22, 5.62976e-23], [3542.37, 1.47124e-22, 3.47555e-23, 1.55519e-23, 3.79859e-23, 1.99697e-22, 5.64609e-23], [3550.42, 1.47458e-22, 3.48313e-23, 1.55858e-23, 3.8075e-23, 2.0016e-22, 5.66245e-23], [3558.5, 1.47793e-22, 3.49072e-23, 1.56197e-23, 3.81644e-23, 2.00624e-22, 5.67884e-23], [3566.59, 1.48129e-22, 3.49833e-23, 1.56537e-23, 3.8254e-23, 2.01089e-22, 5.69526e-23], [3574.71, 1.48465e-22, 3.50596e-23, 1.56878e-23, 3.83437e-23, 2.01555e-22, 5.71171e-23], [3582.84, 1.48803e-22, 3.51361e-23, 1.5722e-23, 3.84337e-23, 2.02022e-22, 5.72819e-23], [3590.99, 1.49141e-22, 3.52128e-23, 1.57562e-23, 3.85238e-23, 2.02491e-22, 5.74471e-23], [3599.15, 1.4948e-22, 3.52896e-23, 1.57905e-23, 3.86142e-23, 2.0296e-22, 5.76125e-23], [3607.34, 1.49819e-22, 3.53666e-23, 1.58249e-23, 3.87047e-23, 2.0343e-22, 5.77783e-23], [3615.55, 1.5016e-22, 3.54438e-23, 1.58594e-23, 3.87955e-23, 2.03901e-22, 5.79444e-23], [3623.77, 1.50501e-22, 3.55212e-23, 1.5894e-23, 3.88865e-23, 2.04374e-22, 5.81108e-23], [3632.01, 1.50843e-22, 3.55988e-23, 1.59287e-23, 3.89776e-23, 2.04847e-22, 5.82775e-23], [3640.27, 1.51186e-22, 3.56765e-23, 1.59634e-23, 3.9069e-23, 2.05322e-22, 5.84446e-23], [3648.55, 1.51529e-22, 3.57545e-23, 1.59982e-23, 3.91605e-23, 2.05797e-22, 5.86119e-23], [3656.85, 1.51874e-22, 3.58326e-23, 1.60331e-23, 3.92523e-23, 2.06274e-22, 5.87796e-23], [3665.17, 1.52219e-22, 3.59109e-23, 1.60681e-23, 3.93443e-23, 2.06752e-22, 5.89476e-23], [3673.51, 1.52565e-22, 3.59894e-23, 1.61032e-23, 3.94364e-23, 2.0723e-22, 5.91159e-23], [3681.86, 1.52911e-22, 3.60681e-23, 1.61384e-23, 3.95288e-23, 2.0771e-22, 5.92846e-23], [3690.24, 1.53259e-22, 3.61469e-23, 1.61736e-23, 3.96214e-23, 2.08191e-22, 5.94535e-23], [3698.63, 1.53607e-22, 3.6226e-23, 1.62089e-23, 3.97142e-23, 2.08673e-22, 5.96228e-23], [3707.04, 1.53956e-22, 3.63052e-23, 1.62443e-23, 3.98072e-23, 2.09156e-22, 5.97925e-23], [3715.48, 1.54306e-22, 3.63846e-23, 1.62798e-23, 3.99004e-23, 2.0964e-22, 5.99624e-23], [3723.93, 1.54657e-22, 3.64643e-23, 1.63154e-23, 3.99938e-23, 2.10125e-22, 6.01327e-23], [3732.4, 1.55008e-22, 3.65441e-23, 1.6351e-23, 4.00874e-23, 2.10612e-22, 6.03033e-23], [3740.89, 1.5536e-22, 3.6624e-23, 1.63868e-23, 4.01812e-23, 2.11099e-22, 6.04742e-23], [3749.4, 1.55714e-22, 3.67042e-23, 1.64226e-23, 4.02752e-23, 2.11587e-22, 6.06454e-23], [3757.92, 1.56067e-22, 3.67846e-23, 1.64585e-23, 4.03695e-23, 2.12077e-22, 6.0817e-23], [3766.47, 1.56422e-22, 3.68651e-23, 1.64945e-23, 4.04639e-23, 2.12568e-22, 6.09889e-23], [3775.04, 1.56777e-22, 3.69459e-23, 1.65306e-23, 4.05586e-23, 2.13059e-22, 6.11612e-23], [3783.62, 1.57134e-22, 3.70268e-23, 1.65667e-23, 4.06534e-23, 2.13552e-22, 6.13337e-23], [3792.23, 1.57491e-22, 3.71079e-23, 1.6603e-23, 4.07485e-23, 2.14046e-22, 6.15066e-23], [3800.86, 1.57849e-22, 3.71893e-23, 1.66393e-23, 4.08438e-23, 2.14541e-22, 6.16799e-23], [3809.5, 1.58207e-22, 3.72708e-23, 1.66758e-23, 4.09393e-23, 2.15037e-22, 6.18535e-23], [3818.17, 1.58567e-22, 3.73525e-23, 1.67123e-23, 4.1035e-23, 2.15535e-22, 6.20274e-23], [3826.85, 1.58927e-22, 3.74344e-23, 1.67489e-23, 4.11309e-23, 2.16033e-22, 6.22016e-23], [3835.56, 1.59288e-22, 3.75165e-23, 1.67855e-23, 4.1227e-23, 2.16533e-22, 6.23762e-23], [3844.28, 1.5965e-22, 3.75988e-23, 1.68223e-23, 4.13234e-23, 2.17033e-22, 6.25511e-23], [3853.02, 1.60013e-22, 3.76812e-23, 1.68592e-23, 4.14199e-23, 2.17535e-22, 6.27264e-23], [3861.79, 1.60377e-22, 3.77639e-23, 1.68961e-23, 4.15167e-23, 2.18038e-22, 6.2902e-23], [3870.57, 1.60741e-22, 3.78468e-23, 1.69331e-23, 4.16137e-23, 2.18542e-22, 6.30779e-23], [3879.38, 1.61107e-22, 3.79298e-23, 1.69702e-23, 4.17109e-23, 2.19047e-22, 6.32542e-23], [3888.2, 1.61473e-22, 3.80131e-23, 1.70074e-23, 4.18083e-23, 2.19553e-22, 6.34308e-23], [3897.04, 1.6184e-22, 3.80965e-23, 1.70447e-23, 4.19059e-23, 2.20061e-22, 6.36077e-23], [3905.91, 1.62207e-22, 3.81802e-23, 1.70821e-23, 4.20038e-23, 2.20569e-22, 6.3785e-23], [3914.79, 1.62576e-22, 3.8264e-23, 1.71196e-23, 4.21018e-23, 2.21079e-22, 6.39627e-23], [3923.7, 1.62946e-22, 3.83481e-23, 1.71571e-23, 4.22001e-23, 2.2159e-22, 6.41407e-23], [3932.62, 1.63316e-22, 3.84323e-23, 1.71948e-23, 4.22986e-23, 2.22101e-22, 6.4319e-23], [3941.57, 1.63687e-22, 3.85167e-23, 1.72325e-23, 4.23973e-23, 2.22615e-22, 6.44977e-23], [3950.53, 1.64059e-22, 3.86014e-23, 1.72703e-23, 4.24963e-23, 2.23129e-22, 6.46767e-23], [3959.52, 1.64432e-22, 3.86862e-23, 1.73083e-23, 4.25954e-23, 2.23644e-22, 6.48561e-23], [3968.52, 1.64805e-22, 3.87712e-23, 1.73463e-23, 4.26948e-23, 2.24161e-22, 6.50358e-23], [3977.55, 1.6518e-22, 3.88565e-23, 1.73843e-23, 4.27944e-23, 2.24678e-22, 6.52159e-23], [3986.6, 1.65555e-22, 3.89419e-23, 1.74225e-23, 4.28943e-23, 2.25197e-22, 6.53963e-23], [3995.67, 1.65932e-22, 3.90275e-23, 1.74608e-23, 4.29943e-23, 2.25717e-22, 6.55771e-23], [4004.75, 1.66309e-22, 3.91134e-23, 1.74992e-23, 4.30946e-23, 2.26239e-22, 6.57582e-23], [4013.86, 1.66687e-22, 3.91994e-23, 1.75376e-23, 4.31951e-23, 2.26761e-22, 6.59397e-23], [4022.99, 1.67066e-22, 3.92857e-23, 1.75761e-23, 4.32958e-23, 2.27284e-22, 6.61216e-23], [4032.14, 1.67445e-22, 3.93721e-23, 1.76148e-23, 4.33967e-23, 2.27809e-22, 6.63037e-23], [4041.32, 1.67826e-22, 3.94587e-23, 1.76535e-23, 4.34979e-23, 2.28335e-22, 6.64863e-23], [4050.51, 1.68207e-22, 3.95456e-23, 1.76923e-23, 4.35993e-23, 2.28862e-22, 6.66692e-23], [4059.72, 1.68589e-22, 3.96327e-23, 1.77312e-23, 4.37009e-23, 2.2939e-22, 6.68525e-23], [4068.95, 1.68973e-22, 3.97199e-23, 1.77702e-23, 4.38027e-23, 2.2992e-22, 6.70361e-23], [4078.21, 1.69357e-22, 3.98074e-23, 1.78093e-23, 4.39048e-23, 2.3045e-22, 6.72201e-23], [4087.49, 1.69741e-22, 3.9895e-23, 1.78485e-23, 4.40071e-23, 2.30982e-22, 6.74044e-23], [4096.78, 1.70127e-22, 3.99829e-23, 1.78877e-23, 4.41096e-23, 2.31515e-22, 6.75891e-23], [4106.1, 1.70514e-22, 4.0071e-23, 1.79271e-23, 4.42123e-23, 2.32049e-22, 6.77742e-23], [4115.44, 1.70901e-22, 4.01593e-23, 1.79666e-23, 4.43153e-23, 2.32585e-22, 6.79596e-23], [4124.8, 1.7129e-22, 4.02478e-23, 1.80061e-23, 4.44185e-23, 2.33121e-22, 6.81454e-23], [4134.18, 1.71679e-22, 4.03365e-23, 1.80458e-23, 4.45219e-23, 2.33659e-22, 6.83315e-23], [4143.59, 1.72069e-22, 4.04254e-23, 1.80855e-23, 4.46256e-23, 2.34198e-22, 6.8518e-23], [4153.01, 1.7246e-22, 4.05145e-23, 1.81253e-23, 4.47295e-23, 2.34738e-22, 6.87049e-23], [4162.46, 1.72852e-22, 4.06038e-23, 1.81652e-23, 4.48336e-23, 2.3528e-22, 6.88922e-23], [4171.93, 1.73245e-22, 4.06934e-23, 1.82053e-23, 4.4938e-23, 2.35823e-22, 6.90798e-23], [4181.42, 1.73639e-22, 4.07831e-23, 1.82454e-23, 4.50426e-23, 2.36366e-22, 6.92678e-23], [4190.93, 1.74033e-22, 4.08731e-23, 1.82856e-23, 4.51474e-23, 2.36912e-22, 6.94561e-23], [4200.46, 1.74429e-22, 4.09633e-23, 1.83259e-23, 4.52524e-23, 2.37458e-22, 6.96449e-23], [4210.01, 1.74825e-22, 4.10536e-23, 1.83663e-23, 4.53577e-23, 2.38005e-22, 6.9834e-23], [4219.59, 1.75223e-22, 4.11442e-23, 1.84068e-23, 4.54633e-23, 2.38554e-22, 7.00234e-23], [4229.19, 1.75621e-22, 4.1235e-23, 1.84473e-23, 4.5569e-23, 2.39104e-22, 7.02133e-23], [4238.81, 1.7602e-22, 4.1326e-23, 1.8488e-23, 4.5675e-23, 2.39655e-22, 7.04035e-23], [4248.45, 1.7642e-22, 4.14173e-23, 1.85288e-23, 4.57812e-23, 2.40208e-22, 7.05941e-23], [4258.11, 1.76821e-22, 4.15087e-23, 1.85697e-23, 4.58877e-23, 2.40762e-22, 7.07851e-23], [4267.8, 1.77223e-22, 4.16004e-23, 1.86106e-23, 4.59944e-23, 2.41316e-22, 7.09764e-23], [4277.5, 1.77626e-22, 4.16922e-23, 1.86517e-23, 4.61014e-23, 2.41873e-22, 7.11681e-23], [4287.23, 1.78029e-22, 4.17843e-23, 1.86928e-23, 4.62085e-23, 2.4243e-22, 7.13602e-23], [4296.99, 1.78434e-22, 4.18766e-23, 1.87341e-23, 4.6316e-23, 2.42989e-22, 7.15527e-23], [4306.76, 1.7884e-22, 4.19692e-23, 1.87755e-23, 4.64236e-23, 2.43549e-22, 7.17456e-23], [4316.56, 1.79246e-22, 4.20619e-23, 1.88169e-23, 4.65315e-23, 2.4411e-22, 7.19388e-23], [4326.37, 1.79653e-22, 4.21548e-23, 1.88584e-23, 4.66396e-23, 2.44672e-22, 7.21325e-23], [4336.21, 1.80062e-22, 4.2248e-23, 1.89001e-23, 4.6748e-23, 2.45236e-22, 7.23265e-23], [4346.08, 1.80471e-22, 4.23414e-23, 1.89418e-23, 4.68566e-23, 2.45801e-22, 7.25209e-23], [4355.96, 1.80881e-22, 4.2435e-23, 1.89837e-23, 4.69655e-23, 2.46367e-22, 7.27157e-23], [4365.87, 1.81292e-22, 4.25288e-23, 1.90256e-23, 4.70746e-23, 2.46935e-22, 7.29108e-23], [4375.8, 1.81704e-22, 4.26229e-23, 1.90676e-23, 4.7184e-23, 2.47504e-22, 7.31064e-23], [4385.75, 1.82117e-22, 4.27171e-23, 1.91098e-23, 4.72936e-23, 2.48074e-22, 7.33023e-23], [4395.73, 1.82531e-22, 4.28116e-23, 1.9152e-23, 4.74034e-23, 2.48645e-22, 7.34987e-23], [4405.73, 1.82946e-22, 4.29063e-23, 1.91943e-23, 4.75135e-23, 2.49218e-22, 7.36954e-23], [4415.75, 1.83362e-22, 4.30013e-23, 1.92368e-23, 4.76238e-23, 2.49792e-22, 7.38925e-23], [4425.79, 1.83779e-22, 4.30964e-23, 1.92793e-23, 4.77344e-23, 2.50367e-22, 7.409e-23], [4435.86, 1.84196e-22, 4.31918e-23, 1.93219e-23, 4.78452e-23, 2.50944e-22, 7.42879e-23], [4445.95, 1.84615e-22, 4.32874e-23, 1.93646e-23, 4.79563e-23, 2.51521e-22, 7.44862e-23], [4456.06, 1.85034e-22, 4.33832e-23, 1.94075e-23, 4.80676e-23, 2.52101e-22, 7.46849e-23], [4466.2, 1.85455e-22, 4.34792e-23, 1.94504e-23, 4.81792e-23, 2.52681e-22, 7.4884e-23], [4476.36, 1.85877e-22, 4.35755e-23, 1.94934e-23, 4.8291e-23, 2.53263e-22, 7.50834e-23], [4486.54, 1.86299e-22, 4.3672e-23, 1.95365e-23, 4.8403e-23, 2.53846e-22, 7.52833e-23], [4496.74, 1.86722e-22, 4.37687e-23, 1.95798e-23, 4.85153e-23, 2.5443e-22, 7.54836e-23], [4506.97, 1.87147e-22, 4.38656e-23, 1.96231e-23, 4.86279e-23, 2.55016e-22, 7.56843e-23], [4517.22, 1.87572e-22, 4.39628e-23, 1.96665e-23, 4.87407e-23, 2.55603e-22, 7.58853e-23], [4527.5, 1.87998e-22, 4.40602e-23, 1.97101e-23, 4.88538e-23, 2.56191e-22, 7.60868e-23], [4537.8, 1.88426e-22, 4.41578e-23, 1.97537e-23, 4.89671e-23, 2.56781e-22, 7.62887e-23], [4548.12, 1.88854e-22, 4.42557e-23, 1.97974e-23, 4.90807e-23, 2.57372e-22, 7.6491e-23], [4558.46, 1.89283e-22, 4.43537e-23, 1.98413e-23, 4.91945e-23, 2.57964e-22, 7.66937e-23], [4568.83, 1.89713e-22, 4.4452e-23, 1.98852e-23, 4.93086e-23, 2.58557e-22, 7.68968e-23], [4579.22, 1.90145e-22, 4.45506e-23, 1.99293e-23, 4.94229e-23, 2.59152e-22, 7.71003e-23], [4589.64, 1.90577e-22, 4.46493e-23, 1.99734e-23, 4.95375e-23, 2.59749e-22, 7.73042e-23], [4600.08, 1.9101e-22, 4.47483e-23, 2.00177e-23, 4.96524e-23, 2.60346e-22, 7.75085e-23], [4610.54, 1.91444e-22, 4.48476e-23, 2.0062e-23, 4.97675e-23, 2.60945e-22, 7.77132e-23], [4621.03, 1.91879e-22, 4.4947e-23, 2.01065e-23, 4.98828e-23, 2.61546e-22, 7.79184e-23], [4631.54, 1.92315e-22, 4.50467e-23, 2.0151e-23, 4.99984e-23, 2.62147e-22, 7.81239e-23], [4642.07, 1.92753e-22, 4.51466e-23, 2.01957e-23, 5.01143e-23, 2.6275e-22, 7.83299e-23], [4652.63, 1.93191e-22, 4.52468e-23, 2.02405e-23, 5.02305e-23, 2.63355e-22, 7.85363e-23], [4663.22, 1.9363e-22, 4.53472e-23, 2.02853e-23, 5.03469e-23, 2.6396e-22, 7.8743e-23], [4673.82, 1.9407e-22, 4.54478e-23, 2.03303e-23, 5.04635e-23, 2.64567e-22, 7.89502e-23], [4684.45, 1.94511e-22, 4.55487e-23, 2.03754e-23, 5.05804e-23, 2.65176e-22, 7.91579e-23], [4695.11, 1.94953e-22, 4.56498e-23, 2.04206e-23, 5.06976e-23, 2.65786e-22, 7.93659e-23], [4705.79, 1.95396e-22, 4.57511e-23, 2.04659e-23, 5.0815e-23, 2.66397e-22, 7.95744e-23], [4716.49, 1.9584e-22, 4.58526e-23, 2.05113e-23, 5.09327e-23, 2.6701e-22, 7.97832e-23], [4727.22, 1.96285e-22, 4.59544e-23, 2.05568e-23, 5.10507e-23, 2.67624e-22, 7.99925e-23], [4737.97, 1.96732e-22, 4.60565e-23, 2.06024e-23, 5.11689e-23, 2.68239e-22, 8.02023e-23], [4748.75, 1.97179e-22, 4.61588e-23, 2.06481e-23, 5.12874e-23, 2.68856e-22, 8.04124e-23], [4759.55, 1.97627e-22, 4.62613e-23, 2.06939e-23, 5.14062e-23, 2.69474e-22, 8.0623e-23], [4770.38, 1.98076e-22, 4.6364e-23, 2.07399e-23, 5.15252e-23, 2.70093e-22, 8.0834e-23], [4781.23, 1.98526e-22, 4.6467e-23, 2.07859e-23, 5.16445e-23, 2.70714e-22, 8.10454e-23], [4792.1, 1.98978e-22, 4.65702e-23, 2.0832e-23, 5.1764e-23, 2.71336e-22, 8.12572e-23], [4803.0, 1.9943e-22, 4.66737e-23, 2.08783e-23, 5.18839e-23, 2.7196e-22, 8.14695e-23], [4813.93, 1.99883e-22, 4.67774e-23, 2.09247e-23, 5.2004e-23, 2.72585e-22, 8.16822e-23], [4824.88, 2.00337e-22, 4.68814e-23, 2.09711e-23, 5.21243e-23, 2.73212e-22, 8.18953e-23], [4835.85, 2.00793e-22, 4.69856e-23, 2.10177e-23, 5.2245e-23, 2.7384e-22, 8.21089e-23], [4846.85, 2.01249e-22, 4.709e-23, 2.10644e-23, 5.23659e-23, 2.74469e-22, 8.23229e-23], [4857.87, 2.01707e-22, 4.71947e-23, 2.11112e-23, 5.2487e-23, 2.751e-22, 8.25373e-23], [4868.92, 2.02165e-22, 4.72996e-23, 2.11581e-23, 5.26085e-23, 2.75732e-22, 8.27522e-23], [4880.0, 2.02625e-22, 4.74048e-23, 2.12051e-23, 5.27302e-23, 2.76365e-22, 8.29675e-23], [4891.1, 2.03085e-22, 4.75102e-23, 2.12522e-23, 5.28522e-23, 2.77e-22, 8.31832e-23], [4902.22, 2.03547e-22, 4.76159e-23, 2.12995e-23, 5.29744e-23, 2.77637e-22, 8.33994e-23], [4913.37, 2.04009e-22, 4.77218e-23, 2.13468e-23, 5.3097e-23, 2.78275e-22, 8.3616e-23], [4924.55, 2.04473e-22, 4.78279e-23, 2.13942e-23, 5.32198e-23, 2.78914e-22, 8.3833e-23], [4935.75, 2.04938e-22, 4.79343e-23, 2.14418e-23, 5.33428e-23, 2.79555e-22, 8.40505e-23], [4946.98, 2.05404e-22, 4.8041e-23, 2.14895e-23, 5.34662e-23, 2.80197e-22, 8.42684e-23], [4958.23, 2.05871e-22, 4.81478e-23, 2.15373e-23, 5.35898e-23, 2.80841e-22, 8.44868e-23], [4969.51, 2.06339e-22, 4.8255e-23, 2.15852e-23, 5.37137e-23, 2.81486e-22, 8.47056e-23], [4980.81, 2.06808e-22, 4.83624e-23, 2.16332e-23, 5.38379e-23, 2.82132e-22, 8.49249e-23], [4992.14, 2.07278e-22, 4.847e-23, 2.16813e-23, 5.39624e-23, 2.8278e-22, 8.51446e-23], [5003.5, 2.07749e-22, 4.85779e-23, 2.17295e-23, 5.40871e-23, 2.8343e-22, 8.53647e-23], [5014.88, 2.08221e-22, 4.8686e-23, 2.17779e-23, 5.42122e-23, 2.84081e-22, 8.55853e-23], [5026.28, 2.08694e-22, 4.87944e-23, 2.18263e-23, 5.43375e-23, 2.84733e-22, 8.58064e-23], [5037.72, 2.09169e-22, 4.89031e-23, 2.18749e-23, 5.4463e-23, 2.85387e-22, 8.60279e-23], [5049.17, 2.09644e-22, 4.9012e-23, 2.19236e-23, 5.45889e-23, 2.86042e-22, 8.62498e-23], [5060.66, 2.10121e-22, 4.91211e-23, 2.19724e-23, 5.47151e-23, 2.86699e-22, 8.64723e-23], [5072.17, 2.10598e-22, 4.92305e-23, 2.20213e-23, 5.48415e-23, 2.87358e-22, 8.66951e-23], [5083.71, 2.11077e-22, 4.93402e-23, 2.20703e-23, 5.49682e-23, 2.88017e-22, 8.69184e-23], [5095.27, 2.11557e-22, 4.94501e-23, 2.21194e-23, 5.50952e-23, 2.88679e-22, 8.71422e-23], [5106.86, 2.12038e-22, 4.95603e-23, 2.21687e-23, 5.52225e-23, 2.89341e-22, 8.73664e-23], [5118.48, 2.1252e-22, 4.96707e-23, 2.2218e-23, 5.535e-23, 2.90006e-22, 8.75911e-23], [5130.12, 2.13003e-22, 4.97814e-23, 2.22675e-23, 5.54779e-23, 2.90671e-22, 8.78163e-23], [5141.79, 2.13487e-22, 4.98923e-23, 2.23171e-23, 5.5606e-23, 2.91339e-22, 8.80419e-23], [5153.48, 2.13972e-22, 5.00035e-23, 2.23668e-23, 5.57345e-23, 2.92007e-22, 8.82679e-23], [5165.2, 2.14459e-22, 5.01149e-23, 2.24166e-23, 5.58632e-23, 2.92678e-22, 8.84945e-23], [5176.95, 2.14946e-22, 5.02266e-23, 2.24666e-23, 5.59922e-23, 2.93349e-22, 8.87214e-23], [5188.73, 2.15435e-22, 5.03386e-23, 2.25166e-23, 5.61215e-23, 2.94023e-22, 8.89489e-23], [5200.53, 2.15924e-22, 5.04508e-23, 2.25668e-23, 5.62511e-23, 2.94698e-22, 8.91768e-23], [5212.36, 2.16415e-22, 5.05633e-23, 2.26171e-23, 5.63809e-23, 2.95374e-22, 8.94052e-23], [5224.22, 2.16907e-22, 5.06761e-23, 2.26675e-23, 5.65111e-23, 2.96052e-22, 8.96341e-23], [5236.1, 2.174e-22, 5.07891e-23, 2.2718e-23, 5.66415e-23, 2.96731e-22, 8.98634e-23], [5248.01, 2.17894e-22, 5.09024e-23, 2.27686e-23, 5.67723e-23, 2.97412e-22, 9.00932e-23], [5259.95, 2.1839e-22, 5.10159e-23, 2.28194e-23, 5.69033e-23, 2.98094e-22, 9.03235e-23], [5271.91, 2.18886e-22, 5.11297e-23, 2.28703e-23, 5.70347e-23, 2.98778e-22, 9.05542e-23], [5283.9, 2.19384e-22, 5.12438e-23, 2.29213e-23, 5.71663e-23, 2.99464e-22, 9.07854e-23], [5295.92, 2.19882e-22, 5.13581e-23, 2.29724e-23, 5.72982e-23, 3.00151e-22, 9.10171e-23], [5307.97, 2.20382e-22, 5.14727e-23, 2.30236e-23, 5.74304e-23, 3.0084e-22, 9.12493e-23], [5320.04, 2.20883e-22, 5.15875e-23, 2.30749e-23, 5.7563e-23, 3.0153e-22, 9.14819e-23], [5332.14, 2.21385e-22, 5.17026e-23, 2.31264e-23, 5.76958e-23, 3.02222e-22, 9.1715e-23], [5344.27, 2.21888e-22, 5.1818e-23, 2.3178e-23, 5.78289e-23, 3.02915e-22, 9.19487e-23], [5356.42, 2.22393e-22, 5.19337e-23, 2.32297e-23, 5.79623e-23, 3.0361e-22, 9.21827e-23], [5368.61, 2.22898e-22, 5.20496e-23, 2.32815e-23, 5.8096e-23, 3.04306e-22, 9.24173e-23], [5380.82, 2.23405e-22, 5.21658e-23, 2.33335e-23, 5.823e-23, 3.05004e-22, 9.26523e-23], [5393.06, 2.23913e-22, 5.22823e-23, 2.33855e-23, 5.83643e-23, 3.05704e-22, 9.28879e-23], [5405.32, 2.24422e-22, 5.2399e-23, 2.34377e-23, 5.8499e-23, 3.06405e-22, 9.31239e-23], [5417.62, 2.24932e-22, 5.2516e-23, 2.349e-23, 5.86339e-23, 3.07108e-22, 9.33604e-23], [5429.94, 2.25443e-22, 5.26333e-23, 2.35425e-23, 5.87691e-23, 3.07812e-22, 9.35974e-23], [5442.29, 2.25956e-22, 5.27508e-23, 2.3595e-23, 5.89046e-23, 3.08518e-22, 9.38349e-23], [5454.67, 2.26469e-22, 5.28687e-23, 2.36477e-23, 5.90404e-23, 3.09225e-22, 9.40729e-23], [5467.08, 2.26984e-22, 5.29867e-23, 2.37005e-23, 5.91766e-23, 3.09934e-22, 9.43113e-23], [5479.51, 2.275e-22, 5.31051e-23, 2.37534e-23, 5.9313e-23, 3.10645e-22, 9.45503e-23], [5491.98, 2.28017e-22, 5.32238e-23, 2.38064e-23, 5.94497e-23, 3.11357e-22, 9.47897e-23], [5504.47, 2.28536e-22, 5.33427e-23, 2.38596e-23, 5.95868e-23, 3.12071e-22, 9.50297e-23], [5516.99, 2.29055e-22, 5.34619e-23, 2.39129e-23, 5.97241e-23, 3.12787e-22, 9.52701e-23], [5529.54, 2.29576e-22, 5.35813e-23, 2.39663e-23, 5.98618e-23, 3.13504e-22, 9.55111e-23], [5542.12, 2.30098e-22, 5.37011e-23, 2.40198e-23, 5.99998e-23, 3.14223e-22, 9.57525e-23], [5554.72, 2.30621e-22, 5.38211e-23, 2.40735e-23, 6.01381e-23, 3.14943e-22, 9.59944e-23], [5567.36, 2.31145e-22, 5.39414e-23, 2.41273e-23, 6.02767e-23, 3.15665e-22, 9.62369e-23], [5580.02, 2.31671e-22, 5.4062e-23, 2.41812e-23, 6.04156e-23, 3.16389e-22, 9.64798e-23], [5592.71, 2.32197e-22, 5.41828e-23, 2.42352e-23, 6.05548e-23, 3.17114e-22, 9.67233e-23], [5605.43, 2.32725e-22, 5.43039e-23, 2.42894e-23, 6.06943e-23, 3.17841e-22, 9.69672e-23], [5618.18, 2.33254e-22, 5.44254e-23, 2.43437e-23, 6.08342e-23, 3.1857e-22, 9.72117e-23], [5630.96, 2.33784e-22, 5.4547e-23, 2.43981e-23, 6.09743e-23, 3.193e-22, 9.74566e-23], [5643.77, 2.34316e-22, 5.4669e-23, 2.44526e-23, 6.11148e-23, 3.20032e-22, 9.77021e-23], [5656.61, 2.34848e-22, 5.47913e-23, 2.45073e-23, 6.12556e-23, 3.20765e-22, 9.79481e-23], [5669.47, 2.35382e-22, 5.49138e-23, 2.4562e-23, 6.13967e-23, 3.215e-22, 9.81946e-23], [5682.37, 2.35917e-22, 5.50366e-23, 2.46169e-23, 6.15381e-23, 3.22237e-22, 9.84416e-23], [5695.29, 2.36454e-22, 5.51598e-23, 2.4672e-23, 6.16799e-23, 3.22975e-22, 9.86891e-23], [5708.25, 2.36991e-22, 5.52831e-23, 2.47272e-23, 6.18219e-23, 3.23716e-22, 9.89371e-23], [5721.23, 2.3753e-22, 5.54068e-23, 2.47825e-23, 6.19643e-23, 3.24457e-22, 9.91857e-23], [5734.25, 2.3807e-22, 5.55308e-23, 2.48379e-23, 6.2107e-23, 3.25201e-22, 9.94347e-23], [5747.29, 2.38611e-22, 5.5655e-23, 2.48934e-23, 6.225e-23, 3.25946e-22, 9.96843e-23], [5760.36, 2.39153e-22, 5.57796e-23, 2.49491e-23, 6.23934e-23, 3.26693e-22, 9.99344e-23], [5773.46, 2.39697e-22, 5.59044e-23, 2.50049e-23, 6.2537e-23, 3.27441e-22, 1.00185e-22], [5786.6, 2.40242e-22, 5.60295e-23, 2.50608e-23, 6.2681e-23, 3.28192e-22, 1.00436e-22], [5799.76, 2.40788e-22, 5.61549e-23, 2.51169e-23, 6.28253e-23, 3.28944e-22, 1.00688e-22], [5812.95, 2.41336e-22, 5.62806e-23, 2.51731e-23, 6.29699e-23, 3.29697e-22, 1.0094e-22], [5826.17, 2.41884e-22, 5.64066e-23, 2.52294e-23, 6.31149e-23, 3.30452e-22, 1.01193e-22], [5839.42, 2.42434e-22, 5.65329e-23, 2.52859e-23, 6.32602e-23, 3.31209e-22, 1.01446e-22], [5852.71, 2.42985e-22, 5.66594e-23, 2.53425e-23, 6.34058e-23, 3.31968e-22, 1.017e-22], [5866.02, 2.43537e-22, 5.67863e-23, 2.53992e-23, 6.35517e-23, 3.32729e-22, 1.01954e-22], [5879.36, 2.44091e-22, 5.69135e-23, 2.5456e-23, 6.3698e-23, 3.33491e-22, 1.02209e-22], [5892.73, 2.44646e-22, 5.70409e-23, 2.5513e-23, 6.38446e-23, 3.34255e-22, 1.02464e-22], [5906.14, 2.45202e-22, 5.71686e-23, 2.55701e-23, 6.39915e-23, 3.3502e-22, 1.0272e-22], [5919.57, 2.4576e-22, 5.72967e-23, 2.56274e-23, 6.41388e-23, 3.35788e-22, 1.02977e-22], [5933.04, 2.46318e-22, 5.7425e-23, 2.56848e-23, 6.42864e-23, 3.36557e-22, 1.03234e-22], [5946.53, 2.46878e-22, 5.75536e-23, 2.57423e-23, 6.44343e-23, 3.37328e-22, 1.03491e-22], [5960.06, 2.47439e-22, 5.76826e-23, 2.57999e-23, 6.45825e-23, 3.381e-22, 1.03749e-22], [5973.61, 2.48002e-22, 5.78118e-23, 2.58577e-23, 6.47311e-23, 3.38874e-22, 1.04008e-22], [5987.2, 2.48566e-22, 5.79413e-23, 2.59156e-23, 6.48801e-23, 3.3965e-22, 1.04267e-22], [6000.82, 2.49131e-22, 5.80711e-23, 2.59736e-23, 6.50293e-23, 3.40428e-22, 1.04527e-22], [6014.47, 2.49697e-22, 5.82012e-23, 2.60318e-23, 6.51789e-23, 3.41208e-22, 1.04787e-22], [6028.15, 2.50265e-22, 5.83317e-23, 2.60901e-23, 6.53288e-23, 3.41989e-22, 1.05048e-22], [6041.86, 2.50834e-22, 5.84624e-23, 2.61486e-23, 6.54791e-23, 3.42772e-22, 1.05309e-22], [6055.6, 2.51404e-22, 5.85934e-23, 2.62071e-23, 6.56297e-23, 3.43557e-22, 1.05571e-22], [6069.38, 2.51975e-22, 5.87247e-23, 2.62659e-23, 6.57806e-23, 3.44344e-22, 1.05833e-22], [6083.18, 2.52548e-22, 5.88564e-23, 2.63247e-23, 6.59319e-23, 3.45132e-22, 1.06096e-22], [6097.02, 2.53122e-22, 5.89883e-23, 2.63837e-23, 6.60835e-23, 3.45922e-22, 1.0636e-22], [6110.89, 2.53698e-22, 5.91205e-23, 2.64428e-23, 6.62355e-23, 3.46714e-22, 1.06624e-22], [6124.79, 2.54275e-22, 5.92531e-23, 2.65021e-23, 6.63878e-23, 3.47508e-22, 1.06888e-22], [6138.72, 2.54853e-22, 5.93859e-23, 2.65615e-23, 6.65405e-23, 3.48304e-22, 1.07154e-22], [6152.68, 2.55432e-22, 5.95191e-23, 2.6621e-23, 6.66934e-23, 3.49101e-22, 1.07419e-22], [6166.67, 2.56013e-22, 5.96525e-23, 2.66807e-23, 6.68468e-23, 3.499e-22, 1.07686e-22], [6180.7, 2.56595e-22, 5.97863e-23, 2.67405e-23, 6.70005e-23, 3.50701e-22, 1.07952e-22], [6194.76, 2.57178e-22, 5.99204e-23, 2.68004e-23, 6.71545e-23, 3.51504e-22, 1.0822e-22], [6208.85, 2.57763e-22, 6.00548e-23, 2.68605e-23, 6.73089e-23, 3.52308e-22, 1.08488e-22], [6222.97, 2.58349e-22, 6.01895e-23, 2.69208e-23, 6.74636e-23, 3.53115e-22, 1.08756e-22], [6237.13, 2.58936e-22, 6.03245e-23, 2.69811e-23, 6.76187e-23, 3.53923e-22, 1.09025e-22], [6251.31, 2.59525e-22, 6.04598e-23, 2.70416e-23, 6.77741e-23, 3.54733e-22, 1.09295e-22], [6265.53, 2.60115e-22, 6.05954e-23, 2.71023e-23, 6.79298e-23, 3.55545e-22, 1.09565e-22], [6279.78, 2.60706e-22, 6.07314e-23, 2.7163e-23, 6.8086e-23, 3.56359e-22, 1.09836e-22], [6294.07, 2.61299e-22, 6.08676e-23, 2.7224e-23, 6.82424e-23, 3.57174e-22, 1.10108e-22], [6308.38, 2.61893e-22, 6.10042e-23, 2.7285e-23, 6.83993e-23, 3.57992e-22, 1.10379e-22], [6322.73, 2.62488e-22, 6.11411e-23, 2.73462e-23, 6.85564e-23, 3.58811e-22, 1.10652e-22], [6337.11, 2.63085e-22, 6.12783e-23, 2.74076e-23, 6.8714e-23, 3.59632e-22, 1.10925e-22], [6351.53, 2.63683e-22, 6.14158e-23, 2.74691e-23, 6.88718e-23, 3.60455e-22, 1.11199e-22], [6365.98, 2.64282e-22, 6.15536e-23, 2.75307e-23, 6.90301e-23, 3.6128e-22, 1.11473e-22], [6380.46, 2.64883e-22, 6.16918e-23, 2.75925e-23, 6.91887e-23, 3.62106e-22, 1.11748e-22], [6394.97, 2.65485e-22, 6.18302e-23, 2.76544e-23, 6.93476e-23, 3.62935e-22, 1.12023e-22], [6409.51, 2.66089e-22, 6.1969e-23, 2.77164e-23, 6.95069e-23, 3.63765e-22, 1.12299e-22], [6424.09, 2.66694e-22, 6.21081e-23, 2.77786e-23, 6.96666e-23, 3.64598e-22, 1.12576e-22], [6438.71, 2.673e-22, 6.22476e-23, 2.7841e-23, 6.98266e-23, 3.65432e-22, 1.12853e-22], [6453.35, 2.67908e-22, 6.23873e-23, 2.79035e-23, 6.9987e-23, 3.66268e-22, 1.13131e-22], [6468.03, 2.68517e-22, 6.25274e-23, 2.79661e-23, 7.01478e-23, 3.67106e-22, 1.13409e-22], [6482.74, 2.69127e-22, 6.26678e-23, 2.80289e-23, 7.03089e-23, 3.67946e-22, 1.13688e-22], [6497.49, 2.69739e-22, 6.28085e-23, 2.80918e-23, 7.04704e-23, 3.68788e-22, 1.13968e-22], [6512.27, 2.70352e-22, 6.29496e-23, 2.81548e-23, 7.06322e-23, 3.69631e-22, 1.14248e-22], [6527.08, 2.70967e-22, 6.30909e-23, 2.8218e-23, 7.07944e-23, 3.70477e-22, 1.14528e-22], [6541.92, 2.71583e-22, 6.32326e-23, 2.82814e-23, 7.0957e-23, 3.71324e-22, 1.1481e-22], [6556.8, 2.72201e-22, 6.33746e-23, 2.83449e-23, 7.11199e-23, 3.72174e-22, 1.15092e-22], [6571.72, 2.72819e-22, 6.3517e-23, 2.84086e-23, 7.12833e-23, 3.73025e-22, 1.15374e-22], [6586.67, 2.7344e-22, 6.36597e-23, 2.84723e-23, 7.14469e-23, 3.73878e-22, 1.15657e-22], [6601.65, 2.74061e-22, 6.38027e-23, 2.85363e-23, 7.1611e-23, 3.74733e-22, 1.15941e-22], [6616.66, 2.74684e-22, 6.3946e-23, 2.86004e-23, 7.17754e-23, 3.75591e-22, 1.16225e-22], [6631.71, 2.75309e-22, 6.40897e-23, 2.86646e-23, 7.19402e-23, 3.7645e-22, 1.1651e-22], [6646.8, 2.75935e-22, 6.42337e-23, 2.8729e-23, 7.21053e-23, 3.77311e-22, 1.16796e-22], [6661.92, 2.76562e-22, 6.4378e-23, 2.87935e-23, 7.22708e-23, 3.78174e-22, 1.17082e-22], [6677.07, 2.77191e-22, 6.45226e-23, 2.88582e-23, 7.24367e-23, 3.79038e-22, 1.17368e-22], [6692.26, 2.77821e-22, 6.46676e-23, 2.8923e-23, 7.2603e-23, 3.79905e-22, 1.17656e-22], [6707.48, 2.78453e-22, 6.4813e-23, 2.8988e-23, 7.27697e-23, 3.80774e-22, 1.17944e-22], [6722.74, 2.79086e-22, 6.49586e-23, 2.90531e-23, 7.29367e-23, 3.81645e-22, 1.18232e-22], [6738.03, 2.7972e-22, 6.51046e-23, 2.91184e-23, 7.31041e-23, 3.82518e-22, 1.18521e-22], [6753.35, 2.80356e-22, 6.52509e-23, 2.91838e-23, 7.32719e-23, 3.83392e-22, 1.18811e-22], [6768.71, 2.80993e-22, 6.53976e-23, 2.92494e-23, 7.344e-23, 3.84269e-22, 1.19102e-22], [6784.11, 2.81632e-22, 6.55446e-23, 2.93152e-23, 7.36086e-23, 3.85148e-22, 1.19393e-22], [6799.54, 2.82273e-22, 6.5692e-23, 2.9381e-23, 7.37775e-23, 3.86028e-22, 1.19684e-22], [6815.01, 2.82914e-22, 6.58396e-23, 2.94471e-23, 7.39468e-23, 3.86911e-22, 1.19976e-22], [6830.51, 2.83557e-22, 6.59877e-23, 2.95133e-23, 7.41165e-23, 3.87796e-22, 1.20269e-22], [6846.04, 2.84202e-22, 6.6136e-23, 2.95796e-23, 7.42865e-23, 3.88682e-22, 1.20563e-22], [6861.62, 2.84848e-22, 6.62847e-23, 2.96461e-23, 7.4457e-23, 3.89571e-22, 1.20857e-22], [6877.22, 2.85496e-22, 6.64338e-23, 2.97127e-23, 7.46278e-23, 3.90462e-22, 1.21152e-22], [6892.87, 2.86145e-22, 6.65832e-23, 2.97795e-23, 7.4799e-23, 3.91355e-22, 1.21447e-22], [6908.54, 2.86796e-22, 6.67329e-23, 2.98465e-23, 7.49706e-23, 3.92249e-22, 1.21743e-22], [6924.26, 2.87448e-22, 6.6883e-23, 2.99136e-23, 7.51426e-23, 3.93146e-22, 1.2204e-22], [6940.01, 2.88101e-22, 6.70334e-23, 2.99808e-23, 7.5315e-23, 3.94045e-22, 1.22337e-22], [6955.79, 2.88756e-22, 6.71842e-23, 3.00483e-23, 7.54877e-23, 3.94946e-22, 1.22635e-22], [6971.62, 2.89413e-22, 6.73353e-23, 3.01158e-23, 7.56609e-23, 3.95848e-22, 1.22933e-22], [6987.47, 2.90071e-22, 6.74868e-23, 3.01836e-23, 7.58344e-23, 3.96753e-22, 1.23233e-22], [7003.37, 2.9073e-22, 6.76386e-23, 3.02514e-23, 7.60084e-23, 3.9766e-22, 1.23532e-22], [7019.3, 2.91391e-22, 6.77907e-23, 3.03195e-23, 7.61827e-23, 3.98569e-22, 1.23833e-22], [7035.26, 2.92053e-22, 6.79432e-23, 3.03877e-23, 7.63574e-23, 3.9948e-22, 1.24134e-22], [7051.26, 2.92717e-22, 6.80961e-23, 3.0456e-23, 7.65325e-23, 4.00393e-22, 1.24436e-22], [7067.3, 2.93383e-22, 6.82493e-23, 3.05245e-23, 7.67081e-23, 4.01309e-22, 1.24738e-22], [7083.38, 2.9405e-22, 6.84029e-23, 3.05932e-23, 7.6884e-23, 4.02226e-22, 1.25041e-22], [7099.49, 2.94718e-22, 6.85568e-23, 3.0662e-23, 7.70603e-23, 4.03145e-22, 1.25345e-22], [7115.64, 2.95389e-22, 6.87111e-23, 3.0731e-23, 7.7237e-23, 4.04067e-22, 1.25649e-22], [7131.82, 2.9606e-22, 6.88657e-23, 3.08001e-23, 7.74141e-23, 4.0499e-22, 1.25954e-22], [7148.04, 2.96733e-22, 6.90207e-23, 3.08694e-23, 7.75916e-23, 4.05916e-22, 1.2626e-22], [7164.3, 2.97408e-22, 6.9176e-23, 3.09389e-23, 7.77695e-23, 4.06843e-22, 1.26566e-22], [7180.6, 2.98084e-22, 6.93317e-23, 3.10085e-23, 7.79478e-23, 4.07773e-22, 1.26873e-22], [7196.93, 2.98762e-22, 6.94878e-23, 3.10783e-23, 7.81265e-23, 4.08705e-22, 1.27181e-22], [7213.3, 2.99441e-22, 6.96442e-23, 3.11482e-23, 7.83056e-23, 4.09639e-22, 1.27489e-22], [7229.71, 3.00122e-22, 6.98009e-23, 3.12183e-23, 7.84851e-23, 4.10575e-22, 1.27798e-22], [7246.15, 3.00804e-22, 6.99581e-23, 3.12886e-23, 7.8665e-23, 4.11513e-22, 1.28107e-22], [7262.64, 3.01488e-22, 7.01156e-23, 3.1359e-23, 7.88453e-23, 4.12454e-22, 1.28418e-22], [7279.15, 3.02173e-22, 7.02734e-23, 3.14296e-23, 7.90261e-23, 4.13396e-22, 1.28729e-22], [7295.71, 3.0286e-22, 7.04316e-23, 3.15003e-23, 7.92072e-23, 4.14341e-22, 1.2904e-22], [7312.31, 3.03549e-22, 7.05902e-23, 3.15712e-23, 7.93888e-23, 4.15288e-22, 1.29353e-22], [7328.94, 3.04239e-22, 7.07492e-23, 3.16423e-23, 7.95707e-23, 4.16236e-22, 1.29665e-22], [7345.61, 3.04931e-22, 7.09085e-23, 3.17135e-23, 7.97531e-23, 4.17188e-22, 1.29979e-22], [7362.32, 3.05624e-22, 7.10682e-23, 3.17849e-23, 7.99359e-23, 4.18141e-22, 1.30293e-22], [7379.06, 3.06319e-22, 7.12282e-23, 3.18565e-23, 8.01191e-23, 4.19096e-22, 1.30608e-22], [7395.85, 3.07015e-22, 7.13886e-23, 3.19282e-23, 8.03027e-23, 4.20054e-22, 1.30924e-22], [7412.67, 3.07713e-22, 7.15494e-23, 3.20001e-23, 8.04867e-23, 4.21013e-22, 1.3124e-22], [7429.53, 3.08413e-22, 7.17105e-23, 3.20722e-23, 8.06711e-23, 4.21975e-22, 1.31557e-22], [7446.43, 3.09114e-22, 7.18721e-23, 3.21444e-23, 8.0856e-23, 4.22939e-22, 1.31875e-22], [7463.37, 3.09817e-22, 7.20339e-23, 3.22168e-23, 8.10413e-23, 4.23905e-22, 1.32193e-22], [7480.34, 3.10521e-22, 7.21962e-23, 3.22893e-23, 8.1227e-23, 4.24874e-22, 1.32512e-22], [7497.36, 3.11227e-22, 7.23588e-23, 3.23621e-23, 8.14131e-23, 4.25844e-22, 1.32832e-22], [7514.41, 3.11935e-22, 7.25218e-23, 3.2435e-23, 8.15996e-23, 4.26817e-22, 1.33152e-22], [7531.5, 3.12644e-22, 7.26852e-23, 3.2508e-23, 8.17865e-23, 4.27792e-22, 1.33473e-22], [7548.63, 3.13355e-22, 7.2849e-23, 3.25812e-23, 8.19739e-23, 4.28769e-22, 1.33795e-22], [7565.8, 3.14068e-22, 7.30131e-23, 3.26546e-23, 8.21617e-23, 4.29749e-22, 1.34118e-22], [7583.01, 3.14782e-22, 7.31776e-23, 3.27282e-23, 8.23499e-23, 4.30731e-22, 1.34441e-22], [7600.26, 3.15497e-22, 7.33425e-23, 3.28019e-23, 8.25386e-23, 4.31714e-22, 1.34765e-22], [7617.55, 3.16215e-22, 7.35078e-23, 3.28758e-23, 8.27276e-23, 4.32701e-22, 1.35089e-22], [7634.87, 3.16934e-22, 7.36734e-23, 3.29499e-23, 8.29171e-23, 4.33689e-22, 1.35415e-22], [7652.24, 3.17654e-22, 7.38395e-23, 3.30241e-23, 8.31071e-23, 4.34679e-22, 1.35741e-22], [7669.65, 3.18376e-22, 7.40059e-23, 3.30985e-23, 8.32974e-23, 4.35672e-22, 1.36067e-22], [7687.09, 3.191e-22, 7.41727e-23, 3.31731e-23, 8.34882e-23, 4.36667e-22, 1.36395e-22], [7704.58, 3.19826e-22, 7.43398e-23, 3.32479e-23, 8.36794e-23, 4.37665e-22, 1.36723e-22], [7722.1, 3.20553e-22, 7.45074e-23, 3.33228e-23, 8.38711e-23, 4.38664e-22, 1.37051e-22], [7739.66, 3.21282e-22, 7.46753e-23, 3.33979e-23, 8.40632e-23, 4.39666e-22, 1.37381e-22], [7757.27, 3.22012e-22, 7.48436e-23, 3.34732e-23, 8.42557e-23, 4.4067e-22, 1.37711e-22], [7774.91, 3.22744e-22, 7.50124e-23, 3.35486e-23, 8.44486e-23, 4.41677e-22, 1.38042e-22], [7792.6, 3.23478e-22, 7.51815e-23, 3.36242e-23, 8.4642e-23, 4.42685e-22, 1.38374e-22], [7810.32, 3.24214e-22, 7.53509e-23, 3.37e-23, 8.48358e-23, 4.43696e-22, 1.38706e-22], [7828.09, 3.24951e-22, 7.55208e-23, 3.3776e-23, 8.50301e-23, 4.44709e-22, 1.39039e-22], [7845.89, 3.2569e-22, 7.56911e-23, 3.38521e-23, 8.52248e-23, 4.45725e-22, 1.39373e-22], [7863.74, 3.2643e-22, 7.58617e-23, 3.39284e-23, 8.54199e-23, 4.46743e-22, 1.39708e-22], [7881.63, 3.27172e-22, 7.60328e-23, 3.40049e-23, 8.56155e-23, 4.47763e-22, 1.40043e-22], [7899.55, 3.27916e-22, 7.62042e-23, 3.40815e-23, 8.58115e-23, 4.48785e-22, 1.40379e-22], [7917.52, 3.28662e-22, 7.63761e-23, 3.41584e-23, 8.6008e-23, 4.4981e-22, 1.40715e-22], [7935.53, 3.29409e-22, 7.65483e-23, 3.42354e-23, 8.62049e-23, 4.50837e-22, 1.41053e-22], [7953.58, 3.30158e-22, 7.67209e-23, 3.43126e-23, 8.64023e-23, 4.51867e-22, 1.41391e-22], [7971.67, 3.30909e-22, 7.68939e-23, 3.439e-23, 8.66001e-23, 4.52898e-22, 1.4173e-22], [7989.8, 3.31661e-22, 7.70674e-23, 3.44675e-23, 8.67983e-23, 4.53933e-22, 1.42069e-22], [8007.98, 3.32415e-22, 7.72412e-23, 3.45452e-23, 8.6997e-23, 4.54969e-22, 1.4241e-22], [8026.19, 3.33171e-22, 7.74154e-23, 3.46231e-23, 8.71962e-23, 4.56008e-22, 1.42751e-22], [8044.45, 3.33928e-22, 7.759e-23, 3.47012e-23, 8.73958e-23, 4.57049e-22, 1.43093e-22], [8062.75, 3.34688e-22, 7.7765e-23, 3.47795e-23, 8.75958e-23, 4.58092e-22, 1.43435e-22], [8081.09, 3.35449e-22, 7.79404e-23, 3.48579e-23, 8.77963e-23, 4.59138e-22, 1.43779e-22], [8099.47, 3.36211e-22, 7.81162e-23, 3.49365e-23, 8.79973e-23, 4.60187e-22, 1.44123e-22], [8117.89, 3.36976e-22, 7.82925e-23, 3.50153e-23, 8.81987e-23, 4.61237e-22, 1.44467e-22], [8136.35, 3.37742e-22, 7.84691e-23, 3.50943e-23, 8.84005e-23, 4.6229e-22, 1.44813e-22], [8154.86, 3.3851e-22, 7.86461e-23, 3.51735e-23, 8.86029e-23, 4.63346e-22, 1.45159e-22], [8173.41, 3.39279e-22, 7.88235e-23, 3.52528e-23, 8.88056e-23, 4.64403e-22, 1.45506e-22], [8192.0, 3.40051e-22, 7.90014e-23, 3.53323e-23, 8.90089e-23, 4.65463e-22, 1.45854e-22]])     
    elif source=='T1800044-v4':
        data = np.array([[5.0, 1.9997e-20], [5.0115, 1.9462e-20], [5.0231, 1.8942e-20], [5.0347, 1.8437e-20], [5.0463, 1.7946e-20], [5.0579, 1.7469e-20], [5.0696, 1.7005e-20], [5.0813, 1.6555e-20], [5.093, 1.6117e-20], [5.1047, 1.5691e-20], [5.1165, 1.5277e-20], [5.1283, 1.4875e-20], [5.1401, 1.4484e-20], [5.152, 1.4104e-20], [5.1639, 1.3734e-20], [5.1758, 1.3375e-20], [5.1877, 1.3026e-20], [5.1997, 1.2686e-20], [5.2117, 1.2355e-20], [5.2237, 1.2034e-20], [5.2357, 1.1722e-20], [5.2478, 1.1418e-20], [5.2599, 1.1122e-20], [5.272, 1.0835e-20], [5.2842, 1.0555e-20], [5.2964, 1.0283e-20], [5.3086, 1.0018e-20], [5.3208, 9.7607e-21], [5.3331, 9.5103e-21], [5.3454, 9.2666e-21], [5.3577, 9.0295e-21], [5.3701, 8.7988e-21], [5.3825, 8.5744e-21], [5.3949, 8.3559e-21], [5.4073, 8.1434e-21], [5.4198, 7.9365e-21], [5.4323, 7.7352e-21], [5.4448, 7.5393e-21], [5.4574, 7.3487e-21], [5.4699, 7.1631e-21], [5.4826, 6.9824e-21], [5.4952, 6.8066e-21], [5.5079, 6.6354e-21], [5.5206, 6.4688e-21], [5.5333, 6.3066e-21], [5.5461, 6.1487e-21], [5.5589, 5.995e-21], [5.5717, 5.8453e-21], [5.5845, 5.6996e-21], [5.5974, 5.5577e-21], [5.6103, 5.4195e-21], [5.6232, 5.285e-21], [5.6362, 5.154e-21], [5.6492, 5.0264e-21], [5.6622, 4.9022e-21], [5.6753, 4.7812e-21], [5.6884, 4.6634e-21], [5.7015, 4.5486e-21], [5.7146, 4.4369e-21], [5.7278, 4.328e-21], [5.741, 4.222e-21], [5.7543, 4.1187e-21], [5.7675, 4.0181e-21], [5.7808, 3.9201e-21], [5.7942, 3.8246e-21], [5.8075, 3.7316e-21], [5.8209, 3.641e-21], [5.8343, 3.5528e-21], [5.8478, 3.4668e-21], [5.8613, 3.383e-21], [5.8748, 3.3014e-21], [5.8884, 3.2218e-21], [5.9019, 3.1443e-21], [5.9155, 3.0688e-21], [5.9292, 2.9953e-21], [5.9429, 2.9236e-21], [5.9566, 2.8537e-21], [5.9703, 2.7856e-21], [5.9841, 2.7193e-21], [5.9979, 2.6546e-21], [6.0117, 2.5916e-21], [6.0256, 2.5302e-21], [6.0394, 2.4703e-21], [6.0534, 2.412e-21], [6.0673, 2.3551e-21], [6.0813, 2.2997e-21], [6.0954, 2.2457e-21], [6.1094, 2.1931e-21], [6.1235, 2.1418e-21], [6.1376, 2.0917e-21], [6.1518, 2.043e-21], [6.166, 1.9955e-21], [6.1802, 1.9491e-21], [6.1944, 1.904e-21], [6.2087, 1.86e-21], [6.223, 1.817e-21], [6.2374, 1.7752e-21], [6.2518, 1.7344e-21], [6.2662, 1.6947e-21], [6.2806, 1.6559e-21], [6.2951, 1.6181e-21], [6.3096, 1.5812e-21], [6.3242, 1.5453e-21], [6.3388, 1.5103e-21], [6.3534, 1.4762e-21], [6.368, 1.4429e-21], [6.3827, 1.4104e-21], [6.3974, 1.3788e-21], [6.4122, 1.3479e-21], [6.427, 1.3178e-21], [6.4418, 1.2885e-21], [6.4566, 1.2599e-21], [6.4715, 1.232e-21], [6.4865, 1.2048e-21], [6.5014, 1.1783e-21], [6.5164, 1.1524e-21], [6.5314, 1.1272e-21], [6.5465, 1.1027e-21], [6.5616, 1.0787e-21], [6.5767, 1.0553e-21], [6.5919, 1.0326e-21], [6.6071, 1.0103e-21], [6.6223, 9.8868e-22], [6.6376, 9.6756e-22], [6.6529, 9.4697e-22], [6.6682, 9.2689e-22], [6.6836, 9.0732e-22], [6.699, 8.8823e-22], [6.7145, 8.6962e-22], [6.73, 8.5148e-22], [6.7455, 8.3378e-22], [6.761, 8.1654e-22], [6.7766, 7.9972e-22], [6.7923, 7.8332e-22], [6.8079, 7.6734e-22], [6.8236, 7.5175e-22], [6.8394, 7.3656e-22], [6.8551, 7.2174e-22], [6.8709, 7.073e-22], [6.8868, 6.9323e-22], [6.9027, 6.795e-22], [6.9186, 6.6612e-22], [6.9345, 6.5308e-22], [6.9505, 6.4037e-22], [6.9666, 6.2798e-22], [6.9826, 6.159e-22], [6.9987, 6.0412e-22], [7.0149, 5.9265e-22], [7.031, 5.8146e-22], [7.0473, 5.7056e-22], [7.0635, 5.5994e-22], [7.0798, 5.4958e-22], [7.0961, 5.3949e-22], [7.1125, 5.2966e-22], [7.1289, 5.2007e-22], [7.1453, 5.1073e-22], [7.1618, 5.0163e-22], [7.1783, 4.9277e-22], [7.1949, 4.8413e-22], [7.2115, 4.7571e-22], [7.2281, 4.6751e-22], [7.2448, 4.5952e-22], [7.2615, 4.5173e-22], [7.2782, 4.4415e-22], [7.295, 4.3676e-22], [7.3118, 4.2957e-22], [7.3287, 4.2256e-22], [7.3456, 4.1573e-22], [7.3625, 4.0908e-22], [7.3795, 4.0261e-22], [7.3965, 3.963e-22], [7.4136, 3.9016e-22], [7.4307, 3.8418e-22], [7.4478, 3.7836e-22], [7.465, 3.7269e-22], [7.4822, 3.6717e-22], [7.4994, 3.618e-22], [7.5167, 3.5656e-22], [7.5341, 3.5147e-22], [7.5514, 3.4651e-22], [7.5689, 3.4169e-22], [7.5863, 3.3699e-22], [7.6038, 3.3242e-22], [7.6213, 3.2797e-22], [7.6389, 3.2365e-22], [7.6565, 3.1943e-22], [7.6742, 3.1534e-22], [7.6919, 3.1135e-22], [7.7096, 3.0747e-22], [7.7274, 3.037e-22], [7.7452, 3.0004e-22], [7.7631, 2.9647e-22], [7.781, 2.93e-22], [7.7989, 2.8963e-22], [7.8169, 2.8635e-22], [7.8349, 2.8316e-22], [7.853, 2.8007e-22], [7.8711, 2.7706e-22], [7.8893, 2.7413e-22], [7.9074, 2.7129e-22], [7.9257, 2.6853e-22], [7.944, 2.6586e-22], [7.9623, 2.6326e-22], [7.9806, 2.6073e-22], [7.999, 2.5828e-22], [8.0175, 2.5591e-22], [8.036, 2.5361e-22], [8.0545, 2.5137e-22], [8.0731, 2.4921e-22], [8.0917, 2.4712e-22], [8.1104, 2.4509e-22], [8.1291, 2.4312e-22], [8.1478, 2.4123e-22], [8.1666, 2.3939e-22], [8.1854, 2.3762e-22], [8.2043, 2.3591e-22], [8.2232, 2.3426e-22], [8.2422, 2.3268e-22], [8.2612, 2.3115e-22], [8.2802, 2.2968e-22], [8.2993, 2.2827e-22], [8.3185, 2.2692e-22], [8.3377, 2.2563e-22], [8.3569, 2.244e-22], [8.3762, 2.2322e-22], [8.3955, 2.221e-22], [8.4148, 2.2104e-22], [8.4342, 2.2004e-22], [8.4537, 2.191e-22], [8.4732, 2.1822e-22], [8.4927, 2.174e-22], [8.5123, 2.1664e-22], [8.5319, 2.1594e-22], [8.5516, 2.1531e-22], [8.5713, 2.1474e-22], [8.5911, 2.1423e-22], [8.6109, 2.138e-22], [8.6308, 2.1343e-22], [8.6507, 2.1314e-22], [8.6706, 2.1292e-22], [8.6906, 2.1278e-22], [8.7106, 2.1272e-22], [8.7307, 2.1274e-22], [8.7509, 2.1285e-22], [8.771, 2.1305e-22], [8.7913, 2.1335e-22], [8.8115, 2.1375e-22], [8.8319, 2.1426e-22], [8.8522, 2.1489e-22], [8.8726, 2.1564e-22], [8.8931, 2.1651e-22], [8.9136, 2.1753e-22], [8.9342, 2.187e-22], [8.9548, 2.2003e-22], [8.9754, 2.2154e-22], [8.9961, 2.2323e-22], [9.0169, 2.2513e-22], [9.0377, 2.2726e-22], [9.0585, 2.2963e-22], [9.0794, 2.3227e-22], [9.1003, 2.3521e-22], [9.1213, 2.3848e-22], [9.1423, 2.4211e-22], [9.1634, 2.4615e-22], [9.1846, 2.5065e-22], [9.2057, 2.5566e-22], [9.227, 2.6125e-22], [9.2482, 2.675e-22], [9.2696, 2.7451e-22], [9.2909, 2.8238e-22], [9.3124, 2.9126e-22], [9.3338, 3.0131e-22], [9.3554, 3.1274e-22], [9.3769, 3.2581e-22], [9.3986, 3.4084e-22], [9.4202, 3.5826e-22], [9.442, 3.786e-22], [9.4637, 4.026e-22], [9.4856, 4.3124e-22], [9.5074, 4.659e-22], [9.5294, 5.0857e-22], [9.5513, 5.6218e-22], [9.5734, 6.3135e-22], [9.5954, 7.237e-22], [9.6176, 8.5276e-22], [9.6397, 1.0452e-21], [9.662, 1.3616e-21], [9.6842, 1.9762e-21], [9.7066, 3.6774e-21], [9.729, 3.0668e-20], [9.7514, 4.7297e-21], [9.7739, 2.1756e-21], [9.7964, 1.4056e-21], [9.819, 1.0349e-21], [9.8417, 8.1736e-22], [9.8643, 6.7459e-22], [9.8871, 5.7392e-22], [9.9099, 4.9927e-22], [9.9327, 4.4183e-22], [9.9557, 3.9636e-22], [9.9786, 3.5955e-22], [10.002, 3.2919e-22], [10.025, 3.0377e-22], [10.048, 2.8223e-22], [10.071, 2.6375e-22], [10.094, 2.4777e-22], [10.117, 2.3383e-22], [10.141, 2.2158e-22], [10.164, 2.1074e-22], [10.188, 2.011e-22], [10.211, 1.9248e-22], [10.235, 1.8473e-22], [10.258, 1.7773e-22], [10.282, 1.7139e-22], [10.306, 1.6562e-22], [10.329, 1.6035e-22], [10.353, 1.5552e-22], [10.377, 1.5108e-22], [10.401, 1.4699e-22], [10.425, 1.432e-22], [10.449, 1.3969e-22], [10.473, 1.3643e-22], [10.497, 1.3339e-22], [10.522, 1.3056e-22], [10.546, 1.279e-22], [10.57, 1.2541e-22], [10.594, 1.2306e-22], [10.619, 1.2085e-22], [10.643, 1.1877e-22], [10.668, 1.1679e-22], [10.693, 1.1492e-22], [10.717, 1.1315e-22], [10.742, 1.1146e-22], [10.767, 1.0986e-22], [10.791, 1.0832e-22], [10.816, 1.0686e-22], [10.841, 1.0545e-22], [10.866, 1.0411e-22], [10.891, 1.0282e-22], [10.916, 1.0158e-22], [10.942, 1.0039e-22], [10.967, 9.9242e-23], [10.992, 9.8134e-23], [11.018, 9.7064e-23], [11.043, 9.603e-23], [11.068, 9.5029e-23], [11.094, 9.406e-23], [11.12, 9.312e-23], [11.145, 9.2207e-23], [11.171, 9.132e-23], [11.197, 9.0458e-23], [11.222, 8.9619e-23], [11.248, 8.8802e-23], [11.274, 8.8005e-23], [11.3, 8.7228e-23], [11.326, 8.6469e-23], [11.352, 8.5728e-23], [11.379, 8.5004e-23], [11.405, 8.4296e-23], [11.431, 8.3602e-23], [11.458, 8.2923e-23], [11.484, 8.2258e-23], [11.51, 8.1606e-23], [11.537, 8.0966e-23], [11.564, 8.0339e-23], [11.59, 7.9722e-23], [11.617, 7.9117e-23], [11.644, 7.8523e-23], [11.671, 7.7938e-23], [11.697, 7.7364e-23], [11.724, 7.6798e-23], [11.752, 7.6242e-23], [11.779, 7.5694e-23], [11.806, 7.5155e-23], [11.833, 7.4624e-23], [11.86, 7.4101e-23], [11.888, 7.3585e-23], [11.915, 7.3076e-23], [11.943, 7.2575e-23], [11.97, 7.208e-23], [11.998, 7.1592e-23], [12.025, 7.1111e-23], [12.053, 7.0636e-23], [12.081, 7.0166e-23], [12.109, 6.9703e-23], [12.137, 6.9246e-23], [12.165, 6.8794e-23], [12.193, 6.8347e-23], [12.221, 6.7906e-23], [12.249, 6.747e-23], [12.277, 6.7039e-23], [12.306, 6.6612e-23], [12.334, 6.6191e-23], [12.362, 6.5774e-23], [12.391, 6.5362e-23], [12.419, 6.4955e-23], [12.448, 6.4551e-23], [12.477, 6.4152e-23], [12.506, 6.3757e-23], [12.534, 6.3367e-23], [12.563, 6.298e-23], [12.592, 6.2597e-23], [12.621, 6.2218e-23], [12.65, 6.1843e-23], [12.68, 6.1471e-23], [12.709, 6.1103e-23], [12.738, 6.0739e-23], [12.767, 6.0378e-23], [12.797, 6.002e-23], [12.826, 5.9666e-23], [12.856, 5.9315e-23], [12.886, 5.8968e-23], [12.915, 5.8623e-23], [12.945, 5.8282e-23], [12.975, 5.7944e-23], [13.005, 5.7609e-23], [13.035, 5.7276e-23], [13.065, 5.6947e-23], [13.095, 5.6621e-23], [13.125, 5.6297e-23], [13.156, 5.5976e-23], [13.186, 5.5658e-23], [13.216, 5.5343e-23], [13.247, 5.503e-23], [13.277, 5.472e-23], [13.308, 5.4413e-23], [13.339, 5.4108e-23], [13.369, 5.3806e-23], [13.4, 5.3506e-23], [13.431, 5.3208e-23], [13.462, 5.2913e-23], [13.493, 5.2621e-23], [13.524, 5.233e-23], [13.555, 5.2042e-23], [13.587, 5.1756e-23], [13.618, 5.1473e-23], [13.649, 5.1192e-23], [13.681, 5.0912e-23], [13.712, 5.0636e-23], [13.744, 5.0361e-23], [13.776, 5.0088e-23], [13.808, 4.9817e-23], [13.839, 4.9549e-23], [13.871, 4.9282e-23], [13.903, 4.9018e-23], [13.935, 4.8755e-23], [13.968, 4.8494e-23], [14.0, 4.8235e-23], [14.032, 4.7979e-23], [14.064, 4.7724e-23], [14.097, 4.7471e-23], [14.129, 4.7219e-23], [14.162, 4.697e-23], [14.195, 4.6722e-23], [14.227, 4.6476e-23], [14.26, 4.6232e-23], [14.293, 4.599e-23], [14.326, 4.5749e-23], [14.359, 4.551e-23], [14.392, 4.5273e-23], [14.425, 4.5037e-23], [14.459, 4.4803e-23], [14.492, 4.4571e-23], [14.525, 4.434e-23], [14.559, 4.411e-23], [14.592, 4.3883e-23], [14.626, 4.3657e-23], [14.66, 4.3432e-23], [14.694, 4.3209e-23], [14.727, 4.2987e-23], [14.761, 4.2767e-23], [14.795, 4.2549e-23], [14.83, 4.2332e-23], [14.864, 4.2116e-23], [14.898, 4.1902e-23], [14.932, 4.1689e-23], [14.967, 4.1477e-23], [15.001, 4.1267e-23], [15.036, 4.1058e-23], [15.071, 4.0851e-23], [15.105, 4.0645e-23], [15.14, 4.044e-23], [15.175, 4.0237e-23], [15.21, 4.0035e-23], [15.245, 3.9834e-23], [15.28, 3.9635e-23], [15.316, 3.9436e-23], [15.351, 3.9239e-23], [15.386, 3.9044e-23], [15.422, 3.8849e-23], [15.457, 3.8656e-23], [15.493, 3.8464e-23], [15.529, 3.8273e-23], [15.564, 3.8083e-23], [15.6, 3.7895e-23], [15.636, 3.7708e-23], [15.672, 3.7522e-23], [15.709, 3.7337e-23], [15.745, 3.7153e-23], [15.781, 3.697e-23], [15.817, 3.6788e-23], [15.854, 3.6608e-23], [15.89, 3.6429e-23], [15.927, 3.625e-23], [15.964, 3.6073e-23], [16.001, 3.5897e-23], [16.038, 3.5722e-23], [16.075, 3.5548e-23], [16.112, 3.5375e-23], [16.149, 3.5203e-23], [16.186, 3.5032e-23], [16.223, 3.4862e-23], [16.261, 3.4693e-23], [16.298, 3.4525e-23], [16.336, 3.4358e-23], [16.374, 3.4192e-23], [16.411, 3.4028e-23], [16.449, 3.3864e-23], [16.487, 3.3701e-23], [16.525, 3.3539e-23], [16.563, 3.3378e-23], [16.601, 3.3217e-23], [16.64, 3.3058e-23], [16.678, 3.29e-23], [16.716, 3.2743e-23], [16.755, 3.2586e-23], [16.794, 3.2431e-23], [16.832, 3.2276e-23], [16.871, 3.2122e-23], [16.91, 3.1969e-23], [16.949, 3.1818e-23], [16.988, 3.1666e-23], [17.027, 3.1516e-23], [17.067, 3.1367e-23], [17.106, 3.1218e-23], [17.145, 3.1071e-23], [17.185, 3.0924e-23], [17.225, 3.0778e-23], [17.264, 3.0633e-23], [17.304, 3.0488e-23], [17.344, 3.0345e-23], [17.384, 3.0202e-23], [17.424, 3.006e-23], [17.464, 2.9919e-23], [17.505, 2.9779e-23], [17.545, 2.964e-23], [17.585, 2.9501e-23], [17.626, 2.9363e-23], [17.667, 2.9226e-23], [17.707, 2.9089e-23], [17.748, 2.8954e-23], [17.789, 2.8819e-23], [17.83, 2.8685e-23], [17.871, 2.8551e-23], [17.912, 2.8419e-23], [17.954, 2.8287e-23], [17.995, 2.8156e-23], [18.037, 2.8025e-23], [18.078, 2.7896e-23], [18.12, 2.7767e-23], [18.162, 2.7638e-23], [18.204, 2.7511e-23], [18.246, 2.7384e-23], [18.288, 2.7258e-23], [18.33, 2.7132e-23], [18.372, 2.7007e-23], [18.414, 2.6883e-23], [18.457, 2.676e-23], [18.499, 2.6637e-23], [18.542, 2.6515e-23], [18.585, 2.6394e-23], [18.628, 2.6273e-23], [18.671, 2.6153e-23], [18.714, 2.6034e-23], [18.757, 2.5915e-23], [18.8, 2.5797e-23], [18.844, 2.5679e-23], [18.887, 2.5562e-23], [18.931, 2.5446e-23], [18.974, 2.5331e-23], [19.018, 2.5216e-23], [19.062, 2.5101e-23], [19.106, 2.4988e-23], [19.15, 2.4875e-23], [19.194, 2.4762e-23], [19.238, 2.465e-23], [19.283, 2.4539e-23], [19.327, 2.4428e-23], [19.372, 2.4318e-23], [19.416, 2.4209e-23], [19.461, 2.41e-23], [19.506, 2.3992e-23], [19.551, 2.3884e-23], [19.596, 2.3777e-23], [19.641, 2.367e-23], [19.686, 2.3564e-23], [19.732, 2.3459e-23], [19.777, 2.3354e-23], [19.823, 2.325e-23], [19.869, 2.3146e-23], [19.915, 2.3043e-23], [19.96, 2.294e-23], [20.006, 2.2838e-23], [20.053, 2.2737e-23], [20.099, 2.2636e-23], [20.145, 2.2535e-23], [20.192, 2.2435e-23], [20.238, 2.2336e-23], [20.285, 2.2237e-23], [20.332, 2.2139e-23], [20.379, 2.2041e-23], [20.426, 2.1943e-23], [20.473, 2.1847e-23], [20.52, 2.175e-23], [20.567, 2.1655e-23], [20.615, 2.1559e-23], [20.662, 2.1465e-23], [20.71, 2.1371e-23], [20.758, 2.1277e-23], [20.805, 2.1184e-23], [20.853, 2.1091e-23], [20.901, 2.0999e-23], [20.95, 2.0907e-23], [20.998, 2.0816e-23], [21.046, 2.0725e-23], [21.095, 2.0634e-23], [21.144, 2.0545e-23], [21.192, 2.0455e-23], [21.241, 2.0366e-23], [21.29, 2.0278e-23], [21.339, 2.019e-23], [21.389, 2.0103e-23], [21.438, 2.0015e-23], [21.487, 1.9929e-23], [21.537, 1.9843e-23], [21.586, 1.9757e-23], [21.636, 1.9672e-23], [21.686, 1.9587e-23], [21.736, 1.9503e-23], [21.786, 1.9419e-23], [21.837, 1.9336e-23], [21.887, 1.9253e-23], [21.937, 1.917e-23], [21.988, 1.9088e-23], [22.039, 1.9006e-23], [22.089, 1.8925e-23], [22.14, 1.8844e-23], [22.191, 1.8764e-23], [22.243, 1.8684e-23], [22.294, 1.8604e-23], [22.345, 1.8525e-23], [22.397, 1.8446e-23], [22.449, 1.8368e-23], [22.5, 1.829e-23], [22.552, 1.8213e-23], [22.604, 1.8136e-23], [22.656, 1.8059e-23], [22.709, 1.7983e-23], [22.761, 1.7907e-23], [22.813, 1.7831e-23], [22.866, 1.7756e-23], [22.919, 1.7681e-23], [22.972, 1.7607e-23], [23.025, 1.7533e-23], [23.078, 1.746e-23], [23.131, 1.7386e-23], [23.184, 1.7314e-23], [23.238, 1.7241e-23], [23.291, 1.7169e-23], [23.345, 1.7098e-23], [23.399, 1.7026e-23], [23.453, 1.6956e-23], [23.507, 1.6885e-23], [23.561, 1.6815e-23], [23.615, 1.6745e-23], [23.67, 1.6676e-23], [23.724, 1.6607e-23], [23.779, 1.6538e-23], [23.834, 1.647e-23], [23.889, 1.6402e-23], [23.944, 1.6334e-23], [23.999, 1.6267e-23], [24.055, 1.62e-23], [24.11, 1.6133e-23], [24.166, 1.6067e-23], [24.221, 1.6001e-23], [24.277, 1.5936e-23], [24.333, 1.5871e-23], [24.389, 1.5806e-23], [24.446, 1.5741e-23], [24.502, 1.5677e-23], [24.558, 1.5613e-23], [24.615, 1.555e-23], [24.672, 1.5487e-23], [24.729, 1.5424e-23], [24.786, 1.5362e-23], [24.843, 1.5299e-23], [24.9, 1.5238e-23], [24.958, 1.5176e-23], [25.015, 1.5115e-23], [25.073, 1.5054e-23], [25.131, 1.4993e-23], [25.189, 1.4933e-23], [25.247, 1.4873e-23], [25.305, 1.4814e-23], [25.363, 1.4754e-23], [25.422, 1.4695e-23], [25.48, 1.4637e-23], [25.539, 1.4578e-23], [25.598, 1.452e-23], [25.657, 1.4463e-23], [25.716, 1.4405e-23], [25.776, 1.4348e-23], [25.835, 1.4291e-23], [25.895, 1.4235e-23], [25.954, 1.4178e-23], [26.014, 1.4122e-23], [26.074, 1.4067e-23], [26.134, 1.4011e-23], [26.194, 1.3956e-23], [26.255, 1.3901e-23], [26.315, 1.3847e-23], [26.376, 1.3793e-23], [26.437, 1.3739e-23], [26.498, 1.3685e-23], [26.559, 1.3632e-23], [26.62, 1.3579e-23], [26.682, 1.3526e-23], [26.743, 1.3473e-23], [26.805, 1.3421e-23], [26.867, 1.3369e-23], [26.929, 1.3317e-23], [26.991, 1.3266e-23], [27.053, 1.3215e-23], [27.115, 1.3164e-23], [27.178, 1.3113e-23], [27.241, 1.3063e-23], [27.303, 1.3013e-23], [27.366, 1.2963e-23], [27.429, 1.2913e-23], [27.493, 1.2864e-23], [27.556, 1.2815e-23], [27.62, 1.2766e-23], [27.683, 1.2718e-23], [27.747, 1.2669e-23], [27.811, 1.2621e-23], [27.875, 1.2574e-23], [27.94, 1.2526e-23], [28.004, 1.2479e-23], [28.069, 1.2432e-23], [28.133, 1.2385e-23], [28.198, 1.2338e-23], [28.263, 1.2292e-23], [28.328, 1.2246e-23], [28.394, 1.22e-23], [28.459, 1.2155e-23], [28.525, 1.2109e-23], [28.591, 1.2064e-23], [28.656, 1.202e-23], [28.723, 1.1975e-23], [28.789, 1.1931e-23], [28.855, 1.1887e-23], [28.922, 1.1843e-23], [28.988, 1.1799e-23], [29.055, 1.1756e-23], [29.122, 1.1712e-23], [29.189, 1.1669e-23], [29.257, 1.1627e-23], [29.324, 1.1584e-23], [29.392, 1.1542e-23], [29.46, 1.15e-23], [29.528, 1.1458e-23], [29.596, 1.1416e-23], [29.664, 1.1375e-23], [29.732, 1.1334e-23], [29.801, 1.1293e-23], [29.87, 1.1252e-23], [29.938, 1.1212e-23], [30.007, 1.1171e-23], [30.077, 1.1131e-23], [30.146, 1.1091e-23], [30.216, 1.1052e-23], [30.285, 1.1012e-23], [30.355, 1.0973e-23], [30.425, 1.0934e-23], [30.495, 1.0895e-23], [30.566, 1.0857e-23], [30.636, 1.0818e-23], [30.707, 1.078e-23], [30.777, 1.0742e-23], [30.848, 1.0704e-23], [30.92, 1.0667e-23], [30.991, 1.0629e-23], [31.062, 1.0592e-23], [31.134, 1.0555e-23], [31.206, 1.0518e-23], [31.278, 1.0482e-23], [31.35, 1.0445e-23], [31.422, 1.0409e-23], [31.495, 1.0373e-23], [31.567, 1.0337e-23], [31.64, 1.0302e-23], [31.713, 1.0266e-23], [31.786, 1.0231e-23], [31.859, 1.0196e-23], [31.933, 1.0161e-23], [32.007, 1.0126e-23], [32.08, 1.0092e-23], [32.154, 1.0057e-23], [32.228, 1.0023e-23], [32.303, 9.9894e-24], [32.377, 9.9556e-24], [32.452, 9.9221e-24], [32.527, 9.8887e-24], [32.602, 9.8554e-24], [32.677, 9.8224e-24], [32.752, 9.7896e-24], [32.828, 9.7569e-24], [32.904, 9.7244e-24], [32.979, 9.692e-24], [33.055, 9.6599e-24], [33.132, 9.6279e-24], [33.208, 9.5961e-24], [33.285, 9.5644e-24], [33.361, 9.533e-24], [33.438, 9.5017e-24], [33.515, 9.4705e-24], [33.593, 9.4396e-24], [33.67, 9.4088e-24], [33.748, 9.3782e-24], [33.826, 9.3477e-24], [33.904, 9.3174e-24], [33.982, 9.2873e-24], [34.06, 9.2573e-24], [34.139, 9.2275e-24], [34.218, 9.1979e-24], [34.296, 9.1684e-24], [34.376, 9.139e-24], [34.455, 9.1099e-24], [34.534, 9.0809e-24], [34.614, 9.052e-24], [34.694, 9.0233e-24], [34.774, 8.9948e-24], [34.854, 8.9664e-24], [34.934, 8.9382e-24], [35.015, 8.9101e-24], [35.096, 8.8822e-24], [35.176, 8.8544e-24], [35.258, 8.8268e-24], [35.339, 8.7993e-24], [35.42, 8.772e-24], [35.502, 8.7448e-24], [35.584, 8.7178e-24], [35.666, 8.6909e-24], [35.748, 8.6642e-24], [35.831, 8.6376e-24], [35.913, 8.6112e-24], [35.996, 8.5849e-24], [36.079, 8.5587e-24], [36.162, 8.5327e-24], [36.246, 8.5069e-24], [36.329, 8.4811e-24], [36.413, 8.4556e-24], [36.497, 8.4301e-24], [36.581, 8.4048e-24], [36.666, 8.3796e-24], [36.75, 8.3546e-24], [36.835, 8.3297e-24], [36.92, 8.3049e-24], [37.005, 8.2803e-24], [37.09, 8.2558e-24], [37.176, 8.2315e-24], [37.262, 8.2073e-24], [37.347, 8.1832e-24], [37.434, 8.1592e-24], [37.52, 8.1354e-24], [37.606, 8.1117e-24], [37.693, 8.0881e-24], [37.78, 8.0647e-24], [37.867, 8.0414e-24], [37.955, 8.0182e-24], [38.042, 7.9952e-24], [38.13, 7.9722e-24], [38.218, 7.9494e-24], [38.306, 7.9268e-24], [38.394, 7.9042e-24], [38.483, 7.8818e-24], [38.571, 7.8595e-24], [38.66, 7.8373e-24], [38.75, 7.8152e-24], [38.839, 7.7933e-24], [38.928, 7.7715e-24], [39.018, 7.7498e-24], [39.108, 7.7282e-24], [39.198, 7.7067e-24], [39.289, 7.6854e-24], [39.379, 7.6641e-24], [39.47, 7.643e-24], [39.561, 7.622e-24], [39.652, 7.6012e-24], [39.744, 7.5804e-24], [39.836, 7.5597e-24], [39.927, 7.5392e-24], [40.019, 7.5188e-24], [40.112, 7.4985e-24], [40.204, 7.4783e-24], [40.297, 7.4582e-24], [40.39, 7.4382e-24], [40.483, 7.4183e-24], [40.576, 7.3986e-24], [40.67, 7.3789e-24], [40.764, 7.3594e-24], [40.858, 7.3399e-24], [40.952, 7.3206e-24], [41.046, 7.3014e-24], [41.141, 7.2822e-24], [41.236, 7.2632e-24], [41.331, 7.2443e-24], [41.426, 7.2255e-24], [41.522, 7.2068e-24], [41.618, 7.1882e-24], [41.714, 7.1697e-24], [41.81, 7.1513e-24], [41.906, 7.133e-24], [42.003, 7.1148e-24], [42.1, 7.0967e-24], [42.197, 7.0787e-24], [42.294, 7.0608e-24], [42.392, 7.043e-24], [42.489, 7.0253e-24], [42.587, 7.0077e-24], [42.686, 6.9902e-24], [42.784, 6.9728e-24], [42.883, 6.9555e-24], [42.981, 6.9383e-24], [43.081, 6.9212e-24], [43.18, 6.9041e-24], [43.28, 6.8872e-24], [43.379, 6.8703e-24], [43.479, 6.8536e-24], [43.58, 6.8369e-24], [43.68, 6.8203e-24], [43.781, 6.8039e-24], [43.882, 6.7875e-24], [43.983, 6.7712e-24], [44.084, 6.755e-24], [44.186, 6.7388e-24], [44.288, 6.7228e-24], [44.39, 6.7069e-24], [44.492, 6.691e-24], [44.595, 6.6752e-24], [44.698, 6.6595e-24], [44.801, 6.6439e-24], [44.904, 6.6284e-24], [45.008, 6.613e-24], [45.112, 6.5977e-24], [45.216, 6.5824e-24], [45.32, 6.5672e-24], [45.424, 6.5521e-24], [45.529, 6.5371e-24], [45.634, 6.5222e-24], [45.739, 6.5073e-24], [45.845, 6.4925e-24], [45.951, 6.4779e-24], [46.057, 6.4632e-24], [46.163, 6.4487e-24], [46.269, 6.4343e-24], [46.376, 6.4199e-24], [46.483, 6.4056e-24], [46.59, 6.3914e-24], [46.697, 6.3772e-24], [46.805, 6.3632e-24], [46.913, 6.3492e-24], [47.021, 6.3353e-24], [47.13, 6.3214e-24], [47.238, 6.3077e-24], [47.347, 6.294e-24], [47.457, 6.2804e-24], [47.566, 6.2668e-24], [47.676, 6.2534e-24], [47.786, 6.24e-24], [47.896, 6.2267e-24], [48.006, 6.2134e-24], [48.117, 6.2002e-24], [48.228, 6.1871e-24], [48.339, 6.1741e-24], [48.451, 6.1611e-24], [48.562, 6.1482e-24], [48.674, 6.1354e-24], [48.787, 6.1227e-24], [48.899, 6.11e-24], [49.012, 6.0974e-24], [49.125, 6.0848e-24], [49.238, 6.0723e-24], [49.352, 6.0599e-24], [49.465, 6.0476e-24], [49.579, 6.0353e-24], [49.694, 6.023e-24], [49.808, 6.0109e-24], [49.923, 5.9988e-24], [50.038, 5.9868e-24], [50.154, 5.9748e-24], [50.269, 5.9629e-24], [50.385, 5.9511e-24], [50.502, 5.9393e-24], [50.618, 5.9276e-24], [50.735, 5.916e-24], [50.852, 5.9044e-24], [50.969, 5.8929e-24], [51.087, 5.8814e-24], [51.204, 5.87e-24], [51.322, 5.8587e-24], [51.441, 5.8474e-24], [51.559, 5.8362e-24], [51.678, 5.825e-24], [51.797, 5.8139e-24], [51.917, 5.8029e-24], [52.037, 5.7919e-24], [52.157, 5.7809e-24], [52.277, 5.7701e-24], [52.397, 5.7593e-24], [52.518, 5.7485e-24], [52.639, 5.7378e-24], [52.761, 5.7272e-24], [52.882, 5.7166e-24], [53.004, 5.706e-24], [53.127, 5.6956e-24], [53.249, 5.6851e-24], [53.372, 5.6748e-24], [53.495, 5.6645e-24], [53.618, 5.6542e-24], [53.742, 5.644e-24], [53.866, 5.6338e-24], [53.99, 5.6237e-24], [54.115, 5.6137e-24], [54.239, 5.6037e-24], [54.365, 5.5938e-24], [54.49, 5.5839e-24], [54.616, 5.574e-24], [54.741, 5.5642e-24], [54.868, 5.5545e-24], [54.994, 5.5448e-24], [55.121, 5.5352e-24], [55.248, 5.5256e-24], [55.376, 5.516e-24], [55.503, 5.5065e-24], [55.631, 5.4971e-24], [55.76, 5.4877e-24], [55.888, 5.4783e-24], [56.017, 5.469e-24], [56.146, 5.4598e-24], [56.276, 5.4506e-24], [56.405, 5.4414e-24], [56.535, 5.4323e-24], [56.666, 5.4232e-24], [56.797, 5.4142e-24], [56.928, 5.4052e-24], [57.059, 5.3963e-24], [57.19, 5.3874e-24], [57.322, 5.3786e-24], [57.454, 5.3698e-24], [57.587, 5.361e-24], [57.72, 5.3523e-24], [57.853, 5.3437e-24], [57.986, 5.3351e-24], [58.12, 5.3265e-24], [58.254, 5.3179e-24], [58.388, 5.3095e-24], [58.523, 5.301e-24], [58.658, 5.2926e-24], [58.793, 5.2842e-24], [58.929, 5.2759e-24], [59.065, 5.2676e-24], [59.201, 5.2594e-24], [59.337, 5.2512e-24], [59.474, 5.243e-24], [59.611, 5.2349e-24], [59.749, 5.2268e-24], [59.887, 5.2188e-24], [60.025, 5.2108e-24], [60.163, 5.2028e-24], [60.302, 5.1949e-24], [60.441, 5.187e-24], [60.58, 5.1792e-24], [60.72, 5.1714e-24], [60.86, 5.1636e-24], [61.0, 5.1559e-24], [61.141, 5.1482e-24], [61.282, 5.1405e-24], [61.423, 5.1329e-24], [61.565, 5.1253e-24], [61.707, 5.1178e-24], [61.849, 5.1103e-24], [61.992, 5.1028e-24], [62.135, 5.0953e-24], [62.278, 5.0879e-24], [62.422, 5.0806e-24], [62.566, 5.0732e-24], [62.71, 5.0659e-24], [62.855, 5.0587e-24], [62.999, 5.0515e-24], [63.145, 5.0443e-24], [63.29, 5.0371e-24], [63.436, 5.03e-24], [63.583, 5.0229e-24], [63.729, 5.0158e-24], [63.876, 5.0088e-24], [64.023, 5.0018e-24], [64.171, 4.9949e-24], [64.319, 4.9879e-24], [64.467, 4.981e-24], [64.616, 4.9742e-24], [64.765, 4.9673e-24], [64.914, 4.9605e-24], [65.064, 4.9538e-24], [65.214, 4.947e-24], [65.365, 4.9403e-24], [65.515, 4.9337e-24], [65.666, 4.927e-24], [65.818, 4.9204e-24], [65.97, 4.9138e-24], [66.122, 4.9073e-24], [66.274, 4.9007e-24], [66.427, 4.8943e-24], [66.58, 4.8878e-24], [66.734, 4.8814e-24], [66.888, 4.875e-24], [67.042, 4.8686e-24], [67.196, 4.8622e-24], [67.351, 4.8559e-24], [67.507, 4.8496e-24], [67.662, 4.8434e-24], [67.818, 4.8371e-24], [67.975, 4.8309e-24], [68.132, 4.8248e-24], [68.289, 4.8186e-24], [68.446, 4.8125e-24], [68.604, 4.8064e-24], [68.762, 4.8003e-24], [68.921, 4.7943e-24], [69.08, 4.7883e-24], [69.239, 4.7823e-24], [69.399, 4.7763e-24], [69.559, 4.7704e-24], [69.719, 4.7645e-24], [69.88, 4.7586e-24], [70.041, 4.7527e-24], [70.202, 4.7469e-24], [70.364, 4.7411e-24], [70.527, 4.7353e-24], [70.689, 4.7296e-24], [70.852, 4.7238e-24], [71.016, 4.7181e-24], [71.179, 4.7124e-24], [71.344, 4.7068e-24], [71.508, 4.7012e-24], [71.673, 4.6955e-24], [71.838, 4.69e-24], [72.004, 4.6844e-24], [72.17, 4.6789e-24], [72.336, 4.6734e-24], [72.503, 4.6679e-24], [72.67, 4.6624e-24], [72.838, 4.657e-24], [73.006, 4.6515e-24], [73.174, 4.6461e-24], [73.343, 4.6408e-24], [73.512, 4.6354e-24], [73.682, 4.6301e-24], [73.852, 4.6248e-24], [74.022, 4.6195e-24], [74.193, 4.6142e-24], [74.364, 4.609e-24], [74.535, 4.6038e-24], [74.707, 4.5986e-24], [74.879, 4.5934e-24], [75.052, 4.5882e-24], [75.225, 4.5831e-24], [75.399, 4.578e-24], [75.572, 4.5729e-24], [75.747, 4.5678e-24], [75.921, 4.5628e-24], [76.096, 4.5577e-24], [76.272, 4.5527e-24], [76.448, 4.5477e-24], [76.624, 4.5427e-24], [76.801, 4.5378e-24], [76.978, 4.5329e-24], [77.155, 4.528e-24], [77.333, 4.5231e-24], [77.512, 4.5182e-24], [77.69, 4.5133e-24], [77.87, 4.5085e-24], [78.049, 4.5037e-24], [78.229, 4.4989e-24], [78.409, 4.4941e-24], [78.59, 4.4894e-24], [78.772, 4.4846e-24], [78.953, 4.4799e-24], [79.135, 4.4752e-24], [79.318, 4.4705e-24], [79.501, 4.4659e-24], [79.684, 4.4612e-24], [79.868, 4.4566e-24], [80.052, 4.452e-24], [80.236, 4.4474e-24], [80.421, 4.4428e-24], [80.607, 4.4382e-24], [80.793, 4.4337e-24], [80.979, 4.4292e-24], [81.166, 4.4247e-24], [81.353, 4.4202e-24], [81.541, 4.4157e-24], [81.729, 4.4113e-24], [81.917, 4.4068e-24], [82.106, 4.4024e-24], [82.295, 4.398e-24], [82.485, 4.3936e-24], [82.675, 4.3893e-24], [82.866, 4.3849e-24], [83.057, 4.3806e-24], [83.249, 4.3762e-24], [83.441, 4.3719e-24], [83.633, 4.3676e-24], [83.826, 4.3634e-24], [84.019, 4.3591e-24], [84.213, 4.3549e-24], [84.407, 4.3506e-24], [84.602, 4.3464e-24], [84.797, 4.3422e-24], [84.992, 4.3381e-24], [85.188, 4.3339e-24], [85.385, 4.3297e-24], [85.582, 4.3256e-24], [85.779, 4.3215e-24], [85.977, 4.3174e-24], [86.175, 4.3133e-24], [86.374, 4.3092e-24], [86.573, 4.3052e-24], [86.773, 4.3011e-24], [86.973, 4.2971e-24], [87.173, 4.2931e-24], [87.374, 4.2891e-24], [87.576, 4.2851e-24], [87.778, 4.2811e-24], [87.98, 4.2771e-24], [88.183, 4.2732e-24], [88.386, 4.2693e-24], [88.59, 4.2654e-24], [88.795, 4.2614e-24], [88.999, 4.2576e-24], [89.205, 4.2537e-24], [89.41, 4.2498e-24], [89.616, 4.246e-24], [89.823, 4.2421e-24], [90.03, 4.2383e-24], [90.238, 4.2345e-24], [90.446, 4.2307e-24], [90.655, 4.2269e-24], [90.864, 4.2232e-24], [91.073, 4.2194e-24], [91.283, 4.2157e-24], [91.494, 4.2119e-24], [91.705, 4.2082e-24], [91.916, 4.2045e-24], [92.128, 4.2008e-24], [92.34, 4.1971e-24], [92.553, 4.1935e-24], [92.767, 4.1898e-24], [92.981, 4.1862e-24], [93.195, 4.1825e-24], [93.41, 4.1789e-24], [93.625, 4.1753e-24], [93.841, 4.1717e-24], [94.058, 4.1681e-24], [94.275, 4.1646e-24], [94.492, 4.161e-24], [94.71, 4.1574e-24], [94.928, 4.1539e-24], [95.147, 4.1504e-24], [95.367, 4.1469e-24], [95.587, 4.1434e-24], [95.807, 4.1399e-24], [96.028, 4.1364e-24], [96.249, 4.1329e-24], [96.471, 4.1295e-24], [96.694, 4.1261e-24], [96.917, 4.1226e-24], [97.14, 4.1192e-24], [97.364, 4.1158e-24], [97.589, 4.1124e-24], [97.814, 4.109e-24], [98.039, 4.1056e-24], [98.266, 4.1023e-24], [98.492, 4.0989e-24], [98.719, 4.0956e-24], [98.947, 4.0922e-24], [99.175, 4.0889e-24], [99.404, 4.0856e-24], [99.633, 4.0823e-24], [99.863, 4.079e-24], [100.09, 4.0757e-24], [100.32, 4.0725e-24], [100.56, 4.0692e-24], [100.79, 4.066e-24], [101.02, 4.0627e-24], [101.25, 4.0595e-24], [101.49, 4.0563e-24], [101.72, 4.0531e-24], [101.95, 4.0499e-24], [102.19, 4.0467e-24], [102.43, 4.0435e-24], [102.66, 4.0403e-24], [102.9, 4.0372e-24], [103.14, 4.034e-24], [103.37, 4.0309e-24], [103.61, 4.0278e-24], [103.85, 4.0247e-24], [104.09, 4.0216e-24], [104.33, 4.0185e-24], [104.57, 4.0154e-24], [104.81, 4.0123e-24], [105.05, 4.0092e-24], [105.3, 4.0062e-24], [105.54, 4.0031e-24], [105.78, 4.0001e-24], [106.03, 3.997e-24], [106.27, 3.994e-24], [106.52, 3.991e-24], [106.76, 3.988e-24], [107.01, 3.985e-24], [107.25, 3.982e-24], [107.5, 3.979e-24], [107.75, 3.9761e-24], [108.0, 3.9731e-24], [108.25, 3.9702e-24], [108.5, 3.9672e-24], [108.75, 3.9643e-24], [109.0, 3.9614e-24], [109.25, 3.9585e-24], [109.5, 3.9556e-24], [109.75, 3.9527e-24], [110.01, 3.9498e-24], [110.26, 3.9469e-24], [110.51, 3.944e-24], [110.77, 3.9412e-24], [111.02, 3.9383e-24], [111.28, 3.9355e-24], [111.54, 3.9327e-24], [111.79, 3.9298e-24], [112.05, 3.927e-24], [112.31, 3.9242e-24], [112.57, 3.9214e-24], [112.83, 3.9186e-24], [113.09, 3.9158e-24], [113.35, 3.913e-24], [113.61, 3.9103e-24], [113.87, 3.9075e-24], [114.14, 3.9048e-24], [114.4, 3.902e-24], [114.66, 3.8993e-24], [114.93, 3.8966e-24], [115.19, 3.8939e-24], [115.46, 3.8911e-24], [115.72, 3.8884e-24], [115.99, 3.8857e-24], [116.26, 3.8831e-24], [116.53, 3.8804e-24], [116.8, 3.8777e-24], [117.06, 3.8751e-24], [117.33, 3.8724e-24], [117.61, 3.8698e-24], [117.88, 3.8671e-24], [118.15, 3.8645e-24], [118.42, 3.8619e-24], [118.69, 3.8593e-24], [118.97, 3.8566e-24], [119.24, 3.854e-24], [119.52, 3.8515e-24], [119.79, 3.8489e-24], [120.07, 3.8463e-24], [120.35, 3.8437e-24], [120.62, 3.8412e-24], [120.9, 3.8386e-24], [121.18, 3.8361e-24], [121.46, 3.8335e-24], [121.74, 3.831e-24], [122.02, 3.8285e-24], [122.3, 3.826e-24], [122.58, 3.8235e-24], [122.87, 3.821e-24], [123.15, 3.8185e-24], [123.43, 3.816e-24], [123.72, 3.8135e-24], [124.0, 3.811e-24], [124.29, 3.8086e-24], [124.58, 3.8061e-24], [124.86, 3.8037e-24], [125.15, 3.8012e-24], [125.44, 3.7988e-24], [125.73, 3.7964e-24], [126.02, 3.7939e-24], [126.31, 3.7915e-24], [126.6, 3.7891e-24], [126.89, 3.7867e-24], [127.19, 3.7843e-24], [127.48, 3.7819e-24], [127.77, 3.7796e-24], [128.07, 3.7772e-24], [128.36, 3.7748e-24], [128.66, 3.7725e-24], [128.96, 3.7701e-24], [129.25, 3.7678e-24], [129.55, 3.7654e-24], [129.85, 3.7631e-24], [130.15, 3.7608e-24], [130.45, 3.7585e-24], [130.75, 3.7562e-24], [131.05, 3.7539e-24], [131.35, 3.7516e-24], [131.66, 3.7493e-24], [131.96, 3.747e-24], [132.26, 3.7447e-24], [132.57, 3.7425e-24], [132.88, 3.7402e-24], [133.18, 3.738e-24], [133.49, 3.7357e-24], [133.8, 3.7335e-24], [134.11, 3.7312e-24], [134.41, 3.729e-24], [134.72, 3.7268e-24], [135.04, 3.7246e-24], [135.35, 3.7224e-24], [135.66, 3.7202e-24], [135.97, 3.718e-24], [136.29, 3.7158e-24], [136.6, 3.7136e-24], [136.91, 3.7114e-24], [137.23, 3.7093e-24], [137.55, 3.7071e-24], [137.86, 3.705e-24], [138.18, 3.7028e-24], [138.5, 3.7007e-24], [138.82, 3.6985e-24], [139.14, 3.6964e-24], [139.46, 3.6943e-24], [139.78, 3.6922e-24], [140.1, 3.6901e-24], [140.43, 3.688e-24], [140.75, 3.6859e-24], [141.08, 3.6838e-24], [141.4, 3.6817e-24], [141.73, 3.6796e-24], [142.05, 3.6776e-24], [142.38, 3.6755e-24], [142.71, 3.6735e-24], [143.04, 3.6714e-24], [143.37, 3.6694e-24], [143.7, 3.6673e-24], [144.03, 3.6653e-24], [144.36, 3.6633e-24], [144.7, 3.6613e-24], [145.03, 3.6592e-24], [145.36, 3.6572e-24], [145.7, 3.6552e-24], [146.04, 3.6533e-24], [146.37, 3.6513e-24], [146.71, 3.6493e-24], [147.05, 3.6473e-24], [147.39, 3.6454e-24], [147.73, 3.6434e-24], [148.07, 3.6414e-24], [148.41, 3.6395e-24], [148.75, 3.6376e-24], [149.09, 3.6356e-24], [149.44, 3.6337e-24], [149.78, 3.6318e-24], [150.13, 3.6299e-24], [150.47, 3.6279e-24], [150.82, 3.626e-24], [151.17, 3.6241e-24], [151.52, 3.6222e-24], [151.87, 3.6204e-24], [152.22, 3.6185e-24], [152.57, 3.6166e-24], [152.92, 3.6147e-24], [153.27, 3.6129e-24], [153.63, 3.611e-24], [153.98, 3.6092e-24], [154.34, 3.6073e-24], [154.69, 3.6055e-24], [155.05, 3.6037e-24], [155.41, 3.6019e-24], [155.76, 3.6e-24], [156.12, 3.5982e-24], [156.48, 3.5964e-24], [156.84, 3.5946e-24], [157.21, 3.5928e-24], [157.57, 3.5911e-24], [157.93, 3.5893e-24], [158.3, 3.5875e-24], [158.66, 3.5857e-24], [159.03, 3.584e-24], [159.39, 3.5822e-24], [159.76, 3.5805e-24], [160.13, 3.5787e-24], [160.5, 3.577e-24], [160.87, 3.5753e-24], [161.24, 3.5735e-24], [161.61, 3.5718e-24], [161.98, 3.5701e-24], [162.36, 3.5684e-24], [162.73, 3.5667e-24], [163.11, 3.565e-24], [163.48, 3.5633e-24], [163.86, 3.5616e-24], [164.24, 3.56e-24], [164.62, 3.5583e-24], [165.0, 3.5566e-24], [165.38, 3.555e-24], [165.76, 3.5533e-24], [166.14, 3.5517e-24], [166.52, 3.55e-24], [166.91, 3.5484e-24], [167.29, 3.5468e-24], [167.68, 3.5452e-24], [168.07, 3.5436e-24], [168.45, 3.5419e-24], [168.84, 3.5403e-24], [169.23, 3.5388e-24], [169.62, 3.5372e-24], [170.01, 3.5356e-24], [170.4, 3.534e-24], [170.8, 3.5324e-24], [171.19, 3.5309e-24], [171.59, 3.5293e-24], [171.98, 3.5278e-24], [172.38, 3.5262e-24], [172.78, 3.5247e-24], [173.17, 3.5231e-24], [173.57, 3.5216e-24], [173.97, 3.5201e-24], [174.37, 3.5186e-24], [174.78, 3.5171e-24], [175.18, 3.5156e-24], [175.58, 3.5141e-24], [175.99, 3.5126e-24], [176.39, 3.5111e-24], [176.8, 3.5096e-24], [177.21, 3.5081e-24], [177.62, 3.5067e-24], [178.03, 3.5052e-24], [178.44, 3.5038e-24], [178.85, 3.5023e-24], [179.26, 3.5009e-24], [179.68, 3.4994e-24], [180.09, 3.498e-24], [180.5, 3.4966e-24], [180.92, 3.4952e-24], [181.34, 3.4938e-24], [181.76, 3.4924e-24], [182.18, 3.491e-24], [182.6, 3.4896e-24], [183.02, 3.4882e-24], [183.44, 3.4868e-24], [183.86, 3.4854e-24], [184.29, 3.4841e-24], [184.71, 3.4827e-24], [185.14, 3.4813e-24], [185.56, 3.48e-24], [185.99, 3.4787e-24], [186.42, 3.4773e-24], [186.85, 3.476e-24], [187.28, 3.4747e-24], [187.71, 3.4733e-24], [188.15, 3.472e-24], [188.58, 3.4707e-24], [189.01, 3.4694e-24], [189.45, 3.4681e-24], [189.89, 3.4669e-24], [190.33, 3.4656e-24], [190.76, 3.4643e-24], [191.2, 3.463e-24], [191.65, 3.4618e-24], [192.09, 3.4605e-24], [192.53, 3.4593e-24], [192.97, 3.458e-24], [193.42, 3.4568e-24], [193.87, 3.4556e-24], [194.31, 3.4544e-24], [194.76, 3.4531e-24], [195.21, 3.4519e-24], [195.66, 3.4507e-24], [196.11, 3.4495e-24], [196.56, 3.4483e-24], [197.02, 3.4472e-24], [197.47, 3.446e-24], [197.93, 3.4448e-24], [198.38, 3.4436e-24], [198.84, 3.4425e-24], [199.3, 3.4413e-24], [199.76, 3.4402e-24], [200.22, 3.439e-24], [200.68, 3.4379e-24], [201.14, 3.4368e-24], [201.61, 3.4357e-24], [202.07, 3.4346e-24], [202.54, 3.4335e-24], [203.0, 3.4324e-24], [203.47, 3.4313e-24], [203.94, 3.4302e-24], [204.41, 3.4291e-24], [204.88, 3.428e-24], [205.36, 3.427e-24], [205.83, 3.4259e-24], [206.3, 3.4248e-24], [206.78, 3.4238e-24], [207.26, 3.4228e-24], [207.73, 3.4217e-24], [208.21, 3.4207e-24], [208.69, 3.4197e-24], [209.18, 3.4187e-24], [209.66, 3.4177e-24], [210.14, 3.4167e-24], [210.63, 3.4157e-24], [211.11, 3.4147e-24], [211.6, 3.4137e-24], [212.09, 3.4127e-24], [212.58, 3.4118e-24], [213.07, 3.4108e-24], [213.56, 3.4098e-24], [214.05, 3.4089e-24], [214.54, 3.408e-24], [215.04, 3.407e-24], [215.53, 3.4061e-24], [216.03, 3.4052e-24], [216.53, 3.4043e-24], [217.03, 3.4034e-24], [217.53, 3.4025e-24], [218.03, 3.4016e-24], [218.53, 3.4007e-24], [219.04, 3.3998e-24], [219.54, 3.3989e-24], [220.05, 3.3981e-24], [220.56, 3.3972e-24], [221.06, 3.3964e-24], [221.57, 3.3955e-24], [222.08, 3.3947e-24], [222.6, 3.3938e-24], [223.11, 3.393e-24], [223.62, 3.3922e-24], [224.14, 3.3914e-24], [224.66, 3.3906e-24], [225.18, 3.3898e-24], [225.69, 3.389e-24], [226.22, 3.3882e-24], [226.74, 3.3875e-24], [227.26, 3.3867e-24], [227.78, 3.3859e-24], [228.31, 3.3852e-24], [228.84, 3.3844e-24], [229.36, 3.3837e-24], [229.89, 3.383e-24], [230.42, 3.3822e-24], [230.95, 3.3815e-24], [231.49, 3.3808e-24], [232.02, 3.3801e-24], [232.56, 3.3794e-24], [233.09, 3.3787e-24], [233.63, 3.378e-24], [234.17, 3.3774e-24], [234.71, 3.3767e-24], [235.25, 3.376e-24], [235.79, 3.3754e-24], [236.34, 3.3747e-24], [236.88, 3.3741e-24], [237.43, 3.3735e-24], [237.97, 3.3729e-24], [238.52, 3.3722e-24], [239.07, 3.3716e-24], [239.62, 3.371e-24], [240.18, 3.3704e-24], [240.73, 3.3699e-24], [241.29, 3.3693e-24], [241.84, 3.3687e-24], [242.4, 3.3681e-24], [242.96, 3.3676e-24], [243.52, 3.367e-24], [244.08, 3.3665e-24], [244.64, 3.366e-24], [245.21, 3.3654e-24], [245.77, 3.3649e-24], [246.34, 3.3644e-24], [246.91, 3.3639e-24], [247.48, 3.3634e-24], [248.05, 3.3629e-24], [248.62, 3.3624e-24], [249.19, 3.362e-24], [249.77, 3.3615e-24], [250.34, 3.3611e-24], [250.92, 3.3606e-24], [251.5, 3.3602e-24], [252.08, 3.3597e-24], [252.66, 3.3593e-24], [253.24, 3.3589e-24], [253.83, 3.3585e-24], [254.41, 3.3581e-24], [255.0, 3.3577e-24], [255.59, 3.3573e-24], [256.18, 3.3569e-24], [256.77, 3.3565e-24], [257.36, 3.3562e-24], [257.95, 3.3558e-24], [258.55, 3.3555e-24], [259.14, 3.3551e-24], [259.74, 3.3548e-24], [260.34, 3.3545e-24], [260.94, 3.3542e-24], [261.54, 3.3539e-24], [262.15, 3.3536e-24], [262.75, 3.3533e-24], [263.36, 3.353e-24], [263.96, 3.3527e-24], [264.57, 3.3525e-24], [265.18, 3.3522e-24], [265.79, 3.352e-24], [266.41, 3.3517e-24], [267.02, 3.3515e-24], [267.64, 3.3513e-24], [268.25, 3.351e-24], [268.87, 3.3508e-24], [269.49, 3.3506e-24], [270.11, 3.3504e-24], [270.74, 3.3503e-24], [271.36, 3.3501e-24], [271.99, 3.3499e-24], [272.61, 3.3498e-24], [273.24, 3.3496e-24], [273.87, 3.3495e-24], [274.5, 3.3493e-24], [275.14, 3.3492e-24], [275.77, 3.3491e-24], [276.41, 3.349e-24], [277.05, 3.3489e-24], [277.68, 3.3488e-24], [278.32, 3.3487e-24], [278.97, 3.3486e-24], [279.61, 3.3486e-24], [280.25, 3.3485e-24], [280.9, 3.3485e-24], [281.55, 3.3484e-24], [282.2, 3.3484e-24], [282.85, 3.3484e-24], [283.5, 3.3484e-24], [284.15, 3.3484e-24], [284.81, 3.3484e-24], [285.47, 3.3484e-24], [286.13, 3.3484e-24], [286.78, 3.3485e-24], [287.45, 3.3485e-24], [288.11, 3.3485e-24], [288.77, 3.3486e-24], [289.44, 3.3487e-24], [290.11, 3.3488e-24], [290.78, 3.3488e-24], [291.45, 3.3489e-24], [292.12, 3.349e-24], [292.79, 3.3492e-24], [293.47, 3.3493e-24], [294.14, 3.3494e-24], [294.82, 3.3495e-24], [295.5, 3.3497e-24], [296.18, 3.3499e-24], [296.87, 3.35e-24], [297.55, 3.3502e-24], [298.24, 3.3504e-24], [298.92, 3.3506e-24], [299.61, 3.3508e-24], [300.31, 3.351e-24], [301.0, 3.3512e-24], [301.69, 3.3515e-24], [302.39, 3.3517e-24], [303.08, 3.352e-24], [303.78, 3.3522e-24], [304.48, 3.3525e-24], [305.19, 3.3528e-24], [305.89, 3.3531e-24], [306.6, 3.3534e-24], [307.3, 3.3537e-24], [308.01, 3.354e-24], [308.72, 3.3543e-24], [309.43, 3.3547e-24], [310.15, 3.355e-24], [310.86, 3.3554e-24], [311.58, 3.3557e-24], [312.3, 3.3561e-24], [313.02, 3.3565e-24], [313.74, 3.3569e-24], [314.46, 3.3573e-24], [315.19, 3.3577e-24], [315.91, 3.3581e-24], [316.64, 3.3586e-24], [317.37, 3.359e-24], [318.11, 3.3595e-24], [318.84, 3.36e-24], [319.57, 3.3604e-24], [320.31, 3.3609e-24], [321.05, 3.3614e-24], [321.79, 3.3619e-24], [322.53, 3.3624e-24], [323.28, 3.363e-24], [324.02, 3.3635e-24], [324.77, 3.3641e-24], [325.52, 3.3646e-24], [326.27, 3.3652e-24], [327.02, 3.3658e-24], [327.77, 3.3663e-24], [328.53, 3.3669e-24], [329.29, 3.3676e-24], [330.05, 3.3682e-24], [330.81, 3.3688e-24], [331.57, 3.3694e-24], [332.34, 3.3701e-24], [333.1, 3.3708e-24], [333.87, 3.3714e-24], [334.64, 3.3721e-24], [335.41, 3.3728e-24], [336.19, 3.3735e-24], [336.96, 3.3742e-24], [337.74, 3.3749e-24], [338.52, 3.3757e-24], [339.3, 3.3764e-24], [340.08, 3.3772e-24], [340.86, 3.378e-24], [341.65, 3.3787e-24], [342.44, 3.3795e-24], [343.23, 3.3803e-24], [344.02, 3.3811e-24], [344.81, 3.382e-24], [345.61, 3.3828e-24], [346.4, 3.3836e-24], [347.2, 3.3845e-24], [348.0, 3.3854e-24], [348.81, 3.3862e-24], [349.61, 3.3871e-24], [350.42, 3.388e-24], [351.23, 3.3889e-24], [352.04, 3.3899e-24], [352.85, 3.3908e-24], [353.66, 3.3917e-24], [354.48, 3.3927e-24], [355.29, 3.3937e-24], [356.11, 3.3946e-24], [356.93, 3.3956e-24], [357.76, 3.3966e-24], [358.58, 3.3976e-24], [359.41, 3.3987e-24], [360.24, 3.3997e-24], [361.07, 3.4007e-24], [361.9, 3.4018e-24], [362.74, 3.4029e-24], [363.57, 3.4039e-24], [364.41, 3.405e-24], [365.25, 3.4061e-24], [366.09, 3.4073e-24], [366.94, 3.4084e-24], [367.78, 3.4095e-24], [368.63, 3.4107e-24], [369.48, 3.4118e-24], [370.33, 3.413e-24], [371.19, 3.4142e-24], [372.04, 3.4154e-24], [372.9, 3.4166e-24], [373.76, 3.4178e-24], [374.62, 3.4191e-24], [375.49, 3.4203e-24], [376.35, 3.4216e-24], [377.22, 3.4228e-24], [378.09, 3.4241e-24], [378.96, 3.4254e-24], [379.84, 3.4267e-24], [380.71, 3.428e-24], [381.59, 3.4294e-24], [382.47, 3.4307e-24], [383.35, 3.4321e-24], [384.24, 3.4334e-24], [385.12, 3.4348e-24], [386.01, 3.4362e-24], [386.9, 3.4376e-24], [387.79, 3.439e-24], [388.69, 3.4405e-24], [389.58, 3.4419e-24], [390.48, 3.4434e-24], [391.38, 3.4448e-24], [392.28, 3.4463e-24], [393.19, 3.4478e-24], [394.1, 3.4493e-24], [395.0, 3.4508e-24], [395.92, 3.4524e-24], [396.83, 3.4539e-24], [397.74, 3.4555e-24], [398.66, 3.457e-24], [399.58, 3.4586e-24], [400.5, 3.4602e-24], [401.43, 3.4618e-24], [402.35, 3.4635e-24], [403.28, 3.4651e-24], [404.21, 3.4667e-24], [405.14, 3.4684e-24], [406.08, 3.4701e-24], [407.01, 3.4718e-24], [407.95, 3.4735e-24], [408.89, 3.4752e-24], [409.83, 3.4769e-24], [410.78, 3.4786e-24], [411.73, 3.4804e-24], [412.68, 3.4822e-24], [413.63, 3.4839e-24], [414.58, 3.4857e-24], [415.54, 3.4875e-24], [416.5, 3.4894e-24], [417.46, 3.4912e-24], [418.42, 3.493e-24], [419.38, 3.4949e-24], [420.35, 3.4968e-24], [421.32, 3.4987e-24], [422.29, 3.5006e-24], [423.27, 3.5025e-24], [424.24, 3.5044e-24], [425.22, 3.5063e-24], [426.2, 3.5083e-24], [427.18, 3.5103e-24], [428.17, 3.5123e-24], [429.16, 3.5143e-24], [430.15, 3.5163e-24], [431.14, 3.5183e-24], [432.13, 3.5203e-24], [433.13, 3.5224e-24], [434.13, 3.5245e-24], [435.13, 3.5265e-24], [436.13, 3.5286e-24], [437.14, 3.5307e-24], [438.14, 3.5329e-24], [439.16, 3.535e-24], [440.17, 3.5372e-24], [441.18, 3.5393e-24], [442.2, 3.5415e-24], [443.22, 3.5437e-24], [444.24, 3.5459e-24], [445.27, 3.5481e-24], [446.29, 3.5504e-24], [447.32, 3.5526e-24], [448.35, 3.5549e-24], [449.39, 3.5572e-24], [450.42, 3.5595e-24], [451.46, 3.5618e-24], [452.5, 3.5641e-24], [453.55, 3.5665e-24], [454.59, 3.5688e-24], [455.64, 3.5712e-24], [456.69, 3.5736e-24], [457.75, 3.576e-24], [458.8, 3.5784e-24], [459.86, 3.5809e-24], [460.92, 3.5833e-24], [461.98, 3.5858e-24], [463.05, 3.5882e-24], [464.12, 3.5907e-24], [465.19, 3.5933e-24], [466.26, 3.5958e-24], [467.33, 3.5983e-24], [468.41, 3.6009e-24], [469.49, 3.6035e-24], [470.57, 3.6061e-24], [471.66, 3.6087e-24], [472.75, 3.6113e-24], [473.84, 3.6139e-24], [474.93, 3.6166e-24], [476.02, 3.6193e-24], [477.12, 3.622e-24], [478.22, 3.6247e-24], [479.33, 3.6274e-24], [480.43, 3.6302e-24], [481.54, 3.633e-24], [482.65, 3.6358e-24], [483.76, 3.6386e-24], [484.88, 3.6415e-24], [486.0, 3.6445e-24], [487.12, 3.6475e-24], [488.24, 3.6506e-24], [489.37, 3.6538e-24], [490.49, 3.6574e-24], [491.63, 3.6615e-24], [492.76, 3.6668e-24], [493.9, 3.6763e-24], [495.03, 3.7073e-24], [496.18, 4.6937e-24], [497.32, 3.773e-24], [498.47, 3.6946e-24], [499.62, 3.6862e-24], [500.77, 3.6856e-24], [501.92, 3.6871e-24], [503.08, 3.6893e-24], [504.24, 3.6919e-24], [505.4, 3.6946e-24], [506.57, 3.6975e-24], [507.74, 3.7005e-24], [508.91, 3.7035e-24], [510.08, 3.7066e-24], [511.26, 3.7097e-24], [512.44, 3.7128e-24], [513.62, 3.7159e-24], [514.8, 3.7191e-24], [515.99, 3.7223e-24], [517.18, 3.7255e-24], [518.37, 3.7288e-24], [519.57, 3.732e-24], [520.77, 3.7353e-24], [521.97, 3.7386e-24], [523.17, 3.7419e-24], [524.38, 3.7452e-24], [525.59, 3.7486e-24], [526.8, 3.7519e-24], [528.01, 3.7553e-24], [529.23, 3.7587e-24], [530.45, 3.7622e-24], [531.67, 3.7656e-24], [532.9, 3.7691e-24], [534.13, 3.7725e-24], [535.36, 3.776e-24], [536.6, 3.7795e-24], [537.83, 3.7831e-24], [539.07, 3.7866e-24], [540.32, 3.7902e-24], [541.56, 3.7938e-24], [542.81, 3.7974e-24], [544.06, 3.801e-24], [545.32, 3.8046e-24], [546.57, 3.8083e-24], [547.84, 3.8119e-24], [549.1, 3.8156e-24], [550.36, 3.8193e-24], [551.63, 3.823e-24], [552.91, 3.8268e-24], [554.18, 3.8305e-24], [555.46, 3.8343e-24], [556.74, 3.8381e-24], [558.02, 3.8419e-24], [559.31, 3.8458e-24], [560.6, 3.8496e-24], [561.89, 3.8535e-24], [563.19, 3.8574e-24], [564.49, 3.8613e-24], [565.79, 3.8652e-24], [567.09, 3.8691e-24], [568.4, 3.8731e-24], [569.71, 3.8771e-24], [571.03, 3.8811e-24], [572.34, 3.8851e-24], [573.66, 3.8891e-24], [574.99, 3.8932e-24], [576.31, 3.8972e-24], [577.64, 3.9013e-24], [578.97, 3.9054e-24], [580.31, 3.9096e-24], [581.65, 3.9137e-24], [582.99, 3.9179e-24], [584.33, 3.9221e-24], [585.68, 3.9263e-24], [587.03, 3.9305e-24], [588.38, 3.9347e-24], [589.74, 3.939e-24], [591.1, 3.9433e-24], [592.46, 3.9476e-24], [593.83, 3.9519e-24], [595.2, 3.9562e-24], [596.57, 3.9606e-24], [597.95, 3.9649e-24], [599.33, 3.9693e-24], [600.71, 3.9737e-24], [602.09, 3.9782e-24], [603.48, 3.9826e-24], [604.87, 3.9871e-24], [606.27, 3.9916e-24], [607.67, 3.9961e-24], [609.07, 4.0006e-24], [610.47, 4.0052e-24], [611.88, 4.0097e-24], [613.29, 4.0143e-24], [614.7, 4.0189e-24], [616.12, 4.0236e-24], [617.54, 4.0282e-24], [618.97, 4.0329e-24], [620.39, 4.0376e-24], [621.83, 4.0423e-24], [623.26, 4.047e-24], [624.7, 4.0517e-24], [626.14, 4.0565e-24], [627.58, 4.0613e-24], [629.03, 4.0661e-24], [630.48, 4.0709e-24], [631.93, 4.0758e-24], [633.39, 4.0806e-24], [634.85, 4.0855e-24], [636.31, 4.0904e-24], [637.78, 4.0953e-24], [639.25, 4.1003e-24], [640.73, 4.1053e-24], [642.2, 4.1103e-24], [643.68, 4.1153e-24], [645.17, 4.1203e-24], [646.66, 4.1253e-24], [648.15, 4.1304e-24], [649.64, 4.1355e-24], [651.14, 4.1406e-24], [652.64, 4.1458e-24], [654.15, 4.1509e-24], [655.66, 4.1561e-24], [657.17, 4.1613e-24], [658.68, 4.1665e-24], [660.2, 4.1717e-24], [661.72, 4.177e-24], [663.25, 4.1823e-24], [664.78, 4.1876e-24], [666.31, 4.1929e-24], [667.85, 4.1982e-24], [669.39, 4.2036e-24], [670.93, 4.209e-24], [672.48, 4.2144e-24], [674.03, 4.2198e-24], [675.59, 4.2253e-24], [677.14, 4.2307e-24], [678.7, 4.2362e-24], [680.27, 4.2417e-24], [681.84, 4.2473e-24], [683.41, 4.2528e-24], [684.99, 4.2584e-24], [686.57, 4.264e-24], [688.15, 4.2696e-24], [689.74, 4.2753e-24], [691.33, 4.2809e-24], [692.92, 4.2866e-24], [694.52, 4.2923e-24], [696.12, 4.2981e-24], [697.73, 4.3038e-24], [699.33, 4.3096e-24], [700.95, 4.3154e-24], [702.56, 4.3212e-24], [704.18, 4.3271e-24], [705.81, 4.3329e-24], [707.44, 4.3388e-24], [709.07, 4.3447e-24], [710.7, 4.3506e-24], [712.34, 4.3566e-24], [713.98, 4.3626e-24], [715.63, 4.3686e-24], [717.28, 4.3746e-24], [718.93, 4.3806e-24], [720.59, 4.3867e-24], [722.25, 4.3928e-24], [723.92, 4.3989e-24], [725.59, 4.405e-24], [727.26, 4.4112e-24], [728.94, 4.4173e-24], [730.62, 4.4235e-24], [732.3, 4.4298e-24], [733.99, 4.436e-24], [735.69, 4.4423e-24], [737.38, 4.4486e-24], [739.08, 4.4549e-24], [740.79, 4.4612e-24], [742.5, 4.4676e-24], [744.21, 4.474e-24], [745.92, 4.4804e-24], [747.64, 4.4868e-24], [749.37, 4.4933e-24], [751.1, 4.4998e-24], [752.83, 4.5063e-24], [754.56, 4.5128e-24], [756.3, 4.5193e-24], [758.05, 4.5259e-24], [759.8, 4.5325e-24], [761.55, 4.5391e-24], [763.3, 4.5458e-24], [765.06, 4.5524e-24], [766.83, 4.5591e-24], [768.6, 4.5658e-24], [770.37, 4.5726e-24], [772.15, 4.5793e-24], [773.93, 4.5861e-24], [775.71, 4.5929e-24], [777.5, 4.5997e-24], [779.29, 4.6066e-24], [781.09, 4.6135e-24], [782.89, 4.6204e-24], [784.7, 4.6273e-24], [786.51, 4.6343e-24], [788.32, 4.6413e-24], [790.14, 4.6483e-24], [791.96, 4.6553e-24], [793.79, 4.6623e-24], [795.62, 4.6694e-24], [797.45, 4.6765e-24], [799.29, 4.6836e-24], [801.13, 4.6908e-24], [802.98, 4.698e-24], [804.83, 4.7052e-24], [806.69, 4.7124e-24], [808.55, 4.7196e-24], [810.41, 4.7269e-24], [812.28, 4.7342e-24], [814.16, 4.7415e-24], [816.03, 4.7489e-24], [817.91, 4.7562e-24], [819.8, 4.7636e-24], [821.69, 4.7711e-24], [823.59, 4.7785e-24], [825.49, 4.786e-24], [827.39, 4.7935e-24], [829.3, 4.801e-24], [831.21, 4.8086e-24], [833.13, 4.8161e-24], [835.05, 4.8237e-24], [836.97, 4.8314e-24], [838.9, 4.839e-24], [840.84, 4.8467e-24], [842.78, 4.8544e-24], [844.72, 4.8621e-24], [846.67, 4.8699e-24], [848.62, 4.8776e-24], [850.58, 4.8855e-24], [852.54, 4.8933e-24], [854.5, 4.9011e-24], [856.47, 4.909e-24], [858.45, 4.9169e-24], [860.43, 4.9249e-24], [862.41, 4.9328e-24], [864.4, 4.9408e-24], [866.4, 4.9488e-24], [868.39, 4.9569e-24], [870.4, 4.9649e-24], [872.4, 4.973e-24], [874.41, 4.9812e-24], [876.43, 4.9893e-24], [878.45, 4.9975e-24], [880.48, 5.0057e-24], [882.51, 5.0139e-24], [884.54, 5.0221e-24], [886.58, 5.0304e-24], [888.63, 5.0387e-24], [890.68, 5.0471e-24], [892.73, 5.0554e-24], [894.79, 5.0638e-24], [896.85, 5.0722e-24], [898.92, 5.0807e-24], [900.99, 5.0891e-24], [903.07, 5.0976e-24], [905.15, 5.1061e-24], [907.24, 5.1147e-24], [909.33, 5.1233e-24], [911.43, 5.1319e-24], [913.53, 5.1405e-24], [915.64, 5.1492e-24], [917.75, 5.1578e-24], [919.87, 5.1665e-24], [921.99, 5.1753e-24], [924.11, 5.1841e-24], [926.25, 5.1929e-24], [928.38, 5.2017e-24], [930.52, 5.2105e-24], [932.67, 5.2194e-24], [934.82, 5.2283e-24], [936.97, 5.2372e-24], [939.13, 5.2462e-24], [941.3, 5.2552e-24], [943.47, 5.2642e-24], [945.65, 5.2733e-24], [947.83, 5.2823e-24], [950.01, 5.2914e-24], [952.2, 5.3006e-24], [954.4, 5.3097e-24], [956.6, 5.3189e-24], [958.81, 5.3281e-24], [961.02, 5.3374e-24], [963.23, 5.3467e-24], [965.45, 5.356e-24], [967.68, 5.3653e-24], [969.91, 5.3747e-24], [972.15, 5.3841e-24], [974.39, 5.3935e-24], [976.64, 5.4029e-24], [978.89, 5.4124e-24], [981.15, 5.4219e-24], [983.41, 5.4315e-24], [985.68, 5.4411e-24], [987.95, 5.4508e-24], [990.23, 5.4612e-24], [992.51, 5.5229e-24], [994.8, 5.4813e-24], [997.1, 5.4896e-24], [999.39, 5.4991e-24], [1001.7, 5.5088e-24], [1004.0, 5.5186e-24], [1006.3, 5.5284e-24], [1008.6, 5.5382e-24], [1011.0, 5.5481e-24], [1013.3, 5.558e-24], [1015.6, 5.568e-24], [1018.0, 5.5779e-24], [1020.3, 5.5879e-24], [1022.7, 5.598e-24], [1025.0, 5.608e-24], [1027.4, 5.6181e-24], [1029.8, 5.6283e-24], [1032.1, 5.6384e-24], [1034.5, 5.6486e-24], [1036.9, 5.6588e-24], [1039.3, 5.6691e-24], [1041.7, 5.6793e-24], [1044.1, 5.6896e-24], [1046.5, 5.7e-24], [1048.9, 5.7103e-24], [1051.3, 5.7207e-24], [1053.8, 5.7312e-24], [1056.2, 5.7416e-24], [1058.6, 5.7521e-24], [1061.1, 5.7626e-24], [1063.5, 5.7732e-24], [1066.0, 5.7838e-24], [1068.4, 5.7944e-24], [1070.9, 5.805e-24], [1073.4, 5.8157e-24], [1075.8, 5.8264e-24], [1078.3, 5.8372e-24], [1080.8, 5.8479e-24], [1083.3, 5.8587e-24], [1085.8, 5.8696e-24], [1088.3, 5.8805e-24], [1090.8, 5.8913e-24], [1093.3, 5.9023e-24], [1095.8, 5.9132e-24], [1098.4, 5.9242e-24], [1100.9, 5.9353e-24], [1103.4, 5.9463e-24], [1106.0, 5.9574e-24], [1108.5, 5.9685e-24], [1111.1, 5.9797e-24], [1113.7, 5.9909e-24], [1116.2, 6.0021e-24], [1118.8, 6.0133e-24], [1121.4, 6.0246e-24], [1124.0, 6.0359e-24], [1126.6, 6.0473e-24], [1129.2, 6.0586e-24], [1131.8, 6.0701e-24], [1134.4, 6.0815e-24], [1137.0, 6.093e-24], [1139.6, 6.1045e-24], [1142.2, 6.116e-24], [1144.9, 6.1276e-24], [1147.5, 6.1392e-24], [1150.2, 6.1509e-24], [1152.8, 6.1625e-24], [1155.5, 6.1742e-24], [1158.1, 6.186e-24], [1160.8, 6.1978e-24], [1163.5, 6.2096e-24], [1166.2, 6.2214e-24], [1168.9, 6.2333e-24], [1171.5, 6.2452e-24], [1174.2, 6.2571e-24], [1177.0, 6.2691e-24], [1179.7, 6.2811e-24], [1182.4, 6.2932e-24], [1185.1, 6.3052e-24], [1187.9, 6.3173e-24], [1190.6, 6.3295e-24], [1193.3, 6.3417e-24], [1196.1, 6.3539e-24], [1198.8, 6.3661e-24], [1201.6, 6.3784e-24], [1204.4, 6.3907e-24], [1207.2, 6.4031e-24], [1209.9, 6.4155e-24], [1212.7, 6.4279e-24], [1215.5, 6.4403e-24], [1218.3, 6.4528e-24], [1221.1, 6.4653e-24], [1224.0, 6.4779e-24], [1226.8, 6.4905e-24], [1229.6, 6.5031e-24], [1232.4, 6.5158e-24], [1235.3, 6.5285e-24], [1238.1, 6.5412e-24], [1241.0, 6.554e-24], [1243.9, 6.5668e-24], [1246.7, 6.5796e-24], [1249.6, 6.5925e-24], [1252.5, 6.6054e-24], [1255.4, 6.6184e-24], [1258.3, 6.6313e-24], [1261.2, 6.6444e-24], [1264.1, 6.6574e-24], [1267.0, 6.6705e-24], [1269.9, 6.6836e-24], [1272.8, 6.6968e-24], [1275.8, 6.71e-24], [1278.7, 6.7232e-24], [1281.7, 6.7365e-24], [1284.6, 6.7498e-24], [1287.6, 6.7631e-24], [1290.5, 6.7765e-24], [1293.5, 6.7899e-24], [1296.5, 6.8034e-24], [1299.5, 6.8168e-24], [1302.5, 6.8304e-24], [1305.5, 6.8439e-24], [1308.5, 6.8575e-24], [1311.5, 6.8712e-24], [1314.5, 6.8848e-24], [1317.6, 6.8985e-24], [1320.6, 6.9123e-24], [1323.7, 6.9261e-24], [1326.7, 6.9399e-24], [1329.8, 6.9537e-24], [1332.8, 6.9676e-24], [1335.9, 6.9816e-24], [1339.0, 6.9955e-24], [1342.1, 7.0096e-24], [1345.2, 7.0236e-24], [1348.3, 7.0377e-24], [1351.4, 7.0518e-24], [1354.5, 7.066e-24], [1357.6, 7.0802e-24], [1360.8, 7.0944e-24], [1363.9, 7.1087e-24], [1367.0, 7.123e-24], [1370.2, 7.1373e-24], [1373.4, 7.1517e-24], [1376.5, 7.1661e-24], [1379.7, 7.1806e-24], [1382.9, 7.1951e-24], [1386.1, 7.2096e-24], [1389.3, 7.2242e-24], [1392.5, 7.2388e-24], [1395.7, 7.2535e-24], [1398.9, 7.2682e-24], [1402.1, 7.2829e-24], [1405.4, 7.2977e-24], [1408.6, 7.3125e-24], [1411.8, 7.3274e-24], [1415.1, 7.3423e-24], [1418.4, 7.3572e-24], [1421.6, 7.3722e-24], [1424.9, 7.3872e-24], [1428.2, 7.4022e-24], [1431.5, 7.4173e-24], [1434.8, 7.4325e-24], [1438.1, 7.4476e-24], [1441.4, 7.4629e-24], [1444.7, 7.4781e-24], [1448.1, 7.4934e-24], [1451.4, 7.5087e-24], [1454.8, 7.5241e-24], [1458.1, 7.5395e-24], [1461.5, 7.555e-24], [1464.8, 7.5705e-24], [1468.2, 7.586e-24], [1471.6, 7.6016e-24], [1475.0, 7.6172e-24], [1478.4, 7.6329e-24], [1481.8, 7.6486e-24], [1485.2, 7.6644e-24], [1488.7, 7.6837e-24], [1492.1, 7.6961e-24], [1495.5, 7.7118e-24], [1499.0, 7.7277e-24], [1502.4, 7.7436e-24], [1505.9, 7.7596e-24], [1509.4, 7.7756e-24], [1512.9, 7.7917e-24], [1516.3, 7.8078e-24], [1519.8, 7.824e-24], [1523.3, 7.8402e-24], [1526.9, 7.8564e-24], [1530.4, 7.8727e-24], [1533.9, 7.889e-24], [1537.4, 7.9054e-24], [1541.0, 7.9218e-24], [1544.5, 7.9382e-24], [1548.1, 7.9547e-24], [1551.7, 7.9713e-24], [1555.3, 7.9879e-24], [1558.8, 8.0045e-24], [1562.4, 8.0211e-24], [1566.0, 8.0378e-24], [1569.6, 8.0546e-24], [1573.3, 8.0714e-24], [1576.9, 8.0882e-24], [1580.5, 8.1051e-24], [1584.2, 8.122e-24], [1587.8, 8.139e-24], [1591.5, 8.156e-24], [1595.2, 8.1731e-24], [1598.8, 8.1902e-24], [1602.5, 8.2073e-24], [1606.2, 8.2245e-24], [1609.9, 8.2418e-24], [1613.6, 8.259e-24], [1617.4, 8.2763e-24], [1621.1, 8.2937e-24], [1624.8, 8.3111e-24], [1628.6, 8.3286e-24], [1632.3, 8.3461e-24], [1636.1, 8.3636e-24], [1639.9, 8.3812e-24], [1643.6, 8.3989e-24], [1647.4, 8.4165e-24], [1651.2, 8.4343e-24], [1655.0, 8.452e-24], [1658.9, 8.4698e-24], [1662.7, 8.4877e-24], [1666.5, 8.5056e-24], [1670.4, 8.5236e-24], [1674.2, 8.5416e-24], [1678.1, 8.5596e-24], [1681.9, 8.5777e-24], [1685.8, 8.5958e-24], [1689.7, 8.614e-24], [1693.6, 8.6322e-24], [1697.5, 8.6505e-24], [1701.4, 8.6688e-24], [1705.4, 8.6872e-24], [1709.3, 8.7056e-24], [1713.2, 8.7241e-24], [1717.2, 8.7426e-24], [1721.1, 8.7611e-24], [1725.1, 8.7797e-24], [1729.1, 8.7984e-24], [1733.1, 8.8171e-24], [1737.1, 8.8358e-24], [1741.1, 8.8546e-24], [1745.1, 8.8735e-24], [1749.1, 8.8923e-24], [1753.1, 8.9113e-24], [1757.2, 8.9303e-24], [1761.2, 8.9493e-24], [1765.3, 8.9684e-24], [1769.4, 8.9875e-24], [1773.5, 9.0067e-24], [1777.5, 9.0259e-24], [1781.6, 9.0452e-24], [1785.8, 9.0645e-24], [1789.9, 9.0838e-24], [1794.0, 9.1033e-24], [1798.1, 9.1227e-24], [1802.3, 9.1422e-24], [1806.4, 9.1618e-24], [1810.6, 9.1814e-24], [1814.8, 9.2011e-24], [1819.0, 9.2208e-24], [1823.2, 9.2405e-24], [1827.4, 9.2603e-24], [1831.6, 9.2802e-24], [1835.8, 9.3001e-24], [1840.0, 9.3201e-24], [1844.3, 9.3401e-24], [1848.5, 9.3601e-24], [1852.8, 9.3803e-24], [1857.1, 9.4004e-24], [1861.3, 9.4206e-24], [1865.6, 9.4409e-24], [1869.9, 9.4612e-24], [1874.3, 9.4816e-24], [1878.6, 9.502e-24], [1882.9, 9.5224e-24], [1887.2, 9.5429e-24], [1891.6, 9.5635e-24], [1896.0, 9.5841e-24], [1900.3, 9.6048e-24], [1904.7, 9.6255e-24], [1909.1, 9.6463e-24], [1913.5, 9.6671e-24], [1917.9, 9.688e-24], [1922.3, 9.7089e-24], [1926.8, 9.7299e-24], [1931.2, 9.7509e-24], [1935.7, 9.772e-24], [1940.1, 9.7931e-24], [1944.6, 9.8143e-24], [1949.1, 9.8356e-24], [1953.6, 9.8569e-24], [1958.1, 9.8782e-24], [1962.6, 9.8996e-24], [1967.1, 9.9211e-24], [1971.7, 9.9426e-24], [1976.2, 9.9641e-24], [1980.8, 9.9857e-24], [1985.3, 1.0011e-23], [1989.9, 1.0029e-23], [1994.5, 1.0051e-23], [1999.1, 1.0073e-23], [2003.7, 1.0095e-23], [2008.3, 1.0117e-23], [2013.0, 1.0139e-23], [2017.6, 1.0161e-23], [2022.3, 1.0183e-23], [2026.9, 1.0205e-23], [2031.6, 1.0227e-23], [2036.3, 1.0249e-23], [2041.0, 1.0272e-23], [2045.7, 1.0294e-23], [2050.4, 1.0316e-23], [2055.1, 1.0339e-23], [2059.9, 1.0361e-23], [2064.6, 1.0384e-23], [2069.4, 1.0407e-23], [2074.2, 1.0429e-23], [2078.9, 1.0452e-23], [2083.7, 1.0475e-23], [2088.5, 1.0498e-23], [2093.4, 1.0521e-23], [2098.2, 1.0544e-23], [2103.0, 1.0567e-23], [2107.9, 1.059e-23], [2112.7, 1.0613e-23], [2117.6, 1.0636e-23], [2122.5, 1.0659e-23], [2127.4, 1.0683e-23], [2132.3, 1.0706e-23], [2137.2, 1.073e-23], [2142.1, 1.0753e-23], [2147.1, 1.0777e-23], [2152.0, 1.08e-23], [2157.0, 1.0824e-23], [2162.0, 1.0848e-23], [2167.0, 1.0871e-23], [2171.9, 1.0895e-23], [2177.0, 1.0919e-23], [2182.0, 1.0943e-23], [2187.0, 1.0967e-23], [2192.1, 1.0991e-23], [2197.1, 1.1015e-23], [2202.2, 1.104e-23], [2207.3, 1.1064e-23], [2212.3, 1.1088e-23], [2217.4, 1.1113e-23], [2222.6, 1.1137e-23], [2227.7, 1.1161e-23], [2232.8, 1.1186e-23], [2238.0, 1.1211e-23], [2243.1, 1.1235e-23], [2248.3, 1.126e-23], [2253.5, 1.1285e-23], [2258.7, 1.131e-23], [2263.9, 1.1335e-23], [2269.1, 1.136e-23], [2274.3, 1.1385e-23], [2279.6, 1.141e-23], [2284.8, 1.1435e-23], [2290.1, 1.146e-23], [2295.4, 1.1485e-23], [2300.7, 1.1511e-23], [2306.0, 1.1536e-23], [2311.3, 1.1562e-23], [2316.6, 1.1587e-23], [2322.0, 1.1613e-23], [2327.3, 1.1638e-23], [2332.7, 1.1664e-23], [2338.1, 1.169e-23], [2343.5, 1.1716e-23], [2348.9, 1.1742e-23], [2354.3, 1.1768e-23], [2359.7, 1.1794e-23], [2365.2, 1.182e-23], [2370.6, 1.1846e-23], [2376.1, 1.1872e-23], [2381.6, 1.1898e-23], [2387.1, 1.1925e-23], [2392.6, 1.1951e-23], [2398.1, 1.1978e-23], [2403.6, 1.2004e-23], [2409.2, 1.2031e-23], [2414.7, 1.2057e-23], [2420.3, 1.2084e-23], [2425.9, 1.2111e-23], [2431.5, 1.2138e-23], [2437.1, 1.2165e-23], [2442.7, 1.2192e-23], [2448.3, 1.2219e-23], [2454.0, 1.2246e-23], [2459.6, 1.2273e-23], [2465.3, 1.23e-23], [2471.0, 1.2328e-23], [2476.7, 1.2355e-23], [2482.4, 1.2384e-23], [2488.1, 1.241e-23], [2493.8, 1.2437e-23], [2499.6, 1.2465e-23], [2505.4, 1.2493e-23], [2511.1, 1.2521e-23], [2516.9, 1.2548e-23], [2522.7, 1.2576e-23], [2528.5, 1.2604e-23], [2534.4, 1.2632e-23], [2540.2, 1.2661e-23], [2546.1, 1.2689e-23], [2552.0, 1.2717e-23], [2557.8, 1.2745e-23], [2563.7, 1.2774e-23], [2569.6, 1.2802e-23], [2575.6, 1.2831e-23], [2581.5, 1.2859e-23], [2587.5, 1.2888e-23], [2593.4, 1.2917e-23], [2599.4, 1.2945e-23], [2605.4, 1.2974e-23], [2611.4, 1.3003e-23], [2617.4, 1.3032e-23], [2623.5, 1.3061e-23], [2629.5, 1.309e-23], [2635.6, 1.312e-23], [2641.7, 1.3149e-23], [2647.8, 1.3178e-23], [2653.9, 1.3208e-23], [2660.0, 1.3237e-23], [2666.1, 1.3267e-23], [2672.3, 1.3296e-23], [2678.4, 1.3326e-23], [2684.6, 1.3356e-23], [2690.8, 1.3386e-23], [2697.0, 1.3416e-23], [2703.2, 1.3446e-23], [2709.5, 1.3476e-23], [2715.7, 1.3506e-23], [2722.0, 1.3536e-23], [2728.2, 1.3566e-23], [2734.5, 1.3597e-23], [2740.8, 1.3627e-23], [2747.2, 1.3658e-23], [2753.5, 1.3688e-23], [2759.8, 1.3719e-23], [2766.2, 1.375e-23], [2772.6, 1.3781e-23], [2779.0, 1.3811e-23], [2785.4, 1.3842e-23], [2791.8, 1.3873e-23], [2798.2, 1.3904e-23], [2804.7, 1.3936e-23], [2811.2, 1.3967e-23], [2817.6, 1.3998e-23], [2824.1, 1.403e-23], [2830.7, 1.4061e-23], [2837.2, 1.4093e-23], [2843.7, 1.4124e-23], [2850.3, 1.4156e-23], [2856.9, 1.4188e-23], [2863.4, 1.422e-23], [2870.1, 1.4251e-23], [2876.7, 1.4283e-23], [2883.3, 1.4316e-23], [2890.0, 1.4348e-23], [2896.6, 1.438e-23], [2903.3, 1.4412e-23], [2910.0, 1.4445e-23], [2916.7, 1.4477e-23], [2923.4, 1.451e-23], [2930.2, 1.4542e-23], [2936.9, 1.4575e-23], [2943.7, 1.4608e-23], [2950.5, 1.4641e-23], [2957.3, 1.4673e-23], [2964.1, 1.4706e-23], [2970.9, 1.474e-23], [2977.8, 1.4773e-23], [2984.7, 1.4806e-23], [2991.5, 1.4839e-23], [2998.4, 1.4873e-23], [3005.4, 1.4906e-23], [3012.3, 1.494e-23], [3019.2, 1.4973e-23], [3026.2, 1.5007e-23], [3033.2, 1.5041e-23], [3040.2, 1.5075e-23], [3047.2, 1.5109e-23], [3054.2, 1.5143e-23], [3061.3, 1.5177e-23], [3068.3, 1.5211e-23], [3075.4, 1.5245e-23], [3082.5, 1.528e-23], [3089.6, 1.5314e-23], [3096.7, 1.5349e-23], [3103.9, 1.5383e-23], [3111.0, 1.5418e-23], [3118.2, 1.5453e-23], [3125.4, 1.5488e-23], [3132.6, 1.5523e-23], [3139.8, 1.5558e-23], [3147.0, 1.5593e-23], [3154.3, 1.5628e-23], [3161.6, 1.5663e-23], [3168.9, 1.5699e-23], [3176.2, 1.5734e-23], [3183.5, 1.577e-23], [3190.8, 1.5805e-23], [3198.2, 1.5841e-23], [3205.6, 1.5877e-23], [3213.0, 1.5913e-23], [3220.4, 1.5949e-23], [3227.8, 1.5985e-23], [3235.2, 1.6021e-23], [3242.7, 1.6057e-23], [3250.2, 1.6093e-23], [3257.7, 1.613e-23], [3265.2, 1.6166e-23], [3272.7, 1.6203e-23], [3280.3, 1.6239e-23], [3287.8, 1.6276e-23], [3295.4, 1.6313e-23], [3303.0, 1.635e-23], [3310.6, 1.6387e-23], [3318.3, 1.6424e-23], [3325.9, 1.6461e-23], [3333.6, 1.6498e-23], [3341.3, 1.6536e-23], [3349.0, 1.6573e-23], [3356.7, 1.6611e-23], [3364.4, 1.6648e-23], [3372.2, 1.6686e-23], [3380.0, 1.6724e-23], [3387.8, 1.6762e-23], [3395.6, 1.6799e-23], [3403.4, 1.6838e-23], [3411.3, 1.6876e-23], [3419.1, 1.6914e-23], [3427.0, 1.6952e-23], [3434.9, 1.6991e-23], [3442.8, 1.7029e-23], [3450.8, 1.7068e-23], [3458.7, 1.7106e-23], [3466.7, 1.7145e-23], [3474.7, 1.7225e-23], [3482.7, 1.7223e-23], [3490.7, 1.7262e-23], [3498.8, 1.7301e-23], [3506.9, 1.7341e-23], [3514.9, 1.738e-23], [3523.1, 1.7419e-23], [3531.2, 1.7459e-23], [3539.3, 1.7499e-23], [3547.5, 1.7538e-23], [3555.7, 1.7578e-23], [3563.9, 1.7618e-23], [3572.1, 1.7658e-23], [3580.3, 1.7698e-23], [3588.6, 1.7738e-23], [3596.8, 1.7778e-23], [3605.1, 1.7819e-23], [3613.5, 1.7859e-23], [3621.8, 1.79e-23], [3630.1, 1.7941e-23], [3638.5, 1.7981e-23], [3646.9, 1.8022e-23], [3655.3, 1.8063e-23], [3663.7, 1.8104e-23], [3672.2, 1.8145e-23], [3680.7, 1.8187e-23], [3689.1, 1.8228e-23], [3697.7, 1.8269e-23], [3706.2, 1.8311e-23], [3714.7, 1.8353e-23], [3723.3, 1.8394e-23], [3731.9, 1.8436e-23], [3740.5, 1.8478e-23], [3749.1, 1.852e-23], [3757.8, 1.8562e-23], [3766.4, 1.8604e-23], [3775.1, 1.8647e-23], [3783.8, 1.8689e-23], [3792.5, 1.8732e-23], [3801.3, 1.8774e-23], [3810.0, 1.8817e-23], [3818.8, 1.886e-23], [3827.6, 1.8903e-23], [3836.5, 1.8946e-23], [3845.3, 1.8989e-23], [3854.2, 1.9032e-23], [3863.1, 1.9076e-23], [3872.0, 1.9119e-23], [3880.9, 1.9163e-23], [3889.9, 1.9206e-23], [3898.8, 1.925e-23], [3907.8, 1.9294e-23], [3916.8, 1.9338e-23], [3925.9, 1.9382e-23], [3934.9, 1.9426e-23], [3944.0, 1.947e-23], [3953.1, 1.9515e-23], [3962.2, 1.9559e-23], [3971.3, 1.9604e-23], [3980.5, 1.9648e-23], [3989.7, 1.9693e-23], [3998.9, 1.9738e-23], [4008.1, 1.9783e-23], [4017.3, 1.9828e-23], [4026.6, 1.9873e-23], [4035.9, 1.9919e-23], [4045.2, 1.9964e-23], [4054.5, 2.001e-23], [4063.9, 2.0055e-23], [4073.2, 2.0101e-23], [4082.6, 2.0147e-23], [4092.1, 2.0193e-23], [4101.5, 2.0239e-23], [4110.9, 2.0285e-23], [4120.4, 2.0332e-23], [4129.9, 2.0378e-23], [4139.5, 2.0424e-23], [4149.0, 2.0471e-23], [4158.6, 2.0518e-23], [4168.2, 2.0565e-23], [4177.8, 2.0612e-23], [4187.4, 2.0659e-23], [4197.1, 2.0706e-23], [4206.7, 2.0753e-23], [4216.4, 2.0801e-23], [4226.2, 2.0848e-23], [4235.9, 2.0896e-23], [4245.7, 2.0943e-23], [4255.5, 2.0991e-23], [4265.3, 2.1039e-23], [4275.1, 2.1087e-23], [4285.0, 2.1136e-23], [4294.8, 2.1184e-23], [4304.8, 2.1232e-23], [4314.7, 2.1281e-23], [4324.6, 2.1329e-23], [4334.6, 2.1378e-23], [4344.6, 2.1427e-23], [4354.6, 2.1476e-23], [4364.7, 2.1525e-23], [4374.7, 2.1574e-23], [4384.8, 2.1624e-23], [4394.9, 2.1673e-23], [4405.1, 2.1723e-23], [4415.2, 2.1772e-23], [4425.4, 2.1822e-23], [4435.6, 2.1872e-23], [4445.8, 2.1922e-23], [4456.1, 2.1972e-23], [4466.4, 2.2023e-23], [4476.7, 2.2073e-23], [4487.0, 2.2124e-23], [4497.3, 2.2174e-23], [4507.7, 2.2225e-23], [4518.1, 2.2276e-23], [4528.5, 2.2327e-23], [4539.0, 2.2378e-23], [4549.4, 2.2429e-23], [4559.9, 2.248e-23], [4570.4, 2.2532e-23], [4581.0, 2.2584e-23], [4591.5, 2.2635e-23], [4602.1, 2.2687e-23], [4612.7, 2.2739e-23], [4623.4, 2.2791e-23], [4634.0, 2.2843e-23], [4644.7, 2.2896e-23], [4655.4, 2.2948e-23], [4666.2, 2.3001e-23], [4676.9, 2.3053e-23], [4687.7, 2.3106e-23], [4698.5, 2.3159e-23], [4709.4, 2.3212e-23], [4720.2, 2.3265e-23], [4731.1, 2.3319e-23], [4742.0, 2.3372e-23], [4752.9, 2.3426e-23], [4763.9, 2.3479e-23], [4774.9, 2.3533e-23], [4785.9, 2.3587e-23], [4796.9, 2.3641e-23], [4808.0, 2.3695e-23], [4819.1, 2.375e-23], [4830.2, 2.3804e-23], [4841.3, 2.3859e-23], [4852.5, 2.3914e-23], [4863.7, 2.3968e-23], [4874.9, 2.4023e-23], [4886.1, 2.4079e-23], [4897.4, 2.4134e-23], [4908.7, 2.4189e-23], [4920.0, 2.4245e-23], [4931.4, 2.43e-23], [4942.7, 2.4356e-23], [4954.1, 2.4412e-23], [4965.6, 2.4468e-23], [4977.0, 2.4524e-23], [4988.5, 2.458e-23], [5000.0, 2.4637e-23]])
    else:
        raise ValueError('Value '+source+' of internal variable source not recognised')

    f1 = data[:,0]
    S1 = data[:,col]**2
    fstop = [f1[0],f1[-1]]

    # % ---- Interpolate to desired frequencies.  Set noise to infinity outside
    # %      band of frequencies covered by f1.
    # S = inf(size(f)); 
    # k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop


# % -------------------------------------------------------------------------
# %                              Advanced Virgo spectra
# % -------------------------------------------------------------------------

def avirgospectrum(IFO,f):
# % AVIRGOSPECTRUM - Design noise curves for advanced Virgo.
# %
# % Arguments are the same as for the parent function.
# %
# % The spectra are taken from the ligovirgo CVS, 
# % ligovirgo/dac/policy/CoordinatedFollowUpProgram/data/AdV_evolution2.fig
# % These are the official design curves as of June 2012.
# % The AVIRGO-BNS spectrum matches that from
# % https://wwwcascina.virgo.infn.it/advirgo/, corresponding to proposal
# % VIR-0337A-10 and used for ER1 simulated data generation.  

# % The following analytic fit is copied from Manzotti & Dietz,
# % https://dcc.ligo.org/cgi-bin/private/DocDB/ShowDocument?docid=p1200014
# % I have not checked its accuracy, and it is not used in this function.
# %   S.^0.5 = 1.259e-24 * ( ...
# %       + 0.07*exp(-0.142-1.437*x+0.407*x.^2) ...
# %       + 3.10*exp(-0.466-1.043*x-0.548*x.^2) ...
# %       + 0.40*exp(-0.304+2.896*x-0.293*x.^2) ...
# %       + 0.09*exp( 1.466+3.722*x-0.984*x.^2) ...
# %   );
# % where 
# %   x := log(f/300Hz)

# % ---- Recognize detector and compute noise spectrum.
    if IFO.upper()=='AVIRGO-NOSRM':
        col = 1
    elif IFO.upper()=='AVIRGO-ZDLP':
        col = 2
    elif IFO.upper()=='AVIRGO-BNS':
        col = 3
    elif IFO.upper()=='AVIRGO-ZDHP':
        col = 4
    else:
        raise ValueError('Detector '+IFO+' not recognized.')

# % ---- Extract pre-sampled noise spectrum from data.  First column is
# %      frequency, other columns map to configurations as above.
    data = np.array([[10.0, 1.10777e-21, 1.09837e-21, 1.10758e-21, 1.10062e-21], [10.0231, 1.09746e-21, 1.08806e-21, 1.09727e-21, 1.09031e-21], [10.0462, 1.08725e-21, 1.07785e-21, 1.08706e-21, 1.0801e-21], [10.0693, 1.07714e-21, 1.06773e-21, 1.07694e-21, 1.06998e-21], [10.0926, 1.06712e-21, 1.05772e-21, 1.06693e-21, 1.05997e-21], [10.1158, 1.05721e-21, 1.0478e-21, 1.05701e-21, 1.05005e-21], [10.1392, 1.04739e-21, 1.03798e-21, 1.04719e-21, 1.04024e-21], [10.1625, 1.03767e-21, 1.02826e-21, 1.03747e-21, 1.03051e-21], [10.186, 1.02804e-21, 1.01863e-21, 1.02784e-21, 1.02088e-21], [10.2095, 1.0185e-21, 1.0091e-21, 1.0183e-21, 1.01135e-21], [10.233, 1.00906e-21, 9.99654e-22, 1.00886e-21, 1.00191e-21], [10.2566, 9.99711e-22, 9.90303e-22, 9.99508e-22, 9.92554e-22], [10.2803, 9.90451e-22, 9.81042e-22, 9.90246e-22, 9.83294e-22], [10.304, 9.81281e-22, 9.71871e-22, 9.81074e-22, 9.74123e-22], [10.3277, 9.72199e-22, 9.62789e-22, 9.7199e-22, 9.6504e-22], [10.3515, 9.63204e-22, 9.53794e-22, 9.62994e-22, 9.56046e-22], [10.3754, 9.54297e-22, 9.44886e-22, 9.54085e-22, 9.47138e-22], [10.3993, 9.45475e-22, 9.36064e-22, 9.45262e-22, 9.38316e-22], [10.4233, 9.36739e-22, 9.27327e-22, 9.36523e-22, 9.29579e-22], [10.4474, 9.28086e-22, 9.18674e-22, 9.27869e-22, 9.20926e-22], [10.4714, 9.19517e-22, 9.10104e-22, 9.19298e-22, 9.12356e-22], [10.4956, 9.11029e-22, 9.01617e-22, 9.10809e-22, 9.03869e-22], [10.5198, 9.02624e-22, 8.93211e-22, 9.02402e-22, 8.95463e-22], [10.5441, 8.94299e-22, 8.84886e-22, 8.94075e-22, 8.87138e-22], [10.5684, 8.86053e-22, 8.76641e-22, 8.85828e-22, 8.78892e-22], [10.5927, 8.77887e-22, 8.68474e-22, 8.7766e-22, 8.70726e-22], [10.6172, 8.69799e-22, 8.60386e-22, 8.6957e-22, 8.62638e-22], [10.6417, 8.61788e-22, 8.52375e-22, 8.61558e-22, 8.54627e-22], [10.6662, 8.53854e-22, 8.44441e-22, 8.53622e-22, 8.46693e-22], [10.6908, 8.45996e-22, 8.36583e-22, 8.45762e-22, 8.38835e-22], [10.7154, 8.38213e-22, 8.288e-22, 8.37976e-22, 8.31051e-22], [10.7401, 8.30504e-22, 8.21091e-22, 8.30266e-22, 8.23342e-22], [10.7649, 8.22868e-22, 8.13455e-22, 8.22628e-22, 8.15707e-22], [10.7897, 8.15305e-22, 8.05893e-22, 8.15064e-22, 8.08144e-22], [10.8146, 8.07815e-22, 7.98402e-22, 8.07572e-22, 8.00654e-22], [10.8396, 8.00395e-22, 7.90983e-22, 8.0015e-22, 7.93235e-22], [10.8646, 7.93047e-22, 7.83635e-22, 7.928e-22, 7.85886e-22], [10.8896, 7.85768e-22, 7.76356e-22, 7.85519e-22, 7.78607e-22], [10.9147, 7.78558e-22, 7.69147e-22, 7.78308e-22, 7.71398e-22], [10.9399, 7.71417e-22, 7.62006e-22, 7.71165e-22, 7.64257e-22], [10.9651, 7.64344e-22, 7.54933e-22, 7.6409e-22, 7.57184e-22], [10.9904, 7.57338e-22, 7.47928e-22, 7.57082e-22, 7.50178e-22], [11.0157, 7.50398e-22, 7.40988e-22, 7.50141e-22, 7.43239e-22], [11.0412, 7.43524e-22, 7.34115e-22, 7.43265e-22, 7.36366e-22], [11.0666, 7.36716e-22, 7.27307e-22, 7.36455e-22, 7.29558e-22], [11.0921, 7.29972e-22, 7.20564e-22, 7.29709e-22, 7.22814e-22], [11.1177, 7.23292e-22, 7.13884e-22, 7.23027e-22, 7.16134e-22], [11.1433, 7.16675e-22, 7.07268e-22, 7.16408e-22, 7.09518e-22], [11.169, 7.10121e-22, 7.00715e-22, 7.09852e-22, 7.02964e-22], [11.1948, 7.03629e-22, 6.94223e-22, 7.03359e-22, 6.96473e-22], [11.2206, 6.97198e-22, 6.87793e-22, 6.96926e-22, 6.90043e-22], [11.2465, 6.90828e-22, 6.81424e-22, 6.90554e-22, 6.83673e-22], [11.2724, 6.84519e-22, 6.75116e-22, 6.84243e-22, 6.77364e-22], [11.2984, 6.78269e-22, 6.68867e-22, 6.77991e-22, 6.71115e-22], [11.3245, 6.72078e-22, 6.62677e-22, 6.71799e-22, 6.64925e-22], [11.3506, 6.65946e-22, 6.56545e-22, 6.65665e-22, 6.58793e-22], [11.3768, 6.59872e-22, 6.50472e-22, 6.59588e-22, 6.5272e-22], [11.403, 6.53855e-22, 6.44456e-22, 6.5357e-22, 6.46703e-22], [11.4293, 6.47894e-22, 6.38497e-22, 6.47608e-22, 6.40744e-22], [11.4556, 6.41991e-22, 6.32594e-22, 6.41702e-22, 6.34841e-22], [11.4821, 6.36142e-22, 6.26747e-22, 6.35852e-22, 6.28994e-22], [11.5085, 6.3035e-22, 6.20955e-22, 6.30057e-22, 6.23202e-22], [11.5351, 6.24611e-22, 6.15218e-22, 6.24317e-22, 6.17464e-22], [11.5617, 6.18927e-22, 6.09535e-22, 6.18631e-22, 6.11781e-22], [11.5883, 6.13297e-22, 6.03906e-22, 6.12999e-22, 6.06151e-22], [11.6151, 6.07719e-22, 5.9833e-22, 6.07419e-22, 6.00575e-22], [11.6418, 6.02195e-22, 5.92806e-22, 6.01893e-22, 5.95051e-22], [11.6687, 5.96722e-22, 5.87335e-22, 5.96418e-22, 5.89579e-22], [11.6956, 5.91301e-22, 5.81915e-22, 5.90995e-22, 5.84159e-22], [11.7226, 5.85931e-22, 5.76547e-22, 5.85623e-22, 5.7879e-22], [11.7496, 5.80612e-22, 5.71229e-22, 5.80302e-22, 5.73472e-22], [11.7767, 5.75342e-22, 5.65961e-22, 5.75031e-22, 5.68204e-22], [11.8039, 5.70123e-22, 5.60743e-22, 5.69809e-22, 5.62985e-22], [11.8311, 5.64952e-22, 5.55574e-22, 5.64637e-22, 5.57816e-22], [11.8584, 5.59831e-22, 5.50454e-22, 5.59513e-22, 5.52696e-22], [11.8857, 5.54758e-22, 5.45382e-22, 5.54438e-22, 5.47623e-22], [11.9131, 5.49732e-22, 5.40358e-22, 5.49411e-22, 5.42599e-22], [11.9406, 5.44754e-22, 5.35382e-22, 5.44431e-22, 5.37622e-22], [11.9681, 5.39823e-22, 5.30452e-22, 5.39497e-22, 5.32692e-22], [11.9957, 5.34938e-22, 5.25569e-22, 5.34611e-22, 5.27809e-22], [12.0234, 5.30099e-22, 5.20732e-22, 5.2977e-22, 5.22971e-22], [12.0511, 5.25306e-22, 5.15941e-22, 5.24975e-22, 5.18179e-22], [12.0789, 5.20558e-22, 5.11195e-22, 5.20225e-22, 5.13433e-22], [12.1068, 5.15855e-22, 5.06493e-22, 5.1552e-22, 5.08731e-22], [12.1347, 5.11196e-22, 5.01837e-22, 5.10859e-22, 5.04073e-22], [12.1627, 5.06581e-22, 4.97224e-22, 5.06242e-22, 4.9946e-22], [12.1907, 5.02009e-22, 4.92654e-22, 5.01668e-22, 4.9489e-22], [12.2188, 4.97481e-22, 4.88128e-22, 4.97138e-22, 4.90363e-22], [12.247, 4.92995e-22, 4.83644e-22, 4.9265e-22, 4.85879e-22], [12.2752, 4.88552e-22, 4.79203e-22, 4.88205e-22, 4.81437e-22], [12.3035, 4.84151e-22, 4.74804e-22, 4.83802e-22, 4.77038e-22], [12.3319, 4.79791e-22, 4.70446e-22, 4.7944e-22, 4.72679e-22], [12.3603, 4.75472e-22, 4.6613e-22, 4.75119e-22, 4.68362e-22], [12.3888, 4.71194e-22, 4.61854e-22, 4.70839e-22, 4.64086e-22], [12.4174, 4.66957e-22, 4.57619e-22, 4.666e-22, 4.5985e-22], [12.4461, 4.62759e-22, 4.53424e-22, 4.624e-22, 4.55655e-22], [12.4748, 4.58602e-22, 4.49269e-22, 4.58241e-22, 4.51499e-22], [12.5035, 4.54483e-22, 4.45153e-22, 4.5412e-22, 4.47382e-22], [12.5324, 4.50404e-22, 4.41076e-22, 4.50038e-22, 4.43305e-22], [12.5613, 4.46363e-22, 4.37038e-22, 4.45996e-22, 4.39266e-22], [12.5902, 4.4236e-22, 4.33038e-22, 4.41991e-22, 4.35265e-22], [12.6193, 4.38395e-22, 4.29076e-22, 4.38024e-22, 4.31302e-22], [12.6484, 4.34468e-22, 4.25151e-22, 4.34095e-22, 4.27377e-22], [12.6775, 4.30578e-22, 4.21264e-22, 4.30203e-22, 4.23489e-22], [12.7068, 4.26725e-22, 4.17414e-22, 4.26348e-22, 4.19638e-22], [12.7361, 4.22909e-22, 4.136e-22, 4.22529e-22, 4.15824e-22], [12.7654, 4.19128e-22, 4.09823e-22, 4.18747e-22, 4.12046e-22], [12.7949, 4.15384e-22, 4.06081e-22, 4.15e-22, 4.08303e-22], [12.8244, 4.11675e-22, 4.02376e-22, 4.1129e-22, 4.04597e-22], [12.8539, 4.08002e-22, 3.98706e-22, 4.07614e-22, 4.00926e-22], [12.8836, 4.04363e-22, 3.9507e-22, 4.03974e-22, 3.9729e-22], [12.9133, 4.0076e-22, 3.9147e-22, 4.00368e-22, 3.93688e-22], [12.9431, 3.9719e-22, 3.87904e-22, 3.96796e-22, 3.90121e-22], [12.9729, 3.93655e-22, 3.84372e-22, 3.93259e-22, 3.86589e-22], [13.0028, 3.90154e-22, 3.80874e-22, 3.89755e-22, 3.8309e-22], [13.0328, 3.86686e-22, 3.77409e-22, 3.86285e-22, 3.79624e-22], [13.0629, 3.83251e-22, 3.73978e-22, 3.82848e-22, 3.76192e-22], [13.093, 3.79849e-22, 3.7058e-22, 3.79445e-22, 3.72793e-22], [13.1232, 3.7648e-22, 3.67215e-22, 3.76073e-22, 3.69427e-22], [13.1534, 3.73143e-22, 3.63882e-22, 3.72735e-22, 3.66093e-22], [13.1838, 3.69839e-22, 3.60581e-22, 3.69428e-22, 3.62791e-22], [13.2142, 3.66566e-22, 3.57312e-22, 3.66153e-22, 3.59521e-22], [13.2447, 3.63325e-22, 3.54075e-22, 3.6291e-22, 3.56283e-22], [13.2752, 3.60115e-22, 3.50869e-22, 3.59698e-22, 3.53076e-22], [13.3058, 3.56937e-22, 3.47695e-22, 3.56517e-22, 3.49901e-22], [13.3365, 3.53789e-22, 3.44551e-22, 3.53367e-22, 3.46756e-22], [13.3672, 3.50672e-22, 3.41438e-22, 3.50248e-22, 3.43642e-22], [13.3981, 3.47585e-22, 3.38356e-22, 3.47159e-22, 3.40559e-22], [13.429, 3.44528e-22, 3.35304e-22, 3.44101e-22, 3.37505e-22], [13.4599, 3.41502e-22, 3.32282e-22, 3.41072e-22, 3.34482e-22], [13.491, 3.38505e-22, 3.2929e-22, 3.38073e-22, 3.31488e-22], [13.5221, 3.35537e-22, 3.26327e-22, 3.35103e-22, 3.28525e-22], [13.5533, 3.32599e-22, 3.23394e-22, 3.32163e-22, 3.2559e-22], [13.5845, 3.2969e-22, 3.20489e-22, 3.29252e-22, 3.22685e-22], [13.6158, 3.2681e-22, 3.17614e-22, 3.26369e-22, 3.19808e-22], [13.6472, 3.23958e-22, 3.14768e-22, 3.23516e-22, 3.1696e-22], [13.6787, 3.21135e-22, 3.1195e-22, 3.20691e-22, 3.14141e-22], [13.7103, 3.18341e-22, 3.09161e-22, 3.17894e-22, 3.11351e-22], [13.7419, 3.15574e-22, 3.064e-22, 3.15125e-22, 3.08588e-22], [13.7736, 3.12836e-22, 3.03667e-22, 3.12385e-22, 3.05854e-22], [13.8053, 3.10125e-22, 3.00962e-22, 3.09672e-22, 3.03147e-22], [13.8372, 3.07442e-22, 2.98285e-22, 3.06987e-22, 3.00469e-22], [13.8691, 3.04786e-22, 2.95636e-22, 3.04329e-22, 2.97818e-22], [13.9011, 3.02158e-22, 2.93014e-22, 3.01699e-22, 2.95194e-22], [13.9331, 2.99557e-22, 2.90419e-22, 2.99096e-22, 2.92598e-22], [13.9652, 2.96984e-22, 2.87852e-22, 2.9652e-22, 2.90029e-22], [13.9974, 2.94437e-22, 2.85312e-22, 2.93971e-22, 2.87487e-22], [14.0297, 2.91917e-22, 2.82798e-22, 2.91449e-22, 2.84972e-22], [14.0621, 2.89424e-22, 2.80312e-22, 2.88954e-22, 2.82484e-22], [14.0945, 2.86957e-22, 2.77853e-22, 2.86485e-22, 2.80023e-22], [14.127, 2.84518e-22, 2.75421e-22, 2.84043e-22, 2.77589e-22], [14.1596, 2.82104e-22, 2.73015e-22, 2.81628e-22, 2.75181e-22], [14.1922, 2.79718e-22, 2.70636e-22, 2.79239e-22, 2.728e-22], [14.225, 2.77358e-22, 2.68284e-22, 2.76877e-22, 2.70446e-22], [14.2578, 2.75024e-22, 2.65959e-22, 2.74541e-22, 2.68119e-22], [14.2906, 2.72717e-22, 2.6366e-22, 2.72232e-22, 2.65818e-22], [14.3236, 2.70436e-22, 2.61388e-22, 2.6995e-22, 2.63543e-22], [14.3566, 2.68182e-22, 2.59143e-22, 2.67693e-22, 2.61296e-22], [14.3897, 2.65954e-22, 2.56925e-22, 2.65464e-22, 2.59075e-22], [14.4229, 2.63753e-22, 2.54733e-22, 2.63261e-22, 2.56881e-22], [14.4562, 2.61579e-22, 2.52569e-22, 2.61084e-22, 2.54714e-22], [14.4895, 2.59432e-22, 2.50432e-22, 2.58935e-22, 2.52574e-22], [14.5229, 2.57311e-22, 2.48322e-22, 2.56813e-22, 2.50462e-22], [14.5564, 2.55218e-22, 2.46239e-22, 2.54717e-22, 2.48376e-22], [14.59, 2.53152e-22, 2.44185e-22, 2.52649e-22, 2.46319e-22], [14.6236, 2.51113e-22, 2.42158e-22, 2.50609e-22, 2.44289e-22], [14.6573, 2.49103e-22, 2.40159e-22, 2.48596e-22, 2.42287e-22], [14.6911, 2.4712e-22, 2.38189e-22, 2.46612e-22, 2.40314e-22], [14.725, 2.45166e-22, 2.36248e-22, 2.44656e-22, 2.3837e-22], [14.759, 2.43241e-22, 2.34336e-22, 2.42729e-22, 2.36454e-22], [14.793, 2.41345e-22, 2.32455e-22, 2.40831e-22, 2.34569e-22], [14.8271, 2.39479e-22, 2.30603e-22, 2.38963e-22, 2.32714e-22], [14.8613, 2.37643e-22, 2.28783e-22, 2.37126e-22, 2.30889e-22], [14.8956, 2.35838e-22, 2.26994e-22, 2.3532e-22, 2.29096e-22], [14.9299, 2.34065e-22, 2.25237e-22, 2.33545e-22, 2.27336e-22], [14.9644, 2.32325e-22, 2.23514e-22, 2.31803e-22, 2.25608e-22], [14.9989, 2.30618e-22, 2.21825e-22, 2.30094e-22, 2.23914e-22], [15.0335, 2.28945e-22, 2.20171e-22, 2.28419e-22, 2.22255e-22], [15.0681, 2.27307e-22, 2.18554e-22, 2.26781e-22, 2.20633e-22], [15.1029, 2.25707e-22, 2.16974e-22, 2.25178e-22, 2.19048e-22], [15.1377, 2.24144e-22, 2.15433e-22, 2.23615e-22, 2.17501e-22], [15.1726, 2.22621e-22, 2.13933e-22, 2.2209e-22, 2.15996e-22], [15.2076, 2.21139e-22, 2.12476e-22, 2.20608e-22, 2.14532e-22], [15.2427, 2.19701e-22, 2.11063e-22, 2.19168e-22, 2.13113e-22], [15.2778, 2.18308e-22, 2.09697e-22, 2.17774e-22, 2.1174e-22], [15.3131, 2.16964e-22, 2.08381e-22, 2.16429e-22, 2.10417e-22], [15.3484, 2.15669e-22, 2.07117e-22, 2.15134e-22, 2.09145e-22], [15.3838, 2.14429e-22, 2.05909e-22, 2.13893e-22, 2.07929e-22], [15.4192, 2.13246e-22, 2.0476e-22, 2.12709e-22, 2.06772e-22], [15.4548, 2.12124e-22, 2.03675e-22, 2.11587e-22, 2.05677e-22], [15.4904, 2.11069e-22, 2.02657e-22, 2.10531e-22, 2.0465e-22], [15.5261, 2.10084e-22, 2.01714e-22, 2.09546e-22, 2.03696e-22], [15.5619, 2.09175e-22, 2.00849e-22, 2.08638e-22, 2.02821e-22], [15.5978, 2.08351e-22, 2.00072e-22, 2.07813e-22, 2.02031e-22], [15.6338, 2.07617e-22, 1.99389e-22, 2.0708e-22, 2.01336e-22], [15.6699, 2.06983e-22, 1.98809e-22, 2.06447e-22, 2.00743e-22], [15.706, 2.0646e-22, 1.98344e-22, 2.05925e-22, 2.00263e-22], [15.7422, 2.06058e-22, 1.98004e-22, 2.05524e-22, 1.99908e-22], [15.7785, 2.05791e-22, 1.97806e-22, 2.05259e-22, 1.99692e-22], [15.8149, 2.05676e-22, 1.97764e-22, 2.05146e-22, 1.99632e-22], [15.8514, 2.05731e-22, 1.97898e-22, 2.05204e-22, 1.99747e-22], [15.8879, 2.05978e-22, 1.98231e-22, 2.05454e-22, 2.00059e-22], [15.9246, 2.06443e-22, 1.9879e-22, 2.05922e-22, 2.00594e-22], [15.9613, 2.07157e-22, 1.99606e-22, 2.06641e-22, 2.01385e-22], [15.9981, 2.08158e-22, 2.00717e-22, 2.07646e-22, 2.02469e-22], [16.035, 2.09489e-22, 2.0217e-22, 2.08983e-22, 2.03892e-22], [16.072, 2.11206e-22, 2.04019e-22, 2.10706e-22, 2.05709e-22], [16.109, 2.13374e-22, 2.06332e-22, 2.12882e-22, 2.07987e-22], [16.1462, 2.16076e-22, 2.09192e-22, 2.15592e-22, 2.10809e-22], [16.1834, 2.19412e-22, 2.12703e-22, 2.18938e-22, 2.14277e-22], [16.2207, 2.23512e-22, 2.16994e-22, 2.23049e-22, 2.18522e-22], [16.2581, 2.28538e-22, 2.22229e-22, 2.28087e-22, 2.23706e-22], [16.2956, 2.34698e-22, 2.28619e-22, 2.34261e-22, 2.30041e-22], [16.3332, 2.42267e-22, 2.3644e-22, 2.41845e-22, 2.37801e-22], [16.3709, 2.51608e-22, 2.46056e-22, 2.51204e-22, 2.47352e-22], [16.4086, 2.63213e-22, 2.57963e-22, 2.62829e-22, 2.59187e-22], [16.4464, 2.77768e-22, 2.72846e-22, 2.77406e-22, 2.73992e-22], [16.4844, 2.96247e-22, 2.91682e-22, 2.95909e-22, 2.92744e-22], [16.5224, 3.20091e-22, 3.15912e-22, 3.1978e-22, 3.16883e-22], [16.5605, 3.51517e-22, 3.47753e-22, 3.51235e-22, 3.48626e-22], [16.5987, 3.94122e-22, 3.908e-22, 3.93871e-22, 3.91569e-22], [16.6369, 4.54151e-22, 4.51299e-22, 4.53935e-22, 4.51959e-22], [16.6753, 5.43501e-22, 5.41143e-22, 5.43321e-22, 5.41688e-22], [16.7138, 6.87979e-22, 6.86136e-22, 6.87838e-22, 6.86562e-22], [16.7523, 9.56104e-22, 9.54792e-22, 9.56003e-22, 9.55095e-22], [16.7909, 1.60994e-21, 1.60917e-21, 1.60988e-21, 1.60934e-21], [16.8297, 5.42605e-21, 5.42582e-21, 5.42603e-21, 5.42588e-21], [16.8685, 3.82716e-21, 3.82684e-21, 3.82713e-21, 3.82691e-21], [16.9074, 1.41349e-21, 1.41264e-21, 1.41343e-21, 1.41284e-21], [16.9464, 8.69075e-22, 8.677e-22, 8.68967e-22, 8.68016e-22], [16.9854, 6.3052e-22, 6.28641e-22, 6.30371e-22, 6.29074e-22], [17.0246, 4.97637e-22, 4.95278e-22, 4.97449e-22, 4.95821e-22], [17.0639, 4.13567e-22, 4.10752e-22, 4.13342e-22, 4.114e-22], [17.1032, 3.5598e-22, 3.52738e-22, 3.55719e-22, 3.53484e-22], [17.1427, 3.14319e-22, 3.10678e-22, 3.14026e-22, 3.11517e-22], [17.1822, 2.82948e-22, 2.78938e-22, 2.82624e-22, 2.79862e-22], [17.2218, 2.58585e-22, 2.54233e-22, 2.58231e-22, 2.55237e-22], [17.2615, 2.3919e-22, 2.34524e-22, 2.38809e-22, 2.35601e-22], [17.3013, 2.23431e-22, 2.18479e-22, 2.23026e-22, 2.19623e-22], [17.3412, 2.10405e-22, 2.05189e-22, 2.09976e-22, 2.06394e-22], [17.3812, 1.99474e-22, 1.94019e-22, 1.99024e-22, 1.9528e-22], [17.4213, 1.9018e-22, 1.84507e-22, 1.8971e-22, 1.85819e-22], [17.4615, 1.82186e-22, 1.76314e-22, 1.81698e-22, 1.77673e-22], [17.5017, 1.75236e-22, 1.69184e-22, 1.74731e-22, 1.70585e-22], [17.5421, 1.69137e-22, 1.62921e-22, 1.68616e-22, 1.6436e-22], [17.5825, 1.63737e-22, 1.57371e-22, 1.63201e-22, 1.58846e-22], [17.6231, 1.58917e-22, 1.52416e-22, 1.58368e-22, 1.53922e-22], [17.6637, 1.54584e-22, 1.47959e-22, 1.54023e-22, 1.49494e-22], [17.7045, 1.50661e-22, 1.43924e-22, 1.50088e-22, 1.45485e-22], [17.7453, 1.47088e-22, 1.40247e-22, 1.46503e-22, 1.41833e-22], [17.7862, 1.43813e-22, 1.36879e-22, 1.43218e-22, 1.38487e-22], [17.8272, 1.40795e-22, 1.33776e-22, 1.4019e-22, 1.35404e-22], [17.8683, 1.38001e-22, 1.30904e-22, 1.37387e-22, 1.32549e-22], [17.9095, 1.35401e-22, 1.28233e-22, 1.34778e-22, 1.29895e-22], [17.9508, 1.32971e-22, 1.25738e-22, 1.3234e-22, 1.27415e-22], [17.9922, 1.30691e-22, 1.23399e-22, 1.30052e-22, 1.25089e-22], [18.0337, 1.28544e-22, 1.21197e-22, 1.27897e-22, 1.229e-22], [18.0753, 1.26514e-22, 1.19118e-22, 1.25861e-22, 1.20832e-22], [18.117, 1.2459e-22, 1.17148e-22, 1.23929e-22, 1.18873e-22], [18.1588, 1.2276e-22, 1.15276e-22, 1.22092e-22, 1.1701e-22], [18.2006, 1.21014e-22, 1.13492e-22, 1.20341e-22, 1.15235e-22], [18.2426, 1.19345e-22, 1.11788e-22, 1.18665e-22, 1.13539e-22], [18.2847, 1.17745e-22, 1.10157e-22, 1.1706e-22, 1.11914e-22], [18.3268, 1.16208e-22, 1.08591e-22, 1.15517e-22, 1.10355e-22], [18.3691, 1.14729e-22, 1.07085e-22, 1.14033e-22, 1.08855e-22], [18.4115, 1.13302e-22, 1.05634e-22, 1.126e-22, 1.07409e-22], [18.4539, 1.11924e-22, 1.04233e-22, 1.11217e-22, 1.06013e-22], [18.4965, 1.1059e-22, 1.02879e-22, 1.09878e-22, 1.04663e-22], [18.5391, 1.09297e-22, 1.01568e-22, 1.0858e-22, 1.03355e-22], [18.5819, 1.08042e-22, 1.00297e-22, 1.07321e-22, 1.02087e-22], [18.6247, 1.06823e-22, 9.90622e-23, 1.06097e-22, 1.00856e-22], [18.6677, 1.05636e-22, 9.78621e-23, 1.04905e-22, 9.9658e-23], [18.7107, 1.0448e-22, 9.66941e-23, 1.03745e-22, 9.84921e-23], [18.7539, 1.03353e-22, 9.55561e-23, 1.02614e-22, 9.7356e-23], [18.7971, 1.02253e-22, 9.44463e-23, 1.01509e-22, 9.62476e-23], [18.8405, 1.01178e-22, 9.3363e-23, 1.0043e-22, 9.51655e-23], [18.8839, 1.00127e-22, 9.23046e-23, 9.93752e-23, 9.41081e-23], [18.9275, 9.9099e-23, 9.12698e-23, 9.83427e-23, 9.30739e-23], [18.9711, 9.8092e-23, 9.02572e-23, 9.73317e-23, 9.20617e-23], [19.0149, 9.71051e-23, 8.92657e-23, 9.63409e-23, 9.10705e-23], [19.0587, 9.61375e-23, 8.82943e-23, 9.53694e-23, 9.0099e-23], [19.1027, 9.51881e-23, 8.73419e-23, 9.44162e-23, 8.91464e-23], [19.1467, 9.42561e-23, 8.64077e-23, 9.34805e-23, 8.82117e-23], [19.1909, 9.33408e-23, 8.54908e-23, 9.25615e-23, 8.72942e-23], [19.2351, 9.24413e-23, 8.45904e-23, 9.16584e-23, 8.63931e-23], [19.2795, 9.15571e-23, 8.37059e-23, 9.07705e-23, 8.55077e-23], [19.3239, 9.06875e-23, 8.28367e-23, 8.98973e-23, 8.46373e-23], [19.3685, 8.98319e-23, 8.1982e-23, 8.90382e-23, 8.37813e-23], [19.4132, 8.89898e-23, 8.11413e-23, 8.81926e-23, 8.29393e-23], [19.4579, 8.81607e-23, 8.03142e-23, 8.736e-23, 8.21106e-23], [19.5028, 8.73441e-23, 7.95001e-23, 8.654e-23, 8.12948e-23], [19.5478, 8.65396e-23, 7.86986e-23, 8.57321e-23, 8.04914e-23], [19.5928, 8.57468e-23, 7.79092e-23, 8.49359e-23, 7.97001e-23], [19.638, 8.49653e-23, 7.71315e-23, 8.41511e-23, 7.89203e-23], [19.6833, 8.41948e-23, 7.63652e-23, 8.33772e-23, 7.81519e-23], [19.7287, 8.34349e-23, 7.561e-23, 8.2614e-23, 7.73943e-23], [19.7742, 8.26853e-23, 7.48654e-23, 8.18611e-23, 7.66473e-23], [19.8198, 8.19457e-23, 7.41313e-23, 8.11183e-23, 7.59106e-23], [19.8655, 8.12159e-23, 7.34072e-23, 8.03852e-23, 7.51839e-23], [19.9113, 8.04955e-23, 7.26929e-23, 7.96616e-23, 7.44669e-23], [19.9572, 7.97843e-23, 7.19882e-23, 7.89472e-23, 7.37594e-23], [20.0032, 7.90821e-23, 7.12928e-23, 7.82418e-23, 7.30611e-23], [20.0494, 7.83887e-23, 7.06065e-23, 7.75453e-23, 7.23717e-23], [20.0956, 7.77038e-23, 6.99291e-23, 7.68573e-23, 7.16911e-23], [20.1419, 7.70273e-23, 6.92602e-23, 7.61776e-23, 7.10191e-23], [20.1884, 7.6359e-23, 6.85999e-23, 7.55062e-23, 7.03554e-23], [20.2349, 7.56986e-23, 6.79478e-23, 7.48427e-23, 6.96999e-23], [20.2816, 7.5046e-23, 6.73037e-23, 7.4187e-23, 6.90524e-23], [20.3284, 7.44011e-23, 6.66676e-23, 7.3539e-23, 6.84127e-23], [20.3753, 7.37636e-23, 6.60392e-23, 7.28985e-23, 6.77807e-23], [20.4222, 7.31335e-23, 6.54184e-23, 7.22654e-23, 6.71561e-23], [20.4693, 7.25106e-23, 6.4805e-23, 7.16394e-23, 6.65389e-23], [20.5165, 7.18947e-23, 6.41988e-23, 7.10205e-23, 6.5929e-23], [20.5638, 7.12857e-23, 6.35999e-23, 7.04085e-23, 6.5326e-23], [20.6113, 7.06834e-23, 6.30079e-23, 6.98033e-23, 6.473e-23], [20.6588, 7.00879e-23, 6.24228e-23, 6.92047e-23, 6.41408e-23], [20.7064, 6.94989e-23, 6.18445e-23, 6.86127e-23, 6.35583e-23], [20.7542, 6.89163e-23, 6.12728e-23, 6.80272e-23, 6.29824e-23], [20.802, 6.834e-23, 6.07077e-23, 6.7448e-23, 6.24129e-23], [20.85, 6.77699e-23, 6.01489e-23, 6.6875e-23, 6.18498e-23], [20.8981, 6.72059e-23, 5.95965e-23, 6.63081e-23, 6.12929e-23], [20.9463, 6.6648e-23, 5.90503e-23, 6.57472e-23, 6.07421e-23], [20.9946, 6.60959e-23, 5.85101e-23, 6.51922e-23, 6.01973e-23], [21.043, 6.55497e-23, 5.7976e-23, 6.46431e-23, 5.96586e-23], [21.0915, 6.50091e-23, 5.74478e-23, 6.40997e-23, 5.91256e-23], [21.1402, 6.44743e-23, 5.69255e-23, 6.35619e-23, 5.85984e-23], [21.1889, 6.39449e-23, 5.64089e-23, 6.30298e-23, 5.80769e-23], [21.2378, 6.34211e-23, 5.58979e-23, 6.2503e-23, 5.7561e-23], [21.2868, 6.29026e-23, 5.53925e-23, 6.19817e-23, 5.70506e-23], [21.3358, 6.23895e-23, 5.48926e-23, 6.14657e-23, 5.65456e-23], [21.385, 6.18815e-23, 5.43981e-23, 6.0955e-23, 5.60459e-23], [21.4344, 6.13788e-23, 5.3909e-23, 6.04494e-23, 5.55516e-23], [21.4838, 6.08811e-23, 5.34251e-23, 5.9949e-23, 5.50624e-23], [21.5333, 6.03885e-23, 5.29465e-23, 5.94535e-23, 5.45783e-23], [21.583, 5.99008e-23, 5.24729e-23, 5.89631e-23, 5.40994e-23], [21.6327, 5.9418e-23, 5.20044e-23, 5.84775e-23, 5.36254e-23], [21.6826, 5.89401e-23, 5.15409e-23, 5.79967e-23, 5.31563e-23], [21.7326, 5.84668e-23, 5.10824e-23, 5.75208e-23, 5.26921e-23], [21.7828, 5.79983e-23, 5.06287e-23, 5.70495e-23, 5.22327e-23], [21.833, 5.75345e-23, 5.01798e-23, 5.65829e-23, 5.1778e-23], [21.8833, 5.70752e-23, 4.97356e-23, 5.61208e-23, 5.13281e-23], [21.9338, 5.66204e-23, 4.92962e-23, 5.56633e-23, 5.08827e-23], [21.9844, 5.61701e-23, 4.88613e-23, 5.52103e-23, 5.04419e-23], [22.0351, 5.57242e-23, 4.84311e-23, 5.47617e-23, 5.00056e-23], [22.0859, 5.52827e-23, 4.80053e-23, 5.43174e-23, 4.95737e-23], [22.1368, 5.48454e-23, 4.7584e-23, 5.38775e-23, 4.91463e-23], [22.1879, 5.44124e-23, 4.71672e-23, 5.34418e-23, 4.87231e-23], [22.239, 5.39836e-23, 4.67546e-23, 5.30103e-23, 4.83043e-23], [22.2903, 5.3559e-23, 4.63464e-23, 5.2583e-23, 4.78897e-23], [22.3417, 5.31384e-23, 4.59425e-23, 5.21598e-23, 4.74793e-23], [22.3932, 5.2722e-23, 4.55427e-23, 5.17406e-23, 4.70731e-23], [22.4449, 5.23095e-23, 4.51471e-23, 5.13255e-23, 4.66709e-23], [22.4966, 5.19009e-23, 4.47556e-23, 5.09143e-23, 4.62728e-23], [22.5485, 5.14963e-23, 4.43682e-23, 5.0507e-23, 4.58787e-23], [22.6005, 5.10956e-23, 4.39848e-23, 5.01037e-23, 4.54885e-23], [22.6526, 5.06986e-23, 4.36054e-23, 4.97041e-23, 4.51022e-23], [22.7049, 5.03055e-23, 4.32299e-23, 4.93083e-23, 4.47198e-23], [22.7572, 4.99161e-23, 4.28584e-23, 4.89163e-23, 4.43413e-23], [22.8097, 4.95304e-23, 4.24906e-23, 4.8528e-23, 4.39665e-23], [22.8623, 4.91483e-23, 4.21267e-23, 4.81434e-23, 4.35954e-23], [22.915, 4.87699e-23, 4.17665e-23, 4.77623e-23, 4.32281e-23], [22.9679, 4.8395e-23, 4.14101e-23, 4.73849e-23, 4.28644e-23], [23.0208, 4.80236e-23, 4.10573e-23, 4.7011e-23, 4.25043e-23], [23.0739, 4.76558e-23, 4.07082e-23, 4.66406e-23, 4.21478e-23], [23.1271, 4.72914e-23, 4.03627e-23, 4.62736e-23, 4.17948e-23], [23.1804, 4.69304e-23, 4.00208e-23, 4.59101e-23, 4.14454e-23], [23.2339, 4.65729e-23, 3.96824e-23, 4.555e-23, 4.10994e-23], [23.2875, 4.62186e-23, 3.93475e-23, 4.51932e-23, 4.07568e-23], [23.3412, 4.58677e-23, 3.9016e-23, 4.48398e-23, 4.04176e-23], [23.395, 4.552e-23, 3.8688e-23, 4.44896e-23, 4.00818e-23], [23.4489, 4.51756e-23, 3.83634e-23, 4.41427e-23, 3.97493e-23], [23.503, 4.48344e-23, 3.80421e-23, 4.3799e-23, 3.94201e-23], [23.5572, 4.44964e-23, 3.77242e-23, 4.34585e-23, 3.90941e-23], [23.6115, 4.41615e-23, 3.74095e-23, 4.31211e-23, 3.87713e-23], [23.666, 4.38297e-23, 3.70981e-23, 4.27868e-23, 3.84518e-23], [23.7206, 4.3501e-23, 3.67899e-23, 4.24556e-23, 3.81353e-23], [23.7753, 4.31753e-23, 3.64849e-23, 4.21275e-23, 3.7822e-23], [23.8301, 4.28526e-23, 3.6183e-23, 4.18024e-23, 3.75118e-23], [23.885, 4.2533e-23, 3.58843e-23, 4.14803e-23, 3.72046e-23], [23.9401, 4.22162e-23, 3.55887e-23, 4.11611e-23, 3.69005e-23], [23.9953, 4.19024e-23, 3.52962e-23, 4.08449e-23, 3.65994e-23], [24.0507, 4.15914e-23, 3.50066e-23, 4.05315e-23, 3.63012e-23], [24.1061, 4.12834e-23, 3.47201e-23, 4.0221e-23, 3.60059e-23], [24.1617, 4.09781e-23, 3.44366e-23, 3.99134e-23, 3.57136e-23], [24.2174, 4.06757e-23, 3.4156e-23, 3.96086e-23, 3.54241e-23], [24.2733, 4.0376e-23, 3.38783e-23, 3.93065e-23, 3.51374e-23], [24.3292, 4.0079e-23, 3.36035e-23, 3.90072e-23, 3.48536e-23], [24.3854, 3.97848e-23, 3.33316e-23, 3.87106e-23, 3.45726e-23], [24.4416, 3.94933e-23, 3.30625e-23, 3.84167e-23, 3.42943e-23], [24.4979, 3.92044e-23, 3.27962e-23, 3.81255e-23, 3.40188e-23], [24.5544, 3.89181e-23, 3.25327e-23, 3.78369e-23, 3.37459e-23], [24.6111, 3.86345e-23, 3.22719e-23, 3.7551e-23, 3.34758e-23], [24.6678, 3.83534e-23, 3.20139e-23, 3.72676e-23, 3.32083e-23], [24.7247, 3.80749e-23, 3.17586e-23, 3.69868e-23, 3.29434e-23], [24.7817, 3.77989e-23, 3.15059e-23, 3.67085e-23, 3.26811e-23], [24.8389, 3.75254e-23, 3.12559e-23, 3.64327e-23, 3.24214e-23], [24.8961, 3.72544e-23, 3.10085e-23, 3.61595e-23, 3.21642e-23], [24.9536, 3.69859e-23, 3.07637e-23, 3.58887e-23, 3.19096e-23], [25.0111, 3.67197e-23, 3.05215e-23, 3.56203e-23, 3.16575e-23], [25.0688, 3.6456e-23, 3.02818e-23, 3.53543e-23, 3.14078e-23], [25.1266, 3.61947e-23, 3.00447e-23, 3.50908e-23, 3.11606e-23], [25.1845, 3.59357e-23, 2.981e-23, 3.48296e-23, 3.09158e-23], [25.2426, 3.56791e-23, 2.95779e-23, 3.45707e-23, 3.06734e-23], [25.3008, 3.54247e-23, 2.93482e-23, 3.43142e-23, 3.04334e-23], [25.3592, 3.51727e-23, 2.91209e-23, 3.40599e-23, 3.01958e-23], [25.4176, 3.49229e-23, 2.8896e-23, 3.3808e-23, 2.99605e-23], [25.4762, 3.46754e-23, 2.86735e-23, 3.35583e-23, 2.97275e-23], [25.535, 3.443e-23, 2.84534e-23, 3.33108e-23, 2.94967e-23], [25.5939, 3.41869e-23, 2.82356e-23, 3.30655e-23, 2.92683e-23], [25.6529, 3.3946e-23, 2.80202e-23, 3.28224e-23, 2.90421e-23], [25.7121, 3.37072e-23, 2.7807e-23, 3.25815e-23, 2.88181e-23], [25.7713, 3.34705e-23, 2.75961e-23, 3.23428e-23, 2.85963e-23], [25.8308, 3.3236e-23, 2.73875e-23, 3.21061e-23, 2.83767e-23], [25.8903, 3.30035e-23, 2.71811e-23, 3.18716e-23, 2.81593e-23], [25.95, 3.27732e-23, 2.69769e-23, 3.16391e-23, 2.7944e-23], [26.0099, 3.25448e-23, 2.67749e-23, 3.14087e-23, 2.77308e-23], [26.0699, 3.23186e-23, 2.65751e-23, 3.11804e-23, 2.75197e-23], [26.13, 3.20943e-23, 2.63774e-23, 3.09541e-23, 2.73107e-23], [26.1902, 3.1872e-23, 2.61819e-23, 3.07297e-23, 2.71037e-23], [26.2506, 3.16517e-23, 2.59884e-23, 3.05074e-23, 2.68988e-23], [26.3112, 3.14334e-23, 2.57971e-23, 3.02871e-23, 2.66959e-23], [26.3718, 3.12169e-23, 2.56078e-23, 3.00686e-23, 2.6495e-23], [26.4326, 3.10024e-23, 2.54206e-23, 2.98521e-23, 2.6296e-23], [26.4936, 3.07898e-23, 2.52354e-23, 2.96376e-23, 2.60991e-23], [26.5547, 3.05791e-23, 2.50523e-23, 2.94249e-23, 2.59041e-23], [26.6159, 3.03703e-23, 2.48711e-23, 2.92141e-23, 2.57109e-23], [26.6773, 3.01633e-23, 2.46919e-23, 2.90051e-23, 2.55198e-23], [26.7388, 2.99581e-23, 2.45147e-23, 2.8798e-23, 2.53304e-23], [26.8005, 2.97547e-23, 2.43394e-23, 2.85927e-23, 2.5143e-23], [26.8623, 2.95531e-23, 2.41661e-23, 2.83892e-23, 2.49574e-23], [26.9242, 2.93533e-23, 2.39946e-23, 2.81875e-23, 2.47736e-23], [26.9863, 2.91553e-23, 2.3825e-23, 2.79876e-23, 2.45917e-23], [27.0486, 2.89589e-23, 2.36573e-23, 2.77894e-23, 2.44116e-23], [27.1109, 2.87644e-23, 2.34915e-23, 2.75929e-23, 2.42332e-23], [27.1734, 2.85715e-23, 2.33275e-23, 2.73982e-23, 2.40566e-23], [27.2361, 2.83803e-23, 2.31653e-23, 2.72052e-23, 2.38817e-23], [27.2989, 2.81908e-23, 2.30049e-23, 2.70138e-23, 2.37086e-23], [27.3619, 2.80029e-23, 2.28463e-23, 2.68242e-23, 2.35372e-23], [27.425, 2.78167e-23, 2.26895e-23, 2.66362e-23, 2.33675e-23], [27.4882, 2.76321e-23, 2.25344e-23, 2.64498e-23, 2.31995e-23], [27.5516, 2.74492e-23, 2.2381e-23, 2.6265e-23, 2.30331e-23], [27.6151, 2.72678e-23, 2.22294e-23, 2.60819e-23, 2.28684e-23], [27.6788, 2.7088e-23, 2.20795e-23, 2.59004e-23, 2.27053e-23], [27.7426, 2.69098e-23, 2.19312e-23, 2.57204e-23, 2.25439e-23], [27.8066, 2.67331e-23, 2.17846e-23, 2.5542e-23, 2.2384e-23], [27.8707, 2.6558e-23, 2.16397e-23, 2.53652e-23, 2.22257e-23], [27.935, 2.63843e-23, 2.14965e-23, 2.51898e-23, 2.2069e-23], [27.9994, 2.62122e-23, 2.13548e-23, 2.5016e-23, 2.19139e-23], [28.064, 2.60416e-23, 2.12148e-23, 2.48438e-23, 2.17603e-23], [28.1287, 2.58725e-23, 2.10763e-23, 2.4673e-23, 2.16082e-23], [28.1936, 2.57048e-23, 2.09394e-23, 2.45037e-23, 2.14577e-23], [28.2586, 2.55386e-23, 2.08041e-23, 2.43358e-23, 2.13086e-23], [28.3237, 2.53738e-23, 2.06704e-23, 2.41694e-23, 2.1161e-23], [28.3891, 2.52105e-23, 2.05381e-23, 2.40044e-23, 2.10149e-23], [28.4545, 2.50485e-23, 2.04074e-23, 2.38409e-23, 2.08703e-23], [28.5201, 2.4888e-23, 2.02782e-23, 2.36788e-23, 2.07271e-23], [28.5859, 2.47288e-23, 2.01505e-23, 2.3518e-23, 2.05853e-23], [28.6518, 2.4571e-23, 2.00243e-23, 2.33587e-23, 2.0445e-23], [28.7179, 2.44146e-23, 1.98995e-23, 2.32007e-23, 2.0306e-23], [28.7841, 2.42595e-23, 1.97761e-23, 2.30441e-23, 2.01685e-23], [28.8505, 2.41057e-23, 1.96542e-23, 2.28888e-23, 2.00323e-23], [28.917, 2.39533e-23, 1.95338e-23, 2.27349e-23, 1.98975e-23], [28.9837, 2.38022e-23, 1.94147e-23, 2.25823e-23, 1.9764e-23], [29.0506, 2.36524e-23, 1.9297e-23, 2.2431e-23, 1.96318e-23], [29.1175, 2.35038e-23, 1.91807e-23, 2.2281e-23, 1.9501e-23], [29.1847, 2.33565e-23, 1.90658e-23, 2.21323e-23, 1.93715e-23], [29.252, 2.32105e-23, 1.89522e-23, 2.19848e-23, 1.92433e-23], [29.3194, 2.30658e-23, 1.88399e-23, 2.18387e-23, 1.91164e-23], [29.3871, 2.29222e-23, 1.8729e-23, 2.16937e-23, 1.89907e-23], [29.4548, 2.27799e-23, 1.86194e-23, 2.155e-23, 1.88663e-23], [29.5227, 2.26389e-23, 1.85111e-23, 2.14076e-23, 1.87432e-23], [29.5908, 2.2499e-23, 1.84041e-23, 2.12663e-23, 1.86213e-23], [29.6591, 2.23603e-23, 1.82983e-23, 2.11263e-23, 1.85006e-23], [29.7275, 2.22228e-23, 1.81938e-23, 2.09875e-23, 1.83812e-23], [29.796, 2.20864e-23, 1.80906e-23, 2.08498e-23, 1.82629e-23], [29.8647, 2.19512e-23, 1.79886e-23, 2.07133e-23, 1.81458e-23], [29.9336, 2.18172e-23, 1.78878e-23, 2.0578e-23, 1.803e-23], [30.0026, 2.16843e-23, 1.77882e-23, 2.04438e-23, 1.79152e-23], [30.0718, 2.15525e-23, 1.76899e-23, 2.03108e-23, 1.78017e-23], [30.1411, 2.14218e-23, 1.75927e-23, 2.01789e-23, 1.76892e-23], [30.2106, 2.12923e-23, 1.74967e-23, 2.00481e-23, 1.7578e-23], [30.2803, 2.11638e-23, 1.74018e-23, 1.99184e-23, 1.74678e-23], [30.3501, 2.10365e-23, 1.73081e-23, 1.97898e-23, 1.73587e-23], [30.4201, 2.09102e-23, 1.72156e-23, 1.96623e-23, 1.72508e-23], [30.4903, 2.07849e-23, 1.71242e-23, 1.95359e-23, 1.71439e-23], [30.5606, 2.06608e-23, 1.70338e-23, 1.94106e-23, 1.70381e-23], [30.6311, 2.05376e-23, 1.69446e-23, 1.92863e-23, 1.69334e-23], [30.7017, 2.04155e-23, 1.68565e-23, 1.91631e-23, 1.68298e-23], [30.7725, 2.02945e-23, 1.67695e-23, 1.90409e-23, 1.67272e-23], [30.8435, 2.01744e-23, 1.66835e-23, 1.89198e-23, 1.66256e-23], [30.9146, 2.00554e-23, 1.65986e-23, 1.87996e-23, 1.65251e-23], [30.9859, 1.99373e-23, 1.65148e-23, 1.86805e-23, 1.64255e-23], [31.0573, 1.98203e-23, 1.64319e-23, 1.85624e-23, 1.6327e-23], [31.1289, 1.97042e-23, 1.63501e-23, 1.84453e-23, 1.62295e-23], [31.2007, 1.95891e-23, 1.62694e-23, 1.83292e-23, 1.6133e-23], [31.2727, 1.9475e-23, 1.61896e-23, 1.8214e-23, 1.60375e-23], [31.3448, 1.93618e-23, 1.61108e-23, 1.80999e-23, 1.59429e-23], [31.4171, 1.92495e-23, 1.6033e-23, 1.79866e-23, 1.58493e-23], [31.4895, 1.91382e-23, 1.59562e-23, 1.78744e-23, 1.57566e-23], [31.5621, 1.90279e-23, 1.58803e-23, 1.77631e-23, 1.56649e-23], [31.6349, 1.89184e-23, 1.58054e-23, 1.76527e-23, 1.55741e-23], [31.7079, 1.88099e-23, 1.57314e-23, 1.75432e-23, 1.54842e-23], [31.781, 1.87022e-23, 1.56583e-23, 1.74347e-23, 1.53953e-23], [31.8543, 1.85955e-23, 1.55862e-23, 1.7327e-23, 1.53072e-23], [31.9277, 1.84896e-23, 1.5515e-23, 1.72203e-23, 1.522e-23], [32.0014, 1.83846e-23, 1.54447e-23, 1.71145e-23, 1.51338e-23], [32.0752, 1.82805e-23, 1.53753e-23, 1.70095e-23, 1.50484e-23], [32.1491, 1.81773e-23, 1.53067e-23, 1.69054e-23, 1.49639e-23], [32.2233, 1.80749e-23, 1.5239e-23, 1.68022e-23, 1.48802e-23], [32.2976, 1.79733e-23, 1.51722e-23, 1.66999e-23, 1.47974e-23], [32.372, 1.78726e-23, 1.51062e-23, 1.65984e-23, 1.47154e-23], [32.4467, 1.77728e-23, 1.50411e-23, 1.64978e-23, 1.46343e-23], [32.5215, 1.76737e-23, 1.49768e-23, 1.6398e-23, 1.4554e-23], [32.5965, 1.75755e-23, 1.49134e-23, 1.6299e-23, 1.44745e-23], [32.6717, 1.74781e-23, 1.48507e-23, 1.62009e-23, 1.43958e-23], [32.747, 1.73815e-23, 1.47888e-23, 1.61035e-23, 1.43179e-23], [32.8225, 1.72856e-23, 1.47278e-23, 1.6007e-23, 1.42409e-23], [32.8982, 1.71906e-23, 1.46675e-23, 1.59113e-23, 1.41646e-23], [32.9741, 1.70964e-23, 1.4608e-23, 1.58164e-23, 1.4089e-23], [33.0501, 1.70029e-23, 1.45493e-23, 1.57223e-23, 1.40143e-23], [33.1263, 1.69102e-23, 1.44913e-23, 1.5629e-23, 1.39403e-23], [33.2027, 1.68182e-23, 1.44341e-23, 1.55364e-23, 1.38671e-23], [33.2793, 1.6727e-23, 1.43776e-23, 1.54446e-23, 1.37946e-23], [33.356, 1.66366e-23, 1.43218e-23, 1.53536e-23, 1.37228e-23], [33.433, 1.65469e-23, 1.42668e-23, 1.52633e-23, 1.36518e-23], [33.51, 1.64579e-23, 1.42125e-23, 1.51738e-23, 1.35815e-23], [33.5873, 1.63697e-23, 1.41589e-23, 1.5085e-23, 1.3512e-23], [33.6648, 1.62821e-23, 1.4106e-23, 1.4997e-23, 1.34431e-23], [33.7424, 1.61953e-23, 1.40537e-23, 1.49097e-23, 1.33749e-23], [33.8202, 1.61092e-23, 1.40022e-23, 1.48231e-23, 1.33075e-23], [33.8982, 1.60238e-23, 1.39513e-23, 1.47372e-23, 1.32407e-23], [33.9764, 1.59391e-23, 1.39011e-23, 1.4652e-23, 1.31746e-23], [34.0547, 1.58551e-23, 1.38516e-23, 1.45676e-23, 1.31092e-23], [34.1333, 1.57718e-23, 1.38026e-23, 1.44838e-23, 1.30444e-23], [34.212, 1.56891e-23, 1.37544e-23, 1.44008e-23, 1.29803e-23], [34.2909, 1.56072e-23, 1.37068e-23, 1.43184e-23, 1.29168e-23], [34.3699, 1.55258e-23, 1.36597e-23, 1.42367e-23, 1.2854e-23], [34.4492, 1.54452e-23, 1.36134e-23, 1.41557e-23, 1.27919e-23], [34.5286, 1.53652e-23, 1.35676e-23, 1.40753e-23, 1.27304e-23], [34.6083, 1.52858e-23, 1.35224e-23, 1.39956e-23, 1.26695e-23], [34.6881, 1.52071e-23, 1.34778e-23, 1.39166e-23, 1.26092e-23], [34.7681, 1.51291e-23, 1.34338e-23, 1.38383e-23, 1.25495e-23], [34.8482, 1.50516e-23, 1.33904e-23, 1.37605e-23, 1.24904e-23], [34.9286, 1.49748e-23, 1.33475e-23, 1.36834e-23, 1.2432e-23], [35.0091, 1.48986e-23, 1.33053e-23, 1.3607e-23, 1.23741e-23], [35.0899, 1.48231e-23, 1.32635e-23, 1.35312e-23, 1.23168e-23], [35.1708, 1.47481e-23, 1.32223e-23, 1.3456e-23, 1.22601e-23], [35.2519, 1.46737e-23, 1.31817e-23, 1.33814e-23, 1.2204e-23], [35.3332, 1.46e-23, 1.31416e-23, 1.33075e-23, 1.21485e-23], [35.4147, 1.45268e-23, 1.3102e-23, 1.32341e-23, 1.20935e-23], [35.4963, 1.44542e-23, 1.3063e-23, 1.31614e-23, 1.20391e-23], [35.5782, 1.43822e-23, 1.30245e-23, 1.30892e-23, 1.19852e-23], [35.6602, 1.43108e-23, 1.29864e-23, 1.30177e-23, 1.19319e-23], [35.7425, 1.424e-23, 1.29489e-23, 1.29467e-23, 1.18791e-23], [35.8249, 1.41697e-23, 1.29119e-23, 1.28764e-23, 1.18268e-23], [35.9075, 1.41e-23, 1.28754e-23, 1.28066e-23, 1.17751e-23], [35.9903, 1.40308e-23, 1.28393e-23, 1.27374e-23, 1.17239e-23], [36.0733, 1.39622e-23, 1.28037e-23, 1.26687e-23, 1.16732e-23], [36.1565, 1.38942e-23, 1.27686e-23, 1.26006e-23, 1.1623e-23], [36.2398, 1.38267e-23, 1.2734e-23, 1.25331e-23, 1.15733e-23], [36.3234, 1.37597e-23, 1.26998e-23, 1.24661e-23, 1.15242e-23], [36.4072, 1.36933e-23, 1.26661e-23, 1.23997e-23, 1.14755e-23], [36.4911, 1.36274e-23, 1.26328e-23, 1.23339e-23, 1.14273e-23], [36.5753, 1.35621e-23, 1.25999e-23, 1.22685e-23, 1.13796e-23], [36.6596, 1.34972e-23, 1.25675e-23, 1.22037e-23, 1.13324e-23], [36.7442, 1.34329e-23, 1.25355e-23, 1.21395e-23, 1.12857e-23], [36.8289, 1.33691e-23, 1.2504e-23, 1.20758e-23, 1.12394e-23], [36.9138, 1.33058e-23, 1.24728e-23, 1.20126e-23, 1.11936e-23], [36.999, 1.3243e-23, 1.24421e-23, 1.19499e-23, 1.11482e-23], [37.0843, 1.31807e-23, 1.24117e-23, 1.18877e-23, 1.11033e-23], [37.1698, 1.31189e-23, 1.23818e-23, 1.18261e-23, 1.10589e-23], [37.2555, 1.30576e-23, 1.23523e-23, 1.17649e-23, 1.10149e-23], [37.3414, 1.29968e-23, 1.23231e-23, 1.17043e-23, 1.09713e-23], [37.4275, 1.29365e-23, 1.22943e-23, 1.16441e-23, 1.09282e-23], [37.5138, 1.28766e-23, 1.2266e-23, 1.15845e-23, 1.08855e-23], [37.6003, 1.28172e-23, 1.22379e-23, 1.15253e-23, 1.08432e-23], [37.687, 1.27583e-23, 1.22103e-23, 1.14667e-23, 1.08014e-23], [37.7739, 1.26999e-23, 1.2183e-23, 1.14085e-23, 1.076e-23], [37.8611, 1.26419e-23, 1.21561e-23, 1.13508e-23, 1.07189e-23], [37.9484, 1.25844e-23, 1.21295e-23, 1.12935e-23, 1.06783e-23], [38.0359, 1.25274e-23, 1.21033e-23, 1.12368e-23, 1.06381e-23], [38.1236, 1.24707e-23, 1.20774e-23, 1.11805e-23, 1.05983e-23], [38.2115, 1.24146e-23, 1.20518e-23, 1.11246e-23, 1.05589e-23], [38.2996, 1.23589e-23, 1.20266e-23, 1.10693e-23, 1.05198e-23], [38.3879, 1.23036e-23, 1.20017e-23, 1.10144e-23, 1.04812e-23], [38.4765, 1.22488e-23, 1.19771e-23, 1.09599e-23, 1.04429e-23], [38.5652, 1.21944e-23, 1.19529e-23, 1.09059e-23, 1.0405e-23], [38.6541, 1.21404e-23, 1.19289e-23, 1.08523e-23, 1.03675e-23], [38.7433, 1.20869e-23, 1.19053e-23, 1.07992e-23, 1.03304e-23], [38.8326, 1.20338e-23, 1.1882e-23, 1.07465e-23, 1.02936e-23], [38.9221, 1.19811e-23, 1.1859e-23, 1.06943e-23, 1.02572e-23], [39.0119, 1.19288e-23, 1.18363e-23, 1.06424e-23, 1.02211e-23], [39.1019, 1.18769e-23, 1.18138e-23, 1.0591e-23, 1.01854e-23], [39.192, 1.18255e-23, 1.17917e-23, 1.05401e-23, 1.015e-23], [39.2824, 1.17745e-23, 1.17698e-23, 1.04895e-23, 1.0115e-23], [39.373, 1.17238e-23, 1.17483e-23, 1.04394e-23, 1.00803e-23], [39.4638, 1.16736e-23, 1.1727e-23, 1.03897e-23, 1.0046e-23], [39.5548, 1.16237e-23, 1.17059e-23, 1.03404e-23, 1.0012e-23], [39.646, 1.15743e-23, 1.16852e-23, 1.02915e-23, 9.97827e-24], [39.7374, 1.15252e-23, 1.16647e-23, 1.0243e-23, 9.94491e-24], [39.8291, 1.14765e-23, 1.16444e-23, 1.01949e-23, 9.91188e-24], [39.9209, 1.14282e-23, 1.16245e-23, 1.01472e-23, 9.87916e-24], [40.013, 1.13803e-23, 1.16047e-23, 1.00999e-23, 9.84676e-24], [40.1052, 1.13328e-23, 1.15853e-23, 1.0053e-23, 9.81467e-24], [40.1977, 1.12856e-23, 1.1566e-23, 1.00065e-23, 9.78288e-24], [40.2904, 1.12389e-23, 1.1547e-23, 9.96032e-24, 9.7514e-24], [40.3833, 1.11924e-23, 1.15283e-23, 9.91457e-24, 9.72022e-24], [40.4764, 1.11464e-23, 1.15098e-23, 9.8692e-24, 9.68934e-24], [40.5698, 1.11007e-23, 1.14915e-23, 9.8242e-24, 9.65875e-24], [40.6633, 1.10554e-23, 1.14734e-23, 9.77958e-24, 9.62845e-24], [40.7571, 1.10104e-23, 1.14556e-23, 9.73532e-24, 9.59844e-24], [40.8511, 1.09658e-23, 1.1438e-23, 9.69144e-24, 9.56872e-24], [40.9453, 1.09215e-23, 1.14206e-23, 9.64792e-24, 9.53927e-24], [41.0397, 1.08776e-23, 1.14034e-23, 9.60476e-24, 9.5101e-24], [41.1344, 1.0834e-23, 1.13864e-23, 9.56196e-24, 9.48121e-24], [41.2292, 1.07908e-23, 1.13697e-23, 9.51952e-24, 9.45259e-24], [41.3243, 1.07479e-23, 1.13531e-23, 9.47742e-24, 9.42423e-24], [41.4196, 1.07054e-23, 1.13368e-23, 9.43569e-24, 9.39615e-24], [41.5151, 1.06631e-23, 1.13206e-23, 9.39429e-24, 9.36832e-24], [41.6108, 1.06213e-23, 1.13047e-23, 9.35325e-24, 9.34075e-24], [41.7068, 1.05797e-23, 1.12889e-23, 9.31254e-24, 9.31344e-24], [41.803, 1.05385e-23, 1.12734e-23, 9.27218e-24, 9.28639e-24], [41.8994, 1.04976e-23, 1.1258e-23, 9.23215e-24, 9.25958e-24], [41.996, 1.0457e-23, 1.12428e-23, 9.19246e-24, 9.23302e-24], [42.0928, 1.04167e-23, 1.12278e-23, 9.15309e-24, 9.20671e-24], [42.1899, 1.03768e-23, 1.12129e-23, 9.11406e-24, 9.18063e-24], [42.2872, 1.03372e-23, 1.11983e-23, 9.07536e-24, 9.1548e-24], [42.3847, 1.02979e-23, 1.11838e-23, 9.03698e-24, 9.1292e-24], [42.4824, 1.02588e-23, 1.11695e-23, 8.99892e-24, 9.10384e-24], [42.5804, 1.02201e-23, 1.11553e-23, 8.96117e-24, 9.07871e-24], [42.6786, 1.01818e-23, 1.11414e-23, 8.92375e-24, 9.05381e-24], [42.777, 1.01437e-23, 1.11276e-23, 8.88664e-24, 9.02913e-24], [42.8757, 1.01059e-23, 1.11139e-23, 8.84984e-24, 9.00468e-24], [42.9745, 1.00684e-23, 1.11004e-23, 8.81335e-24, 8.98044e-24], [43.0736, 1.00312e-23, 1.10871e-23, 8.77717e-24, 8.95643e-24], [43.173, 9.99426e-24, 1.10739e-23, 8.74129e-24, 8.93263e-24], [43.2725, 9.95765e-24, 1.10609e-23, 8.70572e-24, 8.90905e-24], [43.3723, 9.92132e-24, 1.1048e-23, 8.67044e-24, 8.88567e-24], [43.4723, 9.88528e-24, 1.10353e-23, 8.63546e-24, 8.86251e-24], [43.5726, 9.84952e-24, 1.10227e-23, 8.60078e-24, 8.83955e-24], [43.673, 9.81404e-24, 1.10103e-23, 8.56639e-24, 8.81679e-24], [43.7737, 9.77884e-24, 1.0998e-23, 8.53229e-24, 8.79423e-24], [43.8747, 9.74393e-24, 1.09858e-23, 8.49847e-24, 8.77188e-24], [43.9759, 9.70928e-24, 1.09738e-23, 8.46495e-24, 8.74972e-24], [44.0773, 9.67491e-24, 1.09619e-23, 8.43171e-24, 8.72775e-24], [44.1789, 9.64081e-24, 1.09501e-23, 8.39875e-24, 8.70598e-24], [44.2808, 9.60698e-24, 1.09385e-23, 8.36606e-24, 8.6844e-24], [44.3829, 9.57342e-24, 1.0927e-23, 8.33366e-24, 8.66301e-24], [44.4853, 9.54012e-24, 1.09156e-23, 8.30153e-24, 8.6418e-24], [44.5878, 9.50709e-24, 1.09044e-23, 8.26968e-24, 8.62077e-24], [44.6907, 9.47431e-24, 1.08933e-23, 8.23809e-24, 8.59993e-24], [44.7937, 9.44179e-24, 1.08823e-23, 8.20677e-24, 8.57927e-24], [44.897, 9.40954e-24, 1.08714e-23, 8.17572e-24, 8.55878e-24], [45.0005, 9.37753e-24, 1.08606e-23, 8.14494e-24, 8.53847e-24], [45.1043, 9.34578e-24, 1.085e-23, 8.11442e-24, 8.51833e-24], [45.2083, 9.31428e-24, 1.08395e-23, 8.08416e-24, 8.49837e-24], [45.3126, 9.28303e-24, 1.0829e-23, 8.05415e-24, 8.47857e-24], [45.4171, 9.25203e-24, 1.08187e-23, 8.02441e-24, 8.45894e-24], [45.5218, 9.22127e-24, 1.08085e-23, 7.99491e-24, 8.43948e-24], [45.6268, 9.19076e-24, 1.07984e-23, 7.96567e-24, 8.42018e-24], [45.732, 9.16048e-24, 1.07885e-23, 7.93668e-24, 8.40105e-24], [45.8374, 9.13045e-24, 1.07786e-23, 7.90794e-24, 8.38207e-24], [45.9431, 9.10066e-24, 1.07688e-23, 7.87945e-24, 8.36325e-24], [46.0491, 9.0711e-24, 1.07591e-23, 7.8512e-24, 8.34459e-24], [46.1553, 9.04177e-24, 1.07495e-23, 7.82319e-24, 8.32608e-24], [46.2617, 9.01268e-24, 1.07401e-23, 7.79543e-24, 8.30773e-24], [46.3684, 8.98382e-24, 1.07307e-23, 7.7679e-24, 8.28952e-24], [46.4753, 8.95519e-24, 1.07214e-23, 7.74061e-24, 8.27147e-24], [46.5825, 8.92679e-24, 1.07122e-23, 7.71356e-24, 8.25356e-24], [46.6899, 8.89861e-24, 1.07031e-23, 7.68674e-24, 8.2358e-24], [46.7976, 8.87065e-24, 1.06941e-23, 7.66015e-24, 8.21819e-24], [46.9055, 8.84292e-24, 1.06852e-23, 7.63379e-24, 8.20071e-24], [47.0137, 8.81541e-24, 1.06763e-23, 7.60766e-24, 8.18338e-24], [47.1221, 8.78812e-24, 1.06676e-23, 7.58175e-24, 8.16619e-24], [47.2307, 8.76104e-24, 1.06589e-23, 7.55607e-24, 8.14914e-24], [47.3397, 8.73419e-24, 1.06504e-23, 7.53062e-24, 8.13222e-24], [47.4488, 8.70754e-24, 1.06419e-23, 7.50538e-24, 8.11544e-24], [47.5582, 8.68111e-24, 1.06335e-23, 7.48036e-24, 8.09879e-24], [47.6679, 8.65489e-24, 1.06251e-23, 7.45556e-24, 8.08227e-24], [47.7778, 8.62888e-24, 1.06169e-23, 7.43098e-24, 8.06588e-24], [47.888, 8.60307e-24, 1.06087e-23, 7.40661e-24, 8.04963e-24], [47.9984, 8.57747e-24, 1.06006e-23, 7.38245e-24, 8.0335e-24], [48.1091, 8.55208e-24, 1.05926e-23, 7.3585e-24, 8.01749e-24], [48.2201, 8.52689e-24, 1.05847e-23, 7.33477e-24, 8.00161e-24], [48.3313, 8.5019e-24, 1.05768e-23, 7.31124e-24, 7.98586e-24], [48.4427, 8.47712e-24, 1.0569e-23, 7.28791e-24, 7.97022e-24], [48.5544, 8.45253e-24, 1.05613e-23, 7.26479e-24, 7.95471e-24], [48.6664, 8.42814e-24, 1.05536e-23, 7.24188e-24, 7.93931e-24], [48.7786, 8.40394e-24, 1.05461e-23, 7.21916e-24, 7.92404e-24], [48.8911, 8.37994e-24, 1.05385e-23, 7.19664e-24, 7.90888e-24], [49.0038, 8.35613e-24, 1.05311e-23, 7.17432e-24, 7.89383e-24], [49.1168, 8.33251e-24, 1.05237e-23, 7.1522e-24, 7.8789e-24], [49.2301, 8.30909e-24, 1.05164e-23, 7.13027e-24, 7.86409e-24], [49.3436, 8.28585e-24, 1.05092e-23, 7.10853e-24, 7.84938e-24], [49.4574, 8.2628e-24, 1.0502e-23, 7.08699e-24, 7.83478e-24], [49.5715, 8.23993e-24, 1.04949e-23, 7.06564e-24, 7.8203e-24], [49.6858, 8.21725e-24, 1.04878e-23, 7.04447e-24, 7.80592e-24], [49.8004, 8.19475e-24, 1.04808e-23, 7.02349e-24, 7.79164e-24], [49.9152, 8.17244e-24, 1.04739e-23, 7.0027e-24, 7.77748e-24], [50.0303, 8.1503e-24, 1.0467e-23, 6.98209e-24, 7.76341e-24], [50.1457, 8.12834e-24, 1.04602e-23, 6.96166e-24, 7.74945e-24], [50.2613, 8.10657e-24, 1.04534e-23, 6.94142e-24, 7.7356e-24], [50.3772, 8.08496e-24, 1.04467e-23, 6.92135e-24, 7.72184e-24], [50.4934, 8.06354e-24, 1.04401e-23, 6.90146e-24, 7.70818e-24], [50.6098, 8.04228e-24, 1.04335e-23, 6.88175e-24, 7.69462e-24], [50.7265, 8.0212e-24, 1.04269e-23, 6.86221e-24, 7.68116e-24], [50.8435, 8.0003e-24, 1.04204e-23, 6.84285e-24, 7.6678e-24], [50.9607, 7.97956e-24, 1.0414e-23, 6.82366e-24, 7.65453e-24], [51.0783, 7.95899e-24, 1.04076e-23, 6.80464e-24, 7.64136e-24], [51.196, 7.93858e-24, 1.04013e-23, 6.78579e-24, 7.62827e-24], [51.3141, 7.91835e-24, 1.0395e-23, 6.76711e-24, 7.61529e-24], [51.4324, 7.89828e-24, 1.03888e-23, 6.74859e-24, 7.60239e-24], [51.551, 7.87837e-24, 1.03826e-23, 6.73024e-24, 7.58958e-24], [51.6699, 7.85863e-24, 1.03765e-23, 6.71206e-24, 7.57687e-24], [51.7891, 7.83904e-24, 1.03704e-23, 6.69404e-24, 7.56424e-24], [51.9085, 7.81962e-24, 1.03644e-23, 6.67617e-24, 7.5517e-24], [52.0282, 7.80036e-24, 1.03584e-23, 6.65847e-24, 7.53924e-24], [52.1482, 7.78126e-24, 1.03524e-23, 6.64093e-24, 7.52687e-24], [52.2684, 7.76231e-24, 1.03466e-23, 6.62355e-24, 7.51459e-24], [52.389, 7.74352e-24, 1.03407e-23, 6.60632e-24, 7.50239e-24], [52.5098, 7.72488e-24, 1.03349e-23, 6.58924e-24, 7.49027e-24], [52.6309, 7.7064e-24, 1.03291e-23, 6.57232e-24, 7.47824e-24], [52.7522, 7.68807e-24, 1.03234e-23, 6.55555e-24, 7.46629e-24], [52.8739, 7.66989e-24, 1.03177e-23, 6.53894e-24, 7.45441e-24], [52.9958, 7.65186e-24, 1.03121e-23, 6.52247e-24, 7.44262e-24], [53.118, 7.63398e-24, 1.03065e-23, 6.50615e-24, 7.4309e-24], [53.2405, 7.61624e-24, 1.03009e-23, 6.48998e-24, 7.41927e-24], [53.3633, 7.59866e-24, 1.02954e-23, 6.47395e-24, 7.40771e-24], [53.4863, 7.58122e-24, 1.029e-23, 6.45807e-24, 7.39622e-24], [53.6097, 7.56393e-24, 1.02845e-23, 6.44233e-24, 7.38481e-24], [53.7333, 7.54677e-24, 1.02791e-23, 6.42673e-24, 7.37348e-24], [53.8572, 7.52977e-24, 1.02738e-23, 6.41128e-24, 7.36222e-24], [53.9814, 7.5129e-24, 1.02684e-23, 6.39597e-24, 7.35103e-24], [54.1059, 7.49618e-24, 1.02632e-23, 6.38079e-24, 7.33992e-24], [54.2306, 7.47959e-24, 1.02579e-23, 6.36575e-24, 7.32888e-24], [54.3557, 7.46314e-24, 1.02527e-23, 6.35085e-24, 7.3179e-24], [54.481, 7.44683e-24, 1.02475e-23, 6.33608e-24, 7.307e-24], [54.6067, 7.43066e-24, 1.02424e-23, 6.32145e-24, 7.29617e-24], [54.7326, 7.41463e-24, 1.02373e-23, 6.30695e-24, 7.2854e-24], [54.8588, 7.39872e-24, 1.02322e-23, 6.29258e-24, 7.27471e-24], [54.9853, 7.38296e-24, 1.02272e-23, 6.27834e-24, 7.26408e-24], [55.1121, 7.36732e-24, 1.02222e-23, 6.26423e-24, 7.25352e-24], [55.2392, 7.35182e-24, 1.02172e-23, 6.25025e-24, 7.24302e-24], [55.3666, 7.33645e-24, 1.02122e-23, 6.2364e-24, 7.23259e-24], [55.4943, 7.3212e-24, 1.02073e-23, 6.22267e-24, 7.22222e-24], [55.6222, 7.30609e-24, 1.02025e-23, 6.20907e-24, 7.21192e-24], [55.7505, 7.29111e-24, 1.01976e-23, 6.19559e-24, 7.20168e-24], [55.8791, 7.27625e-24, 1.01928e-23, 6.18223e-24, 7.1915e-24], [56.0079, 7.26152e-24, 1.0188e-23, 6.169e-24, 7.18138e-24], [56.1371, 7.24691e-24, 1.01833e-23, 6.15589e-24, 7.17133e-24], [56.2665, 7.23243e-24, 1.01785e-23, 6.14289e-24, 7.16134e-24], [56.3963, 7.21807e-24, 1.01738e-23, 6.13002e-24, 7.1514e-24], [56.5263, 7.20384e-24, 1.01692e-23, 6.11726e-24, 7.14153e-24], [56.6567, 7.18972e-24, 1.01645e-23, 6.10462e-24, 7.13172e-24], [56.7873, 7.17573e-24, 1.01599e-23, 6.09209e-24, 7.12196e-24], [56.9183, 7.16186e-24, 1.01554e-23, 6.07968e-24, 7.11226e-24], [57.0495, 7.14811e-24, 1.01508e-23, 6.06738e-24, 7.10262e-24], [57.1811, 7.13447e-24, 1.01463e-23, 6.0552e-24, 7.09303e-24], [57.3129, 7.12096e-24, 1.01418e-23, 6.04312e-24, 7.0835e-24], [57.4451, 7.10756e-24, 1.01373e-23, 6.03116e-24, 7.07403e-24], [57.5776, 7.09427e-24, 1.01328e-23, 6.0193e-24, 7.06461e-24], [57.7104, 7.08111e-24, 1.01284e-23, 6.00755e-24, 7.05525e-24], [57.8434, 7.06805e-24, 1.0124e-23, 5.99592e-24, 7.04594e-24], [57.9768, 7.05511e-24, 1.01197e-23, 5.98438e-24, 7.03668e-24], [58.1105, 7.04228e-24, 1.01153e-23, 5.97295e-24, 7.02748e-24], [58.2445, 7.02957e-24, 1.0111e-23, 5.96163e-24, 7.01832e-24], [58.3788, 7.01696e-24, 1.01067e-23, 5.95041e-24, 7.00922e-24], [58.5135, 7.00447e-24, 1.01024e-23, 5.93929e-24, 7.00018e-24], [58.6484, 6.99209e-24, 1.00982e-23, 5.92828e-24, 6.99118e-24], [58.7836, 6.97981e-24, 1.00939e-23, 5.91736e-24, 6.98223e-24], [58.9192, 6.96764e-24, 1.00897e-23, 5.90655e-24, 6.97333e-24], [59.0551, 6.95558e-24, 1.00856e-23, 5.89583e-24, 6.96448e-24], [59.1912, 6.94363e-24, 1.00814e-23, 5.88521e-24, 6.95568e-24], [59.3277, 6.93178e-24, 1.00773e-23, 5.87469e-24, 6.94693e-24], [59.4645, 6.92004e-24, 1.00732e-23, 5.86426e-24, 6.93823e-24], [59.6017, 6.9084e-24, 1.00691e-23, 5.85393e-24, 6.92957e-24], [59.7391, 6.89687e-24, 1.0065e-23, 5.8437e-24, 6.92097e-24], [59.8769, 6.88544e-24, 1.0061e-23, 5.83355e-24, 6.9124e-24], [60.0149, 6.87411e-24, 1.00569e-23, 5.8235e-24, 6.90389e-24], [60.1533, 6.86288e-24, 1.00529e-23, 5.81354e-24, 6.89542e-24], [60.2921, 6.85175e-24, 1.00489e-23, 5.80368e-24, 6.88699e-24], [60.4311, 6.84073e-24, 1.0045e-23, 5.7939e-24, 6.87861e-24], [60.5704, 6.8298e-24, 1.0041e-23, 5.78421e-24, 6.87028e-24], [60.7101, 6.81897e-24, 1.00371e-23, 5.77461e-24, 6.86199e-24], [60.8501, 6.80824e-24, 1.00332e-23, 5.76509e-24, 6.85374e-24], [60.9904, 6.79761e-24, 1.00293e-23, 5.75567e-24, 6.84553e-24], [61.1311, 6.78707e-24, 1.00255e-23, 5.74633e-24, 6.83737e-24], [61.2721, 6.77663e-24, 1.00216e-23, 5.73707e-24, 6.82925e-24], [61.4133, 6.76629e-24, 1.00178e-23, 5.7279e-24, 6.82118e-24], [61.555, 6.75604e-24, 1.0014e-23, 5.71881e-24, 6.81314e-24], [61.6969, 6.74588e-24, 1.00102e-23, 5.7098e-24, 6.80515e-24], [61.8392, 6.73582e-24, 1.00064e-23, 5.70087e-24, 6.79719e-24], [61.9818, 6.72585e-24, 1.00027e-23, 5.69203e-24, 6.78928e-24], [62.1247, 6.71597e-24, 9.99892e-24, 5.68327e-24, 6.78141e-24], [62.268, 6.70619e-24, 9.99521e-24, 5.67458e-24, 6.77357e-24], [62.4116, 6.69649e-24, 9.99151e-24, 5.66597e-24, 6.76578e-24], [62.5555, 6.68689e-24, 9.98783e-24, 5.65744e-24, 6.75803e-24], [62.6997, 6.67738e-24, 9.98416e-24, 5.64899e-24, 6.75031e-24], [62.8443, 6.66795e-24, 9.98051e-24, 5.64062e-24, 6.74263e-24], [62.9892, 6.65861e-24, 9.97688e-24, 5.63232e-24, 6.735e-24], [63.1345, 6.64937e-24, 9.97327e-24, 5.62409e-24, 6.72739e-24], [63.2801, 6.64021e-24, 9.96967e-24, 5.61594e-24, 6.71983e-24], [63.426, 6.63113e-24, 9.96609e-24, 5.60786e-24, 6.7123e-24], [63.5723, 6.62214e-24, 9.96252e-24, 5.59985e-24, 6.70481e-24], [63.7189, 6.61324e-24, 9.95898e-24, 5.59192e-24, 6.69736e-24], [63.8658, 6.60442e-24, 9.95544e-24, 5.58406e-24, 6.68994e-24], [64.0131, 6.59569e-24, 9.95193e-24, 5.57626e-24, 6.68256e-24], [64.1607, 6.58704e-24, 9.94842e-24, 5.56854e-24, 6.67521e-24], [64.3087, 6.57848e-24, 9.94494e-24, 5.56088e-24, 6.6679e-24], [64.457, 6.56999e-24, 9.94147e-24, 5.5533e-24, 6.66063e-24], [64.6056, 6.56159e-24, 9.93801e-24, 5.54578e-24, 6.65339e-24], [64.7546, 6.55328e-24, 9.93457e-24, 5.53833e-24, 6.64618e-24], [64.9039, 6.54504e-24, 9.93115e-24, 5.53094e-24, 6.63901e-24], [65.0536, 6.53688e-24, 9.92774e-24, 5.52362e-24, 6.63187e-24], [65.2036, 6.52881e-24, 9.92434e-24, 5.51636e-24, 6.62476e-24], [65.3539, 6.52081e-24, 9.92096e-24, 5.50917e-24, 6.61769e-24], [65.5046, 6.51289e-24, 9.91759e-24, 5.50204e-24, 6.61065e-24], [65.6557, 6.50505e-24, 9.91424e-24, 5.49498e-24, 6.60364e-24], [65.8071, 6.49729e-24, 9.9109e-24, 5.48797e-24, 6.59666e-24], [65.9589, 6.48961e-24, 9.90757e-24, 5.48103e-24, 6.58972e-24], [66.111, 6.482e-24, 9.90426e-24, 5.47415e-24, 6.58281e-24], [66.2634, 6.47448e-24, 9.90096e-24, 5.46733e-24, 6.57593e-24], [66.4162, 6.46702e-24, 9.89768e-24, 5.46057e-24, 6.56908e-24], [66.5694, 6.45965e-24, 9.89441e-24, 5.45387e-24, 6.56226e-24], [66.7229, 6.45234e-24, 9.89115e-24, 5.44722e-24, 6.55547e-24], [66.8767, 6.44512e-24, 9.88791e-24, 5.44064e-24, 6.54871e-24], [67.031, 6.43796e-24, 9.88468e-24, 5.43411e-24, 6.54199e-24], [67.1855, 6.43089e-24, 9.88146e-24, 5.42763e-24, 6.53529e-24], [67.3405, 6.42388e-24, 9.87825e-24, 5.42122e-24, 6.52862e-24], [67.4958, 6.41695e-24, 9.87506e-24, 5.41486e-24, 6.52199e-24], [67.6514, 6.41009e-24, 9.87188e-24, 5.40855e-24, 6.51538e-24], [67.8074, 6.4033e-24, 9.86871e-24, 5.4023e-24, 6.5088e-24], [67.9638, 6.39658e-24, 9.86556e-24, 5.3961e-24, 6.50225e-24], [68.1205, 6.38994e-24, 9.86242e-24, 5.38995e-24, 6.49572e-24], [68.2776, 6.38336e-24, 9.85929e-24, 5.38386e-24, 6.48923e-24], [68.435, 6.37686e-24, 9.85617e-24, 5.37781e-24, 6.48277e-24], [68.5928, 6.37042e-24, 9.85306e-24, 5.37182e-24, 6.47633e-24], [68.751, 6.36406e-24, 9.84997e-24, 5.36588e-24, 6.46992e-24], [68.9096, 6.35776e-24, 9.84689e-24, 5.35999e-24, 6.46353e-24], [69.0685, 6.35153e-24, 9.84382e-24, 5.35415e-24, 6.45718e-24], [69.2277, 6.34537e-24, 9.84076e-24, 5.34836e-24, 6.45085e-24], [69.3874, 6.33928e-24, 9.83771e-24, 5.34262e-24, 6.44455e-24], [69.5474, 6.33325e-24, 9.83468e-24, 5.33692e-24, 6.43827e-24], [69.7078, 6.3273e-24, 9.83165e-24, 5.33127e-24, 6.43202e-24], [69.8685, 6.3214e-24, 9.82864e-24, 5.32567e-24, 6.4258e-24], [70.0296, 6.31558e-24, 9.82564e-24, 5.32012e-24, 6.4196e-24], [70.1911, 6.30982e-24, 9.82265e-24, 5.31461e-24, 6.41343e-24], [70.353, 6.30413e-24, 9.81967e-24, 5.30914e-24, 6.40728e-24], [70.5152, 6.2985e-24, 9.8167e-24, 5.30372e-24, 6.40116e-24], [70.6778, 6.29293e-24, 9.81375e-24, 5.29835e-24, 6.39507e-24], [70.8408, 6.28743e-24, 9.8108e-24, 5.29302e-24, 6.389e-24], [71.0042, 6.28199e-24, 9.80786e-24, 5.28773e-24, 6.38295e-24], [71.1679, 6.27662e-24, 9.80494e-24, 5.28249e-24, 6.37693e-24], [71.332, 6.27131e-24, 9.80202e-24, 5.27728e-24, 6.37093e-24], [71.4965, 6.26606e-24, 9.79912e-24, 5.27212e-24, 6.36496e-24], [71.6614, 6.26088e-24, 9.79623e-24, 5.267e-24, 6.35901e-24], [71.8266, 6.25576e-24, 9.79334e-24, 5.26193e-24, 6.35309e-24], [71.9923, 6.2507e-24, 9.79047e-24, 5.25689e-24, 6.34719e-24], [72.1583, 6.2457e-24, 9.78761e-24, 5.25189e-24, 6.34131e-24], [72.3247, 6.24076e-24, 9.78476e-24, 5.24693e-24, 6.33546e-24], [72.4914, 6.23588e-24, 9.78191e-24, 5.24201e-24, 6.32963e-24], [72.6586, 6.23106e-24, 9.77908e-24, 5.23713e-24, 6.32382e-24], [72.8262, 6.22631e-24, 9.77626e-24, 5.23229e-24, 6.31803e-24], [72.9941, 6.22161e-24, 9.77345e-24, 5.22748e-24, 6.31227e-24], [73.1624, 6.21697e-24, 9.77064e-24, 5.22271e-24, 6.30653e-24], [73.3311, 6.21239e-24, 9.76785e-24, 5.21798e-24, 6.30081e-24], [73.5002, 6.20787e-24, 9.76507e-24, 5.21329e-24, 6.29512e-24], [73.6697, 6.20341e-24, 9.7623e-24, 5.20863e-24, 6.28944e-24], [73.8396, 6.199e-24, 9.75953e-24, 5.204e-24, 6.28379e-24], [74.0099, 6.19466e-24, 9.75678e-24, 5.19942e-24, 6.27816e-24], [74.1806, 6.19037e-24, 9.75403e-24, 5.19486e-24, 6.27255e-24], [74.3516, 6.18614e-24, 9.7513e-24, 5.19034e-24, 6.26697e-24], [74.5231, 6.18196e-24, 9.74857e-24, 5.18585e-24, 6.2614e-24], [74.6949, 6.17784e-24, 9.74585e-24, 5.1814e-24, 6.25586e-24], [74.8672, 6.17378e-24, 9.74315e-24, 5.17698e-24, 6.25033e-24], [75.0398, 6.16978e-24, 9.74045e-24, 5.17259e-24, 6.24483e-24], [75.2129, 6.16583e-24, 9.73776e-24, 5.16824e-24, 6.23935e-24], [75.3863, 6.16193e-24, 9.73508e-24, 5.16392e-24, 6.23388e-24], [75.5601, 6.15809e-24, 9.73241e-24, 5.15962e-24, 6.22844e-24], [75.7344, 6.15431e-24, 9.72974e-24, 5.15536e-24, 6.22302e-24], [75.909, 6.15058e-24, 9.72709e-24, 5.15113e-24, 6.21762e-24], [76.0841, 6.1469e-24, 9.72444e-24, 5.14693e-24, 6.21224e-24], [76.2595, 6.14328e-24, 9.72181e-24, 5.14276e-24, 6.20688e-24], [76.4354, 6.13971e-24, 9.71918e-24, 5.13862e-24, 6.20153e-24], [76.6116, 6.1362e-24, 9.71656e-24, 5.13451e-24, 6.19621e-24], [76.7883, 6.13274e-24, 9.71395e-24, 5.13042e-24, 6.19091e-24], [76.9654, 6.12933e-24, 9.71135e-24, 5.12637e-24, 6.18562e-24], [77.1429, 6.12597e-24, 9.70876e-24, 5.12234e-24, 6.18036e-24], [77.3208, 6.12267e-24, 9.70617e-24, 5.11834e-24, 6.17511e-24], [77.4991, 6.11942e-24, 9.7036e-24, 5.11437e-24, 6.16989e-24], [77.6778, 6.11622e-24, 9.70103e-24, 5.11042e-24, 6.16468e-24], [77.8569, 6.11308e-24, 9.69847e-24, 5.1065e-24, 6.15949e-24], [78.0364, 6.10998e-24, 9.69592e-24, 5.10261e-24, 6.15432e-24], [78.2164, 6.10694e-24, 9.69338e-24, 5.09874e-24, 6.14917e-24], [78.3968, 6.10395e-24, 9.69084e-24, 5.0949e-24, 6.14403e-24], [78.5775, 6.101e-24, 9.68832e-24, 5.09108e-24, 6.13892e-24], [78.7587, 6.09811e-24, 9.6858e-24, 5.08729e-24, 6.13382e-24], [78.9404, 6.09527e-24, 9.68329e-24, 5.08352e-24, 6.12874e-24], [79.1224, 6.09248e-24, 9.68079e-24, 5.07978e-24, 6.12367e-24], [79.3049, 6.08974e-24, 9.67829e-24, 5.07606e-24, 6.11863e-24], [79.4877, 6.08705e-24, 9.67581e-24, 5.07237e-24, 6.1136e-24], [79.671, 6.08441e-24, 9.67333e-24, 5.06869e-24, 6.10859e-24], [79.8548, 6.08182e-24, 9.67086e-24, 5.06504e-24, 6.1036e-24], [80.0389, 6.07928e-24, 9.6684e-24, 5.06142e-24, 6.09863e-24], [80.2235, 6.07678e-24, 9.66594e-24, 5.05781e-24, 6.09367e-24], [80.4085, 6.07434e-24, 9.6635e-24, 5.05423e-24, 6.08873e-24], [80.5939, 6.07194e-24, 9.66106e-24, 5.05067e-24, 6.0838e-24], [80.7797, 6.06959e-24, 9.65863e-24, 5.04713e-24, 6.07889e-24], [80.966, 6.06729e-24, 9.6562e-24, 5.04361e-24, 6.074e-24], [81.1527, 6.06504e-24, 9.65379e-24, 5.04011e-24, 6.06913e-24], [81.3399, 6.06284e-24, 9.65138e-24, 5.03663e-24, 6.06427e-24], [81.5274, 6.06068e-24, 9.64898e-24, 5.03318e-24, 6.05943e-24], [81.7154, 6.05857e-24, 9.64659e-24, 5.02974e-24, 6.05461e-24], [81.9039, 6.05651e-24, 9.6442e-24, 5.02632e-24, 6.0498e-24], [82.0927, 6.05449e-24, 9.64182e-24, 5.02292e-24, 6.045e-24], [82.282, 6.05252e-24, 9.63945e-24, 5.01954e-24, 6.04023e-24], [82.4718, 6.0506e-24, 9.63709e-24, 5.01618e-24, 6.03547e-24], [82.662, 6.04873e-24, 9.63473e-24, 5.01284e-24, 6.03072e-24], [82.8526, 6.0469e-24, 9.63238e-24, 5.00951e-24, 6.02599e-24], [83.0437, 6.04511e-24, 9.63004e-24, 5.00621e-24, 6.02128e-24], [83.2352, 6.04338e-24, 9.62771e-24, 5.00292e-24, 6.01658e-24], [83.4271, 6.04168e-24, 9.62538e-24, 4.99965e-24, 6.0119e-24], [83.6195, 6.04004e-24, 9.62306e-24, 4.99639e-24, 6.00723e-24], [83.8123, 6.03844e-24, 9.62075e-24, 4.99316e-24, 6.00258e-24], [84.0056, 6.03688e-24, 9.61845e-24, 4.98994e-24, 5.99794e-24], [84.1993, 6.03537e-24, 9.61615e-24, 4.98673e-24, 5.99332e-24], [84.3935, 6.03391e-24, 9.61386e-24, 4.98354e-24, 5.98871e-24], [84.5881, 6.03249e-24, 9.61157e-24, 4.98037e-24, 5.98412e-24], [84.7831, 6.03111e-24, 9.6093e-24, 4.97721e-24, 5.97954e-24], [84.9786, 6.02978e-24, 9.60703e-24, 4.97407e-24, 5.97498e-24], [85.1746, 6.02849e-24, 9.60476e-24, 4.97095e-24, 5.97043e-24], [85.371, 6.02725e-24, 9.60251e-24, 4.96783e-24, 5.9659e-24], [85.5679, 6.02605e-24, 9.60026e-24, 4.96474e-24, 5.96138e-24], [85.7652, 6.0249e-24, 9.59802e-24, 4.96166e-24, 5.95687e-24], [85.963, 6.02379e-24, 9.59578e-24, 4.95859e-24, 5.95238e-24], [86.1612, 6.02272e-24, 9.59355e-24, 4.95553e-24, 5.9479e-24], [86.3599, 6.0217e-24, 9.59133e-24, 4.95249e-24, 5.94344e-24], [86.559, 6.02072e-24, 9.58912e-24, 4.94947e-24, 5.93899e-24], [86.7586, 6.01978e-24, 9.58691e-24, 4.94645e-24, 5.93456e-24], [86.9587, 6.01888e-24, 9.58471e-24, 4.94345e-24, 5.93014e-24], [87.1592, 6.01803e-24, 9.58252e-24, 4.94046e-24, 5.92573e-24], [87.3602, 6.01722e-24, 9.58033e-24, 4.93749e-24, 5.92134e-24], [87.5617, 6.01646e-24, 9.57815e-24, 4.93453e-24, 5.91696e-24], [87.7636, 6.01574e-24, 9.57597e-24, 4.93157e-24, 5.9126e-24], [87.966, 6.01506e-24, 9.57381e-24, 4.92864e-24, 5.90824e-24], [88.1688, 6.01442e-24, 9.57165e-24, 4.92571e-24, 5.90391e-24], [88.3722, 6.01382e-24, 9.56949e-24, 4.92279e-24, 5.89958e-24], [88.5759, 6.01327e-24, 9.56734e-24, 4.91989e-24, 5.89527e-24], [88.7802, 6.01276e-24, 9.5652e-24, 4.917e-24, 5.89097e-24], [88.9849, 6.01229e-24, 9.56307e-24, 4.91411e-24, 5.88668e-24], [89.1901, 6.01186e-24, 9.56094e-24, 4.91124e-24, 5.88241e-24], [89.3958, 6.01147e-24, 9.55882e-24, 4.90838e-24, 5.87815e-24], [89.6019, 6.01113e-24, 9.5567e-24, 4.90553e-24, 5.87391e-24], [89.8086, 6.01083e-24, 9.55459e-24, 4.90269e-24, 5.86967e-24], [90.0157, 6.01056e-24, 9.55249e-24, 4.89986e-24, 5.86545e-24], [90.2232, 6.01034e-24, 9.5504e-24, 4.89704e-24, 5.86124e-24], [90.4313, 6.01016e-24, 9.54831e-24, 4.89423e-24, 5.85705e-24], [90.6398, 6.01003e-24, 9.54622e-24, 4.89142e-24, 5.85287e-24], [90.8489, 6.00993e-24, 9.54415e-24, 4.88863e-24, 5.8487e-24], [91.0583, 6.00987e-24, 9.54208e-24, 4.88585e-24, 5.84454e-24], [91.2683, 6.00986e-24, 9.54001e-24, 4.88307e-24, 5.84039e-24], [91.4788, 6.00988e-24, 9.53795e-24, 4.88031e-24, 5.83626e-24], [91.6897, 6.00995e-24, 9.5359e-24, 4.87755e-24, 5.83214e-24], [91.9012, 6.01005e-24, 9.53386e-24, 4.8748e-24, 5.82803e-24], [92.1131, 6.0102e-24, 9.53182e-24, 4.87206e-24, 5.82393e-24], [92.3255, 6.01039e-24, 9.52979e-24, 4.86932e-24, 5.81985e-24], [92.5384, 6.01061e-24, 9.52776e-24, 4.8666e-24, 5.81578e-24], [92.7518, 6.01088e-24, 9.52574e-24, 4.86388e-24, 5.81172e-24], [92.9657, 6.01119e-24, 9.52372e-24, 4.86117e-24, 5.80767e-24], [93.1801, 6.01154e-24, 9.52172e-24, 4.85847e-24, 5.80363e-24], [93.395, 6.01192e-24, 9.51971e-24, 4.85577e-24, 5.79961e-24], [93.6103, 6.01235e-24, 9.51772e-24, 4.85308e-24, 5.79559e-24], [93.8262, 6.01282e-24, 9.51573e-24, 4.8504e-24, 5.79159e-24], [94.0426, 6.01332e-24, 9.51374e-24, 4.84772e-24, 5.7876e-24], [94.2594, 6.01387e-24, 9.51177e-24, 4.84505e-24, 5.78363e-24], [94.4768, 6.01446e-24, 9.5098e-24, 4.84239e-24, 5.77966e-24], [94.6946, 6.01508e-24, 9.50783e-24, 4.83973e-24, 5.7757e-24], [94.913, 6.01575e-24, 9.50587e-24, 4.83708e-24, 5.77176e-24], [95.1319, 6.01645e-24, 9.50392e-24, 4.83443e-24, 5.76783e-24], [95.3513, 6.01719e-24, 9.50197e-24, 4.83179e-24, 5.76391e-24], [95.5711, 6.01798e-24, 9.50003e-24, 4.82916e-24, 5.76e-24], [95.7915, 6.0188e-24, 9.49809e-24, 4.82653e-24, 5.7561e-24], [96.0124, 6.01966e-24, 9.49616e-24, 4.82391e-24, 5.75221e-24], [96.2338, 6.02056e-24, 9.49424e-24, 4.82129e-24, 5.74833e-24], [96.4557, 6.0215e-24, 9.49232e-24, 4.81867e-24, 5.74447e-24], [96.6782, 6.02248e-24, 9.49041e-24, 4.81606e-24, 5.74061e-24], [96.9011, 6.02349e-24, 9.4885e-24, 4.81346e-24, 5.73677e-24], [97.1246, 6.02455e-24, 9.4866e-24, 4.81086e-24, 5.73294e-24], [97.3485, 6.02564e-24, 9.48471e-24, 4.80826e-24, 5.72912e-24], [97.573, 6.02678e-24, 9.48282e-24, 4.80567e-24, 5.72531e-24], [97.798, 6.02795e-24, 9.48094e-24, 4.80309e-24, 5.72151e-24], [98.0236, 6.02916e-24, 9.47906e-24, 4.8005e-24, 5.71772e-24], [98.2496, 6.03041e-24, 9.47719e-24, 4.79792e-24, 5.71394e-24], [98.4762, 6.03169e-24, 9.47533e-24, 4.79535e-24, 5.71017e-24], [98.7032, 6.03302e-24, 9.47347e-24, 4.79277e-24, 5.70641e-24], [98.9309, 6.03438e-24, 9.47161e-24, 4.79021e-24, 5.70266e-24], [99.159, 6.03579e-24, 9.46976e-24, 4.78764e-24, 5.69893e-24], [99.3877, 6.03723e-24, 9.46792e-24, 4.78508e-24, 5.6952e-24], [99.6168, 6.03871e-24, 9.46609e-24, 4.78252e-24, 5.69148e-24], [99.8466, 6.04022e-24, 9.46426e-24, 4.77996e-24, 5.68778e-24], [100.077, 6.04178e-24, 9.46243e-24, 4.77741e-24, 5.68408e-24], [100.308, 6.04337e-24, 9.46061e-24, 4.77486e-24, 5.6804e-24], [100.539, 6.045e-24, 9.4588e-24, 4.77231e-24, 5.67672e-24], [100.771, 6.04667e-24, 9.45699e-24, 4.76977e-24, 5.67306e-24], [101.003, 6.04838e-24, 9.45519e-24, 4.76722e-24, 5.6694e-24], [101.236, 6.05013e-24, 9.45339e-24, 4.76468e-24, 5.66576e-24], [101.469, 6.05191e-24, 9.4516e-24, 4.76214e-24, 5.66212e-24], [101.703, 6.05373e-24, 9.44982e-24, 4.7596e-24, 5.6585e-24], [101.938, 6.05559e-24, 9.44804e-24, 4.75707e-24, 5.65488e-24], [102.173, 6.05749e-24, 9.44626e-24, 4.75454e-24, 5.65128e-24], [102.409, 6.05942e-24, 9.44449e-24, 4.752e-24, 5.64768e-24], [102.645, 6.06139e-24, 9.44273e-24, 4.74947e-24, 5.6441e-24], [102.882, 6.0634e-24, 9.44097e-24, 4.74694e-24, 5.64052e-24], [103.119, 6.06545e-24, 9.43922e-24, 4.74442e-24, 5.63695e-24], [103.357, 6.06754e-24, 9.43748e-24, 4.74189e-24, 5.6334e-24], [103.595, 6.06966e-24, 9.43574e-24, 4.73936e-24, 5.62985e-24], [103.834, 6.07182e-24, 9.434e-24, 4.73684e-24, 5.62631e-24], [104.073, 6.07402e-24, 9.43227e-24, 4.73432e-24, 5.62278e-24], [104.313, 6.07626e-24, 9.43055e-24, 4.73179e-24, 5.61927e-24], [104.554, 6.07853e-24, 9.42883e-24, 4.72927e-24, 5.61576e-24], [104.795, 6.08085e-24, 9.42712e-24, 4.72675e-24, 5.61226e-24], [105.037, 6.08319e-24, 9.42541e-24, 4.72423e-24, 5.60877e-24], [105.279, 6.08558e-24, 9.42371e-24, 4.72171e-24, 5.60529e-24], [105.522, 6.08801e-24, 9.42201e-24, 4.71919e-24, 5.60182e-24], [105.765, 6.09047e-24, 9.42032e-24, 4.71667e-24, 5.59835e-24], [106.009, 6.09297e-24, 9.41863e-24, 4.71415e-24, 5.5949e-24], [106.253, 6.0955e-24, 9.41695e-24, 4.71163e-24, 5.59146e-24], [106.498, 6.09808e-24, 9.41528e-24, 4.70911e-24, 5.58802e-24], [106.744, 6.10069e-24, 9.41361e-24, 4.70659e-24, 5.5846e-24], [106.99, 6.10334e-24, 9.41195e-24, 4.70407e-24, 5.58118e-24], [107.237, 6.10602e-24, 9.41029e-24, 4.70155e-24, 5.57777e-24], [107.484, 6.10875e-24, 9.40864e-24, 4.69902e-24, 5.57438e-24], [107.732, 6.11151e-24, 9.40699e-24, 4.6965e-24, 5.57099e-24], [107.98, 6.1143e-24, 9.40535e-24, 4.69398e-24, 5.56761e-24], [108.229, 6.11714e-24, 9.40371e-24, 4.69146e-24, 5.56424e-24], [108.479, 6.12001e-24, 9.40208e-24, 4.68893e-24, 5.56087e-24], [108.729, 6.12292e-24, 9.40045e-24, 4.68641e-24, 5.55752e-24], [108.98, 6.12587e-24, 9.39883e-24, 4.68388e-24, 5.55418e-24], [109.231, 6.12886e-24, 9.39722e-24, 4.68135e-24, 5.55084e-24], [109.483, 6.13188e-24, 9.39561e-24, 4.67882e-24, 5.54751e-24], [109.735, 6.13494e-24, 9.394e-24, 4.67629e-24, 5.54419e-24], [109.988, 6.13803e-24, 9.3924e-24, 4.67376e-24, 5.54088e-24], [110.242, 6.14117e-24, 9.39081e-24, 4.67123e-24, 5.53758e-24], [110.496, 6.14434e-24, 9.38922e-24, 4.6687e-24, 5.53429e-24], [110.751, 6.14755e-24, 9.38764e-24, 4.66616e-24, 5.53101e-24], [111.007, 6.15079e-24, 9.38606e-24, 4.66362e-24, 5.52773e-24], [111.262, 6.15408e-24, 9.38449e-24, 4.66108e-24, 5.52447e-24], [111.519, 6.1574e-24, 9.38292e-24, 4.65854e-24, 5.52121e-24], [111.776, 6.16075e-24, 9.38136e-24, 4.656e-24, 5.51796e-24], [112.034, 6.16415e-24, 9.3798e-24, 4.65346e-24, 5.51472e-24], [112.292, 6.16758e-24, 9.37825e-24, 4.65091e-24, 5.51148e-24], [112.551, 6.17105e-24, 9.3767e-24, 4.64836e-24, 5.50826e-24], [112.811, 6.17455e-24, 9.37516e-24, 4.64581e-24, 5.50504e-24], [113.071, 6.1781e-24, 9.37362e-24, 4.64325e-24, 5.50184e-24], [113.332, 6.18168e-24, 9.37209e-24, 4.6407e-24, 5.49864e-24], [113.593, 6.1853e-24, 9.37057e-24, 4.63814e-24, 5.49545e-24], [113.855, 6.18895e-24, 9.36905e-24, 4.63558e-24, 5.49226e-24], [114.118, 6.19265e-24, 9.36753e-24, 4.63302e-24, 5.48909e-24], [114.381, 6.19638e-24, 9.36602e-24, 4.63045e-24, 5.48592e-24], [114.644, 6.20014e-24, 9.36452e-24, 4.62788e-24, 5.48276e-24], [114.909, 6.20395e-24, 9.36302e-24, 4.62531e-24, 5.47961e-24], [115.174, 6.20779e-24, 9.36153e-24, 4.62274e-24, 5.47647e-24], [115.439, 6.21167e-24, 9.36004e-24, 4.62016e-24, 5.47334e-24], [115.706, 6.21559e-24, 9.35855e-24, 4.61758e-24, 5.47021e-24], [115.972, 6.21954e-24, 9.35708e-24, 4.615e-24, 5.46709e-24], [116.24, 6.22353e-24, 9.3556e-24, 4.61241e-24, 5.46398e-24], [116.508, 6.22756e-24, 9.35413e-24, 4.60982e-24, 5.46088e-24], [116.777, 6.23162e-24, 9.35267e-24, 4.60723e-24, 5.45779e-24], [117.046, 6.23573e-24, 9.35121e-24, 4.60463e-24, 5.4547e-24], [117.316, 6.23987e-24, 9.34976e-24, 4.60203e-24, 5.45162e-24], [117.586, 6.24405e-24, 9.34831e-24, 4.59943e-24, 5.44855e-24], [117.857, 6.24826e-24, 9.34687e-24, 4.59682e-24, 5.44549e-24], [118.129, 6.25251e-24, 9.34544e-24, 4.59421e-24, 5.44244e-24], [118.402, 6.2568e-24, 9.344e-24, 4.5916e-24, 5.43939e-24], [118.675, 6.26113e-24, 9.34258e-24, 4.58898e-24, 5.43635e-24], [118.948, 6.2655e-24, 9.34116e-24, 4.58636e-24, 5.43332e-24], [119.223, 6.2699e-24, 9.33974e-24, 4.58374e-24, 5.4303e-24], [119.498, 6.27434e-24, 9.33833e-24, 4.58111e-24, 5.42728e-24], [119.773, 6.27882e-24, 9.33692e-24, 4.57848e-24, 5.42427e-24], [120.049, 6.28333e-24, 9.33552e-24, 4.57584e-24, 5.42127e-24], [120.326, 6.28788e-24, 9.33412e-24, 4.5732e-24, 5.41828e-24], [120.604, 6.29247e-24, 9.33273e-24, 4.57056e-24, 5.41529e-24], [120.882, 6.2971e-24, 9.33135e-24, 4.56791e-24, 5.41232e-24], [121.161, 6.30177e-24, 9.32997e-24, 4.56526e-24, 5.40935e-24], [121.44, 6.30647e-24, 9.32859e-24, 4.5626e-24, 5.40638e-24], [121.72, 6.31121e-24, 9.32722e-24, 4.55994e-24, 5.40343e-24], [122.001, 6.31599e-24, 9.32586e-24, 4.55727e-24, 5.40048e-24], [122.282, 6.3208e-24, 9.3245e-24, 4.55461e-24, 5.39754e-24], [122.564, 6.32565e-24, 9.32314e-24, 4.55193e-24, 5.39461e-24], [122.847, 6.33055e-24, 9.32179e-24, 4.54925e-24, 5.39168e-24], [123.13, 6.33547e-24, 9.32045e-24, 4.54657e-24, 5.38877e-24], [123.414, 6.34044e-24, 9.31911e-24, 4.54388e-24, 5.38585e-24], [123.698, 6.34544e-24, 9.31777e-24, 4.54119e-24, 5.38295e-24], [123.984, 6.35049e-24, 9.31644e-24, 4.53849e-24, 5.38006e-24], [124.27, 6.35557e-24, 9.31512e-24, 4.53579e-24, 5.37717e-24], [124.556, 6.36068e-24, 9.3138e-24, 4.53308e-24, 5.37429e-24], [124.843, 6.36584e-24, 9.31249e-24, 4.53037e-24, 5.37141e-24], [125.131, 6.37103e-24, 9.31118e-24, 4.52766e-24, 5.36855e-24], [125.42, 6.37626e-24, 9.30987e-24, 4.52494e-24, 5.36569e-24], [125.709, 6.38153e-24, 9.30857e-24, 4.52221e-24, 5.36283e-24], [125.999, 6.38684e-24, 9.30728e-24, 4.51948e-24, 5.35999e-24], [126.289, 6.39218e-24, 9.30599e-24, 4.51674e-24, 5.35715e-24], [126.581, 6.39756e-24, 9.30471e-24, 4.514e-24, 5.35432e-24], [126.873, 6.40299e-24, 9.30343e-24, 4.51126e-24, 5.3515e-24], [127.165, 6.40844e-24, 9.30215e-24, 4.5085e-24, 5.34868e-24], [127.458, 6.41394e-24, 9.30089e-24, 4.50575e-24, 5.34587e-24], [127.752, 6.41948e-24, 9.29962e-24, 4.50299e-24, 5.34307e-24], [128.047, 6.42505e-24, 9.29836e-24, 4.50022e-24, 5.34028e-24], [128.342, 6.43066e-24, 9.29711e-24, 4.49745e-24, 5.33749e-24], [128.638, 6.43631e-24, 9.29586e-24, 4.49467e-24, 5.33471e-24], [128.935, 6.442e-24, 9.29462e-24, 4.49188e-24, 5.33193e-24], [129.232, 6.44772e-24, 9.29338e-24, 4.4891e-24, 5.32917e-24], [129.53, 6.45348e-24, 9.29215e-24, 4.4863e-24, 5.32641e-24], [129.829, 6.45929e-24, 9.29092e-24, 4.4835e-24, 5.32365e-24], [130.128, 6.46513e-24, 9.28969e-24, 4.4807e-24, 5.32091e-24], [130.428, 6.47101e-24, 9.28848e-24, 4.47788e-24, 5.31817e-24], [130.729, 6.47692e-24, 9.28726e-24, 4.47507e-24, 5.31544e-24], [131.031, 6.48288e-24, 9.28605e-24, 4.47225e-24, 5.31271e-24], [131.333, 6.48887e-24, 9.28485e-24, 4.46942e-24, 5.30999e-24], [131.636, 6.4949e-24, 9.28365e-24, 4.46658e-24, 5.30728e-24], [131.939, 6.50097e-24, 9.28246e-24, 4.46374e-24, 5.30458e-24], [132.243, 6.50708e-24, 9.28127e-24, 4.4609e-24, 5.30188e-24], [132.548, 6.51323e-24, 9.28009e-24, 4.45805e-24, 5.29919e-24], [132.854, 6.51942e-24, 9.27891e-24, 4.45519e-24, 5.2965e-24], [133.16, 6.52564e-24, 9.27774e-24, 4.45233e-24, 5.29382e-24], [133.467, 6.53191e-24, 9.27657e-24, 4.44946e-24, 5.29115e-24], [133.775, 6.53821e-24, 9.27541e-24, 4.44658e-24, 5.28849e-24], [134.084, 6.54455e-24, 9.27425e-24, 4.4437e-24, 5.28583e-24], [134.393, 6.55093e-24, 9.2731e-24, 4.44081e-24, 5.28318e-24], [134.703, 6.55735e-24, 9.27195e-24, 4.43792e-24, 5.28054e-24], [135.013, 6.5638e-24, 9.27081e-24, 4.43502e-24, 5.2779e-24], [135.325, 6.5703e-24, 9.26967e-24, 4.43212e-24, 5.27527e-24], [135.637, 6.57684e-24, 9.26854e-24, 4.4292e-24, 5.27265e-24], [135.95, 6.58341e-24, 9.26741e-24, 4.42629e-24, 5.27003e-24], [136.263, 6.59002e-24, 9.26629e-24, 4.42336e-24, 5.26742e-24], [136.577, 6.59667e-24, 9.26517e-24, 4.42043e-24, 5.26481e-24], [136.892, 6.60336e-24, 9.26406e-24, 4.41749e-24, 5.26222e-24], [137.208, 6.61009e-24, 9.26295e-24, 4.41455e-24, 5.25962e-24], [137.524, 6.61686e-24, 9.26184e-24, 4.4116e-24, 5.25704e-24], [137.841, 6.62367e-24, 9.26075e-24, 4.40865e-24, 5.25446e-24], [138.159, 6.63052e-24, 9.25965e-24, 4.40568e-24, 5.25189e-24], [138.478, 6.6374e-24, 9.25857e-24, 4.40272e-24, 5.24933e-24], [138.797, 6.64433e-24, 9.25749e-24, 4.39974e-24, 5.24677e-24], [139.117, 6.65129e-24, 9.25641e-24, 4.39676e-24, 5.24421e-24], [139.438, 6.6583e-24, 9.25534e-24, 4.39377e-24, 5.24167e-24], [139.76, 6.66534e-24, 9.25427e-24, 4.39078e-24, 5.23913e-24], [140.082, 6.67242e-24, 9.25321e-24, 4.38778e-24, 5.2366e-24], [140.405, 6.67954e-24, 9.25215e-24, 4.38477e-24, 5.23407e-24], [140.729, 6.68671e-24, 9.2511e-24, 4.38175e-24, 5.23155e-24], [141.053, 6.69391e-24, 9.25005e-24, 4.37873e-24, 5.22904e-24], [141.379, 6.70115e-24, 9.24901e-24, 4.37571e-24, 5.22653e-24], [141.705, 6.70843e-24, 9.24797e-24, 4.37267e-24, 5.22403e-24], [142.031, 6.71575e-24, 9.24694e-24, 4.36963e-24, 5.22154e-24], [142.359, 6.72311e-24, 9.24591e-24, 4.36658e-24, 5.21905e-24], [142.687, 6.7305e-24, 9.24489e-24, 4.36353e-24, 5.21657e-24], [143.016, 6.73794e-24, 9.24387e-24, 4.36047e-24, 5.21409e-24], [143.346, 6.74542e-24, 9.24286e-24, 4.3574e-24, 5.21162e-24], [143.677, 6.75294e-24, 9.24185e-24, 4.35433e-24, 5.20916e-24], [144.008, 6.7605e-24, 9.24085e-24, 4.35125e-24, 5.2067e-24], [144.34, 6.76809e-24, 9.23985e-24, 4.34816e-24, 5.20425e-24], [144.673, 6.77573e-24, 9.23886e-24, 4.34506e-24, 5.20181e-24], [145.006, 6.78341e-24, 9.23788e-24, 4.34196e-24, 5.19937e-24], [145.341, 6.79113e-24, 9.23689e-24, 4.33885e-24, 5.19694e-24], [145.676, 6.79888e-24, 9.23592e-24, 4.33574e-24, 5.19452e-24], [146.012, 6.80668e-24, 9.23495e-24, 4.33262e-24, 5.1921e-24], [146.349, 6.81452e-24, 9.23398e-24, 4.32949e-24, 5.18969e-24], [146.686, 6.8224e-24, 9.23302e-24, 4.32635e-24, 5.18728e-24], [147.024, 6.83031e-24, 9.23206e-24, 4.32321e-24, 5.18488e-24], [147.363, 6.83827e-24, 9.23111e-24, 4.32006e-24, 5.18248e-24], [147.703, 6.84627e-24, 9.23016e-24, 4.31691e-24, 5.1801e-24], [148.044, 6.85431e-24, 9.22922e-24, 4.31374e-24, 5.17771e-24], [148.385, 6.86239e-24, 9.22829e-24, 4.31057e-24, 5.17534e-24], [148.727, 6.87051e-24, 9.22736e-24, 4.3074e-24, 5.17297e-24], [149.07, 6.87867e-24, 9.22643e-24, 4.30421e-24, 5.17061e-24], [149.414, 6.88687e-24, 9.22551e-24, 4.30102e-24, 5.16825e-24], [149.759, 6.89511e-24, 9.22459e-24, 4.29782e-24, 5.1659e-24], [150.104, 6.90339e-24, 9.22368e-24, 4.29462e-24, 5.16355e-24], [150.45, 6.91171e-24, 9.22278e-24, 4.29141e-24, 5.16121e-24], [150.797, 6.92007e-24, 9.22188e-24, 4.28819e-24, 5.15888e-24], [151.145, 6.92848e-24, 9.22098e-24, 4.28496e-24, 5.15655e-24], [151.493, 6.93692e-24, 9.22009e-24, 4.28173e-24, 5.15423e-24], [151.843, 6.94541e-24, 9.21921e-24, 4.27849e-24, 5.15192e-24], [152.193, 6.95393e-24, 9.21833e-24, 4.27524e-24, 5.14961e-24], [152.544, 6.9625e-24, 9.21745e-24, 4.27199e-24, 5.14731e-24], [152.896, 6.97111e-24, 9.21658e-24, 4.26873e-24, 5.14501e-24], [153.248, 6.97975e-24, 9.21572e-24, 4.26546e-24, 5.14272e-24], [153.601, 6.98844e-24, 9.21486e-24, 4.26219e-24, 5.14043e-24], [153.956, 6.99717e-24, 9.214e-24, 4.2589e-24, 5.13815e-24], [154.311, 7.00595e-24, 9.21315e-24, 4.25562e-24, 5.13588e-24], [154.667, 7.01476e-24, 9.21231e-24, 4.25232e-24, 5.13361e-24], [155.023, 7.02361e-24, 9.21147e-24, 4.24902e-24, 5.13135e-24], [155.381, 7.03251e-24, 9.21063e-24, 4.24571e-24, 5.1291e-24], [155.739, 7.04144e-24, 9.20981e-24, 4.24239e-24, 5.12685e-24], [156.098, 7.05042e-24, 9.20898e-24, 4.23907e-24, 5.1246e-24], [156.458, 7.05944e-24, 9.20816e-24, 4.23574e-24, 5.12237e-24], [156.819, 7.0685e-24, 9.20735e-24, 4.2324e-24, 5.12014e-24], [157.181, 7.0776e-24, 9.20654e-24, 4.22905e-24, 5.11791e-24], [157.543, 7.08675e-24, 9.20574e-24, 4.2257e-24, 5.11569e-24], [157.906, 7.09593e-24, 9.20494e-24, 4.22234e-24, 5.11348e-24], [158.27, 7.10516e-24, 9.20415e-24, 4.21898e-24, 5.11127e-24], [158.635, 7.11442e-24, 9.20336e-24, 4.21561e-24, 5.10907e-24], [159.001, 7.12373e-24, 9.20258e-24, 4.21223e-24, 5.10687e-24], [159.368, 7.13309e-24, 9.2018e-24, 4.20884e-24, 5.10468e-24], [159.735, 7.14248e-24, 9.20103e-24, 4.20545e-24, 5.10249e-24], [160.104, 7.15191e-24, 9.20026e-24, 4.20205e-24, 5.10032e-24], [160.473, 7.16139e-24, 9.1995e-24, 4.19864e-24, 5.09814e-24], [160.843, 7.17091e-24, 9.19874e-24, 4.19523e-24, 5.09598e-24], [161.214, 7.18047e-24, 9.19799e-24, 4.19181e-24, 5.09381e-24], [161.586, 7.19007e-24, 9.19724e-24, 4.18838e-24, 5.09166e-24], [161.958, 7.19972e-24, 9.1965e-24, 4.18495e-24, 5.08951e-24], [162.332, 7.20941e-24, 9.19577e-24, 4.1815e-24, 5.08736e-24], [162.706, 7.21914e-24, 9.19504e-24, 4.17806e-24, 5.08523e-24], [163.081, 7.22891e-24, 9.19431e-24, 4.1746e-24, 5.08309e-24], [163.457, 7.23872e-24, 9.19359e-24, 4.17114e-24, 5.08097e-24], [163.834, 7.24858e-24, 9.19288e-24, 4.16767e-24, 5.07885e-24], [164.212, 7.25848e-24, 9.19217e-24, 4.1642e-24, 5.07673e-24], [164.591, 7.26842e-24, 9.19146e-24, 4.16072e-24, 5.07462e-24], [164.97, 7.2784e-24, 9.19076e-24, 4.15723e-24, 5.07252e-24], [165.351, 7.28843e-24, 9.19007e-24, 4.15373e-24, 5.07042e-24], [165.732, 7.29849e-24, 9.18938e-24, 4.15023e-24, 5.06833e-24], [166.114, 7.30861e-24, 9.1887e-24, 4.14672e-24, 5.06624e-24], [166.497, 7.31876e-24, 9.18802e-24, 4.14321e-24, 5.06416e-24], [166.881, 7.32896e-24, 9.18735e-24, 4.13969e-24, 5.06208e-24], [167.266, 7.3392e-24, 9.18668e-24, 4.13616e-24, 5.06001e-24], [167.652, 7.34948e-24, 9.18602e-24, 4.13263e-24, 5.05795e-24], [168.038, 7.3598e-24, 9.18536e-24, 4.12909e-24, 5.05589e-24], [168.426, 7.37017e-24, 9.18471e-24, 4.12554e-24, 5.05384e-24], [168.814, 7.38058e-24, 9.18406e-24, 4.12199e-24, 5.05179e-24], [169.204, 7.39104e-24, 9.18342e-24, 4.11843e-24, 5.04975e-24], [169.594, 7.40153e-24, 9.18278e-24, 4.11486e-24, 5.04772e-24], [169.985, 7.41207e-24, 9.18215e-24, 4.11129e-24, 5.04569e-24], [170.377, 7.42266e-24, 9.18153e-24, 4.10771e-24, 5.04366e-24], [170.77, 7.43328e-24, 9.18091e-24, 4.10413e-24, 5.04164e-24], [171.163, 7.44395e-24, 9.1803e-24, 4.10053e-24, 5.03963e-24], [171.558, 7.45467e-24, 9.17969e-24, 4.09694e-24, 5.03762e-24], [171.954, 7.46542e-24, 9.17908e-24, 4.09333e-24, 5.03562e-24], [172.35, 7.47622e-24, 9.17848e-24, 4.08972e-24, 5.03362e-24], [172.748, 7.48707e-24, 9.17789e-24, 4.08611e-24, 5.03163e-24], [173.146, 7.49795e-24, 9.1773e-24, 4.08249e-24, 5.02965e-24], [173.545, 7.50888e-24, 9.17672e-24, 4.07886e-24, 5.02767e-24], [173.946, 7.51986e-24, 9.17614e-24, 4.07523e-24, 5.0257e-24], [174.347, 7.53088e-24, 9.17557e-24, 4.07159e-24, 5.02373e-24], [174.749, 7.54194e-24, 9.17501e-24, 4.06794e-24, 5.02176e-24], [175.152, 7.55304e-24, 9.17445e-24, 4.06429e-24, 5.01981e-24], [175.556, 7.56419e-24, 9.17389e-24, 4.06064e-24, 5.01786e-24], [175.96, 7.57539e-24, 9.17334e-24, 4.05697e-24, 5.01591e-24], [176.366, 7.58662e-24, 9.1728e-24, 4.05331e-24, 5.01397e-24], [176.773, 7.59791e-24, 9.17226e-24, 4.04963e-24, 5.01203e-24], [177.181, 7.60923e-24, 9.17173e-24, 4.04596e-24, 5.0101e-24], [177.589, 7.6206e-24, 9.1712e-24, 4.04227e-24, 5.00818e-24], [177.999, 7.63201e-24, 9.17067e-24, 4.03858e-24, 5.00626e-24], [178.409, 7.64347e-24, 9.17016e-24, 4.03489e-24, 5.00435e-24], [178.821, 7.65498e-24, 9.16965e-24, 4.03119e-24, 5.00244e-24], [179.233, 7.66652e-24, 9.16914e-24, 4.02748e-24, 5.00054e-24], [179.646, 7.67811e-24, 9.16864e-24, 4.02377e-24, 4.99864e-24], [180.061, 7.68975e-24, 9.16814e-24, 4.02006e-24, 4.99675e-24], [180.476, 7.70143e-24, 9.16765e-24, 4.01634e-24, 4.99487e-24], [180.892, 7.71316e-24, 9.16717e-24, 4.01261e-24, 4.99299e-24], [181.309, 7.72493e-24, 9.16669e-24, 4.00888e-24, 4.99112e-24], [181.727, 7.73674e-24, 9.16622e-24, 4.00514e-24, 4.98925e-24], [182.146, 7.7486e-24, 9.16575e-24, 4.00141e-24, 4.98738e-24], [182.566, 7.7605e-24, 9.16529e-24, 3.99766e-24, 4.98553e-24], [182.987, 7.77245e-24, 9.16483e-24, 3.99391e-24, 4.98367e-24], [183.409, 7.78445e-24, 9.16438e-24, 3.99016e-24, 4.98183e-24], [183.832, 7.79648e-24, 9.16394e-24, 3.9864e-24, 4.97999e-24], [184.256, 7.80857e-24, 9.1635e-24, 3.98264e-24, 4.97815e-24], [184.681, 7.8207e-24, 9.16306e-24, 3.97887e-24, 4.97632e-24], [185.107, 7.83287e-24, 9.16264e-24, 3.9751e-24, 4.97449e-24], [185.534, 7.84509e-24, 9.16221e-24, 3.97133e-24, 4.97268e-24], [185.962, 7.85736e-24, 9.1618e-24, 3.96755e-24, 4.97086e-24], [186.39, 7.86967e-24, 9.16138e-24, 3.96376e-24, 4.96905e-24], [186.82, 7.88202e-24, 9.16098e-24, 3.95998e-24, 4.96725e-24], [187.251, 7.89442e-24, 9.16058e-24, 3.95619e-24, 4.96545e-24], [187.683, 7.90687e-24, 9.16018e-24, 3.95239e-24, 4.96366e-24], [188.116, 7.91936e-24, 9.1598e-24, 3.94859e-24, 4.96187e-24], [188.549, 7.9319e-24, 9.15941e-24, 3.94479e-24, 4.96009e-24], [188.984, 7.94448e-24, 9.15903e-24, 3.94099e-24, 4.95832e-24], [189.42, 7.95711e-24, 9.15866e-24, 3.93718e-24, 4.95655e-24], [189.857, 7.96979e-24, 9.1583e-24, 3.93337e-24, 4.95478e-24], [190.295, 7.98251e-24, 9.15794e-24, 3.92955e-24, 4.95302e-24], [190.733, 7.99528e-24, 9.15758e-24, 3.92574e-24, 4.95127e-24], [191.173, 8.00809e-24, 9.15723e-24, 3.92192e-24, 4.94952e-24], [191.614, 8.02095e-24, 9.15689e-24, 3.91809e-24, 4.94778e-24], [192.056, 8.03385e-24, 9.15655e-24, 3.91427e-24, 4.94604e-24], [192.499, 8.04681e-24, 9.15622e-24, 3.91044e-24, 4.94431e-24], [192.943, 8.0598e-24, 9.15589e-24, 3.90661e-24, 4.94258e-24], [193.388, 8.07285e-24, 9.15557e-24, 3.90277e-24, 4.94086e-24], [193.834, 8.08594e-24, 9.15526e-24, 3.89894e-24, 4.93914e-24], [194.281, 8.09907e-24, 9.15495e-24, 3.8951e-24, 4.93743e-24], [194.729, 8.11226e-24, 9.15465e-24, 3.89126e-24, 4.93573e-24], [195.178, 8.12549e-24, 9.15435e-24, 3.88742e-24, 4.93403e-24], [195.628, 8.13876e-24, 9.15406e-24, 3.88358e-24, 4.93233e-24], [196.079, 8.15209e-24, 9.15378e-24, 3.87973e-24, 4.93064e-24], [196.531, 8.16546e-24, 9.1535e-24, 3.87589e-24, 4.92896e-24], [196.984, 8.17887e-24, 9.15322e-24, 3.87204e-24, 4.92728e-24], [197.438, 8.19234e-24, 9.15296e-24, 3.86819e-24, 4.92561e-24], [197.894, 8.20585e-24, 9.15269e-24, 3.86434e-24, 4.92394e-24], [198.35, 8.21941e-24, 9.15244e-24, 3.86049e-24, 4.92228e-24], [198.808, 8.23301e-24, 9.15219e-24, 3.85664e-24, 4.92062e-24], [199.266, 8.24666e-24, 9.15194e-24, 3.85279e-24, 4.91897e-24], [199.725, 8.26036e-24, 9.15171e-24, 3.84894e-24, 4.91732e-24], [200.186, 8.27411e-24, 9.15147e-24, 3.84508e-24, 4.91568e-24], [200.648, 8.2879e-24, 9.15125e-24, 3.84123e-24, 4.91405e-24], [201.11, 8.30175e-24, 9.15103e-24, 3.83738e-24, 4.91242e-24], [201.574, 8.31563e-24, 9.15081e-24, 3.83352e-24, 4.91079e-24], [202.039, 8.32957e-24, 9.15061e-24, 3.82967e-24, 4.90917e-24], [202.505, 8.34355e-24, 9.1504e-24, 3.82582e-24, 4.90756e-24], [202.972, 8.35759e-24, 9.15021e-24, 3.82197e-24, 4.90595e-24], [203.44, 8.37167e-24, 9.15002e-24, 3.81811e-24, 4.90435e-24], [203.909, 8.38579e-24, 9.14983e-24, 3.81426e-24, 4.90275e-24], [204.379, 8.39997e-24, 9.14966e-24, 3.81041e-24, 4.90116e-24], [204.851, 8.41419e-24, 9.14949e-24, 3.80657e-24, 4.89957e-24], [205.323, 8.42846e-24, 9.14932e-24, 3.80272e-24, 4.89799e-24], [205.796, 8.44278e-24, 9.14916e-24, 3.79888e-24, 4.89642e-24], [206.271, 8.45715e-24, 9.14901e-24, 3.79503e-24, 4.89485e-24], [206.747, 8.47156e-24, 9.14886e-24, 3.79119e-24, 4.89328e-24], [207.223, 8.48603e-24, 9.14872e-24, 3.78735e-24, 4.89172e-24], [207.701, 8.50054e-24, 9.14858e-24, 3.78352e-24, 4.89017e-24], [208.18, 8.5151e-24, 9.14845e-24, 3.77968e-24, 4.88862e-24], [208.66, 8.52971e-24, 9.14833e-24, 3.77585e-24, 4.88707e-24], [209.141, 8.54437e-24, 9.14821e-24, 3.77202e-24, 4.88554e-24], [209.624, 8.55907e-24, 9.1481e-24, 3.7682e-24, 4.884e-24], [210.107, 8.57383e-24, 9.148e-24, 3.76437e-24, 4.88247e-24], [210.592, 8.58863e-24, 9.1479e-24, 3.76056e-24, 4.88095e-24], [211.077, 8.60348e-24, 9.14781e-24, 3.75674e-24, 4.87944e-24], [211.564, 8.61839e-24, 9.14772e-24, 3.75293e-24, 4.87792e-24], [212.052, 8.63334e-24, 9.14765e-24, 3.74912e-24, 4.87642e-24], [212.541, 8.64834e-24, 9.14757e-24, 3.74532e-24, 4.87492e-24], [213.031, 8.66338e-24, 9.14751e-24, 3.74153e-24, 4.87342e-24], [213.522, 8.67848e-24, 9.14745e-24, 3.73773e-24, 4.87193e-24], [214.015, 8.69363e-24, 9.14739e-24, 3.73395e-24, 4.87045e-24], [214.508, 8.70883e-24, 9.14734e-24, 3.73016e-24, 4.86897e-24], [215.003, 8.72407e-24, 9.1473e-24, 3.72639e-24, 4.86749e-24], [215.499, 8.73937e-24, 9.14727e-24, 3.72262e-24, 4.86603e-24], [215.996, 8.75471e-24, 9.14724e-24, 3.71885e-24, 4.86456e-24], [216.494, 8.77011e-24, 9.14722e-24, 3.71509e-24, 4.86311e-24], [216.993, 8.78555e-24, 9.1472e-24, 3.71134e-24, 4.86165e-24], [217.493, 8.80104e-24, 9.14719e-24, 3.7076e-24, 4.86021e-24], [217.995, 8.81659e-24, 9.14719e-24, 3.70386e-24, 4.85876e-24], [218.498, 8.83218e-24, 9.14719e-24, 3.70013e-24, 4.85733e-24], [219.001, 8.84782e-24, 9.1472e-24, 3.6964e-24, 4.8559e-24], [219.506, 8.86352e-24, 9.14722e-24, 3.69269e-24, 4.85447e-24], [220.013, 8.87926e-24, 9.14724e-24, 3.68898e-24, 4.85305e-24], [220.52, 8.89506e-24, 9.14727e-24, 3.68528e-24, 4.85164e-24], [221.028, 8.9109e-24, 9.14731e-24, 3.68159e-24, 4.85023e-24], [221.538, 8.92679e-24, 9.14735e-24, 3.67791e-24, 4.84882e-24], [222.049, 8.94274e-24, 9.1474e-24, 3.67424e-24, 4.84743e-24], [222.561, 8.95874e-24, 9.14746e-24, 3.67058e-24, 4.84603e-24], [223.074, 8.97478e-24, 9.14752e-24, 3.66692e-24, 4.84464e-24], [223.589, 8.99088e-24, 9.14759e-24, 3.66328e-24, 4.84326e-24], [224.104, 9.00703e-24, 9.14766e-24, 3.65965e-24, 4.84189e-24], [224.621, 9.02322e-24, 9.14775e-24, 3.65603e-24, 4.84051e-24], [225.139, 9.03947e-24, 9.14784e-24, 3.65242e-24, 4.83915e-24], [225.658, 9.05577e-24, 9.14793e-24, 3.64882e-24, 4.83779e-24], [226.179, 9.07212e-24, 9.14803e-24, 3.64523e-24, 4.83643e-24], [226.7, 9.08853e-24, 9.14814e-24, 3.64166e-24, 4.83508e-24], [227.223, 9.10498e-24, 9.14826e-24, 3.63809e-24, 4.83374e-24], [227.747, 9.12149e-24, 9.14838e-24, 3.63454e-24, 4.8324e-24], [228.272, 9.13804e-24, 9.14851e-24, 3.631e-24, 4.83106e-24], [228.798, 9.15465e-24, 9.14865e-24, 3.62748e-24, 4.82974e-24], [229.326, 9.17131e-24, 9.14879e-24, 3.62397e-24, 4.82841e-24], [229.855, 9.18802e-24, 9.14894e-24, 3.62047e-24, 4.8271e-24], [230.385, 9.20478e-24, 9.1491e-24, 3.61699e-24, 4.82578e-24], [230.916, 9.2216e-24, 9.14926e-24, 3.61353e-24, 4.82448e-24], [231.449, 9.23846e-24, 9.14943e-24, 3.61007e-24, 4.82318e-24], [231.982, 9.25538e-24, 9.14961e-24, 3.60664e-24, 4.82188e-24], [232.517, 9.27235e-24, 9.14979e-24, 3.60322e-24, 4.82059e-24], [233.054, 9.28937e-24, 9.14998e-24, 3.59981e-24, 4.8193e-24], [233.591, 9.30645e-24, 9.15018e-24, 3.59642e-24, 4.81802e-24], [234.13, 9.32357e-24, 9.15038e-24, 3.59305e-24, 4.81675e-24], [234.67, 9.34075e-24, 9.15059e-24, 3.5897e-24, 4.81548e-24], [235.211, 9.35798e-24, 9.15081e-24, 3.58636e-24, 4.81422e-24], [235.753, 9.37526e-24, 9.15104e-24, 3.58305e-24, 4.81296e-24], [236.297, 9.3926e-24, 9.15127e-24, 3.57975e-24, 4.81171e-24], [236.842, 9.40999e-24, 9.15151e-24, 3.57647e-24, 4.81046e-24], [237.388, 9.42743e-24, 9.15176e-24, 3.57321e-24, 4.80922e-24], [237.935, 9.44493e-24, 9.15201e-24, 3.56997e-24, 4.80798e-24], [238.484, 9.46247e-24, 9.15227e-24, 3.56675e-24, 4.80675e-24], [239.034, 9.48007e-24, 9.15254e-24, 3.56355e-24, 4.80553e-24], [239.585, 9.49773e-24, 9.15282e-24, 3.56037e-24, 4.80431e-24], [240.138, 9.51543e-24, 9.1531e-24, 3.55721e-24, 4.8031e-24], [240.691, 9.53319e-24, 9.15339e-24, 3.55408e-24, 4.80189e-24], [241.246, 9.551e-24, 9.15368e-24, 3.55097e-24, 4.80068e-24], [241.803, 9.56887e-24, 9.15399e-24, 3.54788e-24, 4.79949e-24], [242.36, 9.58679e-24, 9.1543e-24, 3.54481e-24, 4.79829e-24], [242.919, 9.60476e-24, 9.15462e-24, 3.54177e-24, 4.79711e-24], [243.479, 9.62279e-24, 9.15494e-24, 3.53875e-24, 4.79593e-24], [244.041, 9.64087e-24, 9.15528e-24, 3.53576e-24, 4.79475e-24], [244.604, 9.659e-24, 9.15562e-24, 3.53279e-24, 4.79358e-24], [245.168, 9.67719e-24, 9.15596e-24, 3.52985e-24, 4.79242e-24], [245.733, 9.69543e-24, 9.15632e-24, 3.52693e-24, 4.79126e-24], [246.3, 9.71373e-24, 9.15668e-24, 3.52404e-24, 4.7901e-24], [246.868, 9.73208e-24, 9.15705e-24, 3.52118e-24, 4.78895e-24], [247.437, 9.75049e-24, 9.15743e-24, 3.51835e-24, 4.78781e-24], [248.007, 9.76894e-24, 9.15781e-24, 3.51554e-24, 4.78667e-24], [248.579, 9.78746e-24, 9.1582e-24, 3.51277e-24, 4.78554e-24], [249.153, 9.80603e-24, 9.1586e-24, 3.51002e-24, 4.78442e-24], [249.727, 9.82465e-24, 9.15901e-24, 3.5073e-24, 4.7833e-24], [250.303, 9.84333e-24, 9.15942e-24, 3.50461e-24, 4.78218e-24], [250.88, 9.86206e-24, 9.15984e-24, 3.50196e-24, 4.78107e-24], [251.459, 9.88084e-24, 9.16027e-24, 3.49933e-24, 4.77997e-24], [252.039, 9.89969e-24, 9.16071e-24, 3.49674e-24, 4.77887e-24], [252.62, 9.91858e-24, 9.16116e-24, 3.49418e-24, 4.77778e-24], [253.202, 9.93754e-24, 9.16161e-24, 3.49165e-24, 4.77669e-24], [253.786, 9.95654e-24, 9.16207e-24, 3.48916e-24, 4.77561e-24], [254.372, 9.97561e-24, 9.16253e-24, 3.4867e-24, 4.77453e-24], [254.958, 9.99472e-24, 9.16301e-24, 3.48428e-24, 4.77346e-24], [255.546, 1.00139e-23, 9.16349e-24, 3.48189e-24, 4.7724e-24], [256.135, 1.00331e-23, 9.16398e-24, 3.47954e-24, 4.77134e-24], [256.726, 1.00524e-23, 9.16448e-24, 3.47722e-24, 4.77028e-24], [257.318, 1.00718e-23, 9.16499e-24, 3.47494e-24, 4.76924e-24], [257.911, 1.00911e-23, 9.1655e-24, 3.4727e-24, 4.76819e-24], [258.506, 1.01106e-23, 9.16602e-24, 3.4705e-24, 4.76716e-24], [259.102, 1.01301e-23, 9.16655e-24, 3.46834e-24, 4.76612e-24], [259.7, 1.01497e-23, 9.16709e-24, 3.46622e-24, 4.7651e-24], [260.299, 1.01693e-23, 9.16763e-24, 3.46413e-24, 4.76408e-24], [260.899, 1.0189e-23, 9.16819e-24, 3.46209e-24, 4.76306e-24], [261.5, 1.02087e-23, 9.16875e-24, 3.46009e-24, 4.76206e-24], [262.103, 1.02285e-23, 9.16932e-24, 3.45813e-24, 4.76105e-24], [262.708, 1.02483e-23, 9.16989e-24, 3.45622e-24, 4.76006e-24], [263.314, 1.02682e-23, 9.17048e-24, 3.45435e-24, 4.75906e-24], [263.921, 1.02882e-23, 9.17107e-24, 3.45252e-24, 4.75808e-24], [264.53, 1.03082e-23, 9.17167e-24, 3.45074e-24, 4.7571e-24], [265.14, 1.03283e-23, 9.17228e-24, 3.449e-24, 4.75612e-24], [265.751, 1.03484e-23, 9.1729e-24, 3.44731e-24, 4.75515e-24], [266.364, 1.03686e-23, 9.17352e-24, 3.44567e-24, 4.75419e-24], [266.978, 1.03889e-23, 9.17416e-24, 3.44408e-24, 4.75323e-24], [267.594, 1.04092e-23, 9.1748e-24, 3.44253e-24, 4.75228e-24], [268.211, 1.04295e-23, 9.17545e-24, 3.44103e-24, 4.75133e-24], [268.829, 1.04499e-23, 9.1761e-24, 3.43958e-24, 4.75039e-24], [269.449, 1.04704e-23, 9.17677e-24, 3.43819e-24, 4.74946e-24], [270.071, 1.0491e-23, 9.17744e-24, 3.43684e-24, 4.74853e-24], [270.693, 1.05115e-23, 9.17813e-24, 3.43555e-24, 4.74761e-24], [271.318, 1.05322e-23, 9.17882e-24, 3.43431e-24, 4.74669e-24], [271.943, 1.05529e-23, 9.17952e-24, 3.43312e-24, 4.74578e-24], [272.57, 1.05737e-23, 9.18022e-24, 3.43198e-24, 4.74487e-24], [273.199, 1.05945e-23, 9.18094e-24, 3.4309e-24, 4.74397e-24], [273.829, 1.06154e-23, 9.18166e-24, 3.42988e-24, 4.74307e-24], [274.46, 1.06363e-23, 9.1824e-24, 3.42891e-24, 4.74218e-24], [275.093, 1.06573e-23, 9.18314e-24, 3.428e-24, 4.7413e-24], [275.728, 1.06784e-23, 9.18389e-24, 3.42715e-24, 4.74042e-24], [276.363, 1.06995e-23, 9.18465e-24, 3.42636e-24, 4.73955e-24], [277.001, 1.07207e-23, 9.18541e-24, 3.42563e-24, 4.73869e-24], [277.639, 1.07419e-23, 9.18619e-24, 3.42495e-24, 4.73783e-24], [278.28, 1.07632e-23, 9.18697e-24, 3.42434e-24, 4.73697e-24], [278.921, 1.07845e-23, 9.18776e-24, 3.42379e-24, 4.73612e-24], [279.565, 1.0806e-23, 9.18857e-24, 3.4233e-24, 4.73528e-24], [280.209, 1.08274e-23, 9.18937e-24, 3.42288e-24, 4.73445e-24], [280.855, 1.0849e-23, 9.19019e-24, 3.42251e-24, 4.73361e-24], [281.503, 1.08705e-23, 9.19102e-24, 3.42222e-24, 4.73279e-24], [282.152, 1.08922e-23, 9.19186e-24, 3.42199e-24, 4.73197e-24], [282.803, 1.09139e-23, 9.1927e-24, 3.42183e-24, 4.73116e-24], [283.455, 1.09357e-23, 9.19355e-24, 3.42173e-24, 4.73035e-24], [284.109, 1.09575e-23, 9.19442e-24, 3.4217e-24, 4.72955e-24], [284.764, 1.09794e-23, 9.19529e-24, 3.42174e-24, 4.72875e-24], [285.421, 1.10013e-23, 9.19617e-24, 3.42185e-24, 4.72796e-24], [286.079, 1.10234e-23, 9.19706e-24, 3.42203e-24, 4.72718e-24], [286.738, 1.10454e-23, 9.19795e-24, 3.42229e-24, 4.7264e-24], [287.4, 1.10676e-23, 9.19886e-24, 3.42261e-24, 4.72563e-24], [288.062, 1.10898e-23, 9.19978e-24, 3.42301e-24, 4.72486e-24], [288.727, 1.1112e-23, 9.2007e-24, 3.42348e-24, 4.7241e-24], [289.392, 1.11343e-23, 9.20163e-24, 3.42403e-24, 4.72335e-24], [290.06, 1.11567e-23, 9.20258e-24, 3.42465e-24, 4.7226e-24], [290.729, 1.11791e-23, 9.20353e-24, 3.42535e-24, 4.72186e-24], [291.399, 1.12016e-23, 9.20449e-24, 3.42613e-24, 4.72112e-24], [292.071, 1.12242e-23, 9.20546e-24, 3.42699e-24, 4.72039e-24], [292.745, 1.12468e-23, 9.20644e-24, 3.42792e-24, 4.71967e-24], [293.42, 1.12695e-23, 9.20743e-24, 3.42893e-24, 4.71895e-24], [294.096, 1.12922e-23, 9.20843e-24, 3.43003e-24, 4.71824e-24], [294.774, 1.1315e-23, 9.20943e-24, 3.43121e-24, 4.71753e-24], [295.454, 1.13379e-23, 9.21045e-24, 3.43246e-24, 4.71683e-24], [296.136, 1.13608e-23, 9.21148e-24, 3.43381e-24, 4.71614e-24], [296.818, 1.13838e-23, 9.21251e-24, 3.43523e-24, 4.71545e-24], [297.503, 1.14069e-23, 9.21356e-24, 3.43674e-24, 4.71477e-24], [298.189, 1.143e-23, 9.21461e-24, 3.43834e-24, 4.71409e-24], [298.877, 1.14532e-23, 9.21567e-24, 3.44002e-24, 4.71342e-24], [299.566, 1.14764e-23, 9.21675e-24, 3.4418e-24, 4.71276e-24], [300.257, 1.14997e-23, 9.21783e-24, 3.44366e-24, 4.7121e-24], [300.949, 1.15231e-23, 9.21892e-24, 3.44561e-24, 4.71145e-24], [301.643, 1.15465e-23, 9.22002e-24, 3.44765e-24, 4.7108e-24], [302.339, 1.157e-23, 9.22113e-24, 3.44978e-24, 4.71016e-24], [303.036, 1.15936e-23, 9.22225e-24, 3.452e-24, 4.70953e-24], [303.735, 1.16172e-23, 9.22339e-24, 3.45432e-24, 4.7089e-24], [304.435, 1.16409e-23, 9.22453e-24, 3.45672e-24, 4.70828e-24], [305.137, 1.16646e-23, 9.22568e-24, 3.45923e-24, 4.70767e-24], [305.841, 1.16884e-23, 9.22684e-24, 3.46183e-24, 4.70706e-24], [306.546, 1.17123e-23, 9.228e-24, 3.46452e-24, 4.70646e-24], [307.253, 1.17362e-23, 9.22918e-24, 3.46732e-24, 4.70586e-24], [307.961, 1.17602e-23, 9.23037e-24, 3.47021e-24, 4.70527e-24], [308.671, 1.17843e-23, 9.23157e-24, 3.47319e-24, 4.70469e-24], [309.383, 1.18084e-23, 9.23278e-24, 3.47628e-24, 4.70411e-24], [310.097, 1.18326e-23, 9.234e-24, 3.47947e-24, 4.70354e-24], [310.812, 1.18569e-23, 9.23523e-24, 3.48276e-24, 4.70297e-24], [311.529, 1.18812e-23, 9.23647e-24, 3.48616e-24, 4.70241e-24], [312.247, 1.19056e-23, 9.23772e-24, 3.48965e-24, 4.70186e-24], [312.967, 1.193e-23, 9.23898e-24, 3.49325e-24, 4.70132e-24], [313.689, 1.19545e-23, 9.24025e-24, 3.49696e-24, 4.70078e-24], [314.412, 1.19791e-23, 9.24153e-24, 3.50077e-24, 4.70024e-24], [315.137, 1.20038e-23, 9.24281e-24, 3.50469e-24, 4.69971e-24], [315.864, 1.20285e-23, 9.24411e-24, 3.50871e-24, 4.69919e-24], [316.592, 1.20533e-23, 9.24542e-24, 3.51284e-24, 4.69868e-24], [317.322, 1.20781e-23, 9.24674e-24, 3.51708e-24, 4.69817e-24], [318.054, 1.2103e-23, 9.24808e-24, 3.52144e-24, 4.69767e-24], [318.787, 1.2128e-23, 9.24942e-24, 3.5259e-24, 4.69717e-24], [319.523, 1.2153e-23, 9.25077e-24, 3.53047e-24, 4.69668e-24], [320.259, 1.21781e-23, 9.25213e-24, 3.53516e-24, 4.6962e-24], [320.998, 1.22033e-23, 9.2535e-24, 3.53995e-24, 4.69573e-24], [321.738, 1.22285e-23, 9.25488e-24, 3.54487e-24, 4.69526e-24], [322.48, 1.22538e-23, 9.25628e-24, 3.54989e-24, 4.69479e-24], [323.224, 1.22792e-23, 9.25768e-24, 3.55503e-24, 4.69434e-24], [323.969, 1.23046e-23, 9.2591e-24, 3.56029e-24, 4.69388e-24], [324.716, 1.23301e-23, 9.26052e-24, 3.56567e-24, 4.69344e-24], [325.465, 1.23557e-23, 9.26196e-24, 3.57116e-24, 4.693e-24], [326.215, 1.23814e-23, 9.2634e-24, 3.57677e-24, 4.69257e-24], [326.968, 1.24071e-23, 9.26486e-24, 3.5825e-24, 4.69215e-24], [327.722, 1.24328e-23, 9.26633e-24, 3.58835e-24, 4.69173e-24], [328.477, 1.24587e-23, 9.26781e-24, 3.59432e-24, 4.69132e-24], [329.235, 1.24846e-23, 9.2693e-24, 3.60042e-24, 4.69091e-24], [329.994, 1.25105e-23, 9.2708e-24, 3.60663e-24, 4.69051e-24], [330.755, 1.25366e-23, 9.27231e-24, 3.61297e-24, 4.69012e-24], [331.518, 1.25627e-23, 9.27383e-24, 3.61943e-24, 4.68974e-24], [332.282, 1.25889e-23, 9.27537e-24, 3.62602e-24, 4.68936e-24], [333.049, 1.26151e-23, 9.27691e-24, 3.63273e-24, 4.68899e-24], [333.817, 1.26414e-23, 9.27847e-24, 3.63957e-24, 4.68862e-24], [334.586, 1.26678e-23, 9.28003e-24, 3.64653e-24, 4.68826e-24], [335.358, 1.26943e-23, 9.28161e-24, 3.65362e-24, 4.68791e-24], [336.131, 1.27208e-23, 9.2832e-24, 3.66084e-24, 4.68756e-24], [336.906, 1.27474e-23, 9.2848e-24, 3.66819e-24, 4.68723e-24], [337.683, 1.2774e-23, 9.28641e-24, 3.67567e-24, 4.68689e-24], [338.462, 1.28008e-23, 9.28803e-24, 3.68328e-24, 4.68657e-24], [339.242, 1.28276e-23, 9.28967e-24, 3.69102e-24, 4.68625e-24], [340.025, 1.28544e-23, 9.29131e-24, 3.69889e-24, 4.68594e-24], [340.809, 1.28814e-23, 9.29297e-24, 3.70689e-24, 4.68563e-24], [341.595, 1.29084e-23, 9.29464e-24, 3.71502e-24, 4.68533e-24], [342.382, 1.29354e-23, 9.29632e-24, 3.72329e-24, 4.68504e-24], [343.172, 1.29626e-23, 9.29801e-24, 3.73169e-24, 4.68476e-24], [343.963, 1.29898e-23, 9.29971e-24, 3.74023e-24, 4.68448e-24], [344.757, 1.30171e-23, 9.30143e-24, 3.7489e-24, 4.68421e-24], [345.552, 1.30444e-23, 9.30315e-24, 3.75771e-24, 4.68394e-24], [346.348, 1.30719e-23, 9.30489e-24, 3.76666e-24, 4.68369e-24], [347.147, 1.30994e-23, 9.30664e-24, 3.77574e-24, 4.68343e-24], [347.948, 1.31269e-23, 9.3084e-24, 3.78495e-24, 4.68319e-24], [348.75, 1.31546e-23, 9.31018e-24, 3.79431e-24, 4.68295e-24], [349.554, 1.31823e-23, 9.31196e-24, 3.80381e-24, 4.68272e-24], [350.36, 1.321e-23, 9.31376e-24, 3.81344e-24, 4.6825e-24], [351.168, 1.32379e-23, 9.31557e-24, 3.82322e-24, 4.68228e-24], [351.978, 1.32658e-23, 9.31739e-24, 3.83313e-24, 4.68207e-24], [352.79, 1.32938e-23, 9.31922e-24, 3.84319e-24, 4.68187e-24], [353.603, 1.33219e-23, 9.32107e-24, 3.85339e-24, 4.68168e-24], [354.419, 1.335e-23, 9.32292e-24, 3.86373e-24, 4.68149e-24], [355.236, 1.33782e-23, 9.32479e-24, 3.87421e-24, 4.68131e-24], [356.055, 1.34065e-23, 9.32667e-24, 3.88483e-24, 4.68113e-24], [356.876, 1.34348e-23, 9.32857e-24, 3.8956e-24, 4.68097e-24], [357.699, 1.34633e-23, 9.33047e-24, 3.90651e-24, 4.68081e-24], [358.524, 1.34918e-23, 9.33239e-24, 3.91757e-24, 4.68065e-24], [359.351, 1.35203e-23, 9.33432e-24, 3.92877e-24, 4.68051e-24], [360.179, 1.3549e-23, 9.33626e-24, 3.94012e-24, 4.68037e-24], [361.01, 1.35777e-23, 9.33822e-24, 3.95161e-24, 4.68024e-24], [361.842, 1.36065e-23, 9.34019e-24, 3.96325e-24, 4.68011e-24], [362.677, 1.36354e-23, 9.34217e-24, 3.97504e-24, 4.67999e-24], [363.513, 1.36643e-23, 9.34416e-24, 3.98697e-24, 4.67988e-24], [364.351, 1.36933e-23, 9.34617e-24, 3.99905e-24, 4.67978e-24], [365.192, 1.37224e-23, 9.34818e-24, 4.01128e-24, 4.67968e-24], [366.034, 1.37515e-23, 9.35021e-24, 4.02366e-24, 4.6796e-24], [366.878, 1.37808e-23, 9.35226e-24, 4.03619e-24, 4.67951e-24], [367.724, 1.38101e-23, 9.35431e-24, 4.04886e-24, 4.67944e-24], [368.572, 1.38394e-23, 9.35638e-24, 4.06169e-24, 4.67937e-24], [369.422, 1.38689e-23, 9.35847e-24, 4.07467e-24, 4.67931e-24], [370.274, 1.38984e-23, 9.36056e-24, 4.08779e-24, 4.67926e-24], [371.128, 1.3928e-23, 9.36267e-24, 4.10107e-24, 4.67922e-24], [371.983, 1.39577e-23, 9.36479e-24, 4.1145e-24, 4.67918e-24], [372.841, 1.39875e-23, 9.36692e-24, 4.12808e-24, 4.67915e-24], [373.701, 1.40173e-23, 9.36907e-24, 4.14181e-24, 4.67913e-24], [374.563, 1.40472e-23, 9.37123e-24, 4.15569e-24, 4.67911e-24], [375.426, 1.40772e-23, 9.3734e-24, 4.16973e-24, 4.6791e-24], [376.292, 1.41072e-23, 9.37559e-24, 4.18392e-24, 4.6791e-24], [377.16, 1.41374e-23, 9.37779e-24, 4.19826e-24, 4.67911e-24], [378.03, 1.41676e-23, 9.38e-24, 4.21276e-24, 4.67912e-24], [378.901, 1.41978e-23, 9.38223e-24, 4.22741e-24, 4.67915e-24], [379.775, 1.42282e-23, 9.38447e-24, 4.24222e-24, 4.67918e-24], [380.651, 1.42586e-23, 9.38673e-24, 4.25717e-24, 4.67921e-24], [381.529, 1.42891e-23, 9.38899e-24, 4.27229e-24, 4.67926e-24], [382.408, 1.43197e-23, 9.39127e-24, 4.28756e-24, 4.67931e-24], [383.29, 1.43504e-23, 9.39357e-24, 4.30298e-24, 4.67937e-24], [384.174, 1.43811e-23, 9.39588e-24, 4.31856e-24, 4.67944e-24], [385.06, 1.4412e-23, 9.3982e-24, 4.3343e-24, 4.67951e-24], [385.948, 1.44428e-23, 9.40053e-24, 4.35019e-24, 4.6796e-24], [386.838, 1.44738e-23, 9.40288e-24, 4.36624e-24, 4.67969e-24], [387.73, 1.45049e-23, 9.40525e-24, 4.38245e-24, 4.67978e-24], [388.624, 1.4536e-23, 9.40762e-24, 4.39881e-24, 4.67989e-24], [389.52, 1.45672e-23, 9.41001e-24, 4.41533e-24, 4.68001e-24], [390.419, 1.45985e-23, 9.41242e-24, 4.43201e-24, 4.68013e-24], [391.319, 1.46299e-23, 9.41484e-24, 4.44884e-24, 4.68026e-24], [392.221, 1.46613e-23, 9.41727e-24, 4.46583e-24, 4.68039e-24], [393.126, 1.46928e-23, 9.41972e-24, 4.48299e-24, 4.68054e-24], [394.032, 1.47244e-23, 9.42218e-24, 4.5003e-24, 4.68069e-24], [394.941, 1.47561e-23, 9.42466e-24, 4.51776e-24, 4.68086e-24], [395.852, 1.47878e-23, 9.42715e-24, 4.53539e-24, 4.68102e-24], [396.765, 1.48197e-23, 9.42965e-24, 4.55318e-24, 4.6812e-24], [397.679, 1.48516e-23, 9.43217e-24, 4.57112e-24, 4.68139e-24], [398.597, 1.48836e-23, 9.43471e-24, 4.58923e-24, 4.68158e-24], [399.516, 1.49157e-23, 9.43726e-24, 4.6075e-24, 4.68178e-24], [400.437, 1.49478e-23, 9.43982e-24, 4.62592e-24, 4.68199e-24], [401.36, 1.49801e-23, 9.4424e-24, 4.64451e-24, 4.68221e-24], [402.286, 1.50124e-23, 9.44499e-24, 4.66325e-24, 4.68244e-24], [403.214, 1.50448e-23, 9.4476e-24, 4.68216e-24, 4.68268e-24], [404.143, 1.50772e-23, 9.45022e-24, 4.70123e-24, 4.68292e-24], [405.075, 1.51098e-23, 9.45286e-24, 4.72045e-24, 4.68317e-24], [406.009, 1.51424e-23, 9.45551e-24, 4.73984e-24, 4.68343e-24], [406.946, 1.51752e-23, 9.45818e-24, 4.7594e-24, 4.6837e-24], [407.884, 1.5208e-23, 9.46087e-24, 4.77911e-24, 4.68398e-24], [408.825, 1.52408e-23, 9.46357e-24, 4.79898e-24, 4.68427e-24], [409.767, 1.52738e-23, 9.46628e-24, 4.81902e-24, 4.68457e-24], [410.712, 1.53068e-23, 9.46901e-24, 4.83922e-24, 4.68487e-24], [411.66, 1.534e-23, 9.47176e-24, 4.85959e-24, 4.68519e-24], [412.609, 1.53732e-23, 9.47452e-24, 4.88011e-24, 4.68552e-24], [413.56, 1.54065e-23, 9.4773e-24, 4.9008e-24, 4.68585e-24], [414.514, 1.54399e-23, 9.48009e-24, 4.92166e-24, 4.6862e-24], [415.47, 1.54733e-23, 9.48291e-24, 4.94268e-24, 4.68656e-24], [416.428, 1.55069e-23, 9.48573e-24, 4.96386e-24, 4.68693e-24], [417.388, 1.55405e-23, 9.48858e-24, 4.98521e-24, 4.68731e-24], [418.351, 1.55742e-23, 9.49144e-24, 5.00673e-24, 4.6877e-24], [419.315, 1.5608e-23, 9.49432e-24, 5.02841e-24, 4.68811e-24], [420.282, 1.56419e-23, 9.49723e-24, 5.05026e-24, 4.68853e-24], [421.252, 1.56758e-23, 9.50015e-24, 5.07229e-24, 4.68897e-24], [422.223, 1.57099e-23, 9.50309e-24, 5.09448e-24, 4.68942e-24], [423.197, 1.5744e-23, 9.50605e-24, 5.11685e-24, 4.6899e-24], [424.173, 1.57783e-23, 9.50904e-24, 5.1394e-24, 4.6904e-24], [425.151, 1.58126e-23, 9.51205e-24, 5.16213e-24, 4.69094e-24], [426.131, 1.5847e-23, 9.51509e-24, 5.18505e-24, 4.69151e-24], [427.114, 1.58815e-23, 9.51817e-24, 5.20817e-24, 4.69213e-24], [428.099, 1.59161e-23, 9.5213e-24, 5.2315e-24, 4.69281e-24], [429.086, 1.59508e-23, 9.52448e-24, 5.25507e-24, 4.69359e-24], [430.075, 1.59857e-23, 9.52774e-24, 5.27891e-24, 4.6945e-24], [431.067, 1.60207e-23, 9.53111e-24, 5.30309e-24, 4.69563e-24], [432.061, 1.60558e-23, 9.53468e-24, 5.32774e-24, 4.69711e-24], [433.057, 1.60913e-23, 9.53858e-24, 5.35312e-24, 4.69925e-24], [434.056, 1.61272e-23, 9.54317e-24, 5.37986e-24, 4.70278e-24], [435.057, 1.61643e-23, 9.54951e-24, 5.40978e-24, 4.70981e-24], [436.06, 1.62049e-23, 9.5618e-24, 5.45024e-24, 4.72887e-24], [437.066, 1.62694e-23, 9.61449e-24, 5.56082e-24, 4.8288e-24], [438.074, 1.06875e-22, 1.06061e-22, 1.05772e-22, 1.05735e-22], [439.084, 1.63376e-23, 9.6166e-24, 5.60229e-24, 4.82147e-24], [440.096, 1.63459e-23, 9.57339e-24, 5.54721e-24, 4.72876e-24], [441.111, 1.63765e-23, 9.56794e-24, 5.55746e-24, 4.71176e-24], [442.129, 1.64105e-23, 9.56815e-24, 5.57765e-24, 4.70618e-24], [443.148, 1.64455e-23, 9.57001e-24, 5.60083e-24, 4.70392e-24], [444.17, 1.6481e-23, 9.57253e-24, 5.62529e-24, 4.70297e-24], [445.194, 1.65167e-23, 9.57536e-24, 5.65044e-24, 4.70264e-24], [446.221, 1.65526e-23, 9.57837e-24, 5.67603e-24, 4.70265e-24], [447.25, 1.65887e-23, 9.5815e-24, 5.70194e-24, 4.70285e-24], [448.281, 1.66249e-23, 9.58469e-24, 5.72812e-24, 4.70318e-24], [449.315, 1.66612e-23, 9.58794e-24, 5.75454e-24, 4.7036e-24], [450.351, 1.66976e-23, 9.59124e-24, 5.78116e-24, 4.70407e-24], [451.39, 1.67341e-23, 9.59457e-24, 5.80798e-24, 4.7046e-24], [452.43, 1.67707e-23, 9.59793e-24, 5.83499e-24, 4.70515e-24], [453.474, 1.68074e-23, 9.60131e-24, 5.86218e-24, 4.70574e-24], [454.519, 1.68442e-23, 9.60473e-24, 5.88955e-24, 4.70635e-24], [455.568, 1.68811e-23, 9.60816e-24, 5.91709e-24, 4.70698e-24], [456.618, 1.69181e-23, 9.61162e-24, 5.94481e-24, 4.70763e-24], [457.671, 1.69551e-23, 9.61509e-24, 5.97271e-24, 4.70829e-24], [458.727, 1.69923e-23, 9.61859e-24, 6.00077e-24, 4.70897e-24], [459.784, 1.70295e-23, 9.62211e-24, 6.02901e-24, 4.70966e-24], [460.845, 1.70669e-23, 9.62565e-24, 6.05742e-24, 4.71037e-24], [461.907, 1.71043e-23, 9.62921e-24, 6.086e-24, 4.71108e-24], [462.972, 1.71418e-23, 9.63278e-24, 6.11475e-24, 4.71181e-24], [464.04, 1.71795e-23, 9.63638e-24, 6.14366e-24, 4.71255e-24], [465.11, 1.72172e-23, 9.64e-24, 6.17275e-24, 4.71331e-24], [466.183, 1.7255e-23, 9.64363e-24, 6.20201e-24, 4.71407e-24], [467.258, 1.72929e-23, 9.64728e-24, 6.23144e-24, 4.71484e-24], [468.335, 1.73309e-23, 9.65096e-24, 6.26103e-24, 4.71563e-24], [469.415, 1.73689e-23, 9.65465e-24, 6.2908e-24, 4.71642e-24], [470.498, 1.74071e-23, 9.65836e-24, 6.32073e-24, 4.71723e-24], [471.583, 1.74454e-23, 9.66209e-24, 6.35084e-24, 4.71804e-24], [472.67, 1.74837e-23, 9.66583e-24, 6.38111e-24, 4.71887e-24], [473.76, 1.75222e-23, 9.6696e-24, 6.41155e-24, 4.71971e-24], [474.853, 1.75607e-23, 9.67339e-24, 6.44217e-24, 4.72055e-24], [475.948, 1.75994e-23, 9.67719e-24, 6.47295e-24, 4.72141e-24], [477.045, 1.76381e-23, 9.68102e-24, 6.5039e-24, 4.72227e-24], [478.145, 1.76769e-23, 9.68486e-24, 6.53502e-24, 4.72315e-24], [479.248, 1.77158e-23, 9.68872e-24, 6.56632e-24, 4.72404e-24], [480.353, 1.77548e-23, 9.69261e-24, 6.59778e-24, 4.72493e-24], [481.461, 1.7794e-23, 9.69651e-24, 6.62941e-24, 4.72584e-24], [482.571, 1.78332e-23, 9.70043e-24, 6.66122e-24, 4.72676e-24], [483.684, 1.78725e-23, 9.70437e-24, 6.69319e-24, 4.72768e-24], [484.799, 1.79118e-23, 9.70833e-24, 6.72534e-24, 4.72862e-24], [485.917, 1.79513e-23, 9.71231e-24, 6.75765e-24, 4.72957e-24], [487.038, 1.79909e-23, 9.7163e-24, 6.79014e-24, 4.73052e-24], [488.161, 1.80306e-23, 9.72032e-24, 6.8228e-24, 4.73149e-24], [489.286, 1.80704e-23, 9.72436e-24, 6.85563e-24, 4.73247e-24], [490.415, 1.81102e-23, 9.72842e-24, 6.88863e-24, 4.73345e-24], [491.546, 1.81502e-23, 9.7325e-24, 6.9218e-24, 4.73445e-24], [492.679, 1.81903e-23, 9.73659e-24, 6.95514e-24, 4.73546e-24], [493.815, 1.82304e-23, 9.74071e-24, 6.98866e-24, 4.73648e-24], [494.954, 1.82707e-23, 9.74485e-24, 7.02234e-24, 4.7375e-24], [496.095, 1.8311e-23, 9.749e-24, 7.0562e-24, 4.73854e-24], [497.239, 1.83515e-23, 9.75318e-24, 7.09024e-24, 4.73959e-24], [498.386, 1.8392e-23, 9.75738e-24, 7.12444e-24, 4.74065e-24], [499.535, 1.84327e-23, 9.7616e-24, 7.15882e-24, 4.74172e-24], [500.687, 1.84734e-23, 9.76584e-24, 7.19337e-24, 4.7428e-24], [501.842, 1.85143e-23, 9.7701e-24, 7.22809e-24, 4.74389e-24], [502.999, 1.85552e-23, 9.77438e-24, 7.26298e-24, 4.74499e-24], [504.159, 1.85962e-23, 9.77868e-24, 7.29805e-24, 4.7461e-24], [505.322, 1.86374e-23, 9.783e-24, 7.3333e-24, 4.74722e-24], [506.487, 1.86786e-23, 9.78734e-24, 7.36871e-24, 4.74835e-24], [507.655, 1.872e-23, 9.7917e-24, 7.4043e-24, 4.74949e-24], [508.825, 1.87614e-23, 9.79608e-24, 7.44006e-24, 4.75064e-24], [509.999, 1.88029e-23, 9.80049e-24, 7.476e-24, 4.75181e-24], [511.175, 1.88446e-23, 9.80491e-24, 7.51211e-24, 4.75298e-24], [512.354, 1.88863e-23, 9.80936e-24, 7.5484e-24, 4.75416e-24], [513.535, 1.89281e-23, 9.81382e-24, 7.58486e-24, 4.75536e-24], [514.719, 1.89701e-23, 9.81831e-24, 7.6215e-24, 4.75656e-24], [515.906, 1.90121e-23, 9.82282e-24, 7.65831e-24, 4.75778e-24], [517.096, 1.90542e-23, 9.82735e-24, 7.69529e-24, 4.759e-24], [518.288, 1.90965e-23, 9.8319e-24, 7.73246e-24, 4.76024e-24], [519.484, 1.91388e-23, 9.83648e-24, 7.76979e-24, 4.76149e-24], [520.682, 1.91813e-23, 9.84107e-24, 7.80731e-24, 4.76275e-24], [521.882, 1.92238e-23, 9.84569e-24, 7.84499e-24, 4.76402e-24], [523.086, 1.92664e-23, 9.85033e-24, 7.88286e-24, 4.7653e-24], [524.292, 1.93092e-23, 9.85499e-24, 7.9209e-24, 4.76659e-24], [525.501, 1.9352e-23, 9.85967e-24, 7.95912e-24, 4.76789e-24], [526.713, 1.9395e-23, 9.86437e-24, 7.99751e-24, 4.76921e-24], [527.927, 1.9438e-23, 9.8691e-24, 8.03608e-24, 4.77053e-24], [529.145, 1.94812e-23, 9.87385e-24, 8.07483e-24, 4.77186e-24], [530.365, 1.95245e-23, 9.87862e-24, 8.11376e-24, 4.77321e-24], [531.588, 1.95678e-23, 9.88341e-24, 8.15286e-24, 4.77457e-24], [532.814, 1.96113e-23, 9.88823e-24, 8.19214e-24, 4.77594e-24], [534.043, 1.96549e-23, 9.89306e-24, 8.2316e-24, 4.77732e-24], [535.274, 1.96985e-23, 9.89792e-24, 8.27124e-24, 4.77871e-24], [536.508, 1.97423e-23, 9.9028e-24, 8.31106e-24, 4.78011e-24], [537.746, 1.97862e-23, 9.90771e-24, 8.35105e-24, 4.78152e-24], [538.986, 1.98302e-23, 9.91264e-24, 8.39122e-24, 4.78295e-24], [540.229, 1.98743e-23, 9.91759e-24, 8.43158e-24, 4.78438e-24], [541.474, 1.99185e-23, 9.92256e-24, 8.47211e-24, 4.78583e-24], [542.723, 1.99628e-23, 9.92755e-24, 8.51282e-24, 4.78729e-24], [543.974, 2.00072e-23, 9.93257e-24, 8.55371e-24, 4.78876e-24], [545.229, 2.00517e-23, 9.93762e-24, 8.59478e-24, 4.79024e-24], [546.486, 2.00964e-23, 9.94268e-24, 8.63603e-24, 4.79173e-24], [547.746, 2.01411e-23, 9.94777e-24, 8.67746e-24, 4.79323e-24], [549.01, 2.01859e-23, 9.95288e-24, 8.71907e-24, 4.79475e-24], [550.276, 2.02309e-23, 9.95801e-24, 8.76087e-24, 4.79628e-24], [551.544, 2.02759e-23, 9.96317e-24, 8.80284e-24, 4.79781e-24], [552.816, 2.03211e-23, 9.96835e-24, 8.84499e-24, 4.79936e-24], [554.091, 2.03664e-23, 9.97356e-24, 8.88733e-24, 4.80093e-24], [555.369, 2.04117e-23, 9.97879e-24, 8.92985e-24, 4.8025e-24], [556.65, 2.04572e-23, 9.98404e-24, 8.97255e-24, 4.80409e-24], [557.933, 2.05028e-23, 9.98932e-24, 9.01543e-24, 4.80568e-24], [559.22, 2.05485e-23, 9.99462e-24, 9.0585e-24, 4.80729e-24], [560.509, 2.05943e-23, 9.99994e-24, 9.10174e-24, 4.80891e-24], [561.802, 2.06402e-23, 1.00053e-23, 9.14517e-24, 4.81054e-24], [563.097, 2.06863e-23, 1.00107e-23, 9.18879e-24, 4.81219e-24], [564.396, 2.07324e-23, 1.00161e-23, 9.23258e-24, 4.81385e-24], [565.697, 2.07787e-23, 1.00215e-23, 9.27656e-24, 4.81551e-24], [567.002, 2.0825e-23, 1.00269e-23, 9.32073e-24, 4.81719e-24], [568.309, 2.08715e-23, 1.00324e-23, 9.36507e-24, 4.81889e-24], [569.62, 2.09181e-23, 1.00379e-23, 9.40961e-24, 4.82059e-24], [570.933, 2.09648e-23, 1.00434e-23, 9.45432e-24, 4.82231e-24], [572.25, 2.10116e-23, 1.0049e-23, 9.49923e-24, 4.82404e-24], [573.57, 2.10585e-23, 1.00545e-23, 9.54431e-24, 4.82578e-24], [574.892, 2.11055e-23, 1.00601e-23, 9.58959e-24, 4.82753e-24], [576.218, 2.11526e-23, 1.00657e-23, 9.63504e-24, 4.82929e-24], [577.547, 2.11999e-23, 1.00714e-23, 9.68069e-24, 4.83107e-24], [578.879, 2.12472e-23, 1.00771e-23, 9.72652e-24, 4.83286e-24], [580.214, 2.12947e-23, 1.00827e-23, 9.77253e-24, 4.83466e-24], [581.552, 2.13423e-23, 1.00885e-23, 9.81874e-24, 4.83648e-24], [582.893, 2.139e-23, 1.00942e-23, 9.86513e-24, 4.8383e-24], [584.237, 2.14378e-23, 1.01e-23, 9.9117e-24, 4.84014e-24], [585.584, 2.14857e-23, 1.01058e-23, 9.95847e-24, 4.842e-24], [586.934, 2.15338e-23, 1.01116e-23, 1.00054e-23, 4.84386e-24], [588.288, 2.15819e-23, 1.01174e-23, 1.00526e-23, 4.84574e-24], [589.644, 2.16302e-23, 1.01233e-23, 1.00999e-23, 4.84763e-24], [591.004, 2.16786e-23, 1.01292e-23, 1.01474e-23, 4.84953e-24], [592.367, 2.17271e-23, 1.01351e-23, 1.01951e-23, 4.85144e-24], [593.733, 2.17757e-23, 1.01411e-23, 1.0243e-23, 4.85337e-24], [595.102, 2.18244e-23, 1.01471e-23, 1.02911e-23, 4.85531e-24], [596.474, 2.18732e-23, 1.01531e-23, 1.03394e-23, 4.85726e-24], [597.85, 2.19222e-23, 1.01591e-23, 1.03878e-23, 4.85923e-24], [599.229, 2.19713e-23, 1.01652e-23, 1.04365e-23, 4.86121e-24], [600.61, 2.20205e-23, 1.01712e-23, 1.04853e-23, 4.8632e-24], [601.995, 2.20698e-23, 1.01773e-23, 1.05343e-23, 4.8652e-24], [603.384, 2.21192e-23, 1.01835e-23, 1.05836e-23, 4.86722e-24], [604.775, 2.21687e-23, 1.01896e-23, 1.0633e-23, 4.86925e-24], [606.17, 2.22184e-23, 1.01958e-23, 1.06826e-23, 4.8713e-24], [607.567, 2.22682e-23, 1.02021e-23, 1.07324e-23, 4.87335e-24], [608.969, 2.23181e-23, 1.02083e-23, 1.07824e-23, 4.87542e-24], [610.373, 2.23681e-23, 1.02146e-23, 1.08326e-23, 4.87751e-24], [611.78, 2.24182e-23, 1.02209e-23, 1.08829e-23, 4.8796e-24], [613.191, 2.24684e-23, 1.02272e-23, 1.09335e-23, 4.88171e-24], [614.605, 2.25188e-23, 1.02335e-23, 1.09843e-23, 4.88383e-24], [616.022, 2.25693e-23, 1.02399e-23, 1.10352e-23, 4.88597e-24], [617.443, 2.26199e-23, 1.02463e-23, 1.10864e-23, 4.88812e-24], [618.867, 2.26706e-23, 1.02528e-23, 1.11377e-23, 4.89028e-24], [620.294, 2.27215e-23, 1.02592e-23, 1.11893e-23, 4.89246e-24], [621.724, 2.27724e-23, 1.02657e-23, 1.1241e-23, 4.89465e-24], [623.158, 2.28235e-23, 1.02722e-23, 1.12929e-23, 4.89685e-24], [624.595, 2.28747e-23, 1.02788e-23, 1.1345e-23, 4.89907e-24], [626.035, 2.29261e-23, 1.02854e-23, 1.13973e-23, 4.9013e-24], [627.479, 2.29775e-23, 1.0292e-23, 1.14499e-23, 4.90354e-24], [628.926, 2.30291e-23, 1.02986e-23, 1.15026e-23, 4.9058e-24], [630.376, 2.30808e-23, 1.03052e-23, 1.15555e-23, 4.90807e-24], [631.83, 2.31326e-23, 1.03119e-23, 1.16086e-23, 4.91035e-24], [633.287, 2.31845e-23, 1.03187e-23, 1.16619e-23, 4.91265e-24], [634.747, 2.32366e-23, 1.03254e-23, 1.17154e-23, 4.91496e-24], [636.211, 2.32888e-23, 1.03322e-23, 1.1769e-23, 4.91729e-24], [637.678, 2.33411e-23, 1.0339e-23, 1.18229e-23, 4.91963e-24], [639.149, 2.33935e-23, 1.03458e-23, 1.1877e-23, 4.92198e-24], [640.623, 2.34461e-23, 1.03527e-23, 1.19313e-23, 4.92435e-24], [642.1, 2.34987e-23, 1.03596e-23, 1.19858e-23, 4.92673e-24], [643.58, 2.35515e-23, 1.03665e-23, 1.20405e-23, 4.92913e-24], [645.065, 2.36045e-23, 1.03734e-23, 1.20954e-23, 4.93154e-24], [646.552, 2.36575e-23, 1.03804e-23, 1.21504e-23, 4.93396e-24], [648.043, 2.37107e-23, 1.03874e-23, 1.22057e-23, 4.9364e-24], [649.537, 2.3764e-23, 1.03944e-23, 1.22612e-23, 4.93885e-24], [651.035, 2.38174e-23, 1.04015e-23, 1.23169e-23, 4.94132e-24], [652.537, 2.3871e-23, 1.04086e-23, 1.23728e-23, 4.9438e-24], [654.041, 2.39247e-23, 1.04157e-23, 1.24288e-23, 4.9463e-24], [655.55, 2.39785e-23, 1.04229e-23, 1.24851e-23, 4.94881e-24], [657.061, 2.40324e-23, 1.04301e-23, 1.25416e-23, 4.95133e-24], [658.576, 2.40865e-23, 1.04373e-23, 1.25983e-23, 4.95387e-24], [660.095, 2.41406e-23, 1.04445e-23, 1.26552e-23, 4.95643e-24], [661.617, 2.4195e-23, 1.04518e-23, 1.27123e-23, 4.95899e-24], [663.143, 2.42494e-23, 1.04591e-23, 1.27696e-23, 4.96158e-24], [664.672, 2.4304e-23, 1.04664e-23, 1.28271e-23, 4.96417e-24], [666.205, 2.43587e-23, 1.04738e-23, 1.28848e-23, 4.96678e-24], [667.741, 2.44135e-23, 1.04812e-23, 1.29427e-23, 4.96941e-24], [669.281, 2.44685e-23, 1.04886e-23, 1.30008e-23, 4.97205e-24], [670.824, 2.45235e-23, 1.04961e-23, 1.30591e-23, 4.97471e-24], [672.371, 2.45788e-23, 1.05036e-23, 1.31176e-23, 4.97738e-24], [673.922, 2.46341e-23, 1.05111e-23, 1.31763e-23, 4.98007e-24], [675.476, 2.46896e-23, 1.05187e-23, 1.32352e-23, 4.98277e-24], [677.034, 2.47452e-23, 1.05262e-23, 1.32944e-23, 4.98548e-24], [678.595, 2.48009e-23, 1.05339e-23, 1.33537e-23, 4.98821e-24], [680.16, 2.48568e-23, 1.05415e-23, 1.34133e-23, 4.99096e-24], [681.728, 2.49128e-23, 1.05492e-23, 1.3473e-23, 4.99372e-24], [683.3, 2.49689e-23, 1.05569e-23, 1.3533e-23, 4.9965e-24], [684.876, 2.50252e-23, 1.05646e-23, 1.35931e-23, 4.99929e-24], [686.455, 2.50816e-23, 1.05724e-23, 1.36535e-23, 5.00209e-24], [688.038, 2.51381e-23, 1.05802e-23, 1.37141e-23, 5.00492e-24], [689.625, 2.51948e-23, 1.05881e-23, 1.37749e-23, 5.00775e-24], [691.215, 2.52516e-23, 1.05959e-23, 1.38358e-23, 5.0106e-24], [692.809, 2.53085e-23, 1.06038e-23, 1.3897e-23, 5.01347e-24], [694.407, 2.53656e-23, 1.06118e-23, 1.39585e-23, 5.01636e-24], [696.008, 2.54228e-23, 1.06197e-23, 1.40201e-23, 5.01925e-24], [697.613, 2.54801e-23, 1.06277e-23, 1.40819e-23, 5.02217e-24], [699.222, 2.55376e-23, 1.06358e-23, 1.41439e-23, 5.0251e-24], [700.834, 2.55952e-23, 1.06438e-23, 1.42062e-23, 5.02804e-24], [702.45, 2.56529e-23, 1.06519e-23, 1.42687e-23, 5.03101e-24], [704.07, 2.57108e-23, 1.06601e-23, 1.43313e-23, 5.03398e-24], [705.694, 2.57688e-23, 1.06682e-23, 1.43942e-23, 5.03698e-24], [707.321, 2.58269e-23, 1.06764e-23, 1.44573e-23, 5.03998e-24], [708.952, 2.58852e-23, 1.06846e-23, 1.45206e-23, 5.04301e-24], [710.587, 2.59436e-23, 1.06929e-23, 1.45841e-23, 5.04605e-24], [712.226, 2.60022e-23, 1.07012e-23, 1.46478e-23, 5.0491e-24], [713.868, 2.60609e-23, 1.07095e-23, 1.47118e-23, 5.05218e-24], [715.514, 2.61197e-23, 1.07179e-23, 1.47759e-23, 5.05526e-24], [717.164, 2.61787e-23, 1.07263e-23, 1.48403e-23, 5.05837e-24], [718.818, 2.62378e-23, 1.07347e-23, 1.49049e-23, 5.06149e-24], [720.475, 2.6297e-23, 1.07432e-23, 1.49697e-23, 5.06463e-24], [722.137, 2.63564e-23, 1.07517e-23, 1.50347e-23, 5.06778e-24], [723.802, 2.64159e-23, 1.07602e-23, 1.50999e-23, 5.07095e-24], [725.471, 2.64756e-23, 1.07688e-23, 1.51653e-23, 5.07413e-24], [727.144, 2.65354e-23, 1.07774e-23, 1.5231e-23, 5.07733e-24], [728.821, 2.65953e-23, 1.0786e-23, 1.52969e-23, 5.08055e-24], [730.502, 2.66554e-23, 1.07947e-23, 1.53629e-23, 5.08379e-24], [732.186, 2.67156e-23, 1.08034e-23, 1.54292e-23, 5.08704e-24], [733.875, 2.6776e-23, 1.08122e-23, 1.54958e-23, 5.09031e-24], [735.567, 2.68365e-23, 1.08209e-23, 1.55625e-23, 5.09359e-24], [737.263, 2.68972e-23, 1.08298e-23, 1.56294e-23, 5.09689e-24], [738.963, 2.6958e-23, 1.08386e-23, 1.56966e-23, 5.10021e-24], [740.667, 2.70189e-23, 1.08475e-23, 1.5764e-23, 5.10354e-24], [742.375, 2.708e-23, 1.08564e-23, 1.58316e-23, 5.10689e-24], [744.087, 2.71412e-23, 1.08654e-23, 1.58994e-23, 5.11026e-24], [745.803, 2.72025e-23, 1.08743e-23, 1.59674e-23, 5.11364e-24], [747.523, 2.72641e-23, 1.08834e-23, 1.60357e-23, 5.11705e-24], [749.247, 2.73257e-23, 1.08924e-23, 1.61042e-23, 5.12046e-24], [750.975, 2.73875e-23, 1.09015e-23, 1.61729e-23, 5.1239e-24], [752.706, 2.74494e-23, 1.09107e-23, 1.62418e-23, 5.12735e-24], [754.442, 2.75115e-23, 1.09198e-23, 1.63109e-23, 5.13082e-24], [756.182, 2.75738e-23, 1.0929e-23, 1.63803e-23, 5.13431e-24], [757.926, 2.76361e-23, 1.09383e-23, 1.64498e-23, 5.13781e-24], [759.673, 2.76987e-23, 1.09475e-23, 1.65196e-23, 5.14133e-24], [761.425, 2.77613e-23, 1.09569e-23, 1.65897e-23, 5.14487e-24], [763.181, 2.78241e-23, 1.09662e-23, 1.66599e-23, 5.14843e-24], [764.941, 2.78871e-23, 1.09756e-23, 1.67304e-23, 5.152e-24], [766.705, 2.79502e-23, 1.0985e-23, 1.6801e-23, 5.15559e-24], [768.473, 2.80135e-23, 1.09945e-23, 1.68719e-23, 5.1592e-24], [770.245, 2.80769e-23, 1.1004e-23, 1.69431e-23, 5.16282e-24], [772.021, 2.81404e-23, 1.10135e-23, 1.70144e-23, 5.16647e-24], [773.801, 2.82041e-23, 1.10231e-23, 1.7086e-23, 5.17013e-24], [775.586, 2.8268e-23, 1.10327e-23, 1.71578e-23, 5.17381e-24], [777.374, 2.8332e-23, 1.10423e-23, 1.72298e-23, 5.1775e-24], [779.167, 2.83961e-23, 1.1052e-23, 1.73021e-23, 5.18122e-24], [780.964, 2.84604e-23, 1.10617e-23, 1.73745e-23, 5.18495e-24], [782.765, 2.85249e-23, 1.10715e-23, 1.74472e-23, 5.1887e-24], [784.57, 2.85895e-23, 1.10813e-23, 1.75202e-23, 5.19247e-24], [786.379, 2.86542e-23, 1.10911e-23, 1.75933e-23, 5.19625e-24], [788.192, 2.87191e-23, 1.1101e-23, 1.76667e-23, 5.20006e-24], [790.01, 2.87842e-23, 1.11109e-23, 1.77403e-23, 5.20388e-24], [791.832, 2.88494e-23, 1.11209e-23, 1.78141e-23, 5.20772e-24], [793.658, 2.89148e-23, 1.11309e-23, 1.78882e-23, 5.21158e-24], [795.488, 2.89803e-23, 1.11409e-23, 1.79625e-23, 5.21545e-24], [797.322, 2.90459e-23, 1.1151e-23, 1.8037e-23, 5.21935e-24], [799.161, 2.91118e-23, 1.11611e-23, 1.81117e-23, 5.22326e-24], [801.004, 2.91777e-23, 1.11712e-23, 1.81867e-23, 5.22719e-24], [802.851, 2.92439e-23, 1.11814e-23, 1.82619e-23, 5.23114e-24], [804.702, 2.93101e-23, 1.11916e-23, 1.83373e-23, 5.23511e-24], [806.558, 2.93766e-23, 1.12019e-23, 1.84129e-23, 5.2391e-24], [808.418, 2.94432e-23, 1.12122e-23, 1.84888e-23, 5.2431e-24], [810.282, 2.95099e-23, 1.12225e-23, 1.85649e-23, 5.24713e-24], [812.151, 2.95768e-23, 1.12329e-23, 1.86413e-23, 5.25117e-24], [814.023, 2.96439e-23, 1.12433e-23, 1.87178e-23, 5.25523e-24], [815.901, 2.97111e-23, 1.12538e-23, 1.87947e-23, 5.25931e-24], [817.782, 2.97785e-23, 1.12643e-23, 1.88717e-23, 5.26341e-24], [819.668, 2.9846e-23, 1.12748e-23, 1.8949e-23, 5.26753e-24], [821.558, 2.99137e-23, 1.12854e-23, 1.90265e-23, 5.27167e-24], [823.452, 2.99815e-23, 1.1296e-23, 1.91042e-23, 5.27583e-24], [825.351, 3.00495e-23, 1.13067e-23, 1.91821e-23, 5.28e-24], [827.255, 3.01177e-23, 1.13174e-23, 1.92603e-23, 5.2842e-24], [829.162, 3.0186e-23, 1.13281e-23, 1.93388e-23, 5.28842e-24], [831.074, 3.02545e-23, 1.13389e-23, 1.94174e-23, 5.29265e-24], [832.991, 3.03232e-23, 1.13497e-23, 1.94963e-23, 5.29691e-24], [834.912, 3.0392e-23, 1.13606e-23, 1.95755e-23, 5.30119e-24], [836.837, 3.04609e-23, 1.13715e-23, 1.96548e-23, 5.30549e-24], [838.767, 3.053e-23, 1.13825e-23, 1.97344e-23, 5.30981e-24], [840.701, 3.05993e-23, 1.13935e-23, 1.98143e-23, 5.31415e-24], [842.64, 3.06688e-23, 1.14045e-23, 1.98943e-23, 5.31852e-24], [844.583, 3.07384e-23, 1.14156e-23, 1.99746e-23, 5.32293e-24], [846.53, 3.08082e-23, 1.14267e-23, 2.00552e-23, 5.32736e-24], [848.482, 3.08781e-23, 1.14379e-23, 2.0136e-23, 5.33185e-24], [850.439, 3.09483e-23, 1.14492e-23, 2.0217e-23, 5.33642e-24], [852.4, 3.10186e-23, 1.14606e-23, 2.02984e-23, 5.34117e-24], [854.366, 3.10891e-23, 1.14722e-23, 2.03801e-23, 5.34642e-24], [856.336, 3.11603e-23, 1.14851e-23, 2.04627e-23, 5.35428e-24], [858.311, 3.12528e-23, 1.15556e-23, 2.05777e-23, 5.48441e-24], [860.29, 3.13032e-23, 1.15114e-23, 2.06288e-23, 5.37083e-24], [862.274, 3.13726e-23, 1.15181e-23, 2.07086e-23, 5.36525e-24], [864.262, 3.14436e-23, 1.1529e-23, 2.07909e-23, 5.36844e-24], [866.255, 3.15149e-23, 1.15403e-23, 2.08737e-23, 5.37259e-24], [868.253, 3.15865e-23, 1.15518e-23, 2.09568e-23, 5.37701e-24], [870.255, 3.16582e-23, 1.15634e-23, 2.10402e-23, 5.38154e-24], [872.262, 3.17301e-23, 1.1575e-23, 2.11238e-23, 5.38613e-24], [874.273, 3.18022e-23, 1.15867e-23, 2.12077e-23, 5.39076e-24], [876.289, 3.18745e-23, 1.15984e-23, 2.12918e-23, 5.39542e-24], [878.31, 3.19469e-23, 1.16102e-23, 2.13761e-23, 5.40011e-24], [880.336, 3.20195e-23, 1.16221e-23, 2.14607e-23, 5.40482e-24], [882.366, 3.20922e-23, 1.16339e-23, 2.15455e-23, 5.40956e-24], [884.4, 3.21651e-23, 1.16459e-23, 2.16306e-23, 5.41432e-24], [886.44, 3.22382e-23, 1.16578e-23, 2.17159e-23, 5.41911e-24], [888.484, 3.23115e-23, 1.16698e-23, 2.18015e-23, 5.42391e-24], [890.533, 3.23849e-23, 1.16819e-23, 2.18873e-23, 5.42874e-24], [892.586, 3.24586e-23, 1.1694e-23, 2.19734e-23, 5.43359e-24], [894.645, 3.25323e-23, 1.17061e-23, 2.20596e-23, 5.43846e-24], [896.708, 3.26063e-23, 1.17183e-23, 2.21462e-23, 5.44335e-24], [898.775, 3.26804e-23, 1.17306e-23, 2.2233e-23, 5.44826e-24], [900.848, 3.27547e-23, 1.17429e-23, 2.232e-23, 5.4532e-24], [902.925, 3.28292e-23, 1.17552e-23, 2.24073e-23, 5.45815e-24], [905.008, 3.29038e-23, 1.17676e-23, 2.24948e-23, 5.46313e-24], [907.095, 3.29786e-23, 1.178e-23, 2.25825e-23, 5.46813e-24], [909.186, 3.30536e-23, 1.17924e-23, 2.26705e-23, 5.47314e-24], [911.283, 3.31288e-23, 1.18049e-23, 2.27588e-23, 5.47818e-24], [913.384, 3.32041e-23, 1.18175e-23, 2.28473e-23, 5.48325e-24], [915.491, 3.32797e-23, 1.18301e-23, 2.2936e-23, 5.48833e-24], [917.602, 3.33553e-23, 1.18428e-23, 2.3025e-23, 5.49344e-24], [919.718, 3.34312e-23, 1.18554e-23, 2.31143e-23, 5.49856e-24], [921.839, 3.35073e-23, 1.18682e-23, 2.32038e-23, 5.50371e-24], [923.964, 3.35835e-23, 1.1881e-23, 2.32935e-23, 5.50888e-24], [926.095, 3.36599e-23, 1.18938e-23, 2.33835e-23, 5.51407e-24], [928.231, 3.37365e-23, 1.19067e-23, 2.34737e-23, 5.51928e-24], [930.371, 3.38132e-23, 1.19196e-23, 2.35642e-23, 5.52452e-24], [932.517, 3.38901e-23, 1.19326e-23, 2.36549e-23, 5.52978e-24], [934.667, 3.39673e-23, 1.19456e-23, 2.37459e-23, 5.53505e-24], [936.822, 3.40446e-23, 1.19587e-23, 2.38371e-23, 5.54036e-24], [938.983, 3.4122e-23, 1.19718e-23, 2.39286e-23, 5.54568e-24], [941.148, 3.41997e-23, 1.19849e-23, 2.40203e-23, 5.55102e-24], [943.318, 3.42775e-23, 1.19981e-23, 2.41123e-23, 5.55639e-24], [945.494, 3.43555e-23, 1.20114e-23, 2.42046e-23, 5.56178e-24], [947.674, 3.44337e-23, 1.20247e-23, 2.4297e-23, 5.56719e-24], [949.859, 3.45121e-23, 1.2038e-23, 2.43898e-23, 5.57263e-24], [952.05, 3.45907e-23, 1.20514e-23, 2.44828e-23, 5.57808e-24], [954.245, 3.46694e-23, 1.20649e-23, 2.4576e-23, 5.58356e-24], [956.445, 3.47483e-23, 1.20784e-23, 2.46695e-23, 5.58906e-24], [958.651, 3.48274e-23, 1.20919e-23, 2.47632e-23, 5.59458e-24], [960.862, 3.49067e-23, 1.21055e-23, 2.48572e-23, 5.60013e-24], [963.077, 3.49862e-23, 1.21192e-23, 2.49515e-23, 5.6057e-24], [965.298, 3.50659e-23, 1.21328e-23, 2.5046e-23, 5.61129e-24], [967.524, 3.51457e-23, 1.21466e-23, 2.51408e-23, 5.61691e-24], [969.755, 3.52258e-23, 1.21604e-23, 2.52358e-23, 5.62254e-24], [971.992, 3.5306e-23, 1.21742e-23, 2.5331e-23, 5.6282e-24], [974.233, 3.53864e-23, 1.21881e-23, 2.54266e-23, 5.63389e-24], [976.48, 3.5467e-23, 1.2202e-23, 2.55224e-23, 5.63959e-24], [978.731, 3.55478e-23, 1.2216e-23, 2.56184e-23, 5.64532e-24], [980.988, 3.56287e-23, 1.223e-23, 2.57147e-23, 5.65107e-24], [983.251, 3.57099e-23, 1.22441e-23, 2.58112e-23, 5.65685e-24], [985.518, 3.57912e-23, 1.22583e-23, 2.5908e-23, 5.66265e-24], [987.791, 3.58728e-23, 1.22724e-23, 2.60051e-23, 5.66847e-24], [990.068, 3.59545e-23, 1.22867e-23, 2.61024e-23, 5.67431e-24], [992.352, 3.60364e-23, 1.2301e-23, 2.62e-23, 5.68018e-24], [994.64, 3.61185e-23, 1.23153e-23, 2.62978e-23, 5.68607e-24], [996.934, 3.62008e-23, 1.23297e-23, 2.63959e-23, 5.69199e-24], [999.233, 3.62833e-23, 1.23441e-23, 2.64943e-23, 5.69793e-24], [1001.54, 3.6366e-23, 1.23586e-23, 2.65929e-23, 5.70389e-24], [1003.85, 3.64488e-23, 1.23731e-23, 2.66918e-23, 5.70988e-24], [1006.16, 3.65319e-23, 1.23877e-23, 2.67909e-23, 5.71589e-24], [1008.48, 3.66152e-23, 1.24023e-23, 2.68903e-23, 5.72192e-24], [1010.81, 3.66986e-23, 1.2417e-23, 2.699e-23, 5.72798e-24], [1013.14, 3.67822e-23, 1.24318e-23, 2.70899e-23, 5.73406e-24], [1015.47, 3.68661e-23, 1.24466e-23, 2.71901e-23, 5.74017e-24], [1017.82, 3.69501e-23, 1.24614e-23, 2.72905e-23, 5.7463e-24], [1020.16, 3.70343e-23, 1.24763e-23, 2.73912e-23, 5.75245e-24], [1022.52, 3.71188e-23, 1.24913e-23, 2.74921e-23, 5.75863e-24], [1024.87, 3.72034e-23, 1.25063e-23, 2.75934e-23, 5.76483e-24], [1027.24, 3.72882e-23, 1.25213e-23, 2.76949e-23, 5.77106e-24], [1029.61, 3.73732e-23, 1.25364e-23, 2.77966e-23, 5.77731e-24], [1031.98, 3.74584e-23, 1.25516e-23, 2.78986e-23, 5.78358e-24], [1034.36, 3.75438e-23, 1.25668e-23, 2.80009e-23, 5.78988e-24], [1036.74, 3.76294e-23, 1.2582e-23, 2.81034e-23, 5.79621e-24], [1039.14, 3.77152e-23, 1.25974e-23, 2.82062e-23, 5.80255e-24], [1041.53, 3.78012e-23, 1.26127e-23, 2.83093e-23, 5.80893e-24], [1043.93, 3.78875e-23, 1.26281e-23, 2.84126e-23, 5.81532e-24], [1046.34, 3.79739e-23, 1.26436e-23, 2.85162e-23, 5.82175e-24], [1048.75, 3.80605e-23, 1.26591e-23, 2.86201e-23, 5.82819e-24], [1051.17, 3.81473e-23, 1.26747e-23, 2.87242e-23, 5.83467e-24], [1053.6, 3.82343e-23, 1.26904e-23, 2.88286e-23, 5.84116e-24], [1056.03, 3.83215e-23, 1.27061e-23, 2.89333e-23, 5.84769e-24], [1058.46, 3.84089e-23, 1.27218e-23, 2.90382e-23, 5.85423e-24], [1060.9, 3.84965e-23, 1.27376e-23, 2.91434e-23, 5.86081e-24], [1063.35, 3.85843e-23, 1.27534e-23, 2.92489e-23, 5.8674e-24], [1065.8, 3.86723e-23, 1.27693e-23, 2.93546e-23, 5.87403e-24], [1068.26, 3.87606e-23, 1.27853e-23, 2.94606e-23, 5.88067e-24], [1070.72, 3.8849e-23, 1.28013e-23, 2.95669e-23, 5.88735e-24], [1073.19, 3.89376e-23, 1.28174e-23, 2.96735e-23, 5.89405e-24], [1075.67, 3.90265e-23, 1.28335e-23, 2.97803e-23, 5.90077e-24], [1078.15, 3.91155e-23, 1.28497e-23, 2.98873e-23, 5.90752e-24], [1080.63, 3.92048e-23, 1.28659e-23, 2.99947e-23, 5.9143e-24], [1083.12, 3.92942e-23, 1.28822e-23, 3.01023e-23, 5.9211e-24], [1085.62, 3.93839e-23, 1.28986e-23, 3.02102e-23, 5.92793e-24], [1088.13, 3.94738e-23, 1.2915e-23, 3.03184e-23, 5.93478e-24], [1090.63, 3.95639e-23, 1.29314e-23, 3.04268e-23, 5.94166e-24], [1093.15, 3.96542e-23, 1.29479e-23, 3.05355e-23, 5.94856e-24], [1095.67, 3.97447e-23, 1.29645e-23, 3.06445e-23, 5.95549e-24], [1098.2, 3.98354e-23, 1.29811e-23, 3.07537e-23, 5.96245e-24], [1100.73, 3.99263e-23, 1.29978e-23, 3.08633e-23, 5.96943e-24], [1103.27, 4.00175e-23, 1.30145e-23, 3.09731e-23, 5.97644e-24], [1105.81, 4.01088e-23, 1.30313e-23, 3.10831e-23, 5.98347e-24], [1108.36, 4.02004e-23, 1.30482e-23, 3.11935e-23, 5.99053e-24], [1110.92, 4.02921e-23, 1.30651e-23, 3.13041e-23, 5.99762e-24], [1113.48, 4.03841e-23, 1.30821e-23, 3.1415e-23, 6.00474e-24], [1116.05, 4.04763e-23, 1.30991e-23, 3.15262e-23, 6.01188e-24], [1118.62, 4.05687e-23, 1.31162e-23, 3.16376e-23, 6.01904e-24], [1121.2, 4.06614e-23, 1.31333e-23, 3.17493e-23, 6.02624e-24], [1123.79, 4.07542e-23, 1.31505e-23, 3.18613e-23, 6.03346e-24], [1126.38, 4.08472e-23, 1.31677e-23, 3.19736e-23, 6.0407e-24], [1128.97, 4.09405e-23, 1.31851e-23, 3.20861e-23, 6.04798e-24], [1131.58, 4.1034e-23, 1.32024e-23, 3.2199e-23, 6.05528e-24], [1134.19, 4.11277e-23, 1.32198e-23, 3.23121e-23, 6.06261e-24], [1136.8, 4.12216e-23, 1.32373e-23, 3.24254e-23, 6.06996e-24], [1139.42, 4.13158e-23, 1.32549e-23, 3.25391e-23, 6.07734e-24], [1142.05, 4.14101e-23, 1.32725e-23, 3.2653e-23, 6.08475e-24], [1144.69, 4.15047e-23, 1.32901e-23, 3.27673e-23, 6.09219e-24], [1147.33, 4.15995e-23, 1.33078e-23, 3.28817e-23, 6.09965e-24], [1149.97, 4.16945e-23, 1.33256e-23, 3.29965e-23, 6.10714e-24], [1152.62, 4.17898e-23, 1.33435e-23, 3.31116e-23, 6.11466e-24], [1155.28, 4.18852e-23, 1.33614e-23, 3.32269e-23, 6.1222e-24], [1157.94, 4.19809e-23, 1.33793e-23, 3.33425e-23, 6.12977e-24], [1160.62, 4.20768e-23, 1.33973e-23, 3.34584e-23, 6.13737e-24], [1163.29, 4.21729e-23, 1.34154e-23, 3.35746e-23, 6.145e-24], [1165.97, 4.22693e-23, 1.34335e-23, 3.36911e-23, 6.15266e-24], [1168.66, 4.23658e-23, 1.34517e-23, 3.38078e-23, 6.16034e-24], [1171.36, 4.24626e-23, 1.347e-23, 3.39248e-23, 6.16805e-24], [1174.06, 4.25596e-23, 1.34883e-23, 3.40421e-23, 6.17579e-24], [1176.77, 4.26569e-23, 1.35067e-23, 3.41597e-23, 6.18356e-24], [1179.48, 4.27544e-23, 1.35251e-23, 3.42776e-23, 6.19135e-24], [1182.2, 4.28521e-23, 1.35436e-23, 3.43957e-23, 6.19917e-24], [1184.93, 4.295e-23, 1.35622e-23, 3.45142e-23, 6.20703e-24], [1187.66, 4.30481e-23, 1.35808e-23, 3.46329e-23, 6.21491e-24], [1190.4, 4.31465e-23, 1.35995e-23, 3.47519e-23, 6.22281e-24], [1193.14, 4.32451e-23, 1.36182e-23, 3.48712e-23, 6.23075e-24], [1195.89, 4.33439e-23, 1.3637e-23, 3.49908e-23, 6.23871e-24], [1198.65, 4.3443e-23, 1.36559e-23, 3.51107e-23, 6.24671e-24], [1201.42, 4.35423e-23, 1.36748e-23, 3.52308e-23, 6.25473e-24], [1204.19, 4.36418e-23, 1.36938e-23, 3.53513e-23, 6.26278e-24], [1206.96, 4.37415e-23, 1.37128e-23, 3.5472e-23, 6.27086e-24], [1209.75, 4.38415e-23, 1.37319e-23, 3.5593e-23, 6.27897e-24], [1212.54, 4.39417e-23, 1.37511e-23, 3.57143e-23, 6.28711e-24], [1215.33, 4.40422e-23, 1.37703e-23, 3.58359e-23, 6.29528e-24], [1218.13, 4.41429e-23, 1.37897e-23, 3.59578e-23, 6.30348e-24], [1220.94, 4.42438e-23, 1.3809e-23, 3.608e-23, 6.31171e-24], [1223.76, 4.43449e-23, 1.38284e-23, 3.62024e-23, 6.31998e-24], [1226.58, 4.44463e-23, 1.38479e-23, 3.63252e-23, 6.32829e-24], [1229.41, 4.45479e-23, 1.38675e-23, 3.64482e-23, 6.33665e-24], [1232.24, 4.46498e-23, 1.38872e-23, 3.65716e-23, 6.34509e-24], [1235.09, 4.47519e-23, 1.3907e-23, 3.66953e-23, 6.35374e-24], [1237.93, 4.48544e-23, 1.39274e-23, 3.68194e-23, 6.36356e-24], [1240.79, 4.50232e-23, 1.41591e-23, 3.70241e-23, 6.82338e-24], [1243.65, 4.50599e-23, 1.39673e-23, 3.70682e-23, 6.38097e-24], [1246.52, 4.51626e-23, 1.39863e-23, 3.71927e-23, 6.38755e-24], [1249.39, 4.52659e-23, 1.40062e-23, 3.73177e-23, 6.39574e-24], [1252.27, 4.53694e-23, 1.40262e-23, 3.74431e-23, 6.40417e-24], [1255.16, 4.54731e-23, 1.40463e-23, 3.75687e-23, 6.4127e-24], [1258.06, 4.55771e-23, 1.40665e-23, 3.76947e-23, 6.42127e-24], [1260.96, 4.56813e-23, 1.40867e-23, 3.78209e-23, 6.42989e-24], [1263.86, 4.57858e-23, 1.41071e-23, 3.79475e-23, 6.43855e-24], [1266.78, 4.58905e-23, 1.41274e-23, 3.80743e-23, 6.44724e-24], [1269.7, 4.59955e-23, 1.41479e-23, 3.82015e-23, 6.45595e-24], [1272.63, 4.61007e-23, 1.41684e-23, 3.83289e-23, 6.46471e-24], [1275.56, 4.62061e-23, 1.4189e-23, 3.84567e-23, 6.47349e-24], [1278.5, 4.63118e-23, 1.42096e-23, 3.85847e-23, 6.4823e-24], [1281.45, 4.64177e-23, 1.42303e-23, 3.8713e-23, 6.49115e-24], [1284.41, 4.65239e-23, 1.42511e-23, 3.88417e-23, 6.50002e-24], [1287.37, 4.66304e-23, 1.4272e-23, 3.89706e-23, 6.50893e-24], [1290.34, 4.6737e-23, 1.42929e-23, 3.90998e-23, 6.51786e-24], [1293.31, 4.68439e-23, 1.43139e-23, 3.92294e-23, 6.52683e-24], [1296.3, 4.69511e-23, 1.43349e-23, 3.93592e-23, 6.53583e-24], [1299.29, 4.70585e-23, 1.4356e-23, 3.94893e-23, 6.54486e-24], [1302.28, 4.71662e-23, 1.43772e-23, 3.96197e-23, 6.55392e-24], [1305.28, 4.72741e-23, 1.43984e-23, 3.97505e-23, 6.56301e-24], [1308.29, 4.73823e-23, 1.44197e-23, 3.98815e-23, 6.57213e-24], [1311.31, 4.74907e-23, 1.44411e-23, 4.00128e-23, 6.58128e-24], [1314.34, 4.75993e-23, 1.44626e-23, 4.01445e-23, 6.59047e-24], [1317.37, 4.77083e-23, 1.44841e-23, 4.02764e-23, 6.59968e-24], [1320.4, 4.78174e-23, 1.45057e-23, 4.04086e-23, 6.60893e-24], [1323.45, 4.79268e-23, 1.45273e-23, 4.05412e-23, 6.61821e-24], [1326.5, 4.80365e-23, 1.4549e-23, 4.0674e-23, 6.62752e-24], [1329.56, 4.81465e-23, 1.45708e-23, 4.08072e-23, 6.63686e-24], [1332.63, 4.82566e-23, 1.45927e-23, 4.09406e-23, 6.64623e-24], [1335.7, 4.83671e-23, 1.46146e-23, 4.10744e-23, 6.65563e-24], [1338.78, 4.84778e-23, 1.46366e-23, 4.12084e-23, 6.66507e-24], [1341.87, 4.85887e-23, 1.46587e-23, 4.13428e-23, 6.67454e-24], [1344.96, 4.86999e-23, 1.46808e-23, 4.14775e-23, 6.68404e-24], [1348.06, 4.88114e-23, 1.4703e-23, 4.16125e-23, 6.69357e-24], [1351.17, 4.89231e-23, 1.47253e-23, 4.17478e-23, 6.70313e-24], [1354.29, 4.90351e-23, 1.47476e-23, 4.18834e-23, 6.71273e-24], [1357.41, 4.91473e-23, 1.477e-23, 4.20193e-23, 6.72236e-24], [1360.54, 4.92598e-23, 1.47925e-23, 4.21555e-23, 6.73201e-24], [1363.68, 4.93726e-23, 1.48151e-23, 4.2292e-23, 6.74171e-24], [1366.82, 4.94856e-23, 1.48377e-23, 4.24289e-23, 6.75143e-24], [1369.97, 4.95989e-23, 1.48604e-23, 4.2566e-23, 6.76119e-24], [1373.13, 4.97124e-23, 1.48832e-23, 4.27035e-23, 6.77098e-24], [1376.3, 4.98262e-23, 1.4906e-23, 4.28412e-23, 6.7808e-24], [1379.47, 4.99403e-23, 1.49289e-23, 4.29793e-23, 6.79065e-24], [1382.65, 5.00546e-23, 1.49519e-23, 4.31177e-23, 6.80054e-24], [1385.84, 5.01692e-23, 1.49749e-23, 4.32564e-23, 6.81046e-24], [1389.04, 5.02841e-23, 1.4998e-23, 4.33954e-23, 6.82041e-24], [1392.24, 5.03992e-23, 1.50212e-23, 4.35347e-23, 6.8304e-24], [1395.45, 5.05146e-23, 1.50445e-23, 4.36744e-23, 6.84041e-24], [1398.67, 5.06303e-23, 1.50678e-23, 4.38143e-23, 6.85047e-24], [1401.9, 5.07462e-23, 1.50912e-23, 4.39546e-23, 6.86055e-24], [1405.13, 5.08624e-23, 1.51147e-23, 4.40952e-23, 6.87067e-24], [1408.37, 5.09789e-23, 1.51383e-23, 4.42361e-23, 6.88082e-24], [1411.62, 5.10956e-23, 1.51619e-23, 4.43773e-23, 6.891e-24], [1414.87, 5.12126e-23, 1.51856e-23, 4.45188e-23, 6.90122e-24], [1418.13, 5.13299e-23, 1.52093e-23, 4.46607e-23, 6.91147e-24], [1421.4, 5.14475e-23, 1.52332e-23, 4.48028e-23, 6.92176e-24], [1424.68, 5.15653e-23, 1.52571e-23, 4.49453e-23, 6.93207e-24], [1427.97, 5.16834e-23, 1.52811e-23, 4.50881e-23, 6.94243e-24], [1431.26, 5.18017e-23, 1.53051e-23, 4.52312e-23, 6.95281e-24], [1434.56, 5.19204e-23, 1.53293e-23, 4.53747e-23, 6.96323e-24], [1437.87, 5.20393e-23, 1.53535e-23, 4.55184e-23, 6.97368e-24], [1441.18, 5.21585e-23, 1.53778e-23, 4.56625e-23, 6.98417e-24], [1444.51, 5.2278e-23, 1.54021e-23, 4.58069e-23, 6.99469e-24], [1447.84, 5.23977e-23, 1.54266e-23, 4.59517e-23, 7.00525e-24], [1451.18, 5.25177e-23, 1.54511e-23, 4.60967e-23, 7.01584e-24], [1454.52, 5.2638e-23, 1.54756e-23, 4.62421e-23, 7.02646e-24], [1457.88, 5.27586e-23, 1.55003e-23, 4.63878e-23, 7.03712e-24], [1461.24, 5.28794e-23, 1.5525e-23, 4.65338e-23, 7.04781e-24], [1464.61, 5.30006e-23, 1.55498e-23, 4.66801e-23, 7.05854e-24], [1467.99, 5.3122e-23, 1.55747e-23, 4.68268e-23, 7.0693e-24], [1471.37, 5.32437e-23, 1.55997e-23, 4.69737e-23, 7.0801e-24], [1474.77, 5.33657e-23, 1.56247e-23, 4.71211e-23, 7.09093e-24], [1478.17, 5.34879e-23, 1.56498e-23, 4.72687e-23, 7.10179e-24], [1481.57, 5.36105e-23, 1.5675e-23, 4.74167e-23, 7.11269e-24], [1484.99, 5.37333e-23, 1.57003e-23, 4.75649e-23, 7.12363e-24], [1488.42, 5.38564e-23, 1.57256e-23, 4.77136e-23, 7.1346e-24], [1491.85, 5.39798e-23, 1.5751e-23, 4.78625e-23, 7.14561e-24], [1495.29, 5.41035e-23, 1.57765e-23, 4.80118e-23, 7.15665e-24], [1498.74, 5.42274e-23, 1.58021e-23, 4.81614e-23, 7.16772e-24], [1502.19, 5.43517e-23, 1.58277e-23, 4.83113e-23, 7.17884e-24], [1505.66, 5.44762e-23, 1.58534e-23, 4.84616e-23, 7.18998e-24], [1509.13, 5.46011e-23, 1.58792e-23, 4.86122e-23, 7.20117e-24], [1512.61, 5.47262e-23, 1.59051e-23, 4.87631e-23, 7.21238e-24], [1516.1, 5.48516e-23, 1.59311e-23, 4.89143e-23, 7.22364e-24], [1519.59, 5.49773e-23, 1.59571e-23, 4.90659e-23, 7.23493e-24], [1523.1, 5.51033e-23, 1.59832e-23, 4.92178e-23, 7.24625e-24], [1526.61, 5.52295e-23, 1.60094e-23, 4.93701e-23, 7.25761e-24], [1530.13, 5.53561e-23, 1.60357e-23, 4.95227e-23, 7.26901e-24], [1533.66, 5.5483e-23, 1.6062e-23, 4.96756e-23, 7.28044e-24], [1537.19, 5.56101e-23, 1.60884e-23, 4.98288e-23, 7.29191e-24], [1540.74, 5.57376e-23, 1.61149e-23, 4.99824e-23, 7.30342e-24], [1544.29, 5.58653e-23, 1.61415e-23, 5.01364e-23, 7.31496e-24], [1547.85, 5.59934e-23, 1.61682e-23, 5.02906e-23, 7.32654e-24], [1551.42, 5.61217e-23, 1.61949e-23, 5.04452e-23, 7.33816e-24], [1555.0, 5.62503e-23, 1.62218e-23, 5.06002e-23, 7.34982e-24], [1558.59, 5.63793e-23, 1.62487e-23, 5.07554e-23, 7.36152e-24], [1562.18, 5.65085e-23, 1.62757e-23, 5.0911e-23, 7.37327e-24], [1565.78, 5.6638e-23, 1.63027e-23, 5.1067e-23, 7.38508e-24], [1569.39, 5.67679e-23, 1.633e-23, 5.12233e-23, 7.39704e-24], [1573.01, 5.68981e-23, 1.63577e-23, 5.13801e-23, 7.4099e-24], [1576.64, 5.70343e-23, 1.6405e-23, 5.15434e-23, 7.46585e-24], [1580.28, 5.71592e-23, 1.64121e-23, 5.16943e-23, 7.43309e-24], [1583.92, 5.72902e-23, 1.64393e-23, 5.18519e-23, 7.44452e-24], [1587.57, 5.74215e-23, 1.64668e-23, 5.20099e-23, 7.45639e-24], [1591.23, 5.75531e-23, 1.64944e-23, 5.21682e-23, 7.46837e-24], [1594.9, 5.76851e-23, 1.65221e-23, 5.23269e-23, 7.4804e-24], [1598.58, 5.78173e-23, 1.65499e-23, 5.24859e-23, 7.49249e-24], [1602.27, 5.79499e-23, 1.65778e-23, 5.26453e-23, 7.50461e-24], [1605.96, 5.80827e-23, 1.66057e-23, 5.2805e-23, 7.51678e-24], [1609.67, 5.82159e-23, 1.66337e-23, 5.29651e-23, 7.52898e-24], [1613.38, 5.83493e-23, 1.66619e-23, 5.31255e-23, 7.54122e-24], [1617.1, 5.84831e-23, 1.66901e-23, 5.32862e-23, 7.5535e-24], [1620.83, 5.86172e-23, 1.67183e-23, 5.34473e-23, 7.56582e-24], [1624.56, 5.87516e-23, 1.67467e-23, 5.36088e-23, 7.57818e-24], [1628.31, 5.88863e-23, 1.67752e-23, 5.37706e-23, 7.59057e-24], [1632.07, 5.90214e-23, 1.68037e-23, 5.39327e-23, 7.60301e-24], [1635.83, 5.91567e-23, 1.68323e-23, 5.40952e-23, 7.61548e-24], [1639.6, 5.92923e-23, 1.6861e-23, 5.42581e-23, 7.62799e-24], [1643.38, 5.94283e-23, 1.68898e-23, 5.44213e-23, 7.64053e-24], [1647.17, 5.95646e-23, 1.69187e-23, 5.45849e-23, 7.65312e-24], [1650.97, 5.97012e-23, 1.69476e-23, 5.47488e-23, 7.66575e-24], [1654.78, 5.98381e-23, 1.69767e-23, 5.4913e-23, 7.67841e-24], [1658.59, 5.99753e-23, 1.70058e-23, 5.50777e-23, 7.69111e-24], [1662.42, 6.01128e-23, 1.7035e-23, 5.52426e-23, 7.70385e-24], [1666.25, 6.02507e-23, 1.70643e-23, 5.54079e-23, 7.71663e-24], [1670.09, 6.03889e-23, 1.70937e-23, 5.55736e-23, 7.72945e-24], [1673.95, 6.05274e-23, 1.71231e-23, 5.57397e-23, 7.74231e-24], [1677.81, 6.06662e-23, 1.71527e-23, 5.59061e-23, 7.75521e-24], [1681.67, 6.08053e-23, 1.71823e-23, 5.60728e-23, 7.76814e-24], [1685.55, 6.09448e-23, 1.7212e-23, 5.62399e-23, 7.78112e-24], [1689.44, 6.10846e-23, 1.72418e-23, 5.64074e-23, 7.79414e-24], [1693.34, 6.12247e-23, 1.72717e-23, 5.65752e-23, 7.80719e-24], [1697.24, 6.13651e-23, 1.73017e-23, 5.67434e-23, 7.82028e-24], [1701.15, 6.15058e-23, 1.73318e-23, 5.6912e-23, 7.83342e-24], [1705.08, 6.16469e-23, 1.73619e-23, 5.70809e-23, 7.84659e-24], [1709.01, 6.17883e-23, 1.73922e-23, 5.72501e-23, 7.8598e-24], [1712.95, 6.193e-23, 1.74225e-23, 5.74198e-23, 7.87305e-24], [1716.9, 6.20721e-23, 1.74529e-23, 5.75898e-23, 7.88635e-24], [1720.86, 6.22145e-23, 1.74834e-23, 5.77601e-23, 7.89968e-24], [1724.83, 6.23572e-23, 1.7514e-23, 5.79309e-23, 7.91305e-24], [1728.8, 6.25002e-23, 1.75447e-23, 5.8102e-23, 7.92647e-24], [1732.79, 6.26436e-23, 1.75755e-23, 5.82734e-23, 7.93992e-24], [1736.79, 6.27873e-23, 1.76063e-23, 5.84452e-23, 7.95341e-24], [1740.79, 6.29314e-23, 1.76373e-23, 5.86174e-23, 7.96694e-24], [1744.81, 6.30757e-23, 1.76683e-23, 5.879e-23, 7.98052e-24], [1748.83, 6.32204e-23, 1.76994e-23, 5.89629e-23, 7.99413e-24], [1752.86, 6.33655e-23, 1.77306e-23, 5.91362e-23, 8.00779e-24], [1756.9, 6.35109e-23, 1.77619e-23, 5.93099e-23, 8.02149e-24], [1760.96, 6.36566e-23, 1.77933e-23, 5.94839e-23, 8.03522e-24], [1765.02, 6.38026e-23, 1.78248e-23, 5.96583e-23, 8.049e-24], [1769.09, 6.3949e-23, 1.78564e-23, 5.98331e-23, 8.06282e-24], [1773.17, 6.40957e-23, 1.7888e-23, 6.00083e-23, 8.07668e-24], [1777.26, 6.42428e-23, 1.79198e-23, 6.01838e-23, 8.09058e-24], [1781.35, 6.43902e-23, 1.79516e-23, 6.03597e-23, 8.10452e-24], [1785.46, 6.45379e-23, 1.79835e-23, 6.0536e-23, 8.11851e-24], [1789.58, 6.4686e-23, 1.80156e-23, 6.07126e-23, 8.13253e-24], [1793.71, 6.48344e-23, 1.80477e-23, 6.08897e-23, 8.1466e-24], [1797.84, 6.49832e-23, 1.80799e-23, 6.10671e-23, 8.16071e-24], [1801.99, 6.51323e-23, 1.81122e-23, 6.12448e-23, 8.17486e-24], [1806.14, 6.52818e-23, 1.81446e-23, 6.1423e-23, 8.18905e-24], [1810.31, 6.54316e-23, 1.8177e-23, 6.16015e-23, 8.20328e-24], [1814.48, 6.55817e-23, 1.82096e-23, 6.17805e-23, 8.21756e-24], [1818.67, 6.57322e-23, 1.82423e-23, 6.19598e-23, 8.23187e-24], [1822.86, 6.58831e-23, 1.8275e-23, 6.21394e-23, 8.24623e-24], [1827.06, 6.60342e-23, 1.83079e-23, 6.23195e-23, 8.26064e-24], [1831.28, 6.61858e-23, 1.83408e-23, 6.25e-23, 8.27508e-24], [1835.5, 6.63377e-23, 1.83739e-23, 6.26808e-23, 8.28957e-24], [1839.73, 6.64899e-23, 1.8407e-23, 6.2862e-23, 8.3041e-24], [1843.98, 6.66425e-23, 1.84402e-23, 6.30436e-23, 8.31867e-24], [1848.23, 6.67955e-23, 1.84735e-23, 6.32256e-23, 8.33328e-24], [1852.49, 6.69488e-23, 1.85069e-23, 6.34079e-23, 8.34794e-24], [1856.76, 6.71024e-23, 1.85404e-23, 6.35907e-23, 8.36264e-24], [1861.04, 6.72564e-23, 1.8574e-23, 6.37738e-23, 8.37738e-24], [1865.34, 6.74108e-23, 1.86077e-23, 6.39574e-23, 8.39217e-24], [1869.64, 6.75655e-23, 1.86415e-23, 6.41413e-23, 8.407e-24], [1873.95, 6.77206e-23, 1.86754e-23, 6.43256e-23, 8.42187e-24], [1878.27, 6.7876e-23, 1.87094e-23, 6.45103e-23, 8.43679e-24], [1882.6, 6.80318e-23, 1.87434e-23, 6.46954e-23, 8.45176e-24], [1886.94, 6.8188e-23, 1.87776e-23, 6.48809e-23, 8.46678e-24], [1891.29, 6.83445e-23, 1.88119e-23, 6.50668e-23, 8.4819e-24], [1895.65, 6.85014e-23, 1.88466e-23, 6.52531e-23, 8.49764e-24], [1900.03, 6.86589e-23, 1.88817e-23, 6.544e-23, 8.51429e-24], [1904.41, 6.88162e-23, 1.89153e-23, 6.56268e-23, 8.52735e-24], [1908.8, 6.89742e-23, 1.89499e-23, 6.58142e-23, 8.54246e-24], [1913.2, 6.91325e-23, 1.89846e-23, 6.60021e-23, 8.5577e-24], [1917.61, 6.92912e-23, 1.90195e-23, 6.61903e-23, 8.57299e-24], [1922.03, 6.94503e-23, 1.90544e-23, 6.6379e-23, 8.58834e-24], [1926.47, 6.96097e-23, 1.90894e-23, 6.6568e-23, 8.60374e-24], [1930.91, 6.97695e-23, 1.91246e-23, 6.67575e-23, 8.61917e-24], [1935.36, 6.99296e-23, 1.91598e-23, 6.69473e-23, 8.63466e-24], [1939.83, 7.00902e-23, 1.91951e-23, 6.71376e-23, 8.65019e-24], [1944.3, 7.02511e-23, 1.92306e-23, 6.73282e-23, 8.66576e-24], [1948.78, 7.04124e-23, 1.92661e-23, 6.75193e-23, 8.68138e-24], [1953.28, 7.0574e-23, 1.93017e-23, 6.77107e-23, 8.69705e-24], [1957.78, 7.07361e-23, 1.93374e-23, 6.79026e-23, 8.71276e-24], [1962.29, 7.08985e-23, 1.93733e-23, 6.80949e-23, 8.72851e-24], [1966.82, 7.10612e-23, 1.94092e-23, 6.82875e-23, 8.74431e-24], [1971.36, 7.12244e-23, 1.94452e-23, 6.84806e-23, 8.76015e-24], [1975.9, 7.13879e-23, 1.94813e-23, 6.86741e-23, 8.77604e-24], [1980.46, 7.15518e-23, 1.95175e-23, 6.8868e-23, 8.79198e-24], [1985.02, 7.17161e-23, 1.95539e-23, 6.90623e-23, 8.80796e-24], [1989.6, 7.18808e-23, 1.95903e-23, 6.9257e-23, 8.82399e-24], [1994.19, 7.20458e-23, 1.96268e-23, 6.94522e-23, 8.84006e-24], [1998.79, 7.22112e-23, 1.96634e-23, 6.96477e-23, 8.85618e-24], [2003.4, 7.2377e-23, 1.97001e-23, 6.98437e-23, 8.87234e-24], [2008.02, 7.25432e-23, 1.9737e-23, 7.004e-23, 8.88855e-24], [2012.65, 7.27098e-23, 1.97739e-23, 7.02368e-23, 8.90481e-24], [2017.29, 7.28768e-23, 1.98109e-23, 7.0434e-23, 8.92111e-24], [2021.94, 7.30441e-23, 1.98481e-23, 7.06316e-23, 8.93745e-24], [2026.6, 7.32118e-23, 1.98853e-23, 7.08296e-23, 8.95385e-24], [2031.28, 7.338e-23, 1.99226e-23, 7.10281e-23, 8.97029e-24], [2035.96, 7.35485e-23, 1.996e-23, 7.1227e-23, 8.98678e-24], [2040.66, 7.37174e-23, 1.99976e-23, 7.14263e-23, 9.00331e-24], [2045.36, 7.38866e-23, 2.00352e-23, 7.1626e-23, 9.01989e-24], [2050.08, 7.40563e-23, 2.0073e-23, 7.18261e-23, 9.03652e-24], [2054.81, 7.42264e-23, 2.01108e-23, 7.20266e-23, 9.05319e-24], [2059.54, 7.43968e-23, 2.01488e-23, 7.22276e-23, 9.06991e-24], [2064.29, 7.45677e-23, 2.01868e-23, 7.2429e-23, 9.08668e-24], [2069.05, 7.4739e-23, 2.0225e-23, 7.26308e-23, 9.1035e-24], [2073.83, 7.49106e-23, 2.02632e-23, 7.28331e-23, 9.12036e-24], [2078.61, 7.50826e-23, 2.03016e-23, 7.30358e-23, 9.13727e-24], [2083.4, 7.52551e-23, 2.03401e-23, 7.32389e-23, 9.15423e-24], [2088.21, 7.54279e-23, 2.03787e-23, 7.34424e-23, 9.17123e-24], [2093.02, 7.56011e-23, 2.04173e-23, 7.36463e-23, 9.18829e-24], [2097.85, 7.57748e-23, 2.04561e-23, 7.38507e-23, 9.20539e-24], [2102.69, 7.59488e-23, 2.0495e-23, 7.40555e-23, 9.22254e-24], [2107.53, 7.61233e-23, 2.0534e-23, 7.42608e-23, 9.23973e-24], [2112.39, 7.62981e-23, 2.05731e-23, 7.44665e-23, 9.25698e-24], [2117.27, 7.64733e-23, 2.06123e-23, 7.46726e-23, 9.27427e-24], [2122.15, 7.6649e-23, 2.06517e-23, 7.48792e-23, 9.29161e-24], [2127.04, 7.6825e-23, 2.06911e-23, 7.50861e-23, 9.309e-24], [2131.95, 7.70015e-23, 2.07306e-23, 7.52936e-23, 9.32644e-24], [2136.86, 7.71784e-23, 2.07702e-23, 7.55014e-23, 9.34393e-24], [2141.79, 7.73556e-23, 2.081e-23, 7.57097e-23, 9.36146e-24], [2146.73, 7.75333e-23, 2.08498e-23, 7.59185e-23, 9.37905e-24], [2151.68, 7.77114e-23, 2.08898e-23, 7.61276e-23, 9.39668e-24], [2156.64, 7.78899e-23, 2.09299e-23, 7.63373e-23, 9.41436e-24], [2161.61, 7.80689e-23, 2.097e-23, 7.65473e-23, 9.43209e-24], [2166.6, 7.82482e-23, 2.10103e-23, 7.67578e-23, 9.44987e-24], [2171.6, 7.84279e-23, 2.10507e-23, 7.69688e-23, 9.4677e-24], [2176.6, 7.86081e-23, 2.10912e-23, 7.71802e-23, 9.48558e-24], [2181.62, 7.87887e-23, 2.11318e-23, 7.7392e-23, 9.50351e-24], [2186.65, 7.89697e-23, 2.11726e-23, 7.76043e-23, 9.52149e-24], [2191.7, 7.91511e-23, 2.12134e-23, 7.7817e-23, 9.53952e-24], [2196.75, 7.93329e-23, 2.12543e-23, 7.80302e-23, 9.55759e-24], [2201.82, 7.95152e-23, 2.12954e-23, 7.82438e-23, 9.57572e-24], [2206.89, 7.96978e-23, 2.13365e-23, 7.84579e-23, 9.5939e-24], [2211.98, 7.98809e-23, 2.13778e-23, 7.86724e-23, 9.61212e-24], [2217.08, 8.00645e-23, 2.14192e-23, 7.88874e-23, 9.6304e-24], [2222.2, 8.02484e-23, 2.14607e-23, 7.91028e-23, 9.64873e-24], [2227.32, 8.04328e-23, 2.15023e-23, 7.93187e-23, 9.66711e-24], [2232.46, 8.06175e-23, 2.1544e-23, 7.95351e-23, 9.68554e-24], [2237.6, 8.08028e-23, 2.15858e-23, 7.97519e-23, 9.70402e-24], [2242.76, 8.09884e-23, 2.16278e-23, 7.99691e-23, 9.72256e-24], [2247.94, 8.11745e-23, 2.16698e-23, 8.01868e-23, 9.74115e-24], [2253.12, 8.1361e-23, 2.1712e-23, 8.0405e-23, 9.75986e-24], [2258.32, 8.15482e-23, 2.17552e-23, 8.06239e-23, 9.7804e-24], [2263.52, 8.17353e-23, 2.17967e-23, 8.08427e-23, 9.79722e-24], [2268.74, 8.19231e-23, 2.18392e-23, 8.10623e-23, 9.81598e-24], [2273.97, 8.21113e-23, 2.18818e-23, 8.12823e-23, 9.83481e-24], [2279.22, 8.23e-23, 2.19245e-23, 8.15028e-23, 9.85369e-24], [2284.47, 8.24891e-23, 2.19674e-23, 8.17237e-23, 9.87263e-24], [2289.74, 8.26786e-23, 2.20103e-23, 8.19451e-23, 9.89162e-24], [2295.02, 8.28686e-23, 2.20534e-23, 8.2167e-23, 9.91067e-24], [2300.31, 8.3059e-23, 2.20966e-23, 8.23894e-23, 9.92976e-24], [2305.62, 8.32499e-23, 2.21399e-23, 8.26122e-23, 9.94891e-24], [2310.94, 8.34412e-23, 2.21833e-23, 8.28355e-23, 9.96811e-24], [2316.27, 8.36329e-23, 2.22269e-23, 8.30592e-23, 9.98736e-24], [2321.61, 8.38251e-23, 2.22705e-23, 8.32835e-23, 1.00067e-23], [2326.96, 8.40177e-23, 2.23143e-23, 8.35082e-23, 1.0026e-23], [2332.33, 8.42108e-23, 2.23581e-23, 8.37333e-23, 1.00454e-23], [2337.7, 8.44043e-23, 2.24021e-23, 8.3959e-23, 1.00649e-23], [2343.1, 8.45982e-23, 2.24463e-23, 8.41851e-23, 1.00844e-23], [2348.5, 8.47926e-23, 2.24905e-23, 8.44117e-23, 1.0104e-23], [2353.91, 8.49875e-23, 2.25348e-23, 8.46388e-23, 1.01236e-23], [2359.34, 8.51828e-23, 2.25793e-23, 8.48664e-23, 1.01433e-23], [2364.78, 8.53785e-23, 2.26239e-23, 8.50944e-23, 1.0163e-23], [2370.24, 8.55748e-23, 2.26686e-23, 8.53229e-23, 1.01828e-23], [2375.7, 8.57714e-23, 2.27134e-23, 8.55519e-23, 1.02026e-23], [2381.18, 8.59685e-23, 2.27583e-23, 8.57814e-23, 1.02225e-23], [2386.67, 8.61661e-23, 2.28034e-23, 8.60114e-23, 1.02424e-23], [2392.17, 8.63641e-23, 2.28486e-23, 8.62419e-23, 1.02624e-23], [2397.69, 8.65626e-23, 2.28939e-23, 8.64728e-23, 1.02825e-23], [2403.22, 8.67615e-23, 2.29393e-23, 8.67043e-23, 1.03026e-23], [2408.76, 8.69609e-23, 2.29848e-23, 8.69362e-23, 1.03227e-23], [2414.32, 8.71608e-23, 2.30305e-23, 8.71686e-23, 1.03429e-23], [2419.88, 8.73611e-23, 2.30762e-23, 8.74015e-23, 1.03632e-23], [2425.46, 8.75619e-23, 2.31221e-23, 8.76349e-23, 1.03835e-23], [2431.06, 8.77632e-23, 2.31682e-23, 8.78688e-23, 1.04039e-23], [2436.66, 8.79649e-23, 2.32143e-23, 8.81032e-23, 1.04243e-23], [2442.28, 8.81671e-23, 2.32606e-23, 8.8338e-23, 1.04448e-23], [2447.91, 8.83697e-23, 2.33069e-23, 8.85734e-23, 1.04654e-23], [2453.56, 8.85728e-23, 2.33534e-23, 8.88093e-23, 1.04859e-23], [2459.22, 8.87764e-23, 2.34001e-23, 8.90457e-23, 1.05066e-23], [2464.89, 8.89804e-23, 2.34468e-23, 8.92825e-23, 1.05273e-23], [2470.57, 8.9185e-23, 2.34937e-23, 8.95199e-23, 1.05481e-23], [2476.27, 8.939e-23, 2.35407e-23, 8.97578e-23, 1.05689e-23], [2481.98, 8.95954e-23, 2.35878e-23, 8.99961e-23, 1.05897e-23], [2487.7, 8.98014e-23, 2.3635e-23, 9.0235e-23, 1.06107e-23], [2493.44, 9.00078e-23, 2.36824e-23, 9.04744e-23, 1.06317e-23], [2499.19, 9.02147e-23, 2.37298e-23, 9.07143e-23, 1.06527e-23], [2504.95, 9.0422e-23, 2.37774e-23, 9.09547e-23, 1.06738e-23], [2510.73, 9.06299e-23, 2.38252e-23, 9.11956e-23, 1.06949e-23], [2516.52, 9.08382e-23, 2.3873e-23, 9.1437e-23, 1.07161e-23], [2522.32, 9.1047e-23, 2.3921e-23, 9.1679e-23, 1.07374e-23], [2528.14, 9.12563e-23, 2.39691e-23, 9.19214e-23, 1.07587e-23], [2533.97, 9.14661e-23, 2.40174e-23, 9.21643e-23, 1.07801e-23], [2539.81, 9.16763e-23, 2.40657e-23, 9.24078e-23, 1.08015e-23], [2545.67, 9.18871e-23, 2.41142e-23, 9.26518e-23, 1.0823e-23], [2551.54, 9.20983e-23, 2.41628e-23, 9.28963e-23, 1.08446e-23], [2557.42, 9.231e-23, 2.42115e-23, 9.31413e-23, 1.08662e-23], [2563.32, 9.25222e-23, 2.42604e-23, 9.33869e-23, 1.08878e-23], [2569.23, 9.27349e-23, 2.43094e-23, 9.36329e-23, 1.09095e-23], [2575.16, 9.29481e-23, 2.43585e-23, 9.38795e-23, 1.09313e-23], [2581.09, 9.31618e-23, 2.44077e-23, 9.41266e-23, 1.09532e-23], [2587.05, 9.3376e-23, 2.44571e-23, 9.43742e-23, 1.0975e-23], [2593.01, 9.35906e-23, 2.45066e-23, 9.46224e-23, 1.0997e-23], [2598.99, 9.38058e-23, 2.45562e-23, 9.4871e-23, 1.1019e-23], [2604.99, 9.40215e-23, 2.4606e-23, 9.51202e-23, 1.10411e-23], [2610.99, 9.42376e-23, 2.46559e-23, 9.537e-23, 1.10632e-23], [2617.01, 9.44543e-23, 2.47059e-23, 9.56202e-23, 1.10854e-23], [2623.05, 9.46714e-23, 2.4756e-23, 9.5871e-23, 1.11076e-23], [2629.1, 9.48891e-23, 2.48063e-23, 9.61223e-23, 1.11299e-23], [2635.16, 9.51072e-23, 2.48567e-23, 9.63742e-23, 1.11523e-23], [2641.24, 9.53259e-23, 2.49073e-23, 9.66266e-23, 1.11747e-23], [2647.33, 9.55451e-23, 2.49579e-23, 9.68795e-23, 1.11972e-23], [2653.43, 9.57647e-23, 2.50087e-23, 9.7133e-23, 1.12197e-23], [2659.55, 9.59849e-23, 2.50597e-23, 9.7387e-23, 1.12423e-23], [2665.68, 9.62056e-23, 2.51107e-23, 9.76415e-23, 1.1265e-23], [2671.83, 9.64268e-23, 2.51619e-23, 9.78966e-23, 1.12876e-23], [2677.99, 9.66485e-23, 2.52132e-23, 9.81522e-23, 1.13104e-23], [2684.17, 9.68707e-23, 2.52647e-23, 9.84084e-23, 1.13332e-23], [2690.36, 9.70935e-23, 2.53163e-23, 9.86651e-23, 1.13561e-23], [2696.56, 9.73167e-23, 2.5368e-23, 9.89224e-23, 1.13791e-23], [2702.78, 9.75405e-23, 2.54198e-23, 9.91802e-23, 1.14021e-23], [2709.01, 9.77647e-23, 2.54718e-23, 9.94385e-23, 1.14252e-23], [2715.26, 9.79895e-23, 2.55239e-23, 9.96974e-23, 1.14483e-23], [2721.52, 9.82149e-23, 2.55762e-23, 9.99569e-23, 1.14715e-23], [2727.8, 9.84407e-23, 2.56286e-23, 1.00217e-22, 1.14947e-23], [2734.09, 9.8667e-23, 2.56811e-23, 1.00477e-22, 1.15181e-23], [2740.39, 9.88939e-23, 2.57338e-23, 1.00739e-22, 1.15414e-23], [2746.71, 9.91213e-23, 2.57865e-23, 1.01e-22, 1.15649e-23], [2753.04, 9.93492e-23, 2.58395e-23, 1.01262e-22, 1.15884e-23], [2759.39, 9.95777e-23, 2.58925e-23, 1.01525e-22, 1.16119e-23], [2765.76, 9.98067e-23, 2.59457e-23, 1.01789e-22, 1.16355e-23], [2772.13, 1.00036e-22, 2.59991e-23, 1.02052e-22, 1.16592e-23], [2778.53, 1.00266e-22, 2.60525e-23, 1.02317e-22, 1.1683e-23], [2784.93, 1.00497e-22, 2.61062e-23, 1.02582e-22, 1.17068e-23], [2791.36, 1.00728e-22, 2.61599e-23, 1.02847e-22, 1.17306e-23], [2797.79, 1.0096e-22, 2.62138e-23, 1.03114e-22, 1.17545e-23], [2804.24, 1.01192e-22, 2.62678e-23, 1.0338e-22, 1.17785e-23], [2810.71, 1.01424e-22, 2.6322e-23, 1.03648e-22, 1.18026e-23], [2817.19, 1.01658e-22, 2.63763e-23, 1.03915e-22, 1.18267e-23], [2823.69, 1.01891e-22, 2.64307e-23, 1.04184e-22, 1.18509e-23], [2830.2, 1.02126e-22, 2.64853e-23, 1.04453e-22, 1.18751e-23], [2836.73, 1.02361e-22, 2.654e-23, 1.04723e-22, 1.18994e-23], [2843.27, 1.02596e-22, 2.65949e-23, 1.04993e-22, 1.19238e-23], [2849.83, 1.02832e-22, 2.66499e-23, 1.05263e-22, 1.19482e-23], [2856.4, 1.03068e-22, 2.6705e-23, 1.05535e-22, 1.19727e-23], [2862.98, 1.03305e-22, 2.67603e-23, 1.05807e-22, 1.19973e-23], [2869.59, 1.03543e-22, 2.68157e-23, 1.06079e-22, 1.20219e-23], [2876.2, 1.03781e-22, 2.68712e-23, 1.06352e-22, 1.20466e-23], [2882.84, 1.0402e-22, 2.69269e-23, 1.06626e-22, 1.20714e-23], [2889.48, 1.04259e-22, 2.69828e-23, 1.069e-22, 1.20962e-23], [2896.15, 1.04499e-22, 2.70388e-23, 1.07175e-22, 1.21211e-23], [2902.83, 1.04739e-22, 2.70949e-23, 1.0745e-22, 1.2146e-23], [2909.52, 1.0498e-22, 2.71512e-23, 1.07726e-22, 1.2171e-23], [2916.23, 1.05222e-22, 2.72076e-23, 1.08003e-22, 1.21961e-23], [2922.95, 1.05464e-22, 2.72641e-23, 1.0828e-22, 1.22212e-23], [2929.69, 1.05706e-22, 2.73208e-23, 1.08558e-22, 1.22464e-23], [2936.45, 1.05949e-22, 2.73777e-23, 1.08836e-22, 1.22717e-23], [2943.22, 1.06193e-22, 2.74347e-23, 1.09115e-22, 1.2297e-23], [2950.01, 1.06437e-22, 2.74918e-23, 1.09394e-22, 1.23224e-23], [2956.81, 1.06682e-22, 2.75491e-23, 1.09674e-22, 1.23479e-23], [2963.63, 1.06927e-22, 2.76065e-23, 1.09955e-22, 1.23734e-23], [2970.46, 1.07173e-22, 2.76641e-23, 1.10237e-22, 1.2399e-23], [2977.31, 1.0742e-22, 2.77218e-23, 1.10518e-22, 1.24246e-23], [2984.18, 1.07667e-22, 2.77796e-23, 1.10801e-22, 1.24504e-23], [2991.06, 1.07914e-22, 2.78376e-23, 1.11084e-22, 1.24762e-23], [2997.96, 1.08163e-22, 2.78958e-23, 1.11368e-22, 1.2502e-23], [3004.87, 1.08411e-22, 2.79541e-23, 1.11652e-22, 1.2528e-23], [3011.8, 1.08661e-22, 2.80126e-23, 1.11937e-22, 1.2554e-23], [3018.75, 1.08911e-22, 2.80712e-23, 1.12223e-22, 1.25801e-23], [3025.71, 1.09161e-22, 2.81299e-23, 1.12509e-22, 1.26062e-23], [3032.68, 1.09412e-22, 2.81888e-23, 1.12796e-22, 1.26325e-23], [3039.68, 1.09664e-22, 2.82479e-23, 1.13083e-22, 1.26587e-23], [3046.69, 1.09916e-22, 2.83071e-23, 1.13371e-22, 1.26851e-23], [3053.71, 1.10169e-22, 2.83664e-23, 1.1366e-22, 1.27115e-23], [3060.76, 1.10423e-22, 2.84259e-23, 1.13949e-22, 1.2738e-23], [3067.81, 1.10677e-22, 2.84856e-23, 1.14239e-22, 1.27645e-23], [3074.89, 1.10931e-22, 2.85454e-23, 1.14529e-22, 1.27911e-23], [3081.98, 1.11186e-22, 2.86053e-23, 1.14821e-22, 1.28178e-23], [3089.09, 1.11442e-22, 2.86654e-23, 1.15112e-22, 1.28446e-23], [3096.21, 1.11699e-22, 2.87258e-23, 1.15405e-22, 1.28717e-23], [3103.35, 1.11955e-22, 2.8786e-23, 1.15698e-22, 1.28983e-23], [3110.51, 1.12213e-22, 2.88466e-23, 1.15991e-22, 1.29253e-23], [3117.68, 1.12471e-22, 2.89073e-23, 1.16286e-22, 1.29523e-23], [3124.87, 1.1273e-22, 2.89682e-23, 1.16581e-22, 1.29794e-23], [3132.07, 1.12989e-22, 2.90292e-23, 1.16876e-22, 1.30065e-23], [3139.3, 1.13249e-22, 2.90903e-23, 1.17172e-22, 1.30338e-23], [3146.54, 1.1351e-22, 2.91516e-23, 1.17469e-22, 1.30611e-23], [3153.79, 1.13771e-22, 2.92131e-23, 1.17767e-22, 1.30884e-23], [3161.06, 1.14032e-22, 2.92747e-23, 1.18065e-22, 1.31159e-23], [3168.35, 1.14295e-22, 2.93365e-23, 1.18363e-22, 1.31434e-23], [3175.66, 1.14558e-22, 2.93984e-23, 1.18663e-22, 1.3171e-23], [3182.98, 1.14821e-22, 2.94605e-23, 1.18963e-22, 1.31986e-23], [3190.32, 1.15085e-22, 2.95228e-23, 1.19263e-22, 1.32263e-23], [3197.68, 1.1535e-22, 2.95852e-23, 1.19565e-22, 1.32541e-23], [3205.05, 1.15616e-22, 2.96477e-23, 1.19867e-22, 1.3282e-23], [3212.44, 1.15882e-22, 2.97104e-23, 1.20169e-22, 1.33099e-23], [3219.85, 1.16148e-22, 2.97733e-23, 1.20473e-22, 1.33379e-23], [3227.28, 1.16415e-22, 2.98363e-23, 1.20776e-22, 1.3366e-23], [3234.72, 1.16683e-22, 2.98995e-23, 1.21081e-22, 1.33941e-23], [3242.18, 1.16952e-22, 2.99629e-23, 1.21386e-22, 1.34223e-23], [3249.66, 1.17221e-22, 3.00263e-23, 1.21692e-22, 1.34506e-23], [3257.15, 1.1749e-22, 3.009e-23, 1.21999e-22, 1.3479e-23], [3264.66, 1.17761e-22, 3.01538e-23, 1.22306e-22, 1.35074e-23], [3272.19, 1.18032e-22, 3.02178e-23, 1.22614e-22, 1.35359e-23], [3279.73, 1.18303e-22, 3.02819e-23, 1.22922e-22, 1.35645e-23], [3287.3, 1.18575e-22, 3.03462e-23, 1.23231e-22, 1.35931e-23], [3294.88, 1.18848e-22, 3.04107e-23, 1.23541e-22, 1.36218e-23], [3302.48, 1.19122e-22, 3.04753e-23, 1.23852e-22, 1.36506e-23], [3310.09, 1.19396e-22, 3.05401e-23, 1.24163e-22, 1.36795e-23], [3317.72, 1.1967e-22, 3.0605e-23, 1.24475e-22, 1.37084e-23], [3325.37, 1.19946e-22, 3.06701e-23, 1.24787e-22, 1.37374e-23], [3333.04, 1.20222e-22, 3.07354e-23, 1.25101e-22, 1.37665e-23], [3340.73, 1.20498e-22, 3.08008e-23, 1.25414e-22, 1.37956e-23], [3348.43, 1.20776e-22, 3.08664e-23, 1.25729e-22, 1.38248e-23], [3356.15, 1.21054e-22, 3.09321e-23, 1.26044e-22, 1.38541e-23], [3363.89, 1.21332e-22, 3.0998e-23, 1.2636e-22, 1.38835e-23], [3371.65, 1.21611e-22, 3.10641e-23, 1.26677e-22, 1.39129e-23], [3379.43, 1.21891e-22, 3.11303e-23, 1.26994e-22, 1.39425e-23], [3387.22, 1.22172e-22, 3.11967e-23, 1.27312e-22, 1.3972e-23], [3395.03, 1.22453e-22, 3.12633e-23, 1.27631e-22, 1.40017e-23], [3402.86, 1.22734e-22, 3.133e-23, 1.2795e-22, 1.40314e-23], [3410.71, 1.23017e-22, 3.13969e-23, 1.2827e-22, 1.40613e-23], [3418.57, 1.233e-22, 3.1464e-23, 1.28591e-22, 1.40911e-23], [3426.45, 1.23584e-22, 3.15312e-23, 1.28912e-22, 1.41211e-23], [3434.36, 1.23868e-22, 3.15986e-23, 1.29234e-22, 1.41511e-23], [3442.28, 1.24153e-22, 3.16662e-23, 1.29557e-22, 1.41812e-23], [3450.21, 1.24439e-22, 3.17339e-23, 1.29881e-22, 1.42114e-23], [3458.17, 1.24725e-22, 3.18018e-23, 1.30205e-22, 1.42417e-23], [3466.14, 1.25012e-22, 3.18699e-23, 1.3053e-22, 1.4272e-23], [3474.14, 1.253e-22, 3.19381e-23, 1.30856e-22, 1.43024e-23], [3482.15, 1.25588e-22, 3.20065e-23, 1.31182e-22, 1.43329e-23], [3490.18, 1.25877e-22, 3.20751e-23, 1.31509e-22, 1.43635e-23], [3498.23, 1.26167e-22, 3.21438e-23, 1.31837e-22, 1.43941e-23], [3506.29, 1.26457e-22, 3.22128e-23, 1.32165e-22, 1.44248e-23], [3514.38, 1.26748e-22, 3.22818e-23, 1.32494e-22, 1.44556e-23], [3522.48, 1.2704e-22, 3.23511e-23, 1.32824e-22, 1.44865e-23], [3530.61, 1.27332e-22, 3.24205e-23, 1.33155e-22, 1.45175e-23], [3538.75, 1.27625e-22, 3.24901e-23, 1.33486e-22, 1.45485e-23], [3546.91, 1.27919e-22, 3.25599e-23, 1.33818e-22, 1.45796e-23], [3555.09, 1.28213e-22, 3.26298e-23, 1.34151e-22, 1.46108e-23], [3563.29, 1.28508e-22, 3.26999e-23, 1.34485e-22, 1.4642e-23], [3571.5, 1.28804e-22, 3.27702e-23, 1.34819e-22, 1.46733e-23], [3579.74, 1.291e-22, 3.28407e-23, 1.35154e-22, 1.47047e-23], [3587.99, 1.29397e-22, 3.29113e-23, 1.35489e-22, 1.47362e-23], [3596.27, 1.29695e-22, 3.29821e-23, 1.35826e-22, 1.47678e-23], [3604.56, 1.29994e-22, 3.30531e-23, 1.36163e-22, 1.47994e-23], [3612.87, 1.30293e-22, 3.31242e-23, 1.36501e-22, 1.48312e-23], [3621.2, 1.30593e-22, 3.31956e-23, 1.36839e-22, 1.4863e-23], [3629.55, 1.30893e-22, 3.32671e-23, 1.37179e-22, 1.48949e-23], [3637.92, 1.31194e-22, 3.33387e-23, 1.37519e-22, 1.49268e-23], [3646.31, 1.31496e-22, 3.34106e-23, 1.3786e-22, 1.49588e-23], [3654.72, 1.31799e-22, 3.34826e-23, 1.38201e-22, 1.4991e-23], [3663.15, 1.32102e-22, 3.35548e-23, 1.38543e-22, 1.50232e-23], [3671.6, 1.32406e-22, 3.36272e-23, 1.38886e-22, 1.50554e-23], [3680.06, 1.32711e-22, 3.36998e-23, 1.3923e-22, 1.50878e-23], [3688.55, 1.33016e-22, 3.37725e-23, 1.39575e-22, 1.51202e-23], [3697.06, 1.33323e-22, 3.38455e-23, 1.3992e-22, 1.51528e-23], [3705.58, 1.33629e-22, 3.39186e-23, 1.40266e-22, 1.51854e-23], [3714.13, 1.33937e-22, 3.39919e-23, 1.40613e-22, 1.5218e-23], [3722.69, 1.34245e-22, 3.40653e-23, 1.4096e-22, 1.52508e-23], [3731.28, 1.34554e-22, 3.4139e-23, 1.41309e-22, 1.52836e-23], [3739.88, 1.34864e-22, 3.42128e-23, 1.41658e-22, 1.53166e-23], [3748.5, 1.35174e-22, 3.42868e-23, 1.42008e-22, 1.53496e-23], [3757.15, 1.35485e-22, 3.4361e-23, 1.42358e-22, 1.53827e-23], [3765.81, 1.35797e-22, 3.44354e-23, 1.4271e-22, 1.54159e-23], [3774.5, 1.3611e-22, 3.45099e-23, 1.43062e-22, 1.54492e-23], [3783.2, 1.36423e-22, 3.45847e-23, 1.43415e-22, 1.54826e-23], [3791.92, 1.36737e-22, 3.46596e-23, 1.43768e-22, 1.5516e-23], [3800.67, 1.37052e-22, 3.47347e-23, 1.44123e-22, 1.55495e-23], [3809.43, 1.37367e-22, 3.481e-23, 1.44478e-22, 1.55832e-23], [3818.22, 1.37683e-22, 3.48855e-23, 1.44834e-22, 1.56169e-23], [3827.02, 1.38e-22, 3.49612e-23, 1.45191e-22, 1.56506e-23], [3835.85, 1.38318e-22, 3.5037e-23, 1.45548e-22, 1.56845e-23], [3844.69, 1.38636e-22, 3.51131e-23, 1.45906e-22, 1.57184e-23], [3853.56, 1.38955e-22, 3.51893e-23, 1.46266e-22, 1.57525e-23], [3862.44, 1.39275e-22, 3.52657e-23, 1.46625e-22, 1.57866e-23], [3871.35, 1.39596e-22, 3.53423e-23, 1.46986e-22, 1.58208e-23], [3880.28, 1.39917e-22, 3.54191e-23, 1.47348e-22, 1.5855e-23], [3889.23, 1.40239e-22, 3.5496e-23, 1.4771e-22, 1.58894e-23], [3898.2, 1.40562e-22, 3.55732e-23, 1.48073e-22, 1.59238e-23], [3907.18, 1.40885e-22, 3.56505e-23, 1.48437e-22, 1.59584e-23], [3916.19, 1.4121e-22, 3.57281e-23, 1.48801e-22, 1.5993e-23], [3925.23, 1.41535e-22, 3.58058e-23, 1.49167e-22, 1.60277e-23], [3934.28, 1.4186e-22, 3.58837e-23, 1.49533e-22, 1.60625e-23], [3943.35, 1.42187e-22, 3.59618e-23, 1.499e-22, 1.60973e-23], [3952.44, 1.42514e-22, 3.60401e-23, 1.50268e-22, 1.61323e-23], [3961.56, 1.42842e-22, 3.61186e-23, 1.50636e-22, 1.61673e-23], [3970.69, 1.43171e-22, 3.61972e-23, 1.51006e-22, 1.62025e-23], [3979.85, 1.43501e-22, 3.62765e-23, 1.51376e-22, 1.62385e-23], [3989.03, 1.43831e-22, 3.63552e-23, 1.51747e-22, 1.6273e-23], [3998.23, 1.44162e-22, 3.64344e-23, 1.52119e-22, 1.63083e-23], [4007.45, 1.44494e-22, 3.65139e-23, 1.52492e-22, 1.63438e-23], [4016.69, 1.44827e-22, 3.65935e-23, 1.52865e-22, 1.63794e-23], [4025.95, 1.4516e-22, 3.66733e-23, 1.5324e-22, 1.6415e-23], [4035.23, 1.45494e-22, 3.67533e-23, 1.53615e-22, 1.64507e-23], [4044.54, 1.45829e-22, 3.68336e-23, 1.53991e-22, 1.64866e-23], [4053.87, 1.46165e-22, 3.6914e-23, 1.54368e-22, 1.65225e-23], [4063.21, 1.46501e-22, 3.69946e-23, 1.54745e-22, 1.65585e-23], [4072.58, 1.46838e-22, 3.70754e-23, 1.55124e-22, 1.65945e-23], [4081.97, 1.47176e-22, 3.71564e-23, 1.55503e-22, 1.66307e-23], [4091.39, 1.47515e-22, 3.72376e-23, 1.55883e-22, 1.6667e-23], [4100.82, 1.47855e-22, 3.7319e-23, 1.56264e-22, 1.67033e-23], [4110.28, 1.48195e-22, 3.74006e-23, 1.56646e-22, 1.67397e-23], [4119.76, 1.48536e-22, 3.74823e-23, 1.57028e-22, 1.67763e-23], [4129.26, 1.48878e-22, 3.75643e-23, 1.57412e-22, 1.68129e-23], [4138.78, 1.49221e-22, 3.76465e-23, 1.57796e-22, 1.68496e-23], [4148.32, 1.49564e-22, 3.77289e-23, 1.58181e-22, 1.68864e-23], [4157.89, 1.49909e-22, 3.78115e-23, 1.58567e-22, 1.69232e-23], [4167.48, 1.50254e-22, 3.78943e-23, 1.58954e-22, 1.69602e-23], [4177.09, 1.506e-22, 3.79773e-23, 1.59342e-22, 1.69973e-23], [4186.72, 1.50946e-22, 3.80604e-23, 1.5973e-22, 1.70344e-23], [4196.38, 1.51294e-22, 3.81438e-23, 1.6012e-22, 1.70717e-23], [4206.05, 1.51642e-22, 3.82274e-23, 1.6051e-22, 1.7109e-23], [4215.75, 1.51991e-22, 3.83112e-23, 1.60901e-22, 1.71464e-23], [4225.47, 1.52341e-22, 3.83952e-23, 1.61293e-22, 1.71839e-23], [4235.22, 1.52692e-22, 3.84794e-23, 1.61686e-22, 1.72215e-23], [4244.98, 1.53043e-22, 3.85638e-23, 1.6208e-22, 1.72592e-23], [4254.77, 1.53396e-22, 3.86484e-23, 1.62474e-22, 1.7297e-23], [4264.58, 1.53749e-22, 3.87332e-23, 1.6287e-22, 1.73349e-23], [4274.42, 1.54103e-22, 3.88183e-23, 1.63266e-22, 1.73729e-23], [4284.27, 1.54458e-22, 3.89035e-23, 1.63663e-22, 1.74109e-23], [4294.15, 1.54813e-22, 3.89889e-23, 1.64061e-22, 1.74491e-23], [4304.06, 1.5517e-22, 3.90745e-23, 1.6446e-22, 1.74873e-23], [4313.98, 1.55527e-22, 3.91604e-23, 1.6486e-22, 1.75257e-23], [4323.93, 1.55885e-22, 3.92464e-23, 1.65261e-22, 1.75641e-23], [4333.9, 1.56244e-22, 3.93327e-23, 1.65662e-22, 1.76026e-23], [4343.89, 1.56604e-22, 3.94191e-23, 1.66065e-22, 1.76412e-23], [4353.91, 1.56964e-22, 3.95058e-23, 1.66468e-22, 1.76799e-23], [4363.95, 1.57325e-22, 3.95927e-23, 1.66872e-22, 1.77187e-23], [4374.02, 1.57688e-22, 3.96798e-23, 1.67277e-22, 1.77576e-23], [4384.1, 1.58051e-22, 3.97671e-23, 1.67683e-22, 1.77966e-23], [4394.21, 1.58415e-22, 3.98546e-23, 1.6809e-22, 1.78357e-23], [4404.34, 1.58779e-22, 3.99423e-23, 1.68498e-22, 1.78749e-23], [4414.5, 1.59145e-22, 4.00303e-23, 1.68907e-22, 1.79142e-23], [4424.68, 1.59511e-22, 4.01184e-23, 1.69316e-22, 1.79536e-23], [4434.88, 1.59878e-22, 4.02068e-23, 1.69727e-22, 1.7993e-23], [4445.11, 1.60247e-22, 4.02954e-23, 1.70138e-22, 1.80326e-23], [4455.36, 1.60616e-22, 4.03841e-23, 1.70551e-22, 1.80723e-23], [4465.64, 1.60985e-22, 4.04731e-23, 1.70964e-22, 1.8112e-23], [4475.93, 1.61356e-22, 4.05624e-23, 1.71378e-22, 1.81519e-23], [4486.26, 1.61727e-22, 4.06518e-23, 1.71793e-22, 1.81918e-23], [4496.6, 1.621e-22, 4.07414e-23, 1.72209e-22, 1.82318e-23], [4506.97, 1.62473e-22, 4.08313e-23, 1.72626e-22, 1.8272e-23], [4517.36, 1.62847e-22, 4.09214e-23, 1.73044e-22, 1.83122e-23], [4527.78, 1.63222e-22, 4.10117e-23, 1.73463e-22, 1.83526e-23], [4538.22, 1.63598e-22, 4.11022e-23, 1.73882e-22, 1.8393e-23], [4548.69, 1.63975e-22, 4.11929e-23, 1.74303e-22, 1.84335e-23], [4559.18, 1.64352e-22, 4.12839e-23, 1.74725e-22, 1.84741e-23], [4569.69, 1.6473e-22, 4.1375e-23, 1.75147e-22, 1.85149e-23], [4580.23, 1.6511e-22, 4.14664e-23, 1.7557e-22, 1.85557e-23], [4590.79, 1.6549e-22, 4.1558e-23, 1.75995e-22, 1.85966e-23], [4601.38, 1.65871e-22, 4.16498e-23, 1.7642e-22, 1.86376e-23], [4611.99, 1.66253e-22, 4.17419e-23, 1.76846e-22, 1.86788e-23], [4622.62, 1.66636e-22, 4.18342e-23, 1.77273e-22, 1.872e-23], [4633.28, 1.67019e-22, 4.19266e-23, 1.77702e-22, 1.87613e-23], [4643.97, 1.67404e-22, 4.20194e-23, 1.78131e-22, 1.88027e-23], [4654.67, 1.67789e-22, 4.21123e-23, 1.78561e-22, 1.88442e-23], [4665.41, 1.68176e-22, 4.22054e-23, 1.78992e-22, 1.88858e-23], [4676.17, 1.68563e-22, 4.22988e-23, 1.79424e-22, 1.89276e-23], [4686.95, 1.68951e-22, 4.23924e-23, 1.79856e-22, 1.89694e-23], [4697.76, 1.6934e-22, 4.24863e-23, 1.8029e-22, 1.90113e-23], [4708.59, 1.6973e-22, 4.25803e-23, 1.80725e-22, 1.90533e-23], [4719.45, 1.70121e-22, 4.26746e-23, 1.81161e-22, 1.90954e-23], [4730.33, 1.70513e-22, 4.27691e-23, 1.81598e-22, 1.91377e-23], [4741.24, 1.70905e-22, 4.28638e-23, 1.82035e-22, 1.918e-23], [4752.17, 1.71299e-22, 4.29588e-23, 1.82474e-22, 1.92224e-23], [4763.13, 1.71693e-22, 4.3054e-23, 1.82914e-22, 1.92649e-23], [4774.12, 1.72088e-22, 4.31494e-23, 1.83354e-22, 1.93076e-23], [4785.13, 1.72485e-22, 4.32451e-23, 1.83796e-22, 1.93503e-23], [4796.16, 1.72882e-22, 4.33409e-23, 1.84238e-22, 1.93931e-23], [4807.22, 1.7328e-22, 4.3437e-23, 1.84682e-22, 1.94361e-23], [4818.31, 1.73679e-22, 4.35333e-23, 1.85126e-22, 1.94791e-23], [4829.42, 1.74079e-22, 4.36299e-23, 1.85572e-22, 1.95222e-23], [4840.55, 1.7448e-22, 4.37267e-23, 1.86018e-22, 1.95655e-23], [4851.72, 1.74881e-22, 4.38237e-23, 1.86466e-22, 1.96088e-23], [4862.9, 1.75284e-22, 4.3921e-23, 1.86914e-22, 1.96523e-23], [4874.12, 1.75688e-22, 4.40185e-23, 1.87364e-22, 1.96959e-23], [4885.36, 1.76092e-22, 4.41162e-23, 1.87814e-22, 1.97395e-23], [4896.62, 1.76498e-22, 4.42142e-23, 1.88266e-22, 1.97833e-23], [4907.91, 1.76904e-22, 4.43124e-23, 1.88718e-22, 1.98272e-23], [4919.23, 1.77312e-22, 4.44108e-23, 1.89172e-22, 1.98711e-23], [4930.58, 1.7772e-22, 4.45095e-23, 1.89626e-22, 1.99152e-23], [4941.95, 1.78129e-22, 4.46084e-23, 1.90082e-22, 1.99594e-23], [4953.34, 1.78539e-22, 4.47075e-23, 1.90539e-22, 2.00037e-23], [4964.76, 1.78951e-22, 4.48069e-23, 1.90996e-22, 2.00481e-23], [4976.21, 1.79363e-22, 4.49065e-23, 1.91455e-22, 2.00926e-23], [4987.69, 1.79776e-22, 4.50063e-23, 1.91914e-22, 2.01372e-23], [4999.19, 1.8019e-22, 4.51064e-23, 1.92375e-22, 2.01819e-23], [5010.72, 1.80605e-22, 4.52067e-23, 1.92836e-22, 2.02268e-23], [5022.27, 1.8102e-22, 4.53073e-23, 1.93299e-22, 2.02717e-23], [5033.85, 1.81437e-22, 4.54081e-23, 1.93763e-22, 2.03168e-23], [5045.46, 1.81855e-22, 4.55091e-23, 1.94228e-22, 2.03619e-23], [5057.1, 1.82274e-22, 4.56104e-23, 1.94693e-22, 2.04072e-23], [5068.76, 1.82694e-22, 4.5712e-23, 1.9516e-22, 2.04525e-23], [5080.45, 1.83114e-22, 4.58137e-23, 1.95628e-22, 2.0498e-23], [5092.16, 1.83536e-22, 4.59157e-23, 1.96097e-22, 2.05436e-23], [5103.91, 1.83959e-22, 4.6018e-23, 1.96567e-22, 2.05893e-23], [5115.68, 1.84382e-22, 4.61205e-23, 1.97038e-22, 2.06351e-23], [5127.47, 1.84807e-22, 4.62233e-23, 1.9751e-22, 2.0681e-23], [5139.3, 1.85233e-22, 4.63262e-23, 1.97983e-22, 2.0727e-23], [5151.15, 1.85659e-22, 4.64295e-23, 1.98457e-22, 2.07731e-23], [5163.03, 1.86087e-22, 4.6533e-23, 1.98932e-22, 2.08194e-23], [5174.93, 1.86515e-22, 4.66367e-23, 1.99408e-22, 2.08657e-23], [5186.87, 1.86945e-22, 4.67407e-23, 1.99886e-22, 2.09122e-23], [5198.83, 1.87375e-22, 4.68449e-23, 2.00364e-22, 2.09588e-23], [5210.81, 1.87807e-22, 4.69494e-23, 2.00843e-22, 2.10055e-23], [5222.83, 1.88239e-22, 4.70541e-23, 2.01324e-22, 2.10523e-23], [5234.87, 1.88673e-22, 4.71591e-23, 2.01806e-22, 2.10992e-23], [5246.95, 1.89107e-22, 4.72643e-23, 2.02288e-22, 2.11462e-23], [5259.05, 1.89543e-22, 4.73698e-23, 2.02772e-22, 2.11933e-23], [5271.17, 1.89979e-22, 4.74755e-23, 2.03257e-22, 2.12406e-23], [5283.33, 1.90417e-22, 4.75814e-23, 2.03743e-22, 2.12879e-23], [5295.51, 1.90855e-22, 4.76877e-23, 2.0423e-22, 2.13354e-23], [5307.72, 1.91295e-22, 4.77942e-23, 2.04718e-22, 2.1383e-23], [5319.96, 1.91735e-22, 4.79009e-23, 2.05207e-22, 2.14307e-23], [5332.23, 1.92177e-22, 4.80079e-23, 2.05697e-22, 2.14785e-23], [5344.53, 1.9262e-22, 4.81151e-23, 2.06188e-22, 2.15264e-23], [5356.85, 1.93063e-22, 4.82226e-23, 2.06681e-22, 2.15744e-23], [5369.2, 1.93508e-22, 4.83304e-23, 2.07174e-22, 2.16226e-23], [5381.59, 1.93954e-22, 4.84384e-23, 2.07669e-22, 2.16709e-23], [5394.0, 1.944e-22, 4.85466e-23, 2.08165e-22, 2.17193e-23], [5406.43, 1.94848e-22, 4.86552e-23, 2.08662e-22, 2.17677e-23], [5418.9, 1.95297e-22, 4.87639e-23, 2.09159e-22, 2.18164e-23], [5431.4, 1.95746e-22, 4.8873e-23, 2.09659e-22, 2.18651e-23], [5443.92, 1.96197e-22, 4.89823e-23, 2.10159e-22, 2.19139e-23], [5456.48, 1.96649e-22, 4.90918e-23, 2.1066e-22, 2.19629e-23], [5469.06, 1.97102e-22, 4.92017e-23, 2.11162e-22, 2.2012e-23], [5481.67, 1.97556e-22, 4.93117e-23, 2.11666e-22, 2.20612e-23], [5494.31, 1.98011e-22, 4.94221e-23, 2.12171e-22, 2.21105e-23], [5506.98, 1.98467e-22, 4.95327e-23, 2.12677e-22, 2.21599e-23], [5519.68, 1.98924e-22, 4.96435e-23, 2.13183e-22, 2.22095e-23], [5532.41, 1.99382e-22, 4.97547e-23, 2.13692e-22, 2.22591e-23], [5545.17, 1.99841e-22, 4.98661e-23, 2.14201e-22, 2.23089e-23], [5557.95, 2.00302e-22, 4.99777e-23, 2.14711e-22, 2.23588e-23], [5570.77, 2.00763e-22, 5.00896e-23, 2.15223e-22, 2.24088e-23], [5583.62, 2.01225e-22, 5.02018e-23, 2.15735e-22, 2.2459e-23], [5596.49, 2.01689e-22, 5.03143e-23, 2.16249e-22, 2.25092e-23], [5609.4, 2.02153e-22, 5.0427e-23, 2.16764e-22, 2.25596e-23], [5622.33, 2.02619e-22, 5.054e-23, 2.1728e-22, 2.26101e-23], [5635.3, 2.03086e-22, 5.06532e-23, 2.17797e-22, 2.26607e-23], [5648.29, 2.03553e-22, 5.07668e-23, 2.18316e-22, 2.27115e-23], [5661.32, 2.04022e-22, 5.08805e-23, 2.18835e-22, 2.27623e-23], [5674.37, 2.04492e-22, 5.09946e-23, 2.19356e-22, 2.28133e-23], [5687.46, 2.04963e-22, 5.11089e-23, 2.19878e-22, 2.28644e-23], [5700.57, 2.05435e-22, 5.12235e-23, 2.20401e-22, 2.29156e-23], [5713.72, 2.05908e-22, 5.13384e-23, 2.20925e-22, 2.2967e-23], [5726.9, 2.06383e-22, 5.14536e-23, 2.2145e-22, 2.30184e-23], [5740.1, 2.06858e-22, 5.1569e-23, 2.21977e-22, 2.307e-23], [5753.34, 2.07334e-22, 5.16847e-23, 2.22505e-22, 2.31217e-23], [5766.61, 2.07812e-22, 5.18007e-23, 2.23034e-22, 2.31736e-23], [5779.9, 2.08291e-22, 5.19169e-23, 2.23564e-22, 2.32255e-23], [5793.23, 2.0877e-22, 5.20334e-23, 2.24095e-22, 2.32776e-23], [5806.59, 2.09251e-22, 5.21502e-23, 2.24628e-22, 2.33298e-23], [5819.98, 2.09733e-22, 5.22673e-23, 2.25161e-22, 2.33821e-23], [5833.4, 2.10216e-22, 5.23846e-23, 2.25696e-22, 2.34346e-23], [5846.85, 2.107e-22, 5.25023e-23, 2.26232e-22, 2.34871e-23], [5860.34, 2.11186e-22, 5.26202e-23, 2.2677e-22, 2.35398e-23], [5873.85, 2.11672e-22, 5.27383e-23, 2.27308e-22, 2.35927e-23], [5887.4, 2.1216e-22, 5.28568e-23, 2.27848e-22, 2.36456e-23], [5900.97, 2.12648e-22, 5.29755e-23, 2.28389e-22, 2.36987e-23], [5914.58, 2.13138e-22, 5.30946e-23, 2.28931e-22, 2.37519e-23], [5928.22, 2.13629e-22, 5.32139e-23, 2.29475e-22, 2.38052e-23], [5941.89, 2.14121e-22, 5.33335e-23, 2.30019e-22, 2.38587e-23], [5955.59, 2.14614e-22, 5.34533e-23, 2.30565e-22, 2.39122e-23], [5969.33, 2.15109e-22, 5.35735e-23, 2.31112e-22, 2.39659e-23], [5983.09, 2.15604e-22, 5.36939e-23, 2.3166e-22, 2.40198e-23], [5996.89, 2.16101e-22, 5.38146e-23, 2.3221e-22, 2.40737e-23], [6010.72, 2.16599e-22, 5.39356e-23, 2.3276e-22, 2.41278e-23], [6024.58, 2.17097e-22, 5.40569e-23, 2.33312e-22, 2.4182e-23], [6038.47, 2.17598e-22, 5.41785e-23, 2.33866e-22, 2.42364e-23], [6052.4, 2.18099e-22, 5.43004e-23, 2.3442e-22, 2.42909e-23], [6066.35, 2.18601e-22, 5.44225e-23, 2.34976e-22, 2.43455e-23], [6080.34, 2.19105e-22, 5.4545e-23, 2.35533e-22, 2.44002e-23], [6094.36, 2.19609e-22, 5.46677e-23, 2.36091e-22, 2.4455e-23], [6108.42, 2.20115e-22, 5.47907e-23, 2.3665e-22, 2.451e-23], [6122.5, 2.20622e-22, 5.49141e-23, 2.37211e-22, 2.45652e-23], [6136.62, 2.2113e-22, 5.50377e-23, 2.37773e-22, 2.46204e-23], [6150.77, 2.2164e-22, 5.51616e-23, 2.38336e-22, 2.46758e-23], [6164.96, 2.2215e-22, 5.52858e-23, 2.38901e-22, 2.47313e-23], [6179.17, 2.22662e-22, 5.54102e-23, 2.39467e-22, 2.47869e-23], [6193.42, 2.23175e-22, 5.5535e-23, 2.40034e-22, 2.48427e-23], [6207.7, 2.23689e-22, 5.56601e-23, 2.40602e-22, 2.48986e-23], [6222.02, 2.24204e-22, 5.57854e-23, 2.41172e-22, 2.49547e-23], [6236.37, 2.24721e-22, 5.59111e-23, 2.41742e-22, 2.50108e-23], [6250.75, 2.25238e-22, 5.60371e-23, 2.42315e-22, 2.50671e-23], [6265.16, 2.25757e-22, 5.61638e-23, 2.42888e-22, 2.51246e-23], [6279.61, 2.26277e-22, 5.62899e-23, 2.43463e-22, 2.51801e-23], [6294.09, 2.26798e-22, 5.64167e-23, 2.44039e-22, 2.52368e-23], [6308.6, 2.27321e-22, 5.65439e-23, 2.44616e-22, 2.52937e-23], [6323.15, 2.27845e-22, 5.66713e-23, 2.45195e-22, 2.53507e-23], [6337.73, 2.28369e-22, 5.67991e-23, 2.45775e-22, 2.54078e-23], [6352.35, 2.28895e-22, 5.69271e-23, 2.46356e-22, 2.5465e-23], [6367.0, 2.29423e-22, 5.70555e-23, 2.46939e-22, 2.55224e-23], [6381.68, 2.29951e-22, 5.71842e-23, 2.47522e-22, 2.55799e-23], [6396.4, 2.30481e-22, 5.73131e-23, 2.48108e-22, 2.56375e-23], [6411.15, 2.31012e-22, 5.74424e-23, 2.48694e-22, 2.56953e-23], [6425.93, 2.31544e-22, 5.75719e-23, 2.49282e-22, 2.57532e-23], [6440.75, 2.32077e-22, 5.77018e-23, 2.49871e-22, 2.58113e-23], [6455.6, 2.32612e-22, 5.7832e-23, 2.50462e-22, 2.58695e-23], [6470.49, 2.33148e-22, 5.79625e-23, 2.51053e-22, 2.59278e-23], [6485.41, 2.33685e-22, 5.80933e-23, 2.51646e-22, 2.59863e-23], [6500.36, 2.34223e-22, 5.82244e-23, 2.52241e-22, 2.60449e-23], [6515.35, 2.34763e-22, 5.83558e-23, 2.52837e-22, 2.61036e-23], [6530.38, 2.35304e-22, 5.84875e-23, 2.53434e-22, 2.61625e-23], [6545.44, 2.35846e-22, 5.86196e-23, 2.54032e-22, 2.62215e-23], [6560.53, 2.36389e-22, 5.87519e-23, 2.54632e-22, 2.62807e-23], [6575.66, 2.36933e-22, 5.88846e-23, 2.55233e-22, 2.634e-23], [6590.82, 2.37479e-22, 5.90175e-23, 2.55836e-22, 2.63994e-23], [6606.02, 2.38026e-22, 5.91508e-23, 2.5644e-22, 2.6459e-23], [6621.26, 2.38575e-22, 5.92844e-23, 2.57045e-22, 2.65187e-23], [6636.52, 2.39124e-22, 5.94183e-23, 2.57652e-22, 2.65786e-23], [6651.83, 2.39675e-22, 5.95525e-23, 2.5826e-22, 2.66386e-23], [6667.17, 2.40227e-22, 5.96871e-23, 2.58869e-22, 2.66988e-23], [6682.54, 2.40781e-22, 5.98219e-23, 2.5948e-22, 2.6759e-23], [6697.95, 2.41335e-22, 5.99571e-23, 2.60092e-22, 2.68195e-23], [6713.4, 2.41891e-22, 6.00926e-23, 2.60706e-22, 2.688e-23], [6728.88, 2.42449e-22, 6.02284e-23, 2.61321e-22, 2.69408e-23], [6744.39, 2.43007e-22, 6.03645e-23, 2.61937e-22, 2.70016e-23], [6759.95, 2.43567e-22, 6.0501e-23, 2.62555e-22, 2.70626e-23], [6775.54, 2.44128e-22, 6.06377e-23, 2.63174e-22, 2.71237e-23], [6791.16, 2.4469e-22, 6.07748e-23, 2.63794e-22, 2.7185e-23], [6806.82, 2.45254e-22, 6.09122e-23, 2.64416e-22, 2.72465e-23], [6822.52, 2.45819e-22, 6.105e-23, 2.65039e-22, 2.7308e-23], [6838.25, 2.46385e-22, 6.1188e-23, 2.65664e-22, 2.73698e-23], [6854.02, 2.46953e-22, 6.13264e-23, 2.6629e-22, 2.74316e-23], [6869.82, 2.47522e-22, 6.14651e-23, 2.66918e-22, 2.74936e-23], [6885.67, 2.48092e-22, 6.16042e-23, 2.67547e-22, 2.75558e-23], [6901.54, 2.48664e-22, 6.17435e-23, 2.68177e-22, 2.76181e-23], [6917.46, 2.49236e-22, 6.18832e-23, 2.68809e-22, 2.76805e-23], [6933.41, 2.49811e-22, 6.20232e-23, 2.69442e-22, 2.77431e-23], [6949.4, 2.50386e-22, 6.21636e-23, 2.70077e-22, 2.78059e-23], [6965.43, 2.50963e-22, 6.23043e-23, 2.70713e-22, 2.78688e-23], [6981.49, 2.51541e-22, 6.24453e-23, 2.7135e-22, 2.79318e-23], [6997.59, 2.52121e-22, 6.25866e-23, 2.71989e-22, 2.7995e-23], [7013.72, 2.52701e-22, 6.27283e-23, 2.7263e-22, 2.80583e-23], [7029.9, 2.53284e-22, 6.28703e-23, 2.73272e-22, 2.81218e-23], [7046.11, 2.53867e-22, 6.30126e-23, 2.73915e-22, 2.81855e-23], [7062.36, 2.54452e-22, 6.31553e-23, 2.7456e-22, 2.82492e-23], [7078.64, 2.55038e-22, 6.32983e-23, 2.75206e-22, 2.83132e-23], [7094.97, 2.55626e-22, 6.34417e-23, 2.75854e-22, 2.83773e-23], [7111.33, 2.56215e-22, 6.35853e-23, 2.76503e-22, 2.84415e-23], [7127.73, 2.56805e-22, 6.37293e-23, 2.77153e-22, 2.85059e-23], [7144.16, 2.57397e-22, 6.38737e-23, 2.77805e-22, 2.85704e-23], [7160.64, 2.5799e-22, 6.40184e-23, 2.78459e-22, 2.86351e-23], [7177.15, 2.58584e-22, 6.41634e-23, 2.79114e-22, 2.86999e-23], [7193.7, 2.5918e-22, 6.43088e-23, 2.7977e-22, 2.87649e-23], [7210.29, 2.59777e-22, 6.44545e-23, 2.80428e-22, 2.88301e-23], [7226.92, 2.60375e-22, 6.46006e-23, 2.81088e-22, 2.88954e-23], [7243.58, 2.60975e-22, 6.4747e-23, 2.81749e-22, 2.89608e-23], [7260.28, 2.61576e-22, 6.48937e-23, 2.82411e-22, 2.90264e-23], [7277.03, 2.62179e-22, 6.50408e-23, 2.83075e-22, 2.90922e-23], [7293.81, 2.62783e-22, 6.51882e-23, 2.83741e-22, 2.91581e-23], [7310.63, 2.63389e-22, 6.5336e-23, 2.84408e-22, 2.92242e-23], [7327.49, 2.63995e-22, 6.54841e-23, 2.85076e-22, 2.92904e-23], [7344.38, 2.64604e-22, 6.56326e-23, 2.85746e-22, 2.93568e-23], [7361.32, 2.65213e-22, 6.57814e-23, 2.86418e-22, 2.94233e-23], [7378.29, 2.65824e-22, 6.59306e-23, 2.87091e-22, 2.949e-23], [7395.31, 2.66437e-22, 6.60801e-23, 2.87766e-22, 2.95568e-23], [7412.36, 2.67051e-22, 6.623e-23, 2.88442e-22, 2.96238e-23], [7429.46, 2.67666e-22, 6.63802e-23, 2.89119e-22, 2.9691e-23], [7446.59, 2.68282e-22, 6.65308e-23, 2.89799e-22, 2.97583e-23], [7463.76, 2.68901e-22, 6.66817e-23, 2.90479e-22, 2.98258e-23], [7480.97, 2.6952e-22, 6.6833e-23, 2.91162e-22, 2.98934e-23], [7498.22, 2.70141e-22, 6.69846e-23, 2.91845e-22, 2.99612e-23], [7515.51, 2.70763e-22, 6.71366e-23, 2.92531e-22, 3.00292e-23], [7532.84, 2.71387e-22, 6.72889e-23, 2.93218e-22, 3.00973e-23], [7550.22, 2.72013e-22, 6.74416e-23, 2.93906e-22, 3.01656e-23], [7567.63, 2.72639e-22, 6.75947e-23, 2.94596e-22, 3.0234e-23], [7585.08, 2.73267e-22, 6.77481e-23, 2.95288e-22, 3.03026e-23], [7602.57, 2.73897e-22, 6.79019e-23, 2.95981e-22, 3.03713e-23], [7620.1, 2.74528e-22, 6.8056e-23, 2.96676e-22, 3.04402e-23], [7637.67, 2.75161e-22, 6.82105e-23, 2.97372e-22, 3.05093e-23], [7655.28, 2.75795e-22, 6.83654e-23, 2.9807e-22, 3.05785e-23], [7672.94, 2.7643e-22, 6.85206e-23, 2.98769e-22, 3.06479e-23], [7690.63, 2.77067e-22, 6.86762e-23, 2.9947e-22, 3.07175e-23], [7708.37, 2.77705e-22, 6.88321e-23, 3.00173e-22, 3.07872e-23], [7726.14, 2.78345e-22, 6.89885e-23, 3.00877e-22, 3.08571e-23], [7743.96, 2.78986e-22, 6.91451e-23, 3.01583e-22, 3.09272e-23], [7761.82, 2.79629e-22, 6.93022e-23, 3.0229e-22, 3.09974e-23], [7779.71, 2.80273e-22, 6.94596e-23, 3.02999e-22, 3.10678e-23], [7797.65, 2.80919e-22, 6.96174e-23, 3.0371e-22, 3.11383e-23], [7815.64, 2.81566e-22, 6.97755e-23, 3.04422e-22, 3.1209e-23], [7833.66, 2.82215e-22, 6.99341e-23, 3.05136e-22, 3.12799e-23], [7851.72, 2.82865e-22, 7.0093e-23, 3.05852e-22, 3.13509e-23], [7869.83, 2.83517e-22, 7.02522e-23, 3.06569e-22, 3.14222e-23], [7887.98, 2.8417e-22, 7.04119e-23, 3.07288e-22, 3.14935e-23], [7906.17, 2.84825e-22, 7.05719e-23, 3.08008e-22, 3.15651e-23], [7924.4, 2.85481e-22, 7.07323e-23, 3.0873e-22, 3.16368e-23], [7942.67, 2.86139e-22, 7.0893e-23, 3.09454e-22, 3.17087e-23], [7960.99, 2.86798e-22, 7.10542e-23, 3.10179e-22, 3.17807e-23], [7979.35, 2.87459e-22, 7.12157e-23, 3.10906e-22, 3.18529e-23], [7997.75, 2.88122e-22, 7.13776e-23, 3.11635e-22, 3.19253e-23], [8016.19, 2.88785e-22, 7.15399e-23, 3.12365e-22, 3.19979e-23], [8034.68, 2.89451e-22, 7.17025e-23, 3.13097e-22, 3.20706e-23], [8053.2, 2.90118e-22, 7.18657e-23, 3.1383e-22, 3.21438e-23], [8071.77, 2.90786e-22, 7.20289e-23, 3.14565e-22, 3.22165e-23], [8090.39, 2.91456e-22, 7.21927e-23, 3.15302e-22, 3.22898e-23], [8109.04, 2.92128e-22, 7.23569e-23, 3.16041e-22, 3.23632e-23], [8127.74, 2.92801e-22, 7.25215e-23, 3.16781e-22, 3.24367e-23], [8146.49, 2.93475e-22, 7.26864e-23, 3.17523e-22, 3.25105e-23], [8165.27, 2.94152e-22, 7.28518e-23, 3.18267e-22, 3.25844e-23], [8184.1, 2.94829e-22, 7.30175e-23, 3.19012e-22, 3.26585e-23], [8202.97, 2.95509e-22, 7.31836e-23, 3.19759e-22, 3.27328e-23], [8221.89, 2.9619e-22, 7.33501e-23, 3.20508e-22, 3.28072e-23], [8240.85, 2.96872e-22, 7.3517e-23, 3.21258e-22, 3.28818e-23], [8259.85, 2.97556e-22, 7.36842e-23, 3.2201e-22, 3.29566e-23], [8278.9, 2.98242e-22, 7.38519e-23, 3.22764e-22, 3.30316e-23], [8297.99, 2.98929e-22, 7.402e-23, 3.2352e-22, 3.31067e-23], [8317.13, 2.99618e-22, 7.41884e-23, 3.24277e-22, 3.31821e-23], [8336.31, 3.00308e-22, 7.43572e-23, 3.25036e-22, 3.32575e-23], [8355.53, 3.01e-22, 7.45265e-23, 3.25797e-22, 3.33332e-23], [8374.8, 3.01694e-22, 7.46961e-23, 3.26559e-22, 3.34091e-23], [8394.11, 3.02389e-22, 7.48661e-23, 3.27323e-22, 3.34851e-23], [8413.47, 3.03085e-22, 7.50366e-23, 3.28089e-22, 3.35613e-23], [8432.87, 3.03784e-22, 7.52074e-23, 3.28857e-22, 3.36377e-23], [8452.31, 3.04484e-22, 7.53786e-23, 3.29626e-22, 3.37142e-23], [8471.81, 3.05185e-22, 7.55502e-23, 3.30397e-22, 3.3791e-23], [8491.34, 3.05889e-22, 7.57223e-23, 3.3117e-22, 3.38679e-23], [8510.92, 3.06593e-22, 7.58947e-23, 3.31945e-22, 3.3945e-23], [8530.55, 3.073e-22, 7.60675e-23, 3.32721e-22, 3.40222e-23], [8550.22, 3.08008e-22, 7.62407e-23, 3.33499e-22, 3.40997e-23], [8569.94, 3.08718e-22, 7.64144e-23, 3.34279e-22, 3.41773e-23], [8589.7, 3.09429e-22, 7.65884e-23, 3.35061e-22, 3.42552e-23], [8609.51, 3.10142e-22, 7.67629e-23, 3.35845e-22, 3.43331e-23], [8629.36, 3.10857e-22, 7.69377e-23, 3.3663e-22, 3.44113e-23], [8649.26, 3.11573e-22, 7.7113e-23, 3.37417e-22, 3.44897e-23], [8669.21, 3.12291e-22, 7.72887e-23, 3.38206e-22, 3.45682e-23], [8689.2, 3.13011e-22, 7.74647e-23, 3.38996e-22, 3.4647e-23], [8709.23, 3.13732e-22, 7.76412e-23, 3.39789e-22, 3.47259e-23], [8729.32, 3.14455e-22, 7.78181e-23, 3.40583e-22, 3.4805e-23], [8749.45, 3.15179e-22, 7.79955e-23, 3.41379e-22, 3.48843e-23], [8769.62, 3.15906e-22, 7.81732e-23, 3.42177e-22, 3.49637e-23], [8789.85, 3.16633e-22, 7.83513e-23, 3.42977e-22, 3.50434e-23], [8810.12, 3.17363e-22, 7.85299e-23, 3.43778e-22, 3.51232e-23], [8830.43, 3.18094e-22, 7.87089e-23, 3.44582e-22, 3.52033e-23], [8850.8, 3.18827e-22, 7.88883e-23, 3.45387e-22, 3.52835e-23], [8871.21, 3.19562e-22, 7.90681e-23, 3.46194e-22, 3.53639e-23], [8891.66, 3.20298e-22, 7.92483e-23, 3.47003e-22, 3.54445e-23], [8912.17, 3.21036e-22, 7.9429e-23, 3.47813e-22, 3.55252e-23], [8932.72, 3.21776e-22, 7.961e-23, 3.48626e-22, 3.56062e-23], [8953.32, 3.22518e-22, 7.97915e-23, 3.4944e-22, 3.56874e-23], [8973.96, 3.23261e-22, 7.99735e-23, 3.50256e-22, 3.57687e-23], [8994.66, 3.24006e-22, 8.01558e-23, 3.51075e-22, 3.58502e-23], [9015.4, 3.24752e-22, 8.03386e-23, 3.51894e-22, 3.5932e-23], [9036.19, 3.25501e-22, 8.05218e-23, 3.52716e-22, 3.60139e-23], [9057.03, 3.26251e-22, 8.07054e-23, 3.5354e-22, 3.6096e-23], [9077.91, 3.27002e-22, 8.08894e-23, 3.54366e-22, 3.61783e-23], [9098.85, 3.27756e-22, 8.10739e-23, 3.55193e-22, 3.62608e-23], [9119.83, 3.28511e-22, 8.12588e-23, 3.56022e-22, 3.63434e-23], [9140.86, 3.29268e-22, 8.14442e-23, 3.56854e-22, 3.64263e-23], [9161.94, 3.30027e-22, 8.163e-23, 3.57687e-22, 3.65094e-23], [9183.07, 3.30788e-22, 8.18162e-23, 3.58522e-22, 3.65926e-23], [9204.24, 3.3155e-22, 8.20028e-23, 3.59359e-22, 3.66761e-23], [9225.47, 3.32314e-22, 8.21899e-23, 3.60197e-22, 3.67597e-23], [9246.74, 3.3308e-22, 8.23774e-23, 3.61038e-22, 3.68436e-23], [9268.06, 3.33847e-22, 8.25653e-23, 3.61881e-22, 3.69276e-23], [9289.44, 3.34616e-22, 8.27537e-23, 3.62725e-22, 3.70119e-23], [9310.86, 3.35387e-22, 8.29426e-23, 3.63572e-22, 3.70963e-23], [9332.33, 3.3616e-22, 8.31318e-23, 3.6442e-22, 3.71809e-23], [9353.85, 3.36935e-22, 8.33215e-23, 3.65271e-22, 3.72658e-23], [9375.42, 3.37711e-22, 8.35117e-23, 3.66123e-22, 3.73508e-23], [9397.04, 3.3849e-22, 8.37023e-23, 3.66977e-22, 3.7436e-23], [9418.71, 3.3927e-22, 8.38933e-23, 3.67833e-22, 3.75214e-23], [9440.43, 3.40051e-22, 8.40848e-23, 3.68692e-22, 3.7607e-23], [9462.2, 3.40835e-22, 8.42767e-23, 3.69552e-22, 3.76929e-23], [9484.02, 3.4162e-22, 8.44691e-23, 3.70414e-22, 3.77789e-23], [9505.89, 3.42408e-22, 8.46619e-23, 3.71278e-22, 3.78651e-23], [9527.81, 3.43197e-22, 8.48552e-23, 3.72144e-22, 3.79515e-23], [9549.78, 3.43988e-22, 8.50489e-23, 3.73012e-22, 3.80381e-23], [9571.8, 3.4478e-22, 8.52431e-23, 3.73882e-22, 3.8125e-23], [9593.87, 3.45575e-22, 8.54377e-23, 3.74754e-22, 3.8212e-23], [9616.0, 3.46371e-22, 8.56328e-23, 3.75628e-22, 3.82992e-23], [9638.17, 3.47169e-22, 8.58283e-23, 3.76504e-22, 3.83867e-23], [9660.4, 3.47969e-22, 8.60243e-23, 3.77381e-22, 3.84743e-23], [9682.67, 3.48771e-22, 8.62207e-23, 3.78261e-22, 3.85621e-23], [9705.0, 3.49575e-22, 8.64177e-23, 3.79143e-22, 3.86502e-23], [9727.38, 3.50381e-22, 8.6615e-23, 3.80027e-22, 3.87384e-23], [9749.81, 3.51188e-22, 8.68128e-23, 3.80913e-22, 3.88269e-23], [9772.3, 3.51997e-22, 8.70111e-23, 3.81801e-22, 3.89155e-23], [9794.83, 3.52808e-22, 8.72099e-23, 3.82691e-22, 3.90044e-23], [9817.42, 3.53621e-22, 8.74091e-23, 3.83583e-22, 3.90935e-23], [9840.06, 3.54436e-22, 8.76087e-23, 3.84477e-22, 3.91828e-23], [9862.75, 3.55253e-22, 8.78089e-23, 3.85373e-22, 3.92723e-23], [9885.49, 3.56072e-22, 8.80095e-23, 3.86272e-22, 3.9362e-23], [9908.29, 3.56892e-22, 8.82105e-23, 3.87172e-22, 3.94519e-23], [9931.14, 3.57715e-22, 8.84121e-23, 3.88074e-22, 3.9542e-23], [9954.04, 3.58539e-22, 8.86141e-23, 3.88978e-22, 3.96323e-23], [9976.99, 3.59365e-22, 8.88165e-23, 3.89885e-22, 3.97228e-23], [10000.0, 3.60194e-22, 8.90195e-23, 3.90793e-22, 3.98136e-23]])

    f1 = data[:,0]
    S1 = data[:,col]**2
    fstop = [f1[0],f1[-1]]

# % ---- Interpolate to desired frequencies.  Set noise to infinity outside
# %      band of frequencies covered by f1.
# S = inf(size(f)); 
# k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop


# % -------------------------------------------------------------------------
# %                              LIGO3 spectrum
# % -------------------------------------------------------------------------

def ligo3spectrum(IFO,f):
# % LIGO3SPECTRUM - Strawman design noise curves for third-generation LIGO.  Data 
# % extracted from Figure 3 of T1400316-v5 and should be accurate to a few percent
# % in amplitude. These spectra are available from 10 Hz to 2000 Hz only.
# %
# % Arguments are the same as for the parent function. Recognised values of IFO
# % are 'LIGO-A', 'LIGO-Voyager', 'LIGO-CosmicExplorer' (case-insensitive).

    if IFO.upper()=='LIGO-A':
        data = np.array([[10.0, 3.9678080739705763e-23], [11.309284110571209, 3.0398221659058543e-23], [12.45718308341748, 2.5130640029321575e-23], [13.601530704926319, 2.0775855751257282e-23], [15.247762082845437, 1.6851937135883917e-23], [17.396345134910742, 1.3158664493151312e-23], [19.674020961625892, 1.0472208012050727e-23], [22.055222852758373, 8.494330785017142e-24], [24.942878136380717, 7.022383468430227e-24], [27.474594902238422, 6.030715466658035e-24], [30.799922117601024, 4.985677820619556e-24], [33.9261322841218, 4.447719945193319e-24], [39.393071773318546, 3.607682774264497e-24], [46.144732996187464, 3.0398221659058542e-24], [52.6470505637754, 2.660706299234269e-24], [60.06561861120157, 2.373614361213412e-24], [69.1344729755857, 2.077585575125728e-24], [82.4196060504594, 1.8890202226795184e-24], [95.70090187362572, 1.7505670753219003e-24], [114.0911370494335, 1.6222616615793647e-24], [132.4760604258842, 1.5616797507666476e-24], [149.8209405205523, 1.5033602172292137e-24], [173.96345134910743, 1.447218574511171e-24], [207.3929041512335, 1.419938926158929e-24], [249.42878136380742, 1.419938926158929e-24], [289.62234198427996, 1.419938926158929e-24], [342.2560588399126, 1.447218574511171e-24], [418.92608639188995, 1.5322425426005607e-24], [495.058462311237, 1.6534283032200178e-24], [595.4004524093444, 1.784198700846088e-24], [703.6039101022948, 2.0000000000000002e-24], [824.1960605045934, 2.2419027645873307e-24], [957.0090187362568, 2.5613446022056183e-24], [1063.4511794422535, 2.7639225134676383e-24], [1213.3035425912797, 3.098222762024156e-24], [1384.271807603447, 3.4729570877446746e-24], [1565.5123158441343, 3.893016048153975e-24], [1739.6345134910744, 4.3638817204696554e-24], [1984.7687969463493, 4.8916992467265186e-24], [2073.929041512335, 5.278585968283095e-24]])
    elif IFO.upper()=='LIGO-VOYAGER':
        data = np.array([[9.912500131952504, 2.3288724222705647e-22], [10.177323697119471, 1.818476451989422e-22], [10.728385021799559, 1.0274809943043322e-22], [11.611424327245047, 5.179086161171733e-23], [12.567145440192537, 2.926302725914811e-23], [13.84271807603448, 1.6534283032200178e-23], [15.518141037378271, 1.0472208012050727e-23], [17.244127344529417, 7.577786975592319e-24], [19.331232401691526, 5.379997350391837e-24], [21.862239943820917, 3.893016048153975e-24], [25.60925511753295, 2.7639225134676383e-24], [29.998479083578562, 2.077585575125728e-24], [34.52772300468027, 1.6534283032200178e-24], [40.80252343563839, 1.393173491227759e-24], [49.072670729843466, 1.196436076021249e-24], [60.595831335812754, 1.0472208012050726e-24], [74.82484080584663, 9.70466188329871e-25], [89.9908747327901, 9.16615127590918e-25], [115.09824517597313, 8.82384955464901e-25], [140.88182270776983, 8.657522562166054e-25], [180.18772035503366, 8.494330785017143e-25], [228.4434056085519, 8.494330785017143e-25], [297.35992787434714, 8.657522562166054e-25], [380.32307148494033, 8.993371995740937e-25], [495.058462311237, 9.70466188329871e-25], [627.6390027182711, 1.0673398462412604e-24], [761.5166254665346, 1.196436076021249e-24], [899.9087473279005, 1.2910627502415792e-24], [1101.5005867412315, 1.5033602172292137e-24], [1372.159447552731, 1.784198700846088e-24], [1679.541888074497, 2.1581809109852513e-24], [2002.288797503804, 2.5130640029321577e-24]])
    elif IFO.upper()=='LIGO-COSMICEXPLORER':
        data = np.array([[9.825765886595839, 1.1964360760212489e-23], [10.088272249062012, 1.0472208012050727e-23], [11.309284110571209, 5.917038237702881e-24], [12.789990709361842, 3.4729570877446746e-24], [14.3379991273717, 2.373614361213412e-24], [15.932725590172907, 1.7175693961372587e-24], [18.663468857943204, 1.2194218219772992e-24], [21.29336383289432, 9.342249965045526e-25], [24.724628281814923, 7.294801298506009e-25], [28.45761327865132, 5.805503791383494e-25], [33.62927907429953, 4.799492141562294e-25], [39.393071773318546, 4.0440369926888656e-25], [47.79575353611991, 3.5396790053924888e-25], [57.990757086193945, 3.2184114594537887e-25], [70.36039101022943, 3.0398221659058544e-25], [84.62153832015815, 2.982522404004116e-25], [105.41459956546393, 3.098222762024156e-25], [129.0289083390161, 3.1577453414167256e-25], [157.9331445149521, 3.343262611490267e-25], [193.31232401691528, 3.7476348457207434e-25], [236.6169225059843, 4.200916460642608e-25], [279.6178457673089, 4.709023113457543e-25], [333.3502332614619, 5.379997350391837e-25], [397.40803277607773, 6.264663730608179e-25], [486.4328550848352, 7.43494789957218e-25], [590.1907063072211, 8.82384955464901e-25], [728.7782784890753, 1.0878454152776334e-24], [868.822976815246, 1.2910627502415792e-24], [1000.0, 1.4750223152736534e-24], [1161.1424327245047, 1.7175693961372587e-24], [1360.153070492633, 2.0000000000000002e-24], [1621.5250570856872, 2.419215875408719e-24], [1984.7687969463493, 2.926302725914811e-24], [2037.7941227312338, 3.0398221659058542e-24]])

    f1 = data[:,0]
    S1 = data[:,1]**2
    fstop = [10,2000]

#     % ---- Interpolate to desired frequencies.  Set noise to infinity outside
#     %      band of frequencies covered by f1.
#     S = inf(size(f));
#     k = find(f>=f1(1) & f<=f1(end));
##### S = exp(interp1d(np.log(f1),np.log(S1),log(f(k))));
    S = interp1d(f1,S1)

    return S, fstop


# % -------------------------------------------------------------------------
# %                              LIGO3 spectrum (OLD)
# % -------------------------------------------------------------------------

def ligo3spectrumold(IFO,f):
# % LIGO3SPECTRUM - Design noise curve for LIGO3.  Data produced by Rana
# % for straw man challenge, November 2011.
# %
# % Arguments are the same as for the parent function.

# %ascii file of 3G LIGO Noise curves
# %Format:   frequency   strain   strain   strain
# %* Column 2:   Standard 3G: 100kg masses, 10dB squeezing, 2.5x thermal noise reduction
# %* Column 3:  mid-freq  3G: 100kg masses, 10dB squeezing, 5x thermal noise reduction
# %* Column 4: high-freq  3G: 100kg masses, 15dB squeezing, 2.5x thermal noise reduction
    data = np.array([[1.0, 2.4969431523912284e-15, 2.496943152391229e-15, 2.4969431523911306e-15], [1.0027727599724743, 2.4799308119902096e-15, 2.4799308119902104e-15, 2.4799308119901126e-15], [1.0055532081428138, 2.4640336853632634e-15, 2.4640336853632638e-15, 2.4640336853631664e-15], [1.0083413658285454, 2.4492474670157327e-15, 2.4492474670157334e-15, 2.4492474670156364e-15], [1.0111372544063049, 2.435570304876475e-15, 2.4355703048764756e-15, 2.4355703048763794e-15], [1.0139408953120004, 2.423002864140451e-15, 2.4230028641404516e-15, 2.4230028641403558e-15], [1.0167523100409763, 2.4115484077513084e-15, 2.4115484077513088e-15, 2.4115484077512137e-15], [1.0195715201481788, 2.401212894902249e-15, 2.4012128949022496e-15, 2.401212894902155e-15], [1.0223985472483206, 2.3920050991996837e-15, 2.392005099199684e-15, 2.3920050991995903e-15], [1.0252334130160465, 2.3839367484398114e-15, 2.3839367484398122e-15, 2.383936748439719e-15], [1.0280761391861009, 2.377022688300828e-15, 2.3770226883008283e-15, 2.377022688300736e-15], [1.030926747553492, 2.3712810726637166e-15, 2.371281072663717e-15, 2.3712810726636255e-15], [1.0337852599736614, 2.366733583753467e-15, 2.3667335837534675e-15, 2.3667335837533768e-15], [1.0366516983626506, 2.363405685855553e-15, 2.3634056858555534e-15, 2.363405685855464e-15], [1.039526084697268, 2.361326917026668e-15, 2.3613269170266683e-15, 2.3613269170265795e-15], [1.0424084410152596, 2.360531224006219e-15, 2.3605312240062194e-15, 2.3605312240061314e-15], [1.0452987894154762, 2.361057346473431e-15, 2.3610573464734315e-15, 2.3610573464733443e-15], [1.0481971520580433, 2.3629492579169994e-15, 2.362949257917e-15, 2.362949257916914e-15], [1.0511035511645315, 2.366256671732868e-15, 2.3662566717328684e-15, 2.366256671732783e-15], [1.0540180090181261, 2.371035622793245e-15, 2.3710356227932452e-15, 2.3710356227931616e-15], [1.0569405479637988, 2.3773491367030194e-15, 2.3773491367030198e-15, 2.3773491367029373e-15], [1.059871190408478, 2.385268001362562e-15, 2.3852680013625625e-15, 2.3852680013624813e-15], [1.0628099588212214, 2.3948716583960887e-15, 2.394871658396089e-15, 2.394871658396009e-15], [1.065756875733388, 2.406249235619524e-15, 2.4062492356195246e-15, 2.4062492356194457e-15], [1.068711963738811, 2.419500746188872e-15, 2.4195007461888724e-15, 2.4195007461887947e-15], [1.0716752454939704, 2.434738485618952e-15, 2.4347384856189526e-15, 2.434738485618876e-15], [1.0746467437181677, 2.4520886647931656e-15, 2.452088664793166e-15, 2.452088664793091e-15], [1.0776264811936995, 2.4716933257911615e-15, 2.471693325791162e-15, 2.4716933257910886e-15], [1.0806144807660316, 2.493712598364362e-15, 2.493712598364362e-15, 2.49371259836429e-15], [1.083610765343976, 2.5183273688822146e-15, 2.5183273688822154e-15, 2.5183273688821448e-15], [1.086615357899864, 2.5457424514878995e-15, 2.5457424514879e-15, 2.545742451487831e-15], [1.0896282814697247, 2.5761903743006607e-15, 2.576190374300661e-15, 2.5761903743005936e-15], [1.09264955915346, 2.6099359235066503e-15, 2.6099359235066507e-15, 2.609935923506585e-15], [1.0956792141150224, 2.647281627463616e-15, 2.647281627463616e-15, 2.6472816274635523e-15], [1.0987172695825929, 2.6885744148087997e-15, 2.6885744148088e-15, 2.6885744148087377e-15], [1.1017637488487577, 2.734213749648235e-15, 2.7342137496482355e-15, 2.7342137496481748e-15], [1.104818675270689, 2.7846616398161957e-15, 2.784661639816196e-15, 2.784661639816137e-15], [1.1078820722703218, 2.840455040420256e-15, 2.840455040420256e-15, 2.8404550404201987e-15], [1.1109539633345349, 2.902221348251617e-15, 2.9022213482516174e-15, 2.9022213482515618e-15], [1.1140343720153305, 2.9706979235505335e-15, 2.970697923550534e-15, 2.9706979235504802e-15], [1.1171233219300154, 3.0467569146370202e-15, 3.0467569146370206e-15, 3.046756914636969e-15], [1.1202208367613806, 3.1314371445458516e-15, 3.131437144545852e-15, 3.1314371445458023e-15], [1.1233269402578843, 3.2259855189898646e-15, 3.225985518989865e-15, 3.2259855189898173e-15], [1.1264416562338333, 3.3319114450995928e-15, 3.331911445099593e-15, 3.3319114450995474e-15], [1.1295650085695663, 3.45105929274363e-15, 3.4510592927436305e-15, 3.4510592927435867e-15], [1.1326970212116356, 3.585706284383107e-15, 3.585706284383107e-15, 3.585706284383065e-15], [1.1358377181729922, 3.738696869593792e-15, 3.738696869593792e-15, 3.738696869593753e-15], [1.138987123533169, 3.913630498175915e-15, 3.913630498175915e-15, 3.913630498175878e-15], [1.1421452614384655, 4.1151293027244286e-15, 4.1151293027244286e-15, 4.115129302724394e-15], [1.1453121561021333, 4.34922839209985e-15, 4.34922839209985e-15, 4.349228392099818e-15], [1.1484878318045617, 4.6239596909511725e-15, 4.6239596909511725e-15, 4.6239596909511426e-15], [1.1516723128934634, 4.9502513875414985e-15, 4.9502513875414985e-15, 4.950251387541471e-15], [1.1548656237840613, 5.343361740625097e-15, 5.343361740625097e-15, 5.343361740625071e-15], [1.1580677889592763, 5.8252583166690915e-15, 5.8252583166690915e-15, 5.8252583166690686e-15], [1.1612788329699146, 6.428759753956306e-15, 6.428759753956306e-15, 6.428759753956286e-15], [1.1644987804348552, 7.205177897026037e-15, 7.205177897026037e-15, 7.20517789702602e-15], [1.1677276560412404, 8.239477499272201e-15, 8.239477499272201e-15, 8.239477499272185e-15], [1.1709654845446629, 9.683272890605722e-15, 9.683272890605722e-15, 9.68327289060571e-15], [1.1742122907693573, 1.1836119246915019e-14, 1.1836119246915019e-14, 1.1836119246915008e-14], [1.17746809960839, 1.5384171068271614e-14, 1.5384171068271614e-14, 1.5384171068271604e-14], [1.1807329360238497, 2.232090924248319e-14, 2.232090924248319e-14, 2.2320909242483183e-14], [1.1840068250470388, 4.189264851061066e-14, 4.189264851061066e-14, 4.189264851061066e-14], [1.1872897917786656, 4.553406014761366e-13, 4.553406014761366e-13, 4.553406014761366e-13], [1.190581861389037, 4.941582493404276e-14, 4.941582493404276e-14, 4.941582493404275e-14], [1.1938830591182505, 2.3019413016073054e-14, 2.3019413016073054e-14, 2.3019413016073048e-14], [1.1971934102763888, 1.4829354505815323e-14, 1.4829354505815323e-14, 1.4829354505815313e-14], [1.2005129402437134, 1.0842734542197864e-14, 1.0842734542197864e-14, 1.0842734542197853e-14], [1.2038416744708587, 8.48624643466405e-15, 8.48624643466405e-15, 8.486246434664035e-15], [1.207179638479028, 6.9310295213142125e-15, 6.931029521314213e-15, 6.931029521314196e-15], [1.2105268578601889, 5.828683412556677e-15, 5.828683412556677e-15, 5.828683412556658e-15], [1.2138833582772688, 5.007201360424949e-15, 5.0072013604249495e-15, 5.0072013604249266e-15], [1.2172491654643527, 4.371879090695165e-15, 4.371879090695165e-15, 4.37187909069514e-15], [1.2206243052268801, 3.8662753923384135e-15, 3.8662753923384135e-15, 3.866275392338385e-15], [1.2240088034418426, 3.4546611552617052e-15, 3.4546611552617056e-15, 3.4546611552616736e-15], [1.2274026860579823, 3.1133172049619063e-15, 3.1133172049619063e-15, 3.1133172049618716e-15], [1.2308059790959915, 2.8258794758986705e-15, 2.825879475898671e-15, 2.825879475898633e-15], [1.2342187086487109, 2.580693379956886e-15, 2.580693379956886e-15, 2.580693379956845e-15], [1.2376409008813312, 2.3692334893353068e-15, 2.369233489335307e-15, 2.369233489335263e-15], [1.241072582031592, 2.1851195444902665e-15, 2.185119544490267e-15, 2.1851195444902195e-15], [1.2445137784099847, 2.0234817969936006e-15, 2.023481796993601e-15, 2.0234817969935505e-15], [1.2479645163999527, 1.8805390802660167e-15, 1.880539080266017e-15, 1.8805390802659634e-15], [1.2514248224580948, 1.753310822468676e-15, 1.7533108224686762e-15, 1.7533108224686192e-15], [1.2548947231143677, 1.6394158875698742e-15, 1.6394158875698746e-15, 1.6394158875698145e-15], [1.2583742449722883, 1.536929164167179e-15, 1.5369291641671793e-15, 1.5369291641671158e-15], [1.2618634147091403, 1.4442774443509534e-15, 1.4442774443509536e-15, 1.444277444350887e-15], [1.2653622590761755, 1.360162582843446e-15, 1.3601625828434463e-15, 1.3601625828433763e-15], [1.2688708048988218, 1.2835039465561415e-15, 1.2835039465561419e-15, 1.2835039465560685e-15], [1.2723890790768866, 1.2133947316720904e-15, 1.213394731672091e-15, 1.213394731672014e-15], [1.2759171085847645, 1.1490684003477406e-15, 1.149068400347741e-15, 1.1490684003476607e-15], [1.2794549204716437, 1.089872603724341e-15, 1.0898726037243413e-15, 1.0898726037242575e-15], [1.2830025418617126, 1.035248712994704e-15, 1.0352487129947044e-15, 1.0352487129946172e-15], [1.2865599999543698, 9.847156001930845e-16, 9.847156001930851e-16, 9.847156001929944e-16], [1.2901273220244298, 9.37856673802768e-16, 9.378566738027683e-16, 9.378566738026742e-16], [1.2937045354223349, 8.94309431852108e-16, 8.943094318521084e-16, 8.943094318520108e-16], [1.2972916675743624, 8.537569800770362e-16, 8.5375698007703665e-16, 8.537569800769356e-16], [1.300888745982837, 8.159210970533625e-16, 8.15921097053363e-16, 8.159210970532584e-16], [1.3044957982263408, 7.805565268669945e-16, 7.80556526866995e-16, 7.805565268668869e-16], [1.3081128519599237, 7.474462531111723e-16, 7.474462531111729e-16, 7.474462531110613e-16], [1.3117399349153176, 7.163975628653961e-16, 7.163975628653967e-16, 7.163975628652815e-16], [1.3153770749011469, 6.872387507942723e-16, 6.872387507942729e-16, 6.872387507941541e-16], [1.3190242998031432, 6.598163451374515e-16, 6.598163451374521e-16, 6.598163451373298e-16], [1.3226816375843584, 6.339927616761077e-16, 6.339927616761083e-16, 6.339927616759825e-16], [1.326349116285379, 6.096443105913427e-16, 6.096443105913432e-16, 6.096443105912138e-16], [1.330026764024542, 5.866594958164552e-16, 5.866594958164559e-16, 5.866594958163229e-16], [1.3337146089981489, 5.649375580168748e-16, 5.649375580168755e-16, 5.649375580167388e-16], [1.3374126794806833, 5.443872214443256e-16, 5.443872214443263e-16, 5.443872214441861e-16], [1.3411210038250272, 5.249256121562101e-16, 5.249256121562108e-16, 5.249256121560669e-16], [1.3448396104626776, 5.064773208836338e-16, 5.064773208836344e-16, 5.06477320883487e-16], [1.3485685279039668, 4.889735884881657e-16, 4.889735884881664e-16, 4.889735884880153e-16], [1.3523077847382776, 4.723515957105006e-16, 4.723515957105013e-16, 4.723515957103466e-16], [1.3560574096342655, 4.565538419701741e-16, 4.565538419701749e-16, 4.565538419700167e-16], [1.3598174313400766, 4.415276004691572e-16, 4.41527600469158e-16, 4.4152760046899623e-16], [1.3635878786835693, 4.2722443889598163e-16, 4.272244388959824e-16, 4.272244388958171e-16], [1.3673687805725343, 4.1359979670951246e-16, 4.135997967095133e-16, 4.1359979670934443e-16], [1.371160165994917, 4.0061261137230853e-16, 4.006126113723093e-16, 4.0061261137213696e-16], [1.374962064019039, 3.8822498705758823e-16, 3.8822498705758907e-16, 3.882249870574131e-16], [1.3787745037938217, 3.7640190031530885e-16, 3.7640190031530974e-16, 3.7640190031513027e-16], [1.3825975145490095, 3.6511093798675194e-16, 3.6511093798675283e-16, 3.6511093798656986e-16], [1.3864311255953936, 3.543220633314336e-16, 3.543220633314345e-16, 3.54322063331248e-16], [1.390275366325037, 3.4400740689799087e-16, 3.440074068979918e-16, 3.440074068978019e-16], [1.3941302662115003, 3.3414107915025214e-16, 3.3414107915025303e-16, 3.3414107915005966e-16], [1.3979958548100666, 3.2469900226606795e-16, 3.246990022660689e-16, 3.246990022658721e-16], [1.401872161757969, 3.156587588717018e-16, 3.156587588717028e-16, 3.1565875887150264e-16], [1.4057592167746176, 3.0699945576901564e-16, 3.069994557690166e-16, 3.0699945576881305e-16], [1.4096570496618273, 2.987016009641229e-16, 2.9870160096412387e-16, 2.9870160096391694e-16], [1.4135656903040459, 2.907469925219e-16, 2.90746992521901e-16, 2.9074699252169076e-16], [1.417485168668584, 2.8311861795597467e-16, 2.831186179559757e-16, 2.831186179557622e-16], [1.4214155148058443, 2.758005630235048e-16, 2.7580056302350586e-16, 2.758005630232891e-16], [1.4253567588495522, 2.687779289319422e-16, 2.687779289319433e-16, 2.6877792893172336e-16], [1.429308931016986, 2.6203675708444356e-16, 2.6203675708444465e-16, 2.620367570842216e-16], [1.43327206160921, 2.555639605942129e-16, 2.55563960594214e-16, 2.555639605939878e-16], [1.437246181011306, 2.493472618883339e-16, 2.4934726188833505e-16, 2.493472618881058e-16], [1.4412313196926059, 2.4337513580029077e-16, 2.433751358002919e-16, 2.433751358000596e-16], [1.445227508206926, 2.376367576192239e-16, 2.3763675761922503e-16, 2.3763675761898975e-16], [1.449234777192801, 2.321219556242744e-16, 2.321219556242756e-16, 2.3212195562403735e-16], [1.4532531573737189, 2.2682116768535213e-16, 2.2682116768535327e-16, 2.268211676851122e-16], [1.4572826795583569, 2.2172540155839023e-16, 2.2172540155839139e-16, 2.2172540155814753e-16], [1.4613233746408163, 2.1682619854437816e-16, 2.1682619854437935e-16, 2.1682619854413273e-16], [1.4653752736008616, 2.1211560021801092e-16, 2.121156002180121e-16, 2.1211560021776282e-16], [1.4694384075041558, 2.0758611796430555e-16, 2.0758611796430674e-16, 2.0758611796405484e-16], [1.4735128075025, 2.0323070509046213e-16, 2.0323070509046336e-16, 2.0323070509020888e-16], [1.4775985048340712, 1.9904273130623124e-16, 1.9904273130623247e-16, 1.9904273130597553e-16], [1.481695530823663, 1.9501595938934585e-16, 1.9501595938934709e-16, 1.9501595938908775e-16], [1.485803916882925, 1.9114452387367378e-16, 1.9114452387367503e-16, 1.9114452387341338e-16], [1.4899236945106038, 1.8742291161691512e-16, 1.8742291161691638e-16, 1.8742291161665248e-16], [1.4940548952927837, 1.8384594412222388e-16, 1.8384594412222514e-16, 1.8384594412195907e-16], [1.4981975509031311, 1.8040876150429879e-16, 1.8040876150430007e-16, 1.8040876150403195e-16], [1.5023516931031344, 1.771068080056169e-16, 1.771068080056182e-16, 1.771068080053481e-16], [1.50651735374235, 1.7393581898262932e-16, 1.7393581898263063e-16, 1.7393581898235865e-16], [1.5106945647586447, 1.7089180929529813e-16, 1.708918092952994e-16, 1.7089180929502567e-16], [1.5148833581784422, 1.679710630464255e-16, 1.6797106304642682e-16, 1.679710630461514e-16], [1.519083766116967, 1.6517012463003378e-16, 1.651701246300351e-16, 1.651701246297581e-16], [1.5232958207784915, 1.6248579106088253e-16, 1.6248579106088386e-16, 1.6248579106060541e-16], [1.5275195544565838, 1.5991510557016195e-16, 1.5991510557016328e-16, 1.5991510556988348e-16], [1.531754999534353, 1.574553524658744e-16, 1.5745535246587572e-16, 1.574553524655947e-16], [1.5360021884846993, 1.5510405327050387e-16, 1.551040532705052e-16, 1.5510405327022309e-16], [1.5402611538705626, 1.5285896416376467e-16, 1.52858964163766e-16, 1.5285896416348292e-16], [1.5445319283451722, 1.5071807477465184e-16, 1.507180747746532e-16, 1.5071807477436928e-16], [1.5488145446522963, 1.4867960838532373e-16, 1.4867960838532506e-16, 1.4867960838504046e-16], [1.5531090356264943, 1.4674202362976798e-16, 1.4674202362976931e-16, 1.4674202362948414e-16], [1.5574154341933677, 1.4490401779350205e-16, 1.4490401779350338e-16, 1.449040177932178e-16], [1.5617337733698131, 1.4316453184729759e-16, 1.4316453184729892e-16, 1.4316453184701308e-16], [1.5660640862642743, 1.415227573790002e-16, 1.4152275737900153e-16, 1.4152275737871557e-16], [1.5704064060769976, 1.3997814562392858e-16, 1.399781456239299e-16, 1.39978145623644e-16], [1.5747607661002851, 1.3853041883736647e-16, 1.385304188373678e-16, 1.385304188370821e-16], [1.579127199718751, 1.3717958430391157e-16, 1.371795843039129e-16, 1.3717958430362758e-16], [1.5835057404095767, 1.3592595133983022e-16, 1.3592595133983157e-16, 1.359259513395468e-16], [1.5878964217427678, 1.347701517186668e-16, 1.3477015171866812e-16, 1.347701517183841e-16], [1.5922992773814115, 1.337131640402556e-16, 1.3371316404025691e-16, 1.3371316403997383e-16], [1.5967143410819344, 1.3275634267304968e-16, 1.32756342673051e-16, 1.3275634267276902e-16], [1.6011416466943624, 1.3190145203447505e-16, 1.3190145203447636e-16, 1.319014520341957e-16], [1.605581228162578, 1.3115070714049648e-16, 1.3115070714049779e-16, 1.3115070714021865e-16], [1.6100331195245836, 1.3050682156238136e-16, 1.3050682156238266e-16, 1.3050682156210525e-16], [1.6144973549127595, 1.2997306418713116e-16, 1.2997306418713244e-16, 1.29973064186857e-16], [1.6189739685541273, 1.2955332650299252e-16, 1.2955332650299378e-16, 1.2955332650272051e-16], [1.6234629947706123, 1.2925220254270859e-16, 1.2925220254270984e-16, 1.2925220254243897e-16], [1.6279644679793055, 1.2907508414093833e-16, 1.2907508414093959e-16, 1.2907508414067135e-16], [1.6324784226927291, 1.2902827483405405e-16, 1.2902827483405528e-16, 1.2902827483378993e-16], [1.6370048935190997, 1.2911912659841632e-16, 1.2911912659841753e-16, 1.291191265981553e-16], [1.6415439151625941, 1.293562047530411e-16, 1.2935620475304228e-16, 1.2935620475278343e-16], [1.6460955224236158, 1.2974948783534848e-16, 1.2974948783534963e-16, 1.2974948783509446e-16], [1.6506597501990614, 1.3031061122140264e-16, 1.303106112214038e-16, 1.3031061122115252e-16], [1.6552366334825879, 1.31053165883949e-16, 1.3105316588395013e-16, 1.3105316588370304e-16], [1.6598262073648815, 1.319930672181629e-16, 1.31993067218164e-16, 1.3199306721792143e-16], [1.6644285070339269, 1.331490136853602e-16, 1.331490136853613e-16, 1.331490136851235e-16], [1.669043567775276, 1.345430616686256e-16, 1.3454306166862667e-16, 1.3454306166839393e-16], [1.673671424972319, 1.3620135220142162e-16, 1.3620135220142268e-16, 1.3620135220119532e-16], [1.6783121141065565, 1.3815503832464718e-16, 1.381550383246482e-16, 1.3815503832442652e-16], [1.68296567075787, 1.4044148059029498e-16, 1.4044148059029596e-16, 1.4044148059008033e-16], [1.6876321306047959, 1.431058055248193e-16, 1.4310580552482025e-16, 1.43105805524611e-16], [1.6923115294247986, 1.4620296223341646e-16, 1.462029622334174e-16, 1.4620296223321483e-16], [1.6970039030945445, 1.4980047311576855e-16, 1.4980047311576944e-16, 1.4980047311557395e-16], [1.701709287590178, 1.5398216803688711e-16, 1.5398216803688795e-16, 1.5398216803669988e-16], [1.7064277189875958, 1.5885333788812006e-16, 1.5885333788812087e-16, 1.588533378879406e-16], [1.7111592334627252, 1.645479792641861e-16, 1.6454797926418688e-16, 1.6454797926401477e-16], [1.7159038672918008, 1.7123919165324119e-16, 1.7123919165324193e-16, 1.7123919165307836e-16], [1.7206616568516415, 1.791544523947867e-16, 1.7915445239478743e-16, 1.7915445239463283e-16], [1.7254326386199312, 1.8859866524991514e-16, 1.8859866524991579e-16, 1.8859866524977059e-16], [1.7302168491754975, 1.9999002535068547e-16, 1.9999002535068609e-16, 1.9999002535055065e-16], [1.735014325198592, 2.1391786375543372e-16, 2.139178637554343e-16, 2.1391786375530908e-16], [1.7398251034711723, 2.3123997386544165e-16, 2.312399738654422e-16, 2.3123997386532766e-16], [1.744649220877183, 2.532548944387014e-16, 2.532548944387019e-16, 2.5325489443859844e-16], [1.74948671440284, 2.820264161272871e-16, 2.8202641612728753e-16, 2.820264161271957e-16], [1.754337621136912, 3.210443248447557e-16, 3.2104432484475607e-16, 3.210443248446763e-16], [1.7592019782710064, 3.767126799807975e-16, 3.767126799807978e-16, 3.7671267998073057e-16], [1.764079823099854, 4.621937767815877e-16, 4.62193776781588e-16, 4.621937767815337e-16], [1.7689711930215948, 6.095166454745557e-16, 6.0951664547455575e-16, 6.095166454745151e-16], [1.7738761255380655, 9.222570101641354e-16, 9.222570101641356e-16, 9.22257010164109e-16], [1.7787946582550853, 2.027967873591055e-15, 2.027967873591055e-15, 2.0279678735910432e-15], [1.7837268288827464, 7.438712740798656e-15, 7.438712740798656e-15, 7.438712740798653e-15], [1.7886726752357012, 1.2502089416223779e-15, 1.2502089416223779e-15, 1.250208941622359e-15], [1.7936322352334533, 6.645354433633157e-16, 6.645354433633159e-16, 6.645354433632806e-16], [1.7986055469006483, 4.442600121492095e-16, 4.442600121492098e-16, 4.442600121491576e-16], [1.803592648367365, 3.2895104078611573e-16, 3.2895104078611603e-16, 3.289510407860464e-16], [1.8085935778694069, 2.5815126317137154e-16, 2.5815126317137194e-16, 2.581512631712842e-16], [1.8136083737485975, 2.1034986644028965e-16, 2.1034986644029014e-16, 2.1034986644018365e-16], [1.818637074453072, 1.7597125505071937e-16, 1.7597125505071993e-16, 1.7597125505059404e-16], [1.8236797185375733, 1.5010572463646688e-16, 1.5010572463646752e-16, 1.5010572463632158e-16], [1.828736344663748, 1.2997620934296834e-16, 1.2997620934296908e-16, 1.299762093428024e-16], [1.8338069916004403, 1.1389381943349114e-16, 1.1389381943349196e-16, 1.1389381943330386e-16], [1.8388916982239938, 1.0077244790392913e-16, 1.0077244790393005e-16, 1.0077244790371984e-16], [1.8439905035185447, 8.988191335159938e-17, 8.988191335160039e-17, 8.988191335136733e-17], [1.8491034465763239, 8.071323173233307e-17, 8.071323173233419e-17, 8.071323173207754e-17], [1.854230566597955, 7.290077840785175e-17, 7.290077840785299e-17, 7.290077840757199e-17], [1.8593719028927564, 6.617515704010905e-17, 6.617515704011038e-17, 6.617515703980427e-17], [1.864527494879041, 6.033351456065114e-17, 6.033351456065258e-17, 6.033351456032055e-17], [1.8696973820844196, 5.5220205908788687e-17, 5.522020590879024e-17, 5.522020590843151e-17], [1.8748816041461034, 5.071382989426388e-17, 5.071382989426555e-17, 5.0713829893879285e-17], [1.8800802008112083, 4.671831450616121e-17, 4.6718314506163e-17, 4.6718314505748363e-17], [1.8852932119370591, 4.3156649043836076e-17, 4.315664904383799e-17, 4.315664904339413e-17], [1.890520677491496, 3.996638950652172e-17, 3.996638950652377e-17, 3.99663895060498e-17], [1.8957626375531795, 3.709637834508815e-17, 3.709637834509032e-17, 3.709637834458536e-17], [1.9010191323118995, 3.450431232629687e-17, 3.450431232629917e-17, 3.450431232576232e-17], [1.906290202068882, 3.2154913288993834e-17, 3.2154913288996274e-17, 3.21549132884266e-17], [1.9115758872370983, 3.0018534392250297e-17, 3.001853439225288e-17, 3.001853439164945e-17], [1.9168762283415768, 2.807008555352612e-17, 2.807008555352885e-17, 2.8070085552890714e-17], [1.9221912660197098, 2.6288195971604493e-17, 2.6288195971607362e-17, 2.6288195970933555e-17], [1.927521041021569, 2.4654554915810456e-17, 2.4654554915813476e-17, 2.465455491510301e-17], [1.932865594210216, 2.31533880750433e-17, 2.3153388075046475e-17, 2.3153388074298367e-17], [1.9382249665620148, 2.177103807247494e-17, 2.1771038072478268e-17, 2.1771038071691512e-17], [1.9435991991669486, 2.049562580241677e-17, 2.0495625802420262e-17, 2.0495625801593844e-17], [1.948988333228932, 1.931677504757361e-17, 1.9316775047577265e-17, 1.931677504671017e-17], [1.9543924100661287, 1.8225387064998376e-17, 1.82253870650022e-17, 1.8225387064093408e-17], [1.9598114711112677, 1.7213454947108997e-17, 1.7213454947112994e-17, 1.7213454946161486e-17], [1.9652455579119614, 1.6273909886118344e-17, 1.6273909886122513e-17, 1.6273909885127273e-17], [1.9706947121310228, 1.5400493216218487e-17, 1.5400493216222835e-17, 1.5400493215182858e-17], [1.9761589755467863, 1.4587649432879465e-17, 1.4587649432883995e-17, 1.4587649431798288e-17], [1.9816383900534287, 1.3830436403133292e-17, 1.3830436403138005e-17, 1.3830436402005604e-17], [1.9871329976612875, 1.3124449764235282e-17, 1.3124449764240185e-17, 1.3124449763060152e-17], [1.9926428404971857, 1.2465759118471515e-17, 1.2465759118476607e-17, 1.2465759117248055e-17], [1.998167960804754, 1.1850854111608152e-17, 1.1850854111613437e-17, 1.1850854110335527e-17], [2.003708400944754, 1.1276598863072545e-17, 1.1276598863078025e-17, 1.127659886174999e-17], [2.0092642033954045, 1.0740193521040772e-17, 1.074019352104645e-17, 1.0740193519667614e-17], [2.0148354107527044, 1.0239141963096673e-17, 1.0239141963102548e-17, 1.0239141961672345e-17], [2.020422065730764, 9.771224866755154e-18, 9.771224866761228e-18, 9.771224865279226e-18], [2.026024211162126, 9.334477544655644e-18, 9.334477544661916e-18, 9.334477543127852e-18], [2.0316418899981, 8.927172085171937e-18, 8.927172085178407e-18, 8.927172083592217e-18], [2.0372751453090894, 8.54780346764154e-18, 8.547803467648205e-18, 8.547803466010069e-18], [2.0429240202849193, 8.195079438500284e-18, 8.195079438507142e-18, 8.195079436817531e-18], [2.048588558235172, 7.867914046042431e-18, 7.867914046049477e-18, 7.867914044309214e-18], [2.0542688025895153, 7.565424843194902e-18, 7.565424843202131e-18, 7.565424841412451e-18], [2.059964796898038, 7.286933886375934e-18, 7.286933886383337e-18, 7.286933884545961e-18], [2.0656765848315835, 7.031972792917566e-18, 7.031972792925133e-18, 7.031972791042353e-18], [2.071404210182082, 6.800292281853434e-18, 6.800292281861151e-18, 6.800292279935922e-18], [2.07714771686289, 6.5918768311770294e-18, 6.591876831184881e-18, 6.591876829220915e-18], [2.0829071489091238, 6.406965366624081e-18, 6.406965366632048e-18, 6.40696536463392e-18], [2.088682550478, 6.2460792957283815e-18, 6.246079295736441e-18, 6.246079293709688e-18], [2.094473965849171, 6.11005978395941e-18, 6.110059783967535e-18, 6.110059781918757e-18], [2.1002814394250673, 6.000117043537639e-18, 6.000117043545798e-18, 6.000117041482736e-18], [2.106105015731236, 5.917895737755478e-18, 5.917895737763636e-18, 5.9178957356952286e-18], [2.111944739416683, 5.865562660172222e-18, 5.865562660180338e-18, 5.865562658116742e-18], [2.1178006552542152, 5.845926055352496e-18, 5.845926055360525e-18, 5.845926053313083e-18], [2.1236728081407845, 5.862601003937307e-18, 5.8626010039452014e-18, 5.862601001926348e-18], [2.1295612430978292, 5.920243376899868e-18, 5.920243376907576e-18, 5.920243374930672e-18], [2.1354660052716237, 6.024888032941462e-18, 6.024888032948931e-18, 6.0248880310280255e-18], [2.1413871399336206, 6.184448954583619e-18, 6.184448954590792e-18, 6.184448952740317e-18], [2.1473246924808, 6.4094770844569185e-18, 6.409477084463742e-18, 6.409477082698149e-18], [2.1532787084360168, 6.714340110084243e-18, 6.714340110090664e-18, 6.7143401084240375e-18], [2.1592492334483495, 7.119117515883071e-18, 7.11911751588904e-18, 7.11911751433471e-18], [2.165236313293451, 7.652760695199659e-18, 7.652760695205133e-18, 7.65276069377532e-18], [2.171239993873899, 8.35860960715257e-18, 8.358609607157511e-18, 8.358609605863045e-18], [2.1772603212195483, 9.304587046308338e-18, 9.304587046312713e-18, 9.304587045162827e-18], [2.1832973414878825, 1.0603438780115864e-17, 1.0603438780119649e-17, 1.0603438779121875e-17], [2.1893511009643696, 1.2456822685770796e-17, 1.2456822685773972e-17, 1.2456822684934129e-17], [2.1954216460628166, 1.5264040195912645e-17, 1.52640401959152e-17, 1.5264040195237463e-17], [2.2015090233257233, 1.9941766293286458e-17, 1.9941766293288384e-17, 1.9941766292775412e-17], [2.207613279424642, 2.9160912751673584e-17, 2.9160912751674884e-17, 2.9160912751328e-17], [2.2137344611605334, 5.538093720069875e-17, 5.538093720069943e-17, 5.5380937200518815e-17], [2.219872615464127, 6.445779388244643e-16, 6.445779388244644e-16, 6.44577938824449e-16], [2.226027789396278, 6.563504201442857e-17, 6.563504201442913e-17, 6.563504201428011e-17], [2.232200030148331, 3.109929949666063e-17, 3.1099299496661776e-17, 3.10992994963508e-17], [2.2383893850424825, 2.032620529302587e-17, 2.0326205293027598e-17, 2.0326205292557114e-17], [2.24459590153214, 1.5075978549581477e-17, 1.5075978549583776e-17, 1.5075978548956527e-17], [2.2508196272022887, 1.197138358786783e-17, 1.197138358787068e-17, 1.1971383587089588e-17], [2.2570606097698547, 9.92165141404759e-18, 9.921651414050979e-18, 9.921651413119046e-18], [2.263318897084073, 8.467731193749688e-18, 8.467731193753599e-18, 8.467731192673848e-18], [2.269594537126853, 7.382894352476305e-18, 7.382894352480724e-18, 7.382894351256152e-18], [2.2758875780131445, 6.542327866748694e-18, 6.542327866753605e-18, 6.5423278653871354e-18], [2.282198067991311, 5.8716930281901275e-18, 5.871693028195516e-18, 5.8716930266899875e-18], [2.288526055443496, 5.323979038827299e-18, 5.323979038833152e-18, 5.3239790371912924e-18], [2.2948715888859943, 4.868019699766396e-18, 4.8680196997727006e-18, 4.8680196979971226e-18], [2.301234716969626, 4.482348624791263e-18, 4.4823486247980046e-18, 4.482348622891204e-18], [2.3076154884801077, 4.1517061281978075e-18, 4.151706128204976e-18, 4.1517061261693256e-18], [2.3140139523384273, 3.864952683221282e-18, 3.864952683228863e-18, 3.864952681066624e-18], [2.3204301576012183, 3.613769968290026e-18, 3.613769968298008e-18, 3.613769966011329e-18], [2.3268641534611376, 3.391823378719823e-18, 3.391823378728197e-18, 3.3918233763191244e-18], [2.3333159892472404, 3.1942055676597955e-18, 3.1942055676685496e-18, 3.1942055651390348e-18], [2.3397857144253593, 3.017056925017265e-18, 3.0170569250263885e-18, 3.0170569223782965e-18], [2.3462733785984855, 2.857300735680672e-18, 2.857300735690155e-18, 2.8573007329252723e-18], [2.3527790315071453, 2.7124545844338957e-18, 2.7124545844437276e-18, 2.712454581563772e-18], [2.3593027230297854, 2.5804936131795296e-18, 2.580493613189701e-18, 2.5804936101963296e-18], [2.3658445031831525, 2.4597497588913784e-18, 2.4597497589018805e-18, 2.459749755796697e-18], [2.372404422122677, 2.348836414912167e-18, 2.34883641492299e-18, 2.3488364117075556e-18], [2.37898253014286, 2.2465913519327024e-18, 2.246591351943837e-18, 2.246591348619675e-18], [2.385578877677656, 2.1520329494879165e-18, 2.152032949499354e-18, 2.152032946067959e-18], [2.392193515300861, 2.0643262623319168e-18, 2.0643262623436476e-18, 2.064326258806492e-18], [2.3988264937265003, 1.9827564441318164e-18, 1.9827564441438323e-18, 1.982756440502372e-18], [2.405477863809216, 1.9067077380397887e-18, 1.906707738052081e-18, 1.9067077343077625e-18], [2.4121476765446594, 1.8356467238587724e-18, 1.8356467238713317e-18, 1.8356467200255967e-18], [2.4188359830698793, 1.7691088516926976e-18, 1.7691088517055154e-18, 1.7691088477598077e-18], [2.425542834663716, 1.7066875360566457e-18, 1.7066875360697134e-18, 1.7066875320254818e-18], [2.4322682827471938, 1.6480252616340744e-18, 1.6480252616473832e-18, 1.6480252575060874e-18], [2.4390123788839144, 1.5928062819443513e-18, 1.592806281957892e-18, 1.5928062777210066e-18], [2.4457751747804535, 1.5407505886402094e-18, 1.5407505886539735e-18, 1.5407505843229906e-18], [2.452556722286756, 1.49160890136611e-18, 1.4916089013800887e-18, 1.491608896956523e-18], [2.4593570733965358, 1.4451584826532479e-18, 1.445158482667432e-18, 1.4451584781528233e-18], [2.466176280247671, 1.401199623873359e-18, 1.4011996238877397e-18, 1.4011996192836563e-18], [2.4730143951226076, 1.359552680169803e-18, 1.3595526801843714e-18, 1.3595526754924113e-18], [2.479871470448757, 1.3200555569539512e-18, 1.320055556968698e-18, 1.3200555521904928e-18], [2.486747558798898, 1.2825615697692843e-18, 1.2825615697842002e-18, 1.2825615649214168e-18], [2.4936427128915843, 1.2469376143901868e-18, 1.246937614405263e-18, 1.2469376094596052e-18], [2.5005569855915426, 1.2130625959085309e-18, 1.2130625959237577e-18, 1.2130625908969682e-18], [2.5074904299100815, 1.1808260749940795e-18, 1.1808260750094477e-18, 1.1808260699033092e-18], [2.5144430990054993, 1.1501270970444948e-18, 1.150127097059995e-18, 1.1501270918763317e-18], [2.521415046183486, 1.1208731759829906e-18, 1.120873175998613e-18, 1.1208731707392914e-18], [2.5284063248975386, 1.0929794093350247e-18, 1.09297940935076e-18, 1.09297940401769e-18], [2.5354169887493656, 1.0663677051661953e-18, 1.066367705182034e-18, 1.0663676997771694e-18], [2.5424470914893007, 1.0409661046801014e-18, 1.0409661046960335e-18, 1.0409660992213727e-18], [2.5494966870167164, 1.0167081869066267e-18, 1.0167081869226423e-18, 1.0167081813802285e-18], [2.5565658293804328, 9.93532544072856e-19, 9.935325440889453e-19, 9.935325384808667e-19], [2.5636545727791344, 9.713823180319874e-19, 9.71382318048141e-19, 9.713823123765303e-19], [2.5707629715617872, 9.502047896026386e-19, 9.502047896188462e-19, 9.50204783885882e-19], [2.577891080228053, 9.299510138985377e-19, 9.299510139147894e-19, 9.299510081226944e-19], [2.585038953428708, 9.105754957531427e-19, 9.105754957694285e-19, 9.105754899204698e-19], [2.592206645966062, 8.920359002015454e-19, 8.920359002178557e-19, 8.920358943143444e-19], [2.599394212794379, 8.74292793702794e-19, 8.742927937191185e-19, 8.742927877634095e-19], [2.606601709020297, 8.573094123935547e-19, 8.573094124098835e-19, 8.573094064043739e-19], [2.6138291899032517, 8.410514541780395e-19, 8.410514541943624e-19, 8.410514481414915e-19], [2.6210767108559008, 8.254868918953989e-19, 8.254868919117062e-19, 8.254868858139542e-19], [2.6283443274445464, 8.105858051770607e-19, 8.105858051933423e-19, 8.105857990532298e-19], [2.635632095389565, 7.963202289236999e-19, 7.963202289399459e-19, 7.963202227600326e-19], [2.64294007056583, 7.8266401660317e-19, 7.826640166193706e-19, 7.826640104022543e-19], [2.6502683090031436, 7.695927168042197e-19, 7.695927168203648e-19, 7.695927105686802e-19], [2.6576168668866647, 7.570834616821355e-19, 7.570834616982155e-19, 7.570834554146332e-19], [2.6649858005573406, 7.451148661066874e-19, 7.451148661226925e-19, 7.451148598099174e-19], [2.6723751665123388, 7.336669364738929e-19, 7.336669364898136e-19, 7.336669301505841e-19], [2.679785021405479, 7.22720988275036e-19, 7.227209882908628e-19, 7.22720981927949e-19], [2.6872154220476685, 7.122595716316451e-19, 7.122595716473684e-19, 7.12259565263571e-19], [2.694666425407338, 7.022664041066538e-19, 7.022664041222643e-19, 7.022663977204131e-19], [2.702138088610878, 6.92726310191489e-19, 6.927263102069775e-19, 6.927263037899301e-19], [2.7096304689430766, 6.836251669486571e-19, 6.836251669640144e-19, 6.8362516053465435e-19], [2.717143623847559, 6.749498553605188e-19, 6.74949855375736e-19, 6.749498489369719e-19], [2.7246776109272277, 6.666882169993872e-19, 6.666882170144555e-19, 6.666882105692189e-19], [2.732232487944704, 6.588290156924487e-19, 6.588290157073594e-19, 6.588290092586038e-19], [2.7398083128227713, 6.513619039088244e-19, 6.513619039235689e-19, 6.51361897474268e-19], [2.7474051436448184, 6.442773936460235e-19, 6.442773936605937e-19, 6.442773872137396e-19], [2.755023038655287, 6.3756683164017895e-19, 6.375668316545666e-19, 6.37566825213169e-19], [2.7626620562601154, 6.312223787694128e-19, 6.3122237878361e-19, 6.312223723506938e-19], [2.770322255027187, 6.252369935634661e-19, 6.252369935774651e-19, 6.252369871560696e-19], [2.778003693686782, 6.196044197759744e-19, 6.196044197897675e-19, 6.196044133829444e-19], [2.7857064311320223, 6.143191780192213e-19, 6.143191780328014e-19, 6.143191716436133e-19], [2.7934305264193293, 6.093765615058143e-19, 6.093765615191742e-19, 6.093765551506934e-19], [2.801176038768873, 6.047726359882154e-19, 6.047726360013482e-19, 6.047726296566547e-19], [2.808943027565026, 6.005042440362737e-19, 6.00504244049173e-19, 6.005042377313536e-19], [2.8167315523568193, 5.965690138461768e-19, 5.965690138588362e-19, 5.965690075709822e-19], [2.8245416728583996, 5.929653728322544e-19, 5.929653728446677e-19, 5.929653665898742e-19], [2.832373448949487, 5.896925663175512e-19, 5.896925663297127e-19, 5.896925601110765e-19], [2.8402269406758336, 5.867506817113417e-19, 5.867506817232459e-19, 5.867506755438644e-19], [2.8481022082496827, 5.841406786436039e-19, 5.841406786552455e-19, 5.84140672518215e-19], [2.855999312050234, 5.818644256201685e-19, 5.818644256315427e-19, 5.818644195399572e-19], [2.863918312624101, 5.799247438701412e-19, 5.799247438812432e-19, 5.799247378381924e-19], [2.871859270685781, 5.783254591826676e-19, 5.783254591934931e-19, 5.783254532020617e-19], [2.8798222471181183, 5.770714626765331e-19, 5.770714626870783e-19, 5.7707145675034375e-19], [2.887807302972768, 5.761687816184346e-19, 5.761687816286955e-19, 5.761687757497275e-19], [2.895814499470671, 5.756246616092963e-19, 5.756246616192699e-19, 5.756246558011277e-19], [2.9038438980025143, 5.75447661699983e-19, 5.754476617096662e-19, 5.754476559553985e-19], [2.911895560129209, 5.756477642865234e-19, 5.756477642959136e-19, 5.756477586085562e-19], [2.9199695475823617, 5.762365019814308e-19, 5.762365019905256e-19, 5.762364963731004e-19], [2.9280659222647416, 5.772271040752601e-19, 5.7722710408405765e-19, 5.772270985395712e-19], [2.9361847462507638, 5.78634665708097e-19, 5.786346657165957e-19, 5.786346602480377e-19], [2.944326081786958, 5.804763434857641e-19, 5.804763434939629e-19, 5.804763381043044e-19], [2.9524899912924494, 5.82771582027402e-19, 5.8277158203529985e-19, 5.827715767274925e-19], [2.9606765373594364, 5.85542376854706e-19, 5.855423768623025e-19, 5.855423716392767e-19], [2.96888578275367, 5.888135801734852e-19, 5.888135801807805e-19, 5.888135750454438e-19], [2.9771177904149377, 5.926132575137446e-19, 5.92613257520739e-19, 5.926132524759752e-19], [2.985372623457542, 5.969731049609918e-19, 5.969731049676861e-19, 5.96973100016353e-19], [2.993650345170786, 6.019289389289179e-19, 6.019289389353132e-19, 6.019289340802419e-19], [3.0019510190194594, 6.075212732235498e-19, 6.075212732296477e-19, 6.075212684736407e-19], [3.0102747086443253, 6.137960017070375e-19, 6.137960017128399e-19, 6.137959970586696e-19], [3.0186214778626064, 6.208052094205719e-19, 6.208052094260814e-19, 6.208052048764883e-19], [3.026991390668475, 6.2860814088877755e-19, 6.286081408939968e-19, 6.286081364516887e-19], [3.035384511233545, 6.372723619364097e-19, 6.37272361941342e-19, 6.372723576089916e-19], [3.0438009039073624, 6.468751612993487e-19, 6.468751613039977e-19, 6.468751570842416e-19], [3.052240633217898, 6.575052514363767e-19, 6.575052514407466e-19, 6.575052473361834e-19], [3.060703763872045, 6.692648454113158e-19, 6.692648454154112e-19, 6.692648414285996e-19], [3.06919036075611, 6.82272210171712e-19, 6.822722101755378e-19, 6.822722063089956e-19], [3.077700488936319, 6.966648283765377e-19, 6.966648283800992e-19, 6.96664824636301e-19], [3.0862342136593073, 7.126033445768714e-19, 7.126033445801745e-19, 7.126033409615501e-19], [3.0947916003526226, 7.3027653212244205e-19, 7.302765321254933e-19, 7.302765286344258e-19], [3.1033727146252303, 7.499076022662979e-19, 7.499076022691039e-19, 7.499075989079282e-19], [3.111977622268012, 7.717622981284216e-19, 7.717622981309895e-19, 7.717622949019903e-19], [3.120606389254273, 7.961593913039779e-19, 7.961593913063154e-19, 7.961593882117253e-19], [3.129259081740245, 8.23484456021952e-19, 8.234844560240673e-19, 8.234844530660645e-19], [3.1379357660655964, 8.542081798739072e-19, 8.542081798758085e-19, 8.542081770565156e-19], [3.1466365087539385, 8.889110549794115e-19, 8.88911054981108e-19, 8.889110523025894e-19], [3.155361376513338, 9.28317202775566e-19, 9.283172027770669e-19, 9.283172002413267e-19], [3.1641104362368258, 9.733415325745986e-19, 9.73341532575914e-19, 9.73341530184894e-19], [3.1728837550029114, 1.025156796656292e-18, 1.0251567966574321e-18, 1.0251567944130096e-18], [3.1816814000760982, 1.0852910754835444e-18, 1.0852910754845197e-18, 1.0852910733885057e-18], [3.190503438907396, 1.1557731214726434e-18, 1.1557731214734652e-18, 1.1557731195276015e-18], [3.19934993913484, 1.23935541422824e-18, 1.23935541422892e-18, 1.239355412434877e-18], [3.208220968584011, 1.3398681438074852e-18, 1.339868143808035e-18, 1.3398681421674095e-18], [3.217116595268554, 1.4628035034374224e-18, 1.4628035034378547e-18, 1.4628035019521673e-18], [3.2260368873906975, 1.6163263710777328e-18, 1.6163263710780604e-18, 1.6163263697487533e-18], [3.2349819133417803, 1.813124661206653e-18, 1.813124661206889e-18, 1.8131246600353232e-18], [3.243951741702773, 2.0740440076438272e-18, 2.0740440076439855e-18, 2.074044006631438e-18], [3.2529464412448053, 2.435898633995364e-18, 2.435898633995458e-18, 2.4358986331431187e-18], [3.2619660809296915, 2.9703824154906375e-18, 2.9703824154906818e-18, 2.9703824147996527e-18], [3.2710107299104623, 3.838195747337749e-18, 3.838195747337758e-18, 3.838195746809048e-18], [3.2800804575318923, 5.489326605483972e-18, 5.489326605483961e-18, 5.489326605118483e-18], [3.2891753333310327, 9.846646939264654e-18, 9.846646939264639e-18, 9.846646939063207e-18], [3.2982954270377425, 5.3439562051538323e-17, 5.343956205153832e-17, 5.343956205150162e-17], [3.3074408085752287, 1.5059653827783358e-17, 1.5059653827783336e-17, 1.505965382765461e-17], [3.316611548060574, 6.505463350787216e-18, 6.50546335078715e-18, 6.50546335049255e-18], [3.3258077158052823, 4.111130646636862e-18, 4.111130646636736e-18, 4.111130646175864e-18], [3.335029382315814, 2.98544942064035e-18, 2.985449420640148e-18, 2.9854494200127236e-18], [3.344276618294126, 2.3318618517594904e-18, 2.331861851759195e-18, 2.331861850965056e-18], [3.353549494638213, 1.905243589990605e-18, 1.9052435899901987e-18, 1.905243589029303e-18], [3.3628480824426576, 1.6051281105718267e-18, 1.6051281105712916e-18, 1.6051281094437222e-18], [3.372172452999167, 1.3827160601455506e-18, 1.3827160601448696e-18, 1.382716058850837e-18], [3.3815226777971237, 1.21144558258705e-18, 1.2114455825862053e-18, 1.2114455811260521e-18], [3.3908988285241346, 1.0756212111698738e-18, 1.0756212111688478e-18, 1.0756212095430511e-18], [3.4003009770665757, 9.653728982899744e-19, 9.65372898288749e-19, 9.653728964979265e-19], [3.409729195510152, 8.741830257965086e-19, 8.74183025795066e-19, 8.741830238399778e-19], [3.41918355614044, 7.975746212057485e-19, 7.975746212040703e-19, 7.975746190856248e-19], [3.428664131443449, 7.323708246404922e-19, 7.323708246385608e-19, 7.323708223578171e-19], [3.4381709941061733, 6.762550193395091e-19, 6.762550193373064e-19, 6.762550168954797e-19], [3.447704217017154, 6.274982352439166e-19, 6.27498235241425e-19, 6.274982326398908e-19], [3.45726387326703, 5.847840375332007e-19, 5.847840375304023e-19, 5.847840347707006e-19], [3.466850036149107, 5.470925161687566e-19, 5.470925161656341e-19, 5.470925132494737e-19], [3.4764627791599123, 5.136213452070047e-19, 5.136213452035407e-19, 5.136213421328045e-19], [3.4861021759997635, 4.837307920574291e-19, 4.837307920536066e-19, 4.837307888303557e-19], [3.4957683005733315, 4.569046083056872e-19, 4.569046083014895e-19, 4.569046049279681e-19], [3.5054612269902066, 4.3272169761067997e-19, 4.3272169760609087e-19, 4.3272169408473153e-19], [3.5151810295654657, 4.108352489944986e-19, 4.108352489895021e-19, 4.108352453229304e-19], [3.524927782820246, 3.909571380808566e-19, 3.909571380754374e-19, 3.909571342664771e-19], [3.534701561482313, 3.7284610836232293e-19, 3.7284610835646617e-19, 3.7284610440814435e-19], [3.5445024404866334, 3.5629870638940784e-19, 3.5629870638309926e-19, 3.5629870229865194e-19], [3.554330494975953, 3.4114225138236004e-19, 3.4114225137558616e-19, 3.4114224715846335e-19], [3.564185800301367, 3.272293270445178e-19, 3.272293270372658e-19, 3.272293226911373e-19], [3.5740684320229046, 3.1443342582168162e-19, 3.144334258139396e-19, 3.144334213427003e-19], [3.5839784659101017, 3.0264547527327883e-19, 3.0264547526503565e-19, 3.0264547067281156e-19], [3.5939159779425878, 2.9177104658604697e-19, 2.9177104657729255e-19, 2.9177104186844656e-19], [3.6038810443106626, 2.817280957168408e-19, 2.817280957075662e-19, 2.817280908867039e-19], [3.613873741415887, 2.724451242759369e-19, 2.724451242661342e-19, 2.7244511933811e-19], [3.623894145871661, 2.6385967415611e-19, 2.6385967414577263e-19, 2.638596691156961e-19], [3.6339423345038186, 2.5591708987978393e-19, 2.559170898689066e-19, 2.5591708474214847e-19], [3.6440183843512104, 2.485694976211817e-19, 2.485694976097607e-19, 2.485694923919594e-19], [3.6541223726662997, 2.417749612267982e-19, 2.417749612148312e-19, 2.417749559118994e-19], [3.6642543769157525, 2.354967842760666e-19, 2.35496784263553e-19, 2.3549677888168424e-19], [3.6744144747810292, 2.297029339930705e-19, 2.2970293398001152e-19, 2.2970292852568687e-19], [3.684602744158982, 2.2436556814937267e-19, 2.2436556813577127e-19, 2.243655626157663e-19], [3.694819263162455, 2.1946065036590824e-19, 2.1946065035176965e-19, 2.194606447731612e-19], [3.705064110120879, 2.1496764271475832e-19, 2.1496764270008967e-19, 2.1496763707026336e-19], [3.7153373635808737, 2.1086926746285592e-19, 2.1086926744766673e-19, 2.1086926177432377e-19], [3.7256391023068485, 2.071513323703094e-19, 2.0715133235461164e-19, 2.0715132664577638e-19], [3.735969405281611, 2.0380261631134366e-19, 2.0380261629515181e-19, 2.0380261055917921e-19], [3.746328351706965, 2.0081481426965567e-19, 2.0081481425298694e-19, 2.008148084985702e-19], [3.7567160210043236, 1.9818254311706772e-19, 1.981825430999422e-19, 1.981825373361204e-19], [3.7671324928153185, 1.9590341217570802e-19, 1.9590341215814888e-19, 1.9590340639431487e-19], [3.7775778470024037, 1.9397816558268795e-19, 1.9397816556472145e-19, 1.939781598106299e-19], [3.7880521636494784, 1.9241090717118917e-19, 1.924109071528451e-19, 1.924109014186205e-19], [3.7985555230624906, 1.9120942328928394e-19, 1.9120942327059556e-19, 1.9120941756674061e-19], [3.80908800577006, 1.9038562516891518e-19, 1.9038562514991956e-19, 1.9038561948732343e-19], [3.819649692524091, 1.8995614081129384e-19, 1.89956140792032e-19, 1.8995613518197888e-19], [3.8302406643003954, 1.8994309787193434e-19, 1.8994309785245142e-19, 1.8994309230662928e-19], [3.8408610022993117, 1.9037515521185672e-19, 1.903751551922023e-19, 1.9037514972271147e-19], [3.8515107879463257, 1.9128886393329992e-19, 1.912888639135281e-19, 1.9128885853289035e-19], [3.862190102892696, 1.9273047243231177e-19, 1.9273047241248158e-19, 1.9273046713364916e-19], [3.872899029016083, 1.947583399998914e-19, 1.947583399800669e-19, 1.9475833481643164e-19], [3.8836376484211743, 1.974461991003588e-19, 1.9744619908060933e-19, 1.974461940460121e-19], [3.894406043440311, 2.008876231593578e-19, 2.0088762313975828e-19, 2.0088761824849848e-19], [3.905204296634124, 2.0520224100220462e-19, 2.0520224098283585e-19, 2.0520223624968092e-19], [3.9160324907921673, 2.1054453756404568e-19, 2.1054453754499452e-19, 2.1054453298518977e-19], [3.9268907089335445, 2.1711657759741047e-19, 2.171165775787702e-19, 2.171165732080487e-19], [3.9377790343075567, 2.251868434711344e-19, 2.25186843453005e-19, 2.2518683928759777e-19], [3.9486975503943342, 2.3511889989246714e-19, 2.3511889987495557e-19, 2.351188959316016e-19], [3.9596463409054747, 2.4741642181733947e-19, 2.4741642180056004e-19, 2.4741641809651663e-19], [3.9706254897846924, 2.627966141236448e-19, 2.627966141077195e-19, 2.62796610660773e-19], [3.9816350812084536, 2.8231534490661776e-19, 2.8231534489167635e-19, 2.823153417201526e-19], [3.9926751995866288, 3.0759212147392093e-19, 3.0759212146010176e-19, 3.075921185828768e-19], [4.0037459295631335, 3.412420826945425e-19, 3.4124208268199244e-19, 3.412420801185037e-19], [4.014847356016583, 3.8777747503492647e-19, 3.877774750238016e-19, 3.8777747279405873e-19], [4.025979564060941, 4.557055096703955e-19, 4.557055096608612e-19, 4.557055077854572e-19], [4.037142639046169, 5.632010038536288e-19, 5.632010038458604e-19, 5.632010023459829e-19], [4.048336666558885, 7.57279164380858e-19, 7.572791643750413e-19, 7.572791632724827e-19], [4.0595617324230195, 1.208589533287596e-18, 1.2085895332839273e-18, 1.2085895326010923e-18], [4.0708179227004715, 3.396920149819104e-18, 3.39692014981779e-18, 3.3969201495776614e-18], [4.082105323691766, 3.558128441803245e-18, 3.558128441801983e-18, 3.5581284415753924e-18], [4.093424021936723, 1.1109256378731482e-18, 1.1109256378690807e-18, 1.1109256371517655e-18], [4.104774104215116, 6.387343182925126e-19, 6.387343182853955e-19, 6.387343170522769e-19], [4.116155657547332, 4.386804406407861e-19, 4.386804406303627e-19, 4.386804388557433e-19], [4.1275687691950536, 3.284732296917604e-19, 3.284732296777601e-19, 3.28473227335257e-19], [4.1390135266619135, 2.5889949290406575e-19, 2.5889949288620386e-19, 2.5889948994873087e-19], [4.150490017694171, 2.1112915373450823e-19, 2.111291537124854e-19, 2.111291501522391e-19], [4.1619983302813885, 1.764035382062883e-19, 1.7640353817978983e-19, 1.764035339682359e-19], [4.173538552657098, 1.5009644108935901e-19, 1.5009644105805442e-19, 1.5009643616591423e-19], [4.185110773299484, 1.295343289214556e-19, 1.2953432888499782e-19, 1.2953432328223484e-19], [4.1967150809320595, 1.1306461828264383e-19, 1.1306461824066867e-19, 1.1306461189647483e-19], [4.208351564524347, 9.961095401402088e-20, 9.961095396614639e-20, 9.961094684892811e-20], [4.220020313292561, 8.844231209831274e-20, 8.844231204413847e-20, 8.844230412150276e-20], [4.231721416700288, 7.904488360488217e-20, 7.904488354398853e-20, 7.904487478272876e-20], [4.243454964459176, 7.104704969659588e-20, 7.104704962854336e-20, 7.1047039994625e-20], [4.255221046529626, 6.417343869415677e-20, 6.417343861848528e-20, 6.41734280770313e-20], [4.267019753121475, 5.82157116613273e-20, 5.821571157755533e-20, 5.821570009283164e-20], [4.278851174694687, 5.301337705164718e-20, 5.3013376959271044e-20, 5.3013364494671625e-20], [4.290715401960056, 4.8440835297411e-20, 4.8440835195904025e-20, 4.844082171393594e-20], [4.302612525879891, 4.4398414509120347e-20, 4.4398414397932024e-20, 4.439839986020026e-20], [4.314542637668716, 4.0806033038378486e-20, 4.0806032916933605e-20, 4.0806017284125766e-20], [4.326505828793978, 3.7598632780645764e-20, 3.759863264834354e-20, 3.759861588021434e-20], [4.338502190976736, 3.4722831779246703e-20, 3.472283163545982e-20, 3.472281369081546e-20], [4.350531816192368, 3.213443258513779e-20, 3.2134432429211494e-20, 3.213441326589384e-20], [4.362594796671283, 2.979654163900641e-20, 2.9796541470257516e-20, 2.9796521045128086e-20], [4.3746912248996175, 2.7678131786109867e-20, 2.7678131603825755e-20, 2.7678109872749633e-20], [4.386821193619955, 2.5752930759448653e-20, 2.575293056288623e-20, 2.5752907480715732e-20], [4.398984795832026, 2.399855257834854e-20, 2.3998552366733184e-20, 2.3998527887291407e-20], [4.411182124793432, 2.2395812140914143e-20, 2.2395811913438588e-20, 2.239578598950278e-20], [4.423413274020354, 2.0928179497042696e-20, 2.0928179252865925e-20, 2.092815183615071e-20], [4.43567833728827, 1.9581341710762556e-20, 1.958134144900861e-20, 1.958131249014907e-20], [4.447977408632674, 1.8342848377059623e-20, 1.8342848096816426e-20, 1.8342817545350982e-20], [4.460310582349801, 1.720182275453591e-20, 1.7201822454854018e-20, 1.720179025920722e-20], [4.472677952997345, 1.6148724786326236e-20, 1.6148724466217576e-20, 1.6148690573682692e-20], [4.4850796153951835, 1.517515546741591e-20, 1.517515512585247e-20, 1.5175119482573958e-20], [4.497515664626113, 1.4273694394055237e-20, 1.427369402996772e-20, 1.4273656580923546e-20], [4.509986196036565, 1.3437764121896788e-20, 1.3437763734173232e-20, 1.3437724423157035e-20], [4.522491305237347, 1.2661516320304057e-20, 1.2661515907788401e-20, 1.2661474677391554e-20], [4.535031088104372, 1.1939735752797182e-20, 1.1939735314287813e-20, 1.1939692105881309e-20], [4.547605640779396, 1.1267758918416247e-20, 1.126775845266449e-20, 1.1267713206380718e-20], [4.560215059670747, 1.0641404814640071e-20, 1.0641404320348628e-20, 1.0641356975063033e-20], [4.572859441454077, 1.005691577252997e-20, 1.0056915248351339e-20, 1.0056865741663934e-20], [4.5855388830730925, 9.510906700965133e-21, 9.510906145499968e-21, 9.510854413716751e-21], [4.5982534817403025, 9.000321383064839e-21, 9.00032079486026e-21, 9.000266772974511e-21], [4.611003334937763, 8.522394712147871e-21, 8.522394089695735e-21, 8.522337711369215e-21], [4.623788540417824, 8.074619950483422e-21, 8.07461929221854e-21, 8.074560489762587e-21], [4.636609196203881, 7.654720252040142e-21, 7.654719556338454e-21, 7.654658260694983e-21], [4.649465400591122, 7.26062381844642e-21, 7.26062308362311e-21, 7.260559224345545e-21], [4.6623572521472845, 6.890442161599564e-21, 6.890441385907133e-21, 6.890374891141247e-21], [4.675284849713414, 6.5424510316238226e-21, 6.542450213250101e-21, 6.5423810097148e-21], [4.688248292404616, 6.215073638928127e-21, 6.215072775994237e-21, 6.215000788962288e-21], [4.701247679610816, 5.906865856907778e-21, 5.906864947466042e-21, 5.906790100744744e-21], [4.714283110997529, 5.61650313971344e-21, 5.616502181745229e-21, 5.6164243976570606e-21], [4.727354686506616, 5.34276892932324e-21, 5.342767920736766e-21, 5.342687120100009e-21], [4.7404625063570505, 5.084544359383733e-21, 5.084543298011788e-21, 5.084459400121126e-21], [4.753606671045692, 4.840799091120134e-21, 4.840797974717758e-21, 4.840710897324901e-21], [4.766787281348056, 4.6105831400107444e-21, 4.610581966252845e-21, 4.610491625547154e-21], [4.780004438319077, 4.393019571646233e-21, 4.393018338125132e-21, 4.392924648714283e-21], [4.793258243293898, 4.187297961881904e-21, 4.187296666104829e-21, 4.187199540995535e-21], [4.806548797888636, 3.992668530549504e-21, 3.992667169936011e-21, 3.992566520514804e-21], [4.819876204001168, 3.8084368700423696e-21, 3.808435441921702e-21, 3.808331177935807e-21], [4.833240563811904, 3.633959200370276e-21, 3.6339577019786526e-21, 3.633849731516976e-21], [4.84664197978458, 3.4686380910765786e-21, 3.468636519554419e-21, 3.468524749028672e-21], [4.860080554667041, 3.311918597959268e-21, 3.311916950348349e-21, 3.3118012844743118e-21], [4.873556391492023, 3.163284769028923e-21, 3.1632830422694565e-21, 3.1631633840484452e-21], [4.887069593577949, 3.0222564797339274e-21, 3.022254670661578e-21, 3.0221309213621106e-21], [4.900620264529717, 2.888386562322164e-21, 2.8883846676649904e-21, 2.8882567268046976e-21], [4.914208508239503, 2.7612581983978103e-21, 2.761256214773129e-21, 2.7611239801009367e-21], [4.927834428887543, 2.640482547372102e-21, 2.6404804712832823e-21, 2.640343838761899e-21], [4.941498130942943, 2.5256965866709406e-21, 2.5256944145041234e-21, 2.5255532782928666e-21], [4.955199719164479, 2.4165611423234517e-21, 2.41655887034419e-21, 2.416413122782205e-21], [4.968939298601393, 2.312759090966622e-21, 2.31275671531645e-21, 2.3126062469063346e-21], [4.982716974594211, 2.2139937164114296e-21, 2.21399123310435e-21, 2.2138359324962596e-21], [4.996532852775535, 2.1199872057681195e-21, 2.1199846106870153e-21, 2.119824364663271e-21], [5.010387039070864, 2.0304792717535223e-21, 2.0304765606464865e-21, 2.030311254106762e-21], [5.0242796396994045, 1.9452258892358807e-21, 1.9452230577124785e-21, 1.9450525736595823e-21], [5.038210761174881, 1.863998135334968e-21, 1.863995178862406e-21, 1.86381939838876e-21], [5.052180510306356, 1.7865811235096197e-21, 1.786578037408844e-21, 1.786396839683687e-21], [5.06618899419905, 1.7127730230520707e-21, 1.7127698024937865e-21, 1.7125830647511798e-21], [5.080236320255153, 1.642384156281948e-21, 1.6423807962825604e-21, 1.6421883938102586e-21], [5.094322596174668, 1.5752361665087062e-21, 1.5752326619261821e-21, 1.5750344680554567e-21], [5.108447929956212, 1.5111612505209284e-21, 1.5111575960505787e-21, 1.5109534821470655e-21], [5.122612429897865, 1.4500014499743735e-21, 1.4499976401445646e-21, 1.4497874756001921e-21], [5.136816204597985, 1.3916079965979208e-21, 1.3916040257657024e-21, 1.3913876779918508e-21], [5.151059362956053, 1.335840706624997e-21, 1.3358365689716636e-21, 1.3356139033935852e-21], [5.165342014173498, 1.282567420294481e-21, 1.2825631098210583e-21, 1.2823339898736832e-21], [5.1796642677545375, 1.2316634826559177e-21, 1.231658993178577e-21, 1.2314232803038348e-21], [5.1940262335070235, 1.1830112622640038e-21, 1.1830065874094093e-21, 1.1827641410551456e-21], [5.208428021543272, 1.1364997046616738e-21, 1.1364948378622596e-21, 1.1362455154828772e-21], [5.222869742280922, 1.092023917833282e-21, 1.0920188523224827e-21, 1.0917625093814239e-21], [5.237351506443766, 1.0494847870635647e-21, 1.0494795158709662e-21, 1.0492160058452135e-21], [5.251873425062612, 1.0087886168665367e-21, 1.0087831328129794e-21, 1.008512307199657e-21], [5.266435609476128, 9.698467978550052e-22, 9.698410935476472e-22, 9.6956280187291e-22], [5.281038171331696, 9.32575496607276e-22, 9.325695644345932e-22, 9.322836542649625e-22], [5.295681222586275, 8.968953667559553e-22, 8.968891988827151e-22, 8.965955158389955e-22], [5.310364875507244, 8.627312796759709e-22, 8.62724868038165e-22, 8.624232558121835e-22], [5.325089242673285, 8.300120732864459e-22, 8.300054095861936e-22, 8.296957099604715e-22], [5.339854436975225, 7.98670317606434e-22, 7.986633933068797e-22, 7.983454461774818e-22], [5.3546605716169084, 7.686420958172426e-22, 7.686349021374224e-22, 7.683085455401862e-22], [5.3695077601160754, 7.398667996875274e-22, 7.398593275972496e-22, 7.395243977376073e-22], [5.384396116305215, 7.122869383105589e-22, 7.122791785252841e-22, 7.119355098118864e-22], [5.3993257543324535, 6.85847959188391e-22, 6.858399021641308e-22, 6.854873272465619e-22], [5.414296788662417, 6.604980807750724e-22, 6.604897167033253e-22, 6.601280665140965e-22], [5.429309334077116, 6.361881356618383e-22, 6.361794544645605e-22, 6.358085582657092e-22], [5.444363505676828, 6.128714236516147e-22, 6.128624149762587e-22, 6.124821004107876e-22], [5.459459418880967, 5.9050357402915705e-22, 5.904942272437685e-22, 5.901043203922294e-22], [5.474597189428989, 5.690424163868631e-22, 5.690327205752794e-22, 5.686330460177946e-22], [5.489776933381259, 5.484478594156075e-22, 5.484378033727618e-22, 5.480281842568155e-22], [5.504998767119951, 5.286817771149742e-22, 5.286713493423169e-22, 5.282516074566015e-22], [5.520262807349943, 5.097079019185632e-22, 5.096970906196462e-22, 5.092670464743577e-22], [5.535569171099701, 4.914917242678971e-22, 4.914805173440968e-22, 4.910399902579894e-22], [5.55091797572219, 4.74000398203137e-22, 4.73988783249604e-22, 4.735375914441678e-22], [5.56630933889576, 4.572026525708285e-22, 4.571906168726269e-22, 4.567285775737938e-22], [5.5817433786250605, 4.410687074780241e-22, 4.410562380065414e-22, 4.4058316755422395e-22], [5.597220213241938, 4.2557019564927475e-22, 4.255572790588704e-22, 4.250729930248701e-22], [5.612739961406338, 4.106800883675702e-22, 4.1066671099255384e-22, 4.10171024307066e-22], [5.628302742107234, 3.9637262570331456e-22, 3.9635877355524683e-22, 3.958515006425232e-22], [5.643908674663516, 3.8262325075637424e-22, 3.826089095217368e-22, 3.820898644452563e-22], [5.659557878724924, 3.694085476555953e-22, 3.6939370269386815e-22, 3.6886269931140284e-22], [5.675250474272955, 3.5670618307822657e-22, 3.5669081942041424e-22, 3.561476715494725e-22], [5.690986581621786, 3.4449485106806514e-22, 3.444789534156837e-22, 3.4392347500977654e-22], [5.706766321419195, 3.3275422094643665e-22, 3.327377736710467e-22, 3.321697790071114e-22], [5.72258981464749, 3.2146488812434674e-22, 3.2144787526763347e-22, 3.208671791451387e-22], [5.738457182624435, 3.1060832763702266e-22, 3.105907329115156e-22, 3.0999715086356033e-22], [5.754368547004175, 3.0016685023429833e-22, 3.001486570247189e-22, 2.9954200554162115e-22], [5.770324029778173, 2.9012356087140696e-22, 2.9010475223674118e-22, 2.89484849002457e-22], [5.786323753276149, 2.8046231945509693e-22, 2.804428781314139e-22, 2.7980954227314653e-22], [5.80236784016701, 2.711677037096929e-22, 2.7114761211380247e-22, 2.705006644651802e-22], [5.818456413459799, 2.6222497403658563e-22, 2.622042142705251e-22, 2.615434776486088e-22], [5.834589596504625, 2.5362004024891957e-22, 2.5359859410535314e-22, 2.529238936018578e-22], [5.850767512993629, 2.4533943007087965e-22, 2.4531727903951925e-22, 2.446284423263468e-22], [5.866990286961913, 2.373702592984002e-22, 2.3734738457340276e-22, 2.366442422228579e-22], [5.883258042788496, 2.2970020352433235e-22, 2.2967658601278675e-22, 2.2895897183242403e-22], [5.899570905197279, 2.223174713376896e-22, 2.2229309166925324e-22, 2.2156084305160733e-22], [5.915928999257983, 2.1521077891216686e-22, 2.1518561744990954e-22, 2.1443857573697374e-22], [5.932332450387127, 2.0836932590455472e-22, 2.0834336275700551e-22, 2.0758137361955783e-22], [5.9487813843489725, 2.0178277258856564e-22, 2.0175598762309362e-22, 2.0097890145449822e-22], [5.965275927256495, 1.954412181545066e-22, 1.954135910120358e-22, 1.9462126333649934e-22], [5.981816205572358, 1.8933518010929116e-22, 1.8930669022038746e-22, 1.8849898211516942e-22], [5.998402346109865, 1.8345557471552489e-22, 1.834262013179702e-22, 1.826029798491215e-22], [6.015034476033955, 1.777936984122896e-22, 1.7776342057007678e-22, 1.769245592411332e-22], [6.0317127228621565, 1.7234121016349143e-22, 1.7231000678736936e-22, 1.714553860002483e-22], [6.048437214465572, 1.67090114683343e-22, 1.6705796455287388e-22, 1.6618747208021403e-22], [6.065208079069868, 1.620327464913263e-22, 1.6199962827852412e-22, 1.6111315974622851e-22], [6.082025445256241, 1.5716175475215594e-22, 1.5712764704668758e-22, 1.5622510642581196e-22], [6.0988894419624184, 1.5247008885883707e-22, 1.5243497019478452e-22, 1.5151627030122897e-22], [6.115800198483637, 1.4795098471944634e-22, 1.4791483360367827e-22, 1.4697989660416808e-22], [6.132757844473645, 1.4359795171086585e-22, 1.4356074665294032e-22, 1.4260950457581195e-22], [6.14976250994568, 1.3940476026459464e-22, 1.393664798082271e-22, 1.3839887505690706e-22], [6.16681432527348, 1.3536543005231092e-22, 1.3532605280828278e-22, 1.3434203867563444e-22], [6.18391342119228, 1.314742187403067e-22, 1.3143372342092183e-22, 1.3043326460203773e-22], [6.2010599287998085, 1.27725611284371e-22, 1.2768397673915909e-22, 1.266670498405674e-22], [6.2182539795573, 1.2411430973772697e-22, 1.2407151499056663e-22, 1.230381090326649e-22], [6.235495705290497, 1.2063522354697505e-22, 1.2059124783435983e-22, 1.1954136474497173e-22], [6.252785238190661, 1.1728346031190085e-22, 1.1723828312234597e-22, 1.1617193821805737e-22], [6.270122710815596, 1.140543169868383e-22, 1.140079181012136e-22, 1.1292514055368852e-22], [6.287508256090646, 1.109432715023352e-22, 1.1089563103507946e-22, 1.0979646431902455e-22], [6.304942007309737, 1.0794597478729749e-22, 1.0789707322829078e-22, 1.0678157554782255e-22], [6.322424098136379, 1.0505824317284274e-22, 1.0500806142975888e-22, 1.0387630611994575e-22], [6.339954662604698, 1.0227605116004898e-22, 1.0222457060106286e-22, 1.0107664650090806e-22], [6.3575338351204715, 9.959552453509625e-23, 9.954272703169984e-23, 9.837873882536414e-23], [6.375161750462144, 9.701293381590592e-23, 9.695880178560037e-23, 9.577887030835346e-23], [6.392838543781877, 9.45246880153117e-23, 9.446920446401531e-23, 9.327346696933248e-23], [6.410564350606563, 9.212732870670998e-23, 9.207047727054428e-23, 9.085908765521373e-23], [6.428339306838899, 8.981752437865776e-23, 8.975928936499095e-23, 8.853241834879968e-23], [6.446163548758387, 8.759206506569097e-23, 8.753243149305289e-23, 8.629026675022845e-23], [6.464037213022405, 8.54478572430971e-23, 8.538681087983512e-23, 8.412955711895636e-23], [6.48196043666726, 8.338191897411643e-23, 8.331944637543963e-23, 8.204732536576235e-23], [6.499933357109212, 8.139137529825776e-23, 8.132746384145072e-23, 8.004071438264094e-23], [6.517956112145557, 7.94734538501578e-23, 7.940809176772988e-23, 7.81069696015919e-23], [6.53602883995566, 7.762548069856712e-23, 7.755865710904552e-23, 7.624343477117475e-23], [6.554151679102025, 7.584487639555866e-23, 7.577658133171812e-23, 7.444754794211416e-23], [6.572324768531367, 7.412915222626073e-23, 7.405937666056208e-23, 7.271683765212902e-23], [6.59054824757565, 7.247590664992466e-23, 7.240464251694032e-23, 7.104891930195905e-23], [6.608822255953189, 7.088282192305953e-23, 7.081006213873899e-23, 6.944149171294043e-23], [6.627146933769696, 6.934766089589725e-23, 6.927339937345966e-23, 6.789233385869919e-23], [6.645522421519358, 6.786826397372392e-23, 6.779249563600858e-23, 6.639930176287369e-23], [6.663948860085928, 6.644254623429226e-23, 6.636526702258077e-23, 6.496032555407858e-23], [6.682426390743791, 6.506849469365474e-23, 6.498970157273037e-23, 6.35734066719904e-23], [6.700955155159052, 6.374416571190488e-23, 6.366385667134332e-23, 6.223661521532827e-23], [6.719535295390622, 6.246768253140214e-23, 6.238585658307854e-23, 6.094808742632514e-23], [6.7381669538913105, 6.123723293961304e-23, 6.115389011143256e-23, 5.97060233026546e-23], [6.7568502735089115, 6.005106704929204e-23, 5.996620837522741e-23, 5.850868433141703e-23], [6.775585397487297, 5.890749518880737e-23, 5.882112269508777e-23, 5.735439133768155e-23], [6.7943724694675325, 5.780488589547214e-23, 5.771700258332496e-23, 5.624152244065145e-23], [6.813211633488953, 5.674166400520379e-23, 5.665227383013503e-23, 5.516851111182282e-23], [6.832103033990289, 5.571630883209488e-23, 5.562541667975058e-23, 5.413384432825856e-23], [6.851046815810759, 5.472735243128606e-23, 5.463496409029251e-23, 5.313606081467801e-23], [6.870043124191186, 5.37733779394666e-23, 5.367950007131929e-23, 5.2173749369380996e-23], [6.889092104775117, 5.285301798706343e-23, 5.2757658093315283e-23, 5.1245547267292724e-23], [6.9081939036099245, 5.196495317668276e-23, 5.18681195636407e-23, 5.035013873522945e-23], [6.927348667147947, 5.11079106226739e-23, 5.100961236396985e-23, 4.948625349423102e-23], [6.946556542247589, 5.028066254692975e-23, 5.01809094440678e-23, 4.8652665363396025e-23], [6.965817676174463, 4.948202492639315e-23, 4.9380827467761085e-23, 4.7848190921047524e-23], [6.985132216602515, 4.8710856188036476e-23, 4.860822550643068e-23, 4.7071688218625275e-23], [7.004500311615151, 4.796605594742829e-23, 4.7862003776548844e-23, 4.632205554277731e-23], [7.023922109706382, 4.724656378726747e-23, 4.7141102417360556e-23, 4.5598230222197315e-23], [7.0433977597819535, 4.655135807258553e-23, 4.6444500305643024e-23, 4.489918747526703e-23], [7.062927411160493, 4.587945479978808e-23, 4.5771213904340134e-23, 4.422393929501808e-23], [7.0825112135746515, 4.5229906476698265e-23, 4.5120296142710145e-23, 4.357153336880697e-23], [7.10214931717225, 4.4601801031352403e-23, 4.449083532531869e-23, 4.294105202923333e-23], [7.121841872517442, 4.3994260747493325e-23, 4.388195406793962e-23, 4.2331611234695086e-23], [7.14158903059185, 4.340644122476788e-23, 4.329280825859926e-23, 4.174235957583122e-23], [7.161390942795738, 4.2837530362340277e-23, 4.272258604187746e-23, 4.1172477307656806e-23], [7.181247760949164, 4.228674736430651e-23, 4.2170506825545495e-23, 4.062117540426304e-23], [7.201159637293144, 4.175334176602687e-23, 4.163582030795493e-23, 4.008769463499965e-23], [7.22112672449083, 4.1236592480224826e-23, 4.1117805525700246e-23, 3.9571304660676514e-23], [7.24114917562866, 4.07358068624518e-23, 4.061576992046192e-23, 3.9071303149342926e-23], [7.26122714421756, 4.0250319794954765e-23, 4.012904842442854e-23, 3.8587014909400747e-23], [7.281360784194092, 3.977949278885256e-23, 3.965700256423966e-23, 3.811779104087151e-23], [7.301550249921649, 3.9322713103970786e-23, 3.9199019582562645e-23, 3.766300810251253e-23], [7.321795696191643, 3.887939288660619e-23, 3.875451157758288e-23, 3.722206729623646e-23], [7.342097278224678, 3.8448968324445476e-23, 3.832291465984131e-23, 3.679439366634733e-23], [7.362455151671753, 3.8030898819332955e-23, 3.790368812680262e-23, 3.6379435315585724e-23], [7.382869472615446, 3.7624666177229224e-23, 3.749631365468234e-23, 3.5976662635666403e-23], [7.403340397571117, 3.722977381588287e-23, 3.7100294507977095e-23, 3.5585567554135726e-23], [7.423868083488107, 3.684574599010188e-23, 3.6715154766506113e-23, 3.520566279613541e-23], [7.44445268775093, 3.6472127034645884e-23, 3.6340438570193486e-23, 3.4836481161775525e-23], [7.465094368180506, 3.610848062516332e-23, 3.597570938171779e-23, 3.4477574819246535e-23], [7.48579328303534, 3.575438905685292e-23, 3.5620549266835926e-23, 3.412851461316591e-23], [7.506549591012758, 3.5409452541436876e-23, 3.5274558193051486e-23, 3.378888938919379e-23], [7.527363451250114, 3.507328852201577e-23, 3.4937353346089975e-23, 3.3458305333875884e-23], [7.548235023326006, 3.4745531006415937e-23, 3.4608568464686836e-23, 3.313638533111221e-23], [7.569164467261515, 3.4425829918606893e-23, 3.4287853193559713e-23, 3.2822768333957306e-23], [7.59015194352141, 3.411385046856404e-23, 3.3974872454550787e-23, 3.251710875331833e-23], [7.611197613015405, 3.380927254029988e-23, 3.366930583607686e-23, 3.221907586226111e-23], [7.632301637099369, 3.3511790098268306e-23, 3.337084700069038e-23, 3.1928353216884966e-23], [7.653464177576567, 3.3221110611888826e-23, 3.3079203110749534e-23, 3.1644638093523535e-23], [7.674685396698919, 3.293695449817917e-23, 3.279409427219235e-23, 3.136764094205426e-23], [7.695965457168219, 3.2659054582291057e-23, 3.251525299606284e-23, 3.1097084855476626e-23], [7.717304522137401, 3.23871555759319e-23, 3.2242423677865047e-23, 3.083270505586874e-23], [7.738702755211778, 3.2121013573339105e-23, 3.197536209436283e-23, 3.057424839634671e-23], [7.760160320450308, 3.1860395564636054e-23, 3.171383491777327e-23, 3.0321472879056605e-23], [7.781677382366838, 3.1605078966291956e-23, 3.145761924691172e-23, 3.007414718888945e-23], [7.803254105931371, 3.135485116843893e-23, 3.1206502155270514e-23, 2.983205024284069e-23], [7.824890656571345, 3.110950909888433e-23, 3.096028025547368e-23, 2.959497075531536e-23], [7.846587200172873, 3.086885880315994e-23, 3.071875927999689e-23, 2.9362706818144293e-23], [7.868343903082044, 3.06327150407015e-23, 3.048175367774182e-23, 2.913506549589951e-23], [7.890160932106173, 3.040090089642233e-23, 3.024908622612337e-23, 2.891186243598885e-23], [7.9120384545150975, 3.017324740760733e-23, 3.0020587658232093e-23, 2.869292149318567e-23], [7.933976638042457, 2.99495932055589e-23, 2.979609630496627e-23, 2.847807436823985e-23], [7.955975650886964, 2.9729784171787657e-23, 2.957545775133967e-23, 2.826716026066572e-23], [7.978035661713726, 2.951367310812893e-23, 2.935852450698533e-23, 2.806002553476545e-23], [8.000156839655501, 2.93011194206614e-23, 2.9145155690183396e-23, 2.7856523399228375e-23], [8.022339354314013, 2.9091988816717086e-23, 2.8935216725101547e-23, 2.765651359926948e-23], [8.044583375761261, 2.888615301488798e-23, 2.87285790519592e-23, 2.7459862121600985e-23], [8.066889074540805, 2.86834894673672e-23, 2.852511984942002e-23, 2.726644091130113e-23], [8.089256621669083, 2.848388109445542e-23, 2.832472176918782e-23, 2.7076127600970897e-23], [8.11168618863672, 2.8287216030538717e-23, 2.812727268211603e-23, 2.6888805250886803e-23], [8.134177947409846, 2.809338738145986e-23, 2.7932665435529416e-23, 2.670436210056794e-23], [8.156732070431408, 2.79022929926434e-23, 2.7740797621395363e-23, 2.652269133101926e-23], [8.179348730622497, 2.771383522772137e-23, 2.7551571354976955e-23, 2.634369083739486e-23], [8.202028101383677, 2.7527920757181414e-23, 2.7364893063418865e-23, 2.6167263011499047e-23], [8.224770356596302, 2.7344460356788323e-23, 2.7180673284130634e-23, 2.5993314534615026e-23], [8.247575670623867, 2.7163368715245253e-23, 2.6998826472453816e-23, 2.5821756178858894e-23], [8.270444218313328, 2.69845642508534e-23, 2.6819270818168844e-23, 2.565250261816102e-23], [8.293376174996448, 2.6807968936820947e-23, 2.664192807083748e-23, 2.5485472247579715e-23], [8.316371716491153, 2.6633508134763686e-23, 2.646672337317697e-23, 2.5320587011285215e-23], [8.339431019102854, 2.6461110436210425e-23, 2.629358510246344e-23, 2.515777223812651e-23], [8.362554259625835, 2.6290707511653402e-23, 2.6122444719529386e-23, 2.499695648550993e-23], [8.385741615344573, 2.6122233966930787e-23, 2.5953236625031155e-23, 2.4838071389952807e-23], [8.40899326403511, 2.5955627206590322e-23, 2.5785898022668285e-23, 2.4681051525543887e-23], [8.432309383966436, 2.5790827303901625e-23, 2.562036878921274e-23, 2.452583426861165e-23], [8.455690153901816, 2.562777687728832e-23, 2.545659135068757e-23, 2.4372359668981782e-23], [8.479135753100202, 2.546642097292177e-23, 2.5294510564984688e-23, 2.422057032802145e-23], [8.502646361317574, 2.5306706953103e-23, 2.513407361006297e-23, 2.407041128216014e-23], [8.526222158808341, 2.5148584390272155e-23, 2.497522987784814e-23, 2.3921829892209885e-23], [8.54986332632671, 2.4992004966341818e-23, 2.48179308735183e-23, 2.377477573839579e-23], [8.573570045128072, 2.4836922377191702e-23, 2.4662130119851144e-23, 2.3629200520536743e-23], [8.59734249697041, 2.468329224194672e-23, 2.450778306637209e-23, 2.3485057963220588e-23], [8.621180864115662, 2.453107201694997e-23, 2.4354847003332716e-23, 2.3342303725581168e-23], [8.645085329331145, 2.4380220914201096e-23, 2.420328097994069e-23, 2.320089531627149e-23], [8.669056075890941, 2.42306998239804e-23, 2.40530457269207e-23, 2.306079201207945e-23], [8.693093287577305, 2.4082471241468957e-23, 2.390410358307478e-23, 2.292195478111326e-23], [8.717197148682088, 2.3935499197330872e-23, 2.3756418425728536e-23, 2.2784346209992766e-23], [8.741367844008117, 2.3789749191781134e-23, 2.3609955604784207e-23, 2.2647930434598658e-23], [8.76560555887066, 2.364518813228025e-23, 2.3464681880288304e-23, 2.2512673074529062e-23], [8.789910479098793, 2.3501784274467734e-23, 2.3320565363264094e-23, 2.237854117122819e-23], [8.814282791036874, 2.335950716626226e-23, 2.3177575459805688e-23, 2.2245503128998335e-23], [8.838722681545931, 2.3218327594903537e-23, 2.303568281801314e-23, 2.2113528659354367e-23], [8.863230338005122, 2.307821753695499e-23, 2.2894859277926237e-23, 2.1982588728447338e-23], [8.887805948313165, 2.293915011086118e-23, 2.275507782406279e-23, 2.185265550684311e-23], [8.912449700889766, 2.2801099532134864e-23, 2.2616312540576854e-23, 2.1723702322483114e-23], [8.937161784677086, 2.2664041071014393e-23, 2.247853856877671e-23, 2.1595703615994427e-23], [8.961942389141168, 2.2527951012344108e-23, 2.234173206704052e-23, 2.1468634898333898e-23], [8.986791704273395, 2.2392806617612116e-23, 2.2205870172812625e-23, 2.13424727104352e-23], [9.01170992059197, 2.2258586089175884e-23, 2.20709309667563e-23, 2.1217194585913692e-23], [9.036697229143337, 2.2125268536292348e-23, 2.1936893438787134e-23, 2.1092779014733183e-23], [9.061753821503677, 2.1992833942988933e-23, 2.1803737455990792e-23, 2.096920540926243e-23], [9.086879889780361, 2.1861263137798592e-23, 2.1671443732219776e-23, 2.084645407232449e-23], [9.112075626613425, 2.1730537764986346e-23, 2.1539993799376113e-23, 2.0724506166763167e-23], [9.13734122517706, 2.1600640257311707e-23, 2.1409369980168584e-23, 2.0603343686375944e-23], [9.162676879181067, 2.1471553810343273e-23, 2.1279555362455914e-23, 2.0482949428882548e-23], [9.188082782872382, 2.1343262357959412e-23, 2.1150533774669073e-23, 2.0363306969644126e-23], [9.213559131036506, 2.1215750549218655e-23, 2.1022289762788927e-23, 2.0244400637331676e-23], [9.239106118999072, 2.1089003726394292e-23, 2.0894808568333535e-23, 2.0126215490579594e-23], [9.264723942627278, 2.0963007904024394e-23, 2.0768076107437786e-23, 2.0008737295383688e-23], [9.290412798331417, 2.0837749749056026e-23, 2.0642078951046754e-23, 1.9891952504411739e-23], [9.316172883066397, 2.0713216561926192e-23, 2.0516804305977447e-23, 1.977584823653102e-23], [9.342004394333213, 2.0589396258412198e-23, 2.039223999687043e-23, 1.9660412257552336e-23], [9.367907530180501, 2.0466277352448313e-23, 2.0268374448948502e-23, 1.9545632961955154e-23], [9.393882489206028, 2.0343848939496865e-23, 2.014519667153477e-23, 1.943149935508262e-23], [9.419929470558223, 2.0222100680707832e-23, 2.0022696242141592e-23, 1.9318001036257247e-23], [9.446048673937721, 2.010102278761801e-23, 1.9900863291338525e-23, 1.9205128182347147e-23], [9.472240299598859, 1.9980606007462686e-23, 1.9779688487993745e-23, 1.9092871532230783e-23], [9.498504548351246, 1.9860841608920287e-23, 1.9659163025104376e-23, 1.8981222371478545e-23], [9.524841621561281, 1.9741721368251872e-23, 1.953927860606105e-23, 1.887017251742393e-23], [9.551251721153703, 1.962323755597677e-23, 1.942002743129369e-23, 1.875971430509186e-23], [9.577735049613146, 1.9505382923724598e-23, 1.9301402185173558e-23, 1.8649840573013253e-23], [9.604291809985678, 1.9388150691437217e-23, 1.9183396023374663e-23, 1.8540544649446725e-23], [9.63092220588037, 1.927153453487557e-23, 1.906600256029649e-23, 1.8431820339295254e-23], [9.657626441470846, 1.9155528573221687e-23, 1.8949215856844352e-23, 1.8323661910582366e-23], [9.684404721496868, 1.9040127356884845e-23, 1.8833030408260327e-23, 1.8216064081362623e-23], [9.711257251265879, 1.8925325855548452e-23, 1.8717441132143332e-23, 1.810902200727884e-23], [9.73818423665459, 1.881111944606679e-23, 1.8602443356564634e-23, 1.8002531268184416e-23], [9.765185884110567, 1.869750390068715e-23, 1.848803280820796e-23, 1.7896587855869548e-23], [9.7922624006538, 1.858447537512162e-23, 1.8374205600595164e-23, 1.779118816113997e-23], [9.8194139938783, 1.8472030396774462e-23, 1.8260958222269908e-23, 1.7686328961430483e-23], [9.846640871953682, 1.8360165852784534e-23, 1.814828752499567e-23, 1.7582007407782121e-23], [9.873943243626762, 1.8248878978186952e-23, 1.8036190711894827e-23, 1.747822101242743e-23], [9.901321318223175, 1.8138167343941805e-23, 1.7924665325565308e-23, 1.7374967635974675e-23], [9.92877530564895, 1.802802884491903e-23, 1.7813709236116202e-23, 1.7272245474489436e-23], [9.956305416392148, 1.7918461687812413e-23, 1.7703320629068322e-23, 1.7170053046902206e-23], [9.98391186152445, 1.7809464378896408e-23, 1.7593497993241072e-23, 1.7068389181586103e-23], [10.011594852702796, 1.770103581718841e-23, 1.748424021517288e-23, 1.6967253113699342e-23], [10.039354602171002, 1.7593175913274852e-23, 1.7375547205957982e-23, 1.6866645130186722e-23], [10.067191322761374, 1.748588406016788e-23, 1.72674183534946e-23, 1.6766564974008666e-23], [10.09510522789637, 1.7379159700791714e-23, 1.7159853096811788e-23, 1.6667012436601294e-23], [10.1230965315902, 1.7273002527051168e-23, 1.705285112779594e-23, 1.6567987565250322e-23], [10.151165448450488, 1.71674124661451e-23, 1.6946412377573154e-23, 1.6469490648437917e-23], [10.17931219367992, 1.7062389666971817e-23, 1.68405370027983e-23, 1.6371522201804435e-23], [10.20753698307787, 1.695793448634591e-23, 1.6735225371954306e-23, 1.6274082953231605e-23], [10.235840033042102, 1.6854047475296796e-23, 1.6630478051617533e-23, 1.6177173828662845e-23], [10.26422156057037, 1.6750729365342438e-23, 1.6526295792675544e-23, 1.6080795937451483e-23], [10.292681783262129, 1.6647981054674367e-23, 1.642267951655641e-23, 1.5984950557769615e-23], [10.321220919320176, 1.6545803594487156e-23, 1.6319630301449675e-23, 1.5889639122337753e-23], [10.349839187552334, 1.644419817522528e-23, 1.621714936859028e-23, 1.5794863203726378e-23], [10.378536807373127, 1.6343166112988943e-23, 1.6115238068563824e-23, 1.5700624500351593e-23], [10.407313998805458, 1.624270883589154e-23, 1.601389786764118e-23, 1.560692482190241e-23], [10.43617098248232, 1.6142827870562714e-23, 1.5913130334207992e-23, 1.551376607504011e-23], [10.465107979648447, 1.6043524828827915e-23, 1.5812937125382207e-23, 1.542115024988613e-23], [10.494125212162036, 1.594480139433343e-23, 1.5713319973599785e-23, 1.532907940533446e-23], [10.523222902496453, 1.5846659309594867e-23, 1.5614280673513385e-23, 1.523755565613342e-23], [10.552401273741923, 1.574910036295404e-23, 1.5515821068945282e-23, 1.5146581158605894e-23], [10.581660549607243, 1.5652126375950136e-23, 1.5417943040107125e-23, 1.5056158097970907e-23], [10.611000954421502, 1.555573919077522e-23, 1.5320648491004243e-23, 1.4966288674773353e-23], [10.640422713135811, 1.5459940658118005e-23, 1.522393933713579e-23, 1.4876975092417664e-23], [10.669926051325003, 1.5364732625158736e-23, 1.512781749337726e-23, 1.4788219544685618e-23], [10.699511195189373, 1.5270116923934915e-23, 1.503228486228642e-23, 1.4700024203258803e-23], [10.729178371556443, 1.5176095360042968e-23, 1.4937343322638926e-23, 1.46123912063632e-23], [10.758927807882628, 1.5082669701671465e-23, 1.4842994718292435e-23, 1.4525322647046723e-23], [10.78875973225507, 1.4989841668956784e-23, 1.4749240847543043e-23, 1.4438820562219715e-23], [10.81867437339331, 1.4897612923820623e-23, 1.4656083452690925e-23, 1.435288692203903e-23], [10.848671960651085, 1.480598506016856e-23, 1.4563524210172185e-23, 1.4267523619767375e-23], [10.878752724018085, 1.4714959594500091e-23, 1.4471564721032278e-23, 1.418273246208832e-23], [10.908916894121692, 1.462453795703202e-23, 1.438020650182917e-23, 1.409851515969944e-23], [10.939164702228762, 1.4534721483217438e-23, 1.4289450976088808e-23, 1.401487331888952e-23], [10.96949638024741, 1.4445511405748525e-23, 1.419929946615164e-23, 1.3931808432889528e-23], [10.999912160728758, 1.4356908847115267e-23, 1.4109753185497977e-23, 1.38493218745235e-23], [11.030412276868763, 1.426891481257977e-23, 1.4020813231654263e-23, 1.3767414888816596e-23], [11.06099696250995, 1.4181530183702397e-23, 1.393248057953218e-23, 1.368608858636342e-23], [11.09166645214326, 1.40947557124151e-23, 1.3844756075307533e-23, 1.3605343937391582e-23], [11.122420980909801, 1.400859201547447e-23, 1.3757640430821783e-23, 1.3525181765864273e-23], [11.153260784602677, 1.392303956966108e-23, 1.3671134218477165e-23, 1.3445602744855198e-23], [11.184186099668795, 1.3838098707273622e-23, 1.3585237866749998e-23, 1.336660739178877e-23], [11.215197163210656, 1.3753769612340416e-23, 1.349995165605982e-23, 1.328819606483163e-23], [11.246294212988216, 1.367005231717371e-23, 1.3415275715368982e-23, 1.3210368959253615e-23], [11.277477487420656, 1.3586946699623483e-23, 1.3331210019090728e-23, 1.313312610486121e-23], [11.30874722558826, 1.3504452480646339e-23, 1.3247754384699225e-23, 1.3056467363520495e-23], [11.340103667234201, 1.34225692225522e-23, 1.3164908470704374e-23, 1.2980392427586156e-23], [11.37154705276642, 1.3341296327642873e-23, 1.3082671775170302e-23, 1.2904900818639725e-23], [11.403077623259442, 1.3260633037331426e-23, 1.3001043634753938e-23, 1.282999188662163e-23], [11.434695620456228, 1.3180578431787161e-23, 1.2920023224153259e-23, 1.2755664809798502e-23], [11.466401286770061, 1.3101131430000126e-23, 1.283960955608486e-23, 1.2681918595026715e-23], [11.498194865286347, 1.302229079022141e-23, 1.2759801481611577e-23, 1.2608752078063345e-23], [11.530076599764518, 1.2944055110997309e-23, 1.2680597690997442e-23, 1.253616392537164e-23], [11.56204673463991, 1.2866422832391643e-23, 1.2601996714892595e-23, 1.2464152634987666e-23], [11.5941055150256, 1.2789392237801642e-23, 1.2523996925986854e-23, 1.2392716539149086e-23], [11.626253186714308, 1.2712961455956936e-23, 1.2446596540928877e-23, 1.232185380611392e-23], [11.658489996180279, 1.2637128463421367e-23, 1.2369793622722256e-23, 1.225156244322381e-23], [11.690816190581181, 1.25618910873152e-23, 1.2293586083392108e-23, 1.2181840299847842e-23], [11.723232017759983, 1.2487247008377464e-23, 1.2217971686921115e-23, 1.211268507066414e-23], [11.75573772624686, 1.2413193764334707e-23, 1.2142948052566584e-23, 1.2044094299525901e-23], [11.788333565261105, 1.2339728753536178e-23, 1.2068512658391805e-23, 1.197606538329094e-23], [11.821019784713034, 1.2266849238751738e-23, 1.1994662845037684e-23, 1.1908595576034383e-23], [11.853796635205914, 1.219455235129021e-23, 1.1921395819771094e-23, 1.1841681993432697e-23], [11.886664368037867, 1.212283509526155e-23, 1.1848708660701438e-23, 1.1775321617437488e-23], [11.919623235203794, 1.2051694352033247e-23, 1.1776598321150584e-23, 1.1709511301093684e-23], [11.952673489397345, 1.1981126884826374e-23, 1.1705061634245019e-23, 1.1644247773389951e-23], [11.985815384012803, 1.1911129343472018e-23, 1.1634095317638467e-23, 1.1579527644517507e-23], [12.019049173147064, 1.1841698269262189e-23, 1.1563695978322029e-23, 1.151534741100416e-23], [12.052375111601568, 1.1772830099892437e-23, 1.1493860117544934e-23, 1.1451703460957649e-23], [12.08579345488426, 1.1704521174515437e-23, 1.1424584135851392e-23, 1.1388592079591506e-23], [12.119304459211557, 1.1636767738790317e-23, 1.1355864338111053e-23, 1.132600945448755e-23], [12.152908381510285, 1.1569565950041912e-23, 1.1287696938714248e-23, 1.1263951681193792e-23], [12.186605479419686, 1.1502911882408483e-23, 1.1220078066633986e-23, 1.1202414768812137e-23], [12.22039601129336, 1.1436801531966019e-23, 1.1153003770638959e-23, 1.1141394645152465e-23], [12.25428023620126, 1.1371230821892271e-23, 1.1086470024421654e-23, 1.1080887162514771e-23], [12.288258413931686, 1.1306195607621902e-23, 1.102047273175794e-23, 1.1020888103158483e-23], [12.322330804993253, 1.1241691681841992e-23, 1.09550077315741e-23, 1.0961393184297907e-23], [12.356497670616928, 1.1177714779636452e-23, 1.0890070803022267e-23, 1.090239806400327e-23], [12.390759272757983, 1.1114260583365585e-23, 1.0825657670469924e-23, 1.0843898346019044e-23], [12.425115874098054, 1.105132472760906e-23, 1.0761764008437781e-23, 1.0785889585166639e-23], [12.45956773804711, 1.0988902803983345e-23, 1.0698385446405982e-23, 1.0728367292467358e-23], [12.494115128745502, 1.0926990365805939e-23, 1.0635517573615016e-23, 1.0671326939854189e-23], [12.528758311065976, 1.0865582932798411e-23, 1.0573155943711428e-23, 1.0614763965462566e-23], [12.5634975506157, 1.0804675995527727e-23, 1.051129607924027e-23, 1.0558673778080681e-23], [12.59833311373833, 1.0744265019800981e-23, 1.0449933476157891e-23, 1.0503051761827416e-23], [12.633265267516004, 1.0684345450993765e-23, 1.0389063608086464e-23, 1.0447893280821871e-23], [12.668294279771418, 1.0624912718098875e-23, 1.0328681930536224e-23, 1.0393193683247153e-23], [12.703420419069896, 1.0565962237819185e-23, 1.0268783884922738e-23, 1.0338948305858933e-23], [12.73864395472141, 1.0507489418375282e-23, 1.0209364902537501e-23, 1.0285152477830755e-23], [12.773965156782666, 1.044948966327099e-23, 1.0150420408278109e-23, 1.0231801524755833e-23], [12.80938429605917, 1.0391958374901275e-23, 1.0091945824338482e-23, 1.017889077254931e-23], [12.844901644107326, 1.0334890957945897e-23, 1.0033936573667937e-23, 1.0126415550752925e-23], [12.880517473236479, 1.0278282822708876e-23, 9.976388083348414e-24, 1.007437119623481e-23], [12.916232056511028, 1.02221293882716e-23, 9.919295787806933e-24, 1.0022753056422886e-23], [12.952045667752515, 1.0166426085488774e-23, 9.86265513185462e-24, 9.971556492372229e-24], [12.987958581541713, 1.0111168359867949e-23, 9.806461573657616e-24, 9.920776881843536e-24], [13.02397107322077, 1.0056351674285045e-23, 9.750710587439057e-24, 9.870409622121753e-24], [13.06008341889526, 1.0001971511573034e-23, 9.69539766618591e-24, 9.820450132723304e-24], [13.096295895436345, 9.948023376928981e-24, 9.64051832407535e-24, 9.77089385776017e-24], [13.132608780482894, 9.894502800239001e-24, 9.586068098889304e-24, 9.721736268502546e-24], [13.169022352443584, 9.841405338231668e-24, 9.532042554150784e-24, 9.67297286559883e-24], [13.205536890499062, 9.78872657650414e-24, 9.478437281279548e-24, 9.624599181117034e-24], [13.242152674604073, 9.736462131369661e-24, 9.4252479014752e-24, 9.576610780327539e-24], [13.278869985489605, 9.684607651723164e-24, 9.37247006756622e-24, 9.529003263978422e-24], [13.315689104665065, 9.633158820562375e-24, 9.320099465662039e-24, 9.481772269444056e-24], [13.352610314420389, 9.582111356575566e-24, 9.268131816792635e-24, 9.43491347262516e-24], [13.389633897828263, 9.531461015529594e-24, 9.216562878253016e-24, 9.388422589299116e-24], [13.42676013874625, 9.481203591521463e-24, 9.165388445013386e-24, 9.342295376331038e-24], [13.463989321818982, 9.43133491819328e-24, 9.114604350895724e-24, 9.296527633011675e-24], [13.501321732480346, 9.38185086973624e-24, 9.064206469679865e-24, 9.251115201928397e-24], [13.53875765695566, 9.332747361919048e-24, 9.01419071613563e-24, 9.206053970177816e-24], [13.576297382263899, 9.284020352888957e-24, 8.964553046908717e-24, 9.161339870040723e-24], [13.613941196219844, 9.235665843985254e-24, 8.915289461316109e-24, 9.116968879886658e-24], [13.651689387436345, 9.187679880355138e-24, 8.866396002076863e-24, 9.072937024588808e-24], [13.689542245326484, 9.140058551600232e-24, 8.817868755933559e-24, 9.02924037642694e-24], [13.727500060105823, 9.092797992245792e-24, 8.769703854169726e-24, 8.985875055385574e-24], [13.765563122794626, 9.04589438215762e-24, 8.7218974731257e-24, 8.942837229561034e-24], [13.803731725220075, 8.99934394689395e-24, 8.674445834497457e-24, 8.900123115448231e-24], [13.842006160018542, 8.953142958006804e-24, 8.627345205712855e-24, 8.857728978411137e-24], [13.88038672063778, 8.907287733206403e-24, 8.58059190014102e-24, 8.815651132550578e-24], [13.918873701339232, 8.861774636516164e-24, 8.53418227727418e-24, 8.773885941000769e-24], [13.957467397200233, 8.816600078380801e-24, 8.48811274282359e-24, 8.732429815997568e-24], [13.996168104116308, 8.771760515651816e-24, 8.44237974877553e-24, 8.691279218785295e-24], [14.034976118803428, 8.727252451556084e-24, 8.396979793421344e-24, 8.650430659465396e-24], [14.073891738800274, 8.68307243564116e-24, 8.351909421233417e-24, 8.609880697061634e-24], [14.112915262470558, 8.639217063645072e-24, 8.307165222832531e-24, 8.569625939314746e-24], [14.152046989005262, 8.595682977285013e-24, 8.262743834774989e-24, 8.529663042305287e-24], [14.191287218424954, 8.552466864069613e-24, 8.21864193941633e-24, 8.489988710380494e-24], [14.230636251582093, 8.509565457051481e-24, 8.174856264625974e-24, 8.4505996957285e-24], [14.270094390163317, 8.466975534507248e-24, 8.131383583555388e-24, 8.41149279811033e-24], [14.309661936691795, 8.424693919652665e-24, 8.088220714309652e-24, 8.372664864510333e-24], [14.349339194529495, 8.382717480235569e-24, 8.045364519626208e-24, 8.334112788598026e-24], [14.389126467879539, 8.341043128202121e-24, 8.002811906504916e-24, 8.29583351042443e-24], [14.42902406178855, 8.299667819290206e-24, 7.960559825816749e-24, 8.257824016080155e-24], [14.469032282148948, 8.25858855253769e-24, 7.918605271877642e-24, 8.220081336839553e-24], [14.509151435701334, 8.217802369911649e-24, 7.876945282043214e-24, 8.182602549092381e-24], [14.54938183003681, 8.177306355764533e-24, 7.835576936208694e-24, 8.145384773496656e-24], [14.589723773599383, 8.137097636397194e-24, 7.794497356349302e-24, 8.108425174624279e-24], [14.630177575688279, 8.0971733795118e-24, 7.753703706038607e-24, 8.071720960314331e-24], [14.670743546460335, 8.057530793740117e-24, 7.713193189911805e-24, 8.035269381227739e-24], [14.7114219969324, 8.018167128060359e-24, 7.672963053191608e-24, 7.999067730072739e-24], [14.752213238983675, 7.979079671296616e-24, 7.63301058110684e-24, 7.963113341197593e-24], [14.793117585358138, 7.940265751568219e-24, 7.59333309838748e-24, 7.92740358999281e-24], [14.834135349666926, 7.901722735696827e-24, 7.5539279686922e-24, 7.891935892104628e-24], [14.875266846390744, 7.863448028665854e-24, 7.5147925940701e-24, 7.856707702977737e-24], [14.916512390882295, 7.82543907307582e-24, 7.475924414376998e-24, 7.821716517288574e-24], [14.957872299368645, 7.787693348500121e-24, 7.437320906739107e-24, 7.786959868068497e-24], [14.99934688895372, 7.75020837101197e-24, 7.398979584939234e-24, 7.752435326499488e-24], [15.040936477620672, 7.712981692517129e-24, 7.360897998908002e-24, 7.718140500911759e-24], [15.08264138423435, 7.676010900214529e-24, 7.323073734090527e-24, 7.684073036323703e-24], [15.124461928543742, 7.63929361601958e-24, 7.285504410886986e-24, 7.650230613854977e-24], [15.166398431184415, 7.602827495980991e-24, 7.248187684115608e-24, 7.616610950034347e-24], [15.208451213681004, 7.566610229712387e-24, 7.211121242390981e-24, 7.58321179627568e-24], [15.250620598449622, 7.53063953979054e-24, 7.174302807579037e-24, 7.55003093806108e-24], [15.292906908800399, 7.494913181245884e-24, 7.137730134241716e-24, 7.517066194705971e-24], [15.335310468939907, 7.459428940921612e-24, 7.101401009036391e-24, 7.484315418289164e-24], [15.377831603973645, 7.424184636986406e-24, 7.06531325019371e-24, 7.451776493520528e-24], [15.420470639908597, 7.38917811830997e-24, 7.029464706931852e-24, 7.419447336774106e-24], [15.463227903655646, 7.35440726394797e-24, 6.993853258927245e-24, 7.387325895655661e-24], [15.506103723032155, 7.319869982603636e-24, 6.958476815736267e-24, 7.355410148527761e-24], [15.549098426764417, 7.285564212045064e-24, 6.923333316308523e-24, 7.323698103672978e-24], [15.592212344490214, 7.251487918636866e-24, 6.888420728400232e-24, 7.292187799080777e-24], [15.63544580676134, 7.217639096729826e-24, 6.853737048083593e-24, 7.260877301465983e-24], [15.678799145046115, 7.184015768218798e-24, 6.819280299174637e-24, 7.229764706130217e-24], [15.722272691731966, 7.150615981967422e-24, 6.78504853278024e-24, 7.198848136093668e-24], [15.76586678012793, 7.117437813350144e-24, 6.751039826773719e-24, 7.168125741823631e-24], [15.809581744467225, 7.084479363706671e-24, 6.717252285235808e-24, 7.137595700423283e-24], [15.853417919909848, 7.051738759907493e-24, 6.683684038066343e-24, 7.107256215479128e-24], [15.897375642545086, 7.019214153793693e-24, 6.650333240395723e-24, 7.077105516108569e-24], [15.941455249394126, 6.986903721786699e-24, 6.617198072161569e-24, 7.047141856897294e-24], [15.98565707841263, 6.95480566435005e-24, 6.584276737666238e-24, 7.017363517076245e-24], [16.029981468493357, 6.922918205576669e-24, 6.5515674650409405e-24, 6.987768800227553e-24], [16.074428759468706, 6.89123959270551e-24, 6.519068505839509e-24, 6.958356033727646e-24], [16.118999292113344, 6.859768095698261e-24, 6.486778134584324e-24, 6.929123568320652e-24], [16.163693408146862, 6.828502006793588e-24, 6.4546946483345604e-24, 6.900069777596202e-24], [16.208511450236323, 6.797439640075849e-24, 6.4228163662398204e-24, 6.871193057609287e-24], [16.25345376199893, 6.766579331091266e-24, 6.391141629133212e-24, 6.8424918265067e-24], [16.29852068800467, 6.735919436371739e-24, 6.359668799107338e-24, 6.813964523875602e-24], [16.343712573778905, 6.705458333091182e-24, 6.3283962591213314e-24, 6.785609610541084e-24], [16.38902976580511, 6.675194418639345e-24, 6.297322412578468e-24, 6.757425568020684e-24], [16.43447261152742, 6.645126110253442e-24, 6.266445682973541e-24, 6.729410898232444e-24], [16.48004145935339, 6.615251844616273e-24, 6.2357645134725806e-24, 6.701564122963268e-24], [16.525736658656605, 6.585570077507848e-24, 6.2052773665696606e-24, 6.673883783617464e-24], [16.571558559779383, 6.556079283436356e-24, 6.174982723692813e-24, 6.64636844076006e-24], [16.617507514035456, 6.526777955254368e-24, 6.1448790848761775e-24, 6.619016673686241e-24], [16.66358387371266, 6.497664603863549e-24, 6.114964968370638e-24, 6.591827080246477e-24], [16.709787992075665, 6.468737757825234e-24, 6.085238910333317e-24, 6.564798276290684e-24], [16.75612022336862, 6.4399959630565686e-24, 6.0556994644772944e-24, 6.537928895472535e-24], [16.802580922817953, 6.411437782493354e-24, 6.026345201745756e-24, 6.511217588832486e-24], [16.849170446635007, 6.383061795763711e-24, 5.997174709980685e-24, 6.484663024435537e-24], [16.895889152018828, 6.354866598888187e-24, 5.968186593617265e-24, 6.458263887103289e-24], [16.94273739715891, 6.326850803991122e-24, 5.9393794733831726e-24, 6.4320188782124054e-24], [16.989715541237892, 6.299013038959355e-24, 5.910751985958599e-24, 6.405926715116137e-24], [17.036823944434367, 6.271351947178421e-24, 5.8823027837595994e-24, 6.37998613103537e-24], [17.08406296792559, 6.2438661872522546e-24, 5.854030534547746e-24, 6.354195874727254e-24], [17.131432973890288, 6.216554432709084e-24, 5.825933921238588e-24, 6.3285547101802266e-24], [17.17893432551142, 6.18941537174653e-24, 5.798011641569251e-24, 6.3030614163653945e-24], [17.226567386978957, 6.162447706950985e-24, 5.770262407869228e-24, 6.2777147869232364e-24], [17.27433252349271, 6.135650155039841e-24, 5.742684946767856e-24, 6.252513629889116e-24], [17.322230101265063, 6.109021446621202e-24, 5.71527799893985e-24, 6.227456767478015e-24], [17.370260487523836, 6.0825603259368236e-24, 5.688040318890668e-24, 6.202543035828198e-24], [17.418424050515096, 6.0562655506109185e-24, 5.6609706746461264e-24, 6.17777128469291e-24], [17.466721159505955, 6.030135891439618e-24, 5.634067847593341e-24, 6.153140377265367e-24], [17.515152184787404, 6.0041701321243246e-24, 5.607330632157926e-24, 6.128649189859041e-24], [17.563717497677175, 5.978367069084718e-24, 5.5807578356762974e-24, 6.104296611810593e-24], [17.612417470522583, 5.95272551120935e-24, 5.554348278061249e-24, 6.080081545098279e-24], [17.66125247670336, 5.927244279658598e-24, 5.52810079169067e-24, 6.0560029042474566e-24], [17.71022289063452, 5.901922207642432e-24, 5.502014221113676e-24, 6.0320596160272115e-24], [17.759329087769274, 5.876758140222363e-24, 5.476087422887396e-24, 6.008250619292322e-24], [17.808571444601842, 5.851750934109957e-24, 5.4503192653473045e-24, 5.984574864763911e-24], [17.85795033867039, 5.826899457474034e-24, 5.4247086284411486e-24, 5.9610313148368405e-24], [17.90746614855989, 5.8022025897512405e-24, 5.399254403510057e-24, 5.937618943399372e-24], [17.957119253905052, 5.777659221440396e-24, 5.373955493097826e-24, 5.9143367355495776e-24], [18.00691003539323, 5.7532682539452984e-24, 5.348810810770581e-24, 5.891183687539898e-24], [18.05683887476731, 5.7290285993954215e-24, 5.323819280952246e-24, 5.868158806559072e-24], [18.106906154828685, 5.7049391804437504e-24, 5.298979838727422e-24, 5.845261110449583e-24], [18.157112259440147, 5.680998930138048e-24, 5.274291429678213e-24, 5.822489627705695e-24], [18.20745757352885, 5.657206791720187e-24, 5.249753009716874e-24, 5.7998433971712874e-24], [18.25794248308926, 5.6335617184964804e-24, 5.225363544913288e-24, 5.777321467980241e-24], [18.3085673751861, 5.61006267364526e-24, 5.201122011339626e-24, 5.754922899285671e-24], [18.35933263795737, 5.5867086300770546e-24, 5.177027394919413e-24, 5.7326467601311666e-24], [18.410238660617242, 5.563498570305388e-24, 5.15307869125705e-24, 5.710492129416904e-24], [18.461285833459105, 5.5404314862702496e-24, 5.129274905516349e-24, 5.68845809561167e-24], [18.51247454785853, 5.51750637920741e-24, 5.105615052248765e-24, 5.666543756663014e-24], [18.563805196276277, 5.494722259498022e-24, 5.0820981552600895e-24, 5.644748219829474e-24], [18.615278172261327, 5.4720781465448815e-24, 5.058723247459631e-24, 5.6230706015675956e-24], [18.66689387045384, 5.4495730686422414e-24, 5.035489370744805e-24, 5.6015100274676714e-24], [18.718652686588264, 5.427206062814496e-24, 5.012395575858871e-24, 5.5800656319215736e-24], [18.770555017496292, 5.404976174728199e-24, 4.9894409222370436e-24, 5.558736558220131e-24], [18.822601261109934, 5.382882458528295e-24, 4.966624477931885e-24, 5.5375219582456766e-24], [18.87479181646459, 5.360923976747408e-24, 4.943945319428792e-24, 5.5164209924825725e-24], [18.92712708370206, 5.339099800178595e-24, 4.9214025315710765e-24, 5.4954328298566624e-24], [18.979607464073688, 5.317409007733746e-24, 4.8989952074068044e-24, 5.474556647534635e-24], [19.03223335994335, 5.29585068636675e-24, 4.876722448092536e-24, 5.4537916309464785e-24], [19.085005174790588, 5.274423930948104e-24, 4.8545833627807434e-24, 5.4331369736281895e-24], [19.137923313213715, 5.253127844131008e-24, 4.8325770684897925e-24, 5.4125918770095155e-24], [19.19098818093288, 5.231961536277381e-24, 4.810702690024585e-24, 5.392155550422571e-24], [19.244200184793204, 5.210924125343958e-24, 4.788959359825651e-24, 5.3718272110017815e-24], [19.297559732767876, 5.190014736790936e-24, 4.767346217917309e-24, 5.351606083590588e-24], [19.35106723396133, 5.169232503428577e-24, 4.745862411770772e-24, 5.331491400427007e-24], [19.404723098612322, 5.1485765654066726e-24, 4.724507096204521e-24, 5.311482401416441e-24], [19.458527738097096, 5.1280460700419174e-24, 4.703279433305604e-24, 5.291578333712656e-24], [19.512481564932575, 5.1076401717648005e-24, 4.6821785923298774e-24, 5.27177845179685e-24], [19.566584992779468, 5.087358032014447e-24, 4.661203749588861e-24, 5.2520820173232446e-24], [19.620838436445467, 5.067198819172971e-24, 4.640354088383639e-24, 5.232488299129928e-24], [19.675242311888432, 5.04716170841908e-24, 4.619628798887558e-24, 5.212996572925258e-24], [19.729797036219566, 5.027245881714224e-24, 4.599027078095758e-24, 5.193606121464814e-24], [19.78450302770664, 5.007450527675344e-24, 4.5785481297038176e-24, 5.174316234346088e-24], [19.839360705777157, 4.987774841494882e-24, 4.558191164047621e-24, 5.155126207869834e-24], [19.89437049102162, 4.968218024884584e-24, 4.537955397996433e-24, 5.136035345114986e-24], [19.9495328051967, 4.948779285964701e-24, 4.517840054909656e-24, 5.117042955705609e-24], [20.00484807122852, 4.929457839204593e-24, 4.4978443645200355e-24, 5.098148355808631e-24], [20.060316713215855, 4.910252905357895e-24, 4.477967562870386e-24, 5.079350868114541e-24], [20.115939156433413, 4.891163711352859e-24, 4.458208892248183e-24, 5.060649821601034e-24], [20.171715827335102, 4.872189490264796e-24, 4.438567601092898e-24, 5.042044551667946e-24], [20.227647153557268, 4.853329481213064e-24, 4.419042943941386e-24, 5.02353439992256e-24], [20.28373356392199, 4.8345829293007616e-24, 4.39963418133881e-24, 5.005118714153864e-24], [20.33997548844037, 4.81594908553928e-24, 4.3803405797907505e-24, 4.9867968482299814e-24], [20.39637335831582, 4.7974272068045814e-24, 4.361161411653518e-24, 4.968568162153916e-24], [20.452927605947405, 4.779016555737922e-24, 4.342095955130688e-24, 4.950432021817135e-24], [20.509638664933085, 4.760716400718591e-24, 4.323143494135733e-24, 4.9323877991366444e-24], [20.566506970073124, 4.742526015761463e-24, 4.3043033182933266e-24, 4.914434871797019e-24], [20.623532957373364, 4.724444680487661e-24, 4.285574722829971e-24, 4.89657262334331e-24], [20.680717064048576, 4.706471680051032e-24, 4.266957008513693e-24, 4.87880044306313e-24], [20.73805972852584, 4.6886063050662435e-24, 4.24844948163463e-24, 4.861117725871604e-24], [20.795561390447872, 4.6708478515790855e-24, 4.230051453892588e-24, 4.843523872380015e-24], [20.853222490676444, 4.6531956209902445e-24, 4.211762242372039e-24, 4.826018288781954e-24], [20.911043471295695, 4.635648919988301e-24, 4.193581169473716e-24, 4.808600386705138e-24], [20.96902477561557, 4.618207060519857e-24, 4.175507562852664e-24, 4.791269583306661e-24], [21.02716684817522, 4.600869359719537e-24, 4.157540755377476e-24, 4.774025301146107e-24], [21.085470134746387, 4.583635139860716e-24, 4.139680085060617e-24, 4.75686696813649e-24], [21.143935082336817, 4.566503728303044e-24, 4.121924895010344e-24, 4.73979401750125e-24], [21.20256213919371, 4.549474457444413e-24, 4.10427453337176e-24, 4.722805887741326e-24], [21.26135175480717, 4.532546664661949e-24, 4.086728353297025e-24, 4.70590202254789e-24], [21.3203043799136, 4.515719692272678e-24, 4.069285712864523e-24, 4.6890818708050145e-24], [21.379420466499187, 4.498992887473295e-24, 4.051945975046558e-24, 4.672344886472867e-24], [21.438700467803397, 4.4823656023022645e-24, 4.034708507658534e-24, 4.655690528599211e-24], [21.498144838322393, 4.4658371935960546e-24, 4.017572683307806e-24, 4.6391182612900276e-24], [21.557754033812554, 4.4494070229261004e-24, 4.000537879345436e-24, 4.6226275535895165e-24], [21.617528511293962, 4.433074456568623e-24, 3.9836034778237894e-24, 4.60621787949959e-24], [21.677468729053892, 4.4168388654515995e-24, 3.9667688654420524e-24, 4.5898887178992194e-24], [21.73757514665038, 4.4006996251174046e-24, 3.9500334335007644e-24, 4.573639552534546e-24], [21.797848224915658, 4.3846561156763675e-24, 3.9333965778904586e-24, 4.5574698719533095e-24], [21.858288425959778, 4.368707721765545e-24, 3.916857698983645e-24, 4.5413791694971126e-24], [21.918896213174083, 4.352853832499311e-24, 3.900416201651974e-24, 4.5253669431808326e-24], [21.979672051234797, 4.337093841442295e-24, 3.8840714951964354e-24, 4.509432695764695e-24], [22.040616406106576, 4.3214271465514114e-24, 3.8678229933060505e-24, 4.4935759346017796e-24], [22.101729745046082, 4.305853150158508e-24, 3.851670114014268e-24, 4.477796171710346e-24], [22.163012536605596, 4.290371258905122e-24, 3.8356122796788796e-24, 4.462092923637823e-24], [22.22446525063655, 4.274980883728287e-24, 3.819648916918528e-24, 4.4464657114945135e-24], [22.286088358293163, 4.259681439801354e-24, 3.803779456577012e-24, 4.4309140608540036e-24], [22.347882332036068, 4.244472346511009e-24, 3.7880033337059634e-24, 4.415437501776111e-24], [22.4098476456359, 4.229353027413992e-24, 3.772319987500225e-24, 4.4000355687251775e-24], [22.471984774176967, 4.214322910208433e-24, 3.756728861263757e-24, 4.384707800583361e-24], [22.534294194060852, 4.1993814266824236e-24, 3.741229402397216e-24, 4.369453740552955e-24], [22.59677638301011, 4.184528012694228e-24, 3.725821062326324e-24, 4.35427293616964e-24], [22.659431820071877, 4.169762108128203e-24, 3.710503296500915e-24, 4.3391649392436276e-24], [22.72226098562159, 4.1550831568665186e-24, 3.69527556432969e-24, 4.3241293058487036e-24], [22.78526436136664, 4.1404906067460735e-24, 3.6801373291557966e-24, 4.309165596236545e-24], [22.848442430350076, 4.1259839095435545e-24, 3.66508805823588e-24, 4.29427337491418e-24], [22.911795676954338, 4.111562520913264e-24, 3.650127222682072e-24, 4.279452210437885e-24], [22.9753245869049, 4.0972259003868116e-24, 3.635254297450463e-24, 4.264701675568269e-24], [23.039029647274077, 4.082973511316606e-24, 3.620468761287552e-24, 4.250021347105281e-24], [23.102911346484696, 4.0688048208607066e-24, 3.6057700967127654e-24, 4.235410805926908e-24], [23.166970174313853, 4.054719299933758e-24, 3.591157789987502e-24, 4.220869636890732e-24], [23.231206621896703, 4.040716423195405e-24, 3.576631331061729e-24, 4.2063974288799866e-24], [23.295621181730183, 4.026795669007988e-24, 3.5621902135703765e-24, 4.191993774727964e-24], [23.360214347676816, 4.012956519409284e-24, 3.54783393478255e-24, 4.177658271188039e-24], [23.424986614968454, 3.999198460082968e-24, 3.533561995580431e-24, 4.163390518916842e-24], [23.48993848021019, 3.985520980326076e-24, 3.519373900423795e-24, 4.1491901224161e-24], [23.555070441384004, 3.9719235730361526e-24, 3.505269157321758e-24, 4.13505669009034e-24], [23.620382997852694, 3.958405734649254e-24, 3.49124727780443e-24, 4.120989834058295e-24], [23.685876650363646, 3.944966965147534e-24, 3.477307776883935e-24, 4.106989170282243e-24], [23.751551901052743, 3.931606768005391e-24, 3.463450173049234e-24, 4.0930543184258646e-24], [23.817409253448133, 3.918324650181914e-24, 3.449673988210102e-24, 4.079184901922329e-24], [23.883449212474137, 3.905120122071978e-24, 3.43597874768e-24, 4.065380547838044e-24], [23.94967228445511, 3.891992697508603e-24, 3.422363980146152e-24, 4.0516408869777844e-24], [24.01607897711932, 3.878941893691708e-24, 3.4088292176555714e-24, 4.037965553664064e-24], [24.08266979960286, 3.865967231205756e-24, 3.395373995559598e-24, 4.024354185893521e-24], [24.149445262453522, 3.8530682339788716e-24, 3.3819978525128844e-24, 4.010806425269176e-24], [24.21640587763472, 3.840244429243177e-24, 3.368700330430683e-24, 3.997321916868623e-24], [24.28355215852942, 3.827495347520087e-24, 3.3554809744620384e-24, 3.983900309306345e-24], [24.350884619944075, 3.81482052260076e-24, 3.342339332980445e-24, 3.970541254708633e-24], [24.4184037781126, 3.8022194915115035e-24, 3.329274957536874e-24, 3.957244408648747e-24], [24.486110150700274, 3.789691794489052e-24, 3.316287402838855e-24, 3.944009430125756e-24], [24.554004256807737, 3.777236974962258e-24, 3.3033762267337915e-24, 3.9308359815595e-24], [24.622086616974983, 3.764854579519472e-24, 3.2905409901779758e-24, 3.917723728723639e-24], [24.690357753185307, 3.75254415789366e-24, 3.2777812571999963e-24, 3.904672340766467e-24], [24.758818188869412, 3.740305262926678e-24, 3.265096594897377e-24, 3.89168149013475e-24], [24.827468448909286, 3.728137450556797e-24, 3.25248657339278e-24, 3.878750852587829e-24], [24.896309059642295, 3.716040279787071e-24, 3.239950765824567e-24, 3.865880107131605e-24], [24.96534054886523, 3.704013312669128e-24, 3.227488748297966e-24, 3.853068936044798e-24], [25.0345634458383, 3.692056114268432e-24, 3.215100099892524e-24, 3.840317024769308e-24], [25.103978281289294, 3.680168252655005e-24, 3.2027844026162747e-24, 3.827624061970804e-24], [25.173585587417524, 3.668349298867175e-24, 3.190541241384934e-24, 3.814989739436074e-24], [25.243385897897976, 3.656598826904424e-24, 3.17837020400114e-24, 3.802413752126109e-24], [25.313379747885396, 3.644916413685624e-24, 3.1662708811320416e-24, 3.7898957980529285e-24], [25.383567674018366, 3.633301639041685e-24, 3.154242866282569e-24, 3.777435578319888e-24], [25.453950214423482, 3.621754085697243e-24, 3.1422857557721752e-24, 3.765032797120838e-24], [25.524527908719396, 3.610273339219004e-24, 3.130399148713015e-24, 3.75268716157388e-24], [25.595301298021003, 3.598858988035201e-24, 3.1185826469858536e-24, 3.7403983818744174e-24], [25.66627092494358, 3.587510623381388e-24, 3.1068358552185914e-24, 3.728166171132604e-24], [25.737437333606923, 3.576227839295697e-24, 3.095158380760592e-24, 3.715990245417425e-24], [25.80880106963962, 3.565010232591924e-24, 3.083549833655243e-24, 3.703870323693018e-24], [25.880362680183076, 3.55385740283338e-24, 3.072009826641111e-24, 3.691806127789155e-24], [25.95212271389581, 3.542768952330983e-24, 3.0605379750980807e-24, 3.6797973824454264e-24], [26.024081720957632, 3.531744486088668e-24, 3.0491338970398974e-24, 3.667843815141978e-24], [26.09624025307391, 3.5207836118228235e-24, 3.0377972130915497e-24, 3.655945156241515e-24], [26.168598863479712, 3.5098859399047945e-24, 3.0265275464753976e-24, 3.6441011388067406e-24], [26.241158106944113, 3.499051083364489e-24, 3.0153245229750112e-24, 3.632311498678352e-24], [26.31391853977442, 3.488278657857325e-24, 3.004187770907912e-24, 3.6205759743931106e-24], [26.386880719820446, 3.477568281655678e-24, 2.993116921132323e-24, 3.6088943071974105e-24], [26.460045206478824, 3.466919575612716e-24, 2.9821116070013546e-24, 3.597266240967067e-24], [26.533412560697215, 3.456332163157085e-24, 2.9711714643466194e-24, 3.585691522228549e-24], [26.606983344978673, 3.445805670259964e-24, 2.9602961314583112e-24, 3.574169900081598e-24], [26.680758123385925, 3.435339725426261e-24, 2.9494852490589414e-24, 3.562701126220593e-24], [26.75473746154571, 3.424933959675392e-24, 2.9387384603008916e-24, 3.551284954902531e-24], [26.82892192665315, 3.414588006505757e-24, 2.928055410723787e-24, 3.53992114285722e-24], [26.903312087476017, 3.404301501891597e-24, 2.917435748234566e-24, 3.528609449331972e-24], [26.977908514359157, 3.394074084256951e-24, 2.9068791231024688e-24, 3.517349636016243e-24], [27.052711779228854, 3.383905394457823e-24, 2.896385187924135e-24, 3.506141467046754e-24], [27.127722455597162, 3.373795075760725e-24, 2.8859535976130826e-24, 3.494984708948517e-24], [27.20294111856644, 3.3637427738282525e-24, 2.8755840093654874e-24, 3.48387913063831e-24], [27.27836834483358, 3.353748136693352e-24, 2.865276082649277e-24, 3.472824503358975e-24], [27.354004712694554, 3.343810814742119e-24, 2.8550294791886216e-24, 3.461820600669901e-24], [27.42985080204879, 3.333930460704164e-24, 2.8448438629284954e-24, 3.4508671984423575e-24], [27.505907194403648, 3.3241067296152455e-24, 2.834718900032027e-24, 3.439964074762817e-24], [27.582174472878886, 3.314339278821521e-24, 2.824654258837759e-24, 3.429111009997411e-24], [27.658653222211093, 3.3046277679390674e-24, 2.8146496098616466e-24, 3.418307786672317e-24], [27.735344028758195, 3.294971858852098e-24, 2.8047046257648406e-24, 3.40755418951215e-24], [27.812247480503945, 3.2853712156821165e-24, 2.7948189813407682e-24, 3.39685000536406e-24], [27.88936416706243, 3.2758255047796403e-24, 2.784992353486199e-24, 3.386195023199743e-24], [27.966694679682625, 3.2663343947009423e-24, 2.7752244211878378e-24, 3.3755890340773755e-24], [28.044239611252863, 3.2568975561854314e-24, 2.7655148655042085e-24, 3.365031831082484e-24], [28.12199955630543, 3.2475146621549387e-24, 2.7558633695330532e-24, 3.354523209389227e-24], [28.19997511102107, 3.2381853876674747e-24, 2.7462696184098633e-24, 3.344062966084679e-24], [28.278166873233687, 3.2289094099312286e-24, 2.7367332992816737e-24, 3.3336509002994286e-24], [28.356575442434746, 3.2196864082587568e-24, 2.727254101276817e-24, 3.3232868130462426e-24], [28.435201419777982, 3.210516064067818e-24, 2.7178317155041814e-24, 3.3129705072661796e-24], [28.514045408083994, 3.201398060857293e-24, 2.7084658350165977e-24, 3.302701787786125e-24], [28.593108011844834, 3.192332084187865e-24, 2.6991561547991842e-24, 3.2924804612670268e-24], [28.672389837228717, 3.1833178216721912e-24, 2.689902371756514e-24, 3.282306336211673e-24], [28.75189149208457, 3.174354962944398e-24, 2.6807041846849056e-24, 3.2721792228830528e-24], [28.831613585946755, 3.165443199654782e-24, 2.6715612942506513e-24, 3.262098933314964e-24], [28.91155673003972, 3.156582225449113e-24, 2.6624734029821952e-24, 3.2520652812813693e-24], [28.991721537282686, 3.1477717359534546e-24, 2.6534402152417783e-24, 3.242078082263125e-24], [29.07210862229439, 3.139011428750434e-24, 2.644461437213707e-24, 3.2321371533922337e-24], [29.15271860139772, 3.130301003365401e-24, 2.6355367768803444e-24, 3.222242313438635e-24], [29.23355209262449, 3.121640161259896e-24, 2.6266659440045398e-24, 3.2123933828240166e-24], [29.314609715720167, 3.1130286057973423e-24, 2.617848650109134e-24, 3.2025901835085836e-24], [29.395892092148614, 3.1044660422392816e-24, 2.6090846084746344e-24, 3.1928325390283006e-24], [29.477399845096905, 3.0959521777290788e-24, 2.6003735340885266e-24, 3.1831202744574126e-24], [29.559133599480013, 3.087486721261653e-24, 2.5917151436647358e-24, 3.173453216322221e-24], [29.64109398194568, 3.0790693836878465e-24, 2.5831091555959836e-24, 3.1638311926584227e-24], [29.723281620879174, 3.0706998776797194e-24, 2.574555289949116e-24, 3.1542540329086264e-24], [29.805697146408104, 3.062377917727787e-24, 2.5660532684482307e-24, 3.144721567940229e-24], [29.888341190407363, 3.0541032201158133e-24, 2.557602814452347e-24, 3.135233629983832e-24], [29.971214386503785, 3.0458755029112616e-24, 2.549203652937634e-24, 3.1257900526286376e-24], [30.054317370081137, 3.0376944859480672e-24, 2.5408555104828028e-24, 3.1163906707893375e-24], [30.137650778284943, 3.029559890806555e-24, 2.532558115248909e-24, 3.1070353206500415e-24], [30.22121525002737, 3.0214714408049585e-24, 2.5243111969663276e-24, 3.097723839677145e-24], [30.305011425992184, 3.013428860979537e-24, 2.51611448691074e-24, 3.088456066563329e-24], [30.389039948639557, 3.005431878069512e-24, 2.5079677178917462e-24, 3.0792318412034502e-24], [30.473301462211072, 2.997480220502587e-24, 2.4998706242315635e-24, 3.070051004667869e-24], [30.55779661273464, 2.98957361837986e-24, 2.4918229417547798e-24, 3.060913399174055e-24], [30.642526048029435, 2.9817118034637113e-24, 2.4838244077628576e-24, 3.0518188680695395e-24], [30.727490417710918, 2.9738945091542663e-24, 2.4758747610201143e-24, 3.0427672557697538e-24], [30.812690373195743, 2.9661214704870768e-24, 2.4679737417453893e-24, 3.03375840778123e-24], [30.89812656770679, 2.9583924241072156e-24, 2.4601210915856612e-24, 3.024792170621979e-24], [30.983799656278148, 2.950707108262621e-24, 2.452316553598739e-24, 3.015868391836384e-24], [31.069710295760245, 2.9430652627819634e-24, 2.444559872244885e-24, 3.0069869199254894e-24], [31.155859144824706, 2.935466629068333e-24, 2.436850793369629e-24, 2.998147604355189e-24], [31.24224686396953, 2.92791095008425e-24, 2.429189064180014e-24, 2.9893502955372593e-24], [31.32887411552413, 2.9203979703287253e-24, 2.4215744332384336e-24, 2.98059484474341e-24], [31.415741563654315, 2.91292743583337e-24, 2.4140066504430277e-24, 2.9718811041368968e-24], [31.50284987436762, 2.905499094146415e-24, 2.4064854670062534e-24, 2.9632089267363058e-24], [31.590199715518143, 2.898112694312196e-24, 2.399010635450467e-24, 2.954578166346831e-24], [31.677791756811807, 2.890767986868734e-24, 2.3915819095889316e-24, 2.9459886775980697e-24], [31.76562666981148, 2.8834647238250033e-24, 2.3841990445014737e-24, 2.9374403158691717e-24], [31.853705127942085, 2.8762026586503658e-24, 2.3768617965341225e-24, 2.9289329372738657e-24], [31.942027806495847, 2.8689815462661017e-24, 2.369569923276247e-24, 2.920466398657453e-24], [32.030595382637365, 2.861801143024661e-24, 2.3623231835442035e-24, 2.9120405575409588e-24], [32.11940853540887, 2.8546612067050117e-24, 2.3551213373728604e-24, 2.9036552721278828e-24], [32.20846794573541, 2.8475614964875537e-24, 2.3479641459956748e-24, 2.8953104012237315e-24], [32.297774296430056, 2.840501772961122e-24, 2.3408513718364345e-24, 2.887005804306255e-24], [32.38732827219921, 2.8334817980862554e-24, 2.3337827784893875e-24, 2.8787413413813423e-24], [32.47713055964776, 2.8265013352035217e-24, 2.326758130708813e-24, 2.8705168730692813e-24], [32.56718184728438, 2.8195601490118187e-24, 2.319777194394285e-24, 2.862332260516953e-24], [32.65748282552683, 2.812658005555284e-24, 2.312839736576984e-24, 2.8541873653849e-24], [32.748034186707194, 2.80579467221257e-24, 2.305945525407802e-24, 2.846082049827464e-24], [32.83883662507732, 2.798969917693811e-24, 2.2990943301432397e-24, 2.8380161765058022e-24], [32.92989083681397, 2.7921835120110105e-24, 2.292285921130982e-24, 2.829989608486428e-24], [33.02119752002424, 2.785435226484562e-24, 2.285520069799229e-24, 2.8220022093014467e-24], [33.112757374750935, 2.7787248337197957e-24, 2.2787965486429485e-24, 2.8140538428697844e-24], [33.20457110297789, 2.7720521076018797e-24, 2.2721151312139806e-24, 2.806144373507498e-24], [33.29663940863541, 2.7654168232814774e-24, 2.2654755921009925e-24, 2.798273665885792e-24], [33.38896299760559, 2.7588187571671057e-24, 2.2588777069304555e-24, 2.7904415850342147e-24], [33.48154257772779, 2.752257686913196e-24, 2.252321252342322e-24, 2.782647996305029e-24], [33.57437885880401, 2.7457333914041784e-24, 2.2458060059845765e-24, 2.7748927653363803e-24], [33.66747255260438, 2.739245650748943e-24, 2.2393317464994767e-24, 2.7671757580518642e-24], [33.76082437287263, 2.732794246277653e-24, 2.232898253511137e-24, 2.7594968406777685e-24], [33.85443503533147, 2.7263789605126397e-24, 2.226505307622653e-24, 2.751855879632884e-24], [33.948305257688176, 2.719999577178144e-24, 2.2201526903927994e-24, 2.7442527416035712e-24], [34.04243575964001, 2.7136558811762624e-24, 2.2138401843325873e-24, 2.736687293461818e-24], [34.13682726287987, 2.707347658587148e-24, 2.2075675728947555e-24, 2.729159402288377e-24], [34.23148049110166, 2.701074696649096e-24, 2.2013346404613533e-24, 2.7216689353166717e-24], [34.32639617000594, 2.6948367837591576e-24, 2.19514117233203e-24, 2.7142157599511057e-24], [34.421575027305444, 2.6886337094588992e-24, 2.188986954715849e-24, 2.7067997437402193e-24], [34.51701779273065, 2.6824652644227917e-24, 2.1828717747240954e-24, 2.699420754337236e-24], [34.61272519803552, 2.6763312404525775e-24, 2.1767954203585787e-24, 2.692078659510228e-24], [34.70869797700289, 2.6702314304726517e-24, 2.170757680498665e-24, 2.6847733271432775e-24], [34.80493686545024, 2.664165628510648e-24, 2.1647583448962748e-24, 2.677504625168539e-24], [34.90144260123526, 2.6581336296944877e-24, 2.158797204168416e-24, 2.6702724215805663e-24], [34.998215924261565, 2.6521352302477177e-24, 2.1528740497779452e-24, 2.663076584440245e-24], [35.09525757648437, 2.6461702274739453e-24, 2.1469886740429195e-24, 2.6559169818310028e-24], [35.19256830191613, 2.6402384197532385e-24, 2.1411408701077847e-24, 2.648793481857675e-24], [35.29014884663226, 2.634339606534266e-24, 2.135330431951893e-24, 2.641705952644648e-24], [35.387999958776874, 2.628473588321028e-24, 2.129557154373302e-24, 2.6346542622906907e-24], [35.48612238856848, 2.6226401666730698e-24, 2.123820832977547e-24, 2.6276382788963904e-24], [35.584516888305835, 2.616839144190519e-24, 2.1181212641812984e-24, 2.620657870522562e-24], [35.683184212373575, 2.611070324511462e-24, 2.1124582451912343e-24, 2.613712905194507e-24], [35.78212511724808, 2.6053335123024492e-24, 2.106831574009763e-24, 2.6068032508880305e-24], [35.881340361503256, 2.599628513252483e-24, 2.1012410494172406e-24, 2.5999287755196047e-24], [35.98083070581633, 2.5939551340628894e-24, 2.0956864709738885e-24, 2.5930893469233406e-24], [36.0805969129738, 2.5883131824452033e-24, 2.0901676390032524e-24, 2.58628483286678e-24], [36.18063974787708, 2.5827024671104007e-24, 2.0846843545949215e-24, 2.579515101021123e-24], [36.28095997754852, 2.577122797763469e-24, 2.079236419595391e-24, 2.5727800189559512e-24], [36.38155837113722, 2.5715739850973463e-24, 2.073823636596697e-24, 2.5660794541371128e-24], [36.48243569992493, 2.566055840785232e-24, 2.0684458089377825e-24, 2.5594132739130023e-24], [36.583592737332054, 2.56056817747673e-24, 2.063102740696064e-24, 2.5527813455159016e-24], [36.68503025892344, 2.5551108087887944e-24, 2.0577942366782992e-24, 2.5461835360400637e-24], [36.7867490424144, 2.5496835493014152e-24, 2.0525201024217684e-24, 2.5396197124516093e-24], [36.88874986767667, 2.5442862145492105e-24, 2.047280144184449e-24, 2.533089741556686e-24], [36.99103351674437, 2.5389186210236246e-24, 2.042074168937653e-24, 2.526593490047408e-24], [37.093600773820064, 2.533580586154583e-24, 2.03690198437083e-24, 2.5201308244253534e-24], [37.19645242528066, 2.528271928315987e-24, 2.0317633988752636e-24, 2.5137016110607984e-24], [37.29958925968353, 2.5229924668125168e-24, 2.0266582215482204e-24, 2.507305716135088e-24], [37.40301206777249, 2.5177420218819656e-24, 2.0215862621829067e-24, 2.5009430056888964e-24], [37.50672164248399, 2.512520414682099e-24, 2.0165473312689813e-24, 2.4946133455699314e-24], [37.61071877895303, 2.507327467293064e-24, 2.0115412399848647e-24, 2.48831660146973e-24], [37.7150042745193, 2.5021630027043132e-24, 2.0065678001934874e-24, 2.4820526388812674e-24], [37.8195789287334, 2.497026844817123e-24, 2.0016268244457957e-24, 2.475821323129245e-24], [37.92444354336279, 2.4919188184386085e-24, 1.9967181259665876e-24, 2.4696225193688067e-24], [38.0295989223982, 2.486838749270447e-24, 1.9918415186590837e-24, 2.4634560925398692e-24], [38.135045872059486, 2.4817864639138306e-24, 1.9869968170987564e-24, 2.45732190742569e-24], [38.24078520080201, 2.476761789857423e-24, 1.98218383653069e-24, 2.4512198286011293e-24], [38.346817719322786, 2.471764555479488e-24, 1.9774023928675996e-24, 2.445149720474865e-24], [38.45314424056669, 2.4667945900366197e-24, 1.972652302684523e-24, 2.439111447244713e-24], [38.55976557973272, 2.461851723665344e-24, 1.9679333832185605e-24, 2.4331048729319646e-24], [38.6666825542802, 2.4569357873766704e-24, 1.9632454523678116e-24, 2.4271298613756847e-24], [38.77389598393509, 2.4520466130496187e-24, 1.9585883286848897e-24, 2.421186276216271e-24], [38.88140669069623, 2.4471840334313367e-24, 1.9539618313790246e-24, 2.415273980915262e-24], [38.98921549884168, 2.442347882129971e-24, 1.9493657803094194e-24, 2.4093928387478025e-24], [39.09732323493505, 2.4375379936131024e-24, 1.9447999959894623e-24, 2.4035427128056985e-24], [39.20573072783178, 2.4327542032037585e-24, 1.940264299581067e-24, 2.3977234660024426e-24], [39.31443880868552, 2.427996347075708e-24, 1.935758512891545e-24, 2.3919349610683395e-24], [39.42344831095454, 2.4232642622525163e-24, 1.9312824583762796e-24, 2.386177060565585e-24], [39.53276007040804, 2.4185577866008857e-24, 1.9268359591347503e-24, 2.3804496268745263e-24], [39.6423749251327, 2.4138767588307784e-24, 1.9224188389117938e-24, 2.374752522211743e-24], [39.75229371553894, 2.4092210184897043e-24, 1.9180309220922307e-24, 2.3690856086268846e-24], [39.862517284367435, 2.40459040596108e-24, 1.913672033703971e-24, 2.363448748004483e-24], [39.9730464766956, 2.3999847624612613e-24, 1.9093419994155424e-24, 2.3578418020760178e-24], [40.08388213994402, 2.3954039300355404e-24, 1.9050406455351434e-24, 2.3522646324157623e-24], [40.19502512388305, 2.3908477515559985e-24, 1.9007677990112756e-24, 2.346717100453091e-24], [40.30647628063916, 2.3863160707184232e-24, 1.8965232874293404e-24, 2.3411990674672725e-24], [40.41823646470161, 2.3818087320395288e-24, 1.892306939014937e-24, 2.3357103946020643e-24], [40.53030653292895, 2.3773255808554973e-24, 1.888118582631243e-24, 2.3302509428723858e-24], [40.642687344555554, 2.3728664633158246e-24, 1.883958047779056e-24, 2.3248205731534956e-24], [40.755379761198334, 2.368431226386588e-24, 1.879825164596808e-24, 2.319419146216774e-24], [40.868384646863184, 2.3640197178419813e-24, 1.8757197638605122e-24, 2.31404652270211e-24], [40.9817028679517, 2.359631786264258e-24, 1.871641676984787e-24, 2.3087025631433654e-24], [41.09533529326775, 2.3552672810441256e-24, 1.867590736020951e-24, 2.3033871279820736e-24], [41.20928279402437, 2.3509260523729127e-24, 1.863566773658908e-24, 2.2981000775546752e-24], [41.32354624385002, 2.3466079512443836e-24, 1.85956962322775e-24, 2.29284127210848e-24], [41.43812651879567, 2.342312829450393e-24, 1.855599118692146e-24, 2.287610571810573e-24], [41.55302449734132, 2.338040539580858e-24, 1.8516550946590503e-24, 2.2824078367606903e-24], [41.668241060402764, 2.33379093501785e-24, 1.847737386374663e-24, 2.277232926977079e-24], [41.783777091338465, 2.329563869937967e-24, 1.8438458297219928e-24, 2.2720857024331196e-24], [41.89963347595612, 2.3253591993070687e-24, 1.8399802612291006e-24, 2.2669660230449443e-24], [42.015811102519606, 2.3211767788790083e-24, 1.8361405180625036e-24, 2.2618737486866597e-24], [42.13231086175573, 2.317016465194221e-24, 1.832326438032423e-24, 2.2568087391984115e-24], [42.24913364686103, 2.312878115575941e-24, 1.8285378595900163e-24, 2.25177085438519e-24], [42.36628035350878, 2.308761588130791e-24, 1.824774621832197e-24, 2.2467599540450778e-24], [42.48375187985562, 2.304666741745057e-24, 1.8210365644999446e-24, 2.2417758979599992e-24], [42.60154912654862, 2.3005934360835282e-24, 1.8173235279792102e-24, 2.236818545914549e-24], [42.71967299673212, 2.2965415315867645e-24, 1.8136353533030384e-24, 2.2318877576961583e-24], [42.83812439605464, 2.292510889470309e-24, 1.8099718821501664e-24, 2.2269833931131968e-24], [42.9569042326759, 2.2885013717221507e-24, 1.806332956850387e-24, 2.2221053119973884e-24], [43.076013417273685, 2.2845128411014134e-24, 1.802718420382213e-24, 2.2172533742173895e-24], [43.19545286305088, 2.280545161135177e-24, 1.7991281163738554e-24, 2.212427439682852e-24], [43.31522348574246, 2.276598196118176e-24, 1.795561889105929e-24, 2.2076273683544335e-24], [43.435326203622466, 2.2726718111109943e-24, 1.7920195835122554e-24, 2.202853020262265e-24], [43.55576193751125, 2.2687658719377887e-24, 1.7885010451804292e-24, 2.1981042555047618e-24], [43.676531610782206, 2.264880245183409e-24, 1.785006120352567e-24, 2.193380934255274e-24], [43.797636149369104, 2.2610147981947565e-24, 1.7815346559290126e-24, 2.1886829167900136e-24], [43.919076481773075, 2.257169399074683e-24, 1.778086499466492e-24, 2.1840100634714227e-24], [44.04085353906976, 2.253343916684958e-24, 1.7746614991813147e-24, 2.179362234783134e-24], [44.16296825491651, 2.2495382206407497e-24, 1.7712595039496816e-24, 2.1747392913224214e-24], [44.2854215655594, 2.2457521813117872e-24, 1.767880363308324e-24, 2.1701410938186244e-24], [44.40821440984055, 2.241985669817586e-24, 1.7645239274579965e-24, 2.1655675031335916e-24], [44.531347729205194, 2.2382385580289573e-24, 1.7611900472632035e-24, 2.1610183802838003e-24], [44.65482246770908, 2.234510718562947e-24, 1.7578785742521547e-24, 2.1564935864343805e-24], [44.7786395720255, 2.2308020247856364e-24, 1.7545893606209557e-24, 2.1519929829329035e-24], [44.902799991452675, 2.227112350805034e-24, 1.75132225923284e-24, 2.147516431288451e-24], [45.027304677921, 2.223441571473899e-24, 1.7480771236202412e-24, 2.1430637932067765e-24], [45.152154586000336, 2.219789562385149e-24, 1.7448538079845464e-24, 2.138634930584284e-24], [45.277350672907374, 2.2161561998716024e-24, 1.741652167199533e-24, 2.1342297055241368e-24], [45.40289389851291, 2.212541361003838e-24, 1.7384720568110506e-24, 2.129847980350154e-24], [45.52878522534921, 2.2089449235878977e-24, 1.73531333303912e-24, 2.125489617602761e-24], [45.655025618617444, 2.2053667661653496e-24, 1.7321758527770956e-24, 2.121154480065174e-24], [45.781616046195, 2.2018067680084118e-24, 1.729059473595183e-24, 2.1168424307556445e-24], [45.90855747864309, 2.1982648091213387e-24, 1.7259640537421503e-24, 2.11255333295014e-24], [46.03585088921392, 2.1947407702373113e-24, 1.7228894521423884e-24, 2.1082870501921498e-24], [46.16349725385834, 2.191234532815977e-24, 1.7198355284008687e-24, 2.1040434462873304e-24], [46.29149755123327, 2.1877459790425496e-24, 1.7168021428032286e-24, 2.09982238532798e-24], [46.41985276270921, 2.184274991826778e-24, 1.7137891563153548e-24, 2.095623731698931e-24], [46.54856387237781, 2.1808214547987443e-24, 1.7107964305862687e-24, 2.091447350080727e-24], [46.67763186705931, 2.177385252309112e-24, 1.707823827947976e-24, 2.0872931054601618e-24], [46.8070577363102, 2.1739662694263713e-24, 1.704871211416993e-24, 2.0831608631469586e-24], [46.93684247243074, 2.170564391935653e-24, 1.7019384446944175e-24, 2.0790504887743733e-24], [47.066987070472614, 2.1671795063356414e-24, 1.6990253921676782e-24, 2.0749618483099722e-24], [47.1974925282466, 2.1638114998379967e-24, 1.696131918910878e-24, 2.0708948080675996e-24], [47.32835984633009, 2.1604602603647916e-24, 1.69325789068657e-24, 2.0668492347107763e-24], [47.45959002807486, 2.1571256765463164e-24, 1.6904031739438728e-24, 2.0628249952665733e-24], [47.591184079614756, 2.153807637719378e-24, 1.6875676358219457e-24, 2.0588219571275273e-24], [47.723143009873326, 2.1505060339260007e-24, 1.684751144150453e-24, 2.054839988069876e-24], [47.85546783057178, 2.1472207559098274e-24, 1.6819535674484353e-24, 2.050878956247305e-24], [47.98815955623643, 2.1439516951162613e-24, 1.6791747749256234e-24, 2.0469387302171243e-24], [48.12121920420668, 2.1406987436886744e-24, 1.676414636484225e-24, 2.0430191789335094e-24], [48.25464779464278, 2.137461794466243e-24, 1.6736730227174875e-24, 2.0391201717563776e-24], [48.3884463505336, 2.134240740984025e-24, 1.6709498049120287e-24, 2.035241578472332e-24], [48.52261589770459, 2.131035477468173e-24, 1.6682448550466402e-24, 2.031383269285558e-24], [48.6571574648255, 2.127845898835776e-24, 1.665558045793557e-24, 2.027545114840191e-24], [48.79207208341836, 2.1246719006920187e-24, 1.6628892505185394e-24, 2.023726986213866e-24], [48.927360787865325, 2.1215133793273444e-24, 1.660238343280538e-24, 2.0199287549343757e-24], [49.06302461541674, 2.118370231716964e-24, 1.6576051988339802e-24, 2.0161502929875968e-24], [49.1990646061989, 2.115242355516833e-24, 1.6549896926262132e-24, 2.0123914728158123e-24], [49.33548180322215, 2.112129649062096e-24, 1.6523917007989307e-24, 2.0086521673305985e-24], [49.47227725238888, 2.109032011366268e-24, 1.6498111001891806e-24, 2.0049322499261504e-24], [49.609452002501435, 2.1059493421164487e-24, 1.6472477683274684e-24, 2.001231594470357e-24], [49.74700710527037, 2.1028815416726043e-24, 1.6447015834387876e-24, 1.997550075322536e-24], [49.884943615322264, 2.0998285110649387e-24, 1.6421724244428517e-24, 1.9938875673379332e-24], [50.02326259020798, 2.09679015199205e-24, 1.6396601709531734e-24, 1.9902439458741257e-24], [50.16196509041069, 2.0937663668176234e-24, 1.6371647032771974e-24, 1.986619086794828e-24], [50.301052179353995, 2.090757058568573e-24, 1.6346859024162727e-24, 1.9830128664761526e-24], [50.44052492341026, 2.0877621309325413e-24, 1.6322236500654232e-24, 1.9794251618161613e-24], [50.580384391908495, 2.0847814882558303e-24, 1.6297778286121491e-24, 1.9758558502344928e-24], [50.720631657142754, 2.0818150355407356e-24, 1.6273483211369568e-24, 1.9723048096852546e-24], [50.8612677943803, 2.078862678442495e-24, 1.624935011413087e-24, 1.9687719186551562e-24], [51.00229388186984, 2.0759243232677676e-24, 1.6225377839046732e-24, 1.9652570561749608e-24], [51.14371100084987, 2.0729998769719382e-24, 1.6201565237675736e-24, 1.9617601018203542e-24], [51.28552023555683, 2.0700892471557707e-24, 1.6177911168483689e-24, 1.9582809357178437e-24], [51.42772267323352, 2.0671923420642985e-24, 1.6154414496832816e-24, 1.9548194385539445e-24], [51.57031940413738, 2.064309070582536e-24, 1.6131074094980906e-24, 1.9513754915711523e-24], [51.713311521548874, 2.0614393422343364e-24, 1.610788884207097e-24, 1.9479489765806076e-24], [51.85670012177992, 2.0585830671793184e-24, 1.6084857624116475e-24, 1.9445397759653593e-24], [52.00048630418221, 2.055740156209844e-24, 1.6061979334006733e-24, 1.941147772677407e-24], [52.14467117115566, 2.0529105207491375e-24, 1.6039252871490328e-24, 1.9377728502544684e-24], [52.28925582815688, 2.0500940728478346e-24, 1.6016677143156115e-24, 1.934414892809614e-24], [52.43424138370762, 2.0472907251817053e-24, 1.599425106244577e-24, 1.931073785047658e-24], [52.579628949403435, 2.0445003910493633e-24, 1.5971973549623833e-24, 1.9277494122619016e-24], [52.7254196399219, 2.0417229843686262e-24, 1.5949843531766656e-24, 1.9244416603399466e-24], [52.8716145730314, 2.0389584196747548e-24, 1.5927859942766348e-24, 1.921150415766845e-24], [53.0182148695996, 2.036206612116709e-24, 1.5906021723299993e-24, 1.9178755656269803e-24], [53.16522165360205, 2.0334674774552752e-24, 1.5884327820837004e-24, 1.914616997610199e-24], [53.3126360521309, 2.0307409320596613e-24, 1.5862777189606685e-24, 1.9113746000125635e-24], [53.460459195403345, 2.028026892905307e-24, 1.5841368790592055e-24, 1.908148261741524e-24], [53.60869221677047, 2.0253252775704488e-24, 1.582010159152562e-24, 1.9049378723148063e-24], [53.75733625272581, 2.0226360042335916e-24, 1.5798974566859304e-24, 1.9017433218670606e-24], [53.906392442914225, 2.0199589916707705e-24, 1.577798669775325e-24, 1.898564501150985e-24], [54.05586193014044, 2.0172941592523714e-24, 1.5757136972075383e-24, 1.895401301538194e-24], [54.20574586037794, 2.0146414269403343e-24, 1.5736424384358778e-24, 1.892253615024051e-24], [54.35604538277772, 2.0120007152858812e-24, 1.5715847935812636e-24, 1.8891213342296174e-24], [54.50676164967707, 2.0093719454254943e-24, 1.569540663428501e-24, 1.8860043523995442e-24], [54.6578958166085, 2.006755039078934e-24, 1.5675099494258618e-24, 1.8829025634099788e-24], [54.80944904230848, 2.0041499185460635e-24, 1.5654925536825236e-24, 1.8798158617666395e-24], [54.96142248872637, 2.0015565067035534e-24, 1.5634883789670377e-24, 1.8767441426056666e-24], [55.11381732103337, 1.998974727002389e-24, 1.5614973287053328e-24, 1.873687301699185e-24], [55.266634707631354, 1.996404503464835e-24, 1.5595193069796783e-24, 1.870645235452022e-24], [55.41987582016204, 1.993845760680926e-24, 1.5575542185252175e-24, 1.867617840906221e-24], [55.5735418335157, 1.991298423806551e-24, 1.5556019687295084e-24, 1.864605015740928e-24], [55.72763392584031, 1.9887624185591113e-24, 1.553662463629529e-24, 1.8616066582732697e-24], [55.88215327855059, 1.9862376712162197e-24, 1.5517356099098662e-24, 1.858622667462373e-24], [56.03710107633701, 1.9837241086103487e-24, 1.549821314900997e-24, 1.8556529429006324e-24], [56.19247850717499, 1.9812216581283205e-24, 1.5479194865764414e-24, 1.85269738482926e-24], [56.348286762333814, 1.9787302477065843e-24, 1.5460300335515214e-24, 1.8497558941237867e-24], [56.50452703638593, 1.9762498058290373e-24, 1.5441528650802586e-24, 1.846828372307613e-24], [56.661200527216025, 1.9737802615231525e-24, 1.5422878910540442e-24, 1.8439147215414812e-24], [56.818308436030215, 1.97132154435775e-24, 1.5404350219986545e-24, 1.8410148446304227e-24], [56.97585196736535, 1.9688735844396392e-24, 1.5385941690720456e-24, 1.8381286450218342e-24], [57.1338323290981, 1.966436312410448e-24, 1.5367652440621433e-24, 1.8352560268052886e-24], [57.29225073245429, 1.9640096594439104e-24, 1.5349481593848442e-24, 1.8323968947139145e-24], [57.45110839201821, 1.961593557241975e-24, 1.5331428280810722e-24, 1.8295511541211093e-24], [57.61040652574184, 1.9591879380327835e-24, 1.5313491638146516e-24, 1.826718711044574e-24], [57.7701463549544, 1.9567927345668573e-24, 1.5295670808694895e-24, 1.8238994721423783e-24], [57.93032910437141, 1.9544078801142578e-24, 1.5277964941478872e-24, 1.821093344713717e-24], [58.090956002104285, 1.952033308461497e-24, 1.5260373191669058e-24, 1.81830023669837e-24], [58.2520282796697, 1.9496689539082962e-24, 1.524289472056935e-24, 1.815520056677578e-24], [58.41354717199899, 1.947314751264864e-24, 1.5225528695582109e-24, 1.8127527138697992e-24], [58.575513917447765, 1.9449706358483247e-24, 1.520827429018804e-24, 1.809998118133177e-24], [58.73792975780518, 1.942636543480061e-24, 1.519113068391825e-24, 1.8072561799637467e-24], [58.900795938303645, 1.9403124104820213e-24, 1.5174097062328023e-24, 1.8045268104931324e-24], [59.06411370762824, 1.9379981736744233e-24, 1.5157172616969147e-24, 1.8018099214898435e-24], [59.22788431792643, 1.9356937703720104e-24, 1.5140356545361381e-24, 1.799105425354352e-24], [59.39210902481753, 1.933399138381049e-24, 1.512364805097413e-24, 1.796413235122774e-24], [59.556789087402386, 1.931114215996612e-24, 1.5107046343183796e-24, 1.793733264461121e-24], [59.72192576827304, 1.9288389419989678e-24, 1.5090550637263044e-24, 1.7910654276668702e-24], [59.887520333522374, 1.9265732556508316e-24, 1.5074160154338865e-24, 1.788409639665943e-24], [60.053574052753916, 1.924317096694185e-24, 1.5057874121372314e-24, 1.785765816012119e-24], [60.22008819909143, 1.9220704053470734e-24, 1.5041691771132329e-24, 1.783133872884082e-24], [60.38706404918875, 1.919833122300633e-24, 1.502561234215684e-24, 1.7805137270867356e-24], [60.5545028832396, 1.9176051887158627e-24, 1.5009635078734984e-24, 1.7779052960436844e-24], [60.722405984987276, 1.9153865462207708e-24, 1.4993759230875917e-24, 1.7753084978019335e-24], [60.890774641734794, 1.9131771369071648e-24, 1.497798405427524e-24, 1.7727232510259214e-24], [61.059610144354366, 1.910976903327222e-24, 1.4962308810290734e-24, 1.7701494749975033e-24], [61.22891378729753, 1.908785788491136e-24, 1.4946732765911912e-24, 1.767587089612812e-24], [61.39868686860504, 1.9066037358636854e-24, 1.4931255193730854e-24, 1.7650360153799887e-24], [61.568930689916776, 1.9044306893607118e-24, 1.4915875371907998e-24, 1.762496173418175e-24], [61.73964655648184, 1.9022665933470463e-24, 1.4900592584152849e-24, 1.7599674854553093e-24], [61.91083577716837, 1.9001113926322925e-24, 1.4885406119686654e-24, 1.7574498738249837e-24], [62.08249966447375, 1.8979650324688578e-24, 1.487031527321086e-24, 1.754943261463901e-24], [62.25463953453456, 1.8958274585481835e-24, 1.4855319344883387e-24, 1.7524475719105347e-24], [62.42725670713672, 1.8936986169979573e-24, 1.4840417640287202e-24, 1.7499627293036947e-24], [62.60035250572566, 1.8915784543792467e-24, 1.4825609470396738e-24, 1.747488658375558e-24], [62.77392825741633, 1.889466917683169e-24, 1.4810894151547831e-24, 1.7450252844563152e-24], [62.947985293003484, 1.8873639543280126e-24, 1.4796271005414313e-24, 1.742572533464644e-24], [63.122524946971836, 1.885269512156205e-24, 1.4781739358966974e-24, 1.7401303319103543e-24], [63.29754855750626, 1.8831835394315297e-24, 1.4767298544454069e-24, 1.737698606886551e-24], [63.473057466502276, 1.881105984835689e-24, 1.4752947899359795e-24, 1.7352772860735425e-24], [63.649053019575966, 1.879036797465606e-24, 1.4738686766385533e-24, 1.7328662977297067e-24], [63.825536566074554, 1.8769759268306267e-24, 1.4724514493408674e-24, 1.7304655706937447e-24], [64.00250945908671, 1.8749233228492227e-24, 1.471043043345692e-24, 1.728075034378903e-24], [64.17997305545272, 1.8728789358458704e-24, 1.4696433944679799e-24, 1.725694618770289e-24], [64.35792871577537, 1.8708427165489905e-24, 1.4682524390311577e-24, 1.7233242544233758e-24], [64.53637780442985, 1.868814616086915e-24, 1.466870113864432e-24, 1.720963872460648e-24], [64.71532168957445, 1.8667945859856603e-24, 1.4654963562999988e-24, 1.7186134045683212e-24], [64.89476174316108, 1.864782578165625e-24, 1.4641311041693297e-24, 1.7162727829926383e-24], [65.0746993409458, 1.8627785449391263e-24, 1.4627742958003514e-24, 1.713941940539896e-24], [65.25513586249919, 1.860782439007098e-24, 1.4614258700145726e-24, 1.7116208105696353e-24], [65.4360726912171, 1.8587942134563907e-24, 1.4600857661237876e-24, 1.7093093269932618e-24], [65.61751121433124, 1.856813821756698e-24, 1.4587539239268557e-24, 1.7070074242714125e-24], [65.79945282291972, 1.854841217758239e-24, 1.457430283706962e-24, 1.7047150374114985e-24], [65.98189891191782, 1.8528763556880575e-24, 1.4561147862283522e-24, 1.702432101962096e-24], [66.16485088012865, 1.8509191901480305e-24, 1.454807372732995e-24, 1.7001585540111153e-24], [66.34831013023383, 1.8489696761115064e-24, 1.453507984937845e-24, 1.6978943301843657e-24], [66.53227806880426, 1.847027768920552e-24, 1.4522165650319964e-24, 1.6956393676400917e-24], [66.71675610631091, 1.8450934242834346e-24, 1.4509330556726607e-24, 1.6933936040655805e-24], [66.90174565713583, 1.8431665982716e-24, 1.4496573999831233e-24, 1.6911569776758418e-24], [67.08724813958261, 1.8412472473168253e-24, 1.448389541549186e-24, 1.6889294272101814e-24], [67.27326497588751, 1.8393353282087176e-24, 1.4471294244158309e-24, 1.6867108919257024e-24], [67.45979759223033, 1.8374307980916698e-24, 1.4458769930848026e-24, 1.684501311598758e-24], [67.64684741874527, 1.8355336144623966e-24, 1.4446321925111872e-24, 1.6823006265191693e-24], [67.83441588953205, 1.833643735167131e-24, 1.4433949681001968e-24, 1.68010877748609e-24], [68.02250444266673, 1.8317611183986536e-24, 1.4421652657046342e-24, 1.6779257058067628e-24], [68.21111452021283, 1.8298857226940364e-24, 1.4409430316212346e-24, 1.6757513532918736e-24], [68.40024756823236, 1.8280175069315746e-24, 1.4397282125878842e-24, 1.673585662252931e-24], [68.58990503679686, 1.826156430328341e-24, 1.4385207557807821e-24, 1.671428575498998e-24], [68.78008837999872, 1.82430245243752e-24, 1.4373206088112924e-24, 1.6692800363312573e-24], [68.97079905596205, 1.8224555331455325e-24, 1.4361277197227287e-24, 1.667139988544082e-24], [69.16203852685399, 1.8206156326697727e-24, 1.4349420369875483e-24, 1.66500837641684e-24], [69.353808258896, 1.818782711555782e-24, 1.4337635095043888e-24, 1.662885144713258e-24], [69.54610972237487, 1.816956730674515e-24, 1.4325920865949906e-24, 1.66077023867845e-24], [69.73894439165439, 1.8151376512200835e-24, 1.4314277180009701e-24, 1.6586636040334845e-24], [69.93231374518619, 1.8133254347069545e-24, 1.4302703538813688e-24, 1.656565186973931e-24], [70.12621926552141, 1.811520042967595e-24, 1.4291199448093649e-24, 1.6544749341648945e-24], [70.32066243932182, 1.8097214381494442e-24, 1.42797644176915e-24, 1.6523927927396475e-24], [70.51564475737139, 1.8079295827131323e-24, 1.4268397961533e-24, 1.6503187102941027e-24], [70.71116771458786, 1.806144439429356e-24, 1.4257099597598432e-24, 1.6482526348845708e-24], [70.9072328100338, 1.804365971376725e-24, 1.4245868847890534e-24, 1.6461945150252907e-24], [71.1038415469284, 1.802594141939027e-24, 1.4234705238405487e-24, 1.6441442996824429e-24], [71.30099543265885, 1.8008289148029177e-24, 1.422360829910813e-24, 1.6421019382735323e-24], [71.49869597879211, 1.7990702539556476e-24, 1.4212577563898958e-24, 1.6400673806622204e-24], [71.69694470108624, 1.79731812368226e-24, 1.4201612570584544e-24, 1.6380405771564028e-24], [71.89574311950213, 1.7955724885632477e-24, 1.4190712860852362e-24, 1.6360214785029059e-24], [72.0950927582152, 1.7938333134724017e-24, 1.4179877980240814e-24, 1.63401003588638e-24], [72.29499514562698, 1.792100563574182e-24, 1.4169107478108822e-24, 1.632006200924615e-24], [72.495451814377, 1.790374204321464e-24, 1.4158400907609095e-24, 1.6300099256648822e-24], [72.69646430135437, 1.7886542014531284e-24, 1.4147757825660596e-24, 1.6280211625819179e-24], [72.8980341477096, 1.7869405209917364e-24, 1.4137177792918166e-24, 1.6260398645735062e-24], [73.10016289886644, 1.7852331292410437e-24, 1.4126660373747499e-24, 1.62406598495771e-24], [73.30285210453371, 1.7835319927841365e-24, 1.4116205136193525e-24, 1.6220994774687171e-24], [73.50610331871738, 1.781837078480454e-24, 1.4105811651954909e-24, 1.6201402962556098e-24], [73.70991809973209, 1.7801483534641912e-24, 1.4095479496358225e-24, 1.6181883958762017e-24], [73.9142980102134, 1.778465785141725e-24, 1.408520824832672e-24, 1.6162437312961926e-24], [74.11924461712968, 1.7767893411891295e-24, 1.4074997490352954e-24, 1.6143062578844046e-24], [74.32475949179407, 1.775118989550274e-24, 1.4064846808476627e-24, 1.6123759314101094e-24], [74.5308442098767, 1.7734546984347103e-24, 1.4054755792251173e-24, 1.6104527080399647e-24], [74.73750035141659, 1.771796436314958e-24, 1.4044724034722302e-24, 1.6085365443343957e-24], [74.9447295008338, 1.770144171924997e-24, 1.4034751132395086e-24, 1.6066273972442838e-24], [75.15253324694164, 1.7684978742571703e-24, 1.402483668521473e-24, 1.6047252241086536e-24], [75.36091318295878, 1.766857512561254e-24, 1.4014980296532347e-24, 1.6028299826498531e-24], [75.56987090652161, 1.765223056341127e-24, 1.4005181573084538e-24, 1.600941630971796e-24], [75.77940801969629, 1.763594475353401e-24, 1.3995440124964014e-24, 1.5990601275563416e-24], [75.98952612899112, 1.7619717396049137e-24, 1.3985755565594096e-24, 1.5971854312599444e-24], [76.20022684536889, 1.7603548193510677e-24, 1.397612751170336e-24, 1.5953175013105318e-24], [76.41151178425912, 1.7587436850932386e-24, 1.3966555583300102e-24, 1.5934562973046844e-24], [76.62338256557078, 1.757138307576846e-24, 1.3957039403644375e-24, 1.5916017792042582e-24], [76.83584081370424, 1.7555386577898223e-24, 1.3947578599224867e-24, 1.5897539073332325e-24], [77.04888815756391, 1.7539447069597804e-24, 1.3938172799733448e-24, 1.5879126423746619e-24], [77.26252623057078, 1.7523564265524843e-24, 1.3928821638037888e-24, 1.5860779453679382e-24], [77.47675667067517, 1.75077378826994e-24, 1.3919524750158705e-24, 1.5842497777047987e-24], [77.69158112036877, 1.7491967640478317e-24, 1.3910281775243287e-24, 1.5824281011279524e-24], [77.90700122669759, 1.747625326054217e-24, 1.3901092355543929e-24, 1.5806128777260849e-24], [78.1230186412745, 1.7460594466873382e-24, 1.38919561363875e-24, 1.5788040699322237e-24], [78.33963502029187, 1.744499098573347e-24, 1.38828727661553e-24, 1.5770016405201718e-24], [78.5568520245344, 1.7429442545649253e-24, 1.3873841896256684e-24, 1.5752055526015858e-24], [78.77467131939162, 1.7413948877388884e-24, 1.3864863181110046e-24, 1.5734157696234716e-24], [78.99309457487087, 1.7398509713947293e-24, 1.3855936278110576e-24, 1.5716322553646964e-24], [79.21212346560996, 1.738312479052366e-24, 1.384706084761187e-24, 1.5698549739326888e-24], [79.43175967089007, 1.736779384450485e-24, 1.3838236552901563e-24, 1.5680838897616746e-24], [79.65200487464874, 1.7352516615446295e-24, 1.3829463060174913e-24, 1.566318967609065e-24], [79.8728607654925, 1.733729284505324e-24, 1.3820740038516304e-24, 1.564560172552718e-24], [80.09432903671009, 1.7322122277163703e-24, 1.3812067159872806e-24, 1.5628074699875002e-24], [80.3164113862853, 1.730700465772901e-24, 1.3803444099028882e-24, 1.5610608256239558e-24], [80.53910951690989, 1.7291939734797266e-24, 1.3794870533591763e-24, 1.559320205483592e-24], [80.76242513599712, 1.7276927258494933e-24, 1.3786346143956509e-24, 1.5575855758976955e-24], [80.98635995569418, 1.7261966981008324e-24, 1.3777870613296094e-24, 1.5558569035038196e-24], [81.21091569289575, 1.724705865656731e-24, 1.3769443627529197e-24, 1.5541341552422902e-24], [81.43609406925701, 1.7232202041428132e-24, 1.3761064875303689e-24, 1.5524172983549686e-24], [81.66189681120687, 1.7217396893855376e-24, 1.3752734047970378e-24, 1.5507063003819875e-24], [81.88832564996133, 1.7202642974105186e-24, 1.3744450839564227e-24, 1.5490011291580226e-24], [82.11538232153652, 1.7187940044408538e-24, 1.373621494678243e-24, 1.5473017528110586e-24], [82.3430685667621, 1.7173287868955307e-24, 1.3728026068961269e-24, 1.5456081397587704e-24], [82.57138613129474, 1.715868621387506e-24, 1.3719883908053485e-24, 1.5439202587065238e-24], [82.8003367656313, 1.7144134847224024e-24, 1.3711788168610648e-24, 1.5422380786438844e-24], [83.02992222512245, 1.7129633538965766e-24, 1.3703738557758803e-24, 1.5405615688429206e-24], [83.26014426998594, 1.7115182060957577e-24, 1.3695734785177956e-24, 1.538890698854976e-24], [83.49100466532022, 1.7100780186931043e-24, 1.3687776563083304e-24, 1.53722543850858e-24], [83.7225051811179, 1.708642769248156e-24, 1.36798636062008e-24, 1.535565757906392e-24], [83.9546475922793, 1.707212435504553e-24, 1.3671995631750848e-24, 1.533911627422757e-24], [84.18743367862642, 1.7057869953890452e-24, 1.36641723594246e-24, 1.532263017701712e-24], [84.42086522491587, 1.704366427009744e-24, 1.365639351136387e-24, 1.530619899654023e-24], [84.65494402085318, 1.7029507086545022e-24, 1.364865881214407e-24, 1.5289822444545878e-24], [84.8896718611062, 1.701539818789513e-24, 1.364096798875156e-24, 1.5273500235401628e-24], [85.12505054531917, 1.70013373605784e-24, 1.3633320770561486e-24, 1.525723208607125e-24], [85.3610818781261, 1.6987324392776357e-24, 1.36257168893266e-24, 1.524101771608929e-24], [85.59776766916488, 1.697335907441036e-24, 1.3618156079147764e-24, 1.5224856847532067e-24], [85.83510973309112, 1.6959441197124716e-24, 1.3610638076459077e-24, 1.520874920500115e-24], [86.07310988959195, 1.694557055427181e-24, 1.3603162620012698e-24, 1.5192694515597616e-24], [86.31176996340021, 1.6931746940899695e-24, 1.3595729450850815e-24, 1.5176692508894366e-24], [86.55109178430814, 1.691797015373572e-24, 1.3588338312294729e-24, 1.5160742916919876e-24], [86.79107718718166, 1.6904239991173247e-24, 1.3580988949922106e-24, 1.5144845474127908e-24], [87.03172801197421, 1.6890556253257518e-24, 1.3573681111548934e-24, 1.5128999917380147e-24], [87.27304610374105, 1.6876918741671807e-24, 1.3566414547213521e-24, 1.5113205985920772e-24], [87.51503331265343, 1.6863327259723545e-24, 1.3559189009153175e-24, 1.5097463421355275e-24], [87.75769149401253, 1.6849781612330847e-24, 1.3552004251790717e-24, 1.5081771967625634e-24], [88.0010225082639, 1.6836281606008275e-24, 1.3544860031715642e-24, 1.5066131370992727e-24], [88.24502822101164, 1.682282704885454e-24, 1.3537756107662734e-24, 1.505054138000915e-24], [88.48971050303265, 1.680941775053818e-24, 1.3530692240499964e-24, 1.5035001745501497e-24], [88.73507123029133, 1.679605352228487e-24, 1.3523668193206569e-24, 1.5019512220547976e-24], [88.98111228395335, 1.678273417686473e-24, 1.3516683730856545e-24, 1.5004072560456378e-24], [89.22783555040056, 1.6769459528578054e-24, 1.3509738620602306e-24, 1.498868252274096e-24], [89.47524292124524, 1.6756229393244426e-24, 1.3502832631657822e-24, 1.4973341867106727e-24], [89.72333629334466, 1.674304358818857e-24, 1.3495965535280167e-24, 1.4958050355421583e-24], [89.97211756881572, 1.672990193222851e-24, 1.3489137104753475e-24, 1.4942807751702915e-24], [90.2215886550493, 1.6716804245661876e-24, 1.3482347115372893e-24, 1.4927613822092894e-24], [90.47175146472509, 1.6703750350256954e-24, 1.347559534442687e-24, 1.4912468334837036e-24], [90.72260791582613, 1.6690740069234613e-24, 1.346888157118015e-24, 1.4897371060270248e-24], [90.97415993165359, 1.6677773227261763e-24, 1.3462205576859784e-24, 1.488232177078927e-24], [91.22640944084158, 1.6664849650437955e-24, 1.3455567144637788e-24, 1.4867320240839228e-24], [91.47935837737171, 1.6651969166281213e-24, 1.3448966059613689e-24, 1.4852366246888752e-24], [91.73300868058814, 1.6639131603718495e-24, 1.3442402108801834e-24, 1.4837459567416945e-24], [91.98736229521234, 1.6626336793073644e-24, 1.3435875081112805e-24, 1.4822599982888127e-24], [92.24242117135796, 1.6613584566055414e-24, 1.3429384767338472e-24, 1.4807787275735575e-24], [92.49818726454605, 1.6600874755745475e-24, 1.3422930960138827e-24, 1.4793021230344196e-24], [92.75466253571963, 1.6588207196590216e-24, 1.341651345402108e-24, 1.4778301633029448e-24], [93.01184895125905, 1.6575581724384178e-24, 1.341013204533153e-24, 1.4763628272018351e-24], [93.26974848299686, 1.6562998176264423e-24, 1.3403786532236473e-24, 1.4749000937435146e-24], [93.52836310823328, 1.6550456390695832e-24, 1.3397476714707368e-24, 1.473441942127939e-24], [93.78769480975085, 1.6537956207461363e-24, 1.3391202394504838e-24, 1.4719883517408479e-24], [94.04774557582998, 1.652549746765271e-24, 1.3384963375169358e-24, 1.4705393021522815e-24], [94.30851740026411, 1.651308001365723e-24, 1.3378759462000492e-24, 1.4690947731144063e-24], [94.57001228237493, 1.650070368914876e-24, 1.3372590462045162e-24, 1.4676547445601543e-24], [94.83223222702793, 1.6488368339076067e-24, 1.3366456184085574e-24, 1.466219196601159e-24], [95.09517924464743, 1.6476073809654399e-24, 1.3360356438618856e-24, 1.4647881095262935e-24], [95.35885535123228, 1.6463819948354545e-24, 1.335429103785071e-24, 1.4633614637998973e-24], [95.62326256837116, 1.645160660389022e-24, 1.3348259795675995e-24, 1.4619392400601635e-24], [95.88840292325813, 1.6439433626211987e-24, 1.3342262527666688e-24, 1.4605214191173255e-24], [96.15427844870824, 1.642730086649428e-24, 1.333629905105868e-24, 1.4591079819521996e-24], [96.42089118317298, 1.6415208177127206e-24, 1.333036918473764e-24, 1.4576989097143918e-24], [96.68824317075601, 1.6403155411705323e-24, 1.3324472749226775e-24, 1.456294183720699e-24], [96.95633646122876, 1.639114242502002e-24, 1.331860956666976e-24, 1.4548937854538504e-24], [97.22517311004613, 1.6379169073047166e-24, 1.331277946082394e-24, 1.4534976965604074e-24], [97.49475517836257, 1.6367235212940848e-24, 1.3306982257041905e-24, 1.4521058988496035e-24], [97.76508473304733, 1.6355340703021594e-24, 1.3301217782259177e-24, 1.4507183742915884e-24], [98.03616384670072, 1.6343485402767372e-24, 1.3295485864985526e-24, 1.449335105015933e-24], [98.30799459766979, 1.6331669172805964e-24, 1.3289786335288622e-24, 1.4479560733102191e-24], [98.5805790700644, 1.63198918749041e-24, 1.3284119024779523e-24, 1.4465812616185766e-24], [98.85391935377325, 1.6308153371959632e-24, 1.3278483766606877e-24, 1.4452106525398433e-24], [99.1280175444796, 1.6296453527991487e-24, 1.3272880395437947e-24, 1.4438442288264577e-24], [99.4028757436777, 1.6284792208132245e-24, 1.3267308747450307e-24, 1.442481973382927e-24], [99.67849605868862, 1.627316927861718e-24, 1.3261768660318115e-24, 1.441123869264356e-24], [99.95488060267655, 1.6261584606779249e-24, 1.3256259973201237e-24, 1.439769899674988e-24], [100.23203149466514, 1.6250038061037232e-24, 1.3250782526730899e-24, 1.438420047966797e-24], [100.5099508595533, 1.6238529510887848e-24, 1.3245336163002254e-24, 1.4370742976382308e-24], [100.78864082813209, 1.622705882689882e-24, 1.3239920725558806e-24, 1.4357326323326253e-24], [101.0681035371004, 1.6215625880699557e-24, 1.3234536059381863e-24, 1.4343950358369456e-24], [101.34834112908193, 1.620423054497274e-24, 1.3229182010880056e-24, 1.4330614920805553e-24], [101.62935575264139, 1.6192872693446915e-24, 1.3223858427877484e-24, 1.4317319851335975e-24], [101.91114956230064, 1.618155220088822e-24, 1.3218565159602714e-24, 1.430406499205865e-24], [102.1937247185559, 1.617026894309137e-24, 1.3213302056676094e-24, 1.429085018645512e-24], [102.47708338789354, 1.61590227968735e-24, 1.3208068971102335e-24, 1.4277675279377342e-24], [102.76122774280736, 1.614781364006503e-24, 1.3202865756255158e-24, 1.4264540117032753e-24], [103.04615996181502, 1.6136641351503693e-24, 1.3197692266868642e-24, 1.4251444546974928e-24], [103.33188222947429, 1.6125505811024675e-24, 1.3192548359029196e-24, 1.423838841808918e-24], [103.61839673640068, 1.6114406899452991e-24, 1.3187433890159487e-24, 1.4225371580580677e-24], [103.9057056792832, 1.6103344498599639e-24, 1.318234871901216e-24, 1.4212393885960994e-24], [104.19381126090249, 1.6092318491249993e-24, 1.3177292705657586e-24, 1.4199455187038453e-24], [104.48271569014622, 1.608132876115841e-24, 1.3172265711474625e-24, 1.4186555337903026e-24], [104.77242118202736, 1.6070375193041089e-24, 1.316726759913993e-24, 1.4173694193916545e-24], [105.06292995770018, 1.6059457672567775e-24, 1.3162298232616457e-24, 1.4160871611700905e-24], [105.35424424447763, 1.604857608635752e-24, 1.3157357477146354e-24, 1.4148087449125819e-24], [105.64636627584906, 1.6037730321966939e-24, 1.315244519923991e-24, 1.4135341565296449e-24], [105.93929829149607, 1.6026920267888063e-24, 1.3147561266663371e-24, 1.4122633820544236e-24], [106.23304253731082, 1.6016145813537246e-24, 1.3142705548433395e-24, 1.4109964076413973e-24], [106.5276012654124, 1.6005406849252946e-24, 1.3137877914804454e-24, 1.4097332195652346e-24], [106.82297673416481, 1.5994703266284669e-24, 1.313307823725944e-24, 1.4084738042199257e-24], [107.11917120819393, 1.598403495679004e-24, 1.312830638850283e-24, 1.4072181481174549e-24], [107.41618695840461, 1.597340181382391e-24, 1.3123562242448356e-24, 1.4059662378867165e-24], [107.71402626199877, 1.5962803731336579e-24, 1.3118845674211786e-24, 1.4047180602726762e-24], [108.01269140249205, 1.5952240604164269e-24, 1.311415656009901e-24, 1.4034736021351507e-24], [108.31218466973206, 1.5941712328023487e-24, 1.3109494777601369e-24, 1.4022328504476811e-24], [108.61250835991562, 1.5931218799505237e-24, 1.3104860205383118e-24, 1.400995792296825e-24], [108.913664775606, 1.5920759916067868e-24, 1.3100252723274893e-24, 1.3997624148807696e-24], [109.21565622575136, 1.5910335576032715e-24, 1.3095672212261377e-24, 1.3985327055085175e-24], [109.51848502570161, 1.5899945678575423e-24, 1.3091118554477957e-24, 1.3973066515988296e-24], [109.82215349722686, 1.5889590123721696e-24, 1.3086591633197358e-24, 1.3960842406794197e-24], [110.12666396853487, 1.5879268812341322e-24, 1.3082091332823197e-24, 1.3948654603856208e-24], [110.43201877428903, 1.5868981646140918e-24, 1.3077617538881973e-24, 1.3936502984597897e-24], [110.73822025562588, 1.5858728527661278e-24, 1.3073170138013328e-24, 1.3924387427500805e-24], [111.04527076017379, 1.5848509360266525e-24, 1.3068749017963025e-24, 1.3912307812097232e-24], [111.35317264207015, 1.5838324048142846e-24, 1.3064354067574036e-24, 1.3900264018959067e-24], [111.66192826198007, 1.582817249629173e-24, 1.3059985176778755e-24, 1.3888255929689875e-24], [111.97153998711427, 1.5818054610523112e-24, 1.3055642236591157e-24, 1.3876283426914997e-24], [112.28201019124681, 1.5807970297450192e-24, 1.305132513909854e-24, 1.386434639427283e-24], [112.59334125473401, 1.579791946448629e-24, 1.3047033777453723e-24, 1.3852444716406415e-24], [112.90553556453236, 1.5787902019836204e-24, 1.3042768045867052e-24, 1.3840578278954157e-24], [113.21859551421645, 1.5777917872492684e-24, 1.3038527839599818e-24, 1.382874696854015e-24], [113.5325235039981, 1.5767966932231427e-24, 1.3034313054955239e-24, 1.3816950672768174e-24], [113.84732194074395, 1.5758049109604006e-24, 1.3030123589271046e-24, 1.3805189280210364e-24], [114.16299323799461, 1.5748164315935621e-24, 1.3025959340913463e-24, 1.3793462680400868e-24], [114.47953981598275, 1.5738312463317347e-24, 1.302182020926886e-24, 1.378177076382663e-24], [114.79696410165185, 1.5728493464602263e-24, 1.3017706094735352e-24, 1.3770113421918627e-24], [115.11526852867458, 1.5718707233400306e-24, 1.3013616898718195e-24, 1.3758490547045683e-24], [115.43445553747148, 1.5708953684073717e-24, 1.3009552523620592e-24, 1.3746902032503904e-24], [115.75452757523011, 1.5699232731730988e-24, 1.300551287283587e-24, 1.3735347772511276e-24], [116.07548709592335, 1.568954429222478e-24, 1.3001497850744084e-24, 1.3723827662197625e-24], [116.39733656032847, 1.5679888282143221e-24, 1.299750736270123e-24, 1.3712341597598668e-24], [116.72007843604553, 1.567026461880808e-24, 1.299354131503468e-24, 1.3700889475646755e-24], [117.04371519751714, 1.5660673220270157e-24, 1.2989599615035554e-24, 1.3689471194164905e-24], [117.36824932604647, 1.5651114005302679e-24, 1.2985682170951352e-24, 1.3678086651858207e-24], [117.69368330981708, 1.5641586893398279e-24, 1.2981788891981171e-24, 1.3666735748306663e-24], [118.0200196439117, 1.563209180476467e-24, 1.2977919688268087e-24, 1.3655418383958113e-24], [118.34726083033092, 1.5622628660319206e-24, 1.2974074470891579e-24, 1.3644134460121131e-24], [118.67540937801333, 1.5613197381685383e-24, 1.2970253151862814e-24, 1.3632883878957284e-24], [119.00446780285364, 1.5603797891188168e-24, 1.2966455644118476e-24, 1.3621666543474825e-24], [119.33443862772297, 1.559443011184964e-24, 1.2962681861512068e-24, 1.3610482357521337e-24], [119.6653243824877, 1.5585093967384566e-24, 1.2958931718810339e-24, 1.3599331225776607e-24], [119.99712760402858, 1.557578938219689e-24, 1.295520513168509e-24, 1.3588213053746508e-24], [120.329850836261, 1.5566516281374542e-24, 1.295150201670986e-24, 1.3577127747755597e-24], [120.66349663015343, 1.555727459068729e-24, 1.29478222913497e-24, 1.3566075214941237e-24], [120.9980675437484, 1.5548064236580277e-24, 1.2944165873959352e-24, 1.355505536324617e-24], [121.33356614218042, 1.5538885146171878e-24, 1.294053268377511e-24, 1.3544068101412387e-24], [121.6699949976971, 1.552973724724894e-24, 1.2936922640908708e-24, 1.3533113338975275e-24], [122.00735668967783, 1.552062046826276e-24, 1.2933335666343587e-24, 1.352219098625651e-24], [122.34565380465433, 1.551153473832614e-24, 1.292977168192648e-24, 1.3511300954357983e-24], [122.68488893633017, 1.5502479987208432e-24, 1.2926230610365072e-24, 1.350044315515617e-24], [123.02506468560024, 1.5493456145333267e-24, 1.292271237521836e-24, 1.3489617501295457e-24], [123.36618366057162, 1.5484463143772964e-24, 1.2919216900894875e-24, 1.3478823906182293e-24], [123.70824847658253, 1.5475500914245701e-24, 1.2915744112645094e-24, 1.3468062283978751e-24], [124.05126175622327, 1.5466569389113762e-24, 1.2912293936556106e-24, 1.3457332549597984e-24], [124.39522612935595, 1.5457668501376357e-24, 1.2908866299548232e-24, 1.3446634618697012e-24], [124.74014423313427, 1.544879818466929e-24, 1.2905461129366518e-24, 1.3435968407671735e-24], [125.08601871202467, 1.543995837325928e-24, 1.2902078354578271e-24, 1.3425333833651064e-24], [125.43285221782538, 1.5431149002042662e-24, 1.289871790456637e-24, 1.3414730814491188e-24], [125.78064740968834, 1.5422370006539762e-24, 1.2895379709525855e-24, 1.340415926877038e-24], [126.12940695413792, 1.5413621322893432e-24, 1.2892063700455492e-24, 1.3393619115783456e-24], [126.47913352509224, 1.5404902887864436e-24, 1.2888769809157264e-24, 1.3383110275535844e-24], [126.82982980388394, 1.5396214638827953e-24, 1.2885497968227927e-24, 1.3372632668738835e-24], [127.18149847927971, 1.5387556513771964e-24, 1.2882248111054965e-24, 1.3362186216804627e-24], [127.53414224750246, 1.5378928451294171e-24, 1.2879020171813441e-24, 1.335177084184017e-24], [127.88776381225013, 1.537033039059554e-24, 1.2875814085458566e-24, 1.3341386466642637e-24], [128.24236588471808, 1.5361762271480756e-24, 1.2872629787723075e-24, 1.3331033014694463e-24], [128.5979511836186, 1.5353224034354673e-24, 1.286946721511187e-24, 1.3320710410158235e-24], [128.9545224352027, 1.5344715620217128e-24, 1.2866326304896695e-24, 1.331041857787091e-24], [129.31208237328067, 1.5336236970662557e-24, 1.2863206995113182e-24, 1.3300157443340894e-24], [129.67063373924256, 1.5327788027874913e-24, 1.2860109224554352e-24, 1.3289926932740906e-24], [130.03017928208018, 1.5319368734625549e-24, 1.285703293276815e-24, 1.3279726972904849e-24], [130.39072175840715, 1.5310979034272253e-24, 1.2853978060051054e-24, 1.3269557491322792e-24], [130.75226393248087, 1.5302618870753035e-24, 1.2850944547445821e-24, 1.3259418416135887e-24], [131.11480857622334, 1.5294288188585865e-24, 1.284793233673478e-24, 1.3249309676132e-24], [131.4783584692421, 1.528598693286451e-24, 1.284494137043867e-24, 1.3239231200741198e-24], [131.84291639885234, 1.527771504925658e-24, 1.284197159180926e-24, 1.3229182920031668e-24], [132.2084851600973, 1.526947248400073e-24, 1.2839022944826905e-24, 1.3219164764704532e-24], [132.57506755577066, 1.5261259183903567e-24, 1.283609537419617e-24, 1.320917666609004e-24], [132.94266639643732, 1.5253075096338423e-24, 1.2833188825341989e-24, 1.3199218556143709e-24], [133.31128450045546, 1.5244920169240412e-24, 1.2830303244405222e-24, 1.3189290367440849e-24], [133.6809246939974, 1.5236794351106219e-24, 1.2827438578238676e-24, 1.3179392033173364e-24], [134.05158981107226, 1.5228697590989917e-24, 1.2824594774403935e-24, 1.3169523487145542e-24], [134.42328269354704, 1.5220629838502136e-24, 1.2821771781166157e-24, 1.3159684663769233e-24], [134.79600619116826, 1.5212591043805616e-24, 1.2818969547491381e-24, 1.314987549806107e-24], [135.16976316158463, 1.520458115761493e-24, 1.2816188023042614e-24, 1.3140095925637334e-24], [135.54455647036787, 1.5196600131192346e-24, 1.2813427158175318e-24, 1.3130345882710662e-24], [135.92038899103565, 1.5188647916346255e-24, 1.281068690393502e-24, 1.3120625306085769e-24], [136.29726360507323, 1.518072446542967e-24, 1.2807967212052227e-24, 1.311093413315644e-24], [136.6751832019551, 1.5172829731336416e-24, 1.2805268034939564e-24, 1.3101272301900788e-24], [137.05415067916817, 1.5164963667499726e-24, 1.28025893256888e-24, 1.3091639750878137e-24], [137.4341689422328, 1.5157126227890267e-24, 1.2799931038066332e-24, 1.3082036419224985e-24], [137.81524090472604, 1.5149317367013268e-24, 1.2797293126509467e-24, 1.3072462246651783e-24], [138.19736948830368, 1.5141537039907158e-24, 1.2794675546125383e-24, 1.3062917173439196e-24], [138.58055762272204, 1.5133785202141119e-24, 1.2792078252683427e-24, 1.3053401140434521e-24], [138.9648082458616, 1.5126061809812128e-24, 1.2789501202616341e-24, 1.3043914089048061e-24], [139.35012430374826, 1.5118366819544724e-24, 1.2786944353014532e-24, 1.3034455961250193e-24], [139.73650875057695, 1.5110700188487902e-24, 1.2784407661622028e-24, 1.3025026699567333e-24], [140.12396454873382, 1.5103061874313454e-24, 1.278189108683649e-24, 1.3015626247079537e-24], [140.51249466881907, 1.5095451835212694e-24, 1.277939458770235e-24, 1.3006254547415855e-24], [140.90210208966923, 1.5087870029897054e-24, 1.2776918123910068e-24, 1.2996911544752485e-24], [141.29278979838105, 1.5080316417594397e-24, 1.2774461655791451e-24, 1.2987597183808884e-24], [141.6845607903332, 1.5072790958046712e-24, 1.2772025144318816e-24, 1.2978311409844116e-24], [142.07741806921018, 1.5065293611510926e-24, 1.2769608551099472e-24, 1.2969054168655069e-24], [142.4713646470251, 1.5057824338753455e-24, 1.2767211838374046e-24, 1.2959825406571945e-24], [142.86640354414212, 1.5050383101051794e-24, 1.2764834969013486e-24, 1.2950625070456623e-24], [143.26253778930078, 1.504296986019022e-24, 1.2762477906516083e-24, 1.294145310769825e-24], [143.659770419638, 1.503558457845976e-24, 1.2760140615003982e-24, 1.2932309466211405e-24], [144.05810448071236, 1.502822721865575e-24, 1.275782305922133e-24, 1.2923194094432774e-24], [144.4575430265271, 1.5020897744076338e-24, 1.2755525204531308e-24, 1.291410694131842e-24], [144.858089119553, 1.5013596118520606e-24, 1.2753247016913008e-24, 1.2905047956340487e-24], [145.25974583075293, 1.5006322306287167e-24, 1.2750988462957817e-24, 1.2896017089485016e-24], [145.66251623960403, 1.4999076272172953e-24, 1.2748749509868405e-24, 1.2887014291249171e-24], [146.0664034341232, 1.4991857981471006e-24, 1.2746530125456307e-24, 1.2878039512638198e-24], [146.47141051088857, 1.498466739996959e-24, 1.2744330278136727e-24, 1.2869092705162824e-24], [146.87754057506513, 1.497750449394957e-24, 1.2742149936929015e-24, 1.286017382083693e-24], [147.2847967404271, 1.4970369230184749e-24, 1.2739989071451644e-24, 1.2851282812174603e-24], [147.69318212938293, 1.4963261575938681e-24, 1.2737847651921542e-24, 1.2842419632187859e-24], [148.10269987299876, 1.4956181498963728e-24, 1.2735725649150744e-24, 1.2833584234383918e-24], [148.51335311102193, 1.494912896750104e-24, 1.2733623034543562e-24, 1.2824776572763063e-24], [148.92514499190622, 1.494210395027671e-24, 1.2731539780095374e-24, 1.2815996601815748e-24], [149.33807867283465, 1.4935106416502688e-24, 1.2729475858389038e-24, 1.2807244276520512e-24], [149.75215731974487, 1.492813633587418e-24, 1.2727431242593624e-24, 1.2798519552341727e-24], [150.16738410735286, 1.4921193678568046e-24, 1.2725405906460535e-24, 1.278982238522667e-24], [150.58376221917683, 1.4914278415243674e-24, 1.272339982432366e-24, 1.2781152731604124e-24], [151.00129484756286, 1.490739051703918e-24, 1.272141297109442e-24, 1.2772510548381326e-24], [151.41998519370793, 1.4900529955570747e-24, 1.271944532226177e-24, 1.2763895792941984e-24], [151.83983646768567, 1.489369670293331e-24, 1.27174968538886e-24, 1.2755308423144461e-24], [152.26085188847043, 1.4886890731696613e-24, 1.2715567542610658e-24, 1.274674839731912e-24], [152.68303468396155, 1.4880112014906692e-24, 1.2713657365632685e-24, 1.2738215674266488e-24], [153.10638809100928, 1.4873360526082254e-24, 1.2711766300728901e-24, 1.2729710213254853e-24], [153.53091535543794, 1.4866636239215366e-24, 1.2709894326238533e-24, 1.2721231974018636e-24], [153.95661973207294, 1.4859939128770173e-24, 1.270804142106581e-24, 1.2712780916756214e-24], [154.38350448476342, 1.4853269169680545e-24, 1.2706207564675807e-24, 1.2704357002127524e-24], [154.8115728864092, 1.4846626337350854e-24, 1.2704392737094456e-24, 1.269596019125298e-24], [155.24082821898438, 1.4840010607653416e-24, 1.2702596918905546e-24, 1.2687590445710602e-24], [155.6712737735637, 1.483342195692818e-24, 1.2700820091249292e-24, 1.2679247727534504e-24], [156.10291285034725, 1.4826860361981987e-24, 1.2699062235819774e-24, 1.267093199921335e-24], [156.5357487586853, 1.4820325800087335e-24, 1.2697323334863866e-24, 1.2662643223687862e-24], [156.96978481710477, 1.4813818248980706e-24, 1.269560337117893e-24, 1.2654381364349682e-24], [157.4050243533335, 1.4807337686863821e-24, 1.2693902328111526e-24, 1.2646146385039155e-24], [157.84147070432672, 1.4800884092399613e-24, 1.2692220189554807e-24, 1.263793825004349e-24], [158.27912721629227, 1.4794457444714297e-24, 1.2690556939947782e-24, 1.2629756924095586e-24], [158.71799724471572, 1.4788057723395065e-24, 1.268891256427255e-24, 1.262160237237198e-24], [159.15808415438727, 1.4781684908489055e-24, 1.2687287048053891e-24, 1.2613474560490849e-24], [159.59939131942622, 1.4775338980503197e-24, 1.2685680377356513e-24, 1.2605373454511177e-24], [160.04192212330793, 1.4769019920403478e-24, 1.268409253878357e-24, 1.2597299020930463e-24], [160.48567995888925, 1.4762727709613586e-24, 1.2682523519475503e-24, 1.2589251226683454e-24], [160.93066822843468, 1.475646233001457e-24, 1.268097330710872e-24, 1.2581230039140614e-24], [161.37689034364197, 1.4750223763943761e-24, 1.2679441889892924e-24, 1.2573235426106363e-24], [161.82434972566915, 1.4744011994195263e-24, 1.2677929256570363e-24, 1.2565267355818006e-24], [162.2730498051603, 1.4737827004017219e-24, 1.2676435396414714e-24, 1.2557325796943716e-24], [162.72299402227134, 1.4731668777112828e-24, 1.2674960299228978e-24, 1.2549410718581504e-24], [163.17418582669757, 1.4725537297639257e-24, 1.2673503955343241e-24, 1.2541522090257845e-24], [163.62662867769887, 1.4719432550206182e-24, 1.2672066355615958e-24, 1.253365988192583e-24], [164.08032604412728, 1.471335451987701e-24, 1.2670647491429198e-24, 1.2525824063964264e-24], [164.5352814044531, 1.470730319216591e-24, 1.266924735468963e-24, 1.2518014607176076e-24], [164.99149824679108, 1.470127855303968e-24, 1.2667865937826348e-24, 1.2510231482787295e-24], [165.44898006892848, 1.4695280588915129e-24, 1.266650323378972e-24, 1.250247466244523e-24], [165.90773037835024, 1.4689309286659823e-24, 1.2665159236049304e-24, 1.2494744118217792e-24], [166.36775269226735, 1.4683364633591294e-24, 1.2663833938594612e-24, 1.2487039822591996e-24], [166.82905053764307, 1.4677446617475784e-24, 1.2662527335931276e-24, 1.2479361748472501e-24], [167.2916274512197, 1.467155522652909e-24, 1.2661239423081316e-24, 1.247170986918087e-24], [167.75548697954665, 1.4665690449415118e-24, 1.2659970195581901e-24, 1.2464084158454173e-24], [168.22063267900643, 1.4659852275245522e-24, 1.2658719649484265e-24, 1.245648459044373e-24], [168.68706811584303, 1.4654040693580094e-24, 1.2657487781351431e-24, 1.2448911139714452e-24], [169.15479686618863, 1.4648255694425004e-24, 1.2656274588258476e-24, 1.2441363781243087e-24], [169.62382251609134, 1.4642497268233462e-24, 1.2655080067790789e-24, 1.2433842490417537e-24], [170.094148661542, 1.4636765405905095e-24, 1.2653904218042676e-24, 1.2426347243035943e-24], [170.56577890850295, 1.4631060098785282e-24, 1.2652747037616722e-24, 1.2418878015305163e-24], [171.0387168729343, 1.4625381338665493e-24, 1.2651608525622968e-24, 1.2411434783840347e-24], [171.51296618082287, 1.4619729117781712e-24, 1.2650488681677218e-24, 1.2404017525663512e-24], [171.98853046820952, 1.4614103428815676e-24, 1.2649387505900587e-24, 1.2396626218202707e-24], [172.4654133812164, 1.4608504264893629e-24, 1.2648304998918375e-24, 1.238926083929132e-24], [172.9436185760762, 1.4602931619585592e-24, 1.2647241161858896e-24, 1.2381921367166516e-24], [173.42314971915874, 1.459738548690673e-24, 1.2646195996352162e-24, 1.2374607780469164e-24], [173.9040104870004, 1.45918658613159e-24, 1.2645169504530903e-24, 1.2367320058242527e-24], [174.38620456633163, 1.4586372737714873e-24, 1.2644161689026826e-24, 1.2360058179931074e-24], [174.86973565410483, 1.4580906111449664e-24, 1.2643172552972224e-24, 1.2352822125380337e-24], [175.35460745752383, 1.457546597830922e-24, 1.264220209999758e-24, 1.2345611874835587e-24], [175.84082369407076, 1.4570052334525995e-24, 1.2641250334230982e-24, 1.2338427408941396e-24], [176.3283880915367, 1.4564665176774651e-24, 1.2640317260298077e-24, 1.2331268708740399e-24], [176.81730438804777, 1.4559304502173236e-24, 1.2639402883320681e-24, 1.2324135755673007e-24], [177.3075763320959, 1.4553970308282369e-24, 1.2638507208915628e-24, 1.2317028531576506e-24], [177.7992076825659, 1.45486625931049e-24, 1.2637630243195004e-24, 1.2309947018684156e-24], [178.29220220876573, 1.4543381355086073e-24, 1.2636771992764593e-24, 1.230289119962472e-24], [178.7865636904546, 1.4538126593113445e-24, 1.2635932464723336e-24, 1.2295861057421583e-24], [179.2822959178717, 1.4532898306517344e-24, 1.2635111666662738e-24, 1.2288856575492517e-24], [179.77940269176617, 1.4527696495069432e-24, 1.263430960666668e-24, 1.2281877737648395e-24], [180.2778878234252, 1.4522521158983935e-24, 1.2633526293309533e-24, 1.2274924528093102e-24], [180.77775513470414, 1.451737229891757e-24, 1.2632761735656877e-24, 1.2267996931422803e-24], [181.27900845805556, 1.4512249915968205e-24, 1.2632015943263403e-24, 1.226109493262511e-24], [181.78165163655783, 1.4507154011676314e-24, 1.2631288926174175e-24, 1.2254218517078725e-24], [182.28568852394608, 1.4502084588024426e-24, 1.2630580694922278e-24, 1.2247367670553049e-24], [182.79112298464014, 1.4497041647437122e-24, 1.2629891260529105e-24, 1.2240542379207273e-24], [183.2979588937755, 1.4492025192780842e-24, 1.2629220634504028e-24, 1.2233742629590016e-24], [183.80620013723257, 1.448703522736457e-24, 1.2628568828842798e-24, 1.2226968408639017e-24], [184.3158506116656, 1.4482071754939264e-24, 1.2627935856028485e-24, 1.2220219703680422e-24], [184.82691422453433, 1.4477134779698095e-24, 1.2627321729030154e-24, 1.2213496502428306e-24], [185.33939489413183, 1.447222430627715e-24, 1.2626726461301678e-24, 1.2206798792984558e-24], [185.853296549617, 1.4467340339754578e-24, 1.2626150066783154e-24, 1.2200126563837977e-24], [186.3686231310421, 1.4462482885650939e-24, 1.2625592559898345e-24, 1.219347980386405e-24], [186.88537858938514, 1.445765194992995e-24, 1.262505395555595e-24, 1.2186858502324824e-24], [187.40356688657846, 1.4452847538998256e-24, 1.262453426914846e-24, 1.2180262648868134e-24], [187.92319199554044, 1.4448069659705096e-24, 1.2624033516551305e-24, 1.2173692233527436e-24], [188.4442579002054, 1.4443318319343183e-24, 1.2623551714123615e-24, 1.2167147246721315e-24], [188.96676859555365, 1.4438593525648794e-24, 1.2623088878706477e-24, 1.2160627679253467e-24], [189.49072808764336, 1.4433895286801331e-24, 1.2622645027624094e-24, 1.215413352231196e-24], [190.01614039363972, 1.4429223611424232e-24, 1.262222017868217e-24, 1.2147664767469203e-24], [190.5430095418472, 1.4424578508584978e-24, 1.2621814350168077e-24, 1.2141221406681654e-24], [191.07133957173977, 1.4419959987795235e-24, 1.2621427560850797e-24, 1.213480343228944e-24], [191.60113453399129, 1.4415368059010978e-24, 1.2621059829980438e-24, 1.2128410837016138e-24], [192.13239849050794, 1.4410802732632972e-24, 1.2620711177287585e-24, 1.2122043613968491e-24], [192.66513551445763, 1.4406264019506951e-24, 1.2620381622983617e-24, 1.2115701756636301e-24], [193.19934969030362, 1.4401751930924104e-24, 1.262007118776043e-24, 1.2109385258892208e-24], [193.73504511383288, 1.4397266478621026e-24, 1.2619779892789665e-24, 1.210309411499131e-24], [194.27222589219016, 1.4392807674779792e-24, 1.2619507759723066e-24, 1.2096828319571082e-24], [194.81089614390748, 1.4388375532029225e-24, 1.2619254810692155e-24, 1.2090587867651341e-24], [195.3510599989371, 1.438397006344437e-24, 1.2619021068307293e-24, 1.2084372754633965e-24], [195.8927215986827, 1.4379591282547066e-24, 1.2618806555659279e-24, 1.2078182976302633e-24], [196.43588509603052, 1.4375239203306132e-24, 1.26186112963174e-24, 1.2072018528822917e-24], [196.9805546553825, 1.437091384013839e-24, 1.2618435314330025e-24, 1.2065879408742116e-24], [197.52673445268667, 1.436661520790794e-24, 1.2618278634224656e-24, 1.2059765612989017e-24], [198.07442867547059, 1.4362343321927388e-24, 1.261814128100738e-24, 1.2053677138873856e-24], [198.6236415228728, 1.435809819795831e-24, 1.2618023280163452e-24, 1.2047613984088528e-24], [199.17437720567443, 1.4353879852210954e-24, 1.261792465765598e-24, 1.2041576146706112e-24], [199.72663994633297, 1.434968830134512e-24, 1.2617845439927703e-24, 1.203556362518104e-24], [200.2804339790129, 1.4345523562470237e-24, 1.2617785653898815e-24, 1.2029576418348947e-24], [200.83576354961963, 1.4341385653146584e-24, 1.2617745326968767e-24, 1.2023614525426902e-24], [201.39263291583148, 1.4337274591384969e-24, 1.2617724487015167e-24, 1.2017677946013165e-24], [201.95104634713164, 1.4333190395646885e-24, 1.261772316239387e-24, 1.2011766680087076e-24], [202.51100812484242, 1.4329133084846467e-24, 1.2617741381939614e-24, 1.2005880728009544e-24], [203.07252254215635, 1.432510267834916e-24, 1.2617779174965048e-24, 1.200002009052257e-24], [203.63559390417055, 1.432109919597339e-24, 1.2617836571261744e-24, 1.199418476874956e-24], [204.20022652791903, 1.4317122657990797e-24, 1.261791360109939e-24, 1.1988374764195361e-24], [204.76642474240595, 1.4313173085126257e-24, 1.261801029522652e-24, 1.198259007874614e-24], [205.3341928886383, 1.4309250498559284e-24, 1.26181266848702e-24, 1.197683071466981e-24], [205.9035353196604, 1.4305354919923645e-24, 1.2618262801736037e-24, 1.1971096674615763e-24], [206.47445640058564, 1.4301486371308491e-24, 1.2618418678008725e-24, 1.196538796161515e-24], [207.0469605086315, 1.4297644875258703e-24, 1.2618594346351566e-24, 1.1959704579080972e-24], [207.62105203315247, 1.4293830454775202e-24, 1.261878983990677e-24, 1.195404653080819e-24], [208.19673537567294, 1.4290043133316118e-24, 1.2619005192296e-24, 1.1948413820973856e-24], [208.77401494992256, 1.4286282934796772e-24, 1.2619240437619732e-24, 1.1942806454137312e-24], [209.3528951818684, 1.4282549883590597e-24, 1.2619495610458198e-24, 1.1937224435240288e-24], [209.93338050975024, 1.4278844004529448e-24, 1.2619770745871126e-24, 1.1931667769607016e-24], [210.51547538411404, 1.4275165322904436e-24, 1.2620065879397537e-24, 1.1926136462944534e-24], [211.09918426784546, 1.4271513864466461e-24, 1.262038104705691e-24, 1.1920630521342837e-24], [211.68451163620526, 1.4267889655427003e-24, 1.2620716285348328e-24, 1.1915149951275096e-24], [212.27146197686284, 1.4264292722458144e-24, 1.262107163125159e-24, 1.190969475959769e-24], [212.86003978993102, 1.426072309269385e-24, 1.2621447122226772e-24, 1.1904264953550659e-24], [213.4502495879998, 1.4257180793730717e-24, 1.2621842796214476e-24, 1.1898860540757983e-24], [214.04209589617219, 1.4253665853627771e-24, 1.2622258691636877e-24, 1.1893481529227507e-24], [214.63558325209755, 1.4250178300907698e-24, 1.2622694847396544e-24, 1.1888127927351368e-24], [215.23071620600757, 1.424671816455775e-24, 1.2623151302878176e-24, 1.1882799743906428e-24], [215.82749932075072, 1.4243285474029906e-24, 1.2623628097948045e-24, 1.1877496988054268e-24], [216.42593717182646, 1.4239880259242262e-24, 1.2624125272954203e-24, 1.1872219669341786e-24], [217.0260343474219, 1.4236502550578171e-24, 1.2624642868727091e-24, 1.1866967797700911e-24], [217.6277954484452, 1.4233152378888925e-24, 1.262518092657991e-24, 1.186174138344962e-24], [218.23122508856238, 1.4229829775493316e-24, 1.2625739488308796e-24, 1.1856540437291833e-24], [218.83632789423217, 1.4226534772178799e-24, 1.2626318596193014e-24, 1.1851364970317942e-24], [219.44310850474048, 1.4223267401201078e-24, 1.2626918292995009e-24, 1.1846214994004624e-24], [220.05157157223792, 1.422002769528705e-24, 1.262753862196186e-24, 1.1841090520216117e-24], [220.66172176177344, 1.4216815687633298e-24, 1.2628179626824449e-24, 1.183599156120365e-24], [221.27356375133166, 1.4213631411908197e-24, 1.2628841351798336e-24, 1.1830918129606343e-24], [221.88710223186825, 1.4210474902252047e-24, 1.262952384158403e-24, 1.1825870238451338e-24], [222.50234190734503, 1.4207346193278341e-24, 1.2630227141367435e-24, 1.18208479011544e-24], [223.11928749476766, 1.4204245320073913e-24, 1.2630951296820144e-24, 1.1815851131519966e-24], [223.73794372421986, 1.4201172318200241e-24, 1.2631696354099927e-24, 1.1810879943741893e-24], [224.35831533890223, 1.419812722369409e-24, 1.2632462359851264e-24, 1.1805934352403653e-24], [224.98040709516565, 1.4195110073067986e-24, 1.2633249361205092e-24, 1.1801014372478749e-24], [225.60422376255025, 1.419212090331129e-24, 1.263405740578043e-24, 1.1796120019331142e-24], [226.22977012382012, 1.4189159751891339e-24, 1.2634886541683439e-24, 1.1791251308715906e-24], [226.85705097500144, 1.418622665675354e-24, 1.2635736817509108e-24, 1.1786408256779307e-24], [227.48607112541868, 1.418332165632262e-24, 1.2636608282340496e-24, 1.1781590880059435e-24], [228.1168353977306, 1.4180444789503309e-24, 1.2637500985750307e-24, 1.1776799195486667e-24], [228.7493486279691, 1.41775960956815e-24, 1.2638414977800907e-24, 1.177203322038414e-24], [229.38361566557424, 1.4174775614724597e-24, 1.263935030904439e-24, 1.1767292972468131e-24], [230.0196413734331, 1.4171983386982678e-24, 1.2640307030523683e-24, 1.1762578469848641e-24], [230.65743062791643, 1.4169219453289188e-24, 1.264128519377288e-24, 1.1757889731029738e-24], [231.2969883189152, 1.4166483854961699e-24, 1.264228485081725e-24, 1.1753226774910131e-24], [231.93831934987992, 1.4163776633803403e-24, 1.2643306054174719e-24, 1.1748589620783861e-24], [232.58142863785594, 1.4161097832102973e-24, 1.2644348856855332e-24, 1.1743978288340351e-24], [233.22632111352402, 1.4158447492636358e-24, 1.264541331236241e-24, 1.173939279766542e-24], [233.87300172123497, 1.4155825658667137e-24, 1.264649947469305e-24, 1.1734833169241462e-24], [234.5214754190502, 1.4153232373947498e-24, 1.2647607398338195e-24, 1.1730299423948119e-24], [235.17174717877793, 1.4150667682719163e-24, 1.264873713828423e-24, 1.1725791583062776e-24], [235.82382198601178, 1.4148131629714599e-24, 1.2649888750011976e-24, 1.1721309668261202e-24], [236.47770484017067, 1.4145624260157094e-24, 1.2651062289498573e-24, 1.171685370161785e-24], [237.13340075453402, 1.4143145619762743e-24, 1.2652257813217815e-24, 1.1712423705606805e-24], [237.79091475628306, 1.4140695754740559e-24, 1.2653475378139734e-24, 1.1708019703101997e-24], [238.45025188653727, 1.4138274711793649e-24, 1.265471504173279e-24, 1.1703641717377912e-24], [239.1114172003946, 1.4135882538120015e-24, 1.2655976861963014e-24, 1.1699289772110134e-24], [239.77441576696964, 1.4133519281414032e-24, 1.2657260897295532e-24, 1.169496389137608e-24], [240.43925266943162, 1.4131184989866625e-24, 1.2658567206694623e-24, 1.1690664099655405e-24], [241.10593300504524, 1.4128879712166568e-24, 1.2659895849624786e-24, 1.1686390421830628e-24], [241.77446188520764, 1.4126603497501491e-24, 1.2661246886050935e-24, 1.1682142883187794e-24], [242.4448444354894, 1.4124356395558814e-24, 1.2662620376439082e-24, 1.167792150941712e-24], [243.11708579567303, 1.4122138456526502e-24, 1.2664016381757499e-24, 1.1673726326613486e-24], [243.79119111979182, 1.4119949731094318e-24, 1.2665434963476448e-24, 1.1669557361277173e-24], [244.46716557617077, 1.4117790270454494e-24, 1.2666876183569814e-24, 1.1665414640314442e-24], [245.14501434746455, 1.4115660126302978e-24, 1.2668340104514897e-24, 1.1661298191038158e-24], [245.82474263069878, 1.411355935084011e-24, 1.2669826789293362e-24, 1.1657208041168437e-24], [246.5063556373089, 1.4111487996772017e-24, 1.2671336301392179e-24, 1.1653144218833367e-24], [247.18985859318073, 1.4109446117311076e-24, 1.2672868704804313e-24, 1.1649106752569518e-24], [247.87525673868942, 1.4107433766177342e-24, 1.2674424064028517e-24, 1.1645095671322705e-24], [248.5625553287414, 1.4105450997599289e-24, 1.267600244407122e-24, 1.1641111004448606e-24], [249.2517596328128, 1.4103497866314914e-24, 1.2677603910446152e-24, 1.1637152781713484e-24], [249.94287493499138, 1.410157442757267e-24, 1.2679228529175881e-24, 1.1633221033294756e-24], [250.63590653401644, 1.4099680737132704e-24, 1.268087636679223e-24, 1.1629315789781883e-24], [251.3308597433187, 1.4097816851267337e-24, 1.2682547490336317e-24, 1.1625437082176708e-24], [252.02773989106274, 1.4095982826762748e-24, 1.268424196736051e-24, 1.1621584941894522e-24], [252.72655232018576, 1.4094178720919498e-24, 1.2685959865927979e-24, 1.1617759400764521e-24], [253.42730238844052, 1.4092404591553616e-24, 1.2687701254613936e-24, 1.1613960491030522e-24], [254.1299954684355, 1.4090660496997847e-24, 1.2689466202506696e-24, 1.1610188245351779e-24], [254.8346369476754, 1.4088946496102474e-24, 1.269125477920768e-24, 1.1606442696803593e-24], [255.54123222860386, 1.4087262648236404e-24, 1.2693067054832365e-24, 1.1602723878878065e-24], [256.249786728644, 1.4085609013288354e-24, 1.2694903100011407e-24, 1.159903182548484e-24], [256.96030588024047, 1.4083985651667407e-24, 1.2696762985890975e-24, 1.1595366570951715e-24], [257.6727951309001, 1.4082392624304792e-24, 1.2698646784133635e-24, 1.159172815002555e-24], [258.3872599432346, 1.4080829992654303e-24, 1.2700554566919064e-24, 1.1588116597872854e-24], [259.10370579500244, 1.4079297818693656e-24, 1.2702486406944833e-24, 1.1584531950080597e-24], [259.8221381791505, 1.4077796164925089e-24, 1.2704442377427135e-24, 1.1580974242656765e-24], [260.54256260385654, 1.407632509437734e-24, 1.2706422552101607e-24, 1.1577443512031483e-24], [261.2649845925703, 1.4074884670605804e-24, 1.2708427005224093e-24, 1.1573939795057426e-24], [261.9894096840579, 1.4073474957694033e-24, 1.2710455811571097e-24, 1.157046312901072e-24], [262.71584343244194, 1.4072096020254321e-24, 1.2712509046441168e-24, 1.1567013551591535e-24], [263.44429140724617, 1.407074792342974e-24, 1.271458678565513e-24, 1.1563591100925198e-24], [264.1747591934372, 1.4069430732894017e-24, 1.2716689105556951e-24, 1.1560195815562511e-24], [264.9072523914667, 1.4068144514853518e-24, 1.2718816083015028e-24, 1.1556827734480909e-24], [265.6417766173161, 1.4066889336047845e-24, 1.2720967795422259e-24, 1.1553486897084951e-24], [266.3783375025375, 1.4065665263750863e-24, 1.2723144320697104e-24, 1.1550173343207197e-24], [267.1169406942987, 1.4064472365772297e-24, 1.272534573728481e-24, 1.1546887113109099e-24], [267.8575918554258, 1.4063310710458197e-24, 1.2727572124157337e-24, 1.1543628247481593e-24], [268.60029666444586, 1.4062180366692175e-24, 1.2729823560815006e-24, 1.1540396787445943e-24], [269.34506081563194, 1.4061081403896995e-24, 1.2732100127286812e-24, 1.1537192774554694e-24], [270.09189001904485, 1.4060013892034553e-24, 1.273440190413147e-24, 1.1534016250792067e-24], [270.8407900005798, 1.4058977901608448e-24, 1.2736728972437847e-24, 1.153086725857534e-24], [271.59176650200664, 1.4057973503663527e-24, 1.273908141382631e-24, 1.1527745840754985e-24], [272.34482528101717, 1.4057000769788159e-24, 1.2741459310449232e-24, 1.1524652040616047e-24], [273.09997211126677, 1.4056059772114444e-24, 1.2743862744991768e-24, 1.1521585901878457e-24], [273.8572127824207, 1.4055150583320298e-24, 1.2746291800672639e-24, 1.1518547468698359e-24], [274.61655310019734, 1.4054273276629352e-24, 1.274874656124535e-24, 1.1515536785668356e-24], [275.37799888641234, 1.405342792581311e-24, 1.2751227110998593e-24, 1.1512553897818817e-24], [276.1415559790249, 1.4052614605191106e-24, 1.2753733534757113e-24, 1.1509598850618256e-24], [276.9072302321802, 1.4051833389632993e-24, 1.2756265917882787e-24, 1.1506671689974604e-24], [277.67502751625665, 1.4051084354558665e-24, 1.2758824346275298e-24, 1.1503772462235551e-24], [278.44495371790964, 1.4050367575940132e-24, 1.2761408906372918e-24, 1.1500901214189807e-24], [279.21701474011604, 1.4049683130301857e-24, 1.276401968515325e-24, 1.1498057993067485e-24], [279.9912165022214, 1.404903109472269e-24, 1.2766656770134616e-24, 1.1495242846541383e-24], [280.7675649399828, 1.4048411546836273e-24, 1.2769320249375992e-24, 1.1492455822727416e-24], [281.5460660056176, 1.4047824564832668e-24, 1.2772010211478848e-24, 1.1489696970185774e-24], [282.32672566784584, 1.4047270227458732e-24, 1.2774726745587131e-24, 1.1486966337921354e-24], [283.1095499119373, 1.4046748614020213e-24, 1.2777469941388471e-24, 1.1484263975385106e-24], [283.8945447397585, 1.4046259804381732e-24, 1.278023988911532e-24, 1.1481589932474299e-24], [284.68171616981635, 1.4045803878968987e-24, 1.2783036679545157e-24, 1.14789442595339e-24], [285.4710702373075, 1.4045380918768985e-24, 1.2785860404002009e-24, 1.1476327007356999e-24], [286.26261299416086, 1.4044991005331415e-24, 1.2788711154356542e-24, 1.147373822718581e-24], [287.05635050908717, 1.4044634220770088e-24, 1.2791589023027867e-24, 1.1471177970712598e-24], [287.8522888676232, 1.4044310647763458e-24, 1.2794494102983324e-24, 1.146864629008033e-24], [288.6504341721804, 1.4044020369556083e-24, 1.2797426487740226e-24, 1.1466143237883616e-24], [289.45079254209054, 1.404376346995985e-24, 1.2800386271366236e-24, 1.146366886716969e-24], [290.2533701136521, 1.4043540033354335e-24, 1.2803373548480318e-24, 1.1461223231438837e-24], [291.05817304017927, 1.4043350144688907e-24, 1.2806388414253677e-24, 1.1458806384645767e-24], [291.86520749204647, 1.4043193889483183e-24, 1.2809430964410522e-24, 1.1456418381200133e-24], [292.67447965673824, 1.404307135382807e-24, 1.2812501295228766e-24, 1.1454059275967352e-24], [293.48599573889544, 1.4042982624387289e-24, 1.281559950354138e-24, 1.1451729124269684e-24], [294.2997619603619, 1.404292778839822e-24, 1.2818725686736741e-24, 1.1449427981886915e-24], [295.1157845602345, 1.4042906933672878e-24, 1.2821879942759756e-24, 1.1447155905057223e-24], [295.9340697949084, 1.4042920148599078e-24, 1.2825062370112539e-24, 1.1444912950478e-24], [296.75462393812705, 1.4042967522141798e-24, 1.282827306785521e-24, 1.1442699175306856e-24], [297.57745328102925, 1.4043049143843837e-24, 1.2831512135607365e-24, 1.1440514637162318e-24], [298.40256413219794, 1.4043165103827132e-24, 1.2834779673548139e-24, 1.1438359394124661e-24], [299.2299628177073, 1.4043315492793943e-24, 1.2838075782417374e-24, 1.1436233504736946e-24], [300.05965568117347, 1.4043500402027763e-24, 1.2841400563516738e-24, 1.1434137028005722e-24], [300.89164908380053, 1.4043719923394163e-24, 1.2844754118710024e-24, 1.1432070023401786e-24], [301.7259494044318, 1.4043974149342574e-24, 1.2848136550424681e-24, 1.1430032550861362e-24], [302.5625630395974, 1.4044263172906522e-24, 1.2851547961651896e-24, 1.1428024670786538e-24], [303.40149640356276, 1.4044587087705518e-24, 1.2854988455948267e-24, 1.142604644404655e-24], [304.2427559283793, 1.4044945987945465e-24, 1.2858458137436009e-24, 1.1424097931978258e-24], [305.086348063933, 1.4045339968420032e-24, 1.2861957110804017e-24, 1.1422179196387186e-24], [305.93227927799296, 1.4045769124511687e-24, 1.2865485481308963e-24, 1.1420290299548379e-24], [306.780556056263, 1.4046233552192935e-24, 1.286904335477582e-24, 1.1418431304207235e-24], [307.63118490242914, 1.4046733348027023e-24, 1.2872630837598672e-24, 1.1416602273580327e-24], [308.48417233821135, 1.4047268609169205e-24, 1.287624803674201e-24, 1.1414803271356217e-24], [309.33952490341284, 1.404783943336793e-24, 1.2879895059741086e-24, 1.1413034361696463e-24], [310.19724915596913, 1.4048445918965759e-24, 1.288357201470315e-24, 1.1411295609236351e-24], [311.05735167200066, 1.4049088164900216e-24, 1.2887279010307888e-24, 1.1409587079085674e-24], [311.91983904586056, 1.4049766270705284e-24, 1.289101615580858e-24, 1.1407908836829774e-24], [312.78471789018744, 1.4050480336512158e-24, 1.2894783561033177e-24, 1.1406260948530293e-24], [313.65199483595495, 1.4051230463050417e-24, 1.289858133638432e-24, 1.1404643480725995e-24], [314.52167653252303, 1.4052016751648807e-24, 1.290240959284114e-24, 1.1403056500433563e-24], [315.3937696476878, 1.405283930423683e-24, 1.2906268441959197e-24, 1.1401500075148693e-24], [316.2682808677349, 1.4053698223345177e-24, 1.291015799587233e-24, 1.1399974272846648e-24], [317.1452168974882, 1.40545936121072e-24, 1.2914078367292425e-24, 1.1398479161983284e-24], [318.0245844603631, 1.4055525574259787e-24, 1.29180296695111e-24, 1.1397014811495858e-24], [318.9063902984178, 1.405649421414443e-24, 1.2922012016400303e-24, 1.1395581290803887e-24], [319.7906411724034, 1.4057499636708066e-24, 1.2926025522412726e-24, 1.139417866980987e-24], [320.6773438618184, 1.4058541947504414e-24, 1.293007030258317e-24, 1.1392807018900333e-24], [321.56650516495773, 1.4059621252694796e-24, 1.2934146472529248e-24, 1.139146640894651e-24], [322.4581318989675, 1.4060737659049313e-24, 1.2938254148452082e-24, 1.1390156911305314e-24], [323.352230899896, 1.4061891273947762e-24, 1.2942393447137175e-24, 1.138887859782012e-24], [324.24880902274543, 1.4063082205380486e-24, 1.29465644859553e-24, 1.1387631540821495e-24], [325.14787314152636, 1.4064310561949726e-24, 1.2950767382863307e-24, 1.1386415813128285e-24], [326.0494301493079, 1.4065576452870525e-24, 1.2955002256405167e-24, 1.1385231488048292e-24], [326.9534869582742, 1.406687998797156e-24, 1.2959269225712103e-24, 1.1384078639379104e-24], [327.8600504997729, 1.4068221277696345e-24, 1.2963568410504335e-24, 1.1382957341409031e-24], [328.7691277243723, 1.4069600433104075e-24, 1.2967899931091133e-24, 1.1381867668917816e-24], [329.68072560191166, 1.4071017565870835e-24, 1.297226390837234e-24, 1.1380809697177706e-24], [330.59485112155687, 1.4072472788290467e-24, 1.2976660463838484e-24, 1.1379783501954007e-24], [331.511511291853, 1.4073966213275487e-24, 1.2981089719572095e-24, 1.1378789159506139e-24], [332.43071314077747, 1.4075497954358218e-24, 1.2985551798248242e-24, 1.1377826746588344e-24], [333.3524637157955, 1.4077068125691807e-24, 1.2990046823135723e-24, 1.1376896340450694e-24], [334.2767700839123, 1.4078676842051064e-24, 1.2994574918097306e-24, 1.1375998018839728e-24], [335.2036393317288, 1.408032421883344e-24, 1.2999136207591229e-24, 1.1375131859999398e-24], [336.1330785654958, 1.4082010372060284e-24, 1.3003730816671317e-24, 1.1374297942671964e-24], [337.0650949111667, 1.4083735418377519e-24, 1.3008358870988576e-24, 1.137349634609878e-24], [337.99969551445486, 1.4085499475056709e-24, 1.3013020496791158e-24, 1.1372727150021073e-24], [338.9368875408858, 1.4087302659996373e-24, 1.3017715820925991e-24, 1.1371990434681059e-24], [339.87667817585407, 1.4089145091722105e-24, 1.3022444970838803e-24, 1.1371286280822309e-24], [340.8190746246778, 1.4091026889388604e-24, 1.3027208074575768e-24, 1.1370614769691158e-24], [341.76408411265277, 1.4092948172779976e-24, 1.3032005260783748e-24, 1.1369975983037185e-24], [342.71171388510993, 1.409490906231102e-24, 1.3036836658711326e-24, 1.1369370003114268e-24], [343.6619712074682, 1.4096909679027764e-24, 1.304170239820946e-24, 1.1368796912681236e-24], [344.6148633652941, 1.40989501446091e-24, 1.3046602609732653e-24, 1.1368256795002974e-24], [345.570397664353, 1.4101030581367315e-24, 1.3051537424339501e-24, 1.136774973385121e-24], [346.52858143066896, 1.4103151112249007e-24, 1.3056506973693541e-24, 1.136727581350525e-24], [347.48942201057815, 1.4105311860836454e-24, 1.3061511390064235e-24, 1.1366835118753101e-24], [348.4529267707872, 1.4107512951348262e-24, 1.306655080632773e-24, 1.1366427734892212e-24], [349.419103098429, 1.410975450864056e-24, 1.3071625355967672e-24, 1.1366053747730439e-24], [350.3879584011181, 1.411203665820783e-24, 1.3076735173076118e-24, 1.1365713243586868e-24], [351.35950010701, 1.4114359526183911e-24, 1.3081880392354408e-24, 1.1365406309292779e-24], [352.33373566485517, 1.411672323934314e-24, 1.3087061149113862e-24, 1.1365133032192608e-24], [353.3106725440589, 1.411912792510135e-24, 1.3092277579277152e-24, 1.1364893500144893e-24], [354.2903182347373, 1.412157371151647e-24, 1.3097529819378231e-24, 1.136468780152301e-24], [355.2726802477737, 1.4124060727290238e-24, 1.3102818006564325e-24, 1.1364516025216485e-24], [356.2577661148786, 1.4126589101768792e-24, 1.3108142278596322e-24, 1.1364378260631792e-24], [357.24558338864495, 1.4129158964943568e-24, 1.3113502773849414e-24, 1.1364274597693208e-24], [358.2361396426081, 1.4131770447452904e-24, 1.3118899631314687e-24, 1.136420512684415e-24], [359.2294424713027, 1.413442368058241e-24, 1.3124332990599444e-24, 1.1364169939047898e-24], [360.2254994903217, 1.4137118796266719e-24, 1.3129802991928524e-24, 1.1364169125788864e-24], [361.2243183363729, 1.4139855927089946e-24, 1.3135309776145219e-24, 1.136420277907352e-24], [362.2259066673406, 1.4142635206287593e-24, 1.314085348471255e-24, 1.1364270991431806e-24], [363.2302721623409, 1.4145456767746495e-24, 1.3146434259713516e-24, 1.1364373855917716e-24], [364.2374225217835, 1.4148320746007458e-24, 1.3152052243853089e-24, 1.1364511466111143e-24], [365.24736546742935, 1.4151227276265142e-24, 1.3157707580458828e-24, 1.1364683916118586e-24], [366.260108742449, 1.4154176494370078e-24, 1.3163400413482004e-24, 1.1364891300574655e-24], [367.27566011148406, 1.415716853682957e-24, 1.3169130887498844e-24, 1.1365133714643282e-24], [368.2940273607055, 1.4160203540808988e-24, 1.3174899147711664e-24, 1.1365411254019078e-24], [369.31521829787255, 1.4163281644133274e-24, 1.3180705339950363e-24, 1.136572401492877e-24], [370.33924075239474, 1.416640298528806e-24, 1.3186549610673469e-24, 1.1366072094132624e-24], [371.3661025753894, 1.4169567703421328e-24, 1.3192432106969394e-24, 1.1366455588926044e-24], [372.3958116397441, 1.4172775938344693e-24, 1.3198352976558212e-24, 1.1366874597141095e-24], [373.42837584017616, 1.4176027830535116e-24, 1.3204312367792885e-24, 1.1367329217148274e-24], [374.4638030932918, 1.4179323521136089e-24, 1.3210310429660816e-24, 1.136781954785813e-24], [375.5021013376496, 1.4182663151959974e-24, 1.3216347311785615e-24, 1.1368345688723402e-24], [376.5432785338186, 1.4186046865489156e-24, 1.3222423164428687e-24, 1.1368907739740744e-24], [377.5873426644413, 1.4189474804878362e-24, 1.3228538138491229e-24, 1.1369505801453079e-24], [378.634301734294, 1.4192947113955948e-24, 1.3234692385515814e-24, 1.1370139974951455e-24], [379.68416377034896, 1.419646393722701e-24, 1.3240886057689083e-24, 1.137081036187806e-24], [380.73693682183364, 1.4200025419874552e-24, 1.3247119307843362e-24, 1.137151706442815e-24], [381.79262896029593, 1.420363170776246e-24, 1.3253392289459335e-24, 1.137226018535317e-24], [382.8512482796627, 1.4207282947437615e-24, 1.3259705156668154e-24, 1.1373039827963575e-24], [383.91280289630424, 1.4210979286132708e-24, 1.3266058064254786e-24, 1.1373856096131963e-24], [384.9773009490958, 1.4214720871769669e-24, 1.3272451167660734e-24, 1.1374709094296848e-24], [386.04475059947856, 1.421850785296157e-24, 1.3278884622986642e-24, 1.1375598927465805e-24], [387.11516003152485, 1.4222340379016956e-24, 1.3285358586996504e-24, 1.1376525701219955e-24], [388.1885374519982, 1.4226218599943317e-24, 1.3291873217121058e-24, 1.1377489521718333e-24], [389.2648910904183, 1.423014266645078e-24, 1.329842867146197e-24, 1.1378490495702556e-24], [390.3442291991237, 1.4234112729956382e-24, 1.3305025108795959e-24, 1.1379528730502007e-24], [391.4265600533332, 1.4238128942588907e-24, 1.3311662688579796e-24, 1.1380604334039728e-24], [392.51189195121265, 1.4242191457194075e-24, 1.331834157095605e-24, 1.1381717414838793e-24], [393.60023321393464, 1.4246300427340041e-24, 1.3325061916758457e-24, 1.1382868082029096e-24], [394.6915921857471, 1.4250456007323682e-24, 1.3331823887518526e-24, 1.138405644535519e-24], [395.78597723403175, 1.4254658352177222e-24, 1.3338627645472638e-24, 1.138528261518459e-24], [396.8833967493732, 1.4258907617676212e-24, 1.3345473353570257e-24, 1.1386546702517507e-24], [397.98385914561936, 1.4263203960347717e-24, 1.335236117548262e-24, 1.1387848818997107e-24], [399.0873728599491, 1.4267547537479417e-24, 1.335929127561218e-24, 1.1389189076921204e-24], [400.19394635293537, 1.4271938507130554e-24, 1.3366263819104078e-24, 1.1390567589255454e-24], [401.3035881086092, 1.4276377028143183e-24, 1.3373278971858168e-24, 1.1391984469647823e-24], [402.41630663452736, 1.4280863260155466e-24, 1.3380336900542735e-24, 1.1393439832445095e-24], [403.5321104618344, 1.4285397363616283e-24, 1.3387437772610133e-24, 1.139493379271133e-24], [404.65100814533093, 1.428997949980169e-24, 1.3394581756313939e-24, 1.139646646624855e-24], [405.773008263538, 1.4294609830833795e-24, 1.3401769020729006e-24, 1.1398037969620428e-24], [406.89811941876144, 1.4299288519702007e-24, 1.3408999735773693e-24, 1.1399648420179025e-24], [408.0263502371612, 1.4304015730286825e-24, 1.341627407223509e-24, 1.1401297936094971e-24], [409.1577093688134, 1.4308791627387688e-24, 1.3423592201798422e-24, 1.1402986636392226e-24], [410.29220548778045, 1.4313616376753821e-24, 1.3430954297079341e-24, 1.1404714640987485e-24], [411.42984729217545, 1.43184901451207e-24, 1.343836053166289e-24, 1.1406482070735718e-24], [412.5706435042283, 1.4323413100250906e-24, 1.344581108014625e-24, 1.1408289047482061e-24], [413.71460287035507, 1.4328385410981898e-24, 1.345330611818951e-24, 1.1410135694122175e-24], [414.86173416122153, 1.4333407247281292e-24, 1.346084582257357e-24, 1.141202213467173e-24], [416.01204617181537, 1.4338478780310626e-24, 1.3468430371268014e-24, 1.1413948494347312e-24], [417.16554772150755, 1.4343600182499703e-24, 1.3476059943509354e-24, 1.1415914899660303e-24], [418.3222476541254, 1.4348771627633703e-24, 1.3483734719893353e-24, 1.1417921478527095e-24], [419.48215483801613, 1.4353993290955756e-24, 1.3491454882482801e-24, 1.1419968360398313e-24], [420.64527816611815, 1.4359265349286608e-24, 1.3499220614934676e-24, 1.1422055676410685e-24], [421.81162655602776, 1.436458798116728e-24, 1.3507032102650318e-24, 1.1424183559567143e-24], [422.98120895006645, 1.4369961367028258e-24, 1.3514889532954461e-24, 1.1426352144950643e-24], [424.1540343153523, 1.4375385689390686e-24, 1.3522793095307702e-24, 1.142856156997905e-24], [425.3301116438653, 1.4380861133108937e-24, 1.3530742981562837e-24, 1.143081197471133e-24], [426.5094499525193, 1.438638788566116e-24, 1.3538739386271885e-24, 1.143310350221566e-24], [427.69205828323, 1.4391966137503004e-24, 1.3546782507059362e-24, 1.1435436299016234e-24], [428.87794570298274, 1.4397596082496914e-24, 1.355487254507654e-24, 1.1437810515636675e-24], [430.0671213039053, 1.4403277918438653e-24, 1.356300970555722e-24, 1.1440226307266068e-24], [431.2595942033339, 1.440901184770686e-24, 1.3571194198503676e-24, 1.1442683834580271e-24], [432.45537354388625, 1.4414798078068039e-24, 1.3579426239537802e-24, 1.1445183264760618e-24], [433.65446849353003, 1.442063682368286e-24, 1.358770605096339e-24, 1.1447724772767003e-24], [434.85688824565386, 1.4426528306373839e-24, 1.359603386310554e-24, 1.145030854294109e-24], [436.062642019136, 1.4432472757232497e-24, 1.3604409916008072e-24, 1.145293477104024e-24], [437.27173905841835, 1.443847041867826e-24, 1.3612834461608504e-24, 1.145560366684182e-24], [438.48418863357364, 1.4444521547116337e-24, 1.3621307766545057e-24, 1.1458315457506532e-24], [439.70000004037956, 1.4450626416407577e-24, 1.3629830115821461e-24, 1.1461070391967862e-24], [440.9191826003888, 1.4456785322441947e-24, 1.3638401817637435e-24, 1.1463868746719437e-24], [442.1417456609991, 1.4462998589242326e-24, 1.3647023209834583e-24, 1.1466710833537457e-24], [443.3676985955277, 1.4469266577208995e-24, 1.3655694668603641e-24, 1.1469597009913397e-24], [444.5970508032818, 1.4475589694415184e-24, 1.3664416620412621e-24, 1.1472527693346257e-24], [445.8298117096291, 1.4481968412313724e-24, 1.3673189558593429e-24, 1.1475503381218005e-24], [447.06599076607364, 1.4488403287946387e-24, 1.3682014066794346e-24, 1.1478524678897696e-24], [448.3055974503242, 1.449489499593055e-24, 1.3690890852756282e-24, 1.148159234021939e-24], [449.5486412663705, 1.4501444375480989e-24, 1.3699820797963123e-24, 1.1484707326987255e-24], [450.79513174455445, 1.4508052501135666e-24, 1.3708805032319857e-24, 1.148787089848073e-24], [452.0450784416419, 1.451472079192269e-24, 1.3717845049415485e-24, 1.149108474961085e-24], [453.2984909408992, 1.452145118488663e-24, 1.3726942889733604e-24, 1.1494351230531706e-24], [454.555378852163, 1.4528246420372367e-24, 1.3736101441848122e-24, 1.149767370769937e-24], [455.81575181191704, 1.4535110529737403e-24, 1.3745324957320561e-24, 1.150105718113941e-24], [457.0796194833643, 1.4542049708269573e-24, 1.375461997224378e-24, 1.1504509389269447e-24], [458.3469915565019, 1.4549073965353877e-24, 1.3763997049272688e-24, 1.150804289749777e-24], [459.6178777481936, 1.4556200457967792e-24, 1.3773474296569825e-24, 1.1511679317456858e-24], [460.8922878022477, 1.4563460803040509e-24, 1.3783085086688376e-24, 1.15154585624075e-24], [462.1702314894878, 1.4570918900874889e-24, 1.3792896870142678e-24, 1.1519461406719576e-24], [463.4517186078309, 1.4578720925087057e-24, 1.3803063940202862e-24, 1.1523872758148976e-24], [464.73675898236144, 1.4587266253777344e-24, 1.3814007843075119e-24, 1.1529197994346097e-24], [466.025362465405, 1.4598002132889938e-24, 1.3827256051327754e-24, 1.153726857536758e-24], [467.31753893660726, 1.4619974147382977e-24, 1.3852354236245662e-24, 1.1559527262357047e-24], [468.61329830300576, 1.5072013988255413e-24, 1.4330466583388866e-24, 1.2120893161372698e-24], [469.91265049910936, 1.4676244695431448e-24, 1.3915505072465786e-24, 1.161952284098947e-24], [471.21560548697295, 1.4631935994646487e-24, 1.387065224898485e-24, 1.1557884556541435e-24], [472.5221732562723, 1.4632778864919948e-24, 1.387342221327183e-24, 1.1553295375291023e-24], [473.83236382438423, 1.4638304140544447e-24, 1.3881124693207782e-24, 1.1554605728496622e-24], [475.14618723645873, 1.4645064102695553e-24, 1.3890122026593463e-24, 1.155745167182989e-24], [476.4636535655021, 1.4652314479337026e-24, 1.3899628940167767e-24, 1.1560890783300439e-24], [477.78477291244724, 1.4659818516763957e-24, 1.390939564240261e-24, 1.1564623162168983e-24], [479.10955540623695, 1.4667482096434547e-24, 1.3919322832282408e-24, 1.1568529510583745e-24], [480.4380112038972, 1.4675261895124388e-24, 1.3929364799782893e-24, 1.1572554852848486e-24], [481.7701504906184, 1.4683135790009373e-24, 1.3939498201262053e-24, 1.1576671061052069e-24], [483.10598347983205, 1.4691091583042072e-24, 1.3949710163264911e-24, 1.1580862576403818e-24], [484.4455204132876, 1.4699122146299025e-24, 1.3959993160843516e-24, 1.1585120261297963e-24], [485.7887715611344, 1.4707223126565716e-24, 1.397034259596115e-24, 1.1589438492258174e-24], [487.13574722199644, 1.4715391776261636e-24, 1.398075556419297e-24, 1.1593813679284535e-24], [488.4864577230548, 1.472362632086738e-24, 1.3991230187423462e-24, 1.159824346465682e-24], [489.8409134201254, 1.4731925598982697e-24, 1.4001765234156384e-24, 1.1602726267026009e-24], [491.1991246977368, 1.4740288848616441e-24, 1.4012359894086807e-24, 1.1607261010715131e-24], [492.5611019692135, 1.4748715575587143e-24, 1.4023013639291006e-24, 1.1611846959022084e-24], [493.9268556767514, 1.4757205469887237e-24, 1.4033726136011054e-24, 1.1616483608272821e-24], [495.2963962915019, 1.4765758351035606e-24, 1.4044497187017117e-24, 1.162117061859482e-24], [496.66973431365005, 1.4774374131489462e-24, 1.4055326693018713e-24, 1.1625907767565577e-24], [498.04688027249426, 1.4783052791608343e-24, 1.4066214626260645e-24, 1.163069491849505e-24], [499.4278447265299, 1.4791794362185463e-24, 1.4077161012100708e-24, 1.1635531998294572e-24], [500.8126382635261, 1.4800598912044201e-24, 1.4088165915929195e-24, 1.1640418981761954e-24], [502.2012715006128, 1.4809466539089237e-24, 1.4099229433732442e-24, 1.1645355880243276e-24], [503.5937550843552, 1.481839736375572e-24, 1.4110351685184955e-24, 1.16503427333332e-24], [504.99009969084153, 1.482739152415157e-24, 1.4121532808527644e-24, 1.1655379602719342e-24], [506.39031602575994, 1.4836449172408555e-24, 1.4132772956720227e-24, 1.1660466567559632e-24], [507.79441482448465, 1.4845570471913592e-24, 1.4144072294522268e-24, 1.1665603720973393e-24], [509.20240685215634, 1.485475559518388e-24, 1.4155430996253128e-24, 1.1670791167349128e-24], [510.6143029037634, 1.4864004722221023e-24, 1.4166849244056752e-24, 1.1676029020259188e-24], [512.0301138042282, 1.4873318039225439e-24, 1.4178327226546772e-24, 1.1681317400830487e-24], [513.4498504084859, 1.4882695737582951e-24, 1.4189865137738058e-24, 1.1686656436460428e-24], [514.8735236015713, 1.4892138013060084e-24, 1.4201463176198907e-24, 1.1692046259797257e-24], [516.3011442987009, 1.4901645065159939e-24, 1.4213121544372032e-24, 1.1697487007924178e-24], [517.7327234453548, 1.491121709660414e-24, 1.4224840448028985e-24, 1.1702978821702618e-24], [519.1682720173646, 1.492085431291218e-24, 1.4236620095826714e-24, 1.1708521845239537e-24], [520.607801020993, 1.4930556922058658e-24, 1.4248460698946673e-24, 1.1714116225453043e-24], [522.0513214930216, 1.4940325134191961e-24, 1.4260362470798534e-24, 1.1719762111716174e-24], [523.4988445008346, 1.495015916140238e-24, 1.4272325626776014e-24, 1.1725459655563053e-24], [524.9503811425035, 1.4960059217529341e-24, 1.4284350384053924e-24, 1.1731209010445077e-24], [526.405942546871, 1.4970025518001624e-24, 1.4296436961420224e-24, 1.1737010331528249e-24], [527.8655398736367, 1.4980058279702892e-24, 1.430858557913433e-24, 1.1742863775523171e-24], [529.3291843134474, 1.499015772085892e-24, 1.4320796458808608e-24, 1.1748769500542187e-24], [530.796887087974, 1.5000324060941813e-24, 1.433306982330781e-24, 1.1754727665978674e-24], [532.2686594500059, 1.5010557520589925e-24, 1.4345405896664769e-24, 1.176073843240506e-24], [533.7445126835313, 1.5020858321538595e-24, 1.4357804904007958e-24, 1.1766801961485613e-24], [535.2244581038278, 1.5031226686560955e-24, 1.437026707149904e-24, 1.1772918415901955e-24], [536.7085070575478, 1.5041662839417953e-24, 1.438279262628078e-24, 1.1779087959289455e-24], [538.1966709228032, 1.5052167004814288e-24, 1.4395381796430886e-24, 1.1785310756182138e-24], [539.6889611092572, 1.5062739408361526e-24, 1.44080348109231e-24, 1.1791586971965507e-24], [541.1853890582071, 1.5073380276544809e-24, 1.4420751899592796e-24, 1.1797916772835257e-24], [542.6859662426759, 1.5084089836694508e-24, 1.443353329310728e-24, 1.1804300325761481e-24], [544.1907041674971, 1.5094868316961869e-24, 1.4446379222940291e-24, 1.18107377984577e-24], [545.6996143694057, 1.5105715946297167e-24, 1.4459289921349394e-24, 1.1817229359353607e-24], [547.2127084171232, 1.5116632954429976e-24, 1.4472265621355553e-24, 1.1823775177570745e-24], [548.7299979114512, 1.512761957185322e-24, 1.448530655672653e-24, 1.1830375422901948e-24], [550.2514944853564, 1.5138676029807761e-24, 1.4498412961960923e-24, 1.183703026579241e-24], [551.7772098040593, 1.5149802560269176e-24, 1.4511585072274673e-24, 1.1843739877323185e-24], [553.307155565128, 1.5160999395936366e-24, 1.4524823123589235e-24, 1.1850504429196816e-24], [554.8413434985625, 1.517226677022037e-24, 1.4538127352520074e-24, 1.1857324093723855e-24], [556.3797853668896, 1.5183604917235955e-24, 1.4551497996368346e-24, 1.1864199043811751e-24], [557.9224929652486, 1.5195014071791782e-24, 1.4564935293110688e-24, 1.1871129452953768e-24], [559.4694781214861, 1.5206494469384012e-24, 1.457843948139268e-24, 1.1878115495220043e-24], [561.0207526962416, 1.521804634618898e-24, 1.4592010800521839e-24, 1.1885157345249104e-24], [562.5763285830457, 1.5229669939056357e-24, 1.4605649490460104e-24, 1.1892255178239756e-24], [564.1362177084021, 1.5241365485504639e-24, 1.4619355791820126e-24, 1.1899409169944815e-24], [565.7004320318874, 1.525313322371443e-24, 1.4633129945858006e-24, 1.1906619496664046e-24], [567.2689835462367, 1.5264973392525208e-24, 1.4646972194470457e-24, 1.191388633523902e-24], [568.8418842774403, 1.5276886231429686e-24, 1.466088278018904e-24, 1.192120986304742e-24], [570.4191462848314, 1.5288871980570554e-24, 1.4674861946177144e-24, 1.1928590257998386e-24], [572.0007816611834, 1.5300930880736141e-24, 1.4688909936225856e-24, 1.193602769852787e-24], [573.5868025327974, 1.5313063173357639e-24, 1.470302699475117e-24, 1.194352236359485e-24], [575.1772210596002, 1.5325269100504977e-24, 1.4717213366790077e-24, 1.1951074432676967e-24], [576.7720494352326, 1.5337548904884334e-24, 1.4731469297998297e-24, 1.1958684085767338e-24], [578.3712998871483, 1.5349902829835721e-24, 1.474579503464808e-24, 1.1966351503371417e-24], [579.9749846767038, 1.536233111932912e-24, 1.4760190823624365e-24, 1.1974076866503254e-24], [581.5831160992516, 1.5374834017962928e-24, 1.477465691242368e-24, 1.1981860356683142e-24], [583.1957064842389, 1.5387411770961278e-24, 1.4789193549151505e-24, 1.1989702155934653e-24], [584.8127681952972, 1.5400064624171965e-24, 1.4803800982520542e-24, 1.1997602446782153e-24], [586.4343136303413, 1.541279282406374e-24, 1.4818479461848095e-24, 1.2005561412248126e-24], [588.0603552216609, 1.5425596617724711e-24, 1.4833229237054905e-24, 1.2013579235851057e-24], [589.6909054360189, 1.5438476252860553e-24, 1.4848050558663424e-24, 1.2021656101603222e-24], [591.3259767747438, 1.5451431977791903e-24, 1.4862943677795565e-24, 1.2029792194008352e-24], [592.965581773829, 1.5464464041453432e-24, 1.487790884617208e-24, 1.2037987698059934e-24], [594.6097330040262, 1.5477572693391554e-24, 1.489294631611041e-24, 1.2046242799239071e-24], [596.2584430709438, 1.5490758183762399e-24, 1.4908056340523007e-24, 1.205455768351241e-24], [597.9117246151407, 1.5504020763331332e-24, 1.4923239172917081e-24, 1.2062932537330933e-24], [599.569590312227, 1.5517360683470309e-24, 1.4938495067392293e-24, 1.2071367547627652e-24], [601.2320528729574, 1.5530778196156489e-24, 1.4953824278639658e-24, 1.2079862901816056e-24], [602.8991250433324, 1.554427355397132e-24, 1.496922706194078e-24, 1.208841878778873e-24], [604.5708196046922, 1.5557847010098052e-24, 1.4984703673165773e-24, 1.2097035393915286e-24], [606.2471493738185, 1.557149881832206e-24, 1.5000254368773967e-24, 1.2105712909041636e-24], [607.9281272030281, 1.5585229233027338e-24, 1.5015879405810406e-24, 1.2114451522487474e-24], [609.6137659802778, 1.559903850919641e-24, 1.5031579041906375e-24, 1.2123251424045427e-24], [611.3040786292577, 1.5612926902408991e-24, 1.5047353535278115e-24, 1.2132112803979619e-24], [612.9990781094909, 1.562689466884012e-24, 1.5063203144725473e-24, 1.214103585302397e-24], [614.698777416437, 1.5640942065259302e-24, 1.5079128129631138e-24, 1.2150020762381026e-24], [616.403189581586, 1.5655069349028735e-24, 1.5095128749959252e-24, 1.2159067723720355e-24], [618.1123276725638, 1.5669276778102852e-24, 1.511120526625544e-24, 1.2168176929177597e-24], [619.826204793227, 1.5683564611026584e-24, 1.5127357939645067e-24, 1.2177348571352874e-24], [621.5448340837688, 1.569793310693395e-24, 1.5143587031832345e-24, 1.2186582843309434e-24], [623.2682287208142, 1.5712382525547511e-24, 1.5159892805100136e-24, 1.2195879938572656e-24], [624.9964019175254, 1.5726913127176235e-24, 1.5176275522307946e-24, 1.2205240051128336e-24], [626.7293669237033, 1.5741525172715397e-24, 1.5192735446892322e-24, 1.2214663375421948e-24], [628.4671370258833, 1.5756218923644893e-24, 1.520927284286533e-24, 1.2224150106357153e-24], [630.2097255474447, 1.5770994642027652e-24, 1.5225887974813424e-24, 1.2233700439294365e-24], [631.9571458487071, 1.578585259050977e-24, 1.5242581107897893e-24, 1.2243314570050137e-24], [633.7094113270354, 1.580079303231787e-24, 1.525935250785238e-24, 1.2252992694895225e-24], [635.4665354169437, 1.5815816231258922e-24, 1.5276202440983248e-24, 1.2262735010553917e-24], [637.2285315901945, 1.5830922451719672e-24, 1.5293131174169191e-24, 1.2272541714202975e-24], [638.9954133559067, 1.5846111958663542e-24, 1.5310138974858419e-24, 1.2282413003469623e-24], [640.7671942606546, 1.5861385017632183e-24, 1.532722611107073e-24, 1.2292349076431602e-24], [642.5438878885743, 1.5876741894742375e-24, 1.5344392851394587e-24, 1.2302350131615015e-24], [644.3255078614702, 1.5892182856685863e-24, 1.5361639464987295e-24, 1.2312416367993696e-24], [646.1120678389125, 1.5907708170728114e-24, 1.537896622157407e-24, 1.2322547984987879e-24], [647.9035815183493, 1.5923318104707658e-24, 1.5396373391447829e-24, 1.233274518246335e-24], [649.700062635206, 1.5939012927034358e-24, 1.5413861245467672e-24, 1.234300816072995e-24], [651.5015249629953, 1.5954792906689036e-24, 1.543143005505882e-24, 1.2353337120540754e-24], [653.3079823134185, 1.5970658313222083e-24, 1.544908009221171e-24, 1.2363732263090852e-24], [655.1194485364756, 1.5986609416752897e-24, 1.5466811629481548e-24, 1.237419379001639e-24], [656.9359375205668, 1.6002646487968278e-24, 1.5484624939987274e-24, 1.238472190339325e-24], [658.7574631926042, 1.6018769798121906e-24, 1.5502520297411118e-24, 1.2395316805736176e-24], [660.5840395181125, 1.6034979619033357e-24, 1.5520497975998215e-24, 1.2405978699997685e-24], [662.4156805013434, 1.6051276223087198e-24, 1.5538558250555741e-24, 1.2416707789567009e-24], [664.2524001853774, 1.606765988323156e-24, 1.5556701396452117e-24, 1.2427504278268877e-24], [666.0942126522311, 1.608413087297784e-24, 1.55749276896168e-24, 1.2438368370362709e-24], [667.9411320229706, 1.6100689466399453e-24, 1.5593237406539545e-24, 1.2449300270541447e-24], [669.7931724578128, 1.6117335938130727e-24, 1.5611630824269468e-24, 1.2460300183930464e-24], [671.650348156241, 1.6134070563366754e-24, 1.5630108220415403e-24, 1.2471368316086844e-24], [673.5126733571068, 1.6150893617861398e-24, 1.5648669873144045e-24, 1.2482504872997869e-24], [675.380162338746, 1.6167805377927398e-24, 1.5667316061180617e-24, 1.2493710061080501e-24], [677.2528294190812, 1.6184806120434945e-24, 1.5686047063807606e-24, 1.2504984087180078e-24], [679.1306889557399, 1.6201896122810798e-24, 1.5704863160864197e-24, 1.2516327158569364e-24], [681.013755346155, 1.6219075663038143e-24, 1.5723764632746613e-24, 1.2527739482947837e-24], [682.9020430276838, 1.6236345019654553e-24, 1.5742751760406327e-24, 1.2539221268440193e-24], [684.7955664777116, 1.6253704471752226e-24, 1.576182482535052e-24, 1.2550772723595869e-24], [686.6943402137695, 1.6271154298976653e-24, 1.5780984109641336e-24, 1.2562394057387828e-24], [688.5983787936386, 1.6288694781525826e-24, 1.580022989589516e-24, 1.2574085479211619e-24], [690.5076968154688, 1.6306326200149818e-24, 1.581956246728266e-24, 1.2585847198884605e-24], [692.422308917884, 1.6324048836149366e-24, 1.5838982107527656e-24, 1.2597679426644758e-24], [694.3422297801003, 1.6341862971375612e-24, 1.5858489100907243e-24, 1.2609582373149919e-24], [696.2674741220321, 1.6359768888229102e-24, 1.5878083732251108e-24, 1.2621556249476827e-24], [698.1980567044133, 1.6377766869659186e-24, 1.5897766286941137e-24, 1.263360126712022e-24], [700.1339923289031, 1.6395857199163233e-24, 1.5917537050911163e-24, 1.264571763799195e-24], [702.0752958382011, 1.6414040160785785e-24, 1.5937396310646198e-24, 1.2657905574420027e-24], [704.0219821161648, 1.6432316039117962e-24, 1.5957344353182438e-24, 1.267016528914779e-24], [705.9740660879183, 1.6450685119296534e-24, 1.59773814661064e-24, 1.2682496995332909e-24], [707.9315627199724, 1.6469147687004413e-24, 1.5997507937555965e-24, 1.2694900906547088e-24], [709.8944870203334, 1.648770402846741e-24, 1.6017724056217253e-24, 1.270737723677404e-24], [711.8628540386239, 1.6506354430456645e-24, 1.6038030111327501e-24, 1.2719926200409998e-24], [713.8366788661933, 1.6525099180285368e-24, 1.60584263926721e-24, 1.2732548012261811e-24], [715.8159766362373, 1.6543938565810205e-24, 1.6078913190586253e-24, 1.274524288754688e-24], [717.8007625239117, 1.6562872875429573e-24, 1.6099490795953678e-24, 1.2758011041891964e-24], [719.7910517464501, 1.6581902398082695e-24, 1.6120159500205844e-24, 1.2770852691332224e-24], [721.7868595632776, 1.6601027423250226e-24, 1.6140919595323087e-24, 1.2783768052310919e-24], [723.7882012761331, 1.6620248240952465e-24, 1.6161771373833018e-24, 1.2796757341678162e-24], [725.7950922291805, 1.66395651417499e-24, 1.6182715128811326e-24, 1.2809820776690553e-24], [727.8075478091324, 1.66589784167416e-24, 1.6203751153880691e-24, 1.2822958575010047e-24], [729.825583445362, 1.6678488357565647e-24, 1.6224879743211293e-24, 1.283617095470358e-24], [731.8492146100275, 1.669809525639813e-24, 1.6246101191520274e-24, 1.2849458134242142e-24], [733.8784568181841, 1.6717799405952303e-24, 1.6267415794071121e-24, 1.2862820332499876e-24], [735.9133256279105, 1.673760109947891e-24, 1.628882384667433e-24, 1.2876257768753807e-24], [737.9538366404225, 1.6757500630765045e-24, 1.6310325645686537e-24, 1.288977066268279e-24], [740.0000055001927, 1.677749829413471e-24, 1.6331921488011482e-24, 1.2903359234367236e-24], [742.051847895075, 1.679759438444659e-24, 1.6353611671097875e-24, 1.291702370428781e-24], [744.1093795564188, 1.68177891970955e-24, 1.637539649294141e-24, 1.2930764293325421e-24], [746.172616259196, 1.683808302801139e-24, 1.6397276252083874e-24, 1.294458122276048e-24], [748.2415738221157, 1.6858476173657794e-24, 1.6419251247611924e-24, 1.2958474714271637e-24], [750.3162681077515, 1.687896893103372e-24, 1.644132177915935e-24, 1.297244498993628e-24], [752.3967150226567, 1.689956159767109e-24, 1.64634881469047e-24, 1.2986492272228892e-24], [754.4829305174917, 1.6920254471636393e-24, 1.6485750651573367e-24, 1.300061678402137e-24], [756.5749305871462, 1.6941047851528504e-24, 1.6508109594435524e-24, 1.3014818748581677e-24], [758.6727312708564, 1.69619420364801e-24, 1.6530565277307924e-24, 1.302909838957402e-24], [760.7763486523318, 1.6982937326156393e-24, 1.655311800255292e-24, 1.3043455931057862e-24], [762.8857988598807, 1.7004034020754953e-24, 1.6575768073078602e-24, 1.305789159748756e-24], [765.0010980665281, 1.702523242100574e-24, 1.6598515792339026e-24, 1.307240561371184e-24], [767.1222624901465, 1.704653282817119e-24, 1.6621361464334575e-24, 1.3086998204973594e-24], [769.2493083935728, 1.7067935544045718e-24, 1.664430539361189e-24, 1.310166959690924e-24], [771.3822520847407, 1.708944087095557e-24, 1.6667347885263806e-24, 1.3116420015548291e-24], [773.5211099167981, 1.7111049111758407e-24, 1.669048924492932e-24, 1.3131249687312875e-24], [775.6658982882382, 1.7132760569844076e-24, 1.6713729778794626e-24, 1.314615883901772e-24], [777.8166336430257, 1.7154575549133617e-24, 1.6737069793592367e-24, 1.316114769786941e-24], [779.9733324707156, 1.71764943540803e-24, 1.6760509596602916e-24, 1.3176216491466556e-24], [782.1360113065883, 1.71985172896683e-24, 1.678404949565342e-24, 1.3191365447798898e-24], [784.3046867317697, 1.7220644661413687e-24, 1.6807689799118843e-24, 1.3206594795247497e-24], [786.4793753733642, 1.724287677536457e-24, 1.6831430815922674e-24, 1.322190476258457e-24], [788.6600939045758, 1.7265213938099867e-24, 1.6855272855535672e-24, 1.3237295578972671e-24], [790.8468590448429, 1.7287656456731767e-24, 1.6879216227978728e-24, 1.32527674739656e-24], [793.0396875599591, 1.7310204638903303e-24, 1.6903261243820695e-24, 1.3268320677507105e-24], [795.2385962622079, 1.7332858792790735e-24, 1.692740821418099e-24, 1.3283955419931873e-24], [797.443602010491, 1.7355619227101836e-24, 1.695165745072805e-24, 1.329967193196453e-24], [799.6547217104511, 1.7378486251078376e-24, 1.697600926568229e-24, 1.3315470444720639e-24], [801.8719723146105, 1.7401460174494488e-24, 1.7000463971814466e-24, 1.3331351189706008e-24], [804.0953708224932, 1.7424541307658162e-24, 1.7025021882447524e-24, 1.33473143988172e-24], [806.3249342807624, 1.7447729961411384e-24, 1.7049683311457038e-24, 1.3363360304341685e-24], [808.5606797833437, 1.747102644713086e-24, 1.7074448573272075e-24, 1.3379489138958144e-24], [810.8026244715643, 1.7494431076728056e-24, 1.7099317982875504e-24, 1.3395701135736662e-24], [813.0507855342759, 1.7517944162651133e-24, 1.7124291855806208e-24, 1.3411996528139616e-24], [815.3051802079947, 1.754156601788371e-24, 1.7149370508158e-24, 1.3428375550021594e-24], [817.5658257770253, 1.756529695594775e-24, 1.717455425658284e-24, 1.3444838435630723e-24], [819.8327395736032, 1.7589137290903404e-24, 1.719984341829085e-24, 1.3461385419609137e-24], [822.1059389780166, 1.7613087337350286e-24, 1.7225238311051833e-24, 1.3478016736994003e-24], [824.385441418749, 1.7637147410429314e-24, 1.7250739253197316e-24, 1.349473262321895e-24], [826.6712643726052, 1.766131782582292e-24, 1.727634656362113e-24, 1.3511533314114872e-24], [828.9634253648528, 1.7685598899758023e-24, 1.730206056178237e-24, 1.352841904591225e-24], [831.2619419693494, 1.7709990949006808e-24, 1.7327881567706637e-24, 1.354539005524245e-24], [833.5668318086839, 1.7734494290889925e-24, 1.7353809901989353e-24, 1.3562446579140566e-24], [835.8781125543051, 1.775910924327711e-24, 1.737984588579663e-24, 1.35795888550472e-24], [838.1958019266635, 1.778383612459258e-24, 1.740598984087093e-24, 1.359681712081276e-24], [840.5199176953407, 1.7808675253815774e-24, 1.7432242089532022e-24, 1.3614131614700027e-24], [842.8504776791935, 1.7833626950486833e-24, 1.7458602954682777e-24, 1.363153257538928e-24], [845.1874997464838, 1.785869153470941e-24, 1.7485072759812097e-24, 1.3649020241982852e-24], [847.5310018150162, 1.788386932715791e-24, 1.751165182900261e-24, 1.366659485401206e-24], [849.8810018522806, 1.790916064908127e-24, 1.7538340486934545e-24, 1.3684256651443765e-24], [852.2375178755827, 1.7934565822312204e-24, 1.7565139058895415e-24, 1.370200587469005e-24], [854.6005679521896, 1.7960085169274616e-24, 1.759204787078772e-24, 1.3719842764618353e-24], [856.970170199461, 1.798571901299428e-24, 1.7619067249139894e-24, 1.3737767562564791e-24], [859.3463427849952, 1.80114676771117e-24, 1.7646197521119616e-24, 1.3755780510350003e-24], [861.7291039267614, 1.803733148589736e-24, 1.76734390145495e-24, 1.377388185029878e-24], [864.1184718932453, 1.8063310764270205e-24, 1.7700792057926056e-24, 1.3792071825263923e-24], [866.5144650035863, 1.808940583782231e-24, 1.7728256980444926e-24, 1.3810350678656951e-24], [868.9171016277189, 1.811561703284643e-24, 1.7755834112029025e-24, 1.3828718654485066e-24], [871.3264001865103, 1.8141944676374327e-24, 1.7783523783367815e-24, 1.3847175997399375e-24], [873.7423791519083, 1.816838909622264e-24, 1.7811326325964003e-24, 1.3865722952755508e-24], [876.1650570470749, 1.819495062105349e-24, 1.783924207219546e-24, 1.3884359766691733e-24], [878.5944524465364, 1.8221629580452005e-24, 1.786727135539416e-24, 1.3903086686230584e-24], [881.0305839763178, 1.8248426305028703e-24, 1.789541450995061e-24, 1.3921903959412225e-24], [883.4734703140937, 1.8275341126554767e-24, 1.7923671871451597e-24, 1.3940811835471509e-24], [885.9231301893224, 1.8302374378143814e-24, 1.7952043776865288e-24, 1.3959810565076008e-24], [888.3795823834002, 1.8329526394499206e-24, 1.798053056479311e-24, 1.3978900400650273e-24], [890.842845729797, 1.835679751225713e-24, 1.8009132575819006e-24, 1.39980815968248e-24], [893.3129391142014, 1.838418807046546e-24, 1.80378501529969e-24, 1.401735441106583e-24], [895.7898814746712, 1.841169841127209e-24, 1.8066683642551308e-24, 1.4036719104577652e-24], [898.2736918017715, 1.843932888092763e-24, 1.809563339489791e-24, 1.4056175943618814e-24], [900.7643891387268, 1.8467079831279796e-24, 1.8124699766164534e-24, 1.4075725201465472e-24], [903.2619925815607, 1.8494951622061027e-24, 1.8153883120519246e-24, 1.4095367161415065e-24], [905.7665212792489, 1.8522944624484888e-24, 1.8183183833830403e-24, 1.4115102121510907e-24], [908.277994433859, 1.8551059227090535e-24, 1.821260229961434e-24, 1.4134930402219063e-24], [910.7964313007042, 1.8579295845595362e-24, 1.8242138939061928e-24, 1.4154852359372323e-24], [913.3218511884869, 1.8607654940247495e-24, 1.8271794218696784e-24, 1.4174868406967516e-24], [915.8542734594491, 1.8636137047998965e-24, 1.8301568683115015e-24, 1.4194979059435957e-24], [918.3937175295167, 1.8664742845985653e-24, 1.833146301958173e-24, 1.42151850150473e-24], [920.9402028684551, 1.8693473286733315e-24, 1.8361478195612996e-24, 1.4235487333556131e-24], [923.4937490000109, 1.8722329915410592e-24, 1.839161578179866e-24, 1.4255887853048985e-24], [926.054375502069, 1.875131571509785e-24, 1.8421878811897086e-24, 1.4276390300585487e-24], [928.6221020067956, 1.878043779079339e-24, 1.845227451387209e-24, 1.429700381891062e-24], [931.1969482007956, 1.8809718447432666e-24, 1.8482825581830752e-24, 1.4317757522927059e-24], [933.7789338252568, 1.883926705984784e-24, 1.851364330265335e-24, 1.4338794937819031e-24], [936.3680786761059, 1.8870724118643454e-24, 1.8546396813843716e-24, 1.4362271450887176e-24], [938.9644026041624, 1.890307660130647e-24, 1.858005494298657e-24, 1.4386856096705216e-24], [941.5679255152812, 1.8928139819824264e-24, 1.8606291254160124e-24, 1.440179331930082e-24], [944.1786673705163, 1.8957715397279324e-24, 1.8637112458816276e-24, 1.4422590645424627e-24], [946.7966481862651, 1.898762011506678e-24, 1.866826253499584e-24, 1.4443751147654962e-24], [949.4218880344295, 1.9017693946942806e-24, 1.869957867462912e-24, 1.4465064327140152e-24], [952.054407042562, 1.9047909041359725e-24, 1.8731032541546743e-24, 1.4486493384450753e-24], [954.6942253940282, 1.9078257644010144e-24, 1.8762616245665207e-24, 1.4508027931889164e-24], [957.3413633281527, 1.9108737073592462e-24, 1.8794327057866294e-24, 1.4529664248601616e-24], [959.9958411403818, 1.9139346338335245e-24, 1.882616396783191e-24, 1.4551400833898302e-24], [962.6576791824383, 1.9170085119838336e-24, 1.8858126650277283e-24, 1.4573237071752533e-24], [965.3268978624701, 1.920095340456147e-24, 1.8890215090060263e-24, 1.4595172746467836e-24], [968.0035176452167, 1.923195133089499e-24, 1.8922429426607624e-24, 1.4617207841680012e-24], [970.6875590521573, 1.9263079118892766e-24, 1.895476988243385e-24, 1.4639342448000911e-24], [973.3790426616767, 1.9294337035299084e-24, 1.8987236727564637e-24, 1.4661576717049266e-24], [976.0779891092141, 1.9325725374979224e-24, 1.9019830260647133e-24, 1.4683910837042504e-24], [978.78441908743, 1.9357244450520958e-24, 1.905255079837242e-24, 1.4706345019126955e-24], [981.4983533463578, 1.9388894586143772e-24, 1.908539866927987e-24, 1.4728879489370885e-24], [984.2198126935658, 1.9420676113991534e-24, 1.911837420998608e-24, 1.4751514483890743e-24], [986.9488179943173, 1.9452589371802974e-24, 1.915147776281532e-24, 1.4774250245786341e-24], [989.6853901717335, 1.9484634701401466e-24, 1.9184709674263327e-24, 1.4797087023155954e-24], [992.4295502069441, 1.9516812447694768e-24, 1.921807029397997e-24, 1.4820025067778525e-24], [995.1813191392592, 1.954912295799351e-24, 1.925155997407592e-24, 1.4843064634218454e-24], [997.9407180663225, 1.958156658154199e-24, 1.928517906864553e-24, 1.486620597920606e-24], [1000.7077681442797, 1.9614143669186426e-24, 1.9318927933429145e-24, 1.4889449361200037e-24], [1003.4824905879341, 1.9646854573136952e-24, 1.9352806925571202e-24, 1.4912795040074337e-24], [1006.2649066709158, 1.967969964679563e-24, 1.938681640344515e-24, 1.4936243276889866e-24], [1009.0550377258382, 1.971267924462978e-24, 1.9420956726524594e-24, 1.4959794333727014e-24], [1011.8529051444687, 1.974579372207788e-24, 1.9455228255287495e-24, 1.4983448473560476e-24], [1014.6585303778838, 1.9779043435478524e-24, 1.9489631351143972e-24, 1.5007205960164703e-24], [1017.4719349366449, 1.981242874201806e-24, 1.9524166376382785e-24, 1.5031067058042835e-24], [1020.2931403909539, 1.9845949999688e-24, 1.9558833694128177e-24, 1.5055032032371074e-24], [1023.1221683708197, 1.987960756725625e-24, 1.959363366831036e-24, 1.5079101148957167e-24], [1025.9590405662302, 1.991340180424071e-24, 1.9628566663638704e-24, 1.5103274674206479e-24], [1028.8037787273101, 1.994733307089273e-24, 1.966363304558475e-24, 1.512755287509697e-24], [1031.6564046644962, 1.9981401728180942e-24, 1.969883318036576e-24, 1.515193601915851e-24], [1034.5169402486965, 2.0015608137782077e-24, 1.9734167434935263e-24, 1.5176424374457468e-24], [1037.3854074114654, 2.004995266207238e-24, 1.9769636176974214e-24, 1.5201018209584401e-24], [1040.2618281451646, 2.0084435664119415e-24, 1.9805239774882576e-24, 1.5225717793643201e-24], [1043.146224503138, 2.011905750767958e-24, 1.9840978597776698e-24, 1.5250523396244696e-24], [1046.0386185998777, 2.0153818557193663e-24, 1.987685301548472e-24, 1.527543528749995e-24], [1048.9390326111945, 2.018871917778451e-24, 1.991286339854404e-24, 1.5300453738015774e-24], [1051.8474887743844, 2.0223759735255395e-24, 1.9949010118199625e-24, 1.5325579018891306e-24], [1054.7640093884065, 2.0258940596089304e-24, 1.9985293546403105e-24, 1.53508114017151e-24], [1057.6886168140447, 2.029426212744918e-24, 2.0021714055812933e-24, 1.5376151158563816e-24], [1060.6213334740894, 2.0329724697177673e-24, 2.005827201979403e-24, 1.5401598562000796e-24], [1063.5621818534983, 2.03653286737972e-24, 2.009496781241771e-24, 1.5427153885074605e-24], [1066.51118449958, 2.0401074426511734e-24, 2.013180180846334e-24, 1.5452817401319724e-24], [1069.4683640221551, 2.0436962325208424e-24, 2.016877438341995e-24, 1.547858938475633e-24], [1072.433743093743, 2.047299274045708e-24, 2.020588591348549e-24, 1.5504470109889827e-24], [1075.4073444497249, 2.050916604351427e-24, 2.0243136775570896e-24, 1.553045985171279e-24], [1078.3891908885196, 2.054548260632401e-24, 2.0280527347300704e-24, 1.5556558885704763e-24], [1081.379305271765, 2.058194280151905e-24, 2.03180580070141e-24, 1.5582767487833641e-24], [1084.3777105244842, 2.0618547002424244e-24, 2.0355729133768343e-24, 1.5609085934556492e-24], [1087.3844296352706, 2.065529558305853e-24, 2.0393541107340535e-24, 1.5635514502821822e-24], [1090.3994856564548, 2.0692188918136008e-24, 2.0431494308228625e-24, 1.5662053470069422e-24], [1093.4229017042906, 2.0729227383070718e-24, 2.0469589117656028e-24, 1.5688703114233645e-24], [1096.4547009591224, 2.0766411353976804e-24, 2.0507825917571774e-24, 1.571546371374378e-24], [1099.494906665573, 2.0803741207672695e-24, 2.054620509065453e-24, 1.5742335547526376e-24], [1102.5435421327143, 2.084121732168266e-24, 2.058472702031398e-24, 1.5769318895006631e-24], [1105.6006307342507, 2.0878840074240475e-24, 2.0623392090694423e-24, 1.5796414036110873e-24], [1108.6661959086925, 2.0916609844291992e-24, 2.0662200686677277e-24, 1.5823621251267985e-24], [1111.7402611595444, 2.0954527011496766e-24, 2.070115319388249e-24, 1.5850940821411172e-24], [1114.8228500554753, 2.0992591956233003e-24, 2.0740249998673436e-24, 1.5878373027981045e-24], [1117.9139862305096, 2.103080505959832e-24, 2.0779491488157506e-24, 1.5905918152926473e-24], [1121.0136933841986, 2.1069166703414824e-24, 2.081887805019109e-24, 1.5933576478707523e-24], [1124.1219952818105, 2.110767727022956e-24, 2.0858410073379847e-24, 1.596134828829714e-24], [1127.2389157545056, 2.1146337143321048e-24, 2.0898087947085203e-24, 1.5989233865184141e-24], [1130.3644786995237, 2.11851467066982e-24, 2.0937912061423035e-24, 1.6017233493374152e-24], [1133.4987080803692, 2.122410634510673e-24, 2.097788280727002e-24, 1.6045347457393259e-24], [1136.6416279269854, 2.126321644403072e-24, 2.1018000576264988e-24, 1.6073576042289545e-24], [1139.7932623359502, 2.1302477389696214e-24, 2.105826576081248e-24, 1.6101919533635565e-24], [1142.953635470651, 2.1341889569074395e-24, 2.1098678754085714e-24, 1.613037821753096e-24], [1146.1227715614789, 2.138145336988505e-24, 2.1139239950029877e-24, 1.6158952380604696e-24], [1149.3006949060054, 2.142116918059847e-24, 2.1179949743363978e-24, 1.6187642310017124e-24], [1152.4874298691786, 2.1461037390441102e-24, 2.1220808529586286e-24, 1.6216448293463375e-24], [1155.6830008834993, 2.1501058389397288e-24, 2.1261816704975913e-24, 1.624537061917515e-24], [1158.8874324492165, 2.1541232568211918e-24, 2.130297466659534e-24, 1.6274409575922995e-24], [1162.100749134516, 2.1581560318396145e-24, 2.1344282812296e-24, 1.630356545301974e-24], [1165.3229755756984, 2.162204203222817e-24, 2.1385741540718857e-24, 1.6332838540322326e-24], [1168.55413647738, 2.166267810275914e-24, 2.1427351251300194e-24, 1.6362229128234815e-24], [1171.7942566126735, 2.170346892381385e-24, 2.1469112344272164e-24, 1.6391737507710555e-24], [1175.0433608233852, 2.1744414889997262e-24, 2.1511025220669037e-24, 1.642136397025529e-24], [1178.3014740201988, 2.1785516396695252e-24, 2.1553090282327975e-24, 1.6451108807929667e-24], [1181.5686211828693, 2.1826773840080276e-24, 2.1595307931894376e-24, 1.648097231335185e-24], [1184.8448273604176, 2.1868187617112643e-24, 2.1637678572823054e-24, 1.6510954779699956e-24], [1188.1301176713155, 2.1909758125547453e-24, 2.1680202609385003e-24, 1.6541056500715783e-24], [1191.424517303684, 2.1951485763933356e-24, 2.1722880446665954e-24, 1.657127777070574e-24], [1194.7280515154891, 2.1993370931621748e-24, 2.1765712490575454e-24, 1.660161888454611e-24], [1198.0407456347232, 2.203541402876607e-24, 2.180869914784595e-24, 1.6632080137683615e-24], [1201.3626250596133, 2.2077615456327392e-24, 2.1851840826038233e-24, 1.666266182613928e-24], [1204.6937152588048, 2.2119975616077307e-24, 2.1895137933544124e-24, 1.6693364246511047e-24], [1208.0340417715665, 2.216249491060318e-24, 2.1938590879591588e-24, 1.6724187695977136e-24], [1211.3836302079767, 2.220517374330941e-24, 2.198220007424575e-24, 1.6755132472297926e-24], [1214.742506249129, 2.224801251842363e-24, 2.2025965928415053e-24, 1.6786198873819816e-24], [1218.1106956473193, 2.229101164099895e-24, 2.206988885385312e-24, 1.681738719947776e-24], [1221.488224226254, 2.2334171516918606e-24, 2.211396926316339e-24, 1.684869774879837e-24], [1224.8751178812356, 2.2377492552899373e-24, 2.2158207569802293e-24, 1.6880130821902573e-24], [1228.2714025793762, 2.2420975156495875e-24, 2.2202604188083358e-24, 1.691168671950949e-24], [1231.677104359784, 2.246461973610313e-24, 2.2247159533179575e-24, 1.6943365742937948e-24], [1235.0922493337655, 2.250842670096264e-24, 2.229187402112932e-24, 1.697516819411142e-24], [1238.5168636850321, 2.2552396461164827e-24, 2.233674806883863e-24, 1.700709437555986e-24], [1241.9509736698922, 2.2596529427653245e-24, 2.2381782094085132e-24, 1.7039144590423098e-24], [1245.3946056174605, 2.2640826012230097e-24, 2.2426976515523456e-24, 1.7071319142455003e-24], [1248.8477859298516, 2.2685286627557746e-24, 2.247233175268651e-24, 1.7103618336025165e-24], [1252.3105410823919, 2.2729911687164977e-24, 2.2517848225991605e-24, 1.7136042476123364e-24], [1255.782897623811, 2.277470160544949e-24, 2.256352635674268e-24, 1.716859186836213e-24], [1259.2648821764612, 2.281965679768462e-24, 2.260936656713687e-24, 1.720126681898137e-24], [1262.7565214365022, 2.2864777680020794e-24, 2.2655369280265783e-24, 1.723406763485063e-24], [1266.2578421741232, 2.2910064669492296e-24, 2.270153492012206e-24, 1.7266994623473335e-24], [1269.768871233735, 2.295551818401974e-24, 2.274786391160158e-24, 1.7300048092990184e-24], [1273.2896355341866, 2.3001138642416407e-24, 2.2794356680509783e-24, 1.7333228352183597e-24], [1276.8201620689617, 2.3046926464391565e-24, 2.2841013653564532e-24, 1.7366535710480812e-24], [1280.3604779063958, 2.309288207055627e-24, 2.2887835258402013e-24, 1.7399970477958812e-24], [1283.9106101898724, 2.313900588242757e-24, 2.2934821923580453e-24, 1.7433532965348185e-24], [1287.4705861380428, 2.3185298322435075e-24, 2.2981974078586763e-24, 1.7467223484038487e-24], [1291.0404330450226, 2.323175981392497e-24, 2.3029292153840133e-24, 1.7501042346082582e-24], [1294.6201782806154, 2.3278390781166617e-24, 2.3076776580698598e-24, 1.7534989864202376e-24], [1298.2098492905104, 2.332519164935844e-24, 2.3124427791464677e-24, 1.7569066351794706e-24], [1301.8094735964946, 2.3372162844635395e-24, 2.317224621939259e-24, 1.760327212293857e-24], [1305.4190787966718, 2.341930479407601e-24, 2.322023229869525e-24, 1.7637607492402332e-24], [1309.038692565663, 2.346661792571058e-24, 2.3268386464552193e-24, 1.767207277565231e-24], [1312.6683426548302, 2.351410266853129e-24, 2.331670915311952e-24, 1.770666828886377e-24], [1316.3080568924772, 2.3561759452502696e-24, 2.3365200801540284e-24, 1.7741394348933192e-24], [1319.957863184075, 2.3609588708576263e-24, 2.3413861847958848e-24, 1.777625127349361e-24], [1323.6177895124642, 2.365759086870544e-24, 2.3462692731535837e-24, 1.7811239380933843e-24], [1327.287863938079, 2.370576636586589e-24, 2.351169389246827e-24, 1.7846358990423174e-24], [1330.968114599157, 2.3754115634083214e-24, 2.3560865772017253e-24, 1.788161042194373e-24], [1334.658569711958, 2.3802639108464183e-24, 2.3610208812539236e-24, 1.7916993996332688e-24], [1338.359257570975, 2.3851337225246756e-24, 2.365972345753606e-24, 1.7952510035342496e-24], [1342.070206549159, 2.3900210421857713e-24, 2.3709410151712856e-24, 1.798815886172016e-24], [1345.7914450981284, 2.3949259137008766e-24, 2.3759269341074605e-24, 1.8023940799327164e-24], [1349.523001748396, 2.399848381082156e-24, 2.3809301473051765e-24, 1.8059856173307457e-24], [1353.2649051095768, 2.4047884885025938e-24, 2.3859506996699797e-24, 1.8095905310346536e-24], [1357.0171838706199, 2.4097462803259153e-24, 2.3909886363000283e-24, 1.8132088539067525e-24], [1360.7798668000157, 2.4147218011543517e-24, 2.3960440025341845e-24, 1.816840619066442e-24], [1364.552982746026, 2.4197150959077976e-24, 2.4011168440317302e-24, 1.820485859995116e-24], [1368.3365606369055, 2.424726209959896e-24, 2.406207206909415e-24, 1.82414461071693e-24], [1372.1306294811122, 2.429755189384027e-24, 2.4113151379892196e-24, 1.827816906125791e-24], [1375.9352183675444, 2.4348020814250252e-24, 2.4164406852734912e-24, 1.8315027826123255e-24], [1379.7503564657513, 2.4398669354680205e-24, 2.4215838989208127e-24, 1.835202279352167e-24], [1383.5760730261677, 2.4449498052114624e-24, 2.4267448334348307e-24, 1.838915441195601e-24], [1387.4123973803273, 2.4500507541191916e-24, 2.4319235531559313e-24, 1.8426423259184416e-24], [1391.259358941099, 2.455169871339493e-24, 2.437120148295894e-24, 1.8463830253932167e-24], [1395.1169872029009, 2.460307329478323e-24, 2.4423347931302164e-24, 1.8501377424254604e-24], [1398.985311741934, 2.465463681621647e-24, 2.4475680451726073e-24, 1.8539071857883643e-24], [1402.8643622164127, 2.4706429370769896e-24, 2.4528239431719824e-24, 1.857696660950852e-24], [1406.7541683667766, 2.4837852437666613e-24, 2.4660997616886567e-24, 1.872053769343315e-24], [1410.6547600159365, 2.4810455292007524e-24, 2.46337887076027e-24, 1.8653028443649564e-24], [1414.56616706949, 2.486267285664233e-24, 2.46867631073028e-24, 1.8691173949085133e-24], [1418.4884195159561, 2.4915128937742477e-24, 2.473997353252844e-24, 1.8729531308309536e-24], [1422.4215474270095, 2.496777721676137e-24, 2.4793373335684864e-24, 1.8768038642584145e-24], [1426.3655809576996, 2.5020612620681035e-24, 2.4846957411531185e-24, 1.8806688931664172e-24], [1430.3205503466952, 2.507363426881324e-24, 2.4900724877276655e-24, 1.8845480730809173e-24], [1434.2864859165036, 2.512684216811589e-24, 2.495467574414449e-24, 1.8884413775291167e-24], [1438.2634180737118, 2.5180236592691933e-24, 2.500881029241638e-24, 1.892348815506214e-24], [1442.2513773092217, 2.523381791658063e-24, 2.5063128903024622e-24, 1.8962704092350906e-24], [1446.25039419847, 2.5287586557425007e-24, 2.511763200081599e-24, 1.9002061866725964e-24], [1450.26049940168, 2.534154295431328e-24, 2.5172320032234376e-24, 1.9041561785622997e-24], [1454.281723664081, 2.539568755801457e-24, 2.522719345548653e-24, 1.9081204171350475e-24], [1458.3140978161587, 2.545002082627255e-24, 2.528225273580184e-24, 1.912098935482993e-24], [1462.357652773878, 2.550454322137684e-24, 2.5337498342986213e-24, 1.916091767236192e-24], [1466.412419538932, 2.555925520883513e-24, 2.5392930750084445e-24, 1.9200989463855913e-24], [1470.4784291989683, 2.5614157256610962e-24, 2.5448550432612378e-24, 1.9241205071815018e-24], [1474.555712927839, 2.5669249834667933e-24, 2.550435786809761e-24, 1.9281564840726588e-24], [1478.6443019858268, 2.572453341469209e-24, 2.5560353535799662e-24, 1.9322069116689824e-24], [1482.7442277198998, 2.5780008469912245e-24, 2.5616537916528758e-24, 1.9362718247174644e-24], [1486.8555215639399, 2.5835675474986552e-24, 2.5672911492531388e-24, 1.9403512580868046e-24], [1490.9782150389844, 2.5891534905926626e-24, 2.572947474741367e-24, 1.9444452467569605e-24], [1495.1123397534768, 2.5947587240046275e-24, 2.578622816608936e-24, 1.948553825812187e-24], [1499.2579274034972, 2.6003832955927923e-24, 2.584317223474593e-24, 1.9526770304362264e-24], [1503.4150097730176, 2.606027253340031e-24, 2.590030744082179e-24, 1.956814895909127e-24], [1507.5836187341329, 2.611690645352414e-24, 2.5957634272991614e-24, 1.9609674576049087e-24], [1511.763786247318, 2.617373519858102e-24, 2.6015153221154934e-24, 1.9651347509899292e-24], [1515.9555443616587, 2.6230759252067777e-24, 2.607286477643021e-24, 1.9693168116219543e-24], [1520.1589252151161, 2.6287979098693305e-24, 2.613076943115138e-24, 1.9735136751493155e-24], [1524.3739610347518, 2.634539522437877e-24, 2.618886767886786e-24, 1.977725377310598e-24], [1528.6006841369922, 2.640300811625565e-24, 2.62471600143423e-24, 1.9819519539343014e-24], [1532.8391269278636, 2.646081826266866e-24, 2.6305646933553307e-24, 1.986193440938836e-24], [1537.0893219032528, 2.6518826153176563e-24, 2.6364328933695984e-24, 1.9904498743324023e-24], [1541.3513016491434, 2.6577032278557053e-24, 2.64232065131867e-24, 1.9947212902132143e-24], [1545.6250988418783, 2.6635437130807464e-24, 2.6482280171663416e-24, 1.999007724769554e-24], [1549.9107462483983, 2.6694041203149808e-24, 2.6541550409990692e-24, 2.003309214280045e-24], [1554.2082767265047, 2.6752844990035444e-24, 2.6601017730264006e-24, 2.0076257951138285e-24], [1558.517723225098, 2.681184898714686e-24, 2.666068263581143e-24, 2.0119575037309164e-24], [1562.8391187844481, 2.6871053691404584e-24, 2.672054563120022e-24, 2.0163043766823697e-24], [1567.1724965364317, 2.6930459600970507e-24, 2.678060722224007e-24, 2.0206664506107728e-24], [1571.5178897047902, 2.6990067215252205e-24, 2.6840867915987134e-24, 2.0250437622504653e-24], [1575.875331605392, 2.7049877034908533e-24, 2.690132822074946e-24, 2.0294363484278862e-24], [1580.2448556464765, 2.7109889561853247e-24, 2.6961988646090395e-24, 2.0338442460619414e-24], [1584.6264953289228, 2.717010529926191e-24, 2.7022849702835287e-24, 2.0382674921644303e-24], [1589.0202842464926, 2.7230524751574024e-24, 2.708391190307347e-24, 2.0427061238402796e-24], [1593.426256086102, 2.729114842450118e-24, 2.7145175760166135e-24, 2.047160178288153e-24], [1597.8444446280669, 2.7351976825029283e-24, 2.7206641788748453e-24, 2.0516296928005184e-24], [1602.2748837463714, 2.741301046142554e-24, 2.726831050473629e-24, 2.056114704764354e-24], [1606.7176074089239, 2.7474249843242807e-24, 2.7330182425330395e-24, 2.0606152516613178e-24], [1611.172649677818, 2.753569548132442e-24, 2.739225806902105e-24, 2.0651313710682413e-24], [1615.6400447095896, 2.759734788781018e-24, 2.7454537955593837e-24, 2.0696631006575247e-24], [1620.1198267554882, 2.7659207576141194e-24, 2.751702260613433e-24, 2.074210478197506e-24], [1624.6120301617273, 2.7721275061065114e-24, 2.7579712543033087e-24, 2.0787735415529073e-24], [1629.1166893697612, 2.778355085864087e-24, 2.764260828999023e-24, 2.083352328685152e-24], [1633.633838916535, 2.7846035486245583e-24, 2.7705710372022185e-24, 2.0879468776529678e-24], [1638.1635134347634, 2.7908729462576857e-24, 2.7769019315463755e-24, 2.092557226612535e-24], [1642.7057476531827, 2.7971633307661234e-24, 2.7832535647976433e-24, 2.097183413818047e-24], [1647.2605763968268, 2.8034747542857867e-24, 2.789625989855183e-24, 2.1018254776221814e-24], [1651.828034587296, 2.8098072690864534e-24, 2.7960192597517548e-24, 2.106483456476419e-24], [1656.4081572430102, 2.8161609275722e-24, 2.8024334276541365e-24, 2.111157388931399e-24], [1661.0009794794946, 2.822535782282003e-24, 2.8088685468637065e-24, 2.1158473136375434e-24], [1665.6065365096356, 2.8289318858903287e-24, 2.8153246708170105e-24, 2.1205532693452665e-24], [1670.2248636439622, 2.835349291207665e-24, 2.8218018530862837e-24, 2.1252752949055476e-24], [1674.855996290905, 2.841788051180906e-24, 2.8283001473798134e-24, 2.130013429270254e-24], [1679.4999699570803, 2.848248218893981e-24, 2.8348196075425455e-24, 2.1347677114925632e-24], [1684.1568202475487, 2.8547298475686795e-24, 2.841360287556902e-24, 2.139538180727566e-24], [1688.8265828660983, 2.8612329905646778e-24, 2.847922241542786e-24, 2.144324876232412e-24], [1693.5092936155213, 2.8677577013805206e-24, 2.854505523758538e-24, 2.1491278373669803e-24], [1698.204988397873, 2.8743040336538967e-24, 2.8611101886012114e-24, 2.1539471035942177e-24], [1702.9137032147582, 2.8808720411625444e-24, 2.8677362906074432e-24, 2.1587827144805936e-24], [1707.6354741676114, 2.8874617778243176e-24, 2.8743838844535124e-24, 2.1636347096965e-24], [1712.37033745796, 2.8940732976981643e-24, 2.881053024956304e-24, 2.168503129016754e-24], [1717.1183293877168, 2.9007066549845524e-24, 2.88774376707371e-24, 2.1733880123210668e-24], [1721.8794863594446, 2.90736190402597e-24, 2.8944561659051188e-24, 2.1782893995944375e-24], [1726.6538448766478, 2.9140390993077362e-24, 2.9011902766922086e-24, 2.1832073309276895e-24], [1731.4414415440401, 2.920738295458223e-24, 2.9079461548191507e-24, 2.1881418465179015e-24], [1736.2423130678344, 2.9274595472498475e-24, 2.914723855813592e-24, 2.1930929866689182e-24], [1741.0564962560265, 2.9342029095993658e-24, 2.921523435346922e-24, 2.1980607917918564e-24], [1745.884028018661, 2.940968437568811e-24, 2.9283449492352066e-24, 2.203045302405699e-24], [1750.7249453681345, 2.9477561863658403e-24, 2.9351884534395096e-24, 2.2080465591377585e-24], [1755.579285419463, 2.9545662113447353e-24, 2.942054004066887e-24, 2.2130646027244064e-24], [1760.4470853905805, 2.961398568006747e-24, 2.948941657370707e-24, 2.2180994740115824e-24], [1765.32838260261, 2.968253312001372e-24, 2.9558514697519212e-24, 2.223151213955769e-24], [1770.2232144801646, 2.975130499126851e-24, 2.9627834977595433e-24, 2.2282198636245802e-24], [1775.1316185516196, 2.9820301853312047e-24, 2.9697377980916685e-24, 2.2333054641979445e-24], [1780.0536324494142, 2.988952426713561e-24, 2.9767144275967953e-24, 2.238408056969242e-24], [1784.9892939103252, 2.9958972795251785e-24, 2.9837134432748277e-24, 2.2435276833467417e-24], [1789.9386407757743, 3.002864800171655e-24, 2.9907349022792822e-24, 2.2486643848557656e-24], [1794.9017109921037, 3.0098550452147672e-24, 2.997778861919114e-24, 2.2538182031412418e-24], [1799.8785426108677, 3.0168680713754567e-24, 3.004845379661694e-24, 2.2589891799713395e-24], [1804.8691737891359, 3.0239039355385266e-24, 3.011934513137511e-24, 2.2641773572431053e-24], [1809.8736427897707, 3.0309626947588304e-24, 3.0190463201463623e-24, 2.269382776990657e-24], [1814.8919879817356, 3.038044406270931e-24, 3.0261808586670354e-24, 2.2746054813979394e-24], [1819.924247840375, 3.045149127505973e-24, 3.0333381868742258e-24, 2.27984551282036e-24], [1824.9704609477237, 3.052276916118458e-24, 3.0405183631653894e-24, 2.285102913821118e-24], [1830.030665992785, 3.0594278300368575e-24, 3.0477214462115276e-24, 2.2903777272375986e-24], [1835.1049017718517, 3.0666019275589113e-24, 3.05494749505282e-24, 2.295669996308963e-24], [1840.1932071887754, 3.0737992675493976e-24, 3.062196569297095e-24, 2.3009797649399278e-24], [1845.295621255289, 3.0810199098876134e-24, 3.0694687295689345e-24, 2.3063070782982102e-24], [1850.4121830912873, 3.0882639166044255e-24, 3.0767640386508393e-24, 2.3116519843335866e-24], [1855.542931925143, 3.0955313552639174e-24, 3.0840825648771743e-24, 2.3170145382959172e-24], [1860.687907093992, 3.1028223115870537e-24, 3.0914243948036794e-24, 2.3223948196015274e-24], [1865.8471480440505, 3.1101369576150458e-24, 3.0987897016181824e-24, 2.3277930229094137e-24], [1871.020694330902, 3.1174763487940766e-24, 3.1061795451134997e-24, 2.3332105231495835e-24], [1876.208585619815, 3.1251405839652005e-24, 3.1138951084419446e-24, 2.3390482543190998e-24], [1881.410861686032, 3.1322228610582402e-24, 3.1210263150026607e-24, 2.3440945804785422e-24], [1886.6275624150926, 3.139631003594938e-24, 3.12848426768992e-24, 2.349562431109461e-24], [1891.8587278031255, 3.1470637611495087e-24, 3.1359666284978204e-24, 2.3550492626582847e-24], [1897.1043979571534, 3.1545205308737482e-24, 3.14347279292113e-24, 2.3605542338182103e-24], [1902.3646130954155, 3.162001284831285e-24, 3.151002733435914e-24, 2.3660772713788894e-24], [1907.6394135476573, 3.169506060762358e-24, 3.158556488429394e-24, 2.3716183897910608e-24], [1912.928839755458, 3.177034911090329e-24, 3.1661341110229756e-24, 2.3771776230285874e-24], [1918.2329322725232, 3.184587892877379e-24, 3.1737356589918364e-24, 2.3827550111668617e-24], [1923.551731765012, 3.19216506504249e-24, 3.1813611919729264e-24, 2.388350596664846e-24], [1928.8852790118328, 3.1997664874066273e-24, 3.189010770506695e-24, 2.393964423089255e-24], [1934.2336149049713, 3.207392220312543e-24, 3.1966844556555315e-24, 2.3995965346060123e-24], [1939.5967804497934, 3.2150423244560647e-24, 3.204382308834439e-24, 2.4052469757543443e-24], [1944.974816765366, 3.2227168608039263e-24, 3.2121043917285484e-24, 2.4109157913368968e-24], [1950.367765084763, 3.2304158905515975e-24, 3.219850766250795e-24, 2.416603026362991e-24], [1955.7756667553956, 3.2381394751005417e-24, 3.2276214945190734e-24, 2.4223087260177206e-24], [1961.1985632393137, 3.245887676044791e-24, 3.2354166388427552e-24, 2.4280329356438307e-24], [1966.6364961135391, 3.2536605551635064e-24, 3.243236261715214e-24, 2.43377570073152e-24], [1972.0895070703693, 3.2614581744164535e-24, 3.2510804258092634e-24, 2.4395370669118363e-24], [1977.557637917712, 3.269280595941515e-24, 3.2589491939746547e-24, 2.445317079953086e-24], [1983.0409305793885, 3.2771278820526995e-24, 3.2668426292360628e-24, 2.4511157857580165e-24], [1988.5394270954766, 3.2850000952396445e-24, 3.2747607947925777e-24, 2.4569332303625486e-24], [1994.0531696226155, 3.2928972981670713e-24, 3.2827037540171417e-24, 2.462769459934908e-24], [1999.58220043433, 3.300819553674844e-24, 3.290671570456598e-24, 2.4686245207751023e-24], [2005.1265619213677, 3.308766924777662e-24, 3.2986643078313704e-24, 2.474498459314627e-24], [2010.686296592008, 3.316739474665787e-24, 3.3066820300361794e-24, 2.480391322116611e-24], [2016.2614470724022, 3.324737266705086e-24, 3.314724801140078e-24, 2.4863031558758092e-24], [2021.8520561068872, 3.332760364437465e-24, 3.322792685386865e-24, 2.492234007418742e-24], [2027.4581665583266, 3.340808831581355e-24, 3.3308857471955745e-24, 2.498183923704038e-24], [2033.079821408425, 3.348882732032326e-24, 3.339004051161068e-24, 2.5041529518228946e-24], [2038.717063758071, 3.3569821298633766e-24, 3.3471476620543216e-24, 2.510141138999058e-24], [2044.369936827659, 3.3651070893255704e-24, 3.355316644823051e-24, 2.516148532589469e-24], [2050.0384839574263, 3.373257674848733e-24, 3.363511064592398e-24, 2.5221751800847197e-24], [2055.7227486077745, 3.381433951041554e-24, 3.3717309866650245e-24, 2.5282211291090486e-24], [2061.4227743596207, 3.389635982693011e-24, 3.379976476522529e-24, 2.534286427421387e-24], [2067.1386049147113, 3.39786383477207e-24, 3.388247599825135e-24, 2.5403711229152204e-24], [2072.870284095977, 3.406117572428646e-24, 3.396544422412646e-24, 2.5464752636192717e-24], [2078.6178558478487, 3.414397260994233e-24, 3.404867010305063e-24, 2.5525988976979305e-24], [2084.3813642366154, 3.422702965982645e-24, 3.413215429703321e-24, 2.5587420734518282e-24], [2090.160853450742, 3.4310347530902905e-24, 3.4215897469895535e-24, 2.5649048393180782e-24], [2095.9563678012196, 3.439392688196775e-24, 3.4299900287276905e-24, 2.5710872438707314e-24], [2101.767951721913, 3.447776837365948e-24, 3.438416341664487e-24, 2.5772893358214605e-24], [2107.595649769876, 3.4561872668459896e-24, 3.446868752729614e-24, 2.5835111640197676e-24], [2113.4395066257207, 3.4646240430704355e-24, 3.4553473290366676e-24, 2.58975277745369e-24], [2119.2995670939376, 3.47308723265847e-24, 3.463852137883462e-24, 2.5960142252500924e-24], [2125.1758761032597, 3.4815769024158746e-24, 3.472383246752955e-24, 2.602295556675173e-24], [2131.068478706986, 3.490093119335187e-24, 3.4809407233134255e-24, 2.6085968211349563e-24], [2136.977420083348, 3.4986359505968745e-24, 3.48952463541961e-24, 2.6149180681757926e-24], [2142.902745535836, 3.5072054635696016e-24, 3.498135051112981e-24, 2.621259347484829e-24], [2148.8445004935643, 3.515801725810864e-24, 3.506772038622376e-24, 2.6276207088904212e-24], [2154.8027305116016, 3.524424805067668e-24, 3.515435666364659e-24, 2.6340022023626646e-24], [2160.777481271342, 3.5330747692774544e-24, 3.5241260029456406e-24, 2.6404038780139393e-24], [2166.7687985808366, 3.5417516865681365e-24, 3.5328431171601165e-24, 2.6468257860992985e-24], [2172.7767283751473, 3.550455625259285e-24, 3.541587077993033e-24, 2.653267977017003e-24], [2178.801316716711, 3.55918665386254e-24, 3.550357954619907e-24, 2.659730501309087e-24], [2184.8426097956767, 3.567944841082197e-24, 3.559155816407396e-24, 2.6662134096616816e-24], [2190.900653930276, 3.576730255816045e-24, 3.56798073291413e-24, 2.6727167529058286e-24], [2196.9754955671633, 3.5855429671559724e-24, 3.576832773891315e-24, 2.679240582017725e-24], [2203.067181281778, 3.5943830443885054e-24, 3.585712009283265e-24, 2.6857849481194185e-24], [2209.1757577787052, 3.6032505569957884e-24, 3.594618509228372e-24, 2.6923499024793377e-24], [2215.3012718920363, 3.612145574656034e-24, 3.603552344059558e-24, 2.6989354965130176e-24], [2221.443770585709, 3.621068167244443e-24, 3.612513584305184e-24, 2.705541781783515e-24], [2227.6033009538933, 3.630018404834082e-24, 3.62150230068992e-24, 2.71216881000237e-24], [2233.7799102213294, 3.6389963576965e-24, 3.6305185641353674e-24, 2.7188166330302797e-24], [2239.9736457437098, 3.6480020963031546e-24, 3.639562445761462e-24, 2.7254853028781467e-24], [2246.1845550080247, 3.657035691325985e-24, 3.6486340168870625e-24, 2.732174871708119e-24], [2252.4126856329426, 3.666097213639569e-24, 3.6577333490320945e-24, 2.7388853918354738e-24], [2258.658085369158, 3.675186734322214e-24, 3.6668605139186345e-24, 2.745616915730236e-24], [2264.9208020997767, 3.684304324659146e-24, 3.676015583474106e-24, 2.752369496020577e-24], [2271.2008838406605, 3.693450056145959e-24, 3.685198629834728e-24, 2.7591431854974195e-24], [2277.4983787408214, 3.702624000494266e-24, 3.694409725351168e-24, 2.7659380371217275e-24], [2283.8133350827707, 3.711826229642268e-24, 3.703648942599134e-24, 2.7727541040376475e-24], [2290.1458012828907, 3.7210568157718586e-24, 3.7129163543965015e-24, 2.779591439595912e-24], [2296.4958258918195, 3.73031583134426e-24, 3.72221203383902e-24, 2.7864500974004264e-24], [2302.8634575948063, 3.739603349173703e-24, 3.731536054374142e-24, 2.793330131406945e-24], [2309.2487452121004, 3.748919442601996e-24, 3.7408884899759554e-24, 2.8002315961563533e-24], [2315.65173769931, 3.758264185968362e-24, 3.75026941561598e-24, 2.8071545474032616e-24], [2322.0724841477954, 3.7676376561249065e-24, 3.759678908781789e-24, 2.8140990441447543e-24], [2328.5110337850238, 3.777039938862457e-24, 3.769117055916179e-24, 2.8210651572237814e-24], [2334.9674359749665, 3.7864711728568644e-24, 3.7785839964545596e-24, 2.8280530281693555e-24], [2341.441740218468, 3.795932620355006e-24, 3.788080995726338e-24, 2.8350643027693664e-24], [2347.9339961536284, 3.805421895672124e-24, 3.7976056635500695e-24, 2.8420957424171554e-24], [2354.444253556174, 3.814938464113304e-24, 3.8071574646577095e-24, 2.8491465865538183e-24], [2360.9725623398585, 3.82448566740114e-24, 3.816739748097856e-24, 2.856221264963132e-24], [2367.518972556824, 3.834062196349262e-24, 3.8263512024484366e-24, 2.8633179799489127e-24], [2374.0835343980048, 3.8436680571301526e-24, 3.835991834355369e-24, 2.870436694881523e-24], [2380.6662981934933, 3.853303311979481e-24, 3.845661706640986e-24, 2.8775774480717945e-24], [2387.2673144129444, 3.862968032339198e-24, 3.855360891350697e-24, 2.8847402900315365e-24], [2393.8866336659416, 3.872662292199824e-24, 3.8650894630816035e-24, 2.8919252745586838e-24], [2400.524306702411, 3.882386166557199e-24, 3.874847497436074e-24, 2.899132456670634e-24], [2407.180384412989, 3.8921397309412746e-24, 3.884635070549561e-24, 2.9063618919728408e-24], [2413.854917829414, 3.901923061243346e-24, 3.8944522589174886e-24, 2.913613636427072e-24], [2420.5479581249333, 3.9117362336439626e-24, 3.904299139322998e-24, 2.9208877462547435e-24], [2427.2595566146756, 3.921579324580393e-24, 3.914175788804346e-24, 2.9281842778926023e-24], [2433.9897647560642, 3.931452410729956e-24, 3.924082284638205e-24, 2.9355032879704274e-24], [2440.7386341491915, 3.941355569002484e-24, 3.934018704332096e-24, 2.9428448332999693e-24], [2447.5062165372337, 3.951288876534994e-24, 3.9439851256190525e-24, 2.9502089708680646e-24], [2454.2925638068295, 3.96125241069034e-24, 3.953981626456265e-24, 2.957595757833871e-24], [2461.0977279884937, 3.9712462490554e-24, 3.9640082850232586e-24, 2.9650052515264852e-24], [2467.921761257007, 3.981270469440842e-24, 3.974065179721657e-24, 2.9724375094441015e-24], [2474.7647159318203, 3.9913251498812736e-24, 3.9841523891753314e-24, 2.979892589253475e-24], [2481.626644477447, 4.00141036863505e-24, 3.994269992230193e-24, 2.9873705487898848e-24], [2488.5075995038815, 4.0115262041848944e-24, 4.00441806795483e-24, 2.99487144605706e-24], [2495.4076337669835, 4.021672735238349e-24, 4.014596695640938e-24, 3.002395339227446e-24], [2502.326800168901, 4.0318500407280015e-24, 4.024805954803558e-24, 3.009942286642507e-24], [2509.265151758458, 4.042058199812418e-24, 4.035045925182001e-24, 3.0175123468131054e-24], [2516.2227417315808, 4.052297291876626e-24, 4.045316686740323e-24, 3.0251055784198517e-24], [2523.199623431683, 4.062567396532487e-24, 4.055618319667714e-24, 3.0327220403134477e-24], [2530.195850350093, 4.0728685936197496e-24, 4.065950904379528e-24, 3.0403617915154308e-24], [2537.211476126466, 4.083200963206507e-24, 4.076314521517754e-24, 3.0480248912183526e-24], [2544.2465545491714, 4.0935645855897185e-24, 4.0867092519515345e-24, 3.0557113987864253e-24], [2551.301139555733, 4.1039595412960506e-24, 4.097135176777997e-24, 3.063421373755896e-24], [2558.3752852332204, 4.114385911082793e-24, 4.107592377323174e-24, 3.0711548758358102e-24], [2565.469045818684, 4.124843775937977e-24, 4.118080935142126e-24, 3.0789119649080142e-24], [2572.5824756995517, 4.135333217081343e-24, 4.128600932019897e-24, 3.0866927010282046e-24], [2579.715629414062, 4.1458543159653e-24, 4.139152449972481e-24, 3.0944971444261997e-24], [2586.868561651667, 4.156407154275272e-24, 4.149735571247164e-24, 3.1023253555064337e-24], [2594.041327253469, 4.1669918139304946e-24, 4.160350378323321e-24, 3.1101773948485308e-24], [2601.233981212618, 4.177608377084676e-24, 4.170996953913073e-24, 3.1180533232077794e-24], [2608.4465786747633, 4.188256926126749e-24, 4.1816753809620465e-24, 3.1259532015158037e-24], [2615.679174938452, 4.19893754368167e-24, 4.1923857426501634e-24, 3.133877090880994e-24], [2622.9318254555556, 4.209650312611084e-24, 4.20312812239231e-24, 3.141825052589006e-24], [2630.204585831709, 4.2203953160138244e-24, 4.213902603838834e-24, 3.149797148103402e-24], [2637.497511826724, 4.231172637227065e-24, 4.2247092708766974e-24, 3.15779343906604e-24], [2644.8106593550165, 4.241982359826348e-24, 4.235548207629506e-24, 3.165813987297814e-24], [2652.1440844860517, 4.2528245676273474e-24, 4.246419498459266e-24, 3.1738588547992785e-24], [2659.4978434447485, 4.2636993446856515e-24, 4.257323227966177e-24, 3.181928103750855e-24], [2666.87199261193, 4.274606775297895e-24, 4.268259480989754e-24, 3.1900217965137513e-24], [2674.266588524759, 4.285546944002745e-24, 4.279228342609816e-24, 3.1981399956306177e-24], [2681.681687877145, 4.296519935581423e-24, 4.2902298981470164e-24, 3.206282763825949e-24], [2689.1173475202104, 4.30752583505858e-24, 4.301264233163705e-24, 3.2144501640069767e-24], [2696.5736244626996, 4.318564727703372e-24, 4.312331433465014e-24, 3.222642259264419e-24], [2704.0505758714417, 4.3296366990303225e-24, 4.3234315850997126e-24, 3.2308591128734436e-24], [2711.5482590717634, 4.340741834800451e-24, 4.334564774361349e-24, 3.239100788294664e-24], [2719.0667315479523, 4.351880221022881e-24, 4.345731087789846e-24, 3.2473673491757897e-24], [2726.606050943674, 4.363051943956492e-24, 4.356930612173161e-24, 3.2556588593535594e-24], [2734.1662750624387, 4.3742570901127386e-24, 4.3681634345501106e-24, 3.263975382857026e-24], [2741.747461868017, 4.385495746259869e-24, 4.379429642214589e-24, 3.272316983912642e-24], [2749.3496694849173, 4.396767999429867e-24, 4.390729322722527e-24, 3.28068372695335e-24], [2756.9729561988024, 4.4080739369326066e-24, 4.4020625639060474e-24, 3.2890756766370174e-24], [2764.617380456944, 4.419413646384547e-24, 4.413429453902222e-24, 3.297492897884483e-24], [2772.2830008686838, 4.430787215778029e-24, 4.424830081222434e-24, 3.3059354559719247e-24], [2779.9698762058624, 4.44219473367026e-24, 4.436264534941632e-24, 3.31440341678388e-24], [2787.6780654032923, 4.453636289814664e-24, 4.447732905330481e-24, 3.322896847658861e-24], [2795.4076275591865, 4.465111977987906e-24, 4.4592352866860884e-24, 3.331415821178232e-24], [2803.158621935634, 4.476621917601651e-24, 4.470771798971851e-24, 3.339960444132574e-24], [2810.9311079590325, 4.48816711774421e-24, 4.482343452980542e-24, 3.348532015632322e-24], [2818.7251452205633, 4.499744956387547e-24, 4.4939476236683905e-24, 3.357126967711227e-24], [2826.5407934766367, 4.5113577238405616e-24, 4.505586604718236e-24, 3.365748316430352e-24], [2834.378112649357, 4.523005323878323e-24, 4.517260300161211e-24, 3.3743958789626915e-24], [2842.237162826967, 4.534687607012911e-24, 4.5289685607030424e-24, 3.3830694009223424e-24], [2850.1180042643346, 4.5464046463989405e-24, 4.540711459977791e-24, 3.391768926189181e-24], [2858.0206973833865, 4.558156529081149e-24, 4.552489085525519e-24, 3.400494517108198e-24], [2865.9453027735963, 4.5699433447876135e-24, 4.564301527571041e-24, 3.4092462394665964e-24], [2873.891881192427, 4.581765184167934e-24, 4.576148877259836e-24, 3.4180241601330594e-24], [2881.860493565818, 4.593622138364352e-24, 4.588031226228615e-24, 3.426828346481914e-24], [2889.8512009886285, 4.605514298876607e-24, 4.5999486664700055e-24, 3.43565886621158e-24], [2897.864064725136, 4.617441757511582e-24, 4.611901290282128e-24, 3.4445157872766846e-24], [2905.8991462094796, 4.62940460636252e-24, 4.623889190247797e-24, 3.4533991778594784e-24], [2913.956507046136, 4.641402937798725e-24, 4.635912459224194e-24, 3.462309106356278e-24], [2922.0362090104063, 4.6534368444624075e-24, 4.6479711903397275e-24, 3.471245641371856e-24], [2930.13831404887, 4.665506419265742e-24, 4.6600654769910754e-24, 3.480208851715884e-24], [2938.2628842798804, 4.677611755390351e-24, 4.672195412842667e-24, 3.489198806401178e-24], [2946.4099819940175, 4.68975294628726e-24, 4.684361091826652e-24, 3.498215574643616e-24], [2954.5796696545913, 4.7019300856768324e-24, 4.696562608142824e-24, 3.5072592258612795e-24], [2962.7720098980953, 4.714143267549489e-24, 4.708800056259349e-24, 3.516329829675074e-24], [2970.9870655347067, 4.72639258616628e-24, 4.721073530913331e-24, 3.525427455908877e-24], [2979.22489954876, 4.738678136058915e-24, 4.7333831271108585e-24, 3.5345521745896596e-24], [2987.485575099229, 4.751000012031213e-24, 4.745728940128434e-24, 3.543704055948118e-24], [2995.769155520208, 4.7633583091591275e-24, 4.758111065513025e-24, 3.5528831704192416e-24], [3004.07570432141, 4.775753122792064e-24, 4.770529599083369e-24, 3.562089588642566e-24], [3012.4052851886336, 4.7881845485530256e-24, 4.782984636930122e-24, 3.571323381462866e-24], [3020.7579619842772, 4.8006526823402026e-24, 4.795476275417458e-24, 3.580584619930909e-24], [3029.133798747799, 4.8131576203268384e-24, 4.808004611182928e-24, 3.589873375303642e-24], [3037.532859696238, 4.825699458962428e-24, 4.820569741138668e-24, 3.599189719044858e-24], [3045.9552092246786, 4.83827829497369e-24, 4.8331717624723704e-24, 3.608533722825869e-24], [3054.4009119067623, 4.8508942253649765e-24, 4.845810772647694e-24, 3.6179054585262485e-24], [3062.8700324951888, 4.8635473474191245e-24, 4.8584868694051224e-24, 3.627304998233914e-24], [3071.3626359221817, 4.876237758698671e-24, 4.871200150763177e-24, 3.6367324142463005e-24], [3079.878787300022, 4.888965557046024e-24, 4.8839507150185985e-24, 3.646187779070518e-24], [3088.418551921519, 4.901730840584708e-24, 4.896738660747581e-24, 3.655671165424271e-24], [3096.981995260536, 4.914533707720065e-24, 4.909564086806488e-24, 3.6651826462361925e-24], [3105.5691829724674, 4.927374257139926e-24, 4.922427092332526e-24, 3.674722294646586e-24], [3114.1801808947657, 4.9402525878158035e-24, 4.935327776744927e-24, 3.684290184008196e-24], [3122.8150550474224, 4.9531687990031126e-24, 4.94826623974519e-24, 3.69388638788656e-24], [3131.473871633494, 4.9661229902426426e-24, 4.9612425813185435e-24, 3.703510980060803e-24], [3140.1566970396107, 4.979115261361164e-24, 4.974256901734555e-24, 3.713164034524445e-24], [3148.863597836458, 4.992145712472165e-24, 4.9873093015478784e-24, 3.7228456254858664e-24], [3157.5946407793226, 5.0052144439771285e-24, 5.000399881599528e-24, 3.732555827369315e-24], [3166.3498928085774, 5.018321556566334e-24, 5.013528743017684e-24, 3.7422947148156426e-24], [3175.1294210502047, 5.031467151219997e-24, 5.0266959872188354e-24, 3.7520623626833596e-24], [3183.9332928163203, 5.044651329209636e-24, 5.039901715909152e-24, 3.761858846050104e-24], [3192.761575605669, 5.057874192099971e-24, 5.0531460310863874e-24, 3.7716842402143516e-24], [3201.6143371041644, 5.071135841751458e-24, 5.0664290350424054e-24, 3.781538620698341e-24], [3210.491645185386, 5.084436380323793e-24, 5.0797508303667154e-24, 3.791422063253058e-24], [3219.3935679111146, 5.097775910284412e-24, 5.0931115199549554e-24, 3.801334643867854e-24], [3228.3201735318626, 5.111154534423333e-24, 5.1065112070237704e-24, 3.811276438790909e-24], [3237.271530487362, 5.124572355891749e-24, 5.119949995149419e-24, 3.821247524580186e-24], [3246.2477074071303, 5.138029478312624e-24, 5.133427988378489e-24, 3.831247978250978e-24], [3255.248773110964, 5.1515260061779195e-24, 5.146945291625474e-24, 3.841277877808338e-24], [3264.274796609495, 5.165062046888416e-24, 5.16050201271436e-24, 3.851337304982655e-24], [3273.3258471046893, 5.1786377318400686e-24, 5.1740982834834266e-24, 3.8614263735075775e-24], [3282.401993990409, 5.192304067650987e-24, 5.187785155306772e-24, 3.8716134270636126e-24], [3291.5033068529146, 5.2059082828109474e-24, 5.201409723923084e-24, 3.881693516848053e-24], [3300.6298554714253, 5.219603346843195e-24, 5.2151250924265775e-24, 3.891871731364864e-24], [3309.781709818626, 5.2333384804196254e-24, 5.2288804375761735e-24, 3.902079905903355e-24], [3318.9589400612376, 5.2471137569832185e-24, 5.2426758331974876e-24, 3.912318075583823e-24], [3328.1616165605274, 5.260929278740623e-24, 5.2565113819029326e-24, 3.92258631392334e-24], [3337.389809872851, 5.274785151090444e-24, 5.2703871894978095e-24, 3.932884698539463e-24], [3346.6435907502123, 5.288681480346681e-24, 5.2843033627017776e-24, 3.9432133080967335e-24], [3355.9230301407815, 5.302618373312237e-24, 5.298260008721991e-24, 3.953572221734101e-24], [3365.228199189463, 5.316595937164591e-24, 5.312257235138676e-24, 3.963961518911277e-24], [3374.559169238416, 5.3306142794181444e-24, 5.3262951498674535e-24, 3.974381279357455e-24], [3383.916011827629, 5.344673507909608e-24, 5.340373861144705e-24, 3.984831583051284e-24], [3393.2987986954386, 5.358773730791917e-24, 5.3544934775214976e-24, 3.995312510212465e-24], [3402.707601779105, 5.372915056532516e-24, 5.368654107861859e-24, 4.005824141298617e-24], [3412.1424932153514, 5.38709759391215e-24, 5.3828558613415836e-24, 4.0163665570028504e-24], [3421.60354534092, 5.4013214520243574e-24, 5.397098847447707e-24, 4.0269398382537834e-24], [3431.0908306931165, 5.415586740277091e-24, 5.411383175980159e-24, 4.0375440662153284e-24], [3440.6044220103886, 5.429893568391813e-24, 5.425708957050833e-24, 4.04817932228677e-24], [3450.144392232856, 5.44424204640541e-24, 5.4400763010855255e-24, 4.0588456881035194e-24], [3459.7108145028983, 5.4586322846699844e-24, 5.4544853188237226e-24, 4.0695432455370626e-24], [3469.3037621656877, 5.4730643938541896e-24, 5.468936121319934e-24, 4.0802720766960034e-24], [3478.9233087697776, 5.4875384849439994e-24, 5.483428819944474e-24, 4.0910322639263646e-24], [3488.5695280676377, 5.5020546692434864e-24, 5.4979635263842475e-24, 4.101823889812328e-24], [3498.2424940162564, 5.516613058375918e-24, 5.5125403526438366e-24, 4.11264703717685e-24], [3507.9422807776755, 5.53121376428385e-24, 5.5271594110455985e-24, 4.1235017890821445e-24], [3517.668962719565, 5.545856899230958e-24, 5.541820814231522e-24, 4.134388228830574e-24], [3527.422614415812, 5.560542575802802e-24, 5.556524675163963e-24, 4.145306439965199e-24], [3537.2033106470635, 5.5752709069068145e-24, 5.571271107125669e-24, 4.1562565062702295e-24], [3547.0111264013317, 5.590042005774158e-24, 5.5860602237216166e-24, 4.1672385117721315e-24], [3556.8461368745375, 5.604855985960569e-24, 5.600892138879868e-24, 4.17825254074004e-24], [3566.7084174711154, 5.619712961346613e-24, 5.6157669668518305e-24, 4.1892986776862644e-24], [3576.5980438045654, 5.634613046139367e-24, 5.630684822213941e-24, 4.200377007367484e-24], [3586.5150916980556, 5.64955635487254e-24, 5.6456458198677995e-24, 4.211487614784891e-24], [3596.4596371849902, 5.6645430024080794e-24, 5.6606500750417504e-24, 4.2226305851852365e-24], [3606.4317565095985, 5.679573103937436e-24, 5.675697703292198e-24, 4.233806004061774e-24], [3616.4315261275074, 5.694646774981081e-24, 5.6907888205030824e-24, 4.245013957154411e-24], [3626.459022706351, 5.709764131391369e-24, 5.7059235428887734e-24, 4.256254530451076e-24], [3636.5143231263282, 5.7249252893523245e-24, 5.721101986993851e-24, 4.267527810188843e-24], [3646.5975044808247, 5.7401303653817934e-24, 5.7363242696952634e-24, 4.278833882854435e-24], [3656.708644076973, 5.7553794763323026e-24, 5.751590508203192e-24, 4.290172835186475e-24], [3666.847819436273, 5.770672739394068e-24, 5.7669008200640574e-24, 4.3015447541774615e-24], [3677.0151082951597, 5.786010272097516e-24, 5.782255323163062e-24, 4.312949727078035e-24], [3687.210588605619, 5.801392192320665e-24, 5.7976541357315576e-24, 4.3243878414050646e-24], [3697.4343385357906, 5.816818618303302e-24, 5.813097376361253e-24, 4.3358591849608334e-24], [3707.6864364705334, 5.832289668685562e-24, 5.8285851640428085e-24, 4.347363845884296e-24], [3717.9669610120673, 5.8478054626409925e-24, 5.844117618298995e-24, 4.3589019128283636e-24], [3728.2759909805427, 5.863366120487097e-24, 5.859694859795613e-24, 4.370473475779196e-24], [3738.613605414673, 5.878971768771253e-24, 5.875317015430563e-24, 4.38207863287848e-24], [3748.9798835723136, 5.8946228776487056e-24, 5.890984555920076e-24, 4.39371794305045e-24], [3759.3749049310973, 5.9103184976545284e-24, 5.906696531481908e-24, 4.405390060449698e-24], [3769.7987491890135, 5.9260597629757756e-24, 5.922454077271551e-24, 4.417096434458406e-24], [3780.2514962650434, 5.9418465245731636e-24, 5.93825704442311e-24, 4.428836792676345e-24], [3790.7332262997757, 5.957678874285683e-24, 5.954105525096645e-24, 4.4406111856598466e-24], [3801.244019655991, 5.973556931183284e-24, 5.96999963869831e-24, 4.452419700292888e-24], [3811.7839569192993, 5.9894808166760006e-24, 5.985939506974173e-24, 4.464262426392061e-24], [3822.3531188987677, 6.005450652897732e-24, 6.0019252523930124e-24, 4.476139454539196e-24], [3832.9515866275115, 6.0214665624385546e-24, 6.0179569978784705e-24, 4.488050875719465e-24], [3843.5794413633466, 6.037528668276108e-24, 6.034034866740391e-24, 4.499996781230718e-24], [3854.236764589383, 6.053637093755624e-24, 6.0501589826548465e-24, 4.511977262654921e-24], [3864.923638014678, 6.0697919625819315e-24, 6.0663294696561476e-24, 4.523992411847388e-24], [3875.640143574834, 6.0859933988171785e-24, 6.082546452134553e-24, 4.536042320932792e-24], [3886.3863634326476, 6.1022415268796196e-24, 6.098810054835072e-24, 4.548127082303375e-24], [3897.1623799787462, 6.118536471544263e-24, 6.115120402858106e-24, 4.56024678861899e-24], [3907.9682758321833, 6.1348783579431436e-24, 6.131477621659732e-24, 4.57240153280664e-24], [3918.8041338411126, 6.151267311566027e-24, 6.147881837052414e-24, 4.5845914080612825e-24], [3929.6700370833933, 6.1677034582612434e-24, 6.164333175205823e-24, 4.596816507846002e-24], [3940.5660688672524, 6.184186924236673e-24, 6.180831762647853e-24, 4.60907692589296e-24], [3951.4923127318966, 6.200717836060347e-24, 6.1973777262652004e-24, 4.621372756203642e-24], [3962.4488524481826, 6.217296320661934e-24, 6.213971193304873e-24, 4.633704093050104e-24], [3973.4357720192265, 6.233922505333334e-24, 6.230612291374778e-24, 4.646071030975067e-24], [3984.4531556810816, 6.250596517729838e-24, 6.2473011484448906e-24, 4.658473664793023e-24], [3995.501087903348, 6.267318485870848e-24, 6.264037892847975e-24, 4.670912089590954e-24], [4006.579653389863, 6.284088538141163e-24, 6.280822653280873e-24, 4.683386400728753e-24], [4017.688937079315, 6.300906803292124e-24, 6.2976555588056634e-24, 4.695896693840547e-24], [4028.8290241459, 6.317773410442199e-24, 6.3145367388502354e-24, 4.708443064834889e-24], [4039.9999999999977, 6.334688489078047e-24, 6.3314663232093744e-24, 4.721025609895621e-24]])
   
    f1 = data[:,0]
    if IFO.upper() in ['LIGO3', 'LIGO3-DEFAULT']:
        S1 = data[:,1]**2
    elif IFO.upper()=='LIGO3-MID':
        S1 = data[:,2]**2
    elif IFO.upper()=='LIGO3-HIGH':
        S1 = data[:,4]**2

    fstop = [f1[0],f1[-1]]

# % ---- Interpolate to desired frequencies.  Set noise to infinity outside
# %      band of frequencies covered by f1.
# S = inf(size(f));
# k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop


# % -------------------------------------------------------------------------
# %                              KAGRA spectrum
# % -------------------------------------------------------------------------

def kagraspectrum(IFO,f):
# % KAGRASPECTRUM - Design noise curve for KAGRA. Data taken from the DCC 
# % entry https://dcc.ligo.org/T1600593-v1, which was used for DCC v36 of the 
# % Observing Scenarios paper (https://dcc.ligo.org/LIGO-P1200087). 
# %
# % Arguments are the same as for the parent function.

# % The following analytic fit is copied from Manzotti & Dietz,
# % https://dcc.ligo.org/cgi-bin/private/DocDB/ShowDocument?docid=p1200014
# % I have not checked its accuracy, and it is not used in this function.
# %   S.^0.5 = 6.499e-25 * ( ...
# %       + 9.72e-9*exp(-1.43-9.88*x-0.23*x.^2) ...
# %          + 1.17*exp( 0.14-3.10*x-0.26*x.^2) ...
# %          + 1.70*exp( 0.14+1.09*x-0.013*x.^2) ...
# %          + 1.25*exp(0.071+2.83*x-4.91*x.^2) ...
# %   );
# % where 
# %   x := log(f/100Hz)

# % ---- Extract pre-sampled noise spectrum from data.  First column is
# %      frequency, second column is KAGRA_Early curve, third is KAGRA_Mid curve,
# %      fourth is KAGRA_Late curve, fifth is KAGRA_Design ("KAGRA") curve.
    data = np.array([[1.0023052, 7.6534e-15, 3.8267e-15, 2.29602e-15, 7.6533992e-16], [1.0046158, 7.51513e-15, 3.75757e-15, 2.25454e-15, 7.5151318e-16], [1.0069317, 7.37938e-15, 3.68969e-15, 2.21381e-15, 7.3793811e-16], [1.0092529, 7.2461e-15, 3.62305e-15, 2.17383e-15, 7.2461015e-16], [1.0115795, 7.11525e-15, 3.55762e-15, 2.13457e-15, 7.1152481e-16], [1.0139114, 6.98678e-15, 3.49339e-15, 2.09603e-15, 6.9867768e-16], [1.0162487, 6.86064e-15, 3.43032e-15, 2.05819e-15, 6.8606444e-16], [1.0185914, 6.73681e-15, 3.3684e-15, 2.02104e-15, 6.7368084e-16], [1.0209395, 6.61523e-15, 3.30761e-15, 1.98457e-15, 6.6152273e-16], [1.023293, 6.49586e-15, 3.24793e-15, 1.94876e-15, 6.49586e-16], [1.0256519, 6.37867e-15, 3.18933e-15, 1.9136e-15, 6.3786665e-16], [1.0280163, 6.26361e-15, 3.1318e-15, 1.87908e-15, 6.2636072e-16], [1.0303861, 6.15064e-15, 3.07532e-15, 1.84519e-15, 6.1506436e-16], [1.0327614, 6.03974e-15, 3.01987e-15, 1.81192e-15, 6.0397374e-16], [1.0351422, 5.93085e-15, 2.96543e-15, 1.77926e-15, 5.9308516e-16], [1.0375284, 5.82395e-15, 2.91197e-15, 1.74718e-15, 5.8239494e-16], [1.0399202, 5.71899e-15, 2.8595e-15, 1.7157e-15, 5.7189949e-16], [1.0423174, 5.61595e-15, 2.80798e-15, 1.68479e-15, 5.6159527e-16], [1.0447202, 5.51479e-15, 2.75739e-15, 1.65444e-15, 5.5147883e-16], [1.0471285, 5.41547e-15, 2.70773e-15, 1.62464e-15, 5.4154676e-16], [1.0495424, 5.31796e-15, 2.65898e-15, 1.59539e-15, 5.3179572e-16], [1.0519619, 5.22222e-15, 2.61111e-15, 1.56667e-15, 5.2222243e-16], [1.0543869, 5.12824e-15, 2.56412e-15, 1.53847e-15, 5.1282368e-16], [1.0568175, 5.03596e-15, 2.51798e-15, 1.51079e-15, 5.0359629e-16], [1.0592537, 4.94537e-15, 2.47269e-15, 1.48361e-15, 4.9453718e-16], [1.0616956, 4.85643e-15, 2.42822e-15, 1.45693e-15, 4.856433e-16], [1.064143, 4.76912e-15, 2.38456e-15, 1.43073e-15, 4.7691165e-16], [1.0665961, 4.68339e-15, 2.3417e-15, 1.40502e-15, 4.683393e-16], [1.0690549, 4.59923e-15, 2.29962e-15, 1.37977e-15, 4.5992337e-16], [1.0715193, 4.51661e-15, 2.25831e-15, 1.35498e-15, 4.5166104e-16], [1.0739894, 4.4355e-15, 2.21775e-15, 1.33065e-15, 4.4354952e-16], [1.0764652, 4.35586e-15, 2.17793e-15, 1.30676e-15, 4.355861e-16], [1.0789467, 4.27768e-15, 2.13884e-15, 1.2833e-15, 4.2776809e-16], [1.081434, 4.20093e-15, 2.10046e-15, 1.26028e-15, 4.2009287e-16], [1.0839269, 4.12558e-15, 2.06279e-15, 1.23767e-15, 4.1255786e-16], [1.0864256, 4.05161e-15, 2.0258e-15, 1.21548e-15, 4.0516053e-16], [1.0889301, 3.97898e-15, 1.98949e-15, 1.1937e-15, 3.978984e-16], [1.0914403, 3.90769e-15, 1.95385e-15, 1.17231e-15, 3.9076902e-16], [1.0939564, 3.8377e-15, 1.91885e-15, 1.15131e-15, 3.8377002e-16], [1.0964782, 3.76899e-15, 1.8845e-15, 1.1307e-15, 3.7689903e-16], [1.0990058, 3.70154e-15, 1.85077e-15, 1.11046e-15, 3.7015374e-16], [1.1015393, 3.63532e-15, 1.81766e-15, 1.0906e-15, 3.6353191e-16], [1.1040786, 3.57031e-15, 1.78516e-15, 1.07109e-15, 3.5703129e-16], [1.1066238, 3.5065e-15, 1.75325e-15, 1.05195e-15, 3.5064972e-16], [1.1091748, 3.44385e-15, 1.72193e-15, 1.03316e-15, 3.4438506e-16], [1.1117317, 3.38235e-15, 1.69118e-15, 1.01471e-15, 3.3823519e-16], [1.1142945, 3.32198e-15, 1.66099e-15, 9.96594e-16, 3.3219806e-16], [1.1168632, 3.26272e-15, 1.63136e-15, 9.78815e-16, 3.2627165e-16], [1.1194379, 3.20454e-15, 1.60227e-15, 9.61362e-16, 3.2045396e-16], [1.1220185, 3.14743e-15, 1.57372e-15, 9.44229e-16, 3.1474305e-16], [1.124605, 3.09137e-15, 1.54569e-15, 9.27411e-16, 3.0913701e-16], [1.1271975, 3.03634e-15, 1.51817e-15, 9.10902e-16, 3.0363396e-16], [1.1297959, 2.98232e-15, 1.49116e-15, 8.94696e-16, 2.9823205e-16], [1.1324004, 2.92929e-15, 1.46465e-15, 8.78788e-16, 2.9292947e-16], [1.1350108, 2.87724e-15, 1.43862e-15, 8.63173e-16, 2.8772445e-16], [1.1376273, 2.82615e-15, 1.41308e-15, 8.47846e-16, 2.8261525e-16], [1.1402498, 2.776e-15, 1.388e-15, 8.328e-16, 2.7760016e-16], [1.1428783, 2.72678e-15, 1.36339e-15, 8.18033e-16, 2.726775e-16], [1.1455129, 2.67846e-15, 1.33923e-15, 8.03537e-16, 2.6784563e-16], [1.1481536, 2.63103e-15, 1.31551e-15, 7.89309e-16, 2.6310292e-16], [1.1508004, 2.58448e-15, 1.29224e-15, 7.75343e-16, 2.584478e-16], [1.1534533, 2.53879e-15, 1.26939e-15, 7.61636e-16, 2.5387871e-16], [1.1561122, 2.49394e-15, 1.24697e-15, 7.48182e-16, 2.4939413e-16], [1.1587774, 2.44993e-15, 1.22496e-15, 7.34978e-16, 2.4499255e-16], [1.1614486, 2.40673e-15, 1.20336e-15, 7.22018e-16, 2.4067251e-16], [1.164126, 2.36433e-15, 1.18216e-15, 7.09298e-16, 2.3643257e-16], [1.1668096, 2.32271e-15, 1.16136e-15, 6.96814e-16, 2.3227132e-16], [1.1694994, 2.28187e-15, 1.14094e-15, 6.84562e-16, 2.2818737e-16], [1.1721954, 2.24179e-15, 1.1209e-15, 6.72538e-16, 2.2417936e-16], [1.1748976, 2.20246e-15, 1.10123e-15, 6.60738e-16, 2.2024595e-16], [1.177606, 2.16386e-15, 1.08193e-15, 6.49158e-16, 2.1638584e-16], [1.1803206, 2.12598e-15, 1.06299e-15, 6.37793e-16, 2.1259775e-16], [1.1830416, 2.0888e-15, 1.0444e-15, 6.26641e-16, 2.0888042e-16], [1.1857687, 2.05233e-15, 1.02616e-15, 6.15698e-16, 2.0523261e-16], [1.1885022, 2.01653e-15, 1.00827e-15, 6.04959e-16, 2.0165312e-16], [1.191242, 1.98141e-15, 9.90704e-16, 5.94422e-16, 1.9814075e-16], [1.1939881, 1.94694e-15, 9.73472e-16, 5.84083e-16, 1.9469435e-16], [1.1967405, 1.91313e-15, 9.56564e-16, 5.73938e-16, 1.9131278e-16], [1.1994993, 1.87995e-15, 9.39975e-16, 5.63985e-16, 1.8799492e-16], [1.2022644, 1.8474e-15, 9.23698e-16, 5.54219e-16, 1.8473967e-16], [1.2050359, 1.81546e-15, 9.0773e-16, 5.44638e-16, 1.8154597e-16], [1.2078138, 1.78413e-15, 8.92064e-16, 5.35238e-16, 1.7841275e-16], [1.2105981, 1.75339e-15, 8.76695e-16, 5.26017e-16, 1.75339e-16], [1.2133889, 1.72324e-15, 8.61618e-16, 5.16971e-16, 1.7232369e-16], [1.216186, 1.69366e-15, 8.46829e-16, 5.08098e-16, 1.6936585e-16], [1.2189896, 1.66464e-15, 8.32322e-16, 4.99393e-16, 1.664645e-16], [1.2217997, 1.63619e-15, 8.18093e-16, 4.90856e-16, 1.636187e-16], [1.2246162, 1.60828e-15, 8.04138e-16, 4.82483e-16, 1.6082751e-16], [1.2274392, 1.5809e-15, 7.9045e-16, 4.7427e-16, 1.5809003e-16], [1.2302688, 1.55405e-15, 7.77027e-16, 4.66216e-16, 1.5540537e-16], [1.2331048, 1.52773e-15, 7.63863e-16, 4.58318e-16, 1.5277266e-16], [1.2359474, 1.50191e-15, 7.50955e-16, 4.50573e-16, 1.5019104e-16], [1.2387966, 1.4766e-15, 7.38298e-16, 4.42979e-16, 1.4765969e-16], [1.2416523, 1.45178e-15, 7.25889e-16, 4.35533e-16, 1.4517778e-16], [1.2445146, 1.42745e-15, 7.13723e-16, 4.28234e-16, 1.4274453e-16], [1.2473835, 1.40359e-15, 7.01796e-16, 4.21077e-16, 1.4035914e-16], [1.250259, 1.38021e-15, 6.90104e-16, 4.14063e-16, 1.3802087e-16], [1.2531412, 1.35729e-15, 6.78645e-16, 4.07187e-16, 1.3572897e-16], [1.25603, 1.33483e-15, 6.67414e-16, 4.00448e-16, 1.3348271e-16], [1.2589254, 1.31281e-15, 6.56407e-16, 3.93844e-16, 1.3128138e-16], [1.2618275, 1.29124e-15, 6.45621e-16, 3.87373e-16, 1.291243e-16], [1.2647363, 1.27011e-15, 6.35054e-16, 3.81032e-16, 1.2701078e-16], [1.2676519, 1.2494e-15, 6.24701e-16, 3.74821e-16, 1.2494017e-16], [1.2705741, 1.22912e-15, 6.14559e-16, 3.68735e-16, 1.2291183e-16], [1.2735031, 1.20925e-15, 6.04626e-16, 3.62775e-16, 1.2092513e-16], [1.2764388, 1.18979e-15, 5.94897e-16, 3.56938e-16, 1.1897947e-16], [1.2793813, 1.17074e-15, 5.85371e-16, 3.51223e-16, 1.1707426e-16], [1.2823306, 1.15209e-15, 5.76045e-16, 3.45627e-16, 1.1520892e-16], [1.2852867, 1.13383e-15, 5.66914e-16, 3.40149e-16, 1.1338288e-16], [1.2882496, 1.11596e-15, 5.57978e-16, 3.34787e-16, 1.1159562e-16], [1.2912193, 1.09847e-15, 5.49233e-16, 3.2954e-16, 1.0984659e-16], [1.2941958, 1.08135e-15, 5.40677e-16, 3.24406e-16, 1.081353e-16], [1.2971793, 1.06461e-15, 5.32306e-16, 3.19384e-16, 1.0646125e-16], [1.3001696, 1.04824e-15, 5.2412e-16, 3.14472e-16, 1.0482396e-16], [1.3031668, 1.03223e-15, 5.16115e-16, 3.09669e-16, 1.0322297e-16], [1.3061709, 1.01658e-15, 5.08289e-16, 3.04974e-16, 1.0165784e-16], [1.3091819, 1.00128e-15, 5.00641e-16, 3.00384e-16, 1.0012813e-16], [1.3121999, 9.86334e-16, 4.93167e-16, 2.959e-16, 9.8633437e-17], [1.3152248, 9.71734e-16, 4.85867e-16, 2.9152e-16, 9.717336e-17], [1.3182567, 9.57475e-16, 4.78738e-16, 2.87243e-16, 9.5747523e-17], [1.3212956, 9.43556e-16, 4.71778e-16, 2.83067e-16, 9.4355562e-17], [1.3243415, 9.29971e-16, 4.64986e-16, 2.78991e-16, 9.2997132e-17], [1.3273945, 9.16719e-16, 4.5836e-16, 2.75016e-16, 9.1671904e-17], [1.3304544, 9.03796e-16, 4.51898e-16, 2.71139e-16, 9.0379567e-17], [1.3335214, 8.91198e-16, 4.45599e-16, 2.67359e-16, 8.9119827e-17], [1.3365955, 8.78924e-16, 4.39462e-16, 2.63677e-16, 8.7892408e-17], [1.3396767, 8.66971e-16, 4.33485e-16, 2.60091e-16, 8.6697051e-17], [1.342765, 8.55335e-16, 4.27668e-16, 2.56601e-16, 8.5533517e-17], [1.3458604, 8.44016e-16, 4.22008e-16, 2.53205e-16, 8.4401585e-17], [1.3489629, 8.33011e-16, 4.16505e-16, 2.49903e-16, 8.3301051e-17], [1.3520726, 8.22317e-16, 4.11159e-16, 2.46695e-16, 8.2231733e-17], [1.3551894, 8.11935e-16, 4.05967e-16, 2.4358e-16, 8.1193467e-17], [1.3583134, 8.01861e-16, 4.00931e-16, 2.40558e-16, 8.0186111e-17], [1.3614447, 7.92095e-16, 3.96048e-16, 2.37629e-16, 7.9209543e-17], [1.3645831, 7.82637e-16, 3.91318e-16, 2.34791e-16, 7.8263661e-17], [1.3677288, 7.73484e-16, 3.86742e-16, 2.32045e-16, 7.7348388e-17], [1.3708818, 7.64637e-16, 3.82318e-16, 2.29391e-16, 7.6463667e-17], [1.374042, 7.56095e-16, 3.78047e-16, 2.26828e-16, 7.5609467e-17], [1.3772095, 7.47858e-16, 3.73929e-16, 2.24357e-16, 7.4785779e-17], [1.3803843, 7.39926e-16, 3.69963e-16, 2.21978e-16, 7.399262e-17], [1.3835664, 7.323e-16, 3.6615e-16, 2.1969e-16, 7.3230035e-17], [1.3867558, 7.24981e-16, 3.6249e-16, 2.17494e-16, 7.2498095e-17], [1.3899526, 7.17969e-16, 3.58984e-16, 2.15391e-16, 7.17969e-17], [1.3931568, 7.11266e-16, 3.55633e-16, 2.1338e-16, 7.1126579e-17], [1.3963684, 7.04873e-16, 3.52436e-16, 2.11462e-16, 7.0487294e-17], [1.3995873, 6.98792e-16, 3.49396e-16, 2.09638e-16, 6.9879239e-17], [1.4028137, 6.93026e-16, 3.46513e-16, 2.07908e-16, 6.9302644e-17], [1.4060475, 6.87578e-16, 3.43789e-16, 2.06273e-16, 6.8757776e-17], [1.4092888, 6.82449e-16, 3.41225e-16, 2.04735e-16, 6.8244939e-17], [1.4125375, 6.77645e-16, 3.38822e-16, 2.03293e-16, 6.776448e-17], [1.4157938, 6.73168e-16, 3.36584e-16, 2.0195e-16, 6.7316792e-17], [1.4190575, 6.69023e-16, 3.34512e-16, 2.00707e-16, 6.6902311e-17], [1.4223288, 6.65215e-16, 3.32608e-16, 1.99565e-16, 6.6521528e-17], [1.4256076, 6.6175e-16, 3.30875e-16, 1.98525e-16, 6.6174984e-17], [1.428894, 6.58633e-16, 3.29316e-16, 1.9759e-16, 6.5863281e-17], [1.4321879, 6.55871e-16, 3.27935e-16, 1.96761e-16, 6.5587083e-17], [1.4354894, 6.53471e-16, 3.26736e-16, 1.96041e-16, 6.5347123e-17], [1.4387986, 6.51442e-16, 3.25721e-16, 1.95433e-16, 6.5144205e-17], [1.4421154, 6.49792e-16, 3.24896e-16, 1.94938e-16, 6.4979216e-17], [1.4454398, 6.48531e-16, 3.24266e-16, 1.94559e-16, 6.4853128e-17], [1.4487719, 6.4767e-16, 3.23835e-16, 1.94301e-16, 6.476701e-17], [1.4521116, 6.4722e-16, 3.2361e-16, 1.94166e-16, 6.4722034e-17], [1.4554591, 6.47195e-16, 3.23597e-16, 1.94158e-16, 6.4719487e-17], [1.4588143, 6.47608e-16, 3.23804e-16, 1.94282e-16, 6.4760782e-17], [1.4621772, 6.48475e-16, 3.24237e-16, 1.94542e-16, 6.4847473e-17], [1.4655478, 6.49813e-16, 3.24906e-16, 1.94944e-16, 6.4981266e-17], [1.4689263, 6.5164e-16, 3.2582e-16, 1.95492e-16, 6.5164039e-17], [1.4723125, 6.53979e-16, 3.26989e-16, 1.96194e-16, 6.5397862e-17], [1.4757065, 6.5685e-16, 3.28425e-16, 1.97055e-16, 6.5685014e-17], [1.4791084, 6.6028e-16, 3.3014e-16, 1.98084e-16, 6.6028015e-17], [1.4825181, 6.64297e-16, 3.32148e-16, 1.99289e-16, 6.6429651e-17], [1.4859356, 6.6893e-16, 3.34465e-16, 2.00679e-16, 6.6893006e-17], [1.4893611, 6.74215e-16, 3.37108e-16, 2.02265e-16, 6.7421507e-17], [1.4927944, 6.8019e-16, 3.40095e-16, 2.04057e-16, 6.8018962e-17], [1.4962357, 6.86896e-16, 3.43448e-16, 2.06069e-16, 6.8689616e-17], [1.4996848, 6.94382e-16, 3.47191e-16, 2.08315e-16, 6.9438208e-17], [1.503142, 7.02701e-16, 3.5135e-16, 2.1081e-16, 7.0270051e-17], [1.5066071, 7.11911e-16, 3.55956e-16, 2.13573e-16, 7.1191104e-17], [1.5100802, 7.22081e-16, 3.6104e-16, 2.16624e-16, 7.2208083e-17], [1.5135612, 7.33286e-16, 3.66643e-16, 2.19986e-16, 7.332857e-17], [1.5170504, 7.45612e-16, 3.72806e-16, 2.23683e-16, 7.4561161e-17], [1.5205475, 7.59156e-16, 3.79578e-16, 2.27747e-16, 7.5915632e-17], [1.5240528, 7.74031e-16, 3.87016e-16, 2.32209e-16, 7.7403142e-17], [1.5275661, 7.90365e-16, 3.95182e-16, 2.37109e-16, 7.9036488e-17], [1.5310875, 8.08304e-16, 4.04152e-16, 2.42491e-16, 8.0830401e-17], [1.534617, 8.28019e-16, 4.1401e-16, 2.48406e-16, 8.2801926e-17], [1.5381546, 8.49709e-16, 4.24854e-16, 2.54913e-16, 8.4970887e-17], [1.5417005, 8.73605e-16, 4.36802e-16, 2.62081e-16, 8.7360464e-17], [1.5452544, 8.99979e-16, 4.4999e-16, 2.69994e-16, 8.9997929e-17], [1.5488166, 9.29156e-16, 4.64578e-16, 2.78747e-16, 9.2915583e-17], [1.552387, 9.6152e-16, 4.8076e-16, 2.88456e-16, 9.6151953e-17], [1.5559656, 9.97534e-16, 4.98767e-16, 2.9926e-16, 9.9753362e-17], [1.5595525, 1.03776e-15, 5.1888e-16, 3.11328e-16, 1.0377599e-16], [1.5631476, 1.08289e-15, 5.41443e-16, 3.24866e-16, 1.0828862e-16], [1.5667511, 1.13376e-15, 5.66882e-16, 3.40129e-16, 1.1337638e-16], [1.5703628, 1.19146e-15, 5.95729e-16, 3.57438e-16, 1.1914588e-16], [1.5739829, 1.25732e-15, 6.28662e-16, 3.77197e-16, 1.2573238e-16], [1.5776113, 1.3331e-15, 6.66551e-16, 3.99931e-16, 1.3331018e-16], [1.581248, 1.42108e-15, 7.10538e-16, 4.26323e-16, 1.4210766e-16], [1.5848932, 1.5243e-15, 7.6215e-16, 4.5729e-16, 1.5243004e-16], [1.5885467, 1.64695e-15, 8.23473e-16, 4.94084e-16, 1.6469462e-16], [1.5922087, 1.79487e-15, 8.97437e-16, 5.38462e-16, 1.7948737e-16], [1.5958791, 1.97658e-15, 9.88288e-16, 5.92973e-16, 1.9765766e-16], [1.599558, 2.20485e-15, 1.10242e-15, 6.61455e-16, 2.2048496e-16], [1.6032454, 2.4999e-15, 1.24995e-15, 7.49971e-16, 2.4999035e-16], [1.6069413, 2.89565e-15, 1.44782e-15, 8.68695e-16, 2.8956486e-16], [1.6106456, 3.45367e-15, 1.72683e-15, 1.0361e-15, 3.4536685e-16], [1.6143586, 4.29868e-15, 2.14934e-15, 1.2896e-15, 4.2986807e-16], [1.61808, 5.72732e-15, 2.86366e-15, 1.7182e-15, 5.7273189e-16], [1.6218101, 8.6601e-15, 4.33005e-15, 2.59803e-15, 8.6600959e-16], [1.6255488, 1.81002e-14, 9.05009e-15, 5.43005e-15, 1.810018e-15], [1.629296, 1.64906e-13, 8.24532e-14, 4.94719e-14, 1.6490642e-14], [1.6330519, 1.46074e-14, 7.30369e-15, 4.38222e-15, 1.4607385e-15], [1.6368165, 7.57974e-15, 3.78987e-15, 2.27392e-15, 7.579735e-16], [1.6405898, 5.08966e-15, 2.54483e-15, 1.5269e-15, 5.0896572e-16], [1.6443717, 3.81542e-15, 1.90771e-15, 1.14463e-15, 3.8154228e-16], [1.6481624, 3.04154e-15, 1.52077e-15, 9.12463e-16, 3.0415419e-16], [1.6519618, 2.52184e-15, 1.26092e-15, 7.56553e-16, 2.5218438e-16], [1.65577, 2.14888e-15, 1.07444e-15, 6.44665e-16, 2.1488828e-16], [1.6595869, 1.86829e-15, 9.34143e-16, 5.60486e-16, 1.8682862e-16], [1.6634127, 1.64959e-15, 8.24793e-16, 4.94876e-16, 1.6495866e-16], [1.6672472, 1.47439e-15, 7.37193e-16, 4.42316e-16, 1.4743858e-16], [1.6710906, 1.33092e-15, 6.65459e-16, 3.99275e-16, 1.3309172e-16], [1.6749429, 1.2113e-15, 6.05652e-16, 3.63391e-16, 1.211305e-16], [1.678804, 1.11008e-15, 5.55039e-16, 3.33024e-16, 1.1100788e-16], [1.6826741, 1.02332e-15, 5.1166e-16, 3.06996e-16, 1.0233207e-16], [1.686553, 9.48151e-16, 4.74076e-16, 2.84445e-16, 9.4815109e-17], [1.6904409, 8.82406e-16, 4.41203e-16, 2.64722e-16, 8.8240609e-17], [1.6943378, 8.24428e-16, 4.12214e-16, 2.47328e-16, 8.2442829e-17], [1.6982437, 7.72927e-16, 3.86463e-16, 2.31878e-16, 7.729268e-17], [1.7021585, 7.26881e-16, 3.63441e-16, 2.18064e-16, 7.2688127e-17], [1.7060824, 6.85475e-16, 3.42737e-16, 2.05642e-16, 6.8547462e-17], [1.7100153, 6.48045e-16, 3.24022e-16, 1.94413e-16, 6.4804481e-17], [1.7139573, 6.1405e-16, 3.07025e-16, 1.84215e-16, 6.1404979e-17], [1.7179084, 5.83042e-16, 2.91521e-16, 1.74912e-16, 5.8304151e-17], [1.7218686, 5.54646e-16, 2.77323e-16, 1.66394e-16, 5.5464639e-17], [1.7258379, 5.2855e-16, 2.64275e-16, 1.58565e-16, 5.285505e-17], [1.7298164, 5.04488e-16, 2.52244e-16, 1.51346e-16, 5.0448813e-17], [1.733804, 4.82233e-16, 2.41116e-16, 1.4467e-16, 4.8223287e-17], [1.7378008, 4.61591e-16, 2.30795e-16, 1.38477e-16, 4.6159063e-17], [1.7418069, 4.42394e-16, 2.21197e-16, 1.32718e-16, 4.4239407e-17], [1.7458222, 4.24498e-16, 2.12249e-16, 1.27349e-16, 4.2449814e-17], [1.7498467, 4.07777e-16, 2.03888e-16, 1.22333e-16, 4.0777654e-17], [1.7538805, 3.92119e-16, 1.96059e-16, 1.17636e-16, 3.9211874e-17], [1.7579236, 3.77428e-16, 1.88714e-16, 1.13228e-16, 3.7742762e-17], [1.761976, 3.63618e-16, 1.81809e-16, 1.09085e-16, 3.6361753e-17], [1.7660378, 3.50613e-16, 1.75306e-16, 1.05184e-16, 3.5061259e-17], [1.770109, 3.38345e-16, 1.69173e-16, 1.01504e-16, 3.3834537e-17], [1.7741895, 3.26756e-16, 1.63378e-16, 9.80267e-17, 3.2675577e-17], [1.7782794, 3.1579e-16, 1.57895e-16, 9.4737e-17, 3.1579002e-17], [1.7823788, 3.054e-16, 1.527e-16, 9.162e-17, 3.0539988e-17], [1.7864876, 2.95542e-16, 1.47771e-16, 8.86626e-17, 2.9554195e-17], [1.7906059, 2.86177e-16, 1.43089e-16, 8.58531e-17, 2.8617707e-17], [1.7947336, 2.7727e-16, 1.38635e-16, 8.3181e-17, 2.7726986e-17], [1.7988709, 2.68788e-16, 1.34394e-16, 8.06365e-17, 2.687882e-17], [1.8030177, 2.60703e-16, 1.30351e-16, 7.82109e-17, 2.6070293e-17], [1.8071741, 2.52987e-16, 1.26494e-16, 7.58962e-17, 2.5298748e-17], [1.8113401, 2.45618e-16, 1.22809e-16, 7.36853e-17, 2.4561763e-17], [1.8155157, 2.38571e-16, 1.19286e-16, 7.15714e-17, 2.3857119e-17], [1.8197009, 2.31828e-16, 1.15914e-16, 6.95484e-17, 2.3182786e-17], [1.8238957, 2.25369e-16, 1.12684e-16, 6.76107e-17, 2.25369e-17], [1.8281002, 2.19177e-16, 1.09589e-16, 6.57532e-17, 2.1917746e-17], [1.8323144, 2.13237e-16, 1.06619e-16, 6.39712e-17, 2.1323746e-17], [1.8365383, 2.07534e-16, 1.03767e-16, 6.22603e-17, 2.0753442e-17], [1.840772, 2.02055e-16, 1.01027e-16, 6.06165e-17, 2.0205488e-17], [1.8450154, 1.96786e-16, 9.83932e-17, 5.90359e-17, 1.9678638e-17], [1.8492686, 1.91717e-16, 9.58587e-17, 5.75152e-17, 1.9171738e-17], [1.8535316, 1.86837e-16, 9.34186e-17, 5.60511e-17, 1.8683715e-17], [1.8578045, 1.82136e-16, 9.10679e-17, 5.46407e-17, 1.8213574e-17], [1.8620871, 1.77604e-16, 8.88019e-17, 5.32812e-17, 1.7760387e-17], [1.8663797, 1.73233e-16, 8.66165e-17, 5.19699e-17, 1.7323292e-17], [1.8706821, 1.69015e-16, 8.45074e-17, 5.07044e-17, 1.690148e-17], [1.8749945, 1.64942e-16, 8.2471e-17, 4.94826e-17, 1.6494202e-17], [1.8793168, 1.61008e-16, 8.05038e-17, 4.83023e-17, 1.6100751e-17], [1.8836491, 1.57205e-16, 7.86023e-17, 4.71614e-17, 1.5720469e-17], [1.8879913, 1.53527e-16, 7.67637e-17, 4.60582e-17, 1.5352739e-17], [1.8923436, 1.4997e-16, 7.49849e-17, 4.49909e-17, 1.4996982e-17], [1.8967059, 1.46527e-16, 7.32633e-17, 4.3958e-17, 1.4652652e-17], [1.9010783, 1.43192e-16, 7.15962e-17, 4.29577e-17, 1.431924e-17], [1.9054607, 1.39963e-16, 6.99813e-17, 4.19888e-17, 1.3996262e-17], [1.9098533, 1.36833e-16, 6.84163e-17, 4.10498e-17, 1.3683267e-17], [1.9142559, 1.33798e-16, 6.68991e-17, 4.01395e-17, 1.3379827e-17], [1.9186687, 1.30855e-16, 6.54277e-17, 3.92566e-17, 1.3085538e-17], [1.9230917, 1.28e-16, 6.40001e-17, 3.84001e-17, 1.280002e-17], [1.9275249, 1.25229e-16, 6.26146e-17, 3.75687e-17, 1.2522913e-17], [1.9319683, 1.22539e-16, 6.12694e-17, 3.67616e-17, 1.2253877e-17], [1.936422, 1.19926e-16, 5.9963e-17, 3.59778e-17, 1.199259e-17], [1.9408859, 1.17387e-16, 5.86937e-17, 3.52162e-17, 1.1738747e-17], [1.9453601, 1.14921e-16, 5.74603e-17, 3.44762e-17, 1.1492058e-17], [1.9498446, 1.12522e-16, 5.62612e-17, 3.37567e-17, 1.1252248e-17], [1.9543395, 1.10191e-16, 5.50953e-17, 3.30572e-17, 1.1019058e-17], [1.9588447, 1.07922e-16, 5.39612e-17, 3.23767e-17, 1.0792239e-17], [1.9633603, 1.05716e-16, 5.28578e-17, 3.17147e-17, 1.0571556e-17], [1.9678863, 1.03568e-16, 5.17839e-17, 3.10704e-17, 1.0356786e-17], [1.9724227, 1.01477e-16, 5.07386e-17, 3.04431e-17, 1.0147714e-17], [1.9769696, 9.94414e-17, 4.97207e-17, 2.98324e-17, 9.944139e-18], [1.981527, 9.74587e-17, 4.87293e-17, 2.92376e-17, 9.7458661e-18], [1.9860949, 9.55271e-17, 4.77636e-17, 2.86581e-17, 9.5527112e-18], [1.9906733, 9.3645e-17, 4.68225e-17, 2.80935e-17, 9.3644979e-18], [1.9952623, 9.18106e-17, 4.59053e-17, 2.75432e-17, 9.1810582e-18], [1.9998619, 9.00223e-17, 4.50112e-17, 2.70067e-17, 9.0022314e-18], [2.004472, 8.82786e-17, 4.41393e-17, 2.64836e-17, 8.8278638e-18], [2.0090928, 8.65781e-17, 4.3289e-17, 2.59734e-17, 8.6578087e-18], [2.0137242, 8.49192e-17, 4.24596e-17, 2.54758e-17, 8.4919253e-18], [2.0183664, 8.33008e-17, 4.16504e-17, 2.49902e-17, 8.3300793e-18], [2.0230192, 8.17214e-17, 4.08607e-17, 2.45164e-17, 8.1721416e-18], [2.0276827, 8.01799e-17, 4.00899e-17, 2.4054e-17, 8.0179887e-18], [2.032357, 7.8675e-17, 3.93375e-17, 2.36025e-17, 7.8675022e-18], [2.0370421, 7.72057e-17, 3.86028e-17, 2.31617e-17, 7.7205685e-18], [2.0417379, 7.57708e-17, 3.78854e-17, 2.27312e-17, 7.5770786e-18], [2.0464446, 7.43693e-17, 3.71846e-17, 2.23108e-17, 7.4369277e-18], [2.0511622, 7.30001e-17, 3.65001e-17, 2.19e-17, 7.3000154e-18], [2.0558906, 7.16624e-17, 3.58312e-17, 2.14987e-17, 7.1662449e-18], [2.0606299, 7.03552e-17, 3.51776e-17, 2.11066e-17, 7.0355235e-18], [2.0653802, 6.90776e-17, 3.45388e-17, 2.07233e-17, 6.9077618e-18], [2.0701413, 6.78287e-17, 3.39144e-17, 2.03486e-17, 6.7828738e-18], [2.0749135, 6.66078e-17, 3.33039e-17, 1.99823e-17, 6.6607768e-18], [2.0796967, 6.54139e-17, 3.2707e-17, 1.96242e-17, 6.5413911e-18], [2.0844909, 6.42464e-17, 3.21232e-17, 1.92739e-17, 6.42464e-18], [2.0892961, 6.31045e-17, 3.15522e-17, 1.89313e-17, 6.3104496e-18], [2.0941125, 6.19875e-17, 3.09937e-17, 1.85962e-17, 6.1987486e-18], [2.0989399, 6.08947e-17, 3.04473e-17, 1.82684e-17, 6.0894682e-18], [2.1037784, 5.98254e-17, 2.99127e-17, 1.79476e-17, 5.9825422e-18], [2.1086281, 5.87791e-17, 2.93895e-17, 1.76337e-17, 5.8779066e-18], [2.113489, 5.7755e-17, 2.88775e-17, 1.73265e-17, 5.7754996e-18], [2.1183611, 5.67526e-17, 2.83763e-17, 1.70258e-17, 5.6752616e-18], [2.1232445, 5.57713e-17, 2.78857e-17, 1.67314e-17, 5.5771349e-18], [2.128139, 5.48106e-17, 2.74053e-17, 1.64432e-17, 5.481064e-18], [2.1330449, 5.38699e-17, 2.6935e-17, 1.6161e-17, 5.386995e-18], [2.1379621, 5.29487e-17, 2.64744e-17, 1.58846e-17, 5.294876e-18], [2.1428906, 5.20466e-17, 2.60233e-17, 1.5614e-17, 5.2046565e-18], [2.1478305, 5.11629e-17, 2.55814e-17, 1.53489e-17, 5.116288e-18], [2.1527817, 5.02972e-17, 2.51486e-17, 1.50892e-17, 5.0297234e-18], [2.1577444, 4.94492e-17, 2.47246e-17, 1.48347e-17, 4.9449172e-18], [2.1627185, 4.86182e-17, 2.43091e-17, 1.45855e-17, 4.8618251e-18], [2.1677041, 4.7804e-17, 2.3902e-17, 1.43412e-17, 4.7804044e-18], [2.1727012, 4.70061e-17, 2.35031e-17, 1.41018e-17, 4.7006138e-18], [2.1777098, 4.62241e-17, 2.31121e-17, 1.38672e-17, 4.6224132e-18], [2.1827299, 4.54576e-17, 2.27288e-17, 1.36373e-17, 4.5457636e-18], [2.1877616, 4.47063e-17, 2.23531e-17, 1.34119e-17, 4.4706274e-18], [2.1928049, 4.39697e-17, 2.19848e-17, 1.31909e-17, 4.396968e-18], [2.1978599, 4.32475e-17, 2.16237e-17, 1.29742e-17, 4.32475e-18], [2.2029265, 4.25394e-17, 2.12697e-17, 1.27618e-17, 4.2539391e-18], [2.2080047, 4.1845e-17, 2.09225e-17, 1.25535e-17, 4.1845017e-18], [2.2130947, 4.1164e-17, 2.0582e-17, 1.23492e-17, 4.1164056e-18], [2.2181964, 4.04962e-17, 2.02481e-17, 1.21489e-17, 4.0496193e-18], [2.2233099, 3.98411e-17, 1.99206e-17, 1.19523e-17, 3.9841122e-18], [2.2284351, 3.91985e-17, 1.95993e-17, 1.17596e-17, 3.9198548e-18], [2.2335722, 3.85682e-17, 1.92841e-17, 1.15705e-17, 3.8568182e-18], [2.2387211, 3.79497e-17, 1.89749e-17, 1.13849e-17, 3.7949744e-18], [2.2438819, 3.73429e-17, 1.86715e-17, 1.12029e-17, 3.7342963e-18], [2.2490546, 3.67476e-17, 1.83738e-17, 1.10243e-17, 3.6747573e-18], [2.2542392, 3.61633e-17, 1.80817e-17, 1.0849e-17, 3.6163319e-18], [2.2594358, 3.55899e-17, 1.7795e-17, 1.0677e-17, 3.5589949e-18], [2.2646443, 3.50272e-17, 1.75136e-17, 1.05082e-17, 3.5027221e-18], [2.2698649, 3.44749e-17, 1.72374e-17, 1.03425e-17, 3.4474899e-18], [2.2750974, 3.39327e-17, 1.69664e-17, 1.01798e-17, 3.3932752e-18], [2.2803421, 3.34005e-17, 1.67003e-17, 1.00202e-17, 3.3400557e-18], [2.2855988, 3.28781e-17, 1.6439e-17, 9.86343e-18, 3.2878095e-18], [2.2908677, 3.23651e-17, 1.61826e-17, 9.70954e-18, 3.2365155e-18], [2.2961486, 3.18615e-17, 1.59308e-17, 9.55846e-18, 3.1861531e-18], [2.3014418, 3.1367e-17, 1.56835e-17, 9.4101e-18, 3.136702e-18], [2.3067472, 3.08814e-17, 1.54407e-17, 9.26443e-18, 3.0881428e-18], [2.3120648, 3.04045e-17, 1.52023e-17, 9.12137e-18, 3.0404563e-18], [2.3173946, 2.99362e-17, 1.49681e-17, 8.98087e-18, 2.9936241e-18], [2.3227368, 2.94763e-17, 1.47381e-17, 8.84288e-18, 2.9476279e-18], [2.3280913, 2.90245e-17, 1.45122e-17, 8.70735e-18, 2.9024501e-18], [2.3334581, 2.85807e-17, 1.42904e-17, 8.57422e-18, 2.8580736e-18], [2.3388372, 2.81448e-17, 1.40724e-17, 8.44344e-18, 2.8144816e-18], [2.3442288, 2.77166e-17, 1.38583e-17, 8.31497e-18, 2.7716577e-18], [2.3496328, 2.72958e-17, 1.36479e-17, 8.18876e-18, 2.729586e-18], [2.3550493, 2.68825e-17, 1.34413e-17, 8.06475e-18, 2.688251e-18], [2.3604782, 2.64764e-17, 1.32382e-17, 7.94291e-18, 2.6476375e-18], [2.3659197, 2.60773e-17, 1.30386e-17, 7.82319e-18, 2.6077307e-18], [2.3713737, 2.56851e-17, 1.28426e-17, 7.70555e-18, 2.5685162e-18], [2.3768403, 2.52998e-17, 1.26499e-17, 7.58994e-18, 2.52998e-18], [2.3823195, 2.49211e-17, 1.24605e-17, 7.47632e-18, 2.4921082e-18], [2.3878113, 2.45489e-17, 1.22744e-17, 7.36466e-18, 2.4548876e-18], [2.3933158, 2.4183e-17, 1.20915e-17, 7.25491e-18, 2.4183049e-18], [2.3988329, 2.38235e-17, 1.19117e-17, 7.14704e-18, 2.3823476e-18], [2.4043628, 2.347e-17, 1.1735e-17, 7.04101e-18, 2.347003e-18], [2.4099054, 2.31226e-17, 1.15613e-17, 6.93677e-18, 2.312259e-18], [2.4154608, 2.2781e-17, 1.13905e-17, 6.83431e-18, 2.2781037e-18], [2.421029, 2.24452e-17, 1.12226e-17, 6.73357e-18, 2.2445255e-18], [2.4266101, 2.21151e-17, 1.10576e-17, 6.63454e-18, 2.2115131e-18], [2.432204, 2.17905e-17, 1.08953e-17, 6.53716e-18, 2.1790554e-18], [2.4378108, 2.14714e-17, 1.07357e-17, 6.44142e-18, 2.1471415e-18], [2.4434306, 2.11576e-17, 1.05788e-17, 6.34728e-18, 2.115761e-18], [2.4490632, 2.0849e-17, 1.04245e-17, 6.25471e-18, 2.0849034e-18], [2.4547089, 2.05456e-17, 1.02728e-17, 6.16367e-18, 2.0545586e-18], [2.4603676, 2.02471e-17, 1.01236e-17, 6.07415e-18, 2.0247169e-18], [2.4660393, 1.99537e-17, 9.97684e-18, 5.9861e-18, 1.9953685e-18], [2.4717241, 1.9665e-17, 9.83252e-18, 5.89951e-18, 1.966504e-18], [2.4774221, 1.93811e-17, 9.69057e-18, 5.81434e-18, 1.9381141e-18], [2.4831331, 1.91019e-17, 9.55095e-18, 5.73057e-18, 1.91019e-18], [2.4888573, 1.88272e-17, 9.41361e-18, 5.64817e-18, 1.8827227e-18], [2.4945947, 1.8557e-17, 9.27851e-18, 5.56711e-18, 1.8557037e-18], [2.5003454, 1.82912e-17, 9.14562e-18, 5.48737e-18, 1.8291244e-18], [2.5061093, 1.80297e-17, 9.01488e-18, 5.40893e-18, 1.8029768e-18], [2.5118864, 1.77725e-17, 8.88626e-18, 5.33176e-18, 1.7772526e-18], [2.5176769, 1.75194e-17, 8.75972e-18, 5.25583e-18, 1.751944e-18], [2.5234808, 1.72704e-17, 8.63521e-18, 5.18113e-18, 1.7270433e-18], [2.529298, 1.70254e-17, 8.51271e-18, 5.10763e-18, 1.702543e-18], [2.5351286, 1.67843e-17, 8.39217e-18, 5.0353e-18, 1.6784355e-18], [2.5409727, 1.65471e-17, 8.27356e-18, 4.96414e-18, 1.6547138e-18], [2.5468303, 1.63137e-17, 8.15685e-18, 4.89411e-18, 1.6313707e-18], [2.5527013, 1.6084e-17, 8.04199e-18, 4.8252e-18, 1.6083993e-18], [2.5585859, 1.58579e-17, 7.92896e-18, 4.75738e-18, 1.5857928e-18], [2.564484, 1.56354e-17, 7.81772e-18, 4.69063e-18, 1.5635445e-18], [2.5703958, 1.54165e-17, 7.70824e-18, 4.62494e-18, 1.5416481e-18], [2.5763212, 1.52009e-17, 7.60048e-18, 4.56029e-18, 1.520097e-18], [2.5822602, 1.49888e-17, 7.49442e-18, 4.49665e-18, 1.4988851e-18], [2.5882129, 1.478e-17, 7.39003e-18, 4.43402e-18, 1.4780063e-18], [2.5941794, 1.45745e-17, 7.28727e-18, 4.37236e-18, 1.4574546e-18], [2.6001596, 1.43722e-17, 7.18612e-18, 4.31167e-18, 1.4372241e-18], [2.6061535, 1.41731e-17, 7.08654e-18, 4.25192e-18, 1.4173092e-18], [2.6121614, 1.3977e-17, 6.98852e-18, 4.19311e-18, 1.3977041e-18], [2.618183, 1.3784e-17, 6.89201e-18, 4.13521e-18, 1.3784035e-18], [2.6242185, 1.3594e-17, 6.797e-18, 4.0782e-18, 1.3594018e-18], [2.630268, 1.34069e-17, 6.70347e-18, 4.02208e-18, 1.3406939e-18], [2.6363314, 1.32227e-17, 6.61137e-18, 3.96682e-18, 1.3222746e-18], [2.6424088, 1.30414e-17, 6.52069e-18, 3.91241e-18, 1.3041388e-18], [2.6485001, 1.28628e-17, 6.4314e-18, 3.85884e-18, 1.2862815e-18], [2.6546056, 1.2687e-17, 6.34348e-18, 3.80609e-18, 1.2686978e-18], [2.6607251, 1.25138e-17, 6.25691e-18, 3.75415e-18, 1.2513831e-18], [2.6668587, 1.23433e-17, 6.17166e-18, 3.703e-18, 1.2343326e-18], [2.6730064, 1.21754e-17, 6.0877e-18, 3.65262e-18, 1.2175418e-18], [2.6791683, 1.201e-17, 6.00503e-18, 3.60302e-18, 1.2010061e-18], [2.6853444, 1.18472e-17, 5.9236e-18, 3.55416e-18, 1.1847212e-18], [2.6915348, 1.16868e-17, 5.84341e-18, 3.50605e-18, 1.1686828e-18], [2.6977394, 1.15288e-17, 5.76443e-18, 3.45866e-18, 1.1528865e-18], [2.7039584, 1.13733e-17, 5.68664e-18, 3.41198e-18, 1.1373284e-18], [2.7101916, 1.122e-17, 5.61002e-18, 3.36601e-18, 1.1220042e-18], [2.7164393, 1.10691e-17, 5.53455e-18, 3.32073e-18, 1.1069101e-18], [2.7227013, 1.09204e-17, 5.46021e-18, 3.27612e-18, 1.0920421e-18], [2.7289778, 1.07739e-17, 5.38698e-18, 3.23219e-18, 1.0773964e-18], [2.7352687, 1.06297e-17, 5.31484e-18, 3.1889e-18, 1.0629692e-18], [2.7415742, 1.04875e-17, 5.24378e-18, 3.14627e-18, 1.0487568e-18], [2.7478942, 1.03475e-17, 5.17377e-18, 3.10426e-18, 1.0347557e-18], [2.7542287, 1.02096e-17, 5.10481e-18, 3.06288e-18, 1.0209622e-18], [2.7605779, 1.00737e-17, 5.03686e-18, 3.02212e-18, 1.007373e-18], [2.7669416, 9.93982e-18, 4.96992e-18, 2.98195e-18, 9.9398448e-19], [2.7733201, 9.80791e-18, 4.90396e-18, 2.94238e-18, 9.8079344e-19], [2.7797133, 9.67794e-18, 4.83898e-18, 2.90339e-18, 9.6779655e-19], [2.7861212, 9.54988e-18, 4.77495e-18, 2.86497e-18, 9.5499059e-19], [2.7925438, 9.4237e-18, 4.71186e-18, 2.82711e-18, 9.423724e-19], [2.7989813, 9.29936e-18, 4.64969e-18, 2.78981e-18, 9.2993888e-19], [2.8054336, 9.17685e-18, 4.58843e-18, 2.75306e-18, 9.1768697e-19], [2.8119008, 9.05611e-18, 4.52806e-18, 2.71684e-18, 9.0561371e-19], [2.8183829, 8.93714e-18, 4.46858e-18, 2.68115e-18, 8.9371615e-19], [2.82488, 8.81989e-18, 4.40995e-18, 2.64597e-18, 8.8199141e-19], [2.831392, 8.70434e-18, 4.35218e-18, 2.61131e-18, 8.7043668e-19], [2.837919, 8.59047e-18, 4.29524e-18, 2.57714e-18, 8.5904919e-19], [2.8444611, 8.47824e-18, 4.23913e-18, 2.54348e-18, 8.478262e-19], [2.8510183, 8.36763e-18, 4.18382e-18, 2.51029e-18, 8.3676506e-19], [2.8575905, 8.25861e-18, 4.12931e-18, 2.47759e-18, 8.2586313e-19], [2.864178, 8.15115e-18, 4.07558e-18, 2.44535e-18, 8.1511786e-19], [2.8707806, 8.04524e-18, 4.02263e-18, 2.41358e-18, 8.0452671e-19], [2.8773984, 7.94085e-18, 3.97043e-18, 2.38226e-18, 7.940872e-19], [2.8840315, 7.83794e-18, 3.91898e-18, 2.35139e-18, 7.8379689e-19], [2.8906799, 7.73651e-18, 3.86826e-18, 2.32096e-18, 7.736534e-19], [2.8973436, 7.63652e-18, 3.81827e-18, 2.29096e-18, 7.6365438e-19], [2.9040227, 7.53795e-18, 3.76898e-18, 2.26139e-18, 7.5379753e-19], [2.9107171, 7.44078e-18, 3.7204e-18, 2.23224e-18, 7.4408057e-19], [2.917427, 7.34499e-18, 3.6725e-18, 2.2035e-18, 7.3450129e-19], [2.9241524, 7.25055e-18, 3.62528e-18, 2.17517e-18, 7.250575e-19], [2.9308932, 7.15744e-18, 3.57873e-18, 2.14724e-18, 7.1574707e-19], [2.9376497, 7.06565e-18, 3.53283e-18, 2.1197e-18, 7.0656788e-19], [2.9444216, 6.97515e-18, 3.48758e-18, 2.09255e-18, 6.9751786e-19], [2.9512092, 6.88592e-18, 3.44297e-18, 2.06578e-18, 6.8859499e-19], [2.9580125, 6.79795e-18, 3.39898e-18, 2.03939e-18, 6.7979727e-19], [2.9648314, 6.7112e-18, 3.35561e-18, 2.01336e-18, 6.7112273e-19], [2.971666, 6.62567e-18, 3.31284e-18, 1.98771e-18, 6.6256946e-19], [2.9785164, 6.54133e-18, 3.27067e-18, 1.9624e-18, 6.5413556e-19], [2.9853826, 6.45816e-18, 3.22909e-18, 1.93745e-18, 6.4581917e-19], [2.9922646, 6.37616e-18, 3.18809e-18, 1.91285e-18, 6.3761847e-19], [2.9991625, 6.29529e-18, 3.14765e-18, 1.88859e-18, 6.2953166e-19], [3.0060763, 6.21554e-18, 3.10778e-18, 1.86467e-18, 6.2155698e-19], [3.013006, 6.1369e-18, 3.06846e-18, 1.84107e-18, 6.136927e-19], [3.0199517, 6.05934e-18, 3.02968e-18, 1.81781e-18, 6.0593712e-19], [3.0269134, 5.98286e-18, 2.99144e-18, 1.79486e-18, 5.9828855e-19], [3.0338912, 5.90743e-18, 2.95372e-18, 1.77223e-18, 5.9074537e-19], [3.040885, 5.83303e-18, 2.91652e-18, 1.74991e-18, 5.8330595e-19], [3.047895, 5.75966e-18, 2.87984e-18, 1.7279e-18, 5.7596871e-19], [3.0549211, 5.68729e-18, 2.84365e-18, 1.70619e-18, 5.6873208e-19], [3.0619634, 5.61592e-18, 2.80797e-18, 1.68478e-18, 5.6159454e-19], [3.069022, 5.54552e-18, 2.77277e-18, 1.66366e-18, 5.5455457e-19], [3.0760968, 5.47608e-18, 2.73805e-18, 1.64283e-18, 5.476107e-19], [3.083188, 5.40759e-18, 2.7038e-18, 1.62228e-18, 5.4076147e-19], [3.0902954, 5.34003e-18, 2.67002e-18, 1.60201e-18, 5.3400544e-19], [3.0974193, 5.27338e-18, 2.6367e-18, 1.58202e-18, 5.2734121e-19], [3.1045596, 5.20765e-18, 2.60383e-18, 1.5623e-18, 5.207674e-19], [3.1117163, 5.1428e-18, 2.57141e-18, 1.54284e-18, 5.1428264e-19], [3.1188896, 5.07883e-18, 2.53942e-18, 1.52365e-18, 5.0788561e-19], [3.1260794, 5.01572e-18, 2.50787e-18, 1.50472e-18, 5.0157497e-19], [3.1332857, 4.95347e-18, 2.47674e-18, 1.48604e-18, 4.9534945e-19], [3.1405087, 4.89205e-18, 2.44603e-18, 1.46762e-18, 4.8920777e-19], [3.1477483, 4.83146e-18, 2.41574e-18, 1.44944e-18, 4.8314869e-19], [3.1550046, 4.77168e-18, 2.38585e-18, 1.43151e-18, 4.7717096e-19], [3.1622777, 4.7127e-18, 2.35636e-18, 1.41382e-18, 4.7127339e-19], [3.1695675, 4.65452e-18, 2.32727e-18, 1.39636e-18, 4.6545478e-19], [3.1768741, 4.59711e-18, 2.29856e-18, 1.37914e-18, 4.5971398e-19], [3.1841975, 4.54047e-18, 2.27024e-18, 1.36215e-18, 4.5404982e-19], [3.1915379, 4.48458e-18, 2.2423e-18, 1.34538e-18, 4.4846118e-19], [3.1988951, 4.42944e-18, 2.21473e-18, 1.32884e-18, 4.4294694e-19], [3.2062693, 4.37503e-18, 2.18752e-18, 1.31251e-18, 4.3750602e-19], [3.2136605, 4.32134e-18, 2.16068e-18, 1.29641e-18, 4.3213734e-19], [3.2210688, 4.26837e-18, 2.13419e-18, 1.28052e-18, 4.2683983e-19], [3.2284941, 4.21609e-18, 2.10806e-18, 1.26483e-18, 4.2161247e-19], [3.2359366, 4.16451e-18, 2.08226e-18, 1.24936e-18, 4.1645422e-19], [3.2433962, 4.11361e-18, 2.05681e-18, 1.23409e-18, 4.1136408e-19], [3.250873, 4.06338e-18, 2.0317e-18, 1.21902e-18, 4.0634106e-19], [3.258367, 4.01381e-18, 2.00691e-18, 1.20415e-18, 4.0138418e-19], [3.2658783, 3.96489e-18, 1.98246e-18, 1.18947e-18, 3.9649249e-19], [3.2734069, 3.91662e-18, 1.95832e-18, 1.17499e-18, 3.9166504e-19], [3.2809529, 3.86898e-18, 1.9345e-18, 1.1607e-18, 3.8690091e-19], [3.2885163, 3.82196e-18, 1.91099e-18, 1.14659e-18, 3.8219918e-19], [3.2960971, 3.77556e-18, 1.88779e-18, 1.13267e-18, 3.7755895e-19], [3.3036954, 3.72976e-18, 1.86489e-18, 1.11893e-18, 3.7297934e-19], [3.3113112, 3.68456e-18, 1.84229e-18, 1.10537e-18, 3.6845949e-19], [3.3189446, 3.63995e-18, 1.81999e-18, 1.09199e-18, 3.6399852e-19], [3.3265955, 3.59592e-18, 1.79797e-18, 1.07878e-18, 3.5959561e-19], [3.3342641, 3.55247e-18, 1.77624e-18, 1.06575e-18, 3.5524992e-19], [3.3419504, 3.50957e-18, 1.7548e-18, 1.05288e-18, 3.5096064e-19], [3.3496544, 3.46724e-18, 1.73363e-18, 1.04018e-18, 3.4672697e-19], [3.3573761, 3.42545e-18, 1.71273e-18, 1.02764e-18, 3.4254811e-19], [3.3651157, 3.3842e-18, 1.69211e-18, 1.01527e-18, 3.3842328e-19], [3.3728731, 3.34348e-18, 1.67175e-18, 1.00305e-18, 3.3435173e-19], [3.3806484, 3.30329e-18, 1.65166e-18, 9.90994e-19, 3.303327e-19], [3.3884416, 3.26362e-18, 1.63182e-18, 9.79092e-19, 3.2636544e-19], [3.3962527, 3.22446e-18, 1.61224e-18, 9.67344e-19, 3.2244923e-19], [3.4040819, 3.1858e-18, 1.59291e-18, 9.55746e-19, 3.1858336e-19], [3.4119291, 3.14764e-18, 1.57383e-18, 9.44297e-19, 3.147671e-19], [3.4197944, 3.10996e-18, 1.55499e-18, 9.32995e-19, 3.1099977e-19], [3.4276779, 3.07277e-18, 1.5364e-18, 9.21838e-19, 3.0728068e-19], [3.4355795, 3.03606e-18, 1.51804e-18, 9.10823e-19, 3.0360916e-19], [3.4434993, 2.99981e-18, 1.49992e-18, 8.9995e-19, 2.9998454e-19], [3.4514374, 2.96403e-18, 1.48202e-18, 8.89214e-19, 2.9640617e-19], [3.4593938, 2.9287e-18, 1.46436e-18, 8.78616e-19, 2.9287341e-19], [3.4673685, 2.89382e-18, 1.44692e-18, 8.68153e-19, 2.8938561e-19], [3.4753616, 2.85939e-18, 1.4297e-18, 8.57822e-19, 2.8594216e-19], [3.4833732, 2.82539e-18, 1.41271e-18, 8.47623e-19, 2.8254245e-19], [3.4914032, 2.79182e-18, 1.39592e-18, 8.37553e-19, 2.7918586e-19], [3.4994517, 2.75868e-18, 1.37935e-18, 8.27611e-19, 2.758718e-19], [3.5075187, 2.72596e-18, 1.36299e-18, 8.17795e-19, 2.7259969e-19], [3.5156044, 2.69366e-18, 1.34684e-18, 8.08103e-19, 2.6936895e-19], [3.5237087, 2.66176e-18, 1.33089e-18, 7.98533e-19, 2.66179e-19], [3.5318317, 2.63026e-18, 1.31514e-18, 7.89084e-19, 2.6302929e-19], [3.5399734, 2.59916e-18, 1.29959e-18, 7.79754e-19, 2.5991927e-19], [3.5481339, 2.56845e-18, 1.28423e-18, 7.70541e-19, 2.5684839e-19], [3.5563132, 2.53813e-18, 1.26907e-18, 7.61444e-19, 2.5381612e-19], [3.5645113, 2.50818e-18, 1.2541e-18, 7.52461e-19, 2.5082193e-19], [3.5727284, 2.47862e-18, 1.23932e-18, 7.43592e-19, 2.478653e-19], [3.5809644, 2.44942e-18, 1.22472e-18, 7.34833e-19, 2.4494572e-19], [3.5892193, 2.42059e-18, 1.21031e-18, 7.26184e-19, 2.4206269e-19], [3.5974934, 2.39212e-18, 1.19607e-18, 7.17643e-19, 2.3921571e-19], [3.6057864, 2.36401e-18, 1.18201e-18, 7.09209e-19, 2.3640429e-19], [3.6140986, 2.33624e-18, 1.16813e-18, 7.00879e-19, 2.3362795e-19], [3.62243, 2.30883e-18, 1.15442e-18, 6.92654e-19, 2.3088621e-19], [3.6307805, 2.28175e-18, 1.14089e-18, 6.84531e-19, 2.2817862e-19], [3.6391504, 2.25501e-18, 1.12752e-18, 6.7651e-19, 2.255047e-19], [3.6475395, 2.2286e-18, 1.11431e-18, 6.68588e-19, 2.2286401e-19], [3.6559479, 2.20253e-18, 1.10127e-18, 6.60764e-19, 2.202561e-19], [3.6643757, 2.17677e-18, 1.0884e-18, 6.53037e-19, 2.1768053e-19], [3.672823, 2.15133e-18, 1.07568e-18, 6.45406e-19, 2.1513687e-19], [3.6812897, 2.12621e-18, 1.06312e-18, 6.3787e-19, 2.1262468e-19], [3.689776, 2.1014e-18, 1.05071e-18, 6.30426e-19, 2.1014355e-19], [3.6982818, 2.07689e-18, 1.03846e-18, 6.23075e-19, 2.0769306e-19], [3.7068072, 2.05269e-18, 1.02636e-18, 6.15814e-19, 2.0527281e-19], [3.7153523, 2.02879e-18, 1.0144e-18, 6.08643e-19, 2.0288239e-19], [3.7239171, 2.00518e-18, 1.0026e-18, 6.0156e-19, 2.005214e-19], [3.7325016, 1.98186e-18, 9.9094e-19, 5.94564e-19, 1.9818946e-19], [3.7411059, 1.95883e-18, 9.79423e-19, 5.87654e-19, 1.9588618e-19], [3.74973, 1.93608e-18, 9.68048e-19, 5.80829e-19, 1.9361118e-19], [3.758374, 1.9136e-18, 9.56813e-19, 5.74088e-19, 1.9136408e-19], [3.767038, 1.89141e-18, 9.45715e-19, 5.67429e-19, 1.8914452e-19], [3.7757219, 1.86948e-18, 9.34753e-19, 5.60852e-19, 1.8695214e-19], [3.7844258, 1.84783e-18, 9.23925e-19, 5.54355e-19, 1.8478657e-19], [3.7931498, 1.82644e-18, 9.1323e-19, 5.47938e-19, 1.8264746e-19], [3.801894, 1.80531e-18, 9.02665e-19, 5.41599e-19, 1.8053446e-19], [3.8106582, 1.78444e-18, 8.92229e-19, 5.35337e-19, 1.7844724e-19], [3.8194427, 1.76382e-18, 8.8192e-19, 5.29152e-19, 1.7638546e-19], [3.8282474, 1.74345e-18, 8.71736e-19, 5.23042e-19, 1.7434877e-19], [3.8370725, 1.72333e-18, 8.61677e-19, 5.17006e-19, 1.7233686e-19], [3.8459178, 1.70346e-18, 8.51739e-19, 5.11044e-19, 1.7034939e-19], [3.8547836, 1.68382e-18, 8.41923e-19, 5.05154e-19, 1.6838606e-19], [3.8636698, 1.66443e-18, 8.32225e-19, 4.99335e-19, 1.6644655e-19], [3.8725764, 1.64527e-18, 8.22645e-19, 4.93587e-19, 1.6453054e-19], [3.8815037, 1.62634e-18, 8.13181e-19, 4.87908e-19, 1.6263773e-19], [3.8904514, 1.60764e-18, 8.03831e-19, 4.82299e-19, 1.6076782e-19], [3.8994199, 1.58917e-18, 7.94595e-19, 4.76757e-19, 1.5892051e-19], [3.908409, 1.57092e-18, 7.8547e-19, 4.71282e-19, 1.5709551e-19], [3.9174188, 1.55289e-18, 7.76455e-19, 4.65873e-19, 1.5529254e-19], [3.9264494, 1.53507e-18, 7.67549e-19, 4.60529e-19, 1.535113e-19], [3.9355008, 1.51748e-18, 7.5875e-19, 4.5525e-19, 1.5175152e-19], [3.944573, 1.50009e-18, 7.50057e-19, 4.50034e-19, 1.5001292e-19], [3.9536662, 1.48291e-18, 7.41468e-19, 4.44881e-19, 1.4829523e-19], [3.9627803, 1.46594e-18, 7.32983e-19, 4.3979e-19, 1.4659817e-19], [3.9719155, 1.44918e-18, 7.24599e-19, 4.3476e-19, 1.4492149e-19], [3.9810717, 1.43261e-18, 7.16317e-19, 4.2979e-19, 1.4326492e-19], [3.990249, 1.41624e-18, 7.08133e-19, 4.2488e-19, 1.4162821e-19], [3.9994475, 1.40007e-18, 7.00047e-19, 4.20028e-19, 1.400111e-19], [4.0086672, 1.38409e-18, 6.92059e-19, 4.15235e-19, 1.3841333e-19], [4.0179081, 1.36831e-18, 6.84165e-19, 4.10499e-19, 1.3683467e-19], [4.0271703, 1.35271e-18, 6.76366e-19, 4.0582e-19, 1.3527487e-19], [4.0364539, 1.3373e-18, 6.6866e-19, 4.01196e-19, 1.3373369e-19], [4.0457589, 1.32207e-18, 6.61046e-19, 3.96628e-19, 1.3221089e-19], [4.0550854, 1.30702e-18, 6.53523e-19, 3.92114e-19, 1.3070624e-19], [4.0644333, 1.29216e-18, 6.46089e-19, 3.87654e-19, 1.2921951e-19], [4.0738028, 1.27746e-18, 6.38744e-19, 3.83246e-19, 1.2775047e-19], [4.0831939, 1.26295e-18, 6.31486e-19, 3.78892e-19, 1.2629889e-19], [4.0926066, 1.24861e-18, 6.24315e-19, 3.74589e-19, 1.2486456e-19], [4.102041, 1.23443e-18, 6.17228e-19, 3.70337e-19, 1.2344726e-19], [4.1114972, 1.22043e-18, 6.10226e-19, 3.66135e-19, 1.2204677e-19], [4.1209752, 1.20659e-18, 6.03306e-19, 3.61984e-19, 1.2066288e-19], [4.130475, 1.19291e-18, 5.96469e-19, 3.57881e-19, 1.1929538e-19], [4.1399967, 1.1794e-18, 5.89712e-19, 3.53827e-19, 1.1794406e-19], [4.1495404, 1.16605e-18, 5.83035e-19, 3.49821e-19, 1.1660873e-19], [4.1591061, 1.15285e-18, 5.76438e-19, 3.45863e-19, 1.1528918e-19], [4.1686938, 1.13981e-18, 5.69918e-19, 3.41951e-19, 1.1398521e-19], [4.1783037, 1.12693e-18, 5.63475e-19, 3.38085e-19, 1.1269663e-19], [4.1879357, 1.11419e-18, 5.57108e-19, 3.34265e-19, 1.1142324e-19], [4.1975898, 1.10161e-18, 5.50816e-19, 3.3049e-19, 1.1016486e-19], [4.2072663, 1.08917e-18, 5.44598e-19, 3.26759e-19, 1.0892129e-19], [4.216965, 1.07688e-18, 5.38453e-19, 3.23072e-19, 1.0769236e-19], [4.2266861, 1.06474e-18, 5.32381e-19, 3.19429e-19, 1.0647787e-19], [4.2364297, 1.05274e-18, 5.2638e-19, 3.15828e-19, 1.0527766e-19], [4.2461956, 1.04087e-18, 5.20449e-19, 3.12269e-19, 1.0409154e-19], [4.2559841, 1.02915e-18, 5.14588e-19, 3.08753e-19, 1.0291933e-19], [4.2657952, 1.01757e-18, 5.08796e-19, 3.05277e-19, 1.0176087e-19], [4.2756289, 1.00612e-18, 5.03071e-19, 3.01843e-19, 1.0061598e-19], [4.2854852, 9.94803e-19, 4.97414e-19, 2.98448e-19, 9.9484493e-20], [4.2953643, 9.83621e-19, 4.91823e-19, 2.95094e-19, 9.8366248e-20], [4.3052661, 9.72569e-19, 4.86297e-19, 2.91778e-19, 9.7261078e-20], [4.3151908, 9.61646e-19, 4.80835e-19, 2.88501e-19, 9.6168821e-20], [4.3251383, 9.50851e-19, 4.75438e-19, 2.85263e-19, 9.5089318e-20], [4.3351088, 9.40182e-19, 4.70103e-19, 2.82062e-19, 9.4022412e-20], [4.3451022, 9.29637e-19, 4.64831e-19, 2.78899e-19, 9.2967947e-20], [4.3551187, 9.19215e-19, 4.5962e-19, 2.75772e-19, 9.1925771e-20], [4.3651583, 9.08915e-19, 4.5447e-19, 2.72682e-19, 9.0895731e-20], [4.3752211, 8.98734e-19, 4.4938e-19, 2.69628e-19, 8.987768e-20], [4.385307, 8.88672e-19, 4.44349e-19, 2.66609e-19, 8.8871469e-20], [4.3954162, 8.78727e-19, 4.39376e-19, 2.63626e-19, 8.7876953e-20], [4.4055486, 8.68897e-19, 4.34461e-19, 2.60677e-19, 8.689399e-20], [4.4157045, 8.59182e-19, 4.29603e-19, 2.57762e-19, 8.5922437e-20], [4.4258837, 8.49579e-19, 4.24802e-19, 2.54881e-19, 8.4962155e-20], [4.4360864, 8.40087e-19, 4.20056e-19, 2.52034e-19, 8.4013006e-20], [4.4463127, 8.30706e-19, 4.15365e-19, 2.49219e-19, 8.3074854e-20], [4.4565625, 8.21433e-19, 4.10729e-19, 2.46437e-19, 8.2147566e-20], [4.4668359, 8.12267e-19, 4.06146e-19, 2.43688e-19, 8.1231008e-20], [4.477133, 8.03207e-19, 4.01616e-19, 2.4097e-19, 8.032505e-20], [4.4874539, 7.94252e-19, 3.97139e-19, 2.38283e-19, 7.9429563e-20], [4.4977985, 7.85401e-19, 3.92713e-19, 2.35628e-19, 7.8544419e-20], [4.508167, 7.76651e-19, 3.88338e-19, 2.33003e-19, 7.7669493e-20], [4.5185594, 7.68003e-19, 3.84014e-19, 2.30409e-19, 7.6804662e-20], [4.5289758, 7.59454e-19, 3.7974e-19, 2.27844e-19, 7.5949801e-20], [4.5394162, 7.51004e-19, 3.75515e-19, 2.25309e-19, 7.5104792e-20], [4.5498806, 7.42651e-19, 3.71339e-19, 2.22803e-19, 7.4269514e-20], [4.5603692, 7.34395e-19, 3.6721e-19, 2.20326e-19, 7.3443849e-20], [4.5708819, 7.26233e-19, 3.63129e-19, 2.17878e-19, 7.2627683e-20], [4.5814189, 7.18165e-19, 3.59095e-19, 2.15457e-19, 7.1820898e-20], [4.5919801, 7.1019e-19, 3.55108e-19, 2.13065e-19, 7.1023384e-20], [4.6025657, 7.02306e-19, 3.51166e-19, 2.107e-19, 7.0235027e-20], [4.6131757, 6.94513e-19, 3.47269e-19, 2.08362e-19, 6.9455718e-20], [4.6238102, 6.86809e-19, 3.43418e-19, 2.06051e-19, 6.8685348e-20], [4.6344692, 6.79194e-19, 3.3961e-19, 2.03766e-19, 6.7923809e-20], [4.6451528, 6.71665e-19, 3.35846e-19, 2.01507e-19, 6.7170995e-20], [4.6558609, 6.64223e-19, 3.32125e-19, 1.99275e-19, 6.6426801e-20], [4.6665938, 6.56866e-19, 3.28446e-19, 1.97068e-19, 6.5691124e-20], [4.6773514, 6.49594e-19, 3.2481e-19, 1.94886e-19, 6.4963861e-20], [4.6881338, 6.42404e-19, 3.21215e-19, 1.92729e-19, 6.4244913e-20], [4.6989411, 6.35297e-19, 3.17662e-19, 1.90597e-19, 6.3534179e-20], [4.7097733, 6.2827e-19, 3.14148e-19, 1.88489e-19, 6.2831562e-20], [4.7206304, 6.21324e-19, 3.10675e-19, 1.86405e-19, 6.2136963e-20], [4.7315126, 6.14458e-19, 3.07242e-19, 1.84345e-19, 6.1450288e-20], [4.7424199, 6.07669e-19, 3.03848e-19, 1.82309e-19, 6.0771442e-20], [4.7533523, 6.00958e-19, 3.00492e-19, 1.80295e-19, 6.0100332e-20], [4.7643099, 5.94323e-19, 2.97175e-19, 1.78305e-19, 5.9436865e-20], [4.7752927, 5.87764e-19, 2.93895e-19, 1.76337e-19, 5.8780951e-20], [4.7863009, 5.81279e-19, 2.90653e-19, 1.74392e-19, 5.8132499e-20], [4.7973345, 5.74868e-19, 2.87448e-19, 1.72469e-19, 5.7491421e-20], [4.8083935, 5.6853e-19, 2.84279e-19, 1.70567e-19, 5.6857629e-20], [4.819478, 5.62264e-19, 2.81146e-19, 1.68687e-19, 5.6231038e-20], [4.830588, 5.5607e-19, 2.78048e-19, 1.66829e-19, 5.561156e-20], [4.8417237, 5.49945e-19, 2.74986e-19, 1.64992e-19, 5.4999113e-20], [4.852885, 5.4389e-19, 2.71959e-19, 1.63175e-19, 5.4393613e-20], [4.8640721, 5.37903e-19, 2.68965e-19, 1.61379e-19, 5.3794978e-20], [4.8752849, 5.31985e-19, 2.66006e-19, 1.59604e-19, 5.3203126e-20], [4.8865236, 5.26133e-19, 2.6308e-19, 1.57848e-19, 5.2617977e-20], [4.8977882, 5.20348e-19, 2.60188e-19, 1.56113e-19, 5.2039453e-20], [4.9090788, 5.14628e-19, 2.57328e-19, 1.54397e-19, 5.1467474e-20], [4.9203954, 5.08973e-19, 2.545e-19, 1.527e-19, 5.0901965e-20], [4.931738, 5.03382e-19, 2.51705e-19, 1.51023e-19, 5.0342848e-20], [4.9431069, 4.97854e-19, 2.48941e-19, 1.49364e-19, 4.9790048e-20], [4.9545019, 4.92388e-19, 2.46208e-19, 1.47725e-19, 4.9243492e-20], [4.9659232, 4.86984e-19, 2.43506e-19, 1.46103e-19, 4.8703105e-20], [4.9773708, 4.81641e-19, 2.40834e-19, 1.44501e-19, 4.8168815e-20], [4.9888449, 4.76358e-19, 2.38193e-19, 1.42916e-19, 4.7640551e-20], [5.0003453, 4.71135e-19, 2.35581e-19, 1.41349e-19, 4.7118241e-20], [5.0118723, 4.65971e-19, 2.32999e-19, 1.398e-19, 4.6601817e-20], [5.0234259, 4.60865e-19, 2.30446e-19, 1.38268e-19, 4.6091208e-20], [5.0350061, 4.55816e-19, 2.27922e-19, 1.36753e-19, 4.5586347e-20], [5.046613, 4.50824e-19, 2.25426e-19, 1.35256e-19, 4.5087167e-20], [5.0582466, 4.45888e-19, 2.22958e-19, 1.33775e-19, 4.4593601e-20], [5.0699071, 4.41008e-19, 2.20518e-19, 1.32311e-19, 4.4105584e-20], [5.0815944, 4.36183e-19, 2.18105e-19, 1.30863e-19, 4.3623051e-20], [5.0933087, 4.31411e-19, 2.1572e-19, 1.29432e-19, 4.3145937e-20], [5.10505, 4.26694e-19, 2.13361e-19, 1.28017e-19, 4.2674179e-20], [5.1168184, 4.22029e-19, 2.11029e-19, 1.26617e-19, 4.2207716e-20], [5.1286138, 4.17417e-19, 2.08722e-19, 1.25233e-19, 4.1746485e-20], [5.1404365, 4.12856e-19, 2.06442e-19, 1.23865e-19, 4.1290426e-20], [5.1522864, 4.08346e-19, 2.04187e-19, 1.22512e-19, 4.0839478e-20], [5.1641637, 4.03887e-19, 2.01958e-19, 1.21175e-19, 4.0393582e-20], [5.1760683, 3.99478e-19, 1.99753e-19, 1.19852e-19, 3.9952678e-20], [5.1880004, 3.95118e-19, 1.97574e-19, 1.18544e-19, 3.951671e-20], [5.19996, 3.90807e-19, 1.95418e-19, 1.17251e-19, 3.908562e-20], [5.2119471, 3.86545e-19, 1.93287e-19, 1.15972e-19, 3.8659352e-20], [5.2239619, 3.8233e-19, 1.91179e-19, 1.14707e-19, 3.8237848e-20], [5.2360044, 3.78162e-19, 1.89095e-19, 1.13457e-19, 3.7821055e-20], [5.2480746, 3.7404e-19, 1.87034e-19, 1.12221e-19, 3.7408917e-20], [5.2601727, 3.69965e-19, 1.84997e-19, 1.10998e-19, 3.7001382e-20], [5.2722986, 3.65935e-19, 1.82982e-19, 1.09789e-19, 3.6598394e-20], [5.2844525, 3.6195e-19, 1.80989e-19, 1.08594e-19, 3.6199903e-20], [5.2966344, 3.58009e-19, 1.79019e-19, 1.07411e-19, 3.5805856e-20], [5.3088444, 3.54113e-19, 1.77071e-19, 1.06242e-19, 3.5416202e-20], [5.3210826, 3.50259e-19, 1.75144e-19, 1.05087e-19, 3.5030889e-20], [5.333349, 3.46449e-19, 1.73239e-19, 1.03943e-19, 3.4649869e-20], [5.3456436, 3.42681e-19, 1.71355e-19, 1.02813e-19, 3.4273091e-20], [5.3579666, 3.38955e-19, 1.69492e-19, 1.01695e-19, 3.3900507e-20], [5.370318, 3.35271e-19, 1.6765e-19, 1.0059e-19, 3.3532068e-20], [5.3826978, 3.31627e-19, 1.65828e-19, 9.9497e-20, 3.3167727e-20], [5.3951062, 3.28024e-19, 1.64027e-19, 9.84161e-20, 3.2807437e-20], [5.4075432, 3.24461e-19, 1.62245e-19, 9.73473e-20, 3.245115e-20], [5.4200089, 3.20938e-19, 1.60484e-19, 9.62903e-20, 3.2098822e-20], [5.4325033, 3.17454e-19, 1.58742e-19, 9.5245e-20, 3.1750406e-20], [5.4450265, 3.14008e-19, 1.57019e-19, 9.42113e-20, 3.1405858e-20], [5.4575786, 3.10601e-19, 1.55315e-19, 9.31892e-20, 3.1065134e-20], [5.4701596, 3.07231e-19, 1.53631e-19, 9.21783e-20, 3.0728189e-20], [5.4827696, 3.03899e-19, 1.51964e-19, 9.11787e-20, 3.039498e-20], [5.4954087, 3.00604e-19, 1.50317e-19, 9.01901e-20, 3.0065465e-20], [5.508077, 2.97345e-19, 1.48688e-19, 8.92125e-20, 2.9739601e-20], [5.5207744, 2.94123e-19, 1.47076e-19, 8.82458e-20, 2.9417346e-20], [5.5335011, 2.90936e-19, 1.45483e-19, 8.72897e-20, 2.909866e-20], [5.5462571, 2.87784e-19, 1.43907e-19, 8.63442e-20, 2.8783501e-20], [5.5590426, 2.84667e-19, 1.42349e-19, 8.54092e-20, 2.8471829e-20], [5.5718575, 2.81585e-19, 1.40807e-19, 8.44845e-20, 2.8163605e-20], [5.5847019, 2.78537e-19, 1.39283e-19, 8.357e-20, 2.7858789e-20], [5.597576, 2.75522e-19, 1.37776e-19, 8.26657e-20, 2.7557342e-20], [5.6104798, 2.72541e-19, 1.36286e-19, 8.17713e-20, 2.7259226e-20], [5.6234133, 2.69592e-19, 1.34811e-19, 8.08868e-20, 2.6964402e-20], [5.6363766, 2.66677e-19, 1.33354e-19, 8.00121e-20, 2.6672835e-20], [5.6493697, 2.63793e-19, 1.31912e-19, 7.91471e-20, 2.6384485e-20], [5.6623929, 2.60941e-19, 1.30486e-19, 7.82916e-20, 2.6099318e-20], [5.6754461, 2.58121e-19, 1.29076e-19, 7.74455e-20, 2.5817296e-20], [5.6885293, 2.55332e-19, 1.27681e-19, 7.66087e-20, 2.5538384e-20], [5.7016427, 2.52573e-19, 1.26302e-19, 7.57812e-20, 2.5262546e-20], [5.7147864, 2.49845e-19, 1.24938e-19, 7.49628e-20, 2.4989748e-20], [5.7279603, 2.47147e-19, 1.23589e-19, 7.41534e-20, 2.4719956e-20], [5.7411646, 2.44479e-19, 1.22255e-19, 7.33529e-20, 2.4453134e-20], [5.7543994, 2.4184e-19, 1.20935e-19, 7.25613e-20, 2.4189251e-20], [5.7676646, 2.3923e-19, 1.19631e-19, 7.17783e-20, 2.3928271e-20], [5.7809605, 2.36649e-19, 1.1834e-19, 7.1004e-20, 2.3670162e-20], [5.794287, 2.34096e-19, 1.17064e-19, 7.02382e-20, 2.3414893e-20], [5.8076442, 2.31572e-19, 1.15801e-19, 6.94808e-20, 2.316243e-20], [5.8210322, 2.29075e-19, 1.14553e-19, 6.87317e-20, 2.2912742e-20], [5.834451, 2.26605e-19, 1.13318e-19, 6.79909e-20, 2.2665798e-20], [5.8479008, 2.24163e-19, 1.12097e-19, 6.72582e-20, 2.2421566e-20], [5.8613816, 2.21747e-19, 1.10889e-19, 6.65335e-20, 2.2180017e-20], [5.8748935, 2.19358e-19, 1.09695e-19, 6.58168e-20, 2.1941119e-20], [5.8884366, 2.16995e-19, 1.08513e-19, 6.5108e-20, 2.1704843e-20], [5.9020108, 2.14658e-19, 1.07345e-19, 6.44069e-20, 2.1471159e-20], [5.9156163, 2.12347e-19, 1.06189e-19, 6.37135e-20, 2.1240039e-20], [5.9292532, 2.10061e-19, 1.05046e-19, 6.30278e-20, 2.1011452e-20], [5.9429216, 2.078e-19, 1.03916e-19, 6.23495e-20, 2.0785372e-20], [5.9566214, 2.05564e-19, 1.02798e-19, 6.16787e-20, 2.0561768e-20], [5.9703529, 2.03352e-19, 1.01692e-19, 6.10152e-20, 2.0340615e-20], [5.984116, 2.01165e-19, 1.00598e-19, 6.0359e-20, 2.0121883e-20], [5.9979108, 1.99002e-19, 9.95167e-20, 5.971e-20, 1.9905546e-20], [6.0117374, 1.96862e-19, 9.84468e-20, 5.90681e-20, 1.9691578e-20], [6.0255959, 1.94745e-19, 9.73886e-20, 5.84332e-20, 1.947995e-20], [6.0394863, 1.92652e-19, 9.63421e-20, 5.78052e-20, 1.9270638e-20], [6.0534087, 1.90582e-19, 9.53069e-20, 5.71842e-20, 1.9063614e-20], [6.0673633, 1.88534e-19, 9.42831e-20, 5.65699e-20, 1.8858854e-20], [6.08135, 1.86509e-19, 9.32705e-20, 5.59623e-20, 1.8656332e-20], [6.095369, 1.84506e-19, 9.22689e-20, 5.53614e-20, 1.8456023e-20], [6.1094202, 1.82524e-19, 9.12783e-20, 5.4767e-20, 1.8257903e-20], [6.1235039, 1.80565e-19, 9.02985e-20, 5.41791e-20, 1.8061946e-20], [6.1376201, 1.78627e-19, 8.93294e-20, 5.35976e-20, 1.7868128e-20], [6.1517687, 1.76709e-19, 8.83709e-20, 5.30225e-20, 1.7676426e-20], [6.16595, 1.74813e-19, 8.74228e-20, 5.24537e-20, 1.7486816e-20], [6.180164, 1.72938e-19, 8.64851e-20, 5.18911e-20, 1.7299274e-20], [6.1944108, 1.71083e-19, 8.55576e-20, 5.13346e-20, 1.7113777e-20], [6.2086903, 1.69248e-19, 8.46402e-20, 5.07841e-20, 1.6930303e-20], [6.2230029, 1.67433e-19, 8.37328e-20, 5.02397e-20, 1.6748829e-20], [6.2373484, 1.65638e-19, 8.28353e-20, 4.97012e-20, 1.6569333e-20], [6.2517269, 1.63862e-19, 8.19476e-20, 4.91686e-20, 1.6391792e-20], [6.2661386, 1.62106e-19, 8.10695e-20, 4.86417e-20, 1.6216185e-20], [6.2805836, 1.60369e-19, 8.02011e-20, 4.81206e-20, 1.604249e-20], [6.2950618, 1.58651e-19, 7.9342e-20, 4.76052e-20, 1.5870686e-20], [6.3095734, 1.56952e-19, 7.84923e-20, 4.70954e-20, 1.5700752e-20], [6.3241185, 1.55271e-19, 7.76519e-20, 4.65911e-20, 1.5532667e-20], [6.3386971, 1.53608e-19, 7.68206e-20, 4.60924e-20, 1.536641e-20], [6.3533093, 1.51964e-19, 7.59983e-20, 4.5599e-20, 1.5201962e-20], [6.3679552, 1.50337e-19, 7.5185e-20, 4.5111e-20, 1.5039301e-20], [6.3826349, 1.48728e-19, 7.43805e-20, 4.46283e-20, 1.4878408e-20], [6.3973484, 1.47136e-19, 7.35848e-20, 4.41509e-20, 1.4719264e-20], [6.4120958, 1.45562e-19, 7.27977e-20, 4.36786e-20, 1.4561849e-20], [6.4268772, 1.44005e-19, 7.20192e-20, 4.32115e-20, 1.4406144e-20], [6.4416927, 1.42465e-19, 7.12491e-20, 4.27494e-20, 1.4252129e-20], [6.4565423, 1.40941e-19, 7.04873e-20, 4.22924e-20, 1.4099787e-20], [6.4714262, 1.39434e-19, 6.97339e-20, 4.18403e-20, 1.3949098e-20], [6.4863443, 1.37944e-19, 6.89886e-20, 4.13932e-20, 1.3800045e-20], [6.5012969, 1.36469e-19, 6.82514e-20, 4.09508e-20, 1.3652608e-20], [6.5162839, 1.35011e-19, 6.75222e-20, 4.05133e-20, 1.350677e-20], [6.5313055, 1.33568e-19, 6.68009e-20, 4.00805e-20, 1.3362514e-20], [6.5463617, 1.32141e-19, 6.60874e-20, 3.96525e-20, 1.3219822e-20], [6.5614527, 1.3073e-19, 6.53817e-20, 3.9229e-20, 1.3078677e-20], [6.5765784, 1.29333e-19, 6.46836e-20, 3.88102e-20, 1.2939061e-20], [6.591739, 1.27952e-19, 6.39931e-20, 3.83958e-20, 1.2800957e-20], [6.6069345, 1.26586e-19, 6.331e-20, 3.7986e-20, 1.266435e-20], [6.622165, 1.25235e-19, 6.26343e-20, 3.75806e-20, 1.2529222e-20], [6.6374307, 1.23898e-19, 6.1966e-20, 3.71796e-20, 1.2395556e-20], [6.6527316, 1.22576e-19, 6.13049e-20, 3.67829e-20, 1.2263338e-20], [6.6680677, 1.21268e-19, 6.06509e-20, 3.63906e-20, 1.213255e-20], [6.6834392, 1.19974e-19, 6.00041e-20, 3.60024e-20, 1.2003178e-20], [6.6988461, 1.18694e-19, 5.93642e-20, 3.56185e-20, 1.1875205e-20], [6.7142885, 1.17428e-19, 5.87312e-20, 3.52387e-20, 1.1748615e-20], [6.7297666, 1.16176e-19, 5.81051e-20, 3.48631e-20, 1.1623395e-20], [6.7452803, 1.14937e-19, 5.74857e-20, 3.44914e-20, 1.1499528e-20], [6.7608298, 1.13712e-19, 5.68731e-20, 3.41239e-20, 1.1377e-20], [6.7764151, 1.125e-19, 5.62671e-20, 3.37602e-20, 1.1255796e-20], [6.7920363, 1.11301e-19, 5.56676e-20, 3.34005e-20, 1.1135901e-20], [6.8076936, 1.10114e-19, 5.50745e-20, 3.30447e-20, 1.1017302e-20], [6.8233869, 1.08941e-19, 5.44879e-20, 3.26928e-20, 1.0899983e-20], [6.8391165, 1.07781e-19, 5.39077e-20, 3.23446e-20, 1.078393e-20], [6.8548823, 1.06633e-19, 5.33336e-20, 3.20002e-20, 1.0669131e-20], [6.8706844, 1.05497e-19, 5.27658e-20, 3.16595e-20, 1.055557e-20], [6.886523, 1.04373e-19, 5.22041e-20, 3.13225e-20, 1.0443234e-20], [6.902398, 1.03262e-19, 5.16485e-20, 3.09891e-20, 1.0332111e-20], [6.9183097, 1.02163e-19, 5.10989e-20, 3.06593e-20, 1.0222186e-20], [6.9342581, 1.01075e-19, 5.05551e-20, 3.03331e-20, 1.0113446e-20], [6.9502432, 9.99995e-20, 5.00173e-20, 3.00104e-20, 1.0005879e-20], [6.9662651, 9.89353e-20, 4.94852e-20, 2.96911e-20, 9.8994709e-21], [6.982324, 9.78826e-20, 4.89589e-20, 2.93753e-20, 9.7942098e-21], [6.99842, 9.68413e-20, 4.84383e-20, 2.9063e-20, 9.6900829e-21], [7.014553, 9.58112e-20, 4.79232e-20, 2.87539e-20, 9.5870778e-21], [7.0307232, 9.47921e-20, 4.74137e-20, 2.84482e-20, 9.4851822e-21], [7.0469307, 9.3784e-20, 4.69097e-20, 2.81458e-20, 9.3843838e-21], [7.0631755, 9.27868e-20, 4.64111e-20, 2.78467e-20, 9.2846708e-21], [7.0794578, 9.18004e-20, 4.59179e-20, 2.75508e-20, 9.1860311e-21], [7.0957777, 9.08245e-20, 4.543e-20, 2.7258e-20, 9.0884531e-21], [7.1121351, 8.98591e-20, 4.49474e-20, 2.69684e-20, 8.9919251e-21], [7.1285303, 8.89041e-20, 4.44699e-20, 2.66819e-20, 8.8964357e-21], [7.1449633, 8.79594e-20, 4.39976e-20, 2.63985e-20, 8.8019734e-21], [7.1614341, 8.70249e-20, 4.35303e-20, 2.61182e-20, 8.7085271e-21], [7.1779429, 8.61004e-20, 4.30681e-20, 2.58409e-20, 8.6160856e-21], [7.1944898, 8.51858e-20, 4.26108e-20, 2.55665e-20, 8.524638e-21], [7.2110748, 8.42811e-20, 4.21585e-20, 2.52951e-20, 8.4341734e-21], [7.227698, 8.33861e-20, 4.1711e-20, 2.50266e-20, 8.3446811e-21], [7.2443596, 8.25007e-20, 4.12684e-20, 2.4761e-20, 8.2561504e-21], [7.2610596, 8.16248e-20, 4.08304e-20, 2.44983e-20, 8.1685709e-21], [7.277798, 8.07584e-20, 4.03972e-20, 2.42383e-20, 8.0819321e-21], [7.2945751, 7.99012e-20, 3.99687e-20, 2.39812e-20, 7.9962239e-21], [7.3113908, 7.90532e-20, 3.95447e-20, 2.37268e-20, 7.911436e-21], [7.3282453, 7.82144e-20, 3.91253e-20, 2.34752e-20, 7.8275584e-21], [7.3451387, 7.73845e-20, 3.87104e-20, 2.32263e-20, 7.7445813e-21], [7.362071, 7.65636e-20, 3.83e-20, 2.298e-20, 7.6624947e-21], [7.3790423, 7.57514e-20, 3.78939e-20, 2.27364e-20, 7.5812891e-21], [7.3960528, 7.4948e-20, 3.74922e-20, 2.24953e-20, 7.5009547e-21], [7.4131024, 7.41532e-20, 3.70949e-20, 2.22569e-20, 7.4214822e-21], [7.4301914, 7.33669e-20, 3.67017e-20, 2.2021e-20, 7.3428621e-21], [7.4473197, 7.2589e-20, 3.63128e-20, 2.17877e-20, 7.2650851e-21], [7.4644876, 7.18195e-20, 3.59281e-20, 2.15569e-20, 7.1881421e-21], [7.481695, 7.10583e-20, 3.55475e-20, 2.13285e-20, 7.1120241e-21], [7.4989421, 7.03052e-20, 3.5171e-20, 2.11026e-20, 7.036722e-21], [7.5162289, 6.95601e-20, 3.47985e-20, 2.08791e-20, 6.9622269e-21], [7.5335556, 6.88231e-20, 3.443e-20, 2.0658e-20, 6.8885302e-21], [7.5509223, 6.80939e-20, 3.40654e-20, 2.04393e-20, 6.8156231e-21], [7.568329, 6.73726e-20, 3.37048e-20, 2.02229e-20, 6.7434971e-21], [7.5857758, 6.6659e-20, 3.3348e-20, 2.00088e-20, 6.6721436e-21], [7.6032628, 6.5953e-20, 3.2995e-20, 1.9797e-20, 6.6015544e-21], [7.6207901, 6.52546e-20, 3.26459e-20, 1.95875e-20, 6.531721e-21], [7.6383578, 6.45636e-20, 3.23004e-20, 1.93803e-20, 6.4626353e-21], [7.6559661, 6.38801e-20, 3.19587e-20, 1.91752e-20, 6.3942893e-21], [7.6736149, 6.32038e-20, 3.16206e-20, 1.89724e-20, 6.3266748e-21], [7.6913044, 6.25349e-20, 3.12861e-20, 1.87717e-20, 6.2597839e-21], [7.7090347, 6.1873e-20, 3.09552e-20, 1.85731e-20, 6.1936089e-21], [7.7268059, 6.12183e-20, 3.06279e-20, 1.83767e-20, 6.1281419e-21], [7.744618, 6.05705e-20, 3.0304e-20, 1.81824e-20, 6.0633753e-21], [7.7624712, 5.99297e-20, 2.99836e-20, 1.79902e-20, 5.9993014e-21], [7.7803655, 5.92957e-20, 2.96667e-20, 1.78e-20, 5.9359129e-21], [7.7983011, 5.86685e-20, 2.93531e-20, 1.76119e-20, 5.8732023e-21], [7.816278, 5.80481e-20, 2.90429e-20, 1.74257e-20, 5.8111622e-21], [7.8342964, 5.74342e-20, 2.8736e-20, 1.72416e-20, 5.7497854e-21], [7.8523563, 5.68269e-20, 2.84324e-20, 1.70594e-20, 5.6890647e-21], [7.8704579, 5.62261e-20, 2.8132e-20, 1.68792e-20, 5.6289931e-21], [7.8886012, 5.56317e-20, 2.78348e-20, 1.67009e-20, 5.5695635e-21], [7.9067863, 5.50437e-20, 2.75409e-20, 1.65245e-20, 5.510769e-21], [7.9250133, 5.4462e-20, 2.725e-20, 1.635e-20, 5.4526027e-21], [7.9432823, 5.38864e-20, 2.69623e-20, 1.61774e-20, 5.3950578e-21], [7.9615935, 5.3317e-20, 2.66776e-20, 1.60066e-20, 5.3381277e-21], [7.9799469, 5.27537e-20, 2.6396e-20, 1.58376e-20, 5.2818057e-21], [7.9983426, 5.21964e-20, 2.61174e-20, 1.56704e-20, 5.2260853e-21], [8.0167806, 5.16451e-20, 2.58417e-20, 1.5505e-20, 5.1709599e-21], [8.0352612, 5.10997e-20, 2.5569e-20, 1.53414e-20, 5.1164231e-21], [8.0537844, 5.056e-20, 2.52992e-20, 1.51795e-20, 5.0624687e-21], [8.0723503, 5.00262e-20, 2.50323e-20, 1.50194e-20, 5.0090903e-21], [8.090959, 4.9498e-20, 2.47683e-20, 1.4861e-20, 4.9562817e-21], [8.1096106, 4.89755e-20, 2.4507e-20, 1.47042e-20, 4.9040368e-21], [8.1283052, 4.84585e-20, 2.42486e-20, 1.45491e-20, 4.8523495e-21], [8.1470428, 4.79471e-20, 2.39929e-20, 1.43957e-20, 4.8012138e-21], [8.1658237, 4.74411e-20, 2.37399e-20, 1.4244e-20, 4.7506239e-21], [8.1846479, 4.69405e-20, 2.34897e-20, 1.40938e-20, 4.7005737e-21], [8.2035154, 4.64452e-20, 2.32421e-20, 1.39452e-20, 4.6510576e-21], [8.2224265, 4.59553e-20, 2.29971e-20, 1.37983e-20, 4.6020698e-21], [8.2413812, 4.54705e-20, 2.27548e-20, 1.36529e-20, 4.5536046e-21], [8.2603795, 4.4991e-20, 2.2515e-20, 1.3509e-20, 4.5056563e-21], [8.2794216, 4.45165e-20, 2.22778e-20, 1.33667e-20, 4.4582195e-21], [8.2985077, 4.40471e-20, 2.20431e-20, 1.32259e-20, 4.4112887e-21], [8.3176377, 4.35827e-20, 2.1811e-20, 1.30866e-20, 4.3648584e-21], [8.3368118, 4.31233e-20, 2.15813e-20, 1.29488e-20, 4.3189233e-21], [8.3560302, 4.26688e-20, 2.1354e-20, 1.28124e-20, 4.273478e-21], [8.3752928, 4.22191e-20, 2.11292e-20, 1.26775e-20, 4.2285173e-21], [8.3945999, 4.17742e-20, 2.09068e-20, 1.25441e-20, 4.1840361e-21], [8.4139514, 4.1334e-20, 2.06867e-20, 1.2412e-20, 4.1400292e-21], [8.4333476, 4.08986e-20, 2.0469e-20, 1.22814e-20, 4.0964914e-21], [8.4527885, 4.04678e-20, 2.02537e-20, 1.21522e-20, 4.0534179e-21], [8.4722741, 4.00415e-20, 2.00406e-20, 1.20243e-20, 4.0108035e-21], [8.4918048, 3.96198e-20, 1.98298e-20, 1.18979e-20, 3.9686435e-21], [8.5113804, 3.92027e-20, 1.96212e-20, 1.17727e-20, 3.926933e-21], [8.5310011, 3.87899e-20, 1.94148e-20, 1.16489e-20, 3.8856672e-21], [8.5506671, 3.83816e-20, 1.92107e-20, 1.15264e-20, 3.8448412e-21], [8.5703785, 3.79776e-20, 1.90087e-20, 1.14052e-20, 3.8044505e-21], [8.5901352, 3.75779e-20, 1.88089e-20, 1.12854e-20, 3.7644904e-21], [8.6099375, 3.71825e-20, 1.86112e-20, 1.11667e-20, 3.7249563e-21], [8.6297855, 3.67913e-20, 1.84157e-20, 1.10494e-20, 3.6858437e-21], [8.6496792, 3.64042e-20, 1.82222e-20, 1.09333e-20, 3.6471481e-21], [8.6696188, 3.60213e-20, 1.80307e-20, 1.08184e-20, 3.608865e-21], [8.6896043, 3.56425e-20, 1.78413e-20, 1.07048e-20, 3.57099e-21], [8.7096359, 3.52677e-20, 1.7654e-20, 1.05924e-20, 3.5335189e-21], [8.7297137, 3.48969e-20, 1.74686e-20, 1.04812e-20, 3.4964474e-21], [8.7498378, 3.453e-20, 1.72852e-20, 1.03711e-20, 3.4597711e-21], [8.7700082, 3.41671e-20, 1.71038e-20, 1.02623e-20, 3.4234859e-21], [8.7902252, 3.3808e-20, 1.69243e-20, 1.01546e-20, 3.3875877e-21], [8.8104887, 3.34528e-20, 1.67467e-20, 1.0048e-20, 3.3520724e-21], [8.830799, 3.31014e-20, 1.6571e-20, 9.94258e-21, 3.3169359e-21], [8.8511561, 3.27537e-20, 1.63971e-20, 9.83829e-21, 3.2821742e-21], [8.8715601, 3.24097e-20, 1.62252e-20, 9.73511e-21, 3.2477834e-21], [8.8920112, 3.20693e-20, 1.6055e-20, 9.63303e-21, 3.2137594e-21], [8.9125094, 3.17327e-20, 1.58867e-20, 9.53203e-21, 3.1800986e-21], [8.9330548, 3.13996e-20, 1.57202e-20, 9.43212e-21, 3.146797e-21], [8.9536477, 3.107e-20, 1.55555e-20, 9.33327e-21, 3.1138509e-21], [8.9742879, 3.0744e-20, 1.53925e-20, 9.23548e-21, 3.0812565e-21], [8.9949758, 3.04214e-20, 1.52312e-20, 9.13873e-21, 3.0490102e-21], [9.0157114, 3.01023e-20, 1.50717e-20, 9.04302e-21, 3.0171082e-21], [9.0364947, 2.97866e-20, 1.49139e-20, 8.94833e-21, 2.9855471e-21], [9.057326, 2.94743e-20, 1.47577e-20, 8.85465e-21, 2.9543231e-21], [9.0782053, 2.91653e-20, 1.46033e-20, 8.76197e-21, 2.9234329e-21], [9.0991327, 2.88597e-20, 1.44505e-20, 8.67028e-21, 2.8928728e-21], [9.1201084, 2.85572e-20, 1.42993e-20, 8.57957e-21, 2.8626396e-21], [9.1411324, 2.82581e-20, 1.41497e-20, 8.48983e-21, 2.8327297e-21], [9.1622049, 2.79621e-20, 1.40018e-20, 8.40106e-21, 2.8031398e-21], [9.183326, 2.76693e-20, 1.38554e-20, 8.31323e-21, 2.7738667e-21], [9.2044957, 2.73796e-20, 1.37106e-20, 8.22634e-21, 2.7449069e-21], [9.2257143, 2.7093e-20, 1.35673e-20, 8.14038e-21, 2.7162573e-21], [9.2469817, 2.68095e-20, 1.34256e-20, 8.05535e-21, 2.6879148e-21], [9.2682982, 2.6529e-20, 1.32854e-20, 7.97122e-21, 2.6598761e-21], [9.2896639, 2.62516e-20, 1.31467e-20, 7.888e-21, 2.632138e-21], [9.3110788, 2.59771e-20, 1.30095e-20, 7.80567e-21, 2.6046977e-21], [9.332543, 2.57056e-20, 1.28737e-20, 7.72423e-21, 2.5775519e-21], [9.3540567, 2.54369e-20, 1.27394e-20, 7.64365e-21, 2.5506977e-21], [9.3756201, 2.51712e-20, 1.26066e-20, 7.56395e-21, 2.5241322e-21], [9.3972331, 2.49083e-20, 1.24752e-20, 7.4851e-21, 2.4978524e-21], [9.418896, 2.46483e-20, 1.23452e-20, 7.4071e-21, 2.4718554e-21], [9.4406088, 2.4391e-20, 1.22166e-20, 7.32994e-21, 2.4461384e-21], [9.4623716, 2.41366e-20, 1.20894e-20, 7.25362e-21, 2.4206986e-21], [9.4841846, 2.38848e-20, 1.19635e-20, 7.17811e-21, 2.3955332e-21], [9.5060479, 2.36358e-20, 1.1839e-20, 7.10342e-21, 2.3706395e-21], [9.5279616, 2.33895e-20, 1.17159e-20, 7.02954e-21, 2.3460149e-21], [9.5499259, 2.31458e-20, 1.15941e-20, 6.95646e-21, 2.3216566e-21], [9.5719407, 2.29048e-20, 1.14736e-20, 6.88417e-21, 2.297562e-21], [9.5940063, 2.26664e-20, 1.13544e-20, 6.81266e-21, 2.2737286e-21], [9.6161228, 2.24306e-20, 1.12365e-20, 6.74193e-21, 2.2501539e-21], [9.6382902, 2.21973e-20, 1.11199e-20, 6.67197e-21, 2.2268354e-21], [9.6605088, 2.19666e-20, 1.10046e-20, 6.60276e-21, 2.2037706e-21], [9.6827786, 2.17384e-20, 1.08905e-20, 6.53432e-21, 2.1809571e-21], [9.7050997, 2.15126e-20, 1.07777e-20, 6.46661e-21, 2.1583925e-21], [9.7274722, 2.12894e-20, 1.06661e-20, 6.39965e-21, 2.1360746e-21], [9.7498964, 2.10686e-20, 1.05557e-20, 6.33343e-21, 2.114001e-21], [9.7723722, 2.08502e-20, 1.04465e-20, 6.26792e-21, 2.0921696e-21], [9.7948999, 2.06342e-20, 1.03386e-20, 6.20314e-21, 2.070578e-21], [9.8174794, 2.04206e-20, 1.02318e-20, 6.13907e-21, 2.0492242e-21], [9.8401111, 2.02093e-20, 1.01262e-20, 6.07571e-21, 2.0281061e-21], [9.8627949, 2.00004e-20, 1.00218e-20, 6.01305e-21, 2.0072217e-21], [9.8855309, 1.97938e-20, 9.91848e-21, 5.95109e-21, 1.9865688e-21], [9.9083194, 1.95895e-20, 9.81635e-21, 5.88981e-21, 1.9661456e-21], [9.9311605, 1.93875e-20, 9.71536e-21, 5.82922e-21, 1.9459501e-21], [9.9540542, 1.91877e-20, 9.6155e-21, 5.7693e-21, 1.9259805e-21], [9.9770006, 1.89902e-20, 9.51676e-21, 5.71006e-21, 1.9062351e-21], [10.0, 1.87949e-20, 9.41914e-21, 5.65148e-21, 1.886712e-21], [10.023052, 1.86018e-20, 9.32261e-21, 5.59357e-21, 1.8674095e-21], [10.046158, 1.84109e-20, 9.22719e-21, 5.53631e-21, 1.8483261e-21], [10.069317, 1.82222e-20, 9.13285e-21, 5.47971e-21, 1.8294602e-21], [10.092529, 1.80356e-20, 9.03959e-21, 5.42375e-21, 1.8108103e-21], [10.115795, 1.78512e-20, 8.9474e-21, 5.36844e-21, 1.792375e-21], [10.139114, 1.76689e-20, 8.85628e-21, 5.31377e-21, 1.7741529e-21], [10.162487, 1.74888e-20, 8.76622e-21, 5.25973e-21, 1.7561427e-21], [10.185914, 1.73107e-20, 8.67721e-21, 5.20633e-21, 1.7383433e-21], [10.209395, 1.71348e-20, 8.58926e-21, 5.15355e-21, 1.7207535e-21], [10.23293, 1.69609e-20, 8.50234e-21, 5.10141e-21, 1.7033723e-21], [10.256519, 1.67891e-20, 8.41647e-21, 5.04988e-21, 1.6861988e-21], [10.280163, 1.66194e-20, 8.33163e-21, 4.99898e-21, 1.6692322e-21], [10.303861, 1.64517e-20, 8.24782e-21, 4.94869e-21, 1.6524717e-21], [10.327614, 1.62861e-20, 8.16503e-21, 4.89902e-21, 1.6359169e-21], [10.351422, 1.61226e-20, 8.08328e-21, 4.84997e-21, 1.6195671e-21], [10.375284, 1.59611e-20, 8.00255e-21, 4.80153e-21, 1.6034221e-21], [10.399202, 1.58016e-20, 7.92284e-21, 4.7537e-21, 1.5874816e-21], [10.423174, 1.56442e-20, 7.84415e-21, 4.70649e-21, 1.5717457e-21], [10.447202, 1.54889e-20, 7.76649e-21, 4.6599e-21, 1.5562145e-21], [10.471285, 1.53356e-20, 7.68986e-21, 4.61391e-21, 1.5408882e-21], [10.495424, 1.51843e-20, 7.61425e-21, 4.56855e-21, 1.5257675e-21], [10.519619, 1.50351e-20, 7.53967e-21, 4.5238e-21, 1.5108531e-21], [10.543869, 1.4888e-20, 7.46614e-21, 4.47968e-21, 1.4961459e-21], [10.568175, 1.4743e-20, 7.39364e-21, 4.43618e-21, 1.4816472e-21], [10.592537, 1.46001e-20, 7.3222e-21, 4.39332e-21, 1.4673586e-21], [10.616956, 1.44593e-20, 7.25181e-21, 4.35109e-21, 1.4532818e-21], [10.64143, 1.43207e-20, 7.1825e-21, 4.3095e-21, 1.4394191e-21], [10.665961, 1.41842e-20, 7.11427e-21, 4.26856e-21, 1.4257731e-21], [10.690549, 1.40499e-20, 7.04714e-21, 4.22828e-21, 1.4123468e-21], [10.715193, 1.39179e-20, 6.98113e-21, 4.18868e-21, 1.3991437e-21], [10.739894, 1.37882e-20, 6.91625e-21, 4.14975e-21, 1.3861679e-21], [10.764652, 1.36607e-20, 6.85254e-21, 4.11152e-21, 1.3734238e-21], [10.789467, 1.35357e-20, 6.79001e-21, 4.07401e-21, 1.360917e-21], [10.81434, 1.34131e-20, 6.7287e-21, 4.03722e-21, 1.3486533e-21], [10.839269, 1.3293e-20, 6.66864e-21, 4.00118e-21, 1.3366397e-21], [10.864256, 1.31754e-20, 6.60987e-21, 3.96592e-21, 1.324884e-21], [10.889301, 1.30606e-20, 6.55244e-21, 3.93147e-21, 1.3133951e-21], [10.914403, 1.29485e-20, 6.4964e-21, 3.89784e-21, 1.3021832e-21], [10.939564, 1.28394e-20, 6.4418e-21, 3.86508e-21, 1.2912597e-21], [10.964782, 1.27332e-20, 6.38871e-21, 3.83323e-21, 1.2806377e-21], [10.990058, 1.26303e-20, 6.3372e-21, 3.80232e-21, 1.2703322e-21], [11.015393, 1.25307e-20, 6.28737e-21, 3.77242e-21, 1.2603601e-21], [11.040786, 1.24346e-20, 6.2393e-21, 3.74358e-21, 1.2507409e-21], [11.066238, 1.23423e-20, 6.19311e-21, 3.71587e-21, 1.2414967e-21], [11.091748, 1.2254e-20, 6.14893e-21, 3.68936e-21, 1.2326528e-21], [11.117317, 1.21701e-20, 6.1069e-21, 3.66414e-21, 1.2242384e-21], [11.142945, 1.20908e-20, 6.06718e-21, 3.64031e-21, 1.2162869e-21], [11.168632, 1.20165e-20, 6.02998e-21, 3.61799e-21, 1.208837e-21], [11.194379, 1.19477e-20, 5.99552e-21, 3.59731e-21, 1.2019333e-21], [11.220185, 1.1885e-20, 5.96405e-21, 3.57843e-21, 1.1956276e-21], [11.24605, 1.18288e-20, 5.93589e-21, 3.56153e-21, 1.1899802e-21], [11.271975, 1.178e-20, 5.91137e-21, 3.54682e-21, 1.1850616e-21], [11.297959, 1.17394e-20, 5.89092e-21, 3.53455e-21, 1.1809543e-21], [11.324004, 1.17079e-20, 5.87502e-21, 3.52501e-21, 1.1777558e-21], [11.350108, 1.16866e-20, 5.86426e-21, 3.51855e-21, 1.1755812e-21], [11.376273, 1.16771e-20, 5.85931e-21, 3.51559e-21, 1.1745675e-21], [11.402498, 1.16809e-20, 5.861e-21, 3.5166e-21, 1.1748785e-21], [11.428783, 1.16999e-20, 5.87031e-21, 3.52219e-21, 1.1767112e-21], [11.455129, 1.17367e-20, 5.88844e-21, 3.53307e-21, 1.1803039e-21], [11.481536, 1.17941e-20, 5.91685e-21, 3.55011e-21, 1.1859472e-21], [11.508004, 1.18756e-20, 5.95731e-21, 3.57439e-21, 1.1939976e-21], [11.534533, 1.19858e-20, 6.01204e-21, 3.60722e-21, 1.2048964e-21], [11.561122, 1.21301e-20, 6.0838e-21, 3.65028e-21, 1.2191946e-21], [11.587774, 1.23155e-20, 6.17606e-21, 3.70563e-21, 1.2375874e-21], [11.614486, 1.2551e-20, 6.29326e-21, 3.77596e-21, 1.2609614e-21], [11.64126, 1.28479e-20, 6.44114e-21, 3.86468e-21, 1.290462e-21], [11.668096, 1.32213e-20, 6.6272e-21, 3.97632e-21, 1.3275903e-21], [11.694994, 1.36912e-20, 6.86145e-21, 4.11687e-21, 1.3743473e-21], [11.721954, 1.42849e-20, 7.15749e-21, 4.2945e-21, 1.4334522e-21], [11.748976, 1.50402e-20, 7.53425e-21, 4.52055e-21, 1.5086878e-21], [11.77606, 1.60112e-20, 8.01877e-21, 4.81126e-21, 1.6054649e-21], [11.803206, 1.7278e-20, 8.6511e-21, 5.19066e-21, 1.7317909e-21], [11.830416, 1.89643e-20, 9.49306e-21, 5.69583e-21, 1.9000286e-21], [11.857687, 2.12713e-20, 1.06453e-20, 6.38719e-21, 2.130315e-21], [11.885022, 2.45509e-20, 1.22837e-20, 7.37024e-21, 2.4578219e-21], [11.91242, 2.9478e-20, 1.47459e-20, 8.84751e-21, 2.9500574e-21], [11.939881, 3.75352e-20, 1.87729e-20, 1.12637e-20, 3.7552705e-21], [11.967405, 5.27387e-20, 2.63731e-20, 1.58239e-20, 5.2751085e-21], [11.994993, 9.11108e-20, 4.55576e-20, 2.73345e-20, 9.1117913e-21], [12.022644, 3.58502e-19, 1.79252e-19, 1.07551e-19, 3.5850394e-20], [12.050359, 1.82236e-19, 9.11193e-20, 5.46716e-20, 1.8223995e-20], [12.078138, 7.27431e-20, 3.63742e-20, 2.18245e-20, 7.2751778e-21], [12.105981, 4.56809e-20, 2.28446e-20, 1.37068e-20, 4.5694524e-21], [12.133889, 3.35176e-20, 1.67644e-20, 1.00586e-20, 3.3535985e-21], [12.16186, 2.66579e-20, 1.33359e-20, 8.00155e-21, 2.6680812e-21], [12.189896, 2.22846e-20, 1.11506e-20, 6.69033e-21, 2.2311737e-21], [12.217997, 1.92719e-20, 9.6454e-21, 5.78724e-21, 1.9302975e-21], [12.246162, 1.70819e-20, 8.5515e-21, 5.1309e-21, 1.7116606e-21], [12.274392, 1.54253e-20, 7.72422e-21, 4.63453e-21, 1.546335e-21], [12.302688, 1.41328e-20, 7.07892e-21, 4.24735e-21, 1.4173954e-21], [12.331048, 1.3099e-20, 6.56287e-21, 3.93772e-21, 1.3142936e-21], [12.359474, 1.22547e-20, 6.14153e-21, 3.68492e-21, 1.2301253e-21], [12.387966, 1.1553e-20, 5.7914e-21, 3.47484e-21, 1.1601912e-21], [12.416523, 1.09609e-20, 5.49597e-21, 3.29758e-21, 1.101189e-21], [12.445146, 1.04544e-20, 5.24333e-21, 3.146e-21, 1.0507357e-21], [12.473835, 1.0016e-20, 5.02467e-21, 3.0148e-21, 1.0070731e-21], [12.50259, 9.63238e-21, 4.83339e-21, 2.90003e-21, 9.6887858e-22], [12.531412, 9.29351e-21, 4.66441e-21, 2.79865e-21, 9.3514068e-22], [12.5603, 8.99148e-21, 4.51382e-21, 2.70829e-21, 9.0507458e-22], [12.589254, 8.72011e-21, 4.37852e-21, 2.62711e-21, 8.7806335e-22], [12.618275, 8.47448e-21, 4.25606e-21, 2.55364e-21, 8.5361648e-22], [12.647363, 8.25065e-21, 4.14448e-21, 2.48669e-21, 8.3133986e-22], [12.676519, 8.04541e-21, 4.04216e-21, 2.4253e-21, 8.1091389e-22], [12.705741, 7.85614e-21, 3.94781e-21, 2.36868e-21, 7.9207713e-22], [12.735031, 7.68066e-21, 3.86033e-21, 2.3162e-21, 7.7461403e-22], [12.764388, 7.5172e-21, 3.77884e-21, 2.26731e-21, 7.5834561e-22], [12.793813, 7.36424e-21, 3.70259e-21, 2.22155e-21, 7.4312219e-22], [12.823306, 7.22052e-21, 3.63094e-21, 2.17857e-21, 7.2881789e-22], [12.852867, 7.08497e-21, 3.56337e-21, 2.13802e-21, 7.1532614e-22], [12.882496, 6.95668e-21, 3.49941e-21, 2.09964e-21, 7.0255629e-22], [12.912193, 6.83487e-21, 3.43868e-21, 2.06321e-21, 6.9043079e-22], [12.941958, 6.71887e-21, 3.38084e-21, 2.02851e-21, 6.78883e-22], [12.971793, 6.6081e-21, 3.32561e-21, 1.99537e-21, 6.6785537e-22], [13.001696, 6.50207e-21, 3.27274e-21, 1.96365e-21, 6.5729802e-22], [13.031668, 6.40033e-21, 3.22201e-21, 1.93321e-21, 6.4716747e-22], [13.061709, 6.3025e-21, 3.17323e-21, 1.90394e-21, 6.3742577e-22], [13.091819, 6.20824e-21, 3.12623e-21, 1.87574e-21, 6.2803956e-22], [13.121999, 6.11727e-21, 3.08086e-21, 1.84852e-21, 6.1897949e-22], [13.152248, 6.02932e-21, 3.037e-21, 1.8222e-21, 6.1021964e-22], [13.182567, 5.94416e-21, 2.99453e-21, 1.79672e-21, 6.01737e-22], [13.212956, 5.86158e-21, 2.95334e-21, 1.772e-21, 5.9351113e-22], [13.243415, 5.78141e-21, 2.91335e-21, 1.74801e-21, 5.8552377e-22], [13.273945, 5.70346e-21, 2.87447e-21, 1.72468e-21, 5.7775859e-22], [13.304544, 5.62761e-21, 2.83664e-21, 1.70198e-21, 5.7020097e-22], [13.335214, 5.55371e-21, 2.79978e-21, 1.67987e-21, 5.6283771e-22], [13.365955, 5.48165e-21, 2.76383e-21, 1.6583e-21, 5.5565692e-22], [13.396767, 5.41132e-21, 2.72874e-21, 1.63725e-21, 5.4864786e-22], [13.42765, 5.34261e-21, 2.69447e-21, 1.61668e-21, 5.4180078e-22], [13.458604, 5.27544e-21, 2.66096e-21, 1.59658e-21, 5.3510682e-22], [13.489629, 5.20974e-21, 2.62818e-21, 1.57691e-21, 5.2855791e-22], [13.520726, 5.14542e-21, 2.59609e-21, 1.55765e-21, 5.2214667e-22], [13.551894, 5.08241e-21, 2.56465e-21, 1.53879e-21, 5.1586637e-22], [13.583134, 5.02066e-21, 2.53384e-21, 1.52031e-21, 5.0971082e-22], [13.614447, 4.96011e-21, 2.50363e-21, 1.50218e-21, 5.0367434e-22], [13.645831, 4.9007e-21, 2.47399e-21, 1.48439e-21, 4.9775169e-22], [13.677288, 4.84238e-21, 2.44489e-21, 1.46694e-21, 4.9193805e-22], [13.708818, 4.78512e-21, 2.41632e-21, 1.44979e-21, 4.8622896e-22], [13.74042, 4.72887e-21, 2.38825e-21, 1.43295e-21, 4.8062027e-22], [13.772095, 4.67359e-21, 2.36067e-21, 1.4164e-21, 4.7510814e-22], [13.803843, 4.61924e-21, 2.33355e-21, 1.40013e-21, 4.6968902e-22], [13.835664, 4.5658e-21, 2.30688e-21, 1.38413e-21, 4.6435957e-22], [13.867558, 4.51322e-21, 2.28064e-21, 1.36838e-21, 4.591167e-22], [13.899526, 4.46148e-21, 2.25482e-21, 1.35289e-21, 4.5395751e-22], [13.931568, 4.41056e-21, 2.22941e-21, 1.33765e-21, 4.4887928e-22], [13.963684, 4.36043e-21, 2.20439e-21, 1.32264e-21, 4.4387948e-22], [13.995873, 4.31106e-21, 2.17976e-21, 1.30785e-21, 4.3895571e-22], [14.028137, 4.26243e-21, 2.15549e-21, 1.29329e-21, 4.3410572e-22], [14.060475, 4.21452e-21, 2.13158e-21, 1.27895e-21, 4.293274e-22], [14.092888, 4.16731e-21, 2.10802e-21, 1.26481e-21, 4.2461875e-22], [14.125375, 4.12079e-21, 2.0848e-21, 1.25088e-21, 4.1997788e-22], [14.157938, 4.07492e-21, 2.06191e-21, 1.23715e-21, 4.1540301e-22], [14.190575, 4.0297e-21, 2.03934e-21, 1.2236e-21, 4.1089244e-22], [14.223288, 3.98511e-21, 2.01709e-21, 1.21025e-21, 4.0644458e-22], [14.256076, 3.94114e-21, 1.99514e-21, 1.19708e-21, 4.0205788e-22], [14.28894, 3.89776e-21, 1.97349e-21, 1.1841e-21, 3.9773092e-22], [14.321879, 3.85497e-21, 1.95214e-21, 1.17128e-21, 3.9346229e-22], [14.354894, 3.81275e-21, 1.93107e-21, 1.15864e-21, 3.8925068e-22], [14.387986, 3.77109e-21, 1.91027e-21, 1.14616e-21, 3.8509484e-22], [14.421154, 3.72998e-21, 1.88976e-21, 1.13385e-21, 3.8099356e-22], [14.454398, 3.68941e-21, 1.86951e-21, 1.1217e-21, 3.769457e-22], [14.487719, 3.64936e-21, 1.84952e-21, 1.10971e-21, 3.7295014e-22], [14.521116, 3.60983e-21, 1.82979e-21, 1.09787e-21, 3.6900583e-22], [14.554591, 3.5708e-21, 1.81031e-21, 1.08618e-21, 3.6511176e-22], [14.588143, 3.53226e-21, 1.79107e-21, 1.07464e-21, 3.6126695e-22], [14.621772, 3.49421e-21, 1.77208e-21, 1.06325e-21, 3.5747047e-22], [14.655478, 3.45664e-21, 1.75333e-21, 1.052e-21, 3.537214e-22], [14.689263, 3.41953e-21, 1.73481e-21, 1.04088e-21, 3.5001888e-22], [14.723125, 3.38288e-21, 1.71651e-21, 1.02991e-21, 3.4636206e-22], [14.757065, 3.34668e-21, 1.69845e-21, 1.01907e-21, 3.4275015e-22], [14.791084, 3.31093e-21, 1.6806e-21, 1.00836e-21, 3.3918235e-22], [14.825181, 3.27561e-21, 1.66297e-21, 9.97783e-22, 3.356579e-22], [14.859356, 3.24072e-21, 1.64556e-21, 9.87334e-22, 3.3217608e-22], [14.893611, 3.20625e-21, 1.62835e-21, 9.7701e-22, 3.2873617e-22], [14.927944, 3.17219e-21, 1.61135e-21, 9.66811e-22, 3.2533749e-22], [14.962357, 3.13854e-21, 1.59455e-21, 9.56733e-22, 3.2197936e-22], [14.996848, 3.10529e-21, 1.57796e-21, 9.46775e-22, 3.1866114e-22], [15.03142, 3.07243e-21, 1.56156e-21, 9.36935e-22, 3.153822e-22], [15.066071, 3.03996e-21, 1.54535e-21, 9.27212e-22, 3.1214193e-22], [15.100802, 3.00787e-21, 1.52934e-21, 9.17603e-22, 3.0893973e-22], [15.135612, 2.97616e-21, 1.51351e-21, 9.08106e-22, 3.0577502e-22], [15.170504, 2.94483e-21, 1.49787e-21, 8.98721e-22, 3.0264725e-22], [15.205475, 2.91385e-21, 1.48241e-21, 8.89444e-22, 2.9955586e-22], [15.240528, 2.88324e-21, 1.46713e-21, 8.80276e-22, 2.9650031e-22], [15.275661, 2.85298e-21, 1.45202e-21, 8.71214e-22, 2.934801e-22], [15.310875, 2.82307e-21, 1.43709e-21, 8.62256e-22, 2.9049469e-22], [15.34617, 2.7935e-21, 1.42234e-21, 8.53401e-22, 2.8754361e-22], [15.381546, 2.76427e-21, 1.40775e-21, 8.44648e-22, 2.8462636e-22], [15.417005, 2.73538e-21, 1.39333e-21, 8.35996e-22, 2.8174247e-22], [15.452544, 2.70682e-21, 1.37907e-21, 8.27442e-22, 2.7889148e-22], [15.488166, 2.67858e-21, 1.36498e-21, 8.18985e-22, 2.7607293e-22], [15.52387, 2.65067e-21, 1.35104e-21, 8.10625e-22, 2.7328638e-22], [15.559656, 2.62307e-21, 1.33727e-21, 8.0236e-22, 2.7053139e-22], [15.595525, 2.59579e-21, 1.32365e-21, 7.94188e-22, 2.6780755e-22], [15.631476, 2.56881e-21, 1.31018e-21, 7.86109e-22, 2.6511443e-22], [15.667511, 2.54214e-21, 1.29687e-21, 7.78121e-22, 2.6245163e-22], [15.703628, 2.51576e-21, 1.2837e-21, 7.70222e-22, 2.5981875e-22], [15.739829, 2.48969e-21, 1.27069e-21, 7.62413e-22, 2.572154e-22], [15.776113, 2.4639e-21, 1.25782e-21, 7.5469e-22, 2.5464118e-22], [15.81248, 2.43841e-21, 1.24509e-21, 7.47055e-22, 2.5209574e-22], [15.848932, 2.4132e-21, 1.23251e-21, 7.39505e-22, 2.4957869e-22], [15.885467, 2.38827e-21, 1.22006e-21, 7.32039e-22, 2.4708967e-22], [15.922087, 2.36362e-21, 1.20776e-21, 7.24656e-22, 2.4462833e-22], [15.958791, 2.33925e-21, 1.19559e-21, 7.17355e-22, 2.4219432e-22], [15.99558, 2.31514e-21, 1.18356e-21, 7.10136e-22, 2.3978729e-22], [16.032454, 2.2913e-21, 1.17166e-21, 7.02996e-22, 2.3740691e-22], [16.069413, 2.26773e-21, 1.15989e-21, 6.95936e-22, 2.3505284e-22], [16.106456, 2.24442e-21, 1.14826e-21, 6.88953e-22, 2.3272476e-22], [16.143586, 2.22136e-21, 1.13675e-21, 6.82048e-22, 2.3042235e-22], [16.1808, 2.19856e-21, 1.12537e-21, 6.75219e-22, 2.2814529e-22], [16.218101, 2.17602e-21, 1.11411e-21, 6.68466e-22, 2.2589327e-22], [16.255488, 2.15372e-21, 1.10298e-21, 6.61786e-22, 2.23666e-22], [16.29296, 2.13166e-21, 1.09197e-21, 6.55181e-22, 2.2146315e-22], [16.330519, 2.10985e-21, 1.08108e-21, 6.48647e-22, 2.1928445e-22], [16.368165, 2.08828e-21, 1.07031e-21, 6.42186e-22, 2.1712961e-22], [16.405898, 2.06694e-21, 1.05966e-21, 6.35795e-22, 2.1499832e-22], [16.443717, 2.04584e-21, 1.04912e-21, 6.29475e-22, 2.1289032e-22], [16.481624, 2.02497e-21, 1.03871e-21, 6.23223e-22, 2.1080533e-22], [16.519618, 2.00433e-21, 1.0284e-21, 6.1704e-22, 2.0874307e-22], [16.5577, 1.98391e-21, 1.01821e-21, 6.10924e-22, 2.0670327e-22], [16.595869, 1.96372e-21, 1.00813e-21, 6.04876e-22, 2.0468567e-22], [16.634127, 1.94375e-21, 9.98154e-22, 5.98893e-22, 2.0269001e-22], [16.672472, 1.92399e-21, 9.88292e-22, 5.92975e-22, 2.0071603e-22], [16.710906, 1.90445e-21, 9.78536e-22, 5.87122e-22, 1.9876347e-22], [16.749429, 1.88513e-21, 9.68887e-22, 5.81332e-22, 1.9683209e-22], [16.78804, 1.86601e-21, 9.59342e-22, 5.75605e-22, 1.9492164e-22], [16.826741, 1.8471e-21, 9.49902e-22, 5.69941e-22, 1.9303187e-22], [16.86553, 1.8284e-21, 9.40563e-22, 5.64338e-22, 1.9116255e-22], [16.904409, 1.8099e-21, 9.31326e-22, 5.58796e-22, 1.8931344e-22], [16.943378, 1.7916e-21, 9.2219e-22, 5.53314e-22, 1.8748431e-22], [16.982437, 1.7735e-21, 9.13152e-22, 5.47891e-22, 1.8567492e-22], [17.021585, 1.7556e-21, 9.04212e-22, 5.42527e-22, 1.8388506e-22], [17.060824, 1.73789e-21, 8.95369e-22, 5.37221e-22, 1.8211449e-22], [17.100153, 1.72037e-21, 8.86621e-22, 5.31973e-22, 1.80363e-22], [17.139573, 1.70305e-21, 8.77968e-22, 5.26781e-22, 1.7863036e-22], [17.179084, 1.68591e-21, 8.69409e-22, 5.21645e-22, 1.7691637e-22], [17.218686, 1.66895e-21, 8.60942e-22, 5.16565e-22, 1.7522082e-22], [17.258379, 1.65218e-21, 8.52566e-22, 5.1154e-22, 1.7354348e-22], [17.298164, 1.63559e-21, 8.44281e-22, 5.06568e-22, 1.7188416e-22], [17.33804, 1.61918e-21, 8.36084e-22, 5.01651e-22, 1.7024265e-22], [17.378008, 1.60295e-21, 8.27977e-22, 4.96786e-22, 1.6861876e-22], [17.418069, 1.58689e-21, 8.19956e-22, 4.91974e-22, 1.6701228e-22], [17.458222, 1.57101e-21, 8.12022e-22, 4.87213e-22, 1.6542301e-22], [17.498467, 1.55529e-21, 8.04174e-22, 4.82504e-22, 1.6385077e-22], [17.538805, 1.53975e-21, 7.96409e-22, 4.77846e-22, 1.6229535e-22], [17.579236, 1.52437e-21, 7.88729e-22, 4.73237e-22, 1.6075659e-22], [17.61976, 1.50916e-21, 7.8113e-22, 4.68678e-22, 1.5923428e-22], [17.660378, 1.49412e-21, 7.73614e-22, 4.64168e-22, 1.5772824e-22], [17.70109, 1.47924e-21, 7.66178e-22, 4.59707e-22, 1.562383e-22], [17.741895, 1.46451e-21, 7.58822e-22, 4.55293e-22, 1.5476426e-22], [17.782794, 1.44995e-21, 7.51545e-22, 4.50927e-22, 1.5330597e-22], [17.823788, 1.43554e-21, 7.44346e-22, 4.46608e-22, 1.5186323e-22], [17.864876, 1.42129e-21, 7.37224e-22, 4.42335e-22, 1.5043588e-22], [17.906059, 1.40719e-21, 7.30179e-22, 4.38107e-22, 1.4902375e-22], [17.947336, 1.39324e-21, 7.23209e-22, 4.33925e-22, 1.4762667e-22], [17.988709, 1.37944e-21, 7.16314e-22, 4.29788e-22, 1.4624446e-22], [18.030177, 1.36579e-21, 7.09492e-22, 4.25695e-22, 1.4487698e-22], [18.071741, 1.35229e-21, 7.02744e-22, 4.21647e-22, 1.4352405e-22], [18.113401, 1.33894e-21, 6.96068e-22, 4.17641e-22, 1.4218551e-22], [18.155157, 1.32572e-21, 6.89464e-22, 4.13678e-22, 1.4086121e-22], [18.197009, 1.31265e-21, 6.8293e-22, 4.09758e-22, 1.3955099e-22], [18.238957, 1.29972e-21, 6.76466e-22, 4.05879e-22, 1.382547e-22], [18.281002, 1.28693e-21, 6.70071e-22, 4.02043e-22, 1.3697217e-22], [18.323144, 1.27428e-21, 6.63744e-22, 3.98247e-22, 1.3570326e-22], [18.365383, 1.26176e-21, 6.57486e-22, 3.94491e-22, 1.3444782e-22], [18.40772, 1.24938e-21, 6.51294e-22, 3.90776e-22, 1.3320571e-22], [18.450154, 1.23713e-21, 6.45168e-22, 3.87101e-22, 1.3197677e-22], [18.492686, 1.22501e-21, 6.39107e-22, 3.83464e-22, 1.3076087e-22], [18.535316, 1.21302e-21, 6.33112e-22, 3.79867e-22, 1.2955785e-22], [18.578045, 1.20116e-21, 6.2718e-22, 3.76308e-22, 1.2836759e-22], [18.620871, 1.18943e-21, 6.21312e-22, 3.72787e-22, 1.2718993e-22], [18.663797, 1.17783e-21, 6.15507e-22, 3.69304e-22, 1.2602475e-22], [18.706821, 1.16635e-21, 6.09763e-22, 3.65858e-22, 1.2487191e-22], [18.749945, 1.15499e-21, 6.04081e-22, 3.62448e-22, 1.2373126e-22], [18.793168, 1.14375e-21, 5.98459e-22, 3.59075e-22, 1.2260269e-22], [18.836491, 1.13264e-21, 5.92897e-22, 3.55738e-22, 1.2148605e-22], [18.879913, 1.12165e-21, 5.87395e-22, 3.52437e-22, 1.2038123e-22], [18.923436, 1.11077e-21, 5.81952e-22, 3.49171e-22, 1.1928808e-22], [18.967059, 1.10001e-21, 5.76566e-22, 3.4594e-22, 1.1820649e-22], [19.010783, 1.08937e-21, 5.71238e-22, 3.42743e-22, 1.1713633e-22], [19.054607, 1.07884e-21, 5.65967e-22, 3.3958e-22, 1.1607747e-22], [19.098533, 1.06842e-21, 5.60751e-22, 3.36451e-22, 1.1502979e-22], [19.142559, 1.05812e-21, 5.55592e-22, 3.33355e-22, 1.1399318e-22], [19.186687, 1.04793e-21, 5.50488e-22, 3.30293e-22, 1.1296751e-22], [19.230917, 1.03785e-21, 5.45438e-22, 3.27263e-22, 1.1195266e-22], [19.275249, 1.02787e-21, 5.40441e-22, 3.24265e-22, 1.1094852e-22], [19.319683, 1.01801e-21, 5.35498e-22, 3.21299e-22, 1.0995497e-22], [19.36422, 1.00825e-21, 5.30608e-22, 3.18365e-22, 1.0897189e-22], [19.408859, 9.98591e-22, 5.2577e-22, 3.15462e-22, 1.0799919e-22], [19.453601, 9.89041e-22, 5.20983e-22, 3.1259e-22, 1.0703673e-22], [19.498446, 9.79593e-22, 5.16248e-22, 3.09749e-22, 1.0608442e-22], [19.543395, 9.70247e-22, 5.11563e-22, 3.06938e-22, 1.0514214e-22], [19.588447, 9.61003e-22, 5.06928e-22, 3.04157e-22, 1.0420979e-22], [19.633603, 9.51858e-22, 5.02342e-22, 3.01405e-22, 1.0328726e-22], [19.678863, 9.42812e-22, 4.97805e-22, 2.98683e-22, 1.0237444e-22], [19.724227, 9.33863e-22, 4.93317e-22, 2.9599e-22, 1.0147123e-22], [19.769696, 9.25012e-22, 4.88876e-22, 2.93326e-22, 1.0057753e-22], [19.81527, 9.16256e-22, 4.84483e-22, 2.9069e-22, 9.969324e-23], [19.860949, 9.07595e-22, 4.80136e-22, 2.88082e-22, 9.8818251e-23], [19.906733, 8.99027e-22, 4.75836e-22, 2.85502e-22, 9.7952468e-23], [19.952623, 8.90553e-22, 4.71582e-22, 2.82949e-22, 9.7095793e-23], [19.998619, 8.8217e-22, 4.67373e-22, 2.80424e-22, 9.6248128e-23], [20.04472, 8.73878e-22, 4.63209e-22, 2.77925e-22, 9.5409378e-23], [20.090928, 8.65676e-22, 4.59089e-22, 2.75453e-22, 9.4579447e-23], [20.137242, 8.57563e-22, 4.55013e-22, 2.73008e-22, 9.3758243e-23], [20.183664, 8.49538e-22, 4.50981e-22, 2.70589e-22, 9.2945672e-23], [20.230192, 8.416e-22, 4.46992e-22, 2.68195e-22, 9.2141642e-23], [20.276827, 8.33749e-22, 4.43045e-22, 2.65827e-22, 9.1346064e-23], [20.32357, 8.25983e-22, 4.39141e-22, 2.63485e-22, 9.0558847e-23], [20.370421, 8.18301e-22, 4.35278e-22, 2.61167e-22, 8.9779903e-23], [20.417379, 8.10704e-22, 4.31457e-22, 2.58874e-22, 8.9009145e-23], [20.464446, 8.03188e-22, 4.27676e-22, 2.56605e-22, 8.8246487e-23], [20.511622, 7.95755e-22, 4.23935e-22, 2.54361e-22, 8.7491843e-23], [20.558906, 7.88403e-22, 4.20235e-22, 2.52141e-22, 8.6745128e-23], [20.606299, 7.81132e-22, 4.16574e-22, 2.49945e-22, 8.6006259e-23], [20.653802, 7.7394e-22, 4.12952e-22, 2.47771e-22, 8.5275154e-23], [20.701413, 7.66826e-22, 4.09369e-22, 2.45622e-22, 8.4551731e-23], [20.749135, 7.5979e-22, 4.05825e-22, 2.43495e-22, 8.383591e-23], [20.796967, 7.52832e-22, 4.02318e-22, 2.41391e-22, 8.3127611e-23], [20.844909, 7.4595e-22, 3.98849e-22, 2.39309e-22, 8.2426756e-23], [20.892961, 7.39143e-22, 3.95417e-22, 2.3725e-22, 8.1733266e-23], [20.941125, 7.32411e-22, 3.92022e-22, 2.35213e-22, 8.1047066e-23], [20.989399, 7.25754e-22, 3.88663e-22, 2.33198e-22, 8.036808e-23], [21.037784, 7.19169e-22, 3.8534e-22, 2.31204e-22, 7.9696232e-23], [21.086281, 7.12657e-22, 3.82052e-22, 2.29231e-22, 7.9031448e-23], [21.13489, 7.06217e-22, 3.788e-22, 2.2728e-22, 7.8373656e-23], [21.183611, 6.99848e-22, 3.75583e-22, 2.2535e-22, 7.7722784e-23], [21.232445, 6.9355e-22, 3.72401e-22, 2.23441e-22, 7.7078759e-23], [21.28139, 6.87321e-22, 3.69253e-22, 2.21552e-22, 7.6441512e-23], [21.330449, 6.81162e-22, 3.66138e-22, 2.19683e-22, 7.5810973e-23], [21.379621, 6.75071e-22, 3.63057e-22, 2.17834e-22, 7.5187073e-23], [21.428906, 6.69047e-22, 3.6001e-22, 2.16006e-22, 7.4569744e-23], [21.478305, 6.63091e-22, 3.56995e-22, 2.14197e-22, 7.395892e-23], [21.527817, 6.57201e-22, 3.54012e-22, 2.12407e-22, 7.3354534e-23], [21.577444, 6.51377e-22, 3.51062e-22, 2.10637e-22, 7.2756522e-23], [21.627185, 6.45618e-22, 3.48144e-22, 2.08886e-22, 7.2164818e-23], [21.677041, 6.39924e-22, 3.45257e-22, 2.07154e-22, 7.1579359e-23], [21.727012, 6.34293e-22, 3.42401e-22, 2.05441e-22, 7.1000082e-23], [21.777098, 6.28726e-22, 3.39577e-22, 2.03746e-22, 7.0426925e-23], [21.827299, 6.23221e-22, 3.36783e-22, 2.0207e-22, 6.9859828e-23], [21.877616, 6.17779e-22, 3.34019e-22, 2.00411e-22, 6.9298729e-23], [21.928049, 6.12398e-22, 3.31285e-22, 1.98771e-22, 6.874357e-23], [21.978599, 6.07078e-22, 3.28581e-22, 1.97149e-22, 6.8194291e-23], [22.029265, 6.01818e-22, 3.25907e-22, 1.95544e-22, 6.7650835e-23], [22.080047, 5.96619e-22, 3.23261e-22, 1.93957e-22, 6.7113145e-23], [22.130947, 5.91478e-22, 3.20645e-22, 1.92387e-22, 6.6581164e-23], [22.181964, 5.86396e-22, 3.18057e-22, 1.90834e-22, 6.6054836e-23], [22.233099, 5.81373e-22, 3.15497e-22, 1.89298e-22, 6.5534108e-23], [22.284351, 5.76407e-22, 3.12966e-22, 1.87779e-22, 6.5018926e-23], [22.335722, 5.71498e-22, 3.10462e-22, 1.86277e-22, 6.4509235e-23], [22.387211, 5.66646e-22, 3.07986e-22, 1.84791e-22, 6.4004985e-23], [22.438819, 5.6185e-22, 3.05537e-22, 1.83322e-22, 6.3506122e-23], [22.490546, 5.57109e-22, 3.03115e-22, 1.81869e-22, 6.3012598e-23], [22.542392, 5.52424e-22, 3.00719e-22, 1.80432e-22, 6.2524361e-23], [22.594358, 5.47793e-22, 2.98351e-22, 1.7901e-22, 6.2041364e-23], [22.646443, 5.43217e-22, 2.96008e-22, 1.77605e-22, 6.1563556e-23], [22.698649, 5.38694e-22, 2.93692e-22, 1.76215e-22, 6.1090892e-23], [22.750974, 5.34225e-22, 2.91401e-22, 1.74841e-22, 6.0623323e-23], [22.803421, 5.29808e-22, 2.89136e-22, 1.73482e-22, 6.0160805e-23], [22.855988, 5.25444e-22, 2.86896e-22, 1.72138e-22, 5.9703293e-23], [22.908677, 5.21132e-22, 2.84682e-22, 1.70809e-22, 5.9250741e-23], [22.961486, 5.16871e-22, 2.82492e-22, 1.69495e-22, 5.8803107e-23], [23.014418, 5.12662e-22, 2.80327e-22, 1.68196e-22, 5.8360347e-23], [23.067472, 5.08503e-22, 2.78187e-22, 1.66912e-22, 5.7922421e-23], [23.120648, 5.04395e-22, 2.76071e-22, 1.65642e-22, 5.7489286e-23], [23.173946, 5.00336e-22, 2.73978e-22, 1.64387e-22, 5.7060904e-23], [23.227368, 4.96327e-22, 2.7191e-22, 1.63146e-22, 5.6637234e-23], [23.280913, 4.92367e-22, 2.69866e-22, 1.61919e-22, 5.6218239e-23], [23.334581, 4.88457e-22, 2.67845e-22, 1.60707e-22, 5.580388e-23], [23.388372, 4.84594e-22, 2.65847e-22, 1.59508e-22, 5.5394121e-23], [23.442288, 4.8078e-22, 2.63872e-22, 1.58323e-22, 5.4988926e-23], [23.496328, 4.77014e-22, 2.61921e-22, 1.57152e-22, 5.458826e-23], [23.550493, 4.73295e-22, 2.59992e-22, 1.55995e-22, 5.4192089e-23], [23.604782, 4.69623e-22, 2.58086e-22, 1.54851e-22, 5.3800381e-23], [23.659197, 4.65998e-22, 2.56202e-22, 1.53721e-22, 5.3413102e-23], [23.713737, 4.6242e-22, 2.54341e-22, 1.52604e-22, 5.3030223e-23], [23.768403, 4.58888e-22, 2.52501e-22, 1.51501e-22, 5.2651711e-23], [23.823195, 4.55402e-22, 2.50684e-22, 1.5041e-22, 5.2277539e-23], [23.878113, 4.51962e-22, 2.48889e-22, 1.49333e-22, 5.1907679e-23], [23.933158, 4.48567e-22, 2.47115e-22, 1.48269e-22, 5.1542102e-23], [23.988329, 4.45218e-22, 2.45363e-22, 1.47218e-22, 5.1180783e-23], [24.043628, 4.41913e-22, 2.43633e-22, 1.4618e-22, 5.0823696e-23], [24.099054, 4.38654e-22, 2.41924e-22, 1.45154e-22, 5.0470819e-23], [24.154608, 4.35439e-22, 2.40236e-22, 1.44141e-22, 5.0122127e-23], [24.21029, 4.32268e-22, 2.38569e-22, 1.43141e-22, 4.97776e-23], [24.266101, 4.29142e-22, 2.36923e-22, 1.42154e-22, 4.9437216e-23], [24.32204, 4.2606e-22, 2.35299e-22, 1.41179e-22, 4.9100957e-23], [24.378108, 4.23022e-22, 2.33695e-22, 1.40217e-22, 4.8768804e-23], [24.434306, 4.20027e-22, 2.32112e-22, 1.39267e-22, 4.8440742e-23], [24.490632, 4.17076e-22, 2.3055e-22, 1.3833e-22, 4.8116754e-23], [24.547089, 4.14169e-22, 2.29008e-22, 1.37405e-22, 4.7796826e-23], [24.603676, 4.11305e-22, 2.27487e-22, 1.36492e-22, 4.7480947e-23], [24.660393, 4.08485e-22, 2.25986e-22, 1.35592e-22, 4.7169105e-23], [24.717241, 4.05708e-22, 2.24506e-22, 1.34704e-22, 4.6861291e-23], [24.774221, 4.02974e-22, 2.23047e-22, 1.33828e-22, 4.6557496e-23], [24.831331, 4.00283e-22, 2.21607e-22, 1.32964e-22, 4.6257715e-23], [24.888573, 3.97636e-22, 2.20188e-22, 1.32113e-22, 4.5961944e-23], [24.945947, 3.95031e-22, 2.1879e-22, 1.31274e-22, 4.5670179e-23], [25.003454, 3.9247e-22, 2.17412e-22, 1.30447e-22, 4.5382419e-23], [25.061093, 3.89953e-22, 2.16054e-22, 1.29633e-22, 4.5098667e-23], [25.118864, 3.87478e-22, 2.14717e-22, 1.2883e-22, 4.4818926e-23], [25.176769, 3.85047e-22, 2.134e-22, 1.2804e-22, 4.45432e-23], [25.234808, 3.8266e-22, 2.12104e-22, 1.27262e-22, 4.4271497e-23], [25.29298, 3.80317e-22, 2.10828e-22, 1.26497e-22, 4.4003828e-23], [25.351286, 3.78017e-22, 2.09573e-22, 1.25744e-22, 4.3740205e-23], [25.409727, 3.75762e-22, 2.08339e-22, 1.25003e-22, 4.3480643e-23], [25.468303, 3.7355e-22, 2.07125e-22, 1.24275e-22, 4.322516e-23], [25.527013, 3.71384e-22, 2.05932e-22, 1.23559e-22, 4.2973778e-23], [25.585859, 3.69262e-22, 2.0476e-22, 1.22856e-22, 4.2726519e-23], [25.64484, 3.67186e-22, 2.0361e-22, 1.22166e-22, 4.2483411e-23], [25.703958, 3.65155e-22, 2.0248e-22, 1.21488e-22, 4.2244484e-23], [25.763212, 3.6317e-22, 2.01373e-22, 1.20824e-22, 4.2009772e-23], [25.822602, 3.61231e-22, 2.00287e-22, 1.20172e-22, 4.1779314e-23], [25.882129, 3.5934e-22, 1.99222e-22, 1.19533e-22, 4.1553152e-23], [25.941794, 3.57495e-22, 1.9818e-22, 1.18908e-22, 4.133133e-23], [26.001596, 3.55699e-22, 1.97161e-22, 1.18296e-22, 4.1113902e-23], [26.061535, 3.53952e-22, 1.96164e-22, 1.17698e-22, 4.0900921e-23], [26.121614, 3.52254e-22, 1.9519e-22, 1.17114e-22, 4.0692449e-23], [26.18183, 3.50606e-22, 1.94239e-22, 1.16543e-22, 4.0488552e-23], [26.242185, 3.49008e-22, 1.93312e-22, 1.15987e-22, 4.0289303e-23], [26.30268, 3.47463e-22, 1.92409e-22, 1.15446e-22, 4.009478e-23], [26.363314, 3.4597e-22, 1.91531e-22, 1.14919e-22, 3.9905068e-23], [26.424088, 3.44531e-22, 1.90678e-22, 1.14407e-22, 3.972026e-23], [26.485001, 3.43147e-22, 1.8985e-22, 1.1391e-22, 3.9540457e-23], [26.546056, 3.41819e-22, 1.89049e-22, 1.13429e-22, 3.9365765e-23], [26.607251, 3.40548e-22, 1.88274e-22, 1.12964e-22, 3.9196303e-23], [26.668587, 3.39336e-22, 1.87526e-22, 1.12516e-22, 3.9032198e-23], [26.730064, 3.38184e-22, 1.86807e-22, 1.12084e-22, 3.8873587e-23], [26.791683, 3.37093e-22, 1.86116e-22, 1.1167e-22, 3.8720619e-23], [26.853444, 3.36067e-22, 1.85455e-22, 1.11273e-22, 3.8573453e-23], [26.915348, 3.35105e-22, 1.84825e-22, 1.10895e-22, 3.8432263e-23], [26.977394, 3.34211e-22, 1.84226e-22, 1.10535e-22, 3.8297236e-23], [27.039584, 3.33387e-22, 1.83659e-22, 1.10196e-22, 3.8168575e-23], [27.101916, 3.32635e-22, 1.83126e-22, 1.09876e-22, 3.8046499e-23], [27.164393, 3.31957e-22, 1.82628e-22, 1.09577e-22, 3.7931244e-23], [27.227013, 3.31356e-22, 1.82166e-22, 1.093e-22, 3.7823067e-23], [27.289778, 3.30835e-22, 1.81742e-22, 1.09045e-22, 3.7722245e-23], [27.352687, 3.30398e-22, 1.81356e-22, 1.08814e-22, 3.7629079e-23], [27.415742, 3.30047e-22, 1.81011e-22, 1.08607e-22, 3.7543895e-23], [27.478942, 3.29787e-22, 1.80709e-22, 1.08425e-22, 3.7467047e-23], [27.542287, 3.29621e-22, 1.80451e-22, 1.08271e-22, 3.7398916e-23], [27.605779, 3.29555e-22, 1.8024e-22, 1.08144e-22, 3.7339922e-23], [27.669416, 3.29592e-22, 1.80078e-22, 1.08047e-22, 3.7290516e-23], [27.733201, 3.29737e-22, 1.79966e-22, 1.0798e-22, 3.7251192e-23], [27.797133, 3.29998e-22, 1.79909e-22, 1.07946e-22, 3.7222487e-23], [27.861212, 3.30378e-22, 1.7991e-22, 1.07946e-22, 3.7204988e-23], [27.925438, 3.30886e-22, 1.7997e-22, 1.07982e-22, 3.7199335e-23], [27.989813, 3.31529e-22, 1.80094e-22, 1.08056e-22, 3.7206228e-23], [28.054336, 3.32314e-22, 1.80286e-22, 1.08171e-22, 3.7226433e-23], [28.119008, 3.33249e-22, 1.80549e-22, 1.0833e-22, 3.726079e-23], [28.183829, 3.34346e-22, 1.8089e-22, 1.08534e-22, 3.7310224e-23], [28.2488, 3.35614e-22, 1.81312e-22, 1.08787e-22, 3.7375748e-23], [28.31392, 3.37064e-22, 1.81821e-22, 1.09093e-22, 3.7458481e-23], [28.37919, 3.38709e-22, 1.82424e-22, 1.09454e-22, 3.7559658e-23], [28.444611, 3.40563e-22, 1.83127e-22, 1.09876e-22, 3.7680642e-23], [28.510183, 3.42642e-22, 1.83938e-22, 1.10363e-22, 3.7822946e-23], [28.575905, 3.44963e-22, 1.84866e-22, 1.1092e-22, 3.7988249e-23], [28.64178, 3.47545e-22, 1.8592e-22, 1.11552e-22, 3.8178419e-23], [28.707806, 3.50408e-22, 1.87109e-22, 1.12266e-22, 3.8395539e-23], [28.773984, 3.53577e-22, 1.88447e-22, 1.13068e-22, 3.864194e-23], [28.840315, 3.57078e-22, 1.89946e-22, 1.13968e-22, 3.8920233e-23], [28.906799, 3.6094e-22, 1.91621e-22, 1.14973e-22, 3.9233356e-23], [28.973436, 3.65198e-22, 1.93489e-22, 1.16093e-22, 3.9584621e-23], [29.040227, 3.69888e-22, 1.95567e-22, 1.1734e-22, 3.9977775e-23], [29.107171, 3.75054e-22, 1.97878e-22, 1.18727e-22, 4.0417073e-23], [29.17427, 3.80744e-22, 2.00447e-22, 1.20268e-22, 4.0907362e-23], [29.241524, 3.87015e-22, 2.03299e-22, 1.2198e-22, 4.1454188e-23], [29.308932, 3.93929e-22, 2.06468e-22, 1.23881e-22, 4.2063921e-23], [29.376497, 4.01561e-22, 2.09991e-22, 1.25994e-22, 4.2743913e-23], [29.444216, 4.09996e-22, 2.13909e-22, 1.28346e-22, 4.3502693e-23], [29.512092, 4.19334e-22, 2.18273e-22, 1.30964e-22, 4.4350201e-23], [29.580125, 4.29691e-22, 2.23141e-22, 1.33885e-22, 4.52981e-23], [29.648314, 4.41204e-22, 2.28582e-22, 1.37149e-22, 4.6360148e-23], [29.71666, 4.54038e-22, 2.34678e-22, 1.40807e-22, 4.7552701e-23], [29.785164, 4.68387e-22, 2.41525e-22, 1.44915e-22, 4.8895336e-23], [29.853826, 4.84489e-22, 2.49244e-22, 1.49546e-22, 5.0411694e-23], [29.922646, 5.02629e-22, 2.57977e-22, 1.54786e-22, 5.2130576e-23], [29.991625, 5.23163e-22, 2.67901e-22, 1.60741e-22, 5.4087431e-23], [30.060763, 5.46532e-22, 2.79238e-22, 1.67543e-22, 5.6326389e-23], [30.13006, 5.7329e-22, 2.92265e-22, 1.75359e-22, 5.8903082e-23], [30.199517, 6.04148e-22, 3.07338e-22, 1.84403e-22, 6.1888652e-23], [30.269134, 6.4003e-22, 3.24919e-22, 1.94952e-22, 6.5375563e-23], [30.338912, 6.82158e-22, 3.4562e-22, 2.07372e-22, 6.9486228e-23], [30.40885, 7.32188e-22, 3.70269e-22, 2.22161e-22, 7.438622e-23], [30.47895, 7.92414e-22, 4.00013e-22, 2.40008e-22, 8.0305141e-23], [30.549211, 8.66113e-22, 4.36491e-22, 2.61895e-22, 8.7570907e-23], [30.619634, 9.58136e-22, 4.82131e-22, 2.89278e-22, 9.6668723e-23], [30.69022, 1.07598e-21, 5.40678e-22, 3.24407e-22, 1.083483e-22], [30.760968, 1.23187e-21, 6.1825e-22, 3.7095e-22, 1.2383267e-22], [30.83188, 1.4472e-21, 7.25544e-22, 4.35326e-22, 1.4526183e-22], [30.902954, 1.76313e-21, 8.83136e-22, 5.29882e-22, 1.7675074e-22], [30.974193, 2.27016e-21, 1.13629e-21, 6.81771e-22, 2.2735144e-22], [31.045596, 3.21398e-21, 1.60783e-21, 9.64697e-22, 3.2163124e-22], [31.117163, 5.5771e-21, 2.78903e-21, 1.67342e-21, 5.5784231e-22], [31.188896, 2.21221e-20, 1.10612e-20, 6.6367e-21, 2.212241e-21], [31.260794, 1.09969e-20, 5.49868e-21, 3.29921e-21, 1.0997551e-21], [31.332857, 4.36927e-21, 2.18522e-21, 1.31113e-21, 4.3708819e-22], [31.405087, 2.71476e-21, 1.3583e-21, 8.1498e-22, 2.7173066e-22], [31.477483, 1.96412e-21, 9.83308e-22, 5.89985e-22, 1.967574e-22], [31.550046, 1.53623e-21, 7.69686e-22, 4.61812e-22, 1.5405735e-22], [31.622777, 1.2602e-21, 6.31987e-22, 3.79192e-22, 1.2654083e-22], [31.695675, 1.06768e-21, 5.36032e-22, 3.21619e-22, 1.0737226e-22], [31.768741, 9.25985e-22, 4.65476e-22, 2.79285e-22, 9.3282408e-23], [31.841975, 8.175e-22, 4.11516e-22, 2.4691e-22, 8.2510866e-23], [31.915379, 7.31911e-22, 3.68993e-22, 2.21396e-22, 7.4025605e-23], [31.988951, 6.62768e-22, 3.34681e-22, 2.00809e-22, 6.7181553e-23], [32.062693, 6.05832e-22, 3.06461e-22, 1.83877e-22, 6.1554765e-23], [32.136605, 5.58203e-22, 2.82883e-22, 1.6973e-22, 5.6855146e-23], [32.210688, 5.1783e-22, 2.62919e-22, 1.57752e-22, 5.2877491e-23], [32.284941, 4.83219e-22, 2.45825e-22, 1.47495e-22, 4.9472557e-23], [32.359366, 4.53259e-22, 2.31044e-22, 1.38626e-22, 4.6529205e-23], [32.433962, 4.27107e-22, 2.18154e-22, 1.30892e-22, 4.3962982e-23], [32.50873, 4.04107e-22, 2.06828e-22, 1.24097e-22, 4.1708572e-23], [32.58367, 3.83747e-22, 1.9681e-22, 1.18086e-22, 3.9714673e-23], [32.658783, 3.65618e-22, 1.87895e-22, 1.12737e-22, 3.7940443e-23], [32.734069, 3.4939e-22, 1.79918e-22, 1.07951e-22, 3.6352969e-23], [32.809529, 3.34793e-22, 1.72746e-22, 1.03648e-22, 3.4925442e-23], [32.885163, 3.21608e-22, 1.66268e-22, 9.97609e-23, 3.3635813e-23], [32.960971, 3.09649e-22, 1.60392e-22, 9.62354e-23, 3.2465787e-23], [33.036954, 2.98762e-22, 1.55042e-22, 9.30253e-23, 3.1400065e-23], [33.113112, 2.88819e-22, 1.50153e-22, 9.00918e-23, 3.0425762e-23], [33.189446, 2.79708e-22, 1.4567e-22, 8.7402e-23, 2.9531953e-23], [33.265955, 2.71336e-22, 1.41547e-22, 8.4928e-23, 2.8709318e-23], [33.342641, 2.6362e-22, 1.37742e-22, 8.26455e-23, 2.7949867e-23], [33.419504, 2.56493e-22, 1.34223e-22, 8.05338e-23, 2.7246708e-23], [33.496544, 2.49892e-22, 1.30958e-22, 7.85749e-23, 2.6593876e-23], [33.573761, 2.43764e-22, 1.27922e-22, 7.67532e-23, 2.598618e-23], [33.651157, 2.38064e-22, 1.25092e-22, 7.5055e-23, 2.5419085e-23], [33.728731, 2.32751e-22, 1.22447e-22, 7.34682e-23, 2.4888617e-23], [33.806484, 2.27789e-22, 1.1997e-22, 7.19822e-23, 2.4391279e-23], [33.884416, 2.23145e-22, 1.17646e-22, 7.05878e-23, 2.3923984e-23], [33.962527, 2.18792e-22, 1.15461e-22, 6.92765e-23, 2.3483997e-23], [34.040819, 2.14704e-22, 1.13402e-22, 6.80412e-23, 2.3068891e-23], [34.119291, 2.10859e-22, 1.11459e-22, 6.68752e-23, 2.2676503e-23], [34.197944, 2.07237e-22, 1.09621e-22, 6.57726e-23, 2.2304902e-23], [34.276779, 2.0382e-22, 1.0788e-22, 6.47282e-23, 2.195236e-23], [34.355795, 2.00592e-22, 1.06229e-22, 6.37372e-23, 2.1617325e-23], [34.434993, 1.97537e-22, 1.04659e-22, 6.27955e-23, 2.1298404e-23], [34.514374, 1.94643e-22, 1.03165e-22, 6.18992e-23, 2.0994338e-23], [34.593938, 1.91897e-22, 1.01742e-22, 6.10449e-23, 2.0703992e-23], [34.673685, 1.8929e-22, 1.00382e-22, 6.02294e-23, 2.0426341e-23], [34.753616, 1.8681e-22, 9.90831e-23, 5.94498e-23, 2.0160451e-23], [34.833732, 1.84448e-22, 9.78395e-23, 5.87037e-23, 1.9905477e-23], [34.914032, 1.82198e-22, 9.66476e-23, 5.79886e-23, 1.966065e-23], [34.994517, 1.8005e-22, 9.55039e-23, 5.73023e-23, 1.9425269e-23], [35.075187, 1.77998e-22, 9.44051e-23, 5.66431e-23, 1.9198692e-23], [35.156044, 1.76037e-22, 9.33482e-23, 5.60089e-23, 1.8980336e-23], [35.237087, 1.74159e-22, 9.23305e-23, 5.53983e-23, 1.8769665e-23], [35.318317, 1.7236e-22, 9.13495e-23, 5.48097e-23, 1.8566186e-23], [35.399734, 1.70635e-22, 9.04028e-23, 5.42417e-23, 1.8369449e-23], [35.481339, 1.68979e-22, 8.94884e-23, 5.36931e-23, 1.817904e-23], [35.563132, 1.67388e-22, 8.86043e-23, 5.31626e-23, 1.7994576e-23], [35.645113, 1.65859e-22, 8.77487e-23, 5.26492e-23, 1.7815705e-23], [35.727284, 1.64387e-22, 8.69199e-23, 5.21519e-23, 1.7642102e-23], [35.809644, 1.6297e-22, 8.61164e-23, 5.16698e-23, 1.7473468e-23], [35.892193, 1.61604e-22, 8.53367e-23, 5.1202e-23, 1.7309524e-23], [35.974934, 1.60286e-22, 8.45797e-23, 5.07478e-23, 1.7150013e-23], [36.057864, 1.59015e-22, 8.38439e-23, 5.03063e-23, 1.6994696e-23], [36.140986, 1.57787e-22, 8.31283e-23, 4.9877e-23, 1.6843352e-23], [36.2243, 1.566e-22, 8.24319e-23, 4.94591e-23, 1.6695775e-23], [36.307805, 1.55452e-22, 8.17536e-23, 4.90522e-23, 1.6551774e-23], [36.391504, 1.54341e-22, 8.10925e-23, 4.86555e-23, 1.6411169e-23], [36.475395, 1.53265e-22, 8.04479e-23, 4.82687e-23, 1.6273795e-23], [36.559479, 1.52223e-22, 7.98188e-23, 4.78913e-23, 1.6139497e-23], [36.643757, 1.51212e-22, 7.92045e-23, 4.75227e-23, 1.600813e-23], [36.72823, 1.50232e-22, 7.86044e-23, 4.71627e-23, 1.5879558e-23], [36.812897, 1.49281e-22, 7.80178e-23, 4.68107e-23, 1.5753654e-23], [36.89776, 1.48357e-22, 7.74441e-23, 4.64664e-23, 1.5630301e-23], [36.982818, 1.47459e-22, 7.68826e-23, 4.61296e-23, 1.5509386e-23], [37.068072, 1.46587e-22, 7.6333e-23, 4.57998e-23, 1.5390805e-23], [37.153523, 1.45738e-22, 7.57946e-23, 4.54767e-23, 1.5274461e-23], [37.239171, 1.44913e-22, 7.5267e-23, 4.51602e-23, 1.516026e-23], [37.325016, 1.44109e-22, 7.47497e-23, 4.48498e-23, 1.5048117e-23], [37.411059, 1.43326e-22, 7.42423e-23, 4.45454e-23, 1.493795e-23], [37.4973, 1.42564e-22, 7.37445e-23, 4.42467e-23, 1.4829683e-23], [37.58374, 1.41821e-22, 7.32559e-23, 4.39535e-23, 1.4723242e-23], [37.67038, 1.41096e-22, 7.2776e-23, 4.36656e-23, 1.4618559e-23], [37.757219, 1.4039e-22, 7.23046e-23, 4.33828e-23, 1.4515571e-23], [37.844258, 1.397e-22, 7.18414e-23, 4.31049e-23, 1.4414215e-23], [37.931498, 1.39027e-22, 7.13861e-23, 4.28316e-23, 1.4314435e-23], [38.01894, 1.3837e-22, 7.09383e-23, 4.2563e-23, 1.4216176e-23], [38.106582, 1.37728e-22, 7.04978e-23, 4.22987e-23, 1.4119386e-23], [38.194427, 1.371e-22, 7.00644e-23, 4.20386e-23, 1.4024017e-23], [38.282474, 1.36487e-22, 6.96378e-23, 4.17827e-23, 1.3930021e-23], [38.370725, 1.35888e-22, 6.92178e-23, 4.15307e-23, 1.3837356e-23], [38.459178, 1.35302e-22, 6.88042e-23, 4.12825e-23, 1.3745979e-23], [38.547836, 1.34728e-22, 6.83968e-23, 4.10381e-23, 1.365585e-23], [38.636698, 1.34167e-22, 6.79953e-23, 4.07972e-23, 1.3566933e-23], [38.725764, 1.33617e-22, 6.75996e-23, 4.05598e-23, 1.347919e-23], [38.815037, 1.33079e-22, 6.72096e-23, 4.03257e-23, 1.3392588e-23], [38.904514, 1.32552e-22, 6.6825e-23, 4.0095e-23, 1.3307094e-23], [38.994199, 1.32036e-22, 6.64457e-23, 3.98674e-23, 1.3222678e-23], [39.08409, 1.3153e-22, 6.60715e-23, 3.96429e-23, 1.313931e-23], [39.174188, 1.31035e-22, 6.57023e-23, 3.94214e-23, 1.3056961e-23], [39.264494, 1.30549e-22, 6.5338e-23, 3.92028e-23, 1.2975604e-23], [39.355008, 1.30072e-22, 6.49784e-23, 3.8987e-23, 1.2895215e-23], [39.44573, 1.29605e-22, 6.46234e-23, 3.8774e-23, 1.2815767e-23], [39.536662, 1.29147e-22, 6.42728e-23, 3.85637e-23, 1.2737239e-23], [39.627803, 1.28697e-22, 6.39266e-23, 3.8356e-23, 1.2659607e-23], [39.719155, 1.28256e-22, 6.35847e-23, 3.81508e-23, 1.2582849e-23], [39.810717, 1.27822e-22, 6.32469e-23, 3.79481e-23, 1.2506946e-23], [39.90249, 1.27397e-22, 6.29131e-23, 3.77479e-23, 1.2431877e-23], [39.994475, 1.26979e-22, 6.25833e-23, 3.755e-23, 1.2357624e-23], [40.086672, 1.26569e-22, 6.22573e-23, 3.73544e-23, 1.2284167e-23], [40.179081, 1.26167e-22, 6.19351e-23, 3.71611e-23, 1.2211491e-23], [40.271703, 1.25771e-22, 6.16165e-23, 3.69699e-23, 1.2139577e-23], [40.364539, 1.25382e-22, 6.13016e-23, 3.67809e-23, 1.206841e-23], [40.457589, 1.25e-22, 6.09901e-23, 3.65941e-23, 1.1997975e-23], [40.550854, 1.24624e-22, 6.06821e-23, 3.64092e-23, 1.1928256e-23], [40.644333, 1.24255e-22, 6.03774e-23, 3.62264e-23, 1.1859239e-23], [40.738028, 1.23892e-22, 6.0076e-23, 3.60456e-23, 1.1790911e-23], [40.831939, 1.23535e-22, 5.97778e-23, 3.58667e-23, 1.1723258e-23], [40.926066, 1.23184e-22, 5.94827e-23, 3.56896e-23, 1.1656266e-23], [41.02041, 1.22839e-22, 5.91908e-23, 3.55145e-23, 1.1589925e-23], [41.114972, 1.225e-22, 5.89019e-23, 3.53411e-23, 1.1524222e-23], [41.209752, 1.22166e-22, 5.86159e-23, 3.51695e-23, 1.1459145e-23], [41.30475, 1.21837e-22, 5.83328e-23, 3.49997e-23, 1.1394684e-23], [41.399967, 1.21513e-22, 5.80526e-23, 3.48316e-23, 1.1330827e-23], [41.495404, 1.21195e-22, 5.77752e-23, 3.46651e-23, 1.1267565e-23], [41.591061, 1.20882e-22, 5.75005e-23, 3.45003e-23, 1.1204887e-23], [41.686938, 1.20573e-22, 5.72286e-23, 3.43371e-23, 1.1142783e-23], [41.783037, 1.2027e-22, 5.69593e-23, 3.41756e-23, 1.1081245e-23], [41.879357, 1.19971e-22, 5.66925e-23, 3.40155e-23, 1.1020263e-23], [41.975898, 1.19676e-22, 5.64284e-23, 3.3857e-23, 1.0959829e-23], [42.072663, 1.19386e-22, 5.61667e-23, 3.37e-23, 1.0899933e-23], [42.16965, 1.19101e-22, 5.59076e-23, 3.35445e-23, 1.0840569e-23], [42.266861, 1.1882e-22, 5.56508e-23, 3.33905e-23, 1.0781726e-23], [42.364297, 1.18543e-22, 5.53965e-23, 3.32379e-23, 1.0723399e-23], [42.461956, 1.1827e-22, 5.51444e-23, 3.30867e-23, 1.066558e-23], [42.559841, 1.18001e-22, 5.48948e-23, 3.29369e-23, 1.060826e-23], [42.657952, 1.17736e-22, 5.46473e-23, 3.27884e-23, 1.0551434e-23], [42.756289, 1.17475e-22, 5.44022e-23, 3.26413e-23, 1.0495094e-23], [42.854852, 1.17217e-22, 5.41592e-23, 3.24955e-23, 1.0439233e-23], [42.953643, 1.16964e-22, 5.39184e-23, 3.2351e-23, 1.0383846e-23], [43.052661, 1.16714e-22, 5.36798e-23, 3.22079e-23, 1.0328925e-23], [43.151908, 1.16467e-22, 5.34432e-23, 3.20659e-23, 1.0274465e-23], [43.251383, 1.16224e-22, 5.32088e-23, 3.19253e-23, 1.022046e-23], [43.351088, 1.15985e-22, 5.29763e-23, 3.17858e-23, 1.0166903e-23], [43.451022, 1.15749e-22, 5.27459e-23, 3.16476e-23, 1.011379e-23], [43.551187, 1.15516e-22, 5.25175e-23, 3.15105e-23, 1.0061114e-23], [43.651583, 1.15286e-22, 5.22911e-23, 3.13747e-23, 1.0008871e-23], [43.752211, 1.1506e-22, 5.20666e-23, 3.124e-23, 9.9570548e-24], [43.85307, 1.14837e-22, 5.1844e-23, 3.11064e-23, 9.9056607e-24], [43.954162, 1.14617e-22, 5.16233e-23, 3.0974e-23, 9.8546835e-24], [44.055486, 1.14399e-22, 5.14045e-23, 3.08427e-23, 9.8041186e-24], [44.157045, 1.14185e-22, 5.11875e-23, 3.07125e-23, 9.7539611e-24], [44.258837, 1.13974e-22, 5.09723e-23, 3.05834e-23, 9.7042064e-24], [44.360864, 1.13765e-22, 5.07589e-23, 3.04554e-23, 9.6548501e-24], [44.463127, 1.1356e-22, 5.05473e-23, 3.03284e-23, 9.6058877e-24], [44.565625, 1.13357e-22, 5.03374e-23, 3.02024e-23, 9.5573149e-24], [44.668359, 1.13156e-22, 5.01293e-23, 3.00776e-23, 9.5091276e-24], [44.77133, 1.12959e-22, 4.99228e-23, 2.99537e-23, 9.4613216e-24], [44.874539, 1.12764e-22, 4.9718e-23, 2.98308e-23, 9.4138928e-24], [44.977985, 1.12571e-22, 4.95149e-23, 2.9709e-23, 9.3668375e-24], [45.08167, 1.12382e-22, 4.93135e-23, 2.95881e-23, 9.3201517e-24], [45.185594, 1.12194e-22, 4.91136e-23, 2.94682e-23, 9.2738317e-24], [45.289758, 1.12009e-22, 4.89154e-23, 2.93492e-23, 9.2278738e-24], [45.394162, 1.11827e-22, 4.87187e-23, 2.92312e-23, 9.1822745e-24], [45.498806, 1.11647e-22, 4.85236e-23, 2.91142e-23, 9.13703e-24], [45.603692, 1.11469e-22, 4.83301e-23, 2.89981e-23, 9.0921371e-24], [45.708819, 1.11293e-22, 4.81381e-23, 2.88829e-23, 9.0475924e-24], [45.814189, 1.1112e-22, 4.79476e-23, 2.87686e-23, 9.0033924e-24], [45.919801, 1.10949e-22, 4.77587e-23, 2.86552e-23, 8.959534e-24], [46.025657, 1.1078e-22, 4.75712e-23, 2.85427e-23, 8.916014e-24], [46.131757, 1.10613e-22, 4.73851e-23, 2.84311e-23, 8.8728293e-24], [46.238102, 1.10448e-22, 4.72006e-23, 2.83203e-23, 8.8299767e-24], [46.344692, 1.10286e-22, 4.70174e-23, 2.82105e-23, 8.7874533e-24], [46.451528, 1.10125e-22, 4.68357e-23, 2.81014e-23, 8.7452562e-24], [46.558609, 1.09966e-22, 4.66554e-23, 2.79932e-23, 8.7033824e-24], [46.665938, 1.0981e-22, 4.64765e-23, 2.78859e-23, 8.661829e-24], [46.773514, 1.09655e-22, 4.6299e-23, 2.77794e-23, 8.6205933e-24], [46.881338, 1.09502e-22, 4.61228e-23, 2.76737e-23, 8.5796726e-24], [46.989411, 1.09352e-22, 4.5948e-23, 2.75688e-23, 8.5390641e-24], [47.097733, 1.09203e-22, 4.57745e-23, 2.74647e-23, 8.4987652e-24], [47.206304, 1.09056e-22, 4.56023e-23, 2.73614e-23, 8.4587733e-24], [47.315126, 1.0891e-22, 4.54315e-23, 2.72589e-23, 8.4190859e-24], [47.424199, 1.08767e-22, 4.52619e-23, 2.71572e-23, 8.3797003e-24], [47.533523, 1.08625e-22, 4.50937e-23, 2.70562e-23, 8.3406141e-24], [47.643099, 1.08485e-22, 4.49267e-23, 2.6956e-23, 8.301825e-24], [47.752927, 1.08346e-22, 4.4761e-23, 2.68566e-23, 8.2633305e-24], [47.863009, 1.0821e-22, 4.45965e-23, 2.67579e-23, 8.2251282e-24], [47.973345, 1.08075e-22, 4.44333e-23, 2.666e-23, 8.1872158e-24], [48.083935, 1.07941e-22, 4.42712e-23, 2.65627e-23, 8.149591e-24], [48.19478, 1.0781e-22, 4.41104e-23, 2.64663e-23, 8.1122516e-24], [48.30588, 1.07679e-22, 4.39509e-23, 2.63705e-23, 8.0751954e-24], [48.417237, 1.07551e-22, 4.37925e-23, 2.62755e-23, 8.0384202e-24], [48.52885, 1.07424e-22, 4.36352e-23, 2.61811e-23, 8.0019238e-24], [48.640721, 1.07298e-22, 4.34792e-23, 2.60875e-23, 7.9657041e-24], [48.752849, 1.07174e-22, 4.33243e-23, 2.59946e-23, 7.929759e-24], [48.865236, 1.07051e-22, 4.31706e-23, 2.59024e-23, 7.8940865e-24], [48.977882, 1.0693e-22, 4.3018e-23, 2.58108e-23, 7.8586845e-24], [49.090788, 1.0681e-22, 4.28665e-23, 2.57199e-23, 7.823551e-24], [49.203954, 1.06692e-22, 4.27162e-23, 2.56297e-23, 7.7886841e-24], [49.31738, 1.06575e-22, 4.2567e-23, 2.55402e-23, 7.7540818e-24], [49.431069, 1.0646e-22, 4.24188e-23, 2.54513e-23, 7.7197422e-24], [49.545019, 1.06345e-22, 4.22718e-23, 2.53631e-23, 7.6856633e-24], [49.659232, 1.06232e-22, 4.21259e-23, 2.52755e-23, 7.6518434e-24], [49.773708, 1.06121e-22, 4.1981e-23, 2.51886e-23, 7.6182806e-24], [49.888449, 1.06011e-22, 4.18372e-23, 2.51023e-23, 7.584973e-24], [50.003453, 1.05902e-22, 4.16944e-23, 2.50166e-23, 7.5519188e-24], [50.118723, 1.05794e-22, 4.15527e-23, 2.49316e-23, 7.5191163e-24], [50.234259, 1.05688e-22, 4.1412e-23, 2.48472e-23, 7.4865637e-24], [50.350061, 1.05582e-22, 4.12724e-23, 2.47634e-23, 7.4542593e-24], [50.46613, 1.05478e-22, 4.11338e-23, 2.46803e-23, 7.4222013e-24], [50.582466, 1.05376e-22, 4.09961e-23, 2.45977e-23, 7.3903881e-24], [50.699071, 1.05274e-22, 4.08595e-23, 2.45157e-23, 7.358818e-24], [50.815944, 1.05174e-22, 4.07239e-23, 2.44344e-23, 7.3274894e-24], [50.933087, 1.05074e-22, 4.05893e-23, 2.43536e-23, 7.2964005e-24], [51.0505, 1.04976e-22, 4.04557e-23, 2.42734e-23, 7.2655498e-24], [51.168184, 1.04879e-22, 4.0323e-23, 2.41938e-23, 7.2349357e-24], [51.286138, 1.04783e-22, 4.01913e-23, 2.41148e-23, 7.2045566e-24], [51.404365, 1.04688e-22, 4.00605e-23, 2.40363e-23, 7.1744109e-24], [51.522864, 1.04595e-22, 3.99307e-23, 2.39584e-23, 7.1444971e-24], [51.641637, 1.04502e-22, 3.98019e-23, 2.38811e-23, 7.1148137e-24], [51.760683, 1.0441e-22, 3.9674e-23, 2.38044e-23, 7.085359e-24], [51.880004, 1.0432e-22, 3.9547e-23, 2.37282e-23, 7.0561318e-24], [51.9996, 1.0423e-22, 3.94209e-23, 2.36525e-23, 7.0271303e-24], [52.119471, 1.04142e-22, 3.92957e-23, 2.35774e-23, 6.9983533e-24], [52.239619, 1.04054e-22, 3.91715e-23, 2.35029e-23, 6.9697991e-24], [52.360044, 1.03968e-22, 3.90481e-23, 2.34289e-23, 6.9414665e-24], [52.480746, 1.03882e-22, 3.89256e-23, 2.33554e-23, 6.9133538e-24], [52.601727, 1.03798e-22, 3.88041e-23, 2.32824e-23, 6.8854598e-24], [52.722986, 1.03714e-22, 3.86834e-23, 2.321e-23, 6.8577831e-24], [52.844525, 1.03631e-22, 3.85635e-23, 2.31381e-23, 6.8303221e-24], [52.966344, 1.03549e-22, 3.84446e-23, 2.30667e-23, 6.8030757e-24], [53.088444, 1.03469e-22, 3.83264e-23, 2.29959e-23, 6.7760423e-24], [53.210826, 1.03389e-22, 3.82092e-23, 2.29255e-23, 6.7492207e-24], [53.33349, 1.0331e-22, 3.80928e-23, 2.28557e-23, 6.7226095e-24], [53.456436, 1.03231e-22, 3.79772e-23, 2.27863e-23, 6.6962073e-24], [53.579666, 1.03154e-22, 3.78625e-23, 2.27175e-23, 6.6700129e-24], [53.70318, 1.03078e-22, 3.77485e-23, 2.26491e-23, 6.644025e-24], [53.826978, 1.03002e-22, 3.76355e-23, 2.25813e-23, 6.6182422e-24], [53.951062, 1.02927e-22, 3.75232e-23, 2.25139e-23, 6.5926633e-24], [54.075432, 1.02853e-22, 3.74117e-23, 2.2447e-23, 6.567287e-24], [54.200089, 1.0278e-22, 3.7301e-23, 2.23806e-23, 6.542112e-24], [54.325033, 1.02708e-22, 3.71912e-23, 2.23147e-23, 6.517137e-24], [54.450265, 1.02637e-22, 3.70821e-23, 2.22493e-23, 6.4923609e-24], [54.575786, 1.02566e-22, 3.69738e-23, 2.21843e-23, 6.4677824e-24], [54.701596, 1.02496e-22, 3.68663e-23, 2.21198e-23, 6.4434003e-24], [54.827696, 1.02427e-22, 3.67595e-23, 2.20557e-23, 6.4192132e-24], [54.954087, 1.02359e-22, 3.66536e-23, 2.19921e-23, 6.3952201e-24], [55.08077, 1.02291e-22, 3.65484e-23, 2.1929e-23, 6.3714198e-24], [55.207744, 1.02224e-22, 3.64439e-23, 2.18663e-23, 6.3478109e-24], [55.335011, 1.02158e-22, 3.63402e-23, 2.18041e-23, 6.3243925e-24], [55.462571, 1.02093e-22, 3.62373e-23, 2.17424e-23, 6.3011632e-24], [55.590426, 1.02028e-22, 3.6135e-23, 2.1681e-23, 6.2781219e-24], [55.718575, 1.01964e-22, 3.60336e-23, 2.16201e-23, 6.2552674e-24], [55.847019, 1.01901e-22, 3.59328e-23, 2.15597e-23, 6.2325987e-24], [55.97576, 1.01838e-22, 3.58328e-23, 2.14997e-23, 6.2101144e-24], [56.104798, 1.01776e-22, 3.57335e-23, 2.14401e-23, 6.1878136e-24], [56.234133, 1.01715e-22, 3.56349e-23, 2.13809e-23, 6.1656951e-24], [56.363766, 1.01655e-22, 3.5537e-23, 2.13222e-23, 6.1437577e-24], [56.493697, 1.01595e-22, 3.54399e-23, 2.12639e-23, 6.1220003e-24], [56.623929, 1.01536e-22, 3.53434e-23, 2.1206e-23, 6.1004219e-24], [56.754461, 1.01477e-22, 3.52476e-23, 2.11486e-23, 6.0790213e-24], [56.885293, 1.01419e-22, 3.51525e-23, 2.10915e-23, 6.0577974e-24], [57.016427, 1.01362e-22, 3.50581e-23, 2.10349e-23, 6.0367491e-24], [57.147864, 1.01305e-22, 3.49644e-23, 2.09786e-23, 6.0158754e-24], [57.279603, 1.01249e-22, 3.48713e-23, 2.09228e-23, 5.9951751e-24], [57.411646, 1.01194e-22, 3.4779e-23, 2.08674e-23, 5.9746472e-24], [57.543994, 1.01139e-22, 3.46872e-23, 2.08123e-23, 5.9542906e-24], [57.676646, 1.01085e-22, 3.45962e-23, 2.07577e-23, 5.9341043e-24], [57.809605, 1.01031e-22, 3.45058e-23, 2.07035e-23, 5.9140871e-24], [57.94287, 1.00978e-22, 3.44161e-23, 2.06496e-23, 5.8942381e-24], [58.076442, 1.00926e-22, 3.4327e-23, 2.05962e-23, 5.8745562e-24], [58.210322, 1.00874e-22, 3.42385e-23, 2.05431e-23, 5.8550404e-24], [58.34451, 1.00823e-22, 3.41507e-23, 2.04904e-23, 5.8356896e-24], [58.479008, 1.00772e-22, 3.40635e-23, 2.04381e-23, 5.8165028e-24], [58.613816, 1.00722e-22, 3.3977e-23, 2.03862e-23, 5.7974789e-24], [58.748935, 1.00672e-22, 3.38911e-23, 2.03346e-23, 5.778617e-24], [58.884366, 1.00623e-22, 3.38058e-23, 2.02835e-23, 5.759916e-24], [59.020108, 1.00575e-22, 3.37211e-23, 2.02326e-23, 5.7413749e-24], [59.156163, 1.00527e-22, 3.3637e-23, 2.01822e-23, 5.7229928e-24], [59.292532, 1.00479e-22, 3.35535e-23, 2.01321e-23, 5.7047685e-24], [59.429216, 1.00432e-22, 3.34707e-23, 2.00824e-23, 5.6867012e-24], [59.566214, 1.00386e-22, 3.33884e-23, 2.00331e-23, 5.6687898e-24], [59.703529, 1.0034e-22, 3.33068e-23, 1.99841e-23, 5.6510333e-24], [59.84116, 1.00295e-22, 3.32257e-23, 1.99354e-23, 5.6334307e-24], [59.979108, 1.0025e-22, 3.31452e-23, 1.98871e-23, 5.6159811e-24], [60.117374, 1.00205e-22, 3.30653e-23, 1.98392e-23, 5.5986835e-24], [60.255959, 1.00161e-22, 3.2986e-23, 1.97916e-23, 5.5815369e-24], [60.394863, 1.00118e-22, 3.29072e-23, 1.97443e-23, 5.5645404e-24], [60.534087, 1.00075e-22, 3.28291e-23, 1.96974e-23, 5.5476929e-24], [60.673633, 1.00032e-22, 3.27515e-23, 1.96509e-23, 5.5309936e-24], [60.8135, 9.99904e-23, 3.26744e-23, 1.96046e-23, 5.5144414e-24], [60.95369, 9.99489e-23, 3.25979e-23, 1.95588e-23, 5.4980354e-24], [61.094202, 9.99077e-23, 3.2522e-23, 1.95132e-23, 5.4817747e-24], [61.235039, 9.98671e-23, 3.24466e-23, 1.9468e-23, 5.4656583e-24], [61.376201, 9.98269e-23, 3.23718e-23, 1.94231e-23, 5.4496853e-24], [61.517687, 9.97871e-23, 3.22975e-23, 1.93785e-23, 5.4338547e-24], [61.6595, 9.97478e-23, 3.22238e-23, 1.93343e-23, 5.4181657e-24], [61.80164, 9.97089e-23, 3.21506e-23, 1.92904e-23, 5.4026172e-24], [61.944108, 9.96704e-23, 3.20779e-23, 1.92467e-23, 5.3872083e-24], [62.086903, 9.96324e-23, 3.20058e-23, 1.92035e-23, 5.3719382e-24], [62.230029, 9.95948e-23, 3.19342e-23, 1.91605e-23, 5.3568059e-24], [62.373484, 9.95576e-23, 3.18631e-23, 1.91178e-23, 5.3418104e-24], [62.517269, 9.95208e-23, 3.17925e-23, 1.90755e-23, 5.3269509e-24], [62.661386, 9.94844e-23, 3.17225e-23, 1.90335e-23, 5.3122265e-24], [62.805836, 9.94485e-23, 3.16529e-23, 1.89917e-23, 5.2976362e-24], [62.950618, 9.94129e-23, 3.15839e-23, 1.89503e-23, 5.2831792e-24], [63.095734, 9.93777e-23, 3.15154e-23, 1.89092e-23, 5.2688545e-24], [63.241185, 9.93429e-23, 3.14473e-23, 1.88684e-23, 5.2546613e-24], [63.386971, 9.93086e-23, 3.13798e-23, 1.88279e-23, 5.2405986e-24], [63.533093, 9.92745e-23, 3.13128e-23, 1.87877e-23, 5.2266656e-24], [63.679552, 9.92409e-23, 3.12462e-23, 1.87477e-23, 5.2128614e-24], [63.826349, 9.92077e-23, 3.11802e-23, 1.87081e-23, 5.199185e-24], [63.973484, 9.91748e-23, 3.11146e-23, 1.86688e-23, 5.1856357e-24], [64.120958, 9.91423e-23, 3.10495e-23, 1.86297e-23, 5.1722125e-24], [64.268772, 9.91102e-23, 3.09849e-23, 1.85909e-23, 5.1589145e-24], [64.416927, 9.90784e-23, 3.09208e-23, 1.85525e-23, 5.1457409e-24], [64.565423, 9.9047e-23, 3.08571e-23, 1.85143e-23, 5.1326908e-24], [64.714262, 9.9016e-23, 3.07939e-23, 1.84763e-23, 5.1197634e-24], [64.863443, 9.89852e-23, 3.07312e-23, 1.84387e-23, 5.1069577e-24], [65.012969, 9.89549e-23, 3.06689e-23, 1.84014e-23, 5.094273e-24], [65.162839, 9.89249e-23, 3.06071e-23, 1.83643e-23, 5.0817083e-24], [65.313055, 9.88952e-23, 3.05458e-23, 1.83275e-23, 5.0692628e-24], [65.463617, 9.88659e-23, 3.04849e-23, 1.82909e-23, 5.0569357e-24], [65.614527, 9.88369e-23, 3.04244e-23, 1.82546e-23, 5.0447261e-24], [65.765784, 9.88083e-23, 3.03644e-23, 1.82186e-23, 5.0326331e-24], [65.91739, 9.87799e-23, 3.03048e-23, 1.81829e-23, 5.020656e-24], [66.069345, 9.87519e-23, 3.02457e-23, 1.81474e-23, 5.0087938e-24], [66.22165, 9.87243e-23, 3.0187e-23, 1.81122e-23, 4.9970458e-24], [66.374307, 9.86969e-23, 3.01288e-23, 1.80773e-23, 4.9854111e-24], [66.527316, 9.86699e-23, 3.0071e-23, 1.80426e-23, 4.9738888e-24], [66.680677, 9.86432e-23, 3.00136e-23, 1.80082e-23, 4.9624782e-24], [66.834392, 9.86167e-23, 2.99566e-23, 1.7974e-23, 4.9511784e-24], [66.988461, 9.85906e-23, 2.99001e-23, 1.79401e-23, 4.9399886e-24], [67.142885, 9.85648e-23, 2.9844e-23, 1.79064e-23, 4.928908e-24], [67.297666, 9.85394e-23, 2.97883e-23, 1.7873e-23, 4.9179358e-24], [67.452803, 9.85142e-23, 2.9733e-23, 1.78398e-23, 4.9070711e-24], [67.608298, 9.84893e-23, 2.96781e-23, 1.78069e-23, 4.8963131e-24], [67.764151, 9.84647e-23, 2.96236e-23, 1.77742e-23, 4.8856611e-24], [67.920363, 9.84404e-23, 2.95696e-23, 1.77417e-23, 4.8751142e-24], [68.076936, 9.84164e-23, 2.95159e-23, 1.77095e-23, 4.8646717e-24], [68.233869, 9.83926e-23, 2.94626e-23, 1.76776e-23, 4.8543326e-24], [68.391165, 9.83692e-23, 2.94098e-23, 1.76459e-23, 4.8440963e-24], [68.548823, 9.8346e-23, 2.93573e-23, 1.76144e-23, 4.833962e-24], [68.706844, 9.83231e-23, 2.93052e-23, 1.75831e-23, 4.8239287e-24], [68.86523, 9.83005e-23, 2.92535e-23, 1.75521e-23, 4.8139959e-24], [69.02398, 9.82782e-23, 2.92022e-23, 1.75213e-23, 4.8041626e-24], [69.183097, 9.82561e-23, 2.91513e-23, 1.74908e-23, 4.7944282e-24], [69.342581, 9.82343e-23, 2.91008e-23, 1.74605e-23, 4.7847917e-24], [69.502432, 9.82128e-23, 2.90506e-23, 1.74304e-23, 4.7752526e-24], [69.662651, 9.81916e-23, 2.90008e-23, 1.74005e-23, 4.7658099e-24], [69.82324, 9.81706e-23, 2.89514e-23, 1.73709e-23, 4.7564629e-24], [69.9842, 9.81498e-23, 2.89024e-23, 1.73414e-23, 4.7472108e-24], [70.14553, 9.81294e-23, 2.88537e-23, 1.73122e-23, 4.738053e-24], [70.307232, 9.81092e-23, 2.88054e-23, 1.72832e-23, 4.7289885e-24], [70.469307, 9.80892e-23, 2.87575e-23, 1.72545e-23, 4.7200168e-24], [70.631755, 9.80695e-23, 2.87099e-23, 1.72259e-23, 4.711137e-24], [70.794578, 9.805e-23, 2.86627e-23, 1.71976e-23, 4.7023483e-24], [70.957777, 9.80308e-23, 2.86158e-23, 1.71695e-23, 4.6936501e-24], [71.121351, 9.80118e-23, 2.85693e-23, 1.71416e-23, 4.6850415e-24], [71.285303, 9.79931e-23, 2.85231e-23, 1.71139e-23, 4.6765219e-24], [71.449633, 9.79746e-23, 2.84773e-23, 1.70864e-23, 4.6680906e-24], [71.614341, 9.79564e-23, 2.84318e-23, 1.70591e-23, 4.6597466e-24], [71.779429, 9.79384e-23, 2.83867e-23, 1.7032e-23, 4.6514895e-24], [71.944898, 9.79206e-23, 2.83419e-23, 1.70051e-23, 4.6433183e-24], [72.110748, 9.79031e-23, 2.82975e-23, 1.69785e-23, 4.6352325e-24], [72.27698, 9.78858e-23, 2.82534e-23, 1.6952e-23, 4.6272312e-24], [72.443596, 9.78687e-23, 2.82096e-23, 1.69257e-23, 4.6193138e-24], [72.610596, 9.78519e-23, 2.81661e-23, 1.68997e-23, 4.6114796e-24], [72.77798, 9.78352e-23, 2.8123e-23, 1.68738e-23, 4.6037278e-24], [72.945751, 9.78188e-23, 2.80802e-23, 1.68481e-23, 4.5960577e-24], [73.113908, 9.78027e-23, 2.80378e-23, 1.68227e-23, 4.5884687e-24], [73.282453, 9.77867e-23, 2.79956e-23, 1.67974e-23, 4.58096e-24], [73.451387, 9.7771e-23, 2.79538e-23, 1.67723e-23, 4.5735309e-24], [73.62071, 9.77555e-23, 2.79123e-23, 1.67474e-23, 4.5661808e-24], [73.790423, 9.77402e-23, 2.78711e-23, 1.67227e-23, 4.5589089e-24], [73.960528, 9.77251e-23, 2.78302e-23, 1.66981e-23, 4.5517146e-24], [74.131024, 9.77102e-23, 2.77897e-23, 1.66738e-23, 4.5445972e-24], [74.301914, 9.76956e-23, 2.77494e-23, 1.66497e-23, 4.537556e-24], [74.473197, 9.76811e-23, 2.77095e-23, 1.66257e-23, 4.5305904e-24], [74.644876, 9.76669e-23, 2.76698e-23, 1.66019e-23, 4.5236996e-24], [74.81695, 9.76528e-23, 2.76305e-23, 1.65783e-23, 4.516883e-24], [74.989421, 9.7639e-23, 2.75915e-23, 1.65549e-23, 4.5101399e-24], [75.162289, 9.76254e-23, 2.75527e-23, 1.65316e-23, 4.5034697e-24], [75.335556, 9.7612e-23, 2.75143e-23, 1.65086e-23, 4.4968717e-24], [75.509223, 9.75987e-23, 2.74761e-23, 1.64857e-23, 4.4903453e-24], [75.68329, 9.75857e-23, 2.74383e-23, 1.6463e-23, 4.4838898e-24], [75.857758, 9.75729e-23, 2.74007e-23, 1.64404e-23, 4.4775045e-24], [76.032628, 9.75603e-23, 2.73635e-23, 1.64181e-23, 4.4711889e-24], [76.207901, 9.75478e-23, 2.73265e-23, 1.63959e-23, 4.4649422e-24], [76.383578, 9.75356e-23, 2.72898e-23, 1.63739e-23, 4.4587639e-24], [76.559661, 9.75235e-23, 2.72533e-23, 1.6352e-23, 4.4526533e-24], [76.736149, 9.75117e-23, 2.72172e-23, 1.63303e-23, 4.4466098e-24], [76.913044, 9.75e-23, 2.71813e-23, 1.63088e-23, 4.4406327e-24], [77.090347, 9.74885e-23, 2.71458e-23, 1.62875e-23, 4.4347215e-24], [77.268059, 9.74772e-23, 2.71105e-23, 1.62663e-23, 4.4288755e-24], [77.44618, 9.74661e-23, 2.70754e-23, 1.62452e-23, 4.4230941e-24], [77.624712, 9.74552e-23, 2.70406e-23, 1.62244e-23, 4.4173767e-24], [77.803655, 9.74444e-23, 2.70062e-23, 1.62037e-23, 4.4117226e-24], [77.983011, 9.74339e-23, 2.69719e-23, 1.61832e-23, 4.4061314e-24], [78.16278, 9.74235e-23, 2.6938e-23, 1.61628e-23, 4.4006024e-24], [78.342964, 9.74133e-23, 2.69043e-23, 1.61426e-23, 4.3951349e-24], [78.523563, 9.74033e-23, 2.68708e-23, 1.61225e-23, 4.3897285e-24], [78.704579, 9.73934e-23, 2.68376e-23, 1.61026e-23, 4.3843825e-24], [78.886012, 9.73837e-23, 2.68047e-23, 1.60828e-23, 4.3790963e-24], [79.067863, 9.73742e-23, 2.67721e-23, 1.60632e-23, 4.3738693e-24], [79.250133, 9.73649e-23, 2.67396e-23, 1.60438e-23, 4.368701e-24], [79.432823, 9.73558e-23, 2.67075e-23, 1.60245e-23, 4.3635909e-24], [79.615935, 9.73468e-23, 2.66756e-23, 1.60053e-23, 4.3585382e-24], [79.799469, 9.7338e-23, 2.66439e-23, 1.59863e-23, 4.3535426e-24], [79.983426, 9.73293e-23, 2.66125e-23, 1.59675e-23, 4.3486033e-24], [80.167806, 9.73209e-23, 2.65813e-23, 1.59488e-23, 4.3437199e-24], [80.352612, 9.73126e-23, 2.65504e-23, 1.59303e-23, 4.3388918e-24], [80.537844, 9.73044e-23, 2.65197e-23, 1.59118e-23, 4.3341185e-24], [80.723503, 9.72965e-23, 2.64893e-23, 1.58936e-23, 4.3293993e-24], [80.90959, 9.72887e-23, 2.64591e-23, 1.58755e-23, 4.3247338e-24], [81.096106, 9.7281e-23, 2.64291e-23, 1.58575e-23, 4.3201215e-24], [81.283052, 9.72735e-23, 2.63994e-23, 1.58396e-23, 4.3155617e-24], [81.470428, 9.72662e-23, 2.63699e-23, 1.5822e-23, 4.311054e-24], [81.658237, 9.7259e-23, 2.63407e-23, 1.58044e-23, 4.3065978e-24], [81.846479, 9.7252e-23, 2.63116e-23, 1.5787e-23, 4.3021926e-24], [82.035154, 9.72452e-23, 2.62828e-23, 1.57697e-23, 4.297838e-24], [82.224265, 9.72385e-23, 2.62543e-23, 1.57526e-23, 4.2935333e-24], [82.413812, 9.7232e-23, 2.62259e-23, 1.57356e-23, 4.289278e-24], [82.603795, 9.72256e-23, 2.61978e-23, 1.57187e-23, 4.2850718e-24], [82.794216, 9.72194e-23, 2.61699e-23, 1.57019e-23, 4.2809139e-24], [82.985077, 9.72134e-23, 2.61422e-23, 1.56853e-23, 4.2768041e-24], [83.176377, 9.72074e-23, 2.61148e-23, 1.56689e-23, 4.2727417e-24], [83.368118, 9.72017e-23, 2.60875e-23, 1.56525e-23, 4.2687263e-24], [83.560302, 9.71961e-23, 2.60605e-23, 1.56363e-23, 4.2647573e-24], [83.752928, 9.71906e-23, 2.60337e-23, 1.56202e-23, 4.2608344e-24], [83.945999, 9.71853e-23, 2.60071e-23, 1.56043e-23, 4.256957e-24], [84.139514, 9.71802e-23, 2.59807e-23, 1.55884e-23, 4.2531246e-24], [84.333476, 9.71752e-23, 2.59546e-23, 1.55727e-23, 4.2493368e-24], [84.527885, 9.71703e-23, 2.59286e-23, 1.55572e-23, 4.245593e-24], [84.722741, 9.71656e-23, 2.59028e-23, 1.55417e-23, 4.241893e-24], [84.918048, 9.71611e-23, 2.58773e-23, 1.55264e-23, 4.2382361e-24], [85.113804, 9.71567e-23, 2.5852e-23, 1.55112e-23, 4.2346219e-24], [85.310011, 9.71524e-23, 2.58268e-23, 1.54961e-23, 4.23105e-24], [85.506671, 9.71483e-23, 2.58019e-23, 1.54811e-23, 4.2275199e-24], [85.703785, 9.71443e-23, 2.57771e-23, 1.54663e-23, 4.2240312e-24], [85.901352, 9.71405e-23, 2.57526e-23, 1.54516e-23, 4.2205834e-24], [86.099375, 9.71368e-23, 2.57283e-23, 1.5437e-23, 4.2171761e-24], [86.297855, 9.71332e-23, 2.57041e-23, 1.54225e-23, 4.2138089e-24], [86.496792, 9.71298e-23, 2.56802e-23, 1.54081e-23, 4.2104813e-24], [86.696188, 9.71266e-23, 2.56565e-23, 1.53939e-23, 4.2071929e-24], [86.896043, 9.71234e-23, 2.56329e-23, 1.53797e-23, 4.2039432e-24], [87.096359, 9.71204e-23, 2.56095e-23, 1.53657e-23, 4.200732e-24], [87.297137, 9.71176e-23, 2.55864e-23, 1.53518e-23, 4.1975587e-24], [87.498378, 9.71149e-23, 2.55634e-23, 1.5338e-23, 4.1944229e-24], [87.700082, 9.71123e-23, 2.55406e-23, 1.53243e-23, 4.1913243e-24], [87.902252, 9.71099e-23, 2.5518e-23, 1.53108e-23, 4.1882624e-24], [88.104887, 9.71076e-23, 2.54955e-23, 1.52973e-23, 4.1852368e-24], [88.30799, 9.71054e-23, 2.54733e-23, 1.5284e-23, 4.1822472e-24], [88.511561, 9.71034e-23, 2.54512e-23, 1.52707e-23, 4.1792931e-24], [88.715601, 9.71015e-23, 2.54294e-23, 1.52576e-23, 4.1763741e-24], [88.920112, 9.70998e-23, 2.54077e-23, 1.52446e-23, 4.17349e-24], [89.125094, 9.70982e-23, 2.53862e-23, 1.52317e-23, 4.1706402e-24], [89.330548, 9.70967e-23, 2.53648e-23, 1.52189e-23, 4.1678244e-24], [89.536477, 9.70953e-23, 2.53437e-23, 1.52062e-23, 4.1650423e-24], [89.742879, 9.70941e-23, 2.53227e-23, 1.51936e-23, 4.1622934e-24], [89.949758, 9.7093e-23, 2.53019e-23, 1.51811e-23, 4.1595774e-24], [90.157114, 9.70921e-23, 2.52812e-23, 1.51687e-23, 4.156894e-24], [90.364947, 9.70912e-23, 2.52608e-23, 1.51565e-23, 4.1542428e-24], [90.57326, 9.70906e-23, 2.52405e-23, 1.51443e-23, 4.1516234e-24], [90.782053, 9.709e-23, 2.52203e-23, 1.51322e-23, 4.1490354e-24], [90.991327, 9.70896e-23, 2.52004e-23, 1.51202e-23, 4.1464787e-24], [91.201084, 9.70893e-23, 2.51806e-23, 1.51084e-23, 4.1439527e-24], [91.411324, 9.70891e-23, 2.5161e-23, 1.50966e-23, 4.1414571e-24], [91.622049, 9.70891e-23, 2.51415e-23, 1.50849e-23, 4.1389917e-24], [91.83326, 9.70891e-23, 2.51222e-23, 1.50733e-23, 4.136556e-24], [92.044957, 9.70894e-23, 2.51031e-23, 1.50619e-23, 4.1341498e-24], [92.257143, 9.70897e-23, 2.50841e-23, 1.50505e-23, 4.1317728e-24], [92.469817, 9.70902e-23, 2.50653e-23, 1.50392e-23, 4.1294245e-24], [92.682982, 9.70908e-23, 2.50467e-23, 1.5028e-23, 4.1271047e-24], [92.896639, 9.70915e-23, 2.50282e-23, 1.50169e-23, 4.1248131e-24], [93.110788, 9.70924e-23, 2.50099e-23, 1.50059e-23, 4.1225494e-24], [93.32543, 9.70934e-23, 2.49917e-23, 1.4995e-23, 4.1203132e-24], [93.540567, 9.70945e-23, 2.49737e-23, 1.49842e-23, 4.1181043e-24], [93.756201, 9.70957e-23, 2.49558e-23, 1.49735e-23, 4.1159224e-24], [93.972331, 9.70971e-23, 2.49381e-23, 1.49629e-23, 4.1137671e-24], [94.18896, 9.70985e-23, 2.49206e-23, 1.49524e-23, 4.1116382e-24], [94.406088, 9.71002e-23, 2.49032e-23, 1.49419e-23, 4.1095353e-24], [94.623716, 9.71019e-23, 2.4886e-23, 1.49316e-23, 4.1074583e-24], [94.841846, 9.71037e-23, 2.48689e-23, 1.49213e-23, 4.1054068e-24], [95.060479, 9.71057e-23, 2.48519e-23, 1.49112e-23, 4.1033805e-24], [95.279616, 9.71078e-23, 2.48351e-23, 1.49011e-23, 4.1013791e-24], [95.499259, 9.71101e-23, 2.48185e-23, 1.48911e-23, 4.0994025e-24], [95.719407, 9.71124e-23, 2.4802e-23, 1.48812e-23, 4.0974503e-24], [95.940063, 9.71149e-23, 2.47856e-23, 1.48714e-23, 4.0955222e-24], [96.161228, 9.71175e-23, 2.47694e-23, 1.48616e-23, 4.0936181e-24], [96.382902, 9.71202e-23, 2.47533e-23, 1.4852e-23, 4.0917376e-24], [96.605088, 9.71231e-23, 2.47374e-23, 1.48425e-23, 4.0898804e-24], [96.827786, 9.71261e-23, 2.47216e-23, 1.4833e-23, 4.0880464e-24], [97.050997, 9.71292e-23, 2.4706e-23, 1.48236e-23, 4.0862354e-24], [97.274722, 9.71324e-23, 2.46905e-23, 1.48143e-23, 4.0844469e-24], [97.498964, 9.71357e-23, 2.46752e-23, 1.48051e-23, 4.0826809e-24], [97.723722, 9.71392e-23, 2.46599e-23, 1.4796e-23, 4.0809371e-24], [97.948999, 9.71427e-23, 2.46449e-23, 1.47869e-23, 4.0792152e-24], [98.174794, 9.71465e-23, 2.46299e-23, 1.4778e-23, 4.077515e-24], [98.401111, 9.71503e-23, 2.46151e-23, 1.47691e-23, 4.0758363e-24], [98.627949, 9.71542e-23, 2.46004e-23, 1.47603e-23, 4.0741788e-24], [98.855309, 9.71583e-23, 2.45859e-23, 1.47515e-23, 4.0725425e-24], [99.083194, 9.71625e-23, 2.45715e-23, 1.47429e-23, 4.0709269e-24], [99.311605, 9.71668e-23, 2.45573e-23, 1.47344e-23, 4.069332e-24], [99.540542, 9.71712e-23, 2.45431e-23, 1.47259e-23, 4.0677575e-24], [99.770006, 9.71758e-23, 2.45291e-23, 1.47175e-23, 4.0662032e-24], [100.0, 9.71804e-23, 2.45153e-23, 1.47092e-23, 4.0646689e-24], [100.23052, 9.71852e-23, 2.45015e-23, 1.47009e-23, 4.0631545e-24], [100.46158, 9.71902e-23, 2.44879e-23, 1.46927e-23, 4.0616596e-24], [100.69317, 9.71952e-23, 2.44744e-23, 1.46847e-23, 4.0601842e-24], [100.92529, 9.72003e-23, 2.44611e-23, 1.46767e-23, 4.058728e-24], [101.15795, 9.72056e-23, 2.44479e-23, 1.46687e-23, 4.0572909e-24], [101.39114, 9.7211e-23, 2.44348e-23, 1.46609e-23, 4.0558726e-24], [101.62487, 9.72165e-23, 2.44218e-23, 1.46531e-23, 4.0544731e-24], [101.85914, 9.72222e-23, 2.4409e-23, 1.46454e-23, 4.053092e-24], [102.09395, 9.72279e-23, 2.43963e-23, 1.46378e-23, 4.0517294e-24], [102.3293, 9.72338e-23, 2.43837e-23, 1.46302e-23, 4.0503849e-24], [102.56519, 9.72398e-23, 2.43712e-23, 1.46227e-23, 4.0490584e-24], [102.80163, 9.72459e-23, 2.43589e-23, 1.46153e-23, 4.0477498e-24], [103.03861, 9.72522e-23, 2.43466e-23, 1.4608e-23, 4.0464589e-24], [103.27614, 9.72585e-23, 2.43346e-23, 1.46007e-23, 4.0451856e-24], [103.51422, 9.7265e-23, 2.43226e-23, 1.45935e-23, 4.0439297e-24], [103.75284, 9.72716e-23, 2.43107e-23, 1.45864e-23, 4.042691e-24], [103.99202, 9.72784e-23, 2.4299e-23, 1.45794e-23, 4.0414695e-24], [104.23174, 9.72852e-23, 2.42874e-23, 1.45724e-23, 4.0402649e-24], [104.47202, 9.72922e-23, 2.42759e-23, 1.45655e-23, 4.0390772e-24], [104.71285, 9.72992e-23, 2.42645e-23, 1.45587e-23, 4.0379061e-24], [104.95424, 9.73065e-23, 2.42533e-23, 1.4552e-23, 4.0367517e-24], [105.19619, 9.73138e-23, 2.42421e-23, 1.45453e-23, 4.0356137e-24], [105.43869, 9.73212e-23, 2.42311e-23, 1.45387e-23, 4.034492e-24], [105.68175, 9.73288e-23, 2.42202e-23, 1.45321e-23, 4.0333866e-24], [105.92537, 9.73365e-23, 2.42094e-23, 1.45257e-23, 4.0322973e-24], [106.16956, 9.73443e-23, 2.41988e-23, 1.45193e-23, 4.0312239e-24], [106.4143, 9.73522e-23, 2.41882e-23, 1.45129e-23, 4.0301664e-24], [106.65961, 9.73603e-23, 2.41778e-23, 1.45067e-23, 4.0291248e-24], [106.90549, 9.73685e-23, 2.41675e-23, 1.45005e-23, 4.0280987e-24], [107.15193, 9.73768e-23, 2.41572e-23, 1.44943e-23, 4.0270883e-24], [107.39894, 9.73852e-23, 2.41472e-23, 1.44883e-23, 4.0260933e-24], [107.64652, 9.73938e-23, 2.41372e-23, 1.44823e-23, 4.0251138e-24], [107.89467, 9.74024e-23, 2.41273e-23, 1.44764e-23, 4.0241495e-24], [108.1434, 9.74112e-23, 2.41176e-23, 1.44705e-23, 4.0232005e-24], [108.39269, 9.74201e-23, 2.41079e-23, 1.44647e-23, 4.0222666e-24], [108.64256, 9.74292e-23, 2.40984e-23, 1.4459e-23, 4.0213478e-24], [108.89301, 9.74383e-23, 2.4089e-23, 1.44534e-23, 4.020444e-24], [109.14403, 9.74476e-23, 2.40797e-23, 1.44478e-23, 4.0195552e-24], [109.39564, 9.7457e-23, 2.40705e-23, 1.44423e-23, 4.0186811e-24], [109.64782, 9.74666e-23, 2.40614e-23, 1.44368e-23, 4.017822e-24], [109.90058, 9.74762e-23, 2.40524e-23, 1.44314e-23, 4.0169775e-24], [110.15393, 9.7486e-23, 2.40435e-23, 1.44261e-23, 4.0161478e-24], [110.40786, 9.74959e-23, 2.40348e-23, 1.44209e-23, 4.0153326e-24], [110.66238, 9.7506e-23, 2.40261e-23, 1.44157e-23, 4.0145322e-24], [110.91748, 9.75161e-23, 2.40176e-23, 1.44105e-23, 4.0137462e-24], [111.17317, 9.75264e-23, 2.40092e-23, 1.44055e-23, 4.0129748e-24], [111.42945, 9.75368e-23, 2.40008e-23, 1.44005e-23, 4.0122179e-24], [111.68632, 9.75473e-23, 2.39926e-23, 1.43956e-23, 4.0114754e-24], [111.94379, 9.7558e-23, 2.39845e-23, 1.43907e-23, 4.0107474e-24], [112.20185, 9.75688e-23, 2.39765e-23, 1.43859e-23, 4.0100338e-24], [112.4605, 9.75797e-23, 2.39686e-23, 1.43812e-23, 4.0093346e-24], [112.71975, 9.75908e-23, 2.39608e-23, 1.43765e-23, 4.0086498e-24], [112.97959, 9.76019e-23, 2.39532e-23, 1.43719e-23, 4.0079793e-24], [113.24004, 9.76133e-23, 2.39456e-23, 1.43674e-23, 4.0073232e-24], [113.50108, 9.76247e-23, 2.39381e-23, 1.43629e-23, 4.0066815e-24], [113.76273, 9.76363e-23, 2.39308e-23, 1.43585e-23, 4.0060542e-24], [114.02498, 9.76479e-23, 2.39235e-23, 1.43541e-23, 4.0054413e-24], [114.28783, 9.76598e-23, 2.39164e-23, 1.43498e-23, 4.0048427e-24], [114.55129, 9.76717e-23, 2.39094e-23, 1.43456e-23, 4.0042586e-24], [114.81536, 9.76838e-23, 2.39024e-23, 1.43415e-23, 4.0036889e-24], [115.08004, 9.7696e-23, 2.38956e-23, 1.43374e-23, 4.0031337e-24], [115.34533, 9.77084e-23, 2.38889e-23, 1.43333e-23, 4.0025929e-24], [115.61122, 9.77209e-23, 2.38823e-23, 1.43294e-23, 4.0020667e-24], [115.87774, 9.77335e-23, 2.38758e-23, 1.43255e-23, 4.0015551e-24], [116.14486, 9.77462e-23, 2.38694e-23, 1.43216e-23, 4.0010581e-24], [116.4126, 9.77591e-23, 2.38631e-23, 1.43178e-23, 4.0005758e-24], [116.68096, 9.77721e-23, 2.38569e-23, 1.43141e-23, 4.0001081e-24], [116.94994, 9.77853e-23, 2.38508e-23, 1.43105e-23, 3.9996553e-24], [117.21954, 9.77986e-23, 2.38448e-23, 1.43069e-23, 3.9992173e-24], [117.48976, 9.7812e-23, 2.3839e-23, 1.43034e-23, 3.9987943e-24], [117.7606, 9.78256e-23, 2.38332e-23, 1.42999e-23, 3.9983863e-24], [118.03206, 9.78393e-23, 2.38276e-23, 1.42965e-23, 3.9979933e-24], [118.30416, 9.78531e-23, 2.3822e-23, 1.42932e-23, 3.9976155e-24], [118.57687, 9.78671e-23, 2.38166e-23, 1.42899e-23, 3.997253e-24], [118.85022, 9.78812e-23, 2.38112e-23, 1.42867e-23, 3.9969059e-24], [119.1242, 9.78955e-23, 2.3806e-23, 1.42836e-23, 3.9965743e-24], [119.39881, 9.79099e-23, 2.38009e-23, 1.42805e-23, 3.9962582e-24], [119.67405, 9.79245e-23, 2.37959e-23, 1.42775e-23, 3.9959578e-24], [119.94993, 9.79391e-23, 2.3791e-23, 1.42746e-23, 3.9956733e-24], [120.22644, 9.7954e-23, 2.37862e-23, 1.42717e-23, 3.9954048e-24], [120.50359, 9.7969e-23, 2.37815e-23, 1.42689e-23, 3.9951523e-24], [120.78138, 9.79841e-23, 2.37769e-23, 1.42662e-23, 3.9949161e-24], [121.05981, 9.79994e-23, 2.37725e-23, 1.42635e-23, 3.9946963e-24], [121.33889, 9.80148e-23, 2.37681e-23, 1.42609e-23, 3.9944931e-24], [121.6186, 9.80303e-23, 2.37639e-23, 1.42583e-23, 3.9943066e-24], [121.89896, 9.80461e-23, 2.37597e-23, 1.42558e-23, 3.994137e-24], [122.17997, 9.80619e-23, 2.37557e-23, 1.42534e-23, 3.9939845e-24], [122.46162, 9.8078e-23, 2.37518e-23, 1.42511e-23, 3.9938492e-24], [122.74392, 9.80941e-23, 2.3748e-23, 1.42488e-23, 3.9937315e-24], [123.02688, 9.81104e-23, 2.37443e-23, 1.42466e-23, 3.9936314e-24], [123.31048, 9.81269e-23, 2.37407e-23, 1.42444e-23, 3.9935492e-24], [123.59474, 9.81436e-23, 2.37373e-23, 1.42424e-23, 3.9934852e-24], [123.87966, 9.81604e-23, 2.37339e-23, 1.42404e-23, 3.9934395e-24], [124.16523, 9.81773e-23, 2.37307e-23, 1.42384e-23, 3.9934124e-24], [124.45146, 9.81944e-23, 2.37276e-23, 1.42366e-23, 3.9934041e-24], [124.73835, 9.82117e-23, 2.37246e-23, 1.42348e-23, 3.9934149e-24], [125.0259, 9.82291e-23, 2.37217e-23, 1.4233e-23, 3.9934452e-24], [125.31412, 9.82467e-23, 2.37189e-23, 1.42314e-23, 3.9934951e-24], [125.603, 9.82645e-23, 2.37163e-23, 1.42298e-23, 3.993565e-24], [125.89254, 9.82824e-23, 2.37138e-23, 1.42283e-23, 3.9936551e-24], [126.18275, 9.83005e-23, 2.37114e-23, 1.42268e-23, 3.9937658e-24], [126.47363, 9.83187e-23, 2.37091e-23, 1.42255e-23, 3.9938975e-24], [126.76519, 9.83372e-23, 2.37069e-23, 1.42242e-23, 3.9940504e-24], [127.05741, 9.83558e-23, 2.37049e-23, 1.42229e-23, 3.9942248e-24], [127.35031, 9.83745e-23, 2.3703e-23, 1.42218e-23, 3.9944213e-24], [127.64388, 9.83935e-23, 2.37012e-23, 1.42207e-23, 3.9946402e-24], [127.93813, 9.84126e-23, 2.36996e-23, 1.42197e-23, 3.9948817e-24], [128.23306, 9.84319e-23, 2.36981e-23, 1.42188e-23, 3.9951465e-24], [128.52867, 9.84514e-23, 2.36967e-23, 1.4218e-23, 3.9954348e-24], [128.82496, 9.84711e-23, 2.36954e-23, 1.42172e-23, 3.9957472e-24], [129.12193, 9.84909e-23, 2.36943e-23, 1.42166e-23, 3.996084e-24], [129.41958, 9.8511e-23, 2.36933e-23, 1.4216e-23, 3.9964458e-24], [129.71793, 9.85312e-23, 2.36924e-23, 1.42155e-23, 3.9968331e-24], [130.01696, 9.85516e-23, 2.36917e-23, 1.4215e-23, 3.9972463e-24], [130.31668, 9.85722e-23, 2.36912e-23, 1.42147e-23, 3.9976861e-24], [130.61709, 9.8593e-23, 2.36907e-23, 1.42144e-23, 3.9981528e-24], [130.91819, 9.8614e-23, 2.36905e-23, 1.42143e-23, 3.9986472e-24], [131.21999, 9.86352e-23, 2.36903e-23, 1.42142e-23, 3.9991698e-24], [131.52248, 9.86566e-23, 2.36903e-23, 1.42142e-23, 3.9997212e-24], [131.82567, 9.86782e-23, 2.36905e-23, 1.42143e-23, 4.000302e-24], [132.12956, 9.87001e-23, 2.36908e-23, 1.42145e-23, 4.000913e-24], [132.43415, 9.87221e-23, 2.36913e-23, 1.42148e-23, 4.0015547e-24], [132.73945, 9.87443e-23, 2.36919e-23, 1.42151e-23, 4.002228e-24], [133.04544, 9.87668e-23, 2.36927e-23, 1.42156e-23, 4.0029336e-24], [133.35214, 9.87895e-23, 2.36937e-23, 1.42162e-23, 4.0036721e-24], [133.65955, 9.88124e-23, 2.36948e-23, 1.42169e-23, 4.0044446e-24], [133.96767, 9.88355e-23, 2.36961e-23, 1.42176e-23, 4.0052516e-24], [134.2765, 9.88589e-23, 2.36975e-23, 1.42185e-23, 4.0060942e-24], [134.58604, 9.88824e-23, 2.36992e-23, 1.42195e-23, 4.0069732e-24], [134.89629, 9.89063e-23, 2.3701e-23, 1.42206e-23, 4.0078896e-24], [135.20726, 9.89303e-23, 2.3703e-23, 1.42218e-23, 4.0088443e-24], [135.51894, 9.89547e-23, 2.37052e-23, 1.42231e-23, 4.0098383e-24], [135.83134, 9.89792e-23, 2.37076e-23, 1.42245e-23, 4.0108728e-24], [136.14447, 9.90041e-23, 2.37102e-23, 1.42261e-23, 4.0119487e-24], [136.45831, 9.90291e-23, 2.37129e-23, 1.42278e-23, 4.0130673e-24], [136.77288, 9.90545e-23, 2.37159e-23, 1.42296e-23, 4.0142297e-24], [137.08818, 9.90801e-23, 2.37191e-23, 1.42315e-23, 4.0154372e-24], [137.4042, 9.9106e-23, 2.37225e-23, 1.42335e-23, 4.0166911e-24], [137.72095, 9.91322e-23, 2.37262e-23, 1.42357e-23, 4.0179927e-24], [138.03843, 9.91586e-23, 2.373e-23, 1.4238e-23, 4.0193435e-24], [138.35664, 9.91854e-23, 2.37341e-23, 1.42405e-23, 4.0207449e-24], [138.67558, 9.92124e-23, 2.37384e-23, 1.4243e-23, 4.0221984e-24], [138.99526, 9.92398e-23, 2.3743e-23, 1.42458e-23, 4.0237057e-24], [139.31568, 9.92674e-23, 2.37478e-23, 1.42487e-23, 4.0252684e-24], [139.63684, 9.92954e-23, 2.37529e-23, 1.42517e-23, 4.0268884e-24], [139.95873, 9.93237e-23, 2.37582e-23, 1.42549e-23, 4.0285674e-24], [140.28137, 9.93524e-23, 2.37638e-23, 1.42583e-23, 4.0303074e-24], [140.60475, 9.93813e-23, 2.37697e-23, 1.42618e-23, 4.0321103e-24], [140.92888, 9.94107e-23, 2.37759e-23, 1.42655e-23, 4.0339784e-24], [141.25375, 9.94404e-23, 2.37823e-23, 1.42694e-23, 4.0359139e-24], [141.57938, 9.94704e-23, 2.37891e-23, 1.42735e-23, 4.037919e-24], [141.90575, 9.95009e-23, 2.37962e-23, 1.42777e-23, 4.0399961e-24], [142.23288, 9.95317e-23, 2.38036e-23, 1.42822e-23, 4.042148e-24], [142.56076, 9.95629e-23, 2.38114e-23, 1.42868e-23, 4.0443772e-24], [142.8894, 9.95946e-23, 2.38195e-23, 1.42917e-23, 4.0466866e-24], [143.21879, 9.96267e-23, 2.3828e-23, 1.42968e-23, 4.0490792e-24], [143.54894, 9.96592e-23, 2.38368e-23, 1.43021e-23, 4.0515581e-24], [143.87986, 9.96921e-23, 2.3846e-23, 1.43076e-23, 4.0541266e-24], [144.21154, 9.97256e-23, 2.38557e-23, 1.43134e-23, 4.0567882e-24], [144.54398, 9.97595e-23, 2.38657e-23, 1.43194e-23, 4.0595466e-24], [144.87719, 9.97939e-23, 2.38762e-23, 1.43257e-23, 4.0624056e-24], [145.21116, 9.98288e-23, 2.38872e-23, 1.43323e-23, 4.0653693e-24], [145.54591, 9.98643e-23, 2.38986e-23, 1.43391e-23, 4.0684421e-24], [145.88143, 9.99003e-23, 2.39105e-23, 1.43463e-23, 4.0716284e-24], [146.21772, 9.99368e-23, 2.39229e-23, 1.43537e-23, 4.0749332e-24], [146.55478, 9.9974e-23, 2.39358e-23, 1.43615e-23, 4.0783616e-24], [146.89263, 1.00012e-22, 2.39493e-23, 1.43696e-23, 4.081919e-24], [147.23125, 1.0005e-22, 2.39634e-23, 1.4378e-23, 4.085611e-24], [147.57065, 1.00089e-22, 2.39781e-23, 1.43868e-23, 4.0894439e-24], [147.91084, 1.00129e-22, 2.39934e-23, 1.4396e-23, 4.0934241e-24], [148.25181, 1.0017e-22, 2.40093e-23, 1.44056e-23, 4.0975585e-24], [148.59356, 1.00211e-22, 2.4026e-23, 1.44156e-23, 4.1018544e-24], [148.93611, 1.00253e-22, 2.40433e-23, 1.4426e-23, 4.1063197e-24], [149.27944, 1.00296e-22, 2.40615e-23, 1.44369e-23, 4.1109626e-24], [149.62357, 1.00339e-22, 2.40804e-23, 1.44482e-23, 4.1157919e-24], [149.96848, 1.00384e-22, 2.41002e-23, 1.44601e-23, 4.1208173e-24], [150.3142, 1.0043e-22, 2.41208e-23, 1.44725e-23, 4.1260488e-24], [150.66071, 1.00476e-22, 2.41424e-23, 1.44855e-23, 4.1314971e-24], [151.00802, 1.00524e-22, 2.4165e-23, 1.4499e-23, 4.137174e-24], [151.35612, 1.00572e-22, 2.41886e-23, 1.45131e-23, 4.1430918e-24], [151.70504, 1.00622e-22, 2.42133e-23, 1.4528e-23, 4.1492639e-24], [152.05475, 1.00673e-22, 2.42391e-23, 1.45435e-23, 4.1557046e-24], [152.40528, 1.00726e-22, 2.42661e-23, 1.45597e-23, 4.1624292e-24], [152.75661, 1.00779e-22, 2.42945e-23, 1.45767e-23, 4.1694545e-24], [153.10875, 1.00834e-22, 2.43242e-23, 1.45945e-23, 4.1767981e-24], [153.4617, 1.00891e-22, 2.43554e-23, 1.46132e-23, 4.1844795e-24], [153.81546, 1.00949e-22, 2.43881e-23, 1.46329e-23, 4.1925194e-24], [154.17005, 1.01009e-22, 2.44224e-23, 1.46535e-23, 4.2009405e-24], [154.52544, 1.01071e-22, 2.44585e-23, 1.46751e-23, 4.209767e-24], [154.88166, 1.01134e-22, 2.44965e-23, 1.46979e-23, 4.2190256e-24], [155.2387, 1.012e-22, 2.45365e-23, 1.47219e-23, 4.228745e-24], [155.59656, 1.01268e-22, 2.45786e-23, 1.47471e-23, 4.2389566e-24], [155.95525, 1.01338e-22, 2.4623e-23, 1.47738e-23, 4.2496943e-24], [156.31476, 1.01411e-22, 2.46698e-23, 1.48019e-23, 4.2609955e-24], [156.67511, 1.01487e-22, 2.47192e-23, 1.48315e-23, 4.2729009e-24], [157.03628, 1.01565e-22, 2.47715e-23, 1.48629e-23, 4.285455e-24], [157.39829, 1.01646e-22, 2.48268e-23, 1.48961e-23, 4.2987066e-24], [157.76113, 1.01731e-22, 2.48854e-23, 1.49313e-23, 4.3127095e-24], [158.1248, 1.0182e-22, 2.49476e-23, 1.49685e-23, 4.3275226e-24], [158.48932, 1.01912e-22, 2.50136e-23, 1.50081e-23, 4.3432112e-24], [158.85467, 1.02009e-22, 2.50837e-23, 1.50502e-23, 4.3598472e-24], [159.22087, 1.02111e-22, 2.51584e-23, 1.5095e-23, 4.3775102e-24], [159.58791, 1.02217e-22, 2.52379e-23, 1.51428e-23, 4.3962887e-24], [159.9558, 1.02329e-22, 2.53229e-23, 1.51937e-23, 4.4162809e-24], [160.32454, 1.02447e-22, 2.54137e-23, 1.52482e-23, 4.4375965e-24], [160.69413, 1.02572e-22, 2.55109e-23, 1.53065e-23, 4.4603578e-24], [161.06456, 1.02704e-22, 2.56151e-23, 1.53691e-23, 4.4847019e-24], [161.43586, 1.02845e-22, 2.5727e-23, 1.54362e-23, 4.5107829e-24], [161.808, 1.02994e-22, 2.58475e-23, 1.55085e-23, 4.5387741e-24], [162.18101, 1.03153e-22, 2.59774e-23, 1.55865e-23, 4.5688714e-24], [162.55488, 1.03323e-22, 2.61178e-23, 1.56707e-23, 4.6012965e-24], [162.9296, 1.03506e-22, 2.62697e-23, 1.57618e-23, 4.6363016e-24], [163.30519, 1.03703e-22, 2.64346e-23, 1.58607e-23, 4.674174e-24], [163.68165, 1.03915e-22, 2.66139e-23, 1.59683e-23, 4.7152428e-24], [164.05898, 1.04144e-22, 2.68094e-23, 1.60856e-23, 4.7598856e-24], [164.43717, 1.04394e-22, 2.70231e-23, 1.62138e-23, 4.808538e-24], [164.81624, 1.04666e-22, 2.72574e-23, 1.63544e-23, 4.8617043e-24], [165.19618, 1.04965e-22, 2.7515e-23, 1.6509e-23, 4.9199711e-24], [165.577, 1.05293e-22, 2.77992e-23, 1.66795e-23, 4.9840238e-24], [165.95869, 1.05656e-22, 2.81137e-23, 1.68682e-23, 5.0546674e-24], [166.34127, 1.06058e-22, 2.84631e-23, 1.70779e-23, 5.1328528e-24], [166.72472, 1.06508e-22, 2.88528e-23, 1.73117e-23, 5.2197092e-24], [167.10906, 1.07012e-22, 2.92891e-23, 1.75735e-23, 5.3165866e-24], [167.49429, 1.07581e-22, 2.97799e-23, 1.78679e-23, 5.42511e-24], [167.8804, 1.08229e-22, 3.03347e-23, 1.82008e-23, 5.5472499e-24], [168.26741, 1.08969e-22, 3.0965e-23, 1.8579e-23, 5.6854158e-24], [168.6553, 1.09823e-22, 3.16854e-23, 1.90113e-23, 5.8425803e-24], [169.04409, 1.10816e-22, 3.2514e-23, 1.95084e-23, 6.0224483e-24], [169.43378, 1.11982e-22, 3.34734e-23, 2.00841e-23, 6.2296888e-24], [169.82437, 1.13363e-22, 3.45931e-23, 2.07559e-23, 6.4702609e-24], [170.21585, 1.15019e-22, 3.5911e-23, 2.15466e-23, 6.7518796e-24], [170.60824, 1.17028e-22, 3.74773e-23, 2.24864e-23, 7.0846975e-24], [171.00153, 1.19501e-22, 3.93596e-23, 2.36158e-23, 7.4823303e-24], [171.39573, 1.22595e-22, 4.16507e-23, 2.49904e-23, 7.9634445e-24], [171.79084, 1.26535e-22, 4.44813e-23, 2.66888e-23, 8.5543065e-24], [172.18686, 1.31664e-22, 4.80418e-23, 2.88251e-23, 9.2930485e-24], [172.58379, 1.38511e-22, 5.26194e-23, 3.15716e-23, 1.0237173e-23], [172.98164, 1.47934e-22, 5.86684e-23, 3.52011e-23, 1.1477585e-23], [173.3804, 1.61412e-22, 6.69522e-23, 4.01713e-23, 1.3166903e-23], [173.78008, 1.81671e-22, 7.88578e-23, 4.73147e-23, 1.5582489e-23], [174.18069, 2.14304e-22, 9.71974e-23, 5.83184e-23, 1.9286601e-23], [174.58222, 2.72703e-22, 1.28669e-22, 7.72017e-23, 2.5618763e-23], [174.98467, 3.98609e-22, 1.94126e-22, 1.16475e-22, 3.8749049e-23], [175.38805, 8.20702e-22, 4.07859e-22, 2.44715e-22, 8.153558e-23], [175.79236, 6.79937e-21, 3.39938e-21, 2.03963e-21, 6.7987215e-22], [176.1976, 6.59817e-22, 3.26798e-22, 1.96079e-22, 6.5314623e-23], [176.60378, 3.56144e-22, 1.72236e-22, 1.03342e-22, 3.4361914e-23], [177.0109, 2.52081e-22, 1.17644e-22, 7.05864e-23, 2.3403975e-23], [177.41895, 2.01458e-22, 8.99915e-23, 5.39949e-23, 1.783506e-23], [177.82794, 1.72554e-22, 7.34446e-23, 4.40668e-23, 1.4488784e-23], [178.23788, 1.54433e-22, 6.25383e-23, 3.7523e-23, 1.2272395e-23], [178.64876, 1.4234e-22, 5.48817e-23, 3.2929e-23, 1.0707957e-23], [179.06059, 1.33892e-22, 4.92629e-23, 2.95577e-23, 9.5532032e-24], [179.47336, 1.27778e-22, 4.50017e-23, 2.7001e-23, 8.6721767e-24], [179.88709, 1.23228e-22, 4.16869e-23, 2.50121e-23, 7.9826708e-24], [180.30177, 1.1976e-22, 3.90556e-23, 2.34334e-23, 7.432055e-24], [180.71741, 1.17064e-22, 3.69319e-23, 2.21592e-23, 6.9850799e-24], [181.13401, 1.14934e-22, 3.51939e-23, 2.11164e-23, 6.6172619e-24], [181.55157, 1.13226e-22, 3.37546e-23, 2.02527e-23, 6.3110782e-24], [181.97009, 1.11839e-22, 3.25502e-23, 1.95301e-23, 6.053671e-24], [182.38957, 1.10701e-22, 3.15333e-23, 1.892e-23, 5.8354042e-24], [182.81002, 1.09758e-22, 3.06678e-23, 1.84007e-23, 5.6489249e-24], [183.23144, 1.08969e-22, 2.99259e-23, 1.79555e-23, 5.4885349e-24], [183.65383, 1.08305e-22, 2.92857e-23, 1.75714e-23, 5.3497575e-24], [184.0772, 1.07743e-22, 2.873e-23, 1.7238e-23, 5.2290341e-24], [184.50154, 1.07264e-22, 2.82451e-23, 1.6947e-23, 5.1235056e-24], [184.92686, 1.06854e-22, 2.78198e-23, 1.66919e-23, 5.0308533e-24], [185.35316, 1.06501e-22, 2.74452e-23, 1.64671e-23, 4.9491804e-24], [185.78045, 1.06196e-22, 2.71138e-23, 1.62683e-23, 4.8769238e-24], [186.20871, 1.05933e-22, 2.68194e-23, 1.60916e-23, 4.8127858e-24], [186.63797, 1.05704e-22, 2.65571e-23, 1.59342e-23, 4.7556823e-24], [187.06821, 1.05506e-22, 2.63225e-23, 1.57935e-23, 4.7047023e-24], [187.49945, 1.05333e-22, 2.6112e-23, 1.56672e-23, 4.6590759e-24], [187.93168, 1.05183e-22, 2.59226e-23, 1.55536e-23, 4.6181492e-24], [188.36491, 1.05052e-22, 2.57518e-23, 1.54511e-23, 4.581364e-24], [188.79913, 1.04938e-22, 2.55973e-23, 1.53584e-23, 4.548242e-24], [189.23436, 1.04839e-22, 2.54573e-23, 1.52744e-23, 4.5183713e-24], [189.67059, 1.04753e-22, 2.53301e-23, 1.51981e-23, 4.4913957e-24], [190.10783, 1.04679e-22, 2.52143e-23, 1.51286e-23, 4.4670065e-24], [190.54607, 1.04616e-22, 2.51087e-23, 1.50652e-23, 4.4449347e-24], [190.98533, 1.04562e-22, 2.50122e-23, 1.50073e-23, 4.4249453e-24], [191.42559, 1.04516e-22, 2.49239e-23, 1.49543e-23, 4.4068325e-24], [191.86687, 1.04477e-22, 2.48429e-23, 1.49058e-23, 4.3904152e-24], [192.30917, 1.04445e-22, 2.47686e-23, 1.48612e-23, 4.3755338e-24], [192.75249, 1.0442e-22, 2.47004e-23, 1.48202e-23, 4.3620473e-24], [193.19683, 1.04399e-22, 2.46376e-23, 1.47826e-23, 4.3498305e-24], [193.6422, 1.04384e-22, 2.45798e-23, 1.47479e-23, 4.3387723e-24], [194.08859, 1.04374e-22, 2.45265e-23, 1.47159e-23, 4.3287737e-24], [194.53601, 1.04367e-22, 2.44773e-23, 1.46864e-23, 4.3197463e-24], [194.98446, 1.04364e-22, 2.44319e-23, 1.46591e-23, 4.311611e-24], [195.43395, 1.04365e-22, 2.43899e-23, 1.46339e-23, 4.3042968e-24], [195.88447, 1.04369e-22, 2.43511e-23, 1.46107e-23, 4.2977399e-24], [196.33603, 1.04376e-22, 2.43153e-23, 1.45892e-23, 4.2918831e-24], [196.78863, 1.04386e-22, 2.42821e-23, 1.45693e-23, 4.2866744e-24], [197.24227, 1.04398e-22, 2.42514e-23, 1.45508e-23, 4.282067e-24], [197.69696, 1.04413e-22, 2.4223e-23, 1.45338e-23, 4.2780187e-24], [198.1527, 1.04429e-22, 2.41967e-23, 1.4518e-23, 4.274491e-24], [198.60949, 1.04448e-22, 2.41724e-23, 1.45034e-23, 4.2714489e-24], [199.06733, 1.04469e-22, 2.41499e-23, 1.44899e-23, 4.2688608e-24], [199.52623, 1.04491e-22, 2.41291e-23, 1.44775e-23, 4.2666977e-24], [199.98619, 1.04515e-22, 2.41099e-23, 1.44659e-23, 4.2649333e-24], [200.4472, 1.04541e-22, 2.40922e-23, 1.44553e-23, 4.2635434e-24], [200.90928, 1.04568e-22, 2.40758e-23, 1.44455e-23, 4.262506e-24], [201.37242, 1.04596e-22, 2.40607e-23, 1.44364e-23, 4.2618009e-24], [201.83664, 1.04626e-22, 2.40469e-23, 1.44281e-23, 4.2614095e-24], [202.30192, 1.04657e-22, 2.40341e-23, 1.44205e-23, 4.2613149e-24], [202.76827, 1.04689e-22, 2.40224e-23, 1.44135e-23, 4.2615014e-24], [203.2357, 1.04723e-22, 2.40117e-23, 1.4407e-23, 4.2619546e-24], [203.70421, 1.04757e-22, 2.4002e-23, 1.44012e-23, 4.2626613e-24], [204.17379, 1.04792e-22, 2.39931e-23, 1.43959e-23, 4.2636093e-24], [204.64446, 1.04828e-22, 2.3985e-23, 1.4391e-23, 4.2647872e-24], [205.11622, 1.04865e-22, 2.39778e-23, 1.43867e-23, 4.2661846e-24], [205.58906, 1.04903e-22, 2.39713e-23, 1.43828e-23, 4.2677919e-24], [206.06299, 1.04942e-22, 2.39654e-23, 1.43793e-23, 4.2696002e-24], [206.53802, 1.04982e-22, 2.39603e-23, 1.43762e-23, 4.2716012e-24], [207.01413, 1.05022e-22, 2.39557e-23, 1.43734e-23, 4.2737872e-24], [207.49135, 1.05063e-22, 2.39518e-23, 1.43711e-23, 4.2761511e-24], [207.96967, 1.05105e-22, 2.39484e-23, 1.4369e-23, 4.2786864e-24], [208.44909, 1.05148e-22, 2.39456e-23, 1.43673e-23, 4.2813869e-24], [208.92961, 1.05191e-22, 2.39432e-23, 1.43659e-23, 4.2842468e-24], [209.41125, 1.05234e-22, 2.39414e-23, 1.43648e-23, 4.287261e-24], [209.89399, 1.05279e-22, 2.394e-23, 1.4364e-23, 4.2904244e-24], [210.37784, 1.05324e-22, 2.39391e-23, 1.43634e-23, 4.2937325e-24], [210.86281, 1.05369e-22, 2.39385e-23, 1.43631e-23, 4.2971809e-24], [211.3489, 1.05415e-22, 2.39384e-23, 1.43631e-23, 4.3007657e-24], [211.83611, 1.05462e-22, 2.39387e-23, 1.43632e-23, 4.3044831e-24], [212.32445, 1.05509e-22, 2.39393e-23, 1.43636e-23, 4.3083298e-24], [212.8139, 1.05557e-22, 2.39403e-23, 1.43642e-23, 4.3123024e-24], [213.30449, 1.05605e-22, 2.39417e-23, 1.4365e-23, 4.3163978e-24], [213.79621, 1.05654e-22, 2.39433e-23, 1.4366e-23, 4.3206134e-24], [214.28906, 1.05703e-22, 2.39453e-23, 1.43672e-23, 4.3249464e-24], [214.78305, 1.05752e-22, 2.39476e-23, 1.43686e-23, 4.3293943e-24], [215.27817, 1.05802e-22, 2.39502e-23, 1.43701e-23, 4.3339549e-24], [215.77444, 1.05853e-22, 2.3953e-23, 1.43718e-23, 4.338626e-24], [216.27185, 1.05904e-22, 2.39561e-23, 1.43737e-23, 4.3434055e-24], [216.77041, 1.05955e-22, 2.39595e-23, 1.43757e-23, 4.3482917e-24], [217.27012, 1.06007e-22, 2.39632e-23, 1.43779e-23, 4.3532826e-24], [217.77098, 1.06059e-22, 2.3967e-23, 1.43802e-23, 4.3583767e-24], [218.27299, 1.06111e-22, 2.39711e-23, 1.43827e-23, 4.3635724e-24], [218.77616, 1.06164e-22, 2.39755e-23, 1.43853e-23, 4.3688683e-24], [219.28049, 1.06218e-22, 2.398e-23, 1.4388e-23, 4.3742631e-24], [219.78599, 1.06272e-22, 2.39848e-23, 1.43909e-23, 4.3797553e-24], [220.29265, 1.06326e-22, 2.39898e-23, 1.43939e-23, 4.385344e-24], [220.80047, 1.0638e-22, 2.3995e-23, 1.4397e-23, 4.3910279e-24], [221.30947, 1.06435e-22, 2.40003e-23, 1.44002e-23, 4.396806e-24], [221.81964, 1.0649e-22, 2.40059e-23, 1.44035e-23, 4.4026774e-24], [222.33099, 1.06546e-22, 2.40116e-23, 1.4407e-23, 4.4086411e-24], [222.84351, 1.06602e-22, 2.40176e-23, 1.44105e-23, 4.4146964e-24], [223.35722, 1.06659e-22, 2.40237e-23, 1.44142e-23, 4.4208425e-24], [223.87211, 1.06715e-22, 2.40299e-23, 1.4418e-23, 4.4270785e-24], [224.38819, 1.06772e-22, 2.40364e-23, 1.44218e-23, 4.4334038e-24], [224.90546, 1.0683e-22, 2.4043e-23, 1.44258e-23, 4.4398179e-24], [225.42392, 1.06888e-22, 2.40497e-23, 1.44298e-23, 4.44632e-24], [225.94358, 1.06946e-22, 2.40567e-23, 1.4434e-23, 4.4529097e-24], [226.46443, 1.07004e-22, 2.40637e-23, 1.44382e-23, 4.4595864e-24], [226.98649, 1.07063e-22, 2.40709e-23, 1.44426e-23, 4.4663497e-24], [227.50974, 1.07122e-22, 2.40783e-23, 1.4447e-23, 4.4731992e-24], [228.03421, 1.07182e-22, 2.40858e-23, 1.44515e-23, 4.4801343e-24], [228.55988, 1.07241e-22, 2.40935e-23, 1.44561e-23, 4.4871548e-24], [229.08677, 1.07302e-22, 2.41012e-23, 1.44607e-23, 4.4942603e-24], [229.61486, 1.07362e-22, 2.41092e-23, 1.44655e-23, 4.5014505e-24], [230.14418, 1.07423e-22, 2.41172e-23, 1.44703e-23, 4.5087251e-24], [230.67472, 1.07484e-22, 2.41254e-23, 1.44752e-23, 4.5160839e-24], [231.20648, 1.07545e-22, 2.41337e-23, 1.44802e-23, 4.5235265e-24], [231.73946, 1.07607e-22, 2.41421e-23, 1.44853e-23, 4.5310528e-24], [232.27368, 1.07669e-22, 2.41507e-23, 1.44904e-23, 4.5386626e-24], [232.80913, 1.07732e-22, 2.41594e-23, 1.44956e-23, 4.5463557e-24], [233.34581, 1.07795e-22, 2.41682e-23, 1.45009e-23, 4.5541319e-24], [233.88372, 1.07858e-22, 2.41771e-23, 1.45063e-23, 4.5619911e-24], [234.42288, 1.07921e-22, 2.41861e-23, 1.45117e-23, 4.5699331e-24], [234.96328, 1.07985e-22, 2.41953e-23, 1.45172e-23, 4.577958e-24], [235.50493, 1.08049e-22, 2.42045e-23, 1.45227e-23, 4.5860655e-24], [236.04782, 1.08113e-22, 2.42139e-23, 1.45284e-23, 4.5942555e-24], [236.59197, 1.08178e-22, 2.42234e-23, 1.4534e-23, 4.6025281e-24], [237.13737, 1.08243e-22, 2.4233e-23, 1.45398e-23, 4.6108831e-24], [237.68403, 1.08308e-22, 2.42427e-23, 1.45456e-23, 4.6193206e-24], [238.23195, 1.08374e-22, 2.42525e-23, 1.45515e-23, 4.6278404e-24], [238.78113, 1.0844e-22, 2.42624e-23, 1.45575e-23, 4.6364426e-24], [239.33158, 1.08506e-22, 2.42725e-23, 1.45635e-23, 4.6451271e-24], [239.88329, 1.08573e-22, 2.42826e-23, 1.45696e-23, 4.653894e-24], [240.43628, 1.0864e-22, 2.42928e-23, 1.45757e-23, 4.6627432e-24], [240.99054, 1.08707e-22, 2.43031e-23, 1.45819e-23, 4.6716748e-24], [241.54608, 1.08774e-22, 2.43136e-23, 1.45881e-23, 4.6806888e-24], [242.1029, 1.08842e-22, 2.43241e-23, 1.45945e-23, 4.6897853e-24], [242.66101, 1.08911e-22, 2.43347e-23, 1.46008e-23, 4.6989642e-24], [243.2204, 1.08979e-22, 2.43455e-23, 1.46073e-23, 4.7082256e-24], [243.78108, 1.09048e-22, 2.43563e-23, 1.46138e-23, 4.7175695e-24], [244.34306, 1.09117e-22, 2.43672e-23, 1.46203e-23, 4.7269961e-24], [244.90632, 1.09186e-22, 2.43783e-23, 1.4627e-23, 4.7365054e-24], [245.47089, 1.09256e-22, 2.43894e-23, 1.46336e-23, 4.7460974e-24], [246.03676, 1.09326e-22, 2.44006e-23, 1.46404e-23, 4.7557723e-24], [246.60393, 1.09397e-22, 2.44119e-23, 1.46471e-23, 4.7655301e-24], [247.17241, 1.09468e-22, 2.44233e-23, 1.4654e-23, 4.7753708e-24], [247.74221, 1.09539e-22, 2.44348e-23, 1.46609e-23, 4.7852947e-24], [248.31331, 1.0961e-22, 2.44464e-23, 1.46678e-23, 4.7953017e-24], [248.88573, 1.09682e-22, 2.44581e-23, 1.46749e-23, 4.8053919e-24], [249.45947, 1.09754e-22, 2.44699e-23, 1.46819e-23, 4.8155656e-24], [250.03454, 1.09826e-22, 2.44817e-23, 1.4689e-23, 4.8258227e-24], [250.61093, 1.09899e-22, 2.44937e-23, 1.46962e-23, 4.8361634e-24], [251.18864, 1.09972e-22, 2.45057e-23, 1.47034e-23, 4.8465877e-24], [251.76769, 1.10045e-22, 2.45179e-23, 1.47107e-23, 4.8570959e-24], [252.34808, 1.10119e-22, 2.45301e-23, 1.47181e-23, 4.8676879e-24], [252.9298, 1.10193e-22, 2.45424e-23, 1.47255e-23, 4.878364e-24], [253.51286, 1.10267e-22, 2.45549e-23, 1.47329e-23, 4.8891242e-24], [254.09727, 1.10342e-22, 2.45674e-23, 1.47404e-23, 4.8999687e-24], [254.68303, 1.10416e-22, 2.45799e-23, 1.4748e-23, 4.9108975e-24], [255.27013, 1.10492e-22, 2.45926e-23, 1.47556e-23, 4.9219109e-24], [255.85859, 1.10567e-22, 2.46054e-23, 1.47632e-23, 4.9330088e-24], [256.4484, 1.10643e-22, 2.46183e-23, 1.4771e-23, 4.9441915e-24], [257.03958, 1.10719e-22, 2.46312e-23, 1.47787e-23, 4.9554591e-24], [257.63212, 1.10796e-22, 2.46442e-23, 1.47865e-23, 4.9668117e-24], [258.22602, 1.10873e-22, 2.46574e-23, 1.47944e-23, 4.9782495e-24], [258.82129, 1.1095e-22, 2.46706e-23, 1.48023e-23, 4.9897725e-24], [259.41794, 1.11028e-22, 2.46839e-23, 1.48103e-23, 5.0013808e-24], [260.01596, 1.11106e-22, 2.46972e-23, 1.48183e-23, 5.0130748e-24], [260.61535, 1.11184e-22, 2.47107e-23, 1.48264e-23, 5.0248543e-24], [261.21614, 1.11262e-22, 2.47243e-23, 1.48346e-23, 5.0367197e-24], [261.8183, 1.11341e-22, 2.47379e-23, 1.48428e-23, 5.048671e-24], [262.42185, 1.1142e-22, 2.47517e-23, 1.4851e-23, 5.0607083e-24], [263.0268, 1.115e-22, 2.47655e-23, 1.48593e-23, 5.0728319e-24], [263.63314, 1.1158e-22, 2.47794e-23, 1.48676e-23, 5.0850417e-24], [264.24088, 1.1166e-22, 2.47934e-23, 1.4876e-23, 5.097338e-24], [264.85001, 1.11741e-22, 2.48075e-23, 1.48845e-23, 5.109721e-24], [265.46056, 1.11822e-22, 2.48216e-23, 1.4893e-23, 5.1221906e-24], [266.07251, 1.11903e-22, 2.48359e-23, 1.49015e-23, 5.1347471e-24], [266.68587, 1.11984e-22, 2.48502e-23, 1.49101e-23, 5.1473906e-24], [267.30064, 1.12066e-22, 2.48647e-23, 1.49188e-23, 5.1601212e-24], [267.91683, 1.12149e-22, 2.48792e-23, 1.49275e-23, 5.1729391e-24], [268.53444, 1.12231e-22, 2.48938e-23, 1.49363e-23, 5.1858444e-24], [269.15348, 1.12314e-22, 2.49085e-23, 1.49451e-23, 5.1988372e-24], [269.77394, 1.12398e-22, 2.49232e-23, 1.49539e-23, 5.2119177e-24], [270.39584, 1.12481e-22, 2.49381e-23, 1.49629e-23, 5.225086e-24], [271.01916, 1.12565e-22, 2.4953e-23, 1.49718e-23, 5.2383421e-24], [271.64393, 1.12649e-22, 2.49681e-23, 1.49808e-23, 5.2516864e-24], [272.27013, 1.12734e-22, 2.49832e-23, 1.49899e-23, 5.2651188e-24], [272.89778, 1.12819e-22, 2.49984e-23, 1.4999e-23, 5.2786396e-24], [273.52687, 1.12905e-22, 2.50137e-23, 1.50082e-23, 5.2922488e-24], [274.15742, 1.1299e-22, 2.50291e-23, 1.50175e-23, 5.3059466e-24], [274.78942, 1.13076e-22, 2.50446e-23, 1.50267e-23, 5.319733e-24], [275.42287, 1.13163e-22, 2.50601e-23, 1.50361e-23, 5.3336083e-24], [276.05779, 1.1325e-22, 2.50758e-23, 1.50455e-23, 5.3475726e-24], [276.69416, 1.13337e-22, 2.50915e-23, 1.50549e-23, 5.361626e-24], [277.33201, 1.13424e-22, 2.51073e-23, 1.50644e-23, 5.3757685e-24], [277.97133, 1.13512e-22, 2.51232e-23, 1.50739e-23, 5.3900004e-24], [278.61212, 1.136e-22, 2.51392e-23, 1.50835e-23, 5.4043218e-24], [279.25438, 1.13689e-22, 2.51553e-23, 1.50932e-23, 5.4187327e-24], [279.89813, 1.13778e-22, 2.51714e-23, 1.51029e-23, 5.4332333e-24], [280.54336, 1.13867e-22, 2.51877e-23, 1.51126e-23, 5.4478238e-24], [281.19008, 1.13957e-22, 2.5204e-23, 1.51224e-23, 5.4625042e-24], [281.83829, 1.14047e-22, 2.52205e-23, 1.51323e-23, 5.4772747e-24], [282.488, 1.14137e-22, 2.5237e-23, 1.51422e-23, 5.4921353e-24], [283.1392, 1.14228e-22, 2.52536e-23, 1.51522e-23, 5.5070863e-24], [283.7919, 1.14319e-22, 2.52703e-23, 1.51622e-23, 5.5221276e-24], [284.44611, 1.1441e-22, 2.52871e-23, 1.51722e-23, 5.5372596e-24], [285.10183, 1.14502e-22, 2.53039e-23, 1.51824e-23, 5.5524821e-24], [285.75905, 1.14594e-22, 2.53209e-23, 1.51925e-23, 5.5677954e-24], [286.4178, 1.14687e-22, 2.53379e-23, 1.52028e-23, 5.5831996e-24], [287.07806, 1.1478e-22, 2.53551e-23, 1.52131e-23, 5.5986948e-24], [287.73984, 1.14873e-22, 2.53723e-23, 1.52234e-23, 5.6142811e-24], [288.40315, 1.14967e-22, 2.53896e-23, 1.52338e-23, 5.6299586e-24], [289.06799, 1.15061e-22, 2.5407e-23, 1.52442e-23, 5.6457274e-24], [289.73436, 1.15155e-22, 2.54245e-23, 1.52547e-23, 5.6615877e-24], [290.40227, 1.1525e-22, 2.54421e-23, 1.52653e-23, 5.6775395e-24], [291.07171, 1.15345e-22, 2.54598e-23, 1.52759e-23, 5.6935829e-24], [291.7427, 1.15441e-22, 2.54776e-23, 1.52865e-23, 5.7097181e-24], [292.41524, 1.15537e-22, 2.54954e-23, 1.52972e-23, 5.7259452e-24], [293.08932, 1.15633e-22, 2.55134e-23, 1.5308e-23, 5.7422642e-24], [293.76497, 1.1573e-22, 2.55314e-23, 1.53188e-23, 5.7586753e-24], [294.44216, 1.15827e-22, 2.55495e-23, 1.53297e-23, 5.7751785e-24], [295.12092, 1.15925e-22, 2.55677e-23, 1.53406e-23, 5.791774e-24], [295.80125, 1.16022e-22, 2.55861e-23, 1.53516e-23, 5.8084619e-24], [296.48314, 1.16121e-22, 2.56045e-23, 1.53627e-23, 5.8252423e-24], [297.1666, 1.16219e-22, 2.56229e-23, 1.53738e-23, 5.8421152e-24], [297.85164, 1.16318e-22, 2.56415e-23, 1.53849e-23, 5.8590808e-24], [298.53826, 1.16418e-22, 2.56602e-23, 1.53961e-23, 5.8761391e-24], [299.22646, 1.16517e-22, 2.5679e-23, 1.54074e-23, 5.8932903e-24], [299.91625, 1.16618e-22, 2.56978e-23, 1.54187e-23, 5.9105345e-24], [300.60763, 1.16718e-22, 2.57168e-23, 1.54301e-23, 5.9278717e-24], [301.3006, 1.16819e-22, 2.57358e-23, 1.54415e-23, 5.9453021e-24], [301.99517, 1.16921e-22, 2.5755e-23, 1.5453e-23, 5.9628257e-24], [302.69134, 1.17022e-22, 2.57742e-23, 1.54645e-23, 5.9804426e-24], [303.38912, 1.17124e-22, 2.57935e-23, 1.54761e-23, 5.998153e-24], [304.0885, 1.17227e-22, 2.5813e-23, 1.54878e-23, 6.0159569e-24], [304.7895, 1.1733e-22, 2.58325e-23, 1.54995e-23, 6.0338543e-24], [305.49211, 1.17433e-22, 2.58521e-23, 1.55113e-23, 6.0518455e-24], [306.19634, 1.17537e-22, 2.58718e-23, 1.55231e-23, 6.0699305e-24], [306.9022, 1.17641e-22, 2.58916e-23, 1.5535e-23, 6.0881093e-24], [307.60968, 1.17746e-22, 2.59115e-23, 1.55469e-23, 6.1063821e-24], [308.3188, 1.17851e-22, 2.59315e-23, 1.55589e-23, 6.1247489e-24], [309.02954, 1.17956e-22, 2.59516e-23, 1.55709e-23, 6.1432099e-24], [309.74193, 1.18062e-22, 2.59717e-23, 1.5583e-23, 6.1617651e-24], [310.45596, 1.18168e-22, 2.5992e-23, 1.55952e-23, 6.1804146e-24], [311.17163, 1.18275e-22, 2.60124e-23, 1.56074e-23, 6.1991585e-24], [311.88896, 1.18382e-22, 2.60329e-23, 1.56197e-23, 6.2179968e-24], [312.60794, 1.18489e-22, 2.60534e-23, 1.56321e-23, 6.2369298e-24], [313.32857, 1.18597e-22, 2.60741e-23, 1.56445e-23, 6.2559573e-24], [314.05087, 1.18705e-22, 2.60949e-23, 1.56569e-23, 6.2750796e-24], [314.77483, 1.18814e-22, 2.61157e-23, 1.56694e-23, 6.2942967e-24], [315.50046, 1.18923e-22, 2.61367e-23, 1.5682e-23, 6.3136087e-24], [316.22777, 1.19032e-22, 2.61578e-23, 1.56947e-23, 6.3330157e-24], [316.95675, 1.19142e-22, 2.61789e-23, 1.57073e-23, 6.3525178e-24], [317.68741, 1.19253e-22, 2.62002e-23, 1.57201e-23, 6.3721149e-24], [318.41975, 1.19363e-22, 2.62215e-23, 1.57329e-23, 6.3918073e-24], [319.15379, 1.19475e-22, 2.6243e-23, 1.57458e-23, 6.411595e-24], [319.88951, 1.19586e-22, 2.62645e-23, 1.57587e-23, 6.4314781e-24], [320.62693, 1.19698e-22, 2.62862e-23, 1.57717e-23, 6.4514567e-24], [321.36605, 1.19811e-22, 2.6308e-23, 1.57848e-23, 6.4715308e-24], [322.10688, 1.19924e-22, 2.63298e-23, 1.57979e-23, 6.4917006e-24], [322.84941, 1.20037e-22, 2.63518e-23, 1.58111e-23, 6.5119661e-24], [323.59366, 1.20151e-22, 2.63738e-23, 1.58243e-23, 6.5323273e-24], [324.33962, 1.20265e-22, 2.6396e-23, 1.58376e-23, 6.5527845e-24], [325.0873, 1.2038e-22, 2.64183e-23, 1.5851e-23, 6.5733376e-24], [325.8367, 1.20495e-22, 2.64407e-23, 1.58644e-23, 6.5939868e-24], [326.58783, 1.2061e-22, 2.64631e-23, 1.58779e-23, 6.6147321e-24], [327.34069, 1.20726e-22, 2.64857e-23, 1.58914e-23, 6.6355737e-24], [328.09529, 1.20842e-22, 2.65084e-23, 1.5905e-23, 6.6565115e-24], [328.85163, 1.20959e-22, 2.65312e-23, 1.59187e-23, 6.6775458e-24], [329.60971, 1.21076e-22, 2.65541e-23, 1.59325e-23, 6.6986765e-24], [330.36954, 1.21194e-22, 2.65771e-23, 1.59463e-23, 6.7199038e-24], [331.13112, 1.21312e-22, 2.66002e-23, 1.59601e-23, 6.7412278e-24], [331.89446, 1.21431e-22, 2.66234e-23, 1.59741e-23, 6.7626485e-24], [332.65955, 1.2155e-22, 2.66468e-23, 1.59881e-23, 6.7841661e-24], [333.42641, 1.21669e-22, 2.66702e-23, 1.60021e-23, 6.8057806e-24], [334.19504, 1.21789e-22, 2.66937e-23, 1.60162e-23, 6.8274921e-24], [334.96544, 1.2191e-22, 2.67174e-23, 1.60304e-23, 6.8493008e-24], [335.73761, 1.22031e-22, 2.67412e-23, 1.60447e-23, 6.8712068e-24], [336.51157, 1.22152e-22, 2.6765e-23, 1.6059e-23, 6.89321e-24], [337.28731, 1.22274e-22, 2.6789e-23, 1.60734e-23, 6.9153107e-24], [338.06484, 1.22396e-22, 2.68131e-23, 1.60879e-23, 6.937509e-24], [338.84416, 1.22519e-22, 2.68373e-23, 1.61024e-23, 6.959805e-24], [339.62527, 1.22642e-22, 2.68616e-23, 1.6117e-23, 6.9821987e-24], [340.40819, 1.22765e-22, 2.68861e-23, 1.61317e-23, 7.0046903e-24], [341.19291, 1.2289e-22, 2.69106e-23, 1.61464e-23, 7.0272799e-24], [341.97944, 1.23014e-22, 2.69353e-23, 1.61612e-23, 7.0499677e-24], [342.76779, 1.23139e-22, 2.69601e-23, 1.61761e-23, 7.0727537e-24], [343.55795, 1.23265e-22, 2.6985e-23, 1.6191e-23, 7.0956382e-24], [344.34993, 1.23391e-22, 2.701e-23, 1.6206e-23, 7.1186212e-24], [345.14374, 1.23517e-22, 2.70352e-23, 1.62211e-23, 7.1417029e-24], [345.93938, 1.23644e-22, 2.70604e-23, 1.62363e-23, 7.1648834e-24], [346.73685, 1.23771e-22, 2.70858e-23, 1.62515e-23, 7.188163e-24], [347.53616, 1.23899e-22, 2.71113e-23, 1.62668e-23, 7.2115417e-24], [348.33732, 1.24028e-22, 2.7137e-23, 1.62822e-23, 7.2350198e-24], [349.14032, 1.24157e-22, 2.71627e-23, 1.62976e-23, 7.2585974e-24], [349.94517, 1.24286e-22, 2.71886e-23, 1.63132e-23, 7.2822747e-24], [350.75187, 1.24416e-22, 2.72147e-23, 1.63288e-23, 7.306052e-24], [351.56044, 1.24546e-22, 2.72408e-23, 1.63445e-23, 7.3299293e-24], [352.37087, 1.24677e-22, 2.72671e-23, 1.63603e-23, 7.3539071e-24], [353.18317, 1.24808e-22, 2.72935e-23, 1.63761e-23, 7.3779854e-24], [353.99734, 1.2494e-22, 2.73201e-23, 1.6392e-23, 7.4021647e-24], [354.81339, 1.25072e-22, 2.73467e-23, 1.6408e-23, 7.426445e-24], [355.63132, 1.25205e-22, 2.73736e-23, 1.64241e-23, 7.4508268e-24], [356.45113, 1.25339e-22, 2.74005e-23, 1.64403e-23, 7.4753103e-24], [357.27284, 1.25472e-22, 2.74277e-23, 1.64566e-23, 7.4998958e-24], [358.09644, 1.25607e-22, 2.74549e-23, 1.6473e-23, 7.5245838e-24], [358.92193, 1.25742e-22, 2.74823e-23, 1.64894e-23, 7.5493746e-24], [359.74934, 1.25877e-22, 2.75099e-23, 1.65059e-23, 7.5742686e-24], [360.57864, 1.26013e-22, 2.75376e-23, 1.65226e-23, 7.5992662e-24], [361.40986, 1.2615e-22, 2.75655e-23, 1.65393e-23, 7.6243679e-24], [362.243, 1.26287e-22, 2.75936e-23, 1.65561e-23, 7.6495743e-24], [363.07805, 1.26424e-22, 2.76218e-23, 1.65731e-23, 7.6748859e-24], [363.91504, 1.26562e-22, 2.76502e-23, 1.65901e-23, 7.7003033e-24], [364.75395, 1.26701e-22, 2.76787e-23, 1.66072e-23, 7.7258271e-24], [365.59479, 1.2684e-22, 2.77075e-23, 1.66245e-23, 7.7514581e-24], [366.43757, 1.2698e-22, 2.77364e-23, 1.66418e-23, 7.7771971e-24], [367.2823, 1.2712e-22, 2.77656e-23, 1.66593e-23, 7.803045e-24], [368.12897, 1.27261e-22, 2.77949e-23, 1.66769e-23, 7.8290026e-24], [368.9776, 1.27403e-22, 2.78244e-23, 1.66947e-23, 7.8550712e-24], [369.82818, 1.27545e-22, 2.78542e-23, 1.67125e-23, 7.8812517e-24], [370.68072, 1.27688e-22, 2.78842e-23, 1.67305e-23, 7.9075455e-24], [371.53523, 1.27831e-22, 2.79145e-23, 1.67487e-23, 7.933954e-24], [372.39171, 1.27975e-22, 2.7945e-23, 1.6767e-23, 7.9604789e-24], [373.25016, 1.28119e-22, 2.79757e-23, 1.67854e-23, 7.9871218e-24], [374.11059, 1.28265e-22, 2.80068e-23, 1.68041e-23, 8.0138848e-24], [374.973, 1.28411e-22, 2.80381e-23, 1.68229e-23, 8.04077e-24], [375.8374, 1.28557e-22, 2.80698e-23, 1.68419e-23, 8.0677801e-24], [376.7038, 1.28705e-22, 2.81018e-23, 1.68611e-23, 8.0949178e-24], [377.57219, 1.28853e-22, 2.81342e-23, 1.68805e-23, 8.1221863e-24], [378.44258, 1.29001e-22, 2.81669e-23, 1.69001e-23, 8.1495894e-24], [379.31498, 1.29151e-22, 2.82001e-23, 1.69201e-23, 8.1771311e-24], [380.1894, 1.29301e-22, 2.82337e-23, 1.69402e-23, 8.2048163e-24], [381.06582, 1.29453e-22, 2.82679e-23, 1.69607e-23, 8.2326505e-24], [381.94427, 1.29605e-22, 2.83026e-23, 1.69816e-23, 8.2606401e-24], [382.82474, 1.29758e-22, 2.83379e-23, 1.70028e-23, 8.2887924e-24], [383.70725, 1.29912e-22, 2.83739e-23, 1.70243e-23, 8.3171162e-24], [384.59178, 1.30067e-22, 2.84107e-23, 1.70464e-23, 8.3456217e-24], [385.47836, 1.30224e-22, 2.84483e-23, 1.7069e-23, 8.3743207e-24], [386.36698, 1.30381e-22, 2.84868e-23, 1.70921e-23, 8.4032277e-24], [387.25764, 1.30541e-22, 2.85264e-23, 1.71159e-23, 8.4323595e-24], [388.15037, 1.30701e-22, 2.85673e-23, 1.71404e-23, 8.4617368e-24], [389.04514, 1.30863e-22, 2.86097e-23, 1.71658e-23, 8.4913844e-24], [389.94199, 1.31027e-22, 2.86536e-23, 1.71922e-23, 8.5213326e-24], [390.8409, 1.31194e-22, 2.86996e-23, 1.72198e-23, 8.5516189e-24], [391.74188, 1.31362e-22, 2.87479e-23, 1.72487e-23, 8.5822898e-24], [392.64494, 1.31534e-22, 2.87989e-23, 1.72793e-23, 8.613404e-24], [393.55008, 1.31709e-22, 2.88533e-23, 1.7312e-23, 8.6450358e-24], [394.4573, 1.31888e-22, 2.89117e-23, 1.7347e-23, 8.677281e-24], [395.36662, 1.32072e-22, 2.89752e-23, 1.73851e-23, 8.7102646e-24], [396.27803, 1.32262e-22, 2.9045e-23, 1.7427e-23, 8.7441522e-24], [397.19155, 1.32459e-22, 2.91228e-23, 1.74737e-23, 8.779167e-24], [398.10717, 1.32666e-22, 2.92109e-23, 1.75265e-23, 8.8156156e-24], [399.0249, 1.32885e-22, 2.93127e-23, 1.75876e-23, 8.8539287e-24], [399.94475, 1.33122e-22, 2.94328e-23, 1.76597e-23, 8.8947267e-24], [400.86672, 1.33381e-22, 2.95783e-23, 1.7747e-23, 8.9389296e-24], [401.79081, 1.33673e-22, 2.97598e-23, 1.78559e-23, 8.9879493e-24], [402.71703, 1.34014e-22, 2.99942e-23, 1.79965e-23, 9.0440432e-24], [403.64539, 1.34428e-22, 3.03099e-23, 1.81859e-23, 9.1110025e-24], [404.57589, 1.34965e-22, 3.07566e-23, 1.84539e-23, 9.1955954e-24], [405.50854, 1.35714e-22, 3.1429e-23, 1.88574e-23, 9.3108761e-24], [406.44333, 1.36874e-22, 3.25249e-23, 1.9515e-23, 9.4847096e-24], [407.38028, 1.38936e-22, 3.45119e-23, 2.07072e-23, 9.785441e-24], [408.31939, 1.43427e-22, 3.87176e-23, 2.32306e-23, 1.0418466e-23], [409.26066, 1.57151e-22, 5.02125e-23, 3.01275e-23, 1.2243674e-23], [410.2041, 2.4715e-22, 1.07745e-22, 6.46468e-23, 2.2669117e-23], [411.14972, 4.0278e-22, 1.9206e-22, 1.15236e-22, 3.9057526e-23], [412.09752, 1.66959e-22, 5.73461e-23, 3.44076e-23, 1.3492185e-23], [413.0475, 1.46234e-22, 4.07015e-23, 2.44209e-23, 1.0828084e-23], [413.99967, 1.40701e-22, 3.53615e-23, 2.12169e-23, 1.007425e-23], [414.95404, 1.38564e-22, 3.3044e-23, 1.98264e-23, 9.7797171e-24], [415.91061, 1.37583e-22, 3.18566e-23, 1.9114e-23, 9.6466141e-24], [416.86938, 1.37099e-22, 3.11807e-23, 1.87084e-23, 9.5839926e-24], [417.83037, 1.36864e-22, 3.07678e-23, 1.84607e-23, 9.5568481e-24], [418.79357, 1.36766e-22, 3.05033e-23, 1.8302e-23, 9.549458e-24], [419.75898, 1.3675e-22, 3.03286e-23, 1.81971e-23, 9.5539324e-24], [420.72663, 1.36786e-22, 3.02114e-23, 1.81268e-23, 9.5659664e-24], [421.6965, 1.36857e-22, 3.01325e-23, 1.80795e-23, 9.583053e-24], [422.66861, 1.36952e-22, 3.00804e-23, 1.80482e-23, 9.6036531e-24], [423.64297, 1.37065e-22, 3.00472e-23, 1.80283e-23, 9.6267803e-24], [424.61956, 1.3719e-22, 3.00281e-23, 1.80168e-23, 9.6517793e-24], [425.59841, 1.37326e-22, 3.00194e-23, 1.80116e-23, 9.678201e-24], [426.57952, 1.37468e-22, 3.00188e-23, 1.80113e-23, 9.7057297e-24], [427.56289, 1.37617e-22, 3.00244e-23, 1.80147e-23, 9.7341381e-24], [428.54852, 1.37771e-22, 3.0035e-23, 1.8021e-23, 9.7632595e-24], [429.53643, 1.37928e-22, 3.00496e-23, 1.80298e-23, 9.792969e-24], [430.52661, 1.38089e-22, 3.00675e-23, 1.80405e-23, 9.8231721e-24], [431.51908, 1.38253e-22, 3.0088e-23, 1.80528e-23, 9.853796e-24], [432.51383, 1.38419e-22, 3.01107e-23, 1.80664e-23, 9.8847837e-24], [433.51088, 1.38587e-22, 3.01353e-23, 1.80812e-23, 9.9160905e-24], [434.51022, 1.38757e-22, 3.01615e-23, 1.80969e-23, 9.9476807e-24], [435.51187, 1.38929e-22, 3.01891e-23, 1.81135e-23, 9.9795255e-24], [436.51583, 1.39102e-22, 3.02178e-23, 1.81307e-23, 1.0011601e-23], [437.52211, 1.39277e-22, 3.02476e-23, 1.81485e-23, 1.004389e-23], [438.5307, 1.39453e-22, 3.02782e-23, 1.81669e-23, 1.0076374e-23], [439.54162, 1.3963e-22, 3.03097e-23, 1.81858e-23, 1.0109042e-23], [440.55486, 1.39808e-22, 3.03418e-23, 1.82051e-23, 1.0141882e-23], [441.57045, 1.39987e-22, 3.03746e-23, 1.82248e-23, 1.0174886e-23], [442.58837, 1.40168e-22, 3.0408e-23, 1.82448e-23, 1.0208045e-23], [443.60864, 1.40349e-22, 3.04419e-23, 1.82651e-23, 1.0241353e-23], [444.63127, 1.40531e-22, 3.04763e-23, 1.82858e-23, 1.0274804e-23], [445.65625, 1.40714e-22, 3.05111e-23, 1.83066e-23, 1.0308394e-23], [446.68359, 1.40898e-22, 3.05463e-23, 1.83278e-23, 1.0342118e-23], [447.7133, 1.41082e-22, 3.05819e-23, 1.83491e-23, 1.0375973e-23], [448.74539, 1.41268e-22, 3.06178e-23, 1.83707e-23, 1.0409955e-23], [449.77985, 1.41454e-22, 3.06541e-23, 1.83924e-23, 1.0444062e-23], [450.8167, 1.41641e-22, 3.06907e-23, 1.84144e-23, 1.0478292e-23], [451.85594, 1.41829e-22, 3.07275e-23, 1.84365e-23, 1.0512643e-23], [452.89758, 1.42018e-22, 3.07647e-23, 1.84588e-23, 1.0547112e-23], [453.94162, 1.42207e-22, 3.08021e-23, 1.84813e-23, 1.0581698e-23], [454.98806, 1.42397e-22, 3.08398e-23, 1.85039e-23, 1.0616399e-23], [456.03692, 1.42588e-22, 3.08777e-23, 1.85266e-23, 1.0651216e-23], [457.08819, 1.42779e-22, 3.09159e-23, 1.85495e-23, 1.0686145e-23], [458.14189, 1.42971e-22, 3.09542e-23, 1.85725e-23, 1.0721187e-23], [459.19801, 1.43164e-22, 3.09928e-23, 1.85957e-23, 1.075634e-23], [460.25657, 1.43358e-22, 3.10317e-23, 1.8619e-23, 1.0791604e-23], [461.31757, 1.43552e-22, 3.10707e-23, 1.86424e-23, 1.0826978e-23], [462.38102, 1.43747e-22, 3.11099e-23, 1.86659e-23, 1.0862462e-23], [463.44692, 1.43943e-22, 3.11493e-23, 1.86896e-23, 1.0898054e-23], [464.51528, 1.44139e-22, 3.1189e-23, 1.87134e-23, 1.0933755e-23], [465.58609, 1.44336e-22, 3.12288e-23, 1.87373e-23, 1.0969563e-23], [466.65938, 1.44534e-22, 3.12688e-23, 1.87613e-23, 1.100548e-23], [467.73514, 1.44732e-22, 3.1309e-23, 1.87854e-23, 1.1041503e-23], [468.81338, 1.44931e-22, 3.13493e-23, 1.88096e-23, 1.1077633e-23], [469.89411, 1.45131e-22, 3.13899e-23, 1.88339e-23, 1.111387e-23], [470.97733, 1.45332e-22, 3.14306e-23, 1.88584e-23, 1.1150213e-23], [472.06304, 1.45533e-22, 3.14715e-23, 1.88829e-23, 1.1186662e-23], [473.15126, 1.45735e-22, 3.15126e-23, 1.89076e-23, 1.1223217e-23], [474.24199, 1.45937e-22, 3.15539e-23, 1.89323e-23, 1.1259877e-23], [475.33523, 1.4614e-22, 3.15953e-23, 1.89572e-23, 1.1296643e-23], [476.43099, 1.46344e-22, 3.16369e-23, 1.89821e-23, 1.1333515e-23], [477.52927, 1.46549e-22, 3.16786e-23, 1.90072e-23, 1.1370492e-23], [478.63009, 1.46754e-22, 3.17205e-23, 1.90323e-23, 1.1407574e-23], [479.73345, 1.4696e-22, 3.17626e-23, 1.90576e-23, 1.1444761e-23], [480.83935, 1.47167e-22, 3.18049e-23, 1.90829e-23, 1.1482053e-23], [481.9478, 1.47374e-22, 3.18473e-23, 1.91084e-23, 1.151945e-23], [483.0588, 1.47582e-22, 3.18899e-23, 1.91339e-23, 1.1556952e-23], [484.17237, 1.47791e-22, 3.19326e-23, 1.91596e-23, 1.1594559e-23], [485.2885, 1.48e-22, 3.19755e-23, 1.91853e-23, 1.163227e-23], [486.40721, 1.4821e-22, 3.20186e-23, 1.92112e-23, 1.1670087e-23], [487.52849, 1.48421e-22, 3.20618e-23, 1.92371e-23, 1.1708008e-23], [488.65236, 1.48632e-22, 3.21052e-23, 1.92631e-23, 1.1746033e-23], [489.77882, 1.48844e-22, 3.21488e-23, 1.92893e-23, 1.1784163e-23], [490.90788, 1.49057e-22, 3.21925e-23, 1.93155e-23, 1.1822398e-23], [492.03954, 1.4927e-22, 3.22364e-23, 1.93418e-23, 1.1860738e-23], [493.1738, 1.49485e-22, 3.22804e-23, 1.93682e-23, 1.1899182e-23], [494.31069, 1.497e-22, 3.23246e-23, 1.93948e-23, 1.1937731e-23], [495.45019, 1.49915e-22, 3.2369e-23, 1.94214e-23, 1.1976385e-23], [496.59232, 1.50131e-22, 3.24135e-23, 1.94481e-23, 1.2015143e-23], [497.73708, 1.50348e-22, 3.24581e-23, 1.94749e-23, 1.2054007e-23], [498.88449, 1.50566e-22, 3.2503e-23, 1.95018e-23, 1.2092974e-23], [500.03453, 1.50784e-22, 3.2548e-23, 1.95288e-23, 1.2132047e-23], [501.18723, 1.51003e-22, 3.25931e-23, 1.95559e-23, 1.2171225e-23], [502.34259, 1.51223e-22, 3.26384e-23, 1.95831e-23, 1.2210507e-23], [503.50061, 1.51444e-22, 3.26839e-23, 1.96103e-23, 1.2249894e-23], [504.6613, 1.51665e-22, 3.27295e-23, 1.96377e-23, 1.2289386e-23], [505.82466, 1.51887e-22, 3.27753e-23, 1.96652e-23, 1.2328983e-23], [506.99071, 1.52109e-22, 3.28213e-23, 1.96928e-23, 1.2368686e-23], [508.15944, 1.52333e-22, 3.28674e-23, 1.97204e-23, 1.2408493e-23], [509.33087, 1.52557e-22, 3.29136e-23, 1.97482e-23, 1.2448405e-23], [510.505, 1.52782e-22, 3.29601e-23, 1.9776e-23, 1.2488423e-23], [511.68184, 1.53007e-22, 3.30067e-23, 1.9804e-23, 1.2528546e-23], [512.86138, 1.53233e-22, 3.30534e-23, 1.9832e-23, 1.2568774e-23], [514.04365, 1.5346e-22, 3.31003e-23, 1.98602e-23, 1.2609107e-23], [515.22864, 1.53688e-22, 3.31474e-23, 1.98884e-23, 1.2649546e-23], [516.41637, 1.53916e-22, 3.31946e-23, 1.99168e-23, 1.269009e-23], [517.60683, 1.54145e-22, 3.3242e-23, 1.99452e-23, 1.273074e-23], [518.80004, 1.54375e-22, 3.32895e-23, 1.99737e-23, 1.2771496e-23], [519.996, 1.54605e-22, 3.33373e-23, 2.00024e-23, 1.2812357e-23], [521.19471, 1.54837e-22, 3.33851e-23, 2.00311e-23, 1.2853324e-23], [522.39619, 1.55068e-22, 3.34332e-23, 2.00599e-23, 1.2894396e-23], [523.60044, 1.55301e-22, 3.34814e-23, 2.00888e-23, 1.2935575e-23], [524.80746, 1.55535e-22, 3.35297e-23, 2.01178e-23, 1.297686e-23], [526.01727, 1.55769e-22, 3.35782e-23, 2.01469e-23, 1.301825e-23], [527.22986, 1.56004e-22, 3.36269e-23, 2.01762e-23, 1.3059747e-23], [528.44525, 1.56239e-22, 3.36758e-23, 2.02055e-23, 1.310135e-23], [529.66344, 1.56476e-22, 3.37248e-23, 2.02349e-23, 1.3143059e-23], [530.88444, 1.56713e-22, 3.3774e-23, 2.02644e-23, 1.3184875e-23], [532.10826, 1.56951e-22, 3.38233e-23, 2.0294e-23, 1.3226797e-23], [533.3349, 1.57189e-22, 3.38728e-23, 2.03237e-23, 1.3268825e-23], [534.56436, 1.57428e-22, 3.39225e-23, 2.03535e-23, 1.331096e-23], [535.79666, 1.57669e-22, 3.39723e-23, 2.03834e-23, 1.3353202e-23], [537.0318, 1.57909e-22, 3.40223e-23, 2.04134e-23, 1.3395551e-23], [538.26978, 1.58151e-22, 3.40725e-23, 2.04435e-23, 1.3438006e-23], [539.51062, 1.58393e-22, 3.41228e-23, 2.04737e-23, 1.3480568e-23], [540.75432, 1.58636e-22, 3.41733e-23, 2.0504e-23, 1.3523238e-23], [542.00089, 1.5888e-22, 3.42239e-23, 2.05344e-23, 1.3566014e-23], [543.25033, 1.59125e-22, 3.42748e-23, 2.05649e-23, 1.3608898e-23], [544.50265, 1.5937e-22, 3.43258e-23, 2.05955e-23, 1.3651889e-23], [545.75786, 1.59616e-22, 3.43769e-23, 2.06262e-23, 1.3694988e-23], [547.01596, 1.59863e-22, 3.44282e-23, 2.06569e-23, 1.3738194e-23], [548.27696, 1.60111e-22, 3.44797e-23, 2.06878e-23, 1.3781507e-23], [549.54087, 1.60359e-22, 3.45314e-23, 2.07188e-23, 1.3824929e-23], [550.8077, 1.60608e-22, 3.45832e-23, 2.07499e-23, 1.3868458e-23], [552.07744, 1.60858e-22, 3.46352e-23, 2.07811e-23, 1.3912095e-23], [553.35011, 1.61109e-22, 3.46874e-23, 2.08124e-23, 1.395584e-23], [554.62571, 1.6136e-22, 3.47397e-23, 2.08438e-23, 1.3999693e-23], [555.90426, 1.61613e-22, 3.47922e-23, 2.08753e-23, 1.4043654e-23], [557.18575, 1.61866e-22, 3.48449e-23, 2.0907e-23, 1.4087724e-23], [558.47019, 1.62119e-22, 3.48978e-23, 2.09387e-23, 1.4131901e-23], [559.7576, 1.62374e-22, 3.49508e-23, 2.09705e-23, 1.4176188e-23], [561.04798, 1.62629e-22, 3.5004e-23, 2.10024e-23, 1.4220583e-23], [562.34133, 1.62885e-22, 3.50573e-23, 2.10344e-23, 1.4265087e-23], [563.63766, 1.63142e-22, 3.51109e-23, 2.10665e-23, 1.4309699e-23], [564.93697, 1.634e-22, 3.51646e-23, 2.10987e-23, 1.4354421e-23], [566.23929, 1.63658e-22, 3.52184e-23, 2.11311e-23, 1.4399251e-23], [567.54461, 1.63918e-22, 3.52725e-23, 2.11635e-23, 1.4444191e-23], [568.85293, 1.64178e-22, 3.53267e-23, 2.1196e-23, 1.448924e-23], [570.16427, 1.64439e-22, 3.53811e-23, 2.12287e-23, 1.4534399e-23], [571.47864, 1.647e-22, 3.54357e-23, 2.12614e-23, 1.4579666e-23], [572.79603, 1.64963e-22, 3.54904e-23, 2.12942e-23, 1.4625044e-23], [574.11646, 1.65226e-22, 3.55453e-23, 2.13272e-23, 1.4670531e-23], [575.43994, 1.6549e-22, 3.56004e-23, 2.13603e-23, 1.4716128e-23], [576.76646, 1.65755e-22, 3.56557e-23, 2.13934e-23, 1.4761835e-23], [578.09605, 1.6602e-22, 3.57111e-23, 2.14267e-23, 1.4807652e-23], [579.4287, 1.66287e-22, 3.57668e-23, 2.14601e-23, 1.4853579e-23], [580.76442, 1.66554e-22, 3.58225e-23, 2.14935e-23, 1.4899616e-23], [582.10322, 1.66822e-22, 3.58785e-23, 2.15271e-23, 1.4945764e-23], [583.4451, 1.67091e-22, 3.59347e-23, 2.15608e-23, 1.4992023e-23], [584.79008, 1.67361e-22, 3.5991e-23, 2.15946e-23, 1.5038392e-23], [586.13816, 1.67631e-22, 3.60475e-23, 2.16285e-23, 1.5084871e-23], [587.48935, 1.67903e-22, 3.61042e-23, 2.16625e-23, 1.5131462e-23], [588.84366, 1.68175e-22, 3.6161e-23, 2.16966e-23, 1.5178164e-23], [590.20108, 1.68448e-22, 3.62181e-23, 2.17308e-23, 1.5224977e-23], [591.56163, 1.68721e-22, 3.62753e-23, 2.17652e-23, 1.5271901e-23], [592.92532, 1.68996e-22, 3.63327e-23, 2.17996e-23, 1.5318937e-23], [594.29216, 1.69271e-22, 3.63903e-23, 2.18342e-23, 1.5366084e-23], [595.66214, 1.69548e-22, 3.6448e-23, 2.18688e-23, 1.5413343e-23], [597.03529, 1.69825e-22, 3.6506e-23, 2.19036e-23, 1.5460713e-23], [598.4116, 1.70103e-22, 3.65641e-23, 2.19385e-23, 1.5508196e-23], [599.79108, 1.70381e-22, 3.66224e-23, 2.19734e-23, 1.5555791e-23], [601.17374, 1.70661e-22, 3.66809e-23, 2.20085e-23, 1.5603497e-23], [602.55959, 1.70941e-22, 3.67396e-23, 2.20437e-23, 1.5651317e-23], [603.94863, 1.71222e-22, 3.67984e-23, 2.2079e-23, 1.5699248e-23], [605.34087, 1.71505e-22, 3.68574e-23, 2.21145e-23, 1.5747292e-23], [606.73633, 1.71787e-22, 3.69167e-23, 2.215e-23, 1.5795449e-23], [608.135, 1.72071e-22, 3.69761e-23, 2.21856e-23, 1.5843719e-23], [609.5369, 1.72356e-22, 3.70356e-23, 2.22214e-23, 1.5892102e-23], [610.94202, 1.72641e-22, 3.70954e-23, 2.22573e-23, 1.5940598e-23], [612.35039, 1.72928e-22, 3.71554e-23, 2.22932e-23, 1.5989207e-23], [613.76201, 1.73215e-22, 3.72155e-23, 2.23293e-23, 1.603793e-23], [615.17687, 1.73503e-22, 3.72758e-23, 2.23655e-23, 1.6086766e-23], [616.595, 1.73792e-22, 3.73364e-23, 2.24018e-23, 1.6135717e-23], [618.0164, 1.74081e-22, 3.73971e-23, 2.24382e-23, 1.618478e-23], [619.44108, 1.74372e-22, 3.7458e-23, 2.24748e-23, 1.6233958e-23], [620.86903, 1.74663e-22, 3.7519e-23, 2.25114e-23, 1.6283251e-23], [622.30029, 1.74956e-22, 3.75803e-23, 2.25482e-23, 1.6332657e-23], [623.73484, 1.75249e-22, 3.76418e-23, 2.25851e-23, 1.6382178e-23], [625.17269, 1.75543e-22, 3.77034e-23, 2.2622e-23, 1.6431814e-23], [626.61386, 1.75838e-22, 3.77653e-23, 2.26592e-23, 1.6481564e-23], [628.05836, 1.76133e-22, 3.78273e-23, 2.26964e-23, 1.6531429e-23], [629.50618, 1.7643e-22, 3.78895e-23, 2.27337e-23, 1.658141e-23], [630.95734, 1.76728e-22, 3.79519e-23, 2.27711e-23, 1.6631506e-23], [632.41185, 1.77026e-22, 3.80145e-23, 2.28087e-23, 1.6681717e-23], [633.86971, 1.77325e-22, 3.80773e-23, 2.28464e-23, 1.6732044e-23], [635.33093, 1.77625e-22, 3.81403e-23, 2.28842e-23, 1.6782486e-23], [636.79552, 1.77926e-22, 3.82035e-23, 2.29221e-23, 1.6833044e-23], [638.26349, 1.78228e-22, 3.82669e-23, 2.29601e-23, 1.6883719e-23], [639.73484, 1.78531e-22, 3.83304e-23, 2.29983e-23, 1.693451e-23], [641.20958, 1.78835e-22, 3.83942e-23, 2.30365e-23, 1.6985417e-23], [642.68772, 1.79139e-22, 3.84582e-23, 2.30749e-23, 1.7036441e-23], [644.16927, 1.79445e-22, 3.85223e-23, 2.31134e-23, 1.7087581e-23], [645.65423, 1.79751e-22, 3.85867e-23, 2.3152e-23, 1.7138839e-23], [647.14262, 1.80058e-22, 3.86512e-23, 2.31907e-23, 1.7190213e-23], [648.63443, 1.80366e-22, 3.8716e-23, 2.32296e-23, 1.7241705e-23], [650.12969, 1.80675e-22, 3.87809e-23, 2.32686e-23, 1.7293315e-23], [651.62839, 1.80985e-22, 3.88461e-23, 2.33077e-23, 1.7345042e-23], [653.13055, 1.81296e-22, 3.89115e-23, 2.33469e-23, 1.7396887e-23], [654.63617, 1.81608e-22, 3.8977e-23, 2.33862e-23, 1.744885e-23], [656.14527, 1.81921e-22, 3.90428e-23, 2.34257e-23, 1.7500932e-23], [657.65784, 1.82234e-22, 3.91087e-23, 2.34652e-23, 1.7553132e-23], [659.1739, 1.82549e-22, 3.91749e-23, 2.3505e-23, 1.7605451e-23], [660.69345, 1.82864e-22, 3.92413e-23, 2.35448e-23, 1.7657889e-23], [662.2165, 1.83181e-22, 3.93079e-23, 2.35847e-23, 1.7710446e-23], [663.74307, 1.83498e-22, 3.93747e-23, 2.36248e-23, 1.7763123e-23], [665.27316, 1.83816e-22, 3.94417e-23, 2.3665e-23, 1.7815919e-23], [666.80677, 1.84135e-22, 3.95089e-23, 2.37053e-23, 1.7868835e-23], [668.34392, 1.84455e-22, 3.95763e-23, 2.37458e-23, 1.7921871e-23], [669.88461, 1.84776e-22, 3.9644e-23, 2.37864e-23, 1.7975028e-23], [671.42885, 1.85098e-22, 3.97118e-23, 2.38271e-23, 1.8028306e-23], [672.97666, 1.85421e-22, 3.97799e-23, 2.38679e-23, 1.8081705e-23], [674.52803, 1.85745e-22, 3.98482e-23, 2.39089e-23, 1.8135226e-23], [676.08298, 1.8607e-22, 3.99167e-23, 2.395e-23, 1.8188868e-23], [677.64151, 1.86395e-22, 3.99855e-23, 2.39913e-23, 1.8242633e-23], [679.20363, 1.86722e-22, 4.00544e-23, 2.40327e-23, 1.829652e-23], [680.76936, 1.8705e-22, 4.01237e-23, 2.40742e-23, 1.835053e-23], [682.33869, 1.87378e-22, 4.01931e-23, 2.41159e-23, 1.8404665e-23], [683.91165, 1.87708e-22, 4.02628e-23, 2.41577e-23, 1.8458923e-23], [685.48823, 1.88038e-22, 4.03327e-23, 2.41996e-23, 1.8513306e-23], [687.06844, 1.8837e-22, 4.04029e-23, 2.42418e-23, 1.8567815e-23], [688.6523, 1.88703e-22, 4.04734e-23, 2.4284e-23, 1.862245e-23], [690.2398, 1.89036e-22, 4.05441e-23, 2.43265e-23, 1.8677213e-23], [691.83097, 1.89371e-22, 4.06152e-23, 2.43691e-23, 1.8732104e-23], [693.42581, 1.89706e-22, 4.06865e-23, 2.44119e-23, 1.8787125e-23], [695.02432, 1.90043e-22, 4.07581e-23, 2.44549e-23, 1.8842277e-23], [696.62651, 1.90381e-22, 4.08301e-23, 2.44981e-23, 1.8897563e-23], [698.2324, 1.9072e-22, 4.09024e-23, 2.45415e-23, 1.8952983e-23], [699.842, 1.91059e-22, 4.09751e-23, 2.45851e-23, 1.9008542e-23], [701.4553, 1.91401e-22, 4.10483e-23, 2.4629e-23, 1.9064242e-23], [703.07232, 1.91743e-22, 4.11219e-23, 2.46731e-23, 1.9120088e-23], [704.69307, 1.92086e-22, 4.11961e-23, 2.47176e-23, 1.9176085e-23], [706.31755, 1.92431e-22, 4.12708e-23, 2.47625e-23, 1.9232241e-23], [707.94578, 1.92778e-22, 4.13464e-23, 2.48078e-23, 1.9288565e-23], [709.57777, 1.93125e-22, 4.14227e-23, 2.48536e-23, 1.9345069e-23], [711.21351, 1.93475e-22, 4.15002e-23, 2.49001e-23, 1.9401772e-23], [712.85303, 1.93826e-22, 4.15791e-23, 2.49475e-23, 1.9458699e-23], [714.49633, 1.9418e-22, 4.16598e-23, 2.49959e-23, 1.9515885e-23], [716.14341, 1.94537e-22, 4.17429e-23, 2.50458e-23, 1.9573387e-23], [717.79429, 1.94898e-22, 4.18295e-23, 2.50977e-23, 1.963129e-23], [719.44898, 1.95263e-22, 4.19213e-23, 2.51528e-23, 1.9689733e-23], [721.10748, 1.95637e-22, 4.2021e-23, 2.52126e-23, 1.9748958e-23], [722.7698, 1.96022e-22, 4.21339e-23, 2.52803e-23, 1.9809412e-23], [724.43596, 1.96429e-22, 4.22707e-23, 2.53624e-23, 1.9871998e-23], [726.10596, 1.96879e-22, 4.24556e-23, 2.54733e-23, 1.9938784e-23], [727.7798, 1.97428e-22, 4.27544e-23, 2.56527e-23, 2.0015402e-23], [729.45751, 1.98282e-22, 4.34016e-23, 2.6041e-23, 2.0122123e-23], [731.13908, 2.00707e-22, 4.57893e-23, 2.74736e-23, 2.0383369e-23], [732.82453, 2.52855e-22, 8.93346e-23, 5.36008e-23, 2.5552234e-23], [734.51387, 2.07817e-22, 5.25794e-23, 3.15476e-23, 2.1127193e-23], [736.2071, 2.00407e-22, 4.44867e-23, 2.6692e-23, 2.0421377e-23], [737.90423, 1.99766e-22, 4.34219e-23, 2.60531e-23, 2.0381237e-23], [739.60528, 1.99819e-22, 4.31404e-23, 2.58843e-23, 2.040925e-23], [741.31024, 2.00048e-22, 4.30612e-23, 2.58367e-23, 2.0454636e-23], [743.01914, 2.00342e-22, 4.30566e-23, 2.5834e-23, 2.050647e-23], [744.73197, 2.00666e-22, 4.3086e-23, 2.58516e-23, 2.0561282e-23], [746.44876, 2.01007e-22, 4.3133e-23, 2.58798e-23, 2.0617691e-23], [748.1695, 2.01357e-22, 4.31902e-23, 2.59141e-23, 2.0675063e-23], [749.89421, 2.01713e-22, 4.32535e-23, 2.59521e-23, 2.073307e-23], [751.62289, 2.02074e-22, 4.3321e-23, 2.59926e-23, 2.0791533e-23], [753.35556, 2.02437e-22, 4.33913e-23, 2.60348e-23, 2.0850344e-23], [755.09223, 2.02804e-22, 4.34636e-23, 2.60782e-23, 2.0909436e-23], [756.8329, 2.03172e-22, 4.35375e-23, 2.61225e-23, 2.0968768e-23], [758.57758, 2.03543e-22, 4.36125e-23, 2.61675e-23, 2.102831e-23], [760.32628, 2.03915e-22, 4.36885e-23, 2.62131e-23, 2.1088043e-23], [762.07901, 2.04288e-22, 4.37653e-23, 2.62592e-23, 2.1147952e-23], [763.83578, 2.04663e-22, 4.38427e-23, 2.63056e-23, 2.1208029e-23], [765.59661, 2.0504e-22, 4.39207e-23, 2.63524e-23, 2.1268264e-23], [767.36149, 2.05417e-22, 4.39992e-23, 2.63995e-23, 2.1328653e-23], [769.13044, 2.05796e-22, 4.40781e-23, 2.64469e-23, 2.1389192e-23], [770.90347, 2.06176e-22, 4.41574e-23, 2.64945e-23, 2.1449878e-23], [772.68059, 2.06558e-22, 4.42371e-23, 2.65423e-23, 2.1510708e-23], [774.4618, 2.0694e-22, 4.43172e-23, 2.65903e-23, 2.157168e-23], [776.24712, 2.07324e-22, 4.43976e-23, 2.66386e-23, 2.1632793e-23], [778.03655, 2.07708e-22, 4.44783e-23, 2.6687e-23, 2.1694046e-23], [779.83011, 2.08094e-22, 4.45593e-23, 2.67356e-23, 2.1755439e-23], [781.6278, 2.08481e-22, 4.46406e-23, 2.67844e-23, 2.1816969e-23], [783.42964, 2.08869e-22, 4.47222e-23, 2.68333e-23, 2.1878637e-23], [785.23563, 2.09258e-22, 4.48041e-23, 2.68824e-23, 2.1940443e-23], [787.04579, 2.09648e-22, 4.48862e-23, 2.69317e-23, 2.2002386e-23], [788.86012, 2.1004e-22, 4.49686e-23, 2.69812e-23, 2.2064466e-23], [790.67863, 2.10432e-22, 4.50513e-23, 2.70308e-23, 2.2126683e-23], [792.50133, 2.10826e-22, 4.51342e-23, 2.70805e-23, 2.2189036e-23], [794.32823, 2.1122e-22, 4.52174e-23, 2.71305e-23, 2.2251527e-23], [796.15935, 2.11616e-22, 4.53009e-23, 2.71805e-23, 2.2314154e-23], [797.99469, 2.12013e-22, 4.53846e-23, 2.72307e-23, 2.2376918e-23], [799.83426, 2.12411e-22, 4.54685e-23, 2.72811e-23, 2.2439819e-23], [801.67806, 2.1281e-22, 4.55527e-23, 2.73316e-23, 2.2502857e-23], [803.52612, 2.1321e-22, 4.56372e-23, 2.73823e-23, 2.2566032e-23], [805.37844, 2.13612e-22, 4.57219e-23, 2.74331e-23, 2.2629344e-23], [807.23503, 2.14014e-22, 4.58068e-23, 2.74841e-23, 2.2692793e-23], [809.0959, 2.14418e-22, 4.5892e-23, 2.75352e-23, 2.2756381e-23], [810.96106, 2.14822e-22, 4.59775e-23, 2.75865e-23, 2.2820106e-23], [812.83052, 2.15228e-22, 4.60632e-23, 2.76379e-23, 2.2883968e-23], [814.70428, 2.15635e-22, 4.61491e-23, 2.76895e-23, 2.2947969e-23], [816.58237, 2.16043e-22, 4.62353e-23, 2.77412e-23, 2.3012109e-23], [818.46479, 2.16452e-22, 4.63217e-23, 2.7793e-23, 2.3076387e-23], [820.35154, 2.16862e-22, 4.64084e-23, 2.7845e-23, 2.3140803e-23], [822.24265, 2.17273e-22, 4.64953e-23, 2.78972e-23, 2.3205359e-23], [824.13812, 2.17686e-22, 4.65825e-23, 2.79495e-23, 2.3270054e-23], [826.03795, 2.18099e-22, 4.66699e-23, 2.80019e-23, 2.3334888e-23], [827.94216, 2.18514e-22, 4.67576e-23, 2.80545e-23, 2.3399862e-23], [829.85077, 2.1893e-22, 4.68455e-23, 2.81073e-23, 2.3464975e-23], [831.76377, 2.19347e-22, 4.69336e-23, 2.81602e-23, 2.3530229e-23], [833.68118, 2.19765e-22, 4.7022e-23, 2.82132e-23, 2.3595624e-23], [835.60302, 2.20184e-22, 4.71107e-23, 2.82664e-23, 2.3661159e-23], [837.52928, 2.20605e-22, 4.71996e-23, 2.83197e-23, 2.3726834e-23], [839.45999, 2.21026e-22, 4.72887e-23, 2.83732e-23, 2.3792651e-23], [841.39514, 2.21449e-22, 4.73781e-23, 2.84269e-23, 2.385861e-23], [843.33476, 2.21872e-22, 4.74677e-23, 2.84806e-23, 2.3924709e-23], [845.27885, 2.22297e-22, 4.75576e-23, 2.85346e-23, 2.3990951e-23], [847.22741, 2.22724e-22, 4.76478e-23, 2.85887e-23, 2.4057335e-23], [849.18048, 2.23151e-22, 4.77381e-23, 2.86429e-23, 2.4123862e-23], [851.13804, 2.23579e-22, 4.78288e-23, 2.86973e-23, 2.4190531e-23], [853.10011, 2.24009e-22, 4.79197e-23, 2.87518e-23, 2.4257343e-23], [855.06671, 2.24439e-22, 4.80108e-23, 2.88065e-23, 2.4324299e-23], [857.03785, 2.24871e-22, 4.81022e-23, 2.88613e-23, 2.4391397e-23], [859.01352, 2.25304e-22, 4.81938e-23, 2.89163e-23, 2.445864e-23], [860.99375, 2.25738e-22, 4.82857e-23, 2.89714e-23, 2.4526027e-23], [862.97855, 2.26174e-22, 4.83779e-23, 2.90267e-23, 2.4593558e-23], [864.96792, 2.2661e-22, 4.84703e-23, 2.90822e-23, 2.4661234e-23], [866.96188, 2.27048e-22, 4.85629e-23, 2.91378e-23, 2.4729054e-23], [868.96043, 2.27487e-22, 4.86558e-23, 2.91935e-23, 2.479702e-23], [870.96359, 2.27927e-22, 4.8749e-23, 2.92494e-23, 2.4865131e-23], [872.97137, 2.28368e-22, 4.88424e-23, 2.93054e-23, 2.4933388e-23], [874.98378, 2.28811e-22, 4.89361e-23, 2.93616e-23, 2.5001791e-23], [877.00082, 2.29254e-22, 4.903e-23, 2.9418e-23, 2.5070341e-23], [879.02252, 2.29699e-22, 4.91242e-23, 2.94745e-23, 2.5139037e-23], [881.04887, 2.30145e-22, 4.92186e-23, 2.95312e-23, 2.520788e-23], [883.0799, 2.30592e-22, 4.93133e-23, 2.9588e-23, 2.527687e-23], [885.11561, 2.31041e-22, 4.94083e-23, 2.9645e-23, 2.5346007e-23], [887.15601, 2.3149e-22, 4.95035e-23, 2.97021e-23, 2.5415292e-23], [889.20112, 2.31941e-22, 4.95989e-23, 2.97594e-23, 2.5484726e-23], [891.25094, 2.32393e-22, 4.96947e-23, 2.98168e-23, 2.5554307e-23], [893.30548, 2.32846e-22, 4.97907e-23, 2.98744e-23, 2.5624038e-23], [895.36477, 2.33301e-22, 4.98869e-23, 2.99322e-23, 2.5693917e-23], [897.42879, 2.33756e-22, 4.99834e-23, 2.99901e-23, 2.5763945e-23], [899.49758, 2.34213e-22, 5.00802e-23, 3.00481e-23, 2.5834124e-23], [901.57114, 2.34671e-22, 5.01772e-23, 3.01063e-23, 2.5904451e-23], [903.64947, 2.3513e-22, 5.02745e-23, 3.01647e-23, 2.597493e-23], [905.7326, 2.35591e-22, 5.03721e-23, 3.02233e-23, 2.6045558e-23], [907.82053, 2.36052e-22, 5.04699e-23, 3.02819e-23, 2.6116337e-23], [909.91327, 2.36515e-22, 5.0568e-23, 3.03408e-23, 2.6187268e-23], [912.01084, 2.3698e-22, 5.06664e-23, 3.03998e-23, 2.625835e-23], [914.11324, 2.37445e-22, 5.0765e-23, 3.0459e-23, 2.6329583e-23], [916.22049, 2.37912e-22, 5.08639e-23, 3.05183e-23, 2.6400969e-23], [918.3326, 2.38379e-22, 5.0963e-23, 3.05778e-23, 2.6472506e-23], [920.44957, 2.38848e-22, 5.10624e-23, 3.06375e-23, 2.6544197e-23], [922.57143, 2.39319e-22, 5.11621e-23, 3.06973e-23, 2.661604e-23], [924.69817, 2.3979e-22, 5.12621e-23, 3.07572e-23, 2.6688037e-23], [926.82982, 2.40263e-22, 5.13623e-23, 3.08174e-23, 2.6760187e-23], [928.96639, 2.40737e-22, 5.14628e-23, 3.08777e-23, 2.6832491e-23], [931.10788, 2.41213e-22, 5.15635e-23, 3.09381e-23, 2.6904949e-23], [933.2543, 2.41689e-22, 5.16645e-23, 3.09987e-23, 2.6977562e-23], [935.40567, 2.42167e-22, 5.17658e-23, 3.10595e-23, 2.705033e-23], [937.56201, 2.42646e-22, 5.18674e-23, 3.11204e-23, 2.7123252e-23], [939.72331, 2.43127e-22, 5.19693e-23, 3.11816e-23, 2.719633e-23], [941.8896, 2.43608e-22, 5.20714e-23, 3.12428e-23, 2.7269565e-23], [944.06088, 2.44091e-22, 5.21738e-23, 3.13043e-23, 2.7342955e-23], [946.23716, 2.44575e-22, 5.22764e-23, 3.13658e-23, 2.7416501e-23], [948.41846, 2.45061e-22, 5.23793e-23, 3.14276e-23, 2.7490205e-23], [950.60479, 2.45547e-22, 5.24826e-23, 3.14895e-23, 2.7564065e-23], [952.79616, 2.46035e-22, 5.2586e-23, 3.15516e-23, 2.7638083e-23], [954.99259, 2.46525e-22, 5.26898e-23, 3.16139e-23, 2.7712259e-23], [957.19407, 2.47015e-22, 5.27938e-23, 3.16763e-23, 2.7786593e-23], [959.40063, 2.47507e-22, 5.28981e-23, 3.17389e-23, 2.7861085e-23], [961.61228, 2.48e-22, 5.30027e-23, 3.18016e-23, 2.7935736e-23], [963.82902, 2.48495e-22, 5.31076e-23, 3.18646e-23, 2.8010546e-23], [966.05088, 2.4899e-22, 5.32127e-23, 3.19276e-23, 2.8085516e-23], [968.27786, 2.49487e-22, 5.33182e-23, 3.19909e-23, 2.8160645e-23], [970.50997, 2.49986e-22, 5.34239e-23, 3.20543e-23, 2.8235935e-23], [972.74722, 2.50485e-22, 5.35299e-23, 3.21179e-23, 2.8311385e-23], [974.98964, 2.50986e-22, 5.36361e-23, 3.21817e-23, 2.8386996e-23], [977.23722, 2.51489e-22, 5.37427e-23, 3.22456e-23, 2.8462768e-23], [979.48999, 2.51992e-22, 5.38495e-23, 3.23097e-23, 2.8538701e-23], [981.74794, 2.52497e-22, 5.39566e-23, 3.2374e-23, 2.8614796e-23], [984.01111, 2.53003e-22, 5.4064e-23, 3.24384e-23, 2.8691054e-23], [986.27949, 2.53511e-22, 5.41717e-23, 3.2503e-23, 2.8767474e-23], [988.55309, 2.5402e-22, 5.42797e-23, 3.25678e-23, 2.8844057e-23], [990.83194, 2.5453e-22, 5.43879e-23, 3.26327e-23, 2.8920803e-23], [993.11605, 2.55041e-22, 5.44964e-23, 3.26979e-23, 2.8997712e-23], [995.40542, 2.55554e-22, 5.46053e-23, 3.27632e-23, 2.9074786e-23], [997.70006, 2.56068e-22, 5.47144e-23, 3.28286e-23, 2.9152024e-23], [1000.0, 2.56584e-22, 5.48238e-23, 3.28943e-23, 2.9229426e-23], [1002.3052, 2.57101e-22, 5.49334e-23, 3.29601e-23, 2.9306994e-23], [1004.6158, 2.57619e-22, 5.50434e-23, 3.30261e-23, 2.9384726e-23], [1006.9317, 2.58138e-22, 5.51537e-23, 3.30922e-23, 2.9462625e-23], [1009.2529, 2.58659e-22, 5.52642e-23, 3.31585e-23, 2.9540689e-23], [1011.5795, 2.59181e-22, 5.53751e-23, 3.3225e-23, 2.961892e-23], [1013.9114, 2.59705e-22, 5.54862e-23, 3.32917e-23, 2.9697318e-23], [1016.2487, 2.6023e-22, 5.55976e-23, 3.33586e-23, 2.9775883e-23], [1018.5914, 2.60756e-22, 5.57094e-23, 3.34256e-23, 2.9854616e-23], [1020.9395, 2.61284e-22, 5.58214e-23, 3.34928e-23, 2.9933516e-23], [1023.293, 2.61813e-22, 5.59337e-23, 3.35602e-23, 3.0012585e-23], [1025.6519, 2.62343e-22, 5.60463e-23, 3.36278e-23, 3.0091822e-23], [1028.0163, 2.62875e-22, 5.61592e-23, 3.36955e-23, 3.0171228e-23], [1030.3861, 2.63408e-22, 5.62724e-23, 3.37634e-23, 3.0250804e-23], [1032.7614, 2.63943e-22, 5.63859e-23, 3.38315e-23, 3.0330549e-23], [1035.1422, 2.64479e-22, 5.64996e-23, 3.38998e-23, 3.0410464e-23], [1037.5284, 2.65016e-22, 5.66137e-23, 3.39682e-23, 3.049055e-23], [1039.9202, 2.65555e-22, 5.67281e-23, 3.40369e-23, 3.0570807e-23], [1042.3174, 2.66095e-22, 5.68428e-23, 3.41057e-23, 3.0651235e-23], [1044.7202, 2.66636e-22, 5.69578e-23, 3.41747e-23, 3.0731835e-23], [1047.1285, 2.67179e-22, 5.70731e-23, 3.42438e-23, 3.0812606e-23], [1049.5424, 2.67724e-22, 5.71887e-23, 3.43132e-23, 3.089355e-23], [1051.9619, 2.68269e-22, 5.73045e-23, 3.43827e-23, 3.0974667e-23], [1054.3869, 2.68816e-22, 5.74207e-23, 3.44524e-23, 3.1055957e-23], [1056.8175, 2.69365e-22, 5.75372e-23, 3.45223e-23, 3.1137421e-23], [1059.2537, 2.69915e-22, 5.7654e-23, 3.45924e-23, 3.1219059e-23], [1061.6956, 2.70466e-22, 5.77711e-23, 3.46627e-23, 3.1300871e-23], [1064.143, 2.71019e-22, 5.78886e-23, 3.47331e-23, 3.1382858e-23], [1066.5961, 2.71573e-22, 5.80063e-23, 3.48038e-23, 3.146502e-23], [1069.0549, 2.72129e-22, 5.81243e-23, 3.48746e-23, 3.1547357e-23], [1071.5193, 2.72686e-22, 5.82427e-23, 3.49456e-23, 3.1629871e-23], [1073.9894, 2.73244e-22, 5.83613e-23, 3.50168e-23, 3.1712561e-23], [1076.4652, 2.73804e-22, 5.84803e-23, 3.50882e-23, 3.1795428e-23], [1078.9467, 2.74365e-22, 5.85996e-23, 3.51597e-23, 3.1878472e-23], [1081.434, 2.74928e-22, 5.87191e-23, 3.52315e-23, 3.1961695e-23], [1083.9269, 2.75492e-22, 5.88391e-23, 3.53034e-23, 3.2045095e-23], [1086.4256, 2.76058e-22, 5.89593e-23, 3.53756e-23, 3.2128674e-23], [1088.9301, 2.76625e-22, 5.90798e-23, 3.54479e-23, 3.2212433e-23], [1091.4403, 2.77194e-22, 5.92007e-23, 3.55204e-23, 3.2296371e-23], [1093.9564, 2.77764e-22, 5.93219e-23, 3.55932e-23, 3.238049e-23], [1096.4782, 2.78335e-22, 5.94435e-23, 3.56661e-23, 3.246479e-23], [1099.0058, 2.78908e-22, 5.95653e-23, 3.57392e-23, 3.2549271e-23], [1101.5393, 2.79483e-22, 5.96875e-23, 3.58125e-23, 3.2633935e-23], [1104.0786, 2.80059e-22, 5.98101e-23, 3.58861e-23, 3.2718781e-23], [1106.6238, 2.80636e-22, 5.9933e-23, 3.59598e-23, 3.2803812e-23], [1109.1748, 2.81216e-22, 6.00563e-23, 3.60338e-23, 3.2889028e-23], [1111.7317, 2.81796e-22, 6.01799e-23, 3.61079e-23, 3.2974431e-23], [1114.2945, 2.82378e-22, 6.0304e-23, 3.61824e-23, 3.3060022e-23], [1116.8632, 2.82962e-22, 6.04284e-23, 3.6257e-23, 3.3145803e-23], [1119.4379, 2.83547e-22, 6.05533e-23, 3.6332e-23, 3.3231777e-23], [1122.0185, 2.84134e-22, 6.06787e-23, 3.64072e-23, 3.3317948e-23], [1124.605, 2.84723e-22, 6.08047e-23, 3.64828e-23, 3.3404322e-23], [1127.1975, 2.85313e-22, 6.09314e-23, 3.65589e-23, 3.3490908e-23], [1129.7959, 2.85905e-22, 6.1059e-23, 3.66354e-23, 3.3577718e-23], [1132.4004, 2.865e-22, 6.11877e-23, 3.67126e-23, 3.3664777e-23], [1135.0108, 2.87097e-22, 6.13182e-23, 3.67909e-23, 3.3752126e-23], [1137.6273, 2.87699e-22, 6.14516e-23, 3.6871e-23, 3.3839847e-23], [1140.2498, 2.88305e-22, 6.15905e-23, 3.69543e-23, 3.3928126e-23], [1142.8783, 2.88924e-22, 6.17414e-23, 3.70448e-23, 3.4017446e-23], [1145.5129, 2.89573e-22, 6.19271e-23, 3.71562e-23, 3.4109448e-23], [1148.1536, 2.90359e-22, 6.22696e-23, 3.73617e-23, 3.4213023e-23], [1150.8004, 2.93501e-22, 6.53032e-23, 3.91819e-23, 3.4516831e-23], [1153.4533, 2.93343e-22, 6.45712e-23, 3.87427e-23, 3.4540335e-23], [1156.1122, 2.92137e-22, 6.26149e-23, 3.75689e-23, 3.4475119e-23], [1158.7774, 2.92585e-22, 6.25575e-23, 3.75345e-23, 3.4550305e-23], [1161.4486, 2.93149e-22, 6.26345e-23, 3.75807e-23, 3.4635423e-23], [1164.126, 2.93741e-22, 6.27425e-23, 3.76455e-23, 3.4722956e-23], [1166.8096, 2.94343e-22, 6.28616e-23, 3.77169e-23, 3.4811455e-23], [1169.4994, 2.94951e-22, 6.29857e-23, 3.77914e-23, 3.4900485e-23], [1172.1954, 2.95563e-22, 6.31124e-23, 3.78675e-23, 3.4989877e-23], [1174.8976, 2.96178e-22, 6.32409e-23, 3.79445e-23, 3.5079557e-23], [1177.606, 2.96794e-22, 6.33704e-23, 3.80223e-23, 3.5169487e-23], [1180.3206, 2.97413e-22, 6.35008e-23, 3.81005e-23, 3.5259647e-23], [1183.0416, 2.98033e-22, 6.36319e-23, 3.81791e-23, 3.5350025e-23], [1185.7687, 2.98655e-22, 6.37635e-23, 3.82581e-23, 3.5440614e-23], [1188.5022, 2.99279e-22, 6.38957e-23, 3.83374e-23, 3.553141e-23], [1191.242, 2.99905e-22, 6.40283e-23, 3.8417e-23, 3.5622409e-23], [1193.9881, 3.00532e-22, 6.41613e-23, 3.84968e-23, 3.571361e-23], [1196.7405, 3.01161e-22, 6.42948e-23, 3.85769e-23, 3.5805012e-23], [1199.4993, 3.01792e-22, 6.44286e-23, 3.86572e-23, 3.5896614e-23], [1202.2644, 3.02424e-22, 6.45628e-23, 3.87377e-23, 3.5988415e-23], [1205.0359, 3.03058e-22, 6.46974e-23, 3.88185e-23, 3.6080416e-23], [1207.8138, 3.03693e-22, 6.48324e-23, 3.88994e-23, 3.6172617e-23], [1210.5981, 3.0433e-22, 6.49677e-23, 3.89806e-23, 3.6265018e-23], [1213.3889, 3.04969e-22, 6.51034e-23, 3.90621e-23, 3.6357618e-23], [1216.186, 3.05609e-22, 6.52395e-23, 3.91437e-23, 3.6450418e-23], [1218.9896, 3.06251e-22, 6.53759e-23, 3.92255e-23, 3.6543418e-23], [1221.7997, 3.06894e-22, 6.55127e-23, 3.93076e-23, 3.6636619e-23], [1224.6162, 3.07539e-22, 6.56498e-23, 3.93899e-23, 3.6730021e-23], [1227.4392, 3.08186e-22, 6.57872e-23, 3.94723e-23, 3.6823625e-23], [1230.2688, 3.08835e-22, 6.59251e-23, 3.9555e-23, 3.691743e-23], [1233.1048, 3.09485e-22, 6.60633e-23, 3.9638e-23, 3.7011437e-23], [1235.9474, 3.10136e-22, 6.62018e-23, 3.97211e-23, 3.7105647e-23], [1238.7966, 3.10789e-22, 6.63407e-23, 3.98044e-23, 3.7200059e-23], [1241.6523, 3.11444e-22, 6.64799e-23, 3.9888e-23, 3.7294675e-23], [1244.5146, 3.12101e-22, 6.66195e-23, 3.99717e-23, 3.7389495e-23], [1247.3835, 3.12759e-22, 6.67595e-23, 4.00557e-23, 3.7484518e-23], [1250.259, 3.13419e-22, 6.68998e-23, 4.01399e-23, 3.7579747e-23], [1253.1412, 3.1408e-22, 6.70405e-23, 4.02243e-23, 3.767518e-23], [1256.03, 3.14744e-22, 6.71815e-23, 4.03089e-23, 3.7770819e-23], [1258.9254, 3.15408e-22, 6.73229e-23, 4.03937e-23, 3.7866665e-23], [1261.8275, 3.16075e-22, 6.74646e-23, 4.04788e-23, 3.7962716e-23], [1264.7363, 3.16743e-22, 6.76067e-23, 4.0564e-23, 3.8058975e-23], [1267.6519, 3.17413e-22, 6.77492e-23, 4.06495e-23, 3.8155441e-23], [1270.5741, 3.18084e-22, 6.7892e-23, 4.07352e-23, 3.8252114e-23], [1273.5031, 3.18758e-22, 6.80351e-23, 4.08211e-23, 3.8348997e-23], [1276.4388, 3.19432e-22, 6.81787e-23, 4.09072e-23, 3.8446088e-23], [1279.3813, 3.20109e-22, 6.83226e-23, 4.09935e-23, 3.8543388e-23], [1282.3306, 3.20787e-22, 6.84668e-23, 4.10801e-23, 3.8640898e-23], [1285.2867, 3.21467e-22, 6.86115e-23, 4.11669e-23, 3.8738618e-23], [1288.2496, 3.22149e-22, 6.87564e-23, 4.12539e-23, 3.8836549e-23], [1291.2193, 3.22832e-22, 6.89018e-23, 4.13411e-23, 3.8934691e-23], [1294.1958, 3.23517e-22, 6.90475e-23, 4.14285e-23, 3.9033045e-23], [1297.1793, 3.24204e-22, 6.91936e-23, 4.15161e-23, 3.9131611e-23], [1300.1696, 3.24892e-22, 6.934e-23, 4.1604e-23, 3.923039e-23], [1303.1668, 3.25582e-22, 6.94868e-23, 4.16921e-23, 3.9329382e-23], [1306.1709, 3.26274e-22, 6.9634e-23, 4.17804e-23, 3.9428588e-23], [1309.1819, 3.26968e-22, 6.97816e-23, 4.18689e-23, 3.9528007e-23], [1312.1999, 3.27663e-22, 6.99295e-23, 4.19577e-23, 3.9627642e-23], [1315.2248, 3.2836e-22, 7.00778e-23, 4.20467e-23, 3.9727491e-23], [1318.2567, 3.29059e-22, 7.02264e-23, 4.21359e-23, 3.9827556e-23], [1321.2956, 3.2976e-22, 7.03754e-23, 4.22253e-23, 3.9927837e-23], [1324.3415, 3.30462e-22, 7.05249e-23, 4.23149e-23, 4.0028335e-23], [1327.3945, 3.31166e-22, 7.06746e-23, 4.24048e-23, 4.012905e-23], [1330.4544, 3.31872e-22, 7.08248e-23, 4.24949e-23, 4.0229983e-23], [1333.5214, 3.32579e-22, 7.09753e-23, 4.25852e-23, 4.0331134e-23], [1336.5955, 3.33288e-22, 7.11262e-23, 4.26757e-23, 4.0432503e-23], [1339.6767, 3.33999e-22, 7.12775e-23, 4.27665e-23, 4.0534092e-23], [1342.765, 3.34712e-22, 7.14291e-23, 4.28575e-23, 4.06359e-23], [1345.8604, 3.35427e-22, 7.15812e-23, 4.29487e-23, 4.0737928e-23], [1348.9629, 3.36143e-22, 7.17336e-23, 4.30401e-23, 4.0840177e-23], [1352.0726, 3.36861e-22, 7.18864e-23, 4.31318e-23, 4.0942648e-23], [1355.1894, 3.37581e-22, 7.20395e-23, 4.32237e-23, 4.104534e-23], [1358.3134, 3.38302e-22, 7.21931e-23, 4.33158e-23, 4.1148254e-23], [1361.4447, 3.39026e-22, 7.2347e-23, 4.34082e-23, 4.1251391e-23], [1364.5831, 3.39751e-22, 7.25013e-23, 4.35008e-23, 4.1354751e-23], [1367.7288, 3.40478e-22, 7.2656e-23, 4.35936e-23, 4.1458335e-23], [1370.8818, 3.41207e-22, 7.28111e-23, 4.36867e-23, 4.1562143e-23], [1374.042, 3.41937e-22, 7.29666e-23, 4.37799e-23, 4.1666177e-23], [1377.2095, 3.4267e-22, 7.31224e-23, 4.38735e-23, 4.1770435e-23], [1380.3843, 3.43404e-22, 7.32787e-23, 4.39672e-23, 4.1874919e-23], [1383.5664, 3.4414e-22, 7.34353e-23, 4.40612e-23, 4.197963e-23], [1386.7558, 3.44878e-22, 7.35923e-23, 4.41554e-23, 4.2084568e-23], [1389.9526, 3.45618e-22, 7.37497e-23, 4.42498e-23, 4.2189733e-23], [1393.1568, 3.46359e-22, 7.39075e-23, 4.43445e-23, 4.2295126e-23], [1396.3684, 3.47102e-22, 7.40657e-23, 4.44394e-23, 4.2400748e-23], [1399.5873, 3.47848e-22, 7.42243e-23, 4.45346e-23, 4.2506599e-23], [1402.8137, 3.48595e-22, 7.43833e-23, 4.463e-23, 4.2612679e-23], [1406.0475, 3.49343e-22, 7.45427e-23, 4.47256e-23, 4.2718989e-23], [1409.2888, 3.50094e-22, 7.47024e-23, 4.48214e-23, 4.282553e-23], [1412.5375, 3.50847e-22, 7.48626e-23, 4.49175e-23, 4.2932303e-23], [1415.7938, 3.51601e-22, 7.50231e-23, 4.50139e-23, 4.3039307e-23], [1419.0575, 3.52357e-22, 7.51841e-23, 4.51104e-23, 4.3146543e-23], [1422.3288, 3.53115e-22, 7.53454e-23, 4.52073e-23, 4.3254012e-23], [1425.6076, 3.53875e-22, 7.55072e-23, 4.53043e-23, 4.3361715e-23], [1428.894, 3.54637e-22, 7.56693e-23, 4.54016e-23, 4.3469651e-23], [1432.1879, 3.55401e-22, 7.58319e-23, 4.54991e-23, 4.3577822e-23], [1435.4894, 3.56166e-22, 7.59948e-23, 4.55969e-23, 4.3686228e-23], [1438.7986, 3.56934e-22, 7.61582e-23, 4.56949e-23, 4.379487e-23], [1442.1154, 3.57703e-22, 7.63219e-23, 4.57932e-23, 4.3903747e-23], [1445.4398, 3.58474e-22, 7.64861e-23, 4.58917e-23, 4.4012862e-23], [1448.7719, 3.59247e-22, 7.66507e-23, 4.59904e-23, 4.4122213e-23], [1452.1116, 3.60022e-22, 7.68156e-23, 4.60894e-23, 4.4231802e-23], [1455.4591, 3.60799e-22, 7.6981e-23, 4.61886e-23, 4.434163e-23], [1458.8143, 3.61578e-22, 7.71468e-23, 4.62881e-23, 4.4451697e-23], [1462.1772, 3.62359e-22, 7.7313e-23, 4.63878e-23, 4.4562003e-23], [1465.5478, 3.63142e-22, 7.74796e-23, 4.64878e-23, 4.4672549e-23], [1468.9263, 3.63926e-22, 7.76466e-23, 4.6588e-23, 4.4783335e-23], [1472.3125, 3.64713e-22, 7.7814e-23, 4.66884e-23, 4.4894363e-23], [1475.7065, 3.65501e-22, 7.79819e-23, 4.67891e-23, 4.5005633e-23], [1479.1084, 3.66291e-22, 7.81501e-23, 4.68901e-23, 4.5117145e-23], [1482.5181, 3.67084e-22, 7.83188e-23, 4.69913e-23, 4.5228899e-23], [1485.9356, 3.67878e-22, 7.84879e-23, 4.70927e-23, 4.5340897e-23], [1489.3611, 3.68674e-22, 7.86574e-23, 4.71944e-23, 4.5453139e-23], [1492.7944, 3.69472e-22, 7.88273e-23, 4.72964e-23, 4.5565626e-23], [1496.2357, 3.70272e-22, 7.89976e-23, 4.73986e-23, 4.5678358e-23], [1499.6848, 3.71074e-22, 7.91684e-23, 4.7501e-23, 4.5791336e-23], [1503.142, 3.71878e-22, 7.93395e-23, 4.76037e-23, 4.590456e-23], [1506.6071, 3.72684e-22, 7.95111e-23, 4.77067e-23, 4.601803e-23], [1510.0802, 3.73492e-22, 7.96831e-23, 4.78099e-23, 4.6131749e-23], [1513.5612, 3.74302e-22, 7.98556e-23, 4.79133e-23, 4.6245715e-23], [1517.0504, 3.75114e-22, 8.00284e-23, 4.80171e-23, 4.6359931e-23], [1520.5475, 3.75928e-22, 8.02017e-23, 4.8121e-23, 4.6474396e-23], [1524.0528, 3.76744e-22, 8.03754e-23, 4.82253e-23, 4.6589111e-23], [1527.5661, 3.77562e-22, 8.05496e-23, 4.83298e-23, 4.6704078e-23], [1531.0875, 3.78382e-22, 8.07242e-23, 4.84345e-23, 4.6819298e-23], [1534.617, 3.79204e-22, 8.08994e-23, 4.85396e-23, 4.6934776e-23], [1538.1546, 3.80028e-22, 8.10753e-23, 4.86452e-23, 4.7050527e-23], [1541.7005, 3.80856e-22, 8.12539e-23, 4.87524e-23, 4.7166693e-23], [1545.2544, 3.81924e-22, 8.17115e-23, 4.90269e-23, 4.7302326e-23], [1548.8166, 3.82513e-22, 8.16056e-23, 4.89634e-23, 4.739931e-23], [1552.387, 3.83343e-22, 8.1781e-23, 4.90686e-23, 4.7515923e-23], [1555.9656, 3.84177e-22, 8.19583e-23, 4.9175e-23, 4.7632888e-23], [1559.5525, 3.85013e-22, 8.21361e-23, 4.92817e-23, 4.7750123e-23], [1563.1476, 3.85851e-22, 8.23145e-23, 4.93887e-23, 4.786762e-23], [1566.7511, 3.86691e-22, 8.24934e-23, 4.9496e-23, 4.7985375e-23], [1570.3628, 3.87533e-22, 8.26727e-23, 4.96036e-23, 4.8103389e-23], [1573.9829, 3.88377e-22, 8.28524e-23, 4.97114e-23, 4.8221662e-23], [1577.6113, 3.89223e-22, 8.30326e-23, 4.98196e-23, 4.8340194e-23], [1581.248, 3.90071e-22, 8.32132e-23, 4.99279e-23, 4.8458985e-23], [1584.8932, 3.90922e-22, 8.33943e-23, 5.00366e-23, 4.8578037e-23], [1588.5467, 3.91774e-22, 8.35758e-23, 5.01455e-23, 4.869735e-23], [1592.2087, 3.92629e-22, 8.37578e-23, 5.02547e-23, 4.8816924e-23], [1595.8791, 3.93485e-22, 8.39402e-23, 5.03641e-23, 4.893676e-23], [1599.558, 3.94344e-22, 8.41231e-23, 5.04738e-23, 4.9056859e-23], [1603.2454, 3.95205e-22, 8.43064e-23, 5.05838e-23, 4.9177221e-23], [1606.9413, 3.96067e-22, 8.44902e-23, 5.06941e-23, 4.9297848e-23], [1610.6456, 3.96932e-22, 8.46744e-23, 5.08046e-23, 4.941874e-23], [1614.3586, 3.97799e-22, 8.48591e-23, 5.09155e-23, 4.9539897e-23], [1618.08, 3.98668e-22, 8.50443e-23, 5.10266e-23, 4.9661321e-23], [1621.8101, 3.9954e-22, 8.52299e-23, 5.11379e-23, 4.9783012e-23], [1625.5488, 4.00413e-22, 8.5416e-23, 5.12496e-23, 4.9904972e-23], [1629.296, 4.01289e-22, 8.56026e-23, 5.13616e-23, 5.0027203e-23], [1633.0519, 4.02166e-22, 8.57897e-23, 5.14738e-23, 5.0149705e-23], [1636.8165, 4.03046e-22, 8.59773e-23, 5.15864e-23, 5.0272481e-23], [1640.5898, 4.03929e-22, 8.61656e-23, 5.16993e-23, 5.0395535e-23], [1644.3717, 4.04813e-22, 8.63545e-23, 5.18127e-23, 5.0518873e-23], [1648.1624, 4.057e-22, 8.65442e-23, 5.19265e-23, 5.0642507e-23], [1651.9618, 4.0659e-22, 8.6735e-23, 5.2041e-23, 5.0766458e-23], [1655.77, 4.07483e-22, 8.69278e-23, 5.21567e-23, 5.0890783e-23], [1659.5869, 4.08381e-22, 8.71251e-23, 5.22751e-23, 5.1015655e-23], [1663.4127, 4.09295e-22, 8.73381e-23, 5.24029e-23, 5.1141845e-23], [1667.2472, 4.10329e-22, 8.76897e-23, 5.26138e-23, 5.1277743e-23], [1671.0906, 4.13067e-22, 9.00111e-23, 5.40067e-23, 5.1550156e-23], [1674.9429, 4.12027e-22, 8.79527e-23, 5.27716e-23, 5.152019e-23], [1678.804, 4.12883e-22, 8.80902e-23, 5.28541e-23, 5.1642194e-23], [1682.6741, 4.13778e-22, 8.82706e-23, 5.29624e-23, 5.1767373e-23], [1686.553, 4.14682e-22, 8.84593e-23, 5.30756e-23, 5.1893358e-23], [1690.4409, 4.1559e-22, 8.86508e-23, 5.31905e-23, 5.2019788e-23], [1694.3378, 4.16501e-22, 8.88438e-23, 5.33063e-23, 5.2146564e-23], [1698.2437, 4.17415e-22, 8.90378e-23, 5.34227e-23, 5.2273652e-23], [1702.1585, 4.18331e-22, 8.92325e-23, 5.35395e-23, 5.2401038e-23], [1706.0824, 4.19249e-22, 8.94279e-23, 5.36567e-23, 5.2528715e-23], [1710.0153, 4.2017e-22, 8.96238e-23, 5.37743e-23, 5.265668e-23], [1713.9573, 4.21093e-22, 8.98202e-23, 5.38921e-23, 5.2784931e-23], [1717.9084, 4.22019e-22, 9.00172e-23, 5.40103e-23, 5.2913467e-23], [1721.8686, 4.22946e-22, 9.02147e-23, 5.41288e-23, 5.3042288e-23], [1725.8379, 4.23876e-22, 9.04127e-23, 5.42476e-23, 5.3171394e-23], [1729.8164, 4.24808e-22, 9.06111e-23, 5.43667e-23, 5.3300786e-23], [1733.804, 4.25742e-22, 9.08101e-23, 5.44861e-23, 5.3430463e-23], [1737.8008, 4.26679e-22, 9.10096e-23, 5.46057e-23, 5.3560427e-23], [1741.8069, 4.27618e-22, 9.12095e-23, 5.47257e-23, 5.3690678e-23], [1745.8222, 4.28559e-22, 9.141e-23, 5.4846e-23, 5.3821216e-23], [1749.8467, 4.29503e-22, 9.16109e-23, 5.49665e-23, 5.3952042e-23], [1753.8805, 4.30448e-22, 9.18123e-23, 5.50874e-23, 5.4083157e-23], [1757.9236, 4.31396e-22, 9.20142e-23, 5.52085e-23, 5.4214561e-23], [1761.976, 4.32347e-22, 9.22166e-23, 5.533e-23, 5.4346255e-23], [1766.0378, 4.33299e-22, 9.24195e-23, 5.54517e-23, 5.4478239e-23], [1770.109, 4.34254e-22, 9.26229e-23, 5.55737e-23, 5.4610515e-23], [1774.1895, 4.35211e-22, 9.28268e-23, 5.56961e-23, 5.4743082e-23], [1778.2794, 4.36171e-22, 9.30311e-23, 5.58187e-23, 5.4875942e-23], [1782.3788, 4.37132e-22, 9.3236e-23, 5.59416e-23, 5.5009095e-23], [1786.4876, 4.38096e-22, 9.34414e-23, 5.60648e-23, 5.5142542e-23], [1790.6059, 4.39063e-22, 9.36472e-23, 5.61883e-23, 5.5276284e-23], [1794.7336, 4.40032e-22, 9.38536e-23, 5.63121e-23, 5.5410321e-23], [1798.8709, 4.41003e-22, 9.40604e-23, 5.64362e-23, 5.5544654e-23], [1803.0177, 4.41976e-22, 9.42678e-23, 5.65607e-23, 5.5679284e-23], [1807.1741, 4.42952e-22, 9.44756e-23, 5.66854e-23, 5.5814211e-23], [1811.3401, 4.4393e-22, 9.4684e-23, 5.68104e-23, 5.5949436e-23], [1815.5157, 4.4491e-22, 9.48928e-23, 5.69357e-23, 5.6084959e-23], [1819.7009, 4.45893e-22, 9.51022e-23, 5.70613e-23, 5.6220783e-23], [1823.8957, 4.46878e-22, 9.5312e-23, 5.71872e-23, 5.6356906e-23], [1828.1002, 4.47866e-22, 9.55224e-23, 5.73134e-23, 5.649333e-23], [1832.3144, 4.48856e-22, 9.57333e-23, 5.744e-23, 5.6630056e-23], [1836.5383, 4.49848e-22, 9.59446e-23, 5.75668e-23, 5.6767084e-23], [1840.772, 4.50843e-22, 9.61565e-23, 5.76939e-23, 5.6904415e-23], [1845.0154, 4.5184e-22, 9.63689e-23, 5.78214e-23, 5.704205e-23], [1849.2686, 4.52839e-22, 9.65818e-23, 5.79491e-23, 5.7179989e-23], [1853.5316, 4.53841e-22, 9.67953e-23, 5.80772e-23, 5.7318234e-23], [1857.8045, 4.54845e-22, 9.70092e-23, 5.82055e-23, 5.7456784e-23], [1862.0871, 4.55852e-22, 9.72236e-23, 5.83342e-23, 5.7595641e-23], [1866.3797, 4.56861e-22, 9.74386e-23, 5.84632e-23, 5.7734805e-23], [1870.6821, 4.57872e-22, 9.76541e-23, 5.85925e-23, 5.7874277e-23], [1874.9945, 4.58886e-22, 9.78701e-23, 5.8722e-23, 5.8014058e-23], [1879.3168, 4.59903e-22, 9.80866e-23, 5.8852e-23, 5.8154148e-23], [1883.6491, 4.60921e-22, 9.83036e-23, 5.89822e-23, 5.8294548e-23], [1887.9913, 4.61942e-22, 9.85212e-23, 5.91127e-23, 5.843526e-23], [1892.3436, 4.62966e-22, 9.87393e-23, 5.92436e-23, 5.8576283e-23], [1896.7059, 4.63992e-22, 9.89579e-23, 5.93747e-23, 5.8717619e-23], [1901.0783, 4.65021e-22, 9.9177e-23, 5.95062e-23, 5.8859267e-23], [1905.4607, 4.66052e-22, 9.93967e-23, 5.9638e-23, 5.900123e-23], [1909.8533, 4.67085e-22, 9.96168e-23, 5.97701e-23, 5.9143507e-23], [1914.2559, 4.68121e-22, 9.98376e-23, 5.99025e-23, 5.92861e-23], [1918.6687, 4.6916e-22, 1.00059e-22, 6.00353e-23, 5.9429009e-23], [1923.0917, 4.70201e-22, 1.00281e-22, 6.01683e-23, 5.9572234e-23], [1927.5249, 4.71244e-22, 1.00503e-22, 6.03017e-23, 5.9715778e-23], [1931.9683, 4.7229e-22, 1.00726e-22, 6.04354e-23, 5.9859639e-23], [1936.422, 4.73338e-22, 1.00949e-22, 6.05694e-23, 6.000382e-23], [1940.8859, 4.74389e-22, 1.01173e-22, 6.07038e-23, 6.014832e-23], [1945.3601, 4.75443e-22, 1.01397e-22, 6.08384e-23, 6.0293142e-23], [1949.8446, 4.76499e-22, 1.01622e-22, 6.09734e-23, 6.0438284e-23], [1954.3395, 4.77557e-22, 1.01848e-22, 6.11087e-23, 6.0583749e-23], [1958.8447, 4.78618e-22, 1.02074e-22, 6.12444e-23, 6.0729536e-23], [1963.3603, 4.79682e-22, 1.02301e-22, 6.13803e-23, 6.0875648e-23], [1967.8863, 4.80748e-22, 1.02528e-22, 6.15166e-23, 6.1022083e-23], [1972.4227, 4.81816e-22, 1.02755e-22, 6.16532e-23, 6.1168844e-23], [1976.9696, 4.82887e-22, 1.02984e-22, 6.17902e-23, 6.1315931e-23], [1981.527, 4.83961e-22, 1.03212e-22, 6.19274e-23, 6.1463344e-23], [1986.0949, 4.85038e-22, 1.03442e-22, 6.2065e-23, 6.1611085e-23], [1990.6733, 4.86116e-22, 1.03672e-22, 6.22029e-23, 6.1759155e-23], [1995.2623, 4.87198e-22, 1.03902e-22, 6.23412e-23, 6.1907553e-23], [1999.8619, 4.88282e-22, 1.04133e-22, 6.24798e-23, 6.2056281e-23], [2004.472, 4.89368e-22, 1.04364e-22, 6.26187e-23, 6.220534e-23], [2009.0928, 4.90458e-22, 1.04597e-22, 6.27579e-23, 6.235473e-23], [2013.7242, 4.91549e-22, 1.04829e-22, 6.28975e-23, 6.2504452e-23], [2018.3664, 4.92644e-22, 1.05062e-22, 6.30374e-23, 6.2654507e-23], [2023.0192, 4.93741e-22, 1.05296e-22, 6.31777e-23, 6.2804896e-23], [2027.6827, 4.9484e-22, 1.0553e-22, 6.33182e-23, 6.295562e-23], [2032.357, 4.95943e-22, 1.05765e-22, 6.34592e-23, 6.3106679e-23], [2037.0421, 4.97048e-22, 1.06001e-22, 6.36004e-23, 6.3258073e-23], [2041.7379, 4.98155e-22, 1.06237e-22, 6.3742e-23, 6.3409805e-23], [2046.4446, 4.99265e-22, 1.06473e-22, 6.38839e-23, 6.3561875e-23], [2051.1622, 5.00378e-22, 1.0671e-22, 6.40262e-23, 6.3714282e-23], [2055.8906, 5.01493e-22, 1.06948e-22, 6.41688e-23, 6.386703e-23], [2060.6299, 5.02612e-22, 1.07186e-22, 6.43117e-23, 6.4020117e-23], [2065.3802, 5.03732e-22, 1.07425e-22, 6.4455e-23, 6.4173545e-23], [2070.1413, 5.04856e-22, 1.07664e-22, 6.45987e-23, 6.4327315e-23], [2074.9135, 5.05982e-22, 1.07904e-22, 6.47426e-23, 6.4481427e-23], [2079.6967, 5.07111e-22, 1.08145e-22, 6.48869e-23, 6.4635883e-23], [2084.4909, 5.08242e-22, 1.08386e-22, 6.50316e-23, 6.4790682e-23], [2089.2961, 5.09376e-22, 1.08628e-22, 6.51766e-23, 6.4945827e-23], [2094.1125, 5.10513e-22, 1.0887e-22, 6.53219e-23, 6.5101317e-23], [2098.9399, 5.11653e-22, 1.09113e-22, 6.54676e-23, 6.5257154e-23], [2103.7784, 5.12795e-22, 1.09356e-22, 6.56137e-23, 6.5413339e-23], [2108.6281, 5.1394e-22, 1.096e-22, 6.57601e-23, 6.5569871e-23], [2113.489, 5.15088e-22, 1.09845e-22, 6.59068e-23, 6.5726753e-23], [2118.3611, 5.16238e-22, 1.1009e-22, 6.60539e-23, 6.5883984e-23], [2123.2445, 5.17391e-22, 1.10336e-22, 6.62013e-23, 6.6041566e-23], [2128.139, 5.18547e-22, 1.10582e-22, 6.63491e-23, 6.6199499e-23], [2133.0449, 5.19706e-22, 1.10829e-22, 6.64973e-23, 6.6357785e-23], [2137.9621, 5.20867e-22, 1.11076e-22, 6.66458e-23, 6.6516424e-23], [2142.8906, 5.22031e-22, 1.11324e-22, 6.67946e-23, 6.6675417e-23], [2147.8305, 5.23198e-22, 1.11573e-22, 6.69438e-23, 6.6834765e-23], [2152.7817, 5.24368e-22, 1.11822e-22, 6.70934e-23, 6.6994468e-23], [2157.7444, 5.25541e-22, 1.12072e-22, 6.72433e-23, 6.7154528e-23], [2162.7185, 5.26716e-22, 1.12323e-22, 6.73936e-23, 6.7314945e-23], [2167.7041, 5.27894e-22, 1.12574e-22, 6.75442e-23, 6.7475721e-23], [2172.7012, 5.29075e-22, 1.12825e-22, 6.76952e-23, 6.7636855e-23], [2177.7098, 5.30258e-22, 1.13078e-22, 6.78465e-23, 6.779835e-23], [2182.7299, 5.31445e-22, 1.1333e-22, 6.79982e-23, 6.7960205e-23], [2187.7616, 5.32634e-22, 1.13584e-22, 6.81503e-23, 6.8122422e-23], [2192.8049, 5.33826e-22, 1.13838e-22, 6.83027e-23, 6.8285001e-23], [2197.8599, 5.35021e-22, 1.14093e-22, 6.84555e-23, 6.8447943e-23], [2202.9265, 5.36219e-22, 1.14348e-22, 6.86087e-23, 6.861125e-23], [2208.0047, 5.3742e-22, 1.14604e-22, 6.87622e-23, 6.8774922e-23], [2213.0947, 5.38623e-22, 1.1486e-22, 6.89161e-23, 6.893896e-23], [2218.1964, 5.3983e-22, 1.15117e-22, 6.90703e-23, 6.9103365e-23], [2223.3099, 5.41039e-22, 1.15375e-22, 6.92249e-23, 6.9268137e-23], [2228.4351, 5.42251e-22, 1.15633e-22, 6.93799e-23, 6.9433279e-23], [2233.5722, 5.43466e-22, 1.15892e-22, 6.95353e-23, 6.959879e-23], [2238.7211, 5.44684e-22, 1.16152e-22, 6.9691e-23, 6.9764673e-23], [2243.8819, 5.45904e-22, 1.16412e-22, 6.98472e-23, 6.9930927e-23], [2249.0546, 5.47128e-22, 1.16673e-22, 7.00037e-23, 7.0097556e-23], [2254.2392, 5.48355e-22, 1.16934e-22, 7.01606e-23, 7.026456e-23], [2259.4358, 5.49584e-22, 1.17197e-22, 7.03179e-23, 7.0431943e-23], [2264.6443, 5.50817e-22, 1.1746e-22, 7.04757e-23, 7.0599711e-23], [2269.8649, 5.52053e-22, 1.17723e-22, 7.06341e-23, 7.0767878e-23], [2275.0974, 5.53292e-22, 1.17989e-22, 7.07934e-23, 7.0936483e-23], [2280.3421, 5.54538e-22, 1.18259e-22, 7.09553e-23, 7.1105713e-23], [2285.5988, 5.55819e-22, 1.18567e-22, 7.11403e-23, 7.1277833e-23], [2290.8677, 5.57272e-22, 1.19074e-22, 7.14444e-23, 7.1463521e-23], [2296.1486, 5.58284e-22, 1.19062e-22, 7.14373e-23, 7.161501e-23], [2301.4418, 5.59528e-22, 1.19319e-22, 7.15913e-23, 7.178471e-23], [2306.7472, 5.60782e-22, 1.19584e-22, 7.17504e-23, 7.1955316e-23], [2312.0648, 5.6204e-22, 1.19851e-22, 7.19108e-23, 7.2126406e-23], [2317.3946, 5.63301e-22, 1.2012e-22, 7.20719e-23, 7.2297912e-23], [2322.7368, 5.64566e-22, 1.20389e-22, 7.22334e-23, 7.2469816e-23], [2328.0913, 5.65833e-22, 1.20659e-22, 7.23954e-23, 7.2642113e-23], [2333.4581, 5.67104e-22, 1.2093e-22, 7.25579e-23, 7.2814798e-23], [2338.8372, 5.68378e-22, 1.21201e-22, 7.27207e-23, 7.2987874e-23], [2344.2288, 5.69654e-22, 1.21473e-22, 7.28839e-23, 7.3161338e-23], [2349.6328, 5.70934e-22, 1.21746e-22, 7.30476e-23, 7.3335192e-23], [2355.0493, 5.72217e-22, 1.22019e-22, 7.32116e-23, 7.3509437e-23], [2360.4782, 5.73503e-22, 1.22293e-22, 7.3376e-23, 7.3684073e-23], [2365.9197, 5.74791e-22, 1.22568e-22, 7.35408e-23, 7.3859101e-23], [2371.3737, 5.76083e-22, 1.22843e-22, 7.3706e-23, 7.4034521e-23], [2376.8403, 5.77378e-22, 1.23119e-22, 7.38716e-23, 7.4210336e-23], [2382.3195, 5.78676e-22, 1.23396e-22, 7.40375e-23, 7.4386544e-23], [2387.8113, 5.79978e-22, 1.23673e-22, 7.42039e-23, 7.4563149e-23], [2393.3158, 5.81282e-22, 1.23951e-22, 7.43707e-23, 7.474015e-23], [2398.8329, 5.82589e-22, 1.2423e-22, 7.45378e-23, 7.4917548e-23], [2404.3628, 5.83899e-22, 1.24509e-22, 7.47054e-23, 7.5095344e-23], [2409.9054, 5.85213e-22, 1.24789e-22, 7.48734e-23, 7.5273539e-23], [2415.4608, 5.86529e-22, 1.2507e-22, 7.50417e-23, 7.5452134e-23], [2421.029, 5.87849e-22, 1.25351e-22, 7.52105e-23, 7.5631131e-23], [2426.6101, 5.89172e-22, 1.25633e-22, 7.53796e-23, 7.5810529e-23], [2432.204, 5.90498e-22, 1.25915e-22, 7.55492e-23, 7.599033e-23], [2437.8108, 5.91827e-22, 1.26199e-22, 7.57191e-23, 7.6170535e-23], [2443.4306, 5.93159e-22, 1.26482e-22, 7.58895e-23, 7.6351145e-23], [2449.0632, 5.94494e-22, 1.26767e-22, 7.60602e-23, 7.6532161e-23], [2454.7089, 5.95833e-22, 1.27052e-22, 7.62314e-23, 7.6713583e-23], [2460.3676, 5.97174e-22, 1.27338e-22, 7.64029e-23, 7.6895412e-23], [2466.0393, 5.98519e-22, 1.27625e-22, 7.65749e-23, 7.7077651e-23], [2471.7241, 5.99867e-22, 1.27912e-22, 7.67473e-23, 7.7260299e-23], [2477.4221, 6.01218e-22, 1.282e-22, 7.692e-23, 7.7443357e-23], [2483.1331, 6.02572e-22, 1.28489e-22, 7.70932e-23, 7.7626827e-23], [2488.8573, 6.0393e-22, 1.28778e-22, 7.72668e-23, 7.7810709e-23], [2494.5947, 6.0529e-22, 1.29068e-22, 7.74408e-23, 7.7995005e-23], [2500.3454, 6.06654e-22, 1.29359e-22, 7.76152e-23, 7.8179715e-23], [2506.1093, 6.08022e-22, 1.2965e-22, 7.77901e-23, 7.836484e-23], [2511.8864, 6.09392e-22, 1.29942e-22, 7.79653e-23, 7.8550382e-23], [2517.6769, 6.10765e-22, 1.30235e-22, 7.8141e-23, 7.8736341e-23], [2523.4808, 6.12142e-22, 1.30528e-22, 7.8317e-23, 7.8922718e-23], [2529.298, 6.13522e-22, 1.30823e-22, 7.84935e-23, 7.9109515e-23], [2535.1286, 6.14906e-22, 1.31117e-22, 7.86704e-23, 7.9296731e-23], [2540.9727, 6.16292e-22, 1.31413e-22, 7.88477e-23, 7.9484369e-23], [2546.8303, 6.17682e-22, 1.31709e-22, 7.90255e-23, 7.9672429e-23], [2552.7013, 6.19075e-22, 1.32006e-22, 7.92036e-23, 7.9860913e-23], [2558.5859, 6.20472e-22, 1.32304e-22, 7.93822e-23, 8.004982e-23], [2564.484, 6.21871e-22, 1.32602e-22, 7.95612e-23, 8.0239153e-23], [2570.3958, 6.23274e-22, 1.32901e-22, 7.97406e-23, 8.0428912e-23], [2576.3212, 6.24681e-22, 1.33201e-22, 7.99204e-23, 8.0619098e-23], [2582.2602, 6.2609e-22, 1.33501e-22, 8.01007e-23, 8.0809713e-23], [2588.2129, 6.27503e-22, 1.33802e-22, 8.02814e-23, 8.1000757e-23], [2594.1794, 6.28919e-22, 1.34104e-22, 8.04625e-23, 8.119223e-23], [2600.1596, 6.30339e-22, 1.34407e-22, 8.06441e-23, 8.1384135e-23], [2606.1535, 6.31762e-22, 1.3471e-22, 8.0826e-23, 8.1576473e-23], [2612.1614, 6.33188e-22, 1.35014e-22, 8.10084e-23, 8.1769243e-23], [2618.183, 6.34618e-22, 1.35319e-22, 8.11912e-23, 8.1962448e-23], [2624.2185, 6.36051e-22, 1.35624e-22, 8.13745e-23, 8.2156088e-23], [2630.268, 6.37487e-22, 1.3593e-22, 8.15582e-23, 8.2350165e-23], [2636.3314, 6.38927e-22, 1.36237e-22, 8.17423e-23, 8.2544678e-23], [2642.4088, 6.4037e-22, 1.36545e-22, 8.19269e-23, 8.273963e-23], [2648.5001, 6.41816e-22, 1.36853e-22, 8.21118e-23, 8.2935022e-23], [2654.6056, 6.43266e-22, 1.37162e-22, 8.22973e-23, 8.3130854e-23], [2660.7251, 6.4472e-22, 1.37472e-22, 8.24831e-23, 8.3327127e-23], [2666.8587, 6.46176e-22, 1.37782e-22, 8.26694e-23, 8.3523843e-23], [2673.0064, 6.47636e-22, 1.38094e-22, 8.28562e-23, 8.3721003e-23], [2679.1683, 6.491e-22, 1.38406e-22, 8.30433e-23, 8.3918607e-23], [2685.3444, 6.50567e-22, 1.38718e-22, 8.3231e-23, 8.4116656e-23], [2691.5348, 6.52038e-22, 1.39032e-22, 8.3419e-23, 8.4315152e-23], [2697.7394, 6.53511e-22, 1.39346e-22, 8.36075e-23, 8.4514096e-23], [2703.9584, 6.54989e-22, 1.39661e-22, 8.37965e-23, 8.4713489e-23], [2710.1916, 6.5647e-22, 1.39976e-22, 8.39858e-23, 8.4913332e-23], [2716.4393, 6.57954e-22, 1.40293e-22, 8.41757e-23, 8.5113625e-23], [2722.7013, 6.59442e-22, 1.4061e-22, 8.4366e-23, 8.531437e-23], [2728.9778, 6.60933e-22, 1.40928e-22, 8.45567e-23, 8.5515569e-23], [2735.2687, 6.62428e-22, 1.41246e-22, 8.47479e-23, 8.5717221e-23], [2741.5742, 6.63927e-22, 1.41566e-22, 8.49395e-23, 8.5919328e-23], [2747.8942, 6.65428e-22, 1.41886e-22, 8.51315e-23, 8.6121892e-23], [2754.2287, 6.66934e-22, 1.42207e-22, 8.53241e-23, 8.6324913e-23], [2760.5779, 6.68443e-22, 1.42528e-22, 8.55171e-23, 8.6528392e-23], [2766.9416, 6.69955e-22, 1.42851e-22, 8.57105e-23, 8.6732331e-23], [2773.3201, 6.71471e-22, 1.43174e-22, 8.59044e-23, 8.693673e-23], [2779.7133, 6.72991e-22, 1.43498e-22, 8.60987e-23, 8.714159e-23], [2786.1212, 6.74514e-22, 1.43823e-22, 8.62935e-23, 8.7346914e-23], [2792.5438, 6.76041e-22, 1.44148e-22, 8.64888e-23, 8.7552701e-23], [2798.9813, 6.77571e-22, 1.44474e-22, 8.66845e-23, 8.7758952e-23], [2805.4336, 6.79105e-22, 1.44801e-22, 8.68807e-23, 8.796567e-23], [2811.9008, 6.80642e-22, 1.45129e-22, 8.70773e-23, 8.8172855e-23], [2818.3829, 6.82184e-22, 1.45457e-22, 8.72744e-23, 8.8380508e-23], [2824.88, 6.83728e-22, 1.45787e-22, 8.74719e-23, 8.8588629e-23], [2831.392, 6.85277e-22, 1.46117e-22, 8.767e-23, 8.8797222e-23], [2837.919, 6.86829e-22, 1.46447e-22, 8.78685e-23, 8.9006285e-23], [2844.4611, 6.88384e-22, 1.46779e-22, 8.80674e-23, 8.9215821e-23], [2851.0183, 6.89943e-22, 1.47111e-22, 8.82668e-23, 8.9425831e-23], [2857.5905, 6.91506e-22, 1.47445e-22, 8.84667e-23, 8.9636315e-23], [2864.178, 6.93073e-22, 1.47778e-22, 8.86671e-23, 8.9847276e-23], [2870.7806, 6.94643e-22, 1.48113e-22, 8.88679e-23, 9.0058713e-23], [2877.3984, 6.96217e-22, 1.48449e-22, 8.90692e-23, 9.0270628e-23], [2884.0315, 6.97795e-22, 1.48785e-22, 8.9271e-23, 9.0483022e-23], [2890.6799, 6.99376e-22, 1.49122e-22, 8.94732e-23, 9.0695897e-23], [2897.3436, 7.00961e-22, 1.4946e-22, 8.96759e-23, 9.0909253e-23], [2904.0227, 7.0255e-22, 1.49799e-22, 8.98791e-23, 9.1123092e-23], [2910.7171, 7.04143e-22, 1.50138e-22, 9.00828e-23, 9.1337414e-23], [2917.427, 7.05739e-22, 1.50478e-22, 9.0287e-23, 9.1552222e-23], [2924.1524, 7.07339e-22, 1.50819e-22, 9.04916e-23, 9.1767515e-23], [2930.8932, 7.08942e-22, 1.51161e-22, 9.06967e-23, 9.1983296e-23], [2937.6497, 7.1055e-22, 1.51504e-22, 9.09023e-23, 9.2199564e-23], [2944.4216, 7.12161e-22, 1.51847e-22, 9.11084e-23, 9.2416323e-23], [2951.2092, 7.13776e-22, 1.52192e-22, 9.13149e-23, 9.2633572e-23], [2958.0125, 7.15395e-22, 1.52537e-22, 9.1522e-23, 9.2851314e-23], [2964.8314, 7.17018e-22, 1.52883e-22, 9.17295e-23, 9.3069549e-23], [2971.666, 7.18644e-22, 1.53229e-22, 9.19376e-23, 9.3288279e-23], [2978.5164, 7.20274e-22, 1.53577e-22, 9.21461e-23, 9.3507508e-23], [2985.3826, 7.21909e-22, 1.53925e-22, 9.23553e-23, 9.372724e-23], [2992.2646, 7.23547e-22, 1.54275e-22, 9.2565e-23, 9.3947488e-23], [2999.1625, 7.2519e-22, 1.54627e-22, 9.27761e-23, 9.4168317e-23], [3006.0763, 7.26852e-22, 1.54996e-22, 9.29978e-23, 9.4390767e-23], [3013.006, 7.28512e-22, 1.55361e-22, 9.32165e-23, 9.461333e-23], [3019.9517, 7.30139e-22, 1.55682e-22, 9.34091e-23, 9.4833488e-23], [3026.9134, 7.31794e-22, 1.56033e-22, 9.36198e-23, 9.505607e-23], [3033.8912, 7.33454e-22, 1.56387e-22, 9.38319e-23, 9.5279258e-23], [3040.885, 7.35119e-22, 1.56741e-22, 9.40447e-23, 9.5502974e-23], [3047.895, 7.36788e-22, 1.57097e-22, 9.42581e-23, 9.5727203e-23], [3054.9211, 7.3846e-22, 1.57453e-22, 9.4472e-23, 9.5951944e-23], [3061.9634, 7.40137e-22, 1.57811e-22, 9.46864e-23, 9.6177196e-23], [3069.022, 7.41817e-22, 1.58169e-22, 9.49013e-23, 9.6402958e-23], [3076.0968, 7.43502e-22, 1.58528e-22, 9.51167e-23, 9.6629233e-23], [3083.188, 7.4519e-22, 1.58888e-22, 9.53327e-23, 9.6856021e-23], [3090.2954, 7.46883e-22, 1.59249e-22, 9.55491e-23, 9.7083323e-23], [3097.4193, 7.48579e-22, 1.5961e-22, 9.57661e-23, 9.7311141e-23], [3104.5596, 7.50279e-22, 1.59973e-22, 9.59835e-23, 9.7539475e-23], [3111.7163, 7.51983e-22, 1.60336e-22, 9.62015e-23, 9.7768327e-23], [3118.8896, 7.53692e-22, 1.607e-22, 9.642e-23, 9.7997697e-23], [3126.0794, 7.55404e-22, 1.61065e-22, 9.6639e-23, 9.8227588e-23], [3133.2857, 7.5712e-22, 1.61431e-22, 9.68585e-23, 9.8458e-23], [3140.5087, 7.5884e-22, 1.61797e-22, 9.70785e-23, 9.8688934e-23], [3147.7483, 7.60565e-22, 1.62165e-22, 9.7299e-23, 9.8920392e-23], [3155.0046, 7.62293e-22, 1.62533e-22, 9.75201e-23, 9.9152375e-23], [3162.2777, 7.64025e-22, 1.62903e-22, 9.77416e-23, 9.9384884e-23], [3169.5675, 7.65762e-22, 1.63273e-22, 9.79637e-23, 9.961792e-23], [3176.8741, 7.67502e-22, 1.63644e-22, 9.81863e-23, 9.9851485e-23], [3184.1975, 7.69246e-22, 1.64016e-22, 9.84095e-23, 1.0008558e-22], [3191.5379, 7.70995e-22, 1.64389e-22, 9.86331e-23, 1.0032021e-22], [3198.8951, 7.72748e-22, 1.64762e-22, 9.88573e-23, 1.0055536e-22], [3206.2693, 7.74505e-22, 1.65137e-22, 9.9082e-23, 1.0079106e-22], [3213.6605, 7.76265e-22, 1.65512e-22, 9.93072e-23, 1.0102728e-22], [3221.0688, 7.7803e-22, 1.65888e-22, 9.95329e-23, 1.0126405e-22], [3228.4941, 7.798e-22, 1.66265e-22, 9.97592e-23, 1.0150135e-22], [3235.9366, 7.81573e-22, 1.66643e-22, 9.9986e-23, 1.0173918e-22], [3243.3962, 7.8335e-22, 1.67022e-22, 1.00213e-22, 1.0197756e-22], [3250.873, 7.85132e-22, 1.67402e-22, 1.00441e-22, 1.0221648e-22], [3258.367, 7.86917e-22, 1.67783e-22, 1.0067e-22, 1.0245594e-22], [3265.8783, 7.88707e-22, 1.68164e-22, 1.00899e-22, 1.0269595e-22], [3273.4069, 7.90501e-22, 1.68547e-22, 1.01128e-22, 1.029365e-22], [3280.9529, 7.923e-22, 1.6893e-22, 1.01358e-22, 1.0317759e-22], [3288.5163, 7.94102e-22, 1.69314e-22, 1.01589e-22, 1.0341924e-22], [3296.0971, 7.95909e-22, 1.69699e-22, 1.0182e-22, 1.0366143e-22], [3303.6954, 7.9772e-22, 1.70085e-22, 1.02051e-22, 1.0390417e-22], [3311.3112, 7.99535e-22, 1.70472e-22, 1.02283e-22, 1.0414747e-22], [3318.9446, 8.01354e-22, 1.7086e-22, 1.02516e-22, 1.0439131e-22], [3326.5955, 8.03178e-22, 1.71249e-22, 1.02749e-22, 1.0463571e-22], [3334.2641, 8.05006e-22, 1.71639e-22, 1.02983e-22, 1.0488067e-22], [3341.9504, 8.06838e-22, 1.72029e-22, 1.03217e-22, 1.0512618e-22], [3349.6544, 8.08674e-22, 1.72421e-22, 1.03452e-22, 1.0537225e-22], [3357.3761, 8.10515e-22, 1.72813e-22, 1.03688e-22, 1.0561888e-22], [3365.1157, 8.1236e-22, 1.73206e-22, 1.03924e-22, 1.0586607e-22], [3372.8731, 8.14209e-22, 1.736e-22, 1.0416e-22, 1.0611382e-22], [3380.6484, 8.16063e-22, 1.73996e-22, 1.04397e-22, 1.0636213e-22], [3388.4416, 8.17921e-22, 1.74392e-22, 1.04635e-22, 1.0661101e-22], [3396.2527, 8.19783e-22, 1.74789e-22, 1.04873e-22, 1.0686045e-22], [3404.0819, 8.2165e-22, 1.75187e-22, 1.05112e-22, 1.0711046e-22], [3411.9291, 8.23521e-22, 1.75585e-22, 1.05351e-22, 1.0736104e-22], [3419.7944, 8.25396e-22, 1.75985e-22, 1.05591e-22, 1.0761219e-22], [3427.6779, 8.27276e-22, 1.76386e-22, 1.05832e-22, 1.0786391e-22], [3435.5795, 8.2916e-22, 1.76788e-22, 1.06073e-22, 1.081162e-22], [3443.4993, 8.31048e-22, 1.7719e-22, 1.06314e-22, 1.0836906e-22], [3451.4374, 8.32941e-22, 1.77594e-22, 1.06556e-22, 1.086225e-22], [3459.3938, 8.34839e-22, 1.77998e-22, 1.06799e-22, 1.0887652e-22], [3467.3685, 8.3674e-22, 1.78404e-22, 1.07042e-22, 1.0913111e-22], [3475.3616, 8.38646e-22, 1.7881e-22, 1.07286e-22, 1.0938628e-22], [3483.3732, 8.40557e-22, 1.79217e-22, 1.0753e-22, 1.0964204e-22], [3491.4032, 8.42472e-22, 1.79625e-22, 1.07775e-22, 1.0989837e-22], [3499.4517, 8.44392e-22, 1.80035e-22, 1.08021e-22, 1.1015529e-22], [3507.5187, 8.46316e-22, 1.80445e-22, 1.08267e-22, 1.1041279e-22], [3515.6044, 8.48244e-22, 1.80856e-22, 1.08514e-22, 1.1067088e-22], [3523.7087, 8.50177e-22, 1.81268e-22, 1.08761e-22, 1.1092955e-22], [3531.8317, 8.52115e-22, 1.81681e-22, 1.09009e-22, 1.1118881e-22], [3539.9734, 8.54057e-22, 1.82095e-22, 1.09257e-22, 1.1144866e-22], [3548.1339, 8.56003e-22, 1.8251e-22, 1.09506e-22, 1.1170911e-22], [3556.3132, 8.57954e-22, 1.82926e-22, 1.09756e-22, 1.1197014e-22], [3564.5113, 8.5991e-22, 1.83343e-22, 1.10006e-22, 1.1223177e-22], [3572.7284, 8.6187e-22, 1.83761e-22, 1.10256e-22, 1.12494e-22], [3580.9644, 8.63835e-22, 1.8418e-22, 1.10508e-22, 1.1275682e-22], [3589.2193, 8.65804e-22, 1.84599e-22, 1.1076e-22, 1.1302024e-22], [3597.4934, 8.67778e-22, 1.8502e-22, 1.11012e-22, 1.1328426e-22], [3605.7864, 8.69757e-22, 1.85442e-22, 1.11265e-22, 1.1354888e-22], [3614.0986, 8.7174e-22, 1.85865e-22, 1.11519e-22, 1.1381411e-22], [3622.43, 8.73728e-22, 1.86288e-22, 1.11773e-22, 1.1407993e-22], [3630.7805, 8.7572e-22, 1.86713e-22, 1.12028e-22, 1.1434637e-22], [3639.1504, 8.77717e-22, 1.87139e-22, 1.12283e-22, 1.146134e-22], [3647.5395, 8.79719e-22, 1.87566e-22, 1.12539e-22, 1.1488105e-22], [3655.9479, 8.81725e-22, 1.87993e-22, 1.12796e-22, 1.1514931e-22], [3664.3757, 8.83737e-22, 1.88422e-22, 1.13053e-22, 1.1541817e-22], [3672.823, 8.85752e-22, 1.88852e-22, 1.13311e-22, 1.1568765e-22], [3681.2897, 8.87773e-22, 1.89283e-22, 1.1357e-22, 1.1595775e-22], [3689.776, 8.89798e-22, 1.89714e-22, 1.13829e-22, 1.1622845e-22], [3698.2818, 8.91828e-22, 1.90147e-22, 1.14088e-22, 1.1649978e-22], [3706.8072, 8.93862e-22, 1.90581e-22, 1.14348e-22, 1.1677172e-22], [3715.3523, 8.95902e-22, 1.91015e-22, 1.14609e-22, 1.1704428e-22], [3723.9171, 8.97946e-22, 1.91451e-22, 1.14871e-22, 1.1731746e-22], [3732.5016, 8.99995e-22, 1.91888e-22, 1.15133e-22, 1.1759127e-22], [3741.1059, 9.02048e-22, 1.92326e-22, 1.15395e-22, 1.178657e-22], [3749.73, 9.04106e-22, 1.92765e-22, 1.15659e-22, 1.1814075e-22], [3758.374, 9.0617e-22, 1.93204e-22, 1.15923e-22, 1.1841643e-22], [3767.038, 9.08238e-22, 1.93645e-22, 1.16187e-22, 1.1869274e-22], [3775.7219, 9.1031e-22, 1.94087e-22, 1.16452e-22, 1.1896968e-22], [3784.4258, 9.12388e-22, 1.9453e-22, 1.16718e-22, 1.1924725e-22], [3793.1498, 9.14471e-22, 1.94974e-22, 1.16984e-22, 1.1952545e-22], [3801.894, 9.16558e-22, 1.95419e-22, 1.17251e-22, 1.1980429e-22], [3810.6582, 9.1865e-22, 1.95865e-22, 1.17519e-22, 1.2008377e-22], [3819.4427, 9.20748e-22, 1.96313e-22, 1.17788e-22, 1.2036392e-22], [3828.2474, 9.22856e-22, 1.96769e-22, 1.18061e-22, 1.2064518e-22], [3837.0725, 9.24962e-22, 1.97217e-22, 1.1833e-22, 1.2092649e-22], [3845.9178, 9.27068e-22, 1.9766e-22, 1.18596e-22, 1.2120807e-22], [3854.7836, 9.29184e-22, 1.98111e-22, 1.18867e-22, 1.2149071e-22], [3863.6698, 9.31306e-22, 1.98563e-22, 1.19138e-22, 1.2177401e-22], [3872.5764, 9.33432e-22, 1.99016e-22, 1.1941e-22, 1.2205797e-22], [3881.5037, 9.35564e-22, 1.99471e-22, 1.19682e-22, 1.2234257e-22], [3890.4514, 9.377e-22, 1.99926e-22, 1.19956e-22, 1.2262783e-22], [3899.4199, 9.39841e-22, 2.00383e-22, 1.2023e-22, 1.2291373e-22], [3908.409, 9.41988e-22, 2.0084e-22, 1.20504e-22, 1.2320029e-22], [3917.4188, 9.44139e-22, 2.01299e-22, 1.20779e-22, 1.234875e-22], [3926.4494, 9.46295e-22, 2.01758e-22, 1.21055e-22, 1.2377537e-22], [3935.5008, 9.48457e-22, 2.02219e-22, 1.21332e-22, 1.2406389e-22], [3944.573, 9.50623e-22, 2.02681e-22, 1.21609e-22, 1.2435308e-22], [3953.6662, 9.52795e-22, 2.03144e-22, 1.21886e-22, 1.2464292e-22], [3962.7803, 9.54971e-22, 2.03608e-22, 1.22165e-22, 1.2493342e-22], [3971.9155, 9.57153e-22, 2.04073e-22, 1.22444e-22, 1.2522458e-22], [3981.0717, 9.59339e-22, 2.04539e-22, 1.22724e-22, 1.2551641e-22], [3990.249, 9.61531e-22, 2.05006e-22, 1.23004e-22, 1.2580891e-22], [3999.4475, 9.63728e-22, 2.05475e-22, 1.23285e-22, 1.2610207e-22], [4008.6672, 9.6593e-22, 2.05944e-22, 1.23567e-22, 1.263959e-22], [4017.9081, 9.68137e-22, 2.06415e-22, 1.23849e-22, 1.266904e-22], [4027.1703, 9.70349e-22, 2.06886e-22, 1.24132e-22, 1.2698557e-22], [4036.4539, 9.72566e-22, 2.07359e-22, 1.24415e-22, 1.2728142e-22], [4045.7589, 9.74789e-22, 2.07833e-22, 1.247e-22, 1.2757794e-22], [4055.0854, 9.77016e-22, 2.08308e-22, 1.24985e-22, 1.2787514e-22], [4064.4333, 9.79249e-22, 2.08784e-22, 1.2527e-22, 1.2817302e-22], [4073.8028, 9.81487e-22, 2.09261e-22, 1.25557e-22, 1.2847157e-22], [4083.1939, 9.8373e-22, 2.09739e-22, 1.25843e-22, 1.2877081e-22], [4092.6066, 9.85979e-22, 2.10218e-22, 1.26131e-22, 1.2907073e-22], [4102.041, 9.88233e-22, 2.10699e-22, 1.26419e-22, 1.2937133e-22], [4111.4972, 9.90491e-22, 2.1118e-22, 1.26708e-22, 1.2967263e-22], [4120.9752, 9.92756e-22, 2.11663e-22, 1.26998e-22, 1.299746e-22], [4130.475, 9.95025e-22, 2.12147e-22, 1.27288e-22, 1.3027727e-22], [4139.9967, 9.973e-22, 2.12632e-22, 1.27579e-22, 1.3058063e-22], [4149.5404, 9.9958e-22, 2.13118e-22, 1.27871e-22, 1.3088468e-22], [4159.1061, 1.00186e-21, 2.13605e-22, 1.28163e-22, 1.3118943e-22], [4168.6938, 1.00416e-21, 2.14093e-22, 1.28456e-22, 1.3149487e-22], [4178.3037, 1.00645e-21, 2.14583e-22, 1.2875e-22, 1.3180101e-22], [4187.9357, 1.00875e-21, 2.15073e-22, 1.29044e-22, 1.3210785e-22], [4197.5898, 1.01106e-21, 2.15565e-22, 1.29339e-22, 1.3241538e-22], [4207.2663, 1.01337e-21, 2.16058e-22, 1.29635e-22, 1.3272363e-22], [4216.965, 1.01569e-21, 2.16552e-22, 1.29931e-22, 1.3303257e-22], [4226.6861, 1.01801e-21, 2.17047e-22, 1.30228e-22, 1.3334222e-22], [4236.4297, 1.02034e-21, 2.17544e-22, 1.30526e-22, 1.3365258e-22], [4246.1956, 1.02267e-21, 2.18041e-22, 1.30825e-22, 1.3396364e-22], [4255.9841, 1.02501e-21, 2.1854e-22, 1.31124e-22, 1.3427542e-22], [4265.7952, 1.02736e-21, 2.19039e-22, 1.31424e-22, 1.3458791e-22], [4275.6289, 1.02971e-21, 2.1954e-22, 1.31724e-22, 1.3490111e-22], [4285.4852, 1.03206e-21, 2.20043e-22, 1.32026e-22, 1.3521503e-22], [4295.3643, 1.03442e-21, 2.20546e-22, 1.32328e-22, 1.3552966e-22], [4305.2661, 1.03679e-21, 2.2105e-22, 1.3263e-22, 1.3584501e-22], [4315.1908, 1.03916e-21, 2.21556e-22, 1.32934e-22, 1.3616109e-22], [4325.1383, 1.04154e-21, 2.22063e-22, 1.33238e-22, 1.3647788e-22], [4335.1088, 1.04392e-21, 2.22571e-22, 1.33542e-22, 1.367954e-22], [4345.1022, 1.04631e-21, 2.2308e-22, 1.33848e-22, 1.3711365e-22], [4355.1187, 1.0487e-21, 2.2359e-22, 1.34154e-22, 1.3743262e-22], [4365.1583, 1.0511e-21, 2.24102e-22, 1.34461e-22, 1.3775232e-22], [4375.2211, 1.05351e-21, 2.24614e-22, 1.34769e-22, 1.3807275e-22], [4385.307, 1.05592e-21, 2.25128e-22, 1.35077e-22, 1.3839392e-22], [4395.4162, 1.05833e-21, 2.25643e-22, 1.35386e-22, 1.3871581e-22], [4405.5486, 1.06076e-21, 2.2616e-22, 1.35696e-22, 1.3903845e-22], [4415.7045, 1.06318e-21, 2.26677e-22, 1.36006e-22, 1.3936182e-22], [4425.8837, 1.06562e-21, 2.27196e-22, 1.36318e-22, 1.3968593e-22], [4436.0864, 1.06805e-21, 2.27716e-22, 1.36629e-22, 1.4001078e-22], [4446.3127, 1.0705e-21, 2.28237e-22, 1.36942e-22, 1.4033637e-22], [4456.5625, 1.07295e-21, 2.28759e-22, 1.37256e-22, 1.4066271e-22], [4466.8359, 1.0754e-21, 2.29283e-22, 1.3757e-22, 1.4098979e-22], [4477.133, 1.07787e-21, 2.29808e-22, 1.37885e-22, 1.4131762e-22], [4487.4539, 1.08033e-21, 2.30333e-22, 1.382e-22, 1.416462e-22], [4497.7985, 1.08281e-21, 2.30861e-22, 1.38516e-22, 1.4197553e-22], [4508.167, 1.08528e-21, 2.31389e-22, 1.38833e-22, 1.4230562e-22], [4518.5594, 1.08777e-21, 2.31919e-22, 1.39151e-22, 1.4263646e-22], [4528.9758, 1.09026e-21, 2.3245e-22, 1.3947e-22, 1.4296805e-22], [4539.4162, 1.09275e-21, 2.32982e-22, 1.39789e-22, 1.4330041e-22], [4549.8806, 1.09526e-21, 2.33515e-22, 1.40109e-22, 1.4363352e-22], [4560.3692, 1.09776e-21, 2.3405e-22, 1.4043e-22, 1.4396739e-22], [4570.8819, 1.10028e-21, 2.34585e-22, 1.40751e-22, 1.4430203e-22], [4581.4189, 1.1028e-21, 2.35122e-22, 1.41073e-22, 1.4463744e-22], [4591.9801, 1.10532e-21, 2.35661e-22, 1.41396e-22, 1.4497361e-22], [4602.5657, 1.10785e-21, 2.362e-22, 1.4172e-22, 1.4531055e-22], [4613.1757, 1.11039e-21, 2.36741e-22, 1.42045e-22, 1.4564826e-22], [4623.8102, 1.11293e-21, 2.37283e-22, 1.4237e-22, 1.4598674e-22], [4634.4692, 1.11548e-21, 2.37827e-22, 1.42696e-22, 1.4632599e-22], [4645.1528, 1.11803e-21, 2.38371e-22, 1.43023e-22, 1.4666603e-22], [4655.8609, 1.12059e-21, 2.38917e-22, 1.4335e-22, 1.4700684e-22], [4666.5938, 1.12316e-21, 2.39464e-22, 1.43678e-22, 1.4734843e-22], [4677.3514, 1.12573e-21, 2.40012e-22, 1.44007e-22, 1.476908e-22], [4688.1338, 1.12831e-21, 2.40562e-22, 1.44337e-22, 1.4803395e-22], [4698.9411, 1.1309e-21, 2.41113e-22, 1.44668e-22, 1.4837789e-22], [4709.7733, 1.13349e-21, 2.41665e-22, 1.44999e-22, 1.4872262e-22], [4720.6304, 1.13608e-21, 2.42219e-22, 1.45331e-22, 1.4906814e-22], [4731.5126, 1.13868e-21, 2.42774e-22, 1.45664e-22, 1.4941444e-22], [4742.4199, 1.14129e-21, 2.4333e-22, 1.45998e-22, 1.4976156e-22], [4753.3523, 1.14391e-21, 2.43893e-22, 1.46336e-22, 1.5010979e-22], [4764.3099, 1.14653e-21, 2.44447e-22, 1.46668e-22, 1.5045822e-22], [4775.2927, 1.14915e-21, 2.45006e-22, 1.47003e-22, 1.5080761e-22], [4786.3009, 1.15179e-21, 2.45567e-22, 1.4734e-22, 1.5115789e-22], [4797.3345, 1.15443e-21, 2.46129e-22, 1.47678e-22, 1.5150897e-22], [4808.3935, 1.15707e-21, 2.46693e-22, 1.48016e-22, 1.5186086e-22], [4819.478, 1.15972e-21, 2.47258e-22, 1.48355e-22, 1.5221356e-22], [4830.588, 1.16238e-21, 2.47825e-22, 1.48695e-22, 1.5256706e-22], [4841.7237, 1.16504e-21, 2.48393e-22, 1.49036e-22, 1.5292137e-22], [4852.885, 1.16771e-21, 2.48962e-22, 1.49377e-22, 1.5327649e-22], [4864.0721, 1.17039e-21, 2.49532e-22, 1.49719e-22, 1.5363243e-22], [4875.2849, 1.17307e-21, 2.50104e-22, 1.50062e-22, 1.5398918e-22], [4886.5236, 1.17576e-21, 2.50677e-22, 1.50406e-22, 1.5434674e-22], [4897.7882, 1.17845e-21, 2.51251e-22, 1.50751e-22, 1.5470513e-22], [4909.0788, 1.18115e-21, 2.51827e-22, 1.51096e-22, 1.5506433e-22], [4920.3954, 1.18386e-21, 2.52404e-22, 1.51442e-22, 1.5542436e-22], [4931.738, 1.18657e-21, 2.52982e-22, 1.51789e-22, 1.5578521e-22], [4943.1069, 1.18929e-21, 2.53562e-22, 1.52137e-22, 1.5614689e-22], [4954.5019, 1.19202e-21, 2.54143e-22, 1.52486e-22, 1.565094e-22], [4965.9232, 1.19475e-21, 2.54726e-22, 1.52835e-22, 1.5687273e-22], [4977.3708, 1.19749e-21, 2.55309e-22, 1.53186e-22, 1.572369e-22], [4988.8449, 1.20023e-21, 2.55895e-22, 1.53537e-22, 1.576019e-22], [5000.3453, 1.20298e-21, 2.56481e-22, 1.53889e-22, 1.5796774e-22], [5011.8723, 1.20574e-21, 2.57069e-22, 1.54241e-22, 1.5833441e-22], [5023.4259, 1.2085e-21, 2.57658e-22, 1.54595e-22, 1.5870193e-22], [5035.0061, 1.21127e-21, 2.58249e-22, 1.54949e-22, 1.5907028e-22], [5046.613, 1.21405e-21, 2.58841e-22, 1.55304e-22, 1.5943948e-22], [5058.2466, 1.21683e-21, 2.59434e-22, 1.5566e-22, 1.5980953e-22], [5069.9071, 1.21962e-21, 2.60029e-22, 1.56017e-22, 1.6018042e-22], [5081.5944, 1.22242e-21, 2.60625e-22, 1.56375e-22, 1.6055216e-22], [5093.3087, 1.22522e-21, 2.61222e-22, 1.56733e-22, 1.6092475e-22], [5105.05, 1.22803e-21, 2.61821e-22, 1.57093e-22, 1.612982e-22], [5116.8184, 1.23084e-21, 2.62421e-22, 1.57453e-22, 1.616725e-22], [5128.6138, 1.23367e-21, 2.63023e-22, 1.57814e-22, 1.6204766e-22], [5140.4365, 1.2365e-21, 2.63626e-22, 1.58176e-22, 1.6242368e-22], [5152.2864, 1.23933e-21, 2.6423e-22, 1.58538e-22, 1.6280056e-22], [5164.1637, 1.24217e-21, 2.64836e-22, 1.58902e-22, 1.631783e-22], [5176.0683, 1.24502e-21, 2.65443e-22, 1.59266e-22, 1.6355691e-22], [5188.0004, 1.24788e-21, 2.66052e-22, 1.59631e-22, 1.6393638e-22], [5199.96, 1.25074e-21, 2.66662e-22, 1.59997e-22, 1.6431672e-22], [5211.9471, 1.2536e-21, 2.67274e-22, 1.60364e-22, 1.6469794e-22], [5223.9619, 1.25648e-21, 2.67886e-22, 1.60732e-22, 1.6508003e-22], [5236.0044, 1.25936e-21, 2.68501e-22, 1.611e-22, 1.6546299e-22], [5248.0746, 1.26225e-21, 2.69116e-22, 1.6147e-22, 1.6584683e-22], [5260.1727, 1.26514e-21, 2.69734e-22, 1.6184e-22, 1.6623156e-22], [5272.2986, 1.26805e-21, 2.70352e-22, 1.62211e-22, 1.6661716e-22], [5284.4525, 1.27095e-21, 2.70972e-22, 1.62583e-22, 1.6700364e-22], [5296.6344, 1.27387e-21, 2.71594e-22, 1.62956e-22, 1.6739101e-22], [5308.8444, 1.27679e-21, 2.72217e-22, 1.6333e-22, 1.6777927e-22], [5321.0826, 1.27972e-21, 2.72841e-22, 1.63705e-22, 1.6816842e-22], [5333.349, 1.28265e-21, 2.73467e-22, 1.6408e-22, 1.6855846e-22], [5345.6436, 1.2856e-21, 2.74094e-22, 1.64456e-22, 1.689494e-22], [5357.9666, 1.28854e-21, 2.74723e-22, 1.64834e-22, 1.6934123e-22], [5370.318, 1.2915e-21, 2.75353e-22, 1.65212e-22, 1.6973395e-22], [5382.6978, 1.29446e-21, 2.75984e-22, 1.65591e-22, 1.7012758e-22], [5395.1062, 1.29743e-21, 2.76617e-22, 1.6597e-22, 1.7052211e-22], [5407.5432, 1.30041e-21, 2.77252e-22, 1.66351e-22, 1.7091755e-22], [5420.0089, 1.30339e-21, 2.77888e-22, 1.66733e-22, 1.7131389e-22], [5432.5033, 1.30638e-21, 2.78525e-22, 1.67115e-22, 1.7171114e-22], [5445.0265, 1.30938e-21, 2.79164e-22, 1.67499e-22, 1.7210929e-22], [5457.5786, 1.31238e-21, 2.79805e-22, 1.67883e-22, 1.7250837e-22], [5470.1596, 1.31539e-21, 2.80447e-22, 1.68268e-22, 1.7290835e-22], [5482.7696, 1.31841e-21, 2.8109e-22, 1.68654e-22, 1.7330926e-22], [5495.4087, 1.32144e-21, 2.81735e-22, 1.69041e-22, 1.7371108e-22], [5508.077, 1.32447e-21, 2.82381e-22, 1.69429e-22, 1.7411382e-22], [5520.7744, 1.32751e-21, 2.83029e-22, 1.69817e-22, 1.7451749e-22], [5533.5011, 1.33055e-21, 2.83678e-22, 1.70207e-22, 1.7492208e-22], [5546.2571, 1.33361e-21, 2.84329e-22, 1.70598e-22, 1.753276e-22], [5559.0426, 1.33667e-21, 2.84982e-22, 1.70989e-22, 1.7573405e-22], [5571.8575, 1.33973e-21, 2.85636e-22, 1.71381e-22, 1.7614143e-22], [5584.7019, 1.34281e-21, 2.86291e-22, 1.71775e-22, 1.7654974e-22], [5597.576, 1.34589e-21, 2.86948e-22, 1.72169e-22, 1.76959e-22], [5610.4798, 1.34898e-21, 2.87606e-22, 1.72564e-22, 1.7736919e-22], [5623.4133, 1.35207e-21, 2.88266e-22, 1.7296e-22, 1.7778032e-22], [5636.3766, 1.35518e-21, 2.88928e-22, 1.73357e-22, 1.7819239e-22], [5649.3697, 1.35829e-21, 2.89591e-22, 1.73755e-22, 1.786054e-22], [5662.3929, 1.3614e-21, 2.90256e-22, 1.74153e-22, 1.7901937e-22], [5675.4461, 1.36453e-21, 2.90922e-22, 1.74553e-22, 1.7943428e-22], [5688.5293, 1.36766e-21, 2.91589e-22, 1.74954e-22, 1.7985015e-22], [5701.6427, 1.3708e-21, 2.92258e-22, 1.75355e-22, 1.8026697e-22], [5714.7864, 1.37394e-21, 2.92929e-22, 1.75758e-22, 1.8068474e-22], [5727.9603, 1.3771e-21, 2.93602e-22, 1.76161e-22, 1.8110347e-22], [5741.1646, 1.38026e-21, 2.94275e-22, 1.76565e-22, 1.8152317e-22], [5754.3994, 1.38343e-21, 2.94951e-22, 1.76971e-22, 1.8194382e-22], [5767.6646, 1.3866e-21, 2.95628e-22, 1.77377e-22, 1.8236545e-22], [5780.9605, 1.38979e-21, 2.96311e-22, 1.77787e-22, 1.8278833e-22], [5794.287, 1.39297e-21, 2.96987e-22, 1.78192e-22, 1.8321161e-22], [5807.6442, 1.39617e-21, 2.97668e-22, 1.78601e-22, 1.8363612e-22], [5821.0322, 1.39938e-21, 2.98351e-22, 1.79011e-22, 1.8406161e-22], [5834.451, 1.40259e-21, 2.99036e-22, 1.79422e-22, 1.8448809e-22], [5847.9008, 1.40581e-21, 2.99723e-22, 1.79834e-22, 1.8491555e-22], [5861.3816, 1.40904e-21, 3.00411e-22, 1.80246e-22, 1.8534398e-22], [5874.8935, 1.41227e-21, 3.011e-22, 1.8066e-22, 1.857734e-22], [5888.4366, 1.41551e-21, 3.01792e-22, 1.81075e-22, 1.862038e-22], [5902.0108, 1.41876e-21, 3.02484e-22, 1.81491e-22, 1.8663519e-22], [5915.6163, 1.42202e-21, 3.03179e-22, 1.81907e-22, 1.8706758e-22], [5929.2532, 1.42528e-21, 3.03875e-22, 1.82325e-22, 1.8750095e-22], [5942.9216, 1.42856e-21, 3.04573e-22, 1.82744e-22, 1.8793531e-22], [5956.6214, 1.43184e-21, 3.05272e-22, 1.83163e-22, 1.8837068e-22], [5970.3529, 1.43512e-21, 3.05973e-22, 1.83584e-22, 1.8880704e-22], [5984.116, 1.43842e-21, 3.06675e-22, 1.84005e-22, 1.892444e-22], [5997.9108, 1.44172e-21, 3.07379e-22, 1.84428e-22, 1.8968277e-22], [6011.7374, 1.44503e-21, 3.08085e-22, 1.84851e-22, 1.9012214e-22], [6025.5959, 1.44835e-21, 3.08792e-22, 1.85275e-22, 1.9056252e-22], [6039.4863, 1.45168e-21, 3.09501e-22, 1.85701e-22, 1.9100391e-22], [6053.4087, 1.45501e-21, 3.10212e-22, 1.86127e-22, 1.9144631e-22], [6067.3633, 1.45835e-21, 3.10924e-22, 1.86555e-22, 1.9188973e-22], [6081.35, 1.4617e-21, 3.11638e-22, 1.86983e-22, 1.9233417e-22], [6095.369, 1.46506e-21, 3.12354e-22, 1.87412e-22, 1.9277962e-22], [6109.4202, 1.46842e-21, 3.13071e-22, 1.87843e-22, 1.932261e-22], [6123.5039, 1.47179e-21, 3.1379e-22, 1.88274e-22, 1.936736e-22], [6137.6201, 1.47517e-21, 3.14511e-22, 1.88706e-22, 1.9412213e-22], [6151.7687, 1.47856e-21, 3.15233e-22, 1.8914e-22, 1.9457169e-22], [6165.95, 1.48196e-21, 3.15957e-22, 1.89574e-22, 1.9502228e-22], [6180.164, 1.48536e-21, 3.16683e-22, 1.9001e-22, 1.9547391e-22], [6194.4108, 1.48877e-21, 3.1741e-22, 1.90446e-22, 1.9592657e-22], [6208.6903, 1.49219e-21, 3.18139e-22, 1.90883e-22, 1.9638027e-22], [6223.0029, 1.49562e-21, 3.1887e-22, 1.91322e-22, 1.9683501e-22], [6237.3484, 1.49905e-21, 3.19602e-22, 1.91761e-22, 1.9729079e-22], [6251.7269, 1.5025e-21, 3.20336e-22, 1.92202e-22, 1.9774762e-22], [6266.1386, 1.50595e-21, 3.21072e-22, 1.92643e-22, 1.982055e-22], [6280.5836, 1.50941e-21, 3.21809e-22, 1.93085e-22, 1.9866443e-22], [6295.0618, 1.51287e-21, 3.22548e-22, 1.93529e-22, 1.9912442e-22], [6309.5734, 1.51635e-21, 3.23289e-22, 1.93973e-22, 1.9958545e-22], [6324.1185, 1.51983e-21, 3.24032e-22, 1.94419e-22, 2.0004755e-22], [6338.6971, 1.52332e-21, 3.24776e-22, 1.94866e-22, 2.0051071e-22], [6353.3093, 1.52682e-21, 3.25522e-22, 1.95313e-22, 2.0097493e-22], [6367.9552, 1.53033e-21, 3.2627e-22, 1.95762e-22, 2.0144022e-22], [6382.6349, 1.53384e-21, 3.27019e-22, 1.96212e-22, 2.0190657e-22], [6397.3484, 1.53737e-21, 3.2777e-22, 1.96662e-22, 2.02374e-22], [6412.0958, 1.5409e-21, 3.28523e-22, 1.97114e-22, 2.028425e-22], [6426.8772, 1.54444e-21, 3.29278e-22, 1.97567e-22, 2.0331207e-22], [6441.6927, 1.54799e-21, 3.30034e-22, 1.98021e-22, 2.0378272e-22], [6456.5423, 1.55154e-21, 3.30793e-22, 1.98476e-22, 2.0425446e-22], [6471.4262, 1.55511e-21, 3.31553e-22, 1.98932e-22, 2.0472727e-22], [6486.3443, 1.55868e-21, 3.32314e-22, 1.99389e-22, 2.0520117e-22], [6501.2969, 1.56226e-21, 3.33078e-22, 1.99847e-22, 2.0567616e-22], [6516.2839, 1.56585e-21, 3.33843e-22, 2.00306e-22, 2.0615224e-22], [6531.3055, 1.56945e-21, 3.3461e-22, 2.00766e-22, 2.0662941e-22], [6546.3617, 1.57305e-21, 3.35379e-22, 2.01227e-22, 2.0710768e-22], [6561.4527, 1.57667e-21, 3.36149e-22, 2.0169e-22, 2.0758705e-22], [6576.5784, 1.58029e-21, 3.36922e-22, 2.02153e-22, 2.0806751e-22], [6591.739, 1.58392e-21, 3.37696e-22, 2.02617e-22, 2.0854908e-22], [6606.9345, 1.58756e-21, 3.38472e-22, 2.03083e-22, 2.0903176e-22], [6622.165, 1.59121e-21, 3.39249e-22, 2.0355e-22, 2.0951554e-22], [6637.4307, 1.59486e-21, 3.40029e-22, 2.04017e-22, 2.1000044e-22], [6652.7316, 1.59853e-21, 3.4081e-22, 2.04486e-22, 2.1048645e-22], [6668.0677, 1.6022e-21, 3.41593e-22, 2.04956e-22, 2.1097357e-22], [6683.4392, 1.60588e-21, 3.42378e-22, 2.05427e-22, 2.1146181e-22], [6698.8461, 1.60957e-21, 3.43165e-22, 2.05899e-22, 2.1195118e-22], [6714.2885, 1.61327e-21, 3.43953e-22, 2.06372e-22, 2.1244167e-22], [6729.7666, 1.61698e-21, 3.44744e-22, 2.06846e-22, 2.1293328e-22], [6745.2803, 1.6207e-21, 3.45536e-22, 2.07322e-22, 2.1342602e-22], [6760.8298, 1.62442e-21, 3.4633e-22, 2.07798e-22, 2.139199e-22], [6776.4151, 1.62815e-21, 3.47126e-22, 2.08275e-22, 2.1441491e-22], [6792.0363, 1.63189e-21, 3.47923e-22, 2.08754e-22, 2.1491106e-22], [6807.6936, 1.63564e-21, 3.48723e-22, 2.09234e-22, 2.1540834e-22], [6823.3869, 1.6394e-21, 3.49524e-22, 2.09715e-22, 2.1590677e-22], [6839.1165, 1.64317e-21, 3.50328e-22, 2.10197e-22, 2.1640634e-22], [6854.8823, 1.64695e-21, 3.51133e-22, 2.1068e-22, 2.1690706e-22], [6870.6844, 1.65073e-21, 3.5194e-22, 2.11164e-22, 2.1740893e-22], [6886.523, 1.65453e-21, 3.52749e-22, 2.11649e-22, 2.1791196e-22], [6902.398, 1.65833e-21, 3.5356e-22, 2.12136e-22, 2.1841615e-22], [6918.3097, 1.66214e-21, 3.54373e-22, 2.12624e-22, 2.1892155e-22], [6934.2581, 1.66596e-21, 3.55186e-22, 2.13112e-22, 2.1942797e-22], [6950.2432, 1.66979e-21, 3.56003e-22, 2.13602e-22, 2.1993563e-22], [6966.2651, 1.67363e-21, 3.56821e-22, 2.14093e-22, 2.2044445e-22], [6982.324, 1.67747e-21, 3.57641e-22, 2.14585e-22, 2.2095445e-22], [6998.42, 1.68133e-21, 3.58463e-22, 2.15078e-22, 2.2146561e-22], [7014.553, 1.68519e-21, 3.59287e-22, 2.15572e-22, 2.2197795e-22], [7030.7232, 1.68907e-21, 3.60113e-22, 2.16068e-22, 2.2249147e-22], [7046.9307, 1.69295e-21, 3.60941e-22, 2.16564e-22, 2.2300617e-22], [7063.1755, 1.69684e-21, 3.6177e-22, 2.17062e-22, 2.2352205e-22], [7079.4578, 1.70074e-21, 3.62602e-22, 2.17561e-22, 2.2403912e-22], [7095.7777, 1.70465e-21, 3.63435e-22, 2.18061e-22, 2.2455737e-22], [7112.1351, 1.70857e-21, 3.64271e-22, 2.18562e-22, 2.2507681e-22], [7128.5303, 1.7125e-21, 3.65108e-22, 2.19065e-22, 2.2559745e-22], [7144.9633, 1.71644e-21, 3.65947e-22, 2.19568e-22, 2.2611928e-22], [7161.4341, 1.72038e-21, 3.66789e-22, 2.20073e-22, 2.2664231e-22], [7177.9429, 1.72434e-21, 3.67632e-22, 2.20579e-22, 2.2716655e-22], [7194.4898, 1.7283e-21, 3.68477e-22, 2.21086e-22, 2.2769199e-22], [7211.0748, 1.73227e-21, 3.69324e-22, 2.21594e-22, 2.2821863e-22], [7227.698, 1.73626e-21, 3.70173e-22, 2.22104e-22, 2.2874649e-22], [7244.3596, 1.74025e-21, 3.71024e-22, 2.22614e-22, 2.2927556e-22], [7261.0596, 1.74425e-21, 3.71877e-22, 2.23126e-22, 2.2980584e-22], [7277.798, 1.74826e-21, 3.72732e-22, 2.23639e-22, 2.3033734e-22], [7294.5751, 1.75228e-21, 3.73589e-22, 2.24153e-22, 2.3087007e-22], [7311.3908, 1.75631e-21, 3.74448e-22, 2.24669e-22, 2.3140401e-22], [7328.2453, 1.76034e-21, 3.75308e-22, 2.25185e-22, 2.3193919e-22], [7345.1387, 1.76439e-21, 3.76171e-22, 2.25703e-22, 2.3247559e-22], [7362.071, 1.76845e-21, 3.77036e-22, 2.26222e-22, 2.3301323e-22], [7379.0423, 1.77251e-21, 3.77903e-22, 2.26742e-22, 2.335521e-22], [7396.0528, 1.77659e-21, 3.78772e-22, 2.27263e-22, 2.3409221e-22], [7413.1024, 1.78067e-21, 3.79643e-22, 2.27786e-22, 2.3463356e-22], [7430.1914, 1.78477e-21, 3.80516e-22, 2.28309e-22, 2.3517616e-22], [7447.3197, 1.78887e-21, 3.81391e-22, 2.28834e-22, 2.3572e-22], [7464.4876, 1.79298e-21, 3.82268e-22, 2.29361e-22, 2.362651e-22], [7481.695, 1.79711e-21, 3.83146e-22, 2.29888e-22, 2.3681144e-22], [7498.9421, 1.80124e-21, 3.84027e-22, 2.30416e-22, 2.3735904e-22], [7516.2289, 1.80538e-21, 3.8491e-22, 2.30946e-22, 2.379079e-22], [7533.5556, 1.80953e-21, 3.85796e-22, 2.31477e-22, 2.3845802e-22], [7550.9223, 1.81369e-21, 3.86683e-22, 2.3201e-22, 2.3900941e-22], [7568.329, 1.81786e-21, 3.87572e-22, 2.32543e-22, 2.3956206e-22], [7585.7758, 1.82204e-21, 3.88463e-22, 2.33078e-22, 2.4011598e-22], [7603.2628, 1.82623e-21, 3.89356e-22, 2.33614e-22, 2.4067118e-22], [7620.7901, 1.83043e-21, 3.90252e-22, 2.34151e-22, 2.4122765e-22], [7638.3578, 1.83464e-21, 3.91149e-22, 2.34689e-22, 2.417854e-22], [7655.9661, 1.83886e-21, 3.92048e-22, 2.35229e-22, 2.4234443e-22], [7673.6149, 1.84309e-21, 3.9295e-22, 2.3577e-22, 2.4290475e-22], [7691.3044, 1.84733e-21, 3.93854e-22, 2.36312e-22, 2.4346636e-22], [7709.0347, 1.85158e-21, 3.94759e-22, 2.36856e-22, 2.4402925e-22], [7726.8059, 1.85584e-21, 3.95667e-22, 2.374e-22, 2.4459344e-22], [7744.618, 1.8601e-21, 3.96577e-22, 2.37946e-22, 2.4515893e-22], [7762.4712, 1.86438e-21, 3.97489e-22, 2.38493e-22, 2.4572572e-22], [7780.3655, 1.86867e-21, 3.98403e-22, 2.39042e-22, 2.4629381e-22], [7798.3011, 1.87297e-21, 3.99319e-22, 2.39592e-22, 2.468632e-22], [7816.278, 1.87727e-21, 4.00238e-22, 2.40143e-22, 2.4743391e-22], [7834.2964, 1.88159e-21, 4.01158e-22, 2.40695e-22, 2.4800592e-22], [7852.3563, 1.88592e-21, 4.02081e-22, 2.41249e-22, 2.4857925e-22], [7870.4579, 1.89026e-21, 4.03006e-22, 2.41803e-22, 2.491539e-22], [7888.6012, 1.8946e-21, 4.03933e-22, 2.4236e-22, 2.4972987e-22], [7906.7863, 1.89896e-21, 4.04862e-22, 2.42917e-22, 2.5030717e-22], [7925.0133, 1.90333e-21, 4.05793e-22, 2.43476e-22, 2.5088579e-22], [7943.2823, 1.90771e-21, 4.06726e-22, 2.44036e-22, 2.5146574e-22], [7961.5935, 1.91209e-21, 4.07661e-22, 2.44597e-22, 2.5204703e-22], [7979.9469, 1.91649e-21, 4.08599e-22, 2.45159e-22, 2.5262965e-22], [7998.3426, 1.9209e-21, 4.09539e-22, 2.45723e-22, 2.5321361e-22], [8016.7806, 1.92532e-21, 4.10481e-22, 2.46289e-22, 2.5379891e-22], [8035.2612, 1.92975e-21, 4.11425e-22, 2.46855e-22, 2.5438556e-22], [8053.7844, 1.93419e-21, 4.12371e-22, 2.47423e-22, 2.5497356e-22], [8072.3503, 1.93863e-21, 4.1332e-22, 2.47992e-22, 2.5556291e-22], [8090.959, 1.94309e-21, 4.14271e-22, 2.48562e-22, 2.5615361e-22], [8109.6106, 1.94756e-21, 4.15223e-22, 2.49134e-22, 2.5674568e-22], [8128.3052, 1.95204e-21, 4.16179e-22, 2.49707e-22, 2.573391e-22], [8147.0428, 1.95654e-21, 4.17141e-22, 2.50284e-22, 2.579342e-22], [8165.8237, 1.96103e-21, 4.18095e-22, 2.50857e-22, 2.5853005e-22], [8184.6479, 1.96554e-21, 4.19057e-22, 2.51434e-22, 2.5912757e-22], [8203.5154, 1.97007e-21, 4.20021e-22, 2.52013e-22, 2.5972647e-22], [8222.4265, 1.9746e-21, 4.20987e-22, 2.52592e-22, 2.6032675e-22], [8241.3812, 1.97914e-21, 4.21956e-22, 2.53173e-22, 2.6092841e-22], [8260.3795, 1.98369e-21, 4.22926e-22, 2.53756e-22, 2.6153145e-22], [8279.4216, 1.98826e-21, 4.23899e-22, 2.5434e-22, 2.6213588e-22], [8298.5077, 1.99283e-21, 4.24874e-22, 2.54925e-22, 2.6274169e-22], [8317.6377, 1.99741e-21, 4.25852e-22, 2.55511e-22, 2.633489e-22], [8336.8118, 2.00201e-21, 4.26831e-22, 2.56099e-22, 2.6395751e-22], [8356.0302, 2.00662e-21, 4.27813e-22, 2.56688e-22, 2.6456752e-22], [8375.2928, 2.01123e-21, 4.28798e-22, 2.57279e-22, 2.6517893e-22], [8394.5999, 2.01586e-21, 4.29784e-22, 2.5787e-22, 2.6579174e-22], [8413.9514, 2.0205e-21, 4.30773e-22, 2.58464e-22, 2.6640597e-22], [8433.3476, 2.02514e-21, 4.31764e-22, 2.59058e-22, 2.670216e-22], [8452.7885, 2.0298e-21, 4.32757e-22, 2.59654e-22, 2.6763866e-22], [8472.2741, 2.03447e-21, 4.33753e-22, 2.60252e-22, 2.6825713e-22], [8491.8048, 2.03915e-21, 4.34751e-22, 2.6085e-22, 2.6887702e-22], [8511.3804, 2.04385e-21, 4.35751e-22, 2.6145e-22, 2.6949834e-22], [8531.0011, 2.04855e-21, 4.36753e-22, 2.62052e-22, 2.7012109e-22], [8550.6671, 2.05326e-21, 4.37758e-22, 2.62655e-22, 2.7074527e-22], [8570.3785, 2.05799e-21, 4.38765e-22, 2.63259e-22, 2.7137089e-22], [8590.1352, 2.06272e-21, 4.39775e-22, 2.63865e-22, 2.7199794e-22], [8609.9375, 2.06747e-21, 4.40787e-22, 2.64472e-22, 2.7262644e-22], [8629.7855, 2.07222e-21, 4.41801e-22, 2.6508e-22, 2.7325638e-22], [8649.6792, 2.07699e-21, 4.42817e-22, 2.6569e-22, 2.7388777e-22], [8669.6188, 2.08177e-21, 4.43836e-22, 2.66302e-22, 2.7452062e-22], [8689.6043, 2.08656e-21, 4.44857e-22, 2.66914e-22, 2.7515492e-22], [8709.6359, 2.09136e-21, 4.45881e-22, 2.67529e-22, 2.7579068e-22], [8729.7137, 2.09617e-21, 4.46907e-22, 2.68144e-22, 2.764279e-22], [8749.8378, 2.101e-21, 4.47935e-22, 2.68761e-22, 2.7706658e-22], [8770.0082, 2.10583e-21, 4.48966e-22, 2.69379e-22, 2.7770674e-22], [8790.2252, 2.11068e-21, 4.49999e-22, 2.69999e-22, 2.7834836e-22], [8810.4887, 2.11553e-21, 4.51034e-22, 2.70621e-22, 2.7899147e-22], [8830.799, 2.1204e-21, 4.52072e-22, 2.71243e-22, 2.7963605e-22], [8851.1561, 2.12528e-21, 4.53112e-22, 2.71867e-22, 2.8028211e-22], [8871.5601, 2.13017e-21, 4.54155e-22, 2.72493e-22, 2.8092966e-22], [8892.0112, 2.13507e-21, 4.552e-22, 2.7312e-22, 2.815787e-22], [8912.5094, 2.13998e-21, 4.56247e-22, 2.73748e-22, 2.8222924e-22], [8933.0548, 2.14491e-21, 4.57297e-22, 2.74378e-22, 2.8288127e-22], [8953.6477, 2.14984e-21, 4.58349e-22, 2.7501e-22, 2.835348e-22], [8974.2879, 2.15479e-21, 4.59404e-22, 2.75643e-22, 2.8418983e-22], [8994.9758, 2.15975e-21, 4.60461e-22, 2.76277e-22, 2.8484637e-22], [9015.7114, 2.16472e-21, 4.61521e-22, 2.76913e-22, 2.8550442e-22], [9036.4947, 2.1697e-21, 4.62583e-22, 2.7755e-22, 2.8616398e-22], [9057.326, 2.17469e-21, 4.63647e-22, 2.78188e-22, 2.8682506e-22], [9078.2053, 2.1797e-21, 4.64714e-22, 2.78829e-22, 2.8748767e-22], [9099.1327, 2.18471e-21, 4.65784e-22, 2.7947e-22, 2.8815179e-22], [9120.1084, 2.18974e-21, 4.66856e-22, 2.80113e-22, 2.8881744e-22], [9141.1324, 2.19478e-21, 4.6793e-22, 2.80758e-22, 2.8948463e-22], [9162.2049, 2.19983e-21, 4.69007e-22, 2.81404e-22, 2.9015335e-22], [9183.326, 2.20489e-21, 4.70086e-22, 2.82052e-22, 2.9082361e-22], [9204.4957, 2.20997e-21, 4.71168e-22, 2.82701e-22, 2.9149541e-22], [9225.7143, 2.21505e-21, 4.72252e-22, 2.83351e-22, 2.9216875e-22], [9246.9817, 2.22015e-21, 4.73339e-22, 2.84004e-22, 2.9284365e-22], [9268.2982, 2.22526e-21, 4.74429e-22, 2.84657e-22, 2.9352009e-22], [9289.6639, 2.23038e-21, 4.7552e-22, 2.85312e-22, 2.941981e-22], [9311.0788, 2.23552e-21, 4.76615e-22, 2.85969e-22, 2.9487766e-22], [9332.543, 2.24066e-21, 4.77712e-22, 2.86627e-22, 2.9555879e-22], [9354.0567, 2.24582e-21, 4.78811e-22, 2.87287e-22, 2.9624148e-22], [9375.6201, 2.25099e-21, 4.79913e-22, 2.87948e-22, 2.9692575e-22], [9397.2331, 2.25617e-21, 4.81018e-22, 2.88611e-22, 2.9761158e-22], [9418.896, 2.26136e-21, 4.82125e-22, 2.89275e-22, 2.98299e-22], [9440.6088, 2.26656e-21, 4.83234e-22, 2.89941e-22, 2.98988e-22], [9462.3716, 2.27178e-21, 4.84346e-22, 2.90608e-22, 2.9967859e-22], [9484.1846, 2.27701e-21, 4.85462e-22, 2.91277e-22, 3.0037078e-22], [9506.0479, 2.28225e-21, 4.86578e-22, 2.91947e-22, 3.0106452e-22], [9527.9616, 2.2875e-21, 4.87698e-22, 2.92619e-22, 3.0175988e-22], [9549.9259, 2.29277e-21, 4.88821e-22, 2.93293e-22, 3.0245684e-22], [9571.9407, 2.29804e-21, 4.89946e-22, 2.93968e-22, 3.0315541e-22], [9594.0063, 2.30333e-21, 4.91074e-22, 2.94644e-22, 3.0385558e-22], [9616.1228, 2.30864e-21, 4.92204e-22, 2.95322e-22, 3.0455736e-22], [9638.2902, 2.31395e-21, 4.93337e-22, 2.96002e-22, 3.0526076e-22], [9660.5088, 2.31928e-21, 4.94472e-22, 2.96683e-22, 3.0596578e-22], [9682.7786, 2.32461e-21, 4.9561e-22, 2.97366e-22, 3.0667241e-22], [9705.0997, 2.32996e-21, 4.96751e-22, 2.98051e-22, 3.0738068e-22], [9727.4722, 2.33533e-21, 4.97895e-22, 2.98737e-22, 3.0809057e-22], [9749.8964, 2.3407e-21, 4.99041e-22, 2.99424e-22, 3.088021e-22], [9772.3722, 2.34609e-21, 5.00189e-22, 3.00114e-22, 3.0951526e-22], [9794.8999, 2.35149e-21, 5.01341e-22, 3.00804e-22, 3.1023007e-22], [9817.4794, 2.3569e-21, 5.02495e-22, 3.01497e-22, 3.1094652e-22], [9840.1111, 2.36233e-21, 5.03651e-22, 3.02191e-22, 3.1166462e-22], [9862.7949, 2.36777e-21, 5.04811e-22, 3.02886e-22, 3.1238437e-22], [9885.5309, 2.37322e-21, 5.05973e-22, 3.03584e-22, 3.1310578e-22], [9908.3194, 2.37868e-21, 5.07137e-22, 3.04282e-22, 3.1382885e-22], [9931.1605, 2.38415e-21, 5.08305e-22, 3.04983e-22, 3.1455358e-22], [9954.0542, 2.38964e-21, 5.09475e-22, 3.05685e-22, 3.1527998e-22], [9977.0006, 2.39514e-21, 5.10647e-22, 3.06388e-22, 3.1600805e-22], [10000.0, 2.40066e-21, 5.11823e-22, 3.07094e-22, 3.167378e-22]])
            
    f1 = data[:,0]
    if IFO.upper()=='KAGRA_EARLY':
        S1 = data[:,1]**2
    elif IFO.upper()=='KAGRA_MID':
        S1 = data[:,2]**2
    elif IFO.upper()=='KAGRA_LATE':
        S1 = data[:,3]**2
    elif IFO.upper()=='KAGRA':
        S1 = data[:,4]**2
    else:
        raise ValueError('IFO '+IFO+' not recognized.')
    fstop = [f1[0],f1[-1]]

    # % ---- Interpolate to desired frequencies.  Set noise to infinity outside
    # %      band of frequencies covered by f1.
    # S = inf(size(f)); 
    # k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop


    # % -------------------------------------------------------------------------
    # %                              ETB spectrum
    # % -------------------------------------------------------------------------

def etb(f):
    # % ETB - Design noise curve "B" for ET.
    # %
    # % Arguments are the same as for the parent function.

    # % ---- Extract pre-sampled noise spectrum from data.  First column is
    # %      frequency, second column is effective noise amplitude for a 5-year
    # %      integration period with SNR=10.  (We'll need to remove these
    # %      scalings.)  Data supplied by B. S. Sathyaprakash.
    data = np.array([[1.0, 4.052959e-24], [1.003076, 3.910123e-24], [1.006161, 3.772411e-24], [1.009256, 3.639651e-24], [1.01236, 3.511674e-24], [1.015474, 3.388315e-24], [1.018598, 3.269419e-24], [1.021731, 3.154832e-24], [1.024873, 3.044407e-24], [1.028026, 2.938001e-24], [1.031188, 2.835476e-24], [1.03436, 2.736698e-24], [1.037541, 2.64154e-24], [1.040733, 2.549875e-24], [1.043934, 2.461583e-24], [1.047145, 2.376549e-24], [1.050366, 2.294658e-24], [1.053596, 2.215802e-24], [1.056837, 2.139876e-24], [1.060088, 2.066778e-24], [1.063348, 1.996409e-24], [1.066619, 1.928674e-24], [1.0699, 1.863482e-24], [1.073191, 1.800742e-24], [1.076492, 1.740371e-24], [1.079803, 1.682284e-24], [1.083124, 1.6264e-24], [1.086456, 1.572644e-24], [1.089797, 1.520939e-24], [1.093149, 1.471213e-24], [1.096512, 1.423396e-24], [1.099885, 1.37742e-24], [1.103268, 1.333216e-24], [1.106661, 1.290716e-24], [1.110065, 1.249858e-24], [1.113479, 1.210583e-24], [1.116904, 1.172834e-24], [1.12034, 1.136553e-24], [1.123786, 1.101689e-24], [1.127242, 1.068189e-24], [1.13071, 1.036003e-24], [1.134188, 1.005084e-24], [1.137676, 9.753842e-25], [1.141175, 9.468589e-25], [1.144686, 9.194648e-25], [1.148206, 8.931599e-25], [1.151738, 8.679035e-25], [1.155281, 8.436564e-25], [1.158834, 8.203807e-25], [1.162399, 7.980397e-25], [1.165974, 7.765978e-25], [1.169561, 7.560208e-25], [1.173158, 7.362755e-25], [1.176766, 7.173297e-25], [1.180386, 6.991523e-25], [1.184017, 6.817133e-25], [1.187658, 6.649835e-25], [1.191312, 6.489347e-25], [1.194976, 6.335398e-25], [1.198651, 6.187722e-25], [1.202338, 6.046072e-25], [1.206036, 5.910222e-25], [1.209746, 5.779931e-25], [1.213467, 5.654966e-25], [1.2172, 5.5351e-25], [1.220943, 5.420115e-25], [1.224699, 5.3098e-25], [1.228466, 5.203951e-25], [1.232244, 5.102373e-25], [1.236035, 5.004875e-25], [1.239837, 4.911275e-25], [1.24365, 4.821397e-25], [1.247475, 4.735071e-25], [1.251313, 4.652136e-25], [1.255161, 4.572434e-25], [1.259022, 4.495814e-25], [1.262895, 4.422133e-25], [1.266779, 4.351251e-25], [1.270676, 4.283035e-25], [1.274584, 4.217358e-25], [1.278504, 4.154098e-25], [1.282437, 4.093138e-25], [1.286382, 4.034366e-25], [1.290338, 3.977675e-25], [1.294307, 3.922962e-25], [1.298288, 3.870132e-25], [1.302282, 3.819092e-25], [1.306287, 3.769757e-25], [1.310305, 3.722043e-25], [1.314336, 3.67587e-25], [1.318378, 3.631159e-25], [1.322433, 3.587838e-25], [1.326501, 3.545837e-25], [1.330581, 3.505089e-25], [1.334674, 3.465531e-25], [1.338779, 3.427105e-25], [1.342897, 3.389754e-25], [1.347028, 3.353424e-25], [1.351171, 3.318065e-25], [1.355327, 3.283628e-25], [1.359496, 3.250068e-25], [1.363677, 3.217342e-25], [1.367872, 3.18541e-25], [1.372079, 3.154233e-25], [1.376299, 3.123775e-25], [1.380533, 3.094001e-25], [1.384779, 3.06488e-25], [1.389038, 3.03638e-25], [1.393311, 3.008473e-25], [1.397596, 2.981131e-25], [1.401895, 2.954329e-25], [1.406207, 2.928043e-25], [1.410533, 2.902251e-25], [1.414871, 2.876931e-25], [1.419223, 2.852062e-25], [1.423588, 2.827624e-25], [1.427967, 2.8036e-25], [1.432359, 2.779973e-25], [1.436765, 2.756725e-25], [1.441185, 2.733842e-25], [1.445617, 2.711309e-25], [1.450064, 2.689113e-25], [1.454524, 2.66724e-25], [1.458998, 2.64568e-25], [1.463486, 2.624419e-25], [1.467987, 2.603447e-25], [1.472502, 2.582755e-25], [1.477032, 2.562332e-25], [1.481575, 2.542169e-25], [1.486132, 2.522258e-25], [1.490703, 2.502591e-25], [1.495288, 2.48316e-25], [1.499888, 2.463957e-25], [1.504501, 2.444977e-25], [1.509129, 2.426212e-25], [1.51377, 2.407657e-25], [1.518427, 2.389305e-25], [1.523097, 2.371152e-25], [1.527782, 2.353191e-25], [1.532481, 2.335418e-25], [1.537195, 2.317829e-25], [1.541923, 2.300418e-25], [1.546666, 2.283182e-25], [1.551423, 2.266116e-25], [1.556195, 2.249216e-25], [1.560982, 2.23248e-25], [1.565783, 2.215903e-25], [1.570599, 2.199482e-25], [1.57543, 2.183215e-25], [1.580276, 2.167097e-25], [1.585137, 2.151127e-25], [1.590012, 2.135301e-25], [1.594903, 2.119617e-25], [1.599809, 2.104072e-25], [1.604729, 2.088664e-25], [1.609665, 2.073391e-25], [1.614616, 2.058251e-25], [1.619583, 2.043241e-25], [1.624564, 2.028359e-25], [1.629561, 2.013604e-25], [1.634574, 1.998973e-25], [1.639601, 1.984465e-25], [1.644645, 1.970078e-25], [1.649703, 1.955811e-25], [1.654777, 1.941661e-25], [1.659867, 1.927628e-25], [1.664973, 1.913709e-25], [1.670094, 1.899904e-25], [1.675231, 1.88621e-25], [1.680384, 1.872627e-25], [1.685552, 1.859153e-25], [1.690737, 1.845787e-25], [1.695937, 1.832527e-25], [1.701154, 1.819373e-25], [1.706386, 1.806324e-25], [1.711635, 1.793377e-25], [1.7169, 1.780532e-25], [1.722181, 1.767789e-25], [1.727478, 1.755145e-25], [1.732791, 1.7426e-25], [1.738121, 1.730153e-25], [1.743467, 1.717803e-25], [1.74883, 1.705549e-25], [1.754209, 1.693389e-25], [1.759605, 1.681324e-25], [1.765017, 1.669351e-25], [1.770446, 1.657471e-25], [1.775892, 1.645683e-25], [1.781354, 1.633985e-25], [1.786833, 1.622376e-25], [1.792329, 1.610856e-25], [1.797842, 1.599425e-25], [1.803372, 1.588081e-25], [1.808919, 1.576823e-25], [1.814483, 1.56565e-25], [1.820064, 1.554563e-25], [1.825662, 1.54356e-25], [1.831278, 1.532641e-25], [1.836911, 1.521804e-25], [1.842561, 1.51105e-25], [1.848228, 1.500376e-25], [1.853913, 1.489784e-25], [1.859615, 1.479271e-25], [1.865335, 1.468838e-25], [1.871073, 1.458483e-25], [1.876828, 1.448206e-25], [1.882601, 1.438007e-25], [1.888392, 1.427884e-25], [1.8942, 1.417837e-25], [1.900026, 1.407866e-25], [1.90587, 1.39797e-25], [1.911733, 1.388148e-25], [1.917613, 1.378399e-25], [1.923511, 1.368724e-25], [1.929428, 1.359121e-25], [1.935362, 1.34959e-25], [1.941315, 1.34013e-25], [1.947286, 1.33074e-25], [1.953276, 1.321421e-25], [1.959284, 1.312172e-25], [1.96531, 1.302991e-25], [1.971355, 1.293879e-25], [1.977419, 1.284835e-25], [1.983501, 1.275859e-25], [1.989602, 1.266949e-25], [1.995722, 1.258105e-25], [2.00186, 1.249328e-25], [2.008018, 1.240615e-25], [2.014194, 1.231968e-25], [2.02039, 1.223384e-25], [2.026604, 1.214865e-25], [2.032838, 1.206409e-25], [2.03909, 1.198015e-25], [2.045362, 1.189684e-25], [2.051654, 1.181415e-25], [2.057964, 1.173207e-25], [2.064294, 1.165059e-25], [2.070644, 1.156973e-25], [2.077013, 1.148946e-25], [2.083401, 1.140978e-25], [2.089809, 1.13307e-25], [2.096237, 1.12522e-25], [2.102685, 1.117428e-25], [2.109153, 1.109693e-25], [2.11564, 1.102016e-25], [2.122148, 1.094396e-25], [2.128675, 1.086832e-25], [2.135223, 1.079324e-25], [2.14179, 1.071871e-25], [2.148378, 1.064473e-25], [2.154986, 1.05713e-25], [2.161615, 1.049841e-25], [2.168263, 1.042605e-25], [2.174933, 1.035423e-25], [2.181622, 1.028294e-25], [2.188333, 1.021218e-25], [2.195064, 1.014193e-25], [2.201815, 1.007221e-25], [2.208588, 1.000299e-25], [2.215381, 9.934287e-26], [2.222196, 9.866088e-26], [2.229031, 9.798389e-26], [2.235887, 9.731188e-26], [2.242764, 9.664481e-26], [2.249662, 9.598265e-26], [2.256582, 9.532534e-26], [2.263523, 9.467285e-26], [2.270485, 9.402516e-26], [2.277469, 9.338221e-26], [2.284474, 9.274398e-26], [2.291501, 9.211042e-26], [2.298549, 9.148152e-26], [2.305619, 9.085721e-26], [2.312711, 9.023748e-26], [2.319825, 8.962229e-26], [2.32696, 8.90116e-26], [2.334118, 8.840537e-26], [2.341297, 8.780358e-26], [2.348498, 8.72062e-26], [2.355722, 8.661317e-26], [2.362968, 8.602448e-26], [2.370236, 8.544008e-26], [2.377527, 8.485996e-26], [2.38484, 8.428406e-26], [2.392175, 8.371237e-26], [2.399533, 8.314485e-26], [2.406913, 8.258147e-26], [2.414317, 8.202219e-26], [2.421743, 8.146699e-26], [2.429192, 8.091582e-26], [2.436664, 8.036868e-26], [2.444159, 7.982551e-26], [2.451676, 7.928629e-26], [2.459217, 7.8751e-26], [2.466782, 7.821959e-26], [2.474369, 7.769205e-26], [2.48198, 7.716834e-26], [2.489614, 7.664843e-26], [2.497272, 7.613229e-26], [2.504953, 7.56199e-26], [2.512658, 7.511123e-26], [2.520387, 7.460624e-26], [2.528139, 7.410491e-26], [2.535915, 7.360721e-26], [2.543715, 7.311312e-26], [2.551539, 7.26226e-26], [2.559388, 7.213563e-26], [2.56726, 7.165218e-26], [2.575156, 7.117222e-26], [2.583077, 7.069573e-26], [2.591022, 7.022268e-26], [2.598992, 6.975305e-26], [2.606986, 6.928681e-26], [2.615005, 6.882393e-26], [2.623048, 6.836438e-26], [2.631116, 6.790815e-26], [2.639209, 6.74552e-26], [2.647327, 6.700551e-26], [2.65547, 6.655906e-26], [2.663638, 6.611582e-26], [2.671831, 6.567576e-26], [2.680049, 6.523887e-26], [2.688292, 6.480512e-26], [2.696561, 6.437448e-26], [2.704855, 6.394693e-26], [2.713175, 6.352244e-26], [2.721521, 6.3101e-26], [2.729891, 6.268257e-26], [2.738288, 6.226715e-26], [2.746711, 6.185469e-26], [2.755159, 6.144519e-26], [2.763634, 6.103861e-26], [2.772134, 6.063494e-26], [2.780661, 6.023416e-26], [2.789214, 5.983623e-26], [2.797793, 5.944114e-26], [2.806399, 5.904887e-26], [2.815031, 5.86594e-26], [2.823689, 5.82727e-26], [2.832375, 5.788875e-26], [2.841087, 5.750753e-26], [2.849826, 5.712903e-26], [2.858591, 5.675322e-26], [2.867384, 5.638007e-26], [2.876203, 5.600957e-26], [2.88505, 5.564171e-26], [2.893924, 5.527645e-26], [2.902826, 5.491378e-26], [2.911754, 5.455368e-26], [2.92071, 5.419613e-26], [2.929694, 5.38411e-26], [2.938705, 5.348859e-26], [2.947744, 5.313857e-26], [2.956811, 5.279102e-26], [2.965906, 5.244592e-26], [2.975029, 5.210326e-26], [2.98418, 5.176301e-26], [2.993358, 5.142516e-26], [3.002566, 5.108969e-26], [3.011801, 5.075657e-26], [3.021065, 5.04258e-26], [3.030357, 5.009736e-26], [3.039678, 4.977122e-26], [3.049028, 4.944737e-26], [3.058406, 4.912579e-26], [3.067813, 4.880646e-26], [3.07725, 4.848937e-26], [3.086715, 4.81745e-26], [3.096209, 4.786183e-26], [3.105733, 4.755135e-26], [3.115285, 4.724304e-26], [3.124868, 4.693688e-26], [3.134479, 4.663285e-26], [3.14412, 4.633095e-26], [3.153791, 4.603114e-26], [3.163492, 4.573343e-26], [3.173222, 4.543779e-26], [3.182983, 4.51442e-26], [3.192773, 4.485265e-26], [3.202594, 4.456312e-26], [3.212444, 4.427561e-26], [3.222325, 4.399008e-26], [3.232237, 4.370654e-26], [3.242179, 4.342496e-26], [3.252151, 4.314533e-26], [3.262154, 4.286762e-26], [3.272188, 4.259184e-26], [3.282253, 4.231797e-26], [3.292349, 4.204598e-26], [3.302476, 4.177586e-26], [3.312634, 4.150761e-26], [3.322823, 4.12412e-26], [3.333043, 4.097663e-26], [3.343295, 4.071388e-26], [3.353579, 4.045292e-26], [3.363894, 4.019376e-26], [3.374241, 3.993638e-26], [3.384619, 3.968076e-26], [3.39503, 3.94269e-26], [3.405473, 3.917477e-26], [3.415947, 3.892436e-26], [3.426454, 3.867567e-26], [3.436994, 3.842867e-26], [3.447565, 3.818336e-26], [3.458169, 3.793972e-26], [3.468806, 3.769774e-26], [3.479476, 3.745741e-26], [3.490178, 3.721871e-26], [3.500914, 3.698163e-26], [3.511682, 3.674617e-26], [3.522483, 3.65123e-26], [3.533318, 3.628002e-26], [3.544186, 3.604932e-26], [3.555087, 3.582017e-26], [3.566022, 3.559258e-26], [3.576991, 3.536653e-26], [3.587993, 3.5142e-26], [3.599029, 3.491899e-26], [3.6101, 3.469748e-26], [3.621204, 3.447747e-26], [3.632342, 3.425894e-26], [3.643515, 3.404188e-26], [3.654721, 3.382628e-26], [3.665963, 3.361212e-26], [3.677239, 3.339941e-26], [3.68855, 3.318812e-26], [3.699895, 3.297826e-26], [3.711275, 3.276979e-26], [3.722691, 3.256273e-26], [3.734141, 3.235704e-26], [3.745627, 3.215274e-26], [3.757148, 3.194979e-26], [3.768704, 3.174821e-26], [3.780296, 3.154796e-26], [3.791924, 3.134906e-26], [3.803587, 3.115147e-26], [3.815287, 3.09552e-26], [3.827022, 3.076024e-26], [3.838793, 3.056657e-26], [3.850601, 3.037419e-26], [3.862445, 3.018309e-26], [3.874325, 2.999325e-26], [3.886242, 2.980467e-26], [3.898195, 2.961734e-26], [3.910186, 2.943125e-26], [3.922213, 2.924639e-26], [3.934277, 2.906275e-26], [3.946378, 2.888032e-26], [3.958517, 2.869909e-26], [3.970693, 2.851906e-26], [3.982906, 2.834022e-26], [3.995157, 2.816255e-26], [4.007446, 2.798605e-26], [4.019772, 2.781072e-26], [4.032136, 2.763653e-26], [4.044538, 2.746349e-26], [4.056979, 2.729158e-26], [4.069458, 2.71208e-26], [4.081975, 2.695113e-26], [4.09453, 2.678258e-26], [4.107124, 2.661513e-26], [4.119757, 2.644878e-26], [4.132429, 2.628351e-26], [4.14514, 2.611932e-26], [4.15789, 2.59562e-26], [4.170679, 2.579415e-26], [4.183507, 2.563315e-26], [4.196375, 2.547319e-26], [4.209282, 2.531428e-26], [4.22223, 2.515641e-26], [4.235217, 2.499955e-26], [4.248244, 2.484372e-26], [4.26131, 2.46889e-26], [4.274418, 2.453508e-26], [4.287565, 2.438226e-26], [4.300753, 2.423042e-26], [4.313982, 2.407957e-26], [4.327251, 2.39297e-26], [4.340561, 2.378079e-26], [4.353912, 2.363285e-26], [4.367304, 2.348586e-26], [4.380737, 2.333982e-26], [4.394211, 2.319472e-26], [4.407728, 2.305056e-26], [4.421285, 2.290732e-26], [4.434884, 2.2765e-26], [4.448525, 2.262361e-26], [4.462208, 2.248312e-26], [4.475934, 2.234353e-26], [4.489701, 2.220484e-26], [4.50351, 2.206704e-26], [4.517363, 2.193012e-26], [4.531257, 2.179408e-26], [4.545195, 2.165891e-26], [4.559175, 2.152461e-26], [4.573199, 2.139116e-26], [4.587265, 2.125857e-26], [4.601375, 2.112683e-26], [4.615528, 2.099593e-26], [4.629725, 2.086586e-26], [4.643965, 2.073663e-26], [4.658249, 2.060821e-26], [4.672578, 2.048062e-26], [4.68695, 2.035384e-26], [4.701366, 2.022787e-26], [4.715827, 2.01027e-26], [4.730332, 1.997832e-26], [4.744882, 1.985474e-26], [4.759476, 1.973194e-26], [4.774116, 1.960992e-26], [4.7888, 1.948868e-26], [4.80353, 1.93682e-26], [4.818305, 1.924849e-26], [4.833126, 1.912954e-26], [4.847992, 1.901134e-26], [4.862903, 1.889389e-26], [4.877861, 1.877718e-26], [4.892864, 1.866121e-26], [4.907914, 1.854598e-26], [4.92301, 1.843147e-26], [4.938153, 1.831769e-26], [4.953342, 1.820463e-26], [4.968578, 1.809228e-26], [4.98386, 1.798064e-26], [4.99919, 1.78697e-26], [5.014567, 1.775946e-26], [5.029991, 1.764992e-26], [5.045462, 1.754107e-26], [5.060981, 1.74329e-26], [5.076548, 1.732541e-26], [5.092163, 1.721861e-26], [5.107826, 1.711247e-26], [5.123537, 1.7007e-26], [5.139296, 1.690219e-26], [5.155104, 1.679804e-26], [5.17096, 1.669455e-26], [5.186865, 1.65917e-26], [5.202819, 1.64895e-26], [5.218822, 1.638794e-26], [5.234875, 1.628702e-26], [5.250977, 1.618673e-26], [5.267128, 1.608707e-26], [5.283329, 1.598803e-26], [5.29958, 1.588962e-26], [5.31588, 1.579182e-26], [5.332231, 1.569463e-26], [5.348632, 1.559805e-26], [5.365084, 1.550207e-26], [5.381586, 1.54067e-26], [5.398139, 1.531192e-26], [5.414743, 1.521773e-26], [5.431398, 1.512413e-26], [5.448104, 1.503112e-26], [5.464862, 1.493869e-26], [5.481671, 1.484683e-26], [5.498532, 1.475554e-26], [5.515445, 1.466483e-26], [5.532409, 1.457468e-26], [5.549426, 1.448509e-26], [5.566496, 1.439607e-26], [5.583617, 1.430759e-26], [5.600792, 1.421967e-26], [5.618019, 1.413229e-26], [5.635299, 1.404546e-26], [5.652633, 1.395917e-26], [5.670019, 1.387342e-26], [5.687459, 1.37882e-26], [5.704953, 1.370351e-26], [5.722501, 1.361934e-26], [5.740102, 1.35357e-26], [5.757758, 1.345258e-26], [5.775468, 1.336998e-26], [5.793233, 1.328788e-26], [5.811052, 1.32063e-26], [5.828926, 1.312523e-26], [5.846855, 1.304466e-26], [5.864839, 1.296458e-26], [5.882878, 1.288501e-26], [5.900973, 1.280593e-26], [5.919124, 1.272734e-26], [5.93733, 1.264923e-26], [5.955592, 1.257162e-26], [5.973911, 1.249448e-26], [5.992286, 1.241782e-26], [6.010717, 1.234163e-26], [6.029205, 1.226592e-26], [6.047751, 1.219068e-26], [6.066352, 1.21159e-26], [6.085012, 1.204158e-26], [6.103728, 1.196773e-26], [6.122503, 1.189433e-26], [6.141335, 1.182139e-26], [6.160224, 1.17489e-26], [6.179172, 1.167685e-26], [6.198179, 1.160526e-26], [6.217243, 1.15341e-26], [6.236367, 1.146339e-26], [6.255549, 1.139311e-26], [6.27479, 1.132327e-26], [6.29409, 1.125385e-26], [6.31345, 1.118487e-26], [6.332869, 1.111632e-26], [6.352348, 1.104818e-26], [6.371887, 1.098047e-26], [6.391486, 1.091318e-26], [6.411146, 1.08463e-26], [6.430866, 1.077984e-26], [6.450646, 1.071378e-26], [6.470487, 1.064814e-26], [6.490389, 1.05829e-26], [6.510353, 1.051806e-26], [6.530378, 1.045362e-26], [6.550464, 1.038958e-26], [6.570613, 1.032593e-26], [6.590823, 1.026268e-26], [6.611095, 1.019982e-26], [6.63143, 1.013734e-26], [6.651827, 1.007525e-26], [6.672288, 1.001355e-26], [6.692811, 9.95222e-27], [6.713397, 9.891273e-27], [6.734046, 9.8307e-27], [6.754759, 9.770502e-27], [6.775536, 9.710675e-27], [6.796376, 9.651216e-27], [6.817281, 9.592123e-27], [6.83825, 9.533395e-27], [6.859284, 9.475029e-27], [6.880382, 9.417022e-27], [6.901545, 9.359373e-27], [6.922773, 9.302079e-27], [6.944066, 9.245137e-27], [6.965426, 9.188548e-27], [6.98685, 9.132306e-27], [7.008341, 9.07641e-27], [7.029897, 9.020859e-27], [7.05152, 8.96565e-27], [7.07321, 8.910781e-27], [7.094966, 8.85625e-27], [7.116789, 8.802054e-27], [7.138679, 8.748192e-27], [7.160637, 8.694661e-27], [7.182662, 8.641461e-27], [7.204755, 8.588588e-27], [7.226916, 8.53604e-27], [7.249145, 8.483815e-27], [7.271442, 8.431912e-27], [7.293808, 8.380328e-27], [7.316243, 8.329062e-27], [7.338746, 8.278111e-27], [7.361319, 8.227474e-27], [7.383962, 8.177148e-27], [7.406674, 8.127132e-27], [7.429456, 8.077424e-27], [7.452307, 8.028022e-27], [7.47523, 7.978923e-27], [7.498222, 7.930126e-27], [7.521286, 7.88163e-27], [7.54442, 7.833432e-27], [7.567626, 7.78553e-27], [7.590903, 7.737923e-27], [7.614251, 7.690609e-27], [7.637672, 7.643586e-27], [7.661164, 7.596852e-27], [7.684729, 7.550406e-27], [7.708366, 7.504245e-27], [7.732076, 7.458368e-27], [7.755859, 7.412773e-27], [7.779714, 7.367458e-27], [7.803644, 7.322422e-27], [7.827647, 7.277663e-27], [7.851723, 7.233179e-27], [7.875874, 7.188969e-27], [7.900099, 7.14503e-27], [7.924399, 7.101362e-27], [7.948773, 7.057962e-27], [7.973222, 7.014829e-27], [7.997747, 6.971962e-27], [8.022347, 6.929357e-27], [8.047022, 6.887015e-27], [8.071774, 6.844933e-27], [8.096602, 6.80311e-27], [8.121506, 6.761543e-27], [8.146486, 6.720233e-27], [8.171544, 6.679176e-27], [8.196678, 6.638372e-27], [8.22189, 6.597818e-27], [8.247179, 6.557514e-27], [8.272546, 6.517458e-27], [8.297992, 6.477647e-27], [8.323515, 6.438082e-27], [8.349117, 6.39876e-27], [8.374798, 6.359679e-27], [8.400557, 6.320839e-27], [8.426396, 6.282237e-27], [8.452315, 6.243873e-27], [8.478313, 6.205744e-27], [8.504391, 6.16785e-27], [8.530549, 6.130189e-27], [8.556788, 6.092759e-27], [8.583107, 6.05556e-27], [8.609508, 6.018589e-27], [8.63599, 5.981845e-27], [8.662553, 5.945327e-27], [8.689197, 5.909034e-27], [8.715924, 5.872964e-27], [8.742733, 5.837115e-27], [8.769624, 5.801487e-27], [8.796598, 5.766078e-27], [8.823656, 5.730886e-27], [8.850796, 5.695911e-27], [8.87802, 5.661151e-27], [8.905327, 5.626605e-27], [8.932719, 5.592271e-27], [8.960195, 5.558148e-27], [8.987755, 5.524234e-27], [9.0154, 5.49053e-27], [9.04313, 5.457032e-27], [9.070945, 5.42374e-27], [9.098846, 5.390653e-27], [9.126833, 5.35777e-27], [9.154906, 5.325089e-27], [9.183065, 5.292608e-27], [9.211311, 5.260327e-27], [9.239644, 5.228245e-27], [9.268063, 5.19636e-27], [9.296571, 5.164671e-27], [9.325166, 5.133177e-27], [9.353848, 5.101877e-27], [9.38262, 5.070768e-27], [9.411479, 5.039852e-27], [9.440428, 5.009125e-27], [9.469465, 4.978587e-27], [9.498592, 4.948238e-27], [9.527808, 4.918074e-27], [9.557114, 4.888097e-27], [9.58651, 4.858303e-27], [9.615997, 4.828693e-27], [9.645575, 4.799265e-27], [9.675243, 4.770018e-27], [9.705003, 4.740951e-27], [9.734854, 4.712062e-27], [9.764797, 4.683351e-27], [9.794832, 4.654817e-27], [9.82496, 4.626459e-27], [9.85518, 4.598274e-27], [9.885493, 4.570264e-27], [9.915899, 4.542425e-27], [9.946399, 4.514758e-27], [9.976993, 4.487261e-27], [10.007681, 4.459933e-27], [10.038463, 4.432773e-27], [10.06934, 4.405781e-27], [10.100312, 4.378954e-27], [10.131379, 4.352293e-27], [10.162542, 4.325795e-27], [10.1938, 4.299461e-27], [10.225155, 4.273288e-27], [10.256606, 4.247277e-27], [10.288154, 4.221426e-27], [10.319799, 4.195734e-27], [10.351541, 4.1702e-27], [10.383381, 4.144823e-27], [10.415319, 4.119602e-27], [10.447355, 4.094537e-27], [10.479489, 4.069626e-27], [10.511723, 4.044868e-27], [10.544055, 4.020263e-27], [10.576487, 3.995809e-27], [10.609019, 3.971506e-27], [10.641651, 3.947352e-27], [10.674383, 3.923347e-27], [10.707216, 3.89949e-27], [10.74015, 3.875779e-27], [10.773185, 3.852215e-27], [10.806322, 3.828796e-27], [10.83956, 3.805521e-27], [10.872901, 3.78239e-27], [10.906345, 3.759401e-27], [10.939891, 3.736553e-27], [10.973541, 3.713846e-27], [11.007294, 3.691279e-27], [11.041151, 3.668851e-27], [11.075112, 3.646562e-27], [11.109177, 3.624409e-27], [11.143347, 3.602394e-27], [11.177623, 3.580513e-27], [11.212004, 3.558768e-27], [11.24649, 3.537156e-27], [11.281083, 3.515678e-27], [11.315782, 3.494332e-27], [11.350587, 3.473118e-27], [11.3855, 3.452034e-27], [11.42052, 3.431081e-27], [11.455648, 3.410257e-27], [11.490884, 3.38956e-27], [11.526229, 3.368992e-27], [11.561682, 3.34855e-27], [11.597244, 3.328235e-27], [11.632915, 3.308045e-27], [11.668696, 3.287979e-27], [11.704588, 3.268037e-27], [11.740589, 3.248218e-27], [11.776702, 3.228521e-27], [11.812925, 3.208945e-27], [11.84926, 3.189491e-27], [11.885707, 3.170156e-27], [11.922265, 3.150941e-27], [11.958937, 3.131844e-27], [11.995721, 3.112866e-27], [12.032618, 3.094004e-27], [12.069628, 3.075258e-27], [12.106753, 3.056629e-27], [12.143991, 3.038114e-27], [12.181345, 3.019714e-27], [12.218813, 3.001427e-27], [12.256396, 2.983253e-27], [12.294095, 2.965191e-27], [12.33191, 2.947241e-27], [12.369841, 2.929402e-27], [12.407889, 2.911673e-27], [12.446054, 2.894053e-27], [12.484336, 2.876542e-27], [12.522736, 2.85914e-27], [12.561254, 2.841844e-27], [12.599891, 2.824656e-27], [12.638646, 2.807574e-27], [12.677521, 2.790598e-27], [12.716515, 2.773726e-27], [12.75563, 2.756959e-27], [12.794864, 2.740295e-27], [12.834219, 2.723734e-27], [12.873696, 2.707276e-27], [12.913293, 2.690919e-27], [12.953013, 2.674664e-27], [12.992854, 2.658509e-27], [13.032818, 2.642454e-27], [13.072906, 2.626498e-27], [13.113116, 2.610641e-27], [13.15345, 2.594882e-27], [13.193908, 2.579221e-27], [13.234491, 2.563656e-27], [13.275198, 2.548188e-27], [13.316031, 2.532815e-27], [13.356989, 2.517538e-27], [13.398073, 2.502355e-27], [13.439284, 2.487267e-27], [13.480621, 2.472271e-27], [13.522086, 2.457369e-27], [13.563678, 2.442558e-27], [13.605398, 2.42784e-27], [13.647246, 2.413213e-27], [13.689223, 2.398676e-27], [13.731329, 2.384229e-27], [13.773565, 2.369872e-27], [13.81593, 2.355604e-27], [13.858426, 2.341424e-27], [13.901052, 2.327332e-27], [13.94381, 2.313328e-27], [13.986699, 2.29941e-27], [14.02972, 2.285579e-27], [14.072874, 2.271833e-27], [14.11616, 2.258172e-27], [14.159579, 2.244597e-27], [14.203132, 2.231105e-27], [14.246819, 2.217698e-27], [14.29064, 2.204373e-27], [14.334596, 2.191131e-27], [14.378687, 2.177972e-27], [14.422914, 2.164894e-27], [14.467277, 2.151897e-27], [14.511776, 2.138981e-27], [14.556412, 2.126145e-27], [14.601186, 2.113389e-27], [14.646097, 2.100712e-27], [14.691146, 2.088114e-27], [14.736334, 2.075594e-27], [14.781661, 2.063152e-27], [14.827127, 2.050788e-27], [14.872734, 2.0385e-27], [14.91848, 2.026288e-27], [14.964367, 2.014153e-27], [15.010395, 2.002093e-27], [15.056565, 1.990108e-27], [15.102877, 1.978198e-27], [15.149331, 1.966361e-27], [15.195929, 1.954599e-27], [15.242669, 1.942909e-27], [15.289553, 1.931293e-27], [15.336582, 1.919748e-27], [15.383755, 1.908276e-27], [15.431073, 1.896875e-27], [15.478537, 1.885545e-27], [15.526147, 1.874286e-27], [15.573903, 1.863096e-27], [15.621806, 1.851977e-27], [15.669857, 1.840927e-27], [15.718055, 1.829946e-27], [15.766401, 1.819033e-27], [15.814897, 1.808188e-27], [15.863541, 1.797411e-27], [15.912335, 1.786701e-27], [15.961279, 1.776058e-27], [16.010374, 1.765482e-27], [16.059619, 1.754971e-27], [16.109016, 1.744527e-27], [16.158565, 1.734147e-27], [16.208267, 1.723832e-27], [16.258121, 1.713582e-27], [16.308129, 1.703396e-27], [16.35829, 1.693273e-27], [16.408606, 1.683214e-27], [16.459077, 1.673217e-27], [16.509702, 1.663284e-27], [16.560484, 1.653412e-27], [16.611422, 1.643602e-27], [16.662516, 1.633853e-27], [16.713768, 1.624166e-27], [16.765177, 1.614539e-27], [16.816744, 1.604972e-27], [16.86847, 1.595466e-27], [16.920355, 1.586019e-27], [16.9724, 1.576631e-27], [17.024604, 1.567302e-27], [17.07697, 1.558031e-27], [17.129496, 1.548819e-27], [17.182184, 1.539664e-27], [17.235034, 1.530567e-27], [17.288046, 1.521527e-27], [17.341222, 1.512544e-27], [17.394561, 1.503617e-27], [17.448064, 1.494746e-27], [17.501732, 1.485931e-27], [17.555565, 1.477171e-27], [17.609563, 1.468467e-27], [17.663728, 1.459817e-27], [17.718059, 1.451221e-27], [17.772557, 1.44268e-27], [17.827223, 1.434192e-27], [17.882057, 1.425758e-27], [17.93706, 1.417377e-27], [17.992231, 1.409048e-27], [18.047573, 1.400772e-27], [18.103085, 1.392549e-27], [18.158767, 1.384377e-27], [18.214621, 1.376256e-27], [18.270647, 1.368187e-27], [18.326845, 1.360169e-27], [18.383215, 1.352201e-27], [18.439759, 1.344284e-27], [18.496478, 1.336416e-27], [18.55337, 1.328598e-27], [18.610438, 1.32083e-27], [18.667681, 1.313111e-27], [18.7251, 1.30544e-27], [18.782696, 1.297818e-27], [18.840468, 1.290245e-27], [18.898419, 1.282719e-27], [18.956548, 1.275241e-27], [19.014856, 1.26781e-27], [19.073343, 1.260426e-27], [19.132009, 1.253089e-27], [19.190857, 1.245799e-27], [19.249885, 1.238555e-27], [19.309095, 1.231356e-27], [19.368487, 1.224204e-27], [19.428062, 1.217097e-27], [19.48782, 1.210034e-27], [19.547762, 1.203017e-27], [19.607888, 1.196045e-27], [19.668199, 1.189116e-27], [19.728695, 1.182232e-27], [19.789378, 1.175392e-27], [19.850247, 1.168595e-27], [19.911304, 1.161841e-27], [19.972548, 1.15513e-27], [20.033981, 1.148462e-27], [20.095603, 1.141837e-27], [20.157414, 1.135254e-27], [20.219415, 1.128713e-27], [20.281607, 1.122213e-27], [20.343991, 1.115755e-27], [20.406566, 1.109339e-27], [20.469333, 1.102963e-27], [20.532294, 1.096628e-27], [20.595449, 1.090333e-27], [20.658797, 1.084079e-27], [20.722341, 1.077865e-27], [20.78608, 1.071691e-27], [20.850015, 1.065556e-27], [20.914147, 1.05946e-27], [20.978475, 1.053404e-27], [21.043002, 1.047386e-27], [21.107728, 1.041407e-27], [21.172652, 1.035466e-27], [21.237776, 1.029564e-27], [21.3031, 1.023699e-27], [21.368626, 1.017872e-27], [21.434353, 1.012083e-27], [21.500282, 1.006331e-27], [21.566413, 1.000616e-27], [21.632749, 9.949374e-28], [21.699288, 9.892957e-28], [21.766032, 9.836903e-28], [21.832981, 9.781211e-28], [21.900136, 9.725878e-28], [21.967498, 9.670902e-28], [22.035067, 9.616281e-28], [22.102844, 9.562014e-28], [22.170829, 9.508097e-28], [22.239023, 9.454529e-28], [22.307427, 9.401308e-28], [22.376042, 9.348431e-28], [22.444867, 9.295897e-28], [22.513904, 9.243703e-28], [22.583154, 9.191848e-28], [22.652617, 9.14033e-28], [22.722293, 9.089146e-28], [22.792183, 9.038295e-28], [22.862289, 8.987774e-28], [22.93261, 8.937581e-28], [23.003147, 8.887716e-28], [23.073902, 8.838175e-28], [23.144874, 8.788957e-28], [23.216064, 8.74006e-28], [23.287474, 8.691481e-28], [23.359102, 8.64322e-28], [23.430952, 8.595273e-28], [23.503022, 8.547641e-28], [23.575314, 8.500319e-28], [23.647828, 8.453307e-28], [23.720566, 8.406603e-28], [23.793527, 8.360205e-28], [23.866712, 8.314111e-28], [23.940123, 8.268319e-28], [24.013759, 8.222828e-28], [24.087622, 8.177635e-28], [24.161712, 8.13274e-28], [24.23603, 8.088139e-28], [24.310577, 8.043832e-28], [24.385353, 7.999817e-28], [24.460359, 7.956092e-28], [24.535595, 7.912654e-28], [24.611063, 7.869504e-28], [24.686763, 7.826638e-28], [24.762696, 7.784056e-28], [24.838863, 7.741755e-28], [24.915264, 7.699734e-28], [24.9919, 7.65799e-28], [25.068771, 7.616524e-28], [25.145879, 7.575332e-28], [25.223224, 7.534414e-28], [25.300807, 7.493767e-28], [25.378629, 7.45339e-28], [25.45669, 7.413282e-28], [25.534991, 7.37344e-28], [25.613533, 7.333864e-28], [25.692317, 7.294551e-28], [25.771343, 7.255501e-28], [25.850612, 7.216711e-28], [25.930124, 7.17818e-28], [26.009882, 7.139906e-28], [26.089885, 7.101889e-28], [26.170133, 7.064125e-28], [26.250629, 7.026615e-28], [26.331372, 6.989356e-28], [26.412364, 6.952348e-28], [26.493604, 6.915587e-28], [26.575095, 6.879074e-28], [26.656836, 6.842806e-28], [26.738829, 6.806782e-28], [26.821074, 6.771001e-28], [26.903571, 6.735461e-28], [26.986323, 6.700161e-28], [27.069329, 6.665099e-28], [27.152591, 6.630274e-28], [27.236108, 6.595685e-28], [27.319882, 6.56133e-28], [27.403915, 6.527207e-28], [27.488205, 6.493316e-28], [27.572755, 6.459655e-28], [27.657565, 6.426222e-28], [27.742636, 6.393017e-28], [27.827968, 6.360037e-28], [27.913563, 6.327282e-28], [27.999421, 6.294751e-28], [28.085543, 6.262441e-28], [28.17193, 6.230352e-28], [28.258583, 6.198482e-28], [28.345503, 6.16683e-28], [28.432689, 6.135395e-28], [28.520144, 6.104176e-28], [28.607868, 6.07317e-28], [28.695862, 6.042377e-28], [28.784126, 6.011796e-28], [28.872662, 5.981425e-28], [28.96147, 5.951263e-28], [29.050552, 5.92131e-28], [29.139907, 5.891562e-28], [29.229537, 5.86202e-28], [29.319443, 5.832682e-28], [29.409626, 5.803547e-28], [29.500085, 5.774614e-28], [29.590824, 5.745881e-28], [29.681841, 5.717348e-28], [29.773138, 5.689013e-28], [29.864716, 5.660875e-28], [29.956575, 5.632932e-28], [30.048718, 5.605184e-28], [30.141143, 5.577629e-28], [30.233853, 5.550267e-28], [30.326848, 5.523096e-28], [30.420129, 5.496115e-28], [30.513697, 5.469323e-28], [30.607553, 5.442718e-28], [30.701697, 5.4163e-28], [30.796131, 5.390067e-28], [30.890856, 5.364019e-28], [30.985872, 5.338154e-28], [31.08118, 5.312471e-28], [31.176781, 5.28697e-28], [31.272677, 5.261648e-28], [31.368867, 5.236505e-28], [31.465353, 5.21154e-28], [31.562136, 5.186751e-28], [31.659217, 5.162139e-28], [31.756596, 5.1377e-28], [31.854275, 5.113436e-28], [31.952254, 5.089344e-28], [32.050535, 5.065423e-28], [32.149118, 5.041673e-28], [32.248004, 5.018092e-28], [32.347194, 4.99468e-28], [32.446689, 4.971435e-28], [32.546491, 4.948356e-28], [32.646599, 4.925442e-28], [32.747015, 4.902693e-28], [32.847741, 4.880108e-28], [32.948776, 4.857684e-28], [33.050121, 4.835422e-28], [33.151779, 4.81332e-28], [33.253749, 4.791378e-28], [33.356033, 4.769594e-28], [33.458631, 4.747967e-28], [33.561545, 4.726497e-28], [33.664776, 4.705182e-28], [33.768324, 4.684022e-28], [33.87219, 4.663015e-28], [33.976376, 4.642162e-28], [34.080883, 4.62146e-28], [34.185711, 4.600908e-28], [34.290861, 4.580507e-28], [34.396335, 4.560254e-28], [34.502133, 4.54015e-28], [34.608257, 4.520192e-28], [34.714707, 4.500381e-28], [34.821485, 4.480715e-28], [34.92859, 4.461194e-28], [35.036026, 4.441816e-28], [35.143792, 4.42258e-28], [35.251889, 4.403486e-28], [35.360319, 4.384534e-28], [35.469082, 4.365721e-28], [35.57818, 4.347047e-28], [35.687613, 4.328511e-28], [35.797383, 4.310113e-28], [35.907491, 4.291851e-28], [36.017937, 4.273724e-28], [36.128724, 4.255733e-28], [36.23985, 4.237875e-28], [36.351319, 4.220151e-28], [36.46313, 4.202559e-28], [36.575286, 4.185098e-28], [36.687786, 4.167768e-28], [36.800633, 4.150567e-28], [36.913826, 4.133496e-28], [37.027368, 4.116552e-28], [37.141259, 4.099736e-28], [37.2555, 4.083046e-28], [37.370093, 4.066482e-28], [37.485038, 4.050043e-28], [37.600336, 4.033728e-28], [37.71599, 4.017536e-28], [37.831999, 4.001467e-28], [37.948365, 3.985519e-28], [38.065088, 3.969692e-28], [38.182171, 3.953986e-28], [38.299614, 3.938398e-28], [38.417418, 3.92293e-28], [38.535585, 3.907579e-28], [38.654115, 3.892345e-28], [38.773009, 3.877228e-28], [38.89227, 3.862226e-28], [39.011897, 3.847338e-28], [39.131892, 3.832565e-28], [39.252256, 3.817905e-28], [39.37299, 3.803358e-28], [39.494096, 3.788923e-28], [39.615574, 3.774598e-28], [39.737426, 3.760384e-28], [39.859653, 3.74628e-28], [39.982255, 3.732284e-28], [40.105235, 3.718397e-28], [40.228593, 3.704617e-28], [40.352331, 3.690943e-28], [40.476449, 3.677376e-28], [40.600948, 3.663914e-28], [40.725831, 3.650557e-28], [40.851098, 3.637303e-28], [40.97675, 3.624153e-28], [41.102789, 3.611106e-28], [41.229215, 3.59816e-28], [41.35603, 3.585315e-28], [41.483236, 3.572571e-28], [41.610832, 3.559927e-28], [41.738821, 3.547381e-28], [41.867204, 3.534934e-28], [41.995982, 3.522585e-28], [42.125155, 3.510333e-28], [42.254726, 3.498178e-28], [42.384696, 3.486118e-28], [42.515065, 3.474153e-28], [42.645835, 3.462283e-28], [42.777008, 3.450507e-28], [42.908584, 3.438823e-28], [43.040565, 3.427233e-28], [43.172951, 3.415734e-28], [43.305745, 3.404326e-28], [43.438947, 3.393009e-28], [43.572559, 3.381782e-28], [43.706582, 3.370645e-28], [43.841018, 3.359596e-28], [43.975866, 3.348635e-28], [44.11113, 3.337762e-28], [44.24681, 3.326976e-28], [44.382906, 3.316275e-28], [44.519422, 3.305661e-28], [44.656357, 3.295132e-28], [44.793714, 3.284687e-28], [44.931493, 3.274325e-28], [45.069696, 3.264048e-28], [45.208324, 3.253852e-28], [45.347378, 3.243739e-28], [45.48686, 3.233708e-28], [45.626772, 3.223757e-28], [45.767113, 3.213886e-28], [45.907886, 3.204096e-28], [46.049092, 3.194384e-28], [46.190733, 3.184751e-28], [46.332809, 3.175196e-28], [46.475322, 3.165719e-28], [46.618273, 3.156318e-28], [46.761665, 3.146994e-28], [46.905497, 3.137745e-28], [47.049772, 3.128572e-28], [47.19449, 3.119473e-28], [47.339654, 3.110449e-28], [47.485264, 3.101498e-28], [47.631321, 3.09262e-28], [47.777829, 3.083815e-28], [47.924786, 3.075081e-28], [48.072196, 3.066419e-28], [48.22006, 3.057828e-28], [48.368378, 3.049308e-28], [48.517152, 3.040857e-28], [48.666384, 3.032475e-28], [48.816075, 3.024162e-28], [48.966226, 3.015918e-28], [49.116839, 3.007741e-28], [49.267916, 2.999632e-28], [49.419457, 2.991589e-28], [49.571464, 2.983612e-28], [49.723939, 2.975702e-28], [49.876882, 2.967856e-28], [50.030297, 2.960076e-28], [50.184183, 2.952359e-28], [50.338542, 2.944707e-28], [50.493376, 2.937117e-28], [50.648687, 2.929591e-28], [50.804475, 2.922126e-28], [50.960742, 2.914724e-28], [51.11749, 2.907383e-28], [51.27472, 2.900103e-28], [51.432434, 2.892883e-28], [51.590633, 2.885723e-28], [51.749319, 2.878623e-28], [51.908492, 2.871582e-28], [52.068155, 2.864599e-28], [52.228309, 2.857674e-28], [52.388956, 2.850807e-28], [52.550097, 2.843998e-28], [52.711734, 2.837245e-28], [52.873868, 2.830548e-28], [53.0365, 2.823907e-28], [53.199633, 2.817322e-28], [53.363268, 2.810792e-28], [53.527405, 2.804316e-28], [53.692048, 2.797894e-28], [53.857197, 2.791526e-28], [54.022854, 2.785212e-28], [54.189021, 2.77895e-28], [54.355699, 2.772741e-28], [54.522889, 2.766583e-28], [54.690594, 2.760478e-28], [54.858814, 2.754423e-28], [55.027552, 2.748419e-28], [55.196809, 2.742466e-28], [55.366587, 2.736563e-28], [55.536886, 2.730709e-28], [55.70771, 2.724904e-28], [55.879059, 2.719148e-28], [56.050935, 2.713441e-28], [56.22334, 2.707782e-28], [56.396275, 2.70217e-28], [56.569742, 2.696605e-28], [56.743742, 2.691087e-28], [56.918278, 2.685616e-28], [57.09335, 2.680191e-28], [57.268961, 2.674811e-28], [57.445113, 2.669477e-28], [57.621806, 2.664188e-28], [57.799042, 2.658943e-28], [57.976824, 2.653743e-28], [58.155152, 2.648586e-28], [58.334029, 2.643473e-28], [58.513456, 2.638403e-28], [58.693435, 2.633376e-28], [58.873968, 2.628392e-28], [59.055056, 2.623449e-28], [59.236701, 2.618548e-28], [59.418905, 2.613689e-28], [59.601669, 2.608871e-28], [59.784995, 2.604093e-28], [59.968885, 2.599356e-28], [60.153341, 2.594659e-28], [60.338364, 2.590002e-28], [60.523956, 2.585384e-28], [60.710119, 2.580805e-28], [60.896855, 2.576265e-28], [61.084165, 2.571763e-28], [61.272051, 2.567299e-28], [61.460516, 2.562874e-28], [61.649559, 2.558485e-28], [61.839185, 2.554134e-28], [62.029393, 2.54982e-28], [62.220187, 2.545542e-28], [62.411567, 2.541301e-28], [62.603536, 2.537095e-28], [62.796096, 2.532925e-28], [62.989248, 2.528791e-28], [63.182994, 2.524691e-28], [63.377336, 2.520627e-28], [63.572275, 2.516597e-28], [63.767815, 2.5126e-28], [63.963955, 2.508638e-28], [64.160699, 2.50471e-28], [64.358049, 2.500815e-28], [64.556005, 2.496952e-28], [64.75457, 2.493123e-28], [64.953746, 2.489326e-28], [65.153534, 2.485562e-28], [65.353937, 2.481829e-28], [65.554957, 2.478128e-28], [65.756594, 2.474458e-28], [65.958852, 2.47082e-28], [66.161732, 2.467212e-28], [66.365237, 2.463636e-28], [66.569367, 2.460089e-28], [66.774125, 2.456573e-28], [66.979512, 2.453086e-28], [67.185532, 2.449629e-28], [67.392185, 2.446202e-28], [67.599474, 2.442804e-28], [67.8074, 2.439434e-28], [68.015966, 2.436093e-28], [68.225173, 2.432781e-28], [68.435024, 2.429497e-28], [68.645521, 2.426241e-28], [68.856665, 2.423012e-28], [69.068458, 2.419811e-28], [69.280903, 2.416637e-28], [69.494001, 2.41349e-28], [69.707755, 2.41037e-28], [69.922166, 2.407277e-28], [70.137236, 2.40421e-28], [70.352969, 2.401169e-28], [70.569364, 2.398154e-28], [70.786426, 2.395165e-28], [71.004155, 2.392201e-28], [71.222553, 2.389262e-28], [71.441624, 2.386349e-28], [71.661368, 2.38346e-28], [71.881788, 2.380596e-28], [72.102886, 2.377756e-28], [72.324665, 2.374941e-28], [72.547125, 2.37215e-28], [72.77027, 2.369382e-28], [72.994101, 2.366638e-28], [73.21862, 2.363918e-28], [73.44383, 2.361221e-28], [73.669733, 2.358547e-28], [73.896331, 2.355896e-28], [74.123625, 2.353267e-28], [74.351619, 2.350661e-28], [74.580314, 2.348078e-28], [74.809713, 2.345516e-28], [75.039817, 2.342977e-28], [75.270628, 2.340459e-28], [75.50215, 2.337963e-28], [75.734384, 2.335488e-28], [75.967332, 2.333035e-28], [76.200997, 2.330603e-28], [76.43538, 2.328191e-28], [76.670485, 2.325801e-28], [76.906312, 2.323431e-28], [77.142865, 2.321081e-28], [77.380146, 2.318752e-28], [77.618156, 2.316442e-28], [77.856898, 2.314153e-28], [78.096375, 2.311883e-28], [78.336589, 2.309633e-28], [78.577541, 2.307403e-28], [78.819234, 2.305192e-28], [79.061671, 2.303e-28], [79.304853, 2.300827e-28], [79.548784, 2.298673e-28], [79.793465, 2.296537e-28], [80.038898, 2.29442e-28], [80.285086, 2.292322e-28], [80.532032, 2.290242e-28], [80.779737, 2.28818e-28], [81.028204, 2.286136e-28], [81.277435, 2.28411e-28], [81.527433, 2.282101e-28], [81.7782, 2.28011e-28], [82.029738, 2.278137e-28], [82.28205, 2.276181e-28], [82.535138, 2.274242e-28], [82.789004, 2.27232e-28], [83.043651, 2.270415e-28], [83.299082, 2.268527e-28], [83.555298, 2.266656e-28], [83.812302, 2.264801e-28], [84.070097, 2.262963e-28], [84.328684, 2.26114e-28], [84.588067, 2.259335e-28], [84.848248, 2.257545e-28], [85.10923, 2.255771e-28], [85.371013, 2.254013e-28], [85.633603, 2.25227e-28], [85.896999, 2.250544e-28], [86.161206, 2.248832e-28], [86.426226, 2.247137e-28], [86.692061, 2.245456e-28], [86.958713, 2.243791e-28], [87.226186, 2.24214e-28], [87.494481, 2.240505e-28], [87.763602, 2.238884e-28], [88.03355, 2.237278e-28], [88.304329, 2.235687e-28], [88.57594, 2.234111e-28], [88.848387, 2.232548e-28], [89.121672, 2.231001e-28], [89.395798, 2.229467e-28], [89.670767, 2.227947e-28], [89.946581, 2.226442e-28], [90.223244, 2.224951e-28], [90.500758, 2.223473e-28], [90.779126, 2.222009e-28], [91.058349, 2.220559e-28], [91.338432, 2.219122e-28], [91.619376, 2.217699e-28], [91.901184, 2.216289e-28], [92.183859, 2.214893e-28], [92.467404, 2.21351e-28], [92.75182, 2.21214e-28], [93.037112, 2.210783e-28], [93.32328, 2.209439e-28], [93.61033, 2.208108e-28], [93.898262, 2.20679e-28], [94.187079, 2.205484e-28], [94.476786, 2.204192e-28], [94.767383, 2.202911e-28], [95.058874, 2.201644e-28], [95.351261, 2.200388e-28], [95.644548, 2.199145e-28], [95.938737, 2.197915e-28], [96.233831, 2.196696e-28], [96.529833, 2.19549e-28], [96.826745, 2.194296e-28], [97.12457, 2.193114e-28], [97.423312, 2.191943e-28], [97.722972, 2.190785e-28], [98.023554, 2.189638e-28], [98.32506, 2.188503e-28], [98.627494, 2.18738e-28], [98.930858, 2.186268e-28], [99.235156, 2.185168e-28], [99.540389, 2.18408e-28], [99.846561, 2.183002e-28], [100.15367, 2.181936e-28], [100.46173, 2.180882e-28], [100.77074, 2.179838e-28], [101.0807, 2.178806e-28], [101.39161, 2.177785e-28], [101.70347, 2.176775e-28], [102.0163, 2.175776e-28], [102.33008, 2.174788e-28], [102.64484, 2.173811e-28], [102.96056, 2.172844e-28], [103.27725, 2.171889e-28], [103.59492, 2.170944e-28], [103.91356, 2.17001e-28], [104.23318, 2.169086e-28], [104.55379, 2.168173e-28], [104.87538, 2.167271e-28], [105.19796, 2.166379e-28], [105.52154, 2.165497e-28], [105.84611, 2.164626e-28], [106.17168, 2.163765e-28], [106.49824, 2.162915e-28], [106.82582, 2.162074e-28], [107.1544, 2.161244e-28], [107.48399, 2.160424e-28], [107.8146, 2.159615e-28], [108.14622, 2.158815e-28], [108.47886, 2.158025e-28], [108.81253, 2.157246e-28], [109.14722, 2.156476e-28], [109.48294, 2.155716e-28], [109.81969, 2.154966e-28], [110.15748, 2.154226e-28], [110.49631, 2.153496e-28], [110.83618, 2.152775e-28], [111.1771, 2.152065e-28], [111.51906, 2.151363e-28], [111.86208, 2.150672e-28], [112.20615, 2.14999e-28], [112.55128, 2.149318e-28], [112.89747, 2.148656e-28], [113.24473, 2.148002e-28], [113.59306, 2.147359e-28], [113.94245, 2.146725e-28], [114.29292, 2.1461e-28], [114.64447, 2.145485e-28], [114.9971, 2.144879e-28], [115.35082, 2.144282e-28], [115.70562, 2.143695e-28], [116.06151, 2.143117e-28], [116.4185, 2.142549e-28], [116.77659, 2.141989e-28], [117.13578, 2.141439e-28], [117.49607, 2.140898e-28], [117.85747, 2.140366e-28], [118.21998, 2.139844e-28], [118.58361, 2.13933e-28], [118.94836, 2.138826e-28], [119.31423, 2.13833e-28], [119.68122, 2.137844e-28], [120.04934, 2.137367e-28], [120.4186, 2.136898e-28], [120.78899, 2.136439e-28], [121.16052, 2.135989e-28], [121.53319, 2.135547e-28], [121.90701, 2.135115e-28], [122.28198, 2.134692e-28], [122.6581, 2.134277e-28], [123.03538, 2.133871e-28], [123.41382, 2.133474e-28], [123.79342, 2.133086e-28], [124.17419, 2.132707e-28], [124.55613, 2.132337e-28], [124.93925, 2.131975e-28], [125.32355, 2.131623e-28], [125.70902, 2.131279e-28], [126.09569, 2.130944e-28], [126.48354, 2.130617e-28], [126.87259, 2.130299e-28], [127.26283, 2.129991e-28], [127.65427, 2.12969e-28], [128.04692, 2.129399e-28], [128.44077, 2.129116e-28], [128.83584, 2.128842e-28], [129.23212, 2.128576e-28], [129.62962, 2.12832e-28], [130.02834, 2.128071e-28], [130.42829, 2.127832e-28], [130.82947, 2.127601e-28], [131.23188, 2.127379e-28], [131.63553, 2.127165e-28], [132.04042, 2.12696e-28], [132.44656, 2.126764e-28], [132.85395, 2.126576e-28], [133.26259, 2.126397e-28], [133.67248, 2.126227e-28], [134.08364, 2.126065e-28], [134.49606, 2.125911e-28], [134.90975, 2.125767e-28], [135.32472, 2.125631e-28], [135.74096, 2.125503e-28], [136.15848, 2.125384e-28], [136.57728, 2.125274e-28], [136.99737, 2.125172e-28], [137.41876, 2.125079e-28], [137.84144, 2.124994e-28], [138.26542, 2.124918e-28], [138.6907, 2.12485e-28], [139.1173, 2.124791e-28], [139.5452, 2.124741e-28], [139.97442, 2.124699e-28], [140.40496, 2.124666e-28], [140.83683, 2.124641e-28], [141.27002, 2.124625e-28], [141.70455, 2.124618e-28], [142.14041, 2.124619e-28], [142.57762, 2.124629e-28], [143.01617, 2.124647e-28], [143.45606, 2.124674e-28], [143.89731, 2.124709e-28], [144.33992, 2.124754e-28], [144.78389, 2.124806e-28], [145.22922, 2.124868e-28], [145.67593, 2.124938e-28], [146.12401, 2.125016e-28], [146.57346, 2.125103e-28], [147.0243, 2.125199e-28], [147.47653, 2.125304e-28], [147.93015, 2.125417e-28], [148.38516, 2.125539e-28], [148.84157, 2.12567e-28], [149.29939, 2.125809e-28], [149.75861, 2.125957e-28], [150.21925, 2.126113e-28], [150.6813, 2.126279e-28], [151.14477, 2.126453e-28], [151.60967, 2.126635e-28], [152.076, 2.126827e-28], [152.54377, 2.127027e-28], [153.01297, 2.127236e-28], [153.48362, 2.127454e-28], [153.95571, 2.127681e-28], [154.42926, 2.127916e-28], [154.90426, 2.12816e-28], [155.38072, 2.128413e-28], [155.85865, 2.128675e-28], [156.33805, 2.128946e-28], [156.81892, 2.129225e-28], [157.30128, 2.129513e-28], [157.78511, 2.129811e-28], [158.27044, 2.130117e-28], [158.75725, 2.130432e-28], [159.24557, 2.130756e-28], [159.73539, 2.131089e-28], [160.22671, 2.131431e-28], [160.71954, 2.131782e-28], [161.21389, 2.132142e-28], [161.70976, 2.132511e-28], [162.20716, 2.132889e-28], [162.70609, 2.133276e-28], [163.20655, 2.133672e-28], [163.70855, 2.134077e-28], [164.21209, 2.134491e-28], [164.71719, 2.134915e-28], [165.22383, 2.135347e-28], [165.73204, 2.135789e-28], [166.24181, 2.136239e-28], [166.75314, 2.136699e-28], [167.26605, 2.137169e-28], [167.78054, 2.137647e-28], [168.29661, 2.138135e-28], [168.81426, 2.138632e-28], [169.33351, 2.139138e-28], [169.85436, 2.139653e-28], [170.37681, 2.140178e-28], [170.90086, 2.140712e-28], [171.42653, 2.141256e-28], [171.95381, 2.141809e-28], [172.48272, 2.142371e-28], [173.01325, 2.142943e-28], [173.54541, 2.143524e-28], [174.07921, 2.144115e-28], [174.61466, 2.144715e-28], [175.15175, 2.145325e-28], [175.69049, 2.145944e-28], [176.23089, 2.146573e-28], [176.77295, 2.147212e-28], [177.31668, 2.14786e-28], [177.86208, 2.148518e-28], [178.40916, 2.149185e-28], [178.95792, 2.149862e-28], [179.50837, 2.150549e-28], [180.06051, 2.151246e-28], [180.61435, 2.151952e-28], [181.16989, 2.152668e-28], [181.72715, 2.153394e-28], [182.28611, 2.15413e-28], [182.8468, 2.154876e-28], [183.40921, 2.155632e-28], [183.97335, 2.156397e-28], [184.53923, 2.157173e-28], [185.10684, 2.157958e-28], [185.67621, 2.158754e-28], [186.24732, 2.15956e-28], [186.82019, 2.160375e-28], [187.39482, 2.161201e-28], [187.97122, 2.162037e-28], [188.54939, 2.162883e-28], [189.12935, 2.163739e-28], [189.71108, 2.164605e-28], [190.2946, 2.165482e-28], [190.87992, 2.166369e-28], [191.46704, 2.167266e-28], [192.05597, 2.168174e-28], [192.64671, 2.169092e-28], [193.23926, 2.17002e-28], [193.83364, 2.170959e-28], [194.42984, 2.171908e-28], [195.02788, 2.172868e-28], [195.62776, 2.173838e-28], [196.22948, 2.174819e-28], [196.83305, 2.175811e-28], [197.43849, 2.176813e-28], [198.04578, 2.177825e-28], [198.65494, 2.178849e-28], [199.26597, 2.179883e-28], [199.87889, 2.180928e-28], [200.49369, 2.181983e-28], [201.11038, 2.18305e-28], [201.72896, 2.184127e-28], [202.34945, 2.185215e-28], [202.97185, 2.186314e-28], [203.59616, 2.187424e-28], [204.2224, 2.188545e-28], [204.85056, 2.189677e-28], [205.48065, 2.19082e-28], [206.11268, 2.191974e-28], [206.74665, 2.19314e-28], [207.38257, 2.194316e-28], [208.02045, 2.195504e-28], [208.66029, 2.196703e-28], [209.3021, 2.197913e-28], [209.94589, 2.199134e-28], [210.59165, 2.200367e-28], [211.2394, 2.201611e-28], [211.88914, 2.202867e-28], [212.54088, 2.204134e-28], [213.19463, 2.205413e-28], [213.85039, 2.206703e-28], [214.50816, 2.208004e-28], [215.16796, 2.209318e-28], [215.82978, 2.210642e-28], [216.49364, 2.211979e-28], [217.15955, 2.213327e-28], [217.8275, 2.214687e-28], [218.4975, 2.216059e-28], [219.16957, 2.217443e-28], [219.84371, 2.218838e-28], [220.51991, 2.220246e-28], [221.1982, 2.221665e-28], [221.87858, 2.223097e-28], [222.56104, 2.22454e-28], [223.24561, 2.225996e-28], [223.93228, 2.227463e-28], [224.62107, 2.228943e-28], [225.31197, 2.230435e-28], [226.005, 2.231939e-28], [226.70016, 2.233456e-28], [227.39745, 2.234985e-28], [228.09689, 2.236526e-28], [228.79849, 2.238079e-28], [229.50224, 2.239645e-28], [230.20816, 2.241224e-28], [230.91624, 2.242815e-28], [231.62651, 2.244419e-28], [232.33896, 2.246035e-28], [233.0536, 2.247664e-28], [233.77044, 2.249305e-28], [234.48949, 2.25096e-28], [235.21074, 2.252627e-28], [235.93422, 2.254307e-28], [236.65992, 2.255999e-28], [237.38785, 2.257705e-28], [238.11802, 2.259424e-28], [238.85044, 2.261155e-28], [239.58511, 2.2629e-28], [240.32204, 2.264658e-28], [241.06123, 2.266429e-28], [241.8027, 2.268213e-28], [242.54646, 2.27001e-28], [243.29249, 2.271821e-28], [244.04083, 2.273645e-28], [244.79146, 2.275482e-28], [245.54441, 2.277333e-28], [246.29967, 2.279197e-28], [247.05725, 2.281075e-28], [247.81716, 2.282966e-28], [248.57941, 2.284871e-28], [249.34401, 2.28679e-28], [250.11095, 2.288722e-28], [250.88026, 2.290668e-28], [251.65193, 2.292627e-28], [252.42598, 2.294601e-28], [253.2024, 2.296588e-28], [253.98122, 2.298589e-28], [254.76243, 2.300605e-28], [255.54604, 2.302634e-28], [256.33206, 2.304677e-28], [257.1205, 2.306735e-28], [257.91137, 2.308806e-28], [258.70467, 2.310892e-28], [259.50041, 2.312992e-28], [260.29859, 2.315107e-28], [261.09924, 2.317235e-28], [261.90234, 2.319378e-28], [262.70792, 2.321536e-28], [263.51597, 2.323708e-28], [264.3265, 2.325895e-28], [265.13954, 2.328096e-28], [265.95507, 2.330312e-28], [266.77311, 2.332543e-28], [267.59366, 2.334788e-28], [268.41674, 2.337048e-28], [269.24236, 2.339323e-28], [270.07051, 2.341613e-28], [270.90121, 2.343918e-28], [271.73446, 2.346238e-28], [272.57028, 2.348573e-28], [273.40866, 2.350923e-28], [274.24963, 2.353288e-28], [275.09318, 2.355668e-28], [275.93933, 2.358064e-28], [276.78808, 2.360475e-28], [277.63944, 2.362902e-28], [278.49342, 2.365343e-28], [279.35003, 2.367801e-28], [280.20927, 2.370273e-28], [281.07115, 2.372762e-28], [281.93569, 2.375266e-28], [282.80288, 2.377785e-28], [283.67274, 2.380321e-28], [284.54528, 2.382872e-28], [285.4205, 2.385439e-28], [286.29841, 2.388022e-28], [287.17903, 2.390621e-28], [288.06235, 2.393236e-28], [288.94839, 2.395866e-28], [289.83715, 2.398514e-28], [290.72865, 2.401177e-28], [291.62289, 2.403856e-28], [292.51988, 2.406552e-28], [293.41963, 2.409264e-28], [294.32215, 2.411992e-28], [295.22744, 2.414737e-28], [296.13552, 2.417498e-28], [297.04639, 2.420276e-28], [297.96006, 2.42307e-28], [298.87654, 2.425882e-28], [299.79584, 2.428709e-28], [300.71797, 2.431554e-28], [301.64294, 2.434415e-28], [302.57075, 2.437294e-28], [303.50142, 2.440189e-28], [304.43494, 2.443101e-28], [305.37134, 2.44603e-28], [306.31062, 2.448977e-28], [307.25279, 2.45194e-28], [308.19785, 2.454921e-28], [309.14583, 2.457919e-28], [310.09672, 2.460934e-28], [311.05053, 2.463967e-28], [312.00728, 2.467017e-28], [312.96697, 2.470085e-28], [313.92961, 2.47317e-28], [314.89521, 2.476273e-28], [315.86378, 2.479393e-28], [316.83534, 2.482531e-28], [317.80988, 2.485688e-28], [318.78741, 2.488861e-28], [319.76796, 2.492053e-28], [320.75152, 2.495263e-28], [321.73811, 2.498491e-28], [322.72773, 2.501737e-28], [323.72039, 2.505001e-28], [324.71611, 2.508283e-28], [325.71489, 2.511583e-28], [326.71674, 2.514902e-28], [327.72168, 2.518239e-28], [328.7297, 2.521595e-28], [329.74083, 2.524969e-28], [330.75506, 2.528361e-28], [331.77242, 2.531773e-28], [332.7929, 2.535203e-28], [333.81653, 2.538651e-28], [334.8433, 2.542119e-28], [335.87323, 2.545605e-28], [336.90633, 2.54911e-28], [337.94261, 2.552634e-28], [338.98207, 2.556178e-28], [340.02473, 2.55974e-28], [341.0706, 2.563322e-28], [342.11968, 2.566922e-28], [343.17199, 2.570542e-28], [344.22754, 2.574182e-28], [345.28634, 2.577841e-28], [346.34839, 2.581519e-28], [347.41371, 2.585216e-28], [348.4823, 2.588934e-28], [349.55418, 2.592671e-28], [350.62936, 2.596428e-28], [351.70785, 2.600204e-28], [352.78965, 2.604001e-28], [353.87478, 2.607817e-28], [354.96325, 2.611653e-28], [356.05507, 2.615509e-28], [357.15024, 2.619386e-28], [358.24879, 2.623282e-28], [359.35071, 2.627199e-28], [360.45602, 2.631136e-28], [361.56473, 2.635093e-28], [362.67685, 2.639071e-28], [363.7924, 2.643069e-28], [364.91137, 2.647088e-28], [366.03379, 2.651127e-28], [367.15965, 2.655187e-28], [368.28898, 2.659268e-28], [369.42179, 2.663369e-28], [370.55808, 2.667492e-28], [371.69786, 2.671635e-28], [372.84115, 2.675799e-28], [373.98796, 2.679985e-28], [375.13829, 2.684191e-28], [376.29216, 2.688419e-28], [377.44959, 2.692667e-28], [378.61057, 2.696938e-28], [379.77512, 2.701229e-28], [380.94325, 2.705542e-28], [382.11498, 2.709877e-28], [383.29031, 2.714233e-28], [384.46926, 2.71861e-28], [385.65183, 2.72301e-28], [386.83804, 2.727431e-28], [388.0279, 2.731874e-28], [389.22142, 2.736339e-28], [390.41861, 2.740826e-28], [391.61948, 2.745335e-28], [392.82405, 2.749866e-28], [394.03232, 2.754419e-28], [395.24431, 2.758995e-28], [396.46002, 2.763592e-28], [397.67948, 2.768212e-28], [398.90268, 2.772855e-28], [400.12965, 2.77752e-28], [401.36039, 2.782208e-28], [402.59492, 2.786918e-28], [403.83324, 2.791651e-28], [405.07538, 2.796407e-28], [406.32133, 2.801185e-28], [407.57112, 2.805987e-28], [408.82475, 2.810811e-28], [410.08224, 2.815659e-28], [411.34359, 2.82053e-28], [412.60883, 2.825424e-28], [413.87795, 2.830341e-28], [415.15098, 2.835281e-28], [416.42793, 2.840245e-28], [417.7088, 2.845232e-28], [418.99361, 2.850243e-28], [420.28238, 2.855278e-28], [421.57511, 2.860336e-28], [422.87181, 2.865418e-28], [424.17251, 2.870524e-28], [425.4772, 2.875653e-28], [426.78591, 2.880807e-28], [428.09864, 2.885984e-28], [429.41541, 2.891186e-28], [430.73623, 2.896412e-28], [432.06111, 2.901662e-28], [433.39007, 2.906937e-28], [434.72312, 2.912235e-28], [436.06027, 2.917559e-28], [437.40153, 2.922907e-28], [438.74691, 2.928279e-28], [440.09643, 2.933676e-28], [441.45011, 2.939098e-28], [442.80795, 2.944545e-28], [444.16996, 2.950016e-28], [445.53616, 2.955513e-28], [446.90657, 2.961034e-28], [448.28119, 2.966581e-28], [449.66004, 2.972152e-28], [451.04313, 2.97775e-28], [452.43047, 2.983372e-28], [453.82209, 2.98902e-28], [455.21798, 2.994693e-28], [456.61816, 3.000392e-28], [458.02266, 3.006116e-28], [459.43147, 3.011866e-28], [460.84462, 3.017642e-28], [462.26211, 3.023444e-28], [463.68396, 3.029271e-28], [465.11019, 3.035125e-28], [466.5408, 3.041005e-28], [467.97581, 3.04691e-28], [469.41524, 3.052842e-28], [470.8591, 3.058801e-28], [472.30739, 3.064785e-28], [473.76014, 3.070797e-28], [475.21736, 3.076834e-28], [476.67906, 3.082898e-28], [478.14526, 3.088989e-28], [479.61597, 3.095107e-28], [481.0912, 3.101251e-28], [482.57096, 3.107423e-28], [484.05528, 3.113621e-28], [485.54417, 3.119846e-28], [487.03763, 3.126099e-28], [488.53569, 3.132379e-28], [490.03836, 3.138686e-28], [491.54565, 3.14502e-28], [493.05757, 3.151382e-28], [494.57415, 3.157771e-28], [496.09539, 3.164188e-28], [497.62131, 3.170632e-28], [499.15192, 3.177105e-28], [500.68724, 3.183605e-28], [502.22728, 3.190133e-28], [503.77206, 3.196689e-28], [505.32159, 3.203273e-28], [506.87589, 3.209885e-28], [508.43497, 3.216526e-28], [509.99884, 3.223194e-28], [511.56753, 3.229892e-28], [513.14103, 3.236617e-28], [514.71938, 3.243371e-28], [516.30259, 3.250154e-28], [517.89066, 3.256965e-28], [519.48362, 3.263806e-28], [521.08148, 3.270675e-28], [522.68425, 3.277573e-28], [524.29195, 3.2845e-28], [525.9046, 3.291456e-28], [527.52221, 3.298441e-28], [529.14479, 3.305456e-28], [530.77236, 3.3125e-28], [532.40495, 3.319573e-28], [534.04255, 3.326676e-28], [535.68519, 3.333808e-28], [537.33288, 3.34097e-28], [538.98564, 3.348162e-28], [540.64348, 3.355384e-28], [542.30642, 3.362636e-28], [543.97448, 3.369917e-28], [545.64767, 3.377229e-28], [547.326, 3.384571e-28], [549.0095, 3.391943e-28], [550.69818, 3.399346e-28], [552.39205, 3.406779e-28], [554.09112, 3.414243e-28], [555.79543, 3.421737e-28], [557.50498, 3.429261e-28], [559.21978, 3.436817e-28], [560.93986, 3.444403e-28], [562.66524, 3.452021e-28], [564.39591, 3.459669e-28], [566.13192, 3.467349e-28], [567.87326, 3.475059e-28], [569.61995, 3.482801e-28], [571.37202, 3.490575e-28], [573.12948, 3.498379e-28], [574.89235, 3.506216e-28], [576.66064, 3.514084e-28], [578.43436, 3.521983e-28], [580.21354, 3.529915e-28], [581.9982, 3.537878e-28], [583.78834, 3.545874e-28], [585.58399, 3.553901e-28], [587.38517, 3.561961e-28], [589.19188, 3.570052e-28], [591.00415, 3.578177e-28], [592.82199, 3.586333e-28], [594.64543, 3.594522e-28], [596.47448, 3.602744e-28], [598.30915, 3.610998e-28], [600.14946, 3.619286e-28], [601.99543, 3.627606e-28], [603.84709, 3.635959e-28], [605.70443, 3.644345e-28], [607.5675, 3.652764e-28], [609.43629, 3.661216e-28], [611.31083, 3.669702e-28], [613.19113, 3.678221e-28], [615.07722, 3.686774e-28], [616.96911, 3.69536e-28], [618.86682, 3.70398e-28], [620.77037, 3.712634e-28], [622.67977, 3.721322e-28], [624.59504, 3.730044e-28], [626.51621, 3.738799e-28], [628.44328, 3.747589e-28], [630.37629, 3.756414e-28], [632.31523, 3.765272e-28], [634.26014, 3.774165e-28], [636.21104, 3.783092e-28], [638.16793, 3.792055e-28], [640.13085, 3.801051e-28], [642.0998, 3.810083e-28], [644.07481, 3.81915e-28], [646.05589, 3.828251e-28], [648.04307, 3.837388e-28], [650.03635, 3.84656e-28], [652.03577, 3.855767e-28], [654.04134, 3.86501e-28], [656.05308, 3.874288e-28], [658.07101, 3.883602e-28], [660.09514, 3.892951e-28], [662.1255, 3.902336e-28], [664.1621, 3.911758e-28], [666.20497, 3.921215e-28], [668.25412, 3.930708e-28], [670.30958, 3.940237e-28], [672.37136, 3.949803e-28], [674.43947, 3.959405e-28], [676.51395, 3.969043e-28], [678.59481, 3.978719e-28], [680.68208, 3.98843e-28], [682.77576, 3.998179e-28], [684.87588, 4.007964e-28], [686.98246, 4.017787e-28], [689.09552, 4.027646e-28], [691.21508, 4.037543e-28], [693.34116, 4.047477e-28], [695.47378, 4.057449e-28], [697.61296, 4.067457e-28], [699.75872, 4.077504e-28], [701.91107, 4.087588e-28], [704.07005, 4.09771e-28], [706.23567, 4.10787e-28], [708.40795, 4.118068e-28], [710.58691, 4.128304e-28], [712.77258, 4.138579e-28], [714.96497, 4.148891e-28], [717.1641, 4.159242e-28], [719.36999, 4.169632e-28], [721.58267, 4.18006e-28], [723.80216, 4.190527e-28], [726.02847, 4.201033e-28], [728.26163, 4.211578e-28], [730.50166, 4.222162e-28], [732.74858, 4.232785e-28], [735.00241, 4.243447e-28], [737.26317, 4.254149e-28], [739.53089, 4.26489e-28], [741.80558, 4.275671e-28], [744.08727, 4.286492e-28], [746.37598, 4.297352e-28], [748.67172, 4.308253e-28], [750.97453, 4.319193e-28], [753.28442, 4.330174e-28], [755.60142, 4.341195e-28], [757.92554, 4.352256e-28], [760.25681, 4.363358e-28], [762.59525, 4.3745e-28], [764.94089, 4.385683e-28], [767.29374, 4.396907e-28], [769.65382, 4.408172e-28], [772.02117, 4.419478e-28], [774.3958, 4.430825e-28], [776.77773, 4.442213e-28], [779.16699, 4.453643e-28], [781.56359, 4.465114e-28], [783.96757, 4.476627e-28], [786.37895, 4.488181e-28], [788.79774, 4.499778e-28], [791.22397, 4.511416e-28], [793.65766, 4.523097e-28], [796.09884, 4.53482e-28], [798.54752, 4.546585e-28], [801.00374, 4.558392e-28], [803.46752, 4.570242e-28], [805.93887, 4.582135e-28], [808.41782, 4.59407e-28], [810.9044, 4.606049e-28], [813.39863, 4.61807e-28], [815.90052, 4.630134e-28], [818.41012, 4.642242e-28], [820.92743, 4.654394e-28], [823.45249, 4.666588e-28], [825.98531, 4.678827e-28], [828.52592, 4.691109e-28], [831.07435, 4.703434e-28], [833.63062, 4.715804e-28], [836.19475, 4.728218e-28], [838.76676, 4.740677e-28], [841.34669, 4.753179e-28], [843.93456, 4.765726e-28], [846.53038, 4.778318e-28], [849.13419, 4.790954e-28], [851.746, 4.803636e-28], [854.36585, 4.816362e-28], [856.99376, 4.829133e-28], [859.62975, 4.84195e-28], [862.27385, 4.854812e-28], [864.92608, 4.867719e-28], [867.58647, 4.880672e-28], [870.25504, 4.893671e-28], [872.93182, 4.906715e-28], [875.61684, 4.919806e-28], [878.31011, 4.932942e-28], [881.01167, 4.946125e-28], [883.72154, 4.959355e-28], [886.43974, 4.97263e-28], [889.1663, 4.985953e-28], [891.90125, 4.999322e-28], [894.64461, 5.012738e-28], [897.39641, 5.026201e-28], [900.15667, 5.039711e-28], [902.92543, 5.053269e-28], [905.7027, 5.066874e-28], [908.48851, 5.080526e-28], [911.28289, 5.094226e-28], [914.08587, 5.107974e-28], [916.89747, 5.12177e-28], [919.71771, 5.135614e-28], [922.54664, 5.149506e-28], [925.38426, 5.163446e-28], [928.23061, 5.177435e-28], [931.08571, 5.191473e-28], [933.9496, 5.205559e-28], [936.8223, 5.219695e-28], [939.70383, 5.233879e-28], [942.59423, 5.248112e-28], [945.49351, 5.262395e-28], [948.40172, 5.276727e-28], [951.31886, 5.291109e-28], [954.24499, 5.305541e-28], [957.18011, 5.320022e-28], [960.12426, 5.334553e-28], [963.07746, 5.349135e-28], [966.03975, 5.363767e-28], [969.01115, 5.378449e-28], [971.99169, 5.393182e-28], [974.9814, 5.407965e-28], [977.98031, 5.4228e-28], [980.98844, 5.437685e-28], [984.00582, 5.452621e-28], [987.03248, 5.467609e-28], [990.06845, 5.482648e-28], [993.11376, 5.497739e-28], [996.16844, 5.512882e-28], [999.23251, 5.528076e-28], [1002.306, 5.543322e-28], [1005.389, 5.558621e-28], [1008.4814, 5.573972e-28], [1011.5833, 5.589375e-28], [1014.6948, 5.604831e-28], [1017.8159, 5.620339e-28], [1020.9465, 5.635901e-28], [1024.0868, 5.651515e-28], [1027.2368, 5.667183e-28], [1030.3964, 5.682904e-28], [1033.5658, 5.698679e-28], [1036.7449, 5.714507e-28], [1039.9337, 5.730389e-28], [1043.1324, 5.746325e-28], [1046.341, 5.762315e-28], [1049.5594, 5.77836e-28], [1052.7876, 5.794458e-28], [1056.0259, 5.810612e-28], [1059.2741, 5.82682e-28], [1062.5322, 5.843083e-28], [1065.8004, 5.859401e-28], [1069.0787, 5.875774e-28], [1072.367, 5.892202e-28], [1075.6655, 5.908686e-28], [1078.9741, 5.925226e-28], [1082.2928, 5.941822e-28], [1085.6218, 5.958473e-28], [1088.961, 5.975181e-28], [1092.3105, 5.991945e-28], [1095.6703, 6.008765e-28], [1099.0404, 6.025642e-28], [1102.4209, 6.042576e-28], [1105.8118, 6.059567e-28], [1109.2131, 6.076615e-28], [1112.6249, 6.09372e-28], [1116.0472, 6.110882e-28], [1119.48, 6.128102e-28], [1122.9234, 6.14538e-28], [1126.3773, 6.162716e-28], [1129.8419, 6.18011e-28], [1133.3171, 6.197562e-28], [1136.803, 6.215073e-28], [1140.2997, 6.232642e-28], [1143.8071, 6.25027e-28], [1147.3253, 6.267957e-28], [1150.8543, 6.285703e-28], [1154.3942, 6.303508e-28], [1157.9449, 6.321373e-28], [1161.5066, 6.339297e-28], [1165.0792, 6.357281e-28], [1168.6628, 6.375325e-28], [1172.2575, 6.393429e-28], [1175.8632, 6.411594e-28], [1179.48, 6.429819e-28], [1183.1079, 6.448104e-28], [1186.7469, 6.46645e-28], [1190.3972, 6.484858e-28], [1194.0587, 6.503326e-28], [1197.7314, 6.521856e-28], [1201.4155, 6.540447e-28], [1205.1109, 6.5591e-28], [1208.8176, 6.577815e-28], [1212.5358, 6.596592e-28], [1216.2654, 6.615431e-28], [1220.0064, 6.634333e-28], [1223.759, 6.653297e-28], [1227.5231, 6.672324e-28], [1231.2988, 6.691414e-28], [1235.0861, 6.710567e-28], [1238.885, 6.729783e-28], [1242.6957, 6.749063e-28], [1246.518, 6.768407e-28], [1250.3521, 6.787814e-28], [1254.1981, 6.807286e-28], [1258.0558, 6.826821e-28], [1261.9254, 6.846421e-28], [1265.8069, 6.866086e-28], [1269.7003, 6.885815e-28], [1273.6058, 6.90561e-28], [1277.5232, 6.925469e-28], [1281.4527, 6.945394e-28], [1285.3942, 6.965385e-28], [1289.3479, 6.985441e-28], [1293.3138, 7.005563e-28], [1297.2918, 7.025751e-28], [1301.2821, 7.046006e-28], [1305.2847, 7.066327e-28], [1309.2995, 7.086715e-28], [1313.3268, 7.107169e-28], [1317.3664, 7.127691e-28], [1321.4184, 7.14828e-28], [1325.4829, 7.168937e-28], [1329.5599, 7.189661e-28], [1333.6494, 7.210453e-28], [1337.7515, 7.231313e-28], [1341.8663, 7.252241e-28], [1345.9937, 7.273238e-28], [1350.1337, 7.294303e-28], [1354.2866, 7.315437e-28], [1358.4522, 7.336641e-28], [1362.6306, 7.357913e-28], [1366.8218, 7.379255e-28], [1371.026, 7.400667e-28], [1375.2431, 7.422148e-28], [1379.4731, 7.4437e-28], [1383.7162, 7.465322e-28], [1387.9723, 7.487014e-28], [1392.2415, 7.508777e-28], [1396.5238, 7.530611e-28], [1400.8193, 7.552516e-28], [1405.1281, 7.574492e-28], [1409.45, 7.59654e-28], [1413.7853, 7.61866e-28], [1418.1339, 7.640851e-28], [1422.4959, 7.663114e-28], [1426.8713, 7.68545e-28], [1431.2601, 7.707859e-28], [1435.6625, 7.73034e-28], [1440.0784, 7.752894e-28], [1444.5079, 7.775522e-28], [1448.951, 7.798223e-28], [1453.4077, 7.820997e-28], [1457.8782, 7.843845e-28], [1462.3624, 7.866768e-28], [1466.8604, 7.889764e-28], [1471.3723, 7.912835e-28], [1475.898, 7.935981e-28], [1480.4377, 7.959202e-28], [1484.9913, 7.982498e-28], [1489.5589, 8.005869e-28], [1494.1406, 8.029316e-28], [1498.7364, 8.052839e-28], [1503.3463, 8.076438e-28], [1507.9703, 8.100113e-28], [1512.6086, 8.123864e-28], [1517.2612, 8.147692e-28], [1521.9281, 8.171597e-28], [1526.6093, 8.19558e-28], [1531.305, 8.219639e-28], [1536.015, 8.243776e-28], [1540.7396, 8.267992e-28], [1545.4787, 8.292285e-28], [1550.2324, 8.316656e-28], [1555.0007, 8.341106e-28], [1559.7836, 8.365635e-28], [1564.5813, 8.390243e-28], [1569.3937, 8.41493e-28], [1574.221, 8.439696e-28], [1579.063, 8.464542e-28], [1583.92, 8.489468e-28], [1588.7919, 8.514475e-28], [1593.6788, 8.539561e-28], [1598.5807, 8.564729e-28], [1603.4978, 8.589977e-28], [1608.4299, 8.615306e-28], [1613.3772, 8.640716e-28], [1618.3397, 8.666209e-28], [1623.3175, 8.691783e-28], [1628.3106, 8.717439e-28], [1633.319, 8.743178e-28], [1638.3429, 8.768999e-28], [1643.3822, 8.794902e-28], [1648.437, 8.820889e-28], [1653.5074, 8.84696e-28], [1658.5933, 8.873113e-28], [1663.6949, 8.899351e-28], [1668.8122, 8.925672e-28], [1673.9452, 8.952078e-28], [1679.0941, 8.978568e-28], [1684.2587, 9.005143e-28], [1689.4393, 9.031803e-28], [1694.6357, 9.058549e-28], [1699.8482, 9.08538e-28], [1705.0767, 9.112297e-28], [1710.3213, 9.1393e-28], [1715.582, 9.166389e-28], [1720.8588, 9.193564e-28], [1726.152, 9.220827e-28], [1731.4614, 9.248176e-28], [1736.7871, 9.275614e-28], [1742.1292, 9.303138e-28], [1747.4877, 9.330751e-28], [1752.8628, 9.358451e-28], [1758.2543, 9.38624e-28], [1763.6625, 9.414118e-28], [1769.0872, 9.442084e-28], [1774.5287, 9.47014e-28], [1779.9869, 9.498284e-28], [1785.4619, 9.526519e-28], [1790.9537, 9.554844e-28], [1796.4624, 9.583258e-28], [1801.9881, 9.611764e-28], [1807.5308, 9.64036e-28], [1813.0905, 9.669047e-28], [1818.6673, 9.697826e-28], [1824.2612, 9.726696e-28], [1829.8724, 9.755658e-28], [1835.5008, 9.784712e-28], [1841.1466, 9.813859e-28], [1846.8097, 9.843098e-28], [1852.4902, 9.87243e-28], [1858.1882, 9.901855e-28], [1863.9037, 9.931374e-28], [1869.6368, 9.960987e-28], [1875.3876, 9.990693e-28], [1881.156, 1.002049e-27], [1886.9422, 1.005039e-27], [1892.7461, 1.008038e-27], [1898.5679, 1.011047e-27], [1904.4077, 1.014065e-27], [1910.2653, 1.017092e-27], [1916.1411, 1.02013e-27], [1922.0348, 1.023177e-27], [1927.9467, 1.026233e-27], [1933.8768, 1.0293e-27], [1939.8252, 1.032376e-27], [1945.7918, 1.035461e-27], [1951.7768, 1.038557e-27], [1957.7802, 1.041662e-27], [1963.802, 1.044777e-27], [1969.8424, 1.047902e-27], [1975.9013, 1.051037e-27], [1981.9789, 1.054182e-27], [1988.0752, 1.057337e-27], [1994.1903, 1.060502e-27], [2000.3241, 1.063677e-27], [2006.4768, 1.066861e-27], [2012.6484, 1.070056e-27], [2018.8391, 1.073261e-27], [2025.0487, 1.076477e-27], [2031.2775, 1.079702e-27], [2037.5254, 1.082937e-27], [2043.7926, 1.086183e-27], [2050.079, 1.089439e-27], [2056.3847, 1.092706e-27], [2062.7099, 1.095982e-27], [2069.0545, 1.099269e-27], [2075.4186, 1.102566e-27], [2081.8023, 1.105874e-27], [2088.2056, 1.109192e-27], [2094.6286, 1.112521e-27], [2101.0714, 1.11586e-27], [2107.534, 1.11921e-27], [2114.0165, 1.12257e-27], [2120.5189, 1.125941e-27], [2127.0413, 1.129323e-27], [2133.5838, 1.132715e-27], [2140.1464, 1.136118e-27], [2146.7292, 1.139531e-27], [2153.3322, 1.142956e-27], [2159.9556, 1.146391e-27], [2166.5993, 1.149837e-27], [2173.2634, 1.153294e-27], [2179.9481, 1.156761e-27], [2186.6533, 1.16024e-27], [2193.3791, 1.16373e-27], [2200.1256, 1.16723e-27], [2206.8929, 1.170742e-27], [2213.681, 1.174264e-27], [2220.49, 1.177798e-27], [2227.3199, 1.181343e-27], [2234.1708, 1.184899e-27], [2241.0428, 1.188466e-27], [2247.9359, 1.192045e-27], [2254.8503, 1.195634e-27], [2261.7859, 1.199235e-27], [2268.7428, 1.202848e-27], [2275.7211, 1.206471e-27], [2282.7209, 1.210106e-27], [2289.7422, 1.213753e-27], [2296.7852, 1.217411e-27], [2303.8497, 1.22108e-27], [2310.9361, 1.224761e-27], [2318.0442, 1.228454e-27], [2325.1741, 1.232158e-27], [2332.3261, 1.235874e-27], [2339.5, 1.239601e-27], [2346.6959, 1.24334e-27], [2353.914, 1.247091e-27], [2361.1543, 1.250854e-27], [2368.4169, 1.254628e-27], [2375.7018, 1.258415e-27], [2383.0091, 1.262213e-27], [2390.3389, 1.266023e-27], [2397.6913, 1.269845e-27], [2405.0662, 1.273679e-27], [2412.4639, 1.277525e-27], [2419.8843, 1.281383e-27], [2427.3275, 1.285253e-27], [2434.7936, 1.289136e-27], [2442.2827, 1.29303e-27], [2449.7948, 1.296937e-27], [2457.33, 1.300856e-27], [2464.8884, 1.304787e-27], [2472.4701, 1.308731e-27], [2480.075, 1.312687e-27], [2487.7034, 1.316655e-27], [2495.3552, 1.320636e-27], [2503.0306, 1.324629e-27], [2510.7295, 1.328635e-27], [2518.4522, 1.332653e-27], [2526.1986, 1.336684e-27], [2533.9688, 1.340727e-27], [2541.763, 1.344783e-27], [2549.5811, 1.348852e-27], [2557.4232, 1.352934e-27], [2565.2895, 1.357028e-27], [2573.1799, 1.361135e-27], [2581.0947, 1.365255e-27], [2589.0338, 1.369388e-27], [2596.9973, 1.373534e-27], [2604.9852, 1.377693e-27], [2612.9978, 1.381865e-27], [2621.035, 1.38605e-27], [2629.097, 1.390247e-27], [2637.1837, 1.394459e-27], [2645.2953, 1.398683e-27], [2653.4318, 1.40292e-27], [2661.5934, 1.407171e-27], [2669.7801, 1.411435e-27], [2677.992, 1.415712e-27], [2686.2291, 1.420003e-27], [2694.4915, 1.424307e-27], [2702.7794, 1.428624e-27], [2711.0928, 1.432955e-27], [2719.4317, 1.4373e-27], [2727.7963, 1.441658e-27], [2736.1866, 1.446029e-27], [2744.6027, 1.450415e-27], [2753.0448, 1.454814e-27], [2761.5127, 1.459226e-27], [2770.0067, 1.463653e-27], [2778.5269, 1.468093e-27], [2787.0732, 1.472548e-27], [2795.6459, 1.477016e-27], [2804.2449, 1.481498e-27], [2812.8704, 1.485994e-27], [2821.5223, 1.490504e-27], [2830.201, 1.495028e-27], [2838.9062, 1.499566e-27], [2847.6383, 1.504119e-27], [2856.3972, 1.508685e-27], [2865.1831, 1.513266e-27], [2873.996, 1.517861e-27], [2882.836, 1.522471e-27], [2891.7032, 1.527094e-27], [2900.5977, 1.531733e-27], [2909.5195, 1.536385e-27], [2918.4688, 1.541053e-27], [2927.4456, 1.545734e-27], [2936.45, 1.550431e-27], [2945.4821, 1.555142e-27], [2954.542, 1.559867e-27], [2963.6297, 1.564608e-27], [2972.7454, 1.569363e-27], [2981.8892, 1.574133e-27], [2991.061, 1.578918e-27], [3000.2611, 1.583717e-27], [3009.4895, 1.588532e-27], [3018.7463, 1.593362e-27], [3028.0315, 1.598206e-27], [3037.3453, 1.603066e-27], [3046.6877, 1.607941e-27], [3056.0589, 1.612831e-27], [3065.4589, 1.617737e-27], [3074.8878, 1.622657e-27], [3084.3457, 1.627593e-27], [3093.8327, 1.632544e-27], [3103.3489, 1.637511e-27], [3112.8944, 1.642493e-27], [3122.4692, 1.647491e-27], [3132.0735, 1.652504e-27], [3141.7073, 1.657533e-27], [3151.3708, 1.662577e-27], [3161.0639, 1.667637e-27], [3170.7869, 1.672713e-27], [3180.5398, 1.677805e-27], [3190.3227, 1.682912e-27], [3200.1357, 1.688036e-27], [3209.9788, 1.693175e-27], [3219.8523, 1.69833e-27], [3229.7561, 1.703502e-27], [3239.6903, 1.708689e-27], [3249.6552, 1.713893e-27], [3259.6506, 1.719113e-27], [3269.6769, 1.724348e-27], [3279.7339, 1.729601e-27], [3289.8219, 1.734869e-27], [3299.941, 1.740154e-27], [3310.0911, 1.745455e-27], [3320.2725, 1.750773e-27], [3330.4852, 1.756108e-27], [3340.7293, 1.761458e-27], [3351.0049, 1.766826e-27], [3361.3121, 1.77221e-27], [3371.651, 1.777611e-27], [3382.0217, 1.783029e-27], [3392.4243, 1.788463e-27], [3402.859, 1.793915e-27], [3413.3257, 1.799383e-27], [3423.8246, 1.804869e-27], [3434.3558, 1.810371e-27], [3444.9194, 1.815891e-27], [3455.5154, 1.821427e-27], [3466.1441, 1.826981e-27], [3476.8055, 1.832552e-27], [3487.4996, 1.83814e-27], [3498.2267, 1.843746e-27], [3508.9867, 1.849369e-27], [3519.7799, 1.85501e-27], [3530.6062, 1.860668e-27], [3541.4659, 1.866343e-27], [3552.3589, 1.872037e-27], [3563.2855, 1.877747e-27], [3574.2456, 1.883476e-27], [3585.2395, 1.889222e-27], [3596.2672, 1.894987e-27], [3607.3288, 1.900769e-27], [3618.4244, 1.906569e-27], [3629.5542, 1.912387e-27], [3640.7182, 1.918223e-27], [3651.9165, 1.924077e-27], [3663.1493, 1.92995e-27], [3674.4166, 1.93584e-27], [3685.7186, 1.941749e-27], [3697.0553, 1.947676e-27], [3708.427, 1.953622e-27], [3719.8336, 1.959586e-27], [3731.2752, 1.965568e-27], [3742.7521, 1.971569e-27], [3754.2643, 1.977589e-27], [3765.8119, 1.983627e-27], [3777.395, 1.989684e-27], [3789.0137, 1.99576e-27], [3800.6682, 2.001855e-27], [3812.3585, 2.007968e-27], [3824.0848, 2.0141e-27], [3835.8471, 2.020252e-27], [3847.6456, 2.026423e-27], [3859.4804, 2.032612e-27], [3871.3516, 2.038821e-27], [3883.2594, 2.045049e-27], [3895.2037, 2.051297e-27], [3907.1848, 2.057563e-27], [3919.2028, 2.06385e-27], [3931.2577, 2.070155e-27], [3943.3497, 2.07648e-27], [3955.4788, 2.082825e-27], [3967.6453, 2.08919e-27], [3979.8493, 2.095574e-27], [3992.0907, 2.101978e-27], [4004.3698, 2.108401e-27], [4016.6867, 2.114845e-27], [4029.0414, 2.121309e-27], [4041.4342, 2.127792e-27], [4053.8651, 2.134296e-27], [4066.3342, 2.14082e-27], [4078.8417, 2.147364e-27], [4091.3876, 2.153928e-27], [4103.9721, 2.160513e-27], [4116.5954, 2.167118e-27], [4129.2574, 2.173744e-27], [4141.9584, 2.18039e-27], [4154.6985, 2.187056e-27], [4167.4778, 2.193744e-27], [4180.2963, 2.200452e-27], [4193.1543, 2.20718e-27], [4206.0519, 2.21393e-27], [4218.9891, 2.2207e-27], [4231.9661, 2.227492e-27], [4244.983, 2.234304e-27], [4258.04, 2.241137e-27], [4271.1371, 2.247992e-27], [4284.2745, 2.254868e-27], [4297.4524, 2.261765e-27], [4310.6707, 2.268684e-27], [4323.9297, 2.275623e-27], [4337.2295, 2.282585e-27], [4350.5702, 2.289568e-27], [4363.9519, 2.296572e-27], [4377.3748, 2.303598e-27], [4390.839, 2.310646e-27], [4404.3446, 2.317716e-27], [4417.8918, 2.324808e-27], [4431.4806, 2.331921e-27], [4445.1112, 2.339057e-27], [4458.7837, 2.346214e-27], [4472.4983, 2.353394e-27], [4486.2551, 2.360596e-27], [4500.0541, 2.36782e-27], [4513.8957, 2.375067e-27], [4527.7798, 2.382336e-27], [4541.7066, 2.389627e-27], [4555.6762, 2.396941e-27], [4569.6888, 2.404278e-27], [4583.7446, 2.411637e-27], [4597.8435, 2.419019e-27], [4611.9858, 2.426424e-27], [4626.1716, 2.433852e-27], [4640.4011, 2.441303e-27], [4654.6743, 2.448777e-27], [4668.9914, 2.456274e-27], [4683.3526, 2.463794e-27], [4697.7579, 2.471337e-27], [4712.2075, 2.478904e-27], [4726.7016, 2.486494e-27], [4741.2403, 2.494107e-27], [4755.8236, 2.501744e-27], [4770.4519, 2.509405e-27], [4785.1251, 2.517089e-27], [4799.8435, 2.524797e-27], [4814.6071, 2.532529e-27], [4829.4162, 2.540285e-27], [4844.2708, 2.548065e-27], [4859.1711, 2.555869e-27], [4874.1172, 2.563697e-27], [4889.1093, 2.571549e-27], [4904.1475, 2.579425e-27], [4919.2319, 2.587326e-27], [4934.3628, 2.595251e-27], [4949.5402, 2.603201e-27], [4964.7643, 2.611175e-27], [4980.0352, 2.619174e-27], [4995.3531, 2.627197e-27], [5010.7181, 2.635246e-27], [5026.1303, 2.643319e-27], [5041.59, 2.651417e-27], [5057.0972, 2.65954e-27], [5072.6521, 2.667689e-27], [5088.2549, 2.675862e-27], [5103.9056, 2.684061e-27], [5119.6045, 2.692285e-27], [5135.3517, 2.700534e-27], [5151.1473, 2.708809e-27], [5166.9915, 2.71711e-27], [5182.8844, 2.725436e-27], [5198.8262, 2.733788e-27], [5214.8171, 2.742165e-27], [5230.8571, 2.750569e-27], [5246.9465, 2.758998e-27], [5263.0854, 2.767454e-27], [5279.2739, 2.775935e-27], [5295.5122, 2.784443e-27], [5311.8004, 2.792977e-27], [5328.1387, 2.801538e-27], [5344.5273, 2.810125e-27], [5360.9663, 2.818738e-27], [5377.4559, 2.827378e-27], [5393.9962, 2.836045e-27], [5410.5874, 2.844738e-27], [5427.2296, 2.853458e-27], [5443.923, 2.862205e-27], [5460.6677, 2.87098e-27], [5477.4639, 2.879781e-27], [5494.3118, 2.888609e-27], [5511.2116, 2.897465e-27], [5528.1633, 2.906348e-27], [5545.1671, 2.915258e-27], [5562.2233, 2.924196e-27], [5579.3319, 2.933162e-27], [5596.4931, 2.942155e-27], [5613.7071, 2.951176e-27], [5630.9741, 2.960225e-27], [5648.2942, 2.969302e-27], [5665.6675, 2.978407e-27], [5683.0943, 2.98754e-27], [5700.5747, 2.996701e-27], [5718.1088, 3.00589e-27], [5735.6969, 3.015108e-27], [5753.3391, 3.024354e-27], [5771.0356, 3.033629e-27], [5788.7865, 3.042932e-27], [5806.5919, 3.052264e-27], [5824.4522, 3.061625e-27], [5842.3674, 3.071018e-27], [5860.3377, 3.080439e-27], [5878.3632, 3.08989e-27], [5896.4443, 3.09937e-27], [5914.5809, 3.10888e-27], [5932.7733, 3.118418e-27], [5951.0217, 3.127986e-27], [5969.3262, 3.137584e-27], [5987.687, 3.147211e-27], [6006.1042, 3.156868e-27], [6024.5781, 3.166554e-27], [6043.1089, 3.176271e-27], [6061.6966, 3.186017e-27], [6080.3416, 3.195794e-27], [6099.0438, 3.2056e-27], [6117.8036, 3.215437e-27], [6136.6211, 3.225304e-27], [6155.4965, 3.235202e-27], [6174.4299, 3.24513e-27], [6193.4216, 3.255089e-27], [6212.4717, 3.265078e-27], [6231.5804, 3.275098e-27], [6250.7478, 3.285149e-27], [6269.9742, 3.295232e-27], [6289.2598, 3.305345e-27], [6308.6046, 3.315489e-27], [6328.009, 3.325665e-27], [6347.4731, 3.335872e-27], [6366.997, 3.34611e-27], [6386.581, 3.35638e-27], [6406.2252, 3.366682e-27], [6425.9298, 3.377015e-27], [6445.6951, 3.38738e-27], [6465.5211, 3.397778e-27], [6485.4081, 3.408207e-27], [6505.3563, 3.418668e-27], [6525.3659, 3.429162e-27], [6545.437, 3.439688e-27], [6565.5698, 3.450247e-27], [6585.7646, 3.460838e-27], [6606.0215, 3.471461e-27], [6626.3406, 3.482118e-27], [6646.7223, 3.492807e-27], [6667.1667, 3.503529e-27], [6687.674, 3.514284e-27], [6708.2443, 3.525073e-27], [6728.8779, 3.535894e-27], [6749.575, 3.546749e-27], [6770.3357, 3.557638e-27], [6791.1603, 3.56856e-27], [6812.0489, 3.579515e-27], [6833.0018, 3.590505e-27], [6854.0192, 3.601528e-27], [6875.1012, 3.612585e-27], [6896.248, 3.623676e-27], [6917.4599, 3.634802e-27], [6938.737, 3.645962e-27], [6960.0796, 3.657156e-27], [6981.4878, 3.668385e-27], [7002.9619, 3.679648e-27], [7024.502, 3.690946e-27], [7046.1083, 3.702279e-27], [7067.7812, 3.713646e-27], [7089.5207, 3.725049e-27], [7111.327, 3.736487e-27], [7133.2004, 3.74796e-27], [7155.1412, 3.759468e-27], [7177.1494, 3.771012e-27], [7199.2252, 3.782592e-27], [7221.369, 3.794207e-27], [7243.5809, 3.805858e-27], [7265.8612, 3.817545e-27], [7288.2099, 3.829268e-27], [7310.6274, 3.841027e-27], [7333.1139, 3.852822e-27], [7355.6695, 3.864653e-27], [7378.2945, 3.876521e-27], [7400.9891, 3.888426e-27], [7423.7535, 3.900367e-27], [7446.5879, 3.912345e-27], [7469.4925, 3.92436e-27], [7492.4676, 3.936412e-27], [7515.5134, 3.948501e-27], [7538.63, 3.960627e-27], [7561.8178, 3.972791e-27], [7585.0769, 3.984992e-27], [7608.4075, 3.997231e-27], [7631.8099, 4.009507e-27], [7655.2842, 4.021821e-27], [7678.8308, 4.034174e-27], [7702.4498, 4.046564e-27], [7726.1414, 4.058992e-27], [7749.906, 4.071459e-27], [7773.7436, 4.083964e-27], [7797.6545, 4.096508e-27], [7821.639, 4.10909e-27], [7845.6972, 4.121711e-27], [7869.8295, 4.134371e-27], [7894.0359, 4.14707e-27], [7918.3169, 4.159808e-27], [7942.6725, 4.172585e-27], [7967.103, 4.185401e-27], [7991.6087, 4.198257e-27], [8016.1898, 4.211153e-27], [8040.8464, 4.224088e-27], [8065.5789, 4.237064e-27], [8090.3875, 4.250079e-27], [8115.2724, 4.263134e-27], [8140.2338, 4.27623e-27], [8165.272, 4.289365e-27], [8190.3872, 4.302542e-27], [8215.5797, 4.315758e-27], [8240.8496, 4.329016e-27], [8266.1973, 4.342314e-27], [8291.623, 4.355654e-27], [8317.1268, 4.369034e-27], [8342.7091, 4.382456e-27], [8368.3701, 4.395919e-27], [8394.11, 4.409423e-27], [8419.9291, 4.422969e-27], [8445.8276, 4.436557e-27], [8471.8058, 4.450186e-27], [8497.8639, 4.463858e-27], [8524.0021, 4.477572e-27], [8550.2207, 4.491328e-27], [8576.52, 4.505126e-27], [8602.9001, 4.518967e-27], [8629.3614, 4.53285e-27], [8655.9041, 4.546776e-27], [8682.5285, 4.560745e-27], [8709.2347, 4.574757e-27], [8736.0231, 4.588812e-27], [8762.8938, 4.602911e-27], [8789.8472, 4.617053e-27], [8816.8836, 4.631238e-27], [8844.003, 4.645467e-27], [8871.2059, 4.65974e-27], [8898.4925, 4.674057e-27], [8925.863, 4.688418e-27], [8953.3177, 4.702823e-27], [8980.8568, 4.717273e-27], [9008.4807, 4.731767e-27], [9036.1895, 4.746305e-27], [9063.9835, 4.760889e-27], [9091.863, 4.775517e-27], [9119.8283, 4.79019e-27], [9147.8796, 4.804909e-27], [9176.0172, 4.819673e-27], [9204.2413, 4.834482e-27], [9232.5522, 4.849337e-27], [9260.9503, 4.864238e-27], [9289.4356, 4.879184e-27], [9318.0086, 4.894177e-27], [9346.6695, 4.909215e-27], [9375.4185, 4.9243e-27], [9404.256, 4.939432e-27], [9433.1821, 4.95461e-27], [9462.1973, 4.969835e-27], [9491.3016, 4.985106e-27], [9520.4955, 5.000425e-27], [9549.7792, 5.015791e-27], [9579.153, 5.031204e-27], [9608.6171, 5.046665e-27], [9638.1718, 5.062173e-27], [9667.8175, 5.077729e-27], [9697.5543, 5.093333e-27], [9727.3826, 5.108985e-27], [9757.3027, 5.124685e-27], [9787.3147, 5.140433e-27], [9817.4191, 5.15623e-27], [9847.6161, 5.172076e-27], [9877.906, 5.18797e-27], [9908.289, 5.203913e-27], [9938.7655, 5.219906e-27], [9969.3357, 5.235947e-27], [10000.0, 5.252038e-27]])

    fiveyr = 5 * 365.25 * 86400
    snr = 10
    f1 = data[:,0]
    S1 = fiveyr*(data[:,2]/10)**2
    fstop = [f1[0],f1[-1]]

    # %---- Interpolate to desired frequencies.  Set noise to infinity outside
    # %      band of frequencies covered by f1.
    # S = inf(size(f)); 
    # k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop

    
# % -------------------------------------------------------------------------
# %                              ETD spectrum
# % -------------------------------------------------------------------------

def etd(f):
# % ETD - Design noise curve "D" for ET.
# %
# % Arguments are the same as for the parent function.

# % ---- Extract pre-sampled noise spectrum from data.  First column is
# %      frequency, second column is effective noise amplitude for a 5-year
# %      integration period with SNR=10.  (We'll need to remove these
# %      scalings.)  Data supplied by B. S. Sathyaprakash.
    data = np.array([[1.0, 2.951094e-20], [1.003076, 2.992112e-20], [1.006161, 3.021851e-20], [1.009256, 3.050756e-20], [1.01236, 3.081522e-20], [1.015474, 3.099403e-20], [1.018598, 3.112464e-20], [1.021731, 3.113518e-20], [1.024873, 3.109389e-20], [1.028026, 3.106523e-20], [1.031188, 3.088506e-20], [1.03436, 3.046101e-20], [1.037541, 3.0082e-20], [1.040733, 2.970451e-20], [1.043934, 2.928851e-20], [1.047145, 2.90148e-20], [1.050366, 2.860714e-20], [1.053596, 2.798962e-20], [1.056837, 2.740944e-20], [1.060088, 2.67701e-20], [1.063348, 2.606687e-20], [1.066619, 2.549494e-20], [1.0699, 2.490659e-20], [1.073191, 2.457898e-20], [1.076492, 3.016319e-20], [1.079803, 2.329392e-20], [1.083124, 2.230382e-20], [1.086456, 2.174331e-20], [1.089797, 2.103648e-20], [1.093149, 2.016804e-20], [1.096512, 1.968525e-20], [1.099885, 1.874334e-20], [1.103268, 1.775369e-20], [1.106661, 1.678721e-20], [1.110065, 1.586929e-20], [1.113479, 1.496301e-20], [1.116904, 1.40904e-20], [1.12034, 1.322406e-20], [1.123786, 1.213252e-20], [1.127242, 1.112174e-20], [1.13071, 1.024744e-20], [1.134188, 9.22745e-21], [1.137676, 8.383408e-21], [1.141175, 7.629707e-21], [1.144686, 6.811006e-21], [1.148206, 6.153557e-21], [1.151738, 5.566647e-21], [1.155281, 4.958615e-21], [1.158834, 4.477038e-21], [1.162399, 4.064043e-21], [1.165974, 3.646402e-21], [1.169561, 3.311999e-21], [1.173158, 3.013304e-21], [1.176766, 2.715794e-21], [1.180386, 2.474642e-21], [1.184017, 2.258262e-21], [1.187658, 2.04474e-21], [1.191312, 1.869453e-21], [1.194976, 1.711444e-21], [1.198651, 1.556481e-21], [1.202338, 1.427722e-21], [1.206036, 1.311146e-21], [1.209746, 1.197309e-21], [1.213467, 1.104935e-21], [1.2172, 1.029731e-21], [1.220943, 9.555007e-22], [1.224699, 8.93124e-22], [1.228466, 8.361043e-22], [1.232244, 7.79897e-22], [1.236035, 7.323545e-22], [1.239837, 6.887888e-22], [1.24365, 6.458856e-22], [1.247475, 6.094664e-22], [1.251313, 5.7606e-22], [1.255161, 5.431839e-22], [1.259022, 5.153034e-22], [1.262895, 4.89761e-22], [1.266779, 4.646325e-22], [1.270676, 4.44735e-22], [1.274584, 4.303032e-22], [1.278504, 4.158979e-22], [1.282437, 4.04775e-22], [1.286382, 3.950916e-22], [1.290338, 3.854272e-22], [1.294307, 3.791325e-22], [1.298288, 3.743214e-22], [1.302282, 3.695179e-22], [1.306287, 3.689215e-22], [1.310305, 3.701767e-22], [1.314336, 3.714139e-22], [1.318378, 3.795316e-22], [1.322433, 3.906622e-22], [1.326501, 4.016967e-22], [1.330581, 4.287603e-22], [1.334674, 4.621643e-22], [1.338779, 4.948977e-22], [1.342897, 6.011248e-22], [1.347028, 7.333729e-22], [1.351171, 8.550193e-22], [1.355327, 1.090507e-21], [1.359496, 1.316665e-21], [1.363677, 1.474657e-21], [1.367872, 1.3488e-21], [1.372079, 1.017119e-21], [1.376299, 5.443354e-22], [1.380533, 3.19649e-22], [1.384779, 2.537994e-22], [1.389038, 1.875733e-22], [1.393311, 1.497312e-22], [1.397596, 1.25554e-22], [1.401895, 1.014496e-22], [1.406207, 8.565099e-23], [1.410533, 7.384629e-23], [1.414871, 6.209802e-23], [1.419223, 5.37975e-23], [1.423588, 4.717226e-23], [1.427967, 4.058264e-23], [1.432359, 3.570309e-23], [1.436765, 3.166395e-23], [1.441185, 2.764783e-23], [1.445617, 2.457709e-23], [1.450064, 2.197624e-23], [1.454524, 1.939086e-23], [1.458998, 1.735936e-23], [1.463486, 1.559216e-23], [1.467987, 1.384142e-23], [1.472502, 1.24464e-23], [1.477032, 1.123353e-23], [1.481575, 1.003238e-23], [1.486132, 9.061863e-24], [1.490703, 8.210868e-24], [1.495288, 7.368398e-24], [1.499888, 6.679989e-24], [1.504501, 6.072352e-24], [1.509129, 5.471049e-24], [1.51377, 4.975106e-24], [1.518427, 4.53505e-24], [1.523097, 4.099782e-24], [1.527782, 3.737394e-24], [1.532481, 3.41328e-24], [1.537195, 3.093147e-24], [1.541923, 2.825239e-24], [1.546666, 2.585767e-24], [1.551423, 2.349401e-24], [1.556195, 2.150459e-24], [1.560982, 1.972148e-24], [1.565783, 1.796297e-24], [1.570599, 1.647561e-24], [1.57543, 1.513962e-24], [1.580276, 1.382347e-24], [1.585137, 1.27056e-24], [1.590012, 1.16999e-24], [1.594903, 1.071049e-24], [1.599809, 9.873562e-25], [1.604729, 9.133269e-25], [1.609665, 8.403461e-25], [1.614616, 7.779554e-25], [1.619583, 7.216551e-25], [1.624564, 6.662647e-25], [1.629561, 6.188009e-25], [1.634574, 5.759674e-25], [1.639601, 5.339369e-25], [1.644645, 4.978568e-25], [1.649703, 4.653117e-25], [1.654777, 4.334821e-25], [1.659867, 4.061219e-25], [1.664973, 3.814642e-25], [1.670094, 3.574438e-25], [1.675231, 3.370782e-25], [1.680384, 3.195184e-25], [1.685552, 3.023209e-25], [1.690737, 2.874592e-25], [1.695937, 2.74018e-25], [1.701154, 2.609014e-25], [1.706386, 2.495285e-25], [1.711635, 2.392325e-25], [1.7169, 2.292177e-25], [1.722181, 2.20491e-25], [1.727478, 2.125714e-25], [1.732791, 2.04887e-25], [1.738121, 1.981434e-25], [1.743467, 1.919973e-25], [1.74883, 1.860419e-25], [1.754209, 1.807443e-25], [1.759605, 1.75835e-25], [1.765017, 1.710926e-25], [1.770446, 1.668486e-25], [1.775892, 1.629327e-25], [1.781354, 1.591453e-25], [1.786833, 1.557102e-25], [1.792329, 1.525107e-25], [1.797842, 1.494095e-25], [1.803372, 1.465568e-25], [1.808919, 1.438734e-25], [1.814483, 1.412653e-25], [1.820064, 1.388332e-25], [1.825662, 1.365235e-25], [1.831278, 1.342722e-25], [1.836911, 1.321732e-25], [1.842561, 1.302171e-25], [1.848228, 1.282888e-25], [1.853913, 1.26438e-25], [1.859615, 1.246425e-25], [1.865335, 1.228721e-25], [1.871073, 1.211623e-25], [1.876828, 1.194975e-25], [1.882601, 1.178553e-25], [1.888392, 1.162616e-25], [1.8942, 1.14705e-25], [1.900026, 1.131693e-25], [1.90587, 1.11673e-25], [1.911733, 1.10208e-25], [1.917613, 1.087623e-25], [1.923511, 1.073502e-25], [1.929428, 1.059666e-25], [1.935362, 1.046006e-25], [1.941315, 1.03262e-25], [1.947286, 1.019466e-25], [1.953276, 1.006479e-25], [1.959284, 9.937305e-26], [1.96531, 9.8119e-26], [1.971355, 9.688082e-26], [1.977419, 9.566372e-26], [1.983501, 9.446546e-26], [1.989602, 9.328235e-26], [1.995722, 9.211817e-26], [2.00186, 9.097126e-26], [2.008018, 8.98388e-26], [2.014194, 8.8722e-26], [2.02039, 8.761846e-26], [2.026604, 8.652969e-26], [2.032838, 8.545746e-26], [2.03909, 8.440072e-26], [2.045362, 8.335787e-26], [2.051654, 8.233008e-26], [2.057964, 8.13166e-26], [2.064294, 8.031628e-26], [2.070644, 7.932984e-26], [2.077013, 7.835677e-26], [2.083401, 7.73962e-26], [2.089809, 7.644859e-26], [2.096237, 7.551356e-26], [2.102685, 7.459044e-26], [2.109153, 7.367945e-26], [2.11564, 7.278025e-26], [2.122148, 7.189248e-26], [2.128675, 7.101628e-26], [2.135223, 7.015142e-26], [2.14179, 6.929749e-26], [2.148378, 6.845455e-26], [2.154986, 6.762241e-26], [2.161615, 6.680075e-26], [2.168263, 6.598955e-26], [2.174933, 6.518868e-26], [2.181622, 6.439786e-26], [2.188333, 6.361705e-26], [2.195064, 6.284611e-26], [2.201815, 6.208484e-26], [2.208588, 6.133315e-26], [2.215381, 6.059096e-26], [2.222196, 5.985804e-26], [2.229031, 5.913429e-26], [2.235887, 5.841964e-26], [2.242764, 5.771389e-26], [2.249662, 5.701696e-26], [2.256582, 5.632876e-26], [2.263523, 5.564913e-26], [2.270485, 5.497797e-26], [2.277469, 5.431521e-26], [2.284474, 5.366068e-26], [2.291501, 5.301431e-26], [2.298549, 5.2376e-26], [2.305619, 5.174562e-26], [2.312711, 5.112309e-26], [2.319825, 5.050835e-26], [2.32696, 4.990123e-26], [2.334118, 4.930164e-26], [2.341297, 4.870951e-26], [2.348498, 4.812472e-26], [2.355722, 4.754718e-26], [2.362968, 4.697682e-26], [2.370236, 4.641352e-26], [2.377527, 4.58572e-26], [2.38484, 4.530779e-26], [2.392175, 4.476518e-26], [2.399533, 4.422928e-26], [2.406913, 4.370004e-26], [2.414317, 4.317734e-26], [2.421743, 4.26611e-26], [2.429192, 4.215125e-26], [2.436664, 4.164771e-26], [2.444159, 4.115039e-26], [2.451676, 4.065923e-26], [2.459217, 4.017415e-26], [2.466782, 3.969505e-26], [2.474369, 3.922188e-26], [2.48198, 3.875456e-26], [2.489614, 3.8293e-26], [2.497272, 3.783715e-26], [2.504953, 3.738692e-26], [2.512658, 3.694225e-26], [2.520387, 3.650307e-26], [2.528139, 3.60693e-26], [2.535915, 3.564087e-26], [2.543715, 3.521774e-26], [2.551539, 3.479982e-26], [2.559388, 3.438704e-26], [2.56726, 3.397936e-26], [2.575156, 3.357669e-26], [2.583077, 3.317897e-26], [2.591022, 3.278616e-26], [2.598992, 3.239818e-26], [2.606986, 3.201496e-26], [2.615005, 3.163647e-26], [2.623048, 3.126263e-26], [2.631116, 3.089337e-26], [2.639209, 3.052866e-26], [2.647327, 3.016843e-26], [2.65547, 2.981262e-26], [2.663638, 2.946118e-26], [2.671831, 2.911406e-26], [2.680049, 2.877119e-26], [2.688292, 2.843253e-26], [2.696561, 2.809803e-26], [2.704855, 2.776761e-26], [2.713175, 2.744126e-26], [2.721521, 2.71189e-26], [2.729891, 2.680049e-26], [2.738288, 2.648598e-26], [2.746711, 2.617532e-26], [2.755159, 2.586846e-26], [2.763634, 2.556536e-26], [2.772134, 2.526597e-26], [2.780661, 2.497023e-26], [2.789214, 2.467811e-26], [2.797793, 2.438956e-26], [2.806399, 2.410453e-26], [2.815031, 2.382299e-26], [2.823689, 2.354488e-26], [2.832375, 2.327016e-26], [2.841087, 2.299881e-26], [2.849826, 2.273076e-26], [2.858591, 2.246597e-26], [2.867384, 2.220442e-26], [2.876203, 2.194605e-26], [2.88505, 2.169083e-26], [2.893924, 2.143872e-26], [2.902826, 2.118967e-26], [2.911754, 2.094366e-26], [2.92071, 2.070064e-26], [2.929694, 2.046058e-26], [2.938705, 2.022343e-26], [2.947744, 1.998916e-26], [2.956811, 1.975775e-26], [2.965906, 1.952914e-26], [2.975029, 1.93033e-26], [2.98418, 1.908021e-26], [2.993358, 1.885983e-26], [3.002566, 1.864211e-26], [3.011801, 1.842704e-26], [3.021065, 1.821457e-26], [3.030357, 1.800468e-26], [3.039678, 1.779733e-26], [3.049028, 1.759248e-26], [3.058406, 1.739012e-26], [3.067813, 1.719021e-26], [3.07725, 1.699271e-26], [3.086715, 1.67976e-26], [3.096209, 1.660484e-26], [3.105733, 1.641441e-26], [3.115285, 1.622629e-26], [3.124868, 1.604043e-26], [3.134479, 1.585681e-26], [3.14412, 1.567541e-26], [3.153791, 1.549619e-26], [3.163492, 1.531913e-26], [3.173222, 1.514421e-26], [3.182983, 1.497139e-26], [3.192773, 1.480065e-26], [3.202594, 1.463196e-26], [3.212444, 1.446531e-26], [3.222325, 1.430065e-26], [3.232237, 1.413797e-26], [3.242179, 1.397724e-26], [3.252151, 1.381844e-26], [3.262154, 1.366155e-26], [3.272188, 1.350654e-26], [3.282253, 1.335339e-26], [3.292349, 1.320207e-26], [3.302476, 1.305256e-26], [3.312634, 1.290484e-26], [3.322823, 1.275889e-26], [3.333043, 1.261468e-26], [3.343295, 1.24722e-26], [3.353579, 1.233142e-26], [3.363894, 1.219232e-26], [3.374241, 1.205487e-26], [3.384619, 1.191907e-26], [3.39503, 1.178489e-26], [3.405473, 1.16523e-26], [3.415947, 1.152129e-26], [3.426454, 1.139185e-26], [3.436994, 1.126394e-26], [3.447565, 1.113755e-26], [3.458169, 1.101267e-26], [3.468806, 1.088927e-26], [3.479476, 1.076733e-26], [3.490178, 1.064684e-26], [3.500914, 1.052778e-26], [3.511682, 1.041013e-26], [3.522483, 1.029388e-26], [3.533318, 1.017899e-26], [3.544186, 1.006547e-26], [3.555087, 9.953297e-27], [3.566022, 9.842443e-27], [3.576991, 9.732899e-27], [3.587993, 9.624651e-27], [3.599029, 9.517677e-27], [3.6101, 9.411966e-27], [3.621204, 9.307501e-27], [3.632342, 9.204265e-27], [3.643515, 9.102246e-27], [3.654721, 9.001427e-27], [3.665963, 8.90179e-27], [3.677239, 8.803328e-27], [3.68855, 8.706021e-27], [3.699895, 8.609854e-27], [3.711275, 8.514818e-27], [3.722691, 8.420895e-27], [3.734141, 8.32807e-27], [3.745627, 8.236335e-27], [3.757148, 8.145672e-27], [3.768704, 8.056068e-27], [3.780296, 7.967513e-27], [3.791924, 7.879992e-27], [3.803587, 7.79349e-27], [3.815287, 7.707999e-27], [3.827022, 7.623505e-27], [3.838793, 7.539993e-27], [3.850601, 7.457455e-27], [3.862445, 7.375877e-27], [3.874325, 7.295246e-27], [3.886242, 7.215554e-27], [3.898195, 7.136787e-27], [3.910186, 7.058932e-27], [3.922213, 6.981981e-27], [3.934277, 6.905923e-27], [3.946378, 6.830743e-27], [3.958517, 6.756435e-27], [3.970693, 6.682987e-27], [3.982906, 6.610385e-27], [3.995157, 6.538624e-27], [4.007446, 6.467691e-27], [4.019772, 6.397575e-27], [4.032136, 6.328269e-27], [4.044538, 6.259761e-27], [4.056979, 6.19204e-27], [4.069458, 6.125099e-27], [4.081975, 6.058928e-27], [4.09453, 5.993516e-27], [4.107124, 5.928856e-27], [4.119757, 5.864938e-27], [4.132429, 5.801752e-27], [4.14514, 5.739291e-27], [4.15789, 5.677545e-27], [4.170679, 5.616504e-27], [4.183507, 5.556162e-27], [4.196375, 5.49651e-27], [4.209282, 5.437538e-27], [4.22223, 5.37924e-27], [4.235217, 5.321607e-27], [4.248244, 5.264629e-27], [4.26131, 5.208302e-27], [4.274418, 5.152616e-27], [4.287565, 5.097561e-27], [4.300753, 5.043134e-27], [4.313982, 4.989325e-27], [4.327251, 4.936126e-27], [4.340561, 4.883531e-27], [4.353912, 4.831533e-27], [4.367304, 4.780123e-27], [4.380737, 4.729295e-27], [4.394211, 4.679043e-27], [4.407728, 4.629358e-27], [4.421285, 4.580236e-27], [4.434884, 4.531669e-27], [4.448525, 4.483648e-27], [4.462208, 4.436171e-27], [4.475934, 4.389228e-27], [4.489701, 4.342814e-27], [4.50351, 4.296923e-27], [4.517363, 4.251549e-27], [4.531257, 4.206684e-27], [4.545195, 4.162324e-27], [4.559175, 4.118462e-27], [4.573199, 4.075093e-27], [4.587265, 4.03221e-27], [4.601375, 3.989809e-27], [4.615528, 3.947882e-27], [4.629725, 3.906426e-27], [4.643965, 3.865434e-27], [4.658249, 3.824901e-27], [4.672578, 3.784821e-27], [4.68695, 3.74519e-27], [4.701366, 3.706001e-27], [4.715827, 3.66725e-27], [4.730332, 3.628932e-27], [4.744882, 3.591042e-27], [4.759476, 3.553574e-27], [4.774116, 3.516525e-27], [4.7888, 3.479889e-27], [4.80353, 3.443661e-27], [4.818305, 3.407837e-27], [4.833126, 3.372412e-27], [4.847992, 3.337381e-27], [4.862903, 3.302741e-27], [4.877861, 3.268486e-27], [4.892864, 3.234613e-27], [4.907914, 3.201117e-27], [4.92301, 3.167993e-27], [4.938153, 3.135238e-27], [4.953342, 3.102848e-27], [4.968578, 3.070817e-27], [4.98386, 3.039144e-27], [4.99919, 3.007823e-27], [5.014567, 2.97685e-27], [5.029991, 2.946222e-27], [5.045462, 2.915935e-27], [5.060981, 2.885985e-27], [5.076548, 2.856369e-27], [5.092163, 2.827083e-27], [5.107826, 2.798123e-27], [5.123537, 2.769486e-27], [5.139296, 2.741169e-27], [5.155104, 2.713167e-27], [5.17096, 2.685479e-27], [5.186865, 2.6581e-27], [5.202819, 2.631027e-27], [5.218822, 2.604257e-27], [5.234875, 2.577787e-27], [5.250977, 2.551613e-27], [5.267128, 2.525733e-27], [5.283329, 2.500144e-27], [5.29958, 2.474842e-27], [5.31588, 2.449825e-27], [5.332231, 2.425091e-27], [5.348632, 2.400634e-27], [5.365084, 2.376454e-27], [5.381586, 2.352548e-27], [5.398139, 2.328913e-27], [5.414743, 2.305545e-27], [5.431398, 2.282444e-27], [5.448104, 2.259604e-27], [5.464862, 2.237026e-27], [5.481671, 2.214705e-27], [5.498532, 2.19264e-27], [5.515445, 2.170828e-27], [5.532409, 2.149268e-27], [5.549426, 2.127955e-27], [5.566496, 2.106889e-27], [5.583617, 2.086067e-27], [5.600792, 2.065487e-27], [5.618019, 2.045147e-27], [5.635299, 2.025045e-27], [5.652633, 2.005179e-27], [5.670019, 1.985547e-27], [5.687459, 1.966147e-27], [5.704953, 1.946978e-27], [5.722501, 1.928037e-27], [5.740102, 1.909323e-27], [5.757758, 1.890835e-27], [5.775468, 1.87257e-27], [5.793233, 1.854528e-27], [5.811052, 1.836707e-27], [5.828926, 1.819104e-27], [5.846855, 1.801721e-27], [5.864839, 1.784553e-27], [5.882878, 1.767602e-27], [5.900973, 1.750865e-27], [5.919124, 1.734341e-27], [5.93733, 1.71803e-27], [5.955592, 1.70193e-27], [5.973911, 1.686041e-27], [5.992286, 1.670361e-27], [6.010717, 1.654891e-27], [6.029205, 1.639628e-27], [6.047751, 1.624573e-27], [6.066352, 1.609725e-27], [6.085012, 1.595082e-27], [6.103728, 1.580646e-27], [6.122503, 1.566416e-27], [6.141335, 1.552389e-27], [6.160224, 1.538568e-27], [6.179172, 1.524951e-27], [6.198179, 1.511537e-27], [6.217243, 1.498327e-27], [6.236367, 1.485321e-27], [6.255549, 1.472517e-27], [6.27479, 1.459917e-27], [6.29409, 1.447519e-27], [6.31345, 1.435324e-27], [6.332869, 1.423331e-27], [6.352348, 1.41154e-27], [6.371887, 1.39995e-27], [6.391486, 1.388563e-27], [6.411146, 1.377377e-27], [6.430866, 1.366392e-27], [6.450646, 1.355608e-27], [6.470487, 1.345024e-27], [6.490389, 1.33464e-27], [6.510353, 1.324456e-27], [6.530378, 1.314471e-27], [6.550464, 1.304683e-27], [6.570613, 1.295094e-27], [6.590823, 1.285701e-27], [6.611095, 1.276504e-27], [6.63143, 1.267502e-27], [6.651827, 1.258694e-27], [6.672288, 1.250079e-27], [6.692811, 1.241655e-27], [6.713397, 1.233421e-27], [6.734046, 1.225376e-27], [6.754759, 1.217518e-27], [6.775536, 1.209845e-27], [6.796376, 1.202355e-27], [6.817281, 1.195047e-27], [6.83825, 1.187918e-27], [6.859284, 1.180967e-27], [6.880382, 1.17419e-27], [6.901545, 1.167587e-27], [6.922773, 1.161153e-27], [6.944066, 1.154888e-27], [6.965426, 1.148788e-27], [6.98685, 1.14285e-27], [7.008341, 1.137072e-27], [7.029897, 1.131451e-27], [7.05152, 1.125984e-27], [7.07321, 1.120668e-27], [7.094966, 1.115501e-27], [7.116789, 1.110479e-27], [7.138679, 1.1056e-27], [7.160637, 1.10086e-27], [7.182662, 1.096256e-27], [7.204755, 1.091786e-27], [7.226916, 1.087446e-27], [7.249145, 1.083234e-27], [7.271442, 1.079146e-27], [7.293808, 1.075179e-27], [7.316243, 1.071331e-27], [7.338746, 1.067599e-27], [7.361319, 1.063979e-27], [7.383962, 1.06047e-27], [7.406674, 1.057068e-27], [7.429456, 1.05377e-27], [7.452307, 1.050575e-27], [7.47523, 1.047478e-27], [7.498222, 1.044479e-27], [7.521286, 1.041574e-27], [7.54442, 1.038761e-27], [7.567626, 1.036038e-27], [7.590903, 1.033402e-27], [7.614251, 1.030851e-27], [7.637672, 1.028383e-27], [7.661164, 1.025996e-27], [7.684729, 1.023688e-27], [7.708366, 1.021457e-27], [7.732076, 1.019301e-27], [7.755859, 1.017218e-27], [7.779714, 1.015206e-27], [7.803644, 1.013264e-27], [7.827647, 1.01139e-27], [7.851723, 1.009581e-27], [7.875874, 1.007838e-27], [7.900099, 1.006157e-27], [7.924399, 1.004537e-27], [7.948773, 1.002977e-27], [7.973222, 1.001475e-27], [7.997747, 1.000029e-27], [8.022347, 9.986376e-28], [8.047022, 9.972994e-28], [8.071774, 9.960126e-28], [8.096602, 9.947753e-28], [8.121506, 9.93586e-28], [8.146486, 9.924429e-28], [8.171544, 9.913444e-28], [8.196678, 9.902888e-28], [8.22189, 9.892748e-28], [8.247179, 9.883008e-28], [8.272546, 9.873656e-28], [8.297992, 9.864681e-28], [8.323515, 9.856073e-28], [8.349117, 9.847824e-28], [8.374798, 9.839928e-28], [8.400557, 9.832377e-28], [8.426396, 9.825168e-28], [8.452315, 9.818297e-28], [8.478313, 9.81176e-28], [8.504391, 9.805551e-28], [8.530549, 9.799667e-28], [8.556788, 9.794105e-28], [8.583107, 9.788858e-28], [8.609508, 9.783923e-28], [8.63599, 9.779293e-28], [8.662553, 9.774961e-28], [8.689197, 9.770925e-28], [8.715924, 9.767174e-28], [8.742733, 9.763703e-28], [8.769624, 9.760505e-28], [8.796598, 9.757574e-28], [8.823656, 9.754901e-28], [8.850796, 9.752481e-28], [8.87802, 9.750305e-28], [8.905327, 9.748367e-28], [8.932719, 9.746659e-28], [8.960195, 9.745175e-28], [8.987755, 9.743908e-28], [9.0154, 9.74285e-28], [9.04313, 9.741996e-28], [9.070945, 9.741338e-28], [9.098846, 9.74087e-28], [9.126833, 9.740585e-28], [9.154906, 9.740478e-28], [9.183065, 9.740541e-28], [9.211311, 9.74077e-28], [9.239644, 9.741157e-28], [9.268063, 9.741697e-28], [9.296571, 9.742383e-28], [9.325166, 9.743211e-28], [9.353848, 9.744175e-28], [9.38262, 9.745269e-28], [9.411479, 9.746487e-28], [9.440428, 9.747824e-28], [9.469465, 9.749275e-28], [9.498592, 9.750835e-28], [9.527808, 9.752498e-28], [9.557114, 9.75426e-28], [9.58651, 9.756116e-28], [9.615997, 9.75806e-28], [9.645575, 9.760088e-28], [9.675243, 9.762195e-28], [9.705003, 9.764378e-28], [9.734854, 9.76663e-28], [9.764797, 9.768948e-28], [9.794832, 9.771327e-28], [9.82496, 9.773763e-28], [9.85518, 9.776251e-28], [9.885493, 9.778787e-28], [9.915899, 9.781369e-28], [9.946399, 9.78399e-28], [9.976993, 9.786647e-28], [10.007681, 9.789336e-28], [10.038463, 9.792054e-28], [10.06934, 9.794796e-28], [10.100312, 9.79756e-28], [10.131379, 9.800341e-28], [10.162542, 9.803135e-28], [10.1938, 9.805938e-28], [10.225155, 9.808749e-28], [10.256606, 9.811563e-28], [10.288154, 9.814376e-28], [10.319799, 9.817187e-28], [10.351541, 9.81999e-28], [10.383381, 9.822782e-28], [10.415319, 9.825562e-28], [10.447355, 9.828326e-28], [10.479489, 9.831069e-28], [10.511723, 9.833792e-28], [10.544055, 9.836488e-28], [10.576487, 9.839156e-28], [10.609019, 9.841794e-28], [10.641651, 9.844398e-28], [10.674383, 9.846965e-28], [10.707216, 9.849493e-28], [10.74015, 9.851979e-28], [10.773185, 9.85442e-28], [10.806322, 9.856815e-28], [10.83956, 9.85916e-28], [10.872901, 9.861454e-28], [10.906345, 9.863692e-28], [10.939891, 9.865874e-28], [10.973541, 9.867997e-28], [11.007294, 9.87006e-28], [11.041151, 9.872058e-28], [11.075112, 9.87399e-28], [11.109177, 9.875855e-28], [11.143347, 9.877649e-28], [11.177623, 9.879371e-28], [11.212004, 9.88102e-28], [11.24649, 9.882592e-28], [11.281083, 9.884086e-28], [11.315782, 9.8855e-28], [11.350587, 9.886832e-28], [11.3855, 9.88808e-28], [11.42052, 9.889243e-28], [11.455648, 9.890317e-28], [11.490884, 9.891303e-28], [11.526229, 9.892198e-28], [11.561682, 9.893001e-28], [11.597244, 9.893709e-28], [11.632915, 9.894322e-28], [11.668696, 9.894837e-28], [11.704588, 9.895253e-28], [11.740589, 9.895568e-28], [11.776702, 9.895782e-28], [11.812925, 9.895892e-28], [11.84926, 9.895897e-28], [11.885707, 9.895796e-28], [11.922265, 9.895587e-28], [11.958937, 9.89527e-28], [11.995721, 9.894842e-28], [12.032618, 9.894303e-28], [12.069628, 9.89365e-28], [12.106753, 9.892884e-28], [12.143991, 9.892002e-28], [12.181345, 9.891005e-28], [12.218813, 9.889889e-28], [12.256396, 9.888655e-28], [12.294095, 9.887302e-28], [12.33191, 9.885828e-28], [12.369841, 9.884232e-28], [12.407889, 9.882513e-28], [12.446054, 9.880671e-28], [12.484336, 9.878704e-28], [12.522736, 9.876612e-28], [12.561254, 9.874393e-28], [12.599891, 9.872048e-28], [12.638646, 9.869573e-28], [12.677521, 9.866972e-28], [12.716515, 9.86424e-28], [12.75563, 9.861378e-28], [12.794864, 9.858387e-28], [12.834219, 9.855263e-28], [12.873696, 9.852008e-28], [12.913293, 9.848621e-28], [12.953013, 9.8451e-28], [12.992854, 9.841446e-28], [13.032818, 9.837661e-28], [13.072906, 9.83374e-28], [13.113116, 9.829687e-28], [13.15345, 9.825501e-28], [13.193908, 9.821177e-28], [13.234491, 9.816722e-28], [13.275198, 9.812137e-28], [13.316031, 9.80741e-28], [13.356989, 9.802557e-28], [13.398073, 9.797585e-28], [13.439284, 9.792474e-28], [13.480621, 9.787219e-28], [13.522086, 9.781831e-28], [13.563678, 9.776336e-28], [13.605398, 9.770716e-28], [13.647246, 9.764963e-28], [13.689223, 9.759103e-28], [13.731329, 9.753125e-28], [13.773565, 9.74703e-28], [13.81593, 9.74084e-28], [13.858426, 9.734551e-28], [13.901052, 9.728166e-28], [13.94381, 9.721713e-28], [13.986699, 9.715193e-28], [14.02972, 9.708615e-28], [14.072874, 9.702021e-28], [14.11616, 9.695424e-28], [14.159579, 9.688848e-28], [14.203132, 9.682364e-28], [14.246819, 9.676013e-28], [14.29064, 9.669861e-28], [14.334596, 9.664058e-28], [14.378687, 9.658729e-28], [14.422914, 9.654062e-28], [14.467277, 9.650463e-28], [14.511776, 9.648395e-28], [14.556412, 9.648521e-28], [14.601186, 9.652375e-28], [14.646097, 9.662402e-28], [14.691146, 9.681712e-28], [14.736334, 9.72068e-28], [14.781661, 9.807499e-28], [14.827127, 1.064759e-27], [14.872734, 1.516649e-27], [14.91848, 1.974594e-27], [14.964367, 1.876603e-27], [15.010395, 1.376186e-27], [15.056565, 1.031027e-27], [15.102877, 9.828599e-28], [15.149331, 9.665147e-28], [15.195929, 9.579684e-28], [15.242669, 9.529715e-28], [15.289553, 9.494343e-28], [15.336582, 9.46728e-28], [15.383755, 9.444711e-28], [15.431073, 9.424871e-28], [15.478537, 9.406898e-28], [15.526147, 9.390129e-28], [15.573903, 9.374132e-28], [15.621806, 9.35869e-28], [15.669857, 9.343619e-28], [15.718055, 9.328778e-28], [15.766401, 9.314093e-28], [15.814897, 9.2995e-28], [15.863541, 9.28494e-28], [15.912335, 9.270384e-28], [15.961279, 9.255806e-28], [16.010374, 9.241178e-28], [16.059619, 9.226488e-28], [16.109016, 9.211722e-28], [16.158565, 9.196868e-28], [16.208267, 9.181919e-28], [16.258121, 9.166869e-28], [16.308129, 9.151711e-28], [16.35829, 9.136441e-28], [16.408606, 9.121057e-28], [16.459077, 9.105555e-28], [16.509702, 9.089933e-28], [16.560484, 9.07419e-28], [16.611422, 9.058324e-28], [16.662516, 9.042335e-28], [16.713768, 9.026223e-28], [16.765177, 9.009985e-28], [16.816744, 8.993623e-28], [16.86847, 8.977139e-28], [16.920355, 8.960529e-28], [16.9724, 8.943796e-28], [17.024604, 8.926941e-28], [17.07697, 8.909963e-28], [17.129496, 8.892864e-28], [17.182184, 8.875646e-28], [17.235034, 8.858306e-28], [17.288046, 8.84085e-28], [17.341222, 8.82328e-28], [17.394561, 8.805591e-28], [17.448064, 8.787788e-28], [17.501732, 8.769877e-28], [17.555565, 8.751855e-28], [17.609563, 8.73372e-28], [17.663728, 8.715476e-28], [17.718059, 8.697133e-28], [17.772557, 8.678688e-28], [17.827223, 8.660138e-28], [17.882057, 8.641493e-28], [17.93706, 8.622751e-28], [17.992231, 8.603914e-28], [18.047573, 8.584988e-28], [18.103085, 8.565974e-28], [18.158767, 8.546873e-28], [18.214621, 8.527692e-28], [18.270647, 8.508431e-28], [18.326845, 8.489093e-28], [18.383215, 8.469684e-28], [18.439759, 8.450207e-28], [18.496478, 8.430663e-28], [18.55337, 8.41106e-28], [18.610438, 8.3914e-28], [18.667681, 8.371687e-28], [18.7251, 8.351929e-28], [18.782696, 8.332128e-28], [18.840468, 8.312291e-28], [18.898419, 8.292424e-28], [18.956548, 8.272533e-28], [19.014856, 8.252624e-28], [19.073343, 8.232706e-28], [19.132009, 8.212788e-28], [19.190857, 8.192876e-28], [19.249885, 8.172983e-28], [19.309095, 8.153119e-28], [19.368487, 8.133296e-28], [19.428062, 8.113529e-28], [19.48782, 8.093833e-28], [19.547762, 8.074226e-28], [19.607888, 8.054732e-28], [19.668199, 8.035374e-28], [19.728695, 8.016179e-28], [19.789378, 7.997187e-28], [19.850247, 7.978439e-28], [19.911304, 7.959983e-28], [19.972548, 7.94189e-28], [20.033981, 7.924241e-28], [20.095603, 7.907134e-28], [20.157414, 7.890713e-28], [20.219415, 7.875163e-28], [20.281607, 7.860711e-28], [20.343991, 7.847708e-28], [20.406566, 7.836664e-28], [20.469333, 7.828219e-28], [20.532294, 7.823462e-28], [20.595449, 7.824244e-28], [20.658797, 7.832931e-28], [20.722341, 7.854401e-28], [20.78608, 7.899394e-28], [20.850015, 7.990139e-28], [20.914147, 8.192164e-28], [20.978475, 8.59472e-28], [21.043002, 9.90733e-28], [21.107728, 1.383108e-27], [21.172652, 1.683521e-27], [21.237776, 1.635834e-27], [21.3031, 1.232569e-27], [21.368626, 8.466577e-28], [21.434353, 7.858574e-28], [21.500282, 7.694076e-28], [21.566413, 7.599311e-28], [21.632749, 7.532754e-28], [21.699288, 7.482607e-28], [21.766032, 7.443135e-28], [21.832981, 7.410058e-28], [21.900136, 7.381299e-28], [21.967498, 7.355681e-28], [22.035067, 7.332307e-28], [22.102844, 7.310633e-28], [22.170829, 7.290325e-28], [22.239023, 7.271116e-28], [22.307427, 7.252823e-28], [22.376042, 7.235331e-28], [22.444867, 7.218543e-28], [22.513904, 7.202389e-28], [22.583154, 7.186826e-28], [22.652617, 7.171816e-28], [22.722293, 7.157332e-28], [22.792183, 7.143356e-28], [22.862289, 7.129877e-28], [22.93261, 7.116883e-28], [23.003147, 7.104371e-28], [23.073902, 7.092337e-28], [23.144874, 7.080781e-28], [23.216064, 7.069703e-28], [23.287474, 7.059106e-28], [23.359102, 7.048991e-28], [23.430952, 7.039363e-28], [23.503022, 7.030225e-28], [23.575314, 7.021582e-28], [23.647828, 7.013438e-28], [23.720566, 7.005799e-28], [23.793527, 6.998667e-28], [23.866712, 6.992049e-28], [23.940123, 6.985948e-28], [24.013759, 6.980368e-28], [24.087622, 6.975314e-28], [24.161712, 6.970789e-28], [24.23603, 6.966796e-28], [24.310577, 6.963338e-28], [24.385353, 6.960418e-28], [24.460359, 6.958036e-28], [24.535595, 6.956196e-28], [24.611063, 6.954896e-28], [24.686763, 6.954139e-28], [24.762696, 6.953923e-28], [24.838863, 6.954248e-28], [24.915264, 6.955112e-28], [24.9919, 6.956514e-28], [25.068771, 6.95845e-28], [25.145879, 6.960918e-28], [25.223224, 6.963914e-28], [25.300807, 6.967434e-28], [25.378629, 6.971471e-28], [25.45669, 6.976022e-28], [25.534991, 6.981079e-28], [25.613533, 6.986636e-28], [25.692317, 6.992686e-28], [25.771343, 6.99922e-28], [25.850612, 7.006231e-28], [25.930124, 7.013709e-28], [26.009882, 7.021645e-28], [26.089885, 7.030029e-28], [26.170133, 7.038851e-28], [26.250629, 7.048099e-28], [26.331372, 7.057764e-28], [26.412364, 7.067832e-28], [26.493604, 7.078292e-28], [26.575095, 7.089131e-28], [26.656836, 7.100338e-28], [26.738829, 7.111897e-28], [26.821074, 7.123796e-28], [26.903571, 7.136021e-28], [26.986323, 7.148558e-28], [27.069329, 7.161391e-28], [27.152591, 7.174507e-28], [27.236108, 7.18789e-28], [27.319882, 7.201524e-28], [27.403915, 7.215395e-28], [27.488205, 7.229485e-28], [27.572755, 7.243779e-28], [27.657565, 7.25826e-28], [27.742636, 7.272911e-28], [27.827968, 7.287715e-28], [27.913563, 7.302655e-28], [27.999421, 7.317713e-28], [28.085543, 7.332871e-28], [28.17193, 7.348112e-28], [28.258583, 7.363415e-28], [28.345503, 7.378765e-28], [28.432689, 7.39414e-28], [28.520144, 7.409523e-28], [28.607868, 7.424893e-28], [28.695862, 7.440233e-28], [28.784126, 7.455521e-28], [28.872662, 7.470739e-28], [28.96147, 7.485866e-28], [29.050552, 7.500883e-28], [29.139907, 7.515768e-28], [29.229537, 7.530503e-28], [29.319443, 7.545067e-28], [29.409626, 7.559439e-28], [29.500085, 7.5736e-28], [29.590824, 7.587529e-28], [29.681841, 7.601206e-28], [29.773138, 7.61461e-28], [29.864716, 7.627723e-28], [29.956575, 7.640524e-28], [30.048718, 7.652994e-28], [30.141143, 7.665113e-28], [30.233853, 7.676862e-28], [30.326848, 7.688223e-28], [30.420129, 7.699177e-28], [30.513697, 7.709706e-28], [30.607553, 7.719794e-28], [30.701697, 7.729421e-28], [30.796131, 7.738573e-28], [30.890856, 7.747234e-28], [30.985872, 7.755386e-28], [31.08118, 7.763017e-28], [31.176781, 7.770111e-28], [31.272677, 7.776656e-28], [31.368867, 7.782637e-28], [31.465353, 7.788044e-28], [31.562136, 7.792865e-28], [31.659217, 7.797089e-28], [31.756596, 7.800707e-28], [31.854275, 7.803709e-28], [31.952254, 7.806088e-28], [32.050535, 7.807836e-28], [32.149118, 7.808947e-28], [32.248004, 7.809416e-28], [32.347194, 7.809237e-28], [32.446689, 7.808407e-28], [32.546491, 7.806922e-28], [32.646599, 7.804782e-28], [32.747015, 7.801985e-28], [32.847741, 7.798529e-28], [32.948776, 7.794417e-28], [33.050121, 7.789649e-28], [33.151779, 7.784227e-28], [33.253749, 7.778155e-28], [33.356033, 7.771436e-28], [33.458631, 7.764074e-28], [33.561545, 7.756076e-28], [33.664776, 7.747446e-28], [33.768324, 7.738192e-28], [33.87219, 7.728321e-28], [33.976376, 7.71784e-28], [34.080883, 7.706759e-28], [34.185711, 7.695086e-28], [34.290861, 7.682832e-28], [34.396335, 7.670006e-28], [34.502133, 7.656619e-28], [34.608257, 7.642682e-28], [34.714707, 7.628206e-28], [34.821485, 7.613205e-28], [34.92859, 7.597689e-28], [35.036026, 7.581671e-28], [35.143792, 7.565165e-28], [35.251889, 7.548183e-28], [35.360319, 7.530739e-28], [35.469082, 7.512846e-28], [35.57818, 7.494517e-28], [35.687613, 7.475768e-28], [35.797383, 7.456611e-28], [35.907491, 7.43706e-28], [36.017937, 7.41713e-28], [36.128724, 7.396835e-28], [36.23985, 7.376188e-28], [36.351319, 7.355204e-28], [36.46313, 7.333897e-28], [36.575286, 7.312281e-28], [36.687786, 7.290369e-28], [36.800633, 7.268176e-28], [36.913826, 7.245714e-28], [37.027368, 7.222998e-28], [37.141259, 7.20004e-28], [37.2555, 7.176855e-28], [37.370093, 7.153454e-28], [37.485038, 7.129851e-28], [37.600336, 7.106059e-28], [37.71599, 7.082089e-28], [37.831999, 7.057954e-28], [37.948365, 7.033666e-28], [38.065088, 7.009236e-28], [38.182171, 6.984677e-28], [38.299614, 6.959998e-28], [38.417418, 6.935212e-28], [38.535585, 6.910329e-28], [38.654115, 6.885359e-28], [38.773009, 6.860313e-28], [38.89227, 6.8352e-28], [39.011897, 6.81003e-28], [39.131892, 6.784813e-28], [39.252256, 6.759557e-28], [39.37299, 6.734272e-28], [39.494096, 6.708966e-28], [39.615574, 6.683647e-28], [39.737426, 6.658324e-28], [39.859653, 6.633004e-28], [39.982255, 6.607695e-28], [40.105235, 6.582404e-28], [40.228593, 6.557139e-28], [40.352331, 6.531905e-28], [40.476449, 6.50671e-28], [40.600948, 6.48156e-28], [40.725831, 6.456461e-28], [40.851098, 6.431419e-28], [40.97675, 6.406439e-28], [41.102789, 6.381527e-28], [41.229215, 6.356688e-28], [41.35603, 6.331927e-28], [41.483236, 6.307249e-28], [41.610832, 6.282658e-28], [41.738821, 6.258159e-28], [41.867204, 6.233756e-28], [41.995982, 6.209453e-28], [42.125155, 6.185253e-28], [42.254726, 6.16116e-28], [42.384696, 6.137178e-28], [42.515065, 6.11331e-28], [42.645835, 6.089558e-28], [42.777008, 6.065927e-28], [42.908584, 6.042418e-28], [43.040565, 6.019034e-28], [43.172951, 5.995778e-28], [43.305745, 5.972652e-28], [43.438947, 5.949659e-28], [43.572559, 5.9268e-28], [43.706582, 5.904077e-28], [43.841018, 5.881492e-28], [43.975866, 5.859048e-28], [44.11113, 5.836745e-28], [44.24681, 5.814586e-28], [44.382906, 5.792572e-28], [44.519422, 5.770706e-28], [44.656357, 5.748989e-28], [44.793714, 5.727426e-28], [44.931493, 5.70602e-28], [45.069696, 5.684785e-28], [45.208324, 5.663789e-28], [45.347378, 5.643485e-28], [45.48686, 5.62363e-28], [45.626772, 5.603081e-28], [45.767113, 5.582178e-28], [45.907886, 5.561338e-28], [46.049092, 5.540696e-28], [46.190733, 5.520173e-28], [46.332809, 5.499967e-28], [46.475322, 5.480103e-28], [46.618273, 5.460444e-28], [46.761665, 5.440951e-28], [46.905497, 5.421617e-28], [47.049772, 5.402439e-28], [47.19449, 5.383414e-28], [47.339654, 5.364543e-28], [47.485264, 5.345824e-28], [47.631321, 5.327255e-28], [47.777829, 5.308837e-28], [47.924786, 5.290569e-28], [48.072196, 5.27245e-28], [48.22006, 5.25448e-28], [48.368378, 5.236658e-28], [48.517152, 5.218983e-28], [48.666384, 5.201455e-28], [48.816075, 5.184073e-28], [48.966226, 5.166836e-28], [49.116839, 5.149744e-28], [49.267916, 5.132795e-28], [49.419457, 5.115989e-28], [49.571464, 5.099326e-28], [49.723939, 5.082804e-28], [49.876882, 5.066422e-28], [50.030297, 5.05018e-28], [50.184183, 5.034076e-28], [50.338542, 5.01811e-28], [50.493376, 5.002281e-28], [50.648687, 4.986588e-28], [50.804475, 4.97103e-28], [50.960742, 4.955606e-28], [51.11749, 4.940315e-28], [51.27472, 4.925157e-28], [51.432434, 4.910129e-28], [51.590633, 4.895231e-28], [51.749319, 4.880462e-28], [51.908492, 4.865822e-28], [52.068155, 4.851308e-28], [52.228309, 4.83692e-28], [52.388956, 4.822657e-28], [52.550097, 4.808518e-28], [52.711734, 4.794502e-28], [52.873868, 4.780608e-28], [53.0365, 4.766834e-28], [53.199633, 4.75318e-28], [53.363268, 4.739645e-28], [53.527405, 4.726227e-28], [53.692048, 4.712926e-28], [53.857197, 4.69974e-28], [54.022854, 4.686668e-28], [54.189021, 4.67371e-28], [54.355699, 4.660864e-28], [54.522889, 4.648129e-28], [54.690594, 4.635504e-28], [54.858814, 4.622988e-28], [55.027552, 4.61058e-28], [55.196809, 4.598279e-28], [55.366587, 4.586084e-28], [55.536886, 4.573994e-28], [55.70771, 4.562007e-28], [55.879059, 4.550123e-28], [56.050935, 4.538341e-28], [56.22334, 4.52666e-28], [56.396275, 4.515078e-28], [56.569742, 4.503595e-28], [56.743742, 4.49221e-28], [56.918278, 4.480921e-28], [57.09335, 4.469728e-28], [57.268961, 4.458629e-28], [57.445113, 4.447624e-28], [57.621806, 4.436712e-28], [57.799042, 4.425891e-28], [57.976824, 4.415161e-28], [58.155152, 4.40452e-28], [58.334029, 4.393969e-28], [58.513456, 4.383505e-28], [58.693435, 4.373127e-28], [58.873968, 4.362836e-28], [59.055056, 4.35263e-28], [59.236701, 4.342507e-28], [59.418905, 4.332468e-28], [59.601669, 4.322511e-28], [59.784995, 4.312635e-28], [59.968885, 4.30284e-28], [60.153341, 4.293124e-28], [60.338364, 4.283486e-28], [60.523956, 4.273926e-28], [60.710119, 4.264444e-28], [60.896855, 4.255037e-28], [61.084165, 4.245705e-28], [61.272051, 4.236447e-28], [61.460516, 4.227263e-28], [61.649559, 4.218152e-28], [61.839185, 4.209112e-28], [62.029393, 4.200143e-28], [62.220187, 4.191245e-28], [62.411567, 4.182416e-28], [62.603536, 4.173655e-28], [62.796096, 4.164962e-28], [62.989248, 4.156337e-28], [63.182994, 4.147777e-28], [63.377336, 4.139283e-28], [63.572275, 4.130854e-28], [63.767815, 4.122488e-28], [63.963955, 4.114186e-28], [64.160699, 4.105947e-28], [64.358049, 4.097769e-28], [64.556005, 4.089652e-28], [64.75457, 4.081596e-28], [64.953746, 4.073599e-28], [65.153534, 4.065662e-28], [65.353937, 4.057782e-28], [65.554957, 4.049961e-28], [65.756594, 4.042196e-28], [65.958852, 4.034488e-28], [66.161732, 4.026835e-28], [66.365237, 4.019237e-28], [66.569367, 4.011694e-28], [66.774125, 4.004204e-28], [66.979512, 3.996768e-28], [67.185532, 3.989384e-28], [67.392185, 3.982052e-28], [67.599474, 3.974771e-28], [67.8074, 3.967541e-28], [68.015966, 3.960361e-28], [68.225173, 3.953231e-28], [68.435024, 3.946149e-28], [68.645521, 3.939116e-28], [68.856665, 3.932131e-28], [69.068458, 3.925193e-28], [69.280903, 3.918302e-28], [69.494001, 3.911456e-28], [69.707755, 3.904657e-28], [69.922166, 3.897903e-28], [70.137236, 3.891193e-28], [70.352969, 3.884528e-28], [70.569364, 3.877906e-28], [70.786426, 3.871327e-28], [71.004155, 3.86479e-28], [71.222553, 3.858296e-28], [71.441624, 3.851844e-28], [71.661368, 3.845432e-28], [71.881788, 3.839062e-28], [72.102886, 3.832731e-28], [72.324665, 3.82644e-28], [72.547125, 3.820189e-28], [72.77027, 3.813976e-28], [72.994101, 3.807802e-28], [73.21862, 3.801666e-28], [73.44383, 3.795567e-28], [73.669733, 3.789506e-28], [73.896331, 3.783481e-28], [74.123625, 3.777493e-28], [74.351619, 3.77154e-28], [74.580314, 3.765623e-28], [74.809713, 3.759741e-28], [75.039817, 3.753894e-28], [75.270628, 3.748081e-28], [75.50215, 3.742302e-28], [75.734384, 3.736557e-28], [75.967332, 3.730845e-28], [76.200997, 3.725166e-28], [76.43538, 3.719519e-28], [76.670485, 3.713904e-28], [76.906312, 3.708321e-28], [77.142865, 3.70277e-28], [77.380146, 3.697249e-28], [77.618156, 3.69176e-28], [77.856898, 3.686301e-28], [78.096375, 3.680871e-28], [78.336589, 3.675472e-28], [78.577541, 3.670102e-28], [78.819234, 3.664761e-28], [79.061671, 3.659449e-28], [79.304853, 3.654166e-28], [79.548784, 3.648911e-28], [79.793465, 3.643683e-28], [80.038898, 3.638484e-28], [80.285086, 3.633312e-28], [80.532032, 3.628166e-28], [80.779737, 3.623048e-28], [81.028204, 3.617956e-28], [81.277435, 3.61289e-28], [81.527433, 3.60785e-28], [81.7782, 3.602836e-28], [82.029738, 3.597848e-28], [82.28205, 3.592884e-28], [82.535138, 3.587946e-28], [82.789004, 3.583032e-28], [83.043651, 3.578143e-28], [83.299082, 3.573277e-28], [83.555298, 3.568436e-28], [83.812302, 3.563619e-28], [84.070097, 3.558825e-28], [84.328684, 3.554054e-28], [84.588067, 3.549306e-28], [84.848248, 3.544581e-28], [85.10923, 3.539878e-28], [85.371013, 3.535198e-28], [85.633603, 3.530541e-28], [85.896999, 3.525905e-28], [86.161206, 3.52129e-28], [86.426226, 3.516698e-28], [86.692061, 3.512126e-28], [86.958713, 3.507576e-28], [87.226186, 3.503047e-28], [87.494481, 3.498538e-28], [87.763602, 3.49405e-28], [88.03355, 3.489582e-28], [88.304329, 3.485135e-28], [88.57594, 3.480707e-28], [88.848387, 3.4763e-28], [89.121672, 3.471912e-28], [89.395798, 3.467543e-28], [89.670767, 3.463194e-28], [89.946581, 3.458864e-28], [90.223244, 3.454553e-28], [90.500758, 3.45026e-28], [90.779126, 3.445987e-28], [91.058349, 3.441732e-28], [91.338432, 3.437495e-28], [91.619376, 3.433276e-28], [91.901184, 3.429075e-28], [92.183859, 3.424892e-28], [92.467404, 3.420727e-28], [92.75182, 3.41658e-28], [93.037112, 3.41245e-28], [93.32328, 3.408337e-28], [93.61033, 3.404241e-28], [93.898262, 3.400163e-28], [94.187079, 3.396101e-28], [94.476786, 3.392056e-28], [94.767383, 3.388028e-28], [95.058874, 3.384016e-28], [95.351261, 3.38002e-28], [95.644548, 3.376041e-28], [95.938737, 3.372078e-28], [96.233831, 3.368131e-28], [96.529833, 3.3642e-28], [96.826745, 3.360284e-28], [97.12457, 3.356384e-28], [97.423312, 3.3525e-28], [97.722972, 3.348631e-28], [98.023554, 3.344778e-28], [98.32506, 3.34094e-28], [98.627494, 3.337116e-28], [98.930858, 3.333308e-28], [99.235156, 3.329515e-28], [99.540389, 3.325736e-28], [99.846561, 3.321973e-28], [100.15367, 3.318223e-28], [100.46173, 3.314489e-28], [100.77074, 3.310768e-28], [101.0807, 3.307063e-28], [101.39161, 3.303371e-28], [101.70347, 3.299693e-28], [102.0163, 3.29603e-28], [102.33008, 3.29238e-28], [102.64484, 3.288744e-28], [102.96056, 3.285122e-28], [103.27725, 3.281514e-28], [103.59492, 3.277919e-28], [103.91356, 3.274338e-28], [104.23318, 3.27077e-28], [104.55379, 3.267216e-28], [104.87538, 3.263675e-28], [105.19796, 3.260147e-28], [105.52154, 3.256633e-28], [105.84611, 3.253131e-28], [106.17168, 3.249643e-28], [106.49824, 3.246167e-28], [106.82582, 3.242704e-28], [107.1544, 3.239254e-28], [107.48399, 3.235817e-28], [107.8146, 3.232392e-28], [108.14622, 3.22898e-28], [108.47886, 3.225581e-28], [108.81253, 3.222194e-28], [109.14722, 3.218819e-28], [109.48294, 3.215457e-28], [109.81969, 3.212107e-28], [110.15748, 3.208769e-28], [110.49631, 3.205444e-28], [110.83618, 3.20213e-28], [111.1771, 3.198829e-28], [111.51906, 3.19554e-28], [111.86208, 3.192262e-28], [112.20615, 3.188997e-28], [112.55128, 3.185743e-28], [112.89747, 3.182501e-28], [113.24473, 3.179271e-28], [113.59306, 3.176052e-28], [113.94245, 3.172846e-28], [114.29292, 3.16965e-28], [114.64447, 3.166467e-28], [114.9971, 3.163294e-28], [115.35082, 3.160134e-28], [115.70562, 3.156984e-28], [116.06151, 3.153846e-28], [116.4185, 3.15072e-28], [116.77659, 3.147605e-28], [117.13578, 3.1445e-28], [117.49607, 3.141408e-28], [117.85747, 3.138326e-28], [118.21998, 3.135255e-28], [118.58361, 3.132196e-28], [118.94836, 3.129147e-28], [119.31423, 3.12611e-28], [119.68122, 3.123084e-28], [120.04934, 3.120068e-28], [120.4186, 3.117064e-28], [120.78899, 3.11407e-28], [121.16052, 3.111087e-28], [121.53319, 3.108115e-28], [121.90701, 3.105154e-28], [122.28198, 3.102203e-28], [122.6581, 3.099264e-28], [123.03538, 3.096334e-28], [123.41382, 3.093416e-28], [123.79342, 3.090508e-28], [124.17419, 3.087611e-28], [124.55613, 3.084725e-28], [124.93925, 3.081849e-28], [125.32355, 3.078983e-28], [125.70902, 3.076128e-28], [126.09569, 3.073284e-28], [126.48354, 3.07045e-28], [126.87259, 3.067626e-28], [127.26283, 3.064813e-28], [127.65427, 3.06201e-28], [128.04692, 3.059218e-28], [128.44077, 3.056436e-28], [128.83584, 3.053664e-28], [129.23212, 3.050903e-28], [129.62962, 3.048151e-28], [130.02834, 3.045411e-28], [130.42829, 3.04268e-28], [130.82947, 3.039959e-28], [131.23188, 3.037249e-28], [131.63553, 3.034549e-28], [132.04042, 3.03186e-28], [132.44656, 3.02918e-28], [132.85395, 3.02651e-28], [133.26259, 3.023851e-28], [133.67248, 3.021202e-28], [134.08364, 3.018563e-28], [134.49606, 3.015934e-28], [134.90975, 3.013315e-28], [135.32472, 3.010706e-28], [135.74096, 3.008107e-28], [136.15848, 3.005518e-28], [136.57728, 3.00294e-28], [136.99737, 3.000371e-28], [137.41876, 2.997812e-28], [137.84144, 2.995264e-28], [138.26542, 2.992725e-28], [138.6907, 2.990196e-28], [139.1173, 2.987678e-28], [139.5452, 2.985169e-28], [139.97442, 2.98267e-28], [140.40496, 2.980181e-28], [140.83683, 2.977703e-28], [141.27002, 2.975234e-28], [141.70455, 2.972775e-28], [142.14041, 2.970326e-28], [142.57762, 2.967886e-28], [143.01617, 2.965457e-28], [143.45606, 2.963038e-28], [143.89731, 2.960628e-28], [144.33992, 2.958229e-28], [144.78389, 2.955839e-28], [145.22922, 2.95346e-28], [145.67593, 2.95109e-28], [146.12401, 2.94873e-28], [146.57346, 2.94638e-28], [147.0243, 2.94404e-28], [147.47653, 2.94171e-28], [147.93015, 2.939389e-28], [148.38516, 2.937079e-28], [148.84157, 2.934779e-28], [149.29939, 2.932488e-28], [149.75861, 2.930207e-28], [150.21925, 2.927936e-28], [150.6813, 2.925675e-28], [151.14477, 2.923424e-28], [151.60967, 2.921183e-28], [152.076, 2.918952e-28], [152.54377, 2.916731e-28], [153.01297, 2.914519e-28], [153.48362, 2.912317e-28], [153.95571, 2.910126e-28], [154.42926, 2.907944e-28], [154.90426, 2.905772e-28], [155.38072, 2.90361e-28], [155.85865, 2.901459e-28], [156.33805, 2.899317e-28], [156.81892, 2.897185e-28], [157.30128, 2.895062e-28], [157.78511, 2.89295e-28], [158.27044, 2.890848e-28], [158.75725, 2.888756e-28], [159.24557, 2.886674e-28], [159.73539, 2.884601e-28], [160.22671, 2.882539e-28], [160.71954, 2.880487e-28], [161.21389, 2.878445e-28], [161.70976, 2.876412e-28], [162.20716, 2.87439e-28], [162.70609, 2.872378e-28], [163.20655, 2.870376e-28], [163.70855, 2.868384e-28], [164.21209, 2.866402e-28], [164.71719, 2.86443e-28], [165.22383, 2.862468e-28], [165.73204, 2.860516e-28], [166.24181, 2.858574e-28], [166.75314, 2.856643e-28], [167.26605, 2.854721e-28], [167.78054, 2.85281e-28], [168.29661, 2.850909e-28], [168.81426, 2.849018e-28], [169.33351, 2.847137e-28], [169.85436, 2.845266e-28], [170.37681, 2.843406e-28], [170.90086, 2.841556e-28], [171.42653, 2.839716e-28], [171.95381, 2.837886e-28], [172.48272, 2.836067e-28], [173.01325, 2.834258e-28], [173.54541, 2.832459e-28], [174.07921, 2.83067e-28], [174.61466, 2.828892e-28], [175.15175, 2.827124e-28], [175.69049, 2.825367e-28], [176.23089, 2.823619e-28], [176.77295, 2.821883e-28], [177.31668, 2.820156e-28], [177.86208, 2.81844e-28], [178.40916, 2.816735e-28], [178.95792, 2.81504e-28], [179.50837, 2.813355e-28], [180.06051, 2.811681e-28], [180.61435, 2.810018e-28], [181.16989, 2.808365e-28], [181.72715, 2.806722e-28], [182.28611, 2.80509e-28], [182.8468, 2.803469e-28], [183.40921, 2.801858e-28], [183.97335, 2.800258e-28], [184.53923, 2.798669e-28], [185.10684, 2.79709e-28], [185.67621, 2.795522e-28], [186.24732, 2.793965e-28], [186.82019, 2.792418e-28], [187.39482, 2.790882e-28], [187.97122, 2.789357e-28], [188.54939, 2.787843e-28], [189.12935, 2.78634e-28], [189.71108, 2.784847e-28], [190.2946, 2.783365e-28], [190.87992, 2.781894e-28], [191.46704, 2.780434e-28], [192.05597, 2.778985e-28], [192.64671, 2.777547e-28], [193.23926, 2.776121e-28], [193.83364, 2.774704e-28], [194.42984, 2.7733e-28], [195.02788, 2.771906e-28], [195.62776, 2.770523e-28], [196.22948, 2.769151e-28], [196.83305, 2.767791e-28], [197.43849, 2.766441e-28], [198.04578, 2.765103e-28], [198.65494, 2.763776e-28], [199.26597, 2.76246e-28], [199.87889, 2.761156e-28], [200.49369, 2.759863e-28], [201.11038, 2.758581e-28], [201.72896, 2.75731e-28], [202.34945, 2.756051e-28], [202.97185, 2.754803e-28], [203.59616, 2.753567e-28], [204.2224, 2.752342e-28], [204.85056, 2.751129e-28], [205.48065, 2.749927e-28], [206.11268, 2.748737e-28], [206.74665, 2.747558e-28], [207.38257, 2.746391e-28], [208.02045, 2.745236e-28], [208.66029, 2.744092e-28], [209.3021, 2.74296e-28], [209.94589, 2.74184e-28], [210.59165, 2.740731e-28], [211.2394, 2.739635e-28], [211.88914, 2.73855e-28], [212.54088, 2.737477e-28], [213.19463, 2.736416e-28], [213.85039, 2.735366e-28], [214.50816, 2.734329e-28], [215.16796, 2.733304e-28], [215.82978, 2.732291e-28], [216.49364, 2.731289e-28], [217.15955, 2.7303e-28], [217.8275, 2.729323e-28], [218.4975, 2.728358e-28], [219.16957, 2.727406e-28], [219.84371, 2.726465e-28], [220.51991, 2.725537e-28], [221.1982, 2.724621e-28], [221.87858, 2.723718e-28], [222.56104, 2.722827e-28], [223.24561, 2.721948e-28], [223.93228, 2.721082e-28], [224.62107, 2.720228e-28], [225.31197, 2.719386e-28], [226.005, 2.718558e-28], [226.70016, 2.717741e-28], [227.39745, 2.716938e-28], [228.09689, 2.716147e-28], [228.79849, 2.715369e-28], [229.50224, 2.714603e-28], [230.20816, 2.71385e-28], [230.91624, 2.71311e-28], [231.62651, 2.712383e-28], [232.33896, 2.711669e-28], [233.0536, 2.710968e-28], [233.77044, 2.710279e-28], [234.48949, 2.709604e-28], [235.21074, 2.708942e-28], [235.93422, 2.708292e-28], [236.65992, 2.707656e-28], [237.38785, 2.707033e-28], [238.11802, 2.706424e-28], [238.85044, 2.705827e-28], [239.58511, 2.705244e-28], [240.32204, 2.704674e-28], [241.06123, 2.704117e-28], [241.8027, 2.703574e-28], [242.54646, 2.703045e-28], [243.29249, 2.702528e-28], [244.04083, 2.702026e-28], [244.79146, 2.701537e-28], [245.54441, 2.701061e-28], [246.29967, 2.700599e-28], [247.05725, 2.700151e-28], [247.81716, 2.699717e-28], [248.57941, 2.699296e-28], [249.34401, 2.69889e-28], [250.11095, 2.698497e-28], [250.88026, 2.698118e-28], [251.65193, 2.697753e-28], [252.42598, 2.697402e-28], [253.2024, 2.697065e-28], [253.98122, 2.696742e-28], [254.76243, 2.696433e-28], [255.54604, 2.696139e-28], [256.33206, 2.695859e-28], [257.1205, 2.695593e-28], [257.91137, 2.695341e-28], [258.70467, 2.695104e-28], [259.50041, 2.694881e-28], [260.29859, 2.694673e-28], [261.09924, 2.694479e-28], [261.90234, 2.6943e-28], [262.70792, 2.694135e-28], [263.51597, 2.693985e-28], [264.3265, 2.69385e-28], [265.13954, 2.693729e-28], [265.95507, 2.693623e-28], [266.77311, 2.693532e-28], [267.59366, 2.693457e-28], [268.41674, 2.693395e-28], [269.24236, 2.693349e-28], [270.07051, 2.693318e-28], [270.90121, 2.693303e-28], [271.73446, 2.693302e-28], [272.57028, 2.693316e-28], [273.40866, 2.693346e-28], [274.24963, 2.693391e-28], [275.09318, 2.693451e-28], [275.93933, 2.693527e-28], [276.78808, 2.693618e-28], [277.63944, 2.693725e-28], [278.49342, 2.693847e-28], [279.35003, 2.693985e-28], [280.20927, 2.694138e-28], [281.07115, 2.694308e-28], [281.93569, 2.694493e-28], [282.80288, 2.694693e-28], [283.67274, 2.69491e-28], [284.54528, 2.695143e-28], [285.4205, 2.695391e-28], [286.29841, 2.695656e-28], [287.17903, 2.695936e-28], [288.06235, 2.696233e-28], [288.94839, 2.696546e-28], [289.83715, 2.696875e-28], [290.72865, 2.697221e-28], [291.62289, 2.697583e-28], [292.51988, 2.697961e-28], [293.41963, 2.698356e-28], [294.32215, 2.698767e-28], [295.22744, 2.699195e-28], [296.13552, 2.699639e-28], [297.04639, 2.7001e-28], [297.96006, 2.700578e-28], [298.87654, 2.701073e-28], [299.79584, 2.701584e-28], [300.71797, 2.702113e-28], [301.64294, 2.702658e-28], [302.57075, 2.703221e-28], [303.50142, 2.7038e-28], [304.43494, 2.704397e-28], [305.37134, 2.70501e-28], [306.31062, 2.705642e-28], [307.25279, 2.70629e-28], [308.19785, 2.706956e-28], [309.14583, 2.707639e-28], [310.09672, 2.708339e-28], [311.05053, 2.709058e-28], [312.00728, 2.709793e-28], [312.96697, 2.710547e-28], [313.92961, 2.711318e-28], [314.89521, 2.712107e-28], [315.86378, 2.712914e-28], [316.83534, 2.713739e-28], [317.80988, 2.714581e-28], [318.78741, 2.715442e-28], [319.76796, 2.716321e-28], [320.75152, 2.717218e-28], [321.73811, 2.718133e-28], [322.72773, 2.719067e-28], [323.72039, 2.720018e-28], [324.71611, 2.720989e-28], [325.71489, 2.721977e-28], [326.71674, 2.722984e-28], [327.72168, 2.72401e-28], [328.7297, 2.725055e-28], [329.74083, 2.726118e-28], [330.75506, 2.7272e-28], [331.77242, 2.728301e-28], [332.7929, 2.72942e-28], [333.81653, 2.730559e-28], [334.8433, 2.731717e-28], [335.87323, 2.732894e-28], [336.90633, 2.73409e-28], [337.94261, 2.735305e-28], [338.98207, 2.73654e-28], [340.02473, 2.737794e-28], [341.0706, 2.739067e-28], [342.11968, 2.74036e-28], [343.17199, 2.741673e-28], [344.22754, 2.743005e-28], [345.28634, 2.744357e-28], [346.34839, 2.745729e-28], [347.41371, 2.747121e-28], [348.4823, 2.748533e-28], [349.55418, 2.749965e-28], [350.62936, 2.751417e-28], [351.70785, 2.752889e-28], [352.78965, 2.754382e-28], [353.87478, 2.755895e-28], [354.96325, 2.757428e-28], [356.05507, 2.758982e-28], [357.15024, 2.760557e-28], [358.24879, 2.762152e-28], [359.35071, 2.763769e-28], [360.45602, 2.765406e-28], [361.56473, 2.767064e-28], [362.67685, 2.768744e-28], [363.7924, 2.770445e-28], [364.91137, 2.772167e-28], [366.03379, 2.773911e-28], [367.15965, 2.775676e-28], [368.28898, 2.777463e-28], [369.42179, 2.779272e-28], [370.55808, 2.781104e-28], [371.69786, 2.782957e-28], [372.84115, 2.784833e-28], [373.98796, 2.786732e-28], [375.13829, 2.788654e-28], [376.29216, 2.790599e-28], [377.44959, 2.792567e-28], [378.61057, 2.794559e-28], [379.77512, 2.796574e-28], [380.94325, 2.798615e-28], [382.11498, 2.80068e-28], [383.29031, 2.80277e-28], [384.46926, 2.804886e-28], [385.65183, 2.807029e-28], [386.83804, 2.809199e-28], [388.0279, 2.811397e-28], [389.22142, 2.813624e-28], [390.41861, 2.815881e-28], [391.61948, 2.81817e-28], [392.82405, 2.820493e-28], [394.03232, 2.822851e-28], [395.24431, 2.825248e-28], [396.46002, 2.827688e-28], [397.67948, 2.830173e-28], [398.90268, 2.832711e-28], [400.12965, 2.835308e-28], [401.36039, 2.837976e-28], [402.59492, 2.840726e-28], [403.83324, 2.843579e-28], [405.07538, 2.846562e-28], [406.32133, 2.849715e-28], [407.57112, 2.853098e-28], [408.82475, 2.856808e-28], [410.08224, 2.861003e-28], [411.34359, 2.865962e-28], [412.60883, 2.872214e-28], [413.87795, 2.880861e-28], [415.15098, 2.894533e-28], [416.42793, 2.920867e-28], [417.7088, 2.990483e-28], [418.99361, 3.340482e-28], [420.28238, 4.970461e-27], [421.57511, 3.279467e-28], [422.87181, 2.986934e-28], [424.17251, 2.930246e-28], [425.4772, 2.911874e-28], [426.78591, 2.904877e-28], [428.09864, 2.902411e-28], [429.41541, 2.902104e-28], [430.73623, 2.902958e-28], [432.06111, 2.904496e-28], [433.39007, 2.906467e-28], [434.72312, 2.908729e-28], [436.06027, 2.911196e-28], [437.40153, 2.913815e-28], [438.74691, 2.916549e-28], [440.09643, 2.919375e-28], [441.45011, 2.922278e-28], [442.80795, 2.925244e-28], [444.16996, 2.928266e-28], [445.53616, 2.931337e-28], [446.90657, 2.934452e-28], [448.28119, 2.937608e-28], [449.66004, 2.940803e-28], [451.04313, 2.944033e-28], [452.43047, 2.947297e-28], [453.82209, 2.950594e-28], [455.21798, 2.953923e-28], [456.61816, 2.957283e-28], [458.02266, 2.960673e-28], [459.43147, 2.964093e-28], [460.84462, 2.967542e-28], [462.26211, 2.971019e-28], [463.68396, 2.974525e-28], [465.11019, 2.978059e-28], [466.5408, 2.981622e-28], [467.97581, 2.985212e-28], [469.41524, 2.98883e-28], [470.8591, 2.992476e-28], [472.30739, 2.996149e-28], [473.76014, 2.99985e-28], [475.21736, 3.003578e-28], [476.67906, 3.007333e-28], [478.14526, 3.011117e-28], [479.61597, 3.014927e-28], [481.0912, 3.018765e-28], [482.57096, 3.02263e-28], [484.05528, 3.026523e-28], [485.54417, 3.030444e-28], [487.03763, 3.034391e-28], [488.53569, 3.038367e-28], [490.03836, 3.04237e-28], [491.54565, 3.0464e-28], [493.05757, 3.050459e-28], [494.57415, 3.054545e-28], [496.09539, 3.058658e-28], [497.62131, 3.0628e-28], [499.15192, 3.066969e-28], [500.68724, 3.071166e-28], [502.22728, 3.075392e-28], [503.77206, 3.079645e-28], [505.32159, 3.083926e-28], [506.87589, 3.088236e-28], [508.43497, 3.092574e-28], [509.99884, 3.09694e-28], [511.56753, 3.101334e-28], [513.14103, 3.105757e-28], [514.71938, 3.110208e-28], [516.30259, 3.114688e-28], [517.89066, 3.119196e-28], [519.48362, 3.123733e-28], [521.08148, 3.128299e-28], [522.68425, 3.132893e-28], [524.29195, 3.137517e-28], [525.9046, 3.142169e-28], [527.52221, 3.14685e-28], [529.14479, 3.151561e-28], [530.77236, 3.1563e-28], [532.40495, 3.161069e-28], [534.04255, 3.165867e-28], [535.68519, 3.170694e-28], [537.33288, 3.175551e-28], [538.98564, 3.180437e-28], [540.64348, 3.185353e-28], [542.30642, 3.190298e-28], [543.97448, 3.195273e-28], [545.64767, 3.200278e-28], [547.326, 3.205313e-28], [549.0095, 3.210377e-28], [550.69818, 3.215472e-28], [552.39205, 3.220596e-28], [554.09112, 3.225751e-28], [555.79543, 3.230936e-28], [557.50498, 3.236151e-28], [559.21978, 3.241397e-28], [560.93986, 3.246673e-28], [562.66524, 3.251979e-28], [564.39591, 3.257316e-28], [566.13192, 3.262683e-28], [567.87326, 3.268082e-28], [569.61995, 3.273511e-28], [571.37202, 3.278971e-28], [573.12948, 3.284461e-28], [574.89235, 3.289983e-28], [576.66064, 3.295536e-28], [578.43436, 3.30112e-28], [580.21354, 3.306735e-28], [581.9982, 3.312382e-28], [583.78834, 3.31806e-28], [585.58399, 3.323769e-28], [587.38517, 3.32951e-28], [589.19188, 3.335282e-28], [591.00415, 3.341086e-28], [592.82199, 3.346922e-28], [594.64543, 3.35279e-28], [596.47448, 3.358689e-28], [598.30915, 3.364621e-28], [600.14946, 3.370584e-28], [601.99543, 3.37658e-28], [603.84709, 3.382608e-28], [605.70443, 3.388668e-28], [607.5675, 3.39476e-28], [609.43629, 3.400885e-28], [611.31083, 3.407042e-28], [613.19113, 3.413232e-28], [615.07722, 3.419455e-28], [616.96911, 3.42571e-28], [618.86682, 3.431998e-28], [620.77037, 3.438319e-28], [622.67977, 3.444673e-28], [624.59504, 3.45106e-28], [626.51621, 3.45748e-28], [628.44328, 3.463933e-28], [630.37629, 3.470419e-28], [632.31523, 3.476939e-28], [634.26014, 3.483492e-28], [636.21104, 3.490079e-28], [638.16793, 3.496699e-28], [640.13085, 3.503353e-28], [642.0998, 3.510041e-28], [644.07481, 3.516762e-28], [646.05589, 3.523517e-28], [648.04307, 3.530306e-28], [650.03635, 3.53713e-28], [652.03577, 3.543987e-28], [654.04134, 3.550879e-28], [656.05308, 3.557805e-28], [658.07101, 3.564765e-28], [660.09514, 3.571759e-28], [662.1255, 3.578789e-28], [664.1621, 3.585852e-28], [666.20497, 3.592951e-28], [668.25412, 3.600084e-28], [670.30958, 3.607252e-28], [672.37136, 3.614455e-28], [674.43947, 3.621693e-28], [676.51395, 3.628966e-28], [678.59481, 3.636274e-28], [680.68208, 3.643617e-28], [682.77576, 3.650996e-28], [684.87588, 3.65841e-28], [686.98246, 3.66586e-28], [689.09552, 3.673345e-28], [691.21508, 3.680866e-28], [693.34116, 3.688422e-28], [695.47378, 3.696015e-28], [697.61296, 3.703643e-28], [699.75872, 3.711307e-28], [701.91107, 3.719007e-28], [704.07005, 3.726744e-28], [706.23567, 3.734516e-28], [708.40795, 3.742325e-28], [710.58691, 3.750171e-28], [712.77258, 3.758053e-28], [714.96497, 3.765971e-28], [717.1641, 3.773926e-28], [719.36999, 3.781918e-28], [721.58267, 3.789946e-28], [723.80216, 3.798012e-28], [726.02847, 3.806114e-28], [728.26163, 3.814254e-28], [730.50166, 3.82243e-28], [732.74858, 3.830644e-28], [735.00241, 3.838896e-28], [737.26317, 3.847184e-28], [739.53089, 3.855511e-28], [741.80558, 3.863874e-28], [744.08727, 3.872276e-28], [746.37598, 3.880715e-28], [748.67172, 3.889192e-28], [750.97453, 3.897707e-28], [753.28442, 3.90626e-28], [755.60142, 3.914852e-28], [757.92554, 3.923481e-28], [760.25681, 3.932149e-28], [762.59525, 3.940855e-28], [764.94089, 3.9496e-28], [767.29374, 3.958383e-28], [769.65382, 3.967205e-28], [772.02117, 3.976066e-28], [774.3958, 3.984965e-28], [776.77773, 3.993904e-28], [779.16699, 4.002882e-28], [781.56359, 4.011899e-28], [783.96757, 4.020955e-28], [786.37895, 4.030051e-28], [788.79774, 4.039186e-28], [791.22397, 4.048361e-28], [793.65766, 4.057575e-28], [796.09884, 4.06683e-28], [798.54752, 4.076125e-28], [801.00374, 4.08546e-28], [803.46752, 4.094835e-28], [805.93887, 4.104251e-28], [808.41782, 4.113709e-28], [810.9044, 4.123208e-28], [813.39863, 4.13275e-28], [815.90052, 4.142334e-28], [818.41012, 4.151964e-28], [820.92743, 4.16164e-28], [823.45249, 4.171368e-28], [825.98531, 4.181156e-28], [828.52592, 4.191021e-28], [831.07435, 4.201007e-28], [833.63062, 4.211257e-28], [836.19475, 4.222444e-28], [838.76676, 4.243432e-28], [841.34669, 4.290291e-28], [843.93456, 4.253396e-28], [846.53038, 4.261331e-28], [849.13419, 4.270914e-28], [851.746, 4.280859e-28], [854.36585, 4.290949e-28], [856.99376, 4.301125e-28], [859.62975, 4.311365e-28], [862.27385, 4.321659e-28], [864.92608, 4.332002e-28], [867.58647, 4.342391e-28], [870.25504, 4.352826e-28], [872.93182, 4.363306e-28], [875.61684, 4.373829e-28], [878.31011, 4.384396e-28], [881.01167, 4.395007e-28], [883.72154, 4.405661e-28], [886.43974, 4.416359e-28], [889.1663, 4.4271e-28], [891.90125, 4.437885e-28], [894.64461, 4.448713e-28], [897.39641, 4.459585e-28], [900.15667, 4.470501e-28], [902.92543, 4.48146e-28], [905.7027, 4.492464e-28], [908.48851, 4.503511e-28], [911.28289, 4.514602e-28], [914.08587, 4.525738e-28], [916.89747, 4.536917e-28], [919.71771, 4.548141e-28], [922.54664, 4.55941e-28], [925.38426, 4.570723e-28], [928.23061, 4.58208e-28], [931.08571, 4.593483e-28], [933.9496, 4.60493e-28], [936.8223, 4.616422e-28], [939.70383, 4.627959e-28], [942.59423, 4.639542e-28], [945.49351, 4.65117e-28], [948.40172, 4.662843e-28], [951.31886, 4.674561e-28], [954.24499, 4.686325e-28], [957.18011, 4.698135e-28], [960.12426, 4.709991e-28], [963.07746, 4.721893e-28], [966.03975, 4.733841e-28], [969.01115, 4.745835e-28], [971.99169, 4.757875e-28], [974.9814, 4.769961e-28], [977.98031, 4.782095e-28], [980.98844, 4.794274e-28], [984.00582, 4.806501e-28], [987.03248, 4.818774e-28], [990.06845, 4.831094e-28], [993.11376, 4.843462e-28], [996.16844, 4.855876e-28], [999.23251, 4.868338e-28], [1002.306, 4.880847e-28], [1005.389, 4.893404e-28], [1008.4814, 4.906008e-28], [1011.5833, 4.918661e-28], [1014.6948, 4.931361e-28], [1017.8159, 4.944109e-28], [1020.9465, 4.956905e-28], [1024.0868, 4.96975e-28], [1027.2368, 4.982643e-28], [1030.3964, 4.995584e-28], [1033.5658, 5.008574e-28], [1036.7449, 5.021613e-28], [1039.9337, 5.034701e-28], [1043.1324, 5.047837e-28], [1046.341, 5.061023e-28], [1049.5594, 5.074258e-28], [1052.7876, 5.087542e-28], [1056.0259, 5.100876e-28], [1059.2741, 5.114259e-28], [1062.5322, 5.127692e-28], [1065.8004, 5.141175e-28], [1069.0787, 5.154708e-28], [1072.367, 5.168291e-28], [1075.6655, 5.181925e-28], [1078.9741, 5.195608e-28], [1082.2928, 5.209342e-28], [1085.6218, 5.223127e-28], [1088.961, 5.236963e-28], [1092.3105, 5.250849e-28], [1095.6703, 5.264786e-28], [1099.0404, 5.278775e-28], [1102.4209, 5.292814e-28], [1105.8118, 5.306906e-28], [1109.2131, 5.321048e-28], [1112.6249, 5.335242e-28], [1116.0472, 5.349489e-28], [1119.48, 5.363786e-28], [1122.9234, 5.378137e-28], [1126.3773, 5.392539e-28], [1129.8419, 5.406993e-28], [1133.3171, 5.4215e-28], [1136.803, 5.43606e-28], [1140.2997, 5.450672e-28], [1143.8071, 5.465338e-28], [1147.3253, 5.480056e-28], [1150.8543, 5.494827e-28], [1154.3942, 5.509652e-28], [1157.9449, 5.52453e-28], [1161.5066, 5.539461e-28], [1165.0792, 5.554446e-28], [1168.6628, 5.569485e-28], [1172.2575, 5.584578e-28], [1175.8632, 5.599726e-28], [1179.48, 5.614927e-28], [1183.1079, 5.630183e-28], [1186.7469, 5.645493e-28], [1190.3972, 5.660858e-28], [1194.0587, 5.676278e-28], [1197.7314, 5.691753e-28], [1201.4155, 5.707283e-28], [1205.1109, 5.722868e-28], [1208.8176, 5.738509e-28], [1212.5358, 5.754205e-28], [1216.2654, 5.769957e-28], [1220.0064, 5.785766e-28], [1223.759, 5.80163e-28], [1227.5231, 5.81755e-28], [1231.2988, 5.833528e-28], [1235.0861, 5.849562e-28], [1238.885, 5.865655e-28], [1242.6957, 5.881807e-28], [1246.518, 5.898021e-28], [1250.3521, 5.914307e-28], [1254.1981, 5.930712e-28], [1258.0558, 5.947811e-28], [1261.9254, 5.967517e-28], [1265.8069, 5.98009e-28], [1269.7003, 5.996479e-28], [1273.6058, 6.013046e-28], [1277.5232, 6.029694e-28], [1281.4527, 6.046407e-28], [1285.3942, 6.063183e-28], [1289.3479, 6.080018e-28], [1293.3138, 6.096913e-28], [1297.2918, 6.113868e-28], [1301.2821, 6.130881e-28], [1305.2847, 6.147955e-28], [1309.2995, 6.165087e-28], [1313.3268, 6.182279e-28], [1317.3664, 6.199531e-28], [1321.4184, 6.216843e-28], [1325.4829, 6.234215e-28], [1329.5599, 6.251647e-28], [1333.6494, 6.269139e-28], [1337.7515, 6.286692e-28], [1341.8663, 6.304306e-28], [1345.9937, 6.32198e-28], [1350.1337, 6.339715e-28], [1354.2866, 6.357512e-28], [1358.4522, 6.375369e-28], [1362.6306, 6.393288e-28], [1366.8218, 6.411269e-28], [1371.026, 6.429312e-28], [1375.2431, 6.447416e-28], [1379.4731, 6.465583e-28], [1383.7162, 6.483812e-28], [1387.9723, 6.502103e-28], [1392.2415, 6.520457e-28], [1396.5238, 6.538874e-28], [1400.8193, 6.557354e-28], [1405.1281, 6.575897e-28], [1409.45, 6.594503e-28], [1413.7853, 6.613173e-28], [1418.1339, 6.631906e-28], [1422.4959, 6.650703e-28], [1426.8713, 6.669565e-28], [1431.2601, 6.68849e-28], [1435.6625, 6.70748e-28], [1440.0784, 6.726534e-28], [1444.5079, 6.745653e-28], [1448.951, 6.764837e-28], [1453.4077, 6.784087e-28], [1457.8782, 6.803401e-28], [1462.3624, 6.822781e-28], [1466.8604, 6.842227e-28], [1471.3723, 6.86174e-28], [1475.898, 6.881319e-28], [1480.4377, 6.900964e-28], [1484.9913, 6.920675e-28], [1489.5589, 6.940453e-28], [1494.1406, 6.960298e-28], [1498.7364, 6.98021e-28], [1503.3463, 7.000188e-28], [1507.9703, 7.020234e-28], [1512.6086, 7.040347e-28], [1517.2612, 7.060528e-28], [1521.9281, 7.080777e-28], [1526.6093, 7.101094e-28], [1531.305, 7.121479e-28], [1536.015, 7.141933e-28], [1540.7396, 7.162455e-28], [1545.4787, 7.183046e-28], [1550.2324, 7.203705e-28], [1555.0007, 7.224436e-28], [1559.7836, 7.245252e-28], [1564.5813, 7.266138e-28], [1569.3937, 7.287094e-28], [1574.221, 7.30812e-28], [1579.063, 7.329216e-28], [1583.92, 7.350382e-28], [1588.7919, 7.371618e-28], [1593.6788, 7.392925e-28], [1598.5807, 7.414303e-28], [1603.4978, 7.435753e-28], [1608.4299, 7.457273e-28], [1613.3772, 7.478865e-28], [1618.3397, 7.500529e-28], [1623.3175, 7.522264e-28], [1628.3106, 7.544072e-28], [1633.319, 7.565952e-28], [1638.3429, 7.587905e-28], [1643.3822, 7.60993e-28], [1648.437, 7.632029e-28], [1653.5074, 7.6542e-28], [1658.5933, 7.676446e-28], [1663.6949, 7.698766e-28], [1668.8122, 7.721164e-28], [1673.9452, 7.743653e-28], [1679.0941, 7.76664e-28], [1684.2587, 7.788923e-28], [1689.4393, 7.811484e-28], [1694.6357, 7.834233e-28], [1699.8482, 7.857068e-28], [1705.0767, 7.87998e-28], [1710.3213, 7.902969e-28], [1715.582, 7.926033e-28], [1720.8588, 7.949174e-28], [1726.152, 7.972391e-28], [1731.4614, 7.995684e-28], [1736.7871, 8.019053e-28], [1742.1292, 8.0425e-28], [1747.4877, 8.066023e-28], [1752.8628, 8.089624e-28], [1758.2543, 8.113302e-28], [1763.6625, 8.137058e-28], [1769.0872, 8.160892e-28], [1774.5287, 8.184803e-28], [1779.9869, 8.208793e-28], [1785.4619, 8.232862e-28], [1790.9537, 8.257009e-28], [1796.4624, 8.281235e-28], [1801.9881, 8.305541e-28], [1807.5308, 8.329926e-28], [1813.0905, 8.35439e-28], [1818.6673, 8.378935e-28], [1824.2612, 8.403559e-28], [1829.8724, 8.428264e-28], [1835.5008, 8.45305e-28], [1841.1466, 8.477916e-28], [1846.8097, 8.502864e-28], [1852.4902, 8.527892e-28], [1858.1882, 8.553003e-28], [1863.9037, 8.578195e-28], [1869.6368, 8.603468e-28], [1875.3876, 8.628825e-28], [1881.156, 8.654263e-28], [1886.9422, 8.679785e-28], [1892.7461, 8.70539e-28], [1898.5679, 8.731077e-28], [1904.4077, 8.756848e-28], [1910.2653, 8.782702e-28], [1916.1411, 8.808641e-28], [1922.0348, 8.834664e-28], [1927.9467, 8.860771e-28], [1933.8768, 8.886963e-28], [1939.8252, 8.91324e-28], [1945.7918, 8.939665e-28], [1951.7768, 8.966203e-28], [1957.7802, 8.992826e-28], [1963.802, 9.019536e-28], [1969.8424, 9.046331e-28], [1975.9013, 9.073212e-28], [1981.9789, 9.10018e-28], [1988.0752, 9.127234e-28], [1994.1903, 9.154376e-28], [2000.3241, 9.181604e-28], [2006.4768, 9.208919e-28], [2012.6484, 9.236323e-28], [2018.8391, 9.263815e-28], [2025.0487, 9.291394e-28], [2031.2775, 9.319063e-28], [2037.5254, 9.346819e-28], [2043.7926, 9.374665e-28], [2050.079, 9.4026e-28], [2056.3847, 9.430624e-28], [2062.7099, 9.458739e-28], [2069.0545, 9.486943e-28], [2075.4186, 9.515238e-28], [2081.8023, 9.543624e-28], [2088.2056, 9.572102e-28], [2094.6286, 9.60068e-28], [2101.0714, 9.672075e-28], [2107.534, 9.658089e-28], [2114.0165, 9.68692e-28], [2120.5189, 9.715852e-28], [2127.0413, 9.744878e-28], [2133.5838, 9.773997e-28], [2140.1464, 9.803209e-28], [2146.7292, 9.832515e-28], [2153.3322, 9.861914e-28], [2159.9556, 9.891407e-28], [2166.5993, 9.920995e-28], [2173.2634, 9.950676e-28], [2179.9481, 9.980452e-28], [2186.6533, 1.001032e-27], [2193.3791, 1.004029e-27], [2200.1256, 1.007035e-27], [2206.8929, 1.010051e-27], [2213.681, 1.013076e-27], [2220.49, 1.016111e-27], [2227.3199, 1.019156e-27], [2234.1708, 1.02221e-27], [2241.0428, 1.025275e-27], [2247.9359, 1.028349e-27], [2254.8503, 1.031432e-27], [2261.7859, 1.034526e-27], [2268.7428, 1.037629e-27], [2275.7211, 1.040742e-27], [2282.7209, 1.043865e-27], [2289.7422, 1.046998e-27], [2296.7852, 1.050141e-27], [2303.8497, 1.053294e-27], [2310.9361, 1.056457e-27], [2318.0442, 1.05963e-27], [2325.1741, 1.062813e-27], [2332.3261, 1.066006e-27], [2339.5, 1.06921e-27], [2346.6959, 1.072423e-27], [2353.914, 1.075647e-27], [2361.1543, 1.078881e-27], [2368.4169, 1.082125e-27], [2375.7018, 1.08538e-27], [2383.0091, 1.088645e-27], [2390.3389, 1.09192e-27], [2397.6913, 1.095205e-27], [2405.0662, 1.098501e-27], [2412.4639, 1.101808e-27], [2419.8843, 1.105125e-27], [2427.3275, 1.108452e-27], [2434.7936, 1.11179e-27], [2442.2827, 1.115139e-27], [2449.7948, 1.118498e-27], [2457.33, 1.121868e-27], [2464.8884, 1.125248e-27], [2472.4701, 1.128639e-27], [2480.075, 1.132041e-27], [2487.7034, 1.135454e-27], [2495.3552, 1.138877e-27], [2503.0306, 1.142312e-27], [2510.7295, 1.145757e-27], [2518.4522, 1.149216e-27], [2526.1986, 1.152681e-27], [2533.9688, 1.156158e-27], [2541.763, 1.159647e-27], [2549.5811, 1.163147e-27], [2557.4232, 1.166658e-27], [2565.2895, 1.17018e-27], [2573.1799, 1.173713e-27], [2581.0947, 1.177258e-27], [2589.0338, 1.180814e-27], [2596.9973, 1.18438e-27], [2604.9852, 1.187959e-27], [2612.9978, 1.191548e-27], [2621.035, 1.195149e-27], [2629.097, 1.198761e-27], [2637.1837, 1.202385e-27], [2645.2953, 1.20602e-27], [2653.4318, 1.209666e-27], [2661.5934, 1.213324e-27], [2669.7801, 1.216994e-27], [2677.992, 1.220675e-27], [2686.2291, 1.224367e-27], [2694.4915, 1.228072e-27], [2702.7794, 1.231788e-27], [2711.0928, 1.235515e-27], [2719.4317, 1.239255e-27], [2727.7963, 1.243006e-27], [2736.1866, 1.246769e-27], [2744.6027, 1.250544e-27], [2753.0448, 1.254331e-27], [2761.5127, 1.25813e-27], [2770.0067, 1.261941e-27], [2778.5269, 1.265764e-27], [2787.0732, 1.269599e-27], [2795.6459, 1.273446e-27], [2804.2449, 1.277305e-27], [2812.8704, 1.281176e-27], [2821.5223, 1.285059e-27], [2830.201, 1.288955e-27], [2838.9062, 1.292863e-27], [2847.6383, 1.296783e-27], [2856.3972, 1.300715e-27], [2865.1831, 1.30466e-27], [2873.996, 1.308618e-27], [2882.836, 1.312587e-27], [2891.7032, 1.31657e-27], [2900.5977, 1.320565e-27], [2909.5195, 1.324572e-27], [2918.4688, 1.328592e-27], [2927.4456, 1.332625e-27], [2936.45, 1.33667e-27], [2945.4821, 1.340729e-27], [2954.542, 1.344799e-27], [2963.6297, 1.348882e-27], [2972.7454, 1.352979e-27], [2981.8892, 1.357088e-27], [2991.061, 1.36121e-27], [3000.2611, 1.365346e-27], [3009.4895, 1.369494e-27], [3018.7463, 1.373655e-27], [3028.0315, 1.377829e-27], [3037.3453, 1.382017e-27], [3046.6877, 1.386217e-27], [3056.0589, 1.390431e-27], [3065.4589, 1.394658e-27], [3074.8878, 1.398899e-27], [3084.3457, 1.403152e-27], [3093.8327, 1.407419e-27], [3103.3489, 1.4117e-27], [3112.8944, 1.415993e-27], [3122.4692, 1.420301e-27], [3132.0735, 1.424622e-27], [3141.7073, 1.428956e-27], [3151.3708, 1.433304e-27], [3161.0639, 1.437666e-27], [3170.7869, 1.442041e-27], [3180.5398, 1.44643e-27], [3190.3227, 1.450833e-27], [3200.1357, 1.45525e-27], [3209.9788, 1.459681e-27], [3219.8523, 1.464125e-27], [3229.7561, 1.468584e-27], [3239.6903, 1.473056e-27], [3249.6552, 1.477543e-27], [3259.6506, 1.482043e-27], [3269.6769, 1.486558e-27], [3279.7339, 1.491087e-27], [3289.8219, 1.49563e-27], [3299.941, 1.500187e-27], [3310.0911, 1.504759e-27], [3320.2725, 1.509345e-27], [3330.4852, 1.513945e-27], [3340.7293, 1.51856e-27], [3351.0049, 1.52319e-27], [3361.3121, 1.527855e-27], [3371.651, 1.532492e-27], [3382.0217, 1.537165e-27], [3392.4243, 1.541852e-27], [3402.859, 1.546555e-27], [3413.3257, 1.551272e-27], [3423.8246, 1.556004e-27], [3434.3558, 1.56075e-27], [3444.9194, 1.565512e-27], [3455.5154, 1.570289e-27], [3466.1441, 1.57508e-27], [3476.8055, 1.579887e-27], [3487.4996, 1.584708e-27], [3498.2267, 1.589545e-27], [3508.9867, 1.594397e-27], [3519.7799, 1.599264e-27], [3530.6062, 1.604146e-27], [3541.4659, 1.609044e-27], [3552.3589, 1.613957e-27], [3563.2855, 1.618886e-27], [3574.2456, 1.62383e-27], [3585.2395, 1.628789e-27], [3596.2672, 1.633764e-27], [3607.3288, 1.638754e-27], [3618.4244, 1.643761e-27], [3629.5542, 1.648783e-27], [3640.7182, 1.65382e-27], [3651.9165, 1.658874e-27], [3663.1493, 1.663943e-27], [3674.4166, 1.669028e-27], [3685.7186, 1.674129e-27], [3697.0553, 1.679246e-27], [3708.427, 1.684379e-27], [3719.8336, 1.689528e-27], [3731.2752, 1.694694e-27], [3742.7521, 1.699875e-27], [3754.2643, 1.705073e-27], [3765.8119, 1.710287e-27], [3777.395, 1.715518e-27], [3789.0137, 1.720765e-27], [3800.6682, 1.726028e-27], [3812.3585, 1.731307e-27], [3824.0848, 1.736604e-27], [3835.8471, 1.741917e-27], [3847.6456, 1.747246e-27], [3859.4804, 1.752593e-27], [3871.3516, 1.757955e-27], [3883.2594, 1.763335e-27], [3895.2037, 1.768732e-27], [3907.1848, 1.774146e-27], [3919.2028, 1.779576e-27], [3931.2577, 1.785024e-27], [3943.3497, 1.790488e-27], [3955.4788, 1.79597e-27], [3967.6453, 1.801469e-27], [3979.8493, 1.806985e-27], [3992.0907, 1.812519e-27], [4004.3698, 1.81807e-27], [4016.6867, 1.823638e-27], [4029.0414, 1.829224e-27], [4041.4342, 1.834827e-27], [4053.8651, 1.840448e-27], [4066.3342, 1.846086e-27], [4078.8417, 1.851742e-27], [4091.3876, 1.857416e-27], [4103.9721, 1.863107e-27], [4116.5954, 1.868817e-27], [4129.2574, 1.874544e-27], [4141.9584, 1.880289e-27], [4154.6985, 1.886053e-27], [4167.4778, 1.891834e-27], [4180.2963, 1.897634e-27], [4193.1543, 1.903451e-27], [4206.0519, 1.909287e-27], [4218.9891, 1.915141e-27], [4231.9661, 1.921014e-27], [4244.983, 1.926905e-27], [4258.04, 1.932814e-27], [4271.1371, 1.938742e-27], [4284.2745, 1.944689e-27], [4297.4524, 1.950654e-27], [4310.6707, 1.956638e-27], [4323.9297, 1.962641e-27], [4337.2295, 1.968662e-27], [4350.5702, 1.974703e-27], [4363.9519, 1.980762e-27], [4377.3748, 1.98684e-27], [4390.839, 1.992938e-27], [4404.3446, 1.999054e-27], [4417.8918, 2.00519e-27], [4431.4806, 2.011345e-27], [4445.1112, 2.017519e-27], [4458.7837, 2.023713e-27], [4472.4983, 2.029926e-27], [4486.2551, 2.036159e-27], [4500.0541, 2.042411e-27], [4513.8957, 2.048683e-27], [4527.7798, 2.054974e-27], [4541.7066, 2.061286e-27], [4555.6762, 2.067617e-27], [4569.6888, 2.073968e-27], [4583.7446, 2.080338e-27], [4597.8435, 2.086729e-27], [4611.9858, 2.09314e-27], [4626.1716, 2.099571e-27], [4640.4011, 2.106022e-27], [4654.6743, 2.112494e-27], [4668.9914, 2.118986e-27], [4683.3526, 2.125498e-27], [4697.7579, 2.13203e-27], [4712.2075, 2.138584e-27], [4726.7016, 2.145157e-27], [4741.2403, 2.151752e-27], [4755.8236, 2.158367e-27], [4770.4519, 2.165003e-27], [4785.1251, 2.171659e-27], [4799.8435, 2.178337e-27], [4814.6071, 2.185036e-27], [4829.4162, 2.191755e-27], [4844.2708, 2.198496e-27], [4859.1711, 2.205258e-27], [4874.1172, 2.212041e-27], [4889.1093, 2.218846e-27], [4904.1475, 2.225672e-27], [4919.2319, 2.232519e-27], [4934.3628, 2.239388e-27], [4949.5402, 2.246279e-27], [4964.7643, 2.253191e-27], [4980.0352, 2.260125e-27], [4995.3531, 2.267081e-27], [5010.7181, 2.274058e-27], [5026.1303, 2.281058e-27], [5041.59, 2.28808e-27], [5057.0972, 2.295123e-27], [5072.6521, 2.302189e-27], [5088.2549, 2.309277e-27], [5103.9056, 2.316388e-27], [5119.6045, 2.323521e-27], [5135.3517, 2.330676e-27], [5151.1473, 2.337854e-27], [5166.9915, 2.345054e-27], [5182.8844, 2.352277e-27], [5198.8262, 2.359523e-27], [5214.8171, 2.366791e-27], [5230.8571, 2.374083e-27], [5246.9465, 2.381397e-27], [5263.0854, 2.388735e-27], [5279.2739, 2.396096e-27], [5295.5122, 2.40348e-27], [5311.8004, 2.410887e-27], [5328.1387, 2.418317e-27], [5344.5273, 2.425771e-27], [5360.9663, 2.433248e-27], [5377.4559, 2.440749e-27], [5393.9962, 2.448274e-27], [5410.5874, 2.455823e-27], [5427.2296, 2.463395e-27], [5443.923, 2.470991e-27], [5460.6677, 2.478611e-27], [5477.4639, 2.486255e-27], [5494.3118, 2.493923e-27], [5511.2116, 2.501616e-27], [5528.1633, 2.509332e-27], [5545.1671, 2.517073e-27], [5562.2233, 2.524839e-27], [5579.3319, 2.532629e-27], [5596.4931, 2.540444e-27], [5613.7071, 2.548283e-27], [5630.9741, 2.556147e-27], [5648.2942, 2.564036e-27], [5665.6675, 2.57195e-27], [5683.0943, 2.579889e-27], [5700.5747, 2.587853e-27], [5718.1088, 2.595842e-27], [5735.6969, 2.603857e-27], [5753.3391, 2.611896e-27], [5771.0356, 2.619962e-27], [5788.7865, 2.628053e-27], [5806.5919, 2.636169e-27], [5824.4522, 2.644311e-27], [5842.3674, 2.652479e-27], [5860.3377, 2.660673e-27], [5878.3632, 2.668893e-27], [5896.4443, 2.677139e-27], [5914.5809, 2.685411e-27], [5932.7733, 2.693709e-27], [5951.0217, 2.702033e-27], [5969.3262, 2.710384e-27], [5987.687, 2.718762e-27], [6006.1042, 2.727166e-27], [6024.5781, 2.735596e-27], [6043.1089, 2.744054e-27], [6061.6966, 2.752538e-27], [6080.3416, 2.761049e-27], [6099.0438, 2.769587e-27], [6117.8036, 2.778152e-27], [6136.6211, 2.786745e-27], [6155.4965, 2.795364e-27], [6174.4299, 2.804011e-27], [6193.4216, 2.812686e-27], [6212.4717, 2.821388e-27], [6231.5804, 2.830118e-27], [6250.7478, 2.838876e-27], [6269.9742, 2.847661e-27], [6289.2598, 2.856475e-27], [6308.6046, 2.865316e-27], [6328.009, 2.874186e-27], [6347.4731, 2.883083e-27], [6366.997, 2.89201e-27], [6386.581, 2.900964e-27], [6406.2252, 2.909947e-27], [6425.9298, 2.918959e-27], [6445.6951, 2.927999e-27], [6465.5211, 2.937069e-27], [6485.4081, 2.946167e-27], [6505.3563, 2.955294e-27], [6525.3659, 2.964451e-27], [6545.437, 2.973636e-27], [6565.5698, 2.982851e-27], [6585.7646, 2.992096e-27], [6606.0215, 3.001369e-27], [6626.3406, 3.010673e-27], [6646.7223, 3.020006e-27], [6667.1667, 3.029369e-27], [6687.674, 3.038763e-27], [6708.2443, 3.048186e-27], [6728.8779, 3.057639e-27], [6749.575, 3.067123e-27], [6770.3357, 3.076637e-27], [6791.1603, 3.086181e-27], [6812.0489, 3.095756e-27], [6833.0018, 3.105362e-27], [6854.0192, 3.114998e-27], [6875.1012, 3.124665e-27], [6896.248, 3.134364e-27], [6917.4599, 3.144093e-27], [6938.737, 3.153854e-27], [6960.0796, 3.163646e-27], [6981.4878, 3.17347e-27], [7002.9619, 3.183325e-27], [7024.502, 3.193212e-27], [7046.1083, 3.203131e-27], [7067.7812, 3.213082e-27], [7089.5207, 3.223064e-27], [7111.327, 3.233079e-27], [7133.2004, 3.243126e-27], [7155.1412, 3.253206e-27], [7177.1494, 3.263318e-27], [7199.2252, 3.273462e-27], [7221.369, 3.28364e-27], [7243.5809, 3.29385e-27], [7265.8612, 3.304093e-27], [7288.2099, 3.31437e-27], [7310.6274, 3.324679e-27], [7333.1139, 3.335022e-27], [7355.6695, 3.345398e-27], [7378.2945, 3.355808e-27], [7400.9891, 3.366252e-27], [7423.7535, 3.37673e-27], [7446.5879, 3.387241e-27], [7469.4925, 3.397787e-27], [7492.4676, 3.408367e-27], [7515.5134, 3.418981e-27], [7538.63, 3.42963e-27], [7561.8178, 3.440313e-27], [7585.0769, 3.451031e-27], [7608.4075, 3.461784e-27], [7631.8099, 3.472572e-27], [7655.2842, 3.483395e-27], [7678.8308, 3.494254e-27], [7702.4498, 3.505147e-27], [7726.1414, 3.516077e-27], [7749.906, 3.527042e-27], [7773.7436, 3.538043e-27], [7797.6545, 3.54908e-27], [7821.639, 3.560153e-27], [7845.6972, 3.571262e-27], [7869.8295, 3.582408e-27], [7894.0359, 3.59359e-27], [7918.3169, 3.604809e-27], [7942.6725, 3.616064e-27], [7967.103, 3.627357e-27], [7991.6087, 3.638687e-27], [8016.1898, 3.650054e-27], [8040.8464, 3.661458e-27], [8065.5789, 3.6729e-27], [8090.3875, 3.68438e-27], [8115.2724, 3.695898e-27], [8140.2338, 3.707454e-27], [8165.272, 3.719047e-27], [8190.3872, 3.73068e-27], [8215.5797, 3.742351e-27], [8240.8496, 3.75406e-27], [8266.1973, 3.765808e-27], [8291.623, 3.777595e-27], [8317.1268, 3.789422e-27], [8342.7091, 3.801288e-27], [8368.3701, 3.813193e-27], [8394.11, 3.825138e-27], [8419.9291, 3.837122e-27], [8445.8276, 3.849147e-27], [8471.8058, 3.861212e-27], [8497.8639, 3.873317e-27], [8524.0021, 3.885463e-27], [8550.2207, 3.897649e-27], [8576.52, 3.909876e-27], [8602.9001, 3.922145e-27], [8629.3614, 3.934454e-27], [8655.9041, 3.946805e-27], [8682.5285, 3.959198e-27], [8709.2347, 3.971633e-27], [8736.0231, 3.984109e-27], [8762.8938, 3.996628e-27], [8789.8472, 4.009189e-27], [8816.8836, 4.021792e-27], [8844.003, 4.034439e-27], [8871.2059, 4.047128e-27], [8898.4925, 4.059861e-27], [8925.863, 4.072637e-27], [8953.3177, 4.085456e-27], [8980.8568, 4.09832e-27], [9008.4807, 4.111227e-27], [9036.1895, 4.124179e-27], [9063.9835, 4.137175e-27], [9091.863, 4.150216e-27], [9119.8283, 4.163301e-27], [9147.8796, 4.176432e-27], [9176.0172, 4.189609e-27], [9204.2413, 4.20283e-27], [9232.5522, 4.216098e-27], [9260.9503, 4.229412e-27], [9289.4356, 4.242772e-27], [9318.0086, 4.256179e-27], [9346.6695, 4.269632e-27], [9375.4185, 4.283133e-27], [9404.256, 4.296681e-27], [9433.1821, 4.310277e-27], [9462.1973, 4.32392e-27], [9491.3016, 4.337612e-27], [9520.4955, 4.351352e-27], [9549.7792, 4.365141e-27], [9579.153, 4.378979e-27], [9608.6171, 4.392866e-27], [9638.1718, 4.406802e-27], [9667.8175, 4.420789e-27], [9697.5543, 4.434826e-27], [9727.3826, 4.448913e-27], [9757.3027, 4.463051e-27], [9787.3147, 4.47724e-27], [9817.4191, 4.491481e-27], [9847.6161, 4.505773e-27], [9877.906, 4.520118e-27], [9908.289, 4.534515e-27], [9938.7655, 4.548965e-27], [9969.3357, 4.563468e-27], [10000.0, 4.578024e-27]])
    
    fiveyr = 5 * 365.25 * 86400
    snr = 10
    f1 = data[:,0]
    S1 = fiveyr*(data[:,1]/10)**2
    fstop = [f1[0],f1[-1]]

    # % ---- Interpolate to desired frequencies.  Set noise to infinity outside
    # %      band of frequencies covered by f1.
    # S = inf(size(f)); 
    # k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop



def virgo2016():

    data = np.array([[ 20.0 , 1.035616431417693e-42 ], [ 20.0625 , 1.0104890165179941e-42 ], [ 20.125 , 9.860083035557636e-43 ], [ 20.1875 , 9.622281150856483e-43 ], 
                    [ 20.25 , 9.390808553222255e-43 ], [ 20.3125 , 9.165906781698933e-43 ], [ 20.375 , 8.947039501131745e-43 ], [ 20.4375 , 8.733792581502592e-43 ], 
                    [ 20.5 , 8.526583680285831e-43 ], [ 20.5625 , 8.324839300604706e-43 ], [ 20.625 , 8.128843510900958e-43 ], [ 20.6875 , 7.937709666180569e-43 ], 
                    [ 20.75 , 7.751868502406285e-43 ], [ 20.8125 , 7.570811867015822e-43 ], [ 20.875 , 7.394756922510494e-43 ], [ 20.9375 , 7.223272123819575e-43 ], 
                    [ 21.0 , 7.0562587832031904e-43 ], [ 21.0625 , 6.893605424640977e-43 ], [ 21.125 , 6.735320347237326e-43 ], [ 21.1875 , 6.581011471414998e-43 ], 
                    [ 21.25 , 6.430940739093148e-43 ], [ 21.3125 , 6.2845585857199656e-43 ], [ 21.375 , 6.141931233014983e-43 ], [ 21.4375 , 6.003230195750131e-43 ], 
                    [ 21.5 , 5.8678806747217334e-43 ], [ 21.5625 , 5.7360473709576185e-43 ], [ 21.625 , 5.607486618927931e-43 ], [ 21.6875 , 5.482290326300895e-43 ], 
                    [ 21.75 , 5.3603239531846796e-43 ], [ 21.8125 , 5.2414796205878875e-43 ], [ 21.875 , 5.125448995774404e-43 ], [ 21.9375 , 5.012489816214817e-43 ], 
                    [ 22.0 , 4.9023388443538704e-43 ], [ 22.0625 , 4.7949065498143686e-43 ], [ 22.125 , 4.6901016175551204e-43 ], [ 22.1875 , 4.588003283563582e-43 ], 
                    [ 22.25 , 4.488332708748791e-43 ], [ 22.3125 , 4.391207772351713e-43 ], [ 22.375 , 4.296420667650044e-43 ], [ 22.4375 , 4.204030022636231e-43 ], 
                    [ 22.5 , 4.1137929095778245e-43 ], [ 22.5625 , 4.025868898938506e-43 ], [ 22.625 , 3.940043868755037e-43 ], [ 22.6875 , 3.8563369977880944e-43 ], 
                    [ 22.75 , 3.774599698588644e-43 ], [ 22.8125 , 3.69486169000291e-43 ], [ 22.875 , 3.6170374573205155e-43 ], [ 22.9375 , 3.541106958590534e-43 ], 
                    [ 23.0 , 3.466972075200786e-43 ], [ 23.0625 , 3.3945731977202818e-43 ], [ 23.125 , 3.3239267105456634e-43 ], [ 23.1875 , 3.2549337552483204e-43 ], 
                    [ 23.25 , 3.187643965484648e-43 ], [ 23.3125 , 3.1219248718510075e-43 ], [ 23.375 , 3.0577136378045307e-43 ], [ 23.4375 , 2.9949649763409474e-43 ], 
                    [ 23.5 , 2.9337473576209214e-43 ], [ 23.5625 , 2.8739362871022213e-43 ], [ 23.625 , 2.815532711901604e-43 ], [ 23.6875 , 2.75853532020373e-43 ], 
                    [ 23.75 , 2.7028168117661726e-43 ], [ 23.8125 , 2.6483522029075423e-43 ], [ 23.875 , 2.595102136347669e-43 ], [ 23.9375 , 2.54312963611291e-43 ], 
                    [ 24.0 , 2.4923620809721394e-43 ], [ 24.0625 , 2.4427898437238022e-43 ], [ 24.125 , 2.3942566381617056e-43 ], [ 24.1875 , 2.3468662555876477e-43 ], 
                    [ 24.25 , 2.300560984871887e-43 ], [ 24.3125 , 2.2552613553314874e-43 ], [ 24.375 , 2.2110305083621222e-43 ], [ 24.4375 , 2.167763676066851e-43 ], 
                    [ 24.5 , 2.125422483349823e-43 ], [ 24.5625 , 2.084089645954105e-43 ], [ 24.625 , 2.0437007629221833e-43 ], [ 24.6875 , 2.004151651671187e-43 ], 
                    [ 24.75 , 1.9655032914682277e-43 ], [ 24.8125 , 1.927710129164445e-43 ], [ 24.875 , 1.890736801972975e-43 ], [ 24.9375 , 1.854595937647331e-43 ], 
                    [ 25.0 , 1.819228471004947e-43 ], [ 25.0625 , 1.7846420680388874e-43 ], [ 25.125 , 1.750799805315774e-43 ], [ 25.1875 , 1.7176714568653748e-43 ], 
                    [ 25.25 , 1.6853056826220976e-43 ], [ 25.3125 , 1.6536375020872258e-43 ], [ 25.375 , 1.6226522223564215e-43 ], [ 25.4375 , 1.5922914041532018e-43 ], 
                    [ 25.5 , 1.5626098611791125e-43 ], [ 25.5625 , 1.5335764177475802e-43 ], [ 25.625 , 1.505156334878434e-43 ], [ 25.6875 , 1.4773496465323737e-43 ], 
                    [ 25.75 , 1.4500981485990923e-43 ], [ 25.8125 , 1.4234498360563176e-43 ], [ 25.875 , 1.3973544993158407e-43 ], [ 25.9375 , 1.3718041387260903e-43 ], 
                    [ 26.0 , 1.3468157203742365e-43 ], [ 26.0625 , 1.3223450023825021e-43 ], [ 26.125 , 1.2983741212865847e-43 ], [ 26.1875 , 1.2748838058591754e-43 ], 
                    [ 26.25 , 1.251900372349967e-43 ], [ 26.3125 , 1.2294277985834053e-43 ], [ 26.375 , 1.207362610497417e-43 ], [ 26.4375 , 1.1857864201356108e-43 ], 
                    [ 26.5 , 1.164656411420234e-43 ], [ 26.5625 , 1.1439527489305232e-43 ], [ 26.625 , 1.1236701292066743e-43 ], [ 26.6875 , 1.1038099645898549e-43 ], 
                    [ 26.75 , 1.0843470381617566e-43 ], [ 26.8125 , 1.0652900347841669e-43 ], [ 26.875 , 1.0466106920227928e-43 ], [ 26.9375 , 1.0283007550858132e-43 ], 
                    [ 27.0 , 1.0103809136053608e-43 ], [ 27.0625 , 9.92817821543936e-44 ], [ 27.125 , 9.755857156169237e-44 ], [ 27.1875 , 9.587129252744717e-44 ], 
                    [ 27.25 , 9.42181219827403e-44 ], [ 27.3125 , 9.259783838369105e-44 ], [ 27.375 , 9.10103123342601e-44 ], [ 27.4375 , 8.945381586843439e-44 ], 
                    [ 27.5 , 8.792664553736765e-44 ], [ 27.5625 , 8.643072427534589e-44 ], [ 27.625 , 8.496362561392555e-44 ], [ 27.6875 , 8.352615019814264e-44 ], 
                    [ 27.75 , 8.211604933968113e-44 ], [ 27.8125 , 8.073442077004549e-44 ], [ 27.875 , 7.937898231238991e-44 ], [ 27.9375 , 7.805073718948788e-44 ], 
                    [ 28.0 , 7.67472448360044e-44 ], [ 28.0625 , 7.54701547493401e-44 ], [ 28.125 , 7.421678839542079e-44 ], [ 28.1875 , 7.298849457196162e-44 ], 
                    [ 28.25 , 7.178287091081728e-44 ], [ 28.3125 , 7.059963525749529e-44 ], [ 28.375 , 6.944094325875875e-44 ], [ 28.4375 , 6.830284690913327e-44 ], 
                    [ 28.5 , 6.718635208759337e-44 ], [ 28.5625 , 6.609183717605072e-44 ], [ 28.625 , 6.501775960376134e-44 ], [ 28.6875 , 6.39640894366635e-44 ], 
                    [ 28.75 , 6.293017207082238e-44 ], [ 28.8125 , 6.191560651711542e-44 ], [ 28.875 , 6.092012575846012e-44 ], [ 28.9375 , 5.994298491088209e-44 ], 
                    [ 29.0 , 5.898465228334976e-44 ], [ 29.0625 , 5.804421794585527e-44 ], [ 29.125 , 5.711972961566495e-44 ], [ 29.1875 , 5.621409139308759e-44 ], 
                    [ 29.25 , 5.53245316161715e-44 ], [ 29.3125 , 5.445077158608036e-44 ], [ 29.375 , 5.359466148058656e-44 ], [ 29.4375 , 5.275292794297576e-44 ], 
                    [ 29.5 , 5.192662182701381e-44 ], [ 29.5625 , 5.111582806391252e-44 ], [ 29.625 , 5.032025369723638e-44 ], [ 29.6875 , 4.953903938418196e-44 ], 
                    [ 29.75 , 4.877133587387504e-44 ], [ 29.8125 , 4.801722194161601e-44 ], [ 29.875 , 4.727718661766913e-44 ], [ 29.9375 , 4.655067988988141e-44 ], 
                    [ 30.0 , 4.583673334238753e-44 ], [ 30.0625 , 4.51361446748527e-44 ], [ 30.125 , 4.444753433764672e-44 ], [ 30.1875 , 4.377138461894901e-44 ], 
                    [ 30.25 , 4.3107621587971645e-44 ], [ 30.3125 , 4.2455453947756927e-44 ], [ 30.375 , 4.181465449663516e-44 ], [ 30.4375 , 4.1185452145070484e-44 ], 
                    [ 30.5 , 4.0567469097566714e-44 ], [ 30.5625 , 3.9959924606655006e-44 ], [ 30.625 , 3.9363481146057154e-44 ], [ 30.6875 , 3.877700791407154e-44 ], 
                    [ 30.75 , 3.82009618417367e-44 ], [ 30.8125 , 3.763519744761051e-44 ], [ 30.875 , 3.7079073601625587e-44 ], [ 30.9375 , 3.6532410914302046e-44 ], 
                    [ 31.0 , 3.5995252720827106e-44 ], [ 31.0625 , 3.546699201115724e-44 ], [ 31.125 , 3.494817584378465e-44 ], [ 31.1875 , 3.4438547985088684e-44 ], 
                    [ 31.25 , 3.3937889356645897e-44 ], [ 31.3125 , 3.3445485388039476e-44 ], [ 31.375 , 3.296140149217809e-44 ], [ 31.4375 , 3.2485531814899653e-44 ], 
                    [ 31.5 , 3.2017451201794296e-44 ], [ 31.5625 , 3.155740709862551e-44 ], [ 31.625 , 3.11055993576836e-44 ], [ 31.6875 , 3.066153570528749e-44 ], 
                    [ 31.75 , 3.0224635830839274e-44 ], [ 31.8125 , 2.97950164848357e-44 ], [ 31.875 , 2.937254186405437e-44 ], [ 31.9375 , 2.895693779324975e-44 ], 
                    [ 32.0 , 2.854885401878215e-44 ], [ 32.0625 , 2.8147361978250987e-44 ], [ 32.125 , 2.7752198562229e-44 ], [ 32.1875 , 2.7363913054036707e-44 ], 
                    [ 32.25 , 2.6982097268381797e-44 ], [ 32.3125 , 2.6606229111217386e-44 ], [ 32.375 , 2.623631376888452e-44 ], [ 32.4375 , 2.587319844271487e-44 ], 
                    [ 32.5 , 2.5515748543977715e-44 ], [ 32.5625 , 2.51641454361506e-44 ], [ 32.625 , 2.48183295422252e-44 ], [ 32.6875 , 2.447807741703855e-44 ], 
                    [ 32.75 , 2.4143096523543986e-44 ], [ 32.8125 , 2.381376369652383e-44 ], [ 32.875 , 2.348986843181497e-44 ], [ 32.9375 , 2.3171263506948107e-44 ], 
                    [ 33.0 , 2.2857662336301324e-44 ], [ 33.0625 , 2.254888634844253e-44 ], [ 33.125 , 2.2245094317202794e-44 ], [ 33.1875 , 2.1946402812046473e-44 ], 
                    [ 33.25 , 2.1652547520774894e-44 ], [ 33.3125 , 2.1363175404176895e-44 ], [ 33.375 , 2.107835515032261e-44 ], [ 33.4375 , 2.0798317382244555e-44 ], 
                    [ 33.5 , 2.0522202923851955e-44 ], [ 33.5625 , 2.0250802960126746e-44 ], [ 33.625 , 1.998375911080436e-44 ], [ 33.6875 , 1.97208158090684e-44 ], 
                    [ 33.75 , 1.9461802293003763e-44 ], [ 33.8125 , 1.9206960917820727e-44 ], [ 33.875 , 1.8956197082008517e-44 ], [ 33.9375 , 1.87093694002234e-44 ], 
                    [ 34.0 , 1.8466263347651896e-44 ], [ 34.0625 , 1.822675747375952e-44 ], [ 34.125 , 1.7991366557308206e-44 ], [ 34.1875 , 1.7759422798699658e-44 ], 
                    [ 34.25 , 1.7530998783076064e-44 ], [ 34.3125 , 1.7305991455763307e-44 ], [ 34.375 , 1.7084294241694054e-44 ], [ 34.4375 , 1.686637254588804e-44 ], 
                    [ 34.5 , 1.665160975892215e-44 ], [ 34.5625 , 1.6440205846616552e-44 ], [ 34.625 , 1.6232056905689018e-44 ], [ 34.6875 , 1.602687606220702e-44 ], 
                    [ 34.75 , 1.582500569849032e-44 ], [ 34.8125 , 1.562593978169788e-44 ], [ 34.875 , 1.542991141515986e-44 ], [ 34.9375 , 1.5236963153315324e-44 ], 
                    [ 35.0 , 1.5046749722606917e-44 ], [ 35.0625 , 1.4859657057000585e-44 ], [ 35.125 , 1.4675238065190157e-44 ], [ 35.1875 , 1.449346240622412e-44 ], 
                    [ 35.25 , 1.4314435604788823e-44 ], [ 35.3125 , 1.4138138771767925e-44 ], [ 35.375 , 1.3964450692906335e-44 ], [ 35.4375 , 1.379323748967085e-44 ], 
                    [ 35.5 , 1.3624403766289867e-44 ], [ 35.5625 , 1.3458392622182625e-44 ], [ 35.625 , 1.329464918816913e-44 ], [ 35.6875 , 1.3133331068402121e-44 ], 
                    [ 35.75 , 1.2974453402920248e-44 ], [ 35.8125 , 1.281774914890034e-44 ], [ 35.875 , 1.2663409588030965e-44 ], [ 35.9375 , 1.2511277064452576e-44 ], 
                    [ 36.0 , 1.2361305138082383e-44 ], [ 36.0625 , 1.221357917013162e-44 ], [ 36.125 , 1.2068029247367064e-44 ], [ 36.1875 , 1.1924455458955207e-44 ], 
                    [ 36.25 , 1.1782856177180627e-44 ], [ 36.3125 , 1.164349019744202e-44 ], [ 36.375 , 1.150618305491924e-44 ], [ 36.4375 , 1.137070770208311e-44 ], 
                    [ 36.5 , 1.1237052096123373e-44 ], [ 36.5625 , 1.1105511353054339e-44 ], [ 36.625 , 1.0975727436019039e-44 ], [ 36.6875 , 1.0847796752918511e-44 ], 
                    [ 36.75 , 1.0721727274989261e-44 ], [ 36.8125 , 1.0597367362399673e-44 ], [ 36.875 , 1.0474765145396096e-44 ], [ 36.9375 , 1.0353906431909636e-44 ], 
                    [ 37.0 , 1.0234792989780812e-44 ], [ 37.0625 , 1.0117298091704164e-44 ], [ 37.125 , 1.0001421409339153e-44 ], [ 37.1875 , 9.887123370056547e-45 ], 
                    [ 37.25 , 9.774403869317273e-45 ], [ 37.3125 , 9.6633312538632e-45 ], [ 37.375 , 9.553760191159007e-45 ], [ 37.4375 , 9.445663146262705e-45 ], 
                    [ 37.5 , 9.339087158640864e-45 ], [ 37.5625 , 9.234012703902251e-45 ], [ 37.625 , 9.130339870777837e-45 ], [ 37.6875 , 9.028039380418528e-45 ], 
                    [ 37.75 , 8.927268356591393e-45 ], [ 37.8125 , 8.827890884078192e-45 ], [ 37.875 , 8.72983306158531e-45 ], [ 37.9375 , 8.633033305702218e-45 ], 
                    [ 38.0 , 8.537610070413568e-45 ], [ 38.0625 , 8.44342684035671e-45 ], [ 38.125 , 8.350452915211139e-45 ], [ 38.1875 , 8.258829360310136e-45 ], 
                    [ 38.25 , 8.168404354289675e-45 ], [ 38.3125 , 8.079157512360375e-45 ], [ 38.375 , 7.991148074625493e-45 ], [ 38.4375 , 7.904394051094903e-45 ], 
                    [ 38.5 , 7.818735727959004e-45 ], [ 38.5625 , 7.734132969744383e-45 ], [ 38.625 , 7.650744885729338e-45 ], [ 38.6875 , 7.568460556844651e-45 ], 
                    [ 38.75 , 7.487248746380314e-45 ], [ 38.8125 , 7.407094971301705e-45 ], [ 38.875 , 7.327982111686143e-45 ], [ 38.9375 , 7.249886820833702e-45 ], 
                    [ 39.0 , 7.17288656385324e-45 ], [ 39.0625 , 7.096936829484992e-45 ], [ 39.125 , 7.02190839337583e-45 ], [ 39.1875 , 6.947779403363691e-45 ], 
                    [ 39.25 , 6.874722806791978e-45 ], [ 39.3125 , 6.802589927605958e-45 ], [ 39.375 , 6.731364602895592e-45 ], [ 39.4375 , 6.661146995115349e-45 ], 
                    [ 39.5 , 6.591782098340632e-45 ], [ 39.5625 , 6.523257317308854e-45 ], [ 39.625 , 6.455665642919606e-45 ], [ 39.6875 , 6.388937548414853e-45 ], 
                    [ 39.75 , 6.323041238987293e-45 ], [ 39.8125 , 6.258061015626998e-45 ], [ 39.875 , 6.19384980643177e-45 ], [ 39.9375 , 6.130418311784686e-45 ], 
                    [ 40.0 , 6.067868056504663e-45 ], [ 40.0625 , 6.00605872397716e-45 ], [ 40.125 , 5.945004505609183e-45 ], [ 40.1875 , 5.884781848566722e-45 ], 
                    [ 40.25 , 5.82532451944111e-45 ], [ 40.3125 , 5.766573841508007e-45 ], [ 40.375 , 5.708553115903812e-45 ], [ 40.4375 , 5.651302489603837e-45 ], 
                    [ 40.5 , 5.59474895791137e-45 ], [ 40.5625 , 5.5389158795851115e-45 ], [ 40.625 , 5.483768003638629e-45 ], [ 40.6875 , 5.4292808173820515e-45 ], 
                    [ 40.75 , 5.375483649673644e-45 ], [ 40.8125 , 5.322390534100208e-45 ], [ 40.875 , 5.2699110330214985e-45 ], [ 40.9375 , 5.218056964316831e-45 ], 
                    [ 41.0 , 5.166878767176962e-45 ], [ 41.0625 , 5.1163105138596453e-45 ], [ 41.125 , 5.066383048260048e-45 ], [ 41.1875 , 5.017050779238971e-45 ], 
                    [ 41.25 , 4.9682950027332366e-45 ], [ 41.3125 , 4.920166318737293e-45 ], [ 41.375 , 4.8726507012078307e-45 ], [ 41.4375 , 4.8256717704626626e-45 ], 
                    [ 41.5 , 4.7792542766855986e-45 ], [ 41.5625 , 4.733408762751127e-45 ], [ 41.625 , 4.6880925618895445e-45 ], [ 41.6875 , 4.6433784453896974e-45 ], 
                    [ 41.75 , 4.5991985530493696e-45 ], [ 41.8125 , 4.5555080682642546e-45 ], [ 41.875 , 4.5123507307083714e-45 ], [ 41.9375 , 4.4697086103166414e-45 ], 
                    [ 42.0 , 4.4275460793928105e-45 ], [ 42.0625 , 4.3859137576289593e-45 ], [ 42.125 , 4.3447768688299684e-45 ], [ 42.1875 , 4.3041227295289905e-45 ], 
                    [ 42.25 , 4.263976437109487e-45 ], [ 42.3125 , 4.2242705147179116e-45 ], [ 42.375 , 4.1850229966669024e-45 ], [ 42.4375 , 4.1462627856111964e-45 ], 
                    [ 42.5 , 4.1079331846710415e-45 ], [ 42.5625 , 4.070082530052763e-45 ], [ 42.625 , 4.032677896589522e-45 ], [ 42.6875 , 3.995675325303895e-45 ], 
                    [ 42.75 , 3.959129416561119e-45 ], [ 42.8125 , 3.923008034113468e-45 ], [ 42.875 , 3.887284234072411e-45 ], [ 42.9375 , 3.8519923966741984e-45 ], 
                    [ 43.0 , 3.8170809623668647e-45 ], [ 43.0625 , 3.782570482164253e-45 ], [ 43.125 , 3.748527297964566e-45 ], [ 43.1875 , 3.714837106428862e-45 ], 
                    [ 43.25 , 3.681493571162139e-45 ], [ 43.3125 , 3.648566719839892e-45 ], [ 43.375 , 3.616004490644371e-45 ], [ 43.4375 , 3.58383559185349e-45 ], 
                    [ 43.5 , 3.552012660457073e-45 ], [ 43.5625 , 3.5205467461267583e-45 ], [ 43.625 , 3.489444940101599e-45 ], [ 43.6875 , 3.4586656578812495e-45 ], 
                    [ 43.75 , 3.428261065252687e-45 ], [ 43.8125 , 3.398216422653342e-45 ], [ 43.875 , 3.368491662491644e-45 ], [ 43.9375 , 3.339123035966878e-45 ], 
                    [ 44.0 , 3.310057075256742e-45 ], [ 44.0625 , 3.281317748198244e-45 ], [ 44.125 , 3.2529136958072604e-45 ], [ 44.1875 , 3.224797453182597e-45 ], 
                    [ 44.25 , 3.1970209148113464e-45 ], [ 44.3125 , 3.169541172968516e-45 ], [ 44.375 , 3.1423530632159395e-45 ], [ 44.4375 , 3.115503970194316e-45 ], 
                    [ 44.5 , 3.0889203762313593e-45 ], [ 44.5625 , 3.0626310444868236e-45 ], [ 44.625 , 3.03664286074506e-45 ], [ 44.6875 , 3.0109273987861906e-45 ], 
                    [ 44.75 , 2.9855026180664996e-45 ], [ 44.8125 , 2.9603346054434767e-45 ], [ 44.875 , 2.935466752693291e-45 ], [ 44.9375 , 2.910860658193823e-45 ], 
                    [ 45.0 , 2.8864892939935545e-45 ], [ 45.0625 , 2.8624469748733673e-45 ], [ 45.125 , 2.838633390536285e-45 ], [ 45.1875 , 2.8150729498022224e-45 ], 
                    [ 45.25 , 2.7917836837732396e-45 ], [ 45.3125 , 2.7687245016575603e-45 ], [ 45.375 , 2.7459391417920783e-45 ], [ 45.4375 , 2.72337062810929e-45 ], 
                    [ 45.5 , 2.7010473693984713e-45 ], [ 45.5625 , 2.6789727657670423e-45 ], [ 45.625 , 2.6571134586408584e-45 ], [ 45.6875 , 2.6355270916319323e-45 ], 
                    [ 45.75 , 2.6141400963214977e-45 ], [ 45.8125 , 2.5929786953857253e-45 ], [ 45.875 , 2.5720478740637966e-45 ], [ 45.9375 , 2.5513249488925322e-45 ], 
                    [ 46.0 , 2.5308540277201474e-45 ], [ 46.0625 , 2.5105727658864168e-45 ], [ 46.125 , 2.4905162063441486e-45 ], [ 46.1875 , 2.4706586437519392e-45 ], 
                    [ 46.25 , 2.4509956013519383e-45 ], [ 46.3125 , 2.4315748174829298e-45 ], [ 46.375 , 2.412334839739549e-45 ], [ 46.4375 , 2.3933086541642326e-45 ], 
                    [ 46.5 , 2.3744642129027153e-45 ], [ 46.5625 , 2.355816256144837e-45 ], [ 46.625 , 2.337379421473158e-45 ], [ 46.6875 , 2.3191170703425077e-45 ], 
                    [ 46.75 , 2.301063481271377e-45 ], [ 46.8125 , 2.2831699327776844e-45 ], [ 46.875 , 2.2654690383383496e-45 ], [ 46.9375 , 2.2479558457620154e-45 ], 
                    [ 47.0 , 2.230625773022788e-45 ], [ 47.0625 , 2.213476260104961e-45 ], [ 47.125 , 2.196474196546233e-45 ], [ 47.1875 , 2.1796540951747784e-45 ], 
                    [ 47.25 , 2.1629944586475124e-45 ], [ 47.3125 , 2.1465324824940528e-45 ], [ 47.375 , 2.1302282084261048e-45 ], [ 47.4375 , 2.114072562773424e-45 ], 
                    [ 47.5 , 2.0981108325999572e-45 ], [ 47.5625 , 2.0822889200685138e-45 ], [ 47.625 , 2.0666355574510323e-45 ], [ 47.6875 , 2.0511227990992733e-45 ], 
                    [ 47.75 , 2.0357763009685987e-45 ], [ 47.8125 , 2.0205929896252304e-45 ], [ 47.875 , 2.00556137812595e-45 ], [ 47.9375 , 1.9906734653599112e-45 ], 
                    [ 48.0 , 1.9759063416274053e-45 ], [ 48.0625 , 1.961322892735098e-45 ], [ 48.125 , 1.9468677016722968e-45 ], [ 48.1875 , 1.9325790442387104e-45 ], 
                    [ 48.25 , 1.918417984811437e-45 ], [ 48.3125 , 1.9043852915443595e-45 ], [ 48.375 , 1.8905030461441724e-45 ], [ 48.4375 , 1.876742412471303e-45 ], 
                    [ 48.5 , 1.8631319862977956e-45 ], [ 48.5625 , 1.8496339496491483e-45 ], [ 48.625 , 1.8362948832529865e-45 ], [ 48.6875 , 1.8230685161200953e-45 ], 
                    [ 48.75 , 1.8099695596909563e-45 ], [ 48.8125 , 1.7970068212818507e-45 ], [ 48.875 , 1.7841721849302516e-45 ], [ 48.9375 , 1.771461089698826e-45 ], 
                    [ 49.0 , 1.7588515337490246e-45 ], [ 49.0625 , 1.746394186251007e-45 ], [ 49.125 , 1.7340368902998113e-45 ], [ 49.1875 , 1.7218183598896245e-45 ], 
                    [ 49.25 , 1.7097025589606914e-45 ], [ 49.3125 , 1.6977073712073337e-45 ], [ 49.375 , 1.6858262227742714e-45 ], [ 49.4375 , 1.6740576123451402e-45 ], 
                    [ 49.5 , 1.6624097008013095e-45 ], [ 49.5625 , 1.650860714118994e-45 ], [ 49.625 , 1.6394326175931982e-45 ], [ 49.6875 , 1.6280920871497445e-45 ], 
                    [ 49.75 , 1.6168890558961765e-45 ], [ 49.8125 , 1.6057712828326235e-45 ], [ 49.875 , 1.5947637495984327e-45 ], [ 49.9375 , 1.5838527700449368e-45 ], 
                    [ 50.0 , 1.5730542130118766e-45 ], [ 50.0625 , 1.5623555304827027e-45 ], [ 50.125 , 1.551754146409925e-45 ], [ 50.1875 , 1.5412547886802657e-45 ], 
                    [ 50.25 , 1.530841271559155e-45 ], [ 50.3125 , 1.5205416454890093e-45 ], [ 50.375 , 1.5103208593086172e-45 ], [ 50.4375 , 1.500220881697917e-45 ], 
                    [ 50.5 , 1.4901948835128662e-45 ], [ 50.5625 , 1.480274164312188e-45 ], [ 50.625 , 1.4704313353359074e-45 ], [ 50.6875 , 1.460690568986054e-45 ], 
                    [ 50.75 , 1.4510333989873927e-45 ], [ 50.8125 , 1.4414715144610024e-45 ], [ 50.875 , 1.4319922392067014e-45 ], [ 50.9375 , 1.422594462286518e-45 ], 
                    [ 51.0 , 1.4132951856740886e-45 ], [ 51.0625 , 1.4040782943912114e-45 ], [ 51.125 , 1.3949444033814523e-45 ], [ 51.1875 , 1.385877625347961e-45 ], 
                    [ 51.25 , 1.376914721570221e-45 ], [ 51.3125 , 1.3680168123628522e-45 ], [ 51.375 , 1.3592176030342461e-45 ], [ 51.4375 , 1.350479719522398e-45 ], 
                    [ 51.5 , 1.3418388592216573e-45 ], [ 51.5625 , 1.333261210227764e-45 ], [ 51.625 , 1.3247733103789981e-45 ], [ 51.6875 , 1.3163530297937295e-45 ], 
                    [ 51.75 , 1.3080207011344204e-45 ], [ 51.8125 , 1.2997533825598462e-45 ], [ 51.875 , 1.2915586865650838e-45 ], [ 51.9375 , 1.2834352831872618e-45 ], 
                    [ 52.0 , 1.2753855027799409e-45 ], [ 52.0625 , 1.2674117289305489e-45 ], [ 52.125 , 1.2595086564355603e-45 ], [ 52.1875 , 1.2516742753599477e-45 ], 
                    [ 52.25 , 1.2439000133973928e-45 ], [ 52.3125 , 1.2362078259066047e-45 ], [ 52.375 , 1.2285772168555079e-45 ], [ 52.4375 , 1.2210210733720258e-45 ], 
                    [ 52.5 , 1.2135192532737168e-45 ], [ 52.5625 , 1.2060978780818705e-45 ], [ 52.625 , 1.1987282384592206e-45 ], [ 52.6875 , 1.1914343889148284e-45 ], 
                    [ 52.75 , 1.1841881068279544e-45 ], [ 52.8125 , 1.1770263454125053e-45 ], [ 52.875 , 1.16990995537312e-45 ], [ 52.9375 , 1.1628694062772864e-45 ], 
                    [ 53.0 , 1.155873669216106e-45 ], [ 53.0625 , 1.1489537094558659e-45 ], [ 53.125 , 1.1420794204871843e-45 ], [ 53.1875 , 1.135279697440076e-45 ], 
                    [ 53.25 , 1.1285258721621675e-45 ], [ 53.3125 , 1.1218422481846077e-45 ], [ 53.375 , 1.1152042968802475e-45 ], [ 53.4375 , 1.1086319054722606e-45 ], 
                    [ 53.5 , 1.1021068940049753e-45 ], [ 53.5625 , 1.095650902913059e-45 ], [ 53.625 , 1.0892407429812614e-45 ], [ 53.6875 , 1.0828926611337334e-45 ], 
                    [ 53.75 , 1.076590328631578e-45 ], [ 53.8125 , 1.0703485404529852e-45 ], [ 53.875 , 1.0641524891824403e-45 ], [ 53.9375 , 1.0580161462155941e-45 ], 
                    [ 54.0 , 1.051925173920877e-45 ], [ 54.0625 , 1.0458927009547928e-45 ], [ 54.125 , 1.0399042657479341e-45 ], [ 54.1875 , 1.0339714534592827e-45 ], 
                    [ 54.25 , 1.0280828361445067e-45 ], [ 54.3125 , 1.022250613050181e-45 ], [ 54.375 , 1.0164608422377889e-45 ], [ 54.4375 , 1.0107247104445553e-45 ], 
                    [ 54.5 , 1.0050321148451556e-45 ], [ 54.5625 , 9.993970013933009e-46 ], [ 54.625 , 9.938014154405366e-46 ], [ 54.6875 , 9.882562100898532e-46 ], 
                    [ 54.75 , 9.827497879307028e-46 ], [ 54.8125 , 9.772942601597223e-46 ], [ 54.875 , 9.718777052816391e-46 ], [ 54.9375 , 9.665158635363984e-46 ], 
                    [ 55.0 , 9.611900868183476e-46 ], [ 55.0625 , 9.559146241048192e-46 ], [ 55.125 , 9.506750321540399e-46 ], [ 55.1875 , 9.45490608181426e-46 ], 
                    [ 55.25 , 9.403389984963097e-46 ], [ 55.3125 , 9.352370800060988e-46 ], [ 55.375 , 9.301664730121074e-46 ], [ 55.4375 , 9.251466128624898e-46 ], 
                    [ 55.5 , 9.201564810600932e-46 ], [ 55.5625 , 9.1521940656341e-46 ], [ 55.625 , 9.103101468893626e-46 ], [ 55.6875 , 9.05455322201412e-46 ], 
                    [ 55.75 , 9.00626389043749e-46 ], [ 55.8125 , 8.958517669268294e-46 ], [ 55.875 , 8.911026862968888e-46 ], [ 55.9375 , 8.863994024751507e-46 ], 
                    [ 56.0 , 8.817223829028534e-46 ], [ 56.0625 , 8.770977835715709e-46 ], [ 56.125 , 8.725014330973182e-46 ], [ 56.1875 , 8.679490717567356e-46 ], 
                    [ 56.25 , 8.634252160935525e-46 ], [ 56.3125 , 8.589460262938954e-46 ], [ 56.375 , 8.544975978757515e-46 ], [ 56.4375 , 8.500857407827578e-46 ], 
                    [ 56.5 , 8.457036230396052e-46 ], [ 56.5625 , 8.413619978720993e-46 ], [ 56.625 , 8.370556019346478e-46 ], [ 56.6875 , 8.327817916458074e-46 ], 
                    [ 56.75 , 8.285405992185635e-46 ], [ 56.8125 , 8.243329145219077e-46 ], [ 56.875 , 8.201641350506759e-46 ], [ 56.9375 , 8.160219834997467e-46 ], 
                    [ 57.0 , 8.119131952511973e-46 ], [ 57.0625 , 8.078318168487028e-46 ], [ 57.125 , 8.037964202650295e-46 ], [ 57.1875 , 7.997829252769046e-46 ], 
                    [ 57.25 , 7.958046762787728e-46 ], [ 57.3125 , 7.918462157314648e-46 ], [ 57.375 , 7.87934311956197e-46 ], [ 57.4375 , 7.840419060947396e-46 ], 
                    [ 57.5 , 7.801906931769208e-46 ], [ 57.5625 , 7.76361423135565e-46 ], [ 57.625 , 7.725683076955566e-46 ], [ 57.6875 , 7.687992278897132e-46 ], 
                    [ 57.75 , 7.650591613636488e-46 ], [ 57.8125 , 7.613434041816885e-46 ], [ 57.875 , 7.576597553707057e-46 ], [ 57.9375 , 7.540076131832814e-46 ], 
                    [ 58.0 , 7.50381380381177e-46 ], [ 58.0625 , 7.467865880585913e-46 ], [ 58.125 , 7.432140740462188e-46 ], [ 58.1875 , 7.396753110567046e-46 ], 
                    [ 58.25 , 7.3615531868479e-46 ], [ 58.3125 , 7.32671891880384e-46 ], [ 58.375 , 7.292049483669426e-46 ], [ 58.4375 , 7.257772324096414e-46 ], 
                    [ 58.5 , 7.223670993458992e-46 ], [ 58.5625 , 7.189876344437396e-46 ], [ 58.625 , 7.156279958748009e-46 ], [ 58.6875 , 7.122957269164346e-46 ], 
                    [ 58.75 , 7.089859266271109e-46 ], [ 58.8125 , 7.057033495231573e-46 ], [ 58.875 , 7.02449676896908e-46 ], [ 58.9375 , 6.992164916226527e-46 ], 
                    [ 59.0 , 6.960113777555192e-46 ], [ 59.0625 , 6.928230621511705e-46 ], [ 59.125 , 6.8966496460094535e-46 ], [ 59.1875 , 6.865212626167383e-46 ], 
                    [ 59.25 , 6.8341112701280935e-46 ], [ 59.3125 , 6.803162861935092e-46 ], [ 59.375 , 6.7725295227547355e-46 ], [ 59.4375 , 6.742090078088648e-46 ], 
                    [ 59.5 , 6.711897059055087e-46 ], [ 59.5625 , 6.6819222494017726e-46 ], [ 59.625 , 6.65214827586025e-46 ], [ 59.6875 , 6.6226191437584874e-46 ], 
                    [ 59.75 , 6.5932557642033854e-46 ], [ 59.8125 , 6.564185984400344e-46 ], [ 59.875 , 6.535244373764148e-46 ], [ 59.9375 , 6.5065940421876366e-46 ], 
                    [ 60.0 , 6.478074422169909e-46 ], [ 60.0625 , 6.449844241042077e-46 ], [ 60.125 , 6.421788329455491e-46 ], [ 60.1875 , 6.393959675939235e-46 ], 
                    [ 60.25 , 6.366338308206306e-46 ], [ 60.3125 , 6.338882538063745e-46 ], [ 60.375 , 6.311640264320214e-46 ], [ 60.4375 , 6.284535137123464e-46 ], 
                    [ 60.5 , 6.257721575662031e-46 ], [ 60.5625 , 6.231022416787627e-46 ], [ 60.625 , 6.204614797028919e-46 ], [ 60.6875 , 6.178344686847415e-46 ], 
                    [ 60.75 , 6.152289221483981e-46 ], [ 60.8125 , 6.1264019692887084e-46 ], [ 60.875 , 6.100688316215304e-46 ], [ 60.9375 , 6.075179837196284e-46 ], 
                    [ 61.0 , 6.049808330414669e-46 ], [ 61.0625 , 6.02470913495546e-46 ], [ 61.125 , 5.999714070003955e-46 ], [ 61.1875 , 5.9749690935778074e-46 ], 
                    [ 61.25 , 5.950341942202429e-46 ], [ 61.3125 , 5.925906791750035e-46 ], [ 61.375 , 5.901621285117427e-46 ], [ 61.4375 , 5.877517843535597e-46 ], 
                    [ 61.5 , 5.853642943144901e-46 ], [ 61.5625 , 5.829881331284836e-46 ], [ 61.625 , 5.8063352749950596e-46 ], [ 61.6875 , 5.782884317856587e-46 ], 
                    [ 61.75 , 5.759675791273171e-46 ], [ 61.8125 , 5.736586452485353e-46 ], [ 61.875 , 5.713674051589771e-46 ], [ 61.9375 , 5.6909158438057145e-46 ], 
                    [ 62.0 , 5.6682988799205375e-46 ], [ 62.0625 , 5.6458944658392634e-46 ], [ 62.125 , 5.623579416463647e-46 ], [ 62.1875 , 5.601519726953556e-46 ], 
                    [ 62.25 , 5.579546571279838e-46 ], [ 62.3125 , 5.557761046438666e-46 ], [ 62.375 , 5.536101454693034e-46 ], [ 62.4375 , 5.514587546432701e-46 ], 
                    [ 62.5 , 5.493243493681809e-46 ], [ 62.5625 , 5.471998645168927e-46 ], [ 62.625 , 5.450967009412746e-46 ], [ 62.6875 , 5.430016208782967e-46 ], 
                    [ 62.75 , 5.409276453077129e-46 ], [ 62.8125 , 5.388647448754231e-46 ], [ 62.875 , 5.368164874207273e-46 ], [ 62.9375 , 5.347830075391284e-46 ], 
                    [ 63.0 , 5.3276028577565505e-46 ], [ 63.0625 , 5.307597849933237e-46 ], [ 63.125 , 5.28766796038512e-46 ], [ 63.1875 , 5.2679174967025576e-46 ], 
                    [ 63.25 , 5.248259482223093e-46 ], [ 63.3125 , 5.228746436648098e-46 ], [ 63.375 , 5.209372222697042e-46 ], [ 63.4375 , 5.1900967564528046e-46 ], 
                    [ 63.5 , 5.171011732996431e-46 ], [ 63.5625 , 5.151996888987037e-46 ], [ 63.625 , 5.133188554845356e-46 ], [ 63.6875 , 5.1144741914957486e-46 ], 
                    [ 63.75 , 5.095885021416143e-46 ], [ 63.8125 , 5.077420722238432e-46 ], [ 63.875 , 5.0590455343481456e-46 ], [ 63.9375 , 5.040865045385391e-46 ], 
                    [ 64.0 , 5.022749890916404e-46 ], [ 64.0625 , 5.004790457101199e-46 ], [ 64.125 , 4.986919184766989e-46 ], [ 64.1875 , 4.9691799544694515e-46 ], 
                    [ 64.25 , 4.951594387552396e-46 ], [ 64.3125 , 4.93407917947051e-46 ], [ 64.375 , 4.916750992082821e-46 ], [ 64.4375 , 4.899483660240185e-46 ], 
                    [ 64.5 , 4.882352932657708e-46 ], [ 64.5625 , 4.865316429579804e-46 ], [ 64.625 , 4.848382855577547e-46 ], [ 64.6875 , 4.831604484879773e-46 ], 
                    [ 64.75 , 4.8148841776089855e-46 ], [ 64.8125 , 4.79834510800837e-46 ], [ 64.875 , 4.781872618208087e-46 ], [ 64.9375 , 4.765517101400315e-46 ], 
                    [ 65.0 , 4.749269320077848e-46 ], [ 65.0625 , 4.733098448854418e-46 ], [ 65.125 , 4.717111767302834e-46 ], [ 65.1875 , 4.7011790829351806e-46 ], 
                    [ 65.25 , 4.685375769031644e-46 ], [ 65.3125 , 4.6696517076453886e-46 ], [ 65.375 , 4.6540198946955475e-46 ], [ 65.4375 , 4.638517611974098e-46 ], 
                    [ 65.5 , 4.623066966498515e-46 ], [ 65.5625 , 4.607807695531412e-46 ], [ 65.625 , 4.592609869465105e-46 ], [ 65.6875 , 4.5775164655545046e-46 ], 
                    [ 65.75 , 4.562524911141605e-46 ], [ 65.8125 , 4.547592145468961e-46 ], [ 65.875 , 4.5328103904469815e-46 ], [ 65.9375 , 4.518076682891655e-46 ], 
                    [ 66.0 , 4.5034661978638385e-46 ], [ 66.0625 , 4.488942057212352e-46 ], [ 66.125 , 4.474492479486741e-46 ], [ 66.1875 , 4.460185139845519e-46 ], 
                    [ 66.25 , 4.445923548402169e-46 ], [ 66.3125 , 4.431791769689211e-46 ], [ 66.375 , 4.4177278195437946e-46 ], [ 66.4375 , 4.403752172587753e-46 ], 
                    [ 66.5 , 4.389904946271138e-46 ], [ 66.5625 , 4.376101261386756e-46 ], [ 66.625 , 4.362442724086147e-46 ], [ 66.6875 , 4.3488394044875186e-46 ], 
                    [ 66.75 , 4.335319193503896e-46 ], [ 66.8125 , 4.321894002950764e-46 ], [ 66.875 , 4.308510386208805e-46 ], [ 66.9375 , 4.295284635189188e-46 ], 
                    [ 67.0 , 4.282102728237312e-46 ], [ 67.0625 , 4.269012404246611e-46 ], [ 67.125 , 4.256012261067202e-46 ], [ 67.1875 , 4.2430553897986934e-46 ], 
                    [ 67.25 , 4.230249050027691e-46 ], [ 67.3125 , 4.217481362200454e-46 ], [ 67.375 , 4.204810389920267e-46 ], [ 67.4375 , 4.192224563712705e-46 ], 
                    [ 67.5 , 4.179683027616118e-46 ], [ 67.5625 , 4.167270604906801e-46 ], [ 67.625 , 4.1548950434227826e-46 ], [ 67.6875 , 4.1426216187909685e-46 ], 
                    [ 67.75 , 4.130432125105581e-46 ], [ 67.8125 , 4.1182868163645175e-46 ], [ 67.875 , 4.106270562822548e-46 ], [ 67.9375 , 4.094289370060826e-46 ], 
                    [ 68.0 , 4.082395159702362e-46 ], [ 68.0625 , 4.070573189894705e-46 ], [ 68.125 , 4.058793566998844e-46 ], [ 68.1875 , 4.047152290717658e-46 ], 
                    [ 68.25 , 4.0355444034993685e-46 ], [ 68.3125 , 4.024022354625026e-46 ], [ 68.375 , 4.012574638916519e-46 ], [ 68.4375 , 4.001163212029085e-46 ], 
                    [ 68.5 , 3.989873309837162e-46 ], [ 68.5625 , 3.978615263854206e-46 ], [ 68.625 , 3.96743676991384e-46 ], [ 68.6875 , 3.9563345238178815e-46 ], 
                    [ 68.75 , 3.945263345606827e-46 ], [ 68.8125 , 3.934332757151514e-46 ], [ 68.875 , 3.923434229890757e-46 ], [ 68.9375 , 3.912606010162264e-46 ], 
                    [ 69.0 , 3.901857289684554e-46 ], [ 69.0625 , 3.891138098117086e-46 ], [ 69.125 , 3.8805352645205734e-46 ], [ 69.1875 , 3.8699705265999215e-46 ], 
                    [ 69.25 , 3.859464671328413e-46 ], [ 69.3125 , 3.849041487024268e-46 ], [ 69.375 , 3.8386464524196486e-46 ], [ 69.4375 , 3.8283596287117835e-46 ], 
                    [ 69.5 , 3.81812015315284e-46 ], [ 69.5625 , 3.8079293303758345e-46 ], [ 69.625 , 3.797832285381656e-46 ], [ 69.6875 , 3.787762013551754e-46 ], 
                    [ 69.75 , 3.777783101277119e-46 ], [ 69.8125 , 3.76786149999796e-46 ], [ 69.875 , 3.757974041642278e-46 ], [ 69.9375 , 3.748182009690837e-46 ], 
                    [ 70.0 , 3.738415492521769e-46 ], [ 70.0625 , 3.728720519908416e-46 ], [ 70.125 , 3.719092055182776e-46 ], [ 70.1875 , 3.709488453498648e-46 ], 
                    [ 70.25 , 3.70000126672167e-46 ], [ 70.3125 , 3.690543929235791e-46 ], [ 70.375 , 3.681139467417732e-46 ], [ 70.4375 , 3.671811674995882e-46 ], 
                    [ 70.5 , 3.662507518655233e-46 ], [ 70.5625 , 3.653289877440729e-46 ], [ 70.625 , 3.6441156097358595e-46 ], [ 70.6875 , 3.634977282681191e-46 ], 
                    [ 70.75 , 3.6259319254093214e-46 ], [ 70.8125 , 3.616909076803093e-46 ], [ 70.875 , 3.607956185109398e-46 ], [ 70.9375 , 3.599064666650704e-46 ], 
                    [ 71.0 , 3.5901950606257066e-46 ], [ 71.0625 , 3.581415810766655e-46 ], [ 71.125 , 3.572662951804287e-46 ], [ 71.1875 , 3.563958637009837e-46 ], 
                    [ 71.25 , 3.5553348157148096e-46 ], [ 71.3125 , 3.546731861747772e-46 ], [ 71.375 , 3.538201152446838e-46 ], [ 71.4375 , 3.5297142275467915e-46 ], 
                    [ 71.5 , 3.521253013860617e-46 ], [ 71.5625 , 3.512902121002149e-46 ], [ 71.625 , 3.504571032858417e-46 ], [ 71.6875 , 3.4962864865381216e-46 ], 
                    [ 71.75 , 3.488058790349647e-46 ], [ 71.8125 , 3.4798504561284816e-46 ], [ 71.875 , 3.471722840725999e-46 ], [ 71.9375 , 3.4636320546792818e-46 ], 
                    [ 72.0 , 3.4555691536441206e-46 ], [ 72.0625 , 3.4475891356094645e-46 ], [ 72.125 , 3.439627545997161e-46 ], [ 72.1875 , 3.431722528708277e-46 ], 
                    [ 72.25 , 3.423879127569588e-46 ], [ 72.3125 , 3.4160536529796884e-46 ], [ 72.375 , 3.4082988455169763e-46 ], [ 72.4375 , 3.400574415245478e-46 ], 
                    [ 72.5 , 3.3928796882577985e-46 ], [ 72.5625 , 3.385278644373996e-46 ], [ 72.625 , 3.3776946290539367e-46 ], [ 72.6875 , 3.370160521925985e-46 ], 
                    [ 72.75 , 3.3626814062043356e-46 ], [ 72.8125 , 3.3552188882593708e-46 ], [ 72.875 , 3.3478230550814504e-46 ], [ 72.9375 , 3.340457565135663e-46 ], 
                    [ 73.0 , 3.3331154556288733e-46 ], [ 73.0625 , 3.3258581716044897e-46 ], [ 73.125 , 3.3186166890643665e-46 ], [ 73.1875 , 3.311418457604688e-46 ], 
                    [ 73.25 , 3.3042767032901675e-46 ], [ 73.3125 , 3.2971503516362725e-46 ], [ 73.375 , 3.2900890239999696e-46 ], [ 73.4375 , 3.2830638125518805e-46 ], 
                    [ 73.5 , 3.2760536017908636e-46 ], [ 73.5625 , 3.269131078907059e-46 ], [ 73.625 , 3.2622234167401892e-46 ], [ 73.6875 , 3.255349041111008e-46 ], 
                    [ 73.75 , 3.2485359898790232e-46 ], [ 73.8125 , 3.241737197538022e-46 ], [ 73.875 , 3.234991074597176e-46 ], [ 73.9375 , 3.2282883570487156e-46 ], 
                    [ 74.0 , 3.221599527149818e-46 ], [ 74.0625 , 3.2149805139750666e-46 ], [ 74.125 , 3.2083856105101616e-46 ], [ 74.1875 , 3.2018111515692995e-46 ], 
                    [ 74.25 , 3.1953057829216268e-46 ], [ 74.3125 , 3.1888136317372538e-46 ], [ 74.375 , 3.182361347357579e-46 ], [ 74.4375 , 3.1759666032428382e-46 ], 
                    [ 74.5 , 3.1695847089423806e-46 ], [ 74.5625 , 3.16325426698013e-46 ], [ 74.625 , 3.156959071039918e-46 ], [ 74.6875 , 3.1506764031763126e-46 ], 
                    [ 74.75 , 3.1444644602493636e-46 ], [ 74.8125 , 3.138272568851781e-46 ], [ 74.875 , 3.1321000649364836e-46 ], [ 74.9375 , 3.1259900583211103e-46 ], 
                    [ 75.0 , 3.1198919709228738e-46 ], [ 75.0625 , 3.1138309274134684e-46 ], [ 75.125 , 3.107825665535406e-46 ], [ 75.1875 , 3.1018319852657007e-46 ], 
                    [ 75.25 , 3.0958839886144025e-46 ], [ 75.3125 , 3.089970715871587e-46 ], [ 75.375 , 3.0840687377363197e-46 ], [ 75.4375 , 3.078228765905665e-46 ], 
                    [ 75.5 , 3.07241102850762e-46 ], [ 75.5625 , 3.066606745807933e-46 ], [ 75.625 , 3.060874899907267e-46 ], [ 75.6875 , 3.055153767495512e-46 ], 
                    [ 75.75 , 3.04945648822531e-46 ], [ 75.8125 , 3.0438093232468324e-46 ], [ 75.875 , 3.0381726160244886e-46 ], [ 75.9375 , 3.032575866428905e-46 ], 
                    [ 76.0 , 3.0270245782827833e-46 ], [ 76.0625 , 3.021483452058882e-46 ], [ 76.125 , 3.0159898584395173e-46 ], [ 76.1875 , 3.0105259633642796e-46 ], 
                    [ 76.25 , 3.005071966912953e-46 ], [ 76.3125 , 2.9996746950639065e-46 ], [ 76.375 , 2.9942954985538235e-46 ], [ 76.4375 , 2.9889278645424365e-46 ], 
                    [ 76.5 , 2.983624867979399e-46 ], [ 76.5625 , 2.978331280065177e-46 ], [ 76.625 , 2.9730596030109163e-46 ], [ 76.6875 , 2.967843047642998e-46 ], 
                    [ 76.75 , 2.9626356452869734e-46 ], [ 76.8125 , 2.9574587768434043e-46 ], [ 76.875 , 2.9523247593153386e-46 ], [ 76.9375 , 2.9471996542144503e-46 ], 
                    [ 77.0 , 2.9421145647144484e-46 ], [ 77.0625 , 2.9370632987374425e-46 ], [ 77.125 , 2.9320207051922156e-46 ], [ 77.1875 , 2.9270243829422865e-46 ], 
                    [ 77.25 , 2.9220516452101106e-46 ], [ 77.3125 , 2.9170873556892847e-46 ], [ 77.375 , 2.912178139285821e-46 ], [ 77.4375 , 2.9072843149074213e-46 ], 
                    [ 77.5 , 2.902399503786221e-46 ], [ 77.5625 , 2.897576835660947e-46 ], [ 77.625 , 2.892762180949358e-46 ], [ 77.6875 , 2.8879634969795e-46 ], 
                    [ 77.75 , 2.883216088356266e-46 ], [ 77.8125 , 2.8784764838096906e-46 ], [ 77.875 , 2.8737592010987847e-46 ], [ 77.9375 , 2.8690852351191883e-46 ], 
                    [ 78.0 , 2.8644188710145055e-46 ], [ 78.0625 , 2.859782942829488e-46 ], [ 78.125 , 2.855186328838514e-46 ], [ 78.1875 , 2.8505971031215563e-46 ], 
                    [ 78.25 , 2.846042081361603e-46 ], [ 78.3125 , 2.841517376379167e-46 ], [ 78.375 , 2.8369998648796566e-46 ], [ 78.4375 , 2.8325203489377196e-46 ], 
                    [ 78.5 , 2.828064384594459e-46 ], [ 78.5625 , 2.8236154301279395e-46 ], [ 78.625 , 2.8192113909973805e-46 ], [ 78.6875 , 2.8148261759764398e-46 ], 
                    [ 78.75 , 2.8104477820512484e-46 ], [ 78.8125 , 2.806120888727326e-46 ], [ 78.875 , 2.8018078727581588e-46 ], [ 78.9375 , 2.7975014859070237e-46 ], 
                    [ 79.0 , 2.7932465946166863e-46 ], [ 79.0625 , 2.7890003551427267e-46 ], [ 79.125 , 2.784762586317197e-46 ], [ 79.1875 , 2.7805776227095115e-46 ], 
                    [ 79.25 , 2.7763989482988075e-46 ], [ 79.3125 , 2.772232266002949e-46 ], [ 79.375 , 2.7681130396316842e-46 ], [ 79.4375 , 2.763999933969009e-46 ], 
                    [ 79.5 , 2.7599017860466758e-46 ], [ 79.5625 , 2.755845764862624e-46 ], [ 79.625 , 2.7517957045094716e-46 ], [ 79.6875 , 2.747765101519152e-46 ], 
                    [ 79.75 , 2.7437779985422248e-46 ], [ 79.8125 , 2.739796680990748e-46 ], [ 79.875 , 2.7358348877641637e-46 ], [ 79.9375 , 2.7319075696057028e-46 ], 
                    [ 80.0 , 2.7279858891515103e-46 ], [ 80.0625 , 2.7240866443601126e-46 ], [ 80.125 , 2.7202206350826802e-46 ], [ 80.1875 , 2.7163601124250522e-46 ], 
                    [ 80.25 , 2.7125243327531488e-46 ], [ 80.3125 , 2.708720103029647e-46 ], [ 80.375 , 2.7049212086182073e-46 ], [ 80.4375 , 2.7011482201019534e-46 ], 
                    [ 80.5 , 2.6974042326305677e-46 ], [ 80.5625 , 2.6936654345975986e-46 ], [ 80.625 , 2.689953836714818e-46 ], [ 80.6875 , 2.6862695528853397e-46 ], 
                    [ 80.75 , 2.682590315219574e-46 ], [ 80.8125 , 2.6789386678925927e-46 ], [ 80.875 , 2.6753123834327255e-46 ], [ 80.9375 , 2.6716910076104628e-46 ], 
                    [ 81.0 , 2.668097305622483e-46 ], [ 81.0625 , 2.664527500172021e-46 ], [ 81.125 , 2.660962470975756e-46 ], [ 81.1875 , 2.6574259743826167e-46 ], 
                    [ 81.25 , 2.6539130675042736e-46 ], [ 81.3125 , 2.6504048044108773e-46 ], [ 81.375 , 2.6469241866946306e-46 ], [ 81.4375 , 2.6434660768682935e-46 ], 
                    [ 81.5 , 2.640012484936172e-46 ], [ 81.5625 , 2.6365860344312922e-46 ], [ 81.625 , 2.633181673439046e-46 ], [ 81.6875 , 2.6297817081591395e-46 ], 
                    [ 81.75 , 2.6264107054188352e-46 ], [ 81.8125 , 2.6230638128322346e-46 ], [ 81.875 , 2.6197211852640347e-46 ], [ 81.9375 , 2.6164019380863626e-46 ], 
                    [ 82.0 , 2.613103321915804e-46 ], [ 82.0625 , 2.609808864459768e-46 ], [ 82.125 , 2.606541846546257e-46 ], [ 82.1875 , 2.6033006634089896e-46 ], 
                    [ 82.25 , 2.6000635106187288e-46 ], [ 82.3125 , 2.596847395843302e-46 ], [ 82.375 , 2.5936530866090665e-46 ], [ 82.4375 , 2.5904627066051537e-46 ], 
                    [ 82.5 , 2.587295269210422e-46 ], [ 82.5625 , 2.5841548087222014e-46 ], [ 82.625 , 2.581018160126731e-46 ], [ 82.6875 , 2.5779003936784985e-46 ], 
                    [ 82.75 , 2.5748082404587892e-46 ], [ 82.8125 , 2.5717197962309606e-46 ], [ 82.875 , 2.568648837486009e-46 ], [ 82.9375 , 2.5656061869226104e-46 ], 
                    [ 83.0 , 2.5625671404806736e-46 ], [ 83.0625 , 2.5595421976430274e-46 ], [ 83.125 , 2.5565452125339603e-46 ], [ 83.1875 , 2.553551736614838e-46 ], 
                    [ 83.25 , 2.550571153186004e-46 ], [ 83.3125 , 2.5476242690713116e-46 ], [ 83.375 , 2.5446807897336683e-46 ], [ 83.4375 , 2.5417467474052895e-46 ], 
                    [ 83.5 , 2.5388462922776317e-46 ], [ 83.5625 , 2.5359491469368187e-46 ], [ 83.625 , 2.5330585164666716e-46 ], [ 83.6875 , 2.5302043987125724e-46 ], 
                    [ 83.75 , 2.527353496829026e-46 ], [ 83.8125 , 2.52450591305474e-46 ], [ 83.875 , 2.5216944768302063e-46 ], [ 83.9375 , 2.5188861715840823e-46 ], 
                    [ 84.0 , 2.5160809938296185e-46 ], [ 84.0625 , 2.5133145957437355e-46 ], [ 84.125 , 2.5105547446139545e-46 ], [ 84.1875 , 2.5077979240551283e-46 ], 
                    [ 84.25 , 2.5050707650396394e-46 ], [ 84.3125 , 2.5023527639416764e-46 ], [ 84.375 , 2.4996377118741296e-46 ], [ 84.4375 , 2.496950347733809e-46 ], 
                    [ 84.5 , 2.4942762645983917e-46 ], [ 84.5625 , 2.4916050452446526e-46 ], [ 84.625 , 2.4889568696863434e-46 ], [ 84.6875 , 2.4863254962807644e-46 ], 
                    [ 84.75 , 2.483696904814146e-46 ], [ 84.8125 , 2.4810868040541616e-46 ], [ 84.875 , 2.4784971167943635e-46 ], [ 84.9375 , 2.4759101325758485e-46 ], 
                    [ 85.0 , 2.473337937128563e-46 ], [ 85.0625 , 2.4707916067493757e-46 ], [ 85.125 , 2.4682478978471235e-46 ], [ 85.1875 , 2.465712792782351e-46 ], 
                    [ 85.25 , 2.463203280311567e-46 ], [ 85.3125 , 2.460696321930967e-46 ], [ 85.375 , 2.4581943440449646e-46 ], [ 85.4375 , 2.455730398446458e-46 ], 
                    [ 85.5 , 2.453268922558295e-46 ], [ 85.5625 , 2.4508099139049346e-46 ], [ 85.625 , 2.4483789152809016e-46 ], [ 85.6875 , 2.44595274146149e-46 ], 
                    [ 85.75 , 2.4435289718121343e-46 ], [ 85.8125 , 2.441132436367471e-46 ], [ 85.875 , 2.4387462235688466e-46 ], [ 85.9375 , 2.436362343298978e-46 ], 
                    [ 86.0 , 2.43400012777054e-46 ], [ 86.0625 , 2.431653514860492e-46 ], [ 86.125 , 2.429309164313627e-46 ], [ 86.1875 , 2.4269795978572053e-46 ], 
                    [ 86.25 , 2.4246694871263207e-46 ], [ 86.3125 , 2.4223615752650083e-46 ], [ 86.375 , 2.420063024661106e-46 ], [ 86.4375 , 2.4177891237495833e-46 ], 
                    [ 86.5 , 2.4155173594044995e-46 ], [ 86.5625 , 2.4132495433645235e-46 ], [ 86.625 , 2.411014402550724e-46 ], [ 86.6875 , 2.4087813319142212e-46 ], 
                    [ 86.75 , 2.406550329537592e-46 ], [ 86.8125 , 2.40434775982522e-46 ], [ 86.875 , 2.4021514090389586e-46 ], [ 86.9375 , 2.3999570646000214e-46 ], 
                    [ 87.0 , 2.397784696222875e-46 ], [ 87.0625 , 2.395624536731228e-46 ], [ 87.125 , 2.3934663233229796e-46 ], [ 87.1875 , 2.3913225447464712e-46 ], 
                    [ 87.25 , 2.3891957674037885e-46 ], [ 87.3125 , 2.3870708815591454e-46 ], [ 87.375 , 2.384954784987989e-46 ], [ 87.4375 , 2.382862781488402e-46 ], 
                    [ 87.5 , 2.380772613025148e-46 ], [ 87.5625 , 2.3786846405532402e-46 ], [ 87.625 , 2.3766264377084278e-46 ], [ 87.6875 , 2.374570015763394e-46 ], 
                    [ 87.75 , 2.3725153731771984e-46 ], [ 87.8125 , 2.370485136947249e-46 ], [ 87.875 , 2.3684629261246128e-46 ], [ 87.9375 , 2.366442440407318e-46 ], 
                    [ 88.0 , 2.3644389948240454e-46 ], [ 88.0625 , 2.362450073408923e-46 ], [ 88.125 , 2.3604628250369044e-46 ], [ 88.1875 , 2.358485891126245e-46 ], 
                    [ 88.25 , 2.356530839729513e-46 ], [ 88.3125 , 2.3545774089597853e-46 ], [ 88.375 , 2.352626635623759e-46 ], [ 88.4375 , 2.3506995951212964e-46 ], 
                    [ 88.5 , 2.3487741330609728e-46 ], [ 88.5625 , 2.3468502481500003e-46 ], [ 88.625 , 2.3449513269899925e-46 ], [ 88.6875 , 2.343060320038193e-46 ], 
                    [ 88.75 , 2.341170838025189e-46 ], [ 88.8125 , 2.33929667759757e-46 ], [ 88.875 , 2.337436908024826e-46 ], [ 88.9375 , 2.3355786169917364e-46 ], 
                    [ 89.0 , 2.3337282607908444e-46 ], [ 89.0625 , 2.3318996250492694e-46 ], [ 89.125 , 2.330072422168927e-46 ], [ 89.1875 , 2.3282466510270397e-46 ], 
                    [ 89.25 , 2.32644781783878e-46 ], [ 89.3125 , 2.3246514807662023e-46 ], [ 89.375 , 2.322856530712465e-46 ], [ 89.4375 , 2.3210795630739353e-46 ], 
                    [ 89.5 , 2.319312226847488e-46 ], [ 89.5625 , 2.3175462363213025e-46 ], [ 89.625 , 2.3157909708798293e-46 ], [ 89.6875 , 2.3140530222986604e-46 ], 
                    [ 89.75 , 2.3123163780084694e-46 ], [ 89.8125 , 2.3105826290176136e-46 ], [ 89.875 , 2.308874082796706e-46 ], [ 89.9375 , 2.307166799949938e-46 ], 
                    [ 90.0 , 2.3054607795429674e-46 ], [ 90.0625 , 2.3037744806643303e-46 ], [ 90.125 , 2.3020956034953452e-46 ], [ 90.1875 , 2.3004179498092187e-46 ], 
                    [ 90.25 , 2.2987519341795436e-46 ], [ 90.3125 , 2.2971009894743133e-46 ], [ 90.375 , 2.2954512304639558e-46 ], [ 90.4375 , 2.2938051644898758e-46 ], 
                    [ 90.5 , 2.2921830387108876e-46 ], [ 90.5625 , 2.290562060061654e-46 ], [ 90.625 , 2.2889422277308856e-46 ], [ 90.6875 , 2.2873419160467996e-46 ], 
                    [ 90.75 , 2.2857484205639466e-46 ], [ 90.8125 , 2.284156035202742e-46 ], [ 90.875 , 2.2825746151078117e-46 ], [ 90.9375 , 2.2810080258709576e-46 ], 
                    [ 91.0 , 2.2794425118243107e-46 ], [ 91.0625 , 2.277879449223615e-46 ], [ 91.125 , 2.276336559411271e-46 ], [ 91.1875 , 2.2747947146537474e-46 ], 
                    [ 91.25 , 2.2732539142432402e-46 ], [ 91.3125 , 2.2717341827280386e-46 ], [ 91.375 , 2.270223752702699e-46 ], [ 91.4375 , 2.2687143269316603e-46 ], 
                    [ 91.5 , 2.2672125683189257e-46 ], [ 91.5625 , 2.2657247766393296e-46 ], [ 91.625 , 2.2642379612793717e-46 ], [ 91.6875 , 2.262752121598307e-46 ], 
                    [ 91.75 , 2.2612892779671702e-46 ], [ 91.8125 , 2.2598281829525106e-46 ], [ 91.875 , 2.258368032000502e-46 ], [ 91.9375 , 2.256922750209675e-46 ], 
                    [ 92.0 , 2.25548843803536e-46 ], [ 92.0625 , 2.2540550373905127e-46 ], [ 92.125 , 2.2526267331160332e-46 ], [ 92.1875 , 2.2512171195397273e-46 ], 
                    [ 92.25 , 2.2498083880493744e-46 ], [ 92.3125 , 2.248400538093044e-46 ], [ 92.375 , 2.2470094448533357e-46 ], [ 92.4375 , 2.245623379111559e-46 ], 
                    [ 92.5 , 2.2442381683631807e-46 ], [ 92.5625 , 2.2428629345010767e-46 ], [ 92.625 , 2.241503069839005e-46 ], [ 92.6875 , 2.2401440296730558e-46 ], 
                    [ 92.75 , 2.238785813503394e-46 ], [ 92.8125 , 2.2374477388906077e-46 ], [ 92.875 , 2.2361110365351737e-46 ], [ 92.9375 , 2.2347751327561777e-46 ], 
                    [ 93.0 , 2.2334519096618144e-46 ], [ 93.0625 , 2.232139233674005e-46 ], [ 93.125 , 2.230827329190718e-46 ], [ 93.1875 , 2.2295187505876654e-46 ], 
                    [ 93.25 , 2.2282299516354382e-46 ], [ 93.3125 , 2.226941897688282e-46 ], [ 93.375 , 2.2256545883154905e-46 ], [ 93.4375 , 2.224381739408429e-46 ], 
                    [ 93.5 , 2.2231160695193973e-46 ], [ 93.5625 , 2.2218511197946435e-46 ], [ 93.625 , 2.220591637374309e-46 ], [ 93.6875 , 2.2193482978369006e-46 ], 
                    [ 93.75 , 2.218105654462284e-46 ], [ 93.8125 , 2.2168637068606066e-46 ], [ 93.875 , 2.2156385430329042e-46 ], [ 93.9375 , 2.214418570442477e-46 ], 
                    [ 94.0 , 2.2131992695921047e-46 ], [ 94.0625 , 2.211986425546989e-46 ], [ 94.125 , 2.210786624311566e-46 ], [ 94.1875 , 2.209587473859088e-46 ], 
                    [ 94.25 , 2.2083889738365167e-46 ], [ 94.3125 , 2.2072095845226974e-46 ], [ 94.375 , 2.206034091315858e-46 ], [ 94.4375 , 2.2048592241412216e-46 ], 
                    [ 94.5 , 2.2036926204007453e-46 ], [ 94.5625 , 2.2025395651920965e-46 ], [ 94.625 , 2.2013871133054817e-46 ], [ 94.6875 , 2.2002352644252198e-46 ], 
                    [ 94.75 , 2.1990980252402885e-46 ], [ 94.8125 , 2.197963168062147e-46 ], [ 94.875 , 2.19682889653356e-46 ], [ 94.9375 , 2.195704407028157e-46 ], 
                    [ 95.0 , 2.194594750121201e-46 ], [ 95.0625 , 2.1934856540085e-46 ], [ 95.125 , 2.192377118406657e-46 ], [ 95.1875 , 2.191282624404799e-46 ], 
                    [ 95.25 , 2.1901903490225146e-46 ], [ 95.3125 , 2.1890986181001425e-46 ], [ 95.375 , 2.1880151341475947e-46 ], [ 95.4375 , 2.18694479073787e-46 ], 
                    [ 95.5 , 2.1858749709237535e-46 ], [ 95.5625 , 2.1848056744490492e-46 ], [ 95.625 , 2.1837524333636638e-46 ], [ 95.6875 , 2.1827021770882646e-46 ], 
                    [ 95.75 , 2.181652425924279e-46 ], [ 95.8125 , 2.1806092839956924e-46 ], [ 95.875 , 2.179578698512149e-46 ], [ 95.9375 , 2.1785486000972486e-46 ], 
                    [ 96.0 , 2.177518988520827e-46 ], [ 96.0625 , 2.1765038065783725e-46 ], [ 96.125 , 2.1754925472975705e-46 ], [ 96.1875 , 2.174481757873685e-46 ], 
                    [ 96.25 , 2.1734759310192357e-46 ], [ 96.3125 , 2.1724834044381294e-46 ], [ 96.375 , 2.1714913310983195e-46 ], [ 96.4375 , 2.1704997107927846e-46 ], 
                    [ 96.5 , 2.1695215850451567e-46 ], [ 96.5625 , 2.168549255695652e-46 ], [ 96.625 , 2.167577362121683e-46 ], [ 96.6875 , 2.1666087650732394e-46 ], 
                    [ 96.75 , 2.1656569594465166e-46 ], [ 96.8125 , 2.1647055719542997e-46 ], [ 96.875 , 2.1637546024129446e-46 ], [ 96.9375 , 2.1628128488386e-46 ], 
                    [ 97.0 , 2.1618778110290365e-46 ], [ 97.0625 , 2.1609431774596298e-46 ], [ 97.125 , 2.160009054590752e-46 ], [ 97.1875 , 2.1590918835458977e-46 ], 
                    [ 97.25 , 2.1581751019451017e-46 ], [ 97.3125 , 2.157258709623001e-46 ], [ 97.375 , 2.1563500900952046e-46 ], [ 97.4375 , 2.1554518797468753e-46 ], 
                    [ 97.5 , 2.154554043540883e-46 ], [ 97.5625 , 2.1536565813213667e-46 ], [ 97.625 , 2.152773587208884e-46 ], [ 97.6875 , 2.1518937999258637e-46 ], 
                    [ 97.75 , 2.151014372190933e-46 ], [ 97.8125 , 2.150138904741494e-46 ], [ 97.875 , 2.1492757089694133e-46 ], [ 97.9375 , 2.148412859736312e-46 ], 
                    [ 98.0 , 2.1475503569030527e-46 ], [ 98.0625 , 2.1466995119943565e-46 ], [ 98.125 , 2.1458558640980143e-46 ], [ 98.1875 , 2.1450125477533105e-46 ], 
                    [ 98.25 , 2.1441696569865408e-46 ], [ 98.3125 , 2.1433417095486315e-46 ], [ 98.375 , 2.142514081813468e-46 ], [ 98.4375 , 2.1416867736576164e-46 ], 
                    [ 98.5 , 2.140865087097678e-46 ], [ 98.5625 , 2.140052334055893e-46 ], [ 98.625 , 2.139239889565827e-46 ], [ 98.6875 , 2.138427753510267e-46 ], 
                    [ 98.75 , 2.1376289036498958e-46 ], [ 98.8125 , 2.1368347042959295e-46 ], [ 98.875 , 2.136040800013066e-46 ], [ 98.9375 , 2.135248832767725e-46 ], 
                    [ 99.0 , 2.13447106193538e-46 ], [ 99.0625 , 2.1336935744083935e-46 ], [ 99.125 , 2.132916370083585e-46 ], [ 99.1875 , 2.1321463432477253e-46 ], 
                    [ 99.25 , 2.13138481635349e-46 ], [ 99.3125 , 2.130623561449581e-46 ], [ 99.375 , 2.129862578438883e-46 ], [ 99.4375 , 2.1291141748132167e-46 ], 
                    [ 99.5 , 2.1283691717568176e-46 ], [ 99.5625 , 2.127624429386119e-46 ], [ 99.625 , 2.1268821810986888e-46 ], [ 99.6875 , 2.1261549767073878e-46 ], 
                    [ 99.75 , 2.125428020955275e-46 ], [ 99.8125 , 2.1247013137573227e-46 ], [ 99.875 , 2.1239799545830778e-46 ], [ 99.9375 , 2.1232649613121682e-46 ], 
                    [ 100.0 , 2.122550208728721e-46 ], [ 100.0625 , 2.1218356967517136e-46 ], [ 100.125 , 2.121134134561289e-46 ], [ 100.1875 , 2.120436642336335e-46 ], 
                    [ 100.25 , 2.119739379467692e-46 ], [ 100.3125 , 2.119043357706811e-46 ], [ 100.375 , 2.1183606015726093e-46 ], [ 100.4375 , 2.1176780654225367e-46 ], 
                    [ 100.5 , 2.116995749185744e-46 ], [ 100.5625 , 2.1163196313251293e-46 ], [ 100.625 , 2.1156536480907634e-46 ], [ 100.6875 , 2.1149878744342546e-46 ], 
                    [ 100.75 , 2.1143223102895896e-46 ], [ 100.8125 , 2.1136642653096596e-46 ], [ 100.875 , 2.113010122886672e-46 ], [ 100.9375 , 2.1123561829094313e-46 ], 
                    [ 101.0 , 2.1117024453152242e-46 ], [ 101.0625 , 2.111063259850308e-46 ], [ 101.125 , 2.1104249911619935e-46 ], [ 101.1875 , 2.1097869154508638e-46 ], 
                    [ 101.25 , 2.109151783364724e-46 ], [ 101.3125 , 2.1085263688708187e-46 ], [ 101.375 , 2.1079011398274157e-46 ], [ 101.4375 , 2.107276096179539e-46 ], 
                    [ 101.5 , 2.1066591460483052e-46 ], [ 101.5625 , 2.1060504099474947e-46 ], [ 101.625 , 2.10544184974586e-46 ], [ 101.6875 , 2.1048334653926045e-46 ], 
                    [ 101.75 , 2.1042368555869884e-46 ], [ 101.8125 , 2.1036442389384567e-46 ], [ 101.875 , 2.103051789188641e-46 ], [ 101.9375 , 2.1024595062905675e-46 ], 
                    [ 102.0 , 2.101875920152238e-46 ], [ 102.0625 , 2.1012925647727092e-46 ], [ 102.125 , 2.1007093712978515e-46 ], [ 102.1875 , 2.1001302686877558e-46 ], 
                    [ 102.25 , 2.0995643285475173e-46 ], [ 102.3125 , 2.0989985409160587e-46 ], [ 102.375 , 2.0984329057522076e-46 ], [ 102.4375 , 2.0978726801675195e-46 ], 
                    [ 102.5 , 2.097318874181638e-46 ], [ 102.5625 , 2.0967652143920285e-46 ], [ 102.625 , 2.096211700760037e-46 ], [ 102.6875 , 2.0956677322518213e-46 ], 
                    [ 102.75 , 2.0951283268316626e-46 ], [ 102.8125 , 2.094589060249452e-46 ], [ 102.875 , 2.0940499324694384e-46 ], [ 102.9375 , 2.0935211507776807e-46 ], 
                    [ 103.0 , 2.0929937897002806e-46 ], [ 103.0625 , 2.0924665614659466e-46 ], [ 103.125 , 2.091940838968864e-46 ], [ 103.1875 , 2.0914281737555557e-46 ], 
                    [ 103.25 , 2.0909156341794178e-46 ], [ 103.3125 , 2.09040322020978e-46 ], [ 103.375 , 2.0898940518822285e-46 ], [ 103.4375 , 2.0893927192138938e-46 ], 
                    [ 103.5 , 2.0888915068073413e-46 ], [ 103.5625 , 2.088390414633781e-46 ], [ 103.625 , 2.0878962322262933e-46 ], [ 103.6875 , 2.0874095204019507e-46 ], 
                    [ 103.75 , 2.08692292203556e-46 ], [ 103.8125 , 2.0864364371006738e-46 ], [ 103.875 , 2.085955656741856e-46 ], [ 103.9375 , 2.085477918455835e-46 ], 
                    [ 104.0 , 2.0850002895843204e-46 ], [ 104.0625 , 2.0845227701023422e-46 ], [ 104.125 , 2.0840570049088873e-46 ], [ 104.1875 , 2.0835936946473106e-46 ], 
                    [ 104.25 , 2.0831304873850122e-46 ], [ 104.3125 , 2.0826673830991384e-46 ], [ 104.375 , 2.0822197832350936e-46 ], [ 104.4375 , 2.0817724037468425e-46 ], 
                    [ 104.5 , 2.081325120381219e-46 ], [ 104.5625 , 2.0808787783154496e-46 ], [ 104.625 , 2.08043790031698e-46 ], [ 104.6875 , 2.0799971157278013e-46 ], 
                    [ 104.75 , 2.0795564245281067e-46 ], [ 104.8125 , 2.079120305949778e-46 ], [ 104.875 , 2.0786957945374335e-46 ], [ 104.9375 , 2.0782713698011114e-46 ], 
                    [ 105.0 , 2.0778470317231726e-46 ], [ 105.0625 , 2.0774262283267857e-46 ], [ 105.125 , 2.077010512187829e-46 ], [ 105.1875 , 2.07659487923832e-46 ], 
                    [ 105.25 , 2.0761793294615955e-46 ], [ 105.3125 , 2.075770822466916e-46 ], [ 105.375 , 2.0753684194339043e-46 ], [ 105.4375 , 2.0749660944096005e-46 ], 
                    [ 105.5 , 2.074563847378898e-46 ], [ 105.5625 , 2.0741689884236623e-46 ], [ 105.625 , 2.0737781747951846e-46 ], [ 105.6875 , 2.073387434803582e-46 ], 
                    [ 105.75 , 2.072996768434921e-46 ], [ 105.8125 , 2.072614237701181e-46 ], [ 105.875 , 2.0722343229955578e-46 ], [ 105.9375 , 2.071854477929143e-46 ], 
                    [ 106.0 , 2.0714747024891265e-46 ], [ 106.0625 , 2.0711040999592747e-46 ], [ 106.125 , 2.0707350948586644e-46 ], [ 106.1875 , 2.0703661555030663e-46 ], 
                    [ 106.25 , 2.069997281880752e-46 ], [ 106.3125 , 2.0696384547774128e-46 ], [ 106.375 , 2.0692801930230674e-46 ], [ 106.4375 , 2.0689219932851313e-46 ], 
                    [ 106.5 , 2.0685642593294193e-46 ], [ 106.5625 , 2.0682187993970127e-46 ], [ 106.625 , 2.0678733971580205e-46 ], [ 106.6875 , 2.0675280526027924e-46 ], 
                    [ 106.75 , 2.0671838378509753e-46 ], [ 106.8125 , 2.066849774636234e-46 ], [ 106.875 , 2.0665157654071124e-46 ], [ 106.9375 , 2.0661818101548866e-46 ], 
                    [ 107.0 , 2.065849456558219e-46 ], [ 107.0625 , 2.06552528049254e-46 ], [ 107.125 , 2.0652011552970788e-46 ], [ 107.1875 , 2.0648770809637944e-46 ], 
                    [ 107.25 , 2.064554913485277e-46 ], [ 107.3125 , 2.0642398622638633e-46 ], [ 107.375 , 2.0639248591192876e-46 ], [ 107.4375 , 2.0636099040442432e-46 ], 
                    [ 107.5 , 2.063298008003839e-46 ], [ 107.5625 , 2.0629949084448916e-46 ], [ 107.625 , 2.0626918534113988e-46 ], [ 107.6875 , 2.0623888428969077e-46 ], 
                    [ 107.75 , 2.062088438471868e-46 ], [ 107.8125 , 2.0617944096822867e-46 ], [ 107.875 , 2.061500422817672e-46 ], [ 107.9375 , 2.0612064778720026e-46 ], 
                    [ 108.0 , 2.0609164889252527e-46 ], [ 108.0625 , 2.0606348570638693e-46 ], [ 108.125 , 2.060353263688507e-46 ], [ 108.1875 , 2.0600717087939645e-46 ], 
                    [ 108.25 , 2.0597933392584607e-46 ], [ 108.3125 , 2.0595212253531068e-46 ], [ 108.375 , 2.0592491473960092e-46 ], [ 108.4375 , 2.0589771053824474e-46 ], 
                    [ 108.5 , 2.0587078686501694e-46 ], [ 108.5625 , 2.0584441391676886e-46 ], [ 108.625 , 2.058180443470064e-46 ], [ 108.6875 , 2.0579167815530258e-46 ], 
                    [ 108.75 , 2.057657193163169e-46 ], [ 108.8125 , 2.057405619851772e-46 ], [ 108.875 , 2.057154077298234e-46 ], [ 108.9375 , 2.05690256549878e-46 ], 
                    [ 109.0 , 2.056653713722739e-46 ], [ 109.0625 , 2.0564104785941566e-46 ], [ 109.125 , 2.056167272232336e-46 ], [ 109.1875 , 2.055924094633919e-46 ], 
                    [ 109.25 , 2.055683905174149e-46 ], [ 109.3125 , 2.055450518419267e-46 ], [ 109.375 , 2.05521715816139e-46 ], [ 109.4375 , 2.054983824397393e-46 ], 
                    [ 109.5 , 2.054752744240075e-46 ], [ 109.5625 , 2.0545276502187363e-46 ], [ 109.625 , 2.0543025808559067e-46 ], [ 109.6875 , 2.0540775361490455e-46 ], 
                    [ 109.75 , 2.0538552054251622e-46 ], [ 109.8125 , 2.0536414140785618e-46 ], [ 109.875 , 2.0534276449860964e-46 ], [ 109.9375 , 2.0532138981454483e-46 ], 
                    [ 110.0 , 2.0530020071162e-46 ], [ 110.0625 , 2.052797853415629e-46 ], [ 110.125 , 2.0525937200163604e-46 ], [ 110.1875 , 2.0523896069165057e-46 ], 
                    [ 110.25 , 2.052186554990239e-46 ], [ 110.3125 , 2.0519906134349656e-46 ], [ 110.375 , 2.051794690588105e-46 ], [ 110.4375 , 2.0515987864477977e-46 ], 
                    [ 110.5 , 2.051403468688838e-46 ], [ 110.5625 , 2.05121647102074e-46 ], [ 110.625 , 2.051029490398639e-46 ], [ 110.6875 , 2.0508425268209227e-46 ], 
                    [ 110.75 , 2.0506555802860236e-46 ], [ 110.8125 , 2.0504773113590682e-46 ], [ 110.875 , 2.050299198739783e-46 ], [ 110.9375 , 2.0501211015920695e-46 ], 
                    [ 111.0 , 2.0499430199145825e-46 ], [ 111.0625 , 2.049773944466775e-46 ], [ 111.125 , 2.049606016843089e-46 ], [ 111.1875 , 2.0494381029768935e-46 ], 
                    [ 111.25 , 2.049270202867033e-46 ], [ 111.3125 , 2.0491081537311818e-46 ], [ 111.375 , 2.0489475043643185e-46 ], [ 111.4375 , 2.0487868675922804e-46 ], 
                    [ 111.5 , 2.0486262434141095e-46 ], [ 111.5625 , 2.048472712143535e-46 ], [ 111.625 , 2.0483222847063197e-46 ], [ 111.6875 , 2.0481718683155557e-46 ], 
                    [ 111.75 , 2.048021462970476e-46 ], [ 111.8125 , 2.047876091280122e-46 ], [ 111.875 , 2.04773430741153e-46 ], [ 111.9375 , 2.0475925333592713e-46 ], 
                    [ 112.0 , 2.0474507691226813e-46 ], [ 112.0625 , 2.0473117041262656e-46 ], [ 112.125 , 2.0471758567990543e-46 ], [ 112.1875 , 2.0470400184858708e-46 ], 
                    [ 112.25 , 2.0469041891860894e-46 ], [ 112.3125 , 2.0467720389068394e-46 ], [ 112.375 , 2.0466474157346122e-46 ], [ 112.4375 , 2.0465228001503993e-46 ], 
                    [ 112.5 , 2.0463981921537385e-46 ], [ 112.5625 , 2.0462750411759526e-46 ], [ 112.625 , 2.046158325149594e-46 ], [ 112.6875 , 2.0460416157804744e-46 ], 
                    [ 112.75 , 2.045924913068301e-46 ], [ 112.8125 , 2.045808421556359e-46 ], [ 112.875 , 2.0457002543332967e-46 ], [ 112.9375 , 2.045592092829288e-46 ], 
                    [ 113.0 , 2.0454839370440895e-46 ], [ 113.0625 , 2.04537578697734e-46 ], [ 113.125 , 2.0452725568927044e-46 ], [ 113.1875 , 2.0451701055216637e-46 ], 
                    [ 113.25 , 2.045067659282582e-46 ], [ 113.3125 , 2.0449652181752154e-46 ], [ 113.375 , 2.0448700674125946e-46 ], [ 113.4375 , 2.0447782246915833e-46 ], 
                    [ 113.5 , 2.044686386095599e-46 ], [ 113.5625 , 2.0445945516243985e-46 ], [ 113.625 , 2.044505591234174e-46 ], [ 113.6875 , 2.0444193700265072e-46 ], 
                    [ 113.75 , 2.0443331524549606e-46 ], [ 113.8125 , 2.0442469385193947e-46 ], [ 113.875 , 2.0441641967913824e-46 ], [ 113.9375 , 2.044088828848466e-46 ], 
                    [ 114.0 , 2.044013463684395e-46 ], [ 114.0625 , 2.043938101298994e-46 ], [ 114.125 , 2.043863328138674e-46 ], [ 114.1875 , 2.0437932072324255e-46 ], 
                    [ 114.25 , 2.043723088731887e-46 ], [ 114.3125 , 2.043652972636975e-46 ], [ 114.375 , 2.0435828589476227e-46 ], [ 114.4375 , 2.043519866139738e-46 ], 
                    [ 114.5 , 2.0434576311961234e-46 ], [ 114.5625 , 2.0433953981478454e-46 ], [ 114.625 , 2.0433331669948174e-46 ], [ 114.6875 , 2.043277086406052e-46 ], 
                    [ 114.75 , 2.0432236929140397e-46 ], [ 114.8125 , 2.0431703008172977e-46 ], [ 114.875 , 2.043116910115703e-46 ], [ 114.9375 , 2.043066834959068e-46 ], 
                    [ 115.0 , 2.0430207146623e-46 ], [ 115.0625 , 2.0429745954066692e-46 ], [ 115.125 , 2.0429284771921367e-46 ], [ 115.1875 , 2.0428836008586534e-46 ], 
                    [ 115.25 , 2.0428432292078608e-46 ], [ 115.3125 , 2.0428028583548965e-46 ], [ 115.375 , 2.0427624882997158e-46 ], [ 115.4375 , 2.0427221190423602e-46 ], 
                    [ 115.5 , 2.0426916259058252e-46 ], [ 115.5625 , 2.0426613760572915e-46 ], [ 115.625 , 2.0426311266567076e-46 ], [ 115.6875 , 2.0426008777040384e-46 ], 
                    [ 115.75 , 2.0425741938058873e-46 ], [ 115.8125 , 2.04254900899357e-46 ], [ 115.875 , 2.0425238244918086e-46 ], [ 115.9375 , 2.0424986403005565e-46 ], 
                    [ 116.0 , 2.0424772463896674e-46 ], [ 116.0625 , 2.0424605224479224e-46 ], [ 116.125 , 2.0424437986430997e-46 ], [ 116.1875 , 2.0424270749751978e-46 ], 
                    [ 116.25 , 2.0424111733092433e-46 ], [ 116.3125 , 2.042399586530273e-46 ], [ 116.375 , 2.04238799981705e-46 ], [ 116.4375 , 2.042376413169516e-46 ], 
                    [ 116.5 , 2.0423648265877437e-46 ], [ 116.5625 , 2.0423617769649004e-46 ], [ 116.625 , 2.0423599804697156e-46 ], [ 116.6875 , 2.0423581839760815e-46 ], 
                    [ 116.75 , 2.0423563874840713e-46 ], [ 116.8125 , 2.0423578037930778e-46 ], [ 116.875 , 2.0423616636457746e-46 ], [ 116.9375 , 2.0423655235057516e-46 ], 
                    [ 117.0 , 2.042369383373023e-46 ], [ 117.0625 , 2.042375123604272e-46 ], [ 117.125 , 2.042386106082008e-46 ], [ 117.1875 , 2.042397088618829e-46 ], 
                    [ 117.25 , 2.0424080712146776e-46 ], [ 117.3125 , 2.0424190538695686e-46 ], [ 117.375 , 2.0424353543131396e-46 ], [ 117.4375 , 2.0424519703483177e-46 ], 
                    [ 117.5 , 2.0424685865186446e-46 ], [ 117.5625 , 2.0424852028241365e-46 ], [ 117.625 , 2.04250655113563e-46 ], [ 117.6875 , 2.042530750960841e-46 ], 
                    [ 117.75 , 2.042554951072774e-46 ], [ 117.8125 , 2.0425791514714183e-46 ], [ 117.875 , 2.042605366924526e-46 ], [ 117.9375 , 2.042636563740224e-46 ], 
                    [ 118.0 , 2.042667761032364e-46 ], [ 118.0625 , 2.042698958800996e-46 ], [ 118.125 , 2.0427301570460844e-46 ], [ 118.1875 , 2.0427688482586667e-46 ], 
                    [ 118.25 , 2.0428080525218713e-46 ], [ 118.3125 , 2.042847257537459e-46 ], [ 118.375 , 2.0428864633055027e-46 ], [ 118.4375 , 2.042928230348003e-46 ], 
                    [ 118.5 , 2.042971945725737e-46 ], [ 118.5625 , 2.0430156620389098e-46 ], [ 118.625 , 2.043059379287556e-46 ], [ 118.6875 , 2.0431043164424077e-46 ], 
                    [ 118.75 , 2.0431541305832105e-46 ], [ 118.8125 , 2.0432039459385615e-46 ], [ 118.875 , 2.0432537625085047e-46 ], [ 118.9375 , 2.0433035802930413e-46 ], 
                    [ 119.0 , 2.0433605986865167e-46 ], [ 119.0625 , 2.0434190724386104e-46 ], [ 119.125 , 2.0434775478639866e-46 ], [ 119.1875 , 2.0435360249627227e-46 ], 
                    [ 119.25 , 2.04359579458341e-46 ], [ 119.3125 , 2.0436572632298663e-46 ], [ 119.375 , 2.043718733725218e-46 ], [ 119.4375 , 2.0437802060694912e-46 ], 
                    [ 119.5 , 2.0438420024016218e-46 ], [ 119.5625 , 2.043913545619755e-46 ], [ 119.625 , 2.0439850913421924e-46 ], [ 119.6875 , 2.0440566395690796e-46 ], 
                    [ 119.75 , 2.044128190300432e-46 ], [ 119.8125 , 2.0442023947279592e-46 ], [ 119.875 , 2.0442781456386504e-46 ], [ 119.9375 , 2.0443538993563878e-46 ], 
                    [ 120.0 , 2.0444296558812904e-46 ], [ 120.0625 , 2.044506651952743e-46 ], [ 120.125 , 2.0445881400049794e-46 ], [ 120.1875 , 2.0446696313050763e-46 ], 
                    [ 120.25 , 2.0447511258532068e-46 ], [ 120.3125 , 2.0448326236494285e-46 ], [ 120.375 , 2.044918920893097e-46 ], [ 120.4375 , 2.045006543238286e-46 ], 
                    [ 120.5 , 2.0450941693379883e-46 ], [ 120.5625 , 2.04518179919235e-46 ], [ 120.625 , 2.045272109843882e-46 ], [ 120.6875 , 2.0453677150727767e-46 ], 
                    [ 120.75 , 2.045463324770676e-46 ], [ 120.8125 , 2.045558938937803e-46 ], [ 120.875 , 2.0456545575743817e-46 ], [ 120.9375 , 2.0457542433447477e-46 ], 
                    [ 121.0 , 2.045854446406473e-46 ], [ 121.0625 , 2.0459546543762577e-46 ], [ 121.125 , 2.0460548672543274e-46 ], [ 121.1875 , 2.046157169353711e-46 ], 
                    [ 121.25 , 2.0462623082324303e-46 ], [ 121.3125 , 2.046367452513562e-46 ], [ 121.375 , 2.0464726021973535e-46 ], [ 121.4375 , 2.0465777572841414e-46 ], 
                    [ 121.5 , 2.0466896545991312e-46 ], [ 121.5625 , 2.0468018387490404e-46 ], [ 121.625 , 2.0469140290480416e-46 ], [ 121.6875 , 2.0470262254964434e-46 ], 
                    [ 121.75 , 2.0471415262886628e-46 ], [ 121.8125 , 2.047260190146501e-46 ], [ 121.875 , 2.047378860882779e-46 ], [ 121.9375 , 2.0474975384978532e-46 ], 
                    [ 122.0 , 2.0476162229921505e-46 ], [ 122.0625 , 2.0477377676878197e-46 ], [ 122.125 , 2.0478593659965037e-46 ], [ 122.1875 , 2.047980971525926e-46 ], 
                    [ 122.25 , 2.0481025842764863e-46 ], [ 122.3125 , 2.048228536776293e-46 ], [ 122.375 , 2.0483590429140224e-46 ], [ 122.4375 , 2.0484895573671577e-46 ], 
                    [ 122.5 , 2.048620080136229e-46 ], [ 122.5625 , 2.0487506112217512e-46 ], [ 122.625 , 2.0488840011835338e-46 ], [ 122.6875 , 2.049017469930394e-46 ], 
                    [ 122.75 , 2.049150947371727e-46 ], [ 122.8125 , 2.0492844335080115e-46 ], [ 122.875 , 2.049421628959463e-46 ], [ 122.9375 , 2.0495633936023006e-46 ], 
                    [ 123.0 , 2.049705168051409e-46 ], [ 123.0625 , 2.049846952307467e-46 ], [ 123.125 , 2.0499887463711675e-46 ], [ 123.1875 , 2.0501336519686853e-46 ], 
                    [ 123.25 , 2.050278837543499e-46 ], [ 123.3125 , 2.0504240334000523e-46 ], [ 123.375 , 2.050569239538986e-46 ], [ 123.4375 , 2.0507168136558497e-46 ], 
                    [ 123.5 , 2.0508683114508357e-46 ], [ 123.5625 , 2.0510198204378163e-46 ], [ 123.625 , 2.0511713406175604e-46 ], [ 123.6875 , 2.051322871990982e-46 ], 
                    [ 123.75 , 2.0514779283765336e-46 ], [ 123.8125 , 2.051633706057766e-46 ], [ 123.875 , 2.0517894955678777e-46 ], [ 123.9375 , 2.0519452969077972e-46 ], 
                    [ 124.0 , 2.052102832500503e-46 ], [ 124.0625 , 2.0522653863695414e-46 ], [ 124.125 , 2.052427953115069e-46 ], [ 124.1875 , 2.0525905327380334e-46 ], 
                    [ 124.25 , 2.052753125239454e-46 ], [ 124.3125 , 2.0529193203074582e-46 ], [ 124.375 , 2.0530872182343996e-46 ], [ 124.4375 , 2.0532551298929104e-46 ], 
                    [ 124.5 , 2.0534230552839964e-46 ], [ 124.5625 , 2.0535913762650888e-46 ], [ 124.625 , 2.0537630011540595e-46 ], [ 124.6875 , 2.0539346403862745e-46 ], 
                    [ 124.75 , 2.054106293962874e-46 ], [ 124.8125 , 2.054277961885086e-46 ], [ 124.875 , 2.054453269337915e-46 ], [ 124.9375 , 2.0546320473051945e-46 ], 
                    [ 125.0 , 2.0548108408296552e-46 ], [ 125.0625 , 2.054989649912694e-46 ], [ 125.125 , 2.0551684745556516e-46 ], [ 125.1875 , 2.0553505736859338e-46 ], 
                    [ 125.25 , 2.0555330350628184e-46 ], [ 125.3125 , 2.055715512637501e-46 ], [ 125.375 , 2.055898006411406e-46 ], [ 125.4375 , 2.05608197662302e-46 ], 
                    [ 125.5 , 2.056269718533031e-46 ], [ 125.5625 , 2.056457477585809e-46 ], [ 125.625 , 2.0566452537829774e-46 ], [ 125.6875 , 2.056833047126088e-46 ], 
                    [ 125.75 , 2.0570247822857576e-46 ], [ 125.8125 , 2.0572185935717147e-46 ], [ 125.875 , 2.0574124231184358e-46 ], [ 125.9375 , 2.057606270927598e-46 ], 
                    [ 126.0 , 2.057800212168437e-46 ], [ 126.0625 , 2.05799879493038e-46 ], [ 126.125 , 2.0581973968560316e-46 ], [ 126.1875 , 2.058396017947226e-46 ], 
                    [ 126.25 , 2.058594658205872e-46 ], [ 126.3125 , 2.0587945595421796e-46 ], [ 126.375 , 2.0589965415312256e-46 ], [ 126.4375 , 2.059198543336102e-46 ], 
                    [ 126.5 , 2.059400564958797e-46 ], [ 126.5625 , 2.0596026064011813e-46 ], [ 126.625 , 2.0598082950713794e-46 ], [ 126.6875 , 2.060015529596954e-46 ], 
                    [ 126.75 , 2.060222784972128e-46 ], [ 126.8125 , 2.0604300611989707e-46 ], [ 126.875 , 2.0606375531517606e-46 ], [ 126.9375 , 2.0608509614855406e-46 ], 
                    [ 127.0 , 2.0610643919207896e-46 ], [ 127.0625 , 2.0612778444598113e-46 ], [ 127.125 , 2.0614913191048504e-46 ], [ 127.1875 , 2.0617064005897592e-46 ], 
                    [ 127.25 , 2.0619243221121857e-46 ], [ 127.3125 , 2.062142266668813e-46 ], [ 127.375 , 2.0623602342620914e-46 ], [ 127.4375 , 2.0625782248944696e-46 ], 
                    [ 127.5 , 2.0627991743906755e-46 ], [ 127.5625 , 2.0630215806715935e-46 ], [ 127.625 , 2.0632440109318185e-46 ], [ 127.6875 , 2.0634664651739934e-46 ], 
                    [ 127.75 , 2.0636889434006752e-46 ], [ 127.8125 , 2.0639147024653412e-46 ], [ 127.875 , 2.0641405939034485e-46 ], [ 127.9375 , 2.0643665100649766e-46 ], 
                    [ 128.0 , 2.0645924509525443e-46 ], [ 128.0625 , 2.064820151540221e-46 ], [ 128.125 , 2.0650531387104874e-46 ], [ 128.1875 , 2.0652861521701912e-46 ], 
                    [ 128.25 , 2.0655191919223587e-46 ], [ 128.3125 , 2.065752257969896e-46 ], [ 128.375 , 2.0659870996895895e-46 ], [ 128.4375 , 2.066223532120017e-46 ], 
                    [ 128.5 , 2.066459991607915e-46 ], [ 128.5625 , 2.0666964781563508e-46 ], [ 128.625 , 2.0669329917683474e-46 ], [ 128.6875 , 2.0671729009628496e-46 ], 
                    [ 128.75 , 2.067413722767934e-46 ], [ 128.8125 , 2.0676545726283113e-46 ], [ 128.875 , 2.067895450547265e-46 ], [ 128.9375 , 2.068136539088756e-46 ], 
                    [ 129.0 , 2.0683820377100494e-46 ], [ 129.0625 , 2.0686275654733273e-46 ], [ 129.125 , 2.0688731223820643e-46 ], [ 129.1875 , 2.069118708439646e-46 ], 
                    [ 129.25 , 2.069365395623039e-46 ], [ 129.3125 , 2.069614762692177e-46 ], [ 129.375 , 2.0698641598110613e-46 ], [ 129.4375 , 2.0701135869833425e-46 ], 
                    [ 129.5 , 2.0703630442126413e-46 ], [ 129.5625 , 2.0706154349157106e-46 ], [ 129.625 , 2.07087053679086e-46 ], [ 129.6875 , 2.0711256700948137e-46 ], 
                    [ 129.75 , 2.071380834831473e-46 ], [ 129.8125 , 2.071636031004666e-46 ], [ 129.875 , 2.0718935513646464e-46 ], [ 129.9375 , 2.0721519262373785e-46 ], 
                    [ 130.0 , 2.0724103333307313e-46 ], [ 130.0625 , 2.072668772648606e-46 ], [ 130.125 , 2.0729272441950787e-46 ], [ 130.1875 , 2.0731905606057653e-46 ], 
                    [ 130.25 , 2.0734541531493777e-46 ], [ 130.3125 , 2.0737177792070186e-46 ], [ 130.375 , 2.073981438783038e-46 ], [ 130.4375 , 2.0742456814380354e-46 ], 
                    [ 130.5 , 2.0745130240994827e-46 ], [ 130.5625 , 2.074780401217841e-46 ], [ 130.625 , 2.075047812797566e-46 ], [ 130.6875 , 2.075315258843055e-46 ], 
                    [ 130.75 , 2.0755847345151706e-46 ], [ 130.8125 , 2.075858188504786e-46 ], [ 130.875 , 2.0761316785213988e-46 ], [ 130.9375 , 2.0764052045697837e-46 ], 
                    [ 131.0 , 2.0766787666546444e-46 ], [ 131.0625 , 2.076953672373079e-46 ], [ 131.125 , 2.077229901490978e-46 ], [ 131.1875 , 2.0775061673466037e-46 ], 
                    [ 131.25 , 2.077782469944783e-46 ], [ 131.3125 , 2.0780588092905057e-46 ], [ 131.375 , 2.0783372482848195e-46 ], [ 131.4375 , 2.0786167316126955e-46 ], 
                    [ 131.5 , 2.078896252523933e-46 ], [ 131.5625 , 2.0791758110236008e-46 ], [ 131.625 , 2.0794554071167674e-46 ], [ 131.6875 , 2.0797406424128562e-46 ], 
                    [ 131.75 , 2.080027113459448e-46 ], [ 131.8125 , 2.0803136239655895e-46 ], [ 131.875 , 2.0806001739367315e-46 ], [ 131.9375 , 2.080886763378339e-46 ], 
                    [ 132.0 , 2.0811749759684977e-46 ], [ 132.0625 , 2.0814632674256478e-46 ], [ 132.125 , 2.0817515988178845e-46 ], [ 132.1875 , 2.082039970150784e-46 ], 
                    [ 132.25 , 2.0823289264889245e-46 ], [ 132.3125 , 2.0826222450742634e-46 ], [ 132.375 , 2.082915604976702e-46 ], [ 132.4375 , 2.083209006202105e-46 ], 
                    [ 132.5 , 2.0835024487562194e-46 ], [ 132.5625 , 2.0837968394825685e-46 ], [ 132.625 , 2.0840942741776285e-46 ], [ 132.6875 , 2.0843917513275937e-46 ], 
                    [ 132.75 , 2.0846892709385243e-46 ], [ 132.8125 , 2.0849868330164663e-46 ], [ 132.875 , 2.0852859151280395e-46 ], [ 132.9375 , 2.08558796050468e-46 ], 
                    [ 133.0 , 2.0858900496314073e-46 ], [ 133.0625 , 2.0861921825145143e-46 ], [ 133.125 , 2.0864943591603835e-46 ], [ 133.1875 , 2.0867980738555447e-46 ], 
                    [ 133.25 , 2.087103734849598e-46 ], [ 133.3125 , 2.0874094406149426e-46 ], [ 133.375 , 2.0877151911581506e-46 ], [ 133.4375 , 2.0880209864857512e-46 ], 
                    [ 133.5 , 2.0883286063375824e-46 ], [ 133.5625 , 2.08863786271921e-46 ], [ 133.625 , 2.0889471648979785e-46 ], [ 133.6875 , 2.089256512880684e-46 ], 
                    [ 133.75 , 2.0895659066741248e-46 ], [ 133.8125 , 2.0898770761154496e-46 ], [ 133.875 , 2.0901894452876753e-46 ], [ 133.9375 , 2.0905018611489812e-46 ], 
                    [ 134.0 , 2.090814323706405e-46 ], [ 134.0625 , 2.0911268329668808e-46 ], [ 134.125 , 2.0914425426954267e-46 ], [ 134.1875 , 2.091759954140044e-46 ], 
                    [ 134.25 , 2.092077413757181e-46 ], [ 134.3125 , 2.0923949215541334e-46 ], [ 134.375 , 2.0927124775381833e-46 ], [ 134.4375 , 2.0930324335089393e-46 ], 
                    [ 134.5 , 2.0933533898308796e-46 ], [ 134.5625 , 2.0936743953698923e-46 ], [ 134.625 , 2.09399545013354e-46 ], [ 134.6875 , 2.0943165541293707e-46 ], 
                    [ 134.75 , 2.0946423480436448e-46 ], [ 134.8125 , 2.0949697233156602e-46 ], [ 134.875 , 2.095297149753738e-46 ], [ 134.9375 , 2.095624627365844e-46 ], 
                    [ 135.0 , 2.095952156159992e-46 ], [ 135.0625 , 2.096279657748684e-46 ], [ 135.125 , 2.0966071899189722e-46 ], [ 135.1875 , 2.096934773264361e-46 ], 
                    [ 135.25 , 2.0972624077928763e-46 ], [ 135.3125 , 2.0975900935124254e-46 ], [ 135.375 , 2.097921630832795e-46 ], [ 135.4375 , 2.098254170806377e-46 ], 
                    [ 135.5 , 2.0985867634906485e-46 ], [ 135.5625 , 2.098919408893891e-46 ], [ 135.625 , 2.0992521070245345e-46 ], [ 135.6875 , 2.0995870845745216e-46 ], 
                    [ 135.75 , 2.0999226447745765e-46 ], [ 135.8125 , 2.1002582586045476e-46 ], [ 135.875 , 2.1005939260729907e-46 ], [ 135.9375 , 2.100929647188494e-46 ], 
                    [ 136.0 , 2.1012706734052604e-46 ], [ 136.0625 , 2.101613068054523e-46 ], [ 136.125 , 2.1019555184957866e-46 ], [ 136.1875 , 2.1022980247381427e-46 ], 
                    [ 136.25 , 2.1026405867906836e-46 ], [ 136.3125 , 2.1029839015990137e-46 ], [ 136.375 , 2.103327455526856e-46 ], [ 136.4375 , 2.1036710655793645e-46 ], 
                    [ 136.5 , 2.1040147317657386e-46 ], [ 136.5625 , 2.1043584540951025e-46 ], [ 136.625 , 2.1047064017917566e-46 ], [ 136.6875 , 2.105055666680481e-46 ], 
                    [ 136.75 , 2.105404989527895e-46 ], [ 136.8125 , 2.105754370343602e-46 ], [ 136.875 , 2.1061038091371763e-46 ], [ 136.9375 , 2.1064539202886574e-46 ], 
                    [ 137.0 , 2.1068043192245302e-46 ], [ 137.0625 , 2.1071547764476864e-46 ], [ 137.125 , 2.1075052919677626e-46 ], [ 137.1875 , 2.1078558657944855e-46 ], 
                    [ 137.25 , 2.10820933484698e-46 ], [ 137.3125 , 2.108564248088733e-46 ], [ 137.375 , 2.108919221079506e-46 ], [ 137.4375 , 2.1092742538293268e-46 ], 
                    [ 137.5 , 2.109629346348271e-46 ], [ 137.5625 , 2.109987519564373e-46 ], [ 137.625 , 2.1103476370827702e-46 ], [ 137.6875 , 2.110707816063439e-46 ], 
                    [ 137.75 , 2.111068056516901e-46 ], [ 137.8125 , 2.1114283584535556e-46 ], [ 137.875 , 2.1117893970212843e-46 ], [ 137.9375 , 2.112151063344769e-46 ], 
                    [ 138.0 , 2.1125127916074314e-46 ], [ 138.0625 , 2.1128745818198937e-46 ], [ 138.125 , 2.1132364339927955e-46 ], [ 138.1875 , 2.1135997337534183e-46 ], 
                    [ 138.25 , 2.1139647492730264e-46 ], [ 138.3125 , 2.1143298278302242e-46 ], [ 138.375 , 2.1146949694359745e-46 ], [ 138.4375 , 2.115060174101135e-46 ], 
                    [ 138.5 , 2.1154269097876192e-46 ], [ 138.5625 , 2.1157964118989327e-46 ], [ 138.625 , 2.116165978551281e-46 ], [ 138.6875 , 2.1165356097559083e-46 ], 
                    [ 138.75 , 2.116905305524119e-46 ], [ 138.8125 , 2.117275698880266e-46 ], [ 138.875 , 2.1176480768398917e-46 ], [ 138.9375 , 2.1180205202918744e-46 ], 
                    [ 139.0 , 2.1183930292477186e-46 ], [ 139.0625 , 2.118765603718959e-46 ], [ 139.125 , 2.119138912821814e-46 ], [ 139.1875 , 2.119516846783306e-46 ], 
                    [ 139.25 , 2.119894848146725e-46 ], [ 139.3125 , 2.1202729169241375e-46 ], [ 139.375 , 2.1206510531275207e-46 ], [ 139.4375 , 2.1210292567689306e-46 ], 
                    [ 139.5 , 2.1214082265158582e-46 ], [ 139.5625 , 2.1217872696096044e-46 ], [ 139.625 , 2.122166380428976e-46 ], [ 139.6875 , 2.1225455589860583e-46 ], 
                    [ 139.75 , 2.12292480529294e-46 ], [ 139.8125 , 2.1233079536476595e-46 ], [ 139.875 , 2.1236919016263105e-46 ], [ 139.9375 , 2.1240759190324873e-46 ], 
                    [ 140.0 , 2.124460005878789e-46 ], [ 140.0625 , 2.124844162177728e-46 ], [ 140.125 , 2.125230089146169e-46 ], [ 140.1875 , 2.1256168578259e-46 ], 
                    [ 140.25 , 2.1260036968933544e-46 ], [ 140.3125 , 2.1263906063612365e-46 ], [ 140.375 , 2.1267775862424637e-46 ], [ 140.4375 , 2.1271661418847095e-46 ], 
                    [ 140.5 , 2.1275561583095946e-46 ], [ 140.5625 , 2.1279462462440996e-46 ], [ 140.625 , 2.128336405701307e-46 ], [ 140.6875 , 2.128726636694375e-46 ], 
                    [ 140.75 , 2.1291175607989074e-46 ], [ 140.8125 , 2.129509785245736e-46 ], [ 140.875 , 2.1299020819478443e-46 ], [ 140.9375 , 2.130294450918573e-46 ], 
                    [ 141.0 , 2.130686892171205e-46 ], [ 141.0625 , 2.131079994786724e-46 ], [ 141.125 , 2.1314764569145468e-46 ], [ 141.1875 , 2.131872992799426e-46 ], 
                    [ 141.25 , 2.132269602455098e-46 ], [ 141.3125 , 2.132666285895302e-46 ], [ 141.375 , 2.1330630431337346e-46 ], [ 141.4375 , 2.133463670563631e-46 ], 
                    [ 141.5 , 2.1338646328690367e-46 ], [ 141.5625 , 2.134265670531122e-46 ], [ 141.625 , 2.1346667835640648e-46 ], [ 141.6875 , 2.1350679719820447e-46 ], 
                    [ 141.75 , 2.1354705495975142e-46 ], [ 141.8125 , 2.135873714139117e-46 ], [ 141.875 , 2.1362769547958483e-46 ], [ 141.9375 , 2.136680271582093e-46 ], 
                    [ 142.0 , 2.1370836645122396e-46 ], [ 142.0625 , 2.137488820071823e-46 ], [ 142.125 , 2.1378957124599775e-46 ], [ 142.1875 , 2.1383026823041247e-46 ], 
                    [ 142.25 , 2.1387097296190846e-46 ], [ 142.3125 , 2.1391168544195594e-46 ], [ 142.375 , 2.139524630242059e-46 ], [ 142.4375 , 2.139934150914269e-46 ], 
                    [ 142.5 , 2.1403437499717167e-46 ], [ 142.5625 , 2.1407534274294675e-46 ], [ 142.625 , 2.141163183302421e-46 ], [ 142.6875 , 2.1415730396978265e-46 ], 
                    [ 142.75 , 2.1419857144980092e-46 ], [ 142.8125 , 2.142398468819407e-46 ], [ 142.875 , 2.142811302677359e-46 ], [ 142.9375 , 2.14322421608716e-46 ], 
                    [ 143.0 , 2.143637209064172e-46 ], [ 143.0625 , 2.1440536414817545e-46 ], [ 143.125 , 2.144471311103303e-46 ], [ 143.1875 , 2.1448890620884616e-46 ], 
                    [ 143.25 , 2.1453068944530657e-46 ], [ 143.3125 , 2.145724808212953e-46 ], [ 143.375 , 2.14614304069305e-46 ], [ 143.4375 , 2.1465616288789166e-46 ], 
                    [ 143.5 , 2.1469802987071054e-46 ], [ 143.5625 , 2.14739905019351e-46 ], [ 143.625 , 2.1478178833540874e-46 ], [ 143.6875 , 2.148237168161819e-46 ], 
                    [ 143.75 , 2.148658367346839e-46 ], [ 143.8125 , 2.14907964911527e-46 ], [ 143.875 , 2.1495010134833192e-46 ], [ 143.9375 , 2.1499224604671362e-46 ], 
                    [ 144.0 , 2.1503439900829652e-46 ], [ 144.0625 , 2.150769491741614e-46 ], [ 144.125 , 2.1511956486307084e-46 ], [ 144.1875 , 2.151621889959192e-46 ], 
                    [ 144.25 , 2.1520482157438714e-46 ], [ 144.3125 , 2.1524746260013743e-46 ], [ 144.375 , 2.1529026128741462e-46 ], [ 144.4375 , 2.153331857463871e-46 ], 
                    [ 144.5 , 2.1537611876361558e-46 ], [ 144.5625 , 2.15419060340808e-46 ], [ 144.625 , 2.1546201047966943e-46 ], [ 144.6875 , 2.1550501963217276e-46 ], 
                    [ 144.75 , 2.1554820441110204e-46 ], [ 144.8125 , 2.155913978437746e-46 ], [ 144.875 , 2.1563459993192163e-46 ], [ 144.9375 , 2.156778106772836e-46 ], 
                    [ 145.0 , 2.1572103008159386e-46 ], [ 145.0625 , 2.15764499295385e-46 ], [ 145.125 , 2.1580800288240703e-46 ], [ 145.1875 , 2.1585151524085385e-46 ], 
                    [ 145.25 , 2.158950363724971e-46 ], [ 145.3125 , 2.1593856627910265e-46 ], [ 145.375 , 2.159822452890373e-46 ], [ 145.4375 , 2.160260507847258e-46 ], 
                    [ 145.5 , 2.160698651650371e-46 ], [ 145.5625 , 2.1611368843178093e-46 ], [ 145.625 , 2.161575205867488e-46 ], [ 145.6875 , 2.1620140151691102e-46 ], 
                    [ 145.75 , 2.1624546827307392e-46 ], [ 145.8125 , 2.162895440110426e-46 ], [ 145.875 , 2.1633362873264776e-46 ], [ 145.9375 , 2.1637772243971896e-46 ], 
                    [ 146.0 , 2.1642182513409217e-46 ], [ 146.0625 , 2.164661449209523e-46 ], [ 146.125 , 2.165105232442648e-46 ], [ 146.1875 , 2.1655491066570175e-46 ], 
                    [ 146.25 , 2.1659930718712366e-46 ], [ 146.3125 , 2.1664371281040083e-46 ], [ 146.375 , 2.1668827135134638e-46 ], [ 146.4375 , 2.167330409912839e-46 ], 
                    [ 146.5 , 2.167778198810083e-46 ], [ 146.5625 , 2.168226080224337e-46 ], [ 146.625 , 2.1686740541746853e-46 ], [ 146.6875 , 2.1691221918234214e-46 ], 
                    [ 146.75 , 2.169573315829162e-46 ], [ 146.8125 , 2.1700245336576036e-46 ], [ 146.875 , 2.1704758453282435e-46 ], [ 146.9375 , 2.1709272508605824e-46 ], 
                    [ 147.0 , 2.1713787502742037e-46 ], [ 147.0625 , 2.1718305589337384e-46 ], [ 147.125 , 2.1722825958723396e-46 ], [ 147.1875 , 2.17273472689624e-46 ], 
                    [ 147.25 , 2.1731869520250516e-46 ], [ 147.3125 , 2.1736392712783777e-46 ], [ 147.375 , 2.1740921661969342e-46 ], [ 147.4375 , 2.1745471823045572e-46 ], 
                    [ 147.5 , 2.175002293642608e-46 ], [ 147.5625 , 2.1754575002309696e-46 ], [ 147.625 , 2.1759128020895934e-46 ], [ 147.6875 , 2.176368199238434e-46 ], 
                    [ 147.75 , 2.176825563326214e-46 ], [ 147.8125 , 2.1772836408984794e-46 ], [ 147.875 , 2.177741814865712e-46 ], [ 147.9375 , 2.1782000852482274e-46 ], 
                    [ 148.0 , 2.178658452066299e-46 ], [ 148.0625 , 2.1791179223046917e-46 ], [ 148.125 , 2.1795798848930882e-46 ], [ 148.1875 , 2.1800419454153153e-46 ], 
                    [ 148.25 , 2.180504103892166e-46 ], [ 148.3125 , 2.1809663603444212e-46 ], [ 148.375 , 2.1814287147928356e-46 ], [ 148.4375 , 2.1818932302351923e-46 ], 
                    [ 148.5 , 2.1823582376238383e-46 ], [ 148.5625 , 2.182823344115319e-46 ], [ 148.625 , 2.183288549730832e-46 ], [ 148.6875 , 2.1837538544914268e-46 ], 
                    [ 148.75 , 2.184220308558335e-46 ], [ 148.8125 , 2.184688666148513e-46 ], [ 148.875 , 2.1851571241675866e-46 ], [ 148.9375 , 2.1856256826371053e-46 ], 
                    [ 149.0 , 2.1860943415786086e-46 ], [ 149.0625 , 2.1865631010135942e-46 ], [ 149.125 , 2.187031828915932e-46 ], [ 149.1875 , 2.187500639287883e-46 ], 
                    [ 149.25 , 2.187969550153646e-46 ], [ 149.3125 , 2.1884385615347467e-46 ], [ 149.375 , 2.1889076734527625e-46 ], [ 149.4375 , 2.1893785974577855e-46 ], 
                    [ 149.5 , 2.1898524638010298e-46 ], [ 149.5625 , 2.190326432707308e-46 ], [ 149.625 , 2.1908005041988186e-46 ], [ 149.6875 , 2.191274678297781e-46 ], 
                    [ 149.75 , 2.1917489550263715e-46 ], [ 149.8125 , 2.1922247168345652e-46 ], [ 149.875 , 2.1927008145246295e-46 ], [ 149.9375 , 2.193177015611521e-46 ], 
                    [ 150.0 , 2.193653320117631e-46 ], [ 150.0625 , 2.1941297280654683e-46 ], [ 150.125 , 2.1946064553973068e-46 ], [ 150.1875 , 2.1950837131010107e-46 ], 
                    [ 150.25 , 2.195561074593216e-46 ], [ 150.3125 , 2.1960385398965247e-46 ], [ 150.375 , 2.196516109033466e-46 ], [ 150.4375 , 2.19699378202662e-46 ], 
                    [ 150.5 , 2.1974748274825914e-46 ], [ 150.5625 , 2.1979567955921722e-46 ], [ 150.625 , 2.198438869410916e-46 ], [ 150.6875 , 2.1989210489620385e-46 ], 
                    [ 150.75 , 2.1994033342686995e-46 ], [ 150.8125 , 2.1998858771116573e-46 ], [ 150.875 , 2.2003689860922503e-46 ], [ 150.9375 , 2.200852201166659e-46 ], 
                    [ 151.0 , 2.2013355223581977e-46 ], [ 151.0625 , 2.201818949690124e-46 ], [ 151.125 , 2.202302483185825e-46 ], [ 151.1875 , 2.2027895452372775e-46 ], 
                    [ 151.25 , 2.203278325892124e-46 ], [ 151.3125 , 2.2037672150032796e-46 ], [ 151.375 , 2.2042562125948575e-46 ], [ 151.4375 , 2.2047453186909756e-46 ], 
                    [ 151.5 , 2.2052344874848887e-46 ], [ 151.5625 , 2.2057234013581627e-46 ], [ 151.625 , 2.2062124236266424e-46 ], [ 151.6875 , 2.206701554314296e-46 ], 
                    [ 151.75 , 2.207190793445224e-46 ], [ 151.8125 , 2.207680141043406e-46 ], [ 151.875 , 2.2081710681156e-46 ], [ 151.9375 , 2.208663506699432e-46 ], 
                    [ 152.0 , 2.209156055100733e-46 ], [ 152.0625 , 2.2096487133439936e-46 ], [ 152.125 , 2.210141481453741e-46 ], [ 152.1875 , 2.210634359454413e-46 ], 
                    [ 152.25 , 2.21112822919429e-46 ], [ 152.3125 , 2.2116222943751014e-46 ], [ 152.375 , 2.212116469952225e-46 ], [ 152.4375 , 2.2126107559502978e-46 ], 
                    [ 152.5 , 2.213105152394086e-46 ], [ 152.5625 , 2.2136008353467435e-46 ], [ 152.625 , 2.2140994270143232e-46 ], [ 152.6875 , 2.2145981309847706e-46 ], 
                    [ 152.75 , 2.2150969472833326e-46 ], [ 152.8125 , 2.215595875935327e-46 ], [ 152.875 , 2.216094916966059e-46 ], [ 152.9375 , 2.2165950927539756e-46 ], 
                    [ 153.0 , 2.2170958988845438e-46 ], [ 153.0625 , 2.2175968181646497e-46 ], [ 153.125 , 2.218097850619937e-46 ], [ 153.1875 , 2.2185989962758966e-46 ], 
                    [ 153.25 , 2.21910032648193e-46 ], [ 153.3125 , 2.219603928005213e-46 ], [ 153.375 , 2.220107643815591e-46 ], [ 153.4375 , 2.2206114739389838e-46 ], 
                    [ 153.5 , 2.221115418401271e-46 ], [ 153.5625 , 2.221619477228543e-46 ], [ 153.625 , 2.2221231423869066e-46 ], [ 153.6875 , 2.2226261065345635e-46 ], 
                    [ 153.75 , 2.2231291845251482e-46 ], [ 153.8125 , 2.223632376384381e-46 ], [ 153.875 , 2.224135682138035e-46 ], [ 153.9375 , 2.2246391018119688e-46 ], 
                    [ 154.0 , 2.2251466721893743e-46 ], [ 154.0625 , 2.2256560560358055e-46 ], [ 154.125 , 2.2261655564911354e-46 ], [ 154.1875 , 2.226675173582027e-46 ], 
                    [ 154.25 , 2.2271849073352588e-46 ], [ 154.3125 , 2.227694759469914e-46 ], [ 154.375 , 2.2282047971630777e-46 ], [ 154.4375 , 2.2287149516309534e-46 ], 
                    [ 154.5 , 2.2292252229003083e-46 ], [ 154.5625 , 2.2297356109978688e-46 ], [ 154.625 , 2.230246115950367e-46 ], [ 154.6875 , 2.2307579209048755e-46 ], 
                    [ 154.75 , 2.231272267822259e-46 ], [ 154.8125 , 2.2317867333328255e-46 ], [ 154.875 , 2.232301317463997e-46 ], [ 154.9375 , 2.2328160202430453e-46 ], 
                    [ 155.0 , 2.233330841697359e-46 ], [ 155.0625 , 2.233846228576826e-46 ], [ 155.125 , 2.2343619945692863e-46 ], [ 155.1875 , 2.234877879645411e-46 ], 
                    [ 155.25 , 2.2353938838326635e-46 ], [ 155.3125 , 2.2359100071586074e-46 ], [ 155.375 , 2.236426249650672e-46 ], [ 155.4375 , 2.2369439497262496e-46 ], 
                    [ 155.5 , 2.237461911804736e-46 ], [ 155.5625 , 2.237979993816811e-46 ], [ 155.625 , 2.238498195790244e-46 ], [ 155.6875 , 2.239016517752782e-46 ], 
                    [ 155.75 , 2.239535632215346e-46 ], [ 155.8125 , 2.2400580162102732e-46 ], [ 155.875 , 2.2405805220541223e-46 ], [ 155.9375 , 2.241103149775347e-46 ], 
                    [ 156.0 , 2.2416258994023765e-46 ], [ 156.0625 , 2.242148770963582e-46 ], [ 156.125 , 2.2426719407945e-46 ], [ 156.1875 , 2.2431954645639654e-46 ], 
                    [ 156.25 , 2.243719110543503e-46 ], [ 156.3125 , 2.2442428787616425e-46 ], [ 156.375 , 2.244766769246982e-46 ], [ 156.4375 , 2.245290782027984e-46 ], 
                    [ 156.5 , 2.245816322057431e-46 ], [ 156.5625 , 2.2463426709941234e-46 ], [ 156.625 , 2.246869143290505e-46 ], [ 156.6875 , 2.247395738975487e-46 ], 
                    [ 156.75 , 2.2479224580779886e-46 ], [ 156.8125 , 2.248449300626934e-46 ], [ 156.875 , 2.2489762565786062e-46 ], [ 156.9375 , 2.2495033348604738e-46 ], 
                    [ 157.0 , 2.2500305366703144e-46 ], [ 157.0625 , 2.250557862037127e-46 ], [ 157.125 , 2.251085310989789e-46 ], [ 157.1875 , 2.2516132504174292e-46 ], 
                    [ 157.25 , 2.2521444750455806e-46 ], [ 157.3125 , 2.2526758250059314e-46 ], [ 157.375 , 2.25320730032802e-46 ], [ 157.4375 , 2.2537389010414384e-46 ], 
                    [ 157.5 , 2.2542706271757868e-46 ], [ 157.5625 , 2.2548031181574955e-46 ], [ 157.625 , 2.2553371452069877e-46 ], [ 157.6875 , 2.2558712987353392e-46 ], 
                    [ 157.75 , 2.256405578772441e-46 ], [ 157.8125 , 2.256939985348383e-46 ], [ 157.875 , 2.257474518493007e-46 ], [ 157.9375 , 2.2580102028075078e-46 ], 
                    [ 158.0 , 2.258547022784356e-46 ], [ 158.0625 , 2.25908397038492e-46 ], [ 158.125 , 2.2596210456395566e-46 ], [ 158.1875 , 2.2601582485786146e-46 ], 
                    [ 158.25 , 2.2606955792324654e-46 ], [ 158.3125 , 2.26123414123176e-46 ], [ 158.375 , 2.2617733509971116e-46 ], [ 158.4375 , 2.2623126893414076e-46 ], 
                    [ 158.5 , 2.2628521562953403e-46 ], [ 158.5625 , 2.263391751889626e-46 ], [ 158.625 , 2.2639314761548443e-46 ], [ 158.6875 , 2.2644719295214696e-46 ], 
                    [ 158.75 , 2.2650126262961157e-46 ], [ 158.8125 , 2.265553452175034e-46 ], [ 158.875 , 2.2660944071890205e-46 ], [ 158.9375 , 2.2666354913690046e-46 ], 
                    [ 159.0 , 2.267176704745732e-46 ], [ 159.0625 , 2.2677196194859027e-46 ], [ 159.125 , 2.2682626898057217e-46 ], [ 159.1875 , 2.2688058901792926e-46 ], 
                    [ 159.25 , 2.269349220637713e-46 ], [ 159.3125 , 2.2698926812121505e-46 ], [ 159.375 , 2.2704366157550737e-46 ], [ 159.4375 , 2.2709834073081113e-46 ], 
                    [ 159.5 , 2.271530330545465e-46 ], [ 159.5625 , 2.272077385498913e-46 ], [ 159.625 , 2.2726245722001595e-46 ], [ 159.6875 , 2.273171890680935e-46 ], 
                    [ 159.75 , 2.2737191466510377e-46 ], [ 159.8125 , 2.2742659188697308e-46 ], [ 159.875 , 2.2748128225733632e-46 ], [ 159.9375 , 2.2753598577935526e-46 ], 
                    [ 160.0 , 2.2759070245619744e-46 ], [ 160.0625 , 2.276454322910214e-46 ], [ 160.125 , 2.2770027767683017e-46 ], [ 160.1875 , 2.277553386669461e-46 ], 
                    [ 160.25 , 2.2781041297155014e-46 ], [ 160.3125 , 2.2786550059386347e-46 ], [ 160.375 , 2.2792060153710166e-46 ], [ 160.4375 , 2.2797571580449075e-46 ], 
                    [ 160.5 , 2.2803092700375807e-46 ], [ 160.5625 , 2.2808626152519753e-46 ], [ 160.625 , 2.2814160947423903e-46 ], [ 160.6875 , 2.281969708541489e-46 ], 
                    [ 160.75 , 2.282523456681799e-46 ], [ 160.8125 , 2.283077339195904e-46 ], [ 160.875 , 2.2836321403404035e-46 ], [ 160.9375 , 2.2841878239501023e-46 ], 
                    [ 161.0 , 2.2847436427761114e-46 ], [ 161.0625 , 2.2852995968513003e-46 ], [ 161.125 , 2.2858556862086125e-46 ], [ 161.1875 , 2.286411910880918e-46 ], 
                    [ 161.25 , 2.2869686313026355e-46 ], [ 161.3125 , 2.2875257526408218e-46 ], [ 161.375 , 2.2880830096976257e-46 ], [ 161.4375 , 2.288640402506078e-46 ], 
                    [ 161.5 , 2.2891979310992314e-46 ], [ 161.5625 , 2.2897555955102456e-46 ], [ 161.625 , 2.2903152746869417e-46 ], [ 161.6875 , 2.2908762231086667e-46 ], 
                    [ 161.75 , 2.2914373089189578e-46 ], [ 161.8125 , 2.291998532151497e-46 ], [ 161.875 , 2.2925598928398935e-46 ], [ 161.9375 , 2.2931213910178287e-46 ], 
                    [ 162.0 , 2.293683277849366e-46 ], [ 162.0625 , 2.2942454249664325e-46 ], [ 162.125 , 2.29480770985728e-46 ], [ 162.1875 , 2.2953701325556427e-46 ], 
                    [ 162.25 , 2.295932693095295e-46 ], [ 162.3125 , 2.2964953915101016e-46 ], [ 162.375 , 2.297060549256047e-46 ], [ 162.4375 , 2.2976268990827674e-46 ], 
                    [ 162.5 , 2.2981933885454245e-46 ], [ 162.5625 , 2.2987600176783642e-46 ], [ 162.625 , 2.299326786516121e-46 ], [ 162.6875 , 2.2998936950930904e-46 ], 
                    [ 162.75 , 2.3004606053601066e-46 ], [ 162.8125 , 2.301027597295023e-46 ], [ 162.875 , 2.3015947289757695e-46 ], [ 162.9375 , 2.3021620004368398e-46 ], 
                    [ 163.0 , 2.3027294117126844e-46 ], [ 163.0625 , 2.303296962837731e-46 ], [ 163.125 , 2.30386621987636e-46 ], [ 163.1875 , 2.3044362762136197e-46 ], 
                    [ 163.25 , 2.3050064736025707e-46 ], [ 163.3125 , 2.305576812078115e-46 ], [ 163.375 , 2.306147291675145e-46 ], [ 163.4375 , 2.3067179124286453e-46 ], 
                    [ 163.5 , 2.3072872115432582e-46 ], [ 163.5625 , 2.307855987621306e-46 ], [ 163.625 , 2.308424903909949e-46 ], [ 163.6875 , 2.3089939604438496e-46 ], 
                    [ 163.75 , 2.309563157257481e-46 ], [ 163.8125 , 2.3101324943855073e-46 ], [ 163.875 , 2.3107077567697554e-46 ], [ 163.9375 , 2.3112861967142528e-46 ], 
                    [ 164.0 , 2.3118647814597236e-46 ], [ 164.0625 , 2.312443511042448e-46 ], [ 164.125 , 2.313022385498651e-46 ], [ 164.1875 , 2.3136014048646143e-46 ], 
                    [ 164.25 , 2.314176945150067e-46 ], [ 164.3125 , 2.314750291487422e-46 ], [ 164.375 , 2.315323779873529e-46 ], [ 164.4375 , 2.3158974103436304e-46 ], 
                    [ 164.5 , 2.316471182932944e-46 ], [ 164.5625 , 2.3170450976765822e-46 ], [ 164.625 , 2.3176217879916294e-46 ], [ 164.6875 , 2.318200829784586e-46 ], 
                    [ 164.75 , 2.318780016247151e-46 ], [ 164.8125 , 2.319359347415436e-46 ], [ 164.875 , 2.319938823325627e-46 ], [ 164.9375 , 2.3205184440138554e-46 ], 
                    [ 165.0 , 2.3210982657676946e-46 ], [ 165.0625 , 2.3216782933542807e-46 ], [ 165.125 , 2.3222584658860606e-46 ], [ 165.1875 , 2.3228387833992566e-46 ], 
                    [ 165.25 , 2.323419245930048e-46 ], [ 165.3125 , 2.3239998535147404e-46 ], [ 165.375 , 2.324582450350677e-46 ], [ 165.4375 , 2.32516815232257e-46 ], 
                    [ 165.5 , 2.3257540018680092e-46 ], [ 165.5625 , 2.3263399990241452e-46 ], [ 165.625 , 2.3269261438282192e-46 ], [ 165.6875 , 2.327512436317316e-46 ], 
                    [ 165.75 , 2.328098723917096e-46 ], [ 165.8125 , 2.3286847818146253e-46 ], [ 165.875 , 2.329270987241904e-46 ], [ 165.9375 , 2.3298573402360855e-46 ], 
                    [ 166.0 , 2.3304438408343356e-46 ], [ 166.0625 , 2.3310304890737935e-46 ], [ 166.125 , 2.3316171921197838e-46 ], [ 166.1875 , 2.3322036079161502e-46 ], 
                    [ 166.25 , 2.3327901711996115e-46 ], [ 166.3125 , 2.3333768820072956e-46 ], [ 166.375 , 2.3339637403763055e-46 ], [ 166.4375 , 2.3345507463437205e-46 ], 
                    [ 166.5 , 2.335137976331427e-46 ], [ 166.5625 , 2.335726869376402e-46 ], [ 166.625 , 2.3363159109329358e-46 ], [ 166.6875 , 2.336905101038497e-46 ], 
                    [ 166.75 , 2.3374944397305155e-46 ], [ 166.8125 , 2.3380839270465448e-46 ], [ 166.875 , 2.338673563023984e-46 ], [ 166.9375 , 2.339266253237882e-46 ], 
                    [ 167.0 , 2.3398594022883524e-46 ], [ 167.0625 , 2.340452701738904e-46 ], [ 167.125 , 2.341046151627623e-46 ], [ 167.1875 , 2.3416397519927213e-46 ], 
                    [ 167.25 , 2.3422335028723037e-46 ], [ 167.3125 , 2.342826991578595e-46 ], [ 167.375 , 2.343420488617309e-46 ], [ 167.4375 , 2.3440141360037424e-46 ], 
                    [ 167.5 , 2.3446079337760326e-46 ], [ 167.5625 , 2.3452018819721927e-46 ], [ 167.625 , 2.3457959806303952e-46 ], [ 167.6875 , 2.3463919841841024e-46 ], 
                    [ 167.75 , 2.346989473835024e-46 ], [ 167.8125 , 2.3475871156318467e-46 ], [ 167.875 , 2.3481849096132956e-46 ], [ 167.9375 , 2.3487828558180577e-46 ], 
                    [ 168.0 , 2.3493809542850117e-46 ], [ 168.0625 , 2.349979838103561e-46 ], [ 168.125 , 2.350579856709506e-46 ], [ 168.1875 , 2.3511800285177565e-46 ], 
                    [ 168.25 , 2.351780353567479e-46 ], [ 168.3125 , 2.3523808318976663e-46 ], [ 168.375 , 2.3529814635475735e-46 ], [ 168.4375 , 2.3535824262786143e-46 ], 
                    [ 168.5 , 2.3541843308594493e-46 ], [ 168.5625 , 2.3547863893712617e-46 ], [ 168.625 , 2.3553886018533334e-46 ], [ 168.6875 , 2.355990968345075e-46 ], 
                    [ 168.75 , 2.356593488885889e-46 ], [ 168.8125 , 2.3571961635151386e-46 ], [ 168.875 , 2.3577953231054588e-46 ], [ 168.9375 , 2.358394544743803e-46 ], 
                    [ 169.0 , 2.3589939186712826e-46 ], [ 169.0625 , 2.359593444926568e-46 ], [ 169.125 , 2.3601931235483882e-46 ], [ 169.1875 , 2.3607929545754344e-46 ], 
                    [ 169.25 , 2.3613975805574025e-46 ], [ 169.3125 , 2.3620040270681296e-46 ], [ 169.375 , 2.3626106293244767e-46 ], [ 169.4375 , 2.363217387366476e-46 ], 
                    [ 169.5 , 2.3638243012341353e-46 ], [ 169.5625 , 2.3644313709674233e-46 ], [ 169.625 , 2.365038517946293e-46 ], [ 169.6875 , 2.3656457408811813e-46 ], 
                    [ 169.75 , 2.366253119720344e-46 ], [ 169.8125 , 2.3668606545038587e-46 ], [ 169.875 , 2.3674683452717484e-46 ], [ 169.9375 , 2.368076192064062e-46 ], 
                    [ 170.0 , 2.3686844403823473e-46 ], [ 170.0625 , 2.369293622425624e-46 ], [ 170.125 , 2.369902961139361e-46 ], [ 170.1875 , 2.3705124565637837e-46 ], 
                    [ 170.25 , 2.3711221087392456e-46 ], [ 170.3125 , 2.3717319177059928e-46 ], [ 170.375 , 2.3723418835043997e-46 ], [ 170.4375 , 2.3729534098148828e-46 ], 
                    [ 170.5 , 2.3735651401737496e-46 ], [ 170.5625 , 2.3741770282323467e-46 ], [ 170.625 , 2.3747890740313105e-46 ], [ 170.6875 , 2.3754012776112554e-46 ], 
                    [ 170.75 , 2.376013639012923e-46 ], [ 170.8125 , 2.3766293019318997e-46 ], [ 170.875 , 2.3772466041307755e-46 ], [ 170.9375 , 2.3778640666668156e-46 ], 
                    [ 171.0 , 2.3784816895816497e-46 ], [ 171.0625 , 2.3790994729169515e-46 ], [ 171.125 , 2.3797174167144217e-46 ], [ 171.1875 , 2.3803346564931342e-46 ], 
                    [ 171.25 , 2.380950715129319e-46 ], [ 171.3125 , 2.381566933208749e-46 ], [ 171.375 , 2.382183310772657e-46 ], [ 171.4375 , 2.382799847862336e-46 ], 
                    [ 171.5 , 2.383416544519073e-46 ], [ 171.5625 , 2.384033501963664e-46 ], [ 171.625 , 2.3846519235394302e-46 ], [ 171.6875 , 2.3852705055346225e-46 ], 
                    [ 171.75 , 2.385889247990803e-46 ], [ 171.8125 , 2.3865081509496463e-46 ], [ 171.875 , 2.3871272144527697e-46 ], [ 171.9375 , 2.3877464385418023e-46 ], 
                    [ 172.0 , 2.3883664983347096e-46 ], [ 172.0625 , 2.388986961357091e-46 ], [ 172.125 , 2.389607585566796e-46 ], [ 172.1875 , 2.390228371005715e-46 ], 
                    [ 172.25 , 2.3908493177156657e-46 ], [ 172.3125 , 2.3914704257385945e-46 ], [ 172.375 , 2.392092269321787e-46 ], [ 172.4375 , 2.392715136132559e-46 ], 
                    [ 172.5 , 2.3933381651290074e-46 ], [ 172.5625 , 2.3939613563533305e-46 ], [ 172.625 , 2.394584709847786e-46 ], [ 172.6875 , 2.3952082256546445e-46 ], 
                    [ 172.75 , 2.3958318810247354e-46 ], [ 172.8125 , 2.3964550091604894e-46 ], [ 172.875 , 2.3970782993646584e-46 ], [ 172.9375 , 2.397701751679376e-46 ], 
                    [ 173.0 , 2.39832536614684e-46 ], [ 173.0625 , 2.398949142809174e-46 ], [ 173.125 , 2.3995730817086126e-46 ], [ 173.1875 , 2.4001991613162066e-46 ], 
                    [ 173.25 , 2.4008264056705474e-46 ], [ 173.3125 , 2.4014538139427188e-46 ], [ 173.375 , 2.4020813861755915e-46 ], [ 173.4375 , 2.4027091224120134e-46 ], 
                    [ 173.5 , 2.4033370226948092e-46 ], [ 173.5625 , 2.4039654737783656e-46 ], [ 173.625 , 2.404595083857616e-46 ], [ 173.6875 , 2.405224858834729e-46 ], 
                    [ 173.75 , 2.4058547987529444e-46 ], [ 173.8125 , 2.4064849036554436e-46 ], [ 173.875 , 2.4071151735854536e-46 ], [ 173.9375 , 2.4077456085861106e-46 ], 
                    [ 174.0 , 2.4083755808479215e-46 ], [ 174.0625 , 2.409005619083653e-46 ], [ 174.125 , 2.4096358221392835e-46 ], [ 174.1875 , 2.410266190057913e-46 ], 
                    [ 174.25 , 2.410896722882653e-46 ], [ 174.3125 , 2.4115274206566612e-46 ], [ 174.375 , 2.4121595963641655e-46 ], [ 174.4375 , 2.412793555950912e-46 ], 
                    [ 174.5 , 2.4134276821537967e-46 ], [ 174.5625 , 2.414061975016643e-46 ], [ 174.625 , 2.4146964345832187e-46 ], [ 174.6875 , 2.4153310608973708e-46 ], 
                    [ 174.75 , 2.4159658428254392e-46 ], [ 174.8125 , 2.4166001039878457e-46 ], [ 174.875 , 2.417234531662235e-46 ], [ 174.9375 , 2.4178691258922696e-46 ], 
                    [ 175.0 , 2.418503886721745e-46 ], [ 175.0625 , 2.419138814194398e-46 ], [ 175.125 , 2.4197739083538736e-46 ], [ 175.1875 , 2.4204099042083848e-46 ], 
                    [ 175.25 , 2.4210466263574398e-46 ], [ 175.3125 , 2.4216835160050565e-46 ], [ 175.375 , 2.4223205731952636e-46 ], [ 175.4375 , 2.4229577979721355e-46 ], 
                    [ 175.5 , 2.4235951903797235e-46 ], [ 175.5625 , 2.4242328408865548e-46 ], [ 175.625 , 2.4248714384058465e-46 ], [ 175.6875 , 2.4255102041461392e-46 ], 
                    [ 175.75 , 2.4261491381516945e-46 ], [ 175.8125 , 2.42678824046682e-46 ], [ 175.875 , 2.4274275111359212e-46 ], [ 175.9375 , 2.428066950203277e-46 ], 
                    [ 176.0 , 2.4287073886456273e-46 ], [ 176.0625 , 2.429348463535933e-46 ], [ 176.125 , 2.4299897076425777e-46 ], [ 176.1875 , 2.4306311210102445e-46 ], 
                    [ 176.25 , 2.4312727036836113e-46 ], [ 176.3125 , 2.431914455707333e-46 ], [ 176.375 , 2.432556557362075e-46 ], [ 176.4375 , 2.4331999002373142e-46 ], 
                    [ 176.5 , 2.4338434132586365e-46 ], [ 176.5625 , 2.4344870964710936e-46 ], [ 176.625 , 2.435130949919678e-46 ], [ 176.6875 , 2.4357749736494126e-46 ], 
                    [ 176.75 , 2.4364191677053324e-46 ], [ 176.8125 , 2.4370630697399724e-46 ], [ 176.875 , 2.4377068726340952e-46 ], [ 176.9375 , 2.4383508456026877e-46 ], 
                    [ 177.0 , 2.4389949886906267e-46 ], [ 177.0625 , 2.439639301942922e-46 ], [ 177.125 , 2.440283785404474e-46 ], [ 177.1875 , 2.4409288985490486e-46 ], 
                    [ 177.25 , 2.441578141441279e-46 ], [ 177.3125 , 2.4422275570203295e-46 ], [ 177.375 , 2.4428771453322004e-46 ], [ 177.4375 , 2.443526906422766e-46 ], 
                    [ 177.5 , 2.4441768403380352e-46 ], [ 177.5625 , 2.444826947123958e-46 ], [ 177.625 , 2.4454779279622798e-46 ], [ 177.6875 , 2.4461295983876794e-46 ], 
                    [ 177.75 , 2.4467814424701413e-46 ], [ 177.8125 , 2.447433460255855e-46 ], [ 177.875 , 2.448085651791109e-46 ], [ 177.9375 , 2.448738017122221e-46 ], 
                    [ 178.0 , 2.449390543976942e-46 ], [ 178.0625 , 2.4500424869204916e-46 ], [ 178.125 , 2.450694603388607e-46 ], [ 178.1875 , 2.4513468934275967e-46 ], 
                    [ 178.25 , 2.4519993570835892e-46 ], [ 178.3125 , 2.4526519944028304e-46 ], [ 178.375 , 2.4533048054315262e-46 ], [ 178.4375 , 2.453955995483608e-46 ], 
                    [ 178.5 , 2.4546052167327712e-46 ], [ 178.5625 , 2.4552546097406446e-46 ], [ 178.625 , 2.455904174552651e-46 ], [ 178.6875 , 2.456553911214156e-46 ], 
                    [ 178.75 , 2.4572038197707465e-46 ], [ 178.8125 , 2.4578539002678293e-46 ], [ 178.875 , 2.4585123655557928e-46 ], [ 178.9375 , 2.4591723003533966e-46 ], 
                    [ 179.0 , 2.4598324122963113e-46 ], [ 179.0625 , 2.4604927014320873e-46 ], [ 179.125 , 2.4611531678083066e-46 ], [ 179.1875 , 2.4618138114725095e-46 ], 
                    [ 179.25 , 2.4624730233780323e-46 ], [ 179.3125 , 2.4631281039623933e-46 ], [ 179.375 , 2.463783358814868e-46 ], [ 179.4375 , 2.4644387879818334e-46 ], 
                    [ 179.5 , 2.465094391509697e-46 ], [ 179.5625 , 2.4657501694447556e-46 ], [ 179.625 , 2.4664061218335114e-46 ], [ 179.6875 , 2.4670653616312663e-46 ], 
                    [ 179.75 , 2.467726353265111e-46 ], [ 179.8125 , 2.4683875219959996e-46 ], [ 179.875 , 2.4690488678713815e-46 ], [ 179.9375 , 2.4697103909386652e-46 ], 
                    [ 180.0 , 2.470372091245396e-46 ], [ 180.0625 , 2.4710339498759158e-46 ], [ 180.125 , 2.4716952144588228e-46 ], [ 180.1875 , 2.4723566560003164e-46 ], 
                    [ 180.25 , 2.4730182745478747e-46 ], [ 180.3125 , 2.4736800701487956e-46 ], [ 180.375 , 2.4743420428504944e-46 ], [ 180.4375 , 2.4750041927003296e-46 ], 
                    [ 180.5 , 2.4756662162340254e-46 ], [ 180.5625 , 2.476327929834758e-46 ], [ 180.625 , 2.4769898203029887e-46 ], [ 180.6875 , 2.477651887685992e-46 ], 
                    [ 180.75 , 2.4783141320310543e-46 ], [ 180.8125 , 2.4789765533854757e-46 ], [ 180.875 , 2.4796391517965864e-46 ], [ 180.9375 , 2.4803039039303558e-46 ], 
                    [ 181.0 , 2.4809695729888905e-46 ], [ 181.0625 , 2.4816354207010547e-46 ], [ 181.125 , 2.4823014471148315e-46 ], [ 181.1875 , 2.4829676522781107e-46 ], 
                    [ 181.25 , 2.483634036238918e-46 ], [ 181.3125 , 2.4843007497966158e-46 ], [ 181.375 , 2.4849701842498245e-46 ], [ 181.4375 , 2.485639799092804e-46 ], 
                    [ 181.5 , 2.486309594374217e-46 ], [ 181.5625 , 2.486979570142596e-46 ], [ 181.625 , 2.487649726446683e-46 ], [ 181.6875 , 2.488320063335055e-46 ], 
                    [ 181.75 , 2.4889909800659627e-46 ], [ 181.8125 , 2.4896627634778465e-46 ], [ 181.875 , 2.490334728205337e-46 ], [ 181.9375 , 2.4910068742973887e-46 ], 
                    [ 182.0 , 2.4916792018029348e-46 ], [ 182.0625 , 2.4923517107709744e-46 ], [ 182.125 , 2.4930244012504845e-46 ], [ 182.1875 , 2.4936956712277088e-46 ], 
                    [ 182.25 , 2.4943663110503953e-46 ], [ 182.3125 , 2.4950371312310227e-46 ], [ 182.375 , 2.495708131818042e-46 ], [ 182.4375 , 2.4963793128600414e-46 ], 
                    [ 182.5 , 2.4970506744055163e-46 ], [ 182.5625 , 2.4977222165029923e-46 ], [ 182.625 , 2.4983946571390318e-46 ], [ 182.6875 , 2.499067321411576e-46 ], 
                    [ 182.75 , 2.499740166791323e-46 ], [ 182.8125 , 2.5004131933269983e-46 ], [ 182.875 , 2.5010864010674117e-46 ], [ 182.9375 , 2.5017597900613327e-46 ], 
                    [ 183.0 , 2.5024335815842208e-46 ], [ 183.0625 , 2.5031083971676006e-46 ], [ 183.125 , 2.5037833947242345e-46 ], [ 183.1875 , 2.5044585743032302e-46 ], 
                    [ 183.25 , 2.50513393595369e-46 ], [ 183.3125 , 2.5058094797246764e-46 ], [ 183.375 , 2.506485205665354e-46 ], [ 183.4375 , 2.50716231162563e-46 ], 
                    [ 183.5 , 2.5078410298426103e-46 ], [ 183.5625 , 2.508519931796512e-46 ], [ 183.625 , 2.509199017537163e-46 ], [ 183.6875 , 2.5098782871142824e-46 ], 
                    [ 183.75 , 2.5105577405776176e-46 ], [ 183.8125 , 2.5112373779770034e-46 ], [ 183.875 , 2.511915529648741e-46 ], [ 183.9375 , 2.512593107053458e-46 ], 
                    [ 184.0 , 2.5132708672315125e-46 ], [ 184.0625 , 2.513948810232135e-46 ], [ 184.125 , 2.5146269361047655e-46 ], [ 184.1875 , 2.515305244898591e-46 ], 
                    [ 184.25 , 2.515983736663061e-46 ], [ 184.3125 , 2.5166647672233596e-46 ], [ 184.375 , 2.517346232364837e-46 ], [ 184.4375 , 2.5180278820341643e-46 ], 
                    [ 184.5 , 2.5187097162813083e-46 ], [ 184.5625 , 2.5193917351562137e-46 ], [ 184.625 , 2.5200739387089815e-46 ], [ 184.6875 , 2.5207565956621858e-46 ], 
                    [ 184.75 , 2.5214417528855587e-46 ], [ 184.8125 , 2.5221270963388647e-46 ], [ 184.875 , 2.522812626072793e-46 ], [ 184.9375 , 2.523498342137922e-46 ], 
                    [ 185.0 , 2.524184244584934e-46 ], [ 185.0625 , 2.524870333464471e-46 ], [ 185.125 , 2.5255563577076516e-46 ], [ 185.1875 , 2.526241947356323e-46 ], 
                    [ 185.25 , 2.5269277231157723e-46 ], [ 185.3125 , 2.5276136850364306e-46 ], [ 185.375 , 2.5282998331689417e-46 ], [ 185.4375 , 2.5289861675637646e-46 ], 
                    [ 185.5 , 2.5296726882714986e-46 ], [ 185.5625 , 2.5303582823165287e-46 ], [ 185.625 , 2.5310427339919084e-46 ], [ 185.6875 , 2.5317273708086605e-46 ], 
                    [ 185.75 , 2.5324121928169368e-46 ], [ 185.8125 , 2.533097200066795e-46 ], [ 185.875 , 2.5337823926083064e-46 ], [ 185.9375 , 2.5344677704916633e-46 ], 
                    [ 186.0 , 2.5351579520236197e-46 ], [ 186.0625 , 2.5358512998800347e-46 ], [ 186.125 , 2.536544837362262e-46 ], [ 186.1875 , 2.5372385645220914e-46 ], 
                    [ 186.25 , 2.5379324814113797e-46 ], [ 186.3125 , 2.5386265880820716e-46 ], [ 186.375 , 2.539320884586071e-46 ], [ 186.4375 , 2.5400122681759493e-46 ], 
                    [ 186.5 , 2.5407028599121886e-46 ], [ 186.5625 , 2.541393639410098e-46 ], [ 186.625 , 2.5420846067207277e-46 ], [ 186.6875 , 2.5427757618951403e-46 ], 
                    [ 186.75 , 2.5434671049844143e-46 ], [ 186.8125 , 2.544158636039622e-46 ], [ 186.875 , 2.54484957486877e-46 ], [ 186.9375 , 2.545540594916342e-46 ], 
                    [ 187.0 , 2.5462318026012505e-46 ], [ 187.0625 , 2.546923197974411e-46 ], [ 187.125 , 2.5476147810867328e-46 ], [ 187.1875 , 2.5483065519893032e-46 ], 
                    [ 187.25 , 2.5489985107330776e-46 ], [ 187.3125 , 2.549693134642213e-46 ], [ 187.375 , 2.5503879881342464e-46 ], [ 187.4375 , 2.5510830309907295e-46 ], 
                    [ 187.5 , 2.551778263263323e-46 ], [ 187.5625 , 2.5524736850037016e-46 ], [ 187.625 , 2.5531692962634094e-46 ], [ 187.6875 , 2.5538652767687935e-46 ], 
                    [ 187.75 , 2.5545637634343887e-46 ], [ 187.8125 , 2.5552624411372845e-46 ], [ 187.875 , 2.555961309929785e-46 ], [ 187.9375 , 2.5566603698641164e-46 ], 
                    [ 188.0 , 2.5573596209925755e-46 ], [ 188.0625 , 2.55805906336749e-46 ], [ 188.125 , 2.5587585722298226e-46 ], [ 188.1875 , 2.5594575302385435e-46 ], 
                    [ 188.25 , 2.5601566791766514e-46 ], [ 188.3125 , 2.560856019096373e-46 ], [ 188.375 , 2.5615555500498423e-46 ], [ 188.4375 , 2.562255272089169e-46 ], 
                    [ 188.5 , 2.5629551852667146e-46 ], [ 188.5625 , 2.5636554758958122e-46 ], [ 188.625 , 2.564356634113405e-46 ], [ 188.6875 , 2.565057984097363e-46 ], 
                    [ 188.75 , 2.565759525900098e-46 ], [ 188.8125 , 2.566461259574091e-46 ], [ 188.875 , 2.5671631851718176e-46 ], [ 188.9375 , 2.567865302745788e-46 ], 
                    [ 189.0 , 2.568567841090371e-46 ], [ 189.0625 , 2.5692712366043645e-46 ], [ 189.125 , 2.569974824741397e-46 ], [ 189.1875 , 2.5706786055541447e-46 ], 
                    [ 189.25 , 2.5713825790954258e-46 ], [ 189.3125 , 2.572086745418037e-46 ], [ 189.375 , 2.5727911045746614e-46 ], [ 189.4375 , 2.5734953960747734e-46 ], 
                    [ 189.5 , 2.574199210219684e-46 ], [ 189.5625 , 2.574903216847686e-46 ], [ 189.625 , 2.5756074160114218e-46 ], [ 189.6875 , 2.5763118077635456e-46 ], 
                    [ 189.75 , 2.5770163921566924e-46 ], [ 189.8125 , 2.5777211692436384e-46 ], [ 189.875 , 2.578425871934262e-46 ], [ 189.9375 , 2.579130106661281e-46 ], 
                    [ 190.0 , 2.579834533733e-46 ], [ 190.0625 , 2.580539153201935e-46 ], [ 190.125 , 2.5812439651206898e-46 ], [ 190.1875 , 2.5819489695417535e-46 ], 
                    [ 190.25 , 2.58265416651774e-46 ], [ 190.3125 , 2.583361194415397e-46 ], [ 190.375 , 2.584072629835061e-46 ], [ 190.4375 , 2.5847842611778743e-46 ], 
                    [ 190.5 , 2.5854960884978485e-46 ], [ 190.5625 , 2.5862081118489343e-46 ], [ 190.625 , 2.5869203312851177e-46 ], [ 190.6875 , 2.587632746860399e-46 ], 
                    [ 190.75 , 2.5883450812830202e-46 ], [ 190.8125 , 2.589056869288144e-46 ], [ 190.875 , 2.5897688530330374e-46 ], [ 190.9375 , 2.5904810325716196e-46 ], 
                    [ 191.0 , 2.5911934079576417e-46 ], [ 191.0625 , 2.5919059792450162e-46 ], [ 191.125 , 2.5926187464875793e-46 ], [ 191.1875 , 2.5933307669741093e-46 ], 
                    [ 191.25 , 2.5940398612730792e-46 ], [ 191.3125 , 2.594749149459611e-46 ], [ 191.375 , 2.595458631586794e-46 ], [ 191.4375 , 2.596168307707676e-46 ], 
                    [ 191.5 , 2.596878177875208e-46 ], [ 191.5625 , 2.597588242142503e-46 ], [ 191.625 , 2.5982991957610308e-46 ], [ 191.6875 , 2.5990135996512687e-46 ], 
                    [ 191.75 , 2.5997281999672575e-46 ], [ 191.8125 , 2.6004429967630417e-46 ], [ 191.875 , 2.6011579900925876e-46 ], [ 191.9375 , 2.6018731800100437e-46 ], 
                    [ 192.0 , 2.6025885665693503e-46 ], [ 192.0625 , 2.603304234944356e-46 ], [ 192.125 , 2.6040208336575527e-46 ], [ 192.1875 , 2.6047376296253597e-46 ], 
                    [ 192.25 , 2.6054546229020203e-46 ], [ 192.3125 , 2.606171813541882e-46 ], [ 192.375 , 2.606889201599329e-46 ], [ 192.4375 , 2.607606787128573e-46 ], 
                    [ 192.5 , 2.6083245546554706e-46 ], [ 192.5625 , 2.6090415644802112e-46 ], [ 192.625 , 2.609758771405823e-46 ], [ 192.6875 , 2.6104761754865237e-46 ], 
                    [ 192.75 , 2.6111937767765103e-46 ], [ 192.8125 , 2.61191157532992e-46 ], [ 192.875 , 2.6126295712010903e-46 ], [ 192.9375 , 2.613347764444225e-46 ], 
                    [ 193.0 , 2.6140668867100775e-46 ], [ 193.0625 , 2.6147862774710044e-46 ], [ 193.125 , 2.61550586620809e-46 ], [ 193.1875 , 2.616225652975873e-46 ], 
                    [ 193.25 , 2.6169456378289266e-46 ], [ 193.3125 , 2.6176658208215766e-46 ], [ 193.375 , 2.6183862020085194e-46 ], [ 193.4375 , 2.6191086322179972e-46 ], 
                    [ 193.5 , 2.619831747946915e-46 ], [ 193.5625 , 2.6205550633225175e-46 ], [ 193.625 , 2.6212785783998897e-46 ], [ 193.6875 , 2.6220022932342235e-46 ], 
                    [ 193.75 , 2.622726207880688e-46 ], [ 193.8125 , 2.623450322394413e-46 ], [ 193.875 , 2.6241739861500625e-46 ], [ 193.9375 , 2.6248975082196566e-46 ], 
                    [ 194.0 , 2.6256212297745887e-46 ], [ 194.0625 , 2.6263451508698227e-46 ], [ 194.125 , 2.6270692715603926e-46 ], [ 194.1875 , 2.6277935919013297e-46 ], 
                    [ 194.25 , 2.6285181119476996e-46 ], [ 194.3125 , 2.629243997674767e-46 ], [ 194.375 , 2.629971231591795e-46 ], [ 194.4375 , 2.6306986666575454e-46 ], 
                    [ 194.5 , 2.631426302927711e-46 ], [ 194.5625 , 2.6321541404579073e-46 ], [ 194.625 , 2.6328821793038193e-46 ], [ 194.6875 , 2.6336104195211857e-46 ], 
                    [ 194.75 , 2.6343380047410735e-46 ], [ 194.8125 , 2.635064098045596e-46 ], [ 194.875 , 2.6357903914806016e-46 ], [ 194.9375 , 2.636516885101365e-46 ], 
                    [ 195.0 , 2.6372435789629867e-46 ], [ 195.0625 , 2.637970473120697e-46 ], [ 195.125 , 2.6386975676296456e-46 ], [ 195.1875 , 2.639425211364614e-46 ], 
                    [ 195.25 , 2.6401550023368316e-46 ], [ 195.3125 , 2.640884995093461e-46 ], [ 195.375 , 2.6416151896902573e-46 ], [ 195.4375 , 2.642345586183122e-46 ], 
                    [ 195.5 , 2.6430761846277843e-46 ], [ 195.5625 , 2.643806985080103e-46 ], [ 195.625 , 2.644537987595988e-46 ], [ 195.6875 , 2.6452698885908884e-46 ], 
                    [ 195.75 , 2.6460020272666347e-46 ], [ 195.8125 , 2.64673436857843e-46 ], [ 195.875 , 2.6474669125823216e-46 ], [ 195.9375 , 2.6481996593344657e-46 ], 
                    [ 196.0 , 2.6489326088910146e-46 ], [ 196.0625 , 2.6496657613079873e-46 ], [ 196.125 , 2.6503983664019105e-46 ], [ 196.1875 , 2.6511309048715145e-46 ], 
                    [ 196.25 , 2.6518636458059502e-46 ], [ 196.3125 , 2.652596589261271e-46 ], [ 196.375 , 2.6533297352933942e-46 ], [ 196.4375 , 2.6540630839583292e-46 ], 
                    [ 196.5 , 2.654796635312081e-46 ], [ 196.5625 , 2.65553152693543e-46 ], [ 196.625 , 2.6562677417681652e-46 ], [ 196.6875 , 2.6570041607078275e-46 ], 
                    [ 196.75 , 2.6577407838109087e-46 ], [ 196.8125 , 2.658477611134106e-46 ], [ 196.875 , 2.65921464273396e-46 ], [ 196.9375 , 2.659951878667181e-46 ], 
                    [ 197.0 , 2.6606890551878664e-46 ], [ 197.0625 , 2.661425669120644e-46 ], [ 197.125 , 2.662162486985589e-46 ], [ 197.1875 , 2.6628995088391787e-46 ], 
                    [ 197.25 , 2.6636367347378315e-46 ], [ 197.3125 , 2.664374164738132e-46 ], [ 197.375 , 2.6651117988965094e-46 ], [ 197.4375 , 2.6658496372694672e-46 ], 
                    [ 197.5 , 2.6665867621723574e-46 ], [ 197.5625 , 2.667324083491918e-46 ], [ 197.625 , 2.6680616086836107e-46 ], [ 197.6875 , 2.6687993378038444e-46 ], 
                    [ 197.75 , 2.669537270908931e-46 ], [ 197.8125 , 2.6702754080552916e-46 ], [ 197.875 , 2.671013749299401e-46 ], [ 197.9375 , 2.671754969374107e-46 ], 
                    [ 198.0 , 2.672497563715709e-46 ], [ 198.0625 , 2.673240364455901e-46 ], [ 198.125 , 2.6739833716519922e-46 ], [ 198.1875 , 2.674726585361385e-46 ], 
                    [ 198.25 , 2.6754700056414783e-46 ], [ 198.3125 , 2.6762136325496674e-46 ], [ 198.375 , 2.6769557833184845e-46 ], [ 198.4375 , 2.67769561496038e-46 ], 
                    [ 198.5 , 2.678435651069907e-46 ], [ 198.5625 , 2.67917589170365e-46 ], [ 198.625 , 2.679916336918038e-46 ], [ 198.6875 , 2.6806569867696496e-46 ], 
                    [ 198.75 , 2.6813978413150217e-46 ], [ 198.8125 , 2.6821392865099064e-46 ], [ 198.875 , 2.6828859119159333e-46 ], [ 198.9375 , 2.6836327451595994e-46 ], 
                    [ 199.0 , 2.684379786298741e-46 ], [ 199.0625 , 2.6851270353912678e-46 ], [ 199.125 , 2.6858744924950874e-46 ], [ 199.1875 , 2.6866221576679886e-46 ], 
                    [ 199.25 , 2.6873700309680256e-46 ], [ 199.3125 , 2.6881185901795794e-46 ], [ 199.375 , 2.6888675223250012e-46 ], [ 199.4375 , 2.6896166631291335e-46 ], 
                    [ 199.5 , 2.690366012650129e-46 ], [ 199.5625 , 2.6911155709460622e-46 ], [ 199.625 , 2.6918653380751564e-46 ], [ 199.6875 , 2.6926153140956714e-46 ], 
                    [ 199.75 , 2.6933638079803144e-46 ], [ 199.8125 , 2.6941099726003686e-46 ], [ 199.875 , 2.6948563439364948e-46 ], [ 199.9375 , 2.695602922046e-46 ], 
                    [ 200.0 , 2.6963497069861498e-46 ], [ 200.0625 , 2.697096698814224e-46 ], [ 200.125 , 2.697843897587559e-46 ], [ 200.1875 , 2.6985913552934277e-46 ], 
                    [ 200.25 , 2.699341132490756e-46 ], [ 200.3125 , 2.7000911180063592e-46 ], [ 200.375 , 2.7008413118980403e-46 ], [ 200.4375 , 2.7015917142238098e-46 ], 
                    [ 200.5 , 2.7023423250414833e-46 ], [ 200.5625 , 2.7030931444090647e-46 ], [ 200.625 , 2.703844172384459e-46 ], [ 200.6875 , 2.7045957580029745e-46 ], 
                    [ 200.75 , 2.7053477557977486e-46 ], [ 200.8125 , 2.706099962681343e-46 ], [ 200.875 , 2.7068523787118545e-46 ], [ 200.9375 , 2.7076050039474167e-46 ], 
                    [ 201.0 , 2.7083578384462934e-46 ], [ 201.0625 , 2.7091108822665733e-46 ], [ 201.125 , 2.7098642590723003e-46 ], [ 201.1875 , 2.7106182369118286e-46 ], 
                    [ 201.25 , 2.7113724245340353e-46 ], [ 201.3125 , 2.712126821997251e-46 ], [ 201.375 , 2.7128814293599374e-46 ], [ 201.4375 , 2.713636246680457e-46 ], 
                    [ 201.5 , 2.7143912740171898e-46 ], [ 201.5625 , 2.7151465114286453e-46 ], [ 201.625 , 2.7159024449827475e-46 ], [ 201.6875 , 2.716658698619857e-46 ], 
                    [ 201.75 , 2.7174151628387436e-46 ], [ 201.8125 , 2.7181718376981025e-46 ], [ 201.875 , 2.718928723256491e-46 ], [ 201.9375 , 2.719685819572579e-46 ], 
                    [ 202.0 , 2.7204431267051496e-46 ], [ 202.0625 , 2.7212002342118743e-46 ], [ 202.125 , 2.721956870977561e-46 ], [ 202.1875 , 2.7227137181280773e-46 ], 
                    [ 202.25 , 2.723470775721902e-46 ], [ 202.3125 , 2.724228043817529e-46 ], [ 202.375 , 2.72498552247353e-46 ], [ 202.4375 , 2.7257432117484302e-46 ], 
                    [ 202.5 , 2.7265011117007747e-46 ], [ 202.5625 , 2.7272611604991744e-46 ], [ 202.625 , 2.728021589749722e-46 ], [ 202.6875 , 2.728782231027139e-46 ], 
                    [ 202.75 , 2.7295430843905244e-46 ], [ 202.8125 , 2.7303041498990515e-46 ], [ 202.875 , 2.7310654276118326e-46 ], [ 202.9375 , 2.7318269175880753e-46 ], 
                    [ 203.0 , 2.7325874427958384e-46 ], [ 203.0625 , 2.7333467289821006e-46 ], [ 203.125 , 2.734106226146206e-46 ], [ 203.1875 , 2.7348659343468375e-46 ], 
                    [ 203.25 , 2.735625853642574e-46 ], [ 203.3125 , 2.73638598409217e-46 ], [ 203.375 , 2.737146325754219e-46 ], [ 203.4375 , 2.737906878687448e-46 ], 
                    [ 203.5 , 2.7386696495527494e-46 ], [ 203.5625 , 2.739432716555005e-46 ], [ 203.625 , 2.7401959961682254e-46 ], [ 203.6875 , 2.7409594884517274e-46 ], 
                    [ 203.75 , 2.7417231934646696e-46 ], [ 203.8125 , 2.7424871112663437e-46 ], [ 203.875 , 2.7432512419161166e-46 ], [ 203.9375 , 2.7440165298101263e-46 ], 
                    [ 204.0 , 2.744783158089548e-46 ], [ 204.0625 , 2.7455500005509394e-46 ], [ 204.125 , 2.7463170572541587e-46 ], [ 204.1875 , 2.747084328259061e-46 ], 
                    [ 204.25 , 2.7478518136255573e-46 ], [ 204.3125 , 2.7486195134134384e-46 ], [ 204.375 , 2.749387427682669e-46 ], [ 204.4375 , 2.750151562518305e-46 ], 
                    [ 204.5 , 2.7509156365612887e-46 ], [ 204.5625 , 2.7516799228867476e-46 ], [ 204.625 , 2.7524444215537783e-46 ], [ 204.6875 , 2.7532091326212765e-46 ], 
                    [ 204.75 , 2.7539740561482745e-46 ], [ 204.8125 , 2.7547391921938968e-46 ], [ 204.875 , 2.7555059538893034e-46 ], [ 204.9375 , 2.75627519644315e-46 ], 
                    [ 205.0 , 2.7570446537431177e-46 ], [ 205.0625 , 2.7578143258490967e-46 ], [ 205.125 , 2.7585842128210738e-46 ], [ 205.1875 , 2.759354314719032e-46 ], 
                    [ 205.25 , 2.760124631602991e-46 ], [ 205.3125 , 2.7608951635329278e-46 ], [ 205.375 , 2.7616676046305557e-46 ], [ 205.4375 , 2.7624406040071265e-46 ], 
                    [ 205.5 , 2.7632138197486537e-46 ], [ 205.5625 , 2.7639872519156387e-46 ], [ 205.625 , 2.7647609005687594e-46 ], [ 205.6875 , 2.765534765768531e-46 ], 
                    [ 205.75 , 2.7663088475755668e-46 ], [ 205.8125 , 2.7670820144212815e-46 ], [ 205.875 , 2.767852241638229e-46 ], [ 205.9375 , 2.76862268325065e-46 ], 
                    [ 206.0 , 2.7693933393182805e-46 ], [ 206.0625 , 2.7701642099007552e-46 ], [ 206.125 , 2.770935295057885e-46 ], [ 206.1875 , 2.7717065948492784e-46 ], 
                    [ 206.25 , 2.7724781093347582e-46 ], [ 206.3125 , 2.7732511794037086e-46 ], [ 206.375 , 2.7740251437149197e-46 ], [ 206.4375 , 2.7747993240255703e-46 ], 
                    [ 206.5 , 2.775573720395922e-46 ], [ 206.5625 , 2.7763483328863718e-46 ], [ 206.625 , 2.777123161557157e-46 ], [ 206.6875 , 2.7778982064686474e-46 ], 
                    [ 206.75 , 2.7786736420890128e-46 ], [ 206.8125 , 2.7794527542255442e-46 ], [ 206.875 , 2.7802320848172613e-46 ], [ 206.9375 , 2.781011633925476e-46 ], 
                    [ 207.0 , 2.7817914016113788e-46 ], [ 207.0625 , 2.782571387936336e-46 ], [ 207.125 , 2.7833515929615927e-46 ], [ 207.1875 , 2.7841320167484706e-46 ], 
                    [ 207.25 , 2.784911453057524e-46 ], [ 207.3125 , 2.7856895210695163e-46 ], [ 207.375 , 2.786467806463584e-46 ], [ 207.4375 , 2.787246309300401e-46 ], 
                    [ 207.5 , 2.7880250296408374e-46 ], [ 207.5625 , 2.788803967545621e-46 ], [ 207.625 , 2.789583123075477e-46 ], [ 207.6875 , 2.7903624962912463e-46 ], 
                    [ 207.75 , 2.7911399967728547e-46 ], [ 207.8125 , 2.7919171377864195e-46 ], [ 207.875 , 2.792694495180483e-46 ], [ 207.9375 , 2.793472069015234e-46 ], 
                    [ 208.0 , 2.794249859350955e-46 ], [ 208.0625 , 2.7950278662479472e-46 ], [ 208.125 , 2.795806089766468e-46 ], [ 208.1875 , 2.796585143235587e-46 ], 
                    [ 208.25 , 2.797368912349812e-46 ], [ 208.3125 , 2.798152901122653e-46 ], [ 208.375 , 2.798937109615692e-46 ], [ 208.4375 , 2.7997215378904482e-46 ], 
                    [ 208.5 , 2.800506186008577e-46 ], [ 208.5625 , 2.801291054031672e-46 ], [ 208.625 , 2.802076142021383e-46 ], [ 208.6875 , 2.802860267730543e-46 ], 
                    [ 208.75 , 2.8036431076903924e-46 ], [ 208.8125 , 2.804426166297766e-46 ], [ 208.875 , 2.8052094436136922e-46 ], [ 208.9375 , 2.805992939699257e-46 ], 
                    [ 209.0 , 2.8067766546156214e-46 ], [ 209.0625 , 2.807560588423867e-46 ], [ 209.125 , 2.808344741185089e-46 ], [ 209.1875 , 2.8091293533963947e-46 ], 
                    [ 209.25 , 2.8099142675711e-46 ], [ 209.3125 , 2.8106994010629817e-46 ], [ 209.375 , 2.8114847539333206e-46 ], [ 209.4375 , 2.8122703262433534e-46 ], 
                    [ 209.5 , 2.8130561180544767e-46 ], [ 209.5625 , 2.813842129427941e-46 ], [ 209.625 , 2.814628367642525e-46 ], [ 209.6875 , 2.8154152695402746e-46 ], 
                    [ 209.75 , 2.8162023914366695e-46 ], [ 209.8125 , 2.8169897333933154e-46 ], [ 209.875 , 2.817777295471677e-46 ], [ 209.9375 , 2.8185650777333343e-46 ], 
                    [ 210.0 , 2.819353080239805e-46 ], [ 210.0625 , 2.820141303052725e-46 ], [ 210.125 , 2.8209298330164276e-46 ], [ 210.1875 , 2.821718798064691e-46 ], 
                    [ 210.25 , 2.8225079837728064e-46 ], [ 210.3125 , 2.823297390202449e-46 ], [ 210.375 , 2.8240870174153714e-46 ], [ 210.4375 , 2.824876865473201e-46 ], 
                    [ 210.5 , 2.825666934437926e-46 ], [ 210.5625 , 2.826457224371089e-46 ], [ 210.625 , 2.827247172025998e-46 ], [ 210.6875 , 2.8280368367535836e-46 ], 
                    [ 210.75 , 2.8288267220385832e-46 ], [ 210.8125 , 2.82961682794266e-46 ], [ 210.875 , 2.8304071545273537e-46 ], [ 210.9375 , 2.8311977018543215e-46 ], 
                    [ 211.0 , 2.8319884699853183e-46 ], [ 211.0625 , 2.8327794589818755e-46 ], [ 211.125 , 2.833570889780687e-46 ], [ 211.1875 , 2.834362608433161e-46 ], 
                    [ 211.25 , 2.8351545482971192e-46 ], [ 211.3125 , 2.8359467094343716e-46 ], [ 211.375 , 2.8367390919067414e-46 ], [ 211.4375 , 2.837531695776032e-46 ], 
                    [ 211.5 , 2.8383245211042442e-46 ], [ 211.5625 , 2.8391175679530534e-46 ], [ 211.625 , 2.8399126887374074e-46 ], [ 211.6875 , 2.840708077765021e-46 ], 
                    [ 211.75 , 2.841503689561372e-46 ], [ 211.8125 , 2.842299524188811e-46 ], [ 211.875 , 2.8430955817097686e-46 ], [ 211.9375 , 2.8438918621866713e-46 ], 
                    [ 212.0 , 2.844688365681943e-46 ], [ 212.0625 , 2.8454851552878593e-46 ], [ 212.125 , 2.8462824803071487e-46 ], [ 212.1875 , 2.847080028742534e-46 ], 
                    [ 212.25 , 2.847877800656599e-46 ], [ 212.3125 , 2.848675796112024e-46 ], [ 212.375 , 2.8494740151713868e-46 ], [ 212.4375 , 2.8502724578973623e-46 ], 
                    [ 212.5 , 2.8510711243526436e-46 ], [ 212.5625 , 2.85186959749736e-46 ], [ 212.625 , 2.8526674986357836e-46 ], [ 212.6875 , 2.8534656230124445e-46 ], 
                    [ 212.75 , 2.8542639706897197e-46 ], [ 212.8125 , 2.8550625417301658e-46 ], [ 212.875 , 2.855861336196195e-46 ], [ 212.9375 , 2.8566603541503373e-46 ], 
                    [ 213.0 , 2.8574595956552017e-46 ], [ 213.0625 , 2.8582600007516238e-46 ], [ 213.125 , 2.85906155536763e-46 ], [ 213.1875 , 2.8598633347671365e-46 ], 
                    [ 213.25 , 2.860665339013281e-46 ], [ 213.3125 , 2.861467568169017e-46 ], [ 213.375 , 2.8622700222974783e-46 ], [ 213.4375 , 2.8630727014616938e-46 ], 
                    [ 213.5 , 2.863875605724832e-46 ], [ 213.5625 , 2.8646768885875752e-46 ], [ 213.625 , 2.8654773922898963e-46 ], [ 213.6875 , 2.8662781196845018e-46 ], 
                    [ 213.75 , 2.867079070833922e-46 ], [ 213.8125 , 2.8678802458006612e-46 ], [ 213.875 , 2.8686816446473243e-46 ], [ 213.9375 , 2.8694832674364108e-46 ], 
                    [ 214.0 , 2.8702851142305e-46 ], [ 214.0625 , 2.8710886854917446e-46 ], [ 214.125 , 2.8718929556657235e-46 ], [ 214.1875 , 2.8726974511376525e-46 ], 
                    [ 214.25 , 2.8735021719707456e-46 ], [ 214.3125 , 2.8743071182280515e-46 ], [ 214.375 , 2.8751122899727167e-46 ], [ 214.4375 , 2.8759176872679684e-46 ], 
                    [ 214.5 , 2.8767233101769077e-46 ], [ 214.5625 , 2.8775293341084587e-46 ], [ 214.625 , 2.8783356096244257e-46 ], [ 214.6875 , 2.8791421110564524e-46 ], 
                    [ 214.75 , 2.879948838467943e-46 ], [ 214.8125 , 2.880755791922154e-46 ], [ 214.875 , 2.8815629714823818e-46 ], [ 214.9375 , 2.8823703772120412e-46 ], 
                    [ 215.0 , 2.883178009174505e-46 ], [ 215.0625 , 2.8839832648174873e-46 ], [ 215.125 , 2.8847886141030052e-46 ], [ 215.1875 , 2.8855941882814617e-46 ], 
                    [ 215.25 , 2.8863999874156788e-46 ], [ 215.3125 , 2.887206011568434e-46 ], [ 215.375 , 2.888012260802604e-46 ], [ 215.4375 , 2.8888187351810017e-46 ], 
                    [ 215.5 , 2.8896254878070408e-46 ], [ 215.5625 , 2.890435728636791e-46 ], [ 215.625 , 2.891246196655222e-46 ], [ 215.6875 , 2.8920568919260573e-46 ], 
                    [ 215.75 , 2.8928678145129967e-46 ], [ 215.8125 , 2.8936789644797588e-46 ], [ 215.875 , 2.8944903418900994e-46 ], [ 215.9375 , 2.8953019468078143e-46 ], 
                    [ 216.0 , 2.8961137984364515e-46 ], [ 216.0625 , 2.8969261577076676e-46 ], [ 216.125 , 2.8977387448454567e-46 ], [ 216.1875 , 2.898551559913776e-46 ], 
                    [ 216.25 , 2.8993646029765404e-46 ], [ 216.3125 , 2.9001778740976597e-46 ], [ 216.375 , 2.9009913733411878e-46 ], [ 216.4375 , 2.901805100771092e-46 ], 
                    [ 216.5 , 2.9026187924457415e-46 ], [ 216.5625 , 2.9034302255383604e-46 ], [ 216.625 , 2.90424188546878e-46 ], [ 216.6875 , 2.905053772300434e-46 ], 
                    [ 216.75 , 2.9058658860966914e-46 ], [ 216.8125 , 2.9066782269210832e-46 ], [ 216.875 , 2.907490794837034e-46 ], [ 216.9375 , 2.9083035899080077e-46 ], 
                    [ 217.0 , 2.9091169797813904e-46 ], [ 217.0625 , 2.9099335123739733e-46 ], [ 217.125 , 2.9107502741513425e-46 ], [ 217.1875 , 2.911567265177909e-46 ], 
                    [ 217.25 , 2.9123844855179977e-46 ], [ 217.3125 , 2.9132019352359103e-46 ], [ 217.375 , 2.914019614396112e-46 ], [ 217.4375 , 2.9148375230628986e-46 ], 
                    [ 217.5 , 2.915655170342478e-46 ], [ 217.5625 , 2.9164691532949654e-46 ], [ 217.625 , 2.917283363492518e-46 ], [ 217.6875 , 2.9180978009986172e-46 ], 
                    [ 217.75 , 2.9189124658766197e-46 ], [ 217.8125 , 2.919727358190064e-46 ], [ 217.875 , 2.9205424780025078e-46 ], [ 217.9375 , 2.921357825377359e-46 ], 
                    [ 218.0 , 2.922173783217377e-46 ], [ 218.0625 , 2.922994372844666e-46 ], [ 218.125 , 2.9238151929056406e-46 ], [ 218.1875 , 2.9246362434650317e-46 ], 
                    [ 218.25 , 2.9254575245875864e-46 ], [ 218.3125 , 2.926279036337988e-46 ], [ 218.375 , 2.9271007787810844e-46 ], [ 218.4375 , 2.927922751981594e-46 ], 
                    [ 218.5 , 2.9287448228519466e-46 ], [ 218.5625 , 2.9295630955460345e-46 ], [ 218.625 , 2.9303815968603133e-46 ], [ 218.6875 , 2.9312003268586382e-46 ], 
                    [ 218.75 , 2.9320192856049643e-46 ], [ 218.8125 , 2.932838473163139e-46 ], [ 218.875 , 2.9336578895971124e-46 ], [ 218.9375 , 2.934477534970831e-46 ], 
                    [ 219.0 , 2.93529740934828e-46 ], [ 219.0625 , 2.9361205997414726e-46 ], [ 219.125 , 2.936944071202938e-46 ], [ 219.1875 , 2.9377677736172353e-46 ], 
                    [ 219.25 , 2.938591707049077e-46 ], [ 219.3125 , 2.939415871563275e-46 ], [ 219.375 , 2.94024026722464e-46 ], [ 219.4375 , 2.941064894097999e-46 ], 
                    [ 219.5 , 2.9418897522482203e-46 ], [ 219.5625 , 2.9427122109824096e-46 ], [ 219.625 , 2.9435346201994592e-46 ], [ 219.6875 , 2.944357259257808e-46 ], 
                    [ 219.75 , 2.9451801282217734e-46 ], [ 219.8125 , 2.9460032271555246e-46 ], [ 219.875 , 2.946826556123416e-46 ], [ 219.9375 , 2.9476501151896734e-46 ], 
                    [ 220.0 , 2.948473904418602e-46 ], [ 220.0625 , 2.949300575961269e-46 ], [ 220.125 , 2.950128175983306e-46 ], [ 220.1875 , 2.950956008237284e-46 ], 
                    [ 220.25 , 2.95178407278839e-46 ], [ 220.3125 , 2.952612369701788e-46 ], [ 220.375 , 2.9534408990426384e-46 ], [ 220.4375 , 2.9542696608762467e-46 ], 
                    [ 220.5 , 2.9550986552677685e-46 ], [ 220.5625 , 2.9559259889071524e-46 ], [ 220.625 , 2.9567526629244094e-46 ], [ 220.6875 , 2.957579568134804e-46 ], 
                    [ 220.75 , 2.9584067046031e-46 ], [ 220.8125 , 2.9592340723938874e-46 ], [ 220.875 , 2.96006167157186e-46 ], [ 220.9375 , 2.9608895022017703e-46 ], 
                    [ 221.0 , 2.9617175643483063e-46 ], [ 221.0625 , 2.9625483709361775e-46 ], [ 221.125 , 2.9633814505740505e-46 ], [ 221.1875 , 2.9642147644770427e-46 ], 
                    [ 221.25 , 2.9650483127109664e-46 ], [ 221.3125 , 2.9658820953418007e-46 ], [ 221.375 , 2.9667161124353953e-46 ], [ 221.4375 , 2.967550364057725e-46 ], 
                    [ 221.5 , 2.9683848502746765e-46 ], [ 221.5625 , 2.969216721548553e-46 ], [ 221.625 , 2.970044404922372e-46 ], [ 221.6875 , 2.970872319016957e-46 ], 
                    [ 221.75 , 2.971700463896516e-46 ], [ 221.8125 , 2.9725288396254242e-46 ], [ 221.875 , 2.9733574462680746e-46 ], [ 221.9375 , 2.9741862838887507e-46 ], 
                    [ 222.0 , 2.9750153525518385e-46 ], [ 222.0625 , 2.975845982251256e-46 ], [ 222.125 , 2.9766816723678063e-46 ], [ 222.1875 , 2.977517597166561e-46 ], 
                    [ 222.25 , 2.978353756713341e-46 ], [ 222.3125 , 2.9791901510741533e-46 ], [ 222.375 , 2.980026780314876e-46 ], [ 222.4375 , 2.980863644501533e-46 ], 
                    [ 222.5 , 2.9817007436999962e-46 ], [ 222.5625 , 2.9825380106708087e-46 ], [ 222.625 , 2.983372774888125e-46 ], [ 222.6875 , 2.9842077727424843e-46 ], 
                    [ 222.75 , 2.9850430042992142e-46 ], [ 222.8125 , 2.985878469623788e-46 ], [ 222.875 , 2.986714168781612e-46 ], [ 222.9375 , 2.9875501018381545e-46 ], 
                    [ 223.0 , 2.9883862688588576e-46 ], [ 223.0625 , 2.9892226699091625e-46 ], [ 223.125 , 2.990059330695179e-46 ], [ 223.1875 , 2.990896231439495e-46 ], 
                    [ 223.25 , 2.9917333664276306e-46 ], [ 223.3125 , 2.9925707357251285e-46 ], [ 223.375 , 2.993408339397549e-46 ], [ 223.4375 , 2.994246177510514e-46 ], 
                    [ 223.5 , 2.9950842501297057e-46 ], [ 223.5625 , 2.9959225573205906e-46 ], [ 223.625 , 2.9967612757906076e-46 ], [ 223.6875 , 2.997600359126784e-46 ], 
                    [ 223.75 , 2.99843967740363e-46 ], [ 223.8125 , 2.9992792306867788e-46 ], [ 223.875 , 3.0001190190420957e-46 ], [ 223.9375 , 3.0009590425354423e-46 ], 
                    [ 224.0 , 3.0017993012325928e-46 ], [ 224.0625 , 3.0026397951994466e-46 ], [ 224.125 , 3.003480030968293e-46 ], [ 224.1875 , 3.004319526273311e-46 ], 
                    [ 224.25 , 3.0051592562236156e-46 ], [ 224.3125 , 3.005999220884749e-46 ], [ 224.375 , 3.0068394203222924e-46 ], [ 224.4375 , 3.0076798546019967e-46 ], 
                    [ 224.5 , 3.0085205237893952e-46 ], [ 224.5625 , 3.0093614279501457e-46 ], [ 224.625 , 3.010202671432956e-46 ], [ 224.6875 , 3.011045675645075e-46 ], 
                    [ 224.75 , 3.0118889159396487e-46 ], [ 224.8125 , 3.0127323923828137e-46 ], [ 224.875 , 3.013576105040682e-46 ], [ 224.9375 , 3.014420053979469e-46 ], 
                    [ 225.0 , 3.0152642392652804e-46 ], [ 225.0625 , 3.0161086609643055e-46 ], [ 225.125 , 3.016953319142795e-46 ], [ 225.1875 , 3.0177971762969747e-46 ], 
                    [ 225.25 , 3.0186409698938135e-46 ], [ 225.3125 , 3.0194849994201944e-46 ], [ 225.375 , 3.020329264942191e-46 ], [ 225.4375 , 3.021173766525682e-46 ], 
                    [ 225.5 , 3.02201850423678e-46 ], [ 225.5625 , 3.022863478141356e-46 ], [ 225.625 , 3.023708688305602e-46 ], [ 225.6875 , 3.0245541282482916e-46 ], 
                    [ 225.75 , 3.0253997972529675e-46 ], [ 225.8125 , 3.0262457027077454e-46 ], [ 225.875 , 3.027091844678736e-46 ], [ 225.9375 , 3.0279382232320058e-46 ], 
                    [ 226.0 , 3.028784838433789e-46 ], [ 226.0625 , 3.0296316903501674e-46 ], [ 226.125 , 3.0304787790474118e-46 ], [ 226.1875 , 3.0313263420278758e-46 ], 
                    [ 226.25 , 3.0321756508950072e-46 ], [ 226.3125 , 3.033025197719193e-46 ], [ 226.375 , 3.0338749825671034e-46 ], [ 226.4375 , 3.034725005505491e-46 ], 
                    [ 226.5 , 3.035575266600978e-46 ], [ 226.5625 , 3.03642576592042e-46 ], [ 226.625 , 3.0372765035304108e-46 ], [ 226.6875 , 3.038127479497778e-46 ], 
                    [ 226.75 , 3.038976309940034e-46 ], [ 226.8125 , 3.0398247813857727e-46 ], [ 226.875 , 3.0406734897217618e-46 ], [ 226.9375 , 3.041522435014098e-46 ], 
                    [ 227.0 , 3.0423716173289587e-46 ], [ 227.0625 , 3.0432210367325426e-46 ], [ 227.125 , 3.0440706932910007e-46 ], [ 227.1875 , 3.0449205870705454e-46 ], 
                    [ 227.25 , 3.045771402085707e-46 ], [ 227.3125 , 3.0466233543532912e-46 ], [ 227.375 , 3.04747554492589e-46 ], [ 227.4375 , 3.048327973870227e-46 ], 
                    [ 227.5 , 3.0491806412528703e-46 ], [ 227.5625 , 3.0500335471406445e-46 ], [ 227.625 , 3.050886691600156e-46 ], [ 227.6875 , 3.051740074698201e-46 ], 
                    [ 227.75 , 3.052593771702113e-46 ], [ 227.8125 , 3.0534491994164762e-46 ], [ 227.875 , 3.054304866847104e-46 ], [ 227.9375 , 3.0551607740612805e-46 ], 
                    [ 228.0 , 3.0560169211261565e-46 ], [ 228.0625 , 3.056873308108945e-46 ], [ 228.125 , 3.0577299350768795e-46 ], [ 228.1875 , 3.058586802097188e-46 ], 
                    [ 228.25 , 3.0594439092372056e-46 ], [ 228.3125 , 3.060301305211203e-46 ], [ 228.375 , 3.0611589678997008e-46 ], [ 228.4375 , 3.0620168709518795e-46 ], 
                    [ 228.5 , 3.0628750144350588e-46 ], [ 228.5625 , 3.063733398416708e-46 ], [ 228.625 , 3.0645920229641616e-46 ], [ 228.6875 , 3.06545088814484e-46 ], 
                    [ 228.75 , 3.0663099940262468e-46 ], [ 228.8125 , 3.0671686437261563e-46 ], [ 228.875 , 3.068025226082673e-46 ], [ 228.9375 , 3.0688820476609232e-46 ], 
                    [ 229.0 , 3.069739108527672e-46 ], [ 229.0625 , 3.0705964087498123e-46 ], [ 229.125 , 3.0714539483940806e-46 ], [ 229.1875 , 3.0723117275274506e-46 ], 
                    [ 229.25 , 3.0731697462167618e-46 ], [ 229.3125 , 3.0740280045289162e-46 ], [ 229.375 , 3.07488542811052e-46 ], [ 229.4375 , 3.075742794752886e-46 ], 
                    [ 229.5 , 3.0766004004537716e-46 ], [ 229.5625 , 3.077458245279812e-46 ], [ 229.625 , 3.0783163292977253e-46 ], [ 229.6875 , 3.07917465257414e-46 ], 
                    [ 229.75 , 3.080033215175855e-46 ], [ 229.8125 , 3.080892017169515e-46 ], [ 229.875 , 3.081752009352066e-46 ], [ 229.9375 , 3.0826142624569533e-46 ], 
                    [ 230.0 , 3.083476756814397e-46 ], [ 230.0625 , 3.084339492491833e-46 ], [ 230.125 , 3.085202469556824e-46 ], [ 230.1875 , 3.0860656880769097e-46 ], 
                    [ 230.25 , 3.0869291481195586e-46 ], [ 230.3125 , 3.087792849752414e-46 ], [ 230.375 , 3.088656793043136e-46 ], [ 230.4375 , 3.089521046087113e-46 ], 
                    [ 230.5 , 3.0903855539235036e-46 ], [ 230.5625 , 3.0912503036659775e-46 ], [ 230.625 , 3.0921152953821362e-46 ], [ 230.6875 , 3.0929805291397773e-46 ], 
                    [ 230.75 , 3.0938460050065624e-46 ], [ 230.8125 , 3.09471172305026e-46 ], [ 230.875 , 3.0955776833386586e-46 ], [ 230.9375 , 3.0964433385596613e-46 ], 
                    [ 231.0 , 3.097308191722465e-46 ], [ 231.0625 , 3.0981732864433537e-46 ], [ 231.125 , 3.0990386227897745e-46 ], [ 231.1875 , 3.099904200829325e-46 ], 
                    [ 231.25 , 3.1007700206294007e-46 ], [ 231.3125 , 3.101636082257593e-46 ], [ 231.375 , 3.102502385781424e-46 ], [ 231.4375 , 3.103368931268456e-46 ], 
                    [ 231.5 , 3.104234705215226e-46 ], [ 231.5625 , 3.105100492080463e-46 ], [ 231.625 , 3.105966520418043e-46 ], [ 231.6875 , 3.1068327902953577e-46 ], 
                    [ 231.75 , 3.10769930177973e-46 ], [ 231.8125 , 3.1085660549385895e-46 ], [ 231.875 , 3.1094330498393403e-46 ], [ 231.9375 , 3.110300286549295e-46 ], 
                    [ 232.0 , 3.1111681516432282e-46 ], [ 232.0625 , 3.1120372146953897e-46 ], [ 232.125 , 3.1129065205086914e-46 ], [ 232.1875 , 3.1137760691508343e-46 ], 
                    [ 232.25 , 3.1146458606897597e-46 ], [ 232.3125 , 3.1155158951932735e-46 ], [ 232.375 , 3.1163861727292444e-46 ], [ 232.4375 , 3.1172566933655612e-46 ], 
                    [ 232.5 , 3.1181274571701073e-46 ], [ 232.5625 , 3.1189972905217772e-46 ], [ 232.625 , 3.1198669278789544e-46 ], [ 232.6875 , 3.1207368077080105e-46 ], 
                    [ 232.75 , 3.1216069300765733e-46 ], [ 232.8125 , 3.1224772950522015e-46 ], [ 232.875 , 3.123347902702538e-46 ], [ 232.9375 , 3.124218753095379e-46 ], 
                    [ 233.0 , 3.125089846298294e-46 ], [ 233.0625 , 3.1259612115225965e-46 ], [ 233.125 , 3.1268330049077704e-46 ], [ 233.1875 , 3.1277050414257464e-46 ], 
                    [ 233.25 , 3.12857732114433e-46 ], [ 233.3125 , 3.1294498441313704e-46 ], [ 233.375 , 3.130322610454667e-46 ], [ 233.4375 , 3.1311956201820394e-46 ], 
                    [ 233.5 , 3.1320688733815472e-46 ], [ 233.5625 , 3.132942370120937e-46 ], [ 233.625 , 3.1338160247363333e-46 ], [ 233.6875 , 3.134689851096195e-46 ], 
                    [ 233.75 , 3.1355639211118747e-46 ], [ 233.8125 , 3.136438234851335e-46 ], [ 233.875 , 3.1373127923825355e-46 ], [ 233.9375 , 3.1381875937733656e-46 ], 
                    [ 234.0 , 3.1390626390919335e-46 ], [ 234.0625 , 3.1399379284061894e-46 ], [ 234.125 , 3.14081346178419e-46 ], [ 234.1875 , 3.1416892265034884e-46 ], 
                    [ 234.25 , 3.1425652343030815e-46 ], [ 234.3125 , 3.1434414863628163e-46 ], [ 234.375 , 3.1443179827509122e-46 ], [ 234.4375 , 3.1451947235354062e-46 ], 
                    [ 234.5 , 3.1460717087845115e-46 ], [ 234.5625 , 3.146948938566438e-46 ], [ 234.625 , 3.1478264129492127e-46 ], [ 234.6875 , 3.1487045351171784e-46 ], 
                    [ 234.75 , 3.1495839391208132e-46 ], [ 234.8125 , 3.1504635887338385e-46 ], [ 234.875 , 3.1513434840248508e-46 ], [ 234.9375 , 3.152223625062488e-46 ], 
                    [ 235.0 , 3.1531040119153626e-46 ], [ 235.0625 , 3.1539846446520832e-46 ], [ 235.125 , 3.1548655233414357e-46 ], [ 235.1875 , 3.155746648051977e-46 ], 
                    [ 235.25 , 3.15662712307715e-46 ], [ 235.3125 , 3.1575073038483077e-46 ], [ 235.375 , 3.1583877300453643e-46 ], [ 235.4375 , 3.1592684017368875e-46 ], 
                    [ 235.5 , 3.1601493189911956e-46 ], [ 235.5625 , 3.1610304818768048e-46 ], [ 235.625 , 3.161911890462184e-46 ], [ 235.6875 , 3.162793544815865e-46 ], 
                    [ 235.75 , 3.1636754450064e-46 ], [ 235.8125 , 3.164557400175851e-46 ], [ 235.875 , 3.165439591583596e-46 ], [ 235.9375 , 3.1663220289219523e-46 ], 
                    [ 236.0 , 3.1672047122595695e-46 ], [ 236.0625 , 3.1680876416649125e-46 ], [ 236.125 , 3.1689708172066903e-46 ], [ 236.1875 , 3.1698542389534744e-46 ], 
                    [ 236.25 , 3.1707379069739004e-46 ], [ 236.3125 , 3.171621102016263e-46 ], [ 236.375 , 3.1725023612973183e-46 ], [ 236.4375 , 3.1733838654429877e-46 ], 
                    [ 236.5 , 3.174265614521377e-46 ], [ 236.5625 , 3.175147608600497e-46 ], [ 236.625 , 3.1760298477484673e-46 ], [ 236.6875 , 3.176912332033361e-46 ], 
                    [ 236.75 , 3.1777950615232903e-46 ], [ 236.8125 , 3.1786780362863885e-46 ], [ 236.875 , 3.1795627739464996e-46 ], [ 236.9375 , 3.1804491148412263e-46 ], 
                    [ 237.0 , 3.1813357028139607e-46 ], [ 237.0625 , 3.1822225379336477e-46 ], [ 237.125 , 3.1831096202691822e-46 ], [ 237.1875 , 3.183996949889433e-46 ], 
                    [ 237.25 , 3.1848845268633337e-46 ], [ 237.3125 , 3.185772351259883e-46 ], [ 237.375 , 3.186660423148029e-46 ], [ 237.4375 , 3.187548688186914e-46 ], 
                    [ 237.5 , 3.1884371865316207e-46 ], [ 237.5625 , 3.189325932536664e-46 ], [ 237.625 , 3.190214926271009e-46 ], [ 237.6875 , 3.1911041678037978e-46 ], 
                    [ 237.75 , 3.191993657204057e-46 ], [ 237.8125 , 3.192883394540878e-46 ], [ 237.875 , 3.193773379883301e-46 ], [ 237.9375 , 3.194663545258628e-46 ], 
                    [ 238.0 , 3.195552325278585e-46 ], [ 238.0625 , 3.1964413525639937e-46 ], [ 238.125 , 3.197330627183645e-46 ], [ 238.1875 , 3.1982201492063492e-46 ], 
                    [ 238.25 , 3.1991099187009585e-46 ], [ 238.3125 , 3.1999999357362988e-46 ], [ 238.375 , 3.2008902003812152e-46 ], [ 238.4375 , 3.2017807127045944e-46 ], 
                    [ 238.5 , 3.202671472866777e-46 ], [ 238.5625 , 3.203562481110991e-46 ], [ 238.625 , 3.204453737240665e-46 ], [ 238.6875 , 3.2053452413248084e-46 ], 
                    [ 238.75 , 3.2062369934322898e-46 ], [ 238.8125 , 3.2071289936321795e-46 ], [ 238.875 , 3.208021241993522e-46 ], [ 238.9375 , 3.2089137385853125e-46 ], 
                    [ 239.0 , 3.209806483476678e-46 ], [ 239.0625 , 3.2106994348598654e-46 ], [ 239.125 , 3.211592584685269e-46 ], [ 239.1875 , 3.212485982966383e-46 ], 
                    [ 239.25 , 3.213379629772323e-46 ], [ 239.3125 , 3.214273525172224e-46 ], [ 239.375 , 3.2151676692352617e-46 ], [ 239.4375 , 3.216062062030541e-46 ], 
                    [ 239.5 , 3.2169567036273456e-46 ], [ 239.5625 , 3.2178515940948177e-46 ], [ 239.625 , 3.2187465635459475e-46 ], [ 239.6875 , 3.21964168628469e-46 ], 
                    [ 239.75 , 3.2205370579540735e-46 ], [ 239.8125 , 3.221432678623393e-46 ], [ 239.875 , 3.222328548361803e-46 ], [ 239.9375 , 3.22322466723866e-46 ], 
                    [ 240.0 , 3.2241210353231575e-46 ], [ 240.0625 , 3.2250176526846676e-46 ], [ 240.125 , 3.225914519392445e-46 ], [ 240.1875 , 3.2268117510342217e-46 ], 
                    [ 240.25 , 3.2277092625717823e-46 ], [ 240.3125 , 3.2286070237449034e-46 ], [ 240.375 , 3.229505034623019e-46 ], [ 240.4375 , 3.2304032952756287e-46 ], 
                    [ 240.5 , 3.231301805772113e-46 ], [ 240.5625 , 3.232200566181965e-46 ], [ 240.625 , 3.2330995765748094e-46 ], [ 240.6875 , 3.233998837020109e-46 ], 
                    [ 240.75 , 3.2348980804491694e-46 ], [ 240.8125 , 3.2357975580696917e-46 ], [ 240.875 , 3.2366972857938946e-46 ], [ 240.9375 , 3.2375972636913666e-46 ], 
                    [ 241.0 , 3.2384974918315998e-46 ], [ 241.0625 , 3.2393979702842676e-46 ], [ 241.125 , 3.2402986991188787e-46 ], [ 241.1875 , 3.2411996784051216e-46 ], 
                    [ 241.25 , 3.242100796346941e-46 ], [ 241.3125 , 3.243000528327172e-46 ], [ 241.375 , 3.2439005099966046e-46 ], [ 241.4375 , 3.244800741424602e-46 ], 
                    [ 241.5 , 3.2457012226804285e-46 ], [ 241.5625 , 3.2466019538333698e-46 ], [ 241.625 , 3.2475029349528446e-46 ], [ 241.6875 , 3.248404166108246e-46 ], 
                    [ 241.75 , 3.2493056473688705e-46 ], [ 241.8125 , 3.2502078252104167e-46 ], [ 241.875 , 3.2511127447135255e-46 ], [ 241.9375 , 3.2520179161634095e-46 ], 
                    [ 242.0 , 3.252923339630147e-46 ], [ 242.0625 , 3.253829015183973e-46 ], [ 242.125 , 3.254734942895051e-46 ], [ 242.1875 , 3.2556411228335855e-46 ], 
                    [ 242.25 , 3.256547555069825e-46 ], [ 242.3125 , 3.2574542396739677e-46 ], [ 242.375 , 3.258360427322287e-46 ], [ 242.4375 , 3.2592644933219717e-46 ], 
                    [ 242.5 , 3.260168810164196e-46 ], [ 242.5625 , 3.2610733779185375e-46 ], [ 242.625 , 3.261978196654543e-46 ], [ 242.6875 , 3.2628832664420135e-46 ], 
                    [ 242.75 , 3.263788587350466e-46 ], [ 242.8125 , 3.2646941594496e-46 ], [ 242.875 , 3.265599982809181e-46 ], [ 242.9375 , 3.2665064423568353e-46 ], 
                    [ 243.0 , 3.267414069110132e-46 ], [ 243.0625 , 3.268321948055331e-46 ], [ 243.125 , 3.2692300792624375e-46 ], [ 243.1875 , 3.2701384628015198e-46 ], 
                    [ 243.25 , 3.2710470987427845e-46 ], [ 243.3125 , 3.271955987156294e-46 ], [ 243.375 , 3.2728651281121995e-46 ], [ 243.4375 , 3.273774521680742e-46 ], 
                    [ 243.5 , 3.2746826316334203e-46 ], [ 243.5625 , 3.275587956627326e-46 ], [ 243.625 , 3.276493531909056e-46 ], [ 243.6875 , 3.2773993575478526e-46 ], 
                    [ 243.75 , 3.2783054336129294e-46 ], [ 243.8125 , 3.2792117601735668e-46 ], [ 243.875 , 3.2801183372989698e-46 ], [ 243.9375 , 3.2810251650583414e-46 ], 
                    [ 244.0 , 3.2819322435210877e-46 ], [ 244.0625 , 3.2828405154374553e-46 ], [ 244.125 , 3.2837508368862375e-46 ], [ 244.1875 , 3.284661410764345e-46 ], 
                    [ 244.25 , 3.285572237141729e-46 ], [ 244.3125 , 3.2864833160884536e-46 ], [ 244.375 , 3.2873946476744854e-46 ], [ 244.4375 , 3.288306231969974e-46 ], 
                    [ 244.5 , 3.2892180690449477e-46 ], [ 244.5625 , 3.2901301589694554e-46 ], [ 244.625 , 3.291042441527952e-46 ], [ 244.6875 , 3.2919548578736105e-46 ], 
                    [ 244.75 , 3.292867527179745e-46 ], [ 244.8125 , 3.2937804495164876e-46 ], [ 244.875 , 3.2946936249539197e-46 ], [ 244.9375 , 3.29560705356228e-46 ], 
                    [ 245.0 , 3.2965207354117596e-46 ], [ 245.0625 , 3.2974346705725204e-46 ], [ 245.125 , 3.2983488591147904e-46 ], [ 245.1875 , 3.299263694338957e-46 ], 
                    [ 245.25 , 3.3001796506665454e-46 ], [ 245.3125 , 3.3010958612860455e-46 ], [ 245.375 , 3.3020123262680544e-46 ], [ 245.4375 , 3.302929045683143e-46 ], 
                    [ 245.5 , 3.303846019601995e-46 ], [ 245.5625 , 3.30476324809522e-46 ], [ 245.625 , 3.305680731233517e-46 ], [ 245.6875 , 3.306598469087582e-46 ], 
                    [ 245.75 , 3.30751521240237e-46 ], [ 245.8125 , 3.3084288651727876e-46 ], [ 245.875 , 3.3093427703263924e-46 ], [ 245.9375 , 3.3102569279329715e-46 ], 
                    [ 246.0 , 3.3111713380622374e-46 ], [ 246.0625 , 3.312086000783947e-46 ], [ 246.125 , 3.3130009161678743e-46 ], [ 246.1875 , 3.3139160842837665e-46 ], 
                    [ 246.25 , 3.314831505201556e-46 ], [ 246.3125 , 3.3157477186368113e-46 ], [ 246.375 , 3.316666344492825e-46 ], [ 246.4375 , 3.317585224853518e-46 ], 
                    [ 246.5 , 3.318504359789447e-46 ], [ 246.5625 , 3.319423749371189e-46 ], [ 246.625 , 3.3203433936692223e-46 ], [ 246.6875 , 3.3212632927541166e-46 ], 
                    [ 246.75 , 3.322183446696484e-46 ], [ 246.8125 , 3.3231038555669556e-46 ], [ 246.875 , 3.3240244952050435e-46 ], [ 246.9375 , 3.3249451977279213e-46 ], 
                    [ 247.0 , 3.3258661552709215e-46 ], [ 247.0625 , 3.326787367904682e-46 ], [ 247.125 , 3.3277088356998816e-46 ], [ 247.1875 , 3.3286305587271724e-46 ], 
                    [ 247.25 , 3.3295525370572027e-46 ], [ 247.3125 , 3.330474770760758e-46 ], [ 247.375 , 3.33139725990855e-46 ], [ 247.4375 , 3.332319992253603e-46 ], 
                    [ 247.5 , 3.3332414523559183e-46 ], [ 247.5625 , 3.3341631672622877e-46 ], [ 247.625 , 3.3350851370431925e-46 ], [ 247.6875 , 3.3360073617691124e-46 ], 
                    [ 247.75 , 3.3369298415105446e-46 ], [ 247.8125 , 3.337852576338006e-46 ], [ 247.875 , 3.3387755663220347e-46 ], [ 247.9375 , 3.3396988115331863e-46 ], 
                    [ 248.0 , 3.340622312042037e-46 ], [ 248.0625 , 3.341548161640358e-46 ], [ 248.125 , 3.3424745319830317e-46 ], [ 248.1875 , 3.3434011591414246e-46 ], 
                    [ 248.25 , 3.344328043186708e-46 ], [ 248.3125 , 3.3452551841900747e-46 ], [ 248.375 , 3.3461825822228324e-46 ], [ 248.4375 , 3.3471102373561647e-46 ], 
                    [ 248.5 , 3.348038149661348e-46 ], [ 248.5625 , 3.3489663192097006e-46 ], [ 248.625 , 3.3498893274459684e-46 ], [ 248.6875 , 3.3508106459010287e-46 ], 
                    [ 248.75 , 3.351732217745885e-46 ], [ 248.8125 , 3.352654043050204e-46 ], [ 248.875 , 3.353576121883694e-46 ], [ 248.9375 , 3.35449845431606e-46 ], 
                    [ 249.0 , 3.35542104041712e-46 ], [ 249.0625 , 3.3563438802565688e-46 ], [ 249.125 , 3.3572669739042164e-46 ], [ 249.1875 , 3.358194194785519e-46 ], 
                    [ 249.25 , 3.3591248162142946e-46 ], [ 249.3125 , 3.3600556955363886e-46 ], [ 249.375 , 3.360986832823482e-46 ], [ 249.4375 , 3.36191822814692e-46 ], 
                    [ 249.5 , 3.362849881578232e-46 ], [ 249.5625 , 3.3637817931890896e-46 ], [ 249.625 , 3.364713963050823e-46 ], [ 249.6875 , 3.3656463912351187e-46 ], 
                    [ 249.75 , 3.366577375129206e-46 ], [ 249.8125 , 3.3675056915640376e-46 ], [ 249.875 , 3.368434263977458e-46 ], [ 249.9375 , 3.3693630924399322e-46 ], 
                    [ 250.0 , 3.3702921770220647e-46 ], [ 250.0625 , 3.371221517794551e-46 ], [ 250.125 , 3.3721511148279377e-46 ], [ 250.1875 , 3.3730809681929124e-46 ], 
                    [ 250.25 , 3.3740110779602046e-46 ], [ 250.3125 , 3.3749416240642017e-46 ], [ 250.375 , 3.375873430538129e-46 ], [ 250.4375 , 3.376805494279708e-46 ], 
                    [ 250.5 , 3.3777378153599446e-46 ], [ 250.5625 , 3.378670393849961e-46 ], [ 250.625 , 3.3796032298207307e-46 ], [ 250.6875 , 3.380536323343416e-46 ], 
                    [ 250.75 , 3.3814696744890762e-46 ], [ 250.8125 , 3.382403283328889e-46 ], [ 250.875 , 3.3833371499339782e-46 ], [ 250.9375 , 3.3842708635303206e-46 ], 
                    [ 251.0 , 3.385204799071957e-46 ], [ 251.0625 , 3.3861389923457756e-46 ], [ 251.125 , 3.387073443422925e-46 ], [ 251.1875 , 3.388008152374574e-46 ], 
                    [ 251.25 , 3.3889431192718613e-46 ], [ 251.3125 , 3.3898783441859723e-46 ], [ 251.375 , 3.390813827188133e-46 ], [ 251.4375 , 3.391749568349494e-46 ], 
                    [ 251.5 , 3.392685380336919e-46 ], [ 251.5625 , 3.3936213522225814e-46 ], [ 251.625 , 3.3945575823235722e-46 ], [ 251.6875 , 3.3954940707111038e-46 ], 
                    [ 251.75 , 3.396430817456456e-46 ], [ 251.8125 , 3.3973678226309536e-46 ], [ 251.875 , 3.3983050863057945e-46 ], [ 251.9375 , 3.3992426085523178e-46 ], 
                    [ 252.0 , 3.4001803894419553e-46 ], [ 252.0625 , 3.401118376957875e-46 ], [ 252.125 , 3.4020565367628735e-46 ], [ 252.1875 , 3.402994955348628e-46 ], 
                    [ 252.25 , 3.403933632786521e-46 ], [ 252.3125 , 3.4048725691478066e-46 ], [ 252.375 , 3.405811764504053e-46 ], [ 252.4375 , 3.4067512189266514e-46 ], 
                    [ 252.5 , 3.4076909324870632e-46 ], [ 252.5625 , 3.408630905256744e-46 ], [ 252.625 , 3.409571113293563e-46 ], [ 252.6875 , 3.4105113044356516e-46 ], 
                    [ 252.75 , 3.411451754836046e-46 ], [ 252.8125 , 3.412392464566092e-46 ], [ 252.875 , 3.413333433697421e-46 ], [ 252.9375 , 3.4142746623014903e-46 ], 
                    [ 253.0 , 3.415216150449922e-46 ], [ 253.0625 , 3.4161578982141884e-46 ], [ 253.125 , 3.4170999056659286e-46 ], [ 253.1875 , 3.4180421728767742e-46 ], 
                    [ 253.25 , 3.418982242328248e-46 ], [ 253.3125 , 3.4199218277276278e-46 ], [ 253.375 , 3.4208616713385385e-46 ], [ 253.4375 , 3.421801773232038e-46 ], 
                    [ 253.5 , 3.4227421334789845e-46 ], [ 253.5625 , 3.4236827521504258e-46 ], [ 253.625 , 3.4246236293174043e-46 ], [ 253.6875 , 3.425564765050886e-46 ], 
                    [ 253.75 , 3.4265061594220746e-46 ], [ 253.8125 , 3.4274476102721318e-46 ], [ 253.875 , 3.4283890449883078e-46 ], [ 253.9375 , 3.4293307382931893e-46 ], 
                    [ 254.0 , 3.4302726902578285e-46 ], [ 254.0625 , 3.431214900953298e-46 ], [ 254.125 , 3.4321573704505897e-46 ], [ 254.1875 , 3.4331000988208657e-46 ], 
                    [ 254.25 , 3.4340430861352304e-46 ], [ 254.3125 , 3.4349863324647856e-46 ], [ 254.375 , 3.435929899889623e-46 ], [ 254.4375 , 3.4368749566891168e-46 ], 
                    [ 254.5 , 3.4378202734276114e-46 ], [ 254.5625 , 3.4387658501767263e-46 ], [ 254.625 , 3.439711687007806e-46 ], [ 254.6875 , 3.4406577839925583e-46 ], 
                    [ 254.75 , 3.441604141202489e-46 ], [ 254.8125 , 3.4425507587090756e-46 ], [ 254.875 , 3.4434976365840357e-46 ], [ 254.9375 , 3.444444774898862e-46 ], 
                    [ 255.0 , 3.445392799838524e-46 ], [ 255.0625 , 3.4463413913929632e-46 ], [ 255.125 , 3.447290244115308e-46 ], [ 255.1875 , 3.448239358077415e-46 ], 
                    [ 255.25 , 3.4491887333511845e-46 ], [ 255.3125 , 3.4501383700086827e-46 ], [ 255.375 , 3.4510882681217534e-46 ], [ 255.4375 , 3.4520384277625026e-46 ], 
                    [ 255.5 , 3.4529888490027875e-46 ], [ 255.5625 , 3.453939440368288e-46 ], [ 255.625 , 3.4548900381364125e-46 ], [ 255.6875 , 3.4558408975293326e-46 ], 
                    [ 255.75 , 3.456792018619029e-46 ], [ 255.8125 , 3.457743401477526e-46 ], [ 255.875 , 3.458695046176844e-46 ], [ 255.9375 , 3.4596469527890955e-46 ], 
                    [ 256.0 , 3.460599121386291e-46 ], [ 256.0625 , 3.461551552040707e-46 ], [ 256.125 , 3.4625042448242205e-46 ], [ 256.1875 , 3.463454489129014e-46 ], 
                    [ 256.25 , 3.4644044777559085e-46 ], [ 256.3125 , 3.4653547269545504e-46 ], [ 256.375 , 3.4663052367964377e-46 ], [ 256.4375 , 3.4672560073530373e-46 ], 
                    [ 256.5 , 3.4682070386958595e-46 ], [ 256.5625 , 3.469158330896387e-46 ], [ 256.625 , 3.470109884026243e-46 ], [ 256.6875 , 3.4710616981569744e-46 ], 
                    [ 256.75 , 3.4720141887483433e-46 ], [ 256.8125 , 3.4729676072464805e-46 ], [ 256.875 , 3.4739212875543213e-46 ], [ 256.9375 , 3.4748752297437337e-46 ], 
                    [ 257.0 , 3.4758294338865564e-46 ], [ 257.0625 , 3.4767839000548455e-46 ], [ 257.125 , 3.477738628320406e-46 ], [ 257.1875 , 3.4786936187553085e-46 ], 
                    [ 257.25 , 3.47964887143152e-46 ], [ 257.3125 , 3.480604386421052e-46 ], [ 257.375 , 3.481561137598903e-46 ], [ 257.4375 , 3.4825182457584802e-46 ], 
                    [ 257.5 , 3.4834756170345143e-46 ], [ 257.5625 , 3.4844332514993625e-46 ], [ 257.625 , 3.485391149225327e-46 ], [ 257.6875 , 3.4863493102848557e-46 ], 
                    [ 257.75 , 3.4873077347502418e-46 ], [ 257.8125 , 3.4882664226939712e-46 ], [ 257.875 , 3.4892253741884526e-46 ], [ 257.9375 , 3.4901826073561235e-46 ], 
                    [ 258.0 , 3.491137409933703e-46 ], [ 258.0625 , 3.4920924737148213e-46 ], [ 258.125 , 3.493047798770886e-46 ], [ 258.1875 , 3.494003385173399e-46 ], 
                    [ 258.25 , 3.4949592329938547e-46 ], [ 258.3125 , 3.495915342303745e-46 ], [ 258.375 , 3.4968717131746554e-46 ], [ 258.4375 , 3.497828345678139e-46 ], 
                    [ 258.5 , 3.4987852398857217e-46 ], [ 258.5625 , 3.499744649365395e-46 ], [ 258.625 , 3.5007045613018947e-46 ], [ 258.6875 , 3.501664736523557e-46 ], 
                    [ 258.75 , 3.50262517510267e-46 ], [ 258.8125 , 3.503585877111368e-46 ], [ 258.875 , 3.504546842621981e-46 ], [ 258.9375 , 3.50550807170673e-46 ], 
                    [ 259.0 , 3.506469564437961e-46 ], [ 259.0625 , 3.50743132088796e-46 ], [ 259.125 , 3.5083926048292e-46 ], [ 259.1875 , 3.509352887369064e-46 ], 
                    [ 259.25 , 3.510313432747921e-46 ], [ 259.3125 , 3.511274241037812e-46 ], [ 259.375 , 3.512235312310627e-46 ], [ 259.4375 , 3.5131966466383924e-46 ], 
                    [ 259.5 , 3.514158244093013e-46 ], [ 259.5625 , 3.515120104746657e-46 ], [ 259.625 , 3.516082228671315e-46 ], [ 259.6875 , 3.5170446159389474e-46 ], 
                    [ 259.75 , 3.518005956961483e-46 ], [ 259.8125 , 3.5189672332503555e-46 ], [ 259.875 , 3.5199287722027774e-46 ], [ 259.9375 , 3.520890573890594e-46 ], 
                    [ 260.0 , 3.521852638385546e-46 ], [ 260.0625 , 3.5228149657594944e-46 ], [ 260.125 , 3.523777556084244e-46 ], [ 260.1875 , 3.524740409431621e-46 ], 
                    [ 260.25 , 3.525703525873544e-46 ], [ 260.3125 , 3.5266677045582364e-46 ], [ 260.375 , 3.527635048064103e-46 ], [ 260.4375 , 3.528602656906412e-46 ], 
                    [ 260.5 , 3.529570531157992e-46 ], [ 260.5625 , 3.530538670891518e-46 ], [ 260.625 , 3.531507076179937e-46 ], [ 260.6875 , 3.532475747096036e-46 ], 
                    [ 260.75 , 3.5334446837126276e-46 ], [ 260.8125 , 3.534413886102692e-46 ], [ 260.875 , 3.535383354339053e-46 ], [ 260.9375 , 3.536352065849852e-46 ], 
                    [ 261.0 , 3.537320405124691e-46 ], [ 261.0625 , 3.538289009554419e-46 ], [ 261.125 , 3.539257879211592e-46 ], [ 261.1875 , 3.540227014168861e-46 ], 
                    [ 261.25 , 3.5411964144988706e-46 ], [ 261.3125 , 3.542166080274287e-46 ], [ 261.375 , 3.543136011567847e-46 ], [ 261.4375 , 3.544106208452153e-46 ], 
                    [ 261.5 , 3.545076671000007e-46 ], [ 261.5625 , 3.546045362933173e-46 ], [ 261.625 , 3.547014319561313e-46 ], [ 261.6875 , 3.5479835409568024e-46 ], 
                    [ 261.75 , 3.5489530271919405e-46 ], [ 261.8125 , 3.549922778339094e-46 ], [ 261.875 , 3.550892794470701e-46 ], [ 261.9375 , 3.551863075659142e-46 ], 
                    [ 262.0 , 3.552833621976896e-46 ], [ 262.0625 , 3.553804433496332e-46 ], [ 262.125 , 3.5547748647351384e-46 ], [ 262.1875 , 3.555744372236108e-46 ], 
                    [ 262.25 , 3.5567141441546085e-46 ], [ 262.3125 , 3.557684180562706e-46 ], [ 262.375 , 3.558654481532636e-46 ], [ 262.4375 , 3.559625047136503e-46 ], 
                    [ 262.5 , 3.560595877446508e-46 ], [ 262.5625 , 3.561566972534819e-46 ], [ 262.625 , 3.562538332473599e-46 ], [ 262.6875 , 3.563509957335159e-46 ], 
                    [ 262.75 , 3.5644815554250556e-46 ], [ 262.8125 , 3.565453275974345e-46 ], [ 262.875 , 3.566425261426345e-46 ], [ 262.9375 , 3.567397511853297e-46 ], 
                    [ 263.0 , 3.568370027327385e-46 ], [ 263.0625 , 3.569342807920965e-46 ], [ 263.125 , 3.5703158537061365e-46 ], [ 263.1875 , 3.571289164755394e-46 ], 
                    [ 263.25 , 3.572262741140877e-46 ], [ 263.3125 , 3.573236582934992e-46 ], [ 263.375 , 3.5742116511095505e-46 ], [ 263.4375 , 3.5751870089968245e-46 ], 
                    [ 263.5 , 3.576162633047106e-46 ], [ 263.5625 , 3.577138523333026e-46 ], [ 263.625 , 3.578114679927287e-46 ], [ 263.6875 , 3.579091102902537e-46 ], 
                    [ 263.75 , 3.580067792331442e-46 ], [ 263.8125 , 3.581044748286766e-46 ], [ 263.875 , 3.582021970841239e-46 ], [ 263.9375 , 3.582999292087221e-46 ], 
                    [ 264.0 , 3.5839764115497535e-46 ], [ 264.0625 , 3.584953797482434e-46 ], [ 264.125 , 3.585931449957982e-46 ], [ 264.1875 , 3.586909369049064e-46 ], 
                    [ 264.25 , 3.58788755482836e-46 ], [ 264.3125 , 3.5888660073686496e-46 ], [ 264.375 , 3.5898447267426316e-46 ], [ 264.4375 , 3.59082371302315e-46 ], 
                    [ 264.5 , 3.591802966282865e-46 ], [ 264.5625 , 3.592782984766212e-46 ], [ 264.625 , 3.5937637306223006e-46 ], [ 264.6875 , 3.594744744199106e-46 ], 
                    [ 264.75 , 3.5957260255697345e-46 ], [ 264.8125 , 3.596707574807238e-46 ], [ 264.875 , 3.597689391984812e-46 ], [ 264.9375 , 3.598671477175472e-46 ], 
                    [ 265.0 , 3.599653830452506e-46 ], [ 265.0625 , 3.600636451888993e-46 ], [ 265.125 , 3.6016193415582125e-46 ], [ 265.1875 , 3.6026001298285526e-46 ], 
                    [ 265.25 , 3.6035804366546e-46 ], [ 265.3125 , 3.604561010232786e-46 ], [ 265.375 , 3.6055418506357484e-46 ], [ 265.4375 , 3.606522957936117e-46 ], 
                    [ 265.5 , 3.607504332206519e-46 ], [ 265.5625 , 3.608485973519472e-46 ], [ 265.625 , 3.609467881947896e-46 ], [ 265.6875 , 3.610450057564196e-46 ], 
                    [ 265.75 , 3.611432500441255e-46 ], [ 265.8125 , 3.612415967759855e-46 ], [ 265.875 , 3.613399715210743e-46 ], [ 265.9375 , 3.6143837305597684e-46 ], 
                    [ 266.0 , 3.615368013879707e-46 ], [ 266.0625 , 3.616352565243612e-46 ], [ 266.125 , 3.6173373847243725e-46 ], [ 266.1875 , 3.61832247239516e-46 ], 
                    [ 266.25 , 3.619307828328956e-46 ], [ 266.3125 , 3.620293452598714e-46 ], [ 266.375 , 3.62127879629266e-46 ], [ 266.4375 , 3.622261837223841e-46 ], 
                    [ 266.5 , 3.623245145013696e-46 ], [ 266.5625 , 3.624228719734565e-46 ], [ 266.625 , 3.625212561458986e-46 ], [ 266.6875 , 3.626196670259364e-46 ], 
                    [ 266.75 , 3.627181046208354e-46 ], [ 266.8125 , 3.628165689378296e-46 ], [ 266.875 , 3.629150599841834e-46 ], [ 266.9375 , 3.630135777671504e-46 ], 
                    [ 267.0 , 3.631121911606427e-46 ], [ 267.0625 , 3.632109581544024e-46 ], [ 267.125 , 3.633097520129144e-46 ], [ 267.1875 , 3.634085727434832e-46 ], 
                    [ 267.25 , 3.635074203534234e-46 ], [ 267.3125 , 3.63606294850041e-46 ], [ 267.375 , 3.637051962406517e-46 ], [ 267.4375 , 3.638041245325734e-46 ], 
                    [ 267.5 , 3.639030797331233e-46 ], [ 267.5625 , 3.640020618496128e-46 ], [ 267.625 , 3.641009861976884e-46 ], [ 267.6875 , 3.641998513492548e-46 ], 
                    [ 267.75 , 3.642987433458985e-46 ], [ 267.8125 , 3.643976621949038e-46 ], [ 267.875 , 3.644966079035643e-46 ], [ 267.9375 , 3.645955804791761e-46 ], 
                    [ 268.0 , 3.646945799290265e-46 ], [ 268.0625 , 3.647936062604258e-46 ], [ 268.125 , 3.648926594806627e-46 ], [ 268.1875 , 3.649917395970412e-46 ], 
                    [ 268.25 , 3.650909034981349e-46 ], [ 268.3125 , 3.651901286248177e-46 ], [ 268.375 , 3.652893807190936e-46 ], [ 268.4375 , 3.653886597883049e-46 ], 
                    [ 268.5 , 3.654879658397777e-46 ], [ 268.5625 , 3.655872988808478e-46 ], [ 268.625 , 3.656866589188427e-46 ], [ 268.6875 , 3.6578604596110495e-46 ], 
                    [ 268.75 , 3.658854600149739e-46 ], [ 268.8125 , 3.659849010877958e-46 ], [ 268.875 , 3.660843170867954e-46 ], [ 268.9375 , 3.6618374139795824e-46 ], 
                    [ 269.0 , 3.662831927116315e-46 ], [ 269.0625 , 3.663826710351383e-46 ], [ 269.125 , 3.664821763758271e-46 ], [ 269.1875 , 3.665817087410225e-46 ], 
                    [ 269.25 , 3.666812681380693e-46 ], [ 269.3125 , 3.6678085457431166e-46 ], [ 269.375 , 3.668804680570852e-46 ], [ 269.4375 , 3.66980108593746e-46 ], 
                    [ 269.5 , 3.670795050979804e-46 ], [ 269.5625 , 3.671788673782462e-46 ], [ 269.625 , 3.67278256554224e-46 ], [ 269.6875 , 3.67377672633189e-46 ], 
                    [ 269.75 , 3.674771156224204e-46 ], [ 269.8125 , 3.6757658552920794e-46 ], [ 269.875 , 3.676760823608323e-46 ], [ 269.9375 , 3.677756061245868e-46 ], 
                    [ 270.0 , 3.6787515682775635e-46 ], [ 270.0625 , 3.679747344776358e-46 ], [ 270.125 , 3.680744313697942e-46 ], [ 270.1875 , 3.681741698040391e-46 ], 
                    [ 270.25 , 3.682739352647587e-46 ], [ 270.3125 , 3.68373727759279e-46 ], [ 270.375 , 3.68473547294923e-46 ], [ 270.4375 , 3.6857339387902574e-46 ], 
                    [ 270.5 , 3.68673267518909e-46 ], [ 270.5625 , 3.6877316822190406e-46 ], [ 270.625 , 3.688730959953496e-46 ], [ 270.6875 , 3.689730508465836e-46 ], 
                    [ 270.75 , 3.690729500910914e-46 ], [ 270.8125 , 3.691728684020643e-46 ], [ 270.875 , 3.692728137637052e-46 ], [ 270.9375 , 3.693727861833375e-46 ], 
                    [ 271.0 , 3.694727856682811e-46 ], [ 271.0625 , 3.6957281222586886e-46 ], [ 271.125 , 3.696728658634221e-46 ], [ 271.1875 , 3.697729465882799e-46 ], 
                    [ 271.25 , 3.698730544077784e-46 ], [ 271.3125 , 3.699731893292396e-46 ], [ 271.375 , 3.700734466947395e-46 ], [ 271.4375 , 3.701737404299871e-46 ], 
                    [ 271.5 , 3.70274061345878e-46 ], [ 271.5625 , 3.70374409449773e-46 ], [ 271.625 , 3.7047478474904825e-46 ], [ 271.6875 , 3.705751872510662e-46 ], 
                    [ 271.75 , 3.706756169631962e-46 ], [ 271.8125 , 3.707760738928285e-46 ], [ 271.875 , 3.708765580473208e-46 ], [ 271.9375 , 3.709770694340593e-46 ], 
                    [ 272.0 , 3.710775394003584e-46 ], [ 272.0625 , 3.711780299495754e-46 ], [ 272.125 , 3.712785477123781e-46 ], [ 272.1875 , 3.7137909269613644e-46 ], 
                    [ 272.25 , 3.714796649082193e-46 ], [ 272.3125 , 3.715802643560055e-46 ], [ 272.375 , 3.716808910468761e-46 ], [ 272.4375 , 3.717815449881901e-46 ], 
                    [ 272.5 , 3.718822261873378e-46 ], [ 272.5625 , 3.719829346517034e-46 ], [ 272.625 , 3.7208337569805174e-46 ], [ 272.6875 , 3.721838036690869e-46 ], 
                    [ 272.75 , 3.722842587463515e-46 ], [ 272.8125 , 3.723847409371671e-46 ], [ 272.875 , 3.724852502488411e-46 ], [ 272.9375 , 3.7258578668869904e-46 ], 
                    [ 273.0 , 3.726863502640656e-46 ], [ 273.0625 , 3.727869409822675e-46 ], [ 273.125 , 3.72887558850615e-46 ], [ 273.1875 , 3.729882038764519e-46 ], 
                    [ 273.25 , 3.730889447722913e-46 ], [ 273.3125 , 3.731897283739078e-46 ], [ 273.375 , 3.732905392004943e-46 ], [ 273.4375 , 3.733913772594e-46 ], 
                    [ 273.5 , 3.734922425579863e-46 ], [ 273.5625 , 3.735931351036091e-46 ], [ 273.625 , 3.736940549036261e-46 ], [ 273.6875 , 3.73795001965402e-46 ], 
                    [ 273.75 , 3.738959762963013e-46 ], [ 273.8125 , 3.739969779036929e-46 ], [ 273.875 , 3.7409806779107595e-46 ], [ 273.9375 , 3.7419920688761145e-46 ], 
                    [ 274.0 , 3.743003733275623e-46 ], [ 274.0625 , 3.744015671183155e-46 ], [ 274.125 , 3.7450278826726015e-46 ], [ 274.1875 , 3.746040367818008e-46 ], 
                    [ 274.25 , 3.747053126693358e-46 ], [ 274.3125 , 3.748066159372628e-46 ], [ 274.375 , 3.74907946592979e-46 ], [ 274.4375 , 3.750093046438967e-46 ], 
                    [ 274.5 , 3.751104748375298e-46 ], [ 274.5625 , 3.752115512086045e-46 ], [ 274.625 , 3.753126548154683e-46 ], [ 274.6875 , 3.754137856654788e-46 ], 
                    [ 274.75 , 3.755149437659608e-46 ], [ 274.8125 , 3.756161291242652e-46 ], [ 274.875 , 3.757173417477262e-46 ], [ 274.9375 , 3.758185816437092e-46 ], 
                    [ 275.0 , 3.75919848819547e-46 ], [ 275.0625 , 3.760211432825984e-46 ], [ 275.125 , 3.761223527115765e-46 ], [ 275.1875 , 3.7622348228996564e-46 ], 
                    [ 275.25 , 3.763246390594871e-46 ], [ 275.3125 , 3.764258230274437e-46 ], [ 275.375 , 3.765270342011538e-46 ], [ 275.4375 , 3.766282725879377e-46 ], 
                    [ 275.5 , 3.767295381951016e-46 ], [ 275.5625 , 3.768308310299723e-46 ], [ 275.625 , 3.769321510998654e-46 ], [ 275.6875 , 3.77033498412109e-46 ], 
                    [ 275.75 , 3.771350482680441e-46 ], [ 275.8125 , 3.772369482628166e-46 ], [ 275.875 , 3.773388757904555e-46 ], [ 275.9375 , 3.774408308583975e-46 ], 
                    [ 276.0 , 3.775428134740918e-46 ], [ 276.0625 , 3.776448236449684e-46 ], [ 276.125 , 3.7774686137848315e-46 ], [ 276.1875 , 3.778489266820806e-46 ], 
                    [ 276.25 , 3.779510195632157e-46 ], [ 276.3125 , 3.780531400293315e-46 ], [ 276.375 , 3.781551703169294e-46 ], [ 276.4375 , 3.782567323880484e-46 ], 
                    [ 276.5 , 3.7835832173593915e-46 ], [ 276.5625 , 3.784599383679328e-46 ], [ 276.625 , 3.7856158229136235e-46 ], [ 276.6875 , 3.786632535135417e-46 ], 
                    [ 276.75 , 3.7876495204182375e-46 ], [ 276.8125 , 3.788666778835289e-46 ], [ 276.875 , 3.7896843104599004e-46 ], [ 276.9375 , 3.790702115365582e-46 ], 
                    [ 277.0 , 3.791720193625596e-46 ], [ 277.0625 , 3.792742029644098e-46 ], [ 277.125 , 3.793764197709726e-46 ], [ 277.1875 , 3.7947866412561405e-46 ], 
                    [ 277.25 , 3.7958093603575304e-46 ], [ 277.3125 , 3.796832355088241e-46 ], [ 277.375 , 3.797855625522555e-46 ], [ 277.4375 , 3.798879171734724e-46 ], 
                    [ 277.5 , 3.799902993799125e-46 ], [ 277.5625 , 3.80092709179002e-46 ], [ 277.625 , 3.801951465781883e-46 ], [ 277.6875 , 3.8029722487479974e-46 ], 
                    [ 277.75 , 3.803992189207744e-46 ], [ 277.8125 , 3.805012403211072e-46 ], [ 277.875 , 3.806032890831315e-46 ], [ 277.9375 , 3.807053652141803e-46 ], 
                    [ 278.0 , 3.808074687216074e-46 ], [ 278.0625 , 3.809095996127388e-46 ], [ 278.125 , 3.810117578949349e-46 ], [ 278.1875 , 3.811139435755283e-46 ], 
                    [ 278.25 , 3.8121615666187266e-46 ], [ 278.3125 , 3.813185204423642e-46 ], [ 278.375 , 3.8142102553795505e-46 ], [ 278.4375 , 3.81523558188708e-46 ], 
                    [ 278.5 , 3.816261184020383e-46 ], [ 278.5625 , 3.817287061853444e-46 ], [ 278.625 , 3.818313215460431e-46 ], [ 278.6875 , 3.819339644915504e-46 ], 
                    [ 278.75 , 3.820366350292789e-46 ], [ 278.8125 , 3.82139333166643e-46 ], [ 278.875 , 3.8224205891106754e-46 ], [ 278.9375 , 3.823447480164795e-46 ], 
                    [ 279.0 , 3.824472855302357e-46 ], [ 279.0625 , 3.825498505425881e-46 ], [ 279.125 , 3.8265244306090566e-46 ], [ 279.1875 , 3.827550630925624e-46 ], 
                    [ 279.25 , 3.8285771064494496e-46 ], [ 279.3125 , 3.8296038572541754e-46 ], [ 279.375 , 3.830630883413818e-46 ], [ 279.4375 , 3.8316581850021665e-46 ], 
                    [ 279.5 , 3.832685762093005e-46 ], [ 279.5625 , 3.8337136147602724e-46 ], [ 279.625 , 3.834743996902305e-46 ], [ 279.6875 , 3.835774749913389e-46 ], 
                    [ 279.75 , 3.836805779983901e-46 ], [ 279.8125 , 3.8378370871881776e-46 ], [ 279.875 , 3.838868671600845e-46 ], [ 279.9375 , 3.8399005332963334e-46 ], 
                    [ 280.0 , 3.8409326723491755e-46 ], [ 280.0625 , 3.841965088834004e-46 ], [ 280.125 , 3.8429977828253375e-46 ], [ 280.1875 , 3.844030754397793e-46 ], 
                    [ 280.25 , 3.8450634360453824e-46 ], [ 280.3125 , 3.846096097404749e-46 ], [ 280.375 , 3.84712903610401e-46 ], [ 280.4375 , 3.848162252217622e-46 ], 
                    [ 280.5 , 3.849195745820142e-46 ], [ 280.5625 , 3.850229516986071e-46 ], [ 280.625 , 3.8512635657899244e-46 ], [ 280.6875 , 3.852297892306294e-46 ], 
                    [ 280.75 , 3.85333249660982e-46 ], [ 280.8125 , 3.854367378774969e-46 ], [ 280.875 , 3.855400995840642e-46 ], [ 280.9375 , 3.856431610260644e-46 ], 
                    [ 281.0 , 3.857462500181478e-46 ], [ 281.0625 , 3.8584936656767094e-46 ], [ 281.125 , 3.859525106820087e-46 ], [ 281.1875 , 3.860556823685267e-46 ], 
                    [ 281.25 , 3.861588816345957e-46 ], [ 281.3125 , 3.8626210848759075e-46 ], [ 281.375 , 3.8636536293488355e-46 ], [ 281.4375 , 3.8646864498384785e-46 ], 
                    [ 281.5 , 3.8657195464186476e-46 ], [ 281.5625 , 3.8667549575780405e-46 ], [ 281.625 , 3.8677907488712605e-46 ], [ 281.6875 , 3.868826817622898e-46 ], 
                    [ 281.75 , 3.869863163907192e-46 ], [ 281.8125 , 3.8708997877985975e-46 ], [ 281.875 , 3.8719366893713925e-46 ], [ 281.9375 , 3.872973868699935e-46 ], 
                    [ 282.0 , 3.874011325858735e-46 ], [ 282.0625 , 3.87504906092219e-46 ], [ 282.125 , 3.876087073964714e-46 ], [ 282.1875 , 3.877124063584132e-46 ], 
                    [ 282.25 , 3.878160340514788e-46 ], [ 282.3125 , 3.879196894421387e-46 ], [ 282.375 , 3.880233725377875e-46 ], [ 282.4375 , 3.8812708334583036e-46 ], 
                    [ 282.5 , 3.882308218736768e-46 ], [ 282.5625 , 3.883345881287386e-46 ], [ 282.625 , 3.884383821184212e-46 ], [ 282.6875 , 3.885422038501403e-46 ], 
                    [ 282.75 , 3.886460533313106e-46 ], [ 282.8125 , 3.8874990287401425e-46 ], [ 282.875 , 3.8885362561412266e-46 ], [ 282.9375 , 3.8895737602860465e-46 ], 
                    [ 283.0 , 3.890611541248274e-46 ], [ 283.0625 , 3.8916495991019065e-46 ], [ 283.125 , 3.8926879339207386e-46 ], [ 283.1875 , 3.8937265457787215e-46 ], 
                    [ 283.25 , 3.894765434749718e-46 ], [ 283.3125 , 3.895804600907721e-46 ], [ 283.375 , 3.896844044326602e-46 ], [ 283.4375 , 3.897883765080449e-46 ], 
                    [ 283.5 , 3.898925956950037e-46 ], [ 283.5625 , 3.8999692808110915e-46 ], [ 283.625 , 3.901012883857953e-46 ], [ 283.6875 , 3.902056766165301e-46 ], 
                    [ 283.75 , 3.903100927807864e-46 ], [ 283.8125 , 3.904145368860361e-46 ], [ 283.875 , 3.905190089397645e-46 ], [ 283.9375 , 3.906235089494364e-46 ], 
                    [ 284.0 , 3.907280369225411e-46 ], [ 284.0625 , 3.908325928665723e-46 ], [ 284.125 , 3.9093709355130525e-46 ], [ 284.1875 , 3.91041380203284e-46 ], 
                    [ 284.25 , 3.911456946748456e-46 ], [ 284.3125 , 3.9125003697340566e-46 ], [ 284.375 , 3.91354407106393e-46 ], [ 284.4375 , 3.9145880508122975e-46 ], 
                    [ 284.5 , 3.915632309053432e-46 ], [ 284.5625 , 3.916676845861596e-46 ], [ 284.625 , 3.917721661311128e-46 ], [ 284.6875 , 3.918766755476387e-46 ], 
                    [ 284.75 , 3.919812128431695e-46 ], [ 284.8125 , 3.920857074566118e-46 ], [ 284.875 , 3.9219020955057675e-46 ], [ 284.9375 , 3.922947394973489e-46 ], 
                    [ 285.0 , 3.92399297304352e-46 ], [ 285.0625 , 3.925038829790086e-46 ], [ 285.125 , 3.9260849652875204e-46 ], [ 285.1875 , 3.92713137961006e-46 ], 
                    [ 285.25 , 3.928178072832078e-46 ], [ 285.3125 , 3.929225045027879e-46 ], [ 285.375 , 3.930272296271791e-46 ], [ 285.4375 , 3.931319340233765e-46 ], 
                    [ 285.5 , 3.9323653067358275e-46 ], [ 285.5625 , 3.933411551527596e-46 ], [ 285.625 , 3.9344580746831705e-46 ], [ 285.6875 , 3.9355048762766636e-46 ], 
                    [ 285.75 , 3.936551956382048e-46 ], [ 285.8125 , 3.937599315073507e-46 ], [ 285.875 , 3.938646952425162e-46 ], [ 285.9375 , 3.939694868511097e-46 ], 
                    [ 286.0 , 3.940743063405501e-46 ], [ 286.0625 , 3.941791537182524e-46 ], [ 286.125 , 3.942842634972785e-46 ], [ 286.1875 , 3.9438948544295166e-46 ], 
                    [ 286.25 , 3.944947354690227e-46 ], [ 286.3125 , 3.9460001358297945e-46 ], [ 286.375 , 3.947053197923208e-46 ], [ 286.4375 , 3.948106541045445e-46 ], 
                    [ 286.5 , 3.9491601652715015e-46 ], [ 286.5625 , 3.950214070676424e-46 ], [ 286.625 , 3.951268257335223e-46 ], [ 286.6875 , 3.9523227253228995e-46 ], 
                    [ 286.75 , 3.953376505911057e-46 ], [ 286.8125 , 3.954426489328506e-46 ], [ 286.875 , 3.955476751612655e-46 ], [ 286.9375 , 3.956527292837626e-46 ], 
                    [ 287.0 , 3.9575781130774764e-46 ], [ 287.0625 , 3.958629212406309e-46 ], [ 287.125 , 3.959680590898275e-46 ], [ 287.1875 , 3.960732248627464e-46 ], 
                    [ 287.25 , 3.9617841856680374e-46 ], [ 287.3125 , 3.962836402094261e-46 ], [ 287.375 , 3.963888897980228e-46 ], [ 287.4375 , 3.964942174827848e-46 ], 
                    [ 287.5 , 3.965996065924001e-46 ], [ 287.5625 , 3.9670502371469806e-46 ], [ 287.625 , 3.9681046885711875e-46 ], [ 287.6875 , 3.969159420271044e-46 ], 
                    [ 287.75 , 3.970214432321132e-46 ], [ 287.8125 , 3.971269724795943e-46 ], [ 287.875 , 3.9723252977700405e-46 ], [ 287.9375 , 3.9733811513178704e-46 ], 
                    [ 288.0 , 3.974437285514207e-46 ], [ 288.0625 , 3.975493691346598e-46 ], [ 288.125 , 3.976549250901963e-46 ], [ 288.1875 , 3.9776050907258805e-46 ], 
                    [ 288.25 , 3.978661210892795e-46 ], [ 288.3125 , 3.979717611477142e-46 ], [ 288.375 , 3.980774292553377e-46 ], [ 288.4375 , 3.981831254195974e-46 ], 
                    [ 288.5 , 3.9828884964794865e-46 ], [ 288.5625 , 3.98394601947826e-46 ], [ 288.625 , 3.9850038232670525e-46 ], [ 288.6875 , 3.986061907920224e-46 ], 
                    [ 288.75 , 3.987121090340579e-46 ], [ 288.8125 , 3.988181957395143e-46 ], [ 288.875 , 3.989243106718234e-46 ], [ 288.9375 , 3.990304538384983e-46 ], 
                    [ 289.0 , 3.991366252470572e-46 ], [ 289.0625 , 3.9924282490500035e-46 ], [ 289.125 , 3.9934905281986115e-46 ], [ 289.1875 , 3.994553089991409e-46 ], 
                    [ 289.25 , 3.995615934503771e-46 ], [ 289.3125 , 3.99667906181081e-46 ], [ 289.375 , 3.997742471987766e-46 ], [ 289.4375 , 3.9988014576539684e-46 ], 
                    [ 289.5 , 3.999858964546944e-46 ], [ 289.5625 , 4.000916751103924e-46 ], [ 289.625 , 4.001974817398896e-46 ], [ 289.6875 , 4.003033163505809e-46 ], 
                    [ 289.75 , 4.004091789498634e-46 ], [ 289.8125 , 4.005150695451387e-46 ], [ 289.875 , 4.0062098814381915e-46 ], [ 289.9375 , 4.007269347533018e-46 ], 
                    [ 290.0 , 4.0083290938099994e-46 ], [ 290.0625 , 4.009389300832098e-46 ], [ 290.125 , 4.010454121162825e-46 ], [ 290.1875 , 4.011519224290348e-46 ], 
                    [ 290.25 , 4.0125846102897175e-46 ], [ 290.3125 , 4.013650279236085e-46 ], [ 290.375 , 4.0147162312045414e-46 ], [ 290.4375 , 4.015782466270422e-46 ], 
                    [ 290.5 , 4.016848984508684e-46 ], [ 290.5625 , 4.017915785994677e-46 ], [ 290.625 , 4.0189828708036814e-46 ], [ 290.6875 , 4.020050239010773e-46 ], 
                    [ 290.75 , 4.021116782166526e-46 ], [ 290.8125 , 4.022181417046041e-46 ], [ 290.875 , 4.023246333799318e-46 ], [ 290.9375 , 4.024311532501042e-46 ], 
                    [ 291.0 , 4.025377013225834e-46 ], [ 291.0625 , 4.026442776048363e-46 ], [ 291.125 , 4.027508821043317e-46 ], [ 291.1875 , 4.028575148285376e-46 ], 
                    [ 291.25 , 4.029641757849354e-46 ], [ 291.3125 , 4.030708649809826e-46 ], [ 291.375 , 4.0317758242417904e-46 ], [ 291.4375 , 4.032843532938024e-46 ], 
                    [ 291.5 , 4.033911681345266e-46 ], [ 291.5625 , 4.034980112664832e-46 ], [ 291.625 , 4.036048826971597e-46 ], [ 291.6875 , 4.037117824340571e-46 ], 
                    [ 291.75 , 4.03818710484667e-46 ], [ 291.8125 , 4.039256668564915e-46 ], [ 291.875 , 4.040326515570374e-46 ], [ 291.9375 , 4.041396645937966e-46 ], 
                    [ 292.0 , 4.042467059742828e-46 ], [ 292.0625 , 4.0435377570599755e-46 ], [ 292.125 , 4.04460451603694e-46 ], [ 292.1875 , 4.045670891706763e-46 ], 
                    [ 292.25 , 4.046737548530667e-46 ], [ 292.3125 , 4.0478044865827536e-46 ], [ 292.375 , 4.0488717059372524e-46 ], [ 292.4375 , 4.049939206668217e-46 ], 
                    [ 292.5 , 4.051006988849947e-46 ], [ 292.5625 , 4.052075052556505e-46 ], [ 292.625 , 4.053143397862289e-46 ], [ 292.6875 , 4.054212024841399e-46 ], 
                    [ 292.75 , 4.055281178665605e-46 ], [ 292.8125 , 4.056353433805237e-46 ], [ 292.875 , 4.0574259724594156e-46 ], [ 292.9375 , 4.0584987947030455e-46 ], 
                    [ 293.0 , 4.059571900611112e-46 ], [ 293.0625 , 4.060645290258675e-46 ], [ 293.125 , 4.0617189637207875e-46 ], [ 293.1875 , 4.06279292107232e-46 ], 
                    [ 293.25 , 4.0638671623885654e-46 ], [ 293.3125 , 4.0649416877444644e-46 ], [ 293.375 , 4.0660164972150585e-46 ], [ 293.4375 , 4.067091447920146e-46 ], 
                    [ 293.5 , 4.068166315117649e-46 ], [ 293.5625 , 4.069241466385326e-46 ], [ 293.625 , 4.070316901798281e-46 ], [ 293.6875 , 4.071392621431666e-46 ], 
                    [ 293.75 , 4.072468625360453e-46 ], [ 293.8125 , 4.0735449136599205e-46 ], [ 293.875 , 4.074621486405107e-46 ], [ 293.9375 , 4.075698343671332e-46 ], 
                    [ 294.0 , 4.076775485533645e-46 ], [ 294.0625 , 4.077852912067348e-46 ], [ 294.125 , 4.0789294951528414e-46 ], [ 294.1875 , 4.080005058861047e-46 ], 
                    [ 294.25 , 4.0810809061821605e-46 ], [ 294.3125 , 4.082157037191025e-46 ], [ 294.375 , 4.083233451962474e-46 ], [ 294.4375 , 4.084310150571274e-46 ], 
                    [ 294.5 , 4.0853871330923e-46 ], [ 294.5625 , 4.086464399600443e-46 ], [ 294.625 , 4.087541950170589e-46 ], [ 294.6875 , 4.088619784877553e-46 ], 
                    [ 294.75 , 4.089697903796346e-46 ], [ 294.8125 , 4.09077567002629e-46 ], [ 294.875 , 4.091853323101855e-46 ], [ 294.9375 , 4.09293126006882e-46 ], 
                    [ 295.0 , 4.09400948100203e-46 ], [ 295.0625 , 4.095087985976292e-46 ], [ 295.125 , 4.096166775066345e-46 ], [ 295.1875 , 4.097245848347121e-46 ], 
                    [ 295.25 , 4.098325205893458e-46 ], [ 295.3125 , 4.099404847780184e-46 ], [ 295.375 , 4.100484774082377e-46 ], [ 295.4375 , 4.101564984874789e-46 ], 
                    [ 295.5 , 4.102644712481683e-46 ], [ 295.5625 , 4.103724448863169e-46 ], [ 295.625 , 4.104804469410352e-46 ], [ 295.6875 , 4.1058847741979315e-46 ], 
                    [ 295.75 , 4.106965363300713e-46 ], [ 295.8125 , 4.108046236793581e-46 ], [ 295.875 , 4.109127394751295e-46 ], [ 295.9375 , 4.1102088372488956e-46 ], 
                    [ 296.0 , 4.111290564361094e-46 ], [ 296.0625 , 4.112372576162792e-46 ], [ 296.125 , 4.113454872729033e-46 ], [ 296.1875 , 4.114537172198705e-46 ], 
                    [ 296.25 , 4.1156196961992586e-46 ], [ 296.3125 , 4.116702505008995e-46 ], [ 296.375 , 4.1177855987029336e-46 ], [ 296.4375 , 4.118868977355999e-46 ], 
                    [ 296.5 , 4.119952641043164e-46 ], [ 296.5625 , 4.1210365898394184e-46 ], [ 296.625 , 4.1221208238198624e-46 ], [ 296.6875 , 4.123205343059352e-46 ], 
                    [ 296.75 , 4.1242901476329656e-46 ], [ 296.8125 , 4.125375237615893e-46 ], [ 296.875 , 4.126460443083709e-46 ], [ 296.9375 , 4.1275459176137624e-46 ], 
                    [ 297.0 , 4.128631677680228e-46 ], [ 297.0625 , 4.129717723358277e-46 ], [ 297.125 , 4.130804054723098e-46 ], [ 297.1875 , 4.131890671849783e-46 ], 
                    [ 297.25 , 4.132977574813473e-46 ], [ 297.3125 , 4.134064763689358e-46 ], [ 297.375 , 4.135152238552735e-46 ], [ 297.4375 , 4.136239999478718e-46 ], 
                    [ 297.5 , 4.137328046542672e-46 ], [ 297.5625 , 4.138414529548581e-46 ], [ 297.625 , 4.1395012045548756e-46 ], [ 297.6875 , 4.140588164902981e-46 ], 
                    [ 297.75 , 4.1416754106677915e-46 ], [ 297.8125 , 4.142762941924224e-46 ], [ 297.875 , 4.143850758747273e-46 ], [ 297.9375 , 4.144938861211924e-46 ], 
                    [ 298.0 , 4.14602724939318e-46 ], [ 298.0625 , 4.147115923366125e-46 ], [ 298.125 , 4.148204883205715e-46 ], [ 298.1875 , 4.149294128986983e-46 ], 
                    [ 298.25 , 4.150382611784852e-46 ], [ 298.3125 , 4.151471354322188e-46 ], [ 298.375 , 4.152560382462156e-46 ], [ 298.4375 , 4.153649696279702e-46 ], 
                    [ 298.5 , 4.154739295849799e-46 ], [ 298.5625 , 4.155829181247316e-46 ], [ 298.625 , 4.15691935254738e-46 ], [ 298.6875 , 4.158009809824812e-46 ], 
                    [ 298.75 , 4.1591005531547804e-46 ], [ 298.8125 , 4.160191582612263e-46 ], [ 298.875 , 4.16128289827223e-46 ], [ 298.9375 , 4.16237498347488e-46 ], 
                    [ 299.0 , 4.1634673712637245e-46 ], [ 299.0625 , 4.16456004574252e-46 ], [ 299.125 , 4.1656530069865065e-46 ], [ 299.1875 , 4.166746255070944e-46 ], 
                    [ 299.25 , 4.167839790071112e-46 ], [ 299.3125 , 4.168933612062368e-46 ], [ 299.375 , 4.170027721119913e-46 ], [ 299.4375 , 4.171122117319087e-46 ], 
                    [ 299.5 , 4.172216800735394e-46 ], [ 299.5625 , 4.1733117714441246e-46 ], [ 299.625 , 4.17440368750326e-46 ], [ 299.6875 , 4.1754956909470585e-46 ], 
                    [ 299.75 , 4.176587980053637e-46 ], [ 299.8125 , 4.177680554897578e-46 ], [ 299.875 , 4.1787734155537756e-46 ], [ 299.9375 , 4.179866562096969e-46 ], 
                    [ 300.0 , 4.180959994601853e-46 ], [ 300.0625 , 4.1820537131433255e-46 ], [ 300.125 , 4.1831477177962106e-46 ], [ 300.1875 , 4.184242008635264e-46 ], 
                    [ 300.25 , 4.185336585735442e-46 ], [ 300.3125 , 4.186434050989785e-46 ], [ 300.375 , 4.187532132261316e-46 ], [ 300.4375 , 4.188630501554189e-46 ], 
                    [ 300.5 , 4.18972915894398e-46 ], [ 300.5625 , 4.190828104506316e-46 ], [ 300.625 , 4.1919273383166634e-46 ], [ 300.6875 , 4.193026860450689e-46 ], 
                    [ 300.75 , 4.194126670983987e-46 ], [ 300.8125 , 4.195226769992293e-46 ], [ 300.875 , 4.196327157551214e-46 ], [ 300.9375 , 4.197427833736465e-46 ], 
                    [ 301.0 , 4.198525895213459e-46 ], [ 301.0625 , 4.199623589085636e-46 ], [ 301.125 , 4.200721569946983e-46 ], [ 301.1875 , 4.201819837872712e-46 ], 
                    [ 301.25 , 4.2029183929377546e-46 ], [ 301.3125 , 4.204017235217213e-46 ], [ 301.375 , 4.2051163647862095e-46 ], [ 301.4375 , 4.206215781719765e-46 ], 
                    [ 301.5 , 4.207315486093158e-46 ], [ 301.5625 , 4.208415477981363e-46 ], [ 301.625 , 4.209515757459666e-46 ], [ 301.6875 , 4.21061553081936e-46 ], 
                    [ 301.75 , 4.211715270339533e-46 ], [ 301.8125 , 4.212815297092565e-46 ], [ 301.875 , 4.2139156111534475e-46 ], [ 301.9375 , 4.2150162125972795e-46 ], 
                    [ 302.0 , 4.216117101498971e-46 ], [ 302.0625 , 4.217218277933718e-46 ], [ 302.125 , 4.2183197419766234e-46 ], [ 302.1875 , 4.2194214937027425e-46 ], 
                    [ 302.25 , 4.2205235331872745e-46 ], [ 302.3125 , 4.221625860505347e-46 ], [ 302.375 , 4.222728738909857e-46 ], [ 302.4375 , 4.223832099214471e-46 ], 
                    [ 302.5 , 4.224935747817048e-46 ], [ 302.5625 , 4.226039684792885e-46 ], [ 302.625 , 4.227143910217334e-46 ], [ 302.6875 , 4.228248424165792e-46 ], 
                    [ 302.75 , 4.229353226713528e-46 ], [ 302.8125 , 4.230458317936161e-46 ], [ 302.875 , 4.2315636979089686e-46 ], [ 302.9375 , 4.2326693667073684e-46 ], 
                    [ 303.0 , 4.2337753244068885e-46 ], [ 303.0625 , 4.2348804979748e-46 ], [ 303.125 , 4.235984501846611e-46 ], [ 303.1875 , 4.237088793524638e-46 ], 
                    [ 303.25 , 4.238193373083759e-46 ], [ 303.3125 , 4.239298240599115e-46 ], [ 303.375 , 4.24040339614574e-46 ], [ 303.4375 , 4.241508839798726e-46 ], 
                    [ 303.5 , 4.242614571633119e-46 ], [ 303.5625 , 4.243720591724165e-46 ], [ 303.625 , 4.244826900146981e-46 ], [ 303.6875 , 4.245933496976613e-46 ], 
                    [ 303.75 , 4.24704024041791e-46 ], [ 303.8125 , 4.248146822968852e-46 ], [ 303.875 , 4.2492536938440435e-46 ], [ 303.9375 , 4.250360853118666e-46 ], 
                    [ 304.0 , 4.251468300867927e-46 ], [ 304.0625 , 4.252576037166897e-46 ], [ 304.125 , 4.253684062090788e-46 ], [ 304.1875 , 4.254792375714745e-46 ], 
                    [ 304.25 , 4.255900978114138e-46 ], [ 304.3125 , 4.257009869364089e-46 ], [ 304.375 , 4.2581190495399195e-46 ], [ 304.4375 , 4.259228472892375e-46 ], 
                    [ 304.5 , 4.2603370852235174e-46 ], [ 304.5625 , 4.261445986109543e-46 ], [ 304.625 , 4.262555175625619e-46 ], [ 304.6875 , 4.263664653846779e-46 ], 
                    [ 304.75 , 4.26477442084829e-46 ], [ 304.8125 , 4.265884476705227e-46 ], [ 304.875 , 4.2669948214928354e-46 ], [ 304.9375 , 4.268105455286288e-46 ], 
                    [ 305.0 , 4.26921637816081e-46 ], [ 305.0625 , 4.270327590191583e-46 ], [ 305.125 , 4.271439091453991e-46 ], [ 305.1875 , 4.272551900243127e-46 ], 
                    [ 305.25 , 4.273665240961009e-46 ], [ 305.3125 , 4.2747788717930225e-46 ], [ 305.375 , 4.2758927928146745e-46 ], [ 305.4375 , 4.277007004101704e-46 ], 
                    [ 305.5 , 4.278121505729657e-46 ], [ 305.5625 , 4.279236297774221e-46 ], [ 305.625 , 4.2803513803111025e-46 ], [ 305.6875 , 4.281466753415937e-46 ], 
                    [ 305.75 , 4.282582417164504e-46 ], [ 305.8125 , 4.283698371632444e-46 ], [ 305.875 , 4.284812765909767e-46 ], [ 305.9375 , 4.285925898129438e-46 ], 
                    [ 306.0 , 4.287039319524747e-46 ], [ 306.0625 , 4.288153030170819e-46 ], [ 306.125 , 4.289267030142705e-46 ], [ 306.1875 , 4.2903813195156595e-46 ], 
                    [ 306.25 , 4.2914958983647734e-46 ], [ 306.3125 , 4.2926107667654e-46 ], [ 306.375 , 4.293725924792609e-46 ], [ 306.4375 , 4.294841372521701e-46 ], 
                    [ 306.5 , 4.29595711002803e-46 ], [ 306.5625 , 4.297073195170166e-46 ], [ 306.625 , 4.298189731405349e-46 ], [ 306.6875 , 4.2993065577573095e-46 ], 
                    [ 306.75 , 4.300423674301493e-46 ], [ 306.8125 , 4.301541081113303e-46 ], [ 306.875 , 4.302658778268098e-46 ], [ 306.9375 , 4.303776765841383e-46 ], 
                    [ 307.0 , 4.304895043908526e-46 ], [ 307.0625 , 4.30601361254513e-46 ], [ 307.125 , 4.307132471826575e-46 ], [ 307.1875 , 4.308251621828472e-46 ], 
                    [ 307.25 , 4.3093710626262684e-46 ], [ 307.3125 , 4.3104888761593484e-46 ], [ 307.375 , 4.311606882905771e-46 ], [ 307.4375 , 4.312725179628308e-46 ], 
                    [ 307.5 , 4.3138437664022955e-46 ], [ 307.5625 , 4.314962643302901e-46 ], [ 307.625 , 4.316081810405343e-46 ], [ 307.6875 , 4.317201267784891e-46 ], 
                    [ 307.75 , 4.318321015516896e-46 ], [ 307.8125 , 4.319441053676667e-46 ], [ 307.875 , 4.3205613823394685e-46 ], [ 307.9375 , 4.321682001580742e-46 ], 
                    [ 308.0 , 4.322803671791737e-46 ], [ 308.0625 , 4.323926091384547e-46 ], [ 308.125 , 4.325048802414573e-46 ], [ 308.1875 , 4.326171804957489e-46 ], 
                    [ 308.25 , 4.327295099088864e-46 ], [ 308.3125 , 4.328418684884562e-46 ], [ 308.375 , 4.329542562420252e-46 ], [ 308.4375 , 4.3306667317716245e-46 ], 
                    [ 308.5 , 4.3317911930145715e-46 ], [ 308.5625 , 4.332915946224822e-46 ], [ 308.625 , 4.3340409914782155e-46 ], [ 308.6875 , 4.335165376136929e-46 ], 
                    [ 308.75 , 4.336287395758159e-46 ], [ 308.8125 , 4.3374097057783754e-46 ], [ 308.875 , 4.338532306272861e-46 ], [ 308.9375 , 4.339655197316828e-46 ], 
                    [ 309.0 , 4.340778378985352e-46 ], [ 309.0625 , 4.341901851353775e-46 ], [ 309.125 , 4.343025614497212e-46 ], [ 309.1875 , 4.344149668490952e-46 ], 
                    [ 309.25 , 4.345274013410426e-46 ], [ 309.3125 , 4.346398649330685e-46 ], [ 309.375 , 4.347523576327199e-46 ], [ 309.4375 , 4.348646536099121e-46 ], 
                    [ 309.5 , 4.349769454339811e-46 ], [ 309.5625 , 4.35089266254319e-46 ], [ 309.625 , 4.352016160784164e-46 ], [ 309.6875 , 4.3531399491376255e-46 ], 
                    [ 309.75 , 4.35426402767852e-46 ], [ 309.8125 , 4.3553883964817504e-46 ], [ 309.875 , 4.356513055622236e-46 ], [ 309.9375 , 4.357638005174981e-46 ], 
                    [ 310.0 , 4.358763245214976e-46 ], [ 310.0625 , 4.359888775817293e-46 ], [ 310.125 , 4.361016456597885e-46 ], [ 310.1875 , 4.3621467208658926e-46 ], 
                    [ 310.25 , 4.363277278069549e-46 ], [ 310.3125 , 4.3644081282848035e-46 ], [ 310.375 , 4.365539271587601e-46 ], [ 310.4375 , 4.366670708053962e-46 ], 
                    [ 310.5 , 4.3678024377597125e-46 ], [ 310.5625 , 4.368934460781008e-46 ], [ 310.625 , 4.370066777193743e-46 ], [ 310.6875 , 4.3711993870741155e-46 ], 
                    [ 310.75 , 4.372332290498026e-46 ], [ 310.8125 , 4.373465439384504e-46 ], [ 310.875 , 4.374592908915144e-46 ], [ 310.9375 , 4.375720669104826e-46 ], 
                    [ 311.0 , 4.376848720028543e-46 ], [ 311.0625 , 4.377977061761306e-46 ], [ 311.125 , 4.379105694378024e-46 ], [ 311.1875 , 4.380234617953624e-46 ], 
                    [ 311.25 , 4.381363832563178e-46 ], [ 311.3125 , 4.382493338281743e-46 ], [ 311.375 , 4.383623135184336e-46 ], [ 311.4375 , 4.384753223346054e-46 ], 
                    [ 311.5 , 4.3858836028419525e-46 ], [ 311.5625 , 4.387015744459943e-46 ], [ 311.625 , 4.3881494518045966e-46 ], [ 311.6875 , 4.389283452125742e-46 ], 
                    [ 311.75 , 4.3904177454990275e-46 ], [ 311.8125 , 4.3915523320002175e-46 ], [ 311.875 , 4.3926872117050615e-46 ], [ 311.9375 , 4.393822384689394e-46 ], 
                    [ 312.0 , 4.394957851028848e-46 ], [ 312.0625 , 4.3960936107993895e-46 ], [ 312.125 , 4.397229664076756e-46 ], [ 312.1875 , 4.3983660109368876e-46 ], 
                    [ 312.25 , 4.399502591533166e-46 ], [ 312.3125 , 4.400638277065742e-46 ], [ 312.375 , 4.401774255763616e-46 ], [ 312.4375 , 4.40291052770256e-46 ], 
                    [ 312.5 , 4.404047092958177e-46 ], [ 312.5625 , 4.4051839516062475e-46 ], [ 312.625 , 4.406321103722475e-46 ], [ 312.6875 , 4.407458549382553e-46 ], 
                    [ 312.75 , 4.40859628866232e-46 ], [ 312.8125 , 4.4097343216376326e-46 ], [ 312.875 , 4.41087264838418e-46 ], [ 312.9375 , 4.412011268977891e-46 ], 
                    [ 313.0 , 4.413147531604675e-46 ], [ 313.0625 , 4.414281715624687e-46 ], [ 313.125 , 4.415416191131304e-46 ], [ 313.1875 , 4.416550958199437e-46 ], 
                    [ 313.25 , 4.417686016904051e-46 ], [ 313.3125 , 4.418821367320002e-46 ], [ 313.375 , 4.419957009522386e-46 ], [ 313.4375 , 4.4210929435861296e-46 ], 
                    [ 313.5 , 4.422229169586242e-46 ], [ 313.5625 , 4.4233656875976864e-46 ], [ 313.625 , 4.424502497695639e-46 ], [ 313.6875 , 4.425639599955068e-46 ], 
                    [ 313.75 , 4.4267809907852594e-46 ], [ 313.8125 , 4.427922774280829e-46 ], [ 313.875 , 4.4290648522723935e-46 ], [ 313.9375 , 4.430207224835975e-46 ], 
                    [ 314.0 , 4.43134989204758e-46 ], [ 314.0625 , 4.4324928539831776e-46 ], [ 314.125 , 4.433636110718752e-46 ], [ 314.1875 , 4.434779662330435e-46 ], 
                    [ 314.25 , 4.43592350889422e-46 ], [ 314.3125 , 4.4370676504861185e-46 ], [ 314.375 , 4.4382120871823225e-46 ], [ 314.4375 , 4.4393553279652704e-46 ], 
                    [ 314.5 , 4.440496699131981e-46 ], [ 314.5625 , 4.441638363748539e-46 ], [ 314.625 , 4.442780321890389e-46 ], [ 314.6875 , 4.4439225736330635e-46 ], 
                    [ 314.75 , 4.445065119051887e-46 ], [ 314.8125 , 4.446207958222525e-46 ], [ 314.875 , 4.447351091220374e-46 ], [ 314.9375 , 4.448494518121042e-46 ], 
                    [ 315.0 , 4.449638239000186e-46 ], [ 315.0625 , 4.4507822539331995e-46 ], [ 315.125 , 4.451926562995875e-46 ], [ 315.1875 , 4.453070138185752e-46 ], 
                    [ 315.25 , 4.454213762769707e-46 ], [ 315.3125 , 4.455357681055955e-46 ], [ 315.375 , 4.456501893119986e-46 ], [ 315.4375 , 4.4576463990372475e-46 ], 
                    [ 315.5 , 4.458791198883206e-46 ], [ 315.5625 , 4.459936292733347e-46 ], [ 315.625 , 4.461081680663209e-46 ], [ 315.6875 , 4.462227362748251e-46 ], 
                    [ 315.75 , 4.463373339064083e-46 ], [ 315.8125 , 4.464519609686298e-46 ], [ 315.875 , 4.465666226233291e-46 ], [ 315.9375 , 4.466813378641464e-46 ], 
                    [ 316.0 , 4.4679608257332145e-46 ], [ 316.0625 , 4.469108567584307e-46 ], [ 316.125 , 4.470256604270365e-46 ], [ 316.1875 , 4.471404935867284e-46 ], 
                    [ 316.25 , 4.472553562450663e-46 ], [ 316.3125 , 4.473702484096375e-46 ], [ 316.375 , 4.474851700880153e-46 ], [ 316.4375 , 4.476001212877906e-46 ], 
                    [ 316.5 , 4.477151020165314e-46 ], [ 316.5625 , 4.478301122818388e-46 ], [ 316.625 , 4.479449587992284e-46 ], [ 316.6875 , 4.480596619334939e-46 ], 
                    [ 316.75 , 4.481743944392516e-46 ], [ 316.8125 , 4.482891563240131e-46 ], [ 316.875 , 4.4840394759530434e-46 ], [ 316.9375 , 4.485187682606534e-46 ], 
                    [ 317.0 , 4.486336183275807e-46 ], [ 317.0625 , 4.487484978036246e-46 ], [ 317.125 , 4.488634066962933e-46 ], [ 317.1875 , 4.489783450131515e-46 ], 
                    [ 317.25 , 4.490933127617111e-46 ], [ 317.3125 , 4.49208309949512e-46 ], [ 317.375 , 4.49323427677599e-46 ], [ 317.4375 , 4.494385912389409e-46 ], 
                    [ 317.5 , 4.49553784317211e-46 ], [ 317.5625 , 4.496690069199648e-46 ], [ 317.625 , 4.497842590547825e-46 ], [ 317.6875 , 4.498995407292268e-46 ], 
                    [ 317.75 , 4.5001485195086574e-46 ], [ 317.8125 , 4.501301927272723e-46 ], [ 317.875 , 4.502455630660312e-46 ], [ 317.9375 , 4.503609629747129e-46 ], 
                    [ 318.0 , 4.504763924608899e-46 ], [ 318.0625 , 4.5059182299111984e-46 ], [ 318.125 , 4.5070710173346196e-46 ], [ 318.1875 , 4.508224099685462e-46 ], 
                    [ 318.25 , 4.509377477039179e-46 ], [ 318.3125 , 4.51053114947118e-46 ], [ 318.375 , 4.511685117057118e-46 ], [ 318.4375 , 4.5128393798723765e-46 ], 
                    [ 318.5 , 4.513993937992518e-46 ], [ 318.5625 , 4.515148791493093e-46 ], [ 318.625 , 4.516303940449671e-46 ], [ 318.6875 , 4.517459384937809e-46 ], 
                    [ 318.75 , 4.518615125033148e-46 ], [ 318.8125 , 4.519770946294638e-46 ], [ 318.875 , 4.520926751865698e-46 ], [ 318.9375 , 4.522082853001882e-46 ], 
                    [ 319.0 , 4.523239249778742e-46 ], [ 319.0625 , 4.524395942271846e-46 ], [ 319.125 , 4.52555293055688e-46 ], [ 319.1875 , 4.526710214709421e-46 ], 
                    [ 319.25 , 4.527867794805162e-46 ], [ 319.3125 , 4.529025670919811e-46 ], [ 319.375 , 4.530183843129006e-46 ], [ 319.4375 , 4.5313423115085264e-46 ], 
                    [ 319.5 , 4.532501076134111e-46 ], [ 319.5625 , 4.5336606715815266e-46 ], [ 319.625 , 4.534820875007595e-46 ], [ 319.6875 , 4.535981375339806e-46 ], 
                    [ 319.75 , 4.537142172654235e-46 ], [ 319.8125 , 4.538303267026917e-46 ], [ 319.875 , 4.5394646585337084e-46 ], [ 319.9375 , 4.5406263472507784e-46 ], 
                    [ 320.0 , 4.541788333254218e-46 ], [ 320.0625 , 4.542950616619945e-46 ], [ 320.125 , 4.544113197424185e-46 ], [ 320.1875 , 4.545276075743022e-46 ], 
                    [ 320.25 , 4.5464392516525945e-46 ], [ 320.3125 , 4.5475990502736604e-46 ], [ 320.375 , 4.548758526385651e-46 ], [ 320.4375 , 4.549918298122805e-46 ], 
                    [ 320.5 , 4.5510783655605595e-46 ], [ 320.5625 , 4.552238728774242e-46 ], [ 320.625 , 4.553399387839332e-46 ], [ 320.6875 , 4.554560342831261e-46 ], 
                    [ 320.75 , 4.555721593825413e-46 ], [ 320.8125 , 4.556883140897291e-46 ], [ 320.875 , 4.558044984122288e-46 ], [ 320.9375 , 4.5592071235760745e-46 ], 
                    [ 321.0 , 4.5603695668360775e-46 ], [ 321.0625 , 4.561532533468966e-46 ], [ 321.125 , 4.5626957966768085e-46 ], [ 321.1875 , 4.563859356535235e-46 ], 
                    [ 321.25 , 4.565023213119863e-46 ], [ 321.3125 , 4.5661873665064615e-46 ], [ 321.375 , 4.567351816770621e-46 ], [ 321.4375 , 4.568516563988148e-46 ], 
                    [ 321.5 , 4.569681608234542e-46 ], [ 321.5625 , 4.570846949585845e-46 ], [ 321.625 , 4.5720125881176245e-46 ], [ 321.6875 , 4.573178523905799e-46 ], 
                    [ 321.75 , 4.574344497991828e-46 ], [ 321.8125 , 4.575509678973751e-46 ], [ 321.875 , 4.576675156751515e-46 ], [ 321.9375 , 4.577840931400818e-46 ], 
                    [ 322.0 , 4.579007002997248e-46 ], [ 322.0625 , 4.580173371616407e-46 ], [ 322.125 , 4.58134003733399e-46 ], [ 322.1875 , 4.5825070002257374e-46 ], 
                    [ 322.25 , 4.583674260367215e-46 ], [ 322.3125 , 4.584841817834139e-46 ], [ 322.375 , 4.586009672702473e-46 ], [ 322.4375 , 4.587177825047742e-46 ], 
                    [ 322.5 , 4.588345533200091e-46 ], [ 322.5625 , 4.589511961993092e-46 ], [ 322.625 , 4.590678687310333e-46 ], [ 322.6875 , 4.591845709227361e-46 ], 
                    [ 322.75 , 4.593013027819509e-46 ], [ 322.8125 , 4.594180643162165e-46 ], [ 322.875 , 4.595348555330768e-46 ], [ 322.9375 , 4.596516764400905e-46 ], 
                    [ 323.0 , 4.597685270447859e-46 ], [ 323.0625 , 4.598854073547319e-46 ], [ 323.125 , 4.600023173774673e-46 ], [ 323.1875 , 4.601192571205421e-46 ], 
                    [ 323.25 , 4.602362753031854e-46 ], [ 323.3125 , 4.603533916471402e-46 ], [ 323.375 , 4.604705377936927e-46 ], [ 323.4375 , 4.605877137504332e-46 ], 
                    [ 323.5 , 4.607049195249442e-46 ], [ 323.5625 , 4.608221551248069e-46 ], [ 323.625 , 4.609394205576274e-46 ], [ 323.6875 , 4.610567158309842e-46 ], 
                    [ 323.75 , 4.611740409524707e-46 ], [ 323.8125 , 4.612913959296857e-46 ], [ 323.875 , 4.614087807702332e-46 ], [ 323.9375 , 4.61526195481696e-46 ], 
                    [ 324.0 , 4.616435242105401e-46 ], [ 324.0625 , 4.617607650068e-46 ], [ 324.125 , 4.618780355779881e-46 ], [ 324.1875 , 4.619953359316759e-46 ], 
                    [ 324.25 , 4.621126660754142e-46 ], [ 324.3125 , 4.622300260167816e-46 ], [ 324.375 , 4.623474157633291e-46 ], [ 324.4375 , 4.624648353226493e-46 ], 
                    [ 324.5 , 4.625822847022839e-46 ], [ 324.5625 , 4.626997639098326e-46 ], [ 324.625 , 4.6281727295286054e-46 ], [ 324.6875 , 4.6293481183894504e-46 ], 
                    [ 324.75 , 4.63052230566427e-46 ], [ 324.8125 , 4.631695533010927e-46 ], [ 324.875 , 4.632869057616113e-46 ], [ 324.9375 , 4.634042879555179e-46 ], 
                    [ 325.0 , 4.635216998903492e-46 ], [ 325.0625 , 4.636391415736306e-46 ], [ 325.125 , 4.637566130129128e-46 ], [ 325.1875 , 4.6387411421572835e-46 ], 
                    [ 325.25 , 4.639916451896117e-46 ], [ 325.3125 , 4.641092059421126e-46 ], [ 325.375 , 4.642267964807859e-46 ], [ 325.4375 , 4.6434441681316196e-46 ], 
                    [ 325.5 , 4.6446228520106165e-46 ], [ 325.5625 , 4.64580355037564e-46 ], [ 325.625 , 4.6469845488831444e-46 ], [ 325.6875 , 4.6481658476094305e-46 ], 
                    [ 325.75 , 4.649347446630813e-46 ], [ 325.8125 , 4.650529346023631e-46 ], [ 325.875 , 4.6517115458643085e-46 ], [ 325.9375 , 4.652894046229021e-46 ], 
                    [ 326.0 , 4.654076847194429e-46 ], [ 326.0625 , 4.655259948836784e-46 ], [ 326.125 , 4.656443351232518e-46 ], [ 326.1875 , 4.6576270544580855e-46 ], 
                    [ 326.25 , 4.658807312981281e-46 ], [ 326.3125 , 4.659984926863994e-46 ], [ 326.375 , 4.661162838413976e-46 ], [ 326.4375 , 4.662341047706406e-46 ], 
                    [ 326.5 , 4.6635195548166744e-46 ], [ 326.5625 , 4.664698359819996e-46 ], [ 326.625 , 4.6658774627916695e-46 ], [ 326.6875 , 4.66705686380698e-46 ], 
                    [ 326.75 , 4.668236562941329e-46 ], [ 326.8125 , 4.669416560270075e-46 ], [ 326.875 , 4.670596855868493e-46 ], [ 326.9375 , 4.671777449812076e-46 ], 
                    [ 327.0 , 4.672957943795494e-46 ], [ 327.0625 , 4.674138356271071e-46 ], [ 327.125 , 4.675319066924658e-46 ], [ 327.1875 , 4.676500075831677e-46 ], 
                    [ 327.25 , 4.677681383067401e-46 ], [ 327.3125 , 4.678862988707288e-46 ], [ 327.375 , 4.680044892826619e-46 ], [ 327.4375 , 4.681227095500756e-46 ], 
                    [ 327.5 , 4.68240959680525e-46 ], [ 327.5625 , 4.68359239681547e-46 ], [ 327.625 , 4.6847754956069015e-46 ], [ 327.6875 , 4.685958893254954e-46 ], 
                    [ 327.75 , 4.6871437042185945e-46 ], [ 327.8125 , 4.688330188180354e-46 ], [ 327.875 , 4.6895169724837036e-46 ], [ 327.9375 , 4.690704057204636e-46 ], 
                    [ 328.0 , 4.6918914424193e-46 ], [ 328.0625 , 4.693079128203658e-46 ], [ 328.125 , 4.6942671146338315e-46 ], [ 328.1875 , 4.695455401785856e-46 ], 
                    [ 328.25 , 4.696643989735989e-46 ], [ 328.3125 , 4.69783287856034e-46 ], [ 328.375 , 4.6990220683349376e-46 ], [ 328.4375 , 4.700211559136196e-46 ], 
                    [ 328.5 , 4.701400247335583e-46 ], [ 328.5625 , 4.70258734018412e-46 ], [ 328.625 , 4.703774732770895e-46 ], [ 328.6875 , 4.704962425171626e-46 ], 
                    [ 328.75 , 4.706150417461948e-46 ], [ 328.8125 , 4.707338709717616e-46 ], [ 328.875 , 4.7085273020144055e-46 ], [ 328.9375 , 4.70971619442804e-46 ], 
                    [ 329.0 , 4.710905387034266e-46 ], [ 329.0625 , 4.712094879908882e-46 ], [ 329.125 , 4.713284673127805e-46 ], [ 329.1875 , 4.714474766766771e-46 ], 
                    [ 329.25 , 4.715664637212006e-46 ], [ 329.3125 , 4.716853149195695e-46 ], [ 329.375 , 4.7180419607259975e-46 ], [ 329.4375 , 4.719231071878174e-46 ], 
                    [ 329.5 , 4.720420482727873e-46 ], [ 329.5625 , 4.7216101933506304e-46 ], [ 329.625 , 4.722800203821897e-46 ], [ 329.6875 , 4.7239905142173475e-46 ], 
                    [ 329.75 , 4.725181124612539e-46 ], [ 329.8125 , 4.726372035083082e-46 ], [ 329.875 , 4.7275632457046386e-46 ], [ 329.9375 , 4.728754756552758e-46 ], 
                    [ 330.0 , 4.729946652195158e-46 ], [ 330.0625 , 4.731139644090702e-46 ], [ 330.125 , 4.732332936883892e-46 ], [ 330.1875 , 4.733526530650589e-46 ], 
                    [ 330.25 , 4.734720425466706e-46 ], [ 330.3125 , 4.735914621408171e-46 ], [ 330.375 , 4.7371091185509365e-46 ], [ 330.4375 , 4.738303916970905e-46 ], 
                    [ 330.5 , 4.7394990167442306e-46 ], [ 330.5625 , 4.7406944179467545e-46 ], [ 330.625 , 4.741890120654672e-46 ], [ 330.6875 , 4.743086124943792e-46 ], 
                    [ 330.75 , 4.74428243089045e-46 ], [ 330.8125 , 4.7454755756138e-46 ], [ 330.875 , 4.746668719200185e-46 ], [ 330.9375 , 4.7478621627757774e-46 ], 
                    [ 331.0 , 4.74905590641607e-46 ], [ 331.0625 , 4.750249950196408e-46 ], [ 331.125 , 4.751444294192387e-46 ], [ 331.1875 , 4.752638938479424e-46 ], 
                    [ 331.25 , 4.753833883132986e-46 ], [ 331.3125 , 4.75502912822866e-46 ], [ 331.375 , 4.756224673841921e-46 ], [ 331.4375 , 4.7574205200483584e-46 ], 
                    [ 331.5 , 4.758616666923584e-46 ], [ 331.5625 , 4.759813197442566e-46 ], [ 331.625 , 4.761010062364092e-46 ], [ 331.6875 , 4.7622072282396546e-46 ], 
                    [ 331.75 , 4.7634046951451334e-46 ], [ 331.8125 , 4.7646024631561195e-46 ], [ 331.875 , 4.765800532348395e-46 ], [ 331.9375 , 4.7669989027976605e-46 ], 
                    [ 332.0 , 4.768197574579665e-46 ], [ 332.0625 , 4.7693965477701805e-46 ], [ 332.125 , 4.7705958224449634e-46 ], [ 332.1875 , 4.771795398679856e-46 ], 
                    [ 332.25 , 4.7729952765507535e-46 ], [ 332.3125 , 4.7741957732929394e-46 ], [ 332.375 , 4.775396904823654e-46 ], [ 332.4375 , 4.776598338544906e-46 ], 
                    [ 332.5 , 4.7778000745327584e-46 ], [ 332.5625 , 4.779002112863291e-46 ], [ 332.625 , 4.7802044536125375e-46 ], [ 332.6875 , 4.781407096856547e-46 ], 
                    [ 332.75 , 4.782610042671492e-46 ], [ 332.8125 , 4.783813291133428e-46 ], [ 332.875 , 4.785016842318532e-46 ], [ 332.9375 , 4.786220696302999e-46 ], 
                    [ 333.0 , 4.787424853162977e-46 ], [ 333.0625 , 4.788628536717324e-46 ], [ 333.125 , 4.789829704675259e-46 ], [ 333.1875 , 4.7910311739313156e-46 ], 
                    [ 333.25 , 4.792232944560937e-46 ], [ 333.3125 , 4.793435016639784e-46 ], [ 333.375 , 4.794637390243441e-46 ], [ 333.4375 , 4.795840065447607e-46 ], 
                    [ 333.5 , 4.797043042327937e-46 ], [ 333.5625 , 4.798246320960032e-46 ], [ 333.625 , 4.799449901419618e-46 ], [ 333.6875 , 4.80065378378244e-46 ], 
                    [ 333.75 , 4.801857968124123e-46 ], [ 333.8125 , 4.8030624545204835e-46 ], [ 333.875 , 4.804266465585882e-46 ], [ 333.9375 , 4.805470718132192e-46 ], 
                    [ 334.0 , 4.806675272540232e-46 ], [ 334.0625 , 4.807880128885672e-46 ], [ 334.125 , 4.809085287244124e-46 ], [ 334.1875 , 4.810290747691431e-46 ], 
                    [ 334.25 , 4.81149651030321e-46 ], [ 334.3125 , 4.8127025751553084e-46 ], [ 334.375 , 4.813908942323279e-46 ], [ 334.4375 , 4.815115611883211e-46 ], 
                    [ 334.5 , 4.816322583910662e-46 ], [ 334.5625 , 4.817529858481417e-46 ], [ 334.625 , 4.818736110111159e-46 ], [ 334.6875 , 4.819941864836794e-46 ], 
                    [ 334.75 , 4.821147921269057e-46 ], [ 334.8125 , 4.822354279483408e-46 ], [ 334.875 , 4.82356093955536e-46 ], [ 334.9375 , 4.8247679015604075e-46 ], 
                    [ 335.0 , 4.825975165574207e-46 ], [ 335.0625 , 4.827182731672153e-46 ], [ 335.125 , 4.828390599929975e-46 ], [ 335.1875 , 4.8295987704233105e-46 ], 
                    [ 335.25 , 4.830807243227753e-46 ], [ 335.3125 , 4.832016018418776e-46 ], [ 335.375 , 4.833225488515202e-46 ], [ 335.4375 , 4.834436311971478e-46 ], 
                    [ 335.5 , 4.835647438764061e-46 ], [ 335.5625 , 4.8368588689691515e-46 ], [ 335.625 , 4.8380706026626566e-46 ], [ 335.6875 , 4.839282639920675e-46 ], 
                    [ 335.75 , 4.840494980819188e-46 ], [ 335.8125 , 4.8417076254342976e-46 ], [ 335.875 , 4.8429205738420916e-46 ], [ 335.9375 , 4.84413382611871e-46 ], 
                    [ 336.0 , 4.845347382340244e-46 ], [ 336.0625 , 4.8465612425828396e-46 ], [ 336.125 , 4.847775406922624e-46 ], [ 336.1875 , 4.8489885068770115e-46 ], 
                    [ 336.25 , 4.850201765025679e-46 ], [ 336.3125 , 4.851415326741783e-46 ], [ 336.375 , 4.852629192101379e-46 ], [ 336.4375 , 4.8538433611804415e-46 ], 
                    [ 336.5 , 4.855057834054931e-46 ], [ 336.5625 , 4.856272610800824e-46 ], [ 336.625 , 4.857487691494187e-46 ], [ 336.6875 , 4.858703076211034e-46 ], 
                    [ 336.75 , 4.859918765027539e-46 ], [ 336.8125 , 4.861134758019653e-46 ], [ 336.875 , 4.862351055263654e-46 ], [ 336.9375 , 4.863565737602892e-46 ], 
                    [ 337.0 , 4.8647788341469e-46 ], [ 337.0625 , 4.8659922332678934e-46 ], [ 337.125 , 4.867205935041342e-46 ], [ 337.1875 , 4.868419939542839e-46 ], 
                    [ 337.25 , 4.8696342468477184e-46 ], [ 337.3125 , 4.8708488570316825e-46 ], [ 337.375 , 4.87206377017017e-46 ], [ 337.4375 , 4.873278986338852e-46 ], 
                    [ 337.5 , 4.874494505613172e-46 ], [ 337.5625 , 4.875710328068837e-46 ], [ 337.625 , 4.876926453781432e-46 ], [ 337.6875 , 4.8781427738128564e-46 ], 
                    [ 337.75 , 4.87935799178167e-46 ], [ 337.8125 , 4.8805735124794014e-46 ], [ 337.875 , 4.881789335981394e-46 ], [ 337.9375 , 4.883005462363187e-46 ], 
                    [ 338.0 , 4.8842218917000575e-46 ], [ 338.0625 , 4.88543862406765e-46 ], [ 338.125 , 4.8866556595413846e-46 ], [ 338.1875 , 4.887872998196769e-46 ], 
                    [ 338.25 , 4.889090640109331e-46 ], [ 338.3125 , 4.890308585354582e-46 ], [ 338.375 , 4.89152683400812e-46 ], [ 338.4375 , 4.892745386145599e-46 ], 
                    [ 338.5 , 4.893965653278226e-46 ], [ 338.5625 , 4.895187134981434e-46 ], [ 338.625 , 4.896408921553465e-46 ], [ 338.6875 , 4.897631013070411e-46 ], 
                    [ 338.75 , 4.89885340960835e-46 ], [ 338.8125 , 4.900076111243481e-46 ], [ 338.875 , 4.901299118051986e-46 ], [ 338.9375 , 4.902522430109825e-46 ], 
                    [ 339.0 , 4.903746047493431e-46 ], [ 339.0625 , 4.904969970278833e-46 ], [ 339.125 , 4.906194198542361e-46 ], [ 339.1875 , 4.907418732360225e-46 ], 
                    [ 339.25 , 4.908642881965464e-46 ], [ 339.3125 , 4.909862636206971e-46 ], [ 339.375 , 4.911082693546558e-46 ], [ 339.4375 , 4.912303054059613e-46 ], 
                    [ 339.5 , 4.913523717821403e-46 ], [ 339.5625 , 4.914744684907317e-46 ], [ 339.625 , 4.915965955392762e-46 ], [ 339.6875 , 4.917187529353027e-46 ], 
                    [ 339.75 , 4.918409406863627e-46 ], [ 339.8125 , 4.919631588000026e-46 ], [ 339.875 , 4.920854072837498e-46 ], [ 339.9375 , 4.922076861451651e-46 ], 
                    [ 340.0 , 4.923299953917969e-46 ], [ 340.0625 , 4.9245241855094725e-46 ], [ 340.125 , 4.9257492784563355e-46 ], [ 340.1875 , 4.9269746761743146e-46 ], 
                    [ 340.25 , 4.928200378739193e-46 ], [ 340.3125 , 4.929426386226844e-46 ], [ 340.375 , 4.930652698713159e-46 ], [ 340.4375 , 4.931879316273978e-46 ], 
                    [ 340.5 , 4.9331062389851274e-46 ], [ 340.5625 , 4.93433346692266e-46 ], [ 340.625 , 4.935561000162402e-46 ], [ 340.6875 , 4.936788838780376e-46 ], 
                    [ 340.75 , 4.938016982852516e-46 ], [ 340.8125 , 4.939245268499227e-46 ], [ 340.875 , 4.9404710951618484e-46 ], [ 340.9375 , 4.94169722605135e-46 ], 
                    [ 341.0 , 4.9429236612431654e-46 ], [ 341.0625 , 4.944150400812816e-46 ], [ 341.125 , 4.94537744483595e-46 ], [ 341.1875 , 4.946604793388018e-46 ], 
                    [ 341.25 , 4.947832446544602e-46 ], [ 341.3125 , 4.949060404381437e-46 ], [ 341.375 , 4.9502886669739635e-46 ], [ 341.4375 , 4.951517234397887e-46 ], 
                    [ 341.5 , 4.9527461067288965e-46 ], [ 341.5625 , 4.953975284042593e-46 ], [ 341.625 , 4.955205425266819e-46 ], [ 341.6875 , 4.956436585886205e-46 ], 
                    [ 341.75 , 4.957668052397273e-46 ], [ 341.8125 , 4.958899824876165e-46 ], [ 341.875 , 4.960131903398725e-46 ], [ 341.9375 , 4.961364288041203e-46 ], 
                    [ 342.0 , 4.962596978879518e-46 ], [ 342.0625 , 4.963829975989709e-46 ], [ 342.125 , 4.965063279448084e-46 ], [ 342.1875 , 4.966296889330547e-46 ], 
                    [ 342.25 , 4.967530805713301e-46 ], [ 342.3125 , 4.968765028672463e-46 ], [ 342.375 , 4.969999558284276e-46 ], [ 342.4375 , 4.971230390042981e-46 ], 
                    [ 342.5 , 4.972461021412755e-46 ], [ 342.5625 , 4.973691957426139e-46 ], [ 342.625 , 4.974923198158511e-46 ], [ 342.6875 , 4.976154743685341e-46 ], 
                    [ 342.75 , 4.977386594082151e-46 ], [ 342.8125 , 4.978618749424234e-46 ], [ 342.875 , 4.9798512097872234e-46 ], [ 342.9375 , 4.981083975246661e-46 ], 
                    [ 343.0 , 4.982317045878003e-46 ], [ 343.0625 , 4.983550421756723e-46 ], [ 343.125 , 4.984784102958494e-46 ], [ 343.1875 , 4.9860184252428806e-46 ], 
                    [ 343.25 , 4.987254071300574e-46 ], [ 343.3125 , 4.9884900235787565e-46 ], [ 343.375 , 4.989726282153388e-46 ], [ 343.4375 , 4.990962847100339e-46 ], 
                    [ 343.5 , 4.992199718495537e-46 ], [ 343.5625 , 4.9934368964149254e-46 ], [ 343.625 , 4.994674380934398e-46 ], [ 343.6875 , 4.995912172130113e-46 ], 
                    [ 343.75 , 4.997150270077931e-46 ], [ 343.8125 , 4.998388674853871e-46 ], [ 343.875 , 4.999627386534007e-46 ], [ 343.9375 , 5.000866405194433e-46 ], 
                    [ 344.0 , 5.0021038638902014e-46 ], [ 344.0625 , 5.003340341745034e-46 ], [ 344.125 , 5.0045771252467906e-46 ], [ 344.1875 , 5.005814214470954e-46 ], 
                    [ 344.25 , 5.007051609493097e-46 ], [ 344.3125 , 5.008289310388916e-46 ], [ 344.375 , 5.009527317233949e-46 ], [ 344.4375 , 5.010765630103755e-46 ], 
                    [ 344.5 , 5.0120042490740854e-46 ], [ 344.5625 , 5.013243174220535e-46 ], [ 344.625 , 5.014482405618823e-46 ], [ 344.6875 , 5.015721943344761e-46 ], 
                    [ 344.75 , 5.01696178747389e-46 ], [ 344.8125 , 5.018201112794603e-46 ], [ 344.875 , 5.019440640146221e-46 ], [ 344.9375 , 5.020680473668954e-46 ], 
                    [ 345.0 , 5.021920613438359e-46 ], [ 345.0625 , 5.023161059530114e-46 ], [ 345.125 , 5.024401812019848e-46 ], [ 345.1875 , 5.0256428709833514e-46 ], 
                    [ 345.25 , 5.026884236496252e-46 ], [ 345.3125 , 5.0281259086341636e-46 ], [ 345.375 , 5.029367887473039e-46 ], [ 345.4375 , 5.030610173088454e-46 ], 
                    [ 345.5 , 5.031852765556332e-46 ], [ 345.5625 , 5.033095807341618e-46 ], [ 345.625 , 5.034339861541041e-46 ], [ 345.6875 , 5.035584223239249e-46 ], 
                    [ 345.75 , 5.036828892512249e-46 ], [ 345.8125 , 5.0380738694361015e-46 ], [ 345.875 , 5.039319154086743e-46 ], [ 345.9375 , 5.0405647465404134e-46 ], 
                    [ 346.0 , 5.041810646873053e-46 ], [ 346.0625 , 5.0430568551607246e-46 ], [ 346.125 , 5.044303371479618e-46 ], [ 346.1875 , 5.045550195905979e-46 ], 
                    [ 346.25 , 5.046797328515821e-46 ], [ 346.3125 , 5.048044769385317e-46 ], [ 346.375 , 5.049290170083346e-46 ], [ 346.4375 , 5.050532789421189e-46 ], 
                    [ 346.5 , 5.0517757145649865e-46 ], [ 346.5625 , 5.05301894558989e-46 ], [ 346.625 , 5.054262482571281e-46 ], [ 346.6875 , 5.055506325584423e-46 ], 
                    [ 346.75 , 5.05675047470459e-46 ], [ 346.8125 , 5.0579949300072236e-46 ], [ 346.875 , 5.059239691567496e-46 ], [ 346.9375 , 5.0604847594608826e-46 ], 
                    [ 347.0 , 5.061730133762952e-46 ], [ 347.0625 , 5.062975814548821e-46 ], [ 347.125 , 5.0642218018939905e-46 ], [ 347.1875 , 5.065469395826706e-46 ], 
                    [ 347.25 , 5.06671800343108e-46 ], [ 347.3125 , 5.06796691880965e-46 ], [ 347.375 , 5.0692161420383516e-46 ], [ 347.4375 , 5.070465673193032e-46 ], 
                    [ 347.5 , 5.071715512349629e-46 ], [ 347.5625 , 5.072965659584028e-46 ], [ 347.625 , 5.0742161149721315e-46 ], [ 347.6875 , 5.0754668785899344e-46 ], 
                    [ 347.75 , 5.076717950513448e-46 ], [ 347.8125 , 5.077969330818598e-46 ], [ 347.875 , 5.079221019581361e-46 ], [ 347.9375 , 5.080473016877949e-46 ], 
                    [ 348.0 , 5.081723384489637e-46 ], [ 348.0625 , 5.082973668349624e-46 ], [ 348.125 , 5.084224259823811e-46 ], [ 348.1875 , 5.085475158987705e-46 ], 
                    [ 348.25 , 5.086726365917114e-46 ], [ 348.3125 , 5.087977880687688e-46 ], [ 348.375 , 5.089229703375238e-46 ], [ 348.4375 , 5.090481834055488e-46 ], 
                    [ 348.5 , 5.0917342728041786e-46 ], [ 348.5625 , 5.09298701969714e-46 ], [ 348.625 , 5.0942400748102614e-46 ], [ 348.6875 , 5.095493438219229e-46 ], 
                    [ 348.75 , 5.096747110000004e-46 ], [ 348.8125 , 5.097999451959769e-46 ], [ 348.875 , 5.099252101637482e-46 ], [ 348.9375 , 5.10050505910861e-46 ], 
                    [ 349.0 , 5.1017583244490355e-46 ], [ 349.0625 , 5.103011897734153e-46 ], [ 349.125 , 5.1042657790398466e-46 ], [ 349.1875 , 5.1055199684415835e-46 ], 
                    [ 349.25 , 5.1067744660152475e-46 ], [ 349.3125 , 5.108029271836417e-46 ], [ 349.375 , 5.109284385980978e-46 ], [ 349.4375 , 5.110539808524579e-46 ], 
                    [ 349.5 , 5.1117955395430726e-46 ], [ 349.5625 , 5.1130513553228515e-46 ], [ 349.625 , 5.114306057549963e-46 ], [ 349.6875 , 5.115561067671045e-46 ], 
                    [ 349.75 , 5.116816385761614e-46 ], [ 349.8125 , 5.118072011897354e-46 ], [ 349.875 , 5.1193279461537116e-46 ], [ 349.9375 , 5.120584188606404e-46 ], 
                    [ 350.0 , 5.121840739331025e-46 ], [ 350.0625 , 5.12309759840326e-46 ], [ 350.125 , 5.124354765898663e-46 ], [ 350.1875 , 5.125612241892994e-46 ], 
                    [ 350.25 , 5.1268700264620644e-46 ], [ 350.3125 , 5.12812811968134e-46 ], [ 350.375 , 5.129386662567995e-46 ], [ 350.4375 , 5.130645960749073e-46 ], 
                    [ 350.5 , 5.131905568096178e-46 ], [ 350.5625 , 5.13316548468514e-46 ], [ 350.625 , 5.134425710591915e-46 ], [ 350.6875 , 5.135686245892443e-46 ], 
                    [ 350.75 , 5.136947090662684e-46 ], [ 350.8125 , 5.13820824497854e-46 ], [ 350.875 , 5.1394697089161896e-46 ], [ 350.9375 , 5.1407314825515e-46 ], 
                    [ 351.0 , 5.141993565960579e-46 ], [ 351.0625 , 5.143255959219403e-46 ], [ 351.125 , 5.14451866240408e-46 ], [ 351.1875 , 5.145780454790086e-46 ], 
                    [ 351.25 , 5.147039863973847e-46 ], [ 351.3125 , 5.1482995813929684e-46 ], [ 351.375 , 5.149559607122926e-46 ], [ 351.4375 , 5.150819941239068e-46 ], 
                    [ 351.5 , 5.152080583817016e-46 ], [ 351.5625 , 5.153341534932194e-46 ], [ 351.625 , 5.154602794660076e-46 ], [ 351.6875 , 5.1558643630761956e-46 ], 
                    [ 351.75 , 5.157126240256175e-46 ], [ 351.8125 , 5.158388426275547e-46 ], [ 351.875 , 5.159650921209973e-46 ], [ 351.9375 , 5.160913725134873e-46 ], 
                    [ 352.0 , 5.1621767311918235e-46 ], [ 352.0625 , 5.163439849434066e-46 ], [ 352.125 , 5.1647032767450575e-46 ], [ 352.1875 , 5.1659670132005326e-46 ], 
                    [ 352.25 , 5.167231058875989e-46 ], [ 352.3125 , 5.1684954138471626e-46 ], [ 352.375 , 5.169760078189771e-46 ], [ 352.4375 , 5.171025051979475e-46 ], 
                    [ 352.5 , 5.17229033529203e-46 ], [ 352.5625 , 5.173555928202988e-46 ], [ 352.625 , 5.174821830788326e-46 ], [ 352.6875 , 5.176088043123815e-46 ], 
                    [ 352.75 , 5.1773545652851026e-46 ], [ 352.8125 , 5.1786208604044476e-46 ], [ 352.875 , 5.179886510438719e-46 ], [ 352.9375 , 5.181152469796623e-46 ], 
                    [ 353.0 , 5.182418738553796e-46 ], [ 353.0625 , 5.183685316785927e-46 ], [ 353.125 , 5.184952204568505e-46 ], [ 353.1875 , 5.186219401977258e-46 ], 
                    [ 353.25 , 5.187486909087967e-46 ], [ 353.3125 , 5.1887547259760676e-46 ], [ 353.375 , 5.1900228527174876e-46 ], [ 353.4375 , 5.1912912893878454e-46 ], 
                    [ 353.5 , 5.192560036062999e-46 ], [ 353.5625 , 5.193829092818527e-46 ], [ 353.625 , 5.195097013651943e-46 ], [ 353.6875 , 5.196362581263123e-46 ], 
                    [ 353.75 , 5.197628457176762e-46 ], [ 353.8125 , 5.198894641468077e-46 ], [ 353.875 , 5.200161134212117e-46 ], [ 353.9375 , 5.201427935483986e-46 ], 
                    [ 354.0 , 5.202695045358844e-46 ], [ 354.0625 , 5.203962463911944e-46 ], [ 354.125 , 5.205230191218407e-46 ], [ 354.1875 , 5.206498227353523e-46 ], 
                    [ 354.25 , 5.207766572392488e-46 ], [ 354.3125 , 5.209035226410589e-46 ], [ 354.375 , 5.210304189482951e-46 ], [ 354.4375 , 5.211574082513536e-46 ], 
                    [ 354.5 , 5.212845761976699e-46 ], [ 354.5625 , 5.2141177517431666e-46 ], [ 354.625 , 5.215390051888655e-46 ], [ 354.6875 , 5.2166626624889e-46 ], 
                    [ 354.75 , 5.217935583619693e-46 ], [ 354.8125 , 5.219208815356771e-46 ], [ 354.875 , 5.220482357775887e-46 ], [ 354.9375 , 5.221756210952852e-46 ], 
                    [ 355.0 , 5.223030374963641e-46 ], [ 355.0625 , 5.224304849883841e-46 ], [ 355.125 , 5.225579635789579e-46 ], [ 355.1875 , 5.22685473275655e-46 ], 
                    [ 355.25 , 5.2281297576863534e-46 ], [ 355.3125 , 5.2294037658927185e-46 ], [ 355.375 , 5.230678084553717e-46 ], [ 355.4375 , 5.23195271374489e-46 ], 
                    [ 355.5 , 5.233227653542018e-46 ], [ 355.5625 , 5.23450290402083e-46 ], [ 355.625 , 5.235778465256847e-46 ], [ 355.6875 , 5.237054337325945e-46 ], 
                    [ 355.75 , 5.23833052030383e-46 ], [ 355.8125 , 5.23960701426634e-46 ], [ 355.875 , 5.24088381928911e-46 ], [ 355.9375 , 5.242160935448014e-46 ], 
                    [ 356.0 , 5.243438362818868e-46 ], [ 356.0625 , 5.244716058905823e-46 ], [ 356.125 , 5.2459937540670306e-46 ], [ 356.1875 , 5.24727176049491e-46 ], 
                    [ 356.25 , 5.2485500782651435e-46 ], [ 356.3125 , 5.249828707453652e-46 ], [ 356.375 , 5.2511076481363384e-46 ], [ 356.4375 , 5.252386900389052e-46 ], 
                    [ 356.5 , 5.253666464287658e-46 ], [ 356.5625 , 5.254946339908116e-46 ], [ 356.625 , 5.2562265273264035e-46 ], [ 356.6875 , 5.257507026618479e-46 ], 
                    [ 356.75 , 5.258787837860246e-46 ], [ 356.8125 , 5.260068961127699e-46 ], [ 356.875 , 5.261350396496891e-46 ], [ 356.9375 , 5.2626290982608725e-46 ], 
                    [ 357.0 , 5.2639080612594274e-46 ], [ 357.0625 , 5.265187335081018e-46 ], [ 357.125 , 5.266466919801221e-46 ], [ 357.1875 , 5.2677468154955534e-46 ], 
                    [ 357.25 , 5.269027022239593e-46 ], [ 357.3125 , 5.270307540108932e-46 ], [ 357.375 , 5.2715883691791456e-46 ], [ 357.4375 , 5.272869509525939e-46 ], 
                    [ 357.5 , 5.274150961224923e-46 ], [ 357.5625 , 5.275432724351727e-46 ], [ 357.625 , 5.276714798982188e-46 ], [ 357.6875 , 5.277997185191709e-46 ], 
                    [ 357.75 , 5.279277771267782e-46 ], [ 357.8125 , 5.2805581917459145e-46 ], [ 357.875 , 5.281838922773419e-46 ], [ 357.9375 , 5.283119964425767e-46 ], 
                    [ 358.0 , 5.284401316778072e-46 ], [ 358.0625 , 5.2856829799058386e-46 ], [ 358.125 , 5.286964953884404e-46 ], [ 358.1875 , 5.288247238789162e-46 ], 
                    [ 358.25 , 5.289529834695523e-46 ], [ 358.3125 , 5.290812741678879e-46 ], [ 358.375 , 5.292095959814753e-46 ], [ 358.4375 , 5.293379489178574e-46 ], 
                    [ 358.5 , 5.294663329845863e-46 ], [ 358.5625 , 5.2959477786715776e-46 ], [ 358.625 , 5.2972327241461445e-46 ], [ 358.6875 , 5.298517981384413e-46 ], 
                    [ 358.75 , 5.299803550462212e-46 ], [ 358.8125 , 5.301089431455164e-46 ], [ 358.875 , 5.3023756244389135e-46 ], [ 358.9375 , 5.303662129489081e-46 ], 
                    [ 359.0 , 5.304948946681457e-46 ], [ 359.0625 , 5.306236076091854e-46 ], [ 359.125 , 5.307523517795948e-46 ], [ 359.1875 , 5.308811271869396e-46 ], 
                    [ 359.25 , 5.3100993383881776e-46 ], [ 359.3125 , 5.311387717428101e-46 ], [ 359.375 , 5.3126754821218616e-46 ], [ 359.4375 , 5.313962071707646e-46 ], 
                    [ 359.5 , 5.315248972871499e-46 ], [ 359.5625 , 5.316536185688687e-46 ], [ 359.625 , 5.3178237102348735e-46 ], [ 359.6875 , 5.319111546585475e-46 ], 
                    [ 359.75 , 5.320399694816003e-46 ], [ 359.8125 , 5.321688155001948e-46 ], [ 359.875 , 5.322976927218896e-46 ], [ 359.9375 , 5.324266011542412e-46 ], 
                    [ 360.0 , 5.325555408048156e-46 ], [ 360.0625 , 5.326845116811617e-46 ], [ 360.125 , 5.328135137908376e-46 ], [ 360.1875 , 5.329425322448826e-46 ], 
                    [ 360.25 , 5.330714872804246e-46 ], [ 360.3125 , 5.332004735189616e-46 ], [ 360.375 , 5.3332949096803626e-46 ], [ 360.4375 , 5.334585396352119e-46 ], 
                    [ 360.5 , 5.3358761952802324e-46 ], [ 360.5625 , 5.33716730654045e-46 ], [ 360.625 , 5.338458730208231e-46 ], [ 360.6875 , 5.339750466359283e-46 ], 
                    [ 360.75 , 5.3410425150690646e-46 ], [ 360.8125 , 5.3423348764133185e-46 ], [ 360.875 , 5.343627550467578e-46 ], [ 360.9375 , 5.3449205373077e-46 ], 
                    [ 361.0 , 5.346213837009179e-46 ], [ 361.0625 , 5.3475065566291905e-46 ], [ 361.125 , 5.34879941869007e-46 ], [ 361.1875 , 5.350092593325136e-46 ], 
                    [ 361.25 , 5.351386080609922e-46 ], [ 361.3125 , 5.352679880620129e-46 ], [ 361.375 , 5.3539739934311005e-46 ], [ 361.4375 , 5.355268419118652e-46 ], 
                    [ 361.5 , 5.356563157758388e-46 ], [ 361.5625 , 5.357858209426009e-46 ], [ 361.625 , 5.359153574197082e-46 ], [ 361.6875 , 5.3604492521474576e-46 ], 
                    [ 361.75 , 5.3617452433527776e-46 ], [ 361.8125 , 5.3630415478887e-46 ], [ 361.875 , 5.364337573384822e-46 ], [ 361.9375 , 5.365633382337731e-46 ], 
                    [ 362.0 , 5.3669295043061084e-46 ], [ 362.0625 , 5.368225939365605e-46 ], [ 362.125 , 5.369522687591888e-46 ], [ 362.1875 , 5.370819749060569e-46 ], 
                    [ 362.25 , 5.3721171238472365e-46 ], [ 362.3125 , 5.37341481202773e-46 ], [ 362.375 , 5.374712813677638e-46 ], [ 362.4375 , 5.376011128872721e-46 ], 
                    [ 362.5 , 5.377309757688716e-46 ], [ 362.5625 , 5.378608700201347e-46 ], [ 362.625 , 5.379907956486464e-46 ], [ 362.6875 , 5.381206887966726e-46 ], 
                    [ 362.75 , 5.382502969446104e-46 ], [ 362.8125 , 5.383799363091031e-46 ], [ 362.875 , 5.385096068976618e-46 ], [ 362.9375 , 5.3863930871781834e-46 ], 
                    [ 363.0 , 5.387690417770873e-46 ], [ 363.0625 , 5.3889880608298514e-46 ], [ 363.125 , 5.390286016430528e-46 ], [ 363.1875 , 5.391584284648068e-46 ], 
                    [ 363.25 , 5.392882865557802e-46 ], [ 363.3125 , 5.394181759235005e-46 ], [ 363.375 , 5.3954809657550854e-46 ], [ 363.4375 , 5.396780485193316e-46 ], 
                    [ 363.5 , 5.398080317625216e-46 ], [ 363.5625 , 5.399381206379324e-46 ], [ 363.625 , 5.4006826038808194e-46 ], [ 363.6875 , 5.401984315054455e-46 ], 
                    [ 363.75 , 5.403286339975834e-46 ], [ 363.8125 , 5.404588678720578e-46 ], [ 363.875 , 5.4058913313643654e-46 ], [ 363.9375 , 5.407194297982701e-46 ], 
                    [ 364.0 , 5.408497578651529e-46 ], [ 364.0625 , 5.409801173446315e-46 ], [ 364.125 , 5.411105082442811e-46 ], [ 364.1875 , 5.4124093057168224e-46 ], 
                    [ 364.25 , 5.413713843344062e-46 ], [ 364.3125 , 5.415018695400335e-46 ], [ 364.375 , 5.4163228957823156e-46 ], [ 364.4375 , 5.4176258599940405e-46 ], 
                    [ 364.5 , 5.418929137650218e-46 ], [ 364.5625 , 5.4202327288259806e-46 ], [ 364.625 , 5.42153663359702e-46 ], [ 364.6875 , 5.422840852038582e-46 ], 
                    [ 364.75 , 5.424145384226201e-46 ], [ 364.8125 , 5.425450230235508e-46 ], [ 364.875 , 5.426755390141687e-46 ], [ 364.9375 , 5.4280608640204426e-46 ], 
                    [ 365.0 , 5.4293666519472666e-46 ], [ 365.0625 , 5.43067275399767e-46 ], [ 365.125 , 5.431979170247293e-46 ], [ 365.1875 , 5.433285900771723e-46 ], 
                    [ 365.25 , 5.4345906650346285e-46 ], [ 365.3125 , 5.435895565640782e-46 ], [ 365.375 , 5.4372007795667924e-46 ], [ 365.4375 , 5.438506306888046e-46 ], 
                    [ 365.5 , 5.439812147679714e-46 ], [ 365.5625 , 5.4411183020170274e-46 ], [ 365.625 , 5.442424769975347e-46 ], [ 365.6875 , 5.443731551629939e-46 ], 
                    [ 365.75 , 5.445038647056046e-46 ], [ 365.8125 , 5.446346056329126e-46 ], [ 365.875 , 5.4476537795245744e-46 ], [ 365.9375 , 5.44896181671765e-46 ], 
                    [ 366.0 , 5.450270167983711e-46 ], [ 366.0625 , 5.4515786273228375e-46 ], [ 366.125 , 5.452887154883424e-46 ], [ 366.1875 , 5.4541959965264294e-46 ], 
                    [ 366.25 , 5.455505152327011e-46 ], [ 366.3125 , 5.456814622360771e-46 ], [ 366.375 , 5.4581244067029375e-46 ], [ 366.4375 , 5.459434505429185e-46 ], 
                    [ 366.5 , 5.46074491861486e-46 ], [ 366.5625 , 5.462055646335403e-46 ], [ 366.625 , 5.4633666886663475e-46 ], [ 366.6875 , 5.464678045683132e-46 ], 
                    [ 366.75 , 5.4659897174615234e-46 ], [ 366.8125 , 5.4673017040768006e-46 ], [ 366.875 , 5.468614005604727e-46 ], [ 366.9375 , 5.469924884641423e-46 ], 
                    [ 367.0 , 5.47123599028343e-46 ], [ 367.0625 , 5.472547410188908e-46 ], [ 367.125 , 5.47385914443334e-46 ], [ 367.1875 , 5.475171193091953e-46 ], 
                    [ 367.25 , 5.47648355624015e-46 ], [ 367.3125 , 5.4777962339533104e-46 ], [ 367.375 , 5.479109226306875e-46 ], [ 367.4375 , 5.48042253337622e-46 ], 
                    [ 367.5 , 5.481736155236742e-46 ], [ 367.5625 , 5.483050091963896e-46 ], [ 367.625 , 5.484364343633269e-46 ], [ 367.6875 , 5.485678910320233e-46 ], 
                    [ 367.75 , 5.486993939892429e-46 ], [ 367.8125 , 5.48830949223122e-46 ], [ 367.875 , 5.489625359984643e-46 ], [ 367.9375 , 5.490941543228163e-46 ], 
                    [ 368.0 , 5.4922580420376586e-46 ], [ 368.0625 , 5.493574856488631e-46 ], [ 368.125 , 5.494891986656798e-46 ], [ 368.1875 , 5.496209432617854e-46 ], 
                    [ 368.25 , 5.497527194447553e-46 ], [ 368.3125 , 5.4988452722214714e-46 ], [ 368.375 , 5.500163666015593e-46 ], [ 368.4375 , 5.501482375905569e-46 ], 
                    [ 368.5 , 5.5028014019670695e-46 ], [ 368.5625 , 5.504120744276134e-46 ], [ 368.625 , 5.50543812957685e-46 ], [ 368.6875 , 5.506755422606101e-46 ], 
                    [ 368.75 , 5.5080730308256545e-46 ], [ 368.8125 , 5.5093909543110044e-46 ], [ 368.875 , 5.5107091931375855e-46 ], [ 368.9375 , 5.512027747380693e-46 ], 
                    [ 369.0 , 5.513346617116032e-46 ], [ 369.0625 , 5.514665802418894e-46 ], [ 369.125 , 5.515985303364943e-46 ], [ 369.1875 , 5.517305120029586e-46 ], 
                    [ 369.25 , 5.518625252488364e-46 ], [ 369.3125 , 5.519945700816916e-46 ], [ 369.375 , 5.5212664650908226e-46 ], [ 369.4375 , 5.5225867491755605e-46 ], 
                    [ 369.5 , 5.523904934085229e-46 ], [ 369.5625 , 5.5252234336322184e-46 ], [ 369.625 , 5.52654224789143e-46 ], [ 369.6875 , 5.527861376938221e-46 ], 
                    [ 369.75 , 5.52918082084745e-46 ], [ 369.8125 , 5.530500579694549e-46 ], [ 369.875 , 5.531820653554491e-46 ], [ 369.9375 , 5.533141042502666e-46 ], 
                    [ 370.0 , 5.534461746614006e-46 ], [ 370.0625 , 5.535782765964052e-46 ], [ 370.125 , 5.537104100627811e-46 ], [ 370.1875 , 5.538425750680585e-46 ], 
                    [ 370.25 , 5.539747716197773e-46 ], [ 370.3125 , 5.541069691215498e-46 ], [ 370.375 , 5.542391790878948e-46 ], [ 370.4375 , 5.54371420599548e-46 ], 
                    [ 370.5 , 5.545036936640363e-46 ], [ 370.5625 , 5.546359982888921e-46 ], [ 370.625 , 5.547683344816339e-46 ], [ 370.6875 , 5.549007022498057e-46 ], 
                    [ 370.75 , 5.550331016009373e-46 ], [ 370.8125 , 5.551655325425605e-46 ], [ 370.875 , 5.552979950822287e-46 ], [ 370.9375 , 5.554304892274613e-46 ], 
                    [ 371.0 , 5.555630149858074e-46 ], [ 371.0625 , 5.5569557236480985e-46 ], [ 371.125 , 5.558281613720134e-46 ], [ 371.1875 , 5.5596075167141226e-46 ], 
                    [ 371.25 , 5.56093372069341e-46 ], [ 371.3125 , 5.562260241028982e-46 ], [ 371.375 , 5.563587077796381e-46 ], [ 371.4375 , 5.564914231071051e-46 ], 
                    [ 371.5 , 5.566241700928452e-46 ], [ 371.5625 , 5.567569487444144e-46 ], [ 371.625 , 5.5688975906936605e-46 ], [ 371.6875 , 5.570226010752639e-46 ], 
                    [ 371.75 , 5.571554747696529e-46 ], [ 371.8125 , 5.5728838016009244e-46 ], [ 371.875 , 5.574213172541512e-46 ], [ 371.9375 , 5.575542860593878e-46 ], 
                    [ 372.0 , 5.5768719332249544e-46 ], [ 372.0625 , 5.578198825980619e-46 ], [ 372.125 , 5.5795260344408894e-46 ], [ 372.1875 , 5.580853558680923e-46 ], 
                    [ 372.25 , 5.582181398775851e-46 ], [ 372.3125 , 5.583509554800784e-46 ], [ 372.375 , 5.584838026831051e-46 ], [ 372.4375 , 5.586166814941559e-46 ], 
                    [ 372.5 , 5.58749591920783e-46 ], [ 372.5625 , 5.588825339704769e-46 ], [ 372.625 , 5.5901550765078955e-46 ], [ 372.6875 , 5.591485129692266e-46 ], 
                    [ 372.75 , 5.592815499333317e-46 ], [ 372.8125 , 5.594146185506183e-46 ], [ 372.875 , 5.595476902112096e-46 ], [ 372.9375 , 5.5968076953260145e-46 ], 
                    [ 373.0 , 5.598138805047385e-46 ], [ 373.0625 , 5.5994702313516414e-46 ], [ 373.125 , 5.600801974313959e-46 ], [ 373.1875 , 5.602134034009651e-46 ], 
                    [ 373.25 , 5.603466410514204e-46 ], [ 373.3125 , 5.6047991039027694e-46 ], [ 373.375 , 5.60613211425075e-46 ], [ 373.4375 , 5.6074654416336514e-46 ], 
                    [ 373.5 , 5.608799086126715e-46 ], [ 373.5625 , 5.6101330478055585e-46 ], [ 373.625 , 5.611467326745382e-46 ], [ 373.6875 , 5.612801923021799e-46 ], 
                    [ 373.75 , 5.614135350348448e-46 ], [ 373.8125 , 5.615468684850901e-46 ], [ 373.875 , 5.616802336014962e-46 ], [ 373.9375 , 5.6181363039157936e-46 ], 
                    [ 374.0 , 5.619470588628583e-46 ], [ 374.0625 , 5.62080519022861e-46 ], [ 374.125 , 5.6221401087911345e-46 ], [ 374.1875 , 5.623475344391473e-46 ], 
                    [ 374.25 , 5.624810897104841e-46 ], [ 374.3125 , 5.626146767006632e-46 ], [ 374.375 , 5.627482954172057e-46 ], [ 374.4375 , 5.628819458676625e-46 ], 
                    [ 374.5 , 5.630156280595581e-46 ], [ 374.5625 , 5.6314934200043514e-46 ], [ 374.625 , 5.632830524420443e-46 ], [ 374.6875 , 5.6341679434657736e-46 ], 
                    [ 374.75 , 5.635505680058369e-46 ], [ 374.8125 , 5.6368437342735465e-46 ], [ 374.875 , 5.638182106186759e-46 ], [ 374.9375 , 5.639520795873479e-46 ], 
                    [ 375.0 , 5.640859803409116e-46 ], [ 375.0625 , 5.642199128869058e-46 ], [ 375.125 , 5.6435387723289105e-46 ], [ 375.1875 , 5.644878733864177e-46 ], 
                    [ 375.25 , 5.64621901355034e-46 ], [ 375.3125 , 5.6475596114629386e-46 ], [ 375.375 , 5.648900527677489e-46 ], [ 375.4375 , 5.650241523542397e-46 ], 
                    [ 375.5 , 5.651581778792315e-46 ], [ 375.5625 , 5.6529223519550775e-46 ], [ 375.625 , 5.6542632431060964e-46 ], [ 375.6875 , 5.655604452320637e-46 ], 
                    [ 375.75 , 5.656945979674388e-46 ], [ 375.8125 , 5.6582878252426115e-46 ], [ 375.875 , 5.65962998910099e-46 ], [ 375.9375 , 5.66097247132478e-46 ], 
                    [ 376.0 , 5.662315271989662e-46 ], [ 376.0625 , 5.66365839117113e-46 ], [ 376.125 , 5.6650018289446974e-46 ], [ 376.1875 , 5.6663455853860554e-46 ], 
                    [ 376.25 , 5.6676896605706324e-46 ], [ 376.3125 , 5.66903315607031e-46 ], [ 376.375 , 5.670375128765265e-46 ], [ 376.4375 , 5.6717174191318e-46 ], 
                    [ 376.5 , 5.6730600272451124e-46 ], [ 376.5625 , 5.67440295318042e-46 ], [ 376.625 , 5.6757461970129576e-46 ], [ 376.6875 , 5.677089758817897e-46 ], 
                    [ 376.75 , 5.6784336386706695e-46 ], [ 376.8125 , 5.679777836646443e-46 ], [ 376.875 , 5.681122352820645e-46 ], [ 376.9375 , 5.6824671872684766e-46 ], 
                    [ 377.0 , 5.6838123400653626e-46 ], [ 377.0625 , 5.685157811286539e-46 ], [ 377.125 , 5.686503601007465e-46 ], [ 377.1875 , 5.687848263818704e-46 ], 
                    [ 377.25 , 5.689191404454006e-46 ], [ 377.3125 , 5.690534862261406e-46 ], [ 377.375 , 5.691878637315761e-46 ], [ 377.4375 , 5.693222729692107e-46 ], 
                    [ 377.5 , 5.694567139465217e-46 ], [ 377.5625 , 5.695911866710162e-46 ], [ 377.625 , 5.697256911501911e-46 ], [ 377.6875 , 5.698602273915368e-46 ], 
                    [ 377.75 , 5.699947954025579e-46 ], [ 377.8125 , 5.701293951907525e-46 ], [ 377.875 , 5.702640267636285e-46 ], [ 377.9375 , 5.703986901286997e-46 ], 
                    [ 378.0 , 5.705333852934654e-46 ], [ 378.0625 , 5.706681626430877e-46 ], [ 378.125 , 5.708030183446767e-46 ], [ 378.1875 , 5.709379059142865e-46 ], 
                    [ 378.25 , 5.7107282535943165e-46 ], [ 378.3125 , 5.7120777668765694e-46 ], [ 378.375 , 5.7134275990649264e-46 ], [ 378.4375 , 5.714777750234748e-46 ], 
                    [ 378.5 , 5.716128220461374e-46 ], [ 378.5625 , 5.717479009820323e-46 ], [ 378.625 , 5.718830118386806e-46 ], [ 378.6875 , 5.720181546236459e-46 ], 
                    [ 378.75 , 5.721533293444652e-46 ], [ 378.8125 , 5.722885360086893e-46 ], [ 378.875 , 5.724237746238504e-46 ], [ 378.9375 , 5.72558920054787e-46 ], 
                    [ 379.0 , 5.726940082288482e-46 ], [ 379.0625 , 5.728291282752924e-46 ], [ 379.125 , 5.729642802016351e-46 ], [ 379.1875 , 5.7309946401539824e-46 ], 
                    [ 379.25 , 5.732346797241052e-46 ], [ 379.3125 , 5.733699273352769e-46 ], [ 379.375 , 5.73505206856457e-46 ], [ 379.4375 , 5.736405182951414e-46 ], 
                    [ 379.5 , 5.7377586165889325e-46 ], [ 379.5625 , 5.739112369552369e-46 ], [ 379.625 , 5.740466441916941e-46 ], [ 379.6875 , 5.74182083375817e-46 ], 
                    [ 379.75 , 5.743175545151311e-46 ], [ 379.8125 , 5.744528588824527e-46 ], [ 379.875 , 5.74588062605345e-46 ], [ 379.9375 , 5.747232981499108e-46 ], 
                    [ 380.0 , 5.748585655236277e-46 ], [ 380.0625 , 5.749938647339867e-46 ], [ 380.125 , 5.751291957884934e-46 ], [ 380.1875 , 5.752645586946262e-46 ], 
                    [ 380.25 , 5.75399953459894e-46 ], [ 380.3125 , 5.755353800917912e-46 ], [ 380.375 , 5.756708385978179e-46 ], [ 380.4375 , 5.75806328985476e-46 ], 
                    [ 380.5 , 5.7594185126226095e-46 ], [ 380.5625 , 5.760774054357029e-46 ], [ 380.625 , 5.762129915132845e-46 ], [ 380.6875 , 5.763485740991525e-46 ], 
                    [ 380.75 , 5.7648416336275864e-46 ], [ 380.8125 , 5.766197845245024e-46 ], [ 380.875 , 5.767554375919e-46 ], [ 380.9375 , 5.768911225724412e-46 ], 
                    [ 381.0 , 5.77026839473646e-46 ], [ 381.0625 , 5.771625883030158e-46 ], [ 381.125 , 5.772983690680741e-46 ], [ 381.1875 , 5.774341817763216e-46 ], 
                    [ 381.25 , 5.775700264352732e-46 ], [ 381.3125 , 5.777059030524577e-46 ], [ 381.375 , 5.778418116353772e-46 ], [ 381.4375 , 5.779777521915559e-46 ], 
                    [ 381.5 , 5.7811372472853186e-46 ], [ 381.5625 , 5.782497063273552e-46 ], [ 381.625 , 5.783857000597698e-46 ], [ 381.6875 , 5.785217257754135e-46 ], 
                    [ 381.75 , 5.786577834818165e-46 ], [ 381.8125 , 5.787938731864982e-46 ], [ 381.875 , 5.7892999489698015e-46 ], [ 381.9375 , 5.790661486207934e-46 ], 
                    [ 382.0 , 5.7920233436546705e-46 ], [ 382.0625 , 5.793385521385316e-46 ], [ 382.125 , 5.7947480194752396e-46 ], [ 382.1875 , 5.7961108379997e-46 ], 
                    [ 382.25 , 5.797473977034056e-46 ], [ 382.3125 , 5.798837436653853e-46 ], [ 382.375 , 5.800201216934321e-46 ], [ 382.4375 , 5.8015636516525e-46 ], 
                    [ 382.5 , 5.802924541966739e-46 ], [ 382.5625 , 5.804285751509132e-46 ], [ 382.625 , 5.805647280354601e-46 ], [ 382.6875 , 5.807009128578006e-46 ], 
                    [ 382.75 , 5.808371296254221e-46 ], [ 382.8125 , 5.8097337834582644e-46 ], [ 382.875 , 5.811096590265007e-46 ], [ 382.9375 , 5.812459716749501e-46 ], 
                    [ 383.0 , 5.8138231629867345e-46 ], [ 383.0625 , 5.815186929051547e-46 ], [ 383.125 , 5.8165510150191686e-46 ], [ 383.1875 , 5.817915420964597e-46 ], 
                    [ 383.25 , 5.81928014696277e-46 ], [ 383.3125 , 5.820644964597522e-46 ], [ 383.375 , 5.822009696025991e-46 ], [ 383.4375 , 5.823374747434745e-46 ], 
                    [ 383.5 , 5.824740118898889e-46 ], [ 383.5625 , 5.826105810493424e-46 ], [ 383.625 , 5.827471822293369e-46 ], [ 383.6875 , 5.828838154373881e-46 ], 
                    [ 383.75 , 5.8302048068100564e-46 ], [ 383.8125 , 5.831571779676839e-46 ], [ 383.875 , 5.832939073049608e-46 ], [ 383.9375 , 5.834306687003388e-46 ], 
                    [ 384.0 , 5.835674621613216e-46 ], [ 384.0625 , 5.837042876954444e-46 ], [ 384.125 , 5.838411453102268e-46 ], [ 384.1875 , 5.839779918033868e-46 ], 
                    [ 384.25 , 5.8411471349970414e-46 ], [ 384.3125 , 5.842514672054808e-46 ], [ 384.375 , 5.843882529282152e-46 ], [ 384.4375 , 5.8452507067540725e-46 ], 
                    [ 384.5 , 5.84661920454542e-46 ], [ 384.5625 , 5.847988022731355e-46 ], [ 384.625 , 5.849357161386808e-46 ], [ 384.6875 , 5.850726620586762e-46 ], 
                    [ 384.75 , 5.852096400406266e-46 ], [ 384.8125 , 5.853466500920425e-46 ], [ 384.875 , 5.854836922204487e-46 ], [ 384.9375 , 5.8562076643331755e-46 ], 
                    [ 385.0 , 5.857578727381983e-46 ], [ 385.0625 , 5.85895003112852e-46 ], [ 385.125 , 5.860319728322672e-46 ], [ 385.1875 , 5.8616897457228565e-46 ], 
                    [ 385.25 , 5.863060083403679e-46 ], [ 385.3125 , 5.864430741440181e-46 ], [ 385.375 , 5.865801719907256e-46 ], [ 385.4375 , 5.867173018879687e-46 ], 
                    [ 385.5 , 5.868544638432528e-46 ], [ 385.5625 , 5.869916578640723e-46 ], [ 385.625 , 5.87128883957911e-46 ], [ 385.6875 , 5.872661421322835e-46 ], 
                    [ 385.75 , 5.874034323946728e-46 ], [ 385.8125 , 5.8754075475259294e-46 ], [ 385.875 , 5.876781092135473e-46 ], [ 385.9375 , 5.878154957850283e-46 ], 
                    [ 386.0 , 5.879528593037755e-46 ], [ 386.0625 , 5.8809024377941985e-46 ], [ 386.125 , 5.882276603571219e-46 ], [ 386.1875 , 5.8836510904437445e-46 ], 
                    [ 386.25 , 5.885025898486887e-46 ], [ 386.3125 , 5.88640102777561e-46 ], [ 386.375 , 5.887776478385061e-46 ], [ 386.4375 , 5.88915225039028e-46 ], 
                    [ 386.5 , 5.890528343866325e-46 ], [ 386.5625 , 5.891904758888354e-46 ], [ 386.625 , 5.893281495531584e-46 ], [ 386.6875 , 5.894658553871e-46 ], 
                    [ 386.75 , 5.896035933981857e-46 ], [ 386.8125 , 5.897413635939341e-46 ], [ 386.875 , 5.898790923603335e-46 ], [ 386.9375 , 5.900168025410319e-46 ], 
                    [ 387.0 , 5.901545448708506e-46 ], [ 387.0625 , 5.902923193572949e-46 ], [ 387.125 , 5.904301260078678e-46 ], [ 387.1875 , 5.9056796483009056e-46 ], 
                    [ 387.25 , 5.907058358314572e-46 ], [ 387.3125 , 5.908437390194886e-46 ], [ 387.375 , 5.909816744017028e-46 ], [ 387.4375 , 5.911196419856202e-46 ], 
                    [ 387.5 , 5.9125764177873285e-46 ], [ 387.5625 , 5.913956737885857e-46 ], [ 387.625 , 5.915337380226914e-46 ], [ 387.6875 , 5.9167183448856865e-46 ], 
                    [ 387.75 , 5.918098727927686e-46 ], [ 387.8125 , 5.919477511547347e-46 ], [ 387.875 , 5.920856616392473e-46 ], [ 387.9375 , 5.922236042537946e-46 ], 
                    [ 388.0 , 5.923615790058663e-46 ], [ 388.0625 , 5.9249958590294124e-46 ], [ 388.125 , 5.926376249525211e-46 ], [ 388.1875 , 5.9277569616207166e-46 ], 
                    [ 388.25 , 5.929137995391147e-46 ], [ 388.3125 , 5.9305193509113205e-46 ], [ 388.375 , 5.931901028256111e-46 ], [ 388.4375 , 5.9332830275006685e-46 ], 
                    [ 388.5 , 5.934665348719942e-46 ], [ 388.5625 , 5.93604799198878e-46 ], [ 388.625 , 5.93743094632419e-46 ], [ 388.6875 , 5.938813542613945e-46 ], 
                    [ 388.75 , 5.940196460856456e-46 ], [ 388.8125 , 5.941579701126775e-46 ], [ 388.875 , 5.942963263499807e-46 ], [ 388.9375 , 5.944347148050643e-46 ], 
                    [ 389.0 , 5.94573135485426e-46 ], [ 389.0625 , 5.947115883985701e-46 ], [ 389.125 , 5.9485007355200224e-46 ], [ 389.1875 , 5.949885909532299e-46 ], 
                    [ 389.25 , 5.9512714060976235e-46 ], [ 389.3125 , 5.9526572252911495e-46 ], [ 389.375 , 5.954043367187876e-46 ], [ 389.4375 , 5.955429831863033e-46 ], 
                    [ 389.5 , 5.956816619391826e-46 ], [ 389.5625 , 5.958200366636039e-46 ], [ 389.625 , 5.959582852264291e-46 ], [ 389.6875 , 5.960965658671577e-46 ], 
                    [ 389.75 , 5.962348785932539e-46 ], [ 389.8125 , 5.963732234121538e-46 ], [ 389.875 , 5.9651160033129584e-46 ], [ 389.9375 , 5.966500093581492e-46 ], 
                    [ 390.0 , 5.9678845050014696e-46 ], [ 390.0625 , 5.969269237647366e-46 ], [ 390.125 , 5.9706542915938005e-46 ], [ 390.1875 , 5.972039666915366e-46 ], 
                    [ 390.25 , 5.973425363686675e-46 ], [ 390.3125 , 5.9748113819821875e-46 ], [ 390.375 , 5.976197721876461e-46 ], [ 390.4375 , 5.97758463171351e-46 ], 
                    [ 390.5 , 5.978972454028863e-46 ], [ 390.5625 , 5.980360598556556e-46 ], [ 390.625 , 5.98174906537123e-46 ], [ 390.6875 , 5.983137854547752e-46 ], 
                    [ 390.75 , 5.984526966161008e-46 ], [ 390.8125 , 5.985916400285773e-46 ], [ 390.875 , 5.987306156996968e-46 ], [ 390.9375 , 5.988696236369445e-46 ], 
                    [ 391.0 , 5.990086638478202e-46 ], [ 391.0625 , 5.991477363398168e-46 ], [ 391.125 , 5.992868411204208e-46 ], [ 391.1875 , 5.9942597819713266e-46 ], 
                    [ 391.25 , 5.995651475774466e-46 ], [ 391.3125 , 5.997043492688794e-46 ], [ 391.375 , 5.998435882933423e-46 ], [ 391.4375 , 5.999828602284185e-46 ], 
                    [ 391.5 , 6.0012216449970236e-46 ], [ 391.5625 , 6.0026150111472296e-46 ], [ 391.625 , 6.004008700809728e-46 ], [ 391.6875 , 6.005402714059631e-46 ], 
                    [ 391.75 , 6.006797050972157e-46 ], [ 391.8125 , 6.008191711622368e-46 ], [ 391.875 , 6.009586696085556e-46 ], [ 391.9375 , 6.0109820044367344e-46 ], 
                    [ 392.0 , 6.01237763675119e-46 ], [ 392.0625 , 6.013773593104224e-46 ], [ 392.125 , 6.015169873570903e-46 ], [ 392.1875 , 6.0165664782265644e-46 ], 
                    [ 392.25 , 6.017961363639218e-46 ], [ 392.3125 , 6.019354211293494e-46 ], [ 392.375 , 6.0207473813202455e-46 ], [ 392.4375 , 6.022140873793872e-46 ], 
                    [ 392.5 , 6.02353468878913e-46 ], [ 392.5625 , 6.024928826380754e-46 ], [ 392.625 , 6.026323286643279e-46 ], [ 392.6875 , 6.0277180696513864e-46 ], 
                    [ 392.75 , 6.029113175479948e-46 ], [ 392.8125 , 6.030508604203421e-46 ], [ 392.875 , 6.031904355896755e-46 ], [ 392.9375 , 6.033300430634528e-46 ], 
                    [ 393.0 , 6.03469682849168e-46 ], [ 393.0625 , 6.036093549542868e-46 ], [ 393.125 , 6.0374905938629384e-46 ], [ 393.1875 , 6.038887425942516e-46 ], 
                    [ 393.25 , 6.040284572482004e-46 ], [ 393.3125 , 6.041682042262981e-46 ], [ 393.375 , 6.043079835360104e-46 ], [ 393.4375 , 6.044477951848176e-46 ], 
                    [ 393.5 , 6.0458763918021444e-46 ], [ 393.5625 , 6.047275155296715e-46 ], [ 393.625 , 6.048674242406829e-46 ], [ 393.6875 , 6.050073653207315e-46 ], 
                    [ 393.75 , 6.0514733877729325e-46 ], [ 393.8125 , 6.052873446178844e-46 ], [ 393.875 , 6.0542738284997156e-46 ], [ 393.9375 , 6.055674534810703e-46 ], 
                    [ 394.0 , 6.057075565186719e-46 ], [ 394.0625 , 6.058476178312392e-46 ], [ 394.125 , 6.059876337278141e-46 ], [ 394.1875 , 6.061276819831053e-46 ], 
                    [ 394.25 , 6.062677626045959e-46 ], [ 394.3125 , 6.064078755997568e-46 ], [ 394.375 , 6.0654802097607455e-46 ], [ 394.4375 , 6.066881987410411e-46 ], 
                    [ 394.5 , 6.068284089021201e-46 ], [ 394.5625 , 6.06968651466816e-46 ], [ 394.625 , 6.071089264426259e-46 ], [ 394.6875 , 6.0724923383701024e-46 ], 
                    [ 394.75 , 6.07389573657487e-46 ], [ 394.8125 , 6.075299459115414e-46 ], [ 394.875 , 6.076703506066605e-46 ], [ 394.9375 , 6.078107877503591e-46 ], 
                    [ 395.0 , 6.079508582794128e-46 ], [ 395.0625 , 6.080909374087026e-46 ], [ 395.125 , 6.08231048813899e-46 ], [ 395.1875 , 6.0837119250243005e-46 ], 
                    [ 395.25 , 6.0851136848174285e-46 ], [ 395.3125 , 6.08651576759269e-46 ], [ 395.375 , 6.087918173424633e-46 ], [ 395.4375 , 6.089320902387477e-46 ], 
                    [ 395.5 , 6.0907239545559794e-46 ], [ 395.5625 , 6.092127330004307e-46 ], [ 395.625 , 6.093531028807076e-46 ], [ 395.6875 , 6.0949350510388796e-46 ], 
                    [ 395.75 , 6.096339396774023e-46 ], [ 395.8125 , 6.097744066087261e-46 ], [ 395.875 , 6.099149558142738e-46 ], [ 395.9375 , 6.100556231485426e-46 ], 
                    [ 396.0 , 6.101963229255208e-46 ], [ 396.0625 , 6.1033705515269924e-46 ], [ 396.125 , 6.104778198375668e-46 ], [ 396.1875 , 6.106186169876002e-46 ], 
                    [ 396.25 , 6.1075944661029616e-46 ], [ 396.3125 , 6.1090030871313925e-46 ], [ 396.375 , 6.1104120330361646e-46 ], [ 396.4375 , 6.111821303892334e-46 ], 
                    [ 396.5 , 6.11323089977476e-46 ], [ 396.5625 , 6.11464082075823e-46 ], [ 396.625 , 6.116051066918029e-46 ], [ 396.6875 , 6.117461638329066e-46 ], 
                    [ 396.75 , 6.118872535066223e-46 ], [ 396.8125 , 6.120281702086983e-46 ], [ 396.875 , 6.121690544502937e-46 ], [ 396.9375 , 6.123099711223725e-46 ], 
                    [ 397.0 , 6.124509202324048e-46 ], [ 397.0625 , 6.1259190178783965e-46 ], [ 397.125 , 6.127329157961721e-46 ], [ 397.1875 , 6.1287396226485505e-46 ], 
                    [ 397.25 , 6.130150412013651e-46 ], [ 397.3125 , 6.131561526131761e-46 ], [ 397.375 , 6.13297296507759e-46 ], [ 397.4375 , 6.134384728926002e-46 ], 
                    [ 397.5 , 6.135796817751742e-46 ], [ 397.5625 , 6.137209231629658e-46 ], [ 397.625 , 6.138621970634534e-46 ], [ 397.6875 , 6.140034796572412e-46 ], 
                    [ 397.75 , 6.1414464336222295e-46 ], [ 397.8125 , 6.142858395217345e-46 ], [ 397.875 , 6.144270681432289e-46 ], [ 397.9375 , 6.145683292341779e-46 ], 
                    [ 398.0 , 6.147096228020378e-46 ], [ 398.0625 , 6.1485094885428394e-46 ], [ 398.125 , 6.149923073983804e-46 ], [ 398.1875 , 6.151336984417929e-46 ], 
                    [ 398.25 , 6.152751219920019e-46 ], [ 398.3125 , 6.154165780564723e-46 ], [ 398.375 , 6.155580666426839e-46 ], [ 398.4375 , 6.156995877581177e-46 ], 
                    [ 398.5 , 6.158411414102526e-46 ], [ 398.5625 , 6.1598272760655604e-46 ], [ 398.625 , 6.161242247129928e-46 ], [ 398.6875 , 6.162656044085414e-46 ], 
                    [ 398.75 , 6.164070165459628e-46 ], [ 398.8125 , 6.165484611326793e-46 ], [ 398.875 , 6.1668993817615025e-46 ], [ 398.9375 , 6.168314476838319e-46 ], 
                    [ 399.0 , 6.169729896631608e-46 ], [ 399.0625 , 6.1711456412159236e-46 ], [ 399.125 , 6.172561710665794e-46 ], [ 399.1875 , 6.173978105055808e-46 ], 
                    [ 399.25 , 6.1753948244604874e-46 ], [ 399.3125 , 6.1768118689543225e-46 ], [ 399.375 , 6.1782292386119974e-46 ], [ 399.4375 , 6.179646933508214e-46 ], 
                    [ 399.5 , 6.181064953717518e-46 ], [ 399.5625 , 6.182483269568215e-46 ], [ 399.625 , 6.183901900630162e-46 ], [ 399.6875 , 6.18532085721075e-46 ], 
                    [ 399.75 , 6.1867401393848025e-46 ], [ 399.8125 , 6.188159747226988e-46 ], [ 399.875 , 6.189579680811989e-46 ], [ 399.9375 , 6.1909999402145945e-46 ], 
                    [ 400.0 , 6.192420525509654e-46 ], [ 400.0625 , 6.193841436771728e-46 ], [ 400.125 , 6.1952626740758765e-46 ], [ 400.1875 , 6.1966842374966925e-46 ], 
                    [ 400.25 , 6.198106127109094e-46 ], [ 400.3125 , 6.199528342987975e-46 ], [ 400.375 , 6.20095088520811e-46 ], [ 400.4375 , 6.2023737365184126e-46 ], 
                    [ 400.5 , 6.203794765382511e-46 ], [ 400.5625 , 6.205216119819137e-46 ], [ 400.625 , 6.206637799903013e-46 ], [ 400.6875 , 6.2080598057087915e-46 ], 
                    [ 400.75 , 6.209482137310925e-46 ], [ 400.8125 , 6.2109047947840995e-46 ], [ 400.875 , 6.212327778203152e-46 ], [ 400.9375 , 6.213751087642495e-46 ], 
                    [ 401.0 , 6.215174723177001e-46 ], [ 401.0625 , 6.2165986848814254e-46 ], [ 401.125 , 6.2180229728303195e-46 ], [ 401.1875 , 6.2194475870984745e-46 ], 
                    [ 401.25 , 6.220872527760786e-46 ], [ 401.3125 , 6.222297794891946e-46 ], [ 401.375 , 6.223722317245668e-46 ], [ 401.4375 , 6.225143772435232e-46 ], 
                    [ 401.5 , 6.226565552275213e-46 ], [ 401.5625 , 6.2279876568399744e-46 ], [ 401.625 , 6.22941008620351e-46 ], [ 401.6875 , 6.230832840440219e-46 ], 
                    [ 401.75 , 6.2322559196240355e-46 ], [ 401.8125 , 6.233679323829397e-46 ], [ 401.875 , 6.235103053130404e-46 ], [ 401.9375 , 6.236527107601393e-46 ], 
                    [ 402.0 , 6.2379514873164985e-46 ], [ 402.0625 , 6.2393761923501815e-46 ], [ 402.125 , 6.2408012227765644e-46 ], [ 402.1875 , 6.242226578670144e-46 ], 
                    [ 402.25 , 6.243652260105118e-46 ], [ 402.3125 , 6.245078237811615e-46 ], [ 402.375 , 6.246504501321939e-46 ], [ 402.4375 , 6.247931090565009e-46 ], 
                    [ 402.5 , 6.249358005615486e-46 ], [ 402.5625 , 6.250785246547686e-46 ], [ 402.625 , 6.252212813435993e-46 ], [ 402.6875 , 6.253640706354981e-46 ], 
                    [ 402.75 , 6.255068925379023e-46 ], [ 402.8125 , 6.2564974705825485e-46 ], [ 402.875 , 6.257926342040008e-46 ], [ 402.9375 , 6.2593555398261346e-46 ], 
                    [ 403.0 , 6.2607850640152325e-46 ], [ 403.0625 , 6.262214914682025e-46 ], [ 403.125 , 6.263645091900894e-46 ], [ 403.1875 , 6.265075595746599e-46 ], 
                    [ 403.25 , 6.266506052250727e-46 ], [ 403.3125 , 6.267936559960081e-46 ], [ 403.375 , 6.269367394223341e-46 ], [ 403.4375 , 6.27079855511514e-46 ], 
                    [ 403.5 , 6.272230042709953e-46 ], [ 403.5625 , 6.2736618570824035e-46 ], [ 403.625 , 6.275093998307176e-46 ], [ 403.6875 , 6.276526466458707e-46 ], 
                    [ 403.75 , 6.2779592616117584e-46 ], [ 403.8125 , 6.279392383840934e-46 ], [ 403.875 , 6.280825833220764e-46 ], [ 403.9375 , 6.282259609826333e-46 ], 
                    [ 404.0 , 6.283693713731892e-46 ], [ 404.0625 , 6.285128145012379e-46 ], [ 404.125 , 6.286562903742439e-46 ], [ 404.1875 , 6.2879962870243164e-46 ], 
                    [ 404.25 , 6.289429308128545e-46 ], [ 404.3125 , 6.290862655815308e-46 ], [ 404.375 , 6.2922963301589454e-46 ], [ 404.4375 , 6.293730331233898e-46 ], 
                    [ 404.5 , 6.295164659114721e-46 ], [ 404.5625 , 6.2965993138759334e-46 ], [ 404.625 , 6.298034295591854e-46 ], [ 404.6875 , 6.299469604337128e-46 ], 
                    [ 404.75 , 6.3009052401861514e-46 ], [ 404.8125 , 6.30234120321365e-46 ], [ 404.875 , 6.3037774934940066e-46 ], [ 404.9375 , 6.3052141111019366e-46 ], 
                    [ 405.0 , 6.306651056112082e-46 ], [ 405.0625 , 6.308088328598834e-46 ], [ 405.125 , 6.3095241625540455e-46 ], [ 405.1875 , 6.310959881709606e-46 ], 
                    [ 405.25 , 6.312395927560085e-46 ], [ 405.3125 , 6.313832300179825e-46 ], [ 405.375 , 6.315268999643179e-46 ], [ 405.4375 , 6.31670602602452e-46 ], 
                    [ 405.5 , 6.318143379398194e-46 ], [ 405.5625 , 6.31958105983874e-46 ], [ 405.625 , 6.321019067420494e-46 ], [ 405.6875 , 6.322457402217803e-46 ], 
                    [ 405.75 , 6.323896064305262e-46 ], [ 405.8125 , 6.32533505375721e-46 ], [ 405.875 , 6.3267743706483625e-46 ], [ 405.9375 , 6.3282140150531375e-46 ], 
                    [ 406.0 , 6.329653987045835e-46 ], [ 406.0625 , 6.331092227878423e-46 ], [ 406.125 , 6.332530449089274e-46 ], [ 406.1875 , 6.333968997017881e-46 ], 
                    [ 406.25 , 6.335407871738466e-46 ], [ 406.3125 , 6.336847073325264e-46 ], [ 406.375 , 6.338286601852529e-46 ], [ 406.4375 , 6.3397264573945305e-46 ], 
                    [ 406.5 , 6.341166640025512e-46 ], [ 406.5625 , 6.3426071498198215e-46 ], [ 406.625 , 6.3440479868518725e-46 ], [ 406.6875 , 6.345489151195864e-46 ], 
                    [ 406.75 , 6.346930642926155e-46 ], [ 406.8125 , 6.348372462117203e-46 ], [ 406.875 , 6.349814608843355e-46 ], [ 406.9375 , 6.351257083179105e-46 ], 
                    [ 407.0 , 6.352699292419903e-46 ], [ 407.0625 , 6.354141735820948e-46 ], [ 407.125 , 6.355584506743078e-46 ], [ 407.1875 , 6.357027605260615e-46 ], 
                    [ 407.25 , 6.3584710314480334e-46 ], [ 407.3125 , 6.359914785379688e-46 ], [ 407.375 , 6.3613588671299965e-46 ], [ 407.4375 , 6.362803276773348e-46 ], 
                    [ 407.5 , 6.364248014384331e-46 ], [ 407.5625 , 6.365693080037231e-46 ], [ 407.625 , 6.367138473806624e-46 ], [ 407.6875 , 6.3685841957670576e-46 ], 
                    [ 407.75 , 6.370030245992915e-46 ], [ 407.8125 , 6.37147662455896e-46 ], [ 407.875 , 6.372923331539516e-46 ], [ 407.9375 , 6.374367139800745e-46 ], 
                    [ 408.0 , 6.375810732143968e-46 ], [ 408.0625 , 6.377254651415242e-46 ], [ 408.125 , 6.37869889768847e-46 ], [ 408.1875 , 6.380143471037755e-46 ], 
                    [ 408.25 , 6.381588371537212e-46 ], [ 408.3125 , 6.383033599260797e-46 ], [ 408.375 , 6.3844791542827954e-46 ], [ 408.4375 , 6.385925036677194e-46 ], 
                    [ 408.5 , 6.387371246518269e-46 ], [ 408.5625 , 6.3888177838799495e-46 ], [ 408.625 , 6.39026464883659e-46 ], [ 408.6875 , 6.3917118414623816e-46 ], 
                    [ 408.75 , 6.393159361831438e-46 ], [ 408.8125 , 6.394607210018074e-46 ], [ 408.875 , 6.39605757071738e-46 ], [ 408.9375 , 6.397508806652161e-46 ], 
                    [ 409.0 , 6.398960371865811e-46 ], [ 409.0625 , 6.400412266432817e-46 ], [ 409.125 , 6.4018644904279524e-46 ], [ 409.1875 , 6.403317043926144e-46 ], 
                    [ 409.25 , 6.404769927001882e-46 ], [ 409.3125 , 6.406223139730219e-46 ], [ 409.375 , 6.407676682185678e-46 ], [ 409.4375 , 6.409130554443255e-46 ], 
                    [ 409.5 , 6.410584756577782e-46 ], [ 409.5625 , 6.412039288664104e-46 ], [ 409.625 , 6.413494150776907e-46 ], [ 409.6875 , 6.414949342991251e-46 ], 
                    [ 409.75 , 6.416404865382036e-46 ], [ 409.8125 , 6.417856431597746e-46 ], [ 409.875 , 6.419306724313928e-46 ], [ 409.9375 , 6.4207573447639855e-46 ], 
                    [ 410.0 , 6.422208293021979e-46 ], [ 410.0625 , 6.423659569162033e-46 ], [ 410.125 , 6.4251111732581015e-46 ], [ 410.1875 , 6.42656310538448e-46 ], 
                    [ 410.25 , 6.428015365615207e-46 ], [ 410.3125 , 6.429467954024286e-46 ], [ 410.375 , 6.4309208706861066e-46 ], [ 410.4375 , 6.432374115674757e-46 ], 
                    [ 410.5 , 6.433827689064477e-46 ], [ 410.5625 , 6.435281590929385e-46 ], [ 410.625 , 6.436735821343758e-46 ], [ 410.6875 , 6.438190380381838e-46 ], 
                    [ 410.75 , 6.439642964957285e-46 ], [ 410.8125 , 6.441094391997713e-46 ], [ 410.875 , 6.442546146174315e-46 ], [ 410.9375 , 6.443998227560869e-46 ], 
                    [ 411.0 , 6.445450636231171e-46 ], [ 411.0625 , 6.446903372259033e-46 ], [ 411.125 , 6.4483564357180544e-46 ], [ 411.1875 , 6.449809826682127e-46 ], 
                    [ 411.25 , 6.45126354522502e-46 ], [ 411.3125 , 6.452717591420751e-46 ], [ 411.375 , 6.454171965342847e-46 ], [ 411.4375 , 6.455626667065542e-46 ], 
                    [ 411.5 , 6.457081696662397e-46 ], [ 411.5625 , 6.458537054207542e-46 ], [ 411.625 , 6.459992739774753e-46 ], [ 411.6875 , 6.461449940427209e-46 ], 
                    [ 411.75 , 6.462908980839628e-46 ], [ 411.8125 , 6.4643683507134456e-46 ], [ 411.875 , 6.4658280501232385e-46 ], [ 411.9375 , 6.467288079143236e-46 ], 
                    [ 412.0 , 6.468748437848005e-46 ], [ 412.0625 , 6.4702091263119896e-46 ], [ 412.125 , 6.471670144609558e-46 ], [ 412.1875 , 6.473131492815192e-46 ], 
                    [ 412.25 , 6.474593171003478e-46 ], [ 412.3125 , 6.4760551792489275e-46 ], [ 412.375 , 6.477517517626025e-46 ], [ 412.4375 , 6.47898018620927e-46 ], 
                    [ 412.5 , 6.48044318507327e-46 ], [ 412.5625 , 6.4819065142926525e-46 ], [ 412.625 , 6.483369056435466e-46 ], [ 412.6875 , 6.484828680093865e-46 ], 
                    [ 412.75 , 6.486288632362422e-46 ], [ 412.8125 , 6.487748913315162e-46 ], [ 412.875 , 6.489209523026086e-46 ], [ 412.9375 , 6.49067046156916e-46 ], 
                    [ 413.0 , 6.492131729018508e-46 ], [ 413.0625 , 6.4935933254481325e-46 ], [ 413.125 , 6.495055250932097e-46 ], [ 413.1875 , 6.496517505544482e-46 ], 
                    [ 413.25 , 6.4979800893593864e-46 ], [ 413.3125 , 6.49944300245088e-46 ], [ 413.375 , 6.500906244893229e-46 ], [ 413.4375 , 6.502369816760492e-46 ], 
                    [ 413.5 , 6.5038337181267855e-46 ], [ 413.5625 , 6.505297879389583e-46 ], [ 413.625 , 6.506760697649626e-46 ], [ 413.6875 , 6.508223844847345e-46 ], 
                    [ 413.75 , 6.5096873210566126e-46 ], [ 413.8125 , 6.511151126351507e-46 ], [ 413.875 , 6.512615260805934e-46 ], [ 413.9375 , 6.514079724494004e-46 ], 
                    [ 414.0 , 6.515544517489657e-46 ], [ 414.0625 , 6.517009639867129e-46 ], [ 414.125 , 6.518475091700162e-46 ], [ 414.1875 , 6.519940873063116e-46 ], 
                    [ 414.25 , 6.5214069840299994e-46 ], [ 414.3125 , 6.522873424674883e-46 ], [ 414.375 , 6.5243401950719e-46 ], [ 414.4375 , 6.525807295295385e-46 ], 
                    [ 414.5 , 6.527274725419273e-46 ], [ 414.5625 , 6.528741745148101e-46 ], [ 414.625 , 6.5302088808299985e-46 ], [ 414.6875 , 6.531676346205916e-46 ], 
                    [ 414.75 , 6.5331441413498954e-46 ], [ 414.8125 , 6.534612266335996e-46 ], [ 414.875 , 6.536080721238571e-46 ], [ 414.9375 , 6.537549506131574e-46 ], 
                    [ 415.0 , 6.5390186210892064e-46 ], [ 415.0625 , 6.5404880661855465e-46 ], [ 415.125 , 6.541957841495064e-46 ], [ 415.1875 , 6.543427947091684e-46 ], 
                    [ 415.25 , 6.544898383049767e-46 ], [ 415.3125 , 6.546369149443414e-46 ], [ 415.375 , 6.547840246346925e-46 ], [ 415.4375 , 6.549311673834853e-46 ], 
                    [ 415.5 , 6.550782323307093e-46 ], [ 415.5625 , 6.552252101680902e-46 ], [ 415.625 , 6.5537222098243896e-46 ], [ 415.6875 , 6.555192647811266e-46 ], 
                    [ 415.75 , 6.5566634157158164e-46 ], [ 415.8125 , 6.558134513611784e-46 ], [ 415.875 , 6.559605941573488e-46 ], [ 415.9375 , 6.561077699674751e-46 ], 
                    [ 416.0 , 6.562549787989738e-46 ], [ 416.0625 , 6.564022206592679e-46 ], [ 416.125 , 6.565494955557351e-46 ], [ 416.1875 , 6.566968034958206e-46 ], 
                    [ 416.25 , 6.568441444869053e-46 ], [ 416.3125 , 6.569915185364471e-46 ], [ 416.375 , 6.571389256518209e-46 ], [ 416.4375 , 6.57286351038727e-46 ], 
                    [ 416.5 , 6.574337269030057e-46 ], [ 416.5625 , 6.575811358116985e-46 ], [ 416.625 , 6.577285777722238e-46 ], [ 416.6875 , 6.578760527919877e-46 ], 
                    [ 416.75 , 6.580235608783936e-46 ], [ 416.8125 , 6.5817110203887866e-46 ], [ 416.875 , 6.583186762808357e-46 ], [ 416.9375 , 6.584662836116913e-46 ], 
                    [ 417.0 , 6.586139240388695e-46 ], [ 417.0625 , 6.587615975697862e-46 ], [ 417.125 , 6.589093042118593e-46 ], [ 417.1875 , 6.5905704397251765e-46 ], 
                    [ 417.25 , 6.5920481685919645e-46 ], [ 417.3125 , 6.593526228792996e-46 ], [ 417.375 , 6.5950046204027985e-46 ], [ 417.4375 , 6.596480283936099e-46 ], 
                    [ 417.5 , 6.597955473955937e-46 ], [ 417.5625 , 6.599430993876798e-46 ], [ 417.625 , 6.600906843772552e-46 ], [ 417.6875 , 6.602383023716758e-46 ], 
                    [ 417.75 , 6.603859533783462e-46 ], [ 417.8125 , 6.605336374046347e-46 ], [ 417.875 , 6.606813544579257e-46 ], [ 417.9375 , 6.608291045456285e-46 ], 
                    [ 418.0 , 6.6097688767508865e-46 ], [ 418.0625 , 6.6112470385372804e-46 ], [ 418.125 , 6.612725530889376e-46 ], [ 418.1875 , 6.614204353880817e-46 ], 
                    [ 418.25 , 6.615683507585875e-46 ], [ 418.3125 , 6.617162992078366e-46 ], [ 418.375 , 6.618643016631343e-46 ], [ 418.4375 , 6.6201237078787384e-46 ], 
                    [ 418.5 , 6.621604730379318e-46 ], [ 418.5625 , 6.623086084207141e-46 ], [ 418.625 , 6.6245677694364224e-46 ], [ 418.6875 , 6.6260497861412105e-46 ], 
                    [ 418.75 , 6.627532134395707e-46 ], [ 418.8125 , 6.629014814274132e-46 ], [ 418.875 , 6.6304978258505804e-46 ], [ 418.9375 , 6.631981169199257e-46 ], 
                    [ 419.0 , 6.633464844394576e-46 ], [ 419.0625 , 6.6349488515104905e-46 ], [ 419.125 , 6.63643319062154e-46 ], [ 419.1875 , 6.637917861801716e-46 ], 
                    [ 419.25 , 6.63940286512554e-46 ], [ 419.3125 , 6.640888200667177e-46 ], [ 419.375 , 6.642370139998229e-46 ], [ 419.4375 , 6.6438522546408434e-46 ], 
                    [ 419.5 , 6.645334699988182e-46 ], [ 419.5625 , 6.646817476114222e-46 ], [ 419.625 , 6.648300583092586e-46 ], [ 419.6875 , 6.649784020997188e-46 ], 
                    [ 419.75 , 6.651267789901963e-46 ], [ 419.8125 , 6.652751889880532e-46 ], [ 419.875 , 6.654236321007003e-46 ], [ 419.9375 , 6.655721083355171e-46 ], 
                    [ 420.0 , 6.657206176999035e-46 ], [ 420.0625 , 6.6586916020122825e-46 ], [ 420.125 , 6.66017735846904e-46 ], [ 420.1875 , 6.6616634464431685e-46 ], 
                    [ 420.25 , 6.6631498660087344e-46 ], [ 420.3125 , 6.664635727522147e-46 ], [ 420.375 , 6.6661209866960174e-46 ], [ 420.4375 , 6.6676065768698515e-46 ], 
                    [ 420.5 , 6.6690924981175565e-46 ], [ 420.5625 , 6.67057875051282e-46 ], [ 420.625 , 6.6720653341295815e-46 ], [ 420.6875 , 6.673552249041516e-46 ], 
                    [ 420.75 , 6.675039495322453e-46 ], [ 420.8125 , 6.6765270730462894e-46 ], [ 420.875 , 6.6780149822869355e-46 ], [ 420.9375 , 6.6795032231181776e-46 ], 
                    [ 421.0 , 6.6809917956140555e-46 ], [ 421.0625 , 6.682480699848387e-46 ], [ 421.125 , 6.6839699358950094e-46 ], [ 421.1875 , 6.685459503827961e-46 ], 
                    [ 421.25 , 6.686949403721256e-46 ], [ 421.3125 , 6.6884381355915515e-46 ], [ 421.375 , 6.689927149302107e-46 ], [ 421.4375 , 6.691416494504474e-46 ], 
                    [ 421.5 , 6.692906171272258e-46 ], [ 421.5625 , 6.6943961796794195e-46 ], [ 421.625 , 6.695886519799787e-46 ], [ 421.6875 , 6.697377191707115e-46 ], 
                    [ 421.75 , 6.698868195475266e-46 ], [ 421.8125 , 6.700359531178218e-46 ], [ 421.875 , 6.701851198889722e-46 ], [ 421.9375 , 6.7033431986839325e-46 ], 
                    [ 422.0 , 6.704835530634587e-46 ], [ 422.0625 , 6.70632819481568e-46 ], [ 422.125 , 6.707821191301222e-46 ], [ 422.1875 , 6.709314520165049e-46 ], 
                    [ 422.25 , 6.710807383690202e-46 ], [ 422.3125 , 6.712299530205639e-46 ], [ 422.375 , 6.713792008499614e-46 ], [ 422.4375 , 6.715284818645706e-46 ], 
                    [ 422.5 , 6.716777960717848e-46 ], [ 422.5625 , 6.71827143478989e-46 ], [ 422.625 , 6.719765240935461e-46 ], [ 422.6875 , 6.7212593792284925e-46 ], 
                    [ 422.75 , 6.7227538497428395e-46 ], [ 422.8125 , 6.724248652552369e-46 ], [ 422.875 , 6.725743787731015e-46 ], [ 422.9375 , 6.727239255352634e-46 ], 
                    [ 423.0 , 6.728735055491094e-46 ], [ 423.0625 , 6.730231188220332e-46 ], [ 423.125 , 6.731727653614441e-46 ], [ 423.1875 , 6.733224451747295e-46 ], 
                    [ 423.25 , 6.7347195604708045e-46 ], [ 423.3125 , 6.736214638628213e-46 ], [ 423.375 , 6.7377100486863566e-46 ], [ 423.4375 , 6.739205790718964e-46 ], 
                    [ 423.5 , 6.74070186479978e-46 ], [ 423.5625 , 6.74219827100247e-46 ], [ 423.625 , 6.743695009400764e-46 ], [ 423.6875 , 6.745192080068358e-46 ], 
                    [ 423.75 , 6.746689483079113e-46 ], [ 423.8125 , 6.748187218506758e-46 ], [ 423.875 , 6.749685286425136e-46 ], [ 423.9375 , 6.751183686907819e-46 ], 
                    [ 424.0 , 6.752682420028971e-46 ], [ 424.0625 , 6.754181485862293e-46 ], [ 424.125 , 6.755680884481643e-46 ], [ 424.1875 , 6.7571803721770766e-46 ], 
                    [ 424.25 , 6.75867938551055e-46 ], [ 424.3125 , 6.760178731385239e-46 ], [ 424.375 , 6.761678409874914e-46 ], [ 424.4375 , 6.7631784210534115e-46 ], 
                    [ 424.5 , 6.76467876499439e-46 ], [ 424.5625 , 6.766179441771813e-46 ], [ 424.625 , 6.767680451459519e-46 ], [ 424.6875 , 6.769181794131216e-46 ], 
                    [ 424.75 , 6.7706834698609645e-46 ], [ 424.8125 , 6.77218547872241e-46 ], [ 424.875 , 6.7736878207897444e-46 ], [ 424.9375 , 6.7751904961367884e-46 ], 
                    [ 425.0 , 6.776693504837285e-46 ], [ 425.0625 , 6.778196846965425e-46 ], [ 425.125 , 6.7797005225950335e-46 ], [ 425.1875 , 6.781203910512025e-46 ], 
                    [ 425.25 , 6.7827071891434095e-46 ], [ 425.3125 , 6.784210801026286e-46 ], [ 425.375 , 6.785714746234626e-46 ], [ 425.4375 , 6.787219024842274e-46 ], 
                    [ 425.5 , 6.7887236369230434e-46 ], [ 425.5625 , 6.7902285825511496e-46 ], [ 425.625 , 6.791733861800195e-46 ], [ 425.6875 , 6.793239474744427e-46 ], 
                    [ 425.75 , 6.794745421457485e-46 ], [ 425.8125 , 6.7962517020135505e-46 ], [ 425.875 , 6.797758316486826e-46 ], [ 425.9375 , 6.799265264950994e-46 ], 
                    [ 426.0 , 6.800772547480292e-46 ], [ 426.0625 , 6.802280164148774e-46 ], [ 426.125 , 6.8037881150303196e-46 ], [ 426.1875 , 6.805292829729406e-46 ], 
                    [ 426.25 , 6.806797497959991e-46 ], [ 426.3125 , 6.808302498876795e-46 ], [ 426.375 , 6.809807832553231e-46 ], [ 426.4375 , 6.811313499062922e-46 ], 
                    [ 426.5 , 6.812819498479408e-46 ], [ 426.5625 , 6.814325830876492e-46 ], [ 426.625 , 6.815832496327455e-46 ], [ 426.6875 , 6.817339494906326e-46 ], 
                    [ 426.75 , 6.81884682668642e-46 ], [ 426.8125 , 6.820354491741554e-46 ], [ 426.875 , 6.821862490145513e-46 ], [ 426.9375 , 6.82337082197181e-46 ], 
                    [ 427.0 , 6.824879487294354e-46 ], [ 427.0625 , 6.826388486186693e-46 ], [ 427.125 , 6.827897753148314e-46 ], [ 427.1875 , 6.829407046722603e-46 ], 
                    [ 427.25 , 6.8309166739234115e-46 ], [ 427.3125 , 6.8324266348243405e-46 ], [ 427.375 , 6.8339369294992025e-46 ], [ 427.4375 , 6.835447558021777e-46 ], 
                    [ 427.5 , 6.836958520465909e-46 ], [ 427.5625 , 6.8384698169053144e-46 ], [ 427.625 , 6.839981447413919e-46 ], [ 427.6875 , 6.8414934120654225e-46 ], 
                    [ 427.75 , 6.843005710933882e-46 ], [ 427.8125 , 6.844518344093125e-46 ], [ 427.875 , 6.8460313116169015e-46 ], [ 427.9375 , 6.84754461357922e-46 ], 
                    [ 428.0 , 6.849058250054054e-46 ], [ 428.0625 , 6.850572221115202e-46 ], [ 428.125 , 6.8520852193736205e-46 ], [ 428.1875 , 6.853596715906647e-46 ], 
                    [ 428.25 , 6.8551085458595066e-46 ], [ 428.3125 , 6.856620709305892e-46 ], [ 428.375 , 6.8581332063193705e-46 ], [ 428.4375 , 6.859646036973374e-46 ], 
                    [ 428.5 , 6.8611592013416505e-46 ], [ 428.5625 , 6.862672699497762e-46 ], [ 428.625 , 6.8641865315153875e-46 ], [ 428.6875 , 6.865700697468077e-46 ], 
                    [ 428.75 , 6.867215197429492e-46 ], [ 428.8125 , 6.86873003147341e-46 ], [ 428.875 , 6.8702451996734275e-46 ], [ 428.9375 , 6.871760702103402e-46 ], 
                    [ 429.0 , 6.873276538836914e-46 ], [ 429.0625 , 6.874792709947758e-46 ], [ 429.125 , 6.876311079815738e-46 ], [ 429.1875 , 6.87783090864384e-46 ], 
                    [ 429.25 , 6.879351073390427e-46 ], [ 429.3125 , 6.880871574129649e-46 ], [ 429.375 , 6.882392410935814e-46 ], [ 429.4375 , 6.883913583883252e-46 ], 
                    [ 429.5 , 6.885435093046112e-46 ], [ 429.5625 , 6.886956938498851e-46 ], [ 429.625 , 6.8884791203157976e-46 ], [ 429.6875 , 6.89000163857115e-46 ], 
                    [ 429.75 , 6.8915244933394144e-46 ], [ 429.8125 , 6.89304768469492e-46 ], [ 429.875 , 6.894571212712062e-46 ], [ 429.9375 , 6.896095077465249e-46 ], 
                    [ 430.0 , 6.897619279028909e-46 ], [ 430.0625 , 6.899143817477483e-46 ], [ 430.125 , 6.900665483390605e-46 ], [ 430.1875 , 6.9021866823698646e-46 ], 
                    [ 430.25 , 6.903708216685815e-46 ], [ 430.3125 , 6.905230086412332e-46 ], [ 430.375 , 6.906752291623449e-46 ], [ 430.4375 , 6.9082748323930754e-46 ], 
                    [ 430.5 , 6.909797708795181e-46 ], [ 430.5625 , 6.9113209209037045e-46 ], [ 430.625 , 6.9128444687926985e-46 ], [ 430.6875 , 6.9143683525362815e-46 ], 
                    [ 430.75 , 6.915892572208343e-46 ], [ 430.8125 , 6.917417127882936e-46 ], [ 430.875 , 6.9189420196342765e-46 ], [ 430.9375 , 6.9204672475363034e-46 ], 
                    [ 431.0 , 6.921992811663118e-46 ], [ 431.0625 , 6.923518712088888e-46 ], [ 431.125 , 6.925045133628406e-46 ], [ 431.1875 , 6.926571906033116e-46 ], 
                    [ 431.25 , 6.928099015046988e-46 ], [ 431.3125 , 6.9296264607443325e-46 ], [ 431.375 , 6.931154243199279e-46 ], [ 431.4375 , 6.932682362486173e-46 ], 
                    [ 431.5 , 6.934210818679176e-46 ], [ 431.5625 , 6.935739611852618e-46 ], [ 431.625 , 6.93726874208079e-46 ], [ 431.6875 , 6.938798209438055e-46 ], 
                    [ 431.75 , 6.9403280139987385e-46 ], [ 431.8125 , 6.941858155837036e-46 ], [ 431.875 , 6.943388635027359e-46 ], [ 431.9375 , 6.9449194516441806e-46 ], 
                    [ 432.0 , 6.946450605761943e-46 ], [ 432.0625 , 6.947982184181827e-46 ], [ 432.125 , 6.949517627965204e-46 ], [ 432.1875 , 6.9510534110682885e-46 ], 
                    [ 432.25 , 6.952589533566163e-46 ], [ 432.3125 , 6.954125995533782e-46 ], [ 432.375 , 6.955662797046068e-46 ], [ 432.4375 , 6.957199938178252e-46 ], 
                    [ 432.5 , 6.9587374190053425e-46 ], [ 432.5625 , 6.9602752396022565e-46 ], [ 432.625 , 6.96181340004423e-46 ], [ 432.6875 , 6.963351900406218e-46 ], 
                    [ 432.75 , 6.964890740763537e-46 ], [ 432.8125 , 6.966429921191324e-46 ], [ 432.875 , 6.967969441764432e-46 ], [ 432.9375 , 6.969509302558331e-46 ], 
                    [ 433.0 , 6.971049503648055e-46 ], [ 433.0625 , 6.972590514558642e-46 ], [ 433.125 , 6.9741367324064495e-46 ], [ 433.1875 , 6.975683293138294e-46 ], 
                    [ 433.25 , 6.9772301968301105e-46 ], [ 433.3125 , 6.978777443558055e-46 ], [ 433.375 , 6.980325033398147e-46 ], [ 433.4375 , 6.981872966426473e-46 ], 
                    [ 433.5 , 6.983421242719088e-46 ], [ 433.5625 , 6.984969862352265e-46 ], [ 433.625 , 6.986518825401939e-46 ], [ 433.6875 , 6.988068131944466e-46 ], 
                    [ 433.75 , 6.989617782055869e-46 ], [ 433.8125 , 6.991167775812336e-46 ], [ 433.875 , 6.992718113290122e-46 ], [ 433.9375 , 6.994268794565549e-46 ], 
                    [ 434.0 , 6.9958198197146085e-46 ], [ 434.0625 , 6.99737316628718e-46 ], [ 434.125 , 6.998943898262146e-46 ], [ 434.1875 , 7.0005149828264636e-46 ], 
                    [ 434.25 , 7.0020864200592845e-46 ], [ 434.3125 , 7.003658210039671e-46 ], [ 434.375 , 7.0052303528469085e-46 ], [ 434.4375 , 7.006802848560146e-46 ], 
                    [ 434.5 , 7.0083756972585e-46 ], [ 434.5625 , 7.009948899021362e-46 ], [ 434.625 , 7.011522453928032e-46 ], [ 434.6875 , 7.013096362057582e-46 ], 
                    [ 434.75 , 7.014670623489403e-46 ], [ 434.8125 , 7.0162452383028e-46 ], [ 434.875 , 7.017820206577151e-46 ], [ 434.9375 , 7.019395528391647e-46 ], 
                    [ 435.0 , 7.0209712038259e-46 ], [ 435.0625 , 7.0225540548962e-46 ], [ 435.125 , 7.024207979219785e-46 ], [ 435.1875 , 7.025862293069273e-46 ], 
                    [ 435.25 , 7.027516996536205e-46 ], [ 435.3125 , 7.029172089712395e-46 ], [ 435.375 , 7.030827572689521e-46 ], [ 435.4375 , 7.03248344555959e-46 ], 
                    [ 435.5 , 7.034139708414279e-46 ], [ 435.5625 , 7.035796361345582e-46 ], [ 435.625 , 7.037453404445122e-46 ], [ 435.6875 , 7.03911083780504e-46 ], 
                    [ 435.75 , 7.040768661517195e-46 ], [ 435.8125 , 7.042426875673425e-46 ], [ 435.875 , 7.044085480365633e-46 ], [ 435.9375 , 7.045744475685998e-46 ], 
                    [ 436.0 , 7.047403861726419e-46 ], [ 436.0625 , 7.049084264727921e-46 ], [ 436.125 , 7.051260232132965e-46 ], [ 436.1875 , 7.053436871232989e-46 ], 
                    [ 436.25 , 7.055614182235233e-46 ], [ 436.3125 , 7.057792165347209e-46 ], [ 436.375 , 7.059970820776385e-46 ], [ 436.4375 , 7.062150148730202e-46 ], 
                    [ 436.5 , 7.064330149416258e-46 ], [ 436.5625 , 7.066510823042267e-46 ], [ 436.625 , 7.068692169816007e-46 ], [ 436.6875 , 7.070874189945222e-46 ], 
                    [ 436.75 , 7.073056883637817e-46 ], [ 436.8125 , 7.075240251101566e-46 ], [ 436.875 , 7.077424292544651e-46 ], [ 436.9375 , 7.079609008174871e-46 ], 
                    [ 437.0 , 7.081794398200645e-46 ], [ 437.0625 , 7.083980462829891e-46 ], [ 437.125 , 8.357118585059707e-46 ], [ 437.1875 , 9.956023472435798e-46 ], 
                    [ 437.25 , 1.1860834852923867e-45 ], [ 437.3125 , 1.413007952400031e-45 ], [ 437.375 , 1.683348177682074e-45 ], [ 437.4375 , 2.0054105728793052e-45 ], 
                    [ 437.5 , 2.38909075325928e-45 ], [ 437.5625 , 2.846177588020781e-45 ], [ 437.625 , 3.390715422384227e-45 ], [ 437.6875 , 4.0394356009208744e-45 ], 
                    [ 437.75 , 4.812270551007519e-45 ], [ 437.8125 , 5.732966222017632e-45 ], [ 437.875 , 6.82981169791336e-45 ], [ 437.9375 , 8.136508401149686e-45 ], 
                    [ 438.0 , 9.69320559484907e-45 ], [ 438.0625 , 1.1547733999848846e-44 ], [ 438.125 , 1.0344750774356558e-44 ], [ 438.1875 , 8.690129214944478e-45 ], 
                    [ 438.25 , 7.300160962759328e-45 ], [ 438.3125 , 6.132515266924723e-45 ], [ 438.375 , 5.151632092897e-45 ], [ 438.4375 , 4.3276391603464885e-45 ], 
                    [ 438.5 , 3.6354421985970505e-45 ], [ 438.5625 , 3.053960713832227e-45 ], [ 438.625 , 2.5654859937616285e-45 ], [ 438.6875 , 2.155141798117009e-45 ], 
                    [ 438.75 , 1.8104313105918562e-45 ], [ 438.8125 , 1.520856554884323e-45 ], [ 438.875 , 1.2775986843590457e-45 ], [ 438.9375 , 1.0732494087189671e-45 ], 
                    [ 439.0 , 9.01585378426941e-46 ], [ 439.0625 , 7.573786558741116e-46 ], [ 439.125 , 7.133660675730396e-46 ], [ 439.1875 , 7.134613545175127e-46 ], 
                    [ 439.25 , 7.135566541898157e-46 ], [ 439.3125 , 7.136519665916431e-46 ], [ 439.375 , 7.13747291724706e-46 ], [ 439.4375 , 7.138426295906891e-46 ], 
                    [ 439.5 , 7.139379801913189e-46 ], [ 439.5625 , 7.14033343528271e-46 ], [ 439.625 , 7.14128719603267e-46 ], [ 439.6875 , 7.142241084179829e-46 ], 
                    [ 439.75 , 7.143195099741457e-46 ], [ 439.8125 , 7.144149242734372e-46 ], [ 439.875 , 7.145103513175899e-46 ], [ 439.9375 , 7.146057911082707e-46 ], 
                    [ 440.0 , 7.147012436471921e-46 ], [ 440.0625 , 7.147967089360725e-46 ], [ 440.125 , 7.149146266256262e-46 ], [ 440.1875 , 7.150584905843397e-46 ], 
                    [ 440.25 , 7.152023834931378e-46 ], [ 440.3125 , 7.153463053578505e-46 ], [ 440.375 , 7.15490256184295e-46 ], [ 440.4375 , 7.156342359782991e-46 ], 
                    [ 440.5 , 7.157782447457126e-46 ], [ 440.5625 , 7.15922282492335e-46 ], [ 440.625 , 7.160663492240237e-46 ], [ 440.6875 , 7.162104449465959e-46 ], 
                    [ 440.75 , 7.163545696658958e-46 ], [ 440.8125 , 7.16498723387738e-46 ], [ 440.875 , 7.166429061179945e-46 ], [ 440.9375 , 7.167871178624721e-46 ], 
                    [ 441.0 , 7.169313586270195e-46 ], [ 441.0625 , 7.170756284174716e-46 ], [ 441.125 , 7.172215363970772e-46 ], [ 441.1875 , 7.173730498086825e-46 ], 
                    [ 441.25 , 7.175245952275724e-46 ], [ 441.3125 , 7.176761726605085e-46 ], [ 441.375 , 7.17827782114259e-46 ], [ 441.4375 , 7.179794235955729e-46 ], 
                    [ 441.5 , 7.181310971112366e-46 ], [ 441.5625 , 7.182828026680067e-46 ], [ 441.625 , 7.184345402726422e-46 ], [ 441.6875 , 7.185863099319235e-46 ], 
                    [ 441.75 , 7.187381116526371e-46 ], [ 441.8125 , 7.188899454415356e-46 ], [ 441.875 , 7.190418113053937e-46 ], [ 441.9375 , 7.19193709250987e-46 ], 
                    [ 442.0 , 7.193456392851083e-46 ], [ 442.0625 , 7.194976014145208e-46 ], [ 442.125 , 7.196495956460049e-46 ], [ 442.1875 , 7.198034908303712e-46 ], 
                    [ 442.25 , 7.199575467365228e-46 ], [ 442.3125 , 7.20111635614482e-46 ], [ 442.375 , 7.202657574713003e-46 ], [ 442.4375 , 7.204199123140517e-46 ], 
                    [ 442.5 , 7.205741001497753e-46 ], [ 442.5625 , 7.207283209855531e-46 ], [ 442.625 , 7.208825748284272e-46 ], [ 442.6875 , 7.210368616854775e-46 ], 
                    [ 442.75 , 7.211911815637646e-46 ], [ 442.8125 , 7.213455344703509e-46 ], [ 442.875 , 7.214999204123205e-46 ], [ 442.9375 , 7.216543393967285e-46 ], 
                    [ 443.0 , 7.21808791430647e-46 ], [ 443.0625 , 7.219632765211593e-46 ], [ 443.125 , 7.221177946753352e-46 ], [ 443.1875 , 7.222726452729021e-46 ], 
                    [ 443.25 , 7.224277034322463e-46 ], [ 443.3125 , 7.225827948796143e-46 ], [ 443.375 , 7.227379196221576e-46 ], [ 443.4375 , 7.228930776670293e-46 ], 
                    [ 443.5 , 7.230482690213737e-46 ], [ 443.5625 , 7.23203493692336e-46 ], [ 443.625 , 7.233587516870797e-46 ], [ 443.6875 , 7.235140430127531e-46 ], 
                    [ 443.75 , 7.236693676765068e-46 ], [ 443.8125 , 7.23824725685503e-46 ], [ 443.875 , 7.239801170469104e-46 ], [ 443.9375 , 7.241355417678687e-46 ], 
                    [ 444.0 , 7.2429099985556e-46 ], [ 444.0625 , 7.244464913171221e-46 ], [ 444.125 , 7.24602016159745e-46 ], [ 444.1875 , 7.247576349485625e-46 ], 
                    [ 444.25 , 7.249134429127359e-46 ], [ 444.3125 , 7.250692843724074e-46 ], [ 444.375 , 7.252251593347934e-46 ], [ 444.4375 , 7.253810678070857e-46 ], 
                    [ 444.5 , 7.255370097964782e-46 ], [ 444.5625 , 7.256929853101916e-46 ], [ 444.625 , 7.258489943554385e-46 ], [ 444.6875 , 7.260050369394011e-46 ], 
                    [ 444.75 , 7.261611130693056e-46 ], [ 444.8125 , 7.26317222752379e-46 ], [ 444.875 , 7.264733659958188e-46 ], [ 444.9375 , 7.266295428068193e-46 ], 
                    [ 445.0 , 7.267857531926227e-46 ], [ 445.0625 , 7.269419971604573e-46 ], [ 445.125 , 7.270982747175268e-46 ], [ 445.1875 , 7.272545858710368e-46 ], 
                    [ 445.25 , 7.274107457525364e-46 ], [ 445.3125 , 7.27566917702117e-46 ], [ 445.375 , 7.27723123181136e-46 ], [ 445.4375 , 7.278793621967875e-46 ], 
                    [ 445.5 , 7.280356347563025e-46 ], [ 445.5625 , 7.281919408668411e-46 ], [ 445.625 , 7.283482805356378e-46 ], [ 445.6875 , 7.285046537698768e-46 ], 
                    [ 445.75 , 7.286610605767954e-46 ], [ 445.8125 , 7.288175009635601e-46 ], [ 445.875 , 7.289739749374115e-46 ], [ 445.9375 , 7.291304825055552e-46 ], 
                    [ 446.0 , 7.292870236751832e-46 ], [ 446.0625 , 7.294435984535406e-46 ], [ 446.125 , 7.296002068478226e-46 ], [ 446.1875 , 7.29756848865246e-46 ], 
                    [ 446.25 , 7.299136224875755e-46 ], [ 446.3125 , 7.300705429913878e-46 ], [ 446.375 , 7.302274972307544e-46 ], [ 446.4375 , 7.303844852129335e-46 ], 
                    [ 446.5 , 7.305415069451841e-46 ], [ 446.5625 , 7.306985624347516e-46 ], [ 446.625 , 7.308556516889092e-46 ], [ 446.6875 , 7.310127747148949e-46 ], 
                    [ 446.75 , 7.311699315199792e-46 ], [ 446.8125 , 7.313271221114192e-46 ], [ 446.875 , 7.314843464965043e-46 ], [ 446.9375 , 7.316416046824688e-46 ], 
                    [ 447.0 , 7.317988966765841e-46 ], [ 447.0625 , 7.319562224861187e-46 ], [ 447.125 , 7.32113582118358e-46 ], [ 447.1875 , 7.322709755805681e-46 ], 
                    [ 447.25 , 7.324284028800012e-46 ], [ 447.3125 , 7.325856039622654e-46 ], [ 447.375 , 7.327428387845922e-46 ], [ 447.4375 , 7.329001073542235e-46 ], 
                    [ 447.5 , 7.330574096784022e-46 ], [ 447.5625 , 7.332147457643733e-46 ], [ 447.625 , 7.333721156193829e-46 ], [ 447.6875 , 7.33529519250679e-46 ], 
                    [ 447.75 , 7.33686956665511e-46 ], [ 447.8125 , 7.338444278711244e-46 ], [ 447.875 , 7.340019328747928e-46 ], [ 447.9375 , 7.34159471683739e-46 ], 
                    [ 448.0 , 7.343170443052394e-46 ], [ 448.0625 , 7.344746507465511e-46 ], [ 448.125 , 7.346322910149123e-46 ], [ 448.1875 , 7.34789965117609e-46 ], 
                    [ 448.25 , 7.349476730618928e-46 ], [ 448.3125 , 7.351053078227742e-46 ], [ 448.375 , 7.352628710379271e-46 ], [ 448.4375 , 7.354204680253547e-46 ], 
                    [ 448.5 , 7.355780987922747e-46 ], [ 448.5625 , 7.357357633459223e-46 ], [ 448.625 , 7.358934616935602e-46 ], [ 448.6875 , 7.36051193842411e-46 ], 
                    [ 448.75 , 7.362089597997353e-46 ], [ 448.8125 , 7.363667595727799e-46 ], [ 448.875 , 7.365245931687767e-46 ], [ 448.9375 , 7.366824605949966e-46 ], 
                    [ 449.0 , 7.368403618586752e-46 ], [ 449.0625 , 7.369982969670756e-46 ], [ 449.125 , 7.371562659274364e-46 ], [ 449.1875 , 7.373142687470291e-46 ], 
                    [ 449.25 , 7.37472305433106e-46 ], [ 449.3125 , 7.376303759929363e-46 ], [ 449.375 , 7.377883569462405e-46 ], [ 449.4375 , 7.37946366588489e-46 ], 
                    [ 449.5 , 7.381044100711293e-46 ], [ 449.5625 , 7.382624874014088e-46 ], [ 449.625 , 7.384205985865714e-46 ], [ 449.6875 , 7.385787436338834e-46 ], 
                    [ 449.75 , 7.387369225505758e-46 ], [ 449.8125 , 7.388951353439181e-46 ], [ 449.875 , 7.390533820211553e-46 ], [ 449.9375 , 7.392116625895545e-46 ], 
                    [ 450.0 , 7.393699770563688e-46 ], [ 450.0625 , 7.395283254288528e-46 ], [ 450.125 , 7.396867077142732e-46 ], [ 450.1875 , 7.398451239198983e-46 ], 
                    [ 450.25 , 7.400035740529877e-46 ], [ 450.3125 , 7.401620581208125e-46 ], [ 450.375 , 7.403204677119582e-46 ], [ 450.4375 , 7.404787372471415e-46 ], 
                    [ 450.5 , 7.406370406180042e-46 ], [ 450.5625 , 7.40795377831774e-46 ], [ 450.625 , 7.409537488957023e-46 ], [ 450.6875 , 7.411121538170097e-46 ], 
                    [ 450.75 , 7.412705926029449e-46 ], [ 450.8125 , 7.414290652607373e-46 ], [ 450.875 , 7.415875717976335e-46 ], [ 450.9375 , 7.41746112220876e-46 ], 
                    [ 451.0 , 7.419046865377096e-46 ], [ 451.0625 , 7.420632947553802e-46 ], [ 451.125 , 7.422219368811455e-46 ], [ 451.1875 , 7.423806129222284e-46 ], 
                    [ 451.25 , 7.425393228859007e-46 ], [ 451.3125 , 7.426980667794145e-46 ], [ 451.375 , 7.428568446100127e-46 ], [ 451.4375 , 7.43015571547937e-46 ], 
                    [ 451.5 , 7.43174305604846e-46 ], [ 451.5625 , 7.433330735728838e-46 ], [ 451.625 , 7.434918754593006e-46 ], [ 451.6875 , 7.43650711271353e-46 ], 
                    [ 451.75 , 7.438095810162676e-46 ], [ 451.8125 , 7.439684847013093e-46 ], [ 451.875 , 7.441274223337239e-46 ], [ 451.9375 , 7.442863939207584e-46 ], 
                    [ 452.0 , 7.444453994696719e-46 ], [ 452.0625 , 7.44604438987725e-46 ], [ 452.125 , 7.44763512482159e-46 ], [ 452.1875 , 7.449226199602536e-46 ], 
                    [ 452.25 , 7.450817614292582e-46 ], [ 452.3125 , 7.452409368964399e-46 ], [ 452.375 , 7.45400146369046e-46 ], [ 452.4375 , 7.4555933746146e-46 ], 
                    [ 452.5 , 7.45718178254978e-46 ], [ 452.5625 , 7.458770528893999e-46 ], [ 452.625 , 7.460359613719189e-46 ], [ 452.6875 , 7.461949037097573e-46 ], 
                    [ 452.75 , 7.463538799101334e-46 ], [ 452.8125 , 7.465128899802665e-46 ], [ 452.875 , 7.46671933927341e-46 ], [ 452.9375 , 7.468310117586113e-46 ], 
                    [ 453.0 , 7.469901234812699e-46 ], [ 453.0625 , 7.471492691025587e-46 ], [ 453.125 , 7.473084486296835e-46 ], [ 453.1875 , 7.474676620698739e-46 ], 
                    [ 453.25 , 7.476269094303543e-46 ], [ 453.3125 , 7.47786190718357e-46 ], [ 453.375 , 7.479455059410893e-46 ], [ 453.4375 , 7.481048551058123e-46 ], 
                    [ 453.5 , 7.482643298042937e-46 ], [ 453.5625 , 7.484239670963092e-46 ], [ 453.625 , 7.485836384458951e-46 ], [ 453.6875 , 7.487433438603276e-46 ], 
                    [ 453.75 , 7.489030833468536e-46 ], [ 453.8125 , 7.490628569127574e-46 ], [ 453.875 , 7.492226645652994e-46 ], [ 453.9375 , 7.49382506311757e-46 ], 
                    [ 454.0 , 7.495423821594147e-46 ], [ 454.0625 , 7.497022921155261e-46 ], [ 454.125 , 7.498622361873736e-46 ], [ 454.1875 , 7.500222143822623e-46 ], 
                    [ 454.25 , 7.501822267074347e-46 ], [ 454.3125 , 7.503422731701828e-46 ], [ 454.375 , 7.505023537778056e-46 ], [ 454.4375 , 7.506624685375774e-46 ], 
                    [ 454.5 , 7.50822617456784e-46 ], [ 454.5625 , 7.50982308696358e-46 ], [ 454.625 , 7.511418190253579e-46 ], [ 454.6875 , 7.513013632347206e-46 ], 
                    [ 454.75 , 7.514609413316264e-46 ], [ 454.8125 , 7.51620553323289e-46 ], [ 454.875 , 7.517801992168865e-46 ], [ 454.9375 , 7.519398790196356e-46 ], 
                    [ 455.0 , 7.520995927387335e-46 ], [ 455.0625 , 7.522593403813734e-46 ], [ 455.125 , 7.524191219547874e-46 ], [ 455.1875 , 7.525789374661556e-46 ], 
                    [ 455.25 , 7.527387869226975e-46 ], [ 455.3125 , 7.528986703316283e-46 ], [ 455.375 , 7.530585877001541e-46 ], [ 455.4375 , 7.532185390354883e-46 ], 
                    [ 455.5 , 7.533785243448454e-46 ], [ 455.5625 , 7.535385436354414e-46 ], [ 455.625 , 7.53698734990998e-46 ], [ 455.6875 , 7.538589737269213e-46 ], 
                    [ 455.75 , 7.540192465301177e-46 ], [ 455.8125 , 7.541795534078032e-46 ], [ 455.875 , 7.543398943672434e-46 ], [ 455.9375 , 7.545002694156843e-46 ], 
                    [ 456.0 , 7.546606785603625e-46 ], [ 456.0625 , 7.548211218085326e-46 ], [ 456.125 , 7.549815991674449e-46 ], [ 456.1875 , 7.551421106443515e-46 ], 
                    [ 456.25 , 7.553026562465166e-46 ], [ 456.3125 , 7.554632359811688e-46 ], [ 456.375 , 7.556238498555913e-46 ], [ 456.4375 , 7.557844978770264e-46 ], 
                    [ 456.5 , 7.559451800527392e-46 ], [ 456.5625 , 7.56105896389991e-46 ], [ 456.625 , 7.562665995936408e-46 ], [ 456.6875 , 7.564269618330551e-46 ], 
                    [ 456.75 , 7.565873580764251e-46 ], [ 456.8125 , 7.567477883309451e-46 ], [ 456.875 , 7.569082526038376e-46 ], [ 456.9375 , 7.570687509023e-46 ], 
                    [ 457.0 , 7.572292832335688e-46 ], [ 457.0625 , 7.573898496048494e-46 ], [ 457.125 , 7.575504500233548e-46 ], [ 457.1875 , 7.577110844963099e-46 ], 
                    [ 457.25 , 7.578717530309352e-46 ], [ 457.3125 , 7.580324556344648e-46 ], [ 457.375 , 7.581931923141009e-46 ], [ 457.4375 , 7.5835396307708e-46 ], 
                    [ 457.5 , 7.5851476793064e-46 ], [ 457.5625 , 7.586756068819828e-46 ], [ 457.625 , 7.5883647993836e-46 ], [ 457.6875 , 7.589973281862395e-46 ], 
                    [ 457.75 , 7.591580462298492e-46 ], [ 457.8125 , 7.593187983055968e-46 ], [ 457.875 , 7.594795844206673e-46 ], [ 457.9375 , 7.596404045822791e-46 ], 
                    [ 458.0 , 7.59801258797631e-46 ], [ 458.0625 , 7.599621470739498e-46 ], [ 458.125 , 7.60123069418421e-46 ], [ 458.1875 , 7.602840258382752e-46 ], 
                    [ 458.25 , 7.604450163407435e-46 ], [ 458.3125 , 7.606060409330107e-46 ], [ 458.375 , 7.607670996223224e-46 ], [ 458.4375 , 7.609281924158824e-46 ], 
                    [ 458.5 , 7.610893193209178e-46 ], [ 458.5625 , 7.612504803446515e-46 ], [ 458.625 , 7.614116754943138e-46 ], [ 458.6875 , 7.615729047771144e-46 ], 
                    [ 458.75 , 7.617341734451101e-46 ], [ 458.8125 , 7.618954852722455e-46 ], [ 458.875 , 7.620568312602446e-46 ], [ 458.9375 , 7.622182114163632e-46 ], 
                    [ 459.0 , 7.623796257478049e-46 ], [ 459.0625 , 7.6254107426185e-46 ], [ 459.125 , 7.627025569657104e-46 ], [ 459.1875 , 7.62864073866621e-46 ], 
                    [ 459.25 , 7.630256249718507e-46 ], [ 459.3125 , 7.631872102886105e-46 ], [ 459.375 , 7.633488298241778e-46 ], [ 459.4375 , 7.635104835857612e-46 ], 
                    [ 459.5 , 7.636721715806629e-46 ], [ 459.5625 , 7.638338938160782e-46 ], [ 459.625 , 7.639956502992855e-46 ], [ 459.6875 , 7.641574410375154e-46 ], 
                    [ 459.75 , 7.643192660380549e-46 ], [ 459.8125 , 7.644809225014238e-46 ], [ 459.875 , 7.646423711597896e-46 ], [ 459.9375 , 7.648038539140607e-46 ], 
                    [ 460.0 , 7.649653707714382e-46 ], [ 460.0625 , 7.651269217391293e-46 ], [ 460.125 , 7.652885068243267e-46 ], [ 460.1875 , 7.654501260342416e-46 ], 
                    [ 460.25 , 7.656117793760854e-46 ], [ 460.3125 , 7.657734668570614e-46 ], [ 460.375 , 7.659351884843684e-46 ], [ 460.4375 , 7.660969442652393e-46 ], 
                    [ 460.5 , 7.662587342068704e-46 ], [ 460.5625 , 7.66420558316482e-46 ], [ 460.625 , 7.665824166012894e-46 ], [ 460.6875 , 7.667443090685103e-46 ], 
                    [ 460.75 , 7.669062357253745e-46 ], [ 460.8125 , 7.670681965790694e-46 ], [ 460.875 , 7.672301335010266e-46 ], [ 460.9375 , 7.673920416159523e-46 ], 
                    [ 461.0 , 7.675539838982502e-46 ], [ 461.0625 , 7.67715960355131e-46 ], [ 461.125 , 7.67877970993801e-46 ], [ 461.1875 , 7.680400158214896e-46 ], 
                    [ 461.25 , 7.682020948453846e-46 ], [ 461.3125 , 7.683642080727354e-46 ], [ 461.375 , 7.685263555107322e-46 ], [ 461.4375 , 7.686885371666058e-46 ], 
                    [ 461.5 , 7.688507530475877e-46 ], [ 461.5625 , 7.690130031608898e-46 ], [ 461.625 , 7.691752875137251e-46 ], [ 461.6875 , 7.693376061133245e-46 ], 
                    [ 461.75 , 7.69499958966937e-46 ], [ 461.8125 , 7.696623460817695e-46 ], [ 461.875 , 7.698247674650518e-46 ], [ 461.9375 , 7.6998707977558e-46 ], 
                    [ 462.0 , 7.701492758784856e-46 ], [ 462.0625 , 7.70311506147664e-46 ], [ 462.125 , 7.704737705902842e-46 ], [ 462.1875 , 7.706360692135619e-46 ], 
                    [ 462.25 , 7.70798402024686e-46 ], [ 462.3125 , 7.709607690308634e-46 ], [ 462.375 , 7.711231702392918e-46 ], [ 462.4375 , 7.712856056571925e-46 ], 
                    [ 462.5 , 7.714480752917604e-46 ], [ 462.5625 , 7.716105791501978e-46 ], [ 462.625 , 7.717731172397196e-46 ], [ 462.6875 , 7.719356895675417e-46 ], 
                    [ 462.75 , 7.720982961408654e-46 ], [ 462.8125 , 7.722609369669153e-46 ], [ 462.875 , 7.724236120528906e-46 ], [ 462.9375 , 7.725863214060406e-46 ], 
                    [ 463.0 , 7.72748962305311e-46 ], [ 463.0625 , 7.729115108404285e-46 ], [ 463.125 , 7.730740935677966e-46 ], [ 463.1875 , 7.732367104946135e-46 ], 
                    [ 463.25 , 7.733993616280673e-46 ], [ 463.3125 , 7.735620469753534e-46 ], [ 463.375 , 7.73724766543658e-46 ], [ 463.4375 , 7.738875203402067e-46 ], 
                    [ 463.5 , 7.740503083721775e-46 ], [ 463.5625 , 7.74213130646772e-46 ], [ 463.625 , 7.743759871711931e-46 ], [ 463.6875 , 7.745388779526673e-46 ], 
                    [ 463.75 , 7.747018029983733e-46 ], [ 463.8125 , 7.748647623155347e-46 ], [ 463.875 , 7.7502775591135e-46 ], [ 463.9375 , 7.751907837930402e-46 ], 
                    [ 464.0 , 7.753538459678067e-46 ], [ 464.0625 , 7.755168194449746e-46 ], [ 464.125 , 7.75679608469083e-46 ], [ 464.1875 , 7.758424316642958e-46 ], 
                    [ 464.25 , 7.760052890377972e-46 ], [ 464.3125 , 7.761681805967393e-46 ], [ 464.375 , 7.763311063483092e-46 ], [ 464.4375 , 7.764940662996952e-46 ], 
                    [ 464.5 , 7.766570604580489e-46 ], [ 464.5625 , 7.76820088830578e-46 ], [ 464.625 , 7.769831514244532e-46 ], [ 464.6875 , 7.771462482468585e-46 ], 
                    [ 464.75 , 7.773093793049727e-46 ], [ 464.8125 , 7.77472544605988e-46 ], [ 464.875 , 7.776357441570925e-46 ], [ 464.9375 , 7.777989779654808e-46 ], 
                    [ 465.0 , 7.779622460383388e-46 ], [ 465.0625 , 7.781255483828585e-46 ], [ 465.125 , 7.782888295942199e-46 ], [ 465.1875 , 7.784519695602379e-46 ], 
                    [ 465.25 , 7.786151437226207e-46 ], [ 465.3125 , 7.78778352088536e-46 ], [ 465.375 , 7.789415946651587e-46 ], [ 465.4375 , 7.791048714596491e-46 ], 
                    [ 465.5 , 7.792681824791852e-46 ], [ 465.5625 , 7.794315277309352e-46 ], [ 465.625 , 7.795949072220861e-46 ], [ 465.6875 , 7.797583209598093e-46 ], 
                    [ 465.75 , 7.799217689512831e-46 ], [ 465.8125 , 7.80085251203688e-46 ], [ 465.875 , 7.802487677241996e-46 ], [ 465.9375 , 7.804123185200234e-46 ], 
                    [ 466.0 , 7.805759035983216e-46 ], [ 466.0625 , 7.807395229662749e-46 ], [ 466.125 , 7.809031766310818e-46 ], [ 466.1875 , 7.810668541602325e-46 ], 
                    [ 466.25 , 7.812304314118362e-46 ], [ 466.3125 , 7.813940429210985e-46 ], [ 466.375 , 7.815576886951772e-46 ], [ 466.4375 , 7.81721368741265e-46 ], 
                    [ 466.5 , 7.818850830665338e-46 ], [ 466.5625 , 7.820488316781515e-46 ], [ 466.625 , 7.822126145833263e-46 ], [ 466.6875 , 7.823764317892184e-46 ], 
                    [ 466.75 , 7.825402833030165e-46 ], [ 466.8125 , 7.827041691318947e-46 ], [ 466.875 , 7.82868089283067e-46 ], [ 466.9375 , 7.830320437637053e-46 ], 
                    [ 467.0 , 7.831960325809877e-46 ], [ 467.0625 , 7.833600557421165e-46 ], [ 467.125 , 7.835241132542952e-46 ], [ 467.1875 , 7.836882051247014e-46 ], 
                    [ 467.25 , 7.838523313605306e-46 ], [ 467.3125 , 7.840165368085947e-46 ], [ 467.375 , 7.841807832386028e-46 ], [ 467.4375 , 7.843450640771831e-46 ], 
                    [ 467.5 , 7.845093793315442e-46 ], [ 467.5625 , 7.846737290088907e-46 ], [ 467.625 , 7.848381131164447e-46 ], [ 467.6875 , 7.850025316614134e-46 ], 
                    [ 467.75 , 7.851669846510118e-46 ], [ 467.8125 , 7.853314720924496e-46 ], [ 467.875 , 7.85495993992956e-46 ], [ 467.9375 , 7.856605503597384e-46 ], 
                    [ 468.0 , 7.858251412000341e-46 ], [ 468.0625 , 7.859897665210486e-46 ], [ 468.125 , 7.861544263300049e-46 ], [ 468.1875 , 7.863191206341451e-46 ], 
                    [ 468.25 , 7.864838494406789e-46 ], [ 468.3125 , 7.866486127568286e-46 ], [ 468.375 , 7.868130063530474e-46 ], [ 468.4375 , 7.869772068735565e-46 ], 
                    [ 468.5 , 7.871414416611796e-46 ], [ 468.5625 , 7.873057107230679e-46 ], [ 468.625 , 7.874700140663797e-46 ], [ 468.6875 , 7.876343516982579e-46 ], 
                    [ 468.75 , 7.877987236258639e-46 ], [ 468.8125 , 7.879631298563549e-46 ], [ 468.875 , 7.881275703968896e-46 ], [ 468.9375 , 7.882920452546226e-46 ], 
                    [ 469.0 , 7.884565544367325e-46 ], [ 469.0625 , 7.886210979503652e-46 ], [ 469.125 , 7.887856758026859e-46 ], [ 469.1875 , 7.889502880008774e-46 ], 
                    [ 469.25 , 7.891149345520904e-46 ], [ 469.3125 , 7.892796154634998e-46 ], [ 469.375 , 7.894443307422875e-46 ], [ 469.4375 , 7.896090434813604e-46 ], 
                    [ 469.5 , 7.897737249477604e-46 ], [ 469.5625 , 7.89938440760253e-46 ], [ 469.625 , 7.901031909260018e-46 ], [ 469.6875 , 7.902679754521767e-46 ], 
                    [ 469.75 , 7.904327943459272e-46 ], [ 469.8125 , 7.905976476144437e-46 ], [ 469.875 , 7.907625352648784e-46 ], [ 469.9375 , 7.909274573044076e-46 ], 
                    [ 470.0 , 7.910924137402092e-46 ], [ 470.0625 , 7.912574045794399e-46 ], [ 470.125 , 7.914224298292978e-46 ], [ 470.1875 , 7.915874894969367e-46 ], 
                    [ 470.25 , 7.917525835895575e-46 ], [ 470.3125 , 7.919177121143231e-46 ], [ 470.375 , 7.920828750784255e-46 ], [ 470.4375 , 7.922480724890312e-46 ], 
                    [ 470.5 , 7.92413295308165e-46 ], [ 470.5625 , 7.925782789106745e-46 ], [ 470.625 , 7.927432968634324e-46 ], [ 470.6875 , 7.929083491735792e-46 ], 
                    [ 470.75 , 7.9307343584828e-46 ], [ 470.8125 , 7.932385568946777e-46 ], [ 470.875 , 7.934037123199404e-46 ], [ 470.9375 , 7.935689021312199e-46 ], 
                    [ 471.0 , 7.937341263356645e-46 ], [ 471.0625 , 7.938993849404632e-46 ], [ 471.125 , 7.9406467795275e-46 ], [ 471.1875 , 7.942300053797111e-46 ], 
                    [ 471.25 , 7.943953672284893e-46 ], [ 471.3125 , 7.94560763506263e-46 ], [ 471.375 , 7.947261942202111e-46 ], [ 471.4375 , 7.948916593774813e-46 ], 
                    [ 471.5 , 7.950571589852502e-46 ], [ 471.5625 , 7.952226930507075e-46 ], [ 471.625 , 7.953881605557891e-46 ], [ 471.6875 , 7.95553613170651e-46 ], 
                    [ 471.75 , 7.957191002021332e-46 ], [ 471.8125 , 7.958846216573779e-46 ], [ 471.875 , 7.960501775435626e-46 ], [ 471.9375 , 7.962157678678437e-46 ], 
                    [ 472.0 , 7.963813926373909e-46 ], [ 472.0625 , 7.965470518593634e-46 ], [ 472.125 , 7.967127455409165e-46 ], [ 472.1875 , 7.968784736892409e-46 ], 
                    [ 472.25 , 7.970442363114838e-46 ], [ 472.3125 , 7.972100334148386e-46 ], [ 472.375 , 7.973758650064613e-46 ], [ 472.4375 , 7.975417310935312e-46 ], 
                    [ 472.5 , 7.977076316832296e-46 ], [ 472.5625 , 7.978735667827227e-46 ], [ 472.625 , 7.980395363991943e-46 ], [ 472.6875 , 7.982054733638756e-46 ], 
                    [ 472.75 , 7.983712720579238e-46 ], [ 472.8125 , 7.985371051907266e-46 ], [ 472.875 , 7.987029727694428e-46 ], [ 472.9375 , 7.98868874801228e-46 ], 
                    [ 473.0 , 7.99034811293238e-46 ], [ 473.0625 , 7.992007822526366e-46 ], [ 473.125 , 7.993667876865718e-46 ], [ 473.1875 , 7.995328276022044e-46 ], 
                    [ 473.25 , 7.996989020067195e-46 ], [ 473.3125 , 7.998650109072525e-46 ], [ 473.375 , 8.000311543109745e-46 ], [ 473.4375 , 8.001973322250635e-46 ], 
                    [ 473.5 , 8.003635446566878e-46 ], [ 473.5625 , 8.005297916130118e-46 ], [ 473.625 , 8.00696073101195e-46 ], [ 473.6875 , 8.008623891284389e-46 ], 
                    [ 473.75 , 8.010287397018835e-46 ], [ 473.8125 , 8.011948685634884e-46 ], [ 473.875 , 8.013609830567454e-46 ], [ 473.9375 , 8.015271319910925e-46 ], 
                    [ 474.0 , 8.016933153736654e-46 ], [ 474.0625 , 8.01859533211617e-46 ], [ 474.125 , 8.02025785512086e-46 ], [ 474.1875 , 8.021920722822118e-46 ], 
                    [ 474.25 , 8.02358393529158e-46 ], [ 474.3125 , 8.025247492600445e-46 ], [ 474.375 , 8.026911394820551e-46 ], [ 474.4375 , 8.028575642023125e-46 ], 
                    [ 474.5 , 8.030240234279976e-46 ], [ 474.5625 , 8.031905171662307e-46 ], [ 474.625 , 8.033570454241843e-46 ], [ 474.6875 , 8.035236082090155e-46 ], 
                    [ 474.75 , 8.036902055278886e-46 ], [ 474.8125 , 8.03856837387958e-46 ], [ 474.875 , 8.040234028927946e-46 ], [ 474.9375 , 8.04189817118528e-46 ], 
                    [ 475.0 , 8.043562657881468e-46 ], [ 475.0625 , 8.045227489087858e-46 ], [ 475.125 , 8.046892664875755e-46 ], [ 475.1875 , 8.04855818531648e-46 ], 
                    [ 475.25 , 8.050224050481369e-46 ], [ 475.3125 , 8.051890260441826e-46 ], [ 475.375 , 8.053556815269049e-46 ], [ 475.4375 , 8.055223715034528e-46 ], 
                    [ 475.5 , 8.056890959809829e-46 ], [ 475.5625 , 8.058558549666017e-46 ], [ 475.625 , 8.060226484674807e-46 ], [ 475.6875 , 8.061894764907461e-46 ], 
                    [ 475.75 , 8.06356339043549e-46 ], [ 475.8125 , 8.065232361330308e-46 ], [ 475.875 , 8.066901677663565e-46 ], [ 475.9375 , 8.068571339506592e-46 ], 
                    [ 476.0 , 8.070241168442006e-46 ], [ 476.0625 , 8.071911306907853e-46 ], [ 476.125 , 8.073581791009283e-46 ], [ 476.1875 , 8.075252620817936e-46 ], 
                    [ 476.25 , 8.076923796405131e-46 ], [ 476.3125 , 8.078595317842596e-46 ], [ 476.375 , 8.080267185201733e-46 ], [ 476.4375 , 8.081939398554363e-46 ], 
                    [ 476.5 , 8.083611957971913e-46 ], [ 476.5625 , 8.085284863526176e-46 ], [ 476.625 , 8.086958115288555e-46 ], [ 476.6875 , 8.088631713330869e-46 ], 
                    [ 476.75 , 8.09030565772461e-46 ], [ 476.8125 , 8.091979948541685e-46 ], [ 476.875 , 8.093654585853555e-46 ], [ 476.9375 , 8.09532956973216e-46 ], 
                    [ 477.0 , 8.097004900249047e-46 ], [ 477.0625 , 8.098678969867213e-46 ], [ 477.125 , 8.100349250897635e-46 ], [ 477.1875 , 8.102019876408656e-46 ], 
                    [ 477.25 , 8.103690846471498e-46 ], [ 477.3125 , 8.105362161157216e-46 ], [ 477.375 , 8.107033820536717e-46 ], [ 477.4375 , 8.108705824681206e-46 ], 
                    [ 477.5 , 8.110378173661902e-46 ], [ 477.5625 , 8.112050867549751e-46 ], [ 477.625 , 8.113723906416005e-46 ], [ 477.6875 , 8.115397290331576e-46 ], 
                    [ 477.75 , 8.11707101936798e-46 ], [ 477.8125 , 8.118745093596162e-46 ], [ 477.875 , 8.120419513087315e-46 ], [ 477.9375 , 8.122094277912818e-46 ], 
                    [ 478.0 , 8.123769388143665e-46 ], [ 478.0625 , 8.125444843851205e-46 ], [ 478.125 , 8.127120645106808e-46 ], [ 478.1875 , 8.128796934906258e-46 ], 
                    [ 478.25 , 8.13047363772768e-46 ], [ 478.3125 , 8.132150686397617e-46 ], [ 478.375 , 8.13382808098735e-46 ], [ 478.4375 , 8.135505821568402e-46 ], 
                    [ 478.5 , 8.137183908211967e-46 ], [ 478.5625 , 8.138862340989426e-46 ], [ 478.625 , 8.14054111997235e-46 ], [ 478.6875 , 8.142220245231973e-46 ], 
                    [ 478.75 , 8.143899716839668e-46 ], [ 478.8125 , 8.145579534867159e-46 ], [ 478.875 , 8.147259699385672e-46 ], [ 478.9375 , 8.148940210466791e-46 ], 
                    [ 479.0 , 8.150621068181945e-46 ], [ 479.0625 , 8.152302272602459e-46 ], [ 479.125 , 8.153983823800193e-46 ], [ 479.1875 , 8.155665721846387e-46 ], 
                    [ 479.25 , 8.157347873615278e-46 ], [ 479.3125 , 8.159027552530032e-46 ], [ 479.375 , 8.160707577307406e-46 ], [ 479.4375 , 8.162387948018563e-46 ], 
                    [ 479.5 , 8.164068664734731e-46 ], [ 479.5625 , 8.165749727527218e-46 ], [ 479.625 , 8.167431136467164e-46 ], [ 479.6875 , 8.169112891625904e-46 ], 
                    [ 479.75 , 8.170794993074787e-46 ], [ 479.8125 , 8.172477440885e-46 ], [ 479.875 , 8.174160235127803e-46 ], [ 479.9375 , 8.175843375874766e-46 ], 
                    [ 480.0 , 8.177526863197236e-46 ], [ 480.0625 , 8.179210697166341e-46 ], [ 480.125 , 8.180894877853463e-46 ], [ 480.1875 , 8.182579405330284e-46 ], 
                    [ 480.25 , 8.184264279668094e-46 ], [ 480.3125 , 8.185949500938143e-46 ], [ 480.375 , 8.187633507099136e-46 ], [ 480.4375 , 8.189314983393899e-46 ], 
                    [ 480.5 , 8.190996805009812e-46 ], [ 480.5625 , 8.192678972017675e-46 ], [ 480.625 , 8.194361484488653e-46 ], [ 480.6875 , 8.196044342493345e-46 ], 
                    [ 480.75 , 8.197727546102943e-46 ], [ 480.8125 , 8.199411095388367e-46 ], [ 480.875 , 8.20109499042055e-46 ], [ 480.9375 , 8.202779231270667e-46 ], 
                    [ 481.0 , 8.204463818009452e-46 ], [ 481.0625 , 8.206148750708347e-46 ], [ 481.125 , 8.207834029437932e-46 ], [ 481.1875 , 8.209519654269679e-46 ], 
                    [ 481.25 , 8.211205625274435e-46 ], [ 481.3125 , 8.212891942523291e-46 ], [ 481.375 , 8.214578606087415e-46 ], [ 481.4375 , 8.216265616037983e-46 ], 
                    [ 481.5 , 8.217951978874506e-46 ], [ 481.5625 , 8.219638089018344e-46 ], [ 481.625 , 8.221324545108213e-46 ], [ 481.6875 , 8.223011347214865e-46 ], 
                    [ 481.75 , 8.224698495409581e-46 ], [ 481.8125 , 8.226385989763255e-46 ], [ 481.875 , 8.228073830346907e-46 ], [ 481.9375 , 8.229762017231519e-46 ], 
                    [ 482.0 , 8.231450550488258e-46 ], [ 482.0625 , 8.233139430188016e-46 ], [ 482.125 , 8.234828656402108e-46 ], [ 482.1875 , 8.23651822920157e-46 ], 
                    [ 482.25 , 8.238208148657338e-46 ], [ 482.3125 , 8.239898414840772e-46 ], [ 482.375 , 8.241589027822775e-46 ], [ 482.4375 , 8.243279987674737e-46 ], 
                    [ 482.5 , 8.244971294467533e-46 ], [ 482.5625 , 8.246662948272639e-46 ], [ 482.625 , 8.248353937627992e-46 ], [ 482.6875 , 8.250045114467606e-46 ], 
                    [ 482.75 , 8.251736638052718e-46 ], [ 482.8125 , 8.253428508454303e-46 ], [ 482.875 , 8.255120725743586e-46 ], [ 482.9375 , 8.256813289991576e-46 ], 
                    [ 483.0 , 8.258506201269468e-46 ], [ 483.0625 , 8.260199459648295e-46 ], [ 483.125 , 8.261893065199521e-46 ], [ 483.1875 , 8.263587017994148e-46 ], 
                    [ 483.25 , 8.265281318103256e-46 ], [ 483.3125 , 8.266975965598289e-46 ], [ 483.375 , 8.268670960550298e-46 ], [ 483.4375 , 8.270366303030522e-46 ], 
                    [ 483.5 , 8.272061993110334e-46 ], [ 483.5625 , 8.273758030861002e-46 ], [ 483.625 , 8.27545441635375e-46 ], [ 483.6875 , 8.277151058690105e-46 ], 
                    [ 483.75 , 8.278846515071372e-46 ], [ 483.8125 , 8.280542318742705e-46 ], [ 483.875 , 8.282238469775124e-46 ], [ 483.9375 , 8.283934968239957e-46 ], 
                    [ 484.0 , 8.28563181420843e-46 ], [ 484.0625 , 8.287329007751491e-46 ], [ 484.125 , 8.289026548940449e-46 ], [ 484.1875 , 8.290724437846518e-46 ], 
                    [ 484.25 , 8.292422674541037e-46 ], [ 484.3125 , 8.294121259094956e-46 ], [ 484.375 , 8.29582019157982e-46 ], [ 484.4375 , 8.29751947206678e-46 ], 
                    [ 484.5 , 8.29921910062706e-46 ], [ 484.5625 , 8.30091907733202e-46 ], [ 484.625 , 8.30261940225291e-46 ], [ 484.6875 , 8.304320075461353e-46 ], 
                    [ 484.75 , 8.30602109702822e-46 ], [ 484.8125 , 8.307721498444604e-46 ], [ 484.875 , 8.309418731658122e-46 ], [ 484.9375 , 8.311116311609439e-46 ], 
                    [ 485.0 , 8.312814238369214e-46 ], [ 485.0625 , 8.314512512008419e-46 ], [ 485.125 , 8.316211132598037e-46 ], [ 485.1875 , 8.317910100208773e-46 ], 
                    [ 485.25 , 8.319609414911579e-46 ], [ 485.3125 , 8.321309076777424e-46 ], [ 485.375 , 8.323009085877054e-46 ], [ 485.4375 , 8.324709442281588e-46 ], 
                    [ 485.5 , 8.326410146061974e-46 ], [ 485.5625 , 8.328111197289007e-46 ], [ 485.625 , 8.329812596033783e-46 ], [ 485.6875 , 8.33151434236748e-46 ], 
                    [ 485.75 , 8.33321643636075e-46 ], [ 485.8125 , 8.334918878084857e-46 ], [ 485.875 , 8.336621667610841e-46 ], [ 485.9375 , 8.338323334662733e-46 ], 
                    [ 486.0 , 8.340022336026258e-46 ], [ 486.0625 , 8.341721683575103e-46 ], [ 486.125 , 8.343421377379925e-46 ], [ 486.1875 , 8.34512141751116e-46 ], 
                    [ 486.25 , 8.34682180403943e-46 ], [ 486.3125 , 8.34852253703532e-46 ], [ 486.375 , 8.350223616569483e-46 ], [ 486.4375 , 8.351925042712412e-46 ], 
                    [ 486.5 , 8.353626815534849e-46 ], [ 486.5625 , 8.355328935107315e-46 ], [ 486.625 , 8.357031401500519e-46 ], [ 486.6875 , 8.358734214785132e-46 ], 
                    [ 486.75 , 8.360437375031897e-46 ], [ 486.8125 , 8.362140882311328e-46 ], [ 486.875 , 8.363844736694374e-46 ], [ 486.9375 , 8.365548938251643e-46 ], 
                    [ 487.0 , 8.367253487053816e-46 ], [ 487.0625 , 8.36895902342072e-46 ], [ 487.125 , 8.370665900676732e-46 ], [ 487.1875 , 8.372373126056086e-46 ], 
                    [ 487.25 , 8.374080699629839e-46 ], [ 487.3125 , 8.37578862146883e-46 ], [ 487.375 , 8.377496891644385e-46 ], [ 487.4375 , 8.379205510227252e-46 ], 
                    [ 487.5 , 8.380914477288666e-46 ], [ 487.5625 , 8.382623792899645e-46 ], [ 487.625 , 8.384333457131096e-46 ], [ 487.6875 , 8.386043470054538e-46 ], 
                    [ 487.75 , 8.38775383174079e-46 ], [ 487.8125 , 8.389464542260924e-46 ], [ 487.875 , 8.391175601686324e-46 ], [ 487.9375 , 8.392887010088031e-46 ], 
                    [ 488.0 , 8.394598767537221e-46 ], [ 488.0625 , 8.396310874104965e-46 ], [ 488.125 , 8.398023329862701e-46 ], [ 488.1875 , 8.39973486369967e-46 ], 
                    [ 488.25 , 8.401445019110464e-46 ], [ 488.3125 , 8.403155522702608e-46 ], [ 488.375 , 8.40486637454705e-46 ], [ 488.4375 , 8.406577574714634e-46 ], 
                    [ 488.5 , 8.408289123276396e-46 ], [ 488.5625 , 8.41000102030333e-46 ], [ 488.625 , 8.411713265866139e-46 ], [ 488.6875 , 8.413425860035788e-46 ], 
                    [ 488.75 , 8.415138802883606e-46 ], [ 488.8125 , 8.416852094480229e-46 ], [ 488.875 , 8.418565734896835e-46 ], [ 488.9375 , 8.420279724204387e-46 ], 
                    [ 489.0 , 8.421994062473797e-46 ], [ 489.0625 , 8.423708749776232e-46 ], [ 489.125 , 8.425423786182873e-46 ], [ 489.1875 , 8.427139171764618e-46 ], 
                    [ 489.25 , 8.428854906592677e-46 ], [ 489.3125 , 8.430568430013231e-46 ], [ 489.375 , 8.432278822354015e-46 ], [ 489.4375 , 8.433989561698977e-46 ], 
                    [ 489.5 , 8.435700648118336e-46 ], [ 489.5625 , 8.437412081682687e-46 ], [ 489.625 , 8.4391238624624e-46 ], [ 489.6875 , 8.440835990527856e-46 ], 
                    [ 489.75 , 8.442548465949633e-46 ], [ 489.8125 , 8.444261288798085e-46 ], [ 489.875 , 8.445974459143813e-46 ], [ 489.9375 , 8.447687977057262e-46 ], 
                    [ 490.0 , 8.449401842608942e-46 ], [ 490.0625 , 8.451116055869387e-46 ], [ 490.125 , 8.452830616909194e-46 ], [ 490.1875 , 8.454545525798806e-46 ], 
                    [ 490.25 , 8.456260782608792e-46 ], [ 490.3125 , 8.457976387409858e-46 ], [ 490.375 , 8.459692340272546e-46 ], [ 490.4375 , 8.461408038555897e-46 ], 
                    [ 490.5 , 8.463123013093273e-46 ], [ 490.5625 , 8.464838335225004e-46 ], [ 490.625 , 8.466554005021539e-46 ], [ 490.6875 , 8.468270022553345e-46 ], 
                    [ 490.75 , 8.469986387890902e-46 ], [ 490.8125 , 8.471703101104643e-46 ], [ 490.875 , 8.473420162265257e-46 ], [ 490.9375 , 8.475137571443085e-46 ], 
                    [ 491.0 , 8.476855328708667e-46 ], [ 491.0625 , 8.478573434132671e-46 ], [ 491.125 , 8.480291887785482e-46 ], [ 491.1875 , 8.482010689737984e-46 ], 
                    [ 491.25 , 8.483729840060406e-46 ], [ 491.3125 , 8.48544933882378e-46 ], [ 491.375 , 8.487169186098366e-46 ], [ 491.4375 , 8.488889381954862e-46 ], 
                    [ 491.5 , 8.490609926464041e-46 ], [ 491.5625 , 8.492330886989683e-46 ], [ 491.625 , 8.494052383979837e-46 ], [ 491.6875 , 8.49577422993795e-46 ], 
                    [ 491.75 , 8.49749642493476e-46 ], [ 491.8125 , 8.499218969041144e-46 ], [ 491.875 , 8.500941862327871e-46 ], [ 491.9375 , 8.502665104865662e-46 ], 
                    [ 492.0 , 8.504388696725257e-46 ], [ 492.0625 , 8.506112637977525e-46 ], [ 492.125 , 8.507836928693293e-46 ], [ 492.1875 , 8.509561568943341e-46 ], 
                    [ 492.25 , 8.511286558798767e-46 ], [ 492.3125 , 8.513011898330136e-46 ], [ 492.375 , 8.514737587608513e-46 ], [ 492.4375 , 8.516463626704676e-46 ], 
                    [ 492.5 , 8.518190015689598e-46 ], [ 492.5625 , 8.519916754634144e-46 ], [ 492.625 , 8.521643843609315e-46 ], [ 492.6875 , 8.52337084344719e-46 ], 
                    [ 492.75 , 8.525095402254801e-46 ], [ 492.8125 , 8.526820309997432e-46 ], [ 492.875 , 8.528545566745801e-46 ], [ 492.9375 , 8.530271172570469e-46 ], 
                    [ 493.0 , 8.531997127542e-46 ], [ 493.0625 , 8.533723431731101e-46 ], [ 493.125 , 8.53545008520849e-46 ], [ 493.1875 , 8.537177088044718e-46 ], 
                    [ 493.25 , 8.538904440310532e-46 ], [ 493.3125 , 8.540632142076636e-46 ], [ 493.375 , 8.542360193413863e-46 ], [ 493.4375 , 8.544088594392701e-46 ], 
                    [ 493.5 , 8.545817345084078e-46 ], [ 493.5625 , 8.54754644555863e-46 ], [ 493.625 , 8.549275895887191e-46 ], [ 493.6875 , 8.551005696140668e-46 ], 
                    [ 493.75 , 8.55273584638962e-46 ], [ 493.8125 , 8.554466346704982e-46 ], [ 493.875 , 8.556192817380725e-46 ], [ 493.9375 , 8.557919453966834e-46 ], 
                    [ 494.0 , 8.559646438987695e-46 ], [ 494.0625 , 8.561373772513505e-46 ], [ 494.125 , 8.563101454614468e-46 ], [ 494.1875 , 8.564829485361108e-46 ], 
                    [ 494.25 , 8.56655786482366e-46 ], [ 494.3125 , 8.568286593072741e-46 ], [ 494.375 , 8.570015670178306e-46 ], [ 494.4375 , 8.571745096211244e-46 ], 
                    [ 494.5 , 8.57347487124148e-46 ], [ 494.5625 , 8.575204995339869e-46 ], [ 494.625 , 8.576935468576607e-46 ], [ 494.6875 , 8.578666291022214e-46 ], 
                    [ 494.75 , 8.580397462747158e-46 ], [ 494.8125 , 8.582128983821984e-46 ], [ 494.875 , 8.583860854316948e-46 ], [ 494.9375 , 8.585593074302991e-46 ], 
                    [ 495.0 , 8.587325803080826e-46 ], [ 495.0625 , 8.589058938681638e-46 ], [ 495.125 , 8.590792424072254e-46 ], [ 495.1875 , 8.59252625932327e-46 ], 
                    [ 495.25 , 8.594260444505296e-46 ], [ 495.3125 , 8.59599497968902e-46 ], [ 495.375 , 8.597729864944895e-46 ], [ 495.4375 , 8.599465100343884e-46 ], 
                    [ 495.5 , 8.601200685956283e-46 ], [ 495.5625 , 8.602936621853083e-46 ], [ 495.625 , 8.604672908104733e-46 ], [ 495.6875 , 8.60640954478219e-46 ], 
                    [ 495.75 , 8.608146531956055e-46 ], [ 495.8125 , 8.609883869696944e-46 ], [ 495.875 , 8.611621558075796e-46 ], [ 495.9375 , 8.613359597163253e-46 ], 
                    [ 496.0 , 8.615097987030223e-46 ], [ 496.0625 , 8.616836727747559e-46 ], [ 496.125 , 8.618572688590361e-46 ], [ 496.1875 , 8.620305606785511e-46 ], 
                    [ 496.25 , 8.622038873414797e-46 ], [ 496.3125 , 8.623772488548466e-46 ], [ 496.375 , 8.625506452256281e-46 ], [ 496.4375 , 8.627240764608636e-46 ], 
                    [ 496.5 , 8.628975425675512e-46 ], [ 496.5625 , 8.63071043552702e-46 ], [ 496.625 , 8.632445794233172e-46 ], [ 496.6875 , 8.634181501864352e-46 ], 
                    [ 496.75 , 8.635917558490537e-46 ], [ 496.8125 , 8.637653964182019e-46 ], [ 496.875 , 8.639390719008924e-46 ], [ 496.9375 , 8.64112782304145e-46 ], 
                    [ 497.0 , 8.642865276349753e-46 ], [ 497.0625 , 8.644603079004242e-46 ], [ 497.125 , 8.646341231074855e-46 ], [ 497.1875 , 8.648079732632215e-46 ], 
                    [ 497.25 , 8.64981835205559e-46 ], [ 497.3125 , 8.651556236061164e-46 ], [ 497.375 , 8.653294469234861e-46 ], [ 497.4375 , 8.65503305164677e-46 ], 
                    [ 497.5 , 8.656771983367302e-46 ], [ 497.5625 , 8.658511264466275e-46 ], [ 497.625 , 8.66025089501419e-46 ], [ 497.6875 , 8.661990875081008e-46 ], 
                    [ 497.75 , 8.663731204737204e-46 ], [ 497.8125 , 8.665471884052763e-46 ], [ 497.875 , 8.66721291309819e-46 ], [ 497.9375 , 8.668954291943624e-46 ], 
                    [ 498.0 , 8.670696020659225e-46 ], [ 498.0625 , 8.67243809931547e-46 ], [ 498.125 , 8.674180527982544e-46 ], [ 498.1875 , 8.675923306730956e-46 ], 
                    [ 498.25 , 8.677666435630857e-46 ], [ 498.3125 , 8.67940991475266e-46 ], [ 498.375 , 8.681153744166855e-46 ], [ 498.4375 , 8.682898070652616e-46 ], 
                    [ 498.5 , 8.684642778972656e-46 ], [ 498.5625 , 8.686387837867712e-46 ], [ 498.625 , 8.688133247408287e-46 ], [ 498.6875 , 8.689879007664715e-46 ], 
                    [ 498.75 , 8.691625118707532e-46 ], [ 498.8125 , 8.693371580607221e-46 ], [ 498.875 , 8.695118393434345e-46 ], [ 498.9375 , 8.696865557259231e-46 ], 
                    [ 499.0 , 8.698613072152655e-46 ], [ 499.0625 , 8.700360938184914e-46 ], [ 499.125 , 8.702109155426871e-46 ], [ 499.1875 , 8.703857723948792e-46 ], 
                    [ 499.25 , 8.705606643821378e-46 ], [ 499.3125 , 8.707355915115234e-46 ], [ 499.375 , 8.709105537901033e-46 ], [ 499.4375 , 8.710855512249281e-46 ], 
                    [ 499.5 , 8.712605838230679e-46 ], [ 499.5625 , 8.714353633833296e-46 ], [ 499.625 , 8.716098111213282e-46 ], [ 499.6875 , 8.717842937810432e-46 ], 
                    [ 499.75 , 8.719588113694654e-46 ], [ 499.8125 , 8.72133363893581e-46 ], [ 499.875 , 8.72307951360377e-46 ], [ 499.9375 , 8.724825737768612e-46 ], 
                    [ 500.0 , 8.726572311500236e-46 ], [ 500.0625 , 8.728319234868557e-46 ], [ 500.125 , 8.730066507943691e-46 ], [ 500.1875 , 8.731814130795583e-46 ], 
                    [ 500.25 , 8.733562103494128e-46 ], [ 500.3125 , 8.73531042610961e-46 ], [ 500.375 , 8.737059098711947e-46 ], [ 500.4375 , 8.738808121371144e-46 ], 
                    [ 500.5 , 8.740557494157403e-46 ], [ 500.5625 , 8.742307217140683e-46 ], [ 500.625 , 8.744057290391217e-46 ], [ 500.6875 , 8.745807713736144e-46 ], 
                    [ 500.75 , 8.747558457373066e-46 ], [ 500.8125 , 8.749309551475361e-46 ], [ 500.875 , 8.75106099611318e-46 ], [ 500.9375 , 8.752812791356886e-46 ], 
                    [ 501.0 , 8.754564937276347e-46 ], [ 501.0625 , 8.756317433942075e-46 ], [ 501.125 , 8.758070281424159e-46 ], [ 501.1875 , 8.759823479792763e-46 ], 
                    [ 501.25 , 8.761577029118252e-46 ], [ 501.3125 , 8.76333092947088e-46 ], [ 501.375 , 8.765085180920605e-46 ], [ 501.4375 , 8.766839783538273e-46 ], 
                    [ 501.5 , 8.768594737393741e-46 ], [ 501.5625 , 8.770350042557325e-46 ], [ 501.625 , 8.772105699099657e-46 ], [ 501.6875 , 8.773861707090765e-46 ], 
                    [ 501.75 , 8.775618066601255e-46 ], [ 501.8125 , 8.777374777701243e-46 ], [ 501.875 , 8.77913051243856e-46 ], [ 501.9375 , 8.780885410965808e-46 ], 
                    [ 502.0 , 8.782640660287156e-46 ], [ 502.0625 , 8.784396260473102e-46 ], [ 502.125 , 8.786152211593404e-46 ], [ 502.1875 , 8.787908513718592e-46 ], 
                    [ 502.25 , 8.789665166918514e-46 ], [ 502.3125 , 8.791422171263474e-46 ], [ 502.375 , 8.793179526823726e-46 ], [ 502.4375 , 8.794937233669413e-46 ], 
                    [ 502.5 , 8.796695291870692e-46 ], [ 502.5625 , 8.798453701497924e-46 ], [ 502.625 , 8.80021246262123e-46 ], [ 502.6875 , 8.801971575310999e-46 ], 
                    [ 502.75 , 8.803731039637443e-46 ], [ 502.8125 , 8.805490855670728e-46 ], [ 502.875 , 8.80725102348147e-46 ], [ 502.9375 , 8.809011543139675e-46 ], 
                    [ 503.0 , 8.810772351954551e-46 ], [ 503.0625 , 8.812529652142767e-46 ], [ 503.125 , 8.814287302822837e-46 ], [ 503.1875 , 8.816045304064609e-46 ], 
                    [ 503.25 , 8.81780365593812e-46 ], [ 503.3125 , 8.819562358513183e-46 ], [ 503.375 , 8.82132141185987e-46 ], [ 503.4375 , 8.823080816048016e-46 ], 
                    [ 503.5 , 8.824840571147658e-46 ], [ 503.5625 , 8.826600677228847e-46 ], [ 503.625 , 8.828361134361527e-46 ], [ 503.6875 , 8.830121942615584e-46 ], 
                    [ 503.75 , 8.831883102061303e-46 ], [ 503.8125 , 8.833644612768541e-46 ], [ 503.875 , 8.835406474807544e-46 ], [ 503.9375 , 8.837168688248008e-46 ], 
                    [ 504.0 , 8.838931253160586e-46 ], [ 504.0625 , 8.840694169614937e-46 ], [ 504.125 , 8.84245743768124e-46 ], [ 504.1875 , 8.844219815992747e-46 ], 
                    [ 504.25 , 8.845981064254365e-46 ], [ 504.3125 , 8.847742663252953e-46 ], [ 504.375 , 8.849504613058423e-46 ], [ 504.4375 , 8.851266913740635e-46 ], 
                    [ 504.5 , 8.853029565369526e-46 ], [ 504.5625 , 8.854792568014794e-46 ], [ 504.625 , 8.85655592174653e-46 ], [ 504.6875 , 8.858319626634589e-46 ], 
                    [ 504.75 , 8.86008368274896e-46 ], [ 504.8125 , 8.861848090159338e-46 ], [ 504.875 , 8.863612848936182e-46 ], [ 504.9375 , 8.865377959148897e-46 ], 
                    [ 505.0 , 8.86714342086766e-46 ], [ 505.0625 , 8.868909234162782e-46 ], [ 505.125 , 8.870675399103897e-46 ], [ 505.1875 , 8.872441915761101e-46 ], 
                    [ 505.25 , 8.874208784204683e-46 ], [ 505.3125 , 8.875976004504321e-46 ], [ 505.375 , 8.877742032699618e-46 ], [ 505.4375 , 8.879508135461306e-46 ], 
                    [ 505.5 , 8.881274589564404e-46 ], [ 505.5625 , 8.883041395078996e-46 ], [ 505.625 , 8.884808552074801e-46 ], [ 505.6875 , 8.886576060621868e-46 ], 
                    [ 505.75 , 8.888343920789944e-46 ], [ 505.8125 , 8.890112132649292e-46 ], [ 505.875 , 8.891880696269438e-46 ], [ 505.9375 , 8.8936496117208e-46 ], 
                    [ 506.0 , 8.895418879073052e-46 ], [ 506.0625 , 8.897188498396585e-46 ], [ 506.125 , 8.898958469760842e-46 ], [ 506.1875 , 8.900728793236368e-46 ], 
                    [ 506.25 , 8.902499468892894e-46 ], [ 506.3125 , 8.904270496800605e-46 ], [ 506.375 , 8.90604187702977e-46 ], [ 506.4375 , 8.907813609650159e-46 ], 
                    [ 506.5 , 8.909585397788461e-46 ], [ 506.5625 , 8.911356407445834e-46 ], [ 506.625 , 8.913127769136945e-46 ], [ 506.6875 , 8.914899482931831e-46 ], 
                    [ 506.75 , 8.916671548900546e-46 ], [ 506.8125 , 8.918443967113033e-46 ], [ 506.875 , 8.920216737639177e-46 ], [ 506.9375 , 8.921989860549267e-46 ], 
                    [ 507.0 , 8.923763335913158e-46 ], [ 507.0625 , 8.925537163801035e-46 ], [ 507.125 , 8.927311344282845e-46 ], [ 507.1875 , 8.929085877428737e-46 ], 
                    [ 507.25 , 8.930860763308941e-46 ], [ 507.3125 , 8.932636001993319e-46 ], [ 507.375 , 8.934411593552187e-46 ], [ 507.4375 , 8.9361875380555e-46 ], 
                    [ 507.5 , 8.937963835573604e-46 ], [ 507.5625 , 8.939740486176671e-46 ], [ 507.625 , 8.941517489934821e-46 ], [ 507.6875 , 8.943292196503488e-46 ], 
                    [ 507.75 , 8.945064808293108e-46 ], [ 507.8125 , 8.946837771424613e-46 ], [ 507.875 , 8.948611085967512e-46 ], [ 507.9375 , 8.950384751991583e-46 ], 
                    [ 508.0 , 8.95215876956643e-46 ], [ 508.0625 , 8.953933138761668e-46 ], [ 508.125 , 8.955707859647118e-46 ], [ 508.1875 , 8.957482932292486e-46 ], 
                    [ 508.25 , 8.959258356767365e-46 ], [ 508.3125 , 8.961034133141492e-46 ], [ 508.375 , 8.962810261484742e-46 ], [ 508.4375 , 8.964586741866684e-46 ], 
                    [ 508.5 , 8.966363574357414e-46 ], [ 508.5625 , 8.96814075902647e-46 ], [ 508.625 , 8.969918295943652e-46 ], [ 508.6875 , 8.971696185178904e-46 ], 
                    [ 508.75 , 8.973474426802062e-46 ], [ 508.8125 , 8.975253020882841e-46 ], [ 508.875 , 8.977029601794418e-46 ], [ 508.9375 , 8.978805942825249e-46 ], 
                    [ 509.0 , 8.980582635351733e-46 ], [ 509.0625 , 8.98235967944355e-46 ], [ 509.125 , 8.984137075170204e-46 ], [ 509.1875 , 8.985914822601274e-46 ], 
                    [ 509.25 , 8.98769292180629e-46 ], [ 509.3125 , 8.98947137285499e-46 ], [ 509.375 , 8.991250175816927e-46 ], [ 509.4375 , 8.993029330761804e-46 ], 
                    [ 509.5 , 8.994808837759077e-46 ], [ 509.5625 , 8.996588696878601e-46 ], [ 509.625 , 8.99836890818986e-46 ], [ 509.6875 , 9.000149471762865e-46 ], 
                    [ 509.75 , 9.001930387667129e-46 ], [ 509.8125 , 9.00371165597224e-46 ], [ 509.875 , 9.005493276748058e-46 ], [ 509.9375 , 9.007275250064394e-46 ], 
                    [ 510.0 , 9.009057576869653e-46 ], [ 510.0625 , 9.010840310409085e-46 ], [ 510.125 , 9.012623396719968e-46 ], [ 510.1875 , 9.014406835872234e-46 ], 
                    [ 510.25 , 9.016190627935514e-46 ], [ 510.3125 , 9.017974772980026e-46 ], [ 510.375 , 9.019759271075108e-46 ], [ 510.4375 , 9.021544122291004e-46 ], 
                    [ 510.5 , 9.0233293266974e-46 ], [ 510.5625 , 9.025114884364316e-46 ], [ 510.625 , 9.026900795361525e-46 ], [ 510.6875 , 9.028687059759073e-46 ], 
                    [ 510.75 , 9.030473677626829e-46 ], [ 510.8125 , 9.032260649034738e-46 ], [ 510.875 , 9.034047974052761e-46 ], [ 510.9375 , 9.03583565275074e-46 ], 
                    [ 511.0 , 9.03762368519892e-46 ], [ 511.0625 , 9.03941207146711e-46 ], [ 511.125 , 9.041200811625387e-46 ], [ 511.1875 , 9.042988578053132e-46 ], 
                    [ 511.25 , 9.044771386174746e-46 ], [ 511.3125 , 9.046554545773677e-46 ], [ 511.375 , 9.048338056919157e-46 ], [ 511.4375 , 9.050121919680491e-46 ], 
                    [ 511.5 , 9.051906134127001e-46 ], [ 511.5625 , 9.053690700327954e-46 ], [ 511.625 , 9.055475618352892e-46 ], [ 511.6875 , 9.05726088827092e-46 ], 
                    [ 511.75 , 9.059046510151606e-46 ], [ 511.8125 , 9.060832484064274e-46 ], [ 511.875 , 9.06261881007826e-46 ], [ 511.9375 , 9.064405488263241e-46 ], 
                    [ 512.0 , 9.06619251868813e-46 ], [ 512.0625 , 9.067979901423016e-46 ], [ 512.125 , 9.069767636536837e-46 ], [ 512.1875 , 9.071555724099197e-46 ], 
                    [ 512.25 , 9.073344164179837e-46 ], [ 512.3125 , 9.075132956848062e-46 ], [ 512.375 , 9.076922115319611e-46 ], [ 512.4375 , 9.078711652508001e-46 ], 
                    [ 512.5 , 9.080501542507919e-46 ], [ 512.5625 , 9.082291785388987e-46 ], [ 512.625 , 9.084082381220777e-46 ], [ 512.6875 , 9.085873330072937e-46 ], 
                    [ 512.75 , 9.087664632014877e-46 ], [ 512.8125 , 9.089456287116396e-46 ], [ 512.875 , 9.091248295447127e-46 ], [ 512.9375 , 9.09304065707664e-46 ], 
                    [ 513.0 , 9.094833372074528e-46 ], [ 513.0625 , 9.096626440510587e-46 ], [ 513.125 , 9.098419862454366e-46 ], [ 513.1875 , 9.100213637975627e-46 ], 
                    [ 513.25 , 9.102007767144079e-46 ], [ 513.3125 , 9.103802250029379e-46 ], [ 513.375 , 9.105597086701329e-46 ], [ 513.4375 , 9.107392277229934e-46 ], 
                    [ 513.5 , 9.109187821684444e-46 ], [ 513.5625 , 9.110983059363566e-46 ], [ 513.625 , 9.112777809700298e-46 ], [ 513.6875 , 9.114572913580458e-46 ], 
                    [ 513.75 , 9.116368371073883e-46 ], [ 513.8125 , 9.118164182250039e-46 ], [ 513.875 , 9.119960347178657e-46 ], [ 513.9375 , 9.121756865929425e-46 ], 
                    [ 514.0 , 9.123553738572043e-46 ], [ 514.0625 , 9.12535096517622e-46 ], [ 514.125 , 9.127148545811748e-46 ], [ 514.1875 , 9.128946480548303e-46 ], 
                    [ 514.25 , 9.130744769455572e-46 ], [ 514.3125 , 9.132543412603324e-46 ], [ 514.375 , 9.134342410061403e-46 ], [ 514.4375 , 9.136141761899735e-46 ], 
                    [ 514.5 , 9.137941468187934e-46 ], [ 514.5625 , 9.139741528995884e-46 ], [ 514.625 , 9.141541944393556e-46 ], [ 514.6875 , 9.143342714450406e-46 ], 
                    [ 514.75 , 9.14514053225293e-46 ], [ 514.8125 , 9.146935342571076e-46 ], [ 514.875 , 9.148730505135825e-46 ], [ 514.9375 , 9.150526020016177e-46 ], 
                    [ 515.0 , 9.152321887281342e-46 ], [ 515.0625 , 9.154118107000541e-46 ], [ 515.125 , 9.155914679242885e-46 ], [ 515.1875 , 9.157711604077426e-46 ], 
                    [ 515.25 , 9.159508881573561e-46 ], [ 515.3125 , 9.161306511800568e-46 ], [ 515.375 , 9.163104494827347e-46 ], [ 515.4375 , 9.164902830723461e-46 ], 
                    [ 515.5 , 9.166701519557972e-46 ], [ 515.5625 , 9.168500561400211e-46 ], [ 515.625 , 9.170299956319588e-46 ], [ 515.6875 , 9.172099704385073e-46 ], 
                    [ 515.75 , 9.1738998056663e-46 ], [ 515.8125 , 9.175700260232457e-46 ], [ 515.875 , 9.177501068152819e-46 ], [ 515.9375 , 9.179302133721323e-46 ], 
                    [ 516.0 , 9.18110345847159e-46 ], [ 516.0625 , 9.18290513670967e-46 ], [ 516.125 , 9.184707168504738e-46 ], [ 516.1875 , 9.186509553926435e-46 ], 
                    [ 516.25 , 9.188312293043962e-46 ], [ 516.3125 , 9.190115385926789e-46 ], [ 516.375 , 9.19191883264434e-46 ], [ 516.4375 , 9.193722633266117e-46 ], 
                    [ 516.5 , 9.195526787861372e-46 ], [ 516.5625 , 9.197331296499698e-46 ], [ 516.625 , 9.199136159250638e-46 ], [ 516.6875 , 9.200941376183683e-46 ], 
                    [ 516.75 , 9.202746947368206e-46 ], [ 516.8125 , 9.204552872873853e-46 ], [ 516.875 , 9.206359152770027e-46 ], [ 516.9375 , 9.208165787126336e-46 ], 
                    [ 517.0 , 9.209972776012405e-46 ], [ 517.0625 , 9.211780119497676e-46 ], [ 517.125 , 9.213587192301348e-46 ], [ 517.1875 , 9.215393897069138e-46 ], 
                    [ 517.25 , 9.217200956115978e-46 ], [ 517.3125 , 9.219008369511534e-46 ], [ 517.375 , 9.220816137325227e-46 ], [ 517.4375 , 9.222624259626487e-46 ], 
                    [ 517.5 , 9.22443273648496e-46 ], [ 517.5625 , 9.226241567970105e-46 ], [ 517.625 , 9.228050754151527e-46 ], [ 517.6875 , 9.229860295098515e-46 ], 
                    [ 517.75 , 9.231670190880967e-46 ], [ 517.8125 , 9.233480441568261e-46 ], [ 517.875 , 9.23529104723026e-46 ], [ 517.9375 , 9.237102007936106e-46 ], 
                    [ 518.0 , 9.238913323755815e-46 ], [ 518.0625 , 9.240724994758894e-46 ], [ 518.125 , 9.242537021014922e-46 ], [ 518.1875 , 9.244349402593695e-46 ], 
                    [ 518.25 , 9.246162139564821e-46 ], [ 518.3125 , 9.247972015125213e-46 ], [ 518.375 , 9.249777254547096e-46 ], [ 518.4375 , 9.251582846358602e-46 ], 
                    [ 518.5 , 9.253388790628584e-46 ], [ 518.5625 , 9.255195087425777e-46 ], [ 518.625 , 9.25700173681893e-46 ], [ 518.6875 , 9.25880873887694e-46 ], 
                    [ 518.75 , 9.260616093668777e-46 ], [ 518.8125 , 9.262423801262967e-46 ], [ 518.875 , 9.264231861728775e-46 ], [ 518.9375 , 9.266040275134818e-46 ], 
                    [ 519.0 , 9.267849041549926e-46 ], [ 519.0625 , 9.269658161043337e-46 ], [ 519.125 , 9.271467633683709e-46 ], [ 519.1875 , 9.27327745953998e-46 ], 
                    [ 519.25 , 9.27508763868103e-46 ], [ 519.3125 , 9.276898171176285e-46 ], [ 519.375 , 9.278709057094259e-46 ], [ 519.4375 , 9.280520296503941e-46 ], 
                    [ 519.5 , 9.282332219504196e-46 ], [ 519.5625 , 9.284145455598904e-46 ], [ 519.625 , 9.285959045896227e-46 ], [ 519.6875 , 9.287772990465161e-46 ], 
                    [ 519.75 , 9.289587289374908e-46 ], [ 519.8125 , 9.291401942694886e-46 ], [ 519.875 , 9.29321695049406e-46 ], [ 519.9375 , 9.295032312841874e-46 ], 
                    [ 520.0 , 9.296848029807588e-46 ], [ 520.0625 , 9.298664101460342e-46 ], [ 520.125 , 9.30048052786955e-46 ], [ 520.1875 , 9.302297309104314e-46 ], 
                    [ 520.25 , 9.304114445234148e-46 ], [ 520.3125 , 9.30593193632831e-46 ], [ 520.375 , 9.307749782456008e-46 ], [ 520.4375 , 9.309567983686991e-46 ], 
                    [ 520.5 , 9.311386540090098e-46 ], [ 520.5625 , 9.313205451735105e-46 ], [ 520.625 , 9.31502471869134e-46 ], [ 520.6875 , 9.316844219230376e-46 ], 
                    [ 520.75 , 9.3186628126583e-46 ], [ 520.8125 , 9.320481761065079e-46 ], [ 520.875 , 9.322301064519867e-46 ], [ 520.9375 , 9.3241207230921e-46 ], 
                    [ 521.0 , 9.325940736851033e-46 ], [ 521.0625 , 9.327761105865923e-46 ], [ 521.125 , 9.329581830206317e-46 ], [ 521.1875 , 9.331402909941309e-46 ], 
                    [ 521.25 , 9.333224345140464e-46 ], [ 521.3125 , 9.335046135873104e-46 ], [ 521.375 , 9.336868282208494e-46 ], [ 521.4375 , 9.33869078421644e-46 ], 
                    [ 521.5 , 9.340513641965908e-46 ], [ 521.5625 , 9.342336855526466e-46 ], [ 521.625 , 9.344160424967831e-46 ], [ 521.6875 , 9.345984350359208e-46 ], 
                    [ 521.75 , 9.347808631769938e-46 ], [ 521.8125 , 9.34963326927005e-46 ], [ 521.875 , 9.351458262928515e-46 ], [ 521.9375 , 9.353277472060404e-46 ], 
                    [ 522.0 , 9.35509626043669e-46 ], [ 522.0625 , 9.356915402484781e-46 ], [ 522.125 , 9.358734898273653e-46 ], [ 522.1875 , 9.360554747871891e-46 ], 
                    [ 522.25 , 9.36237495134836e-46 ], [ 522.3125 , 9.364195508771808e-46 ], [ 522.375 , 9.36601642021126e-46 ], [ 522.4375 , 9.367837685735293e-46 ], 
                    [ 522.5 , 9.369659305413023e-46 ], [ 522.5625 , 9.371481279313187e-46 ], [ 522.625 , 9.373303607504461e-46 ], [ 522.6875 , 9.375126290056209e-46 ], 
                    [ 522.75 , 9.37694932703687e-46 ], [ 522.8125 , 9.378772718515632e-46 ], [ 522.875 , 9.38059646456136e-46 ], [ 522.9375 , 9.382420565242935e-46 ], 
                    [ 523.0 , 9.384245020629521e-46 ], [ 523.0625 , 9.386069830789821e-46 ], [ 523.125 , 9.387894949002131e-46 ], [ 523.1875 , 9.389720393907994e-46 ], 
                    [ 523.25 , 9.391546193765605e-46 ], [ 523.3125 , 9.393372348643918e-46 ], [ 523.375 , 9.395198858611832e-46 ], [ 523.4375 , 9.397025723738592e-46 ], 
                    [ 523.5 , 9.398852944093258e-46 ], [ 523.5625 , 9.400680519744838e-46 ], [ 523.625 , 9.402508450762414e-46 ], [ 523.6875 , 9.40433673721489e-46 ], 
                    [ 523.75 , 9.406165379171978e-46 ], [ 523.8125 , 9.407994376702003e-46 ], [ 523.875 , 9.40982372987464e-46 ], [ 523.9375 , 9.411653438758975e-46 ], 
                    [ 524.0 , 9.413483503424109e-46 ], [ 524.0625 , 9.415313923939155e-46 ], [ 524.125 , 9.417144700373506e-46 ], [ 524.1875 , 9.41897583279604e-46 ], 
                    [ 524.25 , 9.420807321276306e-46 ], [ 524.3125 , 9.422638203358418e-46 ], [ 524.375 , 9.424467468877194e-46 ], [ 524.4375 , 9.426297089520818e-46 ], 
                    [ 524.5 , 9.428127065358166e-46 ], [ 524.5625 , 9.429957396458127e-46 ], [ 524.625 , 9.431788082889738e-46 ], [ 524.6875 , 9.433619124722113e-46 ], 
                    [ 524.75 , 9.435450522023914e-46 ], [ 524.8125 , 9.437282274864485e-46 ], [ 524.875 , 9.439114383312715e-46 ], [ 524.9375 , 9.440946847437641e-46 ], 
                    [ 525.0 , 9.442779667308309e-46 ], [ 525.0625 , 9.444612842993584e-46 ], [ 525.125 , 9.446446374563009e-46 ], [ 525.1875 , 9.448280262085272e-46 ], 
                    [ 525.25 , 9.450114505629613e-46 ], [ 525.3125 , 9.451949105265212e-46 ], [ 525.375 , 9.453784061061132e-46 ], [ 525.4375 , 9.455619373086449e-46 ], 
                    [ 525.5 , 9.45745504141039e-46 ], [ 525.5625 , 9.45928572843025e-46 ], [ 525.625 , 9.461116683009505e-46 ], [ 525.6875 , 9.462947991991276e-46 ], 
                    [ 525.75 , 9.464779655444029e-46 ], [ 525.8125 , 9.466611673436575e-46 ], [ 525.875 , 9.468444046037543e-46 ], [ 525.9375 , 9.470276773315503e-46 ], 
                    [ 526.0 , 9.472109855339037e-46 ], [ 526.0625 , 9.473943292176948e-46 ], [ 526.125 , 9.475777083897712e-46 ], [ 526.1875 , 9.477611230570222e-46 ], 
                    [ 526.25 , 9.479445732263182e-46 ], [ 526.3125 , 9.481280589045245e-46 ], [ 526.375 , 9.483115800985074e-46 ], [ 526.4375 , 9.484951368151548e-46 ], 
                    [ 526.5 , 9.486787290613293e-46 ], [ 526.5625 , 9.488623568439144e-46 ], [ 526.625 , 9.490460201697823e-46 ], [ 526.6875 , 9.492297190458396e-46 ], 
                    [ 526.75 , 9.494135592471318e-46 ], [ 526.8125 , 9.495975079617385e-46 ], [ 526.875 , 9.497814923163774e-46 ], [ 526.9375 , 9.499655123179535e-46 ], 
                    [ 527.0 , 9.501495679733734e-46 ], [ 527.0625 , 9.50333659289545e-46 ], [ 527.125 , 9.505177862733777e-46 ], [ 527.1875 , 9.507019489317819e-46 ], 
                    [ 527.25 , 9.508861472716696e-46 ], [ 527.3125 , 9.510703812999542e-46 ], [ 527.375 , 9.512546510235503e-46 ], [ 527.4375 , 9.514389564493804e-46 ], 
                    [ 527.5 , 9.516232975843488e-46 ], [ 527.5625 , 9.518076744353666e-46 ], [ 527.625 , 9.519920870093883e-46 ], [ 527.6875 , 9.521765353133078e-46 ], 
                    [ 527.75 , 9.523610193540547e-46 ], [ 527.8125 , 9.525455391385526e-46 ], [ 527.875 , 9.527300946737275e-46 ], [ 527.9375 , 9.529145689328239e-46 ], 
                    [ 528.0 , 9.530984992037388e-46 ], [ 528.0625 , 9.532824649766306e-46 ], [ 528.125 , 9.534664662583315e-46 ], [ 528.1875 , 9.536505030557158e-46 ], 
                    [ 528.25 , 9.538345753756251e-46 ], [ 528.3125 , 9.540186832249361e-46 ], [ 528.375 , 9.542028266104863e-46 ], [ 528.4375 , 9.543870055391417e-46 ], 
                    [ 528.5 , 9.545712200177629e-46 ], [ 528.5625 , 9.547554700532116e-46 ], [ 528.625 , 9.54939755652351e-46 ], [ 528.6875 , 9.551240768220387e-46 ], 
                    [ 528.75 , 9.55308433569154e-46 ], [ 528.8125 , 9.554928259005503e-46 ], [ 528.875 , 9.556772538231101e-46 ], [ 528.9375 , 9.558617173437096e-46 ], 
                    [ 529.0 , 9.56046216469186e-46 ], [ 529.0625 , 9.562307512064252e-46 ], [ 529.125 , 9.564153215623212e-46 ], [ 529.1875 , 9.565999179128038e-46 ], 
                    [ 529.25 , 9.567845453588758e-46 ], [ 529.3125 , 9.569692084387449e-46 ], [ 529.375 , 9.571539071593022e-46 ], [ 529.4375 , 9.573386415274196e-46 ], 
                    [ 529.5 , 9.575234115499634e-46 ], [ 529.5625 , 9.577082172338565e-46 ], [ 529.625 , 9.578930585859333e-46 ], [ 529.6875 , 9.580779356131055e-46 ], 
                    [ 529.75 , 9.582628483222448e-46 ], [ 529.8125 , 9.584477967202517e-46 ], [ 529.875 , 9.586327808140075e-46 ], [ 529.9375 , 9.588178006104015e-46 ], 
                    [ 530.0 , 9.590028561163108e-46 ], [ 530.0625 , 9.591879473386615e-46 ], [ 530.125 , 9.59373074284313e-46 ], [ 530.1875 , 9.595582369601805e-46 ], 
                    [ 530.25 , 9.597434353731534e-46 ], [ 530.3125 , 9.59928669530129e-46 ], [ 530.375 , 9.601138611578416e-46 ], [ 530.4375 , 9.602986774631306e-46 ], 
                    [ 530.5 , 9.604835293444709e-46 ], [ 530.5625 , 9.606684168087315e-46 ], [ 530.625 , 9.608533398627344e-46 ], [ 530.6875 , 9.610382985133507e-46 ], 
                    [ 530.75 , 9.612232927674262e-46 ], [ 530.8125 , 9.61408322631814e-46 ], [ 530.875 , 9.61593388113369e-46 ], [ 530.9375 , 9.617784892189404e-46 ], 
                    [ 531.0 , 9.619636259553995e-46 ], [ 531.0625 , 9.621487983295841e-46 ], [ 531.125 , 9.623340063483957e-46 ], [ 531.1875 , 9.625192500186545e-46 ], 
                    [ 531.25 , 9.627045293472435e-46 ], [ 531.3125 , 9.628898443409998e-46 ], [ 531.375 , 9.630751950068293e-46 ], [ 531.4375 , 9.63260581351565e-46 ], 
                    [ 531.5 , 9.634460033820948e-46 ], [ 531.5625 , 9.636314611052818e-46 ], [ 531.625 , 9.638167430428633e-46 ], [ 531.6875 , 9.640019148241932e-46 ], 
                    [ 531.75 , 9.641871221813683e-46 ], [ 531.8125 , 9.643723651212098e-46 ], [ 531.875 , 9.645576436505677e-46 ], [ 531.9375 , 9.647429577762657e-46 ], 
                    [ 532.0 , 9.649283075051568e-46 ], [ 532.0625 , 9.65113692844067e-46 ], [ 532.125 , 9.652991137998519e-46 ], [ 532.1875 , 9.654845703793473e-46 ], 
                    [ 532.25 , 9.656700625893906e-46 ], [ 532.3125 , 9.658555904368478e-46 ], [ 532.375 , 9.660411539285313e-46 ], [ 532.4375 , 9.662267530713234e-46 ], 
                    [ 532.5 , 9.664123878720534e-46 ], [ 532.5625 , 9.665980583375783e-46 ], [ 532.625 , 9.667837644747576e-46 ], [ 532.6875 , 9.669695062904442e-46 ], 
                    [ 532.75 , 9.671552837914722e-46 ], [ 532.8125 , 9.673410969847253e-46 ], [ 532.875 , 9.675267809607311e-46 ], [ 532.9375 , 9.677124965232336e-46 ], 
                    [ 533.0 , 9.678982477336076e-46 ], [ 533.0625 , 9.680840345987021e-46 ], [ 533.125 , 9.682698571253474e-46 ], [ 533.1875 , 9.684557153204026e-46 ], 
                    [ 533.25 , 9.686416091907004e-46 ], [ 533.3125 , 9.688275387430953e-46 ], [ 533.375 , 9.690135039844436e-46 ], [ 533.4375 , 9.691995049215886e-46 ], 
                    [ 533.5 , 9.693855415613756e-46 ], [ 533.5625 , 9.69571613910664e-46 ], [ 533.625 , 9.697577219763155e-46 ], [ 533.6875 , 9.699438657651648e-46 ], 
                    [ 533.75 , 9.701300452840968e-46 ], [ 533.8125 , 9.70316260539956e-46 ], [ 533.875 , 9.70502511539602e-46 ], [ 533.9375 , 9.706887982898819e-46 ], 
                    [ 534.0 , 9.708751207976789e-46 ], [ 534.0625 , 9.710614317055266e-46 ], [ 534.125 , 9.71247673901996e-46 ], [ 534.1875 , 9.71433951818295e-46 ], 
                    [ 534.25 , 9.716202654612817e-46 ], [ 534.3125 , 9.71806614837808e-46 ], [ 534.375 , 9.719929999547271e-46 ], [ 534.4375 , 9.721794208189009e-46 ], 
                    [ 534.5 , 9.723658774371717e-46 ], [ 534.5625 , 9.725523698164034e-46 ], [ 534.625 , 9.72738897963469e-46 ], [ 534.6875 , 9.729254618851866e-46 ], 
                    [ 534.75 , 9.73112061588473e-46 ], [ 534.8125 , 9.732986970801426e-46 ], [ 534.875 , 9.734853683670867e-46 ], [ 534.9375 , 9.736720754561706e-46 ], 
                    [ 535.0 , 9.738588183542542e-46 ], [ 535.0625 , 9.740455970681913e-46 ], [ 535.125 , 9.74232411604865e-46 ], [ 535.1875 , 9.744192619711668e-46 ], 
                    [ 535.25 , 9.746061481739267e-46 ], [ 535.3125 , 9.74792812342421e-46 ], [ 535.375 , 9.749793514766485e-46 ], [ 535.4375 , 9.751659263075298e-46 ], 
                    [ 535.5 , 9.753525368419102e-46 ], [ 535.5625 , 9.75539183086608e-46 ], [ 535.625 , 9.757258650484639e-46 ], [ 535.6875 , 9.759125827343196e-46 ], 
                    [ 535.75 , 9.760993361510044e-46 ], [ 535.8125 , 9.762861253053491e-46 ], [ 535.875 , 9.764729502042062e-46 ], [ 535.9375 , 9.766598108544161e-46 ], 
                    [ 536.0 , 9.768467072627994e-46 ], [ 536.0625 , 9.770336394362126e-46 ], [ 536.125 , 9.77220607381507e-46 ], [ 536.1875 , 9.774076111055209e-46 ], 
                    [ 536.25 , 9.775946506150871e-46 ], [ 536.3125 , 9.777817259170816e-46 ], [ 536.375 , 9.779688370183397e-46 ], [ 536.4375 , 9.78155983925712e-46 ], 
                    [ 536.5 , 9.783431666460366e-46 ], [ 536.5625 , 9.785299873550198e-46 ], [ 536.625 , 9.787167853301936e-46 ], [ 536.6875 , 9.78903618964459e-46 ], 
                    [ 536.75 , 9.79090488264609e-46 ], [ 536.8125 , 9.792773932374733e-46 ], [ 536.875 , 9.794643338898337e-46 ], [ 536.9375 , 9.796513102285152e-46 ], 
                    [ 537.0 , 9.79838322260337e-46 ], [ 537.0625 , 9.800253699920992e-46 ], [ 537.125 , 9.802124534306304e-46 ], [ 537.1875 , 9.803995725827473e-46 ], 
                    [ 537.25 , 9.805867274552531e-46 ], [ 537.3125 , 9.8077391805496e-46 ], [ 537.375 , 9.80961144388716e-46 ], [ 537.4375 , 9.811484064633218e-46 ], 
                    [ 537.5 , 9.81335704285607e-46 ], [ 537.5625 , 9.815230378624029e-46 ], [ 537.625 , 9.817104072005279e-46 ], [ 537.6875 , 9.818978123068014e-46 ], 
                    [ 537.75 , 9.820852517088963e-46 ], [ 537.8125 , 9.822727052553881e-46 ], [ 537.875 , 9.824601945816984e-46 ], [ 537.9375 , 9.826477196946428e-46 ], 
                    [ 538.0 , 9.828352806010726e-46 ], [ 538.0625 , 9.830228773078201e-46 ], [ 538.125 , 9.832105098217046e-46 ], [ 538.1875 , 9.833981781495539e-46 ], 
                    [ 538.25 , 9.835858822982383e-46 ], [ 538.3125 , 9.837736222745607e-46 ], [ 538.375 , 9.839613980853873e-46 ], [ 538.4375 , 9.841492097375298e-46 ], 
                    [ 538.5 , 9.843370572378508e-46 ], [ 538.5625 , 9.845249405931853e-46 ], [ 538.625 , 9.847128598103773e-46 ], [ 538.6875 , 9.849008148962788e-46 ], 
                    [ 538.75 , 9.85088805857715e-46 ], [ 538.8125 , 9.852768327015479e-46 ], [ 538.875 , 9.854648954346403e-46 ], [ 538.9375 , 9.856529940638145e-46 ], 
                    [ 539.0 , 9.858410046282374e-46 ], [ 539.0625 , 9.860286215138133e-46 ], [ 539.125 , 9.862162741050414e-46 ], [ 539.1875 , 9.864039624087164e-46 ], 
                    [ 539.25 , 9.865916864316351e-46 ], [ 539.3125 , 9.86779446180588e-46 ], [ 539.375 , 9.869672416623956e-46 ], [ 539.4375 , 9.87155072883837e-46 ], 
                    [ 539.5 , 9.873429398517209e-46 ], [ 539.5625 , 9.875308425728502e-46 ], [ 539.625 , 9.87718781054022e-46 ], [ 539.6875 , 9.879067553020493e-46 ], 
                    [ 539.75 , 9.880947653237598e-46 ], [ 539.8125 , 9.882828111259264e-46 ], [ 539.875 , 9.884708927153656e-46 ], [ 539.9375 , 9.886590100989094e-46 ], 
                    [ 540.0 , 9.888471632833555e-46 ], [ 540.0625 , 9.890353522754893e-46 ], [ 540.125 , 9.892235770821677e-46 ], [ 540.1875 , 9.89411837710213e-46 ], 
                    [ 540.25 , 9.896001446368725e-46 ], [ 540.3125 , 9.897885080980844e-46 ], [ 540.375 , 9.899769074129417e-46 ], [ 540.4375 , 9.901653425883181e-46 ], 
                    [ 540.5 , 9.903538136309902e-46 ], [ 540.5625 , 9.905423205478132e-46 ], [ 540.625 , 9.907308633456083e-46 ], [ 540.6875 , 9.909194420312053e-46 ], 
                    [ 540.75 , 9.911080566114425e-46 ], [ 540.8125 , 9.912967070931448e-46 ], [ 540.875 , 9.914853934831247e-46 ], [ 540.9375 , 9.916741157882594e-46 ], 
                    [ 541.0 , 9.91862874015357e-46 ], [ 541.0625 , 9.92051668171262e-46 ], [ 541.125 , 9.922404982628131e-46 ], [ 541.1875 , 9.924293642968575e-46 ], 
                    [ 541.25 , 9.926182662802227e-46 ], [ 541.3125 , 9.928072042197654e-46 ], [ 541.375 , 9.929961781223081e-46 ], [ 541.4375 , 9.93185187994718e-46 ], 
                    [ 541.5 , 9.93373994647529e-46 ], [ 541.5625 , 9.935625013343521e-46 ], [ 541.625 , 9.937510437929775e-46 ], [ 541.6875 , 9.939396220301648e-46 ], 
                    [ 541.75 , 9.941282360527392e-46 ], [ 541.8125 , 9.943168858674769e-46 ], [ 541.875 , 9.945055714811775e-46 ], [ 541.9375 , 9.946942929006057e-46 ], 
                    [ 542.0 , 9.94883050132606e-46 ], [ 542.0625 , 9.950718431839175e-46 ], [ 542.125 , 9.952606720613795e-46 ], [ 542.1875 , 9.954495367717632e-46 ], 
                    [ 542.25 , 9.956384373219101e-46 ], [ 542.3125 , 9.958273737185649e-46 ], [ 542.375 , 9.960163459685794e-46 ], [ 542.4375 , 9.96205354078722e-46 ], 
                    [ 542.5 , 9.963943980558259e-46 ], [ 542.5625 , 9.965834779066693e-46 ], [ 542.625 , 9.967725936380808e-46 ], [ 542.6875 , 9.969617452568692e-46 ], 
                    [ 542.75 , 9.971508636151968e-46 ], [ 542.8125 , 9.973399269055162e-46 ], [ 542.875 , 9.975290260428967e-46 ], [ 542.9375 , 9.977181610341347e-46 ], 
                    [ 543.0 , 9.979073318860355e-46 ], [ 543.0625 , 9.980965386053844e-46 ], [ 543.125 , 9.982857811989814e-46 ], [ 543.1875 , 9.984750596736503e-46 ], 
                    [ 543.25 , 9.986643740361656e-46 ], [ 543.3125 , 9.988537242933601e-46 ], [ 543.375 , 9.990431104520183e-46 ], [ 543.4375 , 9.992325325189544e-46 ], 
                    [ 543.5 , 9.994219905009909e-46 ], [ 543.5625 , 9.996114844048949e-46 ], [ 543.625 , 9.998010142375198e-46 ], [ 543.6875 , 9.999905800056638e-46 ], 
                    [ 543.75 , 1.0001801817161401e-45 ], [ 543.8125 , 1.000369819375778e-45 ], [ 543.875 , 1.0005594929913509e-45 ], [ 543.9375 , 1.0007492025697185e-45 ], 
                    [ 544.0 , 1.0009387318094626e-45 ], [ 544.0625 , 1.0011279932227224e-45 ], [ 544.125 , 1.0013172904222854e-45 ], [ 544.1875 , 1.0015066234148894e-45 ], 
                    [ 544.25 , 1.0016959922073168e-45 ], [ 544.3125 , 1.0018853968063511e-45 ], [ 544.375 , 1.0020748372187267e-45 ], [ 544.4375 , 1.0022643134512657e-45 ], 
                    [ 544.5 , 1.0024538255106909e-45 ], [ 544.5625 , 1.002643373403798e-45 ], [ 544.625 , 1.00283295713737e-45 ], [ 544.6875 , 1.0030225767181763e-45 ], 
                    [ 544.75 , 1.0032122321529952e-45 ], [ 544.8125 , 1.0034019234485988e-45 ], [ 544.875 , 1.0035916506117822e-45 ], [ 544.9375 , 1.0037814136493204e-45 ], 
                    [ 545.0 , 1.0039712125679892e-45 ], [ 545.0625 , 1.0041610473745947e-45 ], [ 545.125 , 1.0043509180759015e-45 ], [ 545.1875 , 1.0045408246786965e-45 ], 
                    [ 545.25 , 1.0047307569328146e-45 ], [ 545.3125 , 1.0049207048245863e-45 ], [ 545.375 , 1.0051106886266835e-45 ], [ 545.4375 , 1.0053007083458742e-45 ], 
                    [ 545.5 , 1.0054907639889769e-45 ], [ 545.5625 , 1.0056808555627548e-45 ], [ 545.625 , 1.005870983074022e-45 ], [ 545.6875 , 1.0060611465295654e-45 ], 
                    [ 545.75 , 1.0062513459361806e-45 ], [ 545.8125 , 1.0064415813006641e-45 ], [ 545.875 , 1.0066318526298139e-45 ], [ 545.9375 , 1.006822159930415e-45 ], 
                    [ 546.0 , 1.0070125032093109e-45 ], [ 546.0625 , 1.0072028824732531e-45 ], [ 546.125 , 1.0073932977290809e-45 ], [ 546.1875 , 1.0075837489835768e-45 ], 
                    [ 546.25 , 1.0077742362435467e-45 ], [ 546.3125 , 1.0079647595158116e-45 ], [ 546.375 , 1.0081553188071804e-45 ], [ 546.4375 , 1.0083459141244402e-45 ], 
                    [ 546.5 , 1.0085364622144736e-45 ], [ 546.5625 , 1.0087267578218327e-45 ], [ 546.625 , 1.0089170893351004e-45 ], [ 546.6875 , 1.0091074567610518e-45 ], 
                    [ 546.75 , 1.0092978601064631e-45 ], [ 546.8125 , 1.0094882993781118e-45 ], [ 546.875 , 1.0096787745827763e-45 ], [ 546.9375 , 1.0098692857272369e-45 ], 
                    [ 547.0 , 1.0100598328182822e-45 ], [ 547.0625 , 1.01025041586268e-45 ], [ 547.125 , 1.0104410348672289e-45 ], [ 547.1875 , 1.0106316898386924e-45 ], 
                    [ 547.25 , 1.0108223807838926e-45 ], [ 547.3125 , 1.01101310770956e-45 ], [ 547.375 , 1.0112038706225554e-45 ], [ 547.4375 , 1.0113946695296185e-45 ], 
                    [ 547.5 , 1.0115855044375485e-45 ], [ 547.5625 , 1.0117763753531664e-45 ], [ 547.625 , 1.0119672822832308e-45 ], [ 547.6875 , 1.0121582252345442e-45 ], 
                    [ 547.75 , 1.0123491742905957e-45 ], [ 547.8125 , 1.0125397216590125e-45 ], [ 547.875 , 1.012730304892821e-45 ], [ 547.9375 , 1.0129209239987717e-45 ], 
                    [ 548.0 , 1.0131115789836168e-45 ], [ 548.0625 , 1.0133022698541165e-45 ], [ 548.125 , 1.013492996617004e-45 ], [ 548.1875 , 1.0136837592790491e-45 ], 
                    [ 548.25 , 1.0138745578470308e-45 ], [ 548.3125 , 1.014065392327664e-45 ], [ 548.375 , 1.0142562627277444e-45 ], [ 548.4375 , 1.0144471690540255e-45 ], 
                    [ 548.5 , 1.0146381113132409e-45 ], [ 548.5625 , 1.0148290895121681e-45 ], [ 548.625 , 1.015020103657608e-45 ], [ 548.6875 , 1.0152111537562906e-45 ], 
                    [ 548.75 , 1.015402239814983e-45 ], [ 548.8125 , 1.0155933618404607e-45 ], [ 548.875 , 1.0157845198394868e-45 ], [ 548.9375 , 1.0159757138188463e-45 ], 
                    [ 549.0 , 1.0161669437853045e-45 ], [ 549.0625 , 1.0163581337772686e-45 ], [ 549.125 , 1.01654934526839e-45 ], [ 549.1875 , 1.016740592732903e-45 ], 
                    [ 549.25 , 1.0169318761775535e-45 ], [ 549.3125 , 1.0171231956091179e-45 ], [ 549.375 , 1.017314551034374e-45 ], [ 549.4375 , 1.0175059424600715e-45 ], 
                    [ 549.5 , 1.017697369893012e-45 ], [ 549.5625 , 1.0178888333399486e-45 ], [ 549.625 , 1.0180803328076564e-45 ], [ 549.6875 , 1.0182718683029265e-45 ], 
                    [ 549.75 , 1.0184634398325373e-45 ], [ 549.8125 , 1.0186550474032459e-45 ], [ 549.875 , 1.0188466910218622e-45 ], [ 549.9375 , 1.0190383706951602e-45 ], 
                    [ 550.0 , 1.019230086429895e-45 ], [ 550.0625 , 1.0194218382328793e-45 ], [ 550.125 , 1.019613626110877e-45 ], [ 550.1875 , 1.0198054500707115e-45 ], 
                    [ 550.25 , 1.0199973101191418e-45 ], [ 550.3125 , 1.0201889984583777e-45 ], [ 550.375 , 1.0203805747686772e-45 ], [ 550.4375 , 1.0205721870541418e-45 ], 
                    [ 550.5 , 1.0207638353215638e-45 ], [ 550.5625 , 1.0209555195776704e-45 ], [ 550.625 , 1.0211472398292207e-45 ], [ 550.6875 , 1.0213389960829805e-45 ], 
                    [ 550.75 , 1.0215307883457038e-45 ], [ 550.8125 , 1.021722616624174e-45 ], [ 550.875 , 1.0219144809251326e-45 ], [ 550.9375 , 1.0221063812553514e-45 ], 
                    [ 551.0 , 1.0222983176216033e-45 ], [ 551.0625 , 1.0224902900306483e-45 ], [ 551.125 , 1.0226822984892545e-45 ], [ 551.1875 , 1.0228743430041771e-45 ], 
                    [ 551.25 , 1.0230664235822083e-45 ], [ 551.3125 , 1.0232585402301209e-45 ], [ 551.375 , 1.0234506929546807e-45 ], [ 551.4375 , 1.0236428817626622e-45 ], 
                    [ 551.5 , 1.023835106660827e-45 ], [ 551.5625 , 1.0240272671693864e-45 ], [ 551.625 , 1.0242192247039813e-45 ], [ 551.6875 , 1.02441121822171e-45 ], 
                    [ 551.75 , 1.0246032477292955e-45 ], [ 551.8125 , 1.0247953132334916e-45 ], [ 551.875 , 1.0249874147410462e-45 ], [ 551.9375 , 1.0251795522587008e-45 ], 
                    [ 552.0 , 1.0253717257932274e-45 ], [ 552.0625 , 1.025563935351356e-45 ], [ 552.125 , 1.0257561809398461e-45 ], [ 552.1875 , 1.0259484625654448e-45 ], 
                    [ 552.25 , 1.0261407802349364e-45 ], [ 552.3125 , 1.0263331339550265e-45 ], [ 552.375 , 1.0265255237325386e-45 ], [ 552.4375 , 1.0267179495741587e-45 ], 
                    [ 552.5 , 1.0269104114867055e-45 ], [ 552.5625 , 1.0271029094769044e-45 ], [ 552.625 , 1.0272954435515399e-45 ], [ 552.6875 , 1.0274880137173761e-45 ], 
                    [ 552.75 , 1.0276806199811641e-45 ], [ 552.8125 , 1.0278732623496778e-45 ], [ 552.875 , 1.0280655849531765e-45 ], [ 552.9375 , 1.0282579224263885e-45 ], 
                    [ 553.0 , 1.0284502958833976e-45 ], [ 553.0625 , 1.0286427053309215e-45 ], [ 553.125 , 1.0288351507757154e-45 ], [ 553.1875 , 1.029027632224514e-45 ], 
                    [ 553.25 , 1.0292201496840386e-45 ], [ 553.3125 , 1.0294127031610334e-45 ], [ 553.375 , 1.0296052926622367e-45 ], [ 553.4375 , 1.0297979181943888e-45 ], 
                    [ 553.5 , 1.0299905797642373e-45 ], [ 553.5625 , 1.0301832773785027e-45 ], [ 553.625 , 1.030376011043943e-45 ], [ 553.6875 , 1.030568780767325e-45 ], 
                    [ 553.75 , 1.0307615865553577e-45 ], [ 553.8125 , 1.030954428414796e-45 ], [ 553.875 , 1.0311473063524102e-45 ], [ 553.9375 , 1.031340220374928e-45 ], 
                    [ 554.0 , 1.031533170489108e-45 ], [ 554.0625 , 1.0317261567017168e-45 ], [ 554.125 , 1.0319190035561763e-45 ], [ 554.1875 , 1.032111739349937e-45 ], 
                    [ 554.25 , 1.0323045111417836e-45 ], [ 554.3125 , 1.0324973189383952e-45 ], [ 554.375 , 1.0326901627465339e-45 ], [ 554.4375 , 1.0328830425729255e-45 ], 
                    [ 554.5 , 1.033075958424253e-45 ], [ 554.5625 , 1.0332689103072891e-45 ], [ 554.625 , 1.0334618982287345e-45 ], [ 554.6875 , 1.0336549221953565e-45 ], 
                    [ 554.75 , 1.0338479822138511e-45 ], [ 554.8125 , 1.0340410782909737e-45 ], [ 554.875 , 1.0342342104334443e-45 ], [ 554.9375 , 1.0344273786480214e-45 ], 
                    [ 555.0 , 1.0346205829414126e-45 ], [ 555.0625 , 1.0348138233203711e-45 ], [ 555.125 , 1.0350070997916446e-45 ], [ 555.1875 , 1.0352004123619664e-45 ], 
                    [ 555.25 , 1.0353937610380644e-45 ], [ 555.3125 , 1.035587145826704e-45 ], [ 555.375 , 1.0357805430027478e-45 ], [ 555.4375 , 1.03597375277888e-45 ], 
                    [ 555.5 , 1.0361669985954898e-45 ], [ 555.5625 , 1.0363602804592848e-45 ], [ 555.625 , 1.036553598377004e-45 ], [ 555.6875 , 1.0367469523553584e-45 ], 
                    [ 555.75 , 1.0369403424010814e-45 ], [ 555.8125 , 1.0371337685209086e-45 ], [ 555.875 , 1.0373272307215614e-45 ], [ 555.9375 , 1.0375207290097632e-45 ], 
                    [ 556.0 , 1.0377142633922524e-45 ], [ 556.0625 , 1.0379078338757696e-45 ], [ 556.125 , 1.0381014404670345e-45 ], [ 556.1875 , 1.0382950831727893e-45 ], 
                    [ 556.25 , 1.0384887619997854e-45 ], [ 556.3125 , 1.0386824769547242e-45 ], [ 556.375 , 1.0388762280443888e-45 ], [ 556.4375 , 1.0390700152754828e-45 ], 
                    [ 556.5 , 1.03926383865477e-45 ], [ 556.5625 , 1.0394576981889787e-45 ], [ 556.625 , 1.0396515938848677e-45 ], [ 556.6875 , 1.0398453090626243e-45 ], 
                    [ 556.75 , 1.0400389158501622e-45 ], [ 556.8125 , 1.0402325586849664e-45 ], [ 556.875 , 1.0404262375737701e-45 ], [ 556.9375 , 1.0406199525232574e-45 ], 
                    [ 557.0 , 1.0408137035401634e-45 ], [ 557.0625 , 1.041007490631182e-45 ], [ 557.125 , 1.0412013138030594e-45 ], [ 557.1875 , 1.0413951730624982e-45 ], 
                    [ 557.25 , 1.0415890684162177e-45 ], [ 557.3125 , 1.0417829998709381e-45 ], [ 557.375 , 1.0419769674333812e-45 ], [ 557.4375 , 1.0421709711102697e-45 ], 
                    [ 557.5 , 1.0423650109083279e-45 ], [ 557.5625 , 1.0425590868342733e-45 ], [ 557.625 , 1.042753198894833e-45 ], [ 557.6875 , 1.042947347096771e-45 ], 
                    [ 557.75 , 1.0431415314467654e-45 ], [ 557.8125 , 1.0433357519515833e-45 ], [ 557.875 , 1.0435300086179268e-45 ], [ 557.9375 , 1.0437242860723971e-45 ], 
                    [ 558.0 , 1.0439184014261323e-45 ], [ 558.0625 , 1.0441125528821017e-45 ], [ 558.125 , 1.0443067404470043e-45 ], [ 558.1875 , 1.0445009641275782e-45 ], 
                    [ 558.25 , 1.044695223930511e-45 ], [ 558.3125 , 1.0448895198625427e-45 ], [ 558.375 , 1.0450838519303854e-45 ], [ 558.4375 , 1.045278220140745e-45 ], 
                    [ 558.5 , 1.0454726245003655e-45 ], [ 558.5625 , 1.0456670650159705e-45 ], [ 558.625 , 1.0458615416942764e-45 ], [ 558.6875 , 1.0460560545420167e-45 ], 
                    [ 558.75 , 1.046250603565881e-45 ], [ 558.8125 , 1.0464451887726568e-45 ], [ 558.875 , 1.0466398101690367e-45 ], [ 558.9375 , 1.046834467761751e-45 ], 
                    [ 559.0 , 1.0470291615575393e-45 ], [ 559.0625 , 1.0472238915631496e-45 ], [ 559.125 , 1.0474186577852866e-45 ], [ 559.1875 , 1.0476134602307004e-45 ], 
                    [ 559.25 , 1.0478081243125019e-45 ], [ 559.3125 , 1.0480026353879073e-45 ], [ 559.375 , 1.0481971825716169e-45 ], [ 559.4375 , 1.0483917658703032e-45 ], 
                    [ 559.5 , 1.0485863852907005e-45 ], [ 559.5625 , 1.0487810408394845e-45 ], [ 559.625 , 1.0489757325233922e-45 ], [ 559.6875 , 1.0491704603491084e-45 ], 
                    [ 559.75 , 1.0493652243233502e-45 ], [ 559.8125 , 1.0495600244528284e-45 ], [ 559.875 , 1.0497548607442615e-45 ], [ 559.9375 , 1.049949733204348e-45 ], 
                    [ 560.0 , 1.0501446418398093e-45 ], [ 560.0625 , 1.0503395866573612e-45 ], [ 560.125 , 1.0505345676637128e-45 ], [ 560.1875 , 1.0507295848656043e-45 ], 
                    [ 560.25 , 1.0509246382697327e-45 ], [ 560.3125 , 1.0511197278828257e-45 ], [ 560.375 , 1.0513148537115977e-45 ], [ 560.4375 , 1.051510015762794e-45 ], 
                    [ 560.5 , 1.0517052140431088e-45 ], [ 560.5625 , 1.0519001505435463e-45 ], [ 560.625 , 1.0520950730332217e-45 ], [ 560.6875 , 1.052290031643041e-45 ], 
                    [ 560.75 , 1.052485026379675e-45 ], [ 560.8125 , 1.0526800572498406e-45 ], [ 560.875 , 1.0528751242602262e-45 ], [ 560.9375 , 1.0530702274175362e-45 ], 
                    [ 561.0 , 1.0532653667284462e-45 ], [ 561.0625 , 1.0534605421996859e-45 ], [ 561.125 , 1.0536557538379334e-45 ], [ 561.1875 , 1.0538510016499055e-45 ], 
                    [ 561.25 , 1.0540462856422752e-45 ], [ 561.3125 , 1.0542416058217925e-45 ], [ 561.375 , 1.0544369621951399e-45 ], [ 561.4375 , 1.0546323547690175e-45 ], 
                    [ 561.5 , 1.0548277835501403e-45 ], [ 561.5625 , 1.0550232485452182e-45 ], [ 561.625 , 1.0552187497609614e-45 ], [ 561.6875 , 1.0554142872040894e-45 ], 
                    [ 561.75 , 1.0556098608813008e-45 ], [ 561.8125 , 1.0558054314801317e-45 ], [ 561.875 , 1.056000843552301e-45 ], [ 561.9375 , 1.0561962917919959e-45 ], 
                    [ 562.0 , 1.0563917762059329e-45 ], [ 562.0625 , 1.0565872968007849e-45 ], [ 562.125 , 1.0567828535832482e-45 ], [ 562.1875 , 1.0569784465600434e-45 ], 
                    [ 562.25 , 1.0571740757378314e-45 ], [ 562.3125 , 1.0573697411233653e-45 ], [ 562.375 , 1.057565442723294e-45 ], [ 562.4375 , 1.0577611805443425e-45 ], 
                    [ 562.5 , 1.05795695459323e-45 ], [ 562.5625 , 1.058152764876639e-45 ], [ 562.625 , 1.0583486114012836e-45 ], [ 562.6875 , 1.058544494173856e-45 ], 
                    [ 562.75 , 1.0587404132010956e-45 ], [ 562.8125 , 1.0589363684897123e-45 ], [ 562.875 , 1.0591323600463725e-45 ], [ 562.9375 , 1.0593283878778489e-45 ], 
                    [ 563.0 , 1.0595244519907951e-45 ], [ 563.0625 , 1.059720552391979e-45 ], [ 563.125 , 1.0599164729090081e-45 ], [ 563.1875 , 1.0601121632350314e-45 ], 
                    [ 563.25 , 1.0603078896909804e-45 ], [ 563.3125 , 1.0605036522835338e-45 ], [ 563.375 , 1.0606994510193552e-45 ], [ 563.4375 , 1.0608952859051107e-45 ], 
                    [ 563.5 , 1.0610911569474821e-45 ], [ 563.5625 , 1.0612870641531446e-45 ], [ 563.625 , 1.0614830075287752e-45 ], [ 563.6875 , 1.0616789870810518e-45 ], 
                    [ 563.75 , 1.0618750028166611e-45 ], [ 563.8125 , 1.0620710547422763e-45 ], [ 563.875 , 1.0622671428645563e-45 ], [ 563.9375 , 1.0624632671902216e-45 ], 
                    [ 564.0 , 1.0626594277259413e-45 ], [ 564.0625 , 1.062855624478401e-45 ], [ 564.125 , 1.0630518574542947e-45 ], [ 564.1875 , 1.0632481266602952e-45 ], 
                    [ 564.25 , 1.0634444321030993e-45 ], [ 564.3125 , 1.0636407737894123e-45 ], [ 564.375 , 1.0638371517258877e-45 ], [ 564.4375 , 1.0640334965700228e-45 ], 
                    [ 564.5 , 1.0642298425532696e-45 ], [ 564.5625 , 1.064426224768219e-45 ], [ 564.625 , 1.0646226432215565e-45 ], [ 564.6875 , 1.0648190979199696e-45 ], 
                    [ 564.75 , 1.065015588870154e-45 ], [ 564.8125 , 1.0652121160787839e-45 ], [ 564.875 , 1.065408679552558e-45 ], [ 564.9375 , 1.0656052792981757e-45 ], 
                    [ 565.0 , 1.0658019153223151e-45 ], [ 565.0625 , 1.0659985876316782e-45 ], [ 565.125 , 1.0661952962329684e-45 ], [ 565.1875 , 1.0663920411328598e-45 ], 
                    [ 565.25 , 1.0665888223380735e-45 ], [ 565.3125 , 1.0667856398553085e-45 ], [ 565.375 , 1.0669824936912508e-45 ], [ 565.4375 , 1.0671793838526096e-45 ], 
                    [ 565.5 , 1.067376310346088e-45 ], [ 565.5625 , 1.067573273178398e-45 ], [ 565.625 , 1.0677702723562377e-45 ], [ 565.6875 , 1.067967307886291e-45 ], 
                    [ 565.75 , 1.0681640796360259e-45 ], [ 565.8125 , 1.068360833832276e-45 ], [ 565.875 , 1.068557624270342e-45 ], [ 565.9375 , 1.0687544509569066e-45 ], 
                    [ 566.0 , 1.0689513138986625e-45 ], [ 566.0625 , 1.0691482131022646e-45 ], [ 566.125 , 1.0693451485743924e-45 ], [ 566.1875 , 1.0695421203217344e-45 ], 
                    [ 566.25 , 1.0697391283509873e-45 ], [ 566.3125 , 1.0699361726687962e-45 ], [ 566.375 , 1.0701332532818911e-45 ], [ 566.4375 , 1.0703303701969194e-45 ], 
                    [ 566.5 , 1.0705275234205833e-45 ], [ 566.5625 , 1.0707247129595858e-45 ], [ 566.625 , 1.0709219388205782e-45 ], [ 566.6875 , 1.071119201010289e-45 ], 
                    [ 566.75 , 1.0713164995353946e-45 ], [ 566.8125 , 1.071513834402588e-45 ], [ 566.875 , 1.0717112056185556e-45 ], [ 566.9375 , 1.0719086131900158e-45 ], 
                    [ 567.0 , 1.0721060571236272e-45 ], [ 567.0625 , 1.0723031783740243e-45 ], [ 567.125 , 1.0725003239961668e-45 ], [ 567.1875 , 1.0726975058640022e-45 ], 
                    [ 567.25 , 1.0728947239842406e-45 ], [ 567.3125 , 1.0730919783635158e-45 ], [ 567.375 , 1.0732892690085025e-45 ], [ 567.4375 , 1.0734865959258599e-45 ], 
                    [ 567.5 , 1.0736839591222801e-45 ], [ 567.5625 , 1.0738813586044103e-45 ], [ 567.625 , 1.0740787943789291e-45 ], [ 567.6875 , 1.0742762664525166e-45 ], 
                    [ 567.75 , 1.074473774831839e-45 ], [ 567.8125 , 1.0746713195235631e-45 ], [ 567.875 , 1.0748689005343657e-45 ], [ 567.9375 , 1.0750665178709391e-45 ], 
                    [ 568.0 , 1.0752641715399542e-45 ], [ 568.0625 , 1.0754618615480908e-45 ], [ 568.125 , 1.0756595879020298e-45 ], [ 568.1875 , 1.0758573506084615e-45 ], 
                    [ 568.25 , 1.076055149674054e-45 ], [ 568.3125 , 1.0762529547929215e-45 ], [ 568.375 , 1.0764502851960771e-45 ], [ 568.4375 , 1.0766476517796488e-45 ], 
                    [ 568.5 , 1.0768450545502779e-45 ], [ 568.5625 , 1.0770424935145994e-45 ], [ 568.625 , 1.0772399686792494e-45 ], [ 568.6875 , 1.0774374800508802e-45 ], 
                    [ 568.75 , 1.0776350276361002e-45 ], [ 568.8125 , 1.077832611441564e-45 ], [ 568.875 , 1.0780302314739206e-45 ], [ 568.9375 , 1.0782278877397889e-45 ], 
                    [ 569.0 , 1.0784255802458432e-45 ], [ 569.0625 , 1.0786233089987048e-45 ], [ 569.125 , 1.0788210740050275e-45 ], [ 569.1875 , 1.0790188752714659e-45 ], 
                    [ 569.25 , 1.0792167128046454e-45 ], [ 569.3125 , 1.0794145866112305e-45 ], [ 569.375 , 1.0796124966978877e-45 ], [ 569.4375 , 1.0798104430712531e-45 ], 
                    [ 569.5 , 1.0800084257379725e-45 ], [ 569.5625 , 1.0802064447047002e-45 ], [ 569.625 , 1.0804045061187972e-45 ], [ 569.6875 , 1.0806026744792638e-45 ], 
                    [ 569.75 , 1.0808008791878673e-45 ], [ 569.8125 , 1.08099912025129e-45 ], [ 569.875 , 1.0811973976761847e-45 ], [ 569.9375 , 1.0813957114692438e-45 ], 
                    [ 570.0 , 1.0815940616370997e-45 ], [ 570.0625 , 1.0817924481864702e-45 ], [ 570.125 , 1.0819908711239827e-45 ], [ 570.1875 , 1.0821893304563421e-45 ], 
                    [ 570.25 , 1.0823878261902312e-45 ], [ 570.3125 , 1.0825863583322968e-45 ], [ 570.375 , 1.0827849268892239e-45 ], [ 570.4375 , 1.0829835318677153e-45 ], 
                    [ 570.5 , 1.0831821732744357e-45 ], [ 570.5625 , 1.0833808511160595e-45 ], [ 570.625 , 1.0835795653992847e-45 ], [ 570.6875 , 1.0837783161307879e-45 ], 
                    [ 570.75 , 1.0839771033172387e-45 ], [ 570.8125 , 1.0841759269653627e-45 ], [ 570.875 , 1.0843747870818089e-45 ], [ 570.9375 , 1.0845736483149637e-45 ], 
                    [ 571.0 , 1.0847720902039516e-45 ], [ 571.0625 , 1.0849705684013773e-45 ], [ 571.125 , 1.0851690829138995e-45 ], [ 571.1875 , 1.0853676337481475e-45 ], 
                    [ 571.25 , 1.0855662209107745e-45 ], [ 571.3125 , 1.0857648444084274e-45 ], [ 571.375 , 1.0859635042477623e-45 ], [ 571.4375 , 1.0861622004354127e-45 ], 
                    [ 571.5 , 1.086360932978045e-45 ], [ 571.5625 , 1.0865597018822878e-45 ], [ 571.625 , 1.0867585071548168e-45 ], [ 571.6875 , 1.0869573488022869e-45 ], 
                    [ 571.75 , 1.0871562268313375e-45 ], [ 571.8125 , 1.0873551412486488e-45 ], [ 571.875 , 1.0875540920608398e-45 ], [ 571.9375 , 1.0877530792746083e-45 ], 
                    [ 572.0 , 1.0879521028965915e-45 ], [ 572.0625 , 1.0881511629334665e-45 ], [ 572.125 , 1.0883502593918647e-45 ], [ 572.1875 , 1.0885493922784968e-45 ], 
                    [ 572.25 , 1.0887485616000048e-45 ], [ 572.3125 , 1.0889471876463434e-45 ], [ 572.375 , 1.0891458499290542e-45 ], [ 572.4375 , 1.089344548454763e-45 ], 
                    [ 572.5 , 1.0895432832300743e-45 ], [ 572.5625 , 1.0897420542615936e-45 ], [ 572.625 , 1.0899408615559584e-45 ], [ 572.6875 , 1.0901397051197614e-45 ], 
                    [ 572.75 , 1.0903385849596269e-45 ], [ 572.8125 , 1.0905375010821652e-45 ], [ 572.875 , 1.0907364534939955e-45 ], [ 572.9375 , 1.0909354422017694e-45 ], 
                    [ 573.0 , 1.0911344672120772e-45 ], [ 573.0625 , 1.0913335285315578e-45 ], [ 573.125 , 1.0915326261668038e-45 ], [ 573.1875 , 1.0917317601244797e-45 ], 
                    [ 573.25 , 1.0919309304111881e-45 ], [ 573.3125 , 1.0921301370335806e-45 ], [ 573.375 , 1.0923293799982623e-45 ], [ 573.4375 , 1.0925286593118639e-45 ], 
                    [ 573.5 , 1.0927279749810238e-45 ], [ 573.5625 , 1.0929273270123828e-45 ], [ 573.625 , 1.0931268267836906e-45 ], [ 573.6875 , 1.0933263781608604e-45 ], 
                    [ 573.75 , 1.0935259659663227e-45 ], [ 573.8125 , 1.0937255902067427e-45 ], [ 573.875 , 1.0939252508887488e-45 ], [ 573.9375 , 1.0941249480190241e-45 ], 
                    [ 574.0 , 1.0943246816041914e-45 ], [ 574.0625 , 1.0945244516509445e-45 ], [ 574.125 , 1.0947242581659003e-45 ], [ 574.1875 , 1.0949241011557321e-45 ], 
                    [ 574.25 , 1.0951239806270982e-45 ], [ 574.3125 , 1.0953238965866582e-45 ], [ 574.375 , 1.0955238490410888e-45 ], [ 574.4375 , 1.0957238379970133e-45 ], 
                    [ 574.5 , 1.0959238634611262e-45 ], [ 574.5625 , 1.096123925440092e-45 ], [ 574.625 , 1.096324023940561e-45 ], [ 574.6875 , 1.0965241589692158e-45 ], 
                    [ 574.75 , 1.0967243305327091e-45 ], [ 574.8125 , 1.0969245386377183e-45 ], [ 574.875 , 1.0971247832909136e-45 ], [ 574.9375 , 1.097324530517056e-45 ], 
                    [ 575.0 , 1.0975241145640292e-45 ], [ 575.0625 , 1.097723734911856e-45 ], [ 575.125 , 1.0979233915670925e-45 ], [ 575.1875 , 1.0981230845363891e-45 ], 
                    [ 575.25 , 1.0983228138263195e-45 ], [ 575.3125 , 1.09852257944349e-45 ], [ 575.375 , 1.0987223813945151e-45 ], [ 575.4375 , 1.0989222196860194e-45 ], 
                    [ 575.5 , 1.0991220943245812e-45 ], [ 575.5625 , 1.0993220053168192e-45 ], [ 575.625 , 1.0995219526693845e-45 ], [ 575.6875 , 1.099721936388836e-45 ], 
                    [ 575.75 , 1.0999219564818113e-45 ], [ 575.8125 , 1.1001220129549339e-45 ], [ 575.875 , 1.1003221058148131e-45 ], [ 575.9375 , 1.1005222350680514e-45 ], 
                    [ 576.0 , 1.1007224007212993e-45 ], [ 576.0625 , 1.1009226027811538e-45 ], [ 576.125 , 1.1011228412542523e-45 ], [ 576.1875 , 1.1013231161472099e-45 ], 
                    [ 576.25 , 1.1015233461926152e-45 ], [ 576.3125 , 1.1017235351631877e-45 ], [ 576.375 , 1.1019237605157537e-45 ], [ 576.4375 , 1.1021240222569333e-45 ], 
                    [ 576.5 , 1.1023243203933317e-45 ], [ 576.5625 , 1.1025246549315867e-45 ], [ 576.625 , 1.1027250258782905e-45 ], [ 576.6875 , 1.1029254332400677e-45 ], 
                    [ 576.75 , 1.1031258770235442e-45 ], [ 576.8125 , 1.1033263572353314e-45 ], [ 576.875 , 1.1035268738820419e-45 ], [ 576.9375 , 1.1037274269702973e-45 ], 
                    [ 577.0 , 1.1039280165067516e-45 ], [ 577.0625 , 1.1041286424979821e-45 ], [ 577.125 , 1.104329304950661e-45 ], [ 577.1875 , 1.1045300038713753e-45 ], 
                    [ 577.25 , 1.1047307392667606e-45 ], [ 577.3125 , 1.1049315111434615e-45 ], [ 577.375 , 1.1051323195081003e-45 ], [ 577.4375 , 1.1053331643673158e-45 ], 
                    [ 577.5 , 1.105534045727725e-45 ], [ 577.5625 , 1.1057348514191725e-45 ], [ 577.625 , 1.1059353533735987e-45 ], [ 577.6875 , 1.1061358916848805e-45 ], 
                    [ 577.75 , 1.1063364663596026e-45 ], [ 577.8125 , 1.1065370774043667e-45 ], [ 577.875 , 1.1067377248257522e-45 ], [ 577.9375 , 1.106938408630355e-45 ], 
                    [ 578.0 , 1.1071391288247882e-45 ], [ 578.0625 , 1.1073398854156505e-45 ], [ 578.125 , 1.1075406784095255e-45 ], [ 578.1875 , 1.1077415078130145e-45 ], 
                    [ 578.25 , 1.1079423736327352e-45 ], [ 578.3125 , 1.1081432758752755e-45 ], [ 578.375 , 1.1083442145472709e-45 ], [ 578.4375 , 1.1085451896552804e-45 ], 
                    [ 578.5 , 1.1087462012059497e-45 ], [ 578.5625 , 1.1089472492058717e-45 ], [ 578.625 , 1.1091483336616475e-45 ], [ 578.6875 , 1.1093494545798877e-45 ], 
                    [ 578.75 , 1.1095506119672198e-45 ], [ 578.8125 , 1.1097518058302566e-45 ], [ 578.875 , 1.1099530361755966e-45 ], [ 578.9375 , 1.1101542745547615e-45 ], 
                    [ 579.0 , 1.110355547473169e-45 ], [ 579.0625 , 1.110556856882694e-45 ], [ 579.125 , 1.1107582027899683e-45 ], [ 579.1875 , 1.110959585201617e-45 ], 
                    [ 579.25 , 1.1111610041242423e-45 ], [ 579.3125 , 1.1113624595644877e-45 ], [ 579.375 , 1.1115639515289422e-45 ], [ 579.4375 , 1.1117654800242434e-45 ], 
                    [ 579.5 , 1.1119670450570304e-45 ], [ 579.5625 , 1.1121686466338958e-45 ], [ 579.625 , 1.1123702847614808e-45 ], [ 579.6875 , 1.1125719594464201e-45 ], 
                    [ 579.75 , 1.1127736706953418e-45 ], [ 579.8125 , 1.1129754185148433e-45 ], [ 579.875 , 1.1131772029115941e-45 ], [ 579.9375 , 1.1133790238922184e-45 ], 
                    [ 580.0 , 1.1135808814633327e-45 ], [ 580.0625 , 1.1137827756315788e-45 ], [ 580.125 , 1.113984706403576e-45 ], [ 580.1875 , 1.1141866737860007e-45 ], 
                    [ 580.25 , 1.1143882640540622e-45 ], [ 580.3125 , 1.1145895861886869e-45 ], [ 580.375 , 1.1147909446935577e-45 ], [ 580.4375 , 1.1149923395752928e-45 ], 
                    [ 580.5 , 1.1151937708404162e-45 ], [ 580.5625 , 1.1153952384955404e-45 ], [ 580.625 , 1.1155967425472239e-45 ], [ 580.6875 , 1.1157982830020338e-45 ], 
                    [ 580.75 , 1.1159998598665627e-45 ], [ 580.8125 , 1.1162014731473721e-45 ], [ 580.875 , 1.116403122851065e-45 ], [ 580.9375 , 1.1166048089841815e-45 ], 
                    [ 581.0 , 1.1168065315533428e-45 ], [ 581.0625 , 1.1170082905651153e-45 ], [ 581.125 , 1.1172100860260664e-45 ], [ 581.1875 , 1.1174119179428212e-45 ], 
                    [ 581.25 , 1.1176137863219258e-45 ], [ 581.3125 , 1.117815691169983e-45 ], [ 581.375 , 1.1180176324935895e-45 ], [ 581.4375 , 1.1182196102993186e-45 ], 
                    [ 581.5 , 1.118421624593785e-45 ], [ 581.5625 , 1.1186236919579154e-45 ], [ 581.625 , 1.1188258779273622e-45 ], [ 581.6875 , 1.1190281004409641e-45 ], 
                    [ 581.75 , 1.1192303595053338e-45 ], [ 581.8125 , 1.1194326551270777e-45 ], [ 581.875 , 1.1196349873128036e-45 ], [ 581.9375 , 1.119837356069136e-45 ], 
                    [ 582.0 , 1.1200397614026533e-45 ], [ 582.0625 , 1.1202422033199821e-45 ], [ 582.125 , 1.120444681827743e-45 ], [ 582.1875 , 1.1206471969325337e-45 ], 
                    [ 582.25 , 1.120849748640969e-45 ], [ 582.3125 , 1.1210523369596962e-45 ], [ 582.375 , 1.1212549618952771e-45 ], [ 582.4375 , 1.1214576234543859e-45 ], 
                    [ 582.5 , 1.1216603216436258e-45 ], [ 582.5625 , 1.1218630564696017e-45 ], [ 582.625 , 1.1220658279389358e-45 ], [ 582.6875 , 1.1222686360582588e-45 ], 
                    [ 582.75 , 1.1224714808341953e-45 ], [ 582.8125 , 1.1226743622733707e-45 ], [ 582.875 , 1.1228772803824279e-45 ], [ 582.9375 , 1.1230797220242887e-45 ], 
                    [ 583.0 , 1.1232819925629369e-45 ], [ 583.0625 , 1.1234842995312077e-45 ], [ 583.125 , 1.1236866429356629e-45 ], [ 583.1875 , 1.1238890227828723e-45 ], 
                    [ 583.25 , 1.1240914390793835e-45 ], [ 583.3125 , 1.124293891831761e-45 ], [ 583.375 , 1.1244963810465949e-45 ], [ 583.4375 , 1.124698906730436e-45 ], 
                    [ 583.5 , 1.1249014688898447e-45 ], [ 583.5625 , 1.1251040675313981e-45 ], [ 583.625 , 1.1253067026616589e-45 ], [ 583.6875 , 1.125509374287223e-45 ], 
                    [ 583.75 , 1.1257120824146469e-45 ], [ 583.8125 , 1.1259148270504893e-45 ], [ 583.875 , 1.126117608201357e-45 ], [ 583.9375 , 1.1263204258737947e-45 ], 
                    [ 584.0 , 1.126523280074412e-45 ], [ 584.0625 , 1.1267261708097639e-45 ], [ 584.125 , 1.1269290980864302e-45 ], [ 584.1875 , 1.1271320619110163e-45 ], 
                    [ 584.25 , 1.1273350438934449e-45 ], [ 584.3125 , 1.1275379923689775e-45 ], [ 584.375 , 1.1277409773802986e-45 ], [ 584.4375 , 1.1279439989339937e-45 ], 
                    [ 584.5 , 1.1281470570366493e-45 ], [ 584.5625 , 1.1283501516948206e-45 ], [ 584.625 , 1.128553282915121e-45 ], [ 584.6875 , 1.1287564507041087e-45 ], 
                    [ 584.75 , 1.1289596550683747e-45 ], [ 584.8125 , 1.1291628960145034e-45 ], [ 584.875 , 1.1293661735490887e-45 ], [ 584.9375 , 1.1295694876787093e-45 ], 
                    [ 585.0 , 1.1297728384099372e-45 ], [ 585.0625 , 1.1299762257494018e-45 ], [ 585.125 , 1.1301796497036451e-45 ], [ 585.1875 , 1.1303831102792991e-45 ], 
                    [ 585.25 , 1.1305866074829323e-45 ], [ 585.3125 , 1.1307901413211469e-45 ], [ 585.375 , 1.1309937118005538e-45 ], [ 585.4375 , 1.1311973189277093e-45 ], 
                    [ 585.5 , 1.1314009627092511e-45 ], [ 585.5625 , 1.1316046431517458e-45 ], [ 585.625 , 1.1318081280714666e-45 ], [ 585.6875 , 1.1320115278015666e-45 ], 
                    [ 585.75 , 1.1322149640850795e-45 ], [ 585.8125 , 1.1324184369285667e-45 ], [ 585.875 , 1.1326219463386147e-45 ], [ 585.9375 , 1.1328254923217869e-45 ], 
                    [ 586.0 , 1.1330290748846477e-45 ], [ 586.0625 , 1.1332326940337869e-45 ], [ 586.125 , 1.1334363497757716e-45 ], [ 586.1875 , 1.133640042117178e-45 ], 
                    [ 586.25 , 1.1338437710645836e-45 ], [ 586.3125 , 1.1340475366245506e-45 ], [ 586.375 , 1.1342513388037073e-45 ], [ 586.4375 , 1.1344551776085778e-45 ], 
                    [ 586.5 , 1.1346590530457849e-45 ], [ 586.5625 , 1.1348629651218795e-45 ], [ 586.625 , 1.135066913843462e-45 ], [ 586.6875 , 1.1352708992171346e-45 ], 
                    [ 586.75 , 1.1354749212494432e-45 ], [ 586.8125 , 1.1356789799470248e-45 ], [ 586.875 , 1.1358830753164279e-45 ], [ 586.9375 , 1.1360871761066846e-45 ], 
                    [ 587.0 , 1.1362907865624672e-45 ], [ 587.0625 , 1.1364944335094792e-45 ], [ 587.125 , 1.1366981169542523e-45 ], [ 587.1875 , 1.1369018369033442e-45 ], 
                    [ 587.25 , 1.1371055933632967e-45 ], [ 587.3125 , 1.1373093863406377e-45 ], [ 587.375 , 1.1375132158419198e-45 ], [ 587.4375 , 1.1377170818736966e-45 ], 
                    [ 587.5 , 1.1379209844424993e-45 ], [ 587.5625 , 1.1381249235548841e-45 ], [ 587.625 , 1.1383288992174084e-45 ], [ 587.6875 , 1.1385329114366066e-45 ], 
                    [ 587.75 , 1.1387369602190384e-45 ], [ 587.8125 , 1.138941045571257e-45 ], [ 587.875 , 1.1391451674998e-45 ], [ 587.9375 , 1.1393493260112713e-45 ], 
                    [ 588.0 , 1.1395535211121868e-45 ], [ 588.0625 , 1.1397577528090962e-45 ], [ 588.125 , 1.1399620211085821e-45 ], [ 588.1875 , 1.1401663260172049e-45 ], 
                    [ 588.25 , 1.1403706675415094e-45 ], [ 588.3125 , 1.1405750168413678e-45 ], [ 588.375 , 1.1407793580098918e-45 ], [ 588.4375 , 1.1409837357874095e-45 ], 
                    [ 588.5 , 1.1411881501804964e-45 ], [ 588.5625 , 1.1413926011956955e-45 ], [ 588.625 , 1.1415970888395847e-45 ], [ 588.6875 , 1.1418016131187098e-45 ], 
                    [ 588.75 , 1.1420061740396424e-45 ], [ 588.8125 , 1.1422107716089473e-45 ], [ 588.875 , 1.142415405833206e-45 ], [ 588.9375 , 1.1426200767189533e-45 ], 
                    [ 589.0 , 1.142824784272782e-45 ], [ 589.0625 , 1.1430295285012365e-45 ], [ 589.125 , 1.1432343094109202e-45 ], [ 589.1875 , 1.1434391270083801e-45 ], 
                    [ 589.25 , 1.1436439813001974e-45 ], [ 589.3125 , 1.1438488722929541e-45 ], [ 589.375 , 1.1440537999932174e-45 ], [ 589.4375 , 1.1442587644075473e-45 ], 
                    [ 589.5 , 1.1444637655425701e-45 ], [ 589.5625 , 1.1446688034047997e-45 ], [ 589.625 , 1.1448738780008727e-45 ], [ 589.6875 , 1.1450786903140265e-45 ], 
                    [ 589.75 , 1.145283408635868e-45 ], [ 589.8125 , 1.1454881635574576e-45 ], [ 589.875 , 1.1456929550853542e-45 ], [ 589.9375 , 1.1458977832260706e-45 ], 
                    [ 590.0 , 1.1461026479861762e-45 ], [ 590.0625 , 1.1463075493722101e-45 ], [ 590.125 , 1.1465124873907122e-45 ], [ 590.1875 , 1.1467174620482638e-45 ], 
                    [ 590.25 , 1.1469224733513671e-45 ], [ 590.3125 , 1.1471275213066058e-45 ], [ 590.375 , 1.1473326059205325e-45 ], [ 590.4375 , 1.1475377271996769e-45 ], 
                    [ 590.5 , 1.1477428851506101e-45 ], [ 590.5625 , 1.1479480797798886e-45 ], [ 590.625 , 1.1481533110940858e-45 ], [ 590.6875 , 1.1483585790997362e-45 ], 
                    [ 590.75 , 1.1485638838034073e-45 ], [ 590.8125 , 1.1487692252116437e-45 ], [ 590.875 , 1.1489746033310568e-45 ], [ 590.9375 , 1.1491800181681607e-45 ], 
                    [ 591.0 , 1.1493854697295194e-45 ], [ 591.0625 , 1.1495906220001735e-45 ], [ 591.125 , 1.1497957879082517e-45 ], [ 591.1875 , 1.1500009904320083e-45 ], 
                    [ 591.25 , 1.1502062295780025e-45 ], [ 591.3125 , 1.1504115053527374e-45 ], [ 591.375 , 1.1506168177627666e-45 ], [ 591.4375 , 1.1508221668146284e-45 ], 
                    [ 591.5 , 1.1510275525148704e-45 ], [ 591.5625 , 1.1512329748700166e-45 ], [ 591.625 , 1.1514384338866252e-45 ], [ 591.6875 , 1.1516439295712313e-45 ], 
                    [ 591.75 , 1.1518494619303701e-45 ], [ 591.8125 , 1.1520550309705957e-45 ], [ 591.875 , 1.1522606366984623e-45 ], [ 591.9375 , 1.1524662791205097e-45 ], 
                    [ 592.0 , 1.1526719582432865e-45 ], [ 592.0625 , 1.152877674073334e-45 ], [ 592.125 , 1.1530834266172121e-45 ], [ 592.1875 , 1.1532892158814892e-45 ], 
                    [ 592.25 , 1.1534950418726613e-45 ], [ 592.3125 , 1.1537009045973733e-45 ], [ 592.375 , 1.153906751286369e-45 ], [ 592.4375 , 1.1541122751043444e-45 ], 
                    [ 592.5 , 1.1543178355284404e-45 ], [ 592.5625 , 1.1545234325651681e-45 ], [ 592.625 , 1.1547290662210488e-45 ], [ 592.6875 , 1.1549347365026051e-45 ], 
                    [ 592.75 , 1.1551404434163517e-45 ], [ 592.8125 , 1.15534618696883e-45 ], [ 592.875 , 1.1555519671665495e-45 ], [ 592.9375 , 1.155757784016045e-45 ], 
                    [ 593.0 , 1.1559636375238527e-45 ], [ 593.0625 , 1.1561695276964776e-45 ], [ 593.125 , 1.156375454540491e-45 ], [ 593.1875 , 1.1565814180623918e-45 ], 
                    [ 593.25 , 1.1567874182687206e-45 ], [ 593.3125 , 1.1569934551660118e-45 ], [ 593.375 , 1.1571995287608e-45 ], [ 593.4375 , 1.1574056390596053e-45 ], 
                    [ 593.5 , 1.1576117860689894e-45 ], [ 593.5625 , 1.1578179697954913e-45 ], [ 593.625 , 1.1580241902456424e-45 ], [ 593.6875 , 1.1582304474259753e-45 ], 
                    [ 593.75 , 1.1584366788442725e-45 ], [ 593.8125 , 1.1586427796776577e-45 ], [ 593.875 , 1.1588489171790404e-45 ], [ 593.9375 , 1.1590550913549529e-45 ], 
                    [ 594.0 , 1.1592613022119118e-45 ], [ 594.0625 , 1.1594675497564346e-45 ], [ 594.125 , 1.1596738339950734e-45 ], [ 594.1875 , 1.1598801549343321e-45 ], 
                    [ 594.25 , 1.1600865125807401e-45 ], [ 594.3125 , 1.1602929069408358e-45 ], [ 594.375 , 1.1604993380211599e-45 ], [ 594.4375 , 1.1607058058282369e-45 ], 
                    [ 594.5 , 1.1609123103685927e-45 ], [ 594.5625 , 1.1611188516487873e-45 ], [ 594.625 , 1.1613254296753245e-45 ], [ 594.6875 , 1.1615320444547582e-45 ], 
                    [ 594.75 , 1.1617386959936439e-45 ], [ 594.8125 , 1.1619453842984882e-45 ], [ 594.875 , 1.1621521093758573e-45 ], [ 594.9375 , 1.1623588712322853e-45 ], 
                    [ 595.0 , 1.162565669874299e-45 ], [ 595.0625 , 1.1627725053084845e-45 ], [ 595.125 , 1.1629792248404469e-45 ], [ 595.1875 , 1.1631857188295585e-45 ], 
                    [ 595.25 , 1.1633922494829256e-45 ], [ 595.3125 , 1.1635988168070498e-45 ], [ 595.375 , 1.1638054208084593e-45 ], [ 595.4375 , 1.1640120614936575e-45 ], 
                    [ 595.5 , 1.1642187388691581e-45 ], [ 595.5625 , 1.1644254529414674e-45 ], [ 595.625 , 1.1646322037171177e-45 ], [ 595.6875 , 1.1648389912026092e-45 ], 
                    [ 595.75 , 1.165045815404485e-45 ], [ 595.8125 , 1.1652526763292475e-45 ], [ 595.875 , 1.1654595739833843e-45 ], [ 595.9375 , 1.1656665083734911e-45 ], 
                    [ 596.0 , 1.1658734795060163e-45 ], [ 596.0625 , 1.1660804873875414e-45 ], [ 596.125 , 1.1662875320245423e-45 ], [ 596.1875 , 1.1664946134236027e-45 ], 
                    [ 596.25 , 1.1667017315911836e-45 ], [ 596.3125 , 1.1669088865338634e-45 ], [ 596.375 , 1.1671160782581633e-45 ], [ 596.4375 , 1.1673233067705892e-45 ], 
                    [ 596.5 , 1.1675304127268187e-45 ], [ 596.5625 , 1.1677373316839502e-45 ], [ 596.625 , 1.1679442873129011e-45 ], [ 596.6875 , 1.1681512796201377e-45 ], 
                    [ 596.75 , 1.1683583086122016e-45 ], [ 596.8125 , 1.1685653742955696e-45 ], [ 596.875 , 1.1687724766767528e-45 ], [ 596.9375 , 1.1689796157622633e-45 ], 
                    [ 597.0 , 1.1691867915585896e-45 ], [ 597.0625 , 1.1693940040722462e-45 ], [ 597.125 , 1.16960125330974e-45 ], [ 597.1875 , 1.1698085392775885e-45 ], 
                    [ 597.25 , 1.1700158619822842e-45 ], [ 597.3125 , 1.1702232214303549e-45 ], [ 597.375 , 1.1704306176282957e-45 ], [ 597.4375 , 1.1706380505826196e-45 ], 
                    [ 597.5 , 1.170845520299866e-45 ], [ 597.5625 , 1.1710530267865168e-45 ], [ 597.625 , 1.1712605700491306e-45 ], [ 597.6875 , 1.1714681500941832e-45 ], 
                    [ 597.75 , 1.1716757669282184e-45 ], [ 597.8125 , 1.1718834205577316e-45 ], [ 597.875 , 1.1720909442141701e-45 ], [ 597.9375 , 1.1722982544131491e-45 ], 
                    [ 598.0 , 1.1725056012795203e-45 ], [ 598.0625 , 1.1727129848197693e-45 ], [ 598.125 , 1.1729204050403828e-45 ], [ 598.1875 , 1.1731278619478568e-45 ], 
                    [ 598.25 , 1.1733353555486636e-45 ], [ 598.3125 , 1.1735428858493013e-45 ], [ 598.375 , 1.1737504528562531e-45 ], [ 598.4375 , 1.1739580565760194e-45 ], 
                    [ 598.5 , 1.1741656970151105e-45 ], [ 598.5625 , 1.1743733741800043e-45 ], [ 598.625 , 1.1745810880771797e-45 ], [ 598.6875 , 1.1747888387131675e-45 ], 
                    [ 598.75 , 1.1749966260944318e-45 ], [ 598.8125 , 1.1752044502274972e-45 ], [ 598.875 , 1.1754123111188471e-45 ], [ 598.9375 , 1.175620208775e-45 ], 
                    [ 599.0 , 1.1758281432024586e-45 ], [ 599.0625 , 1.1760361144077014e-45 ], [ 599.125 , 1.1762441223972585e-45 ], [ 599.1875 , 1.1764521671776193e-45 ], 
                    [ 599.25 , 1.1766602204113695e-45 ], [ 599.3125 , 1.176868254415948e-45 ], [ 599.375 , 1.1770763252010229e-45 ], [ 599.4375 , 1.1772844327730886e-45 ], 
                    [ 599.5 , 1.1774925771386744e-45 ], [ 599.5625 , 1.17770075830426e-45 ], [ 599.625 , 1.1779089762763603e-45 ], [ 599.6875 , 1.1781172310614826e-45 ], 
                    [ 599.75 , 1.1783255226661355e-45 ], [ 599.8125 , 1.1785338510968284e-45 ], [ 599.875 , 1.1787422163600725e-45 ], [ 599.9375 , 1.1789506184623799e-45 ], 
                    [ 600.0 , 1.1791590574102633e-45 ], [ 600.0625 , 1.1793675332102375e-45 ], [ 600.125 , 1.1795760458688011e-45 ], [ 600.1875 , 1.1797845953925206e-45 ], 
                    [ 600.25 , 1.1799931817878638e-45 ], [ 600.3125 , 1.1802018050613666e-45 ], [ 600.375 , 1.1804104652195403e-45 ], [ 600.4375 , 1.1806191622689231e-45 ], 
                    [ 600.5 , 1.1808278962160372e-45 ], [ 600.5625 , 1.1810366670673812e-45 ], [ 600.625 , 1.1812453378738678e-45 ], [ 600.6875 , 1.181453611779842e-45 ], 
                    [ 600.75 , 1.1816619224081114e-45 ], [ 600.8125 , 1.1818702697651084e-45 ], [ 600.875 , 1.1820786538573426e-45 ], [ 600.9375 , 1.1822870746912828e-45 ], 
                    [ 601.0 , 1.1824955322734156e-45 ], [ 601.0625 , 1.1827040266102118e-45 ], [ 601.125 , 1.1829125577081268e-45 ], [ 601.1875 , 1.183121125573701e-45 ], 
                    [ 601.25 , 1.1833297302133665e-45 ], [ 601.3125 , 1.1835383716336245e-45 ], [ 601.375 , 1.1837470498409597e-45 ], [ 601.4375 , 1.1839557648418583e-45 ], 
                    [ 601.5 , 1.1841645166428077e-45 ], [ 601.5625 , 1.184373305250305e-45 ], [ 601.625 , 1.1845821306708226e-45 ], [ 601.6875 , 1.1847909929108517e-45 ], 
                    [ 601.75 , 1.1849998919769092e-45 ], [ 601.8125 , 1.1852088278754546e-45 ], [ 601.875 , 1.185417800613024e-45 ], [ 601.9375 , 1.185626810196071e-45 ], 
                    [ 602.0 , 1.1858358255914489e-45 ], [ 602.0625 , 1.1860445208153392e-45 ], [ 602.125 , 1.1862532527674726e-45 ], [ 602.1875 , 1.1864620214543717e-45 ], 
                    [ 602.25 , 1.1866708268824514e-45 ], [ 602.3125 , 1.1868796690581857e-45 ], [ 602.375 , 1.1870885479880758e-45 ], [ 602.4375 , 1.1872974636785563e-45 ], 
                    [ 602.5 , 1.1875064161360968e-45 ], [ 602.5625 , 1.1877154053671845e-45 ], [ 602.625 , 1.187924431378266e-45 ], [ 602.6875 , 1.1881334941758481e-45 ], 
                    [ 602.75 , 1.1883425937663877e-45 ], [ 602.8125 , 1.1885517301563688e-45 ], [ 602.875 , 1.1887609033522253e-45 ], [ 602.9375 , 1.1889701133604936e-45 ], 
                    [ 603.0 , 1.1891793601876277e-45 ], [ 603.0625 , 1.189388643840098e-45 ], [ 603.125 , 1.1895979643244025e-45 ], [ 603.1875 , 1.18980732164699e-45 ], 
                    [ 603.25 , 1.1900167158144109e-45 ], [ 603.3125 , 1.1902261468330816e-45 ], [ 603.375 , 1.1904356147095219e-45 ], [ 603.4375 , 1.190645043501941e-45 ], 
                    [ 603.5 , 1.1908544963597906e-45 ], [ 603.5625 , 1.1910639860636335e-45 ], [ 603.625 , 1.191273512619935e-45 ], [ 603.6875 , 1.1914830760352034e-45 ], 
                    [ 603.75 , 1.191692676315914e-45 ], [ 603.8125 , 1.1919023134685693e-45 ], [ 603.875 , 1.1921119874996213e-45 ], [ 603.9375 , 1.1923216984155663e-45 ], 
                    [ 604.0 , 1.1925314462229184e-45 ], [ 604.0625 , 1.1927412309281417e-45 ], [ 604.125 , 1.1929510525377441e-45 ], [ 604.1875 , 1.1931609110582094e-45 ], 
                    [ 604.25 , 1.1933708064960137e-45 ], [ 604.3125 , 1.1935807388576852e-45 ], [ 604.375 , 1.1937907081496852e-45 ], [ 604.4375 , 1.1940007143785448e-45 ], 
                    [ 604.5 , 1.1942107575507445e-45 ], [ 604.5625 , 1.194420837672766e-45 ], [ 604.625 , 1.1946309547511523e-45 ], [ 604.6875 , 1.1948411087923618e-45 ], 
                    [ 604.75 , 1.1950512998029054e-45 ], [ 604.8125 , 1.1952612617470299e-45 ], [ 604.875 , 1.1954710831873327e-45 ], [ 604.9375 , 1.1956809414606093e-45 ], 
                    [ 605.0 , 1.1958908365733252e-45 ], [ 605.0625 , 1.1961007685319648e-45 ], [ 605.125 , 1.1963107373429961e-45 ], [ 605.1875 , 1.1965207430128456e-45 ], 
                    [ 605.25 , 1.1967307855480433e-45 ], [ 605.3125 , 1.1969408649550436e-45 ], [ 605.375 , 1.1971509812403024e-45 ], [ 605.4375 , 1.1973611344102933e-45 ], 
                    [ 605.5 , 1.1975713244715169e-45 ], [ 605.5625 , 1.1977815514304147e-45 ], [ 605.625 , 1.19799181529349e-45 ], [ 605.6875 , 1.1982021160671953e-45 ], 
                    [ 605.75 , 1.1984124537580525e-45 ], [ 605.8125 , 1.1986228283725e-45 ], [ 605.875 , 1.1988332399170619e-45 ], [ 605.9375 , 1.1990436883981697e-45 ], 
                    [ 606.0 , 1.199254173822342e-45 ], [ 606.0625 , 1.1994646961960467e-45 ], [ 606.125 , 1.199675255525787e-45 ], [ 606.1875 , 1.1998857382435251e-45 ], 
                    [ 606.25 , 1.2000959657904851e-45 ], [ 606.3125 , 1.2003062301706455e-45 ], [ 606.375 , 1.2005165313904424e-45 ], [ 606.4375 , 1.2007268694563475e-45 ], 
                    [ 606.5 , 1.200937244374791e-45 ], [ 606.5625 , 1.2011476561522553e-45 ], [ 606.625 , 1.2013581047952067e-45 ], [ 606.6875 , 1.2015685903100617e-45 ], 
                    [ 606.75 , 1.2017791127033146e-45 ], [ 606.8125 , 1.2019896719814267e-45 ], [ 606.875 , 1.2022002681508434e-45 ], [ 606.9375 , 1.2024109012180284e-45 ], 
                    [ 607.0 , 1.2026215711894717e-45 ], [ 607.0625 , 1.2028322780716142e-45 ], [ 607.125 , 1.2030430218709308e-45 ], [ 607.1875 , 1.2032538025938984e-45 ], 
                    [ 607.25 , 1.2034646202469777e-45 ], [ 607.3125 , 1.203675474836622e-45 ], [ 607.375 , 1.203886366369337e-45 ], [ 607.4375 , 1.2040972948515697e-45 ], 
                    [ 607.5 , 1.2043082602898025e-45 ], [ 607.5625 , 1.2045192626905187e-45 ], [ 607.625 , 1.2047297972472959e-45 ], [ 607.6875 , 1.2049403294294708e-45 ], 
                    [ 607.75 , 1.2051508984031318e-45 ], [ 607.8125 , 1.2053615041747085e-45 ], [ 607.875 , 1.2055721467506231e-45 ], [ 607.9375 , 1.2057828261373243e-45 ], 
                    [ 608.0 , 1.2059935423412363e-45 ], [ 608.0625 , 1.2062042953687847e-45 ], [ 608.125 , 1.2064150852264129e-45 ], [ 608.1875 , 1.2066259119205747e-45 ], 
                    [ 608.25 , 1.2068367754576727e-45 ], [ 608.3125 , 1.2070476758441885e-45 ], [ 608.375 , 1.2072586130865018e-45 ], [ 608.4375 , 1.2074695871911132e-45 ], 
                    [ 608.5 , 1.20768059816443e-45 ], [ 608.5625 , 1.2078916460128956e-45 ], [ 608.625 , 1.2081027307429627e-45 ], [ 608.6875 , 1.2083138523610847e-45 ], 
                    [ 608.75 , 1.2085250108737168e-45 ], [ 608.8125 , 1.2087362062872552e-45 ], [ 608.875 , 1.2089474386081907e-45 ], [ 608.9375 , 1.2091587078429826e-45 ], 
                    [ 609.0 , 1.209369967435801e-45 ], [ 609.0625 , 1.2095812166175509e-45 ], [ 609.125 , 1.2097925026996939e-45 ], [ 609.1875 , 1.2100038256886669e-45 ], 
                    [ 609.25 , 1.2102151855909168e-45 ], [ 609.3125 , 1.2104265824129001e-45 ], [ 609.375 , 1.2106380161610316e-45 ], [ 609.4375 , 1.2108494868418045e-45 ], 
                    [ 609.5 , 1.2110609944616525e-45 ], [ 609.5625 , 1.2112725390270281e-45 ], [ 609.625 , 1.2114841205443851e-45 ], [ 609.6875 , 1.2116957390201691e-45 ], 
                    [ 609.75 , 1.2119073944608623e-45 ], [ 609.8125 , 1.2121190868728869e-45 ], [ 609.875 , 1.2123308162627267e-45 ], [ 609.9375 , 1.2125425826368235e-45 ], 
                    [ 610.0 , 1.2127543860016638e-45 ], [ 610.0625 , 1.2129662263636917e-45 ], [ 610.125 , 1.2131781037293439e-45 ], [ 610.1875 , 1.2133900181051358e-45 ], 
                    [ 610.25 , 1.2136019694974803e-45 ], [ 610.3125 , 1.2138139579128953e-45 ], [ 610.375 , 1.2140259696409038e-45 ], [ 610.4375 , 1.2142376033910834e-45 ], 
                    [ 610.5 , 1.2144492740340937e-45 ], [ 610.5625 , 1.2146609815763573e-45 ], [ 610.625 , 1.2148727260242984e-45 ], [ 610.6875 , 1.2150845073843677e-45 ], 
                    [ 610.75 , 1.2152963256629738e-45 ], [ 610.8125 , 1.2155081808665955e-45 ], [ 610.875 , 1.2157200730016268e-45 ], [ 610.9375 , 1.2159320020745316e-45 ], 
                    [ 611.0 , 1.2161439680917232e-45 ], [ 611.0625 , 1.2163559710596849e-45 ], [ 611.125 , 1.2165680109848149e-45 ], [ 611.1875 , 1.2167800878735817e-45 ], 
                    [ 611.25 , 1.2169922017324032e-45 ], [ 611.3125 , 1.2172043525677586e-45 ], [ 611.375 , 1.2174165403860677e-45 ], [ 611.4375 , 1.2176287651938123e-45 ], 
                    [ 611.5 , 1.21784102699738e-45 ], [ 611.5625 , 1.2180533258032803e-45 ], [ 611.625 , 1.2182656616179208e-45 ], [ 611.6875 , 1.2184780344477701e-45 ], 
                    [ 611.75 , 1.2186904442992895e-45 ], [ 611.8125 , 1.2189026842799681e-45 ], [ 611.875 , 1.2191147702062721e-45 ], [ 611.9375 , 1.219326893034955e-45 ], 
                    [ 612.0 , 1.2195390527724892e-45 ], [ 612.0625 , 1.2197512494252451e-45 ], [ 612.125 , 1.2199634829996894e-45 ], [ 612.1875 , 1.2201757535022199e-45 ], 
                    [ 612.25 , 1.2203880609392708e-45 ], [ 612.3125 , 1.2206004053172775e-45 ], [ 612.375 , 1.2208127866426502e-45 ], [ 612.4375 , 1.2210252049218435e-45 ], 
                    [ 612.5 , 1.2212376601612528e-45 ], [ 612.5625 , 1.2214501523673174e-45 ], [ 612.625 , 1.2216626815464959e-45 ], [ 612.6875 , 1.2218752477051952e-45 ], 
                    [ 612.75 , 1.2220878508498672e-45 ], [ 612.8125 , 1.2223004909869296e-45 ], [ 612.875 , 1.2225131681228455e-45 ], [ 612.9375 , 1.2227258822640177e-45 ], 
                    [ 613.0 , 1.2229386334169022e-45 ], [ 613.0625 , 1.2231514215879478e-45 ], [ 613.125 , 1.2233642467835697e-45 ], [ 613.1875 , 1.2235771090102359e-45 ], 
                    [ 613.25 , 1.2237896016855441e-45 ], [ 613.3125 , 1.224002107139529e-45 ], [ 613.375 , 1.2242146494941376e-45 ], [ 613.4375 , 1.2244272287557345e-45 ], 
                    [ 613.5 , 1.2246398449307542e-45 ], [ 613.5625 , 1.2248524980255982e-45 ], [ 613.625 , 1.2250651880466683e-45 ], [ 613.6875 , 1.2252779150003943e-45 ], 
                    [ 613.75 , 1.225490678893172e-45 ], [ 613.8125 , 1.2257034797314328e-45 ], [ 613.875 , 1.2259163175215837e-45 ], [ 613.9375 , 1.2261291922700324e-45 ], 
                    [ 614.0 , 1.2263421039832139e-45 ], [ 614.0625 , 1.2265550526675384e-45 ], [ 614.125 , 1.226768038329417e-45 ], [ 614.1875 , 1.2269810609752969e-45 ], 
                    [ 614.25 , 1.227194120611565e-45 ], [ 614.3125 , 1.2274072172446712e-45 ], [ 614.375 , 1.2276203508810306e-45 ], [ 614.4375 , 1.2278335215270514e-45 ], 
                    [ 614.5 , 1.2280467291891865e-45 ], [ 614.5625 , 1.2282599738738721e-45 ], [ 614.625 , 1.2284732176140828e-45 ], [ 614.6875 , 1.2286864176689798e-45 ], 
                    [ 614.75 , 1.2288996547244909e-45 ], [ 614.8125 , 1.2291129287870287e-45 ], [ 614.875 , 1.2293262398630331e-45 ], [ 614.9375 , 1.2295395879589191e-45 ], 
                    [ 615.0 , 1.2297529730811116e-45 ], [ 615.0625 , 1.2299663952360452e-45 ], [ 615.125 , 1.2301798544301294e-45 ], [ 615.1875 , 1.230393350669792e-45 ], 
                    [ 615.25 , 1.2306068839614712e-45 ], [ 615.3125 , 1.2308204543115973e-45 ], [ 615.375 , 1.2310340617266017e-45 ], [ 615.4375 , 1.2312477062129522e-45 ], 
                    [ 615.5 , 1.2314613877770122e-45 ], [ 615.5625 , 1.2316751064252516e-45 ], [ 615.625 , 1.2318888621641325e-45 ], [ 615.6875 , 1.2321026550000221e-45 ], 
                    [ 615.75 , 1.2323164849394458e-45 ], [ 615.8125 , 1.2325303519887733e-45 ], [ 615.875 , 1.2327442561544793e-45 ], [ 615.9375 , 1.2329581974429885e-45 ], 
                    [ 616.0 , 1.2331721758607606e-45 ], [ 616.0625 , 1.2333857278726367e-45 ], [ 616.125 , 1.2335990650220178e-45 ], [ 616.1875 , 1.2338124390720697e-45 ], 
                    [ 616.25 , 1.2340258500291489e-45 ], [ 616.3125 , 1.2342392978996392e-45 ], [ 616.375 , 1.234452782689943e-45 ], [ 616.4375 , 1.2346663044064465e-45 ], 
                    [ 616.5 , 1.2348798630555011e-45 ], [ 616.5625 , 1.2350934586435481e-45 ], [ 616.625 , 1.2353070911769505e-45 ], [ 616.6875 , 1.2355207606620896e-45 ], 
                    [ 616.75 , 1.2357344671053574e-45 ], [ 616.8125 , 1.235948210513172e-45 ], [ 616.875 , 1.2361619908918928e-45 ], [ 616.9375 , 1.2363758082479319e-45 ], 
                    [ 617.0 , 1.23658966258772e-45 ], [ 617.0625 , 1.2368035539175668e-45 ], [ 617.125 , 1.237017482243967e-45 ], [ 617.1875 , 1.2372314475732403e-45 ], 
                    [ 617.25 , 1.2374454499118403e-45 ], [ 617.3125 , 1.2376594892661681e-45 ], [ 617.375 , 1.2378735656425739e-45 ], [ 617.4375 , 1.2380876790475138e-45 ], 
                    [ 617.5 , 1.2383017650722101e-45 ], [ 617.5625 , 1.2385158818993535e-45 ], [ 617.625 , 1.2387300357498039e-45 ], [ 617.6875 , 1.2389442266299633e-45 ], 
                    [ 617.75 , 1.2391584545462168e-45 ], [ 617.8125 , 1.2393727195049863e-45 ], [ 617.875 , 1.2395870215126675e-45 ], [ 617.9375 , 1.2398013605756405e-45 ], 
                    [ 618.0 , 1.2400157367003743e-45 ], [ 618.0625 , 1.2402301498932245e-45 ], [ 618.125 , 1.240444600160618e-45 ], [ 618.1875 , 1.2406590875089655e-45 ], 
                    [ 618.25 , 1.2408736119446965e-45 ], [ 618.3125 , 1.2410881734741884e-45 ], [ 618.375 , 1.2413027721038728e-45 ], [ 618.4375 , 1.241517407840165e-45 ], 
                    [ 618.5 , 1.2417320806894808e-45 ], [ 618.5625 , 1.2419467906582554e-45 ], [ 618.625 , 1.2421615377528718e-45 ], [ 618.6875 , 1.242376321979758e-45 ], 
                    [ 618.75 , 1.2425911433453618e-45 ], [ 618.8125 , 1.2428060018560778e-45 ], [ 618.875 , 1.2430208352181843e-45 ], [ 618.9375 , 1.2432352812235992e-45 ], 
                    [ 619.0 , 1.2434497642252388e-45 ], [ 619.0625 , 1.2436642842294944e-45 ], [ 619.125 , 1.2438788412427589e-45 ], [ 619.1875 , 1.2440934352713994e-45 ], 
                    [ 619.25 , 1.244308066321828e-45 ], [ 619.3125 , 1.2445227344003965e-45 ], [ 619.375 , 1.2447374395135108e-45 ], [ 619.4375 , 1.244952181667569e-45 ], 
                    [ 619.5 , 1.2451669608689522e-45 ], [ 619.5625 , 1.2453817771240434e-45 ], [ 619.625 , 1.2455966304392524e-45 ], [ 619.6875 , 1.2458115208209642e-45 ], 
                    [ 619.75 , 1.2460264482755645e-45 ], [ 619.8125 , 1.2462414128094582e-45 ], [ 619.875 , 1.2464564144290509e-45 ], [ 619.9375 , 1.2466714531407318e-45 ], 
                    [ 620.0 , 1.2468865289508824e-45 ], [ 620.0625 , 1.2471016418659471e-45 ], [ 620.125 , 1.247316791892283e-45 ], [ 620.1875 , 1.2475319790363014e-45 ], 
                    [ 620.25 , 1.2477472033044412e-45 ], [ 620.3125 , 1.24796235029988e-45 ], [ 620.375 , 1.2481772622517965e-45 ], [ 620.4375 , 1.2483922112137704e-45 ], 
                    [ 620.5 , 1.2486071971921575e-45 ], [ 620.5625 , 1.248822220193341e-45 ], [ 620.625 , 1.249037280223697e-45 ], [ 620.6875 , 1.249252377289602e-45 ], 
                    [ 620.75 , 1.2494675113974516e-45 ], [ 620.8125 , 1.2496826825535897e-45 ], [ 620.875 , 1.2498978907644137e-45 ], [ 620.9375 , 1.250113136036314e-45 ], 
                    [ 621.0 , 1.250328418375655e-45 ], [ 621.0625 , 1.2505437377888379e-45 ], [ 621.125 , 1.2507590942822294e-45 ], [ 621.1875 , 1.2509744878622329e-45 ], 
                    [ 621.25 , 1.2511899185351996e-45 ], [ 621.3125 , 1.2514053863075793e-45 ], [ 621.375 , 1.2516208911857076e-45 ], [ 621.4375 , 1.2518364331760015e-45 ], 
                    [ 621.5 , 1.2520520122848428e-45 ], [ 621.5625 , 1.2522676285186237e-45 ], [ 621.625 , 1.2524832818837291e-45 ], [ 621.6875 , 1.2526989723865708e-45 ], 
                    [ 621.75 , 1.2529145852061589e-45 ], [ 621.8125 , 1.253130073909076e-45 ], [ 621.875 , 1.253345599673864e-45 ], [ 621.9375 , 1.2535611625069422e-45 ], 
                    [ 622.0 , 1.253776762414632e-45 ], [ 622.0625 , 1.2539923994033456e-45 ], [ 622.125 , 1.2542080734794695e-45 ], [ 622.1875 , 1.2544237846493378e-45 ], 
                    [ 622.25 , 1.2546395329193661e-45 ], [ 622.3125 , 1.2548553182959348e-45 ], [ 622.375 , 1.2550711407854087e-45 ], [ 622.4375 , 1.2552870003941791e-45 ], 
                    [ 622.5 , 1.2555028971286394e-45 ], [ 622.5625 , 1.2557188309951658e-45 ], [ 622.625 , 1.2559348020001268e-45 ], [ 622.6875 , 1.2561508101499457e-45 ], 
                    [ 622.75 , 1.2563668554509932e-45 ], [ 622.8125 , 1.2565829379096409e-45 ], [ 622.875 , 1.2567990575323152e-45 ], [ 622.9375 , 1.2570152143253902e-45 ], 
                    [ 623.0 , 1.257231408295241e-45 ], [ 623.0625 , 1.2574476394482973e-45 ], [ 623.125 , 1.2576639077909364e-45 ], [ 623.1875 , 1.2578799438838838e-45 ], 
                    [ 623.25 , 1.2580957156210953e-45 ], [ 623.3125 , 1.2583115243709261e-45 ], [ 623.375 , 1.2585273701397337e-45 ], [ 623.4375 , 1.2587432529338689e-45 ], 
                    [ 623.5 , 1.258959172759682e-45 ], [ 623.5625 , 1.2591751296235262e-45 ], [ 623.625 , 1.2593911235317632e-45 ], [ 623.6875 , 1.2596071544907295e-45 ], 
                    [ 623.75 , 1.2598232225067901e-45 ], [ 623.8125 , 1.2600393275863012e-45 ], [ 623.875 , 1.2602554697356206e-45 ], [ 623.9375 , 1.260471648961107e-45 ], 
                    [ 624.0 , 1.2606878652691384e-45 ], [ 624.0625 , 1.2609041186660402e-45 ], [ 624.125 , 1.2611204091581919e-45 ], [ 624.1875 , 1.2613367367519568e-45 ], 
                    [ 624.25 , 1.2615531014536996e-45 ], [ 624.3125 , 1.2617695032697849e-45 ], [ 624.375 , 1.2619859422065885e-45 ], [ 624.4375 , 1.2622024182704693e-45 ], 
                    [ 624.5 , 1.2624189314677955e-45 ], [ 624.5625 , 1.2626354818049189e-45 ], [ 624.625 , 1.2628520078629835e-45 ], [ 624.6875 , 1.2630685044971311e-45 ], 
                    [ 624.75 , 1.2632850382463017e-45 ], [ 624.8125 , 1.263501609116867e-45 ], [ 624.875 , 1.263718217115191e-45 ], [ 624.9375 , 1.2639348622476296e-45 ], 
                    [ 625.0 , 1.2641515445205757e-45 ], [ 625.0625 , 1.2643682639403699e-45 ], [ 625.125 , 1.264585020513389e-45 ], [ 625.1875 , 1.2648018142460028e-45 ], 
                    [ 625.25 , 1.2650186451445812e-45 ], [ 625.3125 , 1.2652355132154965e-45 ], [ 625.375 , 1.2654524184651295e-45 ], [ 625.4375 , 1.2656693608998274e-45 ], 
                    [ 625.5 , 1.265886340525983e-45 ], [ 625.5625 , 1.2661033573499899e-45 ], [ 625.625 , 1.2663204113782075e-45 ], [ 625.6875 , 1.2665375026169956e-45 ], 
                    [ 625.75 , 1.2667546310727697e-45 ], [ 625.8125 , 1.2669717967518736e-45 ], [ 625.875 , 1.2671889996607342e-45 ], [ 625.9375 , 1.2674062398056795e-45 ], 
                    [ 626.0 , 1.2676235171931382e-45 ], [ 626.0625 , 1.2678406505640768e-45 ], [ 626.125 , 1.2680575903873253e-45 ], [ 626.1875 , 1.2682745673310622e-45 ], 
                    [ 626.25 , 1.268491581401666e-45 ], [ 626.3125 , 1.2687086326054809e-45 ], [ 626.375 , 1.2689257209488601e-45 ], [ 626.4375 , 1.2691428464381501e-45 ], 
                    [ 626.5 , 1.2693600090797336e-45 ], [ 626.5625 , 1.2695772088799318e-45 ], [ 626.625 , 1.269794445845121e-45 ], [ 626.6875 , 1.2700117199816784e-45 ], 
                    [ 626.75 , 1.2702290312959283e-45 ], [ 626.8125 , 1.2704463797942592e-45 ], [ 626.875 , 1.270663765483016e-45 ], [ 626.9375 , 1.2708811883685798e-45 ], 
                    [ 627.0 , 1.2710986484573067e-45 ], [ 627.0625 , 1.2713161457555441e-45 ], [ 627.125 , 1.2715336802697045e-45 ], [ 627.1875 , 1.2717512520061287e-45 ], 
                    [ 627.25 , 1.2719688609711676e-45 ], [ 627.3125 , 1.2721865071712185e-45 ], [ 627.375 , 1.2724041906126436e-45 ], [ 627.4375 , 1.2726219113018153e-45 ], 
                    [ 627.5 , 1.2728394761500621e-45 ], [ 627.5625 , 1.273056696535028e-45 ], [ 627.625 , 1.2732739539904061e-45 ], [ 627.6875 , 1.273491248522532e-45 ], 
                    [ 627.75 , 1.2737085801377421e-45 ], [ 627.8125 , 1.2739259488423558e-45 ], [ 627.875 , 1.2741433546426937e-45 ], [ 627.9375 , 1.2743607975451137e-45 ], 
                    [ 628.0 , 1.2745782775559112e-45 ], [ 628.0625 , 1.2747957946814283e-45 ], [ 628.125 , 1.275013348928026e-45 ], [ 628.1875 , 1.2752309403020121e-45 ], 
                    [ 628.25 , 1.275448568809741e-45 ], [ 628.3125 , 1.2756662344575314e-45 ], [ 628.375 , 1.2758839372517219e-45 ], [ 628.4375 , 1.276101677198679e-45 ], 
                    [ 628.5 , 1.2763194543047337e-45 ], [ 628.5625 , 1.2765372685762006e-45 ], [ 628.625 , 1.2767551200194404e-45 ], [ 628.6875 , 1.2769730086408059e-45 ], 
                    [ 628.75 , 1.2771909344466326e-45 ], [ 628.8125 , 1.2774088974432481e-45 ], [ 628.875 , 1.277626897637036e-45 ], [ 628.9375 , 1.277844910710948e-45 ], 
                    [ 629.0 , 1.2780628530991115e-45 ], [ 629.0625 , 1.2782808326583622e-45 ], [ 629.125 , 1.2784988493950217e-45 ], [ 629.1875 , 1.2787169033154668e-45 ], 
                    [ 629.25 , 1.2789349944260124e-45 ], [ 629.3125 , 1.2791531227330196e-45 ], [ 629.375 , 1.2793712882428232e-45 ], [ 629.4375 , 1.2795894909617684e-45 ], 
                    [ 629.5 , 1.2798077308961922e-45 ], [ 629.5625 , 1.2800260080524604e-45 ], [ 629.625 , 1.2802443224369028e-45 ], [ 629.6875 , 1.2804626740558781e-45 ], 
                    [ 629.75 , 1.2806810629157457e-45 ], [ 629.8125 , 1.2808994890228486e-45 ], [ 629.875 , 1.2811179523835207e-45 ], [ 629.9375 , 1.2813364530041524e-45 ], 
                    [ 630.0 , 1.2815549908910716e-45 ], [ 630.0625 , 1.281773566050661e-45 ], [ 630.125 , 1.2819921784892324e-45 ], [ 630.1875 , 1.2822108282131987e-45 ], 
                    [ 630.25 , 1.2824295152288822e-45 ], [ 630.3125 , 1.2826482395426435e-45 ], [ 630.375 , 1.2828670011608622e-45 ], [ 630.4375 , 1.2830852130318485e-45 ], 
                    [ 630.5 , 1.2833034524727407e-45 ], [ 630.5625 , 1.2835217290339175e-45 ], [ 630.625 , 1.2837400427216476e-45 ], [ 630.6875 , 1.2839583935422728e-45 ], 
                    [ 630.75 , 1.2841767815020911e-45 ], [ 630.8125 , 1.2843952066074194e-45 ], [ 630.875 , 1.284613668864603e-45 ], [ 630.9375 , 1.2848321682799429e-45 ], 
                    [ 631.0 , 1.28505070485975e-45 ], [ 631.0625 , 1.2852692786103641e-45 ], [ 631.125 , 1.2854878895381076e-45 ], [ 631.1875 , 1.2857065376492582e-45 ], 
                    [ 631.25 , 1.2859252229502041e-45 ], [ 631.3125 , 1.2861439454472347e-45 ], [ 631.375 , 1.2863627051466765e-45 ], [ 631.4375 , 1.2865815020548939e-45 ], 
                    [ 631.5 , 1.2868003361781789e-45 ], [ 631.5625 , 1.2870192075228435e-45 ], [ 631.625 , 1.2872381160952824e-45 ], [ 631.6875 , 1.2874570619017729e-45 ], 
                    [ 631.75 , 1.2876760449486756e-45 ], [ 631.8125 , 1.2878950652423064e-45 ], [ 631.875 , 1.288113931185154e-45 ], [ 631.9375 , 1.2883327597970016e-45 ], 
                    [ 632.0 , 1.2885516255840836e-45 ], [ 632.0625 , 1.2887705285527435e-45 ], [ 632.125 , 1.2889894687092882e-45 ], [ 632.1875 , 1.2892084460600356e-45 ], 
                    [ 632.25 , 1.2894274606113043e-45 ], [ 632.3125 , 1.2896465123694048e-45 ], [ 632.375 , 1.2898656013406945e-45 ], [ 632.4375 , 1.2900847275314498e-45 ], 
                    [ 632.5 , 1.2903038909480115e-45 ], [ 632.5625 , 1.2905230915966948e-45 ], [ 632.625 , 1.290742329483843e-45 ], [ 632.6875 , 1.2909616046157824e-45 ], 
                    [ 632.75 , 1.291180916998831e-45 ], [ 632.8125 , 1.2914002666392989e-45 ], [ 632.875 , 1.291619653543552e-45 ], [ 632.9375 , 1.2918390777178842e-45 ], 
                    [ 633.0 , 1.2920585391686637e-45 ], [ 633.0625 , 1.2922780379021956e-45 ], [ 633.125 , 1.2924975739248136e-45 ], [ 633.1875 , 1.2927171472428802e-45 ], 
                    [ 633.25 , 1.2929367578626944e-45 ], [ 633.3125 , 1.2931563498397422e-45 ], [ 633.375 , 1.29337589791484e-45 ], [ 633.4375 , 1.2935954832641294e-45 ], 
                    [ 633.5 , 1.2938151058939292e-45 ], [ 633.5625 , 1.2940347658105779e-45 ], [ 633.625 , 1.2942544630204062e-45 ], [ 633.6875 , 1.2944741975297546e-45 ], 
                    [ 633.75 , 1.2946939693449283e-45 ], [ 633.8125 , 1.294913778472298e-45 ], [ 633.875 , 1.295133624918161e-45 ], [ 633.9375 , 1.295353508688881e-45 ], 
                    [ 634.0 , 1.2955734297907951e-45 ], [ 634.0625 , 1.2957933882302227e-45 ], [ 634.125 , 1.296013384013512e-45 ], [ 634.1875 , 1.2962334171469942e-45 ], 
                    [ 634.25 , 1.2964534876370376e-45 ], [ 634.3125 , 1.2966735954899666e-45 ], [ 634.375 , 1.2968937407121058e-45 ], [ 634.4375 , 1.2971139233098273e-45 ], 
                    [ 634.5 , 1.2973341432894676e-45 ], [ 634.5625 , 1.2975544006573733e-45 ], [ 634.625 , 1.2977746954198917e-45 ], [ 634.6875 , 1.297995027583381e-45 ], 
                    [ 634.75 , 1.2982153620171812e-45 ], [ 634.8125 , 1.2984350368511573e-45 ], [ 634.875 , 1.2986547488569388e-45 ], [ 634.9375 , 1.2988744980408433e-45 ], 
                    [ 635.0 , 1.2990942844091433e-45 ], [ 635.0625 , 1.2993141079681495e-45 ], [ 635.125 , 1.2995339687241455e-45 ], [ 635.1875 , 1.2997538666834259e-45 ], 
                    [ 635.25 , 1.2999738018522856e-45 ], [ 635.3125 , 1.3001937742370214e-45 ], [ 635.375 , 1.3004137838439209e-45 ], [ 635.4375 , 1.300633830679292e-45 ], 
                    [ 635.5 , 1.3008539147494531e-45 ], [ 635.5625 , 1.3010740360606675e-45 ], [ 635.625 , 1.3012941946192738e-45 ], [ 635.6875 , 1.3015143904315565e-45 ], 
                    [ 635.75 , 1.3017346235038007e-45 ], [ 635.8125 , 1.30195489384233e-45 ], [ 635.875 , 1.3021752014534593e-45 ], [ 635.9375 , 1.3023955463434863e-45 ], 
                    [ 636.0 , 1.3026159285187101e-45 ], [ 636.0625 , 1.3028363479854489e-45 ], [ 636.125 , 1.3030568047500224e-45 ], [ 636.1875 , 1.3032772988187324e-45 ], 
                    [ 636.25 , 1.3034976615436863e-45 ], [ 636.3125 , 1.30371795988618e-45 ], [ 636.375 , 1.3039382954603354e-45 ], [ 636.4375 , 1.3041586682724076e-45 ], 
                    [ 636.5 , 1.3043790783287274e-45 ], [ 636.5625 , 1.3045995256355611e-45 ], [ 636.625 , 1.3048200101992044e-45 ], [ 636.6875 , 1.305040532025991e-45 ], 
                    [ 636.75 , 1.3052610911221631e-45 ], [ 636.8125 , 1.3054816874940472e-45 ], [ 636.875 , 1.3057023211479615e-45 ], [ 636.9375 , 1.3059229920901788e-45 ], 
                    [ 637.0 , 1.3061437003269923e-45 ], [ 637.0625 , 1.3063644458647698e-45 ], [ 637.125 , 1.3065852287097409e-45 ], [ 637.1875 , 1.3068060488682202e-45 ], 
                    [ 637.25 , 1.3070269063465603e-45 ], [ 637.3125 , 1.3072478011510404e-45 ], [ 637.375 , 1.3074687332879505e-45 ], [ 637.4375 , 1.307689702763637e-45 ], 
                    [ 637.5 , 1.3079107095843826e-45 ], [ 637.5625 , 1.308131753756508e-45 ], [ 637.625 , 1.3083528352863258e-45 ], [ 637.6875 , 1.3085738825198141e-45 ], 
                    [ 637.75 , 1.3087945672427373e-45 ], [ 637.8125 , 1.3090152891830919e-45 ], [ 637.875 , 1.3092360483471456e-45 ], [ 637.9375 , 1.3094568447411572e-45 ], 
                    [ 638.0 , 1.3096776783714334e-45 ], [ 638.0625 , 1.3098985492442537e-45 ], [ 638.125 , 1.3101194573658804e-45 ], [ 638.1875 , 1.3103404027426046e-45 ], 
                    [ 638.25 , 1.3105613853807186e-45 ], [ 638.3125 , 1.3107824052864969e-45 ], [ 638.375 , 1.3110034624662154e-45 ], [ 638.4375 , 1.3112245569261786e-45 ], 
                    [ 638.5 , 1.3114456886726459e-45 ], [ 638.5625 , 1.3116668577119424e-45 ], [ 638.625 , 1.3118880640503297e-45 ], [ 638.6875 , 1.3121093076941073e-45 ], 
                    [ 638.75 , 1.3123305886495856e-45 ], [ 638.8125 , 1.3125519069230004e-45 ], [ 638.875 , 1.3127732625207016e-45 ], [ 638.9375 , 1.3129946554489743e-45 ], 
                    [ 639.0 , 1.3132160857140864e-45 ], [ 639.0625 , 1.3134375533223716e-45 ], [ 639.125 , 1.3136590582801182e-45 ], [ 639.1875 , 1.3138805430834471e-45 ], 
                    [ 639.25 , 1.3141020293729004e-45 ], [ 639.3125 , 1.3143235529992344e-45 ], [ 639.375 , 1.3145451139687148e-45 ], [ 639.4375 , 1.3147667122876652e-45 ], 
                    [ 639.5 , 1.3149883479623438e-45 ], [ 639.5625 , 1.3152100209990672e-45 ], [ 639.625 , 1.3154317314041423e-45 ], [ 639.6875 , 1.3156534791838593e-45 ], 
                    [ 639.75 , 1.3158752643445097e-45 ], [ 639.8125 , 1.3160970868924038e-45 ], [ 639.875 , 1.3163189468338444e-45 ], [ 639.9375 , 1.3165408441751442e-45 ], 
                    [ 640.0 , 1.3167627789225894e-45 ], [ 640.0625 , 1.3169847510824949e-45 ], [ 640.125 , 1.3172067606611769e-45 ], [ 640.1875 , 1.3174288076649338e-45 ], 
                    [ 640.25 , 1.3176508921000748e-45 ], [ 640.3125 , 1.3178730139729e-45 ], [ 640.375 , 1.3180951732897301e-45 ], [ 640.4375 , 1.3183173700568772e-45 ], 
                    [ 640.5 , 1.3185396042806637e-45 ], [ 640.5625 , 1.3187618759673851e-45 ], [ 640.625 , 1.3189841605880327e-45 ], [ 640.6875 , 1.3192057403572499e-45 ], 
                    [ 640.75 , 1.3194273573502513e-45 ], [ 640.8125 , 1.3196490115732996e-45 ], [ 640.875 , 1.31987070303264e-45 ], [ 640.9375 , 1.3200924317345462e-45 ], 
                    [ 641.0 , 1.3203141976852752e-45 ], [ 641.0625 , 1.3205360008910653e-45 ], [ 641.125 , 1.3207578413581848e-45 ], [ 641.1875 , 1.320979719092884e-45 ], 
                    [ 641.25 , 1.3212016341014608e-45 ], [ 641.3125 , 1.3214235863901209e-45 ], [ 641.375 , 1.321645575965155e-45 ], [ 641.4375 , 1.3218676028328555e-45 ], 
                    [ 641.5 , 1.3220896669994305e-45 ], [ 641.5625 , 1.3223117684711742e-45 ], [ 641.625 , 1.3225339072543634e-45 ], [ 641.6875 , 1.3227560833552751e-45 ], 
                    [ 641.75 , 1.3229782967801316e-45 ], [ 641.8125 , 1.3232005475352313e-45 ], [ 641.875 , 1.3234228356268547e-45 ], [ 641.9375 , 1.323645161061274e-45 ], 
                    [ 642.0 , 1.323867523844744e-45 ], [ 642.0625 , 1.324089923983548e-45 ], [ 642.125 , 1.3243123756594384e-45 ], [ 642.1875 , 1.324534885974756e-45 ], 
                    [ 642.25 , 1.324757433676157e-45 ], [ 642.3125 , 1.3249800187699038e-45 ], [ 642.375 , 1.3252026412622984e-45 ], [ 642.4375 , 1.3254253011596149e-45 ], 
                    [ 642.5 , 1.3256479984681286e-45 ], [ 642.5625 , 1.3258707331941443e-45 ], [ 642.625 , 1.3260935053439298e-45 ], [ 642.6875 , 1.3263163149237921e-45 ], 
                    [ 642.75 , 1.3265391619400104e-45 ], [ 642.8125 , 1.3267620463988657e-45 ], [ 642.875 , 1.326984968306649e-45 ], [ 642.9375 , 1.3272079276696901e-45 ], 
                    [ 643.0 , 1.3274309244942446e-45 ], [ 643.0625 , 1.3276539587866066e-45 ], [ 643.125 , 1.3278770305530902e-45 ], [ 643.1875 , 1.3281001397999637e-45 ], 
                    [ 643.25 , 1.3283232865335717e-45 ], [ 643.3125 , 1.328546470760175e-45 ], [ 643.375 , 1.328769692486073e-45 ], [ 643.4375 , 1.3289929517175758e-45 ], 
                    [ 643.5 , 1.3292162484609946e-45 ], [ 643.5625 , 1.3294395827226319e-45 ], [ 643.625 , 1.3296624080331389e-45 ], [ 643.6875 , 1.3298850581372039e-45 ], 
                    [ 643.75 , 1.330107745523707e-45 ], [ 643.8125 , 1.3303304701989099e-45 ], [ 643.875 , 1.3305532321690564e-45 ], [ 643.9375 , 1.3307760314403635e-45 ], 
                    [ 644.0 , 1.3309988680191052e-45 ], [ 644.0625 , 1.3312217419115192e-45 ], [ 644.125 , 1.3314446531238536e-45 ], [ 644.1875 , 1.3316676016623484e-45 ], 
                    [ 644.25 , 1.3318905875332726e-45 ], [ 644.3125 , 1.3321136107428677e-45 ], [ 644.375 , 1.3323366712973957e-45 ], [ 644.4375 , 1.3325597692030722e-45 ], 
                    [ 644.5 , 1.3327829044661988e-45 ], [ 644.5625 , 1.333006077093012e-45 ], [ 644.625 , 1.333229287089759e-45 ], [ 644.6875 , 1.3334525344627255e-45 ], 
                    [ 644.75 , 1.3336758192181513e-45 ], [ 644.8125 , 1.3338991413622866e-45 ], [ 644.875 , 1.334122500901392e-45 ], [ 644.9375 , 1.3343458978417483e-45 ], 
                    [ 645.0 , 1.3345693321896085e-45 ], [ 645.0625 , 1.3347928039512274e-45 ], [ 645.125 , 1.3350160172144923e-45 ], [ 645.1875 , 1.3352392554730099e-45 ], 
                    [ 645.25 , 1.3354625310609075e-45 ], [ 645.3125 , 1.3356858439844364e-45 ], [ 645.375 , 1.3359091942498117e-45 ], [ 645.4375 , 1.3361325818633152e-45 ], 
                    [ 645.5 , 1.336356006831164e-45 ], [ 645.5625 , 1.3365794691596138e-45 ], [ 645.625 , 1.3368029688549214e-45 ], [ 645.6875 , 1.3370265059233353e-45 ], 
                    [ 645.75 , 1.3372500803710669e-45 ], [ 645.8125 , 1.3374736922044141e-45 ], [ 645.875 , 1.3376973414296e-45 ], [ 645.9375 , 1.3379210280528962e-45 ], 
                    [ 646.0 , 1.3381447520805468e-45 ], [ 646.0625 , 1.3383685135187969e-45 ], [ 646.125 , 1.3385923123739025e-45 ], [ 646.1875 , 1.3388161486521483e-45 ], 
                    [ 646.25 , 1.3390400223597738e-45 ], [ 646.3125 , 1.3392639335030279e-45 ], [ 646.375 , 1.3394878820881704e-45 ], [ 646.4375 , 1.3397118681215008e-45 ], 
                    [ 646.5 , 1.3399358916092235e-45 ], [ 646.5625 , 1.3401599323965762e-45 ], [ 646.625 , 1.3403839107820999e-45 ], [ 646.6875 , 1.3406079266006965e-45 ], 
                    [ 646.75 , 1.3408319798586414e-45 ], [ 646.8125 , 1.3410560705621722e-45 ], [ 646.875 , 1.3412801987175664e-45 ], [ 646.9375 , 1.3415043643310645e-45 ], 
                    [ 647.0 , 1.3417285674089452e-45 ], [ 647.0625 , 1.3419528079574517e-45 ], [ 647.125 , 1.3421770859828644e-45 ], [ 647.1875 , 1.3424014014914285e-45 ], 
                    [ 647.25 , 1.3426257544894273e-45 ], [ 647.3125 , 1.3428501449831166e-45 ], [ 647.375 , 1.3430745729787442e-45 ], [ 647.4375 , 1.3432990384826156e-45 ], 
                    [ 647.5 , 1.3435235415009808e-45 ], [ 647.5625 , 1.3437480820401188e-45 ], [ 647.625 , 1.3439726601062621e-45 ], [ 647.6875 , 1.3441972757057305e-45 ], 
                    [ 647.75 , 1.3444219288447584e-45 ], [ 647.8125 , 1.3446466195296583e-45 ], [ 647.875 , 1.3448713477666953e-45 ], [ 647.9375 , 1.3450961135621167e-45 ], 
                    [ 648.0 , 1.3453209169222192e-45 ], [ 648.0625 , 1.3455455726823943e-45 ], [ 648.125 , 1.3457698575643134e-45 ], [ 648.1875 , 1.3459941798315754e-45 ], 
                    [ 648.25 , 1.3462185394904502e-45 ], [ 648.3125 , 1.3464429365471321e-45 ], [ 648.375 , 1.3466673710078837e-45 ], [ 648.4375 , 1.34689184287893e-45 ], 
                    [ 648.5 , 1.3471163521665165e-45 ], [ 648.5625 , 1.3473408988768512e-45 ], [ 648.625 , 1.3475654830161913e-45 ], [ 648.6875 , 1.3477901045907853e-45 ], 
                    [ 648.75 , 1.348014763606873e-45 ], [ 648.8125 , 1.3482394600706762e-45 ], [ 648.875 , 1.3484641939884469e-45 ], [ 648.9375 , 1.3486889653664279e-45 ], 
                    [ 649.0 , 1.3489137742108444e-45 ], [ 649.0625 , 1.3491386205279794e-45 ], [ 649.125 , 1.3493635043240604e-45 ], [ 649.1875 , 1.349588425605325e-45 ], 
                    [ 649.25 , 1.3498133843780502e-45 ], [ 649.3125 , 1.350038380648447e-45 ], [ 649.375 , 1.350263414422804e-45 ], [ 649.4375 , 1.350488485707344e-45 ], 
                    [ 649.5 , 1.3507135945083193e-45 ], [ 649.5625 , 1.3509385441893146e-45 ], [ 649.625 , 1.351163245961553e-45 ], [ 649.6875 , 1.3513879851084549e-45 ], 
                    [ 649.75 , 1.3516127616362754e-45 ], [ 649.8125 , 1.3518375755511839e-45 ], [ 649.875 , 1.3520624268594178e-45 ], [ 649.9375 , 1.352287315567207e-45 ], 
                    [ 650.0 , 1.3525122416807527e-45 ], [ 650.0625 , 1.3527372052062862e-45 ], [ 650.125 , 1.3529622061500398e-45 ], [ 650.1875 , 1.3531872445182181e-45 ], 
                    [ 650.25 , 1.3534123203170558e-45 ], [ 650.3125 , 1.3536374335527878e-45 ], [ 650.375 , 1.353862584231622e-45 ], [ 650.4375 , 1.354087772359796e-45 ], 
                    [ 650.5 , 1.3543129979435485e-45 ], [ 650.5625 , 1.3545382609891093e-45 ], [ 650.625 , 1.3547635615026614e-45 ], [ 650.6875 , 1.354988899490504e-45 ], 
                    [ 650.75 , 1.355214274958832e-45 ], [ 650.8125 , 1.3554396879138987e-45 ], [ 650.875 , 1.3556651383619294e-45 ], [ 650.9375 , 1.3558906263091415e-45 ], 
                    [ 651.0 , 1.3561161517618008e-45 ], [ 651.0625 , 1.3563414660871311e-45 ], [ 651.125 , 1.3565665013455368e-45 ], [ 651.1875 , 1.3567915739403225e-45 ], 
                    [ 651.25 , 1.3570166838776543e-45 ], [ 651.3125 , 1.3572418311637756e-45 ], [ 651.375 , 1.3574670158048447e-45 ], [ 651.4375 , 1.3576922378070594e-45 ], 
                    [ 651.5 , 1.3579174971766182e-45 ], [ 651.5625 , 1.3581427939197499e-45 ], [ 651.625 , 1.3583681280426262e-45 ], [ 651.6875 , 1.358593499551478e-45 ], 
                    [ 651.75 , 1.3588189084524695e-45 ], [ 651.8125 , 1.3590443547518239e-45 ], [ 651.875 , 1.3592698384557363e-45 ], [ 651.9375 , 1.3594953595704513e-45 ], 
                    [ 652.0 , 1.3597209181021275e-45 ], [ 652.0625 , 1.3599465140569924e-45 ], [ 652.125 , 1.3601721474412645e-45 ], [ 652.1875 , 1.3603978182611346e-45 ], 
                    [ 652.25 , 1.3606235265228234e-45 ], [ 652.3125 , 1.3608492722325429e-45 ], [ 652.375 , 1.3610750553965161e-45 ], [ 652.4375 , 1.3613008760209473e-45 ], 
                    [ 652.5 , 1.3615267341120421e-45 ], [ 652.5625 , 1.3617526279213141e-45 ], [ 652.625 , 1.361978556662618e-45 ], [ 652.6875 , 1.3622045228878387e-45 ], 
                    [ 652.75 , 1.3624305266031571e-45 ], [ 652.8125 , 1.3626565678148408e-45 ], [ 652.875 , 1.3628826465290532e-45 ], [ 652.9375 , 1.3631087627520454e-45 ], 
                    [ 653.0 , 1.3633349164900598e-45 ], [ 653.0625 , 1.3635611077492817e-45 ], [ 653.125 , 1.3637873365359653e-45 ], [ 653.1875 , 1.364013602856337e-45 ], 
                    [ 653.25 , 1.3642399067166043e-45 ], [ 653.3125 , 1.3644662481230056e-45 ], [ 653.375 , 1.3646926270817698e-45 ], [ 653.4375 , 1.3649190435991272e-45 ], 
                    [ 653.5 , 1.3651454976812999e-45 ], [ 653.5625 , 1.3653719893345491e-45 ], [ 653.625 , 1.3655985185650694e-45 ], [ 653.6875 , 1.3658250853791146e-45 ], 
                    [ 653.75 , 1.3660516897829395e-45 ], [ 653.8125 , 1.3662783317827617e-45 ], [ 653.875 , 1.3665050113847893e-45 ], [ 653.9375 , 1.3667317285953095e-45 ], 
                    [ 654.0 , 1.3669584834205522e-45 ], [ 654.0625 , 1.367185073431874e-45 ], [ 654.125 , 1.3674113149007824e-45 ], [ 654.1875 , 1.367637593808086e-45 ], 
                    [ 654.25 , 1.3678639101599709e-45 ], [ 654.3125 , 1.3680902639626234e-45 ], [ 654.375 , 1.3683166552222604e-45 ], [ 654.4375 , 1.3685430839450705e-45 ], 
                    [ 654.5 , 1.3687695501372435e-45 ], [ 654.5625 , 1.3689960538049991e-45 ], [ 654.625 , 1.369222594954539e-45 ], [ 654.6875 , 1.3694491735920266e-45 ], 
                    [ 654.75 , 1.3696757897237238e-45 ], [ 654.8125 , 1.3699024433557865e-45 ], [ 654.875 , 1.3701291344944492e-45 ], [ 654.9375 , 1.370355863145909e-45 ], 
                    [ 655.0 , 1.3705826293163442e-45 ], [ 655.0625 , 1.370809433012032e-45 ], [ 655.125 , 1.3710362742391425e-45 ], [ 655.1875 , 1.3712631530038575e-45 ], 
                    [ 655.25 , 1.3714900693124471e-45 ], [ 655.3125 , 1.371717023171085e-45 ], [ 655.375 , 1.3719440145859851e-45 ], [ 655.4375 , 1.3721710435633816e-45 ], 
                    [ 655.5 , 1.3723981101094808e-45 ], [ 655.5625 , 1.3726251767951607e-45 ], [ 655.625 , 1.3728521312834897e-45 ], [ 655.6875 , 1.3730791232972316e-45 ], 
                    [ 655.75 , 1.3733061528426201e-45 ], [ 655.8125 , 1.3735332199258219e-45 ], [ 655.875 , 1.3737603245530924e-45 ], [ 655.9375 , 1.3739874667305707e-45 ], 
                    [ 656.0 , 1.3742146464645245e-45 ], [ 656.0625 , 1.3744418637611337e-45 ], [ 656.125 , 1.3746691186266387e-45 ], [ 656.1875 , 1.3748964110672024e-45 ], 
                    [ 656.25 , 1.375123741089106e-45 ], [ 656.3125 , 1.3753511086984945e-45 ], [ 656.375 , 1.3755785139016127e-45 ], [ 656.4375 , 1.3758059567046761e-45 ], 
                    [ 656.5 , 1.376033437113921e-45 ], [ 656.5625 , 1.3762609551355364e-45 ], [ 656.625 , 1.3764885107757508e-45 ], [ 656.6875 , 1.3767161040407843e-45 ], 
                    [ 656.75 , 1.3769437349368676e-45 ], [ 656.8125 , 1.3771714034701935e-45 ], [ 656.875 , 1.3773991096470046e-45 ], [ 656.9375 , 1.3776268534735346e-45 ], 
                    [ 657.0 , 1.377854634956009e-45 ], [ 657.0625 , 1.3780824403717444e-45 ], [ 657.125 , 1.3783097250510591e-45 ], [ 657.1875 , 1.3785370472160405e-45 ], 
                    [ 657.25 , 1.378764406872871e-45 ], [ 657.3125 , 1.3789918040276947e-45 ], [ 657.375 , 1.3792192386867453e-45 ], [ 657.4375 , 1.3794467108561987e-45 ], 
                    [ 657.5 , 1.3796742205422112e-45 ], [ 657.5625 , 1.3799017677509908e-45 ], [ 657.625 , 1.3801293524887553e-45 ], [ 657.6875 , 1.3803569747616253e-45 ], 
                    [ 657.75 , 1.3805846345758407e-45 ], [ 657.8125 , 1.3808123319375833e-45 ], [ 657.875 , 1.3810400668530555e-45 ], [ 657.9375 , 1.381267839328451e-45 ], 
                    [ 658.0 , 1.3814956493699449e-45 ], [ 658.0625 , 1.3817234969837426e-45 ], [ 658.125 , 1.3819513821760312e-45 ], [ 658.1875 , 1.382179304953018e-45 ], 
                    [ 658.25 , 1.3824072653209119e-45 ], [ 658.3125 , 1.3826352632859028e-45 ], [ 658.375 , 1.3828632988542013e-45 ], [ 658.4375 , 1.3830913720319797e-45 ], 
                    [ 658.5 , 1.3833194828254704e-45 ], [ 658.5625 , 1.3835476312408575e-45 ], [ 658.625 , 1.3837754708254587e-45 ], [ 658.6875 , 1.3840032524614766e-45 ], 
                    [ 658.75 , 1.3842310715923785e-45 ], [ 658.8125 , 1.3844589282243072e-45 ], [ 658.875 , 1.384686822363455e-45 ], [ 658.9375 , 1.3849147540159766e-45 ], 
                    [ 659.0 , 1.3851427231880663e-45 ], [ 659.0625 , 1.3853707298858908e-45 ], [ 659.125 , 1.385598774115617e-45 ], [ 659.1875 , 1.3858268558834328e-45 ], 
                    [ 659.25 , 1.3860549751955272e-45 ], [ 659.3125 , 1.3862831320580508e-45 ], [ 659.375 , 1.3865113264772243e-45 ], [ 659.4375 , 1.3867395584592099e-45 ], 
                    [ 659.5 , 1.386967828010191e-45 ], [ 659.5625 , 1.3871961351363614e-45 ], [ 659.625 , 1.3874244798438668e-45 ], [ 659.6875 , 1.3876528621389333e-45 ], 
                    [ 659.75 , 1.3878812820277281e-45 ], [ 659.8125 , 1.388109739516479e-45 ], [ 659.875 , 1.3883382346113259e-45 ], [ 659.9375 , 1.3885667673184888e-45 ], 
                    [ 660.0 , 1.388795337644139e-45 ], [ 660.0625 , 1.389023945594479e-45 ], [ 660.125 , 1.3892524274280304e-45 ], [ 660.1875 , 1.3894807694222742e-45 ], 
                    [ 660.25 , 1.3897091489475518e-45 ], [ 660.3125 , 1.3899375660100222e-45 ], [ 660.375 , 1.3901660206158452e-45 ], [ 660.4375 , 1.3903945127712115e-45 ], 
                    [ 660.5 , 1.3906230424822632e-45 ], [ 660.5625 , 1.3908516097552221e-45 ], [ 660.625 , 1.391080214596213e-45 ], [ 660.6875 , 1.3913088570114403e-45 ], 
                    [ 660.75 , 1.39153753700707e-45 ], [ 660.8125 , 1.3917662545892587e-45 ], [ 660.875 , 1.391995009764234e-45 ], [ 660.9375 , 1.3922238025381256e-45 ], 
                    [ 661.0 , 1.3924526329171427e-45 ], [ 661.0625 , 1.3926815009074563e-45 ], [ 661.125 , 1.3929104065152385e-45 ], [ 661.1875 , 1.3931393497466922e-45 ], 
                    [ 661.25 , 1.3933683306079714e-45 ], [ 661.3125 , 1.3935973491053007e-45 ], [ 661.375 , 1.3938264052448464e-45 ], [ 661.4375 , 1.3940554990327954e-45 ], 
                    [ 661.5 , 1.3942846304753158e-45 ], [ 661.5625 , 1.3945137995786364e-45 ], [ 661.625 , 1.3947429494193176e-45 ], [ 661.6875 , 1.3949717490178846e-45 ], 
                    [ 661.75 , 1.3952005861497263e-45 ], [ 661.8125 , 1.3954294608209798e-45 ], [ 661.875 , 1.3956583730378035e-45 ], [ 661.9375 , 1.3958873228063858e-45 ], 
                    [ 662.0 , 1.3961163101328577e-45 ], [ 662.0625 , 1.39634533502338e-45 ], [ 662.125 , 1.3965743974841447e-45 ], [ 662.1875 , 1.3968034975212753e-45 ], 
                    [ 662.25 , 1.3970326351409657e-45 ], [ 662.3125 , 1.3972618103493613e-45 ], [ 662.375 , 1.3974910231526581e-45 ], [ 662.4375 , 1.3977202735569936e-45 ], 
                    [ 662.5 , 1.3979495615685459e-45 ], [ 662.5625 , 1.3981788871934843e-45 ], [ 662.625 , 1.3984082504379888e-45 ], [ 662.6875 , 1.398637651308201e-45 ], 
                    [ 662.75 , 1.3988670898103132e-45 ], [ 662.8125 , 1.3990965659504984e-45 ], [ 662.875 , 1.3993260797349511e-45 ], [ 662.9375 , 1.3995556311698068e-45 ], 
                    [ 663.0 , 1.3997852202612813e-45 ], [ 663.0625 , 1.4000148470155126e-45 ], [ 663.125 , 1.4002445114387184e-45 ], [ 663.1875 , 1.4004740026523439e-45 ], 
                    [ 663.25 , 1.4007034461624758e-45 ], [ 663.3125 , 1.4009329272629695e-45 ], [ 663.375 , 1.4011624459599835e-45 ], [ 663.4375 , 1.401392002259677e-45 ], 
                    [ 663.5 , 1.4016215961682112e-45 ], [ 663.5625 , 1.401851227691767e-45 ], [ 663.625 , 1.4020808968364677e-45 ], [ 663.6875 , 1.4023106036084962e-45 ], 
                    [ 663.75 , 1.4025403480140073e-45 ], [ 663.8125 , 1.4027701300591967e-45 ], [ 663.875 , 1.402999949750201e-45 ], [ 663.9375 , 1.4032298070931975e-45 ], 
                    [ 664.0 , 1.4034597020943653e-45 ], [ 664.0625 , 1.403689634759854e-45 ], [ 664.125 , 1.4039196050958337e-45 ], [ 664.1875 , 1.404149613108517e-45 ], 
                    [ 664.25 , 1.4043796588040253e-45 ], [ 664.3125 , 1.4046097421885434e-45 ], [ 664.375 , 1.4048398632682858e-45 ], [ 664.4375 , 1.4050700220493778e-45 ], 
                    [ 664.5 , 1.4053002185380063e-45 ], [ 664.5625 , 1.405530452740379e-45 ], [ 664.625 , 1.4057607246626648e-45 ], [ 664.6875 , 1.405990911371003e-45 ], 
                    [ 664.75 , 1.406220762924577e-45 ], [ 664.8125 , 1.4064506520542795e-45 ], [ 664.875 , 1.4066805787663124e-45 ], [ 664.9375 , 1.4069105430667708e-45 ], 
                    [ 665.0 , 1.4071405449618388e-45 ], [ 665.0625 , 1.4073705844576328e-45 ], [ 665.125 , 1.4076006615602995e-45 ], [ 665.1875 , 1.407830776276007e-45 ], 
                    [ 665.25 , 1.4080609286108945e-45 ], [ 665.3125 , 1.408291118571092e-45 ], [ 665.375 , 1.4085213461628e-45 ], [ 665.4375 , 1.408751611392141e-45 ], 
                    [ 665.5 , 1.408981914265237e-45 ], [ 665.5625 , 1.4092122547882936e-45 ], [ 665.625 , 1.4094426329674348e-45 ], [ 665.6875 , 1.409673048808847e-45 ], 
                    [ 665.75 , 1.4099035023186468e-45 ], [ 665.8125 , 1.4101339935030327e-45 ], [ 665.875 , 1.4103645223681435e-45 ], [ 665.9375 , 1.4105950889201193e-45 ], 
                    [ 666.0 , 1.410825693165141e-45 ], [ 666.0625 , 1.4110563351093912e-45 ], [ 666.125 , 1.411287014759002e-45 ], [ 666.1875 , 1.4115177321201384e-45 ], 
                    [ 666.25 , 1.411748177786869e-45 ], [ 666.3125 , 1.4119785407298315e-45 ], [ 666.375 , 1.4122089412623995e-45 ], [ 666.4375 , 1.4124393793907473e-45 ], 
                    [ 666.5 , 1.4126698551209686e-45 ], [ 666.5625 , 1.41290036845923e-45 ], [ 666.625 , 1.4131309194116378e-45 ], [ 666.6875 , 1.4133615079843798e-45 ], 
                    [ 666.75 , 1.4135921341835547e-45 ], [ 666.8125 , 1.4138227980153124e-45 ], [ 666.875 , 1.4140534994857832e-45 ], [ 666.9375 , 1.414284238601129e-45 ], 
                    [ 667.0 , 1.414515015367493e-45 ], [ 667.0625 , 1.4147458297909985e-45 ], [ 667.125 , 1.41497668187777e-45 ], [ 667.1875 , 1.4152075716340334e-45 ], 
                    [ 667.25 , 1.415438499065876e-45 ], [ 667.3125 , 1.4156694641794347e-45 ], [ 667.375 , 1.415900466980879e-45 ], [ 667.4375 , 1.4161315074763677e-45 ], 
                    [ 667.5 , 1.4163625856720325e-45 ], [ 667.5625 , 1.4165937015740244e-45 ], [ 667.625 , 1.416824855188557e-45 ], [ 667.6875 , 1.4170560465216936e-45 ], 
                    [ 667.75 , 1.4172872295570612e-45 ], [ 667.8125 , 1.4175181766851617e-45 ], [ 667.875 , 1.417749161446155e-45 ], [ 667.9375 , 1.4179801838461237e-45 ], 
                    [ 668.0 , 1.418211243891221e-45 ], [ 668.0625 , 1.4184423415875807e-45 ], [ 668.125 , 1.4186734769413284e-45 ], [ 668.1875 , 1.4189046499586103e-45 ], 
                    [ 668.25 , 1.4191358606455842e-45 ], [ 668.3125 , 1.4193671090083475e-45 ], [ 668.375 , 1.4195983950530698e-45 ], [ 668.4375 , 1.4198297187858818e-45 ], 
                    [ 668.5 , 1.4200610802129245e-45 ], [ 668.5625 , 1.42029247934034e-45 ], [ 668.625 , 1.420523916174272e-45 ], [ 668.6875 , 1.420755390720864e-45 ], 
                    [ 668.75 , 1.4209869029862624e-45 ], [ 668.8125 , 1.4212184529766225e-45 ], [ 668.875 , 1.421450040698062e-45 ], [ 668.9375 , 1.421681666156759e-45 ], 
                    [ 669.0 , 1.4219133293588535e-45 ], [ 669.0625 , 1.4221450303104948e-45 ], [ 669.125 , 1.4223767690178245e-45 ], [ 669.1875 , 1.4226085454870255e-45 ], 
                    [ 669.25 , 1.42284035972423e-45 ], [ 669.3125 , 1.4230719696116537e-45 ], [ 669.375 , 1.4233033788811302e-45 ], [ 669.4375 , 1.4235348257806434e-45 ], 
                    [ 669.5 , 1.423766310316322e-45 ], [ 669.5625 , 1.4239978324942663e-45 ], [ 669.625 , 1.4242293923205968e-45 ], [ 669.6875 , 1.4244609898014565e-45 ], 
                    [ 669.75 , 1.4246926249429483e-45 ], [ 669.8125 , 1.424924297751226e-45 ], [ 669.875 , 1.4251560082323945e-45 ], [ 669.9375 , 1.4253877563925604e-45 ], 
                    [ 670.0 , 1.4256195422378804e-45 ], [ 670.0625 , 1.4258513657744725e-45 ], [ 670.125 , 1.4260832270084666e-45 ], [ 670.1875 , 1.4263151259459816e-45 ], 
                    [ 670.25 , 1.4265470625931797e-45 ], [ 670.3125 , 1.4267790369561517e-45 ], [ 670.375 , 1.427011049041062e-45 ], [ 670.4375 , 1.427243098854044e-45 ], 
                    [ 670.5 , 1.4274751864012122e-45 ], [ 670.5625 , 1.4277073116887033e-45 ], [ 670.625 , 1.4279394747226643e-45 ], [ 670.6875 , 1.4281716755092435e-45 ], 
                    [ 670.75 , 1.4284039140545693e-45 ], [ 670.8125 , 1.4286361903647722e-45 ], [ 670.875 , 1.428868134685726e-45 ], [ 670.9375 , 1.429100033272852e-45 ], 
                    [ 671.0 , 1.4293319694960293e-45 ], [ 671.0625 , 1.429563943361366e-45 ], [ 671.125 , 1.429795954874962e-45 ], [ 671.1875 , 1.4300280040429366e-45 ], 
                    [ 671.25 , 1.4302600908714114e-45 ], [ 671.3125 , 1.4304922153664883e-45 ], [ 671.375 , 1.4307243775342808e-45 ], [ 671.4375 , 1.430956577380882e-45 ], 
                    [ 671.5 , 1.4311888149124485e-45 ], [ 671.5625 , 1.431421090135055e-45 ], [ 671.625 , 1.4316534030548596e-45 ], [ 671.6875 , 1.4318857536779598e-45 ], 
                    [ 671.75 , 1.432118142010475e-45 ], [ 671.8125 , 1.432350568058525e-45 ], [ 671.875 , 1.432583031828231e-45 ], [ 671.9375 , 1.4328155333256944e-45 ], 
                    [ 672.0 , 1.4330480725570795e-45 ], [ 672.0625 , 1.4332806495284894e-45 ], [ 672.125 , 1.433513264246029e-45 ], [ 672.1875 , 1.4337459167158552e-45 ], 
                    [ 672.25 , 1.4339786069440945e-45 ], [ 672.3125 , 1.4342113349368547e-45 ], [ 672.375 , 1.4344440716180483e-45 ], [ 672.4375 , 1.4346764206699482e-45 ], 
                    [ 672.5 , 1.4349088073573912e-45 ], [ 672.5625 , 1.435141231686484e-45 ], [ 672.625 , 1.435373693663313e-45 ], [ 672.6875 , 1.4356061932939566e-45 ], 
                    [ 672.75 , 1.4358387305845543e-45 ], [ 672.8125 , 1.4360713055411858e-45 ], [ 672.875 , 1.4363039181699318e-45 ], [ 672.9375 , 1.4365365684769456e-45 ], 
                    [ 673.0 , 1.4367692564682794e-45 ], [ 673.0625 , 1.4370019821500467e-45 ], [ 673.125 , 1.4372347455283937e-45 ], [ 673.1875 , 1.4374675466093657e-45 ], 
                    [ 673.25 , 1.4377003853991098e-45 ], [ 673.3125 , 1.4379332619037348e-45 ], [ 673.375 , 1.438166176129349e-45 ], [ 673.4375 , 1.4383991280820116e-45 ], 
                    [ 673.5 , 1.438632117767905e-45 ], [ 673.5625 , 1.4388651451931005e-45 ], [ 673.625 , 1.439098210363731e-45 ], [ 673.6875 , 1.439331313285891e-45 ], 
                    [ 673.75 , 1.439564453965725e-45 ], [ 673.8125 , 1.4397976324092984e-45 ], [ 673.875 , 1.4400308486227687e-45 ], [ 673.9375 , 1.44026405912326e-45 ], 
                    [ 674.0 , 1.4404971755006114e-45 ], [ 674.0625 , 1.440730329609399e-45 ], [ 674.125 , 1.4409635214557597e-45 ], [ 674.1875 , 1.441196751045761e-45 ], 
                    [ 674.25 , 1.4414300183855428e-45 ], [ 674.3125 , 1.4416633234811943e-45 ], [ 674.375 , 1.4418966663388472e-45 ], [ 674.4375 , 1.4421300469645934e-45 ], 
                    [ 674.5 , 1.4423634653645554e-45 ], [ 674.5625 , 1.4425969215448582e-45 ], [ 674.625 , 1.4428304155116063e-45 ], [ 674.6875 , 1.4430639472709154e-45 ], 
                    [ 674.75 , 1.4432975168288927e-45 ], [ 674.8125 , 1.4435311241916764e-45 ], [ 674.875 , 1.443764769365375e-45 ], [ 674.9375 , 1.443998452356099e-45 ], 
                    [ 675.0 , 1.444232173169999e-45 ], [ 675.0625 , 1.4444659318131464e-45 ], [ 675.125 , 1.4446997282917152e-45 ], [ 675.1875 , 1.4449335626117987e-45 ], 
                    [ 675.25 , 1.4451674347795525e-45 ], [ 675.3125 , 1.4454013448010607e-45 ], [ 675.375 , 1.4456352926824822e-45 ], [ 675.4375 , 1.4458692784299342e-45 ], 
                    [ 675.5 , 1.4461030577668453e-45 ], [ 675.5625 , 1.4463364829696253e-45 ], [ 675.625 , 1.4465699458511342e-45 ], [ 675.6875 , 1.4468034464174544e-45 ], 
                    [ 675.75 , 1.4470369846746577e-45 ], [ 675.8125 , 1.44727056062885e-45 ], [ 675.875 , 1.4475041742861046e-45 ], [ 675.9375 , 1.4477378256525085e-45 ], 
                    [ 676.0 , 1.4479715147341474e-45 ], [ 676.0625 , 1.4482052415371102e-45 ], [ 676.125 , 1.448439006067475e-45 ], [ 676.1875 , 1.448672808331352e-45 ], 
                    [ 676.25 , 1.448906648334812e-45 ], [ 676.3125 , 1.4491405260839774e-45 ], [ 676.375 , 1.4493744415848998e-45 ], [ 676.4375 , 1.4496083948437143e-45 ], 
                    [ 676.5 , 1.449842385866464e-45 ], [ 676.5625 , 1.450076414659306e-45 ], [ 676.625 , 1.4503104812282966e-45 ], [ 676.6875 , 1.450544585579523e-45 ], 
                    [ 676.75 , 1.4507787277191353e-45 ], [ 676.8125 , 1.4510129076531916e-45 ], [ 676.875 , 1.4512471253877932e-45 ], [ 676.9375 , 1.4514813809290823e-45 ], 
                    [ 677.0 , 1.45171567428309e-45 ], [ 677.0625 , 1.4519498646335855e-45 ], [ 677.125 , 1.4521839247378974e-45 ], [ 677.1875 , 1.4524180225736204e-45 ], 
                    [ 677.25 , 1.4526521581468372e-45 ], [ 677.3125 , 1.4528863314636517e-45 ], [ 677.375 , 1.4531205425301172e-45 ], [ 677.4375 , 1.4533547913523608e-45 ], 
                    [ 677.5 , 1.4535890779364482e-45 ], [ 677.5625 , 1.4538234022884664e-45 ], [ 677.625 , 1.4540577644144936e-45 ], [ 677.6875 , 1.45429216432064e-45 ], 
                    [ 677.75 , 1.454526602012975e-45 ], [ 677.8125 , 1.4547610774976104e-45 ], [ 677.875 , 1.4549955907806286e-45 ], [ 677.9375 , 1.4552301418681018e-45 ], 
                    [ 678.0 , 1.455464730766186e-45 ], [ 678.0625 , 1.4556993574808946e-45 ], [ 678.125 , 1.4559340220183954e-45 ], [ 678.1875 , 1.456168724384724e-45 ], 
                    [ 678.25 , 1.45640346458604e-45 ], [ 678.3125 , 1.4566382426283912e-45 ], [ 678.375 , 1.4568730585178986e-45 ], [ 678.4375 , 1.4571079122606738e-45 ], 
                    [ 678.5 , 1.4573428038627975e-45 ], [ 678.5625 , 1.4575777333303938e-45 ], [ 678.625 , 1.457812479552378e-45 ], [ 678.6875 , 1.4580470239986365e-45 ], 
                    [ 678.75 , 1.4582816061802777e-45 ], [ 678.8125 , 1.458516226103332e-45 ], [ 678.875 , 1.458750883773933e-45 ], [ 678.9375 , 1.458985579198113e-45 ], 
                    [ 679.0 , 1.459220312381946e-45 ], [ 679.0625 , 1.4594550833315378e-45 ], [ 679.125 , 1.4596898920529232e-45 ], [ 679.1875 , 1.4599247385522212e-45 ], 
                    [ 679.25 , 1.4601596228354672e-45 ], [ 679.3125 , 1.4603945449087725e-45 ], [ 679.375 , 1.460629504778206e-45 ], [ 679.4375 , 1.4608645024498386e-45 ], 
                    [ 679.5 , 1.4610995379297828e-45 ], [ 679.5625 , 1.461334611224091e-45 ], [ 679.625 , 1.4615697223388465e-45 ], [ 679.6875 , 1.4618048712801554e-45 ], 
                    [ 679.75 , 1.4620400580541037e-45 ], [ 679.8125 , 1.462275282666757e-45 ], [ 679.875 , 1.462510545124203e-45 ], [ 679.9375 , 1.4627458454325412e-45 ], 
                    [ 680.0 , 1.462981183597892e-45 ], [ 680.0625 , 1.463216559626274e-45 ], [ 680.125 , 1.4634519735238412e-45 ], [ 680.1875 , 1.4636872673627517e-45 ], 
                    [ 680.25 , 1.4639223979931306e-45 ], [ 680.3125 , 1.4641575663955218e-45 ], [ 680.375 , 1.464392772576003e-45 ], [ 680.4375 , 1.4646280165406334e-45 ], 
                    [ 680.5 , 1.4648632982954716e-45 ], [ 680.5625 , 1.4650986178465993e-45 ], [ 680.625 , 1.465333975200088e-45 ], [ 680.6875 , 1.4655693703620102e-45 ], 
                    [ 680.75 , 1.4658048033384396e-45 ], [ 680.8125 , 1.466040274135451e-45 ], [ 680.875 , 1.4662757827591303e-45 ], [ 680.9375 , 1.4665113292155225e-45 ], 
                    [ 681.0 , 1.466746913510758e-45 ], [ 681.0625 , 1.466982535650842e-45 ], [ 681.125 , 1.4672181956419163e-45 ], [ 681.1875 , 1.4674538934900618e-45 ], 
                    [ 681.25 , 1.4676896292012968e-45 ], [ 681.3125 , 1.467925402781767e-45 ], [ 681.375 , 1.4681612142375447e-45 ], [ 681.4375 , 1.468397063574694e-45 ], 
                    [ 681.5 , 1.4686329507993003e-45 ], [ 681.5625 , 1.4688688759174702e-45 ], [ 681.625 , 1.469104838935302e-45 ], [ 681.6875 , 1.4693408398588316e-45 ], 
                    [ 681.75 , 1.4695767779280384e-45 ], [ 681.8125 , 1.469812568351688e-45 ], [ 681.875 , 1.4700483966074023e-45 ], [ 681.9375 , 1.4702842627012515e-45 ], 
                    [ 682.0 , 1.4705201666393064e-45 ], [ 682.0625 , 1.470756108427629e-45 ], [ 682.125 , 1.470992088072313e-45 ], [ 682.1875 , 1.4712281055794113e-45 ], 
                    [ 682.25 , 1.47146416095502e-45 ], [ 682.3125 , 1.4717002542052147e-45 ], [ 682.375 , 1.471936385336031e-45 ], [ 682.4375 , 1.4721725543535987e-45 ], 
                    [ 682.5 , 1.4724087612639654e-45 ], [ 682.5625 , 1.472645006073242e-45 ], [ 682.625 , 1.472881288787478e-45 ], [ 682.6875 , 1.4731176094127555e-45 ], 
                    [ 682.75 , 1.4733539679551676e-45 ], [ 682.8125 , 1.4735903644208292e-45 ], [ 682.875 , 1.4738267988157726e-45 ], [ 682.9375 , 1.474063271146084e-45 ], 
                    [ 683.0 , 1.474299781417881e-45 ], [ 683.0625 , 1.474536329637262e-45 ], [ 683.125 , 1.4747729158102835e-45 ], [ 683.1875 , 1.4750095399430467e-45 ], 
                    [ 683.25 , 1.4752462020416305e-45 ], [ 683.3125 , 1.4754827380393782e-45 ], [ 683.375 , 1.4757186555664362e-45 ], [ 683.4375 , 1.475954610814752e-45 ], 
                    [ 683.5 , 1.4761906037903668e-45 ], [ 683.5625 , 1.4764266344993235e-45 ], [ 683.625 , 1.4766627029476133e-45 ], [ 683.6875 , 1.4768988091413337e-45 ], 
                    [ 683.75 , 1.4771349530864778e-45 ], [ 683.8125 , 1.4773711347890922e-45 ], [ 683.875 , 1.477607354255203e-45 ], [ 683.9375 , 1.4778436114908805e-45 ], 
                    [ 684.0 , 1.478079906502132e-45 ], [ 684.0625 , 1.478316239295008e-45 ], [ 684.125 , 1.478552609875528e-45 ], [ 684.1875 , 1.4787890182497873e-45 ], 
                    [ 684.25 , 1.4790254644237863e-45 ], [ 684.3125 , 1.4792619484035895e-45 ], [ 684.375 , 1.4794984701952217e-45 ], [ 684.4375 , 1.4797350298047387e-45 ], 
                    [ 684.5 , 1.4799716272381975e-45 ], [ 684.5625 , 1.480208262501625e-45 ], [ 684.625 , 1.48044493560106e-45 ], [ 684.6875 , 1.4806816465426036e-45 ], 
                    [ 684.75 , 1.4809183953322652e-45 ], [ 684.8125 , 1.4811551819761065e-45 ], [ 684.875 , 1.48139200648017e-45 ], [ 684.9375 , 1.4816287329394443e-45 ], 
                    [ 685.0 , 1.4818654950172906e-45 ], [ 685.0625 , 1.4821022949293874e-45 ], [ 685.125 , 1.4823391326817386e-45 ], [ 685.1875 , 1.4825760082804227e-45 ], 
                    [ 685.25 , 1.482812921731466e-45 ], [ 685.3125 , 1.483049873040949e-45 ], [ 685.375 , 1.4832868622148905e-45 ], [ 685.4375 , 1.4835238892593505e-45 ], 
                    [ 685.5 , 1.483760954180371e-45 ], [ 685.5625 , 1.4839980569840146e-45 ], [ 685.625 , 1.4842351976763564e-45 ], [ 685.6875 , 1.484472376263419e-45 ], 
                    [ 685.75 , 1.484709592751269e-45 ], [ 685.8125 , 1.484946847145973e-45 ], [ 685.875 , 1.4851841394535462e-45 ], [ 685.9375 , 1.4854214696801106e-45 ], 
                    [ 686.0 , 1.4856588378316833e-45 ], [ 686.0625 , 1.4858962439143455e-45 ], [ 686.125 , 1.4861336879341378e-45 ], [ 686.1875 , 1.486371169897112e-45 ], 
                    [ 686.25 , 1.486608689809373e-45 ], [ 686.3125 , 1.4868462476769648e-45 ], [ 686.375 , 1.4870838435059624e-45 ], [ 686.4375 , 1.487321477302411e-45 ], 
                    [ 686.5 , 1.4875589059169503e-45 ], [ 686.5625 , 1.4877962778579034e-45 ], [ 686.625 , 1.4880336876766534e-45 ], [ 686.6875 , 1.4882711353792233e-45 ], 
                    [ 686.75 , 1.4885086209716585e-45 ], [ 686.8125 , 1.4887461444600258e-45 ], [ 686.875 , 1.488983705850373e-45 ], [ 686.9375 , 1.489221305148716e-45 ], 
                    [ 687.0 , 1.4894589423611354e-45 ], [ 687.0625 , 1.4896966174936708e-45 ], [ 687.125 , 1.4899343305523735e-45 ], [ 687.1875 , 1.4901720815432953e-45 ], 
                    [ 687.25 , 1.4904098704724993e-45 ], [ 687.3125 , 1.4906476973460184e-45 ], [ 687.375 , 1.4908855621699076e-45 ], [ 687.4375 , 1.4911234649502645e-45 ], 
                    [ 687.5 , 1.4913614056930622e-45 ], [ 687.5625 , 1.4915993844044632e-45 ], [ 687.625 , 1.4918374010904422e-45 ], [ 687.6875 , 1.492075455757101e-45 ], 
                    [ 687.75 , 1.4923135484104895e-45 ], [ 687.8125 , 1.4925516790566694e-45 ], [ 687.875 , 1.492789847701693e-45 ], [ 687.9375 , 1.4930280543516656e-45 ], 
                    [ 688.0 , 1.4932662990125782e-45 ], [ 688.0625 , 1.4935042768625527e-45 ], [ 688.125 , 1.4937418197687284e-45 ], [ 688.1875 , 1.4939794004562598e-45 ], 
                    [ 688.25 , 1.494217018931188e-45 ], [ 688.3125 , 1.4944546751994804e-45 ], [ 688.375 , 1.494692369267191e-45 ], [ 688.4375 , 1.4949301011403104e-45 ], 
                    [ 688.5 , 1.4951678708248304e-45 ], [ 688.5625 , 1.4954056783267856e-45 ], [ 688.625 , 1.4956435236522236e-45 ], [ 688.6875 , 1.4958814068071063e-45 ], 
                    [ 688.75 , 1.4961193277974513e-45 ], [ 688.8125 , 1.4963572866293182e-45 ], [ 688.875 , 1.4965952833086942e-45 ], [ 688.9375 , 1.4968333178416415e-45 ], 
                    [ 689.0 , 1.497071390234117e-45 ], [ 689.0625 , 1.4973095004921844e-45 ], [ 689.125 , 1.4975476486218776e-45 ], [ 689.1875 , 1.497785834629166e-45 ], 
                    [ 689.25 , 1.4980240585201064e-45 ], [ 689.3125 , 1.4982623203007353e-45 ], [ 689.375 , 1.4985006199770894e-45 ], [ 689.4375 , 1.4987389575551425e-45 ], 
                    [ 689.5 , 1.4989773330409554e-45 ], [ 689.5625 , 1.4992157464405673e-45 ], [ 689.625 , 1.4994541977599987e-45 ], [ 689.6875 , 1.4996924975866333e-45 ], 
                    [ 689.75 , 1.4999308352849303e-45 ], [ 689.8125 , 1.5001692108608767e-45 ], [ 689.875 , 1.5004076243205242e-45 ], [ 689.9375 , 1.5006460756698933e-45 ], 
                    [ 690.0 , 1.500884564914984e-45 ], [ 690.0625 , 1.5011230920618192e-45 ], [ 690.125 , 1.5013616571164544e-45 ], [ 690.1875 , 1.501600260084882e-45 ], 
                    [ 690.25 , 1.5018389009731382e-45 ], [ 690.3125 , 1.50207757978726e-45 ], [ 690.375 , 1.502316296533253e-45 ], [ 690.4375 , 1.5025550512171354e-45 ], 
                    [ 690.5 , 1.5027938438450002e-45 ], [ 690.5625 , 1.5030326744228246e-45 ], [ 690.625 , 1.5032715429566283e-45 ], [ 690.6875 , 1.5035104494524765e-45 ], 
                    [ 690.75 , 1.5037493939163908e-45 ], [ 690.8125 , 1.5039883763544163e-45 ], [ 690.875 , 1.504227396772556e-45 ], [ 690.9375 , 1.5044664551768776e-45 ], 
                    [ 691.0 , 1.5047055515734183e-45 ], [ 691.0625 , 1.5049446859681947e-45 ], [ 691.125 , 1.5051838583672455e-45 ], [ 691.1875 , 1.5054230687766423e-45 ], 
                    [ 691.25 , 1.505662026741327e-45 ], [ 691.3125 , 1.5059007943805467e-45 ], [ 691.375 , 1.5061395998835e-45 ], [ 691.4375 , 1.5063784432562016e-45 ], 
                    [ 691.5 , 1.5066173245046464e-45 ], [ 691.5625 , 1.5068562436348298e-45 ], [ 691.625 , 1.50709520065276e-45 ], [ 691.6875 , 1.507334195564465e-45 ], 
                    [ 691.75 , 1.5075732283759664e-45 ], [ 691.8125 , 1.5078122990932196e-45 ], [ 691.875 , 1.5080514077222898e-45 ], [ 691.9375 , 1.5082905542691672e-45 ], 
                    [ 692.0 , 1.508529738739844e-45 ], [ 692.0625 , 1.508768961140366e-45 ], [ 692.125 , 1.5090082214767585e-45 ], [ 692.1875 , 1.5092475197549844e-45 ], 
                    [ 692.25 , 1.5094868559811028e-45 ], [ 692.3125 , 1.5097262301611316e-45 ], [ 692.375 , 1.509965642301058e-45 ], [ 692.4375 , 1.510205092406954e-45 ], 
                    [ 692.5 , 1.5104445804847773e-45 ], [ 692.5625 , 1.5106841065406023e-45 ], [ 692.625 , 1.5109236705804194e-45 ], [ 692.6875 , 1.511163272610263e-45 ], 
                    [ 692.75 , 1.5114029126361575e-45 ], [ 692.8125 , 1.5116425740523474e-45 ], [ 692.875 , 1.511881993399693e-45 ], [ 692.9375 , 1.5121214506671517e-45 ], 
                    [ 693.0 , 1.5123609458606857e-45 ], [ 693.0625 , 1.512600478986334e-45 ], [ 693.125 , 1.512840050050083e-45 ], [ 693.1875 , 1.5130796590579525e-45 ], 
                    [ 693.25 , 1.513319306015952e-45 ], [ 693.3125 , 1.513558990930103e-45 ], [ 693.375 , 1.5137987138064063e-45 ], [ 693.4375 , 1.514038474650864e-45 ], 
                    [ 693.5 , 1.5142782734695e-45 ], [ 693.5625 , 1.51451811026834e-45 ], [ 693.625 , 1.514757985053388e-45 ], [ 693.6875 , 1.514997897830661e-45 ], 
                    [ 693.75 , 1.515237848606144e-45 ], [ 693.8125 , 1.515477837385941e-45 ], [ 693.875 , 1.5157178641759746e-45 ], [ 693.9375 , 1.5159579289823296e-45 ], 
                    [ 694.0 , 1.516198031810995e-45 ], [ 694.0625 , 1.5164381726680038e-45 ], [ 694.125 , 1.516678351559389e-45 ], [ 694.1875 , 1.5169185684911542e-45 ], 
                    [ 694.25 , 1.517158823469345e-45 ], [ 694.3125 , 1.517399116499977e-45 ], [ 694.375 , 1.517639447589067e-45 ], [ 694.4375 , 1.517879571482242e-45 ], 
                    [ 694.5 , 1.518119476005233e-45 ], [ 694.5625 , 1.5183594184456897e-45 ], [ 694.625 , 1.518599398809649e-45 ], [ 694.6875 , 1.5188394171030833e-45 ], 
                    [ 694.75 , 1.519079473331976e-45 ], [ 694.8125 , 1.519319567502323e-45 ], [ 694.875 , 1.519559699620165e-45 ], [ 694.9375 , 1.5197998696914553e-45 ], 
                    [ 695.0 , 1.5200400777221828e-45 ], [ 695.0625 , 1.520280323718379e-45 ], [ 695.125 , 1.5205206076860334e-45 ], [ 695.1875 , 1.5207609296311586e-45 ], 
                    [ 695.25 , 1.5210012895597464e-45 ], [ 695.3125 , 1.5212416874777782e-45 ], [ 695.375 , 1.521482123391291e-45 ], [ 695.4375 , 1.5217225973063008e-45 ], 
                    [ 695.5 , 1.5219631092287596e-45 ], [ 695.5625 , 1.522203659164729e-45 ], [ 695.625 , 1.5224442471201844e-45 ], [ 695.6875 , 1.5226848731011455e-45 ], 
                    [ 695.75 , 1.5229255371136444e-45 ], [ 695.8125 , 1.523166239163649e-45 ], [ 695.875 , 1.523406979257181e-45 ], [ 695.9375 , 1.5236477574002866e-45 ], 
                    [ 696.0 , 1.5238885735989686e-45 ], [ 696.0625 , 1.5241290122525532e-45 ], [ 696.125 , 1.52436942782641e-45 ], [ 696.1875 , 1.5246098813233243e-45 ], 
                    [ 696.25 , 1.5248503727492885e-45 ], [ 696.3125 , 1.5250909021102965e-45 ], [ 696.375 , 1.5253314694122996e-45 ], [ 696.4375 , 1.5255720746613148e-45 ], 
                    [ 696.5 , 1.5258127178633172e-45 ], [ 696.5625 , 1.526053399024305e-45 ], [ 696.625 , 1.526294118150243e-45 ], [ 696.6875 , 1.5265348752471314e-45 ], 
                    [ 696.75 , 1.526775670320959e-45 ], [ 696.8125 , 1.5270165033777068e-45 ], [ 696.875 , 1.5272573744234082e-45 ], [ 696.9375 , 1.5274982834640128e-45 ], 
                    [ 697.0 , 1.5277392305055137e-45 ], [ 697.0625 , 1.5279802155539053e-45 ], [ 697.125 , 1.5282212386152045e-45 ], [ 697.1875 , 1.5284622996954076e-45 ], 
                    [ 697.25 , 1.5287033988004787e-45 ], [ 697.3125 , 1.5289445359364488e-45 ], [ 697.375 , 1.5291857111093057e-45 ], [ 697.4375 , 1.529426924325071e-45 ], 
                    [ 697.5 , 1.5296681755896918e-45 ], [ 697.5625 , 1.529909464909213e-45 ], [ 697.625 , 1.5301507005370014e-45 ], [ 697.6875 , 1.5303915880163196e-45 ], 
                    [ 697.75 , 1.5306325134178795e-45 ], [ 697.8125 , 1.5308734767476724e-45 ], [ 697.875 , 1.5311144780116913e-45 ], [ 697.9375 , 1.5313555172158755e-45 ], 
                    [ 698.0 , 1.5315965943661978e-45 ], [ 698.0625 , 1.5318377094686536e-45 ], [ 698.125 , 1.5320788625292174e-45 ], [ 698.1875 , 1.5323200535538323e-45 ], 
                    [ 698.25 , 1.5325612825484973e-45 ], [ 698.3125 , 1.5328025495192e-45 ], [ 698.375 , 1.533043854471908e-45 ], [ 698.4375 , 1.5332851974125905e-45 ], 
                    [ 698.5 , 1.5335265783472708e-45 ], [ 698.5625 , 1.533767997281854e-45 ], [ 698.625 , 1.5340094542223773e-45 ], [ 698.6875 , 1.534250949174845e-45 ], 
                    [ 698.75 , 1.5344924821451653e-45 ], [ 698.8125 , 1.5347340531393884e-45 ], [ 698.875 , 1.5349756621634787e-45 ], [ 698.9375 , 1.5352173092234234e-45 ], 
                    [ 699.0 , 1.5354589943251988e-45 ], [ 699.0625 , 1.535700717474827e-45 ], [ 699.125 , 1.535942478678232e-45 ], [ 699.1875 , 1.536184277941514e-45 ], 
                    [ 699.25 , 1.5364259470906702e-45 ], [ 699.3125 , 1.536667447004504e-45 ], [ 699.375 , 1.536908984877977e-45 ], [ 699.4375 , 1.537150560717078e-45 ], 
                    [ 699.5 , 1.5373921745277958e-45 ], [ 699.5625 , 1.537633826316067e-45 ], [ 699.625 , 1.5378755160878603e-45 ], [ 699.6875 , 1.538117243849168e-45 ], 
                    [ 699.75 , 1.538359009605951e-45 ], [ 699.8125 , 1.5386008133641805e-45 ], [ 699.875 , 1.538842655129809e-45 ], [ 699.9375 , 1.5390845349088646e-45 ], 
                    [ 700.0 , 1.5393264527072678e-45 ], [ 700.0625 , 1.5395684085310384e-45 ], [ 700.125 , 1.5398104023861094e-45 ], [ 700.1875 , 1.540052434278492e-45 ], 
                    [ 700.25 , 1.540294504214164e-45 ], [ 700.3125 , 1.5405366121990737e-45 ], [ 700.375 , 1.5407787582392335e-45 ], [ 700.4375 , 1.541020942340615e-45 ], 
                    [ 700.5 , 1.5412631645092e-45 ], [ 700.5625 , 1.541505424750972e-45 ], [ 700.625 , 1.541747723071905e-45 ], [ 700.6875 , 1.541990059478006e-45 ], 
                    [ 700.75 , 1.5422324339752506e-45 ], [ 700.8125 , 1.542474846569615e-45 ], [ 700.875 , 1.5427170974918282e-45 ], [ 700.9375 , 1.5429592816837147e-45 ], 
                    [ 701.0 , 1.5432015038949932e-45 ], [ 701.0625 , 1.543443764131676e-45 ], [ 701.125 , 1.543686062399677e-45 ], [ 701.1875 , 1.54392839870499e-45 ], 
                    [ 701.25 , 1.544170773053585e-45 ], [ 701.3125 , 1.5444131854514346e-45 ], [ 701.375 , 1.5446556359045123e-45 ], [ 701.4375 , 1.5448981244188027e-45 ], 
                    [ 701.5 , 1.5451406510002595e-45 ], [ 701.5625 , 1.545383215654869e-45 ], [ 701.625 , 1.5456258183886198e-45 ], [ 701.6875 , 1.5458684592074563e-45 ], 
                    [ 701.75 , 1.5461111381173792e-45 ], [ 701.8125 , 1.5463538551244017e-45 ], [ 701.875 , 1.5465966102344603e-45 ], [ 701.9375 , 1.5468394034535146e-45 ], 
                    [ 702.0 , 1.547082234787613e-45 ], [ 702.0625 , 1.5473251042427065e-45 ], [ 702.125 , 1.547568011824757e-45 ], [ 702.1875 , 1.547810957539794e-45 ], 
                    [ 702.25 , 1.548053941393771e-45 ], [ 702.3125 , 1.5482969633927085e-45 ], [ 702.375 , 1.548540023542528e-45 ], [ 702.4375 , 1.5487831218492853e-45 ], 
                    [ 702.5 , 1.5490258350346007e-45 ], [ 702.5625 , 1.5492684804183608e-45 ], [ 702.625 , 1.5495111638110412e-45 ], [ 702.6875 , 1.5497538852185734e-45 ], 
                    [ 702.75 , 1.5499966446469233e-45 ], [ 702.8125 , 1.550239442102058e-45 ], [ 702.875 , 1.550482277589934e-45 ], [ 702.9375 , 1.5507251511164647e-45 ], 
                    [ 703.0 , 1.5509680626876638e-45 ], [ 703.0625 , 1.5512110123094574e-45 ], [ 703.125 , 1.5514539999878284e-45 ], [ 703.1875 , 1.551697025728727e-45 ], 
                    [ 703.25 , 1.5519400895380934e-45 ], [ 703.3125 , 1.5521831914219346e-45 ], [ 703.375 , 1.5524263313861817e-45 ], [ 703.4375 , 1.5526695094368335e-45 ], 
                    [ 703.5 , 1.5529127255797892e-45 ], [ 703.5625 , 1.553155979821093e-45 ], [ 703.625 , 1.55339927216668e-45 ], [ 703.6875 , 1.553642602622529e-45 ], 
                    [ 703.75 , 1.553885971194556e-45 ], [ 703.8125 , 1.5541293778888193e-45 ], [ 703.875 , 1.554372822711213e-45 ], [ 703.9375 , 1.5546163056677657e-45 ], 
                    [ 704.0 , 1.5548598267644166e-45 ], [ 704.0625 , 1.555103386007174e-45 ], [ 704.125 , 1.555346557273178e-45 ], [ 704.1875 , 1.5555897084462893e-45 ], 
                    [ 704.25 , 1.5558328976318246e-45 ], [ 704.3125 , 1.556076124835715e-45 ], [ 704.375 , 1.5563193900639382e-45 ], [ 704.4375 , 1.556562693322405e-45 ], 
                    [ 704.5 , 1.5568060346170716e-45 ], [ 704.5625 , 1.557049413953896e-45 ], [ 704.625 , 1.5572928313388023e-45 ], [ 704.6875 , 1.5575362867777504e-45 ], 
                    [ 704.75 , 1.5577797802766783e-45 ], [ 704.8125 , 1.558023311841558e-45 ], [ 704.875 , 1.5582668814783403e-45 ], [ 704.9375 , 1.5585104891929442e-45 ], 
                    [ 705.0 , 1.5587541349913776e-45 ], [ 705.0625 , 1.5589978188795393e-45 ], [ 705.125 , 1.5592415408634056e-45 ], [ 705.1875 , 1.559485300948921e-45 ], 
                    [ 705.25 , 1.5597290991420643e-45 ], [ 705.3125 , 1.5599729354487824e-45 ], [ 705.375 , 1.5602168098750328e-45 ], [ 705.4375 , 1.5604607224267645e-45 ], 
                    [ 705.5 , 1.5607046731099707e-45 ], [ 705.5625 , 1.5609486619305903e-45 ], [ 705.625 , 1.5611926888945634e-45 ], [ 705.6875 , 1.561436754007886e-45 ], 
                    [ 705.75 , 1.5616803986684274e-45 ], [ 705.8125 , 1.5619240281073928e-45 ], [ 705.875 , 1.562167695553678e-45 ], [ 705.9375 , 1.5624114010132118e-45 ], 
                    [ 706.0 , 1.5626551444919472e-45 ], [ 706.0625 , 1.5628989259958154e-45 ], [ 706.125 , 1.563142745530726e-45 ], [ 706.1875 , 1.5633866031026126e-45 ], 
                    [ 706.25 , 1.5636304987174306e-45 ], [ 706.3125 , 1.563874432381115e-45 ], [ 706.375 , 1.5641184040995908e-45 ], [ 706.4375 , 1.5643624138787726e-45 ], 
                    [ 706.5 , 1.5646064617246308e-45 ], [ 706.5625 , 1.564850547643105e-45 ], [ 706.625 , 1.5650946716401e-45 ], [ 706.6875 , 1.5653388337215906e-45 ], 
                    [ 706.75 , 1.5655830338935066e-45 ], [ 706.8125 , 1.5658272721617906e-45 ], [ 706.875 , 1.5660715485323855e-45 ], [ 706.9375 , 1.5663158630112468e-45 ], 
                    [ 707.0 , 1.5665602156042972e-45 ], [ 707.0625 , 1.5668046063174825e-45 ], [ 707.125 , 1.5670490351567833e-45 ], [ 707.1875 , 1.5672935021281142e-45 ], 
                    [ 707.25 , 1.5675380072374457e-45 ], [ 707.3125 , 1.5677825504906834e-45 ], [ 707.375 , 1.568026863720567e-45 ], [ 707.4375 , 1.5682711728036262e-45 ], 
                    [ 707.5 , 1.5685155199516536e-45 ], [ 707.5625 , 1.568759905170614e-45 ], [ 707.625 , 1.5690043284664385e-45 ], [ 707.6875 , 1.569248789845038e-45 ], 
                    [ 707.75 , 1.569493289312368e-45 ], [ 707.8125 , 1.5697378268743292e-45 ], [ 707.875 , 1.5699824025369134e-45 ], [ 707.9375 , 1.570227016306012e-45 ], 
                    [ 708.0 , 1.5704716681875512e-45 ], [ 708.0625 , 1.5707163581875026e-45 ], [ 708.125 , 1.5709610863117943e-45 ], [ 708.1875 , 1.5712058525663775e-45 ], 
                    [ 708.25 , 1.5714506569571814e-45 ], [ 708.3125 , 1.571695499490115e-45 ], [ 708.375 , 1.5719403801712213e-45 ], [ 708.4375 , 1.5721852990062993e-45 ], 
                    [ 708.5 , 1.572430256001427e-45 ], [ 708.5625 , 1.5726752511624723e-45 ], [ 708.625 , 1.5729202844953927e-45 ], [ 708.6875 , 1.573165356006169e-45 ], 
                    [ 708.75 , 1.5734104657007277e-45 ], [ 708.8125 , 1.5736556135850064e-45 ], [ 708.875 , 1.5739007996649667e-45 ], [ 708.9375 , 1.574146023946571e-45 ], 
                    [ 709.0 , 1.5743908760879452e-45 ], [ 709.0625 , 1.5746356423366234e-45 ], [ 709.125 , 1.5748804466384304e-45 ], [ 709.1875 , 1.5751252889993047e-45 ], 
                    [ 709.25 , 1.5753701694251297e-45 ], [ 709.3125 , 1.5756150879218676e-45 ], [ 709.375 , 1.5758600444953932e-45 ], [ 709.4375 , 1.5761050391516704e-45 ], 
                    [ 709.5 , 1.576350071896575e-45 ], [ 709.5625 , 1.5765951427360847e-45 ], [ 709.625 , 1.576840251676066e-45 ], [ 709.6875 , 1.5770853987224757e-45 ], 
                    [ 709.75 , 1.5773305838812048e-45 ], [ 709.8125 , 1.577575807158212e-45 ], [ 709.875 , 1.577821068559412e-45 ], [ 709.9375 , 1.5780663680907434e-45 ], 
                    [ 710.0 , 1.578311705758123e-45 ], [ 710.0625 , 1.5785570815674456e-45 ], [ 710.125 , 1.5788024955247198e-45 ], [ 710.1875 , 1.579047947635798e-45 ], 
                    [ 710.25 , 1.579293437906668e-45 ], [ 710.3125 , 1.5795389663432396e-45 ], [ 710.375 , 1.579784532951435e-45 ], [ 710.4375 , 1.5800301377372008e-45 ], 
                    [ 710.5 , 1.58027578070646e-45 ], [ 710.5625 , 1.5805214618651833e-45 ], [ 710.625 , 1.5807668834728505e-45 ], [ 710.6875 , 1.58101215119146e-45 ], 
                    [ 710.75 , 1.5812574569651882e-45 ], [ 710.8125 , 1.5815028007999064e-45 ], [ 710.875 , 1.5817481827015417e-45 ], [ 710.9375 , 1.581993602676001e-45 ], 
                    [ 711.0 , 1.5822390607291914e-45 ], [ 711.0625 , 1.5824845568670324e-45 ], [ 711.125 , 1.5827300910954108e-45 ], [ 711.1875 , 1.5829756634202592e-45 ], 
                    [ 711.25 , 1.5832212738474767e-45 ], [ 711.3125 , 1.5834669223829646e-45 ], [ 711.375 , 1.5837126090326578e-45 ], [ 711.4375 , 1.5839583338024588e-45 ], 
                    [ 711.5 , 1.58420409669826e-45 ], [ 711.5625 , 1.5844498977260326e-45 ], [ 711.625 , 1.5846957368916485e-45 ], [ 711.6875 , 1.584941614201025e-45 ], 
                    [ 711.75 , 1.5851875296600803e-45 ], [ 711.8125 , 1.5854334832747672e-45 ], [ 711.875 , 1.585679475050995e-45 ], [ 711.9375 , 1.5859255049946398e-45 ], 
                    [ 712.0 , 1.586171573111679e-45 ], [ 712.0625 , 1.5864176794080138e-45 ], [ 712.125 , 1.5866638238895566e-45 ], [ 712.1875 , 1.5869100065622544e-45 ], 
                    [ 712.25 , 1.5871561464119024e-45 ], [ 712.3125 , 1.5874021944495408e-45 ], [ 712.375 , 1.5876482806306455e-45 ], [ 712.4375 , 1.5878944049611407e-45 ], 
                    [ 712.5 , 1.588140567446907e-45 ], [ 712.5625 , 1.5883867680938928e-45 ], [ 712.625 , 1.588633006908003e-45 ], [ 712.6875 , 1.588879283895132e-45 ], 
                    [ 712.75 , 1.5891255990612314e-45 ], [ 712.8125 , 1.5893719524122422e-45 ], [ 712.875 , 1.589618343954005e-45 ], [ 712.9375 , 1.5898647736925195e-45 ], 
                    [ 713.0 , 1.5901112416336505e-45 ], [ 713.0625 , 1.5903577477833432e-45 ], [ 713.125 , 1.590604292147543e-45 ], [ 713.1875 , 1.5908508747321406e-45 ], 
                    [ 713.25 , 1.591097495543061e-45 ], [ 713.3125 , 1.5913441545862414e-45 ], [ 713.375 , 1.59159085186762e-45 ], [ 713.4375 , 1.5918375873931136e-45 ], 
                    [ 713.5 , 1.5920843611686395e-45 ], [ 713.5625 , 1.59233117320015e-45 ], [ 713.625 , 1.592578023493553e-45 ], [ 713.6875 , 1.5928249120548258e-45 ], 
                    [ 713.75 , 1.593071838889844e-45 ], [ 713.8125 , 1.5933188040045515e-45 ], [ 713.875 , 1.5935657375073665e-45 ], [ 713.9375 , 1.5938121550067214e-45 ], 
                    [ 714.0 , 1.594058610610299e-45 ], [ 714.0625 , 1.5943051043239913e-45 ], [ 714.125 , 1.5945516361537148e-45 ], [ 714.1875 , 1.5947982061053175e-45 ], 
                    [ 714.25 , 1.5950448141847284e-45 ], [ 714.3125 , 1.5952914603978207e-45 ], [ 714.375 , 1.5955381447505142e-45 ], [ 714.4375 , 1.5957848672486945e-45 ], 
                    [ 714.5 , 1.596031627898238e-45 ], [ 714.5625 , 1.5962784267051004e-45 ], [ 714.625 , 1.596525263675126e-45 ], [ 714.6875 , 1.596772138814238e-45 ], 
                    [ 714.75 , 1.5970190521283617e-45 ], [ 714.8125 , 1.597266003623332e-45 ], [ 714.875 , 1.597512993305144e-45 ], [ 714.9375 , 1.5977600211796455e-45 ], 
                    [ 715.0 , 1.5980070872527542e-45 ], [ 715.0625 , 1.598254191530342e-45 ], [ 715.125 , 1.5985013340183968e-45 ], [ 715.1875 , 1.5987485147227585e-45 ], 
                    [ 715.25 , 1.5989957336493598e-45 ], [ 715.3125 , 1.5992429908041334e-45 ], [ 715.375 , 1.599490286192957e-45 ], [ 715.4375 , 1.5997376198217422e-45 ], 
                    [ 715.5 , 1.599984991696414e-45 ], [ 715.5625 , 1.6002320842803523e-45 ], [ 715.625 , 1.6004791233479207e-45 ], [ 715.6875 , 1.6007262005526683e-45 ], 
                    [ 715.75 , 1.6009733159004707e-45 ], [ 715.8125 , 1.6012204693971935e-45 ], [ 715.875 , 1.6014676610487267e-45 ], [ 715.9375 , 1.6017148908609714e-45 ], 
                    [ 716.0 , 1.6019621588398416e-45 ], [ 716.0625 , 1.6022094649911953e-45 ], [ 716.125 , 1.602456809320937e-45 ], [ 716.1875 , 1.60270419183496e-45 ], 
                    [ 716.25 , 1.6029516125391366e-45 ], [ 716.3125 , 1.6031990714394198e-45 ], [ 716.375 , 1.6034465685416484e-45 ], [ 716.4375 , 1.6036941038517664e-45 ], 
                    [ 716.5 , 1.603941677375626e-45 ], [ 716.5625 , 1.6041892891191613e-45 ], [ 716.625 , 1.6044369390882723e-45 ], [ 716.6875 , 1.6046846272888137e-45 ], 
                    [ 716.75 , 1.604932353726746e-45 ], [ 716.8125 , 1.6051801184079484e-45 ], [ 716.875 , 1.6054279213383362e-45 ], [ 716.9375 , 1.6056757625237803e-45 ], 
                    [ 717.0 , 1.6059236419702096e-45 ], [ 717.0625 , 1.6061715596835413e-45 ], [ 717.125 , 1.606419515669672e-45 ], [ 717.1875 , 1.6066672744592397e-45 ], 
                    [ 717.25 , 1.6069146806117777e-45 ], [ 717.3125 , 1.6071621248617126e-45 ], [ 717.375 , 1.6074096072148766e-45 ], [ 717.4375 , 1.6076571276771374e-45 ], 
                    [ 717.5 , 1.6079046862543864e-45 ], [ 717.5625 , 1.6081522829524806e-45 ], [ 717.625 , 1.608399917777279e-45 ], [ 717.6875 , 1.6086475907346873e-45 ], 
                    [ 717.75 , 1.6088953018305426e-45 ], [ 717.8125 , 1.6091430510707298e-45 ], [ 717.875 , 1.6093908384611332e-45 ], [ 717.9375 , 1.6096386640076055e-45 ], 
                    [ 718.0 , 1.6098865277160216e-45 ], [ 718.0625 , 1.61013442959227e-45 ], [ 718.125 , 1.6103823696422384e-45 ], [ 718.1875 , 1.610630347871795e-45 ], 
                    [ 718.25 , 1.610878364286829e-45 ], [ 718.3125 , 1.611126418893199e-45 ], [ 718.375 , 1.6113745116967964e-45 ], [ 718.4375 , 1.611622642703515e-45 ], 
                    [ 718.5 , 1.611870811919203e-45 ], [ 718.5625 , 1.612119019349779e-45 ], [ 718.625 , 1.6123672650011266e-45 ], [ 718.6875 , 1.6126155488790863e-45 ], 
                    [ 718.75 , 1.612863870989613e-45 ], [ 718.8125 , 1.6131122313385482e-45 ], [ 718.875 , 1.6133605923869454e-45 ], [ 718.9375 , 1.613608988050665e-45 ], 
                    [ 719.0 , 1.613857421957803e-45 ], [ 719.0625 , 1.6141058941142362e-45 ], [ 719.125 , 1.6143544045258533e-45 ], [ 719.1875 , 1.6146029531985326e-45 ], 
                    [ 719.25 , 1.6148515401381767e-45 ], [ 719.3125 , 1.615100165350688e-45 ], [ 719.375 , 1.615348828841948e-45 ], [ 719.4375 , 1.615597530617862e-45 ], 
                    [ 719.5 , 1.6158462706842897e-45 ], [ 719.5625 , 1.616095049047149e-45 ], [ 719.625 , 1.616343865712336e-45 ], [ 719.6875 , 1.6165927206857713e-45 ], 
                    [ 719.75 , 1.616841613973295e-45 ], [ 719.8125 , 1.6170905455808754e-45 ], [ 719.875 , 1.617339515514343e-45 ], [ 719.9375 , 1.6175885237796563e-45 ], 
                    [ 720.0 , 1.617837570382705e-45 ], [ 720.0625 , 1.618086655329369e-45 ], [ 720.125 , 1.6183357786255627e-45 ], [ 720.1875 , 1.61858494027718e-45 ], 
                    [ 720.25 , 1.618834140290125e-45 ], [ 720.3125 , 1.6190833786703395e-45 ], [ 720.375 , 1.6193326554237184e-45 ], [ 720.4375 , 1.619581970556135e-45 ], 
                    [ 720.5 , 1.6198309939088987e-45 ], [ 720.5625 , 1.6200795602279584e-45 ], [ 720.625 , 1.6203281646900307e-45 ], [ 720.6875 , 1.6205768073009456e-45 ], 
                    [ 720.75 , 1.6208254880665794e-45 ], [ 720.8125 , 1.621074206992765e-45 ], [ 720.875 , 1.6213229640853927e-45 ], [ 720.9375 , 1.6215717593502607e-45 ], 
                    [ 721.0 , 1.621820592793285e-45 ], [ 721.0625 , 1.622069464420301e-45 ], [ 721.125 , 1.6223183742371562e-45 ], [ 721.1875 , 1.622567322249734e-45 ], 
                    [ 721.25 , 1.6228163084638734e-45 ], [ 721.3125 , 1.6230653328854354e-45 ], [ 721.375 , 1.623314395520318e-45 ], [ 721.4375 , 1.6235634963743626e-45 ], 
                    [ 721.5 , 1.623812635453433e-45 ], [ 721.5625 , 1.6240618127633842e-45 ], [ 721.625 , 1.624311028310094e-45 ], [ 721.6875 , 1.6245602820994296e-45 ], 
                    [ 721.75 , 1.6248095741372715e-45 ], [ 721.8125 , 1.6250589044294772e-45 ], [ 721.875 , 1.6253082729819056e-45 ], [ 721.9375 , 1.6255576798004506e-45 ], 
                    [ 722.0 , 1.625807124890961e-45 ], [ 722.0625 , 1.626056608259345e-45 ], [ 722.125 , 1.6263061299114414e-45 ], [ 722.1875 , 1.6265553704785743e-45 ], 
                    [ 722.25 , 1.62680457334063e-45 ], [ 722.3125 , 1.6270538143828325e-45 ], [ 722.375 , 1.6273030936109384e-45 ], [ 722.4375 , 1.6275524110309148e-45 ], 
                    [ 722.5 , 1.6278017666485082e-45 ], [ 722.5625 , 1.628051160469652e-45 ], [ 722.625 , 1.6283005925001297e-45 ], [ 722.6875 , 1.628550062745865e-45 ], 
                    [ 722.75 , 1.6287995712126548e-45 ], [ 722.8125 , 1.6290491179063784e-45 ], [ 722.875 , 1.6292987028329152e-45 ], [ 722.9375 , 1.6295483259980768e-45 ], 
                    [ 723.0 , 1.629797987407733e-45 ], [ 723.0625 , 1.630047687067778e-45 ], [ 723.125 , 1.630297424984061e-45 ], [ 723.1875 , 1.6305472011624193e-45 ], 
                    [ 723.25 , 1.6307970156087274e-45 ], [ 723.3125 , 1.6310468683288245e-45 ], [ 723.375 , 1.6312967593286325e-45 ], [ 723.4375 , 1.6315466886139583e-45 ], 
                    [ 723.5 , 1.6317966561907022e-45 ], [ 723.5625 , 1.6320466620647198e-45 ], [ 723.625 , 1.632296706241855e-45 ], [ 723.6875 , 1.6325467887280455e-45 ], 
                    [ 723.75 , 1.6327969095290913e-45 ], [ 723.8125 , 1.6330470138225747e-45 ], [ 723.875 , 1.633296884853745e-45 ], [ 723.9375 , 1.633546794117476e-45 ], 
                    [ 724.0 , 1.6337967416195724e-45 ], [ 724.0625 , 1.6340467273658957e-45 ], [ 724.125 , 1.6342967513623447e-45 ], [ 724.1875 , 1.6345468136147135e-45 ], 
                    [ 724.25 , 1.634796914128879e-45 ], [ 724.3125 , 1.635047052910707e-45 ], [ 724.375 , 1.6352972299660303e-45 ], [ 724.4375 , 1.6355474453007045e-45 ], 
                    [ 724.5 , 1.635797698920633e-45 ], [ 724.5625 , 1.6360479908316275e-45 ], [ 724.625 , 1.6362983210395358e-45 ], [ 724.6875 , 1.6365486895502633e-45 ], 
                    [ 724.75 , 1.6367990963696594e-45 ], [ 724.8125 , 1.6370495415035277e-45 ], [ 724.875 , 1.6373000249578354e-45 ], [ 724.9375 , 1.6375505467383522e-45 ], 
                    [ 725.0 , 1.6378011068509897e-45 ], [ 725.0625 , 1.6380517053015776e-45 ], [ 725.125 , 1.6383023420960405e-45 ], [ 725.1875 , 1.6385530172401636e-45 ], 
                    [ 725.25 , 1.6388037307398964e-45 ], [ 725.3125 , 1.639054482601084e-45 ], [ 725.375 , 1.6393052728295733e-45 ], [ 725.4375 , 1.6395561014312114e-45 ], 
                    [ 725.5 , 1.6398067345606017e-45 ], [ 725.5625 , 1.6400571358234995e-45 ], [ 725.625 , 1.640307575323082e-45 ], [ 725.6875 , 1.640558053065212e-45 ], 
                    [ 725.75 , 1.6408085690557283e-45 ], [ 725.8125 , 1.641059123300437e-45 ], [ 725.875 , 1.6413097158052263e-45 ], [ 725.9375 , 1.641560346575903e-45 ], 
                    [ 726.0 , 1.641811015618323e-45 ], [ 726.0625 , 1.6420617229383294e-45 ], [ 726.125 , 1.6423124685417682e-45 ], [ 726.1875 , 1.642563252434473e-45 ], 
                    [ 726.25 , 1.6428140746223142e-45 ], [ 726.3125 , 1.6430649351111278e-45 ], [ 726.375 , 1.6433158339067624e-45 ], [ 726.4375 , 1.6435667710150564e-45 ], 
                    [ 726.5 , 1.6438177464418944e-45 ], [ 726.5625 , 1.644068760193093e-45 ], [ 726.625 , 1.6443198122745046e-45 ], [ 726.6875 , 1.6445709026920173e-45 ], 
                    [ 726.75 , 1.6448220314514497e-45 ], [ 726.8125 , 1.645073198558657e-45 ], [ 726.875 , 1.6453244040195298e-45 ], [ 726.9375 , 1.6455756478398898e-45 ], 
                    [ 727.0 , 1.6458269300255946e-45 ], [ 727.0625 , 1.6460782505825263e-45 ], [ 727.125 , 1.6463296095165438e-45 ], [ 727.1875 , 1.6465806240274875e-45 ], 
                    [ 727.25 , 1.646831509582345e-45 ], [ 727.3125 , 1.6470824333640228e-45 ], [ 727.375 , 1.6473333953783813e-45 ], [ 727.4375 , 1.6475843956311987e-45 ], 
                    [ 727.5 , 1.6478354341283367e-45 ], [ 727.5625 , 1.6480865108756104e-45 ], [ 727.625 , 1.6483376258788364e-45 ], [ 727.6875 , 1.6485887791438788e-45 ], 
                    [ 727.75 , 1.6488399706765204e-45 ], [ 727.8125 , 1.6490912004826276e-45 ], [ 727.875 , 1.64934246856802e-45 ], [ 727.9375 , 1.6495937749385298e-45 ], 
                    [ 728.0 , 1.649845119599991e-45 ], [ 728.0625 , 1.6500965025582138e-45 ], [ 728.125 , 1.650347923819081e-45 ], [ 728.1875 , 1.6505993833884048e-45 ], 
                    [ 728.25 , 1.6508508812720223e-45 ], [ 728.3125 , 1.6511024174757714e-45 ], [ 728.375 , 1.6513539920054792e-45 ], [ 728.4375 , 1.6516056048669965e-45 ], 
                    [ 728.5 , 1.651857256066188e-45 ], [ 728.5625 , 1.6521089456088475e-45 ], [ 728.625 , 1.6523606735008414e-45 ], [ 728.6875 , 1.6526124397480127e-45 ], 
                    [ 728.75 , 1.652864244356229e-45 ], [ 728.8125 , 1.6531160873313116e-45 ], [ 728.875 , 1.6533675313004932e-45 ], [ 728.9375 , 1.6536189446579714e-45 ], 
                    [ 729.0 , 1.6538703962457483e-45 ], [ 729.0625 , 1.6541218860695665e-45 ], [ 729.125 , 1.6543734141352997e-45 ], [ 729.1875 , 1.6546249804487152e-45 ], 
                    [ 729.25 , 1.654876585015665e-45 ], [ 729.3125 , 1.655128227841977e-45 ], [ 729.375 , 1.6553799089333992e-45 ], [ 729.4375 , 1.6556316282958315e-45 ], 
                    [ 729.5 , 1.6558833859350593e-45 ], [ 729.5625 , 1.6561351818568903e-45 ], [ 729.625 , 1.6563870160671467e-45 ], [ 729.6875 , 1.6566388885716854e-45 ], 
                    [ 729.75 , 1.6568907993762944e-45 ], [ 729.8125 , 1.6571427484868095e-45 ], [ 729.875 , 1.6573947359090786e-45 ], [ 729.9375 , 1.6576467616488692e-45 ], 
                    [ 730.0 , 1.6578988257120545e-45 ], [ 730.0625 , 1.6581509281044506e-45 ], [ 730.125 , 1.658403068831874e-45 ], [ 730.1875 , 1.6586552479001894e-45 ], 
                    [ 730.25 , 1.6589074653152033e-45 ], [ 730.3125 , 1.659159721082723e-45 ], [ 730.375 , 1.659412015208616e-45 ], [ 730.4375 , 1.6596643476987154e-45 ], 
                    [ 730.5 , 1.6599167185588306e-45 ], [ 730.5625 , 1.6601689074092305e-45 ], [ 730.625 , 1.6604211272877578e-45 ], [ 730.6875 , 1.6606733854845836e-45 ], 
                    [ 730.75 , 1.6609256820055173e-45 ], [ 730.8125 , 1.6611780168564405e-45 ], [ 730.875 , 1.661430390043129e-45 ], [ 730.9375 , 1.6616828015714073e-45 ], 
                    [ 731.0 , 1.6619352514471118e-45 ], [ 731.0625 , 1.662187739676092e-45 ], [ 731.125 , 1.6624402662641517e-45 ], [ 731.1875 , 1.6626928312170947e-45 ], 
                    [ 731.25 , 1.6629454345407963e-45 ], [ 731.3125 , 1.663198076241051e-45 ], [ 731.375 , 1.6634507563237008e-45 ], [ 731.4375 , 1.6637034747945652e-45 ], 
                    [ 731.5 , 1.6639562316595113e-45 ], [ 731.5625 , 1.6642090269243374e-45 ], [ 731.625 , 1.6644618605948887e-45 ], [ 731.6875 , 1.664714732677e-45 ], 
                    [ 731.75 , 1.6649676431765188e-45 ], [ 731.8125 , 1.665220592099234e-45 ], [ 731.875 , 1.6654735794510314e-45 ], [ 731.9375 , 1.6657266052377717e-45 ], 
                    [ 732.0 , 1.6659796694652118e-45 ], [ 732.0625 , 1.666232772139263e-45 ], [ 732.125 , 1.666485913265766e-45 ], [ 732.1875 , 1.66673907701453e-45 ], 
                    [ 732.25 , 1.6669916351287588e-45 ], [ 732.3125 , 1.667244231512674e-45 ], [ 732.375 , 1.667496866172098e-45 ], [ 732.4375 , 1.6677495391128307e-45 ], 
                    [ 732.5 , 1.6680022503406486e-45 ], [ 732.5625 , 1.668254999861366e-45 ], [ 732.625 , 1.6685077876807724e-45 ], [ 732.6875 , 1.6687606138046958e-45 ], 
                    [ 732.75 , 1.669013478238928e-45 ], [ 732.8125 , 1.6692663809892746e-45 ], [ 732.875 , 1.669519322061553e-45 ], [ 732.9375 , 1.6697723014615346e-45 ], 
                    [ 733.0 , 1.670025319195051e-45 ], [ 733.0625 , 1.6702783752679224e-45 ], [ 733.125 , 1.6705314696859468e-45 ], [ 733.1875 , 1.670784602454946e-45 ], 
                    [ 733.25 , 1.671037773580696e-45 ], [ 733.3125 , 1.671290983069044e-45 ], [ 733.375 , 1.6715442309257922e-45 ], [ 733.4375 , 1.6717975171567537e-45 ], 
                    [ 733.5 , 1.6720508417677313e-45 ], [ 733.5625 , 1.6723042047645768e-45 ], [ 733.625 , 1.6725576061530707e-45 ], [ 733.6875 , 1.6728110459390307e-45 ], 
                    [ 733.75 , 1.6730645241282987e-45 ], [ 733.8125 , 1.673318040726682e-45 ], [ 733.875 , 1.6735715957400012e-45 ], [ 733.9375 , 1.6738248278162336e-45 ], 
                    [ 734.0 , 1.6740780982096458e-45 ], [ 734.0625 , 1.6743314069259648e-45 ], [ 734.125 , 1.6745847539710485e-45 ], [ 734.1875 , 1.674838139350661e-45 ], 
                    [ 734.25 , 1.6750915630706383e-45 ], [ 734.3125 , 1.675345025136722e-45 ], [ 734.375 , 1.675598525554763e-45 ], [ 734.4375 , 1.675852064330539e-45 ], 
                    [ 734.5 , 1.676105641469903e-45 ], [ 734.5625 , 1.676359256978588e-45 ], [ 734.625 , 1.6766129108624474e-45 ], [ 734.6875 , 1.6768666031272758e-45 ], 
                    [ 734.75 , 1.6771203337788928e-45 ], [ 734.8125 , 1.6773741028230827e-45 ], [ 734.875 , 1.6776279102656665e-45 ], [ 734.9375 , 1.677881756112467e-45 ], 
                    [ 735.0 , 1.6781356403692713e-45 ], [ 735.0625 , 1.6783895630419146e-45 ], [ 735.125 , 1.678643524136174e-45 ], [ 735.1875 , 1.6788975236579114e-45 ], 
                    [ 735.25 , 1.6791515616129168e-45 ], [ 735.3125 , 1.6794056380069945e-45 ], [ 735.375 , 1.6796597528459837e-45 ], [ 735.4375 , 1.6799139061356666e-45 ], 
                    [ 735.5 , 1.680168097881885e-45 ], [ 735.5625 , 1.6804223280904932e-45 ], [ 735.625 , 1.6806760796240164e-45 ], [ 735.6875 , 1.6809298293459735e-45 ], 
                    [ 735.75 , 1.6811836173792518e-45 ], [ 735.8125 , 1.6814374437296358e-45 ], [ 735.875 , 1.6816913084029105e-45 ], [ 735.9375 , 1.68194521140485e-45 ], 
                    [ 736.0 , 1.6821991527412772e-45 ], [ 736.0625 , 1.682453132417956e-45 ], [ 736.125 , 1.6827071504406387e-45 ], [ 736.1875 , 1.6829612068151754e-45 ], 
                    [ 736.25 , 1.6832153015473437e-45 ], [ 736.3125 , 1.6834694346428874e-45 ], [ 736.375 , 1.6837236061076702e-45 ], [ 736.4375 , 1.6839778159474377e-45 ], 
                    [ 736.5 , 1.6842320641680074e-45 ], [ 736.5625 , 1.6844863507751263e-45 ], [ 736.625 , 1.6847406757746377e-45 ], [ 736.6875 , 1.6849950391723504e-45 ], 
                    [ 736.75 , 1.685249440974014e-45 ], [ 736.8125 , 1.685503881185474e-45 ], [ 736.875 , 1.685758359812482e-45 ], [ 736.9375 , 1.6860128768608742e-45 ], 
                    [ 737.0 , 1.686267432336439e-45 ], [ 737.0625 , 1.6865220262449787e-45 ], [ 737.125 , 1.6867766585922838e-45 ], [ 737.1875 , 1.6870313293841933e-45 ], 
                    [ 737.25 , 1.6872860386264642e-45 ], [ 737.3125 , 1.6875405176024944e-45 ], [ 737.375 , 1.6877949643751953e-45 ], [ 737.4375 , 1.688049449513292e-45 ], 
                    [ 737.5 , 1.688303973022581e-45 ], [ 737.5625 , 1.6885585349088006e-45 ], [ 737.625 , 1.688813135177809e-45 ], [ 737.6875 , 1.6890677738353573e-45 ], 
                    [ 737.75 , 1.689322450887222e-45 ], [ 737.8125 , 1.689577166339216e-45 ], [ 737.875 , 1.6898319201970935e-45 ], [ 737.9375 , 1.690086712466693e-45 ], 
                    [ 738.0 , 1.69034154315377e-45 ], [ 738.0625 , 1.69059641226413e-45 ], [ 738.125 , 1.6908513198035655e-45 ], [ 738.1875 , 1.6911062657778712e-45 ], 
                    [ 738.25 , 1.6913612501928546e-45 ], [ 738.3125 , 1.691616273054287e-45 ], [ 738.375 , 1.6918713343679903e-45 ], [ 738.4375 , 1.6921264341397497e-45 ], 
                    [ 738.5 , 1.6923815723753278e-45 ], [ 738.5625 , 1.6926367490805965e-45 ], [ 738.625 , 1.6928919642613203e-45 ], [ 738.6875 , 1.693147217923276e-45 ], 
                    [ 738.75 , 1.6934025100723023e-45 ], [ 738.8125 , 1.6936578407142022e-45 ], [ 738.875 , 1.6939132098547553e-45 ], [ 738.9375 , 1.694168617499779e-45 ], 
                    [ 739.0 , 1.694423731840758e-45 ], [ 739.0625 , 1.6946786558805737e-45 ], [ 739.125 , 1.6949336182734088e-45 ], [ 739.1875 , 1.695188619025033e-45 ], 
                    [ 739.25 , 1.695443658141206e-45 ], [ 739.3125 , 1.695698735627735e-45 ], [ 739.375 , 1.6959538514903574e-45 ], [ 739.4375 , 1.6962090057348585e-45 ], 
                    [ 739.5 , 1.6964641983670133e-45 ], [ 739.5625 , 1.6967194293926083e-45 ], [ 739.625 , 1.6969746988173967e-45 ], [ 739.6875 , 1.697230006647167e-45 ], 
                    [ 739.75 , 1.6974853528876968e-45 ], [ 739.8125 , 1.6977407375447663e-45 ], [ 739.875 , 1.6979961606241417e-45 ], [ 739.9375 , 1.6982516221316407e-45 ], 
                    [ 740.0 , 1.6985071220729965e-45 ], [ 740.0625 , 1.698762660454027e-45 ], [ 740.125 , 1.699018237280516e-45 ], [ 740.1875 , 1.6992738525581986e-45 ], 
                    [ 740.25 , 1.699529506292933e-45 ], [ 740.3125 , 1.6997851984904806e-45 ], [ 740.375 , 1.7000409291565552e-45 ], [ 740.4375 , 1.7002966982970656e-45 ], 
                    [ 740.5 , 1.7005525059177035e-45 ], [ 740.5625 , 1.7008083520243066e-45 ], [ 740.625 , 1.7010642366226768e-45 ], [ 740.6875 , 1.7013200472673666e-45 ], 
                    [ 740.75 , 1.701575665959199e-45 ], [ 740.8125 , 1.7018313230570296e-45 ], [ 740.875 , 1.7020870185666536e-45 ], [ 740.9375 , 1.7023427524938172e-45 ], 
                    [ 741.0 , 1.7025985248443175e-45 ], [ 741.0625 , 1.7028543356239152e-45 ], [ 741.125 , 1.7031101848383596e-45 ], [ 741.1875 , 1.7033660724934743e-45 ], 
                    [ 741.25 , 1.7036219985950105e-45 ], [ 741.3125 , 1.7038779631487446e-45 ], [ 741.375 , 1.704133966160442e-45 ], [ 741.4375 , 1.7043900076358928e-45 ], 
                    [ 741.5 , 1.704646087580876e-45 ], [ 741.5625 , 1.704902206001172e-45 ], [ 741.625 , 1.7051583629025735e-45 ], [ 741.6875 , 1.70541455829085e-45 ], 
                    [ 741.75 , 1.7056707921717717e-45 ], [ 741.8125 , 1.7059270645511467e-45 ], [ 741.875 , 1.706183375434735e-45 ], [ 741.9375 , 1.7064397248283458e-45 ], 
                    [ 742.0 , 1.706696112737741e-45 ], [ 742.0625 , 1.7069525391687316e-45 ], [ 742.125 , 1.7072090041270807e-45 ], [ 742.1875 , 1.7074655076185778e-45 ], 
                    [ 742.25 , 1.7077220496490608e-45 ], [ 742.3125 , 1.7079786302242466e-45 ], [ 742.375 , 1.70823524935e-45 ], [ 742.4375 , 1.708491213083797e-45 ], 
                    [ 742.5 , 1.708747215171464e-45 ], [ 742.5625 , 1.7090032556187485e-45 ], [ 742.625 , 1.709259334431386e-45 ], [ 742.6875 , 1.709515451615149e-45 ], 
                    [ 742.75 , 1.7097716071758e-45 ], [ 742.8125 , 1.7100278011190156e-45 ], [ 742.875 , 1.7102840334506453e-45 ], [ 742.9375 , 1.7105403041763555e-45 ], 
                    [ 743.0 , 1.7107966133019238e-45 ], [ 743.0625 , 1.7110529608331402e-45 ], [ 743.125 , 1.711309346775748e-45 ], [ 743.1875 , 1.7115657711354776e-45 ], 
                    [ 743.25 , 1.7118222339180864e-45 ], [ 743.3125 , 1.7120787351293674e-45 ], [ 743.375 , 1.7123352747750546e-45 ], [ 743.4375 , 1.712591852860871e-45 ], 
                    [ 743.5 , 1.712848469392637e-45 ], [ 743.5625 , 1.7131051243760892e-45 ], [ 743.625 , 1.7133618178169767e-45 ], [ 743.6875 , 1.7136185497210992e-45 ], 
                    [ 743.75 , 1.7138753200941952e-45 ], [ 743.8125 , 1.7141321289419923e-45 ], [ 743.875 , 1.7143889762703293e-45 ], [ 743.9375 , 1.7146458620849105e-45 ], 
                    [ 744.0 , 1.714902786391564e-45 ], [ 744.0625 , 1.7151597491959847e-45 ], [ 744.125 , 1.71541651360315e-45 ], [ 744.1875 , 1.715673163687136e-45 ], 
                    [ 744.25 , 1.7159298521695317e-45 ], [ 744.3125 , 1.7161865790560814e-45 ], [ 744.375 , 1.7164433443525436e-45 ], [ 744.4375 , 1.7167001480646524e-45 ], 
                    [ 744.5 , 1.7169569901981313e-45 ], [ 744.5625 , 1.7172138707587772e-45 ], [ 744.625 , 1.7174707897523028e-45 ], [ 744.6875 , 1.717727747184483e-45 ], 
                    [ 744.75 , 1.7179847430610437e-45 ], [ 744.8125 , 1.718241777387761e-45 ], [ 744.875 , 1.718498850170364e-45 ], [ 744.9375 , 1.7187559614146178e-45 ], 
                    [ 745.0 , 1.7190131111262774e-45 ], [ 745.0625 , 1.719270299311097e-45 ], [ 745.125 , 1.7195275259748705e-45 ], [ 745.1875 , 1.7197847911232563e-45 ], 
                    [ 745.25 , 1.7200420947621107e-45 ], [ 745.3125 , 1.720299436897155e-45 ], [ 745.375 , 1.7205568175341494e-45 ], [ 745.4375 , 1.7208142366788295e-45 ], 
                    [ 745.5 , 1.7210716943369935e-45 ], [ 745.5625 , 1.7213291905144154e-45 ], [ 745.625 , 1.7215867252167852e-45 ], [ 745.6875 , 1.7218442984499526e-45 ], 
                    [ 745.75 , 1.722101910219645e-45 ], [ 745.8125 , 1.722359500990418e-45 ], [ 745.875 , 1.7226167980644296e-45 ], [ 745.9375 , 1.7228741335751476e-45 ], 
                    [ 746.0 , 1.7231315075282894e-45 ], [ 746.0625 , 1.7233889199295857e-45 ], [ 746.125 , 1.723646370784792e-45 ], [ 746.1875 , 1.723903860099641e-45 ], 
                    [ 746.25 , 1.724161387879927e-45 ], [ 746.3125 , 1.7244189541313344e-45 ], [ 746.375 , 1.724676558859623e-45 ], [ 746.4375 , 1.7249342020705898e-45 ], 
                    [ 746.5 , 1.7251918837699222e-45 ], [ 746.5625 , 1.7254496039634067e-45 ], [ 746.625 , 1.7257073626567687e-45 ], [ 746.6875 , 1.725965159855785e-45 ], 
                    [ 746.75 , 1.726222995566158e-45 ], [ 746.8125 , 1.726480869793727e-45 ], [ 746.875 , 1.7267387825441844e-45 ], [ 746.9375 , 1.7269967338232856e-45 ], 
                    [ 747.0 , 1.72725472363681e-45 ], [ 747.0625 , 1.7275127519904775e-45 ], [ 747.125 , 1.727770818890095e-45 ], [ 747.1875 , 1.7280289243413956e-45 ], 
                    [ 747.25 , 1.7282870683501146e-45 ], [ 747.3125 , 1.7285452509220602e-45 ], [ 747.375 , 1.728803472062957e-45 ], [ 747.4375 , 1.729061731778603e-45 ], 
                    [ 747.5 , 1.7293200300747e-45 ], [ 747.5625 , 1.7295778436090795e-45 ], [ 747.625 , 1.7298353907996328e-45 ], [ 747.6875 , 1.7300929763409184e-45 ], 
                    [ 747.75 , 1.7303506002386222e-45 ], [ 747.8125 , 1.7306082624984805e-45 ], [ 747.875 , 1.7308659631261935e-45 ], [ 747.9375 , 1.731123702127462e-45 ], 
                    [ 748.0 , 1.731381479508012e-45 ], [ 748.0625 , 1.7316392952735717e-45 ], [ 748.125 , 1.7318971494298438e-45 ], [ 748.1875 , 1.732155041982521e-45 ], 
                    [ 748.25 , 1.732412972937382e-45 ], [ 748.3125 , 1.7326709423000832e-45 ], [ 748.375 , 1.7329289500763813e-45 ], [ 748.4375 , 1.7331869962719595e-45 ], 
                    [ 748.5 , 1.733445080892588e-45 ], [ 748.5625 , 1.733703203943963e-45 ], [ 748.625 , 1.7339613654318087e-45 ], [ 748.6875 , 1.7342195653618476e-45 ], 
                    [ 748.75 , 1.7344778037397795e-45 ], [ 748.8125 , 1.7347360805713912e-45 ], [ 748.875 , 1.734994395862347e-45 ], [ 748.9375 , 1.7352527496183987e-45 ], 
                    [ 749.0 , 1.7355111418452863e-45 ], [ 749.0625 , 1.7357695725487015e-45 ], [ 749.125 , 1.7360280417344227e-45 ], [ 749.1875 , 1.736286549408144e-45 ], 
                    [ 749.25 , 1.7365450912965408e-45 ], [ 749.3125 , 1.7368035868019422e-45 ], [ 749.375 , 1.7370621207859905e-45 ], [ 749.4375 , 1.73732069325445e-45 ], 
                    [ 749.5 , 1.7375793042130122e-45 ], [ 749.5625 , 1.737837953667432e-45 ], [ 749.625 , 1.7380966416234022e-45 ], [ 749.6875 , 1.7383553680866914e-45 ], 
                    [ 749.75 , 1.738614133063031e-45 ], [ 749.8125 , 1.7388729365581053e-45 ], [ 749.875 , 1.739131778577734e-45 ], [ 749.9375 , 1.7393906591275527e-45 ], 
                    [ 750.0 , 1.739649578213384e-45 ], [ 750.0625 , 1.7399085358409145e-45 ], [ 750.125 , 1.7401675320158812e-45 ], [ 750.1875 , 1.7404265667440592e-45 ], 
                    [ 750.25 , 1.7406856400311507e-45 ], [ 750.3125 , 1.740944751882932e-45 ], [ 750.375 , 1.7412039023050702e-45 ], [ 750.4375 , 1.74146309130338e-45 ], 
                    [ 750.5 , 1.7417223188835797e-45 ], [ 750.5625 , 1.7419815850514245e-45 ], [ 750.625 , 1.7422408898126464e-45 ], [ 750.6875 , 1.742500233172965e-45 ], 
                    [ 750.75 , 1.7427596151381635e-45 ], [ 750.8125 , 1.7430190357139634e-45 ], [ 750.875 , 1.7432784949061498e-45 ], [ 750.9375 , 1.7435379927204456e-45 ], 
                    [ 751.0 , 1.7437973260187295e-45 ], [ 751.0625 , 1.7440563931291506e-45 ], [ 751.125 , 1.744315498727856e-45 ], [ 751.1875 , 1.7445746428205767e-45 ], 
                    [ 751.25 , 1.7448338254129936e-45 ], [ 751.3125 , 1.7450930465108643e-45 ], [ 751.375 , 1.7453523061198968e-45 ], [ 751.4375 , 1.7456116042458122e-45 ], 
                    [ 751.5 , 1.745870940894333e-45 ], [ 751.5625 , 1.7461303160711822e-45 ], [ 751.625 , 1.746389729782084e-45 ], [ 751.6875 , 1.7466491820327628e-45 ], 
                    [ 751.75 , 1.7469086728289446e-45 ], [ 751.8125 , 1.747168202176356e-45 ], [ 751.875 , 1.747427770080724e-45 ], [ 751.9375 , 1.7476873765477896e-45 ], 
                    [ 752.0 , 1.7479470215832563e-45 ], [ 752.0625 , 1.7482067051928425e-45 ], [ 752.125 , 1.7484664273823653e-45 ], [ 752.1875 , 1.7487261881574324e-45 ], 
                    [ 752.25 , 1.748985987523888e-45 ], [ 752.3125 , 1.7492458254873664e-45 ], [ 752.375 , 1.7495057020536755e-45 ], [ 752.4375 , 1.749765617228526e-45 ], 
                    [ 752.5 , 1.750025571017666e-45 ], [ 752.5625 , 1.7502855634267954e-45 ], [ 752.625 , 1.7505455944617008e-45 ], [ 752.6875 , 1.7508056641280716e-45 ], 
                    [ 752.75 , 1.7510653559025002e-45 ], [ 752.8125 , 1.751324911038707e-45 ], [ 752.875 , 1.751584504647989e-45 ], [ 752.9375 , 1.7518441367360237e-45 ], 
                    [ 753.0 , 1.752103807308565e-45 ], [ 753.0625 , 1.752363516371279e-45 ], [ 753.125 , 1.752623263929897e-45 ], [ 753.1875 , 1.7528830499900996e-45 ], 
                    [ 753.25 , 1.7531428745576062e-45 ], [ 753.3125 , 1.753402737638137e-45 ], [ 753.375 , 1.7536626392373765e-45 ], [ 753.4375 , 1.753922579361071e-45 ], 
                    [ 753.5 , 1.7541825580148685e-45 ], [ 753.5625 , 1.7544425752045425e-45 ], [ 753.625 , 1.7547026309357553e-45 ], [ 753.6875 , 1.7549627252142823e-45 ], 
                    [ 753.75 , 1.7552228580457626e-45 ], [ 753.8125 , 1.755483029435985e-45 ], [ 753.875 , 1.7557432393905904e-45 ], [ 753.9375 , 1.756003487915345e-45 ], 
                    [ 754.0 , 1.7562637750159662e-45 ], [ 754.0625 , 1.7565241006981467e-45 ], [ 754.125 , 1.7567844649676303e-45 ], [ 754.1875 , 1.757044867830112e-45 ], 
                    [ 754.25 , 1.757305309291349e-45 ], [ 754.3125 , 1.7575657893570012e-45 ], [ 754.375 , 1.7578263080328653e-45 ], [ 754.4375 , 1.7580868653246018e-45 ], 
                    [ 754.5 , 1.7583470102717288e-45 ], [ 754.5625 , 1.7586071587187187e-45 ], [ 754.625 , 1.7588673456548325e-45 ], [ 754.6875 , 1.7591275710857276e-45 ], 
                    [ 754.75 , 1.7593878350171244e-45 ], [ 754.8125 , 1.7596481374547563e-45 ], [ 754.875 , 1.7599084784042454e-45 ], [ 754.9375 , 1.760168857871352e-45 ], 
                    [ 755.0 , 1.7604292758617248e-45 ], [ 755.0625 , 1.7606897323811132e-45 ], [ 755.125 , 1.7609502274351805e-45 ], [ 755.1875 , 1.7612107610296278e-45 ], 
                    [ 755.25 , 1.7614713331701944e-45 ], [ 755.3125 , 1.7617319438625585e-45 ], [ 755.375 , 1.7619925931123988e-45 ], [ 755.4375 , 1.7622532809254447e-45 ], 
                    [ 755.5 , 1.762514007307427e-45 ], [ 755.5625 , 1.7627747722640267e-45 ], [ 755.625 , 1.7630355758009137e-45 ], [ 755.6875 , 1.763296417923858e-45 ], 
                    [ 755.75 , 1.7635572986385437e-45 ], [ 755.8125 , 1.7638182179506936e-45 ], [ 755.875 , 1.764079175865992e-45 ], [ 755.9375 , 1.7643401723901638e-45 ], 
                    [ 756.0 , 1.7646012075289204e-45 ], [ 756.0625 , 1.7648622812879756e-45 ], [ 756.125 , 1.765123393673043e-45 ], [ 756.1875 , 1.765384494110806e-45 ], 
                    [ 756.25 , 1.765645108910931e-45 ], [ 756.3125 , 1.7659057621842912e-45 ], [ 756.375 , 1.7661664539366034e-45 ], [ 756.4375 , 1.766427184173485e-45 ], 
                    [ 756.5 , 1.7666879529006925e-45 ], [ 756.5625 , 1.766948760123846e-45 ], [ 756.625 , 1.7672096058486654e-45 ], [ 756.6875 , 1.7674704900808225e-45 ], 
                    [ 756.75 , 1.767731412825989e-45 ], [ 756.8125 , 1.767992374089863e-45 ], [ 756.875 , 1.7682533738781435e-45 ], [ 756.9375 , 1.7685144121965053e-45 ], 
                    [ 757.0 , 1.7687754890506107e-45 ], [ 757.0625 , 1.7690366044462118e-45 ], [ 757.125 , 1.7692977583889352e-45 ], [ 757.1875 , 1.7695589508845096e-45 ], 
                    [ 757.25 , 1.7698201819386132e-45 ], [ 757.3125 , 1.7700814515569134e-45 ], [ 757.375 , 1.7703427597451663e-45 ], [ 757.4375 , 1.7706041065090023e-45 ], 
                    [ 757.5 , 1.7708654918541415e-45 ], [ 757.5625 , 1.7711269157862924e-45 ], [ 757.625 , 1.771388378311126e-45 ], [ 757.6875 , 1.7716498794343896e-45 ], 
                    [ 757.75 , 1.771911419161719e-45 ], [ 757.8125 , 1.7721729974988124e-45 ], [ 757.875 , 1.7724346144514204e-45 ], [ 757.9375 , 1.7726962235628317e-45 ], 
                    [ 758.0 , 1.77295766520655e-45 ], [ 758.0625 , 1.7732191454083185e-45 ], [ 758.125 , 1.7734806641738496e-45 ], [ 758.1875 , 1.7737422215087927e-45 ], 
                    [ 758.25 , 1.7740038174188994e-45 ], [ 758.3125 , 1.7742654519098084e-45 ], [ 758.375 , 1.774527124987197e-45 ], [ 758.4375 , 1.774788836656806e-45 ], 
                    [ 758.5 , 1.7750505869243028e-45 ], [ 758.5625 , 1.7753123757953666e-45 ], [ 758.625 , 1.7755742032757036e-45 ], [ 758.6875 , 1.7758360693710206e-45 ], 
                    [ 758.75 , 1.7760979740869875e-45 ], [ 758.8125 , 1.7763599174293253e-45 ], [ 758.875 , 1.7766218994036927e-45 ], [ 758.9375 , 1.7768839200158378e-45 ], 
                    [ 759.0 , 1.7771459792714214e-45 ], [ 759.0625 , 1.777408077176155e-45 ], [ 759.125 , 1.7776702137357644e-45 ], [ 759.1875 , 1.7779323889558743e-45 ], 
                    [ 759.25 , 1.7781946028422875e-45 ], [ 759.3125 , 1.7784568554006435e-45 ], [ 759.375 , 1.778719146636646e-45 ], [ 759.4375 , 1.778981476556024e-45 ], 
                    [ 759.5 , 1.7792438451644453e-45 ], [ 759.5625 , 1.7795062524676662e-45 ], [ 759.625 , 1.7797686984713682e-45 ], [ 759.6875 , 1.780031064651027e-45 ], 
                    [ 759.75 , 1.780293077073883e-45 ], [ 759.8125 , 1.7805551280637936e-45 ], [ 759.875 , 1.7808172176263973e-45 ], [ 759.9375 , 1.7810793457673974e-45 ], 
                    [ 760.0 , 1.7813415124924218e-45 ], [ 760.0625 , 1.7816037178072383e-45 ], [ 760.125 , 1.7818659617174513e-45 ], [ 760.1875 , 1.7821282442287673e-45 ], 
                    [ 760.25 , 1.7823905653468804e-45 ], [ 760.3125 , 1.7826529250774364e-45 ], [ 760.375 , 1.7829153234261683e-45 ], [ 760.4375 , 1.783177760398736e-45 ], 
                    [ 760.5 , 1.783440236000799e-45 ], [ 760.5625 , 1.7837027502380943e-45 ], [ 760.625 , 1.783965303116258e-45 ], [ 760.6875 , 1.784227894641029e-45 ], 
                    [ 760.75 , 1.7844905248180445e-45 ], [ 760.8125 , 1.7847531936530322e-45 ], [ 760.875 , 1.7850159011516954e-45 ], [ 760.9375 , 1.7852786473196615e-45 ], 
                    [ 761.0 , 1.7855414321626983e-45 ], [ 761.0625 , 1.7858042556864103e-45 ], [ 761.125 , 1.7860671178965926e-45 ], [ 761.1875 , 1.7863300187988504e-45 ], 
                    [ 761.25 , 1.7865929583989427e-45 ], [ 761.3125 , 1.7868559367025277e-45 ], [ 761.375 , 1.7871189537153276e-45 ], [ 761.4375 , 1.7873818934331503e-45 ], 
                    [ 761.5 , 1.787644407729821e-45 ], [ 761.5625 , 1.7879069605821766e-45 ], [ 761.625 , 1.788169551995905e-45 ], [ 761.6875 , 1.7884321819766705e-45 ], 
                    [ 761.75 , 1.7886948505301113e-45 ], [ 761.8125 , 1.788957557661893e-45 ], [ 761.875 , 1.7892203033776944e-45 ], [ 761.9375 , 1.789483087683182e-45 ], 
                    [ 762.0 , 1.789745910584062e-45 ], [ 762.0625 , 1.7900087720859393e-45 ], [ 762.125 , 1.7902716721944957e-45 ], [ 762.1875 , 1.7905346109154143e-45 ], 
                    [ 762.25 , 1.7907975882543667e-45 ], [ 762.3125 , 1.791060604217036e-45 ], [ 762.375 , 1.7913236588090966e-45 ], [ 762.4375 , 1.7915867520361955e-45 ], 
                    [ 762.5 , 1.7918498839040074e-45 ], [ 762.5625 , 1.7921130544182204e-45 ], [ 762.625 , 1.7923762635845228e-45 ], [ 762.6875 , 1.7926395114085788e-45 ], 
                    [ 762.75 , 1.7929027978960666e-45 ], [ 762.8125 , 1.7931661230526517e-45 ], [ 762.875 , 1.7934294868840262e-45 ], [ 762.9375 , 1.793692889395883e-45 ], 
                    [ 763.0 , 1.7939563305938774e-45 ], [ 763.0625 , 1.79421981048373e-45 ], [ 763.125 , 1.7944833290710854e-45 ], [ 763.1875 , 1.7947468304530906e-45 ], 
                    [ 763.25 , 1.795009888783379e-45 ], [ 763.3125 , 1.7952729856704365e-45 ], [ 763.375 , 1.7955361211199656e-45 ], [ 763.4375 , 1.79579929513758e-45 ], 
                    [ 763.5 , 1.796062507728933e-45 ], [ 763.5625 , 1.7963257588997035e-45 ], [ 763.625 , 1.7965890486555466e-45 ], [ 763.6875 , 1.796852377002092e-45 ], 
                    [ 763.75 , 1.797115743944996e-45 ], [ 763.8125 , 1.797379149489941e-45 ], [ 763.875 , 1.7976425936425727e-45 ], [ 763.9375 , 1.7979060764085493e-45 ], 
                    [ 764.0 , 1.7981695977935438e-45 ], [ 764.0625 , 1.798433157803165e-45 ], [ 764.125 , 1.7986967564431383e-45 ], [ 764.1875 , 1.7989603937191257e-45 ], 
                    [ 764.25 , 1.7992240696367395e-45 ], [ 764.3125 , 1.7994877842016298e-45 ], [ 764.375 , 1.7997515374195767e-45 ], [ 764.4375 , 1.800015329296118e-45 ], 
                    [ 764.5 , 1.8002791598369705e-45 ], [ 764.5625 , 1.8005430290478146e-45 ], [ 764.625 , 1.800806936934331e-45 ], [ 764.6875 , 1.8010708835021238e-45 ], 
                    [ 764.75 , 1.8013348687568892e-45 ], [ 764.8125 , 1.8015988927043477e-45 ], [ 764.875 , 1.8018629553501202e-45 ], [ 764.9375 , 1.8021270566999166e-45 ], 
                    [ 765.0 , 1.8023908107588215e-45 ], [ 765.0625 , 1.8026545805183348e-45 ], [ 765.125 , 1.802918388879069e-45 ], [ 765.1875 , 1.8031822358466736e-45 ], 
                    [ 765.25 , 1.8034461214267978e-45 ], [ 765.3125 , 1.803710045625093e-45 ], [ 765.375 , 1.8039740084472104e-45 ], [ 765.4375 , 1.8042380098988022e-45 ], 
                    [ 765.5 , 1.804502049985522e-45 ], [ 765.5625 , 1.8047661287130238e-45 ], [ 765.625 , 1.8050302460869622e-45 ], [ 765.6875 , 1.8052944021129932e-45 ], 
                    [ 765.75 , 1.805558596796773e-45 ], [ 765.8125 , 1.805822830143959e-45 ], [ 765.875 , 1.8060871021601966e-45 ], [ 765.9375 , 1.8063514128511837e-45 ], 
                    [ 766.0 , 1.806615762222541e-45 ], [ 766.0625 , 1.8068801502799423e-45 ], [ 766.125 , 1.8071445770290747e-45 ], [ 766.1875 , 1.8074090424755362e-45 ], 
                    [ 766.25 , 1.8076735466250413e-45 ], [ 766.3125 , 1.8079380894832285e-45 ], [ 766.375 , 1.8082026710557884e-45 ], [ 766.4375 , 1.8084672913483863e-45 ], 
                    [ 766.5 , 1.8087319503666505e-45 ], [ 766.5625 , 1.8089966481162733e-45 ], [ 766.625 , 1.8092613846029623e-45 ], [ 766.6875 , 1.8095261598323086e-45 ], 
                    [ 766.75 , 1.809790584283114e-45 ], [ 766.8125 , 1.810054895869223e-45 ], [ 766.875 , 1.8103192460568358e-45 ], [ 766.9375 , 1.8105836348515512e-45 ], 
                    [ 767.0 , 1.8108480622590335e-45 ], [ 767.0625 , 1.8111125282849343e-45 ], [ 767.125 , 1.8113770329348687e-45 ], [ 767.1875 , 1.8116415762145152e-45 ], 
                    [ 767.25 , 1.8119061581294776e-45 ], [ 767.3125 , 1.812170778685411e-45 ], [ 767.375 , 1.812435437887959e-45 ], [ 767.4375 , 1.8127001357427526e-45 ], 
                    [ 767.5 , 1.8129648722554625e-45 ], [ 767.5625 , 1.813229647431709e-45 ], [ 767.625 , 1.81349446127719e-45 ], [ 767.6875 , 1.813759313797476e-45 ], 
                    [ 767.75 , 1.8140242049982665e-45 ], [ 767.8125 , 1.814289134885185e-45 ], [ 767.875 , 1.8145541034639072e-45 ], [ 767.9375 , 1.8148191107400706e-45 ], 
                    [ 768.0 , 1.8150841567193404e-45 ], [ 768.0625 , 1.8153492414073297e-45 ], [ 768.125 , 1.815614364809718e-45 ], [ 768.1875 , 1.8158795269321587e-45 ], 
                    [ 768.25 , 1.8161447277803205e-45 ], [ 768.3125 , 1.816409967359871e-45 ], [ 768.375 , 1.81667524567639e-45 ], [ 768.4375 , 1.8169405627356253e-45 ], 
                    [ 768.5 , 1.817205684277804e-45 ], [ 768.5625 , 1.817470536445116e-45 ], [ 768.625 , 1.8177354272138336e-45 ], [ 768.6875 , 1.8180003565895314e-45 ], 
                    [ 768.75 , 1.8182653245779138e-45 ], [ 768.8125 , 1.8185303311845437e-45 ], [ 768.875 , 1.8187953764150755e-45 ], [ 768.9375 , 1.8190604602751514e-45 ], 
                    [ 769.0 , 1.8193255827703762e-45 ], [ 769.0625 , 1.8195907439064186e-45 ], [ 769.125 , 1.8198559436888853e-45 ], [ 769.1875 , 1.8201211821233702e-45 ], 
                    [ 769.25 , 1.8203864592155577e-45 ], [ 769.3125 , 1.8206517749710828e-45 ], [ 769.375 , 1.8209171293955803e-45 ], [ 769.4375 , 1.821182522494647e-45 ], 
                    [ 769.5 , 1.8214479542739718e-45 ], [ 769.5625 , 1.821713424739127e-45 ], [ 769.625 , 1.821978933895803e-45 ], [ 769.6875 , 1.8222444817496395e-45 ], 
                    [ 769.75 , 1.8225100683062496e-45 ], [ 769.8125 , 1.8227756935713138e-45 ], [ 769.875 , 1.8230413575504343e-45 ], [ 769.9375 , 1.823307060249267e-45 ], 
                    [ 770.0 , 1.8235728016734548e-45 ], [ 770.0625 , 1.823838581828616e-45 ], [ 770.125 , 1.8241044007204734e-45 ], [ 770.1875 , 1.8243702583545818e-45 ], 
                    [ 770.25 , 1.8246361182094535e-45 ], [ 770.3125 , 1.8249015966835625e-45 ], [ 770.375 , 1.8251671137839048e-45 ], [ 770.4375 , 1.8254326695160873e-45 ], 
                    [ 770.5 , 1.8256982638857826e-45 ], [ 770.5625 , 1.8259638968985348e-45 ], [ 770.625 , 1.826229568560005e-45 ], [ 770.6875 , 1.8264952788758295e-45 ], 
                    [ 770.75 , 1.8267610278516324e-45 ], [ 770.8125 , 1.8270268154929994e-45 ], [ 770.875 , 1.8272926418056087e-45 ], [ 770.9375 , 1.8275585067950474e-45 ], 
                    [ 771.0 , 1.827824410466956e-45 ], [ 771.0625 , 1.8280903528269497e-45 ], [ 771.125 , 1.8283563338806963e-45 ], [ 771.1875 , 1.828622353633774e-45 ], 
                    [ 771.25 , 1.828888412091865e-45 ], [ 771.3125 , 1.829154509260562e-45 ], [ 771.375 , 1.8294206451455106e-45 ], [ 771.4375 , 1.829686819752343e-45 ], 
                    [ 771.5 , 1.8299530330866938e-45 ], [ 771.5625 , 1.8302192851541842e-45 ], [ 771.625 , 1.8304855759604893e-45 ], [ 771.6875 , 1.8307519055112065e-45 ], 
                    [ 771.75 , 1.8310182738119722e-45 ], [ 771.8125 , 1.8312846808684514e-45 ], [ 771.875 , 1.8315511266862692e-45 ], [ 771.9375 , 1.8318176112710656e-45 ], 
                    [ 772.0 , 1.8320841346284808e-45 ], [ 772.0625 , 1.832350363245007e-45 ], [ 772.125 , 1.832616461755198e-45 ], [ 772.1875 , 1.8328825989088813e-45 ], 
                    [ 772.25 , 1.833148774711682e-45 ], [ 772.3125 , 1.8334149891691863e-45 ], [ 772.375 , 1.8336812422870342e-45 ], [ 772.4375 , 1.8339475340708138e-45 ], 
                    [ 772.5 , 1.8342138645261795e-45 ], [ 772.5625 , 1.834480233658682e-45 ], [ 772.625 , 1.8347466414740163e-45 ], [ 772.6875 , 1.835013087977735e-45 ], 
                    [ 772.75 , 1.8352795731754826e-45 ], [ 772.8125 , 1.835546097072878e-45 ], [ 772.875 , 1.8358126596755546e-45 ], [ 772.9375 , 1.8360792609891202e-45 ], 
                    [ 773.0 , 1.8363459010191835e-45 ], [ 773.0625 , 1.8366125797713796e-45 ], [ 773.125 , 1.8368792972513324e-45 ], [ 773.1875 , 1.837146053464679e-45 ], 
                    [ 773.25 , 1.8374128484169916e-45 ], [ 773.3125 , 1.8376796821140012e-45 ], [ 773.375 , 1.8379465545612297e-45 ], [ 773.4375 , 1.8382134657643432e-45 ], 
                    [ 773.5 , 1.8384804157289704e-45 ], [ 773.5625 , 1.8387474044607667e-45 ], [ 773.625 , 1.839014431965296e-45 ], [ 773.6875 , 1.8392814982482414e-45 ], 
                    [ 773.75 , 1.8395486033152348e-45 ], [ 773.8125 , 1.8398156429638756e-45 ], [ 773.875 , 1.840082259170696e-45 ], [ 773.9375 , 1.8403489140140798e-45 ], 
                    [ 774.0 , 1.840615607499639e-45 ], [ 774.0625 , 1.8408823396329995e-45 ], [ 774.125 , 1.841149110419775e-45 ], [ 774.1875 , 1.841415919865515e-45 ], 
                    [ 774.25 , 1.8416827679758605e-45 ], [ 774.3125 , 1.8419496547563756e-45 ], [ 774.375 , 1.8422165802127296e-45 ], [ 774.4375 , 1.8424835443504748e-45 ], 
                    [ 774.5 , 1.8427505471751906e-45 ], [ 774.5625 , 1.8430175886925746e-45 ], [ 774.625 , 1.8432846689081556e-45 ], [ 774.6875 , 1.843551787827568e-45 ], 
                    [ 774.75 , 1.84381894545642e-45 ], [ 774.8125 , 1.844086141800348e-45 ], [ 774.875 , 1.8443533768648964e-45 ], [ 774.9375 , 1.844620650655742e-45 ], 
                    [ 775.0 , 1.8448879631784705e-45 ], [ 775.0625 , 1.845155314438669e-45 ], [ 775.125 , 1.8454227044420027e-45 ], [ 775.1875 , 1.8456901331940475e-45 ], 
                    [ 775.25 , 1.8459576007004315e-45 ], [ 775.3125 , 1.8462251069667837e-45 ], [ 775.375 , 1.846492651998695e-45 ], [ 775.4375 , 1.8467602358017964e-45 ], 
                    [ 775.5 , 1.847027858381719e-45 ], [ 775.5625 , 1.847295519744056e-45 ], [ 775.625 , 1.8475630520699905e-45 ], [ 775.6875 , 1.8478305220013068e-45 ], 
                    [ 775.75 , 1.8480980306540014e-45 ], [ 775.8125 , 1.8483655780336674e-45 ], [ 775.875 , 1.8486331641459106e-45 ], [ 775.9375 , 1.848900788996352e-45 ], 
                    [ 776.0 , 1.849168452590547e-45 ], [ 776.0625 , 1.8494361549341833e-45 ], [ 776.125 , 1.8497038960328048e-45 ], [ 776.1875 , 1.8499716758920488e-45 ], 
                    [ 776.25 , 1.8502394945175652e-45 ], [ 776.3125 , 1.850507351914888e-45 ], [ 776.375 , 1.8507752480896692e-45 ], [ 776.4375 , 1.851043183047549e-45 ], 
                    [ 776.5 , 1.8513111567941025e-45 ], [ 776.5625 , 1.8515791693349585e-45 ], [ 776.625 , 1.8518472206757198e-45 ], [ 776.6875 , 1.8521153108220168e-45 ], 
                    [ 776.75 , 1.85238343977948e-45 ], [ 776.8125 , 1.8526516075537155e-45 ], [ 776.875 , 1.8529198141503558e-45 ], [ 776.9375 , 1.8531880595749946e-45 ], 
                    [ 777.0 , 1.853456343833266e-45 ], [ 777.0625 , 1.8537246669307926e-45 ], [ 777.125 , 1.8539930288731703e-45 ], [ 777.1875 , 1.854261429666075e-45 ], 
                    [ 777.25 , 1.8545298693151058e-45 ], [ 777.3125 , 1.854798347825887e-45 ], [ 777.375 , 1.8550668512315458e-45 ], [ 777.4375 , 1.855334534073054e-45 ], 
                    [ 777.5 , 1.8556022555407196e-45 ], [ 777.5625 , 1.8558700156401028e-45 ], [ 777.625 , 1.8561378143768173e-45 ], [ 777.6875 , 1.8564056517563863e-45 ], 
                    [ 777.75 , 1.8566735277844384e-45 ], [ 777.8125 , 1.8569414424665238e-45 ], [ 777.875 , 1.8572093958081945e-45 ], [ 777.9375 , 1.857477387815068e-45 ], 
                    [ 778.0 , 1.8577454184927104e-45 ], [ 778.0625 , 1.8580134878467023e-45 ], [ 778.125 , 1.8582815958826242e-45 ], [ 778.1875 , 1.858549742606045e-45 ], 
                    [ 778.25 , 1.8588179280225733e-45 ], [ 778.3125 , 1.8590861521377925e-45 ], [ 778.375 , 1.8593544149572607e-45 ], [ 778.4375 , 1.8596227164865622e-45 ], 
                    [ 778.5 , 1.8598910567313364e-45 ], [ 778.5625 , 1.8601594356971032e-45 ], [ 778.625 , 1.8604278533894632e-45 ], [ 778.6875 , 1.8606963098140318e-45 ], 
                    [ 778.75 , 1.860964804976397e-45 ], [ 778.8125 , 1.8612333388821092e-45 ], [ 778.875 , 1.861501911536812e-45 ], [ 778.9375 , 1.861770522946044e-45 ], 
                    [ 779.0 , 1.8620391731154368e-45 ], [ 779.0625 , 1.8623078620505707e-45 ], [ 779.125 , 1.862576589757026e-45 ], [ 779.1875 , 1.8628452749165488e-45 ], 
                    [ 779.25 , 1.863113832196275e-45 ], [ 779.3125 , 1.863382428192602e-45 ], [ 779.375 , 1.863651062911085e-45 ], [ 779.4375 , 1.86391973635732e-45 ], 
                    [ 779.5 , 1.8641884485368895e-45 ], [ 779.5625 , 1.8644571994553915e-45 ], [ 779.625 , 1.864725989118384e-45 ], [ 779.6875 , 1.864994817531465e-45 ], 
                    [ 779.75 , 1.8652636847002356e-45 ], [ 779.8125 , 1.8655325906302556e-45 ], [ 779.875 , 1.8658015353271268e-45 ], [ 779.9375 , 1.8660705187964374e-45 ], 
                    [ 780.0 , 1.866339541043778e-45 ], [ 780.0625 , 1.8666086020747378e-45 ], [ 780.125 , 1.8668777018949088e-45 ], [ 780.1875 , 1.8671468405098962e-45 ], 
                    [ 780.25 , 1.8674160179252928e-45 ], [ 780.3125 , 1.8676852341466528e-45 ], [ 780.375 , 1.8679544891795965e-45 ], [ 780.4375 , 1.86822378302972e-45 ], 
                    [ 780.5 , 1.868493115702632e-45 ], [ 780.5625 , 1.8687624872039032e-45 ], [ 780.625 , 1.8690318975391577e-45 ], [ 780.6875 , 1.869301346713967e-45 ], 
                    [ 780.75 , 1.8695708347339316e-45 ], [ 780.8125 , 1.869840361604677e-45 ], [ 780.875 , 1.8701099273318183e-45 ], [ 780.9375 , 1.8703795319209167e-45 ], 
                    [ 781.0 , 1.8706489312180676e-45 ], [ 781.0625 , 1.870918189563655e-45 ], [ 781.125 , 1.8711874866658902e-45 ], [ 781.1875 , 1.8714568225303383e-45 ], 
                    [ 781.25 , 1.8717261971625784e-45 ], [ 781.3125 , 1.8719956105681647e-45 ], [ 781.375 , 1.8722650627527178e-45 ], [ 781.4375 , 1.8725345537218193e-45 ], 
                    [ 781.5 , 1.872804083481039e-45 ], [ 781.5625 , 1.8730736520359463e-45 ], [ 781.625 , 1.873343259392139e-45 ], [ 781.6875 , 1.8736129055552153e-45 ], 
                    [ 781.75 , 1.873882590530735e-45 ], [ 781.8125 , 1.8741523143243102e-45 ], [ 781.875 , 1.874422076941503e-45 ], [ 781.9375 , 1.8746918783879138e-45 ], 
                    [ 782.0 , 1.8749617186691326e-45 ], [ 782.0625 , 1.8752315977907747e-45 ], [ 782.125 , 1.8755015157583918e-45 ], [ 782.1875 , 1.875771472577575e-45 ], 
                    [ 782.25 , 1.8760414682539295e-45 ], [ 782.3125 , 1.8763115027930486e-45 ], [ 782.375 , 1.8765815762005266e-45 ], [ 782.4375 , 1.876851688481971e-45 ], 
                    [ 782.5 , 1.8771218396429637e-45 ], [ 782.5625 , 1.8773920296891012e-45 ], [ 782.625 , 1.8776622586259544e-45 ], [ 782.6875 , 1.8779325264591603e-45 ], 
                    [ 782.75 , 1.8782028331943314e-45 ], [ 782.8125 , 1.8784726251508218e-45 ], [ 782.875 , 1.8787422809878864e-45 ], [ 782.9375 , 1.879011975534203e-45 ], 
                    [ 783.0 , 1.879281708795315e-45 ], [ 783.0625 , 1.879551480776807e-45 ], [ 783.125 , 1.879821291484224e-45 ], [ 783.1875 , 1.8800911409230975e-45 ], 
                    [ 783.25 , 1.8803610290990552e-45 ], [ 783.3125 , 1.8806309560175905e-45 ], [ 783.375 , 1.8809009216843185e-45 ], [ 783.4375 , 1.881170926104748e-45 ], 
                    [ 783.5 , 1.881440969284483e-45 ], [ 783.5625 , 1.881711051229046e-45 ], [ 783.625 , 1.881981171944056e-45 ], [ 783.6875 , 1.8822513314350654e-45 ], 
                    [ 783.75 , 1.882521529707573e-45 ], [ 783.8125 , 1.8827917667672798e-45 ], [ 783.875 , 1.88306204261958e-45 ], [ 783.9375 , 1.8833323572702164e-45 ], 
                    [ 784.0 , 1.8836027107246374e-45 ], [ 784.0625 , 1.883873102988481e-45 ], [ 784.125 , 1.8841435340672376e-45 ], [ 784.1875 , 1.884414003966586e-45 ], 
                    [ 784.25 , 1.8846845126920192e-45 ], [ 784.3125 , 1.8849550602491638e-45 ], [ 784.375 , 1.885225646643554e-45 ], [ 784.4375 , 1.885496271880765e-45 ], 
                    [ 784.5 , 1.8857669359663864e-45 ], [ 784.5625 , 1.8860376389060076e-45 ], [ 784.625 , 1.8863080431990392e-45 ], [ 784.6875 , 1.8865784402301894e-45 ], 
                    [ 784.75 , 1.886848876022032e-45 ], [ 784.8125 , 1.8871193505800955e-45 ], [ 784.875 , 1.8873898639099376e-45 ], [ 784.9375 , 1.8876604160171027e-45 ], 
                    [ 785.0 , 1.887931006907176e-45 ], [ 785.0625 , 1.8882016365857038e-45 ], [ 785.125 , 1.8884723050582322e-45 ], [ 785.1875 , 1.8887430123303633e-45 ], 
                    [ 785.25 , 1.8890137584076184e-45 ], [ 785.3125 , 1.8892845432955598e-45 ], [ 785.375 , 1.889555366999778e-45 ], [ 785.4375 , 1.8898262295258104e-45 ], 
                    [ 785.5 , 1.8900971308792355e-45 ], [ 785.5625 , 1.8903680710656325e-45 ], [ 785.625 , 1.8906390500905276e-45 ], [ 785.6875 , 1.8909100679595683e-45 ], 
                    [ 785.75 , 1.891181124678243e-45 ], [ 785.8125 , 1.8914522202521473e-45 ], [ 785.875 , 1.8917233546868507e-45 ], [ 785.9375 , 1.8919945279878975e-45 ], 
                    [ 786.0 , 1.892265740160926e-45 ], [ 786.0625 , 1.8925369912114408e-45 ], [ 786.125 , 1.8928082811450287e-45 ], [ 786.1875 , 1.893079609967317e-45 ], 
                    [ 786.25 , 1.893350977683827e-45 ], [ 786.3125 , 1.8936223843001602e-45 ], [ 786.375 , 1.8938938298218534e-45 ], [ 786.4375 , 1.8941649131258437e-45 ], 
                    [ 786.5 , 1.8944360077999417e-45 ], [ 786.5625 , 1.8947071412733796e-45 ], [ 786.625 , 1.8949783135516966e-45 ], [ 786.6875 , 1.895249524640433e-45 ], 
                    [ 786.75 , 1.895520774545171e-45 ], [ 786.8125 , 1.895792063271438e-45 ], [ 786.875 , 1.8960633908248313e-45 ], [ 786.9375 , 1.8963347572108398e-45 ], 
                    [ 787.0 , 1.896606162435076e-45 ], [ 787.0625 , 1.8968776065031118e-45 ], [ 787.125 , 1.897149089420452e-45 ], [ 787.1875 , 1.8974206111926845e-45 ], 
                    [ 787.25 , 1.8976921718253695e-45 ], [ 787.3125 , 1.8979637713240965e-45 ], [ 787.375 , 1.8982354096943734e-45 ], [ 787.4375 , 1.8985070869417907e-45 ], 
                    [ 787.5 , 1.8987788030719265e-45 ], [ 787.5625 , 1.899050558090332e-45 ], [ 787.625 , 1.8993223520025586e-45 ], [ 787.6875 , 1.899594184814201e-45 ], 
                    [ 787.75 , 1.8998660565307854e-45 ], [ 787.8125 , 1.9001379671579474e-45 ], [ 787.875 , 1.9004099167012025e-45 ], [ 787.9375 , 1.9006819051661197e-45 ], 
                    [ 788.0 , 1.900953932558283e-45 ], [ 788.0625 , 1.901225998883278e-45 ], [ 788.125 , 1.9014981041466627e-45 ], [ 788.1875 , 1.90177024835401e-45 ], 
                    [ 788.25 , 1.9020417332228617e-45 ], [ 788.3125 , 1.902313202661963e-45 ], [ 788.375 , 1.9025847108466162e-45 ], [ 788.4375 , 1.902856257782351e-45 ], 
                    [ 788.5 , 1.9031278434746983e-45 ], [ 788.5625 , 1.9033994679291898e-45 ], [ 788.625 , 1.903671131151371e-45 ], [ 788.6875 , 1.9039428331467615e-45 ], 
                    [ 788.75 , 1.9042145739208823e-45 ], [ 788.8125 , 1.9044863534792808e-45 ], [ 788.875 , 1.904758171827493e-45 ], [ 788.9375 , 1.9050300289710686e-45 ], 
                    [ 789.0 , 1.9053019249155177e-45 ], [ 789.0625 , 1.9055738596663913e-45 ], [ 789.125 , 1.9058458332292424e-45 ], [ 789.1875 , 1.9061178456095825e-45 ], 
                    [ 789.25 , 1.9063898968129527e-45 ], [ 789.3125 , 1.906661986844934e-45 ], [ 789.375 , 1.9069341157110274e-45 ], [ 789.4375 , 1.90720628341679e-45 ], 
                    [ 789.5 , 1.9074784899677642e-45 ], [ 789.5625 , 1.9077507353694948e-45 ], [ 789.625 , 1.9080230196275402e-45 ], [ 789.6875 , 1.9082953427474324e-45 ], 
                    [ 789.75 , 1.9085677047346914e-45 ], [ 789.8125 , 1.9088401055949183e-45 ], [ 789.875 , 1.9091125453336068e-45 ], [ 789.9375 , 1.9093850239563607e-45 ], 
                    [ 790.0 , 1.909657541468689e-45 ], [ 790.0625 , 1.9099297714700625e-45 ], [ 790.125 , 1.9102019780975798e-45 ], [ 790.1875 , 1.9104742235204752e-45 ], 
                    [ 790.25 , 1.910746507744265e-45 ], [ 790.3125 , 1.9110188307745057e-45 ], [ 790.375 , 1.911291192616674e-45 ], [ 790.4375 , 1.9115635932763825e-45 ], 
                    [ 790.5 , 1.911836032759137e-45 ], [ 790.5625 , 1.912108511070416e-45 ], [ 790.625 , 1.912381028215821e-45 ], [ 790.6875 , 1.9126535842008196e-45 ], 
                    [ 790.75 , 1.912926179031042e-45 ], [ 790.8125 , 1.9131988127119424e-45 ], [ 790.875 , 1.913471485249086e-45 ], [ 790.9375 , 1.91374419664801e-45 ], 
                    [ 791.0 , 1.914016946914253e-45 ], [ 791.0625 , 1.914289736053355e-45 ], [ 791.125 , 1.9145625640708833e-45 ], [ 791.1875 , 1.9148354309723102e-45 ], 
                    [ 791.25 , 1.9151083367632468e-45 ], [ 791.3125 , 1.9153812814491934e-45 ], [ 791.375 , 1.9156542650357083e-45 ], [ 791.4375 , 1.9159272875283485e-45 ], 
                    [ 791.5 , 1.9162003489326184e-45 ], [ 791.5625 , 1.916473449254118e-45 ], [ 791.625 , 1.916746588498367e-45 ], [ 791.6875 , 1.917019766670899e-45 ], 
                    [ 791.75 , 1.917292983777248e-45 ], [ 791.8125 , 1.917566239823032e-45 ], [ 791.875 , 1.9178391181325366e-45 ], [ 791.9375 , 1.9181118462868863e-45 ], 
                    [ 792.0 , 1.918384613224819e-45 ], [ 792.0625 , 1.9186574189518363e-45 ], [ 792.125 , 1.9189302634734403e-45 ], [ 792.1875 , 1.9192031467951754e-45 ], 
                    [ 792.25 , 1.9194760689225454e-45 ], [ 792.3125 , 1.9197490298610686e-45 ], [ 792.375 , 1.9200220296162507e-45 ], [ 792.4375 , 1.9202950681936252e-45 ], 
                    [ 792.5 , 1.9205681455987407e-45 ], [ 792.5625 , 1.9208412618370633e-45 ], [ 792.625 , 1.9211144169141427e-45 ], [ 792.6875 , 1.9213876108355026e-45 ], 
                    [ 792.75 , 1.92166084360668e-45 ], [ 792.8125 , 1.9219341152331728e-45 ], [ 792.875 , 1.922207425720547e-45 ], [ 792.9375 , 1.9224807750742876e-45 ], 
                    [ 793.0 , 1.922754163299936e-45 ], [ 793.0625 , 1.923027590403006e-45 ], [ 793.125 , 1.92330105638904e-45 ], [ 793.1875 , 1.9235745612635672e-45 ], 
                    [ 793.25 , 1.9238481050321586e-45 ], [ 793.3125 , 1.92412168770025e-45 ], [ 793.375 , 1.924395309273455e-45 ], [ 793.4375 , 1.9246689697572786e-45 ], 
                    [ 793.5 , 1.9249426691572404e-45 ], [ 793.5625 , 1.925216407478902e-45 ], [ 793.625 , 1.925490184727785e-45 ], [ 793.6875 , 1.925763815877114e-45 ], 
                    [ 793.75 , 1.9260372788974468e-45 ], [ 793.8125 , 1.9263107807501754e-45 ], [ 793.875 , 1.9265843214408283e-45 ], [ 793.9375 , 1.9268579009748795e-45 ], 
                    [ 794.0 , 1.9271315193578995e-45 ], [ 794.0625 , 1.9274051765953774e-45 ], [ 794.125 , 1.9276788726928037e-45 ], [ 794.1875 , 1.927952607655737e-45 ], 
                    [ 794.25 , 1.9282263814896834e-45 ], [ 794.3125 , 1.9285001942001758e-45 ], [ 794.375 , 1.9287740457927212e-45 ], [ 794.4375 , 1.9290479362728137e-45 ], 
                    [ 794.5 , 1.9293218656460298e-45 ], [ 794.5625 , 1.9295958339178383e-45 ], [ 794.625 , 1.929869841093831e-45 ], [ 794.6875 , 1.930143887179436e-45 ], 
                    [ 794.75 , 1.9304179721802893e-45 ], [ 794.8125 , 1.9306920961018336e-45 ], [ 794.875 , 1.9309662589496238e-45 ], [ 794.9375 , 1.9312404607291736e-45 ], 
                    [ 795.0 , 1.9315147014460527e-45 ], [ 795.0625 , 1.9317889811057628e-45 ], [ 795.125 , 1.93206329971382e-45 ], [ 795.1875 , 1.9323376572757965e-45 ], 
                    [ 795.25 , 1.932612053797127e-45 ], [ 795.3125 , 1.9328864892834687e-45 ], [ 795.375 , 1.9331609637402988e-45 ], [ 795.4375 , 1.9334354771731377e-45 ], 
                    [ 795.5 , 1.9337099457392238e-45 ], [ 795.5625 , 1.9339841003568445e-45 ], [ 795.625 , 1.9342582938431872e-45 ], [ 795.6875 , 1.9345325262036798e-45 ], 
                    [ 795.75 , 1.9348067974438888e-45 ], [ 795.8125 , 1.935081107569327e-45 ], [ 795.875 , 1.935355456585465e-45 ], [ 795.9375 , 1.935629844497859e-45 ], 
                    [ 796.0 , 1.935904271312037e-45 ], [ 796.0625 , 1.9361787370334317e-45 ], [ 796.125 , 1.936453241667655e-45 ], [ 796.1875 , 1.936727785220184e-45 ], 
                    [ 796.25 , 1.9370023676965207e-45 ], [ 796.3125 , 1.937276989102199e-45 ], [ 796.375 , 1.9375516494427236e-45 ], [ 796.4375 , 1.9378263487236835e-45 ], 
                    [ 796.5 , 1.9381010869504896e-45 ], [ 796.5625 , 1.938375864128732e-45 ], [ 796.625 , 1.9386506802639336e-45 ], [ 796.6875 , 1.9389255353615756e-45 ], 
                    [ 796.75 , 1.9392004294272374e-45 ], [ 796.8125 , 1.939475362466389e-45 ], [ 796.875 , 1.9397503344846107e-45 ], [ 796.9375 , 1.9400253454873873e-45 ], 
                    [ 797.0 , 1.9403003954802598e-45 ], [ 797.0625 , 1.940575484468756e-45 ], [ 797.125 , 1.9408506124584185e-45 ], [ 797.1875 , 1.9411257794547357e-45 ], 
                    [ 797.25 , 1.9414009854632784e-45 ], [ 797.3125 , 1.9416762304895647e-45 ], [ 797.375 , 1.9419508698085308e-45 ], [ 797.4375 , 1.9422254323924843e-45 ], 
                    [ 797.5 , 1.942500033795435e-45 ], [ 797.5625 , 1.942774674022872e-45 ], [ 797.625 , 1.9430493530802975e-45 ], [ 797.6875 , 1.9433240709732294e-45 ], 
                    [ 797.75 , 1.943598827707117e-45 ], [ 797.8125 , 1.9438736232874518e-45 ], [ 797.875 , 1.944148457719754e-45 ], [ 797.9375 , 1.9444233310094884e-45 ], 
                    [ 798.0 , 1.9446982431621773e-45 ], [ 798.0625 , 1.9449731941833007e-45 ], [ 798.125 , 1.9452481840783546e-45 ], [ 798.1875 , 1.9455232128528214e-45 ], 
                    [ 798.25 , 1.9457982805122253e-45 ], [ 798.3125 , 1.9460733870620642e-45 ], [ 798.375 , 1.9463485325077954e-45 ], [ 798.4375 , 1.946623716854959e-45 ], 
                    [ 798.5 , 1.9468989401090423e-45 ], [ 798.5625 , 1.9471742022755037e-45 ], [ 798.625 , 1.9474495033599422e-45 ], [ 798.6875 , 1.947724843367791e-45 ], 
                    [ 798.75 , 1.948000222304553e-45 ], [ 798.8125 , 1.9482756401757737e-45 ], [ 798.875 , 1.9485510969869027e-45 ], [ 798.9375 , 1.948826592743473e-45 ], 
                    [ 799.0 , 1.949102127451033e-45 ], [ 799.0625 , 1.949377701114992e-45 ], [ 799.125 , 1.9496533137409414e-45 ], [ 799.1875 , 1.9499287969887216e-45 ], 
                    [ 799.25 , 1.9502040904336254e-45 ], [ 799.3125 , 1.9504794227448244e-45 ], [ 799.375 , 1.950754793927764e-45 ], [ 799.4375 , 1.9510302039879872e-45 ], 
                    [ 799.5 , 1.951305652930942e-45 ], [ 799.5625 , 1.9515811407621313e-45 ], [ 799.625 , 1.951856667487059e-45 ], [ 799.6875 , 1.9521322331111754e-45 ], 
                    [ 799.75 , 1.952407837640013e-45 ], [ 799.8125 , 1.952683481079051e-45 ], [ 799.875 , 1.9529591634337827e-45 ], [ 799.9375 , 1.9532348847097027e-45 ], 
                    [ 800.0 , 1.9535106449123052e-45 ], [ 800.0625 , 1.9537864440470863e-45 ], [ 800.125 , 1.9540622821195565e-45 ], [ 800.1875 , 1.954338159135185e-45 ], 
                    [ 800.25 , 1.95461407509947e-45 ], [ 800.3125 , 1.9548900300179523e-45 ], [ 800.375 , 1.955166023896089e-45 ], [ 800.4375 , 1.955442056739423e-45 ], 
                    [ 800.5 , 1.955718128553372e-45 ], [ 800.5625 , 1.9559942393435487e-45 ], [ 800.625 , 1.9562703891153584e-45 ], [ 800.6875 , 1.9565465778743745e-45 ], 
                    [ 800.75 , 1.9568228056260456e-45 ], [ 800.8125 , 1.9570990723759185e-45 ], [ 800.875 , 1.9573753781294846e-45 ], [ 800.9375 , 1.957651722892265e-45 ], 
                    [ 801.0 , 1.957928106669725e-45 ], [ 801.0625 , 1.9582041161362838e-45 ], [ 801.125 , 1.9584801362460105e-45 ], [ 801.1875 , 1.9587561952623565e-45 ], 
                    [ 801.25 , 1.959032293190763e-45 ], [ 801.3125 , 1.9593084300367993e-45 ], [ 801.375 , 1.9595846058059085e-45 ], [ 801.4375 , 1.9598608205035642e-45 ], 
                    [ 801.5 , 1.9601370741352808e-45 ], [ 801.5625 , 1.9604133667065178e-45 ], [ 801.625 , 1.9606896982227794e-45 ], [ 801.6875 , 1.960966068689554e-45 ], 
                    [ 801.75 , 1.9612424781123318e-45 ], [ 801.8125 , 1.961518926496591e-45 ], [ 801.875 , 1.961795413847864e-45 ], [ 801.9375 , 1.9620719401716165e-45 ], 
                    [ 802.0 , 1.9623485054733136e-45 ], [ 802.0625 , 1.962625109758491e-45 ], [ 802.125 , 1.9629017530326023e-45 ], [ 802.1875 , 1.9631784353012408e-45 ], 
                    [ 802.25 , 1.963455156569805e-45 ], [ 802.3125 , 1.9637319168437928e-45 ], [ 802.375 , 1.9640087161287295e-45 ], [ 802.4375 , 1.9642855544301418e-45 ], 
                    [ 802.5 , 1.964562431753502e-45 ], [ 802.5625 , 1.964839348104296e-45 ], [ 802.625 , 1.9651163034880254e-45 ], [ 802.6875 , 1.9653932979102194e-45 ], 
                    [ 802.75 , 1.9656703313763673e-45 ], [ 802.8125 , 1.9659474038919722e-45 ], [ 802.875 , 1.9662242292346627e-45 ], [ 802.9375 , 1.9665006343353375e-45 ], 
                    [ 803.0 , 1.9667770782921133e-45 ], [ 803.0625 , 1.9670535611104243e-45 ], [ 803.125 , 1.9673300827957896e-45 ], [ 803.1875 , 1.967606643353589e-45 ], 
                    [ 803.25 , 1.967883242789329e-45 ], [ 803.3125 , 1.968159881108489e-45 ], [ 803.375 , 1.9684365583165346e-45 ], [ 803.4375 , 1.9687132744188916e-45 ], 
                    [ 803.5 , 1.9689900294210688e-45 ], [ 803.5625 , 1.9692668233285073e-45 ], [ 803.625 , 1.9695436561467175e-45 ], [ 803.6875 , 1.9698205278811142e-45 ], 
                    [ 803.75 , 1.9700974385371954e-45 ], [ 803.8125 , 1.970374388120475e-45 ], [ 803.875 , 1.970651376636327e-45 ], [ 803.9375 , 1.9709284040903084e-45 ], 
                    [ 804.0 , 1.9712054704878653e-45 ], [ 804.0625 , 1.9714825758344717e-45 ], [ 804.125 , 1.9717597201356037e-45 ], [ 804.1875 , 1.9720369033967368e-45 ], 
                    [ 804.25 , 1.9723141256233478e-45 ], [ 804.3125 , 1.9725913868209147e-45 ], [ 804.375 , 1.9728686869949018e-45 ], [ 804.4375 , 1.9731460261508165e-45 ], 
                    [ 804.5 , 1.973423404294138e-45 ], [ 804.5625 , 1.973700821430306e-45 ], [ 804.625 , 1.9739782775648433e-45 ], [ 804.6875 , 1.974255772703204e-45 ], 
                    [ 804.75 , 1.9745330829903702e-45 ], [ 804.8125 , 1.9748103645953834e-45 ], [ 804.875 , 1.975087685138733e-45 ], [ 804.9375 , 1.9753650446259574e-45 ], 
                    [ 805.0 , 1.9756424430624687e-45 ], [ 805.0625 , 1.9759198804537507e-45 ], [ 805.125 , 1.976197356805288e-45 ], [ 805.1875 , 1.9764748721225236e-45 ], 
                    [ 805.25 , 1.9767524264109154e-45 ], [ 805.3125 , 1.9770300196759922e-45 ], [ 805.375 , 1.9773076519232135e-45 ], [ 805.4375 , 1.9775853231580113e-45 ], 
                    [ 805.5 , 1.9778630333859027e-45 ], [ 805.5625 , 1.9781407826123633e-45 ], [ 805.625 , 1.978418570842828e-45 ], [ 805.6875 , 1.9786963980828018e-45 ], 
                    [ 805.75 , 1.9789742643377907e-45 ], [ 805.8125 , 1.9792521696132175e-45 ], [ 805.875 , 1.9795301139146046e-45 ], [ 805.9375 , 1.9798080972474312e-45 ], 
                    [ 806.0 , 1.9800861196171372e-45 ], [ 806.0625 , 1.9803641810292603e-45 ], [ 806.125 , 1.9806422814892554e-45 ], [ 806.1875 , 1.9809204210025772e-45 ], 
                    [ 806.25 , 1.9811985995747387e-45 ], [ 806.3125 , 1.9814768172112526e-45 ], [ 806.375 , 1.9817550739175628e-45 ], [ 806.4375 , 1.9820333696991697e-45 ], 
                    [ 806.5 , 1.9823117045615606e-45 ], [ 806.5625 , 1.982590027619128e-45 ], [ 806.625 , 1.9828677337324388e-45 ], [ 806.6875 , 1.983145478744721e-45 ], 
                    [ 806.75 , 1.9834232626614098e-45 ], [ 806.8125 , 1.983701085487954e-45 ], [ 806.875 , 1.9839789472298042e-45 ], [ 806.9375 , 1.9842568478923975e-45 ], 
                    [ 807.0 , 1.984534787481227e-45 ], [ 807.0625 , 1.9848127660016896e-45 ], [ 807.125 , 1.9850907834592668e-45 ], [ 807.1875 , 1.98536883985944e-45 ], 
                    [ 807.25 , 1.9856469352076083e-45 ], [ 807.3125 , 1.9859250695092696e-45 ], [ 807.375 , 1.9862032427698654e-45 ], [ 807.4375 , 1.9864814549948393e-45 ], 
                    [ 807.5 , 1.9867597061897052e-45 ], [ 807.5625 , 1.9870379963598232e-45 ], [ 807.625 , 1.9873163255107512e-45 ], [ 807.6875 , 1.9875946936478932e-45 ], 
                    [ 807.75 , 1.9878731007767103e-45 ], [ 807.8125 , 1.9881515469027057e-45 ], [ 807.875 , 1.9884300320312716e-45 ], [ 807.9375 , 1.988708556167942e-45 ], 
                    [ 808.0 , 1.9889871193181522e-45 ], [ 808.0625 , 1.989265721487353e-45 ], [ 808.125 , 1.9895443626810387e-45 ], [ 808.1875 , 1.9898230429046608e-45 ], 
                    [ 808.25 , 1.9901017621637148e-45 ], [ 808.3125 , 1.9903805204635832e-45 ], [ 808.375 , 1.9906593178098337e-45 ], [ 808.4375 , 1.9909380550535024e-45 ], 
                    [ 808.5 , 1.9912166126480308e-45 ], [ 808.5625 , 1.991495209216315e-45 ], [ 808.625 , 1.9917738447637935e-45 ], [ 808.6875 , 1.992052519295963e-45 ], 
                    [ 808.75 , 1.992331232818221e-45 ], [ 808.8125 , 1.9926099853360653e-45 ], [ 808.875 , 1.9928887768549512e-45 ], [ 808.9375 , 1.9931676073802942e-45 ], 
                    [ 809.0 , 1.9934464769175788e-45 ], [ 809.0625 , 1.9937253854722784e-45 ], [ 809.125 , 1.9940043330498517e-45 ], [ 809.1875 , 1.994283319655702e-45 ], 
                    [ 809.25 , 1.9945623452953745e-45 ], [ 809.3125 , 1.9948414099742743e-45 ], [ 809.375 , 1.995120513697877e-45 ], [ 809.4375 , 1.995399656471646e-45 ], 
                    [ 809.5 , 1.9956788383010024e-45 ], [ 809.5625 , 1.995958059191538e-45 ], [ 809.625 , 1.9962373191485763e-45 ], [ 809.6875 , 1.9965166181776397e-45 ], 
                    [ 809.75 , 1.996795956284209e-45 ], [ 809.8125 , 1.9970753334737378e-45 ], [ 809.875 , 1.9973547497516796e-45 ], [ 809.9375 , 1.9976342051235325e-45 ], 
                    [ 810.0 , 1.9979136995947656e-45 ], [ 810.0625 , 1.9981932331708355e-45 ], [ 810.125 , 1.9984728058571712e-45 ], [ 810.1875 , 1.998752417659329e-45 ], 
                    [ 810.25 , 1.9990320685827114e-45 ], [ 810.3125 , 1.999311402530861e-45 ], [ 810.375 , 1.999590401844291e-45 ], [ 810.4375 , 1.999869440091434e-45 ], 
                    [ 810.5 , 2.000148517277738e-45 ], [ 810.5625 , 2.000427633408608e-45 ], [ 810.625 , 2.0007067884895072e-45 ], [ 810.6875 , 2.000985982525842e-45 ], 
                    [ 810.75 , 2.001265215523078e-45 ], [ 810.8125 , 2.0015444874866367e-45 ], [ 810.875 , 2.0018237984219425e-45 ], [ 810.9375 , 2.0021031483344337e-45 ], 
                    [ 811.0 , 2.0023825372295915e-45 ], [ 811.0625 , 2.002661965112814e-45 ], [ 811.125 , 2.0029414319895697e-45 ], [ 811.1875 , 2.0032209378652717e-45 ], 
                    [ 811.25 , 2.003500482745391e-45 ], [ 811.3125 , 2.0037800666353416e-45 ], [ 811.375 , 2.004059689540568e-45 ], [ 811.4375 , 2.0043393514665853e-45 ], 
                    [ 811.5 , 2.0046190524187105e-45 ], [ 811.5625 , 2.0048987924025322e-45 ], [ 811.625 , 2.0051785714233695e-45 ], [ 811.6875 , 2.0054583894867407e-45 ], 
                    [ 811.75 , 2.005738246598108e-45 ], [ 811.8125 , 2.0060181427628353e-45 ], [ 811.875 , 2.0062980779864723e-45 ], [ 811.9375 , 2.006578052274441e-45 ], 
                    [ 812.0 , 2.0068580656321357e-45 ], [ 812.0625 , 2.0071381180650654e-45 ], [ 812.125 , 2.007418209578683e-45 ], [ 812.1875 , 2.007698088071726e-45 ], 
                    [ 812.25 , 2.0079778259781653e-45 ], [ 812.3125 , 2.00825760286123e-45 ], [ 812.375 , 2.0085374187263363e-45 ], [ 812.4375 , 2.0088172735789444e-45 ], 
                    [ 812.5 , 2.009097167424458e-45 ], [ 812.5625 , 2.0093771002683246e-45 ], [ 812.625 , 2.0096570721160065e-45 ], [ 812.6875 , 2.0099370829728805e-45 ], 
                    [ 812.75 , 2.010217132844411e-45 ], [ 812.8125 , 2.0104972217360337e-45 ], [ 812.875 , 2.0107773496531856e-45 ], [ 812.9375 , 2.0110575166013327e-45 ], 
                    [ 813.0 , 2.0113377225858703e-45 ], [ 813.0625 , 2.0116179676122236e-45 ], [ 813.125 , 2.011898251685889e-45 ], [ 813.1875 , 2.0121785748122647e-45 ], 
                    [ 813.25 , 2.0124589369968057e-45 ], [ 813.3125 , 2.0127393382449834e-45 ], [ 813.375 , 2.0130197785621827e-45 ], [ 813.4375 , 2.013300257953862e-45 ], 
                    [ 813.5 , 2.0135807764255226e-45 ], [ 813.5625 , 2.0138613339825527e-45 ], [ 813.625 , 2.0141419306303836e-45 ], [ 813.6875 , 2.0144225663745336e-45 ], 
                    [ 813.75 , 2.0147032412203787e-45 ], [ 813.8125 , 2.0149839551733805e-45 ], [ 813.875 , 2.0152647082390462e-45 ], [ 813.9375 , 2.015545500422768e-45 ], 
                    [ 814.0 , 2.0158263317299958e-45 ], [ 814.0625 , 2.016106727087667e-45 ], [ 814.125 , 2.0163868847810827e-45 ], [ 814.1875 , 2.0166670814051278e-45 ], 
                    [ 814.25 , 2.016947316965241e-45 ], [ 814.3125 , 2.0172275914668043e-45 ], [ 814.375 , 2.0175079049152573e-45 ], [ 814.4375 , 2.017788257315984e-45 ], 
                    [ 814.5 , 2.018068648674425e-45 ], [ 814.5625 , 2.018349078995966e-45 ], [ 814.625 , 2.0186295482860493e-45 ], [ 814.6875 , 2.018910056550062e-45 ], 
                    [ 814.75 , 2.0191906037934193e-45 ], [ 814.8125 , 2.0194711900215666e-45 ], [ 814.875 , 2.0197518152399215e-45 ], [ 814.9375 , 2.0200324794538872e-45 ], 
                    [ 815.0 , 2.0203131826688833e-45 ], [ 815.0625 , 2.0205939248902998e-45 ], [ 815.125 , 2.0208747061235863e-45 ], [ 815.1875 , 2.0211555263742072e-45 ], 
                    [ 815.25 , 2.0214363856475257e-45 ], [ 815.3125 , 2.0217172839489656e-45 ], [ 815.375 , 2.0219982212839496e-45 ], [ 815.4375 , 2.0222791976579455e-45 ], 
                    [ 815.5 , 2.022560213076334e-45 ], [ 815.5625 , 2.0228412675445705e-45 ], [ 815.625 , 2.0231223610680522e-45 ], [ 815.6875 , 2.0234034936522202e-45 ], 
                    [ 815.75 , 2.023684665302489e-45 ], [ 815.8125 , 2.0239658760243435e-45 ], [ 815.875 , 2.024247125823142e-45 ], [ 815.9375 , 2.0245282367792348e-45 ], 
                    [ 816.0 , 2.02480926001492e-45 ], [ 816.0625 , 2.0250903222592577e-45 ], [ 816.125 , 2.0253714235176188e-45 ], [ 816.1875 , 2.025652563795419e-45 ], 
                    [ 816.25 , 2.025933743098117e-45 ], [ 816.3125 , 2.0262149614310594e-45 ], [ 816.375 , 2.0264962187997487e-45 ], [ 816.4375 , 2.0267775152095465e-45 ], 
                    [ 816.5 , 2.0270588506659013e-45 ], [ 816.5625 , 2.0273402251741895e-45 ], [ 816.625 , 2.0276216387398892e-45 ], [ 816.6875 , 2.0279030913683507e-45 ], 
                    [ 816.75 , 2.028184583065082e-45 ], [ 816.8125 , 2.0284661138354488e-45 ], [ 816.875 , 2.0287476836848892e-45 ], [ 816.9375 , 2.029029292618813e-45 ], 
                    [ 817.0 , 2.0293109406426753e-45 ], [ 817.0625 , 2.0295926277618867e-45 ], [ 817.125 , 2.029874353981889e-45 ], [ 817.1875 , 2.0301561193080807e-45 ], 
                    [ 817.25 , 2.030437923745904e-45 ], [ 817.3125 , 2.0307197673007895e-45 ], [ 817.375 , 2.0310016499781658e-45 ], [ 817.4375 , 2.0312835717834628e-45 ], 
                    [ 817.5 , 2.0315655327221136e-45 ], [ 817.5625 , 2.031847532799549e-45 ], [ 817.625 , 2.032129572021202e-45 ], [ 817.6875 , 2.032411650392507e-45 ], 
                    [ 817.75 , 2.0326937679188972e-45 ], [ 817.8125 , 2.032975628900001e-45 ], [ 817.875 , 2.033257218673361e-45 ], [ 817.9375 , 2.0335388474500273e-45 ], 
                    [ 818.0 , 2.0338205152354307e-45 ], [ 818.0625 , 2.034102222034961e-45 ], [ 818.125 , 2.0343839678540067e-45 ], [ 818.1875 , 2.0346657526979723e-45 ], 
                    [ 818.25 , 2.0349475765723074e-45 ], [ 818.3125 , 2.0352294394823743e-45 ], [ 818.375 , 2.0355113414335947e-45 ], [ 818.4375 , 2.03579328243139e-45 ], 
                    [ 818.5 , 2.0360752624811544e-45 ], [ 818.5625 , 2.036357281588254e-45 ], [ 818.625 , 2.0366393397581993e-45 ], [ 818.6875 , 2.0369214369963153e-45 ], 
                    [ 818.75 , 2.0372035733080557e-45 ], [ 818.8125 , 2.0374857486988036e-45 ], [ 818.875 , 2.037767963173987e-45 ], [ 818.9375 , 2.0380502167390194e-45 ], 
                    [ 819.0 , 2.038332509399315e-45 ], [ 819.0625 , 2.0386148411603034e-45 ], [ 819.125 , 2.038897212027386e-45 ], [ 819.1875 , 2.0391796220059504e-45 ], 
                    [ 819.25 , 2.039462071101458e-45 ], [ 819.3125 , 2.0397445593193262e-45 ], [ 819.375 , 2.040027086664945e-45 ], [ 819.4375 , 2.040309653143749e-45 ], 
                    [ 819.5 , 2.0405922587611875e-45 ], [ 819.5625 , 2.0408749035226088e-45 ], [ 819.625 , 2.0411575874335215e-45 ], [ 819.6875 , 2.0414401270873225e-45 ], 
                    [ 819.75 , 2.041722301347815e-45 ], [ 819.8125 , 2.0420045146113335e-45 ], [ 819.875 , 2.0422867668832695e-45 ], [ 819.9375 , 2.0425690581689713e-45 ], 
                    [ 820.0 , 2.042851388473875e-45 ], [ 820.0625 , 2.043133757803345e-45 ], [ 820.125 , 2.0434161661628188e-45 ], [ 820.1875 , 2.0436986135576475e-45 ], 
                    [ 820.25 , 2.0439810999932414e-45 ], [ 820.3125 , 2.0442636254750115e-45 ], [ 820.375 , 2.044546190008326e-45 ], [ 820.4375 , 2.0448287935985824e-45 ], 
                    [ 820.5 , 2.0451114362512084e-45 ], [ 820.5625 , 2.0453941179716032e-45 ], [ 820.625 , 2.0456768387651233e-45 ], [ 820.6875 , 2.0459595986372135e-45 ], 
                    [ 820.75 , 2.0462423975932747e-45 ], [ 820.8125 , 2.0465252356386805e-45 ], [ 820.875 , 2.0468081127788344e-45 ], [ 820.9375 , 2.0470910290191685e-45 ], 
                    [ 821.0 , 2.0473739843651023e-45 ], [ 821.0625 , 2.047656978821983e-45 ], [ 821.125 , 2.0479400123952308e-45 ], [ 821.1875 , 2.0482230850902673e-45 ], 
                    [ 821.25 , 2.0485061969125294e-45 ], [ 821.3125 , 2.0487893478673814e-45 ], [ 821.375 , 2.0490725379602467e-45 ], [ 821.4375 , 2.0493557671965353e-45 ], 
                    [ 821.5 , 2.0496390355816434e-45 ], [ 821.5625 , 2.0499223100508214e-45 ], [ 821.625 , 2.050205197373503e-45 ], [ 821.6875 , 2.0504881237343946e-45 ], 
                    [ 821.75 , 2.0507710891388537e-45 ], [ 821.8125 , 2.051054093592269e-45 ], [ 821.875 , 2.051337137100014e-45 ], [ 821.9375 , 2.0516202196675075e-45 ], 
                    [ 822.0 , 2.0519033413001105e-45 ], [ 822.0625 , 2.0521865020032435e-45 ], [ 822.125 , 2.052469701782254e-45 ], [ 822.1875 , 2.0527529406425933e-45 ], 
                    [ 822.25 , 2.0530362185896245e-45 ], [ 822.3125 , 2.0533195356287278e-45 ], [ 822.375 , 2.053602891765312e-45 ], [ 822.4375 , 2.0538862870047876e-45 ], 
                    [ 822.5 , 2.0541697213525063e-45 ], [ 822.5625 , 2.054453194813924e-45 ], [ 822.625 , 2.0547367073944085e-45 ], [ 822.6875 , 2.055020259099359e-45 ], 
                    [ 822.75 , 2.0553038499341446e-45 ], [ 822.8125 , 2.0555874799042096e-45 ], [ 822.875 , 2.0558711490149544e-45 ], [ 822.9375 , 2.056154857271766e-45 ], 
                    [ 823.0 , 2.0564386046800314e-45 ], [ 823.0625 , 2.0567223912451975e-45 ], [ 823.125 , 2.0570062169725953e-45 ], [ 823.1875 , 2.0572900818677017e-45 ], 
                    [ 823.25 , 2.057573985935893e-45 ], [ 823.3125 , 2.05785792918256e-45 ], [ 823.375 , 2.0581419116131386e-45 ], [ 823.4375 , 2.0584259332329926e-45 ], 
                    [ 823.5 , 2.0587096564236398e-45 ], [ 823.5625 , 2.0589933167166884e-45 ], [ 823.625 , 2.059277016094006e-45 ], [ 823.6875 , 2.059560754560963e-45 ], 
                    [ 823.75 , 2.0598445321229745e-45 ], [ 823.8125 , 2.0601283487854136e-45 ], [ 823.875 , 2.0604122045536665e-45 ], [ 823.9375 , 2.0606960994331078e-45 ], 
                    [ 824.0 , 2.060980033429155e-45 ], [ 824.0625 , 2.0612640065471542e-45 ], [ 824.125 , 2.061548018792554e-45 ], [ 824.1875 , 2.061832070170717e-45 ], 
                    [ 824.25 , 2.0621161606870348e-45 ], [ 824.3125 , 2.0624002903468853e-45 ], [ 824.375 , 2.0626844591556915e-45 ], [ 824.4375 , 2.062968667118832e-45 ], 
                    [ 824.5 , 2.0632529142416887e-45 ], [ 824.5625 , 2.063537200529685e-45 ], [ 824.625 , 2.0638215259882037e-45 ], [ 824.6875 , 2.064105890622597e-45 ], 
                    [ 824.75 , 2.0643902944383956e-45 ], [ 824.8125 , 2.0646747374408215e-45 ], [ 824.875 , 2.0649592196354212e-45 ], [ 824.9375 , 2.0652437410275066e-45 ], 
                    [ 825.0 , 2.065528301622522e-45 ], [ 825.0625 , 2.06581290142587e-45 ], [ 825.125 , 2.0660975404428793e-45 ], [ 825.1875 , 2.066382218679055e-45 ], 
                    [ 825.25 , 2.0666669361397578e-45 ], [ 825.3125 , 2.0669516928304064e-45 ], [ 825.375 , 2.0672361377564183e-45 ], [ 825.4375 , 2.0675200586867765e-45 ], 
                    [ 825.5 , 2.0678040186117732e-45 ], [ 825.5625 , 2.0680880175367485e-45 ], [ 825.625 , 2.068372055467045e-45 ], [ 825.6875 , 2.0686561324080337e-45 ], 
                    [ 825.75 , 2.0689402483650876e-45 ], [ 825.8125 , 2.0692244033435358e-45 ], [ 825.875 , 2.0695085973487522e-45 ], [ 825.9375 , 2.0697928303860972e-45 ], 
                    [ 826.0 , 2.0700771024609314e-45 ], [ 826.0625 , 2.0703614135786454e-45 ], [ 826.125 , 2.0706457637445585e-45 ], [ 826.1875 , 2.0709301529640468e-45 ], 
                    [ 826.25 , 2.0712145812424605e-45 ], [ 826.3125 , 2.0714990485851786e-45 ], [ 826.375 , 2.071783554997566e-45 ], [ 826.4375 , 2.0720681004849886e-45 ], 
                    [ 826.5 , 2.072352685052843e-45 ], [ 826.5625 , 2.0726373087064523e-45 ], [ 826.625 , 2.0729219714511696e-45 ], [ 826.6875 , 2.0732066732924232e-45 ], 
                    [ 826.75 , 2.073491414235553e-45 ], [ 826.8125 , 2.0737761942859146e-45 ], [ 826.875 , 2.074061013448909e-45 ], [ 826.9375 , 2.074345871729878e-45 ], 
                    [ 827.0 , 2.074630769134209e-45 ], [ 827.0625 , 2.074915705667276e-45 ], [ 827.125 , 2.075200681334467e-45 ], [ 827.1875 , 2.0754856961411276e-45 ], 
                    [ 827.25 , 2.075770750092678e-45 ], [ 827.3125 , 2.0760557114031798e-45 ], [ 827.375 , 2.0763407003714497e-45 ], [ 827.4375 , 2.0766257284613635e-45 ], 
                    [ 827.5 , 2.0769107956782914e-45 ], [ 827.5625 , 2.077195902027575e-45 ], [ 827.625 , 2.0774810475146456e-45 ], [ 827.6875 , 2.0777662321448603e-45 ], 
                    [ 827.75 , 2.0780514559235638e-45 ], [ 827.8125 , 2.0783367188561446e-45 ], [ 827.875 , 2.0786220209479623e-45 ], [ 827.9375 , 2.078907362204408e-45 ], 
                    [ 828.0 , 2.0791927426308866e-45 ], [ 828.0625 , 2.079478162232717e-45 ], [ 828.125 , 2.079763621015306e-45 ], [ 828.1875 , 2.080049118984047e-45 ], 
                    [ 828.25 , 2.0803346561442745e-45 ], [ 828.3125 , 2.080620232501443e-45 ], [ 828.375 , 2.080905848060874e-45 ], [ 828.4375 , 2.0811915028279635e-45 ], 
                    [ 828.5 , 2.081477196808079e-45 ], [ 828.5625 , 2.081762930006648e-45 ], [ 828.625 , 2.082048702428995e-45 ], [ 828.6875 , 2.0823345140805336e-45 ], 
                    [ 828.75 , 2.082620364966649e-45 ], [ 828.8125 , 2.082906255092742e-45 ], [ 828.875 , 2.0831921844641852e-45 ], [ 828.9375 , 2.08347815308635e-45 ], 
                    [ 829.0 , 2.08376416096467e-45 ], [ 829.0625 , 2.0840502081044736e-45 ], [ 829.125 , 2.0843362945111813e-45 ], [ 829.1875 , 2.0846221685728922e-45 ], 
                    [ 829.25 , 2.0849077167014257e-45 ], [ 829.3125 , 2.0851933039438742e-45 ], [ 829.375 , 2.085478930305595e-45 ], [ 829.4375 , 2.0857645957919472e-45 ], 
                    [ 829.5 , 2.0860503004082746e-45 ], [ 829.5625 , 2.0863360441599665e-45 ], [ 829.625 , 2.086621827052355e-45 ], [ 829.6875 , 2.0869076490908303e-45 ], 
                    [ 829.75 , 2.08719351028074e-45 ], [ 829.8125 , 2.087479410627447e-45 ], [ 829.875 , 2.0877653501363302e-45 ], [ 829.9375 , 2.088051328812694e-45 ], 
                    [ 830.0 , 2.0883373466619927e-45 ], [ 830.0625 , 2.0886234036895332e-45 ], [ 830.125 , 2.088909499900681e-45 ], [ 830.1875 , 2.0891956353008347e-45 ], 
                    [ 830.25 , 2.0894818098953462e-45 ], [ 830.3125 , 2.0897680236895844e-45 ], [ 830.375 , 2.0900542766889196e-45 ], [ 830.4375 , 2.0903405688987214e-45 ], 
                    [ 830.5 , 2.090626900324361e-45 ], [ 830.5625 , 2.09091327097121e-45 ], [ 830.625 , 2.091199680844611e-45 ], [ 830.6875 , 2.0914861299500268e-45 ], 
                    [ 830.75 , 2.091772618292727e-45 ], [ 830.8125 , 2.092059145878116e-45 ], [ 830.875 , 2.0923457127116285e-45 ], [ 830.9375 , 2.0926323187985968e-45 ], 
                    [ 831.0 , 2.0929189641443524e-45 ], [ 831.0625 , 2.0932056487543624e-45 ], [ 831.125 , 2.0934916229698505e-45 ], [ 831.1875 , 2.0937774671902515e-45 ], 
                    [ 831.25 , 2.0940633504396987e-45 ], [ 831.3125 , 2.094349272723431e-45 ], [ 831.375 , 2.0946352340468833e-45 ], [ 831.4375 , 2.094921234415326e-45 ], 
                    [ 831.5 , 2.09520727383412e-45 ], [ 831.5625 , 2.0954933523085673e-45 ], [ 831.625 , 2.095779469844001e-45 ], [ 831.6875 , 2.0960656264457993e-45 ], 
                    [ 831.75 , 2.0963518221192657e-45 ], [ 831.8125 , 2.0966380568697203e-45 ], [ 831.875 , 2.0969243307024994e-45 ], [ 831.9375 , 2.097210643622983e-45 ], 
                    [ 832.0 , 2.097496995636479e-45 ], [ 832.0625 , 2.09778338674831e-45 ], [ 832.125 , 2.0980698169638146e-45 ], [ 832.1875 , 2.0983562862883767e-45 ], 
                    [ 832.25 , 2.0986427947272916e-45 ], [ 832.3125 , 2.0989293422859144e-45 ], [ 832.375 , 2.0992159289695866e-45 ], [ 832.4375 , 2.099502554783636e-45 ], 
                    [ 832.5 , 2.0997892197334492e-45 ], [ 832.5625 , 2.1000759238243558e-45 ], [ 832.625 , 2.1003626670616397e-45 ], [ 832.6875 , 2.1006494494507062e-45 ], 
                    [ 832.75 , 2.1009362709968857e-45 ], [ 832.8125 , 2.1012231317055253e-45 ], [ 832.875 , 2.1015100315819863e-45 ], [ 832.9375 , 2.1017969706315874e-45 ], 
                    [ 833.0 , 2.1020839060157712e-45 ], [ 833.0625 , 2.102370625854069e-45 ], [ 833.125 , 2.1026573848003064e-45 ], [ 833.1875 , 2.102944182859893e-45 ], 
                    [ 833.25 , 2.1032310200381488e-45 ], [ 833.3125 , 2.1035178963403643e-45 ], [ 833.375 , 2.103804811771906e-45 ], [ 833.4375 , 2.1040917663381412e-45 ], 
                    [ 833.5 , 2.104378760044333e-45 ], [ 833.5625 , 2.1046657928958943e-45 ], [ 833.625 , 2.1049528648981502e-45 ], [ 833.6875 , 2.105239976056395e-45 ], 
                    [ 833.75 , 2.1055271263760154e-45 ], [ 833.8125 , 2.1058143158623072e-45 ], [ 833.875 , 2.106101544520703e-45 ], [ 833.9375 , 2.1063888123564407e-45 ], 
                    [ 834.0 , 2.106676119374939e-45 ], [ 834.0625 , 2.106963465581467e-45 ], [ 834.125 , 2.1072508509814608e-45 ], [ 834.1875 , 2.107538275580175e-45 ], 
                    [ 834.25 , 2.1078257393830326e-45 ], [ 834.3125 , 2.1081132423953653e-45 ], [ 834.375 , 2.108400784622476e-45 ], [ 834.4375 , 2.1086883660697585e-45 ], 
                    [ 834.5 , 2.108975986742533e-45 ], [ 834.5625 , 2.109263646646149e-45 ], [ 834.625 , 2.109551345786048e-45 ], [ 834.6875 , 2.1098390841674315e-45 ], 
                    [ 834.75 , 2.110126861795742e-45 ], [ 834.8125 , 2.110414678676348e-45 ], [ 834.875 , 2.1107025348145433e-45 ], [ 834.9375 , 2.1109902378363976e-45 ], 
                    [ 835.0 , 2.1112777008974657e-45 ], [ 835.0625 , 2.1115652031036457e-45 ], [ 835.125 , 2.1118527444603273e-45 ], [ 835.1875 , 2.1121403249727973e-45 ], 
                    [ 835.25 , 2.112427944646403e-45 ], [ 835.3125 , 2.1127156034864762e-45 ], [ 835.375 , 2.1130033014983667e-45 ], [ 835.4375 , 2.1132910386873924e-45 ], 
                    [ 835.5 , 2.113578815058874e-45 ], [ 835.5625 , 2.113866630618162e-45 ], [ 835.625 , 2.1141544853705925e-45 ], [ 835.6875 , 2.1144423793215182e-45 ], 
                    [ 835.75 , 2.114730312476231e-45 ], [ 835.8125 , 2.1150182848401154e-45 ], [ 835.875 , 2.11530629641851e-45 ], [ 835.9375 , 2.1155943472167258e-45 ], 
                    [ 836.0 , 2.1158824372401473e-45 ], [ 836.0625 , 2.1161705664940567e-45 ], [ 836.125 , 2.1164587349838413e-45 ], [ 836.1875 , 2.1167469427147984e-45 ], 
                    [ 836.25 , 2.117035189692302e-45 ], [ 836.3125 , 2.1173234759216966e-45 ], [ 836.375 , 2.1176118014083726e-45 ], [ 836.4375 , 2.117900166157615e-45 ], 
                    [ 836.5 , 2.1181885701747856e-45 ], [ 836.5625 , 2.1184770134651875e-45 ], [ 836.625 , 2.1187654960342577e-45 ], [ 836.6875 , 2.1190540178873005e-45 ], 
                    [ 836.75 , 2.1193425790296654e-45 ], [ 836.8125 , 2.119631179466687e-45 ], [ 836.875 , 2.1199194404891325e-45 ], [ 836.9375 , 2.1202074965094774e-45 ], 
                    [ 837.0 , 2.120495591671045e-45 ], [ 837.0625 , 2.1207837259791692e-45 ], [ 837.125 , 2.1210718994391837e-45 ], [ 837.1875 , 2.121360112056379e-45 ], 
                    [ 837.25 , 2.1216483638361052e-45 ], [ 837.3125 , 2.1219366547836542e-45 ], [ 837.375 , 2.1222249849043632e-45 ], [ 837.4375 , 2.1225133542035695e-45 ], 
                    [ 837.5 , 2.1228017626865675e-45 ], [ 837.5625 , 2.123090210358696e-45 ], [ 837.625 , 2.1233786972253098e-45 ], [ 837.6875 , 2.1236672232916603e-45 ], 
                    [ 837.75 , 2.123955788563134e-45 ], [ 837.8125 , 2.1242443930450575e-45 ], [ 837.875 , 2.1245330367427443e-45 ], [ 837.9375 , 2.1248217196615076e-45 ], 
                    [ 838.0 , 2.125110441806692e-45 ], [ 838.0625 , 2.125399203183643e-45 ], [ 838.125 , 2.1256880037976455e-45 ], [ 838.1875 , 2.125976843654107e-45 ], 
                    [ 838.25 , 2.1262657227582997e-45 ], [ 838.3125 , 2.1265546411155407e-45 ], [ 838.375 , 2.1268435987312397e-45 ], [ 838.4375 , 2.1271325956106708e-45 ], 
                    [ 838.5 , 2.1274216317591237e-45 ], [ 838.5625 , 2.1277107071820556e-45 ], [ 838.625 , 2.127999821884742e-45 ], [ 838.6875 , 2.1282889758725056e-45 ], 
                    [ 838.75 , 2.1285781691507146e-45 ], [ 838.8125 , 2.128867085683764e-45 ], [ 838.875 , 2.129155923334841e-45 ], [ 838.9375 , 2.1294448001744556e-45 ], 
                    [ 839.0 , 2.1297337162079104e-45 ], [ 839.0625 , 2.1300226714405682e-45 ], [ 839.125 , 2.130311665877717e-45 ], [ 839.1875 , 2.1306006995246608e-45 ], 
                    [ 839.25 , 2.1308897723867348e-45 ], [ 839.3125 , 2.1311788844692745e-45 ], [ 839.375 , 2.131468035777541e-45 ], [ 839.4375 , 2.131757226316946e-45 ], 
                    [ 839.5 , 2.1320464560927684e-45 ], [ 839.5625 , 2.1323357251103163e-45 ], [ 839.625 , 2.1326250333749282e-45 ], [ 839.6875 , 2.132914380891929e-45 ], 
                    [ 839.75 , 2.1332037676666145e-45 ], [ 839.8125 , 2.1334931937044173e-45 ], [ 839.875 , 2.1337826590105278e-45 ], [ 839.9375 , 2.1340721635903497e-45 ], 
                    [ 840.0 , 2.1343617074492416e-45 ], [ 840.0625 , 2.1346512905924116e-45 ], [ 840.125 , 2.134940913025296e-45 ], [ 840.1875 , 2.1352305747532252e-45 ], 
                    [ 840.25 , 2.1355202757814395e-45 ], [ 840.3125 , 2.1358100161153627e-45 ], [ 840.375 , 2.136099795760312e-45 ], [ 840.4375 , 2.136389614721591e-45 ], 
                    [ 840.5 , 2.1366794730045484e-45 ], [ 840.5625 , 2.13696937061452e-45 ], [ 840.625 , 2.137259307556841e-45 ], [ 840.6875 , 2.137549283836864e-45 ], 
                    [ 840.75 , 2.137838674351273e-45 ], [ 840.8125 , 2.138127931797479e-45 ], [ 840.875 , 2.1384172283813135e-45 ], [ 840.9375 , 2.1387065641080107e-45 ], 
                    [ 841.0 , 2.138995938982928e-45 ], [ 841.0625 , 2.1392853530113162e-45 ], [ 841.125 , 2.139574806198473e-45 ], [ 841.1875 , 2.139864298549743e-45 ], 
                    [ 841.25 , 2.1401538300703785e-45 ], [ 841.3125 , 2.1404434007656803e-45 ], [ 841.375 , 2.140733010640994e-45 ], [ 841.4375 , 2.1410226597015905e-45 ], 
                    [ 841.5 , 2.1413123479527565e-45 ], [ 841.5625 , 2.1416020753997942e-45 ], [ 841.625 , 2.1418918420480534e-45 ], [ 841.6875 , 2.142181647902792e-45 ], 
                    [ 841.75 , 2.14247149296933e-45 ], [ 841.8125 , 2.1427613772529877e-45 ], [ 841.875 , 2.143051300759027e-45 ], [ 841.9375 , 2.1433412634928145e-45 ], 
                    [ 842.0 , 2.1436312654595975e-45 ], [ 842.0625 , 2.143921306664745e-45 ], [ 842.125 , 2.144211387113536e-45 ], [ 842.1875 , 2.1445015068112648e-45 ], 
                    [ 842.25 , 2.1447916657632264e-45 ], [ 842.3125 , 2.1450818639748087e-45 ], [ 842.375 , 2.145372101451293e-45 ], [ 842.4375 , 2.1456623781979464e-45 ], 
                    [ 842.5 , 2.1459526942200826e-45 ], [ 842.5625 , 2.1462430495231062e-45 ], [ 842.625 , 2.146533444112226e-45 ], [ 842.6875 , 2.146823644483845e-45 ], 
                    [ 842.75 , 2.147113810339416e-45 ], [ 842.8125 , 2.1474040154139927e-45 ], [ 842.875 , 2.1476942597128154e-45 ], [ 842.9375 , 2.1479845432412464e-45 ], 
                    [ 843.0 , 2.1482748660045422e-45 ], [ 843.0625 , 2.148565228008006e-45 ], [ 843.125 , 2.1488556292569876e-45 ], [ 843.1875 , 2.1491460697567296e-45 ], 
                    [ 843.25 , 2.149436549512583e-45 ], [ 843.3125 , 2.149727068529839e-45 ], [ 843.375 , 2.150017626813804e-45 ], [ 843.4375 , 2.1503082243697854e-45 ], 
                    [ 843.5 , 2.150598861203106e-45 ], [ 843.5625 , 2.1508895373190297e-45 ], [ 843.625 , 2.151180252722911e-45 ], [ 843.6875 , 2.1514710074200293e-45 ], 
                    [ 843.75 , 2.1517618014157418e-45 ], [ 843.8125 , 2.1520526347153144e-45 ], [ 843.875 , 2.1523435073240586e-45 ], [ 843.9375 , 2.1526344192473503e-45 ], 
                    [ 844.0 , 2.1529253704903795e-45 ], [ 844.0625 , 2.153216361058598e-45 ], [ 844.125 , 2.153507390957261e-45 ], [ 844.1875 , 2.153798460191654e-45 ], 
                    [ 844.25 , 2.1540895687671688e-45 ], [ 844.3125 , 2.1543807166890318e-45 ], [ 844.375 , 2.154671903962622e-45 ], [ 844.4375 , 2.1549631305932278e-45 ], 
                    [ 844.5 , 2.155254396586199e-45 ], [ 844.5625 , 2.1555457019468416e-45 ], [ 844.625 , 2.1558365217137745e-45 ], [ 844.6875 , 2.1561271244489604e-45 ], 
                    [ 844.75 , 2.1564177663568676e-45 ], [ 844.8125 , 2.156708447442745e-45 ], [ 844.875 , 2.15699916771189e-45 ], [ 844.9375 , 2.157289927169584e-45 ], 
                    [ 845.0 , 2.157580725821125e-45 ], [ 845.0625 , 2.1578715636717505e-45 ], [ 845.125 , 2.1581624407268053e-45 ], [ 845.1875 , 2.158453356991513e-45 ], 
                    [ 845.25 , 2.1587443124712204e-45 ], [ 845.3125 , 2.159035307171183e-45 ], [ 845.375 , 2.1593263410966873e-45 ], [ 845.4375 , 2.1596174142530053e-45 ], 
                    [ 845.5 , 2.1599085266454257e-45 ], [ 845.5625 , 2.1601996782792684e-45 ], [ 845.625 , 2.1604908691598224e-45 ], [ 845.6875 , 2.1607820992923632e-45 ], 
                    [ 845.75 , 2.1610733686821818e-45 ], [ 845.8125 , 2.1613646773345545e-45 ], [ 845.875 , 2.1616560252547892e-45 ], [ 845.9375 , 2.1619474124481794e-45 ], 
                    [ 846.0 , 2.162238838920019e-45 ], [ 846.0625 , 2.1625303046755866e-45 ], [ 846.125 , 2.1628218097202245e-45 ], [ 846.1875 , 2.163113354059182e-45 ], 
                    [ 846.25 , 2.1634049376977876e-45 ], [ 846.3125 , 2.1636965606413075e-45 ], [ 846.375 , 2.163988222895055e-45 ], [ 846.4375 , 2.164279924464314e-45 ], 
                    [ 846.5 , 2.1645716653544307e-45 ], [ 846.5625 , 2.1648632738713027e-45 ], [ 846.625 , 2.1651547631603185e-45 ], [ 846.6875 , 2.1654462916970506e-45 ], 
                    [ 846.75 , 2.165737859486829e-45 ], [ 846.8125 , 2.166029466534924e-45 ], [ 846.875 , 2.166321112846591e-45 ], [ 846.9375 , 2.1666127984271625e-45 ], 
                    [ 847.0 , 2.166904523281926e-45 ], [ 847.0625 , 2.1671962874161387e-45 ], [ 847.125 , 2.1674880908351045e-45 ], [ 847.1875 , 2.1677799335441448e-45 ], 
                    [ 847.25 , 2.1680718155484873e-45 ], [ 847.3125 , 2.1683637368534386e-45 ], [ 847.375 , 2.1686556974643363e-45 ], [ 847.4375 , 2.1689476973864424e-45 ], 
                    [ 847.5 , 2.169239736625034e-45 ], [ 847.5625 , 2.1695318151854057e-45 ], [ 847.625 , 2.169823933072913e-45 ], [ 847.6875 , 2.1701160902927746e-45 ], 
                    [ 847.75 , 2.1704082868503164e-45 ], [ 847.8125 , 2.1707005227508362e-45 ], [ 847.875 , 2.1709927979996303e-45 ], [ 847.9375 , 2.1712851126019973e-45 ], 
                    [ 848.0 , 2.171577466563236e-45 ], [ 848.0625 , 2.1718698598886765e-45 ], [ 848.125 , 2.1721622925835572e-45 ], [ 848.1875 , 2.1724547646532257e-45 ], 
                    [ 848.25 , 2.1727472761029377e-45 ], [ 848.3125 , 2.1730398269380872e-45 ], [ 848.375 , 2.1733324171638543e-45 ], [ 848.4375 , 2.1736250467856356e-45 ], 
                    [ 848.5 , 2.1739174959549514e-45 ], [ 848.5625 , 2.1742094408713292e-45 ], [ 848.625 , 2.1745014249943013e-45 ], [ 848.6875 , 2.1747934483291023e-45 ], 
                    [ 848.75 , 2.1750855108809665e-45 ], [ 848.8125 , 2.1753776126552075e-45 ], [ 848.875 , 2.1756697536570618e-45 ], [ 848.9375 , 2.175961933891844e-45 ], 
                    [ 849.0 , 2.1762541533647755e-45 ], [ 849.0625 , 2.1765464120811263e-45 ], [ 849.125 , 2.1768387100461827e-45 ], [ 849.1875 , 2.177131047265261e-45 ], 
                    [ 849.25 , 2.1774234237435404e-45 ], [ 849.3125 , 2.1777158394863553e-45 ], [ 849.375 , 2.1780082944989782e-45 ], [ 849.4375 , 2.178300788786652e-45 ], 
                    [ 849.5 , 2.1785933223546517e-45 ], [ 849.5625 , 2.1788858952082824e-45 ], [ 849.625 , 2.1791785073528052e-45 ], [ 849.6875 , 2.1794711587934965e-45 ], 
                    [ 849.75 , 2.179763849535649e-45 ], [ 849.8125 , 2.1800565795845095e-45 ], [ 849.875 , 2.1803493489453572e-45 ], [ 849.9375 , 2.1806421576235172e-45 ], 
                    [ 850.0 , 2.180935005624208e-45 ], [ 850.0625 , 2.1812278929527567e-45 ], [ 850.125 , 2.1815208196144455e-45 ], [ 850.1875 , 2.1818137856145093e-45 ], 
                    [ 850.25 , 2.182106790958278e-45 ], [ 850.3125 , 2.182399835651035e-45 ], [ 850.375 , 2.182692919698018e-45 ], [ 850.4375 , 2.182986043104559e-45 ], 
                    [ 850.5 , 2.1832788985797037e-45 ], [ 850.5625 , 2.1835717857849488e-45 ], [ 850.625 , 2.1838647122810895e-45 ], [ 850.6875 , 2.1841576780733492e-45 ], 
                    [ 850.75 , 2.1844506831670316e-45 ], [ 850.8125 , 2.1847437275673776e-45 ], [ 850.875 , 2.1850368112796913e-45 ], [ 850.9375 , 2.1853299343091997e-45 ], 
                    [ 851.0 , 2.1856230966612235e-45 ], [ 851.0625 , 2.1859162983410538e-45 ], [ 851.125 , 2.1862095393538887e-45 ], [ 851.1875 , 2.1865028197050823e-45 ], 
                    [ 851.25 , 2.186796139399865e-45 ], [ 851.3125 , 2.187089498443562e-45 ], [ 851.375 , 2.1873828968413894e-45 ], [ 851.4375 , 2.1876763345986577e-45 ], 
                    [ 851.5 , 2.1879698117206472e-45 ], [ 851.5625 , 2.1882633282126695e-45 ], [ 851.625 , 2.1885568840799438e-45 ], [ 851.6875 , 2.1888504793277836e-45 ], 
                    [ 851.75 , 2.189144113961456e-45 ], [ 851.8125 , 2.1894377879862918e-45 ], [ 851.875 , 2.1897315014075595e-45 ], [ 851.9375 , 2.1900252542304657e-45 ], 
                    [ 852.0 , 2.1903190464604217e-45 ], [ 852.0625 , 2.1906128781026047e-45 ], [ 852.125 , 2.1909067491623946e-45 ], [ 852.1875 , 2.191200659645002e-45 ], 
                    [ 852.25 , 2.1914946095557317e-45 ], [ 852.3125 , 2.1917885988999493e-45 ], [ 852.375 , 2.192082627682837e-45 ], [ 852.4375 , 2.1923763133401208e-45 ], 
                    [ 852.5 , 2.1926697832635913e-45 ], [ 852.5625 , 2.1929632924707204e-45 ], [ 852.625 , 2.1932568409668125e-45 ], [ 852.6875 , 2.1935504287570805e-45 ], 
                    [ 852.75 , 2.193844055846769e-45 ], [ 852.8125 , 2.1941377222412006e-45 ], [ 852.875 , 2.194431427945605e-45 ], [ 852.9375 , 2.194725172965261e-45 ], 
                    [ 853.0 , 2.1950189573053835e-45 ], [ 853.0625 , 2.1953127809712985e-45 ], [ 853.125 , 2.1956066439682237e-45 ], [ 853.1875 , 2.1959005463014078e-45 ], 
                    [ 853.25 , 2.1961944879761948e-45 ], [ 853.3125 , 2.1964884689977726e-45 ], [ 853.375 , 2.1967824893714706e-45 ], [ 853.4375 , 2.197076549102525e-45 ], 
                    [ 853.5 , 2.197370648196158e-45 ], [ 853.5625 , 2.1976647866577006e-45 ], [ 853.625 , 2.1979589644924537e-45 ], [ 853.6875 , 2.198253181705595e-45 ], 
                    [ 853.75 , 2.1985474383024568e-45 ], [ 853.8125 , 2.198841734288296e-45 ], [ 853.875 , 2.199136069668354e-45 ], [ 853.9375 , 2.1994304444479817e-45 ], 
                    [ 854.0 , 2.1997248586324073e-45 ], [ 854.0625 , 2.2000193122268573e-45 ], [ 854.125 , 2.2003138052366863e-45 ], [ 854.1875 , 2.2006083376671695e-45 ], 
                    [ 854.25 , 2.200902909523522e-45 ], [ 854.3125 , 2.2011975208110212e-45 ], [ 854.375 , 2.2014922550515612e-45 ], [ 854.4375 , 2.2017875252821704e-45 ], 
                    [ 854.5 , 2.202082835115208e-45 ], [ 854.5625 , 2.2023781845560332e-45 ], [ 854.625 , 2.2026735736099577e-45 ], [ 854.6875 , 2.2029690022822482e-45 ], 
                    [ 854.75 , 2.20326447057828e-45 ], [ 854.8125 , 2.203559978503322e-45 ], [ 854.875 , 2.2038555260626886e-45 ], [ 854.9375 , 2.2041511132617424e-45 ], 
                    [ 855.0 , 2.204446740105769e-45 ], [ 855.0625 , 2.20474240660007e-45 ], [ 855.125 , 2.205038112749979e-45 ], [ 855.1875 , 2.205333858560799e-45 ], 
                    [ 855.25 , 2.2056296440379125e-45 ], [ 855.3125 , 2.2059254691865452e-45 ], [ 855.375 , 2.206221334012096e-45 ], [ 855.4375 , 2.2065172385198402e-45 ], 
                    [ 855.5 , 2.2068131827151003e-45 ], [ 855.5625 , 2.207109166603261e-45 ], [ 855.625 , 2.2074051901895678e-45 ], [ 855.6875 , 2.207701253479377e-45 ], 
                    [ 855.75 , 2.2079973564780138e-45 ], [ 855.8125 , 2.2082934991907877e-45 ], [ 855.875 , 2.2085896816230734e-45 ], [ 855.9375 , 2.2088859037801342e-45 ], 
                    [ 856.0 , 2.2091821656673773e-45 ], [ 856.0625 , 2.209478467290053e-45 ], [ 856.125 , 2.2097748086535058e-45 ], [ 856.1875 , 2.2100711897631294e-45 ], 
                    [ 856.25 , 2.2103676106241756e-45 ], [ 856.3125 , 2.2106640712420077e-45 ], [ 856.375 , 2.2109871194617406e-45 ], [ 856.4375 , 2.2113262141559653e-45 ], 
                    [ 856.5 , 2.2116653608564212e-45 ], [ 856.5625 , 2.212004559571163e-45 ], [ 856.625 , 2.2123438103081365e-45 ], [ 856.6875 , 2.2126831130753048e-45 ], 
                    [ 856.75 , 2.213022467880663e-45 ], [ 856.8125 , 2.2133618747322083e-45 ], [ 856.875 , 2.2137013336378913e-45 ], [ 856.9375 , 2.214040844605711e-45 ], 
                    [ 857.0 , 2.2143804076436674e-45 ], [ 857.0625 , 2.2147200227597157e-45 ], [ 857.125 , 2.2150596899618742e-45 ], [ 857.1875 , 2.2153994092581e-45 ], 
                    [ 857.25 , 2.215739180656398e-45 ], [ 857.3125 , 2.2160790041647748e-45 ], [ 857.375 , 2.2164188797911592e-45 ], [ 857.4375 , 2.2167588075436557e-45 ], 
                    [ 857.5 , 2.2170987874301786e-45 ], [ 857.5625 , 2.217438819458709e-45 ], [ 857.625 , 2.217778903637322e-45 ], [ 857.6875 , 2.2181190399739218e-45 ], 
                    [ 857.75 , 2.21845922847657e-45 ], [ 857.8125 , 2.218799469153283e-45 ], [ 857.875 , 2.2191397620120004e-45 ], [ 857.9375 , 2.2194801070607562e-45 ], 
                    [ 858.0 , 2.2198205043075547e-45 ], [ 858.0625 , 2.2201609537604332e-45 ], [ 858.125 , 2.220501455427319e-45 ], [ 858.1875 , 2.2208420093163156e-45 ], 
                    [ 858.25 , 2.2211826154353368e-45 ], [ 858.3125 , 2.2215211817024475e-45 ], [ 858.375 , 2.2217747099497434e-45 ], [ 858.4375 , 2.222028267130655e-45 ], 
                    [ 858.5 , 2.222281853248423e-45 ], [ 858.5625 , 2.2225354683064118e-45 ], [ 858.625 , 2.2227891123078612e-45 ], [ 858.6875 , 2.223042785256106e-45 ], 
                    [ 858.75 , 2.2232964871544816e-45 ], [ 858.8125 , 2.223550218006228e-45 ], [ 858.875 , 2.223803977814666e-45 ], [ 858.9375 , 2.224057766583179e-45 ], 
                    [ 859.0 , 2.224311584314961e-45 ], [ 859.0625 , 2.2245654310133812e-45 ], [ 859.125 , 2.2248193066817133e-45 ], [ 859.1875 , 2.2250732113232797e-45 ], 
                    [ 859.25 , 2.2253271449414026e-45 ], [ 859.3125 , 2.2255811075393726e-45 ], [ 859.375 , 2.2258350991205134e-45 ], [ 859.4375 , 2.2260891196880697e-45 ], 
                    [ 859.5 , 2.2263431692454278e-45 ], [ 859.5625 , 2.2265972477958646e-45 ], [ 859.625 , 2.2268513553426738e-45 ], [ 859.6875 , 2.2271054918891953e-45 ], 
                    [ 859.75 , 2.2273596574387236e-45 ], [ 859.8125 , 2.2276138519945842e-45 ], [ 859.875 , 2.2278680755600397e-45 ], [ 859.9375 , 2.2281223281384484e-45 ], 
                    [ 860.0 , 2.2283766097331053e-45 ], [ 860.0625 , 2.228630920347307e-45 ], [ 860.125 , 2.2288852599843958e-45 ], [ 860.1875 , 2.2291396286476366e-45 ], 
                    [ 860.25 , 2.229394026340406e-45 ], [ 860.3125 , 2.2296620981026685e-45 ], [ 860.375 , 2.2299544630650757e-45 ], [ 860.4375 , 2.2302468663639385e-45 ], 
                    [ 860.5 , 2.2305393080042046e-45 ], [ 860.5625 , 2.2308317879909963e-45 ], [ 860.625 , 2.2311243063292626e-45 ], [ 860.6875 , 2.2314168630240963e-45 ], 
                    [ 860.75 , 2.231709458080447e-45 ], [ 860.8125 , 2.2320020915034087e-45 ], [ 860.875 , 2.2322947632980117e-45 ], [ 860.9375 , 2.2325874734692566e-45 ], 
                    [ 861.0 , 2.2328802220221906e-45 ], [ 861.0625 , 2.233173008961863e-45 ], [ 861.125 , 2.233465834293291e-45 ], [ 861.1875 , 2.233758698021493e-45 ], 
                    [ 861.25 , 2.234051600151551e-45 ], [ 861.3125 , 2.2343445406884218e-45 ], [ 861.375 , 2.2346375196372366e-45 ], [ 861.4375 , 2.2349305370029364e-45 ], 
                    [ 861.5 , 2.2352235927906383e-45 ], [ 861.5625 , 2.235516687005349e-45 ], [ 861.625 , 2.235809819652075e-45 ], [ 861.6875 , 2.2361029907359195e-45 ], 
                    [ 861.75 , 2.236396200261875e-45 ], [ 861.8125 , 2.2366894482350137e-45 ], [ 861.875 , 2.2369827346603608e-45 ], [ 861.9375 , 2.2372760595429436e-45 ], 
                    [ 862.0 , 2.2375694228878355e-45 ], [ 862.0625 , 2.2378628247000167e-45 ], [ 862.125 , 2.2381562649846585e-45 ], [ 862.1875 , 2.2384497437466782e-45 ], 
                    [ 862.25 , 2.238743260991169e-45 ], [ 862.3125 , 2.2390385256399797e-45 ], [ 862.375 , 2.2393348946707413e-45 ], [ 862.4375 , 2.2396313029302172e-45 ], 
                    [ 862.5 , 2.2399277504235997e-45 ], [ 862.5625 , 2.2402242371560986e-45 ], [ 862.625 , 2.2405207631328905e-45 ], [ 862.6875 , 2.240817328359155e-45 ], 
                    [ 862.75 , 2.241113932840119e-45 ], [ 862.8125 , 2.2414105765809307e-45 ], [ 862.875 , 2.2417072595868498e-45 ], [ 862.9375 , 2.242003981863043e-45 ], 
                    [ 863.0 , 2.2423007434146907e-45 ], [ 863.0625 , 2.2425975442470248e-45 ], [ 863.125 , 2.2428943843652283e-45 ], [ 863.1875 , 2.2431912637744852e-45 ], 
                    [ 863.25 , 2.2434881824800122e-45 ], [ 863.3125 , 2.2437851404870264e-45 ], [ 863.375 , 2.244082137800699e-45 ], [ 863.4375 , 2.2443791744262638e-45 ], 
                    [ 863.5 , 2.2446762503688927e-45 ], [ 863.5625 , 2.2449733656338055e-45 ], [ 863.625 , 2.2452705202262075e-45 ], [ 863.6875 , 2.24556771415132e-45 ], 
                    [ 863.75 , 2.245864947414318e-45 ], [ 863.8125 , 2.2461622200204545e-45 ], [ 863.875 , 2.2464595319748913e-45 ], [ 863.9375 , 2.2467568832828357e-45 ], 
                    [ 864.0 , 2.2470542739495128e-45 ], [ 864.0625 , 2.2473517039801164e-45 ], [ 864.125 , 2.247649173379905e-45 ], [ 864.1875 , 2.247946682154057e-45 ], 
                    [ 864.25 , 2.2482442303077853e-45 ], [ 864.3125 , 2.248541401902824e-45 ], [ 864.375 , 2.2488385139267788e-45 ], [ 864.4375 , 2.2491356652096994e-45 ], 
                    [ 864.5 , 2.249432855756886e-45 ], [ 864.5625 , 2.2497300855734786e-45 ], [ 864.625 , 2.250027354664634e-45 ], [ 864.6875 , 2.2503246630356057e-45 ], 
                    [ 864.75 , 2.250622010691536e-45 ], [ 864.8125 , 2.2509193976376318e-45 ], [ 864.875 , 2.2512168238791005e-45 ], [ 864.9375 , 2.251514289421087e-45 ], 
                    [ 865.0 , 2.2518117942688478e-45 ], [ 865.0625 , 2.2521093384275126e-45 ], [ 865.125 , 2.252406921902356e-45 ], [ 865.1875 , 2.252704544698477e-45 ], 
                    [ 865.25 , 2.2530022068211513e-45 ], [ 865.3125 , 2.2532999082755273e-45 ], [ 865.375 , 2.253597649066802e-45 ], [ 865.4375 , 2.2538954292002215e-45 ], 
                    [ 865.5 , 2.2541932486809684e-45 ], [ 865.5625 , 2.2544911075141614e-45 ], [ 865.625 , 2.2547890057050803e-45 ], [ 865.6875 , 2.255086943258974e-45 ], 
                    [ 865.75 , 2.2553849201809318e-45 ], [ 865.8125 , 2.2556829364761875e-45 ], [ 865.875 , 2.2559809921499757e-45 ], [ 865.9375 , 2.2562790872074833e-45 ], 
                    [ 866.0 , 2.256577221653947e-45 ], [ 866.0625 , 2.256875395494491e-45 ], [ 866.125 , 2.257173608734401e-45 ], [ 866.1875 , 2.2574718613788185e-45 ], 
                    [ 866.25 , 2.2577701534330147e-45 ], [ 866.3125 , 2.2580679269529796e-45 ], [ 866.375 , 2.258365691222094e-45 ], [ 866.4375 , 2.2586634947564245e-45 ], 
                    [ 866.5 , 2.2589613375611636e-45 ], [ 866.5625 , 2.259259219641491e-45 ], [ 866.625 , 2.259557141002601e-45 ], [ 866.6875 , 2.259855101649626e-45 ], 
                    [ 866.75 , 2.26015310158781e-45 ], [ 866.8125 , 2.2604511408222865e-45 ], [ 866.875 , 2.2607492193582377e-45 ], [ 866.9375 , 2.2610473372008932e-45 ], 
                    [ 867.0 , 2.2613454943553894e-45 ], [ 867.0625 , 2.2616436908269416e-45 ], [ 867.125 , 2.2619419266207024e-45 ], [ 867.1875 , 2.262240201741857e-45 ], 
                    [ 867.25 , 2.262538516195656e-45 ], [ 867.3125 , 2.2628368699871893e-45 ], [ 867.375 , 2.2631352631217084e-45 ], [ 867.4375 , 2.2634336956044017e-45 ], 
                    [ 867.5 , 2.26373216744041e-45 ], [ 867.5625 , 2.2640306786350022e-45 ], [ 867.625 , 2.264329229193305e-45 ], [ 867.6875 , 2.264627819120492e-45 ], 
                    [ 867.75 , 2.2649264484218043e-45 ], [ 867.8125 , 2.2652251171024166e-45 ], [ 867.875 , 2.2655238251675385e-45 ], [ 867.9375 , 2.2658225726223154e-45 ], 
                    [ 868.0 , 2.266121359471989e-45 ], [ 868.0625 , 2.2664201857217874e-45 ], [ 868.125 , 2.266719051376777e-45 ], [ 868.1875 , 2.267017956442266e-45 ], 
                    [ 868.25 , 2.2673169009234198e-45 ], [ 868.3125 , 2.2676156944535987e-45 ], [ 868.375 , 2.2679145177597866e-45 ], [ 868.4375 , 2.2682133804445194e-45 ], 
                    [ 868.5 , 2.2685122825129526e-45 ], [ 868.5625 , 2.2688112239703253e-45 ], [ 868.625 , 2.269110204821748e-45 ], [ 868.6875 , 2.2694092250724917e-45 ], 
                    [ 868.75 , 2.269708284727685e-45 ], [ 868.8125 , 2.2700073837925836e-45 ], [ 868.875 , 2.27030652227235e-45 ], [ 868.9375 , 2.2706057001721774e-45 ], 
                    [ 869.0 , 2.2709049174972608e-45 ], [ 869.0625 , 2.271204174252763e-45 ], [ 869.125 , 2.2715034704439452e-45 ], [ 869.1875 , 2.27180280607594e-45 ], 
                    [ 869.25 , 2.2721021811540076e-45 ], [ 869.3125 , 2.2724015956833153e-45 ], [ 869.375 , 2.2727010496690624e-45 ], [ 869.4375 , 2.273000543116447e-45 ], 
                    [ 869.5 , 2.273300076030654e-45 ], [ 869.5625 , 2.273599648416868e-45 ], [ 869.625 , 2.2738992602804198e-45 ], [ 869.6875 , 2.2741989116263176e-45 ], 
                    [ 869.75 , 2.2744986024599263e-45 ], [ 869.8125 , 2.2747983327863522e-45 ], [ 869.875 , 2.275098102610832e-45 ], [ 869.9375 , 2.275397911938571e-45 ], 
                    [ 870.0 , 2.2756977607747744e-45 ], [ 870.0625 , 2.2759976491246493e-45 ], [ 870.125 , 2.276297576993418e-45 ], [ 870.1875 , 2.2765975443862728e-45 ], 
                    [ 870.25 , 2.2768975513084054e-45 ], [ 870.3125 , 2.2771969974253187e-45 ], [ 870.375 , 2.2774964307135827e-45 ], [ 870.4375 , 2.2777959033749253e-45 ], 
                    [ 870.5 , 2.2780954154145242e-45 ], [ 870.5625 , 2.278394966837606e-45 ], [ 870.625 , 2.2786945576493006e-45 ], [ 870.6875 , 2.2789941878548195e-45 ], 
                    [ 870.75 , 2.2792938574593104e-45 ], [ 870.8125 , 2.279593566467954e-45 ], [ 870.875 , 2.2798933148859806e-45 ], [ 870.9375 , 2.280193102718523e-45 ], 
                    [ 871.0 , 2.28049292997078e-45 ], [ 871.0625 , 2.2807927966479356e-45 ], [ 871.125 , 2.2810927027551737e-45 ], [ 871.1875 , 2.2813926482976785e-45 ], 
                    [ 871.25 , 2.2816926332806362e-45 ], [ 871.3125 , 2.2819926577092487e-45 ], [ 871.375 , 2.282292721588654e-45 ], [ 871.4375 , 2.2825928249240888e-45 ], 
                    [ 871.5 , 2.2828929677207567e-45 ], [ 871.5625 , 2.2831931499837822e-45 ], [ 871.625 , 2.2834933717183548e-45 ], [ 871.6875 , 2.283793632929713e-45 ], 
                    [ 871.75 , 2.2840939336230802e-45 ], [ 871.8125 , 2.2843942738035674e-45 ], [ 871.875 , 2.2846946534764313e-45 ], [ 871.9375 , 2.2849950726467834e-45 ], 
                    [ 872.0 , 2.2852955313198828e-45 ], [ 872.0625 , 2.2855960295009074e-45 ], [ 872.125 , 2.2858965671950844e-45 ], [ 872.1875 , 2.2861971444075284e-45 ], 
                    [ 872.25 , 2.2864977611435173e-45 ], [ 872.3125 , 2.2867979048341876e-45 ], [ 872.375 , 2.2870979661083264e-45 ], [ 872.4375 , 2.2873980667549182e-45 ], 
                    [ 872.5 , 2.28769820677908e-45 ], [ 872.5625 , 2.2879983861860116e-45 ], [ 872.625 , 2.288298604980848e-45 ], [ 872.6875 , 2.2885988631687898e-45 ], 
                    [ 872.75 , 2.28889916075499e-45 ], [ 872.8125 , 2.289199497744602e-45 ], [ 872.875 , 2.289499874142812e-45 ], [ 872.9375 , 2.2898002899548074e-45 ], 
                    [ 873.0 , 2.29010074518576e-45 ], [ 873.0625 , 2.290401239840793e-45 ], [ 873.125 , 2.2907017739251128e-45 ], [ 873.1875 , 2.291002347443892e-45 ], 
                    [ 873.25 , 2.291302960402306e-45 ], [ 873.3125 , 2.2916036128055613e-45 ], [ 873.375 , 2.2919043046587368e-45 ], [ 873.4375 , 2.2922050359671388e-45 ], 
                    [ 873.5 , 2.292505806735847e-45 ], [ 873.5625 , 2.2928066169700715e-45 ], [ 873.625 , 2.2931074666750067e-45 ], [ 873.6875 , 2.293408355855816e-45 ], 
                    [ 873.75 , 2.2937092845176628e-45 ], [ 873.8125 , 2.2940102526657603e-45 ], [ 873.875 , 2.2943112603052406e-45 ], [ 873.9375 , 2.2946123074413507e-45 ], 
                    [ 874.0 , 2.2949133940792566e-45 ], [ 874.0625 , 2.2952145202240932e-45 ], [ 874.125 , 2.2955156858810767e-45 ], [ 874.1875 , 2.295816891055424e-45 ], 
                    [ 874.25 , 2.2961181357522395e-45 ], [ 874.3125 , 2.2964190878112434e-45 ], [ 874.375 , 2.296719885877825e-45 ], [ 874.4375 , 2.297020723344685e-45 ], 
                    [ 874.5 , 2.297321600216936e-45 ], [ 874.5625 , 2.297622516499739e-45 ], [ 874.625 , 2.2979234721982885e-45 ], [ 874.6875 , 2.298224467317732e-45 ], 
                    [ 874.75 , 2.2985255018632157e-45 ], [ 874.8125 , 2.2988265758399372e-45 ], [ 874.875 , 2.2991276892530125e-45 ], [ 874.9375 , 2.299428842107705e-45 ], 
                    [ 875.0 , 2.2997300344090664e-45 ], [ 875.0625 , 2.3000312661623128e-45 ], [ 875.125 , 2.3003325373726445e-45 ], [ 875.1875 , 2.3006338480451485e-45 ], 
                    [ 875.25 , 2.3009351981850752e-45 ], [ 875.3125 , 2.3012365877975614e-45 ], [ 875.375 , 2.301538016887761e-45 ], [ 875.4375 , 2.301839485460862e-45 ], 
                    [ 875.5 , 2.302140993522052e-45 ], [ 875.5625 , 2.302442541076487e-45 ], [ 875.625 , 2.302744128129324e-45 ], [ 875.6875 , 2.3030457546857855e-45 ], 
                    [ 875.75 , 2.303347420750997e-45 ], [ 875.8125 , 2.3036491263301165e-45 ], [ 875.875 , 2.3039508714284028e-45 ], [ 875.9375 , 2.304252656050934e-45 ], 
                    [ 876.0 , 2.3045544802029678e-45 ], [ 876.0625 , 2.30485634388965e-45 ], [ 876.125 , 2.3051582471161433e-45 ], [ 876.1875 , 2.3054601898876916e-45 ], 
                    [ 876.25 , 2.3057621722093767e-45 ], [ 876.3125 , 2.3060638981912235e-45 ], [ 876.375 , 2.3063651725318832e-45 ], [ 876.4375 , 2.3066664862322833e-45 ], 
                    [ 876.5 , 2.3069678392976483e-45 ], [ 876.5625 , 2.3072692317331375e-45 ], [ 876.625 , 2.3075706635438448e-45 ], [ 876.6875 , 2.3078721347349146e-45 ], 
                    [ 876.75 , 2.3081736453115245e-45 ], [ 876.8125 , 2.3084751952787875e-45 ], [ 876.875 , 2.3087767846418983e-45 ], [ 876.9375 , 2.3090784134059388e-45 ], 
                    [ 877.0 , 2.3093800815761214e-45 ], [ 877.0625 , 2.3096817891575295e-45 ], [ 877.125 , 2.3099835361553443e-45 ], [ 877.1875 , 2.3102853225747318e-45 ], 
                    [ 877.25 , 2.3105871484208254e-45 ], [ 877.3125 , 2.3108890136987937e-45 ], [ 877.375 , 2.3111909184137702e-45 ], [ 877.4375 , 2.3114928625708586e-45 ], 
                    [ 877.5 , 2.3117948461752777e-45 ], [ 877.5625 , 2.312096869232214e-45 ], [ 877.625 , 2.3123989317467224e-45 ], [ 877.6875 , 2.3127010337240245e-45 ], 
                    [ 877.75 , 2.3130031751692585e-45 ], [ 877.8125 , 2.313305356087565e-45 ], [ 877.875 , 2.3136075764841172e-45 ], [ 877.9375 , 2.3139098363641052e-45 ], 
                    [ 878.0 , 2.3142121357326546e-45 ], [ 878.0625 , 2.314514474594859e-45 ], [ 878.125 , 2.3148168529560088e-45 ], [ 878.1875 , 2.3151192708211668e-45 ], 
                    [ 878.25 , 2.3154217281955587e-45 ], [ 878.3125 , 2.315724205543952e-45 ], [ 878.375 , 2.3160262533775563e-45 ], [ 878.4375 , 2.316328340608327e-45 ], 
                    [ 878.5 , 2.3166304672413358e-45 ], [ 878.5625 , 2.3169326332817555e-45 ], [ 878.625 , 2.3172348387347258e-45 ], [ 878.6875 , 2.3175370836053882e-45 ], 
                    [ 878.75 , 2.3178393678989158e-45 ], [ 878.8125 , 2.3181416916203854e-45 ], [ 878.875 , 2.3184440547749565e-45 ], [ 878.9375 , 2.318746457367838e-45 ], 
                    [ 879.0 , 2.319048899404092e-45 ], [ 879.0625 , 2.3193513808888957e-45 ], [ 879.125 , 2.3196539018273782e-45 ], [ 879.1875 , 2.3199564622247514e-45 ], 
                    [ 879.25 , 2.3202590620860797e-45 ], [ 879.3125 , 2.3205617014165436e-45 ], [ 879.375 , 2.320864380221308e-45 ], [ 879.4375 , 2.3211670985055045e-45 ], 
                    [ 879.5 , 2.32146985627425e-45 ], [ 879.5625 , 2.3217726535327768e-45 ], [ 879.625 , 2.3220754902861037e-45 ], [ 879.6875 , 2.3223783665395805e-45 ], 
                    [ 879.75 , 2.322681282298194e-45 ], [ 879.8125 , 2.3229842375671138e-45 ], [ 879.875 , 2.3232872323515755e-45 ], [ 879.9375 , 2.323590266656618e-45 ], 
                    [ 880.0 , 2.3238933404875115e-45 ], [ 880.0625 , 2.324196453849362e-45 ], [ 880.125 , 2.3244996067473092e-45 ], [ 880.1875 , 2.3248027991865268e-45 ], 
                    [ 880.25 , 2.325106031172155e-45 ], [ 880.3125 , 2.3254093027094348e-45 ], [ 880.375 , 2.325712294649782e-45 ], [ 880.4375 , 2.3260151337332967e-45 ], 
                    [ 880.5 , 2.326318012250509e-45 ], [ 880.5625 , 2.3266209302066015e-45 ], [ 880.625 , 2.3269238876066947e-45 ], [ 880.6875 , 2.3272268844559407e-45 ], 
                    [ 880.75 , 2.3275299207594433e-45 ], [ 880.8125 , 2.3278329965223568e-45 ], [ 880.875 , 2.3281361117498354e-45 ], [ 880.9375 , 2.3284392664469685e-45 ], 
                    [ 881.0 , 2.3287424606189617e-45 ], [ 881.0625 , 2.3290456942709058e-45 ], [ 881.125 , 2.3293489674079577e-45 ], [ 881.1875 , 2.3296522800352595e-45 ], 
                    [ 881.25 , 2.3299556321579522e-45 ], [ 881.3125 , 2.3302590237811804e-45 ], [ 881.375 , 2.330562454910119e-45 ], [ 881.4375 , 2.330865925549814e-45 ], 
                    [ 881.5 , 2.3311694357055254e-45 ], [ 881.5625 , 2.3314729853823332e-45 ], [ 881.625 , 2.331776574585383e-45 ], [ 881.6875 , 2.3320802033198717e-45 ], 
                    [ 881.75 , 2.3323838715908967e-45 ], [ 881.8125 , 2.332687579403623e-45 ], [ 881.875 , 2.3329913267632162e-45 ], [ 881.9375 , 2.3332951136747925e-45 ], 
                    [ 882.0 , 2.3335989401435187e-45 ], [ 882.0625 , 2.3339028061745456e-45 ], [ 882.125 , 2.3342067117730412e-45 ], [ 882.1875 , 2.3345106569441254e-45 ], 
                    [ 882.25 , 2.3348146416929334e-45 ], [ 882.3125 , 2.3351186660247185e-45 ], [ 882.375 , 2.33542263734588e-45 ], [ 882.4375 , 2.3357260977170612e-45 ], 
                    [ 882.5 , 2.3360295975192722e-45 ], [ 882.5625 , 2.3363331367576863e-45 ], [ 882.625 , 2.3366367154373942e-45 ], [ 882.6875 , 2.3369403335635547e-45 ], 
                    [ 882.75 , 2.3372439911412764e-45 ], [ 882.8125 , 2.3375476881756525e-45 ], [ 882.875 , 2.3378514246718925e-45 ], [ 882.9375 , 2.3381552006350414e-45 ], 
                    [ 883.0 , 2.3384590160702775e-45 ], [ 883.0625 , 2.3387628709827295e-45 ], [ 883.125 , 2.3390667653774768e-45 ], [ 883.1875 , 2.3393706992597166e-45 ], 
                    [ 883.25 , 2.3396746726344964e-45 ], [ 883.3125 , 2.3399786855070147e-45 ], [ 883.375 , 2.340282737882403e-45 ], [ 883.4375 , 2.3405868297657786e-45 ], 
                    [ 883.5 , 2.340890961162274e-45 ], [ 883.5625 , 2.3411951320770076e-45 ], [ 883.625 , 2.3414993425151645e-45 ], [ 883.6875 , 2.3418035924818463e-45 ], 
                    [ 883.75 , 2.3421078819821225e-45 ], [ 883.8125 , 2.3424122110212642e-45 ], [ 883.875 , 2.3427165796043084e-45 ], [ 883.9375 , 2.3430209877364437e-45 ], 
                    [ 884.0 , 2.3433254354227748e-45 ], [ 884.0625 , 2.343629922668509e-45 ], [ 884.125 , 2.3439344494786862e-45 ], [ 884.1875 , 2.344239015858514e-45 ], 
                    [ 884.25 , 2.3445436218131008e-45 ], [ 884.3125 , 2.3448482673476054e-45 ], [ 884.375 , 2.3451529524671703e-45 ], [ 884.4375 , 2.345457267106813e-45 ], 
                    [ 884.5 , 2.3457613478197437e-45 ], [ 884.5625 , 2.3460654679557415e-45 ], [ 884.625 , 2.346369627519884e-45 ], [ 884.6875 , 2.3466738265173e-45 ], 
                    [ 884.75 , 2.3469780649531013e-45 ], [ 884.8125 , 2.347282342832418e-45 ], [ 884.875 , 2.347586660160347e-45 ], [ 884.9375 , 2.3478910169419693e-45 ], 
                    [ 885.0 , 2.3481954131824336e-45 ], [ 885.0625 , 2.3484998488868883e-45 ], [ 885.125 , 2.3488043240604172e-45 ], [ 885.1875 , 2.3491088387081034e-45 ], 
                    [ 885.25 , 2.3494133928351318e-45 ], [ 885.3125 , 2.349717986446587e-45 ], [ 885.375 , 2.3500226195475714e-45 ], [ 885.4375 , 2.3503272921432213e-45 ], 
                    [ 885.5 , 2.3506320042386408e-45 ], [ 885.5625 , 2.350936755839034e-45 ], [ 885.625 , 2.3512415469493887e-45 ], [ 885.6875 , 2.351546377574912e-45 ], 
                    [ 885.75 , 2.3518512477206918e-45 ], [ 885.8125 , 2.35215615739192e-45 ], [ 885.875 , 2.352461106593636e-45 ], [ 885.9375 , 2.352766095331e-45 ], 
                    [ 886.0 , 2.353071123609136e-45 ], [ 886.0625 , 2.353376191433205e-45 ], [ 886.125 , 2.3536812988082327e-45 ], [ 886.1875 , 2.3539864457394643e-45 ], 
                    [ 886.25 , 2.3542916322319614e-45 ], [ 886.3125 , 2.3545968582908865e-45 ], [ 886.375 , 2.354902123921335e-45 ], [ 886.4375 , 2.355207429128438e-45 ], 
                    [ 886.5 , 2.3555124406654086e-45 ], [ 886.5625 , 2.3558174778156318e-45 ], [ 886.625 , 2.3561225544679274e-45 ], [ 886.6875 , 2.356427670627445e-45 ], 
                    [ 886.75 , 2.3567328262992833e-45 ], [ 886.8125 , 2.3570380214885597e-45 ], [ 886.875 , 2.3573432562003912e-45 ], [ 886.9375 , 2.3576485304398966e-45 ], 
                    [ 887.0 , 2.3579538442122106e-45 ], [ 887.0625 , 2.3582591975223864e-45 ], [ 887.125 , 2.358564590375644e-45 ], [ 887.1875 , 2.3588700227770378e-45 ], 
                    [ 887.25 , 2.359175494731689e-45 ], [ 887.3125 , 2.3594810062447534e-45 ], [ 887.375 , 2.3597865573213532e-45 ], [ 887.4375 , 2.360092147966579e-45 ], 
                    [ 887.5 , 2.3603977781855546e-45 ], [ 887.5625 , 2.3607034479834384e-45 ], [ 887.625 , 2.3610091573653387e-45 ], [ 887.6875 , 2.361314906336399e-45 ], 
                    [ 887.75 , 2.361620694901695e-45 ], [ 887.8125 , 2.361926523066422e-45 ], [ 887.875 , 2.3622323908356745e-45 ], [ 887.9375 , 2.362538298214547e-45 ], 
                    [ 888.0 , 2.362844245208203e-45 ], [ 888.0625 , 2.3631502318217892e-45 ], [ 888.125 , 2.36345625806042e-45 ], [ 888.1875 , 2.363762323929209e-45 ], 
                    [ 888.25 , 2.3640684294333228e-45 ], [ 888.3125 , 2.364374574577877e-45 ], [ 888.375 , 2.364680759368005e-45 ], [ 888.4375 , 2.3649869838088408e-45 ], 
                    [ 888.5 , 2.365293043558763e-45 ], [ 888.5625 , 2.3655985489572145e-45 ], [ 888.625 , 2.365904093815296e-45 ], [ 888.6875 , 2.3662096781380535e-45 ], 
                    [ 888.75 , 2.3665153019306353e-45 ], [ 888.8125 , 2.366820965198122e-45 ], [ 888.875 , 2.3671266679456126e-45 ], [ 888.9375 , 2.367432410178206e-45 ], 
                    [ 889.0 , 2.367738191901003e-45 ], [ 889.0625 , 2.36804401311912e-45 ], [ 889.125 , 2.3683498738376424e-45 ], [ 889.1875 , 2.368655774061604e-45 ], 
                    [ 889.25 , 2.368961713796243e-45 ], [ 889.3125 , 2.369267693046594e-45 ], [ 889.375 , 2.369573711817745e-45 ], [ 889.4375 , 2.3698797701148176e-45 ], 
                    [ 889.5 , 2.3701858679429002e-45 ], [ 889.5625 , 2.3704920053071484e-45 ], [ 889.625 , 2.3707981822126353e-45 ], [ 889.6875 , 2.3711043986644522e-45 ], 
                    [ 889.75 , 2.371410654667689e-45 ], [ 889.8125 , 2.371716950227539e-45 ], [ 889.875 , 2.3720232853490436e-45 ], [ 889.9375 , 2.3723296600373136e-45 ], 
                    [ 890.0 , 2.3726360742975258e-45 ], [ 890.0625 , 2.3729425281347082e-45 ], [ 890.125 , 2.3732490215540053e-45 ], [ 890.1875 , 2.3735555545605302e-45 ], 
                    [ 890.25 , 2.373862127159379e-45 ], [ 890.3125 , 2.3741687393556826e-45 ], [ 890.375 , 2.3744753911546225e-45 ], [ 890.4375 , 2.3747820825611625e-45 ], 
                    [ 890.5 , 2.3750888135805696e-45 ], [ 890.5625 , 2.3753953386943826e-45 ], [ 890.625 , 2.3757016286790702e-45 ], [ 890.6875 , 2.3760079581576613e-45 ], 
                    [ 890.75 , 2.3763143271351815e-45 ], [ 890.8125 , 2.376620735616774e-45 ], [ 890.875 , 2.3769271836074988e-45 ], [ 890.9375 , 2.377233671112501e-45 ], 
                    [ 891.0 , 2.3775401981368257e-45 ], [ 891.0625 , 2.3778467646855673e-45 ], [ 891.125 , 2.3781533707638404e-45 ], [ 891.1875 , 2.378460016376792e-45 ], 
                    [ 891.25 , 2.3787667015294182e-45 ], [ 891.3125 , 2.379073426226886e-45 ], [ 891.375 , 2.3793801904742767e-45 ], [ 891.4375 , 2.379686994276673e-45 ], 
                    [ 891.5 , 2.3799938376391926e-45 ], [ 891.5625 , 2.380300720566953e-45 ], [ 891.625 , 2.3806076430650395e-45 ], [ 891.6875 , 2.3809146051385365e-45 ], 
                    [ 891.75 , 2.3812216067926324e-45 ], [ 891.8125 , 2.3815286480322786e-45 ], [ 891.875 , 2.3818357288626802e-45 ], [ 891.9375 , 2.3821428492889426e-45 ], 
                    [ 892.0 , 2.3824500093161545e-45 ], [ 892.0625 , 2.3827572089494052e-45 ], [ 892.125 , 2.3830644481938353e-45 ], [ 892.1875 , 2.383371727054519e-45 ], 
                    [ 892.25 , 2.3836790455365977e-45 ], [ 892.3125 , 2.3839864036451807e-45 ], [ 892.375 , 2.3842938013853433e-45 ], [ 892.4375 , 2.3846012387621625e-45 ], 
                    [ 892.5 , 2.3849087157808673e-45 ], [ 892.5625 , 2.3852162324464845e-45 ], [ 892.625 , 2.3855232768060967e-45 ], [ 892.6875 , 2.385830052163742e-45 ], 
                    [ 892.75 , 2.3861368669723543e-45 ], [ 892.8125 , 2.386443721236957e-45 ], [ 892.875 , 2.3867506149626232e-45 ], [ 892.9375 , 2.387057548154445e-45 ], 
                    [ 893.0 , 2.3873645208175144e-45 ], [ 893.0625 , 2.3876715329568906e-45 ], [ 893.125 , 2.3879785845776328e-45 ], [ 893.1875 , 2.3882856756848528e-45 ], 
                    [ 893.25 , 2.3885928062836284e-45 ], [ 893.3125 , 2.388899976378987e-45 ], [ 893.375 , 2.3892071859760584e-45 ], [ 893.4375 , 2.3895144350798894e-45 ], 
                    [ 893.5 , 2.389821723695577e-45 ], [ 893.5625 , 2.3901290518282193e-45 ], [ 893.625 , 2.390436419482898e-45 ], [ 893.6875 , 2.390743826664662e-45 ], 
                    [ 893.75 , 2.3910512733786116e-45 ], [ 893.8125 , 2.391358759629796e-45 ], [ 893.875 , 2.3916662854233338e-45 ], [ 893.9375 , 2.391973850764361e-45 ], 
                    [ 894.0 , 2.392281455657913e-45 ], [ 894.0625 , 2.392589100109025e-45 ], [ 894.125 , 2.392896784122868e-45 ], [ 894.1875 , 2.393204507704497e-45 ], 
                    [ 894.25 , 2.3935122708589657e-45 ], [ 894.3125 , 2.3938200735914653e-45 ], [ 894.375 , 2.3941279159069668e-45 ], [ 894.4375 , 2.394435797810594e-45 ], 
                    [ 894.5 , 2.3947437193075234e-45 ], [ 894.5625 , 2.3950516804027103e-45 ], [ 894.625 , 2.3953596811013827e-45 ], [ 894.6875 , 2.395667480648714e-45 ], 
                    [ 894.75 , 2.3959752064345025e-45 ], [ 894.8125 , 2.3962829717479463e-45 ], [ 894.875 , 2.3965907765941565e-45 ], [ 894.9375 , 2.3968986209782118e-45 ], 
                    [ 895.0 , 2.397206504905156e-45 ], [ 895.0625 , 2.3975144283800867e-45 ], [ 895.125 , 2.3978223914080827e-45 ], [ 895.1875 , 2.398130393994225e-45 ], 
                    [ 895.25 , 2.3984384361435778e-45 ], [ 895.3125 , 2.3987465178612574e-45 ], [ 895.375 , 2.399054639152346e-45 ], [ 895.4375 , 2.3993628000219098e-45 ], 
                    [ 895.5 , 2.399671000474965e-45 ], [ 895.5625 , 2.3999792405167322e-45 ], [ 895.625 , 2.4002875201522286e-45 ], [ 895.6875 , 2.4005958393865227e-45 ], 
                    [ 895.75 , 2.400904198224684e-45 ], [ 895.8125 , 2.401212596671885e-45 ], [ 895.875 , 2.4015210347331455e-45 ], [ 895.9375 , 2.4018295124135704e-45 ], 
                    [ 896.0 , 2.4021380297182497e-45 ], [ 896.0625 , 2.4024465866522896e-45 ], [ 896.125 , 2.402755183220712e-45 ], [ 896.1875 , 2.4030638194287115e-45 ], 
                    [ 896.25 , 2.4033724952812934e-45 ], [ 896.3125 , 2.403681210783619e-45 ], [ 896.375 , 2.4039899659407298e-45 ], [ 896.4375 , 2.4042987607577365e-45 ], 
                    [ 896.5 , 2.404607595239717e-45 ], [ 896.5625 , 2.404916469391817e-45 ], [ 896.625 , 2.4052253832190636e-45 ], [ 896.6875 , 2.4055343367266047e-45 ], 
                    [ 896.75 , 2.4058429763124213e-45 ], [ 896.8125 , 2.4061514828818264e-45 ], [ 896.875 , 2.406460029011712e-45 ], [ 896.9375 , 2.406768614707151e-45 ], 
                    [ 897.0 , 2.407077239973216e-45 ], [ 897.0625 , 2.4073859048149824e-45 ], [ 897.125 , 2.4076946092375416e-45 ], [ 897.1875 , 2.4080033532459348e-45 ], 
                    [ 897.25 , 2.4083121368452557e-45 ], [ 897.3125 , 2.4086209600405808e-45 ], [ 897.375 , 2.408929822836971e-45 ], [ 897.4375 , 2.409238725239572e-45 ], 
                    [ 897.5 , 2.4095476672533774e-45 ], [ 897.5625 , 2.4098566488835185e-45 ], [ 897.625 , 2.410165670135041e-45 ], [ 897.6875 , 2.4104747310130247e-45 ], 
                    [ 897.75 , 2.410783831522569e-45 ], [ 897.8125 , 2.4110929716688243e-45 ], [ 897.875 , 2.4114021514567193e-45 ], [ 897.9375 , 2.411711370891474e-45 ], 
                    [ 898.0 , 2.412020629978104e-45 ], [ 898.0625 , 2.4123299287217112e-45 ], [ 898.125 , 2.412639267127398e-45 ], [ 898.1875 , 2.4129486452001987e-45 ], 
                    [ 898.25 , 2.4132580629452685e-45 ], [ 898.3125 , 2.4135675203676437e-45 ], [ 898.375 , 2.4138770174724464e-45 ], [ 898.4375 , 2.4141865542647304e-45 ], 
                    [ 898.5 , 2.4144961307495683e-45 ], [ 898.5625 , 2.4148057469321523e-45 ], [ 898.625 , 2.4151154028174705e-45 ], [ 898.6875 , 2.4154250984106652e-45 ], 
                    [ 898.75 , 2.415734833716777e-45 ], [ 898.8125 , 2.4160441127978454e-45 ], [ 898.875 , 2.4163531008039577e-45 ], [ 898.9375 , 2.4166621283265446e-45 ], 
                    [ 899.0 , 2.416971195370643e-45 ], [ 899.0625 , 2.4172803019413933e-45 ], [ 899.125 , 2.417589448043799e-45 ], [ 899.1875 , 2.417898633682898e-45 ], 
                    [ 899.25 , 2.4182078588637644e-45 ], [ 899.3125 , 2.4185171235914554e-45 ], [ 899.375 , 2.4188264278710455e-45 ], [ 899.4375 , 2.4191357717075587e-45 ], 
                    [ 899.5 , 2.4194451551060538e-45 ], [ 899.5625 , 2.419754578071642e-45 ], [ 899.625 , 2.4200640406093494e-45 ], [ 899.6875 , 2.420373542724202e-45 ], 
                    [ 899.75 , 2.4206830844213133e-45 ], [ 899.8125 , 2.4209926657057286e-45 ], [ 899.875 , 2.42130228658251e-45 ], [ 899.9375 , 2.4216119470567212e-45 ], 
                    [ 900.0 , 2.4219216471334438e-45 ], [ 900.0625 , 2.422231386817725e-45 ], [ 900.125 , 2.4225411661145963e-45 ], [ 900.1875 , 2.4228509850291575e-45 ], 
                    [ 900.25 , 2.4231608435665273e-45 ], [ 900.3125 , 2.4234707417316698e-45 ], [ 900.375 , 2.4237806795297388e-45 ], [ 900.4375 , 2.424090656965735e-45 ], 
                    [ 900.5 , 2.4244006740447784e-45 ], [ 900.5625 , 2.4247107307719217e-45 ], [ 900.625 , 2.4250208271522363e-45 ], [ 900.6875 , 2.4253309631908098e-45 ], 
                    [ 900.75 , 2.425641138892663e-45 ], [ 900.8125 , 2.425951354262937e-45 ], [ 900.875 , 2.4262613807371913e-45 ], [ 900.9375 , 2.4265711462666458e-45 ], 
                    [ 901.0 , 2.4268809513444875e-45 ], [ 901.0625 , 2.4271907959757313e-45 ], [ 901.125 , 2.4275006801654442e-45 ], [ 901.1875 , 2.427810603918677e-45 ], 
                    [ 901.25 , 2.4281205672404638e-45 ], [ 901.3125 , 2.4284305701359076e-45 ], [ 901.375 , 2.428740612610009e-45 ], [ 901.4375 , 2.429050694667839e-45 ], 
                    [ 901.5 , 2.429360816314451e-45 ], [ 901.5625 , 2.4296709775549165e-45 ], [ 901.625 , 2.4299811783942387e-45 ], [ 901.6875 , 2.4302914188375427e-45 ], 
                    [ 901.75 , 2.4306016988898327e-45 ], [ 901.8125 , 2.4309120185561833e-45 ], [ 901.875 , 2.4312223778416513e-45 ], [ 901.9375 , 2.4315327767512956e-45 ], 
                    [ 902.0 , 2.4318432152901925e-45 ], [ 902.0625 , 2.432153693463332e-45 ], [ 902.125 , 2.4324642112758956e-45 ], [ 902.1875 , 2.4327747687328227e-45 ], 
                    [ 902.25 , 2.433085365839262e-45 ], [ 902.3125 , 2.43339600260024e-45 ], [ 902.375 , 2.4337066790207853e-45 ], [ 902.4375 , 2.434017395106014e-45 ], 
                    [ 902.5 , 2.4343281508609716e-45 ], [ 902.5625 , 2.4346389462907234e-45 ], [ 902.625 , 2.434949781400335e-45 ], [ 902.6875 , 2.4352606561948892e-45 ], 
                    [ 902.75 , 2.435571570679436e-45 ], [ 902.8125 , 2.4358825248590076e-45 ], [ 902.875 , 2.4361935187387066e-45 ], [ 902.9375 , 2.4365043943402836e-45 ], 
                    [ 903.0 , 2.4368146775978662e-45 ], [ 903.0625 , 2.4371250003692764e-45 ], [ 903.125 , 2.4374353626595992e-45 ], [ 903.1875 , 2.4377457644738487e-45 ], 
                    [ 903.25 , 2.438056205817076e-45 ], [ 903.3125 , 2.438366686694264e-45 ], [ 903.375 , 2.4386772071105147e-45 ], [ 903.4375 , 2.4389877670707953e-45 ], 
                    [ 903.5 , 2.4392983665801937e-45 ], [ 903.5625 , 2.439609005643728e-45 ], [ 903.625 , 2.4399196842664536e-45 ], [ 903.6875 , 2.440230402453356e-45 ], 
                    [ 903.75 , 2.4405411602095258e-45 ], [ 903.8125 , 2.4408519575400018e-45 ], [ 903.875 , 2.4411627944497716e-45 ], [ 903.9375 , 2.441473670943928e-45 ], 
                    [ 904.0 , 2.4417845870274937e-45 ], [ 904.0625 , 2.442095542705529e-45 ], [ 904.125 , 2.4424065379830232e-45 ], [ 904.1875 , 2.4427175728650543e-45 ], 
                    [ 904.25 , 2.4430286473567175e-45 ], [ 904.3125 , 2.4433397614629533e-45 ], [ 904.375 , 2.4436509151888927e-45 ], [ 904.4375 , 2.4439621085395303e-45 ], 
                    [ 904.5 , 2.4442733415199108e-45 ], [ 904.5625 , 2.4445846141351165e-45 ], [ 904.625 , 2.4448959263901946e-45 ], [ 904.6875 , 2.4452072782901412e-45 ], 
                    [ 904.75 , 2.4455186698400392e-45 ], [ 904.8125 , 2.4458301010449907e-45 ], [ 904.875 , 2.4461415719099585e-45 ], [ 904.9375 , 2.4464530824399935e-45 ], 
                    [ 905.0 , 2.4467646326402334e-45 ], [ 905.0625 , 2.447075871218898e-45 ], [ 905.125 , 2.4473870978153393e-45 ], [ 905.1875 , 2.4476983639945315e-45 ], 
                    [ 905.25 , 2.4480096697615087e-45 ], [ 905.3125 , 2.4483210151213233e-45 ], [ 905.375 , 2.4486324000789937e-45 ], [ 905.4375 , 2.4489438246395375e-45 ], 
                    [ 905.5 , 2.449255288808028e-45 ], [ 905.5625 , 2.4495667925894494e-45 ], [ 905.625 , 2.4498783359889094e-45 ], [ 905.6875 , 2.4501899190114296e-45 ], 
                    [ 905.75 , 2.4505015416619624e-45 ], [ 905.8125 , 2.4508132039456866e-45 ], [ 905.875 , 2.451124905867539e-45 ], [ 905.9375 , 2.4514366474325962e-45 ], 
                    [ 906.0 , 2.4517484286459164e-45 ], [ 906.0625 , 2.4520602495125075e-45 ], [ 906.125 , 2.4523721100374306e-45 ], [ 906.1875 , 2.4526840102257298e-45 ], 
                    [ 906.25 , 2.4529959500824665e-45 ], [ 906.3125 , 2.4533079296126684e-45 ], [ 906.375 , 2.453619948821364e-45 ], [ 906.4375 , 2.4539320077136e-45 ], 
                    [ 906.5 , 2.4542441062944756e-45 ], [ 906.5625 , 2.4545562445690036e-45 ], [ 906.625 , 2.4548684225422327e-45 ], [ 906.6875 , 2.4551806402191587e-45 ], 
                    [ 906.75 , 2.4554928976049367e-45 ], [ 906.8125 , 2.4558051947045644e-45 ], [ 906.875 , 2.4561175315230927e-45 ], [ 906.9375 , 2.456429908065573e-45 ], 
                    [ 907.0 , 2.4567423243370058e-45 ], [ 907.0625 , 2.457054780342548e-45 ], [ 907.125 , 2.4573669570669433e-45 ], [ 907.1875 , 2.4576788278052763e-45 ], 
                    [ 907.25 , 2.4579907381239064e-45 ], [ 907.3125 , 2.4583026880278917e-45 ], [ 907.375 , 2.458614677522239e-45 ], [ 907.4375 , 2.458926706611938e-45 ], 
                    [ 907.5 , 2.4592387753021177e-45 ], [ 907.5625 , 2.4595508835976826e-45 ], [ 907.625 , 2.4598630315037282e-45 ], [ 907.6875 , 2.4601752190252476e-45 ], 
                    [ 907.75 , 2.460487446167285e-45 ], [ 907.8125 , 2.460799712934887e-45 ], [ 907.875 , 2.461112019333047e-45 ], [ 907.9375 , 2.4614243653668302e-45 ], 
                    [ 908.0 , 2.4617367510412313e-45 ], [ 908.0625 , 2.4620491763612815e-45 ], [ 908.125 , 2.4623616413320822e-45 ], [ 908.1875 , 2.4626741459585957e-45 ], 
                    [ 908.25 , 2.462986690245855e-45 ], [ 908.3125 , 2.4632992741989105e-45 ], [ 908.375 , 2.4636118978227965e-45 ], [ 908.4375 , 2.4639245611225653e-45 ], 
                    [ 908.5 , 2.4642372641032174e-45 ], [ 908.5625 , 2.4645500067698236e-45 ], [ 908.625 , 2.4648627891273856e-45 ], [ 908.6875 , 2.4651756111809934e-45 ], 
                    [ 908.75 , 2.465488472935632e-45 ], [ 908.8125 , 2.4658013743963407e-45 ], [ 908.875 , 2.466114315568175e-45 ], [ 908.9375 , 2.4664272964561758e-45 ], 
                    [ 909.0 , 2.4667403170654007e-45 ], [ 909.0625 , 2.4670533774008733e-45 ], [ 909.125 , 2.467366477467618e-45 ], [ 909.1875 , 2.4676796043975196e-45 ], 
                    [ 909.25 , 2.4679922474893905e-45 ], [ 909.3125 , 2.468304930191632e-45 ], [ 909.375 , 2.468617652509262e-45 ], [ 909.4375 , 2.4689304144473168e-45 ], 
                    [ 909.5 , 2.4692432160107823e-45 ], [ 909.5625 , 2.4695560572047307e-45 ], [ 909.625 , 2.469868938034095e-45 ], [ 909.6875 , 2.4701818585039676e-45 ], 
                    [ 909.75 , 2.4704948186193527e-45 ], [ 909.8125 , 2.4708078183852736e-45 ], [ 909.875 , 2.471120857806754e-45 ], [ 909.9375 , 2.4714339368888348e-45 ], 
                    [ 910.0 , 2.471747055636524e-45 ], [ 910.0625 , 2.472060214054847e-45 ], [ 910.125 , 2.4723734121488117e-45 ], [ 910.1875 , 2.472686649923463e-45 ], 
                    [ 910.25 , 2.4729999273838102e-45 ], [ 910.3125 , 2.4733132445349694e-45 ], [ 910.375 , 2.473626601381829e-45 ], [ 910.4375 , 2.4739399979295232e-45 ], 
                    [ 910.5 , 2.4742534341830293e-45 ], [ 910.5625 , 2.4745669101473957e-45 ], [ 910.625 , 2.4748804258276703e-45 ], [ 910.6875 , 2.4751939812288686e-45 ], 
                    [ 910.75 , 2.4755075763559868e-45 ], [ 910.8125 , 2.4758212112141465e-45 ], [ 910.875 , 2.4761348858082578e-45 ], [ 910.9375 , 2.4764486001434788e-45 ], 
                    [ 911.0 , 2.4767623542247915e-45 ], [ 911.0625 , 2.477076148057214e-45 ], [ 911.125 , 2.4773899816458e-45 ], [ 911.1875 , 2.477703854995604e-45 ], 
                    [ 911.25 , 2.4780177681116283e-45 ], [ 911.3125 , 2.4783314054286108e-45 ], [ 911.375 , 2.4786447293864446e-45 ], [ 911.4375 , 2.478958092956357e-45 ], 
                    [ 911.5 , 2.4792714961434094e-45 ], [ 911.5625 , 2.4795849389525565e-45 ], [ 911.625 , 2.4798984213888264e-45 ], [ 911.6875 , 2.480211943457228e-45 ], 
                    [ 911.75 , 2.4805255051627728e-45 ], [ 911.8125 , 2.480839106510471e-45 ], [ 911.875 , 2.4811527475053348e-45 ], [ 911.9375 , 2.481466428152394e-45 ], 
                    [ 912.0 , 2.481780148456627e-45 ], [ 912.0625 , 2.4820939084230464e-45 ], [ 912.125 , 2.48240770805672e-45 ], [ 912.1875 , 2.4827215473626102e-45 ], 
                    [ 912.25 , 2.48303542634575e-45 ], [ 912.3125 , 2.4833493450111554e-45 ], [ 912.375 , 2.4836633033638607e-45 ], [ 912.4375 , 2.4839773014088486e-45 ], 
                    [ 912.5 , 2.4842913391511544e-45 ], [ 912.5625 , 2.484605416595797e-45 ], [ 912.625 , 2.4849195337477787e-45 ], [ 912.6875 , 2.4852336906122067e-45 ], 
                    [ 912.75 , 2.4855478871939438e-45 ], [ 912.8125 , 2.4858621234981517e-45 ], [ 912.875 , 2.4861763995297824e-45 ], [ 912.9375 , 2.486490715293858e-45 ], 
                    [ 913.0 , 2.4868050707954548e-45 ], [ 913.0625 , 2.4871194660395437e-45 ], [ 913.125 , 2.4874339010311843e-45 ], [ 913.1875 , 2.4877483757753665e-45 ], 
                    [ 913.25 , 2.4880628902771515e-45 ], [ 913.3125 , 2.4883774445414947e-45 ], [ 913.375 , 2.488692038573529e-45 ], [ 913.4375 , 2.4890060654138926e-45 ], 
                    [ 913.5 , 2.4893200297595847e-45 ], [ 913.5625 , 2.489634033708917e-45 ], [ 913.625 , 2.4899480772668126e-45 ], [ 913.6875 , 2.490262160438322e-45 ], 
                    [ 913.75 , 2.490576283228406e-45 ], [ 913.8125 , 2.490890445642116e-45 ], [ 913.875 , 2.4912046476843788e-45 ], [ 913.9375 , 2.491518889360264e-45 ], 
                    [ 914.0 , 2.4918331706747176e-45 ], [ 914.0625 , 2.492147491632776e-45 ], [ 914.125 , 2.492461852239386e-45 ], [ 914.1875 , 2.492776252499602e-45 ], 
                    [ 914.25 , 2.4930906924184087e-45 ], [ 914.3125 , 2.4934051720008082e-45 ], [ 914.375 , 2.4937196912518034e-45 ], [ 914.4375 , 2.494034250176399e-45 ], 
                    [ 914.5 , 2.494348848779599e-45 ], [ 914.5625 , 2.4946634870664083e-45 ], [ 914.625 , 2.4949781650418147e-45 ], [ 914.6875 , 2.4952928827108784e-45 ], 
                    [ 914.75 , 2.4956076400785352e-45 ], [ 914.8125 , 2.495922437149881e-45 ], [ 914.875 , 2.4962372739298005e-45 ], [ 914.9375 , 2.496552150423408e-45 ], 
                    [ 915.0 , 2.4968670666356968e-45 ], [ 915.0625 , 2.49718202257164e-45 ], [ 915.125 , 2.4974970182362495e-45 ], [ 915.1875 , 2.4978120536345715e-45 ], 
                    [ 915.25 , 2.4981271287716002e-45 ], [ 915.3125 , 2.4984422436523313e-45 ], [ 915.375 , 2.4987573982818124e-45 ], [ 915.4375 , 2.499072592665041e-45 ], 
                    [ 915.5 , 2.499387730121957e-45 ], [ 915.5625 , 2.4997023325068386e-45 ], [ 915.625 , 2.500016974491283e-45 ], [ 915.6875 , 2.5003316560802564e-45 ], 
                    [ 915.75 , 2.5006463772787795e-45 ], [ 915.8125 , 2.5009611380918027e-45 ], [ 915.875 , 2.501275938524348e-45 ], [ 915.9375 , 2.5015907785813663e-45 ], 
                    [ 916.0 , 2.501905658267863e-45 ], [ 916.0625 , 2.5022205775888272e-45 ], [ 916.125 , 2.5025355365492648e-45 ], [ 916.1875 , 2.5028505351541478e-45 ], 
                    [ 916.25 , 2.503165573408466e-45 ], [ 916.3125 , 2.5034806513171748e-45 ], [ 916.375 , 2.5037957688853368e-45 ], [ 916.4375 , 2.504110926117908e-45 ], 
                    [ 916.5 , 2.504426123019864e-45 ], [ 916.5625 , 2.5047413595962505e-45 ], [ 916.625 , 2.505056635851991e-45 ], [ 916.6875 , 2.5053719517921153e-45 ], 
                    [ 916.75 , 2.5056873074216363e-45 ], [ 916.8125 , 2.5060027027455143e-45 ], [ 916.875 , 2.5063181377687633e-45 ], [ 916.9375 , 2.5066336124963803e-45 ], 
                    [ 917.0 , 2.506949126933381e-45 ], [ 917.0625 , 2.507264681084728e-45 ], [ 917.125 , 2.5075802749554382e-45 ], [ 917.1875 , 2.5078959085504932e-45 ], 
                    [ 917.25 , 2.508211581874982e-45 ], [ 917.3125 , 2.5085272949337808e-45 ], [ 917.375 , 2.5088430477319268e-45 ], [ 917.4375 , 2.509158840274476e-45 ], 
                    [ 917.5 , 2.5094746725663942e-45 ], [ 917.5625 , 2.5097905446126678e-45 ], [ 917.625 , 2.5101062636614146e-45 ], [ 917.6875 , 2.510421691345318e-45 ], 
                    [ 917.75 , 2.5107371586668714e-45 ], [ 917.8125 , 2.5110526656310203e-45 ], [ 917.875 , 2.511368212242728e-45 ], [ 917.9375 , 2.5116837985069945e-45 ], 
                    [ 918.0 , 2.5119994244288034e-45 ], [ 918.0625 , 2.5123150900131376e-45 ], [ 918.125 , 2.512630795264999e-45 ], [ 918.1875 , 2.512946540189319e-45 ], 
                    [ 918.25 , 2.5132623247911368e-45 ], [ 918.3125 , 2.513578149075438e-45 ], [ 918.375 , 2.513894013047173e-45 ], [ 918.4375 , 2.514209916711348e-45 ], 
                    [ 918.5 , 2.5145258600729508e-45 ], [ 918.5625 , 2.5148418431369688e-45 ], [ 918.625 , 2.51515786590841e-45 ], [ 918.6875 , 2.5154739283922462e-45 ], 
                    [ 918.75 , 2.5157900305934314e-45 ], [ 918.8125 , 2.516106172517029e-45 ], [ 918.875 , 2.5164223541679402e-45 ], [ 918.9375 , 2.516738575551319e-45 ], 
                    [ 919.0 , 2.5170548366719254e-45 ], [ 919.0625 , 2.517371137534986e-45 ], [ 919.125 , 2.5176874781452974e-45 ], [ 919.1875 , 2.51800385850798e-45 ], 
                    [ 919.25 , 2.5183202786280113e-45 ], [ 919.3125 , 2.5186367385103517e-45 ], [ 919.375 , 2.5189532381600153e-45 ], [ 919.4375 , 2.519269777582e-45 ], 
                    [ 919.5 , 2.519586356781285e-45 ], [ 919.5625 , 2.5199029757629415e-45 ], [ 919.625 , 2.520219634531878e-45 ], [ 919.6875 , 2.520536333093131e-45 ], 
                    [ 919.75 , 2.5208527819710678e-45 ], [ 919.8125 , 2.5211689946328082e-45 ], [ 919.875 , 2.5214852469598566e-45 ], [ 919.9375 , 2.5218015389571886e-45 ], 
                    [ 920.0 , 2.5221178706298343e-45 ], [ 920.0625 , 2.522434241982699e-45 ], [ 920.125 , 2.5227506530208312e-45 ], [ 920.1875 , 2.5230671037491198e-45 ], 
                    [ 920.25 , 2.523383594172651e-45 ], [ 920.3125 , 2.5237001242963327e-45 ], [ 920.375 , 2.5240166941251437e-45 ], [ 920.4375 , 2.524333303664102e-45 ], 
                    [ 920.5 , 2.524649952918152e-45 ], [ 920.5625 , 2.5249666418923116e-45 ], [ 920.625 , 2.525283370591563e-45 ], [ 920.6875 , 2.5256001390208536e-45 ], 
                    [ 920.75 , 2.525916947185167e-45 ], [ 920.8125 , 2.5262337950895417e-45 ], [ 920.875 , 2.5265506827388906e-45 ], [ 920.9375 , 2.5268676101382527e-45 ], 
                    [ 921.0 , 2.527184577292597e-45 ], [ 921.0625 , 2.527501584206892e-45 ], [ 921.125 , 2.5278186308861787e-45 ], [ 921.1875 , 2.5281357173353918e-45 ], 
                    [ 921.25 , 2.5284528435595373e-45 ], [ 921.3125 , 2.5287700095635698e-45 ], [ 921.375 , 2.5290872153525678e-45 ], [ 921.4375 , 2.5294044609314327e-45 ], 
                    [ 921.5 , 2.5297217463052273e-45 ], [ 921.5625 , 2.5300390714788896e-45 ], [ 921.625 , 2.53035643645743e-45 ], [ 921.6875 , 2.5306738412458238e-45 ], 
                    [ 921.75 , 2.5309912858491005e-45 ], [ 921.8125 , 2.5313087702722362e-45 ], [ 921.875 , 2.531626001318293e-45 ], [ 921.9375 , 2.531943056264477e-45 ], 
                    [ 922.0 , 2.5322601509178662e-45 ], [ 922.0625 , 2.5325772852834515e-45 ], [ 922.125 , 2.5328944593661885e-45 ], [ 922.1875 , 2.533211673171087e-45 ], 
                    [ 922.25 , 2.53352892670314e-45 ], [ 922.3125 , 2.533846219967251e-45 ], [ 922.375 , 2.5341635529684493e-45 ], [ 922.4375 , 2.5344809257116577e-45 ], 
                    [ 922.5 , 2.534798338201926e-45 ], [ 922.5625 , 2.5351157904441773e-45 ], [ 922.625 , 2.5354332824433905e-45 ], [ 922.6875 , 2.535750814204616e-45 ], 
                    [ 922.75 , 2.5360683857327448e-45 ], [ 922.8125 , 2.5363859970327918e-45 ], [ 922.875 , 2.5367036481097385e-45 ], [ 922.9375 , 2.5370213389685664e-45 ], 
                    [ 923.0 , 2.5373390696142582e-45 ], [ 923.0625 , 2.5376568400518325e-45 ], [ 923.125 , 2.5379746502861823e-45 ], [ 923.1875 , 2.5382925003223817e-45 ], 
                    [ 923.25 , 2.5386103901653256e-45 ], [ 923.3125 , 2.5389283198201435e-45 ], [ 923.375 , 2.539246289291641e-45 ], [ 923.4375 , 2.539564298584949e-45 ], 
                    [ 923.5 , 2.5398823477049826e-45 ], [ 923.5625 , 2.540200436656783e-45 ], [ 923.625 , 2.5405185654452677e-45 ], [ 923.6875 , 2.540836734075443e-45 ], 
                    [ 923.75 , 2.5411549425523344e-45 ], [ 923.8125 , 2.541473190880915e-45 ], [ 923.875 , 2.5417914790661934e-45 ], [ 923.9375 , 2.5421098071131427e-45 ], 
                    [ 924.0 , 2.5424275893233846e-45 ], [ 924.0625 , 2.542744980062174e-45 ], [ 924.125 , 2.5430624104233387e-45 ], [ 924.1875 , 2.543379880411734e-45 ], 
                    [ 924.25 , 2.5436973900323435e-45 ], [ 924.3125 , 2.544014939290096e-45 ], [ 924.375 , 2.5443325281899948e-45 ], [ 924.4375 , 2.5446501567369338e-45 ], 
                    [ 924.5 , 2.544967824935845e-45 ], [ 924.5625 , 2.5452855327917682e-45 ], [ 924.625 , 2.5456032803096004e-45 ], [ 924.6875 , 2.5459210674942566e-45 ], 
                    [ 924.75 , 2.5462388943507606e-45 ], [ 924.8125 , 2.546556760883993e-45 ], [ 924.875 , 2.5468746670989795e-45 ], [ 924.9375 , 2.547192613000601e-45 ], 
                    [ 925.0 , 2.5475105985938668e-45 ], [ 925.0625 , 2.5478286238837134e-45 ], [ 925.125 , 2.5481466888750963e-45 ], [ 925.1875 , 2.548464793572954e-45 ], 
                    [ 925.25 , 2.5487829379822792e-45 ], [ 925.3125 , 2.5491011221080295e-45 ], [ 925.375 , 2.549419345955127e-45 ], [ 925.4375 , 2.5497376095285487e-45 ], 
                    [ 925.5 , 2.5500559128332354e-45 ], [ 925.5625 , 2.5503742558742015e-45 ], [ 925.625 , 2.550692638656353e-45 ], [ 925.6875 , 2.551011061184688e-45 ], 
                    [ 925.75 , 2.5513295234641137e-45 ], [ 925.8125 , 2.5516480254996464e-45 ], [ 925.875 , 2.5519665672962134e-45 ], [ 925.9375 , 2.5522851488588147e-45 ], 
                    [ 926.0 , 2.5526037701923958e-45 ], [ 926.0625 , 2.5529224313019404e-45 ], [ 926.125 , 2.5532408773977446e-45 ], [ 926.1875 , 2.5535590871539563e-45 ], 
                    [ 926.25 , 2.553877336568532e-45 ], [ 926.3125 , 2.5541956256464504e-45 ], [ 926.375 , 2.5545139543926734e-45 ], [ 926.4375 , 2.554832322812108e-45 ], 
                    [ 926.5 , 2.555150730909754e-45 ], [ 926.5625 , 2.5554691786904828e-45 ], [ 926.625 , 2.5557876661592952e-45 ], [ 926.6875 , 2.556106193321083e-45 ], 
                    [ 926.75 , 2.5564247601808657e-45 ], [ 926.8125 , 2.556743366743572e-45 ], [ 926.875 , 2.5570620130141152e-45 ], [ 926.9375 , 2.557380698997443e-45 ], 
                    [ 927.0 , 2.557699424698523e-45 ], [ 927.0625 , 2.558018190122342e-45 ], [ 927.125 , 2.5583369952737777e-45 ], [ 927.1875 , 2.5586558401577995e-45 ], 
                    [ 927.25 , 2.5589747247794504e-45 ], [ 927.3125 , 2.5592936491435368e-45 ], [ 927.375 , 2.559612613255158e-45 ], [ 927.4375 , 2.559931617119122e-45 ], 
                    [ 927.5 , 2.5602506607405105e-45 ], [ 927.5625 , 2.560569744124187e-45 ], [ 927.625 , 2.5608888672751802e-45 ], [ 927.6875 , 2.5612080301983376e-45 ], 
                    [ 927.75 , 2.5615272328987606e-45 ], [ 927.8125 , 2.5618464753812803e-45 ], [ 927.875 , 2.562165757650963e-45 ], [ 927.9375 , 2.5624850797126767e-45 ], 
                    [ 928.0 , 2.5628044415714173e-45 ], [ 928.0625 , 2.5631238432321625e-45 ], [ 928.125 , 2.563443284699855e-45 ], [ 928.1875 , 2.5637627659794558e-45 ], 
                    [ 928.25 , 2.5640821702302682e-45 ], [ 928.3125 , 2.5644013467254538e-45 ], [ 928.375 , 2.564720562951654e-45 ], [ 928.4375 , 2.5650398189138343e-45 ], 
                    [ 928.5 , 2.565359114616958e-45 ], [ 928.5625 , 2.5656784500659733e-45 ], [ 928.625 , 2.5659978252657717e-45 ], [ 928.6875 , 2.5663172402213388e-45 ], 
                    [ 928.75 , 2.566636694937623e-45 ], [ 928.8125 , 2.5669561894195737e-45 ], [ 928.875 , 2.5672757236721595e-45 ], [ 928.9375 , 2.567595297700294e-45 ], 
                    [ 929.0 , 2.5679149115089657e-45 ], [ 929.0625 , 2.5682345651031075e-45 ], [ 929.125 , 2.5685542584876542e-45 ], [ 929.1875 , 2.568873991667595e-45 ], 
                    [ 929.25 , 2.5691937646478288e-45 ], [ 929.3125 , 2.569513577433365e-45 ], [ 929.375 , 2.5698334300291767e-45 ], [ 929.4375 , 2.5701533224401465e-45 ], 
                    [ 929.5 , 2.5704732546713036e-45 ], [ 929.5625 , 2.5707932267275135e-45 ], [ 929.625 , 2.5711132386138063e-45 ], [ 929.6875 , 2.5714332903351592e-45 ], 
                    [ 929.75 , 2.5717533818964385e-45 ], [ 929.8125 , 2.5720735133026955e-45 ], [ 929.875 , 2.5723936845588535e-45 ], [ 929.9375 , 2.5727138956698542e-45 ], 
                    [ 930.0 , 2.573034146640678e-45 ], [ 930.0625 , 2.573354437476304e-45 ], [ 930.125 , 2.5736747681816762e-45 ], [ 930.1875 , 2.57399513876174e-45 ], 
                    [ 930.25 , 2.5743155492214953e-45 ], [ 930.3125 , 2.574635999565888e-45 ], [ 930.375 , 2.5749564434345804e-45 ], [ 930.4375 , 2.575276249009982e-45 ], 
                    [ 930.5 , 2.5755960943047557e-45 ], [ 930.5625 , 2.575915979323798e-45 ], [ 930.625 , 2.5762359040720794e-45 ], [ 930.6875 , 2.576555868554497e-45 ], 
                    [ 930.75 , 2.5768758727760413e-45 ], [ 930.8125 , 2.5771959167415566e-45 ], [ 930.875 , 2.5775160004560875e-45 ], [ 930.9375 , 2.5778361239244623e-45 ], 
                    [ 931.0 , 2.5781562871517274e-45 ], [ 931.0625 , 2.5784764901427112e-45 ], [ 931.125 , 2.578796732902462e-45 ], [ 931.1875 , 2.5791170154358277e-45 ], 
                    [ 931.25 , 2.579437337747785e-45 ], [ 931.3125 , 2.5797576998432924e-45 ], [ 931.375 , 2.5800781017272536e-45 ], [ 931.4375 , 2.5803985434046478e-45 ], 
                    [ 931.5 , 2.5807190248803804e-45 ], [ 931.5625 , 2.581039546159431e-45 ], [ 931.625 , 2.5813601072467246e-45 ], [ 931.6875 , 2.581680708147187e-45 ], 
                    [ 931.75 , 2.5820013488657816e-45 ], [ 931.8125 , 2.5823220294074534e-45 ], [ 931.875 , 2.582642749777185e-45 ], [ 931.9375 , 2.58296350997985e-45 ], 
                    [ 932.0 , 2.583284310020432e-45 ], [ 932.0625 , 2.5836051499038972e-45 ], [ 932.125 , 2.5839260296351573e-45 ], [ 932.1875 , 2.5842469492191982e-45 ], 
                    [ 932.25 , 2.5845679086609325e-45 ], [ 932.3125 , 2.5848889079653658e-45 ], [ 932.375 , 2.5852099471373573e-45 ], [ 932.4375 , 2.585531026181968e-45 ], 
                    [ 932.5 , 2.5858521451040408e-45 ], [ 932.5625 , 2.5861728039197147e-45 ], [ 932.625 , 2.5864933156665497e-45 ], [ 932.6875 , 2.586813867135298e-45 ], 
                    [ 932.75 , 2.587134458330901e-45 ], [ 932.8125 , 2.5874550892583188e-45 ], [ 932.875 , 2.5877757599223836e-45 ], [ 932.9375 , 2.588096470328094e-45 ], 
                    [ 933.0 , 2.5884172204803747e-45 ], [ 933.0625 , 2.588738010384115e-45 ], [ 933.125 , 2.589058840044279e-45 ], [ 933.1875 , 2.589379709465774e-45 ], 
                    [ 933.25 , 2.5897006186534922e-45 ], [ 933.3125 , 2.5900215676124528e-45 ], [ 933.375 , 2.5903425563474938e-45 ], [ 933.4375 , 2.5906635848635813e-45 ], 
                    [ 933.5 , 2.590984653165645e-45 ], [ 933.5625 , 2.591305761258616e-45 ], [ 933.625 , 2.5916269091474626e-45 ], [ 933.6875 , 2.591948096837025e-45 ], 
                    [ 933.75 , 2.592269324332309e-45 ], [ 933.8125 , 2.5925905916382295e-45 ], [ 933.875 , 2.5929118987597577e-45 ], [ 933.9375 , 2.5932332457017716e-45 ], 
                    [ 934.0 , 2.593554632469226e-45 ], [ 934.0625 , 2.5938760590670192e-45 ], [ 934.125 , 2.5941975255001792e-45 ], [ 934.1875 , 2.5945190317735332e-45 ], 
                    [ 934.25 , 2.5948405778920917e-45 ], [ 934.3125 , 2.5951621638607933e-45 ], [ 934.375 , 2.5954837896845763e-45 ], [ 934.4375 , 2.595805455368288e-45 ], 
                    [ 934.5 , 2.596127160916998e-45 ], [ 934.5625 , 2.5964489063355905e-45 ], [ 934.625 , 2.596770691628971e-45 ], [ 934.6875 , 2.5970922846956944e-45 ], 
                    [ 934.75 , 2.597413441995588e-45 ], [ 934.8125 , 2.5977346390099007e-45 ], [ 934.875 , 2.598055875743525e-45 ], [ 934.9375 , 2.5983771522013913e-45 ], 
                    [ 935.0 , 2.59869846838843e-45 ], [ 935.0625 , 2.5990198243095175e-45 ], [ 935.125 , 2.5993412199696038e-45 ], [ 935.1875 , 2.5996626553735663e-45 ], 
                    [ 935.25 , 2.599984130526357e-45 ], [ 935.3125 , 2.6003056454328725e-45 ], [ 935.375 , 2.600627200098029e-45 ], [ 935.4375 , 2.6009487945267055e-45 ], 
                    [ 935.5 , 2.6012704287238938e-45 ], [ 935.5625 , 2.6015921026944555e-45 ], [ 935.625 , 2.6019138164433467e-45 ], [ 935.6875 , 2.6022355699754678e-45 ], 
                    [ 935.75 , 2.6025573632957012e-45 ], [ 935.8125 , 2.6028791964090593e-45 ], [ 935.875 , 2.603201069320371e-45 ], [ 935.9375 , 2.6035229820346312e-45 ], 
                    [ 936.0 , 2.60384493455667e-45 ], [ 936.0625 , 2.604166926891539e-45 ], [ 936.125 , 2.604488959043977e-45 ], [ 936.1875 , 2.6048110310191295e-45 ], 
                    [ 936.25 , 2.6051331428217362e-45 ], [ 936.3125 , 2.6054552944568328e-45 ], [ 936.375 , 2.605777485929327e-45 ], [ 936.4375 , 2.6060997172440894e-45 ], 
                    [ 936.5 , 2.6064219884061022e-45 ], [ 936.5625 , 2.6067442994202747e-45 ], [ 936.625 , 2.6070666502914792e-45 ], [ 936.6875 , 2.6073890410247746e-45 ], 
                    [ 936.75 , 2.6077114716249783e-45 ], [ 936.8125 , 2.608033942097077e-45 ], [ 936.875 , 2.6083558672777628e-45 ], [ 936.9375 , 2.6086777272940688e-45 ], 
                    [ 937.0 , 2.6089996270264938e-45 ], [ 937.0625 , 2.6093215664799583e-45 ], [ 937.125 , 2.6096435456594182e-45 ], [ 937.1875 , 2.609965564569702e-45 ], 
                    [ 937.25 , 2.610287623215805e-45 ], [ 937.3125 , 2.6106097216025377e-45 ], [ 937.375 , 2.6109318597348406e-45 ], [ 937.4375 , 2.6112540376176184e-45 ], 
                    [ 937.5 , 2.6115762552557948e-45 ], [ 937.5625 , 2.6118985126542568e-45 ], [ 937.625 , 2.612220809817874e-45 ], [ 937.6875 , 2.6125431467516262e-45 ], 
                    [ 937.75 , 2.6128655234603667e-45 ], [ 937.8125 , 2.6131879399490213e-45 ], [ 937.875 , 2.6135103962225175e-45 ], [ 937.9375 , 2.613832892285727e-45 ], 
                    [ 938.0 , 2.6141554281435972e-45 ], [ 938.0625 , 2.6144780038010013e-45 ], [ 938.125 , 2.614800619262869e-45 ], [ 938.1875 , 2.615123274534094e-45 ], 
                    [ 938.25 , 2.6154459696196622e-45 ], [ 938.3125 , 2.6157687045244122e-45 ], [ 938.375 , 2.6160914792532392e-45 ], [ 938.4375 , 2.6164142938111133e-45 ], 
                    [ 938.5 , 2.61673714820293e-45 ], [ 938.5625 , 2.6170600424336238e-45 ], [ 938.625 , 2.617382976508092e-45 ], [ 938.6875 , 2.617705950431232e-45 ], 
                    [ 938.75 , 2.6180289642079804e-45 ], [ 938.8125 , 2.618352017843254e-45 ], [ 938.875 , 2.6186751113419723e-45 ], [ 938.9375 , 2.6189982447090903e-45 ], 
                    [ 939.0 , 2.6193213459796816e-45 ], [ 939.0625 , 2.6196442944619535e-45 ], [ 939.125 , 2.6199672827620857e-45 ], [ 939.1875 , 2.620290310884969e-45 ], 
                    [ 939.25 , 2.6206133788355316e-45 ], [ 939.3125 , 2.620936486618665e-45 ], [ 939.375 , 2.6212596342392437e-45 ], [ 939.4375 , 2.621582821702254e-45 ], 
                    [ 939.5 , 2.6219060490125526e-45 ], [ 939.5625 , 2.6222293161750893e-45 ], [ 939.625 , 2.622552623194722e-45 ], [ 939.6875 , 2.6228759700764572e-45 ], 
                    [ 939.75 , 2.6231993568250802e-45 ], [ 939.8125 , 2.6235227834456368e-45 ], [ 939.875 , 2.6238462499429865e-45 ], [ 939.9375 , 2.6241697563220278e-45 ], 
                    [ 940.0 , 2.6244933025876963e-45 ], [ 940.0625 , 2.6248168887449098e-45 ], [ 940.125 , 2.6251405147985865e-45 ], [ 940.1875 , 2.6254641807536276e-45 ], 
                    [ 940.25 , 2.6257878866150452e-45 ], [ 940.3125 , 2.6261116323875914e-45 ], [ 940.375 , 2.6264354180763373e-45 ], [ 940.4375 , 2.626759243686129e-45 ], 
                    [ 940.5 , 2.6270831092219073e-45 ], [ 940.5625 , 2.6274070146886137e-45 ], [ 940.625 , 2.6277309600911155e-45 ], [ 940.6875 , 2.6280549454344108e-45 ], 
                    [ 940.75 , 2.628378970723369e-45 ], [ 940.8125 , 2.6287030359629702e-45 ], [ 940.875 , 2.6290271411580474e-45 ], [ 940.9375 , 2.6293512863136197e-45 ], 
                    [ 941.0 , 2.62967547143454e-45 ], [ 941.0625 , 2.6299996965258103e-45 ], [ 941.125 , 2.630323961592302e-45 ], [ 941.1875 , 2.630647626836395e-45 ], 
                    [ 941.25 , 2.630970959318927e-45 ], [ 941.3125 , 2.6312943315422034e-45 ], [ 941.375 , 2.631617743511127e-45 ], [ 941.4375 , 2.6319411952305458e-45 ], 
                    [ 941.5 , 2.632264686705364e-45 ], [ 941.5625 , 2.632588217940506e-45 ], [ 941.625 , 2.6329117889407642e-45 ], [ 941.6875 , 2.6332353997111202e-45 ], 
                    [ 941.75 , 2.6335590502564247e-45 ], [ 941.8125 , 2.6338827405815476e-45 ], [ 941.875 , 2.634206470691397e-45 ], [ 941.9375 , 2.6345302405909003e-45 ], 
                    [ 942.0 , 2.634854050284836e-45 ], [ 942.0625 , 2.6351778997782254e-45 ], [ 942.125 , 2.6355017890758862e-45 ], [ 942.1875 , 2.635825718182748e-45 ], 
                    [ 942.25 , 2.6361496871036662e-45 ], [ 942.3125 , 2.636473695843572e-45 ], [ 942.375 , 2.6367977444073404e-45 ], [ 942.4375 , 2.6371218327998476e-45 ], 
                    [ 942.5 , 2.637445961025989e-45 ], [ 942.5625 , 2.637770129090698e-45 ], [ 942.625 , 2.6380943369988337e-45 ], [ 942.6875 , 2.6384185847553683e-45 ], 
                    [ 942.75 , 2.638742872365143e-45 ], [ 942.8125 , 2.6390671998330004e-45 ], [ 942.875 , 2.6393915671639517e-45 ], [ 942.9375 , 2.639715974362803e-45 ], 
                    [ 943.0 , 2.6400404214344907e-45 ], [ 943.0625 , 2.640364908383936e-45 ], [ 943.125 , 2.6406894352159644e-45 ], [ 943.1875 , 2.6410140019356092e-45 ], 
                    [ 943.25 , 2.6413386085476603e-45 ], [ 943.3125 , 2.6416632550570586e-45 ], [ 943.375 , 2.6419873071251697e-45 ], [ 943.4375 , 2.6423113377285634e-45 ], 
                    [ 943.5 , 2.6426354080732083e-45 ], [ 943.5625 , 2.6429595181639593e-45 ], [ 943.625 , 2.6432836680056924e-45 ], [ 943.6875 , 2.643607857603282e-45 ], 
                    [ 943.75 , 2.6439320869616043e-45 ], [ 943.8125 , 2.6442563560855358e-45 ], [ 943.875 , 2.644580664979897e-45 ], [ 943.9375 , 2.644905013649697e-45 ], 
                    [ 944.0 , 2.645229402099702e-45 ], [ 944.0625 , 2.645553830334866e-45 ], [ 944.125 , 2.6458782983599928e-45 ], [ 944.1875 , 2.646202806180001e-45 ], 
                    [ 944.25 , 2.64652735379977e-45 ], [ 944.3125 , 2.646851941224201e-45 ], [ 944.375 , 2.6471765684581564e-45 ], [ 944.4375 , 2.6475012355065003e-45 ], 
                    [ 944.5 , 2.647825942374135e-45 ], [ 944.5625 , 2.6481506890659807e-45 ], [ 944.625 , 2.6484754755868473e-45 ], [ 944.6875 , 2.6488003019416758e-45 ], 
                    [ 944.75 , 2.6491251681353145e-45 ], [ 944.8125 , 2.6494500741726678e-45 ], [ 944.875 , 2.6497750200586414e-45 ], [ 944.9375 , 2.650100005798085e-45 ], 
                    [ 945.0 , 2.6504250313959053e-45 ], [ 945.0625 , 2.6507500968570096e-45 ], [ 945.125 , 2.651075202186268e-45 ], [ 945.1875 , 2.6514003473885706e-45 ], 
                    [ 945.25 , 2.651725532468807e-45 ], [ 945.3125 , 2.6520507574318877e-45 ], [ 945.375 , 2.6523760222826474e-45 ], [ 945.4375 , 2.6527013270260345e-45 ], 
                    [ 945.5 , 2.653026645732399e-45 ], [ 945.5625 , 2.6533517600882475e-45 ], [ 945.625 , 2.653676914285137e-45 ], [ 945.6875 , 2.654002108327969e-45 ], 
                    [ 945.75 , 2.6543273422216265e-45 ], [ 945.8125 , 2.6546526159709365e-45 ], [ 945.875 , 2.6549779295808016e-45 ], [ 945.9375 , 2.6553032830561825e-45 ], 
                    [ 946.0 , 2.6556286764018887e-45 ], [ 946.0625 , 2.6559541096228248e-45 ], [ 946.125 , 2.6562795827238778e-45 ], [ 946.1875 , 2.6566050957098967e-45 ], 
                    [ 946.25 , 2.6569306485858638e-45 ], [ 946.3125 , 2.6572562413565532e-45 ], [ 946.375 , 2.6575818740269495e-45 ], [ 946.4375 , 2.6579075466019042e-45 ], 
                    [ 946.5 , 2.6582332590862874e-45 ], [ 946.5625 , 2.658559011485048e-45 ], [ 946.625 , 2.6588848038030007e-45 ], [ 946.6875 , 2.6592106360450754e-45 ], 
                    [ 946.75 , 2.6595365082161646e-45 ], [ 946.8125 , 2.659862420321199e-45 ], [ 946.875 , 2.6601883723650164e-45 ], [ 946.9375 , 2.660514364352548e-45 ], 
                    [ 947.0 , 2.6608403962885944e-45 ], [ 947.0625 , 2.6611664681781837e-45 ], [ 947.125 , 2.6614925800261546e-45 ], [ 947.1875 , 2.6618187318374235e-45 ], 
                    [ 947.25 , 2.6621449236168497e-45 ], [ 947.3125 , 2.662471155369312e-45 ], [ 947.375 , 2.6627974270998032e-45 ], [ 947.4375 , 2.663123738813147e-45 ], 
                    [ 947.5 , 2.6634500905142435e-45 ], [ 947.5625 , 2.66377648220803e-45 ], [ 947.625 , 2.6641029138993698e-45 ], [ 947.6875 , 2.664429165813205e-45 ], 
                    [ 947.75 , 2.6647546598629962e-45 ], [ 947.8125 , 2.6650801936760613e-45 ], [ 947.875 , 2.6654057672572013e-45 ], [ 947.9375 , 2.6657313806113498e-45 ], 
                    [ 948.0 , 2.6660570337433088e-45 ], [ 948.0625 , 2.666382726657976e-45 ], [ 948.125 , 2.666708459360192e-45 ], [ 948.1875 , 2.6670342318547986e-45 ], 
                    [ 948.25 , 2.6673600441466757e-45 ], [ 948.3125 , 2.6676858962407233e-45 ], [ 948.375 , 2.6680117881417274e-45 ], [ 948.4375 , 2.6683377198545895e-45 ], 
                    [ 948.5 , 2.6686636913841727e-45 ], [ 948.5625 , 2.668989702735341e-45 ], [ 948.625 , 2.6693157539129783e-45 ], [ 948.6875 , 2.669641844921931e-45 ], 
                    [ 948.75 , 2.669967975767046e-45 ], [ 948.8125 , 2.6702941464532274e-45 ], [ 948.875 , 2.670620356985286e-45 ], [ 948.9375 , 2.670946607368165e-45 ], 
                    [ 949.0 , 2.6712728976067327e-45 ], [ 949.0625 , 2.6715992277057443e-45 ], [ 949.125 , 2.6719255976701826e-45 ], [ 949.1875 , 2.6722520075048424e-45 ], 
                    [ 949.25 , 2.672578457214651e-45 ], [ 949.3125 , 2.6729049468044607e-45 ], [ 949.375 , 2.6732314762791816e-45 ], [ 949.4375 , 2.673558045643609e-45 ], 
                    [ 949.5 , 2.673884654902675e-45 ], [ 949.5625 , 2.674211304061214e-45 ], [ 949.625 , 2.6745379931241183e-45 ], [ 949.6875 , 2.674864722096284e-45 ], 
                    [ 949.75 , 2.6751914909825462e-45 ], [ 949.8125 , 2.6755182997878207e-45 ], [ 949.875 , 2.6758450450919753e-45 ], [ 949.9375 , 2.6761715296841255e-45 ], 
                    [ 950.0 , 2.67649805411122e-45 ], [ 950.0625 , 2.676824618378176e-45 ], [ 950.125 , 2.6771512224897784e-45 ], [ 950.1875 , 2.677477866450946e-45 ], 
                    [ 950.25 , 2.6778045502665025e-45 ], [ 950.3125 , 2.6781312739413677e-45 ], [ 950.375 , 2.678458037480329e-45 ], [ 950.4375 , 2.6787848408883073e-45 ], 
                    [ 950.5 , 2.6791116841701295e-45 ], [ 950.5625 , 2.679438567330679e-45 ], [ 950.625 , 2.6797654903748407e-45 ], [ 950.6875 , 2.680092453307481e-45 ], 
                    [ 950.75 , 2.6804194561334288e-45 ], [ 950.8125 , 2.680746498857551e-45 ], [ 950.875 , 2.6810735814847352e-45 ], [ 950.9375 , 2.681400704019869e-45 ], 
                    [ 951.0 , 2.6817278664678024e-45 ], [ 951.0625 , 2.6820550688334057e-45 ], [ 951.125 , 2.68238231112153e-45 ], [ 951.1875 , 2.682709593337046e-45 ], 
                    [ 951.25 , 2.6830369154849022e-45 ], [ 951.3125 , 2.6833642775699128e-45 ], [ 951.375 , 2.6836916795969322e-45 ], [ 951.4375 , 2.6840191215708337e-45 ], 
                    [ 951.5 , 2.68434660349651e-45 ], [ 951.5625 , 2.684674125378855e-45 ], [ 951.625 , 2.685001687222725e-45 ], [ 951.6875 , 2.6853292890329764e-45 ], 
                    [ 951.75 , 2.6856569308145433e-45 ], [ 951.8125 , 2.6859846125721874e-45 ], [ 951.875 , 2.6863123343109585e-45 ], [ 951.9375 , 2.686640096035563e-45 ], 
                    [ 952.0 , 2.6869678977509938e-45 ], [ 952.0625 , 2.687295565132085e-45 ], [ 952.125 , 2.687622575066087e-45 ], [ 952.1875 , 2.6879496247931066e-45 ], 
                    [ 952.25 , 2.6882767143179482e-45 ], [ 952.3125 , 2.688603843645416e-45 ], [ 952.375 , 2.688931012780449e-45 ], [ 952.4375 , 2.689258221727796e-45 ], 
                    [ 952.5 , 2.6895854704923593e-45 ], [ 952.5625 , 2.6899127590789255e-45 ], [ 952.625 , 2.6902400874924376e-45 ], [ 952.6875 , 2.6905674557376455e-45 ], 
                    [ 952.75 , 2.6908948638194538e-45 ], [ 952.8125 , 2.691222311742672e-45 ], [ 952.875 , 2.6915497995121673e-45 ], [ 952.9375 , 2.6918773271328076e-45 ], 
                    [ 953.0 , 2.6922048946094042e-45 ], [ 953.0625 , 2.6925325019468447e-45 ], [ 953.125 , 2.6928601491499805e-45 ], [ 953.1875 , 2.693187836223605e-45 ], 
                    [ 953.25 , 2.693515563172589e-45 ], [ 953.3125 , 2.6938433300018425e-45 ], [ 953.375 , 2.6941711367161797e-45 ], [ 953.4375 , 2.6944989833204353e-45 ], 
                    [ 953.5 , 2.6948268698195016e-45 ], [ 953.5625 , 2.695154796218176e-45 ], [ 953.625 , 2.695482762521409e-45 ], [ 953.6875 , 2.6958107687339617e-45 ], 
                    [ 953.75 , 2.696138814860747e-45 ], [ 953.8125 , 2.6964669009065646e-45 ], [ 953.875 , 2.696795026876388e-45 ], [ 953.9375 , 2.6971231927749215e-45 ], 
                    [ 954.0 , 2.6974513986071384e-45 ], [ 954.0625 , 2.6977796443778612e-45 ], [ 954.125 , 2.69810793009191e-45 ], [ 954.1875 , 2.698436255754243e-45 ], 
                    [ 954.25 , 2.6987645872326526e-45 ], [ 954.3125 , 2.6990925660378415e-45 ], [ 954.375 , 2.699420584702006e-45 ], [ 954.4375 , 2.6997486432300867e-45 ], 
                    [ 954.5 , 2.700076741626832e-45 ], [ 954.5625 , 2.700404879897107e-45 ], [ 954.625 , 2.700733058045833e-45 ], [ 954.6875 , 2.7010612760777614e-45 ], 
                    [ 954.75 , 2.7013895339977967e-45 ], [ 954.8125 , 2.7017178318107474e-45 ], [ 954.875 , 2.7020461695214818e-45 ], [ 954.9375 , 2.7023745471348476e-45 ], 
                    [ 955.0 , 2.7027029646557336e-45 ], [ 955.0625 , 2.7030314220888927e-45 ], [ 955.125 , 2.7033599194392918e-45 ], [ 955.1875 , 2.7036884567116845e-45 ], 
                    [ 955.25 , 2.704017033910982e-45 ], [ 955.3125 , 2.704345651041996e-45 ], [ 955.375 , 2.7046743081096004e-45 ], [ 955.4375 , 2.705003005118667e-45 ], 
                    [ 955.5 , 2.705331742074031e-45 ], [ 955.5625 , 2.7056605189805272e-45 ], [ 955.625 , 2.70598933584305e-45 ], [ 955.6875 , 2.706318192666455e-45 ], 
                    [ 955.75 , 2.7066470894555215e-45 ], [ 955.8125 , 2.7069760262151834e-45 ], [ 955.875 , 2.7073050029502988e-45 ], [ 955.9375 , 2.7076340196657057e-45 ], 
                    [ 956.0 , 2.707963076366245e-45 ], [ 956.0625 , 2.708292173056794e-45 ], [ 956.125 , 2.7086213097422326e-45 ], [ 956.1875 , 2.7089504864273832e-45 ], 
                    [ 956.25 , 2.709279703117125e-45 ], [ 956.3125 , 2.7096089598163398e-45 ], [ 956.375 , 2.7099382565298705e-45 ], [ 956.4375 , 2.7102675932625418e-45 ], 
                    [ 956.5 , 2.7105962326329533e-45 ], [ 956.5625 , 2.7109248112883298e-45 ], [ 956.625 , 2.711253429774071e-45 ], [ 956.6875 , 2.711582088094948e-45 ], 
                    [ 956.75 , 2.7119107862558088e-45 ], [ 956.8125 , 2.7122395242614816e-45 ], [ 956.875 , 2.7125683021167982e-45 ], [ 956.9375 , 2.7128971198265877e-45 ], 
                    [ 957.0 , 2.713225977395701e-45 ], [ 957.0625 , 2.7135548748289123e-45 ], [ 957.125 , 2.7138838121311116e-45 ], [ 957.1875 , 2.714212789307132e-45 ], 
                    [ 957.25 , 2.7145418063617682e-45 ], [ 957.3125 , 2.714870863299854e-45 ], [ 957.375 , 2.7151999601262826e-45 ], [ 957.4375 , 2.7155290968458115e-45 ], 
                    [ 957.5 , 2.7158582734633733e-45 ], [ 957.5625 , 2.716187489983689e-45 ], [ 957.625 , 2.7165167464116915e-45 ], [ 957.6875 , 2.7168460427521416e-45 ], 
                    [ 957.75 , 2.717175379009935e-45 ], [ 957.8125 , 2.7175047551899112e-45 ], [ 957.875 , 2.71783417129687e-45 ], [ 957.9375 , 2.7181636273356523e-45 ], 
                    [ 958.0 , 2.718493123311117e-45 ], [ 958.0625 , 2.7188226592281453e-45 ], [ 958.125 , 2.7191522350915192e-45 ], [ 958.1875 , 2.7194818509061014e-45 ], 
                    [ 958.25 , 2.719811506676754e-45 ], [ 958.3125 , 2.7201412024083007e-45 ], [ 958.375 , 2.720470938105547e-45 ], [ 958.4375 , 2.7208007137733965e-45 ], 
                    [ 958.5 , 2.721130529416674e-45 ], [ 958.5625 , 2.7214603850402257e-45 ], [ 958.625 , 2.7217902806488788e-45 ], [ 958.6875 , 2.7221197686828326e-45 ], 
                    [ 958.75 , 2.7224489777514065e-45 ], [ 958.8125 , 2.7227782266339918e-45 ], [ 958.875 , 2.7231075153354603e-45 ], [ 958.9375 , 2.7234368438606093e-45 ], 
                    [ 959.0 , 2.7237662122142542e-45 ], [ 959.0625 , 2.7240956204012124e-45 ], [ 959.125 , 2.7244250684263008e-45 ], [ 959.1875 , 2.7247545562943185e-45 ], 
                    [ 959.25 , 2.7250840840101224e-45 ], [ 959.3125 , 2.7254136515785122e-45 ], [ 959.375 , 2.7257432590042887e-45 ], [ 959.4375 , 2.7260729062923104e-45 ], 
                    [ 959.5 , 2.7264025934473985e-45 ], [ 959.5625 , 2.7267323204743358e-45 ], [ 959.625 , 2.7270620873779437e-45 ], [ 959.6875 , 2.7273918941630845e-45 ], 
                    [ 959.75 , 2.727721740834581e-45 ], [ 959.8125 , 2.728051627397199e-45 ], [ 959.875 , 2.7283815538558008e-45 ], [ 959.9375 , 2.7287115202152512e-45 ], 
                    [ 960.0 , 2.7290415264802784e-45 ], [ 960.0625 , 2.729371572655825e-45 ], [ 960.125 , 2.729701658746601e-45 ], [ 960.1875 , 2.7300317847575895e-45 ], 
                    [ 960.25 , 2.7303619506934234e-45 ], [ 960.3125 , 2.730692156559068e-45 ], [ 960.375 , 2.7310224023593322e-45 ], [ 960.4375 , 2.731352688098988e-45 ], 
                    [ 960.5 , 2.7316830137829235e-45 ], [ 960.5625 , 2.7320133794159497e-45 ], [ 960.625 , 2.7323437850029376e-45 ], [ 960.6875 , 2.7326742305485826e-45 ], 
                    [ 960.75 , 2.7330047160578733e-45 ], [ 960.8125 , 2.7333352415355842e-45 ], [ 960.875 , 2.7336657559653617e-45 ], [ 960.9375 , 2.7339961160642915e-45 ], 
                    [ 961.0 , 2.7343265160868072e-45 ], [ 961.0625 , 2.7346569560377932e-45 ], [ 961.125 , 2.7349874359219968e-45 ], [ 961.1875 , 2.7353179557443403e-45 ], 
                    [ 961.25 , 2.735648515509534e-45 ], [ 961.3125 , 2.735979115222483e-45 ], [ 961.375 , 2.7363097548880337e-45 ], [ 961.4375 , 2.7366404345109177e-45 ], 
                    [ 961.5 , 2.7369711540960412e-45 ], [ 961.5625 , 2.7373019136482146e-45 ], [ 961.625 , 2.7376327131722866e-45 ], [ 961.6875 , 2.7379635526730498e-45 ], 
                    [ 961.75 , 2.738294432155373e-45 ], [ 961.8125 , 2.7386253516240503e-45 ], [ 961.875 , 2.7389563110839132e-45 ], [ 961.9375 , 2.739287310539834e-45 ], 
                    [ 962.0 , 2.7396183499966455e-45 ], [ 962.0625 , 2.7399494294591436e-45 ], [ 962.125 , 2.7402805489321815e-45 ], [ 962.1875 , 2.740611708420615e-45 ], 
                    [ 962.25 , 2.7409429079292204e-45 ], [ 962.3125 , 2.7412741474629324e-45 ], [ 962.375 , 2.7416054270264513e-45 ], [ 962.4375 , 2.7419367466247896e-45 ], 
                    [ 962.5 , 2.742268106262591e-45 ], [ 962.5625 , 2.7425995059448502e-45 ], [ 962.625 , 2.7429309456763475e-45 ], [ 962.6875 , 2.7432624254618846e-45 ], 
                    [ 962.75 , 2.743593945306399e-45 ], [ 962.8125 , 2.7439255052145952e-45 ], [ 962.875 , 2.744257105191432e-45 ], [ 962.9375 , 2.7445887452417324e-45 ], 
                    [ 963.0 , 2.7449204253702997e-45 ], [ 963.0625 , 2.7452521455820165e-45 ], [ 963.125 , 2.7455830993805846e-45 ], [ 963.1875 , 2.745913849417393e-45 ], 
                    [ 963.25 , 2.74624463929845e-45 ], [ 963.3125 , 2.7465754690284765e-45 ], [ 963.375 , 2.7469063386123124e-45 ], [ 963.4375 , 2.7472372480547783e-45 ], 
                    [ 963.5 , 2.7475681973606364e-45 ], [ 963.5625 , 2.747899186534748e-45 ], [ 963.625 , 2.7482302155818377e-45 ], [ 963.6875 , 2.7485612845067472e-45 ], 
                    [ 963.75 , 2.7488923933143206e-45 ], [ 963.8125 , 2.7492235420093028e-45 ], [ 963.875 , 2.7495547305964996e-45 ], [ 963.9375 , 2.749885959080736e-45 ], 
                    [ 964.0 , 2.7502172274668372e-45 ], [ 964.0625 , 2.7505485357595916e-45 ], [ 964.125 , 2.750879883963786e-45 ], [ 964.1875 , 2.751211272084268e-45 ], 
                    [ 964.25 , 2.7515427001258464e-45 ], [ 964.3125 , 2.7518741680932515e-45 ], [ 964.375 , 2.7522056759913912e-45 ], [ 964.4375 , 2.7525372238250168e-45 ], 
                    [ 964.5 , 2.752868811598979e-45 ], [ 964.5625 , 2.7532004393180885e-45 ], [ 964.625 , 2.753532106987099e-45 ], [ 964.6875 , 2.7538638146108815e-45 ], 
                    [ 964.75 , 2.7541955621942106e-45 ], [ 964.8125 , 2.7545273497419392e-45 ], [ 964.875 , 2.7548591772588616e-45 ], [ 964.9375 , 2.755191044749734e-45 ], 
                    [ 965.0 , 2.7555229522194898e-45 ], [ 965.0625 , 2.755854899672906e-45 ], [ 965.125 , 2.7561868871147402e-45 ], [ 965.1875 , 2.7565189145498682e-45 ], 
                    [ 965.25 , 2.7568509819830888e-45 ], [ 965.3125 , 2.757182958142641e-45 ], [ 965.375 , 2.757514539655281e-45 ], [ 965.4375 , 2.7578461610442373e-45 ], 
                    [ 965.5 , 2.7581778223143056e-45 ], [ 965.5625 , 2.7585095234702424e-45 ], [ 965.625 , 2.7588412645169427e-45 ], [ 965.6875 , 2.7591730454591058e-45 ], 
                    [ 965.75 , 2.7595048663015685e-45 ], [ 965.8125 , 2.759836727049149e-45 ], [ 965.875 , 2.7601686277066272e-45 ], [ 965.9375 , 2.7605005682787825e-45 ], 
                    [ 966.0 , 2.7608325487704545e-45 ], [ 966.0625 , 2.7611645691864242e-45 ], [ 966.125 , 2.7614966295314935e-45 ], [ 966.1875 , 2.7618287298104436e-45 ], 
                    [ 966.25 , 2.7621608700280978e-45 ], [ 966.3125 , 2.7624930501892976e-45 ], [ 966.375 , 2.7628252702987684e-45 ], [ 966.4375 , 2.7631575303613735e-45 ], 
                    [ 966.5 , 2.7634898303818974e-45 ], [ 966.5625 , 2.763822170365107e-45 ], [ 966.625 , 2.7641545503159264e-45 ], [ 966.6875 , 2.764486970239004e-45 ], 
                    [ 966.75 , 2.7648194301392653e-45 ], [ 966.8125 , 2.765151930021478e-45 ], [ 966.875 , 2.7654844698904514e-45 ], [ 966.9375 , 2.765817049750994e-45 ], 
                    [ 967.0 , 2.7661496696079147e-45 ], [ 967.0625 , 2.7664823294660046e-45 ], [ 967.125 , 2.7668150293300933e-45 ], [ 967.1875 , 2.767147769205012e-45 ], 
                    [ 967.25 , 2.7674805490955334e-45 ], [ 967.3125 , 2.7678133690065084e-45 ], [ 967.375 , 2.7681462289427308e-45 ], [ 967.4375 , 2.768479128908994e-45 ], 
                    [ 967.5 , 2.7688120689101325e-45 ], [ 967.5625 , 2.769144778444767e-45 ], [ 967.625 , 2.769477359311293e-45 ], [ 967.6875 , 2.76980998012161e-45 ], 
                    [ 967.75 , 2.7701426408804763e-45 ], [ 967.8125 , 2.7704753415926708e-45 ], [ 967.875 , 2.770808082263011e-45 ], [ 967.9375 , 2.7711408628963358e-45 ], 
                    [ 968.0 , 2.7714736834973656e-45 ], [ 968.0625 , 2.7718065440709406e-45 ], [ 968.125 , 2.7721394446218807e-45 ], [ 968.1875 , 2.772472385154968e-45 ], 
                    [ 968.25 , 2.7728053656749646e-45 ], [ 968.3125 , 2.7731383861867525e-45 ], [ 968.375 , 2.7734714466950948e-45 ], [ 968.4375 , 2.7738045472047954e-45 ], 
                    [ 968.5 , 2.7741376877206193e-45 ], [ 968.5625 , 2.77447086824745e-45 ], [ 968.625 , 2.774804088790053e-45 ], [ 968.6875 , 2.7751373493532355e-45 ], 
                    [ 968.75 , 2.775470649941783e-45 ], [ 968.8125 , 2.7758039905605236e-45 ], [ 968.875 , 2.7761373712143036e-45 ], [ 968.9375 , 2.7764707919078525e-45 ], 
                    [ 969.0 , 2.7768042526460194e-45 ], [ 969.0625 , 2.7771377534336326e-45 ], [ 969.125 , 2.777471294275443e-45 ], [ 969.1875 , 2.7778048751763603e-45 ], 
                    [ 969.25 , 2.7781384961410965e-45 ], [ 969.3125 , 2.7784721571745234e-45 ], [ 969.375 , 2.778805858281413e-45 ], [ 969.4375 , 2.7791395994665986e-45 ], 
                    [ 969.5 , 2.7794733807349127e-45 ], [ 969.5625 , 2.7798072020911504e-45 ], [ 969.625 , 2.7801410635401062e-45 ], [ 969.6875 , 2.780474965086635e-45 ], 
                    [ 969.75 , 2.7808089067355328e-45 ], [ 969.8125 , 2.781141952451255e-45 ], [ 969.875 , 2.7814749566498754e-45 ], [ 969.9375 , 2.781808000721273e-45 ], 
                    [ 970.0 , 2.782141084670201e-45 ], [ 970.0625 , 2.7824742085014148e-45 ], [ 970.125 , 2.7828073722197493e-45 ], [ 970.1875 , 2.783140575829921e-45 ], 
                    [ 970.25 , 2.7834738193367263e-45 ], [ 970.3125 , 2.7838071027449423e-45 ], [ 970.375 , 2.7841404260593464e-45 ], [ 970.4375 , 2.7844737892847167e-45 ], 
                    [ 970.5 , 2.7848071924258525e-45 ], [ 970.5625 , 2.7851406354874732e-45 ], [ 970.625 , 2.7854741184743986e-45 ], [ 970.6875 , 2.7858076413914284e-45 ], 
                    [ 970.75 , 2.786141204243325e-45 ], [ 970.8125 , 2.7864748070348888e-45 ], [ 970.875 , 2.7868084497708632e-45 ], [ 970.9375 , 2.7871421324560502e-45 ], 
                    [ 971.0 , 2.7874758550952534e-45 ], [ 971.0625 , 2.787809617693217e-45 ], [ 971.125 , 2.7881434202547254e-45 ], [ 971.1875 , 2.7884772627846037e-45 ], 
                    [ 971.25 , 2.788811145287637e-45 ], [ 971.3125 , 2.789145067768533e-45 ], [ 971.375 , 2.789479030232197e-45 ], [ 971.4375 , 2.789813032683317e-45 ], 
                    [ 971.5 , 2.790147075126761e-45 ], [ 971.5625 , 2.7904811575672372e-45 ], [ 971.625 , 2.7908152800096153e-45 ], [ 971.6875 , 2.791149442458644e-45 ], 
                    [ 971.75 , 2.791483644919075e-45 ], [ 971.8125 , 2.7918178873957777e-45 ], [ 971.875 , 2.7921521698934646e-45 ], [ 971.9375 , 2.792486492416987e-45 ], 
                    [ 972.0 , 2.7928207794301776e-45 ], [ 972.0625 , 2.7931545917765163e-45 ], [ 972.125 , 2.7934884440218285e-45 ], [ 972.1875 , 2.793822336170923e-45 ], 
                    [ 972.25 , 2.7941562682285097e-45 ], [ 972.3125 , 2.7944902401993587e-45 ], [ 972.375 , 2.7948242520883e-45 ], [ 972.4375 , 2.795158303900006e-45 ], 
                    [ 972.5 , 2.795492395639347e-45 ], [ 972.5625 , 2.7958265273110365e-45 ], [ 972.625 , 2.7961606989199066e-45 ], [ 972.6875 , 2.7964949104706516e-45 ], 
                    [ 972.75 , 2.796829161968085e-45 ], [ 972.8125 , 2.7971634534170017e-45 ], [ 972.875 , 2.797497784822157e-45 ], [ 972.9375 , 2.797832156188306e-45 ], 
                    [ 973.0 , 2.798166567520266e-45 ], [ 973.0625 , 2.7985010188227932e-45 ], [ 973.125 , 2.7988355101006454e-45 ], [ 973.1875 , 2.799170041358641e-45 ], 
                    [ 973.25 , 2.7995046126015376e-45 ], [ 973.3125 , 2.7998392238341156e-45 ], [ 973.375 , 2.8001738750611338e-45 ], [ 973.4375 , 2.8005085662874135e-45 ], 
                    [ 973.5 , 2.8008432975176944e-45 ], [ 973.5625 , 2.801178068756779e-45 ], [ 973.625 , 2.801512880009469e-45 ], [ 973.6875 , 2.801847731280507e-45 ], 
                    [ 973.75 , 2.802182622574696e-45 ], [ 973.8125 , 2.8025175538968207e-45 ], [ 973.875 , 2.8028525252517044e-45 ], [ 973.9375 , 2.803187536644072e-45 ], 
                    [ 974.0 , 2.80352258807871e-45 ], [ 974.0625 , 2.803857679560423e-45 ], [ 974.125 , 2.8041928110940196e-45 ], [ 974.1875 , 2.8045279826842655e-45 ], 
                    [ 974.25 , 2.8048630328855726e-45 ], [ 974.3125 , 2.8051976909458004e-45 ], [ 974.375 , 2.805532388935256e-45 ], [ 974.4375 , 2.805867126858744e-45 ], 
                    [ 974.5 , 2.80620190472095e-45 ], [ 974.5625 , 2.806536722526698e-45 ], [ 974.625 , 2.806871580280696e-45 ], [ 974.6875 , 2.807206477987787e-45 ], 
                    [ 974.75 , 2.807541415652681e-45 ], [ 974.8125 , 2.807876393280183e-45 ], [ 974.875 , 2.8082114108750235e-45 ], [ 974.9375 , 2.808546468441969e-45 ], 
                    [ 975.0 , 2.80888156598583e-45 ], [ 975.0625 , 2.8092167035113964e-45 ], [ 975.125 , 2.809551881023338e-45 ], [ 975.1875 , 2.809887098526525e-45 ], 
                    [ 975.25 , 2.8102223560256706e-45 ], [ 975.3125 , 2.8105576535255655e-45 ], [ 975.375 , 2.810892991031023e-45 ], [ 975.4375 , 2.811228368546716e-45 ], 
                    [ 975.5 , 2.811563786077479e-45 ], [ 975.5625 , 2.811899243628145e-45 ], [ 975.625 , 2.812234741203409e-45 ], [ 975.6875 , 2.812570278808028e-45 ], 
                    [ 975.75 , 2.812905856446838e-45 ], [ 975.8125 , 2.8132414741245935e-45 ], [ 975.875 , 2.8135771318460937e-45 ], [ 975.9375 , 2.813912829616095e-45 ], 
                    [ 976.0 , 2.814248567439356e-45 ], [ 976.0625 , 2.814584345320697e-45 ], [ 976.125 , 2.814920163264856e-45 ], [ 976.1875 , 2.815256021276633e-45 ], 
                    [ 976.25 , 2.8155919193608096e-45 ], [ 976.3125 , 2.815927857522166e-45 ], [ 976.375 , 2.8162638357655045e-45 ], [ 976.4375 , 2.816599854095548e-45 ], 
                    [ 976.5 , 2.816935775665534e-45 ], [ 976.5625 , 2.8172714464547583e-45 ], [ 976.625 , 2.817607157243096e-45 ], [ 976.6875 , 2.8179429080352946e-45 ], 
                    [ 976.75 , 2.818278698836099e-45 ], [ 976.8125 , 2.818614529650339e-45 ], [ 976.875 , 2.818950400482702e-45 ], [ 976.9375 , 2.8192863113379966e-45 ], 
                    [ 977.0 , 2.819622262221012e-45 ], [ 977.0625 , 2.819958253136477e-45 ], [ 977.125 , 2.8202942840892435e-45 ], [ 977.1875 , 2.820630355083941e-45 ], 
                    [ 977.25 , 2.820966466125462e-45 ], [ 977.3125 , 2.821302617218538e-45 ], [ 977.375 , 2.8216388083679417e-45 ], [ 977.4375 , 2.821975039578446e-45 ], 
                    [ 977.5 , 2.8223113108548053e-45 ], [ 977.5625 , 2.822647622201834e-45 ], [ 977.625 , 2.822983973624286e-45 ], [ 977.6875 , 2.823320365126918e-45 ], 
                    [ 977.75 , 2.823656796714545e-45 ], [ 977.8125 , 2.823993268391924e-45 ], [ 977.875 , 2.824329780163812e-45 ], [ 977.9375 , 2.824666332035047e-45 ], 
                    [ 978.0 , 2.8250029240103464e-45 ], [ 978.0625 , 2.825339556094491e-45 ], [ 978.125 , 2.8256762282922983e-45 ], [ 978.1875 , 2.82601294060851e-45 ], 
                    [ 978.25 , 2.826349693047945e-45 ], [ 978.3125 , 2.8266864856153855e-45 ], [ 978.375 , 2.8270233183155526e-45 ], [ 978.4375 , 2.827360191153269e-45 ], 
                    [ 978.5 , 2.827697104133357e-45 ], [ 978.5625 , 2.8280340572605217e-45 ], [ 978.625 , 2.8283710505395853e-45 ], [ 978.6875 , 2.8287080839753727e-45 ], 
                    [ 978.75 , 2.8290448851714776e-45 ], [ 978.8125 , 2.8293811027396585e-45 ], [ 978.875 , 2.829717360265584e-45 ], [ 978.9375 , 2.830053657754003e-45 ], 
                    [ 979.0 , 2.8303899952096644e-45 ], [ 979.0625 , 2.830726372637318e-45 ], [ 979.125 , 2.8310627900417356e-45 ], [ 979.1875 , 2.831399247427607e-45 ], 
                    [ 979.25 , 2.831735744799724e-45 ], [ 979.3125 , 2.8320722821628796e-45 ], [ 979.375 , 2.8324088595217456e-45 ], [ 979.4375 , 2.832745476881136e-45 ], 
                    [ 979.5 , 2.833082134245783e-45 ], [ 979.5625 , 2.833418831620423e-45 ], [ 979.625 , 2.83375556900987e-45 ], [ 979.6875 , 2.8340923464188005e-45 ], 
                    [ 979.75 , 2.834429163852029e-45 ], [ 979.8125 , 2.8347660213143135e-45 ], [ 979.875 , 2.83510291881033e-45 ], [ 979.9375 , 2.8354398563449784e-45 ], 
                    [ 980.0 , 2.835776833922875e-45 ], [ 980.0625 , 2.836113851548859e-45 ], [ 980.125 , 2.8364509092276714e-45 ], [ 980.1875 , 2.8367880069640715e-45 ], 
                    [ 980.25 , 2.8371251447628395e-45 ], [ 980.3125 , 2.837462322628697e-45 ], [ 980.375 , 2.837799540566405e-45 ], [ 980.4375 , 2.8381367985807667e-45 ], 
                    [ 980.5 , 2.838474096676525e-45 ], [ 980.5625 , 2.8388114348584434e-45 ], [ 980.625 , 2.8391488131312654e-45 ], [ 980.6875 , 2.8394862314997956e-45 ], 
                    [ 980.75 , 2.83982368996878e-45 ], [ 980.8125 , 2.840161188542963e-45 ], [ 980.875 , 2.840498727227152e-45 ], [ 980.9375 , 2.840836306026094e-45 ], 
                    [ 981.0 , 2.841173752789706e-45 ], [ 981.0625 , 2.841510515065688e-45 ], [ 981.125 , 2.84184731725792e-45 ], [ 981.1875 , 2.8421841593710124e-45 ], 
                    [ 981.25 , 2.842521041409778e-45 ], [ 981.3125 , 2.842857963378969e-45 ], [ 981.375 , 2.843194925283238e-45 ], [ 981.4375 , 2.8435319271273584e-45 ], 
                    [ 981.5 , 2.843868968916064e-45 ], [ 981.5625 , 2.8442060506540894e-45 ], [ 981.625 , 2.8445431723461916e-45 ], [ 981.6875 , 2.8448803339970434e-45 ], 
                    [ 981.75 , 2.845217535611483e-45 ], [ 981.8125 , 2.845554777194147e-45 ], [ 981.875 , 2.8458920587498124e-45 ], [ 981.9375 , 2.846229380283217e-45 ], 
                    [ 982.0 , 2.8465667417991007e-45 ], [ 982.0625 , 2.8469041433022215e-45 ], [ 982.125 , 2.8472415847972986e-45 ], [ 982.1875 , 2.847579066289053e-45 ], 
                    [ 982.25 , 2.8479165877822445e-45 ], [ 982.3125 , 2.848254149281616e-45 ], [ 982.375 , 2.848591750791908e-45 ], [ 982.4375 , 2.848929392317864e-45 ], 
                    [ 982.5 , 2.8492670738642264e-45 ], [ 982.5625 , 2.849604795435779e-45 ], [ 982.625 , 2.849942557037186e-45 ], [ 982.6875 , 2.850280358673252e-45 ], 
                    [ 982.75 , 2.850618200348683e-45 ], [ 982.8125 , 2.850956082068243e-45 ], [ 982.875 , 2.85129400383668e-45 ], [ 982.9375 , 2.851631965658801e-45 ], 
                    [ 983.0 , 2.8519699675392124e-45 ], [ 983.0625 , 2.852308009482763e-45 ], [ 983.125 , 2.8526460914942016e-45 ], [ 983.1875 , 2.8529842135782174e-45 ], 
                    [ 983.25 , 2.8533223757396394e-45 ], [ 983.3125 , 2.853660152222229e-45 ], [ 983.375 , 2.853997961767063e-45 ], [ 983.4375 , 2.8543358113009936e-45 ], 
                    [ 983.5 , 2.8546737008287535e-45 ], [ 983.5625 , 2.8550116303550566e-45 ], [ 983.625 , 2.855349599884719e-45 ], [ 983.6875 , 2.8556876094223354e-45 ], 
                    [ 983.75 , 2.8560256589728024e-45 ], [ 983.8125 , 2.856363748540716e-45 ], [ 983.875 , 2.856701878130894e-45 ], [ 983.9375 , 2.857040047748033e-45 ], 
                    [ 984.0 , 2.857378257396933e-45 ], [ 984.0625 , 2.857716507082291e-45 ], [ 984.125 , 2.858054796808828e-45 ], [ 984.1875 , 2.8583931265813033e-45 ], 
                    [ 984.25 , 2.858731496404457e-45 ], [ 984.3125 , 2.8590699062830313e-45 ], [ 984.375 , 2.859408356221767e-45 ], [ 984.4375 , 2.859746846225366e-45 ], 
                    [ 984.5 , 2.860085376298693e-45 ], [ 984.5625 , 2.860423946446349e-45 ], [ 984.625 , 2.86076255667318e-45 ], [ 984.6875 , 2.8611012069838485e-45 ], 
                    [ 984.75 , 2.86143989738316e-45 ], [ 984.8125 , 2.861778627875842e-45 ], [ 984.875 , 2.862117398466638e-45 ], [ 984.9375 , 2.862456209160296e-45 ], 
                    [ 985.0 , 2.862795059961544e-45 ], [ 985.0625 , 2.863133950875188e-45 ], [ 985.125 , 2.863472881905919e-45 ], [ 985.1875 , 2.863811853058524e-45 ], 
                    [ 985.25 , 2.864150864337692e-45 ], [ 985.3125 , 2.8644899157482546e-45 ], [ 985.375 , 2.864829007294902e-45 ], [ 985.4375 , 2.8651681389823846e-45 ], 
                    [ 985.5 , 2.8655073108154956e-45 ], [ 985.5625 , 2.865845881514462e-45 ], [ 985.625 , 2.866184232790306e-45 ], [ 985.6875 , 2.866522624012983e-45 ], 
                    [ 985.75 , 2.866861055187251e-45 ], [ 985.8125 , 2.867199526317827e-45 ], [ 985.875 , 2.8675380374094476e-45 ], [ 985.9375 , 2.8678765884668114e-45 ], 
                    [ 986.0 , 2.8682151794945756e-45 ], [ 986.0625 , 2.868553810497581e-45 ], [ 986.125 , 2.8688924814805074e-45 ], [ 986.1875 , 2.8692311924480127e-45 ], 
                    [ 986.25 , 2.869569943404879e-45 ], [ 986.3125 , 2.869908734355788e-45 ], [ 986.375 , 2.8702475653054604e-45 ], [ 986.4375 , 2.8705864362587006e-45 ], 
                    [ 986.5 , 2.8709253472201286e-45 ], [ 986.5625 , 2.87126429819451e-45 ], [ 986.625 , 2.871603289186589e-45 ], [ 986.6875 , 2.871942320201028e-45 ], 
                    [ 986.75 , 2.872281391242635e-45 ], [ 986.8125 , 2.8726205023160745e-45 ], [ 986.875 , 2.872959653426112e-45 ], [ 986.9375 , 2.873298844577436e-45 ], 
                    [ 987.0 , 2.8736380757748123e-45 ], [ 987.0625 , 2.8739773470229297e-45 ], [ 987.125 , 2.874316658326557e-45 ], [ 987.1875 , 2.874656009690383e-45 ], 
                    [ 987.25 , 2.874995401119157e-45 ], [ 987.3125 , 2.875334832617629e-45 ], [ 987.375 , 2.8756743041904697e-45 ], [ 987.4375 , 2.8760138158424927e-45 ], 
                    [ 987.5 , 2.876353367578387e-45 ], [ 987.5625 , 2.876692959402866e-45 ], [ 987.625 , 2.877032591320683e-45 ], [ 987.6875 , 2.877372263336571e-45 ], 
                    [ 987.75 , 2.877711975455285e-45 ], [ 987.8125 , 2.878051621965956e-45 ], [ 987.875 , 2.878391106942949e-45 ], [ 987.9375 , 2.878730631964432e-45 ], 
                    [ 988.0 , 2.879070197035087e-45 ], [ 988.0625 , 2.879409802159679e-45 ], [ 988.125 , 2.8797494473429126e-45 ], [ 988.1875 , 2.880089132589492e-45 ], 
                    [ 988.25 , 2.8804288579041843e-45 ], [ 988.3125 , 2.8807686232917166e-45 ], [ 988.375 , 2.8811084287567523e-45 ], [ 988.4375 , 2.881448274304082e-45 ], 
                    [ 988.5 , 2.881788159938412e-45 ], [ 988.5625 , 2.882128085664472e-45 ], [ 988.625 , 2.88246805148697e-45 ], [ 988.6875 , 2.882808057410676e-45 ], 
                    [ 988.75 , 2.88314810344028e-45 ], [ 988.8125 , 2.8834881895805745e-45 ], [ 988.875 , 2.8838283158362075e-45 ], [ 988.9375 , 2.8841684822119535e-45 ], 
                    [ 989.0 , 2.884508688712605e-45 ], [ 989.0625 , 2.884848935342753e-45 ], [ 989.125 , 2.885189222107273e-45 ], [ 989.1875 , 2.8855295490107974e-45 ], 
                    [ 989.25 , 2.8858699160581226e-45 ], [ 989.3125 , 2.886210323253983e-45 ], [ 989.375 , 2.886550770603054e-45 ], [ 989.4375 , 2.8868912581101324e-45 ], 
                    [ 989.5 , 2.8872317857798935e-45 ], [ 989.5625 , 2.8875723536171783e-45 ], [ 989.625 , 2.887912961626642e-45 ], [ 989.6875 , 2.888253609813024e-45 ], 
                    [ 989.75 , 2.888594298181083e-45 ], [ 989.8125 , 2.888935026735559e-45 ], [ 989.875 , 2.889275795481254e-45 ], [ 989.9375 , 2.889616604422765e-45 ], 
                    [ 990.0 , 2.889957453564977e-45 ], [ 990.0625 , 2.89029834291255e-45 ], [ 990.125 , 2.8906383108222e-45 ], [ 990.1875 , 2.8909782259186676e-45 ], 
                    [ 990.25 , 2.891318180986352e-45 ], [ 990.3125 , 2.891658176029934e-45 ], [ 990.375 , 2.891998211054093e-45 ], [ 990.4375 , 2.8923382860635724e-45 ], 
                    [ 990.5 , 2.892678401063073e-45 ], [ 990.5625 , 2.893018556057216e-45 ], [ 990.625 , 2.893358751050827e-45 ], [ 990.6875 , 2.893698986048509e-45 ], 
                    [ 990.75 , 2.894039261055024e-45 ], [ 990.8125 , 2.8943795760750387e-45 ], [ 990.875 , 2.8947199311132985e-45 ], [ 990.9375 , 2.8950603261744673e-45 ], 
                    [ 991.0 , 2.895400761263315e-45 ], [ 991.0625 , 2.895741236384464e-45 ], [ 991.125 , 2.8960817515426643e-45 ], [ 991.1875 , 2.896422306742645e-45 ], 
                    [ 991.25 , 2.8967629019891134e-45 ], [ 991.3125 , 2.897103537286697e-45 ], [ 991.375 , 2.8974442126401876e-45 ], [ 991.4375 , 2.897784928054296e-45 ], 
                    [ 991.5 , 2.898125683533711e-45 ], [ 991.5625 , 2.898466479083125e-45 ], [ 991.625 , 2.8988073147073106e-45 ], [ 991.6875 , 2.8991481904109396e-45 ], 
                    [ 991.75 , 2.899489106198725e-45 ], [ 991.8125 , 2.8998300620752975e-45 ], [ 991.875 , 2.900171058045516e-45 ], [ 991.9375 , 2.9005120941140327e-45 ], 
                    [ 992.0 , 2.9008531702855224e-45 ], [ 992.0625 , 2.901194286564762e-45 ], [ 992.125 , 2.90153544295651e-45 ], [ 992.1875 , 2.901876639465379e-45 ], 
                    [ 992.25 , 2.902217876096128e-45 ], [ 992.3125 , 2.902559152853476e-45 ], [ 992.375 , 2.9029003074955565e-45 ], [ 992.4375 , 2.9032412235620185e-45 ], 
                    [ 992.5 , 2.903582179665642e-45 ], [ 992.5625 , 2.903923175811044e-45 ], [ 992.625 , 2.904264212002969e-45 ], [ 992.6875 , 2.9046052882461624e-45 ], 
                    [ 992.75 , 2.9049464045452434e-45 ], [ 992.8125 , 2.90528756090498e-45 ], [ 992.875 , 2.905628757330056e-45 ], [ 992.9375 , 2.9059699938251534e-45 ], 
                    [ 993.0 , 2.906311270395043e-45 ], [ 993.0625 , 2.906652587044348e-45 ], [ 993.125 , 2.9069939437778155e-45 ], [ 993.1875 , 2.9073353406001746e-45 ], 
                    [ 993.25 , 2.9076767775161324e-45 ], [ 993.3125 , 2.9080182545303573e-45 ], [ 993.375 , 2.908359771647537e-45 ], [ 993.4375 , 2.9087013288724644e-45 ], 
                    [ 993.5 , 2.909042926209808e-45 ], [ 993.5625 , 2.909384563664278e-45 ], [ 993.625 , 2.909726241240588e-45 ], [ 993.6875 , 2.910067958943447e-45 ], 
                    [ 993.75 , 2.910409716777549e-45 ], [ 993.8125 , 2.9107515147476256e-45 ], [ 993.875 , 2.911093352858413e-45 ], [ 993.9375 , 2.911435231114583e-45 ], 
                    [ 994.0 , 2.911777149520933e-45 ], [ 994.0625 , 2.912119108082034e-45 ], [ 994.125 , 2.912461106802747e-45 ], [ 994.1875 , 2.9128031456876624e-45 ], 
                    [ 994.25 , 2.9131452247416016e-45 ], [ 994.3125 , 2.91348734396922e-45 ], [ 994.375 , 2.913829503375298e-45 ], [ 994.4375 , 2.91417170296447e-45 ], 
                    [ 994.5 , 2.914513942741519e-45 ], [ 994.5625 , 2.9148562227111424e-45 ], [ 994.625 , 2.915198542878041e-45 ], [ 994.6875 , 2.915540212604038e-45 ], 
                    [ 994.75 , 2.915881704251347e-45 ], [ 994.8125 , 2.916223235896961e-45 ], [ 994.875 , 2.916564807545482e-45 ], [ 994.9375 , 2.916906419201698e-45 ], 
                    [ 995.0 , 2.917248070870193e-45 ], [ 995.0625 , 2.917589762555694e-45 ], [ 995.125 , 2.917931494262889e-45 ], [ 995.1875 , 2.918273265996485e-45 ], 
                    [ 995.25 , 2.91861507776113e-45 ], [ 995.3125 , 2.918956929561533e-45 ], [ 995.375 , 2.919298821402384e-45 ], [ 995.4375 , 2.9196407532883716e-45 ], 
                    [ 995.5 , 2.919982725224207e-45 ], [ 995.5625 , 2.9203247372145203e-45 ], [ 995.625 , 2.920666789264085e-45 ], [ 995.6875 , 2.921008881377552e-45 ], 
                    [ 995.75 , 2.921351013559593e-45 ], [ 995.8125 , 2.9216931858149224e-45 ], [ 995.875 , 2.922035398148253e-45 ], [ 995.9375 , 2.922377650564219e-45 ], 
                    [ 996.0 , 2.922719943067597e-45 ], [ 996.0625 , 2.9230622756629995e-45 ], [ 996.125 , 2.9234046483552054e-45 ], [ 996.1875 , 2.923747061148891e-45 ], 
                    [ 996.25 , 2.9240895140486694e-45 ], [ 996.3125 , 2.9244320070593217e-45 ], [ 996.375 , 2.9247745401855455e-45 ], [ 996.4375 , 2.9251171134319775e-45 ], 
                    [ 996.5 , 2.925459726803421e-45 ], [ 996.5625 , 2.9258023803044715e-45 ], [ 996.625 , 2.926145073939851e-45 ], [ 996.6875 , 2.9264878077143216e-45 ], 
                    [ 996.75 , 2.926830581632564e-45 ], [ 996.8125 , 2.927173395699197e-45 ], [ 996.875 , 2.927516249918987e-45 ], [ 996.9375 , 2.9278591110444324e-45 ], 
                    [ 997.0 , 2.928201451720828e-45 ], [ 997.0625 , 2.9285438324255566e-45 ], [ 997.125 , 2.928886253163238e-45 ], [ 997.1875 , 2.92922871393853e-45 ], 
                    [ 997.25 , 2.929571214756116e-45 ], [ 997.3125 , 2.9299137556207387e-45 ], [ 997.375 , 2.93025633653704e-45 ], [ 997.4375 , 2.9305989575096614e-45 ], 
                    [ 997.5 , 2.930941618543348e-45 ], [ 997.5625 , 2.931284319642785e-45 ], [ 997.625 , 2.9316270608125935e-45 ], [ 997.6875 , 2.9319698420575226e-45 ], 
                    [ 997.75 , 2.932312663382257e-45 ], [ 997.8125 , 2.932655524791442e-45 ], [ 997.875 , 2.932998426289785e-45 ], [ 997.9375 , 2.9333413678819736e-45 ], 
                    [ 998.0 , 2.933684349572675e-45 ], [ 998.0625 , 2.9340273713666396e-45 ], [ 998.125 , 2.9343704332685154e-45 ], [ 998.1875 , 2.9347135352829507e-45 ], 
                    [ 998.25 , 2.935056677414676e-45 ], [ 998.3125 , 2.935399859668426e-45 ], [ 998.375 , 2.935743082048848e-45 ], [ 998.4375 , 2.936086344560614e-45 ], 
                    [ 998.5 , 2.936429647208416e-45 ], [ 998.5625 , 2.9367729899969686e-45 ], [ 998.625 , 2.937116372931027e-45 ], [ 998.6875 , 2.9374597960152025e-45 ], 
                    [ 998.75 , 2.937803259254147e-45 ], [ 998.8125 , 2.9381467626526394e-45 ], [ 998.875 , 2.938490306215417e-45 ], [ 998.9375 , 2.938833889947073e-45 ], 
                    [ 999.0 , 2.939177513852344e-45 ], [ 999.0625 , 2.9395211779359495e-45 ], [ 999.125 , 2.939864882202545e-45 ], [ 999.1875 , 2.940208626656849e-45 ], 
                    [ 999.25 , 2.9405521175222096e-45 ], [ 999.3125 , 2.940894862127205e-45 ], [ 999.375 , 2.941237646681816e-45 ], [ 999.4375 , 2.941580471190636e-45 ], 
                    [ 999.5 , 2.941923335658384e-45 ], [ 999.5625 , 2.9422662400896774e-45 ], [ 999.625 , 2.942609184489235e-45 ], [ 999.6875 , 2.942952168861676e-45 ], 
                    [ 999.75 , 2.943295193211636e-45 ], [ 999.8125 , 2.943638257543755e-45 ], [ 999.875 , 2.9439813618627984e-45 ], [ 999.9375 , 2.9443245061733634e-45 ], 
                    [ 1000.0 , 2.94466769048007e-45 ], [ 1000.0625 , 2.945010914787663e-45 ], [ 1000.125 , 2.945354179100743e-45 ], [ 1000.1875 , 2.945697483423993e-45 ], 
                    [ 1000.25 , 2.9460408277620566e-45 ], [ 1000.3125 , 2.94638421211966e-45 ], [ 1000.375 , 2.946727636501404e-45 ], [ 1000.4375 , 2.947071100911977e-45 ], 
                    [ 1000.5 , 2.947414605356063e-45 ], [ 1000.5625 , 2.947758149838247e-45 ], [ 1000.625 , 2.948101734363299e-45 ], [ 1000.6875 , 2.948445358935845e-45 ], 
                    [ 1000.75 , 2.9487890235605725e-45 ], [ 1000.8125 , 2.949132728242109e-45 ], [ 1000.875 , 2.949476472985145e-45 ], [ 1000.9375 , 2.949820257794349e-45 ], 
                    [ 1001.0 , 2.9501640826743916e-45 ], [ 1001.0625 , 2.9505079476299225e-45 ], [ 1001.125 , 2.9508518526656956e-45 ], [ 1001.1875 , 2.951195797786258e-45 ], 
                    [ 1001.25 , 2.951539782996344e-45 ], [ 1001.3125 , 2.951883808300669e-45 ], [ 1001.375 , 2.952227873703801e-45 ], [ 1001.4375 , 2.9525719792105815e-45 ], 
                    [ 1001.5 , 2.9529161248254755e-45 ], [ 1001.5625 , 2.953260156266751e-45 ], [ 1001.625 , 2.9536039534705355e-45 ], [ 1001.6875 , 2.953947790696684e-45 ], 
                    [ 1001.75 , 2.954291667949856e-45 ], [ 1001.8125 , 2.954635585234775e-45 ], [ 1001.875 , 2.954979542556017e-45 ], [ 1001.9375 , 2.955323539918327e-45 ], 
                    [ 1002.0 , 2.955667577326281e-45 ], [ 1002.0625 , 2.956011654784626e-45 ], [ 1002.125 , 2.95635577229796e-45 ], [ 1002.1875 , 2.956699929870969e-45 ], 
                    [ 1002.25 , 2.9570441275083156e-45 ], [ 1002.3125 , 2.957388365214684e-45 ], [ 1002.375 , 2.9577326429946555e-45 ], [ 1002.4375 , 2.9580769608529997e-45 ], 
                    [ 1002.5 , 2.9584213187943615e-45 ], [ 1002.5625 , 2.958765716823366e-45 ], [ 1002.625 , 2.959110154944678e-45 ], [ 1002.6875 , 2.959454633163008e-45 ], 
                    [ 1002.75 , 2.9597991514830025e-45 ], [ 1002.8125 , 2.96014370990933e-45 ], [ 1002.875 , 2.96048830844666e-45 ], [ 1002.9375 , 2.9608329470996606e-45 ], 
                    [ 1003.0 , 2.9611776258729816e-45 ], [ 1003.0625 , 2.9615223447713567e-45 ], [ 1003.125 , 2.9618671037993716e-45 ], [ 1003.1875 , 2.962211902961784e-45 ], 
                    [ 1003.25 , 2.962556742263201e-45 ], [ 1003.3125 , 2.962901621708339e-45 ], [ 1003.375 , 2.963246541301829e-45 ], [ 1003.4375 , 2.963591501048343e-45 ], 
                    [ 1003.5 , 2.9639365009525994e-45 ], [ 1003.5625 , 2.964281541019272e-45 ], [ 1003.625 , 2.964626621252974e-45 ], [ 1003.6875 , 2.964971741658422e-45 ], 
                    [ 1003.75 , 2.965316902240315e-45 ], [ 1003.8125 , 2.965662103003329e-45 ], [ 1003.875 , 2.9660073874657895e-45 ], [ 1003.9375 , 2.966352777407026e-45 ], 
                    [ 1004.0 , 2.966698207568711e-45 ], [ 1004.0625 , 2.9670436779555494e-45 ], [ 1004.125 , 2.967389188572226e-45 ], [ 1004.1875 , 2.967734739423467e-45 ], 
                    [ 1004.25 , 2.9680803305138534e-45 ], [ 1004.3125 , 2.968425961848155e-45 ], [ 1004.375 , 2.968771633431036e-45 ], [ 1004.4375 , 2.969117345267205e-45 ], 
                    [ 1004.5 , 2.9694630973613087e-45 ], [ 1004.5625 , 2.969808889718075e-45 ], [ 1004.625 , 2.9701547223421314e-45 ], [ 1004.6875 , 2.9705005952382706e-45 ], 
                    [ 1004.75 , 2.970846508411057e-45 ], [ 1004.8125 , 2.971192461865285e-45 ], [ 1004.875 , 2.971538455605583e-45 ], [ 1004.9375 , 2.971884489636664e-45 ], 
                    [ 1005.0 , 2.972230563963239e-45 ], [ 1005.0625 , 2.972576678589981e-45 ], [ 1005.125 , 2.972922833521561e-45 ], [ 1005.1875 , 2.973269028762694e-45 ], 
                    [ 1005.25 , 2.973615264318074e-45 ], [ 1005.3125 , 2.9739615401924163e-45 ], [ 1005.375 , 2.974307856390353e-45 ], [ 1005.4375 , 2.974654212916686e-45 ], 
                    [ 1005.5 , 2.975000609776004e-45 ], [ 1005.5625 , 2.9753470469730686e-45 ], [ 1005.625 , 2.9756935245125345e-45 ], [ 1005.6875 , 2.9760400423991e-45 ], 
                    [ 1005.75 , 2.9763866006375476e-45 ], [ 1005.8125 , 2.976733199232449e-45 ], [ 1005.875 , 2.9770798381885893e-45 ], [ 1005.9375 , 2.977426517510668e-45 ], 
                    [ 1006.0 , 2.977773237203386e-45 ], [ 1006.0625 , 2.9781199972713795e-45 ], [ 1006.125 , 2.978466797719373e-45 ], [ 1006.1875 , 2.9788130440318435e-45 ], 
                    [ 1006.25 , 2.979158573845415e-45 ], [ 1006.3125 , 2.979504143739017e-45 ], [ 1006.375 , 2.9798497537172546e-45 ], [ 1006.4375 , 2.980195403784842e-45 ], 
                    [ 1006.5 , 2.980541093946345e-45 ], [ 1006.5625 , 2.980886824206456e-45 ], [ 1006.625 , 2.9812325945698475e-45 ], [ 1006.6875 , 2.981578405041129e-45 ], 
                    [ 1006.75 , 2.9819242556249746e-45 ], [ 1006.8125 , 2.982270146326057e-45 ], [ 1006.875 , 2.982616077148967e-45 ], [ 1006.9375 , 2.9829620480984436e-45 ], 
                    [ 1007.0 , 2.9833080591790994e-45 ], [ 1007.0625 , 2.983654110395545e-45 ], [ 1007.125 , 2.984000201752544e-45 ], [ 1007.1875 , 2.984346333254645e-45 ], 
                    [ 1007.25 , 2.98469250490657e-45 ], [ 1007.3125 , 2.985038716712953e-45 ], [ 1007.375 , 2.98538496867839e-45 ], [ 1007.4375 , 2.9857312608076865e-45 ], 
                    [ 1007.5 , 2.986077593105375e-45 ], [ 1007.5625 , 2.986423965576177e-45 ], [ 1007.625 , 2.986770378224712e-45 ], [ 1007.6875 , 2.98711683105566e-45 ], 
                    [ 1007.75 , 2.987463324073725e-45 ], [ 1007.8125 , 2.9878098572835054e-45 ], [ 1007.875 , 2.988156430689641e-45 ], [ 1007.9375 , 2.988503044296838e-45 ], 
                    [ 1008.0 , 2.988849698109823e-45 ], [ 1008.0625 , 2.9891963921331526e-45 ], [ 1008.125 , 2.989543126371555e-45 ], [ 1008.1875 , 2.989889900829675e-45 ], 
                    [ 1008.25 , 2.9902367155121556e-45 ], [ 1008.3125 , 2.990583570423705e-45 ], [ 1008.375 , 2.990930465568926e-45 ], [ 1008.4375 , 2.991277400952614e-45 ], 
                    [ 1008.5 , 2.9916239865227526e-45 ], [ 1008.5625 , 2.9919697832842676e-45 ], [ 1008.625 , 2.992315620015867e-45 ], [ 1008.6875 , 2.992661496722128e-45 ], 
                    [ 1008.75 , 2.9930074134077145e-45 ], [ 1008.8125 , 2.9933533700772054e-45 ], [ 1008.875 , 2.9936993667352423e-45 ], [ 1008.9375 , 2.994045403386449e-45 ], 
                    [ 1009.0 , 2.9943914800354685e-45 ], [ 1009.0625 , 2.994737596686882e-45 ], [ 1009.125 , 2.995083753345356e-45 ], [ 1009.1875 , 2.995429950015449e-45 ], 
                    [ 1009.25 , 2.9957761867018745e-45 ], [ 1009.3125 , 2.996122463409213e-45 ], [ 1009.375 , 2.996468780142069e-45 ], [ 1009.4375 , 2.996815136905092e-45 ], 
                    [ 1009.5 , 2.997161533702929e-45 ], [ 1009.5625 , 2.9975079705401857e-45 ], [ 1009.625 , 2.9978544474214706e-45 ], [ 1009.6875 , 2.9982009643514116e-45 ], 
                    [ 1009.75 , 2.998547521334722e-45 ], [ 1009.8125 , 2.9988941183759264e-45 ], [ 1009.875 , 2.999240755479698e-45 ], [ 1009.9375 , 2.9995874326506866e-45 ], 
                    [ 1010.0 , 2.999934149893504e-45 ], [ 1010.0625 , 3.000280907212782e-45 ], [ 1010.125 , 3.000627704613088e-45 ], [ 1010.1875 , 3.000974542099184e-45 ], 
                    [ 1010.25 , 3.001321419675639e-45 ], [ 1010.3125 , 3.001668337347066e-45 ], [ 1010.375 , 3.0020152951180987e-45 ], [ 1010.4375 , 3.002362292993457e-45 ], 
                    [ 1010.5 , 3.002709330977672e-45 ], [ 1010.5625 , 3.0030564090754414e-45 ], [ 1010.625 , 3.0034035272913824e-45 ], [ 1010.6875 , 3.00375068563011e-45 ], 
                    [ 1010.75 , 3.004097884096305e-45 ], [ 1010.8125 , 3.004445121588935e-45 ], [ 1010.875 , 3.004792372680677e-45 ], [ 1010.9375 , 3.0051396639073485e-45 ], 
                    [ 1011.0 , 3.005486995273652e-45 ], [ 1011.0625 , 3.005834366784227e-45 ], [ 1011.125 , 3.0061817784436704e-45 ], [ 1011.1875 , 3.006529230256644e-45 ], 
                    [ 1011.25 , 3.00687672222781e-45 ], [ 1011.3125 , 3.0072242543617675e-45 ], [ 1011.375 , 3.007571826663158e-45 ], [ 1011.4375 , 3.00791943913671e-45 ], 
                    [ 1011.5 , 3.0082670917869377e-45 ], [ 1011.5625 , 3.008614784618571e-45 ], [ 1011.625 , 3.008962517636232e-45 ], [ 1011.6875 , 3.0093102908445865e-45 ], 
                    [ 1011.75 , 3.009658104248216e-45 ], [ 1011.8125 , 3.010005957851873e-45 ], [ 1011.875 , 3.010353851660076e-45 ], [ 1011.9375 , 3.010701785677557e-45 ], 
                    [ 1012.0 , 3.011049759908963e-45 ], [ 1012.0625 , 3.011397774358878e-45 ], [ 1012.125 , 3.0117458290320365e-45 ], [ 1012.1875 , 3.0120939239330437e-45 ], 
                    [ 1012.25 , 3.0124420590665285e-45 ], [ 1012.3125 , 3.01279023443714e-45 ], [ 1012.375 , 3.013138450049552e-45 ], [ 1012.4375 , 3.0134867059084776e-45 ], 
                    [ 1012.5 , 3.013835002018484e-45 ], [ 1012.5625 , 3.014183338384245e-45 ], [ 1012.625 , 3.0145317150103483e-45 ], [ 1012.6875 , 3.014880131901576e-45 ], 
                    [ 1012.75 , 3.0152285890624966e-45 ], [ 1012.8125 , 3.015577086497785e-45 ], [ 1012.875 , 3.01592562421214e-45 ], [ 1012.9375 , 3.0162742022100886e-45 ], 
                    [ 1013.0 , 3.0166228204964354e-45 ], [ 1013.0625 , 3.016971479075752e-45 ], [ 1013.125 , 3.0173201779527166e-45 ], [ 1013.1875 , 3.017668796502413e-45 ], 
                    [ 1013.25 , 3.018017417233244e-45 ], [ 1013.3125 , 3.018366078238988e-45 ], [ 1013.375 , 3.0187147795243624e-45 ], [ 1013.4375 , 3.019063521093936e-45 ], 
                    [ 1013.5 , 3.0194123029524246e-45 ], [ 1013.5625 , 3.0197611251044625e-45 ], [ 1013.625 , 3.0201099875547056e-45 ], [ 1013.6875 , 3.020458890307786e-45 ], 
                    [ 1013.75 , 3.020807833368404e-45 ], [ 1013.8125 , 3.021156816741195e-45 ], [ 1013.875 , 3.021505840430815e-45 ], [ 1013.9375 , 3.021854904441901e-45 ], 
                    [ 1014.0 , 3.0222040087791534e-45 ], [ 1014.0625 , 3.0225531534472104e-45 ], [ 1014.125 , 3.0229023384507304e-45 ], [ 1014.1875 , 3.0232515637943526e-45 ], 
                    [ 1014.25 , 3.0236008294828016e-45 ], [ 1014.3125 , 3.02395013552063e-45 ], [ 1014.375 , 3.0242994819126075e-45 ], [ 1014.4375 , 3.024648868663353e-45 ], 
                    [ 1014.5 , 3.024998295777529e-45 ], [ 1014.5625 , 3.0253477632597977e-45 ], [ 1014.625 , 3.025697271114824e-45 ], [ 1014.6875 , 3.026046819347228e-45 ], 
                    [ 1014.75 , 3.026396407961783e-45 ], [ 1014.8125 , 3.0267460369630456e-45 ], [ 1014.875 , 3.0270957063557894e-45 ], [ 1014.9375 , 3.0274454161445306e-45 ], 
                    [ 1015.0 , 3.027795166334064e-45 ], [ 1015.0625 , 3.028144956929015e-45 ], [ 1015.125 , 3.0284947879340296e-45 ], [ 1015.1875 , 3.0288446593538405e-45 ], 
                    [ 1015.25 , 3.029194571193031e-45 ], [ 1015.3125 , 3.0295445234563346e-45 ], [ 1015.375 , 3.029894516148423e-45 ], [ 1015.4375 , 3.0302445492738575e-45 ], 
                    [ 1015.5 , 3.0305933814898814e-45 ], [ 1015.5625 , 3.0309409089148995e-45 ], [ 1015.625 , 3.031288476191951e-45 ], [ 1015.6875 , 3.031636083325607e-45 ], 
                    [ 1015.75 , 3.031983730320394e-45 ], [ 1015.8125 , 3.0323314171809916e-45 ], [ 1015.875 , 3.032679143911885e-45 ], [ 1015.9375 , 3.033026910517646e-45 ], 
                    [ 1016.0 , 3.033374717002869e-45 ], [ 1016.0625 , 3.0337225633721485e-45 ], [ 1016.125 , 3.034070449630036e-45 ], [ 1016.1875 , 3.0344183757810847e-45 ], 
                    [ 1016.25 , 3.0347663418299343e-45 ], [ 1016.3125 , 3.0351143477810946e-45 ], [ 1016.375 , 3.035462393639142e-45 ], [ 1016.4375 , 3.035810479408673e-45 ], 
                    [ 1016.5 , 3.036158605094266e-45 ], [ 1016.5625 , 3.0365067707005406e-45 ], [ 1016.625 , 3.0368549762319885e-45 ], [ 1016.6875 , 3.0372032216932745e-45 ], 
                    [ 1016.75 , 3.037551507088869e-45 ], [ 1016.8125 , 3.03789983242346e-45 ], [ 1016.875 , 3.0382481977015844e-45 ], [ 1016.9375 , 3.0385966029277996e-45 ], 
                    [ 1017.0 , 3.0389450481067096e-45 ], [ 1017.0625 , 3.039293533242982e-45 ], [ 1017.125 , 3.039642058340961e-45 ], [ 1017.1875 , 3.03999062340551e-45 ], 
                    [ 1017.25 , 3.040339228440996e-45 ], [ 1017.3125 , 3.0406878734521326e-45 ], [ 1017.375 , 3.041036558443417e-45 ], [ 1017.4375 , 3.041385283419564e-45 ], 
                    [ 1017.5 , 3.041734048385007e-45 ], [ 1017.5625 , 3.042082853344418e-45 ], [ 1017.625 , 3.0424316983023844e-45 ], [ 1017.6875 , 3.042780583263427e-45 ], 
                    [ 1017.75 , 3.043129508232177e-45 ], [ 1017.8125 , 3.043478473213243e-45 ], [ 1017.875 , 3.043828794211517e-45 ], [ 1017.9375 , 3.04417933500898e-45 ], 
                    [ 1018.0 , 3.0445299161762514e-45 ], [ 1018.0625 , 3.044880537718045e-45 ], [ 1018.125 , 3.045231199638924e-45 ], [ 1018.1875 , 3.0455819019435817e-45 ], 
                    [ 1018.25 , 3.0459326446366697e-45 ], [ 1018.3125 , 3.04628342772286e-45 ], [ 1018.375 , 3.046634251206762e-45 ], [ 1018.4375 , 3.046985115093071e-45 ], 
                    [ 1018.5 , 3.047336019386418e-45 ], [ 1018.5625 , 3.047686964091435e-45 ], [ 1018.625 , 3.048037949212798e-45 ], [ 1018.6875 , 3.048388974755161e-45 ], 
                    [ 1018.75 , 3.048740040723222e-45 ], [ 1018.8125 , 3.04909114712153e-45 ], [ 1018.875 , 3.049442293954848e-45 ], [ 1018.9375 , 3.049793481227789e-45 ], 
                    [ 1019.0 , 3.0501447089450116e-45 ], [ 1019.0625 , 3.050495977111108e-45 ], [ 1019.125 , 3.0508472857308873e-45 ], [ 1019.1875 , 3.051198634808901e-45 ], 
                    [ 1019.25 , 3.0515500243497865e-45 ], [ 1019.3125 , 3.0519014543582896e-45 ], [ 1019.375 , 3.052252924839029e-45 ], [ 1019.4375 , 3.052604435796621e-45 ], 
                    [ 1019.5 , 3.0529559872358577e-45 ], [ 1019.5625 , 3.0533075791612275e-45 ], [ 1019.625 , 3.053659211577523e-45 ], [ 1019.6875 , 3.0540108844893424e-45 ], 
                    [ 1019.75 , 3.0543625979013705e-45 ], [ 1019.8125 , 3.054714351818295e-45 ], [ 1019.875 , 3.055066146244756e-45 ], [ 1019.9375 , 3.055417981185378e-45 ], 
                    [ 1020.0 , 3.0557698566449126e-45 ], [ 1020.0625 , 3.056121772627961e-45 ], [ 1020.125 , 3.056473729139255e-45 ], [ 1020.1875 , 3.056824454914359e-45 ], 
                    [ 1020.25 , 3.0571736027711e-45 ], [ 1020.3125 , 3.0575227905072085e-45 ], [ 1020.375 , 3.057872018127239e-45 ], [ 1020.4375 , 3.058221285635748e-45 ], 
                    [ 1020.5 , 3.0585705930372913e-45 ], [ 1020.5625 , 3.0589199403364026e-45 ], [ 1020.625 , 3.0592693275376627e-45 ], [ 1020.6875 , 3.059618754645627e-45 ], 
                    [ 1020.75 , 3.059968221664898e-45 ], [ 1020.8125 , 3.06031772859997e-45 ], [ 1020.875 , 3.060667275455445e-45 ], [ 1020.9375 , 3.0610168622358165e-45 ], 
                    [ 1021.0 , 3.0613664889456895e-45 ], [ 1021.0625 , 3.061716155589646e-45 ], [ 1021.125 , 3.062065862172247e-45 ], [ 1021.1875 , 3.062415608698011e-45 ], 
                    [ 1021.25 , 3.0627653951715e-45 ], [ 1021.3125 , 3.063115221597343e-45 ], [ 1021.375 , 3.063465087979993e-45 ], [ 1021.4375 , 3.0638149943241456e-45 ], 
                    [ 1021.5 , 3.0641649406342346e-45 ], [ 1021.5625 , 3.064514926914976e-45 ], [ 1021.625 , 3.0648649531707843e-45 ], [ 1021.6875 , 3.0652150194062896e-45 ], 
                    [ 1021.75 , 3.0655651256261025e-45 ], [ 1021.8125 , 3.065915271834703e-45 ], [ 1021.875 , 3.066265458036724e-45 ], [ 1021.9375 , 3.0666156842367106e-45 ], 
                    [ 1022.0 , 3.066965950439233e-45 ], [ 1022.0625 , 3.0673162566488587e-45 ], [ 1022.125 , 3.0676666028701367e-45 ], [ 1022.1875 , 3.06801698910768e-45 ], 
                    [ 1022.25 , 3.0683674153660156e-45 ], [ 1022.3125 , 3.0687178816498024e-45 ], [ 1022.375 , 3.069068387963502e-45 ], [ 1022.4375 , 3.0694189343117315e-45 ], 
                    [ 1022.5 , 3.069769520699019e-45 ], [ 1022.5625 , 3.0701212500261083e-45 ], [ 1022.625 , 3.070473538723329e-45 ], [ 1022.6875 , 3.0708258678447916e-45 ], 
                    [ 1022.75 , 3.0711782373951143e-45 ], [ 1022.8125 , 3.0715306473789785e-45 ], [ 1022.875 , 3.071883097801003e-45 ], [ 1022.9375 , 3.072235588665828e-45 ], 
                    [ 1023.0 , 3.072588119978027e-45 ], [ 1023.0625 , 3.072940691742396e-45 ], [ 1023.125 , 3.0732933039634675e-45 ], [ 1023.1875 , 3.073645956645883e-45 ], 
                    [ 1023.25 , 3.0739986497943285e-45 ], [ 1023.3125 , 3.074351383413384e-45 ], [ 1023.375 , 3.0747041575078e-45 ], [ 1023.4375 , 3.075056972082114e-45 ], 
                    [ 1023.5 , 3.0754098271410355e-45 ], [ 1023.5625 , 3.0757627226891875e-45 ], [ 1023.625 , 3.076115658731174e-45 ], [ 1023.6875 , 3.076468635271727e-45 ], 
                    [ 1023.75 , 3.076821652315451e-45 ], [ 1023.8125 , 3.077174709866994e-45 ], [ 1023.875 , 3.07752780793096e-45 ], [ 1023.9375 , 3.0778809465120845e-45 ], 
                    [ 1024.0 , 0.0 ]])

    f1 = data[:,0]
    S1 = data[:,1]
    fstop = [f1[0],f1[-1]]

    #% ---- Interpolate to desired frequencies.  Set noise to infinity outside
    #%      band of frequencies covered by f1.
    ##S = inf(size(f)); 
    ##k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop




def ligo2016():


    data = np.array([[ 20.0 , 3.349892833904415e-45 ], [ 20.0625 , 3.3087874665616637e-45 ], [ 20.125 , 3.268336658266442e-45 ], [ 20.1875 , 3.228504941706327e-45 ], 
                    [ 20.25 , 3.189293446809069e-45 ], [ 20.3125 , 3.150692074425704e-45 ], [ 20.375 , 3.1126768122296244e-45 ], [ 20.4375 , 3.075253188948464e-45 ], 
                    [ 20.5 , 3.038401640621744e-45 ], [ 20.5625 , 3.002104418216431e-45 ], [ 20.625 , 2.9663701500366824e-45 ], [ 20.6875 , 2.931174112311855e-45 ], 
                    [ 20.75 , 2.89650341328305e-45 ], [ 20.8125 , 2.862365164098965e-45 ], [ 20.875 , 2.828736860442717e-45 ], [ 20.9375 , 2.7956053137594827e-45 ], 
                    [ 21.0 , 2.7629770533111635e-45 ], [ 21.0625 , 2.7308330275953395e-45 ], [ 21.125 , 2.699160118702457e-45 ], [ 21.1875 , 2.6679595973791245e-45 ], 
                    [ 21.25 , 2.6372223886981946e-45 ], [ 21.3125 , 2.6069319028140372e-45 ], [ 21.375 , 2.5770833012257257e-45 ], [ 21.4375 , 2.5476797561937448e-45 ], 
                    [ 21.5 , 2.5186997411787173e-45 ], [ 21.5625 , 2.49013633883137e-45 ], [ 21.625 , 2.4619925112352822e-45 ], [ 21.6875 , 2.434255974076136e-45 ], 
                    [ 21.75 , 2.406914196670252e-45 ], [ 21.8125 , 2.3799617259194665e-45 ], [ 21.875 , 2.353404785845167e-45 ], [ 21.9375 , 2.3272230376799977e-45 ], 
                    [ 22.0 , 2.3014103359955458e-45 ], [ 22.0625 , 2.275962482321079e-45 ], [ 22.125 , 2.2508821561086825e-45 ], [ 22.1875 , 2.2261523108807596e-45 ], 
                    [ 22.25 , 2.2017671471796648e-45 ], [ 22.3125 , 2.1777218223237714e-45 ], [ 22.375 , 2.1540207322724886e-45 ], [ 22.4375 , 2.1306473067909935e-45 ], 
                    [ 22.5 , 2.1075962429002086e-45 ], [ 22.5625 , 2.084862284258581e-45 ], [ 22.625 , 2.0624475733056374e-45 ], [ 22.6875 , 2.0403425304231136e-45 ], 
                    [ 22.75 , 2.0185391247142342e-45 ], [ 22.8125 , 1.9970323775047067e-45 ], [ 22.875 , 1.9758178865040247e-45 ], [ 22.9375 , 1.9548995346396205e-45 ], 
                    [ 23.0 , 1.9342632871175142e-45 ], [ 23.0625 , 1.9139048741767587e-45 ], [ 23.125 , 1.8938200747257005e-45 ], [ 23.1875 , 1.8740044542421015e-45 ], 
                    [ 23.25 , 1.8544615499603602e-45 ], [ 23.3125 , 1.83517883671544e-45 ], [ 23.375 , 1.816152472290482e-45 ], [ 23.4375 , 1.7973782690303236e-45 ], 
                    [ 23.5 , 1.778852473600392e-45 ], [ 23.5625 , 1.7605744028049418e-45 ], [ 23.625 , 1.7425403181488424e-45 ], [ 23.6875 , 1.7247425422054293e-45 ], 
                    [ 23.75 , 1.7071778306101184e-45 ], [ 23.8125 , 1.6898424713665347e-45 ], [ 23.875 , 1.672732707688321e-45 ], [ 23.9375 , 1.655847428194093e-45 ], 
                    [ 24.0 , 1.6391843697485864e-45 ], [ 24.0625 , 1.6227367372167613e-45 ], [ 24.125 , 1.6065008860910626e-45 ], [ 24.1875 , 1.5904738553586595e-45 ], 
                    [ 24.25 , 1.574652169651218e-45 ], [ 24.3125 , 1.559033216389884e-45 ], [ 24.375 , 1.5436153924231447e-45 ], [ 24.4375 , 1.5283964868063573e-45 ], 
                    [ 24.5 , 1.513370724781221e-45 ], [ 24.5625 , 1.4985352452514463e-45 ], [ 24.625 , 1.4838870337830758e-45 ], [ 24.6875 , 1.4694234943653354e-45 ], 
                    [ 24.75 , 1.4551417033054063e-45 ], [ 24.8125 , 1.4410391547827978e-45 ], [ 24.875 , 1.4271138562215977e-45 ], [ 24.9375 , 1.413364887493447e-45 ], 
                    [ 25.0 , 1.3997870668244835e-45 ], [ 25.0625 , 1.3863777301371898e-45 ], [ 25.125 , 1.373134398188047e-45 ], [ 25.1875 , 1.3600547001624588e-45 ], 
                    [ 25.25 , 1.3471361927076377e-45 ], [ 25.3125 , 1.334376385429566e-45 ], [ 25.375 , 1.3217732383010426e-45 ], [ 25.4375 , 1.3093241369053904e-45 ], 
                    [ 25.5 , 1.297027120638049e-45 ], [ 25.5625 , 1.2848819875744476e-45 ], [ 25.625 , 1.2728845685505967e-45 ], [ 25.6875 , 1.2610324128460144e-45 ], 
                    [ 25.75 , 1.2493234621678029e-45 ], [ 25.8125 , 1.2377557677610255e-45 ], [ 25.875 , 1.226327138923577e-45 ], [ 25.9375 , 1.215035586324184e-45 ], 
                    [ 26.0 , 1.2038792006532381e-45 ], [ 26.0625 , 1.1928560868420993e-45 ], [ 26.125 , 1.1819642619416844e-45 ], [ 26.1875 , 1.1712018487533178e-45 ], 
                    [ 26.25 , 1.1605671661349622e-45 ], [ 26.3125 , 1.1500581955049092e-45 ], [ 26.375 , 1.1396732282017763e-45 ], [ 26.4375 , 1.1294106924429595e-45 ], 
                    [ 26.5 , 1.1192685792470399e-45 ], [ 26.5625 , 1.1092457053678033e-45 ], [ 26.625 , 1.0993406223123954e-45 ], [ 26.6875 , 1.0895510669609706e-45 ], 
                    [ 26.75 , 1.0798752378078506e-45 ], [ 26.8125 , 1.0703116956731556e-45 ], [ 26.875 , 1.0608589118775295e-45 ], [ 26.9375 , 1.0515152481550625e-45 ], 
                    [ 27.0 , 1.042279285731338e-45 ], [ 27.0625 , 1.0331494043610395e-45 ], [ 27.125 , 1.024124337863321e-45 ], [ 27.1875 , 1.0152024506031906e-45 ], 
                    [ 27.25 , 1.0063824596705193e-45 ], [ 27.3125 , 9.97662872022475e-46 ], [ 27.375 , 9.890423321486234e-46 ], [ 27.4375 , 9.805195082623188e-46 ], 
                    [ 27.5 , 9.720931416814139e-46 ], [ 27.5625 , 9.637617732562383e-46 ], [ 27.625 , 9.555241738192716e-46 ], [ 27.6875 , 9.473792057228527e-46 ], 
                    [ 27.75 , 9.393253920495157e-46 ], [ 27.8125 , 9.313616240053037e-46 ], [ 27.875 , 9.234867428582683e-46 ], [ 27.9375 , 9.156994658322907e-46 ], 
                    [ 28.0 , 9.079985678984864e-46 ], [ 28.0625 , 9.003830386971057e-46 ], [ 28.125 , 8.928516733775443e-46 ], [ 28.1875 , 8.854033330314703e-46 ], 
                    [ 28.25 , 8.780369533031399e-46 ], [ 28.3125 , 8.707513427130313e-46 ], [ 28.375 , 8.635455249876797e-46 ], [ 28.4375 , 8.564183803256355e-46 ], 
                    [ 28.5 , 8.493690288538092e-46 ], [ 28.5625 , 8.423962249293984e-46 ], [ 28.625 , 8.354991512842735e-46 ], [ 28.6875 , 8.286766083891649e-46 ], 
                    [ 28.75 , 8.219284175876878e-46 ], [ 28.8125 , 8.152531185736974e-46 ], [ 28.875 , 8.086495658664458e-46 ], [ 28.9375 , 8.021169625989033e-46 ], 
                    [ 29.0 , 7.956542890814263e-46 ], [ 29.0625 , 7.892606329519241e-46 ], [ 29.125 , 7.82935040027916e-46 ], [ 29.1875 , 7.766768000358578e-46 ], 
                    [ 29.25 , 7.704849092915596e-46 ], [ 29.3125 , 7.643585193568828e-46 ], [ 29.375 , 7.582968920011127e-46 ], [ 29.4375 , 7.522991376684416e-46 ], 
                    [ 29.5 , 7.463643721821522e-46 ], [ 29.5625 , 7.404918630354912e-46 ], [ 29.625 , 7.346808543194792e-46 ], [ 29.6875 , 7.289304884075836e-46 ], 
                    [ 29.75 , 7.232399368935133e-46 ], [ 29.8125 , 7.176086010772399e-46 ], [ 29.875 , 7.12036877531449e-46 ], [ 29.9375 , 7.065228821208915e-46 ], 
                    [ 30.0 , 7.010658125699367e-46 ], [ 30.0625 , 6.9566501618280905e-46 ], [ 30.125 , 6.903196814010715e-46 ], [ 30.1875 , 6.850291841105482e-46 ], 
                    [ 30.25 , 6.797928477461333e-46 ], [ 30.3125 , 6.746099300404742e-46 ], [ 30.375 , 6.694797846819827e-46 ], [ 30.4375 , 6.644018238549547e-46 ], 
                    [ 30.5 , 6.593753287653729e-46 ], [ 30.5625 , 6.544000489185424e-46 ], [ 30.625 , 6.494761804354096e-46 ], [ 30.6875 , 6.446019916632285e-46 ], 
                    [ 30.75 , 6.3977672727401535e-46 ], [ 30.8125 , 6.3499984879542155e-46 ], [ 30.875 , 6.302707934877678e-46 ], [ 30.9375 , 6.255889547014478e-46 ], 
                    [ 31.0 , 6.2095375869403145e-46 ], [ 31.0625 , 6.163646604822592e-46 ], [ 31.125 , 6.1182097798130624e-46 ], [ 31.1875 , 6.073235307699402e-46 ], 
                    [ 31.25 , 6.028710048799671e-46 ], [ 31.3125 , 5.984623231525461e-46 ], [ 31.375 , 5.940969648971047e-46 ], [ 31.4375 , 5.897744330753228e-46 ], 
                    [ 31.5 , 5.854942015906696e-46 ], [ 31.5625 , 5.812557416094639e-46 ], [ 31.625 , 5.770585892505172e-46 ], [ 31.6875 , 5.729031022968548e-46 ], 
                    [ 31.75 , 5.6878884391579355e-46 ], [ 31.8125 , 5.6471446955657454e-46 ], [ 31.875 , 5.606794099128303e-46 ], [ 31.9375 , 5.566832457890067e-46 ], 
                    [ 32.0 , 5.5272552144861074e-46 ], [ 32.0625 , 5.488057073177676e-46 ], [ 32.125 , 5.449234391181298e-46 ], [ 32.1875 , 5.410801046488049e-46 ], 
                    [ 32.25 , 5.372734383203847e-46 ], [ 32.3125 , 5.335029709000523e-46 ], [ 32.375 , 5.297682657803857e-46 ], [ 32.4375 , 5.260689183121748e-46 ], 
                    [ 32.5 , 5.2240452609693204e-46 ], [ 32.5625 , 5.187749630896606e-46 ], [ 32.625 , 5.151811632061158e-46 ], [ 32.6875 , 5.116210955937718e-46 ], 
                    [ 32.75 , 5.0809434938274035e-46 ], [ 32.8125 , 5.046005438643078e-46 ], [ 32.875 , 5.011393141710629e-46 ], [ 32.9375 , 4.977102931771735e-46 ], 
                    [ 33.0 , 4.9431466694710606e-46 ], [ 33.0625 , 4.909508938979007e-46 ], [ 33.125 , 4.876182054847829e-46 ], [ 33.1875 , 4.8431622549121155e-46 ], 
                    [ 33.25 , 4.810445970923776e-46 ], [ 33.3125 , 4.778029892736711e-46 ], [ 33.375 , 4.745927740588325e-46 ], [ 33.4375 , 4.7141223483184174e-46 ], 
                    [ 33.5 , 4.682607014746826e-46 ], [ 33.5625 , 4.651378385451988e-46 ], [ 33.625 , 4.6204331098856e-46 ], [ 33.6875 , 4.589770980870762e-46 ], 
                    [ 33.75 , 4.559403820887697e-46 ], [ 33.8125 , 4.529310317705142e-46 ], [ 33.875 , 4.499487508457628e-46 ], [ 33.9375 , 4.469932336327928e-46 ], 
                    [ 34.0 , 4.440641730856754e-46 ], [ 34.0625 , 4.4116284640334474e-46 ], [ 34.125 , 4.382879364691158e-46 ], [ 34.1875 , 4.354386036640705e-46 ], 
                    [ 34.25 , 4.326145799507617e-46 ], [ 34.3125 , 4.298155771255498e-46 ], [ 34.375 , 4.270423607033512e-46 ], [ 34.4375 , 4.242946630608768e-46 ], 
                    [ 34.5 , 4.215711313318351e-46 ], [ 34.5625 , 4.1887149383833236e-46 ], [ 34.625 , 4.161954911096803e-46 ], [ 34.6875 , 4.135439716593057e-46 ], 
                    [ 34.75 , 4.109166380871688e-46 ], [ 34.8125 , 4.083121545275164e-46 ], [ 34.875 , 4.057302681174975e-46 ], [ 34.9375 , 4.031707337994279e-46 ], 
                    [ 35.0 , 4.006347737831042e-46 ], [ 35.0625 , 3.981213820769583e-46 ], [ 35.125 , 3.95629616004895e-46 ], [ 35.1875 , 3.9315923708601695e-46 ], 
                    [ 35.25 , 3.907100097803129e-46 ], [ 35.3125 , 3.882838887908055e-46 ], [ 35.375 , 3.858784381471385e-46 ], [ 35.4375 , 3.834934484791492e-46 ], 
                    [ 35.5 , 3.811287009034743e-46 ], [ 35.5625 , 3.787849256975951e-46 ], [ 35.625 , 3.764621747969791e-46 ], [ 35.6875 , 3.741590103736566e-46 ], 
                    [ 35.75 , 3.718752139720651e-46 ], [ 35.8125 , 3.696106376023252e-46 ], [ 35.875 , 3.6736713262065534e-46 ], [ 35.9375 , 3.651423876168696e-46 ], 
                    [ 36.0 , 3.629361813955562e-46 ], [ 36.0625 , 3.60748301378565e-46 ], [ 36.125 , 3.585801464095102e-46 ], [ 36.1875 , 3.564305589240774e-46 ], 
                    [ 36.25 , 3.5429872702623955e-46 ], [ 36.3125 , 3.5218446174440805e-46 ], [ 36.375 , 3.500887123027355e-46 ], [ 36.4375 , 3.48011192799827e-46 ], 
                    [ 36.5 , 3.459506805349317e-46 ], [ 36.5625 , 3.439069967306835e-46 ], [ 36.625 , 3.4188082169522026e-46 ], [ 36.6875 , 3.398724628811146e-46 ], 
                    [ 36.75 , 3.3788037028172916e-46 ], [ 36.8125 , 3.3590437492390592e-46 ], [ 36.875 , 3.339451562893992e-46 ], [ 36.9375 , 3.320031109306174e-46 ], 
                    [ 37.0 , 3.300766569524918e-46 ], [ 37.0625 , 3.281656315397523e-46 ], [ 37.125 , 3.2627082722917593e-46 ], [ 37.1875 , 3.243924021081637e-46 ], 
                    [ 37.25 , 3.2252893209808857e-46 ], [ 37.3125 , 3.2068025909990882e-46 ], [ 37.375 , 3.188474394238479e-46 ], [ 37.4375 , 3.1703008045353446e-46 ], 
                    [ 37.5 , 3.1522705291762388e-46 ], [ 37.5625 , 3.1343821058133923e-46 ], [ 37.625 , 3.1166506919660547e-46 ], [ 37.6875 , 3.099063395968442e-46 ], 
                    [ 37.75 , 3.0816135172069522e-46 ], [ 37.8125 , 3.0643001473302557e-46 ], [ 37.875 , 3.047142740255173e-46 ], [ 37.9375 , 3.0301186050337335e-46 ], 
                    [ 38.0 , 3.0132262279479597e-46 ], [ 38.0625 , 2.9964715359906767e-46 ], [ 38.125 , 2.979860347433917e-46 ], [ 38.1875 , 2.963376755861066e-46 ], 
                    [ 38.25 , 2.9470194666544254e-46 ], [ 38.3125 , 2.9308032621697287e-46 ], [ 38.375 , 2.9147167299341737e-46 ], [ 38.4375 , 2.8987525743472585e-46 ], 
                    [ 38.5 , 2.8829131210805784e-46 ], [ 38.5625 , 2.867211981495048e-46 ], [ 38.625 , 2.8516294984467814e-46 ], [ 38.6875 , 2.836164452335824e-46 ], 
                    [ 38.75 , 2.8208298164447555e-46 ], [ 38.8125 , 2.805617862654204e-46 ], [ 38.875 , 2.7905197408705296e-46 ], [ 38.9375 , 2.7755384064301467e-46 ], 
                    [ 39.0 , 2.7606862035562254e-46 ], [ 39.0625 , 2.7459444648140812e-46 ], [ 39.125 , 2.731312047419517e-46 ], [ 39.1875 , 2.7168042499493417e-46 ], 
                    [ 39.25 , 2.7024084824631908e-46 ], [ 39.3125 , 2.688118701632965e-46 ], [ 39.375 , 2.673942528292329e-46 ], [ 39.4375 , 2.659882966207553e-46 ], 
                    [ 39.5 , 2.6459260726296145e-46 ], [ 39.5625 , 2.6320724418787658e-46 ], [ 39.625 , 2.618339309411343e-46 ], [ 39.6875 , 2.6047058141903966e-46 ], 
                    [ 39.75 , 2.5911709861257962e-46 ], [ 39.8125 , 2.5777507846689947e-46 ], [ 39.875 , 2.5644317308180728e-46 ], [ 39.9375 , 2.551208190679564e-46 ], 
                    [ 40.0 , 2.5380905519236515e-46 ], [ 40.0625 , 2.525076700404026e-46 ], [ 40.125 , 2.5121556247507398e-46 ], [ 40.1875 , 2.499333233760137e-46 ], 
                    [ 40.25 , 2.48661607283969e-46 ], [ 40.3125 , 2.473988884204552e-46 ], [ 40.375 , 2.4614538077564033e-46 ], [ 40.4375 , 2.449025285361583e-46 ], 
                    [ 40.5 , 2.436683902557721e-46 ], [ 40.5625 , 2.424428811090524e-46 ], [ 40.625 , 2.4122804085534502e-46 ], [ 40.6875 , 2.4002167520242396e-46 ], 
                    [ 40.75 , 2.3882368339796474e-46 ], [ 40.8125 , 2.376358497984219e-46 ], [ 40.875 , 2.364564789855871e-46 ], [ 40.9375 , 2.3528523271872557e-46 ], 
                    [ 41.0 , 2.3412372330057427e-46 ], [ 41.0625 , 2.329706169271925e-46 ], [ 41.125 , 2.318253958009489e-46 ], [ 41.1875 , 2.3068952675101123e-46 ], 
                    [ 41.25 , 2.2956193150921252e-46 ], [ 41.3125 , 2.284419884891799e-46 ], [ 41.375 , 2.2733113169070783e-46 ], [ 41.4375 , 2.2622836506032606e-46 ], 
                    [ 41.5 , 2.2513302078240935e-46 ], [ 41.5625 , 2.2404653644112535e-46 ], [ 41.625 , 2.229679140996972e-46 ], [ 41.6875 , 2.218964929920316e-46 ], 
                    [ 41.75 , 2.2083375792890546e-46 ], [ 41.8125 , 2.1977861211473283e-46 ], [ 41.875 , 2.18730457032606e-46 ], [ 41.9375 , 2.1769088102842306e-46 ], 
                    [ 42.0 , 2.1665857728285473e-46 ], [ 42.0625 , 2.1563306298549212e-46 ], [ 42.125 , 2.1461608114148855e-46 ], [ 42.1875 , 2.136059957699258e-46 ], 
                    [ 42.25 , 2.1260249453515627e-46 ], [ 42.3125 , 2.1160754253262942e-46 ], [ 42.375 , 2.106190743519213e-46 ], [ 42.4375 , 2.0963721628848775e-46 ], 
                    [ 42.5 , 2.0866352242571252e-46 ], [ 42.5625 , 2.076960903540284e-46 ], [ 42.625 , 2.0673538747670433e-46 ], [ 42.6875 , 2.0578235356396315e-46 ], 
                    [ 42.75 , 2.0483539540929043e-46 ], [ 42.8125 , 2.0389531977547716e-46 ], [ 42.875 , 2.0296237352245612e-46 ], [ 42.9375 , 2.0203533432960846e-46 ], 
                    [ 43.0 , 2.0111540863752235e-46 ], [ 43.0625 , 2.002020222037093e-46 ], [ 43.125 , 1.9929435951639885e-46 ], [ 43.1875 , 1.9839405881552153e-46 ], 
                    [ 43.25 , 1.9749971906188274e-46 ], [ 43.3125 , 1.9661108336822713e-46 ], [ 43.375 , 1.9572981273335013e-46 ], [ 43.4375 , 1.948539974732401e-46 ], 
                    [ 43.5 , 1.9398422670371854e-46 ], [ 43.5625 , 1.9312116022333617e-46 ], [ 43.625 , 1.922633868262813e-46 ], [ 43.6875 , 1.9141203477693375e-46 ], 
                    [ 43.75 , 1.905666984001756e-46 ], [ 43.8125 , 1.8972650388364257e-46 ], [ 43.875 , 1.8889313188778435e-46 ], [ 43.9375 , 1.8806504216921355e-46 ], 
                    [ 44.0 , 1.87242305395729e-46 ], [ 44.0625 , 1.864261193123672e-46 ], [ 44.125 , 1.8561482572086846e-46 ], [ 44.1875 , 1.848093817103141e-46 ], 
                    [ 44.25 , 1.8400970548952817e-46 ], [ 44.3125 , 1.832147783500931e-46 ], [ 44.375 , 1.8242620968753164e-46 ], [ 44.4375 , 1.81642599890441e-46 ], 
                    [ 44.5 , 1.8086400479805165e-46 ], [ 44.5625 , 1.80091533531199e-46 ], [ 44.625 , 1.7932358072597613e-46 ], [ 44.6875 , 1.78561223630771e-46 ], 
                    [ 44.75 , 1.7780411936666432e-46 ], [ 44.8125 , 1.7705140393748293e-46 ], [ 44.875 , 1.7630491354172753e-46 ], [ 44.9375 , 1.7556279136721283e-46 ], 
                    [ 45.0 , 1.7482564060249374e-46 ], [ 45.0625 , 1.7409392439865573e-46 ], [ 45.125 , 1.7336638675528197e-46 ], [ 45.1875 , 1.7264450895455064e-46 ], 
                    [ 45.25 , 1.7192713331701666e-46 ], [ 45.3125 , 1.7121423268476588e-46 ], [ 45.375 , 1.7050684610140647e-46 ], [ 45.4375 , 1.6980343694512554e-46 ], 
                    [ 45.5 , 1.6910523926584612e-46 ], [ 45.5625 , 1.684115736359614e-46 ], [ 45.625 , 1.677220095507583e-46 ], [ 45.6875 , 1.6703792866732767e-46 ], 
                    [ 45.75 , 1.6635763360475858e-46 ], [ 45.8125 , 1.656822272720176e-46 ], [ 45.875 , 1.650112868664229e-46 ], [ 45.9375 , 1.6434418953406914e-46 ], 
                    [ 46.0 , 1.6368245049639375e-46 ], [ 46.0625 , 1.6302431452264367e-46 ], [ 46.125 , 1.6237085266153129e-46 ], [ 46.1875 , 1.6172169553014778e-46 ], 
                    [ 46.25 , 1.610762271968385e-46 ], [ 46.3125 , 1.6043590155270948e-46 ], [ 46.375 , 1.5979900694236163e-46 ], [ 46.4375 , 1.5916668020160879e-46 ], 
                    [ 46.5 , 1.5853840994794838e-46 ], [ 46.5625 , 1.5791377281598722e-46 ], [ 46.625 , 1.5729398183221988e-46 ], [ 46.6875 , 1.5667745577444667e-46 ], 
                    [ 46.75 , 1.5606548053235255e-46 ], [ 46.8125 , 1.5545723719865341e-46 ], [ 46.875 , 1.5485266697296538e-46 ], [ 46.9375 , 1.5425256947900103e-46 ], 
                    [ 47.0 , 1.536555797227613e-46 ], [ 47.0625 , 1.5306321872107404e-46 ], [ 47.125 , 1.524741877098056e-46 ], [ 47.1875 , 1.518889538443194e-46 ], 
                    [ 47.25 , 1.5130774914982856e-46 ], [ 47.3125 , 1.5072954045602647e-46 ], [ 47.375 , 1.5015604187316667e-46 ], [ 47.4375 , 1.4958544101770984e-46 ], 
                    [ 47.5 , 1.4901884211536056e-46 ], [ 47.5625 , 1.4845576192397474e-46 ], [ 47.625 , 1.4789593103682523e-46 ], [ 47.6875 , 1.4734025512897185e-46 ], 
                    [ 47.75 , 1.4678733809609223e-46 ], [ 47.8125 , 1.4623870532348738e-46 ], [ 47.875 , 1.4569301654755203e-46 ], [ 47.9375 , 1.4515090173177747e-46 ], 
                    [ 48.0 , 1.4461233024183463e-46 ], [ 48.0625 , 1.4407664548438657e-46 ], [ 48.125 , 1.4354508042674297e-46 ], [ 48.1875 , 1.4301608736274356e-46 ], 
                    [ 48.25 , 1.424910584769288e-46 ], [ 48.3125 , 1.4196890507177775e-46 ], [ 48.375 , 1.414500742164799e-46 ], [ 48.4375 , 1.4093466013680973e-46 ], 
                    [ 48.5 , 1.4042193957040828e-46 ], [ 48.5625 , 1.3991315343263507e-46 ], [ 48.625 , 1.3940676482240314e-46 ], [ 48.6875 , 1.3890418773436533e-46 ], 
                    [ 48.75 , 1.3840427413491798e-46 ], [ 48.8125 , 1.3790758835248603e-46 ], [ 48.875 , 1.3741405588025173e-46 ], [ 48.9375 , 1.3692317958041774e-46 ], 
                    [ 49.0 , 1.3643592337347533e-46 ], [ 49.0625 , 1.3595090230638968e-46 ], [ 49.125 , 1.3546969452304488e-46 ], [ 49.1875 , 1.3499080779778274e-46 ], 
                    [ 49.25 , 1.3451519681448708e-46 ], [ 49.3125 , 1.340423480372317e-46 ], [ 49.375 , 1.3357225805191746e-46 ], [ 49.4375 , 1.3310535901351154e-46 ], 
                    [ 49.5 , 1.3264072114623786e-46 ], [ 49.5625 , 1.3217967478743939e-46 ], [ 49.625 , 1.3172066861405974e-46 ], [ 49.6875 , 1.3126512625528825e-46 ], 
                    [ 49.75 , 1.3081184521993086e-46 ], [ 49.8125 , 1.3036155919195826e-46 ], [ 49.875 , 1.2991391231680723e-46 ], [ 49.9375 , 1.294688048652391e-46 ], 
                    [ 50.0 , 1.2902670384279288e-46 ], [ 50.0625 , 1.2858670872687505e-46 ], [ 50.125 , 1.2815007697487364e-46 ], [ 50.1875 , 1.2771530636973175e-46 ], 
                    [ 50.25 , 1.2728387816516362e-46 ], [ 50.3125 , 1.2685444675501108e-46 ], [ 50.375 , 1.264279560244806e-46 ], [ 50.4375 , 1.2600378796496484e-46 ], 
                    [ 50.5 , 1.255821665238751e-46 ], [ 50.5625 , 1.2516317526047286e-46 ], [ 50.625 , 1.2474635029352822e-46 ], [ 50.6875 , 1.243324639986824e-46 ], 
                    [ 50.75 , 1.2392038593635646e-46 ], [ 50.8125 , 1.2351152278380283e-46 ], [ 50.875 , 1.2310432030750865e-46 ], [ 50.9375 , 1.2270020996579205e-46 ], 
                    [ 51.0 , 1.2229792562667585e-46 ], [ 51.0625 , 1.2189838891402084e-46 ], [ 51.125 , 1.2150094288680157e-46 ], [ 51.1875 , 1.211059220633612e-46 ], 
                    [ 51.25 , 1.2071324367014847e-46 ], [ 51.3125 , 1.2032268006572633e-46 ], [ 51.375 , 1.1993469529721728e-46 ], [ 51.4375 , 1.1954852656525825e-46 ], 
                    [ 51.5 , 1.1916517366081255e-46 ], [ 51.5625 , 1.1878335476519458e-46 ], [ 51.625 , 1.1840455054817158e-46 ], [ 51.6875 , 1.1802719416791887e-46 ], 
                    [ 51.75 , 1.1765269507864426e-46 ], [ 51.8125 , 1.1727979977090137e-46 ], [ 51.875 , 1.1690949605428764e-46 ], [ 51.9375 , 1.1654099117060434e-46 ], 
                    [ 52.0 , 1.1617482094589073e-46 ], [ 52.0625 , 1.1581063917142483e-46 ], [ 52.125 , 1.154485571694911e-46 ], [ 52.1875 , 1.15088639098922e-46 ], 
                    [ 52.25 , 1.147305919683081e-46 ], [ 52.3125 , 1.1437487314414144e-46 ], [ 52.375 , 1.1402080786807768e-46 ], [ 52.4375 , 1.1366922342350767e-46 ], 
                    [ 52.5 , 1.133190862147397e-46 ], [ 52.5625 , 1.1297157257509693e-46 ], [ 52.625 , 1.1262531148502604e-46 ], [ 52.6875 , 1.1228181909763682e-46 ], 
                    [ 52.75 , 1.1193953076768767e-46 ], [ 52.8125 , 1.1159985467417294e-46 ], [ 52.875 , 1.1126149188588545e-46 ], [ 52.9375 , 1.109255684017723e-46 ], 
                    [ 53.0 , 1.1059107794039796e-46 ], [ 53.0625 , 1.1025885559056193e-46 ], [ 53.125 , 1.0992817752358932e-46 ], [ 53.1875 , 1.0959960986755828e-46 ], 
                    [ 53.25 , 1.0927268917416891e-46 ], [ 53.3125 , 1.0894773167879571e-46 ], [ 53.375 , 1.0862451738829001e-46 ], [ 53.4375 , 1.0830313097269782e-46 ], 
                    [ 53.5 , 1.0798356622630014e-46 ], [ 53.5625 , 1.0766569416477977e-46 ], [ 53.625 , 1.073497219141935e-46 ], [ 53.6875 , 1.0703532428568666e-46 ], 
                    [ 53.75 , 1.067228975872571e-46 ], [ 53.8125 , 1.0641193640781195e-46 ], [ 53.875 , 1.0610300030153487e-46 ], [ 53.9375 , 1.0579542502920803e-46 ], 
                    [ 54.0 , 1.0548993206756476e-46 ], [ 54.0625 , 1.0518570843375445e-46 ], [ 54.125 , 1.0488360409170957e-46 ], [ 54.1875 , 1.0458268205161863e-46 ], 
                    [ 54.25 , 1.0428392223797746e-46 ], [ 54.3125 , 1.0398626939622795e-46 ], [ 54.375 , 1.0369080344168869e-46 ], [ 54.4375 , 1.0339637496023396e-46 ], 
                    [ 54.5 , 1.0310415351672365e-46 ], [ 54.5625 , 1.028129085939667e-46 ], [ 54.625 , 1.0252388615733748e-46 ], [ 54.6875 , 1.0223578731173745e-46 ], 
                    [ 54.75 , 1.0194991985385382e-46 ], [ 54.8125 , 1.0166493055974339e-46 ], [ 54.875 , 1.0138217123235273e-46 ], [ 54.9375 , 1.0110025188971496e-46 ], 
                    [ 55.0 , 1.0082055401926717e-46 ], [ 55.0625 , 1.0054166524364812e-46 ], [ 55.125 , 1.0026499008598434e-46 ], [ 55.1875 , 9.998910037059033e-47 ], 
                    [ 55.25 , 9.971539803467875e-47 ], [ 55.3125 , 9.944246429062616e-47 ], [ 55.375 , 9.917169819908169e-47 ], [ 55.4375 , 9.890169078362952e-47 ], 
                    [ 55.5 , 9.863381146835544e-47 ], [ 55.5625 , 9.83666873868865e-47 ], [ 55.625 , 9.810166460134337e-47 ], [ 55.6875 , 9.783740050175443e-47 ], 
                    [ 55.75 , 9.75751822156991e-47 ], [ 55.8125 , 9.731373172647599e-47 ], [ 55.875 , 9.705428175456308e-47 ], [ 55.9375 , 9.679561552781679e-47 ], 
                    [ 56.0 , 9.653890105781323e-47 ], [ 56.0625 , 9.628299275022824e-47 ], [ 56.125 , 9.602896724531397e-47 ], [ 56.1875 , 9.577577480761148e-47 ], 
                    [ 56.25 , 9.552440368551365e-47 ], [ 56.3125 , 9.527389990106237e-47 ], [ 56.375 , 9.502514335129323e-47 ], [ 56.4375 , 9.47772928684678e-47 ], 
                    [ 56.5 , 9.453111288259345e-47 ], [ 56.5625 , 9.428588360304565e-47 ], [ 56.625 , 9.404224781080168e-47 ], [ 56.6875 , 9.379961453165628e-47 ], 
                    [ 56.75 , 9.355848658835576e-47 ], [ 56.8125 , 9.33184160553786e-47 ], [ 56.875 , 9.307975818073034e-47 ], [ 56.9375 , 9.284221749605178e-47 ], 
                    [ 57.0 , 9.260599623715826e-47 ], [ 57.0625 , 9.237095968540113e-47 ], [ 57.125 , 9.213714154338594e-47 ], [ 57.1875 , 9.190457824152675e-47 ], 
                    [ 57.25 , 9.16731271404477e-47 ], [ 57.3125 , 9.14430042632734e-47 ], [ 57.375 , 9.121388728800704e-47 ], [ 57.4375 , 9.098618311564716e-47 ], 
                    [ 57.5 , 9.07593697172729e-47 ], [ 57.5625 , 9.053404924710785e-47 ], [ 57.625 , 9.030950404671237e-47 ], [ 57.6875 , 9.008654736107456e-47 ], 
                    [ 57.75 , 8.986424378324259e-47 ], [ 57.8125 , 8.964361663710766e-47 ], [ 57.875 , 8.942353115621822e-47 ], [ 57.9375 , 8.920519859831665e-47 ], 
                    [ 58.0 , 8.898741249631094e-47 ], [ 58.0625 , 8.877123667485966e-47 ], [ 58.125 , 8.855571881657549e-47 ], [ 58.1875 , 8.834167381852112e-47 ], 
                    [ 58.25 , 8.81284034462596e-47 ], [ 58.3125 , 8.791645588783775e-47 ], [ 58.375 , 8.77054009111165e-47 ], [ 58.4375 , 8.749552392074459e-47 ], 
                    [ 58.5 , 8.728666537345942e-47 ], [ 58.5625 , 8.707882916158333e-47 ], [ 58.625 , 8.687213508144847e-47 ], [ 58.6875 , 8.666631057358362e-47 ], 
                    [ 58.75 , 8.646176303032855e-47 ], [ 58.8125 , 8.62579252839494e-47 ], [ 58.875 , 8.605549596330601e-47 ], [ 58.9375 , 8.5853614910288215e-47 ], 
                    [ 59.0 , 8.565328020842937e-47 ], [ 59.0625 , 8.545341297660837e-47 ], [ 59.125 , 8.5255065727182215e-47 ], [ 59.1875 , 8.505726231807777e-47 ], 
                    [ 59.25 , 8.486080532972243e-47 ], [ 59.3125 , 8.466504561021444e-47 ], [ 59.375 , 8.447044960297704e-47 ], [ 59.4375 , 8.4276705223317e-47 ], 
                    [ 59.5 , 8.408394498571496e-47 ], [ 59.5625 , 8.389219647217121e-47 ], [ 59.625 , 8.370124793343938e-47 ], [ 59.6875 , 8.351147312019995e-47 ], 
                    [ 59.75 , 8.332231154097264e-47 ], [ 59.8125 , 8.313448861185789e-47 ], [ 59.875 , 8.294708964163043e-47 ], [ 59.9375 , 8.276119618137769e-47 ], 
                    [ 60.0 , 8.257571932861187e-47 ], [ 60.0625 , 8.2391545522693e-47 ], [ 60.125 , 8.220796503210712e-47 ], [ 60.1875 , 8.202549449141947e-47 ], 
                    [ 60.25 , 8.184380105804694e-47 ], [ 60.3125 , 8.166300703263191e-47 ], [ 60.375 , 8.148316837917292e-47 ], [ 60.4375 , 8.130402553838493e-47 ], 
                    [ 60.5 , 8.112602633832323e-47 ], [ 60.5625 , 8.094851478747306e-47 ], [ 60.625 , 8.077233406193812e-47 ], [ 60.6875 , 8.059653678566322e-47 ], 
                    [ 60.75 , 8.042204598054776e-47 ], [ 60.8125 , 8.02480378764609e-47 ], [ 60.875 , 8.00751226265442e-47 ], [ 60.9375 , 7.990289119536382e-47 ], 
                    [ 61.0 , 7.97315291658611e-47 ], [ 61.0625 , 7.956105175929195e-47 ], [ 61.125 , 7.939121979833705e-47 ], [ 61.1875 , 7.922247336509095e-47 ], 
                    [ 61.25 , 7.905414907756745e-47 ], [ 61.3125 , 7.88871202887849e-47 ], [ 61.375 , 7.872044440515164e-47 ], [ 61.4375 , 7.8554954642856405e-47 ], 
                    [ 61.5 , 7.838996883177413e-47 ], [ 61.5625 , 7.822593595323146e-47 ], [ 61.625 , 7.806262373511683e-47 ], [ 61.6875 , 7.790002825849766e-47 ], 
                    [ 61.75 , 7.773837141011308e-47 ], [ 61.8125 , 7.757719343997323e-47 ], [ 61.875 , 7.741717132954529e-47 ], [ 61.9375 , 7.725747930422066e-47 ], 
                    [ 62.0 , 7.709898711623922e-47 ], [ 62.0625 , 7.694090918212353e-47 ], [ 62.125 , 7.67837864512034e-47 ], [ 62.1875 , 7.662730829057477e-47 ], 
                    [ 62.25 , 7.647153434262606e-47 ], [ 62.3125 , 7.631663281726495e-47 ], [ 62.375 , 7.616218812728104e-47 ], [ 62.4375 , 7.60088506608246e-47 ], 
                    [ 62.5 , 7.585582190895627e-47 ], [ 62.5625 , 7.570392549998406e-47 ], [ 62.625 , 7.555243435940756e-47 ], [ 62.6875 , 7.540182403886252e-47 ], 
                    [ 62.75 , 7.525186186667656e-47 ], [ 62.8125 , 7.510251874961277e-47 ], [ 62.875 , 7.49540624119849e-47 ], [ 62.9375 , 7.48059671048633e-47 ], 
                    [ 63.0 , 7.465900285127246e-47 ], [ 63.0625 , 7.451232732453951e-47 ], [ 63.125 , 7.436665374328126e-47 ], [ 63.1875 , 7.422145194229079e-47 ], 
                    [ 63.25 , 7.407698449755931e-47 ], [ 63.3125 , 7.393324003297469e-47 ], [ 63.375 , 7.378996166304848e-47 ], [ 63.4375 , 7.364765575867159e-47 ], 
                    [ 63.5 , 7.350562429501863e-47 ], [ 63.5625 , 7.336467013888295e-47 ], [ 63.625 , 7.322406173991459e-47 ], [ 63.6875 , 7.30842539760136e-47 ], 
                    [ 63.75 , 7.294505083815989e-47 ], [ 63.8125 , 7.280637482226981e-47 ], [ 63.875 , 7.26685652029985e-47 ], [ 63.9375 , 7.253101643301542e-47 ], 
                    [ 64.0 , 7.239457726599387e-47 ], [ 64.0625 , 7.225840417768384e-47 ], [ 64.125 , 7.212305392132572e-47 ], [ 64.1875 , 7.198823554935947e-47 ], 
                    [ 64.25 , 7.185396262794734e-47 ], [ 64.3125 , 7.172049341648524e-47 ], [ 64.375 , 7.158728614416195e-47 ], [ 64.4375 , 7.145514442427481e-47 ], 
                    [ 64.5 , 7.13232466224791e-47 ], [ 64.5625 , 7.119215897050005e-47 ], [ 64.625 , 7.106157574243666e-47 ], [ 64.6875 , 7.093151353877983e-47 ], 
                    [ 64.75 , 7.080222923547657e-47 ], [ 64.8125 , 7.067318057399523e-47 ], [ 64.875 , 7.054518217624233e-47 ], [ 64.9375 , 7.041741795025058e-47 ], 
                    [ 65.0 , 7.029040847635545e-47 ], [ 65.0625 , 7.016391362305338e-47 ], [ 65.125 , 7.003788005853676e-47 ], [ 65.1875 , 6.991264181572129e-47 ], 
                    [ 65.25 , 6.978762751768178e-47 ], [ 65.3125 , 6.966357588982898e-47 ], [ 65.375 , 6.95398006743173e-47 ], [ 65.4375 , 6.941669006734212e-47 ], 
                    [ 65.5 , 6.929414191354578e-47 ], [ 65.5625 , 6.917195981743931e-47 ], [ 65.625 , 6.905062560375973e-47 ], [ 65.6875 , 6.8929504221860745e-47 ], 
                    [ 65.75 , 6.880922988335692e-47 ], [ 65.8125 , 6.868931029917325e-47 ], [ 65.875 , 6.856993125172629e-47 ], [ 65.9375 , 6.84511962636094e-47 ], 
                    [ 66.0 , 6.833269993857266e-47 ], [ 66.0625 , 6.821514089303982e-47 ], [ 66.125 , 6.809778409517316e-47 ], [ 66.1875 , 6.798112260750883e-47 ], 
                    [ 66.25 , 6.786492661756757e-47 ], [ 66.3125 , 6.774911710135778e-47 ], [ 66.375 , 6.763406560326653e-47 ], [ 66.4375 , 6.751920948554055e-47 ], 
                    [ 66.5 , 6.74051830778731e-47 ], [ 66.5625 , 6.729146420479203e-47 ], [ 66.625 , 6.717825978400978e-47 ], [ 66.6875 , 6.706566135712617e-47 ], 
                    [ 66.75 , 6.695326733167765e-47 ], [ 66.8125 , 6.684177857601168e-47 ], [ 66.875 , 6.67304754683816e-47 ], [ 66.9375 , 6.661979695953405e-47 ], 
                    [ 67.0 , 6.650959359801147e-47 ], [ 67.0625 , 6.639969834762342e-47 ], [ 67.125 , 6.629057921617799e-47 ], [ 67.1875 , 6.618163940760865e-47 ], 
                    [ 67.25 , 6.60734088806008e-47 ], [ 67.3125 , 6.596553874602138e-47 ], [ 67.375 , 6.585806235005003e-47 ], [ 67.4375 , 6.575125503548181e-47 ], 
                    [ 67.5 , 6.564462093892364e-47 ], [ 67.5625 , 6.553876888559334e-47 ], [ 67.625 , 6.543318427381045e-47 ], [ 67.6875 , 6.532805906106085e-47 ], 
                    [ 67.75 , 6.522351082467637e-47 ], [ 67.8125 , 6.511912990282408e-47 ], [ 67.875 , 6.501558170972208e-47 ], [ 67.9375 , 6.491222542680334e-47 ], 
                    [ 68.0 , 6.480937842344115e-47 ], [ 68.0625 , 6.470703712747464e-47 ], [ 68.125 , 6.460488218358772e-47 ], [ 68.1875 , 6.45035455079951e-47 ], 
                    [ 68.25 , 6.440236778512406e-47 ], [ 68.3125 , 6.430173129932418e-47 ], [ 68.375 , 6.420154531599706e-47 ], [ 68.4375 , 6.410157549206615e-47 ], 
                    [ 68.5 , 6.400237127354788e-47 ], [ 68.5625 , 6.390332058444043e-47 ], [ 68.625 , 6.380482869973425e-47 ], [ 68.6875 , 6.370674992730916e-47 ], 
                    [ 68.75 , 6.360890121785284e-47 ], [ 68.8125 , 6.351178150528823e-47 ], [ 68.875 , 6.341481007760854e-47 ], [ 68.9375 , 6.33183981549331e-47 ], 
                    [ 69.0 , 6.322237686551315e-47 ], [ 69.0625 , 6.312658481315321e-47 ], [ 69.125 , 6.30315034730707e-47 ], [ 69.1875 , 6.293656534462061e-47 ], 
                    [ 69.25 , 6.284217202539656e-47 ], [ 69.3125 , 6.27481617386178e-47 ], [ 69.375 , 6.265436520721532e-47 ], [ 69.4375 , 6.256127487318375e-47 ], 
                    [ 69.5 , 6.246832285050899e-47 ], [ 69.5625 , 6.237588799950148e-47 ], [ 69.625 , 6.2283843797501275e-47 ], [ 69.6875 , 6.219198392056853e-47 ], 
                    [ 69.75 , 6.210083907516739e-47 ], [ 69.8125 , 6.200982780619116e-47 ], [ 69.875 , 6.191929268544798e-47 ], [ 69.9375 , 6.182916980336054e-47 ], 
                    [ 70.0 , 6.173918870460316e-47 ], [ 70.0625 , 6.164994625274991e-47 ], [ 70.125 , 6.156083279862651e-47 ], [ 70.1875 , 6.147214168699001e-47 ], 
                    [ 70.25 , 6.138389776725494e-47 ], [ 70.3125 , 6.129578052261413e-47 ], [ 70.375 , 6.120835747628466e-47 ], [ 70.4375 , 6.112109906182485e-47 ], 
                    [ 70.5 , 6.103419683854442e-47 ], [ 70.5625 , 6.094778941026263e-47 ], [ 70.625 , 6.08615043108399e-47 ], [ 70.6875 , 6.077583649992341e-47 ], 
                    [ 70.75 , 6.06903918794075e-47 ], [ 70.8125 , 6.060522554242231e-47 ], [ 70.875 , 6.052061352858889e-47 ], [ 70.9375 , 6.043611964306569e-47 ], 
                    [ 71.0 , 6.035215536485445e-47 ], [ 71.0625 , 6.026848480056961e-47 ], [ 71.125 , 6.018500355347699e-47 ], [ 71.1875 , 6.010214759054955e-47 ], 
                    [ 71.25 , 6.001940569441928e-47 ], [ 71.3125 , 5.99370938414089e-47 ], [ 71.375 , 5.985515688785477e-47 ], [ 71.4375 , 5.977333194614384e-47 ], 
                    [ 71.5 , 5.96921716825626e-47 ], [ 71.5625 , 5.961114374097838e-47 ], [ 71.625 , 5.953043672864107e-47 ], [ 71.6875 , 5.945019673348942e-47 ], 
                    [ 71.75 , 5.937006489237082e-47 ], [ 71.8125 , 5.929047816147004e-47 ], [ 71.875 , 5.92111256701918e-47 ], [ 71.9375 , 5.9131975289877285e-47 ], 
                    [ 72.0 , 5.9053394241252215e-47 ], [ 72.0625 , 5.89749176197357e-47 ], [ 72.125 , 5.889685796784993e-47 ], [ 72.1875 , 5.881914427930938e-47 ], 
                    [ 72.25 , 5.87415331330388e-47 ], [ 72.3125 , 5.866454715880488e-47 ], [ 72.375 , 5.858769054238456e-47 ], [ 72.4375 , 5.851111381959472e-47 ], 
                    [ 72.5 , 5.843500396299909e-47 ], [ 72.5625 , 5.835899310828433e-47 ], [ 72.625 , 5.828346116798897e-47 ], [ 72.6875 , 5.820818837723198e-47 ], 
                    [ 72.75 , 5.813305012228876e-47 ], [ 72.8125 , 5.805850872886936e-47 ], [ 72.875 , 5.798406291652421e-47 ], [ 72.9375 , 5.79099425214077e-47 ], 
                    [ 73.0 , 5.78362198477026e-47 ], [ 73.0625 , 5.776259102718372e-47 ], [ 73.125 , 5.768947203731769e-47 ], [ 73.1875 , 5.76165580845442e-47 ], 
                    [ 73.25 , 5.754380808353294e-47 ], [ 73.3125 , 5.747160239651098e-47 ], [ 73.375 , 5.739948731282955e-47 ], [ 73.4375 , 5.73277125163489e-47 ], 
                    [ 73.5 , 5.725629710676824e-47 ], [ 73.5625 , 5.71849706622008e-47 ], [ 73.625 , 5.711415526067201e-47 ], [ 73.6875 , 5.7043521130545685e-47 ], 
                    [ 73.75 , 5.69730511814809e-47 ], [ 73.8125 , 5.690310266731135e-47 ], [ 73.875 , 5.683324003224702e-47 ], [ 73.9375 , 5.676370429727237e-47 ], 
                    [ 74.0 , 5.669451894484629e-47 ], [ 74.0625 , 5.662541791765781e-47 ], [ 74.125 , 5.655680166530007e-47 ], [ 74.1875 , 5.648837142296635e-47 ], 
                    [ 74.25 , 5.642007840782627e-47 ], [ 74.3125 , 5.635231217281687e-47 ], [ 74.375 , 5.628462733192743e-47 ], [ 74.4375 , 5.621723002267078e-47 ], 
                    [ 74.5 , 5.615020198596246e-47 ], [ 74.5625 , 5.608325386705893e-47 ], [ 74.625 , 5.601673708963171e-47 ], [ 74.6875 , 5.5950437189330765e-47 ], 
                    [ 74.75 , 5.588422217252234e-47 ], [ 74.8125 , 5.581856671658208e-47 ], [ 74.875 , 5.575298839581112e-47 ], [ 74.9375 , 5.568763285212398e-47 ], 
                    [ 75.0 , 5.562268994434446e-47 ], [ 75.0625 , 5.555782277297273e-47 ], [ 75.125 , 5.549331141904273e-47 ], [ 75.1875 , 5.542907330598009e-47 ], 
                    [ 75.25 , 5.536490955386449e-47 ], [ 75.3125 , 5.530122918866903e-47 ], [ 75.375 , 5.523768745387263e-47 ], [ 75.4375 , 5.517428119497002e-47 ], 
                    [ 75.5 , 5.511135550096535e-47 ], [ 75.5625 , 5.504850157306002e-47 ], [ 75.625 , 5.498590493606321e-47 ], [ 75.6875 , 5.492366087831971e-47 ], 
                    [ 75.75 , 5.486148728086549e-47 ], [ 75.8125 , 5.479968760877199e-47 ], [ 75.875 , 5.473811695549546e-47 ], [ 75.9375 , 5.467661548044803e-47 ], 
                    [ 76.0 , 5.461559902534977e-47 ], [ 76.0625 , 5.455469302264823e-47 ], [ 76.125 , 5.449392486989196e-47 ], [ 76.1875 , 5.443361021877314e-47 ], 
                    [ 76.25 , 5.437336232476765e-47 ], [ 76.3125 , 5.431335839800926e-47 ], [ 76.375 , 5.425369436411505e-47 ], [ 76.4375 , 5.419409587205232e-47 ], 
                    [ 76.5 , 5.413484337491046e-47 ], [ 76.5625 , 5.407582293906947e-47 ], [ 76.625 , 5.401686685017224e-47 ], [ 76.6875 , 5.39583537077678e-47 ], 
                    [ 76.75 , 5.38999690153093e-47 ], [ 76.8125 , 5.384168009283931e-47 ], [ 76.875 , 5.37838616548435e-47 ], [ 76.9375 , 5.372610530576854e-47 ], 
                    [ 77.0 , 5.366853707190815e-47 ], [ 77.0625 , 5.3611341216767e-47 ], [ 77.125 , 5.355420631662859e-47 ], [ 77.1875 , 5.349734724890992e-47 ], 
                    [ 77.25 , 5.344076619725708e-47 ], [ 77.3125 , 5.338424498811064e-47 ], [ 77.375 , 5.332808382724125e-47 ], [ 77.4375 , 5.327211126278324e-47 ], 
                    [ 77.5 , 5.321619744650757e-47 ], [ 77.5625 , 5.316072263613815e-47 ], [ 77.625 , 5.310535074766314e-47 ], [ 77.6875 , 5.305007161412572e-47 ], 
                    [ 77.75 , 5.2995237068179915e-47 ], [ 77.8125 , 5.294045920127936e-47 ], [ 77.875 , 5.288584927063145e-47 ], [ 77.9375 , 5.283160287336502e-47 ], 
                    [ 78.0 , 5.277741211804511e-47 ], [ 78.0625 , 5.272346079285461e-47 ], [ 78.125 , 5.26697963415791e-47 ], [ 78.1875 , 5.261618651253984e-47 ], 
                    [ 78.25 , 5.256288310137635e-47 ], [ 78.3125 , 5.250979312386195e-47 ], [ 78.375 , 5.245675676870584e-47 ], [ 78.4375 , 5.240409111400428e-47 ], 
                    [ 78.5 , 5.235156996331995e-47 ], [ 78.5625 , 5.229910145110698e-47 ], [ 78.625 , 5.224706287525238e-47 ], [ 78.6875 , 5.219510324351418e-47 ], 
                    [ 78.75 , 5.214322884133176e-47 ], [ 78.8125 , 5.209177415163929e-47 ], [ 78.875 , 5.20403702371901e-47 ], [ 78.9375 , 5.198910766995457e-47 ], 
                    [ 79.0 , 5.193820272722397e-47 ], [ 79.0625 , 5.188734762788e-47 ], [ 79.125 , 5.183668672855168e-47 ], [ 79.1875 , 5.178632581562315e-47 ], 
                    [ 79.25 , 5.173601382984781e-47 ], [ 79.3125 , 5.168594536052106e-47 ], [ 79.375 , 5.16361224009234e-47 ], [ 79.4375 , 5.158634746844894e-47 ], 
                    [ 79.5 , 5.153686156444518e-47 ], [ 79.5625 , 5.148756983907759e-47 ], [ 79.625 , 5.143832525810525e-47 ], [ 79.6875 , 5.138941285322276e-47 ], 
                    [ 79.75 , 5.134064720093908e-47 ], [ 79.8125 , 5.129192782450808e-47 ], [ 79.875 , 5.124357985862617e-47 ], [ 79.9375 , 5.119533417858248e-47 ], 
                    [ 80.0 , 5.114713392170567e-47 ], [ 80.0625 , 5.109934057296597e-47 ], [ 80.125 , 5.105160829901517e-47 ], [ 80.1875 , 5.100394134430654e-47 ], 
                    [ 80.25 , 5.095667385245301e-47 ], [ 80.3125 , 5.090945016536722e-47 ], [ 80.375 , 5.086232506222135e-47 ], [ 80.4375 , 5.081556100681123e-47 ], 
                    [ 80.5 , 5.076883994740834e-47 ], [ 80.5625 , 5.07222477127597e-47 ], [ 80.625 , 5.06759813153058e-47 ], [ 80.6875 , 5.062975711983648e-47 ], 
                    [ 80.75 , 5.058368913798387e-47 ], [ 80.8125 , 5.053791503486861e-47 ], [ 80.875 , 5.049218235357433e-47 ], [ 80.9375 , 5.044663040779695e-47 ], 
                    [ 81.0 , 5.040134327680871e-47 ], [ 81.0625 , 5.035609680114632e-47 ], [ 81.125 , 5.031105264592164e-47 ], [ 81.1875 , 5.026624690748452e-47 ], 
                    [ 81.25 , 5.022148107189256e-47 ], [ 81.3125 , 5.017693630091214e-47 ], [ 81.375 , 5.013260631880626e-47 ], [ 81.4375 , 5.008831550105402e-47 ], 
                    [ 81.5 , 5.004426242378717e-47 ], [ 81.5625 , 5.000040379366659e-47 ], [ 81.625 , 4.995658360110757e-47 ], [ 81.6875 , 4.99130143041162e-47 ], 
                    [ 81.75 , 4.986962109595945e-47 ], [ 81.8125 , 4.982626561284485e-47 ], [ 81.875 , 4.978317217211057e-47 ], [ 81.9375 , 4.974023949998978e-47 ], 
                    [ 82.0 , 4.969734385271677e-47 ], [ 82.0625 , 4.965471903800499e-47 ], [ 82.125 , 4.961224228952463e-47 ], [ 82.1875 , 4.956980187745256e-47 ], 
                    [ 82.25 , 4.952763759719693e-47 ], [ 82.3125 , 4.94856107314226e-47 ], [ 82.375 , 4.944361952770515e-47 ], [ 82.4375 , 4.940190890223324e-47 ], 
                    [ 82.5 , 4.936032814294043e-47 ], [ 82.5625 , 4.931878238147697e-47 ], [ 82.625 , 4.927751803029641e-47 ], [ 82.6875 , 4.923637790869744e-47 ], 
                    [ 82.75 , 4.919527213358351e-47 ], [ 82.8125 , 4.915444676347612e-47 ], [ 82.875 , 4.91137424999477e-47 ], [ 82.9375 , 4.907307194317715e-47 ], 
                    [ 83.0 , 4.903267888989906e-47 ], [ 83.0625 , 4.899240624200817e-47 ], [ 83.125 , 4.895216667177498e-47 ], [ 83.1875 , 4.891219883577786e-47 ], 
                    [ 83.25 , 4.887235270223839e-47 ], [ 83.3125 , 4.883253902919883e-47 ], [ 83.375 , 4.879298942914567e-47 ], [ 83.4375 , 4.875356510936019e-47 ], 
                    [ 83.5 , 4.871417264408993e-47 ], [ 83.5625 , 4.867503497264757e-47 ], [ 83.625 , 4.863602849349635e-47 ], [ 83.6875 , 4.859705327278079e-47 ], 
                    [ 83.75 , 4.85583205487251e-47 ], [ 83.8125 , 4.851972657548971e-47 ], [ 83.875 , 4.848116327660226e-47 ], [ 83.9375 , 4.844282923881528e-47 ], 
                    [ 84.0 , 4.840464407735023e-47 ], [ 84.0625 , 4.836648901541721e-47 ], [ 84.125 , 4.832854731532931e-47 ], [ 84.1875 , 4.829076620688907e-47 ], 
                    [ 84.25 , 4.825301463403942e-47 ], [ 84.3125 , 4.821545892706525e-47 ], [ 84.375 , 4.817807766591731e-47 ], [ 84.4375 , 4.814072538631734e-47 ], 
                    [ 84.5 , 4.810354946320087e-47 ], [ 84.5625 , 4.806656358604891e-47 ], [ 84.625 , 4.802960614661426e-47 ], [ 84.6875 , 4.799280410626691e-47 ], 
                    [ 84.75 , 4.795620999774607e-47 ], [ 84.8125 , 4.791964379192445e-47 ], [ 84.875 , 4.7883209726539165e-47 ], [ 84.9375 , 4.784700251120927e-47 ], 
                    [ 85.0 , 4.781082267421097e-47 ], [ 85.0625 , 4.777475033088223e-47 ], [ 85.125 , 4.773892601977407e-47 ], [ 85.1875 , 4.770312857183579e-47 ], 
                    [ 85.25 , 4.76674123250338e-47 ], [ 85.3125 , 4.763196723922111e-47 ], [ 85.375 , 4.759654851007532e-47 ], [ 85.4375 , 4.75611829433114e-47 ], 
                    [ 85.5 , 4.752611289774637e-47 ], [ 85.5625 , 4.749106871167473e-47 ], [ 85.625 , 4.74560503660282e-47 ], [ 85.6875 , 4.742134885329637e-47 ], 
                    [ 85.75 , 4.738667498976901e-47 ], [ 85.8125 , 4.735202647931239e-47 ], [ 85.875 , 4.73176616270891e-47 ], [ 85.9375 , 4.728335462369314e-47 ], 
                    [ 86.0 , 4.724907249410592e-47 ], [ 86.0625 , 4.72150385573274e-47 ], [ 86.125 , 4.718109409236565e-47 ], [ 86.1875 , 4.714717403121117e-47 ], 
                    [ 86.25 , 4.711346595242581e-47 ], [ 86.3125 , 4.707988064679672e-47 ], [ 86.375 , 4.704631928278846e-47 ], [ 86.4375 , 4.701293178215426e-47 ], 
                    [ 86.5 , 4.697970155651957e-47 ], [ 86.5625 , 4.694649481905832e-47 ], [ 86.625 , 4.691342268560196e-47 ], [ 86.6875 , 4.688054387238062e-47 ], 
                    [ 86.75 , 4.684768810195348e-47 ], [ 86.8125 , 4.681492637794772e-47 ], [ 86.875 , 4.678239553099211e-47 ], [ 86.9375 , 4.67498872891346e-47 ], 
                    [ 87.0 , 4.671743111655932e-47 ], [ 87.0625 , 4.668524426920855e-47 ], [ 87.125 , 4.665307959758884e-47 ], [ 87.1875 , 4.66209370864195e-47 ], 
                    [ 87.25 , 4.658907761689296e-47 ], [ 87.3125 , 4.655725314762145e-47 ], [ 87.375 , 4.652545041728343e-47 ], [ 87.4375 , 4.649388406646692e-47 ], 
                    [ 87.5 , 4.646239628093113e-47 ], [ 87.5625 , 4.64309298203657e-47 ], [ 87.625 , 4.639965194461381e-47 ], [ 87.6875 , 4.636849742628752e-47 ], 
                    [ 87.75 , 4.633736382630805e-47 ], [ 87.8125 , 4.630636964577663e-47 ], [ 87.875 , 4.62755445375212e-47 ], [ 87.9375 , 4.62447399488467e-47 ], 
                    [ 88.0 , 4.621402485688349e-47 ], [ 88.0625 , 4.618352605174503e-47 ], [ 88.125 , 4.615304737420123e-47 ], [ 88.1875 , 4.612260707616355e-47 ], 
                    [ 88.25 , 4.609243128320962e-47 ], [ 88.3125 , 4.606227523281878e-47 ], [ 88.375 , 4.603213891207509e-47 ], [ 88.4375 , 4.600224926380855e-47 ], 
                    [ 88.5 , 4.597241249329672e-47 ], [ 88.5625 , 4.594259507472755e-47 ], [ 88.625 , 4.591296887304355e-47 ], [ 88.6875 , 4.588344788639593e-47 ], 
                    [ 88.75 , 4.585394588106609e-47 ], [ 88.8125 , 4.582457894284364e-47 ], [ 88.875 , 4.579537063649663e-47 ], [ 88.9375 , 4.576618094734577e-47 ], 
                    [ 89.0 , 4.573706919935959e-47 ], [ 89.0625 , 4.570817049891932e-47 ], [ 89.125 , 4.567929005795569e-47 ], [ 89.1875 , 4.565042948505215e-47 ], 
                    [ 89.25 , 4.562183662926932e-47 ], [ 89.3125 , 4.559326168243865e-47 ], [ 89.375 , 4.556470463334395e-47 ], [ 89.4375 , 4.5536358341301035e-47 ], 
                    [ 89.5 , 4.550808662062168e-47 ], [ 89.5625 , 4.547983245273321e-47 ], [ 89.625 , 4.545172686591014e-47 ], [ 89.6875 , 4.542375471368798e-47 ], 
                    [ 89.75 , 4.539579977624223e-47 ], [ 89.8125 , 4.536793098703762e-47 ], [ 89.875 , 4.534025562825817e-47 ], [ 89.9375 , 4.5312597152009e-47 ], 
                    [ 90.0 , 4.528496148570742e-47 ], [ 90.0625 , 4.525757946369596e-47 ], [ 90.125 , 4.523021399850833e-47 ], [ 90.1875 , 4.520286508013264e-47 ], 
                    [ 90.25 , 4.5175716094217095e-47 ], [ 90.3125 , 4.514864112135622e-47 ], [ 90.375 , 4.512158237522583e-47 ], [ 90.4375 , 4.509465668173344e-47 ], 
                    [ 90.5 , 4.506786896205771e-47 ], [ 90.5625 , 4.504109715517502e-47 ], [ 90.625 , 4.501439120705495e-47 ], [ 90.6875 , 4.498788833191593e-47 ], 
                    [ 90.75 , 4.49614010607287e-47 ], [ 90.8125 , 4.493492938430654e-47 ], [ 90.875 , 4.490869018723742e-47 ], [ 90.9375 , 4.488248405226397e-47 ], 
                    [ 91.0 , 4.485629320968765e-47 ], [ 91.0625 , 4.483026442722371e-47 ], [ 91.125 , 4.480433662914649e-47 ], [ 91.1875 , 4.477842382653542e-47 ], 
                    [ 91.25 , 4.475260216589794e-47 ], [ 91.3125 , 4.472695014862556e-47 ], [ 91.375 , 4.47013128349899e-47 ], [ 91.4375 , 4.46756950497601e-47 ], 
                    [ 91.5 , 4.465031582937347e-47 ], [ 91.5625 , 4.46249510263294e-47 ], [ 91.625 , 4.459960063243803e-47 ], [ 91.6875 , 4.457442457284279e-47 ], 
                    [ 91.75 , 4.45493295865957e-47 ], [ 91.8125 , 4.452424872859233e-47 ], [ 91.875 , 4.44992678751269e-47 ], [ 91.9375 , 4.44744401956259e-47 ], 
                    [ 92.0 , 4.444962636834539e-47 ], [ 92.0625 , 4.442483761972337e-47 ], [ 92.125 , 4.44002742437791e-47 ], [ 92.1875 , 4.437572444941343e-47 ], 
                    [ 92.25 , 4.43511882291181e-47 ], [ 92.3125 , 4.432682285081079e-47 ], [ 92.375 , 4.4302534519845085e-47 ], [ 92.4375 , 4.427825949736714e-47 ], 
                    [ 92.5 , 4.425407797203641e-47 ], [ 92.5625 , 4.423004878360268e-47 ], [ 92.625 , 4.420603264259666e-47 ], [ 92.6875 , 4.418203209139173e-47 ], 
                    [ 92.75 , 4.41582592872039e-47 ], [ 92.8125 , 4.413449927433009e-47 ], [ 92.875 , 4.411075204588618e-47 ], [ 92.9375 , 4.408715754549205e-47 ], 
                    [ 93.0 , 4.406365087756087e-47 ], [ 93.0625 , 4.404015674306242e-47 ], [ 93.125 , 4.401673528285572e-47 ], [ 93.1875 , 4.39934798727193e-47 ], 
                    [ 93.25 , 4.397023674913928e-47 ], [ 93.3125 , 4.394700590562522e-47 ], [ 93.375 , 4.392397856530413e-47 ], [ 93.4375 , 4.39009834419814e-47 ], 
                    [ 93.5 , 4.387800035708703e-47 ], [ 93.5625 , 4.385513858459762e-47 ], [ 93.625 , 4.383238941639352e-47 ], [ 93.6875 , 4.380965204896512e-47 ], 
                    [ 93.75 , 4.378695321984694e-47 ], [ 93.8125 , 4.376444770199441e-47 ], [ 93.875 , 4.3741953751476676e-47 ], [ 93.9375 , 4.3719471362348375e-47 ], 
                    [ 94.0 , 4.369715066490289e-47 ], [ 94.0625 , 4.3674897243926205e-47 ], [ 94.125 , 4.365265515583395e-47 ], [ 94.1875 , 4.363049010688781e-47 ], 
                    [ 94.25 , 4.360847506223738e-47 ], [ 94.3125 , 4.3586471125924347e-47 ], [ 94.375 , 4.3564478292342724e-47 ], [ 94.4375 , 4.354268029758241e-47 ], 
                    [ 94.5 , 4.352091222454551e-47 ], [ 94.5625 , 4.349915503391243e-47 ], [ 94.625 , 4.3477505947861113e-47 ], [ 94.6875 , 4.34559715200444e-47 ], 
                    [ 94.75 , 4.343444775823864e-47 ], [ 94.8125 , 4.3412945152142367e-47 ], [ 94.875 , 4.339164167338091e-47 ], [ 94.9375 , 4.3370348648603137e-47 ], 
                    [ 95.0 , 4.3349066072678493e-47 ], [ 95.0625 , 4.332791551535718e-47 ], [ 95.125 , 4.3306851562110503e-47 ], [ 95.1875 , 4.328579784914674e-47 ], 
                    [ 95.25 , 4.326478726270124e-47 ], [ 95.3125 , 4.3243949480961015e-47 ], [ 95.375 , 4.322312173539945e-47 ], [ 95.4375 , 4.320230402118309e-47 ], 
                    [ 95.5 , 4.31816351765863e-47 ], [ 95.5625 , 4.3161031756668143e-47 ], [ 95.625 , 4.3140438167338787e-47 ], [ 95.6875 , 4.3119902926436453e-47 ], 
                    [ 95.75 , 4.30995211388044e-47 ], [ 95.8125 , 4.307914898517489e-47 ], [ 95.875 , 4.305878646099444e-47 ], [ 95.9375 , 4.30385830296624e-47 ], 
                    [ 96.0 , 4.301843055023927e-47 ], [ 96.0625 , 4.299828750705557e-47 ], [ 96.125 , 4.297821161118788e-47 ], [ 96.1875 , 4.295827642923293e-47 ], 
                    [ 96.25 , 4.2938350494091146e-47 ], [ 96.3125 , 4.291843380147313e-47 ], [ 96.375 , 4.2898679914211017e-47 ], [ 96.4375 , 4.2878968864549987e-47 ], 
                    [ 96.5 , 4.2859266871706226e-47 ], [ 96.5625 , 4.2839634666489395e-47 ], [ 96.625 , 4.282013669163503e-47 ], [ 96.6875 , 4.2800647591063544e-47 ], 
                    [ 96.75 , 4.2781167360735295e-47 ], [ 96.8125 , 4.2761848219203756e-47 ], [ 96.875 , 4.274257002152417e-47 ], [ 96.9375 , 4.2723300514977485e-47 ], 
                    [ 97.0 , 4.270409742437938e-47 ], [ 97.0625 , 4.268502781088216e-47 ], [ 97.125 , 4.266596671296526e-47 ], [ 97.1875 , 4.2646914126825435e-47 ], 
                    [ 97.25 , 4.2628014795338075e-47 ], [ 97.3125 , 4.260916024174442e-47 ], [ 97.375 , 4.2590314027601214e-47 ], [ 97.4375 , 4.2571525371004614e-47 ], 
                    [ 97.5 , 4.255287543893689e-47 ], [ 97.5625 , 4.253423367711735e-47 ], [ 97.625 , 4.2515600081967926e-47 ], [ 97.6875 , 4.2497106444302185e-47 ], 
                    [ 97.75 , 4.2478666941088555e-47 ], [ 97.8125 , 4.2460235438779146e-47 ], [ 97.875 , 4.2441847152362077e-47 ], [ 97.9375 , 4.2423607882207364e-47 ], 
                    [ 98.0 , 4.2405376450329853e-47 ], [ 98.0625 , 4.2387152853361963e-47 ], [ 98.125 , 4.236905107862305e-47 ], [ 98.1875 , 4.235101821978553e-47 ], 
                    [ 98.25 , 4.233299303598474e-47 ], [ 98.3125 , 4.2314991734586245e-47 ], [ 98.375 , 4.2297155268106403e-47 ], [ 98.4375 , 4.2279326319990875e-47 ], 
                    [ 98.5 , 4.2261504887071448e-47 ], [ 98.5625 , 4.224378182165113e-47 ], [ 98.625 , 4.222614729593613e-47 ], [ 98.6875 , 4.220852013169532e-47 ], 
                    [ 98.75 , 4.2190900325855384e-47 ], [ 98.8125 , 4.217345035838723e-47 ], [ 98.875 , 4.215601527123771e-47 ], [ 98.9375 , 4.2138587391994437e-47 ], 
                    [ 99.0 , 4.212122997773095e-47 ], [ 99.0625 , 4.210398575705015e-47 ], [ 99.125 , 4.2086748596066665e-47 ], [ 99.1875 , 4.206951849189119e-47 ], 
                    [ 99.25 , 4.205242665646981e-47 ], [ 99.3125 , 4.2035377894633114e-47 ], [ 99.375 , 4.201833604465213e-47 ], [ 99.4375 , 4.200133205018894e-47 ], 
                    [ 99.5 , 4.198446998635003e-47 ], [ 99.5625 , 4.1967614692040103e-47 ], [ 99.625 , 4.195076616454023e-47 ], [ 99.6875 , 4.1934020052801244e-47 ], 
                    [ 99.75 , 4.1917349704331335e-47 ], [ 99.8125 , 4.190068598295066e-47 ], [ 99.875 , 4.1884028886025604e-47 ], [ 99.9375 , 4.186753578206096e-47 ], 
                    [ 100.0 , 4.1851054609790607e-47 ], [ 100.0625 , 4.183457992533974e-47 ], [ 100.125 , 4.1818167539197474e-47 ], [ 100.1875 , 4.1801867758536165e-47 ], 
                    [ 100.25 , 4.178557433116316e-47 ], [ 100.3125 , 4.17692872546003e-47 ], [ 100.375 , 4.1753121614806382e-47 ], [ 100.4375 , 4.173700829551695e-47 ], 
                    [ 100.5 , 4.1720901194661583e-47 ], [ 100.5625 , 4.170481220319847e-47 ], [ 100.625 , 4.1688875287389465e-47 ], [ 100.6875 , 4.1672944461652487e-47 ], 
                    [ 100.75 , 4.165701972365941e-47 ], [ 100.8125 , 4.1641169028253194e-47 ], [ 100.875 , 4.16254149140081e-47 ], [ 100.9375 , 4.160966676002171e-47 ], 
                    [ 101.0 , 4.1593924564037896e-47 ], [ 101.0625 , 4.157830889329603e-47 ], [ 101.125 , 4.156273438980796e-47 ], [ 101.1875 , 4.154716572025507e-47 ], 
                    [ 101.25 , 4.1531620020164464e-47 ], [ 101.3125 , 4.151621863166687e-47 ], [ 101.375 , 4.150082295454755e-47 ], [ 101.4375 , 4.1485432986687934e-47 ], 
                    [ 101.5 , 4.147011568325565e-47 ], [ 101.5625 , 4.145488983327568e-47 ], [ 101.625 , 4.143966957350214e-47 ], [ 101.6875 , 4.142445490188257e-47 ], 
                    [ 101.75 , 4.1409361265574607e-47 ], [ 101.8125 , 4.139430992576695e-47 ], [ 101.875 , 4.137926405677039e-47 ], [ 101.9375 , 4.136423424753432e-47 ], 
                    [ 102.0 , 4.1349350727793867e-47 ], [ 102.0625 , 4.1334472563384857e-47 ], [ 102.125 , 4.131959975237888e-47 ], [ 102.1875 , 4.130478809035553e-47 ], 
                    [ 102.25 , 4.1290075449922266e-47 ], [ 102.3125 , 4.127536805008638e-47 ], [ 102.375 , 4.1260665888982073e-47 ], [ 102.4375 , 4.1246067678675174e-47 ], 
                    [ 102.5 , 4.123152367742682e-47 ], [ 102.5625 , 4.1216984804618193e-47 ], [ 102.625 , 4.1202451058440646e-47 ], [ 102.6875 , 4.118806322059557e-47 ], 
                    [ 102.75 , 4.1173687106363495e-47 ], [ 102.8125 , 4.11593160099125e-47 ], [ 102.875 , 4.114498392307496e-47 ], [ 102.9375 , 4.113076891536335e-47 ], 
                    [ 103.0 , 4.111655881873448e-47 ], [ 103.0625 , 4.1102353631493387e-47 ], [ 103.125 , 4.1088225934806625e-47 ], [ 103.1875 , 4.1074175004912767e-47 ], 
                    [ 103.25 , 4.106012888001232e-47 ], [ 103.3125 , 4.1046087558460957e-47 ], [ 103.375 , 4.103215984046054e-47 ], [ 103.4375 , 4.101827056535689e-47 ], 
                    [ 103.5 , 4.1004385991735285e-47 ], [ 103.5625 , 4.099050842680917e-47 ], [ 103.625 , 4.097677607405504e-47 ], [ 103.6875 , 4.0963048321818036e-47 ], 
                    [ 103.75 , 4.094932516855634e-47 ], [ 103.8125 , 4.0935643241284034e-47 ], [ 103.875 , 4.09220699281304e-47 ], [ 103.9375 , 4.0908501115573405e-47 ], 
                    [ 104.0 , 4.0894936802121327e-47 ], [ 104.0625 , 4.0881446194475744e-47 ], [ 104.125 , 4.086803033854921e-47 ], [ 104.1875 , 4.0854618885236063e-47 ], 
                    [ 104.25 , 4.084121183309151e-47 ], [ 104.3125 , 4.0827909340157986e-47 ], [ 104.375 , 4.0814649502970744e-47 ], [ 104.4375 , 4.0801393972232424e-47 ], 
                    [ 104.5 , 4.0788142746543525e-47 ], [ 104.5625 , 4.0775024950978364e-47 ], [ 104.625 , 4.076191935704284e-47 ], [ 104.6875 , 4.0748817975405914e-47 ], 
                    [ 104.75 , 4.0735741480389564e-47 ], [ 104.8125 , 4.072278463665039e-47 ], [ 104.875 , 4.070983191410263e-47 ], [ 104.9375 , 4.0696883311436024e-47 ], 
                    [ 105.0 , 4.068398650869252e-47 ], [ 105.0625 , 4.06711810516604e-47 ], [ 105.125 , 4.06583796252001e-47 ], [ 105.1875 , 4.064558222804241e-47 ], 
                    [ 105.25 , 4.063286162159215e-47 ], [ 105.3125 , 4.0620205460177137e-47 ], [ 105.375 , 4.0607553240851857e-47 ], [ 105.4375 , 4.0594904962390756e-47 ], 
                    [ 105.5 , 4.058235775052647e-47 ], [ 105.5625 , 4.056985025658201e-47 ], [ 105.625 , 4.0557346617450676e-47 ], [ 105.6875 , 4.0544846831944105e-47 ], 
                    [ 105.75 , 4.0532469994479203e-47 ], [ 105.8125 , 4.052010929955385e-47 ], [ 105.875 , 4.0507752374119745e-47 ], [ 105.9375 , 4.0495408667360154e-47 ], 
                    [ 106.0 , 4.0483189047487753e-47 ], [ 106.0625 , 4.0470973114924404e-47 ], [ 106.125 , 4.0458760868558296e-47 ], [ 106.1875 , 4.044658230179866e-47 ], 
                    [ 106.25 , 4.0434507200456877e-47 ], [ 106.3125 , 4.042243570406861e-47 ], [ 106.375 , 4.0410367811558775e-47 ], [ 106.4375 , 4.039835210427007e-47 ], 
                    [ 106.5 , 4.0386419051124146e-47 ], [ 106.5625 , 4.037448952281983e-47 ], [ 106.625 , 4.036256351831392e-47 ], [ 106.6875 , 4.035070727941245e-47 ], 
                    [ 106.75 , 4.0338915339865673e-47 ], [ 106.8125 , 4.0327126846351993e-47 ], [ 106.875 , 4.0315341797862354e-47 ], [ 106.9375 , 4.030364240309836e-47 ], 
                    [ 107.0 , 4.0291990022620454e-47 ], [ 107.0625 , 4.028034101101876e-47 ], [ 107.125 , 4.0268695367318436e-47 ], [ 107.1875 , 4.0257149875478773e-47 ], 
                    [ 107.25 , 4.0245635546735033e-47 ], [ 107.3125 , 4.0234124511314376e-47 ], [ 107.375 , 4.0222616768272554e-47 ], [ 107.4375 , 4.021122288413065e-47 ], 
                    [ 107.5 , 4.019984574437079e-47 ], [ 107.5625 , 4.0188471823595284e-47 ], [ 107.625 , 4.0177101120894527e-47 ], [ 107.6875 , 4.016585557300833e-47 ], 
                    [ 107.75 , 4.0154613583722474e-47 ], [ 107.8125 , 4.014337474094747e-47 ], [ 107.875 , 4.013215040555472e-47 ], [ 107.9375 , 4.0121039410234744e-47 ], 
                    [ 108.0 , 4.0109931491107206e-47 ], [ 108.0625 , 4.0098826647319836e-47 ], [ 108.125 , 4.0087746713049263e-47 ], [ 108.1875 , 4.0076768321199375e-47 ], 
                    [ 108.25 , 4.00657929358822e-47 ], [ 108.3125 , 4.0054820556272376e-47 ], [ 108.375 , 4.004388241249309e-47 ], [ 108.4375 , 4.0033035824756364e-47 ], 
                    [ 108.5 , 4.002219217500842e-47 ], [ 108.5625 , 4.001135146245286e-47 ], [ 108.625 , 4.000055292094071e-47 ], [ 108.6875 , 3.9989836614535173e-47 ], 
                    [ 108.75 , 3.9979123179070805e-47 ], [ 108.8125 , 3.9968412613777623e-47 ], [ 108.875 , 3.9957750988084023e-47 ], [ 108.9375 , 3.994716359531432e-47 ], 
                    [ 109.0 , 3.993657900782977e-47 ], [ 109.0625 , 3.992599722488679e-47 ], [ 109.125 , 3.991546993266727e-47 ], [ 109.1875 , 3.990500999859274e-47 ], 
                    [ 109.25 , 3.989455280556626e-47 ], [ 109.3125 , 3.98840983528698e-47 ], [ 109.375 , 3.9873703181986055e-47 ], [ 109.4375 , 3.9863370028031423e-47 ], 
                    [ 109.5 , 3.985303955188437e-47 ], [ 109.5625 , 3.9842711752849244e-47 ], [ 109.625 , 3.9832446014805217e-47 ], [ 109.6875 , 3.982223719975324e-47 ], 
                    [ 109.75 , 3.981203100115911e-47 ], [ 109.8125 , 3.9801827418351394e-47 ], [ 109.875 , 3.97916889374125e-47 ], [ 109.9375 , 3.978160456610041e-47 ], 
                    [ 110.0 , 3.977152275046104e-47 ], [ 110.0625 , 3.976144348984754e-47 ], [ 110.125 , 3.9751430488200353e-47 ], [ 110.1875 , 3.9741469146551197e-47 ], 
                    [ 110.25 , 3.9731510301121495e-47 ], [ 110.3125 , 3.9721553951287113e-47 ], [ 110.375 , 3.9711663669764827e-47 ], [ 110.4375 , 3.970182343900392e-47 ], 
                    [ 110.5 , 3.96919856465733e-47 ], [ 110.5625 , 3.968215029186794e-47 ], [ 110.625 , 3.9672380455101935e-47 ], [ 110.6875 , 3.966266055276056e-47 ], 
                    [ 110.75 , 3.965294303183586e-47 ], [ 110.8125 , 3.9643227891746055e-47 ], [ 110.875 , 3.9633576634819264e-47 ], [ 110.9375 , 3.9623976102318095e-47 ], 
                    [ 111.0 , 3.961437789537544e-47 ], [ 111.0625 , 3.960478201342884e-47 ], [ 111.125 , 3.9595246972197078e-47 ], [ 111.1875 , 3.9585764093952207e-47 ], 
                    [ 111.25 , 3.9576283486812466e-47 ], [ 111.3125 , 3.956680515023422e-47 ], [ 111.375 , 3.955738440172028e-47 ], [ 111.4375 , 3.954801881170525e-47 ], 
                    [ 111.5 , 3.9538655439084007e-47 ], [ 111.5625 , 3.9529294283330163e-47 ], [ 111.625 , 3.951998578981528e-47 ], [ 111.6875 , 3.9510735689199767e-47 ], 
                    [ 111.75 , 3.950148775367543e-47 ], [ 111.8125 , 3.9492241982734936e-47 ], [ 111.875 , 3.94830435765627e-47 ], [ 111.9375 , 3.9473908262764264e-47 ], 
                    [ 112.0 , 3.94647750626316e-47 ], [ 112.0625 , 3.9455643975676244e-47 ], [ 112.125 , 3.944655365071882e-47 ], [ 112.1875 , 3.9437531371627565e-47 ], 
                    [ 112.25 , 3.942851115612647e-47 ], [ 112.3125 , 3.9419493003743545e-47 ], [ 112.375 , 3.9410508676277025e-47 ], [ 112.4375 , 3.9401599098886244e-47 ], 
                    [ 112.5 , 3.939269153569406e-47 ], [ 112.5625 , 3.938378598624344e-47 ], [ 112.625 , 3.9374906194077807e-47 ], [ 112.6875 , 3.93661082010883e-47 ], 
                    [ 112.75 , 3.9357312173936915e-47 ], [ 112.8125 , 3.9348518112183563e-47 ], [ 112.875 , 3.933974087452511e-47 ], [ 112.9375 , 3.9331052995522713e-47 ], 
                    [ 113.0 , 3.932236703517153e-47 ], [ 113.0625 , 3.9313682993047283e-47 ], [ 113.125 , 3.9305006131674776e-47 ], [ 113.1875 , 3.9296427111888424e-47 ], 
                    [ 113.25 , 3.928784996462641e-47 ], [ 113.3125 , 3.92792746894803e-47 ], [ 113.375 , 3.9270701286040915e-47 ], [ 113.4375 , 3.9262225860692964e-47 ], 
                    [ 113.5 , 3.9253757364287734e-47 ], [ 113.5625 , 3.924529069445775e-47 ], [ 113.625 , 3.923682585081125e-47 ], [ 113.6875 , 3.9228446220104047e-47 ], 
                    [ 113.75 , 3.9220084524263963e-47 ], [ 113.8125 , 3.9211724610752687e-47 ], [ 113.875 , 3.920336647918921e-47 ], [ 113.9375 , 3.9195081232357474e-47 ], 
                    [ 114.0 , 3.9186825844870206e-47 ], [ 114.0625 , 3.9178572196157824e-47 ], [ 114.125 , 3.9170320285854086e-47 ], [ 114.1875 , 3.91621275772912e-47 ], 
                    [ 114.25 , 3.9153977157679484e-47 ], [ 114.3125 , 3.914582843433205e-47 ], [ 114.375 , 3.913768140689753e-47 ], [ 114.4375 , 3.912957945295956e-47 ], 
                    [ 114.5 , 3.912153299484875e-47 ], [ 114.5625 , 3.9113488191381313e-47 ], [ 114.625 , 3.910544504221642e-47 ], [ 114.6875 , 3.9097432323786264e-47 ], 
                    [ 114.75 , 3.908948930888553e-47 ], [ 114.8125 , 3.90815479076832e-47 ], [ 114.875 , 3.907360811985282e-47 ], [ 114.9375 , 3.9065683130722484e-47 ], 
                    [ 115.0 , 3.905784181883623e-47 ], [ 115.0625 , 3.905000208086773e-47 ], [ 115.125 , 3.904216391650162e-47 ], [ 115.1875 , 3.903432732542037e-47 ], 
                    [ 115.25 , 3.9026584560837507e-47 ], [ 115.3125 , 3.90188461455363e-47 ], [ 115.375 , 3.9011109264652794e-47 ], [ 115.4375 , 3.900337391788163e-47 ], 
                    [ 115.5 , 3.8995714692376095e-47 ], [ 115.5625 , 3.8988076441748094e-47 ], [ 115.625 , 3.8980439687255585e-47 ], [ 115.6875 , 3.897280442860524e-47 ], 
                    [ 115.75 , 3.8965227196294087e-47 ], [ 115.8125 , 3.89576881967296e-47 ], [ 115.875 , 3.8950150655811903e-47 ], [ 115.9375 , 3.8942614573259057e-47 ], 
                    [ 116.0 , 3.893511788162079e-47 ], [ 116.0625 , 3.892767729930099e-47 ], [ 116.125 , 3.8920238138891737e-47 ], [ 116.1875 , 3.8912800400121576e-47 ], 
                    [ 116.25 , 3.8905382770648857e-47 ], [ 116.3125 , 3.8898039676907266e-47 ], [ 116.375 , 3.8890697969117965e-47 ], [ 116.4375 , 3.888335764702047e-47 ], 
                    [ 116.5 , 3.8876018710352684e-47 ], [ 116.5625 , 3.8868770645421407e-47 ], [ 116.625 , 3.886152501850337e-47 ], [ 116.6875 , 3.8854280742261443e-47 ], 
                    [ 116.75 , 3.8847037816443303e-47 ], [ 116.8125 , 3.883986528285892e-47 ], [ 116.875 , 3.8832715510469044e-47 ], [ 116.9375 , 3.8825567054232653e-47 ], 
                    [ 117.0 , 3.8818419913908277e-47 ], [ 117.0625 , 3.881132118413579e-47 ], [ 117.125 , 3.880426579081433e-47 ], [ 117.1875 , 3.879721168007159e-47 ], 
                    [ 117.25 , 3.87901588516747e-47 ], [ 117.3125 , 3.8783132543904436e-47 ], [ 117.375 , 3.8776171076864597e-47 ], [ 117.4375 , 3.8769210859388364e-47 ], 
                    [ 117.5 , 3.8762251891253386e-47 ], [ 117.5625 , 3.8755296849870417e-47 ], [ 117.625 , 3.8748428443700254e-47 ], [ 117.6875 , 3.8741561254783967e-47 ], 
                    [ 117.75 , 3.873469528290501e-47 ], [ 117.8125 , 3.872783052784658e-47 ], [ 117.875 , 3.872103377795825e-47 ], [ 117.9375 , 3.871425852008104e-47 ], 
                    [ 118.0 , 3.8707484447713005e-47 ], [ 118.0625 , 3.8700711560646143e-47 ], [ 118.125 , 3.869398266160402e-47 ], [ 118.1875 , 3.868729880973038e-47 ], 
                    [ 118.25 , 3.868061611239915e-47 ], [ 118.3125 , 3.867393456941227e-47 ], [ 118.375 , 3.866727226896194e-47 ], [ 118.4375 , 3.866067879199718e-47 ], 
                    [ 118.5 , 3.8654086439340817e-47 ], [ 118.5625 , 3.864749521080139e-47 ], [ 118.625 , 3.864090510618668e-47 ], [ 118.6875 , 3.863439397144853e-47 ], 
                    [ 118.75 , 3.862789086411295e-47 ], [ 118.8125 , 3.8621388851409184e-47 ], [ 118.875 , 3.861488793315133e-47 ], [ 118.9375 , 3.8608440085121806e-47 ], 
                    [ 119.0 , 3.860202570897e-47 ], [ 119.0625 , 3.859561239849793e-47 ], [ 119.125 , 3.8589200153527706e-47 ], [ 119.1875 , 3.858281439059039e-47 ], 
                    [ 119.25 , 3.8576487865689424e-47 ], [ 119.3125 , 3.8570162378165563e-47 ], [ 119.375 , 3.8563837927847595e-47 ], [ 119.4375 , 3.8557514514567644e-47 ], 
                    [ 119.5 , 3.8551273759279033e-47 ], [ 119.5625 , 3.8545035489656974e-47 ], [ 119.625 , 3.85387982294951e-47 ], [ 119.6875 , 3.853256197863254e-47 ], 
                    [ 119.75 , 3.8526380174026e-47 ], [ 119.8125 , 3.8520228059664716e-47 ], [ 119.875 , 3.8514076927708503e-47 ], [ 119.9375 , 3.850792677800048e-47 ], 
                    [ 120.0 , 3.850180280727218e-47 ], [ 120.0625 , 3.8495736116336286e-47 ], [ 120.125 , 3.84896703813226e-47 ], [ 120.1875 , 3.848360560208105e-47 ], 
                    [ 120.25 , 3.847754177846103e-47 ], [ 120.3125 , 3.847155647177485e-47 ], [ 120.375 , 3.8465575502755983e-47 ], [ 120.4375 , 3.8459595463566937e-47 ], 
                    [ 120.5 , 3.8453616354062627e-47 ], [ 120.5625 , 3.844768604753019e-47 ], [ 120.625 , 3.844178899405824e-47 ], [ 120.6875 , 3.843589284506801e-47 ], 
                    [ 120.75 , 3.8429997600421316e-47 ], [ 120.8125 , 3.842412116931337e-47 ], [ 120.875 , 3.8418307047206506e-47 ], [ 120.9375 , 3.8412493804860207e-47 ], 
                    [ 121.0 , 3.840668144214134e-47 ], [ 121.0625 , 3.8400869958916e-47 ], [ 121.125 , 3.839512631667009e-47 ], [ 121.1875 , 3.838939581833075e-47 ], 
                    [ 121.25 , 3.838366617527212e-47 ], [ 121.3125 , 3.837793738736684e-47 ], [ 121.375 , 3.8372244879036952e-47 ], [ 121.4375 , 3.8366595798136545e-47 ], 
                    [ 121.5 , 3.8360947548881626e-47 ], [ 121.5625 , 3.8355300131150043e-47 ], [ 121.625 , 3.834965768782672e-47 ], [ 121.6875 , 3.8344089679943564e-47 ], 
                    [ 121.75 , 3.833852248048334e-47 ], [ 121.8125 , 3.833295608932733e-47 ], [ 121.875 , 3.832739050635871e-47 ], [ 121.9375 , 3.832187506445201e-47 ], 
                    [ 122.0 , 3.831638785501956e-47 ], [ 122.0625 , 3.8310901431286872e-47 ], [ 122.125 , 3.8305415793139814e-47 ], [ 122.1875 , 3.8299947929374215e-47 ], 
                    [ 122.25 , 3.8294540458182086e-47 ], [ 122.3125 , 3.82891337504569e-47 ], [ 122.375 , 3.8283727806090895e-47 ], [ 122.4375 , 3.827832262497627e-47 ], 
                    [ 122.5 , 3.8272977866058424e-47 ], [ 122.5625 , 3.8267649625944643e-47 ], [ 122.625 , 3.82623221276112e-47 ], [ 122.6875 , 3.8256995370954835e-47 ], 
                    [ 122.75 , 3.825169567761585e-47 ], [ 122.8125 , 3.824644557776813e-47 ], [ 122.875 , 3.824119619850356e-47 ], [ 122.9375 , 3.8235947539724064e-47 ], 
                    [ 123.0 , 3.8230699601330748e-47 ], [ 123.0625 , 3.822551919628107e-47 ], [ 123.125 , 3.8220346981908743e-47 ], [ 123.1875 , 3.821517546737784e-47 ], 
                    [ 123.25 , 3.8210004652593654e-47 ], [ 123.3125 , 3.8204866925091994e-47 ], [ 123.375 , 3.8199771399104185e-47 ], [ 123.4375 , 3.8194676552725735e-47 ], 
                    [ 123.5 , 3.8189582385866e-47 ], [ 123.5625 , 3.81844888984349e-47 ], [ 123.625 , 3.817946719941223e-47 ], [ 123.6875 , 3.8174448571959877e-47 ], 
                    [ 123.75 , 3.81694306041983e-47 ], [ 123.8125 , 3.8164413296039688e-47 ], [ 123.875 , 3.815943180665958e-47 ], [ 123.9375 , 3.8154488159486506e-47 ], 
                    [ 124.0 , 3.8149545152775207e-47 ], [ 124.0625 , 3.814460278644218e-47 ], [ 124.125 , 3.8139661060405267e-47 ], [ 124.1875 , 3.813479180001847e-47 ], 
                    [ 124.25 , 3.8129923554104393e-47 ], [ 124.3125 , 3.812505592966501e-47 ], [ 124.375 , 3.8120188926621517e-47 ], [ 124.4375 , 3.8115357756723104e-47 ], 
                    [ 124.5 , 3.811056341210282e-47 ], [ 124.5625 , 3.8105769670539994e-47 ], [ 124.625 , 3.8100976531957956e-47 ], [ 124.6875 , 3.8096183996282475e-47 ], 
                    [ 124.75 , 3.8091461523989457e-47 ], [ 124.8125 , 3.8086740973590104e-47 ], [ 124.875 , 3.8082021008193144e-47 ], [ 124.9375 , 3.807730162772634e-47 ], 
                    [ 125.0 , 3.807261519259331e-47 ], [ 125.0625 , 3.806796744054163e-47 ], [ 125.125 , 3.8063320255868855e-47 ], [ 125.1875 , 3.8058673638505734e-47 ], 
                    [ 125.25 , 3.805402758838329e-47 ], [ 125.3125 , 3.8049446472994036e-47 ], [ 125.375 , 3.8044870986342204e-47 ], [ 125.4375 , 3.804029604989699e-47 ], 
                    [ 125.5 , 3.8035721663593327e-47 ], [ 125.5625 , 3.8031174493313375e-47 ], [ 125.625 , 3.8026670551514407e-47 ], [ 125.6875 , 3.8022167143106316e-47 ], 
                    [ 125.75 , 3.8017664268025934e-47 ], [ 125.8125 , 3.801316192621091e-47 ], [ 125.875 , 3.800871724910466e-47 ], [ 125.9375 , 3.8004284652940076e-47 ], 
                    [ 126.0 , 3.7999852573706866e-47 ], [ 126.0625 , 3.799542101134502e-47 ], [ 126.125 , 3.799100817548959e-47 ], [ 126.1875 , 3.798664546187955e-47 ], 
                    [ 126.25 , 3.798228324926332e-47 ], [ 126.3125 , 3.797792153758417e-47 ], [ 126.375 , 3.7973560326783765e-47 ], [ 126.4375 , 3.796924668330158e-47 ], 
                    [ 126.5 , 3.7964954178688114e-47 ], [ 126.5625 , 3.7960662159351226e-47 ], [ 126.625 , 3.79563706252363e-47 ], [ 126.6875 , 3.7952087021885293e-47 ], 
                    [ 126.75 , 3.794786340650902e-47 ], [ 126.8125 , 3.7943640261170297e-47 ], [ 126.875 , 3.793941758581791e-47 ], [ 126.9375 , 3.7935195380398467e-47 ], 
                    [ 127.0 , 3.7931007952177034e-47 ], [ 127.0625 , 3.7926853038933853e-47 ], [ 127.125 , 3.7922698580813656e-47 ], [ 127.1875 , 3.791854457776766e-47 ], 
                    [ 127.25 , 3.7914391029746024e-47 ], [ 127.3125 , 3.7910298271391956e-47 ], [ 127.375 , 3.790621175278465e-47 ], [ 127.4375 , 3.790212567468016e-47 ], 
                    [ 127.5 , 3.789804003703317e-47 ], [ 127.5625 , 3.7893974163181985e-47 ], [ 127.625 , 3.7889954432917673e-47 ], [ 127.6875 , 3.788593512906026e-47 ], 
                    [ 127.75 , 3.788191625156397e-47 ], [ 127.8125 , 3.787789780038304e-47 ], [ 127.875 , 3.7873923400025785e-47 ], [ 127.9375 , 3.786997073833436e-47 ], 
                    [ 128.0 , 3.786601848915722e-47 ], [ 128.0625 , 3.7862066652451046e-47 ], [ 128.125 , 3.7858117521950214e-47 ], [ 128.1875 , 3.7854230790627685e-47 ], 
                    [ 128.25 , 3.785034445833926e-47 ], [ 128.3125 , 3.784645852504426e-47 ], [ 128.375 , 3.784257299070144e-47 ], [ 128.4375 , 3.7838712716335383e-47 ], 
                    [ 128.5 , 3.7834891861222676e-47 ], [ 128.5625 , 3.783107139193027e-47 ], [ 128.625 , 3.782725130841869e-47 ], [ 128.6875 , 3.7823431610649507e-47 ], 
                    [ 128.75 , 3.7819658777807037e-47 ], [ 128.8125 , 3.781590333660035e-47 ], [ 128.875 , 3.7812148268303924e-47 ], [ 128.9375 , 3.780839357288045e-47 ], 
                    [ 129.0 , 3.780464325990473e-47 ], [ 129.0625 , 3.7800951642066995e-47 ], [ 129.125 , 3.7797260384715196e-47 ], [ 129.1875 , 3.7793569487814666e-47 ], 
                    [ 129.25 , 3.778987895132861e-47 ], [ 129.3125 , 3.778621299578995e-47 ], [ 129.375 , 3.7782585880205227e-47 ], [ 129.4375 , 3.77789591127887e-47 ], 
                    [ 129.5 , 3.7775332693506933e-47 ], [ 129.5625 , 3.7771706622327593e-47 ], [ 129.625 , 3.776812393404128e-47 ], [ 129.6875 , 3.77645602671361e-47 ], 
                    [ 129.75 , 3.776099693648568e-47 ], [ 129.8125 , 3.7757433942058843e-47 ], [ 129.875 , 3.775387128382385e-47 ], [ 129.9375 , 3.7750370003049126e-47 ], 
                    [ 130.0 , 3.7746869174201414e-47 ], [ 130.0625 , 3.7743368670007543e-47 ], [ 130.125 , 3.7739868490437406e-47 ], [ 130.1875 , 3.7736386176148103e-47 ], 
                    [ 130.25 , 3.77329475775413e-47 ], [ 130.3125 , 3.7729509292264713e-47 ], [ 130.375 , 3.77260713202898e-47 ], [ 130.4375 , 3.7722633661588547e-47 ], 
                    [ 130.5 , 3.771923050692531e-47 ], [ 130.5625 , 3.7715853763119257e-47 ], [ 130.625 , 3.77124773216099e-47 ], [ 130.6875 , 3.770910118236964e-47 ], 
                    [ 130.75 , 3.770572534537302e-47 ], [ 130.8125 , 3.7702399781659136e-47 ], [ 130.875 , 3.769908434767051e-47 ], [ 130.9375 , 3.769576920523022e-47 ], 
                    [ 131.0 , 3.769245435431371e-47 ], [ 131.0625 , 3.7689145356163086e-47 ], [ 131.125 , 3.768589059953321e-47 ], [ 131.1875 , 3.768263612397739e-47 ], 
                    [ 131.25 , 3.767938192947161e-47 ], [ 131.3125 , 3.7676128015991336e-47 ], [ 131.375 , 3.767289441648203e-47 ], [ 131.4375 , 3.766969989566602e-47 ], 
                    [ 131.5 , 3.766650564573345e-47 ], [ 131.5625 , 3.766331166666108e-47 ], [ 131.625 , 3.766011795842621e-47 ], [ 131.6875 , 3.765695827044428e-47 ], 
                    [ 131.75 , 3.7653823607573095e-47 ], [ 131.8125 , 3.7650689205639106e-47 ], [ 131.875 , 3.7647555064620867e-47 ], [ 131.9375 , 3.764442118449665e-47 ], 
                    [ 132.0 , 3.7641334267918506e-47 ], [ 132.0625 , 3.76382591162035e-47 ], [ 132.125 , 3.7635184215716784e-47 ], [ 132.1875 , 3.763210956643757e-47 ], 
                    [ 132.25 , 3.762903614010191e-47 ], [ 132.3125 , 3.762601928545386e-47 ], [ 132.375 , 3.762300267267813e-47 ], [ 132.4375 , 3.761998630175532e-47 ], 
                    [ 132.5 , 3.761697017266524e-47 ], [ 132.5625 , 3.7613966934979396e-47 ], [ 132.625 , 3.761100857375302e-47 ], [ 132.6875 , 3.7608050445203537e-47 ], 
                    [ 132.75 , 3.760509254931239e-47 ], [ 132.8125 , 3.7602134886061804e-47 ], [ 132.875 , 3.7599200872937376e-47 ], [ 132.9375 , 3.759630028740167e-47 ], 
                    [ 133.0 , 3.7593399925631805e-47 ], [ 133.0625 , 3.7590499787609444e-47 ], [ 133.125 , 3.7587599873317857e-47 ], [ 133.1875 , 3.7584733601978946e-47 ], 
                    [ 133.25 , 3.758189023092841e-47 ], [ 133.3125 , 3.7579047074985664e-47 ], [ 133.375 , 3.757620413413416e-47 ], [ 133.4375 , 3.757336140835737e-47 ], 
                    [ 133.5 , 3.757056169623216e-47 ], [ 133.5625 , 3.7567775201891666e-47 ], [ 133.625 , 3.7564988914216717e-47 ], [ 133.6875 , 3.756220283319279e-47 ], 
                    [ 133.75 , 3.7559416958803758e-47 ], [ 133.8125 , 3.755668267085947e-47 ], [ 133.875 , 3.7553952625378354e-47 ], [ 133.9375 , 3.755122277834709e-47 ], 
                    [ 134.0 , 3.7548493129752594e-47 ], [ 134.0625 , 3.7545767868028613e-47 ], [ 134.125 , 3.7543093492106375e-47 ], [ 134.1875 , 3.754041930667928e-47 ], 
                    [ 134.25 , 3.753774531173376e-47 ], [ 134.3125 , 3.7535071507256256e-47 ], [ 134.375 , 3.7532409616834063e-47 ], [ 134.4375 , 3.752979082877065e-47 ], 
                    [ 134.5 , 3.7527172223430403e-47 ], [ 134.5625 , 3.7524553800801114e-47 ], [ 134.625 , 3.7521935560869494e-47 ], [ 134.6875 , 3.7519335944134477e-47 ], 
                    [ 134.75 , 3.751677202204961e-47 ], [ 134.8125 , 3.7514208275172976e-47 ], [ 134.875 , 3.7511644703492594e-47 ], [ 134.9375 , 3.75090813069965e-47 ], 
                    [ 135.0 , 3.7506542709937127e-47 ], [ 135.0625 , 3.7504033432635613e-47 ], [ 135.125 , 3.7501524323210476e-47 ], [ 135.1875 , 3.7499015381651284e-47 ], 
                    [ 135.25 , 3.749650660794654e-47 ], [ 135.3125 , 3.749402793950668e-47 ], [ 135.375 , 3.749157268212821e-47 ], [ 135.4375 , 3.7489117585529986e-47 ], 
                    [ 135.5 , 3.7486662649701204e-47 ], [ 135.5625 , 3.748420787463107e-47 ], [ 135.625 , 3.748178799015212e-47 ], [ 135.6875 , 3.747938639536422e-47 ], 
                    [ 135.75 , 3.7476984954455526e-47 ], [ 135.8125 , 3.747458366741563e-47 ], [ 135.875 , 3.747218253423522e-47 ], [ 135.9375 , 3.74698203906626e-47 ], 
                    [ 136.0 , 3.7467472051486845e-47 ], [ 136.0625 , 3.7465123859488644e-47 ], [ 136.125 , 3.7462775814658255e-47 ], [ 136.1875 , 3.746042791698671e-47 ], 
                    [ 136.25 , 3.745812233533814e-47 ], [ 136.3125 , 3.745582668894936e-47 ], [ 136.375 , 3.745353118325162e-47 ], [ 136.4375 , 3.745123581823443e-47 ], 
                    [ 136.5 , 3.744894059389131e-47 ], [ 136.5625 , 3.744669062974158e-47 ], [ 136.625 , 3.744444742342928e-47 ], [ 136.6875 , 3.7442204351493452e-47 ], 
                    [ 136.75 , 3.74399614139274e-47 ], [ 136.8125 , 3.743771861072119e-47 ], [ 136.875 , 3.743552301380358e-47 ], [ 136.9375 , 3.7433331558669877e-47 ], 
                    [ 137.0 , 3.7431140231823027e-47 ], [ 137.0625 , 3.7428949033254993e-47 ], [ 137.125 , 3.7426757962958535e-47 ], [ 137.1875 , 3.742461591727931e-47 ], 
                    [ 137.25 , 3.74224760569063e-47 ], [ 137.3125 , 3.7420336318885983e-47 ], [ 137.375 , 3.741819670321215e-47 ], [ 137.4375 , 3.7416057209876487e-47 ], 
                    [ 137.5 , 3.741396761324266e-47 ], [ 137.5625 , 3.741187881953452e-47 ], [ 137.625 , 3.7409790142442165e-47 ], [ 137.6875 , 3.740770158195881e-47 ], 
                    [ 137.75 , 3.740561324227202e-47 ], [ 137.8125 , 3.7403575004985874e-47 ], [ 137.875 , 3.7401536878764103e-47 ], [ 137.9375 , 3.739949886359986e-47 ], 
                    [ 138.0 , 3.7397460959486827e-47 ], [ 138.0625 , 3.739542348491201e-47 ], [ 138.125 , 3.7393435564527197e-47 ], [ 138.1875 , 3.7391447749819686e-47 ], 
                    [ 138.25 , 3.7389460040782795e-47 ], [ 138.3125 , 3.7387472437411435e-47 ], [ 138.375 , 3.738548493969999e-47 ], [ 138.4375 , 3.7383546889698836e-47 ], 
                    [ 138.5 , 3.738160897176273e-47 ], [ 138.5625 , 3.737967115428595e-47 ], [ 138.625 , 3.737773343726329e-47 ], [ 138.6875 , 3.7375795820689275e-47 ], 
                    [ 138.75 , 3.737390633784137e-47 ], [ 138.8125 , 3.737201787478662e-47 ], [ 138.875 , 3.73701295071544e-47 ], [ 138.9375 , 3.7368241234939087e-47 ], 
                    [ 139.0 , 3.7366353058136124e-47 ], [ 139.0625 , 3.736451126316567e-47 ], [ 139.125 , 3.7362671925339795e-47 ], [ 139.1875 , 3.7360832678058764e-47 ], 
                    [ 139.25 , 3.735899352131784e-47 ], [ 139.3125 , 3.735715445511338e-47 ], [ 139.375 , 3.735535930435026e-47 ], [ 139.4375 , 3.735356856128015e-47 ], 
                    [ 139.5 , 3.7351777904054244e-47 ], [ 139.5625 , 3.7349987332668954e-47 ], [ 139.625 , 3.734819684712095e-47 ], [ 139.6875 , 3.734644755836647e-47 ], 
                    [ 139.75 , 3.734470517748917e-47 ], [ 139.8125 , 3.7342962877901824e-47 ], [ 139.875 , 3.7341220659600917e-47 ], [ 139.9375 , 3.733947852258211e-47 ], 
                    [ 140.0 , 3.7337774129004937e-47 ], [ 140.0625 , 3.7336079607985232e-47 ], [ 140.125 , 3.733438516386866e-47 ], [ 140.1875 , 3.7332690796651995e-47 ], 
                    [ 140.25 , 3.7330996506332005e-47 ], [ 140.3125 , 3.732933632614186e-47 ], [ 140.375 , 3.7327689570174554e-47 ], [ 140.4375 , 3.73260428868524e-47 ], 
                    [ 140.5 , 3.7324396276172453e-47 ], [ 140.5625 , 3.7322749738131514e-47 ], [ 140.625 , 3.7321132730268766e-47 ], [ 140.6875 , 3.731953300329581e-47 ], 
                    [ 140.75 , 3.731793334489302e-47 ], [ 140.8125 , 3.7316333755057467e-47 ], [ 140.875 , 3.7314734233787004e-47 ], [ 140.9375 , 3.7313159529981133e-47 ], 
                    [ 141.0 , 3.7311606607419027e-47 ], [ 141.0625 , 3.7310053749487704e-47 ], [ 141.125 , 3.7308500956184456e-47 ], [ 141.1875 , 3.730694822750607e-47 ], 
                    [ 141.25 , 3.7305415111653525e-47 ], [ 141.3125 , 3.730390877088646e-47 ], [ 141.375 , 3.730240249094334e-47 ], [ 141.4375 , 3.7300896271822235e-47 ], 
                    [ 141.5 , 3.7299390113519367e-47 ], [ 141.5625 , 3.729789779649681e-47 ], [ 141.625 , 3.7296437545810017e-47 ], [ 141.6875 , 3.729497735229405e-47 ], 
                    [ 141.75 , 3.729351721594589e-47 ], [ 141.8125 , 3.7292057136763553e-47 ], [ 141.875 , 3.7290604718500003e-47 ], [ 141.9375 , 3.7289190098421314e-47 ], 
                    [ 142.0 , 3.7287775532006273e-47 ], [ 142.0625 , 3.7286361019252584e-47 ], [ 142.125 , 3.728494656015874e-47 ], [ 142.1875 , 3.7283533181877724e-47 ], 
                    [ 142.25 , 3.728216403765937e-47 ], [ 142.3125 , 3.728079494371967e-47 ], [ 142.375 , 3.727942590005599e-47 ], [ 142.4375 , 3.727805690666726e-47 ], 
                    [ 142.5 , 3.727668796355138e-47 ], [ 142.5625 , 3.727535776928021e-47 ], [ 142.625 , 3.7274033597886377e-47 ], [ 142.6875 , 3.7272709473532474e-47 ], 
                    [ 142.75 , 3.727138539621682e-47 ], [ 142.8125 , 3.727006136593775e-47 ], [ 142.875 , 3.7268768560842624e-47 ], [ 142.9375 , 3.726748929493776e-47 ], 
                    [ 143.0 , 3.7266210072944495e-47 ], [ 143.0625 , 3.726493089486106e-47 ], [ 143.125 , 3.7263651760685934e-47 ], [ 143.1875 , 3.7262395514579907e-47 ], 
                    [ 143.25 , 3.7261160573470233e-47 ], [ 143.3125 , 3.725992567328841e-47 ], [ 143.375 , 3.7258690814033874e-47 ], [ 143.4375 , 3.7257455995703946e-47 ], 
                    [ 143.5 , 3.725623548620494e-47 ], [ 143.5625 , 3.725504448221934e-47 ], [ 143.625 , 3.7253853516307916e-47 ], [ 143.6875 , 3.7252662588468923e-47 ], 
                    [ 143.75 , 3.725147169870166e-47 ], [ 143.8125 , 3.7250286202783176e-47 ], [ 143.875 , 3.724913897305537e-47 ], [ 143.9375 , 3.724799177866006e-47 ], 
                    [ 144.0 , 3.724684461959591e-47 ], [ 144.0625 , 3.724569749586155e-47 ], [ 144.125 , 3.724455040745644e-47 ], [ 144.1875 , 3.724344248103607e-47 ], 
                    [ 144.25 , 3.724233857462393e-47 ], [ 144.3125 , 3.72412347009319e-47 ], [ 144.375 , 3.7240130859959e-47 ], [ 144.4375 , 3.723902705170427e-47 ], 
                    [ 144.5 , 3.7237952549569196e-47 ], [ 144.5625 , 3.7236891823401207e-47 ], [ 144.625 , 3.723583112744862e-47 ], [ 144.6875 , 3.723477046170978e-47 ], 
                    [ 144.75 , 3.723370982618436e-47 ], [ 144.8125 , 3.723266799544105e-47 ], [ 144.875 , 3.7231649948915184e-47 ], [ 144.9375 , 3.7230631930225593e-47 ], 
                    [ 145.0 , 3.7229613939371516e-47 ], [ 145.0625 , 3.722859597635272e-47 ], [ 145.125 , 3.722758608409857e-47 ], [ 145.1875 , 3.722661034681273e-47 ], 
                    [ 145.25 , 3.7225634635100493e-47 ], [ 145.3125 , 3.7224658948961984e-47 ], [ 145.375 , 3.722368328839627e-47 ], [ 145.4375 , 3.7222707653402677e-47 ], 
                    [ 145.5 , 3.722177102941436e-47 ], [ 145.5625 , 3.7220837453652433e-47 ], [ 145.625 , 3.7219903901305675e-47 ], [ 145.6875 , 3.7218970372373493e-47 ], 
                    [ 145.75 , 3.721803686685556e-47 ], [ 145.8125 , 3.7217130689474417e-47 ], [ 145.875 , 3.721623902544418e-47 ], [ 145.9375 , 3.7215347382776813e-47 ], 
                    [ 146.0 , 3.721445576147207e-47 ], [ 146.0625 , 3.721356416152838e-47 ], [ 146.125 , 3.721268773027662e-47 ], [ 146.1875 , 3.721183747662668e-47 ], 
                    [ 146.25 , 3.7210987242403218e-47 ], [ 146.3125 , 3.7210137027606323e-47 ], [ 146.375 , 3.7209286832235815e-47 ], [ 146.4375 , 3.7208439461027924e-47 ], 
                    [ 146.5 , 3.7207630452841694e-47 ], [ 146.5625 , 3.7206821462245414e-47 ], [ 146.625 , 3.720601248923948e-47 ], [ 146.6875 , 3.7205203533821924e-47 ], 
                    [ 146.75 , 3.720439459599291e-47 ], [ 146.8125 , 3.7203616456633546e-47 ], [ 146.875 , 3.7202848181929475e-47 ], [ 146.9375 , 3.720207992309122e-47 ], 
                    [ 147.0 , 3.720131168011792e-47 ], [ 147.0625 , 3.7200543453008985e-47 ], [ 147.125 , 3.719979288054706e-47 ], [ 147.1875 , 3.7199065244109096e-47 ], 
                    [ 147.25 , 3.7198337621904393e-47 ], [ 147.3125 , 3.719761001393215e-47 ], [ 147.375 , 3.7196882420191818e-47 ], [ 147.4375 , 3.7196158885266757e-47 ], 
                    [ 147.5 , 3.719547163596591e-47 ], [ 147.5625 , 3.7194784399362654e-47 ], [ 147.625 , 3.7194097175457027e-47 ], [ 147.6875 , 3.7193409964249585e-47 ], 
                    [ 147.75 , 3.7192722765738505e-47 ], [ 147.8125 , 3.719206561649882e-47 ], [ 147.875 , 3.719141830804165e-47 ], [ 147.9375 , 3.7190771010850817e-47 ], 
                    [ 148.0 , 3.719012372492532e-47 ], [ 148.0625 , 3.718947645026576e-47 ], [ 148.125 , 3.718884482532493e-47 ], [ 148.1875 , 3.718823720743556e-47 ], 
                    [ 148.25 , 3.7187629599473896e-47 ], [ 148.3125 , 3.718702200143976e-47 ], [ 148.375 , 3.7186414413332994e-47 ], [ 148.4375 , 3.7185807763226446e-47 ], 
                    [ 148.5 , 3.718523965322003e-47 ], [ 148.5625 , 3.71846715518935e-47 ], [ 148.625 , 3.7184103459246197e-47 ], [ 148.6875 , 3.718353537527773e-47 ], 
                    [ 148.75 , 3.718296729998795e-47 ], [ 148.8125 , 3.7182424187154674e-47 ], [ 148.875 , 3.7181895099071616e-47 ], [ 148.9375 , 3.71813660185167e-47 ], 
                    [ 149.0 , 3.7180836945490607e-47 ], [ 149.0625 , 3.7180307879992977e-47 ], [ 149.125 , 3.717978834017614e-47 ], [ 149.1875 , 3.7179298170413835e-47 ], 
                    [ 149.25 , 3.717880800711382e-47 ], [ 149.3125 , 3.717831785027652e-47 ], [ 149.375 , 3.717782769990055e-47 ], [ 149.4375 , 3.7177337555986613e-47 ], 
                    [ 149.5 , 3.7176879761299944e-47 ], [ 149.5625 , 3.7176428265168495e-47 ], [ 149.625 , 3.7175976774520266e-47 ], [ 149.6875 , 3.717552528935518e-47 ], 
                    [ 149.75 , 3.717507380967265e-47 ], [ 149.8125 , 3.7174638296903416e-47 ], [ 149.875 , 3.71742250786818e-47 ], [ 149.9375 , 3.717381186505336e-47 ], 
                    [ 150.0 , 3.717339865601803e-47 ], [ 150.0625 , 3.717298545157602e-47 ], [ 150.125 , 3.717257225172704e-47 ], [ 150.1875 , 3.7172196471625525e-47 ], 
                    [ 150.25 , 3.717182126046473e-47 ], [ 150.3125 , 3.717144605309127e-47 ], [ 150.375 , 3.7171070849505365e-47 ], [ 150.4375 , 3.7170695649706186e-47 ], 
                    [ 150.5 , 3.7170340901076934e-47 ], [ 150.5625 , 3.717000352899249e-47 ], [ 150.625 , 3.71696661599699e-47 ], [ 150.6875 , 3.7169328794009666e-47 ], 
                    [ 150.75 , 3.71689914311107e-47 ], [ 150.8125 , 3.7168657222579947e-47 ], [ 150.875 , 3.716835752445904e-47 ], [ 150.9375 , 3.716805782875465e-47 ], 
                    [ 151.0 , 3.7167758135466766e-47 ], [ 151.0625 , 3.716745844459537e-47 ], [ 151.125 , 3.716715875614017e-47 ], [ 151.1875 , 3.7166881834493305e-47 ], 
                    [ 151.25 , 3.716661919219114e-47 ], [ 151.3125 , 3.716635655174548e-47 ], [ 151.375 , 3.7166093913155525e-47 ], [ 151.4375 , 3.7165831276421523e-47 ], 
                    [ 151.5 , 3.716557355927005e-47 ], [ 151.5625 , 3.7165347957050424e-47 ], [ 151.625 , 3.716512235620024e-47 ], [ 151.6875 , 3.716489675671977e-47 ], 
                    [ 151.75 , 3.716467115860899e-47 ], [ 151.8125 , 3.716444556186684e-47 ], [ 151.875 , 3.7164243498001643e-47 ], [ 151.9375 , 3.7164054736099185e-47 ], 
                    [ 152.0 , 3.716386597515521e-47 ], [ 152.0625 , 3.7163677215170244e-47 ], [ 152.125 , 3.716348845614401e-47 ], [ 152.1875 , 3.716330451356022e-47 ], 
                    [ 152.25 , 3.7163152148981102e-47 ], [ 152.3125 , 3.716299978502666e-47 ], [ 152.375 , 3.716284742169662e-47 ], [ 152.4375 , 3.716269505899126e-47 ], 
                    [ 152.5 , 3.7162542696911087e-47 ], [ 152.5625 , 3.7162412672284206e-47 ], [ 152.625 , 3.7162296733152093e-47 ], [ 152.6875 , 3.7162180794381425e-47 ], 
                    [ 152.75 , 3.7162064855971935e-47 ], [ 152.8125 , 3.7161948917924406e-47 ], [ 152.875 , 3.7161835916983176e-47 ], [ 152.9375 , 3.7161755828487923e-47 ], 
                    [ 153.0 , 3.7161675740165e-47 ], [ 153.0625 , 3.716159565201468e-47 ], [ 153.125 , 3.7161515564037493e-47 ], [ 153.1875 , 3.7161435476231843e-47 ], 
                    [ 153.25 , 3.7161374637589133e-47 ], [ 153.3125 , 3.7161330334149907e-47 ], [ 153.375 , 3.716128603076323e-47 ], [ 153.4375 , 3.7161241727429645e-47 ], 
                    [ 153.5 , 3.7161197424148337e-47 ], [ 153.5625 , 3.7161153120920115e-47 ], [ 153.625 , 3.716114379890818e-47 ], [ 153.6875 , 3.716113514423909e-47 ], 
                    [ 153.75 , 3.7161126489570947e-47 ], [ 153.8125 , 3.716111783490562e-47 ], [ 153.875 , 3.7161109180242303e-47 ], [ 153.9375 , 3.716111501625593e-47 ], 
                    [ 154.0 , 3.7161141533150455e-47 ], [ 154.0625 , 3.71611680500639e-47 ], [ 154.125 , 3.7161194566996267e-47 ], [ 154.1875 , 3.716122108394782e-47 ], 
                    [ 154.25 , 3.716124760091803e-47 ], [ 154.3125 , 3.7161303187452405e-47 ], [ 154.375 , 3.7161364717960234e-47 ], [ 154.4375 , 3.7161426248570204e-47 ], 
                    [ 154.5 , 3.7161487779282056e-47 ], [ 154.5625 , 3.716154931009578e-47 ], [ 154.625 , 3.7161619089663825e-47 ], [ 154.6875 , 3.7161715596403566e-47 ], 
                    [ 154.75 , 3.716181210339419e-47 ], [ 154.8125 , 3.716190861063544e-47 ], [ 154.875 , 3.7162005118127315e-47 ], [ 154.9375 , 3.7162101625869544e-47 ], 
                    [ 155.0 , 3.7162219825459115e-47 ], [ 155.0625 , 3.716235083946837e-47 ], [ 155.125 , 3.7162481853939516e-47 ], [ 155.1875 , 3.716261286887334e-47 ], 
                    [ 155.25 , 3.716274388426799e-47 ], [ 155.3125 , 3.7162875203217565e-47 ], [ 155.375 , 3.716304066197179e-47 ], [ 155.4375 , 3.716320612146269e-47 ], 
                    [ 155.5 , 3.716337158169025e-47 ], [ 155.5625 , 3.716353704265449e-47 ], [ 155.625 , 3.716370250435566e-47 ], [ 155.6875 , 3.7163880793889414e-47 ], 
                    [ 155.75 , 3.7164080357399893e-47 ], [ 155.8125 , 3.7164279921982267e-47 ], [ 155.875 , 3.7164479487636784e-47 ], [ 155.9375 , 3.716467905436215e-47 ], 
                    [ 156.0 , 3.716487862215968e-47 ], [ 156.0625 , 3.716510292320709e-47 ], [ 156.125 , 3.7165336332393793e-47 ], [ 156.1875 , 3.716556974304665e-47 ], 
                    [ 156.25 , 3.716580315516487e-47 ], [ 156.3125 , 3.7166036568749793e-47 ], [ 156.375 , 3.7166272431700124e-47 ], [ 156.4375 , 3.7166539547528594e-47 ], 
                    [ 156.5 , 3.7166806665276305e-47 ], [ 156.5625 , 3.716707378494354e-47 ], [ 156.625 , 3.716734090653137e-47 ], [ 156.6875 , 3.7167608030038755e-47 ], 
                    [ 156.75 , 3.7167888627325602e-47 ], [ 156.8125 , 3.7168189287140208e-47 ], [ 156.875 , 3.7168489949386398e-47 ], [ 156.9375 , 3.716879061406524e-47 ], 
                    [ 157.0 , 3.716909128117597e-47 ], [ 157.0625 , 3.7169391950719124e-47 ], [ 157.125 , 3.7169716449867177e-47 ], [ 157.1875 , 3.717005031393726e-47 ], 
                    [ 157.25 , 3.7170384181005637e-47 ], [ 157.3125 , 3.717071805107339e-47 ], [ 157.375 , 3.717105192413974e-47 ], [ 157.4375 , 3.7171386498182197e-47 ], 
                    [ 157.5 , 3.717175342571557e-47 ], [ 157.5625 , 3.7172120356870967e-47 ], [ 157.625 , 3.7172487291648166e-47 ], [ 157.6875 , 3.7172854230047995e-47 ], 
                    [ 157.75 , 3.7173221172069425e-47 ], [ 157.8125 , 3.717359836319359e-47 ], [ 157.875 , 3.7173998201580336e-47 ], [ 157.9375 , 3.7174398044267473e-47 ], 
                    [ 158.0 , 3.7174797891255307e-47 ], [ 158.0625 , 3.717519774254389e-47 ], [ 158.125 , 3.7175597598133785e-47 ], [ 158.1875 , 3.7176016558565473e-47 ], 
                    [ 158.25 , 3.7176448854299833e-47 ], [ 158.3125 , 3.717688115506107e-47 ], [ 158.375 , 3.7177313460849e-47 ], [ 158.4375 , 3.7177745771663403e-47 ], 
                    [ 158.5 , 3.7178178087506183e-47 ], [ 158.5625 , 3.7178638164702036e-47 ], [ 158.625 , 3.717910301420588e-47 ], [ 158.6875 , 3.717956786952179e-47 ], 
                    [ 158.75 , 3.718003273064986e-47 ], [ 158.8125 , 3.718049759759041e-47 ], [ 158.875 , 3.718096585451177e-47 ], [ 158.9375 , 3.7181462976006133e-47 ], 
                    [ 159.0 , 3.7181960104147427e-47 ], [ 159.0625 , 3.7182457238934965e-47 ], [ 159.125 , 3.718295438036961e-47 ], [ 159.1875 , 3.7183451528451206e-47 ], 
                    [ 159.25 , 3.7183959630332356e-47 ], [ 159.3125 , 3.718448867473304e-47 ], [ 159.375 , 3.7185017726661377e-47 ], [ 159.4375 , 3.71855467861164e-47 ], 
                    [ 159.5 , 3.718607585309902e-47 ], [ 159.5625 , 3.7186604927609343e-47 ], [ 159.625 , 3.718715220022143e-47 ], [ 159.6875 , 3.7187713280137335e-47 ], 
                    [ 159.75 , 3.718827436851803e-47 ], [ 159.8125 , 3.7188835465365223e-47 ], [ 159.875 , 3.718939657067745e-47 ], [ 159.9375 , 3.7189957684456436e-47 ], 
                    [ 160.0 , 3.719054333337089e-47 ], [ 160.0625 , 3.7191135870744165e-47 ], [ 160.125 , 3.719172841755829e-47 ], [ 160.1875 , 3.7192320973813156e-47 ], 
                    [ 160.25 , 3.71929135395089e-47 ], [ 160.3125 , 3.7193506114645426e-47 ], [ 160.375 , 3.719412944075116e-47 ], [ 160.4375 , 3.719475339957113e-47 ], 
                    [ 160.5 , 3.7195377368858466e-47 ], [ 160.5625 , 3.719600134861308e-47 ], [ 160.625 , 3.7196625338835674e-47 ], [ 160.6875 , 3.719725453035919e-47 ], 
                    [ 160.75 , 3.7197909797159573e-47 ], [ 160.8125 , 3.719856507550339e-47 ], [ 160.875 , 3.719922036539085e-47 ], [ 160.9375 , 3.7199875666821645e-47 ], 
                    [ 161.0 , 3.720053097979596e-47 ], [ 161.0625 , 3.7201196802926347e-47 ], [ 161.125 , 3.720188310544132e-47 ], [ 161.1875 , 3.720256942061721e-47 ], 
                    [ 161.25 , 3.720325574845426e-47 ], [ 161.3125 , 3.7203942088953214e-47 ], [ 161.375 , 3.7204628442112995e-47 ], [ 161.4375 , 3.7205330147950164e-47 ], 
                    [ 161.5 , 3.7206047295946687e-47 ], [ 161.5625 , 3.720676445776627e-47 ], [ 161.625 , 3.720748163340892e-47 ], [ 161.6875 , 3.7208198822875946e-47 ], 
                    [ 161.75 , 3.7208916026167096e-47 ], [ 161.8125 , 3.720965304446709e-47 ], [ 161.875 , 3.721040097781701e-47 ], [ 161.9375 , 3.7211148926199716e-47 ], 
                    [ 162.0 , 3.7211896889616843e-47 ], [ 162.0625 , 3.7212644868068946e-47 ], [ 162.125 , 3.7213392861555284e-47 ], [ 162.1875 , 3.721416450556196e-47 ], 
                    [ 162.25 , 3.7214942906903684e-47 ], [ 162.3125 , 3.721572132452708e-47 ], [ 162.375 , 3.7216499758432484e-47 ], [ 162.4375 , 3.721727820862024e-47 ], 
                    [ 162.5 , 3.7218056675090943e-47 ], [ 162.5625 , 3.721886220452819e-47 ], [ 162.625 , 3.72196708262689e-47 ], [ 162.6875 , 3.722047946557782e-47 ], 
                    [ 162.75 , 3.7221288122455873e-47 ], [ 162.8125 , 3.722209679690212e-47 ], [ 162.875 , 3.7222905623772627e-47 ], [ 162.9375 , 3.722374443546779e-47 ], 
                    [ 163.0 , 3.722458326606543e-47 ], [ 163.0625 , 3.7225422115565703e-47 ], [ 163.125 , 3.722626098396904e-47 ], [ 163.1875 , 3.722709987127613e-47 ], 
                    [ 163.25 , 3.722794169053075e-47 ], [ 163.3125 , 3.7228810463145646e-47 ], [ 163.375 , 3.7229679256034455e-47 ], [ 163.4375 , 3.7230548069198437e-47 ], 
                    [ 163.5 , 3.723141690263755e-47 ], [ 163.5625 , 3.723228575635199e-47 ], [ 163.625 , 3.723315990151882e-47 ], [ 163.6875 , 3.72340585302404e-47 ], 
                    [ 163.75 , 3.7234957180650804e-47 ], [ 163.8125 , 3.723585585275029e-47 ], [ 163.875 , 3.7236754546539375e-47 ], [ 163.9375 , 3.7237653262018327e-47 ], 
                    [ 164.0 , 3.723855914924362e-47 ], [ 164.0625 , 3.7239487250988427e-47 ], [ 164.125 , 3.724041537586497e-47 ], [ 164.1875 , 3.724134352387303e-47 ], 
                    [ 164.25 , 3.7242271695013457e-47 ], [ 164.3125 , 3.724319988928655e-47 ], [ 164.375 , 3.7244136822190573e-47 ], [ 164.4375 , 3.724509446027385e-47 ], 
                    [ 164.5 , 3.7246052122980085e-47 ], [ 164.5625 , 3.724700981031018e-47 ], [ 164.625 , 3.724796752226476e-47 ], [ 164.6875 , 3.7248925258844465e-47 ], 
                    [ 164.75 , 3.724989278728353e-47 ], [ 164.8125 , 3.725087964327679e-47 ], [ 164.875 , 3.7251866525414145e-47 ], [ 164.9375 , 3.7252853433697364e-47 ], 
                    [ 165.0 , 3.7253840368126067e-47 ], [ 165.0625 , 3.725482732870201e-47 ], [ 165.125 , 3.7255824853171566e-47 ], [ 165.1875 , 3.725684099621752e-47 ], 
                    [ 165.25 , 3.725785716697825e-47 ], [ 165.3125 , 3.7258873365454523e-47 ], [ 165.375 , 3.725988959164786e-47 ], [ 165.4375 , 3.7260905845558516e-47 ], 
                    [ 165.5 , 3.726193282842952e-47 ], [ 165.5625 , 3.726297773346226e-47 ], [ 165.625 , 3.7264022667796654e-47 ], [ 165.6875 , 3.7265067631433523e-47 ], 
                    [ 165.75 , 3.726611262437291e-47 ], [ 165.8125 , 3.7267157646616144e-47 ], [ 165.875 , 3.726821337516422e-47 ], [ 165.9375 , 3.726928708486922e-47 ], 
                    [ 166.0 , 3.7270360825508155e-47 ], [ 166.0625 , 3.7271434597081657e-47 ], [ 166.125 , 3.72725083995914e-47 ], [ 166.1875 , 3.7273582233037496e-47 ], 
                    [ 166.25 , 3.727466629256717e-47 ], [ 166.3125 , 3.7275768660718184e-47 ], [ 166.375 , 3.7276871061470056e-47 ], [ 166.4375 , 3.7277973494824806e-47 ], 
                    [ 166.5 , 3.727907596078313e-47 ], [ 166.5625 , 3.7280178459346517e-47 ], [ 166.625 , 3.7281290309819854e-47 ], [ 166.6875 , 3.7282421149311326e-47 ], 
                    [ 166.75 , 3.728355202310386e-47 ], [ 166.8125 , 3.7284682931199033e-47 ], [ 166.875 , 3.7285813873597626e-47 ], [ 166.9375 , 3.72869448503004e-47 ], 
                    [ 167.0 , 3.7288083937741726e-47 ], [ 167.0625 , 3.7289243074151835e-47 ], [ 167.125 , 3.729040224659536e-47 ], [ 167.1875 , 3.7291561455072096e-47 ], 
                    [ 167.25 , 3.729272069958475e-47 ], [ 167.3125 , 3.729387998013312e-47 ], [ 167.375 , 3.729504576088179e-47 ], [ 167.4375 , 3.729623301583922e-47 ], 
                    [ 167.5 , 3.7297420308591864e-47 ], [ 167.5625 , 3.7298607639140926e-47 ], [ 167.625 , 3.729979500748842e-47 ], [ 167.6875 , 3.7300982413633663e-47 ], 
                    [ 167.75 , 3.730217435697202e-47 ], [ 167.8125 , 3.73033895726998e-47 ], [ 167.875 , 3.7304604828016407e-47 ], [ 167.9375 , 3.730582012292313e-47 ], 
                    [ 168.0 , 3.7307035457421258e-47 ], [ 168.0625 , 3.7308250831512346e-47 ], [ 168.125 , 3.7309468416190663e-47 ], [ 168.1875 , 3.7310711393186287e-47 ], 
                    [ 168.25 , 3.7311954411592625e-47 ], [ 168.3125 , 3.731319747141078e-47 ], [ 168.375 , 3.731444057264135e-47 ], [ 168.4375 , 3.731568371528598e-47 ], 
                    [ 168.5 , 3.7316926899346825e-47 ], [ 168.5625 , 3.7318197242075795e-47 ], [ 168.625 , 3.731946813187529e-47 ], [ 168.6875 , 3.7320739064955567e-47 ], 
                    [ 168.75 , 3.73220100413181e-47 ], [ 168.8125 , 3.732328106096437e-47 ], [ 168.875 , 3.7324552123895844e-47 ], [ 168.9375 , 3.732584687121714e-47 ], 
                    [ 169.0 , 3.732714516869462e-47 ], [ 169.0625 , 3.7328443511330254e-47 ], [ 169.125 , 3.732974189912668e-47 ], [ 169.1875 , 3.7331040332084144e-47 ], 
                    [ 169.25 , 3.7332338810204736e-47 ], [ 169.3125 , 3.7333657583835606e-47 ], [ 169.375 , 3.7334983296277475e-47 ], [ 169.4375 , 3.733630905579519e-47 ], 
                    [ 169.5 , 3.7337634862390414e-47 ], [ 169.5625 , 3.733896071606483e-47 ], [ 169.625 , 3.734028661681958e-47 ], [ 169.6875 , 3.734162896378441e-47 ], 
                    [ 169.75 , 3.734298195382193e-47 ], [ 169.8125 , 3.734433499288147e-47 ], [ 169.875 , 3.7345688080965607e-47 ], [ 169.9375 , 3.734704121807612e-47 ], 
                    [ 170.0 , 3.734839440421425e-47 ], [ 170.0625 , 3.7349759825866132e-47 ], [ 170.125 , 3.7351139902039374e-47 ], [ 170.1875 , 3.735252002920652e-47 ], 
                    [ 170.25 , 3.735390020736893e-47 ], [ 170.3125 , 3.735528043652955e-47 ], [ 170.375 , 3.735666071669e-47 ], [ 170.4375 , 3.735804867869257e-47 ], 
                    [ 170.5 , 3.7359455528699017e-47 ], [ 170.5625 , 3.7360862431685127e-47 ], [ 170.625 , 3.7362269387653434e-47 ], [ 170.6875 , 3.7363676396605665e-47 ], 
                    [ 170.75 , 3.7365083458543556e-47 ], [ 170.8125 , 3.736649345719687e-47 ], [ 170.875 , 3.7367927367034735e-47 ], [ 170.9375 , 3.7369361331897496e-47 ], 
                    [ 171.0 , 3.7370795351787003e-47 ], [ 171.0625 , 3.7372229426706696e-47 ], [ 171.125 , 3.7373663556657355e-47 ], [ 171.1875 , 3.737509774164215e-47 ], 
                    [ 171.25 , 3.7376555885359766e-47 ], [ 171.3125 , 3.7378016351232473e-47 ], [ 171.375 , 3.7379476874171714e-47 ], [ 171.4375 , 3.7380937454180253e-47 ], 
                    [ 171.5 , 3.7382398091259793e-47 ], [ 171.5625 , 3.7383858785412824e-47 ], [ 171.625 , 3.7385338069874345e-47 ], [ 171.6875 , 3.738682517568203e-47 ], 
                    [ 171.75 , 3.738831234064322e-47 ], [ 171.8125 , 3.738979956476054e-47 ], [ 171.875 , 3.739128684803659e-47 ], [ 171.9375 , 3.7392774190473206e-47 ], 
                    [ 172.0 , 3.7394274164508226e-47 ], [ 172.0625 , 3.739578769773394e-47 ], [ 172.125 , 3.7397301292219637e-47 ], [ 172.1875 , 3.739881494796807e-47 ], 
                    [ 172.25 , 3.740032866498145e-47 ], [ 172.3125 , 3.7401842443263067e-47 ], [ 172.375 , 3.7403362619107247e-47 ], [ 172.4375 , 3.74049023204821e-47 ], 
                    [ 172.5 , 3.740644208523842e-47 ], [ 172.5625 , 3.7407981913378294e-47 ], [ 172.625 , 3.740952180490593e-47 ], [ 172.6875 , 3.74110617598218e-47 ], 
                    [ 172.75 , 3.741260177812985e-47 ], [ 172.8125 , 3.741416755144519e-47 ], [ 172.875 , 3.7415733526667867e-47 ], [ 172.9375 , 3.7417299567434395e-47 ], 
                    [ 173.0 , 3.7418865673747767e-47 ], [ 173.0625 , 3.7420431845611286e-47 ], [ 173.125 , 3.742199808302634e-47 ], [ 173.1875 , 3.7423582943702236e-47 ], 
                    [ 173.25 , 3.7425174743961026e-47 ], [ 173.3125 , 3.742676661192627e-47 ], [ 173.375 , 3.742835854760191e-47 ], [ 173.4375 , 3.7429950550989237e-47 ], 
                    [ 173.5 , 3.7431542622092184e-47 ], [ 173.5625 , 3.743314628503981e-47 ], [ 173.625 , 3.74347640795993e-47 ], [ 173.6875 , 3.7436381944076753e-47 ], 
                    [ 173.75 , 3.7437999878476253e-47 ], [ 173.8125 , 3.7439617882798953e-47 ], [ 173.875 , 3.7441235957049753e-47 ], [ 173.9375 , 3.74428580649179e-47 ], 
                    [ 174.0 , 3.7444501533154164e-47 ], [ 174.0625 , 3.744614507352643e-47 ], [ 174.125 , 3.7447788686038645e-47 ], [ 174.1875 , 3.7449432370692934e-47 ], 
                    [ 174.25 , 3.745107612749245e-47 ], [ 174.3125 , 3.745271995644195e-47 ], [ 174.375 , 3.7454385277158907e-47 ], [ 174.4375 , 3.7456054473389928e-47 ], 
                    [ 174.5 , 3.7457723744010255e-47 ], [ 174.5625 , 3.7459393089023756e-47 ], [ 174.625 , 3.746106250843293e-47 ], [ 174.6875 , 3.74627320022427e-47 ], 
                    [ 174.75 , 3.7464414785732206e-47 ], [ 174.8125 , 3.7466109489379143e-47 ], [ 174.875 , 3.7467804269686044e-47 ], [ 174.9375 , 3.746949912665663e-47 ], 
                    [ 175.0 , 3.747119406029331e-47 ], [ 175.0625 , 3.7472889070601424e-47 ], [ 175.125 , 3.747458890577037e-47 ], [ 175.1875 , 3.7476308925195665e-47 ], 
                    [ 175.25 , 3.7478029023566903e-47 ], [ 175.3125 , 3.747974920088824e-47 ], [ 175.375 , 3.7481469457162244e-47 ], [ 175.4375 , 3.748318979239306e-47 ], 
                    [ 175.5 , 3.748491020658405e-47 ], [ 175.5625 , 3.748665150616256e-47 ], [ 175.625 , 3.748839684051105e-47 ], [ 175.6875 , 3.7490142256119724e-47 ], 
                    [ 175.75 , 3.749188775299289e-47 ], [ 175.8125 , 3.7493633331134867e-47 ], [ 175.875 , 3.749537899054864e-47 ], [ 175.9375 , 3.749713647114422e-47 ], 
                    [ 176.0 , 3.7498907000011915e-47 ], [ 176.0625 , 3.7500677612480197e-47 ], [ 176.125 , 3.7502448308552997e-47 ], [ 176.1875 , 3.7504219088233223e-47 ], 
                    [ 176.25 , 3.750598995152587e-47 ], [ 176.3125 , 3.75077632222781e-47 ], [ 176.375 , 3.7509558541729383e-47 ], [ 176.4375 , 3.7511353947114933e-47 ], 
                    [ 176.5 , 3.7513149438436983e-47 ], [ 176.5625 , 3.751494501570126e-47 ], [ 176.625 , 3.751674067891161e-47 ], [ 176.6875 , 3.75185364280716e-47 ], 
                    [ 176.75 , 3.7520349449079554e-47 ], [ 176.8125 , 3.752216986744816e-47 ], [ 176.875 , 3.7523990374140386e-47 ], [ 176.9375 , 3.7525810969160255e-47 ], 
                    [ 177.0 , 3.7527631652512045e-47 ], [ 177.0625 , 3.752945242420004e-47 ], [ 177.125 , 3.7531280380583346e-47 ], [ 177.1875 , 3.7533125540211524e-47 ], 
                    [ 177.25 , 3.753497079055375e-47 ], [ 177.3125 , 3.7536816131614734e-47 ], [ 177.375 , 3.753866156339841e-47 ], [ 177.4375 , 3.754050708590978e-47 ], 
                    [ 177.5 , 3.7542352699153016e-47 ], [ 177.5625 , 3.754421936545595e-47 ], [ 177.625 , 3.7546089250385478e-47 ], [ 177.6875 , 3.7547959228444113e-47 ], 
                    [ 177.75 , 3.75498292996365e-47 ], [ 177.8125 , 3.7551699463968057e-47 ], [ 177.875 , 3.7553569721442646e-47 ], [ 177.9375 , 3.7555450380740124e-47 ], 
                    [ 178.0 , 3.755734465359929e-47 ], [ 178.0625 , 3.7559239022004367e-47 ], [ 178.125 , 3.756113348596017e-47 ], [ 178.1875 , 3.756302804547152e-47 ], 
                    [ 178.25 , 3.7564922700543775e-47 ], [ 178.3125 , 3.7566817451180686e-47 ], [ 178.375 , 3.75687357943611e-47 ], [ 178.4375 , 3.757065470331778e-47 ], 
                    [ 178.5 , 3.7572573710287106e-47 ], [ 178.5625 , 3.7574492815274097e-47 ], [ 178.625 , 3.7576412018283487e-47 ], [ 178.6875 , 3.7578331319320823e-47 ], 
                    [ 178.75 , 3.758026285100422e-47 ], [ 178.8125 , 3.7582205972639003e-47 ], [ 178.875 , 3.7584149194744926e-47 ], [ 178.9375 , 3.758609251732691e-47 ], 
                    [ 179.0 , 3.758803594038882e-47 ], [ 179.0625 , 3.7589979463938785e-47 ], [ 179.125 , 3.7591923850174657e-47 ], [ 179.1875 , 3.7593891276079187e-47 ], 
                    [ 179.25 , 3.759585880495171e-47 ], [ 179.3125 , 3.759782643679734e-47 ], [ 179.375 , 3.7599794171622266e-47 ], [ 179.4375 , 3.7601762009430812e-47 ], 
                    [ 179.5 , 3.76037299502289e-47 ], [ 179.5625 , 3.760571053190491e-47 ], [ 179.625 , 3.760770206518254e-47 ], [ 179.6875 , 3.7609693703928052e-47 ], 
                    [ 179.75 , 3.7611685448147573e-47 ], [ 179.8125 , 3.7613677297845876e-47 ], [ 179.875 , 3.761566925302882e-47 ], [ 179.9375 , 3.7617661903801974e-47 ], 
                    [ 180.0 , 3.76196774137574e-47 ], [ 180.0625 , 3.7621693031701467e-47 ], [ 180.125 , 3.7623708757640234e-47 ], [ 180.1875 , 3.762572459157842e-47 ], 
                    [ 180.25 , 3.7627740533523945e-47 ], [ 180.3125 , 3.7629756583480187e-47 ], [ 180.375 , 3.7631784272146083e-47 ], [ 180.4375 , 3.763382361616124e-47 ], 
                    [ 180.5 , 3.763586307069238e-47 ], [ 180.5625 , 3.763790263574658e-47 ], [ 180.625 , 3.7639942311328475e-47 ], [ 180.6875 , 3.7641982097444063e-47 ], 
                    [ 180.75 , 3.7644021994100136e-47 ], [ 180.8125 , 3.764608427996667e-47 ], [ 180.875 , 3.7648147626498006e-47 ], [ 180.9375 , 3.765021108611969e-47 ], 
                    [ 181.0 , 3.7652274658837393e-47 ], [ 181.0625 , 3.765433834465732e-47 ], [ 181.125 , 3.7656402143586722e-47 ], [ 181.1875 , 3.7658475288597205e-47 ], 
                    [ 181.25 , 3.7660562169408516e-47 ], [ 181.3125 , 3.7662649165866614e-47 ], [ 181.375 , 3.7664736277977374e-47 ], [ 181.4375 , 3.7666823505748e-47 ], 
                    [ 181.5 , 3.766891084918357e-47 ], [ 181.5625 , 3.767099830829156e-47 ], [ 181.625 , 3.767310495582483e-47 ], [ 181.6875 , 3.7675215519827764e-47 ], 
                    [ 181.75 , 3.767732620207103e-47 ], [ 181.8125 , 3.767943700256152e-47 ], [ 181.875 , 3.7681547921305586e-47 ], [ 181.9375 , 3.7683658958309866e-47 ], 
                    [ 182.0 , 3.7685775800595468e-47 ], [ 182.0625 , 3.768790991883462e-47 ], [ 182.125 , 3.7690044157926805e-47 ], [ 182.1875 , 3.769217851787995e-47 ], 
                    [ 182.25 , 3.7694312998699824e-47 ], [ 182.3125 , 3.769644760039433e-47 ], [ 182.375 , 3.7698582322969523e-47 ], [ 182.4375 , 3.770073170043928e-47 ], 
                    [ 182.5 , 3.770288905722214e-47 ], [ 182.5625 , 3.7705046537455874e-47 ], [ 182.625 , 3.7707204141147816e-47 ], [ 182.6875 , 3.770936186830422e-47 ], 
                    [ 182.75 , 3.771151971893323e-47 ], [ 182.8125 , 3.771367846664005e-47 ], [ 182.875 , 3.7715859159002906e-47 ], [ 182.9375 , 3.7718039977458183e-47 ], 
                    [ 183.0 , 3.7720220922013438e-47 ], [ 183.0625 , 3.7722401992675967e-47 ], [ 183.125 , 3.772458318945306e-47 ], [ 183.1875 , 3.772676451235201e-47 ], 
                    [ 183.25 , 3.7728954989653105e-47 ], [ 183.3125 , 3.773115889616358e-47 ], [ 183.375 , 3.7733362931413465e-47 ], [ 183.4375 , 3.7735567095410843e-47 ], 
                    [ 183.5 , 3.773777138816214e-47 ], [ 183.5625 , 3.7739975809675425e-47 ], [ 183.625 , 3.7742180359958755e-47 ], [ 183.6875 , 3.7744401994120157e-47 ], 
                    [ 183.75 , 3.774662912771353e-47 ], [ 183.8125 , 3.774885639272122e-47 ], [ 183.875 , 3.775108378914962e-47 ], [ 183.9375 , 3.77533113170073e-47 ], 
                    [ 184.0 , 3.7755538976301745e-47 ], [ 184.0625 , 3.7757768962484034e-47 ], [ 184.125 , 3.77600189464346e-47 ], [ 184.1875 , 3.776226906446135e-47 ], 
                    [ 184.25 , 3.776451931657281e-47 ], [ 184.3125 , 3.7766769702776706e-47 ], [ 184.375 , 3.7769020223081027e-47 ], [ 184.4375 , 3.7771270877493763e-47 ], 
                    [ 184.5 , 3.777353101507659e-47 ], [ 184.5625 , 3.7775803964330943e-47 ], [ 184.625 , 3.777807705035536e-47 ], [ 184.6875 , 3.7780350273158335e-47 ], 
                    [ 184.75 , 3.778262363274784e-47 ], [ 184.8125 , 3.7784897129133426e-47 ], [ 184.875 , 3.778717076232093e-47 ], [ 184.9375 , 3.7789460725185284e-47 ], 
                    [ 185.0 , 3.7791756688230067e-47 ], [ 185.0625 , 3.7794052790769727e-47 ], [ 185.125 , 3.7796349032813006e-47 ], [ 185.1875 , 3.7798645414368645e-47 ], 
                    [ 185.25 , 3.780094193544513e-47 ], [ 185.3125 , 3.7803239224688336e-47 ], [ 185.375 , 3.7805557733320744e-47 ], [ 185.4375 , 3.7807876384150033e-47 ], 
                    [ 185.5 , 3.7810195177183846e-47 ], [ 185.5625 , 3.7812514112431444e-47 ], [ 185.625 , 3.7814833189901813e-47 ], [ 185.6875 , 3.7817152409603683e-47 ], 
                    [ 185.75 , 3.781947844976681e-47 ], [ 185.8125 , 3.78218196045032e-47 ], [ 185.875 , 3.7824160904165063e-47 ], [ 185.9375 , 3.7826502348761903e-47 ], 
                    [ 186.0 , 3.7828843938301603e-47 ], [ 186.0625 , 3.783118567279396e-47 ], [ 186.125 , 3.7833527552247134e-47 ], [ 186.1875 , 3.783588194496142e-47 ], 
                    [ 186.25 , 3.783824571519616e-47 ], [ 186.3125 , 3.7840609633105797e-47 ], [ 186.375 , 3.7842973698699546e-47 ], [ 186.4375 , 3.784533791198637e-47 ], 
                    [ 186.5 , 3.784770227297603e-47 ], [ 186.5625 , 3.785006678167722e-47 ], [ 186.625 , 3.785244918159579e-47 ], [ 186.6875 , 3.7854835584488367e-47 ], 
                    [ 186.75 , 3.7857222137831403e-47 ], [ 186.8125 , 3.78596088416344e-47 ], [ 186.875 , 3.786199569590711e-47 ], [ 186.9375 , 3.7864382700658477e-47 ], 
                    [ 187.0 , 3.786677102800818e-47 ], [ 187.0625 , 3.7869179507674943e-47 ], [ 187.125 , 3.7871588140530745e-47 ], [ 187.1875 , 3.7873996926585325e-47 ], 
                    [ 187.25 , 3.7876405865848174e-47 ], [ 187.3125 , 3.787881495832956e-47 ], [ 187.375 , 3.788122420403842e-47 ], [ 187.4375 , 3.7883639482064823e-47 ], 
                    [ 187.5 , 3.7886070523870515e-47 ], [ 187.5625 , 3.788850172167955e-47 ], [ 187.625 , 3.7890933075501677e-47 ], [ 187.6875 , 3.7893364585346357e-47 ], 
                    [ 187.75 , 3.789579625122442e-47 ], [ 187.8125 , 3.789822807314614e-47 ], [ 187.875 , 3.7900670091716334e-47 ], [ 187.9375 , 3.7903123293122704e-47 ], 
                    [ 188.0 , 3.7905576653318827e-47 ], [ 188.0625 , 3.790803017231336e-47 ], [ 188.125 , 3.7910483850117126e-47 ], [ 188.1875 , 3.79129376867404e-47 ], 
                    [ 188.25 , 3.7915391682193465e-47 ], [ 188.3125 , 3.7917859704979145e-47 ], [ 188.375 , 3.792033490163363e-47 ], [ 188.4375 , 3.7922810259863916e-47 ], 
                    [ 188.5 , 3.792528577967975e-47 ], [ 188.5625 , 3.7927761461092757e-47 ], [ 188.625 , 3.793023730411294e-47 ], [ 188.6875 , 3.793271330875004e-47 ], 
                    [ 188.75 , 3.793520713394085e-47 ], [ 188.8125 , 3.793770458346176e-47 ], [ 188.875 , 3.7940202197400744e-47 ], [ 188.9375 , 3.794269997576971e-47 ], 
                    [ 189.0 , 3.7945197918578393e-47 ], [ 189.0625 , 3.7947696025838433e-47 ], [ 189.125 , 3.795019429756012e-47 ], [ 189.1875 , 3.7952713540121585e-47 ], 
                    [ 189.25 , 3.7955233100820243e-47 ], [ 189.3125 , 3.795775282878457e-47 ], [ 189.375 , 3.79602727240254e-47 ], [ 189.4375 , 3.7962792786553846e-47 ], 
                    [ 189.5 , 3.7965313016381817e-47 ], [ 189.5625 , 3.796783618193728e-47 ], [ 189.625 , 3.797037742013443e-47 ], [ 189.6875 , 3.797291882841952e-47 ], 
                    [ 189.75 , 3.797546040680448e-47 ], [ 189.8125 , 3.797800215530122e-47 ], [ 189.875 , 3.7980544073919525e-47 ], [ 189.9375 , 3.798308616267293e-47 ], 
                    [ 190.0 , 3.7985633759991733e-47 ], [ 190.0625 , 3.7988196761151033e-47 ], [ 190.125 , 3.7990759935243207e-47 ], [ 190.1875 , 3.799332328228046e-47 ], 
                    [ 190.25 , 3.799588680227418e-47 ], [ 190.3125 , 3.799845049523633e-47 ], [ 190.375 , 3.800101436117748e-47 ], [ 190.4375 , 3.8003585980468965e-47 ], 
                    [ 190.5 , 3.800617070393489e-47 ], [ 190.5625 , 3.8008755603195174e-47 ], [ 190.625 , 3.8011340678260415e-47 ], [ 190.6875 , 3.8013925929143384e-47 ], 
                    [ 190.75 , 3.8016511355856585e-47 ], [ 190.8125 , 3.801909695841143e-47 ], [ 190.875 , 3.802169228923964e-47 ], [ 190.9375 , 3.8024298832683445e-47 ], 
                    [ 191.0 , 3.8026905554816264e-47 ], [ 191.0625 , 3.8029512455651145e-47 ], [ 191.125 , 3.8032119535198996e-47 ], [ 191.1875 , 3.8034726793473683e-47 ], 
                    [ 191.25 , 3.8037334230486117e-47 ], [ 191.3125 , 3.803995286666878e-47 ], [ 191.375 , 3.804258096925239e-47 ], [ 191.4375 , 3.804520925340651e-47 ], 
                    [ 191.5 , 3.8047837719143676e-47 ], [ 191.5625 , 3.805046636647563e-47 ], [ 191.625 , 3.805309519541572e-47 ], [ 191.6875 , 3.805572420597568e-47 ], 
                    [ 191.75 , 3.805836564844618e-47 ], [ 191.8125 , 3.806101524752474e-47 ], [ 191.875 , 3.8063665031066957e-47 ], [ 191.9375 , 3.8066314999085415e-47 ], 
                    [ 192.0 , 3.8068965151592415e-47 ], [ 192.0625 , 3.8071615488601875e-47 ], [ 192.125 , 3.80742660101261e-47 ], [ 192.1875 , 3.807692985714584e-47 ], 
                    [ 192.25 , 3.807960087529023e-47 ], [ 192.3125 , 3.80822720808005e-47 ], [ 192.375 , 3.8084943473690606e-47 ], [ 192.4375 , 3.8087615053973685e-47 ], 
                    [ 192.5 , 3.809028682166207e-47 ], [ 192.5625 , 3.8092958776769725e-47 ], [ 192.625 , 3.8095644665776464e-47 ], [ 192.6875 , 3.8098337081224815e-47 ], 
                    [ 192.75 , 3.81010296869595e-47 ], [ 192.8125 , 3.8103722482994496e-47 ], [ 192.875 , 3.810641546934326e-47 ], [ 192.9375 , 3.8109108646019243e-47 ], 
                    [ 193.0 , 3.8111802013036163e-47 ], [ 193.0625 , 3.8114509562455322e-47 ], [ 193.125 , 3.811722329211083e-47 ], [ 193.1875 , 3.8119937214982216e-47 ], 
                    [ 193.25 , 3.812265133108298e-47 ], [ 193.3125 , 3.8125365640427945e-47 ], [ 193.375 , 3.812808014302926e-47 ], [ 193.4375 , 3.813079483890204e-47 ], 
                    [ 193.5 , 3.8133523664275024e-47 ], [ 193.5625 , 3.8136258653480114e-47 ], [ 193.625 , 3.8138993838841603e-47 ], [ 193.6875 , 3.814172922037438e-47 ], 
                    [ 193.75 , 3.814446479809279e-47 ], [ 193.8125 , 3.8147200572010624e-47 ], [ 193.875 , 3.814993654214196e-47 ], [ 193.9375 , 3.815268624374757e-47 ], 
                    [ 194.0 , 3.8155442387865334e-47 ], [ 194.0625 , 3.815819873108598e-47 ], [ 194.125 , 3.816095527342578e-47 ], [ 194.1875 , 3.816371201489669e-47 ], 
                    [ 194.25 , 3.816646895551444e-47 ], [ 194.3125 , 3.816922609529289e-47 ], [ 194.375 , 3.8171996274586106e-47 ], [ 194.4375 , 3.81747734943715e-47 ], 
                    [ 194.5 , 3.817755091621471e-47 ], [ 194.5625 , 3.818032854013042e-47 ], [ 194.625 , 3.818310636613333e-47 ], [ 194.6875 , 3.818588439423789e-47 ], 
                    [ 194.75 , 3.8188662624459055e-47 ], [ 194.8125 , 3.8191452874472385e-47 ], [ 194.875 , 3.819425105621384e-47 ], [ 194.9375 , 3.819704944297004e-47 ], 
                    [ 195.0 , 3.819984803475654e-47 ], [ 195.0625 , 3.8202646831588363e-47 ], [ 195.125 , 3.8205445833479986e-47 ], [ 195.1875 , 3.8208245040446444e-47 ], 
                    [ 195.25 , 3.8211054956902886e-47 ], [ 195.3125 , 3.821387399433656e-47 ], [ 195.375 , 3.82166932397457e-47 ], [ 195.4375 , 3.821951269314563e-47 ], 
                    [ 195.5 , 3.822233235455226e-47 ], [ 195.5625 , 3.822515222398092e-47 ], [ 195.625 , 3.822797230144615e-47 ], [ 195.6875 , 3.8230801605860287e-47 ], 
                    [ 195.75 , 3.823364166023645e-47 ], [ 195.8125 , 3.823648192559249e-47 ], [ 195.875 , 3.8239322401942987e-47 ], [ 195.9375 , 3.824216308930416e-47 ], 
                    [ 196.0 , 3.8245003987691686e-47 ], [ 196.0625 , 3.824784509712151e-47 ], [ 196.125 , 3.8250693469543376e-47 ], [ 196.1875 , 3.8253554308628025e-47 ], 
                    [ 196.25 , 3.825641536168004e-47 ], [ 196.3125 , 3.825927662871597e-47 ], [ 196.375 , 3.82621381097502e-47 ], [ 196.4375 , 3.826499980480035e-47 ], 
                    [ 196.5 , 3.8267861713882155e-47 ], [ 196.5625 , 3.827072870310385e-47 ], [ 196.625 , 3.827361024227867e-47 ], [ 196.6875 , 3.827649199841509e-47 ], 
                    [ 196.75 , 3.8279373971528895e-47 ], [ 196.8125 , 3.82822561616367e-47 ], [ 196.875 , 3.8285138568754837e-47 ], [ 196.9375 , 3.828802119289993e-47 ], 
                    [ 197.0 , 3.8290906432872135e-47 ], [ 197.0625 , 3.829380861570869e-47 ], [ 197.125 , 3.8296711018510397e-47 ], [ 197.1875 , 3.8299613641293935e-47 ], 
                    [ 197.25 , 3.8302516484075696e-47 ], [ 197.3125 , 3.830541954687291e-47 ], [ 197.375 , 3.8308322829702525e-47 ], [ 197.4375 , 3.8311226332579844e-47 ], 
                    [ 197.5 , 3.831414875781489e-47 ], [ 197.5625 , 3.8317071762790387e-47 ], [ 197.625 , 3.8319994990762534e-47 ], [ 197.6875 , 3.832291844174916e-47 ], 
                    [ 197.75 , 3.832584211576809e-47 ], [ 197.8125 , 3.8328766012834446e-47 ], [ 197.875 , 3.833169013296687e-47 ], [ 197.9375 , 3.833463013058998e-47 ], 
                    [ 198.0 , 3.8337573771697452e-47 ], [ 198.0625 , 3.834051763884108e-47 ], [ 198.125 , 3.83434617320385e-47 ], [ 198.1875 , 3.8346406051307327e-47 ], 
                    [ 198.25 , 3.8349350596664946e-47 ], [ 198.3125 , 3.835229536812733e-47 ], [ 198.375 , 3.835525242877504e-47 ], [ 198.4375 , 3.835821637703103e-47 ], 
                    [ 198.5 , 3.83611805543305e-47 ], [ 198.5625 , 3.8364144960690603e-47 ], [ 198.625 , 3.8367109596128233e-47 ], [ 198.6875 , 3.8370074460661085e-47 ], 
                    [ 198.75 , 3.8373039554308227e-47 ], [ 198.8125 , 3.837601341245868e-47 ], [ 198.875 , 3.8378997876856117e-47 ], [ 198.9375 , 3.8381982573352387e-47 ], 
                    [ 199.0 , 3.8384967501965254e-47 ], [ 199.0625 , 3.838795266271361e-47 ], [ 199.125 , 3.8390938055614385e-47 ], [ 199.1875 , 3.8393923680686485e-47 ], 
                    [ 199.25 , 3.8396914063281e-47 ], [ 199.3125 , 3.8399918835757878e-47 ], [ 199.375 , 3.840292384337522e-47 ], [ 199.4375 , 3.840592908615088e-47 ], 
                    [ 199.5 , 3.840893456410326e-47 ], [ 199.5625 , 3.8411940277251856e-47 ], [ 199.625 , 3.8414946225613434e-47 ], [ 199.6875 , 3.8417952719412747e-47 ], 
                    [ 199.75 , 3.842097778186275e-47 ], [ 199.8125 , 3.8424003082509313e-47 ], [ 199.875 , 3.8427028621370085e-47 ], [ 199.9375 , 3.843005439846437e-47 ], 
                    [ 200.0 , 3.8433080413811763e-47 ], [ 200.0625 , 3.843610666742936e-47 ], [ 200.125 , 3.8439133159337033e-47 ], [ 200.1875 , 3.8442174181099905e-47 ], 
                    [ 200.25 , 3.8445219595790513e-47 ], [ 200.3125 , 3.844826525174063e-47 ], [ 200.375 , 3.84513111489702e-47 ], [ 200.4375 , 3.84543572874964e-47 ], 
                    [ 200.5 , 3.8457403667341376e-47 ], [ 200.5625 , 3.84604502885215e-47 ], [ 200.625 , 3.846350673579679e-47 ], [ 200.6875 , 3.8466572399755667e-47 ], 
                    [ 200.75 , 3.846963830805827e-47 ], [ 200.8125 , 3.8472704460723255e-47 ], [ 200.875 , 3.8475770857770097e-47 ], [ 200.9375 , 3.8478837499218553e-47 ], 
                    [ 201.0 , 3.8481904385088905e-47 ], [ 201.0625 , 3.8484975962570227e-47 ], [ 201.125 , 3.848806173415028e-47 ], [ 201.1875 , 3.8491147753150938e-47 ], 
                    [ 201.25 , 3.8494234019592294e-47 ], [ 201.3125 , 3.849732053349421e-47 ], [ 201.375 , 3.8500407294876787e-47 ], [ 201.4375 , 3.8503494303759596e-47 ], 
                    [ 201.5 , 3.850658156016166e-47 ], [ 201.5625 , 3.850968641502604e-47 ], [ 201.625 , 3.85127924227598e-47 ], [ 201.6875 , 3.8515898681009067e-47 ], 
                    [ 201.75 , 3.851900518979486e-47 ], [ 201.8125 , 3.852211194913603e-47 ], [ 201.875 , 3.852521895905414e-47 ], [ 201.9375 , 3.8528326219568855e-47 ], 
                    [ 202.0 , 3.8531445432259943e-47 ], [ 202.0625 , 3.8534571402757735e-47 ], [ 202.125 , 3.8537697626859117e-47 ], [ 202.1875 , 3.854082410458355e-47 ], 
                    [ 202.25 , 3.8543950835952726e-47 ], [ 202.3125 , 3.854707782098584e-47 ], [ 202.375 , 3.85502050597054e-47 ], [ 202.4375 , 3.855333835779072e-47 ], 
                    [ 202.5 , 3.855648433975616e-47 ], [ 202.5625 , 3.8559630578436417e-47 ], [ 202.625 , 3.8562777073852163e-47 ], [ 202.6875 , 3.856592382602462e-47 ], 
                    [ 202.75 , 3.856907083497419e-47 ], [ 202.8125 , 3.857221810072183e-47 ], [ 202.875 , 3.8575365623289314e-47 ], [ 202.9375 , 3.85785309479985e-47 ], 
                    [ 203.0 , 3.858169693743171e-47 ], [ 203.0625 , 3.8584863186685034e-47 ], [ 203.125 , 3.858802969578035e-47 ], [ 203.1875 , 3.859119646473898e-47 ], 
                    [ 203.25 , 3.8594363493581706e-47 ], [ 203.3125 , 3.8597530782329846e-47 ], [ 203.375 , 3.860070951669084e-47 ], [ 203.4375 , 3.860389537380599e-47 ], 
                    [ 203.5 , 3.8607081493860976e-47 ], [ 203.5625 , 3.8610267876878606e-47 ], [ 203.625 , 3.861345452287948e-47 ], [ 203.6875 , 3.8616641431886396e-47 ], 
                    [ 203.75 , 3.861982860392025e-47 ], [ 203.8125 , 3.8623020424760397e-47 ], [ 203.875 , 3.862622598207133e-47 ], [ 203.9375 , 3.862943180543106e-47 ], 
                    [ 204.0 , 3.863263789486168e-47 ], [ 204.0625 , 3.8635844250384435e-47 ], [ 204.125 , 3.8639050872022237e-47 ], [ 204.1875 , 3.8642257759797454e-47 ], 
                    [ 204.25 , 3.8645464913730523e-47 ], [ 204.3125 , 3.8648687560730736e-47 ], [ 204.375 , 3.865191306485527e-47 ], [ 204.4375 , 3.865513883817074e-47 ], 
                    [ 204.5 , 3.865836488069991e-47 ], [ 204.5625 , 3.8661591192464957e-47 ], [ 204.625 , 3.866481777348808e-47 ], [ 204.6875 , 3.86680446237923e-47 ], 
                    [ 204.75 , 3.867127967382474e-47 ], [ 204.8125 , 3.8674524811114463e-47 ], [ 204.875 , 3.867777022072352e-47 ], [ 204.9375 , 3.8681015902674224e-47 ], 
                    [ 205.0 , 3.8684261856988866e-47 ], [ 205.0625 , 3.868750808369141e-47 ], [ 205.125 , 3.869075458280334e-47 ], [ 205.1875 , 3.869400175287232e-47 ], 
                    [ 205.25 , 3.8697266589632143e-47 ], [ 205.3125 , 3.870053170186514e-47 ], [ 205.375 , 3.8703797089594264e-47 ], [ 205.4375 , 3.8707062752843034e-47 ], 
                    [ 205.5 , 3.8710328691634986e-47 ], [ 205.5625 , 3.8713594905993075e-47 ], [ 205.625 , 3.8716861395940575e-47 ], [ 205.6875 , 3.872013828525886e-47 ], 
                    [ 205.75 , 3.872342274990033e-47 ], [ 205.8125 , 3.872670749314842e-47 ], [ 205.875 , 3.8729992515027584e-47 ], [ 205.9375 , 3.873327781556145e-47 ], 
                    [ 206.0 , 3.8736563394772837e-47 ], [ 206.0625 , 3.873984925268594e-47 ], [ 206.125 , 3.874313756186734e-47 ], [ 206.1875 , 3.8746441659140274e-47 ], 
                    [ 206.25 , 3.8749746038193665e-47 ], [ 206.3125 , 3.8753050699051815e-47 ], [ 206.375 , 3.875635564173765e-47 ], [ 206.4375 , 3.8759660866276864e-47 ], 
                    [ 206.5 , 3.876296637269266e-47 ], [ 206.5625 , 3.8766272161008813e-47 ], [ 206.625 , 3.876958942548412e-47 ], [ 206.6875 , 3.8772913059428447e-47 ], 
                    [ 206.75 , 3.8776236978300524e-47 ], [ 206.8125 , 3.8779561182125063e-47 ], [ 206.875 , 3.8782885670926476e-47 ], [ 206.9375 , 3.8786210444729197e-47 ], 
                    [ 207.0 , 3.8789535503557944e-47 ], [ 207.0625 , 3.8792863571219595e-47 ], [ 207.125 , 3.8796206697948897e-47 ], [ 207.1875 , 3.8799550112784937e-47 ], 
                    [ 207.25 , 3.880289381575311e-47 ], [ 207.3125 , 3.8806237806877956e-47 ], [ 207.375 , 3.8809582086183765e-47 ], [ 207.4375 , 3.8812926653696474e-47 ], 
                    [ 207.5 , 3.881627150944009e-47 ], [ 207.5625 , 3.8819627804702937e-47 ], [ 207.625 , 3.882299048223623e-47 ], [ 207.6875 , 3.88263534510549e-47 ], 
                    [ 207.75 , 3.8829716711184437e-47 ], [ 207.8125 , 3.883308026265064e-47 ], [ 207.875 , 3.8836444105477634e-47 ], [ 207.9375 , 3.8839808239691487e-47 ], 
                    [ 208.0 , 3.884317472783266e-47 ], [ 208.0625 , 3.8846556543011056e-47 ], [ 208.125 , 3.884993865262118e-47 ], [ 208.1875 , 3.8853321056688943e-47 ], 
                    [ 208.25 , 3.885670375524026e-47 ], [ 208.3125 , 3.8860086748300777e-47 ], [ 208.375 , 3.8863470035895293e-47 ], [ 208.4375 , 3.886685361805001e-47 ], 
                    [ 208.5 , 3.8870247494560963e-47 ], [ 208.5625 , 3.887364898571655e-47 ], [ 208.625 , 3.8877050774533035e-47 ], [ 208.6875 , 3.8880452861035637e-47 ], 
                    [ 208.75 , 3.888385524525096e-47 ], [ 208.8125 , 3.8887257927205613e-47 ], [ 208.875 , 3.8890660906924813e-47 ], [ 208.9375 , 3.8894064479241086e-47 ], 
                    [ 209.0 , 3.8897485116628123e-47 ], [ 209.0625 , 3.8900906054852374e-47 ], [ 209.125 , 3.890432729393945e-47 ], [ 209.1875 , 3.8907748833916096e-47 ], 
                    [ 209.25 , 3.8911170674809055e-47 ], [ 209.3125 , 3.891459281664451e-47 ], [ 209.375 , 3.8918015259448095e-47 ], [ 209.4375 , 3.8921445446919574e-47 ], 
                    [ 209.5 , 3.892488541132544e-47 ], [ 209.5625 , 3.892832567976194e-47 ], [ 209.625 , 3.8931766252257623e-47 ], [ 209.6875 , 3.893520712883907e-47 ], 
                    [ 209.75 , 3.893864830953288e-47 ], [ 209.8125 , 3.894208979436567e-47 ], [ 209.875 , 3.894553158336486e-47 ], [ 209.9375 , 3.8948988042573545e-47 ], 
                    [ 210.0 , 3.895244743511945e-47 ], [ 210.0625 , 3.895590713492386e-47 ], [ 210.125 , 3.8959367142013767e-47 ], [ 210.1875 , 3.8962827456416487e-47 ], 
                    [ 210.25 , 3.8966288078158476e-47 ], [ 210.3125 , 3.896974900726897e-47 ], [ 210.375 , 3.8973214187229407e-47 ], [ 210.4375 , 3.8976692614242897e-47 ], 
                    [ 210.5 , 3.898017135171259e-47 ], [ 210.5625 , 3.8983650399664807e-47 ], [ 210.625 , 3.898712975812837e-47 ], [ 210.6875 , 3.899060942713044e-47 ], 
                    [ 210.75 , 3.899408940669956e-47 ], [ 210.8125 , 3.8997569696862617e-47 ], [ 210.875 , 3.9001060406546863e-47 ], [ 210.9375 , 3.900455803925448e-47 ], 
                    [ 211.0 , 3.900805598563165e-47 ], [ 211.0625 , 3.9011554245706236e-47 ], [ 211.125 , 3.9015052819506085e-47 ], [ 211.1875 , 3.901855170706045e-47 ], 
                    [ 211.25 , 3.9022050908395515e-47 ], [ 211.3125 , 3.9025550423541375e-47 ], [ 211.375 , 3.90290662959046e-47 ], [ 211.4375 , 3.9032583181454403e-47 ], 
                    [ 211.5 , 3.903610038390947e-47 ], [ 211.5625 , 3.903961790329752e-47 ], [ 211.625 , 3.904313573964628e-47 ], [ 211.6875 , 3.9046653892985153e-47 ], 
                    [ 211.75 , 3.9050172363343794e-47 ], [ 211.8125 , 3.9053696073343882e-47 ], [ 211.875 , 3.9057231882396537e-47 ], [ 211.9375 , 3.9060768011570864e-47 ], 
                    [ 212.0 , 3.9064304460895865e-47 ], [ 212.0625 , 3.9067841230400787e-47 ], [ 212.125 , 3.9071378320114903e-47 ], [ 212.1875 , 3.9074915730066644e-47 ], 
                    [ 212.25 , 3.907845346028557e-47 ], [ 212.3125 , 3.9082001662201093e-47 ], [ 212.375 , 3.908555657515387e-47 ], [ 212.4375 , 3.9089111811462797e-47 ], 
                    [ 212.5 , 3.9092667371157025e-47 ], [ 212.5625 , 3.9096223254266525e-47 ], [ 212.625 , 3.909977946082071e-47 ], [ 212.6875 , 3.9103335990848174e-47 ], 
                    [ 212.75 , 3.910689284437916e-47 ], [ 212.8125 , 3.9110465097327943e-47 ], [ 212.875 , 3.911403907072718e-47 ], [ 212.9375 , 3.91176133707215e-47 ], 
                    [ 213.0 , 3.9121187997339925e-47 ], [ 213.0625 , 3.912476295061397e-47 ], [ 213.125 , 3.9128338230572364e-47 ], [ 213.1875 , 3.9131913837245246e-47 ], 
                    [ 213.25 , 3.9135493071244483e-47 ], [ 213.3125 , 3.9139085875865573e-47 ], [ 213.375 , 3.9142679010321403e-47 ], [ 213.4375 , 3.914627247464224e-47 ], 
                    [ 213.5 , 3.914986626885811e-47 ], [ 213.5625 , 3.9153460393000117e-47 ], [ 213.625 , 3.9157054847098e-47 ], [ 213.6875 , 3.9160649631181487e-47 ], 
                    [ 213.75 , 3.916425237292712e-47 ], [ 213.8125 , 3.9167864179193075e-47 ], [ 213.875 , 3.91714763185468e-47 ], [ 213.9375 , 3.917508879101985e-47 ], 
                    [ 214.0 , 3.9178701596641836e-47 ], [ 214.0625 , 3.9182314735444314e-47 ], [ 214.125 , 3.918592820745801e-47 ], [ 214.1875 , 3.91895420127131e-47 ], 
                    [ 214.25 , 3.91931678045695e-47 ], [ 214.3125 , 3.9196798569143103e-47 ], [ 214.375 , 3.9200429670062353e-47 ], [ 214.4375 , 3.9204061107358974e-47 ], 
                    [ 214.5 , 3.9207692881062997e-47 ], [ 214.5625 , 3.9211324991205876e-47 ], [ 214.625 , 3.92149574378196e-47 ], [ 214.6875 , 3.921859022093481e-47 ], 
                    [ 214.75 , 3.9222238681602613e-47 ], [ 214.8125 , 3.922588834239483e-47 ], [ 214.875 , 3.922953834279064e-47 ], [ 214.9375 , 3.923318868282219e-47 ], 
                    [ 215.0 , 3.923683936252081e-47 ], [ 215.0625 , 3.924049038191811e-47 ], [ 215.125 , 3.9244141741045415e-47 ], [ 215.1875 , 3.9247796042221116e-47 ], 
                    [ 215.25 , 3.9251464266027005e-47 ], [ 215.3125 , 3.9255132832677027e-47 ], [ 215.375 , 3.925880174220324e-47 ], [ 215.4375 , 3.926247099463712e-47 ], 
                    [ 215.5 , 3.9266140590010993e-47 ], [ 215.5625 , 3.926981052835721e-47 ], [ 215.625 , 3.927348080970696e-47 ], [ 215.6875 , 3.927715723969089e-47 ], 
                    [ 215.75 , 3.928084446592548e-47 ], [ 215.8125 , 3.928453203830568e-47 ], [ 215.875 , 3.9288219956864823e-47 ], [ 215.9375 , 3.92919082216354e-47 ], 
                    [ 216.0 , 3.9295596832649637e-47 ], [ 216.0625 , 3.929928578993976e-47 ], [ 216.125 , 3.9302975093539103e-47 ], [ 216.1875 , 3.9306673201743445e-47 ], 
                    [ 216.25 , 3.9310378985394154e-47 ], [ 216.3125 , 3.9314085118421757e-47 ], [ 216.375 , 3.931779160085836e-47 ], [ 216.4375 , 3.9321498432737456e-47 ], 
                    [ 216.5 , 3.9325205614092837e-47 ], [ 216.5625 , 3.932891314495576e-47 ], [ 216.625 , 3.933262102536059e-47 ], [ 216.6875 , 3.9336340583063626e-47 ], 
                    [ 216.75 , 3.9340065425769843e-47 ], [ 216.8125 , 3.9343790621189705e-47 ], [ 216.875 , 3.9347516169356624e-47 ], [ 216.9375 , 3.9351242070303147e-47 ], 
                    [ 217.0 , 3.9354968324063245e-47 ], [ 217.0625 , 3.935869493067061e-47 ], [ 217.125 , 3.936242189015837e-47 ], [ 217.1875 , 3.936616237150634e-47 ], 
                    [ 217.25 , 3.936990571989682e-47 ], [ 217.3125 , 3.9373649424244477e-47 ], [ 217.375 , 3.9377393484582326e-47 ], [ 217.4375 , 3.9381137900944777e-47 ], 
                    [ 217.5 , 3.938488267336624e-47 ], [ 217.5625 , 3.938862780187946e-47 ], [ 217.625 , 3.9392373286519133e-47 ], [ 217.6875 , 3.939613464578391e-47 ], 
                    [ 217.75 , 3.9399896925099505e-47 ], [ 217.8125 , 3.9403659563707855e-47 ], [ 217.875 , 3.940742256164355e-47 ], [ 217.9375 , 3.941118591893978e-47 ], 
                    [ 218.0 , 3.941494963563283e-47 ], [ 218.0625 , 3.94187137117559e-47 ], [ 218.125 , 3.942247928654359e-47 ], [ 218.1875 , 3.9426259766072453e-47 ], 
                    [ 218.25 , 3.9430040608135947e-47 ], [ 218.3125 , 3.943382181276942e-47 ], [ 218.375 , 3.9437603380006506e-47 ], [ 218.4375 , 3.9441385309883666e-47 ], 
                    [ 218.5 , 3.9445167602434826e-47 ], [ 218.5625 , 3.9448950257694776e-47 ], [ 218.625 , 3.945273581635425e-47 ], [ 218.6875 , 3.9456534858751584e-47 ], 
                    [ 218.75 , 3.946033426697149e-47 ], [ 218.8125 , 3.946413404104974e-47 ], [ 218.875 , 3.946793418102185e-47 ], [ 218.9375 , 3.947173468692306e-47 ], 
                    [ 219.0 , 3.947553555878776e-47 ], [ 219.0625 , 3.9479336796652315e-47 ], [ 219.125 , 3.9483142081046953e-47 ], [ 219.1875 , 3.948695979765213e-47 ], 
                    [ 219.25 , 3.949077788340174e-47 ], [ 219.3125 , 3.94945963383312e-47 ], [ 219.375 , 3.949841516247536e-47 ], [ 219.4375 , 3.9502234355870777e-47 ], 
                    [ 219.5 , 3.950605391855286e-47 ], [ 219.5625 , 3.9509873850557596e-47 ], [ 219.625 , 3.9513698655331066e-47 ], [ 219.6875 , 3.9517534990990847e-47 ], 
                    [ 219.75 , 3.9521371699115673e-47 ], [ 219.8125 , 3.952520877974198e-47 ], [ 219.875 , 3.952904623290565e-47 ], [ 219.9375 , 3.953288405864286e-47 ], 
                    [ 220.0 , 3.953672225698949e-47 ], [ 220.0625 , 3.954056082798229e-47 ], [ 220.125 , 3.954440474829323e-47 ], [ 220.1875 , 3.954825946467792e-47 ], 
                    [ 220.25 , 3.955211455681305e-47 ], [ 220.3125 , 3.955597002473524e-47 ], [ 220.375 , 3.955982586848197e-47 ], [ 220.4375 , 3.956368208808874e-47 ], 
                    [ 220.5 , 3.9567538683593324e-47 ], [ 220.5625 , 3.95713956550318e-47 ], [ 220.625 , 3.957525831113486e-47 ], [ 220.6875 , 3.9579131613745655e-47 ], 
                    [ 220.75 , 3.9583005295443904e-47 ], [ 220.8125 , 3.958687935626616e-47 ], [ 220.875 , 3.959075379625009e-47 ], [ 220.9375 , 3.959462861543223e-47 ], 
                    [ 221.0 , 3.959850381384998e-47 ], [ 221.0625 , 3.9602379391540457e-47 ], [ 221.125 , 3.960626059001827e-47 ], [ 221.1875 , 3.96101523300054e-47 ], 
                    [ 221.25 , 3.9614044452397443e-47 ], [ 221.3125 , 3.961793695723199e-47 ], [ 221.375 , 3.962182984454717e-47 ], [ 221.4375 , 3.962572311438029e-47 ], 
                    [ 221.5 , 3.9629616766768944e-47 ], [ 221.5625 , 3.963351080175071e-47 ], [ 221.625 , 3.963741016691137e-47 ], [ 221.6875 , 3.964132034570939e-47 ], 
                    [ 221.75 , 3.964523091024145e-47 ], [ 221.8125 , 3.964914186054587e-47 ], [ 221.875 , 3.9653053196660156e-47 ], [ 221.9375 , 3.965696491862265e-47 ], 
                    [ 222.0 , 3.966087702647169e-47 ], [ 222.0625 , 3.9664789520244785e-47 ], [ 222.125 , 3.9668706717328946e-47 ], [ 222.1875 , 3.9672635135706825e-47 ], 
                    [ 222.25 , 3.9676563943118867e-47 ], [ 222.3125 , 3.968049313960303e-47 ], [ 222.375 , 3.9684422725198136e-47 ], [ 222.4375 , 3.9688352699943288e-47 ], 
                    [ 222.5 , 3.969228306387587e-47 ], [ 222.5625 , 3.969621381703529e-47 ], [ 222.625 , 3.9700148479496617e-47 ], [ 222.6875 , 3.970409532292881e-47 ], 
                    [ 222.75 , 3.970804255874116e-47 ], [ 222.8125 , 3.9711990186973255e-47 ], [ 222.875 , 3.971593820766409e-47 ], [ 222.9375 , 3.971988662085382e-47 ], 
                    [ 223.0 , 3.972383542657921e-47 ], [ 223.0625 , 3.9727784624880124e-47 ], [ 223.125 , 3.9731736611191383e-47 ], [ 223.1875 , 3.973570188887985e-47 ], 
                    [ 223.25 , 3.973966756230806e-47 ], [ 223.3125 , 3.974363363151579e-47 ], [ 223.375 , 3.9747600096542254e-47 ], [ 223.4375 , 3.9751566957426676e-47 ], 
                    [ 223.5 , 3.9755534214209126e-47 ], [ 223.5625 , 3.9759501866928268e-47 ], [ 223.625 , 3.976347089500909e-47 ], [ 223.6875 , 3.976745438100756e-47 ], 
                    [ 223.75 , 3.977143826607037e-47 ], [ 223.8125 , 3.977542255023665e-47 ], [ 223.875 , 3.9779407233546654e-47 ], [ 223.9375 , 3.9783392316040667e-47 ], 
                    [ 224.0 , 3.978737779775838e-47 ], [ 224.0625 , 3.9791363678738957e-47 ], [ 224.125 , 3.9795349959024366e-47 ], [ 224.1875 , 3.979935100725916e-47 ], 
                    [ 224.25 , 3.980335313173468e-47 ], [ 224.3125 , 3.9807355658653966e-47 ], [ 224.375 , 3.9811358588056903e-47 ], [ 224.4375 , 3.9815361919985675e-47 ], 
                    [ 224.5 , 3.9819365654478203e-47 ], [ 224.5625 , 3.982336979157724e-47 ], [ 224.625 , 3.982737433132241e-47 ], [ 224.6875 , 3.983139164094908e-47 ], 
                    [ 224.75 , 3.983541194531762e-47 ], [ 224.8125 , 3.9839432655467786e-47 ], [ 224.875 , 3.9843453771440835e-47 ], [ 224.9375 , 3.9847475293276863e-47 ], 
                    [ 225.0 , 3.985149722101826e-47 ], [ 225.0625 , 3.9855519554704017e-47 ], [ 225.125 , 3.985954229437679e-47 ], [ 225.1875 , 3.986357571969677e-47 ], 
                    [ 225.25 , 3.986761438756455e-47 ], [ 225.3125 , 3.9871653464598793e-47 ], [ 225.375 , 3.9875692950841513e-47 ], [ 225.4375 , 3.987973284633275e-47 ], 
                    [ 225.5 , 3.9883773151115397e-47 ], [ 225.5625 , 3.9887813865230056e-47 ], [ 225.625 , 3.9891854988717924e-47 ], [ 225.6875 , 3.9895904095551217e-47 ], 
                    [ 225.75 , 3.989996074568915e-47 ], [ 225.8125 , 3.990401780831079e-47 ], [ 225.875 , 3.9908075283458364e-47 ], [ 225.9375 , 3.9912133171172677e-47 ], 
                    [ 226.0 , 3.991619147149738e-47 ], [ 226.0625 , 3.992025018447359e-47 ], [ 226.125 , 3.992430931014297e-47 ], [ 226.1875 , 3.9928373795951344e-47 ], 
                    [ 226.25 , 3.99324486797596e-47 ], [ 226.3125 , 3.993652397942919e-47 ], [ 226.375 , 3.9940599695003116e-47 ], [ 226.4375 , 3.994467582652355e-47 ], 
                    [ 226.5 , 3.994875237403265e-47 ], [ 226.5625 , 3.995282933757344e-47 ], [ 226.625 , 3.9956906717187803e-47 ], [ 226.6875 , 3.9960986409631497e-47 ], 
                    [ 226.75 , 3.9965079312751554e-47 ], [ 226.8125 , 3.99691726350766e-47 ], [ 226.875 , 3.997326637665013e-47 ], [ 226.9375 , 3.997736053751452e-47 ], 
                    [ 227.0 , 3.998145511771356e-47 ], [ 227.0625 , 3.9985550117289074e-47 ], [ 227.125 , 3.998964553628486e-47 ], [ 227.1875 , 3.9993741374744165e-47 ], 
                    [ 227.25 , 3.9997851045926306e-47 ], [ 227.3125 , 4.0001962493392305e-47 ], [ 227.375 , 4.000607436348158e-47 ], [ 227.4375 , 4.0010186656237014e-47 ], 
                    [ 227.5 , 4.0014299371702043e-47 ], [ 227.5625 , 4.001841250991984e-47 ], [ 227.625 , 4.002252607093444e-47 ], [ 227.6875 , 4.002664005478871e-47 ], 
                    [ 227.75 , 4.00307642304635e-47 ], [ 227.8125 , 4.0034893669302624e-47 ], [ 227.875 , 4.003902353412103e-47 ], [ 227.9375 , 4.0043153824961534e-47 ], 
                    [ 228.0 , 4.004728454187007e-47 ], [ 228.0625 , 4.0051415684888595e-47 ], [ 228.125 , 4.0055547254062215e-47 ], [ 228.1875 , 4.0059679249434026e-47 ], 
                    [ 228.25 , 4.0063817660768624e-47 ], [ 228.3125 , 4.0067965094257333e-47 ], [ 228.375 , 4.007211295709145e-47 ], [ 228.4375 , 4.0076261249314556e-47 ], 
                    [ 228.5 , 4.008040997097225e-47 ], [ 228.5625 , 4.008455912210841e-47 ], [ 228.625 , 4.0088708702766937e-47 ], [ 228.6875 , 4.009285871299371e-47 ], 
                    [ 228.75 , 4.009701111351028e-47 ], [ 228.8125 , 4.0101176642910455e-47 ], [ 228.875 , 4.0105342605052275e-47 ], [ 228.9375 , 4.0109508999980416e-47 ], 
                    [ 229.0 , 4.0113675827739553e-47 ], [ 229.0625 , 4.0117843088374924e-47 ], [ 229.125 , 4.0122010781932657e-47 ], [ 229.1875 , 4.012617890845486e-47 ], 
                    [ 229.25 , 4.013034746798909e-47 ], [ 229.3125 , 4.0134528542684253e-47 ], [ 229.375 , 4.0138712384790666e-47 ], [ 229.4375 , 4.014289666304359e-47 ], 
                    [ 229.5 , 4.0147081377488504e-47 ], [ 229.5625 , 4.0151266528171436e-47 ], [ 229.625 , 4.0155452115136443e-47 ], [ 229.6875 , 4.0159638138430433e-47 ], 
                    [ 229.75 , 4.01638245980986e-47 ], [ 229.8125 , 4.016801900951818e-47 ], [ 229.875 , 4.017222069097238e-47 ], [ 229.9375 , 4.017642281193306e-47 ], 
                    [ 230.0 , 4.0180625372447313e-47 ], [ 230.0625 , 4.0184828372559986e-47 ], [ 230.125 , 4.018903181231763e-47 ], [ 230.1875 , 4.0193235691767377e-47 ], 
                    [ 230.25 , 4.019744001095323e-47 ], [ 230.3125 , 4.0201647542451433e-47 ], [ 230.375 , 4.0205867292443025e-47 ], [ 230.4375 , 4.0210087485358715e-47 ], 
                    [ 230.5 , 4.021430812124528e-47 ], [ 230.5625 , 4.021852920014893e-47 ], [ 230.625 , 4.0222750722117023e-47 ], [ 230.6875 , 4.022697268719464e-47 ], 
                    [ 230.75 , 4.023119509543029e-47 ], [ 230.8125 , 4.023541794686906e-47 ], [ 230.875 , 4.0239653116816656e-47 ], [ 230.9375 , 4.0243891021732187e-47 ], 
                    [ 231.0 , 4.024812937296989e-47 ], [ 231.0625 , 4.025236817057621e-47 ], [ 231.125 , 4.0256607414598715e-47 ], [ 231.1875 , 4.026084710508386e-47 ], 
                    [ 231.25 , 4.026508724207894e-47 ], [ 231.3125 , 4.026932782563099e-47 ], [ 231.375 , 4.0273575540951614e-47 ], [ 231.4375 , 4.0277831328973457e-47 ], 
                    [ 231.5 , 4.028208756671223e-47 ], [ 231.5625 , 4.0286344254216304e-47 ], [ 231.625 , 4.0290601391532647e-47 ], [ 231.6875 , 4.0294858978709067e-47 ], 
                    [ 231.75 , 4.0299117015793113e-47 ], [ 231.8125 , 4.0303375502832317e-47 ], [ 231.875 , 4.030763553346159e-47 ], [ 231.9375 , 4.031190907313603e-47 ], 
                    [ 232.0 , 4.031618306590431e-47 ], [ 232.0625 , 4.032045751181476e-47 ], [ 232.125 , 4.0324732410914834e-47 ], [ 232.1875 , 4.032900776325316e-47 ], 
                    [ 232.25 , 4.0333283568877235e-47 ], [ 232.3125 , 4.0337559827835394e-47 ], [ 232.375 , 4.0341836540175983e-47 ], [ 232.4375 , 4.03461232489012e-47 ], 
                    [ 232.5 , 4.0350415153119955e-47 ], [ 232.5625 , 4.0354707513898523e-47 ], [ 232.625 , 4.035900033128634e-47 ], [ 232.6875 , 4.036329360533198e-47 ], 
                    [ 232.75 , 4.0367587336083143e-47 ], [ 232.8125 , 4.037188152359015e-47 ], [ 232.875 , 4.0376176167899583e-47 ], [ 232.9375 , 4.0380474644324415e-47 ], 
                    [ 233.0 , 4.0384784267842853e-47 ], [ 233.0625 , 4.038909435130742e-47 ], [ 233.125 , 4.039340489476778e-47 ], [ 233.1875 , 4.0397715898272465e-47 ], 
                    [ 233.25 , 4.040202736187142e-47 ], [ 233.3125 , 4.0406339285612613e-47 ], [ 233.375 , 4.0410651669546003e-47 ], [ 233.4375 , 4.0414964513721006e-47 ], 
                    [ 233.5 , 4.0419288876002344e-47 ], [ 233.5625 , 4.042361659128125e-47 ], [ 233.625 , 4.0427944769931852e-47 ], [ 233.6875 , 4.043227341200175e-47 ], 
                    [ 233.75 , 4.0436602517542285e-47 ], [ 233.8125 , 4.0440932086602517e-47 ], [ 233.875 , 4.0445262119231764e-47 ], [ 233.9375 , 4.044959261548026e-47 ], 
                    [ 234.0 , 4.045392818305986e-47 ], [ 234.0625 , 4.045827372531584e-47 ], [ 234.125 , 4.0462619734369114e-47 ], [ 234.1875 , 4.04669662102678e-47 ], 
                    [ 234.25 , 4.0471313153062925e-47 ], [ 234.3125 , 4.047566056280549e-47 ], [ 234.375 , 4.048000843954394e-47 ], [ 234.4375 , 4.048435678332901e-47 ], 
                    [ 234.5 , 4.0488705594212024e-47 ], [ 234.5625 , 4.04930666002743e-47 ], [ 234.625 , 4.049743029692344e-47 ], [ 234.6875 , 4.050179446382161e-47 ], 
                    [ 234.75 , 4.050615910102034e-47 ], [ 234.8125 , 4.0510524208570324e-47 ], [ 234.875 , 4.051488978652109e-47 ], [ 234.9375 , 4.051925583492534e-47 ], 
                    [ 235.0 , 4.0523622353831767e-47 ], [ 235.0625 , 4.052799400167723e-47 ], [ 235.125 , 4.05323753279818e-47 ], [ 235.1875 , 4.0536757127934505e-47 ], 
                    [ 235.25 , 4.0541139401587403e-47 ], [ 235.3125 , 4.0545522148990554e-47 ], [ 235.375 , 4.054990537019604e-47 ], [ 235.4375 , 4.055428906525479e-47 ], 
                    [ 235.5 , 4.055867323421833e-47 ], [ 235.5625 , 4.056305787713702e-47 ], [ 235.625 , 4.0567454226605906e-47 ], [ 235.6875 , 4.0571853653484524e-47 ], 
                    [ 235.75 , 4.057625355746839e-47 ], [ 235.8125 , 4.05806539386101e-47 ], [ 235.875 , 4.058505479696055e-47 ], [ 235.9375 , 4.058945613257234e-47 ], 
                    [ 236.0 , 4.05938579454958e-47 ], [ 236.0625 , 4.0598260235783547e-47 ], [ 236.125 , 4.060266668091476e-47 ], [ 236.1875 , 4.060708362518456e-47 ], 
                    [ 236.25 , 4.0611501049949516e-47 ], [ 236.3125 , 4.06159189552619e-47 ], [ 236.375 , 4.0620337341174577e-47 ], [ 236.4375 , 4.0624756207739826e-47 ], 
                    [ 236.5 , 4.0629175555009353e-47 ], [ 236.5625 , 4.0633595383035164e-47 ], [ 236.625 , 4.0638015691871e-47 ], [ 236.6875 , 4.0642446244217496e-47 ], 
                    [ 236.75 , 4.0646881350217446e-47 ], [ 236.8125 , 4.0651316940198237e-47 ], [ 236.875 , 4.0655753014212386e-47 ], [ 236.9375 , 4.066018957231301e-47 ], 
                    [ 237.0 , 4.066462661455293e-47 ], [ 237.0625 , 4.066906414098528e-47 ], [ 237.125 , 4.067350215166288e-47 ], [ 237.1875 , 4.067794232621694e-47 ], 
                    [ 237.25 , 4.068239490484231e-47 ], [ 237.3125 , 4.0686847970843755e-47 ], [ 237.375 , 4.069130152427522e-47 ], [ 237.4375 , 4.069575556518919e-47 ], 
                    [ 237.5 , 4.070021009363872e-47 ], [ 237.5625 , 4.070466510967893e-47 ], [ 237.625 , 4.0709120613361156e-47 ], [ 237.6875 , 4.0713576604739936e-47 ], 
                    [ 237.75 , 4.0718040250197695e-47 ], [ 237.8125 , 4.0722510950125265e-47 ], [ 237.875 , 4.0726982140919924e-47 ], [ 237.9375 , 4.0731453822636444e-47 ], 
                    [ 238.0 , 4.0735925995326983e-47 ], [ 238.0625 , 4.074039865904719e-47 ], [ 238.125 , 4.0744871813850393e-47 ], [ 238.1875 , 4.0749345459790524e-47 ], 
                    [ 238.25 , 4.075381959692178e-47 ], [ 238.3125 , 4.0758306384868403e-47 ], [ 238.375 , 4.0762795005999264e-47 ], [ 238.4375 , 4.076728412145222e-47 ], 
                    [ 238.5 , 4.077177373128142e-47 ], [ 238.5625 , 4.07762638355416e-47 ], [ 238.625 , 4.078075443428605e-47 ], [ 238.6875 , 4.0785245527570975e-47 ], 
                    [ 238.75 , 4.0789737115449664e-47 ], [ 238.8125 , 4.0794232722006833e-47 ], [ 238.875 , 4.0798738846746143e-47 ], [ 238.9375 , 4.080324546923131e-47 ], 
                    [ 239.0 , 4.080775258951789e-47 ], [ 239.0625 , 4.081226020766e-47 ], [ 239.125 , 4.081676832371235e-47 ], [ 239.1875 , 4.082127693773109e-47 ], 
                    [ 239.25 , 4.082578604977093e-47 ], [ 239.3125 , 4.083029565988603e-47 ], [ 239.375 , 4.0834813864879186e-47 ], [ 239.4375 , 4.083933797187851e-47 ], 
                    [ 239.5 , 4.0843862580105353e-47 ], [ 239.5625 , 4.084838768961524e-47 ], [ 239.625 , 4.085291330046458e-47 ], [ 239.6875 , 4.0857439412708047e-47 ], 
                    [ 239.75 , 4.0861966026401485e-47 ], [ 239.8125 , 4.0866493141601024e-47 ], [ 239.875 , 4.0871020758360776e-47 ], [ 239.9375 , 4.087556126250205e-47 ], 
                    [ 240.0 , 4.088010334820659e-47 ], [ 240.0625 , 4.088464593862666e-47 ], [ 240.125 , 4.0889189033818636e-47 ], [ 240.1875 , 4.089373263383919e-47 ], 
                    [ 240.25 , 4.0898276738742967e-47 ], [ 240.3125 , 4.09028213485881e-47 ], [ 240.375 , 4.0907366463428383e-47 ], [ 240.4375 , 4.091191501071347e-47 ], 
                    [ 240.5 , 4.0916474414874652e-47 ], [ 240.5625 , 4.092103432715681e-47 ], [ 240.625 , 4.092559474761482e-47 ], [ 240.6875 , 4.0930155676306777e-47 ], 
                    [ 240.75 , 4.0934717113288725e-47 ], [ 240.8125 , 4.093927905861674e-47 ], [ 240.875 , 4.094384151234892e-47 ], [ 240.9375 , 4.094840447454019e-47 ], 
                    [ 241.0 , 4.0952974616980375e-47 ], [ 241.0625 , 4.0957551912134123e-47 ], [ 241.125 , 4.0962129718889463e-47 ], [ 241.1875 , 4.096670803730416e-47 ], 
                    [ 241.25 , 4.0971286867435697e-47 ], [ 241.3125 , 4.097586620934155e-47 ], [ 241.375 , 4.0980446063077185e-47 ], [ 241.4375 , 4.098502642870125e-47 ], 
                    [ 241.5 , 4.098960730627097e-47 ], [ 241.5625 , 4.0994198863643864e-47 ], [ 241.625 , 4.099879411007735e-47 ], [ 241.6875 , 4.1003389871614883e-47 ], 
                    [ 241.75 , 4.1007986148314203e-47 ], [ 241.8125 , 4.101258294023335e-47 ], [ 241.875 , 4.1017180247430086e-47 ], [ 241.9375 , 4.1021778069963034e-47 ], 
                    [ 242.0 , 4.1026376407888217e-47 ], [ 242.0625 , 4.1030975303537965e-47 ], [ 242.125 , 4.103558788439195e-47 ], [ 242.1875 , 4.104020098377834e-47 ], 
                    [ 242.25 , 4.104481460175601e-47 ], [ 242.3125 , 4.104942873838238e-47 ], [ 242.375 , 4.1054043393716915e-47 ], [ 242.4375 , 4.105865856781736e-47 ], 
                    [ 242.5 , 4.1063274260741726e-47 ], [ 242.5625 , 4.106789047254922e-47 ], [ 242.625 , 4.107251016091273e-47 ], [ 242.6875 , 4.1077140587885265e-47 ], 
                    [ 242.75 , 4.108177153688224e-47 ], [ 242.8125 , 4.1086403007962504e-47 ], [ 242.875 , 4.1091035001185494e-47 ], [ 242.9375 , 4.109566751660921e-47 ], 
                    [ 243.0 , 4.11003005542928e-47 ], [ 243.0625 , 4.1104934114294865e-47 ], [ 243.125 , 4.110956819667487e-47 ], [ 243.1875 , 4.1114208377122983e-47 ], 
                    [ 243.25 , 4.1118856597337746e-47 ], [ 243.3125 , 4.1123505343063964e-47 ], [ 243.375 , 4.112815461435957e-47 ], [ 243.4375 , 4.1132804411284595e-47 ], 
                    [ 243.5 , 4.113745473389844e-47 ], [ 243.5625 , 4.1142105582261134e-47 ], [ 243.625 , 4.114675695643065e-47 ], [ 243.6875 , 4.1151408856467607e-47 ], 
                    [ 243.75 , 4.115606922232646e-47 ], [ 243.8125 , 4.1160735249454492e-47 ], [ 243.875 , 4.1165401805588276e-47 ], [ 243.9375 , 4.1170068890788063e-47 ], 
                    [ 244.0 , 4.117473650511386e-47 ], [ 244.0625 , 4.117940464862622e-47 ], [ 244.125 , 4.118407332138369e-47 ], [ 244.1875 , 4.118874252344802e-47 ], 
                    [ 244.25 , 4.1193412254877174e-47 ], [ 244.3125 , 4.1198092578846022e-47 ], [ 244.375 , 4.12027764611926e-47 ], [ 244.4375 , 4.1207460876058145e-47 ], 
                    [ 244.5 , 4.1212145823502325e-47 ], [ 244.5625 , 4.121683130358687e-47 ], [ 244.625 , 4.122151731637204e-47 ], [ 244.6875 , 4.1226203861917506e-47 ], 
                    [ 244.75 , 4.1230890940284435e-47 ], [ 244.8125 , 4.1235578551533997e-47 ], [ 244.875 , 4.1240278519388845e-47 ], [ 244.9375 , 4.1244980189018544e-47 ], 
                    [ 245.0 , 4.1249682394670843e-47 ], [ 245.0625 , 4.125438513640628e-47 ], [ 245.125 , 4.1259088414285955e-47 ], [ 245.1875 , 4.1263792228370706e-47 ], 
                    [ 245.25 , 4.1268496578722256e-47 ], [ 245.3125 , 4.1273201465401145e-47 ], [ 245.375 , 4.1277907294646686e-47 ], [ 245.4375 , 4.1282626076940195e-47 ], 
                    [ 245.5 , 4.128734539867225e-47 ], [ 245.5625 , 4.1292065259905394e-47 ], [ 245.625 , 4.129678566070013e-47 ], [ 245.6875 , 4.130150660111873e-47 ], 
                    [ 245.75 , 4.1306228081222874e-47 ], [ 245.8125 , 4.1310950101074853e-47 ], [ 245.875 , 4.1315672660735485e-47 ], [ 245.9375 , 4.1320397470597657e-47 ], 
                    [ 246.0 , 4.132513397522855e-47 ], [ 246.0625 , 4.1329871022798936e-47 ], [ 246.125 , 4.1334608613371356e-47 ], [ 246.1875 , 4.133934674700747e-47 ], 
                    [ 246.25 , 4.1344085423770114e-47 ], [ 246.3125 , 4.1348824643721246e-47 ], [ 246.375 , 4.1353564406923434e-47 ], [ 246.4375 , 4.135830471343777e-47 ], 
                    [ 246.5 , 4.1363048319283503e-47 ], [ 246.5625 , 4.1367802597520004e-47 ], [ 246.625 , 4.1372557422214043e-47 ], [ 246.6875 , 4.1377312793429006e-47 ], 
                    [ 246.75 , 4.138206871122714e-47 ], [ 246.8125 , 4.1386825175672134e-47 ], [ 246.875 , 4.139158218682507e-47 ], [ 246.9375 , 4.1396339744750535e-47 ], 
                    [ 247.0 , 4.14010978495111e-47 ], [ 247.0625 , 4.1405860035325354e-47 ], [ 247.125 , 4.1410632100406994e-47 ], [ 247.1875 , 4.1415404715473124e-47 ], 
                    [ 247.25 , 4.1420177880588603e-47 ], [ 247.3125 , 4.1424951595815347e-47 ], [ 247.375 , 4.1429725861217354e-47 ], [ 247.4375 , 4.143450067685773e-47 ], 
                    [ 247.5 , 4.143927604280107e-47 ], [ 247.5625 , 4.144405195910903e-47 ], [ 247.625 , 4.144883240992625e-47 ], [ 247.6875 , 4.1453622062506975e-47 ], 
                    [ 247.75 , 4.1458412268560087e-47 ], [ 247.8125 , 4.1463203028149247e-47 ], [ 247.875 , 4.1467994341338414e-47 ], [ 247.9375 , 4.1472786208190967e-47 ], 
                    [ 248.0 , 4.1477578628772373e-47 ], [ 248.0625 , 4.1482371603145135e-47 ], [ 248.125 , 4.1487165131373837e-47 ], [ 248.1875 , 4.149196348919387e-47 ], 
                    [ 248.25 , 4.14967708646376e-47 ], [ 248.3125 , 4.150157879707736e-47 ], [ 248.375 , 4.150638728657768e-47 ], [ 248.4375 , 4.1511196333202804e-47 ], 
                    [ 248.5 , 4.1516005937017875e-47 ], [ 248.5625 , 4.1520816098087156e-47 ], [ 248.625 , 4.1525626816474907e-47 ], [ 248.6875 , 4.1530438092246603e-47 ], 
                    [ 248.75 , 4.1535254157490196e-47 ], [ 248.8125 , 4.154007916350563e-47 ], [ 248.875 , 4.154490473002522e-47 ], [ 248.9375 , 4.1549730857114105e-47 ], 
                    [ 249.0 , 4.155455754483738e-47 ], [ 249.0625 , 4.155938479326078e-47 ], [ 249.125 , 4.1564212602447946e-47 ], [ 249.1875 , 4.15690409724658e-47 ], 
                    [ 249.25 , 4.157386990337802e-47 ], [ 249.3125 , 4.157870335607403e-47 ], [ 249.375 , 4.158354598334142e-47 ], [ 249.4375 , 4.158838917462474e-47 ], 
                    [ 249.5 , 4.1593232929989385e-47 ], [ 249.5625 , 4.1598077249501056e-47 ], [ 249.625 , 4.1602922133225165e-47 ], [ 249.6875 , 4.1607767581227713e-47 ], 
                    [ 249.75 , 4.1612613593575316e-47 ], [ 249.8125 , 4.161746017033192e-47 ], [ 249.875 , 4.1622310763101765e-47 ], [ 249.9375 , 4.1627171063920327e-47 ], 
                    [ 250.0 , 4.1632031932283056e-47 ], [ 250.0625 , 4.1636893368257397e-47 ], [ 250.125 , 4.1641755371909043e-47 ], [ 250.1875 , 4.164661794330429e-47 ], 
                    [ 250.25 , 4.165148108250883e-47 ], [ 250.3125 , 4.165634478959046e-47 ], [ 250.375 , 4.1661209064614294e-47 ], [ 250.4375 , 4.166607654205363e-47 ], 
                    [ 250.5 , 4.1670954334648347e-47 ], [ 250.5625 , 4.167583269828039e-47 ], [ 250.625 , 4.168071163301574e-47 ], [ 250.6875 , 4.168559113892124e-47 ], 
                    [ 250.75 , 4.169047121606465e-47 ], [ 250.8125 , 4.169535186451255e-47 ], [ 250.875 , 4.170023308433063e-47 ], [ 250.9375 , 4.170511487558667e-47 ], 
                    [ 251.0 , 4.1709998870568e-47 ], [ 251.0625 , 4.171489429293903e-47 ], [ 251.125 , 4.171979028987667e-47 ], [ 251.1875 , 4.172468686144806e-47 ], 
                    [ 251.25 , 4.172958400772034e-47 ], [ 251.3125 , 4.173448172876156e-47 ], [ 251.375 , 4.1739380024638283e-47 ], [ 251.4375 , 4.1744278895419473e-47 ], 
                    [ 251.5 , 4.174917834117081e-47 ], [ 251.5625 , 4.1754078713040896e-47 ], [ 251.625 , 4.1758991777761544e-47 ], [ 251.6875 , 4.1763905420586013e-47 ], 
                    [ 251.75 , 4.176881964158292e-47 ], [ 251.8125 , 4.177373444082001e-47 ], [ 251.875 , 4.177864981836531e-47 ], [ 251.9375 , 4.1783565774286576e-47 ], 
                    [ 252.0 , 4.1788482308652756e-47 ], [ 252.0625 , 4.179339942153102e-47 ], [ 252.125 , 4.179831711298974e-47 ], [ 252.1875 , 4.180324648932148e-47 ], 
                    [ 252.25 , 4.180817761134387e-47 ], [ 252.3125 , 4.1813109315042725e-47 ], [ 252.375 , 4.181804160048635e-47 ], [ 252.4375 , 4.1822974467743975e-47 ], 
                    [ 252.5 , 4.182790791688363e-47 ], [ 252.5625 , 4.1832841947974846e-47 ], [ 252.625 , 4.183777656108508e-47 ], [ 252.6875 , 4.184271175628388e-47 ], 
                    [ 252.75 , 4.1847657007297386e-47 ], [ 252.8125 , 4.1852605816463193e-47 ], [ 252.875 , 4.1857555210864567e-47 ], [ 252.9375 , 4.186250519057012e-47 ], 
                    [ 253.0 , 4.186745575564906e-47 ], [ 253.0625 , 4.1872406906170026e-47 ], [ 253.125 , 4.1877358642203138e-47 ], [ 253.1875 , 4.1882310963818233e-47 ], 
                    [ 253.25 , 4.1887263871082774e-47 ], [ 253.3125 , 4.1892224624038904e-47 ], [ 253.375 , 4.189719087320687e-47 ], [ 253.4375 , 4.19021577111146e-47 ], 
                    [ 253.5 , 4.190712513783278e-47 ], [ 253.5625 , 4.1912093153430317e-47 ], [ 253.625 , 4.191706175797732e-47 ], [ 253.6875 , 4.192203095154361e-47 ], 
                    [ 253.75 , 4.1927000734199307e-47 ], [ 253.8125 , 4.1931971106013654e-47 ], [ 253.875 , 4.193694711492591e-47 ], [ 253.9375 , 4.1941930943271863e-47 ], 
                    [ 254.0 , 4.194691536390121e-47 ], [ 254.0625 , 4.1951900376883745e-47 ], [ 254.125 , 4.195688598228986e-47 ], [ 254.1875 , 4.196187218019086e-47 ], 
                    [ 254.25 , 4.196685897065596e-47 ], [ 254.3125 , 4.1971846353756477e-47 ], [ 254.375 , 4.1976834329562547e-47 ], [ 254.4375 , 4.1981825360396573e-47 ], 
                    [ 254.5 , 4.1986826527505023e-47 ], [ 254.5625 , 4.199182829038729e-47 ], [ 254.625 , 4.199683064911465e-47 ], [ 254.6875 , 4.2001833603757784e-47 ], 
                    [ 254.75 , 4.200683715438768e-47 ], [ 254.8125 , 4.201184130107444e-47 ], [ 254.875 , 4.201684604389146e-47 ], [ 254.9375 , 4.2021851382907075e-47 ], 
                    [ 255.0 , 4.2026857318193785e-47 ], [ 255.0625 , 4.2031875865651333e-47 ], [ 255.125 , 4.203689528171001e-47 ], [ 255.1875 , 4.204191529718371e-47 ], 
                    [ 255.25 , 4.2046935912144296e-47 ], [ 255.3125 , 4.2051957126662777e-47 ], [ 255.375 , 4.205697894081105e-47 ], [ 255.4375 , 4.2062001354660707e-47 ], 
                    [ 255.5 , 4.206702436828308e-47 ], [ 255.5625 , 4.207204798175099e-47 ], [ 255.625 , 4.207708095546618e-47 ], [ 255.6875 , 4.20821177704601e-47 ], 
                    [ 255.75 , 4.2087155188383026e-47 ], [ 255.8125 , 4.2092193209307434e-47 ], [ 255.875 , 4.2097231833305803e-47 ], [ 255.9375 , 4.2102271060449727e-47 ], 
                    [ 256.0 , 4.2107310890811705e-47 ], [ 256.0625 , 4.2112351324463937e-47 ], [ 256.125 , 4.211739236147865e-47 ], [ 256.1875 , 4.212243955445809e-47 ], 
                    [ 256.25 , 4.212749384192162e-47 ], [ 256.3125 , 4.2132548735851623e-47 ], [ 256.375 , 4.213760423632026e-47 ], [ 256.4375 , 4.2142660343399707e-47 ], 
                    [ 256.5 , 4.2147717057163365e-47 ], [ 256.5625 , 4.215277437768491e-47 ], [ 256.625 , 4.215783230503537e-47 ], [ 256.6875 , 4.2162890839288444e-47 ], 
                    [ 256.75 , 4.21679520414304e-47 ], [ 256.8125 , 4.2173023829444314e-47 ], [ 256.875 , 4.217809622747175e-47 ], [ 256.9375 , 4.2183169235586985e-47 ], 
                    [ 257.0 , 4.218824285386219e-47 ], [ 257.0625 , 4.2193317082371653e-47 ], [ 257.125 , 4.2198391921188777e-47 ], [ 257.1875 , 4.2203467370386363e-47 ], 
                    [ 257.25 , 4.2208543430037823e-47 ], [ 257.3125 , 4.2213620100217177e-47 ], [ 257.375 , 4.221870757124552e-47 ], [ 257.4375 , 4.2223797310071905e-47 ], 
                    [ 257.5 , 4.222888766249959e-47 ], [ 257.5625 , 4.223397862860195e-47 ], [ 257.625 , 4.223907020845267e-47 ], [ 257.6875 , 4.2244162402127245e-47 ], 
                    [ 257.75 , 4.224925520969877e-47 ], [ 257.8125 , 4.2254348631240654e-47 ], [ 257.875 , 4.2259442666828117e-47 ], [ 257.9375 , 4.2264543583990396e-47 ], 
                    [ 258.0 , 4.2269650716664936e-47 ], [ 258.0625 , 4.2274758466471823e-47 ], [ 258.125 , 4.227986683348443e-47 ], [ 258.1875 , 4.228497581777943e-47 ], 
                    [ 258.25 , 4.2290085419429616e-47 ], [ 258.3125 , 4.2295195638509595e-47 ], [ 258.375 , 4.2300306475095467e-47 ], [ 258.4375 , 4.230541792926066e-47 ], 
                    [ 258.5 , 4.23105320942787e-47 ], [ 258.5625 , 4.231565675487012e-47 ], [ 258.625 , 4.2320782036162116e-47 ], [ 258.6875 , 4.2325907938229273e-47 ], 
                    [ 258.75 , 4.233103446114676e-47 ], [ 258.8125 , 4.2336161604989496e-47 ], [ 258.875 , 4.234128936983297e-47 ], [ 258.9375 , 4.2346417755752417e-47 ], 
                    [ 259.0 , 4.235154676282394e-47 ], [ 259.0625 , 4.2356676391120686e-47 ], [ 259.125 , 4.236181605117575e-47 ], [ 259.1875 , 4.236695868073498e-47 ], 
                    [ 259.25 , 4.237210193459822e-47 ], [ 259.3125 , 4.2377245812840957e-47 ], [ 259.375 , 4.238239031553869e-47 ], [ 259.4375 , 4.238753544276692e-47 ], 
                    [ 259.5 , 4.239268119460327e-47 ], [ 259.5625 , 4.2397827571121765e-47 ], [ 259.625 , 4.240297457239884e-47 ], [ 259.6875 , 4.2408126917457217e-47 ], 
                    [ 259.75 , 4.2413286827945797e-47 ], [ 259.8125 , 4.241844736625483e-47 ], [ 259.875 , 4.2423608532460085e-47 ], [ 259.9375 , 4.2428770326638267e-47 ], 
                    [ 260.0 , 4.2433932748866394e-47 ], [ 260.0625 , 4.2439095799219362e-47 ], [ 260.125 , 4.244425947777481e-47 ], [ 260.1875 , 4.2449423784609474e-47 ], 
                    [ 260.25 , 4.2454588719798897e-47 ], [ 260.3125 , 4.2459765984772853e-47 ], [ 260.375 , 4.24649440251633e-47 ], [ 260.4375 , 4.2470122697024536e-47 ], 
                    [ 260.5 , 4.2475302000433286e-47 ], [ 260.5625 , 4.2480481935466857e-47 ], [ 260.625 , 4.248566250220228e-47 ], [ 260.6875 , 4.2490843700717197e-47 ], 
                    [ 260.75 , 4.2496025531088045e-47 ], [ 260.8125 , 4.250120799339159e-47 ], [ 260.875 , 4.250639747975258e-47 ], [ 260.9375 , 4.2511592842103923e-47 ], 
                    [ 261.0 , 4.2516788839461227e-47 ], [ 261.0625 , 4.2521985471901796e-47 ], [ 261.125 , 4.2527182739502053e-47 ], [ 261.1875 , 4.253238064234083e-47 ], 
                    [ 261.25 , 4.2537579180495477e-47 ], [ 261.3125 , 4.254277835404425e-47 ], [ 261.375 , 4.254797816306359e-47 ], [ 261.4375 , 4.255317964362063e-47 ], 
                    [ 261.5 , 4.255839226711968e-47 ], [ 261.5625 , 4.2563605529148147e-47 ], [ 261.625 , 4.256881942978302e-47 ], [ 261.6875 , 4.257403396910405e-47 ], 
                    [ 261.75 , 4.2579249147188266e-47 ], [ 261.8125 , 4.2584464964115425e-47 ], [ 261.875 , 4.2589681419962566e-47 ], [ 261.9375 , 4.2594898514807964e-47 ], 
                    [ 262.0 , 4.26001162487308e-47 ], [ 262.0625 , 4.2605341712462543e-47 ], [ 262.125 , 4.2610572377070104e-47 ], [ 262.1875 , 4.2615803683846893e-47 ], 
                    [ 262.25 , 4.2621035632872045e-47 ], [ 262.3125 , 4.262626822422503e-47 ], [ 262.375 , 4.263150145798348e-47 ], [ 262.4375 , 4.263673533422688e-47 ], 
                    [ 262.5 , 4.2641969853034094e-47 ], [ 262.5625 , 4.264720501448372e-47 ], [ 262.625 , 4.265244203681514e-47 ], [ 262.6875 , 4.2657689986102783e-47 ], 
                    [ 262.75 , 4.2662938581097517e-47 ], [ 262.8125 , 4.2668187821878204e-47 ], [ 262.875 , 4.2673437708523977e-47 ], [ 262.9375 , 4.267868824111584e-47 ], 
                    [ 263.0 , 4.2683939419732033e-47 ], [ 263.0625 , 4.2689191244451764e-47 ], [ 263.125 , 4.2694443715356327e-47 ], [ 263.1875 , 4.269969683252373e-47 ], 
                    [ 263.25 , 4.270495727843968e-47 ], [ 263.3125 , 4.271022317714361e-47 ], [ 263.375 , 4.2715489725179397e-47 ], [ 263.4375 , 4.272075692262771e-47 ], 
                    [ 263.5 , 4.2726024769567725e-47 ], [ 263.5625 , 4.2731293266079833e-47 ], [ 263.625 , 4.2736562412244126e-47 ], [ 263.6875 , 4.2741832208141023e-47 ], 
                    [ 263.75 , 4.2747102653850333e-47 ], [ 263.8125 , 4.2752374159008343e-47 ], [ 263.875 , 4.275765728261349e-47 ], [ 263.9375 , 4.27629410590805e-47 ], 
                    [ 264.0 , 4.2768225488489765e-47 ], [ 264.0625 , 4.277351057092287e-47 ], [ 264.125 , 4.2778796306459e-47 ], [ 264.1875 , 4.2784082695180687e-47 ], 
                    [ 264.25 , 4.2789369737167125e-47 ], [ 264.3125 , 4.279465743250025e-47 ], [ 264.375 , 4.2799945781259596e-47 ], [ 264.4375 , 4.2805240152803647e-47 ], 
                    [ 264.5 , 4.281054127180085e-47 ], [ 264.5625 , 4.2815843047302946e-47 ], [ 264.625 , 4.2821145479390935e-47 ], [ 264.6875 , 4.282644856814735e-47 ], 
                    [ 264.75 , 4.28317523136532e-47 ], [ 264.8125 , 4.283705671598922e-47 ], [ 264.875 , 4.284236177523705e-47 ], [ 264.9375 , 4.2847667491478058e-47 ], 
                    [ 265.0 , 4.2852973864793894e-47 ], [ 265.0625 , 4.2858290955692994e-47 ], [ 265.125 , 4.2863610086692254e-47 ], [ 265.1875 , 4.2868929877847737e-47 ], 
                    [ 265.25 , 4.2874250329239865e-47 ], [ 265.3125 , 4.28795714409527e-47 ], [ 265.375 , 4.288489321306758e-47 ], [ 265.4375 , 4.2890215645665873e-47 ], 
                    [ 265.5 , 4.2895538738829237e-47 ], [ 265.5625 , 4.2900862492640567e-47 ], [ 265.625 , 4.290618991088542e-47 ], [ 265.6875 , 4.2911526169147477e-47 ], 
                    [ 265.75 , 4.2916863091081584e-47 ], [ 265.8125 , 4.29222006767709e-47 ], [ 265.875 , 4.2927538926297665e-47 ], [ 265.9375 , 4.2932877839744136e-47 ], 
                    [ 266.0 , 4.2938217417193806e-47 ], [ 266.0625 , 4.294355765872803e-47 ], [ 266.125 , 4.2948898564430316e-47 ], [ 266.1875 , 4.295424013438266e-47 ], 
                    [ 266.25 , 4.295958951552373e-47 ], [ 266.3125 , 4.2964943653707923e-47 ], [ 266.375 , 4.297029845918859e-47 ], [ 266.4375 , 4.29756539320495e-47 ], 
                    [ 266.5 , 4.2981010072373846e-47 ], [ 266.5625 , 4.2986366880244194e-47 ], [ 266.625 , 4.299172435574374e-47 ], [ 266.6875 , 4.299708249895631e-47 ], 
                    [ 266.75 , 4.3002441309964506e-47 ], [ 266.8125 , 4.3007800788852477e-47 ], [ 266.875 , 4.301317197200803e-47 ], [ 266.9375 , 4.301854403469715e-47 ], 
                    [ 267.0 , 4.302391676832101e-47 ], [ 267.0625 , 4.302929017296494e-47 ], [ 267.125 , 4.30346642487112e-47 ], [ 267.1875 , 4.3040038995644227e-47 ], 
                    [ 267.25 , 4.304541441384754e-47 ], [ 267.3125 , 4.3050790503405894e-47 ], [ 267.375 , 4.3056167264402224e-47 ], [ 267.4375 , 4.3061548140685976e-47 ], 
                    [ 267.5 , 4.306693753324292e-47 ], [ 267.5625 , 4.3072327600313277e-47 ], [ 267.625 , 4.307771834198024e-47 ], [ 267.6875 , 4.308310975832884e-47 ], 
                    [ 267.75 , 4.3088501849443234e-47 ], [ 267.8125 , 4.309389461540785e-47 ], [ 267.875 , 4.3099288056306547e-47 ], [ 267.9375 , 4.310468217222502e-47 ], 
                    [ 268.0 , 4.3110076963246224e-47 ], [ 268.0625 , 4.311547908323241e-47 ], [ 268.125 , 4.3120886190829464e-47 ], [ 268.1875 , 4.3126293976530997e-47 ], 
                    [ 268.25 , 4.313170244042236e-47 ], [ 268.3125 , 4.313711158258859e-47 ], [ 268.375 , 4.3142521403115376e-47 ], [ 268.4375 , 4.314793190208594e-47 ], 
                    [ 268.5 , 4.3153343079587836e-47 ], [ 268.5625 , 4.3158754935703993e-47 ], [ 268.625 , 4.3164167470520443e-47 ], [ 268.6875 , 4.316959059975342e-47 ], 
                    [ 268.75 , 4.317501567926215e-47 ], [ 268.8125 , 4.318044144053553e-47 ], [ 268.875 , 4.3185867883658937e-47 ], [ 268.9375 , 4.3191295008717747e-47 ], 
                    [ 269.0 , 4.319672281579858e-47 ], [ 269.0625 , 4.32021513049856e-47 ], [ 269.125 , 4.3207580476366387e-47 ], [ 269.1875 , 4.3213010330025737e-47 ], 
                    [ 269.25 , 4.321844245540693e-47 ], [ 269.3125 , 4.322388468820169e-47 ], [ 269.375 , 4.3229327606303364e-47 ], [ 269.4375 , 4.3234771209798573e-47 ], 
                    [ 269.5 , 4.3240215498773e-47 ], [ 269.5625 , 4.324566047331327e-47 ], [ 269.625 , 4.325110613350633e-47 ], [ 269.6875 , 4.325655247943759e-47 ], 
                    [ 269.75 , 4.3261999511193415e-47 ], [ 269.8125 , 4.3267447228860463e-47 ], [ 269.875 , 4.327289980415701e-47 ], [ 269.9375 , 4.327835987379373e-47 ], 
                    [ 270.0 , 4.328382063236788e-47 ], [ 270.0625 , 4.3289282079968237e-47 ], [ 270.125 , 4.3294744216680814e-47 ], [ 270.1875 , 4.330020704259196e-47 ], 
                    [ 270.25 , 4.3305670557789536e-47 ], [ 270.3125 , 4.331113476236054e-47 ], [ 270.375 , 4.331659965639132e-47 ], [ 270.4375 , 4.3322065239969186e-47 ], 
                    [ 270.5 , 4.332753803275508e-47 ], [ 270.5625 , 4.333301598396901e-47 ], [ 270.625 , 4.3338494627767324e-47 ], [ 270.6875 , 4.3343973964237274e-47 ], 
                    [ 270.75 , 4.334945399346552e-47 ], [ 270.8125 , 4.3354934715540873e-47 ], [ 270.875 , 4.336041613054908e-47 ], [ 270.9375 , 4.336589823858053e-47 ], 
                    [ 271.0 , 4.3371381039720063e-47 ], [ 271.0625 , 4.3376864534057164e-47 ], [ 271.125 , 4.3382357232890165e-47 ], [ 271.1875 , 4.338785296140381e-47 ], 
                    [ 271.25 , 4.3393349386123475e-47 ], [ 271.3125 , 4.339884650713644e-47 ], [ 271.375 , 4.3404344324530917e-47 ], [ 271.4375 , 4.3409842838396044e-47 ], 
                    [ 271.5 , 4.3415342048819124e-47 ], [ 271.5625 , 4.3420841955889017e-47 ], [ 271.625 , 4.3426342559693054e-47 ], [ 271.6875 , 4.3431843860320417e-47 ], 
                    [ 271.75 , 4.3437356253532e-47 ], [ 271.8125 , 4.344286985044434e-47 ], [ 271.875 , 4.3448384147209377e-47 ], [ 271.9375 , 4.3453899143915947e-47 ], 
                    [ 272.0 , 4.345941484065289e-47 ], [ 272.0625 , 4.3464931237509377e-47 ], [ 272.125 , 4.3470448334573336e-47 ], [ 272.1875 , 4.3475966131934285e-47 ], 
                    [ 272.25 , 4.3481484629682024e-47 ], [ 272.3125 , 4.348700492400054e-47 ], [ 272.375 , 4.3492535715572905e-47 ], [ 272.4375 , 4.3498067210565576e-47 ], 
                    [ 272.5 , 4.3503599409068316e-47 ], [ 272.5625 , 4.3509132311170605e-47 ], [ 272.625 , 4.3514665916962224e-47 ], [ 272.6875 , 4.352020022653268e-47 ], 
                    [ 272.75 , 4.352573523997025e-47 ], [ 272.8125 , 4.35312709573663e-47 ], [ 272.875 , 4.3536807378809126e-47 ], [ 272.9375 , 4.3542346911237638e-47 ], 
                    [ 273.0 , 4.3547895434116917e-47 ], [ 273.0625 , 4.355344466403479e-47 ], [ 273.125 , 4.355899460108073e-47 ], [ 273.1875 , 4.356454524534547e-47 ], 
                    [ 273.25 , 4.3570096596918504e-47 ], [ 273.3125 , 4.357564865589027e-47 ], [ 273.375 , 4.358120142235154e-47 ], [ 273.4375 , 4.3586754896391225e-47 ], 
                    [ 273.5 , 4.359230907810011e-47 ], [ 273.5625 , 4.35978674678883e-47 ], [ 273.625 , 4.360343376481608e-47 ], [ 273.6875 , 4.3609000772412756e-47 ], 
                    [ 273.75 , 4.361456849076936e-47 ], [ 273.8125 , 4.362013691997725e-47 ], [ 273.875 , 4.362570606012627e-47 ], [ 273.9375 , 4.36312759113075e-47 ], 
                    [ 274.0 , 4.363684647361139e-47 ], [ 274.0625 , 4.364241774712937e-47 ], [ 274.125 , 4.3647989731951933e-47 ], [ 274.1875 , 4.3653566825669437e-47 ], 
                    [ 274.25 , 4.3659151047421966e-47 ], [ 274.3125 , 4.3664735983515806e-47 ], [ 274.375 , 4.3670321634041713e-47 ], [ 274.4375 , 4.367590799909108e-47 ], 
                    [ 274.5 , 4.368149507875593e-47 ], [ 274.5625 , 4.3687082873128e-47 ], [ 274.625 , 4.369267138229714e-47 ], [ 274.6875 , 4.3698260606356356e-47 ], 
                    [ 274.75 , 4.370385054539648e-47 ], [ 274.8125 , 4.370944608078996e-47 ], [ 274.875 , 4.371504797203969e-47 ], [ 274.9375 , 4.372065058123979e-47 ], 
                    [ 275.0 , 4.372625390848196e-47 ], [ 275.0625 , 4.3731857953857594e-47 ], [ 275.125 , 4.3737462717458754e-47 ], [ 275.1875 , 4.374306819937808e-47 ], 
                    [ 275.25 , 4.3748674399707648e-47 ], [ 275.3125 , 4.37542813185386e-47 ], [ 275.375 , 4.3759888955963327e-47 ], [ 275.4375 , 4.376550262780098e-47 ], 
                    [ 275.5 , 4.377112236689973e-47 ], [ 275.5625 , 4.377674282760489e-47 ], [ 275.625 , 4.378236401000851e-47 ], [ 275.6875 , 4.37879859142045e-47 ], 
                    [ 275.75 , 4.37936085402849e-47 ], [ 275.8125 , 4.379923188834213e-47 ], [ 275.875 , 4.380485595846981e-47 ], [ 275.9375 , 4.38104807507594e-47 ], 
                    [ 276.0 , 4.381610626530459e-47 ], [ 276.0625 , 4.382173790424315e-47 ], [ 276.125 , 4.382737544625926e-47 ], [ 276.1875 , 4.383301371352975e-47 ], 
                    [ 276.25 , 4.383865270614668e-47 ], [ 276.3125 , 4.384429242420429e-47 ], [ 276.375 , 4.384993286779497e-47 ], [ 276.4375 , 4.385557403701394e-47 ], 
                    [ 276.5 , 4.386121593195235e-47 ], [ 276.5625 , 4.386685855270451e-47 ], [ 276.625 , 4.38725018993638e-47 ], [ 276.6875 , 4.387815121385457e-47 ], 
                    [ 276.75 , 4.388380649829454e-47 ], [ 276.8125 , 4.388946251162213e-47 ], [ 276.875 , 4.389511925393101e-47 ], [ 276.9375 , 4.390077672531574e-47 ], 
                    [ 277.0 , 4.390643492587029e-47 ], [ 277.0625 , 4.391209385568769e-47 ], [ 277.125 , 4.391775351486257e-47 ], [ 277.1875 , 4.392341390348926e-47 ], 
                    [ 277.25 , 4.392907502166144e-47 ], [ 277.3125 , 4.393474177896601e-47 ], [ 277.375 , 4.394041487764309e-47 ], [ 277.4375 , 4.394608870886133e-47 ], 
                    [ 277.5 , 4.395176327271628e-47 ], [ 277.5625 , 4.395743856930315e-47 ], [ 277.625 , 4.3963114598715335e-47 ], [ 277.6875 , 4.396879136104868e-47 ], 
                    [ 277.75 , 4.397446885639627e-47 ], [ 277.8125 , 4.398014708485433e-47 ], [ 277.875 , 4.398582604651657e-47 ], [ 277.9375 , 4.399150996733219e-47 ], 
                    [ 278.0 , 4.399720068645548e-47 ], [ 278.0625 , 4.400289214172759e-47 ], [ 278.125 , 4.400858433324311e-47 ], [ 278.1875 , 4.40142772610976e-47 ], 
                    [ 278.25 , 4.401997092538663e-47 ], [ 278.3125 , 4.402566532620451e-47 ], [ 278.375 , 4.403136046364779e-47 ], [ 278.4375 , 4.403705633781113e-47 ], 
                    [ 278.5 , 4.40427529487895e-47 ], [ 278.5625 , 4.404845375252906e-47 ], [ 278.625 , 4.40541622534276e-47 ], [ 278.6875 , 4.405987149412468e-47 ], 
                    [ 278.75 , 4.406558147471621e-47 ], [ 278.8125 , 4.407129219529805e-47 ], [ 278.875 , 4.40770036559658e-47 ], [ 278.9375 , 4.408271585681599e-47 ], 
                    [ 279.0 , 4.408842879794455e-47 ], [ 279.0625 , 4.409414247944711e-47 ], [ 279.125 , 4.409985690141867e-47 ], [ 279.1875 , 4.410557448880341e-47 ], 
                    [ 279.25 , 4.411130093283088e-47 ], [ 279.3125 , 4.411702812035079e-47 ], [ 279.375 , 4.412275605145936e-47 ], [ 279.4375 , 4.412848472625406e-47 ], 
                    [ 279.5 , 4.413421414483052e-47 ], [ 279.5625 , 4.413994430728531e-47 ], [ 279.625 , 4.414567521371564e-47 ], [ 279.6875 , 4.415140686421747e-47 ], 
                    [ 279.75 , 4.415713925888866e-47 ], [ 279.8125 , 4.416287346291878e-47 ], [ 279.875 , 4.41686174337915e-47 ], [ 279.9375 , 4.417436215174396e-47 ], 
                    [ 280.0 , 4.418010761687424e-47 ], [ 280.0625 , 4.418585382927829e-47 ], [ 280.125 , 4.419160078905393e-47 ], [ 280.1875 , 4.419734849629867e-47 ], 
                    [ 280.25 , 4.420309695110878e-47 ], [ 280.3125 , 4.420884615358308e-47 ], [ 280.375 , 4.421459610381723e-47 ], [ 280.4375 , 4.422034680190974e-47 ], 
                    [ 280.5 , 4.422610814287852e-47 ], [ 280.5625 , 4.423187066747587e-47 ], [ 280.625 , 4.423763394291323e-47 ], [ 280.6875 , 4.424339796928623e-47 ], 
                    [ 280.75 , 4.424916274669461e-47 ], [ 280.8125 , 4.425492827523527e-47 ], [ 280.875 , 4.426069455500672e-47 ], [ 280.9375 , 4.426646158610682e-47 ], 
                    [ 281.0 , 4.427222936863319e-47 ], [ 281.0625 , 4.427799790268309e-47 ], [ 281.125 , 4.428377521151435e-47 ], [ 281.1875 , 4.428955543916113e-47 ], 
                    [ 281.25 , 4.429533642128423e-47 ], [ 281.3125 , 4.430111815798117e-47 ], [ 281.375 , 4.430690064934983e-47 ], [ 281.4375 , 4.431268389549091e-47 ], 
                    [ 281.5 , 4.43184678965004e-47 ], [ 281.5625 , 4.432425265247937e-47 ], [ 281.625 , 4.433003816352447e-47 ], [ 281.6875 , 4.43358244297355e-47 ], 
                    [ 281.75 , 4.434161749721615e-47 ], [ 281.8125 , 4.434741545600598e-47 ], [ 281.875 , 4.435321417291713e-47 ], [ 281.9375 , 4.435901364804875e-47 ], 
                    [ 282.0 , 4.436481388149996e-47 ], [ 282.0625 , 4.437061487336995e-47 ], [ 282.125 , 4.437641662375817e-47 ], [ 282.1875 , 4.438221913276318e-47 ], 
                    [ 282.25 , 4.438802240048512e-47 ], [ 282.3125 , 4.439382642702195e-47 ], [ 282.375 , 4.439963502561733e-47 ], [ 282.4375 , 4.440545067883835e-47 ], 
                    [ 282.5 , 4.441126709381902e-47 ], [ 282.5625 , 4.441708427065787e-47 ], [ 282.625 , 4.442290220945562e-47 ], [ 282.6875 , 4.442872091031175e-47 ], 
                    [ 282.75 , 4.44345403733261e-47 ], [ 282.8125 , 4.444036059859816e-47 ], [ 282.875 , 4.444618158622876e-47 ], [ 282.9375 , 4.445200333631614e-47 ], 
                    [ 283.0 , 4.4457827239036945e-47 ], [ 283.0625 , 4.446366062989491e-47 ], [ 283.125 , 4.446949478616251e-47 ], [ 283.1875 , 4.447532970794017e-47 ], 
                    [ 283.25 , 4.448116539532864e-47 ], [ 283.3125 , 4.448700184842743e-47 ], [ 283.375 , 4.449283906733829e-47 ], [ 283.4375 , 4.449867705216106e-47 ], 
                    [ 283.5 , 4.450451580299591e-47 ], [ 283.5625 , 4.451035531994338e-47 ], [ 283.625 , 4.451619560310492e-47 ], [ 283.6875 , 4.452204534619269e-47 ], 
                    [ 283.75 , 4.452789710690505e-47 ], [ 283.8125 , 4.453374963674401e-47 ], [ 283.875 , 4.453960293581098e-47 ], [ 283.9375 , 4.454545700420706e-47 ], 
                    [ 284.0 , 4.455131184203337e-47 ], [ 284.0625 , 4.455716744939103e-47 ], [ 284.125 , 4.456302382638183e-47 ], [ 284.1875 , 4.456888097310533e-47 ], 
                    [ 284.25 , 4.457473888966367e-47 ], [ 284.3125 , 4.458060350447132e-47 ], [ 284.375 , 4.458647304527379e-47 ], [ 284.4375 , 4.459234335886709e-47 ], 
                    [ 284.5 , 4.459821444535389e-47 ], [ 284.5625 , 4.460408630483597e-47 ], [ 284.625 , 4.460995893741416e-47 ], [ 284.6875 , 4.46158323431918e-47 ], 
                    [ 284.75 , 4.462170652226882e-47 ], [ 284.8125 , 4.462758147474858e-47 ], [ 284.875 , 4.463345720073233e-47 ], [ 284.9375 , 4.463933647057191e-47 ], 
                    [ 285.0 , 4.464522369615901e-47 ], [ 285.0625 , 4.465111169817814e-47 ], [ 285.125 , 4.465700047673172e-47 ], [ 285.1875 , 4.466289003192309e-47 ], 
                    [ 285.25 , 4.466878036385343e-47 ], [ 285.3125 , 4.467467147262611e-47 ], [ 285.375 , 4.468056335834296e-47 ], [ 285.4375 , 4.468645602110676e-47 ], 
                    [ 285.5 , 4.469234946102063e-47 ], [ 285.5625 , 4.469824367818578e-47 ], [ 285.625 , 4.470414800148657e-47 ], [ 285.6875 , 4.471005364312829e-47 ], 
                    [ 285.75 , 4.471596006493546e-47 ], [ 285.8125 , 4.472186726700986e-47 ], [ 285.875 , 4.472777524945555e-47 ], [ 285.9375 , 4.473368401237495e-47 ], 
                    [ 286.0 , 4.473959355587214e-47 ], [ 286.0625 , 4.474550388004895e-47 ], [ 286.125 , 4.475141498500917e-47 ], [ 286.1875 , 4.475732687085656e-47 ], 
                    [ 286.25 , 4.4763245354444e-47 ], [ 286.3125 , 4.476916868075733e-47 ], [ 286.375 , 4.477509279087814e-47 ], [ 286.4375 , 4.47810176849114e-47 ], 
                    [ 286.5 , 4.478694336295958e-47 ], [ 286.5625 , 4.479286982512803e-47 ], [ 286.625 , 4.479879707151826e-47 ], [ 286.6875 , 4.480472510223597e-47 ], 
                    [ 286.75 , 4.481065391738397e-47 ], [ 286.8125 , 4.481658351706671e-47 ], [ 286.875 , 4.482251596626729e-47 ], [ 286.9375 , 4.482845701104793e-47 ], 
                    [ 287.0 , 4.483439884328958e-47 ], [ 287.0625 , 4.48403414630976e-47 ], [ 287.125 , 4.484628487057605e-47 ], [ 287.1875 , 4.485222906582965e-47 ], 
                    [ 287.25 , 4.485817404896217e-47 ], [ 287.3125 , 4.486411982007838e-47 ], [ 287.375 , 4.487006637928271e-47 ], [ 287.4375 , 4.487601372667961e-47 ], 
                    [ 287.5 , 4.488196186237358e-47 ], [ 287.5625 , 4.488791857029774e-47 ], [ 287.625 , 4.489387794544545e-47 ], [ 287.6875 , 4.489983811176744e-47 ], 
                    [ 287.75 , 4.490579906936873e-47 ], [ 287.8125 , 4.491176081835344e-47 ], [ 287.875 , 4.491772335882726e-47 ], [ 287.9375 , 4.492368669089558e-47 ], 
                    [ 288.0 , 4.492965081466255e-47 ], [ 288.0625 , 4.493561573023487e-47 ], [ 288.125 , 4.494158143771606e-47 ], [ 288.1875 , 4.49475516351595e-47 ], 
                    [ 288.25 , 4.495352869289485e-47 ], [ 288.3125 , 4.495950654545067e-47 ], [ 288.375 , 4.496548519293198e-47 ], [ 288.4375 , 4.497146463544516e-47 ], 
                    [ 288.5 , 4.497744487309528e-47 ], [ 288.5625 , 4.498342590598872e-47 ], [ 288.625 , 4.49894077342309e-47 ], [ 288.6875 , 4.4995390357927285e-47 ], 
                    [ 288.75 , 4.500137377718428e-47 ], [ 288.8125 , 4.50073579921064e-47 ], [ 288.875 , 4.501335205929604e-47 ], [ 288.9375 , 4.501934762752614e-47 ], 
                    [ 289.0 , 4.502534399433791e-47 ], [ 289.0625 , 4.503134115983773e-47 ], [ 289.125 , 4.503733912413135e-47 ], [ 289.1875 , 4.504333788732675e-47 ], 
                    [ 289.25 , 4.504933744952938e-47 ], [ 289.3125 , 4.505533781084472e-47 ], [ 289.375 , 4.5061338971380155e-47 ], [ 289.4375 , 4.506734093124311e-47 ], 
                    [ 289.5 , 4.50733480381564e-47 ], [ 289.5625 , 4.507936118016248e-47 ], [ 289.625 , 4.508537512436937e-47 ], [ 289.6875 , 4.509138987088443e-47 ], 
                    [ 289.75 , 4.509740541981468e-47 ], [ 289.8125 , 4.510342177126653e-47 ], [ 289.875 , 4.51094389253477e-47 ], [ 289.9375 , 4.51154568821646e-47 ], 
                    [ 290.0 , 4.512147564182403e-47 ], [ 290.0625 , 4.5127495204434665e-47 ], [ 290.125 , 4.513351557010173e-47 ], [ 290.1875 , 4.51395459262531e-47 ], 
                    [ 290.25 , 4.514557751425133e-47 ], [ 290.3125 , 4.515160990819595e-47 ], [ 290.375 , 4.515764310819496e-47 ], [ 290.4375 , 4.516367711435542e-47 ], 
                    [ 290.5 , 4.516971192678507e-47 ], [ 290.5625 , 4.517574754559259e-47 ], [ 290.625 , 4.518178397088478e-47 ], [ 290.6875 , 4.518782120277004e-47 ], 
                    [ 290.75 , 4.51938592413555e-47 ], [ 290.8125 , 4.519990226063116e-47 ], [ 290.875 , 4.520595153734752e-47 ], [ 290.9375 , 4.521200162366249e-47 ], 
                    [ 291.0 , 4.521805251968377e-47 ], [ 291.0625 , 4.522410422551876e-47 ], [ 291.125 , 4.523015674127746e-47 ], [ 291.1875 , 4.523621006706826e-47 ], 
                    [ 291.25 , 4.524226420299764e-47 ], [ 291.3125 , 4.524831914917531e-47 ], [ 291.375 , 4.525437490571099e-47 ], [ 291.4375 , 4.526043147271088e-47 ], 
                    [ 291.5 , 4.526649719381831e-47 ], [ 291.5625 , 4.527256477862917e-47 ], [ 291.625 , 4.527863317674697e-47 ], [ 291.6875 , 4.528470238828138e-47 ], 
                    [ 291.75 , 4.529077241334172e-47 ], [ 291.8125 , 4.5296843252036765e-47 ], [ 291.875 , 4.530291490447553e-47 ], [ 291.9375 , 4.530898737076679e-47 ], 
                    [ 292.0 , 4.531506065101997e-47 ], [ 292.0625 , 4.532113474534478e-47 ], [ 292.125 , 4.532721271132904e-47 ], [ 292.1875 , 4.533329803845924e-47 ], 
                    [ 292.25 , 4.533938418256459e-47 ], [ 292.3125 , 4.534547114375478e-47 ], [ 292.375 , 4.535155892213916e-47 ], [ 292.4375 , 4.535764751782843e-47 ], 
                    [ 292.5 , 4.5363736930931e-47 ], [ 292.5625 , 4.536982716155826e-47 ], [ 292.625 , 4.5375918209818e-47 ], [ 292.6875 , 4.538201007582097e-47 ], 
                    [ 292.75 , 4.53881027596776e-47 ], [ 292.8125 , 4.539420303449411e-47 ], [ 292.875 , 4.540030669913308e-47 ], [ 292.9375 , 4.540641118446449e-47 ], 
                    [ 293.0 , 4.541251649059993e-47 ], [ 293.0625 , 4.54186226176498e-47 ], [ 293.125 , 4.542472956572415e-47 ], [ 293.1875 , 4.543083733493207e-47 ], 
                    [ 293.25 , 4.543694592538657e-47 ], [ 293.3125 , 4.544305533719709e-47 ], [ 293.375 , 4.544916557047278e-47 ], [ 293.4375 , 4.54552776287939e-47 ], 
                    [ 293.5 , 4.546139885348769e-47 ], [ 293.5625 , 4.546752090249548e-47 ], [ 293.625 , 4.547364377592727e-47 ], [ 293.6875 , 4.5479767473894445e-47 ], 
                    [ 293.75 , 4.548589199650802e-47 ], [ 293.8125 , 4.54920173438784e-47 ], [ 293.875 , 4.549814351611827e-47 ], [ 293.9375 , 4.550427051333743e-47 ], 
                    [ 294.0 , 4.551039833564695e-47 ], [ 294.0625 , 4.551652698315828e-47 ], [ 294.125 , 4.552266080562502e-47 ], [ 294.1875 , 4.552880044773352e-47 ], 
                    [ 294.25 , 4.5534940917896e-47 ], [ 294.3125 , 4.554108221622346e-47 ], [ 294.375 , 4.554722434282828e-47 ], [ 294.4375 , 4.555336729782118e-47 ], 
                    [ 294.5 , 4.555951108131551e-47 ], [ 294.5625 , 4.55656556934217e-47 ], [ 294.625 , 4.557180113425087e-47 ], [ 294.6875 , 4.557794740391673e-47 ], 
                    [ 294.75 , 4.558409450252978e-47 ], [ 294.8125 , 4.559024992239504e-47 ], [ 294.875 , 4.559640804862812e-47 ], [ 294.9375 , 4.560256700667357e-47 ], 
                    [ 295.0 , 4.560872679664276e-47 ], [ 295.0625 , 4.561488741864938e-47 ], [ 295.125 , 4.562104887280481e-47 ], [ 295.1875 , 4.562721115922149e-47 ], 
                    [ 295.25 , 4.563337427801312e-47 ], [ 295.3125 , 4.563953822929051e-47 ], [ 295.375 , 4.564570301316643e-47 ], [ 295.4375 , 4.56518696580679e-47 ], 
                    [ 295.5 , 4.565804546399265e-47 ], [ 295.5625 , 4.566422210538325e-47 ], [ 295.625 , 4.567039958235309e-47 ], [ 295.6875 , 4.567657789501419e-47 ], 
                    [ 295.75 , 4.568275704348124e-47 ], [ 295.8125 , 4.568893702786602e-47 ], [ 295.875 , 4.56951178482816e-47 ], [ 295.9375 , 4.570129950484141e-47 ], 
                    [ 296.0 , 4.570748199765824e-47 ], [ 296.0625 , 4.571366532684652e-47 ], [ 296.125 , 4.571985305599481e-47 ], [ 296.1875 , 4.572604706228051e-47 ], 
                    [ 296.25 , 4.5732241907714e-47 ], [ 296.3125 , 4.57384375924093e-47 ], [ 296.375 , 4.574463411647944e-47 ], [ 296.4375 , 4.575083148003882e-47 ], 
                    [ 296.5 , 4.57570296832005e-47 ], [ 296.5625 , 4.576322872607856e-47 ], [ 296.625 , 4.576942860878675e-47 ], [ 296.6875 , 4.577562933143919e-47 ], 
                    [ 296.75 , 4.578183089414901e-47 ], [ 296.8125 , 4.578803937532724e-47 ], [ 296.875 , 4.579425186822386e-47 ], [ 296.9375 , 4.580046520402801e-47 ], 
                    [ 297.0 , 4.580667938285339e-47 ], [ 297.0625 , 4.581289440481502e-47 ], [ 297.125 , 4.581911027002734e-47 ], [ 297.1875 , 4.582532697860311e-47 ], 
                    [ 297.25 , 4.58315445306587e-47 ], [ 297.3125 , 4.583776292630824e-47 ], [ 297.375 , 4.584398216566618e-47 ], [ 297.4375 , 4.585020224884571e-47 ], 
                    [ 297.5 , 4.585643137295243e-47 ], [ 297.5625 , 4.586266230963153e-47 ], [ 297.625 , 4.586889409296487e-47 ], [ 297.6875 , 4.587512672306814e-47 ], 
                    [ 297.75 , 4.58813602000564e-47 ], [ 297.8125 , 4.588759452404505e-47 ], [ 297.875 , 4.589382969514789e-47 ], [ 297.9375 , 4.590006571348229e-47 ], 
                    [ 298.0 , 4.590630257916141e-47 ], [ 298.0625 , 4.591254029230105e-47 ], [ 298.125 , 4.591877982184911e-47 ], [ 298.1875 , 4.592502825576275e-47 ], 
                    [ 298.25 , 4.59312775399366e-47 ], [ 298.3125 , 4.593752767448668e-47 ], [ 298.375 , 4.594377865952938e-47 ], [ 298.4375 , 4.595003049517878e-47 ], 
                    [ 298.5 , 4.595628318155225e-47 ], [ 298.5625 , 4.596253671876428e-47 ], [ 298.625 , 4.596879110693159e-47 ], [ 298.6875 , 4.597504634617001e-47 ], 
                    [ 298.75 , 4.598130243659467e-47 ], [ 298.8125 , 4.598756205968388e-47 ], [ 298.875 , 4.599382892288742e-47 ], [ 298.9375 , 4.600009664009459e-47 ], 
                    [ 299.0 , 4.600636521142306e-47 ], [ 299.0625 , 4.601263463698824e-47 ], [ 299.125 , 4.60189049169072e-47 ], [ 299.1875 , 4.602517605129603e-47 ], 
                    [ 299.25 , 4.603144804027086e-47 ], [ 299.3125 , 4.603772088394814e-47 ], [ 299.375 , 4.604399458244532e-47 ], [ 299.4375 , 4.6050269135878225e-47 ], 
                    [ 299.5 , 4.605654869224994e-47 ], [ 299.5625 , 4.606283398426501e-47 ], [ 299.625 , 4.606912013402706e-47 ], [ 299.6875 , 4.607540714165413e-47 ], 
                    [ 299.75 , 4.608169500726296e-47 ], [ 299.8125 , 4.608798373097031e-47 ], [ 299.875 , 4.609427331289394e-47 ], [ 299.9375 , 4.610056375315066e-47 ], 
                    [ 300.0 , 4.610685505185758e-47 ], [ 300.0625 , 4.611314720913218e-47 ], [ 300.125 , 4.6119440225091e-47 ], [ 300.1875 , 4.612573936375692e-47 ], 
                    [ 300.25 , 4.613204288893871e-47 ], [ 300.3125 , 4.613834727555774e-47 ], [ 300.375 , 4.61446525237314e-47 ], [ 300.4375 , 4.61509586335781e-47 ], 
                    [ 300.5 , 4.615726560521525e-47 ], [ 300.5625 , 4.616357343876032e-47 ], [ 300.625 , 4.616988213433239e-47 ], [ 300.6875 , 4.61761916920473e-47 ], 
                    [ 300.75 , 4.618250211202386e-47 ], [ 300.8125 , 4.618881339438023e-47 ], [ 300.875 , 4.619513200204516e-47 ], [ 300.9375 , 4.62014540702985e-47 ], 
                    [ 301.0 , 4.620777700376207e-47 ], [ 301.0625 , 4.621410080255563e-47 ], [ 301.125 , 4.622042546679757e-47 ], [ 301.1875 , 4.622675099660601e-47 ], 
                    [ 301.25 , 4.623307739209877e-47 ], [ 301.3125 , 4.623940465339562e-47 ], [ 301.375 , 4.624573278061407e-47 ], [ 301.4375 , 4.625206177387294e-47 ], 
                    [ 301.5 , 4.625839163329113e-47 ], [ 301.5625 , 4.626472942250016e-47 ], [ 301.625 , 4.627106980747497e-47 ], [ 301.6875 , 4.627741106137333e-47 ], 
                    [ 301.75 , 4.628375318431331e-47 ], [ 301.8125 , 4.629009617641468e-47 ], [ 301.875 , 4.629644003779622e-47 ], [ 301.9375 , 4.630278476857673e-47 ], 
                    [ 302.0 , 4.6309130368876e-47 ], [ 302.0625 , 4.631547683881291e-47 ], [ 302.125 , 4.632182417850659e-47 ], [ 302.1875 , 4.632817238807561e-47 ], 
                    [ 302.25 , 4.633452915545108e-47 ], [ 302.3125 , 4.634088795616718e-47 ], [ 302.375 , 4.63472476295445e-47 ], [ 302.4375 , 4.635360817570213e-47 ], 
                    [ 302.5 , 4.635996959476118e-47 ], [ 302.5625 , 4.636633188684077e-47 ], [ 302.625 , 4.63726950520604e-47 ], [ 302.6875 , 4.637905909054087e-47 ], 
                    [ 302.75 , 4.638542400240072e-47 ], [ 302.8125 , 4.639178978776046e-47 ], [ 302.875 , 4.639815644674028e-47 ], [ 302.9375 , 4.640453190220427e-47 ], 
                    [ 303.0 , 4.641090902427431e-47 ], [ 303.0625 , 4.6417287022718165e-47 ], [ 303.125 , 4.642366589765528e-47 ], [ 303.1875 , 4.64300456492061e-47 ], 
                    [ 303.25 , 4.643642627749176e-47 ], [ 303.3125 , 4.644280778263274e-47 ], [ 303.375 , 4.644919016474855e-47 ], [ 303.4375 , 4.645557342396039e-47 ], 
                    [ 303.5 , 4.646195756038909e-47 ], [ 303.5625 , 4.646834257415488e-47 ], [ 303.625 , 4.647473654700238e-47 ], [ 303.6875 , 4.648113205689724e-47 ], 
                    [ 303.75 , 4.648752844689558e-47 ], [ 303.8125 , 4.649392571711721e-47 ], [ 303.875 , 4.650032386768421e-47 ], [ 303.9375 , 4.650672289871744e-47 ], 
                    [ 304.0 , 4.651312281033738e-47 ], [ 304.0625 , 4.651952360266686e-47 ], [ 304.125 , 4.65259252758261e-47 ], [ 304.1875 , 4.653232782993565e-47 ], 
                    [ 304.25 , 4.653873126511706e-47 ], [ 304.3125 , 4.654514362071882e-47 ], [ 304.375 , 4.655155760177473e-47 ], [ 304.4375 , 4.655797246668553e-47 ], 
                    [ 304.5 , 4.656438821557302e-47 ], [ 304.5625 , 4.657080484855902e-47 ], [ 304.625 , 4.657722236576534e-47 ], [ 304.6875 , 4.658364076731386e-47 ], 
                    [ 304.75 , 4.659006005332642e-47 ], [ 304.8125 , 4.659648022392489e-47 ], [ 304.875 , 4.660290127923119e-47 ], [ 304.9375 , 4.660932321936722e-47 ], 
                    [ 305.0 , 4.661575351373767e-47 ], [ 305.0625 , 4.662218570005296e-47 ], [ 305.125 , 4.662861877390061e-47 ], [ 305.1875 , 4.663505273540376e-47 ], 
                    [ 305.25 , 4.66414875846852e-47 ], [ 305.3125 , 4.664792332186646e-47 ], [ 305.375 , 4.665435994707171e-47 ], [ 305.4375 , 4.66607974604218e-47 ], 
                    [ 305.5 , 4.666723586204062e-47 ], [ 305.5625 , 4.667367515204909e-47 ], [ 305.625 , 4.668011533057142e-47 ], [ 305.6875 , 4.668656362784033e-47 ], 
                    [ 305.75 , 4.669301438086836e-47 ], [ 305.8125 , 4.669946602520734e-47 ], [ 305.875 , 4.670591856097911e-47 ], [ 305.9375 , 4.671237198830781e-47 ], 
                    [ 306.0 , 4.6718826307316665e-47 ], [ 306.0625 , 4.672528151812819e-47 ], [ 306.125 , 4.67317376208656e-47 ], [ 306.1875 , 4.673819461565282e-47 ], 
                    [ 306.25 , 4.674465250261344e-47 ], [ 306.3125 , 4.675111128186872e-47 ], [ 306.375 , 4.675757717603029e-47 ], [ 306.4375 , 4.676404617474167e-47 ], 
                    [ 306.5 , 4.677051606845086e-47 ], [ 306.5625 , 4.6776986857282e-47 ], [ 306.625 , 4.678345854135895e-47 ], [ 306.6875 , 4.678993112080555e-47 ], 
                    [ 306.75 , 4.679640459574603e-47 ], [ 306.8125 , 4.680287896630327e-47 ], [ 306.875 , 4.680935423260284e-47 ], [ 306.9375 , 4.681583039476667e-47 ], 
                    [ 307.0 , 4.682230745292007e-47 ], [ 307.0625 , 4.682879075710513e-47 ], [ 307.125 , 4.683527812199932e-47 ], [ 307.1875 , 4.684176638561308e-47 ], 
                    [ 307.25 , 4.684825554806889e-47 ], [ 307.3125 , 4.685474560949295e-47 ], [ 307.375 , 4.686123657000846e-47 ], [ 307.4375 , 4.686772842974097e-47 ], 
                    [ 307.5 , 4.687422118881507e-47 ], [ 307.5625 , 4.688071484735466e-47 ], [ 307.625 , 4.688720940548469e-47 ], [ 307.6875 , 4.689370486332946e-47 ], 
                    [ 307.75 , 4.690020541907992e-47 ], [ 307.8125 , 4.690671117797384e-47 ], [ 307.875 , 4.691321783931367e-47 ], [ 307.9375 , 4.691972540322428e-47 ], 
                    [ 308.0 , 4.692623386983151e-47 ], [ 308.0625 , 4.693274323926025e-47 ], [ 308.125 , 4.693925351163576e-47 ], [ 308.1875 , 4.694576468708325e-47 ], 
                    [ 308.25 , 4.695227676572802e-47 ], [ 308.3125 , 4.695878974769535e-47 ], [ 308.375 , 4.696530363311054e-47 ], [ 308.4375 , 4.697182122067375e-47 ], 
                    [ 308.5 , 4.697834529187098e-47 ], [ 308.5625 , 4.698487026921767e-47 ], [ 308.625 , 4.699139615284031e-47 ], [ 308.6875 , 4.699792294286478e-47 ], 
                    [ 308.75 , 4.700445063941665e-47 ], [ 308.8125 , 4.701097924262149e-47 ], [ 308.875 , 4.701750875260557e-47 ], [ 308.9375 , 4.702403916949484e-47 ], 
                    [ 309.0 , 4.70305704934149e-47 ], [ 309.0625 , 4.703710272449243e-47 ], [ 309.125 , 4.704363711087377e-47 ], [ 309.1875 , 4.705017953620518e-47 ], 
                    [ 309.25 , 4.705672287140096e-47 ], [ 309.3125 , 4.706326711658764e-47 ], [ 309.375 , 4.706981227189179e-47 ], [ 309.4375 , 4.707635833743963e-47 ], 
                    [ 309.5 , 4.708290531335809e-47 ], [ 309.5625 , 4.708945319977477e-47 ], [ 309.625 , 4.709600199681432e-47 ], [ 309.6875 , 4.710255170460535e-47 ], 
                    [ 309.75 , 4.710910232327219e-47 ], [ 309.8125 , 4.711565385294353e-47 ], [ 309.875 , 4.712221415251597e-47 ], [ 309.9375 , 4.712877588148224e-47 ], 
                    [ 310.0 , 4.713533852416479e-47 ], [ 310.0625 , 4.714190208068918e-47 ], [ 310.125 , 4.714846655118369e-47 ], [ 310.1875 , 4.715503193577521e-47 ], 
                    [ 310.25 , 4.716159823459072e-47 ], [ 310.3125 , 4.716816544775752e-47 ], [ 310.375 , 4.717473357540394e-47 ], [ 310.4375 , 4.718130261765565e-47 ], 
                    [ 310.5 , 4.718787257464235e-47 ], [ 310.5625 , 4.719444919931641e-47 ], [ 310.625 , 4.720102917647326e-47 ], [ 310.6875 , 4.720761007102791e-47 ], 
                    [ 310.75 , 4.721419188310826e-47 ], [ 310.8125 , 4.722077461284357e-47 ], [ 310.875 , 4.722735826036011e-47 ], [ 310.9375 , 4.723394282578651e-47 ], 
                    [ 311.0 , 4.724052830925076e-47 ], [ 311.0625 , 4.724711471088083e-47 ], [ 311.125 , 4.725370203080441e-47 ], [ 311.1875 , 4.726029026915021e-47 ], 
                    [ 311.25 , 4.726688305997141e-47 ], [ 311.3125 , 4.727348137515304e-47 ], [ 311.375 , 4.728008061144037e-47 ], [ 311.4375 , 4.728668076896061e-47 ], 
                    [ 311.5 , 4.7293281847843065e-47 ], [ 311.5625 , 4.729988384821701e-47 ], [ 311.625 , 4.730648677020975e-47 ], [ 311.6875 , 4.731309061395027e-47 ], 
                    [ 311.75 , 4.731969537956927e-47 ], [ 311.8125 , 4.7326301067191715e-47 ], [ 311.875 , 4.733290767694937e-47 ], [ 311.9375 , 4.733951648182624e-47 ], 
                    [ 312.0 , 4.734613319000835e-47 ], [ 312.0625 , 4.735275082301671e-47 ], [ 312.125 , 4.735936938098024e-47 ], [ 312.1875 , 4.736598886402923e-47 ], 
                    [ 312.25 , 4.737260927229198e-47 ], [ 312.3125 , 4.737923060589813e-47 ], [ 312.375 , 4.738585286497738e-47 ], [ 312.4375 , 4.739247604965769e-47 ], 
                    [ 312.5 , 4.739910016007083e-47 ], [ 312.5625 , 4.740572519634383e-47 ], [ 312.625 , 4.741235115860744e-47 ], [ 312.6875 , 4.741898498177295e-47 ], 
                    [ 312.75 , 4.742562103695841e-47 ], [ 312.8125 , 4.743225802082803e-47 ], [ 312.875 , 4.74388959335111e-47 ], [ 312.9375 , 4.744553477513694e-47 ], 
                    [ 313.0 , 4.745217454583621e-47 ], [ 313.0625 , 4.745881524573962e-47 ], [ 313.125 , 4.746545687497652e-47 ], [ 313.1875 , 4.74720994336763e-47 ], 
                    [ 313.25 , 4.747874292197004e-47 ], [ 313.3125 , 4.748538733998782e-47 ], [ 313.375 , 4.749203667908244e-47 ], [ 313.4375 , 4.749869088273215e-47 ], 
                    [ 313.5 , 4.750534601871489e-47 ], [ 313.5625 , 4.751200208716228e-47 ], [ 313.625 , 4.751865908820466e-47 ], [ 313.6875 , 4.752531702197269e-47 ], 
                    [ 313.75 , 4.753197588859773e-47 ], [ 313.8125 , 4.753863568820845e-47 ], [ 313.875 , 4.754529642093761e-47 ], [ 313.9375 , 4.755195808691561e-47 ], 
                    [ 314.0 , 4.755862068627252e-47 ], [ 314.0625 , 4.756528535933594e-47 ], [ 314.125 , 4.75719579807252e-47 ], [ 314.1875 , 4.757863153817268e-47 ], 
                    [ 314.25 , 4.758530603180969e-47 ], [ 314.3125 , 4.759198146176757e-47 ], [ 314.375 , 4.759865782817799e-47 ], [ 314.4375 , 4.760533513117166e-47 ], 
                    [ 314.5 , 4.761201337087996e-47 ], [ 314.5625 , 4.761869254743531e-47 ], [ 314.625 , 4.762537266096778e-47 ], [ 314.6875 , 4.763205371160983e-47 ], 
                    [ 314.75 , 4.763873569949326e-47 ], [ 314.8125 , 4.764542467966953e-47 ], [ 314.875 , 4.765211661505598e-47 ], [ 314.9375 , 4.765880949034446e-47 ], 
                    [ 315.0 , 4.766550330566631e-47 ], [ 315.0625 , 4.767219806115356e-47 ], [ 315.125 , 4.767889375693791e-47 ], [ 315.1875 , 4.768559039315211e-47 ], 
                    [ 315.25 , 4.769228796992825e-47 ], [ 315.3125 , 4.769898648739708e-47 ], [ 315.375 , 4.770568594569208e-47 ], [ 315.4375 , 4.771238634494607e-47 ], 
                    [ 315.5 , 4.771909032850799e-47 ], [ 315.5625 , 4.772580050458723e-47 ], [ 315.625 , 4.77325116242392e-47 ], [ 315.6875 , 4.773922368759753e-47 ], 
                    [ 315.75 , 4.774593669479497e-47 ], [ 315.8125 , 4.775265064596389e-47 ], [ 315.875 , 4.77593655412367e-47 ], [ 315.9375 , 4.776608138074647e-47 ], 
                    [ 316.0 , 4.7772798164626e-47 ], [ 316.0625 , 4.777951589300876e-47 ], [ 316.125 , 4.778623456602655e-47 ], [ 316.1875 , 4.779295418381151e-47 ], 
                    [ 316.25 , 4.779968182148562e-47 ], [ 316.3125 , 4.780641124871324e-47 ], [ 316.375 , 4.781314162333679e-47 ], [ 316.4375 , 4.781987294548929e-47 ], 
                    [ 316.5 , 4.782660521530277e-47 ], [ 316.5625 , 4.783333843291202e-47 ], [ 316.625 , 4.784007259845047e-47 ], [ 316.6875 , 4.784680771205158e-47 ], 
                    [ 316.75 , 4.785354377384744e-47 ], [ 316.8125 , 4.786028078397293e-47 ], [ 316.875 , 4.7867018742561545e-47 ], [ 316.9375 , 4.787376104795996e-47 ], 
                    [ 317.0 , 4.788050887502092e-47 ], [ 317.0625 , 4.788725765319064e-47 ], [ 317.125 , 4.789400738260387e-47 ], [ 317.1875 , 4.790075806339434e-47 ], 
                    [ 317.25 , 4.7907509695695805e-47 ], [ 317.3125 , 4.791426227964308e-47 ], [ 317.375 , 4.792101581536959e-47 ], [ 317.4375 , 4.792777030301019e-47 ], 
                    [ 317.5 , 4.793452574269837e-47 ], [ 317.5625 , 4.794128213456865e-47 ], [ 317.625 , 4.79480394787556e-47 ], [ 317.6875 , 4.795480504798947e-47 ], 
                    [ 317.75 , 4.796157209184961e-47 ], [ 317.8125 , 4.796834009062759e-47 ], [ 317.875 , 4.797510904445714e-47 ], [ 317.9375 , 4.798187895347371e-47 ], 
                    [ 318.0 , 4.798864981781277e-47 ], [ 318.0625 , 4.799542163760744e-47 ], [ 318.125 , 4.800219441299389e-47 ], [ 318.1875 , 4.800896814410698e-47 ], 
                    [ 318.25 , 4.801574283108157e-47 ], [ 318.3125 , 4.802251847405152e-47 ], [ 318.375 , 4.80292982702195e-47 ], [ 318.4375 , 4.803608354802077e-47 ], 
                    [ 318.5 , 4.804286978440348e-47 ], [ 318.5625 , 4.80496569795031e-47 ], [ 318.625 , 4.805644513345468e-47 ], [ 318.6875 , 4.80632342463944e-47 ], 
                    [ 318.75 , 4.807002431845669e-47 ], [ 318.8125 , 4.807681534977946e-47 ], [ 318.875 , 4.808360734049479e-47 ], [ 318.9375 , 4.809040029074063e-47 ], 
                    [ 319.0 , 4.809719420065218e-47 ], [ 319.0625 , 4.810398907036435e-47 ], [ 319.125 , 4.8110791675118145e-47 ], [ 319.1875 , 4.811759626656976e-47 ], 
                    [ 319.25 , 4.812440182043432e-47 ], [ 319.3125 , 4.813120833684929e-47 ], [ 319.375 , 4.813801581594844e-47 ], [ 319.4375 , 4.814482425786964e-47 ], 
                    [ 319.5 , 4.815163366274973e-47 ], [ 319.5625 , 4.81584440307232e-47 ], [ 319.625 , 4.816525536192663e-47 ], [ 319.6875 , 4.817206765649658e-47 ], 
                    [ 319.75 , 4.817888091456964e-47 ], [ 319.8125 , 4.81856974287101e-47 ], [ 319.875 , 4.81925203311121e-47 ], [ 319.9375 , 4.819934419960986e-47 ], 
                    [ 320.0 , 4.820616903433912e-47 ], [ 320.0625 , 4.821299483543841e-47 ], [ 320.125 , 4.82198216030439e-47 ], [ 320.1875 , 4.822664933729175e-47 ], 
                    [ 320.25 , 4.823347803832019e-47 ], [ 320.3125 , 4.824030770626578e-47 ], [ 320.375 , 4.824713834126475e-47 ], [ 320.4375 , 4.825396994345436e-47 ], 
                    [ 320.5 , 4.826080251297088e-47 ], [ 320.5625 , 4.826764135990061e-47 ], [ 320.625 , 4.827448345715543e-47 ], [ 320.6875 , 4.82813265243001e-47 ], 
                    [ 320.75 , 4.828817056147243e-47 ], [ 320.8125 , 4.829501556881028e-47 ], [ 320.875 , 4.830186154644945e-47 ], [ 320.9375 , 4.830870849452852e-47 ], 
                    [ 321.0 , 4.831555641318538e-47 ], [ 321.0625 , 4.832240530255694e-47 ], [ 321.125 , 4.832925516278183e-47 ], [ 321.1875 , 4.833610599399664e-47 ], 
                    [ 321.25 , 4.834295837822667e-47 ], [ 321.3125 , 4.834981887444414e-47 ], [ 321.375 , 4.83566803442555e-47 ], [ 321.4375 , 4.836354278779924e-47 ], 
                    [ 321.5 , 4.837040620521218e-47 ], [ 321.5625 , 4.837727059663494e-47 ], [ 321.625 , 4.838413596220298e-47 ], [ 321.6875 , 4.839100230205729e-47 ], 
                    [ 321.75 , 4.839786961633339e-47 ], [ 321.8125 , 4.840473790517266e-47 ], [ 321.875 , 4.841160716871101e-47 ], [ 321.9375 , 4.841847740708707e-47 ], 
                    [ 322.0 , 4.842535177599719e-47 ], [ 322.0625 , 4.843223138871281e-47 ], [ 322.125 , 4.843911197878951e-47 ], [ 322.1875 , 4.844599354636649e-47 ], 
                    [ 322.25 , 4.845287609158262e-47 ], [ 322.3125 , 4.845975961457714e-47 ], [ 322.375 , 4.846664411548792e-47 ], [ 322.4375 , 4.847352959445527e-47 ], 
                    [ 322.5 , 4.8480416051617105e-47 ], [ 322.5625 , 4.848730348711273e-47 ], [ 322.625 , 4.849419190108148e-47 ], [ 322.6875 , 4.850108129366204e-47 ], 
                    [ 322.75 , 4.850797728648141e-47 ], [ 322.8125 , 4.851487615071419e-47 ], [ 322.875 , 4.852177599611229e-47 ], [ 322.9375 , 4.852867682281488e-47 ], 
                    [ 323.0 , 4.853557863096154e-47 ], [ 323.0625 , 4.854248142069186e-47 ], [ 323.125 , 4.854938519214577e-47 ], [ 323.1875 , 4.855628994546152e-47 ], 
                    [ 323.25 , 4.856319568078013e-47 ], [ 323.3125 , 4.85701023982416e-47 ], [ 323.375 , 4.85770100979846e-47 ], [ 323.4375 , 4.858391912134145e-47 ], 
                    [ 323.5 , 4.859083637599256e-47 ], [ 323.5625 , 4.859775461550513e-47 ], [ 323.625 , 4.860467384001867e-47 ], [ 323.6875 , 4.861159404967377e-47 ], 
                    [ 323.75 , 4.861851524461068e-47 ], [ 323.8125 , 4.862543742497006e-47 ], [ 323.875 , 4.86323605908908e-47 ], [ 323.9375 , 4.8639284742516e-47 ], 
                    [ 324.0 , 4.86462098799822e-47 ], [ 324.0625 , 4.865313600343356e-47 ], [ 324.125 , 4.866006311300804e-47 ], [ 324.1875 , 4.866699354009226e-47 ], 
                    [ 324.25 , 4.867393001980488e-47 ], [ 324.3125 , 4.868086748817046e-47 ], [ 324.375 , 4.868780594532959e-47 ], [ 324.4375 , 4.869474539142352e-47 ], 
                    [ 324.5 , 4.870168582659184e-47 ], [ 324.5625 , 4.870862725097759e-47 ], [ 324.625 , 4.871556966472003e-47 ], [ 324.6875 , 4.872251306796086e-47 ], 
                    [ 324.75 , 4.872945746084113e-47 ], [ 324.8125 , 4.873640284350223e-47 ], [ 324.875 , 4.874334921608488e-47 ], [ 324.9375 , 4.875030065022589e-47 ], 
                    [ 325.0 , 4.875725628979246e-47 ], [ 325.0625 , 4.876421292178173e-47 ], [ 325.125 , 4.877117054633633e-47 ], [ 325.1875 , 4.87781291635965e-47 ], 
                    [ 325.25 , 4.87850887737056e-47 ], [ 325.3125 , 4.879204937680392e-47 ], [ 325.375 , 4.879901097303278e-47 ], [ 325.4375 , 4.880597356253492e-47 ], 
                    [ 325.5 , 4.881293714545206e-47 ], [ 325.5625 , 4.881990172192525e-47 ], [ 325.625 , 4.882686729209727e-47 ], [ 325.6875 , 4.883383962830062e-47 ], 
                    [ 325.75 , 4.884081462810658e-47 ], [ 325.8125 , 4.884779062415999e-47 ], [ 325.875 , 4.885476761660453e-47 ], [ 325.9375 , 4.886174560558184e-47 ], 
                    [ 326.0 , 4.886872459123355e-47 ], [ 326.0625 , 4.887570457370271e-47 ], [ 326.125 , 4.888268555313239e-47 ], [ 326.1875 , 4.88896675296636e-47 ], 
                    [ 326.25 , 4.88966505034405e-47 ], [ 326.3125 , 4.890363447460379e-47 ], [ 326.375 , 4.891061944329626e-47 ], [ 326.4375 , 4.891761240072877e-47 ], 
                    [ 326.5 , 4.892460660347244e-47 ], [ 326.5625 , 4.893160180624151e-47 ], [ 326.625 , 4.893859800917967e-47 ], [ 326.6875 , 4.894559521242956e-47 ], 
                    [ 326.75 , 4.895259341613386e-47 ], [ 326.8125 , 4.895959262043596e-47 ], [ 326.875 , 4.896659282547963e-47 ], [ 326.9375 , 4.897359403140656e-47 ], 
                    [ 327.0 , 4.89805962383602e-47 ], [ 327.0625 , 4.898759944648507e-47 ], [ 327.125 , 4.899460460497829e-47 ], [ 327.1875 , 4.900161714001511e-47 ], 
                    [ 327.25 , 4.900863067874743e-47 ], [ 327.3125 , 4.901564522131823e-47 ], [ 327.375 , 4.902266076787152e-47 ], [ 327.4375 , 4.902967731855137e-47 ], 
                    [ 327.5 , 4.903669487350112e-47 ], [ 327.5625 , 4.904371343286419e-47 ], [ 327.625 , 4.905073299678502e-47 ], [ 327.6875 , 4.90577535654067e-47 ], 
                    [ 327.75 , 4.906477513887339e-47 ], [ 327.8125 , 4.907179771732889e-47 ], [ 327.875 , 4.907882319664219e-47 ], [ 327.9375 , 4.908585490288039e-47 ], 
                    [ 328.0 , 4.90928876165774e-47 ], [ 328.0625 , 4.909992133787756e-47 ], [ 328.125 , 4.910695606692488e-47 ], [ 328.1875 , 4.911399180386444e-47 ], 
                    [ 328.25 , 4.912102854883996e-47 ], [ 328.3125 , 4.912806630199724e-47 ], [ 328.375 , 4.913510506347796e-47 ], [ 328.4375 , 4.914214483343005e-47 ], 
                    [ 328.5 , 4.914918561199522e-47 ], [ 328.5625 , 4.915622739931903e-47 ], [ 328.625 , 4.916327286027919e-47 ], [ 328.6875 , 4.917032375099271e-47 ], 
                    [ 328.75 , 4.917737565293013e-47 ], [ 328.8125 , 4.918442856623615e-47 ], [ 328.875 , 4.919148249105475e-47 ], [ 328.9375 , 4.919853742753346e-47 ], 
                    [ 329.0 , 4.920559337581492e-47 ], [ 329.0625 , 4.921265033604598e-47 ], [ 329.125 , 4.921970830837039e-47 ], [ 329.1875 , 4.922676729293397e-47 ], 
                    [ 329.25 , 4.923382728988299e-47 ], [ 329.3125 , 4.924088829936049e-47 ], [ 329.375 , 4.924795357710999e-47 ], [ 329.4375 , 4.92550237082258e-47 ], 
                    [ 329.5 , 4.9262094854343255e-47 ], [ 329.5625 , 4.926916701560807e-47 ], [ 329.625 , 4.927624019216634e-47 ], [ 329.6875 , 4.928331438416346e-47 ], 
                    [ 329.75 , 4.929038959174522e-47 ], [ 329.8125 , 4.929746581505672e-47 ], [ 329.875 , 4.930454305424515e-47 ], [ 329.9375 , 4.931162130945535e-47 ], 
                    [ 330.0 , 4.931870058083385e-47 ], [ 330.0625 , 4.932578086852549e-47 ], [ 330.125 , 4.933286585409416e-47 ], [ 330.1875 , 4.933995532867784e-47 ], 
                    [ 330.25 , 4.934704582206884e-47 ], [ 330.3125 , 4.935413733441216e-47 ], [ 330.375 , 4.936122986585493e-47 ], [ 330.4375 , 4.93683234165443e-47 ], 
                    [ 330.5 , 4.9375417986625e-47 ], [ 330.5625 , 4.938251357624526e-47 ], [ 330.625 , 4.938961018555091e-47 ], [ 330.6875 , 4.939670781468814e-47 ], 
                    [ 330.75 , 4.940380646380348e-47 ], [ 330.8125 , 4.941090613304529e-47 ], [ 330.875 , 4.941801061234859e-47 ], [ 330.9375 , 4.942511929650664e-47 ], 
                    [ 331.0 , 4.943222900323535e-47 ], [ 331.0625 , 4.943933973268146e-47 ], [ 331.125 , 4.944645148499175e-47 ], [ 331.1875 , 4.945356426031404e-47 ], 
                    [ 331.25 , 4.946067805879516e-47 ], [ 331.3125 , 4.946779288058156e-47 ], [ 331.375 , 4.947490872582151e-47 ], [ 331.4375 , 4.948202559466258e-47 ], 
                    [ 331.5 , 4.948914348725061e-47 ], [ 331.5625 , 4.949626240373428e-47 ], [ 331.625 , 4.950338609917397e-47 ], [ 331.6875 , 4.951051395365679e-47 ], 
                    [ 331.75 , 4.95176428344588e-47 ], [ 331.8125 , 4.952477274172846e-47 ], [ 331.875 , 4.953190367561394e-47 ], [ 331.9375 , 4.953903563626268e-47 ], 
                    [ 332.0 , 4.95461686238222e-47 ], [ 332.0625 , 4.955330263844141e-47 ], [ 332.125 , 4.956043768026678e-47 ], [ 332.1875 , 4.956757374944692e-47 ], 
                    [ 332.25 , 4.957471084613011e-47 ], [ 332.3125 , 4.958184897046394e-47 ], [ 332.375 , 4.958899177966433e-47 ], [ 332.4375 , 4.959613900060402e-47 ], 
                    [ 332.5 , 4.960328725166683e-47 ], [ 332.5625 , 4.961043653300124e-47 ], [ 332.625 , 4.961758684475575e-47 ], [ 332.6875 , 4.96247381870792e-47 ], 
                    [ 332.75 , 4.963189056011873e-47 ], [ 332.8125 , 4.963904396402537e-47 ], [ 332.875 , 4.964619839894557e-47 ], [ 332.9375 , 4.965335386502899e-47 ], 
                    [ 333.0 , 4.96605103624239e-47 ], [ 333.0625 , 4.96676678912786e-47 ], [ 333.125 , 4.967482962235051e-47 ], [ 333.1875 , 4.968199600939307e-47 ], 
                    [ 333.25 , 4.968916343030204e-47 ], [ 333.3125 , 4.969633188522518e-47 ], [ 333.375 , 4.970350137431271e-47 ], [ 333.4375 , 4.971067189771312e-47 ], 
                    [ 333.5 , 4.971784345557669e-47 ], [ 333.5625 , 4.972501604805089e-47 ], [ 333.625 , 4.973218967528639e-47 ], [ 333.6875 , 4.973936433743282e-47 ], 
                    [ 333.75 , 4.974654003463878e-47 ], [ 333.8125 , 4.975371676705289e-47 ], [ 333.875 , 4.976089717357479e-47 ], [ 333.9375 , 4.9768082804457385e-47 ], 
                    [ 334.0 , 4.97752694729685e-47 ], [ 334.0625 , 4.978245717925621e-47 ], [ 334.125 , 4.97896459234718e-47 ], [ 334.1875 , 4.979683570576476e-47 ], 
                    [ 334.25 , 4.980402652628503e-47 ], [ 334.3125 , 4.981121838518252e-47 ], [ 334.375 , 4.981841128260681e-47 ], [ 334.4375 , 4.982560521870894e-47 ], 
                    [ 334.5 , 4.983280019363784e-47 ], [ 334.5625 , 4.983999620754422e-47 ], [ 334.625 , 4.984719513352937e-47 ], [ 334.6875 , 4.985439993920883e-47 ], 
                    [ 334.75 , 4.986160578625532e-47 ], [ 334.8125 , 4.986881267481934e-47 ], [ 334.875 , 4.987602060505142e-47 ], [ 334.9375 , 4.98832295771018e-47 ], 
                    [ 335.0 , 4.989043959112138e-47 ], [ 335.0625 , 4.989765064726149e-47 ], [ 335.125 , 4.990486274567169e-47 ], [ 335.1875 , 4.991207588650263e-47 ], 
                    [ 335.25 , 4.991929006990568e-47 ], [ 335.3125 , 4.992650529603154e-47 ], [ 335.375 , 4.993372254784693e-47 ], [ 335.4375 , 4.994094668611785e-47 ], 
                    [ 335.5 , 4.9948171869538e-47 ], [ 335.5625 , 4.995539809825787e-47 ], [ 335.625 , 4.996262537242904e-47 ], [ 335.6875 , 4.996985369220242e-47 ], 
                    [ 335.75 , 4.997708305773033e-47 ], [ 335.8125 , 4.998431346916373e-47 ], [ 335.875 , 4.999154492665249e-47 ], [ 335.9375 , 4.999877743034869e-47 ], 
                    [ 336.0 , 5.000601098040506e-47 ], [ 336.0625 , 5.001324557697126e-47 ], [ 336.125 , 5.002048122020043e-47 ], [ 336.1875 , 5.002772440478671e-47 ], 
                    [ 336.25 , 5.00349687599752e-47 ], [ 336.3125 , 5.004221416419599e-47 ], [ 336.375 , 5.00494606176003e-47 ], [ 336.4375 , 5.005670812034078e-47 ], 
                    [ 336.5 , 5.006395667256899e-47 ], [ 336.5625 , 5.007120627443695e-47 ], [ 336.625 , 5.007845692609627e-47 ], [ 336.6875 , 5.008570862769932e-47 ], 
                    [ 336.75 , 5.0092961379398865e-47 ], [ 336.8125 , 5.010021518134554e-47 ], [ 336.875 , 5.010747003369248e-47 ], [ 336.9375 , 5.011473123324697e-47 ], 
                    [ 337.0 , 5.012199488033973e-47 ], [ 337.0625 , 5.01292595802281e-47 ], [ 337.125 , 5.013652533306468e-47 ], [ 337.1875 , 5.014379213900171e-47 ], 
                    [ 337.25 , 5.01510599981922e-47 ], [ 337.3125 , 5.015832891078916e-47 ], [ 337.375 , 5.016559887694491e-47 ], [ 337.4375 , 5.017286989681146e-47 ], 
                    [ 337.5 , 5.018014197054257e-47 ], [ 337.5625 , 5.018741509829137e-47 ], [ 337.625 , 5.019468928020921e-47 ], [ 337.6875 , 5.020196825757158e-47 ], 
                    [ 337.75 , 5.020925106831846e-47 ], [ 337.8125 , 5.021653493558361e-47 ], [ 337.875 , 5.0223819859521e-47 ], [ 337.9375 , 5.023110584028394e-47 ], 
                    [ 338.0 , 5.023839287802646e-47 ], [ 338.0625 , 5.024568097290044e-47 ], [ 338.125 , 5.025297012505963e-47 ], [ 338.1875 , 5.026026033465882e-47 ], 
                    [ 338.25 , 5.026755160184997e-47 ], [ 338.3125 , 5.027484392678618e-47 ], [ 338.375 , 5.028213730962268e-47 ], [ 338.4375 , 5.028943389496099e-47 ], 
                    [ 338.5 , 5.029673592371673e-47 ], [ 338.5625 , 5.030403901272712e-47 ], [ 338.625 , 5.031134316214718e-47 ], [ 338.6875 , 5.0318648372130165e-47 ], 
                    [ 338.75 , 5.032595464283007e-47 ], [ 338.8125 , 5.033326197440018e-47 ], [ 338.875 , 5.0340570366996355e-47 ], [ 338.9375 , 5.034787982077084e-47 ], 
                    [ 339.0 , 5.03551903358788e-47 ], [ 339.0625 , 5.036250191247362e-47 ], [ 339.125 , 5.036981455071016e-47 ], [ 339.1875 , 5.037712860860016e-47 ], 
                    [ 339.25 , 5.038444983577179e-47 ], [ 339.3125 , 5.0391772126925625e-47 ], [ 339.375 , 5.039909548221662e-47 ], [ 339.4375 , 5.040641990179872e-47 ], 
                    [ 339.5 , 5.04137453858273e-47 ], [ 339.5625 , 5.042107193445635e-47 ], [ 339.625 , 5.04283995478413e-47 ], [ 339.6875 , 5.043572822613618e-47 ], 
                    [ 339.75 , 5.044305796949646e-47 ], [ 339.8125 , 5.045038877807584e-47 ], [ 339.875 , 5.045772065203059e-47 ], [ 339.9375 , 5.046505359151444e-47 ], 
                    [ 340.0 , 5.047239250771586e-47 ], [ 340.0625 , 5.047973409023719e-47 ], [ 340.125 , 5.04870767406456e-47 ], [ 340.1875 , 5.049442045909679e-47 ], 
                    [ 340.25 , 5.050176524574645e-47 ], [ 340.3125 , 5.050911110074997e-47 ], [ 340.375 , 5.051645802426169e-47 ], [ 340.4375 , 5.052380601643736e-47 ], 
                    [ 340.5 , 5.053115507743389e-47 ], [ 340.5625 , 5.053850520740493e-47 ], [ 340.625 , 5.054585640650706e-47 ], [ 340.6875 , 5.055320867489507e-47 ], 
                    [ 340.75 , 5.056056470022733e-47 ], [ 340.8125 , 5.056792542358316e-47 ], [ 340.875 , 5.057528721853039e-47 ], [ 340.9375 , 5.058265008522468e-47 ], 
                    [ 341.0 , 5.059001402382167e-47 ], [ 341.0625 , 5.059737903447851e-47 ], [ 341.125 , 5.060474511734983e-47 ], [ 341.1875 , 5.061211227259244e-47 ], 
                    [ 341.25 , 5.061948050036355e-47 ], [ 341.3125 , 5.062684980081714e-47 ], [ 341.375 , 5.063422017411044e-47 ], [ 341.4375 , 5.064159162039964e-47 ], 
                    [ 341.5 , 5.064896457540392e-47 ], [ 341.5625 , 5.065634449998141e-47 ], [ 341.625 , 5.066372549986753e-47 ], [ 341.6875 , 5.067110757521968e-47 ], 
                    [ 341.75 , 5.06784907261942e-47 ], [ 341.8125 , 5.0685874952948175e-47 ], [ 341.875 , 5.069326025563728e-47 ], [ 341.9375 , 5.070064663441971e-47 ], 
                    [ 342.0 , 5.070803408945084e-47 ], [ 342.0625 , 5.071542262088892e-47 ], [ 342.125 , 5.072281222888934e-47 ], [ 342.1875 , 5.073020291361042e-47 ], 
                    [ 342.25 , 5.07375946752076e-47 ], [ 342.3125 , 5.074499184010311e-47 ], [ 342.375 , 5.075239208242601e-47 ], [ 342.4375 , 5.075979340394046e-47 ], 
                    [ 342.5 , 5.076719580480427e-47 ], [ 342.5625 , 5.077459928517479e-47 ], [ 342.625 , 5.078200384520983e-47 ], [ 342.6875 , 5.078940948506648e-47 ], 
                    [ 342.75 , 5.079681620490149e-47 ], [ 342.8125 , 5.080422400487343e-47 ], [ 342.875 , 5.08116328851391e-47 ], [ 342.9375 , 5.081904284585748e-47 ], 
                    [ 343.0 , 5.082645388718362e-47 ], [ 343.0625 , 5.083386769460714e-47 ], [ 343.125 , 5.084128711312005e-47 ], [ 343.1875 , 5.08487076145289e-47 ], 
                    [ 343.25 , 5.085612919899067e-47 ], [ 343.3125 , 5.086355186666489e-47 ], [ 343.375 , 5.087097561770856e-47 ], [ 343.4375 , 5.087840045227944e-47 ], 
                    [ 343.5 , 5.08858263705375e-47 ], [ 343.5625 , 5.089325337263944e-47 ], [ 343.625 , 5.09006814587442e-47 ], [ 343.6875 , 5.090811062900925e-47 ], 
                    [ 343.75 , 5.091554088359321e-47 ], [ 343.8125 , 5.092297222265395e-47 ], [ 343.875 , 5.093040983873305e-47 ], [ 343.9375 , 5.0937849604173905e-47 ], 
                    [ 344.0 , 5.094529045639433e-47 ], [ 344.0625 , 5.095273239555196e-47 ], [ 344.125 , 5.096017542180669e-47 ], [ 344.1875 , 5.09676195353173e-47 ], 
                    [ 344.25 , 5.09750647362419e-47 ], [ 344.3125 , 5.09825110247397e-47 ], [ 344.375 , 5.098995840097064e-47 ], [ 344.4375 , 5.099740686509073e-47 ], 
                    [ 344.5 , 5.100485641726179e-47 ], [ 344.5625 , 5.101230705764129e-47 ], [ 344.625 , 5.101976098591164e-47 ], [ 344.6875 , 5.102721989793286e-47 ], 
                    [ 344.75 , 5.103467990042185e-47 ], [ 344.8125 , 5.104214099353588e-47 ], [ 344.875 , 5.104960317743728e-47 ], [ 344.9375 , 5.105706645228335e-47 ], 
                    [ 345.0 , 5.106453081823431e-47 ], [ 345.0625 , 5.107199627544967e-47 ], [ 345.125 , 5.107946282408933e-47 ], [ 345.1875 , 5.108693046431249e-47 ], 
                    [ 345.25 , 5.109439919627838e-47 ], [ 345.3125 , 5.110186902014769e-47 ], [ 345.375 , 5.110933993607824e-47 ], [ 345.4375 , 5.111681728753401e-47 ], 
                    [ 345.5 , 5.112429657757226e-47 ], [ 345.5625 , 5.113177696196156e-47 ], [ 345.625 , 5.113925844086239e-47 ], [ 345.6875 , 5.114674101443598e-47 ], 
                    [ 345.75 , 5.115422468284142e-47 ], [ 345.8125 , 5.116170944623926e-47 ], [ 345.875 , 5.1169195304789e-47 ], [ 345.9375 , 5.117668225865267e-47 ], 
                    [ 346.0 , 5.118417030798912e-47 ], [ 346.0625 , 5.119165945295933e-47 ], [ 346.125 , 5.11991496937229e-47 ], [ 346.1875 , 5.12066430371429e-47 ], 
                    [ 346.25 , 5.121414141234599e-47 ], [ 346.3125 , 5.122164088556346e-47 ], [ 346.375 , 5.122914145695647e-47 ], [ 346.4375 , 5.123664312668475e-47 ], 
                    [ 346.5 , 5.124414589491055e-47 ], [ 346.5625 , 5.125164976179331e-47 ], [ 346.625 , 5.125915472749571e-47 ], [ 346.6875 , 5.126666079217647e-47 ], 
                    [ 346.75 , 5.1274167955997975e-47 ], [ 346.8125 , 5.128167621912155e-47 ], [ 346.875 , 5.12891855817067e-47 ], [ 346.9375 , 5.129669604391517e-47 ], 
                    [ 347.0 , 5.130421230863441e-47 ], [ 347.0625 , 5.131173097419082e-47 ], [ 347.125 , 5.131925074161291e-47 ], [ 347.1875 , 5.132677161106178e-47 ], 
                    [ 347.25 , 5.1334293582698955e-47 ], [ 347.3125 , 5.1341816656685575e-47 ], [ 347.375 , 5.134934083318393e-47 ], [ 347.4375 , 5.135686611235522e-47 ], 
                    [ 347.5 , 5.136439249436142e-47 ], [ 347.5625 , 5.137191997936305e-47 ], [ 347.625 , 5.13794485675232e-47 ], [ 347.6875 , 5.13869782590021e-47 ], 
                    [ 347.75 , 5.139451025320474e-47 ], [ 347.8125 , 5.140204812042291e-47 ], [ 347.875 , 5.140958709319577e-47 ], [ 347.9375 , 5.141712717168511e-47 ], 
                    [ 348.0 , 5.142466835605419e-47 ], [ 348.0625 , 5.143221064646377e-47 ], [ 348.125 , 5.143975404307785e-47 ], [ 348.1875 , 5.144729854605727e-47 ], 
                    [ 348.25 , 5.145484415556389e-47 ], [ 348.3125 , 5.146239087176183e-47 ], [ 348.375 , 5.146993869481087e-47 ], [ 348.4375 , 5.147748762487625e-47 ], 
                    [ 348.5 , 5.148503766211779e-47 ], [ 348.5625 , 5.149259229400742e-47 ], [ 348.625 , 5.150015046056649e-47 ], [ 348.6875 , 5.150770973652628e-47 ], 
                    [ 348.75 , 5.151527012204854e-47 ], [ 348.8125 , 5.15228316172965e-47 ], [ 348.875 , 5.153039422243341e-47 ], [ 348.9375 , 5.15379579376211e-47 ], 
                    [ 349.0 , 5.154552276302321e-47 ], [ 349.0625 , 5.155308869880382e-47 ], [ 349.125 , 5.15606557451237e-47 ], [ 349.1875 , 5.15682239021477e-47 ], 
                    [ 349.25 , 5.157579317003774e-47 ], [ 349.3125 , 5.158336354895726e-47 ], [ 349.375 , 5.159094054993907e-47 ], [ 349.4375 , 5.15985189534952e-47 ], 
                    [ 349.5 , 5.160609847027317e-47 ], [ 349.5625 , 5.161367910043837e-47 ], [ 349.625 , 5.162126084415215e-47 ], [ 349.6875 , 5.162884370157988e-47 ], 
                    [ 349.75 , 5.163642767288445e-47 ], [ 349.8125 , 5.164401275822912e-47 ], [ 349.875 , 5.165159895777754e-47 ], [ 349.9375 , 5.165918627169445e-47 ], 
                    [ 350.0 , 5.166677470014247e-47 ], [ 350.0625 , 5.167436424328604e-47 ], [ 350.125 , 5.168195653720843e-47 ], [ 350.1875 , 5.168955412070244e-47 ], 
                    [ 350.25 , 5.169715282109023e-47 ], [ 350.3125 , 5.170475263853637e-47 ], [ 350.375 , 5.171235357320579e-47 ], [ 350.4375 , 5.171995562526091e-47 ], 
                    [ 350.5 , 5.172755879486782e-47 ], [ 350.5625 , 5.173516308219008e-47 ], [ 350.625 , 5.174276848739237e-47 ], [ 350.6875 , 5.175037501063753e-47 ], 
                    [ 350.75 , 5.175798265209216e-47 ], [ 350.8125 , 5.176559141191879e-47 ], [ 350.875 , 5.177320129028256e-47 ], [ 350.9375 , 5.178081569921439e-47 ], 
                    [ 351.0 , 5.178843365394515e-47 ], [ 351.0625 , 5.179605272942332e-47 ], [ 351.125 , 5.180367292581416e-47 ], [ 351.1875 , 5.181129424328256e-47 ], 
                    [ 351.25 , 5.181891668199384e-47 ], [ 351.3125 , 5.182654024211221e-47 ], [ 351.375 , 5.183416492380339e-47 ], [ 351.4375 , 5.184179072723165e-47 ], 
                    [ 351.5 , 5.184941765256275e-47 ], [ 351.5625 , 5.185704569996101e-47 ], [ 351.625 , 5.186467486959225e-47 ], [ 351.6875 , 5.187230516162083e-47 ], 
                    [ 351.75 , 5.187994144805658e-47 ], [ 351.8125 , 5.188757966030696e-47 ], [ 351.875 , 5.189521899712081e-47 ], [ 351.9375 , 5.190285945866368e-47 ], 
                    [ 352.0 , 5.191050104510117e-47 ], [ 352.0625 , 5.191814375659963e-47 ], [ 352.125 , 5.192578759332248e-47 ], [ 352.1875 , 5.1933432555438e-47 ], 
                    [ 352.25 , 5.194107864311038e-47 ], [ 352.3125 , 5.194872585650607e-47 ], [ 352.375 , 5.195637419578971e-47 ], [ 352.4375 , 5.196402366112852e-47 ], 
                    [ 352.5 , 5.197167483366144e-47 ], [ 352.5625 , 5.197933213879853e-47 ], [ 352.625 , 5.198699057213402e-47 ], [ 352.6875 , 5.199465013383303e-47 ], 
                    [ 352.75 , 5.200231082406291e-47 ], [ 352.8125 , 5.200997264298809e-47 ], [ 352.875 , 5.201763559077635e-47 ], [ 352.9375 , 5.202529966759399e-47 ], 
                    [ 353.0 , 5.2032964873607395e-47 ], [ 353.0625 , 5.204063120898143e-47 ], [ 353.125 , 5.204829867388396e-47 ], [ 353.1875 , 5.205596726848181e-47 ], 
                    [ 353.25 , 5.206363699293955e-47 ], [ 353.3125 , 5.207130964559598e-47 ], [ 353.375 , 5.207898724503901e-47 ], [ 353.4375 , 5.208666597649723e-47 ], 
                    [ 353.5 , 5.209434584013757e-47 ], [ 353.5625 , 5.210202683612805e-47 ], [ 353.625 , 5.210970896463416e-47 ], [ 353.6875 , 5.211739222582398e-47 ], 
                    [ 353.75 , 5.212507661986456e-47 ], [ 353.8125 , 5.213276214692176e-47 ], [ 353.875 , 5.214044880716379e-47 ], [ 353.9375 , 5.214813660075772e-47 ], 
                    [ 354.0 , 5.215582552786955e-47 ], [ 354.0625 , 5.216351558866751e-47 ], [ 354.125 , 5.217120964691566e-47 ], [ 354.1875 , 5.217890761739768e-47 ], 
                    [ 354.25 , 5.218660672373089e-47 ], [ 354.3125 , 5.219430696608325e-47 ], [ 354.375 , 5.2202008344622e-47 ], [ 354.4375 , 5.220971085951555e-47 ], 
                    [ 354.5 , 5.221741451093156e-47 ], [ 354.5625 , 5.222511929903661e-47 ], [ 354.625 , 5.223282522399918e-47 ], [ 354.6875 , 5.224053228598811e-47 ], 
                    [ 354.75 , 5.224824048516894e-47 ], [ 354.8125 , 5.225594982171098e-47 ], [ 354.875 , 5.226366029578129e-47 ], [ 354.9375 , 5.227137550069531e-47 ], 
                    [ 355.0 , 5.227909364611981e-47 ], [ 355.0625 , 5.22868129311701e-47 ], [ 355.125 , 5.229453335601296e-47 ], [ 355.1875 , 5.230225492081818e-47 ], 
                    [ 355.25 , 5.230997762575261e-47 ], [ 355.3125 , 5.231770147098605e-47 ], [ 355.375 , 5.232542645668577e-47 ], [ 355.4375 , 5.233315258302092e-47 ], 
                    [ 355.5 , 5.234087985015843e-47 ], [ 355.5625 , 5.234860825826933e-47 ], [ 355.625 , 5.235633780751914e-47 ], [ 355.6875 , 5.236406849807932e-47 ], 
                    [ 355.75 , 5.237180465559742e-47 ], [ 355.8125 , 5.237954306911799e-47 ], [ 355.875 , 5.238728262605925e-47 ], [ 355.9375 , 5.239502332659164e-47 ], 
                    [ 356.0 , 5.240276517088264e-47 ], [ 356.0625 , 5.241050815910275e-47 ], [ 356.125 , 5.241825229141911e-47 ], [ 356.1875 , 5.242599756800304e-47 ], 
                    [ 356.25 , 5.24337439890221e-47 ], [ 356.3125 , 5.244149155464577e-47 ], [ 356.375 , 5.244924026504318e-47 ], [ 356.4375 , 5.245699012038311e-47 ], 
                    [ 356.5 , 5.246474112083623e-47 ], [ 356.5625 , 5.247249820076033e-47 ], [ 356.625 , 5.248025700433327e-47 ], [ 356.6875 , 5.24880169551562e-47 ], 
                    [ 356.75 , 5.249577805339797e-47 ], [ 356.8125 , 5.250354029922791e-47 ], [ 356.875 , 5.251130369281642e-47 ], [ 356.9375 , 5.251906823433286e-47 ], 
                    [ 357.0 , 5.25268339239466e-47 ], [ 357.0625 , 5.253460076182775e-47 ], [ 357.125 , 5.254236874814685e-47 ], [ 357.1875 , 5.255013788307263e-47 ], 
                    [ 357.25 , 5.255790816677412e-47 ], [ 357.3125 , 5.256567959942385e-47 ], [ 357.375 , 5.257345722300797e-47 ], [ 357.4375 , 5.258123617355552e-47 ], 
                    [ 357.5 , 5.258901627510312e-47 ], [ 357.5625 , 5.259679752782144e-47 ], [ 357.625 , 5.260457993188159e-47 ], [ 357.6875 , 5.261236348745239e-47 ], 
                    [ 357.75 , 5.262014819470575e-47 ], [ 357.8125 , 5.262793405381059e-47 ], [ 357.875 , 5.263572106493767e-47 ], [ 357.9375 , 5.264350922825784e-47 ], 
                    [ 358.0 , 5.265129854394273e-47 ], [ 358.0625 , 5.26590890121617e-47 ], [ 358.125 , 5.266688067457513e-47 ], [ 358.1875 , 5.267467885159017e-47 ], 
                    [ 358.25 , 5.268247818325122e-47 ], [ 358.3125 , 5.269027866972851e-47 ], [ 358.375 , 5.269808031119265e-47 ], [ 358.4375 , 5.270588310781503e-47 ], 
                    [ 358.5 , 5.271368705976743e-47 ], [ 358.5625 , 5.272149216721982e-47 ], [ 358.625 , 5.272929843034251e-47 ], [ 358.6875 , 5.273710584930924e-47 ], 
                    [ 358.75 , 5.274491442428817e-47 ], [ 358.8125 , 5.275272415545383e-47 ], [ 358.875 , 5.276053504297367e-47 ], [ 358.9375 , 5.276834718958981e-47 ], 
                    [ 359.0 , 5.277616556085914e-47 ], [ 359.0625 , 5.278398509052992e-47 ], [ 359.125 , 5.2791805778773e-47 ], [ 359.1875 , 5.279962762576158e-47 ], 
                    [ 359.25 , 5.280745063166695e-47 ], [ 359.3125 , 5.281527479665968e-47 ], [ 359.375 , 5.282310012091191e-47 ], [ 359.4375 , 5.283092660459726e-47 ], 
                    [ 359.5 , 5.283875424788489e-47 ], [ 359.5625 , 5.284658305094886e-47 ], [ 359.625 , 5.285441301395801e-47 ], [ 359.6875 , 5.286224413708758e-47 ], 
                    [ 359.75 , 5.287007643402437e-47 ], [ 359.8125 , 5.287791515881646e-47 ], [ 359.875 , 5.288575504580824e-47 ], [ 359.9375 , 5.289359609517316e-47 ], 
                    [ 360.0 , 5.290143830708165e-47 ], [ 360.0625 , 5.290928168170837e-47 ], [ 360.125 , 5.29171262192238e-47 ], [ 360.1875 , 5.292497191980149e-47 ], 
                    [ 360.25 , 5.29328187836135e-47 ], [ 360.3125 , 5.294066681083231e-47 ], [ 360.375 , 5.2948516001630005e-47 ], [ 360.4375 , 5.2956366356180265e-47 ], 
                    [ 360.5 , 5.296421787465374e-47 ], [ 360.5625 , 5.297207055722487e-47 ], [ 360.625 , 5.297992921050575e-47 ], [ 360.6875 , 5.298778925174477e-47 ], 
                    [ 360.75 , 5.29956504590912e-47 ], [ 360.8125 , 5.300351283271654e-47 ], [ 360.875 , 5.301137637279495e-47 ], [ 360.9375 , 5.301924107949948e-47 ], 
                    [ 361.0 , 5.302710695300283e-47 ], [ 361.0625 , 5.303497399347773e-47 ], [ 361.125 , 5.304284220109807e-47 ], [ 361.1875 , 5.305071157603587e-47 ], 
                    [ 361.25 , 5.30585821184662e-47 ], [ 361.3125 , 5.306645382856037e-47 ], [ 361.375 , 5.307432670649201e-47 ], [ 361.4375 , 5.308220531011126e-47 ], 
                    [ 361.5 , 5.309008570863088e-47 ], [ 361.5625 , 5.309796727704652e-47 ], [ 361.625 , 5.310585001553337e-47 ], [ 361.6875 , 5.311373392426325e-47 ], 
                    [ 361.75 , 5.312161900341026e-47 ], [ 361.8125 , 5.31295052531493e-47 ], [ 361.875 , 5.313739267365262e-47 ], [ 361.9375 , 5.314528126509444e-47 ], 
                    [ 362.0 , 5.315317102765083e-47 ], [ 362.0625 , 5.316106196149226e-47 ], [ 362.125 , 5.316895406679526e-47 ], [ 362.1875 , 5.317684734373147e-47 ], 
                    [ 362.25 , 5.318474562418582e-47 ], [ 362.3125 , 5.319264620033079e-47 ], [ 362.375 , 5.320054795010484e-47 ], [ 362.4375 , 5.320845087368004e-47 ], 
                    [ 362.5 , 5.321635497123189e-47 ], [ 362.5625 , 5.322426024293591e-47 ], [ 362.625 , 5.323216668896427e-47 ], [ 362.6875 , 5.324007430949253e-47 ], 
                    [ 362.75 , 5.324798310469594e-47 ], [ 362.8125 , 5.325589307474859e-47 ], [ 362.875 , 5.326380421982314e-47 ], [ 362.9375 , 5.327171654009676e-47 ], 
                    [ 363.0 , 5.327963003574254e-47 ], [ 363.0625 , 5.328754795002554e-47 ], [ 363.125 , 5.329546888457977e-47 ], [ 363.1875 , 5.330339099654198e-47 ], 
                    [ 363.25 , 5.33113142860887e-47 ], [ 363.3125 , 5.33192387533927e-47 ], [ 363.375 , 5.332716439863056e-47 ], [ 363.4375 , 5.333509122197738e-47 ], 
                    [ 363.5 , 5.334301922360751e-47 ], [ 363.5625 , 5.335094840369611e-47 ], [ 363.625 , 5.335887876242022e-47 ], [ 363.6875 , 5.336681029995243e-47 ], 
                    [ 363.75 , 5.337474301646906e-47 ], [ 363.8125 , 5.338267691214538e-47 ], [ 363.875 , 5.33906142731245e-47 ], [ 363.9375 , 5.339855537021654e-47 ], 
                    [ 364.0 , 5.340649764843384e-47 ], [ 364.0625 , 5.341444110795359e-47 ], [ 364.125 , 5.342238574894958e-47 ], [ 364.1875 , 5.343033157159832e-47 ], 
                    [ 364.25 , 5.343827857607517e-47 ], [ 364.3125 , 5.3446226762556665e-47 ], [ 364.375 , 5.345417613121862e-47 ], [ 364.4375 , 5.346212668223687e-47 ], 
                    [ 364.5 , 5.347007841578536e-47 ], [ 364.5625 , 5.347803133204188e-47 ], [ 364.625 , 5.348598543118313e-47 ], [ 364.6875 , 5.349394211764085e-47 ], 
                    [ 364.75 , 5.35019036572255e-47 ], [ 364.8125 , 5.350986638173184e-47 ], [ 364.875 , 5.351783029133548e-47 ], [ 364.9375 , 5.352579538621355e-47 ], 
                    [ 365.0 , 5.353376166654247e-47 ], [ 365.0625 , 5.35417291324975e-47 ], [ 365.125 , 5.35496977842559e-47 ], [ 365.1875 , 5.355766762199334e-47 ], 
                    [ 365.25 , 5.356563864588943e-47 ], [ 365.3125 , 5.357361085611648e-47 ], [ 365.375 , 5.358158425285412e-47 ], [ 365.4375 , 5.358955883627664e-47 ], 
                    [ 365.5 , 5.359753484847251e-47 ], [ 365.5625 , 5.360551648744787e-47 ], [ 365.625 , 5.361349931503274e-47 ], [ 365.6875 , 5.362148333140453e-47 ], 
                    [ 365.75 , 5.362946853674063e-47 ], [ 365.8125 , 5.363745493121813e-47 ], [ 365.875 , 5.364544251501255e-47 ], [ 365.9375 , 5.365343128830294e-47 ], 
                    [ 366.0 , 5.366142125126642e-47 ], [ 366.0625 , 5.3669412404077865e-47 ], [ 366.125 , 5.367740474691638e-47 ], [ 366.1875 , 5.368539827995994e-47 ], 
                    [ 366.25 , 5.369339300338312e-47 ], [ 366.3125 , 5.370138891736548e-47 ], [ 366.375 , 5.370938986912012e-47 ], [ 366.4375 , 5.371739299609229e-47 ], 
                    [ 366.5 , 5.372539731559298e-47 ], [ 366.5625 , 5.373340282780219e-47 ], [ 366.625 , 5.374140953289573e-47 ], [ 366.6875 , 5.37494174310525e-47 ], 
                    [ 366.75 , 5.375742652245027e-47 ], [ 366.8125 , 5.37654368072657e-47 ], [ 366.875 , 5.377344828567854e-47 ], [ 366.9375 , 5.378146095786394e-47 ], 
                    [ 367.0 , 5.378947482400249e-47 ], [ 367.0625 , 5.379748988426978e-47 ], [ 367.125 , 5.380550613884568e-47 ], [ 367.1875 , 5.381352597314246e-47 ], 
                    [ 367.25 , 5.382154928387146e-47 ], [ 367.3125 , 5.382957379083308e-47 ], [ 367.375 , 5.38375994942068e-47 ], [ 367.4375 , 5.384562639416948e-47 ], 
                    [ 367.5 , 5.385365449090066e-47 ], [ 367.5625 , 5.386168378457763e-47 ], [ 367.625 , 5.386971427538078e-47 ], [ 367.6875 , 5.387774596348703e-47 ], 
                    [ 367.75 , 5.38857788490753e-47 ], [ 367.8125 , 5.389381293232412e-47 ], [ 367.875 , 5.390184821341167e-47 ], [ 367.9375 , 5.390988469251694e-47 ], 
                    [ 368.0 , 5.391792328142158e-47 ], [ 368.0625 , 5.392596681276358e-47 ], [ 368.125 , 5.393401154404824e-47 ], [ 368.1875 , 5.394205747545422e-47 ], 
                    [ 368.25 , 5.395010460715975e-47 ], [ 368.3125 , 5.395815293934508e-47 ], [ 368.375 , 5.396620247218888e-47 ], [ 368.4375 , 5.397425320587068e-47 ], 
                    [ 368.5 , 5.398230514056845e-47 ], [ 368.5625 , 5.39903582764629e-47 ], [ 368.625 , 5.399841261373169e-47 ], [ 368.6875 , 5.400646815255598e-47 ], 
                    [ 368.75 , 5.401452489311345e-47 ], [ 368.8125 , 5.402258283558224e-47 ], [ 368.875 , 5.403064595755753e-47 ], [ 368.9375 , 5.403871098367133e-47 ], 
                    [ 369.0 , 5.404677721363203e-47 ], [ 369.0625 , 5.405484464761932e-47 ], [ 369.125 , 5.406291328581408e-47 ], [ 369.1875 , 5.407098312839489e-47 ], 
                    [ 369.25 , 5.407905417554194e-47 ], [ 369.3125 , 5.408712642743501e-47 ], [ 369.375 , 5.409519988425354e-47 ], [ 369.4375 , 5.410327454617856e-47 ], 
                    [ 369.5 , 5.41113504133888e-47 ], [ 369.5625 , 5.411942748606415e-47 ], [ 369.625 , 5.412750576438571e-47 ], [ 369.6875 , 5.413558742577489e-47 ], 
                    [ 369.75 , 5.414367270640286e-47 ], [ 369.8125 , 5.415175919458703e-47 ], [ 369.875 , 5.415984689050771e-47 ], [ 369.9375 , 5.416793579434529e-47 ], 
                    [ 370.0 , 5.41760259062798e-47 ], [ 370.0625 , 5.418411722649283e-47 ], [ 370.125 , 5.419220975516327e-47 ], [ 370.1875 , 5.420030349247242e-47 ], 
                    [ 370.25 , 5.420839843860114e-47 ], [ 370.3125 , 5.421649459372999e-47 ], [ 370.375 , 5.422459195803876e-47 ], [ 370.4375 , 5.423269053170881e-47 ], 
                    [ 370.5 , 5.424079062669527e-47 ], [ 370.5625 , 5.42488961838251e-47 ], [ 370.625 , 5.425700295222188e-47 ], [ 370.6875 , 5.426511093206547e-47 ], 
                    [ 370.75 , 5.427322012353729e-47 ], [ 370.8125 , 5.428133052682032e-47 ], [ 370.875 , 5.428944214209296e-47 ], [ 370.9375 , 5.429755496953786e-47 ], 
                    [ 371.0 , 5.430566900933501e-47 ], [ 371.0625 , 5.431378426166635e-47 ], [ 371.125 , 5.432190072671423e-47 ], [ 371.1875 , 5.433001840465794e-47 ], 
                    [ 371.25 , 5.433813729567992e-47 ], [ 371.3125 , 5.434625739996026e-47 ], [ 371.375 , 5.435438148442128e-47 ], [ 371.4375 , 5.4362508395693e-47 ], 
                    [ 371.5 , 5.437063652207642e-47 ], [ 371.5625 , 5.437876586375438e-47 ], [ 371.625 , 5.438689642090858e-47 ], [ 371.6875 , 5.439502819372e-47 ], 
                    [ 371.75 , 5.440316118237154e-47 ], [ 371.8125 , 5.441129538704306e-47 ], [ 371.875 , 5.441943080791792e-47 ], [ 371.9375 , 5.442756744517759e-47 ], 
                    [ 372.0 , 5.443570529900394e-47 ], [ 372.0625 , 5.444384436957963e-47 ], [ 372.125 , 5.445198465708504e-47 ], [ 372.1875 , 5.446012689487799e-47 ], 
                    [ 372.25 , 5.446827406644188e-47 ], [ 372.3125 , 5.447642245681239e-47 ], [ 372.375 , 5.448457206617376e-47 ], [ 372.4375 , 5.449272289470642e-47 ], 
                    [ 372.5 , 5.450087494259316e-47 ], [ 372.5625 , 5.450902821001755e-47 ], [ 372.625 , 5.451718269716047e-47 ], [ 372.6875 , 5.452533840420594e-47 ], 
                    [ 372.75 , 5.453349533133608e-47 ], [ 372.8125 , 5.454165347873225e-47 ], [ 372.875 , 5.454981284657738e-47 ], [ 372.9375 , 5.455797343505521e-47 ], 
                    [ 373.0 , 5.456613524434719e-47 ], [ 373.0625 , 5.457430112043478e-47 ], [ 373.125 , 5.458246963722921e-47 ], [ 373.1875 , 5.459063937666299e-47 ], 
                    [ 373.25 , 5.459881033891912e-47 ], [ 373.3125 , 5.460698252417947e-47 ], [ 373.375 , 5.461515593262862e-47 ], [ 373.4375 , 5.462333056444813e-47 ], 
                    [ 373.5 , 5.463150641982265e-47 ], [ 373.5625 , 5.463968349893416e-47 ], [ 373.625 , 5.464786180196582e-47 ], [ 373.6875 , 5.465604132910236e-47 ], 
                    [ 373.75 , 5.466422208052471e-47 ], [ 373.8125 , 5.467240405641764e-47 ], [ 373.875 , 5.468058787553592e-47 ], [ 373.9375 , 5.468877674543989e-47 ], 
                    [ 374.0 , 5.469696684169481e-47 ], [ 374.0625 , 5.470515816448354e-47 ], [ 374.125 , 5.471335071399094e-47 ], [ 374.1875 , 5.472154449040032e-47 ], 
                    [ 374.25 , 5.472973949389584e-47 ], [ 374.3125 , 5.473793572466122e-47 ], [ 374.375 , 5.474613318287835e-47 ], [ 374.4375 , 5.475433186873375e-47 ], 
                    [ 374.5 , 5.476253178241011e-47 ], [ 374.5625 , 5.477073292409131e-47 ], [ 374.625 , 5.477893529396048e-47 ], [ 374.6875 , 5.478713889220347e-47 ], 
                    [ 374.75 , 5.479534615518672e-47 ], [ 374.8125 , 5.480355640856563e-47 ], [ 374.875 , 5.481176789212599e-47 ], [ 374.9375 , 5.481998060605368e-47 ], 
                    [ 375.0 , 5.482819455053186e-47 ], [ 375.0625 , 5.483640972574532e-47 ], [ 375.125 , 5.484462613187847e-47 ], [ 375.1875 , 5.48528437691165e-47 ], 
                    [ 375.25 , 5.486106263764235e-47 ], [ 375.3125 , 5.486928273764125e-47 ], [ 375.375 , 5.487750406929774e-47 ], [ 375.4375 , 5.488572663279598e-47 ], 
                    [ 375.5 , 5.48939504283217e-47 ], [ 375.5625 , 5.490217545605833e-47 ], [ 375.625 , 5.491040581424339e-47 ], [ 375.6875 , 5.491863744329468e-47 ], 
                    [ 375.75 , 5.492687030635022e-47 ], [ 375.8125 , 5.493510440359577e-47 ], [ 375.875 , 5.494333973521636e-47 ], [ 375.9375 , 5.495157630139703e-47 ], 
                    [ 376.0 , 5.495981410232285e-47 ], [ 376.0625 , 5.496805313817932e-47 ], [ 376.125 , 5.497629340915077e-47 ], [ 376.1875 , 5.498453491542316e-47 ], 
                    [ 376.25 , 5.49927776571809e-47 ], [ 376.3125 , 5.5001021634609555e-47 ], [ 376.375 , 5.500926684789437e-47 ], [ 376.4375 , 5.501751487337156e-47 ], 
                    [ 376.5 , 5.502576681139121e-47 ], [ 376.5625 , 5.503401998709898e-47 ], [ 376.625 , 5.504227440067933e-47 ], [ 376.6875 , 5.50505300523187e-47 ], 
                    [ 376.75 , 5.5058786942202e-47 ], [ 376.8125 , 5.506704507051535e-47 ], [ 376.875 , 5.507530443744411e-47 ], [ 376.9375 , 5.508356504317443e-47 ], 
                    [ 377.0 , 5.509182688789333e-47 ], [ 377.0625 , 5.510008997178424e-47 ], [ 377.125 , 5.510835429503462e-47 ], [ 377.1875 , 5.511661985783036e-47 ], 
                    [ 377.25 , 5.512488666035657e-47 ], [ 377.3125 , 5.51331576191496e-47 ], [ 377.375 , 5.514143095593043e-47 ], [ 377.4375 , 5.514970553421645e-47 ], 
                    [ 377.5 , 5.515798135419358e-47 ], [ 377.5625 , 5.516625841604852e-47 ], [ 377.625 , 5.517453671996766e-47 ], [ 377.6875 , 5.518281626613814e-47 ], 
                    [ 377.75 , 5.519109705474482e-47 ], [ 377.8125 , 5.519937908597492e-47 ], [ 377.875 , 5.52076623600153e-47 ], [ 377.9375 , 5.521594687705089e-47 ], 
                    [ 378.0 , 5.522423263727135e-47 ], [ 378.0625 , 5.523251964085972e-47 ], [ 378.125 , 5.524080803689047e-47 ], [ 378.1875 , 5.524910150504563e-47 ], 
                    [ 378.25 , 5.525739621832442e-47 ], [ 378.3125 , 5.526569217691341e-47 ], [ 378.375 , 5.527398938099912e-47 ], [ 378.4375 , 5.528228783076897e-47 ], 
                    [ 378.5 , 5.529058752640997e-47 ], [ 378.5625 , 5.529888846810878e-47 ], [ 378.625 , 5.530719065605324e-47 ], [ 378.6875 , 5.531549409043046e-47 ], 
                    [ 378.75 , 5.532379877142758e-47 ], [ 378.8125 , 5.5332104699230575e-47 ], [ 378.875 , 5.534041187402742e-47 ], [ 378.9375 , 5.534872029600611e-47 ], 
                    [ 379.0 , 5.535703127720991e-47 ], [ 379.0625 , 5.536534623986526e-47 ], [ 379.125 , 5.537366245147817e-47 ], [ 379.1875 , 5.53819799122366e-47 ], 
                    [ 379.25 , 5.539029862232861e-47 ], [ 379.3125 , 5.539861858194185e-47 ], [ 379.375 , 5.540693979126282e-47 ], [ 379.4375 , 5.541526225047962e-47 ], 
                    [ 379.5 , 5.542358595978158e-47 ], [ 379.5625 , 5.543191091935488e-47 ], [ 379.625 , 5.544023712938734e-47 ], [ 379.6875 , 5.544856459006795e-47 ], 
                    [ 379.75 , 5.545689330158299e-47 ], [ 379.8125 , 5.546522326412114e-47 ], [ 379.875 , 5.547355675216429e-47 ], [ 379.9375 , 5.54818931383585e-47 ], 
                    [ 380.0 , 5.5490230777317e-47 ], [ 380.0625 , 5.549856966922924e-47 ], [ 380.125 , 5.55069098142835e-47 ], [ 380.1875 , 5.551525121266692e-47 ], 
                    [ 380.25 , 5.552359386456982e-47 ], [ 380.3125 , 5.553193777017861e-47 ], [ 380.375 , 5.554028292968208e-47 ], [ 380.4375 , 5.554862934326943e-47 ], 
                    [ 380.5 , 5.555697701112876e-47 ], [ 380.5625 , 5.556532593344893e-47 ], [ 380.625 , 5.557367611041808e-47 ], [ 380.6875 , 5.558202754222434e-47 ], 
                    [ 380.75 , 5.55903833614106e-47 ], [ 380.8125 , 5.559874119660366e-47 ], [ 380.875 , 5.560710028837073e-47 ], [ 380.9375 , 5.561546063689996e-47 ], 
                    [ 381.0 , 5.562382224238106e-47 ], [ 381.0625 , 5.563218510500186e-47 ], [ 381.125 , 5.56405492249529e-47 ], [ 381.1875 , 5.564891460242207e-47 ], 
                    [ 381.25 , 5.565728123759881e-47 ], [ 381.3125 , 5.566564913067221e-47 ], [ 381.375 , 5.56740182818318e-47 ], [ 381.4375 , 5.568238869126515e-47 ], 
                    [ 381.5 , 5.569076035916378e-47 ], [ 381.5625 , 5.569913328571535e-47 ], [ 381.625 , 5.570751134920189e-47 ], [ 381.6875 , 5.571589067972515e-47 ], 
                    [ 381.75 , 5.5724271270638225e-47 ], [ 381.8125 , 5.57326531221315e-47 ], [ 381.875 , 5.574103623439259e-47 ], [ 381.9375 , 5.574942060761314e-47 ], 
                    [ 382.0 , 5.5757806241981605e-47 ], [ 382.0625 , 5.576619313768809e-47 ], [ 382.125 , 5.577458129492272e-47 ], [ 382.1875 , 5.578297071387485e-47 ], 
                    [ 382.25 , 5.579136139473386e-47 ], [ 382.3125 , 5.579975333768958e-47 ], [ 382.375 , 5.580814654293301e-47 ], [ 382.4375 , 5.581654159284675e-47 ], 
                    [ 382.5 , 5.582494095169093e-47 ], [ 382.5625 , 5.583334157448432e-47 ], [ 382.625 , 5.584174346141631e-47 ], [ 382.6875 , 5.585014661267676e-47 ], 
                    [ 382.75 , 5.585855102845588e-47 ], [ 382.8125 , 5.586695670894481e-47 ], [ 382.875 , 5.587536365433423e-47 ], [ 382.9375 , 5.58837718648125e-47 ], 
                    [ 383.0 , 5.589218134057159e-47 ], [ 383.0625 , 5.590059208180229e-47 ], [ 383.125 , 5.590900408869304e-47 ], [ 383.1875 , 5.59174173614359e-47 ], 
                    [ 383.25 , 5.592583190022136e-47 ], [ 383.3125 , 5.593424880550299e-47 ], [ 383.375 , 5.594266965605732e-47 ], [ 383.4375 , 5.59510917743635e-47 ], 
                    [ 383.5 , 5.595951516061115e-47 ], [ 383.5625 , 5.596793981499239e-47 ], [ 383.625 , 5.597636573769773e-47 ], [ 383.6875 , 5.598479292891809e-47 ], 
                    [ 383.75 , 5.599322138884409e-47 ], [ 383.8125 , 5.60016511176671e-47 ], [ 383.875 , 5.601008211557934e-47 ], [ 383.9375 , 5.601851438276991e-47 ], 
                    [ 384.0 , 5.602694791942987e-47 ], [ 384.0625 , 5.603538272575194e-47 ], [ 384.125 , 5.604381880192687e-47 ], [ 384.1875 , 5.60522575746853e-47 ], 
                    [ 384.25 , 5.606069984445388e-47 ], [ 384.3125 , 5.606914338574795e-47 ], [ 384.375 , 5.607758819876102e-47 ], [ 384.4375 , 5.608603428368302e-47 ], 
                    [ 384.5 , 5.609448164070591e-47 ], [ 384.5625 , 5.610293027002131e-47 ], [ 384.625 , 5.611138017182043e-47 ], [ 384.6875 , 5.611983134629573e-47 ], 
                    [ 384.75 , 5.612828379363849e-47 ], [ 384.8125 , 5.613673751404002e-47 ], [ 384.875 , 5.614519250769286e-47 ], [ 384.9375 , 5.61536487747884e-47 ], 
                    [ 385.0 , 5.616210631551882e-47 ], [ 385.0625 , 5.617056674063741e-47 ], [ 385.125 , 5.617903034342721e-47 ], [ 385.1875 , 5.618749522148622e-47 ], 
                    [ 385.25 , 5.619596137500579e-47 ], [ 385.3125 , 5.620442880417852e-47 ], [ 385.375 , 5.621289750919703e-47 ], [ 385.4375 , 5.622136749025233e-47 ], 
                    [ 385.5 , 5.62298387475383e-47 ], [ 385.5625 , 5.623831128124604e-47 ], [ 385.625 , 5.62467850915691e-47 ], [ 385.6875 , 5.625526017869861e-47 ], 
                    [ 385.75 , 5.626373654282654e-47 ], [ 385.8125 , 5.627221418414614e-47 ], [ 385.875 , 5.628069310285024e-47 ], [ 385.9375 , 5.628917510042103e-47 ], 
                    [ 386.0 , 5.629766021749768e-47 ], [ 386.0625 , 5.630614661363439e-47 ], [ 386.125 , 5.631463428902273e-47 ], [ 386.1875 , 5.632312324385715e-47 ], 
                    [ 386.25 , 5.633161347832932e-47 ], [ 386.3125 , 5.634010499263253e-47 ], [ 386.375 , 5.634859778696009e-47 ], [ 386.4375 , 5.635709186150379e-47 ], 
                    [ 386.5 , 5.636558721645856e-47 ], [ 386.5625 , 5.637408385201466e-47 ], [ 386.625 , 5.638258176836831e-47 ], [ 386.6875 , 5.639108096570937e-47 ], 
                    [ 386.75 , 5.639958144423375e-47 ], [ 386.8125 , 5.640808489755039e-47 ], [ 386.875 , 5.64165913320469e-47 ], [ 386.9375 , 5.642509904932771e-47 ], 
                    [ 387.0 , 5.643360804958666e-47 ], [ 387.0625 , 5.644211833301762e-47 ], [ 387.125 , 5.645062989981249e-47 ], [ 387.1875 , 5.645914275016722e-47 ], 
                    [ 387.25 , 5.646765688427416e-47 ], [ 387.3125 , 5.6476172302326915e-47 ], [ 387.375 , 5.648468900451749e-47 ], [ 387.4375 , 5.649320699104355e-47 ], 
                    [ 387.5 , 5.650172626209437e-47 ], [ 387.5625 , 5.651024681786685e-47 ], [ 387.625 , 5.6518768658552725e-47 ], [ 387.6875 , 5.652729344606457e-47 ], 
                    [ 387.75 , 5.653582137852921e-47 ], [ 387.8125 , 5.654435059755087e-47 ], [ 387.875 , 5.655288110332527e-47 ], [ 387.9375 , 5.656141289604572e-47 ], 
                    [ 388.0 , 5.656994597590597e-47 ], [ 388.0625 , 5.657848034310062e-47 ], [ 388.125 , 5.658701599782346e-47 ], [ 388.1875 , 5.659555294026994e-47 ], 
                    [ 388.25 , 5.660409117063355e-47 ], [ 388.3125 , 5.661263068910817e-47 ], [ 388.375 , 5.662117149588852e-47 ], [ 388.4375 , 5.662971359116979e-47 ], 
                    [ 388.5 , 5.663825697514473e-47 ], [ 388.5625 , 5.664680306405582e-47 ], [ 388.625 , 5.665535245442572e-47 ], [ 388.6875 , 5.666390313510689e-47 ], 
                    [ 388.75 , 5.667245510629565e-47 ], [ 388.8125 , 5.668100836818718e-47 ], [ 388.875 , 5.668956292097428e-47 ], [ 388.9375 , 5.669811876485336e-47 ], 
                    [ 389.0 , 5.67066759000189e-47 ], [ 389.0625 , 5.671523432666618e-47 ], [ 389.125 , 5.672379404498971e-47 ], [ 389.1875 , 5.673235505518364e-47 ], 
                    [ 389.25 , 5.674091735744495e-47 ], [ 389.3125 , 5.674948095196581e-47 ], [ 389.375 , 5.675804583894411e-47 ], [ 389.4375 , 5.67666130480643e-47 ], 
                    [ 389.5 , 5.677518371438847e-47 ], [ 389.5625 , 5.678375567471755e-47 ], [ 389.625 , 5.679232892924728e-47 ], [ 389.6875 , 5.680090347817391e-47 ], 
                    [ 389.75 , 5.680947932169326e-47 ], [ 389.8125 , 5.681805645999874e-47 ], [ 389.875 , 5.68266348932883e-47 ], [ 389.9375 , 5.6835214621755e-47 ], 
                    [ 390.0 , 5.684379564559604e-47 ], [ 390.0625 , 5.685237796500737e-47 ], [ 390.125 , 5.686096158018299e-47 ], [ 390.1875 , 5.686954649131975e-47 ], 
                    [ 390.25 , 5.68781326986125e-47 ], [ 390.3125 , 5.688672084543517e-47 ], [ 390.375 , 5.689531286725405e-47 ], [ 390.4375 , 5.69039061867895e-47 ], 
                    [ 390.5 , 5.691250080423789e-47 ], [ 390.5625 , 5.692109671979487e-47 ], [ 390.625 , 5.692969393365609e-47 ], [ 390.6875 , 5.693829244601805e-47 ], 
                    [ 390.75 , 5.694689225707727e-47 ], [ 390.8125 , 5.69554933670287e-47 ], [ 390.875 , 5.696409577607094e-47 ], [ 390.9375 , 5.697269948439657e-47 ], 
                    [ 391.0 , 5.698130449220546e-47 ], [ 391.0625 , 5.698991079969187e-47 ], [ 391.125 , 5.699851840705167e-47 ], [ 391.1875 , 5.700712746554551e-47 ], 
                    [ 391.25 , 5.70157410457079e-47 ], [ 391.3125 , 5.7024355927351725e-47 ], [ 391.375 , 5.703297211067482e-47 ], [ 391.4375 , 5.704158959587347e-47 ], 
                    [ 391.5 , 5.705020838314398e-47 ], [ 391.5625 , 5.70588284726843e-47 ], [ 391.625 , 5.706744986468959e-47 ], [ 391.6875 , 5.707607255935785e-47 ], 
                    [ 391.75 , 5.708469655688632e-47 ], [ 391.8125 , 5.709332185746983e-47 ], [ 391.875 , 5.710194846130647e-47 ], [ 391.9375 , 5.711057636859439e-47 ], 
                    [ 392.0 , 5.711920557952851e-47 ], [ 392.0625 , 5.712783609430662e-47 ], [ 392.125 , 5.713647053051515e-47 ], [ 392.1875 , 5.714510670471611e-47 ], 
                    [ 392.25 , 5.715374418427348e-47 ], [ 392.3125 , 5.716238296938661e-47 ], [ 392.375 , 5.717102306024997e-47 ], [ 392.4375 , 5.717966445706379e-47 ], 
                    [ 392.5 , 5.718830716002383e-47 ], [ 392.5625 , 5.719695116932832e-47 ], [ 392.625 , 5.72055964851731e-47 ], [ 392.6875 , 5.721424310775849e-47 ], 
                    [ 392.75 , 5.722289103727917e-47 ], [ 392.8125 , 5.72315402739347e-47 ], [ 392.875 , 5.724019081792144e-47 ], [ 392.9375 , 5.72488426694374e-47 ], 
                    [ 393.0 , 5.725749781229896e-47 ], [ 393.0625 , 5.726615538429483e-47 ], [ 393.125 , 5.7274814265352445e-47 ], [ 393.1875 , 5.72834744556685e-47 ], 
                    [ 393.25 , 5.7292135955441385e-47 ], [ 393.3125 , 5.730079876486908e-47 ], [ 393.375 , 5.730946288415044e-47 ], [ 393.4375 , 5.731812831348189e-47 ], 
                    [ 393.5 , 5.732679505306231e-47 ], [ 393.5625 , 5.733546310308984e-47 ], [ 393.625 , 5.73441324637622e-47 ], [ 393.6875 , 5.73528031352792e-47 ], 
                    [ 393.75 , 5.736147511783743e-47 ], [ 393.8125 , 5.737014841163468e-47 ], [ 393.875 , 5.737882423077063e-47 ], [ 393.9375 , 5.738750327819881e-47 ], 
                    [ 394.0 , 5.739618363840776e-47 ], [ 394.0625 , 5.740486531159769e-47 ], [ 394.125 , 5.741354829796598e-47 ], [ 394.1875 , 5.742223259771166e-47 ], 
                    [ 394.25 , 5.743091821103338e-47 ], [ 394.3125 , 5.743960513813027e-47 ], [ 394.375 , 5.744829337919979e-47 ], [ 394.4375 , 5.745698293444193e-47 ], 
                    [ 394.5 , 5.746567380405547e-47 ], [ 394.5625 , 5.747436598823841e-47 ], [ 394.625 , 5.748305948719039e-47 ], [ 394.6875 , 5.7491754301109075e-47 ], 
                    [ 394.75 , 5.750045072981756e-47 ], [ 394.8125 , 5.750915104690287e-47 ], [ 394.875 , 5.751785268042042e-47 ], [ 394.9375 , 5.752655563057146e-47 ], 
                    [ 395.0 , 5.753525989755352e-47 ], [ 395.0625 , 5.754396548156752e-47 ], [ 395.125 , 5.755267238281192e-47 ], [ 395.1875 , 5.756138060148519e-47 ], 
                    [ 395.25 , 5.757009013778874e-47 ], [ 395.3125 , 5.757880099191947e-47 ], [ 395.375 , 5.758751316407883e-47 ], [ 395.4375 , 5.759622665446459e-47 ], 
                    [ 395.5 , 5.76049414632783e-47 ], [ 395.5625 , 5.761365759071736e-47 ], [ 395.625 , 5.762237503698253e-47 ], [ 395.6875 , 5.763109618182447e-47 ], 
                    [ 395.75 , 5.763981928505326e-47 ], [ 395.8125 , 5.764854370861969e-47 ], [ 395.875 , 5.765726945272526e-47 ], [ 395.9375 , 5.766599651756822e-47 ], 
                    [ 396.0 , 5.767472490334805e-47 ], [ 396.0625 , 5.768345461026674e-47 ], [ 396.125 , 5.769218563852263e-47 ], [ 396.1875 , 5.77009179883153e-47 ], 
                    [ 396.25 , 5.770965165984684e-47 ], [ 396.3125 , 5.771838665331484e-47 ], [ 396.375 , 5.772712296892186e-47 ], [ 396.4375 , 5.773586060686596e-47 ], 
                    [ 396.5 , 5.774459956734852e-47 ], [ 396.5625 , 5.775334110660664e-47 ], [ 396.625 , 5.776208556698004e-47 ], [ 396.6875 , 5.777083135135586e-47 ], 
                    [ 396.75 , 5.77795784599354e-47 ], [ 396.8125 , 5.778832689291955e-47 ], [ 396.875 , 5.779707665050721e-47 ], [ 396.9375 , 5.7805827732899345e-47 ], 
                    [ 397.0 , 5.781458014029737e-47 ], [ 397.0625 , 5.782333387290189e-47 ], [ 397.125 , 5.783208893091235e-47 ], [ 397.1875 , 5.784084531453105e-47 ], 
                    [ 397.25 , 5.784960302395747e-47 ], [ 397.3125 , 5.785836205939278e-47 ], [ 397.375 , 5.786712242103855e-47 ], [ 397.4375 , 5.787588426862189e-47 ], 
                    [ 397.5 , 5.788464997118874e-47 ], [ 397.5625 , 5.789341700138223e-47 ], [ 397.625 , 5.790218535940302e-47 ], [ 397.6875 , 5.791095504545221e-47 ], 
                    [ 397.75 , 5.791972605973096e-47 ], [ 397.8125 , 5.792849840244041e-47 ], [ 397.875 , 5.793727207378179e-47 ], [ 397.9375 , 5.794604707395673e-47 ], 
                    [ 398.0 , 5.795482340316442e-47 ], [ 398.0625 , 5.796360106160946e-47 ], [ 398.125 , 5.797238004949151e-47 ], [ 398.1875 , 5.798116036701112e-47 ], 
                    [ 398.25 , 5.798994201437132e-47 ], [ 398.3125 , 5.799872499177227e-47 ], [ 398.375 , 5.800751116225182e-47 ], [ 398.4375 , 5.801629968577941e-47 ], 
                    [ 398.5 , 5.802508954082551e-47 ], [ 398.5625 , 5.803388072759389e-47 ], [ 398.625 , 5.80426732462851e-47 ], [ 398.6875 , 5.805146709710092e-47 ], 
                    [ 398.75 , 5.806026228024442e-47 ], [ 398.8125 , 5.8069058795916625e-47 ], [ 398.875 , 5.807785664431861e-47 ], [ 398.9375 , 5.808665582565392e-47 ], 
                    [ 399.0 , 5.80954563401237e-47 ], [ 399.0625 , 5.81042581879291e-47 ], [ 399.125 , 5.811306136927377e-47 ], [ 399.1875 , 5.812186588435852e-47 ], 
                    [ 399.25 , 5.81306722641394e-47 ], [ 399.3125 , 5.813948210581597e-47 ], [ 399.375 , 5.814829328264458e-47 ], [ 399.4375 , 5.815710579482759e-47 ], 
                    [ 399.5 , 5.816591964256821e-47 ], [ 399.5625 , 5.8174734826068e-47 ], [ 399.625 , 5.818355134553024e-47 ], [ 399.6875 , 5.819236920115698e-47 ], 
                    [ 399.75 , 5.820118839315072e-47 ], [ 399.8125 , 5.821000892171398e-47 ], [ 399.875 , 5.821883078704933e-47 ], [ 399.9375 , 5.822765398935977e-47 ], 
                    [ 400.0 , 5.823647852884711e-47 ], [ 400.0625 , 5.824530440571357e-47 ], [ 400.125 , 5.825413162016431e-47 ], [ 400.1875 , 5.826296206566663e-47 ], 
                    [ 400.25 , 5.827179458319567e-47 ], [ 400.3125 , 5.828062843971075e-47 ], [ 400.375 , 5.828946363541694e-47 ], [ 400.4375 , 5.82983001705156e-47 ], 
                    [ 400.5 , 5.8307138045211845e-47 ], [ 400.5625 , 5.83159772597067e-47 ], [ 400.625 , 5.832481781420408e-47 ], [ 400.6875 , 5.833365970890673e-47 ], 
                    [ 400.75 , 5.834250294401947e-47 ], [ 400.8125 , 5.835134751974343e-47 ], [ 400.875 , 5.836019343628268e-47 ], [ 400.9375 , 5.836904069384048e-47 ], 
                    [ 401.0 , 5.837788929262012e-47 ], [ 401.0625 , 5.838673979210222e-47 ], [ 401.125 , 5.839559374885466e-47 ], [ 401.1875 , 5.8404449048250575e-47 ], 
                    [ 401.25 , 5.841330569049313e-47 ], [ 401.3125 , 5.842216367578597e-47 ], [ 401.375 , 5.843102300433275e-47 ], [ 401.4375 , 5.843988367633718e-47 ], 
                    [ 401.5 , 5.844874569200298e-47 ], [ 401.5625 , 5.84576090515339e-47 ], [ 401.625 , 5.8466473755133745e-47 ], [ 401.6875 , 5.84753398030059e-47 ], 
                    [ 401.75 , 5.848420719535506e-47 ], [ 401.8125 , 5.849307593238427e-47 ], [ 401.875 , 5.850194601429909e-47 ], [ 401.9375 , 5.85108174413014e-47 ], 
                    [ 402.0 , 5.85196918488938e-47 ], [ 402.0625 , 5.852856842773315e-47 ], [ 402.125 , 5.853744635301971e-47 ], [ 402.1875 , 5.854632562495607e-47 ], 
                    [ 402.25 , 5.855520624374899e-47 ], [ 402.3125 , 5.856408820960066e-47 ], [ 402.375 , 5.857297152271669e-47 ], [ 402.4375 , 5.858185618330101e-47 ], 
                    [ 402.5 , 5.859074219155759e-47 ], [ 402.5625 , 5.859962954769211e-47 ], [ 402.625 , 5.86085182519082e-47 ], [ 402.6875 , 5.861740830440949e-47 ], 
                    [ 402.75 , 5.862629970540134e-47 ], [ 402.8125 , 5.86351924550883e-47 ], [ 402.875 , 5.864408677123237e-47 ], [ 402.9375 , 5.865298471883699e-47 ], 
                    [ 403.0 , 5.866188401650946e-47 ], [ 403.0625 , 5.867078466445379e-47 ], [ 403.125 , 5.867968666287526e-47 ], [ 403.1875 , 5.868859001197921e-47 ], 
                    [ 403.25 , 5.86974947119693e-47 ], [ 403.3125 , 5.870640076305261e-47 ], [ 403.375 , 5.871530816543204e-47 ], [ 403.4375 , 5.872421691931304e-47 ], 
                    [ 403.5 , 5.873312702490191e-47 ], [ 403.5625 , 5.874203848240251e-47 ], [ 403.625 , 5.875095129202036e-47 ], [ 403.6875 , 5.875986545396147e-47 ], 
                    [ 403.75 , 5.876878096842892e-47 ], [ 403.8125 , 5.877769896480467e-47 ], [ 403.875 , 5.878661955530118e-47 ], [ 403.9375 , 5.879554149965954e-47 ], 
                    [ 404.0 , 5.880446479808693e-47 ], [ 404.0625 , 5.881338945078802e-47 ], [ 404.125 , 5.882231545796791e-47 ], [ 404.1875 , 5.883124281983258e-47 ], 
                    [ 404.25 , 5.884017153658764e-47 ], [ 404.3125 , 5.884910160843956e-47 ], [ 404.375 , 5.885803303559233e-47 ], [ 404.4375 , 5.886696581825288e-47 ], 
                    [ 404.5 , 5.88758999566261e-47 ], [ 404.5625 , 5.88848354509186e-47 ], [ 404.625 , 5.889377230133531e-47 ], [ 404.6875 , 5.890271050808247e-47 ], 
                    [ 404.75 , 5.891165197540362e-47 ], [ 404.8125 , 5.89205951823746e-47 ], [ 404.875 , 5.892953974698705e-47 ], [ 404.9375 , 5.893848566944873e-47 ], 
                    [ 405.0 , 5.894743294996453e-47 ], [ 405.0625 , 5.895638158874104e-47 ], [ 405.125 , 5.896533158598527e-47 ], [ 405.1875 , 5.897428294190136e-47 ], 
                    [ 405.25 , 5.898323565669681e-47 ], [ 405.3125 , 5.8992189730578765e-47 ], [ 405.375 , 5.900114516375271e-47 ], [ 405.4375 , 5.901010195642455e-47 ], 
                    [ 405.5 , 5.901906010880111e-47 ], [ 405.5625 , 5.902801962108837e-47 ], [ 405.625 , 5.903698086107102e-47 ], [ 405.6875 , 5.904594545289528e-47 ], 
                    [ 405.75 , 5.905491140596601e-47 ], [ 405.8125 , 5.906387872049075e-47 ], [ 405.875 , 5.90728473966758e-47 ], [ 405.9375 , 5.908181743472754e-47 ], 
                    [ 406.0 , 5.909078883485315e-47 ], [ 406.0625 , 5.909976159725989e-47 ], [ 406.125 , 5.910873572215421e-47 ], [ 406.1875 , 5.91177112097434e-47 ], 
                    [ 406.25 , 5.912668806023312e-47 ], [ 406.3125 , 5.913566627383203e-47 ], [ 406.375 , 5.914464585074498e-47 ], [ 406.4375 , 5.915362679118112e-47 ], 
                    [ 406.5 , 5.916260909534662e-47 ], [ 406.5625 , 5.917159380535643e-47 ], [ 406.625 , 5.918058117632921e-47 ], [ 406.6875 , 5.918956991236345e-47 ], 
                    [ 406.75 , 5.919856001366562e-47 ], [ 406.8125 , 5.920755148044312e-47 ], [ 406.875 , 5.921654431290459e-47 ], [ 406.9375 , 5.922553851125662e-47 ], 
                    [ 407.0 , 5.923453407570667e-47 ], [ 407.0625 , 5.924353100646098e-47 ], [ 407.125 , 5.925252930372999e-47 ], [ 407.1875 , 5.926152896771876e-47 ], 
                    [ 407.25 , 5.927052999863613e-47 ], [ 407.3125 , 5.927953239668887e-47 ], [ 407.375 , 5.928853616208506e-47 ], [ 407.4375 , 5.929754129503238e-47 ], 
                    [ 407.5 , 5.930654925946493e-47 ], [ 407.5625 , 5.931555921304532e-47 ], [ 407.625 , 5.932457053543258e-47 ], [ 407.6875 , 5.933358322683639e-47 ], 
                    [ 407.75 , 5.934259728746345e-47 ], [ 407.8125 , 5.935161271752176e-47 ], [ 407.875 , 5.936062951722066e-47 ], [ 407.9375 , 5.93696476867678e-47 ], 
                    [ 408.0 , 5.937866722636916e-47 ], [ 408.0625 , 5.938768813623543e-47 ], [ 408.125 , 5.939671041657436e-47 ], [ 408.1875 , 5.940573406759289e-47 ], 
                    [ 408.25 , 5.941475908950008e-47 ], [ 408.3125 , 5.942378548250592e-47 ], [ 408.375 , 5.943281324681531e-47 ], [ 408.4375 , 5.944184434710081e-47 ], 
                    [ 408.5 , 5.94508769783672e-47 ], [ 408.5625 , 5.945991098220965e-47 ], [ 408.625 , 5.946894635883547e-47 ], [ 408.6875 , 5.947798310845452e-47 ], 
                    [ 408.75 , 5.948702123127462e-47 ], [ 408.8125 , 5.949606072750525e-47 ], [ 408.875 , 5.950510159735513e-47 ], [ 408.9375 , 5.951414384103258e-47 ], 
                    [ 409.0 , 5.952318745874592e-47 ], [ 409.0625 , 5.953223245070436e-47 ], [ 409.125 , 5.954127881711675e-47 ], [ 409.1875 , 5.955032655819151e-47 ], 
                    [ 409.25 , 5.955937567413881e-47 ], [ 409.3125 , 5.956842640075656e-47 ], [ 409.375 , 5.9577480318168255e-47 ], [ 409.4375 , 5.958653561170235e-47 ], 
                    [ 409.5 , 5.959559228156885e-47 ], [ 409.5625 , 5.960465032797398e-47 ], [ 409.625 , 5.961370975112949e-47 ], [ 409.6875 , 5.962277055124382e-47 ], 
                    [ 409.75 , 5.96318327285258e-47 ], [ 409.8125 , 5.96408962831852e-47 ], [ 409.875 , 5.964996121543178e-47 ], [ 409.9375 , 5.9659027525474075e-47 ], 
                    [ 410.0 , 5.966809521352276e-47 ], [ 410.0625 , 5.967716427978518e-47 ], [ 410.125 , 5.96862347244725e-47 ], [ 410.1875 , 5.969530654779423e-47 ], 
                    [ 410.25 , 5.9704380278876605e-47 ], [ 410.3125 , 5.971345681306168e-47 ], [ 410.375 , 5.972253472710317e-47 ], [ 410.4375 , 5.973161402121047e-47 ], 
                    [ 410.5 , 5.974069469559416e-47 ], [ 410.5625 , 5.974977675046371e-47 ], [ 410.625 , 5.975886018602854e-47 ], [ 410.6875 , 5.976794500249939e-47 ], 
                    [ 410.75 , 5.977703120008536e-47 ], [ 410.8125 , 5.978611877899598e-47 ], [ 410.875 , 5.979520773944465e-47 ], [ 410.9375 , 5.98042980816367e-47 ], 
                    [ 411.0 , 5.981338980578479e-47 ], [ 411.0625 , 5.982248291209896e-47 ], [ 411.125 , 5.983157740078851e-47 ], [ 411.1875 , 5.984067406480508e-47 ], 
                    [ 411.25 , 5.984977333584e-47 ], [ 411.3125 , 5.985887399049416e-47 ], [ 411.375 , 5.986797602897881e-47 ], [ 411.4375 , 5.987707945150392e-47 ], 
                    [ 411.5 , 5.988618425827956e-47 ], [ 411.5625 , 5.989529044951745e-47 ], [ 411.625 , 5.990439802542644e-47 ], [ 411.6875 , 5.991350698621834e-47 ], 
                    [ 411.75 , 5.992261733210288e-47 ], [ 411.8125 , 5.993172906329154e-47 ], [ 411.875 , 5.994084217999453e-47 ], [ 411.9375 , 5.994995668242169e-47 ], 
                    [ 412.0 , 5.995907257078584e-47 ], [ 412.0625 , 5.996818984529477e-47 ], [ 412.125 , 5.997730943625879e-47 ], [ 412.1875 , 5.998643141474872e-47 ], 
                    [ 412.25 , 5.999555478060569e-47 ], [ 412.3125 , 6.000467953403986e-47 ], [ 412.375 , 6.001380567526227e-47 ], [ 412.4375 , 6.002293320448314e-47 ], 
                    [ 412.5 , 6.003206212191527e-47 ], [ 412.5625 , 6.004119242776851e-47 ], [ 412.625 , 6.005032412225446e-47 ], [ 412.6875 , 6.005945720558388e-47 ], 
                    [ 412.75 , 6.006859167796931e-47 ], [ 412.8125 , 6.0077727539621115e-47 ], [ 412.875 , 6.00868647907502e-47 ], [ 412.9375 , 6.009600343156873e-47 ], 
                    [ 413.0 , 6.010514346228677e-47 ], [ 413.0625 , 6.011428586601167e-47 ], [ 413.125 , 6.01234305091643e-47 ], [ 413.1875 , 6.01325765434093e-47 ], 
                    [ 413.25 , 6.014172396895784e-47 ], [ 413.3125 , 6.015087278602117e-47 ], [ 413.375 , 6.016002299481179e-47 ], [ 413.4375 , 6.016917459554014e-47 ], 
                    [ 413.5 , 6.017832758842012e-47 ], [ 413.5625 , 6.018748197366134e-47 ], [ 413.625 , 6.019663775147648e-47 ], [ 413.6875 , 6.020579492207779e-47 ], 
                    [ 413.75 , 6.021495348567629e-47 ], [ 413.8125 , 6.022411344248477e-47 ], [ 413.875 , 6.023327479271471e-47 ], [ 413.9375 , 6.024243753657679e-47 ], 
                    [ 414.0 , 6.025160265095393e-47 ], [ 414.0625 , 6.02607699346918e-47 ], [ 414.125 , 6.026993861323269e-47 ], [ 414.1875 , 6.027910868678797e-47 ], 
                    [ 414.25 , 6.028828015556986e-47 ], [ 414.3125 , 6.029745301979111e-47 ], [ 414.375 , 6.030662727966486e-47 ], [ 414.4375 , 6.031580293540261e-47 ], 
                    [ 414.5 , 6.032497998721632e-47 ], [ 414.5625 , 6.033415843531841e-47 ], [ 414.625 , 6.034333827992172e-47 ], [ 414.6875 , 6.035251952123963e-47 ], 
                    [ 414.75 , 6.036170215948377e-47 ], [ 414.8125 , 6.037088619486539e-47 ], [ 414.875 , 6.038007162760007e-47 ], [ 414.9375 , 6.038925935783574e-47 ], 
                    [ 415.0 , 6.039844923865301e-47 ], [ 415.0625 , 6.04076405179625e-47 ], [ 415.125 , 6.041683319597661e-47 ], [ 415.1875 , 6.042602727290903e-47 ], 
                    [ 415.25 , 6.043522274897223e-47 ], [ 415.3125 , 6.044441962437913e-47 ], [ 415.375 , 6.045361789934265e-47 ], [ 415.4375 , 6.046281757407581e-47 ], 
                    [ 415.5 , 6.047201864879117e-47 ], [ 415.5625 , 6.048122112370349e-47 ], [ 415.625 , 6.049042499902327e-47 ], [ 415.6875 , 6.049963027496494e-47 ], 
                    [ 415.75 , 6.050883695174162e-47 ], [ 415.8125 , 6.051804502956647e-47 ], [ 415.875 , 6.052725536541767e-47 ], [ 415.9375 , 6.053646796816473e-47 ], 
                    [ 416.0 , 6.05456819731243e-47 ], [ 416.0625 , 6.05548973805081e-47 ], [ 416.125 , 6.056411419053217e-47 ], [ 416.1875 , 6.057333240340784e-47 ], 
                    [ 416.25 , 6.058255201934994e-47 ], [ 416.3125 , 6.05917730385707e-47 ], [ 416.375 , 6.060099546128547e-47 ], [ 416.4375 , 6.061021928770697e-47 ], 
                    [ 416.5 , 6.061944451804931e-47 ], [ 416.5625 , 6.062867115252402e-47 ], [ 416.625 , 6.063789919134825e-47 ], [ 416.6875 , 6.0647128634732735e-47 ], 
                    [ 416.75 , 6.065635948289342e-47 ], [ 416.8125 , 6.066559245743917e-47 ], [ 416.875 , 6.067482784829215e-47 ], [ 416.9375 , 6.068406464508945e-47 ], 
                    [ 417.0 , 6.069330284804507e-47 ], [ 417.0625 , 6.07025424573731e-47 ], [ 417.125 , 6.071178347328763e-47 ], [ 417.1875 , 6.07210258960028e-47 ], 
                    [ 417.25 , 6.073026972573232e-47 ], [ 417.3125 , 6.073951496269085e-47 ], [ 417.375 , 6.074876160709345e-47 ], [ 417.4375 , 6.075800965915311e-47 ], 
                    [ 417.5 , 6.076725911908455e-47 ], [ 417.5625 , 6.077650998710208e-47 ], [ 417.625 , 6.078576226342007e-47 ], [ 417.6875 , 6.07950159482525e-47 ], 
                    [ 417.75 , 6.080427145898413e-47 ], [ 417.8125 , 6.081352934642117e-47 ], [ 417.875 , 6.082278864343858e-47 ], [ 417.9375 , 6.08320493502505e-47 ], 
                    [ 418.0 , 6.084131146707118e-47 ], [ 418.0625 , 6.085057499411573e-47 ], [ 418.125 , 6.08598399315993e-47 ], [ 418.1875 , 6.086910627973577e-47 ], 
                    [ 418.25 , 6.087837403874122e-47 ], [ 418.3125 , 6.088764320882788e-47 ], [ 418.375 , 6.089691379021362e-47 ], [ 418.4375 , 6.090618578311156e-47 ], 
                    [ 418.5 , 6.09154591877371e-47 ], [ 418.5625 , 6.092473400430558e-47 ], [ 418.625 , 6.093401023303114e-47 ], [ 418.6875 , 6.094328810061285e-47 ], 
                    [ 418.75 , 6.095256864772696e-47 ], [ 418.8125 , 6.096185060809892e-47 ], [ 418.875 , 6.097113398194353e-47 ], [ 418.9375 , 6.098041876947604e-47 ], 
                    [ 419.0 , 6.098970497091128e-47 ], [ 419.0625 , 6.0998992586465e-47 ], [ 419.125 , 6.100828161635298e-47 ], [ 419.1875 , 6.101757206079058e-47 ], 
                    [ 419.25 , 6.102686391999151e-47 ], [ 419.3125 , 6.103615719417292e-47 ], [ 419.375 , 6.1045451883550735e-47 ], [ 419.4375 , 6.105474798833784e-47 ], 
                    [ 419.5 , 6.106404550875327e-47 ], [ 419.5625 , 6.10733444450104e-47 ], [ 419.625 , 6.108264479732487e-47 ], [ 419.6875 , 6.109194793176754e-47 ], 
                    [ 419.75 , 6.11012525285066e-47 ], [ 419.8125 , 6.111055854238168e-47 ], [ 419.875 , 6.111986597360687e-47 ], [ 419.9375 , 6.11291748223985e-47 ], 
                    [ 420.0 , 6.113848508897289e-47 ], [ 420.0625 , 6.114779677354424e-47 ], [ 420.125 , 6.115710987633112e-47 ], [ 420.1875 , 6.11664243975478e-47 ], 
                    [ 420.25 , 6.117574033740987e-47 ], [ 420.3125 , 6.118505769613557e-47 ], [ 420.375 , 6.119437647393972e-47 ], [ 420.4375 , 6.1203696671037555e-47 ], 
                    [ 420.5 , 6.121301828764698e-47 ], [ 420.5625 , 6.122234132398247e-47 ], [ 420.625 , 6.123166688435339e-47 ], [ 420.6875 , 6.124099423440662e-47 ], 
                    [ 420.75 , 6.125032300528533e-47 ], [ 420.8125 , 6.12596531972042e-47 ], [ 420.875 , 6.126898481038058e-47 ], [ 420.9375 , 6.127831784503097e-47 ], 
                    [ 421.0 , 6.128765230137233e-47 ], [ 421.0625 , 6.129698817962122e-47 ], [ 421.125 , 6.130632547999293e-47 ], [ 421.1875 , 6.131566420270454e-47 ], 
                    [ 421.25 , 6.132500434797401e-47 ], [ 421.3125 , 6.133434591601717e-47 ], [ 421.375 , 6.134368890705031e-47 ], [ 421.4375 , 6.135303332129106e-47 ], 
                    [ 421.5 , 6.136237915895579e-47 ], [ 421.5625 , 6.137172704394769e-47 ], [ 421.625 , 6.138107698521013e-47 ], [ 421.6875 , 6.139042835092894e-47 ], 
                    [ 421.75 , 6.139978114132289e-47 ], [ 421.8125 , 6.140913535660816e-47 ], [ 421.875 , 6.141849099700139e-47 ], [ 421.9375 , 6.142784806272013e-47 ], 
                    [ 422.0 , 6.143720655398196e-47 ], [ 422.0625 , 6.144656647100275e-47 ], [ 422.125 , 6.145592781400104e-47 ], [ 422.1875 , 6.146529058319362e-47 ], 
                    [ 422.25 , 6.147465477879821e-47 ], [ 422.3125 , 6.148402040103126e-47 ], [ 422.375 , 6.149338745011143e-47 ], [ 422.4375 , 6.150275592625345e-47 ], 
                    [ 422.5 , 6.151212609143551e-47 ], [ 422.5625 , 6.15214986733383e-47 ], [ 422.625 , 6.1530872683337805e-47 ], [ 422.6875 , 6.154024812165247e-47 ], 
                    [ 422.75 , 6.154962498849952e-47 ], [ 422.8125 , 6.155900328409617e-47 ], [ 422.875 , 6.156838300866143e-47 ], [ 422.9375 , 6.157776416241129e-47 ], 
                    [ 423.0 , 6.158714674556525e-47 ], [ 423.0625 , 6.15965307583398e-47 ], [ 423.125 , 6.160591620095278e-47 ], [ 423.1875 , 6.161530307362379e-47 ], 
                    [ 423.25 , 6.162469137656768e-47 ], [ 423.3125 , 6.163408111000543e-47 ], [ 423.375 , 6.164347227415326e-47 ], [ 423.4375 , 6.165286486922961e-47 ], 
                    [ 423.5 , 6.166225999055942e-47 ], [ 423.5625 , 6.167165668284338e-47 ], [ 423.625 , 6.168105480708543e-47 ], [ 423.6875 , 6.16904543635051e-47 ], 
                    [ 423.75 , 6.16998553523202e-47 ], [ 423.8125 , 6.1709257773749e-47 ], [ 423.875 , 6.171866162800984e-47 ], [ 423.9375 , 6.172806691532104e-47 ], 
                    [ 424.0 , 6.173747363590013e-47 ], [ 424.0625 , 6.174688178996813e-47 ], [ 424.125 , 6.175629137774132e-47 ], [ 424.1875 , 6.1765702399437735e-47 ], 
                    [ 424.25 , 6.177511485527678e-47 ], [ 424.3125 , 6.178452874547786e-47 ], [ 424.375 , 6.179394407025869e-47 ], [ 424.4375 , 6.180336144121946e-47 ], 
                    [ 424.5 , 6.181278078876675e-47 ], [ 424.5625 , 6.182220157190157e-47 ], [ 424.625 , 6.18316237908414e-47 ], [ 424.6875 , 6.184104744580683e-47 ], 
                    [ 424.75 , 6.185047253701498e-47 ], [ 424.8125 , 6.185989906468603e-47 ], [ 424.875 , 6.1869327029038045e-47 ], [ 424.9375 , 6.187875643029088e-47 ], 
                    [ 425.0 , 6.188818726866263e-47 ], [ 425.0625 , 6.189761954437367e-47 ], [ 425.125 , 6.190705325764171e-47 ], [ 425.1875 , 6.191648840868543e-47 ], 
                    [ 425.25 , 6.1925924997726155e-47 ], [ 425.3125 , 6.193536302498128e-47 ], [ 425.375 , 6.194480265084819e-47 ], [ 425.4375 , 6.19542447087432e-47 ], 
                    [ 425.5 , 6.196368820586191e-47 ], [ 425.5625 , 6.197313314242418e-47 ], [ 425.625 , 6.198257951864985e-47 ], [ 425.6875 , 6.199202733475792e-47 ], 
                    [ 425.75 , 6.2001476590967865e-47 ], [ 425.8125 , 6.201092728749877e-47 ], [ 425.875 , 6.202037942457059e-47 ], [ 425.9375 , 6.202983300240293e-47 ], 
                    [ 426.0 , 6.203928802121628e-47 ], [ 426.0625 , 6.204874448122894e-47 ], [ 426.125 , 6.205820238266149e-47 ], [ 426.1875 , 6.206766172573143e-47 ], 
                    [ 426.25 , 6.207712251066203e-47 ], [ 426.3125 , 6.208658473767087e-47 ], [ 426.375 , 6.209604910051882e-47 ], [ 426.4375 , 6.210551518601674e-47 ], 
                    [ 426.5 , 6.211498271455014e-47 ], [ 426.5625 , 6.212445168633857e-47 ], [ 426.625 , 6.213392210160203e-47 ], [ 426.6875 , 6.214339396055968e-47 ], 
                    [ 426.75 , 6.215286726343293e-47 ], [ 426.8125 , 6.2162342010441905e-47 ], [ 426.875 , 6.217181820180542e-47 ], [ 426.9375 , 6.218129583774542e-47 ], 
                    [ 427.0 , 6.219077491848036e-47 ], [ 427.0625 , 6.220025544423223e-47 ], [ 427.125 , 6.220973741522092e-47 ], [ 427.1875 , 6.221922083166538e-47 ], 
                    [ 427.25 , 6.222870569378684e-47 ], [ 427.3125 , 6.223819224502147e-47 ], [ 427.375 , 6.224768096817813e-47 ], [ 427.4375 , 6.225717113796805e-47 ], 
                    [ 427.5 , 6.226666275461269e-47 ], [ 427.5625 , 6.227615581833172e-47 ], [ 427.625 , 6.228565032934622e-47 ], [ 427.6875 , 6.229514628787729e-47 ], 
                    [ 427.75 , 6.230464369414427e-47 ], [ 427.8125 , 6.231414254836967e-47 ], [ 427.875 , 6.232364285077379e-47 ], [ 427.9375 , 6.233314460157609e-47 ], 
                    [ 428.0 , 6.234264780099826e-47 ], [ 428.0625 , 6.235215244926118e-47 ], [ 428.125 , 6.236165854658617e-47 ], [ 428.1875 , 6.237116609319371e-47 ], 
                    [ 428.25 , 6.238067508930341e-47 ], [ 428.3125 , 6.239018623201527e-47 ], [ 428.375 , 6.239969905159508e-47 ], [ 428.4375 , 6.240921332162314e-47 ], 
                    [ 428.5 , 6.24187290423206e-47 ], [ 428.5625 , 6.24282462139091e-47 ], [ 428.625 , 6.243776483660896e-47 ], [ 428.6875 , 6.244728491064233e-47 ], 
                    [ 428.75 , 6.245680643623007e-47 ], [ 428.8125 , 6.246632941359348e-47 ], [ 428.875 , 6.247585384295304e-47 ], [ 428.9375 , 6.248537972453193e-47 ], 
                    [ 429.0 , 6.24949070585502e-47 ], [ 429.0625 , 6.250443584523069e-47 ], [ 429.125 , 6.251396608479264e-47 ], [ 429.1875 , 6.25234977774598e-47 ], 
                    [ 429.25 , 6.253303114556061e-47 ], [ 429.3125 , 6.254256674182328e-47 ], [ 429.375 , 6.255210379215814e-47 ], [ 429.4375 , 6.256164229678911e-47 ], 
                    [ 429.5 , 6.25711822559353e-47 ], [ 429.5625 , 6.25807236698203e-47 ], [ 429.625 , 6.259026653866504e-47 ], [ 429.6875 , 6.259981086269271e-47 ], 
                    [ 429.75 , 6.260935664212389e-47 ], [ 429.8125 , 6.261890387718096e-47 ], [ 429.875 , 6.262845256808587e-47 ], [ 429.9375 , 6.263800271506064e-47 ], 
                    [ 430.0 , 6.264755431832729e-47 ], [ 430.0625 , 6.265710737810879e-47 ], [ 430.125 , 6.266666189462455e-47 ], [ 430.1875 , 6.267621786809985e-47 ], 
                    [ 430.25 , 6.268577582023974e-47 ], [ 430.3125 , 6.269533550325127e-47 ], [ 430.375 , 6.270489664413049e-47 ], [ 430.4375 , 6.271445924309883e-47 ], 
                    [ 430.5 , 6.272402330037866e-47 ], [ 430.5625 , 6.273358881619372e-47 ], [ 430.625 , 6.274315579076465e-47 ], [ 430.6875 , 6.275272422431568e-47 ], 
                    [ 430.75 , 6.276229411706797e-47 ], [ 430.8125 , 6.277186546924364e-47 ], [ 430.875 , 6.278143828106744e-47 ], [ 430.9375 , 6.279101255276066e-47 ], 
                    [ 431.0 , 6.280058828454591e-47 ], [ 431.0625 , 6.281016547664499e-47 ], [ 431.125 , 6.281974412928147e-47 ], [ 431.1875 , 6.282932427520449e-47 ], 
                    [ 431.25 , 6.283890645112173e-47 ], [ 431.3125 , 6.284849008842772e-47 ], [ 431.375 , 6.285807518734623e-47 ], [ 431.4375 , 6.286766174809927e-47 ], 
                    [ 431.5 , 6.287724977090978e-47 ], [ 431.5625 , 6.28868392560021e-47 ], [ 431.625 , 6.28964302035979e-47 ], [ 431.6875 , 6.290602261392021e-47 ], 
                    [ 431.75 , 6.291561648719302e-47 ], [ 431.8125 , 6.292521182363945e-47 ], [ 431.875 , 6.29348086234813e-47 ], [ 431.9375 , 6.294440688694355e-47 ], 
                    [ 432.0 , 6.295400661424809e-47 ], [ 432.0625 , 6.296360780561857e-47 ], [ 432.125 , 6.297321046127789e-47 ], [ 432.1875 , 6.298281500153036e-47 ], 
                    [ 432.25 , 6.299242159529996e-47 ], [ 432.3125 , 6.300202965433663e-47 ], [ 432.375 , 6.301163917886429e-47 ], [ 432.4375 , 6.302125016910557e-47 ], 
                    [ 432.5 , 6.303086262528406e-47 ], [ 432.5625 , 6.304047654762465e-47 ], [ 432.625 , 6.305009193634923e-47 ], [ 432.6875 , 6.305970879168232e-47 ], 
                    [ 432.75 , 6.306932711384853e-47 ], [ 432.8125 , 6.307894690307025e-47 ], [ 432.875 , 6.3088568159571245e-47 ], [ 432.9375 , 6.309819088357711e-47 ], 
                    [ 433.0 , 6.3107815075309415e-47 ], [ 433.0625 , 6.311744073499341e-47 ], [ 433.125 , 6.312706786285162e-47 ], [ 433.1875 , 6.313669687600427e-47 ], 
                    [ 433.25 , 6.314632750161895e-47 ], [ 433.3125 , 6.315595959625127e-47 ], [ 433.375 , 6.316559316012576e-47 ], [ 433.4375 , 6.317522819346653e-47 ], 
                    [ 433.5 , 6.318486469649817e-47 ], [ 433.5625 , 6.319450266944353e-47 ], [ 433.625 , 6.320414211252859e-47 ], [ 433.6875 , 6.321378302597585e-47 ], 
                    [ 433.75 , 6.322342541001178e-47 ], [ 433.8125 , 6.323306926485851e-47 ], [ 433.875 , 6.324271459074079e-47 ], [ 433.9375 , 6.325236138788394e-47 ], 
                    [ 434.0 , 6.326200965651193e-47 ], [ 434.0625 , 6.327165939684874e-47 ], [ 434.125 , 6.328131063252403e-47 ], [ 434.1875 , 6.329096398192325e-47 ], 
                    [ 434.25 , 6.330061880390821e-47 ], [ 434.3125 , 6.331027509870307e-47 ], [ 434.375 , 6.331993286653341e-47 ], [ 434.4375 , 6.33295921076235e-47 ], 
                    [ 434.5 , 6.33392528221976e-47 ], [ 434.5625 , 6.334891501048187e-47 ], [ 434.625 , 6.335857867269932e-47 ], [ 434.6875 , 6.3368243809076115e-47 ], 
                    [ 434.75 , 6.337791041983715e-47 ], [ 434.8125 , 6.338757850520644e-47 ], [ 434.875 , 6.339724806540846e-47 ], [ 434.9375 , 6.340691910067047e-47 ], 
                    [ 435.0 , 6.341659161121567e-47 ], [ 435.0625 , 6.342626559727002e-47 ], [ 435.125 , 6.343594124601224e-47 ], [ 435.1875 , 6.344561882589963e-47 ], 
                    [ 435.25 , 6.345529788216744e-47 ], [ 435.3125 , 6.346497841504089e-47 ], [ 435.375 , 6.34746604247448e-47 ], [ 435.4375 , 6.348434391150402e-47 ], 
                    [ 435.5 , 6.349402887554433e-47 ], [ 435.5625 , 6.350371531709155e-47 ], [ 435.625 , 6.351340323637019e-47 ], [ 435.6875 , 6.352309263360615e-47 ], 
                    [ 435.75 , 6.353278350902488e-47 ], [ 435.8125 , 6.3542475862852785e-47 ], [ 435.875 , 6.355216969531362e-47 ], [ 435.9375 , 6.35618650066334e-47 ], 
                    [ 436.0 , 6.357156179703955e-47 ], [ 436.0625 , 6.358126006675591e-47 ], [ 436.125 , 6.359096007752962e-47 ], [ 436.1875 , 6.360066181792569e-47 ], 
                    [ 436.25 , 6.361036503846606e-47 ], [ 436.3125 , 6.362006973937606e-47 ], [ 436.375 , 6.362977592088155e-47 ], [ 436.4375 , 6.363948358320841e-47 ], 
                    [ 436.5 , 6.364919272658213e-47 ], [ 436.5625 , 6.365890335123001e-47 ], [ 436.625 , 6.366861545737623e-47 ], [ 436.6875 , 6.367832904524863e-47 ], 
                    [ 436.75 , 6.3688044115071455e-47 ], [ 436.8125 , 6.36977606670708e-47 ], [ 436.875 , 6.370747870147594e-47 ], [ 436.9375 , 6.371719821850945e-47 ], 
                    [ 437.0 , 6.372691921839978e-47 ], [ 437.0625 , 6.373664170137135e-47 ], [ 437.125 , 6.374636605310448e-47 ], [ 437.1875 , 6.375609205687445e-47 ], 
                    [ 437.25 , 6.376581954457494e-47 ], [ 437.3125 , 6.377554851643146e-47 ], [ 437.375 , 6.378527897267134e-47 ], [ 437.4375 , 6.379501091351927e-47 ], 
                    [ 437.5 , 6.380474433920447e-47 ], [ 437.5625 , 6.381447924995166e-47 ], [ 437.625 , 6.382421564598789e-47 ], [ 437.6875 , 6.383395352753932e-47 ], 
                    [ 437.75 , 6.38436928948335e-47 ], [ 437.8125 , 6.385343374809668e-47 ], [ 437.875 , 6.386317608755602e-47 ], [ 437.9375 , 6.387291991343737e-47 ], 
                    [ 438.0 , 6.38826652259675e-47 ], [ 438.0625 , 6.389241202537461e-47 ], [ 438.125 , 6.390216079966382e-47 ], [ 438.1875 , 6.391191115116699e-47 ], 
                    [ 438.25 , 6.392166299040229e-47 ], [ 438.3125 , 6.39314163175976e-47 ], [ 438.375 , 6.394117113297995e-47 ], [ 438.4375 , 6.395092743677735e-47 ], 
                    [ 438.5 , 6.396068522921507e-47 ], [ 438.5625 , 6.397044451052161e-47 ], [ 438.625 , 6.39802052809228e-47 ], [ 438.6875 , 6.39899675406472e-47 ], 
                    [ 438.75 , 6.399973128992205e-47 ], [ 438.8125 , 6.400949652897419e-47 ], [ 438.875 , 6.401926325803092e-47 ], [ 438.9375 , 6.40290314773196e-47 ], 
                    [ 439.0 , 6.403880118706759e-47 ], [ 439.0625 , 6.404857238750189e-47 ], [ 439.125 , 6.405834541684452e-47 ], [ 439.1875 , 6.406811995245872e-47 ], 
                    [ 439.25 , 6.407789597955079e-47 ], [ 439.3125 , 6.408767349834603e-47 ], [ 439.375 , 6.40974525090748e-47 ], [ 439.4375 , 6.410723301196202e-47 ], 
                    [ 439.5 , 6.411701500723764e-47 ], [ 439.5625 , 6.412679849512758e-47 ], [ 439.625 , 6.413658347586185e-47 ], [ 439.6875 , 6.4146369949665975e-47 ], 
                    [ 439.75 , 6.415615791676778e-47 ], [ 439.8125 , 6.416594737739785e-47 ], [ 439.875 , 6.417573833178134e-47 ], [ 439.9375 , 6.418553078014757e-47 ], 
                    [ 440.0 , 6.419532472272358e-47 ], [ 440.0625 , 6.420512017675791e-47 ], [ 440.125 , 6.421491760791003e-47 ], [ 440.1875 , 6.422471653410891e-47 ], 
                    [ 440.25 , 6.423451695558269e-47 ], [ 440.3125 , 6.424431887256002e-47 ], [ 440.375 , 6.425412228526771e-47 ], [ 440.4375 , 6.426392719393582e-47 ], 
                    [ 440.5 , 6.42737335987913e-47 ], [ 440.5625 , 6.42835415000638e-47 ], [ 440.625 , 6.429335089797938e-47 ], [ 440.6875 , 6.430316179276781e-47 ], 
                    [ 440.75 , 6.431297418465841e-47 ], [ 440.8125 , 6.432278807387827e-47 ], [ 440.875 , 6.433260346065632e-47 ], [ 440.9375 , 6.434242034522154e-47 ], 
                    [ 441.0 , 6.435223872780205e-47 ], [ 441.0625 , 6.436205864213287e-47 ], [ 441.125 , 6.437188046993931e-47 ], [ 441.1875 , 6.438170379658423e-47 ], 
                    [ 441.25 , 6.439152862229592e-47 ], [ 441.3125 , 6.440135494730358e-47 ], [ 441.375 , 6.4411182771834195e-47 ], [ 441.4375 , 6.442101209611934e-47 ], 
                    [ 441.5 , 6.443084292038558e-47 ], [ 441.5625 , 6.444067524486274e-47 ], [ 441.625 , 6.4450509069780665e-47 ], [ 441.6875 , 6.446034439536605e-47 ], 
                    [ 441.75 , 6.447018122185018e-47 ], [ 441.8125 , 6.448001954946118e-47 ], [ 441.875 , 6.44898593784286e-47 ], [ 441.9375 , 6.449970070898016e-47 ], 
                    [ 442.0 , 6.450954354134639e-47 ], [ 442.0625 , 6.451938791246083e-47 ], [ 442.125 , 6.452923421267433e-47 ], [ 442.1875 , 6.453908201553117e-47 ], 
                    [ 442.25 , 6.4548931321259745e-47 ], [ 442.3125 , 6.455878213009124e-47 ], [ 442.375 , 6.45686344422532e-47 ], [ 442.4375 , 6.45784882579769e-47 ], 
                    [ 442.5 , 6.45883435774904e-47 ], [ 442.5625 , 6.459820040102367e-47 ], [ 442.625 , 6.460805872880624e-47 ], [ 442.6875 , 6.461791856106766e-47 ], 
                    [ 442.75 , 6.4627779898037525e-47 ], [ 442.8125 , 6.463764273994549e-47 ], [ 442.875 , 6.464750708702121e-47 ], [ 442.9375 , 6.465737293949485e-47 ], 
                    [ 443.0 , 6.466724029759478e-47 ], [ 443.0625 , 6.467710917583889e-47 ], [ 443.125 , 6.468697984656874e-47 ], [ 443.1875 , 6.469685202370791e-47 ], 
                    [ 443.25 , 6.470672570748447e-47 ], [ 443.3125 , 6.471660089812925e-47 ], [ 443.375 , 6.472647759587318e-47 ], [ 443.4375 , 6.473635580094439e-47 ], 
                    [ 443.5 , 6.474623551357431e-47 ], [ 443.5625 , 6.475611673399256e-47 ], [ 443.625 , 6.476599946242879e-47 ], [ 443.6875 , 6.477588369911407e-47 ], 
                    [ 443.75 , 6.47857694442781e-47 ], [ 443.8125 , 6.479565669815065e-47 ], [ 443.875 , 6.480554546096245e-47 ], [ 443.9375 , 6.481543573294468e-47 ], 
                    [ 444.0 , 6.482532751432581e-47 ], [ 444.0625 , 6.483522080533762e-47 ], [ 444.125 , 6.484511601325698e-47 ], [ 444.1875 , 6.485501273997119e-47 ], 
                    [ 444.25 , 6.486491097713332e-47 ], [ 444.3125 , 6.487481072497572e-47 ], [ 444.375 , 6.48847119837262e-47 ], [ 444.4375 , 6.489461475361718e-47 ], 
                    [ 444.5 , 6.490451903487885e-47 ], [ 444.5625 , 6.491442482774187e-47 ], [ 444.625 , 6.492433213243694e-47 ], [ 444.6875 , 6.493424094919527e-47 ], 
                    [ 444.75 , 6.494415127824577e-47 ], [ 444.8125 , 6.495406311982203e-47 ], [ 444.875 , 6.496397647415258e-47 ], [ 444.9375 , 6.497389134146923e-47 ], 
                    [ 445.0 , 6.498380772200333e-47 ], [ 445.0625 , 6.499372561598631e-47 ], [ 445.125 , 6.500364552330351e-47 ], [ 445.1875 , 6.501356701625713e-47 ], 
                    [ 445.25 , 6.502349002352616e-47 ], [ 445.3125 , 6.503341454534081e-47 ], [ 445.375 , 6.50433405819341e-47 ], [ 445.4375 , 6.505326813353583e-47 ], 
                    [ 445.5 , 6.506319720037818e-47 ], [ 445.5625 , 6.507312778269193e-47 ], [ 445.625 , 6.508305988070748e-47 ], [ 445.6875 , 6.509299349465755e-47 ], 
                    [ 445.75 , 6.5102928624774e-47 ], [ 445.8125 , 6.511286527128684e-47 ], [ 445.875 , 6.512280343442751e-47 ], [ 445.9375 , 6.513274311442703e-47 ], 
                    [ 446.0 , 6.51426843115197e-47 ], [ 446.0625 , 6.515262702593431e-47 ], [ 446.125 , 6.516257152344209e-47 ], [ 446.1875 , 6.517251763517996e-47 ], 
                    [ 446.25 , 6.518246526504605e-47 ], [ 446.3125 , 6.519241441327208e-47 ], [ 446.375 , 6.520236508008979e-47 ], [ 446.4375 , 6.521231726573147e-47 ], 
                    [ 446.5 , 6.522227097042799e-47 ], [ 446.5625 , 6.523222619441168e-47 ], [ 446.625 , 6.524218293791537e-47 ], [ 446.6875 , 6.525214120116821e-47 ], 
                    [ 446.75 , 6.52621009844054e-47 ], [ 446.8125 , 6.527206228785619e-47 ], [ 446.875 , 6.528202511175584e-47 ], [ 446.9375 , 6.529198945633318e-47 ], 
                    [ 447.0 , 6.530195532182311e-47 ], [ 447.0625 , 6.5311922708454535e-47 ], [ 447.125 , 6.532189190590147e-47 ], [ 447.1875 , 6.533186285606224e-47 ], 
                    [ 447.25 , 6.534183532822239e-47 ], [ 447.3125 , 6.535180932261469e-47 ], [ 447.375 , 6.536178483946961e-47 ], [ 447.4375 , 6.537176187902097e-47 ], 
                    [ 447.5 , 6.53817404414998e-47 ], [ 447.5625 , 6.539172052714089e-47 ], [ 447.625 , 6.540170213617348e-47 ], [ 447.6875 , 6.541168526883383e-47 ], 
                    [ 447.75 , 6.542166992535217e-47 ], [ 447.8125 , 6.543165610596021e-47 ], [ 447.875 , 6.544164381089382e-47 ], [ 447.9375 , 6.545163304038291e-47 ], 
                    [ 448.0 , 6.546162379466155e-47 ], [ 448.0625 , 6.547161607396206e-47 ], [ 448.125 , 6.548161007012483e-47 ], [ 448.1875 , 6.549160596093425e-47 ], 
                    [ 448.25 , 6.5501603377633785e-47 ], [ 448.3125 , 6.55116023204587e-47 ], [ 448.375 , 6.552160278964149e-47 ], [ 448.4375 , 6.553160478541328e-47 ], 
                    [ 448.5 , 6.554160830800993e-47 ], [ 448.5625 , 6.555161335766215e-47 ], [ 448.625 , 6.556161993460355e-47 ], [ 448.6875 , 6.557162803906911e-47 ], 
                    [ 448.75 , 6.558163767129016e-47 ], [ 448.8125 , 6.559164883150084e-47 ], [ 448.875 , 6.560166151993346e-47 ], [ 448.9375 , 6.561167573682132e-47 ], 
                    [ 449.0 , 6.562169148239868e-47 ], [ 449.0625 , 6.563170875689795e-47 ], [ 449.125 , 6.564172762158353e-47 ], [ 449.1875 , 6.565174862746523e-47 ], 
                    [ 449.25 , 6.566177116317595e-47 ], [ 449.3125 , 6.567179522894786e-47 ], [ 449.375 , 6.568182082501544e-47 ], [ 449.4375 , 6.56918479516114e-47 ], 
                    [ 449.5 , 6.570187660896984e-47 ], [ 449.5625 , 6.571190679732494e-47 ], [ 449.625 , 6.572193851690948e-47 ], [ 449.6875 , 6.573197176795816e-47 ], 
                    [ 449.75 , 6.57420065507043e-47 ], [ 449.8125 , 6.575204286538173e-47 ], [ 449.875 , 6.576208071222293e-47 ], [ 449.9375 , 6.57721200914646e-47 ], 
                    [ 450.0 , 6.578216100333787e-47 ], [ 450.0625 , 6.579220344808e-47 ], [ 450.125 , 6.580224742592311e-47 ], [ 450.1875 , 6.581229355599058e-47 ], 
                    [ 450.25 , 6.582234137178827e-47 ], [ 450.3125 , 6.583239072162416e-47 ], [ 450.375 , 6.584244160573339e-47 ], [ 450.4375 , 6.585249402434925e-47 ], 
                    [ 450.5 , 6.586254797770695e-47 ], [ 450.5625 , 6.58726034660399e-47 ], [ 450.625 , 6.588266048958336e-47 ], [ 450.6875 , 6.589271904857033e-47 ], 
                    [ 450.75 , 6.590277914323755e-47 ], [ 450.8125 , 6.591284077381716e-47 ], [ 450.875 , 6.592290394054505e-47 ], [ 450.9375 , 6.593296864365481e-47 ], 
                    [ 451.0 , 6.594303488338195e-47 ], [ 451.0625 , 6.595310265996012e-47 ], [ 451.125 , 6.596317197362397e-47 ], [ 451.1875 , 6.597324321236891e-47 ], 
                    [ 451.25 , 6.59833164121678e-47 ], [ 451.3125 , 6.599339115000469e-47 ], [ 451.375 , 6.600346742611486e-47 ], [ 451.4375 , 6.601354524073227e-47 ], 
                    [ 451.5 , 6.602362459409228e-47 ], [ 451.5625 , 6.603370548643031e-47 ], [ 451.625 , 6.604378791797993e-47 ], [ 451.6875 , 6.605387188897758e-47 ], 
                    [ 451.75 , 6.606395739965782e-47 ], [ 451.8125 , 6.607404445025527e-47 ], [ 451.875 , 6.608413304100695e-47 ], [ 451.9375 , 6.609422317214614e-47 ], 
                    [ 452.0 , 6.610431484390754e-47 ], [ 452.0625 , 6.61144080565283e-47 ], [ 452.125 , 6.612450281024226e-47 ], [ 452.1875 , 6.613459923994906e-47 ], 
                    [ 452.25 , 6.614469823605102e-47 ], [ 452.3125 , 6.615479877430582e-47 ], [ 452.375 , 6.616490085495036e-47 ], [ 452.4375 , 6.617500447821972e-47 ], 
                    [ 452.5 , 6.618510964434944e-47 ], [ 452.5625 , 6.619521635357511e-47 ], [ 452.625 , 6.62053246061324e-47 ], [ 452.6875 , 6.621543440225743e-47 ], 
                    [ 452.75 , 6.622554574218497e-47 ], [ 452.8125 , 6.623565862615124e-47 ], [ 452.875 , 6.624577305439248e-47 ], [ 452.9375 , 6.62558890271436e-47 ], 
                    [ 453.0 , 6.626600654464088e-47 ], [ 453.0625 , 6.627612560712023e-47 ], [ 453.125 , 6.628624621481709e-47 ], [ 453.1875 , 6.629636836796838e-47 ], 
                    [ 453.25 , 6.6306493076755055e-47 ], [ 453.3125 , 6.631661972991199e-47 ], [ 453.375 , 6.632674792966205e-47 ], [ 453.4375 , 6.633687767624003e-47 ], 
                    [ 453.5 , 6.63470089698817e-47 ], [ 453.5625 , 6.635714181082571e-47 ], [ 453.625 , 6.636727619930596e-47 ], [ 453.6875 , 6.637741213555979e-47 ], 
                    [ 453.75 , 6.638754961982451e-47 ], [ 453.8125 , 6.639768865233465e-47 ], [ 453.875 , 6.640782923332761e-47 ], [ 453.9375 , 6.64179713630399e-47 ], 
                    [ 454.0 , 6.642811504170849e-47 ], [ 454.0625 , 6.643826026956902e-47 ], [ 454.125 , 6.644840704685997e-47 ], [ 454.1875 , 6.645855537381421e-47 ], 
                    [ 454.25 , 6.646870567238902e-47 ], [ 454.3125 , 6.647885859895662e-47 ], [ 454.375 , 6.648901307635804e-47 ], [ 454.4375 , 6.649916910483114e-47 ], 
                    [ 454.5 , 6.650932668461094e-47 ], [ 454.5625 , 6.651948581593487e-47 ], [ 454.625 , 6.652964649904088e-47 ], [ 454.6875 , 6.653980873416551e-47 ], 
                    [ 454.75 , 6.654997252154584e-47 ], [ 454.8125 , 6.656013786141896e-47 ], [ 454.875 , 6.657030475402202e-47 ], [ 454.9375 , 6.658047319959268e-47 ], 
                    [ 455.0 , 6.659064319836765e-47 ], [ 455.0625 , 6.660081475058374e-47 ], [ 455.125 , 6.661098785647867e-47 ], [ 455.1875 , 6.662116251628931e-47 ], 
                    [ 455.25 , 6.663133873025444e-47 ], [ 455.3125 , 6.664151822631343e-47 ], [ 455.375 , 6.665169968855872e-47 ], [ 455.4375 , 6.666188270632292e-47 ], 
                    [ 455.5 , 6.667206727984416e-47 ], [ 455.5625 , 6.668225340936013e-47 ], [ 455.625 , 6.669244109510901e-47 ], [ 455.6875 , 6.670263033732764e-47 ], 
                    [ 455.75 , 6.671282113625427e-47 ], [ 455.8125 , 6.672301349212768e-47 ], [ 455.875 , 6.673320740518291e-47 ], [ 455.9375 , 6.674340287566164e-47 ], 
                    [ 456.0 , 6.675359990379898e-47 ], [ 456.0625 , 6.676379848983435e-47 ], [ 456.125 , 6.677399863400525e-47 ], [ 456.1875 , 6.678420033654833e-47 ], 
                    [ 456.25 , 6.679440359770453e-47 ], [ 456.3125 , 6.680460912889185e-47 ], [ 456.375 , 6.68148179007547e-47 ], [ 456.4375 , 6.682502823267539e-47 ], 
                    [ 456.5 , 6.683524012489138e-47 ], [ 456.5625 , 6.684545357764158e-47 ], [ 456.625 , 6.685566859116494e-47 ], [ 456.6875 , 6.686588516569948e-47 ], 
                    [ 456.75 , 6.687610330148331e-47 ], [ 456.8125 , 6.688632299875593e-47 ], [ 456.875 , 6.689654425775454e-47 ], [ 456.9375 , 6.690676707871969e-47 ], 
                    [ 457.0 , 6.691699146188961e-47 ], [ 457.0625 , 6.692721740750113e-47 ], [ 457.125 , 6.693744491579443e-47 ], [ 457.1875 , 6.694767398700926e-47 ], 
                    [ 457.25 , 6.695790462138306e-47 ], [ 457.3125 , 6.696813681915657e-47 ], [ 457.375 , 6.697837302814841e-47 ], [ 457.4375 , 6.698861162158621e-47 ], 
                    [ 457.5 , 6.699885178013716e-47 ], [ 457.5625 , 6.700909350404193e-47 ], [ 457.625 , 6.7019336793539335e-47 ], [ 457.6875 , 6.702958164886774e-47 ], 
                    [ 457.75 , 6.703982807026935e-47 ], [ 457.8125 , 6.705007605798025e-47 ], [ 457.875 , 6.706032561224081e-47 ], [ 457.9375 , 6.707057673329288e-47 ], 
                    [ 458.0 , 6.708082942137314e-47 ], [ 458.0625 , 6.709108367672252e-47 ], [ 458.125 , 6.710133949957822e-47 ], [ 458.1875 , 6.711159689018369e-47 ], 
                    [ 458.25 , 6.712185584877524e-47 ], [ 458.3125 , 6.713211637559542e-47 ], [ 458.375 , 6.714237910065053e-47 ], [ 458.4375 , 6.715264661665429e-47 ], 
                    [ 458.5 , 6.716291570278193e-47 ], [ 458.5625 , 6.717318635927356e-47 ], [ 458.625 , 6.718345858637028e-47 ], [ 458.6875 , 6.719373238431225e-47 ], 
                    [ 458.75 , 6.72040077533397e-47 ], [ 458.8125 , 6.721428469369003e-47 ], [ 458.875 , 6.722456320560828e-47 ], [ 458.9375 , 6.7234843289331445e-47 ], 
                    [ 459.0 , 6.724512494510037e-47 ], [ 459.0625 , 6.725540817315688e-47 ], [ 459.125 , 6.726569297373951e-47 ], [ 459.1875 , 6.727597934709064e-47 ], 
                    [ 459.25 , 6.728626729344933e-47 ], [ 459.3125 , 6.729655681305663e-47 ], [ 459.375 , 6.730684790615359e-47 ], [ 459.4375 , 6.731714328830117e-47 ], 
                    [ 459.5 , 6.732744255567804e-47 ], [ 459.5625 , 6.733774339880277e-47 ], [ 459.625 , 6.734804581791787e-47 ], [ 459.6875 , 6.735834981326447e-47 ], 
                    [ 459.75 , 6.736865538508228e-47 ], [ 459.8125 , 6.737896253361395e-47 ], [ 459.875 , 6.738927125909972e-47 ], [ 459.9375 , 6.739958156178183e-47 ], 
                    [ 460.0 , 6.740989344190113e-47 ], [ 460.0625 , 6.74202068996994e-47 ], [ 460.125 , 6.743052193541614e-47 ], [ 460.1875 , 6.744083854929611e-47 ], 
                    [ 460.25 , 6.745115674157739e-47 ], [ 460.3125 , 6.746147651250292e-47 ], [ 460.375 , 6.747179786231519e-47 ], [ 460.4375 , 6.748212079125432e-47 ], 
                    [ 460.5 , 6.74924508045742e-47 ], [ 460.5625 , 6.750278315340301e-47 ], [ 460.625 , 6.751311708400081e-47 ], [ 460.6875 , 6.752345259660829e-47 ], 
                    [ 460.75 , 6.753378969146956e-47 ], [ 460.8125 , 6.754412836882541e-47 ], [ 460.875 , 6.75544686289181e-47 ], [ 460.9375 , 6.756481047199089e-47 ], 
                    [ 461.0 , 6.757515389828661e-47 ], [ 461.0625 , 6.758549890804521e-47 ], [ 461.125 , 6.759584550151056e-47 ], [ 461.1875 , 6.760619367892606e-47 ], 
                    [ 461.25 , 6.761654344053228e-47 ], [ 461.3125 , 6.762689478657365e-47 ], [ 461.375 , 6.763724771729131e-47 ], [ 461.4375 , 6.764760223292832e-47 ], 
                    [ 461.5 , 6.765795981949694e-47 ], [ 461.5625 , 6.766832572494291e-47 ], [ 461.625 , 6.767869321855231e-47 ], [ 461.6875 , 6.768906230057185e-47 ], 
                    [ 461.75 , 6.769943297124103e-47 ], [ 461.8125 , 6.770980523080613e-47 ], [ 461.875 , 6.772017907950915e-47 ], [ 461.9375 , 6.773055451759452e-47 ], 
                    [ 462.0 , 6.774093154530527e-47 ], [ 462.0625 , 6.775131016288399e-47 ], [ 462.125 , 6.776169037057665e-47 ], [ 462.1875 , 6.777207216862451e-47 ], 
                    [ 462.25 , 6.778245555727214e-47 ], [ 462.3125 , 6.779284053676375e-47 ], [ 462.375 , 6.780322710734259e-47 ], [ 462.4375 , 6.781361526925242e-47 ], 
                    [ 462.5 , 6.782400502273608e-47 ], [ 462.5625 , 6.783440104036036e-47 ], [ 462.625 , 6.78448044957204e-47 ], [ 462.6875 , 6.785520954661136e-47 ], 
                    [ 462.75 , 6.786561619327841e-47 ], [ 462.8125 , 6.78760244359658e-47 ], [ 462.875 , 6.788643427491734e-47 ], [ 462.9375 , 6.789684571037978e-47 ], 
                    [ 463.0 , 6.790725874259699e-47 ], [ 463.0625 , 6.791767337181387e-47 ], [ 463.125 , 6.792808959827487e-47 ], [ 463.1875 , 6.793850742222542e-47 ], 
                    [ 463.25 , 6.794892684391099e-47 ], [ 463.3125 , 6.795934786357665e-47 ], [ 463.375 , 6.796977048146552e-47 ], [ 463.4375 , 6.798019469782417e-47 ], 
                    [ 463.5 , 6.799062051289872e-47 ], [ 463.5625 , 6.800104792693385e-47 ], [ 463.625 , 6.801148616183563e-47 ], [ 463.6875 , 6.802193054337215e-47 ], 
                    [ 463.75 , 6.803237652883047e-47 ], [ 463.8125 , 6.80428241184569e-47 ], [ 463.875 , 6.805327331249829e-47 ], [ 463.9375 , 6.806372411120004e-47 ], 
                    [ 464.0 , 6.807417651480859e-47 ], [ 464.0625 , 6.808463052357183e-47 ], [ 464.125 , 6.809508613773483e-47 ], [ 464.1875 , 6.810554335754509e-47 ], 
                    [ 464.25 , 6.811600218324823e-47 ], [ 464.3125 , 6.812646261509133e-47 ], [ 464.375 , 6.813692465332154e-47 ], [ 464.4375 , 6.814738829818358e-47 ], 
                    [ 464.5 , 6.815785354992663e-47 ], [ 464.5625 , 6.816832040879698e-47 ], [ 464.625 , 6.817878887504044e-47 ], [ 464.6875 , 6.818927465551731e-47 ], 
                    [ 464.75 , 6.819976466051003e-47 ], [ 464.8125 , 6.821025627924831e-47 ], [ 464.875 , 6.822074951198136e-47 ], [ 464.9375 , 6.823124435895844e-47 ], 
                    [ 465.0 , 6.824174082042692e-47 ], [ 465.0625 , 6.825223889663419e-47 ], [ 465.125 , 6.826273858782965e-47 ], [ 465.1875 , 6.82732398942617e-47 ], 
                    [ 465.25 , 6.828374281617885e-47 ], [ 465.3125 , 6.829424735382962e-47 ], [ 465.375 , 6.830475350746402e-47 ], [ 465.4375 , 6.83152612773287e-47 ], 
                    [ 465.5 , 6.832577066367183e-47 ], [ 465.5625 , 6.833628166674255e-47 ], [ 465.625 , 6.834679428679103e-47 ], [ 465.6875 , 6.835730869379811e-47 ], 
                    [ 465.75 , 6.836784922289628e-47 ], [ 465.8125 , 6.837839137731717e-47 ], [ 465.875 , 6.838893515731334e-47 ], [ 465.9375 , 6.839948056313401e-47 ], 
                    [ 466.0 , 6.8410027595030845e-47 ], [ 466.0625 , 6.842057625325359e-47 ], [ 466.125 , 6.843112653805305e-47 ], [ 466.1875 , 6.844167844968197e-47 ], 
                    [ 466.25 , 6.84522319883878e-47 ], [ 466.3125 , 6.846278715442434e-47 ], [ 466.375 , 6.847334394804254e-47 ], [ 466.4375 , 6.848390236949141e-47 ], 
                    [ 466.5 , 6.849446241902244e-47 ], [ 466.5625 , 6.850502409688718e-47 ], [ 466.625 , 6.851558740333718e-47 ], [ 466.6875 , 6.852615233862359e-47 ], 
                    [ 466.75 , 6.853672289183722e-47 ], [ 466.8125 , 6.854732482649487e-47 ], [ 466.875 , 6.855792840116311e-47 ], [ 466.9375 , 6.856853361609709e-47 ], 
                    [ 467.0 , 6.857914047155009e-47 ], [ 467.0625 , 6.858974896777583e-47 ], [ 467.125 , 6.860035910502865e-47 ], [ 467.1875 , 6.861097088356092e-47 ], 
                    [ 467.25 , 6.862158430362799e-47 ], [ 467.3125 , 6.863219936548379e-47 ], [ 467.375 , 6.864281606938134e-47 ], [ 467.4375 , 6.865343441557411e-47 ], 
                    [ 467.5 , 6.86640544043191e-47 ], [ 467.5625 , 6.867467603586699e-47 ], [ 467.625 , 6.868529931047384e-47 ], [ 467.6875 , 6.869592422839334e-47 ], 
                    [ 467.75 , 6.870655078987918e-47 ], [ 467.8125 , 6.871718807395542e-47 ], [ 467.875 , 6.872786541520389e-47 ], [ 467.9375 , 6.873854441550814e-47 ], 
                    [ 468.0 , 6.874922507512551e-47 ], [ 468.0625 , 6.87599073943138e-47 ], [ 468.125 , 6.87705913733304e-47 ], [ 468.1875 , 6.878127701243369e-47 ], 
                    [ 468.25 , 6.879196431188212e-47 ], [ 468.3125 , 6.880265327193318e-47 ], [ 468.375 , 6.88133438928444e-47 ], [ 468.4375 , 6.8824036174874345e-47 ], 
                    [ 468.5 , 6.883473011828209e-47 ], [ 468.5625 , 6.884542572332384e-47 ], [ 468.625 , 6.885612299025924e-47 ], [ 468.6875 , 6.886682191934554e-47 ], 
                    [ 468.75 , 6.887752251084198e-47 ], [ 468.8125 , 6.888822476500638e-47 ], [ 468.875 , 6.889894412979384e-47 ], [ 468.9375 , 6.890971814844778e-47 ], 
                    [ 469.0 , 6.89204938518805e-47 ], [ 469.0625 , 6.893127124035592e-47 ], [ 469.125 , 6.894205031413657e-47 ], [ 469.1875 , 6.895283107348649e-47 ], 
                    [ 469.25 , 6.896361351866922e-47 ], [ 469.3125 , 6.897439764994841e-47 ], [ 469.375 , 6.898518346758773e-47 ], [ 469.4375 , 6.899597097185185e-47 ], 
                    [ 469.5 , 6.900676016300256e-47 ], [ 469.5625 , 6.901755104130362e-47 ], [ 469.625 , 6.902834360702184e-47 ], [ 469.6875 , 6.903913786041811e-47 ], 
                    [ 469.75 , 6.904993380175736e-47 ], [ 469.8125 , 6.90607314313035e-47 ], [ 469.875 , 6.907153074932054e-47 ], [ 469.9375 , 6.908235438997497e-47 ], 
                    [ 470.0 , 6.909325909657348e-47 ], [ 470.0625 , 6.910416552448844e-47 ], [ 470.125 , 6.911507367399204e-47 ], [ 470.1875 , 6.912598354535603e-47 ], 
                    [ 470.25 , 6.91368951388527e-47 ], [ 470.3125 , 6.914780845475291e-47 ], [ 470.375 , 6.915872349332902e-47 ], [ 470.4375 , 6.916964025485348e-47 ], 
                    [ 470.5 , 6.918055873959675e-47 ], [ 470.5625 , 6.919147894783332e-47 ], [ 470.625 , 6.920240087983278e-47 ], [ 470.6875 , 6.921332453586723e-47 ], 
                    [ 470.75 , 6.922424991621225e-47 ], [ 470.8125 , 6.923517702113559e-47 ], [ 470.875 , 6.924610585091195e-47 ], [ 470.9375 , 6.925703640581408e-47 ], 
                    [ 471.0 , 6.926799718516119e-47 ], [ 471.0625 , 6.927908997229143e-47 ], [ 471.125 , 6.929018453585326e-47 ], [ 471.1875 , 6.9301280876133125e-47 ], 
                    [ 471.25 , 6.931237899341359e-47 ], [ 471.3125 , 6.932347888797971e-47 ], [ 471.375 , 6.93345805601176e-47 ], [ 471.4375 , 6.934568401011092e-47 ], 
                    [ 471.5 , 6.935678923824442e-47 ], [ 471.5625 , 6.936789624480281e-47 ], [ 471.625 , 6.937900503007042e-47 ], [ 471.6875 , 6.93901155943331e-47 ], 
                    [ 471.75 , 6.940122793787423e-47 ], [ 471.8125 , 6.941234206098174e-47 ], [ 471.875 , 6.942345796393714e-47 ], [ 471.9375 , 6.943457564702746e-47 ], 
                    [ 472.0 , 6.944569511053675e-47 ], [ 472.0625 , 6.945684196851257e-47 ], [ 472.125 , 6.946822491985819e-47 ], [ 472.1875 , 6.947960973670008e-47 ], 
                    [ 472.25 , 6.94909964193464e-47 ], [ 472.3125 , 6.950238496810148e-47 ], [ 472.375 , 6.951377538327163e-47 ], [ 472.4375 , 6.952516766516271e-47 ], 
                    [ 472.5 , 6.953656181408116e-47 ], [ 472.5625 , 6.954795783033247e-47 ], [ 472.625 , 6.955935571422217e-47 ], [ 472.6875 , 6.95707554660578e-47 ], 
                    [ 472.75 , 6.958215708614305e-47 ], [ 472.8125 , 6.959356057478605e-47 ], [ 472.875 , 6.960496593229254e-47 ], [ 472.9375 , 6.961637315896931e-47 ], 
                    [ 473.0 , 6.962778225512118e-47 ], [ 473.0625 , 6.963919322105703e-47 ], [ 473.125 , 6.965060605708029e-47 ], [ 473.1875 , 6.966246792281569e-47 ], 
                    [ 473.25 , 6.967434153281962e-47 ], [ 473.3125 , 6.968621716662039e-47 ], [ 473.375 , 6.969809482456149e-47 ], [ 473.4375 , 6.970997450698839e-47 ], 
                    [ 473.5 , 6.972185621424619e-47 ], [ 473.5625 , 6.973373994668047e-47 ], [ 473.625 , 6.974562570463544e-47 ], [ 473.6875 , 6.975751348845782e-47 ], 
                    [ 473.75 , 6.976940329849044e-47 ], [ 473.8125 , 6.978129513508012e-47 ], [ 473.875 , 6.97931889985733e-47 ], [ 473.9375 , 6.980508488931395e-47 ], 
                    [ 474.0 , 6.98169828076486e-47 ], [ 474.0625 , 6.982888275392135e-47 ], [ 474.125 , 6.984078472848083e-47 ], [ 474.1875 , 6.985268873166878e-47 ], 
                    [ 474.25 , 6.986531723146081e-47 ], [ 474.3125 , 6.987810646254231e-47 ], [ 474.375 , 6.989089803476199e-47 ], [ 474.4375 , 6.990369194855092e-47 ], 
                    [ 474.5 , 6.99164882043362e-47 ], [ 474.5625 , 6.992928680254561e-47 ], [ 474.625 , 6.9942087743609385e-47 ], [ 474.6875 , 6.995489102795593e-47 ], 
                    [ 474.75 , 6.996769665601368e-47 ], [ 474.8125 , 6.998050462821316e-47 ], [ 474.875 , 6.9993314944982e-47 ], [ 474.9375 , 7.000612760674988e-47 ], 
                    [ 475.0 , 7.001894261394554e-47 ], [ 475.0625 , 7.003175996699985e-47 ], [ 475.125 , 7.004457966634072e-47 ], [ 475.1875 , 7.005740171239767e-47 ], 
                    [ 475.25 , 7.0070226105601245e-47 ], [ 475.3125 , 7.008426105684013e-47 ], [ 475.375 , 7.009903098018823e-47 ], [ 475.4375 , 7.011380401622766e-47 ], 
                    [ 475.5 , 7.012858016561388e-47 ], [ 475.5625 , 7.014335942900303e-47 ], [ 475.625 , 7.0158141807050865e-47 ], [ 475.6875 , 7.017292730041528e-47 ], 
                    [ 475.75 , 7.018771590975083e-47 ], [ 475.8125 , 7.020250763571615e-47 ], [ 475.875 , 7.021730247896662e-47 ], [ 475.9375 , 7.023210044015963e-47 ], 
                    [ 476.0 , 7.024690151995129e-47 ], [ 476.0625 , 7.026170571900136e-47 ], [ 476.125 , 7.027651303796517e-47 ], [ 476.1875 , 7.029132347750074e-47 ], 
                    [ 476.25 , 7.030613703826574e-47 ], [ 476.3125 , 7.032095372091792e-47 ], [ 476.375 , 7.033780138341908e-47 ], [ 476.4375 , 7.035787419460299e-47 ], 
                    [ 476.5 , 7.037795273411186e-47 ], [ 476.5625 , 7.039803700358045e-47 ], [ 476.625 , 7.041812700464295e-47 ], [ 476.6875 , 7.0438222738935e-47 ], 
                    [ 476.75 , 7.045832420809326e-47 ], [ 476.8125 , 7.047843141375483e-47 ], [ 476.875 , 7.049854435755677e-47 ], [ 476.9375 , 7.051866304113562e-47 ], 
                    [ 477.0 , 7.053878746613038e-47 ], [ 477.0625 , 7.055891763417752e-47 ], [ 477.125 , 7.057905354691898e-47 ], [ 477.1875 , 7.059919520599165e-47 ], 
                    [ 477.25 , 7.061934261303692e-47 ], [ 477.3125 , 7.063949576969513e-47 ], [ 477.375 , 7.065965467760558e-47 ], [ 477.4375 , 7.0682086157538875e-47 ], 
                    [ 477.5 , 7.072271023517484e-47 ], [ 477.5625 , 7.076335766123999e-47 ], [ 477.625 , 7.080402844915268e-47 ], [ 477.6875 , 7.084472261234202e-47 ], 
                    [ 477.75 , 7.088544016424176e-47 ], [ 477.8125 , 7.092618111829393e-47 ], [ 477.875 , 7.096694548794926e-47 ], [ 477.9375 , 7.100773328666624e-47 ], 
                    [ 478.0 , 7.104854452791005e-47 ], [ 478.0625 , 7.108937922515313e-47 ], [ 478.125 , 7.113023739187818e-47 ], [ 478.1875 , 7.117111904157413e-47 ], 
                    [ 478.25 , 7.121202418773668e-47 ], [ 478.3125 , 7.125295284387125e-47 ], [ 478.375 , 7.129390502348957e-47 ], [ 478.4375 , 7.133488074011109e-47 ], 
                    [ 478.5 , 7.137588000726406e-47 ], [ 478.5625 , 7.157137872227025e-47 ], [ 478.625 , 7.180712537733411e-47 ], [ 478.6875 , 7.204364855069854e-47 ], 
                    [ 478.75 , 7.228095080011262e-47 ], [ 478.8125 , 7.251903469174824e-47 ], [ 478.875 , 7.275790280023308e-47 ], [ 478.9375 , 7.299755770867317e-47 ], 
                    [ 479.0 , 7.323800200868351e-47 ], [ 479.0625 , 7.347923830041552e-47 ], [ 479.125 , 7.372126919258571e-47 ], [ 479.1875 , 7.396409730250176e-47 ], 
                    [ 479.25 , 7.420772525609455e-47 ], [ 479.3125 , 7.445215568794122e-47 ], [ 479.375 , 7.469739124130274e-47 ], [ 479.4375 , 7.494343456813862e-47 ], 
                    [ 479.5 , 7.519028832915264e-47 ], [ 479.5625 , 7.543795519380192e-47 ], [ 479.625 , 7.621454849078273e-47 ], [ 479.6875 , 7.767692467623243e-47 ], 
                    [ 479.75 , 7.91673603877185e-47 ], [ 479.8125 , 8.068639402090768e-47 ], [ 479.875 , 8.223457430200207e-47 ], [ 479.9375 , 8.381246048595473e-47 ], 
                    [ 480.0 , 8.542062255849431e-47 ], [ 480.0625 , 8.705964144201884e-47 ], [ 480.125 , 8.87301092054531e-47 ], [ 480.1875 , 9.04326292781148e-47 ], 
                    [ 480.25 , 9.216781666769584e-47 ], [ 480.3125 , 9.393629818242897e-47 ], [ 480.375 , 9.573871265750582e-47 ], [ 480.4375 , 9.757571118584955e-47 ], 
                    [ 480.5 , 9.944795735330737e-47 ], [ 480.5625 , 1.0135612747836655e-46 ], [ 480.625 , 1.0330091085645994e-46 ], [ 480.6875 , 1.050750782533393e-46 ], 
                    [ 480.75 , 1.028667027096339e-46 ], [ 480.8125 , 1.0070474086005112e-46 ], [ 480.875 , 9.858821722240602e-47 ], [ 480.9375 , 9.651617681633923e-47 ], 
                    [ 481.0 , 9.448768473242831e-47 ], [ 481.0625 , 9.250182571035511e-47 ], [ 481.125 , 9.055770372593562e-47 ], [ 481.1875 , 8.86544415868357e-47 ], 
                    [ 481.25 , 8.679118053677601e-47 ], [ 481.3125 , 8.4967079868066e-47 ], [ 481.375 , 8.318131654226234e-47 ], [ 481.4375 , 8.143308481882442e-47 ], 
                    [ 481.5 , 7.972159589155726e-47 ], [ 481.5625 , 7.804607753269599e-47 ], [ 481.625 , 7.640577374448615e-47 ], [ 481.6875 , 7.479994441806392e-47 ], 
                    [ 481.75 , 7.322786499952614e-47 ], [ 481.8125 , 7.260207966336276e-47 ], [ 481.875 , 7.255074415275707e-47 ], [ 481.9375 , 7.249944494048744e-47 ], 
                    [ 482.0 , 7.2448182000887e-47 ], [ 482.0625 , 7.239695530830908e-47 ], [ 482.125 , 7.234576483712414e-47 ], [ 482.1875 , 7.229461056171968e-47 ], 
                    [ 482.25 , 7.22434924565029e-47 ], [ 482.3125 , 7.219241049589807e-47 ], [ 482.375 , 7.214136465434904e-47 ], [ 482.4375 , 7.209035490631624e-47 ], 
                    [ 482.5 , 7.203938122627866e-47 ], [ 482.5625 , 7.198844358873379e-47 ], [ 482.625 , 7.193754196819572e-47 ], [ 482.6875 , 7.188667633919802e-47 ], 
                    [ 482.75 , 7.183584667629175e-47 ], [ 482.8125 , 7.178505295404603e-47 ], [ 482.875 , 7.174108473085578e-47 ], [ 482.9375 , 7.173738773447742e-47 ], 
                    [ 483.0 , 7.173369092861396e-47 ], [ 483.0625 , 7.172999431325661e-47 ], [ 483.125 , 7.172629788839452e-47 ], [ 483.1875 , 7.172260165401891e-47 ], 
                    [ 483.25 , 7.171890561011893e-47 ], [ 483.3125 , 7.171520975668578e-47 ], [ 483.375 , 7.171151409370864e-47 ], [ 483.4375 , 7.1707818621178695e-47 ], 
                    [ 483.5 , 7.170412333908564e-47 ], [ 483.5625 , 7.170042824741864e-47 ], [ 483.625 , 7.169673334617042e-47 ], [ 483.6875 , 7.169303863532964e-47 ], 
                    [ 483.75 , 7.1689344114884965e-47 ], [ 483.8125 , 7.168564978482913e-47 ], [ 483.875 , 7.16819556451513e-47 ], [ 483.9375 , 7.167826169584117e-47 ], 
                    [ 484.0 , 7.168025518676179e-47 ], [ 484.0625 , 7.168555376620068e-47 ], [ 484.125 , 7.16908527373077e-47 ], [ 484.1875 , 7.169615210011231e-47 ], 
                    [ 484.25 , 7.170145185464601e-47 ], [ 484.3125 , 7.170675200093418e-47 ], [ 484.375 , 7.171205253900783e-47 ], [ 484.4375 , 7.171735346889439e-47 ], 
                    [ 484.5 , 7.172265479062484e-47 ], [ 484.5625 , 7.172795650422765e-47 ], [ 484.625 , 7.173325860972976e-47 ], [ 484.6875 , 7.173856110716267e-47 ], 
                    [ 484.75 , 7.174386399655332e-47 ], [ 484.8125 , 7.174916727793323e-47 ], [ 484.875 , 7.175447095132884e-47 ], [ 484.9375 , 7.175977501677012e-47 ], 
                    [ 485.0 , 7.17650794742871e-47 ], [ 485.0625 , 7.177061772099971e-47 ], [ 485.125 , 7.177880043813566e-47 ], [ 485.1875 , 7.178698408820121e-47 ], 
                    [ 485.25 , 7.179516867130066e-47 ], [ 485.3125 , 7.180335418754246e-47 ], [ 485.375 , 7.181154063703195e-47 ], [ 485.4375 , 7.181972801987555e-47 ], 
                    [ 485.5 , 7.1827916336179145e-47 ], [ 485.5625 , 7.183610558604969e-47 ], [ 485.625 , 7.184429576959362e-47 ], [ 485.6875 , 7.185248688691788e-47 ], 
                    [ 485.75 , 7.186067893812743e-47 ], [ 485.8125 , 7.186887192333075e-47 ], [ 485.875 , 7.187706584263333e-47 ], [ 485.9375 , 7.188526069614115e-47 ], 
                    [ 486.0 , 7.189345648396123e-47 ], [ 486.0625 , 7.19016532062001e-47 ], [ 486.125 , 7.19098508629643e-47 ], [ 486.1875 , 7.191863487044269e-47 ], 
                    [ 486.25 , 7.192802953613752e-47 ], [ 486.3125 , 7.193742542904948e-47 ], [ 486.375 , 7.194682254933735e-47 ], [ 486.4375 , 7.195622089716246e-47 ], 
                    [ 486.5 , 7.19656204726857e-47 ], [ 486.5625 , 7.197502127606589e-47 ], [ 486.625 , 7.198442330746598e-47 ], [ 486.6875 , 7.199382656704284e-47 ], 
                    [ 486.75 , 7.200323105496094e-47 ], [ 486.8125 , 7.20126367713772e-47 ], [ 486.875 , 7.202204371645516e-47 ], [ 486.9375 , 7.203145189035273e-47 ], 
                    [ 487.0 , 7.204086129323099e-47 ], [ 487.0625 , 7.205027192525147e-47 ], [ 487.125 , 7.205968378657526e-47 ], [ 487.1875 , 7.206909687736089e-47 ], 
                    [ 487.25 , 7.207851119776998e-47 ], [ 487.3125 , 7.208842671575389e-47 ], [ 487.375 , 7.2098425866091035e-47 ], [ 487.4375 , 7.210842640337614e-47 ], 
                    [ 487.5 , 7.211842832780571e-47 ], [ 487.5625 , 7.212843163956801e-47 ], [ 487.625 , 7.213843633885703e-47 ], [ 487.6875 , 7.214844242586575e-47 ], 
                    [ 487.75 , 7.215844990078562e-47 ], [ 487.8125 , 7.216845876380967e-47 ], [ 487.875 , 7.217846901513145e-47 ], [ 487.9375 , 7.218848065494148e-47 ], 
                    [ 488.0 , 7.2198493683432885e-47 ], [ 488.0625 , 7.220850810080084e-47 ], [ 488.125 , 7.221852390723439e-47 ], [ 488.1875 , 7.222854110292776e-47 ], 
                    [ 488.25 , 7.223855968807415e-47 ], [ 488.3125 , 7.224857966286527e-47 ], [ 488.375 , 7.225866033704038e-47 ], [ 488.4375 , 7.226899915878236e-47 ], 
                    [ 488.5 , 7.2279339459811e-47 ], [ 488.5625 , 7.22896812403364e-47 ], [ 488.625 , 7.230002450057234e-47 ], [ 488.6875 , 7.231036924072946e-47 ], 
                    [ 488.75 , 7.232071546101903e-47 ], [ 488.8125 , 7.233106316165386e-47 ], [ 488.875 , 7.23414123428442e-47 ], [ 488.9375 , 7.235176300480394e-47 ], 
                    [ 489.0 , 7.236211514774342e-47 ], [ 489.0625 , 7.237246877187557e-47 ], [ 489.125 , 7.238282387741181e-47 ], [ 489.1875 , 7.239318046456356e-47 ], 
                    [ 489.25 , 7.240353853354283e-47 ], [ 489.3125 , 7.24138980845637e-47 ], [ 489.375 , 7.242425911783513e-47 ], [ 489.4375 , 7.243462163357177e-47 ], 
                    [ 489.5 , 7.244507397700242e-47 ], [ 489.5625 , 7.245562490963022e-47 ], [ 489.625 , 7.246617737890154e-47 ], [ 489.6875 , 7.247673138503864e-47 ], 
                    [ 489.75 , 7.2487286928265345e-47 ], [ 489.8125 , 7.249784400880604e-47 ], [ 489.875 , 7.250840262688617e-47 ], [ 489.9375 , 7.251896278272655e-47 ], 
                    [ 490.0 , 7.252952447655322e-47 ], [ 490.0625 , 7.254008770858967e-47 ], [ 490.125 , 7.255065247905936e-47 ], [ 490.1875 , 7.256121878818743e-47 ], 
                    [ 490.25 , 7.257178663619743e-47 ], [ 490.3125 , 7.258235602331298e-47 ], [ 490.375 , 7.259292694975977e-47 ], [ 490.4375 , 7.260349941575994e-47 ], 
                    [ 490.5 , 7.261407342153977e-47 ], [ 490.5625 , 7.262464896732195e-47 ], [ 490.625 , 7.263530942736107e-47 ], [ 490.6875 , 7.264600303712896e-47 ], 
                    [ 490.75 , 7.265669822124532e-47 ], [ 490.8125 , 7.266739497994247e-47 ], [ 490.875 , 7.26780933134512e-47 ], [ 490.9375 , 7.268879322200439e-47 ], 
                    [ 491.0 , 7.269949470583339e-47 ], [ 491.0625 , 7.27101977651696e-47 ], [ 491.125 , 7.272090240024551e-47 ], [ 491.1875 , 7.27316086112931e-47 ], 
                    [ 491.25 , 7.274231639854489e-47 ], [ 491.3125 , 7.275302576223192e-47 ], [ 491.375 , 7.276373670258678e-47 ], [ 491.4375 , 7.277444921984264e-47 ], 
                    [ 491.5 , 7.278516331422905e-47 ], [ 491.5625 , 7.279587898598081e-47 ], [ 491.625 , 7.280659623532911e-47 ], [ 491.6875 , 7.281731506250618e-47 ], 
                    [ 491.75 , 7.282810526867991e-47 ], [ 491.8125 , 7.283890202611247e-47 ], [ 491.875 , 7.284970038416244e-47 ], [ 491.9375 , 7.286050034306762e-47 ], 
                    [ 492.0 , 7.287130190306588e-47 ], [ 492.0625 , 7.288210506439301e-47 ], [ 492.125 , 7.289290982728796e-47 ], [ 492.1875 , 7.290371619198816e-47 ], 
                    [ 492.25 , 7.291452415873057e-47 ], [ 492.3125 , 7.292533372775218e-47 ], [ 492.375 , 7.293614489929102e-47 ], [ 492.4375 , 7.294695767358522e-47 ], 
                    [ 492.5 , 7.295777205087182e-47 ], [ 492.5625 , 7.296858803138799e-47 ], [ 492.625 , 7.29794056153719e-47 ], [ 492.6875 , 7.299022480306076e-47 ], 
                    [ 492.75 , 7.300104559469438e-47 ], [ 492.8125 , 7.301187307755999e-47 ], [ 492.875 , 7.302274699817334e-47 ], [ 492.9375 , 7.303362253827929e-47 ], 
                    [ 493.0 , 7.304449969811693e-47 ], [ 493.0625 , 7.305537847793065e-47 ], [ 493.125 , 7.3066258877960115e-47 ], [ 493.1875 , 7.307714089844563e-47 ], 
                    [ 493.25 , 7.30880245396311e-47 ], [ 493.3125 , 7.309890980175484e-47 ], [ 493.375 , 7.310979668506078e-47 ], [ 493.4375 , 7.312068518978835e-47 ], 
                    [ 493.5 , 7.313157531618109e-47 ], [ 493.5625 , 7.314246706447949e-47 ], [ 493.625 , 7.315336043492459e-47 ], [ 493.6875 , 7.316425542775796e-47 ], 
                    [ 493.75 , 7.317515204322176e-47 ], [ 493.8125 , 7.318605028155871e-47 ], [ 493.875 , 7.319695014300946e-47 ], [ 493.9375 , 7.320785951780703e-47 ], 
                    [ 494.0 , 7.321879699659098e-47 ], [ 494.0625 , 7.32297361094673e-47 ], [ 494.125 , 7.324067685668064e-47 ], [ 494.1875 , 7.325161923847465e-47 ], 
                    [ 494.25 , 7.326256325509564e-47 ], [ 494.3125 , 7.327350890678524e-47 ], [ 494.375 , 7.328445619378825e-47 ], [ 494.4375 , 7.329540511635005e-47 ], 
                    [ 494.5 , 7.330635567471499e-47 ], [ 494.5625 , 7.331730786912589e-47 ], [ 494.625 , 7.33282616998298e-47 ], [ 494.6875 , 7.333921716706857e-47 ], 
                    [ 494.75 , 7.335017427108724e-47 ], [ 494.8125 , 7.336113301213294e-47 ], [ 494.875 , 7.337209339044505e-47 ], [ 494.9375 , 7.33830554062739e-47 ], 
                    [ 495.0 , 7.339401905986051e-47 ], [ 495.0625 , 7.340499188146801e-47 ], [ 495.125 , 7.341598257298382e-47 ], [ 495.1875 , 7.342697491010025e-47 ], 
                    [ 495.25 , 7.343796889306419e-47 ], [ 495.3125 , 7.344896452212158e-47 ], [ 495.375 , 7.345996179751782e-47 ], [ 495.4375 , 7.347096071950099e-47 ], 
                    [ 495.5 , 7.348196128831656e-47 ], [ 495.5625 , 7.349296350421271e-47 ], [ 495.625 , 7.350396736743446e-47 ], [ 495.6875 , 7.351497287822897e-47 ], 
                    [ 495.75 , 7.352598003684295e-47 ], [ 495.8125 , 7.353698884352206e-47 ], [ 495.875 , 7.354799929851621e-47 ], [ 495.9375 , 7.355901140206958e-47 ], 
                    [ 496.0 , 7.3570025154429e-47 ], [ 496.0625 , 7.358104055584134e-47 ], [ 496.125 , 7.359205760655454e-47 ], [ 496.1875 , 7.360308239880652e-47 ], 
                    [ 496.25 , 7.361411951062219e-47 ], [ 496.3125 , 7.362515827750272e-47 ], [ 496.375 , 7.363619869969473e-47 ], [ 496.4375 , 7.364724077744855e-47 ], 
                    [ 496.5 , 7.365828451101136e-47 ], [ 496.5625 , 7.366932990063043e-47 ], [ 496.625 , 7.368037694655672e-47 ], [ 496.6875 , 7.369142564903649e-47 ], 
                    [ 496.75 , 7.370247600831762e-47 ], [ 496.8125 , 7.371352802465118e-47 ], [ 496.875 , 7.372458169828305e-47 ], [ 496.9375 , 7.373563702946276e-47 ], 
                    [ 497.0 , 7.374669401843942e-47 ], [ 497.0625 , 7.375775266546059e-47 ], [ 497.125 , 7.3768812970775375e-47 ], [ 497.1875 , 7.377987493463353e-47 ], 
                    [ 497.25 , 7.379093855728166e-47 ], [ 497.3125 , 7.380200815530085e-47 ], [ 497.375 , 7.381308678211982e-47 ], [ 497.4375 , 7.382416707198312e-47 ], 
                    [ 497.5 , 7.383524902514039e-47 ], [ 497.5625 , 7.384633264183973e-47 ], [ 497.625 , 7.385741792233244e-47 ], [ 497.6875 , 7.386850486686776e-47 ], 
                    [ 497.75 , 7.387959347569548e-47 ], [ 497.8125 , 7.389068374906595e-47 ], [ 497.875 , 7.390177568722747e-47 ], [ 497.9375 , 7.391286929043204e-47 ], 
                    [ 498.0 , 7.392396455892856e-47 ], [ 498.0625 , 7.393506149296597e-47 ], [ 498.125 , 7.394616009279585e-47 ], [ 498.1875 , 7.395726035866721e-47 ], 
                    [ 498.25 , 7.396836229083066e-47 ], [ 498.3125 , 7.397946588953688e-47 ], [ 498.375 , 7.39905711550355e-47 ], [ 498.4375 , 7.400168082310966e-47 ], 
                    [ 498.5 , 7.4012797601091e-47 ], [ 498.5625 , 7.40239160490716e-47 ], [ 498.625 , 7.403503616730128e-47 ], [ 498.6875 , 7.4046157956034135e-47 ], 
                    [ 498.75 , 7.405728141551739e-47 ], [ 498.8125 , 7.406840654600363e-47 ], [ 498.875 , 7.407953334774441e-47 ], [ 498.9375 , 7.409066182099026e-47 ], 
                    [ 499.0 , 7.410179196599122e-47 ], [ 499.0625 , 7.411292378300054e-47 ], [ 499.125 , 7.412405727226886e-47 ], [ 499.1875 , 7.413519243404635e-47 ], 
                    [ 499.25 , 7.414632926858426e-47 ], [ 499.3125 , 7.415746777613544e-47 ], [ 499.375 , 7.416860795694913e-47 ], [ 499.4375 , 7.41797498112772e-47 ], 
                    [ 499.5 , 7.419089333937318e-47 ], [ 499.5625 , 7.420203994909689e-47 ], [ 499.625 , 7.421319226381915e-47 ], [ 499.6875 , 7.4224346254696455e-47 ], 
                    [ 499.75 , 7.423550192198019e-47 ], [ 499.8125 , 7.424665926592392e-47 ], [ 499.875 , 7.425781828677803e-47 ], [ 499.9375 , 7.42689789847951e-47 ], 
                    [ 500.0 , 7.428014136022667e-47 ], [ 500.0625 , 7.429130541332591e-47 ], [ 500.125 , 7.430247114434495e-47 ], [ 500.1875 , 7.431363855353441e-47 ], 
                    [ 500.25 , 7.432480764114862e-47 ], [ 500.3125 , 7.433597840743666e-47 ], [ 500.375 , 7.43471508526556e-47 ], [ 500.4375 , 7.435832497705303e-47 ], 
                    [ 500.5 , 7.436950078088448e-47 ], [ 500.5625 , 7.438067826440129e-47 ], [ 500.625 , 7.439185742785595e-47 ], [ 500.6875 , 7.440303874955401e-47 ], 
                    [ 500.75 , 7.441422463511445e-47 ], [ 500.8125 , 7.442541220238141e-47 ], [ 500.875 , 7.443660145160613e-47 ], [ 500.9375 , 7.444779238304306e-47 ], 
                    [ 501.0 , 7.445898499694353e-47 ], [ 501.0625 , 7.447017929356313e-47 ], [ 501.125 , 7.448137527315272e-47 ], [ 501.1875 , 7.449257293596531e-47 ], 
                    [ 501.25 , 7.450377228225506e-47 ], [ 501.3125 , 7.451497331227343e-47 ], [ 501.375 , 7.452617602627621e-47 ], [ 501.4375 , 7.453738042451344e-47 ], 
                    [ 501.5 , 7.454858650724145e-47 ], [ 501.5625 , 7.455979427471139e-47 ], [ 501.625 , 7.457100372717709e-47 ], [ 501.6875 , 7.458221486489082e-47 ], 
                    [ 501.75 , 7.459342768810804e-47 ], [ 501.8125 , 7.460464219708111e-47 ], [ 501.875 , 7.461586033655756e-47 ], [ 501.9375 , 7.462708019354044e-47 ], 
                    [ 502.0 , 7.463830173763473e-47 ], [ 502.0625 , 7.464952496909201e-47 ], [ 502.125 , 7.466074988816813e-47 ], [ 502.1875 , 7.467197649511417e-47 ], 
                    [ 502.25 , 7.468320479018609e-47 ], [ 502.3125 , 7.469443477363825e-47 ], [ 502.375 , 7.470566644572294e-47 ], [ 502.4375 , 7.471689980669354e-47 ], 
                    [ 502.5 , 7.472813485680665e-47 ], [ 502.5625 , 7.473937159631416e-47 ], [ 502.625 , 7.475061002547061e-47 ], [ 502.6875 , 7.476185014452956e-47 ], 
                    [ 502.75 , 7.477309195374617e-47 ], [ 502.8125 , 7.478433545337353e-47 ], [ 502.875 , 7.479558064366688e-47 ], [ 502.9375 , 7.480682752487937e-47 ], 
                    [ 503.0 , 7.481807657849114e-47 ], [ 503.0625 , 7.482932745502725e-47 ], [ 503.125 , 7.48405800234302e-47 ], [ 503.1875 , 7.48518342839544e-47 ], 
                    [ 503.25 , 7.486309023685432e-47 ], [ 503.3125 , 7.487434788238444e-47 ], [ 503.375 , 7.48856072207993e-47 ], [ 503.4375 , 7.4896868252354e-47 ], 
                    [ 503.5 , 7.490813097730209e-47 ], [ 503.5625 , 7.491939539589874e-47 ], [ 503.625 , 7.49306615083981e-47 ], [ 503.6875 , 7.494192931505649e-47 ], 
                    [ 503.75 , 7.495319881612656e-47 ], [ 503.8125 , 7.4964470011865745e-47 ], [ 503.875 , 7.497574290252572e-47 ], [ 503.9375 , 7.498701748836399e-47 ], 
                    [ 504.0 , 7.499829376963284e-47 ], [ 504.0625 , 7.500957174659039e-47 ], [ 504.125 , 7.502085125863904e-47 ], [ 504.1875 , 7.503213233276837e-47 ], 
                    [ 504.25 , 7.504341510325959e-47 ], [ 504.3125 , 7.505469957037046e-47 ], [ 504.375 , 7.506598573435343e-47 ], [ 504.4375 , 7.507727359546742e-47 ], 
                    [ 504.5 , 7.508856315396387e-47 ], [ 504.5625 , 7.509985441010017e-47 ], [ 504.625 , 7.511114736413105e-47 ], [ 504.6875 , 7.512244201631131e-47 ], 
                    [ 504.75 , 7.513373836689844e-47 ], [ 504.8125 , 7.514503641614567e-47 ], [ 504.875 , 7.515633616430795e-47 ], [ 504.9375 , 7.516763761164229e-47 ], 
                    [ 505.0 , 7.517894075840372e-47 ], [ 505.0625 , 7.519024560484721e-47 ], [ 505.125 , 7.520155215122946e-47 ], [ 505.1875 , 7.5212860397804965e-47 ], 
                    [ 505.25 , 7.522417008107032e-47 ], [ 505.3125 , 7.523548072890868e-47 ], [ 505.375 , 7.524679307740674e-47 ], [ 505.4375 , 7.525810712682235e-47 ], 
                    [ 505.5 , 7.526942287740857e-47 ], [ 505.5625 , 7.528074032942175e-47 ], [ 505.625 , 7.529205948311928e-47 ], [ 505.6875 , 7.530338033875547e-47 ], 
                    [ 505.75 , 7.531470289658724e-47 ], [ 505.8125 , 7.532602715687003e-47 ], [ 505.875 , 7.533735311985927e-47 ], [ 505.9375 , 7.534868078581151e-47 ], 
                    [ 506.0 , 7.536001015498282e-47 ], [ 506.0625 , 7.537134122762981e-47 ], [ 506.125 , 7.538267400400757e-47 ], [ 506.1875 , 7.539400848437278e-47 ], 
                    [ 506.25 , 7.540534466898274e-47 ], [ 506.3125 , 7.541668255809101e-47 ], [ 506.375 , 7.542802215195603e-47 ], [ 506.4375 , 7.543936185222973e-47 ], 
                    [ 506.5 , 7.545070315697322e-47 ], [ 506.5625 , 7.546204616673105e-47 ], [ 506.625 , 7.547339088176007e-47 ], [ 506.6875 , 7.548473730231558e-47 ], 
                    [ 506.75 , 7.549608542865506e-47 ], [ 506.8125 , 7.550743526103333e-47 ], [ 506.875 , 7.551878679970901e-47 ], [ 506.9375 , 7.553014004493704e-47 ], 
                    [ 507.0 , 7.55414949969745e-47 ], [ 507.0625 , 7.555285165607797e-47 ], [ 507.125 , 7.556421002250463e-47 ], [ 507.1875 , 7.557557009650957e-47 ], 
                    [ 507.25 , 7.558693187835161e-47 ], [ 507.3125 , 7.5598295368286445e-47 ], [ 507.375 , 7.560966056657033e-47 ], [ 507.4375 , 7.562102747346008e-47 ], 
                    [ 507.5 , 7.563239608921311e-47 ], [ 507.5625 , 7.564376513176993e-47 ], [ 507.625 , 7.565513494109997e-47 ], [ 507.6875 , 7.56665064593947e-47 ], 
                    [ 507.75 , 7.567787968691259e-47 ], [ 507.8125 , 7.568925462390949e-47 ], [ 507.875 , 7.570063127064127e-47 ], [ 507.9375 , 7.571200962736707e-47 ], 
                    [ 508.0 , 7.572338969434282e-47 ], [ 508.0625 , 7.573477147182507e-47 ], [ 508.125 , 7.574615496007143e-47 ], [ 508.1875 , 7.575754015933907e-47 ], 
                    [ 508.25 , 7.576892706988515e-47 ], [ 508.3125 , 7.578031569196689e-47 ], [ 508.375 , 7.579170602584103e-47 ], [ 508.4375 , 7.58030980717659e-47 ], 
                    [ 508.5 , 7.58144918299994e-47 ], [ 508.5625 , 7.58258873007973e-47 ], [ 508.625 , 7.583728448441644e-47 ], [ 508.6875 , 7.584868295175694e-47 ], 
                    [ 508.75 , 7.586008103926471e-47 ], [ 508.8125 , 7.587148083960905e-47 ], [ 508.875 , 7.58828823530479e-47 ], [ 508.9375 , 7.589428557983709e-47 ], 
                    [ 509.0 , 7.59056905202357e-47 ], [ 509.0625 , 7.59170971745007e-47 ], [ 509.125 , 7.592850554288964e-47 ], [ 509.1875 , 7.593991562566012e-47 ], 
                    [ 509.25 , 7.595132742306976e-47 ], [ 509.3125 , 7.596274093537622e-47 ], [ 509.375 , 7.597415616283721e-47 ], [ 509.4375 , 7.598557310571046e-47 ], 
                    [ 509.5 , 7.599699176425377e-47 ], [ 509.5625 , 7.600841213872495e-47 ], [ 509.625 , 7.601983422938187e-47 ], [ 509.6875 , 7.603125803648242e-47 ], 
                    [ 509.75 , 7.604268356028509e-47 ], [ 509.8125 , 7.605411080104621e-47 ], [ 509.875 , 7.606553753024106e-47 ], [ 509.9375 , 7.607696512033086e-47 ], 
                    [ 510.0 , 7.6088394427227235e-47 ], [ 510.0625 , 7.609982545118757e-47 ], [ 510.125 , 7.611125819246983e-47 ], [ 510.1875 , 7.61226926513347e-47 ], 
                    [ 510.25 , 7.613412882803647e-47 ], [ 510.3125 , 7.614556672283535e-47 ], [ 510.375 , 7.615700633598891e-47 ], [ 510.4375 , 7.616844766775532e-47 ], 
                    [ 510.5 , 7.617989071839169e-47 ], [ 510.5625 , 7.619133548815949e-47 ], [ 510.625 , 7.620278197731212e-47 ], [ 510.6875 , 7.621423018611277e-47 ], 
                    [ 510.75 , 7.622568011481654e-47 ], [ 510.8125 , 7.623713176368399e-47 ], [ 510.875 , 7.624858513297244e-47 ], [ 510.9375 , 7.626004022293928e-47 ], 
                    [ 511.0 , 7.627149628197441e-47 ], [ 511.0625 , 7.628295158379962e-47 ], [ 511.125 , 7.629440860610948e-47 ], [ 511.1875 , 7.630586734916241e-47 ], 
                    [ 511.25 , 7.631732781321685e-47 ], [ 511.3125 , 7.632878999853128e-47 ], [ 511.375 , 7.634025390536421e-47 ], [ 511.4375 , 7.635171953397476e-47 ], 
                    [ 511.5 , 7.636318688461987e-47 ], [ 511.5625 , 7.637465595755982e-47 ], [ 511.625 , 7.638612675305274e-47 ], [ 511.6875 , 7.639759927135679e-47 ], 
                    [ 511.75 , 7.640907351273182e-47 ], [ 511.8125 , 7.642054947743607e-47 ], [ 511.875 , 7.64320271657289e-47 ], [ 511.9375 , 7.644350657786702e-47 ], 
                    [ 512.0 , 7.645498771411206e-47 ], [ 512.0625 , 7.646647057472239e-47 ], [ 512.125 , 7.647795515995486e-47 ], [ 512.1875 , 7.648943890558677e-47 ], 
                    [ 512.25 , 7.650092328304585e-47 ], [ 512.3125 , 7.651240938480844e-47 ], [ 512.375 , 7.652389721113181e-47 ], [ 512.4375 , 7.653538676227544e-47 ], 
                    [ 512.5 , 7.65468780384983e-47 ], [ 512.5625 , 7.655837104005939e-47 ], [ 512.625 , 7.656986576721777e-47 ], [ 512.6875 , 7.658136222023252e-47 ], 
                    [ 512.75 , 7.659286039936277e-47 ], [ 512.8125 , 7.660436030486878e-47 ], [ 512.875 , 7.6615861937007e-47 ], [ 512.9375 , 7.662736529603888e-47 ], 
                    [ 513.0 , 7.663887038222261e-47 ], [ 513.0625 , 7.665037719581913e-47 ], [ 513.125 , 7.666188573708674e-47 ], [ 513.1875 , 7.667339600628316e-47 ], 
                    [ 513.25 , 7.668490800367113e-47 ], [ 513.3125 , 7.669642125072986e-47 ], [ 513.375 , 7.670793295481037e-47 ], [ 513.4375 , 7.671944638673278e-47 ], 
                    [ 513.5 , 7.673096154675695e-47 ], [ 513.5625 , 7.674247843514227e-47 ], [ 513.625 , 7.67539970521487e-47 ], [ 513.6875 , 7.676551739803515e-47 ], 
                    [ 513.75 , 7.677703947306057e-47 ], [ 513.8125 , 7.67885632774845e-47 ], [ 513.875 , 7.680008881156814e-47 ], [ 513.9375 , 7.6811616075570555e-47 ], 
                    [ 514.0 , 7.682314506974868e-47 ], [ 514.0625 , 7.683467579436603e-47 ], [ 514.125 , 7.684620824968121e-47 ], [ 514.1875 , 7.685774243595185e-47 ], 
                    [ 514.25 , 7.686927835343936e-47 ], [ 514.3125 , 7.688081600240523e-47 ], [ 514.375 , 7.689235538310717e-47 ], [ 514.4375 , 7.690389649580401e-47 ], 
                    [ 514.5 , 7.691543720632185e-47 ], [ 514.5625 , 7.692697761532658e-47 ], [ 514.625 , 7.69385197558557e-47 ], [ 514.6875 , 7.69500636281701e-47 ], 
                    [ 514.75 , 7.696160923253017e-47 ], [ 514.8125 , 7.69731565691958e-47 ], [ 514.875 , 7.69847056384258e-47 ], [ 514.9375 , 7.69962564404812e-47 ], 
                    [ 515.0 , 7.700780897562038e-47 ], [ 515.0625 , 7.701936324410501e-47 ], [ 515.125 , 7.703091924619515e-47 ], [ 515.1875 , 7.704247698214927e-47 ], 
                    [ 515.25 , 7.705403645222916e-47 ], [ 515.3125 , 7.706559765669448e-47 ], [ 515.375 , 7.707716059580545e-47 ], [ 515.4375 , 7.708872526982232e-47 ], 
                    [ 515.5 , 7.71002916790054e-47 ], [ 515.5625 , 7.711185982361505e-47 ], [ 515.625 , 7.712342970391164e-47 ], [ 515.6875 , 7.71349977170258e-47 ], 
                    [ 515.75 , 7.714656685036792e-47 ], [ 515.8125 , 7.715813771891315e-47 ], [ 515.875 , 7.71697103229212e-47 ], [ 515.9375 , 7.718128466265183e-47 ], 
                    [ 516.0 , 7.719286073836647e-47 ], [ 516.0625 , 7.720443855032493e-47 ], [ 516.125 , 7.721601809878708e-47 ], [ 516.1875 , 7.722759938401445e-47 ], 
                    [ 516.25 , 7.723918240626701e-47 ], [ 516.3125 , 7.725076716580417e-47 ], [ 516.375 , 7.726235366288871e-47 ], [ 516.4375 , 7.727394189778068e-47 ], 
                    [ 516.5 , 7.728553187073852e-47 ], [ 516.5625 , 7.729712358202512e-47 ], [ 516.625 , 7.730871703190067e-47 ], [ 516.6875 , 7.732031222062645e-47 ], 
                    [ 516.75 , 7.733190914846384e-47 ], [ 516.8125 , 7.734350712402275e-47 ], [ 516.875 , 7.735510304119708e-47 ], [ 516.9375 , 7.73667006969175e-47 ], 
                    [ 517.0 , 7.737830009144633e-47 ], [ 517.0625 , 7.7389901225042055e-47 ], [ 517.125 , 7.740150409796762e-47 ], [ 517.1875 , 7.741310871048159e-47 ], 
                    [ 517.25 , 7.742471506284642e-47 ], [ 517.3125 , 7.743632315532298e-47 ], [ 517.375 , 7.744793298816996e-47 ], [ 517.4375 , 7.745954456165102e-47 ], 
                    [ 517.5 , 7.747115787602714e-47 ], [ 517.5625 , 7.748277293155712e-47 ], [ 517.625 , 7.749438972850257e-47 ], [ 517.6875 , 7.750600826712567e-47 ], 
                    [ 517.75 , 7.7517628547687e-47 ], [ 517.8125 , 7.752925057044717e-47 ], [ 517.875 , 7.754087433566848e-47 ], [ 517.9375 , 7.755249984361162e-47 ], 
                    [ 518.0 , 7.756412524696755e-47 ], [ 518.0625 , 7.757574975056387e-47 ], [ 518.125 , 7.75873759963192e-47 ], [ 518.1875 , 7.7599003984495775e-47 ], 
                    [ 518.25 , 7.761063371535527e-47 ], [ 518.3125 , 7.762226518915553e-47 ], [ 518.375 , 7.763389840616112e-47 ], [ 518.4375 , 7.764553336663213e-47 ], 
                    [ 518.5 , 7.765717007083046e-47 ], [ 518.5625 , 7.766880851901631e-47 ], [ 518.625 , 7.768044871145107e-47 ], [ 518.6875 , 7.769209064839668e-47 ], 
                    [ 518.75 , 7.770373433011571e-47 ], [ 518.8125 , 7.771537975686798e-47 ], [ 518.875 , 7.772702692891557e-47 ], [ 518.9375 , 7.773867584652005e-47 ], 
                    [ 519.0 , 7.775032650994304e-47 ], [ 519.0625 , 7.776197891944617e-47 ], [ 519.125 , 7.777363307529056e-47 ], [ 519.1875 , 7.778528583974311e-47 ], 
                    [ 519.25 , 7.779693847954281e-47 ], [ 519.3125 , 7.780859286496901e-47 ], [ 519.375 , 7.782024899628213e-47 ], [ 519.4375 , 7.78319068737437e-47 ], 
                    [ 519.5 , 7.784356649761697e-47 ], [ 519.5625 , 7.785522786816188e-47 ], [ 519.625 , 7.786689098564069e-47 ], [ 519.6875 , 7.787855585031506e-47 ], 
                    [ 519.75 , 7.789022246244783e-47 ], [ 519.8125 , 7.790189082229805e-47 ], [ 519.875 , 7.7913560930130265e-47 ], [ 519.9375 , 7.79252327862036e-47 ], 
                    [ 520.0 , 7.793690639078381e-47 ], [ 520.0625 , 7.794858174412949e-47 ], [ 520.125 , 7.79602588465043e-47 ], [ 520.1875 , 7.797193769816913e-47 ], 
                    [ 520.25 , 7.79836182993866e-47 ], [ 520.3125 , 7.799530065041935e-47 ], [ 520.375 , 7.800698111015797e-47 ], [ 520.4375 , 7.801866240319691e-47 ], 
                    [ 520.5 , 7.803034544547112e-47 ], [ 520.5625 , 7.804203023724311e-47 ], [ 520.625 , 7.8053716778775425e-47 ], [ 520.6875 , 7.806540507032951e-47 ], 
                    [ 520.75 , 7.807709511216746e-47 ], [ 520.8125 , 7.808878690455189e-47 ], [ 520.875 , 7.810048044774331e-47 ], [ 520.9375 , 7.811217574200608e-47 ], 
                    [ 521.0 , 7.812387278760022e-47 ], [ 521.0625 , 7.813557158479022e-47 ], [ 521.125 , 7.814727213383557e-47 ], [ 521.1875 , 7.815897443500141e-47 ], 
                    [ 521.25 , 7.817067848854896e-47 ], [ 521.3125 , 7.818238429474068e-47 ], [ 521.375 , 7.8194091853839e-47 ], [ 521.4375 , 7.820580116610699e-47 ], 
                    [ 521.5 , 7.821751223180491e-47 ], [ 521.5625 , 7.822922022443847e-47 ], [ 521.625 , 7.824092959862186e-47 ], [ 521.6875 , 7.82526407254697e-47 ], 
                    [ 521.75 , 7.826435360524046e-47 ], [ 521.8125 , 7.827606823819984e-47 ], [ 521.875 , 7.828778462461026e-47 ], [ 521.9375 , 7.829950276473085e-47 ], 
                    [ 522.0 , 7.8311222658828e-47 ], [ 522.0625 , 7.8322944307161435e-47 ], [ 522.125 , 7.8334667709996e-47 ], [ 522.1875 , 7.834639286759315e-47 ], 
                    [ 522.25 , 7.835811978021446e-47 ], [ 522.3125 , 7.836984844812428e-47 ], [ 522.375 , 7.838157887158589e-47 ], [ 522.4375 , 7.839331105085931e-47 ], 
                    [ 522.5 , 7.840504498620842e-47 ], [ 522.5625 , 7.841678067789888e-47 ], [ 522.625 , 7.842851812618968e-47 ], [ 522.6875 , 7.844025724835689e-47 ], 
                    [ 522.75 , 7.84519929171951e-47 ], [ 522.8125 , 7.846373034183827e-47 ], [ 522.875 , 7.847546952255188e-47 ], [ 522.9375 , 7.848721045959641e-47 ], 
                    [ 523.0 , 7.849895315323742e-47 ], [ 523.0625 , 7.851069760373438e-47 ], [ 523.125 , 7.852244381135291e-47 ], [ 523.1875 , 7.853419177635372e-47 ], 
                    [ 523.25 , 7.854594149900135e-47 ], [ 523.3125 , 7.855769297955879e-47 ], [ 523.375 , 7.856944621828793e-47 ], [ 523.4375 , 7.858120121545293e-47 ], 
                    [ 523.5 , 7.859295797131576e-47 ], [ 523.5625 , 7.8604716486140105e-47 ], [ 523.625 , 7.861647676018968e-47 ], [ 523.6875 , 7.862823879372601e-47 ], 
                    [ 523.75 , 7.864000258701347e-47 ], [ 523.8125 , 7.865176814031532e-47 ], [ 523.875 , 7.866353515303085e-47 ], [ 523.9375 , 7.867529918504545e-47 ], 
                    [ 524.0 , 7.868706497635502e-47 ], [ 524.0625 , 7.869883252722489e-47 ], [ 524.125 , 7.871060183791598e-47 ], [ 524.1875 , 7.872237290869367e-47 ], 
                    [ 524.25 , 7.873414573981954e-47 ], [ 524.3125 , 7.874592033155737e-47 ], [ 524.375 , 7.875769668417049e-47 ], [ 524.4375 , 7.876947479792165e-47 ], 
                    [ 524.5 , 7.878125467307649e-47 ], [ 524.5625 , 7.879303630989449e-47 ], [ 524.625 , 7.880481970864246e-47 ], [ 524.6875 , 7.881660486958279e-47 ], 
                    [ 524.75 , 7.882839179297845e-47 ], [ 524.8125 , 7.884018047909526e-47 ], [ 524.875 , 7.885197092819344e-47 ], [ 524.9375 , 7.886376314053891e-47 ], 
                    [ 525.0 , 7.887555711639536e-47 ], [ 525.0625 , 7.888735251161995e-47 ], [ 525.125 , 7.889914436469976e-47 ], [ 525.1875 , 7.891093798039161e-47 ], 
                    [ 525.25 , 7.892273335895956e-47 ], [ 525.3125 , 7.893453050066655e-47 ], [ 525.375 , 7.894632940577558e-47 ], [ 525.4375 , 7.895813007455022e-47 ], 
                    [ 525.5 , 7.896993250725579e-47 ], [ 525.5625 , 7.898173670415427e-47 ], [ 525.625 , 7.899354266550993e-47 ], [ 525.6875 , 7.900535039158765e-47 ], 
                    [ 525.75 , 7.901715988264838e-47 ], [ 525.8125 , 7.902897113895822e-47 ], [ 525.875 , 7.904078416078101e-47 ], [ 525.9375 , 7.905259894837955e-47 ], 
                    [ 526.0 , 7.90644155020189e-47 ], [ 526.0625 , 7.9076233821962485e-47 ], [ 526.125 , 7.908805390847318e-47 ], [ 526.1875 , 7.909987576181733e-47 ], 
                    [ 526.25 , 7.911169928122032e-47 ], [ 526.3125 , 7.912351931669084e-47 ], [ 526.375 , 7.913534111818581e-47 ], [ 526.4375 , 7.91471646859702e-47 ], 
                    [ 526.5 , 7.915899002030678e-47 ], [ 526.5625 , 7.9170817121460055e-47 ], [ 526.625 , 7.918264598969401e-47 ], [ 526.6875 , 7.919447662527323e-47 ], 
                    [ 526.75 , 7.920630902846121e-47 ], [ 526.8125 , 7.921814319952145e-47 ], [ 526.875 , 7.922997913871927e-47 ], [ 526.9375 , 7.924181684631823e-47 ], 
                    [ 527.0 , 7.925365632258203e-47 ], [ 527.0625 , 7.926549756777546e-47 ], [ 527.125 , 7.927734058216338e-47 ], [ 527.1875 , 7.928918536600899e-47 ], 
                    [ 527.25 , 7.930103191957782e-47 ], [ 527.3125 , 7.931288024313314e-47 ], [ 527.375 , 7.932473033693995e-47 ], [ 527.4375 , 7.933658220126276e-47 ], 
                    [ 527.5 , 7.934843075191259e-47 ], [ 527.5625 , 7.936028071228071e-47 ], [ 527.625 , 7.937213244233117e-47 ], [ 527.6875 , 7.938398594232937e-47 ], 
                    [ 527.75 , 7.939584121253965e-47 ], [ 527.8125 , 7.940769825322579e-47 ], [ 527.875 , 7.941955706465051e-47 ], [ 527.9375 , 7.943141764708223e-47 ], 
                    [ 528.0 , 7.944328000078144e-47 ], [ 528.0625 , 7.945514412601495e-47 ], [ 528.125 , 7.946701002304731e-47 ], [ 528.1875 , 7.947887769214086e-47 ], 
                    [ 528.25 , 7.949074713356365e-47 ], [ 528.3125 , 7.950261834757865e-47 ], [ 528.375 , 7.951449133444888e-47 ], [ 528.4375 , 7.952636609444251e-47 ], 
                    [ 528.5 , 7.953824262782205e-47 ], [ 528.5625 , 7.955012093485406e-47 ], [ 528.625 , 7.95620010158006e-47 ], [ 528.6875 , 7.957387789915913e-47 ], 
                    [ 528.75 , 7.958575535433253e-47 ], [ 528.8125 , 7.959763458237283e-47 ], [ 528.875 , 7.960951558354582e-47 ], [ 528.9375 , 7.96213983581156e-47 ], 
                    [ 529.0 , 7.96332829063463e-47 ], [ 529.0625 , 7.964516922850264e-47 ], [ 529.125 , 7.965705732485113e-47 ], [ 529.1875 , 7.966894719565544e-47 ], 
                    [ 529.25 , 7.968083884117876e-47 ], [ 529.3125 , 7.969273226168992e-47 ], [ 529.375 , 7.970462745745105e-47 ], [ 529.4375 , 7.971652442872713e-47 ], 
                    [ 529.5 , 7.972842317578429e-47 ], [ 529.5625 , 7.974032369888649e-47 ], [ 529.625 , 7.975222599829993e-47 ], [ 529.6875 , 7.976413007428749e-47 ], 
                    [ 529.75 , 7.977603592711834e-47 ], [ 529.8125 , 7.978794355705426e-47 ], [ 529.875 , 7.979984941439449e-47 ], [ 529.9375 , 7.981175499653472e-47 ], 
                    [ 530.0 , 7.982366235490553e-47 ], [ 530.0625 , 7.983557148977018e-47 ], [ 530.125 , 7.984748240139601e-47 ], [ 530.1875 , 7.985939509004639e-47 ], 
                    [ 530.25 , 7.9871309555987e-47 ], [ 530.3125 , 7.988322579948301e-47 ], [ 530.375 , 7.989514382080018e-47 ], [ 530.4375 , 7.990706362020262e-47 ], 
                    [ 530.5 , 7.991898519795617e-47 ], [ 530.5625 , 7.993090855432615e-47 ], [ 530.625 , 7.994283368957792e-47 ], [ 530.6875 , 7.995476060397744e-47 ], 
                    [ 530.75 , 7.996668929778957e-47 ], [ 530.8125 , 7.997861977127922e-47 ], [ 530.875 , 7.99905520247125e-47 ], [ 530.9375 , 8.000248605835551e-47 ], 
                    [ 531.0 , 8.00144218724727e-47 ], [ 531.0625 , 8.002635692886074e-47 ], [ 531.125 , 8.003829023867465e-47 ], [ 531.1875 , 8.005022532795028e-47 ], 
                    [ 531.25 , 8.006216219695352e-47 ], [ 531.3125 , 8.007410084594976e-47 ], [ 531.375 , 8.008604127520559e-47 ], [ 531.4375 , 8.009798348498419e-47 ], 
                    [ 531.5 , 8.010992747555162e-47 ], [ 531.5625 , 8.012187324717517e-47 ], [ 531.625 , 8.013382080011867e-47 ], [ 531.6875 , 8.014577013464836e-47 ], 
                    [ 531.75 , 8.015772125103044e-47 ], [ 531.8125 , 8.016967414953007e-47 ], [ 531.875 , 8.018162883041185e-47 ], [ 531.9375 , 8.019358529394382e-47 ], 
                    [ 532.0 , 8.020554354039014e-47 ], [ 532.0625 , 8.021750357001719e-47 ], [ 532.125 , 8.022946538309149e-47 ], [ 532.1875 , 8.024142897987781e-47 ], 
                    [ 532.25 , 8.025339342409102e-47 ], [ 532.3125 , 8.026535472877402e-47 ], [ 532.375 , 8.027731781622037e-47 ], [ 532.4375 , 8.028928268669638e-47 ], 
                    [ 532.5 , 8.030124934046778e-47 ], [ 532.5625 , 8.031321777779866e-47 ], [ 532.625 , 8.032518799895654e-47 ], [ 532.6875 , 8.033716000420617e-47 ], 
                    [ 532.75 , 8.034913379381461e-47 ], [ 532.8125 , 8.036110936804777e-47 ], [ 532.875 , 8.037308672716996e-47 ], [ 532.9375 , 8.038506587144948e-47 ], 
                    [ 533.0 , 8.039704680115068e-47 ], [ 533.0625 , 8.040902951653909e-47 ], [ 533.125 , 8.042101401788258e-47 ], [ 533.1875 , 8.043300030544792e-47 ], 
                    [ 533.25 , 8.044498837950017e-47 ], [ 533.3125 , 8.045697824030445e-47 ], [ 533.375 , 8.046896988812881e-47 ], [ 533.4375 , 8.0480963323239e-47 ], 
                    [ 533.5 , 8.049295312476481e-47 ], [ 533.5625 , 8.050494381102867e-47 ], [ 533.625 , 8.051693628349251e-47 ], [ 533.6875 , 8.052893054242298e-47 ], 
                    [ 533.75 , 8.054092658808734e-47 ], [ 533.8125 , 8.055292442075006e-47 ], [ 533.875 , 8.056492404067732e-47 ], [ 533.9375 , 8.057692544813594e-47 ], 
                    [ 534.0 , 8.058892864339334e-47 ], [ 534.0625 , 8.06009336267147e-47 ], [ 534.125 , 8.061294039836524e-47 ], [ 534.1875 , 8.062494895861249e-47 ], 
                    [ 534.25 , 8.063695930772348e-47 ], [ 534.3125 , 8.064897144596354e-47 ], [ 534.375 , 8.066098537359974e-47 ], [ 534.4375 , 8.067300109089981e-47 ], 
                    [ 534.5 , 8.068501859812747e-47 ], [ 534.5625 , 8.069703789555282e-47 ], [ 534.625 , 8.070905898343964e-47 ], [ 534.6875 , 8.072107873395201e-47 ], 
                    [ 534.75 , 8.073309728496031e-47 ], [ 534.8125 , 8.074511762540912e-47 ], [ 534.875 , 8.075713975556546e-47 ], [ 534.9375 , 8.076916367569463e-47 ], 
                    [ 535.0 , 8.078118938606429e-47 ], [ 535.0625 , 8.079321688693984e-47 ], [ 535.125 , 8.080524617858959e-47 ], [ 535.1875 , 8.081727726127788e-47 ], 
                    [ 535.25 , 8.082931013527309e-47 ], [ 535.3125 , 8.084134480084078e-47 ], [ 535.375 , 8.085338125824944e-47 ], [ 535.4375 , 8.086541950776297e-47 ], 
                    [ 535.5 , 8.087745954965105e-47 ], [ 535.5625 , 8.088950138417943e-47 ], [ 535.625 , 8.090154501161441e-47 ], [ 535.6875 , 8.091359043222469e-47 ], 
                    [ 535.75 , 8.092563764627552e-47 ], [ 535.8125 , 8.093768665403391e-47 ], [ 535.875 , 8.094973669437374e-47 ], [ 535.9375 , 8.096178301958286e-47 ], 
                    [ 536.0 , 8.097383113743522e-47 ], [ 536.0625 , 8.098588104819756e-47 ], [ 536.125 , 8.099793275213556e-47 ], [ 536.1875 , 8.100998624951665e-47 ], 
                    [ 536.25 , 8.102204154060828e-47 ], [ 536.3125 , 8.103409862567564e-47 ], [ 536.375 , 8.104615750498744e-47 ], [ 536.4375 , 8.105821817880952e-47 ], 
                    [ 536.5 , 8.1070280647410675e-47 ], [ 536.5625 , 8.108234491105623e-47 ], [ 536.625 , 8.109441097001449e-47 ], [ 536.6875 , 8.110647882455031e-47 ], 
                    [ 536.75 , 8.111854847493433e-47 ], [ 536.8125 , 8.11306199214321e-47 ], [ 536.875 , 8.114269316430972e-47 ], [ 536.9375 , 8.115476820383684e-47 ], 
                    [ 537.0 , 8.116684504027967e-47 ], [ 537.0625 , 8.117892367390502e-47 ], [ 537.125 , 8.11909996530455e-47 ], [ 537.1875 , 8.120307540485966e-47 ], 
                    [ 537.25 , 8.12151529527317e-47 ], [ 537.3125 , 8.122723229692991e-47 ], [ 537.375 , 8.123931343772089e-47 ], [ 537.4375 , 8.125139637537185e-47 ], 
                    [ 537.5 , 8.12634811101506e-47 ], [ 537.5625 , 8.127556764232216e-47 ], [ 537.625 , 8.12876559721573e-47 ], [ 537.6875 , 8.129974609992166e-47 ], 
                    [ 537.75 , 8.131183802588092e-47 ], [ 537.8125 , 8.132393175030541e-47 ], [ 537.875 , 8.133602727346263e-47 ], [ 537.9375 , 8.134812459561725e-47 ], 
                    [ 538.0 , 8.13602237170391e-47 ], [ 538.0625 , 8.137232463799581e-47 ], [ 538.125 , 8.138442735875215e-47 ], [ 538.1875 , 8.139653187958043e-47 ], 
                    [ 538.25 , 8.140863820074489e-47 ], [ 538.3125 , 8.142074499798231e-47 ], [ 538.375 , 8.143284859531992e-47 ], [ 538.4375 , 8.144495399191787e-47 ], 
                    [ 538.5 , 8.145706118804129e-47 ], [ 538.5625 , 8.146917018396116e-47 ], [ 538.625 , 8.148128097994329e-47 ], [ 538.6875 , 8.149339357625528e-47 ], 
                    [ 538.75 , 8.150550797316475e-47 ], [ 538.8125 , 8.151762417093822e-47 ], [ 538.875 , 8.152974216984513e-47 ], [ 538.9375 , 8.154186197015437e-47 ], 
                    [ 539.0 , 8.155398357212969e-47 ], [ 539.0625 , 8.156610697604181e-47 ], [ 539.125 , 8.157823218215743e-47 ], [ 539.1875 , 8.159035919074679e-47 ], 
                    [ 539.25 , 8.160248800207551e-47 ], [ 539.3125 , 8.161461861641098e-47 ], [ 539.375 , 8.162675103402358e-47 ], [ 539.4375 , 8.163888525517901e-47 ], 
                    [ 539.5 , 8.165102128014834e-47 ], [ 539.5625 , 8.166315420085057e-47 ], [ 539.625 , 8.167528669158898e-47 ], [ 539.6875 , 8.168742098482238e-47 ], 
                    [ 539.75 , 8.169955708081683e-47 ], [ 539.8125 , 8.171169497984131e-47 ], [ 539.875 , 8.172383468216252e-47 ], [ 539.9375 , 8.173597618804956e-47 ], 
                    [ 540.0 , 8.174811949776979e-47 ], [ 540.0625 , 8.17602646115912e-47 ], [ 540.125 , 8.177241152978123e-47 ], [ 540.1875 , 8.178456025260796e-47 ], 
                    [ 540.25 , 8.179671078034183e-47 ], [ 540.3125 , 8.180886311324807e-47 ], [ 540.375 , 8.182101725159721e-47 ], [ 540.4375 , 8.183317319565456e-47 ], 
                    [ 540.5 , 8.18453309456913e-47 ], [ 540.5625 , 8.185749050197459e-47 ], [ 540.625 , 8.186965186477219e-47 ], [ 540.6875 , 8.188181503435365e-47 ], 
                    [ 540.75 , 8.189397924586377e-47 ], [ 540.8125 , 8.190613970334723e-47 ], [ 540.875 , 8.191830196654124e-47 ], [ 540.9375 , 8.193046603571099e-47 ], 
                    [ 541.0 , 8.194263191112758e-47 ], [ 541.0625 , 8.195479959305748e-47 ], [ 541.125 , 8.196696908177011e-47 ], [ 541.1875 , 8.197914037753257e-47 ], 
                    [ 541.25 , 8.199131348061379e-47 ], [ 541.3125 , 8.200348839128213e-47 ], [ 541.375 , 8.201566510980602e-47 ], [ 541.4375 , 8.202784363645503e-47 ], 
                    [ 541.5 , 8.20400239714942e-47 ], [ 541.5625 , 8.20522061151967e-47 ], [ 541.625 , 8.206439006782702e-47 ], [ 541.6875 , 8.207657582965551e-47 ], 
                    [ 541.75 , 8.208876340095316e-47 ], [ 541.8125 , 8.210095278198398e-47 ], [ 541.875 , 8.211314397302024e-47 ], [ 541.9375 , 8.212533697432948e-47 ], 
                    [ 542.0 , 8.213752827221839e-47 ], [ 542.0625 , 8.214971800645699e-47 ], [ 542.125 , 8.216190954972947e-47 ], [ 542.1875 , 8.21741029023055e-47 ], 
                    [ 542.25 , 8.218629806445299e-47 ], [ 542.3125 , 8.219849503643992e-47 ], [ 542.375 , 8.221069381853545e-47 ], [ 542.4375 , 8.222289441100882e-47 ], 
                    [ 542.5 , 8.22350968141275e-47 ], [ 542.5625 , 8.224730102816138e-47 ], [ 542.625 , 8.225950705337865e-47 ], [ 542.6875 , 8.227171489004751e-47 ], 
                    [ 542.75 , 8.228392453843911e-47 ], [ 542.8125 , 8.229613599881707e-47 ], [ 542.875 , 8.230834927145673e-47 ], [ 542.9375 , 8.232056435662237e-47 ], 
                    [ 543.0 , 8.233278125458591e-47 ], [ 543.0625 , 8.23449999656146e-47 ], [ 543.125 , 8.235722048997812e-47 ], [ 543.1875 , 8.236944282794615e-47 ], 
                    [ 543.25 , 8.238166130333819e-47 ], [ 543.3125 , 8.239388062868225e-47 ], [ 543.375 , 8.240610176646628e-47 ], [ 543.4375 , 8.241832471696145e-47 ], 
                    [ 543.5 , 8.243054948043545e-47 ], [ 543.5625 , 8.244277605715604e-47 ], [ 543.625 , 8.24550044473951e-47 ], [ 543.6875 , 8.246723465141807e-47 ], 
                    [ 543.75 , 8.247946666949754e-47 ], [ 543.8125 , 8.249170050189964e-47 ], [ 543.875 , 8.250393614889464e-47 ], [ 543.9375 , 8.251617361075228e-47 ], 
                    [ 544.0 , 8.252841288774117e-47 ], [ 544.0625 , 8.254065398013114e-47 ], [ 544.125 , 8.255289688818968e-47 ], [ 544.1875 , 8.256514161218729e-47 ], 
                    [ 544.25 , 8.257738815239388e-47 ], [ 544.3125 , 8.258963650907712e-47 ], [ 544.375 , 8.260188668250875e-47 ], [ 544.4375 , 8.261413750788206e-47 ], 
                    [ 544.5 , 8.262638381513888e-47 ], [ 544.5625 , 8.263863193772593e-47 ], [ 544.625 , 8.26508818759153e-47 ], [ 544.6875 , 8.2663133629973765e-47 ], 
                    [ 544.75 , 8.267538720017048e-47 ], [ 544.8125 , 8.268764258677468e-47 ], [ 544.875 , 8.269989979005737e-47 ], [ 544.9375 , 8.271215881028607e-47 ], 
                    [ 545.0 , 8.272441964773133e-47 ], [ 545.0625 , 8.273668230266131e-47 ], [ 545.125 , 8.274894677534547e-47 ], [ 545.1875 , 8.276121306605497e-47 ], 
                    [ 545.25 , 8.277348117505878e-47 ], [ 545.3125 , 8.278575110262461e-47 ], [ 545.375 , 8.279802284902326e-47 ], [ 545.4375 , 8.281029641452431e-47 ], 
                    [ 545.5 , 8.282257179939743e-47 ], [ 545.5625 , 8.283484900391349e-47 ], [ 545.625 , 8.284712802833988e-47 ], [ 545.6875 , 8.285940620695828e-47 ], 
                    [ 545.75 , 8.287168238470102e-47 ], [ 545.8125 , 8.288396038124139e-47 ], [ 545.875 , 8.289624019685002e-47 ], [ 545.9375 , 8.290852183179526e-47 ], 
                    [ 546.0 , 8.292080528634782e-47 ], [ 546.0625 , 8.293309056077672e-47 ], [ 546.125 , 8.294537765535037e-47 ], [ 546.1875 , 8.295766657034081e-47 ], 
                    [ 546.25 , 8.296995730601599e-47 ], [ 546.3125 , 8.298224986264741e-47 ], [ 546.375 , 8.299454424050252e-47 ], [ 546.4375 , 8.300684043985289e-47 ], 
                    [ 546.5 , 8.301913846096665e-47 ], [ 546.5625 , 8.303143830411602e-47 ], [ 546.625 , 8.304373996956979e-47 ], [ 546.6875 , 8.305604345759739e-47 ], 
                    [ 546.75 , 8.306834876847057e-47 ], [ 546.8125 , 8.308065590245704e-47 ], [ 546.875 , 8.309296485982869e-47 ], [ 546.9375 , 8.310527097350371e-47 ], 
                    [ 547.0 , 8.311757607190838e-47 ], [ 547.0625 , 8.312988299228468e-47 ], [ 547.125 , 8.314219173490238e-47 ], [ 547.1875 , 8.315450230003191e-47 ], 
                    [ 547.25 , 8.316681468794133e-47 ], [ 547.3125 , 8.317912889890291e-47 ], [ 547.375 , 8.31914449331842e-47 ], [ 547.4375 , 8.320376279105755e-47 ], 
                    [ 547.5 , 8.321608247279061e-47 ], [ 547.5625 , 8.322840397865579e-47 ], [ 547.625 , 8.324072730892141e-47 ], [ 547.6875 , 8.3253052463857e-47 ], 
                    [ 547.75 , 8.326537944373572e-47 ], [ 547.8125 , 8.327770824882479e-47 ], [ 547.875 , 8.329003887939746e-47 ], [ 547.9375 , 8.330237133572102e-47 ], 
                    [ 548.0 , 8.331470561806643e-47 ], [ 548.0625 , 8.332704172670582e-47 ], [ 548.125 , 8.333937966190841e-47 ], [ 548.1875 , 8.335171386702216e-47 ], 
                    [ 548.25 , 8.336404840093289e-47 ], [ 548.3125 , 8.337638476012916e-47 ], [ 548.375 , 8.338872294488225e-47 ], [ 548.4375 , 8.340106295546114e-47 ], 
                    [ 548.5 , 8.341340479213719e-47 ], [ 548.5625 , 8.342574845518004e-47 ], [ 548.625 , 8.343809394485997e-47 ], [ 548.6875 , 8.345044126144725e-47 ], 
                    [ 548.75 , 8.346279040521169e-47 ], [ 548.8125 , 8.347514137642424e-47 ], [ 548.875 , 8.348749417535593e-47 ], [ 548.9375 , 8.349984880227545e-47 ], 
                    [ 549.0 , 8.351220525745567e-47 ], [ 549.0625 , 8.352456354116657e-47 ], [ 549.125 , 8.353692365367694e-47 ], [ 549.1875 , 8.35492855952592e-47 ], 
                    [ 549.25 , 8.356164936618341e-47 ], [ 549.3125 , 8.35740149667197e-47 ], [ 549.375 , 8.358638239713941e-47 ], [ 549.4375 , 8.359874464838859e-47 ], 
                    [ 549.5 , 8.361110802583969e-47 ], [ 549.5625 , 8.362347323170583e-47 ], [ 549.625 , 8.363584026625621e-47 ], [ 549.6875 , 8.364820912976128e-47 ], 
                    [ 549.75 , 8.366057982249092e-47 ], [ 549.8125 , 8.367295234471687e-47 ], [ 549.875 , 8.368532669670906e-47 ], [ 549.9375 , 8.369770287873754e-47 ], 
                    [ 550.0 , 8.371008089107411e-47 ], [ 550.0625 , 8.372246073398827e-47 ], [ 550.125 , 8.373484240775255e-47 ], [ 550.1875 , 8.37472259126347e-47 ], 
                    [ 550.25 , 8.375961124890853e-47 ], [ 550.3125 , 8.377199841684367e-47 ], [ 550.375 , 8.378438741671162e-47 ], [ 550.4375 , 8.379677824878092e-47 ], 
                    [ 550.5 , 8.380917091332609e-47 ], [ 550.5625 , 8.382156541061518e-47 ], [ 550.625 , 8.38339617409204e-47 ], [ 550.6875 , 8.384635295542635e-47 ], 
                    [ 550.75 , 8.385874589530917e-47 ], [ 550.8125 , 8.387114066693487e-47 ], [ 550.875 , 8.388353727057536e-47 ], [ 550.9375 , 8.389593570649905e-47 ], 
                    [ 551.0 , 8.390833597497795e-47 ], [ 551.0625 , 8.392073807628294e-47 ], [ 551.125 , 8.393314201068548e-47 ], [ 551.1875 , 8.394554777845534e-47 ], 
                    [ 551.25 , 8.395795537986411e-47 ], [ 551.3125 , 8.397036481518279e-47 ], [ 551.375 , 8.398277608468245e-47 ], [ 551.4375 , 8.399518918863538e-47 ], 
                    [ 551.5 , 8.400760412731035e-47 ], [ 551.5625 , 8.402002090098034e-47 ], [ 551.625 , 8.403243950991534e-47 ], [ 551.6875 , 8.404485995438723e-47 ], 
                    [ 551.75 , 8.405728223466733e-47 ], [ 551.8125 , 8.406970635102754e-47 ], [ 551.875 , 8.40821321803244e-47 ], [ 551.9375 , 8.409455209768227e-47 ], 
                    [ 552.0 , 8.410697384960703e-47 ], [ 552.0625 , 8.411939743637208e-47 ], [ 552.125 , 8.413182285824545e-47 ], [ 552.1875 , 8.414425011549879e-47 ], 
                    [ 552.25 , 8.415667920840503e-47 ], [ 552.3125 , 8.416911013723412e-47 ], [ 552.375 , 8.418154290225662e-47 ], [ 552.4375 , 8.419397750374497e-47 ], 
                    [ 552.5 , 8.420641394196865e-47 ], [ 552.5625 , 8.421885221720136e-47 ], [ 552.625 , 8.423129232971383e-47 ], [ 552.6875 , 8.424373427977507e-47 ], 
                    [ 552.75 , 8.425617806765951e-47 ], [ 552.8125 , 8.42686236936368e-47 ], [ 552.875 , 8.428107115797966e-47 ], [ 552.9375 , 8.429352046095783e-47 ], 
                    [ 553.0 , 8.43059716028441e-47 ], [ 553.0625 , 8.431842458391131e-47 ], [ 553.125 , 8.433087939513263e-47 ], [ 553.1875 , 8.43433287909386e-47 ], 
                    [ 553.25 , 8.435578002459337e-47 ], [ 553.3125 , 8.436823309637005e-47 ], [ 553.375 , 8.43806880065382e-47 ], [ 553.4375 , 8.439314475536981e-47 ], 
                    [ 553.5 , 8.44056033431357e-47 ], [ 553.5625 , 8.441806377010858e-47 ], [ 553.625 , 8.443052603655873e-47 ], [ 553.6875 , 8.444299014275951e-47 ], 
                    [ 553.75 , 8.445545608898073e-47 ], [ 553.8125 , 8.4467923875494e-47 ], [ 553.875 , 8.448039350257221e-47 ], [ 553.9375 , 8.449286497048586e-47 ], 
                    [ 554.0 , 8.450533827950793e-47 ], [ 554.0625 , 8.451781342990901e-47 ], [ 554.125 , 8.453029042196331e-47 ], [ 554.1875 , 8.454276925593912e-47 ], 
                    [ 554.25 , 8.455524993211077e-47 ], [ 554.3125 , 8.456773245074959e-47 ], [ 554.375 , 8.458021681212699e-47 ], [ 554.4375 , 8.459269574053185e-47 ], 
                    [ 554.5 , 8.46051760574188e-47 ], [ 554.5625 , 8.461765821558129e-47 ], [ 554.625 , 8.463014221528857e-47 ], [ 554.6875 , 8.464262805681355e-47 ], 
                    [ 554.75 , 8.465511574042615e-47 ], [ 554.8125 , 8.466760526639993e-47 ], [ 554.875 , 8.468009663500611e-47 ], [ 554.9375 , 8.469258984651654e-47 ], 
                    [ 555.0 , 8.470508490120252e-47 ], [ 555.0625 , 8.471758179933656e-47 ], [ 555.125 , 8.473008054119245e-47 ], [ 555.1875 , 8.474258112703922e-47 ], 
                    [ 555.25 , 8.475508355715007e-47 ], [ 555.3125 , 8.476758783179712e-47 ], [ 555.375 , 8.478009395125192e-47 ], [ 555.4375 , 8.479260191578721e-47 ], 
                    [ 555.5 , 8.480511172567582e-47 ], [ 555.5625 , 8.4817623381189425e-47 ], [ 555.625 , 8.483013688260088e-47 ], [ 555.6875 , 8.484264555863533e-47 ], 
                    [ 555.75 , 8.48551547732337e-47 ], [ 555.8125 , 8.486766583219286e-47 ], [ 555.875 , 8.488017873578477e-47 ], [ 555.9375 , 8.489269348428078e-47 ], 
                    [ 556.0 , 8.4905210077954115e-47 ], [ 556.0625 , 8.491772851707622e-47 ], [ 556.125 , 8.493024880191798e-47 ], [ 556.1875 , 8.494277093275396e-47 ], 
                    [ 556.25 , 8.495529490985511e-47 ], [ 556.3125 , 8.496782073349305e-47 ], [ 556.375 , 8.498034840394124e-47 ], [ 556.4375 , 8.499287792147075e-47 ], 
                    [ 556.5 , 8.5005409286355145e-47 ], [ 556.5625 , 8.501794249886618e-47 ], [ 556.625 , 8.503047755927627e-47 ], [ 556.6875 , 8.504301446785668e-47 ], 
                    [ 556.75 , 8.50555532248823e-47 ], [ 556.8125 , 8.506809383062445e-47 ], [ 556.875 , 8.508063628535512e-47 ], [ 556.9375 , 8.509317535725411e-47 ], 
                    [ 557.0 , 8.510571387569625e-47 ], [ 557.0625 , 8.511825424169364e-47 ], [ 557.125 , 8.51307964555197e-47 ], [ 557.1875 , 8.514334051744734e-47 ], 
                    [ 557.25 , 8.515588642774706e-47 ], [ 557.3125 , 8.516843418669362e-47 ], [ 557.375 , 8.518098379455822e-47 ], [ 557.4375 , 8.519353525161329e-47 ], 
                    [ 557.5 , 8.520608855813073e-47 ], [ 557.5625 , 8.521864371438364e-47 ], [ 557.625 , 8.523120072064518e-47 ], [ 557.6875 , 8.524375957718797e-47 ], 
                    [ 557.75 , 8.525632028428162e-47 ], [ 557.8125 , 8.526888284220303e-47 ], [ 557.875 , 8.528144725122251e-47 ], [ 557.9375 , 8.529401351161465e-47 ], 
                    [ 558.0 , 8.53065816236498e-47 ], [ 558.0625 , 8.531915158760204e-47 ], [ 558.125 , 8.533172340374361e-47 ], [ 558.1875 , 8.534429304408974e-47 ], 
                    [ 558.25 , 8.535686032520113e-47 ], [ 558.3125 , 8.536942945689322e-47 ], [ 558.375 , 8.538200043944037e-47 ], [ 558.4375 , 8.539457327311267e-47 ], 
                    [ 558.5 , 8.540714795818393e-47 ], [ 558.5625 , 8.541972449492737e-47 ], [ 558.625 , 8.543230288361447e-47 ], [ 558.6875 , 8.54448831245185e-47 ], 
                    [ 558.75 , 8.545746521791286e-47 ], [ 558.8125 , 8.54700491640697e-47 ], [ 558.875 , 8.548263496326063e-47 ], [ 558.9375 , 8.549522261576097e-47 ], 
                    [ 559.0 , 8.550781212184238e-47 ], [ 559.0625 , 8.552040348177663e-47 ], [ 559.125 , 8.55329966958391e-47 ], [ 559.1875 , 8.554559176430163e-47 ], 
                    [ 559.25 , 8.555818868743727e-47 ], [ 559.3125 , 8.557078746551853e-47 ], [ 559.375 , 8.558338809881977e-47 ], [ 559.4375 , 8.559598864116775e-47 ], 
                    [ 559.5 , 8.560858508986977e-47 ], [ 559.5625 , 8.562118339228684e-47 ], [ 559.625 , 8.563378354869053e-47 ], [ 559.6875 , 8.564638555935367e-47 ], 
                    [ 559.75 , 8.565898942455099e-47 ], [ 559.8125 , 8.567159514455355e-47 ], [ 559.875 , 8.568420271963493e-47 ], [ 559.9375 , 8.569681215006752e-47 ], 
                    [ 560.0 , 8.570942343612618e-47 ], [ 560.0625 , 8.572203657808215e-47 ], [ 560.125 , 8.573465157620917e-47 ], [ 560.1875 , 8.5747268430781e-47 ], 
                    [ 560.25 , 8.575988714207021e-47 ], [ 560.3125 , 8.577250771035007e-47 ], [ 560.375 , 8.578513013589264e-47 ], [ 560.4375 , 8.579775441897243e-47 ], 
                    [ 560.5 , 8.581038055986404e-47 ], [ 560.5625 , 8.582300855883902e-47 ], [ 560.625 , 8.583563841617082e-47 ], [ 560.6875 , 8.584827013213413e-47 ], 
                    [ 560.75 , 8.586089609000727e-47 ], [ 560.8125 , 8.587352357414759e-47 ], [ 560.875 , 8.588615291539935e-47 ], [ 560.9375 , 8.589878411403808e-47 ], 
                    [ 561.0 , 8.591141717033514e-47 ], [ 561.0625 , 8.592405208456496e-47 ], [ 561.125 , 8.593668885699893e-47 ], [ 561.1875 , 8.594932748791216e-47 ], 
                    [ 561.25 , 8.59619679775774e-47 ], [ 561.3125 , 8.597461032626797e-47 ], [ 561.375 , 8.598725453425855e-47 ], [ 561.4375 , 8.599990060182009e-47 ], 
                    [ 561.5 , 8.601254852922672e-47 ], [ 561.5625 , 8.602519831675318e-47 ], [ 561.625 , 8.60378499646718e-47 ], [ 561.6875 , 8.605050347325866e-47 ], 
                    [ 561.75 , 8.606315884278371e-47 ], [ 561.8125 , 8.60758160735237e-47 ], [ 561.875 , 8.608847516575052e-47 ], [ 561.9375 , 8.610113611973919e-47 ], 
                    [ 562.0 , 8.611379375032754e-47 ], [ 562.0625 , 8.612645012094327e-47 ], [ 562.125 , 8.613910835169907e-47 ], [ 562.1875 , 8.615176844286956e-47 ], 
                    [ 562.25 , 8.616443039472568e-47 ], [ 562.3125 , 8.617709420754215e-47 ], [ 562.375 , 8.618975988159432e-47 ], [ 562.4375 , 8.620242741715144e-47 ], 
                    [ 562.5 , 8.621509681449139e-47 ], [ 562.5625 , 8.622776807388476e-47 ], [ 562.625 , 8.624044119560641e-47 ], [ 562.6875 , 8.62531161799313e-47 ], 
                    [ 562.75 , 8.626579302713196e-47 ], [ 562.8125 , 8.627847173748034e-47 ], [ 562.875 , 8.629115231125331e-47 ], [ 562.9375 , 8.630383474872356e-47 ], 
                    [ 563.0 , 8.631651905016435e-47 ], [ 563.0625 , 8.632920521585213e-47 ], [ 563.125 , 8.634189324605593e-47 ], [ 563.1875 , 8.635458314105474e-47 ], 
                    [ 563.25 , 8.636727281267199e-47 ], [ 563.3125 , 8.637995787729849e-47 ], [ 563.375 , 8.639264480502498e-47 ], [ 563.4375 , 8.640533359612447e-47 ], 
                    [ 563.5 , 8.641802425087188e-47 ], [ 563.5625 , 8.643071676953969e-47 ], [ 563.625 , 8.644341115240353e-47 ], [ 563.6875 , 8.645610739973472e-47 ], 
                    [ 563.75 , 8.646880551180896e-47 ], [ 563.8125 , 8.648150548889949e-47 ], [ 563.875 , 8.649420733127965e-47 ], [ 563.9375 , 8.650691103922523e-47 ], 
                    [ 564.0 , 8.651961661300715e-47 ], [ 564.0625 , 8.653232405290254e-47 ], [ 564.125 , 8.654503335918547e-47 ], [ 564.1875 , 8.65577445321264e-47 ], 
                    [ 564.25 , 8.657045757200315e-47 ], [ 564.3125 , 8.65831724790881e-47 ], [ 564.375 , 8.6595889253656705e-47 ], [ 564.4375 , 8.660860789598265e-47 ], 
                    [ 564.5 , 8.662132840634025e-47 ], [ 564.5625 , 8.663404410801026e-47 ], [ 564.625 , 8.664676018649753e-47 ], [ 564.6875 , 8.665947813144226e-47 ], 
                    [ 564.75 , 8.667219794311658e-47 ], [ 564.8125 , 8.66849196217951e-47 ], [ 564.875 , 8.669764316775124e-47 ], [ 564.9375 , 8.671036858126089e-47 ], 
                    [ 565.0 , 8.672309586259576e-47 ], [ 565.0625 , 8.673582501203121e-47 ], [ 565.125 , 8.674855602984143e-47 ], [ 565.1875 , 8.676128891630314e-47 ], 
                    [ 565.25 , 8.67740236716857e-47 ], [ 565.3125 , 8.678676029626587e-47 ], [ 565.375 , 8.6799498790318e-47 ], [ 565.4375 , 8.681223915411714e-47 ], 
                    [ 565.5 , 8.682498138793772e-47 ], [ 565.5625 , 8.683772549205298e-47 ], [ 565.625 , 8.685047146673744e-47 ], [ 565.6875 , 8.686321931226688e-47 ], 
                    [ 565.75 , 8.687596902891533e-47 ], [ 565.8125 , 8.688871771318816e-47 ], [ 565.875 , 8.690146274963164e-47 ], [ 565.9375 , 8.69142096555463e-47 ], 
                    [ 566.0 , 8.692695843120635e-47 ], [ 566.0625 , 8.693970907688546e-47 ], [ 566.125 , 8.695246159285915e-47 ], [ 566.1875 , 8.696521597940112e-47 ], 
                    [ 566.25 , 8.697797223678578e-47 ], [ 566.3125 , 8.699073036528754e-47 ], [ 566.375 , 8.700349036518022e-47 ], [ 566.4375 , 8.701625223673959e-47 ], 
                    [ 566.5 , 8.702901598023956e-47 ], [ 566.5625 , 8.70417815959547e-47 ], [ 566.625 , 8.705454908415964e-47 ], [ 566.6875 , 8.706731844512966e-47 ], 
                    [ 566.75 , 8.708008967913822e-47 ], [ 566.8125 , 8.709286278645944e-47 ], [ 566.875 , 8.710563776737059e-47 ], [ 566.9375 , 8.711841462214587e-47 ], 
                    [ 567.0 , 8.713119335105827e-47 ], [ 567.0625 , 8.714397395438457e-47 ], [ 567.125 , 8.715674918209427e-47 ], [ 567.1875 , 8.71695247525604e-47 ], 
                    [ 567.25 , 8.718230219568992e-47 ], [ 567.3125 , 8.719508151175611e-47 ], [ 567.375 , 8.720786270103474e-47 ], [ 567.4375 , 8.722064576379915e-47 ], 
                    [ 567.5 , 8.72334307003252e-47 ], [ 567.5625 , 8.724621751088631e-47 ], [ 567.625 , 8.725900619575842e-47 ], [ 567.6875 , 8.727179675521626e-47 ], 
                    [ 567.75 , 8.728458918953275e-47 ], [ 567.8125 , 8.729738349898521e-47 ], [ 567.875 , 8.731017968384599e-47 ], [ 567.9375 , 8.732297774439125e-47 ], 
                    [ 568.0 , 8.733577768089717e-47 ], [ 568.0625 , 8.734857949363625e-47 ], [ 568.125 , 8.736138318288536e-47 ], [ 568.1875 , 8.737418874891834e-47 ], 
                    [ 568.25 , 8.738699619201089e-47 ], [ 568.3125 , 8.739980551243942e-47 ], [ 568.375 , 8.74126145168044e-47 ], [ 568.4375 , 8.742541915915757e-47 ], 
                    [ 568.5 , 8.743822567719897e-47 ], [ 568.5625 , 8.745103407120522e-47 ], [ 568.625 , 8.746384434144801e-47 ], [ 568.6875 , 8.747665648820528e-47 ], 
                    [ 568.75 , 8.748947051175069e-47 ], [ 568.8125 , 8.750228641235851e-47 ], [ 568.875 , 8.751510419030435e-47 ], [ 568.9375 , 8.752792384586381e-47 ], 
                    [ 569.0 , 8.754074537931072e-47 ], [ 569.0625 , 8.755356879092076e-47 ], [ 569.125 , 8.756639408096967e-47 ], [ 569.1875 , 8.7579221249732e-47 ], 
                    [ 569.25 , 8.759205029748236e-47 ], [ 569.3125 , 8.760488122449657e-47 ], [ 569.375 , 8.761771403104992e-47 ], [ 569.4375 , 8.763054871741716e-47 ], 
                    [ 569.5 , 8.76433852838767e-47 ], [ 569.5625 , 8.765622373069902e-47 ], [ 569.625 , 8.766906405816326e-47 ], [ 569.6875 , 8.768190044082017e-47 ], 
                    [ 569.75 , 8.769473553568838e-47 ], [ 569.8125 , 8.770757250938913e-47 ], [ 569.875 , 8.772041136219804e-47 ], [ 569.9375 , 8.773325209438955e-47 ], 
                    [ 570.0 , 8.774609470624066e-47 ], [ 570.0625 , 8.775893919802525e-47 ], [ 570.125 , 8.777178557001792e-47 ], [ 570.1875 , 8.77846338224945e-47 ], 
                    [ 570.25 , 8.779748395573025e-47 ], [ 570.3125 , 8.781033597000174e-47 ], [ 570.375 , 8.782318986558184e-47 ], [ 570.4375 , 8.783604564274594e-47 ], 
                    [ 570.5 , 8.784890330177255e-47 ], [ 570.5625 , 8.78617628429353e-47 ], [ 570.625 , 8.787462426650909e-47 ], [ 570.6875 , 8.78874875727707e-47 ], 
                    [ 570.75 , 8.79003527619951e-47 ], [ 570.8125 , 8.791321983445542e-47 ], [ 570.875 , 8.792608879043422e-47 ], [ 570.9375 , 8.793895963019972e-47 ], 
                    [ 571.0 , 8.795182345774896e-47 ], [ 571.0625 , 8.796468906625413e-47 ], [ 571.125 , 8.797755655674347e-47 ], [ 571.1875 , 8.799042592949098e-47 ], 
                    [ 571.25 , 8.800329718477266e-47 ], [ 571.3125 , 8.801617032286448e-47 ], [ 571.375 , 8.802904534404063e-47 ], [ 571.4375 , 8.804192224857844e-47 ], 
                    [ 571.5 , 8.80548010367509e-47 ], [ 571.5625 , 8.806768170883543e-47 ], [ 571.625 , 8.808056426510698e-47 ], [ 571.6875 , 8.809344870584055e-47 ], 
                    [ 571.75 , 8.810633503131304e-47 ], [ 571.8125 , 8.81192232417989e-47 ], [ 571.875 , 8.813211333757451e-47 ], [ 571.9375 , 8.814500531891564e-47 ], 
                    [ 572.0 , 8.81578991860981e-47 ], [ 572.0625 , 8.817079493939967e-47 ], [ 572.125 , 8.81836925790912e-47 ], [ 572.1875 , 8.819659210545427e-47 ], 
                    [ 572.25 , 8.820949109702453e-47 ], [ 572.3125 , 8.822238583978454e-47 ], [ 572.375 , 8.823528246753826e-47 ], [ 572.4375 , 8.82481809805625e-47 ], 
                    [ 572.5 , 8.826108137913033e-47 ], [ 572.5625 , 8.827398366352178e-47 ], [ 572.625 , 8.828688783400878e-47 ], [ 572.6875 , 8.829979389086766e-47 ], 
                    [ 572.75 , 8.831270183437671e-47 ], [ 572.8125 , 8.832561166480728e-47 ], [ 572.875 , 8.833852338244027e-47 ], [ 572.9375 , 8.835143698754713e-47 ], 
                    [ 573.0 , 8.836435248040758e-47 ], [ 573.0625 , 8.837726986129443e-47 ], [ 573.125 , 8.83901891304843e-47 ], [ 573.1875 , 8.840311028825448e-47 ], 
                    [ 573.25 , 8.84160333348804e-47 ], [ 573.3125 , 8.842895827063885e-47 ], [ 573.375 , 8.844188509580533e-47 ], [ 573.4375 , 8.84548138106548e-47 ], 
                    [ 573.5 , 8.846774441546663e-47 ], [ 573.5625 , 8.84806719739046e-47 ], [ 573.625 , 8.849359704724572e-47 ], [ 573.6875 , 8.850652400865352e-47 ], 
                    [ 573.75 , 8.851945285840448e-47 ], [ 573.8125 , 8.853238359677627e-47 ], [ 573.875 , 8.854531622404166e-47 ], [ 573.9375 , 8.855825074047971e-47 ], 
                    [ 574.0 , 8.857118714636641e-47 ], [ 574.0625 , 8.85841254419746e-47 ], [ 574.125 , 8.859706562758286e-47 ], [ 574.1875 , 8.861000770346662e-47 ], 
                    [ 574.25 , 8.862295166990204e-47 ], [ 574.3125 , 8.863589752716526e-47 ], [ 574.375 , 8.864884527553188e-47 ], [ 574.4375 , 8.866179491528005e-47 ], 
                    [ 574.5 , 8.867474644668348e-47 ], [ 574.5625 , 8.868769987002171e-47 ], [ 574.625 , 8.87006551855673e-47 ], [ 574.6875 , 8.87136123935992e-47 ], 
                    [ 574.75 , 8.872657149439258e-47 ], [ 574.8125 , 8.87395324882252e-47 ], [ 574.875 , 8.875248878778801e-47 ], [ 574.9375 , 8.87654445471014e-47 ], 
                    [ 575.0 , 8.877840219764851e-47 ], [ 575.0625 , 8.87913617397054e-47 ], [ 575.125 , 8.88043231735476e-47 ], [ 575.1875 , 8.881728649945182e-47 ], 
                    [ 575.25 , 8.883025171769497e-47 ], [ 575.3125 , 8.88432188285526e-47 ], [ 575.375 , 8.8856187832301e-47 ], [ 575.4375 , 8.88691587292165e-47 ], 
                    [ 575.5 , 8.88821315195742e-47 ], [ 575.5625 , 8.889510620365362e-47 ], [ 575.625 , 8.890808278172806e-47 ], [ 575.6875 , 8.892106125407593e-47 ], 
                    [ 575.75 , 8.893404162097366e-47 ], [ 575.8125 , 8.8947023882696e-47 ], [ 575.875 , 8.896000803952017e-47 ], [ 575.9375 , 8.897299409172592e-47 ], 
                    [ 576.0 , 8.898598203958492e-47 ], [ 576.0625 , 8.899897188337765e-47 ], [ 576.125 , 8.901196362337964e-47 ], [ 576.1875 , 8.902494901916194e-47 ], 
                    [ 576.25 , 8.903793522271928e-47 ], [ 576.3125 , 8.905092332059435e-47 ], [ 576.375 , 8.906391331306221e-47 ], [ 576.4375 , 8.907690520039986e-47 ], 
                    [ 576.5 , 8.908989898288435e-47 ], [ 576.5625 , 8.91028946607902e-47 ], [ 576.625 , 8.911589223439645e-47 ], [ 576.6875 , 8.912889170397837e-47 ], 
                    [ 576.75 , 8.914189306981187e-47 ], [ 576.8125 , 8.915489633217424e-47 ], [ 576.875 , 8.916790149134145e-47 ], [ 576.9375 , 8.91809085475921e-47 ], 
                    [ 577.0 , 8.919391750120166e-47 ], [ 577.0625 , 8.920692835244629e-47 ], [ 577.125 , 8.921994110160403e-47 ], [ 577.1875 , 8.923295574894981e-47 ], 
                    [ 577.25 , 8.92459722947625e-47 ], [ 577.3125 , 8.925899073931771e-47 ], [ 577.375 , 8.927201108289496e-47 ], [ 577.4375 , 8.928503332576808e-47 ], 
                    [ 577.5 , 8.92980483839814e-47 ], [ 577.5625 , 8.931106524094397e-47 ], [ 577.625 , 8.932408399535702e-47 ], [ 577.6875 , 8.93371046474958e-47 ], 
                    [ 577.75 , 8.935012719763762e-47 ], [ 577.8125 , 8.936315164605976e-47 ], [ 577.875 , 8.937617799303708e-47 ], [ 577.9375 , 8.938920623884816e-47 ], 
                    [ 578.0 , 8.940223638377047e-47 ], [ 578.0625 , 8.941526842807828e-47 ], [ 578.125 , 8.942830237204975e-47 ], [ 578.1875 , 8.944133821596244e-47 ], 
                    [ 578.25 , 8.945437596009135e-47 ], [ 578.3125 , 8.946741560471541e-47 ], [ 578.375 , 8.948045715011164e-47 ], [ 578.4375 , 8.949350059655648e-47 ], 
                    [ 578.5 , 8.95065459443258e-47 ], [ 578.5625 , 8.951959319369795e-47 ], [ 578.625 , 8.95326423449521e-47 ], [ 578.6875 , 8.954569339836168e-47 ], 
                    [ 578.75 , 8.955874590958396e-47 ], [ 578.8125 , 8.957179147277404e-47 ], [ 578.875 , 8.958483893624378e-47 ], [ 578.9375 , 8.959788830027123e-47 ], 
                    [ 579.0 , 8.961093956513135e-47 ], [ 579.0625 , 8.96239927311029e-47 ], [ 579.125 , 8.963704779846221e-47 ], [ 579.1875 , 8.965010476748623e-47 ], 
                    [ 579.25 , 8.966316363845132e-47 ], [ 579.3125 , 8.967622441163644e-47 ], [ 579.375 , 8.968928708731676e-47 ], [ 579.4375 , 8.970235166577068e-47 ], 
                    [ 579.5 , 8.971541814727285e-47 ], [ 579.5625 , 8.972848653210425e-47 ], [ 579.625 , 8.974155682053963e-47 ], [ 579.6875 , 8.975462901285688e-47 ], 
                    [ 579.75 , 8.976770310933335e-47 ], [ 579.8125 , 8.978077911024702e-47 ], [ 579.875 , 8.979385701587405e-47 ], [ 579.9375 , 8.980693682649189e-47 ], 
                    [ 580.0 , 8.982001854237865e-47 ], [ 580.0625 , 8.983310160899096e-47 ], [ 580.125 , 8.984617768618937e-47 ], [ 580.1875 , 8.985925566673668e-47 ], 
                    [ 580.25 , 8.987233555091e-47 ], [ 580.3125 , 8.988541733898827e-47 ], [ 580.375 , 8.989850103124739e-47 ], [ 580.4375 , 8.991158662796262e-47 ], 
                    [ 580.5 , 8.992467412941496e-47 ], [ 580.5625 , 8.993776353587916e-47 ], [ 580.625 , 8.995085484763314e-47 ], [ 580.6875 , 8.99639480649542e-47 ], 
                    [ 580.75 , 8.997704318811974e-47 ], [ 580.8125 , 8.999014021740782e-47 ], [ 580.875 , 9.000323915309459e-47 ], [ 580.9375 , 9.001633999545885e-47 ], 
                    [ 581.0 , 9.002944274477618e-47 ], [ 581.0625 , 9.004254740132549e-47 ], [ 581.125 , 9.005565396538628e-47 ], [ 581.1875 , 9.006876243723238e-47 ], 
                    [ 581.25 , 9.008187281714401e-47 ], [ 581.3125 , 9.009498510539833e-47 ], [ 581.375 , 9.010809908194991e-47 ], [ 581.4375 , 9.012120527961736e-47 ], 
                    [ 581.5 , 9.013431338357885e-47 ], [ 581.5625 , 9.014742339411038e-47 ], [ 581.625 , 9.016053531148926e-47 ], [ 581.6875 , 9.017364913599348e-47 ], 
                    [ 581.75 , 9.01867648678985e-47 ], [ 581.8125 , 9.01998825074843e-47 ], [ 581.875 , 9.021300205502454e-47 ], [ 581.9375 , 9.022612351080121e-47 ], 
                    [ 582.0 , 9.023924687508991e-47 ], [ 582.0625 , 9.025237214816828e-47 ], [ 582.125 , 9.026549933031328e-47 ], [ 582.1875 , 9.027862842180322e-47 ], 
                    [ 582.25 , 9.029175942291712e-47 ], [ 582.3125 , 9.030489233393014e-47 ], [ 582.375 , 9.03180271551233e-47 ], [ 582.4375 , 9.033116388677061e-47 ], 
                    [ 582.5 , 9.034430252915182e-47 ], [ 582.5625 , 9.035744308254551e-47 ], [ 582.625 , 9.037058554722902e-47 ], [ 582.6875 , 9.038372992348029e-47 ], 
                    [ 582.75 , 9.039686742622754e-47 ], [ 582.8125 , 9.041000625696605e-47 ], [ 582.875 , 9.042314699738215e-47 ], [ 582.9375 , 9.043628964775343e-47 ], 
                    [ 583.0 , 9.044943420835813e-47 ], [ 583.0625 , 9.046258067947195e-47 ], [ 583.125 , 9.04757290613758e-47 ], [ 583.1875 , 9.048887935434358e-47 ], 
                    [ 583.25 , 9.050203155865557e-47 ], [ 583.3125 , 9.051518567458897e-47 ], [ 583.375 , 9.052834170242225e-47 ], [ 583.4375 , 9.054149964243202e-47 ], 
                    [ 583.5 , 9.055465949489685e-47 ], [ 583.5625 , 9.056782126009408e-47 ], [ 583.625 , 9.058098493830234e-47 ], [ 583.6875 , 9.059415052980225e-47 ], 
                    [ 583.75 , 9.060731803486808e-47 ], [ 583.8125 , 9.062048745377854e-47 ], [ 583.875 , 9.063365878681184e-47 ], [ 583.9375 , 9.06468320342481e-47 ], 
                    [ 584.0 , 9.06600071963637e-47 ], [ 584.0625 , 9.067317611494318e-47 ], [ 584.125 , 9.06863449813914e-47 ], [ 584.1875 , 9.069951576041168e-47 ], 
                    [ 584.25 , 9.071268845228313e-47 ], [ 584.3125 , 9.072586305728286e-47 ], [ 584.375 , 9.073903957568747e-47 ], [ 584.4375 , 9.075221800777873e-47 ], 
                    [ 584.5 , 9.076539835383002e-47 ], [ 584.5625 , 9.077858061412127e-47 ], [ 584.625 , 9.07917647889318e-47 ], [ 584.6875 , 9.080495087853769e-47 ], 
                    [ 584.75 , 9.08181388832164e-47 ], [ 584.8125 , 9.083132880324803e-47 ], [ 584.875 , 9.084452063891137e-47 ], [ 584.9375 , 9.085771439048274e-47 ], 
                    [ 585.0 , 9.08709100582397e-47 ], [ 585.0625 , 9.088410764246443e-47 ], [ 585.125 , 9.089730714343076e-47 ], [ 585.1875 , 9.091050856141964e-47 ], 
                    [ 585.25 , 9.09237118967089e-47 ], [ 585.3125 , 9.09369171495756e-47 ], [ 585.375 , 9.09501184758629e-47 ], [ 585.4375 , 9.09633182251633e-47 ], 
                    [ 585.5 , 9.097651989016634e-47 ], [ 585.5625 , 9.09897234711507e-47 ], [ 585.625 , 9.100292896839249e-47 ], [ 585.6875 , 9.101613638217243e-47 ], 
                    [ 585.75 , 9.10293457127667e-47 ], [ 585.8125 , 9.104255696045415e-47 ], [ 585.875 , 9.105577012551304e-47 ], [ 585.9375 , 9.10689852082216e-47 ], 
                    [ 586.0 , 9.108220220885815e-47 ], [ 586.0625 , 9.109542112770237e-47 ], [ 586.125 , 9.110864196502938e-47 ], [ 586.1875 , 9.112186472111957e-47 ], 
                    [ 586.25 , 9.11350893962527e-47 ], [ 586.3125 , 9.1148315990706e-47 ], [ 586.375 , 9.116154450475607e-47 ], [ 586.4375 , 9.117477493868476e-47 ], 
                    [ 586.5 , 9.118800729277002e-47 ], [ 586.5625 , 9.120124156728994e-47 ], [ 586.625 , 9.121447776252188e-47 ], [ 586.6875 , 9.122771180399716e-47 ], 
                    [ 586.75 , 9.124094159929379e-47 ], [ 586.8125 , 9.12541733131691e-47 ], [ 586.875 , 9.126740694590068e-47 ], [ 586.9375 , 9.128064249776615e-47 ], 
                    [ 587.0 , 9.129387996904446e-47 ], [ 587.0625 , 9.13071193600146e-47 ], [ 587.125 , 9.132036067095435e-47 ], [ 587.1875 , 9.133360390214211e-47 ], 
                    [ 587.25 , 9.134684905385509e-47 ], [ 587.3125 , 9.136009612637435e-47 ], [ 587.375 , 9.137334511997655e-47 ], [ 587.4375 , 9.138659603494218e-47 ], 
                    [ 587.5 , 9.139984887154666e-47 ], [ 587.5625 , 9.141310363007064e-47 ], [ 587.625 , 9.142636031079407e-47 ], [ 587.6875 , 9.143961891399444e-47 ], 
                    [ 587.75 , 9.145287943995056e-47 ], [ 587.8125 , 9.146614188894061e-47 ], [ 587.875 , 9.147940626124346e-47 ], [ 587.9375 , 9.149267255714127e-47 ], 
                    [ 588.0 , 9.150593954546256e-47 ], [ 588.0625 , 9.151919958730191e-47 ], [ 588.125 , 9.153246155064112e-47 ], [ 588.1875 , 9.154572543575863e-47 ], 
                    [ 588.25 , 9.155899124293487e-47 ], [ 588.3125 , 9.157225897244643e-47 ], [ 588.375 , 9.158552862457252e-47 ], [ 588.4375 , 9.159880019959173e-47 ], 
                    [ 588.5 , 9.161207369778338e-47 ], [ 588.5625 , 9.162534911942416e-47 ], [ 588.625 , 9.163862646479544e-47 ], [ 588.6875 , 9.165190573417336e-47 ], 
                    [ 588.75 , 9.166518692783936e-47 ], [ 588.8125 , 9.16784700460703e-47 ], [ 588.875 , 9.16917550891457e-47 ], [ 588.9375 , 9.170504205734456e-47 ], 
                    [ 589.0 , 9.171833095094514e-47 ], [ 589.0625 , 9.17316217702297e-47 ], [ 589.125 , 9.174491451547279e-47 ], [ 589.1875 , 9.175820918695473e-47 ], 
                    [ 589.25 , 9.177150578495663e-47 ], [ 589.3125 , 9.178480430975506e-47 ], [ 589.375 , 9.179809701323226e-47 ], [ 589.4375 , 9.181138967529124e-47 ], 
                    [ 589.5 , 9.18246842621708e-47 ], [ 589.5625 , 9.18379807741516e-47 ], [ 589.625 , 9.185127921150847e-47 ], [ 589.6875 , 9.186457957452218e-47 ], 
                    [ 589.75 , 9.187788186347225e-47 ], [ 589.8125 , 9.189118607863685e-47 ], [ 589.875 , 9.190449222029432e-47 ], [ 589.9375 , 9.191780028872553e-47 ], 
                    [ 590.0 , 9.19311102842082e-47 ], [ 590.0625 , 9.194442220701942e-47 ], [ 590.125 , 9.195773605744085e-47 ], [ 590.1875 , 9.197105183575295e-47 ], 
                    [ 590.25 , 9.198436954223097e-47 ], [ 590.3125 , 9.199768917715605e-47 ], [ 590.375 , 9.20110107408081e-47 ], [ 590.4375 , 9.202433423346579e-47 ], 
                    [ 590.5 , 9.203765965540905e-47 ], [ 590.5625 , 9.20509870069153e-47 ], [ 590.625 , 9.20643162882672e-47 ], [ 590.6875 , 9.207764292486257e-47 ], 
                    [ 590.75 , 9.209096529908889e-47 ], [ 590.8125 , 9.210428960088098e-47 ], [ 590.875 , 9.211761583051712e-47 ], [ 590.9375 , 9.21309439882756e-47 ], 
                    [ 591.0 , 9.214427407443598e-47 ], [ 591.0625 , 9.215760608927865e-47 ], [ 591.125 , 9.217094003308132e-47 ], [ 591.1875 , 9.218427590612112e-47 ], 
                    [ 591.25 , 9.219761370867983e-47 ], [ 591.3125 , 9.22109534410366e-47 ], [ 591.375 , 9.222429510347065e-47 ], [ 591.4375 , 9.223763869625862e-47 ], 
                    [ 591.5 , 9.225098421968308e-47 ], [ 591.5625 , 9.22643316740214e-47 ], [ 591.625 , 9.22776810595536e-47 ], [ 591.6875 , 9.22910323765611e-47 ], 
                    [ 591.75 , 9.23043856253194e-47 ], [ 591.8125 , 9.23177408061106e-47 ], [ 591.875 , 9.233109791921362e-47 ], [ 591.9375 , 9.234445696490671e-47 ], 
                    [ 592.0 , 9.235781790514235e-47 ], [ 592.0625 , 9.237117083354338e-47 ], [ 592.125 , 9.238452569248762e-47 ], [ 592.1875 , 9.239788248225417e-47 ], 
                    [ 592.25 , 9.241124120312023e-47 ], [ 592.3125 , 9.242460185536826e-47 ], [ 592.375 , 9.243796443927424e-47 ], [ 592.4375 , 9.245132895511806e-47 ], 
                    [ 592.5 , 9.246469540318039e-47 ], [ 592.5625 , 9.247806378374056e-47 ], [ 592.625 , 9.249143409707668e-47 ], [ 592.6875 , 9.250480634347013e-47 ], 
                    [ 592.75 , 9.251818052319777e-47 ], [ 592.8125 , 9.253155663653913e-47 ], [ 592.875 , 9.25449346837777e-47 ], [ 592.9375 , 9.255831466518916e-47 ], 
                    [ 593.0 , 9.257169658105374e-47 ], [ 593.0625 , 9.258508043165315e-47 ], [ 593.125 , 9.25984662172658e-47 ], [ 593.1875 , 9.26118539381701e-47 ], 
                    [ 593.25 , 9.262524359464786e-47 ], [ 593.3125 , 9.263863518697953e-47 ], [ 593.375 , 9.26520232162089e-47 ], [ 593.4375 , 9.266540838523935e-47 ], 
                    [ 593.5 , 9.267879548798701e-47 ], [ 593.5625 , 9.269218452472926e-47 ], [ 593.625 , 9.270557549574814e-47 ], [ 593.6875 , 9.271896840132107e-47 ], 
                    [ 593.75 , 9.273236324172824e-47 ], [ 593.8125 , 9.274576001724915e-47 ], [ 593.875 , 9.275915872816333e-47 ], [ 593.9375 , 9.27725593747511e-47 ], 
                    [ 594.0 , 9.27859619572914e-47 ], [ 594.0625 , 9.27993664760626e-47 ], [ 594.125 , 9.28127729313464e-47 ], [ 594.1875 , 9.282618132342324e-47 ], 
                    [ 594.25 , 9.28395916525696e-47 ], [ 594.3125 , 9.285300391906729e-47 ], [ 594.375 , 9.286641812319823e-47 ], [ 594.4375 , 9.287983426523768e-47 ], 
                    [ 594.5 , 9.289325234547088e-47 ], [ 594.5625 , 9.290667236417323e-47 ], [ 594.625 , 9.292009432162742e-47 ], [ 594.6875 , 9.293351805026231e-47 ], 
                    [ 594.75 , 9.29469332878389e-47 ], [ 594.8125 , 9.296035046194539e-47 ], [ 594.875 , 9.297376957286266e-47 ], [ 594.9375 , 9.298719062086962e-47 ], 
                    [ 595.0 , 9.30006136062446e-47 ], [ 595.0625 , 9.301403852926992e-47 ], [ 595.125 , 9.302746539022327e-47 ], [ 595.1875 , 9.30408941893857e-47 ], 
                    [ 595.25 , 9.305432492703505e-47 ], [ 595.3125 , 9.306775760345313e-47 ], [ 595.375 , 9.308119221891978e-47 ], [ 595.4375 , 9.309462877371429e-47 ], 
                    [ 595.5 , 9.310806726811657e-47 ], [ 595.5625 , 9.312150770240595e-47 ], [ 595.625 , 9.313495007686312e-47 ], [ 595.6875 , 9.314839439176884e-47 ], 
                    [ 595.75 , 9.316184064740056e-47 ], [ 595.8125 , 9.317528884404304e-47 ], [ 595.875 , 9.318873898197183e-47 ], [ 595.9375 , 9.320219106147052e-47 ], 
                    [ 596.0 , 9.321564508281804e-47 ], [ 596.0625 , 9.32290961744746e-47 ], [ 596.125 , 9.324254322287813e-47 ], [ 596.1875 , 9.325599221083896e-47 ], 
                    [ 596.25 , 9.326944313863548e-47 ], [ 596.3125 , 9.328289600654752e-47 ], [ 596.375 , 9.329635081485623e-47 ], [ 596.4375 , 9.330980756384147e-47 ], 
                    [ 596.5 , 9.332326625378188e-47 ], [ 596.5625 , 9.333672688495803e-47 ], [ 596.625 , 9.335018945765057e-47 ], [ 596.6875 , 9.33636539721376e-47 ], 
                    [ 596.75 , 9.337712042870184e-47 ], [ 596.8125 , 9.339058882762138e-47 ], [ 596.875 , 9.340405916917643e-47 ], [ 596.9375 , 9.341753145364914e-47 ], 
                    [ 597.0 , 9.343100568131776e-47 ], [ 597.0625 , 9.344448185246459e-47 ], [ 597.125 , 9.345795996736662e-47 ], [ 597.1875 , 9.347144002630751e-47 ], 
                    [ 597.25 , 9.348492202956436e-47 ], [ 597.3125 , 9.349840597742225e-47 ], [ 597.375 , 9.351189187015703e-47 ], [ 597.4375 , 9.352537118910441e-47 ], 
                    [ 597.5 , 9.353885074045246e-47 ], [ 597.5625 , 9.355233223457205e-47 ], [ 597.625 , 9.356581567173987e-47 ], [ 597.6875 , 9.357930105223858e-47 ], 
                    [ 597.75 , 9.359278837634765e-47 ], [ 597.8125 , 9.360627764434787e-47 ], [ 597.875 , 9.361976885651805e-47 ], [ 597.9375 , 9.363326201313908e-47 ], 
                    [ 598.0 , 9.364675711449119e-47 ], [ 598.0625 , 9.36602541608547e-47 ], [ 598.125 , 9.367375315250992e-47 ], [ 598.1875 , 9.368725408973723e-47 ], 
                    [ 598.25 , 9.370075697281702e-47 ], [ 598.3125 , 9.37142618020311e-47 ], [ 598.375 , 9.372776857765727e-47 ], [ 598.4375 , 9.374127729997739e-47 ], 
                    [ 598.5 , 9.375478796927205e-47 ], [ 598.5625 , 9.376830058582254e-47 ], [ 598.625 , 9.378181514990816e-47 ], [ 598.6875 , 9.379533166181027e-47 ], 
                    [ 598.75 , 9.380884826299416e-47 ], [ 598.8125 , 9.382235759458733e-47 ], [ 598.875 , 9.383586887164969e-47 ], [ 598.9375 , 9.38493820944568e-47 ], 
                    [ 599.0 , 9.386289726329285e-47 ], [ 599.0625 , 9.387641437843674e-47 ], [ 599.125 , 9.388993344016876e-47 ], [ 599.1875 , 9.390345444876923e-47 ], 
                    [ 599.25 , 9.39169774045172e-47 ], [ 599.3125 , 9.393050230769573e-47 ], [ 599.375 , 9.394402915858396e-47 ], [ 599.4375 , 9.395755795746167e-47 ], 
                    [ 599.5 , 9.397108870461076e-47 ], [ 599.5625 , 9.398462140031042e-47 ], [ 599.625 , 9.399815604484198e-47 ], [ 599.6875 , 9.401169263848605e-47 ], 
                    [ 599.75 , 9.402523118152401e-47 ], [ 599.8125 , 9.403877167423658e-47 ], [ 599.875 , 9.405231411690252e-47 ], [ 599.9375 , 9.406585850980467e-47 ], 
                    [ 600.0 , 9.407940485322252e-47 ], [ 600.0625 , 9.409295314743899e-47 ], [ 600.125 , 9.410649847293931e-47 ], [ 600.1875 , 9.412003982504477e-47 ], 
                    [ 600.25 , 9.413358312566898e-47 ], [ 600.3125 , 9.414712837509035e-47 ], [ 600.375 , 9.416067557359133e-47 ], [ 600.4375 , 9.417422472145166e-47 ], 
                    [ 600.5 , 9.41877758189512e-47 ], [ 600.5625 , 9.420132886637248e-47 ], [ 600.625 , 9.421488386399345e-47 ], [ 600.6875 , 9.422844081209736e-47 ], 
                    [ 600.75 , 9.42419997109622e-47 ], [ 600.8125 , 9.425556056087074e-47 ], [ 600.875 , 9.426912336210364e-47 ], [ 600.9375 , 9.428268811494039e-47 ], 
                    [ 601.0 , 9.429625481966112e-47 ], [ 601.0625 , 9.430982347655008e-47 ], [ 601.125 , 9.432339408588547e-47 ], [ 601.1875 , 9.433696664794892e-47 ], 
                    [ 601.25 , 9.435054116302072e-47 ], [ 601.3125 , 9.436411763138462e-47 ], [ 601.375 , 9.437769605331761e-47 ], [ 601.4375 , 9.439127642910352e-47 ], 
                    [ 601.5 , 9.440485112720658e-47 ], [ 601.5625 , 9.441842435722e-47 ], [ 601.625 , 9.44319995387494e-47 ], [ 601.6875 , 9.444557667207465e-47 ], 
                    [ 601.75 , 9.445915575747706e-47 ], [ 601.8125 , 9.447273679523798e-47 ], [ 601.875 , 9.448631978563741e-47 ], [ 601.9375 , 9.449990472895542e-47 ], 
                    [ 602.0 , 9.451349162547486e-47 ], [ 602.0625 , 9.45270804754738e-47 ], [ 602.125 , 9.45406712792352e-47 ], [ 602.1875 , 9.45542640370399e-47 ], 
                    [ 602.25 , 9.456785874916617e-47 ], [ 602.3125 , 9.458145541589838e-47 ], [ 602.375 , 9.45950540375162e-47 ], [ 602.4375 , 9.460865461429933e-47 ], 
                    [ 602.5 , 9.46222571465323e-47 ], [ 602.5625 , 9.46358616344935e-47 ], [ 602.625 , 9.464946807846485e-47 ], [ 602.6875 , 9.466307647872686e-47 ], 
                    [ 602.75 , 9.46766868355615e-47 ], [ 602.8125 , 9.46902991492514e-47 ], [ 602.875 , 9.47039039643322e-47 ], [ 602.9375 , 9.471750941904445e-47 ], 
                    [ 603.0 , 9.47311168283596e-47 ], [ 603.0625 , 9.474472619255643e-47 ], [ 603.125 , 9.475833751191646e-47 ], [ 603.1875 , 9.477195078672125e-47 ], 
                    [ 603.25 , 9.478556601725107e-47 ], [ 603.3125 , 9.479918320378551e-47 ], [ 603.375 , 9.48128023466083e-47 ], [ 603.4375 , 9.48264234459991e-47 ], 
                    [ 603.5 , 9.484004650223767e-47 ], [ 603.5625 , 9.48536715156065e-47 ], [ 603.625 , 9.486729848638943e-47 ], [ 603.6875 , 9.488092741486296e-47 ], 
                    [ 603.75 , 9.489455830131034e-47 ], [ 603.8125 , 9.490819114601425e-47 ], [ 603.875 , 9.49218259492526e-47 ], [ 603.9375 , 9.493546271130883e-47 ], 
                    [ 604.0 , 9.494910143246499e-47 ], [ 604.0625 , 9.496274211300252e-47 ], [ 604.125 , 9.497638475320228e-47 ], [ 604.1875 , 9.499002917636646e-47 ], 
                    [ 604.25 , 9.5003664562211e-47 ], [ 604.3125 , 9.501730190535371e-47 ], [ 604.375 , 9.50309412060749e-47 ], [ 604.4375 , 9.504458246465555e-47 ], 
                    [ 604.5 , 9.50582256813767e-47 ], [ 604.5625 , 9.507187085651945e-47 ], [ 604.625 , 9.508551799036423e-47 ], [ 604.6875 , 9.50991670831929e-47 ], 
                    [ 604.75 , 9.5112818135288e-47 ], [ 604.8125 , 9.512647114692874e-47 ], [ 604.875 , 9.514012611839642e-47 ], [ 604.9375 , 9.51537830499737e-47 ], 
                    [ 605.0 , 9.516744194194128e-47 ], [ 605.0625 , 9.518110279457921e-47 ], [ 605.125 , 9.519476560817165e-47 ], [ 605.1875 , 9.52084303829994e-47 ], 
                    [ 605.25 , 9.522209711934194e-47 ], [ 605.3125 , 9.523576581748223e-47 ], [ 605.375 , 9.52494364777025e-47 ], [ 605.4375 , 9.526310910028377e-47 ], 
                    [ 605.5 , 9.527678368550703e-47 ], [ 605.5625 , 9.529045903575356e-47 ], [ 605.625 , 9.530412606232594e-47 ], [ 605.6875 , 9.53177950490919e-47 ], 
                    [ 605.75 , 9.533146599632985e-47 ], [ 605.8125 , 9.5345138904323e-47 ], [ 605.875 , 9.53588137733519e-47 ], [ 605.9375 , 9.537249060369713e-47 ], 
                    [ 606.0 , 9.538616939564205e-47 ], [ 606.0625 , 9.539985014946525e-47 ], [ 606.125 , 9.541353286545018e-47 ], [ 606.1875 , 9.542721754387756e-47 ], 
                    [ 606.25 , 9.544090418502887e-47 ], [ 606.3125 , 9.54545927891856e-47 ], [ 606.375 , 9.546828335662864e-47 ], [ 606.4375 , 9.548197588764089e-47 ], 
                    [ 606.5 , 9.549567038250336e-47 ], [ 606.5625 , 9.550936684149763e-47 ], [ 606.625 , 9.552306526490548e-47 ], [ 606.6875 , 9.553676565300863e-47 ], 
                    [ 606.75 , 9.555046800608816e-47 ], [ 606.8125 , 9.556417232442797e-47 ], [ 606.875 , 9.557787860830787e-47 ], [ 606.9375 , 9.559158527006624e-47 ], 
                    [ 607.0 , 9.560528438265596e-47 ], [ 607.0625 , 9.561898545844932e-47 ], [ 607.125 , 9.563268849772702e-47 ], [ 607.1875 , 9.564639350077044e-47 ], 
                    [ 607.25 , 9.566010046785963e-47 ], [ 607.3125 , 9.567380939927948e-47 ], [ 607.375 , 9.568752029530806e-47 ], [ 607.4375 , 9.5701233156229e-47 ], 
                    [ 607.5 , 9.571494798232317e-47 ], [ 607.5625 , 9.572866477387086e-47 ], [ 607.625 , 9.57423835311578e-47 ], [ 607.6875 , 9.575610425445958e-47 ], 
                    [ 607.75 , 9.576982694406544e-47 ], [ 607.8125 , 9.5783551600249e-47 ], [ 607.875 , 9.57972782232989e-47 ], [ 607.9375 , 9.58110068134929e-47 ], 
                    [ 608.0 , 9.582473737111496e-47 ], [ 608.0625 , 9.58384698964457e-47 ], [ 608.125 , 9.585220438976774e-47 ], [ 608.1875 , 9.58659408513645e-47 ], 
                    [ 608.25 , 9.587967928151535e-47 ], [ 608.3125 , 9.589341818045083e-47 ], [ 608.375 , 9.590714918267099e-47 ], [ 608.4375 , 9.592088215103676e-47 ], 
                    [ 608.5 , 9.593461708582965e-47 ], [ 608.5625 , 9.594835398733123e-47 ], [ 608.625 , 9.596209285582384e-47 ], [ 608.6875 , 9.597583369158773e-47 ], 
                    [ 608.75 , 9.598957649490665e-47 ], [ 608.8125 , 9.600332126605892e-47 ], [ 608.875 , 9.601706800532903e-47 ], [ 608.9375 , 9.603081671299884e-47 ], 
                    [ 609.0 , 9.604456738934947e-47 ], [ 609.0625 , 9.60583200346615e-47 ], [ 609.125 , 9.607207464921888e-47 ], [ 609.1875 , 9.60858312333029e-47 ], 
                    [ 609.25 , 9.609958978719559e-47 ], [ 609.3125 , 9.611335031117901e-47 ], [ 609.375 , 9.612711280553529e-47 ], [ 609.4375 , 9.614087727054717e-47 ], 
                    [ 609.5 , 9.615464370649622e-47 ], [ 609.5625 , 9.616841211366394e-47 ], [ 609.625 , 9.618218249233398e-47 ], [ 609.6875 , 9.619595399985805e-47 ], 
                    [ 609.75 , 9.620971645500105e-47 ], [ 609.8125 , 9.622348087909692e-47 ], [ 609.875 , 9.623724727242537e-47 ], [ 609.9375 , 9.625101563526944e-47 ], 
                    [ 610.0 , 9.626478596791026e-47 ], [ 610.0625 , 9.627855827062895e-47 ], [ 610.125 , 9.62923325437087e-47 ], [ 610.1875 , 9.630610878743077e-47 ], 
                    [ 610.25 , 9.631988700207706e-47 ], [ 610.3125 , 9.633366718792749e-47 ], [ 610.375 , 9.634744934526888e-47 ], [ 610.4375 , 9.636123347437917e-47 ], 
                    [ 610.5 , 9.637501957554181e-47 ], [ 610.5625 , 9.638880764904033e-47 ], [ 610.625 , 9.640259769515417e-47 ], [ 610.6875 , 9.641638971416687e-47 ], 
                    [ 610.75 , 9.643018370636212e-47 ], [ 610.8125 , 9.644397967201875e-47 ], [ 610.875 , 9.645777761142185e-47 ], [ 610.9375 , 9.64715775248531e-47 ], 
                    [ 611.0 , 9.64853794125963e-47 ], [ 611.0625 , 9.64991832749305e-47 ], [ 611.125 , 9.651297813285864e-47 ], [ 611.1875 , 9.65267745052688e-47 ], 
                    [ 611.25 , 9.65405728498465e-47 ], [ 611.3125 , 9.655437316687564e-47 ], [ 611.375 , 9.656817545663894e-47 ], [ 611.4375 , 9.658197971941561e-47 ], 
                    [ 611.5 , 9.659578595548974e-47 ], [ 611.5625 , 9.660959416514209e-47 ], [ 611.625 , 9.66234043486547e-47 ], [ 611.6875 , 9.663721650631187e-47 ], 
                    [ 611.75 , 9.665103063839438e-47 ], [ 611.8125 , 9.66648467451831e-47 ], [ 611.875 , 9.667866482696445e-47 ], [ 611.9375 , 9.669248488401524e-47 ], 
                    [ 612.0 , 9.670630691662062e-47 ], [ 612.0625 , 9.672013092506568e-47 ], [ 612.125 , 9.67339569096274e-47 ], [ 612.1875 , 9.674778487059238e-47 ], 
                    [ 612.25 , 9.676161480824107e-47 ], [ 612.3125 , 9.677544672285675e-47 ], [ 612.375 , 9.678928061472268e-47 ], [ 612.4375 , 9.680311648412017e-47 ], 
                    [ 612.5 , 9.681694517601901e-47 ], [ 612.5625 , 9.6830773540493e-47 ], [ 612.625 , 9.684460388007368e-47 ], [ 612.6875 , 9.685843619503973e-47 ], 
                    [ 612.75 , 9.687227048567604e-47 ], [ 612.8125 , 9.68861067522648e-47 ], [ 612.875 , 9.689994499508617e-47 ], [ 612.9375 , 9.691378521442448e-47 ], 
                    [ 613.0 , 9.692762741056134e-47 ], [ 613.0625 , 9.69414715837798e-47 ], [ 613.125 , 9.695531773436155e-47 ], [ 613.1875 , 9.696916586258832e-47 ], 
                    [ 613.25 , 9.698301596874328e-47 ], [ 613.3125 , 9.69968680531089e-47 ], [ 613.375 , 9.70107221159685e-47 ], [ 613.4375 , 9.702457815760318e-47 ], 
                    [ 613.5 , 9.703843617829566e-47 ], [ 613.5625 , 9.705229617832925e-47 ], [ 613.625 , 9.706615815798807e-47 ], [ 613.6875 , 9.708002211755416e-47 ], 
                    [ 613.75 , 9.709388805730683e-47 ], [ 613.8125 , 9.710775597753312e-47 ], [ 613.875 , 9.71216187376671e-47 ], [ 613.9375 , 9.713547846601922e-47 ], 
                    [ 614.0 , 9.714934017222282e-47 ], [ 614.0625 , 9.716320385656017e-47 ], [ 614.125 , 9.717706951931007e-47 ], [ 614.1875 , 9.719093716076039e-47 ], 
                    [ 614.25 , 9.720480678119072e-47 ], [ 614.3125 , 9.721867838088212e-47 ], [ 614.375 , 9.723255196011908e-47 ], [ 614.4375 , 9.724642751918205e-47 ], 
                    [ 614.5 , 9.7260305058357e-47 ], [ 614.5625 , 9.72741845779237e-47 ], [ 614.625 , 9.728806607816552e-47 ], [ 614.6875 , 9.73019495593651e-47 ], 
                    [ 614.75 , 9.731583502180508e-47 ], [ 614.8125 , 9.732972246576687e-47 ], [ 614.875 , 9.734361189153595e-47 ], [ 614.9375 , 9.73575032993938e-47 ], 
                    [ 615.0 , 9.737139668962327e-47 ], [ 615.0625 , 9.738529206250513e-47 ], [ 615.125 , 9.739918941832722e-47 ], [ 615.1875 , 9.741308875736904e-47 ], 
                    [ 615.25 , 9.742698627109553e-47 ], [ 615.3125 , 9.744087789458846e-47 ], [ 615.375 , 9.745477149881945e-47 ], [ 615.4375 , 9.746866708406883e-47 ], 
                    [ 615.5 , 9.748256465061974e-47 ], [ 615.5625 , 9.749646419875471e-47 ], [ 615.625 , 9.751036572875559e-47 ], [ 615.6875 , 9.752426924090633e-47 ], 
                    [ 615.75 , 9.753817473548956e-47 ], [ 615.8125 , 9.755208221278658e-47 ], [ 615.875 , 9.756599167308147e-47 ], [ 615.9375 , 9.757990311665419e-47 ], 
                    [ 616.0 , 9.75938165437917e-47 ], [ 616.0625 , 9.760773195477402e-47 ], [ 616.125 , 9.762164934988478e-47 ], [ 616.1875 , 9.763556872940682e-47 ], 
                    [ 616.25 , 9.76494900936231e-47 ], [ 616.3125 , 9.766341344281663e-47 ], [ 616.375 , 9.767733877727183e-47 ], [ 616.4375 , 9.769126609726757e-47 ], 
                    [ 616.5 , 9.770519540309253e-47 ], [ 616.5625 , 9.7719126695025e-47 ], [ 616.625 , 9.773305979974009e-47 ], [ 616.6875 , 9.774698259093298e-47 ], 
                    [ 616.75 , 9.776090736552957e-47 ], [ 616.8125 , 9.777483412381316e-47 ], [ 616.875 , 9.778876286606418e-47 ], [ 616.9375 , 9.780269359256742e-47 ], 
                    [ 617.0 , 9.781662630360481e-47 ], [ 617.0625 , 9.783056099945978e-47 ], [ 617.125 , 9.784449768041369e-47 ], [ 617.1875 , 9.785843634674933e-47 ], 
                    [ 617.25 , 9.78723769987509e-47 ], [ 617.3125 , 9.78863196367006e-47 ], [ 617.375 , 9.790026426088065e-47 ], [ 617.4375 , 9.791421087157747e-47 ], 
                    [ 617.5 , 9.792815946906779e-47 ], [ 617.5625 , 9.794211005364022e-47 ], [ 617.625 , 9.795606262557505e-47 ], [ 617.6875 , 9.79700171851554e-47 ], 
                    [ 617.75 , 9.79839737326665e-47 ], [ 617.8125 , 9.799793226838876e-47 ], [ 617.875 , 9.80118927926082e-47 ], [ 617.9375 , 9.802585530560533e-47 ], 
                    [ 618.0 , 9.803981980766552e-47 ], [ 618.0625 , 9.805377883249033e-47 ], [ 618.125 , 9.806773584494955e-47 ], [ 618.1875 , 9.808169484405526e-47 ], 
                    [ 618.25 , 9.809565583008955e-47 ], [ 618.3125 , 9.810961880333804e-47 ], [ 618.375 , 9.812358376407942e-47 ], [ 618.4375 , 9.813755071259935e-47 ], 
                    [ 618.5 , 9.81515196491801e-47 ], [ 618.5625 , 9.81654905741039e-47 ], [ 618.625 , 9.817946348765594e-47 ], [ 618.6875 , 9.819343839011645e-47 ], 
                    [ 618.75 , 9.82074152817699e-47 ], [ 618.8125 , 9.82213941629009e-47 ], [ 618.875 , 9.823537503379046e-47 ], [ 618.9375 , 9.82493578947226e-47 ], 
                    [ 619.0 , 9.826334274598187e-47 ], [ 619.0625 , 9.827732958784745e-47 ], [ 619.125 , 9.829131842060686e-47 ], [ 619.1875 , 9.83053092445414e-47 ], 
                    [ 619.25 , 9.831930205993654e-47 ], [ 619.3125 , 9.833329686707443e-47 ], [ 619.375 , 9.834729366623642e-47 ], [ 619.4375 , 9.83612895056524e-47 ], 
                    [ 619.5 , 9.83752778797572e-47 ], [ 619.5625 , 9.838926824320909e-47 ], [ 619.625 , 9.840326059628754e-47 ], [ 619.6875 , 9.841725493927897e-47 ], 
                    [ 619.75 , 9.843125127246498e-47 ], [ 619.8125 , 9.84452495961279e-47 ], [ 619.875 , 9.84592499105515e-47 ], [ 619.9375 , 9.84732522160196e-47 ], 
                    [ 620.0 , 9.848725651281327e-47 ], [ 620.0625 , 9.850126280121847e-47 ], [ 620.125 , 9.851527108151708e-47 ], [ 620.1875 , 9.852928135399232e-47 ], 
                    [ 620.25 , 9.854329361892757e-47 ], [ 620.3125 , 9.855730787660474e-47 ], [ 620.375 , 9.857132412731004e-47 ], [ 620.4375 , 9.85853423713255e-47 ], 
                    [ 620.5 , 9.859936260893394e-47 ], [ 620.5625 , 9.86133848404195e-47 ], [ 620.625 , 9.86274090660658e-47 ], [ 620.6875 , 9.864143528615641e-47 ], 
                    [ 620.75 , 9.865546350097638e-47 ], [ 620.8125 , 9.866949371080657e-47 ], [ 620.875 , 9.86835161125906e-47 ], [ 620.9375 , 9.869753764035705e-47 ], 
                    [ 621.0 , 9.871156116038228e-47 ], [ 621.0625 , 9.872558667295077e-47 ], [ 621.125 , 9.873961417834775e-47 ], [ 621.1875 , 9.875364367685215e-47 ], 
                    [ 621.25 , 9.876767516875069e-47 ], [ 621.3125 , 9.878170865432377e-47 ], [ 621.375 , 9.879574413385677e-47 ], [ 621.4375 , 9.880978160763162e-47 ], 
                    [ 621.5 , 9.882382107593308e-47 ], [ 621.5625 , 9.88378625390431e-47 ], [ 621.625 , 9.885190599724726e-47 ], [ 621.6875 , 9.886595145082692e-47 ], 
                    [ 621.75 , 9.887999890006698e-47 ], [ 621.8125 , 9.889404834524893e-47 ], [ 621.875 , 9.890809978665842e-47 ], [ 621.9375 , 9.89221532245777e-47 ], 
                    [ 622.0 , 9.893620865929258e-47 ], [ 622.0625 , 9.895026609108465e-47 ], [ 622.125 , 9.896432552023834e-47 ], [ 622.1875 , 9.89783869470375e-47 ], 
                    [ 622.25 , 9.899244731978635e-47 ], [ 622.3125 , 9.900650093686402e-47 ], [ 622.375 , 9.902055654908533e-47 ], [ 622.4375 , 9.903461415673355e-47 ], 
                    [ 622.5 , 9.904867376009126e-47 ], [ 622.5625 , 9.906273535944243e-47 ], [ 622.625 , 9.907679895507192e-47 ], [ 622.6875 , 9.909086454726025e-47 ], 
                    [ 622.75 , 9.910493213629372e-47 ], [ 622.8125 , 9.911900172245299e-47 ], [ 622.875 , 9.913307330602298e-47 ], [ 622.9375 , 9.914714688728796e-47 ], 
                    [ 623.0 , 9.916122246653016e-47 ], [ 623.0625 , 9.91753000440353e-47 ], [ 623.125 , 9.918937962008354e-47 ], [ 623.1875 , 9.920346119496288e-47 ], 
                    [ 623.25 , 9.921754476895423e-47 ], [ 623.3125 , 9.923163034234141e-47 ], [ 623.375 , 9.924571791540827e-47 ], [ 623.4375 , 9.92598074884415e-47 ], 
                    [ 623.5 , 9.927389906172224e-47 ], [ 623.5625 , 9.928799263553443e-47 ], [ 623.625 , 9.93020882101642e-47 ], [ 623.6875 , 9.931617689049726e-47 ], 
                    [ 623.75 , 9.933026361881824e-47 ], [ 623.8125 , 9.934435234516202e-47 ], [ 623.875 , 9.93584430698106e-47 ], [ 623.9375 , 9.937253579304806e-47 ], 
                    [ 624.0 , 9.93866305151572e-47 ], [ 624.0625 , 9.940072723642439e-47 ], [ 624.125 , 9.941482595712889e-47 ], [ 624.1875 , 9.942892667755856e-47 ], 
                    [ 624.25 , 9.944302939799281e-47 ], [ 624.3125 , 9.945713411871813e-47 ], [ 624.375 , 9.947124084001752e-47 ], [ 624.4375 , 9.948534956217543e-47 ], 
                    [ 624.5 , 9.94994602854757e-47 ], [ 624.5625 , 9.95135730101986e-47 ], [ 624.625 , 9.952768773663227e-47 ], [ 624.6875 , 9.954180446505916e-47 ], 
                    [ 624.75 , 9.95559231957633e-47 ], [ 624.8125 , 9.957004392902866e-47 ], [ 624.875 , 9.958416666514066e-47 ], [ 624.9375 , 9.959829140437989e-47 ], 
                    [ 625.0 , 9.961241814703396e-47 ], [ 625.0625 , 9.962654597742658e-47 ], [ 625.125 , 9.964066454559597e-47 ], [ 625.1875 , 9.965478511457714e-47 ], 
                    [ 625.25 , 9.966890768465367e-47 ], [ 625.3125 , 9.968303225610909e-47 ], [ 625.375 , 9.969715882922704e-47 ], [ 625.4375 , 9.971128740429121e-47 ], 
                    [ 625.5 , 9.972541798158457e-47 ], [ 625.5625 , 9.973955056139231e-47 ], [ 625.625 , 9.975368514399818e-47 ], [ 625.6875 , 9.976782172968462e-47 ], 
                    [ 625.75 , 9.978196031873618e-47 ], [ 625.8125 , 9.97961009114361e-47 ], [ 625.875 , 9.981024350807111e-47 ], [ 625.9375 , 9.98243881089217e-47 ], 
                    [ 626.0 , 9.983853471427327e-47 ], [ 626.0625 , 9.985268332440922e-47 ], [ 626.125 , 9.986683393961576e-47 ], [ 626.1875 , 9.988098656017421e-47 ], 
                    [ 626.25 , 9.98951411863716e-47 ], [ 626.3125 , 9.99092978184893e-47 ], [ 626.375 , 9.992345645681515e-47 ], [ 626.4375 , 9.993761710162848e-47 ], 
                    [ 626.5 , 9.99517743762078e-47 ], [ 626.5625 , 9.996592581180184e-47 ], [ 626.625 , 9.99800792509927e-47 ], [ 626.6875 , 9.999423469406476e-47 ], 
                    [ 626.75 , 1.000083921413025e-46 ], [ 626.8125 , 1.0002255159298748e-46 ], [ 626.875 , 1.0003671304940566e-46 ], [ 626.9375 , 1.0005087651084015e-46 ], 
                    [ 627.0 , 1.0006504197757412e-46 ], [ 627.0625 , 1.0007920944989362e-46 ], [ 627.125 , 1.0009337892808047e-46 ], [ 627.1875 , 1.001075504124201e-46 ], 
                    [ 627.25 , 1.001217239031951e-46 ], [ 627.3125 , 1.0013589940068957e-46 ], [ 627.375 , 1.0015007690518832e-46 ], [ 627.4375 , 1.0016425641697624e-46 ], 
                    [ 627.5 , 1.001784379363368e-46 ], [ 627.5625 , 1.0019262146355494e-46 ], [ 627.625 , 1.0020680699891282e-46 ], [ 627.6875 , 1.002209945426976e-46 ], 
                    [ 627.75 , 1.0023518409519078e-46 ], [ 627.8125 , 1.0024937565667963e-46 ], [ 627.875 , 1.0026356922744644e-46 ], [ 627.9375 , 1.0027775632150654e-46 ], 
                    [ 628.0 , 1.0029194155104891e-46 ], [ 628.0625 , 1.0030612878722584e-46 ], [ 628.125 , 1.0032031803031973e-46 ], [ 628.1875 , 1.0033450928061589e-46 ], 
                    [ 628.25 , 1.0034870253839686e-46 ], [ 628.3125 , 1.0036289780394875e-46 ], [ 628.375 , 1.003770950775527e-46 ], [ 628.4375 , 1.0039129435949424e-46 ], 
                    [ 628.5 , 1.0040549565005743e-46 ], [ 628.5625 , 1.0041969894952714e-46 ], [ 628.625 , 1.0043390425818612e-46 ], [ 628.6875 , 1.0044811157632e-46 ], 
                    [ 628.75 , 1.0046232090421235e-46 ], [ 628.8125 , 1.0047653224214602e-46 ], [ 628.875 , 1.0049074559040894e-46 ], [ 628.9375 , 1.005049609492819e-46 ], 
                    [ 629.0 , 1.0051917831905218e-46 ], [ 629.0625 , 1.0053339770000069e-46 ], [ 629.125 , 1.0054761909241475e-46 ], [ 629.1875 , 1.0056184249657895e-46 ], 
                    [ 629.25 , 1.0057606791277781e-46 ], [ 629.3125 , 1.0059029534129457e-46 ], [ 629.375 , 1.0060451282787466e-46 ], [ 629.4375 , 1.0061873121841376e-46 ], 
                    [ 629.5 , 1.0063295161843232e-46 ], [ 629.5625 , 1.006471740282129e-46 ], [ 629.625 , 1.0066139844804098e-46 ], [ 629.6875 , 1.006756248781992e-46 ], 
                    [ 629.75 , 1.006898533189731e-46 ], [ 629.8125 , 1.0070408377064614e-46 ], [ 629.875 , 1.0071831623350107e-46 ], [ 629.9375 , 1.007325507078243e-46 ], 
                    [ 630.0 , 1.0074678719390081e-46 ], [ 630.0625 , 1.0076102569201135e-46 ], [ 630.125 , 1.0077526620244314e-46 ], [ 630.1875 , 1.00789508725482e-46 ], 
                    [ 630.25 , 1.0080375326140738e-46 ], [ 630.3125 , 1.0081799981050731e-46 ], [ 630.375 , 1.0083224837306563e-46 ], [ 630.4375 , 1.008464989493676e-46 ], 
                    [ 630.5 , 1.008607515396971e-46 ], [ 630.5625 , 1.0087500614433881e-46 ], [ 630.625 , 1.0088926276357664e-46 ], [ 630.6875 , 1.0090352139769607e-46 ], 
                    [ 630.75 , 1.009177808667304e-46 ], [ 630.8125 , 1.0093203074514615e-46 ], [ 630.875 , 1.0094628263568536e-46 ], [ 630.9375 , 1.0096053653863145e-46 ], 
                    [ 631.0 , 1.0097479245426927e-46 ], [ 631.0625 , 1.0098905038288303e-46 ], [ 631.125 , 1.0100331032475697e-46 ], [ 631.1875 , 1.0101757228017534e-46 ], 
                    [ 631.25 , 1.010318362494225e-46 ], [ 631.3125 , 1.0104610223278353e-46 ], [ 631.375 , 1.0106037023054133e-46 ], [ 631.4375 , 1.010746402429811e-46 ], 
                    [ 631.5 , 1.0108891227038662e-46 ], [ 631.5625 , 1.0110318631304597e-46 ], [ 631.625 , 1.0111746237123797e-46 ], [ 631.6875 , 1.0113174044525154e-46 ], 
                    [ 631.75 , 1.0114602053536916e-46 ], [ 631.8125 , 1.011603026418762e-46 ], [ 631.875 , 1.0117458676505813e-46 ], [ 631.9375 , 1.0118887290519827e-46 ], 
                    [ 632.0 , 1.0120316106258212e-46 ], [ 632.0625 , 1.0121745123749454e-46 ], [ 632.125 , 1.012317434302204e-46 ], [ 632.1875 , 1.0124603476087444e-46 ], 
                    [ 632.25 , 1.0126031775172807e-46 ], [ 632.3125 , 1.0127460275751326e-46 ], [ 632.375 , 1.0128888977851565e-46 ], [ 632.4375 , 1.013031788150174e-46 ], 
                    [ 632.5 , 1.0131746986730356e-46 ], [ 632.5625 , 1.0133176293565777e-46 ], [ 632.625 , 1.0134605802036588e-46 ], [ 632.6875 , 1.0136035512171162e-46 ], 
                    [ 632.75 , 1.013746542399795e-46 ], [ 632.8125 , 1.013889553754533e-46 ], [ 632.875 , 1.0140325852841908e-46 ], [ 632.9375 , 1.0141756369915995e-46 ], 
                    [ 633.0 , 1.0143187088796277e-46 ], [ 633.0625 , 1.0144618009511006e-46 ], [ 633.125 , 1.0146049132088797e-46 ], [ 633.1875 , 1.0147480456557842e-46 ], 
                    [ 633.25 , 1.0148911982946981e-46 ], [ 633.3125 , 1.0150343711284557e-46 ], [ 633.375 , 1.0151775641599128e-46 ], [ 633.4375 , 1.0153207773919045e-46 ], 
                    [ 633.5 , 1.0154640108272807e-46 ], [ 633.5625 , 1.0156072644688984e-46 ], [ 633.625 , 1.0157505001577636e-46 ], [ 633.6875 , 1.0158936630741806e-46 ], 
                    [ 633.75 , 1.016036846168415e-46 ], [ 633.8125 , 1.0161800494432886e-46 ], [ 633.875 , 1.0163232729016752e-46 ], [ 633.9375 , 1.0164665165463975e-46 ], 
                    [ 634.0 , 1.016609780380308e-46 ], [ 634.0625 , 1.0167530644062594e-46 ], [ 634.125 , 1.0168963686270902e-46 ], [ 634.1875 , 1.0170396930456397e-46 ], 
                    [ 634.25 , 1.0171830376647692e-46 ], [ 634.3125 , 1.0173264024873039e-46 ], [ 634.375 , 1.0174697875161205e-46 ], [ 634.4375 , 1.0176131927540521e-46 ], 
                    [ 634.5 , 1.0177566182039401e-46 ], [ 634.5625 , 1.0179000638686404e-46 ], [ 634.625 , 1.0180435297510094e-46 ], [ 634.6875 , 1.0181870158538822e-46 ], 
                    [ 634.75 , 1.018330522180123e-46 ], [ 634.8125 , 1.0184740487325678e-46 ], [ 634.875 , 1.018617595514082e-46 ], [ 634.9375 , 1.018761162527502e-46 ], 
                    [ 635.0 , 1.0189047497756865e-46 ], [ 635.0625 , 1.019048314879551e-46 ], [ 635.125 , 1.019191806949832e-46 ], [ 635.1875 , 1.0193353192252283e-46 ], 
                    [ 635.25 , 1.0194788517085638e-46 ], [ 635.3125 , 1.0196224044026909e-46 ], [ 635.375 , 1.0197659773104553e-46 ], [ 635.4375 , 1.0199095704347111e-46 ], 
                    [ 635.5 , 1.0200531837782827e-46 ], [ 635.5625 , 1.0201968173440465e-46 ], [ 635.625 , 1.0203404711348208e-46 ], [ 635.6875 , 1.0204841451534899e-46 ], 
                    [ 635.75 , 1.0206278394028584e-46 ], [ 635.8125 , 1.0207715538858114e-46 ], [ 635.875 , 1.020915288605176e-46 ], [ 635.9375 , 1.021059043563802e-46 ], 
                    [ 636.0 , 1.0212028187645607e-46 ], [ 636.0625 , 1.021346614210281e-46 ], [ 636.125 , 1.0214904299038134e-46 ], [ 636.1875 , 1.021634265848038e-46 ], 
                    [ 636.25 , 1.0217781220457554e-46 ], [ 636.3125 , 1.0219219984998685e-46 ], [ 636.375 , 1.0220658952132006e-46 ], [ 636.4375 , 1.022209812188626e-46 ], 
                    [ 636.5 , 1.0223537127902417e-46 ], [ 636.5625 , 1.0224975402397143e-46 ], [ 636.625 , 1.022641387923224e-46 ], [ 636.6875 , 1.022785255843617e-46 ], 
                    [ 636.75 , 1.0229291440037111e-46 ], [ 636.8125 , 1.0230730524063905e-46 ], [ 636.875 , 1.0232169810544882e-46 ], [ 636.9375 , 1.0233609299508447e-46 ], 
                    [ 637.0 , 1.0235048990983164e-46 ], [ 637.0625 , 1.0236488884997594e-46 ], [ 637.125 , 1.0237928981580229e-46 ], [ 637.1875 , 1.023936928075928e-46 ], 
                    [ 637.25 , 1.024080978256368e-46 ], [ 637.3125 , 1.0242250487021501e-46 ], [ 637.375 , 1.0243691394161691e-46 ], [ 637.4375 , 1.0245132504012471e-46 ], 
                    [ 637.5 , 1.0246573816602213e-46 ], [ 637.5625 , 1.024801533196002e-46 ], [ 637.625 , 1.0249457050113768e-46 ], [ 637.6875 , 1.025089897109249e-46 ], 
                    [ 637.75 , 1.0252341094924287e-46 ], [ 637.8125 , 1.0253783421638206e-46 ], [ 637.875 , 1.0255225951262207e-46 ], [ 637.9375 , 1.0256668417200767e-46 ], 
                    [ 638.0 , 1.0258110011875392e-46 ], [ 638.0625 , 1.0259551809168876e-46 ], [ 638.125 , 1.0260993809109847e-46 ], [ 638.1875 , 1.0262436011726635e-46 ], 
                    [ 638.25 , 1.0263878417047878e-46 ], [ 638.3125 , 1.0265321025101918e-46 ], [ 638.375 , 1.0266763835917397e-46 ], [ 638.4375 , 1.0268206849522737e-46 ], 
                    [ 638.5 , 1.026965006594637e-46 ], [ 638.5625 , 1.0271093485216878e-46 ], [ 638.625 , 1.0272537107362842e-46 ], [ 638.6875 , 1.0273980932412702e-46 ], 
                    [ 638.75 , 1.0275424960394906e-46 ], [ 638.8125 , 1.0276869191338048e-46 ], [ 638.875 , 1.0278313625270727e-46 ], [ 638.9375 , 1.0279758262221402e-46 ], 
                    [ 639.0 , 1.0281203102218537e-46 ], [ 639.0625 , 1.0282648145290811e-46 ], [ 639.125 , 1.0284093391466699e-46 ], [ 639.1875 , 1.0285538840774599e-46 ], 
                    [ 639.25 , 1.0286984493243351e-46 ], [ 639.3125 , 1.0288430348901295e-46 ], [ 639.375 , 1.028987632104664e-46 ], [ 639.4375 , 1.0291321220989551e-46 ], 
                    [ 639.5 , 1.029276632382483e-46 ], [ 639.5625 , 1.0294211629580598e-46 ], [ 639.625 , 1.0295657138285789e-46 ], [ 639.6875 , 1.0297102849968609e-46 ], 
                    [ 639.75 , 1.0298548764657706e-46 ], [ 639.8125 , 1.0299994882381439e-46 ], [ 639.875 , 1.0301441203168465e-46 ], [ 639.9375 , 1.0302887727047078e-46 ], 
                    [ 640.0 , 1.030433445404609e-46 ], [ 640.0625 , 1.0305781384193801e-46 ], [ 640.125 , 1.0307228517518814e-46 ], [ 640.1875 , 1.030867585404973e-46 ], 
                    [ 640.25 , 1.0310123393814863e-46 ], [ 640.3125 , 1.0311571136842973e-46 ], [ 640.375 , 1.0313019083162526e-46 ], [ 640.4375 , 1.031446723280207e-46 ], 
                    [ 640.5 , 1.0315915585790226e-46 ], [ 640.5625 , 1.0317364142155404e-46 ], [ 640.625 , 1.0318812901926236e-46 ], [ 640.6875 , 1.0320261865131283e-46 ], 
                    [ 640.75 , 1.0321711031799108e-46 ], [ 640.8125 , 1.032316040195836e-46 ], [ 640.875 , 1.032460875902655e-46 ], [ 640.9375 , 1.0326057147316894e-46 ], 
                    [ 641.0 , 1.0327505738794468e-46 ], [ 641.0625 , 1.0328954533487846e-46 ], [ 641.125 , 1.0330403531425317e-46 ], [ 641.1875 , 1.033185273263569e-46 ], 
                    [ 641.25 , 1.0333302137147184e-46 ], [ 641.3125 , 1.033475174498854e-46 ], [ 641.375 , 1.0336201556188357e-46 ], [ 641.4375 , 1.0337651570774721e-46 ], 
                    [ 641.5 , 1.0339101788776755e-46 ], [ 641.5625 , 1.0340552210222476e-46 ], [ 641.625 , 1.034200283514072e-46 ], [ 641.6875 , 1.0343453663559959e-46 ], 
                    [ 641.75 , 1.0344904695508666e-46 ], [ 641.8125 , 1.0346355931015615e-46 ], [ 641.875 , 1.0347807370109216e-46 ], [ 641.9375 , 1.0349259012817879e-46 ], 
                    [ 642.0 , 1.0350710859170464e-46 ], [ 642.0625 , 1.035216290919539e-46 ], [ 642.125 , 1.0353615162921086e-46 ], [ 642.1875 , 1.035506762037642e-46 ], 
                    [ 642.25 , 1.0356520281589754e-46 ], [ 642.3125 , 1.035797229361681e-46 ], [ 642.375 , 1.0359424020634518e-46 ], [ 642.4375 , 1.0360875951119784e-46 ], 
                    [ 642.5 , 1.0362328085101129e-46 ], [ 642.5625 , 1.0363780422607067e-46 ], [ 642.625 , 1.0365232963666202e-46 ], [ 642.6875 , 1.0366685708306916e-46 ], 
                    [ 642.75 , 1.0368138656557811e-46 ], [ 642.8125 , 1.0369591808447356e-46 ], [ 642.875 , 1.037104516400431e-46 ], [ 642.9375 , 1.0372498723256995e-46 ], 
                    [ 643.0 , 1.037395248623404e-46 ], [ 643.0625 , 1.0375406452963919e-46 ], [ 643.125 , 1.0376860623475339e-46 ], [ 643.1875 , 1.0378314997796711e-46 ], 
                    [ 643.25 , 1.0379769575956824e-46 ], [ 643.3125 , 1.0381224357984027e-46 ], [ 643.375 , 1.0382679343907037e-46 ], [ 643.4375 , 1.0384134533754213e-46 ], 
                    [ 643.5 , 1.0385589927554354e-46 ], [ 643.5625 , 1.038704552533605e-46 ], [ 643.625 , 1.0388501327127738e-46 ], [ 643.6875 , 1.0389957332958086e-46 ], 
                    [ 643.75 , 1.0391413060368314e-46 ], [ 643.8125 , 1.0392868067592786e-46 ], [ 643.875 , 1.0394323278547368e-46 ], [ 643.9375 , 1.0395778693260957e-46 ], 
                    [ 644.0 , 1.0397234311761858e-46 ], [ 644.0625 , 1.0398690134078755e-46 ], [ 644.125 , 1.040014616024004e-46 ], [ 644.1875 , 1.0401602390274399e-46 ], 
                    [ 644.25 , 1.0403058824210236e-46 ], [ 644.3125 , 1.040451546207617e-46 ], [ 644.375 , 1.0405972303900834e-46 ], [ 644.4375 , 1.0407429349712636e-46 ], 
                    [ 644.5 , 1.040888659954021e-46 ], [ 644.5625 , 1.0410344053412275e-46 ], [ 644.625 , 1.0411801711357177e-46 ], [ 644.6875 , 1.0413259573403415e-46 ], 
                    [ 644.75 , 1.0414717639579938e-46 ], [ 644.8125 , 1.0416175909914813e-46 ], [ 644.875 , 1.0417634384437216e-46 ], [ 644.9375 , 1.0419093063175218e-46 ], 
                    [ 645.0 , 1.0420551946157711e-46 ], [ 645.0625 , 1.0422011033413369e-46 ], [ 645.125 , 1.0423470324970571e-46 ], [ 645.1875 , 1.042492982085792e-46 ], 
                    [ 645.25 , 1.0426388140036918e-46 ], [ 645.3125 , 1.0427846659671602e-46 ], [ 645.375 , 1.0429305383334782e-46 ], [ 645.4375 , 1.0430764311054853e-46 ], 
                    [ 645.5 , 1.0432223442860431e-46 ], [ 645.5625 , 1.0433682778780066e-46 ], [ 645.625 , 1.0435142318842313e-46 ], [ 645.6875 , 1.0436602063075802e-46 ], 
                    [ 645.75 , 1.0438062011509017e-46 ], [ 645.8125 , 1.0439522164170378e-46 ], [ 645.875 , 1.044098252108875e-46 ], [ 645.9375 , 1.0442443082292558e-46 ], 
                    [ 646.0 , 1.0443903847810303e-46 ], [ 646.0625 , 1.044536481767079e-46 ], [ 646.125 , 1.0446825991902232e-46 ], [ 646.1875 , 1.0448287370533663e-46 ], 
                    [ 646.25 , 1.0449748953593158e-46 ], [ 646.3125 , 1.0451210741109757e-46 ], [ 646.375 , 1.0452672733112064e-46 ], [ 646.4375 , 1.0454134929628311e-46 ], 
                    [ 646.5 , 1.045559733068755e-46 ], [ 646.5625 , 1.0457059936317953e-46 ], [ 646.625 , 1.0458522746548581e-46 ], [ 646.6875 , 1.0459984922291775e-46 ], 
                    [ 646.75 , 1.0461446740353215e-46 ], [ 646.8125 , 1.046290876270842e-46 ], [ 646.875 , 1.0464370989386239e-46 ], [ 646.9375 , 1.0465833420415302e-46 ], 
                    [ 647.0 , 1.0467296055823719e-46 ], [ 647.0625 , 1.04687588956405e-46 ], [ 647.125 , 1.047022193989414e-46 ], [ 647.1875 , 1.0471685188613134e-46 ], 
                    [ 647.25 , 1.047314864182598e-46 ], [ 647.3125 , 1.0474612299561333e-46 ], [ 647.375 , 1.0476076161847851e-46 ], [ 647.4375 , 1.047754022871397e-46 ], 
                    [ 647.5 , 1.0479004500188282e-46 ], [ 647.5625 , 1.0480468976299605e-46 ], [ 647.625 , 1.0481933657076314e-46 ], [ 647.6875 , 1.0483398542547085e-46 ], 
                    [ 647.75 , 1.048486363274075e-46 ], [ 647.8125 , 1.04863289276854e-46 ], [ 647.875 , 1.048779442740994e-46 ], [ 647.9375 , 1.048926013194322e-46 ], 
                    [ 648.0 , 1.0490726041313339e-46 ], [ 648.0625 , 1.0492192155549518e-46 ], [ 648.125 , 1.0493658261825026e-46 ], [ 648.1875 , 1.0495123339622195e-46 ], 
                    [ 648.25 , 1.0496588621966995e-46 ], [ 648.3125 , 1.0498054108887985e-46 ], [ 648.375 , 1.0499519800413803e-46 ], [ 648.4375 , 1.050098569657279e-46 ], 
                    [ 648.5 , 1.050245179739374e-46 ], [ 648.5625 , 1.050391810290515e-46 ], [ 648.625 , 1.050538461313553e-46 ], [ 648.6875 , 1.0506851328113678e-46 ], 
                    [ 648.75 , 1.0508318247867887e-46 ], [ 648.8125 , 1.0509785372426893e-46 ], [ 648.875 , 1.051125270181944e-46 ], [ 648.9375 , 1.0512720236073754e-46 ], 
                    [ 649.0 , 1.0514187975218959e-46 ], [ 649.0625 , 1.0515655919283137e-46 ], [ 649.125 , 1.0517124068295124e-46 ], [ 649.1875 , 1.051859242228368e-46 ], 
                    [ 649.25 , 1.0520060981277201e-46 ], [ 649.3125 , 1.052152974530431e-46 ], [ 649.375 , 1.0522998714393782e-46 ], [ 649.4375 , 1.0524467888574168e-46 ], 
                    [ 649.5 , 1.0525937267873958e-46 ], [ 649.5625 , 1.0527406852322085e-46 ], [ 649.625 , 1.0528875730303583e-46 ], [ 649.6875 , 1.0530344305687533e-46 ], 
                    [ 649.75 , 1.0531813085909463e-46 ], [ 649.8125 , 1.0533282070998023e-46 ], [ 649.875 , 1.0534751260981559e-46 ], [ 649.9375 , 1.0536220655888803e-46 ], 
                    [ 650.0 , 1.0537690255748488e-46 ], [ 650.0625 , 1.0539160060588897e-46 ], [ 650.125 , 1.0540630070438775e-46 ], [ 650.1875 , 1.0542100285326866e-46 ], 
                    [ 650.25 , 1.0543570705281468e-46 ], [ 650.3125 , 1.0545041330331483e-46 ], [ 650.375 , 1.0546512160505221e-46 ], [ 650.4375 , 1.054798319583144e-46 ], 
                    [ 650.5 , 1.0549454436338755e-46 ], [ 650.5625 , 1.0550925882056012e-46 ], [ 650.625 , 1.055239753301138e-46 ], [ 650.6875 , 1.0553869389233789e-46 ], 
                    [ 650.75 , 1.055534145075172e-46 ], [ 650.8125 , 1.055681371759388e-46 ], [ 650.875 , 1.0558286189789135e-46 ], [ 650.9375 , 1.0559758867365753e-46 ], 
                    [ 651.0 , 1.0561231750352454e-46 ], [ 651.0625 , 1.0562704719882372e-46 ], [ 651.125 , 1.0564176606179593e-46 ], [ 651.1875 , 1.0565648697580313e-46 ], 
                    [ 651.25 , 1.0567120994113337e-46 ], [ 651.3125 , 1.0568593495807324e-46 ], [ 651.375 , 1.0570066202690639e-46 ], [ 651.4375 , 1.057153911479187e-46 ], 
                    [ 651.5 , 1.0573012232139696e-46 ], [ 651.5625 , 1.0574485554762863e-46 ], [ 651.625 , 1.05759590826896e-46 ], [ 651.6875 , 1.0577432815948893e-46 ], 
                    [ 651.75 , 1.0578906754569204e-46 ], [ 651.8125 , 1.058038089857915e-46 ], [ 651.875 , 1.0581855248007425e-46 ], [ 651.9375 , 1.0583329802882429e-46 ], 
                    [ 652.0 , 1.0584804563232865e-46 ], [ 652.0625 , 1.0586279529087592e-46 ], [ 652.125 , 1.0587754700475097e-46 ], [ 652.1875 , 1.0589230077424092e-46 ], 
                    [ 652.25 , 1.0590705659963e-46 ], [ 652.3125 , 1.0592181448120617e-46 ], [ 652.375 , 1.0593657441925596e-46 ], [ 652.4375 , 1.0595133641406746e-46 ], 
                    [ 652.5 , 1.05966100465925e-46 ], [ 652.5625 , 1.0598085968226238e-46 ], [ 652.625 , 1.0599561327222365e-46 ], [ 652.6875 , 1.0601036891603067e-46 ], 
                    [ 652.75 , 1.060251266139701e-46 ], [ 652.8125 , 1.0603988636632713e-46 ], [ 652.875 , 1.0605464817338926e-46 ], [ 652.9375 , 1.0606941203544255e-46 ], 
                    [ 653.0 , 1.060841779527708e-46 ], [ 653.0625 , 1.0609894592566239e-46 ], [ 653.125 , 1.0611371595440345e-46 ], [ 653.1875 , 1.0612848803927872e-46 ], 
                    [ 653.25 , 1.0614326218057514e-46 ], [ 653.3125 , 1.06158038378579e-46 ], [ 653.375 , 1.0617281663357586e-46 ], [ 653.4375 , 1.061875969458551e-46 ], 
                    [ 653.5 , 1.0620237931569932e-46 ], [ 653.5625 , 1.062171637433965e-46 ], [ 653.625 , 1.0623195022923305e-46 ], [ 653.6875 , 1.062467387734948e-46 ], 
                    [ 653.75 , 1.0626152937647052e-46 ], [ 653.8125 , 1.0627632203844383e-46 ], [ 653.875 , 1.0629111675970358e-46 ], [ 653.9375 , 1.0630591354053646e-46 ], 
                    [ 654.0 , 1.0632071238122692e-46 ], [ 654.0625 , 1.063355014518554e-46 ], [ 654.125 , 1.0635028976809353e-46 ], [ 654.1875 , 1.0636508014097669e-46 ], 
                    [ 654.25 , 1.0637987257078857e-46 ], [ 654.3125 , 1.0639466705781908e-46 ], [ 654.375 , 1.0640946360234974e-46 ], [ 654.4375 , 1.0642426220467048e-46 ], 
                    [ 654.5 , 1.0643906286506446e-46 ], [ 654.5625 , 1.0645386558382094e-46 ], [ 654.625 , 1.0646867036122465e-46 ], [ 654.6875 , 1.064834771975619e-46 ], 
                    [ 654.75 , 1.0649828609311828e-46 ], [ 654.8125 , 1.0651309704817938e-46 ], [ 654.875 , 1.0652791006303696e-46 ], [ 654.9375 , 1.065427251379714e-46 ], 
                    [ 655.0 , 1.0655754227327147e-46 ], [ 655.0625 , 1.0657236146922448e-46 ], [ 655.125 , 1.0658718272611625e-46 ], [ 655.1875 , 1.0660200604423112e-46 ], 
                    [ 655.25 , 1.0661683142386106e-46 ], [ 655.3125 , 1.0663165886528745e-46 ], [ 655.375 , 1.0664648836880084e-46 ], [ 655.4375 , 1.0666131993468497e-46 ], 
                    [ 655.5 , 1.0667615226881114e-46 ], [ 655.5625 , 1.0669097348561649e-46 ], [ 655.625 , 1.0670579676163058e-46 ], [ 655.6875 , 1.0672062209714028e-46 ], 
                    [ 655.75 , 1.0673544949243098e-46 ], [ 655.8125 , 1.0675027894778962e-46 ], [ 655.875 , 1.0676511046350016e-46 ], [ 655.9375 , 1.0677994403985112e-46 ], 
                    [ 656.0 , 1.0679477967712727e-46 ], [ 656.0625 , 1.0680961737561722e-46 ], [ 656.125 , 1.068244571356051e-46 ], [ 656.1875 , 1.0683929895737729e-46 ], 
                    [ 656.25 , 1.0685414284122256e-46 ], [ 656.3125 , 1.068689887874236e-46 ], [ 656.375 , 1.0688383679627148e-46 ], [ 656.4375 , 1.0689868686804976e-46 ], 
                    [ 656.5 , 1.0691353900304508e-46 ], [ 656.5625 , 1.0692839320154558e-46 ], [ 656.625 , 1.0694324946383643e-46 ], [ 656.6875 , 1.0695810779020666e-46 ], 
                    [ 656.75 , 1.0697296818093927e-46 ], [ 656.8125 , 1.069878306363256e-46 ], [ 656.875 , 1.070026951566495e-46 ], [ 656.9375 , 1.0701756174219857e-46 ], 
                    [ 657.0 , 1.0703242564485792e-46 ], [ 657.0625 , 1.0704728138546274e-46 ], [ 657.125 , 1.0706213918799365e-46 ], [ 657.1875 , 1.0707699905273832e-46 ], 
                    [ 657.25 , 1.0709186097998145e-46 ], [ 657.3125 , 1.0710672497001008e-46 ], [ 657.375 , 1.071215910231113e-46 ], [ 657.4375 , 1.0713645913957066e-46 ], 
                    [ 657.5 , 1.0715132931967382e-46 ], [ 657.5625 , 1.0716620156370795e-46 ], [ 657.625 , 1.0718107587196027e-46 ], [ 657.6875 , 1.0719595224471655e-46 ], 
                    [ 657.75 , 1.0721083068226332e-46 ], [ 657.8125 , 1.072257111848864e-46 ], [ 657.875 , 1.072405937528732e-46 ], [ 657.9375 , 1.072554783865111e-46 ], 
                    [ 658.0 , 1.0727036508608532e-46 ], [ 658.0625 , 1.0728525385188414e-46 ], [ 658.125 , 1.0730014468419356e-46 ], [ 658.1875 , 1.0731503758329964e-46 ], 
                    [ 658.25 , 1.0732993254949076e-46 ], [ 658.3125 , 1.0734482958305387e-46 ], [ 658.375 , 1.0735972868427284e-46 ], [ 658.4375 , 1.0737462985343843e-46 ], 
                    [ 658.5 , 1.0738952582258785e-46 ], [ 658.5625 , 1.074044163188762e-46 ], [ 658.625 , 1.0741930887986047e-46 ], [ 658.6875 , 1.0743420350583078e-46 ], 
                    [ 658.75 , 1.0744910019706958e-46 ], [ 658.8125 , 1.0746399895386484e-46 ], [ 658.875 , 1.0747889977650445e-46 ], [ 658.9375 , 1.0749380266527185e-46 ], 
                    [ 659.0 , 1.07508707620455e-46 ], [ 659.0625 , 1.0752361464234044e-46 ], [ 659.125 , 1.0753852373121475e-46 ], [ 659.1875 , 1.0755343488736531e-46 ], 
                    [ 659.25 , 1.0756834811107798e-46 ], [ 659.3125 , 1.0758326340263868e-46 ], [ 659.375 , 1.075981807623349e-46 ], [ 659.4375 , 1.0761310019045267e-46 ], 
                    [ 659.5 , 1.0762802168728258e-46 ], [ 659.5625 , 1.0764294525310462e-46 ], [ 659.625 , 1.07657870888211e-46 ], [ 659.6875 , 1.0767279859288866e-46 ], 
                    [ 659.75 , 1.0768772836742152e-46 ], [ 659.8125 , 1.0770266021209885e-46 ], [ 659.875 , 1.0771759412720772e-46 ], [ 659.9375 , 1.0773253011303289e-46 ], 
                    [ 660.0 , 1.077474590881684e-46 ], [ 660.0625 , 1.0776238433860343e-46 ], [ 660.125 , 1.0777731165649416e-46 ], [ 660.1875 , 1.0779224104212626e-46 ], 
                    [ 660.25 , 1.0780717249578762e-46 ], [ 660.3125 , 1.0782210601776324e-46 ], [ 660.375 , 1.0783704160834112e-46 ], [ 660.4375 , 1.078519792678078e-46 ], 
                    [ 660.5 , 1.0786691899644832e-46 ], [ 660.5625 , 1.0788186079455011e-46 ], [ 660.625 , 1.0789680466239908e-46 ], [ 660.6875 , 1.0791175060028337e-46 ], 
                    [ 660.75 , 1.0792669860848907e-46 ], [ 660.8125 , 1.0794164868730212e-46 ], [ 660.875 , 1.0795660083701089e-46 ], [ 660.9375 , 1.079715550579015e-46 ], 
                    [ 661.0 , 1.0798651135026006e-46 ], [ 661.0625 , 1.0800146971437584e-46 ], [ 661.125 , 1.0801643015053274e-46 ], [ 661.1875 , 1.080313926590193e-46 ], 
                    [ 661.25 , 1.0804635724012486e-46 ], [ 661.3125 , 1.0806132389413199e-46 ], [ 661.375 , 1.0807629262133084e-46 ], [ 661.4375 , 1.0809126342200707e-46 ], 
                    [ 661.5 , 1.081062261256615e-46 ], [ 661.5625 , 1.081211860286819e-46 ], [ 661.625 , 1.0813614800187552e-46 ], [ 661.6875 , 1.0815111204552961e-46 ], 
                    [ 661.75 , 1.081660781599307e-46 ], [ 661.8125 , 1.0818104634536608e-46 ], [ 661.875 , 1.0819601660212084e-46 ], [ 661.9375 , 1.0821098893048236e-46 ], 
                    [ 662.0 , 1.0822596333073807e-46 ], [ 662.0625 , 1.0824093980317241e-46 ], [ 662.125 , 1.0825591834807594e-46 ], [ 662.1875 , 1.0827089896573086e-46 ], 
                    [ 662.25 , 1.0828588165642782e-46 ], [ 662.3125 , 1.083008664204522e-46 ], [ 662.375 , 1.0831585325809086e-46 ], [ 662.4375 , 1.0833084216962925e-46 ], 
                    [ 662.5 , 1.0834583315535741e-46 ], [ 662.5625 , 1.083608262155593e-46 ], [ 662.625 , 1.0837582135052429e-46 ], [ 662.6875 , 1.0839081856053952e-46 ], 
                    [ 662.75 , 1.0840581784589211e-46 ], [ 662.8125 , 1.0842081920686693e-46 ], [ 662.875 , 1.0843582264375355e-46 ], [ 662.9375 , 1.0845082815683691e-46 ], 
                    [ 663.0 , 1.084658253582905e-46 ], [ 663.0625 , 1.0848081991155977e-46 ], [ 663.125 , 1.0849581653770738e-46 ], [ 663.1875 , 1.0851081523702302e-46 ], 
                    [ 663.25 , 1.0852581600979166e-46 ], [ 663.3125 , 1.0854081885629923e-46 ], [ 663.375 , 1.0855582377683238e-46 ], [ 663.4375 , 1.0857083077168017e-46 ], 
                    [ 663.5 , 1.0858583984112855e-46 ], [ 663.5625 , 1.0860085098546126e-46 ], [ 663.625 , 1.08615864204969e-46 ], [ 663.6875 , 1.0863087949993629e-46 ], 
                    [ 663.75 , 1.0864589687065085e-46 ], [ 663.8125 , 1.086609163174004e-46 ], [ 663.875 , 1.086759378404704e-46 ], [ 663.9375 , 1.0869096144014866e-46 ], 
                    [ 664.0 , 1.08705987116723e-46 ], [ 664.0625 , 1.0872101487047902e-46 ], [ 664.125 , 1.0873604470170385e-46 ], [ 664.1875 , 1.0875107661068627e-46 ], 
                    [ 664.25 , 1.0876611059771345e-46 ], [ 664.3125 , 1.0878114666307193e-46 ], [ 664.375 , 1.0879618480704827e-46 ], [ 664.4375 , 1.0881122502993132e-46 ], 
                    [ 664.5 , 1.0882625735523814e-46 ], [ 664.5625 , 1.088412862567819e-46 ], [ 664.625 , 1.0885631723381631e-46 ], [ 664.6875 , 1.0887135028662803e-46 ], 
                    [ 664.75 , 1.088863854155037e-46 ], [ 664.8125 , 1.0890142262073081e-46 ], [ 664.875 , 1.0891646190259223e-46 ], [ 664.9375 , 1.089315032613817e-46 ], 
                    [ 665.0 , 1.0894654669738065e-46 ], [ 665.0625 , 1.0896159221087748e-46 ], [ 665.125 , 1.089766398021583e-46 ], [ 665.1875 , 1.089916894715124e-46 ], 
                    [ 665.25 , 1.0900674121922364e-46 ], [ 665.3125 , 1.0902179504558217e-46 ], [ 665.375 , 1.0903685095087268e-46 ], [ 665.4375 , 1.090519089353831e-46 ], 
                    [ 665.5 , 1.0906696899939898e-46 ], [ 665.5625 , 1.090820311432114e-46 ], [ 665.625 , 1.0909709536710447e-46 ], [ 665.6875 , 1.0911216167136622e-46 ], 
                    [ 665.75 , 1.0912723005628474e-46 ], [ 665.8125 , 1.09142300522145e-46 ], [ 665.875 , 1.09157373069236e-46 ], [ 665.9375 , 1.0917244769784738e-46 ], 
                    [ 666.0 , 1.0918751598789204e-46 ], [ 666.0625 , 1.0920257938450963e-46 ], [ 666.125 , 1.092176448592578e-46 ], [ 666.1875 , 1.0923271241242325e-46 ], 
                    [ 666.25 , 1.0924778204429199e-46 ], [ 666.3125 , 1.0926285375515228e-46 ], [ 666.375 , 1.092779275452902e-46 ], [ 666.4375 , 1.0929300341499257e-46 ], 
                    [ 666.5 , 1.0930808136454712e-46 ], [ 666.5625 , 1.0932316139423842e-46 ], [ 666.625 , 1.0933824350435576e-46 ], [ 666.6875 , 1.093533276951854e-46 ], 
                    [ 666.75 , 1.093684139670136e-46 ], [ 666.8125 , 1.0938350232012901e-46 ], [ 666.875 , 1.0939859275481801e-46 ], [ 666.9375 , 1.0941368527136773e-46 ], 
                    [ 667.0 , 1.094287798700654e-46 ], [ 667.0625 , 1.0944387655119828e-46 ], [ 667.125 , 1.0945897531505365e-46 ], [ 667.1875 , 1.0947407616191807e-46 ], 
                    [ 667.25 , 1.0948917909208123e-46 ], [ 667.3125 , 1.0950428410582822e-46 ], [ 667.375 , 1.0951939120344649e-46 ], [ 667.4375 , 1.0953450038522585e-46 ], 
                    [ 667.5 , 1.095496055956584e-46 ], [ 667.5625 , 1.0956470361410183e-46 ], [ 667.625 , 1.0957980371333847e-46 ], [ 667.6875 , 1.0959490589365664e-46 ], 
                    [ 667.75 , 1.0961001015534164e-46 ], [ 667.8125 , 1.0962511649868337e-46 ], [ 667.875 , 1.0964022492396487e-46 ], [ 667.9375 , 1.0965533543147385e-46 ], 
                    [ 668.0 , 1.0967044802149962e-46 ], [ 668.0625 , 1.0968556269432682e-46 ], [ 668.125 , 1.0970067945024252e-46 ], [ 668.1875 , 1.097157982895354e-46 ], 
                    [ 668.25 , 1.0973091921249173e-46 ], [ 668.3125 , 1.0974604221939876e-46 ], [ 668.375 , 1.097611673105421e-46 ], [ 668.4375 , 1.0977629448621368e-46 ], 
                    [ 668.5 , 1.097914237466961e-46 ], [ 668.5625 , 1.098065550922775e-46 ], [ 668.625 , 1.0982168852324599e-46 ], [ 668.6875 , 1.0983682403989057e-46 ], 
                    [ 668.75 , 1.0985196164249628e-46 ], [ 668.8125 , 1.0986710133135068e-46 ], [ 668.875 , 1.0988224310674282e-46 ], [ 668.9375 , 1.098973869689595e-46 ], 
                    [ 669.0 , 1.0991252994938882e-46 ], [ 669.0625 , 1.0992766220494637e-46 ], [ 669.125 , 1.0994279654384303e-46 ], [ 669.1875 , 1.0995793296636717e-46 ], 
                    [ 669.25 , 1.0997307147280491e-46 ], [ 669.3125 , 1.0998821206344392e-46 ], [ 669.375 , 1.1000335473856879e-46 ], [ 669.4375 , 1.1001849949846806e-46 ], 
                    [ 669.5 , 1.1003364634342956e-46 ], [ 669.5625 , 1.1004879527373954e-46 ], [ 669.625 , 1.1006394628968434e-46 ], [ 669.6875 , 1.1007909939155187e-46 ], 
                    [ 669.75 , 1.1009425457963088e-46 ], [ 669.8125 , 1.1010941185420547e-46 ], [ 669.875 , 1.1012457121556447e-46 ], [ 669.9375 , 1.1013973266399514e-46 ], 
                    [ 670.0 , 1.1015489619978642e-46 ], [ 670.0625 , 1.1017006182322332e-46 ], [ 670.125 , 1.1018522953459327e-46 ], [ 670.1875 , 1.1020039933418453e-46 ], 
                    [ 670.25 , 1.1021557122228612e-46 ], [ 670.3125 , 1.1023074519918244e-46 ], [ 670.375 , 1.1024592126516269e-46 ], [ 670.4375 , 1.10261099420516e-46 ], 
                    [ 670.5 , 1.1027627966552692e-46 ], [ 670.5625 , 1.1029144767169433e-46 ], [ 670.625 , 1.1030661662561203e-46 ], [ 670.6875 , 1.103217876657945e-46 ], 
                    [ 670.75 , 1.103369607925279e-46 ], [ 670.8125 , 1.1035213600610156e-46 ], [ 670.875 , 1.1036731330680092e-46 ], [ 670.9375 , 1.1038249269491149e-46 ], 
                    [ 671.0 , 1.1039767417072268e-46 ], [ 671.0625 , 1.1041285773452165e-46 ], [ 671.125 , 1.1042804338659396e-46 ], [ 671.1875 , 1.1044323112722765e-46 ], 
                    [ 671.25 , 1.1045842095670997e-46 ], [ 671.3125 , 1.104736128753274e-46 ], [ 671.375 , 1.1048880688336963e-46 ], [ 671.4375 , 1.1050400298112244e-46 ], 
                    [ 671.5 , 1.1051920116887173e-46 ], [ 671.5625 , 1.1053440144690805e-46 ], [ 671.625 , 1.1054960381551498e-46 ], [ 671.6875 , 1.105648082749847e-46 ], 
                    [ 671.75 , 1.1058001482560249e-46 ], [ 671.8125 , 1.1059522346765513e-46 ], [ 671.875 , 1.1061043420143107e-46 ], [ 671.9375 , 1.10625647027218e-46 ], 
                    [ 672.0 , 1.1064086194530363e-46 ], [ 672.0625 , 1.1065606933119511e-46 ], [ 672.125 , 1.106712726467098e-46 ], [ 672.1875 , 1.1068647805104624e-46 ], 
                    [ 672.25 , 1.1070168554449218e-46 ], [ 672.3125 , 1.107168951273331e-46 ], [ 672.375 , 1.1073210679985603e-46 ], [ 672.4375 , 1.1074732056235046e-46 ], 
                    [ 672.5 , 1.1076253641510118e-46 ], [ 672.5625 , 1.1077775435839692e-46 ], [ 672.625 , 1.1079297439252338e-46 ], [ 672.6875 , 1.1080819651776858e-46 ], 
                    [ 672.75 , 1.1082342073441906e-46 ], [ 672.8125 , 1.1083864704276528e-46 ], [ 672.875 , 1.108538754430907e-46 ], [ 672.9375 , 1.1086910593568352e-46 ], 
                    [ 673.0 , 1.1088433852083282e-46 ], [ 673.0625 , 1.1089957319882526e-46 ], [ 673.125 , 1.109148099699492e-46 ], [ 673.1875 , 1.109300488344914e-46 ], 
                    [ 673.25 , 1.1094528979273875e-46 ], [ 673.3125 , 1.1096053284497966e-46 ], [ 673.375 , 1.1097577799150183e-46 ], [ 673.4375 , 1.1099102523259379e-46 ], 
                    [ 673.5 , 1.1100627456854016e-46 ], [ 673.5625 , 1.11021522215895e-46 ], [ 673.625 , 1.110367598363054e-46 ], [ 673.6875 , 1.1105199954806539e-46 ], 
                    [ 673.75 , 1.1106724135146672e-46 ], [ 673.8125 , 1.1108248524679489e-46 ], [ 673.875 , 1.1109773123433465e-46 ], [ 673.9375 , 1.111129793143755e-46 ], 
                    [ 674.0 , 1.1112822948720312e-46 ], [ 674.0625 , 1.1114348175310704e-46 ], [ 674.125 , 1.111587361123714e-46 ], [ 674.1875 , 1.1117399256528589e-46 ], 
                    [ 674.25 , 1.1118925111213628e-46 ], [ 674.3125 , 1.1120451175321073e-46 ], [ 674.375 , 1.112197744887975e-46 ], [ 674.4375 , 1.1123503931918403e-46 ], 
                    [ 674.5 , 1.1125030624465783e-46 ], [ 674.5625 , 1.1126557526550172e-46 ], [ 674.625 , 1.1128084638200803e-46 ], [ 674.6875 , 1.1129611959446676e-46 ], 
                    [ 674.75 , 1.1131139490316004e-46 ], [ 674.8125 , 1.113266723083795e-46 ], [ 674.875 , 1.113419518104082e-46 ], [ 674.9375 , 1.1135723340954021e-46 ], 
                    [ 675.0 , 1.1137251710605942e-46 ], [ 675.0625 , 1.1138780290025448e-46 ], [ 675.125 , 1.1140307757613052e-46 ], [ 675.1875 , 1.114183512605838e-46 ], 
                    [ 675.25 , 1.1143362703910346e-46 ], [ 675.3125 , 1.1144890491197735e-46 ], [ 675.375 , 1.1146418487949106e-46 ], [ 675.4375 , 1.1147946694193256e-46 ], 
                    [ 675.5 , 1.1149475109958985e-46 ], [ 675.5625 , 1.1151003735274861e-46 ], [ 675.625 , 1.1152532570169612e-46 ], [ 675.6875 , 1.1154061614672215e-46 ], 
                    [ 675.75 , 1.1155590868811244e-46 ], [ 675.8125 , 1.1157120332615365e-46 ], [ 675.875 , 1.1158650006113478e-46 ], [ 675.9375 , 1.116017988933418e-46 ], 
                    [ 676.0 , 1.1161709982306218e-46 ], [ 676.0625 , 1.1163240285058515e-46 ], [ 676.125 , 1.1164770797619986e-46 ], [ 676.1875 , 1.116630152001884e-46 ], 
                    [ 676.25 , 1.1167832452284171e-46 ], [ 676.3125 , 1.1169363594444904e-46 ], [ 676.375 , 1.117089494652974e-46 ], [ 676.4375 , 1.1172426508567064e-46 ], 
                    [ 676.5 , 1.1173958280586291e-46 ], [ 676.5625 , 1.1175490262615661e-46 ], [ 676.625 , 1.1177021935645287e-46 ], [ 676.6875 , 1.117855275660056e-46 ], 
                    [ 676.75 , 1.1180083787219108e-46 ], [ 676.8125 , 1.1181615027529726e-46 ], [ 676.875 , 1.1183146477561372e-46 ], [ 676.9375 , 1.1184678137342531e-46 ], 
                    [ 677.0 , 1.118621000690209e-46 ], [ 677.0625 , 1.1187742086268703e-46 ], [ 677.125 , 1.11892743754711e-46 ], [ 677.1875 , 1.119080687453779e-46 ], 
                    [ 677.25 , 1.1192339583498063e-46 ], [ 677.3125 , 1.1193872502380278e-46 ], [ 677.375 , 1.119540563121326e-46 ], [ 677.4375 , 1.1196938970025607e-46 ], 
                    [ 677.5 , 1.1198472518846395e-46 ], [ 677.5625 , 1.120000627770431e-46 ], [ 677.625 , 1.1201540246627877e-46 ], [ 677.6875 , 1.1203074425645951e-46 ], 
                    [ 677.75 , 1.120460881478762e-46 ], [ 677.8125 , 1.1206143414081191e-46 ], [ 677.875 , 1.1207678223555918e-46 ], [ 677.9375 , 1.1209213243240035e-46 ], 
                    [ 678.0 , 1.1210748473163047e-46 ], [ 678.0625 , 1.1212283913353114e-46 ], [ 678.125 , 1.1213819563839269e-46 ], [ 678.1875 , 1.1215354161537212e-46 ], 
                    [ 678.25 , 1.1216888634242017e-46 ], [ 678.3125 , 1.1218423316891807e-46 ], [ 678.375 , 1.1219958209515149e-46 ], [ 678.4375 , 1.122149331214085e-46 ], 
                    [ 678.5 , 1.1223028624797803e-46 ], [ 678.5625 , 1.1224564147514424e-46 ], [ 678.625 , 1.1226099880319613e-46 ], [ 678.6875 , 1.1227635823242274e-46 ], 
                    [ 678.75 , 1.1229171976310755e-46 ], [ 678.8125 , 1.1230708339554208e-46 ], [ 678.875 , 1.1232244913001151e-46 ], [ 678.9375 , 1.12337816966805e-46 ], 
                    [ 679.0 , 1.123531869062086e-46 ], [ 679.0625 , 1.1236855894851161e-46 ], [ 679.125 , 1.1238393309400093e-46 ], [ 679.1875 , 1.123993093429643e-46 ], 
                    [ 679.25 , 1.1241468769568874e-46 ], [ 679.3125 , 1.1243006815246288e-46 ], [ 679.375 , 1.1244545071357538e-46 ], [ 679.4375 , 1.1246083537931258e-46 ], 
                    [ 679.5 , 1.1247622214996478e-46 ], [ 679.5625 , 1.1249161102581684e-46 ], [ 679.625 , 1.1250700200715754e-46 ], [ 679.6875 , 1.125223917961723e-46 ], 
                    [ 679.75 , 1.125377702279508e-46 ], [ 679.8125 , 1.1255315076149909e-46 ], [ 679.875 , 1.1256853339710363e-46 ], [ 679.9375 , 1.1258391813505328e-46 ], 
                    [ 680.0 , 1.125993049756338e-46 ], [ 680.0625 , 1.1261469391913494e-46 ], [ 680.125 , 1.126300849658401e-46 ], [ 680.1875 , 1.1264547811604075e-46 ], 
                    [ 680.25 , 1.1266087337002194e-46 ], [ 680.3125 , 1.1267627072807363e-46 ], [ 680.375 , 1.1269167019048094e-46 ], [ 680.4375 , 1.1270707175753072e-46 ], 
                    [ 680.5 , 1.1272247542951538e-46 ], [ 680.5625 , 1.127378812067146e-46 ], [ 680.625 , 1.1275328908942574e-46 ], [ 680.6875 , 1.127686990779293e-46 ], 
                    [ 680.75 , 1.127841111725147e-46 ], [ 680.8125 , 1.1279952537347462e-46 ], [ 680.875 , 1.128149416810889e-46 ], [ 680.9375 , 1.1283036009565024e-46 ], 
                    [ 681.0 , 1.1284578061744662e-46 ], [ 681.0625 , 1.1286120324676445e-46 ], [ 681.125 , 1.1287662798389333e-46 ], [ 681.1875 , 1.1289205482912135e-46 ], 
                    [ 681.25 , 1.1290747490326496e-46 ], [ 681.3125 , 1.1292288974451094e-46 ], [ 681.375 , 1.1293830669028594e-46 ], [ 681.4375 , 1.1295372574088131e-46 ], 
                    [ 681.5 , 1.1296914689658284e-46 ], [ 681.5625 , 1.1298457015767626e-46 ], [ 681.625 , 1.1299999552445228e-46 ], [ 681.6875 , 1.130154229971967e-46 ], 
                    [ 681.75 , 1.1303085257619633e-46 ], [ 681.8125 , 1.1304628426174029e-46 ], [ 681.875 , 1.130617180541154e-46 ], [ 681.9375 , 1.1307715395360763e-46 ], 
                    [ 682.0 , 1.1309259196050871e-46 ], [ 682.0625 , 1.1310803207510317e-46 ], [ 682.125 , 1.1312347429767874e-46 ], [ 682.1875 , 1.1313891862852402e-46 ], 
                    [ 682.25 , 1.1315436506792765e-46 ], [ 682.3125 , 1.131698136161767e-46 ], [ 682.375 , 1.1318526427355667e-46 ], [ 682.4375 , 1.1320071704036112e-46 ], 
                    [ 682.5 , 1.1321617191687326e-46 ], [ 682.5625 , 1.132316289033835e-46 ], [ 682.625 , 1.1324708800017912e-46 ], [ 682.6875 , 1.1326254920754663e-46 ], 
                    [ 682.75 , 1.1327801252577737e-46 ], [ 682.8125 , 1.1329346406510475e-46 ], [ 682.875 , 1.1330891511310833e-46 ], [ 682.9375 , 1.133243682683383e-46 ], 
                    [ 683.0 , 1.1333982353108282e-46 ], [ 683.0625 , 1.1335528090162613e-46 ], [ 683.125 , 1.1337074038025808e-46 ], [ 683.1875 , 1.1338620196726375e-46 ], 
                    [ 683.25 , 1.1340166566293395e-46 ], [ 683.3125 , 1.1341713146755458e-46 ], [ 683.375 , 1.134325993814117e-46 ], [ 683.4375 , 1.134480694047962e-46 ], 
                    [ 683.5 , 1.1346354153799413e-46 ], [ 683.5625 , 1.1347901578129245e-46 ], [ 683.625 , 1.1349449213498135e-46 ], [ 683.6875 , 1.1350997059934539e-46 ], 
                    [ 683.75 , 1.1352545117467328e-46 ], [ 683.8125 , 1.1354093386125529e-46 ], [ 683.875 , 1.1355641865937859e-46 ], [ 683.9375 , 1.1357190556932951e-46 ], 
                    [ 684.0 , 1.1358739459139448e-46 ], [ 684.0625 , 1.1360288572586716e-46 ], [ 684.125 , 1.1361837897303082e-46 ], [ 684.1875 , 1.136338743331768e-46 ], 
                    [ 684.25 , 1.1364937180658928e-46 ], [ 684.3125 , 1.1366487005853895e-46 ], [ 684.375 , 1.136803552995785e-46 ], [ 684.4375 , 1.136958426502629e-46 ], 
                    [ 684.5 , 1.1371133211088197e-46 ], [ 684.5625 , 1.1372682368172235e-46 ], [ 684.625 , 1.1374231736307153e-46 ], [ 684.6875 , 1.1375781315521703e-46 ], 
                    [ 684.75 , 1.1377331105844643e-46 ], [ 684.8125 , 1.1378881107304735e-46 ], [ 684.875 , 1.1380431319930658e-46 ], [ 684.9375 , 1.1381981743751345e-46 ], 
                    [ 685.0 , 1.1383532378795488e-46 ], [ 685.0625 , 1.138508322509178e-46 ], [ 685.125 , 1.1386634282669245e-46 ], [ 685.1875 , 1.1388185551556345e-46 ], 
                    [ 685.25 , 1.1389737031782028e-46 ], [ 685.3125 , 1.1391288723375167e-46 ], [ 685.375 , 1.1392840626364722e-46 ], [ 685.4375 , 1.1394392740778841e-46 ], 
                    [ 685.5 , 1.1395945066646897e-46 ], [ 685.5625 , 1.1397497603997536e-46 ], [ 685.625 , 1.139905035285973e-46 ], [ 685.6875 , 1.140060331326181e-46 ], 
                    [ 685.75 , 1.1402156485233e-46 ], [ 685.8125 , 1.1403709868802124e-46 ], [ 685.875 , 1.1405263012675971e-46 ], [ 685.9375 , 1.1406815108643049e-46 ], 
                    [ 686.0 , 1.140836741582857e-46 ], [ 686.0625 , 1.1409919934261278e-46 ], [ 686.125 , 1.141147266396992e-46 ], [ 686.1875 , 1.1413025604983248e-46 ], 
                    [ 686.25 , 1.1414578757330017e-46 ], [ 686.3125 , 1.1416132121038907e-46 ], [ 686.375 , 1.141768569613884e-46 ], [ 686.4375 , 1.141923948265867e-46 ], 
                    [ 686.5 , 1.142079348062684e-46 ], [ 686.5625 , 1.142234769007229e-46 ], [ 686.625 , 1.1423902111023793e-46 ], [ 686.6875 , 1.1425456743509976e-46 ], 
                    [ 686.75 , 1.1427011587560112e-46 ], [ 686.8125 , 1.1428566643202584e-46 ], [ 686.875 , 1.1430121910466187e-46 ], [ 686.9375 , 1.1431677389379641e-46 ], 
                    [ 687.0 , 1.1433233079972151e-46 ], [ 687.0625 , 1.1434788982272283e-46 ], [ 687.125 , 1.1436345096308763e-46 ], [ 687.1875 , 1.1437901422110408e-46 ], 
                    [ 687.25 , 1.14394579597062e-46 ], [ 687.3125 , 1.1441014709124954e-46 ], [ 687.375 , 1.1442571670395342e-46 ], [ 687.4375 , 1.1444128202813256e-46 ], 
                    [ 687.5 , 1.144568394566848e-46 ], [ 687.5625 , 1.1447239900015205e-46 ], [ 687.625 , 1.1448796065882016e-46 ], [ 687.6875 , 1.1450352443297912e-46 ], 
                    [ 687.75 , 1.1451909032291572e-46 ], [ 687.8125 , 1.145346583289192e-46 ], [ 687.875 , 1.1455022845127314e-46 ], [ 687.9375 , 1.1456580069026934e-46 ], 
                    [ 688.0 , 1.1458137504619308e-46 ], [ 688.0625 , 1.1459695151933379e-46 ], [ 688.125 , 1.1461253010997682e-46 ], [ 688.1875 , 1.146281108184133e-46 ], 
                    [ 688.25 , 1.146436936449287e-46 ], [ 688.3125 , 1.1465927858981337e-46 ], [ 688.375 , 1.146748656533504e-46 ], [ 688.4375 , 1.1469045483583352e-46 ], 
                    [ 688.5 , 1.1470604613754836e-46 ], [ 688.5625 , 1.1472163955878134e-46 ], [ 688.625 , 1.1473723509982472e-46 ], [ 688.6875 , 1.1475283276096172e-46 ], 
                    [ 688.75 , 1.1476843254248386e-46 ], [ 688.8125 , 1.1478403444467938e-46 ], [ 688.875 , 1.1479963846783575e-46 ], [ 688.9375 , 1.1481524461223966e-46 ], 
                    [ 689.0 , 1.1483084489331708e-46 ], [ 689.0625 , 1.1484643793130782e-46 ], [ 689.125 , 1.1486203308669696e-46 ], [ 689.1875 , 1.1487763035977447e-46 ], 
                    [ 689.25 , 1.1489322975082708e-46 ], [ 689.3125 , 1.149088312601416e-46 ], [ 689.375 , 1.149244348880073e-46 ], [ 689.4375 , 1.1494004063471103e-46 ], 
                    [ 689.5 , 1.1495564850054051e-46 ], [ 689.5625 , 1.149712584857827e-46 ], [ 689.625 , 1.1498687059072703e-46 ], [ 689.6875 , 1.1500248481565967e-46 ], 
                    [ 689.75 , 1.15018101160871e-46 ], [ 689.8125 , 1.1503371962664644e-46 ], [ 689.875 , 1.1504934021327561e-46 ], [ 689.9375 , 1.1506496292104484e-46 ], 
                    [ 690.0 , 1.15080587750243e-46 ], [ 690.0625 , 1.1509621470115653e-46 ], [ 690.125 , 1.151118437740784e-46 ], [ 690.1875 , 1.151274749692911e-46 ], 
                    [ 690.25 , 1.1514310828708607e-46 ], [ 690.3125 , 1.1515874372774906e-46 ], [ 690.375 , 1.1517438129157248e-46 ], [ 690.4375 , 1.1519002097884212e-46 ], 
                    [ 690.5 , 1.1520566278984554e-46 ], [ 690.5625 , 1.1522129878205849e-46 ], [ 690.625 , 1.1523692807822648e-46 ], [ 690.6875 , 1.1525255949444269e-46 ], 
                    [ 690.75 , 1.1526819303099797e-46 ], [ 690.8125 , 1.152838286881775e-46 ], [ 690.875 , 1.1529946646627056e-46 ], [ 690.9375 , 1.1531510636556481e-46 ], 
                    [ 691.0 , 1.1533074838634475e-46 ], [ 691.0625 , 1.1534639252890226e-46 ], [ 691.125 , 1.1536203879352347e-46 ], [ 691.1875 , 1.153776871804946e-46 ], 
                    [ 691.25 , 1.1539333769010683e-46 ], [ 691.3125 , 1.1540899032264643e-46 ], [ 691.375 , 1.1542464507840055e-46 ], [ 691.4375 , 1.1544030195765803e-46 ], 
                    [ 691.5 , 1.1545596096070854e-46 ], [ 691.5625 , 1.1547162208783773e-46 ], [ 691.625 , 1.1548728533933286e-46 ], [ 691.6875 , 1.1550295071548622e-46 ], 
                    [ 691.75 , 1.155186182165819e-46 ], [ 691.8125 , 1.1553428784290979e-46 ], [ 691.875 , 1.1554995959475816e-46 ], [ 691.9375 , 1.1556563347241534e-46 ], 
                    [ 692.0 , 1.1558130947617132e-46 ], [ 692.0625 , 1.155969876063112e-46 ], [ 692.125 , 1.1561266040509043e-46 ], [ 692.1875 , 1.156283254766643e-46 ], 
                    [ 692.25 , 1.1564399267079688e-46 ], [ 692.3125 , 1.1565966198777245e-46 ], [ 692.375 , 1.156753334278803e-46 ], [ 692.4375 , 1.1569100699140892e-46 ], 
                    [ 692.5 , 1.1570668267864443e-46 ], [ 692.5625 , 1.1572236048987616e-46 ], [ 692.625 , 1.157380404253895e-46 ], [ 692.6875 , 1.1575372248547554e-46 ], 
                    [ 692.75 , 1.1576940667041966e-46 ], [ 692.8125 , 1.1578509298051063e-46 ], [ 692.875 , 1.1580078141603802e-46 ], [ 692.9375 , 1.1581647197728739e-46 ], 
                    [ 693.0 , 1.1583216466454756e-46 ], [ 693.0625 , 1.1584785947810578e-46 ], [ 693.125 , 1.1586355641825096e-46 ], [ 693.1875 , 1.1587925548527207e-46 ], 
                    [ 693.25 , 1.1589495667945486e-46 ], [ 693.3125 , 1.1591066000109082e-46 ], [ 693.375 , 1.1592636545046656e-46 ], [ 693.4375 , 1.1594207302786955e-46 ], 
                    [ 693.5 , 1.1595778273358897e-46 ], [ 693.5625 , 1.1597349456791405e-46 ], [ 693.625 , 1.1598920853113314e-46 ], [ 693.6875 , 1.1600491894615142e-46 ], 
                    [ 693.75 , 1.1602062001479905e-46 ], [ 693.8125 , 1.1603632320855884e-46 ], [ 693.875 , 1.1605202852772005e-46 ], [ 693.9375 , 1.1606773597256785e-46 ], 
                    [ 694.0 , 1.1608344554339245e-46 ], [ 694.0625 , 1.1609915724048077e-46 ], [ 694.125 , 1.1611487106412224e-46 ], [ 694.1875 , 1.1613058701459973e-46 ], 
                    [ 694.25 , 1.1614630509220689e-46 ], [ 694.3125 , 1.1616202529722914e-46 ], [ 694.375 , 1.1617774762995277e-46 ], [ 694.4375 , 1.1619347209066908e-46 ], 
                    [ 694.5 , 1.1620919867966358e-46 ], [ 694.5625 , 1.1622492739722521e-46 ], [ 694.625 , 1.1624065824364037e-46 ], [ 694.6875 , 1.1625639121920215e-46 ], 
                    [ 694.75 , 1.162721263241938e-46 ], [ 694.8125 , 1.1628786355890433e-46 ], [ 694.875 , 1.163036029236245e-46 ], [ 694.9375 , 1.1631934441863928e-46 ], 
                    [ 695.0 , 1.1633508804423948e-46 ], [ 695.0625 , 1.1635083380071348e-46 ], [ 695.125 , 1.1636658168834804e-46 ], [ 695.1875 , 1.1638233170743326e-46 ], 
                    [ 695.25 , 1.1639808086602834e-46 ], [ 695.3125 , 1.1641381797313182e-46 ], [ 695.375 , 1.1642955720790388e-46 ], [ 695.4375 , 1.1644529857063216e-46 ], 
                    [ 695.5 , 1.1646104206160443e-46 ], [ 695.5625 , 1.1647678768110835e-46 ], [ 695.625 , 1.1649253542943093e-46 ], [ 695.6875 , 1.1650828530686247e-46 ], 
                    [ 695.75 , 1.1652403731368748e-46 ], [ 695.8125 , 1.1653979145019552e-46 ], [ 695.875 , 1.165555477166762e-46 ], [ 695.9375 , 1.1657130611341499e-46 ], 
                    [ 696.0 , 1.165870666406999e-46 ], [ 696.0625 , 1.1660282929882067e-46 ], [ 696.125 , 1.1661859408806453e-46 ], [ 696.1875 , 1.1663436100871962e-46 ], 
                    [ 696.25 , 1.1665013006107497e-46 ], [ 696.3125 , 1.166659012454154e-46 ], [ 696.375 , 1.1668167456203337e-46 ], [ 696.4375 , 1.1669745001121545e-46 ], 
                    [ 696.5 , 1.1671322759325001e-46 ], [ 696.5625 , 1.1672900730842455e-46 ], [ 696.625 , 1.1674478915703e-46 ], [ 696.6875 , 1.1676057313935227e-46 ], 
                    [ 696.75 , 1.167763592556807e-46 ], [ 696.8125 , 1.1679214750630295e-46 ], [ 696.875 , 1.168079208687496e-46 ], [ 696.9375 , 1.168236956609528e-46 ], 
                    [ 697.0 , 1.1683947258352585e-46 ], [ 697.0625 , 1.1685525163675556e-46 ], [ 697.125 , 1.1687103282093223e-46 ], [ 697.1875 , 1.1688681613634108e-46 ], 
                    [ 697.25 , 1.1690260158326999e-46 ], [ 697.3125 , 1.1691838916201012e-46 ], [ 697.375 , 1.1693417887284437e-46 ], [ 697.4375 , 1.1694997071606404e-46 ], 
                    [ 697.5 , 1.1696576469195621e-46 ], [ 697.5625 , 1.1698156080080897e-46 ], [ 697.625 , 1.1699735904291115e-46 ], [ 697.6875 , 1.1701315941854922e-46 ], 
                    [ 697.75 , 1.1702896192801212e-46 ], [ 697.8125 , 1.1704476657158634e-46 ], [ 697.875 , 1.170605733495668e-46 ], [ 697.9375 , 1.1707638226223169e-46 ], 
                    [ 698.0 , 1.170921933098752e-46 ], [ 698.0625 , 1.1710800649278728e-46 ], [ 698.125 , 1.171238218112513e-46 ], [ 698.1875 , 1.1713963926555816e-46 ], 
                    [ 698.25 , 1.17155458855998e-46 ], [ 698.3125 , 1.1717128058285758e-46 ], [ 698.375 , 1.1718710444642298e-46 ], [ 698.4375 , 1.172029183679532e-46 ], 
                    [ 698.5 , 1.1721872898580735e-46 ], [ 698.5625 , 1.1723454173650461e-46 ], [ 698.625 , 1.1725035662033611e-46 ], [ 698.6875 , 1.1726617363758622e-46 ], 
                    [ 698.75 , 1.1728199278854525e-46 ], [ 698.8125 , 1.1729781407350023e-46 ], [ 698.875 , 1.1731363749273898e-46 ], [ 698.9375 , 1.1732946304654946e-46 ], 
                    [ 699.0 , 1.1734529073521879e-46 ], [ 699.0625 , 1.1736112055903662e-46 ], [ 699.125 , 1.173769525182901e-46 ], [ 699.1875 , 1.1739278661326823e-46 ], 
                    [ 699.25 , 1.1740862284425737e-46 ], [ 699.3125 , 1.1742446121154651e-46 ], [ 699.375 , 1.1744030171542387e-46 ], [ 699.4375 , 1.1745614435617683e-46 ], 
                    [ 699.5 , 1.1747198913409612e-46 ], [ 699.5625 , 1.1748783604946594e-46 ], [ 699.625 , 1.1750368510257874e-46 ], [ 699.6875 , 1.1751953629372208e-46 ], 
                    [ 699.75 , 1.175353896231819e-46 ], [ 699.8125 , 1.1755124509124998e-46 ], [ 699.875 , 1.1756710269821148e-46 ], [ 699.9375 , 1.1758296244435825e-46 ], 
                    [ 700.0 , 1.1759881791972108e-46 ], [ 700.0625 , 1.1761466420652462e-46 ], [ 700.125 , 1.1763051262859374e-46 ], [ 700.1875 , 1.176463631862178e-46 ], 
                    [ 700.25 , 1.1766221587968296e-46 ], [ 700.3125 , 1.1767807070927863e-46 ], [ 700.375 , 1.1769392767529187e-46 ], [ 700.4375 , 1.1770978677801048e-46 ], 
                    [ 700.5 , 1.1772564801772245e-46 ], [ 700.5625 , 1.177415113947157e-46 ], [ 700.625 , 1.1775737690927742e-46 ], [ 700.6875 , 1.1777324456169728e-46 ], 
                    [ 700.75 , 1.1778911435226171e-46 ], [ 700.8125 , 1.1780498628126132e-46 ], [ 700.875 , 1.1782086034898175e-46 ], [ 700.9375 , 1.1783673655571035e-46 ], 
                    [ 701.0 , 1.1785261490173957e-46 ], [ 701.0625 , 1.1786849538735426e-46 ], [ 701.125 , 1.1788437801284362e-46 ], [ 701.1875 , 1.1790026277849516e-46 ], 
                    [ 701.25 , 1.1791614968459972e-46 ], [ 701.3125 , 1.179320387314416e-46 ], [ 701.375 , 1.1794792991931257e-46 ], [ 701.4375 , 1.1796382324850283e-46 ], 
                    [ 701.5 , 1.1797971871929587e-46 ], [ 701.5625 , 1.1799561633198195e-46 ], [ 701.625 , 1.180114987089502e-46 ], [ 701.6875 , 1.1802738279340522e-46 ], 
                    [ 701.75 , 1.1804326901582344e-46 ], [ 701.8125 , 1.180591573764909e-46 ], [ 701.875 , 1.180750478756971e-46 ], [ 701.9375 , 1.1809094051372822e-46 ], 
                    [ 702.0 , 1.1810683529087294e-46 ], [ 702.0625 , 1.1812273220742006e-46 ], [ 702.125 , 1.1813863126365668e-46 ], [ 702.1875 , 1.1815453245986828e-46 ], 
                    [ 702.25 , 1.1817043579634961e-46 ], [ 702.3125 , 1.1818634127338119e-46 ], [ 702.375 , 1.182022488912562e-46 ], [ 702.4375 , 1.1821815865026105e-46 ], 
                    [ 702.5 , 1.1823407055068233e-46 ], [ 702.5625 , 1.1824998459281073e-46 ], [ 702.625 , 1.1826590077693536e-46 ], [ 702.6875 , 1.1828181910333953e-46 ], 
                    [ 702.75 , 1.1829773957231659e-46 ], [ 702.8125 , 1.1831366218415579e-46 ], [ 702.875 , 1.1832958693913964e-46 ], [ 702.9375 , 1.1834551383756086e-46 ], 
                    [ 703.0 , 1.1836144287970788e-46 ], [ 703.0625 , 1.1837737406586761e-46 ], [ 703.125 , 1.1839330739633029e-46 ], [ 703.1875 , 1.1840923349076392e-46 ], 
                    [ 703.25 , 1.1842515350080193e-46 ], [ 703.3125 , 1.1844107565126777e-46 ], [ 703.375 , 1.1845699994245176e-46 ], [ 703.4375 , 1.1847292637464338e-46 ], 
                    [ 703.5 , 1.18488854948128e-46 ], [ 703.5625 , 1.1850478566319434e-46 ], [ 703.625 , 1.1852071852013034e-46 ], [ 703.6875 , 1.1853665351922396e-46 ], 
                    [ 703.75 , 1.1855259066076405e-46 ], [ 703.8125 , 1.1856852994503614e-46 ], [ 703.875 , 1.1858447137233085e-46 ], [ 703.9375 , 1.186004149429363e-46 ], 
                    [ 704.0 , 1.1861636065713814e-46 ], [ 704.0625 , 1.186323085152271e-46 ], [ 704.125 , 1.1864825851749055e-46 ], [ 704.1875 , 1.186642106642185e-46 ], 
                    [ 704.25 , 1.1868016495569334e-46 ], [ 704.3125 , 1.186961213922119e-46 ], [ 704.375 , 1.1871207997405496e-46 ], [ 704.4375 , 1.1872804070151852e-46 ], 
                    [ 704.5 , 1.1874400357488267e-46 ], [ 704.5625 , 1.1875996859444184e-46 ], [ 704.625 , 1.1877593576048281e-46 ], [ 704.6875 , 1.1879190507329344e-46 ], 
                    [ 704.75 , 1.1880787594759717e-46 ], [ 704.8125 , 1.1882383125325663e-46 ], [ 704.875 , 1.1883978870163578e-46 ], [ 704.9375 , 1.188557482930182e-46 ], 
                    [ 705.0 , 1.1887171002769677e-46 ], [ 705.0625 , 1.188876739059559e-46 ], [ 705.125 , 1.1890363992808432e-46 ], [ 705.1875 , 1.1891960809436911e-46 ], 
                    [ 705.25 , 1.189355784050999e-46 ], [ 705.3125 , 1.1895155086056383e-46 ], [ 705.375 , 1.1896752546104807e-46 ], [ 705.4375 , 1.1898350220684324e-46 ], 
                    [ 705.5 , 1.1899948109823409e-46 ], [ 705.5625 , 1.1901546213551039e-46 ], [ 705.625 , 1.1903144531896207e-46 ], [ 705.6875 , 1.1904743064887391e-46 ], 
                    [ 705.75 , 1.1906341812553762e-46 ], [ 705.8125 , 1.1907940774923807e-46 ], [ 705.875 , 1.1909539952026781e-46 ], [ 705.9375 , 1.1911139343890935e-46 ], 
                    [ 706.0 , 1.19127389505457e-46 ], [ 706.0625 , 1.191433877201941e-46 ], [ 706.125 , 1.1915938808341598e-46 ], [ 706.1875 , 1.1917539059540265e-46 ], 
                    [ 706.25 , 1.1919139525644948e-46 ], [ 706.3125 , 1.1920740206684083e-46 ], [ 706.375 , 1.1922340244825185e-46 ], [ 706.4375 , 1.1923939542874004e-46 ], 
                    [ 706.5 , 1.192553905545732e-46 ], [ 706.5625 , 1.1927138782604246e-46 ], [ 706.625 , 1.1928738724343143e-46 ], [ 706.6875 , 1.1930338880702883e-46 ], 
                    [ 706.75 , 1.1931939251712505e-46 ], [ 706.8125 , 1.1933539837400468e-46 ], [ 706.875 , 1.1935140637795907e-46 ], [ 706.9375 , 1.1936741652927372e-46 ], 
                    [ 707.0 , 1.1938342882823665e-46 ], [ 707.0625 , 1.193994432751368e-46 ], [ 707.125 , 1.1941545987026316e-46 ], [ 707.1875 , 1.1943147861390392e-46 ], 
                    [ 707.25 , 1.1944749950634555e-46 ], [ 707.3125 , 1.1946352254787634e-46 ], [ 707.375 , 1.1947954773878707e-46 ], [ 707.4375 , 1.194955750793644e-46 ], 
                    [ 707.5 , 1.1951160456989414e-46 ], [ 707.5625 , 1.1952763621067065e-46 ], [ 707.625 , 1.195436700019764e-46 ], [ 707.6875 , 1.1955970594410329e-46 ], 
                    [ 707.75 , 1.1957574403734066e-46 ], [ 707.8125 , 1.1959178428197283e-46 ], [ 707.875 , 1.196078266782901e-46 ], [ 707.9375 , 1.196238712265845e-46 ], 
                    [ 708.0 , 1.1963990236035447e-46 ], [ 708.0625 , 1.1965593304042111e-46 ], [ 708.125 , 1.1967196586845681e-46 ], [ 708.1875 , 1.1968800084474848e-46 ], 
                    [ 708.25 , 1.1970403796958398e-46 ], [ 708.3125 , 1.1972007724324952e-46 ], [ 708.375 , 1.1973611866603723e-46 ], [ 708.4375 , 1.1975216223823088e-46 ], 
                    [ 708.5 , 1.1976820796012096e-46 ], [ 708.5625 , 1.1978425583199473e-46 ], [ 708.625 , 1.198003058541402e-46 ], [ 708.6875 , 1.1981635802684553e-46 ], 
                    [ 708.75 , 1.19832412350398e-46 ], [ 708.8125 , 1.1984846882508667e-46 ], [ 708.875 , 1.1986452745120147e-46 ], [ 708.9375 , 1.1988058822902812e-46 ], 
                    [ 709.0 , 1.198966511588558e-46 ], [ 709.0625 , 1.1991271624097198e-46 ], [ 709.125 , 1.1992878347566592e-46 ], [ 709.1875 , 1.19944852863226e-46 ], 
                    [ 709.25 , 1.1996092440394332e-46 ], [ 709.3125 , 1.1997699809810033e-46 ], [ 709.375 , 1.199930739459916e-46 ], [ 709.4375 , 1.2000915194790145e-46 ], 
                    [ 709.5 , 1.200252321041193e-46 ], [ 709.5625 , 1.200413110990955e-46 ], [ 709.625 , 1.200573773669981e-46 ], [ 709.6875 , 1.2007344578520354e-46 ], 
                    [ 709.75 , 1.2008951635399616e-46 ], [ 709.8125 , 1.201055890736664e-46 ], [ 709.875 , 1.201216639445012e-46 ], [ 709.9375 , 1.2013774096678943e-46 ], 
                    [ 710.0 , 1.201538201408164e-46 ], [ 710.0625 , 1.2016990146687265e-46 ], [ 710.125 , 1.2018598494524628e-46 ], [ 710.1875 , 1.2020207057622358e-46 ], 
                    [ 710.25 , 1.2021815836009355e-46 ], [ 710.3125 , 1.202342482971443e-46 ], [ 710.375 , 1.2025034038766404e-46 ], [ 710.4375 , 1.2026643463194096e-46 ], 
                    [ 710.5 , 1.2028253103026249e-46 ], [ 710.5625 , 1.2029862958292032e-46 ], [ 710.625 , 1.2031473029019682e-46 ], [ 710.6875 , 1.2033083315238635e-46 ], 
                    [ 710.75 , 1.2034693816977475e-46 ], [ 710.8125 , 1.203630453426479e-46 ], [ 710.875 , 1.2037915467129941e-46 ], [ 710.9375 , 1.2039526615601356e-46 ], 
                    [ 711.0 , 1.2041137979708233e-46 ], [ 711.0625 , 1.2042749559479174e-46 ], [ 711.125 , 1.2044361354943132e-46 ], [ 711.1875 , 1.2045972534050054e-46 ], 
                    [ 711.25 , 1.2047582909259832e-46 ], [ 711.3125 , 1.2049193499753788e-46 ], [ 711.375 , 1.2050804305560702e-46 ], [ 711.4375 , 1.205241532670953e-46 ], 
                    [ 711.5 , 1.2054026563228971e-46 ], [ 711.5625 , 1.2055638015147736e-46 ], [ 711.625 , 1.2057249682494788e-46 ], [ 711.6875 , 1.2058861565298844e-46 ], 
                    [ 711.75 , 1.2060473663588624e-46 ], [ 711.8125 , 1.2062085977393188e-46 ], [ 711.875 , 1.2063698506741004e-46 ], [ 711.9375 , 1.2065311251661063e-46 ], 
                    [ 712.0 , 1.2066924212182264e-46 ], [ 712.0625 , 1.2068537388333349e-46 ], [ 712.125 , 1.2070150780143142e-46 ], [ 712.1875 , 1.2071764387640475e-46 ], 
                    [ 712.25 , 1.2073378210854267e-46 ], [ 712.3125 , 1.2074992249813013e-46 ], [ 712.375 , 1.2076606504545899e-46 ], [ 712.4375 , 1.2078220975081771e-46 ], 
                    [ 712.5 , 1.2079835661449391e-46 ], [ 712.5625 , 1.2081450563677702e-46 ], [ 712.625 , 1.2083065681795302e-46 ], [ 712.6875 , 1.2084681015831222e-46 ], 
                    [ 712.75 , 1.208629656581459e-46 ], [ 712.8125 , 1.2087911121242355e-46 ], [ 712.875 , 1.208952526375643e-46 ], [ 712.9375 , 1.2091139621812797e-46 ], 
                    [ 713.0 , 1.2092754195440406e-46 ], [ 713.0625 , 1.209436898466762e-46 ], [ 713.125 , 1.2095983989523738e-46 ], [ 713.1875 , 1.2097599210037128e-46 ], 
                    [ 713.25 , 1.209921464623693e-46 ], [ 713.3125 , 1.2100830298151514e-46 ], [ 713.375 , 1.2102446165810119e-46 ], [ 713.4375 , 1.2104062249241378e-46 ], 
                    [ 713.5 , 1.2105678548474107e-46 ], [ 713.5625 , 1.2107295063537034e-46 ], [ 713.625 , 1.210891179445924e-46 ], [ 713.6875 , 1.211052874126929e-46 ], 
                    [ 713.75 , 1.21121459039961e-46 ], [ 713.8125 , 1.2113763282668328e-46 ], [ 713.875 , 1.211538087731524e-46 ], [ 713.9375 , 1.2116998687965332e-46 ], 
                    [ 714.0 , 1.2118616714647365e-46 ], [ 714.0625 , 1.212023495739044e-46 ], [ 714.125 , 1.2121853416223497e-46 ], [ 714.1875 , 1.2123472091175133e-46 ], 
                    [ 714.25 , 1.2125090982274208e-46 ], [ 714.3125 , 1.2126710089549667e-46 ], [ 714.375 , 1.2128329413030465e-46 ], [ 714.4375 , 1.2129947396775798e-46 ], 
                    [ 714.5 , 1.2131565228489173e-46 ], [ 714.5625 , 1.2133183275980763e-46 ], [ 714.625 , 1.2134801539279523e-46 ], [ 714.6875 , 1.2136420018414145e-46 ], 
                    [ 714.75 , 1.2138038713413161e-46 ], [ 714.8125 , 1.2139657624305878e-46 ], [ 714.875 , 1.2141276751120919e-46 ], [ 714.9375 , 1.2142896093886913e-46 ], 
                    [ 715.0 , 1.2144515652632654e-46 ], [ 715.0625 , 1.2146135427387044e-46 ], [ 715.125 , 1.2147755418178973e-46 ], [ 715.1875 , 1.2149375625037087e-46 ], 
                    [ 715.25 , 1.2150996047990287e-46 ], [ 715.3125 , 1.215261668706748e-46 ], [ 715.375 , 1.215423754229741e-46 ], [ 715.4375 , 1.215585861370899e-46 ], 
                    [ 715.5 , 1.215747990133062e-46 ], [ 715.5625 , 1.2159101405191743e-46 ], [ 715.625 , 1.2160723125320941e-46 ], [ 715.6875 , 1.2162345061746885e-46 ], 
                    [ 715.75 , 1.2163967214498858e-46 ], [ 715.8125 , 1.2165589583605276e-46 ], [ 715.875 , 1.2167212169095171e-46 ], [ 715.9375 , 1.2168834970997406e-46 ], 
                    [ 716.0 , 1.2170457989340925e-46 ], [ 716.0625 , 1.2172079595778112e-46 ], [ 716.125 , 1.217370123828973e-46 ], [ 716.1875 , 1.2175323096846878e-46 ], 
                    [ 716.25 , 1.2176945171478597e-46 ], [ 716.3125 , 1.2178567462213417e-46 ], [ 716.375 , 1.2180189969080298e-46 ], [ 716.4375 , 1.2181812692107952e-46 ], 
                    [ 716.5 , 1.2183435631325001e-46 ], [ 716.5625 , 1.218505878676051e-46 ], [ 716.625 , 1.2186682158443195e-46 ], [ 716.6875 , 1.2188305746401954e-46 ], 
                    [ 716.75 , 1.2189929550665514e-46 ], [ 716.8125 , 1.2191553571262606e-46 ], [ 716.875 , 1.2193177808222312e-46 ], [ 716.9375 , 1.2194802261572936e-46 ], 
                    [ 717.0 , 1.2196426931343825e-46 ], [ 717.0625 , 1.2198051817563558e-46 ], [ 717.125 , 1.2199676920261228e-46 ], [ 717.1875 , 1.2201302239465414e-46 ], 
                    [ 717.25 , 1.2202927775205222e-46 ], [ 717.3125 , 1.220455352750924e-46 ], [ 717.375 , 1.2206179496406235e-46 ], [ 717.4375 , 1.220780568192532e-46 ], 
                    [ 717.5 , 1.2209432084095358e-46 ], [ 717.5625 , 1.2211058702945124e-46 ], [ 717.625 , 1.2212685538503398e-46 ], [ 717.6875 , 1.2214310780998645e-46 ], 
                    [ 717.75 , 1.2215936111043584e-46 ], [ 717.8125 , 1.2217561657367345e-46 ], [ 717.875 , 1.2219187419998873e-46 ], [ 717.9375 , 1.2220813398966957e-46 ], 
                    [ 718.0 , 1.2222439594300036e-46 ], [ 718.0625 , 1.2224066006027422e-46 ], [ 718.125 , 1.2225692634177476e-46 ], [ 718.1875 , 1.2227319478779341e-46 ], 
                    [ 718.25 , 1.222894653986148e-46 ], [ 718.3125 , 1.2230573817452866e-46 ], [ 718.375 , 1.2232201311582573e-46 ], [ 718.4375 , 1.2233829022278894e-46 ], 
                    [ 718.5 , 1.2235456949570907e-46 ], [ 718.5625 , 1.2237085093487522e-46 ], [ 718.625 , 1.2238713454057474e-46 ], [ 718.6875 , 1.2240342031309598e-46 ], 
                    [ 718.75 , 1.2241970825272724e-46 ], [ 718.8125 , 1.2243599835975777e-46 ], [ 718.875 , 1.224522906344725e-46 ], [ 718.9375 , 1.22468585077166e-46 ], 
                    [ 719.0 , 1.224848816881206e-46 ], [ 719.0625 , 1.2250118046762749e-46 ], [ 719.125 , 1.2251748141597697e-46 ], [ 719.1875 , 1.2253378453345325e-46 ], 
                    [ 719.25 , 1.2255008982035023e-46 ], [ 719.3125 , 1.2256638029239543e-46 ], [ 719.375 , 1.2258267120549054e-46 ], [ 719.4375 , 1.2259896428389274e-46 ], 
                    [ 719.5 , 1.2261525952789065e-46 ], [ 719.5625 , 1.2263155693776957e-46 ], [ 719.625 , 1.2264785651382079e-46 ], [ 719.6875 , 1.2266415825632876e-46 ], 
                    [ 719.75 , 1.2268046216558407e-46 ], [ 719.8125 , 1.2269676824187297e-46 ], [ 719.875 , 1.2271307648548697e-46 ], [ 719.9375 , 1.2272938689670977e-46 ], 
                    [ 720.0 , 1.2274569947583123e-46 ], [ 720.0625 , 1.2276201422313864e-46 ], [ 720.125 , 1.2277833113892277e-46 ], [ 720.1875 , 1.2279465022346838e-46 ], 
                    [ 720.25 , 1.2281097147706634e-46 ], [ 720.3125 , 1.2282729490000496e-46 ], [ 720.375 , 1.228436204925699e-46 ], [ 720.4375 , 1.228599482550514e-46 ], 
                    [ 720.5 , 1.2287627818774036e-46 ], [ 720.5625 , 1.2289261029092185e-46 ], [ 720.625 , 1.2290894456488254e-46 ], [ 720.6875 , 1.2292528100991624e-46 ], 
                    [ 720.75 , 1.2294161962630889e-46 ], [ 720.8125 , 1.2295796041434821e-46 ], [ 720.875 , 1.2297430337432286e-46 ], [ 720.9375 , 1.2299063283963748e-46 ], 
                    [ 721.0 , 1.2300696110805332e-46 ], [ 721.0625 , 1.230232915442152e-46 ], [ 721.125 , 1.2303962414840916e-46 ], [ 721.1875 , 1.2305595892092477e-46 ], 
                    [ 721.25 , 1.2307229586204815e-46 ], [ 721.3125 , 1.2308863497206898e-46 ], [ 721.375 , 1.2310497625127254e-46 ], [ 721.4375 , 1.231213196999512e-46 ], 
                    [ 721.5 , 1.2313766531838862e-46 ], [ 721.5625 , 1.231540131068746e-46 ], [ 721.625 , 1.2317036306569896e-46 ], [ 721.6875 , 1.2318671519514637e-46 ], 
                    [ 721.75 , 1.2320306949550676e-46 ], [ 721.8125 , 1.2321942596706918e-46 ], [ 721.875 , 1.2323578461012014e-46 ], [ 721.9375 , 1.2325214542495057e-46 ], 
                    [ 722.0 , 1.232685084118444e-46 ], [ 722.0625 , 1.2328487357109438e-46 ], [ 722.125 , 1.2330124090298715e-46 ], [ 722.1875 , 1.233176104078103e-46 ], 
                    [ 722.25 , 1.2333398208585227e-46 ], [ 722.3125 , 1.2335035593740513e-46 ], [ 722.375 , 1.2336673196275215e-46 ], [ 722.4375 , 1.2338311016218806e-46 ], 
                    [ 722.5 , 1.2339949053599536e-46 ], [ 722.5625 , 1.2341586003690333e-46 ], [ 722.625 , 1.2343222555037937e-46 ], [ 722.6875 , 1.2344859323399902e-46 ], 
                    [ 722.75 , 1.2346496308804654e-46 ], [ 722.8125 , 1.2348133511281412e-46 ], [ 722.875 , 1.2349770930858786e-46 ], [ 722.9375 , 1.2351408567565564e-46 ], 
                    [ 723.0 , 1.235304642143054e-46 ], [ 723.0625 , 1.2354684492482418e-46 ], [ 723.125 , 1.235632278075018e-46 ], [ 723.1875 , 1.2357961286262448e-46 ], 
                    [ 723.25 , 1.235960000904821e-46 ], [ 723.3125 , 1.2361238949136186e-46 ], [ 723.375 , 1.2362878106555193e-46 ], [ 723.4375 , 1.236451748133414e-46 ], 
                    [ 723.5 , 1.2366157073501581e-46 ], [ 723.5625 , 1.2367796883086523e-46 ], [ 723.625 , 1.2369436910117884e-46 ], [ 723.6875 , 1.2371077154624232e-46 ], 
                    [ 723.75 , 1.2372717616634849e-46 ], [ 723.8125 , 1.2374358296178222e-46 ], [ 723.875 , 1.2375999193283198e-46 ], [ 723.9375 , 1.2377640307978715e-46 ], 
                    [ 724.0 , 1.23792816402938e-46 ], [ 724.0625 , 1.2380923190256961e-46 ], [ 724.125 , 1.2382564957897323e-46 ], [ 724.1875 , 1.2384206007487886e-46 ], 
                    [ 724.25 , 1.2385846244971676e-46 ], [ 724.3125 , 1.2387486699698135e-46 ], [ 724.375 , 1.2389127371696206e-46 ], [ 724.4375 , 1.2390768260994667e-46 ], 
                    [ 724.5 , 1.239240936762221e-46 ], [ 724.5625 , 1.2394050691607447e-46 ], [ 724.625 , 1.2395692232979425e-46 ], [ 724.6875 , 1.2397333991766853e-46 ], 
                    [ 724.75 , 1.2398975967998432e-46 ], [ 724.8125 , 1.2400618161703146e-46 ], [ 724.875 , 1.2402260572909705e-46 ], [ 724.9375 , 1.2403903201646828e-46 ], 
                    [ 725.0 , 1.2405546047943592e-46 ], [ 725.0625 , 1.2407189111828459e-46 ], [ 725.125 , 1.24088323933306e-46 ], [ 725.1875 , 1.2410475892478575e-46 ], 
                    [ 725.25 , 1.2412119609301296e-46 ], [ 725.3125 , 1.2413763543827506e-46 ], [ 725.375 , 1.2415407696086126e-46 ], [ 725.4375 , 1.241705206610617e-46 ], 
                    [ 725.5 , 1.2418696653916132e-46 ], [ 725.5625 , 1.242034145954529e-46 ], [ 725.625 , 1.2421986483021883e-46 ], [ 725.6875 , 1.2423631724375377e-46 ], 
                    [ 725.75 , 1.2425277183634101e-46 ], [ 725.8125 , 1.2426922454338854e-46 ], [ 725.875 , 1.2428566446152666e-46 ], [ 725.9375 , 1.2430210655454864e-46 ], 
                    [ 726.0 , 1.2431855082273863e-46 ], [ 726.0625 , 1.2433499726638796e-46 ], [ 726.125 , 1.2435144588578175e-46 ], [ 726.1875 , 1.243678966812087e-46 ], 
                    [ 726.25 , 1.2438434965295584e-46 ], [ 726.3125 , 1.2440080480131372e-46 ], [ 726.375 , 1.2441726212656675e-46 ], [ 726.4375 , 1.2443372162900556e-46 ], 
                    [ 726.5 , 1.244501833089173e-46 ], [ 726.5625 , 1.2446664716659006e-46 ], [ 726.625 , 1.2448311320231366e-46 ], [ 726.6875 , 1.2449958141637184e-46 ], 
                    [ 726.75 , 1.2451605180905546e-46 ], [ 726.8125 , 1.2453252438065359e-46 ], [ 726.875 , 1.2454899913145538e-46 ], [ 726.9375 , 1.245654760617447e-46 ], 
                    [ 727.0 , 1.2458195517181339e-46 ], [ 727.0625 , 1.245984364619481e-46 ], [ 727.125 , 1.2461491993243983e-46 ], [ 727.1875 , 1.2463140558357443e-46 ], 
                    [ 727.25 , 1.2464789341564125e-46 ], [ 727.3125 , 1.2466438342892792e-46 ], [ 727.375 , 1.2468087562372566e-46 ], [ 727.4375 , 1.2469737000031951e-46 ], 
                    [ 727.5 , 1.2471384914826513e-46 ], [ 727.5625 , 1.2473032835516522e-46 ], [ 727.625 , 1.2474680973956502e-46 ], [ 727.6875 , 1.2476329330175137e-46 ], 
                    [ 727.75 , 1.2477977904201207e-46 ], [ 727.8125 , 1.2479626696063395e-46 ], [ 727.875 , 1.248127570579067e-46 ], [ 727.9375 , 1.2482924933411642e-46 ], 
                    [ 728.0 , 1.2484574378955364e-46 ], [ 728.0625 , 1.248622404245028e-46 ], [ 728.125 , 1.2487873923925544e-46 ], [ 728.1875 , 1.2489524023409516e-46 ], 
                    [ 728.25 , 1.2491174340931533e-46 ], [ 728.3125 , 1.2492824876519875e-46 ], [ 728.375 , 1.2494475630203888e-46 ], [ 728.4375 , 1.2496126602012036e-46 ], 
                    [ 728.5 , 1.249777779197314e-46 ], [ 728.5625 , 1.2499429200116293e-46 ], [ 728.625 , 1.250108082647006e-46 ], [ 728.6875 , 1.2502732671063181e-46 ], 
                    [ 728.75 , 1.2504384733924855e-46 ], [ 728.8125 , 1.250603701508383e-46 ], [ 728.875 , 1.2507689514568684e-46 ], [ 728.9375 , 1.2509342232408535e-46 ], 
                    [ 729.0 , 1.2510995168632145e-46 ], [ 729.0625 , 1.2512648323268194e-46 ], [ 729.125 , 1.251430068147949e-46 ], [ 729.1875 , 1.251595228238398e-46 ], 
                    [ 729.25 , 1.2517604101261855e-46 ], [ 729.3125 , 1.2519256138142144e-46 ], [ 729.375 , 1.2520908393053263e-46 ], [ 729.4375 , 1.2522560866024347e-46 ], 
                    [ 729.5 , 1.2524213557083813e-46 ], [ 729.5625 , 1.2525866466260806e-46 ], [ 729.625 , 1.252751959358375e-46 ], [ 729.6875 , 1.2529172939081708e-46 ], 
                    [ 729.75 , 1.2530826502783382e-46 ], [ 729.8125 , 1.2532480284717658e-46 ], [ 729.875 , 1.2534134284913073e-46 ], [ 729.9375 , 1.2535788503398789e-46 ], 
                    [ 730.0 , 1.2537442940203167e-46 ], [ 730.0625 , 1.2539097595355469e-46 ], [ 730.125 , 1.2540752468884331e-46 ], [ 730.1875 , 1.2542407560818574e-46 ], 
                    [ 730.25 , 1.2544062871187024e-46 ], [ 730.3125 , 1.2545718400018599e-46 ], [ 730.375 , 1.254737414734204e-46 ], [ 730.4375 , 1.2549030113185826e-46 ], 
                    [ 730.5 , 1.2550686297579512e-46 ], [ 730.5625 , 1.2552342700551493e-46 ], [ 730.625 , 1.2553999322130795e-46 ], [ 730.6875 , 1.2555656162346096e-46 ], 
                    [ 730.75 , 1.2557313085768303e-46 ], [ 730.8125 , 1.2558968380365144e-46 ], [ 730.875 , 1.2560623893161278e-46 ], [ 730.9375 , 1.2562279624186092e-46 ], 
                    [ 731.0 , 1.256393557346764e-46 ], [ 731.0625 , 1.2565591741035317e-46 ], [ 731.125 , 1.2567248126917451e-46 ], [ 731.1875 , 1.2568904731143177e-46 ], 
                    [ 731.25 , 1.257056155374092e-46 ], [ 731.3125 , 1.2572218594739731e-46 ], [ 731.375 , 1.2573875854168046e-46 ], [ 731.4375 , 1.2575533332055279e-46 ], 
                    [ 731.5 , 1.2577191028429786e-46 ], [ 731.5625 , 1.2578848943320453e-46 ], [ 731.625 , 1.2580507076756087e-46 ], [ 731.6875 , 1.2582165428765494e-46 ], 
                    [ 731.75 , 1.25838239993774e-46 ], [ 731.8125 , 1.2585482788620885e-46 ], [ 731.875 , 1.2587141796524415e-46 ], [ 731.9375 , 1.2588801023116902e-46 ], 
                    [ 732.0 , 1.259046046842762e-46 ], [ 732.0625 , 1.2592120132484862e-46 ], [ 732.125 , 1.2593780015317467e-46 ], [ 732.1875 , 1.2595440116954717e-46 ], 
                    [ 732.25 , 1.2597100437425011e-46 ], [ 732.3125 , 1.259876097675737e-46 ], [ 732.375 , 1.2600421734980737e-46 ], [ 732.4375 , 1.2602081596579994e-46 ], 
                    [ 732.5 , 1.2603740824399725e-46 ], [ 732.5625 , 1.2605400270678182e-46 ], [ 732.625 , 1.2607059935444396e-46 ], [ 732.6875 , 1.2608719818727046e-46 ], 
                    [ 732.75 , 1.2610379920554903e-46 ], [ 732.8125 , 1.261204024095674e-46 ], [ 732.875 , 1.2613700779961246e-46 ], [ 732.9375 , 1.2615361537597382e-46 ], 
                    [ 733.0 , 1.2617022513893754e-46 ], [ 733.0625 , 1.2618683708879331e-46 ], [ 733.125 , 1.2620345122582906e-46 ], [ 733.1875 , 1.2622006755033012e-46 ], 
                    [ 733.25 , 1.262366860625871e-46 ], [ 733.3125 , 1.262533067628872e-46 ], [ 733.375 , 1.262699296515167e-46 ], [ 733.4375 , 1.2628655472876733e-46 ], 
                    [ 733.5 , 1.2630318199492632e-46 ], [ 733.5625 , 1.2631981145027832e-46 ], [ 733.625 , 1.2633644309511601e-46 ], [ 733.6875 , 1.2635307692972679e-46 ], 
                    [ 733.75 , 1.2636971295439629e-46 ], [ 733.8125 , 1.2638635116941552e-46 ], [ 733.875 , 1.26402991575072e-46 ], [ 733.9375 , 1.2641963417165324e-46 ], 
                    [ 734.0 , 1.2643627895945042e-46 ], [ 734.0625 , 1.2645292583020306e-46 ], [ 734.125 , 1.264695550453966e-46 ], [ 734.1875 , 1.2648618644741996e-46 ], 
                    [ 734.25 , 1.2650282003655797e-46 ], [ 734.3125 , 1.265194558130983e-46 ], [ 734.375 , 1.2653609377733034e-46 ], [ 734.4375 , 1.2655273392954094e-46 ], 
                    [ 734.5 , 1.2656937627001867e-46 ], [ 734.5625 , 1.2658602079904866e-46 ], [ 734.625 , 1.266026675169214e-46 ], [ 734.6875 , 1.266193164239229e-46 ], 
                    [ 734.75 , 1.266359675203438e-46 ], [ 734.8125 , 1.2665262080646926e-46 ], [ 734.875 , 1.266692762825882e-46 ], [ 734.9375 , 1.2668593394898766e-46 ], 
                    [ 735.0 , 1.2670259380595929e-46 ], [ 735.0625 , 1.2671925585378487e-46 ], [ 735.125 , 1.2673592009275968e-46 ], [ 735.1875 , 1.2675258652316382e-46 ], 
                    [ 735.25 , 1.2676925514529443e-46 ], [ 735.3125 , 1.2678592595943438e-46 ], [ 735.375 , 1.2680259896587186e-46 ], [ 735.4375 , 1.2681927416489434e-46 ], 
                    [ 735.5 , 1.268359515567946e-46 ], [ 735.5625 , 1.2685263114185745e-46 ], [ 735.625 , 1.268693129203713e-46 ], [ 735.6875 , 1.2688599689262727e-46 ], 
                    [ 735.75 , 1.2690267537383758e-46 ], [ 735.8125 , 1.2691934370089331e-46 ], [ 735.875 , 1.2693601421728924e-46 ], [ 735.9375 , 1.26952686923312e-46 ], 
                    [ 736.0 , 1.2696936181925193e-46 ], [ 736.0625 , 1.2698603890539394e-46 ], [ 736.125 , 1.2700271818202752e-46 ], [ 736.1875 , 1.2701939964943858e-46 ], 
                    [ 736.25 , 1.2703608330791578e-46 ], [ 736.3125 , 1.2705276915774691e-46 ], [ 736.375 , 1.270694571992207e-46 ], [ 736.4375 , 1.270861474326223e-46 ], 
                    [ 736.5 , 1.2710283985824324e-46 ], [ 736.5625 , 1.2711953447636873e-46 ], [ 736.625 , 1.2713623128728766e-46 ], [ 736.6875 , 1.2715293029128808e-46 ], 
                    [ 736.75 , 1.27169631488658e-46 ], [ 736.8125 , 1.2718633487968464e-46 ], [ 736.875 , 1.2720304046465702e-46 ], [ 736.9375 , 1.2721974824386692e-46 ], 
                    [ 737.0 , 1.2723645821759625e-46 ], [ 737.0625 , 1.2725317038613502e-46 ], [ 737.125 , 1.2726988474977426e-46 ], [ 737.1875 , 1.2728660130879866e-46 ], 
                    [ 737.25 , 1.2730332006350109e-46 ], [ 737.3125 , 1.2732004101416454e-46 ], [ 737.375 , 1.2733676416108193e-46 ], [ 737.4375 , 1.2735347510076877e-46 ], 
                    [ 737.5 , 1.2737018210658324e-46 ], [ 737.5625 , 1.2738689130412194e-46 ], [ 737.625 , 1.2740360269367601e-46 ], [ 737.6875 , 1.2742031627553122e-46 ], 
                    [ 737.75 , 1.2743703204997697e-46 ], [ 737.8125 , 1.2745375001730088e-46 ], [ 737.875 , 1.2747047017778884e-46 ], [ 737.9375 , 1.2748719253172946e-46 ], 
                    [ 738.0 , 1.275039170794114e-46 ], [ 738.0625 , 1.275206438211197e-46 ], [ 738.125 , 1.2753737275714404e-46 ], [ 738.1875 , 1.2755410388777317e-46 ], 
                    [ 738.25 , 1.275708372132941e-46 ], [ 738.3125 , 1.2758757273399385e-46 ], [ 738.375 , 1.2760431045016218e-46 ], [ 738.4375 , 1.2762105036208716e-46 ], 
                    [ 738.5 , 1.27637792470055e-46 ], [ 738.5625 , 1.276545367743529e-46 ], [ 738.625 , 1.2767128327527441e-46 ], [ 738.6875 , 1.2768803197310227e-46 ], 
                    [ 738.75 , 1.2770478286812556e-46 ], [ 738.8125 , 1.277215359606362e-46 ], [ 738.875 , 1.2773829125091698e-46 ], [ 738.9375 , 1.2775504873925898e-46 ], 
                    [ 739.0 , 1.2777180842595231e-46 ], [ 739.0625 , 1.2778857031128272e-46 ], [ 739.125 , 1.278053153011077e-46 ], [ 739.1875 , 1.2782206166952153e-46 ], 
                    [ 739.25 , 1.2783881023221789e-46 ], [ 739.3125 , 1.2785556098948246e-46 ], [ 739.375 , 1.2787231394160375e-46 ], [ 739.4375 , 1.2788906908887024e-46 ], 
                    [ 739.5 , 1.2790582643156864e-46 ], [ 739.5625 , 1.2792258596998572e-46 ], [ 739.625 , 1.279393477044101e-46 ], [ 739.6875 , 1.2795611163513135e-46 ], 
                    [ 739.75 , 1.2797287776243451e-46 ], [ 739.8125 , 1.2798964608660738e-46 ], [ 739.875 , 1.2800641660793966e-46 ], [ 739.9375 , 1.2802318932671652e-46 ], 
                    [ 740.0 , 1.280399642432304e-46 ], [ 740.0625 , 1.2805674135776572e-46 ], [ 740.125 , 1.2807352067061037e-46 ], [ 740.1875 , 1.2809030218205342e-46 ], 
                    [ 740.25 , 1.2810708589238379e-46 ], [ 740.3125 , 1.2812387180188866e-46 ], [ 740.375 , 1.281406599108563e-46 ], [ 740.4375 , 1.28157450219573e-46 ], 
                    [ 740.5 , 1.2817424272833156e-46 ], [ 740.5625 , 1.281910374374166e-46 ], [ 740.625 , 1.2820783434711649e-46 ], [ 740.6875 , 1.2822463345771952e-46 ], 
                    [ 740.75 , 1.2824143155453581e-46 ], [ 740.8125 , 1.2825821445583158e-46 ], [ 740.875 , 1.2827499955349937e-46 ], [ 740.9375 , 1.282917868478239e-46 ], 
                    [ 741.0 , 1.2830857633909625e-46 ], [ 741.0625 , 1.2832536802760215e-46 ], [ 741.125 , 1.2834216191362824e-46 ], [ 741.1875 , 1.2835895799746484e-46 ], 
                    [ 741.25 , 1.2837575627939501e-46 ], [ 741.3125 , 1.2839255675971097e-46 ], [ 741.375 , 1.2840935943869861e-46 ], [ 741.4375 , 1.2842616431664473e-46 ], 
                    [ 741.5 , 1.2844297139383806e-46 ], [ 741.5625 , 1.2845978067056731e-46 ], [ 741.625 , 1.284765921471194e-46 ], [ 741.6875 , 1.2849340582378043e-46 ], 
                    [ 741.75 , 1.2851022170084193e-46 ], [ 741.8125 , 1.2852703977859009e-46 ], [ 741.875 , 1.2854386005731287e-46 ], [ 741.9375 , 1.2856068253729654e-46 ], 
                    [ 742.0 , 1.2857750721883278e-46 ], [ 742.0625 , 1.285943341022088e-46 ], [ 742.125 , 1.2861116318771003e-46 ], [ 742.1875 , 1.2862799447562738e-46 ], 
                    [ 742.25 , 1.2864482796624638e-46 ], [ 742.3125 , 1.2866166365985892e-46 ], [ 742.375 , 1.2867850155675054e-46 ], [ 742.4375 , 1.2869533566047867e-46 ], 
                    [ 742.5 , 1.2871215742456867e-46 ], [ 742.5625 , 1.28728981387431e-46 ], [ 742.625 , 1.2874580754935218e-46 ], [ 742.6875 , 1.2876263591062243e-46 ], 
                    [ 742.75 , 1.2877946647152642e-46 ], [ 742.8125 , 1.287962992323517e-46 ], [ 742.875 , 1.2881313419338761e-46 ], [ 742.9375 , 1.2882997135491996e-46 ], 
                    [ 743.0 , 1.2884681071723818e-46 ], [ 743.0625 , 1.2886365228062904e-46 ], [ 743.125 , 1.2888049604538021e-46 ], [ 743.1875 , 1.2889734201177946e-46 ], 
                    [ 743.25 , 1.2891419018011454e-46 ], [ 743.3125 , 1.289310405506733e-46 ], [ 743.375 , 1.2894789312374356e-46 ], [ 743.4375 , 1.2896474789961322e-46 ], 
                    [ 743.5 , 1.289816048785702e-46 ], [ 743.5625 , 1.2899846406090159e-46 ], [ 743.625 , 1.2901532544689716e-46 ], [ 743.6875 , 1.2903218903684502e-46 ], 
                    [ 743.75 , 1.2904905483103046e-46 ], [ 743.8125 , 1.2906592282974433e-46 ], [ 743.875 , 1.290827930332748e-46 ], [ 743.9375 , 1.2909966544190734e-46 ], 
                    [ 744.0 , 1.2911654005593289e-46 ], [ 744.0625 , 1.291334168756379e-46 ], [ 744.125 , 1.2915028810279535e-46 ], [ 744.1875 , 1.2916714809314303e-46 ], 
                    [ 744.25 , 1.2918401028448771e-46 ], [ 744.3125 , 1.2920087467711483e-46 ], [ 744.375 , 1.2921774127131363e-46 ], [ 744.4375 , 1.2923461006737058e-46 ], 
                    [ 744.5 , 1.292514810655713e-46 ], [ 744.5625 , 1.2926835426620602e-46 ], [ 744.625 , 1.2928522966956134e-46 ], [ 744.6875 , 1.2930210727592573e-46 ], 
                    [ 744.75 , 1.2931898708558497e-46 ], [ 744.8125 , 1.2933586909882755e-46 ], [ 744.875 , 1.293527533159421e-46 ], [ 744.9375 , 1.293696397372145e-46 ], 
                    [ 745.0 , 1.2938652836293244e-46 ], [ 745.0625 , 1.2940341919338654e-46 ], [ 745.125 , 1.2942031222886178e-46 ], [ 745.1875 , 1.2943720746964698e-46 ], 
                    [ 745.25 , 1.2945410491603093e-46 ], [ 745.3125 , 1.2947100456829882e-46 ], [ 745.375 , 1.2948790642674413e-46 ], [ 745.4375 , 1.2950481049164658e-46 ], 
                    [ 745.5 , 1.2952171676330065e-46 ], [ 745.5625 , 1.2953862524199352e-46 ], [ 745.625 , 1.2955553592801147e-46 ], [ 745.6875 , 1.2957244882164447e-46 ], 
                    [ 745.75 , 1.295893639231789e-46 ], [ 745.8125 , 1.2960627374701968e-46 ], [ 745.875 , 1.2962317363635798e-46 ], [ 745.9375 , 1.2964007572934343e-46 ], 
                    [ 746.0 , 1.2965698002626067e-46 ], [ 746.0625 , 1.2967388652739887e-46 ], [ 746.125 , 1.296907952330436e-46 ], [ 746.1875 , 1.2970770614348418e-46 ], 
                    [ 746.25 , 1.2972461925900717e-46 ], [ 746.3125 , 1.2974153457989918e-46 ], [ 746.375 , 1.2975845210644779e-46 ], [ 746.4375 , 1.2977537183894337e-46 ], 
                    [ 746.5 , 1.2979229377767171e-46 ], [ 746.5625 , 1.2980921792292044e-46 ], [ 746.625 , 1.2982614427497553e-46 ], [ 746.6875 , 1.2984307283412744e-46 ], 
                    [ 746.75 , 1.2986000360066306e-46 ], [ 746.8125 , 1.2987693657487111e-46 ], [ 746.875 , 1.2989387175703857e-46 ], [ 746.9375 , 1.2991080914745243e-46 ], 
                    [ 747.0 , 1.2992774874640152e-46 ], [ 747.0625 , 1.2994469055417476e-46 ], [ 747.125 , 1.2996163457105925e-46 ], [ 747.1875 , 1.2997858079734117e-46 ], 
                    [ 747.25 , 1.2999552923331235e-46 ], [ 747.3125 , 1.3001247987925904e-46 ], [ 747.375 , 1.3002943273546757e-46 ], [ 747.4375 , 1.3004638780223078e-46 ], 
                    [ 747.5 , 1.3006333790690874e-46 ], [ 747.5625 , 1.300802759417789e-46 ], [ 747.625 , 1.3009721618247175e-46 ], [ 747.6875 , 1.301141586292792e-46 ], 
                    [ 747.75 , 1.3013110328248858e-46 ], [ 747.8125 , 1.3014805014238345e-46 ], [ 747.875 , 1.3016499920925125e-46 ], [ 747.9375 , 1.3018195048338399e-46 ], 
                    [ 748.0 , 1.3019890396506635e-46 ], [ 748.0625 , 1.3021585965458398e-46 ], [ 748.125 , 1.3023281755222623e-46 ], [ 748.1875 , 1.3024977765828346e-46 ], 
                    [ 748.25 , 1.3026673997303863e-46 ], [ 748.3125 , 1.302837044967803e-46 ], [ 748.375 , 1.3030067122979987e-46 ], [ 748.4375 , 1.303176401723804e-46 ], 
                    [ 748.5 , 1.3033461132481055e-46 ], [ 748.5625 , 1.3035158468737909e-46 ], [ 748.625 , 1.3036856026037657e-46 ], [ 748.6875 , 1.303855380440853e-46 ], 
                    [ 748.75 , 1.304025180387969e-46 ], [ 748.8125 , 1.304195002447984e-46 ], [ 748.875 , 1.3043648466237867e-46 ], [ 748.9375 , 1.3045347129182294e-46 ], 
                    [ 749.0 , 1.3047046013342205e-46 ], [ 749.0625 , 1.3048745118746317e-46 ], [ 749.125 , 1.3050444445423532e-46 ], [ 749.1875 , 1.3052143511836744e-46 ], 
                    [ 749.25 , 1.3053841150738331e-46 ], [ 749.3125 , 1.3055539010444861e-46 ], [ 749.375 , 1.3057237090985059e-46 ], [ 749.4375 , 1.3058935392387828e-46 ], 
                    [ 749.5 , 1.3060633914681526e-46 ], [ 749.5625 , 1.3062332657895071e-46 ], [ 749.625 , 1.3064031622057285e-46 ], [ 749.6875 , 1.3065730807196725e-46 ], 
                    [ 749.75 , 1.3067430213342409e-46 ], [ 749.8125 , 1.3069129840522711e-46 ], [ 749.875 , 1.3070829688766567e-46 ], [ 749.9375 , 1.3072529758102822e-46 ], 
                    [ 750.0 , 1.3074230048560139e-46 ], [ 750.0625 , 1.3075930560167185e-46 ], [ 750.125 , 1.307763129295291e-46 ], [ 750.1875 , 1.3079332246945899e-46 ], 
                    [ 750.25 , 1.3081033422175105e-46 ], [ 750.3125 , 1.3082734818669214e-46 ], [ 750.375 , 1.3084436436456816e-46 ], [ 750.4375 , 1.3086138275567161e-46 ], 
                    [ 750.5 , 1.3087840336028571e-46 ], [ 750.5625 , 1.3089542617870113e-46 ], [ 750.625 , 1.3091245121120395e-46 ], [ 750.6875 , 1.3092947845808497e-46 ], 
                    [ 750.75 , 1.309465079196275e-46 ], [ 750.8125 , 1.3096353959612336e-46 ], [ 750.875 , 1.3098057232933569e-46 ], [ 750.9375 , 1.3099758786336887e-46 ], 
                    [ 751.0 , 1.3101460560786845e-46 ], [ 751.0625 , 1.3103162556312533e-46 ], [ 751.125 , 1.310486477294239e-46 ], [ 751.1875 , 1.3106567210705324e-46 ], 
                    [ 751.25 , 1.3108269869629786e-46 ], [ 751.3125 , 1.3109972749744784e-46 ], [ 751.375 , 1.3111675851079147e-46 ], [ 751.4375 , 1.3113379173661052e-46 ], 
                    [ 751.5 , 1.3115082717519897e-46 ], [ 751.5625 , 1.311678648268405e-46 ], [ 751.625 , 1.311849046918236e-46 ], [ 751.6875 , 1.3120194677043576e-46 ], 
                    [ 751.75 , 1.3121899106296549e-46 ], [ 751.8125 , 1.3123603756969854e-46 ], [ 751.875 , 1.312530862909235e-46 ], [ 751.9375 , 1.3127013722692706e-46 ], 
                    [ 752.0 , 1.3128719037800073e-46 ], [ 752.0625 , 1.3130424574442759e-46 ], [ 752.125 , 1.3132130332649822e-46 ], [ 752.1875 , 1.3133836312449765e-46 ], 
                    [ 752.25 , 1.3135542513871659e-46 ], [ 752.3125 , 1.3137248936944197e-46 ], [ 752.375 , 1.3138955581696267e-46 ], [ 752.4375 , 1.3140662448156481e-46 ], 
                    [ 752.5 , 1.3142369536353828e-46 ], [ 752.5625 , 1.3144076846316834e-46 ], [ 752.625 , 1.3145782591832713e-46 ], [ 752.6875 , 1.3147488179654325e-46 ], 
                    [ 752.75 , 1.3149193988765906e-46 ], [ 752.8125 , 1.315090001919607e-46 ], [ 752.875 , 1.3152606270973724e-46 ], [ 752.9375 , 1.3154312744127492e-46 ], 
                    [ 753.0 , 1.3156019438686096e-46 ], [ 753.0625 , 1.3157726354678071e-46 ], [ 753.125 , 1.3159433492132618e-46 ], [ 753.1875 , 1.3161140851078004e-46 ], 
                    [ 753.25 , 1.3162848431543244e-46 ], [ 753.3125 , 1.3164556233556988e-46 ], [ 753.375 , 1.3166264257147884e-46 ], [ 753.4375 , 1.3167972502344868e-46 ], 
                    [ 753.5 , 1.316968096917651e-46 ], [ 753.5625 , 1.3171389657671745e-46 ], [ 753.625 , 1.3173098567859243e-46 ], [ 753.6875 , 1.3174807699767673e-46 ], 
                    [ 753.75 , 1.317651705342599e-46 ], [ 753.8125 , 1.317822662886287e-46 ], [ 753.875 , 1.3179936426106997e-46 ], [ 753.9375 , 1.3181646445187334e-46 ], 
                    [ 754.0 , 1.3183356686132473e-46 ], [ 754.0625 , 1.3185067148971486e-46 ], [ 754.125 , 1.3186777833732876e-46 ], [ 754.1875 , 1.3188488740445535e-46 ], 
                    [ 754.25 , 1.3190199869138067e-46 ], [ 754.3125 , 1.3191910101466705e-46 ], [ 754.375 , 1.3193619571583705e-46 ], [ 754.4375 , 1.3195329263221833e-46 ], 
                    [ 754.5 , 1.3197039176409896e-46 ], [ 754.5625 , 1.31987493111765e-46 ], [ 754.625 , 1.3200459667550554e-46 ], [ 754.6875 , 1.3202170245560678e-46 ], 
                    [ 754.75 , 1.3203881045235497e-46 ], [ 754.8125 , 1.320559206660402e-46 ], [ 754.875 , 1.320730330969469e-46 ], [ 754.9375 , 1.3209014774536245e-46 ], 
                    [ 755.0 , 1.3210726461157697e-46 ], [ 755.0625 , 1.3212438369587408e-46 ], [ 755.125 , 1.321415049985441e-46 ], [ 755.1875 , 1.3215862851987441e-46 ], 
                    [ 755.25 , 1.3217575426015165e-46 ], [ 755.3125 , 1.321928822196605e-46 ], [ 755.375 , 1.322100123986942e-46 ], [ 755.4375 , 1.3222714479753564e-46 ], 
                    [ 755.5 , 1.3224427941647343e-46 ], [ 755.5625 , 1.3226141625579904e-46 ], [ 755.625 , 1.3227855531579732e-46 ], [ 755.6875 , 1.3229569659675516e-46 ], 
                    [ 755.75 , 1.3231284009896128e-46 ], [ 755.8125 , 1.3232998582270256e-46 ], [ 755.875 , 1.3234713376826877e-46 ], [ 755.9375 , 1.323642839359469e-46 ], 
                    [ 756.0 , 1.3238143245950657e-46 ], [ 756.0625 , 1.3239856523900984e-46 ], [ 756.125 , 1.3241570023583416e-46 ], [ 756.1875 , 1.3243283745026647e-46 ], 
                    [ 756.25 , 1.3244997688259473e-46 ], [ 756.3125 , 1.3246711853310405e-46 ], [ 756.375 , 1.3248426240208156e-46 ], [ 756.4375 , 1.3250140848981717e-46 ], 
                    [ 756.5 , 1.325185567965952e-46 ], [ 756.5625 , 1.325357073227038e-46 ], [ 756.625 , 1.3255286006842923e-46 ], [ 756.6875 , 1.3257001503406161e-46 ], 
                    [ 756.75 , 1.3258717221988543e-46 ], [ 756.8125 , 1.3260433162618795e-46 ], [ 756.875 , 1.3262149325325942e-46 ], [ 756.9375 , 1.3263865710138348e-46 ], 
                    [ 757.0 , 1.3265582317085135e-46 ], [ 757.0625 , 1.326729914619477e-46 ], [ 757.125 , 1.3269016197495906e-46 ], [ 757.1875 , 1.327073347101787e-46 ], 
                    [ 757.25 , 1.3272450966789042e-46 ], [ 757.3125 , 1.3274168684837903e-46 ], [ 757.375 , 1.3275886625193505e-46 ], [ 757.4375 , 1.3277604787884709e-46 ], 
                    [ 757.5 , 1.3279323172940199e-46 ], [ 757.5625 , 1.328104178038894e-46 ], [ 757.625 , 1.3282760610259146e-46 ], [ 757.6875 , 1.328447966258046e-46 ], 
                    [ 757.75 , 1.3286197312702882e-46 ], [ 757.8125 , 1.328791467224072e-46 ], [ 757.875 , 1.328963225376279e-46 ], [ 757.9375 , 1.32913500572975e-46 ], 
                    [ 758.0 , 1.3293068082873642e-46 ], [ 758.0625 , 1.3294786330520011e-46 ], [ 758.125 , 1.3296504800265125e-46 ], [ 758.1875 , 1.3298223492137879e-46 ], 
                    [ 758.25 , 1.3299942406166701e-46 ], [ 758.3125 , 1.3301661542380595e-46 ], [ 758.375 , 1.3303380900808275e-46 ], [ 758.4375 , 1.3305100481478369e-46 ], 
                    [ 758.5 , 1.3306820284419418e-46 ], [ 758.5625 , 1.3308540309660531e-46 ], [ 758.625 , 1.3310260557230157e-46 ], [ 758.6875 , 1.3311981027157223e-46 ], 
                    [ 758.75 , 1.3313701719470189e-46 ], [ 758.8125 , 1.3315422634198271e-46 ], [ 758.875 , 1.3317143771369556e-46 ], [ 758.9375 , 1.3318865131013364e-46 ], 
                    [ 759.0 , 1.3320586713158454e-46 ], [ 759.0625 , 1.3322308517833112e-46 ], [ 759.125 , 1.3324030545066574e-46 ], [ 759.1875 , 1.3325752794887327e-46 ], 
                    [ 759.25 , 1.3327475267324232e-46 ], [ 759.3125 , 1.3329197962406259e-46 ], [ 759.375 , 1.3330920880161707e-46 ], [ 759.4375 , 1.3332643418027682e-46 ], 
                    [ 759.5 , 1.3334364637321964e-46 ], [ 759.5625 , 1.3336086078822343e-46 ], [ 759.625 , 1.3337807742557497e-46 ], [ 759.6875 , 1.3339529628556409e-46 ], 
                    [ 759.75 , 1.334125173684748e-46 ], [ 759.8125 , 1.3342974067459508e-46 ], [ 759.875 , 1.3344696620421285e-46 ], [ 759.9375 , 1.3346419395761234e-46 ], 
                    [ 760.0 , 1.3348142393508445e-46 ], [ 760.0625 , 1.3349865613691244e-46 ], [ 760.125 , 1.335158905633864e-46 ], [ 760.1875 , 1.3353312721479346e-46 ], 
                    [ 760.25 , 1.3355036609141903e-46 ], [ 760.3125 , 1.335676071935503e-46 ], [ 760.375 , 1.335848505214784e-46 ], [ 760.4375 , 1.3360209607548407e-46 ], 
                    [ 760.5 , 1.3361934385586223e-46 ], [ 760.5625 , 1.3363659386289371e-46 ], [ 760.625 , 1.3365384609687067e-46 ], [ 760.6875 , 1.336711005580759e-46 ], 
                    [ 760.75 , 1.3368835724680259e-46 ], [ 760.8125 , 1.3370561616333547e-46 ], [ 760.875 , 1.337228773079593e-46 ], [ 760.9375 , 1.3374014068096839e-46 ], 
                    [ 761.0 , 1.3375740628264468e-46 ], [ 761.0625 , 1.3377467411327877e-46 ], [ 761.125 , 1.3379194417315555e-46 ], [ 761.1875 , 1.3380920011951004e-46 ], 
                    [ 761.25 , 1.3382645253901663e-46 ], [ 761.3125 , 1.3384370718292916e-46 ], [ 761.375 , 1.3386096405153539e-46 ], [ 761.4375 , 1.3387822314512023e-46 ], 
                    [ 761.5 , 1.3389548446397152e-46 ], [ 761.5625 , 1.339127480083771e-46 ], [ 761.625 , 1.339300137786211e-46 ], [ 761.6875 , 1.339472817749933e-46 ], 
                    [ 761.75 , 1.3396455199777983e-46 ], [ 761.8125 , 1.3398182444726867e-46 ], [ 761.875 , 1.339990991237441e-46 ], [ 761.9375 , 1.3401637602749697e-46 ], 
                    [ 762.0 , 1.3403365515881167e-46 ], [ 762.0625 , 1.3405093651797635e-46 ], [ 762.125 , 1.3406822010527824e-46 ], [ 762.1875 , 1.3408550592100556e-46 ], 
                    [ 762.25 , 1.341027939654447e-46 ], [ 762.3125 , 1.3412008423888206e-46 ], [ 762.375 , 1.3413737674160594e-46 ], [ 762.4375 , 1.3415467147390288e-46 ], 
                    [ 762.5 , 1.3417196843606318e-46 ], [ 762.5625 , 1.341892676283705e-46 ], [ 762.625 , 1.342065690511153e-46 ], [ 762.6875 , 1.3422387270458604e-46 ], 
                    [ 762.75 , 1.3424117858906657e-46 ], [ 762.8125 , 1.3425848670484547e-46 ], [ 762.875 , 1.3427579407573978e-46 ], [ 762.9375 , 1.3429308476594005e-46 ], 
                    [ 763.0 , 1.343103776826639e-46 ], [ 763.0625 , 1.343276728261952e-46 ], [ 763.125 , 1.3434497019682069e-46 ], [ 763.1875 , 1.3436226979482906e-46 ], 
                    [ 763.25 , 1.3437957162050712e-46 ], [ 763.3125 , 1.3439687567413984e-46 ], [ 763.375 , 1.3441418195601506e-46 ], [ 763.4375 , 1.3443149046641972e-46 ], 
                    [ 763.5 , 1.344488012056417e-46 ], [ 763.5625 , 1.3446611417396517e-46 ], [ 763.625 , 1.3448342937168005e-46 ], [ 763.6875 , 1.3450074679907245e-46 ], 
                    [ 763.75 , 1.3451806645642854e-46 ], [ 763.8125 , 1.345353883440402e-46 ], [ 763.875 , 1.3455271246218608e-46 ], [ 763.9375 , 1.3457003881116008e-46 ], 
                    [ 764.0 , 1.345873673912485e-46 ], [ 764.0625 , 1.346046982027358e-46 ], [ 764.125 , 1.3462203124591306e-46 ], [ 764.1875 , 1.3463936652106586e-46 ], 
                    [ 764.25 , 1.3465670402848056e-46 ], [ 764.3125 , 1.346740437684456e-46 ], [ 764.375 , 1.3469138574124654e-46 ], [ 764.4375 , 1.3470872994717477e-46 ], 
                    [ 764.5 , 1.347260763865159e-46 ], [ 764.5625 , 1.3474342505956038e-46 ], [ 764.625 , 1.3476076567915674e-46 ], [ 764.6875 , 1.347780969095194e-46 ], 
                    [ 764.75 , 1.347954303688064e-46 ], [ 764.8125 , 1.3481276605730435e-46 ], [ 764.875 , 1.3483010397529995e-46 ], [ 764.9375 , 1.3484744412308187e-46 ], 
                    [ 765.0 , 1.3486478650093106e-46 ], [ 765.0625 , 1.3488213110914207e-46 ], [ 765.125 , 1.3489947794799595e-46 ], [ 765.1875 , 1.349168270177815e-46 ], 
                    [ 765.25 , 1.3493417831878564e-46 ], [ 765.3125 , 1.3495153185129435e-46 ], [ 765.375 , 1.3496888761559848e-46 ], [ 765.4375 , 1.3498624561197928e-46 ], 
                    [ 765.5 , 1.3500360584072764e-46 ], [ 765.5625 , 1.3502096830213066e-46 ], [ 765.625 , 1.3503833299647455e-46 ], [ 765.6875 , 1.3505569992404454e-46 ], 
                    [ 765.75 , 1.3507306908513262e-46 ], [ 765.8125 , 1.3509044048001936e-46 ], [ 765.875 , 1.3510781410899968e-46 ], [ 765.9375 , 1.3512518997235614e-46 ], 
                    [ 766.0 , 1.3514256807037604e-46 ], [ 766.0625 , 1.351599484033478e-46 ], [ 766.125 , 1.3517733097155787e-46 ], [ 766.1875 , 1.3519471577529561e-46 ], 
                    [ 766.25 , 1.3521210281484855e-46 ], [ 766.3125 , 1.3522949209050326e-46 ], [ 766.375 , 1.3524686728663779e-46 ], [ 766.4375 , 1.3526423899619497e-46 ], 
                    [ 766.5 , 1.3528161293705278e-46 ], [ 766.5625 , 1.3529898910949688e-46 ], [ 766.625 , 1.3531636751381295e-46 ], [ 766.6875 , 1.3533374815028764e-46 ], 
                    [ 766.75 , 1.3535113101921057e-46 ], [ 766.8125 , 1.3536851612086656e-46 ], [ 766.875 , 1.3538590345554141e-46 ], [ 766.9375 , 1.3540329302352386e-46 ], 
                    [ 767.0 , 1.3542068482509885e-46 ], [ 767.0625 , 1.3543807886055425e-46 ], [ 767.125 , 1.354554751301789e-46 ], [ 767.1875 , 1.354728736342559e-46 ], 
                    [ 767.25 , 1.354902743730752e-46 ], [ 767.3125 , 1.3550767734692183e-46 ], [ 767.375 , 1.3552508255608388e-46 ], [ 767.4375 , 1.3554249000084944e-46 ], 
                    [ 767.5 , 1.3555989968150372e-46 ], [ 767.5625 , 1.3557731159833482e-46 ], [ 767.625 , 1.3559472575163005e-46 ], [ 767.6875 , 1.356121421416776e-46 ], 
                    [ 767.75 , 1.3562956076876282e-46 ], [ 767.8125 , 1.3564698163317597e-46 ], [ 767.875 , 1.3566440473520149e-46 ], [ 767.9375 , 1.3568183007512585e-46 ], 
                    [ 768.0 , 1.3569925765324036e-46 ], [ 768.0625 , 1.3571668746983058e-46 ], [ 768.125 , 1.3573409866127065e-46 ], [ 768.1875 , 1.357515109566755e-46 ], 
                    [ 768.25 , 1.357689254857694e-46 ], [ 768.3125 , 1.3578634224884276e-46 ], [ 768.375 , 1.3580376124617924e-46 ], [ 768.4375 , 1.358211824780674e-46 ], 
                    [ 768.5 , 1.35838605944791e-46 ], [ 768.5625 , 1.3585603164663962e-46 ], [ 768.625 , 1.3587345958389902e-46 ], [ 768.6875 , 1.3589088975685793e-46 ], 
                    [ 768.75 , 1.3590832216579827e-46 ], [ 768.8125 , 1.3592575681101175e-46 ], [ 768.875 , 1.3594319369278327e-46 ], [ 768.9375 , 1.3596063281139881e-46 ], 
                    [ 769.0 , 1.3597807416714435e-46 ], [ 769.0625 , 1.3599551776031263e-46 ], [ 769.125 , 1.3601296359118491e-46 ], [ 769.1875 , 1.360304116600502e-46 ], 
                    [ 769.25 , 1.3604786196719654e-46 ], [ 769.3125 , 1.3606531451290813e-46 ], [ 769.375 , 1.3608276929747605e-46 ], [ 769.4375 , 1.3610022632118651e-46 ], 
                    [ 769.5 , 1.3611768558432195e-46 ], [ 769.5625 , 1.3613514708717638e-46 ], [ 769.625 , 1.3615261083003232e-46 ], [ 769.6875 , 1.3617007681318099e-46 ], 
                    [ 769.75 , 1.3618754503690778e-46 ], [ 769.8125 , 1.3620501333936041e-46 ], [ 769.875 , 1.362224629815203e-46 ], [ 769.9375 , 1.3623991485920933e-46 ], 
                    [ 770.0 , 1.3625736897270906e-46 ], [ 770.0625 , 1.362748253223117e-46 ], [ 770.125 , 1.3629228390829895e-46 ], [ 770.1875 , 1.3630974473096015e-46 ], 
                    [ 770.25 , 1.3632720779058094e-46 ], [ 770.3125 , 1.3634467308744786e-46 ], [ 770.375 , 1.3636214062184658e-46 ], [ 770.4375 , 1.363796103940657e-46 ], 
                    [ 770.5 , 1.3639708240438999e-46 ], [ 770.5625 , 1.364145566531071e-46 ], [ 770.625 , 1.364320331405048e-46 ], [ 770.6875 , 1.364495118668699e-46 ], 
                    [ 770.75 , 1.3646699283248438e-46 ], [ 770.8125 , 1.3648447603764385e-46 ], [ 770.875 , 1.3650196148262747e-46 ], [ 770.9375 , 1.3651944916772508e-46 ], 
                    [ 771.0 , 1.3653693909322371e-46 ], [ 771.0625 , 1.3655443125941227e-46 ], [ 771.125 , 1.3657192566657496e-46 ], [ 771.1875 , 1.3658942231499887e-46 ], 
                    [ 771.25 , 1.3660692120497204e-46 ], [ 771.3125 , 1.366244223367827e-46 ], [ 771.375 , 1.3664192571071608e-46 ], [ 771.4375 , 1.3665943132706333e-46 ], 
                    [ 771.5 , 1.366769391861049e-46 ], [ 771.5625 , 1.3669444538173141e-46 ], [ 771.625 , 1.3671193548779604e-46 ], [ 771.6875 , 1.3672942783172826e-46 ], 
                    [ 771.75 , 1.3674692241380864e-46 ], [ 771.8125 , 1.367644192343294e-46 ], [ 771.875 , 1.3678191829357396e-46 ], [ 771.9375 , 1.3679941959183174e-46 ], 
                    [ 772.0 , 1.3681692312938338e-46 ], [ 772.0625 , 1.3683442890652313e-46 ], [ 772.125 , 1.3685193692353276e-46 ], [ 772.1875 , 1.3686944718069784e-46 ], 
                    [ 772.25 , 1.3688695967830892e-46 ], [ 772.3125 , 1.3690447441665069e-46 ], [ 772.375 , 1.3692199139600986e-46 ], [ 772.4375 , 1.3693951061667222e-46 ], 
                    [ 772.5 , 1.3695703207892644e-46 ], [ 772.5625 , 1.3697455578305843e-46 ], [ 772.625 , 1.36992081729354e-46 ], [ 772.6875 , 1.3700960991810104e-46 ], 
                    [ 772.75 , 1.370271403495894e-46 ], [ 772.8125 , 1.370446730240992e-46 ], [ 772.875 , 1.3706220794192235e-46 ], [ 772.9375 , 1.3707974510334485e-46 ], 
                    [ 773.0 , 1.3709728450865477e-46 ], [ 773.0625 , 1.3711482615813827e-46 ], [ 773.125 , 1.3713237005208243e-46 ], [ 773.1875 , 1.371499161907735e-46 ], 
                    [ 773.25 , 1.3716746457450065e-46 ], [ 773.3125 , 1.3718501086793314e-46 ], [ 773.375 , 1.37202541436632e-46 ], [ 773.4375 , 1.3722007424552144e-46 ], 
                    [ 773.5 , 1.3723760929489164e-46 ], [ 773.5625 , 1.37255146585027e-46 ], [ 773.625 , 1.3727268611621378e-46 ], [ 773.6875 , 1.372902278887384e-46 ], 
                    [ 773.75 , 1.3730777190288732e-46 ], [ 773.8125 , 1.3732531815894594e-46 ], [ 773.875 , 1.3734286665720179e-46 ], [ 773.9375 , 1.3736041739794134e-46 ], 
                    [ 774.0 , 1.3737797038145412e-46 ], [ 774.0625 , 1.373955256080199e-46 ], [ 774.125 , 1.374130830779311e-46 ], [ 774.1875 , 1.3743064279147155e-46 ], 
                    [ 774.25 , 1.3744820474892987e-46 ], [ 774.3125 , 1.3746576895059087e-46 ], [ 774.375 , 1.3748333539674426e-46 ], [ 774.4375 , 1.3750090408767391e-46 ], 
                    [ 774.5 , 1.3751847502366568e-46 ], [ 774.5625 , 1.3753604820501141e-46 ], [ 774.625 , 1.3755362363199312e-46 ], [ 774.6875 , 1.3757120130490065e-46 ], 
                    [ 774.75 , 1.3758878122402109e-46 ], [ 774.8125 , 1.3760636338963847e-46 ], [ 774.875 , 1.3762394780204187e-46 ], [ 774.9375 , 1.376415344615184e-46 ], 
                    [ 775.0 , 1.3765912336835614e-46 ], [ 775.0625 , 1.376767110572921e-46 ], [ 775.125 , 1.3769428129087025e-46 ], [ 775.1875 , 1.3771185376675387e-46 ], 
                    [ 775.25 , 1.3772942848522814e-46 ], [ 775.3125 , 1.3774700544658123e-46 ], [ 775.375 , 1.3776458465109349e-46 ], [ 775.4375 , 1.3778216609905805e-46 ], 
                    [ 775.5 , 1.3779974979075729e-46 ], [ 775.5625 , 1.3781733572647951e-46 ], [ 775.625 , 1.3783492390650818e-46 ], [ 775.6875 , 1.3785251433113266e-46 ], 
                    [ 775.75 , 1.378701070006384e-46 ], [ 775.8125 , 1.3788770191531188e-46 ], [ 775.875 , 1.3790529907544064e-46 ], [ 775.9375 , 1.3792289848131024e-46 ], 
                    [ 776.0 , 1.3794050013320534e-46 ], [ 776.0625 , 1.3795810403141549e-46 ], [ 776.125 , 1.379757101762264e-46 ], [ 776.1875 , 1.3799331856792576e-46 ], 
                    [ 776.25 , 1.3801092920679835e-46 ], [ 776.3125 , 1.3802854209312998e-46 ], [ 776.375 , 1.3804615722721334e-46 ], [ 776.4375 , 1.3806377460932943e-46 ], 
                    [ 776.5 , 1.380813942397671e-46 ], [ 776.5625 , 1.3809901611881325e-46 ], [ 776.625 , 1.3811664024675584e-46 ], [ 776.6875 , 1.3813426662387993e-46 ], 
                    [ 776.75 , 1.3815189525047251e-46 ], [ 776.8125 , 1.3816952519598389e-46 ], [ 776.875 , 1.3818713576997234e-46 ], [ 776.9375 , 1.3820474858853708e-46 ], 
                    [ 777.0 , 1.3822236365196818e-46 ], [ 777.0625 , 1.3823998096054977e-46 ], [ 777.125 , 1.3825760051456702e-46 ], [ 777.1875 , 1.3827522231430812e-46 ], 
                    [ 777.25 , 1.382928463600573e-46 ], [ 777.3125 , 1.3831047265210281e-46 ], [ 777.375 , 1.3832810119072998e-46 ], [ 777.4375 , 1.383457319762242e-46 ], 
                    [ 777.5 , 1.383633650088747e-46 ], [ 777.5625 , 1.3838100028896403e-46 ], [ 777.625 , 1.383986378167816e-46 ], [ 777.6875 , 1.3841627759261088e-46 ], 
                    [ 777.75 , 1.3843391961674437e-46 ], [ 777.8125 , 1.3845156388946074e-46 ], [ 777.875 , 1.384692104110515e-46 ], [ 777.9375 , 1.3848685918180423e-46 ], 
                    [ 778.0 , 1.3850451020200173e-46 ], [ 778.0625 , 1.3852216347193457e-46 ], [ 778.125 , 1.3853981899188563e-46 ], [ 778.1875 , 1.3855747676214362e-46 ], 
                    [ 778.25 , 1.3857513678299633e-46 ], [ 778.3125 , 1.3859279905472963e-46 ], [ 778.375 , 1.3861046357762848e-46 ], [ 778.4375 , 1.386281303519837e-46 ], 
                    [ 778.5 , 1.3864579937808028e-46 ], [ 778.5625 , 1.3866347065620526e-46 ], [ 778.625 , 1.386811242919626e-46 ], [ 778.6875 , 1.3869877718316311e-46 ], 
                    [ 778.75 , 1.3871643232142286e-46 ], [ 778.8125 , 1.3873408970702694e-46 ], [ 778.875 , 1.3875174934026139e-46 ], [ 778.9375 , 1.387694112214123e-46 ], 
                    [ 779.0 , 1.3878707535076582e-46 ], [ 779.0625 , 1.3880474172860815e-46 ], [ 779.125 , 1.388224103552255e-46 ], [ 779.1875 , 1.388400812309021e-46 ], 
                    [ 779.25 , 1.3885775435592924e-46 ], [ 779.3125 , 1.3887542973058823e-46 ], [ 779.375 , 1.3889310735516943e-46 ], [ 779.4375 , 1.3891078722995625e-46 ], 
                    [ 779.5 , 1.3892846935523512e-46 ], [ 779.5625 , 1.3894615373129552e-46 ], [ 779.625 , 1.3896384035841992e-46 ], [ 779.6875 , 1.3898152923689791e-46 ], 
                    [ 779.75 , 1.38999220367015e-46 ], [ 779.8125 , 1.390169137490569e-46 ], [ 779.875 , 1.3903460938331316e-46 ], [ 779.9375 , 1.390523072700675e-46 ], 
                    [ 780.0 , 1.3907000740960568e-46 ], [ 780.0625 , 1.3908770980221843e-46 ], [ 780.125 , 1.3910541444819252e-46 ], [ 780.1875 , 1.3912312134780984e-46 ], 
                    [ 780.25 , 1.3914083050135924e-46 ], [ 780.3125 , 1.391585419091316e-46 ], [ 780.375 , 1.3917624069826367e-46 ], [ 780.4375 , 1.3919393313049471e-46 ], 
                    [ 780.5 , 1.3921162781183306e-46 ], [ 780.5625 , 1.3922932474256557e-46 ], [ 780.625 , 1.3924702392297325e-46 ], [ 780.6875 , 1.3926472535334706e-46 ], 
                    [ 780.75 , 1.3928242903396906e-46 ], [ 780.8125 , 1.3930013496512923e-46 ], [ 780.875 , 1.3931784314710971e-46 ], [ 780.9375 , 1.3933555358020064e-46 ], 
                    [ 781.0 , 1.3935326626468617e-46 ], [ 781.0625 , 1.3937098120085053e-46 ], [ 781.125 , 1.3938869838898388e-46 ], [ 781.1875 , 1.394064178293696e-46 ], 
                    [ 781.25 , 1.3942413952229492e-46 ], [ 781.3125 , 1.3944186346804624e-46 ], [ 781.375 , 1.3945958966690995e-46 ], [ 781.4375 , 1.3947731811917243e-46 ], 
                    [ 781.5 , 1.3949504882512015e-46 ], [ 781.5625 , 1.3951278178504158e-46 ], [ 781.625 , 1.3953051699921832e-46 ], [ 781.6875 , 1.3954825446793992e-46 ], 
                    [ 781.75 , 1.3956599419149492e-46 ], [ 781.8125 , 1.39583736170168e-46 ], [ 781.875 , 1.3960148040424385e-46 ], [ 781.9375 , 1.3961922689401318e-46 ], 
                    [ 782.0 , 1.396369756397607e-46 ], [ 782.0625 , 1.3965472664177124e-46 ], [ 782.125 , 1.3967247230771882e-46 ], [ 782.1875 , 1.396902050596277e-46 ], 
                    [ 782.25 , 1.3970794006287747e-46 ], [ 782.3125 , 1.3972567731775402e-46 ], [ 782.375 , 1.3974341682454813e-46 ], [ 782.4375 , 1.3976115858353775e-46 ], 
                    [ 782.5 , 1.3977890259501384e-46 ], [ 782.5625 , 1.3979664885926132e-46 ], [ 782.625 , 1.3981439737656624e-46 ], [ 782.6875 , 1.3983214814721465e-46 ], 
                    [ 782.75 , 1.3984990117149356e-46 ], [ 782.8125 , 1.3986765644968723e-46 ], [ 782.875 , 1.3988541398208268e-46 ], [ 782.9375 , 1.3990317376896616e-46 ], 
                    [ 783.0 , 1.3992093581062392e-46 ], [ 783.0625 , 1.399387001073412e-46 ], [ 783.125 , 1.3995646665940928e-46 ], [ 783.1875 , 1.3997423546710552e-46 ], 
                    [ 783.25 , 1.399920065307243e-46 ], [ 783.3125 , 1.4000977985054702e-46 ], [ 783.375 , 1.4002755542686415e-46 ], [ 783.4375 , 1.4004533325996014e-46 ], 
                    [ 783.5 , 1.4006311335012054e-46 ], [ 783.5625 , 1.4008089569763385e-46 ], [ 783.625 , 1.4009868030278472e-46 ], [ 783.6875 , 1.4011646716585979e-46 ], 
                    [ 783.75 , 1.4013425628714769e-46 ], [ 783.8125 , 1.4015204766693312e-46 ], [ 783.875 , 1.401698413055058e-46 ], [ 783.9375 , 1.40187614350309e-46 ], 
                    [ 784.0 , 1.402053889320736e-46 ], [ 784.0625 , 1.4022316576750108e-46 ], [ 784.125 , 1.402409448568781e-46 ], [ 784.1875 , 1.4025872620049048e-46 ], 
                    [ 784.25 , 1.4027650979862505e-46 ], [ 784.3125 , 1.4029429565156468e-46 ], [ 784.375 , 1.403120837596002e-46 ], [ 784.4375 , 1.403298741230116e-46 ], 
                    [ 784.5 , 1.4034766674208982e-46 ], [ 784.5625 , 1.403654616171178e-46 ], [ 784.625 , 1.403832587483827e-46 ], [ 784.6875 , 1.4040105813617049e-46 ], 
                    [ 784.75 , 1.4041885978076728e-46 ], [ 784.8125 , 1.4043666368246027e-46 ], [ 784.875 , 1.404544698415326e-46 ], [ 784.9375 , 1.404722782582745e-46 ], 
                    [ 785.0 , 1.4049008893296822e-46 ], [ 785.0625 , 1.4050790186590406e-46 ], [ 785.125 , 1.4052571705736533e-46 ], [ 785.1875 , 1.4054353450764138e-46 ], 
                    [ 785.25 , 1.4056135421701367e-46 ], [ 785.3125 , 1.4057917618577254e-46 ], [ 785.375 , 1.4059700041420354e-46 ], [ 785.4375 , 1.4061482690259316e-46 ], 
                    [ 785.5 , 1.4063265565122787e-46 ], [ 785.5625 , 1.4065048666039535e-46 ], [ 785.625 , 1.4066831993038119e-46 ], [ 785.6875 , 1.4068614263931008e-46 ], 
                    [ 785.75 , 1.4070395673857786e-46 ], [ 785.8125 , 1.4072177309351808e-46 ], [ 785.875 , 1.4073959170442133e-46 ], [ 785.9375 , 1.4075741257156825e-46 ], 
                    [ 786.0 , 1.4077523569524557e-46 ], [ 786.0625 , 1.4079306107574203e-46 ], [ 786.125 , 1.4081088871334033e-46 ], [ 786.1875 , 1.4082871860832834e-46 ], 
                    [ 786.25 , 1.4084655076098985e-46 ], [ 786.3125 , 1.4086438517161076e-46 ], [ 786.375 , 1.4088222184047896e-46 ], [ 786.4375 , 1.4090006076787942e-46 ], 
                    [ 786.5 , 1.409179019540991e-46 ], [ 786.5625 , 1.4093574539942302e-46 ], [ 786.625 , 1.4095359110413727e-46 ], [ 786.6875 , 1.409714390685269e-46 ], 
                    [ 786.75 , 1.4098928929287903e-46 ], [ 786.8125 , 1.4100714177747987e-46 ], [ 786.875 , 1.4102499652261558e-46 ], [ 786.9375 , 1.4104285352857338e-46 ], 
                    [ 787.0 , 1.4106071279563561e-46 ], [ 787.0625 , 1.4107857432409352e-46 ], [ 787.125 , 1.4109643811423148e-46 ], [ 787.1875 , 1.4111430416633683e-46 ], 
                    [ 787.25 , 1.4113217248069406e-46 ], [ 787.3125 , 1.4115004305759056e-46 ], [ 787.375 , 1.4116791589731186e-46 ], [ 787.4375 , 1.4118578988115338e-46 ], 
                    [ 787.5 , 1.4120364416277246e-46 ], [ 787.5625 , 1.412215007022317e-46 ], [ 787.625 , 1.4123935949981969e-46 ], [ 787.6875 , 1.4125722055582491e-46 ], 
                    [ 787.75 , 1.4127508387052799e-46 ], [ 787.8125 , 1.4129294944421557e-46 ], [ 787.875 , 1.413108172771753e-46 ], [ 787.9375 , 1.4132868736969192e-46 ], 
                    [ 788.0 , 1.4134655972205216e-46 ], [ 788.0625 , 1.4136443433453978e-46 ], [ 788.125 , 1.413823112074416e-46 ], [ 788.1875 , 1.4140019034104348e-46 ], 
                    [ 788.25 , 1.4141807173563032e-46 ], [ 788.3125 , 1.41435955391491e-46 ], [ 788.375 , 1.4145384130890953e-46 ], [ 788.4375 , 1.4147172948816986e-46 ], 
                    [ 788.5 , 1.4148961992956105e-46 ], [ 788.5625 , 1.4150751263336819e-46 ], [ 788.625 , 1.415254075998763e-46 ], [ 788.6875 , 1.4154330482937462e-46 ], 
                    [ 788.75 , 1.4156120432214525e-46 ], [ 788.8125 , 1.4157910607847946e-46 ], [ 788.875 , 1.4159701009865646e-46 ], [ 788.9375 , 1.4161491638297056e-46 ], 
                    [ 789.0 , 1.4163282493170207e-46 ], [ 789.0625 , 1.4165073574513731e-46 ], [ 789.125 , 1.4166864882356977e-46 ], [ 789.1875 , 1.4168656416727778e-46 ], 
                    [ 789.25 , 1.4170447014793145e-46 ], [ 789.3125 , 1.4172236634469966e-46 ], [ 789.375 , 1.4174026480161937e-46 ], [ 789.4375 , 1.4175816551897906e-46 ], 
                    [ 789.5 , 1.417760684970622e-46 ], [ 789.5625 , 1.4179397373615732e-46 ], [ 789.625 , 1.4181188123654593e-46 ], [ 789.6875 , 1.4182979099851665e-46 ], 
                    [ 789.75 , 1.4184770302235407e-46 ], [ 789.8125 , 1.4186561730834287e-46 ], [ 789.875 , 1.4188353385677074e-46 ], [ 789.9375 , 1.4190145266792238e-46 ], 
                    [ 790.0 , 1.4191937374208359e-46 ], [ 790.0625 , 1.4193729707954014e-46 ], [ 790.125 , 1.419552226805759e-46 ], [ 790.1875 , 1.4197315054548174e-46 ], 
                    [ 790.25 , 1.4199108067453955e-46 ], [ 790.3125 , 1.420090130680373e-46 ], [ 790.375 , 1.4202694772625891e-46 ], [ 790.4375 , 1.4204488464948946e-46 ], 
                    [ 790.5 , 1.42062823838018e-46 ], [ 790.5625 , 1.420807652921326e-46 ], [ 790.625 , 1.4209870901211337e-46 ], [ 790.6875 , 1.421166549982505e-46 ], 
                    [ 790.75 , 1.4213460325082817e-46 ], [ 790.8125 , 1.4215255377013667e-46 ], [ 790.875 , 1.4217050655645517e-46 ], [ 790.9375 , 1.4218846161007607e-46 ], 
                    [ 791.0 , 1.4220641893128266e-46 ], [ 791.0625 , 1.4222435756220502e-46 ], [ 791.125 , 1.4224229529952733e-46 ], [ 791.1875 , 1.4226023529920774e-46 ], 
                    [ 791.25 , 1.4227817756153047e-46 ], [ 791.3125 , 1.4229612208678402e-46 ], [ 791.375 , 1.4231406887524969e-46 ], [ 791.4375 , 1.42332017927217e-46 ], 
                    [ 791.5 , 1.4234996924296838e-46 ], [ 791.5625 , 1.4236792282279035e-46 ], [ 791.625 , 1.4238587866696848e-46 ], [ 791.6875 , 1.4240383677578835e-46 ], 
                    [ 791.75 , 1.4242179714953556e-46 ], [ 791.8125 , 1.4243975978849582e-46 ], [ 791.875 , 1.424577246929558e-46 ], [ 791.9375 , 1.424756918631982e-46 ], 
                    [ 792.0 , 1.4249366129951283e-46 ], [ 792.0625 , 1.4251163300218246e-46 ], [ 792.125 , 1.4252960697149395e-46 ], [ 792.1875 , 1.4254758320773419e-46 ], 
                    [ 792.25 , 1.4256556171118604e-46 ], [ 792.3125 , 1.4258354248213952e-46 ], [ 792.375 , 1.4260152552087654e-46 ], [ 792.4375 , 1.4261951082768615e-46 ], 
                    [ 792.5 , 1.4263749840285243e-46 ], [ 792.5625 , 1.4265548824666445e-46 ], [ 792.625 , 1.4267348035940532e-46 ], [ 792.6875 , 1.4269147474136325e-46 ], 
                    [ 792.75 , 1.4270947139282034e-46 ], [ 792.8125 , 1.4272746600563963e-46 ], [ 792.875 , 1.4274544353115078e-46 ], [ 792.9375 , 1.4276342332105739e-46 ], 
                    [ 793.0 , 1.427814053756436e-46 ], [ 793.0625 , 1.4279938969519578e-46 ], [ 793.125 , 1.4281737628000114e-46 ], [ 793.1875 , 1.4283536513034102e-46 ], 
                    [ 793.25 , 1.428533562465048e-46 ], [ 793.3125 , 1.4287134962877285e-46 ], [ 793.375 , 1.4288934527743662e-46 ], [ 793.4375 , 1.4290734319277757e-46 ], 
                    [ 793.5 , 1.4292534337508017e-46 ], [ 793.5625 , 1.42943345824634e-46 ], [ 793.625 , 1.4296135054172266e-46 ], [ 793.6875 , 1.4297935752663173e-46 ], 
                    [ 793.75 , 1.4299736677964687e-46 ], [ 793.8125 , 1.4301537830105374e-46 ], [ 793.875 , 1.4303339209113705e-46 ], [ 793.9375 , 1.4305140815018563e-46 ], 
                    [ 794.0 , 1.4306942647847915e-46 ], [ 794.0625 , 1.4308744707631053e-46 ], [ 794.125 , 1.4310546994396267e-46 ], [ 794.1875 , 1.4312349508172138e-46 ], 
                    [ 794.25 , 1.4314152248987364e-46 ], [ 794.3125 , 1.431595521687034e-46 ], [ 794.375 , 1.4317758411849666e-46 ], [ 794.4375 , 1.431956183395415e-46 ], 
                    [ 794.5 , 1.43213654832123e-46 ], [ 794.5625 , 1.432316935965262e-46 ], [ 794.625 , 1.4324972435716326e-46 ], [ 794.6875 , 1.4326774376882824e-46 ], 
                    [ 794.75 , 1.432857654471607e-46 ], [ 794.8125 , 1.4330378939244164e-46 ], [ 794.875 , 1.4332181560496035e-46 ], [ 794.9375 , 1.4333984408499788e-46 ], 
                    [ 795.0 , 1.4335787483284362e-46 ], [ 795.0625 , 1.4337590784877874e-46 ], [ 795.125 , 1.4339394313309155e-46 ], [ 795.1875 , 1.4341198068606539e-46 ], 
                    [ 795.25 , 1.4343002050798972e-46 ], [ 795.3125 , 1.4344806259914279e-46 ], [ 795.375 , 1.4346610695981822e-46 ], [ 795.4375 , 1.4348415359029334e-46 ], 
                    [ 795.5 , 1.4350220249085877e-46 ], [ 795.5625 , 1.4352025366180104e-46 ], [ 795.625 , 1.4353830710340479e-46 ], [ 795.6875 , 1.4355636281595153e-46 ], 
                    [ 795.75 , 1.4357442079972997e-46 ], [ 795.8125 , 1.4359248105502888e-46 ], [ 795.875 , 1.4361054358212689e-46 ], [ 795.9375 , 1.436286083813138e-46 ], 
                    [ 796.0 , 1.4364667545288053e-46 ], [ 796.0625 , 1.4366474479710274e-46 ], [ 796.125 , 1.4368281641427446e-46 ], [ 796.1875 , 1.4370089030467855e-46 ], 
                    [ 796.25 , 1.4371896646859787e-46 ], [ 796.3125 , 1.4373704490632357e-46 ], [ 796.375 , 1.4375512561814057e-46 ], [ 796.4375 , 1.4377319375338827e-46 ], 
                    [ 796.5 , 1.4379125478108505e-46 ], [ 796.5625 , 1.4380931807763918e-46 ], [ 796.625 , 1.4382738364333367e-46 ], [ 796.6875 , 1.4384545147845454e-46 ], 
                    [ 796.75 , 1.4386352158328895e-46 ], [ 796.8125 , 1.4388159395811794e-46 ], [ 796.875 , 1.4389966860322973e-46 ], [ 796.9375 , 1.439177455189085e-46 ], 
                    [ 797.0 , 1.4393582470543847e-46 ], [ 797.0625 , 1.439539061631059e-46 ], [ 797.125 , 1.4397198989219818e-46 ], [ 797.1875 , 1.439900758929965e-46 ], 
                    [ 797.25 , 1.4400816416578938e-46 ], [ 797.3125 , 1.4402625471086012e-46 ], [ 797.375 , 1.440443475284942e-46 ], [ 797.4375 , 1.4406244261898118e-46 ], 
                    [ 797.5 , 1.4408053998260354e-46 ], [ 797.5625 , 1.4409863961964476e-46 ], [ 797.625 , 1.441167415303935e-46 ], [ 797.6875 , 1.4413484571513234e-46 ], 
                    [ 797.75 , 1.44152952174151e-46 ], [ 797.8125 , 1.4417106090773317e-46 ], [ 797.875 , 1.4418917191616459e-46 ], [ 797.9375 , 1.4420728519972893e-46 ], 
                    [ 798.0 , 1.4422540075871715e-46 ], [ 798.0625 , 1.4424351859340996e-46 ], [ 798.125 , 1.442616387040953e-46 ], [ 798.1875 , 1.442797610910591e-46 ], 
                    [ 798.25 , 1.442978679643152e-46 ], [ 798.3125 , 1.44315970547846e-46 ], [ 798.375 , 1.4433407540239805e-46 ], [ 798.4375 , 1.443521825282563e-46 ], 
                    [ 798.5 , 1.4437029192570468e-46 ], [ 798.5625 , 1.4438840359503224e-46 ], [ 798.625 , 1.4440651753651786e-46 ], [ 798.6875 , 1.4442463375044966e-46 ], 
                    [ 798.75 , 1.4444275223711478e-46 ], [ 798.8125 , 1.4446087299679526e-46 ], [ 798.875 , 1.4447899602977625e-46 ], [ 798.9375 , 1.4449712133634603e-46 ], 
                    [ 799.0 , 1.4451524891678572e-46 ], [ 799.0625 , 1.4453337877138366e-46 ], [ 799.125 , 1.4455151090042515e-46 ], [ 799.1875 , 1.4456964530419348e-46 ], 
                    [ 799.25 , 1.4458778198297401e-46 ], [ 799.3125 , 1.446059209370542e-46 ], [ 799.375 , 1.446240621667185e-46 ], [ 799.4375 , 1.4464220567225233e-46 ], 
                    [ 799.5 , 1.4466035145394125e-46 ], [ 799.5625 , 1.4467849951207079e-46 ], [ 799.625 , 1.4469664984692654e-46 ], [ 799.6875 , 1.4471480245879413e-46 ], 
                    [ 799.75 , 1.4473295734795818e-46 ], [ 799.8125 , 1.4475111451470745e-46 ], [ 799.875 , 1.4476927395932362e-46 ], [ 799.9375 , 1.4478743568209752e-46 ], 
                    [ 800.0 , 1.4480559968330884e-46 ], [ 800.0625 , 1.448237468486947e-46 ], [ 800.125 , 1.4484189098150905e-46 ], [ 800.1875 , 1.4486003738749726e-46 ], 
                    [ 800.25 , 1.4487818606694412e-46 ], [ 800.3125 , 1.448963370201355e-46 ], [ 800.375 , 1.449144902473542e-46 ], [ 800.4375 , 1.4493264574888614e-46 ], 
                    [ 800.5 , 1.4495080352501626e-46 ], [ 800.5625 , 1.4496896357602953e-46 ], [ 800.625 , 1.4498712590221096e-46 ], [ 800.6875 , 1.4500529050384663e-46 ], 
                    [ 800.75 , 1.450234573812185e-46 ], [ 800.8125 , 1.4504162653461688e-46 ], [ 800.875 , 1.4505979796432174e-46 ], [ 800.9375 , 1.450779716706203e-46 ], 
                    [ 801.0 , 1.4509614765379785e-46 ], [ 801.0625 , 1.451143259141396e-46 ], [ 801.125 , 1.4513250645193082e-46 ], [ 801.1875 , 1.4515068926745793e-46 ], 
                    [ 801.25 , 1.4516887436100316e-46 ], [ 801.3125 , 1.4518706173285604e-46 ], [ 801.375 , 1.4520525138329894e-46 ], [ 801.4375 , 1.4522344331261832e-46 ], 
                    [ 801.5 , 1.4524163752109971e-46 ], [ 801.5625 , 1.452598340090297e-46 ], [ 801.625 , 1.452780327766918e-46 ], [ 801.6875 , 1.4529623382437365e-46 ], 
                    [ 801.75 , 1.4531443715235783e-46 ], [ 801.8125 , 1.4533264276093417e-46 ], [ 801.875 , 1.4535083156916317e-46 ], [ 801.9375 , 1.453690169251202e-46 ], 
                    [ 802.0 , 1.4538720455631135e-46 ], [ 802.0625 , 1.454053944630213e-46 ], [ 802.125 , 1.4542358664553369e-46 ], [ 802.1875 , 1.4544178110413532e-46 ], 
                    [ 802.25 , 1.454599778391089e-46 ], [ 802.3125 , 1.454781768507413e-46 ], [ 802.375 , 1.4549637813931636e-46 ], [ 802.4375 , 1.4551458170511888e-46 ], 
                    [ 802.5 , 1.4553278754843384e-46 ], [ 802.5625 , 1.4555099566954516e-46 ], [ 802.625 , 1.4556920606873878e-46 ], [ 802.6875 , 1.4558741874630081e-46 ], 
                    [ 802.75 , 1.4560563370251527e-46 ], [ 802.8125 , 1.4562385093766724e-46 ], [ 802.875 , 1.4564207045204188e-46 ], [ 802.9375 , 1.4566029224592426e-46 ], 
                    [ 803.0 , 1.4567851631959657e-46 ], [ 803.0625 , 1.4569674267335126e-46 ], [ 803.125 , 1.457149713074684e-46 ], [ 803.1875 , 1.457332022222333e-46 ], 
                    [ 803.25 , 1.4575143541793442e-46 ], [ 803.3125 , 1.4576967089485297e-46 ], [ 803.375 , 1.4578790865327648e-46 ], [ 803.4375 , 1.4580614869349243e-46 ], 
                    [ 803.5 , 1.4582439101578113e-46 ], [ 803.5625 , 1.4584263562043325e-46 ], [ 803.625 , 1.4586088250773338e-46 ], [ 803.6875 , 1.458791153045613e-46 ], 
                    [ 803.75 , 1.4589734301517183e-46 ], [ 803.8125 , 1.459155730033492e-46 ], [ 803.875 , 1.4593380526937696e-46 ], [ 803.9375 , 1.4595203981354072e-46 ], 
                    [ 804.0 , 1.459702766361252e-46 ], [ 804.0625 , 1.4598851573741608e-46 ], [ 804.125 , 1.4600675711769602e-46 ], [ 804.1875 , 1.4602500077725187e-46 ], 
                    [ 804.25 , 1.4604324671636636e-46 ], [ 804.3125 , 1.4606149493532636e-46 ], [ 804.375 , 1.4607974543441366e-46 ], [ 804.4375 , 1.460979982139173e-46 ], 
                    [ 804.5 , 1.461162532741191e-46 ], [ 804.5625 , 1.4613451061530613e-46 ], [ 804.625 , 1.461527702377613e-46 ], [ 804.6875 , 1.4617103214177283e-46 ], 
                    [ 804.75 , 1.4618929632762155e-46 ], [ 804.8125 , 1.462075627955937e-46 ], [ 804.875 , 1.4622583154597647e-46 ], [ 804.9375 , 1.4624410257905299e-46 ], 
                    [ 805.0 , 1.4626237589511264e-46 ], [ 805.0625 , 1.4628065149443546e-46 ], [ 805.125 , 1.4629892937731093e-46 ], [ 805.1875 , 1.463172095440202e-46 ], 
                    [ 805.25 , 1.4633549199485075e-46 ], [ 805.3125 , 1.4635377673008796e-46 ], [ 805.375 , 1.4637206375001832e-46 ], [ 805.4375 , 1.4639035305492414e-46 ], 
                    [ 805.5 , 1.464086307701907e-46 ], [ 805.5625 , 1.4642689916562554e-46 ], [ 805.625 , 1.4644516984053272e-46 ], [ 805.6875 , 1.4646344279519457e-46 ], 
                    [ 805.75 , 1.4648171802989867e-46 ], [ 805.8125 , 1.4649999554492644e-46 ], [ 805.875 , 1.4651827534056342e-46 ], [ 805.9375 , 1.4653655741709416e-46 ], 
                    [ 806.0 , 1.4655484177480328e-46 ], [ 806.0625 , 1.4657312841397546e-46 ], [ 806.125 , 1.4659141733489633e-46 ], [ 806.1875 , 1.466097085378465e-46 ], 
                    [ 806.25 , 1.4662800202311591e-46 ], [ 806.3125 , 1.4664629779098826e-46 ], [ 806.375 , 1.4666459584174634e-46 ], [ 806.4375 , 1.4668289617567497e-46 ], 
                    [ 806.5 , 1.4670119879306006e-46 ], [ 806.5625 , 1.4671950369418866e-46 ], [ 806.625 , 1.4673781087934256e-46 ], [ 806.6875 , 1.4675612034880782e-46 ], 
                    [ 806.75 , 1.4677443210286842e-46 ], [ 806.8125 , 1.467927461418136e-46 ], [ 806.875 , 1.4681106246592115e-46 ], [ 806.9375 , 1.468293810754835e-46 ], 
                    [ 807.0 , 1.4684770197078173e-46 ], [ 807.0625 , 1.4686602515210091e-46 ], [ 807.125 , 1.468843506197295e-46 ], [ 807.1875 , 1.4690267837394754e-46 ], 
                    [ 807.25 , 1.4692100841504663e-46 ], [ 807.3125 , 1.4693933255740878e-46 ], [ 807.375 , 1.4695764260741018e-46 ], [ 807.4375 , 1.4697595493901835e-46 ], 
                    [ 807.5 , 1.4699426955251972e-46 ], [ 807.5625 , 1.4701258644819753e-46 ], [ 807.625 , 1.4703090562633625e-46 ], [ 807.6875 , 1.4704922708722023e-46 ], 
                    [ 807.75 , 1.4706755083113391e-46 ], [ 807.8125 , 1.4708587685836184e-46 ], [ 807.875 , 1.4710420516918643e-46 ], [ 807.9375 , 1.4712253576389848e-46 ], 
                    [ 808.0 , 1.4714086864277529e-46 ], [ 808.0625 , 1.4715920380610568e-46 ], [ 808.125 , 1.4717754125417219e-46 ], [ 808.1875 , 1.471958809872627e-46 ], 
                    [ 808.25 , 1.4721422300565774e-46 ], [ 808.3125 , 1.4723256730964417e-46 ], [ 808.375 , 1.4725091389950994e-46 ], [ 808.4375 , 1.472692627755336e-46 ], 
                    [ 808.5 , 1.472876139380032e-46 ], [ 808.5625 , 1.4730596738720465e-46 ], [ 808.625 , 1.4732432312342087e-46 ], [ 808.6875 , 1.4734268114693888e-46 ], 
                    [ 808.75 , 1.473610414580427e-46 ], [ 808.8125 , 1.4737940405701734e-46 ], [ 808.875 , 1.4739776894414687e-46 ], [ 808.9375 , 1.4741613611971851e-46 ], 
                    [ 809.0 , 1.4743450558401845e-46 ], [ 809.0625 , 1.4745287733732563e-46 ], [ 809.125 , 1.4747124998407776e-46 ], [ 809.1875 , 1.4748960131559528e-46 ], 
                    [ 809.25 , 1.475079549307568e-46 ], [ 809.3125 , 1.475263108298402e-46 ], [ 809.375 , 1.4754466901313386e-46 ], [ 809.4375 , 1.4756302948092101e-46 ], 
                    [ 809.5 , 1.4758139223348589e-46 ], [ 809.5625 , 1.4759975727111284e-46 ], [ 809.625 , 1.476181245940862e-46 ], [ 809.6875 , 1.4763649420269042e-46 ], 
                    [ 809.75 , 1.4765486609720984e-46 ], [ 809.8125 , 1.4767324027793004e-46 ], [ 809.875 , 1.476916167451323e-46 ], [ 809.9375 , 1.4770999549910435e-46 ], 
                    [ 810.0 , 1.4772837654013073e-46 ], [ 810.0625 , 1.4774675986849288e-46 ], [ 810.125 , 1.4776514548447863e-46 ], [ 810.1875 , 1.4778353338837157e-46 ], 
                    [ 810.25 , 1.4780192358045537e-46 ], [ 810.3125 , 1.4782031606101792e-46 ], [ 810.375 , 1.4783871083034082e-46 ], [ 810.4375 , 1.4785710788871205e-46 ], 
                    [ 810.5 , 1.4787550723641124e-46 ], [ 810.5625 , 1.4789390887372536e-46 ], [ 810.625 , 1.4791231280094352e-46 ], [ 810.6875 , 1.4793071901834753e-46 ], 
                    [ 810.75 , 1.4794912752622132e-46 ], [ 810.8125 , 1.4796753832484888e-46 ], [ 810.875 , 1.4798595141452155e-46 ], [ 810.9375 , 1.480043667955192e-46 ], 
                    [ 811.0 , 1.4802276657952876e-46 ], [ 811.0625 , 1.4804116141425017e-46 ], [ 811.125 , 1.4805955853490334e-46 ], [ 811.1875 , 1.4807795794177236e-46 ], 
                    [ 811.25 , 1.4809635963514029e-46 ], [ 811.3125 , 1.4811476361529334e-46 ], [ 811.375 , 1.4813316988251466e-46 ], [ 811.4375 , 1.4815157843708954e-46 ], 
                    [ 811.5 , 1.48169989279298e-46 ], [ 811.5625 , 1.4818840240943167e-46 ], [ 811.625 , 1.482068178277686e-46 ], [ 811.6875 , 1.4822523553459419e-46 ], 
                    [ 811.75 , 1.4824365553019492e-46 ], [ 811.8125 , 1.4826207781485414e-46 ], [ 811.875 , 1.4828050238885531e-46 ], [ 811.9375 , 1.4829892925248604e-46 ], 
                    [ 812.0 , 1.4831735840602985e-46 ], [ 812.0625 , 1.4833578984976495e-46 ], [ 812.125 , 1.4835422358398548e-46 ], [ 812.1875 , 1.483726596089718e-46 ], 
                    [ 812.25 , 1.4839109792500759e-46 ], [ 812.3125 , 1.484095385323775e-46 ], [ 812.375 , 1.4842798143137055e-46 ], [ 812.4375 , 1.4844642662226515e-46 ], 
                    [ 812.5 , 1.4846487410535033e-46 ], [ 812.5625 , 1.4848332388090888e-46 ], [ 812.625 , 1.4850177594922778e-46 ], [ 812.6875 , 1.485202303105898e-46 ], 
                    [ 812.75 , 1.4853868696528309e-46 ], [ 812.8125 , 1.4855713826555843e-46 ], [ 812.875 , 1.4857557414547743e-46 ], [ 812.9375 , 1.4859401231327945e-46 ], 
                    [ 813.0 , 1.486124527692516e-46 ], [ 813.0625 , 1.4863089551367674e-46 ], [ 813.125 , 1.4864934054683786e-46 ], [ 813.1875 , 1.4866778786902109e-46 ], 
                    [ 813.25 , 1.4868623748050943e-46 ], [ 813.3125 , 1.4870468938158699e-46 ], [ 813.375 , 1.487231435725379e-46 ], [ 813.4375 , 1.4874160005364634e-46 ], 
                    [ 813.5 , 1.487600588251965e-46 ], [ 813.5625 , 1.4877851988747265e-46 ], [ 813.625 , 1.48796983240758e-46 ], [ 813.6875 , 1.4881544888533897e-46 ], 
                    [ 813.75 , 1.4883391682149887e-46 ], [ 813.8125 , 1.48852387049521e-46 ], [ 813.875 , 1.4887085956969088e-46 ], [ 813.9375 , 1.48889334382295e-46 ], 
                    [ 814.0 , 1.489078114876137e-46 ], [ 814.0625 , 1.4892629088593572e-46 ], [ 814.125 , 1.4894477257754135e-46 ], [ 814.1875 , 1.4896325656271838e-46 ], 
                    [ 814.25 , 1.4898174284175147e-46 ], [ 814.3125 , 1.4900023141492208e-46 ], [ 814.375 , 1.4901872228251915e-46 ], [ 814.4375 , 1.4903721544482318e-46 ], 
                    [ 814.5 , 1.4905571090212318e-46 ], [ 814.5625 , 1.490742086546987e-46 ], [ 814.625 , 1.4909270870284191e-46 ], [ 814.6875 , 1.4911119024866967e-46 ], 
                    [ 814.75 , 1.4912966966218574e-46 ], [ 814.8125 , 1.4914815136586449e-46 ], [ 814.875 , 1.4916663535998866e-46 ], [ 814.9375 , 1.4918512164484105e-46 ], 
                    [ 815.0 , 1.492036102207077e-46 ], [ 815.0625 , 1.4922210108787145e-46 ], [ 815.125 , 1.4924059424661629e-46 ], [ 815.1875 , 1.4925908969722511e-46 ], 
                    [ 815.25 , 1.492775874399862e-46 ], [ 815.3125 , 1.4929608747517624e-46 ], [ 815.375 , 1.4931458980308564e-46 ], [ 815.4375 , 1.4933309442399646e-46 ], 
                    [ 815.5 , 1.4935160133819391e-46 ], [ 815.5625 , 1.4937011054595902e-46 ], [ 815.625 , 1.4938862204758028e-46 ], [ 815.6875 , 1.494071358433388e-46 ], 
                    [ 815.75 , 1.494256519335199e-46 ], [ 815.8125 , 1.49444170318408e-46 ], [ 815.875 , 1.4946269099828637e-46 ], [ 815.9375 , 1.4948121397344159e-46 ], 
                    [ 816.0 , 1.494997392441581e-46 ], [ 816.0625 , 1.4951826681071824e-46 ], [ 816.125 , 1.4953679667340765e-46 ], [ 816.1875 , 1.4955532883250978e-46 ], 
                    [ 816.25 , 1.4957386328831137e-46 ], [ 816.3125 , 1.4959240004109492e-46 ], [ 816.375 , 1.496109390911483e-46 ], [ 816.4375 , 1.49629480438753e-46 ], 
                    [ 816.5 , 1.4964801693630298e-46 ], [ 816.5625 , 1.4966653742037505e-46 ], [ 816.625 , 1.4968506019654677e-46 ], [ 816.6875 , 1.4970358526510505e-46 ], 
                    [ 816.75 , 1.4972211262633037e-46 ], [ 816.8125 , 1.497406422805086e-46 ], [ 816.875 , 1.4975917422792143e-46 ], [ 816.9375 , 1.4977770846885261e-46 ], 
                    [ 817.0 , 1.4979624500358925e-46 ], [ 817.0625 , 1.4981478383241196e-46 ], [ 817.125 , 1.498333249556047e-46 ], [ 817.1875 , 1.4985186837345464e-46 ], 
                    [ 817.25 , 1.4987041408624362e-46 ], [ 817.3125 , 1.4988896209425349e-46 ], [ 817.375 , 1.4990751239777153e-46 ], [ 817.4375 , 1.4992606499708292e-46 ], 
                    [ 817.5 , 1.4994461989246645e-46 ], [ 817.5625 , 1.4996317708420945e-46 ], [ 817.625 , 1.4998173657259618e-46 ], [ 817.6875 , 1.5000029835791292e-46 ], 
                    [ 817.75 , 1.5001886244044082e-46 ], [ 817.8125 , 1.5003742882046629e-46 ], [ 817.875 , 1.5005599749827152e-46 ], [ 817.9375 , 1.5007456847414089e-46 ], 
                    [ 818.0 , 1.5009314174836094e-46 ], [ 818.0625 , 1.50111717321214e-46 ], [ 818.125 , 1.5013029519298987e-46 ], [ 818.1875 , 1.5014887536396452e-46 ], 
                    [ 818.25 , 1.5016745783442787e-46 ], [ 818.3125 , 1.501860426046634e-46 ], [ 818.375 , 1.5020461253581593e-46 ], [ 818.4375 , 1.5022317632341509e-46 ], 
                    [ 818.5 , 1.5024174240531056e-46 ], [ 818.5625 , 1.5026031078179015e-46 ], [ 818.625 , 1.5027888145313428e-46 ], [ 818.6875 , 1.5029745441962977e-46 ], 
                    [ 818.75 , 1.50316029681556e-46 ], [ 818.8125 , 1.5033460723919878e-46 ], [ 818.875 , 1.503531870928429e-46 ], [ 818.9375 , 1.5037176924277003e-46 ], 
                    [ 819.0 , 1.5039035368926503e-46 ], [ 819.0625 , 1.5040894043261274e-46 ], [ 819.125 , 1.5042752947309499e-46 ], [ 819.1875 , 1.5044612081099771e-46 ], 
                    [ 819.25 , 1.5046471444660492e-46 ], [ 819.3125 , 1.504833103801952e-46 ], [ 819.375 , 1.5050190861206009e-46 ], [ 819.4375 , 1.505205091424793e-46 ], 
                    [ 819.5 , 1.505391119717391e-46 ], [ 819.5625 , 1.5055771710012036e-46 ], [ 819.625 , 1.5057632452791153e-46 ], [ 819.6875 , 1.505949342553957e-46 ], 
                    [ 819.75 , 1.5061354628285499e-46 ], [ 819.8125 , 1.5063216061057465e-46 ], [ 819.875 , 1.5065077723883797e-46 ], [ 819.9375 , 1.5066939616793247e-46 ], 
                    [ 820.0 , 1.5068801739813822e-46 ], [ 820.0625 , 1.5070664092974174e-46 ], [ 820.125 , 1.5072526676302748e-46 ], [ 820.1875 , 1.5074389489828097e-46 ], 
                    [ 820.25 , 1.5076249988043618e-46 ], [ 820.3125 , 1.5078110695766715e-46 ], [ 820.375 , 1.5079971633138089e-46 ], [ 820.4375 , 1.5081832800185871e-46 ], 
                    [ 820.5 , 1.5083694196938623e-46 ], [ 820.5625 , 1.5085555823424582e-46 ], [ 820.625 , 1.5087417679672e-46 ], [ 820.6875 , 1.508927976570934e-46 ], 
                    [ 820.75 , 1.5091142081565067e-46 ], [ 820.8125 , 1.5093004627267335e-46 ], [ 820.875 , 1.5094867402844722e-46 ], [ 820.9375 , 1.5096730408325494e-46 ], 
                    [ 821.0 , 1.509859364373792e-46 ], [ 821.0625 , 1.510045710911059e-46 ], [ 821.125 , 1.5102320804471565e-46 ], [ 821.1875 , 1.5104184729849657e-46 ], 
                    [ 821.25 , 1.5106048885272937e-46 ], [ 821.3125 , 1.5107913270770009e-46 ], [ 821.375 , 1.510977788636927e-46 ], [ 821.4375 , 1.5111642732098472e-46 ], 
                    [ 821.5 , 1.511350780798731e-46 ], [ 821.5625 , 1.5115373114062898e-46 ], [ 821.625 , 1.51172386503544e-46 ], [ 821.6875 , 1.5119104416889907e-46 ], 
                    [ 821.75 , 1.512097041369805e-46 ], [ 821.8125 , 1.5122836640807143e-46 ], [ 821.875 , 1.512470309824539e-46 ], [ 821.9375 , 1.5126569786041546e-46 ], 
                    [ 822.0 , 1.5128436704224038e-46 ], [ 822.0625 , 1.5130303212440474e-46 ], [ 822.125 , 1.5132167997186568e-46 ], [ 822.1875 , 1.5134033011764396e-46 ], 
                    [ 822.25 , 1.5135898256202064e-46 ], [ 822.3125 , 1.5137763730528227e-46 ], [ 822.375 , 1.5139629434770895e-46 ], [ 822.4375 , 1.5141495368958407e-46 ], 
                    [ 822.5 , 1.5143361533119316e-46 ], [ 822.5625 , 1.514522792728186e-46 ], [ 822.625 , 1.5147094551474276e-46 ], [ 822.6875 , 1.5148961405725135e-46 ], 
                    [ 822.75 , 1.5150828490062784e-46 ], [ 822.8125 , 1.515269580451537e-46 ], [ 822.875 , 1.5154563349111256e-46 ], [ 822.9375 , 1.515643112387891e-46 ], 
                    [ 823.0 , 1.5158299128846813e-46 ], [ 823.0625 , 1.5160167364043222e-46 ], [ 823.125 , 1.5162035829496413e-46 ], [ 823.1875 , 1.5163904525235191e-46 ], 
                    [ 823.25 , 1.5165773451287183e-46 ], [ 823.3125 , 1.5167642607681533e-46 ], [ 823.375 , 1.5169511994445872e-46 ], [ 823.4375 , 1.5171381611609673e-46 ], 
                    [ 823.5 , 1.5173251459200146e-46 ], [ 823.5625 , 1.5175121537246552e-46 ], [ 823.625 , 1.5176991845776972e-46 ], [ 823.6875 , 1.5178862384819488e-46 ], 
                    [ 823.75 , 1.5180733154403161e-46 ], [ 823.8125 , 1.518260415455586e-46 ], [ 823.875 , 1.5184475385306334e-46 ], [ 823.9375 , 1.5186345721477233e-46 ], 
                    [ 824.0 , 1.5188214821844296e-46 ], [ 824.0625 , 1.5190084152255593e-46 ], [ 824.125 , 1.5191953712739868e-46 ], [ 824.1875 , 1.5193823503325331e-46 ], 
                    [ 824.25 , 1.5195693524040303e-46 ], [ 824.3125 , 1.5197563774913322e-46 ], [ 824.375 , 1.5199434255972284e-46 ], [ 824.4375 , 1.5201304967245627e-46 ], 
                    [ 824.5 , 1.5203175908762014e-46 ], [ 824.5625 , 1.520504708054945e-46 ], [ 824.625 , 1.5206918482636392e-46 ], [ 824.6875 , 1.5208790115051182e-46 ], 
                    [ 824.75 , 1.5210661977822167e-46 ], [ 824.8125 , 1.52125340709777e-46 ], [ 824.875 , 1.5214406394546136e-46 ], [ 824.9375 , 1.5216278948555618e-46 ], 
                    [ 825.0 , 1.5218151733035373e-46 ], [ 825.0625 , 1.5220024748012682e-46 ], [ 825.125 , 1.5221897993516354e-46 ], [ 825.1875 , 1.5223771469574538e-46 ], 
                    [ 825.25 , 1.5225645176216156e-46 ], [ 825.3125 , 1.522751911346915e-46 ], [ 825.375 , 1.5229393281362127e-46 ], [ 825.4375 , 1.523126767992325e-46 ], 
                    [ 825.5 , 1.5233142309180803e-46 ], [ 825.5625 , 1.523501716916383e-46 ], [ 825.625 , 1.5236892259899971e-46 ], [ 825.6875 , 1.5238767581418383e-46 ], 
                    [ 825.75 , 1.5240643133746603e-46 ], [ 825.8125 , 1.524251748840854e-46 ], [ 825.875 , 1.5244390923793519e-46 ], [ 825.9375 , 1.5246264589439579e-46 ], 
                    [ 826.0 , 1.5248138485375229e-46 ], [ 826.0625 , 1.5250012611628673e-46 ], [ 826.125 , 1.5251886968228106e-46 ], [ 826.1875 , 1.525376155520195e-46 ], 
                    [ 826.25 , 1.5255636372578736e-46 ], [ 826.3125 , 1.5257511420386348e-46 ], [ 826.375 , 1.5259386698653325e-46 ], [ 826.4375 , 1.5261262207407995e-46 ], 
                    [ 826.5 , 1.5263137946678902e-46 ], [ 826.5625 , 1.5265013916493943e-46 ], [ 826.625 , 1.5266890116881673e-46 ], [ 826.6875 , 1.526876654787043e-46 ], 
                    [ 826.75 , 1.5270643209488666e-46 ], [ 826.8125 , 1.527252010176462e-46 ], [ 826.875 , 1.527439722472653e-46 ], [ 826.9375 , 1.5276274578402861e-46 ], 
                    [ 827.0 , 1.5278152162822078e-46 ], [ 827.0625 , 1.5280029978012213e-46 ], [ 827.125 , 1.5281908024002283e-46 ], [ 827.1875 , 1.5283786300819787e-46 ], 
                    [ 827.25 , 1.5285664808493639e-46 ], [ 827.3125 , 1.52875435470521e-46 ], [ 827.375 , 1.5289422516523666e-46 ], [ 827.4375 , 1.529130171693671e-46 ], 
                    [ 827.5 , 1.5293181148319292e-46 ], [ 827.5625 , 1.5295060810700016e-46 ], [ 827.625 , 1.5296940704107282e-46 ], [ 827.6875 , 1.5298819201427225e-46 ], 
                    [ 827.75 , 1.530069687245305e-46 ], [ 827.8125 , 1.5302574773931309e-46 ], [ 827.875 , 1.530445290589018e-46 ], [ 827.9375 , 1.5306331268357732e-46 ], 
                    [ 828.0 , 1.530820986136269e-46 ], [ 828.0625 , 1.5310088684933023e-46 ], [ 828.125 , 1.5311967739097139e-46 ], [ 828.1875 , 1.5313847023883447e-46 ], 
                    [ 828.25 , 1.5315726539320144e-46 ], [ 828.3125 , 1.531760628543532e-46 ], [ 828.375 , 1.5319486262257505e-46 ], [ 828.4375 , 1.532136646981523e-46 ], 
                    [ 828.5 , 1.53232469081366e-46 ], [ 828.5625 , 1.5325127577249826e-46 ], [ 828.625 , 1.5327008477183454e-46 ], [ 828.6875 , 1.5328889607965482e-46 ], 
                    [ 828.75 , 1.5330770969624573e-46 ], [ 828.8125 , 1.5332652562189062e-46 ], [ 828.875 , 1.5334534385686855e-46 ], [ 828.9375 , 1.5336416440146946e-46 ], 
                    [ 829.0 , 1.5338298725597356e-46 ], [ 829.0625 , 1.5340181242066327e-46 ], [ 829.125 , 1.5342063989582539e-46 ], [ 829.1875 , 1.5343946968173693e-46 ], 
                    [ 829.25 , 1.5345830177869025e-46 ], [ 829.3125 , 1.5347713618696135e-46 ], [ 829.375 , 1.5349597290683934e-46 ], [ 829.4375 , 1.5351481193860357e-46 ], 
                    [ 829.5 , 1.5353365328254108e-46 ], [ 829.5625 , 1.5355248150663291e-46 ], [ 829.625 , 1.5357130133194299e-46 ], [ 829.6875 , 1.5359012346386396e-46 ], 
                    [ 829.75 , 1.536089479026785e-46 ], [ 829.8125 , 1.5362777464866938e-46 ], [ 829.875 , 1.5364660370211936e-46 ], [ 829.9375 , 1.5366543506331014e-46 ], 
                    [ 830.0 , 1.5368426873252788e-46 ], [ 830.0625 , 1.5370310471005324e-46 ], [ 830.125 , 1.5372194299616695e-46 ], [ 830.1875 , 1.5374078359115632e-46 ], 
                    [ 830.25 , 1.5375962649530112e-46 ], [ 830.3125 , 1.5377847170888535e-46 ], [ 830.375 , 1.5379731923219106e-46 ], [ 830.4375 , 1.5381616906550458e-46 ], 
                    [ 830.5 , 1.5383502120910573e-46 ], [ 830.5625 , 1.5385387566327771e-46 ], [ 830.625 , 1.5387273242830698e-46 ], [ 830.6875 , 1.5389159150447564e-46 ], 
                    [ 830.75 , 1.5391045289206151e-46 ], [ 830.8125 , 1.5392931659135554e-46 ], [ 830.875 , 1.5394818260263667e-46 ], [ 830.9375 , 1.5396705092618827e-46 ], 
                    [ 831.0 , 1.5398592156229703e-46 ], [ 831.0625 , 1.5400479451124307e-46 ], [ 831.125 , 1.54023669773311e-46 ], [ 831.1875 , 1.5404254734878537e-46 ], 
                    [ 831.25 , 1.5406142723794756e-46 ], [ 831.3125 , 1.540803094410833e-46 ], [ 831.375 , 1.5409919395847298e-46 ], [ 831.4375 , 1.541180674316508e-46 ], 
                    [ 831.5 , 1.5413693020202771e-46 ], [ 831.5625 , 1.5415579528105094e-46 ], [ 831.625 , 1.541746626690041e-46 ], [ 831.6875 , 1.5419353236616762e-46 ], 
                    [ 831.75 , 1.5421240437282522e-46 ], [ 831.8125 , 1.5423127868925848e-46 ], [ 831.875 , 1.5425015531575335e-46 ], [ 831.9375 , 1.5426903425258934e-46 ], 
                    [ 832.0 , 1.5428791550004914e-46 ], [ 832.0625 , 1.5430679905841998e-46 ], [ 832.125 , 1.5432568492797813e-46 ], [ 832.1875 , 1.543445731090119e-46 ], 
                    [ 832.25 , 1.5436346360179984e-46 ], [ 832.3125 , 1.5438235640662817e-46 ], [ 832.375 , 1.5440125152378095e-46 ], [ 832.4375 , 1.5442014895353572e-46 ], 
                    [ 832.5 , 1.5443904869617772e-46 ], [ 832.5625 , 1.544579507519955e-46 ], [ 832.625 , 1.544768551212656e-46 ], [ 832.6875 , 1.5449576180427227e-46 ], 
                    [ 832.75 , 1.5451467080130306e-46 ], [ 832.8125 , 1.5453358211263464e-46 ], [ 832.875 , 1.5455249573855352e-46 ], [ 832.9375 , 1.545714116793452e-46 ], 
                    [ 833.0 , 1.5459032993529077e-46 ], [ 833.0625 , 1.5460925050667144e-46 ], [ 833.125 , 1.5462817339377386e-46 ], [ 833.1875 , 1.5464709859688148e-46 ], 
                    [ 833.25 , 1.5466602611627334e-46 ], [ 833.3125 , 1.546849466479825e-46 ], [ 833.375 , 1.547038525547505e-46 ], [ 833.4375 , 1.5472276077223769e-46 ], 
                    [ 833.5 , 1.5474167130072434e-46 ], [ 833.5625 , 1.54760584140494e-46 ], [ 833.625 , 1.5477949929183027e-46 ], [ 833.6875 , 1.5479841675501344e-46 ], 
                    [ 833.75 , 1.548173365303272e-46 ], [ 833.8125 , 1.5483625861805523e-46 ], [ 833.875 , 1.5485518301847906e-46 ], [ 833.9375 , 1.5487410973188027e-46 ], 
                    [ 834.0 , 1.548930387585426e-46 ], [ 834.0625 , 1.5491197009874996e-46 ], [ 834.125 , 1.5493090375278175e-46 ], [ 834.1875 , 1.5494983972092632e-46 ], 
                    [ 834.25 , 1.5496877800346096e-46 ], [ 834.3125 , 1.5498771860067078e-46 ], [ 834.375 , 1.5500666151283864e-46 ], [ 834.4375 , 1.5502560674024974e-46 ], 
                    [ 834.5 , 1.550445542831837e-46 ], [ 834.5625 , 1.5506350414192242e-46 ], [ 834.625 , 1.5508245631675342e-46 ], [ 834.6875 , 1.551014108079553e-46 ], 
                    [ 834.75 , 1.5512036761581559e-46 ], [ 834.8125 , 1.5513932674061306e-46 ], [ 834.875 , 1.5515828818263748e-46 ], [ 834.9375 , 1.5517725194216435e-46 ], 
                    [ 835.0 , 1.5519621801948026e-46 ], [ 835.0625 , 1.5521518641486954e-46 ], [ 835.125 , 1.5523415712861332e-46 ], [ 835.1875 , 1.5525312654436983e-46 ], 
                    [ 835.25 , 1.5527207495044084e-46 ], [ 835.3125 , 1.5529102566913355e-46 ], [ 835.375 , 1.5530997870073574e-46 ], [ 835.4375 , 1.5532893404552529e-46 ], 
                    [ 835.5 , 1.553478917037845e-46 ], [ 835.5625 , 1.5536685167579903e-46 ], [ 835.625 , 1.5538581396184686e-46 ], [ 835.6875 , 1.5540477856221375e-46 ], 
                    [ 835.75 , 1.5542374547718105e-46 ], [ 835.8125 , 1.5544271470703122e-46 ], [ 835.875 , 1.5546168625204573e-46 ], [ 835.9375 , 1.554806601125093e-46 ], 
                    [ 836.0 , 1.554996362887035e-46 ], [ 836.0625 , 1.5551861478090977e-46 ], [ 836.125 , 1.5553759558941301e-46 ], [ 836.1875 , 1.5555657871449485e-46 ], 
                    [ 836.25 , 1.5557556415643799e-46 ], [ 836.3125 , 1.555945519155241e-46 ], [ 836.375 , 1.5561354199203933e-46 ], [ 836.4375 , 1.5563253438626205e-46 ], 
                    [ 836.5 , 1.5565152909847962e-46 ], [ 836.5625 , 1.5567052612897043e-46 ], [ 836.625 , 1.5568952547801972e-46 ], [ 836.6875 , 1.5570852714591042e-46 ], 
                    [ 836.75 , 1.5572753113292888e-46 ], [ 836.8125 , 1.5574653743935147e-46 ], [ 836.875 , 1.5576554606546688e-46 ], [ 836.9375 , 1.5578455701155262e-46 ], 
                    [ 837.0 , 1.558035702778974e-46 ], [ 837.0625 , 1.558225858647822e-46 ], [ 837.125 , 1.558415811955337e-46 ], [ 837.1875 , 1.5586057471394807e-46 ], 
                    [ 837.25 , 1.5587957054723938e-46 ], [ 837.3125 , 1.5589856869568648e-46 ], [ 837.375 , 1.5591756915957152e-46 ], [ 837.4375 , 1.5593657193917891e-46 ], 
                    [ 837.5 , 1.55955577034792e-46 ], [ 837.5625 , 1.5597458444668751e-46 ], [ 837.625 , 1.5599359417515109e-46 ], [ 837.6875 , 1.5601260622046727e-46 ], 
                    [ 837.75 , 1.5603162058291395e-46 ], [ 837.8125 , 1.5605063726278026e-46 ], [ 837.875 , 1.5606965626033974e-46 ], [ 837.9375 , 1.560886775758793e-46 ], 
                    [ 838.0 , 1.5610770120968255e-46 ], [ 838.0625 , 1.5612672716203315e-46 ], [ 838.125 , 1.561457554332081e-46 ], [ 838.1875 , 1.5616478602349562e-46 ], 
                    [ 838.25 , 1.5618381893317386e-46 ], [ 838.3125 , 1.562028541625289e-46 ], [ 838.375 , 1.5622189171184226e-46 ], [ 838.4375 , 1.5624093158139675e-46 ], 
                    [ 838.5 , 1.5625997377147625e-46 ], [ 838.5625 , 1.5627901828236132e-46 ], [ 838.625 , 1.5629806511433487e-46 ], [ 838.6875 , 1.5631711426768199e-46 ], 
                    [ 838.75 , 1.5633616574268559e-46 ], [ 838.8125 , 1.5635521953962642e-46 ], [ 838.875 , 1.5637427565878855e-46 ], [ 838.9375 , 1.5639333410045616e-46 ], 
                    [ 839.0 , 1.5641238154354358e-46 ], [ 839.0625 , 1.5643141731928253e-46 ], [ 839.125 , 1.564504554117205e-46 ], [ 839.1875 , 1.564694958211439e-46 ], 
                    [ 839.25 , 1.5648853854783133e-46 ], [ 839.3125 , 1.5650758359206707e-46 ], [ 839.375 , 1.5652663095413092e-46 ], [ 839.4375 , 1.565456806343061e-46 ], 
                    [ 839.5 , 1.5656473263287697e-46 ], [ 839.5625 , 1.565837869501212e-46 ], [ 839.625 , 1.566028435863243e-46 ], [ 839.6875 , 1.566219025417652e-46 ], 
                    [ 839.75 , 1.566409638167306e-46 ], [ 839.8125 , 1.5666002741150053e-46 ], [ 839.875 , 1.5667909332635737e-46 ], [ 839.9375 , 1.5669816156158012e-46 ], 
                    [ 840.0 , 1.5671723211745783e-46 ], [ 840.0625 , 1.5673630499427074e-46 ], [ 840.125 , 1.5675538019229796e-46 ], [ 840.1875 , 1.567744577118253e-46 ], 
                    [ 840.25 , 1.5679353755313308e-46 ], [ 840.3125 , 1.5681261971650612e-46 ], [ 840.375 , 1.5683170420222477e-46 ], [ 840.4375 , 1.5685079101057278e-46 ], 
                    [ 840.5 , 1.5686988014183394e-46 ], [ 840.5625 , 1.5688897159628983e-46 ], [ 840.625 , 1.569080653742221e-46 ], [ 840.6875 , 1.569271614759146e-46 ], 
                    [ 840.75 , 1.5694625990165017e-46 ], [ 840.8125 , 1.5696536065171387e-46 ], [ 840.875 , 1.5698446207521584e-46 ], [ 840.9375 , 1.5700354066665405e-46 ], 
                    [ 841.0 , 1.5702262157674628e-46 ], [ 841.0625 , 1.5704170480577319e-46 ], [ 841.125 , 1.5706079035401882e-46 ], [ 841.1875 , 1.5707987822176393e-46 ], 
                    [ 841.25 , 1.5709896840928928e-46 ], [ 841.3125 , 1.5711806091687684e-46 ], [ 841.375 , 1.57137155744813e-46 ], [ 841.4375 , 1.5715625289337415e-46 ], 
                    [ 841.5 , 1.5717535236284461e-46 ], [ 841.5625 , 1.571944541535075e-46 ], [ 841.625 , 1.5721355826564386e-46 ], [ 841.6875 , 1.5723266469953466e-46 ], 
                    [ 841.75 , 1.5725177345546433e-46 ], [ 841.8125 , 1.5727088453371284e-46 ], [ 841.875 , 1.5728999793456463e-46 ], [ 841.9375 , 1.5730911365830087e-46 ], 
                    [ 842.0 , 1.5732823170520275e-46 ], [ 842.0625 , 1.5734735207555374e-46 ], [ 842.125 , 1.5736647476963616e-46 ], [ 842.1875 , 1.5738559978773472e-46 ], 
                    [ 842.25 , 1.5740472713012848e-46 ], [ 842.3125 , 1.5742385679709877e-46 ], [ 842.375 , 1.5744298878893372e-46 ], [ 842.4375 , 1.5746212310591025e-46 ], 
                    [ 842.5 , 1.5748125974831326e-46 ], [ 842.5625 , 1.575003987164275e-46 ], [ 842.625 , 1.575195400105301e-46 ], [ 842.6875 , 1.5753868363090702e-46 ], 
                    [ 842.75 , 1.5755782957784441e-46 ], [ 842.8125 , 1.575769623896155e-46 ], [ 842.875 , 1.575960856788679e-46 ], [ 842.9375 , 1.5761521128889112e-46 ], 
                    [ 843.0 , 1.5763433921996903e-46 ], [ 843.0625 , 1.5765346947238107e-46 ], [ 843.125 , 1.5767260204641123e-46 ], [ 843.1875 , 1.5769173694233897e-46 ], 
                    [ 843.25 , 1.5771087416044611e-46 ], [ 843.3125 , 1.5773001370101895e-46 ], [ 843.375 , 1.5774915556433478e-46 ], [ 843.4375 , 1.5776829975067557e-46 ], 
                    [ 843.5 , 1.5778744626032658e-46 ], [ 843.5625 , 1.5780659509356413e-46 ], [ 843.625 , 1.57825746250677e-46 ], [ 843.6875 , 1.5784489973194038e-46 ], 
                    [ 843.75 , 1.5786405553764091e-46 ], [ 843.8125 , 1.5788321366806175e-46 ], [ 843.875 , 1.579023741234817e-46 ], [ 843.9375 , 1.5792153690418171e-46 ], 
                    [ 844.0 , 1.579407020104497e-46 ], [ 844.0625 , 1.5795986944256329e-46 ], [ 844.125 , 1.5797903920080371e-46 ], [ 844.1875 , 1.5799821128545885e-46 ], 
                    [ 844.25 , 1.580173856968065e-46 ], [ 844.3125 , 1.5803656243512795e-46 ], [ 844.375 , 1.5805574150070896e-46 ], [ 844.4375 , 1.5807492289383195e-46 ], 
                    [ 844.5 , 1.5809410661477938e-46 ], [ 844.5625 , 1.581132926638304e-46 ], [ 844.625 , 1.5813248104126866e-46 ], [ 844.6875 , 1.5815167169571299e-46 ], 
                    [ 844.75 , 1.5817083781669164e-46 ], [ 844.8125 , 1.5819000626037967e-46 ], [ 844.875 , 1.5820917702705744e-46 ], [ 844.9375 , 1.582283501170076e-46 ], 
                    [ 845.0 , 1.5824752553051057e-46 ], [ 845.0625 , 1.5826670326784455e-46 ], [ 845.125 , 1.5828588332929794e-46 ], [ 845.1875 , 1.5830506571514676e-46 ], 
                    [ 845.25 , 1.5832425042567717e-46 ], [ 845.3125 , 1.5834343746116644e-46 ], [ 845.375 , 1.583626268219008e-46 ], [ 845.4375 , 1.5838181850815756e-46 ], 
                    [ 845.5 , 1.5840101252022303e-46 ], [ 845.5625 , 1.5842020885837458e-46 ], [ 845.625 , 1.5843940752289861e-46 ], [ 845.6875 , 1.5845860851407252e-46 ], 
                    [ 845.75 , 1.5847781183218169e-46 ], [ 845.8125 , 1.5849701747750694e-46 ], [ 845.875 , 1.5851622545033033e-46 ], [ 845.9375 , 1.585354357509317e-46 ], 
                    [ 846.0 , 1.5855464837959874e-46 ], [ 846.0625 , 1.5857386333660797e-46 ], [ 846.125 , 1.5859308062224607e-46 ], [ 846.1875 , 1.5861230023679184e-46 ], 
                    [ 846.25 , 1.5863152218052865e-46 ], [ 846.3125 , 1.5865074645373878e-46 ], [ 846.375 , 1.5866997305670566e-46 ], [ 846.4375 , 1.5868920198970824e-46 ], 
                    [ 846.5 , 1.5870843325303343e-46 ], [ 846.5625 , 1.5872766684696023e-46 ], [ 846.625 , 1.5874689254655133e-46 ], [ 846.6875 , 1.5876610302407731e-46 ], 
                    [ 846.75 , 1.5878531582632564e-46 ], [ 846.8125 , 1.5880453095357875e-46 ], [ 846.875 , 1.5882374840611575e-46 ], [ 846.9375 , 1.5884296818422027e-46 ], 
                    [ 847.0 , 1.5886219028817036e-46 ], [ 847.0625 , 1.5888141471825312e-46 ], [ 847.125 , 1.5890064147474441e-46 ], [ 847.1875 , 1.5891987055792915e-46 ], 
                    [ 847.25 , 1.589391019680866e-46 ], [ 847.3125 , 1.5895833570549953e-46 ], [ 847.375 , 1.5897757177045068e-46 ], [ 847.4375 , 1.5899681016322056e-46 ], 
                    [ 847.5 , 1.5901605088408978e-46 ], [ 847.5625 , 1.5903529393334003e-46 ], [ 847.625 , 1.590545393112565e-46 ], [ 847.6875 , 1.5907378701811988e-46 ], 
                    [ 847.75 , 1.5909303705420854e-46 ], [ 847.8125 , 1.5911228941980552e-46 ], [ 847.875 , 1.591315441151961e-46 ], [ 847.9375 , 1.5915080114066113e-46 ], 
                    [ 848.0 , 1.5917006049648024e-46 ], [ 848.0625 , 1.5918932218293436e-46 ], [ 848.125 , 1.5920858620030892e-46 ], [ 848.1875 , 1.5922785254888481e-46 ], 
                    [ 848.25 , 1.5924712122894646e-46 ], [ 848.3125 , 1.5926639224077147e-46 ], [ 848.375 , 1.5928566558464423e-46 ], [ 848.4375 , 1.5930494126084815e-46 ], 
                    [ 848.5 , 1.5932421926966319e-46 ], [ 848.5625 , 1.5934348139991184e-46 ], [ 848.625 , 1.5936273668856898e-46 ], [ 848.6875 , 1.5938199430406209e-46 ], 
                    [ 848.75 , 1.594012542466723e-46 ], [ 848.8125 , 1.594205165166809e-46 ], [ 848.875 , 1.5943978111436907e-46 ], [ 848.9375 , 1.5945904804001695e-46 ], 
                    [ 849.0 , 1.594783172939093e-46 ], [ 849.0625 , 1.5949758887632516e-46 ], [ 849.125 , 1.5951686278754256e-46 ], [ 849.1875 , 1.5953613902784855e-46 ], 
                    [ 849.25 , 1.5955541759752343e-46 ], [ 849.3125 , 1.595746984968476e-46 ], [ 849.375 , 1.5959398172610027e-46 ], [ 849.4375 , 1.5961326728556641e-46 ], 
                    [ 849.5 , 1.5963255517552535e-46 ], [ 849.5625 , 1.596518453962632e-46 ], [ 849.625 , 1.5967113794805484e-46 ], [ 849.6875 , 1.5969043283118538e-46 ], 
                    [ 849.75 , 1.5970973004593994e-46 ], [ 849.8125 , 1.5972902959259458e-46 ], [ 849.875 , 1.5974833147143336e-46 ], [ 849.9375 , 1.5976763568273697e-46 ], 
                    [ 850.0 , 1.5978694222679185e-46 ], [ 850.0625 , 1.5980625110387306e-46 ], [ 850.125 , 1.5982556231426815e-46 ], [ 850.1875 , 1.5984487585825346e-46 ], 
                    [ 850.25 , 1.5986419173611893e-46 ], [ 850.3125 , 1.5988350994813977e-46 ], [ 850.375 , 1.5990283049459914e-46 ], [ 850.4375 , 1.599221533757826e-46 ], 
                    [ 850.5 , 1.5994145339136608e-46 ], [ 850.5625 , 1.5996075269472187e-46 ], [ 850.625 , 1.5998005432682314e-46 ], [ 850.6875 , 1.5999935828795308e-46 ], 
                    [ 850.75 , 1.6001866457839273e-46 ], [ 850.8125 , 1.6003797319841865e-46 ], [ 850.875 , 1.600572841483176e-46 ], [ 850.9375 , 1.600765974283673e-46 ], 
                    [ 851.0 , 1.600959130388512e-46 ], [ 851.0625 , 1.6011523098004708e-46 ], [ 851.125 , 1.60134551252243e-46 ], [ 851.1875 , 1.6015387385571344e-46 ], 
                    [ 851.25 , 1.6017319879074077e-46 ], [ 851.3125 , 1.601925260576064e-46 ], [ 851.375 , 1.602118556565962e-46 ], [ 851.4375 , 1.6023118758798478e-46 ], 
                    [ 851.5 , 1.6025052185206039e-46 ], [ 851.5625 , 1.6026985844909995e-46 ], [ 851.625 , 1.6028919737938723e-46 ], [ 851.6875 , 1.603085386431981e-46 ], 
                    [ 851.75 , 1.6032788224082213e-46 ], [ 851.8125 , 1.6034722817253746e-46 ], [ 851.875 , 1.6036657643862576e-46 ], [ 851.9375 , 1.6038592703936868e-46 ], 
                    [ 852.0 , 1.6040527997504685e-46 ], [ 852.0625 , 1.604246352459431e-46 ], [ 852.125 , 1.604439928523415e-46 ], [ 852.1875 , 1.6046335279451817e-46 ], 
                    [ 852.25 , 1.6048271507276182e-46 ], [ 852.3125 , 1.6050207968734858e-46 ], [ 852.375 , 1.6052144527175872e-46 ], [ 852.4375 , 1.6054078717651734e-46 ], 
                    [ 852.5 , 1.6056013141186461e-46 ], [ 852.5625 , 1.605794779780814e-46 ], [ 852.625 , 1.6059882687544397e-46 ], [ 852.6875 , 1.6061817810424007e-46 ], 
                    [ 852.75 , 1.6063753166474491e-46 ], [ 852.8125 , 1.606568875572429e-46 ], [ 852.875 , 1.6067624578201499e-46 ], [ 852.9375 , 1.606956063393388e-46 ], 
                    [ 853.0 , 1.6071496922949881e-46 ], [ 853.0625 , 1.6073433445277501e-46 ], [ 853.125 , 1.6075370200944731e-46 ], [ 853.1875 , 1.6077307189979921e-46 ], 
                    [ 853.25 , 1.6079244412411192e-46 ], [ 853.3125 , 1.6081181868266316e-46 ], [ 853.375 , 1.6083119557573656e-46 ], [ 853.4375 , 1.6085057480361568e-46 ], 
                    [ 853.5 , 1.608699563665784e-46 ], [ 853.5625 , 1.6088934026490609e-46 ], [ 853.625 , 1.6090872649888246e-46 ], [ 853.6875 , 1.6092811506879007e-46 ], 
                    [ 853.75 , 1.6094750597490349e-46 ], [ 853.8125 , 1.6096689921751227e-46 ], [ 853.875 , 1.609862947968911e-46 ], [ 853.9375 , 1.6100569271332833e-46 ], 
                    [ 854.0 , 1.6102509296709879e-46 ], [ 854.0625 , 1.6104449555849206e-46 ], [ 854.125 , 1.6106390048778297e-46 ], [ 854.1875 , 1.610833077552555e-46 ], 
                    [ 854.25 , 1.6110271736119027e-46 ], [ 854.3125 , 1.6112212541004412e-46 ], [ 854.375 , 1.6114151134421323e-46 ], [ 854.4375 , 1.6116089961086424e-46 ], 
                    [ 854.5 , 1.6118029021027782e-46 ], [ 854.5625 , 1.6119968314273574e-46 ], [ 854.625 , 1.6121907840851648e-46 ], [ 854.6875 , 1.6123847600790075e-46 ], 
                    [ 854.75 , 1.6125787594117046e-46 ], [ 854.8125 , 1.6127727820860875e-46 ], [ 854.875 , 1.6129668281049298e-46 ], [ 854.9375 , 1.613160897471052e-46 ], 
                    [ 855.0 , 1.6133549901872743e-46 ], [ 855.0625 , 1.6135491062563723e-46 ], [ 855.125 , 1.6137432456811897e-46 ], [ 855.1875 , 1.613937408464514e-46 ], 
                    [ 855.25 , 1.6141315946092017e-46 ], [ 855.3125 , 1.614325804118006e-46 ], [ 855.375 , 1.6145200369937606e-46 ], [ 855.4375 , 1.6147142932392545e-46 ], 
                    [ 855.5 , 1.6149085728573337e-46 ], [ 855.5625 , 1.6151028758508334e-46 ], [ 855.625 , 1.615297202222509e-46 ], [ 855.6875 , 1.6154915519752071e-46 ], 
                    [ 855.75 , 1.615685925111707e-46 ], [ 855.8125 , 1.6158803216348564e-46 ], [ 855.875 , 1.6160747415474576e-46 ], [ 855.9375 , 1.6162691848523135e-46 ], 
                    [ 856.0 , 1.6164636515522844e-46 ], [ 856.0625 , 1.6166581416501048e-46 ], [ 856.125 , 1.6168526551486704e-46 ], [ 856.1875 , 1.6170471920507504e-46 ], 
                    [ 856.25 , 1.6172417098952131e-46 ], [ 856.3125 , 1.6174360196843632e-46 ], [ 856.375 , 1.6176303528196177e-46 ], [ 856.4375 , 1.6178247093037704e-46 ], 
                    [ 856.5 , 1.6180190891396612e-46 ], [ 856.5625 , 1.618213492330061e-46 ], [ 856.625 , 1.6184079188777758e-46 ], [ 856.6875 , 1.6186023687856354e-46 ], 
                    [ 856.75 , 1.6187968420564346e-46 ], [ 856.8125 , 1.6189913386929808e-46 ], [ 856.875 , 1.6191858586980807e-46 ], [ 856.9375 , 1.6193804020745428e-46 ], 
                    [ 857.0 , 1.6195749688251746e-46 ], [ 857.0625 , 1.6197695589527963e-46 ], [ 857.125 , 1.6199641724601818e-46 ], [ 857.1875 , 1.6201588093501866e-46 ], 
                    [ 857.25 , 1.620353469625585e-46 ], [ 857.3125 , 1.6205481532891873e-46 ], [ 857.375 , 1.6207428603438263e-46 ], [ 857.4375 , 1.6209375907923123e-46 ], 
                    [ 857.5 , 1.6211323446374216e-46 ], [ 857.5625 , 1.621327121882e-46 ], [ 857.625 , 1.6215219225288587e-46 ], [ 857.6875 , 1.6217167465807752e-46 ], 
                    [ 857.75 , 1.6219115940405957e-46 ], [ 857.8125 , 1.6221064649111216e-46 ], [ 857.875 , 1.6223013591951653e-46 ], [ 857.9375 , 1.6224962768955285e-46 ], 
                    [ 858.0 , 1.622691218015059e-46 ], [ 858.0625 , 1.6228861825565367e-46 ], [ 858.125 , 1.6230811705227864e-46 ], [ 858.1875 , 1.6232761527823215e-46 ], 
                    [ 858.25 , 1.6234709030540776e-46 ], [ 858.3125 , 1.6236656766907348e-46 ], [ 858.375 , 1.6238604736950967e-46 ], [ 858.4375 , 1.6240552940699319e-46 ], 
                    [ 858.5 , 1.624250137818079e-46 ], [ 858.5625 , 1.6244450049423305e-46 ], [ 858.625 , 1.6246398954454794e-46 ], [ 858.6875 , 1.6248348093303653e-46 ], 
                    [ 858.75 , 1.6250297465997468e-46 ], [ 858.8125 , 1.6252247072564533e-46 ], [ 858.875 , 1.625419691303313e-46 ], [ 858.9375 , 1.6256146987430863e-46 ], 
                    [ 859.0 , 1.625809729578614e-46 ], [ 859.0625 , 1.6260047838126804e-46 ], [ 859.125 , 1.6261998614481157e-46 ], [ 859.1875 , 1.626394962487727e-46 ], 
                    [ 859.25 , 1.6265900869342996e-46 ], [ 859.3125 , 1.6267852347906414e-46 ], [ 859.375 , 1.6269804060595726e-46 ], [ 859.4375 , 1.6271756007439255e-46 ], 
                    [ 859.5 , 1.6273708188464623e-46 ], [ 859.5625 , 1.6275660603700282e-46 ], [ 859.625 , 1.6277613253174207e-46 ], [ 859.6875 , 1.6279566136914508e-46 ], 
                    [ 859.75 , 1.6281519254949283e-46 ], [ 859.8125 , 1.6283472607306649e-46 ], [ 859.875 , 1.628542619401483e-46 ], [ 859.9375 , 1.628738001510148e-46 ], 
                    [ 860.0 , 1.6289334070595295e-46 ], [ 860.0625 , 1.6291288360524052e-46 ], [ 860.125 , 1.6293242884916109e-46 ], [ 860.1875 , 1.6295194879678238e-46 ], 
                    [ 860.25 , 1.62971470500363e-46 ], [ 860.3125 , 1.6299099454265095e-46 ], [ 860.375 , 1.6301052092392636e-46 ], [ 860.4375 , 1.6303004964446827e-46 ], 
                    [ 860.5 , 1.630495807045593e-46 ], [ 860.5625 , 1.630691141044785e-46 ], [ 860.625 , 1.6308864984450509e-46 ], [ 860.6875 , 1.6310818792492166e-46 ], 
                    [ 860.75 , 1.631277283460052e-46 ], [ 860.8125 , 1.6314727110804184e-46 ], [ 860.875 , 1.6316681621130628e-46 ], [ 860.9375 , 1.6318636365608245e-46 ], 
                    [ 861.0 , 1.6320591344264972e-46 ], [ 861.0625 , 1.632254655712863e-46 ], [ 861.125 , 1.6324502004227743e-46 ], [ 861.1875 , 1.632645768559014e-46 ], 
                    [ 861.25 , 1.6328413601244005e-46 ], [ 861.3125 , 1.6330369751216817e-46 ], [ 861.375 , 1.6332326135537468e-46 ], [ 861.4375 , 1.6334282754233683e-46 ], 
                    [ 861.5 , 1.6336239607333424e-46 ], [ 861.5625 , 1.6338196694864888e-46 ], [ 861.625 , 1.6340154016856391e-46 ], [ 861.6875 , 1.6342111573335677e-46 ], 
                    [ 861.75 , 1.6344069364330715e-46 ], [ 861.8125 , 1.6346027389869959e-46 ], [ 861.875 , 1.6347985649981727e-46 ], [ 861.9375 , 1.6349944144693434e-46 ], 
                    [ 862.0 , 1.6351902874033412e-46 ], [ 862.0625 , 1.6353861838030236e-46 ], [ 862.125 , 1.6355818792383163e-46 ], [ 862.1875 , 1.6357775368914856e-46 ], 
                    [ 862.25 , 1.6359732179503536e-46 ], [ 862.3125 , 1.6361689224176963e-46 ], [ 862.375 , 1.6363646502963377e-46 ], [ 862.4375 , 1.636560401589067e-46 ], 
                    [ 862.5 , 1.636756176298673e-46 ], [ 862.5625 , 1.636951974427969e-46 ], [ 862.625 , 1.637147795979756e-46 ], [ 862.6875 , 1.6373436409568598e-46 ], 
                    [ 862.75 , 1.637539509362036e-46 ], [ 862.8125 , 1.6377354011981102e-46 ], [ 862.875 , 1.6379313164678857e-46 ], [ 862.9375 , 1.6381272551741769e-46 ], 
                    [ 863.0 , 1.6383232173197649e-46 ], [ 863.0625 , 1.6385192029074645e-46 ], [ 863.125 , 1.6387152119400923e-46 ], [ 863.1875 , 1.638911244420429e-46 ], 
                    [ 863.25 , 1.6391073003513034e-46 ], [ 863.3125 , 1.6393033797355087e-46 ], [ 863.375 , 1.6394994825758392e-46 ], [ 863.4375 , 1.6396956088751123e-46 ], 
                    [ 863.5 , 1.6398917586361459e-46 ], [ 863.5625 , 1.6400879318617351e-46 ], [ 863.625 , 1.6402841285546753e-46 ], [ 863.6875 , 1.6404803487177854e-46 ], 
                    [ 863.75 , 1.6406765923538727e-46 ], [ 863.8125 , 1.6408728594657572e-46 ], [ 863.875 , 1.6410691500562358e-46 ], [ 863.9375 , 1.6412654641281048e-46 ], 
                    [ 864.0 , 1.6414618016841969e-46 ], [ 864.0625 , 1.6416580125737802e-46 ], [ 864.125 , 1.6418541090949896e-46 ], [ 864.1875 , 1.6420502290399847e-46 ], 
                    [ 864.25 , 1.6422463724115867e-46 ], [ 864.3125 , 1.6424425392125476e-46 ], [ 864.375 , 1.6426387294457012e-46 ], [ 864.4375 , 1.6428349431138224e-46 ], 
                    [ 864.5 , 1.643031180219746e-46 ], [ 864.5625 , 1.6432274407662369e-46 ], [ 864.625 , 1.643423724756083e-46 ], [ 864.6875 , 1.643620032192143e-46 ], 
                    [ 864.75 , 1.6438163630771594e-46 ], [ 864.8125 , 1.6440127174139565e-46 ], [ 864.875 , 1.644209095205347e-46 ], [ 864.9375 , 1.6444054964541328e-46 ], 
                    [ 865.0 , 1.644601921163081e-46 ], [ 865.0625 , 1.6447983693350286e-46 ], [ 865.125 , 1.6449948409727551e-46 ], [ 865.1875 , 1.6451913360790986e-46 ], 
                    [ 865.25 , 1.6453878546567923e-46 ], [ 865.3125 , 1.645584396708745e-46 ], [ 865.375 , 1.6457809622376788e-46 ], [ 865.4375 , 1.6459775512464097e-46 ], 
                    [ 865.5 , 1.6461741637377778e-46 ], [ 865.5625 , 1.646370799714541e-46 ], [ 865.625 , 1.6465674591795397e-46 ], [ 865.6875 , 1.6467641421355681e-46 ], 
                    [ 865.75 , 1.6469608485854088e-46 ], [ 865.8125 , 1.6471575785319031e-46 ], [ 865.875 , 1.6473543319778696e-46 ], [ 865.9375 , 1.6475511089260677e-46 ], 
                    [ 866.0 , 1.6477478556963634e-46 ], [ 866.0625 , 1.647944391370488e-46 ], [ 866.125 , 1.6481409504864471e-46 ], [ 866.1875 , 1.6483375330470955e-46 ], 
                    [ 866.25 , 1.6485341390551589e-46 ], [ 866.3125 , 1.648730768513493e-46 ], [ 866.375 , 1.6489274214248596e-46 ], [ 866.4375 , 1.6491240977920557e-46 ], 
                    [ 866.5 , 1.6493207976179028e-46 ], [ 866.5625 , 1.6495175209052106e-46 ], [ 866.625 , 1.6497142676567187e-46 ], [ 866.6875 , 1.6499110378752964e-46 ], 
                    [ 866.75 , 1.6501078315636605e-46 ], [ 866.8125 , 1.6503046487247043e-46 ], [ 866.875 , 1.650501489361145e-46 ], [ 866.9375 , 1.650698353475842e-46 ], 
                    [ 867.0 , 1.6508952410715718e-46 ], [ 867.0625 , 1.6510921521510997e-46 ], [ 867.125 , 1.6512890867172858e-46 ], [ 867.1875 , 1.6514860447729196e-46 ], 
                    [ 867.25 , 1.6516830263207793e-46 ], [ 867.3125 , 1.651880031363702e-46 ], [ 867.375 , 1.6520770599044434e-46 ], [ 867.4375 , 1.6522741119458064e-46 ], 
                    [ 867.5 , 1.652471187490652e-46 ], [ 867.5625 , 1.6526682865417609e-46 ], [ 867.625 , 1.6528654091018894e-46 ], [ 867.6875 , 1.6530625551738651e-46 ], 
                    [ 867.75 , 1.6532597247605513e-46 ], [ 867.8125 , 1.6534569178647057e-46 ], [ 867.875 , 1.6536541344890861e-46 ], [ 867.9375 , 1.6538513746365685e-46 ], 
                    [ 868.0 , 1.6540484156302013e-46 ], [ 868.0625 , 1.6542454173433718e-46 ], [ 868.125 , 1.6544424425200229e-46 ], [ 868.1875 , 1.6546394911629025e-46 ], 
                    [ 868.25 , 1.6548365632748053e-46 ], [ 868.3125 , 1.65503365885855e-46 ], [ 868.375 , 1.6552307779169093e-46 ], [ 868.4375 , 1.6554279204527134e-46 ], 
                    [ 868.5 , 1.6556250864687239e-46 ], [ 868.5625 , 1.655822275967737e-46 ], [ 868.625 , 1.6560194889525616e-46 ], [ 868.6875 , 1.6562167254260183e-46 ], 
                    [ 868.75 , 1.6564139853908578e-46 ], [ 868.8125 , 1.6566112688499246e-46 ], [ 868.875 , 1.6568085758059704e-46 ], [ 868.9375 , 1.6570059062618287e-46 ], 
                    [ 869.0 , 1.6572032602202869e-46 ], [ 869.0625 , 1.657400637684144e-46 ], [ 869.125 , 1.6575980386561994e-46 ], [ 869.1875 , 1.6577954631392415e-46 ], 
                    [ 869.25 , 1.6579929111360943e-46 ], [ 869.3125 , 1.6581903826495343e-46 ], [ 869.375 , 1.6583878776823864e-46 ], [ 869.4375 , 1.6585853962374395e-46 ], 
                    [ 869.5 , 1.6587829383174838e-46 ], [ 869.5625 , 1.6589805039253447e-46 ], [ 869.625 , 1.6591780930638127e-46 ], [ 869.6875 , 1.65937570573569e-46 ], 
                    [ 869.75 , 1.65957334194378e-46 ], [ 869.8125 , 1.6597710016908618e-46 ], [ 869.875 , 1.6599686849797983e-46 ], [ 869.9375 , 1.6601663039100272e-46 ], 
                    [ 870.0 , 1.660363743867702e-46 ], [ 870.0625 , 1.6605612073064565e-46 ], [ 870.125 , 1.660758694229131e-46 ], [ 870.1875 , 1.660956204638494e-46 ], 
                    [ 870.25 , 1.6611537385373279e-46 ], [ 870.3125 , 1.6613512959284374e-46 ], [ 870.375 , 1.66154887681464e-46 ], [ 870.4375 , 1.661746481198683e-46 ], 
                    [ 870.5 , 1.6619441090833846e-46 ], [ 870.5625 , 1.6621417604715392e-46 ], [ 870.625 , 1.6623394353659425e-46 ], [ 870.6875 , 1.6625371337693897e-46 ], 
                    [ 870.75 , 1.6627348556846887e-46 ], [ 870.8125 , 1.662932601114612e-46 ], [ 870.875 , 1.66313037006198e-46 ], [ 870.9375 , 1.6633281625295656e-46 ], 
                    [ 871.0 , 1.6635259785202017e-46 ], [ 871.0625 , 1.6637238180366028e-46 ], [ 871.125 , 1.6639216810816855e-46 ], [ 871.1875 , 1.6641195676581767e-46 ], 
                    [ 871.25 , 1.6643174777688632e-46 ], [ 871.3125 , 1.6645154114166027e-46 ], [ 871.375 , 1.6647133686041242e-46 ], [ 871.4375 , 1.6649113493342856e-46 ], 
                    [ 871.5 , 1.6651093536098638e-46 ], [ 871.5625 , 1.6653073814336343e-46 ], [ 871.625 , 1.6655054328084344e-46 ], [ 871.6875 , 1.6657035077370642e-46 ], 
                    [ 871.75 , 1.6659016062222896e-46 ], [ 871.8125 , 1.6660997282669712e-46 ], [ 871.875 , 1.6662978738738521e-46 ], [ 871.9375 , 1.666495823093273e-46 ], 
                    [ 872.0 , 1.6666937240271017e-46 ], [ 872.0625 , 1.6668916484621834e-46 ], [ 872.125 , 1.667089596401333e-46 ], [ 872.1875 , 1.6672875678473532e-46 ], 
                    [ 872.25 , 1.6674855628030118e-46 ], [ 872.3125 , 1.6676835812711243e-46 ], [ 872.375 , 1.6678816232544473e-46 ], [ 872.4375 , 1.668079688755833e-46 ], 
                    [ 872.5 , 1.6682777777780144e-46 ], [ 872.5625 , 1.6684758903238206e-46 ], [ 872.625 , 1.6686740263960328e-46 ], [ 872.6875 , 1.6688721859974216e-46 ], 
                    [ 872.75 , 1.6690703691308284e-46 ], [ 872.8125 , 1.6692685757990238e-46 ], [ 872.875 , 1.6694668060047795e-46 ], [ 872.9375 , 1.669665059750961e-46 ], 
                    [ 873.0 , 1.6698633370402694e-46 ], [ 873.0625 , 1.6700616378755476e-46 ], [ 873.125 , 1.6702599622596397e-46 ], [ 873.1875 , 1.6704583101952467e-46 ], 
                    [ 873.25 , 1.6706566816852254e-46 ], [ 873.3125 , 1.6708550767323725e-46 ], [ 873.375 , 1.6710534953394617e-46 ], [ 873.4375 , 1.6712519375093146e-46 ], 
                    [ 873.5 , 1.6714504032447178e-46 ], [ 873.5625 , 1.6716488925484577e-46 ], [ 873.625 , 1.6718474054233564e-46 ], [ 873.6875 , 1.6720459418721899e-46 ], 
                    [ 873.75 , 1.6722445018977696e-46 ], [ 873.8125 , 1.6724430855029065e-46 ], [ 873.875 , 1.6726416541855599e-46 ], [ 873.9375 , 1.672839997054426e-46 ], 
                    [ 874.0 , 1.6730383634429222e-46 ], [ 874.0625 , 1.6732367533538025e-46 ], [ 874.125 , 1.6734351667899268e-46 ], [ 874.1875 , 1.673633603754002e-46 ], 
                    [ 874.25 , 1.6738320642488656e-46 ], [ 874.3125 , 1.6740305482772957e-46 ], [ 874.375 , 1.6742290558420594e-46 ], [ 874.4375 , 1.6744275869460072e-46 ], 
                    [ 874.5 , 1.6746261415918587e-46 ], [ 874.5625 , 1.6748247197824889e-46 ], [ 874.625 , 1.6750233215205946e-46 ], [ 874.6875 , 1.6752219468090397e-46 ], 
                    [ 874.75 , 1.6754205956505929e-46 ], [ 874.8125 , 1.6756192680480468e-46 ], [ 874.875 , 1.6758179640041948e-46 ], [ 874.9375 , 1.6760166835218072e-46 ], 
                    [ 875.0 , 1.6762154266037249e-46 ], [ 875.0625 , 1.6764141932527186e-46 ], [ 875.125 , 1.6766129834715713e-46 ], [ 875.1875 , 1.676811797263125e-46 ], 
                    [ 875.25 , 1.6770106346300801e-46 ], [ 875.3125 , 1.6772094955753392e-46 ], [ 875.375 , 1.6774083801016388e-46 ], [ 875.4375 , 1.6776072882117751e-46 ], 
                    [ 875.5 , 1.6778062199085683e-46 ], [ 875.5625 , 1.6780051751947798e-46 ], [ 875.625 , 1.6782041540732665e-46 ], [ 875.6875 , 1.6784031565467784e-46 ], 
                    [ 875.75 , 1.678602182618113e-46 ], [ 875.8125 , 1.6788012322900806e-46 ], [ 875.875 , 1.6790001741373925e-46 ], [ 875.9375 , 1.6791989734580267e-46 ], 
                    [ 876.0 , 1.6793977963171871e-46 ], [ 876.0625 , 1.6795966427176372e-46 ], [ 876.125 , 1.679795512662176e-46 ], [ 876.1875 , 1.6799944061536033e-46 ], 
                    [ 876.25 , 1.6801933231946957e-46 ], [ 876.3125 , 1.6803922637882287e-46 ], [ 876.375 , 1.6805912279369913e-46 ], [ 876.4375 , 1.6807902156438085e-46 ], 
                    [ 876.5 , 1.680989226911434e-46 ], [ 876.5625 , 1.6811882617426812e-46 ], [ 876.625 , 1.6813873201403039e-46 ], [ 876.6875 , 1.6815864021071409e-46 ], 
                    [ 876.75 , 1.6817855076459346e-46 ], [ 876.8125 , 1.6819846367595239e-46 ], [ 876.875 , 1.6821837894506522e-46 ], [ 876.9375 , 1.6823829657221595e-46 ], 
                    [ 877.0 , 1.6825821655767894e-46 ], [ 877.0625 , 1.6827813890174062e-46 ], [ 877.125 , 1.6829806360467189e-46 ], [ 877.1875 , 1.6831799066675323e-46 ], 
                    [ 877.25 , 1.6833792008827114e-46 ], [ 877.3125 , 1.6835785186949545e-46 ], [ 877.375 , 1.683777860107139e-46 ], [ 877.4375 , 1.6839772251220119e-46 ], 
                    [ 877.5 , 1.684176613742391e-46 ], [ 877.5625 , 1.6843760259710477e-46 ], [ 877.625 , 1.6845754618107892e-46 ], [ 877.6875 , 1.6847749212644112e-46 ], 
                    [ 877.75 , 1.6849744043347218e-46 ], [ 877.8125 , 1.685173911024505e-46 ], [ 877.875 , 1.6853732436005232e-46 ], [ 877.9375 , 1.6855725058681286e-46 ], 
                    [ 878.0 , 1.6857717916945931e-46 ], [ 878.0625 , 1.6859711010826428e-46 ], [ 878.125 , 1.686170434035147e-46 ], [ 878.1875 , 1.686369790554868e-46 ], 
                    [ 878.25 , 1.6865691706445556e-46 ], [ 878.3125 , 1.686768574307021e-46 ], [ 878.375 , 1.6869680015450633e-46 ], [ 878.4375 , 1.6871674523614455e-46 ], 
                    [ 878.5 , 1.6873669267589677e-46 ], [ 878.5625 , 1.6875664247404294e-46 ], [ 878.625 , 1.687765946308583e-46 ], [ 878.6875 , 1.6879654914662775e-46 ], 
                    [ 878.75 , 1.6881650602162418e-46 ], [ 878.8125 , 1.6883646525613011e-46 ], [ 878.875 , 1.6885642685042211e-46 ], [ 878.9375 , 1.688763908047792e-46 ], 
                    [ 879.0 , 1.6889635711948518e-46 ], [ 879.0625 , 1.689163257948119e-46 ], [ 879.125 , 1.6893629683104331e-46 ], [ 879.1875 , 1.6895627022845731e-46 ], 
                    [ 879.25 , 1.6897624598733547e-46 ], [ 879.3125 , 1.6899622410795456e-46 ], [ 879.375 , 1.6901620459059023e-46 ], [ 879.4375 , 1.6903618743552656e-46 ], 
                    [ 879.5 , 1.6905617264304163e-46 ], [ 879.5625 , 1.6907616021341477e-46 ], [ 879.625 , 1.6909615014692655e-46 ], [ 879.6875 , 1.6911614244385515e-46 ], 
                    [ 879.75 , 1.6913613710447759e-46 ], [ 879.8125 , 1.6915613412907936e-46 ], [ 879.875 , 1.6917610836493136e-46 ], [ 879.9375 , 1.6919608025308081e-46 ], 
                    [ 880.0 , 1.692160544989894e-46 ], [ 880.0625 , 1.6923603110293312e-46 ], [ 880.125 , 1.692560100651927e-46 ], [ 880.1875 , 1.6927599138604537e-46 ], 
                    [ 880.25 , 1.6929597506576957e-46 ], [ 880.3125 , 1.6931596110464377e-46 ], [ 880.375 , 1.6933594950294408e-46 ], [ 880.4375 , 1.6935594026095383e-46 ], 
                    [ 880.5 , 1.6937593337894922e-46 ], [ 880.5625 , 1.6939592885720882e-46 ], [ 880.625 , 1.6941592669601009e-46 ], [ 880.6875 , 1.6943592689563287e-46 ], 
                    [ 880.75 , 1.6945592945635712e-46 ], [ 880.8125 , 1.6947593437846152e-46 ], [ 880.875 , 1.6949594166222129e-46 ], [ 880.9375 , 1.695159513079188e-46 ], 
                    [ 881.0 , 1.6953596331583049e-46 ], [ 881.0625 , 1.6955597768623644e-46 ], [ 881.125 , 1.6957599441941794e-46 ], [ 881.1875 , 1.695960135156479e-46 ], 
                    [ 881.25 , 1.6961603497521135e-46 ], [ 881.3125 , 1.6963605879838607e-46 ], [ 881.375 , 1.6965608498544624e-46 ], [ 881.4375 , 1.6967611353667698e-46 ], 
                    [ 881.5 , 1.6969614445235739e-46 ], [ 881.5625 , 1.6971617773275939e-46 ], [ 881.625 , 1.6973621337816932e-46 ], [ 881.6875 , 1.697562513888628e-46 ], 
                    [ 881.75 , 1.6977629176512146e-46 ], [ 881.8125 , 1.6979633450722218e-46 ], [ 881.875 , 1.6981635192038992e-46 ], [ 881.9375 , 1.698363697488508e-46 ], 
                    [ 882.0 , 1.6985638993700117e-46 ], [ 882.0625 , 1.698764124851144e-46 ], [ 882.125 , 1.698964373934711e-46 ], [ 882.1875 , 1.6991646466235067e-46 ], 
                    [ 882.25 , 1.699364942920314e-46 ], [ 882.3125 , 1.6995652628278794e-46 ], [ 882.375 , 1.69976560634901e-46 ], [ 882.4375 , 1.699965973486514e-46 ], 
                    [ 882.5 , 1.700166364243115e-46 ], [ 882.5625 , 1.700366778621657e-46 ], [ 882.625 , 1.7005672166248884e-46 ], [ 882.6875 , 1.7007676782556187e-46 ], 
                    [ 882.75 , 1.7009681635166327e-46 ], [ 882.8125 , 1.7011686724106918e-46 ], [ 882.875 , 1.7013692049405698e-46 ], [ 882.9375 , 1.7015697611091013e-46 ], 
                    [ 883.0 , 1.7017703409190364e-46 ], [ 883.0625 , 1.701970944373174e-46 ], [ 883.125 , 1.7021715714743259e-46 ], [ 883.1875 , 1.7023722222252305e-46 ], 
                    [ 883.25 , 1.7025728966287245e-46 ], [ 883.3125 , 1.7027735946875477e-46 ], [ 883.375 , 1.7029743164045122e-46 ], [ 883.4375 , 1.7031750617824314e-46 ], 
                    [ 883.5 , 1.7033758308240463e-46 ], [ 883.5625 , 1.7035766235321824e-46 ], [ 883.625 , 1.703777439909605e-46 ], [ 883.6875 , 1.7039782799591288e-46 ], 
                    [ 883.75 , 1.7041791436835199e-46 ], [ 883.8125 , 1.7043800310855935e-46 ], [ 883.875 , 1.7045806561496024e-46 ], [ 883.9375 , 1.704781291112045e-46 ], 
                    [ 884.0 , 1.7049819496899078e-46 ], [ 884.0625 , 1.7051826318859582e-46 ], [ 884.125 , 1.705383337702988e-46 ], [ 884.1875 , 1.7055840671437898e-46 ], 
                    [ 884.25 , 1.7057848202111078e-46 ], [ 884.3125 , 1.7059855969077716e-46 ], [ 884.375 , 1.7061863972365135e-46 ], [ 884.4375 , 1.706387221200164e-46 ], 
                    [ 884.5 , 1.7065880688014806e-46 ], [ 884.5625 , 1.706788940043209e-46 ], [ 884.625 , 1.706989834928205e-46 ], [ 884.6875 , 1.7071907534592024e-46 ], 
                    [ 884.75 , 1.707391695639009e-46 ], [ 884.8125 , 1.7075926614703841e-46 ], [ 884.875 , 1.7077936509561237e-46 ], [ 884.9375 , 1.7079946640989994e-46 ], 
                    [ 885.0 , 1.708195700901857e-46 ], [ 885.0625 , 1.7083967613673843e-46 ], [ 885.125 , 1.7085978454984265e-46 ], [ 885.1875 , 1.7087989532977457e-46 ], 
                    [ 885.25 , 1.7090000847681635e-46 ], [ 885.3125 , 1.7092012399124298e-46 ], [ 885.375 , 1.7094024187333556e-46 ], [ 885.4375 , 1.709603621233691e-46 ], 
                    [ 885.5 , 1.709804847416272e-46 ], [ 885.5625 , 1.7100060972838493e-46 ], [ 885.625 , 1.7102073708392352e-46 ], [ 885.6875 , 1.7104086680851815e-46 ], 
                    [ 885.75 , 1.7106099890245253e-46 ], [ 885.8125 , 1.7108113336600185e-46 ], [ 885.875 , 1.711012431476433e-46 ], [ 885.9375 , 1.7112135233500387e-46 ], 
                    [ 886.0 , 1.7114146388575832e-46 ], [ 886.0625 , 1.7116157780018435e-46 ], [ 886.125 , 1.7118169407855857e-46 ], [ 886.1875 , 1.712018127211564e-46 ], 
                    [ 886.25 , 1.7122193372825932e-46 ], [ 886.3125 , 1.7124205710014521e-46 ], [ 886.375 , 1.7126218283708962e-46 ], [ 886.4375 , 1.7128231093937166e-46 ], 
                    [ 886.5 , 1.7130244140727182e-46 ], [ 886.5625 , 1.7132257424106321e-46 ], [ 886.625 , 1.7134270944102514e-46 ], [ 886.6875 , 1.7136284700743935e-46 ], 
                    [ 886.75 , 1.713829869405815e-46 ], [ 886.8125 , 1.7140312924072734e-46 ], [ 886.875 , 1.7142327390816114e-46 ], [ 886.9375 , 1.7144342094315382e-46 ], 
                    [ 887.0 , 1.7146357034598728e-46 ], [ 887.0625 , 1.7148372211694222e-46 ], [ 887.125 , 1.7150387625629457e-46 ], [ 887.1875 , 1.7152403276432268e-46 ], 
                    [ 887.25 , 1.7154419164130366e-46 ], [ 887.3125 , 1.71564352887516e-46 ], [ 887.375 , 1.7158451650324055e-46 ], [ 887.4375 , 1.7160468248875703e-46 ], 
                    [ 887.5 , 1.716248508443415e-46 ], [ 887.5625 , 1.7164502157026765e-46 ], [ 887.625 , 1.7166519466682135e-46 ], [ 887.6875 , 1.7168537013427883e-46 ], 
                    [ 887.75 , 1.7170554797292112e-46 ], [ 887.8125 , 1.7172572818302204e-46 ], [ 887.875 , 1.7174588768917029e-46 ], [ 887.9375 , 1.7176604290624016e-46 ], 
                    [ 888.0 , 1.7178620048862544e-46 ], [ 888.0625 , 1.718063604365988e-46 ], [ 888.125 , 1.7182652275044034e-46 ], [ 888.1875 , 1.718466874304289e-46 ], 
                    [ 888.25 , 1.7186685447683974e-46 ], [ 888.3125 , 1.7188702388995298e-46 ], [ 888.375 , 1.7190719567004518e-46 ], [ 888.4375 , 1.7192736981739286e-46 ], 
                    [ 888.5 , 1.7194754633227507e-46 ], [ 888.5625 , 1.7196772521497088e-46 ], [ 888.625 , 1.7198790646575573e-46 ], [ 888.6875 , 1.720080900849099e-46 ], 
                    [ 888.75 , 1.720282760727102e-46 ], [ 888.8125 , 1.7204846442943456e-46 ], [ 888.875 , 1.7206865515535853e-46 ], [ 888.9375 , 1.720888482507675e-46 ], 
                    [ 889.0 , 1.7210904371592975e-46 ], [ 889.0625 , 1.721292415511295e-46 ], [ 889.125 , 1.7214944175664612e-46 ], [ 889.1875 , 1.7216964433275164e-46 ], 
                    [ 889.25 , 1.7218984927973284e-46 ], [ 889.3125 , 1.7221005659785812e-46 ], [ 889.375 , 1.7223026628741318e-46 ], [ 889.4375 , 1.722504783486738e-46 ], 
                    [ 889.5 , 1.7227069278191833e-46 ], [ 889.5625 , 1.7229090958742391e-46 ], [ 889.625 , 1.7231112876547016e-46 ], [ 889.6875 , 1.723313503163355e-46 ], 
                    [ 889.75 , 1.7235157424030087e-46 ], [ 889.8125 , 1.7237180053764106e-46 ], [ 889.875 , 1.7239201132608287e-46 ], [ 889.9375 , 1.7241221142233015e-46 ], 
                    [ 890.0 , 1.7243241388553222e-46 ], [ 890.0625 , 1.7245261871596522e-46 ], [ 890.125 , 1.7247282591390534e-46 ], [ 890.1875 , 1.7249303547963237e-46 ], 
                    [ 890.25 , 1.7251324741342382e-46 ], [ 890.3125 , 1.7253346171555346e-46 ], [ 890.375 , 1.7255367838630124e-46 ], [ 890.4375 , 1.72573897425946e-46 ], 
                    [ 890.5 , 1.7259411883476397e-46 ], [ 890.5625 , 1.7261434261303288e-46 ], [ 890.625 , 1.7263456876103153e-46 ], [ 890.6875 , 1.7265479727903393e-46 ], 
                    [ 890.75 , 1.7267502816732025e-46 ], [ 890.8125 , 1.7269526142616819e-46 ], [ 890.875 , 1.7271549705585804e-46 ], [ 890.9375 , 1.7273573505666386e-46 ], 
                    [ 891.0 , 1.7275597542886352e-46 ], [ 891.0625 , 1.7277621817273858e-46 ], [ 891.125 , 1.7279646328856322e-46 ], [ 891.1875 , 1.7281671077661541e-46 ], 
                    [ 891.25 , 1.7283696063717433e-46 ], [ 891.3125 , 1.7285721287052045e-46 ], [ 891.375 , 1.7287746747692687e-46 ], [ 891.4375 , 1.7289772445667411e-46 ], 
                    [ 891.5 , 1.7291798381004396e-46 ], [ 891.5625 , 1.7293824553730595e-46 ], [ 891.625 , 1.7295850963874439e-46 ], [ 891.6875 , 1.7297877611463872e-46 ], 
                    [ 891.75 , 1.7299904496526343e-46 ], [ 891.8125 , 1.730193161908968e-46 ], [ 891.875 , 1.7303958073086802e-46 ], [ 891.9375 , 1.7305982716138786e-46 ], 
                    [ 892.0 , 1.7308007596083608e-46 ], [ 892.0625 , 1.7310032712948619e-46 ], [ 892.125 , 1.7312058066761534e-46 ], [ 892.1875 , 1.7314083657550205e-46 ], 
                    [ 892.25 , 1.7316109485342728e-46 ], [ 892.3125 , 1.7318135550165972e-46 ], [ 892.375 , 1.7320161852048407e-46 ], [ 892.4375 , 1.732218839101765e-46 ], 
                    [ 892.5 , 1.7324215167101314e-46 ], [ 892.5625 , 1.7326242180326777e-46 ], [ 892.625 , 1.7328269430722647e-46 ], [ 892.6875 , 1.7330296918315811e-46 ], 
                    [ 892.75 , 1.7332324643134514e-46 ], [ 892.8125 , 1.7334352605206515e-46 ], [ 892.875 , 1.7336380804559325e-46 ], [ 892.9375 , 1.7338409241220833e-46 ], 
                    [ 893.0 , 1.7340437915218677e-46 ], [ 893.0625 , 1.734246682658088e-46 ], [ 893.125 , 1.7344495975335084e-46 ], [ 893.1875 , 1.734652536150907e-46 ], 
                    [ 893.25 , 1.7348554985130738e-46 ], [ 893.3125 , 1.735058484622775e-46 ], [ 893.375 , 1.7352614944827641e-46 ], [ 893.4375 , 1.7354645280958571e-46 ], 
                    [ 893.5 , 1.735667585464809e-46 ], [ 893.5625 , 1.7358706665924237e-46 ], [ 893.625 , 1.7360737714814808e-46 ], [ 893.6875 , 1.7362769001347363e-46 ], 
                    [ 893.75 , 1.7364800525549827e-46 ], [ 893.8125 , 1.7366832287449888e-46 ], [ 893.875 , 1.73688642870756e-46 ], [ 893.9375 , 1.7370893549638017e-46 ], 
                    [ 894.0 , 1.7372922933563259e-46 ], [ 894.0625 , 1.7374952554574572e-46 ], [ 894.125 , 1.737698241269978e-46 ], [ 894.1875 , 1.7379012507966713e-46 ], 
                    [ 894.25 , 1.7381042840402947e-46 ], [ 894.3125 , 1.738307341003619e-46 ], [ 894.375 , 1.738510421689403e-46 ], [ 894.4375 , 1.7387135261004306e-46 ], 
                    [ 894.5 , 1.7389166542394734e-46 ], [ 894.5625 , 1.7391198061093159e-46 ], [ 894.625 , 1.7393229817127305e-46 ], [ 894.6875 , 1.7395261810524402e-46 ], 
                    [ 894.75 , 1.7397294041312803e-46 ], [ 894.8125 , 1.739932650951999e-46 ], [ 894.875 , 1.740135921517358e-46 ], [ 894.9375 , 1.740339215830156e-46 ], 
                    [ 895.0 , 1.7405425338931299e-46 ], [ 895.0625 , 1.740745875709104e-46 ], [ 895.125 , 1.7409492412808164e-46 ], [ 895.1875 , 1.7411526306110666e-46 ], 
                    [ 895.25 , 1.741356043702606e-46 ], [ 895.3125 , 1.7415594805582352e-46 ], [ 895.375 , 1.7417629411807057e-46 ], [ 895.4375 , 1.7419664255728063e-46 ], 
                    [ 895.5 , 1.7421699337373144e-46 ], [ 895.5625 , 1.7423734656770188e-46 ], [ 895.625 , 1.7425770213946856e-46 ], [ 895.6875 , 1.742780600893092e-46 ], 
                    [ 895.75 , 1.742984204175029e-46 ], [ 895.8125 , 1.743187831243238e-46 ], [ 895.875 , 1.7433914821005104e-46 ], [ 895.9375 , 1.743594988737403e-46 ], 
                    [ 896.0 , 1.743798378733034e-46 ], [ 896.0625 , 1.7440017924540785e-46 ], [ 896.125 , 1.744205229903279e-46 ], [ 896.1875 , 1.7444086910834162e-46 ], 
                    [ 896.25 , 1.7446121759972704e-46 ], [ 896.3125 , 1.744815684647598e-46 ], [ 896.375 , 1.745019217037155e-46 ], [ 896.4375 , 1.7452227731687235e-46 ], 
                    [ 896.5 , 1.7454263530450849e-46 ], [ 896.5625 , 1.7456299566689846e-46 ], [ 896.625 , 1.7458335840432051e-46 ], [ 896.6875 , 1.746037235170529e-46 ], 
                    [ 896.75 , 1.7462409100537149e-46 ], [ 896.8125 , 1.7464446086955216e-46 ], [ 896.875 , 1.7466483310987327e-46 ], [ 896.9375 , 1.7468520772661328e-46 ], 
                    [ 897.0 , 1.7470558472004809e-46 ], [ 897.0625 , 1.7472596409045375e-46 ], [ 897.125 , 1.747463458381063e-46 ], [ 897.1875 , 1.7476672996329045e-46 ], 
                    [ 897.25 , 1.747871164662786e-46 ], [ 897.3125 , 1.748075053473444e-46 ], [ 897.375 , 1.7482789660677143e-46 ], [ 897.4375 , 1.7484829024483346e-46 ], 
                    [ 897.5 , 1.7486868626181038e-46 ], [ 897.5625 , 1.7488908465797723e-46 ], [ 897.625 , 1.7490948543361402e-46 ], [ 897.6875 , 1.7492988858899461e-46 ], 
                    [ 897.75 , 1.7495029412440155e-46 ], [ 897.8125 , 1.7497070204011e-46 ], [ 897.875 , 1.7499111233639508e-46 ], [ 897.9375 , 1.7501152337618012e-46 ], 
                    [ 898.0 , 1.750319079738594e-46 ], [ 898.0625 , 1.7505229494584988e-46 ], [ 898.125 , 1.7507268429243061e-46 ], [ 898.1875 , 1.7509307601387196e-46 ], 
                    [ 898.25 , 1.7511347011045676e-46 ], [ 898.3125 , 1.7513386658245793e-46 ], [ 898.375 , 1.7515426543015214e-46 ], [ 898.4375 , 1.751746666538211e-46 ], 
                    [ 898.5 , 1.7519507025373523e-46 ], [ 898.5625 , 1.7521547623017394e-46 ], [ 898.625 , 1.7523588458341645e-46 ], [ 898.6875 , 1.752562953137346e-46 ], 
                    [ 898.75 , 1.7527670842140658e-46 ], [ 898.8125 , 1.75297123906713e-46 ], [ 898.875 , 1.7531754176992823e-46 ], [ 898.9375 , 1.7533796201132933e-46 ], 
                    [ 899.0 , 1.753583846311896e-46 ], [ 899.0625 , 1.7537880962979462e-46 ], [ 899.125 , 1.753992370074154e-46 ], [ 899.1875 , 1.754196667643277e-46 ], 
                    [ 899.25 , 1.7544009890081373e-46 ], [ 899.3125 , 1.754605334171468e-46 ], [ 899.375 , 1.7548097031360785e-46 ], [ 899.4375 , 1.755014095904692e-46 ], 
                    [ 899.5 , 1.755218512480093e-46 ], [ 899.5625 , 1.7554229528650923e-46 ], [ 899.625 , 1.7556274170624373e-46 ], [ 899.6875 , 1.7558319050748903e-46 ], 
                    [ 899.75 , 1.7560364169052365e-46 ], [ 899.8125 , 1.756240952556251e-46 ], [ 899.875 , 1.756445512030733e-46 ], [ 899.9375 , 1.756650095331382e-46 ], 
                    [ 900.0 , 1.75685454884907e-46 ], [ 900.0625 , 1.757058865595389e-46 ], [ 900.125 , 1.7572632061031006e-46 ], [ 900.1875 , 1.7574675703749806e-46 ], 
                    [ 900.25 , 1.757671958413793e-46 ], [ 900.3125 , 1.7578763702223145e-46 ], [ 900.375 , 1.758080805803284e-46 ], [ 900.4375 , 1.758285265159491e-46 ], 
                    [ 900.5 , 1.758489748293676e-46 ], [ 900.5625 , 1.758694255208617e-46 ], [ 900.625 , 1.758898785907091e-46 ], [ 900.6875 , 1.7591033403918527e-46 ], 
                    [ 900.75 , 1.759307918665655e-46 ], [ 900.8125 , 1.75951252073129e-46 ], [ 900.875 , 1.7597171465914994e-46 ], [ 900.9375 , 1.7599217962490756e-46 ], 
                    [ 901.0 , 1.7601264697067737e-46 ], [ 901.0625 , 1.7603311669673488e-46 ], [ 901.125 , 1.7605358880335815e-46 ], [ 901.1875 , 1.7607406329082407e-46 ], 
                    [ 901.25 , 1.760945401594108e-46 ], [ 901.3125 , 1.7611501940939267e-46 ], [ 901.375 , 1.761355010410492e-46 ], [ 901.4375 , 1.761559850546561e-46 ], 
                    [ 901.5 , 1.7617647145049036e-46 ], [ 901.5625 , 1.761969602288266e-46 ], [ 901.625 , 1.76217451389948e-46 ], [ 901.6875 , 1.7623794493412558e-46 ], 
                    [ 901.75 , 1.7625844086163767e-46 ], [ 901.8125 , 1.762789391727664e-46 ], [ 901.875 , 1.762994398677828e-46 ], [ 901.9375 , 1.7631994294696535e-46 ], 
                    [ 902.0 , 1.763404484105975e-46 ], [ 902.0625 , 1.7636093042760104e-46 ], [ 902.125 , 1.763814102788972e-46 ], [ 902.1875 , 1.764018925084076e-46 ], 
                    [ 902.25 , 1.7642237711640977e-46 ], [ 902.3125 , 1.764428641031799e-46 ], [ 902.375 , 1.7646335346899544e-46 ], [ 902.4375 , 1.764838452141289e-46 ], 
                    [ 902.5 , 1.7650433933886033e-46 ], [ 902.5625 , 1.7652483584346483e-46 ], [ 902.625 , 1.765453347282187e-46 ], [ 902.6875 , 1.7656583599339847e-46 ], 
                    [ 902.75 , 1.7658633963928044e-46 ], [ 902.8125 , 1.7660684566614235e-46 ], [ 902.875 , 1.766273540742582e-46 ], [ 902.9375 , 1.766478648639058e-46 ], 
                    [ 903.0 , 1.7666837803536285e-46 ], [ 903.0625 , 1.7668889358890105e-46 ], [ 903.125 , 1.767094115248057e-46 ], [ 903.1875 , 1.7672993184334603e-46 ], 
                    [ 903.25 , 1.7675045454480235e-46 ], [ 903.3125 , 1.7677097962944895e-46 ], [ 903.375 , 1.7679150709756383e-46 ], [ 903.4375 , 1.7681203694942378e-46 ], 
                    [ 903.5 , 1.7683256918530556e-46 ], [ 903.5625 , 1.7685310380548605e-46 ], [ 903.625 , 1.768736408102421e-46 ], [ 903.6875 , 1.768941801998494e-46 ], 
                    [ 903.75 , 1.7691472197458737e-46 ], [ 903.8125 , 1.7693526613473172e-46 ], [ 903.875 , 1.76955812680562e-46 ], [ 903.9375 , 1.76976361612349e-46 ], 
                    [ 904.0 , 1.7699691293037475e-46 ], [ 904.0625 , 1.7701746233513014e-46 ], [ 904.125 , 1.770379872464053e-46 ], [ 904.1875 , 1.770585145375121e-46 ], 
                    [ 904.25 , 1.770790442087277e-46 ], [ 904.3125 , 1.7709957626032948e-46 ], [ 904.375 , 1.771201106925895e-46 ], [ 904.4375 , 1.7714064750578888e-46 ], 
                    [ 904.5 , 1.7716118670019994e-46 ], [ 904.5625 , 1.771817282761013e-46 ], [ 904.625 , 1.772022722337665e-46 ], [ 904.6875 , 1.7722281857347432e-46 ], 
                    [ 904.75 , 1.772433672954984e-46 ], [ 904.8125 , 1.772639184001149e-46 ], [ 904.875 , 1.7728447188760396e-46 ], [ 904.9375 , 1.7730502775823923e-46 ], 
                    [ 905.0 , 1.7732558601229593e-46 ], [ 905.0625 , 1.773461466500528e-46 ], [ 905.125 , 1.7736670967178503e-46 ], [ 905.1875 , 1.7738727507776774e-46 ], 
                    [ 905.25 , 1.774078428682787e-46 ], [ 905.3125 , 1.7742841304359434e-46 ], [ 905.375 , 1.7744898560399114e-46 ], [ 905.4375 , 1.7746956054974702e-46 ], 
                    [ 905.5 , 1.774901378811372e-46 ], [ 905.5625 , 1.7751071759843585e-46 ], [ 905.625 , 1.775312997019233e-46 ], [ 905.6875 , 1.77551884191875e-46 ], 
                    [ 905.75 , 1.7757247106856902e-46 ], [ 905.8125 , 1.775930603322807e-46 ], [ 905.875 , 1.7761365198328435e-46 ], [ 905.9375 , 1.7763424602186177e-46 ], 
                    [ 906.0 , 1.7765484244828604e-46 ], [ 906.0625 , 1.7767544126283664e-46 ], [ 906.125 , 1.7769603130629267e-46 ], [ 906.1875 , 1.7771660311761524e-46 ], 
                    [ 906.25 , 1.777371773105322e-46 ], [ 906.3125 , 1.7775775388531678e-46 ], [ 906.375 , 1.7777833284224212e-46 ], [ 906.4375 , 1.777989141815891e-46 ], 
                    [ 906.5 , 1.778194979036322e-46 ], [ 906.5625 , 1.778400840086423e-46 ], [ 906.625 , 1.778606724969016e-46 ], [ 906.6875 , 1.7788126336868085e-46 ], 
                    [ 906.75 , 1.7790185662425986e-46 ], [ 906.8125 , 1.779224522639159e-46 ], [ 906.875 , 1.7794305028791984e-46 ], [ 906.9375 , 1.7796365069654897e-46 ], 
                    [ 907.0 , 1.779842534900832e-46 ], [ 907.0625 , 1.7800485866879615e-46 ], [ 907.125 , 1.7802546623296126e-46 ], [ 907.1875 , 1.7804607618286246e-46 ], 
                    [ 907.25 , 1.780666885187657e-46 ], [ 907.3125 , 1.7808730324095234e-46 ], [ 907.375 , 1.7810792034970366e-46 ], [ 907.4375 , 1.7812853984528584e-46 ], 
                    [ 907.5 , 1.781491617279828e-46 ], [ 907.5625 , 1.7816978599806712e-46 ], [ 907.625 , 1.781904126558164e-46 ], [ 907.6875 , 1.782110417015084e-46 ], 
                    [ 907.75 , 1.782316731354157e-46 ], [ 907.8125 , 1.782523069578211e-46 ], [ 907.875 , 1.7827294316899363e-46 ], [ 907.9375 , 1.782935817692148e-46 ], 
                    [ 908.0 , 1.783142227587612e-46 ], [ 908.0625 , 1.7833486613790826e-46 ], [ 908.125 , 1.7835551190693255e-46 ], [ 908.1875 , 1.7837614477413846e-46 ], 
                    [ 908.25 , 1.783967641255452e-46 ], [ 908.3125 , 1.7841738586044007e-46 ], [ 908.375 , 1.7843800997910115e-46 ], [ 908.4375 , 1.7845863648180146e-46 ], 
                    [ 908.5 , 1.7847926536881906e-46 ], [ 908.5625 , 1.7849989664042965e-46 ], [ 908.625 , 1.78520530296905e-46 ], [ 908.6875 , 1.785411663385246e-46 ], 
                    [ 908.75 , 1.7856180476556167e-46 ], [ 908.8125 , 1.785824455782932e-46 ], [ 908.875 , 1.7860308877699494e-46 ], [ 908.9375 , 1.786237343619453e-46 ], 
                    [ 909.0 , 1.7864438233341494e-46 ], [ 909.0625 , 1.7866503269168365e-46 ], [ 909.125 , 1.7868568543702467e-46 ], [ 909.1875 , 1.787063405697153e-46 ], 
                    [ 909.25 , 1.7872699809003274e-46 ], [ 909.3125 , 1.7874765799825045e-46 ], [ 909.375 , 1.7876832029464825e-46 ], [ 909.4375 , 1.7878898497949842e-46 ], 
                    [ 909.5 , 1.7880965205307705e-46 ], [ 909.5625 , 1.788303215156628e-46 ], [ 909.625 , 1.788509933675318e-46 ], [ 909.6875 , 1.7887166760895394e-46 ], 
                    [ 909.75 , 1.7889234424021554e-46 ], [ 909.8125 , 1.789130232615853e-46 ], [ 909.875 , 1.789337046733407e-46 ], [ 909.9375 , 1.7895438847576195e-46 ], 
                    [ 910.0 , 1.789750746691228e-46 ], [ 910.0625 , 1.7899576325369962e-46 ], [ 910.125 , 1.790164542297701e-46 ], [ 910.1875 , 1.7903714759760566e-46 ], 
                    [ 910.25 , 1.79057825975957e-46 ], [ 910.3125 , 1.7907849278865016e-46 ], [ 910.375 , 1.7909916198670203e-46 ], [ 910.4375 , 1.791198335703918e-46 ], 
                    [ 910.5 , 1.7914050753998713e-46 ], [ 910.5625 , 1.791611838957672e-46 ], [ 910.625 , 1.7918186263800874e-46 ], [ 910.6875 , 1.7920254376698462e-46 ], 
                    [ 910.75 , 1.7922322728297287e-46 ], [ 910.8125 , 1.792439131862478e-46 ], [ 910.875 , 1.7926460147708355e-46 ], [ 910.9375 , 1.79285292155757e-46 ], 
                    [ 911.0 , 1.793059852225451e-46 ], [ 911.0625 , 1.7932668067772088e-46 ], [ 911.125 , 1.793473785215613e-46 ], [ 911.1875 , 1.7936807875434332e-46 ], 
                    [ 911.25 , 1.793887813763414e-46 ], [ 911.3125 , 1.794094863878313e-46 ], [ 911.375 , 1.794301937890888e-46 ], [ 911.4375 , 1.794509035803872e-46 ], 
                    [ 911.5 , 1.7947161576200747e-46 ], [ 911.5625 , 1.7949233033422296e-46 ], [ 911.625 , 1.7951304729730696e-46 ], [ 911.6875 , 1.795337666515406e-46 ], 
                    [ 911.75 , 1.7955448839719984e-46 ], [ 911.8125 , 1.7957521253455305e-46 ], [ 911.875 , 1.7959593906388393e-46 ], [ 911.9375 , 1.7961666798546347e-46 ], 
                    [ 912.0 , 1.7963739929957415e-46 ], [ 912.0625 , 1.7965813300648574e-46 ], [ 912.125 , 1.7967886910647696e-46 ], [ 912.1875 , 1.796996075998253e-46 ], 
                    [ 912.25 , 1.797203484868032e-46 ], [ 912.3125 , 1.7974107421005154e-46 ], [ 912.375 , 1.7976178764223214e-46 ], [ 912.4375 , 1.797825034614355e-46 ], 
                    [ 912.5 , 1.798032216679355e-46 ], [ 912.5625 , 1.7982394226201233e-46 ], [ 912.625 , 1.7984466524393856e-46 ], [ 912.6875 , 1.7986539061399074e-46 ], 
                    [ 912.75 , 1.798861183724401e-46 ], [ 912.8125 , 1.799068485195658e-46 ], [ 912.875 , 1.799275810556418e-46 ], [ 912.9375 , 1.799483159809435e-46 ], 
                    [ 913.0 , 1.7996905329574483e-46 ], [ 913.0625 , 1.7998979300032504e-46 ], [ 913.125 , 1.8001053509495437e-46 ], [ 913.1875 , 1.8003127957991217e-46 ], 
                    [ 913.25 , 1.800520264554726e-46 ], [ 913.3125 , 1.8007277572191115e-46 ], [ 913.375 , 1.800935273795033e-46 ], [ 913.4375 , 1.801142814285234e-46 ], 
                    [ 913.5 , 1.801350378692508e-46 ], [ 913.5625 , 1.8015579670195612e-46 ], [ 913.625 , 1.801765579269187e-46 ], [ 913.6875 , 1.8019732154441304e-46 ], 
                    [ 913.75 , 1.8021808755471484e-46 ], [ 913.8125 , 1.8023885595809988e-46 ], [ 913.875 , 1.8025962675484516e-46 ], [ 913.9375 , 1.8028039994522267e-46 ], 
                    [ 914.0 , 1.803011755295121e-46 ], [ 914.0625 , 1.8032195350798804e-46 ], [ 914.125 , 1.8034273388092513e-46 ], [ 914.1875 , 1.8036351664860187e-46 ], 
                    [ 914.25 , 1.8038430181129293e-46 ], [ 914.3125 , 1.8040508936927305e-46 ], [ 914.375 , 1.804258650286317e-46 ], [ 914.4375 , 1.8044662627683844e-46 ], 
                    [ 914.5 , 1.804673899139997e-46 ], [ 914.5625 , 1.804881559403943e-46 ], [ 914.625 , 1.805089243562933e-46 ], [ 914.6875 , 1.8052969516197294e-46 ], 
                    [ 914.75 , 1.805504683577082e-46 ], [ 914.8125 , 1.8057124394377277e-46 ], [ 914.875 , 1.805920219204443e-46 ], [ 914.9375 , 1.8061280228799792e-46 ], 
                    [ 915.0 , 1.8063358504670485e-46 ], [ 915.0625 , 1.806543701968441e-46 ], [ 915.125 , 1.8067515773868956e-46 ], [ 915.1875 , 1.8069594767251638e-46 ], 
                    [ 915.25 , 1.807167399985999e-46 ], [ 915.3125 , 1.8073753471721534e-46 ], [ 915.375 , 1.8075833182863803e-46 ], [ 915.4375 , 1.8077913133314326e-46 ], 
                    [ 915.5 , 1.8079993323100643e-46 ], [ 915.5625 , 1.8082073752250292e-46 ], [ 915.625 , 1.808415442079095e-46 ], [ 915.6875 , 1.808623532874977e-46 ], 
                    [ 915.75 , 1.8088316476154693e-46 ], [ 915.8125 , 1.8090397863033138e-46 ], [ 915.875 , 1.8092479489412668e-46 ], [ 915.9375 , 1.8094561355320832e-46 ], 
                    [ 916.0 , 1.809664346078507e-46 ], [ 916.0625 , 1.8098725805833075e-46 ], [ 916.125 , 1.81008083904928e-46 ], [ 916.1875 , 1.8102891214791047e-46 ], 
                    [ 916.25 , 1.8104974278755905e-46 ], [ 916.3125 , 1.810705758241482e-46 ], [ 916.375 , 1.810914112579538e-46 ], [ 916.4375 , 1.811122394222008e-46 ], 
                    [ 916.5 , 1.8113304738853422e-46 ], [ 916.5625 , 1.811538577454927e-46 ], [ 916.625 , 1.8117467049335352e-46 ], [ 916.6875 , 1.811954856323862e-46 ], 
                    [ 916.75 , 1.812163031628693e-46 ], [ 916.8125 , 1.81237123085075e-46 ], [ 916.875 , 1.812579453992807e-46 ], [ 916.9375 , 1.8127877010575866e-46 ], 
                    [ 917.0 , 1.8129959720478622e-46 ], [ 917.0625 , 1.8132042669663577e-46 ], [ 917.125 , 1.8134125858158474e-46 ], [ 917.1875 , 1.8136209285990422e-46 ], 
                    [ 917.25 , 1.8138292953187565e-46 ], [ 917.3125 , 1.814037685977663e-46 ], [ 917.375 , 1.814246100578589e-46 ], [ 917.4375 , 1.8144545391242213e-46 ], 
                    [ 917.5 , 1.8146630016173615e-46 ], [ 917.5625 , 1.814871488060685e-46 ], [ 917.625 , 1.815079998457033e-46 ], [ 917.6875 , 1.8152885328091053e-46 ], 
                    [ 917.75 , 1.8154970911196936e-46 ], [ 917.8125 , 1.8157056733915245e-46 ], [ 917.875 , 1.8159142796272994e-46 ], [ 917.9375 , 1.8161229098298745e-46 ], 
                    [ 918.0 , 1.8163315640019132e-46 ], [ 918.0625 , 1.816540242146221e-46 ], [ 918.125 , 1.8167489442655647e-46 ], [ 918.1875 , 1.8169576703626602e-46 ], 
                    [ 918.25 , 1.8171664204402498e-46 ], [ 918.3125 , 1.8173751945011527e-46 ], [ 918.375 , 1.817583992548047e-46 ], [ 918.4375 , 1.8177928145837397e-46 ], 
                    [ 918.5 , 1.8180016397889686e-46 ], [ 918.5625 , 1.818210186848721e-46 ], [ 918.625 , 1.8184187578313627e-46 ], [ 918.6875 , 1.8186273527396516e-46 ], 
                    [ 918.75 , 1.8188359715763448e-46 ], [ 918.8125 , 1.819044614344174e-46 ], [ 918.875 , 1.819253281045859e-46 ], [ 918.9375 , 1.819461971684197e-46 ], 
                    [ 919.0 , 1.8196706862619075e-46 ], [ 919.0625 , 1.8198794247817117e-46 ], [ 919.125 , 1.8200881872464193e-46 ], [ 919.1875 , 1.8202969736587135e-46 ], 
                    [ 919.25 , 1.82050578402138e-46 ], [ 919.3125 , 1.8207146183371275e-46 ], [ 919.375 , 1.8209234766087677e-46 ], [ 919.4375 , 1.821132358838985e-46 ], 
                    [ 919.5 , 1.8213412650305525e-46 ], [ 919.5625 , 1.8215501951862195e-46 ], [ 919.625 , 1.821759149308748e-46 ], [ 919.6875 , 1.821968127400874e-46 ], 
                    [ 919.75 , 1.8221771294653346e-46 ], [ 919.8125 , 1.822386155504918e-46 ], [ 919.875 , 1.8225952055223103e-46 ], [ 919.9375 , 1.822804279520326e-46 ], 
                    [ 920.0 , 1.8230133775016782e-46 ], [ 920.0625 , 1.823222499469143e-46 ], [ 920.125 , 1.8234316454254464e-46 ], [ 920.1875 , 1.8236408153733662e-46 ], 
                    [ 920.25 , 1.8238500093156154e-46 ], [ 920.3125 , 1.8240592272550246e-46 ], [ 920.375 , 1.8242684691942425e-46 ], [ 920.4375 , 1.8244777351361006e-46 ], 
                    [ 920.5 , 1.8246870250833124e-46 ], [ 920.5625 , 1.824896339038671e-46 ], [ 920.625 , 1.8251054362846696e-46 ], [ 920.6875 , 1.825314478976636e-46 ], 
                    [ 920.75 , 1.8255235456117924e-46 ], [ 920.8125 , 1.8257326361928817e-46 ], [ 920.875 , 1.825941750722647e-46 ], [ 920.9375 , 1.82615088920383e-46 ], 
                    [ 921.0 , 1.8263600516391626e-46 ], [ 921.0625 , 1.8265692380314268e-46 ], [ 921.125 , 1.8267784483833274e-46 ], [ 921.1875 , 1.8269876826976215e-46 ], 
                    [ 921.25 , 1.8271969409770805e-46 ], [ 921.3125 , 1.827406223224357e-46 ], [ 921.375 , 1.8276155294423016e-46 ], [ 921.4375 , 1.827824859633633e-46 ], 
                    [ 921.5 , 1.8280342138010713e-46 ], [ 921.5625 , 1.8282435919473758e-46 ], [ 921.625 , 1.8284529940753053e-46 ], [ 921.6875 , 1.8286624201875945e-46 ], 
                    [ 921.75 , 1.8288718702869768e-46 ], [ 921.8125 , 1.8290813443762133e-46 ], [ 921.875 , 1.829290842458051e-46 ], [ 921.9375 , 1.8295003645352387e-46 ], 
                    [ 922.0 , 1.829709910610524e-46 ], [ 922.0625 , 1.829919480686643e-46 ], [ 922.125 , 1.8301290747663838e-46 ], [ 922.1875 , 1.8303386928524565e-46 ], 
                    [ 922.25 , 1.8305483349476234e-46 ], [ 922.3125 , 1.830758001054609e-46 ], [ 922.375 , 1.8309676911762285e-46 ], [ 922.4375 , 1.8311774053151674e-46 ], 
                    [ 922.5 , 1.831387143474228e-46 ], [ 922.5625 , 1.8315969056561235e-46 ], [ 922.625 , 1.831806691863631e-46 ], [ 922.6875 , 1.8320163840650823e-46 ], 
                    [ 922.75 , 1.832225897675572e-46 ], [ 922.8125 , 1.83243543524652e-46 ], [ 922.875 , 1.8326449967806803e-46 ], [ 922.9375 , 1.83285458228078e-46 ], 
                    [ 923.0 , 1.8330641917495466e-46 ], [ 923.0625 , 1.8332738251897606e-46 ], [ 923.125 , 1.8334834826041114e-46 ], [ 923.1875 , 1.8336931639953795e-46 ], 
                    [ 923.25 , 1.8339028693662816e-46 ], [ 923.3125 , 1.8341125987195855e-46 ], [ 923.375 , 1.8343223520580207e-46 ], [ 923.4375 , 1.8345321293843044e-46 ], 
                    [ 923.5 , 1.834741930701232e-46 ], [ 923.5625 , 1.8349517560115216e-46 ], [ 923.625 , 1.835161605317916e-46 ], [ 923.6875 , 1.8353714786231604e-46 ], 
                    [ 923.75 , 1.8355813759299864e-46 ], [ 923.8125 , 1.835791297241177e-46 ], [ 923.875 , 1.836001242559413e-46 ], [ 923.9375 , 1.836211211887492e-46 ], 
                    [ 924.0 , 1.836421205228186e-46 ], [ 924.0625 , 1.8366312225841494e-46 ], [ 924.125 , 1.83684126395822e-46 ], [ 924.1875 , 1.8370513293530798e-46 ], 
                    [ 924.25 , 1.8372614187715015e-46 ], [ 924.3125 , 1.8374715322162455e-46 ], [ 924.375 , 1.837681669690047e-46 ], [ 924.4375 , 1.8378918311956405e-46 ], 
                    [ 924.5 , 1.8381020167358005e-46 ], [ 924.5625 , 1.838312226313263e-46 ], [ 924.625 , 1.83852245993075e-46 ], [ 924.6875 , 1.838732717591064e-46 ], 
                    [ 924.75 , 1.8389429992969146e-46 ], [ 924.8125 , 1.839153011252043e-46 ], [ 924.875 , 1.8393630169054695e-46 ], [ 924.9375 , 1.8395730465385904e-46 ], 
                    [ 925.0 , 1.839783100154196e-46 ], [ 925.0625 , 1.8399931777549977e-46 ], [ 925.125 , 1.8402032793437354e-46 ], [ 925.1875 , 1.8404134049231608e-46 ], 
                    [ 925.25 , 1.8406235544959867e-46 ], [ 925.3125 , 1.8408337280649536e-46 ], [ 925.375 , 1.841043925632801e-46 ], [ 925.4375 , 1.841254147202309e-46 ], 
                    [ 925.5 , 1.841464392776191e-46 ], [ 925.5625 , 1.841674662357202e-46 ], [ 925.625 , 1.841884955948031e-46 ], [ 925.6875 , 1.8420952735514715e-46 ], 
                    [ 925.75 , 1.842305615170266e-46 ], [ 925.8125 , 1.8425159808071426e-46 ], [ 925.875 , 1.842726370464858e-46 ], [ 925.9375 , 1.8429367841461016e-46 ], 
                    [ 926.0 , 1.8431472218536566e-46 ], [ 926.0625 , 1.8433576835902927e-46 ], [ 926.125 , 1.843568169358727e-46 ], [ 926.1875 , 1.8437786791617037e-46 ], 
                    [ 926.25 , 1.843989213001968e-46 ], [ 926.3125 , 1.8441997708822635e-46 ], [ 926.375 , 1.8444103528053098e-46 ], [ 926.4375 , 1.8446209587739043e-46 ], 
                    [ 926.5 , 1.8448315887907664e-46 ], [ 926.5625 , 1.8450422428586426e-46 ], [ 926.625 , 1.8452529209802654e-46 ], [ 926.6875 , 1.845463623158395e-46 ], 
                    [ 926.75 , 1.8456743493958043e-46 ], [ 926.8125 , 1.845885099695175e-46 ], [ 926.875 , 1.8460957531805963e-46 ], [ 926.9375 , 1.8463062240829723e-46 ], 
                    [ 927.0 , 1.8465167189808144e-46 ], [ 927.0625 , 1.8467272378769235e-46 ], [ 927.125 , 1.8469377807740105e-46 ], [ 927.1875 , 1.8471483476748106e-46 ], 
                    [ 927.25 , 1.847358938582034e-46 ], [ 927.3125 , 1.8475695534984717e-46 ], [ 927.375 , 1.8477801924268333e-46 ], [ 927.4375 , 1.8479908553698568e-46 ], 
                    [ 927.5 , 1.8482015423302534e-46 ], [ 927.5625 , 1.8484122533108018e-46 ], [ 927.625 , 1.8486229883142785e-46 ], [ 927.6875 , 1.8488337473433053e-46 ], 
                    [ 927.75 , 1.8490445304006995e-46 ], [ 927.8125 , 1.8492553374892006e-46 ], [ 927.875 , 1.8494661686115355e-46 ], [ 927.9375 , 1.8496770237704306e-46 ], 
                    [ 928.0 , 1.849887902968627e-46 ], [ 928.0625 , 1.8500988062089436e-46 ], [ 928.125 , 1.8503097334940035e-46 ], [ 928.1875 , 1.850520684826587e-46 ], 
                    [ 928.25 , 1.8507316602094495e-46 ], [ 928.3125 , 1.8509426596453717e-46 ], [ 928.375 , 1.8511536831370573e-46 ], [ 928.4375 , 1.8513647306872085e-46 ], 
                    [ 928.5 , 1.8515758022986342e-46 ], [ 928.5625 , 1.851786897974065e-46 ], [ 928.625 , 1.851998017716191e-46 ], [ 928.6875 , 1.8522091615278352e-46 ], 
                    [ 928.75 , 1.8524203294116896e-46 ], [ 928.8125 , 1.852631521370485e-46 ], [ 928.875 , 1.852842737407032e-46 ], [ 928.9375 , 1.8530539775240097e-46 ], 
                    [ 929.0 , 1.8532649947693086e-46 ], [ 929.0625 , 1.853475956122625e-46 ], [ 929.125 , 1.8536869414901626e-46 ], [ 929.1875 , 1.853897950874656e-46 ], 
                    [ 929.25 , 1.854108984278813e-46 ], [ 929.3125 , 1.854320041705367e-46 ], [ 929.375 , 1.8545311231571057e-46 ], [ 929.4375 , 1.8547422286366714e-46 ], 
                    [ 929.5 , 1.8549533581468917e-46 ], [ 929.5625 , 1.85516451169045e-46 ], [ 929.625 , 1.855375689270094e-46 ], [ 929.6875 , 1.85558689088856e-46 ], 
                    [ 929.75 , 1.8557981165485588e-46 ], [ 929.8125 , 1.856009366252906e-46 ], [ 929.875 , 1.8562206400042456e-46 ], [ 929.9375 , 1.856431937805395e-46 ], 
                    [ 930.0 , 1.8566432596590252e-46 ], [ 930.0625 , 1.856854605567901e-46 ], [ 930.125 , 1.8570659755347605e-46 ], [ 930.1875 , 1.857277369562329e-46 ], 
                    [ 930.25 , 1.8574887876533584e-46 ], [ 930.3125 , 1.8577002298106013e-46 ], [ 930.375 , 1.8579116960367975e-46 ], [ 930.4375 , 1.85812318633466e-46 ], 
                    [ 930.5 , 1.8583347007069424e-46 ], [ 930.5625 , 1.8585462391563847e-46 ], [ 930.625 , 1.8587578016857283e-46 ], [ 930.6875 , 1.8589693882977008e-46 ], 
                    [ 930.75 , 1.859180998995083e-46 ], [ 930.8125 , 1.85939263378059e-46 ], [ 930.875 , 1.859604292656925e-46 ], [ 930.9375 , 1.859815975626909e-46 ], 
                    [ 931.0 , 1.860027682693232e-46 ], [ 931.0625 , 1.8602393924900605e-46 ], [ 931.125 , 1.860450823558574e-46 ], [ 931.1875 , 1.860662278657905e-46 ], 
                    [ 931.25 , 1.86087375779081e-46 ], [ 931.3125 , 1.8610852609600082e-46 ], [ 931.375 , 1.861296788168218e-46 ], [ 931.4375 , 1.8615083394181985e-46 ], 
                    [ 931.5 , 1.861719914712655e-46 ], [ 931.5625 , 1.8619315140543342e-46 ], [ 931.625 , 1.8621431374459955e-46 ], [ 931.6875 , 1.8623547848903195e-46 ], 
                    [ 931.75 , 1.8625664563900664e-46 ], [ 931.8125 , 1.8627781519479835e-46 ], [ 931.875 , 1.862989871566792e-46 ], [ 931.9375 , 1.8632016152492265e-46 ], 
                    [ 932.0 , 1.863413382998009e-46 ], [ 932.0625 , 1.8636251748159016e-46 ], [ 932.125 , 1.8638369907055993e-46 ], [ 932.1875 , 1.864048830669892e-46 ], 
                    [ 932.25 , 1.8642606947115016e-46 ], [ 932.3125 , 1.8644725828331393e-46 ], [ 932.375 , 1.8646844950375284e-46 ], [ 932.4375 , 1.864896431327459e-46 ], 
                    [ 932.5 , 1.865108391705642e-46 ], [ 932.5625 , 1.8653203761748154e-46 ], [ 932.625 , 1.865532384737704e-46 ], [ 932.6875 , 1.8657444173970727e-46 ], 
                    [ 932.75 , 1.8659564741556604e-46 ], [ 932.8125 , 1.8661685550161532e-46 ], [ 932.875 , 1.8663806599813564e-46 ], [ 932.9375 , 1.866592789053997e-46 ], 
                    [ 933.0 , 1.8668049422367482e-46 ], [ 933.0625 , 1.8670171195324434e-46 ], [ 933.125 , 1.8672293209437697e-46 ], [ 933.1875 , 1.867441447776973e-46 ], 
                    [ 933.25 , 1.8676533682151674e-46 ], [ 933.3125 , 1.867865312702454e-46 ], [ 933.375 , 1.868077281241589e-46 ], [ 933.4375 , 1.8682892738352487e-46 ], 
                    [ 933.5 , 1.8685012904861888e-46 ], [ 933.5625 , 1.8687133311971665e-46 ], [ 933.625 , 1.8689253959708723e-46 ], [ 933.6875 , 1.869137484810037e-46 ], 
                    [ 933.75 , 1.8693495977174174e-46 ], [ 933.8125 , 1.8695617346957586e-46 ], [ 933.875 , 1.869773895747713e-46 ], [ 933.9375 , 1.869986080876091e-46 ], 
                    [ 934.0 , 1.8701982900835865e-46 ], [ 934.0625 , 1.8704105233729577e-46 ], [ 934.125 , 1.8706227807469238e-46 ], [ 934.1875 , 1.8708350622082322e-46 ], 
                    [ 934.25 , 1.8710473677595625e-46 ], [ 934.3125 , 1.8712596974037153e-46 ], [ 934.375 , 1.8714720511434113e-46 ], [ 934.4375 , 1.871684428981345e-46 ], 
                    [ 934.5 , 1.8718968309202775e-46 ], [ 934.5625 , 1.8721092569629842e-46 ], [ 934.625 , 1.87232170711212e-46 ], [ 934.6875 , 1.8725341813704615e-46 ], 
                    [ 934.75 , 1.87274667974077e-46 ], [ 934.8125 , 1.8729592022257424e-46 ], [ 934.875 , 1.8731717488281154e-46 ], [ 934.9375 , 1.8733843195506526e-46 ], 
                    [ 935.0 , 1.8735969143960777e-46 ], [ 935.0625 , 1.8738095333671143e-46 ], [ 935.125 , 1.874022176466528e-46 ], [ 935.1875 , 1.8742348436970298e-46 ], 
                    [ 935.25 , 1.874447535061384e-46 ], [ 935.3125 , 1.874660100095336e-46 ], [ 935.375 , 1.8748725106333852e-46 ], [ 935.4375 , 1.875084945238844e-46 ], 
                    [ 935.5 , 1.8752974039144516e-46 ], [ 935.5625 , 1.875509886662937e-46 ], [ 935.625 , 1.8757223934870264e-46 ], [ 935.6875 , 1.875934924389461e-46 ], 
                    [ 935.75 , 1.8761474793729566e-46 ], [ 935.8125 , 1.8763600584402274e-46 ], [ 935.875 , 1.8765726615940033e-46 ], [ 935.9375 , 1.8767852888370526e-46 ], 
                    [ 936.0 , 1.8769979401720653e-46 ], [ 936.0625 , 1.8772106156017974e-46 ], [ 936.125 , 1.8774233151289528e-46 ], [ 936.1875 , 1.8776360387562882e-46 ], 
                    [ 936.25 , 1.8778487864865077e-46 ], [ 936.3125 , 1.8780615583223687e-46 ], [ 936.375 , 1.8782743542665763e-46 ], [ 936.4375 , 1.8784871743218753e-46 ], 
                    [ 936.5 , 1.8787000184909976e-46 ], [ 936.5625 , 1.8789128867766754e-46 ], [ 936.625 , 1.879125779181655e-46 ], [ 936.6875 , 1.879338695708669e-46 ], 
                    [ 936.75 , 1.87955163636037e-46 ], [ 936.8125 , 1.8797646011396122e-46 ], [ 936.875 , 1.879977590049076e-46 ], [ 936.9375 , 1.8801906030914952e-46 ], 
                    [ 937.0 , 1.880403640269591e-46 ], [ 937.0625 , 1.880616701586125e-46 ], [ 937.125 , 1.8808297870437913e-46 ], [ 937.1875 , 1.88104289664538e-46 ], 
                    [ 937.25 , 1.881256030393586e-46 ], [ 937.3125 , 1.881469188291172e-46 ], [ 937.375 , 1.8816823703408343e-46 ], [ 937.4375 , 1.8818953993217725e-46 ], 
                    [ 937.5 , 1.88210830016565e-46 ], [ 937.5625 , 1.8823212250952013e-46 ], [ 937.625 , 1.8825341741132052e-46 ], [ 937.6875 , 1.8827471472223464e-46 ], 
                    [ 937.75 , 1.882960144425364e-46 ], [ 937.8125 , 1.8831731657249827e-46 ], [ 937.875 , 1.8833862111239298e-46 ], [ 937.9375 , 1.8835992806249313e-46 ], 
                    [ 938.0 , 1.883812374230714e-46 ], [ 938.0625 , 1.8840254919440045e-46 ], [ 938.125 , 1.884238633767544e-46 ], [ 938.1875 , 1.884451799704033e-46 ], 
                    [ 938.25 , 1.884664989756199e-46 ], [ 938.3125 , 1.8848782039268115e-46 ], [ 938.375 , 1.885091442218558e-46 ], [ 938.4375 , 1.885304704634181e-46 ], 
                    [ 938.5 , 1.88551799117641e-46 ], [ 938.5625 , 1.885731301847974e-46 ], [ 938.625 , 1.8859446366515764e-46 ], [ 938.6875 , 1.8861579955900405e-46 ], 
                    [ 938.75 , 1.8863713786659897e-46 ], [ 938.8125 , 1.8865847858822088e-46 ], [ 938.875 , 1.8867982172414013e-46 ], [ 938.9375 , 1.8870116727462855e-46 ], 
                    [ 939.0 , 1.8872251523996603e-46 ], [ 939.0625 , 1.8874386562042034e-46 ], [ 939.125 , 1.8876521841626744e-46 ], [ 939.1875 , 1.8878657362777788e-46 ], 
                    [ 939.25 , 1.8880793125522495e-46 ], [ 939.3125 , 1.888292912988846e-46 ], [ 939.375 , 1.888506537590303e-46 ], [ 939.4375 , 1.8887201863593266e-46 ], 
                    [ 939.5 , 1.8889338592986508e-46 ], [ 939.5625 , 1.889147384838123e-46 ], [ 939.625 , 1.8893607893756957e-46 ], [ 939.6875 , 1.8895742180201604e-46 ], 
                    [ 939.75 , 1.8897876707742397e-46 ], [ 939.8125 , 1.8900011476406845e-46 ], [ 939.875 , 1.890214648622218e-46 ], [ 939.9375 , 1.8904281737215112e-46 ], 
                    [ 940.0 , 1.8906417229413686e-46 ], [ 940.0625 , 1.8908552962844347e-46 ], [ 940.125 , 1.891068893753515e-46 ], [ 940.1875 , 1.891282515351281e-46 ], 
                    [ 940.25 , 1.891496161080472e-46 ], [ 940.3125 , 1.8917098309438007e-46 ], [ 940.375 , 1.8919235249440198e-46 ], [ 940.4375 , 1.8921372430838425e-46 ], 
                    [ 940.5 , 1.8923509853659825e-46 ], [ 940.5625 , 1.8925647517932074e-46 ], [ 940.625 , 1.8927785423682038e-46 ], [ 940.6875 , 1.892992357093687e-46 ], 
                    [ 940.75 , 1.8932061959724514e-46 ], [ 940.8125 , 1.893420059007172e-46 ], [ 940.875 , 1.893633946200578e-46 ], [ 940.9375 , 1.8938478575554248e-46 ], 
                    [ 941.0 , 1.894061793074402e-46 ], [ 941.0625 , 1.8942757527603187e-46 ], [ 941.125 , 1.894489736615812e-46 ], [ 941.1875 , 1.8947037446436653e-46 ], 
                    [ 941.25 , 1.8949177768465962e-46 ], [ 941.3125 , 1.8951318332273086e-46 ], [ 941.375 , 1.895345913788601e-46 ], [ 941.4375 , 1.8955600185331505e-46 ], 
                    [ 941.5 , 1.8957741474636758e-46 ], [ 941.5625 , 1.895988300582977e-46 ], [ 941.625 , 1.8962024778937316e-46 ], [ 941.6875 , 1.896416523391693e-46 ], 
                    [ 941.75 , 1.8966304161453567e-46 ], [ 941.8125 , 1.8968443330235525e-46 ], [ 941.875 , 1.8970582740289343e-46 ], [ 941.9375 , 1.8972722391643036e-46 ], 
                    [ 942.0 , 1.897486228432328e-46 ], [ 942.0625 , 1.89770024183573e-46 ], [ 942.125 , 1.8979142793772584e-46 ], [ 942.1875 , 1.898128341059622e-46 ], 
                    [ 942.25 , 1.8983424268855303e-46 ], [ 942.3125 , 1.8985565368577337e-46 ], [ 942.375 , 1.8987706709789413e-46 ], [ 942.4375 , 1.8989848292518777e-46 ], 
                    [ 942.5 , 1.8991990116792664e-46 ], [ 942.5625 , 1.8994132182638184e-46 ], [ 942.625 , 1.8996274490082717e-46 ], [ 942.6875 , 1.899841703915365e-46 ], 
                    [ 942.75 , 1.9000559829877966e-46 ], [ 942.8125 , 1.900270286228332e-46 ], [ 942.875 , 1.9004846136396437e-46 ], [ 942.9375 , 1.900698965224498e-46 ], 
                    [ 943.0 , 1.9009133409856214e-46 ], [ 943.0625 , 1.9011277409256865e-46 ], [ 943.125 , 1.9013421650475022e-46 ], [ 943.1875 , 1.9015566133537277e-46 ], 
                    [ 943.25 , 1.9017710858471442e-46 ], [ 943.3125 , 1.9019855825304137e-46 ], [ 943.375 , 1.902200103406331e-46 ], [ 943.4375 , 1.9024146484775843e-46 ], 
                    [ 943.5 , 1.9026292177469437e-46 ], [ 943.5625 , 1.9028438112170837e-46 ], [ 943.625 , 1.9030584288907743e-46 ], [ 943.6875 , 1.903273070770719e-46 ], 
                    [ 943.75 , 1.9034877368596873e-46 ], [ 943.8125 , 1.903702328874808e-46 ], [ 943.875 , 1.9039167343476446e-46 ], [ 943.9375 , 1.9041311639680232e-46 ], 
                    [ 944.0 , 1.9043456177386637e-46 ], [ 944.0625 , 1.9045600956622592e-46 ], [ 944.125 , 1.9047745977415434e-46 ], [ 944.1875 , 1.9049891239792504e-46 ], 
                    [ 944.25 , 1.9052036743780732e-46 ], [ 944.3125 , 1.9054182489407613e-46 ], [ 944.375 , 1.9056328476700085e-46 ], [ 944.4375 , 1.905847470568536e-46 ], 
                    [ 944.5 , 1.9060621176391075e-46 ], [ 944.5625 , 1.9062767888844178e-46 ], [ 944.625 , 1.9064914843071626e-46 ], [ 944.6875 , 1.906706203910119e-46 ], 
                    [ 944.75 , 1.906920947695983e-46 ], [ 944.8125 , 1.907135715667465e-46 ], [ 944.875 , 1.9073505078273298e-46 ], [ 944.9375 , 1.907565324178247e-46 ], 
                    [ 945.0 , 1.9077801647229545e-46 ], [ 945.0625 , 1.9079950294642184e-46 ], [ 945.125 , 1.9082099184046955e-46 ], [ 945.1875 , 1.9084248315472067e-46 ], 
                    [ 945.25 , 1.9086397688944096e-46 ], [ 945.3125 , 1.9088547304490303e-46 ], [ 945.375 , 1.909069716213822e-46 ], [ 945.4375 , 1.9092847261914855e-46 ], 
                    [ 945.5 , 1.90949976038476e-46 ], [ 945.5625 , 1.9097148187963866e-46 ], [ 945.625 , 1.9099299014290528e-46 ], [ 945.6875 , 1.91014500828554e-46 ], 
                    [ 945.75 , 1.910360139368523e-46 ], [ 945.8125 , 1.9105752946807973e-46 ], [ 945.875 , 1.910790474225038e-46 ], [ 945.9375 , 1.9110056505267554e-46 ], 
                    [ 946.0 , 1.9112205740155273e-46 ], [ 946.0625 , 1.9114355216758807e-46 ], [ 946.125 , 1.9116504935106018e-46 ], [ 946.1875 , 1.911865489522368e-46 ], 
                    [ 946.25 , 1.9120805097139263e-46 ], [ 946.3125 , 1.9122955540879827e-46 ], [ 946.375 , 1.9125106226472286e-46 ], [ 946.4375 , 1.912725715394426e-46 ], 
                    [ 946.5 , 1.912940832332281e-46 ], [ 946.5625 , 1.913155973463541e-46 ], [ 946.625 , 1.9133711387908606e-46 ], [ 946.6875 , 1.9135863283170135e-46 ], 
                    [ 946.75 , 1.9138015420447226e-46 ], [ 946.8125 , 1.914016779976682e-46 ], [ 946.875 , 1.9142320421156542e-46 ], [ 946.9375 , 1.914447328464281e-46 ], 
                    [ 947.0 , 1.9146626390253936e-46 ], [ 947.0625 , 1.9148779738016195e-46 ], [ 947.125 , 1.9150933327957506e-46 ], [ 947.1875 , 1.9153087160104836e-46 ], 
                    [ 947.25 , 1.9155241234485145e-46 ], [ 947.3125 , 1.915739555112636e-46 ], [ 947.375 , 1.9159550110054776e-46 ], [ 947.4375 , 1.9161704911298593e-46 ], 
                    [ 947.5 , 1.9163859954884796e-46 ], [ 947.5625 , 1.9166015240840222e-46 ], [ 947.625 , 1.9168170769192272e-46 ], [ 947.6875 , 1.9170326539968754e-46 ], 
                    [ 947.75 , 1.9172482553196113e-46 ], [ 947.8125 , 1.9174638808902294e-46 ], [ 947.875 , 1.9176795307114165e-46 ], [ 947.9375 , 1.9178952047858997e-46 ], 
                    [ 948.0 , 1.918110903116461e-46 ], [ 948.0625 , 1.9183266257057607e-46 ], [ 948.125 , 1.9185421577453724e-46 ], [ 948.1875 , 1.9187576550909193e-46 ], 
                    [ 948.25 , 1.918973176641866e-46 ], [ 948.3125 , 1.919188722400972e-46 ], [ 948.375 , 1.919404292370916e-46 ], [ 948.4375 , 1.9196198865544437e-46 ], 
                    [ 948.5 , 1.9198355049542626e-46 ], [ 948.5625 , 1.9200511475730644e-46 ], [ 948.625 , 1.9202668144136113e-46 ], [ 948.6875 , 1.920482505478623e-46 ], 
                    [ 948.75 , 1.92069822077078e-46 ], [ 948.8125 , 1.9209139602928716e-46 ], [ 948.875 , 1.921129724047552e-46 ], [ 948.9375 , 1.921345512037569e-46 ], 
                    [ 949.0 , 1.921561324265646e-46 ], [ 949.0625 , 1.921777160734505e-46 ], [ 949.125 , 1.9219930214468823e-46 ], [ 949.1875 , 1.922208906405488e-46 ], 
                    [ 949.25 , 1.9224248156130176e-46 ], [ 949.3125 , 1.9226407490722495e-46 ], [ 949.375 , 1.922856706785867e-46 ], [ 949.4375 , 1.923072688756594e-46 ], 
                    [ 949.5 , 1.923288694987197e-46 ], [ 949.5625 , 1.9235047254803728e-46 ], [ 949.625 , 1.923720780238833e-46 ], [ 949.6875 , 1.923936859265304e-46 ], 
                    [ 949.75 , 1.9241529625625656e-46 ], [ 949.8125 , 1.9243690901332618e-46 ], [ 949.875 , 1.9245852419801607e-46 ], [ 949.9375 , 1.924801418106016e-46 ], 
                    [ 950.0 , 1.9250176185135415e-46 ], [ 950.0625 , 1.9252338432053957e-46 ], [ 950.125 , 1.925450092184403e-46 ], [ 950.1875 , 1.9256663654532353e-46 ], 
                    [ 950.25 , 1.925882588155909e-46 ], [ 950.3125 , 1.9260987235820402e-46 ], [ 950.375 , 1.926314883264321e-46 ], [ 950.4375 , 1.9265310672055146e-46 ], 
                    [ 950.5 , 1.9267472754082884e-46 ], [ 950.5625 , 1.9269635078754074e-46 ], [ 950.625 , 1.9271797646095794e-46 ], [ 950.6875 , 1.9273960456135156e-46 ], 
                    [ 950.75 , 1.9276123508899935e-46 ], [ 950.8125 , 1.927828680441655e-46 ], [ 950.875 , 1.9280450342712665e-46 ], [ 950.9375 , 1.9282614123815788e-46 ], 
                    [ 951.0 , 1.9284778147752763e-46 ], [ 951.0625 , 1.9286942414550843e-46 ], [ 951.125 , 1.9289106924237695e-46 ], [ 951.1875 , 1.9291271676840166e-46 ], 
                    [ 951.25 , 1.9293436672385652e-46 ], [ 951.3125 , 1.9295601910901143e-46 ], [ 951.375 , 1.929776739241474e-46 ], [ 951.4375 , 1.9299933116952737e-46 ], 
                    [ 951.5 , 1.930209908454297e-46 ], [ 951.5625 , 1.930426529521271e-46 ], [ 951.625 , 1.9306431748988686e-46 ], [ 951.6875 , 1.930859844589929e-46 ], 
                    [ 951.75 , 1.9310765385970284e-46 ], [ 951.8125 , 1.93129325692302e-46 ], [ 951.875 , 1.9315099995706195e-46 ], [ 951.9375 , 1.9317267665424872e-46 ], 
                    [ 952.0 , 1.931943557841422e-46 ], [ 952.0625 , 1.932160373470085e-46 ], [ 952.125 , 1.93237721343129e-46 ], [ 952.1875 , 1.932594077727698e-46 ], 
                    [ 952.25 , 1.9328109663620812e-46 ], [ 952.3125 , 1.933027879337158e-46 ], [ 952.375 , 1.933244816655674e-46 ], [ 952.4375 , 1.9334618269600056e-46 ], 
                    [ 952.5 , 1.933678864702753e-46 ], [ 952.5625 , 1.9338959268087064e-46 ], [ 952.625 , 1.9341130132806557e-46 ], [ 952.6875 , 1.9343301241212946e-46 ], 
                    [ 952.75 , 1.9345472593333998e-46 ], [ 952.8125 , 1.9347644189196657e-46 ], [ 952.875 , 1.9349816028828426e-46 ], [ 952.9375 , 1.935198811225653e-46 ], 
                    [ 953.0 , 1.935416043950847e-46 ], [ 953.0625 , 1.9356333010611895e-46 ], [ 953.125 , 1.9358505825593628e-46 ], [ 953.1875 , 1.9360678884481453e-46 ], 
                    [ 953.25 , 1.936285218730262e-46 ], [ 953.3125 , 1.9365025734084363e-46 ], [ 953.375 , 1.9367199524854493e-46 ], [ 953.4375 , 1.9369373559639974e-46 ], 
                    [ 953.5 , 1.9371547838468346e-46 ], [ 953.5625 , 1.9373722361366714e-46 ], [ 953.625 , 1.9375897128362903e-46 ], [ 953.6875 , 1.937807213948444e-46 ], 
                    [ 953.75 , 1.9380247394758327e-46 ], [ 953.8125 , 1.938242289421183e-46 ], [ 953.875 , 1.9384598637873042e-46 ], [ 953.9375 , 1.9386774625768826e-46 ], 
                    [ 954.0 , 1.9388950857927017e-46 ], [ 954.0625 , 1.9391127334374064e-46 ], [ 954.125 , 1.939330405513836e-46 ], [ 954.1875 , 1.9395481020247188e-46 ], 
                    [ 954.25 , 1.939765822972798e-46 ], [ 954.3125 , 1.9399835683607884e-46 ], [ 954.375 , 1.9402013381914204e-46 ], [ 954.4375 , 1.940419132467507e-46 ], 
                    [ 954.5 , 1.9406369511917226e-46 ], [ 954.5625 , 1.9408552251088207e-46 ], [ 954.625 , 1.941074166755076e-46 ], [ 954.6875 , 1.9412931330993945e-46 ], 
                    [ 954.75 , 1.941512124144631e-46 ], [ 954.8125 , 1.94173113989353e-46 ], [ 954.875 , 1.941950180348906e-46 ], [ 954.9375 , 1.9421692455135327e-46 ], 
                    [ 955.0 , 1.9423883353901833e-46 ], [ 955.0625 , 1.9426074499816873e-46 ], [ 955.125 , 1.9428265892907773e-46 ], [ 955.1875 , 1.9430457533202825e-46 ], 
                    [ 955.25 , 1.9432649420729507e-46 ], [ 955.3125 , 1.94348415555164e-46 ], [ 955.375 , 1.9437033937590563e-46 ], [ 955.4375 , 1.9439226566980725e-46 ], 
                    [ 955.5 , 1.9441419443714093e-46 ], [ 955.5625 , 1.944361256781884e-46 ], [ 955.625 , 1.9445805939323018e-46 ], [ 955.6875 , 1.9447999558254388e-46 ], 
                    [ 955.75 , 1.945019342464073e-46 ], [ 955.8125 , 1.9452387538510237e-46 ], [ 955.875 , 1.945458189989068e-46 ], [ 955.9375 , 1.9456776508809712e-46 ], 
                    [ 956.0 , 1.9458971365296083e-46 ], [ 956.0625 , 1.946116646937675e-46 ], [ 956.125 , 1.9463361821080198e-46 ], [ 956.1875 , 1.9465557420433945e-46 ], 
                    [ 956.25 , 1.9467753267466486e-46 ], [ 956.3125 , 1.946994936220534e-46 ], [ 956.375 , 1.9472145704678727e-46 ], [ 956.4375 , 1.947434229491432e-46 ], 
                    [ 956.5 , 1.9476539132940346e-46 ], [ 956.5625 , 1.947873621878475e-46 ], [ 956.625 , 1.9480933552474937e-46 ], [ 956.6875 , 1.94831311340397e-46 ], 
                    [ 956.75 , 1.9485413611801443e-46 ], [ 956.8125 , 1.9487719725317546e-46 ], [ 956.875 , 1.949002611176394e-46 ], [ 956.9375 , 1.9492332771173065e-46 ], 
                    [ 957.0 , 1.9494639703576662e-46 ], [ 957.0625 , 1.9496946909007883e-46 ], [ 957.125 , 1.949925438749862e-46 ], [ 957.1875 , 1.9501562139080915e-46 ], 
                    [ 957.25 , 1.9503870163787367e-46 ], [ 957.3125 , 1.950617846165043e-46 ], [ 957.375 , 1.9508487032702304e-46 ], [ 957.4375 , 1.9510795876975457e-46 ], 
                    [ 957.5 , 1.9513104994501803e-46 ], [ 957.5625 , 1.9515414385314105e-46 ], [ 957.625 , 1.9517724049444567e-46 ], [ 957.6875 , 1.9520033986925534e-46 ], 
                    [ 957.75 , 1.952234419778936e-46 ], [ 957.8125 , 1.9524654682068255e-46 ], [ 957.875 , 1.9526965439794862e-46 ], [ 957.9375 , 1.9529276471001545e-46 ], 
                    [ 958.0 , 1.9531587775720526e-46 ], [ 958.0625 , 1.9533899353983897e-46 ], [ 958.125 , 1.953621120582418e-46 ], [ 958.1875 , 1.95385233312743e-46 ], 
                    [ 958.25 , 1.9540835730366085e-46 ], [ 958.3125 , 1.95431484031322e-46 ], [ 958.375 , 1.954546134960448e-46 ], [ 958.4375 , 1.954777456981601e-46 ], 
                    [ 958.5 , 1.955008806379891e-46 ], [ 958.5625 , 1.9552401831585584e-46 ], [ 958.625 , 1.955471587320871e-46 ], [ 958.6875 , 1.955703018869987e-46 ], 
                    [ 958.75 , 1.955934477809216e-46 ], [ 958.8125 , 1.956165964141829e-46 ], [ 958.875 , 1.9564686647038945e-46 ], [ 958.9375 , 1.957512387251173e-46 ], 
                    [ 959.0 , 1.9585566665958835e-46 ], [ 959.0625 , 1.9596015030350898e-46 ], [ 959.125 , 1.9606468968659595e-46 ], [ 959.1875 , 1.9616928483858453e-46 ], 
                    [ 959.25 , 1.9627393578922867e-46 ], [ 959.3125 , 1.963786425682941e-46 ], [ 959.375 , 1.9648340520556507e-46 ], [ 959.4375 , 1.965882237308377e-46 ], 
                    [ 959.5 , 1.9669309817392664e-46 ], [ 959.5625 , 1.9679802856466676e-46 ], [ 959.625 , 1.9690301493290178e-46 ], [ 959.6875 , 1.9700805730849276e-46 ], 
                    [ 959.75 , 1.971131557213195e-46 ], [ 959.8125 , 1.9721831020127634e-46 ], [ 959.875 , 1.9732352077827355e-46 ], [ 959.9375 , 1.9742878748223735e-46 ], 
                    [ 960.0 , 1.9753411034311003e-46 ], [ 960.0625 , 1.9763948939084964e-46 ], [ 960.125 , 1.9774492465543322e-46 ], [ 960.1875 , 1.978504161668439e-46 ], 
                    [ 960.25 , 1.97955963955092e-46 ], [ 960.3125 , 1.9806156805020113e-46 ], [ 960.375 , 1.981672284822108e-46 ], [ 960.4375 , 1.9827294528116826e-46 ], 
                    [ 960.5 , 1.983787184771493e-46 ], [ 960.5625 , 1.984845481002402e-46 ], [ 960.625 , 1.985904341805405e-46 ], [ 960.6875 , 1.9869637674816992e-46 ], 
                    [ 960.75 , 1.988023758332644e-46 ], [ 960.8125 , 1.9890843146597013e-46 ], [ 960.875 , 1.9901454367645663e-46 ], [ 960.9375 , 1.991207124949108e-46 ], 
                    [ 961.0 , 1.992269379515243e-46 ], [ 961.0625 , 1.9928040255211148e-46 ], [ 961.125 , 1.9921794934957947e-46 ], [ 961.1875 , 1.991555157194827e-46 ], 
                    [ 961.25 , 1.990931016556818e-46 ], [ 961.3125 , 1.990307071520517e-46 ], [ 961.375 , 1.989683322024569e-46 ], [ 961.4375 , 1.9890597680077626e-46 ], 
                    [ 961.5 , 1.9884364094087508e-46 ], [ 961.5625 , 1.987813246166362e-46 ], [ 961.625 , 1.9871902782193307e-46 ], [ 961.6875 , 1.986567505506452e-46 ], 
                    [ 961.75 , 1.9859449279665828e-46 ], [ 961.8125 , 1.9853225455385157e-46 ], [ 961.875 , 1.9847003581611035e-46 ], [ 961.9375 , 1.9840783657732606e-46 ], 
                    [ 962.0 , 1.9834565683138366e-46 ], [ 962.0625 , 1.982834965721756e-46 ], [ 962.125 , 1.9822135579359635e-46 ], [ 962.1875 , 1.981592344895393e-46 ], 
                    [ 962.25 , 1.9809713265389704e-46 ], [ 962.3125 , 1.980350502805768e-46 ], [ 962.375 , 1.9797298736347352e-46 ], [ 962.4375 , 1.9791094389649404e-46 ], 
                    [ 962.5 , 1.978489198735371e-46 ], [ 962.5625 , 1.9778691528850916e-46 ], [ 962.625 , 1.9772493013532545e-46 ], [ 962.6875 , 1.97662964407892e-46 ], 
                    [ 962.75 , 1.9760101810011947e-46 ], [ 962.8125 , 1.9753909120592326e-46 ], [ 962.875 , 1.9747718371921645e-46 ], [ 962.9375 , 1.974152956339253e-46 ], 
                    [ 963.0 , 1.97353426943964e-46 ], [ 963.0625 , 1.9729157764325124e-46 ], [ 963.125 , 1.972297477257134e-46 ], [ 963.1875 , 1.9716793718527317e-46 ], 
                    [ 963.25 , 1.9714425470259644e-46 ], [ 963.3125 , 1.971654019308049e-46 ], [ 963.375 , 1.971865514274283e-46 ], [ 963.4375 , 1.9720770319271286e-46 ], 
                    [ 963.5 , 1.9722885722690195e-46 ], [ 963.5625 , 1.9725001353023468e-46 ], [ 963.625 , 1.9727117210296005e-46 ], [ 963.6875 , 1.9729233294531595e-46 ], 
                    [ 963.75 , 1.9731349605755e-46 ], [ 963.8125 , 1.973346614399057e-46 ], [ 963.875 , 1.9735582909262237e-46 ], [ 963.9375 , 1.9737699901594773e-46 ], 
                    [ 964.0 , 1.973981712101239e-46 ], [ 964.0625 , 1.9741934567539455e-46 ], [ 964.125 , 1.9744052241200327e-46 ], [ 964.1875 , 1.974617014201937e-46 ], 
                    [ 964.25 , 1.9748288270020947e-46 ], [ 964.3125 , 1.975040662522915e-46 ], [ 964.375 , 1.9752525207669196e-46 ], [ 964.4375 , 1.9754644017364614e-46 ], 
                    [ 964.5 , 1.975676305433992e-46 ], [ 964.5625 , 1.9758882318620056e-46 ], [ 964.625 , 1.9761001810228425e-46 ], [ 964.6875 , 1.976312152919025e-46 ], 
                    [ 964.75 , 1.976524147552964e-46 ], [ 964.8125 , 1.976736164927099e-46 ], [ 964.875 , 1.9769482050438827e-46 ], [ 964.9375 , 1.9771602679057125e-46 ], 
                    [ 965.0 , 1.9773723535150567e-46 ], [ 965.0625 , 1.977584461874369e-46 ], [ 965.125 , 1.977796592986076e-46 ], [ 965.1875 , 1.9780087468526326e-46 ], 
                    [ 965.25 , 1.9782209234764366e-46 ], [ 965.3125 , 1.9784331228599864e-46 ], [ 965.375 , 1.9786453450056804e-46 ], [ 965.4375 , 1.9788604847602513e-46 ], 
                    [ 965.5 , 1.979078245995225e-46 ], [ 965.5625 , 1.979296031193477e-46 ], [ 965.625 , 1.979513840357615e-46 ], [ 965.6875 , 1.979731673490277e-46 ], 
                    [ 965.75 , 1.979949530594129e-46 ], [ 965.8125 , 1.980167411671808e-46 ], [ 965.875 , 1.9803853167259245e-46 ], [ 965.9375 , 1.980603245759117e-46 ], 
                    [ 966.0 , 1.9808211987740524e-46 ], [ 966.0625 , 1.9810391757733416e-46 ], [ 966.125 , 1.981257176759666e-46 ], [ 966.1875 , 1.981475201735609e-46 ], 
                    [ 966.25 , 1.9816932507038387e-46 ], [ 966.3125 , 1.981911323667023e-46 ], [ 966.375 , 1.982129420627747e-46 ], [ 966.4375 , 1.9823475415886935e-46 ], 
                    [ 966.5 , 1.9825656865524884e-46 ], [ 966.5625 , 1.9827838555217885e-46 ], [ 966.625 , 1.9830020484991925e-46 ], [ 966.6875 , 1.983220265487371e-46 ], 
                    [ 966.75 , 1.9834385064889654e-46 ], [ 966.8125 , 1.983656771506661e-46 ], [ 966.875 , 1.9838750605430164e-46 ], [ 966.9375 , 1.984093373600731e-46 ], 
                    [ 967.0 , 1.9843117106824192e-46 ], [ 967.0625 , 1.9845300717907534e-46 ], [ 967.125 , 1.9847484569283642e-46 ], [ 967.1875 , 1.984966866097923e-46 ], 
                    [ 967.25 , 1.98518529930199e-46 ], [ 967.3125 , 1.985403756543281e-46 ], [ 967.375 , 1.9856222378244406e-46 ], [ 967.4375 , 1.9858407431481147e-46 ], 
                    [ 967.5 , 1.9860592725168924e-46 ], [ 967.5625 , 1.9862778259335042e-46 ], [ 967.625 , 1.9864967253108833e-46 ], [ 967.6875 , 1.9867159515672493e-46 ], 
                    [ 967.75 , 1.986935202017079e-46 ], [ 967.8125 , 1.987154476662971e-46 ], [ 967.875 , 1.9873737755076384e-46 ], [ 967.9375 , 1.9875930985537375e-46 ], 
                    [ 968.0 , 1.9878124458039386e-46 ], [ 968.0625 , 1.988031817260886e-46 ], [ 968.125 , 1.988251212927334e-46 ], [ 968.1875 , 1.988470632805871e-46 ], 
                    [ 968.25 , 1.9886900768991684e-46 ], [ 968.3125 , 1.988909545209955e-46 ], [ 968.375 , 1.9891290377408897e-46 ], [ 968.4375 , 1.9893485544946023e-46 ], 
                    [ 968.5 , 1.989568095473809e-46 ], [ 968.5625 , 1.9897876606811548e-46 ], [ 968.625 , 1.9900072501193557e-46 ], [ 968.6875 , 1.9902268637910434e-46 ], 
                    [ 968.75 , 1.9904465016989066e-46 ], [ 968.8125 , 1.9906661638456057e-46 ], [ 968.875 , 1.9908858502338302e-46 ], [ 968.9375 , 1.9911055608662544e-46 ], 
                    [ 969.0 , 1.9913252957455974e-46 ], [ 969.0625 , 1.991545054874449e-46 ], [ 969.125 , 1.991764838255529e-46 ], [ 969.1875 , 1.9919846458915273e-46 ], 
                    [ 969.25 , 1.9922044777851208e-46 ], [ 969.3125 , 1.992424333938958e-46 ], [ 969.375 , 1.992644214355702e-46 ], [ 969.4375 , 1.99286411903813e-46 ], 
                    [ 969.5 , 1.9930840479887778e-46 ], [ 969.5625 , 1.993304001210409e-46 ], [ 969.625 , 1.9935239787057017e-46 ], [ 969.6875 , 1.993743980477307e-46 ], 
                    [ 969.75 , 1.9939640065279456e-46 ], [ 969.8125 , 1.9941840322951077e-46 ], [ 969.875 , 1.994404049009739e-46 ], [ 969.9375 , 1.9946240899986223e-46 ], 
                    [ 970.0 , 1.994844155264465e-46 ], [ 970.0625 , 1.9950642448099303e-46 ], [ 970.125 , 1.9952843586376693e-46 ], [ 970.1875 , 1.9955044967504317e-46 ], 
                    [ 970.25 , 1.9957246591508123e-46 ], [ 970.3125 , 1.9959448458415607e-46 ], [ 970.375 , 1.996165056825343e-46 ], [ 970.4375 , 1.9963852921047835e-46 ], 
                    [ 970.5 , 1.9966055516826325e-46 ], [ 970.5625 , 1.996825835561543e-46 ], [ 970.625 , 1.9970461437441677e-46 ], [ 970.6875 , 1.997266476233245e-46 ], 
                    [ 970.75 , 1.9974868330314274e-46 ], [ 970.8125 , 1.99770721414137e-46 ], [ 970.875 , 1.99792761956581e-46 ], [ 970.9375 , 1.998148049307389e-46 ], 
                    [ 971.0 , 1.998368503368803e-46 ], [ 971.0625 , 1.9985889817527786e-46 ], [ 971.125 , 1.9988094844618993e-46 ], [ 971.1875 , 1.9990300114989343e-46 ], 
                    [ 971.25 , 1.9992505628665395e-46 ], [ 971.3125 , 1.9994711385673705e-46 ], [ 971.375 , 1.9996917386041693e-46 ], [ 971.4375 , 1.9999123629795775e-46 ], 
                    [ 971.5 , 2.000133011696281e-46 ], [ 971.5625 , 2.0003536847570356e-46 ], [ 971.625 , 2.0005743821644288e-46 ], [ 971.6875 , 2.0007951039211747e-46 ], 
                    [ 971.75 , 2.001015850030016e-46 ], [ 971.8125 , 2.001236620493541e-46 ], [ 971.875 , 2.0014574153145634e-46 ], [ 971.9375 , 2.001678234495658e-46 ], 
                    [ 972.0 , 2.0018990181454614e-46 ], [ 972.0625 , 2.0021196501444745e-46 ], [ 972.125 , 2.002340306459596e-46 ], [ 972.1875 , 2.002560987093577e-46 ], 
                    [ 972.25 , 2.0027816920490548e-46 ], [ 972.3125 , 2.003002421328753e-46 ], [ 972.375 , 2.003223174935295e-46 ], [ 972.4375 , 2.003443952871391e-46 ], 
                    [ 972.5 , 2.0036647551397214e-46 ], [ 972.5625 , 2.0038855817430118e-46 ], [ 972.625 , 2.0041064326838437e-46 ], [ 972.6875 , 2.004327307964985e-46 ], 
                    [ 972.75 , 2.0045482075890617e-46 ], [ 972.8125 , 2.004769131558813e-46 ], [ 972.875 , 2.0049900798768935e-46 ], [ 972.9375 , 2.0052110525459876e-46 ], 
                    [ 973.0 , 2.0054320495687784e-46 ], [ 973.0625 , 2.0056530709479358e-46 ], [ 973.125 , 2.0058741166861728e-46 ], [ 973.1875 , 2.0060951867861308e-46 ], 
                    [ 973.25 , 2.006316281250566e-46 ], [ 973.3125 , 2.0065374000820933e-46 ], [ 973.375 , 2.0067585432834547e-46 ], [ 973.4375 , 2.0069797108572504e-46 ], 
                    [ 973.5 , 2.007200902806252e-46 ], [ 973.5625 , 2.007422119133146e-46 ], [ 973.625 , 2.007643359840562e-46 ], [ 973.6875 , 2.00786462493123e-46 ], 
                    [ 973.75 , 2.0080859144078093e-46 ], [ 973.8125 , 2.008307228273029e-46 ], [ 973.875 , 2.0085285665295067e-46 ], [ 973.9375 , 2.0087499291800015e-46 ], 
                    [ 974.0 , 2.008971316227188e-46 ], [ 974.0625 , 2.009192727673712e-46 ], [ 974.125 , 2.0094141635222905e-46 ], [ 974.1875 , 2.0096356227224422e-46 ], 
                    [ 974.25 , 2.009856808152942e-46 ], [ 974.3125 , 2.010078017927667e-46 ], [ 974.375 , 2.010299252049268e-46 ], [ 974.4375 , 2.0105205105204404e-46 ], 
                    [ 974.5 , 2.0107417933438633e-46 ], [ 974.5625 , 2.0109631005222024e-46 ], [ 974.625 , 2.0111844320581676e-46 ], [ 974.6875 , 2.011405787954439e-46 ], 
                    [ 974.75 , 2.011627168213669e-46 ], [ 974.8125 , 2.0118485728385693e-46 ], [ 974.875 , 2.012070001831791e-46 ], [ 974.9375 , 2.012291455196032e-46 ], 
                    [ 975.0 , 2.012512932933988e-46 ], [ 975.0625 , 2.0127344350482844e-46 ], [ 975.125 , 2.01295596154169e-46 ], [ 975.1875 , 2.0131775124168316e-46 ], 
                    [ 975.25 , 2.013399087676406e-46 ], [ 975.3125 , 2.0136206873231125e-46 ], [ 975.375 , 2.0138423113596053e-46 ], [ 975.4375 , 2.0140639597885837e-46 ], 
                    [ 975.5 , 2.014285632612747e-46 ], [ 975.5625 , 2.014507329834708e-46 ], [ 975.625 , 2.0147290514572666e-46 ], [ 975.6875 , 2.014950797483052e-46 ], 
                    [ 975.75 , 2.01517256791472e-46 ], [ 975.8125 , 2.0153943627550296e-46 ], [ 975.875 , 2.0156161820065953e-46 ], [ 975.9375 , 2.015838025672132e-46 ], 
                    [ 976.0 , 2.016059893754327e-46 ], [ 976.0625 , 2.0162817862558397e-46 ], [ 976.125 , 2.0165037031794434e-46 ], [ 976.1875 , 2.0167256445277252e-46 ], 
                    [ 976.25 , 2.016947610303431e-46 ], [ 976.3125 , 2.0171696005092637e-46 ], [ 976.375 , 2.0173916151478548e-46 ], [ 976.4375 , 2.017613433847784e-46 ], 
                    [ 976.5 , 2.0178351700632964e-46 ], [ 976.5625 , 2.0180569306477026e-46 ], [ 976.625 , 2.018278715603623e-46 ], [ 976.6875 , 2.0185005249337933e-46 ], 
                    [ 976.75 , 2.0187223586408214e-46 ], [ 976.8125 , 2.0189442167274858e-46 ], [ 976.875 , 2.0191660991963515e-46 ], [ 976.9375 , 2.019388006050198e-46 ], 
                    [ 977.0 , 2.0196099372916494e-46 ], [ 977.0625 , 2.0198318929234127e-46 ], [ 977.125 , 2.0200538729481117e-46 ], [ 977.1875 , 2.0202758773685137e-46 ], 
                    [ 977.25 , 2.020497906187228e-46 ], [ 977.3125 , 2.0207199594069925e-46 ], [ 977.375 , 2.0209420370304325e-46 ], [ 977.4375 , 2.0211641390602724e-46 ], 
                    [ 977.5 , 2.0213862654991803e-46 ], [ 977.5625 , 2.0216084163498535e-46 ], [ 977.625 , 2.0218305916149453e-46 ], [ 977.6875 , 2.022052791297154e-46 ], 
                    [ 977.75 , 2.022275015399148e-46 ], [ 977.8125 , 2.0224972639236265e-46 ], [ 977.875 , 2.0227195368732864e-46 ], [ 977.9375 , 2.0229418342508136e-46 ], 
                    [ 978.0 , 2.0231641560588627e-46 ], [ 978.0625 , 2.0233865023001332e-46 ], [ 978.125 , 2.0236088729773112e-46 ], [ 978.1875 , 2.023831268093067e-46 ], 
                    [ 978.25 , 2.0240536876501303e-46 ], [ 978.3125 , 2.024276131651114e-46 ], [ 978.375 , 2.0244986000987922e-46 ], [ 978.4375 , 2.0247210929958075e-46 ], 
                    [ 978.5 , 2.024943610344833e-46 ], [ 978.5625 , 2.0251661521485414e-46 ], [ 978.625 , 2.0253886291938593e-46 ], [ 978.6875 , 2.025610880902079e-46 ], 
                    [ 978.75 , 2.02583315699866e-46 ], [ 978.8125 , 2.0260554574862063e-46 ], [ 978.875 , 2.026277782367423e-46 ], [ 978.9375 , 2.0265001316449875e-46 ], 
                    [ 979.0 , 2.0267225053215764e-46 ], [ 979.0625 , 2.0269449033998812e-46 ], [ 979.125 , 2.0271673258825514e-46 ], [ 979.1875 , 2.0273897727722648e-46 ], 
                    [ 979.25 , 2.0276122440717424e-46 ], [ 979.3125 , 2.0278347397836344e-46 ], [ 979.375 , 2.0280572599106054e-46 ], [ 979.4375 , 2.028279804455363e-46 ], 
                    [ 979.5 , 2.0285023734205435e-46 ], [ 979.5625 , 2.028724966808884e-46 ], [ 979.625 , 2.0289475846230217e-46 ], [ 979.6875 , 2.0291702268656513e-46 ], 
                    [ 979.75 , 2.0293928935394678e-46 ], [ 979.8125 , 2.0296155846471374e-46 ], [ 979.875 , 2.0298383001913415e-46 ], [ 979.9375 , 2.030061040174733e-46 ], 
                    [ 980.0 , 2.030283804600022e-46 ], [ 980.0625 , 2.0305065934699194e-46 ], [ 980.125 , 2.030729406787094e-46 ], [ 980.1875 , 2.03095224455417e-46 ], 
                    [ 980.25 , 2.031175106773888e-46 ], [ 980.3125 , 2.031397993448932e-46 ], [ 980.375 , 2.0316209045819414e-46 ], [ 980.4375 , 2.0318438401756583e-46 ], 
                    [ 980.5 , 2.0320668002327234e-46 ], [ 980.5625 , 2.03228978475585e-46 ], [ 980.625 , 2.0325127937476942e-46 ], [ 980.6875 , 2.0327358272109696e-46 ], 
                    [ 980.75 , 2.032958885148318e-46 ], [ 980.8125 , 2.0331819675624688e-46 ], [ 980.875 , 2.0334048050824466e-46 ], [ 980.9375 , 2.0336275877498604e-46 ], 
                    [ 981.0 , 2.033850394825668e-46 ], [ 981.0625 , 2.0340732263125446e-46 ], [ 981.125 , 2.034296082213121e-46 ], [ 981.1875 , 2.034518962530115e-46 ], 
                    [ 981.25 , 2.0347418672662018e-46 ], [ 981.3125 , 2.0349647964239996e-46 ], [ 981.375 , 2.03518775000627e-46 ], [ 981.4375 , 2.035410728015646e-46 ], 
                    [ 981.5 , 2.035633730454775e-46 ], [ 981.5625 , 2.0358567573263763e-46 ], [ 981.625 , 2.0360798086331278e-46 ], [ 981.6875 , 2.0363028843776626e-46 ], 
                    [ 981.75 , 2.036525984562716e-46 ], [ 981.8125 , 2.0367491091909226e-46 ], [ 981.875 , 2.036972258264975e-46 ], [ 981.9375 , 2.0371954317875513e-46 ], 
                    [ 982.0 , 2.03741862976133e-46 ], [ 982.0625 , 2.037641852188975e-46 ], [ 982.125 , 2.0378650990731956e-46 ], [ 982.1875 , 2.038088370416642e-46 ], 
                    [ 982.25 , 2.038311666222037e-46 ], [ 982.3125 , 2.0385349864920464e-46 ], [ 982.375 , 2.038758331229264e-46 ], [ 982.4375 , 2.038981700436515e-46 ], 
                    [ 982.5 , 2.03920509411635e-46 ], [ 982.5625 , 2.0394285122715517e-46 ], [ 982.625 , 2.0396519549047433e-46 ], [ 982.6875 , 2.0398754220186076e-46 ], 
                    [ 982.75 , 2.0400989136158984e-46 ], [ 982.8125 , 2.0403224296991678e-46 ], [ 982.875 , 2.040545970271186e-46 ], [ 982.9375 , 2.0407695353346647e-46 ], 
                    [ 983.0 , 2.0409931248921996e-46 ], [ 983.0625 , 2.041216666066066e-46 ], [ 983.125 , 2.041439943348774e-46 ], [ 983.1875 , 2.041663245054521e-46 ], 
                    [ 983.25 , 2.0418865711859937e-46 ], [ 983.3125 , 2.042109921745893e-46 ], [ 983.375 , 2.0423332967368326e-46 ], [ 983.4375 , 2.0425566961615285e-46 ], 
                    [ 983.5 , 2.0427801200226387e-46 ], [ 983.5625 , 2.0430035683228075e-46 ], [ 983.625 , 2.043227041064766e-46 ], [ 983.6875 , 2.043450538251173e-46 ], 
                    [ 983.75 , 2.0436740598846597e-46 ], [ 983.8125 , 2.0438976059679283e-46 ], [ 983.875 , 2.0441211765036533e-46 ], [ 983.9375 , 2.044344771494539e-46 ], 
                    [ 984.0 , 2.0445683909432022e-46 ], [ 984.0625 , 2.044792034852361e-46 ], [ 984.125 , 2.0450157032246627e-46 ], [ 984.1875 , 2.0452393960628125e-46 ], 
                    [ 984.25 , 2.045463113369457e-46 ], [ 984.3125 , 2.0456868551473017e-46 ], [ 984.375 , 2.0459106213990092e-46 ], [ 984.4375 , 2.0461344121272414e-46 ], 
                    [ 984.5 , 2.0463582273346906e-46 ], [ 984.5625 , 2.046582067024064e-46 ], [ 984.625 , 2.0468059311979806e-46 ], [ 984.6875 , 2.047029819859148e-46 ], 
                    [ 984.75 , 2.0472537330102593e-46 ], [ 984.8125 , 2.0474776706539643e-46 ], [ 984.875 , 2.0477016327929708e-46 ], [ 984.9375 , 2.047925619429944e-46 ], 
                    [ 985.0 , 2.0481496305675638e-46 ], [ 985.0625 , 2.0483736662084955e-46 ], [ 985.125 , 2.0485977263554626e-46 ], [ 985.1875 , 2.048821811011088e-46 ], 
                    [ 985.25 , 2.0490459201781107e-46 ], [ 985.3125 , 2.0492698593272237e-46 ], [ 985.375 , 2.049493663822091e-46 ], [ 985.4375 , 2.0497174927590694e-46 ], 
                    [ 985.5 , 2.0499413461408283e-46 ], [ 985.5625 , 2.050165223969994e-46 ], [ 985.625 , 2.0503891262492794e-46 ], [ 985.6875 , 2.0506130529813265e-46 ], 
                    [ 985.75 , 2.05083700416882e-46 ], [ 985.8125 , 2.051060979814445e-46 ], [ 985.875 , 2.0512849799208727e-46 ], [ 985.9375 , 2.051509004490746e-46 ], 
                    [ 986.0 , 2.051733053526736e-46 ], [ 986.0625 , 2.051957127031529e-46 ], [ 986.125 , 2.052181225007842e-46 ], [ 986.1875 , 2.0524053474582446e-46 ], 
                    [ 986.25 , 2.0526294943854978e-46 ], [ 986.3125 , 2.052853665792231e-46 ], [ 986.375 , 2.0530778616811457e-46 ], [ 986.4375 , 2.0533020820548875e-46 ], 
                    [ 986.5 , 2.0535263269161747e-46 ], [ 986.5625 , 2.053750596267622e-46 ], [ 986.625 , 2.0539748901119484e-46 ], [ 986.6875 , 2.0541992084518287e-46 ], 
                    [ 986.75 , 2.0544235512898945e-46 ], [ 986.8125 , 2.0546479186288505e-46 ], [ 986.875 , 2.0548723104714013e-46 ], [ 986.9375 , 2.0550967268201796e-46 ], 
                    [ 987.0 , 2.055321167677862e-46 ], [ 987.0625 , 2.0555456330471385e-46 ], [ 987.125 , 2.0557701229307168e-46 ], [ 987.1875 , 2.0559946373312145e-46 ], 
                    [ 987.25 , 2.0562191762513394e-46 ], [ 987.3125 , 2.056443739693798e-46 ], [ 987.375 , 2.0566683276612245e-46 ], [ 987.4375 , 2.056892940156298e-46 ], 
                    [ 987.5 , 2.057117577181725e-46 ], [ 987.5625 , 2.0573419392406653e-46 ], [ 987.625 , 2.0575662590184017e-46 ], [ 987.6875 , 2.0577906032545434e-46 ], 
                    [ 987.75 , 2.058014971951815e-46 ], [ 987.8125 , 2.0582393651128405e-46 ], [ 987.875 , 2.0584637827403163e-46 ], [ 987.9375 , 2.058688224836895e-46 ], 
                    [ 988.0 , 2.0589126914052452e-46 ], [ 988.0625 , 2.05913718244802e-46 ], [ 988.125 , 2.0593616979679027e-46 ], [ 988.1875 , 2.0595862379675625e-46 ], 
                    [ 988.25 , 2.0598108024496826e-46 ], [ 988.3125 , 2.060035391416904e-46 ], [ 988.375 , 2.0602600048719393e-46 ], [ 988.4375 , 2.0604846428174e-46 ], 
                    [ 988.5 , 2.0607093052559865e-46 ], [ 988.5625 , 2.0609339921903688e-46 ], [ 988.625 , 2.061158703623233e-46 ], [ 988.6875 , 2.06138343955722e-46 ], 
                    [ 988.75 , 2.061608199995031e-46 ], [ 988.8125 , 2.0618329849393234e-46 ], [ 988.875 , 2.062057794392754e-46 ], [ 988.9375 , 2.0622826283580108e-46 ], 
                    [ 989.0 , 2.0625074868377506e-46 ], [ 989.0625 , 2.0627323698347202e-46 ], [ 989.125 , 2.0629572773514898e-46 ], [ 989.1875 , 2.0631822093907772e-46 ], 
                    [ 989.25 , 2.063407165955256e-46 ], [ 989.3125 , 2.0636321470476007e-46 ], [ 989.375 , 2.0638571526705294e-46 ], [ 989.4375 , 2.0640821828265845e-46 ], 
                    [ 989.5 , 2.0643072375186316e-46 ], [ 989.5625 , 2.0645323167491706e-46 ], [ 989.625 , 2.0647574205209937e-46 ], [ 989.6875 , 2.064982548836689e-46 ], 
                    [ 989.75 , 2.0652077001606674e-46 ], [ 989.8125 , 2.0654325214378625e-46 ], [ 989.875 , 2.0656573671894175e-46 ], [ 989.9375 , 2.065882237417953e-46 ], 
                    [ 990.0 , 2.066107132126163e-46 ], [ 990.0625 , 2.066332051316756e-46 ], [ 990.125 , 2.0665569949923245e-46 ], [ 990.1875 , 2.066781963155548e-46 ], 
                    [ 990.25 , 2.0670069558091366e-46 ], [ 990.3125 , 2.0672319729557267e-46 ], [ 990.375 , 2.067457014597956e-46 ], [ 990.4375 , 2.0676820807385635e-46 ], 
                    [ 990.5 , 2.0679071713801285e-46 ], [ 990.5625 , 2.0681322865253775e-46 ], [ 990.625 , 2.0683574261769772e-46 ], [ 990.6875 , 2.068582590337566e-46 ], 
                    [ 990.75 , 2.068807779009828e-46 ], [ 990.8125 , 2.0690329921964297e-46 ], [ 990.875 , 2.0692582299000406e-46 ], [ 990.9375 , 2.0694834921233595e-46 ], 
                    [ 991.0 , 2.069708778868996e-46 ], [ 991.0625 , 2.0699340901396206e-46 ], [ 991.125 , 2.0701594259379903e-46 ], [ 991.1875 , 2.070384786266688e-46 ], 
                    [ 991.25 , 2.070610171128413e-46 ], [ 991.3125 , 2.0708355805258503e-46 ], [ 991.375 , 2.0710610144616423e-46 ], [ 991.4375 , 2.0712864729384742e-46 ], 
                    [ 991.5 , 2.0715119559590328e-46 ], [ 991.5625 , 2.0717374635259457e-46 ], [ 991.625 , 2.071962995641944e-46 ], [ 991.6875 , 2.0721885523096707e-46 ], 
                    [ 991.75 , 2.072414133531754e-46 ], [ 991.8125 , 2.0726397393109557e-46 ], [ 991.875 , 2.0728653696499195e-46 ], [ 991.9375 , 2.0730910245512604e-46 ], 
                    [ 992.0 , 2.0733166633838672e-46 ], [ 992.0625 , 2.073541997167053e-46 ], [ 992.125 , 2.0737673554401513e-46 ], [ 992.1875 , 2.0739927382058097e-46 ], 
                    [ 992.25 , 2.07421814546669e-46 ], [ 992.3125 , 2.074443577225468e-46 ], [ 992.375 , 2.0746690334847784e-46 ], [ 992.4375 , 2.0748945142472976e-46 ], 
                    [ 992.5 , 2.0751200195156892e-46 ], [ 992.5625 , 2.0753455492926164e-46 ], [ 992.625 , 2.075571103580743e-46 ], [ 992.6875 , 2.075796682382747e-46 ], 
                    [ 992.75 , 2.076022285701264e-46 ], [ 992.8125 , 2.0762479135389585e-46 ], [ 992.875 , 2.076473565898539e-46 ], [ 992.9375 , 2.076699242782612e-46 ], 
                    [ 993.0 , 2.0769249441939015e-46 ], [ 993.0625 , 2.0771506701350445e-46 ], [ 993.125 , 2.0773764206086765e-46 ], [ 993.1875 , 2.0776021956174936e-46 ], 
                    [ 993.25 , 2.077827995164192e-46 ], [ 993.3125 , 2.078053819251394e-46 ], [ 993.375 , 2.0782796678817675e-46 ], [ 993.4375 , 2.078505541058009e-46 ], 
                    [ 993.5 , 2.0787314387827707e-46 ], [ 993.5625 , 2.0789573610587513e-46 ], [ 993.625 , 2.0791833078885444e-46 ], [ 993.6875 , 2.079409279274893e-46 ], 
                    [ 993.75 , 2.079635275220436e-46 ], [ 993.8125 , 2.0798612957278278e-46 ], [ 993.875 , 2.0800873407997533e-46 ], [ 993.9375 , 2.0803134104388814e-46 ], 
                    [ 994.0 , 2.080539504647868e-46 ], [ 994.0625 , 2.080765623429427e-46 ], [ 994.125 , 2.0809917667861707e-46 ], [ 994.1875 , 2.0812179347208138e-46 ], 
                    [ 994.25 , 2.081444078138952e-46 ], [ 994.3125 , 2.0816699285771738e-46 ], [ 994.375 , 2.0818958035216753e-46 ], [ 994.4375 , 2.0821217029751017e-46 ], 
                    [ 994.5 , 2.082347626940111e-46 ], [ 994.5625 , 2.082573575419379e-46 ], [ 994.625 , 2.08279954841552e-46 ], [ 994.6875 , 2.0830255459312545e-46 ], 
                    [ 994.75 , 2.083251567969183e-46 ], [ 994.8125 , 2.0834776145320262e-46 ], [ 994.875 , 2.0837036856224008e-46 ], [ 994.9375 , 2.0839297812429677e-46 ], 
                    [ 995.0 , 2.0841559013964037e-46 ], [ 995.0625 , 2.084382046085415e-46 ], [ 995.125 , 2.0846082153125904e-46 ], [ 995.1875 , 2.084834409080636e-46 ], 
                    [ 995.25 , 2.085060627392156e-46 ], [ 995.3125 , 2.0852868702498875e-46 ], [ 995.375 , 2.085513137656479e-46 ], [ 995.4375 , 2.0857394296145644e-46 ], 
                    [ 995.5 , 2.0859657461268526e-46 ], [ 995.5625 , 2.0861920871959188e-46 ], [ 995.625 , 2.086418452824517e-46 ], [ 995.6875 , 2.086644843015296e-46 ], 
                    [ 995.75 , 2.0868712577708775e-46 ], [ 995.8125 , 2.0870976970939857e-46 ], [ 995.875 , 2.0873241609872275e-46 ], [ 995.9375 , 2.0875506494532983e-46 ], 
                    [ 996.0 , 2.087777162494894e-46 ], [ 996.0625 , 2.088003700114652e-46 ], [ 996.125 , 2.0882302623152093e-46 ], [ 996.1875 , 2.088456849099277e-46 ], 
                    [ 996.25 , 2.088683460469494e-46 ], [ 996.3125 , 2.0889100964285567e-46 ], [ 996.375 , 2.0891367569791634e-46 ], [ 996.4375 , 2.0893634421238782e-46 ], 
                    [ 996.5 , 2.0895901235021593e-46 ], [ 996.5625 , 2.0898164861635747e-46 ], [ 996.625 , 2.090042873346557e-46 ], [ 996.6875 , 2.090269285053792e-46 ], 
                    [ 996.75 , 2.0904957212879065e-46 ], [ 996.8125 , 2.0907221820515873e-46 ], [ 996.875 , 2.0909486673474767e-46 ], [ 996.9375 , 2.0911751771782328e-46 ], 
                    [ 997.0 , 2.0914017115465128e-46 ], [ 997.0625 , 2.0916282704549606e-46 ], [ 997.125 , 2.0918548539062635e-46 ], [ 997.1875 , 2.0920814619030514e-46 ], 
                    [ 997.25 , 2.092308094448012e-46 ], [ 997.3125 , 2.0925347515437904e-46 ], [ 997.375 , 2.092761433193046e-46 ], [ 997.4375 , 2.092988139398424e-46 ], 
                    [ 997.5 , 2.0932148701626138e-46 ], [ 997.5625 , 2.0934416254882613e-46 ], [ 997.625 , 2.093668405378042e-46 ], [ 997.6875 , 2.0938952098345723e-46 ], 
                    [ 997.75 , 2.094122038860528e-46 ], [ 997.8125 , 2.0943488924586307e-46 ], [ 997.875 , 2.0945757706314677e-46 ], [ 997.9375 , 2.0948026733817306e-46 ], 
                    [ 998.0 , 2.0950296007121123e-46 ], [ 998.0625 , 2.0952565526252304e-46 ], [ 998.125 , 2.095483529123748e-46 ], [ 998.1875 , 2.095710530210373e-46 ], 
                    [ 998.25 , 2.0959375558877093e-46 ], [ 998.3125 , 2.096164606158466e-46 ], [ 998.375 , 2.096391681025277e-46 ], [ 998.4375 , 2.0966187804908215e-46 ], 
                    [ 998.5 , 2.09684590455778e-46 ], [ 998.5625 , 2.0970730532287865e-46 ], [ 998.625 , 2.097300226506552e-46 ], [ 998.6875 , 2.0975274243936825e-46 ], 
                    [ 998.75 , 2.0977546468928884e-46 ], [ 998.8125 , 2.097981551398383e-46 ], [ 998.875 , 2.09820845745586e-46 ], [ 998.9375 , 2.0984353880542384e-46 ], 
                    [ 999.0 , 2.098662343196188e-46 ], [ 999.0625 , 2.0988893228843033e-46 ], [ 999.125 , 2.099116327121313e-46 ], [ 999.1875 , 2.099343355909858e-46 ], 
                    [ 999.25 , 2.099570409252549e-46 ], [ 999.3125 , 2.099797487152101e-46 ], [ 999.375 , 2.100024589611125e-46 ], [ 999.4375 , 2.1002517166322925e-46 ], 
                    [ 999.5 , 2.1004788682182744e-46 ], [ 999.5625 , 2.1007060443716837e-46 ], [ 999.625 , 2.1009332450952065e-46 ], [ 999.6875 , 2.1011604703915302e-46 ], 
                    [ 999.75 , 2.1013877202632822e-46 ], [ 999.8125 , 2.101614994713077e-46 ], [ 999.875 , 2.101842293743661e-46 ], [ 999.9375 , 2.1020696173576038e-46 ], 
                    [ 1000.0 , 2.1022969655576683e-46 ], [ 1000.0625 , 2.1025243383464095e-46 ], [ 1000.125 , 2.1027517357265465e-46 ], [ 1000.1875 , 2.102979157700754e-46 ], 
                    [ 1000.25 , 2.1032066042716015e-46 ], [ 1000.3125 , 2.1034340754418546e-46 ], [ 1000.375 , 2.103661571214114e-46 ], [ 1000.4375 , 2.1038890915910403e-46 ], 
                    [ 1000.5 , 2.1041166365753546e-46 ], [ 1000.5625 , 2.1043442061696284e-46 ], [ 1000.625 , 2.1045718003765685e-46 ], [ 1000.6875 , 2.104799419198866e-46 ], 
                    [ 1000.75 , 2.1050270626391098e-46 ], [ 1000.8125 , 2.1052547307000656e-46 ], [ 1000.875 , 2.1054824233842925e-46 ], [ 1000.9375 , 2.1057101406945284e-46 ], 
                    [ 1001.0 , 2.1059378826333764e-46 ], [ 1001.0625 , 2.106165373846852e-46 ], [ 1001.125 , 2.106392780832551e-46 ], [ 1001.1875 , 2.106620212371833e-46 ], 
                    [ 1001.25 , 2.1068476684673484e-46 ], [ 1001.3125 , 2.1070751491217938e-46 ], [ 1001.375 , 2.1073026543377913e-46 ], [ 1001.4375 , 2.1075301841179624e-46 ], 
                    [ 1001.5 , 2.10775773846502e-46 ], [ 1001.5625 , 2.107985317381585e-46 ], [ 1001.625 , 2.1082129208702824e-46 ], [ 1001.6875 , 2.108440548933824e-46 ], 
                    [ 1001.75 , 2.1086682015748186e-46 ], [ 1001.8125 , 2.1088958787959346e-46 ], [ 1001.875 , 2.109123580599826e-46 ], [ 1001.9375 , 2.1093513069891622e-46 ], 
                    [ 1002.0 , 2.1095790579665673e-46 ], [ 1002.0625 , 2.1098068335346972e-46 ], [ 1002.125 , 2.110034633696251e-46 ], [ 1002.1875 , 2.1102624584538545e-46 ], 
                    [ 1002.25 , 2.1104903078101185e-46 ], [ 1002.3125 , 2.110718181767789e-46 ], [ 1002.375 , 2.1109460803294763e-46 ], [ 1002.4375 , 2.1111740034978234e-46 ], 
                    [ 1002.5 , 2.1114019512755006e-46 ], [ 1002.5625 , 2.1116299236651814e-46 ], [ 1002.625 , 2.1118579206694927e-46 ], [ 1002.6875 , 2.112085942291107e-46 ], 
                    [ 1002.75 , 2.1123139885326826e-46 ], [ 1002.8125 , 2.1125420593969075e-46 ], [ 1002.875 , 2.1127701548863654e-46 ], [ 1002.9375 , 2.11299827500379e-46 ], 
                    [ 1003.0 , 2.1132264197517954e-46 ], [ 1003.0625 , 2.1134545891330566e-46 ], [ 1003.125 , 2.1136827831502175e-46 ], [ 1003.1875 , 2.1139110018059835e-46 ], 
                    [ 1003.25 , 2.1141392451029546e-46 ], [ 1003.3125 , 2.1143673615485242e-46 ], [ 1003.375 , 2.1145952870855866e-46 ], [ 1003.4375 , 2.1148232371926403e-46 ], 
                    [ 1003.5 , 2.1150512118723783e-46 ], [ 1003.5625 , 2.1152792111274197e-46 ], [ 1003.625 , 2.1155072349604587e-46 ], [ 1003.6875 , 2.11573528337407e-46 ], 
                    [ 1003.75 , 2.1159633563709783e-46 ], [ 1003.8125 , 2.1161914539537885e-46 ], [ 1003.875 , 2.116419576125181e-46 ], [ 1003.9375 , 2.116647722887791e-46 ], 
                    [ 1004.0 , 2.1168758942442702e-46 ], [ 1004.0625 , 2.117104090197239e-46 ], [ 1004.125 , 2.1173323107494244e-46 ], [ 1004.1875 , 2.1175605559033877e-46 ], 
                    [ 1004.25 , 2.1177888256618414e-46 ], [ 1004.3125 , 2.118017120027438e-46 ], [ 1004.375 , 2.118245439002814e-46 ], [ 1004.4375 , 2.1184737825906235e-46 ], 
                    [ 1004.5 , 2.118702150793519e-46 ], [ 1004.5625 , 2.118930543614139e-46 ], [ 1004.625 , 2.1191589610551378e-46 ], [ 1004.6875 , 2.119387403119229e-46 ], 
                    [ 1004.75 , 2.1196158698089767e-46 ], [ 1004.8125 , 2.119844361127111e-46 ], [ 1004.875 , 2.120072877076272e-46 ], [ 1004.9375 , 2.120301417659084e-46 ], 
                    [ 1005.0 , 2.1205299828782032e-46 ], [ 1005.0625 , 2.1207585727363e-46 ], [ 1005.125 , 2.1209871872360305e-46 ], [ 1005.1875 , 2.1212158263800665e-46 ], 
                    [ 1005.25 , 2.121444490171034e-46 ], [ 1005.3125 , 2.1216731786116195e-46 ], [ 1005.375 , 2.1219018917044515e-46 ], [ 1005.4375 , 2.1221306294522166e-46 ], 
                    [ 1005.5 , 2.1223593918575425e-46 ], [ 1005.5625 , 2.122588168872271e-46 ], [ 1005.625 , 2.1228166110434167e-46 ], [ 1005.6875 , 2.1230450778004875e-46 ], 
                    [ 1005.75 , 2.1232735691461596e-46 ], [ 1005.8125 , 2.1235020850830792e-46 ], [ 1005.875 , 2.123730625613848e-46 ], [ 1005.9375 , 2.1239591907411574e-46 ], 
                    [ 1006.0 , 2.124187780467655e-46 ], [ 1006.0625 , 2.1244163947959433e-46 ], [ 1006.125 , 2.1246450337287e-46 ], [ 1006.1875 , 2.1248736972685883e-46 ], 
                    [ 1006.25 , 2.1251023854182565e-46 ], [ 1006.3125 , 2.125331098180323e-46 ], [ 1006.375 , 2.1255598355574514e-46 ], [ 1006.4375 , 2.1257885975522912e-46 ], 
                    [ 1006.5 , 2.1260173841674623e-46 ], [ 1006.5625 , 2.126246195405674e-46 ], [ 1006.625 , 2.126475031269546e-46 ], [ 1006.6875 , 2.1267038917617446e-46 ], 
                    [ 1006.75 , 2.1269327768848742e-46 ], [ 1006.8125 , 2.127161686641647e-46 ], [ 1006.875 , 2.1273906210346377e-46 ], [ 1006.9375 , 2.127619580066589e-46 ], 
                    [ 1007.0 , 2.127848563740092e-46 ], [ 1007.0625 , 2.128077572057814e-46 ], [ 1007.125 , 2.1283066050224075e-46 ], [ 1007.1875 , 2.1285356626364946e-46 ], 
                    [ 1007.25 , 2.1287647449028187e-46 ], [ 1007.3125 , 2.1289938518239277e-46 ], [ 1007.375 , 2.129222983402535e-46 ], [ 1007.4375 , 2.129452139641279e-46 ], 
                    [ 1007.5 , 2.1296813205427843e-46 ], [ 1007.5625 , 2.1299105261097808e-46 ], [ 1007.625 , 2.1301397563448615e-46 ], [ 1007.6875 , 2.1303690112506977e-46 ], 
                    [ 1007.75 , 2.1305982908299587e-46 ], [ 1007.8125 , 2.1308275950852555e-46 ], [ 1007.875 , 2.1310567024706898e-46 ], [ 1007.9375 , 2.1312856603972854e-46 ], 
                    [ 1008.0 , 2.1315146429228044e-46 ], [ 1008.0625 , 2.1317436500499193e-46 ], [ 1008.125 , 2.131972681781229e-46 ], [ 1008.1875 , 2.1322017381194356e-46 ], 
                    [ 1008.25 , 2.132430819067139e-46 ], [ 1008.3125 , 2.1326599246270128e-46 ], [ 1008.375 , 2.132889054801656e-46 ], [ 1008.4375 , 2.133118209593759e-46 ], 
                    [ 1008.5 , 2.133347389005966e-46 ], [ 1008.5625 , 2.1335765930408928e-46 ], [ 1008.625 , 2.1338058217011987e-46 ], [ 1008.6875 , 2.134035074989515e-46 ], 
                    [ 1008.75 , 2.134264352908533e-46 ], [ 1008.8125 , 2.1344936554608542e-46 ], [ 1008.875 , 2.1347229826491544e-46 ], [ 1008.9375 , 2.1349523344760356e-46 ], 
                    [ 1009.0 , 2.1351817109441903e-46 ], [ 1009.0625 , 2.135411112056266e-46 ], [ 1009.125 , 2.1356405378148803e-46 ], [ 1009.1875 , 2.13586998822268e-46 ], 
                    [ 1009.25 , 2.136099463282376e-46 ], [ 1009.3125 , 2.136328962996509e-46 ], [ 1009.375 , 2.136558487367865e-46 ], [ 1009.4375 , 2.1367880363989415e-46 ], 
                    [ 1009.5 , 2.1370176100925096e-46 ], [ 1009.5625 , 2.1372472084511574e-46 ], [ 1009.625 , 2.137476831477566e-46 ], [ 1009.6875 , 2.13770647917434e-46 ], 
                    [ 1009.75 , 2.1379361515441753e-46 ], [ 1009.8125 , 2.1381658485897074e-46 ], [ 1009.875 , 2.1383955703135882e-46 ], [ 1009.9375 , 2.1386253167184984e-46 ], 
                    [ 1010.0 , 2.1388550878070298e-46 ], [ 1010.0625 , 2.1390848835818493e-46 ], [ 1010.125 , 2.139314701190681e-46 ], [ 1010.1875 , 2.139544189437255e-46 ], 
                    [ 1010.25 , 2.1397737023014883e-46 ], [ 1010.3125 , 2.140003239785976e-46 ], [ 1010.375 , 2.1402328018933437e-46 ], [ 1010.4375 , 2.140462388626279e-46 ], 
                    [ 1010.5 , 2.1406919999874228e-46 ], [ 1010.5625 , 2.1409216359794028e-46 ], [ 1010.625 , 2.141151296604845e-46 ], [ 1010.6875 , 2.1413809818664223e-46 ], 
                    [ 1010.75 , 2.141610691766748e-46 ], [ 1010.8125 , 2.1418404263084946e-46 ], [ 1010.875 , 2.1420701854942908e-46 ], [ 1010.9375 , 2.1422999693267496e-46 ], 
                    [ 1011.0 , 2.1425297778085906e-46 ], [ 1011.0625 , 2.1427596109423828e-46 ], [ 1011.125 , 2.142989468730815e-46 ], [ 1011.1875 , 2.1432193511765182e-46 ], 
                    [ 1011.25 , 2.143449258282121e-46 ], [ 1011.3125 , 2.1436791900503153e-46 ], [ 1011.375 , 2.14390914648367e-46 ], [ 1011.4375 , 2.1441391275848922e-46 ], 
                    [ 1011.5 , 2.1443691333566427e-46 ], [ 1011.5625 , 2.144599163801508e-46 ], [ 1011.625 , 2.144829218922165e-46 ], [ 1011.6875 , 2.1450592987212898e-46 ], 
                    [ 1011.75 , 2.145289403201471e-46 ], [ 1011.8125 , 2.1455195323654005e-46 ], [ 1011.875 , 2.1457496862157263e-46 ], [ 1011.9375 , 2.145979864755021e-46 ], 
                    [ 1012.0 , 2.146210067986023e-46 ], [ 1012.0625 , 2.1464402959113675e-46 ], [ 1012.125 , 2.1466705485336878e-46 ], [ 1012.1875 , 2.1469008258556326e-46 ], 
                    [ 1012.25 , 2.147131127879852e-46 ], [ 1012.3125 , 2.1473614546089805e-46 ], [ 1012.375 , 2.147591806045683e-46 ], [ 1012.4375 , 2.147822033637096e-46 ], 
                    [ 1012.5 , 2.14805203934416e-46 ], [ 1012.5625 , 2.1482820696820465e-46 ], [ 1012.625 , 2.1485121246533794e-46 ], [ 1012.6875 , 2.1487422042608265e-46 ], 
                    [ 1012.75 , 2.1489723085069803e-46 ], [ 1012.8125 , 2.1492024373945403e-46 ], [ 1012.875 , 2.149432590926115e-46 ], [ 1012.9375 , 2.1496627691043433e-46 ], 
                    [ 1013.0 , 2.1498929719318494e-46 ], [ 1013.0625 , 2.150123199411288e-46 ], [ 1013.125 , 2.1503534515452992e-46 ], [ 1013.1875 , 2.1505837283365538e-46 ], 
                    [ 1013.25 , 2.1508140297876464e-46 ], [ 1013.3125 , 2.1510443559012023e-46 ], [ 1013.375 , 2.151274706679924e-46 ], [ 1013.4375 , 2.1515050821264223e-46 ], 
                    [ 1013.5 , 2.1517354822433077e-46 ], [ 1013.5625 , 2.151965907033284e-46 ], [ 1013.625 , 2.152196356498962e-46 ], [ 1013.6875 , 2.152426830642985e-46 ], 
                    [ 1013.75 , 2.1526573294679797e-46 ], [ 1013.8125 , 2.152887852976636e-46 ], [ 1013.875 , 2.1531184011715503e-46 ], [ 1013.9375 , 2.1533489740553513e-46 ], 
                    [ 1014.0 , 2.153579571630744e-46 ], [ 1014.0625 , 2.1538101939003266e-46 ], [ 1014.125 , 2.154040840866775e-46 ], [ 1014.1875 , 2.1542715125327023e-46 ], 
                    [ 1014.25 , 2.1545022089007686e-46 ], [ 1014.3125 , 2.1547329299736358e-46 ], [ 1014.375 , 2.154963675753918e-46 ], [ 1014.4375 , 2.1551944462442764e-46 ], 
                    [ 1014.5 , 2.1554252414473572e-46 ], [ 1014.5625 , 2.1556560613658068e-46 ], [ 1014.625 , 2.155886906002272e-46 ], [ 1014.6875 , 2.156117775359431e-46 ], 
                    [ 1014.75 , 2.1563484206170446e-46 ], [ 1014.8125 , 2.156578961894606e-46 ], [ 1014.875 , 2.1568095278199847e-46 ], [ 1014.9375 , 2.157040118395815e-46 ], 
                    [ 1015.0 , 2.1572707336247478e-46 ], [ 1015.0625 , 2.1575013735093888e-46 ], [ 1015.125 , 2.1577320380523883e-46 ], [ 1015.1875 , 2.1579627272563837e-46 ], 
                    [ 1015.25 , 2.158193441124026e-46 ], [ 1015.3125 , 2.1584241796579373e-46 ], [ 1015.375 , 2.1586549428607393e-46 ], [ 1015.4375 , 2.1588857307350843e-46 ], 
                    [ 1015.5 , 2.1591165432835955e-46 ], [ 1015.5625 , 2.1593473805089717e-46 ], [ 1015.625 , 2.1595782424137588e-46 ], [ 1015.6875 , 2.1598091290006728e-46 ], 
                    [ 1015.75 , 2.160040040272306e-46 ], [ 1015.8125 , 2.1602709762313285e-46 ], [ 1015.875 , 2.1605019368803643e-46 ], [ 1015.9375 , 2.160732922222007e-46 ], 
                    [ 1016.0 , 2.1609639322590037e-46 ], [ 1016.0625 , 2.161194966993903e-46 ], [ 1016.125 , 2.161426026429376e-46 ], [ 1016.1875 , 2.1616571105680786e-46 ], 
                    [ 1016.25 , 2.1618882194126063e-46 ], [ 1016.3125 , 2.1621193529656768e-46 ], [ 1016.375 , 2.1623505112298703e-46 ], [ 1016.4375 , 2.162581694207814e-46 ], 
                    [ 1016.5 , 2.1628129019021956e-46 ], [ 1016.5625 , 2.1630441343156578e-46 ], [ 1016.625 , 2.163275391450797e-46 ], [ 1016.6875 , 2.163506673310318e-46 ], 
                    [ 1016.75 , 2.1637379798968185e-46 ], [ 1016.8125 , 2.163969311212972e-46 ], [ 1016.875 , 2.1642006672613925e-46 ], [ 1016.9375 , 2.164432048044754e-46 ], 
                    [ 1017.0 , 2.1646634535656716e-46 ], [ 1017.0625 , 2.164894555185183e-46 ], [ 1017.125 , 2.1651256276159556e-46 ], [ 1017.1875 , 2.1653567247105323e-46 ], 
                    [ 1017.25 , 2.1655878464714684e-46 ], [ 1017.3125 , 2.1658189929014736e-46 ], [ 1017.375 , 2.1660501640031507e-46 ], [ 1017.4375 , 2.1662813597791327e-46 ], 
                    [ 1017.5 , 2.1665125802320532e-46 ], [ 1017.5625 , 2.166743825364546e-46 ], [ 1017.625 , 2.1669750951792616e-46 ], [ 1017.6875 , 2.167206389678787e-46 ], 
                    [ 1017.75 , 2.167437708865819e-46 ], [ 1017.8125 , 2.167669052742916e-46 ], [ 1017.875 , 2.16790042131279e-46 ], [ 1017.9375 , 2.1681318145780613e-46 ], 
                    [ 1018.0 , 2.16836323254132e-46 ], [ 1018.0625 , 2.168594675205248e-46 ], [ 1018.125 , 2.1688261425724813e-46 ], [ 1018.1875 , 2.1690576346456112e-46 ], 
                    [ 1018.25 , 2.169289151427336e-46 ], [ 1018.3125 , 2.1695206929202315e-46 ], [ 1018.375 , 2.169752259126982e-46 ], [ 1018.4375 , 2.1699838500502086e-46 ], 
                    [ 1018.5 , 2.170215465692535e-46 ], [ 1018.5625 , 2.1704471060566463e-46 ], [ 1018.625 , 2.170678771145134e-46 ], [ 1018.6875 , 2.170910460960637e-46 ], 
                    [ 1018.75 , 2.1711421755058422e-46 ], [ 1018.8125 , 2.1713739147833414e-46 ], [ 1018.875 , 2.1716056787957748e-46 ], [ 1018.9375 , 2.1718374675458293e-46 ], 
                    [ 1019.0 , 2.172069281036083e-46 ], [ 1019.0625 , 2.1723011192692233e-46 ], [ 1019.125 , 2.1725329822478754e-46 ], [ 1019.1875 , 2.172764869974681e-46 ], 
                    [ 1019.25 , 2.172996782452281e-46 ], [ 1019.3125 , 2.173228719683302e-46 ], [ 1019.375 , 2.1734603024752684e-46 ], [ 1019.4375 , 2.1736919006195934e-46 ], 
                    [ 1019.5 , 2.1739235234424154e-46 ], [ 1019.5625 , 2.174155170946349e-46 ], [ 1019.625 , 2.1743868431340087e-46 ], [ 1019.6875 , 2.174618540008055e-46 ], 
                    [ 1019.75 , 2.174850261571104e-46 ], [ 1019.8125 , 2.17508200782577e-46 ], [ 1019.875 , 2.1753137787747158e-46 ], [ 1019.9375 , 2.175545574420557e-46 ], 
                    [ 1020.0 , 2.1757773947659096e-46 ], [ 1020.0625 , 2.1760092398134212e-46 ], [ 1020.125 , 2.1762411095657395e-46 ], [ 1020.1875 , 2.1764730040254812e-46 ], 
                    [ 1020.25 , 2.176704923195279e-46 ], [ 1020.3125 , 2.1769368670777668e-46 ], [ 1020.375 , 2.1771688356755613e-46 ], [ 1020.4375 , 2.1774008289913124e-46 ], 
                    [ 1020.5 , 2.1776328470276534e-46 ], [ 1020.5625 , 2.17786488978725e-46 ], [ 1020.625 , 2.1780969572726743e-46 ], [ 1020.6875 , 2.178329049486577e-46 ], 
                    [ 1020.75 , 2.1785611664316697e-46 ], [ 1020.8125 , 2.178793308110511e-46 ], [ 1020.875 , 2.1790254745256737e-46 ], [ 1020.9375 , 2.1792576656799496e-46 ], 
                    [ 1021.0 , 2.1794898815758813e-46 ], [ 1021.0625 , 2.179722122216105e-46 ], [ 1021.125 , 2.1799543876032732e-46 ], [ 1021.1875 , 2.1801866777400538e-46 ], 
                    [ 1021.25 , 2.180418992629022e-46 ], [ 1021.3125 , 2.1806513322728464e-46 ], [ 1021.375 , 2.1808836966741804e-46 ], [ 1021.4375 , 2.1811160858356152e-46 ], 
                    [ 1021.5 , 2.181348499759883e-46 ], [ 1021.5625 , 2.1815809384494975e-46 ], [ 1021.625 , 2.1818133982215132e-46 ], [ 1021.6875 , 2.182045502484158e-46 ], 
                    [ 1021.75 , 2.1822776314383383e-46 ], [ 1021.8125 , 2.182509785086726e-46 ], [ 1021.875 , 2.1827419634319174e-46 ], [ 1021.9375 , 2.1829741664765867e-46 ], 
                    [ 1022.0 , 2.1832063942232833e-46 ], [ 1022.0625 , 2.183438646674713e-46 ], [ 1022.125 , 2.183670923833489e-46 ], [ 1022.1875 , 2.183903225702192e-46 ], 
                    [ 1022.25 , 2.1841355522834824e-46 ], [ 1022.3125 , 2.1843679035800045e-46 ], [ 1022.375 , 2.1846002795943877e-46 ], [ 1022.4375 , 2.1848326803292145e-46 ], 
                    [ 1022.5 , 2.185065105787146e-46 ], [ 1022.5625 , 2.1852975559708284e-46 ], [ 1022.625 , 2.1855300308828756e-46 ], [ 1022.6875 , 2.185762530525903e-46 ], 
                    [ 1022.75 , 2.1859950549025572e-46 ], [ 1022.8125 , 2.186227604015485e-46 ], [ 1022.875 , 2.186460177867302e-46 ], [ 1022.9375 , 2.186692776460656e-46 ], 
                    [ 1023.0 , 2.186925399798132e-46 ], [ 1023.0625 , 2.1871580478823935e-46 ], [ 1023.125 , 2.187390720716089e-46 ], [ 1023.1875 , 2.1876234183018037e-46 ], 
                    [ 1023.25 , 2.187856140642234e-46 ], [ 1023.3125 , 2.1880888877399977e-46 ], [ 1023.375 , 2.1883216595976967e-46 ], [ 1023.4375 , 2.1885544562179967e-46 ], 
                    [ 1023.5 , 2.18878727760347e-46 ], [ 1023.5625 , 2.1890201237568128e-46 ], [ 1023.625 , 2.1892529946806445e-46 ], [ 1023.6875 , 2.1894858903776466e-46 ], 
                    [ 1023.75 , 2.1897188108503464e-46 ], [ 1023.8125 , 2.1899517561014568e-46 ], [ 1023.875 , 2.1901847261336295e-46 ], [ 1023.9375 , 2.1904177209494382e-46 ], 
                    [ 1024.0 , 0.0 ] ])

    f1 = data[:,0]
    S1 = data[:,1]
    fstop = [f1[0],f1[-1]]

    #% ---- Interpolate to desired frequencies.  Set noise to infinity outside
    #%      band of frequencies covered by f1.
    ##S = inf(size(f)); 
    ##k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop



def ligo2015():


    data = np.array([[ 20.0 , 3.242742153914687e-44 ], [ 20.0625 , 3.202635158124716e-44 ], [ 20.125 , 3.1631610622235574e-44 ], [ 20.1875 , 3.1242870281105673e-44 ], 
                    [ 20.25 , 3.086013534594844e-44 ], [ 20.3125 , 3.048330783472969e-44 ], [ 20.375 , 3.0112165656186626e-44 ], [ 20.4375 , 2.974675313444758e-44 ], 
                    [ 20.5 , 2.9386886627206774e-44 ], [ 20.5625 , 2.9032400004422106e-44 ], [ 20.625 , 2.8683367554165596e-44 ], [ 20.6875 , 2.8339558232851493e-44 ], 
                    [ 20.75 , 2.8000850771346545e-44 ], [ 20.8125 , 2.7667304608082897e-44 ], [ 20.875 , 2.7338711075254157e-44 ], [ 20.9375 , 2.701494474517525e-44 ], 
                    [ 21.0 , 2.669606079586542e-44 ], [ 21.0625 , 2.6381880402071953e-44 ], [ 21.125 , 2.607228080572093e-44 ], [ 21.1875 , 2.576726976183844e-44 ], 
                    [ 21.25 , 2.546675986011371e-44 ], [ 21.3125 , 2.5170595201702876e-44 ], [ 21.375 , 2.4878728848941934e-44 ], [ 21.4375 , 2.459118398571392e-44 ], 
                    [ 21.5 , 2.4307761677109756e-44 ], [ 21.5625 , 2.4028396038262343e-44 ], [ 21.625 , 2.3753107475374133e-44 ], [ 21.6875 , 2.3481782787235393e-44 ], 
                    [ 21.75 , 2.3214302847775795e-44 ], [ 21.8125 , 2.2950614735496875e-44 ], [ 21.875 , 2.269077158747339e-44 ], [ 21.9375 , 2.243458345135279e-44 ], 
                    [ 22.0 , 2.2181992575160544e-44 ], [ 22.0625 , 2.193295614475202e-44 ], [ 22.125 , 2.1687495515432196e-44 ], [ 22.1875 , 2.1445452080821946e-44 ], 
                    [ 22.25 , 2.120676921039971e-44 ], [ 22.3125 , 2.097140048342082e-44 ], [ 22.375 , 2.0739381758563823e-44 ], [ 22.4375 , 2.0510559337520182e-44 ], 
                    [ 22.5 , 2.028488199233011e-44 ], [ 22.5625 , 2.0062298951120582e-44 ], [ 22.625 , 1.9842825913495578e-44 ], [ 22.6875 , 1.9626373417496472e-44 ], 
                    [ 22.75 , 1.9412865008940995e-44 ], [ 22.8125 , 1.9202252607645898e-44 ], [ 22.875 , 1.8994493773856468e-44 ], [ 22.9375 , 1.8789619978942772e-44 ], 
                    [ 23.0 , 1.8587500679647992e-44 ], [ 23.0625 , 1.8388095444063285e-44 ], [ 23.125 , 1.819136255401246e-44 ], [ 23.1875 , 1.799726015920951e-44 ], 
                    [ 23.25 , 1.7805815124114517e-44 ], [ 23.3125 , 1.7616912410829723e-44 ], [ 23.375 , 1.7430515373989482e-44 ], [ 23.4375 , 1.7246582590440276e-44 ], 
                    [ 23.5 , 1.7065077944070545e-44 ], [ 23.5625 , 1.6885990962862504e-44 ], [ 23.625 , 1.6709287563885676e-44 ], [ 23.6875 , 1.6534894367103664e-44 ], 
                    [ 23.75 , 1.636278023115405e-44 ], [ 23.8125 , 1.6192908431955393e-44 ], [ 23.875 , 1.6025243455656568e-44 ], [ 23.9375 , 1.5859771460614613e-44 ], 
                    [ 24.0 , 1.5696470568102646e-44 ], [ 24.0625 , 1.553527694747179e-44 ], [ 24.125 , 1.5376155019452278e-44 ], [ 24.1875 , 1.5219076506534645e-44 ], 
                    [ 24.25 , 1.5064007202978305e-44 ], [ 24.3125 , 1.4910921589461253e-44 ], [ 24.375 , 1.4759802454489972e-44 ], [ 24.4375 , 1.4610628029098785e-44 ], 
                    [ 24.5 , 1.446334404683052e-44 ], [ 24.5625 , 1.4317922752757573e-44 ], [ 24.625 , 1.4174334455558988e-44 ], [ 24.6875 , 1.4032554131397734e-44 ], 
                    [ 24.75 , 1.389255361574823e-44 ], [ 24.8125 , 1.3754308136224289e-44 ], [ 24.875 , 1.3617796709867915e-44 ], [ 24.9375 , 1.3483009912432474e-44 ], 
                    [ 25.0 , 1.3349899244550263e-44 ], [ 25.0625 , 1.3218439255307118e-44 ], [ 25.125 , 1.3088604968017243e-44 ], [ 25.1875 , 1.2960373573328843e-44 ], 
                    [ 25.25 , 1.2833721131881465e-44 ], [ 25.3125 , 1.2708623624433377e-44 ], [ 25.375 , 1.2585061288701433e-44 ], [ 25.4375 , 1.2463008037568255e-44 ], 
                    [ 25.5 , 1.2342444512040646e-44 ], [ 25.5625 , 1.2223367212438843e-44 ], [ 25.625 , 1.2105737708272852e-44 ], [ 25.6875 , 1.198953176428495e-44 ], 
                    [ 25.75 , 1.1874729199887552e-44 ], [ 25.8125 , 1.1761311097003784e-44 ], [ 25.875 , 1.1649256370788065e-44 ], [ 25.9375 , 1.1538545264482643e-44 ], 
                    [ 26.0 , 1.142915930809792e-44 ], [ 26.0625 , 1.1321080158566376e-44 ], [ 26.125 , 1.1214287790910309e-44 ], [ 26.1875 , 1.110876431510451e-44 ], 
                    [ 26.25 , 1.1004492832243928e-44 ], [ 26.3125 , 1.0901454482882294e-44 ], [ 26.375 , 1.079963200969886e-44 ], [ 26.4375 , 1.0699010248357336e-44 ], 
                    [ 26.5 , 1.0599569100721493e-44 ], [ 26.5625 , 1.0501297298573745e-44 ], [ 26.625 , 1.040418024720935e-44 ], [ 26.6875 , 1.0308195908198604e-44 ], 
                    [ 26.75 , 1.0213327489687918e-44 ], [ 26.8125 , 1.0119560488005402e-44 ], [ 26.875 , 1.0026880165119877e-44 ], [ 26.9375 , 9.93526992062606e-45 ], 
                    [ 27.0 , 9.844716904834301e-45 ], [ 27.0625 , 9.75520459893164e-45 ], [ 27.125 , 9.666720366130542e-45 ], [ 27.1875 , 9.579249152484205e-45 ], 
                    [ 27.25 , 9.492777351943571e-45 ], [ 27.3125 , 9.407290604683763e-45 ], [ 27.375 , 9.322776247007162e-45 ], [ 27.4375 , 9.239220905607202e-45 ], 
                    [ 27.5 , 9.156611930164019e-45 ], [ 27.5625 , 9.074935748525574e-45 ], [ 27.625 , 8.99417996302894e-45 ], [ 27.6875 , 8.914332860751794e-45 ], 
                    [ 27.75 , 8.835380811934833e-45 ], [ 27.8125 , 8.75731264551183e-45 ], [ 27.875 , 8.680117271647807e-45 ], [ 27.9375 , 8.603781658102829e-45 ], 
                    [ 28.0 , 8.52829408264785e-45 ], [ 28.0625 , 8.453644625410863e-45 ], [ 28.125 , 8.379821450414667e-45 ], [ 28.1875 , 8.306813492999865e-45 ], 
                    [ 28.25 , 8.234610513616063e-45 ], [ 28.3125 , 8.163200492937981e-45 ], [ 28.375 , 8.09257393221592e-45 ], [ 28.4375 , 8.022720095172682e-45 ], 
                    [ 28.5 , 7.953630071768415e-45 ], [ 28.5625 , 7.885291969519468e-45 ], [ 28.625 , 7.817697759214116e-45 ], [ 28.6875 , 7.750835663191864e-45 ], 
                    [ 28.75 , 7.684702438987056e-45 ], [ 28.8125 , 7.619285626468778e-45 ], [ 28.875 , 7.554573199373776e-45 ], [ 28.9375 , 7.490557653000318e-45 ], 
                    [ 29.0 , 7.427228673983252e-45 ], [ 29.0625 , 7.36457817391099e-45 ], [ 29.125 , 7.302596415635965e-45 ], [ 29.1875 , 7.241275993996061e-45 ], 
                    [ 29.25 , 7.180607434229641e-45 ], [ 29.3125 , 7.12058264219739e-45 ], [ 29.375 , 7.061193997131719e-45 ], [ 29.4375 , 7.002432774543888e-45 ], 
                    [ 29.5 , 6.944290657045959e-45 ], [ 29.5625 , 6.886760342244931e-45 ], [ 29.625 , 6.829834262598151e-45 ], [ 29.6875 , 6.773504308921989e-45 ], 
                    [ 29.75 , 6.717762540487532e-45 ], [ 29.8125 , 6.662602163680505e-45 ], [ 29.875 , 6.608026230358832e-45 ], [ 29.9375 , 6.554016992907089e-45 ], 
                    [ 30.0 , 6.500567821033709e-45 ], [ 30.0625 , 6.447671619830584e-45 ], [ 30.125 , 6.395320493468775e-45 ], [ 30.1875 , 6.34350859576899e-45 ], 
                    [ 30.25 , 6.292229069548205e-45 ], [ 30.3125 , 6.241474600566913e-45 ], [ 30.375 , 6.191239099837586e-45 ], [ 30.4375 , 6.141516590988207e-45 ], 
                    [ 30.5 , 6.092300035355858e-45 ], [ 30.5625 , 6.043586672689346e-45 ], [ 30.625 , 5.995376838386539e-45 ], [ 30.6875 , 5.947655374666682e-45 ], 
                    [ 30.75 , 5.900415351859844e-45 ], [ 30.8125 , 5.853650951558505e-45 ], [ 30.875 , 5.80735658043371e-45 ], [ 30.9375 , 5.761526794846017e-45 ], 
                    [ 31.0 , 5.716156053558398e-45 ], [ 31.0625 , 5.671238401932422e-45 ], [ 31.125 , 5.626767788925796e-45 ], [ 31.1875 , 5.5827494012998246e-45 ], 
                    [ 31.25 , 5.5391724187371447e-45 ], [ 31.3125 , 5.496026570783337e-45 ], [ 31.375 , 5.453307080279222e-45 ], [ 31.4375 , 5.411008888392928e-45 ], 
                    [ 31.5 , 5.369126844567301e-45 ], [ 31.5625 , 5.3276559295165576e-45 ], [ 31.625 , 5.2865912562176755e-45 ], [ 31.6875 , 5.245935406863302e-45 ], 
                    [ 31.75 , 5.205683486276196e-45 ], [ 31.8125 , 5.1658237806645487e-45 ], [ 31.875 , 5.126351240204792e-45 ], [ 31.9375 , 5.087261517760689e-45 ], 
                    [ 32.0 , 5.0485500129654206e-45 ], [ 32.0625 , 5.0102117859382193e-45 ], [ 32.125 , 4.972243049743793e-45 ], [ 32.1875 , 4.934654346392758e-45 ], 
                    [ 32.25 , 4.897426267937339e-45 ], [ 32.3125 , 4.8605547796306816e-45 ], [ 32.375 , 4.824035322096994e-45 ], [ 32.4375 , 4.787863918377919e-45 ], 
                    [ 32.5 , 4.7520369533085805e-45 ], [ 32.5625 , 4.7165521998758166e-45 ], [ 32.625 , 4.6814166992680874e-45 ], [ 32.6875 , 4.646613248599058e-45 ], 
                    [ 32.75 , 4.612138142488826e-45 ], [ 32.8125 , 4.577987531960921e-45 ], [ 32.875 , 4.544157628328284e-45 ], [ 32.9375 , 4.510644860315545e-45 ], 
                    [ 33.0 , 4.4774582376648e-45 ], [ 33.0625 , 4.444584489815198e-45 ], [ 33.125 , 4.412016973205019e-45 ], [ 33.1875 , 4.379752155883414e-45 ], 
                    [ 33.25 , 4.3477865027157304e-45 ], [ 33.3125 , 4.316116500421193e-45 ], [ 33.375 , 4.2847528058944634e-45 ], [ 33.4375 , 4.253680606975854e-45 ], 
                    [ 33.5 , 4.222894387620875e-45 ], [ 33.5625 , 4.1923908424426695e-45 ], [ 33.625 , 4.162166542385551e-45 ], [ 33.6875 , 4.132220570113956e-45 ], 
                    [ 33.75 , 4.1025619516171766e-45 ], [ 33.8125 , 4.0731732883826333e-45 ], [ 33.875 , 4.0440514153288395e-45 ], [ 33.9375 , 4.015193432717585e-45 ], 
                    [ 34.0 , 3.9865966283312117e-45 ], [ 34.0625 , 3.9582699275628436e-45 ], [ 34.125 , 3.9302028558650396e-45 ], [ 34.1875 , 3.9023879322739026e-45 ], 
                    [ 34.25 , 3.874822583064912e-45 ], [ 34.3125 , 3.847504173459489e-45 ], [ 34.375 , 3.8204379718197976e-45 ], [ 34.4375 , 3.79362126803232e-45 ], 
                    [ 34.5 , 3.767043075012681e-45 ], [ 34.5625 , 3.7407006741053396e-45 ], [ 34.625 , 3.7145914972519923e-45 ], [ 34.6875 , 3.688721886084756e-45 ], 
                    [ 34.75 , 3.6630886119625105e-45 ], [ 34.8125 , 3.6376809335091714e-45 ], [ 34.875 , 3.612496328019357e-45 ], [ 34.9375 , 3.5875323498778115e-45 ], 
                    [ 35.0 , 3.562798182389111e-45 ], [ 35.0625 , 3.5382852160300927e-45 ], [ 35.125 , 3.5139859393064726e-45 ], [ 35.1875 , 3.489897975061694e-45 ], 
                    [ 35.25 , 3.4660188305389065e-45 ], [ 35.3125 , 3.4423631190775675e-45 ], [ 35.375 , 3.4189117901293867e-45 ], [ 35.4375 , 3.395662709038054e-45 ], 
                    [ 35.5 , 3.372613673109918e-45 ], [ 35.5625 , 3.349769674242512e-45 ], [ 35.625 , 3.3271307474328267e-45 ], [ 35.6875 , 3.30468542594018e-45 ], 
                    [ 35.75 , 3.2824316339219057e-45 ], [ 35.8125 , 3.2603678681765857e-45 ], [ 35.875 , 3.238507513532894e-45 ], [ 35.9375 , 3.2168326619257535e-45 ], 
                    [ 36.0 , 3.19534122828751e-45 ], [ 36.0625 , 3.1740311753434775e-45 ], [ 36.125 , 3.152912495434912e-45 ], [ 36.1875 , 3.1319761902803744e-45 ], 
                    [ 36.25 , 3.1112156798697607e-45 ], [ 36.3125 , 3.090629052908181e-45 ], [ 36.375 , 3.070222775789116e-45 ], [ 36.4375 , 3.0499946645958107e-45 ], 
                    [ 36.5 , 3.029934938853208e-45 ], [ 36.5625 , 3.010041823358648e-45 ], [ 36.625 , 2.9903200467062874e-45 ], [ 36.6875 , 2.9707714236518674e-45 ], 
                    [ 36.75 , 2.95138407011387e-45 ], [ 36.8125 , 2.9321563014148466e-45 ], [ 36.875 , 2.9130926596554965e-45 ], [ 36.9375 , 2.894195846140666e-45 ], 
                    [ 37.0 , 2.875453607173731e-45 ], [ 37.0625 , 2.85686438675064e-45 ], [ 37.125 , 2.838433829670954e-45 ], [ 37.1875 , 2.8201624080047484e-45 ], 
                    [ 37.25 , 2.8020392917482754e-45 ], [ 37.3125 , 2.7840629633584686e-45 ], [ 37.375 , 2.7662408465048526e-45 ], [ 37.4375 , 2.7485698016316836e-45 ], 
                    [ 37.5 , 2.731041010149905e-45 ], [ 37.5625 , 2.7136530087803365e-45 ], [ 37.625 , 2.696416623160474e-45 ], [ 37.6875 , 2.679321926019727e-45 ], 
                    [ 37.75 , 2.662363737669676e-45 ], [ 37.8125 , 2.6455411222468096e-45 ], [ 37.875 , 2.6288678135319832e-45 ], [ 37.9375 , 2.6123268797428675e-45 ], 
                    [ 38.0 , 2.5959168904273653e-45 ], [ 38.0625 , 2.5796418739300137e-45 ], [ 38.125 , 2.5635058145681947e-45 ], [ 38.1875 , 2.5474966050007956e-45 ], 
                    [ 38.25 , 2.5316129773600228e-45 ], [ 38.3125 , 2.5158655285765484e-45 ], [ 38.375 , 2.500245423663987e-45 ], [ 38.4375 , 2.484747091659578e-45 ], 
                    [ 38.5 , 2.469371960345766e-45 ], [ 38.5625 , 2.454129797518875e-45 ], [ 38.625 , 2.4390056398114632e-45 ], [ 38.6875 , 2.4239983365432694e-45 ], 
                    [ 38.75 , 2.4091172358253618e-45 ], [ 38.8125 , 2.3943563203108025e-45 ], [ 38.875 , 2.379708851484505e-45 ], [ 38.9375 , 2.3651766554279953e-45 ], 
                    [ 39.0 , 2.350768205235042e-45 ], [ 39.0625 , 2.33646989056299e-45 ], [ 39.125 , 2.3222806436265734e-45 ], [ 39.1875 , 2.308211376950249e-45 ], 
                    [ 39.25 , 2.2942523361467908e-45 ], [ 39.3125 , 2.2803990125475033e-45 ], [ 39.375 , 2.2666567407544805e-45 ], [ 39.4375 , 2.253027231140558e-45 ], 
                    [ 39.5 , 2.2395002971031425e-45 ], [ 39.5625 , 2.2260761340989976e-45 ], [ 39.625 , 2.21276666722192e-45 ], [ 39.6875 , 2.1995567741540508e-45 ], 
                    [ 39.75 , 2.1864455073448615e-45 ], [ 39.8125 , 2.1734439818132544e-45 ], [ 39.875 , 2.1605422871237795e-45 ], [ 39.9375 , 2.147736184801044e-45 ], 
                    [ 40.0 , 2.1350328753513768e-45 ], [ 40.0625 , 2.1224305362585044e-45 ], [ 40.125 , 2.1099210437061782e-45 ], [ 40.1875 , 2.097508299657094e-45 ], 
                    [ 40.25 , 2.0851968546358622e-45 ], [ 40.3125 , 2.0729755432243955e-45 ], [ 40.375 , 2.060845615992145e-45 ], [ 40.4375 , 2.0488171129738505e-45 ], 
                    [ 40.5 , 2.036876030444541e-45 ], [ 40.5625 , 2.02502153978493e-45 ], [ 40.625 , 2.0132677398059068e-45 ], [ 40.6875 , 2.001598980842543e-45 ], 
                    [ 40.75 , 1.990014335760532e-45 ], [ 40.8125 , 1.978526190832247e-45 ], [ 40.875 , 1.9671222565322683e-45 ], [ 40.9375 , 1.9557999537953732e-45 ], 
                    [ 41.0 , 1.9445704040285135e-45 ], [ 41.0625 , 1.933423997736663e-45 ], [ 41.125 , 1.9223568985395726e-45 ], [ 41.1875 , 1.9113791387178418e-45 ], 
                    [ 41.25 , 1.900483051061534e-45 ], [ 41.3125 , 1.8896640422391586e-45 ], [ 41.375 , 1.878931810926126e-45 ], [ 41.4375 , 1.868279244493006e-45 ], 
                    [ 41.5 , 1.8577014850414462e-45 ], [ 41.5625 , 1.8472083054993575e-45 ], [ 41.625 , 1.8367925962391314e-45 ], [ 41.6875 , 1.826449573309455e-45 ], 
                    [ 41.75 , 1.8161893965981094e-45 ], [ 41.8125 , 1.806004130414982e-45 ], [ 41.875 , 1.7958894470005065e-45 ], [ 41.9375 , 1.7858561451859958e-45 ], 
                    [ 42.0 , 1.7758950100044734e-45 ], [ 42.0625 , 1.766002515442282e-45 ], [ 42.125 , 1.7561904979334192e-45 ], [ 42.1875 , 1.7464474715875195e-45 ], 
                    [ 42.25 , 1.7367711009420692e-45 ], [ 42.3125 , 1.7271746567239786e-45 ], [ 42.375 , 1.7176438096992615e-45 ], [ 42.4375 , 1.7081792526526764e-45 ], 
                    [ 42.5 , 1.698791438881779e-45 ], [ 42.5625 , 1.689467164318049e-45 ], [ 42.625 , 1.6802094721837618e-45 ], [ 42.6875 , 1.6710245128704006e-45 ], 
                    [ 42.75 , 1.66190127605713e-45 ], [ 42.8125 , 1.6528451421651868e-45 ], [ 42.875 , 1.6438575167533518e-45 ], [ 42.9375 , 1.6349299053879784e-45 ], 
                    [ 43.0 , 1.626070347680624e-45 ], [ 43.0625 , 1.617274896395427e-45 ], [ 43.125 , 1.6085377726568177e-45 ], [ 43.1875 , 1.5998697846545288e-45 ], 
                    [ 43.25 , 1.5912613826271312e-45 ], [ 43.3125 , 1.582710660778557e-45 ], [ 43.375 , 1.5742287623573e-45 ], [ 43.4375 , 1.5658024954748345e-45 ], 
                    [ 43.5 , 1.5574356827129082e-45 ], [ 43.5625 , 1.5491326756633112e-45 ], [ 43.625 , 1.540883745502711e-45 ], [ 43.6875 , 1.5326963374552103e-45 ], 
                    [ 43.75 , 1.5245675765303997e-45 ], [ 43.8125 , 1.516491416241659e-45 ], [ 43.875 , 1.508479020787622e-45 ], [ 43.9375 , 1.5005198746036172e-45 ], 
                    [ 44.0 , 1.4926143080402613e-45 ], [ 44.0625 , 1.4847701932571225e-45 ], [ 44.125 , 1.4769762639044925e-45 ], [ 44.1875 , 1.469238813829786e-45 ], 
                    [ 44.25 , 1.4615571098492512e-45 ], [ 44.3125 , 1.4539241890398895e-45 ], [ 44.375 , 1.446350684821866e-45 ], [ 44.4375 , 1.438827085616957e-45 ], 
                    [ 44.5 , 1.4313536276119967e-45 ], [ 44.5625 , 1.4239375948849214e-45 ], [ 44.625 , 1.4165681126566465e-45 ], [ 44.6875 , 1.4092522208577385e-45 ], 
                    [ 44.75 , 1.4019874907615638e-45 ], [ 44.8125 , 1.3947680421410464e-45 ], [ 44.875 , 1.3876059614392715e-45 ], [ 44.9375 , 1.3804887655562081e-45 ], 
                    [ 45.0 , 1.373420333386934e-45 ], [ 45.0625 , 1.3664035772411965e-45 ], [ 45.125 , 1.3594300596339905e-45 ], [ 45.1875 , 1.3525095027054277e-45 ], 
                    [ 45.25 , 1.3456340514314576e-45 ], [ 45.3125 , 1.3388034023102294e-45 ], [ 45.375 , 1.332024335947189e-45 ], [ 45.4375 , 1.3252865563625172e-45 ], 
                    [ 45.5 , 1.3185980782942067e-45 ], [ 45.5625 , 1.3119542888683172e-45 ], [ 45.625 , 1.3053522338205585e-45 ], [ 45.6875 , 1.2988008907402325e-45 ], 
                    [ 45.75 , 1.2922889724496251e-45 ], [ 45.8125 , 1.2858236153227963e-45 ], [ 45.875 , 1.2794019212604332e-45 ], [ 45.9375 , 1.2730196841158078e-45 ], 
                    [ 46.0 , 1.2666866957982277e-45 ], [ 46.0625 , 1.2603913556339062e-45 ], [ 46.125 , 1.2541405670090063e-45 ], [ 46.1875 , 1.2479317945364504e-45 ], 
                    [ 46.25 , 1.2417608995094283e-45 ], [ 46.3125 , 1.2356372758366095e-45 ], [ 46.375 , 1.2295496231674282e-45 ], [ 46.4375 , 1.223505243723752e-45 ], 
                    [ 46.5 , 1.2175007007067394e-45 ], [ 46.5625 , 1.21153312870529e-45 ], [ 46.625 , 1.2056102935068883e-45 ], [ 46.6875 , 1.1997218200482383e-45 ], 
                    [ 46.75 , 1.1938759687614081e-45 ], [ 46.8125 , 1.1880673022583498e-45 ], [ 46.875 , 1.1822953359224934e-45 ], [ 46.9375 , 1.1765651442236815e-45 ], 
                    [ 47.0 , 1.1708677859576613e-45 ], [ 47.0625 , 1.1652130171803092e-45 ], [ 47.125 , 1.1595923208566135e-45 ], [ 47.1875 , 1.1540086182296433e-45 ], 
                    [ 47.25 , 1.1484632976556367e-45 ], [ 47.3125 , 1.1429495965061069e-45 ], [ 47.375 , 1.1374784940622338e-45 ], [ 47.4375 , 1.1320381833458474e-45 ], 
                    [ 47.5 , 1.126635671767386e-45 ], [ 47.5625 , 1.1212677697422493e-45 ], [ 47.625 , 1.11593264943396e-45 ], [ 47.6875 , 1.1106360375958384e-45 ], 
                    [ 47.75 , 1.1053688704544431e-45 ], [ 47.8125 , 1.1001408442809718e-45 ], [ 47.875 , 1.0949432695427964e-45 ], [ 47.9375 , 1.0897800936884481e-45 ], 
                    [ 48.0 , 1.0846510145606676e-45 ], [ 48.0625 , 1.0795517455239294e-45 ], [ 48.125 , 1.074490099808082e-45 ], [ 48.1875 , 1.0694560850914009e-45 ], 
                    [ 48.25 , 1.0644584971273878e-45 ], [ 48.3125 , 1.0594903087901577e-45 ], [ 48.375 , 1.0545542859342009e-45 ], [ 48.4375 , 1.0496509391210775e-45 ], 
                    [ 48.5 , 1.044775576768677e-45 ], [ 48.5625 , 1.039936008385365e-45 ], [ 48.625 , 1.0351223741258166e-45 ], [ 48.6875 , 1.0303435643567944e-45 ], 
                    [ 48.75 , 1.0255922244980302e-45 ], [ 48.8125 , 1.0208718573774086e-45 ], [ 48.875 , 1.0161819010023262e-45 ], [ 48.9375 , 1.0115191230112696e-45 ], 
                    [ 49.0 , 1.0068895360584715e-45 ], [ 49.0625 , 1.0022843032218068e-45 ], [ 49.125 , 9.977133259940042e-46 ], [ 49.1875 , 9.931670968771711e-46 ], 
                    [ 49.25 , 9.886515703378835e-46 ], [ 49.3125 , 9.841634177189478e-46 ], [ 49.375 , 9.797025402375405e-46 ], [ 49.4375 , 9.752716014474677e-46 ], 
                    [ 49.5 , 9.70864663068339e-46 ], [ 49.5625 , 9.664899963828089e-46 ], [ 49.625 , 9.621377920768179e-46 ], [ 49.6875 , 9.578169460502993e-46 ], 
                    [ 49.75 , 9.535198024895916e-46 ], [ 49.8125 , 9.492508878615138e-46 ], [ 49.875 , 9.450079105778383e-46 ], [ 49.9375 , 9.40790172871863e-46 ], 
                    [ 50.0 , 9.366005245282094e-46 ], [ 50.0625 , 9.324332591653736e-46 ], [ 50.125 , 9.28296186603477e-46 ], [ 50.1875 , 9.241798333935754e-46 ], 
                    [ 50.25 , 9.20093376525268e-46 ], [ 50.3125 , 9.160283556420054e-46 ], [ 50.375 , 9.119905962541147e-46 ], [ 50.4375 , 9.079761713670972e-46 ], 
                    [ 50.5 , 9.039864086344085e-46 ], [ 50.5625 , 9.000217521978742e-46 ], [ 50.625 , 8.960792518272868e-46 ], [ 50.6875 , 8.921636572656631e-46 ], 
                    [ 50.75 , 8.882678733244385e-46 ], [ 50.8125 , 8.844005770046937e-46 ], [ 50.875 , 8.805520504396964e-46 ], [ 50.9375 , 8.767311089380657e-46 ], 
                    [ 51.0 , 8.729298398648633e-46 ], [ 51.0625 , 8.691538886878277e-46 ], [ 51.125 , 8.653991310082359e-46 ], [ 51.1875 , 8.616675748100342e-46 ], 
                    [ 51.25 , 8.579586404456456e-46 ], [ 51.3125 , 8.542708597577838e-46 ], [ 51.375 , 8.506070487882346e-46 ], [ 51.4375 , 8.469624349729965e-46 ], 
                    [ 51.5 , 8.433431241965473e-46 ], [ 51.5625 , 8.397411549745417e-46 ], [ 51.625 , 8.361655761430958e-46 ], [ 51.6875 , 8.326066972112577e-46 ], 
                    [ 51.75 , 8.290731397656969e-46 ], [ 51.8125 , 8.25557134606758e-46 ], [ 51.875 , 8.22064703061124e-46 ], [ 51.9375 , 8.18590900536639e-46 ], 
                    [ 52.0 , 8.151389515425129e-46 ], [ 52.0625 , 8.117066995202396e-46 ], [ 52.125 , 8.082947707504881e-46 ], [ 52.1875 , 8.049034967010544e-46 ], 
                    [ 52.25 , 8.015310276678646e-46 ], [ 52.3125 , 7.9818011284457e-46 ], [ 52.375 , 7.94846569597678e-46 ], [ 52.4375 , 7.915353996664896e-46 ], 
                    [ 52.5 , 7.882402370601154e-46 ], [ 52.5625 , 7.84968202327418e-46 ], [ 52.625 , 7.817108867083773e-46 ], [ 52.6875 , 7.78477495202059e-46 ], 
                    [ 52.75 , 7.752584305254309e-46 ], [ 52.8125 , 7.720622179100972e-46 ], [ 52.875 , 7.688809004119035e-46 ], [ 52.9375 , 7.657212749893238e-46 ], 
                    [ 53.0 , 7.62577175465735e-46 ], [ 53.0625 , 7.594536215692595e-46 ], [ 53.125 , 7.563461755212662e-46 ], [ 53.1875 , 7.532582225160864e-46 ], 
                    [ 53.25 , 7.501869048831818e-46 ], [ 53.3125 , 7.471340893752177e-46 ], [ 53.375 , 7.440984054921614e-46 ], [ 53.4375 , 7.41080317171005e-46 ], 
                    [ 53.5 , 7.380797233131378e-46 ], [ 53.5625 , 7.350957966238323e-46 ], [ 53.625 , 7.321297536851357e-46 ], [ 53.6875 , 7.291795820555426e-46 ], 
                    [ 53.75 , 7.262476347387254e-46 ], [ 53.8125 , 7.233308129001494e-46 ], [ 53.875 , 7.204324426894491e-46 ], [ 53.9375 , 7.175484760730767e-46 ], 
                    [ 54.0 , 7.146832199175373e-46 ], [ 54.0625 , 7.11831732606853e-46 ], [ 54.125 , 7.089990751198335e-46 ], [ 54.1875 , 7.061795758425436e-46 ], 
                    [ 54.25 , 7.033790905149133e-46 ], [ 54.3125 , 7.005912320392341e-46 ], [ 54.375 , 6.978224357326926e-46 ], [ 54.4375 , 6.950657665589911e-46 ], 
                    [ 54.5 , 6.92328202644341e-46 ], [ 54.5625 , 6.896023224482058e-46 ], [ 54.625 , 6.868955418594423e-46 ], [ 54.6875 , 6.842000491249716e-46 ], 
                    [ 54.75 , 6.815236400758973e-46 ], [ 54.8125 , 6.788581715273892e-46 ], [ 54.875 , 6.762116623842075e-46 ], [ 54.9375 , 6.735757890864815e-46 ], 
                    [ 55.0 , 6.709587676837089e-46 ], [ 55.0625 , 6.683521244915056e-46 ], [ 55.125 , 6.657641974879716e-46 ], [ 55.1875 , 6.631864344285026e-46 ], 
                    [ 55.25 , 6.6062714283861556e-46 ], [ 55.3125 , 6.580778428522317e-46 ], [ 55.375 , 6.555468252004439e-46 ], [ 55.4375 , 6.530256692417568e-46 ], 
                    [ 55.5 , 6.505224691973963e-46 ], [ 55.5625 , 6.480290402480223e-46 ], [ 55.625 , 6.455533421984693e-46 ], [ 55.6875 , 6.43087366771132e-46 ], 
                    [ 55.75 , 6.406387120098541e-46 ], [ 55.8125 , 6.381997655159511e-46 ], [ 55.875 , 6.3577778716274654e-46 ], [ 55.9375 , 6.3336554349228716e-46 ], 
                    [ 56.0 , 6.309699225525817e-46 ], [ 56.0625 , 6.285841035083827e-46 ], [ 56.125 , 6.2621441204131284e-46 ], [ 56.1875 , 6.238546138890872e-46 ], 
                    [ 56.25 , 6.215105123481245e-46 ], [ 56.3125 , 6.1917644151818374e-46 ], [ 56.375 , 6.168575594073237e-46 ], [ 56.4375 , 6.1454887154141874e-46 ], 
                    [ 56.5 , 6.122548473233538e-46 ], [ 56.5625 , 6.0997121503099586e-46 ], [ 56.625 , 6.07701737307401e-46 ], [ 56.6875 , 6.054428970081575e-46 ], 
                    [ 56.75 , 6.031976138186725e-46 ], [ 56.8125 , 6.0096322238252905e-46 ], [ 56.875 , 5.9874177629883685e-46 ], [ 56.9375 , 5.9653150858445605e-46 ], 
                    [ 57.0 , 5.943335861635638e-46 ], [ 57.0625 , 5.921471838409745e-46 ], [ 57.125 , 5.899724682317269e-46 ], [ 57.1875 , 5.8780961929572424e-46 ], 
                    [ 57.25 , 5.8565776971450374e-46 ], [ 57.3125 , 5.835181499738736e-46 ], [ 57.375 , 5.813888506227865e-46 ], [ 57.4375 , 5.792722190927106e-46 ], 
                    [ 57.5 , 5.771651740324398e-46 ], [ 57.5625 , 5.750711981879419e-46 ], [ 57.625 , 5.729860805033007e-46 ], [ 57.6875 , 5.7091454244579275e-46 ], 
                    [ 57.75 , 5.688510917502309e-46 ], [ 57.8125 , 5.668016611385281e-46 ], [ 57.875 , 5.6475961412139515e-46 ], [ 57.9375 , 5.6273197015507915e-46 ], 
                    [ 58.0 , 5.607116836799834e-46 ], [ 58.0625 , 5.587049208960565e-46 ], [ 58.125 , 5.567061114661792e-46 ], [ 58.1875 , 5.547199666431869e-46 ], 
                    [ 58.25 , 5.527423997368305e-46 ], [ 58.3125 , 5.507765726431495e-46 ], [ 58.375 , 5.488199567462487e-46 ], [ 58.4375 , 5.468741763841269e-46 ], 
                    [ 58.5 , 5.449382782734638e-46 ], [ 58.5625 , 5.430122672110017e-46 ], [ 58.625 , 5.410968084500271e-46 ], [ 58.6875 , 5.391902986127141e-46 ], 
                    [ 58.75 , 5.372950669713768e-46 ], [ 58.8125 , 5.354078086268654e-46 ], [ 58.875 , 5.335325337595832e-46 ], [ 58.9375 , 5.316642497624984e-46 ], 
                    [ 59.0 , 5.2980869462186384e-46 ], [ 59.0625 , 5.27959615532379e-46 ], [ 59.125 , 5.2612306229474255e-46 ], [ 59.1875 , 5.242933780890667e-46 ], 
                    [ 59.25 , 5.22475172252265e-46 ], [ 59.3125 , 5.206646737273831e-46 ], [ 59.375 , 5.188645382413514e-46 ], [ 59.4375 , 5.1707294578178795e-46 ], 
                    [ 59.5 , 5.152906445514607e-46 ], [ 59.5625 , 5.1351776670188514e-46 ], [ 59.625 , 5.117530718477533e-46 ], [ 59.6875 , 5.099986817500841e-46 ], 
                    [ 59.75 , 5.082513553915888e-46 ], [ 59.8125 , 5.065152297749119e-46 ], [ 59.875 , 5.047850378461852e-46 ], [ 59.9375 , 5.030669649555565e-46 ], 
                    [ 60.0 , 5.01354739651992e-46 ], [ 60.0625 , 4.9965341468917026e-46 ], [ 60.125 , 4.979589053729323e-46 ], [ 60.1875 , 4.962741502255716e-46 ], 
                    [ 60.25 , 4.945972201181102e-46 ], [ 60.3125 , 4.92928798608865e-46 ], [ 60.375 , 4.912691813986005e-46 ], [ 60.4375 , 4.896168615662608e-46 ], 
                    [ 60.5 , 4.879743704114584e-46 ], [ 60.5625 , 4.8633794831184455e-46 ], [ 60.625 , 4.8471238650583054e-46 ], [ 60.6875 , 4.830922580639878e-46 ], 
                    [ 60.75 , 4.8148279818299525e-46 ], [ 60.8125 , 4.798792963133402e-46 ], [ 60.875 , 4.782852155280819e-46 ], [ 60.9375 , 4.766982009240412e-46 ], 
                    [ 61.0 , 4.751192849991968e-46 ], [ 61.0625 , 4.7354853117104015e-46 ], [ 61.125 , 4.7198456765824925e-46 ], [ 61.1875 , 4.704298753987499e-46 ], 
                    [ 61.25 , 4.688806645399669e-46 ], [ 61.3125 , 4.6734187324947156e-46 ], [ 61.375 , 4.658081320256845e-46 ], [ 61.4375 , 4.64284149305391e-46 ], 
                    [ 61.5 , 4.627660357014193e-46 ], [ 61.5625 , 4.6125631757514525e-46 ], [ 61.625 , 4.597536579187236e-46 ], [ 61.6875 , 4.582580187398587e-46 ], 
                    [ 61.75 , 4.56770633969572e-46 ], [ 61.8125 , 4.5528888832152875e-46 ], [ 61.875 , 4.538165932884547e-46 ], [ 61.9375 , 4.523490593042979e-46 ], 
                    [ 62.0 , 4.508911772898998e-46 ], [ 62.0625 , 4.494384955962682e-46 ], [ 62.125 , 4.479940591191755e-46 ], [ 62.1875 , 4.465560913041461e-46 ], 
                    [ 62.25 , 4.451249069286084e-46 ], [ 62.3125 , 4.43701432438992e-46 ], [ 62.375 , 4.422833167321957e-46 ], [ 62.4375 , 4.408742086639892e-46 ], 
                    [ 62.5 , 4.3946958999312135e-46 ], [ 62.5625 , 4.380740768549653e-46 ], [ 62.625 , 4.3668356241657625e-46 ], [ 62.6875 , 4.353007064505658e-46 ], 
                    [ 62.75 , 4.339241815318808e-46 ], [ 62.8125 , 4.325538094617352e-46 ], [ 62.875 , 4.311910738131725e-46 ], [ 62.9375 , 4.298330106471418e-46 ], 
                    [ 63.0 , 4.2848392503498914e-46 ], [ 63.0625 , 4.27139073699736e-46 ], [ 63.125 , 4.258024401371857e-46 ], [ 63.1875 , 4.24471035034864e-46 ], 
                    [ 63.25 , 4.231463180609939e-46 ], [ 63.3125 , 4.218282067418107e-46 ], [ 63.375 , 4.205152480401479e-46 ], [ 63.4375 , 4.192102569398348e-46 ], 
                    [ 63.5 , 4.17909315637433e-46 ], [ 63.5625 , 4.1661689612230386e-46 ], [ 63.625 , 4.153288942663035e-46 ], [ 63.6875 , 4.140478385436438e-46 ], 
                    [ 63.75 , 4.127726164754088e-46 ], [ 63.8125 , 4.115027824078886e-46 ], [ 63.875 , 4.102402110013229e-46 ], [ 63.9375 , 4.089815134119465e-46 ], 
                    [ 64.0 , 4.077314134799334e-46 ], [ 64.0625 , 4.064851870979526e-46 ], [ 64.125 , 4.052459167782436e-46 ], [ 64.1875 , 4.040119686293805e-46 ], 
                    [ 64.25 , 4.0278341085193365e-46 ], [ 64.3125 , 4.015616644624974e-46 ], [ 64.375 , 4.003437018295224e-46 ], [ 64.4375 , 3.991339731864794e-46 ], 
                    [ 64.5 , 3.979279000109393e-46 ], [ 64.5625 , 3.967286162516815e-46 ], [ 64.625 , 3.955344075415027e-46 ], [ 64.6875 , 3.94345355181475e-46 ], 
                    [ 64.75 , 3.931628685807492e-46 ], [ 64.8125 , 3.919839277921995e-46 ], [ 64.875 , 3.908130331704573e-46 ], [ 64.9375 , 3.896456491166667e-46 ], 
                    [ 65.0 , 3.884846471287101e-46 ], [ 65.0625 , 3.873286821261969e-46 ], [ 65.125 , 3.861774493866902e-46 ], [ 65.1875 , 3.850327821509184e-46 ], 
                    [ 65.25 , 3.838915078193143e-46 ], [ 65.3125 , 3.827576931200564e-46 ], [ 65.375 , 3.8162753509778466e-46 ], [ 65.4375 , 3.8050317315779195e-46 ], 
                    [ 65.5 , 3.793840266707075e-46 ], [ 65.5625 , 3.782689961196961e-46 ], [ 65.625 , 3.771607206876182e-46 ], [ 65.6875 , 3.760556923480722e-46 ], 
                    [ 65.75 , 3.7495739925699925e-46 ], [ 65.8125 , 3.738631114466668e-46 ], [ 65.875 , 3.72773842125303e-46 ], [ 65.9375 , 3.716901466677229e-46 ], 
                    [ 66.0 , 3.7060978357337025e-46 ], [ 66.0625 , 3.69536584833924e-46 ], [ 66.125 , 3.684664938255317e-46 ], [ 66.1875 , 3.6740220969427495e-46 ], 
                    [ 66.25 , 3.663424559365633e-46 ], [ 66.3125 , 3.6528679134068704e-46 ], [ 66.375 , 3.6423724458880676e-46 ], [ 66.4375 , 3.631907134082869e-46 ], 
                    [ 66.5 , 3.621506663433367e-46 ], [ 66.5625 , 3.611142267413743e-46 ], [ 66.625 , 3.600825235807797e-46 ], [ 66.6875 , 3.59056045640392e-46 ], 
                    [ 66.75 , 3.5803257960513e-46 ], [ 66.8125 , 3.570159485257751e-46 ], [ 66.875 , 3.560022041634636e-46 ], [ 66.9375 , 3.549937426593221e-46 ], 
                    [ 67.0 , 3.539897319665549e-46 ], [ 67.0625 , 3.529892486251158e-46 ], [ 67.125 , 3.519948392250544e-46 ], [ 67.1875 , 3.510032311852627e-46 ], 
                    [ 67.25 , 3.5001731037835775e-46 ], [ 67.3125 , 3.4903515857081694e-46 ], [ 67.375 , 3.480569496026488e-46 ], [ 67.4375 , 3.4708417301092362e-46 ], 
                    [ 67.5 , 3.4611411521076462e-46 ], [ 67.5625 , 3.4515008949804686e-46 ], [ 67.625 , 3.4418927576564027e-46 ], [ 67.6875 , 3.432327133312116e-46 ], 
                    [ 67.75 , 3.4228103070712755e-46 ], [ 67.8125 , 3.4133198681700152e-46 ], [ 67.875 , 3.403891909401711e-46 ], [ 67.9375 , 3.394491470657847e-46 ], 
                    [ 68.0 , 3.385135746601734e-46 ], [ 68.0625 , 3.3758244197265396e-46 ], [ 68.125 , 3.3665400507113107e-46 ], [ 68.1875 , 3.3573169037966186e-46 ], 
                    [ 68.25 , 3.3481190250914383e-46 ], [ 68.3125 , 3.338967113687924e-46 ], [ 68.375 , 3.3298561206458675e-46 ], [ 68.4375 , 3.3207732454622047e-46 ], 
                    [ 68.5 , 3.3117481796252517e-46 ], [ 68.5625 , 3.3027476417544006e-46 ], [ 68.625 , 3.2937935529741187e-46 ], [ 68.6875 , 3.284877891872093e-46 ], 
                    [ 68.75 , 3.27599065823277e-46 ], [ 68.8125 , 3.2671587725393453e-46 ], [ 68.875 , 3.258350697110674e-46 ], [ 68.9375 , 3.2495886113613697e-46 ], 
                    [ 69.0 , 3.240863266512717e-46 ], [ 69.0625 , 3.2321658727463257e-46 ], [ 69.125 , 3.2235223749333536e-46 ], [ 69.1875 , 3.21490199166874e-46 ], 
                    [ 69.25 , 3.2063263708251566e-46 ], [ 69.3125 , 3.19778665574933e-46 ], [ 69.375 , 3.1892736281769028e-46 ], [ 69.4375 , 3.1808137486362878e-46 ], 
                    [ 69.5 , 3.1723763098048063e-46 ], [ 69.5625 , 3.1639817202773147e-46 ], [ 69.625 , 3.1556230362528895e-46 ], [ 69.6875 , 3.1472890484672535e-46 ], 
                    [ 69.75 , 3.1390082702524388e-46 ], [ 69.8125 , 3.130749279451083e-46 ], [ 69.875 , 3.122530427575023e-46 ], [ 69.9375 , 3.1143481751388775e-46 ], 
                    [ 70.0 , 3.1061879356471004e-46 ], [ 70.0625 , 3.0980818135842043e-46 ], [ 70.125 , 3.089996845815958e-46 ], [ 70.1875 , 3.081948750684018e-46 ], 
                    [ 70.25 , 3.0739387617491787e-46 ], [ 70.3125 , 3.065949590786297e-46 ], [ 70.375 , 3.0580116605350116e-46 ], [ 70.4375 , 3.0500964239826856e-46 ], 
                    [ 70.5 , 3.042214117334263e-46 ], [ 70.5625 , 3.0343720849801648e-46 ], [ 70.625 , 3.0265502673345226e-46 ], [ 70.6875 , 3.0187751463813594e-46 ], 
                    [ 70.75 , 3.011025491715406e-46 ], [ 70.8125 , 3.003304212405924e-46 ], [ 70.875 , 2.995626021638021e-46 ], [ 70.9375 , 2.987967460787427e-46 ], 
                    [ 71.0 , 2.9803504886917035e-46 ], [ 71.0625 , 2.9727623897225038e-46 ], [ 71.125 , 2.965197539985381e-46 ], [ 71.1875 , 2.957679224322644e-46 ], 
                    [ 71.25 , 2.9501799714945767e-46 ], [ 71.3125 , 2.942716618538638e-46 ], [ 71.375 , 2.9352861249214375e-46 ], [ 71.4375 , 2.9278743936393894e-46 ], 
                    [ 71.5 , 2.9205108694673186e-46 ], [ 71.5625 , 2.9131670431977008e-46 ], [ 71.625 , 2.9058529725367817e-46 ], [ 71.6875 , 2.8985763890757987e-46 ], 
                    [ 71.75 , 2.8913180270001727e-46 ], [ 71.8125 , 2.884101095310719e-46 ], [ 71.875 , 2.8769089557418866e-46 ], [ 71.9375 , 2.869739875835228e-46 ], 
                    [ 72.0 , 2.8626133990154238e-46 ], [ 72.0625 , 2.8555046195041074e-46 ], [ 72.125 , 2.848430122659405e-46 ], [ 72.1875 , 2.8413860424017162e-46 ], 
                    [ 72.25 , 2.8343593819383593e-46 ], [ 72.3125 , 2.8273779534087775e-46 ], [ 72.375 , 2.8204152358978097e-46 ], [ 72.4375 , 2.8134791739952417e-46 ], 
                    [ 72.5 , 2.8065796095482303e-46 ], [ 72.5625 , 2.7996969650735517e-46 ], [ 72.625 , 2.792851433272528e-46 ], [ 72.6875 , 2.7860312223751624e-46 ], 
                    [ 72.75 , 2.7792296476296535e-46 ], [ 72.8125 , 2.7724711803166585e-46 ], [ 72.875 , 2.7657291480904556e-46 ], [ 72.9375 , 2.7590157349093085e-46 ], 
                    [ 73.0 , 2.7523347042678883e-46 ], [ 73.0625 , 2.745669851921474e-46 ], [ 73.125 , 2.7390431810639055e-46 ], [ 73.1875 , 2.7324384553086537e-46 ], 
                    [ 73.25 , 2.7258534641893487e-46 ], [ 73.3125 , 2.7193083160071136e-46 ], [ 73.375 , 2.712778883623725e-46 ], [ 73.4375 , 2.7062783422900608e-46 ], 
                    [ 73.5 , 2.6998076240001633e-46 ], [ 73.5625 , 2.693352377213816e-46 ], [ 73.625 , 2.68693493120074e-46 ], [ 73.6875 , 2.6805377277506393e-46 ], 
                    [ 73.75 , 2.6741598205661355e-46 ], [ 73.8125 , 2.6678200483790256e-46 ], [ 73.875 , 2.6614953062253017e-46 ], [ 73.9375 , 2.6551983071284935e-46 ], 
                    [ 74.0 , 2.648930217481914e-46 ], [ 74.0625 , 2.6426769248271683e-46 ], [ 74.125 , 2.6364595490184597e-46 ], [ 74.1875 , 2.6302622271606945e-46 ], 
                    [ 74.25 , 2.6240823464976558e-46 ], [ 74.3125 , 2.6179404108688634e-46 ], [ 74.375 , 2.6118128510744906e-46 ], [ 74.4375 , 2.6057105175549853e-46 ], 
                    [ 74.5 , 2.5996376536227778e-46 ], [ 74.5625 , 2.5935789430955234e-46 ], [ 74.625 , 2.587552904056114e-46 ], [ 74.6875 , 2.581548139967408e-46 ], 
                    [ 74.75 , 2.5755576480140796e-46 ], [ 74.8125 , 2.5696063366034677e-46 ], [ 74.875 , 2.563668776819665e-46 ], [ 74.9375 , 2.55775263194421e-46 ], 
                    [ 75.0 , 2.5518678624581347e-46 ], [ 75.0625 , 2.54599663240176e-46 ], [ 75.125 , 2.540153631737659e-46 ], [ 75.1875 , 2.5343344499336183e-46 ], 
                    [ 75.25 , 2.528528599164574e-46 ], [ 75.3125 , 2.5227575823729605e-46 ], [ 75.375 , 2.5170031747830873e-46 ], [ 75.4375 , 2.511265179769554e-46 ], 
                    [ 75.5 , 2.505561697721728e-46 ], [ 75.5625 , 2.4998711691874087e-46 ], [ 75.625 , 2.494203314821231e-46 ], [ 75.6875 , 2.4885630763841276e-46 ], 
                    [ 75.75 , 2.482935592436414e-46 ], [ 75.8125 , 2.4773367454278647e-46 ], [ 75.875 , 2.4717588696502024e-46 ], [ 75.9375 , 2.466193552802288e-46 ], 
                    [ 76.0 , 2.4606626125040095e-46 ], [ 76.0625 , 2.4551462989384248e-46 ], [ 76.125 , 2.4496460222319175e-46 ], [ 76.1875 , 2.4441782952545317e-46 ], 
                    [ 76.25 , 2.438722772505016e-46 ], [ 76.3125 , 2.4332887083980853e-46 ], [ 76.375 , 2.4278810647714154e-46 ], [ 76.4375 , 2.4224854388759647e-46 ], 
                    [ 76.5 , 2.4171165055679484e-46 ], [ 76.5625 , 2.4117682281336272e-46 ], [ 76.625 , 2.4064317846639834e-46 ], [ 76.6875 , 2.4011269944205567e-46 ], 
                    [ 76.75 , 2.395837304360724e-46 ], [ 76.8125 , 2.3905609716492213e-46 ], [ 76.875 , 2.3853175369429374e-46 ], [ 76.9375 , 2.3800856031386782e-46 ], 
                    [ 77.0 , 2.3748717497653118e-46 ], [ 77.0625 , 2.3696855738652218e-46 ], [ 77.125 , 2.364510723385289e-46 ], [ 77.1875 , 2.359358417455357e-46 ], 
                    [ 77.25 , 2.3542287111607666e-46 ], [ 77.3125 , 2.3491101578501376e-46 ], [ 77.375 , 2.344018404332468e-46 ], [ 77.4375 , 2.338944439108142e-46 ], 
                    [ 77.5 , 2.3338814572118527e-46 ], [ 77.5625 , 2.328849287152826e-46 ], [ 77.625 , 2.3238303183118474e-46 ], [ 77.6875 , 2.3188240014772275e-46 ], 
                    [ 77.75 , 2.3138486987646407e-46 ], [ 77.8125 , 2.308884071134398e-46 ], [ 77.875 , 2.3039359001118344e-46 ], [ 77.9375 , 2.299014334355553e-46 ], 
                    [ 78.0 , 2.2941032818299097e-46 ], [ 78.0625 , 2.2892122981656103e-46 ], [ 78.125 , 2.2843438010136405e-46 ], [ 78.1875 , 2.279485657756999e-46 ], 
                    [ 78.25 , 2.2746509878260543e-46 ], [ 78.3125 , 2.269834881932789e-46 ], [ 78.375 , 2.265028973153762e-46 ], [ 78.4375 , 2.2602497744364735e-46 ], 
                    [ 78.5 , 2.2554854318207425e-46 ], [ 78.5625 , 2.250731131882973e-46 ], [ 78.625 , 2.2460064745091994e-46 ], [ 78.6875 , 2.2412931458967573e-46 ], 
                    [ 78.75 , 2.2365914601003222e-46 ], [ 78.8125 , 2.231918835044594e-46 ], [ 78.875 , 2.2272559719079736e-46 ], [ 78.9375 , 2.2226075557842874e-46 ], 
                    [ 79.0 , 2.217984771017469e-46 ], [ 79.0625 , 2.213371601145977e-46 ], [ 79.125 , 2.208775553438566e-46 ], [ 79.1875 , 2.2042020937575655e-46 ], 
                    [ 79.25 , 2.1996381038180334e-46 ], [ 79.3125 , 2.1950936776888247e-46 ], [ 79.375 , 2.1905688683747743e-46 ], [ 79.4375 , 2.186053386179426e-46 ], 
                    [ 79.5 , 2.1815597577304867e-46 ], [ 79.5625 , 2.1770829936865567e-46 ], [ 79.625 , 2.172615416379884e-46 ], [ 79.6875 , 2.1681717903491545e-46 ], 
                    [ 79.75 , 2.1637424454381756e-46 ], [ 79.8125 , 2.1593221492088992e-46 ], [ 79.875 , 2.1549277816282643e-46 ], [ 79.9375 , 2.1505453008438406e-46 ], 
                    [ 80.0 , 2.1461717327190384e-46 ], [ 80.0625 , 2.1418258474751822e-46 ], [ 80.125 , 2.1374896145831955e-46 ], [ 80.1875 , 2.1331632363731105e-46 ], 
                    [ 80.25 , 2.1288640130400404e-46 ], [ 80.3125 , 2.12457345445468e-46 ], [ 80.375 , 2.1202943891086355e-46 ], [ 80.4375 , 2.1160404032889132e-46 ], 
                    [ 80.5 , 2.111794952319577e-46 ], [ 80.5625 , 2.1075624751179167e-46 ], [ 80.625 , 2.1033531547442847e-46 ], [ 80.6875 , 2.099152241418233e-46 ], 
                    [ 80.75 , 2.0949656360895053e-46 ], [ 80.8125 , 2.090800440033823e-46 ], [ 80.875 , 2.0866435251918665e-46 ], [ 80.9375 , 2.0825020877728157e-46 ], 
                    [ 81.0 , 2.0783804369817705e-46 ], [ 81.0625 , 2.0742669436880068e-46 ], [ 81.125 , 2.070169987447659e-46 ], [ 81.1875 , 2.066091398928879e-46 ], 
                    [ 81.25 , 2.0620208459261905e-46 ], [ 81.3125 , 2.0579677123261924e-46 ], [ 81.375 , 2.053931628518135e-46 ], [ 81.4375 , 2.049903460272785e-46 ], 
                    [ 81.5 , 2.04589345415078e-46 ], [ 81.5625 , 2.0418993287359126e-46 ], [ 81.625 , 2.0379130009107824e-46 ], [ 81.6875 , 2.033945473232936e-46 ], 
                    [ 81.75 , 2.029992822105356e-46 ], [ 81.8125 , 2.0260478523296957e-46 ], [ 81.875 , 2.022122171459514e-46 ], [ 81.9375 , 2.018210487983516e-46 ], 
                    [ 82.0 , 2.0143063714428214e-46 ], [ 82.0625 , 2.010421891107229e-46 ], [ 82.125 , 2.0065506578743647e-46 ], [ 82.1875 , 2.002686879020509e-46 ], 
                    [ 82.25 , 1.9988429683678834e-46 ], [ 82.3125 , 1.995011692447929e-46 ], [ 82.375 , 1.9911877601139136e-46 ], [ 82.4375 , 1.9873838215837543e-46 ], 
                    [ 82.5 , 1.9835920399523724e-46 ], [ 82.5625 , 1.979807492760492e-46 ], [ 82.625 , 1.9760429041040465e-46 ], [ 82.6875 , 1.9722901054116907e-46 ], 
                    [ 82.75 , 1.968544433840887e-46 ], [ 82.8125 , 1.9648186159574927e-46 ], [ 82.875 , 1.9611043608488597e-46 ], [ 82.9375 , 1.9573971270962027e-46 ], 
                    [ 83.0 , 1.953709500046914e-46 ], [ 83.0625 , 1.950033318980404e-46 ], [ 83.125 , 1.9463640551691122e-46 ], [ 83.1875 , 1.942714044546597e-46 ], 
                    [ 83.25 , 1.9390754821783682e-46 ], [ 83.3125 , 1.9354437345733576e-46 ], [ 83.375 , 1.9318307436019133e-46 ], [ 83.4375 , 1.9282293045754728e-46 ], 
                    [ 83.5 , 1.924634579575706e-46 ], [ 83.5625 , 1.9210580926684087e-46 ], [ 83.625 , 1.9174934075433004e-46 ], [ 83.6875 , 1.9139353369917933e-46 ], 
                    [ 83.75 , 1.9103947751848523e-46 ], [ 83.8125 , 1.906866326451615e-46 ], [ 83.875 , 1.9033443946700975e-46 ], [ 83.9375 , 1.899839246541976e-46 ], 
                    [ 84.0 , 1.8963466830898022e-46 ], [ 84.0625 , 1.8928605401805176e-46 ], [ 84.125 , 1.889390267021308e-46 ], [ 84.1875 , 1.8859331021784258e-46 ], 
                    [ 84.25 , 1.88248226317991e-46 ], [ 84.3125 , 1.8790463401084812e-46 ], [ 84.375 , 1.8756241891549357e-46 ], [ 84.4375 , 1.872208270680521e-46 ], 
                    [ 84.5 , 1.868806174006895e-46 ], [ 84.5625 , 1.8654185702884843e-46 ], [ 84.625 , 1.8620371073133657e-46 ], [ 84.6875 , 1.858668343543845e-46 ], 
                    [ 84.75 , 1.8553149634236312e-46 ], [ 84.8125 , 1.851967633418945e-46 ], [ 84.875 , 1.8486317388462297e-46 ], [ 84.9375 , 1.845312149919902e-46 ], 
                    [ 85.0 , 1.841998521981048e-46 ], [ 85.0625 , 1.8386949768243653e-46 ], [ 85.125 , 1.8354087317830454e-46 ], [ 85.1875 , 1.832128360150072e-46 ], 
                    [ 85.25 , 1.82885666393604e-46 ], [ 85.3125 , 1.825603443349445e-46 ], [ 85.375 , 1.8223560096811988e-46 ], [ 85.4375 , 1.819115736155976e-46 ], 
                    [ 85.5 , 1.815895095958735e-46 ], [ 85.5625 , 1.812680157720485e-46 ], [ 85.625 , 1.8094709113461822e-46 ], [ 85.6875 , 1.806282399987126e-46 ], 
                    [ 85.75 , 1.8030996246987236e-46 ], [ 85.8125 , 1.7999224576466407e-46 ], [ 85.875 , 1.7967642384656803e-46 ], [ 85.9375 , 1.7936132597420152e-46 ], 
                    [ 86.0 , 1.7904678068779503e-46 ], [ 86.0625 , 1.787339380618034e-46 ], [ 86.125 , 1.784219764892252e-46 ], [ 86.1875 , 1.7811055941325236e-46 ], 
                    [ 86.25 , 1.7780065623011038e-46 ], [ 86.3125 , 1.7749180213083014e-46 ], [ 86.375 , 1.7718348453606378e-46 ], [ 86.4375 , 1.768764768923838e-46 ], 
                    [ 86.5 , 1.765706903938526e-46 ], [ 86.5625 , 1.7626543254324754e-46 ], [ 86.625 , 1.759612763121037e-46 ], [ 86.6875 , 1.7565852201443203e-46 ], 
                    [ 86.75 , 1.7535628862776114e-46 ], [ 86.8125 , 1.7505494254407315e-46 ], [ 86.875 , 1.7475518874542993e-46 ], [ 86.9375 , 1.7445594822758008e-46 ], 
                    [ 87.0 , 1.741573726158555e-46 ], [ 87.0625 , 1.7386058340573474e-46 ], [ 87.125 , 1.7356429996711447e-46 ], [ 87.1875 , 1.7326852143808118e-46 ], 
                    [ 87.25 , 1.7297459491683727e-46 ], [ 87.3125 , 1.726812352790657e-46 ], [ 87.375 , 1.7238837317029442e-46 ], [ 87.4375 , 1.7209711874675924e-46 ], 
                    [ 87.5 , 1.7180665189044469e-46 ], [ 87.5625 , 1.7151667528635434e-46 ], [ 87.625 , 1.712280501047255e-46 ], [ 87.6875 , 1.70940437191509e-46 ], 
                    [ 87.75 , 1.7065330738364582e-46 ], [ 87.8125 , 1.7036727422192835e-46 ], [ 87.875 , 1.7008248866933957e-46 ], [ 87.9375 , 1.697981791636403e-46 ], 
                    [ 88.0 , 1.6951470157897405e-46 ], [ 88.0625 , 1.6923270893851551e-46 ], [ 88.125 , 1.6895118540102642e-46 ], [ 88.1875 , 1.6867022465748434e-46 ], 
                    [ 88.25 , 1.6839099410792282e-46 ], [ 88.3125 , 1.6811222581955984e-46 ], [ 88.375 , 1.6783391902713145e-46 ], [ 88.4375 , 1.6755724572494779e-46 ], 
                    [ 88.5 , 1.6728120129428846e-46 ], [ 88.5625 , 1.6700561163672936e-46 ], [ 88.625 , 1.6673136605713224e-46 ], [ 88.6875 , 1.6645801629751674e-46 ], 
                    [ 88.75 , 1.6618511468448009e-46 ], [ 88.8125 , 1.6591326163240251e-46 ], [ 88.875 , 1.6564257657032678e-46 ], [ 88.9375 , 1.6537233312697584e-46 ], 
                    [ 89.0 , 1.6510283779499782e-46 ], [ 89.0625 , 1.648347878232691e-46 ], [ 89.125 , 1.645671730396226e-46 ], [ 89.1875 , 1.6430000114826852e-46 ], 
                    [ 89.25 , 1.6403455689370933e-46 ], [ 89.3125 , 1.637695414927801e-46 ], [ 89.375 , 1.6350495425262536e-46 ], [ 89.4375 , 1.632417918832645e-46 ], 
                    [ 89.5 , 1.629793472255366e-46 ], [ 89.5625 , 1.6271732450141645e-46 ], [ 89.625 , 1.6245640080423802e-46 ], [ 89.6875 , 1.6219649524315053e-46 ], 
                    [ 89.75 , 1.6193700549147736e-46 ], [ 89.8125 , 1.6167828875616277e-46 ], [ 89.875 , 1.6142089970012502e-46 ], [ 89.9375 , 1.6116392040304045e-46 ], 
                    [ 90.0 , 1.6090738105596253e-46 ], [ 90.0625 , 1.6065247885956325e-46 ], [ 90.125 , 1.603979804676973e-46 ], [ 90.1875 , 1.6014388524067236e-46 ], 
                    [ 90.25 , 1.5989114194127376e-46 ], [ 90.3125 , 1.5963909597486924e-46 ], [ 90.375 , 1.593874473235917e-46 ], [ 90.4375 , 1.5913680290212212e-46 ], 
                    [ 90.5 , 1.588871859210393e-46 ], [ 90.5625 , 1.5863796048129882e-46 ], [ 90.625 , 1.583893844073404e-46 ], [ 90.6875 , 1.5814216555029987e-46 ], 
                    [ 90.75 , 1.578953325597998e-46 ], [ 90.8125 , 1.576488848335615e-46 ], [ 90.875 , 1.5740395108625591e-46 ], [ 90.9375 , 1.571594901630439e-46 ], 
                    [ 91.0 , 1.5691540890719674e-46 ], [ 91.0625 , 1.5667246814587596e-46 ], [ 91.125 , 1.564303485947047e-46 ], [ 91.1875 , 1.5618860321187168e-46 ], 
                    [ 91.25 , 1.5594762714427092e-46 ], [ 91.3125 , 1.5570782623492493e-46 ], [ 91.375 , 1.5546839406781328e-46 ], [ 91.4375 , 1.5522935522255538e-46 ], 
                    [ 91.5 , 1.5499184763011142e-46 ], [ 91.5625 , 1.547547034345008e-46 ], [ 91.625 , 1.545179220797129e-46 ], [ 91.6875 , 1.542823344021102e-46 ], 
                    [ 91.75 , 1.5404745315410476e-46 ], [ 91.8125 , 1.5381292949208636e-46 ], [ 91.875 , 1.5357920930339014e-46 ], [ 91.9375 , 1.5334656691484196e-46 ], 
                    [ 92.0 , 1.5311427693389634e-46 ], [ 92.0625 , 1.5288239750020656e-46 ], [ 92.125 , 1.5265197304247115e-46 ], [ 92.1875 , 1.5242189588064045e-46 ], 
                    [ 92.25 , 1.5219216549127418e-46 ], [ 92.3125 , 1.519635982597716e-46 ], [ 92.375 , 1.5173570464987222e-46 ], [ 92.4375 , 1.5150815280272408e-46 ], 
                    [ 92.5 , 1.5128136066564645e-46 ], [ 92.5625 , 1.5105563330301988e-46 ], [ 92.625 , 1.508302427488552e-46 ], [ 92.6875 , 1.506052017824673e-46 ], 
                    [ 92.75 , 1.5038161412509084e-46 ], [ 92.8125 , 1.5015835840472438e-46 ], [ 92.875 , 1.4993543412857712e-46 ], [ 92.9375 , 1.4971357172732323e-46 ], 
                    [ 93.0 , 1.4949243006058377e-46 ], [ 93.0625 , 1.4927161504182985e-46 ], [ 93.125 , 1.4905144038730555e-46 ], [ 93.1875 , 1.4883239207751292e-46 ], 
                    [ 93.25 , 1.486136656845124e-46 ], [ 93.3125 , 1.483952607352175e-46 ], [ 93.375 , 1.4817817199916882e-46 ], [ 93.4375 , 1.4796150562103032e-46 ], 
                    [ 93.5 , 1.477451560528418e-46 ], [ 93.5625 , 1.47529695182862e-46 ], [ 93.625 , 1.4731507540547647e-46 ], [ 93.6875 , 1.471007678476e-46 ], 
                    [ 93.75 , 1.468869118523407e-46 ], [ 93.8125 , 1.466743148951215e-46 ], [ 93.875 , 1.4646202564037768e-46 ], [ 93.9375 , 1.4625004364275455e-46 ], 
                    [ 94.0 , 1.4603915299674843e-46 ], [ 94.0625 , 1.4582885818641545e-46 ], [ 94.125 , 1.456188661983496e-46 ], [ 94.1875 , 1.4540952088854565e-46 ], 
                    [ 94.25 , 1.4520120282993298e-46 ], [ 94.3125 , 1.4499318321404763e-46 ], [ 94.375 , 1.4478546161332642e-46 ], [ 94.4375 , 1.4457900034482193e-46 ], 
                    [ 94.5 , 1.443729330252756e-46 ], [ 94.5625 , 1.4416715941187076e-46 ], [ 94.625 , 1.4396218754867464e-46 ], [ 94.6875 , 1.4375804987724273e-46 ], 
                    [ 94.75 , 1.4355420167202143e-46 ], [ 94.8125 , 1.4335069770292646e-46 ], [ 94.875 , 1.43148472309686e-46 ], [ 94.9375 , 1.4294653219660342e-46 ], 
                    [ 95.0 , 1.4274487696123081e-46 ], [ 95.0625 , 1.4254414310357035e-46 ], [ 95.125 , 1.4234409079577208e-46 ], [ 95.1875 , 1.4214431924960203e-46 ], 
                    [ 95.25 , 1.4194500111317043e-46 ], [ 95.3125 , 1.4174681716097086e-46 ], [ 95.375 , 1.4154890991368434e-46 ], [ 95.4375 , 1.413512789849698e-46 ], 
                    [ 95.5 , 1.4115465173406464e-46 ], [ 95.5625 , 1.4095858882989977e-46 ], [ 95.625 , 1.4076279825584883e-46 ], [ 95.6875 , 1.405675353094966e-46 ], 
                    [ 95.75 , 1.4037329961088108e-46 ], [ 95.8125 , 1.4017933230643508e-46 ], [ 95.875 , 1.399856330252895e-46 ], [ 95.9375 , 1.3979298880960387e-46 ], 
                    [ 96.0 , 1.396008279879724e-46 ], [ 96.0625 , 1.3940893131249926e-46 ], [ 96.125 , 1.392176021849694e-46 ], [ 96.1875 , 1.3902722628183533e-46 ], 
                    [ 96.25 , 1.3883711071205138e-46 ], [ 96.3125 , 1.3864725511961547e-46 ], [ 96.375 , 1.3845846899687939e-46 ], [ 96.4375 , 1.3827011773035236e-46 ], 
                    [ 96.5 , 1.3808202268650345e-46 ], [ 96.5625 , 1.3789450391601435e-46 ], [ 96.625 , 1.3770789988086983e-46 ], [ 96.6875 , 1.3752154836533327e-46 ], 
                    [ 96.75 , 1.3733544902768463e-46 ], [ 96.8125 , 1.3715040436337262e-46 ], [ 96.875 , 1.3696577880452273e-46 ], [ 96.9375 , 1.3678140178009658e-46 ], 
                    [ 97.0 , 1.3659757810294271e-46 ], [ 97.0625 , 1.3641466080139179e-46 ], [ 97.125 , 1.3623198844371032e-46 ], [ 97.1875 , 1.3604956070189195e-46 ], 
                    [ 97.25 , 1.3586814252331084e-46 ], [ 97.3125 , 1.3568715854932249e-46 ], [ 97.375 , 1.355064156561268e-46 ], [ 97.4375 , 1.353261736607328e-46 ], 
                    [ 97.5 , 1.3514686015623092e-46 ], [ 97.5625 , 1.3496778425049978e-46 ], [ 97.625 , 1.3478894562871272e-46 ], [ 97.6875 , 1.3461104256543146e-46 ], 
                    [ 97.75 , 1.3443361839857664e-46 ], [ 97.8125 , 1.3425642808575517e-46 ], [ 97.875 , 1.3407965844467238e-46 ], [ 97.9375 , 1.3390387034191254e-46 ], 
                    [ 98.0 , 1.3372831271003416e-46 ], [ 98.0625 , 1.3355298524687333e-46 ], [ 98.125 , 1.3337849232865542e-46 ], [ 98.1875 , 1.3320455167372818e-46 ], 
                    [ 98.25 , 1.3303083785710928e-46 ], [ 98.3125 , 1.3285743651481086e-46 ], [ 98.375 , 1.326850936697932e-46 ], [ 98.4375 , 1.3251297438814543e-46 ], 
                    [ 98.5 , 1.3234107837986282e-46 ], [ 98.5625 , 1.321698892233393e-46 ], [ 98.625 , 1.319993530825589e-46 ], [ 98.6875 , 1.318290369811194e-46 ], 
                    [ 98.75 , 1.3165894063510754e-46 ], [ 98.8125 , 1.3148992752445648e-46 ], [ 98.875 , 1.3132117208094033e-46 ], [ 98.9375 , 1.3115263321979006e-46 ], 
                    [ 99.0 , 1.3098464788516846e-46 ], [ 99.0625 , 1.308174425128966e-46 ], [ 99.125 , 1.3065045058271193e-46 ], [ 99.1875 , 1.3048367182215001e-46 ], 
                    [ 99.25 , 1.3031780445687932e-46 ], [ 99.3125 , 1.3015234010075941e-46 ], [ 99.375 , 1.2998708583453e-46 ], [ 99.4375 , 1.2982220712798571e-46 ], 
                    [ 99.5 , 1.2965826020288877e-46 ], [ 99.5625 , 1.2949452031936877e-46 ], [ 99.625 , 1.2933098721595806e-46 ], [ 99.6875 , 1.2916816978027906e-46 ], 
                    [ 99.75 , 1.2900592470413787e-46 ], [ 99.8125 , 1.2884388342019175e-46 ], [ 99.875 , 1.2868204567246031e-46 ], [ 99.9375 , 1.2852125276107988e-46 ], 
                    [ 100.0 , 1.2836068981647737e-46 ], [ 100.0625 , 1.2820032746484001e-46 ], [ 100.125 , 1.280404647581005e-46 ], [ 100.1875 , 1.2788137023862767e-46 ], 
                    [ 100.25 , 1.2772247339936523e-46 ], [ 100.3125 , 1.2756377399468895e-46 ], [ 100.375 , 1.2740588874179047e-46 ], [ 100.4375 , 1.2724844563185452e-46 ], 
                    [ 100.5 , 1.2709119708382431e-46 ], [ 100.5625 , 1.269342062530727e-46 ], [ 100.625 , 1.2677818623232072e-46 ], [ 100.6875 , 1.2662235798215372e-46 ], 
                    [ 100.75 , 1.2646672126685686e-46 ], [ 100.8125 , 1.263116435013161e-46 ], [ 100.875 , 1.2615724536127548e-46 ], [ 100.9375 , 1.260030359511491e-46 ], 
                    [ 101.0 , 1.2584901504024135e-46 ], [ 101.0625 , 1.256958283813546e-46 ], [ 101.125 , 1.255430171802155e-46 ], [ 101.1875 , 1.2539039175503735e-46 ], 
                    [ 101.25 , 1.2523804436521926e-46 ], [ 101.3125 , 1.2508662884794854e-46 ], [ 101.375 , 1.2493539639533046e-46 ], [ 101.4375 , 1.2478434678603324e-46 ], 
                    [ 101.5 , 1.2463383962486726e-46 ], [ 101.5625 , 1.2448397468580219e-46 ], [ 101.625 , 1.2433428995060582e-46 ], [ 101.6875 , 1.2418478520259347e-46 ], 
                    [ 101.75 , 1.2403608230159093e-46 ], [ 101.8125 , 1.2388775557048572e-46 ], [ 101.875 , 1.2373960621372706e-46 ], [ 101.9375 , 1.2359169132017415e-46 ], 
                    [ 102.0 , 1.2344471455227027e-46 ], [ 102.0625 , 1.2329791257095744e-46 ], [ 102.125 , 1.231512851683747e-46 ], [ 102.1875 , 1.2300513321935842e-46 ], 
                    [ 102.25 , 1.228596599695602e-46 ], [ 102.3125 , 1.2271435876515521e-46 ], [ 102.375 , 1.2256922940266963e-46 ], [ 102.4375 , 1.2242480603334629e-46 ], 
                    [ 102.5 , 1.222808181034418e-46 ], [ 102.5625 , 1.221369995225821e-46 ], [ 102.625 , 1.2199335009159093e-46 ], [ 102.6875 , 1.218506318191564e-46 ], 
                    [ 102.75 , 1.2170811675188938e-46 ], [ 102.8125 , 1.2156576836857137e-46 ], [ 102.875 , 1.2142377139073123e-46 ], [ 102.9375 , 1.2128254950903174e-46 ], 
                    [ 103.0 , 1.2114149187539778e-46 ], [ 103.0625 , 1.210005982988058e-46 ], [ 103.125 , 1.2086026256735897e-46 ], [ 103.1875 , 1.20720479609423e-46 ], 
                    [ 103.25 , 1.2058085831980558e-46 ], [ 103.3125 , 1.204413985115257e-46 ], [ 103.375 , 1.203026894104515e-46 ], [ 103.4375 , 1.2016432256373267e-46 ], 
                    [ 103.5 , 1.2002611486045919e-46 ], [ 103.5625 , 1.1988807874369e-46 ], [ 103.625 , 1.1975096918591573e-46 ], [ 103.6875 , 1.1961401643298003e-46 ], 
                    [ 103.75 , 1.1947722030555447e-46 ], [ 103.8125 , 1.193407798318793e-46 ], [ 103.875 , 1.1920506055875237e-46 ], [ 103.9375 , 1.1906949563119972e-46 ], 
                    [ 104.0 , 1.1893408487369237e-46 ], [ 104.0625 , 1.1879920555584423e-46 ], [ 104.125 , 1.1866486229932904e-46 ], [ 104.1875 , 1.1853067096395242e-46 ], 
                    [ 104.25 , 1.1839663137791627e-46 ], [ 104.3125 , 1.1826329224378656e-46 ], [ 104.375 , 1.1813031311917146e-46 ], [ 104.4375 , 1.1799748352064726e-46 ], 
                    [ 104.5 , 1.1786480328007936e-46 ], [ 104.5625 , 1.1773297631228068e-46 ], [ 104.625 , 1.176013402806441e-46 ], [ 104.6875 , 1.1746985142990167e-46 ], 
                    [ 104.75 , 1.1733862316203942e-46 ], [ 104.8125 , 1.1720817515471275e-46 ], [ 104.875 , 1.1707787216939215e-46 ], [ 104.9375 , 1.1694771404485581e-46 ], 
                    [ 105.0 , 1.1681796217410802e-46 ], [ 105.0625 , 1.1668883253657052e-46 ], [ 105.125 , 1.1655984563789686e-46 ], [ 105.1875 , 1.164310013203046e-46 ], 
                    [ 105.25 , 1.1630269773877541e-46 ], [ 105.3125 , 1.1617486627102629e-46 ], [ 105.375 , 1.1604717530632724e-46 ], [ 105.4375 , 1.1591962469025414e-46 ], 
                    [ 105.5 , 1.1579275049728749e-46 ], [ 105.5625 , 1.1566621288120014e-46 ], [ 105.625 , 1.1553981354465369e-46 ], [ 105.6875 , 1.1541355233653384e-46 ], 
                    [ 105.75 , 1.1528808267362497e-46 ], [ 105.8125 , 1.1516281721227252e-46 ], [ 105.875 , 1.1503768785723337e-46 ], [ 105.9375 , 1.1491274675238232e-46 ], 
                    [ 106.0 , 1.1478860662503362e-46 ], [ 106.0625 , 1.1466460060614408e-46 ], [ 106.125 , 1.1454072855083323e-46 ], [ 106.1875 , 1.1441715552278013e-46 ], 
                    [ 106.25 , 1.1429426507647163e-46 ], [ 106.3125 , 1.1417150662139683e-46 ], [ 106.375 , 1.1404888001579464e-46 ], [ 106.4375 , 1.1392665433220693e-46 ], 
                    [ 106.5 , 1.1380499751931348e-46 ], [ 106.5625 , 1.1368347061789952e-46 ], [ 106.625 , 1.1356207348923637e-46 ], [ 106.6875 , 1.1344117252009565e-46 ], 
                    [ 106.75 , 1.1332073650809908e-46 ], [ 106.8125 , 1.1320042835826007e-46 ], [ 106.875 , 1.1308024793482922e-46 ], [ 106.9375 , 1.1296065150230178e-46 ], 
                    [ 107.0 , 1.1284142353095151e-46 ], [ 107.0625 , 1.1272232140261871e-46 ], [ 107.125 , 1.1260334498447594e-46 ], [ 107.1875 , 1.1248502845223248e-46 ], 
                    [ 107.25 , 1.1236698988614843e-46 ], [ 107.3125 , 1.1224907518635613e-46 ], [ 107.375 , 1.1213128422286747e-46 ], [ 107.4375 , 1.120142356885967e-46 ], 
                    [ 107.5 , 1.1189738492915403e-46 ], [ 107.5625 , 1.1178065606582482e-46 ], [ 107.625 , 1.1166404897145054e-46 ], [ 107.6875 , 1.115482397776374e-46 ], 
                    [ 107.75 , 1.1143255296956951e-46 ], [ 107.8125 , 1.1131698614042363e-46 ], [ 107.875 , 1.1120160254167208e-46 ], [ 107.9375 , 1.1108695278271111e-46 ], 
                    [ 108.0 , 1.109724212285865e-46 ], [ 108.0625 , 1.108580077574273e-46 ], [ 108.125 , 1.1074383432336688e-46 ], [ 108.1875 , 1.1063032865268101e-46 ], 
                    [ 108.25 , 1.1051693931838033e-46 ], [ 108.3125 , 1.1040366620122398e-46 ], [ 108.375 , 1.102906835765469e-46 ], [ 108.4375 , 1.101783107753399e-46 ], 
                    [ 108.5 , 1.100660524683589e-46 ], [ 108.5625 , 1.0995390853894976e-46 ], [ 108.625 , 1.098420990753721e-46 ], [ 108.6875 , 1.0973084816443032e-46 ], 
                    [ 108.75 , 1.0961970993128129e-46 ], [ 108.8125 , 1.095086842617981e-46 ], [ 108.875 , 1.0939802935488729e-46 ], [ 108.9375 , 1.0928788623747525e-46 ], 
                    [ 109.0 , 1.0917785401334332e-46 ], [ 109.0625 , 1.0906793257084298e-46 ], [ 109.125 , 1.0895841140068091e-46 ], [ 109.1875 , 1.088493617951129e-46 ], 
                    [ 109.25 , 1.0874042133041939e-46 ], [ 109.3125 , 1.0863158989737125e-46 ], [ 109.375 , 1.085231863666118e-46 ], [ 109.4375 , 1.0841522534614287e-46 ], 
                    [ 109.5 , 1.0830737172744217e-46 ], [ 109.5625 , 1.081996254036604e-46 ], [ 109.625 , 1.0809232026392274e-46 ], [ 109.6875 , 1.0798542659333813e-46 ], 
                    [ 109.75 , 1.0787863863106713e-46 ], [ 109.8125 , 1.0777195627257294e-46 ], [ 109.875 , 1.076657314850953e-46 ], [ 109.9375 , 1.0755990150697168e-46 ], 
                    [ 110.0 , 1.0745417555437101e-46 ], [ 110.0625 , 1.0734855352504529e-46 ], [ 110.125 , 1.0724339584204744e-46 ], [ 110.1875 , 1.0713861905923791e-46 ], 
                    [ 110.25 , 1.0703394464333061e-46 ], [ 110.3125 , 1.0692937249431232e-46 ], [ 110.375 , 1.0682526284087604e-46 ], [ 110.4375 , 1.0672152408261285e-46 ], 
                    [ 110.5 , 1.066178860657833e-46 ], [ 110.5625 , 1.065143486925529e-46 ], [ 110.625 , 1.0641126669223583e-46 ], [ 110.6875 , 1.0630855160870922e-46 ], 
                    [ 110.75 , 1.062059356724672e-46 ], [ 110.8125 , 1.061034187878094e-46 ], [ 110.875 , 1.060013526565379e-46 ], [ 110.9375 , 1.0589966121374879e-46 ], 
                    [ 111.0 , 1.0579806732772775e-46 ], [ 111.0625 , 1.0569657090488836e-46 ], [ 111.125 , 1.0559550353804097e-46 ], [ 111.1875 , 1.0549481527222834e-46 ], 
                    [ 111.25 , 1.0539422301549234e-46 ], [ 111.3125 , 1.0529372667628642e-46 ], [ 111.375 , 1.051936411560068e-46 ], [ 111.4375 , 1.0509395186833065e-46 ], 
                    [ 111.5 , 1.0499435705361205e-46 ], [ 111.5625 , 1.0489485662231539e-46 ], [ 111.625 , 1.04795738800018e-46 ], [ 111.6875 , 1.0469703561604795e-46 ], 
                    [ 111.75 , 1.045984253969125e-46 ], [ 111.8125 , 1.0449990805505081e-46 ], [ 111.875 , 1.044017403651393e-46 ], [ 111.9375 , 1.043040110104898e-46 ], 
                    [ 112.0 , 1.0420637313924663e-46 ], [ 112.0625 , 1.0410882666577319e-46 ], [ 112.125 , 1.0401159273213553e-46 ], [ 112.1875 , 1.0391482685115049e-46 ], 
                    [ 112.25 , 1.0381815099508936e-46 ], [ 112.3125 , 1.0372156508019935e-46 ], [ 112.375 , 1.036252513926714e-46 ], [ 112.4375 , 1.0352944428454243e-46 ], 
                    [ 112.5 , 1.0343372575522917e-46 ], [ 112.5625 , 1.0333809572283427e-46 ], [ 112.625 , 1.0324268998183825e-46 ], [ 112.6875 , 1.0314782864813167e-46 ], 
                    [ 112.75 , 1.0305305447481251e-46 ], [ 112.8125 , 1.0295836738179178e-46 ], [ 112.875 , 1.0286385314593639e-46 ], [ 112.9375 , 1.0276993032880002e-46 ], 
                    [ 113.0 , 1.026760932706092e-46 ], [ 113.0625 , 1.0258234189305845e-46 ], [ 113.125 , 1.0248870628634301e-46 ], [ 113.1875 , 1.02395705641945e-46 ], 
                    [ 113.25 , 1.023027893885018e-46 ], [ 113.3125 , 1.0220995744943803e-46 ], [ 113.375 , 1.0211720974824124e-46 ], [ 113.4375 , 1.0202509936635358e-46 ], 
                    [ 113.5 , 1.0193310140106876e-46 ], [ 113.5625 , 1.0184118639209244e-46 ], [ 113.625 , 1.0174935426462419e-46 ], [ 113.6875 , 1.0165808776475704e-46 ], 
                    [ 113.75 , 1.0156699654767598e-46 ], [ 113.8125 , 1.0147598695331832e-46 ], [ 113.875 , 1.0138505890854342e-46 ], [ 113.9375 , 1.0129462355914122e-46 ], 
                    [ 114.0 , 1.0120443132963515e-46 ], [ 114.0625 , 1.011143194068414e-46 ], [ 114.125 , 1.010242877192581e-46 ], [ 114.1875 , 1.0093466915886041e-46 ], 
                    [ 114.25 , 1.0084536504705142e-46 ], [ 114.3125 , 1.0075613994896794e-46 ], [ 114.375 , 1.0066699379470371e-46 ], [ 114.4375 , 1.0057817677460307e-46 ], 
                    [ 114.5 , 1.004897484028591e-46 ], [ 114.5625 , 1.0040139777737365e-46 ], [ 114.625 , 1.0031312482979141e-46 ], [ 114.6875 , 1.0022509790585439e-46 ], 
                    [ 114.75 , 1.001375462865221e-46 ], [ 114.8125 , 1.0005007114789241e-46 ], [ 114.875 , 9.996267242315845e-47 ], [ 114.9375 , 9.987542650787956e-47 ], 
                    [ 115.0 , 9.978873233642275e-47 ], [ 115.0625 , 9.97021134175043e-47 ], [ 115.125 , 9.961556968580398e-47 ], [ 115.1875 , 9.95291010760547e-47 ], 
                    [ 115.25 , 9.944324437637185e-47 ], [ 115.3125 , 9.935747810067244e-47 ], [ 115.375 , 9.927178579534702e-47 ], [ 115.4375 , 9.918616739659877e-47 ], 
                    [ 115.5 , 9.910106083454979e-47 ], [ 115.5625 , 9.901614155943847e-47 ], [ 115.625 , 9.893129505129213e-47 ], [ 115.6875 , 9.8846521247755e-47 ], 
                    [ 115.75 , 9.87621477391461e-47 ], [ 115.8125 , 9.867805916849722e-47 ], [ 115.875 , 9.859404219296799e-47 ], [ 115.9375 , 9.851009675159908e-47 ], 
                    [ 116.0 , 9.842644695695258e-47 ], [ 116.0625 , 9.834319104503426e-47 ], [ 116.125 , 9.826000555673838e-47 ], [ 116.1875 , 9.817689043249366e-47 ], 
                    [ 116.25 , 9.809395473775671e-47 ], [ 116.3125 , 9.80115158069833e-47 ], [ 116.375 , 9.79291461585347e-47 ], [ 116.4375 , 9.784684573418629e-47 ], 
                    [ 116.5 , 9.776461447576291e-47 ], [ 116.5625 , 9.7682979034219e-47 ], [ 116.625 , 9.760141815188183e-47 ], [ 116.6875 , 9.75199253692033e-47 ], 
                    [ 116.75 , 9.743850062932329e-47 ], [ 116.8125 , 9.735754999349387e-47 ], [ 116.875 , 9.727679262482553e-47 ], [ 116.9375 , 9.719610224379803e-47 ], 
                    [ 117.0 , 9.711547879484634e-47 ], [ 117.0625 , 9.703520002708415e-47 ], [ 117.125 , 9.695523543779903e-47 ], [ 117.1875 , 9.68753367455843e-47 ], 
                    [ 117.25 , 9.679550389613837e-47 ], [ 117.3125 , 9.671588552289451e-47 ], [ 117.375 , 9.6636706897759e-47 ], [ 117.4375 , 9.655759309397707e-47 ], 
                    [ 117.5 , 9.647854405848401e-47 ], [ 117.5625 , 9.639957558444937e-47 ], [ 117.625 , 9.63211767447137e-47 ], [ 117.6875 , 9.624284166436738e-47 ], 
                    [ 117.75 , 9.616457029155817e-47 ], [ 117.8125 , 9.608636257447129e-47 ], [ 117.875 , 9.600861492138005e-47 ], [ 117.9375 , 9.59310505979974e-47 ], 
                    [ 118.0 , 9.58535489379948e-47 ], [ 118.0625 , 9.57761098907479e-47 ], [ 118.125 , 9.56989868662687e-47 ], [ 118.1875 , 9.562218560291918e-47 ], 
                    [ 118.25 , 9.554544597485243e-47 ], [ 118.3125 , 9.546876793260296e-47 ], [ 118.375 , 9.539225955414607e-47 ], [ 118.4375 , 9.531621678243262e-47 ], 
                    [ 118.5 , 9.524023462887792e-47 ], [ 118.5625 , 9.516431304515959e-47 ], [ 118.625 , 9.508845198299394e-47 ], [ 118.6875 , 9.50131127226017e-47 ], 
                    [ 118.75 , 9.493787419824693e-47 ], [ 118.8125 , 9.486269525341378e-47 ], [ 118.875 , 9.478757584091865e-47 ], [ 118.9375 , 9.471282516925423e-47 ], 
                    [ 119.0 , 9.463832607811895e-47 ], [ 119.0625 , 9.456388558638345e-47 ], [ 119.125 , 9.448950364795268e-47 ], [ 119.1875 , 9.441533227661419e-47 ], 
                    [ 119.25 , 9.434156699219385e-47 ], [ 119.3125 , 9.426785933948565e-47 ], [ 119.375 , 9.419420927346007e-47 ], [ 119.4375 , 9.412061674912954e-47 ], 
                    [ 119.5 , 9.404757161365535e-47 ], [ 119.5625 , 9.397459201800373e-47 ], [ 119.625 , 9.390166905349341e-47 ], [ 119.6875 , 9.382880267618212e-47 ], 
                    [ 119.75 , 9.375631205943531e-47 ], [ 119.8125 , 9.368404877923283e-47 ], [ 119.875 , 9.36118411964181e-47 ], [ 119.9375 , 9.353968926806265e-47 ], 
                    [ 120.0 , 9.346774417878754e-47 ], [ 120.0625 , 9.339619215819667e-47 ], [ 120.125 , 9.33246949125633e-47 ], [ 120.1875 , 9.325325239995506e-47 ], 
                    [ 120.25 , 9.318186457847307e-47 ], [ 120.3125 , 9.311099895110321e-47 ], [ 120.375 , 9.304020774064988e-47 ], [ 120.4375 , 9.296947035193188e-47 ], 
                    [ 120.5 , 9.289878674402781e-47 ], [ 120.5625 , 9.282844421661891e-47 ], [ 120.625 , 9.275834893396675e-47 ], [ 120.6875 , 9.268830658065784e-47 ], 
                    [ 120.75 , 9.261831711672628e-47 ], [ 120.8125 , 9.254848845701202e-47 ], [ 120.875 , 9.247908239478112e-47 ], [ 120.9375 , 9.240972838312042e-47 ], 
                    [ 121.0 , 9.234042638299368e-47 ], [ 121.0625 , 9.22711763553939e-47 ], [ 121.125 , 9.220238397105375e-47 ], [ 121.1875 , 9.213371726212783e-47 ], 
                    [ 121.25 , 9.206510169197763e-47 ], [ 121.3125 , 9.199653722251747e-47 ], [ 121.375 , 9.192823760559675e-47 ], [ 121.4375 , 9.186024553518993e-47 ], 
                    [ 121.5 , 9.179230375315694e-47 ], [ 121.5625 , 9.172441222230344e-47 ], [ 121.625 , 9.165659598578244e-47 ], [ 121.6875 , 9.158927520709669e-47 ], 
                    [ 121.75 , 9.15220038747969e-47 ], [ 121.8125 , 9.145478195256202e-47 ], [ 121.875 , 9.138760940410082e-47 ], [ 121.9375 , 9.1320787709742e-47 ], 
                    [ 122.0 , 9.12541824786296e-47 ], [ 122.0625 , 9.118762582634617e-47 ], [ 122.125 , 9.112111771745732e-47 ], [ 122.1875 , 9.105476087424772e-47 ], 
                    [ 122.25 , 9.098881270789603e-47 ], [ 122.3125 , 9.092291230577676e-47 ], [ 122.375 , 9.08570596332932e-47 ], [ 122.4375 , 9.079125465587815e-47 ], 
                    [ 122.5 , 9.072586115039414e-47 ], [ 122.5625 , 9.066061087328157e-47 ], [ 122.625 , 9.059540752433932e-47 ], [ 122.6875 , 9.05302510698165e-47 ], 
                    [ 122.75 , 9.046530176509522e-47 ], [ 122.8125 , 9.040069641947114e-47 ], [ 122.875 , 9.03361372114329e-47 ], [ 122.9375 , 9.02716241080322e-47 ], 
                    [ 123.0 , 9.020715707634302e-47 ], [ 123.0625 , 9.014314603332582e-47 ], [ 123.125 , 9.007922633531859e-47 ], [ 123.1875 , 9.001535196219452e-47 ], 
                    [ 123.25 , 8.9951522881818e-47 ], [ 123.3125 , 8.988793608756158e-47 ], [ 123.375 , 8.982464662300085e-47 ], [ 123.4375 , 8.97614017201083e-47 ], 
                    [ 123.5 , 8.969820134750712e-47 ], [ 123.5625 , 8.963504547384575e-47 ], [ 123.625 , 8.95723707353486e-47 ], [ 123.6875 , 8.950975461846254e-47 ], 
                    [ 123.75 , 8.944718227376015e-47 ], [ 123.8125 , 8.938465367063715e-47 ], [ 123.875 , 8.93223846434371e-47 ], [ 123.9375 , 8.926038721157777e-47 ], 
                    [ 124.0 , 8.919843281127974e-47 ], [ 124.0625 , 8.913652141267223e-47 ], [ 124.125 , 8.907465298591215e-47 ], [ 124.1875 , 8.901326865825328e-47 ], 
                    [ 124.25 , 8.8951929043957e-47 ], [ 124.3125 , 8.889063169918172e-47 ], [ 124.375 , 8.882937659480046e-47 ], [ 124.4375 , 8.876838130630441e-47 ], 
                    [ 124.5 , 8.870765156735862e-47 ], [ 124.5625 , 8.864696337587819e-47 ], [ 124.625 , 8.858631670343771e-47 ], [ 124.6875 , 8.852571152163436e-47 ], 
                    [ 124.75 , 8.846557449115964e-47 ], [ 124.8125 , 8.840548651815413e-47 ], [ 124.875 , 8.834543935836286e-47 ], [ 124.9375 , 8.828543298406576e-47 ], 
                    [ 125.0 , 8.822566747744912e-47 ], [ 125.0625 , 8.816617794008507e-47 ], [ 125.125 , 8.81067285158208e-47 ], [ 125.1875 , 8.80473191776097e-47 ], 
                    [ 125.25 , 8.798794989842166e-47 ], [ 125.3125 , 8.792901756707574e-47 ], [ 125.375 , 8.787015599839413e-47 ], [ 125.4375 , 8.781133383290548e-47 ], 
                    [ 125.5 , 8.775255104423686e-47 ], [ 125.5625 , 8.769397337087908e-47 ], [ 125.625 , 8.763569999904798e-47 ], [ 125.6875 , 8.757746535035524e-47 ], 
                    [ 125.75 , 8.75192693990678e-47 ], [ 125.8125 , 8.7461112119473975e-47 ], [ 125.875 , 8.740334776608525e-47 ], [ 125.9375 , 8.734569321760993e-47 ], 
                    [ 126.0 , 8.728807670025179e-47 ], [ 126.0625 , 8.723049818892406e-47 ], [ 126.125 , 8.717307074600209e-47 ], [ 126.1875 , 8.711598903405481e-47 ], 
                    [ 126.25 , 8.7058944699726075e-47 ], [ 126.3125 , 8.70019377185419e-47 ], [ 126.375 , 8.694496806604251e-47 ], [ 126.4375 , 8.688832883539365e-47 ], 
                    [ 126.5 , 8.683185502827856e-47 ], [ 126.5625 , 8.677541792679333e-47 ], [ 126.625 , 8.671901750708204e-47 ], [ 126.6875 , 8.666270007817927e-47 ], 
                    [ 126.75 , 8.660678931325819e-47 ], [ 126.8125 , 8.655091461937424e-47 ], [ 126.875 , 8.649507597325855e-47 ], [ 126.9375 , 8.643927335165353e-47 ], 
                    [ 127.0 , 8.63837215132729e-47 ], [ 127.0625 , 8.632840593701235e-47 ], [ 127.125 , 8.62731257819281e-47 ], [ 127.1875 , 8.62178810253371e-47 ], 
                    [ 127.25 , 8.616267164457321e-47 ], [ 127.3125 , 8.610787407014974e-47 ], [ 127.375 , 8.605314750250855e-47 ], [ 127.4375 , 8.599845571678573e-47 ], 
                    [ 127.5 , 8.594379869087963e-47 ], [ 127.5625 , 8.588929745292367e-47 ], [ 127.625 , 8.58351169576907e-47 ], [ 127.6875 , 8.578097064047702e-47 ], 
                    [ 127.75 , 8.572685847972311e-47 ], [ 127.8125 , 8.567278045388005e-47 ], [ 127.875 , 8.561901057365965e-47 ], [ 127.9375 , 8.556540830480753e-47 ], 
                    [ 128.0 , 8.551183959396088e-47 ], [ 128.0625 , 8.545830442010931e-47 ], [ 128.125 , 8.540481716055653e-47 ], [ 128.1875 , 8.5351752276760795e-47 ], 
                    [ 128.25 , 8.529872036396113e-47 ], [ 128.3125 , 8.524572140167096e-47 ], [ 128.375 , 8.519275536941886e-47 ], [ 128.4375 , 8.513997814100118e-47 ], 
                    [ 128.5 , 8.508747819583333e-47 ], [ 128.5625 , 8.503501062375747e-47 ], [ 128.625 , 8.498257540480956e-47 ], [ 128.6875 , 8.493017251904086e-47 ], 
                    [ 128.75 , 8.487809528440696e-47 ], [ 128.8125 , 8.482615731367504e-47 ], [ 128.875 , 8.477425112443862e-47 ], [ 128.9375 , 8.472237669725022e-47 ], 
                    [ 129.0 , 8.467055926438638e-47 ], [ 129.0625 , 8.461914064330458e-47 ], [ 129.125 , 8.456775324765303e-47 ], [ 129.1875 , 8.451639705846924e-47 ], 
                    [ 129.25 , 8.446507205679917e-47 ], [ 129.3125 , 8.441393156441476e-47 ], [ 129.375 , 8.436306555692311e-47 ], [ 129.4375 , 8.431223020018685e-47 ], 
                    [ 129.5 , 8.426142547573472e-47 ], [ 129.5625 , 8.421065136511134e-47 ], [ 129.625 , 8.416017870888947e-47 ], [ 129.6875 , 8.410985382569291e-47 ], 
                    [ 129.75 , 8.40595590350396e-47 ], [ 129.8125 , 8.400929431893703e-47 ], [ 129.875 , 8.395905965940107e-47 ], [ 129.9375 , 8.390924304978883e-47 ], 
                    [ 130.0 , 8.385945680683582e-47 ], [ 130.0625 , 8.380970010377372e-47 ], [ 130.125 , 8.375997292307731e-47 ], [ 130.1875 , 8.371038642104952e-47 ], 
                    [ 130.25 , 8.366110417334619e-47 ], [ 130.3125 , 8.36118509392455e-47 ], [ 130.375 , 8.356262670166708e-47 ], [ 130.4375 , 8.351343144353999e-47 ], 
                    [ 130.5 , 8.346448242871675e-47 ], [ 130.5625 , 8.341572791064372e-47 ], [ 130.625 , 8.336700187178622e-47 ], [ 130.6875 , 8.331830429550854e-47 ], 
                    [ 130.75 , 8.326963516518644e-47 ], [ 130.8125 , 8.322131216989975e-47 ], [ 130.875 , 8.317307972460531e-47 ], [ 130.9375 , 8.31248752333123e-47 ], 
                    [ 131.0 , 8.307669867982126e-47 ], [ 131.0625 , 8.302858537193628e-47 ], [ 131.125 , 8.298084244619129e-47 ], [ 131.1875 , 8.293312697348489e-47 ], 
                    [ 131.25 , 8.288543893803052e-47 ], [ 131.3125 , 8.283777832405418e-47 ], [ 131.375 , 8.279027308344025e-47 ], [ 131.4375 , 8.274304282141788e-47 ], 
                    [ 131.5 , 8.269583950335366e-47 ], [ 131.5625 , 8.264866311387652e-47 ], [ 131.625 , 8.26015136376231e-47 ], [ 131.6875 , 8.255460747275508e-47 ], 
                    [ 131.75 , 8.250788663754105e-47 ], [ 131.8125 , 8.246119224344837e-47 ], [ 131.875 , 8.241452427551359e-47 ], [ 131.9375 , 8.236788271878054e-47 ], 
                    [ 132.0 , 8.232156375960848e-47 ], [ 132.0625 , 8.227534382317288e-47 ], [ 132.125 , 8.22291498371982e-47 ], [ 132.1875 , 8.218298178711559e-47 ], 
                    [ 132.25 , 8.213684589588762e-47 ], [ 132.3125 , 8.209109724345559e-47 ], [ 132.375 , 8.204537407214925e-47 ], [ 132.4375 , 8.199967636777379e-47 ], 
                    [ 132.5 , 8.195400411614636e-47 ], [ 132.5625 , 8.190843861991305e-47 ], [ 132.625 , 8.186318526456629e-47 ], [ 132.6875 , 8.18179569111138e-47 ], 
                    [ 132.75 , 8.177275354574295e-47 ], [ 132.8125 , 8.172757515464929e-47 ], [ 132.875 , 8.168257159304164e-47 ], [ 132.9375 , 8.163780517978883e-47 ], 
                    [ 133.0 , 8.159306330092395e-47 ], [ 133.0625 , 8.154834594299736e-47 ], [ 133.125 , 8.1503653092572e-47 ], [ 133.1875 , 8.145919974999982e-47 ], 
                    [ 133.25 , 8.141491651508412e-47 ], [ 133.3125 , 8.137065735363089e-47 ], [ 133.375 , 8.132642225255149e-47 ], [ 133.4375 , 8.128221119876665e-47 ], 
                    [ 133.5 , 8.123829916433326e-47 ], [ 133.5625 , 8.119449439938258e-47 ], [ 133.625 , 8.115071325453857e-47 ], [ 133.6875 , 8.110695571706784e-47 ], 
                    [ 133.75 , 8.106322177423983e-47 ], [ 133.8125 , 8.101984224549851e-47 ], [ 133.875 , 8.097651195384551e-47 ], [ 133.9375 , 8.093320483570199e-47 ], 
                    [ 134.0 , 8.088992087867794e-47 ], [ 134.0625 , 8.08466871215229e-47 ], [ 134.125 , 8.0803803690051e-47 ], [ 134.1875 , 8.076094300519694e-47 ], 
                    [ 134.25 , 8.071810505489583e-47 ], [ 134.3125 , 8.067528982708685e-47 ], [ 134.375 , 8.063257284872772e-47 ], [ 134.4375 , 8.059015486860123e-47 ], 
                    [ 134.5 , 8.054775920309244e-47 ], [ 134.5625 , 8.050538584046243e-47 ], [ 134.625 , 8.046303476897898e-47 ], [ 134.6875 , 8.042082516508506e-47 ], 
                    [ 134.75 , 8.037886714372629e-47 ], [ 134.8125 , 8.033693101315951e-47 ], [ 134.875 , 8.029501676196365e-47 ], [ 134.9375 , 8.025312437872354e-47 ], 
                    [ 135.0 , 8.021141385584635e-47 ], [ 135.0625 , 8.016991432397259e-47 ], [ 135.125 , 8.012843626299644e-47 ], [ 135.1875 , 8.008697966181164e-47 ], 
                    [ 135.25 , 8.004554450931481e-47 ], [ 135.3125 , 8.000432423689149e-47 ], [ 135.375 , 7.996327533733306e-47 ], [ 135.4375 , 7.992224749928927e-47 ], 
                    [ 135.5 , 7.988124071195209e-47 ], [ 135.5625 , 7.984025496452132e-47 ], [ 135.625 , 7.979951517945534e-47 ], [ 135.6875 , 7.975891357459733e-47 ], 
                    [ 135.75 , 7.971833262763878e-47 ], [ 135.8125 , 7.967777232806793e-47 ], [ 135.875 , 7.963723266538119e-47 ], [ 135.9375 , 7.959696625624364e-47 ], 
                    [ 136.0 , 7.955680899672543e-47 ], [ 136.0625 , 7.95166719968423e-47 ], [ 136.125 , 7.94765552463726e-47 ], [ 136.1875 , 7.943645873510201e-47 ], 
                    [ 136.25 , 7.939665692257311e-47 ], [ 136.3125 , 7.935693876814074e-47 ], [ 136.375 , 7.93172404827041e-47 ], [ 136.4375 , 7.927756205632199e-47 ], 
                    [ 136.5 , 7.923790347906224e-47 ], [ 136.5625 , 7.919855574317745e-47 ], [ 136.625 , 7.915927024127705e-47 ], [ 136.6875 , 7.912000422648275e-47 ], 
                    [ 136.75 , 7.908075768912934e-47 ], [ 136.8125 , 7.904153061955189e-47 ], [ 136.875 , 7.90026337468455e-47 ], [ 136.9375 , 7.896378249549166e-47 ], 
                    [ 137.0 , 7.892495035008087e-47 ], [ 137.0625 , 7.888613730121566e-47 ], [ 137.125 , 7.884734333950493e-47 ], [ 137.1875 , 7.880888442327253e-47 ], 
                    [ 137.25 , 7.877045758942004e-47 ], [ 137.3125 , 7.873204949230676e-47 ], [ 137.375 , 7.869366012279902e-47 ], [ 137.4375 , 7.865528947176248e-47 ], 
                    [ 137.5 , 7.861726286708831e-47 ], [ 137.5625 , 7.857925913014561e-47 ], [ 137.625 , 7.854127376428342e-47 ], [ 137.6875 , 7.850330676062001e-47 ], 
                    [ 137.75 , 7.846535879066456e-47 ], [ 137.8125 , 7.842775543457665e-47 ], [ 137.875 , 7.839017009933836e-47 ], [ 137.9375 , 7.835260277631351e-47 ], 
                    [ 138.0 , 7.83150534568683e-47 ], [ 138.0625 , 7.827752421883396e-47 ], [ 138.125 , 7.824033673300315e-47 ], [ 138.1875 , 7.820316691391901e-47 ], 
                    [ 138.25 , 7.816601475318687e-47 ], [ 138.3125 , 7.812888024241831e-47 ], [ 138.375 , 7.809176337322774e-47 ], [ 138.4375 , 7.805498614967151e-47 ], 
                    [ 138.5 , 7.801822645242798e-47 ], [ 138.5625 , 7.798148406702275e-47 ], [ 138.625 , 7.794475898530341e-47 ], [ 138.6875 , 7.790805119912089e-47 ], 
                    [ 138.75 , 7.787167629138375e-47 ], [ 138.8125 , 7.783532443711724e-47 ], [ 138.875 , 7.779898955252919e-47 ], [ 138.9375 , 7.776267162969677e-47 ], 
                    [ 139.0 , 7.772637066070358e-47 ], [ 139.0625 , 7.769038965709863e-47 ], [ 139.125 , 7.765444079505681e-47 ], [ 139.1875 , 7.761850856725567e-47 ], 
                    [ 139.25 , 7.75825929659982e-47 ], [ 139.3125 , 7.754669398359374e-47 ], [ 139.375 , 7.751110015606867e-47 ], [ 139.4375 , 7.74755511067483e-47 ], 
                    [ 139.5 , 7.744001836134853e-47 ], [ 139.5625 , 7.740450191239351e-47 ], [ 139.625 , 7.736900175240973e-47 ], [ 139.6875 , 7.73337876739471e-47 ], 
                    [ 139.75 , 7.729863440156493e-47 ], [ 139.8125 , 7.726349710864784e-47 ], [ 139.875 , 7.722837578793265e-47 ], [ 139.9375 , 7.719327043215731e-47 ], 
                    [ 140.0 , 7.71584279008148e-47 ], [ 140.0625 , 7.712366527267071e-47 ], [ 140.125 , 7.70889183063308e-47 ], [ 140.1875 , 7.705418699474215e-47 ], 
                    [ 140.25 , 7.701947133084955e-47 ], [ 140.3125 , 7.698499458029478e-47 ], [ 140.375 , 7.695062081192287e-47 ], [ 140.4375 , 7.691626239142439e-47 ], 
                    [ 140.5 , 7.688191931194704e-47 ], [ 140.5625 , 7.684759156664054e-47 ], [ 140.625 , 7.681347460023937e-47 ], [ 140.6875 , 7.677948692902259e-47 ], 
                    [ 140.75 , 7.674551429633594e-47 ], [ 140.8125 , 7.671155669552533e-47 ], [ 140.875 , 7.6677614119941245e-47 ], [ 140.9375 , 7.66438496288625e-47 ], 
                    [ 141.0 , 7.661024302371664e-47 ], [ 141.0625 , 7.657665115431168e-47 ], [ 141.125 , 7.654307401418633e-47 ], [ 141.1875 , 7.650951159688158e-47 ], 
                    [ 141.25 , 7.647609235630658e-47 ], [ 141.3125 , 7.644286318153893e-47 ], [ 141.375 , 7.640964844498302e-47 ], [ 141.4375 , 7.637644814036701e-47 ], 
                    [ 141.5 , 7.634326226141746e-47 ], [ 141.5625 , 7.631018176741258e-47 ], [ 141.625 , 7.627732680219843e-47 ], [ 141.6875 , 7.62444859825255e-47 ], 
                    [ 141.75 , 7.621165930230081e-47 ], [ 141.8125 , 7.617884675543992e-47 ], [ 141.875 , 7.614609884719776e-47 ], [ 141.9375 , 7.611361563459759e-47 ], 
                    [ 142.0 , 7.608114627903115e-47 ], [ 142.0625 , 7.604869077458713e-47 ], [ 142.125 , 7.601624911535843e-47 ], [ 142.1875 , 7.598382804298493e-47 ], 
                    [ 142.25 , 7.595171091458653e-47 ], [ 142.3125 , 7.591960736157707e-47 ], [ 142.375 , 7.58875173782163e-47 ], [ 142.4375 , 7.585544095877122e-47 ], 
                    [ 142.5 , 7.58233780975059e-47 ], [ 142.5625 , 7.579158624993301e-47 ], [ 142.625 , 7.575984747145907e-47 ], [ 142.6875 , 7.572812198403958e-47 ], 
                    [ 142.75 , 7.569640978210713e-47 ], [ 142.8125 , 7.566471086009982e-47 ], [ 142.875 , 7.563323063399866e-47 ], [ 142.9375 , 7.560185238079185e-47 ], 
                    [ 143.0 , 7.557048714560272e-47 ], [ 143.0625 , 7.553913492303205e-47 ], [ 143.125 , 7.550779570767962e-47 ], [ 143.1875 , 7.547662143060813e-47 ], 
                    [ 143.25 , 7.544560138856458e-47 ], [ 143.3125 , 7.541459409540924e-47 ], [ 143.375 , 7.538359954590245e-47 ], [ 143.4375 , 7.535261773480618e-47 ], 
                    [ 143.5 , 7.532174261315805e-47 ], [ 143.5625 , 7.529107410433223e-47 ], [ 143.625 , 7.526041808270333e-47 ], [ 143.6875 , 7.522977454318646e-47 ], 
                    [ 143.75 , 7.519914348070092e-47 ], [ 143.8125 , 7.516856060476824e-47 ], [ 143.875 , 7.51382449950387e-47 ], [ 143.9375 , 7.510794161164672e-47 ], 
                    [ 144.0 , 7.507765044966135e-47 ], [ 144.0625 , 7.504737150415422e-47 ], [ 144.125 , 7.501710477019733e-47 ], [ 144.1875 , 7.498710995162592e-47 ], 
                    [ 144.25 , 7.495715358102721e-47 ], [ 144.3125 , 7.492720917760768e-47 ], [ 144.375 , 7.489727673658603e-47 ], [ 144.4375 , 7.486735625318399e-47 ], 
                    [ 144.5 , 7.483764215795788e-47 ], [ 144.5625 , 7.480803111718051e-47 ], [ 144.625 , 7.477843179261797e-47 ], [ 144.6875 , 7.474884417963137e-47 ], 
                    [ 144.75 , 7.471926827358781e-47 ], [ 144.8125 , 7.468982840744073e-47 ], [ 144.875 , 7.466055740512462e-47 ], [ 144.9375 , 7.463129787413778e-47 ], 
                    [ 145.0 , 7.46020498099841e-47 ], [ 145.0625 , 7.457281320816967e-47 ], [ 145.125 , 7.454364194367768e-47 ], [ 145.1875 , 7.451471062751575e-47 ], 
                    [ 145.25 , 7.448579053995751e-47 ], [ 145.3125 , 7.445688167664506e-47 ], [ 145.375 , 7.442798403322315e-47 ], [ 145.4375 , 7.439909760533721e-47 ], 
                    [ 145.5 , 7.437047993183396e-47 ], [ 145.5625 , 7.434189324029537e-47 ], [ 145.625 , 7.431331753697219e-47 ], [ 145.6875 , 7.428475281764183e-47 ], 
                    [ 145.75 , 7.42561990780822e-47 ], [ 145.8125 , 7.422783827647687e-47 ], [ 145.875 , 7.419958484427474e-47 ], [ 145.9375 , 7.417134216621135e-47 ], 
                    [ 146.0 , 7.414311023819279e-47 ], [ 146.0625 , 7.411488905612621e-47 ], [ 146.125 , 7.408677922995367e-47 ], [ 146.1875 , 7.405885374748481e-47 ], 
                    [ 146.25 , 7.40309387909492e-47 ], [ 146.3125 , 7.400303435638141e-47 ], [ 146.375 , 7.397514043981331e-47 ], [ 146.4375 , 7.394727567469475e-47 ], 
                    [ 146.5 , 7.391967610890607e-47 ], [ 146.5625 , 7.389208684418749e-47 ], [ 146.625 , 7.386450787669379e-47 ], [ 146.6875 , 7.383693920258169e-47 ], 
                    [ 146.75 , 7.380938081800939e-47 ], [ 146.8125 , 7.378203972330093e-47 ], [ 146.875 , 7.37547749646892e-47 ], [ 146.9375 , 7.372752028125333e-47 ], 
                    [ 147.0 , 7.370027566927123e-47 ], [ 147.0625 , 7.367304112502016e-47 ], [ 147.125 , 7.364593323991824e-47 ], [ 147.1875 , 7.361898678726893e-47 ], 
                    [ 147.25 , 7.359205019410969e-47 ], [ 147.3125 , 7.356512345683201e-47 ], [ 147.375 , 7.353820657183071e-47 ], [ 147.4375 , 7.351132647481451e-47 ], 
                    [ 147.5 , 7.348469770981966e-47 ], [ 147.5625 , 7.345807859083664e-47 ], [ 147.625 , 7.343146911437023e-47 ], [ 147.6875 , 7.340486927693015e-47 ], 
                    [ 147.75 , 7.337827907502213e-47 ], [ 147.8125 , 7.335189983640936e-47 ], [ 147.875 , 7.332559594226173e-47 ], [ 147.9375 , 7.329930148065676e-47 ], 
                    [ 148.0 , 7.327301644820985e-47 ], [ 148.0625 , 7.324674084154234e-47 ], [ 148.125 , 7.32205789569469e-47 ], [ 148.1875 , 7.319458640443488e-47 ], 
                    [ 148.25 , 7.31686030790122e-47 ], [ 148.3125 , 7.314262897740283e-47 ], [ 148.375 , 7.311666409633294e-47 ], [ 148.4375 , 7.309071460342532e-47 ], 
                    [ 148.5 , 7.306503042683446e-47 ], [ 148.5625 , 7.303935527569875e-47 ], [ 148.625 , 7.301368914684608e-47 ], [ 148.6875 , 7.298803203710603e-47 ], 
                    [ 148.75 , 7.296238394330873e-47 ], [ 148.8125 , 7.293691177368235e-47 ], [ 148.875 , 7.291154222150069e-47 ], [ 148.9375 , 7.288618149357772e-47 ], 
                    [ 149.0 , 7.286082958684411e-47 ], [ 149.0625 , 7.283548649823107e-47 ], [ 149.125 , 7.28102160424354e-47 ], [ 149.1875 , 7.278515116204509e-47 ], 
                    [ 149.25 , 7.276009491022728e-47 ], [ 149.3125 , 7.273504728401264e-47 ], [ 149.375 , 7.271000828042975e-47 ], [ 149.4375 , 7.268497789651077e-47 ], 
                    [ 149.5 , 7.266017020376662e-47 ], [ 149.5625 , 7.263541261672126e-47 ], [ 149.625 , 7.261066346535723e-47 ], [ 149.6875 , 7.258592274680176e-47 ], 
                    [ 149.75 , 7.256119045818106e-47 ], [ 149.8125 , 7.253657377591196e-47 ], [ 149.875 , 7.251211506736009e-47 ], [ 149.9375 , 7.248766460607432e-47 ], 
                    [ 150.0 , 7.246322238927318e-47 ], [ 150.0625 , 7.243878841417827e-47 ], [ 150.125 , 7.241436267800901e-47 ], [ 150.1875 , 7.239019352743496e-47 ], 
                    [ 150.25 , 7.236603619361571e-47 ], [ 150.3125 , 7.234188692134009e-47 ], [ 150.375 , 7.231774570791991e-47 ], [ 150.4375 , 7.229361255066383e-47 ], 
                    [ 150.5 , 7.226962512056769e-47 ], [ 150.5625 , 7.224576258680616e-47 ], [ 150.625 , 7.222190793215676e-47 ], [ 150.6875 , 7.219806115401946e-47 ], 
                    [ 150.75 , 7.217422224978995e-47 ], [ 150.8125 , 7.21504121263233e-47 ], [ 150.875 , 7.212683874710401e-47 ], [ 150.9375 , 7.210327306990924e-47 ], 
                    [ 151.0 , 7.207971509222103e-47 ], [ 151.0625 , 7.205616481152529e-47 ], [ 151.125 , 7.203262222530568e-47 ], [ 151.1875 , 7.200924004215181e-47 ], 
                    [ 151.25 , 7.198596119546977e-47 ], [ 151.3125 , 7.196268987427614e-47 ], [ 151.375 , 7.193942607613918e-47 ], [ 151.4375 , 7.191616979862478e-47 ], 
                    [ 151.5 , 7.189295379577311e-47 ], [ 151.5625 , 7.186995912253621e-47 ], [ 151.625 , 7.184697180405395e-47 ], [ 151.6875 , 7.182399183797288e-47 ], 
                    [ 151.75 , 7.180101922194292e-47 ], [ 151.8125 , 7.177805395361063e-47 ], [ 151.875 , 7.175525387458845e-47 ], [ 151.9375 , 7.173255022934291e-47 ], 
                    [ 152.0 , 7.170985376762023e-47 ], [ 152.0625 , 7.168716448714804e-47 ], [ 152.125 , 7.166448238565366e-47 ], [ 152.1875 , 7.164183954655925e-47 ], 
                    [ 152.25 , 7.161941419539851e-47 ], [ 152.3125 , 7.159699586382826e-47 ], [ 152.375 , 7.157458454965172e-47 ], [ 152.4375 , 7.155218025067128e-47 ], 
                    [ 152.5 , 7.152978296469207e-47 ], [ 152.5625 , 7.150754155682149e-47 ], [ 152.625 , 7.148540090787985e-47 ], [ 152.6875 , 7.146326711427578e-47 ], 
                    [ 152.75 , 7.144114017388617e-47 ], [ 152.8125 , 7.141902008458958e-47 ], [ 152.875 , 7.1396926531943e-47 ], [ 152.9375 , 7.137506038648898e-47 ], 
                    [ 153.0 , 7.135320093779826e-47 ], [ 153.0625 , 7.133134818381987e-47 ], [ 153.125 , 7.130950212250398e-47 ], [ 153.1875 , 7.128766275179888e-47 ], 
                    [ 153.25 , 7.1265958389883975e-47 ], [ 153.3125 , 7.124437083141582e-47 ], [ 153.375 , 7.122278981215171e-47 ], [ 153.4375 , 7.120121533011287e-47 ], 
                    [ 153.5 , 7.117964738331602e-47 ], [ 153.5625 , 7.115808596978408e-47 ], [ 153.625 , 7.113676549074185e-47 ], [ 153.6875 , 7.111545587017662e-47 ], 
                    [ 153.75 , 7.109415263308818e-47 ], [ 153.8125 , 7.107285577756534e-47 ], [ 153.875 , 7.105156530169543e-47 ], [ 153.9375 , 7.103037743052903e-47 ], 
                    [ 154.0 , 7.100933317056946e-47 ], [ 154.0625 , 7.098829514542046e-47 ], [ 154.125 , 7.096726335323284e-47 ], [ 154.1875 , 7.094623779216145e-47 ], 
                    [ 154.25 , 7.092521846035969e-47 ], [ 154.3125 , 7.090440021031941e-47 ], [ 154.375 , 7.088362790136478e-47 ], [ 154.4375 , 7.086286167791295e-47 ], 
                    [ 154.5 , 7.0842101538179595e-47 ], [ 154.5625 , 7.082134748038241e-47 ], [ 154.625 , 7.080065455338883e-47 ], [ 154.6875 , 7.078014599380097e-47 ], 
                    [ 154.75 , 7.075964337485101e-47 ], [ 154.8125 , 7.073914669481766e-47 ], [ 154.875 , 7.071865595198109e-47 ], [ 154.9375 , 7.069817114462001e-47 ], 
                    [ 155.0 , 7.067783643339067e-47 ], [ 155.0625 , 7.06575927084585e-47 ], [ 155.125 , 7.063735478178632e-47 ], [ 155.1875 , 7.0617122651713355e-47 ], 
                    [ 155.25 , 7.059689631657533e-47 ], [ 155.3125 , 7.057667780468212e-47 ], [ 155.375 , 7.055669366598017e-47 ], [ 155.4375 , 7.053671518588687e-47 ], 
                    [ 155.5 , 7.051674236279947e-47 ], [ 155.5625 , 7.049677519511714e-47 ], [ 155.625 , 7.047681368123803e-47 ], [ 155.6875 , 7.045694306596426e-47 ], 
                    [ 155.75 , 7.043721938980422e-47 ], [ 155.8125 , 7.0417501235079755e-47 ], [ 155.875 , 7.039778860024419e-47 ], [ 155.9375 , 7.037808148375128e-47 ], 
                    [ 156.0 , 7.035837988405876e-47 ], [ 156.0625 , 7.033884872875602e-47 ], [ 156.125 , 7.031938370761252e-47 ], [ 156.1875 , 7.029992407306576e-47 ], 
                    [ 156.25 , 7.028046982362509e-47 ], [ 156.3125 , 7.026102095780177e-47 ], [ 156.375 , 7.024159382931263e-47 ], [ 156.4375 , 7.022238081411009e-47 ], 
                    [ 156.5 , 7.020317305419764e-47 ], [ 156.5625 , 7.018397054813784e-47 ], [ 156.625 , 7.01647732944946e-47 ], [ 156.6875 , 7.014558129182975e-47 ], 
                    [ 156.75 , 7.012648418911637e-47 ], [ 156.8125 , 7.010752574323004e-47 ], [ 156.875 , 7.008857242269222e-47 ], [ 156.9375 , 7.006962422611727e-47 ], 
                    [ 157.0 , 7.005068115212045e-47 ], [ 157.0625 , 7.00317431993159e-47 ], [ 157.125 , 7.001296981922432e-47 ], [ 157.1875 , 6.999426410610703e-47 ], 
                    [ 157.25 , 6.997556339068758e-47 ], [ 157.3125 , 6.995686767163122e-47 ], [ 157.375 , 6.993817694760253e-47 ], [ 157.4375 , 6.991949583332166e-47 ], 
                    [ 157.5 , 6.990103359918638e-47 ], [ 157.5625 , 6.988257624000122e-47 ], [ 157.625 , 6.986412375447895e-47 ], [ 157.6875 , 6.984567614133367e-47 ], 
                    [ 157.75 , 6.982723339927784e-47 ], [ 157.8125 , 6.980886356960614e-47 ], [ 157.875 , 6.979064891190338e-47 ], [ 157.9375 , 6.977243900680179e-47 ], 
                    [ 158.0 , 6.975423385306228e-47 ], [ 158.0625 , 6.973603344944564e-47 ], [ 158.125 , 6.971783779471243e-47 ], [ 158.1875 , 6.969977493074188e-47 ], 
                    [ 158.25 , 6.968180608591966e-47 ], [ 158.3125 , 6.966384187352921e-47 ], [ 158.375 , 6.964588229237481e-47 ], [ 158.4375 , 6.962792734126253e-47 ], 
                    [ 158.5 , 6.960997701900067e-47 ], [ 158.5625 , 6.959221528798869e-47 ], [ 158.625 , 6.957448967285939e-47 ], [ 158.6875 , 6.955676857256597e-47 ], 
                    [ 158.75 , 6.95390519859585e-47 ], [ 158.8125 , 6.952133991188828e-47 ], [ 158.875 , 6.950365471215344e-47 ], [ 158.9375 , 6.94861646617559e-47 ], 
                    [ 159.0 , 6.946867901259305e-47 ], [ 159.0625 , 6.945119776355639e-47 ], [ 159.125 , 6.943372091353915e-47 ], [ 159.1875 , 6.941624846143385e-47 ], 
                    [ 159.25 , 6.939885368665101e-47 ], [ 159.3125 , 6.938160337716822e-47 ], [ 159.375 , 6.93643573555537e-47 ], [ 159.4375 , 6.934711562074163e-47 ], 
                    [ 159.5 , 6.932987817166743e-47 ], [ 159.5625 , 6.931264500726533e-47 ], [ 159.625 , 6.929553549846531e-47 ], [ 159.6875 , 6.927852074428581e-47 ], 
                    [ 159.75 , 6.926151016789098e-47 ], [ 159.8125 , 6.9244503768256e-47 ], [ 159.875 , 6.922750154435482e-47 ], [ 159.9375 , 6.921050349516211e-47 ], 
                    [ 160.0 , 6.919367408777988e-47 ], [ 160.0625 , 6.917689489078313e-47 ], [ 160.125 , 6.916011976267643e-47 ], [ 160.1875 , 6.91433487024731e-47 ], 
                    [ 160.25 , 6.912658170918667e-47 ], [ 160.3125 , 6.910981878183045e-47 ], [ 160.375 , 6.909326268398082e-47 ], [ 160.4375 , 6.907671465564648e-47 ], 
                    [ 160.5 , 6.90601705906104e-47 ], [ 160.5625 , 6.904363048792338e-47 ], [ 160.625 , 6.902709434663736e-47 ], [ 160.6875 , 6.901059647121734e-47 ], 
                    [ 160.75 , 6.899427475213227e-47 ], [ 160.8125 , 6.897795689330289e-47 ], [ 160.875 , 6.896164289381521e-47 ], [ 160.9375 , 6.894533275275747e-47 ], 
                    [ 161.0 , 6.892902646921563e-47 ], [ 161.0625 , 6.891279385793226e-47 ], [ 161.125 , 6.889670114346986e-47 ], [ 161.1875 , 6.888061218702458e-47 ], 
                    [ 161.25 , 6.886452698771834e-47 ], [ 161.3125 , 6.884844554467473e-47 ], [ 161.375 , 6.883236785701416e-47 ], [ 161.4375 , 6.881639516266798e-47 ], 
                    [ 161.5 , 6.880052797496235e-47 ], [ 161.5625 , 6.878466444579836e-47 ], [ 161.625 , 6.876880457433242e-47 ], [ 161.6875 , 6.875294835972117e-47 ], 
                    [ 161.75 , 6.873709580112193e-47 ], [ 161.8125 , 6.872137702454571e-47 ], [ 161.875 , 6.870573345897847e-47 ], [ 161.9375 , 6.869009345447359e-47 ], 
                    [ 162.0 , 6.867445701021997e-47 ], [ 162.0625 , 6.865882412540959e-47 ], [ 162.125 , 6.864319479922929e-47 ], [ 162.1875 , 6.862772545124129e-47 ], 
                    [ 162.25 , 6.861230419283559e-47 ], [ 162.3125 , 6.859688639972388e-47 ], [ 162.375 , 6.858147207112698e-47 ], [ 162.4375 , 6.856606120626543e-47 ], 
                    [ 162.5 , 6.855065380436331e-47 ], [ 162.5625 , 6.853542877847722e-47 ], [ 162.625 , 6.852022746937742e-47 ], [ 162.6875 , 6.850502953196243e-47 ], 
                    [ 162.75 , 6.848983496548346e-47 ], [ 162.8125 , 6.847464376919134e-47 ], [ 162.875 , 6.8459456829884025e-47 ], [ 162.9375 , 6.844447043998905e-47 ], 
                    [ 163.0 , 6.842948733075038e-47 ], [ 163.0625 , 6.841450750144791e-47 ], [ 163.125 , 6.83995309513641e-47 ], [ 163.1875 , 6.83845576797821e-47 ], 
                    [ 163.25 , 6.836960688235473e-47 ], [ 163.3125 , 6.835483680279437e-47 ], [ 163.375 , 6.834006991405609e-47 ], [ 163.4375 , 6.832530621545199e-47 ], 
                    [ 163.5 , 6.831054570629049e-47 ], [ 163.5625 , 6.829578838588354e-47 ], [ 163.625 , 6.828106882080011e-47 ], [ 163.6875 , 6.826651281336357e-47 ], 
                    [ 163.75 , 6.825195990894552e-47 ], [ 163.8125 , 6.82374101068825e-47 ], [ 163.875 , 6.822286340651367e-47 ], [ 163.9375 , 6.820831980717927e-47 ], 
                    [ 164.0 , 6.819382662861563e-47 ], [ 164.0625 , 6.817948336896018e-47 ], [ 164.125 , 6.816514312613493e-47 ], [ 164.1875 , 6.815080589950294e-47 ], 
                    [ 164.25 , 6.81364716884308e-47 ], [ 164.3125 , 6.81221404922842e-47 ], [ 164.375 , 6.810786918934477e-47 ], [ 164.4375 , 6.809373581657646e-47 ], 
                    [ 164.5 , 6.807960537668841e-47 ], [ 164.5625 , 6.806547786907202e-47 ], [ 164.625 , 6.805135329311879e-47 ], [ 164.6875 , 6.803723164821989e-47 ], 
                    [ 164.75 , 6.802317730505662e-47 ], [ 164.8125 , 6.800925304890445e-47 ], [ 164.875 , 6.799533164302847e-47 ], [ 164.9375 , 6.79814130868472e-47 ], 
                    [ 165.0 , 6.796749737977584e-47 ], [ 165.0625 , 6.795358452123214e-47 ], [ 165.125 , 6.793974294878483e-47 ], [ 165.1875 , 6.792602491426118e-47 ], 
                    [ 165.25 , 6.791230964961022e-47 ], [ 165.3125 , 6.789859715427412e-47 ], [ 165.375 , 6.788488742769323e-47 ], [ 165.4375 , 6.787118046930704e-47 ], 
                    [ 165.5 , 6.785754695014336e-47 ], [ 165.5625 , 6.784403431471001e-47 ], [ 165.625 , 6.783052437008051e-47 ], [ 165.6875 , 6.781701711571904e-47 ], 
                    [ 165.75 , 6.780351255108846e-47 ], [ 165.8125 , 6.779001067565455e-47 ], [ 165.875 , 6.77765811605738e-47 ], [ 165.9375 , 6.776327141861845e-47 ], 
                    [ 166.0 , 6.774996429038593e-47 ], [ 166.0625 , 6.773665977536299e-47 ], [ 166.125 , 6.772335787303836e-47 ], [ 166.1875 , 6.771005858289704e-47 ], 
                    [ 166.25 , 6.769682844240868e-47 ], [ 166.3125 , 6.768372014464924e-47 ], [ 166.375 , 6.767061438508066e-47 ], [ 166.4375 , 6.765751116321097e-47 ], 
                    [ 166.5 , 6.764441047854973e-47 ], [ 166.5625 , 6.763131233060519e-47 ], [ 166.625 , 6.761827740421242e-47 ], [ 166.6875 , 6.760536839926077e-47 ], 
                    [ 166.75 , 6.759246185876668e-47 ], [ 166.8125 , 6.757955778226112e-47 ], [ 166.875 , 6.756665616927179e-47 ], [ 166.9375 , 6.755375701933026e-47 ], 
                    [ 167.0 , 6.754091302000704e-47 ], [ 167.0625 , 6.752820200003006e-47 ], [ 167.125 , 6.751549337223464e-47 ], [ 167.1875 , 6.750278713616767e-47 ], 
                    [ 167.25 , 6.749008329138147e-47 ], [ 167.3125 , 6.747738183742454e-47 ], [ 167.375 , 6.746472482366662e-47 ], [ 167.4375 , 6.745220960781035e-47 ], 
                    [ 167.5 , 6.743969671362015e-47 ], [ 167.5625 , 6.742718614066726e-47 ], [ 167.625 , 6.741467788852108e-47 ], [ 167.6875 , 6.74021719567482e-47 ], 
                    [ 167.75 , 6.738969766920735e-47 ], [ 167.8125 , 6.737737708455109e-47 ], [ 167.875 , 6.736505875241726e-47 ], [ 167.9375 , 6.73527426723931e-47 ], 
                    [ 168.0 , 6.734042884406924e-47 ], [ 168.0625 , 6.732811726703164e-47 ], [ 168.125 , 6.731582205618089e-47 ], [ 168.1875 , 6.730369389208437e-47 ], 
                    [ 168.25 , 6.729156791309666e-47 ], [ 168.3125 , 6.727944411882409e-47 ], [ 168.375 , 6.726732250887113e-47 ], [ 168.4375 , 6.725520308284662e-47 ], 
                    [ 168.5 , 6.724308584035564e-47 ], [ 168.5625 , 6.723114543822276e-47 ], [ 168.625 , 6.721921040071836e-47 ], [ 168.6875 , 6.720727748195246e-47 ], 
                    [ 168.75 , 6.719534668154561e-47 ], [ 168.8125 , 6.718341799912556e-47 ], [ 168.875 , 6.717149143431297e-47 ], [ 168.9375 , 6.715972077820235e-47 ], 
                    [ 169.0 , 6.714797498244354e-47 ], [ 169.0625 , 6.713623124094707e-47 ], [ 169.125 , 6.71244895533556e-47 ], [ 169.1875 , 6.711274991930655e-47 ], 
                    [ 169.25 , 6.710101233844317e-47 ], [ 169.3125 , 6.708940761896828e-47 ], [ 169.375 , 6.707784941805235e-47 ], [ 169.4375 , 6.706629320839025e-47 ], 
                    [ 169.5 , 6.705473898963702e-47 ], [ 169.5625 , 6.704318676145254e-47 ], [ 169.625 , 6.703163652349196e-47 ], [ 169.6875 , 6.702019408059173e-47 ], 
                    [ 169.75 , 6.700882193335329e-47 ], [ 169.8125 , 6.699745171576758e-47 ], [ 169.875 , 6.698608342750765e-47 ], [ 169.9375 , 6.697471706824614e-47 ], 
                    [ 170.0 , 6.69633526376562e-47 ], [ 170.0625 , 6.695206871492807e-47 ], [ 170.125 , 6.694088084617533e-47 ], [ 170.1875 , 6.692969484694521e-47 ], 
                    [ 170.25 , 6.691851071692531e-47 ], [ 170.3125 , 6.690732845580329e-47 ], [ 170.375 , 6.689614806326732e-47 ], [ 170.4375 , 6.688501894957154e-47 ], 
                    [ 170.5 , 6.687401363071464e-47 ], [ 170.5625 , 6.6863010122681935e-47 ], [ 170.625 , 6.685200842517644e-47 ], [ 170.6875 , 6.684100853790023e-47 ], 
                    [ 170.75 , 6.683001046055404e-47 ], [ 170.8125 , 6.681903256520797e-47 ], [ 170.875 , 6.680820844291413e-47 ], [ 170.9375 , 6.679738607403817e-47 ], 
                    [ 171.0 , 6.678656545829368e-47 ], [ 171.0625 , 6.677574659540002e-47 ], [ 171.125 , 6.676492948507034e-47 ], [ 171.1875 , 6.675411412702125e-47 ], 
                    [ 171.25 , 6.674345508775471e-47 ], [ 171.3125 , 6.673281239532089e-47 ], [ 171.375 , 6.672217139993543e-47 ], [ 171.4375 , 6.671153210132867e-47 ], 
                    [ 171.5 , 6.670089449922908e-47 ], [ 171.5625 , 6.669025859336665e-47 ], [ 171.625 , 6.667974270610876e-47 ], [ 171.6875 , 6.666927802778698e-47 ], 
                    [ 171.75 , 6.665881499178543e-47 ], [ 171.8125 , 6.664835359784596e-47 ], [ 171.875 , 6.663789384571175e-47 ], [ 171.9375 , 6.66274357351247e-47 ], 
                    [ 172.0 , 6.661705952065201e-47 ], [ 172.0625 , 6.660677107490314e-47 ], [ 172.125 , 6.659648421811837e-47 ], [ 172.1875 , 6.658619895005229e-47 ], 
                    [ 172.25 , 6.65759152704586e-47 ], [ 172.3125 , 6.656563317909338e-47 ], [ 172.375 , 6.6555393345854e-47 ], [ 172.4375 , 6.654527999365479e-47 ], 
                    [ 172.5 , 6.65351681782185e-47 ], [ 172.5625 , 6.652505789931162e-47 ], [ 172.625 , 6.65149491567021e-47 ], [ 172.6875 , 6.650484195015458e-47 ], 
                    [ 172.75 , 6.649473627943709e-47 ], [ 172.8125 , 6.648479550419917e-47 ], [ 172.875 , 6.647485708206902e-47 ], [ 172.9375 , 6.646492014557436e-47 ], 
                    [ 173.0 , 6.645498469449405e-47 ], [ 173.0625 , 6.644505072860747e-47 ], [ 173.125 , 6.643511824768977e-47 ], [ 173.1875 , 6.642530632501253e-47 ], 
                    [ 173.25 , 6.641553993954127e-47 ], [ 173.3125 , 6.640577499000391e-47 ], [ 173.375 , 6.639601147618788e-47 ], [ 173.4375 , 6.638624939788211e-47 ], 
                    [ 173.5 , 6.637648875487694e-47 ], [ 173.5625 , 6.6366802397725e-47 ], [ 173.625 , 6.635720633729302e-47 ], [ 173.6875 , 6.634761166436762e-47 ], 
                    [ 173.75 , 6.633801837874914e-47 ], [ 173.8125 , 6.632842648023462e-47 ], [ 173.875 , 6.631883596862539e-47 ], [ 173.9375 , 6.630927205423511e-47 ], 
                    [ 174.0 , 6.629984490664035e-47 ], [ 174.0625 , 6.629041909929679e-47 ], [ 174.125 , 6.628099463201527e-47 ], [ 174.1875 , 6.627157150460294e-47 ], 
                    [ 174.25 , 6.626214971687026e-47 ], [ 174.3125 , 6.625272926862864e-47 ], [ 174.375 , 6.624344549609058e-47 ], [ 174.4375 , 6.623418703899811e-47 ], 
                    [ 174.5 , 6.622492987590483e-47 ], [ 174.5625 , 6.6215674006631325e-47 ], [ 174.625 , 6.620641943099531e-47 ], [ 174.6875 , 6.619716614881837e-47 ], 
                    [ 174.75 , 6.618799775883587e-47 ], [ 174.8125 , 6.617890555291586e-47 ], [ 174.875 , 6.616981459598774e-47 ], [ 174.9375 , 6.616072488787802e-47 ], 
                    [ 175.0 , 6.61516364284161e-47 ], [ 175.0625 , 6.614254921743093e-47 ], [ 175.125 , 6.613349330678577e-47 ], [ 175.1875 , 6.612456587428018e-47 ], 
                    [ 175.25 , 6.611563964689846e-47 ], [ 175.3125 , 6.610671462447932e-47 ], [ 175.375 , 6.609779080685778e-47 ], [ 175.4375 , 6.608886819387214e-47 ], 
                    [ 175.5 , 6.607994678535881e-47 ], [ 175.5625 , 6.607115764490247e-47 ], [ 175.625 , 6.606239457529363e-47 ], [ 175.6875 , 6.605363266793618e-47 ], 
                    [ 175.75 , 6.604487192267788e-47 ], [ 175.8125 , 6.603611233936409e-47 ], [ 175.875 , 6.602735391784074e-47 ], [ 175.9375 , 6.60186703143606e-47 ], 
                    [ 176.0 , 6.60100691867528e-47 ], [ 176.0625 , 6.600146917972842e-47 ], [ 176.125 , 6.599287029314146e-47 ], [ 176.1875 , 6.598427252684454e-47 ], 
                    [ 176.25 , 6.597567588069311e-47 ], [ 176.3125 , 6.596709506192967e-47 ], [ 176.375 , 6.595865434823377e-47 ], [ 176.4375 , 6.595021471455723e-47 ], 
                    [ 176.5 , 6.594177616075906e-47 ], [ 176.5625 , 6.593333868670296e-47 ], [ 176.625 , 6.592490229225076e-47 ], [ 176.6875 , 6.591646697726341e-47 ], 
                    [ 176.75 , 6.590813982623369e-47 ], [ 176.8125 , 6.589985926600902e-47 ], [ 176.875 , 6.5891579746137e-47 ], [ 176.9375 , 6.5883301266486e-47 ], 
                    [ 177.0 , 6.587502382692627e-47 ], [ 177.0625 , 6.586674742732618e-47 ], [ 177.125 , 6.585851641495396e-47 ], [ 177.1875 , 6.585039336487031e-47 ], 
                    [ 177.25 , 6.584227131669114e-47 ], [ 177.3125 , 6.583415027029334e-47 ], [ 177.375 , 6.582603022555289e-47 ], [ 177.4375 , 6.581791118234576e-47 ], 
                    [ 177.5 , 6.580979314054984e-47 ], [ 177.5625 , 6.580180722968424e-47 ], [ 177.625 , 6.579384183806199e-47 ], [ 177.6875 , 6.578587741066183e-47 ], 
                    [ 177.75 , 6.577791394736423e-47 ], [ 177.8125 , 6.576995144805436e-47 ], [ 177.875 , 6.576198991261553e-47 ], [ 177.9375 , 6.575409388669216e-47 ], 
                    [ 178.0 , 6.57462834434782e-47 ], [ 178.0625 , 6.573847392800818e-47 ], [ 178.125 , 6.57306653401733e-47 ], [ 178.1875 , 6.57228576798629e-47 ], 
                    [ 178.25 , 6.571505094696727e-47 ], [ 178.3125 , 6.570724514137531e-47 ], [ 178.375 , 6.569958574085981e-47 ], [ 178.4375 , 6.569193012911895e-47 ], 
                    [ 178.5 , 6.568427540944447e-47 ], [ 178.5625 , 6.567662158173239e-47 ], [ 178.625 , 6.566896864587834e-47 ], [ 178.6875 , 6.566131660177934e-47 ], 
                    [ 178.75 , 6.565374087795324e-47 ], [ 178.8125 , 6.564623745397988e-47 ], [ 178.875 , 6.563873488755709e-47 ], [ 178.9375 , 6.563123317858637e-47 ], 
                    [ 179.0 , 6.562373232696787e-47 ], [ 179.0625 , 6.56162323326078e-47 ], [ 179.125 , 6.560873789047481e-47 ], [ 179.1875 , 6.560138558424817e-47 ], 
                    [ 179.25 , 6.559403410194284e-47 ], [ 179.3125 , 6.558668344346603e-47 ], [ 179.375 , 6.557933360872634e-47 ], [ 179.4375 , 6.55719845976296e-47 ], 
                    [ 179.5 , 6.556463641008399e-47 ], [ 179.5625 , 6.555736690100829e-47 ], [ 179.625 , 6.555016554407857e-47 ], [ 179.6875 , 6.554296497820382e-47 ], 
                    [ 179.75 , 6.553576520329994e-47 ], [ 179.8125 , 6.552856621927629e-47 ], [ 179.875 , 6.552136802604836e-47 ], [ 179.9375 , 6.551417425508924e-47 ], 
                    [ 180.0 , 6.55071212806583e-47 ], [ 180.0625 , 6.550006906552074e-47 ], [ 180.125 , 6.549301760959482e-47 ], [ 180.1875 , 6.548596691279694e-47 ], 
                    [ 180.25 , 6.547891697504816e-47 ], [ 180.3125 , 6.547186779626492e-47 ], [ 180.375 , 6.546489083431021e-47 ], [ 180.4375 , 6.545798615777266e-47 ], 
                    [ 180.5 , 6.545108220948224e-47 ], [ 180.5625 , 6.544417898936168e-47 ], [ 180.625 , 6.543727649733324e-47 ], [ 180.6875 , 6.543037473331966e-47 ], 
                    [ 180.75 , 6.542347369724558e-47 ], [ 180.8125 , 6.541670921420378e-47 ], [ 180.875 , 6.540995120737046e-47 ], [ 180.9375 , 6.540319389868732e-47 ], 
                    [ 181.0 , 6.53964372880813e-47 ], [ 181.0625 , 6.538968137548073e-47 ], [ 181.125 , 6.538292616081447e-47 ], [ 181.1875 , 6.537622851208227e-47 ], 
                    [ 181.25 , 6.536961543542213e-47 ], [ 181.3125 , 6.536300302770166e-47 ], [ 181.375 , 6.535639128885365e-47 ], [ 181.4375 , 6.534978021881184e-47 ], 
                    [ 181.5 , 6.534316981750533e-47 ], [ 181.5625 , 6.533656008486832e-47 ], [ 181.625 , 6.533006743665956e-47 ], [ 181.6875 , 6.532359861624656e-47 ], 
                    [ 181.75 , 6.531713043636004e-47 ], [ 181.8125 , 6.531066289693655e-47 ], [ 181.875 , 6.530419599791316e-47 ], [ 181.9375 , 6.529772973922553e-47 ], 
                    [ 182.0 , 6.529129863681503e-47 ], [ 182.0625 , 6.528497228241444e-47 ], [ 182.125 , 6.527864654100104e-47 ], [ 182.1875 , 6.527232141251637e-47 ], 
                    [ 182.25 , 6.526599689689964e-47 ], [ 182.3125 , 6.525967299409378e-47 ], [ 182.375 , 6.52533497040371e-47 ], [ 182.4375 , 6.524711630365132e-47 ], 
                    [ 182.5 , 6.524093175258068e-47 ], [ 182.5625 , 6.523474778772294e-47 ], [ 182.625 , 6.52285644090221e-47 ], [ 182.6875 , 6.522238161642164e-47 ], 
                    [ 182.75 , 6.52161994098679e-47 ], [ 182.8125 , 6.521002247155725e-47 ], [ 182.875 , 6.520397817171635e-47 ], [ 182.9375 , 6.519793443211954e-47 ], 
                    [ 183.0 , 6.519189125271538e-47 ], [ 183.0625 , 6.518584863345239e-47 ], [ 183.125 , 6.517980657427726e-47 ], [ 183.1875 , 6.517376507513993e-47 ], 
                    [ 183.25 , 6.516777883613736e-47 ], [ 183.3125 , 6.516187372875707e-47 ], [ 183.375 , 6.515596915646165e-47 ], [ 183.4375 , 6.515006511920265e-47 ], 
                    [ 183.5 , 6.514416161693153e-47 ], [ 183.5625 , 6.513825864959986e-47 ], [ 183.625 , 6.513235621715916e-47 ], [ 183.6875 , 6.512655668988256e-47 ], 
                    [ 183.75 , 6.5120790088751085e-47 ], [ 183.8125 , 6.511502399822251e-47 ], [ 183.875 , 6.510925841824837e-47 ], [ 183.9375 , 6.510349334878535e-47 ], 
                    [ 184.0 , 6.509772878978776e-47 ], [ 184.0625 , 6.50919780326186e-47 ], [ 184.125 , 6.508634803670543e-47 ], [ 184.1875 , 6.508071852774664e-47 ], 
                    [ 184.25 , 6.507508950570058e-47 ], [ 184.3125 , 6.506946097052558e-47 ], [ 184.375 , 6.506383292217863e-47 ], [ 184.4375 , 6.505820536061808e-47 ], 
                    [ 184.5 , 6.505263450979413e-47 ], [ 184.5625 , 6.504714035935974e-47 ], [ 184.625 , 6.50416466729453e-47 ], [ 184.6875 , 6.503615345051114e-47 ], 
                    [ 184.75 , 6.503066069201714e-47 ], [ 184.8125 , 6.502516839742643e-47 ], [ 184.875 , 6.501967656669707e-47 ], [ 184.9375 , 6.501428173516112e-47 ], 
                    [ 185.0 , 6.500892228960075e-47 ], [ 185.0625 , 6.500356328584525e-47 ], [ 185.125 , 6.499820472385914e-47 ], [ 185.1875 , 6.499284660360505e-47 ], 
                    [ 185.25 , 6.498748892504753e-47 ], [ 185.3125 , 6.498213547320812e-47 ], [ 185.375 , 6.497690936652005e-47 ], [ 185.4375 , 6.497168368013598e-47 ], 
                    [ 185.5 , 6.49664584140198e-47 ], [ 185.5625 , 6.496123356813955e-47 ], [ 185.625 , 6.495600914246143e-47 ], [ 185.6875 , 6.495078513695165e-47 ], 
                    [ 185.75 , 6.494560150956571e-47 ], [ 185.8125 , 6.494050786269814e-47 ], [ 185.875 , 6.493541461532196e-47 ], [ 185.9375 , 6.493032176740724e-47 ], 
                    [ 186.0 , 6.492522931892076e-47 ], [ 186.0625 , 6.492013726983214e-47 ], [ 186.125 , 6.491504562010961e-47 ], [ 186.1875 , 6.491002797439713e-47 ], 
                    [ 186.25 , 6.490506564334451e-47 ], [ 186.3125 , 6.490010369165946e-47 ], [ 186.375 , 6.489514211931204e-47 ], [ 186.4375 , 6.489018092627373e-47 ], 
                    [ 186.5 , 6.488522011251552e-47 ], [ 186.5625 , 6.488025967800798e-47 ], [ 186.625 , 6.487540451499176e-47 ], [ 186.6875 , 6.487057248964053e-47 ], 
                    [ 186.75 , 6.486574082418638e-47 ], [ 186.8125 , 6.486090951860253e-47 ], [ 186.875 , 6.485607857286307e-47 ], [ 186.9375 , 6.485124798693935e-47 ], 
                    [ 187.0 , 6.484642478223766e-47 ], [ 187.0625 , 6.484172173825863e-47 ], [ 187.125 , 6.483701903537248e-47 ], [ 187.1875 , 6.483231667355306e-47 ], 
                    [ 187.25 , 6.482761465277656e-47 ], [ 187.3125 , 6.482291297301825e-47 ], [ 187.375 , 6.481821163425341e-47 ], [ 187.4375 , 6.4813545012532e-47 ], 
                    [ 187.5 , 6.480896998777578e-47 ], [ 187.5625 , 6.480439528595903e-47 ], [ 187.625 , 6.479982090705895e-47 ], [ 187.6875 , 6.47952468510523e-47 ], 
                    [ 187.75 , 6.479067311791625e-47 ], [ 187.8125 , 6.478609970762988e-47 ], [ 187.875 , 6.478158615108179e-47 ], [ 187.9375 , 6.477713827059754e-47 ], 
                    [ 188.0 , 6.477269069550273e-47 ], [ 188.0625 , 6.476824342577775e-47 ], [ 188.125 , 6.476379646140026e-47 ], [ 188.1875 , 6.475934980234928e-47 ], 
                    [ 188.25 , 6.475490344860478e-47 ], [ 188.3125 , 6.475053968843093e-47 ], [ 188.375 , 6.474621782956218e-47 ], [ 188.4375 , 6.474189625916197e-47 ], 
                    [ 188.5 , 6.47375749772106e-47 ], [ 188.5625 , 6.473325398368927e-47 ], [ 188.625 , 6.472893327857778e-47 ], [ 188.6875 , 6.472461286185647e-47 ], 
                    [ 188.75 , 6.472039547860217e-47 ], [ 188.8125 , 6.471619850107392e-47 ], [ 188.875 , 6.471200179571103e-47 ], [ 188.9375 , 6.47078053624954e-47 ], 
                    [ 189.0 , 6.470360920140891e-47 ], [ 189.0625 , 6.469941331243484e-47 ], [ 189.125 , 6.469521769555418e-47 ], [ 189.1875 , 6.469114349707034e-47 ], 
                    [ 189.25 , 6.4687070433684125e-47 ], [ 189.3125 , 6.468299762674578e-47 ], [ 189.375 , 6.467892507623728e-47 ], [ 189.4375 , 6.467485278214251e-47 ], 
                    [ 189.5 , 6.467078074444808e-47 ], [ 189.5625 , 6.466672518750491e-47 ], [ 189.625 , 6.466277480374946e-47 ], [ 189.6875 , 6.465882466131555e-47 ], 
                    [ 189.75 , 6.465487476019024e-47 ], [ 189.8125 , 6.465092510035743e-47 ], [ 189.875 , 6.464697568180238e-47 ], [ 189.9375 , 6.464302650451171e-47 ], 
                    [ 190.0 , 6.463910871193811e-47 ], [ 190.0625 , 6.46352800121559e-47 ], [ 190.125 , 6.4631451539155e-47 ], [ 190.1875 , 6.462762329292245e-47 ], 
                    [ 190.25 , 6.462379527344342e-47 ], [ 190.3125 , 6.461996748070632e-47 ], [ 190.375 , 6.461613991469636e-47 ], [ 190.4375 , 6.461235665080899e-47 ], 
                    [ 190.5 , 6.460864878033638e-47 ], [ 190.5625 , 6.460494112264558e-47 ], [ 190.625 , 6.460123367772343e-47 ], [ 190.6875 , 6.459752644555775e-47 ], 
                    [ 190.75 , 6.459381942613722e-47 ], [ 190.8125 , 6.459011261944965e-47 ], [ 190.875 , 6.458646103083101e-47 ], [ 190.9375 , 6.458287318578357e-47 ], 
                    [ 191.0 , 6.457928554004466e-47 ], [ 191.0625 , 6.457569809360319e-47 ], [ 191.125 , 6.457211084644672e-47 ], [ 191.1875 , 6.456852379856785e-47 ], 
                    [ 191.25 , 6.456493694995139e-47 ], [ 191.3125 , 6.456141400239614e-47 ], [ 191.375 , 6.455794491746705e-47 ], [ 191.4375 , 6.455447601894311e-47 ], 
                    [ 191.5 , 6.455100730681429e-47 ], [ 191.5625 , 6.454753878106876e-47 ], [ 191.625 , 6.45440704416988e-47 ], [ 191.6875 , 6.454060228869252e-47 ], 
                    [ 191.75 , 6.453720504742521e-47 ], [ 191.8125 , 6.4533854012008165e-47 ], [ 191.875 , 6.453050315059102e-47 ], [ 191.9375 , 6.452715246316384e-47 ], 
                    [ 192.0 , 6.452380194971758e-47 ], [ 192.0625 , 6.452045161024414e-47 ], [ 192.125 , 6.451710144473446e-47 ], [ 192.1875 , 6.451382685749891e-47 ], 
                    [ 192.25 , 6.451059251091425e-47 ], [ 192.3125 , 6.450735832648037e-47 ], [ 192.375 , 6.450412430418871e-47 ], [ 192.4375 , 6.450089044403341e-47 ], 
                    [ 192.5 , 6.449765674600452e-47 ], [ 192.5625 , 6.44944232100939e-47 ], [ 192.625 , 6.449126856812967e-47 ], [ 192.6875 , 6.448815037119565e-47 ], 
                    [ 192.75 , 6.448503232502812e-47 ], [ 192.8125 , 6.448191442962073e-47 ], [ 192.875 , 6.447879668496524e-47 ], [ 192.9375 , 6.447567909105437e-47 ], 
                    [ 193.0 , 6.4472561647882245e-47 ], [ 193.0625 , 6.446952413001518e-47 ], [ 193.125 , 6.446652089070692e-47 ], [ 193.1875 , 6.446351779130161e-47 ], 
                    [ 193.25 , 6.446051483179134e-47 ], [ 193.3125 , 6.445751201217144e-47 ], [ 193.375 , 6.445450933243446e-47 ], [ 193.4375 , 6.445150679257344e-47 ], 
                    [ 193.5 , 6.44485835484639e-47 ], [ 193.5625 , 6.444569433354176e-47 ], [ 193.625 , 6.444280524814106e-47 ], [ 193.6875 , 6.443991629225832e-47 ], 
                    [ 193.75 , 6.443702746588632e-47 ], [ 193.8125 , 6.443413876901973e-47 ], [ 193.875 , 6.44312502016523e-47 ], [ 193.9375 , 6.442843852555878e-47 ], 
                    [ 194.0 , 6.442566238121275e-47 ], [ 194.0625 , 6.442288635648744e-47 ], [ 194.125 , 6.442011045137909e-47 ], [ 194.1875 , 6.441733466588025e-47 ], 
                    [ 194.25 , 6.441455899998578e-47 ], [ 194.3125 , 6.441178345369189e-47 ], [ 194.375 , 6.440908074889534e-47 ], [ 194.4375 , 6.440641688861596e-47 ], 
                    [ 194.5 , 6.440375313851065e-47 ], [ 194.5625 , 6.440108949857394e-47 ], [ 194.625 , 6.439842596880127e-47 ], [ 194.6875 , 6.439576254918762e-47 ], 
                    [ 194.75 , 6.439309923972891e-47 ], [ 194.8125 , 6.439050290400432e-47 ], [ 194.875 , 6.438795039177633e-47 ], [ 194.9375 , 6.438539798073234e-47 ], 
                    [ 195.0 , 6.438284567086926e-47 ], [ 195.0625 , 6.438029346218309e-47 ], [ 195.125 , 6.43777413546689e-47 ], [ 195.1875 , 6.437518934832314e-47 ], 
                    [ 195.25 , 6.437269673032206e-47 ], [ 195.3125 , 6.437025451598829e-47 ], [ 195.375 , 6.436781239430887e-47 ], [ 195.4375 , 6.436537036527939e-47 ], 
                    [ 195.5 , 6.436292842889816e-47 ], [ 195.5625 , 6.436048658516072e-47 ], [ 195.625 , 6.435804483406313e-47 ], [ 195.6875 , 6.435565346590829e-47 ], 
                    [ 195.75 , 6.435332095538153e-47 ], [ 195.8125 , 6.435098852939534e-47 ], [ 195.875 , 6.434865618794484e-47 ], [ 195.9375 , 6.434632393102834e-47 ], 
                    [ 196.0 , 6.434399175864185e-47 ], [ 196.0625 , 6.434165967078369e-47 ], [ 196.125 , 6.43393670462122e-47 ], [ 196.1875 , 6.433714292651377e-47 ], 
                    [ 196.25 , 6.433491888369951e-47 ], [ 196.3125 , 6.433269491776722e-47 ], [ 196.375 , 6.433047102871334e-47 ], [ 196.4375 , 6.432824721653702e-47 ], 
                    [ 196.5 , 6.432602348123516e-47 ], [ 196.5625 , 6.43238270457028e-47 ], [ 196.625 , 6.432171087318166e-47 ], [ 196.6875 , 6.431959477028081e-47 ], 
                    [ 196.75 , 6.431747873699574e-47 ], [ 196.8125 , 6.431536277332636e-47 ], [ 196.875 , 6.431324687926954e-47 ], [ 196.9375 , 6.431113105482293e-47 ], 
                    [ 197.0 , 6.430902863315852e-47 ], [ 197.0625 , 6.430701923534234e-47 ], [ 197.125 , 6.430500990031174e-47 ], [ 197.1875 , 6.430300062806521e-47 ], 
                    [ 197.25 , 6.4300991418599875e-47 ], [ 197.3125 , 6.42989822719147e-47 ], [ 197.375 , 6.429697318800725e-47 ], [ 197.4375 , 6.429496416687511e-47 ], 
                    [ 197.5 , 6.429305886610761e-47 ], [ 197.5625 , 6.429115559923361e-47 ], [ 197.625 , 6.428925238870155e-47 ], [ 197.6875 , 6.428734923451023e-47 ], 
                    [ 197.75 , 6.428544613665933e-47 ], [ 197.8125 , 6.428354309514307e-47 ], [ 197.875 , 6.4281640109964825e-47 ], [ 197.9375 , 6.427982327444065e-47 ], 
                    [ 198.0 , 6.427802528595458e-47 ], [ 198.0625 , 6.427622734776008e-47 ], [ 198.125 , 6.427442945985575e-47 ], [ 198.1875 , 6.427263162224153e-47 ], 
                    [ 198.25 , 6.427083383491559e-47 ], [ 198.3125 , 6.426903609787467e-47 ], [ 198.375 , 6.426730533545851e-47 ], [ 198.4375 , 6.426561156251684e-47 ], 
                    [ 198.5 , 6.426391783421431e-47 ], [ 198.5625 , 6.426222415055159e-47 ], [ 198.625 , 6.426053051152519e-47 ], [ 198.6875 , 6.425883691713486e-47 ], 
                    [ 198.75 , 6.425714336737942e-47 ], [ 198.8125 , 6.425549642754584e-47 ], [ 198.875 , 6.425390612985498e-47 ], [ 198.9375 , 6.425231587152289e-47 ], 
                    [ 199.0 , 6.4250725652549975e-47 ], [ 199.0625 , 6.424913547293343e-47 ], [ 199.125 , 6.424754533267409e-47 ], [ 199.1875 , 6.424595523176964e-47 ], 
                    [ 199.25 , 6.424439001226286e-47 ], [ 199.3125 , 6.424290253838406e-47 ], [ 199.375 , 6.424141509894619e-47 ], [ 199.4375 , 6.42399276939462e-47 ], 
                    [ 199.5 , 6.423844032338509e-47 ], [ 199.5625 , 6.423695298726207e-47 ], [ 199.625 , 6.423546568557497e-47 ], [ 199.6875 , 6.423398010860731e-47 ], 
                    [ 199.75 , 6.423259445292846e-47 ], [ 199.8125 , 6.42312088271419e-47 ], [ 199.875 , 6.422982323124516e-47 ], [ 199.9375 , 6.422843766523847e-47 ], 
                    [ 200.0 , 6.422705212912216e-47 ], [ 200.0625 , 6.422566662289371e-47 ], [ 200.125 , 6.422428114655479e-47 ], [ 200.1875 , 6.422297380516932e-47 ], 
                    [ 200.25 , 6.422168918122499e-47 ], [ 200.3125 , 6.422040458297595e-47 ], [ 200.375 , 6.421912001042264e-47 ], [ 200.4375 , 6.421783546356268e-47 ], 
                    [ 200.5 , 6.42165509423988e-47 ], [ 200.5625 , 6.421526644692725e-47 ], [ 200.625 , 6.421403372194068e-47 ], [ 200.6875 , 6.421284946234072e-47 ], 
                    [ 200.75 , 6.421166522458086e-47 ], [ 200.8125 , 6.421048100866118e-47 ], [ 200.875 , 6.420929681458079e-47 ], [ 200.9375 , 6.420811264234022e-47 ], 
                    [ 201.0 , 6.420692849193906e-47 ], [ 201.0625 , 6.420576835858796e-47 ], [ 201.125 , 6.420468350188902e-47 ], [ 201.1875 , 6.420359866351907e-47 ], 
                    [ 201.25 , 6.420251384347913e-47 ], [ 201.3125 , 6.42014290417694e-47 ], [ 201.375 , 6.420034425838954e-47 ], [ 201.4375 , 6.419925949333789e-47 ], 
                    [ 201.5 , 6.419817474661456e-47 ], [ 201.5625 , 6.419718371496903e-47 ], [ 201.625 , 6.419619757185302e-47 ], [ 201.6875 , 6.419521144388578e-47 ], 
                    [ 201.75 , 6.41942253310666e-47 ], [ 201.8125 , 6.419323923339434e-47 ], [ 201.875 , 6.41922531508706e-47 ], [ 201.9375 , 6.419126708349377e-47 ], 
                    [ 202.0 , 6.419034383611725e-47 ], [ 202.0625 , 6.4189455513973575e-47 ], [ 202.125 , 6.418856720412465e-47 ], [ 202.1875 , 6.418767890656801e-47 ], 
                    [ 202.25 , 6.418679062130442e-47 ], [ 202.3125 , 6.418590234833414e-47 ], [ 202.375 , 6.418501408765703e-47 ], [ 202.4375 , 6.418415677660573e-47 ], 
                    [ 202.5 , 6.418336570382909e-47 ], [ 202.5625 , 6.41825746408029e-47 ], [ 202.625 , 6.418178358752662e-47 ], [ 202.6875 , 6.418099254400056e-47 ], 
                    [ 202.75 , 6.418020151022368e-47 ], [ 202.8125 , 6.417941048619633e-47 ], [ 202.875 , 6.417861947191885e-47 ], [ 202.9375 , 6.417792260231382e-47 ], 
                    [ 203.0 , 6.417722791295426e-47 ], [ 203.0625 , 6.417653323111432e-47 ], [ 203.125 , 6.417583855679389e-47 ], [ 203.1875 , 6.41751438899934e-47 ], 
                    [ 203.25 , 6.41744492307118e-47 ], [ 203.3125 , 6.41737545789486e-47 ], [ 203.375 , 6.417311916077486e-47 ], [ 203.4375 , 6.41725200730661e-47 ], 
                    [ 203.5 , 6.417192099095059e-47 ], [ 203.5625 , 6.417132191442827e-47 ], [ 203.625 , 6.417072284349724e-47 ], [ 203.6875 , 6.417012377816068e-47 ], 
                    [ 203.75 , 6.416952471841484e-47 ], [ 203.8125 , 6.416894894645322e-47 ], [ 203.875 , 6.416844469407031e-47 ], [ 203.9375 , 6.41679404456499e-47 ], 
                    [ 204.0 , 6.416743620119291e-47 ], [ 204.0625 , 6.4166931960696985e-47 ], [ 204.125 , 6.416642772416394e-47 ], [ 204.1875 , 6.416592349159375e-47 ], 
                    [ 204.25 , 6.416541926298592e-47 ], [ 204.3125 , 6.416499551788719e-47 ], [ 204.375 , 6.416458545492037e-47 ], [ 204.4375 , 6.416417539457417e-47 ], 
                    [ 204.5 , 6.416376533684903e-47 ], [ 204.5625 , 6.416335528174354e-47 ], [ 204.625 , 6.416294522925999e-47 ], [ 204.6875 , 6.416253517939654e-47 ], 
                    [ 204.75 , 6.416216687969121e-47 ], [ 204.8125 , 6.416185025432274e-47 ], [ 204.875 , 6.41615336305172e-47 ], [ 204.9375 , 6.416121700827458e-47 ], 
                    [ 205.0 , 6.416090038759351e-47 ], [ 205.0625 , 6.416058376847534e-47 ], [ 205.125 , 6.41602671509187e-47 ], [ 205.1875 , 6.415995261034678e-47 ], 
                    [ 205.25 , 6.415972864826471e-47 ], [ 205.3125 , 6.415950468696396e-47 ], [ 205.375 , 6.415928072644409e-47 ], [ 205.4375 , 6.41590567667069e-47 ], 
                    [ 205.5 , 6.415883280775148e-47 ], [ 205.5625 , 6.415860884957782e-47 ], [ 205.625 , 6.415838489218593e-47 ], [ 205.6875 , 6.415821441895599e-47 ], 
                    [ 205.75 , 6.41580824977254e-47 ], [ 205.8125 , 6.415795057676514e-47 ], [ 205.875 , 6.415781865607752e-47 ], [ 205.9375 , 6.415768673566114e-47 ], 
                    [ 206.0 , 6.415755481551465e-47 ], [ 206.0625 , 6.415742289564032e-47 ], [ 206.125 , 6.41573021937674e-47 ], [ 206.1875 , 6.415726156442019e-47 ], 
                    [ 206.25 , 6.415722093509778e-47 ], [ 206.3125 , 6.415718030580111e-47 ], [ 206.375 , 6.415713967653062e-47 ], [ 206.4375 , 6.415709904728585e-47 ], 
                    [ 206.5 , 6.415705841806683e-47 ], [ 206.5625 , 6.415701778887354e-47 ], [ 206.625 , 6.415703574343714e-47 ], [ 206.6875 , 6.415708554362714e-47 ], 
                    [ 206.75 , 6.415713534385534e-47 ], [ 206.8125 , 6.415718514412264e-47 ], [ 206.875 , 6.415723494442815e-47 ], [ 206.9375 , 6.415728474477231e-47 ], 
                    [ 207.0 , 6.415733454515558e-47 ], [ 207.0625 , 6.415739834287133e-47 ], [ 207.125 , 6.415753805416364e-47 ], [ 207.1875 , 6.415767776575974e-47 ], 
                    [ 207.25 , 6.415781747766007e-47 ], [ 207.3125 , 6.415795718986512e-47 ], [ 207.375 , 6.415809690237302e-47 ], [ 207.4375 , 6.415823661518654e-47 ], 
                    [ 207.5 , 6.41583763283043e-47 ], [ 207.5625 , 6.4158573789363175e-47 ], [ 207.625 , 6.415880279661887e-47 ], [ 207.6875 , 6.415903180469062e-47 ], 
                    [ 207.75 , 6.415926081358116e-47 ], [ 207.8125 , 6.415948982328866e-47 ], [ 207.875 , 6.415971883381313e-47 ], [ 207.9375 , 6.415994784515548e-47 ], 
                    [ 208.0 , 6.416018751181516e-47 ], [ 208.0625 , 6.416050484169373e-47 ], [ 208.125 , 6.416082217314132e-47 ], [ 208.1875 , 6.416113950615841e-47 ], 
                    [ 208.25 , 6.416145684074499e-47 ], [ 208.3125 , 6.416177417690198e-47 ], [ 208.375 , 6.416209151462804e-47 ], [ 208.4375 , 6.416240885392317e-47 ], 
                    [ 208.5 , 6.416277696964007e-47 ], [ 208.5625 , 6.416318224381474e-47 ], [ 208.625 , 6.416358752054878e-47 ], [ 208.6875 , 6.416399279984271e-47 ], 
                    [ 208.75 , 6.41643980816956e-47 ], [ 208.8125 , 6.416480336610975e-47 ], [ 208.875 , 6.416520865308383e-47 ], [ 208.9375 , 6.41656154477788e-47 ], 
                    [ 209.0 , 6.41661078447841e-47 ], [ 209.0625 , 6.416660024556888e-47 ], [ 209.125 , 6.4167092650130455e-47 ], [ 209.1875 , 6.416758505847157e-47 ], 
                    [ 209.25 , 6.416807747059226e-47 ], [ 209.3125 , 6.416856988649073e-47 ], [ 209.375 , 6.416906230616748e-47 ], [ 209.4375 , 6.416959280130076e-47 ], 
                    [ 209.5 , 6.417017175611511e-47 ], [ 209.5625 , 6.417075071615203e-47 ], [ 209.625 , 6.417132968141249e-47 ], [ 209.6875 , 6.417190865189742e-47 ], 
                    [ 209.75 , 6.417248762760552e-47 ], [ 209.8125 , 6.41730666085373e-47 ], [ 209.875 , 6.417364559469279e-47 ], [ 209.9375 , 6.417429707330311e-47 ], 
                    [ 210.0 , 6.417496181049971e-47 ], [ 210.0625 , 6.417562655458278e-47 ], [ 210.125 , 6.417629130555101e-47 ], [ 210.1875 , 6.417695606340539e-47 ], 
                    [ 210.25 , 6.417762082814373e-47 ], [ 210.3125 , 6.417828559976972e-47 ], [ 210.375 , 6.417897026391643e-47 ], [ 210.4375 , 6.417972017733453e-47 ], 
                    [ 210.5 , 6.418047009951561e-47 ], [ 210.5625 , 6.418122003045933e-47 ], [ 210.625 , 6.418196997016486e-47 ], [ 210.6875 , 6.41827199186337e-47 ], 
                    [ 210.75 , 6.418346987586591e-47 ], [ 210.8125 , 6.418421984186118e-47 ], [ 210.875 , 6.418502090203798e-47 ], [ 210.9375 , 6.418585537621637e-47 ], 
                    [ 211.0 , 6.418668986124337e-47 ], [ 211.0625 , 6.418752435712001e-47 ], [ 211.125 , 6.418835886384555e-47 ], [ 211.1875 , 6.418919338142148e-47 ], 
                    [ 211.25 , 6.419002790984613e-47 ], [ 211.3125 , 6.4190862449121e-47 ], [ 211.375 , 6.419177736020254e-47 ], [ 211.4375 , 6.419269577249778e-47 ], 
                    [ 211.5 , 6.419361419793348e-47 ], [ 211.5625 , 6.419453263650939e-47 ], [ 211.625 , 6.419545108822524e-47 ], [ 211.6875 , 6.419636955308122e-47 ], 
                    [ 211.75 , 6.419728803107933e-47 ], [ 211.8125 , 6.419823106500185e-47 ], [ 211.875 , 6.419923284581387e-47 ], [ 211.9375 , 6.420023464225817e-47 ], 
                    [ 212.0 , 6.420123645433547e-47 ], [ 212.0625 , 6.420223828204507e-47 ], [ 212.125 , 6.420324012538768e-47 ], [ 212.1875 , 6.42042419843631e-47 ], 
                    [ 212.25 , 6.420524385897293e-47 ], [ 212.3125 , 6.420629641270612e-47 ], [ 212.375 , 6.420738086684106e-47 ], [ 212.4375 , 6.42084653392917e-47 ], 
                    [ 212.5 , 6.420954983006014e-47 ], [ 212.5625 , 6.421063433914488e-47 ], [ 212.625 , 6.421171886654852e-47 ], [ 212.6875 , 6.421280341226863e-47 ], 
                    [ 212.75 , 6.421388797630734e-47 ], [ 212.8125 , 6.421504744715505e-47 ], [ 212.875 , 6.4216213863733345e-47 ], [ 212.9375 , 6.421738030149915e-47 ], 
                    [ 213.0 , 6.421854676045101e-47 ], [ 213.0625 , 6.421971324059207e-47 ], [ 213.125 , 6.422087974192089e-47 ], [ 213.1875 , 6.422204626443829e-47 ], 
                    [ 213.25 , 6.422322903282061e-47 ], [ 213.3125 , 6.422447691809611e-47 ], [ 213.375 , 6.422572482761857e-47 ], [ 213.4375 , 6.422697276138799e-47 ], 
                    [ 213.5 , 6.422822071940531e-47 ], [ 213.5625 , 6.422946870167191e-47 ], [ 213.625 , 6.42307167081878e-47 ], [ 213.6875 , 6.423196473895072e-47 ], 
                    [ 213.75 , 6.423325046619836e-47 ], [ 213.8125 , 6.423457934792774e-47 ], [ 213.875 , 6.423590825714907e-47 ], [ 213.9375 , 6.423723719386291e-47 ], 
                    [ 214.0 , 6.423856615807123e-47 ], [ 214.0625 , 6.423989514977319e-47 ], [ 214.125 , 6.424122416896983e-47 ], [ 214.1875 , 6.424255321566172e-47 ], 
                    [ 214.25 , 6.424393950623465e-47 ], [ 214.3125 , 6.424534859348731e-47 ], [ 214.375 , 6.424675771164693e-47 ], [ 214.4375 , 6.424816686071376e-47 ], 
                    [ 214.5 , 6.424957604068708e-47 ], [ 214.5625 , 6.425098525156758e-47 ], [ 214.625 , 6.425239449335913e-47 ], [ 214.6875 , 6.425380376605831e-47 ], 
                    [ 214.75 , 6.425528814471558e-47 ], [ 214.8125 , 6.425677676947694e-47 ], [ 214.875 , 6.425826542872624e-47 ], [ 214.9375 , 6.425975412246338e-47 ], 
                    [ 215.0 , 6.42612428506896e-47 ], [ 215.0625 , 6.426273161340618e-47 ], [ 215.125 , 6.426422041061297e-47 ], [ 215.1875 , 6.426572192783555e-47 ], 
                    [ 215.25 , 6.426728968627906e-47 ], [ 215.3125 , 6.426885748296838e-47 ], [ 215.375 , 6.427042531790448e-47 ], [ 215.4375 , 6.42719931910869e-47 ], 
                    [ 215.5 , 6.427356110251795e-47 ], [ 215.5625 , 6.427512905219811e-47 ], [ 215.625 , 6.42766970401283e-47 ], [ 215.6875 , 6.427829306686232e-47 ], 
                    [ 215.75 , 6.427993954052174e-47 ], [ 215.8125 , 6.42815860563543e-47 ], [ 215.875 , 6.428323261436244e-47 ], [ 215.9375 , 6.428487921454723e-47 ], 
                    [ 216.0 , 6.42865258569093e-47 ], [ 216.0625 , 6.428817254144884e-47 ], [ 216.125 , 6.428981926816963e-47 ], [ 216.1875 , 6.429150747151838e-47 ], 
                    [ 216.25 , 6.429323161187281e-47 ], [ 216.3125 , 6.429495579846492e-47 ], [ 216.375 , 6.429668003129547e-47 ], [ 216.4375 , 6.42984043103648e-47 ], 
                    [ 216.5 , 6.430012863567689e-47 ], [ 216.5625 , 6.430185300722977e-47 ], [ 216.625 , 6.430357742502698e-47 ], [ 216.6875 , 6.430535567832566e-47 ], 
                    [ 216.75 , 6.430715740462086e-47 ], [ 216.8125 , 6.430895918139826e-47 ], [ 216.875 , 6.431076100865838e-47 ], [ 216.9375 , 6.431256288640216e-47 ], 
                    [ 217.0 , 6.431436481463148e-47 ], [ 217.0625 , 6.431616679334776e-47 ], [ 217.125 , 6.431796882255241e-47 ], [ 217.1875 , 6.431983493957394e-47 ], 
                    [ 217.25 , 6.432171332321032e-47 ], [ 217.3125 , 6.432359176170265e-47 ], [ 217.375 , 6.432547025505252e-47 ], [ 217.4375 , 6.432734880326154e-47 ], 
                    [ 217.5 , 6.432922740633176e-47 ], [ 217.5625 , 6.433110606426296e-47 ], [ 217.625 , 6.433298477705903e-47 ], [ 217.6875 , 6.433493690931492e-47 ], 
                    [ 217.75 , 6.433689175232701e-47 ], [ 217.8125 , 6.433884665473781e-47 ], [ 217.875 , 6.434080161655002e-47 ], [ 217.9375 , 6.434275663776269e-47 ], 
                    [ 218.0 , 6.434471171838084e-47 ], [ 218.0625 , 6.434666685840537e-47 ], [ 218.125 , 6.434862751050289e-47 ], [ 218.1875 , 6.435065783607718e-47 ], 
                    [ 218.25 , 6.435268822571222e-47 ], [ 218.3125 , 6.435471867940959e-47 ], [ 218.375 , 6.435674919717176e-47 ], [ 218.4375 , 6.435877977900122e-47 ], 
                    [ 218.5 , 6.43608104248995e-47 ], [ 218.5625 , 6.436284113486913e-47 ], [ 218.625 , 6.436488403686274e-47 ], [ 218.6875 , 6.436698963093171e-47 ], 
                    [ 218.75 , 6.436909529388181e-47 ], [ 218.8125 , 6.437120102571531e-47 ], [ 218.875 , 6.437330682643398e-47 ], [ 218.9375 , 6.437541269604148e-47 ], 
                    [ 219.0 , 6.437751863453775e-47 ], [ 219.0625 , 6.43796246419278e-47 ], [ 219.125 , 6.438174804864138e-47 ], [ 219.1875 , 6.438392833456852e-47 ], 
                    [ 219.25 , 6.438610869433189e-47 ], [ 219.3125 , 6.438828912793353e-47 ], [ 219.375 , 6.439046963537457e-47 ], [ 219.4375 , 6.439265021665888e-47 ], 
                    [ 219.5 , 6.439483087178852e-47 ], [ 219.5625 , 6.439701160076644e-47 ], [ 219.625 , 6.439921353425854e-47 ], [ 219.6875 , 6.440146790704996e-47 ], 
                    [ 219.75 , 6.440372235875797e-47 ], [ 219.8125 , 6.440597688938671e-47 ], [ 219.875 , 6.440823149893759e-47 ], [ 219.9375 , 6.441048618741474e-47 ], 
                    [ 220.0 , 6.441274095481862e-47 ], [ 220.0625 , 6.441499580115384e-47 ], [ 220.125 , 6.441727429104356e-47 ], [ 220.1875 , 6.441960220901813e-47 ], 
                    [ 220.25 , 6.4421930211118305e-47 ], [ 220.3125 , 6.4424258297348525e-47 ], [ 220.375 , 6.442658646771182e-47 ], [ 220.4375 , 6.442891472221031e-47 ], 
                    [ 220.5 , 6.443124306084796e-47 ], [ 220.5625 , 6.443357148362733e-47 ], [ 220.625 , 6.443592464087182e-47 ], [ 220.6875 , 6.443832556625597e-47 ], 
                    [ 220.75 , 6.44407265811011e-47 ], [ 220.8125 , 6.444312768540919e-47 ], [ 220.875 , 6.444552887918401e-47 ], [ 220.9375 , 6.444793016242891e-47 ], 
                    [ 221.0 , 6.445033153514676e-47 ], [ 221.0625 , 6.445273299734227e-47 ], [ 221.125 , 6.445515902851239e-47 ], [ 221.1875 , 6.44576326041347e-47 ], 
                    [ 221.25 , 6.446010627468371e-47 ], [ 221.3125 , 6.446258004016402e-47 ], [ 221.375 , 6.446505390057971e-47 ], [ 221.4375 , 6.446752785593352e-47 ], 
                    [ 221.5 , 6.447000190623047e-47 ], [ 221.5625 , 6.447247605147235e-47 ], [ 221.625 , 6.447497322263206e-47 ], [ 221.6875 , 6.447751882874747e-47 ], 
                    [ 221.75 , 6.448006453536826e-47 ], [ 221.8125 , 6.448261034249976e-47 ], [ 221.875 , 6.448515625014365e-47 ], [ 221.9375 , 6.448770225830618e-47 ], 
                    [ 222.0 , 6.449024836699043e-47 ], [ 222.0625 , 6.449279457620034e-47 ], [ 222.125 , 6.449536104152293e-47 ], [ 222.1875 , 6.449797818670691e-47 ], 
                    [ 222.25 , 6.450059543809199e-47 ], [ 222.3125 , 6.450321279568111e-47 ], [ 222.375 , 6.450583025948041e-47 ], [ 222.4375 , 6.450844782949329e-47 ], 
                    [ 222.5 , 6.451106550572313e-47 ], [ 222.5625 , 6.451368328817609e-47 ], [ 222.625 , 6.451631729350469e-47 ], [ 222.6875 , 6.451900538873259e-47 ], 
                    [ 222.75 , 6.452169359596141e-47 ], [ 222.8125 , 6.452438191519487e-47 ], [ 222.875 , 6.452707034643809e-47 ], [ 222.9375 , 6.452975888969668e-47 ], 
                    [ 223.0 , 6.453244754497346e-47 ], [ 223.0625 , 6.453513631227355e-47 ], [ 223.125 , 6.453783611386232e-47 ], [ 223.1875 , 6.454059483704982e-47 ], 
                    [ 223.25 , 6.454335367815985e-47 ], [ 223.3125 , 6.454611263720067e-47 ], [ 223.375 , 6.454887171417454e-47 ], [ 223.4375 , 6.455163090908699e-47 ], 
                    [ 223.5 , 6.455439022194443e-47 ], [ 223.5625 , 6.455714965275005e-47 ], [ 223.625 , 6.455991370740096e-47 ], [ 223.6875 , 6.456274257094897e-47 ], 
                    [ 223.75 , 6.456557155845159e-47 ], [ 223.8125 , 6.456840066991379e-47 ], [ 223.875 , 6.457122990534145e-47 ], [ 223.9375 , 6.45740592647391e-47 ], 
                    [ 224.0 , 6.457688874811353e-47 ], [ 224.0625 , 6.457971835546698e-47 ], [ 224.125 , 6.458254808680947e-47 ], [ 224.1875 , 6.458544333071057e-47 ], 
                    [ 224.25 , 6.458834177134572e-47 ], [ 224.3125 , 6.459124034205596e-47 ], [ 224.375 , 6.459413904284621e-47 ], [ 224.4375 , 6.45970378737251e-47 ], 
                    [ 224.5 , 6.459993683469429e-47 ], [ 224.5625 , 6.460283592576331e-47 ], [ 224.625 , 6.460573514693709e-47 ], [ 224.6875 , 6.460869078137207e-47 ], 
                    [ 224.75 , 6.461165833531005e-47 ], [ 224.8125 , 6.461462602555178e-47 ], [ 224.875 , 6.461759385210259e-47 ], [ 224.9375 , 6.462056181496918e-47 ], 
                    [ 225.0 , 6.462352991415877e-47 ], [ 225.0625 , 6.462649814967485e-47 ], [ 225.125 , 6.462946652152642e-47 ], [ 225.1875 , 6.463248102034924e-47 ], 
                    [ 225.25 , 6.463551728753609e-47 ], [ 225.3125 , 6.463855369735895e-47 ], [ 225.375 , 6.464159024982451e-47 ], [ 225.4375 , 6.4644626944939e-47 ], 
                    [ 225.5 , 6.464766378271051e-47 ], [ 225.5625 , 6.465070076314391e-47 ], [ 225.625 , 6.465373788624636e-47 ], [ 225.6875 , 6.465680964627121e-47 ], 
                    [ 225.75 , 6.465991403346703e-47 ], [ 225.8125 , 6.466301856971474e-47 ], [ 225.875 , 6.466612325502104e-47 ], [ 225.9375 , 6.466922808939263e-47 ], 
                    [ 226.0 , 6.467233307283895e-47 ], [ 226.0625 , 6.467543820536627e-47 ], [ 226.125 , 6.467854348698035e-47 ], [ 226.1875 , 6.468167094121042e-47 ], 
                    [ 226.25 , 6.468484298934698e-47 ], [ 226.3125 , 6.468801519304323e-47 ], [ 226.375 , 6.469118755230768e-47 ], [ 226.4375 , 6.4694360067148e-47 ], 
                    [ 226.5 , 6.469753273757041e-47 ], [ 226.5625 , 6.470070556358348e-47 ], [ 226.625 , 6.470387854519575e-47 ], [ 226.6875 , 6.470706020059239e-47 ], 
                    [ 226.75 , 6.471029946814192e-47 ], [ 226.8125 , 6.471353889785125e-47 ], [ 226.875 , 6.471677848972758e-47 ], [ 226.9375 , 6.472001824377903e-47 ], 
                    [ 227.0 , 6.472325816001509e-47 ], [ 227.0625 , 6.472649823844204e-47 ], [ 227.125 , 6.472973847906986e-47 ], [ 227.1875 , 6.473297888190618e-47 ], 
                    [ 227.25 , 6.473627899085023e-47 ], [ 227.3125 , 6.473958527832232e-47 ], [ 227.375 , 6.474289173465794e-47 ], [ 227.4375 , 6.474619835986525e-47 ], 
                    [ 227.5 , 6.474950515395197e-47 ], [ 227.5625 , 6.4752812116926695e-47 ], [ 227.625 , 6.475611924879946e-47 ], [ 227.6875 , 6.475942654957703e-47 ], 
                    [ 227.75 , 6.476277734630134e-47 ], [ 227.8125 , 6.476614977373357e-47 ], [ 227.875 , 6.476952237678097e-47 ], [ 227.9375 , 6.477289515545038e-47 ], 
                    [ 228.0 , 6.477626810975418e-47 ], [ 228.0625 , 6.477964123969829e-47 ], [ 228.125 , 6.478301454529417e-47 ], [ 228.1875 , 6.478638802654817e-47 ], 
                    [ 228.25 , 6.478978820823357e-47 ], [ 228.3125 , 6.479322662675224e-47 ], [ 228.375 , 6.479666522774907e-47 ], [ 228.4375 , 6.480010401123377e-47 ], 
                    [ 228.5 , 6.480354297721598e-47 ], [ 228.5625 , 6.48069821257059e-47 ], [ 228.625 , 6.481042145671179e-47 ], [ 228.6875 , 6.48138609702452e-47 ], 
                    [ 228.75 , 6.481730923756258e-47 ], [ 228.8125 , 6.482081319937854e-47 ], [ 228.875 , 6.48243173506144e-47 ], [ 228.9375 , 6.482782169128271e-47 ], 
                    [ 229.0 , 6.483132622139142e-47 ], [ 229.0625 , 6.483483094095121e-47 ], [ 229.125 , 6.483833584997511e-47 ], [ 229.1875 , 6.484184094846779e-47 ], 
                    [ 229.25 , 6.484534623644554e-47 ], [ 229.3125 , 6.484890485989539e-47 ], [ 229.375 , 6.485247393509205e-47 ], [ 229.4375 , 6.48560432067196e-47 ], 
                    [ 229.5 , 6.485961267478745e-47 ], [ 229.5625 , 6.486318233930874e-47 ], [ 229.625 , 6.486675220029104e-47 ], [ 229.6875 , 6.487032225774792e-47 ], 
                    [ 229.75 , 6.487389251169022e-47 ], [ 229.8125 , 6.48774959216257e-47 ], [ 229.875 , 6.488112949291119e-47 ], [ 229.9375 , 6.488476326770083e-47 ], 
                    [ 230.0 , 6.488839724600651e-47 ], [ 230.0625 , 6.489203142783868e-47 ], [ 230.125 , 6.489566581320874e-47 ], [ 230.1875 , 6.489930040212948e-47 ], 
                    [ 230.25 , 6.490293519461091e-47 ], [ 230.3125 , 6.490658212303504e-47 ], [ 230.375 , 6.491027994427593e-47 ], [ 230.4375 , 6.4913977976186565e-47 ], 
                    [ 230.5 , 6.491767621877984e-47 ], [ 230.5625 , 6.492137467206639e-47 ], [ 230.625 , 6.492507333606006e-47 ], [ 230.6875 , 6.492877221077102e-47 ], 
                    [ 230.75 , 6.493247129621355e-47 ], [ 230.8125 , 6.493617059239739e-47 ], [ 230.875 , 6.493992204782451e-47 ], [ 230.9375 , 6.494368373353395e-47 ], 
                    [ 231.0 , 6.494744563714141e-47 ], [ 231.0625 , 6.495120775865948e-47 ], [ 231.125 , 6.495497009810128e-47 ], [ 231.1875 , 6.495873265547848e-47 ], 
                    [ 231.25 , 6.496249543080417e-47 ], [ 231.3125 , 6.496625842409099e-47 ], [ 231.375 , 6.497005041039372e-47 ], [ 231.4375 , 6.497387543533357e-47 ], 
                    [ 231.5 , 6.497770068546665e-47 ], [ 231.5625 , 6.498152616080622e-47 ], [ 231.625 , 6.4985351861366e-47 ], [ 231.6875 , 6.498917778715879e-47 ], 
                    [ 231.75 , 6.499300393819741e-47 ], [ 231.8125 , 6.499683031449508e-47 ], [ 231.875 , 6.500066163053656e-47 ], [ 231.9375 , 6.500454946634615e-47 ], 
                    [ 232.0 , 6.500843753469594e-47 ], [ 232.0625 , 6.501232583559938e-47 ], [ 232.125 , 6.501621436907131e-47 ], [ 232.1875 , 6.502010313512518e-47 ], 
                    [ 232.25 , 6.502399213377489e-47 ], [ 232.3125 , 6.50278813650339e-47 ], [ 232.375 , 6.503177082891659e-47 ], [ 232.4375 , 6.503570106728874e-47 ], 
                    [ 232.5 , 6.503965167942153e-47 ], [ 232.5625 , 6.504360253153535e-47 ], [ 232.625 , 6.504755362364529e-47 ], [ 232.6875 , 6.505150495576589e-47 ], 
                    [ 232.75 , 6.505545652791036e-47 ], [ 232.8125 , 6.505940834009606e-47 ], [ 232.875 , 6.506336039233385e-47 ], [ 232.9375 , 6.506732710803159e-47 ], 
                    [ 233.0 , 6.507133974244685e-47 ], [ 233.0625 , 6.507535262431616e-47 ], [ 233.125 , 6.507936575365622e-47 ], [ 233.1875 , 6.50833791304813e-47 ], 
                    [ 233.25 , 6.508739275480901e-47 ], [ 233.3125 , 6.509140662665091e-47 ], [ 233.375 , 6.509542074602505e-47 ], [ 233.4375 , 6.509943511294621e-47 ], 
                    [ 233.5 , 6.510349706415974e-47 ], [ 233.5625 , 6.51075716411593e-47 ], [ 233.625 , 6.51116464731714e-47 ], [ 233.6875 , 6.5115721560210615e-47 ], 
                    [ 233.75 , 6.511979690229429e-47 ], [ 233.8125 , 6.512387249943792e-47 ], [ 233.875 , 6.512794835165746e-47 ], [ 233.9375 , 6.513202445896888e-47 ], 
                    [ 234.0 , 6.513612016441462e-47 ], [ 234.0625 , 6.514025604863641e-47 ], [ 234.125 , 6.514439219547132e-47 ], [ 234.1875 , 6.514852860493414e-47 ], 
                    [ 234.25 , 6.515266527704249e-47 ], [ 234.3125 , 6.51568022118135e-47 ], [ 234.375 , 6.51609394092634e-47 ], [ 234.4375 , 6.516507686940747e-47 ], 
                    [ 234.5 , 6.516921459226565e-47 ], [ 234.5625 , 6.517340203461451e-47 ], [ 234.625 , 6.517759911094844e-47 ], [ 234.6875 , 6.518179645756825e-47 ], 
                    [ 234.75 , 6.5185994074491325e-47 ], [ 234.8125 , 6.519019196173556e-47 ], [ 234.875 , 6.519439011931697e-47 ], [ 234.9375 , 6.519858854725478e-47 ], 
                    [ 235.0 , 6.520278724556505e-47 ], [ 235.0625 , 6.520700577302921e-47 ], [ 235.125 , 6.52112632277842e-47 ], [ 235.1875 , 6.521552096051471e-47 ], 
                    [ 235.25 , 6.5219778971238875e-47 ], [ 235.3125 , 6.52240372599744e-47 ], [ 235.375 , 6.522829582673896e-47 ], [ 235.4375 , 6.523255467155164e-47 ], 
                    [ 235.5 , 6.523681379442969e-47 ], [ 235.5625 , 6.524107319539124e-47 ], [ 235.625 , 6.524537996559399e-47 ], [ 235.6875 , 6.524969792369885e-47 ], 
                    [ 235.75 , 6.525401616756697e-47 ], [ 235.8125 , 6.525833469721771e-47 ], [ 235.875 , 6.526265351267046e-47 ], [ 235.9375 , 6.526697261394366e-47 ], 
                    [ 236.0 , 6.527129200105577e-47 ], [ 236.0625 , 6.52756116740257e-47 ], [ 236.125 , 6.52799470261322e-47 ], [ 236.1875 , 6.528432461056295e-47 ], 
                    [ 236.25 , 6.528870248854857e-47 ], [ 236.3125 , 6.529308066010919e-47 ], [ 236.375 , 6.52974591252631e-47 ], [ 236.4375 , 6.530183788403234e-47 ], 
                    [ 236.5 , 6.530621693643424e-47 ], [ 236.5625 , 6.531059628248897e-47 ], [ 236.625 , 6.531497592221716e-47 ], [ 236.6875 , 6.531939604863576e-47 ], 
                    [ 236.75 , 6.532383323227612e-47 ], [ 236.8125 , 6.532827071733688e-47 ], [ 236.875 , 6.533270850383853e-47 ], [ 236.9375 , 6.533714659180244e-47 ], 
                    [ 237.0 , 6.53415849812468e-47 ], [ 237.0625 , 6.534602367219441e-47 ], [ 237.125 , 6.535046266466482e-47 ], [ 237.1875 , 6.535490894665054e-47 ], 
                    [ 237.25 , 6.535940512018354e-47 ], [ 237.3125 , 6.5363901603036e-47 ], [ 237.375 , 6.536839839523108e-47 ], [ 237.4375 , 6.537289549678727e-47 ], 
                    [ 237.5 , 6.537739290772722e-47 ], [ 237.5625 , 6.538189062807366e-47 ], [ 237.625 , 6.538638865784504e-47 ], [ 237.6875 , 6.539088699706407e-47 ], 
                    [ 237.75 , 6.539541497080467e-47 ], [ 237.8125 , 6.539997012177267e-47 ], [ 237.875 , 6.540452559003249e-47 ], [ 237.9375 , 6.540908137560671e-47 ], 
                    [ 238.0 , 6.541363747851511e-47 ], [ 238.0625 , 6.541819389878116e-47 ], [ 238.125 , 6.542275063642746e-47 ], [ 238.1875 , 6.542730769147612e-47 ], 
                    [ 238.25 , 6.54318650639483e-47 ], [ 238.3125 , 6.543647280969133e-47 ], [ 238.375 , 6.544108639266875e-47 ], [ 238.4375 , 6.544570030092526e-47 ], 
                    [ 238.5 , 6.545031453448522e-47 ], [ 238.5625 , 6.5454929093370615e-47 ], [ 238.625 , 6.545954397760254e-47 ], [ 238.6875 , 6.546415918720856e-47 ], 
                    [ 238.75 , 6.546877472220698e-47 ], [ 238.8125 , 6.547340500251376e-47 ], [ 238.875 , 6.547807662033479e-47 ], [ 238.9375 , 6.548274857148128e-47 ], 
                    [ 239.0 , 6.548742085597839e-47 ], [ 239.0625 , 6.549209347384991e-47 ], [ 239.125 , 6.54967664251187e-47 ], [ 239.1875 , 6.55014397098095e-47 ], 
                    [ 239.25 , 6.550611332794512e-47 ], [ 239.3125 , 6.551078727954939e-47 ], [ 239.375 , 6.551549438197982e-47 ], [ 239.4375 , 6.552022371701951e-47 ], 
                    [ 239.5 , 6.552495339345294e-47 ], [ 239.5625 , 6.552968341130567e-47 ], [ 239.625 , 6.553441377060236e-47 ], [ 239.6875 , 6.553914447136627e-47 ], 
                    [ 239.75 , 6.554387551362389e-47 ], [ 239.8125 , 6.554860689739896e-47 ], [ 239.875 , 6.555333862271472e-47 ], [ 239.9375 , 6.555812074524156e-47 ], 
                    [ 240.0 , 6.556290756960834e-47 ], [ 240.0625 , 6.5567694743492265e-47 ], [ 240.125 , 6.557248226691841e-47 ], [ 240.1875 , 6.55772701399132e-47 ], 
                    [ 240.25 , 6.558205836250171e-47 ], [ 240.3125 , 6.558684693470994e-47 ], [ 240.375 , 6.559163585656154e-47 ], [ 240.4375 , 6.559643705137694e-47 ], 
                    [ 240.5 , 6.560128075623449e-47 ], [ 240.5625 , 6.560612481875687e-47 ], [ 240.625 , 6.561096923896863e-47 ], [ 240.6875 , 6.561581401689757e-47 ], 
                    [ 240.75 , 6.562065915256961e-47 ], [ 240.8125 , 6.5625504646010275e-47 ], [ 240.875 , 6.563035049724828e-47 ], [ 240.9375 , 6.563519670630775e-47 ], 
                    [ 241.0 , 6.564007005990592e-47 ], [ 241.0625 , 6.564497044429209e-47 ], [ 241.125 , 6.564987119451691e-47 ], [ 241.1875 , 6.565477231060957e-47 ], 
                    [ 241.25 , 6.565967379259737e-47 ], [ 241.3125 , 6.566457564050766e-47 ], [ 241.375 , 6.566947785436585e-47 ], [ 241.4375 , 6.567438043420116e-47 ], 
                    [ 241.5 , 6.56792833800409e-47 ], [ 241.5625 , 6.568422733992411e-47 ], [ 241.625 , 6.568918436315672e-47 ], [ 241.6875 , 6.569414176048388e-47 ], 
                    [ 241.75 , 6.569909953193286e-47 ], [ 241.8125 , 6.570405767753236e-47 ], [ 241.875 , 6.570901619731015e-47 ], [ 241.9375 , 6.571397509129542e-47 ], 
                    [ 242.0 , 6.571893435951593e-47 ], [ 242.0625 , 6.572389417039741e-47 ], [ 242.125 , 6.572890681535195e-47 ], [ 242.1875 , 6.573391984261204e-47 ], 
                    [ 242.25 , 6.573893325220683e-47 ], [ 242.3125 , 6.574394704416457e-47 ], [ 242.375 , 6.574896121851718e-47 ], [ 242.4375 , 6.575397577529014e-47 ], 
                    [ 242.5 , 6.575899071451538e-47 ], [ 242.5625 , 6.576400603622209e-47 ], [ 242.625 , 6.576903352478652e-47 ], [ 242.6875 , 6.577410211268018e-47 ], 
                    [ 242.75 , 6.577917109119253e-47 ], [ 242.8125 , 6.578424046035228e-47 ], [ 242.875 , 6.578931022019186e-47 ], [ 242.9375 , 6.579438037073953e-47 ], 
                    [ 243.0 , 6.57994509120249e-47 ], [ 243.0625 , 6.580452184408046e-47 ], [ 243.125 , 6.580959316693395e-47 ], [ 243.1875 , 6.581468698797965e-47 ], 
                    [ 243.25 , 6.58198110032196e-47 ], [ 243.3125 , 6.582493541739124e-47 ], [ 243.375 , 6.583006023052511e-47 ], [ 243.4375 , 6.583518544265186e-47 ], 
                    [ 243.5 , 6.584031105380299e-47 ], [ 243.5625 , 6.584543706401097e-47 ], [ 243.625 , 6.585056347330453e-47 ], [ 243.6875 , 6.585569028171522e-47 ], 
                    [ 243.75 , 6.586084882450471e-47 ], [ 243.8125 , 6.5866028030249405e-47 ], [ 243.875 , 6.587120764327872e-47 ], [ 243.9375 , 6.587638766362607e-47 ], 
                    [ 244.0 , 6.58815680913235e-47 ], [ 244.0625 , 6.588674892640258e-47 ], [ 244.125 , 6.589193016889488e-47 ], [ 244.1875 , 6.589711181883288e-47 ], 
                    [ 244.25 , 6.590229387624771e-47 ], [ 244.3125 , 6.590751567092072e-47 ], [ 244.375 , 6.591274970785808e-47 ], [ 244.4375 , 6.591798416045676e-47 ], 
                    [ 244.5 , 6.592321902874693e-47 ], [ 244.5625 , 6.592845431276538e-47 ], [ 244.625 , 6.593369001254231e-47 ], [ 244.6875 , 6.593892612811214e-47 ], 
                    [ 244.75 , 6.594416265950647e-47 ], [ 244.8125 , 6.59493996067607e-47 ], [ 244.875 , 6.595468317978013e-47 ], [ 244.9375 , 6.595997173328054e-47 ], 
                    [ 245.0 , 6.596526071084225e-47 ], [ 245.0625 , 6.597055011249974e-47 ], [ 245.125 , 6.597583993828514e-47 ], [ 245.1875 , 6.598113018823339e-47 ], 
                    [ 245.25 , 6.598642086237852e-47 ], [ 245.3125 , 6.599171196075452e-47 ], [ 245.375 , 6.599700507810263e-47 ], [ 245.4375 , 6.600234736720899e-47 ], 
                    [ 245.5 , 6.600769008876095e-47 ], [ 245.5625 , 6.601303324279212e-47 ], [ 245.625 , 6.60183768293375e-47 ], [ 245.6875 , 6.602372084843303e-47 ], 
                    [ 245.75 , 6.602906530011282e-47 ], [ 245.8125 , 6.603441018441279e-47 ], [ 245.875 , 6.603975550136748e-47 ], [ 245.9375 , 6.604510791457043e-47 ], 
                    [ 246.0 , 6.605050421857532e-47 ], [ 246.0625 , 6.605590096349245e-47 ], [ 246.125 , 6.606129814935833e-47 ], [ 246.1875 , 6.606669577620709e-47 ], 
                    [ 246.25 , 6.607209384407761e-47 ], [ 246.3125 , 6.6077492353004e-47 ], [ 246.375 , 6.608289130302328e-47 ], [ 246.4375 , 6.608829069417006e-47 ], 
                    [ 246.5 , 6.609370112405916e-47 ], [ 246.5625 , 6.609915094241673e-47 ], [ 246.625 , 6.61046012101442e-47 ], [ 246.6875 , 6.611005192727816e-47 ], 
                    [ 246.75 , 6.611550309385707e-47 ], [ 246.8125 , 6.612095470991753e-47 ], [ 246.875 , 6.612640677549376e-47 ], [ 246.9375 , 6.613185929062661e-47 ], 
                    [ 247.0 , 6.613731225535173e-47 ], [ 247.0625 , 6.614277921799076e-47 ], [ 247.125 , 6.614828240391868e-47 ], [ 247.1875 , 6.615378604772056e-47 ], 
                    [ 247.25 , 6.615929014943493e-47 ], [ 247.3125 , 6.616479470909898e-47 ], [ 247.375 , 6.617029972675128e-47 ], [ 247.4375 , 6.617580520242946e-47 ], 
                    [ 247.5 , 6.6181311136173965e-47 ], [ 247.5625 , 6.618681752801917e-47 ], [ 247.625 , 6.619233976508984e-47 ], [ 247.6875 , 6.619789587073996e-47 ], 
                    [ 247.75 , 6.62034524427634e-47 ], [ 247.8125 , 6.620900948119884e-47 ], [ 247.875 , 6.62145669860859e-47 ], [ 247.9375 , 6.622012495746231e-47 ], 
                    [ 248.0 , 6.622568339536911e-47 ], [ 248.0625 , 6.623124229984407e-47 ], [ 248.125 , 6.623680167092679e-47 ], [ 248.1875 , 6.624237782017421e-47 ], 
                    [ 248.25 , 6.624798672252643e-47 ], [ 248.3125 , 6.625359609979749e-47 ], [ 248.375 , 6.625920595202763e-47 ], [ 248.4375 , 6.626481627925752e-47 ], 
                    [ 248.5 , 6.627042708152833e-47 ], [ 248.5625 , 6.627603835887841e-47 ], [ 248.625 , 6.62816501113489e-47 ], [ 248.6875 , 6.628726233898147e-47 ], 
                    [ 248.75 , 6.629289114849434e-47 ], [ 248.8125 , 6.6298552337147955e-47 ], [ 248.875 , 6.630421400924745e-47 ], [ 248.9375 , 6.630987616483413e-47 ], 
                    [ 249.0 , 6.631553880395067e-47 ], [ 249.0625 , 6.632120192663792e-47 ], [ 249.125 , 6.632686553293527e-47 ], [ 249.1875 , 6.633252962288686e-47 ], 
                    [ 249.25 , 6.63381941965321e-47 ], [ 249.3125 , 6.634387433373881e-47 ], [ 249.375 , 6.634958774668058e-47 ], [ 249.4375 , 6.635530165165101e-47 ], 
                    [ 249.5 , 6.636101604869294e-47 ], [ 249.5625 , 6.636673093784732e-47 ], [ 249.625 , 6.637244631915797e-47 ], [ 249.6875 , 6.637816219266679e-47 ], 
                    [ 249.75 , 6.638387855841711e-47 ], [ 249.8125 , 6.638959541644943e-47 ], [ 249.875 , 6.639532575433034e-47 ], [ 249.9375 , 6.640109098815997e-47 ], 
                    [ 250.0 , 6.640685672259546e-47 ], [ 250.0625 , 6.641262295768029e-47 ], [ 250.125 , 6.641838969345934e-47 ], [ 250.1875 , 6.6424156929975155e-47 ], 
                    [ 250.25 , 6.642992466727026e-47 ], [ 250.3125 , 6.643569290539051e-47 ], [ 250.375 , 6.644146164437703e-47 ], [ 250.4375 , 6.644724087787089e-47 ], 
                    [ 250.5 , 6.645305758607797e-47 ], [ 250.5625 , 6.645887480347287e-47 ], [ 250.625 , 6.646469253009922e-47 ], [ 250.6875 , 6.647051076600254e-47 ], 
                    [ 250.75 , 6.647632951122647e-47 ], [ 250.8125 , 6.648214876581701e-47 ], [ 250.875 , 6.648796852981686e-47 ], [ 250.9375 , 6.649378880327109e-47 ], 
                    [ 251.0 , 6.649961567619872e-47 ], [ 251.0625 , 6.650548356124469e-47 ], [ 251.125 , 6.651135196407016e-47 ], [ 251.1875 , 6.651722088471987e-47 ], 
                    [ 251.25 , 6.652309032323905e-47 ], [ 251.3125 , 6.652896027967387e-47 ], [ 251.375 , 6.653483075406953e-47 ], [ 251.4375 , 6.654070174647367e-47 ], 
                    [ 251.5 , 6.654657325692864e-47 ], [ 251.5625 , 6.655244659583394e-47 ], [ 251.625 , 6.655836567910005e-47 ], [ 251.6875 , 6.656428528880175e-47 ], 
                    [ 251.75 , 6.657020542498495e-47 ], [ 251.8125 , 6.657612608769742e-47 ], [ 251.875 , 6.658204727698502e-47 ], [ 251.9375 , 6.658796899289508e-47 ], 
                    [ 252.0 , 6.65938912354749e-47 ], [ 252.0625 , 6.659981400477132e-47 ], [ 252.125 , 6.660573730082928e-47 ], [ 252.1875 , 6.661170254998573e-47 ], 
                    [ 252.25 , 6.661767267633242e-47 ], [ 252.3125 , 6.662364333775544e-47 ], [ 252.375 , 6.662961453430323e-47 ], [ 252.4375 , 6.663558626602468e-47 ], 
                    [ 252.5 , 6.664155853296731e-47 ], [ 252.5625 , 6.664753133517953e-47 ], [ 252.625 , 6.665350467270792e-47 ], [ 252.6875 , 6.665947854560234e-47 ], 
                    [ 252.75 , 6.666548778836521e-47 ], [ 252.8125 , 6.667150850674957e-47 ], [ 252.875 , 6.667752976887883e-47 ], [ 252.9375 , 6.668355157480304e-47 ], 
                    [ 253.0 , 6.668957392457083e-47 ], [ 253.0625 , 6.66955968182299e-47 ], [ 253.125 , 6.67016202558322e-47 ], [ 253.1875 , 6.670764423742544e-47 ], 
                    [ 253.25 , 6.671366876305828e-47 ], [ 253.3125 , 6.671972082352829e-47 ], [ 253.375 , 6.672579168184814e-47 ], [ 253.4375 , 6.673186309255824e-47 ], 
                    [ 253.5 , 6.673793505570832e-47 ], [ 253.5625 , 6.67440075713501e-47 ], [ 253.625 , 6.675008063953193e-47 ], [ 253.6875 , 6.6756154260306e-47 ], 
                    [ 253.75 , 6.676222843372164e-47 ], [ 253.8125 , 6.676830315982868e-47 ], [ 253.875 , 6.677439692527303e-47 ], [ 253.9375 , 6.678051772038344e-47 ], 
                    [ 254.0 , 6.6786639076550155e-47 ], [ 254.0625 , 6.679276099382366e-47 ], [ 254.125 , 6.679888347225393e-47 ], [ 254.1875 , 6.68050065118953e-47 ], 
                    [ 254.25 , 6.681113011279633e-47 ], [ 254.3125 , 6.681725427501132e-47 ], [ 254.375 , 6.682337899858984e-47 ], [ 254.4375 , 6.682951340197772e-47 ], 
                    [ 254.5 , 6.683568370731582e-47 ], [ 254.5625 , 6.684185458235139e-47 ], [ 254.625 , 6.684802602713892e-47 ], [ 254.6875 , 6.68541980417296e-47 ], 
                    [ 254.75 , 6.686037062617651e-47 ], [ 254.8125 , 6.686654378053227e-47 ], [ 254.875 , 6.687271750485044e-47 ], [ 254.9375 , 6.687889179918127e-47 ], 
                    [ 255.0 , 6.688506666357929e-47 ], [ 255.0625 , 6.689128547116138e-47 ], [ 255.125 , 6.689750582908727e-47 ], [ 255.1875 , 6.690372676545797e-47 ], 
                    [ 255.25 , 6.690994828032582e-47 ], [ 255.3125 , 6.691617037374561e-47 ], [ 255.375 , 6.692239304576966e-47 ], [ 255.4375 , 6.692861629645324e-47 ], 
                    [ 255.5 , 6.693484012584967e-47 ], [ 255.5625 , 6.694106453401325e-47 ], [ 255.625 , 6.69473214680905e-47 ], [ 255.6875 , 6.695359079936065e-47 ], 
                    [ 255.75 , 6.695986071772646e-47 ], [ 255.8125 , 6.696613122324435e-47 ], [ 255.875 , 6.697240231596742e-47 ], [ 255.9375 , 6.697867399595205e-47 ], 
                    [ 256.0 , 6.69849462632523e-47 ], [ 256.0625 , 6.699121911792365e-47 ], [ 256.125 , 6.699749256002205e-47 ], [ 256.1875 , 6.700378675279176e-47 ], 
                    [ 256.25 , 6.701010510247025e-47 ], [ 256.3125 , 6.701642404795947e-47 ], [ 256.375 , 6.702274358931512e-47 ], [ 256.4375 , 6.70290637265929e-47 ], 
                    [ 256.5 , 6.703538445984903e-47 ], [ 256.5625 , 6.70417057891416e-47 ], [ 256.625 , 6.704802771452397e-47 ], [ 256.6875 , 6.705435023605376e-47 ], 
                    [ 256.75 , 6.706068074674886e-47 ], [ 256.8125 , 6.706704764870909e-47 ], [ 256.875 , 6.707341515515873e-47 ], [ 256.9375 , 6.707978326615425e-47 ], 
                    [ 257.0 , 6.708615198175447e-47 ], [ 257.0625 , 6.709252130201534e-47 ], [ 257.125 , 6.709889122699573e-47 ], [ 257.1875 , 6.710526175675064e-47 ], 
                    [ 257.25 , 6.711163289133894e-47 ], [ 257.3125 , 6.711800463081853e-47 ], [ 257.375 , 6.712441389778866e-47 ], [ 257.4375 , 6.71308297720472e-47 ], 
                    [ 257.5 , 6.713724625954673e-47 ], [ 257.5625 , 6.714366336034587e-47 ], [ 257.625 , 6.715008107450277e-47 ], [ 257.6875 , 6.7156499402077e-47 ], 
                    [ 257.75 , 6.7162918343127195e-47 ], [ 257.8125 , 6.716933789771057e-47 ], [ 257.875 , 6.717575806588719e-47 ], [ 257.9375 , 6.718220127658796e-47 ], 
                    [ 258.0 , 6.718866514440722e-47 ], [ 258.0625 , 6.719512963414102e-47 ], [ 258.125 , 6.720159474584822e-47 ], [ 258.1875 , 6.720806047959107e-47 ], 
                    [ 258.25 , 6.721452683542751e-47 ], [ 258.3125 , 6.722099381341689e-47 ], [ 258.375 , 6.722746141362149e-47 ], [ 258.4375 , 6.723392963609877e-47 ], 
                    [ 258.5 , 6.724040590560458e-47 ], [ 258.5625 , 6.724691783356666e-47 ], [ 258.625 , 6.725343039217988e-47 ], [ 258.6875 , 6.725994358150435e-47 ], 
                    [ 258.75 , 6.726645740160162e-47 ], [ 258.8125 , 6.727297185253279e-47 ], [ 258.875 , 6.727948693435894e-47 ], [ 258.9375 , 6.728600264714119e-47 ], 
                    [ 259.0 , 6.729251899094159e-47 ], [ 259.0625 , 6.729903596581887e-47 ], [ 259.125 , 6.730558707156205e-47 ], [ 259.1875 , 6.731214716564846e-47 ], 
                    [ 259.25 , 6.731870789913004e-47 ], [ 259.3125 , 6.732526927206813e-47 ], [ 259.375 , 6.733183128452556e-47 ], [ 259.4375 , 6.733839393656369e-47 ], 
                    [ 259.5 , 6.734495722824821e-47 ], [ 259.5625 , 6.735152115963716e-47 ], [ 259.625 , 6.735808573079576e-47 ], [ 259.6875 , 6.736466776802575e-47 ], 
                    [ 259.75 , 6.737127519141832e-47 ], [ 259.8125 , 6.737788326289702e-47 ], [ 259.875 , 6.738449198252354e-47 ], [ 259.9375 , 6.739110135036287e-47 ], 
                    [ 260.0 , 6.739771136647812e-47 ], [ 260.0625 , 6.740432203093239e-47 ], [ 260.125 , 6.741093334378927e-47 ], [ 260.1875 , 6.741754530511429e-47 ], 
                    [ 260.25 , 6.742415791496912e-47 ], [ 260.3125 , 6.743081194518851e-47 ], [ 260.375 , 6.743746713402669e-47 ], [ 260.4375 , 6.744412297970812e-47 ], 
                    [ 260.5 , 6.745077948229858e-47 ], [ 260.5625 , 6.745743664186292e-47 ], [ 260.625 , 6.746409445846598e-47 ], [ 260.6875 , 6.7470752932173065e-47 ], 
                    [ 260.75 , 6.747741206304954e-47 ], [ 260.8125 , 6.748407185115737e-47 ], [ 260.875 , 6.749075490274869e-47 ], [ 260.9375 , 6.749745715587536e-47 ], 
                    [ 261.0 , 6.750416007457761e-47 ], [ 261.0625 , 6.751086365892197e-47 ], [ 261.125 , 6.751756790897314e-47 ], [ 261.1875 , 6.752427282479912e-47 ], 
                    [ 261.25 , 6.753097840646414e-47 ], [ 261.3125 , 6.753768465403622e-47 ], [ 261.375 , 6.754439156758102e-47 ], [ 261.4375 , 6.755110278229759e-47 ], 
                    [ 261.5 , 6.75578515307675e-47 ], [ 261.5625 , 6.756460095347669e-47 ], [ 261.625 , 6.7571351050492495e-47 ], [ 261.6875 , 6.757810182188181e-47 ], 
                    [ 261.75 , 6.7584853267712965e-47 ], [ 261.8125 , 6.759160538805336e-47 ], [ 261.875 , 6.75983581829694e-47 ], [ 261.9375 , 6.760511165252802e-47 ], 
                    [ 262.0 , 6.7611865796799e-47 ], [ 262.0625 , 6.761864518118188e-47 ], [ 262.125 , 6.762544104260776e-47 ], [ 262.1875 , 6.763223758703562e-47 ], 
                    [ 262.25 , 6.763903481453552e-47 ], [ 262.3125 , 6.764583272517661e-47 ], [ 262.375 , 6.765263131902562e-47 ], [ 262.4375 , 6.765943059615217e-47 ], 
                    [ 262.5 , 6.766623055662541e-47 ], [ 262.5625 , 6.767303120051307e-47 ], [ 262.625 , 6.767983678346377e-47 ], [ 262.6875 , 6.768667897616271e-47 ], 
                    [ 262.75 , 6.769352186058367e-47 ], [ 262.8125 , 6.770036543679562e-47 ], [ 262.875 , 6.770720970486898e-47 ], [ 262.9375 , 6.771405466487419e-47 ], 
                    [ 263.0 , 6.77209003168802e-47 ], [ 263.0625 , 6.772774666095651e-47 ], [ 263.125 , 6.773459369717599e-47 ], [ 263.1875 , 6.774144142560667e-47 ], 
                    [ 263.25 , 6.774831293773452e-47 ], [ 263.3125 , 6.775520174950493e-47 ], [ 263.375 , 6.776209126174632e-47 ], [ 263.4375 , 6.776898147453039e-47 ], 
                    [ 263.5 , 6.777587238792695e-47 ], [ 263.5625 , 6.778276400200916e-47 ], [ 263.625 , 6.778965631684682e-47 ], [ 263.6875 , 6.779654933251167e-47 ], 
                    [ 263.75 , 6.780344304907546e-47 ], [ 263.8125 , 6.78103388841753e-47 ], [ 263.875 , 6.781727338378022e-47 ], [ 263.9375 , 6.782420859252898e-47 ], 
                    [ 264.0 , 6.783114451049364e-47 ], [ 264.0625 , 6.783808113774766e-47 ], [ 264.125 , 6.784501847436264e-47 ], [ 264.1875 , 6.785195652041207e-47 ], 
                    [ 264.25 , 6.785889527596803e-47 ], [ 264.3125 , 6.786583474110306e-47 ], [ 264.375 , 6.787277491588925e-47 ], [ 264.4375 , 6.787973416560724e-47 ], 
                    [ 264.5 , 6.788671496750831e-47 ], [ 264.5625 , 6.789369648731923e-47 ], [ 264.625 , 6.79006787251143e-47 ], [ 264.6875 , 6.790766168096931e-47 ], 
                    [ 264.75 , 6.791464535495616e-47 ], [ 264.8125 , 6.792162974714871e-47 ], [ 264.875 , 6.792861485762178e-47 ], [ 264.9375 , 6.793560068644925e-47 ], 
                    [ 265.0 , 6.794258723370403e-47 ], [ 265.0625 , 6.794960873401473e-47 ], [ 265.125 , 6.795663565710852e-47 ], [ 265.1875 , 6.796366330688271e-47 ], 
                    [ 265.25 , 6.797069168341148e-47 ], [ 265.3125 , 6.797772078677191e-47 ], [ 265.375 , 6.798475061703868e-47 ], [ 265.4375 , 6.799178117428647e-47 ], 
                    [ 265.5 , 6.799881245858952e-47 ], [ 265.5625 , 6.800584447002349e-47 ], [ 265.625 , 6.801288754610931e-47 ], [ 265.6875 , 6.801995949775106e-47 ], 
                    [ 265.75 , 6.802703218473091e-47 ], [ 265.8125 , 6.803410560712629e-47 ], [ 265.875 , 6.804117976501317e-47 ], [ 265.9375 , 6.804825465846755e-47 ], 
                    [ 266.0 , 6.805533028756688e-47 ], [ 266.0625 , 6.806240665238716e-47 ], [ 266.125 , 6.80694837530049e-47 ], [ 266.1875 , 6.807656158949663e-47 ], 
                    [ 266.25 , 6.808366441512116e-47 ], [ 266.3125 , 6.80907818640299e-47 ], [ 266.375 , 6.809790005699495e-47 ], [ 266.4375 , 6.810501899409406e-47 ], 
                    [ 266.5 , 6.811213867540554e-47 ], [ 266.5625 , 6.811925910100619e-47 ], [ 266.625 , 6.812638027097432e-47 ], [ 266.6875 , 6.813350218538775e-47 ], 
                    [ 266.75 , 6.814062484432379e-47 ], [ 266.8125 , 6.814774824786176e-47 ], [ 266.875 , 6.815490980625982e-47 ], [ 266.9375 , 6.816207282479595e-47 ], 
                    [ 267.0 , 6.816923659615779e-47 ], [ 267.0625 , 6.817640112042592e-47 ], [ 267.125 , 6.818356639767946e-47 ], [ 267.1875 , 6.819073242799562e-47 ], 
                    [ 267.25 , 6.819789921145548e-47 ], [ 267.3125 , 6.820506674813868e-47 ], [ 267.375 , 6.821223503812294e-47 ], [ 267.4375 , 6.821941556960546e-47 ], 
                    [ 267.5 , 6.822662302275385e-47 ], [ 267.5625 , 6.823383123737871e-47 ], [ 267.625 , 6.824104021355808e-47 ], [ 267.6875 , 6.824824995137336e-47 ], 
                    [ 267.75 , 6.825546045090555e-47 ], [ 267.8125 , 6.82626717122341e-47 ], [ 267.875 , 6.826988373544002e-47 ], [ 267.9375 , 6.827709652060429e-47 ], 
                    [ 268.0 , 6.828431006780545e-47 ], [ 268.0625 , 6.829154702741272e-47 ], [ 268.125 , 6.829879942779762e-47 ], [ 268.1875 , 6.830605259836927e-47 ], 
                    [ 268.25 , 6.831330653921094e-47 ], [ 268.3125 , 6.832056125040392e-47 ], [ 268.375 , 6.832781673203051e-47 ], [ 268.4375 , 6.833507298417107e-47 ], 
                    [ 268.5 , 6.834233000690938e-47 ], [ 268.5625 , 6.834958780032582e-47 ], [ 268.625 , 6.835684636450222e-47 ], [ 268.6875 , 6.83641387254154e-47 ], 
                    [ 268.75 , 6.837143609056975e-47 ], [ 268.8125 , 6.83787342346657e-47 ], [ 268.875 , 6.838603315778347e-47 ], [ 268.9375 , 6.839333286000718e-47 ], 
                    [ 269.0 , 6.84006333414205e-47 ], [ 269.0625 , 6.840793460210465e-47 ], [ 269.125 , 6.841523664214621e-47 ], [ 269.1875 , 6.8422539461625945e-47 ], 
                    [ 269.25 , 6.842984836697103e-47 ], [ 269.3125 , 6.843718951798442e-47 ], [ 269.375 , 6.844453145655616e-47 ], [ 269.4375 , 6.845187418277074e-47 ], 
                    [ 269.5 , 6.845921769671265e-47 ], [ 269.5625 , 6.846656199846639e-47 ], [ 269.625 , 6.847390708811748e-47 ], [ 269.6875 , 6.848125296574846e-47 ], 
                    [ 269.75 , 6.848859963144536e-47 ], [ 269.8125 , 6.849594708529173e-47 ], [ 269.875 , 6.850330918165293e-47 ], [ 269.9375 , 6.851069467609763e-47 ], 
                    [ 270.0 , 6.851808096678845e-47 ], [ 270.0625 , 6.852546805381222e-47 ], [ 270.125 , 6.853285593725382e-47 ], [ 270.1875 , 6.854024461720008e-47 ], 
                    [ 270.25 , 6.854763409373592e-47 ], [ 270.3125 , 6.855502436694866e-47 ], [ 270.375 , 6.856241543692226e-47 ], [ 270.4375 , 6.856980730374406e-47 ], 
                    [ 270.5 , 6.85772215075882e-47 ], [ 270.5625 , 6.858465127160901e-47 ], [ 270.625 , 6.859208184058213e-47 ], [ 270.6875 , 6.859951321459577e-47 ], 
                    [ 270.75 , 6.860694539373468e-47 ], [ 270.8125 , 6.861437837808903e-47 ], [ 270.875 , 6.862181216774314e-47 ], [ 270.9375 , 6.862924676278716e-47 ], 
                    [ 271.0 , 6.863668216330545e-47 ], [ 271.0625 , 6.864411836938672e-47 ], [ 271.125 , 6.865158366248045e-47 ], [ 271.1875 , 6.86590575233271e-47 ], 
                    [ 271.25 , 6.866653219782719e-47 ], [ 271.3125 , 6.867400768606976e-47 ], [ 271.375 , 6.868148398814244e-47 ], [ 271.4375 , 6.868896110413431e-47 ], 
                    [ 271.5 , 6.869643903413544e-47 ], [ 271.5625 , 6.870391777823152e-47 ], [ 271.625 , 6.871139733651265e-47 ], [ 271.6875 , 6.871887770906794e-47 ], 
                    [ 271.75 , 6.872639290684802e-47 ], [ 271.8125 , 6.873391057558871e-47 ], [ 271.875 , 6.874142906665458e-47 ], [ 271.9375 , 6.874894838013314e-47 ], 
                    [ 272.0 , 6.875646851611535e-47 ], [ 272.0625 , 6.876398947469164e-47 ], [ 272.125 , 6.877151125595053e-47 ], [ 272.1875 , 6.87790338599835e-47 ], 
                    [ 272.25 , 6.878655728688102e-47 ], [ 272.3125 , 6.879408511593061e-47 ], [ 272.375 , 6.880164575749991e-47 ], [ 272.4375 , 6.880920723000307e-47 ], 
                    [ 272.5 , 6.881676953353041e-47 ], [ 272.5625 , 6.88243326681738e-47 ], [ 272.625 , 6.883189663402454e-47 ], [ 272.6875 , 6.883946143117547e-47 ], 
                    [ 272.75 , 6.884702705971403e-47 ], [ 272.8125 , 6.88545935197355e-47 ], [ 272.875 , 6.886216081132932e-47 ], [ 272.9375 , 6.886973680183078e-47 ], 
                    [ 273.0 , 6.88773407084455e-47 ], [ 273.0625 , 6.888494545460842e-47 ], [ 273.125 , 6.889255104041076e-47 ], [ 273.1875 , 6.89001574659457e-47 ], 
                    [ 273.25 , 6.890776473130648e-47 ], [ 273.3125 , 6.891537283658432e-47 ], [ 273.375 , 6.892298178187394e-47 ], [ 273.4375 , 6.89305915672666e-47 ], 
                    [ 273.5 , 6.893820219285507e-47 ], [ 273.5625 , 6.894582510291214e-47 ], [ 273.625 , 6.895347239040521e-47 ], [ 273.6875 , 6.896112052611506e-47 ], 
                    [ 273.75 , 6.896876951013627e-47 ], [ 273.8125 , 6.897641934256145e-47 ], [ 273.875 , 6.898407002348619e-47 ], [ 273.9375 , 6.89917215530046e-47 ], 
                    [ 274.0 , 6.899937393120931e-47 ], [ 274.0625 , 6.900702715819546e-47 ], [ 274.125 , 6.901468123405867e-47 ], [ 274.1875 , 6.902235030322574e-47 ], 
                    [ 274.25 , 6.903004085915213e-47 ], [ 274.3125 , 6.903773227197029e-47 ], [ 274.375 , 6.904542454177522e-47 ], [ 274.4375 , 6.905311766866142e-47 ], 
                    [ 274.5 , 6.906081165272584e-47 ], [ 274.5625 , 6.906850649406401e-47 ], [ 274.625 , 6.9076202192769975e-47 ], [ 274.6875 , 6.908389874894221e-47 ], 
                    [ 274.75 , 6.909159616267329e-47 ], [ 274.8125 , 6.90993103128581e-47 ], [ 274.875 , 6.910704366913844e-47 ], [ 274.9375 , 6.91147778909103e-47 ], 
                    [ 275.0 , 6.912251297826907e-47 ], [ 275.0625 , 6.913024893131212e-47 ], [ 275.125 , 6.913798575013584e-47 ], [ 275.1875 , 6.914572343483958e-47 ], 
                    [ 275.25 , 6.91534619855168e-47 ], [ 275.3125 , 6.916120140226542e-47 ], [ 275.375 , 6.916894168518283e-47 ], [ 275.4375 , 6.917669986621466e-47 ], 
                    [ 275.5 , 6.918447604958604e-47 ], [ 275.5625 , 6.919225310708164e-47 ], [ 275.625 , 6.920003103879823e-47 ], [ 275.6875 , 6.920780984483701e-47 ], 
                    [ 275.75 , 6.921558952529434e-47 ], [ 275.8125 , 6.922337008026898e-47 ], [ 275.875 , 6.923115150986022e-47 ], [ 275.9375 , 6.923893381416393e-47 ], 
                    [ 276.0 , 6.924671699328087e-47 ], [ 276.0625 , 6.925451834850776e-47 ], [ 276.125 , 6.926233717002269e-47 ], [ 276.1875 , 6.927015687428152e-47 ], 
                    [ 276.25 , 6.927797746138341e-47 ], [ 276.3125 , 6.928579893142856e-47 ], [ 276.375 , 6.929362128451514e-47 ], [ 276.4375 , 6.930144452074484e-47 ], 
                    [ 276.5 , 6.930926864021539e-47 ], [ 276.5625 , 6.931709364302798e-47 ], [ 276.625 , 6.932491952928088e-47 ], [ 276.6875 , 6.933276304268751e-47 ], 
                    [ 276.75 , 6.934062418902985e-47 ], [ 276.8125 , 6.934848622669089e-47 ], [ 276.875 , 6.935634915577217e-47 ], [ 276.9375 , 6.936421297637526e-47 ], 
                    [ 277.0 , 6.937207768860124e-47 ], [ 277.0625 , 6.937994329254973e-47 ], [ 277.125 , 6.938780978832282e-47 ], [ 277.1875 , 6.939567717602162e-47 ], 
                    [ 277.25 , 6.940354545574827e-47 ], [ 277.3125 , 6.941143024989587e-47 ], [ 277.375 , 6.941933379207827e-47 ], [ 277.4375 , 6.942723823419714e-47 ], 
                    [ 277.5 , 6.9435143576356895e-47 ], [ 277.5625 , 6.944304981866052e-47 ], [ 277.625 , 6.945095696120855e-47 ], [ 277.6875 , 6.945886500410546e-47 ], 
                    [ 277.75 , 6.946677394745079e-47 ], [ 277.8125 , 6.947468379135054e-47 ], [ 277.875 , 6.948259453590479e-47 ], [ 277.9375 , 6.94905197270232e-47 ], 
                    [ 278.0 , 6.949846525913681e-47 ], [ 278.0625 , 6.950641169974247e-47 ], [ 278.125 , 6.951435904894058e-47 ], [ 278.1875 , 6.95223073068375e-47 ], 
                    [ 278.25 , 6.953025647353762e-47 ], [ 278.3125 , 6.953820654914288e-47 ], [ 278.375 , 6.954615753375819e-47 ], [ 278.4375 , 6.9554109427488e-47 ], 
                    [ 278.5 , 6.956206223043574e-47 ], [ 278.5625 , 6.957002680247896e-47 ], [ 278.625 , 6.957801415510037e-47 ], [ 278.6875 , 6.958600242475177e-47 ], 
                    [ 278.75 , 6.959399161153896e-47 ], [ 278.8125 , 6.960198171556623e-47 ], [ 278.875 , 6.96099727369394e-47 ], [ 278.9375 , 6.961796467576428e-47 ], 
                    [ 279.0 , 6.962595753214521e-47 ], [ 279.0625 , 6.963395130618902e-47 ], [ 279.125 , 6.964194599799908e-47 ], [ 279.1875 , 6.964994914622627e-47 ], 
                    [ 279.25 , 6.965797844340673e-47 ], [ 279.3125 , 6.966600866621047e-47 ], [ 279.375 , 6.967403981474467e-47 ], [ 279.4375 , 6.968207188911507e-47 ], 
                    [ 279.5 , 6.969010488942988e-47 ], [ 279.5625 , 6.969813881579286e-47 ], [ 279.625 , 6.970617366831425e-47 ], [ 279.6875 , 6.971420944709884e-47 ], 
                    [ 279.75 , 6.972224615225488e-47 ], [ 279.8125 , 6.973028718465288e-47 ], [ 279.875 , 6.973835795120638e-47 ], [ 279.9375 , 6.974642965188961e-47 ], 
                    [ 280.0 , 6.975450228681516e-47 ], [ 280.0625 , 6.976257585608669e-47 ], [ 280.125 , 6.977065035981483e-47 ], [ 280.1875 , 6.977872579810723e-47 ], 
                    [ 280.25 , 6.978680217107205e-47 ], [ 280.3125 , 6.979487947881801e-47 ], [ 280.375 , 6.980295772145227e-47 ], [ 280.4375 , 6.981103689908404e-47 ], 
                    [ 280.5 , 6.981914780280637e-47 ], [ 280.5625 , 6.98272609962744e-47 ], [ 280.625 , 6.983537513252023e-47 ], [ 280.6875 , 6.984349021165193e-47 ], 
                    [ 280.75 , 6.985160623378056e-47 ], [ 280.8125 , 6.98597231990147e-47 ], [ 280.875 , 6.986784110746494e-47 ], [ 280.9375 , 6.987595995924088e-47 ], 
                    [ 281.0 , 6.988407975445113e-47 ], [ 281.0625 , 6.989220049320535e-47 ], [ 281.125 , 6.990034720039e-47 ], [ 281.1875 , 6.990850160986305e-47 ], 
                    [ 281.25 , 6.991665697061074e-47 ], [ 281.3125 , 6.992481328274209e-47 ], [ 281.375 , 6.993297054636954e-47 ], [ 281.4375 , 6.9941128761605105e-47 ], 
                    [ 281.5 , 6.99492879285573e-47 ], [ 281.5625 , 6.995744804734065e-47 ], [ 281.625 , 6.996560911806269e-47 ], [ 281.6875 , 6.9973771140837475e-47 ], 
                    [ 281.75 , 6.998195287274249e-47 ], [ 281.8125 , 6.999014838747773e-47 ], [ 281.875 , 6.999834486198131e-47 ], [ 281.9375 , 7.000654229636611e-47 ], 
                    [ 282.0 , 7.001474069074355e-47 ], [ 282.0625 , 7.002294004522654e-47 ], [ 282.125 , 7.003114035992802e-47 ], [ 282.1875 , 7.003934163496046e-47 ], 
                    [ 282.25 , 7.004754387043529e-47 ], [ 282.3125 , 7.005574706646502e-47 ], [ 282.375 , 7.006396306400746e-47 ], [ 282.4375 , 7.007219957132405e-47 ], 
                    [ 282.5 , 7.008043704689949e-47 ], [ 282.5625 , 7.008867549084761e-47 ], [ 282.625 , 7.009691490328275e-47 ], [ 282.6875 , 7.010515528431826e-47 ], 
                    [ 282.75 , 7.011339663406801e-47 ], [ 282.8125 , 7.01216389526444e-47 ], [ 282.875 , 7.012988224016478e-47 ], [ 282.9375 , 7.013812649673958e-47 ], 
                    [ 283.0 , 7.014637599674195e-47 ], [ 283.0625 , 7.015465329500722e-47 ], [ 283.125 , 7.016293156999725e-47 ], [ 283.1875 , 7.017121082182629e-47 ], 
                    [ 283.25 , 7.017949105061059e-47 ], [ 283.3125 , 7.018777225646446e-47 ], [ 283.375 , 7.019605443950467e-47 ], [ 283.4375 , 7.020433759984504e-47 ], 
                    [ 283.5 , 7.021262173760089e-47 ], [ 283.5625 , 7.022090685288755e-47 ], [ 283.625 , 7.022919294582236e-47 ], [ 283.6875 , 7.023750693025232e-47 ], 
                    [ 283.75 , 7.024582576530452e-47 ], [ 283.8125 , 7.025414558562927e-47 ], [ 283.875 , 7.026246639134075e-47 ], [ 283.9375 , 7.027078818255867e-47 ], 
                    [ 284.0 , 7.027911095939776e-47 ], [ 284.0625 , 7.028743472197524e-47 ], [ 284.125 , 7.0295759470407885e-47 ], [ 284.1875 , 7.030408520481243e-47 ], 
                    [ 284.25 , 7.031241192530518e-47 ], [ 284.3125 , 7.032075766027368e-47 ], [ 284.375 , 7.032911701939404e-47 ], [ 284.4375 , 7.033747737223022e-47 ], 
                    [ 284.5 , 7.034583871890084e-47 ], [ 284.5625 , 7.035420105952459e-47 ], [ 284.625 , 7.036256439421806e-47 ], [ 284.6875 , 7.037092872310246e-47 ], 
                    [ 284.75 , 7.037929404629197e-47 ], [ 284.8125 , 7.038766036390727e-47 ], [ 284.875 , 7.03960276760671e-47 ], [ 284.9375 , 7.040440445904788e-47 ], 
                    [ 285.0 , 7.041280421126445e-47 ], [ 285.0625 , 7.042120496563242e-47 ], [ 285.125 , 7.042960672226986e-47 ], [ 285.1875 , 7.043800948129833e-47 ], 
                    [ 285.25 , 7.044641324283544e-47 ], [ 285.3125 , 7.045481800700229e-47 ], [ 285.375 , 7.04632237739175e-47 ], [ 285.4375 , 7.047163054370171e-47 ], 
                    [ 285.5 , 7.048003831647457e-47 ], [ 285.5625 , 7.048844709235423e-47 ], [ 285.625 , 7.049688533808001e-47 ], [ 285.6875 , 7.050532623692071e-47 ], 
                    [ 285.75 , 7.051376814642797e-47 ], [ 285.8125 , 7.052221106672129e-47 ], [ 285.875 , 7.053065499792268e-47 ], [ 285.9375 , 7.05390999401522e-47 ], 
                    [ 286.0 , 7.054754589353142e-47 ], [ 286.0625 , 7.055599285818137e-47 ], [ 286.125 , 7.056444083422316e-47 ], [ 286.1875 , 7.057288982177838e-47 ], 
                    [ 286.25 , 7.05813574324212e-47 ], [ 286.3125 , 7.058983835169568e-47 ], [ 286.375 , 7.059832029002001e-47 ], [ 286.4375 , 7.060680324751912e-47 ], 
                    [ 286.5 , 7.061528722431249e-47 ], [ 286.5625 , 7.06237722205256e-47 ], [ 286.625 , 7.063225823627841e-47 ], [ 286.6875 , 7.064074527169446e-47 ], 
                    [ 286.75 , 7.064923332689523e-47 ], [ 286.8125 , 7.06577224020058e-47 ], [ 286.875 , 7.066621873958492e-47 ], [ 286.9375 , 7.067473971020832e-47 ], 
                    [ 287.0 , 7.068326170829444e-47 ], [ 287.0625 , 7.069178473396767e-47 ], [ 287.125 , 7.07003087873519e-47 ], [ 287.1875 , 7.070883386857208e-47 ], 
                    [ 287.25 , 7.071735997774963e-47 ], [ 287.3125 , 7.0725887115011e-47 ], [ 287.375 , 7.073441528047817e-47 ], [ 287.4375 , 7.074294447427711e-47 ], 
                    [ 287.5 , 7.075147469653032e-47 ], [ 287.5625 , 7.076002948153553e-47 ], [ 287.625 , 7.076859097486477e-47 ], [ 287.6875 , 7.07771535040778e-47 ], 
                    [ 287.75 , 7.078571706929947e-47 ], [ 287.8125 , 7.079428167065611e-47 ], [ 287.875 , 7.08028473082726e-47 ], [ 287.9375 , 7.081141398227383e-47 ], 
                    [ 288.0 , 7.081998169278566e-47 ], [ 288.0625 , 7.082855043993355e-47 ], [ 288.125 , 7.083712022384339e-47 ], [ 288.1875 , 7.084570212954018e-47 ], 
                    [ 288.25 , 7.085430326161961e-47 ], [ 288.3125 , 7.08629054379333e-47 ], [ 288.375 , 7.087150865860799e-47 ], [ 288.4375 , 7.088011292376901e-47 ], 
                    [ 288.5 , 7.088871823354515e-47 ], [ 288.5625 , 7.089732458806175e-47 ], [ 288.625 , 7.090593198744663e-47 ], [ 288.6875 , 7.091454043182565e-47 ], 
                    [ 288.75 , 7.092314992132618e-47 ], [ 288.8125 , 7.093176045607459e-47 ], [ 288.875 , 7.094039916345266e-47 ], [ 288.9375 , 7.094904102788723e-47 ], 
                    [ 289.0 , 7.095768394506215e-47 ], [ 289.0625 , 7.096632791510566e-47 ], [ 289.125 , 7.097497293814551e-47 ], [ 289.1875 , 7.098361901431052e-47 ], 
                    [ 289.25 , 7.099226614372994e-47 ], [ 289.3125 , 7.100091432653009e-47 ], [ 289.375 , 7.100956356283977e-47 ], [ 289.4375 , 7.101821385278937e-47 ], 
                    [ 289.5 , 7.102687814896633e-47 ], [ 289.5625 , 7.103555909637976e-47 ], [ 289.625 , 7.104424110478428e-47 ], [ 289.6875 , 7.105292417430904e-47 ], 
                    [ 289.75 , 7.106160830508425e-47 ], [ 289.8125 , 7.107029349723858e-47 ], [ 289.875 , 7.1078979750902795e-47 ], [ 289.9375 , 7.10876670662051e-47 ], 
                    [ 290.0 , 7.109635544327677e-47 ], [ 290.0625 , 7.110504488224757e-47 ], [ 290.125 , 7.111373538324477e-47 ], [ 290.1875 , 7.11224542887264e-47 ], 
                    [ 290.25 , 7.11311755313557e-47 ], [ 290.3125 , 7.113989784340869e-47 ], [ 290.375 , 7.1148621225018e-47 ], [ 290.4375 , 7.115734567631229e-47 ], 
                    [ 290.5 , 7.116607119742472e-47 ], [ 290.5625 , 7.117479778848698e-47 ], [ 290.625 , 7.118352544962878e-47 ], [ 290.6875 , 7.119225418098182e-47 ], 
                    [ 290.75 , 7.120098398267684e-47 ], [ 290.8125 , 7.12097272094563e-47 ], [ 290.875 , 7.121848763984225e-47 ], [ 290.9375 , 7.122724914796266e-47 ], 
                    [ 291.0 , 7.123601173394961e-47 ], [ 291.0625 , 7.124477539793469e-47 ], [ 291.125 , 7.125354014005304e-47 ], [ 291.1875 , 7.126230596043529e-47 ], 
                    [ 291.25 , 7.127107285921405e-47 ], [ 291.3125 , 7.127984083652203e-47 ], [ 291.375 , 7.128860989249443e-47 ], [ 291.4375 , 7.12973800272599e-47 ], 
                    [ 291.5 , 7.130617603499307e-47 ], [ 291.5625 , 7.131497625042818e-47 ], [ 291.625 , 7.132377755193689e-47 ], [ 291.6875 , 7.133257993965374e-47 ], 
                    [ 291.75 , 7.13413834137133e-47 ], [ 291.8125 , 7.135018797424863e-47 ], [ 291.875 , 7.135899362139431e-47 ], [ 291.9375 , 7.136780035528447e-47 ], 
                    [ 292.0 , 7.13766081760537e-47 ], [ 292.0625 , 7.138541708383565e-47 ], [ 292.125 , 7.139423600672274e-47 ], [ 292.1875 , 7.140307513357375e-47 ], 
                    [ 292.25 , 7.141191535477359e-47 ], [ 292.3125 , 7.142075667045624e-47 ], [ 292.375 , 7.142959908075975e-47 ], [ 292.4375 , 7.14384425858171e-47 ], 
                    [ 292.5 , 7.144728718576483e-47 ], [ 292.5625 , 7.145613288073952e-47 ], [ 292.625 , 7.146497967087422e-47 ], [ 292.6875 , 7.147382755630602e-47 ], 
                    [ 292.75 , 7.148267653717206e-47 ], [ 292.8125 , 7.149154665911061e-47 ], [ 292.875 , 7.150042548766076e-47 ], [ 292.9375 , 7.150930541890806e-47 ], 
                    [ 293.0 , 7.151818645299148e-47 ], [ 293.0625 , 7.152706859004748e-47 ], [ 293.125 , 7.153595183021202e-47 ], [ 293.1875 , 7.154483617362159e-47 ], 
                    [ 293.25 , 7.155372162041473e-47 ], [ 293.3125 , 7.156260817072898e-47 ], [ 293.375 , 7.157149582469988e-47 ], [ 293.4375 , 7.158038751377023e-47 ], 
                    [ 293.5 , 7.158930468274836e-47 ], [ 293.5625 , 7.15982229625891e-47 ], [ 293.625 , 7.160714235342827e-47 ], [ 293.6875 , 7.161606285540683e-47 ], 
                    [ 293.75 , 7.162498446866168e-47 ], [ 293.8125 , 7.163390719333177e-47 ], [ 293.875 , 7.164283102955554e-47 ], [ 293.9375 , 7.165175597747248e-47 ], 
                    [ 294.0 , 7.166068203721855e-47 ], [ 294.0625 , 7.166960920893377e-47 ], [ 294.125 , 7.167855025725754e-47 ], [ 294.1875 , 7.168750707350176e-47 ], 
                    [ 294.25 , 7.169646500897385e-47 ], [ 294.3125 , 7.17054240638116e-47 ], [ 294.375 , 7.171438423815799e-47 ], [ 294.4375 , 7.172334553214826e-47 ], 
                    [ 294.5 , 7.173230794592695e-47 ], [ 294.5625 , 7.174127147963091e-47 ], [ 294.625 , 7.175023613340112e-47 ], [ 294.6875 , 7.175920190737803e-47 ], 
                    [ 294.75 , 7.176816880170063e-47 ], [ 294.8125 , 7.177715844741877e-47 ], [ 294.875 , 7.1786154632990565e-47 ], [ 294.9375 , 7.179515194609819e-47 ], 
                    [ 295.0 , 7.180415038688351e-47 ], [ 295.0625 , 7.18131499554878e-47 ], [ 295.125 , 7.182215065205296e-47 ], [ 295.1875 , 7.183115247671934e-47 ], 
                    [ 295.25 , 7.1840155429629875e-47 ], [ 295.3125 , 7.184915951092441e-47 ], [ 295.375 , 7.185816472074389e-47 ], [ 295.4375 , 7.18671740115868e-47 ], 
                    [ 295.5 , 7.187620834175457e-47 ], [ 295.5625 , 7.188524380761689e-47 ], [ 295.625 , 7.189428040931552e-47 ], [ 295.6875 , 7.190331814699425e-47 ], 
                    [ 295.75 , 7.191235702079538e-47 ], [ 295.8125 , 7.192139703086175e-47 ], [ 295.875 , 7.193043817733565e-47 ], [ 295.9375 , 7.193948046036045e-47 ], 
                    [ 296.0 , 7.194852388007956e-47 ], [ 296.0625 , 7.195756843663537e-47 ], [ 296.125 , 7.196662459008784e-47 ], [ 296.1875 , 7.197569784999692e-47 ], 
                    [ 296.25 , 7.1984772253826275e-47 ], [ 296.3125 , 7.199384780172016e-47 ], [ 296.375 , 7.200292449382125e-47 ], [ 296.4375 , 7.201200233027689e-47 ], 
                    [ 296.5 , 7.202108131122725e-47 ], [ 296.5625 , 7.203016143681868e-47 ], [ 296.625 , 7.203924270719599e-47 ], [ 296.6875 , 7.204832512250302e-47 ], 
                    [ 296.75 , 7.205740868288462e-47 ], [ 296.8125 , 7.206651086250377e-47 ], [ 296.875 , 7.207562330438141e-47 ], [ 296.9375 , 7.208473689848074e-47 ], 
                    [ 297.0 , 7.209385164494694e-47 ], [ 297.0625 , 7.210296754392673e-47 ], [ 297.125 , 7.211208459556588e-47 ], [ 297.1875 , 7.212120280000754e-47 ], 
                    [ 297.25 , 7.213032215740057e-47 ], [ 297.3125 , 7.213944266789022e-47 ], [ 297.375 , 7.214856433162232e-47 ], [ 297.4375 , 7.215768714874062e-47 ], 
                    [ 297.5 , 7.216683452746073e-47 ], [ 297.5625 , 7.21759858252015e-47 ], [ 297.625 , 7.218513828339466e-47 ], [ 297.6875 , 7.219429190218842e-47 ], 
                    [ 297.75 , 7.220344668173043e-47 ], [ 297.8125 , 7.221260262216687e-47 ], [ 297.875 , 7.222175972364548e-47 ], [ 297.9375 , 7.223091798631397e-47 ], 
                    [ 298.0 , 7.224007741031857e-47 ], [ 298.0625 , 7.224923799580709e-47 ], [ 298.125 , 7.225840252313661e-47 ], [ 298.1875 , 7.226759133181354e-47 ], 
                    [ 298.25 , 7.227678130899307e-47 ], [ 298.3125 , 7.228597245482583e-47 ], [ 298.375 , 7.229516476945942e-47 ], [ 298.4375 , 7.230435825304145e-47 ], 
                    [ 298.5 , 7.231355290572312e-47 ], [ 298.5625 , 7.232274872765004e-47 ], [ 298.625 , 7.233194571897344e-47 ], [ 298.6875 , 7.2341143879840505e-47 ], 
                    [ 298.75 , 7.235034321039996e-47 ], [ 298.8125 , 7.235955134054713e-47 ], [ 298.875 , 7.236877881925854e-47 ], [ 298.9375 , 7.237800747468158e-47 ], 
                    [ 299.0 , 7.2387237306966815e-47 ], [ 299.0625 , 7.239646831626435e-47 ], [ 299.125 , 7.240570050272478e-47 ], [ 299.1875 , 7.241493386649771e-47 ], 
                    [ 299.25 , 7.242416840773275e-47 ], [ 299.3125 , 7.243340412658005e-47 ], [ 299.375 , 7.244264102319135e-47 ], [ 299.4375 , 7.245187909771528e-47 ], 
                    [ 299.5 , 7.246113017308813e-47 ], [ 299.5625 , 7.247039633723843e-47 ], [ 299.625 , 7.247966368632468e-47 ], [ 299.6875 , 7.248893222049842e-47 ], 
                    [ 299.75 , 7.249820193991168e-47 ], [ 299.8125 , 7.250747284471451e-47 ], [ 299.875 , 7.251674493506002e-47 ], [ 299.9375 , 7.252601821109931e-47 ], 
                    [ 300.0 , 7.253529267298505e-47 ], [ 300.0625 , 7.25445683208673e-47 ], [ 300.125 , 7.255384515489723e-47 ], [ 300.1875 , 7.2563138208446635e-47 ], 
                    [ 300.25 , 7.257244252258886e-47 ], [ 300.3125 , 7.258174802976467e-47 ], [ 300.375 , 7.259105473012808e-47 ], [ 300.4375 , 7.260036262383003e-47 ], 
                    [ 300.5 , 7.260967171102508e-47 ], [ 300.5625 , 7.261898199186573e-47 ], [ 300.625 , 7.262829346650557e-47 ], [ 300.6875 , 7.263760613509612e-47 ], 
                    [ 300.75 , 7.264691999779203e-47 ], [ 300.8125 , 7.265623505474535e-47 ], [ 300.875 , 7.266556941900081e-47 ], [ 300.9375 , 7.267491225862075e-47 ], 
                    [ 301.0 , 7.26842562994783e-47 ], [ 301.0625 , 7.269360154172843e-47 ], [ 301.125 , 7.270294798552661e-47 ], [ 301.1875 , 7.271229563102634e-47 ], 
                    [ 301.25 , 7.272164447838104e-47 ], [ 301.3125 , 7.273099452774683e-47 ], [ 301.375 , 7.274034577927719e-47 ], [ 301.4375 , 7.274969823312775e-47 ], 
                    [ 301.5 , 7.275905188945256e-47 ], [ 301.5625 , 7.27684268411174e-47 ], [ 301.625 , 7.277780791452314e-47 ], [ 301.6875 , 7.278719019730747e-47 ], 
                    [ 301.75 , 7.279657368962523e-47 ], [ 301.8125 , 7.28059583916334e-47 ], [ 301.875 , 7.28153443034874e-47 ], [ 301.9375 , 7.282473142534271e-47 ], 
                    [ 302.0 , 7.283411975735632e-47 ], [ 302.0625 , 7.2843509299684265e-47 ], [ 302.125 , 7.285290005248101e-47 ], [ 302.1875 , 7.286229201590364e-47 ], 
                    [ 302.25 , 7.287170657771236e-47 ], [ 302.3125 , 7.288112558620254e-47 ], [ 302.375 , 7.289054581214254e-47 ], [ 302.4375 , 7.2899967255690235e-47 ], 
                    [ 302.5 , 7.290938991700403e-47 ], [ 302.5625 , 7.291881379624237e-47 ], [ 302.625 , 7.292823889355906e-47 ], [ 302.6875 , 7.293766520911463e-47 ], 
                    [ 302.75 , 7.294709274306501e-47 ], [ 302.8125 , 7.295652149556819e-47 ], [ 302.875 , 7.296595146678168e-47 ], [ 302.9375 , 7.297540510305378e-47 ], 
                    [ 303.0 , 7.298486220393515e-47 ], [ 303.0625 , 7.299432053039049e-47 ], [ 303.125 , 7.300378008257812e-47 ], [ 303.1875 , 7.301324086065739e-47 ], 
                    [ 303.25 , 7.30227028647882e-47 ], [ 303.3125 , 7.30321660951284e-47 ], [ 303.375 , 7.304163055183589e-47 ], [ 303.4375 , 7.30510962350706e-47 ], 
                    [ 303.5 , 7.306056314499254e-47 ], [ 303.5625 , 7.30700312817586e-47 ], [ 303.625 , 7.307952316410438e-47 ], [ 303.6875 , 7.308901811078894e-47 ], 
                    [ 303.75 , 7.309851429111763e-47 ], [ 303.8125 , 7.310801170524765e-47 ], [ 303.875 , 7.311751035334189e-47 ], [ 303.9375 , 7.312701023555912e-47 ], 
                    [ 304.0 , 7.313651135205967e-47 ], [ 304.0625 , 7.314601370300495e-47 ], [ 304.125 , 7.315551728855589e-47 ], [ 304.1875 , 7.316502210887027e-47 ], 
                    [ 304.25 , 7.31745281641101e-47 ], [ 304.3125 , 7.318405766490228e-47 ], [ 304.375 , 7.319359045646626e-47 ], [ 304.4375 , 7.320312448975036e-47 ], 
                    [ 304.5 , 7.321265976491681e-47 ], [ 304.5625 , 7.322219628212582e-47 ], [ 304.625 , 7.323173404154127e-47 ], [ 304.6875 , 7.324127304332339e-47 ], 
                    [ 304.75 , 7.325081328763506e-47 ], [ 304.8125 , 7.326035477463602e-47 ], [ 304.875 , 7.32698975044913e-47 ], [ 304.9375 , 7.327944147736122e-47 ], 
                    [ 305.0 , 7.328900769329693e-47 ], [ 305.0625 , 7.329857798305486e-47 ], [ 305.125 , 7.330814952252777e-47 ], [ 305.1875 , 7.331772231187986e-47 ], 
                    [ 305.25 , 7.332729635127386e-47 ], [ 305.3125 , 7.333687164087402e-47 ], [ 305.375 , 7.334644818084358e-47 ], [ 305.4375 , 7.335602597134482e-47 ], 
                    [ 305.5 , 7.336560501254152e-47 ], [ 305.5625 , 7.337518530459703e-47 ], [ 305.625 , 7.338476684767415e-47 ], [ 305.6875 , 7.339436943775841e-47 ], 
                    [ 305.75 , 7.340397756926977e-47 ], [ 305.8125 , 7.34135869585914e-47 ], [ 305.875 , 7.342319760588846e-47 ], [ 305.9375 , 7.343280951132461e-47 ], 
                    [ 306.0 , 7.34424226750656e-47 ], [ 306.0625 , 7.345203709727562e-47 ], [ 306.125 , 7.346165277811839e-47 ], [ 306.1875 , 7.347126971776075e-47 ], 
                    [ 306.25 , 7.348088791636698e-47 ], [ 306.3125 , 7.349050737409928e-47 ], [ 306.375 , 7.350014537221023e-47 ], [ 306.4375 , 7.350979077267094e-47 ], 
                    [ 306.5 , 7.351943743889393e-47 ], [ 306.5625 , 7.352908537104583e-47 ], [ 306.625 , 7.353873456929331e-47 ], [ 306.6875 , 7.354838503380142e-47 ], 
                    [ 306.75 , 7.355803676473795e-47 ], [ 306.8125 , 7.356768976226646e-47 ], [ 306.875 , 7.357734402655524e-47 ], [ 306.9375 , 7.358699955777057e-47 ], 
                    [ 307.0 , 7.359665635607762e-47 ], [ 307.0625 , 7.360632916074114e-47 ], [ 307.125 , 7.361601195072525e-47 ], [ 307.1875 , 7.362569601446478e-47 ], 
                    [ 307.25 , 7.363538135212623e-47 ], [ 307.3125 , 7.364506796387823e-47 ], [ 307.375 , 7.365475584988684e-47 ], [ 307.4375 , 7.366444501032175e-47 ], 
                    [ 307.5 , 7.367413544534958e-47 ], [ 307.5625 , 7.368382715513748e-47 ], [ 307.625 , 7.369352013985364e-47 ], [ 307.6875 , 7.370321439966477e-47 ], 
                    [ 307.75 , 7.371292145534479e-47 ], [ 307.8125 , 7.372264159527728e-47 ], [ 307.875 , 7.373236301695412e-47 ], [ 307.9375 , 7.374208572054329e-47 ], 
                    [ 308.0 , 7.375180970621589e-47 ], [ 308.0625 , 7.376153497413996e-47 ], [ 308.125 , 7.377126152448458e-47 ], [ 308.1875 , 7.378098935741886e-47 ], 
                    [ 308.25 , 7.379071847311189e-47 ], [ 308.3125 , 7.380044887173289e-47 ], [ 308.375 , 7.381018055345098e-47 ], [ 308.4375 , 7.381992117218995e-47 ], 
                    [ 308.5 , 7.382967833458882e-47 ], [ 308.5625 , 7.383943678664253e-47 ], [ 308.625 , 7.384919652852153e-47 ], [ 308.6875 , 7.385895756039735e-47 ], 
                    [ 308.75 , 7.386871988243788e-47 ], [ 308.8125 , 7.387848349481627e-47 ], [ 308.875 , 7.388824839770203e-47 ], [ 308.9375 , 7.389801459126466e-47 ], 
                    [ 309.0 , 7.390778207567583e-47 ], [ 309.0625 , 7.391755085110717e-47 ], [ 309.125 , 7.392732431937298e-47 ], [ 309.1875 , 7.393711851678947e-47 ], 
                    [ 309.25 , 7.394691401178117e-47 ], [ 309.3125 , 7.395671080452106e-47 ], [ 309.375 , 7.396650889517948e-47 ], [ 309.4375 , 7.397630828392997e-47 ], 
                    [ 309.5 , 7.398610897094397e-47 ], [ 309.5625 , 7.399591095639401e-47 ], [ 309.625 , 7.400571424045107e-47 ], [ 309.6875 , 7.40155188232877e-47 ], 
                    [ 309.75 , 7.402532470507599e-47 ], [ 309.8125 , 7.403513188598749e-47 ], [ 309.875 , 7.404496178112867e-47 ], [ 309.9375 , 7.405479438735899e-47 ], 
                    [ 310.0 , 7.406462829928548e-47 ], [ 310.0625 , 7.407446351707943e-47 ], [ 310.125 , 7.408430004091581e-47 ], [ 310.1875 , 7.409413787096755e-47 ], 
                    [ 310.25 , 7.410397700740809e-47 ], [ 310.3125 , 7.4113817450409855e-47 ], [ 310.375 , 7.412365920014793e-47 ], [ 310.4375 , 7.413350225679479e-47 ], 
                    [ 310.5 , 7.41433466205261e-47 ], [ 310.5625 , 7.415320794582668e-47 ], [ 310.625 , 7.416307721110556e-47 ], [ 310.6875 , 7.4172947789913015e-47 ], 
                    [ 310.75 , 7.418281968242388e-47 ], [ 310.8125 , 7.419269288881458e-47 ], [ 310.875 , 7.420256740925786e-47 ], [ 310.9375 , 7.421244324393073e-47 ], 
                    [ 311.0 , 7.422232039300653e-47 ], [ 311.0625 , 7.423219885666017e-47 ], [ 311.125 , 7.424207863506663e-47 ], [ 311.1875 , 7.425195972840303e-47 ], 
                    [ 311.25 , 7.426185199583609e-47 ], [ 311.3125 , 7.427175807375806e-47 ], [ 311.375 , 7.42816654730909e-47 ], [ 311.4375 , 7.429157419400982e-47 ], 
                    [ 311.5 , 7.430148423669166e-47 ], [ 311.5625 , 7.43113956013143e-47 ], [ 311.625 , 7.432130828805093e-47 ], [ 311.6875 , 7.433122229707948e-47 ], 
                    [ 311.75 , 7.434113762857687e-47 ], [ 311.8125 , 7.435105428271845e-47 ], [ 311.875 , 7.436097225968172e-47 ], [ 311.9375 , 7.4370894982187805e-47 ], 
                    [ 312.0 , 7.438083779848512e-47 ], [ 312.0625 , 7.439078194406044e-47 ], [ 312.125 , 7.440072741909096e-47 ], [ 312.1875 , 7.441067422375551e-47 ], 
                    [ 312.25 , 7.442062235823074e-47 ], [ 312.3125 , 7.443057182269552e-47 ], [ 312.375 , 7.444052261732712e-47 ], [ 312.4375 , 7.445047474230287e-47 ], 
                    [ 312.5 , 7.446042819780165e-47 ], [ 312.5625 , 7.447038298400084e-47 ], [ 312.625 , 7.448033910107726e-47 ], [ 312.6875 , 7.449031508437801e-47 ], 
                    [ 312.75 , 7.450029588871058e-47 ], [ 312.8125 , 7.451027803035168e-47 ], [ 312.875 , 7.452026150947997e-47 ], [ 312.9375 , 7.453024632627253e-47 ], 
                    [ 313.0 , 7.454023248091018e-47 ], [ 313.0625 , 7.455021997357378e-47 ], [ 313.125 , 7.456020880444049e-47 ], [ 313.1875 , 7.457019897368958e-47 ], 
                    [ 313.25 , 7.458019048150147e-47 ], [ 313.3125 , 7.459018332805497e-47 ], [ 313.375 , 7.460018841190696e-47 ], [ 313.4375 , 7.461020557834889e-47 ], 
                    [ 313.5 , 7.462022408987675e-47 ], [ 313.5625 , 7.463024394667062e-47 ], [ 313.625 , 7.464026514891113e-47 ], [ 313.6875 , 7.4650287696778955e-47 ], 
                    [ 313.75 , 7.466031159045583e-47 ], [ 313.8125 , 7.46703368301193e-47 ], [ 313.875 , 7.46803634159538e-47 ], [ 313.9375 , 7.469039134813799e-47 ], 
                    [ 314.0 , 7.470042062685264e-47 ], [ 314.0625 , 7.471045426802557e-47 ], [ 314.125 , 7.472050780933925e-47 ], [ 314.1875 , 7.473056270352295e-47 ], 
                    [ 314.25 , 7.474061895076244e-47 ], [ 314.3125 , 7.475067655123769e-47 ], [ 314.375 , 7.476073550513182e-47 ], [ 314.4375 , 7.477079581262647e-47 ], 
                    [ 314.5 , 7.478085747390325e-47 ], [ 314.5625 , 7.479092048914592e-47 ], [ 314.625 , 7.480098485853402e-47 ], [ 314.6875 , 7.481105058225136e-47 ], 
                    [ 314.75 , 7.48211176604818e-47 ], [ 314.8125 , 7.483120219733341e-47 ], [ 314.875 , 7.484129345522938e-47 ], [ 314.9375 , 7.485138607396794e-47 ], 
                    [ 315.0 , 7.486148005373259e-47 ], [ 315.0625 , 7.48715753947074e-47 ], [ 315.125 , 7.488167209707485e-47 ], [ 315.1875 , 7.48917701610191e-47 ], 
                    [ 315.25 , 7.490186958672482e-47 ], [ 315.3125 , 7.491197037437243e-47 ], [ 315.375 , 7.492207252414935e-47 ], [ 315.4375 , 7.493217603623763e-47 ], 
                    [ 315.5 , 7.494228799415733e-47 ], [ 315.5625 , 7.495241538709823e-47 ], [ 315.625 , 7.496254414861235e-47 ], [ 315.6875 , 7.497267427888675e-47 ], 
                    [ 315.75 , 7.498280577810482e-47 ], [ 315.8125 , 7.499293864645206e-47 ], [ 315.875 , 7.500307288411298e-47 ], [ 315.9375 , 7.50132084912742e-47 ], 
                    [ 316.0 , 7.502334546811868e-47 ], [ 316.0625 , 7.503348381483416e-47 ], [ 316.125 , 7.504362353160257e-47 ], [ 316.1875 , 7.505376461861013e-47 ], 
                    [ 316.25 , 7.506392591589639e-47 ], [ 316.3125 , 7.507409083243268e-47 ], [ 316.375 , 7.508425712547048e-47 ], [ 316.4375 , 7.509442479519567e-47 ], 
                    [ 316.5 , 7.510459384179256e-47 ], [ 316.5625 , 7.511476426544916e-47 ], [ 316.625 , 7.512493606635359e-47 ], [ 316.6875 , 7.513510924468966e-47 ], 
                    [ 316.75 , 7.514528380064445e-47 ], [ 316.8125 , 7.515545973440609e-47 ], [ 316.875 , 7.516563704616011e-47 ], [ 316.9375 , 7.517582468192964e-47 ], 
                    [ 317.0 , 7.518602573420201e-47 ], [ 317.0625 , 7.519622817071549e-47 ], [ 317.125 , 7.520643199165792e-47 ], [ 317.1875 , 7.521663719721769e-47 ], 
                    [ 317.25 , 7.52268437875811e-47 ], [ 317.3125 , 7.523705176293766e-47 ], [ 317.375 , 7.52472611234753e-47 ], [ 317.4375 , 7.525747186938092e-47 ], 
                    [ 317.5 , 7.52676840008425e-47 ], [ 317.5625 , 7.527789751804915e-47 ], [ 317.625 , 7.528811242118888e-47 ], [ 317.6875 , 7.5298347937709965e-47 ], 
                    [ 317.75 , 7.530858622050292e-47 ], [ 317.8125 , 7.531882589539222e-47 ], [ 317.875 , 7.532906696256342e-47 ], [ 317.9375 , 7.533930942220955e-47 ], 
                    [ 318.0 , 7.534955327451782e-47 ], [ 318.0625 , 7.535979851967759e-47 ], [ 318.125 , 7.53700451578793e-47 ], [ 318.1875 , 7.538029318931131e-47 ], 
                    [ 318.25 , 7.539054261416465e-47 ], [ 318.3125 , 7.540079343262666e-47 ], [ 318.375 , 7.541105408730289e-47 ], [ 318.4375 , 7.54213280858444e-47 ], 
                    [ 318.5 , 7.543160348411499e-47 ], [ 318.5625 , 7.544188028230538e-47 ], [ 318.625 , 7.545215848060627e-47 ], [ 318.6875 , 7.546243807920843e-47 ], 
                    [ 318.75 , 7.547271907830155e-47 ], [ 318.8125 , 7.54830014780802e-47 ], [ 318.875 , 7.549328527872931e-47 ], [ 318.9375 , 7.550357048044509e-47 ], 
                    [ 319.0 , 7.551385708341628e-47 ], [ 319.0625 , 7.552414508783271e-47 ], [ 319.125 , 7.553445229994537e-47 ], [ 319.1875 , 7.554476361183063e-47 ], 
                    [ 319.25 , 7.555507633132618e-47 ], [ 319.3125 , 7.556539045862684e-47 ], [ 319.375 , 7.557570599392212e-47 ], [ 319.4375 , 7.558602293740475e-47 ], 
                    [ 319.5 , 7.559634128926914e-47 ], [ 319.5625 , 7.560666104970539e-47 ], [ 319.625 , 7.561698221890577e-47 ], [ 319.6875 , 7.56273047970626e-47 ], 
                    [ 319.75 , 7.563762878436984e-47 ], [ 319.8125 , 7.564796018089788e-47 ], [ 319.875 , 7.565830718411657e-47 ], [ 319.9375 , 7.566865560258074e-47 ], 
                    [ 320.0 , 7.5679005436484e-47 ], [ 320.0625 , 7.568935668602153e-47 ], [ 320.125 , 7.569970935138537e-47 ], [ 320.1875 , 7.571006343276862e-47 ], 
                    [ 320.25 , 7.572041893036712e-47 ], [ 320.3125 , 7.573077584437351e-47 ], [ 320.375 , 7.57411341749815e-47 ], [ 320.4375 , 7.575149392238595e-47 ], 
                    [ 320.5 , 7.576185508677741e-47 ], [ 320.5625 , 7.577223154121508e-47 ], [ 320.625 , 7.578261537675403e-47 ], [ 320.6875 , 7.579300063529502e-47 ], 
                    [ 320.75 , 7.580338731703361e-47 ], [ 320.8125 , 7.581377542216536e-47 ], [ 320.875 , 7.582416495088211e-47 ], [ 320.9375 , 7.583455590338164e-47 ], 
                    [ 321.0 , 7.584494827985906e-47 ], [ 321.0625 , 7.585534208050898e-47 ], [ 321.125 , 7.586573730552656e-47 ], [ 321.1875 , 7.587613395510755e-47 ], 
                    [ 321.25 , 7.588653354352158e-47 ], [ 321.3125 , 7.589695313700893e-47 ], [ 321.375 , 7.590737416115761e-47 ], [ 321.4375 , 7.591779661616402e-47 ], 
                    [ 321.5 , 7.592822050222414e-47 ], [ 321.5625 , 7.59386458195355e-47 ], [ 321.625 , 7.59490725682941e-47 ], [ 321.6875 , 7.595950074869649e-47 ], 
                    [ 321.75 , 7.596993036093921e-47 ], [ 321.8125 , 7.598036140521996e-47 ], [ 321.875 , 7.599079388173319e-47 ], [ 321.9375 , 7.600122779067667e-47 ], 
                    [ 322.0 , 7.601167139140503e-47 ], [ 322.0625 , 7.602212759743173e-47 ], [ 322.125 , 7.603258524181918e-47 ], [ 322.1875 , 7.604304432476684e-47 ], 
                    [ 322.25 , 7.605350484647044e-47 ], [ 322.3125 , 7.606396680712952e-47 ], [ 322.375 , 7.607443020694094e-47 ], [ 322.4375 , 7.608489504610431e-47 ], 
                    [ 322.5 , 7.609536132481652e-47 ], [ 322.5625 , 7.610582904327453e-47 ], [ 322.625 , 7.611629820167801e-47 ], [ 322.6875 , 7.612676880022504e-47 ], 
                    [ 322.75 , 7.613725542719021e-47 ], [ 322.8125 , 7.614774840574699e-47 ], [ 322.875 , 7.615824283041108e-47 ], [ 322.9375 , 7.61687387013812e-47 ], 
                    [ 323.0 , 7.617923601885671e-47 ], [ 323.0625 , 7.618973478303692e-47 ], [ 323.125 , 7.620023499412125e-47 ], [ 323.1875 , 7.621073665230856e-47 ], 
                    [ 323.25 , 7.6221239757798265e-47 ], [ 323.3125 , 7.623174431079254e-47 ], [ 323.375 , 7.624225031148764e-47 ], [ 323.4375 , 7.625275863238762e-47 ], 
                    [ 323.5 , 7.626328693366363e-47 ], [ 323.5625 , 7.627381668859471e-47 ], [ 323.625 , 7.62843478973789e-47 ], [ 323.6875 , 7.629488056021852e-47 ], 
                    [ 323.75 , 7.6305414677314355e-47 ], [ 323.8125 , 7.631595024886826e-47 ], [ 323.875 , 7.632648727507838e-47 ], [ 323.9375 , 7.633702575614985e-47 ], 
                    [ 324.0 , 7.634756569227762e-47 ], [ 324.0625 , 7.6358107083668e-47 ], [ 324.125 , 7.636864993051919e-47 ], [ 324.1875 , 7.637920029806763e-47 ], 
                    [ 324.25 , 7.638976530180703e-47 ], [ 324.3125 , 7.640033176692939e-47 ], [ 324.375 , 7.641089969363741e-47 ], [ 324.4375 , 7.642146908213327e-47 ], 
                    [ 324.5 , 7.6432039932618595e-47 ], [ 324.5625 , 7.64426122452978e-47 ], [ 324.625 , 7.645318602036936e-47 ], [ 324.6875 , 7.6463761258039315e-47 ], 
                    [ 324.75 , 7.647433795850675e-47 ], [ 324.8125 , 7.648491612197671e-47 ], [ 324.875 , 7.649549574865048e-47 ], [ 324.9375 , 7.650608738209599e-47 ], 
                    [ 325.0 , 7.651668880502379e-47 ], [ 325.0625 , 7.652729169698774e-47 ], [ 325.125 , 7.653789605819035e-47 ], [ 325.1875 , 7.6548501888836295e-47 ], 
                    [ 325.25 , 7.655910918912863e-47 ], [ 325.3125 , 7.656971795927102e-47 ], [ 325.375 , 7.658032819946713e-47 ], [ 325.4375 , 7.659093990991961e-47 ], 
                    [ 325.5 , 7.660155309083488e-47 ], [ 325.5625 , 7.661216774241452e-47 ], [ 325.625 , 7.662278386486342e-47 ], [ 325.6875 , 7.663341609411287e-47 ], 
                    [ 325.75 , 7.664405402792208e-47 ], [ 325.8125 , 7.66546934384458e-47 ], [ 325.875 , 7.666533432588738e-47 ], [ 325.9375 , 7.667597669045402e-47 ], 
                    [ 326.0 , 7.668662053234915e-47 ], [ 326.0625 , 7.669726585177837e-47 ], [ 326.125 , 7.67079126489468e-47 ], [ 326.1875 , 7.671856092405958e-47 ], 
                    [ 326.25 , 7.6729210677322405e-47 ], [ 326.3125 , 7.673986190893937e-47 ], [ 326.375 , 7.67505146191157e-47 ], [ 326.4375 , 7.676118679348886e-47 ], 
                    [ 326.5 , 7.677186108340409e-47 ], [ 326.5625 , 7.678253685766819e-47 ], [ 326.625 , 7.679321411648922e-47 ], [ 326.6875 , 7.680389286007308e-47 ], 
                    [ 326.75 , 7.681457308862622e-47 ], [ 326.8125 , 7.68252548023546e-47 ], [ 326.875 , 7.683593800146638e-47 ], [ 326.9375 , 7.684662268616595e-47 ], 
                    [ 327.0 , 7.685730885666096e-47 ], [ 327.0625 , 7.686799651315913e-47 ], [ 327.125 , 7.687868807682965e-47 ], [ 327.1875 , 7.68893973891633e-47 ], 
                    [ 327.25 , 7.690010819332126e-47 ], [ 327.3125 , 7.691082048950919e-47 ], [ 327.375 , 7.692153427793599e-47 ], [ 327.4375 , 7.6932249558808995e-47 ], 
                    [ 327.5 , 7.694296633233503e-47 ], [ 327.5625 , 7.695368459872364e-47 ], [ 327.625 , 7.696440435818279e-47 ], [ 327.6875 , 7.697512561091883e-47 ], 
                    [ 327.75 , 7.698584835714142e-47 ], [ 327.8125 , 7.699657259705804e-47 ], [ 327.875 , 7.700730318627539e-47 ], [ 327.9375 , 7.701804864338842e-47 ], 
                    [ 328.0 , 7.702879559990272e-47 ], [ 328.0625 , 7.703954405602752e-47 ], [ 328.125 , 7.705029401197207e-47 ], [ 328.1875 , 7.706104546794621e-47 ], 
                    [ 328.25 , 7.70717984241576e-47 ], [ 328.3125 , 7.70825528808183e-47 ], [ 328.375 , 7.709330883813387e-47 ], [ 328.4375 , 7.710406629631755e-47 ], 
                    [ 328.5 , 7.711482525557656e-47 ], [ 328.5625 , 7.712558571612036e-47 ], [ 328.625 , 7.713635445281087e-47 ], [ 328.6875 , 7.714713593003962e-47 ], 
                    [ 328.75 , 7.715791891421454e-47 ], [ 328.8125 , 7.716870340554519e-47 ], [ 328.875 , 7.717948940424059e-47 ], [ 328.9375 , 7.719027691051471e-47 ], 
                    [ 329.0 , 7.720106592457496e-47 ], [ 329.0625 , 7.721185644663485e-47 ], [ 329.125 , 7.722264847690459e-47 ], [ 329.1875 , 7.723344201559225e-47 ], 
                    [ 329.25 , 7.724423706291306e-47 ], [ 329.3125 , 7.7255033619075125e-47 ], [ 329.375 , 7.726584000084835e-47 ], [ 329.4375 , 7.727665770262657e-47 ], 
                    [ 329.5 , 7.728747691895192e-47 ], [ 329.5625 , 7.729829765003426e-47 ], [ 329.625 , 7.730911989608676e-47 ], [ 329.6875 , 7.731994365732262e-47 ], 
                    [ 329.75 , 7.733076893395179e-47 ], [ 329.8125 , 7.734159572618806e-47 ], [ 329.875 , 7.735242403424309e-47 ], [ 329.9375 , 7.736325385832855e-47 ], 
                    [ 330.0 , 7.73740851986578e-47 ], [ 330.0625 , 7.738491805544202e-47 ], [ 330.125 , 7.739576164808054e-47 ], [ 330.1875 , 7.740661545328474e-47 ], 
                    [ 330.25 , 7.741747078060183e-47 ], [ 330.3125 , 7.742832763024473e-47 ], [ 330.375 , 7.74391860024275e-47 ], [ 330.4375 , 7.745004589736362e-47 ], 
                    [ 330.5 , 7.746090731526609e-47 ], [ 330.5625 , 7.747177025635072e-47 ], [ 330.625 , 7.748263472082778e-47 ], [ 330.6875 , 7.749350070891315e-47 ], 
                    [ 330.75 , 7.750436822081936e-47 ], [ 330.8125 , 7.75152372567623e-47 ], [ 330.875 , 7.752611741978939e-47 ], [ 330.9375 , 7.753700717613265e-47 ], 
                    [ 331.0 , 7.75478984621126e-47 ], [ 331.0625 , 7.75587912779441e-47 ], [ 331.125 , 7.756968562384258e-47 ], [ 331.1875 , 7.758058150002188e-47 ], 
                    [ 331.25 , 7.759147890669859e-47 ], [ 331.3125 , 7.760237784408383e-47 ], [ 331.375 , 7.761327831239758e-47 ], [ 331.4375 , 7.762418031185269e-47 ], 
                    [ 331.5 , 7.763508384266421e-47 ], [ 331.5625 , 7.764598890504671e-47 ], [ 331.625 , 7.765690500490903e-47 ], [ 331.6875 , 7.766783057266666e-47 ], 
                    [ 331.75 , 7.767875767754439e-47 ], [ 331.8125 , 7.768968631975901e-47 ], [ 331.875 , 7.77006164995279e-47 ], [ 331.9375 , 7.771154821706573e-47 ], 
                    [ 332.0 , 7.772248147258831e-47 ], [ 332.0625 , 7.773341626631477e-47 ], [ 332.125 , 7.774435259845876e-47 ], [ 332.1875 , 7.775529046923782e-47 ], 
                    [ 332.25 , 7.776622987886954e-47 ], [ 332.3125 , 7.777717082756875e-47 ], [ 332.375 , 7.778812240856877e-47 ], [ 332.4375 , 7.779908394252465e-47 ], 
                    [ 332.5 , 7.781004702112857e-47 ], [ 332.5625 , 7.78210116445965e-47 ], [ 332.625 , 7.783197781314838e-47 ], [ 332.6875 , 7.784294552700027e-47 ], 
                    [ 332.75 , 7.785391478636936e-47 ], [ 332.8125 , 7.786488559147676e-47 ], [ 332.875 , 7.787585794253699e-47 ], [ 332.9375 , 7.788683183976897e-47 ], 
                    [ 333.0 , 7.789780728339171e-47 ], [ 333.0625 , 7.79087842736209e-47 ], [ 333.125 , 7.791977083510257e-47 ], [ 333.1875 , 7.793076811599234e-47 ], 
                    [ 333.25 , 7.794176694899436e-47 ], [ 333.3125 , 7.795276733432712e-47 ], [ 333.375 , 7.796376927221027e-47 ], [ 333.4375 , 7.797477276286128e-47 ], 
                    [ 333.5 , 7.798577780650149e-47 ], [ 333.5625 , 7.799678440334733e-47 ], [ 333.625 , 7.800779255362023e-47 ], [ 333.6875 , 7.801880225753999e-47 ], 
                    [ 333.75 , 7.802981351532476e-47 ], [ 333.8125 , 7.804082632719276e-47 ], [ 333.875 , 7.805184728224889e-47 ], [ 333.9375 , 7.806288025119979e-47 ], 
                    [ 334.0 , 7.807391477971062e-47 ], [ 334.0625 , 7.808495086799907e-47 ], [ 334.125 , 7.809598851628784e-47 ], [ 334.1875 , 7.810702772479464e-47 ], 
                    [ 334.25 , 7.811806849374226e-47 ], [ 334.3125 , 7.812911082335127e-47 ], [ 334.375 , 7.81401547138406e-47 ], [ 334.4375 , 7.815120016543423e-47 ], 
                    [ 334.5 , 7.816224717834839e-47 ], [ 334.5625 , 7.81732957528071e-47 ], [ 334.625 , 7.818435057738916e-47 ], [ 334.6875 , 7.81954190803861e-47 ], 
                    [ 334.75 , 7.82064891503437e-47 ], [ 334.8125 , 7.821756078748269e-47 ], [ 334.875 , 7.822863399202605e-47 ], [ 334.9375 , 7.823970876419455e-47 ], 
                    [ 335.0 , 7.825078510421014e-47 ], [ 335.0625 , 7.826186301229642e-47 ], [ 335.125 , 7.827294248867484e-47 ], [ 335.1875 , 7.82840235335652e-47 ], 
                    [ 335.25 , 7.829510614719122e-47 ], [ 335.3125 , 7.830619032977664e-47 ], [ 335.375 , 7.831727852179357e-47 ], [ 335.4375 , 7.832838279190383e-47 ], 
                    [ 335.5 , 7.833948863644246e-47 ], [ 335.5625 , 7.835059605562991e-47 ], [ 335.625 , 7.836170504969167e-47 ], [ 335.6875 , 7.837281561884826e-47 ], 
                    [ 335.75 , 7.838392776332577e-47 ], [ 335.8125 , 7.839504148334702e-47 ], [ 335.875 , 7.840615677913428e-47 ], [ 335.9375 , 7.841727365091042e-47 ], 
                    [ 336.0 , 7.842839209890111e-47 ], [ 336.0625 , 7.843951212332816e-47 ], [ 336.125 , 7.845063372441678e-47 ], [ 336.1875 , 7.84617730919361e-47 ], 
                    [ 336.25 , 7.847291434467175e-47 ], [ 336.3125 , 7.848405717941994e-47 ], [ 336.375 , 7.849520159640595e-47 ], [ 336.4375 , 7.850634759585383e-47 ], 
                    [ 336.5 , 7.851749517798777e-47 ], [ 336.5625 , 7.852864434303304e-47 ], [ 336.625 , 7.853979509121497e-47 ], [ 336.6875 , 7.855094742275724e-47 ], 
                    [ 336.75 , 7.856210133788636e-47 ], [ 336.8125 , 7.857325683682552e-47 ], [ 336.875 , 7.858441391979904e-47 ], [ 336.9375 , 7.859558562869428e-47 ], 
                    [ 337.0 , 7.860676236133843e-47 ], [ 337.0625 , 7.861794068337548e-47 ], [ 337.125 , 7.862912059503368e-47 ], [ 337.1875 , 7.864030209653683e-47 ], 
                    [ 337.25 , 7.865148518811215e-47 ], [ 337.3125 , 7.866266986998686e-47 ], [ 337.375 , 7.867385614238599e-47 ], [ 337.4375 , 7.868504400553463e-47 ], 
                    [ 337.5 , 7.86962334596601e-47 ], [ 337.5625 , 7.870742450498974e-47 ], [ 337.625 , 7.871861714174763e-47 ], [ 337.6875 , 7.872982065740374e-47 ], 
                    [ 337.75 , 7.87410326634427e-47 ], [ 337.8125 , 7.875224626619602e-47 ], [ 337.875 , 7.87634614658922e-47 ], [ 337.9375 , 7.877467826275868e-47 ], 
                    [ 338.0 , 7.87858966570218e-47 ], [ 338.0625 , 7.879711664890959e-47 ], [ 338.125 , 7.880833823864958e-47 ], [ 338.1875 , 7.881956142646931e-47 ], 
                    [ 338.25 , 7.883078621259637e-47 ], [ 338.3125 , 7.884201259725781e-47 ], [ 338.375 , 7.885324058068298e-47 ], [ 338.4375 , 7.886447547338572e-47 ], 
                    [ 338.5 , 7.887572282432169e-47 ], [ 338.5625 , 7.888697177931312e-47 ], [ 338.625 , 7.889822233858763e-47 ], [ 338.6875 , 7.890947450237346e-47 ], 
                    [ 338.75 , 7.892072827090058e-47 ], [ 338.8125 , 7.893198364439615e-47 ], [ 338.875 , 7.894324062309188e-47 ], [ 338.9375 , 7.895449920721389e-47 ], 
                    [ 339.0 , 7.896575939699282e-47 ], [ 339.0625 , 7.897702119265598e-47 ], [ 339.125 , 7.89882845944352e-47 ], [ 339.1875 , 7.899955048846543e-47 ], 
                    [ 339.25 , 7.901083310825885e-47 ], [ 339.3125 , 7.902211733942231e-47 ], [ 339.375 , 7.903340318218707e-47 ], [ 339.4375 , 7.904469063678107e-47 ], 
                    [ 339.5 , 7.90559797034356e-47 ], [ 339.5625 , 7.906727038238092e-47 ], [ 339.625 , 7.907856267384729e-47 ], [ 339.6875 , 7.908985657806555e-47 ], 
                    [ 339.75 , 7.910115209526495e-47 ], [ 339.8125 , 7.911244922567636e-47 ], [ 339.875 , 7.912374796953022e-47 ], [ 339.9375 , 7.913504832705694e-47 ], 
                    [ 340.0 , 7.914636225023703e-47 ], [ 340.0625 , 7.9157681682496e-47 ], [ 340.125 , 7.916900273364923e-47 ], [ 340.1875 , 7.918032540392711e-47 ], 
                    [ 340.25 , 7.919164969356123e-47 ], [ 340.3125 , 7.920297560278543e-47 ], [ 340.375 , 7.921430313182852e-47 ], [ 340.4375 , 7.922563228092329e-47 ], 
                    [ 340.5 , 7.923696305030202e-47 ], [ 340.5625 , 7.924829544019473e-47 ], [ 340.625 , 7.925962945083546e-47 ], [ 340.6875 , 7.927096508245429e-47 ], 
                    [ 340.75 , 7.928230898050157e-47 ], [ 340.8125 , 7.929366347172047e-47 ], [ 340.875 , 7.930501958908531e-47 ], [ 340.9375 , 7.931637733282673e-47 ], 
                    [ 341.0 , 7.932773670317652e-47 ], [ 341.0625 , 7.933909770037215e-47 ], [ 341.125 , 7.935046032464153e-47 ], [ 341.1875 , 7.936182457622163e-47 ], 
                    [ 341.25 , 7.937319045534383e-47 ], [ 341.3125 , 7.938455796223951e-47 ], [ 341.375 , 7.939592709714518e-47 ], [ 341.4375 , 7.940729786029063e-47 ], 
                    [ 341.5 , 7.941867132151602e-47 ], [ 341.5625 , 7.943006089078825e-47 ], [ 341.625 , 7.944145209345892e-47 ], [ 341.6875 , 7.94528449297617e-47 ], 
                    [ 341.75 , 7.9464239399930305e-47 ], [ 341.8125 , 7.947563550420018e-47 ], [ 341.875 , 7.948703324280513e-47 ], [ 341.9375 , 7.949843261597953e-47 ], 
                    [ 342.0 , 7.950983362395664e-47 ], [ 342.0625 , 7.95212362669732e-47 ], [ 342.125 , 7.9532640545262e-47 ], [ 342.1875 , 7.954404645905922e-47 ], 
                    [ 342.25 , 7.955545400859718e-47 ], [ 342.3125 , 7.956687375313158e-47 ], [ 342.375 , 7.957830001575102e-47 ], [ 342.4375 , 7.958972791924831e-47 ], 
                    [ 342.5 , 7.960115746385856e-47 ], [ 342.5625 , 7.961258864981684e-47 ], [ 342.625 , 7.962402147736003e-47 ], [ 342.6875 , 7.963545594672272e-47 ], 
                    [ 342.75 , 7.964689205814011e-47 ], [ 342.8125 , 7.965832981185143e-47 ], [ 342.875 , 7.966976920808853e-47 ], [ 342.9375 , 7.968121024709127e-47 ], 
                    [ 343.0 , 7.969265292909047e-47 ], [ 343.0625 , 7.970410134270675e-47 ], [ 343.125 , 7.971556238931481e-47 ], [ 343.1875 , 7.972702508396398e-47 ], 
                    [ 343.25 , 7.973848942689008e-47 ], [ 343.3125 , 7.974995541833071e-47 ], [ 343.375 , 7.976142305852348e-47 ], [ 343.4375 , 7.977289234770377e-47 ], 
                    [ 343.5 , 7.978436328611095e-47 ], [ 343.5625 , 7.979583587398051e-47 ], [ 343.625 , 7.980731011155072e-47 ], [ 343.6875 , 7.9818785999058245e-47 ], 
                    [ 343.75 , 7.98302635367398e-47 ], [ 343.8125 , 7.984174272483321e-47 ], [ 343.875 , 7.985323624353838e-47 ], [ 343.9375 , 7.986473401278071e-47 ], 
                    [ 344.0 , 7.98762334375439e-47 ], [ 344.0625 , 7.988773451806574e-47 ], [ 344.125 , 7.989923725458633e-47 ], [ 344.1875 , 7.991074164734187e-47 ], 
                    [ 344.25 , 7.992224769657308e-47 ], [ 344.3125 , 7.993375540251678e-47 ], [ 344.375 , 7.994526476541264e-47 ], [ 344.4375 , 7.995677578549699e-47 ], 
                    [ 344.5 , 7.996828846301183e-47 ], [ 344.5625 , 7.997980279819353e-47 ], [ 344.625 , 7.999132414566085e-47 ], [ 344.6875 , 8.000285663352138e-47 ], 
                    [ 344.75 , 8.001439078404125e-47 ], [ 344.8125 , 8.002592659745788e-47 ], [ 344.875 , 8.003746407401444e-47 ], [ 344.9375 , 8.004900321394842e-47 ], 
                    [ 345.0 , 8.006054401749965e-47 ], [ 345.0625 , 8.007208648490852e-47 ], [ 345.125 , 8.008363061641492e-47 ], [ 345.1875 , 8.009517641226048e-47 ], 
                    [ 345.25 , 8.010672387268117e-47 ], [ 345.3125 , 8.011827299792153e-47 ], [ 345.375 , 8.012982378821757e-47 ], [ 345.4375 , 8.014138909564045e-47 ], 
                    [ 345.5 , 8.015295810384424e-47 ], [ 345.5625 , 8.016452878212136e-47 ], [ 345.625 , 8.017610113071062e-47 ], [ 345.6875 , 8.018767514985713e-47 ], 
                    [ 345.75 , 8.019925083979862e-47 ], [ 345.8125 , 8.021082820077629e-47 ], [ 345.875 , 8.022240723303364e-47 ], [ 345.9375 , 8.023398793681024e-47 ], 
                    [ 346.0 , 8.024557031234907e-47 ], [ 346.0625 , 8.025715435988977e-47 ], [ 346.125 , 8.0268740079674825e-47 ], [ 346.1875 , 8.028033235089782e-47 ], 
                    [ 346.25 , 8.029193586371615e-47 ], [ 346.3125 , 8.030354105367581e-47 ], [ 346.375 , 8.03151479210198e-47 ], [ 346.4375 , 8.032675646599053e-47 ], 
                    [ 346.5 , 8.033836668883167e-47 ], [ 346.5625 , 8.034997858978285e-47 ], [ 346.625 , 8.036159216909003e-47 ], [ 346.6875 , 8.037320742699298e-47 ], 
                    [ 346.75 , 8.038482436373543e-47 ], [ 346.8125 , 8.039644297956062e-47 ], [ 346.875 , 8.040806327471068e-47 ], [ 346.9375 , 8.041968524942773e-47 ], 
                    [ 347.0 , 8.043132037655993e-47 ], [ 347.0625 , 8.044296035730697e-47 ], [ 347.125 , 8.04546020225851e-47 ], [ 347.1875 , 8.046624537264038e-47 ], 
                    [ 347.25 , 8.047789040771551e-47 ], [ 347.3125 , 8.048953712805378e-47 ], [ 347.375 , 8.050118553389903e-47 ], [ 347.4375 , 8.051283562549695e-47 ], 
                    [ 347.5 , 8.052448740309034e-47 ], [ 347.5625 , 8.053614086692205e-47 ], [ 347.625 , 8.054779601723898e-47 ], [ 347.6875 , 8.055945285428232e-47 ], 
                    [ 347.75 , 8.057111424480981e-47 ], [ 347.8125 , 8.058278872261416e-47 ], [ 347.875 , 8.059446489200968e-47 ], [ 347.9375 , 8.060614275324149e-47 ], 
                    [ 348.0 , 8.061782230655643e-47 ], [ 348.0625 , 8.062950355219798e-47 ], [ 348.125 , 8.06411864904119e-47 ], [ 348.1875 , 8.065287112144461e-47 ], 
                    [ 348.25 , 8.066455744553794e-47 ], [ 348.3125 , 8.067624546294237e-47 ], [ 348.375 , 8.06879351738981e-47 ], [ 348.4375 , 8.069962657865452e-47 ], 
                    [ 348.5 , 8.071131967745306e-47 ], [ 348.5625 , 8.072302283898414e-47 ], [ 348.625 , 8.073473351944945e-47 ], [ 348.6875 , 8.074644589881152e-47 ], 
                    [ 348.75 , 8.075815997731568e-47 ], [ 348.8125 , 8.076987575520901e-47 ], [ 348.875 , 8.078159323273863e-47 ], [ 348.9375 , 8.079331241014935e-47 ], 
                    [ 349.0 , 8.080503328768951e-47 ], [ 349.0625 , 8.081675586560579e-47 ], [ 349.125 , 8.082848014414367e-47 ], [ 349.1875 , 8.084020612355103e-47 ], 
                    [ 349.25 , 8.085193380407347e-47 ], [ 349.3125 , 8.086366318595835e-47 ], [ 349.375 , 8.087540758977681e-47 ], [ 349.4375 , 8.088715439930996e-47 ], 
                    [ 349.5 , 8.089890291501673e-47 ], [ 349.5625 , 8.091065313714609e-47 ], [ 349.625 , 8.092240506594361e-47 ], [ 349.6875 , 8.093415870166002e-47 ], 
                    [ 349.75 , 8.09459140445427e-47 ], [ 349.8125 , 8.095767109483843e-47 ], [ 349.875 , 8.09694298527958e-47 ], [ 349.9375 , 8.098119031866282e-47 ], 
                    [ 350.0 , 8.099295249268758e-47 ], [ 350.0625 , 8.100471637511816e-47 ], [ 350.125 , 8.101648587478938e-47 ], [ 350.1875 , 8.102826705728177e-47 ], 
                    [ 350.25 , 8.104004995296129e-47 ], [ 350.3125 , 8.105183456207362e-47 ], [ 350.375 , 8.106362088487194e-47 ], [ 350.4375 , 8.107540892160085e-47 ], 
                    [ 350.5 , 8.10871986725136e-47 ], [ 350.5625 , 8.109899013785661e-47 ], [ 350.625 , 8.111078331788204e-47 ], [ 350.6875 , 8.112257821283522e-47 ], 
                    [ 350.75 , 8.113437482296956e-47 ], [ 350.8125 , 8.114617314853159e-47 ], [ 350.875 , 8.115797318977249e-47 ], [ 350.9375 , 8.116978312507041e-47 ], 
                    [ 351.0 , 8.118160059378841e-47 ], [ 351.0625 , 8.119341978300469e-47 ], [ 351.125 , 8.120524069297207e-47 ], [ 351.1875 , 8.12170633239399e-47 ], 
                    [ 351.25 , 8.122888767615816e-47 ], [ 351.3125 , 8.124071374987863e-47 ], [ 351.375 , 8.125254154535076e-47 ], [ 351.4375 , 8.126437106282637e-47 ], 
                    [ 351.5 , 8.127620230255503e-47 ], [ 351.5625 , 8.1288035264788625e-47 ], [ 351.625 , 8.129986994977679e-47 ], [ 351.6875 , 8.131170635777147e-47 ], 
                    [ 351.75 , 8.132355607039141e-47 ], [ 351.8125 , 8.13354094155825e-47 ], [ 351.875 , 8.134726448846291e-47 ], [ 351.9375 , 8.135912128928274e-47 ], 
                    [ 352.0 , 8.137097981829614e-47 ], [ 352.0625 , 8.138284007575384e-47 ], [ 352.125 , 8.139470206190664e-47 ], [ 352.1875 , 8.140656577700878e-47 ], 
                    [ 352.25 , 8.141843122131e-47 ], [ 352.3125 , 8.14302983950652e-47 ], [ 352.375 , 8.144216729852357e-47 ], [ 352.4375 , 8.145403793193896e-47 ], 
                    [ 352.5 , 8.146591172793391e-47 ], [ 352.5625 , 8.147779959703516e-47 ], [ 352.625 , 8.14896892008667e-47 ], [ 352.6875 , 8.150158053968281e-47 ], 
                    [ 352.75 , 8.151347361373728e-47 ], [ 352.8125 , 8.152536842328041e-47 ], [ 352.875 , 8.153726496856717e-47 ], [ 352.9375 , 8.154916324985203e-47 ], 
                    [ 353.0 , 8.156106326738716e-47 ], [ 353.0625 , 8.157296502142475e-47 ], [ 353.125 , 8.158486851221936e-47 ], [ 353.1875 , 8.159677374002617e-47 ], 
                    [ 353.25 , 8.160868070509515e-47 ], [ 353.3125 , 8.162059348759046e-47 ], [ 353.375 , 8.163251666738227e-47 ], [ 353.4375 , 8.164444158891796e-47 ], 
                    [ 353.5 , 8.165636825245312e-47 ], [ 353.5625 , 8.166829665824107e-47 ], [ 353.625 , 8.168022680653632e-47 ], [ 353.6875 , 8.169215869759571e-47 ], 
                    [ 353.75 , 8.170409233167097e-47 ], [ 353.8125 , 8.171602770901725e-47 ], [ 353.875 , 8.1727964829891e-47 ], [ 353.9375 , 8.173990369454571e-47 ], 
                    [ 354.0 , 8.175184430323497e-47 ], [ 354.0625 , 8.17637866562153e-47 ], [ 354.125 , 8.177573763264216e-47 ], [ 354.1875 , 8.178769702763177e-47 ], 
                    [ 354.25 , 8.179965817163865e-47 ], [ 354.3125 , 8.181162106491797e-47 ], [ 354.375 , 8.182358570772443e-47 ], [ 354.4375 , 8.183555210031619e-47 ], 
                    [ 354.5 , 8.184752024294801e-47 ], [ 354.5625 , 8.185949013587462e-47 ], [ 354.625 , 8.187146177935321e-47 ], [ 354.6875 , 8.188343517364208e-47 ], 
                    [ 354.75 , 8.189541031899265e-47 ], [ 354.8125 , 8.190738721566449e-47 ], [ 354.875 , 8.191936586391197e-47 ], [ 354.9375 , 8.193135508373255e-47 ], 
                    [ 355.0 , 8.194335048039266e-47 ], [ 355.0625 , 8.195534763327288e-47 ], [ 355.125 , 8.196734654263033e-47 ], [ 355.1875 , 8.19793472087245e-47 ], 
                    [ 355.25 , 8.199134963180853e-47 ], [ 355.3125 , 8.200335381214431e-47 ], [ 355.375 , 8.201535974998559e-47 ], [ 355.4375 , 8.202736744559089e-47 ], 
                    [ 355.5 , 8.203937689921696e-47 ], [ 355.5625 , 8.2051388111123515e-47 ], [ 355.625 , 8.206340108156449e-47 ], [ 355.6875 , 8.207541581080026e-47 ], 
                    [ 355.75 , 8.208744220535179e-47 ], [ 355.8125 , 8.209947291129374e-47 ], [ 355.875 , 8.211150538045161e-47 ], [ 355.9375 , 8.212353961308324e-47 ], 
                    [ 356.0 , 8.213557560944884e-47 ], [ 356.0625 , 8.214761336980456e-47 ], [ 356.125 , 8.215965289441068e-47 ], [ 356.1875 , 8.21716941835246e-47 ], 
                    [ 356.25 , 8.218373723740669e-47 ], [ 356.3125 , 8.219578205631325e-47 ], [ 356.375 , 8.220782864050528e-47 ], [ 356.4375 , 8.221987699023919e-47 ], 
                    [ 356.5 , 8.2231927105776065e-47 ], [ 356.5625 , 8.224399070006215e-47 ], [ 356.625 , 8.225605743315553e-47 ], [ 356.6875 , 8.226812593666574e-47 ], 
                    [ 356.75 , 8.228019621085016e-47 ], [ 356.8125 , 8.229226825597036e-47 ], [ 356.875 , 8.230434207228616e-47 ], [ 356.9375 , 8.231641766005625e-47 ], 
                    [ 357.0 , 8.232849501954055e-47 ], [ 357.0625 , 8.234057415100017e-47 ], [ 357.125 , 8.235265505469567e-47 ], [ 357.1875 , 8.236473773088415e-47 ], 
                    [ 357.25 , 8.237682217982684e-47 ], [ 357.3125 , 8.238890840178674e-47 ], [ 357.375 , 8.240100805907176e-47 ], [ 357.4375 , 8.241310990084274e-47 ], 
                    [ 357.5 , 8.24252135199533e-47 ], [ 357.5625 , 8.243731891666508e-47 ], [ 357.625 , 8.244942609123854e-47 ], [ 357.6875 , 8.246153504393424e-47 ], 
                    [ 357.75 , 8.247364577501444e-47 ], [ 357.8125 , 8.248575828473861e-47 ], [ 357.875 , 8.249787257337029e-47 ], [ 357.9375 , 8.250998864116841e-47 ], 
                    [ 358.0 , 8.252210648839602e-47 ], [ 358.0625 , 8.253422611531562e-47 ], [ 358.125 , 8.254634762511061e-47 ], [ 358.1875 , 8.255848421598695e-47 ], 
                    [ 358.25 , 8.257062259127702e-47 ], [ 358.3125 , 8.258276275124381e-47 ], [ 358.375 , 8.25949046961491e-47 ], [ 358.4375 , 8.260704842625475e-47 ], 
                    [ 358.5 , 8.261919394182442e-47 ], [ 358.5625 , 8.263134124312e-47 ], [ 358.625 , 8.264349033040348e-47 ], [ 358.6875 , 8.265564120393805e-47 ], 
                    [ 358.75 , 8.266779386398629e-47 ], [ 358.8125 , 8.267994831081209e-47 ], [ 358.875 , 8.269210454467461e-47 ], [ 358.9375 , 8.27042627990093e-47 ], 
                    [ 359.0 , 8.271643436192212e-47 ], [ 359.0625 , 8.272860771611981e-47 ], [ 359.125 , 8.274078286186659e-47 ], [ 359.1875 , 8.275295979942729e-47 ], 
                    [ 359.25 , 8.276513852906386e-47 ], [ 359.3125 , 8.277731905104061e-47 ], [ 359.375 , 8.278950136562073e-47 ], [ 359.4375 , 8.280168547307042e-47 ], 
                    [ 359.5 , 8.281387137364938e-47 ], [ 359.5625 , 8.282605906762683e-47 ], [ 359.625 , 8.28382485552602e-47 ], [ 359.6875 , 8.285043983681937e-47 ], 
                    [ 359.75 , 8.286263294465833e-47 ], [ 359.8125 , 8.2874840353268275e-47 ], [ 359.875 , 8.288704956028568e-47 ], [ 359.9375 , 8.289926056597607e-47 ], 
                    [ 360.0 , 8.291147337060382e-47 ], [ 360.0625 , 8.292368797443574e-47 ], [ 360.125 , 8.29359043777351e-47 ], [ 360.1875 , 8.294812258076762e-47 ], 
                    [ 360.25 , 8.2960342583799e-47 ], [ 360.3125 , 8.297256438709385e-47 ], [ 360.375 , 8.298478799091618e-47 ], [ 360.4375 , 8.299701339553364e-47 ], 
                    [ 360.5 , 8.300924060120913e-47 ], [ 360.5625 , 8.302146960821037e-47 ], [ 360.625 , 8.30337112149855e-47 ], [ 360.6875 , 8.304595512575097e-47 ], 
                    [ 360.75 , 8.305820084196868e-47 ], [ 360.8125 , 8.307044836390249e-47 ], [ 360.875 , 8.308269769182162e-47 ], [ 360.9375 , 8.309494882599181e-47 ], 
                    [ 361.0 , 8.310720176667819e-47 ], [ 361.0625 , 8.311945651414715e-47 ], [ 361.125 , 8.313171306866629e-47 ], [ 361.1875 , 8.314397143050091e-47 ], 
                    [ 361.25 , 8.315623159991927e-47 ], [ 361.3125 , 8.316849357718674e-47 ], [ 361.375 , 8.318075736256871e-47 ], [ 361.4375 , 8.319303419099417e-47 ], 
                    [ 361.5 , 8.320531437290253e-47 ], [ 361.5625 , 8.321759636749655e-47 ], [ 361.625 , 8.322988017504613e-47 ], [ 361.6875 , 8.324216579581535e-47 ], 
                    [ 361.75 , 8.325445323007306e-47 ], [ 361.8125 , 8.326674247808869e-47 ], [ 361.875 , 8.327903354012941e-47 ], [ 361.9375 , 8.329132641646003e-47 ], 
                    [ 362.0 , 8.330362110735366e-47 ], [ 362.0625 , 8.331591761307283e-47 ], [ 362.125 , 8.332821593389018e-47 ], [ 362.1875 , 8.334051607006888e-47 ], 
                    [ 362.25 , 8.335282712423095e-47 ], [ 362.3125 , 8.336514266368119e-47 ], [ 362.375 , 8.337746002277593e-47 ], [ 362.4375 , 8.338977920178399e-47 ], 
                    [ 362.5 , 8.340210020097369e-47 ], [ 362.5625 , 8.341442302061635e-47 ], [ 362.625 , 8.342674766097736e-47 ], [ 362.6875 , 8.343907412232813e-47 ], 
                    [ 362.75 , 8.345140240493711e-47 ], [ 362.8125 , 8.346373250907518e-47 ], [ 362.875 , 8.347606443500673e-47 ], [ 362.9375 , 8.348839818300626e-47 ], 
                    [ 363.0 , 8.350073375333886e-47 ], [ 363.0625 , 8.351307824342266e-47 ], [ 363.125 , 8.352542859295955e-47 ], [ 363.1875 , 8.353778076893051e-47 ], 
                    [ 363.25 , 8.355013477160629e-47 ], [ 363.3125 , 8.35624906012558e-47 ], [ 363.375 , 8.357484825815044e-47 ], [ 363.4375 , 8.358720774255922e-47 ], 
                    [ 363.5 , 8.359956905475421e-47 ], [ 363.5625 , 8.361193219500274e-47 ], [ 363.625 , 8.362429716357868e-47 ], [ 363.6875 , 8.363666396075008e-47 ], 
                    [ 363.75 , 8.364903258678794e-47 ], [ 363.8125 , 8.366140304196152e-47 ], [ 363.875 , 8.367378103105868e-47 ], [ 363.9375 , 8.368616722969295e-47 ], 
                    [ 364.0 , 8.369855526185038e-47 ], [ 364.0625 , 8.371094512780539e-47 ], [ 364.125 , 8.372333682782645e-47 ], [ 364.1875 , 8.373573036218625e-47 ], 
                    [ 364.25 , 8.374812573115571e-47 ], [ 364.3125 , 8.376052293500704e-47 ], [ 364.375 , 8.377292197401361e-47 ], [ 364.4375 , 8.378532284844412e-47 ], 
                    [ 364.5 , 8.379772555857146e-47 ], [ 364.5625 , 8.381013010466675e-47 ], [ 364.625 , 8.382253648700359e-47 ], [ 364.6875 , 8.383494774383172e-47 ], 
                    [ 364.75 , 8.384736877689965e-47 ], [ 364.8125 , 8.385979165027603e-47 ], [ 364.875 , 8.387221636423114e-47 ], [ 364.9375 , 8.388464291904006e-47 ], 
                    [ 365.0 , 8.389707131497432e-47 ], [ 365.0625 , 8.390950155230554e-47 ], [ 365.125 , 8.392193363130832e-47 ], [ 365.1875 , 8.393436755225432e-47 ], 
                    [ 365.25 , 8.394680331541886e-47 ], [ 365.3125 , 8.395924092107126e-47 ], [ 365.375 , 8.397168036948753e-47 ], [ 365.4375 , 8.398412166093827e-47 ], 
                    [ 365.5 , 8.399656539994072e-47 ], [ 365.5625 , 8.400902207209633e-47 ], [ 365.625 , 8.402148059157249e-47 ], [ 365.6875 , 8.403394095864433e-47 ], 
                    [ 365.75 , 8.404640317358767e-47 ], [ 365.8125 , 8.405886723667355e-47 ], [ 365.875 , 8.407133314817663e-47 ], [ 365.9375 , 8.408380090837224e-47 ], 
                    [ 366.0 , 8.409627051753515e-47 ], [ 366.0625 , 8.410874197593656e-47 ], [ 366.125 , 8.412121528385249e-47 ], [ 366.1875 , 8.413369044155845e-47 ], 
                    [ 366.25 , 8.414616744932635e-47 ], [ 366.3125 , 8.415864630743234e-47 ], [ 366.375 , 8.4171135270858845e-47 ], [ 366.4375 , 8.418362819721203e-47 ], 
                    [ 366.5 , 8.419612297780074e-47 ], [ 366.5625 , 8.420861961290315e-47 ], [ 366.625 , 8.422111810279154e-47 ], [ 366.6875 , 8.4233618447743e-47 ], 
                    [ 366.75 , 8.424612064803165e-47 ], [ 366.8125 , 8.425862470393348e-47 ], [ 366.875 , 8.42711306157263e-47 ], [ 366.9375 , 8.428363838368016e-47 ], 
                    [ 367.0 , 8.4296148008074175e-47 ], [ 367.0625 , 8.430865948918326e-47 ], [ 367.125 , 8.43211728272836e-47 ], [ 367.1875 , 8.433369391789916e-47 ], 
                    [ 367.25 , 8.434622250557206e-47 ], [ 367.3125 , 8.435875295448928e-47 ], [ 367.375 , 8.437128526492616e-47 ], [ 367.4375 , 8.438381943715864e-47 ], 
                    [ 367.5 , 8.439635547146448e-47 ], [ 367.5625 , 8.440889336811914e-47 ], [ 367.625 , 8.442143312739988e-47 ], [ 367.6875 , 8.443397474958339e-47 ], 
                    [ 367.75 , 8.444651823494766e-47 ], [ 367.8125 , 8.445906358376705e-47 ], [ 367.875 , 8.447161079631901e-47 ], [ 367.9375 , 8.448415987288223e-47 ], 
                    [ 368.0 , 8.449671290462265e-47 ], [ 368.0625 , 8.450927638845966e-47 ], [ 368.125 , 8.452184174031225e-47 ], [ 368.1875 , 8.4534408960457585e-47 ], 
                    [ 368.25 , 8.454697804917284e-47 ], [ 368.3125 , 8.455954900673645e-47 ], [ 368.375 , 8.457212183342687e-47 ], [ 368.4375 , 8.458469652952145e-47 ], 
                    [ 368.5 , 8.459727309529692e-47 ], [ 368.5625 , 8.460985153103368e-47 ], [ 368.625 , 8.462243183700736e-47 ], [ 368.6875 , 8.463501401349906e-47 ], 
                    [ 368.75 , 8.464759806078448e-47 ], [ 368.8125 , 8.466018397914057e-47 ], [ 368.875 , 8.467278130767481e-47 ], [ 368.9375 , 8.468538219107981e-47 ], 
                    [ 369.0 , 8.469798494972986e-47 ], [ 369.0625 , 8.47105895839052e-47 ], [ 369.125 , 8.472319609388314e-47 ], [ 369.1875 , 8.473580447994524e-47 ], 
                    [ 369.25 , 8.47484147423695e-47 ], [ 369.3125 , 8.476102688143517e-47 ], [ 369.375 , 8.477364089742031e-47 ], [ 369.4375 , 8.478625679060604e-47 ], 
                    [ 369.5 , 8.479887456127232e-47 ], [ 369.5625 , 8.481149420969498e-47 ], [ 369.625 , 8.482411573615824e-47 ], [ 369.6875 , 8.483674379928942e-47 ], 
                    [ 369.75 , 8.484937890428083e-47 ], [ 369.8125 , 8.4862015891073935e-47 ], [ 369.875 , 8.487465475994722e-47 ], [ 369.9375 , 8.488729551118398e-47 ], 
                    [ 370.0 , 8.489993814506096e-47 ], [ 370.0625 , 8.491258266186095e-47 ], [ 370.125 , 8.492522906186318e-47 ], [ 370.1875 , 8.493787734534934e-47 ], 
                    [ 370.25 , 8.495052751259934e-47 ], [ 370.3125 , 8.496317956389433e-47 ], [ 370.375 , 8.497583349951369e-47 ], [ 370.4375 , 8.498848931973991e-47 ], 
                    [ 370.5 , 8.500114776999617e-47 ], [ 370.5625 , 8.501381826870708e-47 ], [ 370.625 , 8.502649065611829e-47 ], [ 370.6875 , 8.503916493250887e-47 ], 
                    [ 370.75 , 8.505184109815985e-47 ], [ 370.8125 , 8.506451915335584e-47 ], [ 370.875 , 8.507719909837547e-47 ], [ 370.9375 , 8.508988093350225e-47 ], 
                    [ 371.0 , 8.510256465901736e-47 ], [ 371.0625 , 8.511525027520073e-47 ], [ 371.125 , 8.512793778233844e-47 ], [ 371.1875 , 8.514062718070873e-47 ], 
                    [ 371.25 , 8.515331847059473e-47 ], [ 371.3125 , 8.516601165227839e-47 ], [ 371.375 , 8.51787130315636e-47 ], [ 371.4375 , 8.5191419979733305e-47 ], 
                    [ 371.5 , 8.52041288235225e-47 ], [ 371.5625 , 8.52168395632164e-47 ], [ 371.625 , 8.522955219909662e-47 ], [ 371.6875 , 8.5242266731446635e-47 ], 
                    [ 371.75 , 8.525498316054876e-47 ], [ 371.8125 , 8.526770148668474e-47 ], [ 371.875 , 8.528042171013879e-47 ], [ 371.9375 , 8.529314383119455e-47 ], 
                    [ 372.0 , 8.530586785013449e-47 ], [ 372.0625 , 8.531859376724177e-47 ], [ 372.125 , 8.533132158279951e-47 ], [ 372.1875 , 8.53440530428452e-47 ], 
                    [ 372.25 , 8.535679525116016e-47 ], [ 372.3125 , 8.536953936193717e-47 ], [ 372.375 , 8.538228537546146e-47 ], [ 372.4375 , 8.539503329201714e-47 ], 
                    [ 372.5 , 8.54077831118859e-47 ], [ 372.5625 , 8.542053483535497e-47 ], [ 372.625 , 8.54332884627061e-47 ], [ 372.6875 , 8.544604399422723e-47 ], 
                    [ 372.75 , 8.545880143019898e-47 ], [ 372.8125 , 8.5471560770906925e-47 ], [ 372.875 , 8.548432201663423e-47 ], [ 372.9375 , 8.5497085167668365e-47 ], 
                    [ 373.0 , 8.550985022429077e-47 ], [ 373.0625 , 8.552262363909484e-47 ], [ 373.125 , 8.553540217856086e-47 ], [ 373.1875 , 8.554818262736059e-47 ], 
                    [ 373.25 , 8.556096498577684e-47 ], [ 373.3125 , 8.55737492540956e-47 ], [ 373.375 , 8.558653543260222e-47 ], [ 373.4375 , 8.559932352158209e-47 ], 
                    [ 373.5 , 8.561211352132132e-47 ], [ 373.5625 , 8.562490543210355e-47 ], [ 373.625 , 8.563769925421617e-47 ], [ 373.6875 , 8.565049498794474e-47 ], 
                    [ 373.75 , 8.566329263357371e-47 ], [ 373.8125 , 8.567609219138932e-47 ], [ 373.875 , 8.56888951032753e-47 ], [ 373.9375 , 8.57017088443961e-47 ], 
                    [ 374.0 , 8.571452450165739e-47 ], [ 374.0625 , 8.57273420753451e-47 ], [ 374.125 , 8.574016156574641e-47 ], [ 374.1875 , 8.575298297314854e-47 ], 
                    [ 374.25 , 8.576580629783697e-47 ], [ 374.3125 , 8.577863154010083e-47 ], [ 374.375 , 8.579145870022198e-47 ], [ 374.4375 , 8.58042877784921e-47 ], 
                    [ 374.5 , 8.58171187751962e-47 ], [ 374.5625 , 8.582995169062115e-47 ], [ 374.625 , 8.584278652505327e-47 ], [ 374.6875 , 8.585562327878134e-47 ], 
                    [ 374.75 , 8.586846742753352e-47 ], [ 374.8125 , 8.588131745555769e-47 ], [ 374.875 , 8.589416940655857e-47 ], [ 374.9375 , 8.590702328082698e-47 ], 
                    [ 375.0 , 8.59198790786495e-47 ], [ 375.0625 , 8.593273680031342e-47 ], [ 375.125 , 8.59455964461072e-47 ], [ 375.1875 , 8.595845801632001e-47 ], 
                    [ 375.25 , 8.597132151123741e-47 ], [ 375.3125 , 8.598418693114926e-47 ], [ 375.375 , 8.599705427634179e-47 ], [ 375.4375 , 8.600992354710493e-47 ], 
                    [ 375.5 , 8.602279474372748e-47 ], [ 375.5625 , 8.60356678664958e-47 ], [ 375.625 , 8.604855257132272e-47 ], [ 375.6875 , 8.606143929307455e-47 ], 
                    [ 375.75 , 8.607432794475477e-47 ], [ 375.8125 , 8.608721852665303e-47 ], [ 375.875 , 8.610011103905841e-47 ], [ 375.9375 , 8.611300548225939e-47 ], 
                    [ 376.0 , 8.612590185654696e-47 ], [ 376.0625 , 8.61388001622079e-47 ], [ 376.125 , 8.61517003995345e-47 ], [ 376.1875 , 8.616460256881237e-47 ], 
                    [ 376.25 , 8.617750667033268e-47 ], [ 376.3125 , 8.619041270438481e-47 ], [ 376.375 , 8.620332067126002e-47 ], [ 376.4375 , 8.621623398089903e-47 ], 
                    [ 376.5 , 8.622915501355778e-47 ], [ 376.5625 , 8.624207798266258e-47 ], [ 376.625 , 8.625500288850363e-47 ], [ 376.6875 , 8.626792973137119e-47 ], 
                    [ 376.75 , 8.628085851155494e-47 ], [ 376.8125 , 8.629378922934645e-47 ], [ 376.875 , 8.630672188503548e-47 ], [ 376.9375 , 8.631965647891187e-47 ], 
                    [ 377.0 , 8.633259301126729e-47 ], [ 377.0625 , 8.634553148239167e-47 ], [ 377.125 , 8.635847189257553e-47 ], [ 377.1875 , 8.637141424210952e-47 ], 
                    [ 377.25 , 8.638435853128426e-47 ], [ 377.3125 , 8.639731165496772e-47 ], [ 377.375 , 8.641026940895682e-47 ], [ 377.4375 , 8.642322910633142e-47 ], 
                    [ 377.5 , 8.643619074738544e-47 ], [ 377.5625 , 8.644915433240855e-47 ], [ 377.625 , 8.646211986169291e-47 ], [ 377.6875 , 8.647508733553074e-47 ], 
                    [ 377.75 , 8.648805675421305e-47 ], [ 377.8125 , 8.650102811803032e-47 ], [ 377.875 , 8.651400142727673e-47 ], [ 377.9375 , 8.65269766822422e-47 ], 
                    [ 378.0 , 8.65399538832204e-47 ], [ 378.0625 , 8.655293303050073e-47 ], [ 378.125 , 8.656591447438158e-47 ], [ 378.1875 , 8.657890686505222e-47 ], 
                    [ 378.25 , 8.659190120570826e-47 ], [ 378.3125 , 8.660489749664113e-47 ], [ 378.375 , 8.661789573814414e-47 ], [ 378.4375 , 8.663089593050945e-47 ], 
                    [ 378.5 , 8.664389807403108e-47 ], [ 378.5625 , 8.665690216900064e-47 ], [ 378.625 , 8.666990821571226e-47 ], [ 378.6875 , 8.668291621445885e-47 ], 
                    [ 378.75 , 8.669592616553218e-47 ], [ 378.8125 , 8.670893806922585e-47 ], [ 378.875 , 8.672195192583295e-47 ], [ 378.9375 , 8.673496773564659e-47 ], 
                    [ 379.0 , 8.674798852899528e-47 ], [ 379.0625 , 8.676101759069358e-47 ], [ 379.125 , 8.677404860928332e-47 ], [ 379.1875 , 8.678708158505902e-47 ], 
                    [ 379.25 , 8.680011651831463e-47 ], [ 379.3125 , 8.681315340934601e-47 ], [ 379.375 , 8.682619225844289e-47 ], [ 379.4375 , 8.683923306590183e-47 ], 
                    [ 379.5 , 8.68522758320182e-47 ], [ 379.5625 , 8.686532055708493e-47 ], [ 379.625 , 8.687836724139566e-47 ], [ 379.6875 , 8.689141588524647e-47 ], 
                    [ 379.75 , 8.690446648892924e-47 ], [ 379.8125 , 8.691751905274013e-47 ], [ 379.875 , 8.693057846428557e-47 ], [ 379.9375 , 8.694364337531303e-47 ], 
                    [ 380.0 , 8.695671024988255e-47 ], [ 380.0625 , 8.696977908829045e-47 ], [ 380.125 , 8.698284989083128e-47 ], [ 380.1875 , 8.699592265780024e-47 ], 
                    [ 380.25 , 8.700899738949314e-47 ], [ 380.3125 , 8.7022074086204065e-47 ], [ 380.375 , 8.703515274822833e-47 ], [ 380.4375 , 8.704823337586254e-47 ], 
                    [ 380.5 , 8.706131596940087e-47 ], [ 380.5625 , 8.707440052914e-47 ], [ 380.625 , 8.708748705537547e-47 ], [ 380.6875 , 8.710057554840157e-47 ], 
                    [ 380.75 , 8.711367315696734e-47 ], [ 380.8125 , 8.712677446914643e-47 ], [ 380.875 , 8.713987775167438e-47 ], [ 380.9375 , 8.715298300484815e-47 ], 
                    [ 381.0 , 8.716609022896409e-47 ], [ 381.0625 , 8.717919942431864e-47 ], [ 381.125 , 8.719231059120825e-47 ], [ 381.1875 , 8.720542372993004e-47 ], 
                    [ 381.25 , 8.72185388407787e-47 ], [ 381.3125 , 8.72316559240533e-47 ], [ 381.375 , 8.724477498004865e-47 ], [ 381.4375 , 8.725789600906202e-47 ], 
                    [ 381.5 , 8.727101901139015e-47 ], [ 381.5625 , 8.728414398732921e-47 ], [ 381.625 , 8.729728048114797e-47 ], [ 381.6875 , 8.731041896887267e-47 ], 
                    [ 381.75 , 8.732355943397758e-47 ], [ 381.8125 , 8.733670187676027e-47 ], [ 381.875 , 8.734984629751656e-47 ], [ 381.9375 , 8.736299269654598e-47 ], 
                    [ 382.0 , 8.737614107414504e-47 ], [ 382.0625 , 8.738929143061213e-47 ], [ 382.125 , 8.74024437662457e-47 ], [ 382.1875 , 8.741559808134361e-47 ], 
                    [ 382.25 , 8.742875437620194e-47 ], [ 382.3125 , 8.744191265111985e-47 ], [ 382.375 , 8.745507290639538e-47 ], [ 382.4375 , 8.746823638282031e-47 ], 
                    [ 382.5 , 8.7481408330945735e-47 ], [ 382.5625 , 8.749458226265211e-47 ], [ 382.625 , 8.750775817823687e-47 ], [ 382.6875 , 8.752093607799755e-47 ], 
                    [ 382.75 , 8.75341159622348e-47 ], [ 382.8125 , 8.754729783124686e-47 ], [ 382.875 , 8.756048168533324e-47 ], [ 382.9375 , 8.757366752479161e-47 ], 
                    [ 383.0 , 8.758685534992096e-47 ], [ 383.0625 , 8.760004516102281e-47 ], [ 383.125 , 8.761323695839312e-47 ], [ 383.1875 , 8.762643074233163e-47 ], 
                    [ 383.25 , 8.763962651313935e-47 ], [ 383.3125 , 8.76528267645117e-47 ], [ 383.375 , 8.766603507410958e-47 ], [ 383.4375 , 8.767924537405335e-47 ], 
                    [ 383.5 , 8.769245766464352e-47 ], [ 383.5625 , 8.770567194617826e-47 ], [ 383.625 , 8.771888821896062e-47 ], [ 383.6875 , 8.773210648328758e-47 ], 
                    [ 383.75 , 8.77453267394605e-47 ], [ 383.8125 , 8.775854898777888e-47 ], [ 383.875 , 8.777177322854604e-47 ], [ 383.9375 , 8.778499946205787e-47 ], 
                    [ 384.0 , 8.779822768861588e-47 ], [ 384.0625 , 8.78114579085229e-47 ], [ 384.125 , 8.782469012207556e-47 ], [ 384.1875 , 8.783792779092718e-47 ], 
                    [ 384.25 , 8.785117285706446e-47 ], [ 384.3125 , 8.786441992042301e-47 ], [ 384.375 , 8.787766898130397e-47 ], [ 384.4375 , 8.78909200400092e-47 ], 
                    [ 384.5 , 8.79041730968393e-47 ], [ 384.5625 , 8.791742815209495e-47 ], [ 384.625 , 8.79306852060775e-47 ], [ 384.6875 , 8.794394425908958e-47 ], 
                    [ 384.75 , 8.795720531143262e-47 ], [ 384.8125 , 8.797046836340622e-47 ], [ 384.875 , 8.798373341531444e-47 ], [ 384.9375 , 8.79970004674563e-47 ], 
                    [ 385.0 , 8.801026952013719e-47 ], [ 385.0625 , 8.802354425014517e-47 ], [ 385.125 , 8.803682532683522e-47 ], [ 385.1875 , 8.80501084073881e-47 ], 
                    [ 385.25 , 8.806339349210363e-47 ], [ 385.3125 , 8.80766805812855e-47 ], [ 385.375 , 8.808996967523672e-47 ], [ 385.4375 , 8.810326077425857e-47 ], 
                    [ 385.5 , 8.81165538786554e-47 ], [ 385.5625 , 8.812984898872794e-47 ], [ 385.625 , 8.81431461047794e-47 ], [ 385.6875 , 8.815644522711253e-47 ], 
                    [ 385.75 , 8.816974635602929e-47 ], [ 385.8125 , 8.818304949183316e-47 ], [ 385.875 , 8.81963546348269e-47 ], [ 385.9375 , 8.820966575080085e-47 ], 
                    [ 386.0 , 8.822298292910322e-47 ], [ 386.0625 , 8.82363021179249e-47 ], [ 386.125 , 8.824962331756882e-47 ], [ 386.1875 , 8.826294652833978e-47 ], 
                    [ 386.25 , 8.827627175054082e-47 ], [ 386.3125 , 8.828959898447497e-47 ], [ 386.375 , 8.83029282304478e-47 ], [ 386.4375 , 8.831625948875999e-47 ], 
                    [ 386.5 , 8.832959275971967e-47 ], [ 386.5625 , 8.83429280436251e-47 ], [ 386.625 , 8.83562653407864e-47 ], [ 386.6875 , 8.836960465150195e-47 ], 
                    [ 386.75 , 8.83829459760788e-47 ], [ 386.8125 , 8.839629316253704e-47 ], [ 386.875 , 8.84096462252553e-47 ], [ 386.9375 , 8.842300130507493e-47 ], 
                    [ 387.0 , 8.84363584023006e-47 ], [ 387.0625 , 8.844971751723775e-47 ], [ 387.125 , 8.846307865018798e-47 ], [ 387.1875 , 8.84764418014612e-47 ], 
                    [ 387.25 , 8.848980697135911e-47 ], [ 387.3125 , 8.850317416018792e-47 ], [ 387.375 , 8.851654336825008e-47 ], [ 387.4375 , 8.852991459585505e-47 ], 
                    [ 387.5 , 8.854328784330283e-47 ], [ 387.5625 , 8.85566631109023e-47 ], [ 387.625 , 8.85700403989568e-47 ], [ 387.6875 , 8.85834233450368e-47 ], 
                    [ 387.75 , 8.85968123826997e-47 ], [ 387.8125 , 8.861020344406263e-47 ], [ 387.875 , 8.862359652943203e-47 ], [ 387.9375 , 8.86369916391145e-47 ], 
                    [ 388.0 , 8.865038877341537e-47 ], [ 388.0625 , 8.866378793264e-47 ], [ 388.125 , 8.86771891170951e-47 ], [ 388.1875 , 8.869059232708739e-47 ], 
                    [ 388.25 , 8.870399756292177e-47 ], [ 388.3125 , 8.871740482490571e-47 ], [ 388.375 , 8.87308141133448e-47 ], [ 388.4375 , 8.874422542854537e-47 ], 
                    [ 388.5 , 8.875763877081309e-47 ], [ 388.5625 , 8.877105750791204e-47 ], [ 388.625 , 8.878448305825979e-47 ], [ 388.6875 , 8.879791063905924e-47 ], 
                    [ 388.75 , 8.881134025061686e-47 ], [ 388.8125 , 8.882477189324295e-47 ], [ 388.875 , 8.883820556724147e-47 ], [ 388.9375 , 8.885164127292097e-47 ], 
                    [ 389.0 , 8.886507901058869e-47 ], [ 389.0625 , 8.887851878055256e-47 ], [ 389.125 , 8.889196058311995e-47 ], [ 389.1875 , 8.890540441859639e-47 ], 
                    [ 389.25 , 8.891885028729123e-47 ], [ 389.3125 , 8.893229818951066e-47 ], [ 389.375 , 8.894574812556353e-47 ], [ 389.4375 , 8.89592024056489e-47 ], 
                    [ 389.5 , 8.897266357640443e-47 ], [ 389.5625 , 8.898612678408317e-47 ], [ 389.625 , 8.899959202899267e-47 ], [ 389.6875 , 8.901305931144438e-47 ], 
                    [ 389.75 , 8.902652863174409e-47 ], [ 389.8125 , 8.90399999902008e-47 ], [ 389.875 , 8.905347338712354e-47 ], [ 389.9375 , 8.906694882281824e-47 ], 
                    [ 390.0 , 8.908042629759658e-47 ], [ 390.0625 , 8.909390581176646e-47 ], [ 390.125 , 8.910738736563586e-47 ], [ 390.1875 , 8.912087095951278e-47 ], 
                    [ 390.25 , 8.913435659370656e-47 ], [ 390.3125 , 8.9147845787853e-47 ], [ 390.375 , 8.916134311380025e-47 ], [ 390.4375 , 8.917484248329446e-47 ], 
                    [ 390.5 , 8.918834389664498e-47 ], [ 390.5625 , 8.92018473541607e-47 ], [ 390.625 , 8.921535285615232e-47 ], [ 390.6875 , 8.922886040292877e-47 ], 
                    [ 390.75 , 8.924236999479964e-47 ], [ 390.8125 , 8.92558816320739e-47 ], [ 390.875 , 8.92693953150632e-47 ], [ 390.9375 , 8.928291104407467e-47 ], 
                    [ 391.0 , 8.929642881942066e-47 ], [ 391.0625 , 8.930994864140905e-47 ], [ 391.125 , 8.932347051034976e-47 ], [ 391.1875 , 8.933699477739103e-47 ], 
                    [ 391.25 , 8.93505285735896e-47 ], [ 391.3125 , 8.936406442004144e-47 ], [ 391.375 , 8.937760231705976e-47 ], [ 391.4375 , 8.939114226495452e-47 ], 
                    [ 391.5 , 8.940468426403645e-47 ], [ 391.5625 , 8.941822831461624e-47 ], [ 391.625 , 8.94317744170047e-47 ], [ 391.6875 , 8.944532257151203e-47 ], 
                    [ 391.75 , 8.945887277845227e-47 ], [ 391.8125 , 8.947242503813127e-47 ], [ 391.875 , 8.948597935086256e-47 ], [ 391.9375 , 8.949953571695842e-47 ], 
                    [ 392.0 , 8.951309413672927e-47 ], [ 392.0625 , 8.952665461048434e-47 ], [ 392.125 , 8.95402228160408e-47 ], [ 392.1875 , 8.955379401708708e-47 ], 
                    [ 392.25 , 8.956736727505745e-47 ], [ 392.3125 , 8.958094259026493e-47 ], [ 392.375 , 8.959451996301945e-47 ], [ 392.4375 , 8.960809939363536e-47 ], 
                    [ 392.5 , 8.962168088242267e-47 ], [ 392.5625 , 8.963526442969398e-47 ], [ 392.625 , 8.964885003576062e-47 ], [ 392.6875 , 8.96624377009372e-47 ], 
                    [ 392.75 , 8.967602742553264e-47 ], [ 392.8125 , 8.96896192098603e-47 ], [ 392.875 , 8.970321305423305e-47 ], [ 392.9375 , 8.971680895896118e-47 ], 
                    [ 393.0 , 8.973041119970665e-47 ], [ 393.0625 , 8.974401791801349e-47 ], [ 393.125 , 8.975762669964416e-47 ], [ 393.1875 , 8.977123754490966e-47 ], 
                    [ 393.25 , 8.978485045412355e-47 ], [ 393.3125 , 8.979846542759817e-47 ], [ 393.375 , 8.981208246564844e-47 ], [ 393.4375 , 8.98257015685849e-47 ], 
                    [ 393.5 , 8.983932273672258e-47 ], [ 393.5625 , 8.985294597037337e-47 ], [ 393.625 , 8.986657126985112e-47 ], [ 393.6875 , 8.9880198635471e-47 ], 
                    [ 393.75 , 8.989382806754313e-47 ], [ 393.8125 , 8.990745956638155e-47 ], [ 393.875 , 8.992109613041083e-47 ], [ 393.9375 , 8.9934739495625e-47 ], 
                    [ 394.0 , 8.99483849308913e-47 ], [ 394.0625 , 8.996203243652577e-47 ], [ 394.125 , 8.997568201284056e-47 ], [ 394.1875 , 8.99893336601505e-47 ], 
                    [ 394.25 , 9.000298737876984e-47 ], [ 394.3125 , 9.001664316901281e-47 ], [ 394.375 , 9.003030103119312e-47 ], [ 394.4375 , 9.004396096562707e-47 ], 
                    [ 394.5 , 9.00576229726271e-47 ], [ 394.5625 , 9.007128705250839e-47 ], [ 394.625 , 9.0084953205586e-47 ], [ 394.6875 , 9.009862143217265e-47 ], 
                    [ 394.75 , 9.011229238981248e-47 ], [ 394.8125 , 9.012597106553675e-47 ], [ 394.875 , 9.013965181762586e-47 ], [ 394.9375 , 9.015333464639693e-47 ], 
                    [ 395.0 , 9.016701955216644e-47 ], [ 395.0625 , 9.018070653524781e-47 ], [ 395.125 , 9.019439559595693e-47 ], [ 395.1875 , 9.020808673460926e-47 ], 
                    [ 395.25 , 9.022177995152082e-47 ], [ 395.3125 , 9.023547524700578e-47 ], [ 395.375 , 9.024917262137968e-47 ], [ 395.4375 , 9.026287207496004e-47 ], 
                    [ 395.5 , 9.027657360806115e-47 ], [ 395.5625 , 9.029027722099804e-47 ], [ 395.625 , 9.03039829140877e-47 ], [ 395.6875 , 9.031769600242748e-47 ], 
                    [ 395.75 , 9.033141259913968e-47 ], [ 395.8125 , 9.034513127899874e-47 ], [ 395.875 , 9.035885204232297e-47 ], [ 395.9375 , 9.037257488942557e-47 ], 
                    [ 396.0 , 9.038629982062427e-47 ], [ 396.0625 , 9.040002683623625e-47 ], [ 396.125 , 9.041375593657677e-47 ], [ 396.1875 , 9.04274871219631e-47 ], 
                    [ 396.25 , 9.044122039271251e-47 ], [ 396.3125 , 9.04549557491398e-47 ], [ 396.375 , 9.04686931915643e-47 ], [ 396.4375 , 9.048243272030088e-47 ], 
                    [ 396.5 , 9.049617433566765e-47 ], [ 396.5625 , 9.050992096861436e-47 ], [ 396.625 , 9.05236734177556e-47 ], [ 396.6875 , 9.053742795650107e-47 ], 
                    [ 396.75 , 9.055118458516702e-47 ], [ 396.8125 , 9.056494330407423e-47 ], [ 396.875 , 9.057870411353574e-47 ], [ 396.9375 , 9.059246701387054e-47 ], 
                    [ 397.0 , 9.060623200539691e-47 ], [ 397.0625 , 9.06199990884346e-47 ], [ 397.125 , 9.063376826329683e-47 ], [ 397.1875 , 9.064753953030662e-47 ], 
                    [ 397.25 , 9.0661312889778e-47 ], [ 397.3125 , 9.067508834203019e-47 ], [ 397.375 , 9.068886588738181e-47 ], [ 397.4375 , 9.070264589397367e-47 ], 
                    [ 397.5 , 9.071643382396671e-47 ], [ 397.5625 , 9.073022384989851e-47 ], [ 397.625 , 9.074401597208377e-47 ], [ 397.6875 , 9.07578101908444e-47 ], 
                    [ 397.75 , 9.077160650649779e-47 ], [ 397.8125 , 9.07854049193627e-47 ], [ 397.875 , 9.079920542975795e-47 ], [ 397.9375 , 9.081300803800237e-47 ], 
                    [ 398.0 , 9.082681274441359e-47 ], [ 398.0625 , 9.084061954931248e-47 ], [ 398.125 , 9.085442845301867e-47 ], [ 398.1875 , 9.086823945584864e-47 ], 
                    [ 398.25 , 9.08820525581234e-47 ], [ 398.3125 , 9.089586776016214e-47 ], [ 398.375 , 9.090968925644116e-47 ], [ 398.4375 , 9.09235151555106e-47 ], 
                    [ 398.5 , 9.093734315727463e-47 ], [ 398.5625 , 9.095117326205686e-47 ], [ 398.625 , 9.096500547017327e-47 ], [ 398.6875 , 9.097883978194696e-47 ], 
                    [ 398.75 , 9.09926761976966e-47 ], [ 398.8125 , 9.100651471774216e-47 ], [ 398.875 , 9.102035534240301e-47 ], [ 398.9375 , 9.103419807199988e-47 ], 
                    [ 399.0 , 9.10480429068542e-47 ], [ 399.0625 , 9.106188984728352e-47 ], [ 399.125 , 9.107573889361007e-47 ], [ 399.1875 , 9.10895900461534e-47 ], 
                    [ 399.25 , 9.110344462876776e-47 ], [ 399.3125 , 9.11173066241944e-47 ], [ 399.375 , 9.11311707288159e-47 ], [ 399.4375 , 9.114503694295318e-47 ], 
                    [ 399.5 , 9.11589052669266e-47 ], [ 399.5625 , 9.117277570105844e-47 ], [ 399.625 , 9.118664824566848e-47 ], [ 399.6875 , 9.120052290107984e-47 ], 
                    [ 399.75 , 9.121439966761104e-47 ], [ 399.8125 , 9.122827854558593e-47 ], [ 399.875 , 9.124215953532317e-47 ], [ 399.9375 , 9.125604263714603e-47 ], 
                    [ 400.0 , 9.126992785137525e-47 ], [ 400.0625 , 9.128381517833156e-47 ], [ 400.125 , 9.12977046183384e-47 ], [ 400.1875 , 9.131159997914277e-47 ], 
                    [ 400.25 , 9.132549892982158e-47 ], [ 400.3125 , 9.133939999612108e-47 ], [ 400.375 , 9.13533031783659e-47 ], [ 400.4375 , 9.136720847687554e-47 ], 
                    [ 400.5 , 9.138111589197472e-47 ], [ 400.5625 , 9.139502542398366e-47 ], [ 400.625 , 9.140893707322523e-47 ], [ 400.6875 , 9.14228508400204e-47 ], 
                    [ 400.75 , 9.143676672469541e-47 ], [ 400.8125 , 9.145068472756873e-47 ], [ 400.875 , 9.146460484896405e-47 ], [ 400.9375 , 9.147852708920324e-47 ], 
                    [ 401.0 , 9.149245144861007e-47 ], [ 401.0625 , 9.150637929455322e-47 ], [ 401.125 , 9.15203144316015e-47 ], [ 401.1875 , 9.153425169077553e-47 ], 
                    [ 401.25 , 9.15481910723985e-47 ], [ 401.3125 , 9.156213257679493e-47 ], [ 401.375 , 9.157607620428545e-47 ], [ 401.4375 , 9.159002195519535e-47 ], 
                    [ 401.5 , 9.160396982984672e-47 ], [ 401.5625 , 9.16179198285636e-47 ], [ 401.625 , 9.163187195166947e-47 ], [ 401.6875 , 9.164582619948912e-47 ], 
                    [ 401.75 , 9.16597825723429e-47 ], [ 401.8125 , 9.1673741070557e-47 ], [ 401.875 , 9.168770169445638e-47 ], [ 401.9375 , 9.170166444436085e-47 ], 
                    [ 402.0 , 9.171563291449954e-47 ], [ 402.0625 , 9.172960532605581e-47 ], [ 402.125 , 9.174357986623812e-47 ], [ 402.1875 , 9.17575565353701e-47 ], 
                    [ 402.25 , 9.177153533377675e-47 ], [ 402.3125 , 9.178551626178245e-47 ], [ 402.375 , 9.179949931971162e-47 ], [ 402.4375 , 9.181348450788943e-47 ], 
                    [ 402.5 , 9.182747182663904e-47 ], [ 402.5625 , 9.184146127628577e-47 ], [ 402.625 , 9.185545285715548e-47 ], [ 402.6875 , 9.186944656956965e-47 ], 
                    [ 402.75 , 9.188344241385623e-47 ], [ 402.8125 , 9.189744039033805e-47 ], [ 402.875 , 9.191144098259482e-47 ], [ 402.9375 , 9.19254487774515e-47 ], 
                    [ 403.0 , 9.193945870717164e-47 ], [ 403.0625 , 9.195347077207929e-47 ], [ 403.125 , 9.196748497250054e-47 ], [ 403.1875 , 9.198150130876016e-47 ], 
                    [ 403.25 , 9.199551978118567e-47 ], [ 403.3125 , 9.200954039010129e-47 ], [ 403.375 , 9.202356313583198e-47 ], [ 403.4375 , 9.20375880187041e-47 ], 
                    [ 403.5 , 9.205161503904265e-47 ], [ 403.5625 , 9.20656441971754e-47 ], [ 403.625 , 9.207967549342683e-47 ], [ 403.6875 , 9.209370892812345e-47 ], 
                    [ 403.75 , 9.210774450158855e-47 ], [ 403.8125 , 9.21217849953671e-47 ], [ 403.875 , 9.21358306856697e-47 ], [ 403.9375 , 9.214987851750133e-47 ], 
                    [ 404.0 , 9.216392849118717e-47 ], [ 404.0625 , 9.217798060705515e-47 ], [ 404.125 , 9.219203486543118e-47 ], [ 404.1875 , 9.22060912666413e-47 ], 
                    [ 404.25 , 9.222014981101286e-47 ], [ 404.3125 , 9.22342104988746e-47 ], [ 404.375 , 9.22482733305494e-47 ], [ 404.4375 , 9.226233830636675e-47 ], 
                    [ 404.5 , 9.227640542665225e-47 ], [ 404.5625 , 9.229047469173484e-47 ], [ 404.625 , 9.23045461019389e-47 ], [ 404.6875 , 9.231861965759346e-47 ], 
                    [ 404.75 , 9.233269953018017e-47 ], [ 404.8125 , 9.234678238770808e-47 ], [ 404.875 , 9.236086739319525e-47 ], [ 404.9375 , 9.23749545469706e-47 ], 
                    [ 405.0 , 9.238904384935983e-47 ], [ 405.0625 , 9.240313530069062e-47 ], [ 405.125 , 9.241722890129276e-47 ], [ 405.1875 , 9.243132465149071e-47 ], 
                    [ 405.25 , 9.244542255161502e-47 ], [ 405.3125 , 9.245952260199356e-47 ], [ 405.375 , 9.247362480295435e-47 ], [ 405.4375 , 9.248772915482337e-47 ], 
                    [ 405.5 , 9.250183565793068e-47 ], [ 405.5625 , 9.25159443126024e-47 ], [ 405.625 , 9.253005597909032e-47 ], [ 405.6875 , 9.254417445681838e-47 ], 
                    [ 405.75 , 9.255829508878198e-47 ], [ 405.8125 , 9.257241787530717e-47 ], [ 405.875 , 9.258654281672537e-47 ], [ 405.9375 , 9.260066991336274e-47 ], 
                    [ 406.0 , 9.261479916555007e-47 ], [ 406.0625 , 9.26289305736156e-47 ], [ 406.125 , 9.264306413788833e-47 ], [ 406.1875 , 9.265719985869787e-47 ], 
                    [ 406.25 , 9.26713377363733e-47 ], [ 406.3125 , 9.268547777124106e-47 ], [ 406.375 , 9.269961996363362e-47 ], [ 406.4375 , 9.27137643138775e-47 ], 
                    [ 406.5 , 9.2727910822304e-47 ], [ 406.5625 , 9.27420620113871e-47 ], [ 406.625 , 9.275621849852405e-47 ], [ 406.6875 , 9.277037714655894e-47 ], 
                    [ 406.75 , 9.27845379558229e-47 ], [ 406.8125 , 9.279870092664252e-47 ], [ 406.875 , 9.281286605935107e-47 ], [ 406.9375 , 9.282703335427657e-47 ], 
                    [ 407.0 , 9.284120281175037e-47 ], [ 407.0625 , 9.28553744321006e-47 ], [ 407.125 , 9.286954821566008e-47 ], [ 407.1875 , 9.288372416275762e-47 ], 
                    [ 407.25 , 9.289790227372355e-47 ], [ 407.3125 , 9.291208254888743e-47 ], [ 407.375 , 9.292626498858036e-47 ], [ 407.4375 , 9.29404495931333e-47 ], 
                    [ 407.5 , 9.295463943964302e-47 ], [ 407.5625 , 9.296883275740918e-47 ], [ 407.625 , 9.29830282423623e-47 ], [ 407.6875 , 9.299722589483596e-47 ], 
                    [ 407.75 , 9.301142571516046e-47 ], [ 407.8125 , 9.302562770366614e-47 ], [ 407.875 , 9.303983186068538e-47 ], [ 407.9375 , 9.305403818655e-47 ], 
                    [ 408.0 , 9.306824668158648e-47 ], [ 408.0625 , 9.308245734613135e-47 ], [ 408.125 , 9.309667018051387e-47 ], [ 408.1875 , 9.311088518506537e-47 ], 
                    [ 408.25 , 9.31251023601159e-47 ], [ 408.3125 , 9.31393217060015e-47 ], [ 408.375 , 9.315354322304698e-47 ], [ 408.4375 , 9.316777164529675e-47 ], 
                    [ 408.5 , 9.31820026231438e-47 ], [ 408.5625 , 9.319623577471108e-47 ], [ 408.625 , 9.321047110032993e-47 ], [ 408.6875 , 9.32247086003364e-47 ], 
                    [ 408.75 , 9.323894827505733e-47 ], [ 408.8125 , 9.325319012482888e-47 ], [ 408.875 , 9.326743414998263e-47 ], [ 408.9375 , 9.32816803508495e-47 ], 
                    [ 409.0 , 9.329592872776252e-47 ], [ 409.0625 , 9.331017928105269e-47 ], [ 409.125 , 9.33244320110558e-47 ], [ 409.1875 , 9.333868691810037e-47 ], 
                    [ 409.25 , 9.335294400252288e-47 ], [ 409.3125 , 9.33672038167503e-47 ], [ 409.375 , 9.338147006359548e-47 ], [ 409.4375 , 9.339573849028386e-47 ], 
                    [ 409.5 , 9.341000909714911e-47 ], [ 409.5625 , 9.342428188452175e-47 ], [ 409.625 , 9.343855685273625e-47 ], [ 409.6875 , 9.345283400212587e-47 ], 
                    [ 409.75 , 9.346711333302386e-47 ], [ 409.8125 , 9.348139484576356e-47 ], [ 409.875 , 9.3495678540679e-47 ], [ 409.9375 , 9.35099644181023e-47 ], 
                    [ 410.0 , 9.352425247836895e-47 ], [ 410.0625 , 9.35385427218091e-47 ], [ 410.125 , 9.355283514875905e-47 ], [ 410.1875 , 9.35671297595524e-47 ], 
                    [ 410.25 , 9.358142781162633e-47 ], [ 410.3125 , 9.359573143282626e-47 ], [ 410.375 , 9.361003724028882e-47 ], [ 410.4375 , 9.36243452343482e-47 ], 
                    [ 410.5 , 9.36386554153386e-47 ], [ 410.5625 , 9.365296778359494e-47 ], [ 410.625 , 9.366728233944955e-47 ], [ 410.6875 , 9.368159908323945e-47 ], 
                    [ 410.75 , 9.36959180152964e-47 ], [ 410.8125 , 9.371023913595556e-47 ], [ 410.875 , 9.372456244555544e-47 ], [ 410.9375 , 9.373888794442456e-47 ], 
                    [ 411.0 , 9.375321563290095e-47 ], [ 411.0625 , 9.376754551131859e-47 ], [ 411.125 , 9.378187758001218e-47 ], [ 411.1875 , 9.379621366323888e-47 ], 
                    [ 411.25 , 9.381055475408466e-47 ], [ 411.3125 , 9.382489803762902e-47 ], [ 411.375 , 9.383924351420854e-47 ], [ 411.4375 , 9.38535911841572e-47 ], 
                    [ 411.5 , 9.386794104781166e-47 ], [ 411.5625 , 9.388229310550672e-47 ], [ 411.625 , 9.389664735757714e-47 ], [ 411.6875 , 9.391100380436046e-47 ], 
                    [ 411.75 , 9.392536244618952e-47 ], [ 411.8125 , 9.393972328340132e-47 ], [ 411.875 , 9.395408631633216e-47 ], [ 411.9375 , 9.396845154531644e-47 ], 
                    [ 412.0 , 9.398281897069128e-47 ], [ 412.0625 , 9.399718859279047e-47 ], [ 412.125 , 9.401156253288563e-47 ], [ 412.1875 , 9.402594095290693e-47 ], 
                    [ 412.25 , 9.40403215720084e-47 ], [ 412.3125 , 9.40547043905277e-47 ], [ 412.375 , 9.406908940879923e-47 ], [ 412.4375 , 9.408347662715941e-47 ], 
                    [ 412.5 , 9.40978660459454e-47 ], [ 412.5625 , 9.411225766549507e-47 ], [ 412.625 , 9.412665148614166e-47 ], [ 412.6875 , 9.414104750822452e-47 ], 
                    [ 412.75 , 9.415544573208033e-47 ], [ 412.8125 , 9.416984615804581e-47 ], [ 412.875 , 9.418424878645647e-47 ], [ 412.9375 , 9.419865361764911e-47 ], 
                    [ 413.0 , 9.421306065196131e-47 ], [ 413.0625 , 9.422747216350281e-47 ], [ 413.125 , 9.42418878431163e-47 ], [ 413.1875 , 9.42563057281562e-47 ], 
                    [ 413.25 , 9.427072581896193e-47 ], [ 413.3125 , 9.428514811586826e-47 ], [ 413.375 , 9.429957261921468e-47 ], [ 413.4375 , 9.431399932933677e-47 ], 
                    [ 413.5 , 9.43284282465755e-47 ], [ 413.5625 , 9.434285937126517e-47 ], [ 413.625 , 9.435729270374483e-47 ], [ 413.6875 , 9.437172824435359e-47 ], 
                    [ 413.75 , 9.438616599342591e-47 ], [ 413.8125 , 9.4400605951303e-47 ], [ 413.875 , 9.441504811832147e-47 ], [ 413.9375 , 9.442949249481793e-47 ], 
                    [ 414.0 , 9.444394131751143e-47 ], [ 414.0625 , 9.44583941256599e-47 ], [ 414.125 , 9.44728491455302e-47 ], [ 414.1875 , 9.448730637745878e-47 ], 
                    [ 414.25 , 9.450176582178549e-47 ], [ 414.3125 , 9.451622747885024e-47 ], [ 414.375 , 9.453069134898895e-47 ], [ 414.4375 , 9.454515743254235e-47 ], 
                    [ 414.5 , 9.455962572984908e-47 ], [ 414.5625 , 9.457409624124464e-47 ], [ 414.625 , 9.458856896707247e-47 ], [ 414.6875 , 9.46030439076695e-47 ], 
                    [ 414.75 , 9.461752106337594e-47 ], [ 414.8125 , 9.463200043452681e-47 ], [ 414.875 , 9.464648202146512e-47 ], [ 414.9375 , 9.466096823698229e-47 ], 
                    [ 415.0 , 9.467545868700419e-47 ], [ 415.0625 , 9.468995135518576e-47 ], [ 415.125 , 9.470444624186587e-47 ], [ 415.1875 , 9.471894334738548e-47 ], 
                    [ 415.25 , 9.473344267208356e-47 ], [ 415.3125 , 9.474794421629983e-47 ], [ 415.375 , 9.476244798037402e-47 ], [ 415.4375 , 9.477695396464598e-47 ], 
                    [ 415.5 , 9.479146216945485e-47 ], [ 415.5625 , 9.48059725951426e-47 ], [ 415.625 , 9.482048524204718e-47 ], [ 415.6875 , 9.483500011050858e-47 ], 
                    [ 415.75 , 9.484951720086755e-47 ], [ 415.8125 , 9.486403651346491e-47 ], [ 415.875 , 9.487855991969772e-47 ], [ 415.9375 , 9.489308743881676e-47 ], 
                    [ 416.0 , 9.490761718234505e-47 ], [ 416.0625 , 9.492214915062456e-47 ], [ 416.125 , 9.49366833439946e-47 ], [ 416.1875 , 9.495121976279717e-47 ], 
                    [ 416.25 , 9.496575840737239e-47 ], [ 416.3125 , 9.498029927805966e-47 ], [ 416.375 , 9.49948423752026e-47 ], [ 416.4375 , 9.500938769914075e-47 ], 
                    [ 416.5 , 9.502393525021503e-47 ], [ 416.5625 , 9.503848502876448e-47 ], [ 416.625 , 9.505303703513553e-47 ], [ 416.6875 , 9.506759126966323e-47 ], 
                    [ 416.75 , 9.508214773269214e-47 ], [ 416.8125 , 9.50967079561032e-47 ], [ 416.875 , 9.511127255529915e-47 ], [ 416.9375 , 9.512583938514524e-47 ], 
                    [ 417.0 , 9.51404084459845e-47 ], [ 417.0625 , 9.515497973815719e-47 ], [ 417.125 , 9.516955326200576e-47 ], [ 417.1875 , 9.518412901787205e-47 ], 
                    [ 417.25 , 9.519870700609784e-47 ], [ 417.3125 , 9.521328722702508e-47 ], [ 417.375 , 9.522786968099568e-47 ], [ 417.4375 , 9.524245436835303e-47 ], 
                    [ 417.5 , 9.525704128943578e-47 ], [ 417.5625 , 9.52716304445901e-47 ], [ 417.625 , 9.528622183415477e-47 ], [ 417.6875 , 9.530081545847338e-47 ], 
                    [ 417.75 , 9.531541246177818e-47 ], [ 417.8125 , 9.533001435415586e-47 ], [ 417.875 , 9.534461848347845e-47 ], [ 417.9375 , 9.535922485008728e-47 ], 
                    [ 418.0 , 9.537383345432577e-47 ], [ 418.0625 , 9.538844429653668e-47 ], [ 418.125 , 9.540305737706358e-47 ], [ 418.1875 , 9.541767269624732e-47 ], 
                    [ 418.25 , 9.543229025443355e-47 ], [ 418.3125 , 9.54469100519626e-47 ], [ 418.375 , 9.546153208917951e-47 ], [ 418.4375 , 9.547615636642677e-47 ], 
                    [ 418.5 , 9.54907828840468e-47 ], [ 418.5625 , 9.55054116423849e-47 ], [ 418.625 , 9.552004264178228e-47 ], [ 418.6875 , 9.553467646893087e-47 ], 
                    [ 418.75 , 9.554931581784986e-47 ], [ 418.8125 , 9.556395741004368e-47 ], [ 418.875 , 9.557860124585609e-47 ], [ 418.9375 , 9.55932473256309e-47 ], 
                    [ 419.0 , 9.5607895649712e-47 ], [ 419.0625 , 9.562254621844256e-47 ], [ 419.125 , 9.56371990321686e-47 ], [ 419.1875 , 9.56518540912321e-47 ], 
                    [ 419.25 , 9.566651139597714e-47 ], [ 419.3125 , 9.568117094674986e-47 ], [ 419.375 , 9.56958327438924e-47 ], [ 419.4375 , 9.57104967877483e-47 ], 
                    [ 419.5 , 9.572516307866524e-47 ], [ 419.5625 , 9.573983161698484e-47 ], [ 419.625 , 9.575450240305145e-47 ], [ 419.6875 , 9.576917830427031e-47 ], 
                    [ 419.75 , 9.57838565500865e-47 ], [ 419.8125 , 9.579853704559342e-47 ], [ 419.875 , 9.58132197911338e-47 ], [ 419.9375 , 9.582790478705318e-47 ], 
                    [ 420.0 , 9.584259203369716e-47 ], [ 420.0625 , 9.585728153140865e-47 ], [ 420.125 , 9.587197328053471e-47 ], [ 420.1875 , 9.588666728141973e-47 ], 
                    [ 420.25 , 9.590136353440747e-47 ], [ 420.3125 , 9.591606203984652e-47 ], [ 420.375 , 9.593076279808002e-47 ], [ 420.4375 , 9.594546580945259e-47 ], 
                    [ 420.5 , 9.596017107431025e-47 ], [ 420.5625 , 9.597487859299842e-47 ], [ 420.625 , 9.598959102148294e-47 ], [ 420.6875 , 9.600430659329112e-47 ], 
                    [ 420.75 , 9.601902442105414e-47 ], [ 420.8125 , 9.603374450511518e-47 ], [ 420.875 , 9.604846684582213e-47 ], [ 420.9375 , 9.606319144352027e-47 ], 
                    [ 421.0 , 9.607791829855493e-47 ], [ 421.0625 , 9.60926474112749e-47 ], [ 421.125 , 9.61073787820222e-47 ], [ 421.1875 , 9.612211241114435e-47 ], 
                    [ 421.25 , 9.61368482989903e-47 ], [ 421.3125 , 9.61515864459029e-47 ], [ 421.375 , 9.616632685223053e-47 ], [ 421.4375 , 9.61810695183196e-47 ], 
                    [ 421.5 , 9.619581444451375e-47 ], [ 421.5625 , 9.621056330704424e-47 ], [ 421.625 , 9.622531612970815e-47 ], [ 421.6875 , 9.624007121455312e-47 ], 
                    [ 421.75 , 9.625482856192595e-47 ], [ 421.8125 , 9.626958817217504e-47 ], [ 421.875 , 9.62843500456466e-47 ], [ 421.9375 , 9.629911418268774e-47 ], 
                    [ 422.0 , 9.631388058364483e-47 ], [ 422.0625 , 9.632864924886572e-47 ], [ 422.125 , 9.634342017869825e-47 ], [ 422.1875 , 9.635819337348905e-47 ], 
                    [ 422.25 , 9.637296883358472e-47 ], [ 422.3125 , 9.638774655933397e-47 ], [ 422.375 , 9.64025265510849e-47 ], [ 422.4375 , 9.64173088091802e-47 ], 
                    [ 422.5 , 9.64320939602781e-47 ], [ 422.5625 , 9.64468837453263e-47 ], [ 422.625 , 9.646167579868304e-47 ], [ 422.6875 , 9.647647012069624e-47 ], 
                    [ 422.75 , 9.649126671171309e-47 ], [ 422.8125 , 9.6506065572083e-47 ], [ 422.875 , 9.652086670215333e-47 ], [ 422.9375 , 9.653567010227216e-47 ], 
                    [ 423.0 , 9.65504757727877e-47 ], [ 423.0625 , 9.65652837140488e-47 ], [ 423.125 , 9.658009392640168e-47 ], [ 423.1875 , 9.65949064101981e-47 ], 
                    [ 423.25 , 9.660972116578158e-47 ], [ 423.3125 , 9.662453819350543e-47 ], [ 423.375 , 9.663935749371531e-47 ], [ 423.4375 , 9.665417906676115e-47 ], 
                    [ 423.5 , 9.666900561885223e-47 ], [ 423.5625 , 9.668383478938737e-47 ], [ 423.625 , 9.669866623473884e-47 ], [ 423.6875 , 9.671349995525626e-47 ], 
                    [ 423.75 , 9.672833595128868e-47 ], [ 423.8125 , 9.674317422318517e-47 ], [ 423.875 , 9.67580147712948e-47 ], [ 423.9375 , 9.67728575959675e-47 ], 
                    [ 424.0 , 9.67877026975504e-47 ], [ 424.0625 , 9.68025500763962e-47 ], [ 424.125 , 9.681739973285154e-47 ], [ 424.1875 , 9.683225166726509e-47 ], 
                    [ 424.25 , 9.684710587998835e-47 ], [ 424.3125 , 9.68619623713708e-47 ], [ 424.375 , 9.687682114176134e-47 ], [ 424.4375 , 9.68916836888212e-47 ], 
                    [ 424.5 , 9.690654984192713e-47 ], [ 424.5625 , 9.692141827595782e-47 ], [ 424.625 , 9.69362889912598e-47 ], [ 424.6875 , 9.695116198818654e-47 ], 
                    [ 424.75 , 9.696603726708605e-47 ], [ 424.8125 , 9.69809148283098e-47 ], [ 424.875 , 9.699579467220661e-47 ], [ 424.9375 , 9.701067679912744e-47 ], 
                    [ 425.0 , 9.70255612094225e-47 ], [ 425.0625 , 9.704044790344286e-47 ], [ 425.125 , 9.705533688153824e-47 ], [ 425.1875 , 9.707022814405767e-47 ], 
                    [ 425.25 , 9.708512169135513e-47 ], [ 425.3125 , 9.71000175237777e-47 ], [ 425.375 , 9.711491602009145e-47 ], [ 425.4375 , 9.712981914904917e-47 ], 
                    [ 425.5 , 9.714472456502171e-47 ], [ 425.5625 , 9.715963226836003e-47 ], [ 425.625 , 9.717454225941515e-47 ], [ 425.6875 , 9.718945453853886e-47 ], 
                    [ 425.75 , 9.720436910608088e-47 ], [ 425.8125 , 9.721928596239236e-47 ], [ 425.875 , 9.723420510782526e-47 ], [ 425.9375 , 9.724912654273153e-47 ], 
                    [ 426.0 , 9.726405026746254e-47 ], [ 426.0625 , 9.727897628236967e-47 ], [ 426.125 , 9.729390458780229e-47 ], [ 426.1875 , 9.73088351841119e-47 ], 
                    [ 426.25 , 9.73237680716535e-47 ], [ 426.3125 , 9.733870325077667e-47 ], [ 426.375 , 9.73536428025618e-47 ], [ 426.4375 , 9.736858548707534e-47 ], 
                    [ 426.5 , 9.738353046512287e-47 ], [ 426.5625 , 9.739847773705574e-47 ], [ 426.625 , 9.74134273032253e-47 ], [ 426.6875 , 9.742837916398511e-47 ], 
                    [ 426.75 , 9.744333331968594e-47 ], [ 426.8125 , 9.745828977068147e-47 ], [ 426.875 , 9.74732485173219e-47 ], [ 426.9375 , 9.748820955996237e-47 ], 
                    [ 427.0 , 9.75031728989525e-47 ], [ 427.0625 , 9.751813853464683e-47 ], [ 427.125 , 9.753310646739928e-47 ], [ 427.1875 , 9.754807669755826e-47 ], 
                    [ 427.25 , 9.756304922547847e-47 ], [ 427.3125 , 9.757802449129823e-47 ], [ 427.375 , 9.759300336799035e-47 ], [ 427.4375 , 9.76079845440398e-47 ], 
                    [ 427.5 , 9.762296801980026e-47 ], [ 427.5625 , 9.763795379562336e-47 ], [ 427.625 , 9.765294187186215e-47 ], [ 427.6875 , 9.766793224887256e-47 ], 
                    [ 427.75 , 9.768292492700362e-47 ], [ 427.8125 , 9.769791990661202e-47 ], [ 427.875 , 9.771291718804965e-47 ], [ 427.9375 , 9.77279167716692e-47 ], 
                    [ 428.0 , 9.77429186578254e-47 ], [ 428.0625 , 9.775792284686968e-47 ], [ 428.125 , 9.777292933915833e-47 ], [ 428.1875 , 9.778793813504278e-47 ], 
                    [ 428.25 , 9.780294923487665e-47 ], [ 428.3125 , 9.781796470377548e-47 ], [ 428.375 , 9.783298314968313e-47 ], [ 428.4375 , 9.784800390144386e-47 ], 
                    [ 428.5 , 9.78630269594089e-47 ], [ 428.5625 , 9.787805232393441e-47 ], [ 428.625 , 9.789307999537385e-47 ], [ 428.6875 , 9.79081099740814e-47 ], 
                    [ 428.75 , 9.79231422604113e-47 ], [ 428.8125 , 9.793817685471787e-47 ], [ 428.875 , 9.795321375735403e-47 ], [ 428.9375 , 9.796825296867772e-47 ], 
                    [ 429.0 , 9.79832944890406e-47 ], [ 429.0625 , 9.799833831879857e-47 ], [ 429.125 , 9.80133844583041e-47 ], [ 429.1875 , 9.802843290791256e-47 ], 
                    [ 429.25 , 9.804348426133478e-47 ], [ 429.3125 , 9.805853999492673e-47 ], [ 429.375 , 9.807359804050348e-47 ], [ 429.4375 , 9.808865839842078e-47 ], 
                    [ 429.5 , 9.81037210690337e-47 ], [ 429.5625 , 9.811878605269736e-47 ], [ 429.625 , 9.813385334976629e-47 ], [ 429.6875 , 9.814892296059713e-47 ], 
                    [ 429.75 , 9.816399488554518e-47 ], [ 429.8125 , 9.81790691249644e-47 ], [ 429.875 , 9.819414567921091e-47 ], [ 429.9375 , 9.82092245486395e-47 ], 
                    [ 430.0 , 9.822430573360704e-47 ], [ 430.0625 , 9.823938923446915e-47 ], [ 430.125 , 9.825447505158006e-47 ], [ 430.1875 , 9.826956318529615e-47 ], 
                    [ 430.25 , 9.828465479180914e-47 ], [ 430.3125 , 9.829974932176833e-47 ], [ 430.375 , 9.831484616994053e-47 ], [ 430.4375 , 9.832994533668318e-47 ], 
                    [ 430.5 , 9.834504682235025e-47 ], [ 430.5625 , 9.836015062730072e-47 ], [ 430.625 , 9.837525675188863e-47 ], [ 430.6875 , 9.839036519647235e-47 ], 
                    [ 430.75 , 9.84054759614047e-47 ], [ 430.8125 , 9.842058904704482e-47 ], [ 430.875 , 9.843570445374914e-47 ], [ 430.9375 , 9.845082218187271e-47 ], 
                    [ 431.0 , 9.846594223177347e-47 ], [ 431.0625 , 9.848106460380658e-47 ], [ 431.125 , 9.84961892983287e-47 ], [ 431.1875 , 9.851131642522841e-47 ], 
                    [ 431.25 , 9.852644779145565e-47 ], [ 431.3125 , 9.854158148186575e-47 ], [ 431.375 , 9.855671749681499e-47 ], [ 431.4375 , 9.857185583665905e-47 ], 
                    [ 431.5 , 9.85869965017564e-47 ], [ 431.5625 , 9.860213949246493e-47 ], [ 431.625 , 9.861728480914182e-47 ], [ 431.6875 , 9.8632432452143e-47 ], 
                    [ 431.75 , 9.864758242182571e-47 ], [ 431.8125 , 9.86627347185488e-47 ], [ 431.875 , 9.867788934266897e-47 ], [ 431.9375 , 9.869304629454371e-47 ], 
                    [ 432.0 , 9.870820557453128e-47 ], [ 432.0625 , 9.872336718298716e-47 ], [ 432.125 , 9.873853112027041e-47 ], [ 432.1875 , 9.875369838143652e-47 ], 
                    [ 432.25 , 9.876886936630791e-47 ], [ 432.3125 , 9.878404268181597e-47 ], [ 432.375 , 9.87992183283159e-47 ], [ 432.4375 , 9.881439630616655e-47 ], 
                    [ 432.5 , 9.882957661572532e-47 ], [ 432.5625 , 9.884475925735118e-47 ], [ 432.625 , 9.885994423140304e-47 ], [ 432.6875 , 9.887513153823713e-47 ], 
                    [ 432.75 , 9.889032117821465e-47 ], [ 432.8125 , 9.89055131516919e-47 ], [ 432.875 , 9.892070745902875e-47 ], [ 432.9375 , 9.893590410058379e-47 ], 
                    [ 433.0 , 9.895110307671346e-47 ], [ 433.0625 , 9.896630438777852e-47 ], [ 433.125 , 9.8981508034137e-47 ], [ 433.1875 , 9.899671532903043e-47 ], 
                    [ 433.25 , 9.901192541291394e-47 ], [ 433.3125 , 9.902713783370844e-47 ], [ 433.375 , 9.90423525917751e-47 ], [ 433.4375 , 9.90575696874716e-47 ], 
                    [ 433.5 , 9.907278912115924e-47 ], [ 433.5625 , 9.908801089319439e-47 ], [ 433.625 , 9.910323500393844e-47 ], [ 433.6875 , 9.91184614537493e-47 ], 
                    [ 433.75 , 9.913369024298847e-47 ], [ 433.8125 , 9.914892137201327e-47 ], [ 433.875 , 9.91641548411839e-47 ], [ 433.9375 , 9.91793906508592e-47 ], 
                    [ 434.0 , 9.919462880139945e-47 ], [ 434.0625 , 9.920986929316576e-47 ], [ 434.125 , 9.922511219103785e-47 ], [ 434.1875 , 9.924035919962471e-47 ], 
                    [ 434.25 , 9.92556085510789e-47 ], [ 434.3125 , 9.927086024575967e-47 ], [ 434.375 , 9.928611428402781e-47 ], [ 434.4375 , 9.930137066624415e-47 ], 
                    [ 434.5 , 9.931662939276816e-47 ], [ 434.5625 , 9.933189046396006e-47 ], [ 434.625 , 9.93471538801794e-47 ], [ 434.6875 , 9.936241964178727e-47 ], 
                    [ 434.75 , 9.937768774914547e-47 ], [ 434.8125 , 9.939295820261158e-47 ], [ 434.875 , 9.940823100254757e-47 ], [ 434.9375 , 9.94235061493154e-47 ], 
                    [ 435.0 , 9.943878364327355e-47 ], [ 435.0625 , 9.945406348478411e-47 ], [ 435.125 , 9.946934595084835e-47 ], [ 435.1875 , 9.948463143873881e-47 ], 
                    [ 435.25 , 9.949991927555676e-47 ], [ 435.3125 , 9.951520946166173e-47 ], [ 435.375 , 9.953050199741477e-47 ], [ 435.4375 , 9.954579688317691e-47 ], 
                    [ 435.5 , 9.956109411930859e-47 ], [ 435.5625 , 9.95763937061717e-47 ], [ 435.625 , 9.959169564412818e-47 ], [ 435.6875 , 9.960699993353793e-47 ], 
                    [ 435.75 , 9.962230657476367e-47 ], [ 435.8125 , 9.963761556816685e-47 ], [ 435.875 , 9.965292691410748e-47 ], [ 435.9375 , 9.966824061294567e-47 ], 
                    [ 436.0 , 9.968355666504796e-47 ], [ 436.0625 , 9.969887507077386e-47 ], [ 436.125 , 9.97141968847643e-47 ], [ 436.1875 , 9.972952206037958e-47 ], 
                    [ 436.25 , 9.974484959133517e-47 ], [ 436.3125 , 9.976017947799448e-47 ], [ 436.375 , 9.977551172072097e-47 ], [ 436.4375 , 9.979084631987388e-47 ], 
                    [ 436.5 , 9.980618327581685e-47 ], [ 436.5625 , 9.982152258891346e-47 ], [ 436.625 , 9.98368642595232e-47 ], [ 436.6875 , 9.98522082880119e-47 ], 
                    [ 436.75 , 9.986755467473772e-47 ], [ 436.8125 , 9.988290342006378e-47 ], [ 436.875 , 9.989825452435756e-47 ], [ 436.9375 , 9.991360798797521e-47 ], 
                    [ 437.0 , 9.992896381128505e-47 ], [ 437.0625 , 9.994432199464475e-47 ], [ 437.125 , 9.995968336763055e-47 ], [ 437.1875 , 9.997504746388893e-47 ], 
                    [ 437.25 , 9.999041392165534e-47 ], [ 437.3125 , 1.0000578274128992e-46 ], [ 437.375 , 1.0002115392315856e-46 ], [ 437.4375 , 1.0003652746762005e-46 ], 
                    [ 437.5 , 1.000519033750432e-46 ], [ 437.5625 , 1.0006728164578767e-46 ], [ 437.625 , 1.0008266228021742e-46 ], [ 437.6875 , 1.0009804527869434e-46 ], 
                    [ 437.75 , 1.001134306415839e-46 ], [ 437.8125 , 1.0012881836924883e-46 ], [ 437.875 , 1.0014420846205402e-46 ], [ 437.9375 , 1.0015960092036017e-46 ], 
                    [ 438.0 , 1.0017499574453224e-46 ], [ 438.0625 , 1.0019039293493604e-46 ], [ 438.125 , 1.0020579408205601e-46 ], [ 438.1875 , 1.0022119788914918e-46 ], 
                    [ 438.25 , 1.0023660406414211e-46 ], [ 438.3125 , 1.0025201260739876e-46 ], [ 438.375 , 1.0026742351928319e-46 ], [ 438.4375 , 1.0028283680016094e-46 ], 
                    [ 438.5 , 1.0029825245039335e-46 ], [ 438.5625 , 1.00313670470346e-46 ], [ 438.625 , 1.0032909086038324e-46 ], [ 438.6875 , 1.0034451362086935e-46 ], 
                    [ 438.75 , 1.0035993875216945e-46 ], [ 438.8125 , 1.0037536625464728e-46 ], [ 438.875 , 1.003907961286659e-46 ], [ 438.9375 , 1.0040622837459276e-46 ], 
                    [ 439.0 , 1.00421662992791e-46 ], [ 439.0625 , 1.0043709998362462e-46 ], [ 439.125 , 1.0045254008812805e-46 ], [ 439.1875 , 1.0046798259915409e-46 ], 
                    [ 439.25 , 1.0048342748414916e-46 ], [ 439.3125 , 1.0049887474347604e-46 ], [ 439.375 , 1.0051432437750262e-46 ], [ 439.4375 , 1.0052977638659111e-46 ], 
                    [ 439.5 , 1.0054523077110875e-46 ], [ 439.5625 , 1.005606875314193e-46 ], [ 439.625 , 1.0057614666789009e-46 ], [ 439.6875 , 1.0059160818088503e-46 ], 
                    [ 439.75 , 1.0060707207076728e-46 ], [ 439.8125 , 1.0062253833790658e-46 ], [ 439.875 , 1.0063800698266475e-46 ], [ 439.9375 , 1.006534780054109e-46 ], 
                    [ 440.0 , 1.006689514065063e-46 ], [ 440.0625 , 1.0068442723673481e-46 ], [ 440.125 , 1.0069990687517997e-46 ], [ 440.1875 , 1.007153888935292e-46 ], 
                    [ 440.25 , 1.007308732921477e-46 ], [ 440.3125 , 1.0074636007140065e-46 ], [ 440.375 , 1.0076184923165485e-46 ], [ 440.4375 , 1.0077734077327847e-46 ], 
                    [ 440.5 , 1.0079283469663263e-46 ], [ 440.5625 , 1.0080833100208994e-46 ], [ 440.625 , 1.0082382969000876e-46 ], [ 440.6875 , 1.0083933076076184e-46 ], 
                    [ 440.75 , 1.0085483421471338e-46 ], [ 440.8125 , 1.0087034005223047e-46 ], [ 440.875 , 1.0088584827367745e-46 ], [ 440.9375 , 1.0090135887942224e-46 ], 
                    [ 441.0 , 1.0091687186983357e-46 ], [ 441.0625 , 1.0093238730024327e-46 ], [ 441.125 , 1.0094790579688707e-46 ], [ 441.1875 , 1.0096342667952226e-46 ], 
                    [ 441.25 , 1.0097894994851434e-46 ], [ 441.3125 , 1.0099447560423157e-46 ], [ 441.375 , 1.010100036470395e-46 ], [ 441.4375 , 1.010255340773059e-46 ], 
                    [ 441.5 , 1.0104106689539708e-46 ], [ 441.5625 , 1.0105660210168163e-46 ], [ 441.625 , 1.0107213969652673e-46 ], [ 441.6875 , 1.010876796802982e-46 ], 
                    [ 441.75 , 1.0110322205336406e-46 ], [ 441.8125 , 1.0111876681609164e-46 ], [ 441.875 , 1.011343139688491e-46 ], [ 441.9375 , 1.0114986351200248e-46 ], 
                    [ 442.0 , 1.0116541544591998e-46 ], [ 442.0625 , 1.0118096990656811e-46 ], [ 442.125 , 1.0119652833551492e-46 ], [ 442.1875 , 1.0121208915685469e-46 ], 
                    [ 442.25 , 1.0122765237095526e-46 ], [ 442.3125 , 1.0124321797818673e-46 ], [ 442.375 , 1.0125878597891567e-46 ], [ 442.4375 , 1.0127435637350865e-46 ], 
                    [ 442.5 , 1.0128992916233742e-46 ], [ 442.5625 , 1.0130550434576578e-46 ], [ 442.625 , 1.0132108192416556e-46 ], [ 442.6875 , 1.013366618979014e-46 ], 
                    [ 442.75 , 1.013522442673467e-46 ], [ 442.8125 , 1.013678290328655e-46 ], [ 442.875 , 1.0138341619482773e-46 ], [ 442.9375 , 1.0139900575360183e-46 ], 
                    [ 443.0 , 1.0141459770955637e-46 ], [ 443.0625 , 1.0143019207799227e-46 ], [ 443.125 , 1.0144578914370377e-46 ], [ 443.1875 , 1.0146138860779767e-46 ], 
                    [ 443.25 , 1.0147699047064346e-46 ], [ 443.3125 , 1.0149259473260933e-46 ], [ 443.375 , 1.0150820139406629e-46 ], [ 443.4375 , 1.0152381045538048e-46 ], 
                    [ 443.5 , 1.0153942191692305e-46 ], [ 443.5625 , 1.0155503577906242e-46 ], [ 443.625 , 1.0157065204216696e-46 ], [ 443.6875 , 1.0158627070660807e-46 ], 
                    [ 443.75 , 1.016018917727521e-46 ], [ 443.8125 , 1.0161751524097054e-46 ], [ 443.875 , 1.016331411116313e-46 ], [ 443.9375 , 1.0164876938510673e-46 ], 
                    [ 444.0 , 1.0166440006176195e-46 ], [ 444.0625 , 1.0168003314196866e-46 ], [ 444.125 , 1.0169566995599594e-46 ], [ 444.1875 , 1.0171130920275374e-46 ], 
                    [ 444.25 , 1.0172695085459045e-46 ], [ 444.3125 , 1.0174259491187524e-46 ], [ 444.375 , 1.0175824137497802e-46 ], [ 444.4375 , 1.0177389024426802e-46 ], 
                    [ 444.5 , 1.0178954152011678e-46 ], [ 444.5625 , 1.0180519520289289e-46 ], [ 444.625 , 1.0182085129296797e-46 ], [ 444.6875 , 1.0183650979071223e-46 ], 
                    [ 444.75 , 1.0185217069649159e-46 ], [ 444.8125 , 1.0186783401068431e-46 ], [ 444.875 , 1.0188349973365353e-46 ], [ 444.9375 , 1.0189916786577335e-46 ], 
                    [ 445.0 , 1.019148384074142e-46 ], [ 445.0625 , 1.0193051135894666e-46 ], [ 445.125 , 1.0194618697391414e-46 ], [ 445.1875 , 1.019618650358585e-46 ], 
                    [ 445.25 , 1.0197754550889407e-46 ], [ 445.3125 , 1.019932283933916e-46 ], [ 445.375 , 1.020089136897227e-46 ], [ 445.4375 , 1.0202460139825969e-46 ], 
                    [ 445.5 , 1.0204029151937142e-46 ], [ 445.5625 , 1.0205598405342957e-46 ], [ 445.625 , 1.0207167900080385e-46 ], [ 445.6875 , 1.0208737636186968e-46 ], 
                    [ 445.75 , 1.0210307613699393e-46 ], [ 445.8125 , 1.0211877832655076e-46 ], [ 445.875 , 1.0213448293090787e-46 ], [ 445.9375 , 1.0215018995043803e-46 ], 
                    [ 446.0 , 1.021658993855156e-46 ], [ 446.0625 , 1.0218161123650914e-46 ], [ 446.125 , 1.0219732671656098e-46 ], [ 446.1875 , 1.0221304505375714e-46 ], 
                    [ 446.25 , 1.0222876580849478e-46 ], [ 446.3125 , 1.0224448898114277e-46 ], [ 446.375 , 1.0226021457207518e-46 ], [ 446.4375 , 1.0227594258166394e-46 ], 
                    [ 446.5 , 1.0229167301027964e-46 ], [ 446.5625 , 1.023074058582943e-46 ], [ 446.625 , 1.023231411260837e-46 ], [ 446.6875 , 1.0233887881401343e-46 ], 
                    [ 446.75 , 1.0235461892246153e-46 ], [ 446.8125 , 1.0237036145179523e-46 ], [ 446.875 , 1.0238610640239486e-46 ], [ 446.9375 , 1.0240185377462335e-46 ], 
                    [ 447.0 , 1.024176035688597e-46 ], [ 447.0625 , 1.024333557854728e-46 ], [ 447.125 , 1.0244911033945241e-46 ], [ 447.1875 , 1.024648672483736e-46 ], 
                    [ 447.25 , 1.0248062658074507e-46 ], [ 447.3125 , 1.0249638833693815e-46 ], [ 447.375 , 1.0251215251732484e-46 ], [ 447.4375 , 1.0252791912228022e-46 ], 
                    [ 447.5 , 1.0254368815217427e-46 ], [ 447.5625 , 1.0255945960738214e-46 ], [ 447.625 , 1.0257523348827393e-46 ], [ 447.6875 , 1.0259100979522854e-46 ], 
                    [ 447.75 , 1.02606788528614e-46 ], [ 447.8125 , 1.026225696888035e-46 ], [ 447.875 , 1.0263835327617322e-46 ], [ 447.9375 , 1.0265413929109427e-46 ], 
                    [ 448.0 , 1.0266992773394148e-46 ], [ 448.0625 , 1.0268571860508824e-46 ], [ 448.125 , 1.0270151245284653e-46 ], [ 448.1875 , 1.0271730978533029e-46 ], 
                    [ 448.25 , 1.0273310954772755e-46 ], [ 448.3125 , 1.0274891174041057e-46 ], [ 448.375 , 1.0276471636375539e-46 ], [ 448.4375 , 1.0278052341813147e-46 ], 
                    [ 448.5 , 1.0279633290391863e-46 ], [ 448.5625 , 1.028121448214872e-46 ], [ 448.625 , 1.0282795917121049e-46 ], [ 448.6875 , 1.02843775953467e-46 ], 
                    [ 448.75 , 1.0285959516862576e-46 ], [ 448.8125 , 1.0287541681706397e-46 ], [ 448.875 , 1.0289124089915514e-46 ], [ 448.9375 , 1.029070674152729e-46 ], 
                    [ 449.0 , 1.0292289636579383e-46 ], [ 449.0625 , 1.029387277510887e-46 ], [ 449.125 , 1.0295456159811682e-46 ], [ 449.1875 , 1.029703981471353e-46 ], 
                    [ 449.25 , 1.0298623713214527e-46 ], [ 449.3125 , 1.030020785535185e-46 ], [ 449.375 , 1.0301792241163269e-46 ], [ 449.4375 , 1.0303376870685974e-46 ], 
                    [ 449.5 , 1.0304961743957669e-46 ], [ 449.5625 , 1.0306546861015777e-46 ], [ 449.625 , 1.0308132221897798e-46 ], [ 449.6875 , 1.0309717826641238e-46 ], 
                    [ 449.75 , 1.0311303675283677e-46 ], [ 449.8125 , 1.0312889767862489e-46 ], [ 449.875 , 1.031447610441512e-46 ], [ 449.9375 , 1.031606268497947e-46 ], 
                    [ 450.0 , 1.0317649509592629e-46 ], [ 450.0625 , 1.0319236578292578e-46 ], [ 450.125 , 1.032082389111664e-46 ], [ 450.1875 , 1.0322411553045338e-46 ], 
                    [ 450.25 , 1.032399948497485e-46 ], [ 450.3125 , 1.0325587661181299e-46 ], [ 450.375 , 1.032717608170248e-46 ], [ 450.4375 , 1.0328764746575684e-46 ], 
                    [ 450.5 , 1.0330353655838795e-46 ], [ 450.5625 , 1.0331942809529039e-46 ], [ 450.625 , 1.033353220768446e-46 ], [ 450.6875 , 1.0335121850342298e-46 ], 
                    [ 450.75 , 1.0336711737540386e-46 ], [ 450.8125 , 1.0338301869316266e-46 ], [ 450.875 , 1.0339892245707564e-46 ], [ 450.9375 , 1.0341482866751837e-46 ], 
                    [ 451.0 , 1.0343073732486943e-46 ], [ 451.0625 , 1.0344664842950227e-46 ], [ 451.125 , 1.0346256198179409e-46 ], [ 451.1875 , 1.034784782824607e-46 ], 
                    [ 451.25 , 1.034943973595599e-46 ], [ 451.3125 , 1.035103188856412e-46 ], [ 451.375 , 1.0352624286108279e-46 ], [ 451.4375 , 1.0354216928626076e-46 ], 
                    [ 451.5 , 1.03558098161552e-46 ], [ 451.5625 , 1.0357402948733193e-46 ], [ 451.625 , 1.0358996326398046e-46 ], [ 451.6875 , 1.0360589949187173e-46 ], 
                    [ 451.75 , 1.0362183817138576e-46 ], [ 451.8125 , 1.0363777930289752e-46 ], [ 451.875 , 1.0365372288678564e-46 ], [ 451.9375 , 1.0366966892342671e-46 ], 
                    [ 452.0 , 1.0368561741319733e-46 ], [ 452.0625 , 1.037015683564763e-46 ], [ 452.125 , 1.0371752175363966e-46 ], [ 452.1875 , 1.0373347776602564e-46 ], 
                    [ 452.25 , 1.037494374579968e-46 ], [ 452.3125 , 1.0376539960541069e-46 ], [ 452.375 , 1.0378136420864805e-46 ], [ 452.4375 , 1.0379733126808452e-46 ], 
                    [ 452.5 , 1.0381330078409869e-46 ], [ 452.5625 , 1.0382927275706853e-46 ], [ 452.625 , 1.0384524718737208e-46 ], [ 452.6875 , 1.0386122407538808e-46 ], 
                    [ 452.75 , 1.0387720342149323e-46 ], [ 452.8125 , 1.0389318522606643e-46 ], [ 452.875 , 1.039091694894867e-46 ], [ 452.9375 , 1.039251562121308e-46 ], 
                    [ 453.0 , 1.0394114539437788e-46 ], [ 453.0625 , 1.0395713703660634e-46 ], [ 453.125 , 1.0397313113919463e-46 ], [ 453.1875 , 1.0398912770252132e-46 ], 
                    [ 453.25 , 1.0400512768450765e-46 ], [ 453.3125 , 1.040211305057626e-46 ], [ 453.375 , 1.0403713578930428e-46 ], [ 453.4375 , 1.0405314353550928e-46 ], 
                    [ 453.5 , 1.0406915374475506e-46 ], [ 453.5625 , 1.040851664174258e-46 ], [ 453.625 , 1.0410118155389604e-46 ], [ 453.6875 , 1.0411719915454712e-46 ], 
                    [ 453.75 , 1.0413321921975672e-46 ], [ 453.8125 , 1.0414924174990555e-46 ], [ 453.875 , 1.0416526674537208e-46 ], [ 453.9375 , 1.0418129420653423e-46 ], 
                    [ 454.0 , 1.0419732413377503e-46 ], [ 454.0625 , 1.042133565274695e-46 ], [ 454.125 , 1.0422939138800161e-46 ], [ 454.1875 , 1.0424542871574646e-46 ], 
                    [ 454.25 , 1.0426146898555366e-46 ], [ 454.3125 , 1.0427751293601226e-46 ], [ 454.375 , 1.0429355935534404e-46 ], [ 454.4375 , 1.0430960824392968e-46 ], 
                    [ 454.5 , 1.0432565960214768e-46 ], [ 454.5625 , 1.0434171343037805e-46 ], [ 454.625 , 1.0435776972900237e-46 ], [ 454.6875 , 1.0437382849840004e-46 ], 
                    [ 454.75 , 1.0438988973895128e-46 ], [ 454.8125 , 1.0440595345103633e-46 ], [ 454.875 , 1.0442201963503557e-46 ], [ 454.9375 , 1.044380882913293e-46 ], 
                    [ 455.0 , 1.0445415942029879e-46 ], [ 455.0625 , 1.04470233022323e-46 ], [ 455.125 , 1.0448630909778325e-46 ], [ 455.1875 , 1.0450238764706016e-46 ], 
                    [ 455.25 , 1.0451846867053514e-46 ], [ 455.3125 , 1.045345534672292e-46 ], [ 455.375 , 1.045506410482377e-46 ], [ 455.4375 , 1.0456673110508003e-46 ], 
                    [ 455.5 , 1.0458282363813724e-46 ], [ 455.5625 , 1.0459891864779186e-46 ], [ 455.625 , 1.0461501613442506e-46 ], [ 455.6875 , 1.0463111609841805e-46 ], 
                    [ 455.75 , 1.0464721854014982e-46 ], [ 455.8125 , 1.0466332346000397e-46 ], [ 455.875 , 1.046794308583596e-46 ], [ 455.9375 , 1.0469554073560188e-46 ], 
                    [ 456.0 , 1.047116530921086e-46 ], [ 456.0625 , 1.0472776792826276e-46 ], [ 456.125 , 1.0474388524444677e-46 ], [ 456.1875 , 1.0476000504103928e-46 ], 
                    [ 456.25 , 1.0477612731842648e-46 ], [ 456.3125 , 1.047922527981404e-46 ], [ 456.375 , 1.0480838246457885e-46 ], [ 456.4375 , 1.048245146137029e-46 ], 
                    [ 456.5 , 1.0484064924589318e-46 ], [ 456.5625 , 1.0485678636153268e-46 ], [ 456.625 , 1.0487292596100435e-46 ], [ 456.6875 , 1.0488906804469053e-46 ], 
                    [ 456.75 , 1.0490521261296986e-46 ], [ 456.8125 , 1.0492135966623073e-46 ], [ 456.875 , 1.0493750920485114e-46 ], [ 456.9375 , 1.0495366122921516e-46 ], 
                    [ 457.0 , 1.049698157397061e-46 ], [ 457.0625 , 1.049859727367052e-46 ], [ 457.125 , 1.0500213222059512e-46 ], [ 457.1875 , 1.0501829419176165e-46 ], 
                    [ 457.25 , 1.0503445865058315e-46 ], [ 457.3125 , 1.0505062559744704e-46 ], [ 457.375 , 1.0506679687427932e-46 ], [ 457.4375 , 1.0508297125731565e-46 ], 
                    [ 457.5 , 1.050991481302959e-46 ], [ 457.5625 , 1.0511532749360785e-46 ], [ 457.625 , 1.0513150934763115e-46 ], [ 457.6875 , 1.0514769369274995e-46 ], 
                    [ 457.75 , 1.0516388052935149e-46 ], [ 457.8125 , 1.0518006985781335e-46 ], [ 457.875 , 1.0519626167852213e-46 ], [ 457.9375 , 1.0521245599186445e-46 ], 
                    [ 458.0 , 1.0522865279821885e-46 ], [ 458.0625 , 1.0524485209797282e-46 ], [ 458.125 , 1.0526105389150648e-46 ], [ 458.1875 , 1.0527725817920745e-46 ], 
                    [ 458.25 , 1.0529346496145672e-46 ], [ 458.3125 , 1.053096742386413e-46 ], [ 458.375 , 1.0532588661851244e-46 ], [ 458.4375 , 1.0534210460172324e-46 ], 
                    [ 458.5 , 1.0535832508216419e-46 ], [ 458.5625 , 1.0537454806021983e-46 ], [ 458.625 , 1.0539077353627396e-46 ], [ 458.6875 , 1.0540700151071345e-46 ], 
                    [ 458.75 , 1.054232319839208e-46 ], [ 458.8125 , 1.0543946495628075e-46 ], [ 458.875 , 1.0545570042818037e-46 ], [ 458.9375 , 1.0547193840000227e-46 ], 
                    [ 459.0 , 1.0548817887213066e-46 ], [ 459.0625 , 1.0550442184495426e-46 ], [ 459.125 , 1.0552066731885438e-46 ], [ 459.1875 , 1.055369152942184e-46 ], 
                    [ 459.25 , 1.0555316577143072e-46 ], [ 459.3125 , 1.0556941875087659e-46 ], [ 459.375 , 1.0558567423294053e-46 ], [ 459.4375 , 1.0560193486584996e-46 ], 
                    [ 459.5 , 1.0561820025597505e-46 ], [ 459.5625 , 1.0563446815138413e-46 ], [ 459.625 , 1.0565073855246458e-46 ], [ 459.6875 , 1.056670114596016e-46 ], 
                    [ 459.75 , 1.0568328687318189e-46 ], [ 459.8125 , 1.0569956479358856e-46 ], [ 459.875 , 1.0571584522121143e-46 ], [ 459.9375 , 1.057321281564352e-46 ], 
                    [ 460.0 , 1.0574841359964613e-46 ], [ 460.0625 , 1.0576470155123047e-46 ], [ 460.125 , 1.0578099201157308e-46 ], [ 460.1875 , 1.057972849810649e-46 ], 
                    [ 460.25 , 1.0581358046008713e-46 ], [ 460.3125 , 1.0582987844902779e-46 ], [ 460.375 , 1.058461789482765e-46 ], [ 460.4375 , 1.0586248195821766e-46 ], 
                    [ 460.5 , 1.0587879212035211e-46 ], [ 460.5625 , 1.058951054312384e-46 ], [ 460.625 , 1.059114212556036e-46 ], [ 460.6875 , 1.0592773959383497e-46 ], 
                    [ 460.75 , 1.0594406044632063e-46 ], [ 460.8125 , 1.0596038381344792e-46 ], [ 460.875 , 1.059767096956013e-46 ], [ 460.9375 , 1.05993038093172e-46 ], 
                    [ 461.0 , 1.0600936900654612e-46 ], [ 461.0625 , 1.0602570243611054e-46 ], [ 461.125 , 1.0604203838225365e-46 ], [ 461.1875 , 1.0605837684536397e-46 ], 
                    [ 461.25 , 1.060747178258278e-46 ], [ 461.3125 , 1.060910613240345e-46 ], [ 461.375 , 1.061074073403705e-46 ], [ 461.4375 , 1.061237558752245e-46 ], 
                    [ 461.5 , 1.0614010839778806e-46 ], [ 461.5625 , 1.0615647009623221e-46 ], [ 461.625 , 1.061728343168616e-46 ], [ 461.6875 , 1.0618920106006882e-46 ], 
                    [ 461.75 , 1.0620557032623819e-46 ], [ 461.8125 , 1.0622194211576164e-46 ], [ 461.875 , 1.0623831642902667e-46 ], [ 461.9375 , 1.062546932664238e-46 ], 
                    [ 462.0 , 1.062710726283414e-46 ], [ 462.0625 , 1.062874545151671e-46 ], [ 462.125 , 1.0630383892729466e-46 ], [ 462.1875 , 1.0632022586510883e-46 ], 
                    [ 462.25 , 1.0633661532899892e-46 ], [ 462.3125 , 1.0635300731935815e-46 ], [ 462.375 , 1.0636940183657293e-46 ], [ 462.4375 , 1.0638579888103432e-46 ], 
                    [ 462.5 , 1.0640219845312884e-46 ], [ 462.5625 , 1.0641860469127209e-46 ], [ 462.625 , 1.0643501863504005e-46 ], [ 462.6875 , 1.0645143511048515e-46 ], 
                    [ 462.75 , 1.064678541179979e-46 ], [ 462.8125 , 1.0648427565796806e-46 ], [ 462.875 , 1.0650069973078778e-46 ], [ 462.9375 , 1.0651712633684617e-46 ], 
                    [ 463.0 , 1.0653355547653476e-46 ], [ 463.0625 , 1.0654998715024658e-46 ], [ 463.125 , 1.0656642135836717e-46 ], [ 463.1875 , 1.0658285810129047e-46 ], 
                    [ 463.25 , 1.065992973794105e-46 ], [ 463.3125 , 1.066157391931137e-46 ], [ 463.375 , 1.0663218354279043e-46 ], [ 463.4375 , 1.0664863042883639e-46 ], 
                    [ 463.5 , 1.0666507985164046e-46 ], [ 463.5625 , 1.0668153181159474e-46 ], [ 463.625 , 1.0669799514326031e-46 ], [ 463.6875 , 1.0671446536899072e-46 ], 
                    [ 463.75 , 1.0673093813711505e-46 ], [ 463.8125 , 1.0674741344802651e-46 ], [ 463.875 , 1.0676389130211614e-46 ], [ 463.9375 , 1.0678037169977723e-46 ], 
                    [ 464.0 , 1.0679685464140318e-46 ], [ 464.0625 , 1.068133401273844e-46 ], [ 464.125 , 1.0682982815811668e-46 ], [ 464.1875 , 1.068463187339906e-46 ], 
                    [ 464.25 , 1.0686281185539974e-46 ], [ 464.3125 , 1.0687930752273784e-46 ], [ 464.375 , 1.0689580573639558e-46 ], [ 464.4375 , 1.0691230649676832e-46 ], 
                    [ 464.5 , 1.0692880980424766e-46 ], [ 464.5625 , 1.0694531565922906e-46 ], [ 464.625 , 1.0696182406210345e-46 ], [ 464.6875 , 1.0697835003242496e-46 ], 
                    [ 464.75 , 1.069948810535734e-46 ], [ 464.8125 , 1.0701141462920524e-46 ], [ 464.875 , 1.070279507597182e-46 ], [ 464.9375 , 1.0704448944550786e-46 ], 
                    [ 465.0 , 1.0706103068696759e-46 ], [ 465.0625 , 1.0707757448448998e-46 ], [ 465.125 , 1.0709412083847386e-46 ], [ 465.1875 , 1.0711066974931117e-46 ], 
                    [ 465.25 , 1.071272212174001e-46 ], [ 465.3125 , 1.0714377524313277e-46 ], [ 465.375 , 1.0716033182690897e-46 ], [ 465.4375 , 1.071768909691194e-46 ], 
                    [ 465.5 , 1.0719345267016018e-46 ], [ 465.5625 , 1.072100169304267e-46 ], [ 465.625 , 1.072265837503175e-46 ], [ 465.6875 , 1.0724315329396773e-46 ], 
                    [ 465.75 , 1.072597490378515e-46 ], [ 465.8125 , 1.0727634734990592e-46 ], [ 465.875 , 1.0729294823052912e-46 ], [ 465.9375 , 1.0730955168011785e-46 ], 
                    [ 466.0 , 1.0732615769906964e-46 ], [ 466.0625 , 1.0734276628778134e-46 ], [ 466.125 , 1.0735937744665138e-46 ], [ 466.1875 , 1.07375991176079e-46 ], 
                    [ 466.25 , 1.073926074764582e-46 ], [ 466.3125 , 1.0740922634819063e-46 ], [ 466.375 , 1.0742584779167268e-46 ], [ 466.4375 , 1.0744247180730232e-46 ], 
                    [ 466.5 , 1.074590983954761e-46 ], [ 466.5625 , 1.0747572755659358e-46 ], [ 466.625 , 1.0749235929105523e-46 ], [ 466.6875 , 1.0750899359925626e-46 ], 
                    [ 466.75 , 1.0752563436738268e-46 ], [ 466.8125 , 1.0754230669351727e-46 ], [ 466.875 , 1.0755898160476765e-46 ], [ 466.9375 , 1.0757565910153918e-46 ], 
                    [ 467.0 , 1.0759233918422898e-46 ], [ 467.0625 , 1.0760902185323952e-46 ], [ 467.125 , 1.0762570710897183e-46 ], [ 467.1875 , 1.0764239495182774e-46 ], 
                    [ 467.25 , 1.0765908538220687e-46 ], [ 467.3125 , 1.0767577840051274e-46 ], [ 467.375 , 1.076924740071428e-46 ], [ 467.4375 , 1.0770917220250138e-46 ], 
                    [ 467.5 , 1.0772587298699071e-46 ], [ 467.5625 , 1.0774257636100917e-46 ], [ 467.625 , 1.0775928232495978e-46 ], [ 467.6875 , 1.0777599087924336e-46 ], 
                    [ 467.75 , 1.0779270202426387e-46 ], [ 467.8125 , 1.0780942448829905e-46 ], [ 467.875 , 1.0782618647233355e-46 ], [ 467.9375 , 1.0784295106248555e-46 ], 
                    [ 468.0 , 1.0785971825916177e-46 ], [ 468.0625 , 1.0787648806276592e-46 ], [ 468.125 , 1.0789326047370483e-46 ], [ 468.1875 , 1.0791003549238165e-46 ], 
                    [ 468.25 , 1.0792681311920482e-46 ], [ 468.3125 , 1.0794359335457837e-46 ], [ 468.375 , 1.0796037619890633e-46 ], [ 468.4375 , 1.0797716165259737e-46 ], 
                    [ 468.5 , 1.0799394971605646e-46 ], [ 468.5625 , 1.0801074038968624e-46 ], [ 468.625 , 1.0802753367389643e-46 ], [ 468.6875 , 1.0804432956909132e-46 ], 
                    [ 468.75 , 1.0806112807567618e-46 ], [ 468.8125 , 1.0807792919405848e-46 ], [ 468.875 , 1.0809474800558905e-46 ], [ 468.9375 , 1.0811162116232545e-46 ], 
                    [ 469.0 , 1.0812849695289549e-46 ], [ 469.0625 , 1.0814537537770803e-46 ], [ 469.125 , 1.081622564371742e-46 ], [ 469.1875 , 1.0817914013170609e-46 ], 
                    [ 469.25 , 1.081960264617165e-46 ], [ 469.3125 , 1.0821291542761383e-46 ], [ 469.375 , 1.08229807029811e-46 ], [ 469.4375 , 1.0824670126872108e-46 ], 
                    [ 469.5 , 1.0826359814475333e-46 ], [ 469.5625 , 1.0828049765831863e-46 ], [ 469.625 , 1.0829739980983258e-46 ], [ 469.6875 , 1.083143045997038e-46 ], 
                    [ 469.75 , 1.0833121202834574e-46 ], [ 469.8125 , 1.0834812209616794e-46 ], [ 469.875 , 1.0836503480358392e-46 ], [ 469.9375 , 1.0838197238049957e-46 ], 
                    [ 470.0 , 1.0839899055644705e-46 ], [ 470.0625 , 1.0841601140459456e-46 ], [ 470.125 , 1.0843303492536097e-46 ], [ 470.1875 , 1.0845006111916743e-46 ], 
                    [ 470.25 , 1.0846708998643293e-46 ], [ 470.3125 , 1.0848412152757724e-46 ], [ 470.375 , 1.0850115574302019e-46 ], [ 470.4375 , 1.0851819263318174e-46 ], 
                    [ 470.5 , 1.085352321984811e-46 ], [ 470.5625 , 1.0855227443933982e-46 ], [ 470.625 , 1.085693193561773e-46 ], [ 470.6875 , 1.0858636694941214e-46 ], 
                    [ 470.75 , 1.0860341721947003e-46 ], [ 470.8125 , 1.0862047016676274e-46 ], [ 470.875 , 1.0863752579171843e-46 ], [ 470.9375 , 1.086545840947544e-46 ], 
                    [ 471.0 , 1.0867167307495514e-46 ], [ 471.0625 , 1.0868889272847656e-46 ], [ 471.125 , 1.0870611511054986e-46 ], [ 471.1875 , 1.0872334022161206e-46 ], 
                    [ 471.25 , 1.0874056806209016e-46 ], [ 471.3125 , 1.0875779863241818e-46 ], [ 471.375 , 1.0877503193303103e-46 ], [ 471.4375 , 1.08792267964359e-46 ], 
                    [ 471.5 , 1.0880950672683637e-46 ], [ 471.5625 , 1.088267482208943e-46 ], [ 471.625 , 1.0884399244696722e-46 ], [ 471.6875 , 1.0886123940548645e-46 ], 
                    [ 471.75 , 1.08878489096885e-46 ], [ 471.8125 , 1.0889574152159898e-46 ], [ 471.875 , 1.0891299668005764e-46 ], [ 471.9375 , 1.0893025457269648e-46 ], 
                    [ 472.0 , 1.0894751519994719e-46 ], [ 472.0625 , 1.0896480380847464e-46 ], [ 472.125 , 1.0898232610205351e-46 ], [ 472.1875 , 1.089998512133383e-46 ], 
                    [ 472.25 , 1.0901737914278289e-46 ], [ 472.3125 , 1.0903490989083735e-46 ], [ 472.375 , 1.0905244345795804e-46 ], [ 472.4375 , 1.090699798445983e-46 ], 
                    [ 472.5 , 1.0908751905121073e-46 ], [ 472.5625 , 1.0910506107824878e-46 ], [ 472.625 , 1.091226059261645e-46 ], [ 472.6875 , 1.0914015359541612e-46 ], 
                    [ 472.75 , 1.0915770408645193e-46 ], [ 472.8125 , 1.0917525739972643e-46 ], [ 472.875 , 1.0919281353569814e-46 ], [ 472.9375 , 1.0921037249481784e-46 ], 
                    [ 473.0 , 1.0922793427753722e-46 ], [ 473.0625 , 1.0924549888431418e-46 ], [ 473.125 , 1.0926306631560055e-46 ], [ 473.1875 , 1.0928107797247122e-46 ], 
                    [ 473.25 , 1.0929910219733952e-46 ], [ 473.3125 , 1.0931712939502594e-46 ], [ 473.375 , 1.0933515956602e-46 ], [ 473.4375 , 1.0935319271081133e-46 ], 
                    [ 473.5 , 1.0937122882989117e-46 ], [ 473.5625 , 1.0938926792375246e-46 ], [ 473.625 , 1.0940730999288036e-46 ], [ 473.6875 , 1.0942535503777183e-46 ], 
                    [ 473.75 , 1.0944340305891147e-46 ], [ 473.8125 , 1.0946145405679482e-46 ], [ 473.875 , 1.094795080319113e-46 ], [ 473.9375 , 1.0949756498475038e-46 ], 
                    [ 474.0 , 1.095156249158071e-46 ], [ 474.0625 , 1.0953368782556879e-46 ], [ 474.125 , 1.095517537145306e-46 ], [ 474.1875 , 1.0956982258318002e-46 ], 
                    [ 474.25 , 1.0958860839799743e-46 ], [ 474.3125 , 1.0960755401550277e-46 ], [ 474.375 , 1.0962650290831514e-46 ], [ 474.4375 , 1.096454550770023e-46 ], 
                    [ 474.5 , 1.0966441052212981e-46 ], [ 474.5625 , 1.0968336924426334e-46 ], [ 474.625 , 1.0970233124397017e-46 ], [ 474.6875 , 1.0972129652181695e-46 ], 
                    [ 474.75 , 1.0974026507837113e-46 ], [ 474.8125 , 1.0975923691419881e-46 ], [ 474.875 , 1.097782120298669e-46 ], [ 474.9375 , 1.0979719042594238e-46 ], 
                    [ 475.0 , 1.0981617210299163e-46 ], [ 475.0625 , 1.098351570615826e-46 ], [ 475.125 , 1.0985414530228343e-46 ], [ 475.1875 , 1.098731368256607e-46 ], 
                    [ 475.25 , 1.0989213163228273e-46 ], [ 475.3125 , 1.099123246887645e-46 ], [ 475.375 , 1.099332455750911e-46 ], [ 475.4375 , 1.0995417044353242e-46 ], 
                    [ 475.5 , 1.0997509929484637e-46 ], [ 475.5625 , 1.0999603212979114e-46 ], [ 475.625 , 1.1001696894912413e-46 ], [ 475.6875 , 1.1003790975360614e-46 ], 
                    [ 475.75 , 1.1005885454399252e-46 ], [ 475.8125 , 1.1007980332104587e-46 ], [ 475.875 , 1.1010075608552039e-46 ], [ 475.9375 , 1.101217128381789e-46 ], 
                    [ 476.0 , 1.101426735797774e-46 ], [ 476.0625 , 1.101636383110775e-46 ], [ 476.125 , 1.1018460703283863e-46 ], [ 476.1875 , 1.1020557974581874e-46 ], 
                    [ 476.25 , 1.1022655645077756e-46 ], [ 476.3125 , 1.1024753714847805e-46 ], [ 476.375 , 1.1027052993982259e-46 ], [ 476.4375 , 1.1029671711307206e-46 ], 
                    [ 476.5 , 1.10322910505281e-46 ], [ 476.5625 , 1.1034911011792865e-46 ], [ 476.625 , 1.1037531595248832e-46 ], [ 476.6875 , 1.1040152801044074e-46 ], 
                    [ 476.75 , 1.1042774629326385e-46 ], [ 476.8125 , 1.1045397080243515e-46 ], [ 476.875 , 1.1048020153943326e-46 ], [ 476.9375 , 1.1050643850573643e-46 ], 
                    [ 477.0 , 1.1053268170282555e-46 ], [ 477.0625 , 1.1055893113217873e-46 ], [ 477.125 , 1.1058518679527682e-46 ], [ 477.1875 , 1.10611448693601e-46 ], 
                    [ 477.25 , 1.1063771682863046e-46 ], [ 477.3125 , 1.1066399120184784e-46 ], [ 477.375 , 1.1069027181473303e-46 ], [ 477.4375 , 1.107188067579648e-46 ], 
                    [ 477.5 , 1.1076538224101309e-46 ], [ 477.5625 , 1.1081197731671772e-46 ], [ 477.625 , 1.1085859199332146e-46 ], [ 477.6875 , 1.1090522627906892e-46 ], 
                    [ 477.75 , 1.1095188018221207e-46 ], [ 477.8125 , 1.109985537109986e-46 ], [ 477.875 , 1.110452468736882e-46 ], [ 477.9375 , 1.1109195967853787e-46 ], 
                    [ 478.0 , 1.1113869213381188e-46 ], [ 478.0625 , 1.111854442477741e-46 ], [ 478.125 , 1.1123221602869745e-46 ], [ 478.1875 , 1.1127900748485195e-46 ], 
                    [ 478.25 , 1.113258186245174e-46 ], [ 478.3125 , 1.1137264945597163e-46 ], [ 478.375 , 1.1141949998749983e-46 ], [ 478.4375 , 1.1146637022738835e-46 ], 
                    [ 478.5 , 1.1151326018392619e-46 ], [ 478.5625 , 1.1171492632809723e-46 ], [ 478.625 , 1.119566855316156e-46 ], [ 478.6875 , 1.1219896791958413e-46 ], 
                    [ 478.75 , 1.1244177462421514e-46 ], [ 478.8125 , 1.1268510678016472e-46 ], [ 478.875 , 1.1292896552455e-46 ], [ 478.9375 , 1.1317335199694578e-46 ], 
                    [ 479.0 , 1.1341826733939376e-46 ], [ 479.0625 , 1.1366371269640706e-46 ], [ 479.125 , 1.1390968921497647e-46 ], [ 479.1875 , 1.1415619804457169e-46 ], 
                    [ 479.25 , 1.1440324033715563e-46 ], [ 479.3125 , 1.1465081724717685e-46 ], [ 479.375 , 1.1489892993159193e-46 ], [ 479.4375 , 1.1514757954984901e-46 ], 
                    [ 479.5 , 1.1539676726391842e-46 ], [ 479.5625 , 1.1564649423827197e-46 ], [ 479.625 , 1.1645786961200177e-46 ], [ 479.6875 , 1.1799101725761366e-46 ], 
                    [ 479.75 , 1.195443485259483e-46 ], [ 479.8125 , 1.2111812913089485e-46 ], [ 479.875 , 1.2271262828442232e-46 ], [ 479.9375 , 1.2432811874262761e-46 ], 
                    [ 480.0 , 1.2596487685239627e-46 ], [ 480.0625 , 1.2762318259866528e-46 ], [ 480.125 , 1.293033196523366e-46 ], [ 480.1875 , 1.3100557541878125e-46 ], 
                    [ 480.25 , 1.3273024108701393e-46 ], [ 480.3125 , 1.3447761167950397e-46 ], [ 480.375 , 1.362479861026395e-46 ], [ 480.4375 , 1.3804166719785932e-46 ], 
                    [ 480.5 , 1.3985896179345822e-46 ], [ 480.5625 , 1.417001807570693e-46 ], [ 480.625 , 1.4356563904884602e-46 ], [ 480.6875 , 1.4525945913519344e-46 ], 
                    [ 480.75 , 1.4318274812296975e-46 ], [ 480.8125 , 1.4113572694061103e-46 ], [ 480.875 , 1.3911797112559646e-46 ], [ 480.9375 , 1.3712906228375646e-46 ], 
                    [ 481.0 , 1.3516858800252e-46 ], [ 481.0625 , 1.3323614176540229e-46 ], [ 481.125 , 1.3133132286770035e-46 ], [ 481.1875 , 1.2945373633341789e-46 ], 
                    [ 481.25 , 1.2760299283335228e-46 ], [ 481.3125 , 1.257787086043785e-46 ], [ 481.375 , 1.239805053698536e-46 ], [ 481.4375 , 1.222080102611994e-46 ], 
                    [ 481.5 , 1.2046085574057589e-46 ], [ 481.5625 , 1.1873867952466584e-46 ], [ 481.625 , 1.1704112450956424e-46 ], [ 481.6875 , 1.1536783869671954e-46 ], 
                    [ 481.75 , 1.1371847511995787e-46 ], [ 481.8125 , 1.1306488624658379e-46 ], [ 481.875 , 1.1302061761535267e-46 ], [ 481.9375 , 1.129763663167496e-46 ], 
                    [ 482.0 , 1.1293213234398992e-46 ], [ 482.0625 , 1.1288791569028917e-46 ], [ 482.125 , 1.1284371634886798e-46 ], [ 482.1875 , 1.12799534312944e-46 ], 
                    [ 482.25 , 1.1275536957574558e-46 ], [ 482.3125 , 1.1271122213049645e-46 ], [ 482.375 , 1.126670919704295e-46 ], [ 482.4375 , 1.126229790887738e-46 ], 
                    [ 482.5 , 1.1257888347876747e-46 ], [ 482.5625 , 1.1253480513364488e-46 ], [ 482.625 , 1.1249074404665025e-46 ], [ 482.6875 , 1.1244670021102171e-46 ], 
                    [ 482.75 , 1.1240267362000953e-46 ], [ 482.8125 , 1.123586642668579e-46 ], [ 482.875 , 1.123214192466029e-46 ], [ 482.9375 , 1.1232418551635663e-46 ], 
                    [ 483.0 , 1.1232695185423769e-46 ], [ 483.0625 , 1.1232971826024936e-46 ], [ 483.125 , 1.1233248473439172e-46 ], [ 483.1875 , 1.1233525127666725e-46 ], 
                    [ 483.25 , 1.1233801788707761e-46 ], [ 483.3125 , 1.123407845656245e-46 ], [ 483.375 , 1.1234355131230958e-46 ], [ 483.4375 , 1.1234631812713535e-46 ], 
                    [ 483.5 , 1.1234908501010106e-46 ], [ 483.5625 , 1.123518519612108e-46 ], [ 483.625 , 1.1235461898046627e-46 ], [ 483.6875 , 1.1235738606786751e-46 ], 
                    [ 483.75 , 1.1236015322341703e-46 ], [ 483.8125 , 1.123629204471173e-46 ], [ 483.875 , 1.1236568773896758e-46 ], [ 483.9375 , 1.12368455098972e-46 ], 
                    [ 484.0 , 1.1237685665390222e-46 ], [ 484.0625 , 1.1238853307676712e-46 ], [ 484.125 , 1.1240021071286016e-46 ], [ 484.1875 , 1.1241188956230817e-46 ], 
                    [ 484.25 , 1.124235696252372e-46 ], [ 484.3125 , 1.1243525090177416e-46 ], [ 484.375 , 1.1244693339204355e-46 ], [ 484.4375 , 1.1245861709617227e-46 ], 
                    [ 484.5 , 1.1247030201428647e-46 ], [ 484.5625 , 1.1248198814651384e-46 ], [ 484.625 , 1.1249367549297656e-46 ], [ 484.6875 , 1.1250536405380561e-46 ], 
                    [ 484.75 , 1.1251705382912233e-46 ], [ 484.8125 , 1.1252874481905614e-46 ], [ 484.875 , 1.1254043702373325e-46 ], [ 484.9375 , 1.1255213044327825e-46 ], 
                    [ 485.0 , 1.1256382507781979e-46 ], [ 485.0625 , 1.1257575187419658e-46 ], [ 485.125 , 1.1259029627412682e-46 ], [ 485.1875 , 1.1260484255314431e-46 ], 
                    [ 485.25 , 1.126193907114902e-46 ], [ 485.3125 , 1.1263394074940814e-46 ], [ 485.375 , 1.126484926671425e-46 ], [ 485.4375 , 1.1266304646493302e-46 ], 
                    [ 485.5 , 1.1267760214302413e-46 ], [ 485.5625 , 1.1269215970165959e-46 ], [ 485.625 , 1.1270671914107995e-46 ], [ 485.6875 , 1.127212804615314e-46 ], 
                    [ 485.75 , 1.1273584366325378e-46 ], [ 485.8125 , 1.1275040874649333e-46 ], [ 485.875 , 1.1276497571149072e-46 ], [ 485.9375 , 1.1277954455849072e-46 ], 
                    [ 486.0 , 1.1279411528773402e-46 ], [ 486.0625 , 1.1280868789946701e-46 ], [ 486.125 , 1.1282326239393052e-46 ], [ 486.1875 , 1.1283841719630508e-46 ], 
                    [ 486.25 , 1.128541763452441e-46 ], [ 486.3125 , 1.1286993769512638e-46 ], [ 486.375 , 1.1288570124625612e-46 ], [ 486.4375 , 1.1290146699894314e-46 ], 
                    [ 486.5 , 1.1291723495349415e-46 ], [ 486.5625 , 1.1293300511021583e-46 ], [ 486.625 , 1.1294877746941816e-46 ], [ 486.6875 , 1.129645520314039e-46 ], 
                    [ 486.75 , 1.1298032879648716e-46 ], [ 486.8125 , 1.1299610776496996e-46 ], [ 486.875 , 1.1301188893716487e-46 ], [ 486.9375 , 1.1302767231337564e-46 ], 
                    [ 487.0 , 1.1304345789391171e-46 ], [ 487.0625 , 1.1305924567908093e-46 ], [ 487.125 , 1.130750356691928e-46 ], [ 487.1875 , 1.1309082786455286e-46 ], 
                    [ 487.25 , 1.131066222654683e-46 ], [ 487.3125 , 1.1312291256193111e-46 ], [ 487.375 , 1.1313928644029712e-46 ], [ 487.4375 , 1.131556626886852e-46 ], 
                    [ 487.5 , 1.1317204130744164e-46 ], [ 487.5625 , 1.131884222969047e-46 ], [ 487.625 , 1.1320480565742155e-46 ], [ 487.6875 , 1.1322119138933218e-46 ], 
                    [ 487.75 , 1.1323757949298305e-46 ], [ 487.8125 , 1.1325396996871503e-46 ], [ 487.875 , 1.1327036281687228e-46 ], [ 487.9375 , 1.1328675803779817e-46 ], 
                    [ 488.0 , 1.133031556318354e-46 ], [ 488.0625 , 1.1331955559933143e-46 ], [ 488.125 , 1.133359579406234e-46 ], [ 488.1875 , 1.1335236265605808e-46 ], 
                    [ 488.25 , 1.1336876974598075e-46 ], [ 488.3125 , 1.133851792107327e-46 ], [ 488.375 , 1.1340164932363259e-46 ], [ 488.4375 , 1.1341837408408333e-46 ], 
                    [ 488.5 , 1.134351013111438e-46 ], [ 488.5625 , 1.1345183100517618e-46 ], [ 488.625 , 1.1346856316654833e-46 ], [ 488.6875 , 1.134852977956201e-46 ], 
                    [ 488.75 , 1.1350203489275787e-46 ], [ 488.8125 , 1.1351877445832398e-46 ], [ 488.875 , 1.1353551649268336e-46 ], [ 488.9375 , 1.1355226099620168e-46 ], 
                    [ 489.0 , 1.1356900796923987e-46 ], [ 489.0625 , 1.1358575741216457e-46 ], [ 489.125 , 1.1360250932533842e-46 ], [ 489.1875 , 1.1361926370912738e-46 ], 
                    [ 489.25 , 1.136360205638942e-46 ], [ 489.3125 , 1.1365277989000571e-46 ], [ 489.375 , 1.1366954168782316e-46 ], [ 489.4375 , 1.1368630595771273e-46 ], 
                    [ 489.5 , 1.1370315909549135e-46 ], [ 489.5625 , 1.1372010967052351e-46 ], [ 489.625 , 1.1373706277250387e-46 ], [ 489.6875 , 1.1375401840180916e-46 ], 
                    [ 489.75 , 1.1377097655881612e-46 ], [ 489.8125 , 1.1378793724389997e-46 ], [ 489.875 , 1.1380490045744164e-46 ], [ 489.9375 , 1.1382186619981403e-46 ], 
                    [ 490.0 , 1.1383883447139652e-46 ], [ 490.0625 , 1.1385580527256618e-46 ], [ 490.125 , 1.138727786036969e-46 ], [ 490.1875 , 1.1388975446517067e-46 ], 
                    [ 490.25 , 1.1390673285736148e-46 ], [ 490.3125 , 1.1392371378064576e-46 ], [ 490.375 , 1.1394069723540414e-46 ], [ 490.4375 , 1.139576832220123e-46 ], 
                    [ 490.5 , 1.1397467174084778e-46 ], [ 490.5625 , 1.1399166279228882e-46 ], [ 490.625 , 1.140087378474955e-46 ], [ 490.6875 , 1.1402584632437761e-46 ], 
                    [ 490.75 , 1.1404295736860827e-46 ], [ 490.8125 , 1.1406007098057028e-46 ], [ 490.875 , 1.1407718716064898e-46 ], [ 490.9375 , 1.1409430590923219e-46 ], 
                    [ 491.0 , 1.141114272267037e-46 ], [ 491.0625 , 1.141285511134474e-46 ], [ 491.125 , 1.1414567756985209e-46 ], [ 491.1875 , 1.1416280659630012e-46 ], 
                    [ 491.25 , 1.1417993819317961e-46 ], [ 491.3125 , 1.1419707236087545e-46 ], [ 491.375 , 1.1421420909977265e-46 ], [ 491.4375 , 1.1423134841026108e-46 ], 
                    [ 491.5 , 1.1424849029271934e-46 ], [ 491.5625 , 1.1426563474753991e-46 ], [ 491.625 , 1.1428278177510551e-46 ], [ 491.6875 , 1.1429993137580305e-46 ], 
                    [ 491.75 , 1.143171501534902e-46 ], [ 491.8125 , 1.1433437625060689e-46 ], [ 491.875 , 1.1435160494347076e-46 ], [ 491.9375 , 1.1436883623247296e-46 ], 
                    [ 492.0 , 1.1438607011800555e-46 ], [ 492.0625 , 1.1440330660045729e-46 ], [ 492.125 , 1.14420545680222e-46 ], [ 492.1875 , 1.1443778735769019e-46 ], 
                    [ 492.25 , 1.1445503163325335e-46 ], [ 492.3125 , 1.1447227850730376e-46 ], [ 492.375 , 1.144895279802322e-46 ], [ 492.4375 , 1.1450678005242862e-46 ], 
                    [ 492.5 , 1.1452403472428795e-46 ], [ 492.5625 , 1.1454129199619953e-46 ], [ 492.625 , 1.1455855186855592e-46 ], [ 492.6875 , 1.1457581434174819e-46 ], 
                    [ 492.75 , 1.1459307941617148e-46 ], [ 492.8125 , 1.1461035193845316e-46 ], [ 492.875 , 1.1462766977386955e-46 ], [ 492.9375 , 1.1464499022604279e-46 ], 
                    [ 493.0 , 1.1466231329536746e-46 ], [ 493.0625 , 1.1467963898224066e-46 ], [ 493.125 , 1.1469696728705707e-46 ], [ 493.1875 , 1.1471429821021066e-46 ], 
                    [ 493.25 , 1.1473163175210111e-46 ], [ 493.3125 , 1.1474896791312006e-46 ], [ 493.375 , 1.1476630669366654e-46 ], [ 493.4375 , 1.1478364809413305e-46 ], 
                    [ 493.5 , 1.1480099211491798e-46 ], [ 493.5625 , 1.1481833875641722e-46 ], [ 493.625 , 1.1483568801902434e-46 ], [ 493.6875 , 1.14853039903137e-46 ], 
                    [ 493.75 , 1.1487039440915134e-46 ], [ 493.8125 , 1.1488775153746516e-46 ], [ 493.875 , 1.1490511128847225e-46 ], [ 493.9375 , 1.1492248095532147e-46 ], 
                    [ 494.0 , 1.1493987772167542e-46 ], [ 494.0625 , 1.1495727712152176e-46 ], [ 494.125 , 1.1497467915525835e-46 ], [ 494.1875 , 1.1499208382328387e-46 ], 
                    [ 494.25 , 1.1500949112600038e-46 ], [ 494.3125 , 1.1502690106380345e-46 ], [ 494.375 , 1.1504431363709196e-46 ], [ 494.4375 , 1.1506172884626571e-46 ], 
                    [ 494.5 , 1.150791466917245e-46 ], [ 494.5625 , 1.1509656717386496e-46 ], [ 494.625 , 1.1511399029309033e-46 ], [ 494.6875 , 1.1513141604979573e-46 ], 
                    [ 494.75 , 1.15148844444382e-46 ], [ 494.8125 , 1.1516627547725015e-46 ], [ 494.875 , 1.1518370914879463e-46 ], [ 494.9375 , 1.152011454594214e-46 ], 
                    [ 495.0 , 1.1521858440952592e-46 ], [ 495.0625 , 1.1523603288925338e-46 ], [ 495.125 , 1.1525349886080864e-46 ], [ 495.1875 , 1.1527096747962758e-46 ], 
                    [ 495.25 , 1.1528843874611138e-46 ], [ 495.3125 , 1.1530591266066222e-46 ], [ 495.375 , 1.1532338922367899e-46 ], [ 495.4375 , 1.1534086843556637e-46 ], 
                    [ 495.5 , 1.1535835029672092e-46 ], [ 495.5625 , 1.1537583480755075e-46 ], [ 495.625 , 1.1539332196845419e-46 ], [ 495.6875 , 1.1541081177983123e-46 ], 
                    [ 495.75 , 1.1542830424208613e-46 ], [ 495.8125 , 1.1544579935561812e-46 ], [ 495.875 , 1.1546329712083322e-46 ], [ 495.9375 , 1.1548079753813083e-46 ], 
                    [ 496.0 , 1.154983006079113e-46 ], [ 496.0625 , 1.1551580633057912e-46 ], [ 496.125 , 1.1553331470653556e-46 ], [ 496.1875 , 1.1555083076591993e-46 ], 
                    [ 496.25 , 1.1556835828872398e-46 ], [ 496.3125 , 1.1558588847021988e-46 ], [ 496.375 , 1.1560342131081011e-46 ], [ 496.4375 , 1.1562095681090054e-46 ], 
                    [ 496.5 , 1.1563849497089286e-46 ], [ 496.5625 , 1.1565603579118898e-46 ], [ 496.625 , 1.1567357927219564e-46 ], [ 496.6875 , 1.1569112541431404e-46 ], 
                    [ 496.75 , 1.157086742179486e-46 ], [ 496.8125 , 1.157262256835039e-46 ], [ 496.875 , 1.1574377981138205e-46 ], [ 496.9375 , 1.157613366019877e-46 ], 
                    [ 497.0 , 1.157788960557256e-46 ], [ 497.0625 , 1.1579645817299806e-46 ], [ 497.125 , 1.1581402295420993e-46 ], [ 497.1875 , 1.158315903997661e-46 ], 
                    [ 497.25 , 1.1584916051006908e-46 ], [ 497.3125 , 1.1586673656891267e-46 ], [ 497.375 , 1.158843208992474e-46 ], [ 497.4375 , 1.1590190789823994e-46 ], 
                    [ 497.5 , 1.1591949756629533e-46 ], [ 497.5625 , 1.159370899038186e-46 ], [ 497.625 , 1.1595468491121327e-46 ], [ 497.6875 , 1.1597228258888862e-46 ], 
                    [ 497.75 , 1.1598988293724497e-46 ], [ 497.8125 , 1.1600748595669255e-46 ], [ 497.875 , 1.1602509164763264e-46 ], [ 497.9375 , 1.1604270001047397e-46 ], 
                    [ 498.0 , 1.160603110456187e-46 ], [ 498.0625 , 1.1607792475347492e-46 ], [ 498.125 , 1.1609554113444656e-46 ], [ 498.1875 , 1.1611316018894016e-46 ], 
                    [ 498.25 , 1.1613078191736225e-46 ], [ 498.3125 , 1.1614840632011617e-46 ], [ 498.375 , 1.1616603339761195e-46 ], [ 498.4375 , 1.1618366522152257e-46 ], 
                    [ 498.5 , 1.1620130384199795e-46 ], [ 498.5625 , 1.1621894514030963e-46 ], [ 498.625 , 1.1623658911686416e-46 ], [ 498.6875 , 1.162542357720706e-46 ], 
                    [ 498.75 , 1.1627188510633314e-46 ], [ 498.8125 , 1.1628953712006017e-46 ], [ 498.875 , 1.1630719181365684e-46 ], [ 498.9375 , 1.1632484918753079e-46 ], 
                    [ 499.0 , 1.1634250924208729e-46 ], [ 499.0625 , 1.1636017197773829e-46 ], [ 499.125 , 1.1637783739488665e-46 ], [ 499.1875 , 1.1639550549393951e-46 ], 
                    [ 499.25 , 1.1641317627530567e-46 ], [ 499.3125 , 1.1643084973939067e-46 ], [ 499.375 , 1.1644852588660266e-46 ], [ 499.4375 , 1.164662047173473e-46 ], 
                    [ 499.5 , 1.1648388623203697e-46 ], [ 499.5625 , 1.1650157112338513e-46 ], [ 499.625 , 1.165192606819473e-46 ], [ 499.6875 , 1.1653695292648511e-46 ], 
                    [ 499.75 , 1.1655464785740725e-46 ], [ 499.8125 , 1.1657234547512247e-46 ], [ 499.875 , 1.1659004578003788e-46 ], [ 499.9375 , 1.1660774877256069e-46 ], 
                    [ 500.0 , 1.16625454453099e-46 ], [ 500.0625 , 1.1664316282206337e-46 ], [ 500.125 , 1.1666087387986127e-46 ], [ 500.1875 , 1.1667858762689677e-46 ], 
                    [ 500.25 , 1.1669630406358487e-46 ], [ 500.3125 , 1.167140231903273e-46 ], [ 500.375 , 1.167317450075392e-46 ], [ 500.4375 , 1.1674946951562322e-46 ], 
                    [ 500.5 , 1.1676719671499213e-46 ], [ 500.5625 , 1.1678492660605125e-46 ], [ 500.625 , 1.168026591892126e-46 ], [ 500.6875 , 1.1682039458524764e-46 ], 
                    [ 500.75 , 1.1683813340019421e-46 ], [ 500.8125 , 1.1685587490872734e-46 ], [ 500.875 , 1.1687361911125192e-46 ], [ 500.9375 , 1.1689136600817703e-46 ], 
                    [ 501.0 , 1.169091155999151e-46 ], [ 501.0625 , 1.169268678868745e-46 ], [ 501.125 , 1.1694462286946285e-46 ], [ 501.1875 , 1.1696238054808943e-46 ], 
                    [ 501.25 , 1.1698014092316535e-46 ], [ 501.3125 , 1.1699790399509836e-46 ], [ 501.375 , 1.170156697643013e-46 ], [ 501.4375 , 1.1703343823117708e-46 ], 
                    [ 501.5 , 1.1705120939614365e-46 ], [ 501.5625 , 1.1706898325960406e-46 ], [ 501.625 , 1.170867598219714e-46 ], [ 501.6875 , 1.1710453908365467e-46 ], 
                    [ 501.75 , 1.1712232104506289e-46 ], [ 501.8125 , 1.1714010570660852e-46 ], [ 501.875 , 1.171578923310998e-46 ], [ 501.9375 , 1.1717568164469255e-46 ], 
                    [ 502.0 , 1.1719347365942615e-46 ], [ 502.0625 , 1.1721126837570745e-46 ], [ 502.125 , 1.1722906579394661e-46 ], [ 502.1875 , 1.1724686591455477e-46 ], 
                    [ 502.25 , 1.1726466873794307e-46 ], [ 502.3125 , 1.1728247426452025e-46 ], [ 502.375 , 1.173002824946984e-46 ], [ 502.4375 , 1.1731809342888555e-46 ], 
                    [ 502.5 , 1.1733590706749563e-46 ], [ 502.5625 , 1.1735372341093845e-46 ], [ 502.625 , 1.1737154245962133e-46 ], [ 502.6875 , 1.1738936421395926e-46 ], 
                    [ 502.75 , 1.1740718867436053e-46 ], [ 502.8125 , 1.1742501584123689e-46 ], [ 502.875 , 1.1744284571500012e-46 ], [ 502.9375 , 1.1746067829605954e-46 ], 
                    [ 503.0 , 1.1747851225553659e-46 ], [ 503.0625 , 1.174963485598933e-46 ], [ 503.125 , 1.1751418757226651e-46 ], [ 503.1875 , 1.1753202929306745e-46 ], 
                    [ 503.25 , 1.1754987372270724e-46 ], [ 503.3125 , 1.175677208615972e-46 ], [ 503.375 , 1.175855707101495e-46 ], [ 503.4375 , 1.1760342326877384e-46 ], 
                    [ 503.5 , 1.1762127853788172e-46 ], [ 503.5625 , 1.1763913651788627e-46 ], [ 503.625 , 1.1765699720919832e-46 ], [ 503.6875 , 1.1767486061223029e-46 ], 
                    [ 503.75 , 1.176927267273914e-46 ], [ 503.8125 , 1.1771059555509678e-46 ], [ 503.875 , 1.1772846709575576e-46 ], [ 503.9375 , 1.1774634134978104e-46 ], 
                    [ 504.0 , 1.177642183175846e-46 ], [ 504.0625 , 1.1778209799957848e-46 ], [ 504.125 , 1.1779997848399936e-46 ], [ 504.1875 , 1.1781786008904155e-46 ], 
                    [ 504.25 , 1.1783574440844503e-46 ], [ 504.3125 , 1.1785363144262266e-46 ], [ 504.375 , 1.178715211919857e-46 ], [ 504.4375 , 1.1788941365694795e-46 ], 
                    [ 504.5 , 1.1790730883791997e-46 ], [ 504.5625 , 1.1792520673531574e-46 ], [ 504.625 , 1.1794310734954672e-46 ], [ 504.6875 , 1.1796101068102454e-46 ], 
                    [ 504.75 , 1.1797891673016413e-46 ], [ 504.8125 , 1.179968254973755e-46 ], [ 504.875 , 1.1801473698307046e-46 ], [ 504.9375 , 1.1803265118766579e-46 ], 
                    [ 505.0 , 1.1805056811157176e-46 ], [ 505.0625 , 1.1806848775519858e-46 ], [ 505.125 , 1.1808641011896333e-46 ], [ 505.1875 , 1.181043352032764e-46 ], 
                    [ 505.25 , 1.1812226211452782e-46 ], [ 505.3125 , 1.1814018925201378e-46 ], [ 505.375 , 1.1815811911025833e-46 ], [ 505.4375 , 1.1817605168967693e-46 ], 
                    [ 505.5 , 1.1819398699068001e-46 ], [ 505.5625 , 1.1821192501368066e-46 ], [ 505.625 , 1.182298657590936e-46 ], [ 505.6875 , 1.1824780922733126e-46 ], 
                    [ 505.75 , 1.1826575541880767e-46 ], [ 505.8125 , 1.1828370433393441e-46 ], [ 505.875 , 1.1830165597312659e-46 ], [ 505.9375 , 1.183196103367959e-46 ], 
                    [ 506.0 , 1.1833756742535589e-46 ], [ 506.0625 , 1.1835552723922173e-46 ], [ 506.125 , 1.1837348977880622e-46 ], [ 506.1875 , 1.1839145504452472e-46 ], 
                    [ 506.25 , 1.1840942303678759e-46 ], [ 506.3125 , 1.1842739375600864e-46 ], [ 506.375 , 1.184453672026051e-46 ], [ 506.4375 , 1.1846333897142274e-46 ], 
                    [ 506.5 , 1.184813131906388e-46 ], [ 506.5625 , 1.184992901370485e-46 ], [ 506.625 , 1.1851726981106637e-46 ], [ 506.6875 , 1.185352522131063e-46 ], 
                    [ 506.75 , 1.1855323734358222e-46 ], [ 506.8125 , 1.185712252029081e-46 ], [ 506.875 , 1.1858921579149967e-46 ], [ 506.9375 , 1.186072091097668e-46 ], 
                    [ 507.0 , 1.186252051581279e-46 ], [ 507.0625 , 1.1864320393699469e-46 ], [ 507.125 , 1.1866120544678226e-46 ], [ 507.1875 , 1.1867920968790415e-46 ], 
                    [ 507.25 , 1.1869721666077816e-46 ], [ 507.3125 , 1.1871522636581457e-46 ], [ 507.375 , 1.187332388034304e-46 ], [ 507.4375 , 1.1875125397403777e-46 ], 
                    [ 507.5 , 1.1876927187805306e-46 ], [ 507.5625 , 1.1878728989519177e-46 ], [ 507.625 , 1.1880530872014014e-46 ], [ 507.6875 , 1.1882333027836033e-46 ], 
                    [ 507.75 , 1.1884135457026947e-46 ], [ 507.8125 , 1.1885938159627972e-46 ], [ 507.875 , 1.1887741135680497e-46 ], [ 507.9375 , 1.1889544385226339e-46 ], 
                    [ 508.0 , 1.1891347908306813e-46 ], [ 508.0625 , 1.1893151704963416e-46 ], [ 508.125 , 1.1894955775237643e-46 ], [ 508.1875 , 1.1896760119170917e-46 ], 
                    [ 508.25 , 1.1898564736804834e-46 ], [ 508.3125 , 1.1900369628180994e-46 ], [ 508.375 , 1.1902174793340669e-46 ], [ 508.4375 , 1.1903980232325809e-46 ], 
                    [ 508.5 , 1.190578594517753e-46 ], [ 508.5625 , 1.1907591931937714e-46 ], [ 508.625 , 1.190939819264757e-46 ], [ 508.6875 , 1.1911204629608173e-46 ], 
                    [ 508.75 , 1.191301086410501e-46 ], [ 508.8125 , 1.191481737250211e-46 ], [ 508.875 , 1.1916624154841174e-46 ], [ 508.9375 , 1.1918431211163573e-46 ], 
                    [ 509.0 , 1.1920238541511023e-46 ], [ 509.0625 , 1.1922046145925082e-46 ], [ 509.125 , 1.1923854024447136e-46 ], [ 509.1875 , 1.1925662177118668e-46 ], 
                    [ 509.25 , 1.1927470603981593e-46 ], [ 509.3125 , 1.1929279305077399e-46 ], [ 509.375 , 1.1931088280447419e-46 ], [ 509.4375 , 1.1932897530133415e-46 ], 
                    [ 509.5 , 1.1934707054177155e-46 ], [ 509.5625 , 1.1936516852619902e-46 ], [ 509.625 , 1.1938326925503435e-46 ], [ 509.6875 , 1.1940137272869459e-46 ], 
                    [ 509.75 , 1.1941947894759511e-46 ], [ 509.8125 , 1.1943758791215217e-46 ], [ 509.875 , 1.1945569489714876e-46 ], [ 509.9375 , 1.194738028124349e-46 ], 
                    [ 510.0 , 1.1949191347264242e-46 ], [ 510.0625 , 1.1951002687818829e-46 ], [ 510.125 , 1.1952814302948783e-46 ], [ 510.1875 , 1.195462619269598e-46 ], 
                    [ 510.25 , 1.1956438357101794e-46 ], [ 510.3125 , 1.1958250796207946e-46 ], [ 510.375 , 1.1960063510056072e-46 ], [ 510.4375 , 1.1961876498687908e-46 ], 
                    [ 510.5 , 1.1963689762144852e-46 ], [ 510.5625 , 1.196550330046899e-46 ], [ 510.625 , 1.196731711370139e-46 ], [ 510.6875 , 1.1969131201884408e-46 ], 
                    [ 510.75 , 1.1970945565059127e-46 ], [ 510.8125 , 1.1972760203267657e-46 ], [ 510.875 , 1.1974575116551517e-46 ], [ 510.9375 , 1.1976390304952238e-46 ], 
                    [ 511.0 , 1.1978205639893139e-46 ], [ 511.0625 , 1.1980020826064487e-46 ], [ 511.125 , 1.1981836287310326e-46 ], [ 511.1875 , 1.1983652023672762e-46 ], 
                    [ 511.25 , 1.1985468035193151e-46 ], [ 511.3125 , 1.1987284321913186e-46 ], [ 511.375 , 1.1989100883874746e-46 ], [ 511.4375 , 1.1990917721119453e-46 ], 
                    [ 511.5 , 1.1992734833688938e-46 ], [ 511.5625 , 1.199455222162501e-46 ], [ 511.625 , 1.199636988496957e-46 ], [ 511.6875 , 1.1998187823763928e-46 ], 
                    [ 511.75 , 1.2000006038050332e-46 ], [ 511.8125 , 1.200182452787011e-46 ], [ 511.875 , 1.200364329326527e-46 ], [ 511.9375 , 1.2005462334277402e-46 ], 
                    [ 512.0 , 1.2007281650948359e-46 ], [ 512.0625 , 1.2009101243320084e-46 ], [ 512.125 , 1.2010921111433846e-46 ], [ 512.1875 , 1.201274076341209e-46 ], 
                    [ 512.25 , 1.2014560481494849e-46 ], [ 512.3125 , 1.2016380475232933e-46 ], [ 512.375 , 1.2018200744667843e-46 ], [ 512.4375 , 1.2020021289841426e-46 ], 
                    [ 512.5 , 1.2021842110795536e-46 ], [ 512.5625 , 1.202366320757178e-46 ], [ 512.625 , 1.2025484580212027e-46 ], [ 512.6875 , 1.2027306228758065e-46 ], 
                    [ 512.75 , 1.202912815325177e-46 ], [ 512.8125 , 1.2030950353734864e-46 ], [ 512.875 , 1.2032772830249064e-46 ], [ 512.9375 , 1.203459558283627e-46 ], 
                    [ 513.0 , 1.2036418611538221e-46 ], [ 513.0625 , 1.2038241916397081e-46 ], [ 513.125 , 1.2040065497454259e-46 ], [ 513.1875 , 1.2041889354751588e-46 ], 
                    [ 513.25 , 1.2043713488331435e-46 ], [ 513.3125 , 1.2045537813670848e-46 ], [ 513.375 , 1.2047361837513577e-46 ], [ 513.4375 , 1.2049186137563485e-46 ], 
                    [ 513.5 , 1.2051010713862054e-46 ], [ 513.5625 , 1.2052835566451459e-46 ], [ 513.625 , 1.2054660695373535e-46 ], [ 513.6875 , 1.2056486100670045e-46 ], 
                    [ 513.75 , 1.2058311782382664e-46 ], [ 513.8125 , 1.206013774055351e-46 ], [ 513.875 , 1.2061963975224443e-46 ], [ 513.9375 , 1.206379048643725e-46 ], 
                    [ 514.0 , 1.2065617274233634e-46 ], [ 514.0625 , 1.2067444338655734e-46 ], [ 514.125 , 1.2069271679745608e-46 ], [ 514.1875 , 1.2071099297544557e-46 ], 
                    [ 514.25 , 1.207292719209482e-46 ], [ 514.3125 , 1.207475536343848e-46 ], [ 514.375 , 1.2076583811617277e-46 ], [ 514.4375 , 1.207841253667296e-46 ], 
                    [ 514.5 , 1.2080241142678336e-46 ], [ 514.5625 , 1.2082069648300559e-46 ], [ 514.625 , 1.2083898430691493e-46 ], [ 514.6875 , 1.2085727489893034e-46 ], 
                    [ 514.75 , 1.2087556825947168e-46 ], [ 514.8125 , 1.2089386438895716e-46 ], [ 514.875 , 1.2091216328780497e-46 ], [ 514.9375 , 1.2093046495643691e-46 ], 
                    [ 515.0 , 1.209487693952688e-46 ], [ 515.0625 , 1.2096707660472253e-46 ], [ 515.125 , 1.209853865852157e-46 ], [ 515.1875 , 1.2100369933716864e-46 ], 
                    [ 515.25 , 1.2102201486100258e-46 ], [ 515.3125 , 1.2104033315713274e-46 ], [ 515.375 , 1.2105865422598308e-46 ], [ 515.4375 , 1.210769780679707e-46 ], 
                    [ 515.5 , 1.2109530468351622e-46 ], [ 515.5625 , 1.2111363407304027e-46 ], [ 515.625 , 1.2113196623696277e-46 ], [ 515.6875 , 1.2115029496144758e-46 ], 
                    [ 515.75 , 1.2116862539878288e-46 ], [ 515.8125 , 1.2118695860957525e-46 ], [ 515.875 , 1.2120529459424174e-46 ], [ 515.9375 , 1.212236333532029e-46 ], 
                    [ 516.0 , 1.2124197488687938e-46 ], [ 516.0625 , 1.2126031919569009e-46 ], [ 516.125 , 1.2127866628005407e-46 ], [ 516.1875 , 1.2129701614039217e-46 ], 
                    [ 516.25 , 1.2131536877712522e-46 ], [ 516.3125 , 1.2133372419067073e-46 ], [ 516.375 , 1.213520823814523e-46 ], [ 516.4375 , 1.2137044334988755e-46 ], 
                    [ 516.5 , 1.2138880709639846e-46 ], [ 516.5625 , 1.2140717362140188e-46 ], [ 516.625 , 1.2142554292532343e-46 ], [ 516.6875 , 1.2144391500857927e-46 ], 
                    [ 516.75 , 1.2146228987159504e-46 ], [ 516.8125 , 1.2148066635383358e-46 ], [ 516.875 , 1.2149903924157893e-46 ], [ 516.9375 , 1.2151741490806363e-46 ], 
                    [ 517.0 , 1.2153579335370622e-46 ], [ 517.0625 , 1.215541745789279e-46 ], [ 517.125 , 1.2157255858414993e-46 ], [ 517.1875 , 1.2159094536979099e-46 ], 
                    [ 517.25 , 1.2160933493627335e-46 ], [ 517.3125 , 1.2162772728401584e-46 ], [ 517.375 , 1.2164612241344e-46 ], [ 517.4375 , 1.2166452032496563e-46 ], 
                    [ 517.5 , 1.2168292101901787e-46 ], [ 517.5625 , 1.2170132449601144e-46 ], [ 517.625 , 1.217197307563699e-46 ], [ 517.6875 , 1.2173813980051415e-46 ], 
                    [ 517.75 , 1.2175655162886523e-46 ], [ 517.8125 , 1.2177496624184337e-46 ], [ 517.875 , 1.2179338363987231e-46 ], [ 517.9375 , 1.2181180382337155e-46 ], 
                    [ 518.0 , 1.2183022362255156e-46 ], [ 518.0625 , 1.2184864167294948e-46 ], [ 518.125 , 1.2186706250775156e-46 ], [ 518.1875 , 1.2188548612738045e-46 ], 
                    [ 518.25 , 1.2190391253225545e-46 ], [ 518.3125 , 1.2192234172279678e-46 ], [ 518.375 , 1.2194077369942548e-46 ], [ 518.4375 , 1.2195920846256459e-46 ], 
                    [ 518.5 , 1.2197764601263443e-46 ], [ 518.5625 , 1.2199608635005632e-46 ], [ 518.625 , 1.2201452947525084e-46 ], [ 518.6875 , 1.220329753886411e-46 ], 
                    [ 518.75 , 1.2205142409064782e-46 ], [ 518.8125 , 1.2206987558169252e-46 ], [ 518.875 , 1.2208832986219687e-46 ], [ 518.9375 , 1.2210678693258343e-46 ], 
                    [ 519.0 , 1.2212524679327223e-46 ], [ 519.0625 , 1.2214370944468686e-46 ], [ 519.125 , 1.2216217488724485e-46 ], [ 519.1875 , 1.2218063755537033e-46 ], 
                    [ 519.25 , 1.2219909969587246e-46 ], [ 519.3125 , 1.2221756462610077e-46 ], [ 519.375 , 1.2223603234647856e-46 ], [ 519.4375 , 1.2225450285742571e-46 ], 
                    [ 519.5 , 1.2227297615936476e-46 ], [ 519.5625 , 1.2229145225271827e-46 ], [ 519.625 , 1.2230993113790549e-46 ], [ 519.6875 , 1.2232841281535172e-46 ], 
                    [ 519.75 , 1.2234689728547886e-46 ], [ 519.8125 , 1.2236538454870288e-46 ], [ 519.875 , 1.2238387460545363e-46 ], [ 519.9375 , 1.2240236745614713e-46 ], 
                    [ 520.0 , 1.224208631012099e-46 ], [ 520.0625 , 1.2243936154106157e-46 ], [ 520.125 , 1.2245786277612709e-46 ], [ 520.1875 , 1.2247636680682446e-46 ], 
                    [ 520.25 , 1.2249487363358048e-46 ], [ 520.3125 , 1.2251338325681501e-46 ], [ 520.375 , 1.2253188992452063e-46 ], [ 520.4375 , 1.2255039794084664e-46 ], 
                    [ 520.5 , 1.2256890875274422e-46 ], [ 520.5625 , 1.2258742236063475e-46 ], [ 520.625 , 1.2260593876494318e-46 ], [ 520.6875 , 1.226244579660893e-46 ], 
                    [ 520.75 , 1.2264297996449642e-46 ], [ 520.8125 , 1.2266150476058793e-46 ], [ 520.875 , 1.226800323547838e-46 ], [ 520.9375 , 1.2269856274751101e-46 ], 
                    [ 521.0 , 1.227170959391862e-46 ], [ 521.0625 , 1.227356319302391e-46 ], [ 521.125 , 1.2275417072108556e-46 ], [ 521.1875 , 1.227727123121546e-46 ], 
                    [ 521.25 , 1.2279125670386657e-46 ], [ 521.3125 , 1.2280980389664273e-46 ], [ 521.375 , 1.2282835389090882e-46 ], [ 521.4375 , 1.2284690668708884e-46 ], 
                    [ 521.5 , 1.2286546228560165e-46 ], [ 521.5625 , 1.2288401200591159e-46 ], [ 521.625 , 1.2290256385960554e-46 ], [ 521.6875 , 1.2292111851408193e-46 ], 
                    [ 521.75 , 1.2293967596976098e-46 ], [ 521.8125 , 1.229582362270691e-46 ], [ 521.875 , 1.2297679928642658e-46 ], [ 521.9375 , 1.2299536514825734e-46 ], 
                    [ 522.0 , 1.230139338129845e-46 ], [ 522.0625 , 1.2303250528103032e-46 ], [ 522.125 , 1.2305107955282062e-46 ], [ 522.1875 , 1.2306965662877785e-46 ], 
                    [ 522.25 , 1.2308823650932097e-46 ], [ 522.3125 , 1.2310681919487862e-46 ], [ 522.375 , 1.2312540468587426e-46 ], [ 522.4375 , 1.2314399298272625e-46 ], 
                    [ 522.5 , 1.2316258408586247e-46 ], [ 522.5625 , 1.2318117799570842e-46 ], [ 522.625 , 1.2319977471268256e-46 ], [ 522.6875 , 1.2321837409918041e-46 ], 
                    [ 522.75 , 1.2323696762727084e-46 ], [ 522.8125 , 1.2325556396110674e-46 ], [ 522.875 , 1.2327416310111061e-46 ], [ 522.9375 , 1.2329276504770592e-46 ], 
                    [ 523.0 , 1.2331136980131705e-46 ], [ 523.0625 , 1.2332997736236579e-46 ], [ 523.125 , 1.2334858773127582e-46 ], [ 523.1875 , 1.2336720090847257e-46 ], 
                    [ 523.25 , 1.2338581689437897e-46 ], [ 523.3125 , 1.2340443568941793e-46 ], [ 523.375 , 1.23423057294016e-46 ], [ 523.4375 , 1.2344168170859536e-46 ], 
                    [ 523.5 , 1.234603089335792e-46 ], [ 523.5625 , 1.234789389693907e-46 ], [ 523.625 , 1.234975718164584e-46 ], [ 523.6875 , 1.2351620747520212e-46 ], 
                    [ 523.75 , 1.2353484594604793e-46 ], [ 523.8125 , 1.2355348722942012e-46 ], [ 523.875 , 1.2357213084071082e-46 ], [ 523.9375 , 1.2359076962185558e-46 ], 
                    [ 524.0 , 1.2360941121434754e-46 ], [ 524.0625 , 1.2362805561861162e-46 ], [ 524.125 , 1.2364670283506839e-46 ], [ 524.1875 , 1.2366535286414732e-46 ], 
                    [ 524.25 , 1.2368400570626998e-46 ], [ 524.3125 , 1.2370266136185984e-46 ], [ 524.375 , 1.2372131983134123e-46 ], [ 524.4375 , 1.2373998111514121e-46 ], 
                    [ 524.5 , 1.2375864521368165e-46 ], [ 524.5625 , 1.237773121273871e-46 ], [ 524.625 , 1.2379598185668483e-46 ], [ 524.6875 , 1.2381465440199512e-46 ], 
                    [ 524.75 , 1.238333297637471e-46 ], [ 524.8125 , 1.2385200794236562e-46 ], [ 524.875 , 1.2387068893827111e-46 ], [ 524.9375 , 1.2388937275189208e-46 ], 
                    [ 525.0 , 1.239080593836535e-46 ], [ 525.0625 , 1.2392674822553626e-46 ], [ 525.125 , 1.2394543051226042e-46 ], [ 525.1875 , 1.2396411561538809e-46 ], 
                    [ 525.25 , 1.239828035353456e-46 ], [ 525.3125 , 1.2400149427255494e-46 ], [ 525.375 , 1.2402018782744437e-46 ], [ 525.4375 , 1.2403888420043513e-46 ], 
                    [ 525.5 , 1.240575833919556e-46 ], [ 525.5625 , 1.2407628540242799e-46 ], [ 525.625 , 1.2409499023227818e-46 ], [ 525.6875 , 1.241136978819312e-46 ], 
                    [ 525.75 , 1.2413240835181214e-46 ], [ 525.8125 , 1.2415112164234613e-46 ], [ 525.875 , 1.241698377539593e-46 ], [ 525.9375 , 1.2418855668707518e-46 ], 
                    [ 526.0 , 1.2420727844212175e-46 ], [ 526.0625 , 1.2422600301952001e-46 ], [ 526.125 , 1.2424473041969897e-46 ], [ 526.1875 , 1.2426346064308415e-46 ], 
                    [ 526.25 , 1.2428219352741437e-46 ], [ 526.3125 , 1.2430092078043316e-46 ], [ 526.375 , 1.2431965085533741e-46 ], [ 526.4375 , 1.2433838375255058e-46 ], 
                    [ 526.5 , 1.243571194724988e-46 ], [ 526.5625 , 1.2437585801560829e-46 ], [ 526.625 , 1.2439459938230273e-46 ], [ 526.6875 , 1.2441334357300844e-46 ], 
                    [ 526.75 , 1.2443209058815098e-46 ], [ 526.8125 , 1.2445084042815678e-46 ], [ 526.875 , 1.2446959309344977e-46 ], [ 526.9375 , 1.2448834858445741e-46 ], 
                    [ 527.0 , 1.2450710690160466e-46 ], [ 527.0625 , 1.2452586804531557e-46 ], [ 527.125 , 1.2454463201601866e-46 ], [ 527.1875 , 1.2456339881414e-46 ], 
                    [ 527.25 , 1.245821684401047e-46 ], [ 527.3125 , 1.2460094089433711e-46 ], [ 527.375 , 1.2461971617726694e-46 ], [ 527.4375 , 1.2463849428931775e-46 ], 
                    [ 527.5 , 1.2465726654779745e-46 ], [ 527.5625 , 1.2467604101916313e-46 ], [ 527.625 , 1.2469481831812796e-46 ], [ 527.6875 , 1.247135984451186e-46 ], 
                    [ 527.75 , 1.2473238140055928e-46 ], [ 527.8125 , 1.2475116718487859e-46 ], [ 527.875 , 1.2476995579849817e-46 ], [ 527.9375 , 1.2478874724184948e-46 ], 
                    [ 528.0 , 1.2480754151535421e-46 ], [ 528.0625 , 1.2482633861944134e-46 ], [ 528.125 , 1.2484513855453624e-46 ], [ 528.1875 , 1.2486394132106266e-46 ], 
                    [ 528.25 , 1.2488274691945322e-46 ], [ 528.3125 , 1.2490155535012999e-46 ], [ 528.375 , 1.2492036661351865e-46 ], [ 528.4375 , 1.249391807100503e-46 ], 
                    [ 528.5 , 1.2495799764014803e-46 ], [ 528.5625 , 1.2497681740423954e-46 ], [ 528.625 , 1.2499564000275075e-46 ], [ 528.6875 , 1.2501445745238833e-46 ], 
                    [ 528.75 , 1.2503327580608072e-46 ], [ 528.8125 , 1.2505209699248718e-46 ], [ 528.875 , 1.250709210120377e-46 ], [ 528.9375 , 1.2508974786515605e-46 ], 
                    [ 529.0 , 1.2510857755226965e-46 ], [ 529.0625 , 1.2512741007380513e-46 ], [ 529.125 , 1.2514624543018915e-46 ], [ 529.1875 , 1.251650836218493e-46 ], 
                    [ 529.25 , 1.25183924649208e-46 ], [ 529.3125 , 1.252027685126991e-46 ], [ 529.375 , 1.2522161521274605e-46 ], [ 529.4375 , 1.2524046474977406e-46 ], 
                    [ 529.5 , 1.2525931712421369e-46 ], [ 529.5625 , 1.252781723364903e-46 ], [ 529.625 , 1.2529703038703019e-46 ], [ 529.6875 , 1.2531589127625969e-46 ], 
                    [ 529.75 , 1.2533475500461054e-46 ], [ 529.8125 , 1.253536215725057e-46 ], [ 529.875 , 1.2537248521145246e-46 ], [ 529.9375 , 1.2539134835479171e-46 ], 
                    [ 530.0 , 1.2541021433621833e-46 ], [ 530.0625 , 1.2542908315616108e-46 ], [ 530.125 , 1.2544795481504617e-46 ], [ 530.1875 , 1.2546682931330073e-46 ], 
                    [ 530.25 , 1.2548570665135106e-46 ], [ 530.3125 , 1.2550458682962622e-46 ], [ 530.375 , 1.2552346984855265e-46 ], [ 530.4375 , 1.2554235570855773e-46 ], 
                    [ 530.5 , 1.2556124441006717e-46 ], [ 530.5625 , 1.2558013595351292e-46 ], [ 530.625 , 1.2559903033931814e-46 ], [ 530.6875 , 1.2561792756791312e-46 ], 
                    [ 530.75 , 1.256368276397247e-46 ], [ 530.8125 , 1.2565573055517978e-46 ], [ 530.875 , 1.2567463631470709e-46 ], [ 530.9375 , 1.2569354491873545e-46 ], 
                    [ 531.0 , 1.257124563676919e-46 ], [ 531.0625 , 1.257313662917484e-46 ], [ 531.125 , 1.2575027298844776e-46 ], [ 531.1875 , 1.257691825282197e-46 ], 
                    [ 531.25 , 1.2578809491148913e-46 ], [ 531.3125 , 1.258070101386845e-46 ], [ 531.375 , 1.2582592821023345e-46 ], [ 531.4375 , 1.2584484912656372e-46 ], 
                    [ 531.5 , 1.2586377288810308e-46 ], [ 531.5625 , 1.2588269949527937e-46 ], [ 531.625 , 1.2590162894852051e-46 ], [ 531.6875 , 1.2592056124825447e-46 ], 
                    [ 531.75 , 1.2593949639490927e-46 ], [ 531.8125 , 1.2595843438891394e-46 ], [ 531.875 , 1.2597737523069396e-46 ], [ 531.9375 , 1.2599631892068198e-46 ], 
                    [ 532.0 , 1.2601526545930189e-46 ], [ 532.0625 , 1.260342148469829e-46 ], [ 532.125 , 1.2605316708415703e-46 ], [ 532.1875 , 1.2607212217124831e-46 ], 
                    [ 532.25 , 1.2609107849139718e-46 ], [ 532.3125 , 1.2611002915966891e-46 ], [ 532.375 , 1.2612898267610272e-46 ], [ 532.4375 , 1.2614793904112754e-46 ], 
                    [ 532.5 , 1.2616689825517063e-46 ], [ 532.5625 , 1.261858603186602e-46 ], [ 532.625 , 1.2620482523202353e-46 ], [ 532.6875 , 1.262237929956899e-46 ], 
                    [ 532.75 , 1.2624276361008859e-46 ], [ 532.8125 , 1.262617370756471e-46 ], [ 532.875 , 1.2628071339279487e-46 ], [ 532.9375 , 1.2629969256195779e-46 ], 
                    [ 533.0 , 1.2631867458356719e-46 ], [ 533.0625 , 1.2633765945804906e-46 ], [ 533.125 , 1.263566471858358e-46 ], [ 533.1875 , 1.263756377673544e-46 ], 
                    [ 533.25 , 1.2639463120303375e-46 ], [ 533.3125 , 1.2641362749330199e-46 ], [ 533.375 , 1.2643262663858987e-46 ], [ 533.4375 , 1.264516286393256e-46 ], 
                    [ 533.5 , 1.2647062502691168e-46 ], [ 533.5625 , 1.264896228599667e-46 ], [ 533.625 , 1.2650862354678844e-46 ], [ 533.6875 , 1.2652762708780557e-46 ], 
                    [ 533.75 , 1.2654663348344777e-46 ], [ 533.8125 , 1.265656427341429e-46 ], [ 533.875 , 1.2658465484031808e-46 ], [ 533.9375 , 1.2660366980240493e-46 ], 
                    [ 534.0 , 1.266226876208324e-46 ], [ 534.0625 , 1.2664170829602964e-46 ], [ 534.125 , 1.2666073182842213e-46 ], [ 534.1875 , 1.2667975821844356e-46 ], 
                    [ 534.25 , 1.2669878746652053e-46 ], [ 534.3125 , 1.2671781957308232e-46 ], [ 534.375 , 1.2673685453856014e-46 ], [ 534.4375 , 1.2675589236338344e-46 ], 
                    [ 534.5 , 1.2677493304797904e-46 ], [ 534.5625 , 1.2679397659278014e-46 ], [ 534.625 , 1.2681302299821366e-46 ], [ 534.6875 , 1.2683206710777348e-46 ], 
                    [ 534.75 , 1.2685110914872324e-46 ], [ 534.8125 , 1.2687015404856703e-46 ], [ 534.875 , 1.2688920180773314e-46 ], [ 534.9375 , 1.2690825242664999e-46 ], 
                    [ 535.0 , 1.269273059057496e-46 ], [ 535.0625 , 1.269463622454569e-46 ], [ 535.125 , 1.269654214462068e-46 ], [ 535.1875 , 1.2698448350842432e-46 ], 
                    [ 535.25 , 1.2700354843254265e-46 ], [ 535.3125 , 1.2702261621898878e-46 ], [ 535.375 , 1.2704168686819243e-46 ], [ 535.4375 , 1.2706076038058345e-46 ], 
                    [ 535.5 , 1.2707983675659526e-46 ], [ 535.5625 , 1.2709891599665332e-46 ], [ 535.625 , 1.271179981011912e-46 ], [ 535.6875 , 1.2713708307063631e-46 ], 
                    [ 535.75 , 1.2715617090541963e-46 ], [ 535.8125 , 1.2717526160597133e-46 ], [ 535.875 , 1.2719435390863636e-46 ], [ 535.9375 , 1.272134399311296e-46 ], 
                    [ 536.0 , 1.2723252881755882e-46 ], [ 536.0625 , 1.2725162056835281e-46 ], [ 536.125 , 1.2727071518393783e-46 ], [ 536.1875 , 1.2728981266474824e-46 ], 
                    [ 536.25 , 1.2730891301121214e-46 ], [ 536.3125 , 1.2732801622375778e-46 ], [ 536.375 , 1.2734712230281883e-46 ], [ 536.4375 , 1.2736623124882176e-46 ], 
                    [ 536.5 , 1.2738534306220046e-46 ], [ 536.5625 , 1.2740445774338243e-46 ], [ 536.625 , 1.274235752928007e-46 ], [ 536.6875 , 1.2744269571087936e-46 ], 
                    [ 536.75 , 1.2746181899805879e-46 ], [ 536.8125 , 1.2748094515476229e-46 ], [ 536.875 , 1.2750007418142044e-46 ], [ 536.9375 , 1.2751920607846747e-46 ], 
                    [ 537.0 , 1.2753834084633236e-46 ], [ 537.0625 , 1.27557478485444e-46 ], [ 537.125 , 1.2757661161621409e-46 ], [ 537.1875 , 1.2759574426213372e-46 ], 
                    [ 537.25 , 1.2761487977737349e-46 ], [ 537.3125 , 1.276340181623637e-46 ], [ 537.375 , 1.2765315941753472e-46 ], [ 537.4375 , 1.27672303543317e-46 ], 
                    [ 537.5 , 1.2769145054014286e-46 ], [ 537.5625 , 1.2771060040843834e-46 ], [ 537.625 , 1.277297531486377e-46 ], [ 537.6875 , 1.277489087611707e-46 ], 
                    [ 537.75 , 1.2776806724646726e-46 ], [ 537.8125 , 1.2778722860495995e-46 ], [ 537.875 , 1.2780639283707971e-46 ], [ 537.9375 , 1.2782555994325569e-46 ], 
                    [ 538.0 , 1.2784472992391978e-46 ], [ 538.0625 , 1.2786390277950399e-46 ], [ 538.125 , 1.2788307851043676e-46 ], [ 538.1875 , 1.2790225711715202e-46 ], 
                    [ 538.25 , 1.2792143860008016e-46 ], [ 538.3125 , 1.2794062070971004e-46 ], [ 538.375 , 1.2795979720174957e-46 ], [ 538.4375 , 1.2797897656807804e-46 ], 
                    [ 538.5 , 1.279981588091181e-46 ], [ 538.5625 , 1.2801734392530697e-46 ], [ 538.625 , 1.280365319170756e-46 ], [ 538.6875 , 1.2805572278485138e-46 ], 
                    [ 538.75 , 1.2807491652906812e-46 ], [ 538.8125 , 1.280941131501542e-46 ], [ 538.875 , 1.2811331264854356e-46 ], [ 538.9375 , 1.2813251502466837e-46 ], 
                    [ 539.0 , 1.2815172027895634e-46 ], [ 539.0625 , 1.2817092841184068e-46 ], [ 539.125 , 1.2819013942375102e-46 ], [ 539.1875 , 1.2820935331512435e-46 ], 
                    [ 539.25 , 1.2822857008638774e-46 ], [ 539.3125 , 1.282477897379728e-46 ], [ 539.375 , 1.2826701227031308e-46 ], [ 539.4375 , 1.2828623768383674e-46 ], 
                    [ 539.5 , 1.2830546597898202e-46 ], [ 539.5625 , 1.2832468936101224e-46 ], [ 539.625 , 1.283439120770902e-46 ], [ 539.6875 , 1.2836313767268454e-46 ], 
                    [ 539.75 , 1.28382366148223e-46 ], [ 539.8125 , 1.284015975041406e-46 ], [ 539.875 , 1.284208317408661e-46 ], [ 539.9375 , 1.284400688588319e-46 ], 
                    [ 540.0 , 1.2845930885847059e-46 ], [ 540.0625 , 1.284785517402129e-46 ], [ 540.125 , 1.284977975044896e-46 ], [ 540.1875 , 1.2851704615173164e-46 ], 
                    [ 540.25 , 1.2853629768237543e-46 ], [ 540.3125 , 1.2855555209685012e-46 ], [ 540.375 , 1.2857480939558681e-46 ], [ 540.4375 , 1.2859406957901758e-46 ], 
                    [ 540.5 , 1.2861333264757633e-46 ], [ 540.5625 , 1.286325986016962e-46 ], [ 540.625 , 1.2865186744180395e-46 ], [ 540.6875 , 1.2867113916833917e-46 ], 
                    [ 540.75 , 1.2869041248151545e-46 ], [ 540.8125 , 1.2870967923126989e-46 ], [ 540.875 , 1.2872894886552734e-46 ], [ 540.9375 , 1.2874822138471598e-46 ], 
                    [ 541.0 , 1.2876749678927047e-46 ], [ 541.0625 , 1.2878677507962187e-46 ], [ 541.125 , 1.2880605625620134e-46 ], [ 541.1875 , 1.288253403194437e-46 ], 
                    [ 541.25 , 1.2884462726977747e-46 ], [ 541.3125 , 1.2886391710763675e-46 ], [ 541.375 , 1.2888320983345473e-46 ], [ 541.4375 , 1.2890250544766375e-46 ], 
                    [ 541.5 , 1.289218039506917e-46 ], [ 541.5625 , 1.2894110534297934e-46 ], [ 541.625 , 1.2896040962495095e-46 ], [ 541.6875 , 1.2897971679704465e-46 ], 
                    [ 541.75 , 1.2899902685969226e-46 ], [ 541.8125 , 1.2901833981332464e-46 ], [ 541.875 , 1.290376556583747e-46 ], [ 541.9375 , 1.2905697439527703e-46 ], 
                    [ 542.0 , 1.290762903138557e-46 ], [ 542.0625 , 1.2909560364122604e-46 ], [ 542.125 , 1.29114919858396e-46 ], [ 542.1875 , 1.2913423896579613e-46 ], 
                    [ 542.25 , 1.2915356096386442e-46 ], [ 542.3125 , 1.2917288585302692e-46 ], [ 542.375 , 1.2919221363371993e-46 ], [ 542.4375 , 1.2921154430637332e-46 ], 
                    [ 542.5 , 1.2923087787142347e-46 ], [ 542.5625 , 1.2925021432930136e-46 ], [ 542.625 , 1.2926955368044074e-46 ], [ 542.6875 , 1.2928889592527176e-46 ], 
                    [ 542.75 , 1.2930824106423198e-46 ], [ 542.8125 , 1.2932758909774712e-46 ], [ 542.875 , 1.2934694002625761e-46 ], [ 542.9375 , 1.2936629385019204e-46 ], 
                    [ 543.0 , 1.2938565056998823e-46 ], [ 543.0625 , 1.2940501018607486e-46 ], [ 543.125 , 1.2942437269888717e-46 ], [ 543.1875 , 1.2944373810885953e-46 ], 
                    [ 543.25 , 1.2946309680268757e-46 ], [ 543.3125 , 1.294824567615467e-46 ], [ 543.375 , 1.2950181961549933e-46 ], [ 543.4375 , 1.2952118536498207e-46 ], 
                    [ 543.5 , 1.2954055401042515e-46 ], [ 543.5625 , 1.2955992555226255e-46 ], [ 543.625 , 1.295792999909274e-46 ], [ 543.6875 , 1.295986773268529e-46 ], 
                    [ 543.75 , 1.2961805756047413e-46 ], [ 543.8125 , 1.2963744069221984e-46 ], [ 543.875 , 1.2965682672252709e-46 ], [ 543.9375 , 1.2967621565182931e-46 ], 
                    [ 544.0 , 1.2969560748055912e-46 ], [ 544.0625 , 1.297150022091501e-46 ], [ 544.125 , 1.2973439983803587e-46 ], [ 544.1875 , 1.2975380036764832e-46 ], 
                    [ 544.25 , 1.2977320379842492e-46 ], [ 544.3125 , 1.297926101307949e-46 ], [ 544.375 , 1.2981201936519858e-46 ], [ 544.4375 , 1.2983142966888555e-46 ], 
                    [ 544.5 , 1.2985083290705681e-46 ], [ 544.5625 , 1.2987023904502984e-46 ], [ 544.625 , 1.2988964808324173e-46 ], [ 544.6875 , 1.2990906002212408e-46 ], 
                    [ 544.75 , 1.2992847486210946e-46 ], [ 544.8125 , 1.2994789260363142e-46 ], [ 544.875 , 1.299673132471264e-46 ], [ 544.9375 , 1.2998673679302529e-46 ], 
                    [ 545.0 , 1.3000616324176278e-46 ], [ 545.0625 , 1.3002559259377271e-46 ], [ 545.125 , 1.3004502484948899e-46 ], [ 545.1875 , 1.3006446000934647e-46 ], 
                    [ 545.25 , 1.300838980737783e-46 ], [ 545.3125 , 1.3010333904321664e-46 ], [ 545.375 , 1.3012278291809848e-46 ], [ 545.4375 , 1.3014222969885893e-46 ], 
                    [ 545.5 , 1.3016167938592672e-46 ], [ 545.5625 , 1.301811319797436e-46 ], [ 545.625 , 1.3020058748073755e-46 ], [ 545.6875 , 1.3022004126356117e-46 ], 
                    [ 545.75 , 1.3023949132399058e-46 ], [ 545.8125 , 1.3025894428954e-46 ], [ 545.875 , 1.3027840016064339e-46 ], [ 545.9375 , 1.302978589377338e-46 ], 
                    [ 546.0 , 1.3031732062124616e-46 ], [ 546.0625 , 1.3033678521161647e-46 ], [ 546.125 , 1.303562527092743e-46 ], [ 546.1875 , 1.303757231146594e-46 ], 
                    [ 546.25 , 1.3039519642820053e-46 ], [ 546.3125 , 1.304146726503367e-46 ], [ 546.375 , 1.3043415178150053e-46 ], [ 546.4375 , 1.3045363382212554e-46 ], 
                    [ 546.5 , 1.3047311877264725e-46 ], [ 546.5625 , 1.3049260663350118e-46 ], [ 546.625 , 1.3051209740512023e-46 ], [ 546.6875 , 1.30531591087941e-46 ], 
                    [ 546.75 , 1.3055108768239829e-46 ], [ 546.8125 , 1.305705871889242e-46 ], [ 546.875 , 1.3059008960795558e-46 ], [ 546.9375 , 1.306095875929799e-46 ], 
                    [ 547.0 , 1.3062908402213584e-46 ], [ 547.0625 , 1.3064858336157483e-46 ], [ 547.125 , 1.3066808561172854e-46 ], [ 547.1875 , 1.3068759077303329e-46 ], 
                    [ 547.25 , 1.307070988459246e-46 ], [ 547.3125 , 1.3072660983083611e-46 ], [ 547.375 , 1.3074612372820162e-46 ], [ 547.4375 , 1.3076564053845865e-46 ], 
                    [ 547.5 , 1.307851602620383e-46 ], [ 547.5625 , 1.308046828993773e-46 ], [ 547.625 , 1.3082420845091338e-46 ], [ 547.6875 , 1.3084373691707414e-46 ], 
                    [ 547.75 , 1.3086326829830206e-46 ], [ 547.8125 , 1.3088280259502953e-46 ], [ 547.875 , 1.3090233980769264e-46 ], [ 547.9375 , 1.3092187993672386e-46 ], 
                    [ 548.0 , 1.309414229825614e-46 ], [ 548.0625 , 1.3096096894564056e-46 ], [ 548.125 , 1.3098051782639405e-46 ], [ 548.1875 , 1.3100005999143435e-46 ], 
                    [ 548.25 , 1.3101960247743826e-46 ], [ 548.3125 , 1.31039147878773e-46 ], [ 548.375 , 1.3105869619587628e-46 ], [ 548.4375 , 1.3107824742918027e-46 ], 
                    [ 548.5 , 1.310978015791247e-46 ], [ 548.5625 , 1.311173586461409e-46 ], [ 548.625 , 1.3113691863066406e-46 ], [ 548.6875 , 1.3115648153313036e-46 ], 
                    [ 548.75 , 1.31176047353976e-46 ], [ 548.8125 , 1.311956160936354e-46 ], [ 548.875 , 1.312151877525431e-46 ], [ 548.9375 , 1.3123476233113548e-46 ], 
                    [ 549.0 , 1.3125433982984997e-46 ], [ 549.0625 , 1.3127392024911847e-46 ], [ 549.125 , 1.3129350358937851e-46 ], [ 549.1875 , 1.3131308985106588e-46 ], 
                    [ 549.25 , 1.3133267903461732e-46 ], [ 549.3125 , 1.3135227114046776e-46 ], [ 549.375 , 1.3137186616905223e-46 ], [ 549.4375 , 1.3139145293634658e-46 ], 
                    [ 549.5 , 1.3141104150352555e-46 ], [ 549.5625 , 1.3143063299107683e-46 ], [ 549.625 , 1.3145022739943584e-46 ], [ 549.6875 , 1.3146982472903802e-46 ], 
                    [ 549.75 , 1.3148942498031888e-46 ], [ 549.8125 , 1.3150902815371493e-46 ], [ 549.875 , 1.3152863424966092e-46 ], [ 549.9375 , 1.3154824326859156e-46 ], 
                    [ 550.0 , 1.3156785521094362e-46 ], [ 550.0625 , 1.3158747007715198e-46 ], [ 550.125 , 1.3160708786765627e-46 ], [ 550.1875 , 1.3162670858288687e-46 ], 
                    [ 550.25 , 1.3164633222328355e-46 ], [ 550.3125 , 1.316659587892824e-46 ], [ 550.375 , 1.3168558828131868e-46 ], [ 550.4375 , 1.3170522069982578e-46 ], 
                    [ 550.5 , 1.3172485604524655e-46 ], [ 550.5625 , 1.3174449431801082e-46 ], [ 550.625 , 1.3176413551855874e-46 ], [ 550.6875 , 1.3178376845899373e-46 ], 
                    [ 550.75 , 1.3180340415388048e-46 ], [ 550.8125 , 1.3182304277447379e-46 ], [ 550.875 , 1.3184268432120958e-46 ], [ 550.9375 , 1.31862328794522e-46 ], 
                    [ 551.0 , 1.318819761948471e-46 ], [ 551.0625 , 1.3190162652262288e-46 ], [ 551.125 , 1.319212797782855e-46 ], [ 551.1875 , 1.3194093596226843e-46 ], 
                    [ 551.25 , 1.319605950750108e-46 ], [ 551.3125 , 1.3198025711694805e-46 ], [ 551.375 , 1.319999220885157e-46 ], [ 551.4375 , 1.3201958999015396e-46 ], 
                    [ 551.5 , 1.3203926082229292e-46 ], [ 551.5625 , 1.3205893458537574e-46 ], [ 551.625 , 1.320786112798354e-46 ], [ 551.6875 , 1.3209829090610772e-46 ], 
                    [ 551.75 , 1.3211797346463331e-46 ], [ 551.8125 , 1.3213765895584905e-46 ], [ 551.875 , 1.32157347173825e-46 ], [ 551.9375 , 1.3217702536887485e-46 ], 
                    [ 552.0 , 1.3219670649400286e-46 ], [ 552.0625 , 1.3221639054964539e-46 ], [ 552.125 , 1.3223607753623964e-46 ], [ 552.1875 , 1.3225576745421927e-46 ], 
                    [ 552.25 , 1.3227546030402357e-46 ], [ 552.3125 , 1.322951560860891e-46 ], [ 552.375 , 1.3231485480085058e-46 ], [ 552.4375 , 1.3233455644874563e-46 ], 
                    [ 552.5 , 1.3235426103021098e-46 ], [ 552.5625 , 1.3237396854568345e-46 ], [ 552.625 , 1.3239367899560086e-46 ], [ 552.6875 , 1.3241339238039824e-46 ], 
                    [ 552.75 , 1.3243310870051358e-46 ], [ 552.8125 , 1.3245282795638486e-46 ], [ 552.875 , 1.3247255014844731e-46 ], [ 552.9375 , 1.3249227527713817e-46 ], 
                    [ 553.0 , 1.3251200334289562e-46 ], [ 553.0625 , 1.3253173434615888e-46 ], [ 553.125 , 1.3255146827152746e-46 ], [ 553.1875 , 1.3257119277841345e-46 ], 
                    [ 553.25 , 1.3259092022043329e-46 ], [ 553.3125 , 1.3261065059802273e-46 ], [ 553.375 , 1.3263038391161772e-46 ], [ 553.4375 , 1.3265012016165607e-46 ], 
                    [ 553.5 , 1.3266985934857475e-46 ], [ 553.5625 , 1.326896014728127e-46 ], [ 553.625 , 1.3270934653480224e-46 ], [ 553.6875 , 1.3272909453498528e-46 ], 
                    [ 553.75 , 1.327488454737962e-46 ], [ 553.8125 , 1.327685993516732e-46 ], [ 553.875 , 1.3278835616905465e-46 ], [ 553.9375 , 1.3280811592637602e-46 ], 
                    [ 554.0 , 1.328278786240748e-46 ], [ 554.0625 , 1.3284764426259048e-46 ], [ 554.125 , 1.3286741284236155e-46 ], [ 554.1875 , 1.3288718436382192e-46 ], 
                    [ 554.25 , 1.3290695882741123e-46 ], [ 554.3125 , 1.329267362335692e-46 ], [ 554.375 , 1.3294651658272893e-46 ], [ 554.4375 , 1.3296628859491537e-46 ], 
                    [ 554.5 , 1.3298606284584897e-46 ], [ 554.5625 , 1.3300584003753737e-46 ], [ 554.625 , 1.3302562017041601e-46 ], [ 554.6875 , 1.330454032449223e-46 ], 
                    [ 554.75 , 1.3306518926149182e-46 ], [ 554.8125 , 1.3308497822056682e-46 ], [ 554.875 , 1.331047701225802e-46 ], [ 554.9375 , 1.3312456496797243e-46 ], 
                    [ 555.0 , 1.3314436275717934e-46 ], [ 555.0625 , 1.3316416349063873e-46 ], [ 555.125 , 1.3318396716879132e-46 ], [ 555.1875 , 1.3320377379207408e-46 ], 
                    [ 555.25 , 1.3322358336092216e-46 ], [ 555.3125 , 1.3324339587577546e-46 ], [ 555.375 , 1.3326321133707312e-46 ], [ 555.4375 , 1.3328302974525042e-46 ], 
                    [ 555.5 , 1.3330285110074944e-46 ], [ 555.5625 , 1.3332267540400563e-46 ], [ 555.625 , 1.3334250265545927e-46 ], [ 555.6875 , 1.3336232143070303e-46 ], 
                    [ 555.75 , 1.3338214091528288e-46 ], [ 555.8125 , 1.334019633453121e-46 ], [ 555.875 , 1.3342178872122848e-46 ], [ 555.9375 , 1.3344161704346978e-46 ], 
                    [ 556.0 , 1.3346144831247483e-46 ], [ 556.0625 , 1.3348128252867964e-46 ], [ 556.125 , 1.3350111969252127e-46 ], [ 556.1875 , 1.3352095980444253e-46 ], 
                    [ 556.25 , 1.3354080286487776e-46 ], [ 556.3125 , 1.3356064887426513e-46 ], [ 556.375 , 1.3358049783304479e-46 ], [ 556.4375 , 1.336003497416541e-46 ], 
                    [ 556.5 , 1.336202046005315e-46 ], [ 556.5625 , 1.3364006241011634e-46 ], [ 556.625 , 1.3365992317084525e-46 ], [ 556.6875 , 1.3367978688315679e-46 ], 
                    [ 556.75 , 1.3369965354749158e-46 ], [ 556.8125 , 1.3371952316428634e-46 ], [ 556.875 , 1.3373939573398272e-46 ], [ 556.9375 , 1.3375926267688144e-46 ], 
                    [ 557.0 , 1.337791286328561e-46 ], [ 557.0625 , 1.3379899753932512e-46 ], [ 557.125 , 1.3381886939672863e-46 ], [ 557.1875 , 1.338387442055068e-46 ], 
                    [ 557.25 , 1.3385862196609417e-46 ], [ 557.3125 , 1.3387850267893293e-46 ], [ 557.375 , 1.3389838634445779e-46 ], [ 557.4375 , 1.339182729631101e-46 ], 
                    [ 557.5 , 1.339381625353275e-46 ], [ 557.5625 , 1.3395805506154867e-46 ], [ 557.625 , 1.3397795054221235e-46 ], [ 557.6875 , 1.3399784897775827e-46 ], 
                    [ 557.75 , 1.3401775036862242e-46 ], [ 557.8125 , 1.3403765471524566e-46 ], [ 557.875 , 1.3405756201806793e-46 ], [ 557.9375 , 1.340774722775273e-46 ], 
                    [ 558.0 , 1.3409738549406293e-46 ], [ 558.0625 , 1.3411730166811491e-46 ], [ 558.125 , 1.3413722080011874e-46 ], [ 558.1875 , 1.3415713647620664e-46 ], 
                    [ 558.25 , 1.341770484042634e-46 ], [ 558.3125 , 1.3419696328769782e-46 ], [ 558.375 , 1.3421688112694764e-46 ], [ 558.4375 , 1.342368019224496e-46 ], 
                    [ 558.5 , 1.342567256746473e-46 ], [ 558.5625 , 1.3427665238397573e-46 ], [ 558.625 , 1.3429658205087382e-46 ], [ 558.6875 , 1.343165146757824e-46 ], 
                    [ 558.75 , 1.3433645025913868e-46 ], [ 558.8125 , 1.3435638880138359e-46 ], [ 558.875 , 1.3437633030295346e-46 ], [ 558.9375 , 1.3439627476429132e-46 ], 
                    [ 559.0 , 1.3441622218583461e-46 ], [ 559.0625 , 1.3443617256802265e-46 ], [ 559.125 , 1.3445612591129391e-46 ], [ 559.1875 , 1.3447608221608885e-46 ], 
                    [ 559.25 , 1.3449604148284893e-46 ], [ 559.3125 , 1.3451600371200997e-46 ], [ 559.375 , 1.3453596890401447e-46 ], [ 559.4375 , 1.3455593385045972e-46 ], 
                    [ 559.5 , 1.3457589195262012e-46 ], [ 559.5625 , 1.3459585301507888e-46 ], [ 559.625 , 1.3461581703827609e-46 ], [ 559.6875 , 1.3463578402265087e-46 ], 
                    [ 559.75 , 1.3465575396864245e-46 ], [ 559.8125 , 1.346757268766901e-46 ], [ 559.875 , 1.3469570274723318e-46 ], [ 559.9375 , 1.347156815807111e-46 ], 
                    [ 560.0 , 1.347356633775643e-46 ], [ 560.0625 , 1.3475564813822855e-46 ], [ 560.125 , 1.3477563586314813e-46 ], [ 560.1875 , 1.3479562655276187e-46 ], 
                    [ 560.25 , 1.3481562020750847e-46 ], [ 560.3125 , 1.3483561682782682e-46 ], [ 560.375 , 1.3485561641415774e-46 ], [ 560.4375 , 1.3487561896694117e-46 ], 
                    [ 560.5 , 1.348956244866171e-46 ], [ 560.5625 , 1.3491563297362562e-46 ], [ 560.625 , 1.3493564442840586e-46 ], [ 560.6875 , 1.3495565885139997e-46 ], 
                    [ 560.75 , 1.3497566345496282e-46 ], [ 560.8125 , 1.3499567046866853e-46 ], [ 560.875 , 1.3501568044795029e-46 ], [ 560.9375 , 1.3503569339324769e-46 ], 
                    [ 561.0 , 1.350557093050004e-46 ], [ 561.0625 , 1.3507572818364811e-46 ], [ 561.125 , 1.3509575002962963e-46 ], [ 561.1875 , 1.351157748433867e-46 ], 
                    [ 561.25 , 1.3513580262535735e-46 ], [ 561.3125 , 1.3515583337598244e-46 ], [ 561.375 , 1.3517586709570497e-46 ], [ 561.4375 , 1.3519590378496018e-46 ], 
                    [ 561.5 , 1.3521594344418826e-46 ], [ 561.5625 , 1.3523598607383326e-46 ], [ 561.625 , 1.3525603167433164e-46 ], [ 561.6875 , 1.3527608024612954e-46 ], 
                    [ 561.75 , 1.3529613178966066e-46 ], [ 561.8125 , 1.353161863053693e-46 ], [ 561.875 , 1.353362437936951e-46 ], [ 561.9375 , 1.3535630425507866e-46 ], 
                    [ 562.0 , 1.35376359057808e-46 ], [ 562.0625 , 1.3539641163681193e-46 ], [ 562.125 , 1.3541646718609756e-46 ], [ 562.1875 , 1.3543652570610684e-46 ], 
                    [ 562.25 , 1.3545658719727687e-46 ], [ 562.3125 , 1.354766516600497e-46 ], [ 562.375 , 1.3549671909486452e-46 ], [ 562.4375 , 1.3551678950215969e-46 ], 
                    [ 562.5 , 1.3553686288237927e-46 ], [ 562.5625 , 1.3555693923596078e-46 ], [ 562.625 , 1.3557701856334564e-46 ], [ 562.6875 , 1.3559710086497526e-46 ], 
                    [ 562.75 , 1.3561718614128927e-46 ], [ 562.8125 , 1.3563727439272634e-46 ], [ 562.875 , 1.3565736561973103e-46 ], [ 562.9375 , 1.3567745982273927e-46 ], 
                    [ 563.0 , 1.3569755700219767e-46 ], [ 563.0625 , 1.3571765715854518e-46 ], [ 563.125 , 1.3573776029221793e-46 ], [ 563.1875 , 1.3575786640366273e-46 ], 
                    [ 563.25 , 1.3577797221494763e-46 ], [ 563.3125 , 1.3579807084483314e-46 ], [ 563.375 , 1.3581817244983249e-46 ], [ 563.4375 , 1.3583827703038608e-46 ], 
                    [ 563.5 , 1.3585838458693433e-46 ], [ 563.5625 , 1.3587849511991685e-46 ], [ 563.625 , 1.358986086297761e-46 ], [ 563.6875 , 1.3591872511695086e-46 ], 
                    [ 563.75 , 1.3593884458188567e-46 ], [ 563.8125 , 1.3595896702501557e-46 ], [ 563.875 , 1.3597909244678424e-46 ], [ 563.9375 , 1.3599922084763553e-46 ], 
                    [ 564.0 , 1.3601935222800366e-46 ], [ 564.0625 , 1.3603948658833547e-46 ], [ 564.125 , 1.3605962392907113e-46 ], [ 564.1875 , 1.3607976425064788e-46 ], 
                    [ 564.25 , 1.3609990755351283e-46 ], [ 564.3125 , 1.3612005383810243e-46 ], [ 564.375 , 1.3614020310485998e-46 ], [ 564.4375 , 1.3616035535422884e-46 ], 
                    [ 564.5 , 1.3618051058664762e-46 ], [ 564.5625 , 1.3620065718632504e-46 ], [ 564.625 , 1.3622080417462836e-46 ], [ 564.6875 , 1.362409541431038e-46 ], 
                    [ 564.75 , 1.3626110709218833e-46 ], [ 564.8125 , 1.3628126302232476e-46 ], [ 564.875 , 1.3630142193395407e-46 ], [ 564.9375 , 1.3632158382751824e-46 ], 
                    [ 565.0 , 1.3634174870345546e-46 ], [ 565.0625 , 1.363619165622098e-46 ], [ 565.125 , 1.3638208740422152e-46 ], [ 565.1875 , 1.3640226122993382e-46 ], 
                    [ 565.25 , 1.3642243803978421e-46 ], [ 565.3125 , 1.3644261783421506e-46 ], [ 565.375 , 1.3646280061367076e-46 ], [ 565.4375 , 1.3648298637858898e-46 ], 
                    [ 565.5 , 1.365031751294152e-46 ], [ 565.5625 , 1.365233668665892e-46 ], [ 565.625 , 1.3654356159055074e-46 ], [ 565.6875 , 1.3656375930174455e-46 ], 
                    [ 565.75 , 1.3658396000061157e-46 ], [ 565.8125 , 1.3660415910077913e-46 ], [ 565.875 , 1.3662435247084493e-46 ], [ 565.9375 , 1.3664454882597304e-46 ], 
                    [ 566.0 , 1.3666474816660861e-46 ], [ 566.0625 , 1.3668495049318911e-46 ], [ 566.125 , 1.3670515580615886e-46 ], [ 566.1875 , 1.3672536410595833e-46 ], 
                    [ 566.25 , 1.3674557539302998e-46 ], [ 566.3125 , 1.3676578966781353e-46 ], [ 566.375 , 1.3678600693075058e-46 ], [ 566.4375 , 1.3680622718228482e-46 ], 
                    [ 566.5 , 1.3682645042285705e-46 ], [ 566.5625 , 1.3684667665290913e-46 ], [ 566.625 , 1.3686690587288395e-46 ], [ 566.6875 , 1.3688713808322154e-46 ], 
                    [ 566.75 , 1.3690737328436492e-46 ], [ 566.8125 , 1.3692761147675524e-46 ], [ 566.875 , 1.3694785266083662e-46 ], [ 566.9375 , 1.3696809683705033e-46 ], 
                    [ 567.0 , 1.3698834400583868e-46 ], [ 567.0625 , 1.3700859416764403e-46 ], [ 567.125 , 1.3702883550968055e-46 ], [ 567.1875 , 1.370490773490985e-46 ], 
                    [ 567.25 , 1.3706932217863402e-46 ], [ 567.3125 , 1.3708956999872688e-46 ], [ 567.375 , 1.3710982080981786e-46 ], [ 567.4375 , 1.3713007461235074e-46 ], 
                    [ 567.5 , 1.371503314067635e-46 ], [ 567.5625 , 1.37170591193504e-46 ], [ 567.625 , 1.3719085397301128e-46 ], [ 567.6875 , 1.3721111974572744e-46 ], 
                    [ 567.75 , 1.3723138851209466e-46 ], [ 567.8125 , 1.3725166027255515e-46 ], [ 567.875 , 1.3727193502754922e-46 ], [ 567.9375 , 1.372922127775222e-46 ], 
                    [ 568.0 , 1.3731249352291744e-46 ], [ 568.0625 , 1.3733277726417357e-46 ], [ 568.125 , 1.37353064001735e-46 ], [ 568.1875 , 1.3737335373604443e-46 ], 
                    [ 568.25 , 1.3739364646754448e-46 ], [ 568.3125 , 1.3741394219667892e-46 ], [ 568.375 , 1.3743423728086998e-46 ], [ 568.4375 , 1.374545250009977e-46 ], 
                    [ 568.5 , 1.3747481571595112e-46 ], [ 568.5625 , 1.3749510942617235e-46 ], [ 568.625 , 1.3751540613210354e-46 ], [ 568.6875 , 1.375357058341879e-46 ], 
                    [ 568.75 , 1.375560085328657e-46 ], [ 568.8125 , 1.3757631422858036e-46 ], [ 568.875 , 1.3759662292177325e-46 ], [ 568.9375 , 1.3761693461288884e-46 ], 
                    [ 569.0 , 1.3763724930236868e-46 ], [ 569.0625 , 1.3765756699065541e-46 ], [ 569.125 , 1.3767788767819167e-46 ], [ 569.1875 , 1.3769821136542117e-46 ], 
                    [ 569.25 , 1.3771853805278482e-46 ], [ 569.3125 , 1.3773886774072544e-46 ], [ 569.375 , 1.3775920042968697e-46 ], [ 569.4375 , 1.3777953612011338e-46 ], 
                    [ 569.5 , 1.3779987481244776e-46 ], [ 569.5625 , 1.378202165071303e-46 ], [ 569.625 , 1.378405612046081e-46 ], [ 569.6875 , 1.3786089931517108e-46 ], 
                    [ 569.75 , 1.378812352126926e-46 ], [ 569.8125 , 1.3790157410996866e-46 ], [ 569.875 , 1.3792191600744085e-46 ], [ 569.9375 , 1.3794226090555067e-46 ], 
                    [ 570.0 , 1.3796260880474372e-46 ], [ 570.0625 , 1.3798295970545978e-46 ], [ 570.125 , 1.3800331360814253e-46 ], [ 570.1875 , 1.3802367051323384e-46 ], 
                    [ 570.25 , 1.3804403042117956e-46 ], [ 570.3125 , 1.3806439333241967e-46 ], [ 570.375 , 1.3808475924739817e-46 ], [ 570.4375 , 1.3810512816655617e-46 ], 
                    [ 570.5 , 1.3812550009034175e-46 ], [ 570.5625 , 1.3814587501919516e-46 ], [ 570.625 , 1.3816625295355773e-46 ], [ 570.6875 , 1.3818663389387671e-46 ], 
                    [ 570.75 , 1.382070178405926e-46 ], [ 570.8125 , 1.3822740479414882e-46 ], [ 570.875 , 1.3824779475499289e-46 ], [ 570.9375 , 1.382681877235625e-46 ], 
                    [ 571.0 , 1.3828856889619293e-46 ], [ 571.0625 , 1.3830895290536411e-46 ], [ 571.125 , 1.3832933991917963e-46 ], [ 571.1875 , 1.383497299380804e-46 ], 
                    [ 571.25 , 1.383701229625104e-46 ], [ 571.3125 , 1.383905189929136e-46 ], [ 571.375 , 1.3841091802973111e-46 ], [ 571.4375 , 1.3843132007340906e-46 ], 
                    [ 571.5 , 1.3845172512438675e-46 ], [ 571.5625 , 1.3847213318310934e-46 ], [ 571.625 , 1.3849254425002127e-46 ], [ 571.6875 , 1.385129583255649e-46 ], 
                    [ 571.75 , 1.3853337541018178e-46 ], [ 571.8125 , 1.3855379550432033e-46 ], [ 571.875 , 1.3857421860841928e-46 ], [ 571.9375 , 1.3859464472292426e-46 ], 
                    [ 572.0 , 1.3861507384827902e-46 ], [ 572.0625 , 1.3863550598492929e-46 ], [ 572.125 , 1.3865594113331507e-46 ], [ 572.1875 , 1.386763792938822e-46 ], 
                    [ 572.25 , 1.3869681659309425e-46 ], [ 572.3125 , 1.3871724708971843e-46 ], [ 572.375 , 1.3873768059582188e-46 ], [ 572.4375 , 1.3875811711184985e-46 ], 
                    [ 572.5 , 1.387785566382408e-46 ], [ 572.5625 , 1.387989991754441e-46 ], [ 572.625 , 1.3881944472389827e-46 ], [ 572.6875 , 1.388398932840489e-46 ], 
                    [ 572.75 , 1.3886034485634157e-46 ], [ 572.8125 , 1.3888079944121505e-46 ], [ 572.875 , 1.3890125703911802e-46 ], [ 572.9375 , 1.3892171765049238e-46 ], 
                    [ 573.0 , 1.3894218127578297e-46 ], [ 573.0625 , 1.389626479154318e-46 ], [ 573.125 , 1.3898311756988088e-46 ], [ 573.1875 , 1.3900359023958128e-46 ], 
                    [ 573.25 , 1.3902406592497216e-46 ], [ 573.3125 , 1.3904454462649777e-46 ], [ 573.375 , 1.3906502634460435e-46 ], [ 573.4375 , 1.390855110797343e-46 ], 
                    [ 573.5 , 1.3910599883233498e-46 ], [ 573.5625 , 1.3912648131579118e-46 ], [ 573.625 , 1.3914695947226457e-46 ], [ 573.6875 , 1.391674406429349e-46 ], 
                    [ 573.75 , 1.3918792482824986e-46 ], [ 573.8125 , 1.3920841202865012e-46 ], [ 573.875 , 1.3922890224457854e-46 ], [ 573.9375 , 1.3924939547648292e-46 ], 
                    [ 574.0 , 1.3926989172480719e-46 ], [ 574.0625 , 1.3929039098999038e-46 ], [ 574.125 , 1.393108932724805e-46 ], [ 574.1875 , 1.3933139857272071e-46 ], 
                    [ 574.25 , 1.3935190689115518e-46 ], [ 574.3125 , 1.3937241822822814e-46 ], [ 574.375 , 1.3939293258438196e-46 ], [ 574.4375 , 1.3941344996006596e-46 ], 
                    [ 574.5 , 1.394339703557206e-46 ], [ 574.5625 , 1.394544937717924e-46 ], [ 574.625 , 1.3947502020872297e-46 ], [ 574.6875 , 1.394955496669609e-46 ], 
                    [ 574.75 , 1.3951608214694695e-46 ], [ 574.8125 , 1.3953661764912885e-46 ], [ 574.875 , 1.3955714584158586e-46 ], [ 574.9375 , 1.3957767323971987e-46 ], 
                    [ 575.0 , 1.3959820365721869e-46 ], [ 575.0625 , 1.3961873709452742e-46 ], [ 575.125 , 1.396392735520883e-46 ], [ 575.1875 , 1.396598130303485e-46 ], 
                    [ 575.25 , 1.3968035552975335e-46 ], [ 575.3125 , 1.3970090105074427e-46 ], [ 575.375 , 1.397214495937657e-46 ], [ 575.4375 , 1.3974200115926507e-46 ], 
                    [ 575.5 , 1.3976255574768107e-46 ], [ 575.5625 , 1.3978311335946725e-46 ], [ 575.625 , 1.3980367399505935e-46 ], [ 575.6875 , 1.3982423765490915e-46 ], 
                    [ 575.75 , 1.3984480433945846e-46 ], [ 575.8125 , 1.3986537404914921e-46 ], [ 575.875 , 1.3988594678442935e-46 ], [ 575.9375 , 1.3990652254574892e-46 ], 
                    [ 576.0 , 1.3992710133354301e-46 ], [ 576.0625 , 1.3994768314826378e-46 ], [ 576.125 , 1.3996826799035347e-46 ], [ 576.1875 , 1.3998884202450395e-46 ], 
                    [ 576.25 , 1.400094172585229e-46 ], [ 576.3125 , 1.4002999551664185e-46 ], [ 576.375 , 1.4005057679930527e-46 ], [ 576.4375 , 1.4007116110695767e-46 ], 
                    [ 576.5 , 1.4009174844004368e-46 ], [ 576.5625 , 1.4011233879900896e-46 ], [ 576.625 , 1.4013293218429727e-46 ], [ 576.6875 , 1.401535285963514e-46 ], 
                    [ 576.75 , 1.401741280356182e-46 ], [ 576.8125 , 1.401947305025436e-46 ], [ 576.875 , 1.4021533599757162e-46 ], [ 576.9375 , 1.4023594452114834e-46 ], 
                    [ 577.0 , 1.4025655607371684e-46 ], [ 577.0625 , 1.4027717065572235e-46 ], [ 577.125 , 1.4029778826761313e-46 ], [ 577.1875 , 1.4031840890983049e-46 ], 
                    [ 577.25 , 1.4033903258282281e-46 ], [ 577.3125 , 1.4035965928703458e-46 ], [ 577.375 , 1.403802890229133e-46 ], [ 577.4375 , 1.4040092179089954e-46 ], 
                    [ 577.5 , 1.404215420887729e-46 ], [ 577.5625 , 1.4044216524708007e-46 ], [ 577.625 , 1.4046279143422922e-46 ], [ 577.6875 , 1.4048342065066515e-46 ], 
                    [ 577.75 , 1.4050405289683177e-46 ], [ 577.8125 , 1.4052468817317505e-46 ], [ 577.875 , 1.4054532648014005e-46 ], [ 577.9375 , 1.405659678181728e-46 ], 
                    [ 578.0 , 1.405866121877175e-46 ], [ 578.0625 , 1.4060725958921738e-46 ], [ 578.125 , 1.406279100231217e-46 ], [ 578.1875 , 1.4064856348987383e-46 ], 
                    [ 578.25 , 1.406692199899172e-46 ], [ 578.3125 , 1.4068987952370132e-46 ], [ 578.375 , 1.4071054209166971e-46 ], [ 578.4375 , 1.4073120769426802e-46 ], 
                    [ 578.5 , 1.4075187633194088e-46 ], [ 578.5625 , 1.407725480051341e-46 ], [ 578.625 , 1.4079322271429646e-46 ], [ 578.6875 , 1.4081390045987085e-46 ], 
                    [ 578.75 , 1.4083458054077256e-46 ], [ 578.8125 , 1.4085524969412646e-46 ], [ 578.875 , 1.4087592188092295e-46 ], [ 578.9375 , 1.4089659710161226e-46 ], 
                    [ 579.0 , 1.4091727535663364e-46 ], [ 579.0625 , 1.4093795664643644e-46 ], [ 579.125 , 1.40958640971464e-46 ], [ 579.1875 , 1.409793283321658e-46 ], 
                    [ 579.25 , 1.4100001872898138e-46 ], [ 579.3125 , 1.410207121623613e-46 ], [ 579.375 , 1.4104140863274822e-46 ], [ 579.4375 , 1.4106210814058988e-46 ], 
                    [ 579.5 , 1.41082810686329e-46 ], [ 579.5625 , 1.4110351627041449e-46 ], [ 579.625 , 1.4112422489329224e-46 ], [ 579.6875 , 1.4114493655540524e-46 ], 
                    [ 579.75 , 1.4116565125720452e-46 ], [ 579.8125 , 1.4118636899913118e-46 ], [ 579.875 , 1.4120708978163342e-46 ], [ 579.9375 , 1.4122781360515648e-46 ], 
                    [ 580.0 , 1.4124854047014965e-46 ], [ 580.0625 , 1.4126926948747766e-46 ], [ 580.125 , 1.4128998728490278e-46 ], [ 580.1875 , 1.4131070812068754e-46 ], 
                    [ 580.25 , 1.4133143199528153e-46 ], [ 580.3125 , 1.4135215890912742e-46 ], [ 580.375 , 1.413728888626729e-46 ], [ 580.4375 , 1.413936218563608e-46 ], 
                    [ 580.5 , 1.4141435789063995e-46 ], [ 580.5625 , 1.4143509696595426e-46 ], [ 580.625 , 1.414558390827517e-46 ], [ 580.6875 , 1.4147658424147738e-46 ], 
                    [ 580.75 , 1.4149733244257833e-46 ], [ 580.8125 , 1.415180836864998e-46 ], [ 580.875 , 1.4153883797368595e-46 ], [ 580.9375 , 1.4155959530458718e-46 ], 
                    [ 581.0 , 1.415803556796458e-46 ], [ 581.0625 , 1.4160111909931027e-46 ], [ 581.125 , 1.4162188556403014e-46 ], [ 581.1875 , 1.416426550742469e-46 ], 
                    [ 581.25 , 1.4166342763040927e-46 ], [ 581.3125 , 1.4168420323296386e-46 ], [ 581.375 , 1.4170498150839127e-46 ], [ 581.4375 , 1.4172574638750996e-46 ], 
                    [ 581.5 , 1.417465143094327e-46 ], [ 581.5625 , 1.4176728527460334e-46 ], [ 581.625 , 1.4178805928346783e-46 ], [ 581.6875 , 1.418088363364732e-46 ], 
                    [ 581.75 , 1.418296164340625e-46 ], [ 581.8125 , 1.4185039957668586e-46 ], [ 581.875 , 1.4187118576478552e-46 ], [ 581.9375 , 1.4189197499880973e-46 ], 
                    [ 582.0 , 1.4191276727920686e-46 ], [ 582.0625 , 1.4193356260642029e-46 ], [ 582.125 , 1.4195436098089645e-46 ], [ 582.1875 , 1.4197516240308193e-46 ], 
                    [ 582.25 , 1.4199596687342834e-46 ], [ 582.3125 , 1.4201677439237627e-46 ], [ 582.375 , 1.4203758496037655e-46 ], [ 582.4375 , 1.4205839857786986e-46 ], 
                    [ 582.5 , 1.4207921524530815e-46 ], [ 582.5625 , 1.4210003496313933e-46 ], [ 582.625 , 1.4212085773180736e-46 ], [ 582.6875 , 1.4214168355176033e-46 ], 
                    [ 582.75 , 1.4216249818274558e-46 ], [ 582.8125 , 1.4218331491902069e-46 ], [ 582.875 , 1.4220413470347518e-46 ], [ 582.9375 , 1.4222495753655035e-46 ], 
                    [ 583.0 , 1.4224578341869967e-46 ], [ 583.0625 , 1.422666123503636e-46 ], [ 583.125 , 1.422874443319916e-46 ], [ 583.1875 , 1.4230827936402736e-46 ], 
                    [ 583.25 , 1.4232911744692048e-46 ], [ 583.3125 , 1.4234995858111779e-46 ], [ 583.375 , 1.4237080276706401e-46 ], [ 583.4375 , 1.4239165000520701e-46 ], 
                    [ 583.5 , 1.4241250029599375e-46 ], [ 583.5625 , 1.4243335363987123e-46 ], [ 583.625 , 1.4245421003728444e-46 ], [ 583.6875 , 1.4247506948868664e-46 ], 
                    [ 583.75 , 1.4249593199451991e-46 ], [ 583.8125 , 1.4251679755523156e-46 ], [ 583.875 , 1.4253766617126987e-46 ], [ 583.9375 , 1.4255853784308227e-46 ], 
                    [ 584.0 , 1.4257941257111721e-46 ], [ 584.0625 , 1.426002772221777e-46 ], [ 584.125 , 1.4262114176323304e-46 ], [ 584.1875 , 1.42642009357081e-46 ], 
                    [ 584.25 , 1.4266288000416826e-46 ], [ 584.3125 , 1.4268375370494154e-46 ], [ 584.375 , 1.4270463045984661e-46 ], [ 584.4375 , 1.4272551026933242e-46 ], 
                    [ 584.5 , 1.427463931338448e-46 ], [ 584.5625 , 1.4276727905382979e-46 ], [ 584.625 , 1.427881680297385e-46 ], [ 584.6875 , 1.4280906006201197e-46 ], 
                    [ 584.75 , 1.4282995515110041e-46 ], [ 584.8125 , 1.428508532974501e-46 ], [ 584.875 , 1.4287175450151342e-46 ], [ 584.9375 , 1.4289265876373063e-46 ], 
                    [ 585.0 , 1.4291356608455228e-46 ], [ 585.0625 , 1.429344764644289e-46 ], [ 585.125 , 1.42955389903802e-46 ], [ 585.1875 , 1.4297630640312226e-46 ], 
                    [ 585.25 , 1.4299722596283843e-46 ], [ 585.3125 , 1.4301814858339727e-46 ], [ 585.375 , 1.4303906441882806e-46 ], [ 585.4375 , 1.4305997742756119e-46 ], 
                    [ 585.5 , 1.4308089349387815e-46 ], [ 585.5625 , 1.4310181261822704e-46 ], [ 585.625 , 1.431227348010529e-46 ], [ 585.6875 , 1.4314366004280493e-46 ], 
                    [ 585.75 , 1.4316458834392934e-46 ], [ 585.8125 , 1.431855197048724e-46 ], [ 585.875 , 1.4320645412608253e-46 ], [ 585.9375 , 1.432273916080071e-46 ], 
                    [ 586.0 , 1.4324833215109465e-46 ], [ 586.0625 , 1.4326927575579173e-46 ], [ 586.125 , 1.432902224225449e-46 ], [ 586.1875 , 1.4331117215180292e-46 ], 
                    [ 586.25 , 1.4333212494401351e-46 ], [ 586.3125 , 1.4335308079962652e-46 ], [ 586.375 , 1.4337403971908477e-46 ], [ 586.4375 , 1.4339500170284025e-46 ], 
                    [ 586.5 , 1.4341596675134102e-46 ], [ 586.5625 , 1.4343693486503407e-46 ], [ 586.625 , 1.434579060443666e-46 ], [ 586.6875 , 1.434788736682449e-46 ], 
                    [ 586.75 , 1.4349983433641879e-46 ], [ 586.8125 , 1.4352079806671223e-46 ], [ 586.875 , 1.435417648595736e-46 ], [ 586.9375 , 1.435627347154513e-46 ], 
                    [ 587.0 , 1.4358370763478977e-46 ], [ 587.0625 , 1.436046836180406e-46 ], [ 587.125 , 1.4362566266564831e-46 ], [ 587.1875 , 1.4364664477806268e-46 ], 
                    [ 587.25 , 1.4366762995572833e-46 ], [ 587.3125 , 1.4368861819909715e-46 ], [ 587.375 , 1.4370960950861394e-46 ], [ 587.4375 , 1.437306038847297e-46 ], 
                    [ 587.5 , 1.4375160132788734e-46 ], [ 587.5625 , 1.4377260183853795e-46 ], [ 587.625 , 1.437936054171307e-46 ], [ 587.6875 , 1.438146120641148e-46 ], 
                    [ 587.75 , 1.4383562177993544e-46 ], [ 587.8125 , 1.4385663456504096e-46 ], [ 587.875 , 1.4387765041987868e-46 ], [ 587.9375 , 1.4389866934490424e-46 ], 
                    [ 588.0 , 1.4391968932841664e-46 ], [ 588.0625 , 1.4394069788867412e-46 ], [ 588.125 , 1.4396170951563817e-46 ], [ 588.1875 , 1.4398272420975748e-46 ], 
                    [ 588.25 , 1.440037419714808e-46 ], [ 588.3125 , 1.4402476280125489e-46 ], [ 588.375 , 1.4404578669952655e-46 ], [ 588.4375 , 1.440668136667448e-46 ], 
                    [ 588.5 , 1.440878437033586e-46 ], [ 588.5625 , 1.4410887680981496e-46 ], [ 588.625 , 1.4412991298656205e-46 ], [ 588.6875 , 1.4415095223404598e-46 ], 
                    [ 588.75 , 1.4417199455271907e-46 ], [ 588.8125 , 1.4419303994302662e-46 ], [ 588.875 , 1.44214088405419e-46 ], [ 588.9375 , 1.4423513994034367e-46 ], 
                    [ 589.0 , 1.442561945482481e-46 ], [ 589.0625 , 1.4427725222958392e-46 ], [ 589.125 , 1.4429831298479571e-46 ], [ 589.1875 , 1.4431937681433422e-46 ], 
                    [ 589.25 , 1.4434044371864922e-46 ], [ 589.3125 , 1.4436151369818545e-46 ], [ 589.375 , 1.4438257394621605e-46 ], [ 589.4375 , 1.444036340161891e-46 ], 
                    [ 589.5 , 1.4442469715804535e-46 ], [ 589.5625 , 1.4444576337223382e-46 ], [ 589.625 , 1.4446683265920376e-46 ], [ 589.6875 , 1.4448790501940128e-46 ], 
                    [ 589.75 , 1.4450898045327568e-46 ], [ 589.8125 , 1.4453005896127633e-46 ], [ 589.875 , 1.4455114054384856e-46 ], [ 589.9375 , 1.4457222520144594e-46 ], 
                    [ 590.0 , 1.4459331293451293e-46 ], [ 590.0625 , 1.446144037434971e-46 ], [ 590.125 , 1.4463549762885015e-46 ], [ 590.1875 , 1.4465659459102084e-46 ], 
                    [ 590.25 , 1.4467769463045693e-46 ], [ 590.3125 , 1.4469879774760523e-46 ], [ 590.375 , 1.4471990394291876e-46 ], [ 590.4375 , 1.4474101321684346e-46 ], 
                    [ 590.5 , 1.4476212556983042e-46 ], [ 590.5625 , 1.4478324100232566e-46 ], [ 590.625 , 1.4480435951478457e-46 ], [ 590.6875 , 1.4482547361533561e-46 ], 
                    [ 590.75 , 1.4484658065476449e-46 ], [ 590.8125 , 1.4486769077035881e-46 ], [ 590.875 , 1.448888039625679e-46 ], [ 590.9375 , 1.4490992023183805e-46 ], 
                    [ 591.0 , 1.4493103957862088e-46 ], [ 591.0625 , 1.4495216200336075e-46 ], [ 591.125 , 1.4497328750650935e-46 ], [ 591.1875 , 1.4499441608851428e-46 ], 
                    [ 591.25 , 1.4501554774982323e-46 ], [ 591.3125 , 1.45036682490886e-46 ], [ 591.375 , 1.4505782031215458e-46 ], [ 591.4375 , 1.4507896121407063e-46 ], 
                    [ 591.5 , 1.4510010519708934e-46 ], [ 591.5625 , 1.4512125226165662e-46 ], [ 591.625 , 1.4514240240822262e-46 ], [ 591.6875 , 1.451635556372386e-46 ], 
                    [ 591.75 , 1.4518471194914968e-46 ], [ 591.8125 , 1.4520587134440923e-46 ], [ 591.875 , 1.452270338234615e-46 ], [ 591.9375 , 1.452481993867601e-46 ], 
                    [ 592.0 , 1.4526936797312176e-46 ], [ 592.0625 , 1.4529052365325866e-46 ], [ 592.125 , 1.4531168241431523e-46 ], [ 592.1875 , 1.45332844256736e-46 ], 
                    [ 592.25 , 1.4535400918097076e-46 ], [ 592.3125 , 1.453751771874683e-46 ], [ 592.375 , 1.4539634827667853e-46 ], [ 592.4375 , 1.4541752244904732e-46 ], 
                    [ 592.5 , 1.4543869970502672e-46 ], [ 592.5625 , 1.454598800450648e-46 ], [ 592.625 , 1.4548106346961171e-46 ], [ 592.6875 , 1.4550224997911563e-46 ], 
                    [ 592.75 , 1.4552343957402374e-46 ], [ 592.8125 , 1.4554463225478745e-46 ], [ 592.875 , 1.4556582802185822e-46 ], [ 592.9375 , 1.4558702687568242e-46 ], 
                    [ 593.0 , 1.4560822881670954e-46 ], [ 593.0625 , 1.4562943384539128e-46 ], [ 593.125 , 1.4565064196217832e-46 ], [ 593.1875 , 1.456718531675152e-46 ], 
                    [ 593.25 , 1.4569306746185586e-46 ], [ 593.3125 , 1.4571428484565116e-46 ], [ 593.375 , 1.457354962095962e-46 ], [ 593.4375 , 1.4575670271804533e-46 ], 
                    [ 593.5 , 1.4577791231233156e-46 ], [ 593.5625 , 1.4579912499290287e-46 ], [ 593.625 , 1.458203407602115e-46 ], [ 593.6875 , 1.4584155961470347e-46 ], 
                    [ 593.75 , 1.4586278155682802e-46 ], [ 593.8125 , 1.4588400658703547e-46 ], [ 593.875 , 1.4590523470577728e-46 ], [ 593.9375 , 1.4592646591349868e-46 ], 
                    [ 594.0 , 1.4594770021065337e-46 ], [ 594.0625 , 1.4596893759768667e-46 ], [ 594.125 , 1.4599017807505037e-46 ], [ 594.1875 , 1.4601142164319719e-46 ], 
                    [ 594.25 , 1.460326683025707e-46 ], [ 594.3125 , 1.4605391805362279e-46 ], [ 594.375 , 1.460751708968085e-46 ], [ 594.4375 , 1.4609642683256952e-46 ], 
                    [ 594.5 , 1.4611768586136307e-46 ], [ 594.5625 , 1.4613894798363304e-46 ], [ 594.625 , 1.4616021319983476e-46 ], [ 594.6875 , 1.461814812382952e-46 ], 
                    [ 594.75 , 1.4620273546178545e-46 ], [ 594.8125 , 1.4622399277555676e-46 ], [ 594.875 , 1.4624525318006053e-46 ], [ 594.9375 , 1.4626651667574408e-46 ], 
                    [ 595.0 , 1.462877832630589e-46 ], [ 595.0625 , 1.4630905294245248e-46 ], [ 595.125 , 1.4633032571437536e-46 ], [ 595.1875 , 1.4635160157927825e-46 ], 
                    [ 595.25 , 1.4637288053760777e-46 ], [ 595.3125 , 1.4639416258981785e-46 ], [ 595.375 , 1.4641544773635514e-46 ], [ 595.4375 , 1.4643673597767172e-46 ], 
                    [ 595.5 , 1.4645802731421642e-46 ], [ 595.5625 , 1.4647932174643833e-46 ], [ 595.625 , 1.4650061927478852e-46 ], [ 595.6875 , 1.4652191989971822e-46 ], 
                    [ 595.75 , 1.4654322362167453e-46 ], [ 595.8125 , 1.4656453044111193e-46 ], [ 595.875 , 1.4658584035847659e-46 ], [ 595.9375 , 1.4660715337422418e-46 ], 
                    [ 596.0 , 1.4662846948879892e-46 ], [ 596.0625 , 1.4664978054579976e-46 ], [ 596.125 , 1.4667108468074975e-46 ], [ 596.1875 , 1.4669239191059602e-46 ], 
                    [ 596.25 , 1.4671370223579133e-46 ], [ 596.3125 , 1.4673501565678321e-46 ], [ 596.375 , 1.4675633217402456e-46 ], [ 596.4375 , 1.4677765178796097e-46 ], 
                    [ 596.5 , 1.4679897449904546e-46 ], [ 596.5625 , 1.4682030030772586e-46 ], [ 596.625 , 1.4684162921445428e-46 ], [ 596.6875 , 1.468629612196797e-46 ], 
                    [ 596.75 , 1.4688429632385227e-46 ], [ 596.8125 , 1.4690563452742112e-46 ], [ 596.875 , 1.4692697583083755e-46 ], [ 596.9375 , 1.4694832023455294e-46 ], 
                    [ 597.0 , 1.4696966773901663e-46 ], [ 597.0625 , 1.4699101834468011e-46 ], [ 597.125 , 1.4701237205199075e-46 ], [ 597.1875 , 1.4703372886140232e-46 ], 
                    [ 597.25 , 1.4705508877336227e-46 ], [ 597.3125 , 1.4707645178832763e-46 ], [ 597.375 , 1.4709781790674081e-46 ], [ 597.4375 , 1.4711917372324655e-46 ], 
                    [ 597.5 , 1.4714052994835609e-46 ], [ 597.5625 , 1.4716188927359522e-46 ], [ 597.625 , 1.471832516994119e-46 ], [ 597.6875 , 1.4720461722625723e-46 ], 
                    [ 597.75 , 1.4722598585458242e-46 ], [ 597.8125 , 1.4724735758483561e-46 ], [ 597.875 , 1.4726873241746812e-46 ], [ 597.9375 , 1.4729011035293029e-46 ], 
                    [ 598.0 , 1.4731149139167257e-46 ], [ 598.0625 , 1.473328755341454e-46 ], [ 598.125 , 1.4735426278080038e-46 ], [ 598.1875 , 1.473756531320871e-46 ], 
                    [ 598.25 , 1.4739704658845413e-46 ], [ 598.3125 , 1.4741844315035637e-46 ], [ 598.375 , 1.474398428182426e-46 ], [ 598.4375 , 1.4746124559256258e-46 ], 
                    [ 598.5 , 1.4748265147376833e-46 ], [ 598.5625 , 1.4750406046231188e-46 ], [ 598.625 , 1.4752547255864224e-46 ], [ 598.6875 , 1.4754688776321264e-46 ], 
                    [ 598.75 , 1.4756830294100596e-46 ], [ 598.8125 , 1.4758970567896323e-46 ], [ 598.875 , 1.476111115210932e-46 ], [ 598.9375 , 1.4763252046784196e-46 ], 
                    [ 599.0 , 1.4765393251966392e-46 ], [ 599.0625 , 1.476753476770063e-46 ], [ 599.125 , 1.476967659403206e-46 ], [ 599.1875 , 1.4771818731005828e-46 ], 
                    [ 599.25 , 1.477396117866647e-46 ], [ 599.3125 , 1.4776103937059988e-46 ], [ 599.375 , 1.4778247006230611e-46 ], [ 599.4375 , 1.4780390386223832e-46 ], 
                    [ 599.5 , 1.478253407708463e-46 ], [ 599.5625 , 1.4784678078858086e-46 ], [ 599.625 , 1.4786822391589297e-46 ], [ 599.6875 , 1.4788967015323257e-46 ], 
                    [ 599.75 , 1.4791111950105286e-46 ], [ 599.8125 , 1.479325719598039e-46 ], [ 599.875 , 1.479540275299379e-46 ], [ 599.9375 , 1.4797548621190512e-46 ], 
                    [ 600.0 , 1.4799694800615372e-46 ], [ 600.0625 , 1.4801841291314143e-46 ], [ 600.125 , 1.4803987288918419e-46 ], [ 600.1875 , 1.4806132629153945e-46 ], 
                    [ 600.25 , 1.4808278280284514e-46 ], [ 600.3125 , 1.4810424242354973e-46 ], [ 600.375 , 1.4812570515410589e-46 ], [ 600.4375 , 1.4814717099496114e-46 ], 
                    [ 600.5 , 1.481686399465694e-46 ], [ 600.5625 , 1.4819011200938142e-46 ], [ 600.625 , 1.48211587183846e-46 ], [ 600.6875 , 1.4823306547041612e-46 ], 
                    [ 600.75 , 1.4825454686954075e-46 ], [ 600.8125 , 1.48276031381673e-46 ], [ 600.875 , 1.4829751900726402e-46 ], [ 600.9375 , 1.4831900974676077e-46 ], 
                    [ 601.0 , 1.4834050360061871e-46 ], [ 601.0625 , 1.4836200056929026e-46 ], [ 601.125 , 1.4838350065322256e-46 ], [ 601.1875 , 1.4840500385286917e-46 ], 
                    [ 601.25 , 1.4842651016868162e-46 ], [ 601.3125 , 1.4844801960111464e-46 ], [ 601.375 , 1.4846953215061147e-46 ], [ 601.4375 , 1.4849104781763222e-46 ], 
                    [ 601.5 , 1.4851255437026846e-46 ], [ 601.5625 , 1.4853405855567295e-46 ], [ 601.625 , 1.485555658548209e-46 ], [ 601.6875 , 1.4857707626816314e-46 ], 
                    [ 601.75 , 1.4859858979614846e-46 ], [ 601.8125 , 1.4862010643923107e-46 ], [ 601.875 , 1.4864162619786196e-46 ], [ 601.9375 , 1.486631490724902e-46 ], 
                    [ 602.0 , 1.4868467506356906e-46 ], [ 602.0625 , 1.4870620417154768e-46 ], [ 602.125 , 1.4872773639688057e-46 ], [ 602.1875 , 1.4874927174001487e-46 ], 
                    [ 602.25 , 1.4877081020140414e-46 ], [ 602.3125 , 1.4879235178149991e-46 ], [ 602.375 , 1.4881389648075586e-46 ], [ 602.4375 , 1.4883544429961835e-46 ], 
                    [ 602.5 , 1.4885699523854437e-46 ], [ 602.5625 , 1.488785492979836e-46 ], [ 602.625 , 1.489001064783868e-46 ], [ 602.6875 , 1.4892166678020803e-46 ], 
                    [ 602.75 , 1.4894323020389712e-46 ], [ 602.8125 , 1.489647967499082e-46 ], [ 602.875 , 1.4898635081114318e-46 ], [ 602.9375 , 1.490079058204489e-46 ], 
                    [ 603.0 , 1.4902946394828474e-46 ], [ 603.0625 , 1.4905102519510089e-46 ], [ 603.125 , 1.490725895613517e-46 ], [ 603.1875 , 1.4909415704748637e-46 ], 
                    [ 603.25 , 1.4911572765395423e-46 ], [ 603.3125 , 1.4913730138120983e-46 ], [ 603.375 , 1.4915887822970473e-46 ], [ 603.4375 , 1.4918045819988834e-46 ], 
                    [ 603.5 , 1.4920204129221232e-46 ], [ 603.5625 , 1.492236275071305e-46 ], [ 603.625 , 1.492452168450957e-46 ], [ 603.6875 , 1.4926680930655445e-46 ], 
                    [ 603.75 , 1.4928840489196291e-46 ], [ 603.8125 , 1.4931000360177407e-46 ], [ 603.875 , 1.4933160543643578e-46 ], [ 603.9375 , 1.4935321039640114e-46 ], 
                    [ 604.0 , 1.4937481848212446e-46 ], [ 604.0625 , 1.4939642969405692e-46 ], [ 604.125 , 1.4941804403265074e-46 ], [ 604.1875 , 1.4943966120423922e-46 ], 
                    [ 604.25 , 1.4946126322660996e-46 ], [ 604.3125 , 1.494828683716281e-46 ], [ 604.375 , 1.4950447663974501e-46 ], [ 604.4375 , 1.4952608803141433e-46 ], 
                    [ 604.5 , 1.4954770254708433e-46 ], [ 604.5625 , 1.4956932018720665e-46 ], [ 604.625 , 1.4959094095223291e-46 ], [ 604.6875 , 1.4961256484261695e-46 ], 
                    [ 604.75 , 1.4963419185881056e-46 ], [ 604.8125 , 1.4965582200126556e-46 ], [ 604.875 , 1.4967745527043072e-46 ], [ 604.9375 , 1.4969909166676012e-46 ], 
                    [ 605.0 , 1.4972073119070685e-46 ], [ 605.0625 , 1.4974237384271987e-46 ], [ 605.125 , 1.4976401962325448e-46 ], [ 605.1875 , 1.4978566853276405e-46 ], 
                    [ 605.25 , 1.4980732057169659e-46 ], [ 605.3125 , 1.4982897574050555e-46 ], [ 605.375 , 1.4985063403964759e-46 ], [ 605.4375 , 1.498722954695699e-46 ], 
                    [ 605.5 , 1.4989396003072722e-46 ], [ 605.5625 , 1.4991562578812515e-46 ], [ 605.625 , 1.4993727805715925e-46 ], [ 605.6875 , 1.4995893345342306e-46 ], 
                    [ 605.75 , 1.4998059197737035e-46 ], [ 605.8125 , 1.5000225362945177e-46 ], [ 605.875 , 1.5002391841011913e-46 ], [ 605.9375 , 1.500455863198243e-46 ], 
                    [ 606.0 , 1.5006725735901813e-46 ], [ 606.0625 , 1.5008893152815578e-46 ], [ 606.125 , 1.5011060882768616e-46 ], [ 606.1875 , 1.5013228925806135e-46 ], 
                    [ 606.25 , 1.5015397281973571e-46 ], [ 606.3125 , 1.5017565951315931e-46 ], [ 606.375 , 1.5019734933878771e-46 ], [ 606.4375 , 1.5021904229707007e-46 ], 
                    [ 606.5 , 1.5024073838845884e-46 ], [ 606.5625 , 1.502624376134087e-46 ], [ 606.625 , 1.5028413997237218e-46 ], [ 606.6875 , 1.503058454657998e-46 ], 
                    [ 606.75 , 1.503275540941443e-46 ], [ 606.8125 , 1.5034926585786059e-46 ], [ 606.875 , 1.5037098075740146e-46 ], [ 606.9375 , 1.5039269620245266e-46 ], 
                    [ 607.0 , 1.5041439925984473e-46 ], [ 607.0625 , 1.5043610544918879e-46 ], [ 607.125 , 1.5045781477093676e-46 ], [ 607.1875 , 1.5047952722554178e-46 ], 
                    [ 607.25 , 1.5050124281345169e-46 ], [ 607.3125 , 1.50522961535125e-46 ], [ 607.375 , 1.5054468339100865e-46 ], [ 607.4375 , 1.5056640838155918e-46 ], 
                    [ 607.5 , 1.505881365072258e-46 ], [ 607.5625 , 1.5060986776846088e-46 ], [ 607.625 , 1.506316021657212e-46 ], [ 607.6875 , 1.5065333969945186e-46 ], 
                    [ 607.75 , 1.50675080370114e-46 ], [ 607.8125 , 1.5069682417815175e-46 ], [ 607.875 , 1.5071857112402433e-46 ], [ 607.9375 , 1.5074032120818343e-46 ], 
                    [ 608.0 , 1.5076207443107983e-46 ], [ 608.0625 , 1.5078383079316649e-46 ], [ 608.125 , 1.5080559029489851e-46 ], [ 608.1875 , 1.5082735293672796e-46 ], 
                    [ 608.25 , 1.5084911871910688e-46 ], [ 608.3125 , 1.5087088520740672e-46 ], [ 608.375 , 1.508926388216127e-46 ], [ 608.4375 , 1.5091439557240617e-46 ], 
                    [ 608.5 , 1.5093615546023939e-46 ], [ 608.5625 , 1.5095791848556466e-46 ], [ 608.625 , 1.5097968464883546e-46 ], [ 608.6875 , 1.510014539505021e-46 ], 
                    [ 608.75 , 1.5102322639101811e-46 ], [ 608.8125 , 1.5104500197083617e-46 ], [ 608.875 , 1.5106678069040993e-46 ], [ 608.9375 , 1.510885625501889e-46 ], 
                    [ 609.0 , 1.5111034755063121e-46 ], [ 609.0625 , 1.5113213569218437e-46 ], [ 609.125 , 1.5115392697530124e-46 ], [ 609.1875 , 1.5117572140044126e-46 ], 
                    [ 609.25 , 1.5119751896805104e-46 ], [ 609.3125 , 1.5121931967858474e-46 ], [ 609.375 , 1.512411235324977e-46 ], [ 609.4375 , 1.5126293053024311e-46 ], 
                    [ 609.5 , 1.5128474067227322e-46 ], [ 609.5625 , 1.5130655395904033e-46 ], [ 609.625 , 1.5132837039099893e-46 ], [ 609.6875 , 1.5135018859067393e-46 ], 
                    [ 609.75 , 1.5137199191486064e-46 ], [ 609.8125 , 1.5139379838000556e-46 ], [ 609.875 , 1.5141560798656551e-46 ], [ 609.9375 , 1.5143742073499192e-46 ], 
                    [ 610.0 , 1.5145923662573423e-46 ], [ 610.0625 , 1.5148105565924723e-46 ], [ 610.125 , 1.5150287783598473e-46 ], [ 610.1875 , 1.515247031563996e-46 ], 
                    [ 610.25 , 1.515465316209414e-46 ], [ 610.3125 , 1.5156836323006422e-46 ], [ 610.375 , 1.5159019798422424e-46 ], [ 610.4375 , 1.5161203588387136e-46 ], 
                    [ 610.5 , 1.5163387692945975e-46 ], [ 610.5625 , 1.5165572112144583e-46 ], [ 610.625 , 1.5167756846027534e-46 ], [ 610.6875 , 1.516994189464081e-46 ], 
                    [ 610.75 , 1.5172127258029533e-46 ], [ 610.8125 , 1.5174312936238834e-46 ], [ 610.875 , 1.5176498929314388e-46 ], [ 610.9375 , 1.5178685237301553e-46 ], 
                    [ 611.0 , 1.5180871860245699e-46 ], [ 611.0625 , 1.5183058798191657e-46 ], [ 611.125 , 1.5185244246423326e-46 ], [ 611.1875 , 1.5187429934018683e-46 ], 
                    [ 611.25 , 1.5189615936210907e-46 ], [ 611.3125 , 1.519180225304517e-46 ], [ 611.375 , 1.5193988884567194e-46 ], [ 611.4375 , 1.5196175830821628e-46 ], 
                    [ 611.5 , 1.5198363091854198e-46 ], [ 611.5625 , 1.5200550667710111e-46 ], [ 611.625 , 1.5202738558434783e-46 ], [ 611.6875 , 1.5204926764073324e-46 ], 
                    [ 611.75 , 1.5207115284671166e-46 ], [ 611.8125 , 1.5209304120273533e-46 ], [ 611.875 , 1.52114932709262e-46 ], [ 611.9375 , 1.5213682736673864e-46 ], 
                    [ 612.0 , 1.5215872517562205e-46 ], [ 612.0625 , 1.5218062613637012e-46 ], [ 612.125 , 1.522025302494279e-46 ], [ 612.1875 , 1.5222443751525557e-46 ], 
                    [ 612.25 , 1.522463479343037e-46 ], [ 612.3125 , 1.5226826150702613e-46 ], [ 612.375 , 1.5229017823388005e-46 ], [ 612.4375 , 1.523120981153162e-46 ], 
                    [ 612.5 , 1.5233400610160246e-46 ], [ 612.5625 , 1.5235591345335827e-46 ], [ 612.625 , 1.5237782395563773e-46 ], [ 612.6875 , 1.523997376088951e-46 ], 
                    [ 612.75 , 1.5242165441358237e-46 ], [ 612.8125 , 1.5244357437015603e-46 ], [ 612.875 , 1.5246549747906395e-46 ], [ 612.9375 , 1.5248742374076592e-46 ], 
                    [ 613.0 , 1.5250935315571108e-46 ], [ 613.0625 , 1.5253128572435394e-46 ], [ 613.125 , 1.5255322144714804e-46 ], [ 613.1875 , 1.5257516032454702e-46 ], 
                    [ 613.25 , 1.5259710235700453e-46 ], [ 613.3125 , 1.5261904754497426e-46 ], [ 613.375 , 1.5264099588891114e-46 ], [ 613.4375 , 1.526629473892647e-46 ], 
                    [ 613.5 , 1.5268490204649536e-46 ], [ 613.5625 , 1.5270685986105061e-46 ], [ 613.625 , 1.5272882083338991e-46 ], [ 613.6875 , 1.5275078496396741e-46 ], 
                    [ 613.75 , 1.5277275225323188e-46 ], [ 613.8125 , 1.5279472270164082e-46 ], [ 613.875 , 1.528166845080599e-46 ], [ 613.9375 , 1.52838641189874e-46 ], 
                    [ 614.0 , 1.5286060102642245e-46 ], [ 614.0625 , 1.5288256401815743e-46 ], [ 614.125 , 1.5290453016553119e-46 ], [ 614.1875 , 1.5292649946899928e-46 ], 
                    [ 614.25 , 1.529484719290141e-46 ], [ 614.3125 , 1.5297044754602922e-46 ], [ 614.375 , 1.5299242632049817e-46 ], [ 614.4375 , 1.5301440825287356e-46 ], 
                    [ 614.5 , 1.5303639334361238e-46 ], [ 614.5625 , 1.5305838159316516e-46 ], [ 614.625 , 1.5308037300198684e-46 ], [ 614.6875 , 1.5310236757053134e-46 ], 
                    [ 614.75 , 1.5312436529925265e-46 ], [ 614.8125 , 1.5314636618860375e-46 ], [ 614.875 , 1.5316837023903985e-46 ], [ 614.9375 , 1.5319037745101727e-46 ], 
                    [ 615.0 , 1.5321238782498594e-46 ], [ 615.0625 , 1.5323440136140236e-46 ], [ 615.125 , 1.5325641806072198e-46 ], [ 615.1875 , 1.5327843792339815e-46 ], 
                    [ 615.25 , 1.5330045496954242e-46 ], [ 615.3125 , 1.5332246281668732e-46 ], [ 615.375 , 1.533444738232858e-46 ], [ 615.4375 , 1.5336648798978926e-46 ], 
                    [ 615.5 , 1.5338850531664802e-46 ], [ 615.5625 , 1.5341052580432235e-46 ], [ 615.625 , 1.5343254945326063e-46 ], [ 615.6875 , 1.5345457626391985e-46 ], 
                    [ 615.75 , 1.5347660623675505e-46 ], [ 615.8125 , 1.5349863937221586e-46 ], [ 615.875 , 1.5352067567075844e-46 ], [ 615.9375 , 1.5354271513283472e-46 ], 
                    [ 616.0 , 1.5356475775890432e-46 ], [ 616.0625 , 1.5358680354941708e-46 ], [ 616.125 , 1.5360885250482839e-46 ], [ 616.1875 , 1.5363090462559152e-46 ], 
                    [ 616.25 , 1.5365295991216308e-46 ], [ 616.3125 , 1.5367501836499641e-46 ], [ 616.375 , 1.536970799845483e-46 ], [ 616.4375 , 1.5371914477126682e-46 ], 
                    [ 616.5 , 1.5374121272561642e-46 ], [ 616.5625 , 1.5376328384804097e-46 ], [ 616.625 , 1.5378535784321486e-46 ], [ 616.6875 , 1.5380741405088566e-46 ], 
                    [ 616.75 , 1.5382947342190461e-46 ], [ 616.8125 , 1.5385153595672328e-46 ], [ 616.875 , 1.538736016557954e-46 ], [ 616.9375 , 1.538956705195737e-46 ], 
                    [ 617.0 , 1.5391774254851315e-46 ], [ 617.0625 , 1.5393981774306885e-46 ], [ 617.125 , 1.5396189610369366e-46 ], [ 617.1875 , 1.5398397763083953e-46 ], 
                    [ 617.25 , 1.5400606232496495e-46 ], [ 617.3125 , 1.5402815018652194e-46 ], [ 617.375 , 1.540502412159637e-46 ], [ 617.4375 , 1.540723354137479e-46 ], 
                    [ 617.5 , 1.540944327803245e-46 ], [ 617.5625 , 1.541165333161502e-46 ], [ 617.625 , 1.5413863702168173e-46 ], [ 617.6875 , 1.5416074389736823e-46 ], 
                    [ 617.75 , 1.541828539436687e-46 ], [ 617.8125 , 1.5420496716103794e-46 ], [ 617.875 , 1.5422708354992852e-46 ], [ 617.9375 , 1.542492031107986e-46 ], 
                    [ 618.0 , 1.5427132584409766e-46 ], [ 618.0625 , 1.5429343997616176e-46 ], [ 618.125 , 1.5431555097075381e-46 ], [ 618.1875 , 1.543376651339592e-46 ], 
                    [ 618.25 , 1.5435978246622978e-46 ], [ 618.3125 , 1.5438190296802302e-46 ], [ 618.375 , 1.5440402663978765e-46 ], [ 618.4375 , 1.5442615348198229e-46 ], 
                    [ 618.5 , 1.544482834950613e-46 ], [ 618.5625 , 1.5447041667947582e-46 ], [ 618.625 , 1.5449255303568688e-46 ], [ 618.6875 , 1.545146925641402e-46 ], 
                    [ 618.75 , 1.545368352652959e-46 ], [ 618.8125 , 1.5455898113960864e-46 ], [ 618.875 , 1.5458113018752984e-46 ], [ 618.9375 , 1.5460328240951873e-46 ], 
                    [ 619.0 , 1.546254378060301e-46 ], [ 619.0625 , 1.5464759637751235e-46 ], [ 619.125 , 1.5466975812442592e-46 ], [ 619.1875 , 1.546919230472259e-46 ], 
                    [ 619.25 , 1.5471409114636625e-46 ], [ 619.3125 , 1.547362624223011e-46 ], [ 619.375 , 1.5475843687548682e-46 ], [ 619.4375 , 1.547806096754514e-46 ], 
                    [ 619.5 , 1.5480277017605647e-46 ], [ 619.5625 , 1.5482493384946392e-46 ], [ 619.625 , 1.5484710069612151e-46 ], [ 619.6875 , 1.5486927071648903e-46 ], 
                    [ 619.75 , 1.548914439110187e-46 ], [ 619.8125 , 1.5491362028016383e-46 ], [ 619.875 , 1.5493579982438009e-46 ], [ 619.9375 , 1.5495798254412312e-46 ], 
                    [ 620.0 , 1.5498016843984432e-46 ], [ 620.0625 , 1.5500235751200276e-46 ], [ 620.125 , 1.5502454976104995e-46 ], [ 620.1875 , 1.5504674518744295e-46 ], 
                    [ 620.25 , 1.5506894379163444e-46 ], [ 620.3125 , 1.5509114557407942e-46 ], [ 620.375 , 1.551133505352362e-46 ], [ 620.4375 , 1.5513555867555883e-46 ], 
                    [ 620.5 , 1.5515776999550028e-46 ], [ 620.5625 , 1.5517998449551463e-46 ], [ 620.625 , 1.5520220217606054e-46 ], [ 620.6875 , 1.5522442303759226e-46 ], 
                    [ 620.75 , 1.5524664708056744e-46 ], [ 620.8125 , 1.5526887430543604e-46 ], [ 620.875 , 1.5529108835640371e-46 ], [ 620.9375 , 1.553133008024035e-46 ], 
                    [ 621.0 , 1.553355164256134e-46 ], [ 621.0625 , 1.553577352264901e-46 ], [ 621.125 , 1.5537995720549026e-46 ], [ 621.1875 , 1.554021823630641e-46 ], 
                    [ 621.25 , 1.5542441069967071e-46 ], [ 621.3125 , 1.5544664221576037e-46 ], [ 621.375 , 1.5546887691179005e-46 ], [ 621.4375 , 1.5549111478821574e-46 ], 
                    [ 621.5 , 1.5551335584549127e-46 ], [ 621.5625 , 1.5553560008406935e-46 ], [ 621.625 , 1.5555784750441061e-46 ], [ 621.6875 , 1.555800981069635e-46 ], 
                    [ 621.75 , 1.556023518921898e-46 ], [ 621.8125 , 1.5562460886053928e-46 ], [ 621.875 , 1.5564686901247165e-46 ], [ 621.9375 , 1.5566913234843673e-46 ], 
                    [ 622.0 , 1.556913988688955e-46 ], [ 622.0625 , 1.557136685743013e-46 ], [ 622.125 , 1.5573594146510958e-46 ], [ 622.1875 , 1.557582175417761e-46 ], 
                    [ 622.25 , 1.5578049191019296e-46 ], [ 622.3125 , 1.5580275542645797e-46 ], [ 622.375 , 1.5582502212453443e-46 ], [ 622.4375 , 1.5584729200487602e-46 ], 
                    [ 622.5 , 1.558695650679375e-46 ], [ 622.5625 , 1.558918413141748e-46 ], [ 622.625 , 1.5591412074404623e-46 ], [ 622.6875 , 1.5593640335800124e-46 ], 
                    [ 622.75 , 1.5595868915649708e-46 ], [ 622.8125 , 1.559809781399889e-46 ], [ 622.875 , 1.5600327030893182e-46 ], [ 622.9375 , 1.5602556566378005e-46 ], 
                    [ 623.0 , 1.5604786420499222e-46 ], [ 623.0625 , 1.560701659330215e-46 ], [ 623.125 , 1.5609247084831997e-46 ], [ 623.1875 , 1.5611477895135095e-46 ], 
                    [ 623.25 , 1.5613709024256444e-46 ], [ 623.3125 , 1.56159404722415e-46 ], [ 623.375 , 1.561817223913605e-46 ], [ 623.4375 , 1.5620404324985902e-46 ], 
                    [ 623.5 , 1.5622636729836412e-46 ], [ 623.5625 , 1.5624869453732949e-46 ], [ 623.625 , 1.5627102496721556e-46 ], [ 623.6875 , 1.5629334428708275e-46 ], 
                    [ 623.75 , 1.563156604427513e-46 ], [ 623.8125 , 1.5633797978480733e-46 ], [ 623.875 , 1.5636030231370135e-46 ], [ 623.9375 , 1.5638262802989278e-46 ], 
                    [ 624.0 , 1.5640495693383344e-46 ], [ 624.0625 , 1.5642728902598069e-46 ], [ 624.125 , 1.5644962430678753e-46 ], [ 624.1875 , 1.5647196277671147e-46 ], 
                    [ 624.25 , 1.5649430443620562e-46 ], [ 624.3125 , 1.5651664928572656e-46 ], [ 624.375 , 1.5653899732572973e-46 ], [ 624.4375 , 1.5656134855667183e-46 ], 
                    [ 624.5 , 1.5658370297900843e-46 ], [ 624.5625 , 1.5660606059318966e-46 ], [ 624.625 , 1.566284213996802e-46 ], [ 624.6875 , 1.5665078539892804e-46 ], 
                    [ 624.75 , 1.566731525913924e-46 ], [ 624.8125 , 1.5669552297753145e-46 ], [ 624.875 , 1.5671789655780006e-46 ], [ 624.9375 , 1.5674027333264986e-46 ], 
                    [ 625.0 , 1.5676265330254147e-46 ], [ 625.0625 , 1.5678503491850761e-46 ], [ 625.125 , 1.5680740067265955e-46 ], [ 625.1875 , 1.5682976961734114e-46 ], 
                    [ 625.25 , 1.56852141753002e-46 ], [ 625.3125 , 1.5687451708010178e-46 ], [ 625.375 , 1.5689689559909571e-46 ], [ 625.4375 , 1.5691927731043581e-46 ], 
                    [ 625.5 , 1.5694166221458194e-46 ], [ 625.5625 , 1.569640503119862e-46 ], [ 625.625 , 1.5698644160310631e-46 ], [ 625.6875 , 1.570088360883957e-46 ], 
                    [ 625.75 , 1.5703123376831114e-46 ], [ 625.8125 , 1.570536346433072e-46 ], [ 625.875 , 1.5707603871384409e-46 ], [ 625.9375 , 1.5709844598037216e-46 ], 
                    [ 626.0 , 1.5712085644334952e-46 ], [ 626.0625 , 1.5714327010323104e-46 ], [ 626.125 , 1.5716568696047612e-46 ], [ 626.1875 , 1.571881070155375e-46 ], 
                    [ 626.25 , 1.5721053026887362e-46 ], [ 626.3125 , 1.5723295672093624e-46 ], [ 626.375 , 1.572553863721895e-46 ], [ 626.4375 , 1.5727781922308303e-46 ], 
                    [ 626.5 , 1.5730024680543616e-46 ], [ 626.5625 , 1.573226652310141e-46 ], [ 626.625 , 1.573450868516686e-46 ], [ 626.6875 , 1.5736751166784835e-46 ], 
                    [ 626.75 , 1.5738993968001432e-46 ], [ 626.8125 , 1.5741237088861868e-46 ], [ 626.875 , 1.574348052941192e-46 ], [ 626.9375 , 1.574572428969693e-46 ], 
                    [ 627.0 , 1.5747968369762572e-46 ], [ 627.0625 , 1.5750212769654427e-46 ], [ 627.125 , 1.5752457489418299e-46 ], [ 627.1875 , 1.575470252909944e-46 ], 
                    [ 627.25 , 1.5756947888743444e-46 ], [ 627.3125 , 1.5759193568395912e-46 ], [ 627.375 , 1.5761439568102789e-46 ], [ 627.4375 , 1.576368588790924e-46 ], 
                    [ 627.5 , 1.5765932527861223e-46 ], [ 627.5625 , 1.5768179488004475e-46 ], [ 627.625 , 1.5770426768384072e-46 ], [ 627.6875 , 1.5772674369046322e-46 ], 
                    [ 627.75 , 1.5774922290036203e-46 ], [ 627.8125 , 1.5777170531399817e-46 ], [ 627.875 , 1.5779419093182934e-46 ], [ 627.9375 , 1.5781666559418004e-46 ], 
                    [ 628.0 , 1.5783913699687069e-46 ], [ 628.0625 , 1.57861611599245e-46 ], [ 628.125 , 1.578840894017642e-46 ], [ 628.1875 , 1.5790657040488057e-46 ], 
                    [ 628.25 , 1.57929054609051e-46 ], [ 628.3125 , 1.5795154201473236e-46 ], [ 628.375 , 1.5797403262237827e-46 ], [ 628.4375 , 1.5799652643244584e-46 ], 
                    [ 628.5 , 1.5801902344539099e-46 ], [ 628.5625 , 1.5804152366167094e-46 ], [ 628.625 , 1.5806402708173842e-46 ], [ 628.6875 , 1.5808653370605409e-46 ], 
                    [ 628.75 , 1.5810904353507198e-46 ], [ 628.8125 , 1.581315565692473e-46 ], [ 628.875 , 1.5815407280903866e-46 ], [ 628.9375 , 1.5817659225490023e-46 ], 
                    [ 629.0 , 1.581991149072897e-46 ], [ 629.0625 , 1.5822164076666471e-46 ], [ 629.125 , 1.5824416983347857e-46 ], [ 629.1875 , 1.5826670210819246e-46 ], 
                    [ 629.25 , 1.582892375912609e-46 ], [ 629.3125 , 1.5831177628314075e-46 ], [ 629.375 , 1.5833429875441e-46 ], [ 629.4375 , 1.5835682263300919e-46 ], 
                    [ 629.5 , 1.5837934971574745e-46 ], [ 629.5625 , 1.5840188000308169e-46 ], [ 629.625 , 1.584244134954666e-46 ], [ 629.6875 , 1.5844695019335707e-46 ], 
                    [ 629.75 , 1.5846949009721015e-46 ], [ 629.8125 , 1.584920332074831e-46 ], [ 629.875 , 1.585145795246286e-46 ], [ 629.9375 , 1.5853712904910632e-46 ], 
                    [ 630.0 , 1.5855968178137241e-46 ], [ 630.0625 , 1.5858223772187989e-46 ], [ 630.125 , 1.586047968710896e-46 ], [ 630.1875 , 1.5862735922945693e-46 ], 
                    [ 630.25 , 1.5864992479743384e-46 ], [ 630.3125 , 1.5867249357548148e-46 ], [ 630.375 , 1.5869506556405531e-46 ], [ 630.4375 , 1.5871764076361318e-46 ], 
                    [ 630.5 , 1.587402191746108e-46 ], [ 630.5625 , 1.5876280079750381e-46 ], [ 630.625 , 1.5878538563275028e-46 ], [ 630.6875 , 1.588079736808072e-46 ], 
                    [ 630.75 , 1.588305630225486e-46 ], [ 630.8125 , 1.5885313670166442e-46 ], [ 630.875 , 1.588757135890481e-46 ], [ 630.9375 , 1.5889829368515445e-46 ], 
                    [ 631.0 , 1.589208769904407e-46 ], [ 631.0625 , 1.5894346350536398e-46 ], [ 631.125 , 1.5896605323037832e-46 ], [ 631.1875 , 1.5898864616594098e-46 ], 
                    [ 631.25 , 1.590112423125083e-46 ], [ 631.3125 , 1.590338416705378e-46 ], [ 631.375 , 1.590564442404836e-46 ], [ 631.4375 , 1.5907905002280332e-46 ], 
                    [ 631.5 , 1.5910165901795354e-46 ], [ 631.5625 , 1.5912427122639314e-46 ], [ 631.625 , 1.591468866485743e-46 ], [ 631.6875 , 1.59169505284956e-46 ], 
                    [ 631.75 , 1.5919212713599619e-46 ], [ 631.8125 , 1.592147522021495e-46 ], [ 631.875 , 1.59237380483874e-46 ], [ 631.9375 , 1.5926001198162786e-46 ], 
                    [ 632.0 , 1.5928264669586584e-46 ], [ 632.0625 , 1.5930528462704628e-46 ], [ 632.125 , 1.593279257756263e-46 ], [ 632.1875 , 1.5935056541156428e-46 ], 
                    [ 632.25 , 1.5937319125311762e-46 ], [ 632.3125 , 1.5939582030726518e-46 ], [ 632.375 , 1.594184525744654e-46 ], [ 632.4375 , 1.5944108805517107e-46 ], 
                    [ 632.5 , 1.5946372674983962e-46 ], [ 632.5625 , 1.5948636865892735e-46 ], [ 632.625 , 1.595090137828907e-46 ], [ 632.6875 , 1.5953166212218615e-46 ], 
                    [ 632.75 , 1.5955431367727136e-46 ], [ 632.8125 , 1.5957696844860068e-46 ], [ 632.875 , 1.595996264366319e-46 ], [ 632.9375 , 1.5962228764182062e-46 ], 
                    [ 633.0 , 1.5964495206462817e-46 ], [ 633.0625 , 1.5966761970550571e-46 ], [ 633.125 , 1.5969029056491252e-46 ], [ 633.1875 , 1.5971296464330551e-46 ], 
                    [ 633.25 , 1.5973564194114064e-46 ], [ 633.3125 , 1.5975832245887843e-46 ], [ 633.375 , 1.5978100619697382e-46 ], [ 633.4375 , 1.59803693155884e-46 ], 
                    [ 633.5 , 1.5982638333606409e-46 ], [ 633.5625 , 1.5984907673797368e-46 ], [ 633.625 , 1.5987176725919105e-46 ], [ 633.6875 , 1.5989444613287006e-46 ], 
                    [ 633.75 , 1.5991712822369817e-46 ], [ 633.8125 , 1.5993981353213062e-46 ], [ 633.875 , 1.5996250205862726e-46 ], [ 633.9375 , 1.5998519380364116e-46 ], 
                    [ 634.0 , 1.6000788876763004e-46 ], [ 634.0625 , 1.6003058695104939e-46 ], [ 634.125 , 1.6005328835435933e-46 ], [ 634.1875 , 1.600759929780132e-46 ], 
                    [ 634.25 , 1.6009870082247013e-46 ], [ 634.3125 , 1.6012141188818355e-46 ], [ 634.375 , 1.60144126175615e-46 ], [ 634.4375 , 1.6016684368521925e-46 ], 
                    [ 634.5 , 1.6018956441744991e-46 ], [ 634.5625 , 1.6021228837276987e-46 ], [ 634.625 , 1.6023501555163518e-46 ], [ 634.6875 , 1.6025774595450088e-46 ], 
                    [ 634.75 , 1.6028047958182655e-46 ], [ 634.8125 , 1.6030321643406735e-46 ], [ 634.875 , 1.603259565116819e-46 ], [ 634.9375 , 1.6034869981512772e-46 ], 
                    [ 635.0 , 1.6037144634486244e-46 ], [ 635.0625 , 1.6039418909778875e-46 ], [ 635.125 , 1.604169196644921e-46 ], [ 635.1875 , 1.6043965345250196e-46 ], 
                    [ 635.25 , 1.6046239046227258e-46 ], [ 635.3125 , 1.6048513069426164e-46 ], [ 635.375 , 1.6050787414892695e-46 ], [ 635.4375 , 1.6053062082672292e-46 ], 
                    [ 635.5 , 1.6055337072810636e-46 ], [ 635.5625 , 1.6057612385353746e-46 ], [ 635.625 , 1.6059888020346978e-46 ], [ 635.6875 , 1.6062163977836365e-46 ], 
                    [ 635.75 , 1.6064440257867158e-46 ], [ 635.8125 , 1.6066716860485291e-46 ], [ 635.875 , 1.6068993785736596e-46 ], [ 635.9375 , 1.6071271033666453e-46 ], 
                    [ 636.0 , 1.607354860432082e-46 ], [ 636.0625 , 1.6075826497745889e-46 ], [ 636.125 , 1.6078104713986372e-46 ], [ 636.1875 , 1.6080383253088935e-46 ], 
                    [ 636.25 , 1.6082662115098872e-46 ], [ 636.3125 , 1.6084941300062175e-46 ], [ 636.375 , 1.608722080802427e-46 ], [ 636.4375 , 1.6089500639031504e-46 ], 
                    [ 636.5 , 1.6091780190909462e-46 ], [ 636.5625 , 1.6094058530405975e-46 ], [ 636.625 , 1.6096337192479262e-46 ], [ 636.6875 , 1.6098616177174767e-46 ], 
                    [ 636.75 , 1.6100895484537939e-46 ], [ 636.8125 , 1.610317511461492e-46 ], [ 636.875 , 1.6105455067450826e-46 ], [ 636.9375 , 1.6107735343092045e-46 ], 
                    [ 637.0 , 1.6110015941583824e-46 ], [ 637.0625 , 1.6112296862971983e-46 ], [ 637.125 , 1.6114578107302244e-46 ], [ 637.1875 , 1.6116859674620214e-46 ], 
                    [ 637.25 , 1.6119141564972078e-46 ], [ 637.3125 , 1.6121423778402892e-46 ], [ 637.375 , 1.6123706314958852e-46 ], [ 637.4375 , 1.6125989174685708e-46 ], 
                    [ 637.5 , 1.6128272357628758e-46 ], [ 637.5625 , 1.6130555863834565e-46 ], [ 637.625 , 1.6132839693347985e-46 ], [ 637.6875 , 1.6135123846215707e-46 ], 
                    [ 637.75 , 1.6137408322482595e-46 ], [ 637.8125 , 1.6139693122195358e-46 ], [ 637.875 , 1.6141978245398987e-46 ], [ 637.9375 , 1.6144263258115825e-46 ], 
                    [ 638.0 , 1.6146546845728744e-46 ], [ 638.0625 , 1.6148830756352637e-46 ], [ 638.125 , 1.615111499003297e-46 ], [ 638.1875 , 1.6153399546815434e-46 ], 
                    [ 638.25 , 1.6155684426746078e-46 ], [ 638.3125 , 1.6157969629870148e-46 ], [ 638.375 , 1.6160255156233824e-46 ], [ 638.4375 , 1.6162541005882594e-46 ], 
                    [ 638.5 , 1.6164827178861845e-46 ], [ 638.5625 , 1.6167113675217886e-46 ], [ 638.625 , 1.616940049499634e-46 ], [ 638.6875 , 1.6171687638242841e-46 ], 
                    [ 638.75 , 1.6173975105002917e-46 ], [ 638.8125 , 1.6176262895322552e-46 ], [ 638.875 , 1.6178551009247749e-46 ], [ 638.9375 , 1.6180839446823932e-46 ], 
                    [ 639.0 , 1.6183128208096887e-46 ], [ 639.0625 , 1.6185417293112624e-46 ], [ 639.125 , 1.618770670191694e-46 ], [ 639.1875 , 1.618999643455529e-46 ], 
                    [ 639.25 , 1.6192286491073936e-46 ], [ 639.3125 , 1.6194576871518234e-46 ], [ 639.375 , 1.6196867435818585e-46 ], [ 639.4375 , 1.6199156263609428e-46 ], 
                    [ 639.5 , 1.6201445414841371e-46 ], [ 639.5625 , 1.6203734889559891e-46 ], [ 639.625 , 1.620602468781128e-46 ], [ 639.6875 , 1.620831480964068e-46 ], 
                    [ 639.75 , 1.6210605255094276e-46 ], [ 639.8125 , 1.6212896024217342e-46 ], [ 639.875 , 1.6215187117055961e-46 ], [ 639.9375 , 1.6217478533655762e-46 ], 
                    [ 640.0 , 1.6219770274062266e-46 ], [ 640.0625 , 1.622206233832181e-46 ], [ 640.125 , 1.6224354726479696e-46 ], [ 640.1875 , 1.6226647438581807e-46 ], 
                    [ 640.25 , 1.6228940474673696e-46 ], [ 640.3125 , 1.6231233834801606e-46 ], [ 640.375 , 1.6233527519010982e-46 ], [ 640.4375 , 1.623582152734785e-46 ], 
                    [ 640.5 , 1.6238115859857897e-46 ], [ 640.5625 , 1.6240410516587051e-46 ], [ 640.625 , 1.6242705497580896e-46 ], [ 640.6875 , 1.6245000802885139e-46 ], 
                    [ 640.75 , 1.6247296432546073e-46 ], [ 640.8125 , 1.6249592386609418e-46 ], [ 640.875 , 1.6251886679798569e-46 ], [ 640.9375 , 1.6254181016267772e-46 ], 
                    [ 641.0 , 1.625647567663671e-46 ], [ 641.0625 , 1.6258770660950871e-46 ], [ 641.125 , 1.6261065969256103e-46 ], [ 641.1875 , 1.6263361601598268e-46 ], 
                    [ 641.25 , 1.6265657558022987e-46 ], [ 641.3125 , 1.62679538385759e-46 ], [ 641.375 , 1.6270250443302882e-46 ], [ 641.4375 , 1.6272547372249583e-46 ], 
                    [ 641.5 , 1.6274844625462347e-46 ], [ 641.5625 , 1.627714220298603e-46 ], [ 641.625 , 1.6279440104866997e-46 ], [ 641.6875 , 1.628173833115092e-46 ], 
                    [ 641.75 , 1.6284036881883712e-46 ], [ 641.8125 , 1.6286335757111063e-46 ], [ 641.875 , 1.6288634956878778e-46 ], [ 641.9375 , 1.6290934481232448e-46 ], 
                    [ 642.0 , 1.629323433021847e-46 ], [ 642.0625 , 1.6295534503882098e-46 ], [ 642.125 , 1.6297835002269398e-46 ], [ 642.1875 , 1.6300135825426212e-46 ], 
                    [ 642.25 , 1.6302436973398622e-46 ], [ 642.3125 , 1.6304737084227114e-46 ], [ 642.375 , 1.6307036739417308e-46 ], [ 642.4375 , 1.6309336718955816e-46 ], 
                    [ 642.5 , 1.6311637022888389e-46 ], [ 642.5625 , 1.6313937651260774e-46 ], [ 642.625 , 1.6316238604118735e-46 ], [ 642.6875 , 1.6318539881508037e-46 ], 
                    [ 642.75 , 1.6320841483474334e-46 ], [ 642.8125 , 1.6323143410063755e-46 ], [ 642.875 , 1.6325445661321623e-46 ], [ 642.9375 , 1.6327748237294191e-46 ], 
                    [ 643.0 , 1.6330051138026913e-46 ], [ 643.0625 , 1.6332354363565592e-46 ], [ 643.125 , 1.6334657913956154e-46 ], [ 643.1875 , 1.633696178924465e-46 ], 
                    [ 643.25 , 1.6339265989476442e-46 ], [ 643.3125 , 1.6341570514697706e-46 ], [ 643.375 , 1.6343875364954395e-46 ], [ 643.4375 , 1.6346180540291776e-46 ], 
                    [ 643.5 , 1.634848604075616e-46 ], [ 643.5625 , 1.6350791866393287e-46 ], [ 643.625 , 1.6353098017248908e-46 ], [ 643.6875 , 1.6355404493369124e-46 ], 
                    [ 643.75 , 1.635771048813536e-46 ], [ 643.8125 , 1.6360015262959889e-46 ], [ 643.875 , 1.6362320362523325e-46 ], [ 643.9375 , 1.636462578687166e-46 ], 
                    [ 644.0 , 1.6366931536050422e-46 ], [ 644.0625 , 1.6369237610105732e-46 ], [ 644.125 , 1.6371544009082893e-46 ], [ 644.1875 , 1.6373850733027923e-46 ], 
                    [ 644.25 , 1.6376157781986726e-46 ], [ 644.3125 , 1.6378465156004859e-46 ], [ 644.375 , 1.6380772855128361e-46 ], [ 644.4375 , 1.6383080879403031e-46 ], 
                    [ 644.5 , 1.6385389228874227e-46 ], [ 644.5625 , 1.6387697903588455e-46 ], [ 644.625 , 1.639000690359097e-46 ], [ 644.6875 , 1.6392316228927947e-46 ], 
                    [ 644.75 , 1.639462587964523e-46 ], [ 644.8125 , 1.6396935855788077e-46 ], [ 644.875 , 1.6399246157403394e-46 ], [ 644.9375 , 1.6401556784535986e-46 ], 
                    [ 645.0 , 1.6403867737232418e-46 ], [ 645.0625 , 1.6406179015538332e-46 ], [ 645.125 , 1.6408490619499602e-46 ], [ 645.1875 , 1.6410802549161995e-46 ], 
                    [ 645.25 , 1.6413112584610932e-46 ], [ 645.3125 , 1.6415422939528891e-46 ], [ 645.375 , 1.6417733619658885e-46 ], [ 645.4375 , 1.6420044625046458e-46 ], 
                    [ 645.5 , 1.6422355955737514e-46 ], [ 645.5625 , 1.6424667611777958e-46 ], [ 645.625 , 1.6426979593213474e-46 ], [ 645.6875 , 1.6429291900089743e-46 ], 
                    [ 645.75 , 1.6431604532452812e-46 ], [ 645.8125 , 1.6433917490348379e-46 ], [ 645.875 , 1.6436230773822152e-46 ], [ 645.9375 , 1.643854438292019e-46 ], 
                    [ 646.0 , 1.6440858317688219e-46 ], [ 646.0625 , 1.6443172578172192e-46 ], [ 646.125 , 1.6445487164417608e-46 ], [ 646.1875 , 1.6447802076470557e-46 ], 
                    [ 646.25 , 1.6450117314376902e-46 ], [ 646.3125 , 1.6452432878182622e-46 ], [ 646.375 , 1.64547487679336e-46 ], [ 646.4375 , 1.6457064983675245e-46 ], 
                    [ 646.5 , 1.6459381525454144e-46 ], [ 646.5625 , 1.6461698393315373e-46 ], [ 646.625 , 1.6464015587305772e-46 ], [ 646.6875 , 1.646633172461113e-46 ], 
                    [ 646.75 , 1.6468647261401321e-46 ], [ 646.8125 , 1.647096312380798e-46 ], [ 646.875 , 1.647327931187689e-46 ], [ 646.9375 , 1.6475595825654198e-46 ], 
                    [ 647.0 , 1.6477912665185236e-46 ], [ 647.0625 , 1.6480229830515935e-46 ], [ 647.125 , 1.6482547321692575e-46 ], [ 647.1875 , 1.648486513876051e-46 ], 
                    [ 647.25 , 1.648718328176568e-46 ], [ 647.3125 , 1.6489501750753926e-46 ], [ 647.375 , 1.6491820545770965e-46 ], [ 647.4375 , 1.6494139666862883e-46 ], 
                    [ 647.5 , 1.6496459114075416e-46 ], [ 647.5625 , 1.6498778887454422e-46 ], [ 647.625 , 1.6501098987045768e-46 ], [ 647.6875 , 1.6503419412895093e-46 ], 
                    [ 647.75 , 1.6505740165049095e-46 ], [ 647.8125 , 1.650806124355249e-46 ], [ 647.875 , 1.651038264845187e-46 ], [ 647.9375 , 1.6512704379793254e-46 ], 
                    [ 648.0 , 1.651502643762208e-46 ], [ 648.0625 , 1.6517348821984605e-46 ], [ 648.125 , 1.651967119605196e-46 ], [ 648.1875 , 1.652199194469984e-46 ], 
                    [ 648.25 , 1.652431301937577e-46 ], [ 648.3125 , 1.6526634420125322e-46 ], [ 648.375 , 1.6528956146994534e-46 ], [ 648.4375 , 1.6531278200028871e-46 ], 
                    [ 648.5 , 1.6533600579274507e-46 ], [ 648.5625 , 1.6535923284777268e-46 ], [ 648.625 , 1.6538246316582634e-46 ], [ 648.6875 , 1.6540569674737154e-46 ], 
                    [ 648.75 , 1.6542893359285729e-46 ], [ 648.8125 , 1.654521737027469e-46 ], [ 648.875 , 1.6547541707750124e-46 ], [ 648.9375 , 1.654986637175767e-46 ], 
                    [ 649.0 , 1.6552191362343195e-46 ], [ 649.0625 , 1.6554516679552343e-46 ], [ 649.125 , 1.6556842323431352e-46 ], [ 649.1875 , 1.6559168294026236e-46 ], 
                    [ 649.25 , 1.6561494591382415e-46 ], [ 649.3125 , 1.6563821215545918e-46 ], [ 649.375 , 1.6566148166563007e-46 ], [ 649.4375 , 1.6568475444479246e-46 ], 
                    [ 649.5 , 1.6570803049340562e-46 ], [ 649.5625 , 1.6573130981193233e-46 ], [ 649.625 , 1.6575457733521443e-46 ], [ 649.6875 , 1.6577783973745453e-46 ], 
                    [ 649.75 , 1.6580110540439695e-46 ], [ 649.8125 , 1.6582437433650225e-46 ], [ 649.875 , 1.658476465342228e-46 ], [ 649.9375 , 1.658709219980204e-46 ], 
                    [ 650.0 , 1.658942007283558e-46 ], [ 650.0625 , 1.6591748272568275e-46 ], [ 650.125 , 1.6594076799046323e-46 ], [ 650.1875 , 1.659640565231558e-46 ], 
                    [ 650.25 , 1.659873483242156e-46 ], [ 650.3125 , 1.66010643394106e-46 ], [ 650.375 , 1.6603394173328104e-46 ], [ 650.4375 , 1.6605724334220192e-46 ], 
                    [ 650.5 , 1.660805482213287e-46 ], [ 650.5625 , 1.6610385637112147e-46 ], [ 650.625 , 1.6612716779203457e-46 ], [ 650.6875 , 1.6615048248452943e-46 ], 
                    [ 650.75 , 1.6617380044906398e-46 ], [ 650.8125 , 1.661971216860986e-46 ], [ 650.875 , 1.6622044619609617e-46 ], [ 650.9375 , 1.662437739795101e-46 ], 
                    [ 651.0 , 1.6626710503680208e-46 ], [ 651.0625 , 1.662904373831337e-46 ], [ 651.125 , 1.6631375148577565e-46 ], [ 651.1875 , 1.663370688570805e-46 ], 
                    [ 651.25 , 1.6636038949750649e-46 ], [ 651.3125 , 1.6638371340751318e-46 ], [ 651.375 , 1.6640704058755775e-46 ], [ 651.4375 , 1.664303710380975e-46 ], 
                    [ 651.5 , 1.6645370475959095e-46 ], [ 651.5625 , 1.6647704175250024e-46 ], [ 651.625 , 1.6650038201727933e-46 ], [ 651.6875 , 1.665237255543916e-46 ], 
                    [ 651.75 , 1.6654707236429235e-46 ], [ 651.8125 , 1.6657042244744157e-46 ], [ 651.875 , 1.6659377580429937e-46 ], [ 651.9375 , 1.666171324353212e-46 ], 
                    [ 652.0 , 1.6664049234096841e-46 ], [ 652.0625 , 1.6666385552170254e-46 ], [ 652.125 , 1.6668722197798038e-46 ], [ 652.1875 , 1.6671059171026113e-46 ], 
                    [ 652.25 , 1.6673396471900173e-46 ], [ 652.3125 , 1.6675734100466633e-46 ], [ 652.375 , 1.6678072056770953e-46 ], [ 652.4375 , 1.6680410340859565e-46 ], 
                    [ 652.5 , 1.6682748952778064e-46 ], [ 652.5625 , 1.668508679805881e-46 ], [ 652.625 , 1.668742375111928e-46 ], [ 652.6875 , 1.6689761031498855e-46 ], 
                    [ 652.75 , 1.6692098639243734e-46 ], [ 652.8125 , 1.6694436574399174e-46 ], [ 652.875 , 1.6696774837011504e-46 ], [ 652.9375 , 1.6699113427126598e-46 ], 
                    [ 653.0 , 1.67014523447902e-46 ], [ 653.0625 , 1.670379159004795e-46 ], [ 653.125 , 1.6706131162946328e-46 ], [ 653.1875 , 1.6708471063530872e-46 ], 
                    [ 653.25 , 1.6710811291847231e-46 ], [ 653.3125 , 1.6713151847941553e-46 ], [ 653.375 , 1.6715492731859745e-46 ], [ 653.4375 , 1.6717833943647962e-46 ], 
                    [ 653.5 , 1.6720175483351886e-46 ], [ 653.5625 , 1.6722517351017089e-46 ], [ 653.625 , 1.6724859546690101e-46 ], [ 653.6875 , 1.6727202070416509e-46 ], 
                    [ 653.75 , 1.672954492224238e-46 ], [ 653.8125 , 1.6731888102213903e-46 ], [ 653.875 , 1.6734231610376565e-46 ], [ 653.9375 , 1.6736575446776689e-46 ], 
                    [ 654.0 , 1.6738919611460012e-46 ], [ 654.0625 , 1.6741262206924123e-46 ], [ 654.125 , 1.6743604679266654e-46 ], [ 654.1875 , 1.6745947479372801e-46 ], 
                    [ 654.25 , 1.6748290607288306e-46 ], [ 654.3125 , 1.675063406305928e-46 ], [ 654.375 , 1.675297784673136e-46 ], [ 654.4375 , 1.6755321958350656e-46 ], 
                    [ 654.5 , 1.6757666397962826e-46 ], [ 654.5625 , 1.6760011165613995e-46 ], [ 654.625 , 1.6762356261349948e-46 ], [ 654.6875 , 1.6764701685216589e-46 ], 
                    [ 654.75 , 1.6767047437259596e-46 ], [ 654.8125 , 1.676939351752524e-46 ], [ 654.875 , 1.6771739926059689e-46 ], [ 654.9375 , 1.6774086662908039e-46 ], 
                    [ 655.0 , 1.6776433728116946e-46 ], [ 655.0625 , 1.6778781121731994e-46 ], [ 655.125 , 1.6781128843799495e-46 ], [ 655.1875 , 1.6783476894364686e-46 ], 
                    [ 655.25 , 1.6785825273474492e-46 ], [ 655.3125 , 1.6788173981174046e-46 ], [ 655.375 , 1.6790523017510038e-46 ], [ 655.4375 , 1.679287238252774e-46 ], 
                    [ 655.5 , 1.6795221859782515e-46 ], [ 655.5625 , 1.6797569461411857e-46 ], [ 655.625 , 1.6799917391184135e-46 ], [ 655.6875 , 1.6802265649145331e-46 ], 
                    [ 655.75 , 1.6804614235341206e-46 ], [ 655.8125 , 1.6806963149817639e-46 ], [ 655.875 , 1.6809312392620392e-46 ], [ 655.9375 , 1.6811661963795603e-46 ], 
                    [ 656.0 , 1.681401186338905e-46 ], [ 656.0625 , 1.6816362091446753e-46 ], [ 656.125 , 1.6818712648014273e-46 ], [ 656.1875 , 1.6821063533137882e-46 ], 
                    [ 656.25 , 1.6823414746863268e-46 ], [ 656.3125 , 1.6825766289236601e-46 ], [ 656.375 , 1.6828118160303819e-46 ], [ 656.4375 , 1.6830470360110625e-46 ], 
                    [ 656.5 , 1.6832822888702973e-46 ], [ 656.5625 , 1.6835175746126936e-46 ], [ 656.625 , 1.68375289324286e-46 ], [ 656.6875 , 1.6839882447653814e-46 ], 
                    [ 656.75 , 1.6842236291848433e-46 ], [ 656.8125 , 1.6844590465058803e-46 ], [ 656.875 , 1.684694496733043e-46 ], [ 656.9375 , 1.6849299798709668e-46 ], 
                    [ 657.0 , 1.6851654193028636e-46 ], [ 657.0625 , 1.6854007266152515e-46 ], [ 657.125 , 1.6856360667846578e-46 ], [ 657.1875 , 1.6858714398156946e-46 ], 
                    [ 657.25 , 1.686106845712938e-46 ], [ 657.3125 , 1.6863422844809776e-46 ], [ 657.375 , 1.6865777561243915e-46 ], [ 657.4375 , 1.6868132606477938e-46 ], 
                    [ 657.5 , 1.687048798055764e-46 ], [ 657.5625 , 1.6872843683528816e-46 ], [ 657.625 , 1.6875199715437635e-46 ], [ 657.6875 , 1.6877556076329905e-46 ], 
                    [ 657.75 , 1.6879912766251562e-46 ], [ 657.8125 , 1.6882269785248557e-46 ], [ 657.875 , 1.6884627133366711e-46 ], [ 657.9375 , 1.6886984810652105e-46 ], 
                    [ 658.0 , 1.6889342817150825e-46 ], [ 658.0625 , 1.6891701152908717e-46 ], [ 658.125 , 1.6894059817971634e-46 ], [ 658.1875 , 1.6896418812385927e-46 ], 
                    [ 658.25 , 1.6898778136197338e-46 ], [ 658.3125 , 1.6901137789451624e-46 ], [ 658.375 , 1.6903497772195026e-46 ], [ 658.4375 , 1.6905858084473793e-46 ], 
                    [ 658.5 , 1.6908217553118134e-46 ], [ 658.5625 , 1.6910576134083327e-46 ], [ 658.625 , 1.6912935044054592e-46 ], [ 658.6875 , 1.691529428307758e-46 ], 
                    [ 658.75 , 1.6917653851198312e-46 ], [ 658.8125 , 1.6920013748462815e-46 ], [ 658.875 , 1.6922373974916881e-46 ], [ 658.9375 , 1.6924734530606312e-46 ], 
                    [ 659.0 , 1.6927095415577033e-46 ], [ 659.0625 , 1.6929456629875215e-46 ], [ 659.125 , 1.693181817354656e-46 ], [ 659.1875 , 1.6934180046637492e-46 ], 
                    [ 659.25 , 1.6936542249193361e-46 ], [ 659.3125 , 1.6938904781260367e-46 ], [ 659.375 , 1.6941267642884476e-46 ], [ 659.4375 , 1.6943630834111655e-46 ], 
                    [ 659.5 , 1.6945994354988125e-46 ], [ 659.5625 , 1.6948358205559147e-46 ], [ 659.625 , 1.6950722385871555e-46 ], [ 659.6875 , 1.695308689597098e-46 ], 
                    [ 659.75 , 1.6955451735903189e-46 ], [ 659.8125 , 1.6957816905714791e-46 ], [ 659.875 , 1.6960182405451203e-46 ], [ 659.9375 , 1.6962548235158685e-46 ], 
                    [ 660.0 , 1.6964912924359185e-46 ], [ 660.0625 , 1.6967277005121784e-46 ], [ 660.125 , 1.6969641415322088e-46 ], [ 660.1875 , 1.6972006155005882e-46 ], 
                    [ 660.25 , 1.6974371224218716e-46 ], [ 660.3125 , 1.6976736623006755e-46 ], [ 660.375 , 1.6979102351415922e-46 ], [ 660.4375 , 1.6981468409492396e-46 ], 
                    [ 660.5 , 1.6983834797281753e-46 ], [ 660.5625 , 1.6986201514829936e-46 ], [ 660.625 , 1.698856856218302e-46 ], [ 660.6875 , 1.6990935939387087e-46 ], 
                    [ 660.75 , 1.6993303646487976e-46 ], [ 660.8125 , 1.699567168353142e-46 ], [ 660.875 , 1.699804005056388e-46 ], [ 660.9375 , 1.7000408747631098e-46 ], 
                    [ 661.0 , 1.7002777774778944e-46 ], [ 661.0625 , 1.7005147132053782e-46 ], [ 661.125 , 1.7007516819501003e-46 ], [ 661.1875 , 1.7009886837167107e-46 ], 
                    [ 661.25 , 1.7012257185098106e-46 ], [ 661.3125 , 1.7014627863339907e-46 ], [ 661.375 , 1.7016998871938535e-46 ], [ 661.4375 , 1.7019370210939787e-46 ], 
                    [ 661.5 , 1.7021740228733216e-46 ], [ 661.5625 , 1.7024109785406657e-46 ], [ 661.625 , 1.7026479671940664e-46 ], [ 661.6875 , 1.7028849888380795e-46 ], 
                    [ 661.75 , 1.7031220434773457e-46 ], [ 661.8125 , 1.7033591311164224e-46 ], [ 661.875 , 1.7035962517599149e-46 ], [ 661.9375 , 1.7038334054124182e-46 ], 
                    [ 662.0 , 1.7040705920785272e-46 ], [ 662.0625 , 1.7043078117628375e-46 ], [ 662.125 , 1.7045450644699576e-46 ], [ 662.1875 , 1.7047823502044487e-46 ], 
                    [ 662.25 , 1.7050196689709562e-46 ], [ 662.3125 , 1.7052570207740547e-46 ], [ 662.375 , 1.705494405618343e-46 ], [ 662.4375 , 1.7057318235084083e-46 ], 
                    [ 662.5 , 1.7059692744488512e-46 ], [ 662.5625 , 1.7062067584442966e-46 ], [ 662.625 , 1.7064442754993463e-46 ], [ 662.6875 , 1.7066818256185775e-46 ], 
                    [ 662.75 , 1.706919408806618e-46 ], [ 662.8125 , 1.7071570250680468e-46 ], [ 662.875 , 1.707394674407492e-46 ], [ 662.9375 , 1.7076323568295344e-46 ], 
                    [ 663.0 , 1.7078698997654967e-46 ], [ 663.0625 , 1.708107397299277e-46 ], [ 663.125 , 1.7083449278595997e-46 ], [ 663.1875 , 1.708582491451093e-46 ], 
                    [ 663.25 , 1.7088200880783392e-46 ], [ 663.3125 , 1.7090577177459193e-46 ], [ 663.375 , 1.7092953804584405e-46 ], [ 663.4375 , 1.709533076220522e-46 ], 
                    [ 663.5 , 1.7097708050367358e-46 ], [ 663.5625 , 1.7100085669116537e-46 ], [ 663.625 , 1.7102463618498976e-46 ], [ 663.6875 , 1.7104841898560895e-46 ], 
                    [ 663.75 , 1.7107220509347788e-46 ], [ 663.8125 , 1.7109599450905895e-46 ], [ 663.875 , 1.7111978723281454e-46 ], [ 663.9375 , 1.7114358326519981e-46 ], 
                    [ 664.0 , 1.711673826066773e-46 ], [ 664.0625 , 1.711911852577084e-46 ], [ 664.125 , 1.7121499121875214e-46 ], [ 664.1875 , 1.7123880049026755e-46 ], 
                    [ 664.25 , 1.7126261307271624e-46 ], [ 664.3125 , 1.7128642896656227e-46 ], [ 664.375 , 1.713102481722588e-46 ], [ 664.4375 , 1.7133407069027126e-46 ], 
                    [ 664.5 , 1.7135788055080817e-46 ], [ 664.5625 , 1.7138168491526347e-46 ], [ 664.625 , 1.714054925865289e-46 ], [ 664.6875 , 1.714293035650614e-46 ], 
                    [ 664.75 , 1.7145311785132162e-46 ], [ 664.8125 , 1.7147693544576905e-46 ], [ 664.875 , 1.7150075634886326e-46 ], [ 664.9375 , 1.7152458056106385e-46 ], 
                    [ 665.0 , 1.7154840808283056e-46 ], [ 665.0625 , 1.715722389146243e-46 ], [ 665.125 , 1.715960730569025e-46 ], [ 665.1875 , 1.716199105101262e-46 ], 
                    [ 665.25 , 1.716437512747554e-46 ], [ 665.3125 , 1.716675953512513e-46 ], [ 665.375 , 1.7169144274007154e-46 ], [ 665.4375 , 1.7171529344167866e-46 ], 
                    [ 665.5 , 1.7173914745652928e-46 ], [ 665.5625 , 1.7176300478508845e-46 ], [ 665.625 , 1.7178686542781048e-46 ], [ 665.6875 , 1.7181072938516423e-46 ], 
                    [ 665.75 , 1.7183459665760289e-46 ], [ 665.8125 , 1.718584672455906e-46 ], [ 665.875 , 1.7188234114958308e-46 ], [ 665.9375 , 1.719062183700495e-46 ], 
                    [ 666.0 , 1.7193008542752804e-46 ], [ 666.0625 , 1.7195394463546577e-46 ], [ 666.125 , 1.7197780715441347e-46 ], [ 666.1875 , 1.7200167298482693e-46 ], 
                    [ 666.25 , 1.7202554212716568e-46 ], [ 666.3125 , 1.72049414581893e-46 ], [ 666.375 , 1.720732903494637e-46 ], [ 666.4375 , 1.7209716943033993e-46 ], 
                    [ 666.5 , 1.7212105182498392e-46 ], [ 666.5625 , 1.7214493753385311e-46 ], [ 666.625 , 1.7216882655740498e-46 ], [ 666.6875 , 1.7219271889610314e-46 ], 
                    [ 666.75 , 1.7221661455040648e-46 ], [ 666.8125 , 1.7224051352077385e-46 ], [ 666.875 , 1.722644158076679e-46 ], [ 666.9375 , 1.7228832141154762e-46 ], 
                    [ 667.0 , 1.7231223033287337e-46 ], [ 667.0625 , 1.7233614257210546e-46 ], [ 667.125 , 1.7236005812970438e-46 ], [ 667.1875 , 1.7238397700612935e-46 ], 
                    [ 667.25 , 1.7240789920184465e-46 ], [ 667.3125 , 1.7243182471730722e-46 ], [ 667.375 , 1.7245575355297899e-46 ], [ 667.4375 , 1.7247968570932189e-46 ], 
                    [ 667.5 , 1.7250361104199725e-46 ], [ 667.5625 , 1.7252752415566773e-46 ], [ 667.625 , 1.7255144058426597e-46 ], [ 667.6875 , 1.7257536032825144e-46 ], 
                    [ 667.75 , 1.7259928338808134e-46 ], [ 667.8125 , 1.726232097642202e-46 ], [ 667.875 , 1.726471394571253e-46 ], [ 667.9375 , 1.7267107246725519e-46 ], 
                    [ 668.0 , 1.7269500879507218e-46 ], [ 668.0625 , 1.727189484410337e-46 ], [ 668.125 , 1.7274289140560342e-46 ], [ 668.1875 , 1.7276683768923526e-46 ], 
                    [ 668.25 , 1.7279078729239663e-46 ], [ 668.3125 , 1.7281474021554165e-46 ], [ 668.375 , 1.7283869645913169e-46 ], [ 668.4375 , 1.72862656023632e-46 ], 
                    [ 668.5 , 1.7288661890949803e-46 ], [ 668.5625 , 1.7291058511718893e-46 ], [ 668.625 , 1.729345546471689e-46 ], [ 668.6875 , 1.729585274998985e-46 ], 
                    [ 668.75 , 1.7298250367583832e-46 ], [ 668.8125 , 1.7300648317544533e-46 ], [ 668.875 , 1.7303046599918279e-46 ], [ 668.9375 , 1.7305445214751267e-46 ], 
                    [ 669.0 , 1.730784369186456e-46 ], [ 669.0625 , 1.7310240472273767e-46 ], [ 669.125 , 1.731263758458777e-46 ], [ 669.1875 , 1.7315035028852528e-46 ], 
                    [ 669.25 , 1.7317432805113892e-46 ], [ 669.3125 , 1.7319830913418202e-46 ], [ 669.375 , 1.7322229353811072e-46 ], [ 669.4375 , 1.7324628126338488e-46 ], 
                    [ 669.5 , 1.7327027231046691e-46 ], [ 669.5625 , 1.732942666798156e-46 ], [ 669.625 , 1.7331826437189096e-46 ], [ 669.6875 , 1.733422653871507e-46 ], 
                    [ 669.75 , 1.733662697260636e-46 ], [ 669.8125 , 1.7339027738907889e-46 ], [ 669.875 , 1.7341428837666177e-46 ], [ 669.9375 , 1.7343830268927383e-46 ], 
                    [ 670.0 , 1.7346232032737435e-46 ], [ 670.0625 , 1.7348634129142258e-46 ], [ 670.125 , 1.7351036558188281e-46 ], [ 670.1875 , 1.735343931992095e-46 ], 
                    [ 670.25 , 1.7355842414387076e-46 ], [ 670.3125 , 1.735824584163224e-46 ], [ 670.375 , 1.7360649601702528e-46 ], [ 670.4375 , 1.7363053694644395e-46 ], 
                    [ 670.5 , 1.7365458120503573e-46 ], [ 670.5625 , 1.7367860511538894e-46 ], [ 670.625 , 1.7370263046787716e-46 ], [ 670.6875 , 1.737266591438477e-46 ], 
                    [ 670.75 , 1.7375069114376033e-46 ], [ 670.8125 , 1.7377472646807238e-46 ], [ 670.875 , 1.7379876511724744e-46 ], [ 670.9375 , 1.7382280709173925e-46 ], 
                    [ 671.0 , 1.73846852392014e-46 ], [ 671.0625 , 1.7387090101852929e-46 ], [ 671.125 , 1.738949529717452e-46 ], [ 671.1875 , 1.7391900825212072e-46 ], 
                    [ 671.25 , 1.7394306686011734e-46 ], [ 671.3125 , 1.7396712879619413e-46 ], [ 671.375 , 1.7399119406081646e-46 ], [ 671.4375 , 1.7401526265443854e-46 ], 
                    [ 671.5 , 1.7403933457752458e-46 ], [ 671.5625 , 1.7406340983053272e-46 ], [ 671.625 , 1.7408748841392481e-46 ], [ 671.6875 , 1.7411157032816155e-46 ], 
                    [ 671.75 , 1.741356555737062e-46 ], [ 671.8125 , 1.7415974415101459e-46 ], [ 671.875 , 1.7418383606055014e-46 ], [ 671.9375 , 1.7420793130277501e-46 ], 
                    [ 672.0 , 1.74232029878149e-46 ], [ 672.0625 , 1.7425611649474135e-46 ], [ 672.125 , 1.7428019665285185e-46 ], [ 672.1875 , 1.7430428013856116e-46 ], 
                    [ 672.25 , 1.743283669523254e-46 ], [ 672.3125 , 1.743524570946057e-46 ], [ 672.375 , 1.743765505658608e-46 ], [ 672.4375 , 1.7440064736655321e-46 ], 
                    [ 672.5 , 1.7442474749714298e-46 ], [ 672.5625 , 1.7444885095808908e-46 ], [ 672.625 , 1.7447295774984919e-46 ], [ 672.6875 , 1.7449706787288857e-46 ], 
                    [ 672.75 , 1.7452118132766264e-46 ], [ 672.8125 , 1.745452981146392e-46 ], [ 672.875 , 1.745694182342688e-46 ], [ 672.9375 , 1.7459354168701696e-46 ], 
                    [ 673.0 , 1.7461766847334428e-46 ], [ 673.0625 , 1.7464179859371264e-46 ], [ 673.125 , 1.7466593204858152e-46 ], [ 673.1875 , 1.7469006883841423e-46 ], 
                    [ 673.25 , 1.7471420896366538e-46 ], [ 673.3125 , 1.7473835242480086e-46 ], [ 673.375 , 1.7476249922228168e-46 ], [ 673.4375 , 1.747866493565676e-46 ], 
                    [ 673.5 , 1.7481080282811604e-46 ], [ 673.5625 , 1.748349533560128e-46 ], [ 673.625 , 1.7485908709771934e-46 ], [ 673.6875 , 1.7488322417078075e-46 ], 
                    [ 673.75 , 1.7490736457565936e-46 ], [ 673.8125 , 1.7493150831281384e-46 ], [ 673.875 , 1.7495565538270416e-46 ], [ 673.9375 , 1.7497980578579036e-46 ], 
                    [ 674.0 , 1.7500395952253258e-46 ], [ 674.0625 , 1.7502811659339094e-46 ], [ 674.125 , 1.750522769988257e-46 ], [ 674.1875 , 1.7507644073929838e-46 ], 
                    [ 674.25 , 1.7510060781526564e-46 ], [ 674.3125 , 1.7512477822719037e-46 ], [ 674.375 , 1.7514895197553557e-46 ], [ 674.4375 , 1.7517312906076047e-46 ], 
                    [ 674.5 , 1.7519730948332456e-46 ], [ 674.5625 , 1.7522149324368345e-46 ], [ 674.625 , 1.7524568034230285e-46 ], [ 674.6875 , 1.7526987077964862e-46 ], 
                    [ 674.75 , 1.7529406455617404e-46 ], [ 674.8125 , 1.7531826167234393e-46 ], [ 674.875 , 1.753424621286142e-46 ], [ 674.9375 , 1.753666659254534e-46 ], 
                    [ 675.0 , 1.753908730633189e-46 ], [ 675.0625 , 1.754150835426695e-46 ], [ 675.125 , 1.754392762471589e-46 ], [ 675.1875 , 1.7546346735738026e-46 ], 
                    [ 675.25 , 1.754876618032834e-46 ], [ 675.3125 , 1.7551185958533077e-46 ], [ 675.375 , 1.7553606070397983e-46 ], [ 675.4375 , 1.7556026515969202e-46 ], 
                    [ 675.5 , 1.7558447295292737e-46 ], [ 675.5625 , 1.756086840841474e-46 ], [ 675.625 , 1.756328985538086e-46 ], [ 675.6875 , 1.756571163623763e-46 ], 
                    [ 675.75 , 1.7568133751030723e-46 ], [ 675.8125 , 1.757055619980642e-46 ], [ 675.875 , 1.7572978982610542e-46 ], [ 675.9375 , 1.7575402099489257e-46 ], 
                    [ 676.0 , 1.7577825550488512e-46 ], [ 676.0625 , 1.758024933565463e-46 ], [ 676.125 , 1.7582673455033934e-46 ], [ 676.1875 , 1.758509790867176e-46 ], 
                    [ 676.25 , 1.7587522696614324e-46 ], [ 676.3125 , 1.7589947818908478e-46 ], [ 676.375 , 1.7592373275599563e-46 ], [ 676.4375 , 1.759479906673395e-46 ], 
                    [ 676.5 , 1.7597225192357752e-46 ], [ 676.5625 , 1.7599651652516965e-46 ], [ 676.625 , 1.760207759891009e-46 ], [ 676.6875 , 1.7604502143899775e-46 ], 
                    [ 676.75 , 1.7606927022851084e-46 ], [ 676.8125 , 1.760935223581002e-46 ], [ 676.875 , 1.7611777782822582e-46 ], [ 676.9375 , 1.761420366393479e-46 ], 
                    [ 677.0 , 1.7616629879192663e-46 ], [ 677.0625 , 1.761905642864235e-46 ], [ 677.125 , 1.7621483312329503e-46 ], [ 677.1875 , 1.762391053030055e-46 ], 
                    [ 677.25 , 1.7626338082601518e-46 ], [ 677.3125 , 1.7628765969278225e-46 ], [ 677.375 , 1.7631194190376965e-46 ], [ 677.4375 , 1.7633622745943438e-46 ], 
                    [ 677.5 , 1.7636051636024204e-46 ], [ 677.5625 , 1.7638480860665095e-46 ], [ 677.625 , 1.764091041991195e-46 ], [ 677.6875 , 1.764334031381122e-46 ], 
                    [ 677.75 , 1.764577054240901e-46 ], [ 677.8125 , 1.764820110575092e-46 ], [ 677.875 , 1.765063200388393e-46 ], [ 677.9375 , 1.7653063236853286e-46 ], 
                    [ 678.0 , 1.7655494804705857e-46 ], [ 678.0625 , 1.7657926707487143e-46 ], [ 678.125 , 1.7660358945243404e-46 ], [ 678.1875 , 1.7662789448316183e-46 ], 
                    [ 678.25 , 1.766521973696173e-46 ], [ 678.3125 , 1.7667650359999828e-46 ], [ 678.375 , 1.7670081317476354e-46 ], [ 678.4375 , 1.767251260943745e-46 ], 
                    [ 678.5 , 1.767494423592902e-46 ], [ 678.5625 , 1.7677376196996833e-46 ], [ 678.625 , 1.7679808492687434e-46 ], [ 678.6875 , 1.7682241123046486e-46 ], 
                    [ 678.75 , 1.7684674088120163e-46 ], [ 678.8125 , 1.768710738795464e-46 ], [ 678.875 , 1.768954102259561e-46 ], [ 678.9375 , 1.7691974992089513e-46 ], 
                    [ 679.0 , 1.7694409296482296e-46 ], [ 679.0625 , 1.7696843935820167e-46 ], [ 679.125 , 1.769927891014896e-46 ], [ 679.1875 , 1.770171421951502e-46 ], 
                    [ 679.25 , 1.7704149863964192e-46 ], [ 679.3125 , 1.7706585843542584e-46 ], [ 679.375 , 1.7709022158296682e-46 ], [ 679.4375 , 1.7711458808272226e-46 ], 
                    [ 679.5 , 1.7713895793515724e-46 ], [ 679.5625 , 1.77163331140728e-46 ], [ 679.625 , 1.7718770769989716e-46 ], [ 679.6875 , 1.772120823807987e-46 ], 
                    [ 679.75 , 1.772364390571303e-46 ], [ 679.8125 , 1.7726079908113374e-46 ], [ 679.875 , 1.772851624532654e-46 ], [ 679.9375 , 1.7730952917398923e-46 ], 
                    [ 680.0 , 1.7733389924376422e-46 ], [ 680.0625 , 1.773582726630519e-46 ], [ 680.125 , 1.7738264943231147e-46 ], [ 680.1875 , 1.7740702955200076e-46 ], 
                    [ 680.25 , 1.7743141302258535e-46 ], [ 680.3125 , 1.77455799844522e-46 ], [ 680.375 , 1.7748019001827384e-46 ], [ 680.4375 , 1.775045835442978e-46 ], 
                    [ 680.5 , 1.7752898042305966e-46 ], [ 680.5625 , 1.7755338065501518e-46 ], [ 680.625 , 1.775777842406303e-46 ], [ 680.6875 , 1.776021911803634e-46 ], 
                    [ 680.75 , 1.7762660147467173e-46 ], [ 680.8125 , 1.7765101512402646e-46 ], [ 680.875 , 1.776754321288786e-46 ], [ 680.9375 , 1.7769985248969182e-46 ], 
                    [ 681.0 , 1.7772427620693128e-46 ], [ 681.0625 , 1.7774870328105316e-46 ], [ 681.125 , 1.777731337125239e-46 ], [ 681.1875 , 1.7779756750180246e-46 ], 
                    [ 681.25 , 1.7782199002336903e-46 ], [ 681.3125 , 1.7784640381082655e-46 ], [ 681.375 , 1.778708209501349e-46 ], [ 681.4375 , 1.7789524144175817e-46 ], 
                    [ 681.5 , 1.7791966528615273e-46 ], [ 681.5625 , 1.7794409248377894e-46 ], [ 681.625 , 1.7796852303510095e-46 ], [ 681.6875 , 1.7799295694057406e-46 ], 
                    [ 681.75 , 1.7801739420066526e-46 ], [ 681.8125 , 1.7804183481582743e-46 ], [ 681.875 , 1.780662787865263e-46 ], [ 681.9375 , 1.7809072611321865e-46 ], 
                    [ 682.0 , 1.7811517679637173e-46 ], [ 682.0625 , 1.7813963083644247e-46 ], [ 682.125 , 1.7816408823388922e-46 ], [ 682.1875 , 1.7818854898917677e-46 ], 
                    [ 682.25 , 1.782130131027648e-46 ], [ 682.3125 , 1.7823748057511574e-46 ], [ 682.375 , 1.782619514066843e-46 ], [ 682.4375 , 1.782864255979431e-46 ], 
                    [ 682.5 , 1.7831090314934457e-46 ], [ 682.5625 , 1.783353840613513e-46 ], [ 682.625 , 1.783598683344272e-46 ], [ 682.6875 , 1.7838435596903114e-46 ], 
                    [ 682.75 , 1.784088469656285e-46 ], [ 682.8125 , 1.7843331880868944e-46 ], [ 682.875 , 1.784577897955123e-46 ], [ 682.9375 , 1.784822641383755e-46 ], 
                    [ 683.0 , 1.785067418377381e-46 ], [ 683.0625 , 1.7853122289406034e-46 ], [ 683.125 , 1.7855570730780396e-46 ], [ 683.1875 , 1.785801950794243e-46 ], 
                    [ 683.25 , 1.786046862093882e-46 ], [ 683.3125 , 1.7862918069815373e-46 ], [ 683.375 , 1.7865367854618023e-46 ], [ 683.4375 , 1.7867817975393224e-46 ], 
                    [ 683.5 , 1.7870268432186665e-46 ], [ 683.5625 , 1.7872719225044564e-46 ], [ 683.625 , 1.7875170354013007e-46 ], [ 683.6875 , 1.7877621819138085e-46 ], 
                    [ 683.75 , 1.788007362046565e-46 ], [ 683.8125 , 1.788252575804245e-46 ], [ 683.875 , 1.7884978231914208e-46 ], [ 683.9375 , 1.7887431042126925e-46 ], 
                    [ 684.0 , 1.7889884188726593e-46 ], [ 684.0625 , 1.789233767175999e-46 ], [ 684.125 , 1.7894791491272616e-46 ], [ 684.1875 , 1.7897245647311127e-46 ], 
                    [ 684.25 , 1.7899700139921295e-46 ], [ 684.3125 , 1.7902154752120123e-46 ], [ 684.375 , 1.7904607242209e-46 ], [ 684.4375 , 1.7907060068274614e-46 ], 
                    [ 684.5 , 1.7909513230362872e-46 ], [ 684.5625 , 1.7911966728519933e-46 ], [ 684.625 , 1.7914420562791963e-46 ], [ 684.6875 , 1.791687473322488e-46 ], 
                    [ 684.75 , 1.7919329239864485e-46 ], [ 684.8125 , 1.7921784082757468e-46 ], [ 684.875 , 1.792423926194913e-46 ], [ 684.9375 , 1.792669477748618e-46 ], 
                    [ 685.0 , 1.7929150629414446e-46 ], [ 685.0625 , 1.7931606817779747e-46 ], [ 685.125 , 1.793406334262869e-46 ], [ 685.1875 , 1.7936520204006985e-46 ], 
                    [ 685.25 , 1.7938977401960735e-46 ], [ 685.3125 , 1.7941434936536436e-46 ], [ 685.375 , 1.794389280778007e-46 ], [ 685.4375 , 1.7946351015737504e-46 ], 
                    [ 685.5 , 1.794880956045474e-46 ], [ 685.5625 , 1.7951268441978298e-46 ], [ 685.625 , 1.7953727660354568e-46 ], [ 685.6875 , 1.7956187215628678e-46 ], 
                    [ 685.75 , 1.7958647107847674e-46 ], [ 685.8125 , 1.796110733705746e-46 ], [ 685.875 , 1.7963567181268478e-46 ], [ 685.9375 , 1.7966025347481524e-46 ], 
                    [ 686.0 , 1.7968483850074364e-46 ], [ 686.0625 , 1.7970942689093168e-46 ], [ 686.125 , 1.797340186458383e-46 ], [ 686.1875 , 1.7975861376592155e-46 ], 
                    [ 686.25 , 1.797832122516457e-46 ], [ 686.3125 , 1.7980781410346873e-46 ], [ 686.375 , 1.7983241932185635e-46 ], [ 686.4375 , 1.798570279072629e-46 ], 
                    [ 686.5 , 1.798816398601517e-46 ], [ 686.5625 , 1.799062551809848e-46 ], [ 686.625 , 1.799308738702218e-46 ], [ 686.6875 , 1.799554959283198e-46 ], 
                    [ 686.75 , 1.7998012135574754e-46 ], [ 686.8125 , 1.8000475015296348e-46 ], [ 686.875 , 1.8002938232042486e-46 ], [ 686.9375 , 1.800540178585942e-46 ], 
                    [ 687.0 , 1.8007865676793662e-46 ], [ 687.0625 , 1.8010329904891214e-46 ], [ 687.125 , 1.8012794470197698e-46 ], [ 687.1875 , 1.8015259372759647e-46 ], 
                    [ 687.25 , 1.8017724612623333e-46 ], [ 687.3125 , 1.802019018983479e-46 ], [ 687.375 , 1.8022656104440445e-46 ], [ 687.4375 , 1.802512128977928e-46 ], 
                    [ 687.5 , 1.8027585145328742e-46 ], [ 687.5625 , 1.80300493376626e-46 ], [ 687.625 , 1.80325138668274e-46 ], [ 687.6875 , 1.803497873286854e-46 ], 
                    [ 687.75 , 1.803744393583259e-46 ], [ 687.8125 , 1.8039909475765717e-46 ], [ 687.875 , 1.8042375352713606e-46 ], [ 687.9375 , 1.804484156672232e-46 ], 
                    [ 688.0 , 1.8047308117838194e-46 ], [ 688.0625 , 1.8049775006107175e-46 ], [ 688.125 , 1.8052242231575222e-46 ], [ 688.1875 , 1.805470979428855e-46 ], 
                    [ 688.25 , 1.805717769429352e-46 ], [ 688.3125 , 1.805964593163585e-46 ], [ 688.375 , 1.806211450636152e-46 ], [ 688.4375 , 1.806458341851716e-46 ], 
                    [ 688.5 , 1.8067052668148638e-46 ], [ 688.5625 , 1.8069522255301828e-46 ], [ 688.625 , 1.8071992180023637e-46 ], [ 688.6875 , 1.8074462442359175e-46 ], 
                    [ 688.75 , 1.8076933042355365e-46 ], [ 688.8125 , 1.807940398005811e-46 ], [ 688.875 , 1.808187525551356e-46 ], [ 688.9375 , 1.8084346868767883e-46 ], 
                    [ 689.0 , 1.8086817558123267e-46 ], [ 689.0625 , 1.808928710555221e-46 ], [ 689.125 , 1.8091756990169384e-46 ], [ 689.1875 , 1.8094227212020967e-46 ], 
                    [ 689.25 , 1.8096697771153133e-46 ], [ 689.3125 , 1.8099168667611803e-46 ], [ 689.375 , 1.8101639901442912e-46 ], [ 689.4375 , 1.8104111472692642e-46 ], 
                    [ 689.5 , 1.8106583381407206e-46 ], [ 689.5625 , 1.810905562763241e-46 ], [ 689.625 , 1.81115282114146e-46 ], [ 689.6875 , 1.8114001132799613e-46 ], 
                    [ 689.75 , 1.81164743918338e-46 ], [ 689.8125 , 1.8118947988563e-46 ], [ 689.875 , 1.8121421923033586e-46 ], [ 689.9375 , 1.812389619529141e-46 ], 
                    [ 690.0 , 1.812637080538298e-46 ], [ 690.0625 , 1.8128845753353776e-46 ], [ 690.125 , 1.8131321039250718e-46 ], [ 690.1875 , 1.8133796663119163e-46 ], 
                    [ 690.25 , 1.8136272625005896e-46 ], [ 690.3125 , 1.8138748924956563e-46 ], [ 690.375 , 1.814122556301797e-46 ], [ 690.4375 , 1.8143702539235756e-46 ], 
                    [ 690.5 , 1.8146179853656225e-46 ], [ 690.5625 , 1.814865620123189e-46 ], [ 690.625 , 1.8151131437815927e-46 ], [ 690.6875 , 1.815360701198923e-46 ], 
                    [ 690.75 , 1.815608292379822e-46 ], [ 690.8125 , 1.8158559173288826e-46 ], [ 690.875 , 1.81610357605071e-46 ], [ 690.9375 , 1.8163512685499363e-46 ], 
                    [ 691.0 , 1.8165989948311032e-46 ], [ 691.0625 , 1.8168467548988703e-46 ], [ 691.125 , 1.8170945487578453e-46 ], [ 691.1875 , 1.8173423764125857e-46 ], 
                    [ 691.25 , 1.817590237867791e-46 ], [ 691.3125 , 1.8178381331279935e-46 ], [ 691.375 , 1.8180860621978293e-46 ], [ 691.4375 , 1.818334025081897e-46 ], 
                    [ 691.5 , 1.8185820217848602e-46 ], [ 691.5625 , 1.81883005231128e-46 ], [ 691.625 , 1.8190781166657425e-46 ], [ 691.6875 , 1.819326214852953e-46 ], 
                    [ 691.75 , 1.8195743468774346e-46 ], [ 691.8125 , 1.8198225127438543e-46 ], [ 691.875 , 1.820070712456802e-46 ], [ 691.9375 , 1.8203189460209325e-46 ], 
                    [ 692.0 , 1.8205672134408497e-46 ], [ 692.0625 , 1.8208155147211323e-46 ], [ 692.125 , 1.8210637292544302e-46 ], [ 692.1875 , 1.821311818294618e-46 ], 
                    [ 692.25 , 1.8215599411327674e-46 ], [ 692.3125 , 1.821808097773406e-46 ], [ 692.375 , 1.8220562882211894e-46 ], [ 692.4375 , 1.822304512480724e-46 ], 
                    [ 692.5 , 1.8225527705565906e-46 ], [ 692.5625 , 1.822801062453421e-46 ], [ 692.625 , 1.823049388175771e-46 ], [ 692.6875 , 1.8232977477283148e-46 ], 
                    [ 692.75 , 1.823546141115647e-46 ], [ 692.8125 , 1.8237945683423385e-46 ], [ 692.875 , 1.824043029413064e-46 ], [ 692.9375 , 1.82429152433237e-46 ], 
                    [ 693.0 , 1.8245400531049056e-46 ], [ 693.0625 , 1.8247886157352448e-46 ], [ 693.125 , 1.8250372122280388e-46 ], [ 693.1875 , 1.8252858425879003e-46 ], 
                    [ 693.25 , 1.825534506819405e-46 ], [ 693.3125 , 1.8257832049272052e-46 ], [ 693.375 , 1.826031936915917e-46 ], [ 693.4375 , 1.82628070279013e-46 ], 
                    [ 693.5 , 1.8265295025544726e-46 ], [ 693.5625 , 1.8267783362135758e-46 ], [ 693.625 , 1.8270272037720307e-46 ], [ 693.6875 , 1.8272760133961472e-46 ], 
                    [ 693.75 , 1.8275246712933594e-46 ], [ 693.8125 , 1.8277733630282218e-46 ], [ 693.875 , 1.8280220886053647e-46 ], [ 693.9375 , 1.8282708480293675e-46 ], 
                    [ 694.0 , 1.828519641304836e-46 ], [ 694.0625 , 1.828768468436403e-46 ], [ 694.125 , 1.8290173294286757e-46 ], [ 694.1875 , 1.8292662242862098e-46 ], 
                    [ 694.25 , 1.829515153013691e-46 ], [ 694.3125 , 1.8297641156156906e-46 ], [ 694.375 , 1.8300131120967918e-46 ], [ 694.4375 , 1.8302621424616437e-46 ], 
                    [ 694.5 , 1.8305112067148444e-46 ], [ 694.5625 , 1.8307603048609922e-46 ], [ 694.625 , 1.831009436904713e-46 ], [ 694.6875 , 1.831258602850645e-46 ], 
                    [ 694.75 , 1.8315078027033624e-46 ], [ 694.8125 , 1.83175703646748e-46 ], [ 694.875 , 1.8320063041476516e-46 ], [ 694.9375 , 1.8322556057484394e-46 ], 
                    [ 695.0 , 1.8325049412744995e-46 ], [ 695.0625 , 1.832754310730448e-46 ], [ 695.125 , 1.8330037141208763e-46 ], [ 695.1875 , 1.833253151450415e-46 ], 
                    [ 695.25 , 1.833502574901984e-46 ], [ 695.3125 , 1.8337518056210393e-46 ], [ 695.375 , 1.8340010702184213e-46 ], [ 695.4375 , 1.8342503686986695e-46 ], 
                    [ 695.5 , 1.8344997010664428e-46 ], [ 695.5625 , 1.8347490673263464e-46 ], [ 695.625 , 1.834998467482936e-46 ], [ 695.6875 , 1.835247901540897e-46 ], 
                    [ 695.75 , 1.835497369504786e-46 ], [ 695.8125 , 1.8357468713792245e-46 ], [ 695.875 , 1.835996407168835e-46 ], [ 695.9375 , 1.8362459768781887e-46 ], 
                    [ 696.0 , 1.8364955805119356e-46 ], [ 696.0625 , 1.836745218074674e-46 ], [ 696.125 , 1.8369948895710287e-46 ], [ 696.1875 , 1.8372445950055863e-46 ], 
                    [ 696.25 , 1.8374943343829864e-46 ], [ 696.3125 , 1.837744107707791e-46 ], [ 696.375 , 1.837993914984692e-46 ], [ 696.4375 , 1.83824375621824e-46 ], 
                    [ 696.5 , 1.8384936314130898e-46 ], [ 696.5625 , 1.8387435405738445e-46 ], [ 696.625 , 1.8389934837051346e-46 ], [ 696.6875 , 1.839243460811551e-46 ], 
                    [ 696.75 , 1.8394934718977393e-46 ], [ 696.8125 , 1.839743516968278e-46 ], [ 696.875 , 1.8399933140280825e-46 ], [ 696.9375 , 1.840243133400107e-46 ], 
                    [ 697.0 , 1.8404929866905934e-46 ], [ 697.0625 , 1.840742873904133e-46 ], [ 697.125 , 1.8409927950453325e-46 ], [ 697.1875 , 1.841242750118785e-46 ], 
                    [ 697.25 , 1.8414927391291237e-46 ], [ 697.3125 , 1.841742762080956e-46 ], [ 697.375 , 1.841992818978838e-46 ], [ 697.4375 , 1.842242909827444e-46 ], 
                    [ 697.5 , 1.842493034631345e-46 ], [ 697.5625 , 1.842743193395163e-46 ], [ 697.625 , 1.842993386123523e-46 ], [ 697.6875 , 1.8432436128210357e-46 ], 
                    [ 697.75 , 1.8434938734922743e-46 ], [ 697.8125 , 1.8437441681418775e-46 ], [ 697.875 , 1.843994496774498e-46 ], [ 697.9375 , 1.8442448593946708e-46 ], 
                    [ 698.0 , 1.844495256007024e-46 ], [ 698.0625 , 1.8447456866162644e-46 ], [ 698.125 , 1.8449961512269026e-46 ], [ 698.1875 , 1.845246649843595e-46 ], 
                    [ 698.25 , 1.8454971824709844e-46 ], [ 698.3125 , 1.8457477491136494e-46 ], [ 698.375 , 1.8459983497761948e-46 ], [ 698.4375 , 1.8462487904263658e-46 ], 
                    [ 698.5 , 1.846499177702017e-46 ], [ 698.5625 , 1.846749598935057e-46 ], [ 698.625 , 1.8470000541300777e-46 ], [ 698.6875 , 1.847250543291712e-46 ], 
                    [ 698.75 , 1.847501066424539e-46 ], [ 698.8125 , 1.8477516235331925e-46 ], [ 698.875 , 1.8480022146222547e-46 ], [ 698.9375 , 1.8482528396963728e-46 ], 
                    [ 699.0 , 1.848503498760117e-46 ], [ 699.0625 , 1.8487541918180827e-46 ], [ 699.125 , 1.849004918874947e-46 ], [ 699.1875 , 1.8492556799352804e-46 ], 
                    [ 699.25 , 1.8495064750036694e-46 ], [ 699.3125 , 1.8497573040847512e-46 ], [ 699.375 , 1.8500081671831654e-46 ], [ 699.4375 , 1.8502590643034603e-46 ], 
                    [ 699.5 , 1.8505099954503273e-46 ], [ 699.5625 , 1.85076096062833e-46 ], [ 699.625 , 1.8510119598420698e-46 ], [ 699.6875 , 1.851262993096228e-46 ], 
                    [ 699.75 , 1.8515140603953433e-46 ], [ 699.8125 , 1.8517651617441113e-46 ], [ 699.875 , 1.8520162971470834e-46 ], [ 699.9375 , 1.8522674666089313e-46 ], 
                    [ 700.0 , 1.8525185671797874e-46 ], [ 700.0625 , 1.852769519875433e-46 ], [ 700.125 , 1.8530205065665448e-46 ], [ 700.1875 , 1.8532715272577534e-46 ], 
                    [ 700.25 , 1.8535225819536384e-46 ], [ 700.3125 , 1.853773670658833e-46 ], [ 700.375 , 1.8540247933779176e-46 ], [ 700.4375 , 1.8542759501155396e-46 ], 
                    [ 700.5 , 1.854527140876268e-46 ], [ 700.5625 , 1.8547783656647245e-46 ], [ 700.625 , 1.8550296244855056e-46 ], [ 700.6875 , 1.855280917343235e-46 ], 
                    [ 700.75 , 1.85553224424255e-46 ], [ 700.8125 , 1.8557836051880087e-46 ], [ 700.875 , 1.856035000184263e-46 ], [ 700.9375 , 1.8562864292358865e-46 ], 
                    [ 701.0 , 1.8565378923475583e-46 ], [ 701.0625 , 1.8567893895238387e-46 ], [ 701.125 , 1.8570409207693566e-46 ], [ 701.1875 , 1.8572924860887138e-46 ], 
                    [ 701.25 , 1.8575440854865655e-46 ], [ 701.3125 , 1.857795718967462e-46 ], [ 701.375 , 1.858047386536087e-46 ], [ 701.4375 , 1.8582990881970446e-46 ], 
                    [ 701.5 , 1.858550823954913e-46 ], [ 701.5625 , 1.8588025938143256e-46 ], [ 701.625 , 1.8590541132268876e-46 ], [ 701.6875 , 1.8593056596272072e-46 ], 
                    [ 701.75 , 1.8595572400639845e-46 ], [ 701.8125 , 1.859808854541824e-46 ], [ 701.875 , 1.8600605030652935e-46 ], [ 701.9375 , 1.8603121856390382e-46 ], 
                    [ 702.0 , 1.8605639022676394e-46 ], [ 702.0625 , 1.8608156529557312e-46 ], [ 702.125 , 1.8610674377079222e-46 ], [ 702.1875 , 1.861319256528769e-46 ], 
                    [ 702.25 , 1.8615711094229607e-46 ], [ 702.3125 , 1.8618229963950412e-46 ], [ 702.375 , 1.862074917449675e-46 ], [ 702.4375 , 1.8623268725914196e-46 ], 
                    [ 702.5 , 1.8625788618249154e-46 ], [ 702.5625 , 1.862830885154788e-46 ], [ 702.625 , 1.863082942585637e-46 ], [ 702.6875 , 1.863335034122051e-46 ], 
                    [ 702.75 , 1.863587159768697e-46 ], [ 702.8125 , 1.863839319530178e-46 ], [ 702.875 , 1.8640915134110697e-46 ], [ 702.9375 , 1.864343741416015e-46 ], 
                    [ 703.0 , 1.8645960035496585e-46 ], [ 703.0625 , 1.8648482998165644e-46 ], [ 703.125 , 1.8651006302214048e-46 ], [ 703.1875 , 1.86535284504066e-46 ], 
                    [ 703.25 , 1.865604962655787e-46 ], [ 703.3125 , 1.8658571143466215e-46 ], [ 703.375 , 1.8661093001178218e-46 ], [ 703.4375 , 1.8663615199739815e-46 ], 
                    [ 703.5 , 1.86661377391972e-46 ], [ 703.5625 , 1.8668660619596186e-46 ], [ 703.625 , 1.867118384098286e-46 ], [ 703.6875 , 1.8673707403403424e-46 ], 
                    [ 703.75 , 1.8676231306904388e-46 ], [ 703.8125 , 1.8678755551531046e-46 ], [ 703.875 , 1.868128013733004e-46 ], [ 703.9375 , 1.8683805064347343e-46 ], 
                    [ 704.0 , 1.8686330332628946e-46 ], [ 704.0625 , 1.868885594222124e-46 ], [ 704.125 , 1.8691381893170356e-46 ], [ 704.1875 , 1.869390818552229e-46 ], 
                    [ 704.25 , 1.8696434819322796e-46 ], [ 704.3125 , 1.8698961794618683e-46 ], [ 704.375 , 1.8701489111455712e-46 ], [ 704.4375 , 1.8704016769880705e-46 ], 
                    [ 704.5 , 1.870654476993864e-46 ], [ 704.5625 , 1.8709073111676615e-46 ], [ 704.625 , 1.871160179514055e-46 ], [ 704.6875 , 1.8714130820376636e-46 ], 
                    [ 704.75 , 1.87166600913992e-46 ], [ 704.8125 , 1.8719186798902836e-46 ], [ 704.875 , 1.8721713847506495e-46 ], [ 704.9375 , 1.87242412372561e-46 ], 
                    [ 705.0 , 1.8726768968197837e-46 ], [ 705.0625 , 1.87292970403775e-46 ], [ 705.125 , 1.873182545384115e-46 ], [ 705.1875 , 1.873435420863513e-46 ], 
                    [ 705.25 , 1.873688330480539e-46 ], [ 705.3125 , 1.8739412742397876e-46 ], [ 705.375 , 1.8741942521459076e-46 ], [ 705.4375 , 1.874447264203456e-46 ], 
                    [ 705.5 , 1.8747003104170966e-46 ], [ 705.5625 , 1.874953390791387e-46 ], [ 705.625 , 1.875206505330978e-46 ], [ 705.6875 , 1.8754596540404565e-46 ], 
                    [ 705.75 , 1.8757128369244746e-46 ], [ 705.8125 , 1.8759660539875786e-46 ], [ 705.875 , 1.876219305234477e-46 ], [ 705.9375 , 1.8764725906696776e-46 ], 
                    [ 706.0 , 1.876725910297902e-46 ], [ 706.0625 , 1.8769792641236593e-46 ], [ 706.125 , 1.877232652151687e-46 ], [ 706.1875 , 1.877486074386482e-46 ], 
                    [ 706.25 , 1.877739530832728e-46 ], [ 706.3125 , 1.877993021495018e-46 ], [ 706.375 , 1.878246408787835e-46 ], [ 706.4375 , 1.8784996771221372e-46 ], 
                    [ 706.5 , 1.878752979607869e-46 ], [ 706.5625 , 1.8790063162497014e-46 ], [ 706.625 , 1.8792596870521736e-46 ], [ 706.6875 , 1.8795130920199186e-46 ], 
                    [ 706.75 , 1.879766531157557e-46 ], [ 706.8125 , 1.88002000446967e-46 ], [ 706.875 , 1.880273511960891e-46 ], [ 706.9375 , 1.8805270536358167e-46 ], 
                    [ 707.0 , 1.880780629499043e-46 ], [ 707.0625 , 1.8810342395551933e-46 ], [ 707.125 , 1.881287883808891e-46 ], [ 707.1875 , 1.881541562264735e-46 ], 
                    [ 707.25 , 1.88179527492731e-46 ], [ 707.3125 , 1.882049021801282e-46 ], [ 707.375 , 1.8823028028912374e-46 ], [ 707.4375 , 1.8825566182017904e-46 ], 
                    [ 707.5 , 1.882810467737528e-46 ], [ 707.5625 , 1.883064351503133e-46 ], [ 707.625 , 1.883318269503153e-46 ], [ 707.6875 , 1.883572221742245e-46 ], 
                    [ 707.75 , 1.8838262082250263e-46 ], [ 707.8125 , 1.8840802289560733e-46 ], [ 707.875 , 1.8843342839400185e-46 ], [ 707.9375 , 1.8845883731815203e-46 ], 
                    [ 708.0 , 1.8848422441706606e-46 ], [ 708.0625 , 1.8850961071491976e-46 ], [ 708.125 , 1.885350004319674e-46 ], [ 708.1875 , 1.885603935686695e-46 ], 
                    [ 708.25 , 1.88585790125488e-46 ], [ 708.3125 , 1.886111901028795e-46 ], [ 708.375 , 1.8863659350131008e-46 ], [ 708.4375 , 1.886620003212365e-46 ], 
                    [ 708.5 , 1.8868741056311957e-46 ], [ 708.5625 , 1.887128242274242e-46 ], [ 708.625 , 1.887382413146073e-46 ], [ 708.6875 , 1.887636618251299e-46 ], 
                    [ 708.75 , 1.8878908575945716e-46 ], [ 708.8125 , 1.888145131180448e-46 ], [ 708.875 , 1.888399439013594e-46 ], [ 708.9375 , 1.888653781098582e-46 ], 
                    [ 709.0 , 1.888908157440025e-46 ], [ 709.0625 , 1.8891625680425776e-46 ], [ 709.125 , 1.889417012910814e-46 ], [ 709.1875 , 1.889671492049362e-46 ], 
                    [ 709.25 , 1.889926005462852e-46 ], [ 709.3125 , 1.8901805531558458e-46 ], [ 709.375 , 1.8904351351330545e-46 ], [ 709.4375 , 1.890689751399002e-46 ], 
                    [ 709.5 , 1.8909444019583465e-46 ], [ 709.5625 , 1.8911990324776347e-46 ], [ 709.625 , 1.891453453422466e-46 ], [ 709.6875 , 1.8917079085942698e-46 ], 
                    [ 709.75 , 1.8919623979976504e-46 ], [ 709.8125 , 1.8922169216371997e-46 ], [ 709.875 , 1.892471479517564e-46 ], [ 709.9375 , 1.892726071643295e-46 ], 
                    [ 710.0 , 1.8929806980190415e-46 ], [ 710.0625 , 1.8932353586493962e-46 ], [ 710.125 , 1.8934900535389816e-46 ], [ 710.1875 , 1.893744782692353e-46 ], 
                    [ 710.25 , 1.893999546114186e-46 ], [ 710.3125 , 1.894254343809065e-46 ], [ 710.375 , 1.8945091757816002e-46 ], [ 710.4375 , 1.8947640420364027e-46 ], 
                    [ 710.5 , 1.8950189425780713e-46 ], [ 710.5625 , 1.8952738774112586e-46 ], [ 710.625 , 1.895528846540525e-46 ], [ 710.6875 , 1.8957838499705234e-46 ], 
                    [ 710.75 , 1.8960388877058824e-46 ], [ 710.8125 , 1.8962939597511228e-46 ], [ 710.875 , 1.896549066110995e-46 ], [ 710.9375 , 1.8968042067899935e-46 ], 
                    [ 711.0 , 1.89705938179283e-46 ], [ 711.0625 , 1.8973145911240548e-46 ], [ 711.125 , 1.897569834788326e-46 ], [ 711.1875 , 1.8978249815514173e-46 ], 
                    [ 711.25 , 1.8980800018334777e-46 ], [ 711.3125 , 1.898335056383909e-46 ], [ 711.375 , 1.8985901452072885e-46 ], [ 711.4375 , 1.8988452683082356e-46 ], 
                    [ 711.5 , 1.8991004256913558e-46 ], [ 711.5625 , 1.8993556173612564e-46 ], [ 711.625 , 1.8996108433225444e-46 ], [ 711.6875 , 1.8998661035798275e-46 ], 
                    [ 711.75 , 1.9001213981377143e-46 ], [ 711.8125 , 1.9003767270008414e-46 ], [ 711.875 , 1.9006320901737503e-46 ], [ 711.9375 , 1.9008874876610927e-46 ], 
                    [ 712.0 , 1.901142919467493e-46 ], [ 712.0625 , 1.901398385597562e-46 ], [ 712.125 , 1.9016538860558857e-46 ], [ 712.1875 , 1.901909420847117e-46 ], 
                    [ 712.25 , 1.902164989975816e-46 ], [ 712.3125 , 1.9024205934466096e-46 ], [ 712.375 , 1.9026762312641266e-46 ], [ 712.4375 , 1.9029319034330088e-46 ], 
                    [ 712.5 , 1.9031876099578324e-46 ], [ 712.5625 , 1.90344335084324e-46 ], [ 712.625 , 1.903699126093809e-46 ], [ 712.6875 , 1.9039549357141977e-46 ], 
                    [ 712.75 , 1.9042107797090514e-46 ], [ 712.8125 , 1.904466459289398e-46 ], [ 712.875 , 1.9047220699721015e-46 ], [ 712.9375 , 1.904977714961944e-46 ], 
                    [ 713.0 , 1.9052333942635563e-46 ], [ 713.0625 , 1.9054891078815046e-46 ], [ 713.125 , 1.905744855820434e-46 ], [ 713.1875 , 1.9060006380849514e-46 ], 
                    [ 713.25 , 1.9062564546796367e-46 ], [ 713.3125 , 1.906512305609098e-46 ], [ 713.375 , 1.90676819087797e-46 ], [ 713.4375 , 1.9070241104908487e-46 ], 
                    [ 713.5 , 1.9072800644523432e-46 ], [ 713.5625 , 1.9075360527670637e-46 ], [ 713.625 , 1.9077920754396345e-46 ], [ 713.6875 , 1.9080481324746404e-46 ], 
                    [ 713.75 , 1.908304223876706e-46 ], [ 713.8125 , 1.9085603496504313e-46 ], [ 713.875 , 1.908816509800483e-46 ], [ 713.9375 , 1.9090727043313944e-46 ], 
                    [ 714.0 , 1.9093289332478337e-46 ], [ 714.0625 , 1.909585196554389e-46 ], [ 714.125 , 1.9098414942557165e-46 ], [ 714.1875 , 1.9100978263563923e-46 ], 
                    [ 714.25 , 1.910354192861033e-46 ], [ 714.3125 , 1.9106105937742837e-46 ], [ 714.375 , 1.910867029100748e-46 ], [ 714.4375 , 1.9111232470658275e-46 ], 
                    [ 714.5 , 1.9113794398573197e-46 ], [ 714.5625 , 1.911635666992357e-46 ], [ 714.625 , 1.9118919284755292e-46 ], [ 714.6875 , 1.9121482243114675e-46 ], 
                    [ 714.75 , 1.9124045545047505e-46 ], [ 714.8125 , 1.9126609190600246e-46 ], [ 714.875 , 1.912917317981855e-46 ], [ 714.9375 , 1.9131737512748627e-46 ], 
                    [ 715.0 , 1.913430218943641e-46 ], [ 715.0625 , 1.9136867209928124e-46 ], [ 715.125 , 1.9139432574270127e-46 ], [ 715.1875 , 1.914199828250797e-46 ], 
                    [ 715.25 , 1.914456433468816e-46 ], [ 715.3125 , 1.9147130730856664e-46 ], [ 715.375 , 1.91496974710596e-46 ], [ 715.4375 , 1.915226455534336e-46 ], 
                    [ 715.5 , 1.9154831983752976e-46 ], [ 715.5625 , 1.915739975633622e-46 ], [ 715.625 , 1.9159967873137996e-46 ], [ 715.6875 , 1.916253633420486e-46 ], 
                    [ 715.75 , 1.9165105139583235e-46 ], [ 715.8125 , 1.916767428931873e-46 ], [ 715.875 , 1.9170243783457917e-46 ], [ 715.9375 , 1.917281362204656e-46 ], 
                    [ 716.0 , 1.9175383805131506e-46 ], [ 716.0625 , 1.917795170044948e-46 ], [ 716.125 , 1.918051964869163e-46 ], [ 716.1875 , 1.918308794078452e-46 ], 
                    [ 716.25 , 1.918565657677474e-46 ], [ 716.3125 , 1.9188225556708204e-46 ], [ 716.375 , 1.919079488063042e-46 ], [ 716.4375 , 1.9193364548588264e-46 ], 
                    [ 716.5 , 1.919593456062726e-46 ], [ 716.5625 , 1.9198504916793345e-46 ], [ 716.625 , 1.9201075617133144e-46 ], [ 716.6875 , 1.9203646661692464e-46 ], 
                    [ 716.75 , 1.9206218050517265e-46 ], [ 716.8125 , 1.9208789783653914e-46 ], [ 716.875 , 1.9211361861148523e-46 ], [ 716.9375 , 1.9213934283046643e-46 ], 
                    [ 717.0 , 1.9216507049394945e-46 ], [ 717.0625 , 1.9219080160239548e-46 ], [ 717.125 , 1.9221653615626447e-46 ], [ 717.1875 , 1.9224227415601636e-46 ], 
                    [ 717.25 , 1.9226801560211666e-46 ], [ 717.3125 , 1.922937604950228e-46 ], [ 717.375 , 1.9231950883519486e-46 ], [ 717.4375 , 1.9234526062309995e-46 ], 
                    [ 717.5 , 1.92371015859197e-46 ], [ 717.5625 , 1.9239677454394905e-46 ], [ 717.625 , 1.9242253667781516e-46 ], [ 717.6875 , 1.9244827284199496e-46 ], 
                    [ 717.75 , 1.9247401035568988e-46 ], [ 717.8125 , 1.9249975131144785e-46 ], [ 717.875 , 1.925254957097319e-46 ], [ 717.9375 , 1.925512435510052e-46 ], 
                    [ 718.0 , 1.9257699483572404e-46 ], [ 718.0625 , 1.9260274956435175e-46 ], [ 718.125 , 1.9262850773734612e-46 ], [ 718.1875 , 1.926542693551705e-46 ], 
                    [ 718.25 , 1.926800344182816e-46 ], [ 718.3125 , 1.9270580292714416e-46 ], [ 718.375 , 1.9273157488222047e-46 ], [ 718.4375 , 1.9275735028396728e-46 ], 
                    [ 718.5 , 1.9278312913284274e-46 ], [ 718.5625 , 1.9280891142931616e-46 ], [ 718.625 , 1.928346971738431e-46 ], [ 718.6875 , 1.9286048636688607e-46 ], 
                    [ 718.75 , 1.928862790089076e-46 ], [ 718.8125 , 1.92912075100369e-46 ], [ 718.875 , 1.9293787464172473e-46 ], [ 718.9375 , 1.929636776334471e-46 ], 
                    [ 719.0 , 1.92989484075988e-46 ], [ 719.0625 , 1.930152939698157e-46 ], [ 719.125 , 1.930411073153878e-46 ], [ 719.1875 , 1.9306692411316714e-46 ], 
                    [ 719.25 , 1.9309274436361553e-46 ], [ 719.3125 , 1.931185407968903e-46 ], [ 719.375 , 1.9314433790773213e-46 ], [ 719.4375 , 1.931701384645983e-46 ], 
                    [ 719.5 , 1.9319594246795052e-46 ], [ 719.5625 , 1.932217499182437e-46 ], [ 719.625 , 1.932475608159396e-46 ], [ 719.6875 , 1.9327337516150157e-46 ], 
                    [ 719.75 , 1.9329919295538733e-46 ], [ 719.8125 , 1.933250141980617e-46 ], [ 719.875 , 1.9335083888998263e-46 ], [ 719.9375 , 1.933766670316081e-46 ], 
                    [ 720.0 , 1.9340249862340166e-46 ], [ 720.0625 , 1.93428333665827e-46 ], [ 720.125 , 1.9345417215934093e-46 ], [ 720.1875 , 1.934800141044044e-46 ], 
                    [ 720.25 , 1.9350585950148262e-46 ], [ 720.3125 , 1.9353170835103256e-46 ], [ 720.375 , 1.9355756065351412e-46 ], [ 720.4375 , 1.9358341640939536e-46 ], 
                    [ 720.5 , 1.9360927561913346e-46 ], [ 720.5625 , 1.9363513828318982e-46 ], [ 720.625 , 1.9366100440202177e-46 ], [ 720.6875 , 1.9368687397609897e-46 ], 
                    [ 720.75 , 1.937127470058776e-46 ], [ 720.8125 , 1.937386234918219e-46 ], [ 720.875 , 1.937645034343895e-46 ], [ 720.9375 , 1.937903616856971e-46 ], 
                    [ 721.0 , 1.9381621798596423e-46 ], [ 721.0625 , 1.9384207773608716e-46 ], [ 721.125 , 1.9386794093652065e-46 ], [ 721.1875 , 1.9389380758772914e-46 ], 
                    [ 721.25 , 1.9391967769017173e-46 ], [ 721.3125 , 1.9394555124431025e-46 ], [ 721.375 , 1.9397142825060115e-46 ], [ 721.4375 , 1.9399730870951046e-46 ], 
                    [ 721.5 , 1.9402319262149475e-46 ], [ 721.5625 , 1.9404907998701476e-46 ], [ 721.625 , 1.940749708065354e-46 ], [ 721.6875 , 1.94100865080512e-46 ], 
                    [ 721.75 , 1.941267628094096e-46 ], [ 721.8125 , 1.9415266399368777e-46 ], [ 721.875 , 1.941785686338062e-46 ], [ 721.9375 , 1.9420447673023287e-46 ], 
                    [ 722.0 , 1.9423038828341787e-46 ], [ 722.0625 , 1.9425630329382934e-46 ], [ 722.125 , 1.942822217619272e-46 ], [ 722.1875 , 1.943081436881727e-46 ], 
                    [ 722.25 , 1.9433406907302597e-46 ], [ 722.3125 , 1.9435999791695257e-46 ], [ 722.375 , 1.9438593022040848e-46 ], [ 722.4375 , 1.944118659838608e-46 ], 
                    [ 722.5 , 1.944378052077685e-46 ], [ 722.5625 , 1.9446372657370478e-46 ], [ 722.625 , 1.944896413320433e-46 ], [ 722.6875 , 1.9451555954385067e-46 ], 
                    [ 722.75 , 1.9454148120958858e-46 ], [ 722.8125 , 1.9456740632971725e-46 ], [ 722.875 , 1.945933349046984e-46 ], [ 722.9375 , 1.946192669349883e-46 ], 
                    [ 723.0 , 1.94645202421053e-46 ], [ 723.0625 , 1.946711413633488e-46 ], [ 723.125 , 1.9469708376233906e-46 ], [ 723.1875 , 1.947230296184817e-46 ], 
                    [ 723.25 , 1.947489789322388e-46 ], [ 723.3125 , 1.947749317040711e-46 ], [ 723.375 , 1.948008879344381e-46 ], [ 723.4375 , 1.948268476238076e-46 ], 
                    [ 723.5 , 1.9485281077262943e-46 ], [ 723.5625 , 1.948787773813716e-46 ], [ 723.625 , 1.9490474745049513e-46 ], [ 723.6875 , 1.9493072098045564e-46 ], 
                    [ 723.75 , 1.9495669797172267e-46 ], [ 723.8125 , 1.949826784247533e-46 ], [ 723.875 , 1.950086623400075e-46 ], [ 723.9375 , 1.95034649717948e-46 ], 
                    [ 724.0 , 1.950606405590376e-46 ], [ 724.0625 , 1.950866348637351e-46 ], [ 724.125 , 1.951126326325035e-46 ], [ 724.1875 , 1.951386187825741e-46 ], 
                    [ 724.25 , 1.9516459179783335e-46 ], [ 724.3125 , 1.9519056827010822e-46 ], [ 724.375 , 1.952165481998602e-46 ], [ 724.4375 , 1.9524253158755087e-46 ], 
                    [ 724.5 , 1.952685184336377e-46 ], [ 724.5625 , 1.9529450873858247e-46 ], [ 724.625 , 1.953205025028455e-46 ], [ 724.6875 , 1.9534649972688716e-46 ], 
                    [ 724.75 , 1.9537250041116943e-46 ], [ 724.8125 , 1.953985045561501e-46 ], [ 724.875 , 1.9542451216228973e-46 ], [ 724.9375 , 1.9545052323005318e-46 ], 
                    [ 725.0 , 1.954765377598957e-46 ], [ 725.0625 , 1.955025557522835e-46 ], [ 725.125 , 1.955285772076735e-46 ], [ 725.1875 , 1.955546021265278e-46 ], 
                    [ 725.25 , 1.955806305093089e-46 ], [ 725.3125 , 1.9560666235647362e-46 ], [ 725.375 , 1.9563269766848734e-46 ], [ 725.4375 , 1.956587364458097e-46 ], 
                    [ 725.5 , 1.9568477868890207e-46 ], [ 725.5625 , 1.9571082439822843e-46 ], [ 725.625 , 1.957368735742432e-46 ], [ 725.6875 , 1.957629262174148e-46 ], 
                    [ 725.75 , 1.9578898232819773e-46 ], [ 725.8125 , 1.9581503542030235e-46 ], [ 725.875 , 1.9584106809406948e-46 ], [ 725.9375 , 1.9586710422875765e-46 ], 
                    [ 726.0 , 1.9589314382482418e-46 ], [ 726.0625 , 1.95919186882732e-46 ], [ 726.125 , 1.9594523340293866e-46 ], [ 726.1875 , 1.959712833859058e-46 ], 
                    [ 726.25 , 1.959973368320937e-46 ], [ 726.3125 , 1.9602339374196427e-46 ], [ 726.375 , 1.9604945411597514e-46 ], [ 726.4375 , 1.9607551795458972e-46 ], 
                    [ 726.5 , 1.9610158525826574e-46 ], [ 726.5625 , 1.9612765602746527e-46 ], [ 726.625 , 1.9615373026265187e-46 ], [ 726.6875 , 1.9617980796428073e-46 ], 
                    [ 726.75 , 1.9620588913281546e-46 ], [ 726.8125 , 1.96231973768717e-46 ], [ 726.875 , 1.9625806187245045e-46 ], [ 726.9375 , 1.9628415344446717e-46 ], 
                    [ 727.0 , 1.963102484852365e-46 ], [ 727.0625 , 1.9633634699521417e-46 ], [ 727.125 , 1.963624489748682e-46 ], [ 727.1875 , 1.96388554424653e-46 ], 
                    [ 727.25 , 1.9641466334503413e-46 ], [ 727.3125 , 1.9644077573646867e-46 ], [ 727.375 , 1.9646689159942516e-46 ], [ 727.4375 , 1.964930109343567e-46 ], 
                    [ 727.5 , 1.965191051099966e-46 ], [ 727.5625 , 1.9654519926656318e-46 ], [ 727.625 , 1.96571296887957e-46 ], [ 727.6875 , 1.9659739797464214e-46 ], 
                    [ 727.75 , 1.9662350252707187e-46 ], [ 727.8125 , 1.9664961054571473e-46 ], [ 727.875 , 1.966757220310254e-46 ], [ 727.9375 , 1.9670183698346412e-46 ], 
                    [ 728.0 , 1.9672795540349415e-46 ], [ 728.0625 , 1.9675407729157446e-46 ], [ 728.125 , 1.967802026481656e-46 ], [ 728.1875 , 1.968063314737266e-46 ], 
                    [ 728.25 , 1.9683246376872106e-46 ], [ 728.3125 , 1.968585995336067e-46 ], [ 728.375 , 1.968847387688472e-46 ], [ 728.4375 , 1.9691088147490178e-46 ], 
                    [ 728.5 , 1.9693702765223005e-46 ], [ 728.5625 , 1.969631773012957e-46 ], [ 728.625 , 1.9698933042255828e-46 ], [ 728.6875 , 1.9701548701647744e-46 ], 
                    [ 728.75 , 1.9704164708351713e-46 ], [ 728.8125 , 1.9706781062413706e-46 ], [ 728.875 , 1.970939776387971e-46 ], [ 728.9375 , 1.9712014812796128e-46 ], 
                    [ 729.0 , 1.971463220920911e-46 ], [ 729.0625 , 1.9717249953164213e-46 ], [ 729.125 , 1.9719866422227062e-46 ], [ 729.1875 , 1.9722481678933105e-46 ], 
                    [ 729.25 , 1.972509728247556e-46 ], [ 729.3125 , 1.97277132329007e-46 ], [ 729.375 , 1.9730329530253827e-46 ], [ 729.4375 , 1.9732946174581655e-46 ], 
                    [ 729.5 , 1.973556316592978e-46 ], [ 729.5625 , 1.9738180504344636e-46 ], [ 729.625 , 1.9740798189871563e-46 ], [ 729.6875 , 1.9743416222557005e-46 ], 
                    [ 729.75 , 1.9746034602447292e-46 ], [ 729.8125 , 1.9748653329588183e-46 ], [ 729.875 , 1.975127240402573e-46 ], [ 729.9375 , 1.9753891825805993e-46 ], 
                    [ 730.0 , 1.975651159497462e-46 ], [ 730.0625 , 1.9759131711578383e-46 ], [ 730.125 , 1.976175217566321e-46 ], [ 730.1875 , 1.9764372987275055e-46 ], 
                    [ 730.25 , 1.976699414646014e-46 ], [ 730.3125 , 1.9769615653264145e-46 ], [ 730.375 , 1.9772237507733583e-46 ], [ 730.4375 , 1.977485970991415e-46 ], 
                    [ 730.5 , 1.977748225985252e-46 ], [ 730.5625 , 1.9780105157594385e-46 ], [ 730.625 , 1.9782728403186155e-46 ], [ 730.6875 , 1.9785351996673824e-46 ], 
                    [ 730.75 , 1.978797571722703e-46 ], [ 730.8125 , 1.979059677321962e-46 ], [ 730.875 , 1.9793218176389436e-46 ], [ 730.9375 , 1.9795839926782747e-46 ], 
                    [ 731.0 , 1.9798462024444977e-46 ], [ 731.0625 , 1.9801084469422553e-46 ], [ 731.125 , 1.9803707261761196e-46 ], [ 731.1875 , 1.9806330401507197e-46 ], 
                    [ 731.25 , 1.9808953888706426e-46 ], [ 731.3125 , 1.981157772340492e-46 ], [ 731.375 , 1.9814201905648413e-46 ], [ 731.4375 , 1.9816826435483504e-46 ], 
                    [ 731.5 , 1.981945131295596e-46 ], [ 731.5625 , 1.9822076538111684e-46 ], [ 731.625 , 1.9824702110997147e-46 ], [ 731.6875 , 1.982732803165785e-46 ], 
                    [ 731.75 , 1.9829954300140136e-46 ], [ 731.8125 , 1.983258091649036e-46 ], [ 731.875 , 1.9835207880754183e-46 ], [ 731.9375 , 1.9837835192977398e-46 ], 
                    [ 732.0 , 1.9840462853207224e-46 ], [ 732.0625 , 1.984309086148863e-46 ], [ 732.125 , 1.9845719217868144e-46 ], [ 732.1875 , 1.984834792239229e-46 ], 
                    [ 732.25 , 1.9850976975106065e-46 ], [ 732.3125 , 1.9853606376056565e-46 ], [ 732.375 , 1.9856236125289783e-46 ], [ 732.4375 , 1.985886443065703e-46 ], 
                    [ 732.5 , 1.9861491714431675e-46 ], [ 732.5625 , 1.9864119345790004e-46 ], [ 732.625 , 1.986674732477828e-46 ], [ 732.6875 , 1.9869375651442217e-46 ], 
                    [ 732.75 , 1.987200432582809e-46 ], [ 732.8125 , 1.9874633347981905e-46 ], [ 732.875 , 1.987726271794924e-46 ], [ 732.9375 , 1.987989243577654e-46 ], 
                    [ 733.0 , 1.9882522501509547e-46 ], [ 733.0625 , 1.988515291519442e-46 ], [ 733.125 , 1.988778367687748e-46 ], [ 733.1875 , 1.98904147866042e-46 ], 
                    [ 733.25 , 1.9893046244421193e-46 ], [ 733.3125 , 1.9895678050373943e-46 ], [ 733.375 , 1.989831020450879e-46 ], [ 733.4375 , 1.99009427068718e-46 ], 
                    [ 733.5 , 1.9903575557509324e-46 ], [ 733.5625 , 1.990620875646673e-46 ], [ 733.625 , 1.9908842303790807e-46 ], [ 733.6875 , 1.9911476199527506e-46 ], 
                    [ 733.75 , 1.991411044372249e-46 ], [ 733.8125 , 1.991674503642229e-46 ], [ 733.875 , 1.9919379977672866e-46 ], [ 733.9375 , 1.9922015267520193e-46 ], 
                    [ 734.0 , 1.992465090601081e-46 ], [ 734.0625 , 1.9927286875714447e-46 ], [ 734.125 , 1.992991999872541e-46 ], [ 734.1875 , 1.9932553469668313e-46 ], 
                    [ 734.25 , 1.9935187288588994e-46 ], [ 734.3125 , 1.993782145553329e-46 ], [ 734.375 , 1.994045597054761e-46 ], [ 734.4375 , 1.994309083367738e-46 ], 
                    [ 734.5 , 1.9945726044969166e-46 ], [ 734.5625 , 1.9948361604468406e-46 ], [ 734.625 , 1.9950997512221536e-46 ], [ 734.6875 , 1.9953633768274436e-46 ], 
                    [ 734.75 , 1.9956270372673124e-46 ], [ 734.8125 , 1.995890732546363e-46 ], [ 734.875 , 1.9961544626691993e-46 ], [ 734.9375 , 1.9964182276404103e-46 ], 
                    [ 735.0 , 1.996682027464673e-46 ], [ 735.0625 , 1.99694586214645e-46 ], [ 735.125 , 1.997209731690489e-46 ], [ 735.1875 , 1.9974736361012836e-46 ], 
                    [ 735.25 , 1.9977375753835825e-46 ], [ 735.3125 , 1.998001549541852e-46 ], [ 735.375 , 1.9982655585807855e-46 ], [ 735.4375 , 1.9985296025049206e-46 ], 
                    [ 735.5 , 1.998793681318953e-46 ], [ 735.5625 , 1.9990577950274352e-46 ], [ 735.625 , 1.9993219436349926e-46 ], [ 735.6875 , 1.9995861271462365e-46 ], 
                    [ 735.75 , 1.9998502215580994e-46 ], [ 735.8125 , 2.000114151625496e-46 ], [ 735.875 , 2.000378116525056e-46 ], [ 735.9375 , 2.000642116261362e-46 ], 
                    [ 736.0 , 2.0009061508390113e-46 ], [ 736.0625 , 2.001170220262602e-46 ], [ 736.125 , 2.001434324536733e-46 ], [ 736.1875 , 2.0016984636660183e-46 ], 
                    [ 736.25 , 2.0019626376550286e-46 ], [ 736.3125 , 2.0022268465083934e-46 ], [ 736.375 , 2.0024910902306573e-46 ], [ 736.4375 , 2.002755368826493e-46 ], 
                    [ 736.5 , 2.003019682300489e-46 ], [ 736.5625 , 2.003284030657219e-46 ], [ 736.625 , 2.003548413901302e-46 ], [ 736.6875 , 2.0038128320373273e-46 ], 
                    [ 736.75 , 2.004077285069914e-46 ], [ 736.8125 , 2.0043417730036684e-46 ], [ 736.875 , 2.0046062958432097e-46 ], [ 736.9375 , 2.0048708535931592e-46 ], 
                    [ 737.0 , 2.0051354462580676e-46 ], [ 737.0625 , 2.005400073842557e-46 ], [ 737.125 , 2.005664736351264e-46 ], [ 737.1875 , 2.0059294337887837e-46 ], 
                    [ 737.25 , 2.0061941661597402e-46 ], [ 737.3125 , 2.006458933468715e-46 ], [ 737.375 , 2.0067237357203335e-46 ], [ 737.4375 , 2.006988339572888e-46 ], 
                    [ 737.5 , 2.0072528790573703e-46 ], [ 737.5625 , 2.0075174534105276e-46 ], [ 737.625 , 2.0077820626370277e-46 ], [ 737.6875 , 2.0080467067414524e-46 ], 
                    [ 737.75 , 2.008311385728399e-46 ], [ 737.8125 , 2.0085760996024943e-46 ], [ 737.875 , 2.008840848368279e-46 ], [ 737.9375 , 2.009105632030381e-46 ], 
                    [ 738.0 , 2.0093704505934005e-46 ], [ 738.0625 , 2.0096353040619372e-46 ], [ 738.125 , 2.0099001924405916e-46 ], [ 738.1875 , 2.01016511573398e-46 ], 
                    [ 738.25 , 2.0104300739466904e-46 ], [ 738.3125 , 2.0106950670832962e-46 ], [ 738.375 , 2.010960095148444e-46 ], [ 738.4375 , 2.011225158146738e-46 ], 
                    [ 738.5 , 2.0114902560827532e-46 ], [ 738.5625 , 2.01175538896111e-46 ], [ 738.625 , 2.0120205567864425e-46 ], [ 738.6875 , 2.012285759563285e-46 ], 
                    [ 738.75 , 2.012550997296302e-46 ], [ 738.8125 , 2.0128162699901155e-46 ], [ 738.875 , 2.0130815776492762e-46 ], [ 738.9375 , 2.0133469202784213e-46 ], 
                    [ 739.0 , 2.0136122978821607e-46 ], [ 739.0625 , 2.013877710465118e-46 ], [ 739.125 , 2.0141428530328464e-46 ], [ 739.1875 , 2.014408017480598e-46 ], 
                    [ 739.25 , 2.0146732168375982e-46 ], [ 739.3125 , 2.0149384511083995e-46 ], [ 739.375 , 2.0152037202976558e-46 ], [ 739.4375 , 2.0154690244099215e-46 ], 
                    [ 739.5 , 2.0157343634498223e-46 ], [ 739.5625 , 2.015999737421914e-46 ], [ 739.625 , 2.016265146330838e-46 ], [ 739.6875 , 2.0165305901811946e-46 ], 
                    [ 739.75 , 2.0167960689775542e-46 ], [ 739.8125 , 2.017061582724547e-46 ], [ 739.875 , 2.0173271314267445e-46 ], [ 739.9375 , 2.01759271508875e-46 ], 
                    [ 740.0 , 2.017858333715208e-46 ], [ 740.0625 , 2.018123987310708e-46 ], [ 740.125 , 2.0183896758797814e-46 ], [ 740.1875 , 2.018655399427133e-46 ], 
                    [ 740.25 , 2.018921157957311e-46 ], [ 740.3125 , 2.0191869514749488e-46 ], [ 740.375 , 2.0194527799846243e-46 ], [ 740.4375 , 2.0197186434909586e-46 ], 
                    [ 740.5 , 2.0199845419985588e-46 ], [ 740.5625 , 2.0202504755120473e-46 ], [ 740.625 , 2.020516444036004e-46 ], [ 740.6875 , 2.0207824475750526e-46 ], 
                    [ 740.75 , 2.02104843381843e-46 ], [ 740.8125 , 2.021314171995409e-46 ], [ 740.875 , 2.0215799451130545e-46 ], [ 740.9375 , 2.02184575317596e-46 ], 
                    [ 741.0 , 2.0221115961887356e-46 ], [ 741.0625 , 2.022377474155961e-46 ], [ 741.125 , 2.0226433870822193e-46 ], [ 741.1875 , 2.022909334972149e-46 ], 
                    [ 741.25 , 2.0231753178302905e-46 ], [ 741.3125 , 2.023441335661256e-46 ], [ 741.375 , 2.023707388469701e-46 ], [ 741.4375 , 2.0239734762601385e-46 ], 
                    [ 741.5 , 2.0242395990372255e-46 ], [ 741.5625 , 2.0245057568055772e-46 ], [ 741.625 , 2.0247719495697363e-46 ], [ 741.6875 , 2.025038177334333e-46 ], 
                    [ 741.75 , 2.0253044401039838e-46 ], [ 741.8125 , 2.0255707378832623e-46 ], [ 741.875 , 2.025837070676787e-46 ], [ 741.9375 , 2.0261034384891756e-46 ], 
                    [ 742.0 , 2.026369841325018e-46 ], [ 742.0625 , 2.0266362791889195e-46 ], [ 742.125 , 2.026902752085472e-46 ], [ 742.1875 , 2.027169260019309e-46 ], 
                    [ 742.25 , 2.02743580299501e-46 ], [ 742.3125 , 2.0277023810171962e-46 ], [ 742.375 , 2.027968994090476e-46 ], [ 742.4375 , 2.028235544198403e-46 ], 
                    [ 742.5 , 2.028501891642156e-46 ], [ 742.5625 , 2.028768274062567e-46 ], [ 742.625 , 2.0290346914642733e-46 ], [ 742.6875 , 2.029301143851854e-46 ], 
                    [ 742.75 , 2.0295676312299034e-46 ], [ 742.8125 , 2.0298341536030162e-46 ], [ 742.875 , 2.0301007109757886e-46 ], [ 742.9375 , 2.030367303352787e-46 ], 
                    [ 743.0 , 2.030633930738681e-46 ], [ 743.0625 , 2.0309005931379955e-46 ], [ 743.125 , 2.0311672905553865e-46 ], [ 743.1875 , 2.031434022995394e-46 ], 
                    [ 743.25 , 2.0317007904626614e-46 ], [ 743.3125 , 2.0319675929617737e-46 ], [ 743.375 , 2.0322344304973172e-46 ], [ 743.4375 , 2.0325013030739352e-46 ], 
                    [ 743.5 , 2.032768210696187e-46 ], [ 743.5625 , 2.0330351533686743e-46 ], [ 743.625 , 2.0333021310960287e-46 ], [ 743.6875 , 2.033569143882839e-46 ], 
                    [ 743.75 , 2.0338361917337243e-46 ], [ 743.8125 , 2.03410327465326e-46 ], [ 743.875 , 2.0343703926460514e-46 ], [ 743.9375 , 2.0346375457167473e-46 ], 
                    [ 744.0 , 2.0349047338698965e-46 ], [ 744.0625 , 2.0351719571101205e-46 ], [ 744.125 , 2.0354390943962456e-46 ], [ 744.1875 , 2.035706058119809e-46 ], 
                    [ 744.25 , 2.035973056857792e-46 ], [ 744.3125 , 2.0362400906147143e-46 ], [ 744.375 , 2.0365071593951984e-46 ], [ 744.4375 , 2.0367742632038516e-46 ], 
                    [ 744.5 , 2.0370414020452395e-46 ], [ 744.5625 , 2.037308575923957e-46 ], [ 744.625 , 2.0375757848446284e-46 ], [ 744.6875 , 2.03784302881185e-46 ], 
                    [ 744.75 , 2.038110307830189e-46 ], [ 744.8125 , 2.0383776219042573e-46 ], [ 744.875 , 2.038644971038653e-46 ], [ 744.9375 , 2.0389123552379738e-46 ], 
                    [ 745.0 , 2.0391797745068052e-46 ], [ 745.0625 , 2.0394472288498043e-46 ], [ 745.125 , 2.039714718271499e-46 ], [ 745.1875 , 2.0399822427765042e-46 ], 
                    [ 745.25 , 2.0402498023694657e-46 ], [ 745.3125 , 2.0405173970549267e-46 ], [ 745.375 , 2.0407850268375633e-46 ], [ 745.4375 , 2.0410526917219053e-46 ], 
                    [ 745.5 , 2.0413203917125715e-46 ], [ 745.5625 , 2.0415881268142095e-46 ], [ 745.625 , 2.0418558970313963e-46 ], [ 745.6875 , 2.0421237023687076e-46 ], 
                    [ 745.75 , 2.0423915428308085e-46 ], [ 745.8125 , 2.0426592915380648e-46 ], [ 745.875 , 2.0429268695580264e-46 ], [ 745.9375 , 2.0431944826293817e-46 ], 
                    [ 746.0 , 2.0434621307566794e-46 ], [ 746.0625 , 2.0437298139445117e-46 ], [ 746.125 , 2.0439975321974852e-46 ], [ 746.1875 , 2.044265285520223e-46 ], 
                    [ 746.25 , 2.0445330739172754e-46 ], [ 746.3125 , 2.0448008973932512e-46 ], [ 746.375 , 2.045068755952731e-46 ], [ 746.4375 , 2.0453366496003397e-46 ], 
                    [ 746.5 , 2.045604578340674e-46 ], [ 746.5625 , 2.0458725421783013e-46 ], [ 746.625 , 2.0461405411178197e-46 ], [ 746.6875 , 2.0464085751638417e-46 ], 
                    [ 746.75 , 2.0466766443209804e-46 ], [ 746.8125 , 2.0469447485938205e-46 ], [ 746.875 , 2.047212887986977e-46 ], [ 746.9375 , 2.047481062504992e-46 ], 
                    [ 747.0 , 2.0477492721525394e-46 ], [ 747.0625 , 2.048017516934207e-46 ], [ 747.125 , 2.0482857968545816e-46 ], [ 747.1875 , 2.0485541119182233e-46 ], 
                    [ 747.25 , 2.048822462129823e-46 ], [ 747.3125 , 2.049090847493941e-46 ], [ 747.375 , 2.0493592680151385e-46 ], [ 747.4375 , 2.0496277236981084e-46 ], 
                    [ 747.5 , 2.049896100347792e-46 ], [ 747.5625 , 2.05016428480236e-46 ], [ 747.625 , 2.0504325043430333e-46 ], [ 747.6875 , 2.050700758974418e-46 ], 
                    [ 747.75 , 2.050969048701133e-46 ], [ 747.8125 , 2.0512373735277267e-46 ], [ 747.875 , 2.051505733458791e-46 ], [ 747.9375 , 2.0517741284989333e-46 ], 
                    [ 748.0 , 2.0520425586527612e-46 ], [ 748.0625 , 2.0523110239248687e-46 ], [ 748.125 , 2.0525795243197915e-46 ], [ 748.1875 , 2.052848059842213e-46 ], 
                    [ 748.25 , 2.05311663049667e-46 ], [ 748.3125 , 2.0533852362877587e-46 ], [ 748.375 , 2.05365387722012e-46 ], [ 748.4375 , 2.0539225532982926e-46 ], 
                    [ 748.5 , 2.0541912645269042e-46 ], [ 748.5625 , 2.0544600109105536e-46 ], [ 748.625 , 2.0547287924538833e-46 ], [ 748.6875 , 2.0549976091614056e-46 ], 
                    [ 748.75 , 2.0552664610377507e-46 ], [ 748.8125 , 2.055535348087563e-46 ], [ 748.875 , 2.05580427031543e-46 ], [ 748.9375 , 2.05607322772591e-46 ], 
                    [ 749.0 , 2.0563422203236495e-46 ], [ 749.0625 , 2.0566112481132375e-46 ], [ 749.125 , 2.0568803110992785e-46 ], [ 749.1875 , 2.0571493336828736e-46 ], 
                    [ 749.25 , 2.0574181326116064e-46 ], [ 749.3125 , 2.057686966663164e-46 ], [ 749.375 , 2.0579558358420904e-46 ], [ 749.4375 , 2.05822474015305e-46 ], 
                    [ 749.5 , 2.0584936796005887e-46 ], [ 749.5625 , 2.0587626541892686e-46 ], [ 749.625 , 2.0590316639237257e-46 ], [ 749.6875 , 2.0593007088085076e-46 ], 
                    [ 749.75 , 2.059569788848281e-46 ], [ 749.8125 , 2.0598389040475807e-46 ], [ 749.875 , 2.0601080544110158e-46 ], [ 749.9375 , 2.0603772399431808e-46 ], 
                    [ 750.0 , 2.0606464606486713e-46 ], [ 750.0625 , 2.0609157165320827e-46 ], [ 750.125 , 2.0611850075979975e-46 ], [ 750.1875 , 2.0614543338510417e-46 ], 
                    [ 750.25 , 2.0617236952957984e-46 ], [ 750.3125 , 2.0619930919368667e-46 ], [ 750.375 , 2.0622625237788152e-46 ], [ 750.4375 , 2.0625319908263175e-46 ], 
                    [ 750.5 , 2.0628014930839145e-46 ], [ 750.5625 , 2.063071030556222e-46 ], [ 750.625 , 2.063340603247827e-46 ], [ 750.6875 , 2.063610211163375e-46 ], 
                    [ 750.75 , 2.063879854307395e-46 ], [ 750.8125 , 2.0641495326845643e-46 ], [ 750.875 , 2.0644192269971587e-46 ], [ 750.9375 , 2.064688633081461e-46 ], 
                    [ 751.0 , 2.064958074323146e-46 ], [ 751.0625 , 2.065227550726876e-46 ], [ 751.125 , 2.0654970622971507e-46 ], [ 751.1875 , 2.0657666090386333e-46 ], 
                    [ 751.25 , 2.066036190955869e-46 ], [ 751.3125 , 2.0663058080534494e-46 ], [ 751.375 , 2.066575460336023e-46 ], [ 751.4375 , 2.0668451478080938e-46 ], 
                    [ 751.5 , 2.0671148704743275e-46 ], [ 751.5625 , 2.067384628339273e-46 ], [ 751.625 , 2.0676544214075527e-46 ], [ 751.6875 , 2.067924249683746e-46 ], 
                    [ 751.75 , 2.0681941131724334e-46 ], [ 751.8125 , 2.0684640118782096e-46 ], [ 751.875 , 2.0687339458056997e-46 ], [ 751.9375 , 2.0690039149595004e-46 ], 
                    [ 752.0 , 2.0692739193442086e-46 ], [ 752.0625 , 2.069543958964378e-46 ], [ 752.125 , 2.069814033824636e-46 ], [ 752.1875 , 2.070084143929567e-46 ], 
                    [ 752.25 , 2.070354289283814e-46 ], [ 752.3125 , 2.0706244698919337e-46 ], [ 752.375 , 2.0708946857585555e-46 ], [ 752.4375 , 2.0711649368882663e-46 ], 
                    [ 752.5 , 2.0714352232856525e-46 ], [ 752.5625 , 2.0717055449553176e-46 ], [ 752.625 , 2.0719756214644353e-46 ], [ 752.6875 , 2.0722456736709266e-46 ], 
                    [ 752.75 , 2.0725157610748226e-46 ], [ 752.8125 , 2.0727858836807244e-46 ], [ 752.875 , 2.073056041493236e-46 ], [ 752.9375 , 2.073326234516931e-46 ], 
                    [ 753.0 , 2.0735964627563835e-46 ], [ 753.0625 , 2.0738667262161987e-46 ], [ 753.125 , 2.0741370249009667e-46 ], [ 753.1875 , 2.0744073588152934e-46 ], 
                    [ 753.25 , 2.0746777279637558e-46 ], [ 753.3125 , 2.0749481323509313e-46 ], [ 753.375 , 2.0752185719814273e-46 ], [ 753.4375 , 2.075489046859838e-46 ], 
                    [ 753.5 , 2.0757595569907713e-46 ], [ 753.5625 , 2.076030102378793e-46 ], [ 753.625 , 2.076300683028542e-46 ], [ 753.6875 , 2.076571298944541e-46 ], 
                    [ 753.75 , 2.076841950131445e-46 ], [ 753.8125 , 2.077112636593867e-46 ], [ 753.875 , 2.0773833583363434e-46 ], [ 753.9375 , 2.077654115363489e-46 ], 
                    [ 754.0 , 2.0779249076799318e-46 ], [ 754.0625 , 2.0781957352903007e-46 ], [ 754.125 , 2.078466598199092e-46 ], [ 754.1875 , 2.0787374964109955e-46 ], 
                    [ 754.25 , 2.0790084299305678e-46 ], [ 754.3125 , 2.07927921265882e-46 ], [ 754.375 , 2.0795498669189692e-46 ], [ 754.4375 , 2.0798205564094362e-46 ], 
                    [ 754.5 , 2.0800912811348658e-46 ], [ 754.5625 , 2.0803620410997704e-46 ], [ 754.625 , 2.0806328363088108e-46 ], [ 754.6875 , 2.0809036667665604e-46 ], 
                    [ 754.75 , 2.081174532477548e-46 ], [ 754.8125 , 2.081445433446451e-46 ], [ 754.875 , 2.0817163696778145e-46 ], [ 754.9375 , 2.0819873411761993e-46 ], 
                    [ 755.0 , 2.08225834794624e-46 ], [ 755.0625 , 2.0825293899925134e-46 ], [ 755.125 , 2.0828004673196107e-46 ], [ 755.1875 , 2.083071579932125e-46 ], 
                    [ 755.25 , 2.0833427278346633e-46 ], [ 755.3125 , 2.0836139110317758e-46 ], [ 755.375 , 2.0838851295281e-46 ], [ 755.4375 , 2.084156383328217e-46 ], 
                    [ 755.5 , 2.084427672436692e-46 ], [ 755.5625 , 2.084698996858195e-46 ], [ 755.625 , 2.0849703565972783e-46 ], [ 755.6875 , 2.08524175165851e-46 ], 
                    [ 755.75 , 2.085513182046546e-46 ], [ 755.8125 , 2.0857846477659262e-46 ], [ 755.875 , 2.0860561488212788e-46 ], [ 755.9375 , 2.086327685217219e-46 ], 
                    [ 756.0 , 2.086599197094831e-46 ], [ 756.0625 , 2.086870466151072e-46 ], [ 756.125 , 2.087141770473772e-46 ], [ 756.1875 , 2.0874131100674698e-46 ], 
                    [ 756.25 , 2.0876844849367813e-46 ], [ 756.3125 , 2.087955895086277e-46 ], [ 756.375 , 2.088227340520529e-46 ], [ 756.4375 , 2.088498821244169e-46 ], 
                    [ 756.5 , 2.0887703372617108e-46 ], [ 756.5625 , 2.0890418885778164e-46 ], [ 756.625 , 2.089313475197031e-46 ], [ 756.6875 , 2.0895850971239885e-46 ], 
                    [ 756.75 , 2.089856754363205e-46 ], [ 756.8125 , 2.090128446919345e-46 ], [ 756.875 , 2.0904001747969556e-46 ], [ 756.9375 , 2.0906719380006438e-46 ], 
                    [ 757.0 , 2.090943736535017e-46 ], [ 757.0625 , 2.0912155704046534e-46 ], [ 757.125 , 2.0914874396141168e-46 ], [ 757.1875 , 2.0917593441680764e-46 ], 
                    [ 757.25 , 2.0920312840710825e-46 ], [ 757.3125 , 2.092303259327686e-46 ], [ 757.375 , 2.0925752699425423e-46 ], [ 757.4375 , 2.092847315920263e-46 ], 
                    [ 757.5 , 2.0931193972653714e-46 ], [ 757.5625 , 2.0933915139825695e-46 ], [ 757.625 , 2.0936636660763216e-46 ], [ 757.6875 , 2.0939358535513763e-46 ], 
                    [ 757.75 , 2.094207806295117e-46 ], [ 757.8125 , 2.0944797091226725e-46 ], [ 757.875 , 2.094751647252952e-46 ], [ 757.9375 , 2.09502362069048e-46 ], 
                    [ 758.0 , 2.095295629439856e-46 ], [ 758.0625 , 2.0955676735056786e-46 ], [ 758.125 , 2.0958397528924895e-46 ], [ 758.1875 , 2.0961118676049334e-46 ], 
                    [ 758.25 , 2.0963840176475373e-46 ], [ 758.3125 , 2.0966562030249338e-46 ], [ 758.375 , 2.0969284237416948e-46 ], [ 758.4375 , 2.097200679802409e-46 ], 
                    [ 758.5 , 2.097472971211665e-46 ], [ 758.5625 , 2.097745297974068e-46 ], [ 758.625 , 2.098017660094177e-46 ], [ 758.6875 , 2.098290057576599e-46 ], 
                    [ 758.75 , 2.0985624904259098e-46 ], [ 758.8125 , 2.0988349586467455e-46 ], [ 758.875 , 2.09910746224364e-46 ], [ 758.9375 , 2.09938000122123e-46 ], 
                    [ 759.0 , 2.0996525755841246e-46 ], [ 759.0625 , 2.0999251853368136e-46 ], [ 759.125 , 2.1001978304840267e-46 ], [ 759.1875 , 2.100470511030284e-46 ], 
                    [ 759.25 , 2.1007432269801964e-46 ], [ 759.3125 , 2.101015978338376e-46 ], [ 759.375 , 2.1012887651093748e-46 ], [ 759.4375 , 2.1015614919206445e-46 ], 
                    [ 759.5 , 2.1018340101989885e-46 ], [ 759.5625 , 2.102106563815933e-46 ], [ 759.625 , 2.1023791527760317e-46 ], [ 759.6875 , 2.1026517770838976e-46 ], 
                    [ 759.75 , 2.102924436744084e-46 ], [ 759.8125 , 2.1031971317612047e-46 ], [ 759.875 , 2.103469862139815e-46 ], [ 759.9375 , 2.103742627884516e-46 ], 
                    [ 760.0 , 2.104015428999923e-46 ], [ 760.0625 , 2.1042882654905633e-46 ], [ 760.125 , 2.1045611373610383e-46 ], [ 760.1875 , 2.1048340446159963e-46 ], 
                    [ 760.25 , 2.1051069872599664e-46 ], [ 760.3125 , 2.1053799652975063e-46 ], [ 760.375 , 2.105652978733312e-46 ], [ 760.4375 , 2.1059260275718675e-46 ], 
                    [ 760.5 , 2.1061991118178546e-46 ], [ 760.5625 , 2.106472231475789e-46 ], [ 760.625 , 2.106745386550308e-46 ], [ 760.6875 , 2.10701857704596e-46 ], 
                    [ 760.75 , 2.1072918029673967e-46 ], [ 760.8125 , 2.1075650643191828e-46 ], [ 760.875 , 2.107838361105898e-46 ], [ 760.9375 , 2.1081116933321666e-46 ], 
                    [ 761.0 , 2.1083850610025693e-46 ], [ 761.0625 , 2.108658464121688e-46 ], [ 761.125 , 2.1089319026941334e-46 ], [ 761.1875 , 2.1092051090166214e-46 ], 
                    [ 761.25 , 2.1094782565040226e-46 ], [ 761.3125 , 2.1097514393647264e-46 ], [ 761.375 , 2.1100246576033132e-46 ], [ 761.4375 , 2.11029791122435e-46 ], 
                    [ 761.5 , 2.1105712002324633e-46 ], [ 761.5625 , 2.1108445246321915e-46 ], [ 761.625 , 2.1111178844281175e-46 ], [ 761.6875 , 2.1113912796248553e-46 ], 
                    [ 761.75 , 2.1116647102269594e-46 ], [ 761.8125 , 2.1119381762390596e-46 ], [ 761.875 , 2.112211677665682e-46 ], [ 761.9375 , 2.1124852145114575e-46 ], 
                    [ 762.0 , 2.1127587867809586e-46 ], [ 762.0625 , 2.113032394478757e-46 ], [ 762.125 , 2.1133060376094713e-46 ], [ 762.1875 , 2.1135797161776747e-46 ], 
                    [ 762.25 , 2.113853430187957e-46 ], [ 762.3125 , 2.1141271796449226e-46 ], [ 762.375 , 2.1144009645531168e-46 ], [ 762.4375 , 2.1146747849171615e-46 ], 
                    [ 762.5 , 2.114948640741662e-46 ], [ 762.5625 , 2.115222532031167e-46 ], [ 762.625 , 2.1154964587903432e-46 ], [ 762.6875 , 2.115770421023724e-46 ], 
                    [ 762.75 , 2.1160444187359193e-46 ], [ 762.8125 , 2.1163184519315235e-46 ], [ 762.875 , 2.1165924735078955e-46 ], [ 762.9375 , 2.1168662312565143e-46 ], 
                    [ 763.0 , 2.1171400244126755e-46 ], [ 763.0625 , 2.117413852980898e-46 ], [ 763.125 , 2.117687716965837e-46 ], [ 763.1875 , 2.1179616163720286e-46 ], 
                    [ 763.25 , 2.1182355512040684e-46 ], [ 763.3125 , 2.1185095214665696e-46 ], [ 763.375 , 2.1187835271640687e-46 ], [ 763.4375 , 2.119057568301164e-46 ], 
                    [ 763.5 , 2.1193316448824247e-46 ], [ 763.5625 , 2.119605756912464e-46 ], [ 763.625 , 2.1198799043958236e-46 ], [ 763.6875 , 2.1201540873371625e-46 ], 
                    [ 763.75 , 2.120428305740977e-46 ], [ 763.8125 , 2.1207025596119595e-46 ], [ 763.875 , 2.1209768489545762e-46 ], [ 763.9375 , 2.121251173773506e-46 ], 
                    [ 764.0 , 2.1215255340732916e-46 ], [ 764.0625 , 2.121799929858553e-46 ], [ 764.125 , 2.122074361133864e-46 ], [ 764.1875 , 2.12234882790383e-46 ], 
                    [ 764.25 , 2.122623330172997e-46 ], [ 764.3125 , 2.122897867945986e-46 ], [ 764.375 , 2.1231724412273897e-46 ], [ 764.4375 , 2.1234470500218153e-46 ], 
                    [ 764.5 , 2.1237216943338264e-46 ], [ 764.5625 , 2.1239963741680766e-46 ], [ 764.625 , 2.124270920962954e-46 ], [ 764.6875 , 2.1245453128303486e-46 ], 
                    [ 764.75 , 2.1248197401409144e-46 ], [ 764.8125 , 2.125094202899214e-46 ], [ 764.875 , 2.1253687011098564e-46 ], [ 764.9375 , 2.1256432347774058e-46 ], 
                    [ 765.0 , 2.125917803906442e-46 ], [ 765.0625 , 2.1261924085015463e-46 ], [ 765.125 , 2.126467048567299e-46 ], [ 765.1875 , 2.1267417241082668e-46 ], 
                    [ 765.25 , 2.127016435129048e-46 ], [ 765.3125 , 2.1272911816342395e-46 ], [ 765.375 , 2.1275659636283956e-46 ], [ 765.4375 , 2.127840781116145e-46 ], 
                    [ 765.5 , 2.1281156341019972e-46 ], [ 765.5625 , 2.128390522590628e-46 ], [ 765.625 , 2.1286654465865627e-46 ], [ 765.6875 , 2.1289404060943883e-46 ], 
                    [ 765.75 , 2.1292154011187368e-46 ], [ 765.8125 , 2.1294904316641203e-46 ], [ 765.875 , 2.129765497735218e-46 ], [ 765.9375 , 2.1300405993365733e-46 ], 
                    [ 766.0 , 2.1303157364727604e-46 ], [ 766.0625 , 2.130590909148369e-46 ], [ 766.125 , 2.1308661173679906e-46 ], [ 766.1875 , 2.1311413611362613e-46 ], 
                    [ 766.25 , 2.1314166404577575e-46 ], [ 766.3125 , 2.1316919553370418e-46 ], [ 766.375 , 2.131967045547716e-46 ], [ 766.4375 , 2.132242080041907e-46 ], 
                    [ 766.5 , 2.132517150016903e-46 ], [ 766.5625 , 2.132792255477328e-46 ], [ 766.625 , 2.1330673964277287e-46 ], [ 766.6875 , 2.1333425728726987e-46 ], 
                    [ 766.75 , 2.133617784816832e-46 ], [ 766.8125 , 2.1338930322646784e-46 ], [ 766.875 , 2.1341683152208628e-46 ], [ 766.9375 , 2.134443633689906e-46 ], 
                    [ 767.0 , 2.1347189876764194e-46 ], [ 767.0625 , 2.134994377184984e-46 ], [ 767.125 , 2.1352698022202145e-46 ], [ 767.1875 , 2.135545262786632e-46 ], 
                    [ 767.25 , 2.135820758888851e-46 ], [ 767.3125 , 2.136096290531471e-46 ], [ 767.375 , 2.136371857719046e-46 ], [ 767.4375 , 2.136647460456177e-46 ], 
                    [ 767.5 , 2.136923098747481e-46 ], [ 767.5625 , 2.137198772597468e-46 ], [ 767.625 , 2.1374744820107863e-46 ], [ 767.6875 , 2.1377502269920236e-46 ], 
                    [ 767.75 , 2.1380260075457537e-46 ], [ 767.8125 , 2.1383018236765648e-46 ], [ 767.875 , 2.138577675389047e-46 ], [ 767.9375 , 2.138853562687775e-46 ], 
                    [ 768.0 , 2.1391294855773707e-46 ], [ 768.0625 , 2.1394054440624095e-46 ], [ 768.125 , 2.139681096354789e-46 ], [ 768.1875 , 2.1399567656564188e-46 ], 
                    [ 768.25 , 2.1402324704743066e-46 ], [ 768.3125 , 2.140508210813135e-46 ], [ 768.375 , 2.1407839866773884e-46 ], [ 768.4375 , 2.1410597980716603e-46 ], 
                    [ 768.5 , 2.141335645000573e-46 ], [ 768.5625 , 2.1416115274686438e-46 ], [ 768.625 , 2.1418874454805125e-46 ], [ 768.6875 , 2.142163399040713e-46 ], 
                    [ 768.75 , 2.14243938815384e-46 ], [ 768.8125 , 2.1427154128244894e-46 ], [ 768.875 , 2.1429914730572123e-46 ], [ 768.9375 , 2.143267568856605e-46 ], 
                    [ 769.0 , 2.1435437002272347e-46 ], [ 769.0625 , 2.1438198671737303e-46 ], [ 769.125 , 2.144096069700629e-46 ], [ 769.1875 , 2.1443723078125307e-46 ], 
                    [ 769.25 , 2.1446485815140355e-46 ], [ 769.3125 , 2.1449248908096817e-46 ], [ 769.375 , 2.145201235704117e-46 ], [ 769.4375 , 2.1454776162019125e-46 ], 
                    [ 769.5 , 2.1457540323075638e-46 ], [ 769.5625 , 2.1460304840257956e-46 ], [ 769.625 , 2.14630697136112e-46 ], [ 769.6875 , 2.1465834943181408e-46 ], 
                    [ 769.75 , 2.1468600529014626e-46 ], [ 769.8125 , 2.147136613349557e-46 ], [ 769.875 , 2.1474128830154254e-46 ], [ 769.9375 , 2.1476891882286183e-46 ], 
                    [ 770.0 , 2.147965528993663e-46 ], [ 770.0625 , 2.148241905315181e-46 ], [ 770.125 , 2.14851831719773e-46 ], [ 770.1875 , 2.1487947646458717e-46 ], 
                    [ 770.25 , 2.149071247664228e-46 ], [ 770.3125 , 2.149347766257329e-46 ], [ 770.375 , 2.149624320429769e-46 ], [ 770.4375 , 2.1499009101861243e-46 ], 
                    [ 770.5 , 2.1501775355309595e-46 ], [ 770.5625 , 2.150454196468883e-46 ], [ 770.625 , 2.15073089300446e-46 ], [ 770.6875 , 2.1510076251422706e-46 ], 
                    [ 770.75 , 2.151284392886865e-46 ], [ 770.8125 , 2.151561196242917e-46 ], [ 770.875 , 2.1518380352149014e-46 ], [ 770.9375 , 2.1521149098074616e-46 ], 
                    [ 771.0 , 2.1523918200251508e-46 ], [ 771.0625 , 2.152668765872598e-46 ], [ 771.125 , 2.1529457473543426e-46 ], [ 771.1875 , 2.1532227644749846e-46 ], 
                    [ 771.25 , 2.1534998172391094e-46 ], [ 771.3125 , 2.1537769056512884e-46 ], [ 771.375 , 2.1540540297161235e-46 ], [ 771.4375 , 2.1543311894382628e-46 ], 
                    [ 771.5 , 2.1546083848221418e-46 ], [ 771.5625 , 2.1548855514232772e-46 ], [ 771.625 , 2.1551624512335046e-46 ], [ 771.6875 , 2.155439386624943e-46 ], 
                    [ 771.75 , 2.155716357602196e-46 ], [ 771.8125 , 2.1559933641698517e-46 ], [ 771.875 , 2.1562704063324676e-46 ], [ 771.9375 , 2.1565474840946183e-46 ], 
                    [ 772.0 , 2.1568245974608466e-46 ], [ 772.0625 , 2.1571017464357895e-46 ], [ 772.125 , 2.1573789310239774e-46 ], [ 772.1875 , 2.1576561512300003e-46 ], 
                    [ 772.25 , 2.157933407058452e-46 ], [ 772.3125 , 2.1582106985138933e-46 ], [ 772.375 , 2.1584880256008874e-46 ], [ 772.4375 , 2.1587653883240287e-46 ], 
                    [ 772.5 , 2.159042786687911e-46 ], [ 772.5625 , 2.159320220697099e-46 ], [ 772.625 , 2.159597690356158e-46 ], [ 772.6875 , 2.1598751956696686e-46 ], 
                    [ 772.75 , 2.1601527366422735e-46 ], [ 772.8125 , 2.160430313278479e-46 ], [ 772.875 , 2.160707925582913e-46 ], [ 772.9375 , 2.160985573560113e-46 ], 
                    [ 773.0 , 2.161263257214725e-46 ], [ 773.0625 , 2.1615409765513014e-46 ], [ 773.125 , 2.1618187315744133e-46 ], [ 773.1875 , 2.1620965222886764e-46 ], 
                    [ 773.25 , 2.1623743486986615e-46 ], [ 773.3125 , 2.1626521416680364e-46 ], [ 773.375 , 2.1629296837655882e-46 ], [ 773.4375 , 2.163207261481235e-46 ], 
                    [ 773.5 , 2.1634848748196093e-46 ], [ 773.5625 , 2.163762523785252e-46 ], [ 773.625 , 2.16404020838272e-46 ], [ 773.6875 , 2.1643179286166163e-46 ], 
                    [ 773.75 , 2.1645956844914844e-46 ], [ 773.8125 , 2.1648734760119126e-46 ], [ 773.875 , 2.165151303182492e-46 ], [ 773.9375 , 2.1654291660077662e-46 ], 
                    [ 774.0 , 2.1657070644923417e-46 ], [ 774.0625 , 2.165984998640764e-46 ], [ 774.125 , 2.1662629684576413e-46 ], [ 774.1875 , 2.1665409739475195e-46 ], 
                    [ 774.25 , 2.166819015114993e-46 ], [ 774.3125 , 2.167097091964655e-46 ], [ 774.375 , 2.1673752045011005e-46 ], [ 774.4375 , 2.1676533527288324e-46 ], 
                    [ 774.5 , 2.1679315366524466e-46 ], [ 774.5625 , 2.168209756276616e-46 ], [ 774.625 , 2.1684880116058306e-46 ], [ 774.6875 , 2.1687663026447023e-46 ], 
                    [ 774.75 , 2.1690446293978458e-46 ], [ 774.8125 , 2.1693229918697667e-46 ], [ 774.875 , 2.1696013900651262e-46 ], [ 774.9375 , 2.1698798239884624e-46 ], 
                    [ 775.0 , 2.170158293644392e-46 ], [ 775.0625 , 2.1704367430859296e-46 ], [ 775.125 , 2.170714910158032e-46 ], [ 775.1875 , 2.1709931128804954e-46 ], 
                    [ 775.25 , 2.1712713512579347e-46 ], [ 775.3125 , 2.1715496252949196e-46 ], [ 775.375 , 2.1718279349959903e-46 ], [ 775.4375 , 2.1721062803657316e-46 ], 
                    [ 775.5 , 2.1723846614087003e-46 ], [ 775.5625 , 2.1726630781295298e-46 ], [ 775.625 , 2.1729415305327003e-46 ], [ 775.6875 , 2.1732200186228463e-46 ], 
                    [ 775.75 , 2.173498542404542e-46 ], [ 775.8125 , 2.1737771018823453e-46 ], [ 775.875 , 2.1740556970608014e-46 ], [ 775.9375 , 2.1743343279445466e-46 ], 
                    [ 776.0 , 2.1746129945380953e-46 ], [ 776.0625 , 2.174891696846071e-46 ], [ 776.125 , 2.1751704348730503e-46 ], [ 776.1875 , 2.1754492086235808e-46 ], 
                    [ 776.25 , 2.175728018102256e-46 ], [ 776.3125 , 2.1760068633136548e-46 ], [ 776.375 , 2.1762857442623567e-46 ], [ 776.4375 , 2.176564660952942e-46 ], 
                    [ 776.5 , 2.1768436133899914e-46 ], [ 776.5625 , 2.177122601578102e-46 ], [ 776.625 , 2.17740162552184e-46 ], [ 776.6875 , 2.177680685225772e-46 ], 
                    [ 776.75 , 2.1779597806944814e-46 ], [ 776.8125 , 2.1782388967824246e-46 ], [ 776.875 , 2.178517696735793e-46 ], [ 776.9375 , 2.1787965323736517e-46 ], 
                    [ 777.0 , 2.179075403700616e-46 ], [ 777.0625 , 2.179354310721253e-46 ], [ 777.125 , 2.1796332534401004e-46 ], [ 777.1875 , 2.1799122318617427e-46 ], 
                    [ 777.25 , 2.1801912459907503e-46 ], [ 777.3125 , 2.1804702958316925e-46 ], [ 777.375 , 2.1807493813891256e-46 ], [ 777.4375 , 2.181028502667636e-46 ], 
                    [ 777.5 , 2.181307659671843e-46 ], [ 777.5625 , 2.18158685240624e-46 ], [ 777.625 , 2.1818660808754486e-46 ], [ 777.6875 , 2.1821453450839954e-46 ], 
                    [ 777.75 , 2.1824246450365327e-46 ], [ 777.8125 , 2.182703980737526e-46 ], [ 777.875 , 2.1829833521916296e-46 ], [ 777.9375 , 2.1832627594034336e-46 ], 
                    [ 778.0 , 2.18354220237747e-46 ], [ 778.0625 , 2.1838216811183462e-46 ], [ 778.125 , 2.184101195630609e-46 ], [ 778.1875 , 2.184380745918837e-46 ], 
                    [ 778.25 , 2.1846603319876406e-46 ], [ 778.3125 , 2.1849399538416145e-46 ], [ 778.375 , 2.1852196114852465e-46 ], [ 778.4375 , 2.1854993049232103e-46 ], 
                    [ 778.5 , 2.1857790341600867e-46 ], [ 778.5625 , 2.1860587992003813e-46 ], [ 778.625 , 2.1863382828966357e-46 ], [ 778.6875 , 2.1866177546257878e-46 ], 
                    [ 778.75 , 2.186897262078802e-46 ], [ 778.8125 , 2.1871768052602605e-46 ], [ 778.875 , 2.187456384174699e-46 ], [ 778.9375 , 2.1877359988267006e-46 ], 
                    [ 779.0 , 2.1880156492208336e-46 ], [ 779.0625 , 2.1882953353616824e-46 ], [ 779.125 , 2.1885750572537854e-46 ], [ 779.1875 , 2.1888548149017124e-46 ], 
                    [ 779.25 , 2.1891346083100807e-46 ], [ 779.3125 , 2.1894144374834146e-46 ], [ 779.375 , 2.189694302426317e-46 ], [ 779.4375 , 2.1899742031433287e-46 ], 
                    [ 779.5 , 2.1902541396390388e-46 ], [ 779.5625 , 2.190534111918036e-46 ], [ 779.625 , 2.1908141199848476e-46 ], [ 779.6875 , 2.191094163844111e-46 ], 
                    [ 779.75 , 2.191374243500354e-46 ], [ 779.8125 , 2.1916543589581683e-46 ], [ 779.875 , 2.1919345102221612e-46 ], [ 779.9375 , 2.1922146972968476e-46 ], 
                    [ 780.0 , 2.1924949201868055e-46 ], [ 780.0625 , 2.19277517889669e-46 ], [ 780.125 , 2.193055473431049e-46 ], [ 780.1875 , 2.1933358037944153e-46 ], 
                    [ 780.25 , 2.1936161699913688e-46 ], [ 780.3125 , 2.1938965720265836e-46 ], [ 780.375 , 2.1941767758502927e-46 ], [ 780.4375 , 2.194456880001406e-46 ], 
                    [ 780.5 , 2.1947370199100467e-46 ], [ 780.5625 , 2.1950171955808105e-46 ], [ 780.625 , 2.1952974070181848e-46 ], [ 780.6875 , 2.195577654226829e-46 ], 
                    [ 780.75 , 2.195857937211215e-46 ], [ 780.8125 , 2.19613825597599e-46 ], [ 780.875 , 2.196418610525657e-46 ], [ 780.9375 , 2.1966990008648636e-46 ], 
                    [ 781.0 , 2.1969794269981162e-46 ], [ 781.0625 , 2.197259888929968e-46 ], [ 781.125 , 2.1975403866650515e-46 ], [ 781.1875 , 2.1978209202079064e-46 ], 
                    [ 781.25 , 2.198101489563088e-46 ], [ 781.3125 , 2.1983820947351835e-46 ], [ 781.375 , 2.1986627357287815e-46 ], [ 781.4375 , 2.1989434125484383e-46 ], 
                    [ 781.5 , 2.1992241251987127e-46 ], [ 781.5625 , 2.1995048736842092e-46 ], [ 781.625 , 2.1997856580094716e-46 ], [ 781.6875 , 2.2000664781790755e-46 ], 
                    [ 781.75 , 2.200347334197643e-46 ], [ 781.8125 , 2.2006282260697347e-46 ], [ 781.875 , 2.200909153799866e-46 ], [ 781.9375 , 2.2011901173926914e-46 ], 
                    [ 782.0 , 2.201471116852759e-46 ], [ 782.0625 , 2.2017521521846155e-46 ], [ 782.125 , 2.2020330973600225e-46 ], [ 782.1875 , 2.20231382658863e-46 ], 
                    [ 782.25 , 2.2025945916063597e-46 ], [ 782.3125 , 2.202875392417805e-46 ], [ 782.375 , 2.2031562290275604e-46 ], [ 782.4375 , 2.2034371014401435e-46 ], 
                    [ 782.5 , 2.203718009660149e-46 ], [ 782.5625 , 2.203998953692095e-46 ], [ 782.625 , 2.2042799335405793e-46 ], [ 782.6875 , 2.204560949210183e-46 ], 
                    [ 782.75 , 2.204842000705441e-46 ], [ 782.8125 , 2.2051230880309686e-46 ], [ 782.875 , 2.205404211191286e-46 ], [ 782.9375 , 2.205685370190978e-46 ], 
                    [ 783.0 , 2.2059665650345967e-46 ], [ 783.0625 , 2.206247795726728e-46 ], [ 783.125 , 2.2065290622719898e-46 ], [ 783.1875 , 2.206810364674858e-46 ], 
                    [ 783.25 , 2.207091702939967e-46 ], [ 783.3125 , 2.2073730770718574e-46 ], [ 783.375 , 2.2076544870751335e-46 ], [ 783.4375 , 2.2079359329543673e-46 ], 
                    [ 783.5 , 2.208217414714102e-46 ], [ 783.5625 , 2.208498932358943e-46 ], [ 783.625 , 2.208780485893418e-46 ], [ 783.6875 , 2.209062075322134e-46 ], 
                    [ 783.75 , 2.209343700649667e-46 ], [ 783.8125 , 2.2096253618805938e-46 ], [ 783.875 , 2.2099070590195223e-46 ], [ 783.9375 , 2.2101884278825645e-46 ], 
                    [ 784.0 , 2.210469821150049e-46 ], [ 784.0625 , 2.2107512502434844e-46 ], [ 784.125 , 2.211032715167495e-46 ], [ 784.1875 , 2.2113142159265786e-46 ], 
                    [ 784.25 , 2.2115957525253456e-46 ], [ 784.3125 , 2.211877324968343e-46 ], [ 784.375 , 2.212158933260135e-46 ], [ 784.4375 , 2.2124405774052847e-46 ], 
                    [ 784.5 , 2.2127222574083733e-46 ], [ 784.5625 , 2.2130039732739183e-46 ], [ 784.625 , 2.2132857250065333e-46 ], [ 784.6875 , 2.213567512610769e-46 ], 
                    [ 784.75 , 2.2138493360911766e-46 ], [ 784.8125 , 2.214131195452355e-46 ], [ 784.875 , 2.2144130906988567e-46 ], [ 784.9375 , 2.214695021835267e-46 ], 
                    [ 785.0 , 2.214976988866091e-46 ], [ 785.0625 , 2.215258991795994e-46 ], [ 785.125 , 2.2155410306294676e-46 ], [ 785.1875 , 2.215823105371162e-46 ], 
                    [ 785.25 , 2.216105216025554e-46 ], [ 785.3125 , 2.2163873625972783e-46 ], [ 785.375 , 2.2166695450908617e-46 ], [ 785.4375 , 2.21695176351094e-46 ], 
                    [ 785.5 , 2.2172340178620717e-46 ], [ 785.5625 , 2.2175163081487998e-46 ], [ 785.625 , 2.2177986343757463e-46 ], [ 785.6875 , 2.218080790739493e-46 ], 
                    [ 785.75 , 2.218362808561518e-46 ], [ 785.8125 , 2.218644862240676e-46 ], [ 785.875 , 2.2189269517815726e-46 ], [ 785.9375 , 2.219209077188689e-46 ], 
                    [ 786.0 , 2.2194912384666483e-46 ], [ 786.0625 , 2.2197734356199957e-46 ], [ 786.125 , 2.2200556686532926e-46 ], [ 786.1875 , 2.220337937571148e-46 ], 
                    [ 786.25 , 2.2206202423780302e-46 ], [ 786.3125 , 2.2209025830785338e-46 ], [ 786.375 , 2.2211849596772376e-46 ], [ 786.4375 , 2.221467372178754e-46 ], 
                    [ 786.5 , 2.221749820587569e-46 ], [ 786.5625 , 2.2220323049083105e-46 ], [ 786.625 , 2.2223148251454815e-46 ], [ 786.6875 , 2.2225973813036807e-46 ], 
                    [ 786.75 , 2.222879973387475e-46 ], [ 786.8125 , 2.223162601401432e-46 ], [ 786.875 , 2.2234452653501365e-46 ], [ 786.9375 , 2.2237279652381404e-46 ], 
                    [ 787.0 , 2.2240107010699826e-46 ], [ 787.0625 , 2.22429347285028e-46 ], [ 787.125 , 2.224576280583588e-46 ], [ 787.1875 , 2.224859124274478e-46 ], 
                    [ 787.25 , 2.2251420039275363e-46 ], [ 787.3125 , 2.225424919547305e-46 ], [ 787.375 , 2.225707871138357e-46 ], [ 787.4375 , 2.225990840736324e-46 ], 
                    [ 787.5 , 2.2262734935867022e-46 ], [ 787.5625 , 2.2265561823278992e-46 ], [ 787.625 , 2.2268389069644565e-46 ], [ 787.6875 , 2.2271216675009638e-46 ], 
                    [ 787.75 , 2.227404463941964e-46 ], [ 787.8125 , 2.2276872962920152e-46 ], [ 787.875 , 2.227970164555647e-46 ], [ 787.9375 , 2.228253068737481e-46 ], 
                    [ 788.0 , 2.2285360088420323e-46 ], [ 788.0625 , 2.228818984873909e-46 ], [ 788.125 , 2.2291019968376095e-46 ], [ 788.1875 , 2.229385044737729e-46 ], 
                    [ 788.25 , 2.2296681285788457e-46 ], [ 788.3125 , 2.229951248365507e-46 ], [ 788.375 , 2.230234404102279e-46 ], [ 788.4375 , 2.2305175957937248e-46 ], 
                    [ 788.5 , 2.230800823444379e-46 ], [ 788.5625 , 2.2310840870588713e-46 ], [ 788.625 , 2.231367386641704e-46 ], [ 788.6875 , 2.231650722197524e-46 ], 
                    [ 788.75 , 2.2319340937308203e-46 ], [ 788.8125 , 2.2322175012462245e-46 ], [ 788.875 , 2.2325009447482416e-46 ], [ 788.9375 , 2.232784424241489e-46 ], 
                    [ 789.0 , 2.233067939730521e-46 ], [ 789.0625 , 2.2333514912198767e-46 ], [ 789.125 , 2.2336350787142223e-46 ], [ 789.1875 , 2.2339187022180185e-46 ], 
                    [ 789.25 , 2.2342021737476665e-46 ], [ 789.3125 , 2.234485486499575e-46 ], [ 789.375 , 2.2347688351775744e-46 ], [ 789.4375 , 2.2350522197862048e-46 ], 
                    [ 789.5 , 2.2353356403300064e-46 ], [ 789.5625 , 2.2356190968135687e-46 ], [ 789.625 , 2.2359025892414154e-46 ], [ 789.6875 , 2.2361861176181383e-46 ], 
                    [ 789.75 , 2.236469681948263e-46 ], [ 789.8125 , 2.2367532822363644e-46 ], [ 789.875 , 2.237036918487019e-46 ], [ 789.9375 , 2.2373205907047867e-46 ], 
                    [ 790.0 , 2.2376042988941806e-46 ], [ 790.0625 , 2.2378880430597944e-46 ], [ 790.125 , 2.2381718232061894e-46 ], [ 790.1875 , 2.23845563933796e-46 ], 
                    [ 790.25 , 2.2387394914596067e-46 ], [ 790.3125 , 2.2390233795757567e-46 ], [ 790.375 , 2.239307303690926e-46 ], [ 790.4375 , 2.2395912638096636e-46 ], 
                    [ 790.5 , 2.2398752599365833e-46 ], [ 790.5625 , 2.240159292076267e-46 ], [ 790.625 , 2.2404433602332174e-46 ], [ 790.6875 , 2.24072746441205e-46 ], 
                    [ 790.75 , 2.241011604617316e-46 ], [ 790.8125 , 2.2412957808535846e-46 ], [ 790.875 , 2.2415799931253923e-46 ], [ 790.9375 , 2.2418642414373726e-46 ], 
                    [ 791.0 , 2.2421485257940643e-46 ], [ 791.0625 , 2.2424325122790903e-46 ], [ 791.125 , 2.242716484442122e-46 ], [ 791.1875 , 2.2430004925661712e-46 ], 
                    [ 791.25 , 2.243284536655807e-46 ], [ 791.3125 , 2.2435686167156004e-46 ], [ 791.375 , 2.2438527327500582e-46 ], [ 791.4375 , 2.2441368847638155e-46 ], 
                    [ 791.5 , 2.24442107276135e-46 ], [ 791.5625 , 2.2447052967472497e-46 ], [ 791.625 , 2.244989556726088e-46 ], [ 791.6875 , 2.2452738527023912e-46 ], 
                    [ 791.75 , 2.245558184680733e-46 ], [ 791.8125 , 2.2458425526656898e-46 ], [ 791.875 , 2.2461269566618206e-46 ], [ 791.9375 , 2.246411396673606e-46 ], 
                    [ 792.0 , 2.246695872705734e-46 ], [ 792.0625 , 2.2469803847626873e-46 ], [ 792.125 , 2.2472649328490116e-46 ], [ 792.1875 , 2.247549516969349e-46 ], 
                    [ 792.25 , 2.2478341371281827e-46 ], [ 792.3125 , 2.248118793330125e-46 ], [ 792.375 , 2.2484034855796926e-46 ], [ 792.4375 , 2.248688213881513e-46 ], 
                    [ 792.5 , 2.2489729782400732e-46 ], [ 792.5625 , 2.24925777866005e-46 ], [ 792.625 , 2.2495426151459307e-46 ], [ 792.6875 , 2.249827487702315e-46 ], 
                    [ 792.75 , 2.2501123963336897e-46 ], [ 792.8125 , 2.2503972713952633e-46 ], [ 792.875 , 2.2506818696119327e-46 ], [ 792.9375 , 2.25096650382055e-46 ], 
                    [ 793.0 , 2.2512511740256185e-46 ], [ 793.0625 , 2.251535880231739e-46 ], [ 793.125 , 2.2518206224434642e-46 ], [ 793.1875 , 2.252105400665315e-46 ], 
                    [ 793.25 , 2.2523902149018624e-46 ], [ 793.3125 , 2.2526750651576612e-46 ], [ 793.375 , 2.2529599514372812e-46 ], [ 793.4375 , 2.2532448737452313e-46 ], 
                    [ 793.5 , 2.2535298320861155e-46 ], [ 793.5625 , 2.253814826464491e-46 ], [ 793.625 , 2.254099856884899e-46 ], [ 793.6875 , 2.2543849233518816e-46 ], 
                    [ 793.75 , 2.254670025870014e-46 ], [ 793.8125 , 2.2549551644438705e-46 ], [ 793.875 , 2.2552403390779956e-46 ], [ 793.9375 , 2.2555255497769335e-46 ], 
                    [ 794.0 , 2.2558107965452292e-46 ], [ 794.0625 , 2.2560960793875238e-46 ], [ 794.125 , 2.2563813983083006e-46 ], [ 794.1875 , 2.2566667533122007e-46 ], 
                    [ 794.25 , 2.2569521444037247e-46 ], [ 794.3125 , 2.2572375715874356e-46 ], [ 794.375 , 2.2575230348679145e-46 ], [ 794.4375 , 2.257808534249742e-46 ], 
                    [ 794.5 , 2.2580940697374515e-46 ], [ 794.5625 , 2.2583796413356096e-46 ], [ 794.625 , 2.2586650809528532e-46 ], [ 794.6875 , 2.258950333866867e-46 ], 
                    [ 794.75 , 2.2592356228062097e-46 ], [ 794.8125 , 2.2595209477754952e-46 ], [ 794.875 , 2.259806308779226e-46 ], [ 794.9375 , 2.2600917058219686e-46 ], 
                    [ 795.0 , 2.2603771389083066e-46 ], [ 795.0625 , 2.2606626080427123e-46 ], [ 795.125 , 2.2609481132298015e-46 ], [ 795.1875 , 2.2612336544741127e-46 ], 
                    [ 795.25 , 2.2615192317802144e-46 ], [ 795.3125 , 2.2618048451526296e-46 ], [ 795.375 , 2.2620904945959614e-46 ], [ 795.4375 , 2.2623761801146845e-46 ], 
                    [ 795.5 , 2.2626619017134195e-46 ], [ 795.5625 , 2.2629476593967072e-46 ], [ 795.625 , 2.2632334531691368e-46 ], [ 795.6875 , 2.2635192830351857e-46 ], 
                    [ 795.75 , 2.2638051489994923e-46 ], [ 795.8125 , 2.2640910510665995e-46 ], [ 795.875 , 2.264376989240987e-46 ], [ 795.9375 , 2.264662963527311e-46 ], 
                    [ 796.0 , 2.2649489739301165e-46 ], [ 796.0625 , 2.265235020453916e-46 ], [ 796.125 , 2.2655211031033515e-46 ], [ 796.1875 , 2.265807221882922e-46 ], 
                    [ 796.25 , 2.266093376797173e-46 ], [ 796.3125 , 2.2663795678507183e-46 ], [ 796.375 , 2.2666657950481202e-46 ], [ 796.4375 , 2.26695182303719e-46 ], 
                    [ 796.5 , 2.267237738490261e-46 ], [ 796.5625 , 2.2675236900039336e-46 ], [ 796.625 , 2.2678096775827714e-46 ], [ 796.6875 , 2.2680957012312917e-46 ], 
                    [ 796.75 , 2.2683817609540754e-46 ], [ 796.8125 , 2.26866785675564e-46 ], [ 796.875 , 2.2689539886405527e-46 ], [ 796.9375 , 2.2692401566133794e-46 ], 
                    [ 797.0 , 2.269526360678657e-46 ], [ 797.0625 , 2.2698126008409366e-46 ], [ 797.125 , 2.270098877104755e-46 ], [ 797.1875 , 2.270385189474681e-46 ], 
                    [ 797.25 , 2.2706715379552862e-46 ], [ 797.3125 , 2.270957922551091e-46 ], [ 797.375 , 2.2712443432666506e-46 ], [ 797.4375 , 2.2715308001065857e-46 ], 
                    [ 797.5 , 2.2718172930754204e-46 ], [ 797.5625 , 2.2721038221776297e-46 ], [ 797.625 , 2.2723903874178525e-46 ], [ 797.6875 , 2.2726769888006137e-46 ], 
                    [ 797.75 , 2.2729636263305204e-46 ], [ 797.8125 , 2.2732503000120996e-46 ], [ 797.875 , 2.2735370098498943e-46 ], [ 797.9375 , 2.273823755848497e-46 ], 
                    [ 798.0 , 2.2741105380124525e-46 ], [ 798.0625 , 2.2743973563463055e-46 ], [ 798.125 , 2.2746842108546183e-46 ], [ 798.1875 , 2.2749711015420017e-46 ], 
                    [ 798.25 , 2.275257741042399e-46 ], [ 798.3125 , 2.2755443106591454e-46 ], [ 798.375 , 2.2758309163694646e-46 ], [ 798.4375 , 2.276117558177855e-46 ], 
                    [ 798.5 , 2.2764042360889107e-46 ], [ 798.5625 , 2.2766909501071797e-46 ], [ 798.625 , 2.2769777002371773e-46 ], [ 798.6875 , 2.2772644864834514e-46 ], 
                    [ 798.75 , 2.2775513088505997e-46 ], [ 798.8125 , 2.2778381673431226e-46 ], [ 798.875 , 2.278125061965587e-46 ], [ 798.9375 , 2.278411992722559e-46 ], 
                    [ 799.0 , 2.2786989596185415e-46 ], [ 799.0625 , 2.278985962658135e-46 ], [ 799.125 , 2.2792730018458757e-46 ], [ 799.1875 , 2.2795600771863318e-46 ], 
                    [ 799.25 , 2.279847188684009e-46 ], [ 799.3125 , 2.280134336343526e-46 ], [ 799.375 , 2.280421520169372e-46 ], [ 799.4375 , 2.2807087401661516e-46 ], 
                    [ 799.5 , 2.28099599633842e-46 ], [ 799.5625 , 2.281283288690685e-46 ], [ 799.625 , 2.281570617227552e-46 ], [ 799.6875 , 2.2818579819535627e-46 ], 
                    [ 799.75 , 2.2821453828732746e-46 ], [ 799.8125 , 2.282432819991263e-46 ], [ 799.875 , 2.2827202933120384e-46 ], [ 799.9375 , 2.283007802840241e-46 ], 
                    [ 800.0 , 2.2832953485803347e-46 ], [ 800.0625 , 2.283582627604677e-46 ], [ 800.125 , 2.2838698586704663e-46 ], [ 800.1875 , 2.284157125864403e-46 ], 
                    [ 800.25 , 2.2844444291910648e-46 ], [ 800.3125 , 2.2847317686550116e-46 ], [ 800.375 , 2.285019144260758e-46 ], [ 800.4375 , 2.2853065560128646e-46 ], 
                    [ 800.5 , 2.2855940039158786e-46 ], [ 800.5625 , 2.285881487974363e-46 ], [ 800.625 , 2.2861690081928495e-46 ], [ 800.6875 , 2.2864565645758863e-46 ], 
                    [ 800.75 , 2.2867441571279897e-46 ], [ 800.8125 , 2.2870317858537907e-46 ], [ 800.875 , 2.2873194507577576e-46 ], [ 800.9375 , 2.2876071518444892e-46 ], 
                    [ 801.0 , 2.2878948891184888e-46 ], [ 801.0625 , 2.288182662584389e-46 ], [ 801.125 , 2.2884704722466436e-46 ], [ 801.1875 , 2.2887583181099047e-46 ], 
                    [ 801.25 , 2.2890462001786267e-46 ], [ 801.3125 , 2.2893341184574617e-46 ], [ 801.375 , 2.289622072950883e-46 ], [ 801.4375 , 2.2899100636634944e-46 ], 
                    [ 801.5 , 2.290198090599819e-46 ], [ 801.5625 , 2.290486153764429e-46 ], [ 801.625 , 2.2907742531618816e-46 ], [ 801.6875 , 2.2910623887967505e-46 ], 
                    [ 801.75 , 2.291350560673561e-46 ], [ 801.8125 , 2.291638768796921e-46 ], [ 801.875 , 2.291926705075259e-46 ], [ 801.9375 , 2.292214585033687e-46 ], 
                    [ 802.0 , 2.292502501151592e-46 ], [ 802.0625 , 2.2927904534335e-46 ], [ 802.125 , 2.293078441883952e-46 ], [ 802.1875 , 2.2933664665075245e-46 ], 
                    [ 802.25 , 2.2936545273087284e-46 ], [ 802.3125 , 2.2939426242921565e-46 ], [ 802.375 , 2.2942307574623213e-46 ], [ 802.4375 , 2.294518926823751e-46 ], 
                    [ 802.5 , 2.2948071323810088e-46 ], [ 802.5625 , 2.295095374138624e-46 ], [ 802.625 , 2.295383652101193e-46 ], [ 802.6875 , 2.295671966273214e-46 ], 
                    [ 802.75 , 2.2959603166592514e-46 ], [ 802.8125 , 2.2962487032638544e-46 ], [ 802.875 , 2.2965371260915885e-46 ], [ 802.9375 , 2.296825585146971e-46 ], 
                    [ 803.0 , 2.297114080434551e-46 ], [ 803.0625 , 2.297402611958914e-46 ], [ 803.125 , 2.2976911797245944e-46 ], [ 803.1875 , 2.2979797837361605e-46 ], 
                    [ 803.25 , 2.2982684239981487e-46 ], [ 803.3125 , 2.2985571005150964e-46 ], [ 803.375 , 2.2988458132915574e-46 ], [ 803.4375 , 2.299134562332118e-46 ], 
                    [ 803.5 , 2.2994233476413177e-46 ], [ 803.5625 , 2.299712169223728e-46 ], [ 803.625 , 2.3000010270838724e-46 ], [ 803.6875 , 2.3002896547202475e-46 ], 
                    [ 803.75 , 2.3005781986925935e-46 ], [ 803.8125 , 2.300866778859332e-46 ], [ 803.875 , 2.301155395224987e-46 ], [ 803.9375 , 2.3014440477941477e-46 ], 
                    [ 804.0 , 2.3017327365713394e-46 ], [ 804.0625 , 2.30202146156112e-46 ], [ 804.125 , 2.302310222768e-46 ], [ 804.1875 , 2.3025990201965542e-46 ], 
                    [ 804.25 , 2.3028878538512937e-46 ], [ 804.3125 , 2.3031767237367633e-46 ], [ 804.375 , 2.3034656298575557e-46 ], [ 804.4375 , 2.3037545722181682e-46 ], 
                    [ 804.5 , 2.3040435508231786e-46 ], [ 804.5625 , 2.3043325656771006e-46 ], [ 804.625 , 2.3046216167844982e-46 ], [ 804.6875 , 2.304910704149935e-46 ], 
                    [ 804.75 , 2.305199827777959e-46 ], [ 804.8125 , 2.3054889876730374e-46 ], [ 804.875 , 2.3057781838398005e-46 ], [ 804.9375 , 2.3060674162827994e-46 ], 
                    [ 805.0 , 2.306356685006583e-46 ], [ 805.0625 , 2.306645990015704e-46 ], [ 805.125 , 2.30693533131468e-46 ], [ 805.1875 , 2.3072247089080636e-46 ], 
                    [ 805.25 , 2.307514122800407e-46 ], [ 805.3125 , 2.3078035729962974e-46 ], [ 805.375 , 2.3080930595002875e-46 ], [ 805.4375 , 2.308382582316883e-46 ], 
                    [ 805.5 , 2.308671921610696e-46 ], [ 805.5625 , 2.3089611133452987e-46 ], [ 805.625 , 2.309250341304996e-46 ], [ 805.6875 , 2.3095396054943095e-46 ], 
                    [ 805.75 , 2.3098289059178263e-46 ], [ 805.8125 , 2.3101182425800368e-46 ], [ 805.875 , 2.310407615485496e-46 ], [ 805.9375 , 2.3106970246387608e-46 ], 
                    [ 806.0 , 2.3109864700443217e-46 ], [ 806.0625 , 2.3112759517068025e-46 ], [ 806.125 , 2.311565469630663e-46 ], [ 806.1875 , 2.311855023820461e-46 ], 
                    [ 806.25 , 2.3121446142807565e-46 ], [ 806.3125 , 2.3124342410161253e-46 ], [ 806.375 , 2.3127239040310786e-46 ], [ 806.4375 , 2.3130136033301442e-46 ], 
                    [ 806.5 , 2.3133033389178678e-46 ], [ 806.5625 , 2.313593110798844e-46 ], [ 806.625 , 2.313882918977585e-46 ], [ 806.6875 , 2.314172763458607e-46 ], 
                    [ 806.75 , 2.314462644246505e-46 ], [ 806.8125 , 2.314752561345827e-46 ], [ 806.875 , 2.315042514761057e-46 ], [ 806.9375 , 2.3153325044968575e-46 ], 
                    [ 807.0 , 2.31562253055768e-46 ], [ 807.0625 , 2.315912592948092e-46 ], [ 807.125 , 2.3162026916726925e-46 ], [ 807.1875 , 2.3164928267359675e-46 ], 
                    [ 807.25 , 2.316782998142535e-46 ], [ 807.3125 , 2.317073074362997e-46 ], [ 807.375 , 2.317362923739001e-46 ], [ 807.4375 , 2.317652809373101e-46 ], 
                    [ 807.5 , 2.317942731269818e-46 ], [ 807.5625 , 2.3182326894337208e-46 ], [ 807.625 , 2.318522683869329e-46 ], [ 807.6875 , 2.3188127145811806e-46 ], 
                    [ 807.75 , 2.319102781573813e-46 ], [ 807.8125 , 2.319392884851765e-46 ], [ 807.875 , 2.3196830244195757e-46 ], [ 807.9375 , 2.3199732002817676e-46 ], 
                    [ 808.0 , 2.3202634124429146e-46 ], [ 808.0625 , 2.320553660907557e-46 ], [ 808.125 , 2.3208439456802034e-46 ], [ 808.1875 , 2.321134266765412e-46 ], 
                    [ 808.25 , 2.3214246241676924e-46 ], [ 808.3125 , 2.3217150178916533e-46 ], [ 808.375 , 2.3220054479418386e-46 ], [ 808.4375 , 2.32229591432271e-46 ], 
                    [ 808.5 , 2.3225864170388775e-46 ], [ 808.5625 , 2.322876956094871e-46 ], [ 808.625 , 2.323167531495219e-46 ], [ 808.6875 , 2.323458143244501e-46 ], 
                    [ 808.75 , 2.323748791347248e-46 ], [ 808.8125 , 2.3240394758080073e-46 ], [ 808.875 , 2.3243301966313098e-46 ], [ 808.9375 , 2.3246209538217386e-46 ], 
                    [ 809.0 , 2.3249117473838586e-46 ], [ 809.0625 , 2.325202577322137e-46 ], [ 809.125 , 2.325493421108519e-46 ], [ 809.1875 , 2.3257839202376882e-46 ], 
                    [ 809.25 , 2.3260744556558375e-46 ], [ 809.3125 , 2.326365027367466e-46 ], [ 809.375 , 2.326655635377124e-46 ], [ 809.4375 , 2.3269462796893457e-46 ], 
                    [ 809.5 , 2.3272369603086834e-46 ], [ 809.5625 , 2.3275276772396385e-46 ], [ 809.625 , 2.327818430486764e-46 ], [ 809.6875 , 2.328109220054597e-46 ], 
                    [ 809.75 , 2.3284000459477065e-46 ], [ 809.8125 , 2.328690908170549e-46 ], [ 809.875 , 2.328981806727711e-46 ], [ 809.9375 , 2.3292727416237654e-46 ], 
                    [ 810.0 , 2.3295637128632182e-46 ], [ 810.0625 , 2.329854720450576e-46 ], [ 810.125 , 2.330145764390446e-46 ], [ 810.1875 , 2.3304368446873197e-46 ], 
                    [ 810.25 , 2.3307279613457388e-46 ], [ 810.3125 , 2.3310191143702617e-46 ], [ 810.375 , 2.3313103037654813e-46 ], [ 810.4375 , 2.3316015295359077e-46 ], 
                    [ 810.5 , 2.331892791686002e-46 ], [ 810.5625 , 2.332184090220359e-46 ], [ 810.625 , 2.3324754251435893e-46 ], [ 810.6875 , 2.332766796460189e-46 ], 
                    [ 810.75 , 2.3330582041747046e-46 ], [ 810.8125 , 2.3333496482916333e-46 ], [ 810.875 , 2.3336411288156213e-46 ], [ 810.9375 , 2.3339326457511675e-46 ], 
                    [ 811.0 , 2.334223908454646e-46 ], [ 811.0625 , 2.3345150899259998e-46 ], [ 811.125 , 2.3348063077206084e-46 ], [ 811.1875 , 2.335097561843037e-46 ], 
                    [ 811.25 , 2.3353888522978002e-46 ], [ 811.3125 , 2.3356801790894133e-46 ], [ 811.375 , 2.3359715422224427e-46 ], [ 811.4375 , 2.3362629417014047e-46 ], 
                    [ 811.5 , 2.336554377530834e-46 ], [ 811.5625 , 2.3368458497152483e-46 ], [ 811.625 , 2.3371373582592322e-46 ], [ 811.6875 , 2.3374289031672383e-46 ], 
                    [ 811.75 , 2.337720484443903e-46 ], [ 811.8125 , 2.3380121020936963e-46 ], [ 811.875 , 2.338303756121155e-46 ], [ 811.9375 , 2.3385954465308513e-46 ], 
                    [ 812.0 , 2.338887173327323e-46 ], [ 812.0625 , 2.339178936515026e-46 ], [ 812.125 , 2.3394707360986337e-46 ], [ 812.1875 , 2.3397625720826184e-46 ], 
                    [ 812.25 , 2.3400544444715053e-46 ], [ 812.3125 , 2.340346353269819e-46 ], [ 812.375 , 2.340638298482201e-46 ], [ 812.4375 , 2.340930280113077e-46 ], 
                    [ 812.5 , 2.3412222981670735e-46 ], [ 812.5625 , 2.341514352648651e-46 ], [ 812.625 , 2.341806443562436e-46 ], [ 812.6875 , 2.342098570912908e-46 ], 
                    [ 812.75 , 2.3423907347046946e-46 ], [ 812.8125 , 2.342682812954018e-46 ], [ 812.875 , 2.3429746451059604e-46 ], [ 812.9375 , 2.343266513611952e-46 ], 
                    [ 813.0 , 2.3435584184765374e-46 ], [ 813.0625 , 2.343850359704196e-46 ], [ 813.125 , 2.3441423372995074e-46 ], [ 813.1875 , 2.3444343512670027e-46 ], 
                    [ 813.25 , 2.3447264016111625e-46 ], [ 813.3125 , 2.3450184883365846e-46 ], [ 813.375 , 2.3453106114477677e-46 ], [ 813.4375 , 2.3456027709492287e-46 ], 
                    [ 813.5 , 2.345894966845516e-46 ], [ 813.5625 , 2.3461871991411646e-46 ], [ 813.625 , 2.346479467840709e-46 ], [ 813.6875 , 2.3467717729486994e-46 ], 
                    [ 813.75 , 2.347064114469639e-46 ], [ 813.8125 , 2.3473564924080973e-46 ], [ 813.875 , 2.3476489067685603e-46 ], [ 813.9375 , 2.347941357555632e-46 ], 
                    [ 814.0 , 2.348233844773783e-46 ], [ 814.0625 , 2.3485263684276194e-46 ], [ 814.125 , 2.348818928521612e-46 ], [ 814.1875 , 2.349111525060368e-46 ], 
                    [ 814.25 , 2.3494041580483938e-46 ], [ 814.3125 , 2.349696827490213e-46 ], [ 814.375 , 2.349989533390384e-46 ], [ 814.4375 , 2.3502822757534638e-46 ], 
                    [ 814.5 , 2.3505750545839624e-46 ], [ 814.5625 , 2.3508678698864225e-46 ], [ 814.625 , 2.3511607216654543e-46 ], [ 814.6875 , 2.3514532759241994e-46 ], 
                    [ 814.75 , 2.3517457955510193e-46 ], [ 814.8125 , 2.3520383515671696e-46 ], [ 814.875 , 2.352330943977127e-46 ], [ 814.9375 , 2.3526235727854357e-46 ], 
                    [ 815.0 , 2.3529162379966406e-46 ], [ 815.0625 , 2.3532089396152366e-46 ], [ 815.125 , 2.353501677645753e-46 ], [ 815.1875 , 2.353794452092769e-46 ], 
                    [ 815.25 , 2.3540872629607986e-46 ], [ 815.3125 , 2.3543801102543054e-46 ], [ 815.375 , 2.3546729939779045e-46 ], [ 815.4375 , 2.354965914136061e-46 ], 
                    [ 815.5 , 2.35525887073339e-46 ], [ 815.5625 , 2.355551863774309e-46 ], [ 815.625 , 2.3558448932634843e-46 ], [ 815.6875 , 2.35613795920535e-46 ], 
                    [ 815.75 , 2.3564310616044742e-46 ], [ 815.8125 , 2.3567242004654086e-46 ], [ 815.875 , 2.3570173757926393e-46 ], [ 815.9375 , 2.3573105875907863e-46 ], 
                    [ 816.0 , 2.3576038358643028e-46 ], [ 816.0625 , 2.35789712061776e-46 ], [ 816.125 , 2.358190441855729e-46 ], [ 816.1875 , 2.358483799582665e-46 ], 
                    [ 816.25 , 2.358777193803174e-46 ], [ 816.3125 , 2.3590706245217795e-46 ], [ 816.375 , 2.3593640917430216e-46 ], [ 816.4375 , 2.359657595471458e-46 ], 
                    [ 816.5 , 2.3599510217029807e-46 ], [ 816.5625 , 2.3602441923812137e-46 ], [ 816.625 , 2.3605373994792558e-46 ], [ 816.6875 , 2.3608306430016992e-46 ], 
                    [ 816.75 , 2.361123922953002e-46 ], [ 816.8125 , 2.3614172393377563e-46 ], [ 816.875 , 2.361710592160421e-46 ], [ 816.9375 , 2.3620039814255734e-46 ], 
                    [ 817.0 , 2.3622974071377238e-46 ], [ 817.0625 , 2.3625908693013996e-46 ], [ 817.125 , 2.3628843679211457e-46 ], [ 817.1875 , 2.363177903001458e-46 ], 
                    [ 817.25 , 2.363471474546866e-46 ], [ 817.3125 , 2.3637650825619154e-46 ], [ 817.375 , 2.3640587270511547e-46 ], [ 817.4375 , 2.364352408019131e-46 ], 
                    [ 817.5 , 2.3646461254702924e-46 ], [ 817.5625 , 2.364939879409205e-46 ], [ 817.625 , 2.365233669840434e-46 ], [ 817.6875 , 2.365527496768514e-46 ], 
                    [ 817.75 , 2.365821360197978e-46 ], [ 817.8125 , 2.3661152601333275e-46 ], [ 817.875 , 2.366409196579131e-46 ], [ 817.9375 , 2.3667031695398566e-46 ], 
                    [ 818.0 , 2.3669971790201418e-46 ], [ 818.0625 , 2.367291225024422e-46 ], [ 818.125 , 2.367585307557369e-46 ], [ 818.1875 , 2.36787942662337e-46 ], 
                    [ 818.25 , 2.3681735822270296e-46 ], [ 818.3125 , 2.3684677743729213e-46 ], [ 818.375 , 2.368761726080242e-46 ], [ 818.4375 , 2.3690555778769267e-46 ], 
                    [ 818.5 , 2.3693494661267828e-46 ], [ 818.5625 , 2.3696433908343333e-46 ], [ 818.625 , 2.3699373520041e-46 ], [ 818.6875 , 2.370231349640624e-46 ], 
                    [ 818.75 , 2.370525383748395e-46 ], [ 818.8125 , 2.3708194543319537e-46 ], [ 818.875 , 2.3711135613958425e-46 ], [ 818.9375 , 2.371407704944536e-46 ], 
                    [ 819.0 , 2.371701884982628e-46 ], [ 819.0625 , 2.3719961015145774e-46 ], [ 819.125 , 2.372290354544979e-46 ], [ 819.1875 , 2.37258464407831e-46 ], 
                    [ 819.25 , 2.3728789701191487e-46 ], [ 819.3125 , 2.3731733326719234e-46 ], [ 819.375 , 2.3734677317412478e-46 ], [ 819.4375 , 2.3737621673316683e-46 ], 
                    [ 819.5 , 2.3740566394476823e-46 ], [ 819.5625 , 2.374351148093787e-46 ], [ 819.625 , 2.374645693274581e-46 ], [ 819.6875 , 2.374940274994564e-46 ], 
                    [ 819.75 , 2.3752348932582504e-46 ], [ 819.8125 , 2.3755295480702084e-46 ], [ 819.875 , 2.3758242394349206e-46 ], [ 819.9375 , 2.3761189673570064e-46 ], 
                    [ 820.0 , 2.3764137318409167e-46 ], [ 820.0625 , 2.37670853289122e-46 ], [ 820.125 , 2.377003370512436e-46 ], [ 820.1875 , 2.377298244709135e-46 ], 
                    [ 820.25 , 2.3775927495103234e-46 ], [ 820.3125 , 2.3778872875870003e-46 ], [ 820.375 , 2.3781818621512872e-46 ], [ 820.4375 , 2.3784764732076886e-46 ], 
                    [ 820.5 , 2.3787711207607583e-46 ], [ 820.5625 , 2.379065804815001e-46 ], [ 820.625 , 2.3793605253749207e-46 ], [ 820.6875 , 2.379655282445075e-46 ], 
                    [ 820.75 , 2.379950076029968e-46 ], [ 820.8125 , 2.380244906134091e-46 ], [ 820.875 , 2.380539772762086e-46 ], [ 820.9375 , 2.380834675918308e-46 ], 
                    [ 821.0 , 2.3811296156074005e-46 ], [ 821.0625 , 2.3814245918338563e-46 ], [ 821.125 , 2.3817196046021674e-46 ], [ 821.1875 , 2.3820146539169447e-46 ], 
                    [ 821.25 , 2.3823097397826317e-46 ], [ 821.3125 , 2.382604862203807e-46 ], [ 821.375 , 2.382900021184999e-46 ], [ 821.4375 , 2.3831952167306516e-46 ], 
                    [ 821.5 , 2.3834904488454475e-46 ], [ 821.5625 , 2.3837857175337644e-46 ], [ 821.625 , 2.3840810228002515e-46 ], [ 821.6875 , 2.3843763646493725e-46 ], 
                    [ 821.75 , 2.384671743085676e-46 ], [ 821.8125 , 2.384967158113695e-46 ], [ 821.875 , 2.385262609737928e-46 ], [ 821.9375 , 2.3855580979629767e-46 ], 
                    [ 822.0 , 2.385853622793341e-46 ], [ 822.0625 , 2.3861490796267528e-46 ], [ 822.125 , 2.38644425390075e-46 ], [ 822.1875 , 2.3867394646887827e-46 ], 
                    [ 822.25 , 2.3870347119953003e-46 ], [ 822.3125 , 2.387329995824871e-46 ], [ 822.375 , 2.387625316182013e-46 ], [ 822.4375 , 2.387920673071194e-46 ], 
                    [ 822.5 , 2.3882160664969842e-46 ], [ 822.5625 , 2.3885114964639028e-46 ], [ 822.625 , 2.3888069629764023e-46 ], [ 822.6875 , 2.3891024660390886e-46 ], 
                    [ 822.75 , 2.389398005656483e-46 ], [ 822.8125 , 2.3896935818330226e-46 ], [ 822.875 , 2.3899891945732806e-46 ], [ 822.9375 , 2.3902848438817973e-46 ], 
                    [ 823.0 , 2.390580529763062e-46 ], [ 823.0625 , 2.3908762522216503e-46 ], [ 823.125 , 2.3911720112620186e-46 ], [ 823.1875 , 2.39146780688876e-46 ], 
                    [ 823.25 , 2.391763639106333e-46 ], [ 823.3125 , 2.3920595079193486e-46 ], [ 823.375 , 2.3923554133321976e-46 ], [ 823.4375 , 2.392651355349578e-46 ], 
                    [ 823.5 , 2.3929473339758813e-46 ], [ 823.5625 , 2.3932433492157213e-46 ], [ 823.625 , 2.3935394010735766e-46 ], [ 823.6875 , 2.393835489553994e-46 ], 
                    [ 823.75 , 2.3941316146615035e-46 ], [ 823.8125 , 2.394427776400636e-46 ], [ 823.875 , 2.3947239747759228e-46 ], [ 823.9375 , 2.3950200312340375e-46 ], 
                    [ 824.0 , 2.3953158916262122e-46 ], [ 824.0625 , 2.3956117885664105e-46 ], [ 824.125 , 2.395907722059199e-46 ], [ 824.1875 , 2.3962036921091093e-46 ], 
                    [ 824.25 , 2.3964996987206243e-46 ], [ 824.3125 , 2.3967957418983106e-46 ], [ 824.375 , 2.3970918216466174e-46 ], [ 824.4375 , 2.397387937970097e-46 ], 
                    [ 824.5 , 2.397684090873284e-46 ], [ 824.5625 , 2.397980280360663e-46 ], [ 824.625 , 2.3982765064367883e-46 ], [ 824.6875 , 2.3985727691061614e-46 ], 
                    [ 824.75 , 2.3988690683733032e-46 ], [ 824.8125 , 2.3991654042427355e-46 ], [ 824.875 , 2.399461776718962e-46 ], [ 824.9375 , 2.3997581858064873e-46 ], 
                    [ 825.0 , 2.4000546315099718e-46 ], [ 825.0625 , 2.400351113833767e-46 ], [ 825.125 , 2.4006476327824824e-46 ], [ 825.1875 , 2.4009441883605914e-46 ], 
                    [ 825.25 , 2.4012407805727034e-46 ], [ 825.3125 , 2.4015374094232593e-46 ], [ 825.375 , 2.4018340749168533e-46 ], [ 825.4375 , 2.402130777057926e-46 ], 
                    [ 825.5 , 2.4024275158510737e-46 ], [ 825.5625 , 2.4027242913008235e-46 ], [ 825.625 , 2.4030211034116865e-46 ], [ 825.6875 , 2.403317952188209e-46 ], 
                    [ 825.75 , 2.4036148376348682e-46 ], [ 825.8125 , 2.403911530458713e-46 ], [ 825.875 , 2.4042080753452712e-46 ], [ 825.9375 , 2.404504656813422e-46 ], 
                    [ 826.0 , 2.4048012748676598e-46 ], [ 826.0625 , 2.4050979295124814e-46 ], [ 826.125 , 2.405394620752418e-46 ], [ 826.1875 , 2.4056913485919832e-46 ], 
                    [ 826.25 , 2.4059881130357098e-46 ], [ 826.3125 , 2.4062849140880953e-46 ], [ 826.375 , 2.406581751753639e-46 ], [ 826.4375 , 2.406878626036875e-46 ], 
                    [ 826.5 , 2.40717553694232e-46 ], [ 826.5625 , 2.4074724844744916e-46 ], [ 826.625 , 2.407769468637908e-46 ], [ 826.6875 , 2.4080664894370885e-46 ], 
                    [ 826.75 , 2.4083635468765693e-46 ], [ 826.8125 , 2.408660640960853e-46 ], [ 826.875 , 2.4089577716944265e-46 ], [ 826.9375 , 2.409254939081861e-46 ], 
                    [ 827.0 , 2.4095521431276966e-46 ], [ 827.0625 , 2.409849383836369e-46 ], [ 827.125 , 2.4101466612125035e-46 ], [ 827.1875 , 2.4104439752605042e-46 ], 
                    [ 827.25 , 2.4107413259849975e-46 ], [ 827.3125 , 2.411038713390473e-46 ], [ 827.375 , 2.411336137481473e-46 ], [ 827.4375 , 2.4116335982625233e-46 ], 
                    [ 827.5 , 2.4119310957380983e-46 ], [ 827.5625 , 2.412228629912742e-46 ], [ 827.625 , 2.4125262007910153e-46 ], [ 827.6875 , 2.412823550205677e-46 ], 
                    [ 827.75 , 2.413120768563072e-46 ], [ 827.8125 , 2.4134180235326905e-46 ], [ 827.875 , 2.4137153151189906e-46 ], [ 827.9375 , 2.4140126433264832e-46 ], 
                    [ 828.0 , 2.4143100081597135e-46 ], [ 828.0625 , 2.414607409623176e-46 ], [ 828.125 , 2.414904847721366e-46 ], [ 828.1875 , 2.415202322458831e-46 ], 
                    [ 828.25 , 2.4154998338400837e-46 ], [ 828.3125 , 2.4157973818696037e-46 ], [ 828.375 , 2.416094966551888e-46 ], [ 828.4375 , 2.416392587891538e-46 ], 
                    [ 828.5 , 2.4166902458930003e-46 ], [ 828.5625 , 2.4169879405607912e-46 ], [ 828.625 , 2.417285671899479e-46 ], [ 828.6875 , 2.4175834399135117e-46 ], 
                    [ 828.75 , 2.4178812446074588e-46 ], [ 828.8125 , 2.4181790859858215e-46 ], [ 828.875 , 2.4184769640530844e-46 ], [ 828.9375 , 2.4187748788138703e-46 ], 
                    [ 829.0 , 2.419072830272561e-46 ], [ 829.0625 , 2.419370818433798e-46 ], [ 829.125 , 2.4196688433020336e-46 ], [ 829.1875 , 2.4199669048817893e-46 ], 
                    [ 829.25 , 2.4202650031776213e-46 ], [ 829.3125 , 2.420563138194019e-46 ], [ 829.375 , 2.4208613099355566e-46 ], [ 829.4375 , 2.4211595184066716e-46 ], 
                    [ 829.5 , 2.4214577636119755e-46 ], [ 829.5625 , 2.4217557963011703e-46 ], [ 829.625 , 2.4220536927261485e-46 ], [ 829.6875 , 2.4223516257949038e-46 ], 
                    [ 829.75 , 2.4226495955119442e-46 ], [ 829.8125 , 2.42294760188176e-46 ], [ 829.875 , 2.4232456449088948e-46 ], [ 829.9375 , 2.42354372459784e-46 ], 
                    [ 830.0 , 2.42384184095314e-46 ], [ 830.0625 , 2.424139993979236e-46 ], [ 830.125 , 2.4244381836806563e-46 ], [ 830.1875 , 2.4247364100619465e-46 ], 
                    [ 830.25 , 2.4250346731276014e-46 ], [ 830.3125 , 2.425332972882133e-46 ], [ 830.375 , 2.4256313093300554e-46 ], [ 830.4375 , 2.4259296824758986e-46 ], 
                    [ 830.5 , 2.4262280923241598e-46 ], [ 830.5625 , 2.426526538879302e-46 ], [ 830.625 , 2.4268250221459613e-46 ], [ 830.6875 , 2.427123542128567e-46 ], 
                    [ 830.75 , 2.427422098831601e-46 ], [ 830.8125 , 2.4277206922596837e-46 ], [ 830.875 , 2.428019322417281e-46 ], [ 830.9375 , 2.4283179893088767e-46 ], 
                    [ 831.0 , 2.428616692939058e-46 ], [ 831.0625 , 2.428915433312292e-46 ], [ 831.125 , 2.4292142104330993e-46 ], [ 831.1875 , 2.4295130243060512e-46 ], 
                    [ 831.25 , 2.429811874935634e-46 ], [ 831.3125 , 2.4301107623263694e-46 ], [ 831.375 , 2.430409686482745e-46 ], [ 831.4375 , 2.430708431701466e-46 ], 
                    [ 831.5 , 2.4310070034950763e-46 ], [ 831.5625 , 2.43130561196321e-46 ], [ 831.625 , 2.4316042571104238e-46 ], [ 831.6875 , 2.4319029389411544e-46 ], 
                    [ 831.75 , 2.4322016574599595e-46 ], [ 831.8125 , 2.432500412671311e-46 ], [ 831.875 , 2.4327992045797333e-46 ], [ 831.9375 , 2.4330980331897514e-46 ], 
                    [ 832.0 , 2.4333968985058554e-46 ], [ 832.0625 , 2.4336958005325555e-46 ], [ 832.125 , 2.4339947392743423e-46 ], [ 832.1875 , 2.434293714735761e-46 ], 
                    [ 832.25 , 2.4345927269212874e-46 ], [ 832.3125 , 2.4348917758354665e-46 ], [ 832.375 , 2.435190861482793e-46 ], [ 832.4375 , 2.4354899838677616e-46 ], 
                    [ 832.5 , 2.4357891429948846e-46 ], [ 832.5625 , 2.4360883388687273e-46 ], [ 832.625 , 2.4363875714937683e-46 ], [ 832.6875 , 2.436686840874523e-46 ], 
                    [ 832.75 , 2.4369861470155223e-46 ], [ 832.8125 , 2.4372854899212133e-46 ], [ 832.875 , 2.4375848695961807e-46 ], [ 832.9375 , 2.4378842860449582e-46 ], 
                    [ 833.0 , 2.4381837392720116e-46 ], [ 833.0625 , 2.438483229281857e-46 ], [ 833.125 , 2.4387827560790317e-46 ], [ 833.1875 , 2.4390823196680706e-46 ], 
                    [ 833.25 , 2.439381920053441e-46 ], [ 833.3125 , 2.439681406991675e-46 ], [ 833.375 , 2.439980657153655e-46 ], [ 833.4375 , 2.4402799440215175e-46 ], 
                    [ 833.5 , 2.4405792675997824e-46 ], [ 833.5625 , 2.440878627892918e-46 ], [ 833.625 , 2.4411780249054614e-46 ], [ 833.6875 , 2.4414774586418835e-46 ], 
                    [ 833.75 , 2.441776929106705e-46 ], [ 833.8125 , 2.4420764363044492e-46 ], [ 833.875 , 2.4423759802396037e-46 ], [ 833.9375 , 2.4426755609166404e-46 ], 
                    [ 834.0 , 2.442975178340135e-46 ], [ 834.0625 , 2.4432748325145435e-46 ], [ 834.125 , 2.44357452344439e-46 ], [ 834.1875 , 2.4438742511342014e-46 ], 
                    [ 834.25 , 2.4441740155884513e-46 ], [ 834.3125 , 2.4444738168116666e-46 ], [ 834.375 , 2.4447736548083574e-46 ], [ 834.4375 , 2.4450735295830693e-46 ], 
                    [ 834.5 , 2.4453734411402435e-46 ], [ 834.5625 , 2.4456733894844272e-46 ], [ 834.625 , 2.445973374620132e-46 ], [ 834.6875 , 2.446273396551871e-46 ], 
                    [ 834.75 , 2.446573455284175e-46 ], [ 834.8125 , 2.446873550821506e-46 ], [ 834.875 , 2.4471736831684475e-46 ], [ 834.9375 , 2.4474738523294624e-46 ], 
                    [ 835.0 , 2.447774058309102e-46 ], [ 835.0625 , 2.448074301111864e-46 ], [ 835.125 , 2.448374580742248e-46 ], [ 835.1875 , 2.4486748398276974e-46 ], 
                    [ 835.25 , 2.448974765645769e-46 ], [ 835.3125 , 2.4492747282001876e-46 ], [ 835.375 , 2.44957472749554e-46 ], [ 835.4375 , 2.449874763536326e-46 ], 
                    [ 835.5 , 2.4501748363269772e-46 ], [ 835.5625 , 2.450474945872047e-46 ], [ 835.625 , 2.4507750921759845e-46 ], [ 835.6875 , 2.4510752752433277e-46 ], 
                    [ 835.75 , 2.4513754950786143e-46 ], [ 835.8125 , 2.4516757516863123e-46 ], [ 835.875 , 2.451976045070909e-46 ], [ 835.9375 , 2.452276375236926e-46 ], 
                    [ 836.0 , 2.452576742188869e-46 ], [ 836.0625 , 2.4528771459312435e-46 ], [ 836.125 , 2.4531775864685726e-46 ], [ 836.1875 , 2.4534780638053118e-46 ], 
                    [ 836.25 , 2.453778577946055e-46 ], [ 836.3125 , 2.4540791288951884e-46 ], [ 836.375 , 2.4543797166573423e-46 ], [ 836.4375 , 2.454680341236974e-46 ], 
                    [ 836.5 , 2.454981002638575e-46 ], [ 836.5625 , 2.455281700866638e-46 ], [ 836.625 , 2.4555824359257787e-46 ], [ 836.6875 , 2.455883207820386e-46 ], 
                    [ 836.75 , 2.4561840165550763e-46 ], [ 836.8125 , 2.4564848621342748e-46 ], [ 836.875 , 2.4567857445625297e-46 ], [ 836.9375 , 2.4570866638443364e-46 ], 
                    [ 837.0 , 2.4573876199842446e-46 ], [ 837.0625 , 2.4576886129867505e-46 ], [ 837.125 , 2.457989276144711e-46 ], [ 837.1875 , 2.458289909035467e-46 ], 
                    [ 837.25 , 2.4585905786962046e-46 ], [ 837.3125 , 2.4588912851313866e-46 ], [ 837.375 , 2.4591920283454757e-46 ], [ 837.4375 , 2.45949280834304e-46 ], 
                    [ 837.5 , 2.4597936251285607e-46 ], [ 837.5625 , 2.460094478706468e-46 ], [ 837.625 , 2.4603953690813314e-46 ], [ 837.6875 , 2.4606962962576697e-46 ], 
                    [ 837.75 , 2.4609972602399128e-46 ], [ 837.8125 , 2.4612982610326156e-46 ], [ 837.875 , 2.4615992986402625e-46 ], [ 837.9375 , 2.461900373067322e-46 ], 
                    [ 838.0 , 2.462201484318367e-46 ], [ 838.0625 , 2.4625026323978663e-46 ], [ 838.125 , 2.4628038173103064e-46 ], [ 838.1875 , 2.4631050390602457e-46 ], 
                    [ 838.25 , 2.4634062976521015e-46 ], [ 838.3125 , 2.4637075930904682e-46 ], [ 838.375 , 2.464008925379799e-46 ], [ 838.4375 , 2.464310294524637e-46 ], 
                    [ 838.5 , 2.4646117005294886e-46 ], [ 838.5625 , 2.4649131433987936e-46 ], [ 838.625 , 2.4652146231371298e-46 ], [ 838.6875 , 2.4655161397489897e-46 ], 
                    [ 838.75 , 2.4658176932389002e-46 ], [ 838.8125 , 2.466119283611355e-46 ], [ 838.875 , 2.4664209108708645e-46 ], [ 838.9375 , 2.466722575021941e-46 ], 
                    [ 839.0 , 2.46702406702477e-46 ], [ 839.0625 , 2.4673253763845324e-46 ], [ 839.125 , 2.4676267225446547e-46 ], [ 839.1875 , 2.4679281055095967e-46 ], 
                    [ 839.25 , 2.4682295252838704e-46 ], [ 839.3125 , 2.468530981872007e-46 ], [ 839.375 , 2.4688324752784508e-46 ], [ 839.4375 , 2.469134005507698e-46 ], 
                    [ 839.5 , 2.4694355725642803e-46 ], [ 839.5625 , 2.4697371764526786e-46 ], [ 839.625 , 2.4700388171773742e-46 ], [ 839.6875 , 2.4703404947429185e-46 ], 
                    [ 839.75 , 2.470642209153741e-46 ], [ 839.8125 , 2.4709439604143935e-46 ], [ 839.875 , 2.4712457485293605e-46 ], [ 839.9375 , 2.471547573503107e-46 ], 
                    [ 840.0 , 2.4718494353402056e-46 ], [ 840.0625 , 2.472151334045123e-46 ], [ 840.125 , 2.47245326962238e-46 ], [ 840.1875 , 2.4727552420764447e-46 ], 
                    [ 840.25 , 2.473057251411821e-46 ], [ 840.3125 , 2.4733592976330488e-46 ], [ 840.375 , 2.473661380744615e-46 ], [ 840.4375 , 2.473963500751026e-46 ], 
                    [ 840.5 , 2.4742656576567692e-46 ], [ 840.5625 , 2.4745678514664053e-46 ], [ 840.625 , 2.474870082184371e-46 ], [ 840.6875 , 2.4751723498152257e-46 ], 
                    [ 840.75 , 2.4754746543634435e-46 ], [ 840.8125 , 2.4757769958335865e-46 ], [ 840.875 , 2.4760793474130272e-46 ], [ 840.9375 , 2.4763813273421955e-46 ], 
                    [ 841.0 , 2.476683344100506e-46 ], [ 841.0625 , 2.4769853976924323e-46 ], [ 841.125 , 2.477287488122502e-46 ], [ 841.1875 , 2.4775896153951726e-46 ], 
                    [ 841.25 , 2.477891779514955e-46 ], [ 841.3125 , 2.4781939804863426e-46 ], [ 841.375 , 2.4784962183138654e-46 ], [ 841.4375 , 2.478798493001948e-46 ], 
                    [ 841.5 , 2.479100804555121e-46 ], [ 841.5625 , 2.479403152977898e-46 ], [ 841.625 , 2.479705538274758e-46 ], [ 841.6875 , 2.4800079604501807e-46 ], 
                    [ 841.75 , 2.4803104195086985e-46 ], [ 841.8125 , 2.4806129154547746e-46 ], [ 841.875 , 2.4809154482929435e-46 ], [ 841.9375 , 2.4812180180276858e-46 ], 
                    [ 842.0 , 2.4815206246634848e-46 ], [ 842.0625 , 2.4818232682048405e-46 ], [ 842.125 , 2.4821259486562887e-46 ], [ 842.1875 , 2.4824286660223493e-46 ], 
                    [ 842.25 , 2.4827314203074892e-46 ], [ 842.3125 , 2.4830342115161575e-46 ], [ 842.375 , 2.4833370396529456e-46 ], [ 842.4375 , 2.483639904722304e-46 ], 
                    [ 842.5 , 2.4839428067287736e-46 ], [ 842.5625 , 2.48424574567684e-46 ], [ 842.625 , 2.4845487215709916e-46 ], [ 842.6875 , 2.4848517344157343e-46 ], 
                    [ 842.75 , 2.4851547842156273e-46 ], [ 842.8125 , 2.4854576215351566e-46 ], [ 842.875 , 2.485760304654771e-46 ], [ 842.9375 , 2.486063024635634e-46 ], 
                    [ 843.0 , 2.4863657814822343e-46 ], [ 843.0625 , 2.486668575199062e-46 ], [ 843.125 , 2.4869714057906068e-46 ], [ 843.1875 , 2.4872742732613593e-46 ], 
                    [ 843.25 , 2.487577177615793e-46 ], [ 843.3125 , 2.487880118858453e-46 ], [ 843.375 , 2.4881830969937607e-46 ], [ 843.4375 , 2.4884861120262624e-46 ], 
                    [ 843.5 , 2.488789163960433e-46 ], [ 843.5625 , 2.4890922528007666e-46 ], [ 843.625 , 2.4893953785517766e-46 ], [ 843.6875 , 2.4896985412179034e-46 ], 
                    [ 843.75 , 2.490001740803661e-46 ], [ 843.8125 , 2.490304977313634e-46 ], [ 843.875 , 2.4906082507522128e-46 ], [ 843.9375 , 2.490911561123912e-46 ], 
                    [ 844.0 , 2.491214908433284e-46 ], [ 844.0625 , 2.4915182926847892e-46 ], [ 844.125 , 2.4918217138828755e-46 ], [ 844.1875 , 2.492125172032148e-46 ], 
                    [ 844.25 , 2.492428667137035e-46 ], [ 844.3125 , 2.4927321992020553e-46 ], [ 844.375 , 2.4930357682316927e-46 ], [ 844.4375 , 2.4933393742304837e-46 ], 
                    [ 844.5 , 2.4936430172029485e-46 ], [ 844.5625 , 2.4939466971534834e-46 ], [ 844.625 , 2.4942504140866975e-46 ], [ 844.6875 , 2.4945541671859906e-46 ], 
                    [ 844.75 , 2.4948575303967062e-46 ], [ 844.8125 , 2.495160930499463e-46 ], [ 844.875 , 2.495464367498764e-46 ], [ 844.9375 , 2.4957678413991325e-46 ], 
                    [ 845.0 , 2.496071352205002e-46 ], [ 845.0625 , 2.4963748999208616e-46 ], [ 845.125 , 2.4966784845512167e-46 ], [ 845.1875 , 2.496982106100575e-46 ], 
                    [ 845.25 , 2.497285764573408e-46 ], [ 845.3125 , 2.4975894599741884e-46 ], [ 845.375 , 2.4978931923074425e-46 ], [ 845.4375 , 2.498196961577644e-46 ], 
                    [ 845.5 , 2.4985007677892847e-46 ], [ 845.5625 , 2.498804610946839e-46 ], [ 845.625 , 2.4991084910548184e-46 ], [ 845.6875 , 2.499412408117716e-46 ], 
                    [ 845.75 , 2.499716362140043e-46 ], [ 845.8125 , 2.5000203531262775e-46 ], [ 845.875 , 2.5003243810809137e-46 ], [ 845.9375 , 2.5006284460084307e-46 ], 
                    [ 846.0 , 2.5009325479133417e-46 ], [ 846.0625 , 2.5012366868001607e-46 ], [ 846.125 , 2.5015408626733688e-46 ], [ 846.1875 , 2.501845075537463e-46 ], 
                    [ 846.25 , 2.5021493253969244e-46 ], [ 846.3125 , 2.502453612256287e-46 ], [ 846.375 , 2.5027579361200153e-46 ], [ 846.4375 , 2.5030622969926446e-46 ], 
                    [ 846.5 , 2.5033666948786584e-46 ], [ 846.5625 , 2.5036711297825574e-46 ], [ 846.625 , 2.503975436814736e-46 ], [ 846.6875 , 2.5042794978063932e-46 ], 
                    [ 846.75 , 2.5045835957206074e-46 ], [ 846.8125 , 2.5048877305618268e-46 ], [ 846.875 , 2.5051919023345353e-46 ], [ 846.9375 , 2.5054961110432e-46 ], 
                    [ 847.0 , 2.5058003566923048e-46 ], [ 847.0625 , 2.5061046392863903e-46 ], [ 847.125 , 2.506408958829889e-46 ], [ 847.1875 , 2.506713315327323e-46 ], 
                    [ 847.25 , 2.5070177087831267e-46 ], [ 847.3125 , 2.507322139201859e-46 ], [ 847.375 , 2.5076266065879546e-46 ], [ 847.4375 , 2.507931110945939e-46 ], 
                    [ 847.5 , 2.5082356522802646e-46 ], [ 847.5625 , 2.508540230595441e-46 ], [ 847.625 , 2.5088448458959577e-46 ], [ 847.6875 , 2.509149498186324e-46 ], 
                    [ 847.75 , 2.509454187471014e-46 ], [ 847.8125 , 2.5097589137544663e-46 ], [ 847.875 , 2.5100636770412987e-46 ], [ 847.9375 , 2.5103684773359152e-46 ], 
                    [ 848.0 , 2.510673314642792e-46 ], [ 848.0625 , 2.5109781889664595e-46 ], [ 848.125 , 2.5112831003114297e-46 ], [ 848.1875 , 2.511588048682199e-46 ], 
                    [ 848.25 , 2.5118930340832627e-46 ], [ 848.3125 , 2.512198056519047e-46 ], [ 848.375 , 2.5125031159941016e-46 ], [ 848.4375 , 2.512808212512979e-46 ], 
                    [ 848.5 , 2.513113346080069e-46 ], [ 848.5625 , 2.5134182242400304e-46 ], [ 848.625 , 2.5137229921182925e-46 ], [ 848.6875 , 2.51402779695159e-46 ], 
                    [ 848.75 , 2.5143326387443864e-46 ], [ 848.8125 , 2.5146375175011987e-46 ], [ 848.875 , 2.5149424332264914e-46 ], [ 848.9375 , 2.5152473859247287e-46 ], 
                    [ 849.0 , 2.515552375600448e-46 ], [ 849.0625 , 2.515857402258079e-46 ], [ 849.125 , 2.5161624659021244e-46 ], [ 849.1875 , 2.5164675665370686e-46 ], 
                    [ 849.25 , 2.516772704167397e-46 ], [ 849.3125 , 2.5170778787975957e-46 ], [ 849.375 , 2.5173830904321515e-46 ], [ 849.4375 , 2.5176883390755326e-46 ], 
                    [ 849.5 , 2.5179936247322456e-46 ], [ 849.5625 , 2.518298947406849e-46 ], [ 849.625 , 2.518604307103689e-46 ], [ 849.6875 , 2.5189097038273264e-46 ], 
                    [ 849.75 , 2.5192151375822865e-46 ], [ 849.8125 , 2.519520608373006e-46 ], [ 849.875 , 2.519826116203993e-46 ], [ 849.9375 , 2.5201316610797222e-46 ], 
                    [ 850.0 , 2.5204372430047562e-46 ], [ 850.0625 , 2.52074286198348e-46 ], [ 850.125 , 2.5210485180204767e-46 ], [ 850.1875 , 2.52135421112015e-46 ], 
                    [ 850.25 , 2.5216599412871014e-46 ], [ 850.3125 , 2.521965708525736e-46 ], [ 850.375 , 2.522271512840585e-46 ], [ 850.4375 , 2.52257735423618e-46 ], 
                    [ 850.5 , 2.5228828347957085e-46 ], [ 850.5625 , 2.523188304323845e-46 ], [ 850.625 , 2.5234938108380963e-46 ], [ 850.6875 , 2.5237993543429395e-46 ], 
                    [ 850.75 , 2.524104934842872e-46 ], [ 850.8125 , 2.524410552342337e-46 ], [ 850.875 , 2.524716206845851e-46 ], [ 850.9375 , 2.5250218983578394e-46 ], 
                    [ 851.0 , 2.525327626882856e-46 ], [ 851.0625 , 2.52563339242531e-46 ], [ 851.125 , 2.525939194989773e-46 ], [ 851.1875 , 2.5262450345806387e-46 ], 
                    [ 851.25 , 2.5265509112024252e-46 ], [ 851.3125 , 2.526856824859617e-46 ], [ 851.375 , 2.527162775556716e-46 ], [ 851.4375 , 2.527468763298152e-46 ], 
                    [ 851.5 , 2.527774788088484e-46 ], [ 851.5625 , 2.528080849932161e-46 ], [ 851.625 , 2.5283869488336695e-46 ], [ 851.6875 , 2.5286930847974423e-46 ], 
                    [ 851.75 , 2.528999257828075e-46 ], [ 851.8125 , 2.529305467930002e-46 ], [ 851.875 , 2.5296117151076934e-46 ], [ 851.9375 , 2.529917999365674e-46 ], 
                    [ 852.0 , 2.5302243207083986e-46 ], [ 852.0625 , 2.5305306791403564e-46 ], [ 852.125 , 2.5308370746660922e-46 ], [ 852.1875 , 2.5311435072900442e-46 ], 
                    [ 852.25 , 2.531449977016739e-46 ], [ 852.3125 , 2.5317564838506336e-46 ], [ 852.375 , 2.5320630054796653e-46 ], [ 852.4375 , 2.532369139438068e-46 ], 
                    [ 852.5 , 2.5326753104089785e-46 ], [ 852.5625 , 2.5329815183968886e-46 ], [ 852.625 , 2.5332877634062027e-46 ], [ 852.6875 , 2.5335940454414682e-46 ], 
                    [ 852.75 , 2.5339003645071797e-46 ], [ 852.8125 , 2.5342067206077436e-46 ], [ 852.875 , 2.5345131137477076e-46 ], [ 852.9375 , 2.534819543931479e-46 ], 
                    [ 853.0 , 2.53512601116359e-46 ], [ 853.0625 , 2.5354325154484846e-46 ], [ 853.125 , 2.5357390567906593e-46 ], [ 853.1875 , 2.5360456351946316e-46 ], 
                    [ 853.25 , 2.536352250664828e-46 ], [ 853.3125 , 2.53665890320573e-46 ], [ 853.375 , 2.5369655928218552e-46 ], [ 853.4375 , 2.5372723195176687e-46 ], 
                    [ 853.5 , 2.537579083297635e-46 ], [ 853.5625 , 2.537885884166256e-46 ], [ 853.625 , 2.5381927221280522e-46 ], [ 853.6875 , 2.538499597187453e-46 ], 
                    [ 853.75 , 2.538806509348945e-46 ], [ 853.8125 , 2.5391134586170672e-46 ], [ 853.875 , 2.539420444996234e-46 ], [ 853.9375 , 2.5397274684910046e-46 ], 
                    [ 854.0 , 2.540034529105758e-46 ], [ 854.0625 , 2.540341626845144e-46 ], [ 854.125 , 2.5406487617135072e-46 ], [ 854.1875 , 2.5409559337154267e-46 ], 
                    [ 854.25 , 2.54126314285532e-46 ], [ 854.3125 , 2.5415703276242206e-46 ], [ 854.375 , 2.5418771634258164e-46 ], [ 854.4375 , 2.542184036270753e-46 ], 
                    [ 854.5 , 2.542490946163466e-46 ], [ 854.5625 , 2.5427978931084823e-46 ], [ 854.625 , 2.543104877110203e-46 ], [ 854.6875 , 2.5434118981731557e-46 ], 
                    [ 854.75 , 2.543718956301761e-46 ], [ 854.8125 , 2.5440260515005838e-46 ], [ 854.875 , 2.544333183774009e-46 ], [ 854.9375 , 2.544640353126567e-46 ], 
                    [ 855.0 , 2.5449475595627154e-46 ], [ 855.0625 , 2.5452548030869322e-46 ], [ 855.125 , 2.5455620837036945e-46 ], [ 855.1875 , 2.5458694014174623e-46 ], 
                    [ 855.25 , 2.5461767562327503e-46 ], [ 855.3125 , 2.5464841481540197e-46 ], [ 855.375 , 2.5467915771857323e-46 ], [ 855.4375 , 2.5470990433323855e-46 ], 
                    [ 855.5 , 2.547406546598461e-46 ], [ 855.5625 , 2.547714086988458e-46 ], [ 855.625 , 2.5480216645068584e-46 ], [ 855.6875 , 2.548329279158108e-46 ], 
                    [ 855.75 , 2.5486369309467083e-46 ], [ 855.8125 , 2.5489446198771245e-46 ], [ 855.875 , 2.549252345953859e-46 ], [ 855.9375 , 2.5495601091814138e-46 ], 
                    [ 856.0 , 2.5498679095642744e-46 ], [ 856.0625 , 2.5501757471068716e-46 ], [ 856.125 , 2.550483621813729e-46 ], [ 856.1875 , 2.5507915336893687e-46 ], 
                    [ 856.25 , 2.5510994141300607e-46 ], [ 856.3125 , 2.551406957776075e-46 ], [ 856.375 , 2.5517145384974964e-46 ], [ 856.4375 , 2.5520221562988117e-46 ], 
                    [ 856.5 , 2.5523298111844913e-46 ], [ 856.5625 , 2.5526375031590055e-46 ], [ 856.625 , 2.5529452322268263e-46 ], [ 856.6875 , 2.5532529983924246e-46 ], 
                    [ 856.75 , 2.5535608016602736e-46 ], [ 856.8125 , 2.553868642034864e-46 ], [ 856.875 , 2.5541765195206135e-46 ], [ 856.9375 , 2.5544844341220516e-46 ], 
                    [ 857.0 , 2.5547923858436347e-46 ], [ 857.0625 , 2.5551003746898557e-46 ], [ 857.125 , 2.5554084006651346e-46 ], [ 857.1875 , 2.5557164637740215e-46 ], 
                    [ 857.25 , 2.556024564020938e-46 ], [ 857.3125 , 2.556332701410361e-46 ], [ 857.375 , 2.5566408759468046e-46 ], [ 857.4375 , 2.556949087634747e-46 ], 
                    [ 857.5 , 2.557257336478613e-46 ], [ 857.5625 , 2.5575656224829352e-46 ], [ 857.625 , 2.5578739456521942e-46 ], [ 857.6875 , 2.5581823059908156e-46 ], 
                    [ 857.75 , 2.5584907035033347e-46 ], [ 857.8125 , 2.5587991381942155e-46 ], [ 857.875 , 2.5591076100679576e-46 ], [ 857.9375 , 2.559416119128989e-46 ], 
                    [ 858.0 , 2.5597246653818833e-46 ], [ 858.0625 , 2.5600332488310156e-46 ], [ 858.125 , 2.5603418694809428e-46 ], [ 858.1875 , 2.5606504805831383e-46 ], 
                    [ 858.25 , 2.5609587190151614e-46 ], [ 858.3125 , 2.561266994551402e-46 ], [ 858.375 , 2.5615753071963256e-46 ], [ 858.4375 , 2.5618836569543815e-46 ], 
                    [ 858.5 , 2.562192043830055e-46 ], [ 858.5625 , 2.5625004678278147e-46 ], [ 858.625 , 2.5628089289521287e-46 ], [ 858.6875 , 2.5631174272074845e-46 ], 
                    [ 858.75 , 2.563425962598315e-46 ], [ 858.8125 , 2.563734535129109e-46 ], [ 858.875 , 2.5640431448043554e-46 ], [ 858.9375 , 2.5643517916285077e-46 ], 
                    [ 859.0 , 2.5646604756060183e-46 ], [ 859.0625 , 2.5649691967413783e-46 ], [ 859.125 , 2.5652779550390793e-46 ], [ 859.1875 , 2.5655867505035942e-46 ], 
                    [ 859.25 , 2.5658955831393424e-46 ], [ 859.3125 , 2.5662044529508346e-46 ], [ 859.375 , 2.566513359942547e-46 ], [ 859.4375 , 2.5668223041189535e-46 ], 
                    [ 859.5 , 2.567131285484531e-46 ], [ 859.5625 , 2.5674403040437747e-46 ], [ 859.625 , 2.567749359801143e-46 ], [ 859.6875 , 2.568058452761114e-46 ], 
                    [ 859.75 , 2.568367582928147e-46 ], [ 859.8125 , 2.568676750306758e-46 ], [ 859.875 , 2.5689859549014084e-46 ], [ 859.9375 , 2.569295196716523e-46 ], 
                    [ 860.0 , 2.5696044757567093e-46 ], [ 860.0625 , 2.569913792026339e-46 ], [ 860.125 , 2.570223145529931e-46 ], [ 860.1875 , 2.5705320900277213e-46 ], 
                    [ 860.25 , 2.570841062255351e-46 ], [ 860.3125 , 2.571150071620736e-46 ], [ 860.375 , 2.571459118128394e-46 ], [ 860.4375 , 2.571768201782735e-46 ], 
                    [ 860.5 , 2.5720773225882426e-46 ], [ 860.5625 , 2.572386480549364e-46 ], [ 860.625 , 2.5726956756706195e-46 ], [ 860.6875 , 2.5730049079564027e-46 ], 
                    [ 860.75 , 2.5733141774112355e-46 ], [ 860.8125 , 2.5736234840395675e-46 ], [ 860.875 , 2.573932827845867e-46 ], [ 860.9375 , 2.5742422088345842e-46 ], 
                    [ 861.0 , 2.574551627010225e-46 ], [ 861.0625 , 2.5748610823772224e-46 ], [ 861.125 , 2.5751705749400836e-46 ], [ 861.1875 , 2.575480104703261e-46 ], 
                    [ 861.25 , 2.575789671671244e-46 ], [ 861.3125 , 2.57609927584845e-46 ], [ 861.375 , 2.5764089172394067e-46 ], [ 861.4375 , 2.5767185958485683e-46 ], 
                    [ 861.5 , 2.5770283116804084e-46 ], [ 861.5625 , 2.5773380647393826e-46 ], [ 861.625 , 2.577647855030021e-46 ], [ 861.6875 , 2.5779576825567436e-46 ], 
                    [ 861.75 , 2.578267547324007e-46 ], [ 861.8125 , 2.578577449336362e-46 ], [ 861.875 , 2.5788873885982666e-46 ], [ 861.9375 , 2.579197365114161e-46 ], 
                    [ 862.0 , 2.579507378888505e-46 ], [ 862.0625 , 2.5798174299258686e-46 ], [ 862.125 , 2.5801271617686682e-46 ], [ 862.1875 , 2.580436833595609e-46 ], 
                    [ 862.25 , 2.5807465425899223e-46 ], [ 862.3125 , 2.5810562887561047e-46 ], [ 862.375 , 2.5813660720986363e-46 ], [ 862.4375 , 2.5816758926219796e-46 ], 
                    [ 862.5 , 2.581985750330541e-46 ], [ 862.5625 , 2.5822956452288212e-46 ], [ 862.625 , 2.5826055773212835e-46 ], [ 862.6875 , 2.5829155466124286e-46 ], 
                    [ 862.75 , 2.5832255531066657e-46 ], [ 862.8125 , 2.583535596808461e-46 ], [ 862.875 , 2.5838456777223164e-46 ], [ 862.9375 , 2.584155795852679e-46 ], 
                    [ 863.0 , 2.584465951204017e-46 ], [ 863.0625 , 2.584776143780779e-46 ], [ 863.125 , 2.5850863735874687e-46 ], [ 863.1875 , 2.585396640628537e-46 ], 
                    [ 863.25 , 2.5857069449084347e-46 ], [ 863.3125 , 2.5860172864316485e-46 ], [ 863.375 , 2.5863276652027048e-46 ], [ 863.4375 , 2.5866380812259447e-46 ], 
                    [ 863.5 , 2.5869485345059497e-46 ], [ 863.5625 , 2.5872590250471548e-46 ], [ 863.625 , 2.5875695528540324e-46 ], [ 863.6875 , 2.587880117931036e-46 ], 
                    [ 863.75 , 2.5881907202826584e-46 ], [ 863.8125 , 2.5885013599133906e-46 ], [ 863.875 , 2.588812036827708e-46 ], [ 863.9375 , 2.5891227510300293e-46 ], 
                    [ 864.0 , 2.5894335025248852e-46 ], [ 864.0625 , 2.589744050417851e-46 ], [ 864.125 , 2.59005441443916e-46 ], [ 864.1875 , 2.5903648156556013e-46 ], 
                    [ 864.25 , 2.5906752540716146e-46 ], [ 864.3125 , 2.5909857296916393e-46 ], [ 864.375 , 2.591296242520171e-46 ], [ 864.4375 , 2.5916067925616507e-46 ], 
                    [ 864.5 , 2.5919173798205377e-46 ], [ 864.5625 , 2.5922280043012742e-46 ], [ 864.625 , 2.5925386660083394e-46 ], [ 864.6875 , 2.592849364946231e-46 ], 
                    [ 864.75 , 2.5931601011193376e-46 ], [ 864.8125 , 2.593470874532159e-46 ], [ 864.875 , 2.5937816851891935e-46 ], [ 864.9375 , 2.5940925330948324e-46 ], 
                    [ 865.0 , 2.5944034182535753e-46 ], [ 865.0625 , 2.5947143406699065e-46 ], [ 865.125 , 2.5950253003482714e-46 ], [ 865.1875 , 2.5953362972931363e-46 ], 
                    [ 865.25 , 2.5956473315089486e-46 ], [ 865.3125 , 2.595958403000212e-46 ], [ 865.375 , 2.596269511771394e-46 ], [ 865.4375 , 2.5965806578269067e-46 ], 
                    [ 865.5 , 2.5968918411712924e-46 ], [ 865.5625 , 2.597203061808927e-46 ], [ 865.625 , 2.597514319744373e-46 ], [ 865.6875 , 2.597825614982063e-46 ], 
                    [ 865.75 , 2.5981369475264305e-46 ], [ 865.8125 , 2.598448317382002e-46 ], [ 865.875 , 2.5987597245532485e-46 ], [ 865.9375 , 2.5990711690446057e-46 ], 
                    [ 866.0 , 2.5993825659657142e-46 ], [ 866.0625 , 2.599693629210112e-46 ], [ 866.125 , 2.6000047296788467e-46 ], [ 866.1875 , 2.6003158673764298e-46 ], 
                    [ 866.25 , 2.600627042307259e-46 ], [ 866.3125 , 2.6009382544758107e-46 ], [ 866.375 , 2.6012495038865394e-46 ], [ 866.4375 , 2.6015607905438838e-46 ], 
                    [ 866.5 , 2.6018721144523376e-46 ], [ 866.5625 , 2.602183475616341e-46 ], [ 866.625 , 2.602494874040352e-46 ], [ 866.6875 , 2.602806309728829e-46 ], 
                    [ 866.75 , 2.6031177826862135e-46 ], [ 866.8125 , 2.6034292929170207e-46 ], [ 866.875 , 2.603740840425655e-46 ], [ 866.9375 , 2.6040524252165965e-46 ], 
                    [ 867.0 , 2.604364047294325e-46 ], [ 867.0625 , 2.6046757066632285e-46 ], [ 867.125 , 2.6049874033278437e-46 ], [ 867.1875 , 2.6052991372926702e-46 ], 
                    [ 867.25 , 2.605610908562043e-46 ], [ 867.3125 , 2.6059227171405738e-46 ], [ 867.375 , 2.6062345630325974e-46 ], [ 867.4375 , 2.6065464462426358e-46 ], 
                    [ 867.5 , 2.6068583667751717e-46 ], [ 867.5625 , 2.607170324634691e-46 ], [ 867.625 , 2.607482319825549e-46 ], [ 867.6875 , 2.607794352352306e-46 ], 
                    [ 867.75 , 2.6081064222194482e-46 ], [ 867.8125 , 2.608418529431407e-46 ], [ 867.875 , 2.608730673992615e-46 ], [ 867.9375 , 2.609042855907577e-46 ], 
                    [ 868.0 , 2.6093547121536733e-46 ], [ 868.0625 , 2.60966650336655e-46 ], [ 868.125 , 2.6099783318353263e-46 ], [ 868.1875 , 2.6102901975644166e-46 ], 
                    [ 868.25 , 2.6106021005582547e-46 ], [ 868.3125 , 2.6109140408213312e-46 ], [ 868.375 , 2.6112260183580795e-46 ], [ 868.4375 , 2.6115380331729545e-46 ], 
                    [ 868.5 , 2.61185008527041e-46 ], [ 868.5625 , 2.6121621746548827e-46 ], [ 868.625 , 2.6124743013308834e-46 ], [ 868.6875 , 2.612786465302794e-46 ], 
                    [ 868.75 , 2.6130986665751267e-46 ], [ 868.8125 , 2.6134109051523007e-46 ], [ 868.875 , 2.6137231810388118e-46 ], [ 868.9375 , 2.6140354942390802e-46 ], 
                    [ 869.0 , 2.6143478447576204e-46 ], [ 869.0625 , 2.6146602325988176e-46 ], [ 869.125 , 2.6149726577672054e-46 ], [ 869.1875 , 2.6152851202671506e-46 ], 
                    [ 869.25 , 2.615597620103208e-46 ], [ 869.3125 , 2.6159101572797824e-46 ], [ 869.375 , 2.616222731801373e-46 ], [ 869.4375 , 2.616535343672405e-46 ], 
                    [ 869.5 , 2.616847992897378e-46 ], [ 869.5625 , 2.617160679480719e-46 ], [ 869.625 , 2.6174734034269474e-46 ], [ 869.6875 , 2.6177861647404715e-46 ], 
                    [ 869.75 , 2.618098963425757e-46 ], [ 869.8125 , 2.6184117994872874e-46 ], [ 869.875 , 2.618724672929566e-46 ], [ 869.9375 , 2.6190374433515153e-46 ], 
                    [ 870.0 , 2.6193499276834488e-46 ], [ 870.0625 , 2.61966244929872e-46 ], [ 870.125 , 2.6199750082017774e-46 ], [ 870.1875 , 2.620287604397089e-46 ], 
                    [ 870.25 , 2.620600237889047e-46 ], [ 870.3125 , 2.6209129086821776e-46 ], [ 870.375 , 2.621225616780874e-46 ], [ 870.4375 , 2.621538362189588e-46 ], 
                    [ 870.5 , 2.621851144912827e-46 ], [ 870.5625 , 2.622163964954968e-46 ], [ 870.625 , 2.6224768223205388e-46 ], [ 870.6875 , 2.6227897170139178e-46 ], 
                    [ 870.75 , 2.623102649039633e-46 ], [ 870.8125 , 2.623415618402064e-46 ], [ 870.875 , 2.623728625105741e-46 ], [ 870.9375 , 2.6240416691550625e-46 ], 
                    [ 871.0 , 2.624354750554503e-46 ], [ 871.0625 , 2.624667869308463e-46 ], [ 871.125 , 2.6249810254215484e-46 ], [ 871.1875 , 2.6252942188981058e-46 ], 
                    [ 871.25 , 2.625607449742573e-46 ], [ 871.3125 , 2.6259207179594836e-46 ], [ 871.375 , 2.6262340235532225e-46 ], [ 871.4375 , 2.6265473665283045e-46 ], 
                    [ 871.5 , 2.6268607468891905e-46 ], [ 871.5625 , 2.627174164640284e-46 ], [ 871.625 , 2.627487619786084e-46 ], [ 871.6875 , 2.6278011123310898e-46 ], 
                    [ 871.75 , 2.62811464227967e-46 ], [ 871.8125 , 2.6284282096364173e-46 ], [ 871.875 , 2.6287418144056844e-46 ], [ 871.9375 , 2.629055105290248e-46 ], 
                    [ 872.0 , 2.6293683188338373e-46 ], [ 872.0625 , 2.629681569692211e-46 ], [ 872.125 , 2.629994857869834e-46 ], [ 872.1875 , 2.6303081833711888e-46 ], 
                    [ 872.25 , 2.630621546200686e-46 ], [ 872.3125 , 2.630934946362809e-46 ], [ 872.375 , 2.6312483838619685e-46 ], [ 872.4375 , 2.6315618587026685e-46 ], 
                    [ 872.5 , 2.6318753708892645e-46 ], [ 872.5625 , 2.632188920426299e-46 ], [ 872.625 , 2.6325025073181843e-46 ], [ 872.6875 , 2.6328161315693524e-46 ], 
                    [ 872.75 , 2.633129793184291e-46 ], [ 872.8125 , 2.6334434921674336e-46 ], [ 872.875 , 2.633757228523212e-46 ], [ 872.9375 , 2.6340710022561357e-46 ], 
                    [ 873.0 , 2.6343848133705828e-46 ], [ 873.0625 , 2.6346986618710437e-46 ], [ 873.125 , 2.6350125477619915e-46 ], [ 873.1875 , 2.6353264710478435e-46 ], 
                    [ 873.25 , 2.6356404317330725e-46 ], [ 873.3125 , 2.6359544298221542e-46 ], [ 873.375 , 2.6362684653195434e-46 ], [ 873.4375 , 2.63658253822966e-46 ], 
                    [ 873.5 , 2.6368966485569803e-46 ], [ 873.5625 , 2.6372107963059608e-46 ], [ 873.625 , 2.6375249814810423e-46 ], [ 873.6875 , 2.63783920408672e-46 ], 
                    [ 873.75 , 2.6381534641274167e-46 ], [ 873.8125 , 2.638467761607629e-46 ], [ 873.875 , 2.638782034663967e-46 ], [ 873.9375 , 2.6390959444455555e-46 ], 
                    [ 874.0 , 2.639409891569903e-46 ], [ 874.0625 , 2.6397238760413947e-46 ], [ 874.125 , 2.640037897864568e-46 ], [ 874.1875 , 2.640351957043772e-46 ], 
                    [ 874.25 , 2.6406660535835254e-46 ], [ 874.3125 , 2.640980187488217e-46 ], [ 874.375 , 2.6412943587622725e-46 ], [ 874.4375 , 2.6416085674102312e-46 ], 
                    [ 874.5 , 2.6419228134364266e-46 ], [ 874.5625 , 2.6422370968454366e-46 ], [ 874.625 , 2.6425514176415394e-46 ], [ 874.6875 , 2.642865775829351e-46 ], 
                    [ 874.75 , 2.643180171413208e-46 ], [ 874.8125 , 2.6434946043976148e-46 ], [ 874.875 , 2.643809074787001e-46 ], [ 874.9375 , 2.6441235825857796e-46 ], 
                    [ 875.0 , 2.644438127798476e-46 ], [ 875.0625 , 2.6447527104294844e-46 ], [ 875.125 , 2.6450673304833125e-46 ], [ 875.1875 , 2.6453819879643555e-46 ], 
                    [ 875.25 , 2.645696682877047e-46 ], [ 875.3125 , 2.6460114152259155e-46 ], [ 875.375 , 2.6463261850153764e-46 ], [ 875.4375 , 2.646640992249865e-46 ], 
                    [ 875.5 , 2.646955836933892e-46 ], [ 875.5625 , 2.6472707190718184e-46 ], [ 875.625 , 2.647585638668213e-46 ], [ 875.6875 , 2.647900595727456e-46 ], 
                    [ 875.75 , 2.6482155902540045e-46 ], [ 875.8125 , 2.648530622252334e-46 ], [ 875.875 , 2.648845483048614e-46 ], [ 875.9375 , 2.6491601175435086e-46 ], 
                    [ 876.0 , 2.649474789411276e-46 ], [ 876.0625 , 2.649789498656299e-46 ], [ 876.125 , 2.650104245283018e-46 ], [ 876.1875 , 2.650419029295928e-46 ], 
                    [ 876.25 , 2.650733850699415e-46 ], [ 876.3125 , 2.6510487094979565e-46 ], [ 876.375 , 2.6513636056959577e-46 ], [ 876.4375 , 2.6516785392978795e-46 ], 
                    [ 876.5 , 2.651993510308184e-46 ], [ 876.5625 , 2.6523085187312765e-46 ], [ 876.625 , 2.6526235645716007e-46 ], [ 876.6875 , 2.652938647833658e-46 ], 
                    [ 876.75 , 2.6532537685218373e-46 ], [ 876.8125 , 2.6535689266406027e-46 ], [ 876.875 , 2.6538841221943996e-46 ], [ 876.9375 , 2.654199355187694e-46 ], 
                    [ 877.0 , 2.654514625624858e-46 ], [ 877.0625 , 2.6548299335104895e-46 ], [ 877.125 , 2.6551452788488862e-46 ], [ 877.1875 , 2.655460661644534e-46 ], 
                    [ 877.25 , 2.6557760819019393e-46 ], [ 877.3125 , 2.656091539625457e-46 ], [ 877.375 , 2.656407034819651e-46 ], [ 877.4375 , 2.6567225674888973e-46 ], 
                    [ 877.5 , 2.6570381376376838e-46 ], [ 877.5625 , 2.657353745270444e-46 ], [ 877.625 , 2.6576693903915926e-46 ], [ 877.6875 , 2.6579850730056388e-46 ], 
                    [ 877.75 , 2.6583007931170177e-46 ], [ 877.8125 , 2.6586165507302022e-46 ], [ 877.875 , 2.658932028211932e-46 ], [ 877.9375 , 2.659247392311884e-46 ], 
                    [ 878.0 , 2.6595627938157306e-46 ], [ 878.0625 , 2.6598782327279472e-46 ], [ 878.125 , 2.6601937090529885e-46 ], [ 878.1875 , 2.660509222795292e-46 ], 
                    [ 878.25 , 2.6608247739592394e-46 ], [ 878.3125 , 2.6611403625493253e-46 ], [ 878.375 , 2.6614559885699512e-46 ], [ 878.4375 , 2.661771652025576e-46 ], 
                    [ 878.5 , 2.6620873529206575e-46 ], [ 878.5625 , 2.6624030912596367e-46 ], [ 878.625 , 2.662718867046879e-46 ], [ 878.6875 , 2.6630346802869586e-46 ], 
                    [ 878.75 , 2.663350530984166e-46 ], [ 878.8125 , 2.663666419143076e-46 ], [ 878.875 , 2.6639823447680566e-46 ], [ 878.9375 , 2.6642983078635514e-46 ], 
                    [ 879.0 , 2.6646143084340237e-46 ], [ 879.0625 , 2.6649303464839367e-46 ], [ 879.125 , 2.6652464220176973e-46 ], [ 879.1875 , 2.6655625350397907e-46 ], 
                    [ 879.25 , 2.6658786855546433e-46 ], [ 879.3125 , 2.6661948735667214e-46 ], [ 879.375 , 2.6665110990804145e-46 ], [ 879.4375 , 2.6668273621002284e-46 ], 
                    [ 879.5 , 2.6671436626305543e-46 ], [ 879.5625 , 2.667460000675898e-46 ], [ 879.625 , 2.667776376240747e-46 ], [ 879.6875 , 2.6680927893294563e-46 ], 
                    [ 879.75 , 2.668409239946477e-46 ], [ 879.8125 , 2.668725728096373e-46 ], [ 879.875 , 2.6690418460180004e-46 ], [ 879.9375 , 2.6693579250890384e-46 ], 
                    [ 880.0 , 2.669674041591516e-46 ], [ 880.0625 , 2.6699901955297913e-46 ], [ 880.125 , 2.6703063869083343e-46 ], [ 880.1875 , 2.6706226157315992e-46 ], 
                    [ 880.25 , 2.6709388820039818e-46 ], [ 880.3125 , 2.6712551857299735e-46 ], [ 880.375 , 2.6715715269139156e-46 ], [ 880.4375 , 2.6718879055603186e-46 ], 
                    [ 880.5 , 2.672204321673659e-46 ], [ 880.5625 , 2.672520775258239e-46 ], [ 880.625 , 2.6728372663186108e-46 ], [ 880.6875 , 2.6731537948591368e-46 ], 
                    [ 880.75 , 2.673470360884331e-46 ], [ 880.8125 , 2.6737869643985946e-46 ], [ 880.875 , 2.674103605406367e-46 ], [ 880.9375 , 2.674420283912089e-46 ], 
                    [ 881.0 , 2.674736999920162e-46 ], [ 881.0625 , 2.6750537534351053e-46 ], [ 881.125 , 2.67537054446134e-46 ], [ 881.1875 , 2.6756873730032326e-46 ], 
                    [ 881.25 , 2.676004239065302e-46 ], [ 881.3125 , 2.6763211426520293e-46 ], [ 881.375 , 2.6766380837677258e-46 ], [ 881.4375 , 2.67695506241693e-46 ], 
                    [ 881.5 , 2.677272078604107e-46 ], [ 881.5625 , 2.6775891323336062e-46 ], [ 881.625 , 2.677906223609988e-46 ], [ 881.6875 , 2.6782233524376037e-46 ], 
                    [ 881.75 , 2.678540518820938e-46 ], [ 881.8125 , 2.678857722764439e-46 ], [ 881.875 , 2.6791745263817063e-46 ], [ 881.9375 , 2.679491336718367e-46 ], 
                    [ 882.0 , 2.679808184517665e-46 ], [ 882.0625 , 2.6801250697839155e-46 ], [ 882.125 , 2.680441992521625e-46 ], [ 882.1875 , 2.6807589527352245e-46 ], 
                    [ 882.25 , 2.681075950429146e-46 ], [ 882.3125 , 2.6813929856077643e-46 ], [ 882.375 , 2.681710058275549e-46 ], [ 882.4375 , 2.6820271684369726e-46 ], 
                    [ 882.5 , 2.6823443160963726e-46 ], [ 882.5625 , 2.682661501258298e-46 ], [ 882.625 , 2.6829787239270686e-46 ], [ 882.6875 , 2.683295984107215e-46 ], 
                    [ 882.75 , 2.683613281803135e-46 ], [ 882.8125 , 2.683930617019264e-46 ], [ 882.875 , 2.684247989760002e-46 ], [ 882.9375 , 2.6845654000298614e-46 ], 
                    [ 883.0 , 2.6848828478332043e-46 ], [ 883.0625 , 2.6852003331745447e-46 ], [ 883.125 , 2.685517856058303e-46 ], [ 883.1875 , 2.6858354164888616e-46 ], 
                    [ 883.25 , 2.6861530144707356e-46 ], [ 883.3125 , 2.6864706500082702e-46 ], [ 883.375 , 2.6867883231060026e-46 ], [ 883.4375 , 2.6871060337683164e-46 ], 
                    [ 883.5 , 2.6874237819996918e-46 ], [ 883.5625 , 2.6877415678045526e-46 ], [ 883.625 , 2.688059391187322e-46 ], [ 883.6875 , 2.6883772521524635e-46 ], 
                    [ 883.75 , 2.688695150704421e-46 ], [ 883.8125 , 2.6890130868476765e-46 ], [ 883.875 , 2.689330601379732e-46 ], [ 883.9375 , 2.6896481313798802e-46 ], 
                    [ 884.0 , 2.6899656988708626e-46 ], [ 884.0625 , 2.6902833038571264e-46 ], [ 884.125 , 2.6906009463430976e-46 ], [ 884.1875 , 2.690918626333204e-46 ], 
                    [ 884.25 , 2.6912363438318548e-46 ], [ 884.3125 , 2.6915540988435355e-46 ], [ 884.375 , 2.6918718913725992e-46 ], [ 884.4375 , 2.6921897214235335e-46 ], 
                    [ 884.5 , 2.6925075890007675e-46 ], [ 884.5625 , 2.6928254941086558e-46 ], [ 884.625 , 2.693143436751745e-46 ], [ 884.6875 , 2.6934614169343706e-46 ], 
                    [ 884.75 , 2.693779434661061e-46 ], [ 884.8125 , 2.694097489936134e-46 ], [ 884.875 , 2.6944155827640803e-46 ], [ 884.9375 , 2.6947337131493335e-46 ], 
                    [ 885.0 , 2.695051881096386e-46 ], [ 885.0625 , 2.6953700866095773e-46 ], [ 885.125 , 2.6956883296933606e-46 ], [ 885.1875 , 2.6960066103521925e-46 ], 
                    [ 885.25 , 2.6963249285905473e-46 ], [ 885.3125 , 2.696643284412804e-46 ], [ 885.375 , 2.6969616778234202e-46 ], [ 885.4375 , 2.6972801088268136e-46 ], 
                    [ 885.5 , 2.6975985774274818e-46 ], [ 885.5625 , 2.6979170836297867e-46 ], [ 885.625 , 2.6982356274382245e-46 ], [ 885.6875 , 2.6985542088572174e-46 ], 
                    [ 885.75 , 2.6988728278912052e-46 ], [ 885.8125 , 2.6991914845446303e-46 ], [ 885.875 , 2.6995097445330592e-46 ], [ 885.9375 , 2.6998279945582026e-46 ], 
                    [ 886.0 , 2.7001462821024295e-46 ], [ 886.0625 , 2.700464607170143e-46 ], [ 886.125 , 2.7007829697657865e-46 ], [ 886.1875 , 2.7011013698937265e-46 ], 
                    [ 886.25 , 2.7014198075584455e-46 ], [ 886.3125 , 2.70173828276435e-46 ], [ 886.375 , 2.7020567955158648e-46 ], [ 886.4375 , 2.702375345817417e-46 ], 
                    [ 886.5 , 2.702693933673452e-46 ], [ 886.5625 , 2.7030125590883405e-46 ], [ 886.625 , 2.7033312220665676e-46 ], [ 886.6875 , 2.703649922612542e-46 ], 
                    [ 886.75 , 2.7039686607306928e-46 ], [ 886.8125 , 2.7042874364254497e-46 ], [ 886.875 , 2.7046062497012615e-46 ], [ 886.9375 , 2.7049251005625023e-46 ], 
                    [ 887.0 , 2.7052439890136397e-46 ], [ 887.0625 , 2.7055629150591255e-46 ], [ 887.125 , 2.7058818787033727e-46 ], [ 887.1875 , 2.7062008799508325e-46 ], 
                    [ 887.25 , 2.7065199188059e-46 ], [ 887.3125 , 2.706838995273009e-46 ], [ 887.375 , 2.707158109356632e-46 ], [ 887.4375 , 2.7074772610612028e-46 ], 
                    [ 887.5 , 2.707796450391158e-46 ], [ 887.5625 , 2.7081156773508356e-46 ], [ 887.625 , 2.7084349419447696e-46 ], [ 887.6875 , 2.7087542441773958e-46 ], 
                    [ 887.75 , 2.709073584053171e-46 ], [ 887.8125 , 2.7093929615764552e-46 ], [ 887.875 , 2.709712009819901e-46 ], [ 887.9375 , 2.7100309897998323e-46 ], 
                    [ 888.0 , 2.7103500073293025e-46 ], [ 888.0625 , 2.7106690624125783e-46 ], [ 888.125 , 2.7109881550541945e-46 ], [ 888.1875 , 2.711307285258535e-46 ], 
                    [ 888.25 , 2.7116264530300025e-46 ], [ 888.3125 , 2.7119456583730566e-46 ], [ 888.375 , 2.7122649012921024e-46 ], [ 888.4375 , 2.7125841817915237e-46 ], 
                    [ 888.5 , 2.7129034998758215e-46 ], [ 888.5625 , 2.713222855549382e-46 ], [ 888.625 , 2.71354224881663e-46 ], [ 888.6875 , 2.713861679681971e-46 ], 
                    [ 888.75 , 2.7141811481498515e-46 ], [ 888.8125 , 2.7145006542247356e-46 ], [ 888.875 , 2.7148201979109737e-46 ], [ 888.9375 , 2.7151397792130895e-46 ], 
                    [ 889.0 , 2.7154593981353754e-46 ], [ 889.0625 , 2.715779054682338e-46 ], [ 889.125 , 2.716098748858483e-46 ], [ 889.1875 , 2.716418480668086e-46 ], 
                    [ 889.25 , 2.7167382501157124e-46 ], [ 889.3125 , 2.717058057205715e-46 ], [ 889.375 , 2.717377901942506e-46 ], [ 889.4375 , 2.717697784330594e-46 ], 
                    [ 889.5 , 2.7180177043743922e-46 ], [ 889.5625 , 2.7183376620782564e-46 ], [ 889.625 , 2.718657657446696e-46 ], [ 889.6875 , 2.7189776904841275e-46 ], 
                    [ 889.75 , 2.7192977611950216e-46 ], [ 889.8125 , 2.7196178695837564e-46 ], [ 889.875 , 2.7199377287168403e-46 ], [ 889.9375 , 2.720257415881281e-46 ], 
                    [ 890.0 , 2.7205771406200677e-46 ], [ 890.0625 , 2.7208969029375984e-46 ], [ 890.125 , 2.7212167028383283e-46 ], [ 890.1875 , 2.7215365403266547e-46 ], 
                    [ 890.25 , 2.7218564154070157e-46 ], [ 890.3125 , 2.722176328083791e-46 ], [ 890.375 , 2.722496278361418e-46 ], [ 890.4375 , 2.7228162662443363e-46 ], 
                    [ 890.5 , 2.7231362917369267e-46 ], [ 890.5625 , 2.7234563548436486e-46 ], [ 890.625 , 2.7237764555688835e-46 ], [ 890.6875 , 2.7240965939170734e-46 ], 
                    [ 890.75 , 2.7244167698926396e-46 ], [ 890.8125 , 2.7247369835000043e-46 ], [ 890.875 , 2.7250572347435913e-46 ], [ 890.9375 , 2.7253775236278435e-46 ], 
                    [ 891.0 , 2.7256978501571454e-46 ], [ 891.0625 , 2.726018214335942e-46 ], [ 891.125 , 2.7263386161686773e-46 ], [ 891.1875 , 2.7266590556597187e-46 ], 
                    [ 891.25 , 2.7269795328135505e-46 ], [ 891.3125 , 2.7273000476346e-46 ], [ 891.375 , 2.7276206001272355e-46 ], [ 891.4375 , 2.7279411902959238e-46 ], 
                    [ 891.5 , 2.7282618181451706e-46 ], [ 891.5625 , 2.7285824836792498e-46 ], [ 891.625 , 2.7289031869026872e-46 ], [ 891.6875 , 2.729223927819913e-46 ], 
                    [ 891.75 , 2.7295447064353186e-46 ], [ 891.8125 , 2.7298655227533543e-46 ], [ 891.875 , 2.7301862304634998e-46 ], [ 891.9375 , 2.7305066450957698e-46 ], 
                    [ 892.0 , 2.730827097331937e-46 ], [ 892.0625 , 2.7311475871763755e-46 ], [ 892.125 , 2.7314681146334803e-46 ], [ 892.1875 , 2.731788679707685e-46 ], 
                    [ 892.25 , 2.7321092824034227e-46 ], [ 892.3125 , 2.732429922725052e-46 ], [ 892.375 , 2.7327506006770646e-46 ], [ 892.4375 , 2.7330713162638396e-46 ], 
                    [ 892.5 , 2.733392069489773e-46 ], [ 892.5625 , 2.7337128603593027e-46 ], [ 892.625 , 2.7340336888768458e-46 ], [ 892.6875 , 2.7343545550468408e-46 ], 
                    [ 892.75 , 2.7346754588736668e-46 ], [ 892.8125 , 2.7349964003617826e-46 ], [ 892.875 , 2.7353173795155886e-46 ], [ 892.9375 , 2.7356383963394667e-46 ], 
                    [ 893.0 , 2.7359594508379157e-46 ], [ 893.0625 , 2.736280543015298e-46 ], [ 893.125 , 2.736601672876055e-46 ], [ 893.1875 , 2.7369228404246297e-46 ], 
                    [ 893.25 , 2.737244045665444e-46 ], [ 893.3125 , 2.7375652886028643e-46 ], [ 893.375 , 2.737886569241333e-46 ], [ 893.4375 , 2.7382078875853533e-46 ], 
                    [ 893.5 , 2.7385292436392336e-46 ], [ 893.5625 , 2.7388506374074766e-46 ], [ 893.625 , 2.7391720688944895e-46 ], [ 893.6875 , 2.7394935381047376e-46 ], 
                    [ 893.75 , 2.7398150450425904e-46 ], [ 893.8125 , 2.7401365897124555e-46 ], [ 893.875 , 2.74045817211882e-46 ], [ 893.9375 , 2.7407793149618183e-46 ], 
                    [ 894.0 , 2.741100476870647e-46 ], [ 894.0625 , 2.741421676412903e-46 ], [ 894.125 , 2.7417429135930347e-46 ], [ 894.1875 , 2.7420641884154524e-46 ], 
                    [ 894.25 , 2.7423855008845674e-46 ], [ 894.3125 , 2.7427068510047325e-46 ], [ 894.375 , 2.7430282387803986e-46 ], [ 894.4375 , 2.7433496642159974e-46 ], 
                    [ 894.5 , 2.743671127315883e-46 ], [ 894.5625 , 2.7439926280845287e-46 ], [ 894.625 , 2.7443141665263472e-46 ], [ 894.6875 , 2.744635742645695e-46 ], 
                    [ 894.75 , 2.7449573564470846e-46 ], [ 894.8125 , 2.745279007934834e-46 ], [ 894.875 , 2.745600697113398e-46 ], [ 894.9375 , 2.745922423987233e-46 ], 
                    [ 895.0 , 2.7462441885606384e-46 ], [ 895.0625 , 2.7465659908381686e-46 ], [ 895.125 , 2.746887830824163e-46 ], [ 895.1875 , 2.7472097085230803e-46 ], 
                    [ 895.25 , 2.7475316239393194e-46 ], [ 895.3125 , 2.7478535770773007e-46 ], [ 895.375 , 2.748175567941424e-46 ], [ 895.4375 , 2.7484975965361107e-46 ], 
                    [ 895.5 , 2.748819662865821e-46 ], [ 895.5625 , 2.7491417669349763e-46 ], [ 895.625 , 2.74946390874796e-46 ], [ 895.6875 , 2.749786088309234e-46 ], 
                    [ 895.75 , 2.750108305623221e-46 ], [ 895.8125 , 2.7504305606942874e-46 ], [ 895.875 , 2.7507528535268765e-46 ], [ 895.9375 , 2.751074919578007e-46 ], 
                    [ 896.0 , 2.7513968022727297e-46 ], [ 896.0625 , 2.751718722628578e-46 ], [ 896.125 , 2.7520406806498988e-46 ], [ 896.1875 , 2.752362676341119e-46 ], 
                    [ 896.25 , 2.752684709706666e-46 ], [ 896.3125 , 2.7530067807509077e-46 ], [ 896.375 , 2.7533288894782735e-46 ], [ 896.4375 , 2.7536510358932104e-46 ], 
                    [ 896.5 , 2.7539732200000306e-46 ], [ 896.5625 , 2.7542954418032225e-46 ], [ 896.625 , 2.7546177013071573e-46 ], [ 896.6875 , 2.7549399985163245e-46 ], 
                    [ 896.75 , 2.7552623334350376e-46 ], [ 896.8125 , 2.7555847060677482e-46 ], [ 896.875 , 2.7559071164188887e-46 ], [ 896.9375 , 2.756229564492852e-46 ], 
                    [ 897.0 , 2.756552050294052e-46 ], [ 897.0625 , 2.7568745738268636e-46 ], [ 897.125 , 2.757197135095761e-46 ], [ 897.1875 , 2.757519734105139e-46 ], 
                    [ 897.25 , 2.757842370859453e-46 ], [ 897.3125 , 2.7581650453630212e-46 ], [ 897.375 , 2.758487757620358e-46 ], [ 897.4375 , 2.7588105076358026e-46 ], 
                    [ 897.5 , 2.7591332954138702e-46 ], [ 897.5625 , 2.7594561209588824e-46 ], [ 897.625 , 2.7597789842752964e-46 ], [ 897.6875 , 2.760101885367512e-46 ], 
                    [ 897.75 , 2.760424824240027e-46 ], [ 897.8125 , 2.7607478008971453e-46 ], [ 897.875 , 2.7610708153433653e-46 ], [ 897.9375 , 2.7613938411614896e-46 ], 
                    [ 898.0 , 2.7617164396646165e-46 ], [ 898.0625 , 2.7620390758551773e-46 ], [ 898.125 , 2.762361749737536e-46 ], [ 898.1875 , 2.7626844613161154e-46 ], 
                    [ 898.25 , 2.763007210595339e-46 ], [ 898.3125 , 2.763329997579611e-46 ], [ 898.375 , 2.7636528222732577e-46 ], [ 898.4375 , 2.7639756846808022e-46 ], 
                    [ 898.5 , 2.764298584806533e-46 ], [ 898.5625 , 2.7646215226549545e-46 ], [ 898.625 , 2.7649444982304538e-46 ], [ 898.6875 , 2.7652675115373798e-46 ], 
                    [ 898.75 , 2.765590562580179e-46 ], [ 898.8125 , 2.7659136513633006e-46 ], [ 898.875 , 2.7662367778910735e-46 ], [ 898.9375 , 2.766559942167967e-46 ], 
                    [ 899.0 , 2.7668831441983314e-46 ], [ 899.0625 , 2.7672063839866766e-46 ], [ 899.125 , 2.7675296615373344e-46 ], [ 899.1875 , 2.767852976854736e-46 ], 
                    [ 899.25 , 2.768176329943254e-46 ], [ 899.3125 , 2.7684997208074e-46 ], [ 899.375 , 2.7688231494515073e-46 ], [ 899.4375 , 2.7691466158799908e-46 ], 
                    [ 899.5 , 2.769470120097264e-46 ], [ 899.5625 , 2.76979366210784e-46 ], [ 899.625 , 2.7701172419160154e-46 ], [ 899.6875 , 2.7704408595262266e-46 ], 
                    [ 899.75 , 2.7707645149429285e-46 ], [ 899.8125 , 2.771088208170479e-46 ], [ 899.875 , 2.771411939213394e-46 ], [ 899.9375 , 2.771735708075992e-46 ], 
                    [ 900.0 , 2.7720592728703908e-46 ], [ 900.0625 , 2.7723826225918417e-46 ], [ 900.125 , 2.772706010030735e-46 ], [ 900.1875 , 2.7730294351914913e-46 ], 
                    [ 900.25 , 2.7733528980785295e-46 ], [ 900.3125 , 2.7736763986962305e-46 ], [ 900.375 , 2.7739999370489963e-46 ], [ 900.4375 , 2.7743235131412075e-46 ], 
                    [ 900.5 , 2.774647126977287e-46 ], [ 900.5625 , 2.774970778561657e-46 ], [ 900.625 , 2.775294467898701e-46 ], [ 900.6875 , 2.7756181949928226e-46 ], 
                    [ 900.75 , 2.7759419598484057e-46 ], [ 900.8125 , 2.7762657624698954e-46 ], [ 900.875 , 2.7765896028616573e-46 ], [ 900.9375 , 2.776913481028136e-46 ], 
                    [ 901.0 , 2.777237396973719e-46 ], [ 901.0625 , 2.7775613507028117e-46 ], [ 901.125 , 2.7778853422198035e-46 ], [ 901.1875 , 2.7782093715291005e-46 ], 
                    [ 901.25 , 2.7785334386351908e-46 ], [ 901.3125 , 2.7788575435424047e-46 ], [ 901.375 , 2.77918168625517e-46 ], [ 901.4375 , 2.779505866777898e-46 ], 
                    [ 901.5 , 2.7798300851149783e-46 ], [ 901.5625 , 2.780154341270842e-46 ], [ 901.625 , 2.78047863524992e-46 ], [ 901.6875 , 2.780802967056605e-46 ], 
                    [ 901.75 , 2.78112733669527e-46 ], [ 901.8125 , 2.7814517441704253e-46 ], [ 901.875 , 2.781776189486387e-46 ], [ 901.9375 , 2.782100672647608e-46 ], 
                    [ 902.0 , 2.782425193658543e-46 ], [ 902.0625 , 2.782749329489309e-46 ], [ 902.125 , 2.7830734286523314e-46 ], [ 902.1875 , 2.7833975655622723e-46 ], 
                    [ 902.25 , 2.7837217402235084e-46 ], [ 902.3125 , 2.784045952640515e-46 ], [ 902.375 , 2.784370202817651e-46 ], [ 902.4375 , 2.7846944907593136e-46 ], 
                    [ 902.5 , 2.7850188164698812e-46 ], [ 902.5625 , 2.7853431799537923e-46 ], [ 902.625 , 2.785667581215426e-46 ], [ 902.6875 , 2.785992020259183e-46 ], 
                    [ 902.75 , 2.7863164970894243e-46 ], [ 902.8125 , 2.7866410117106486e-46 ], [ 902.875 , 2.786965564127159e-46 ], [ 902.9375 , 2.787290154343437e-46 ], 
                    [ 903.0 , 2.787614782363825e-46 ], [ 903.0625 , 2.787939448192706e-46 ], [ 903.125 , 2.7882641518345832e-46 ], [ 903.1875 , 2.7885888932937806e-46 ], 
                    [ 903.25 , 2.7889136725747233e-46 ], [ 903.3125 , 2.7892384896818154e-46 ], [ 903.375 , 2.7895633446194633e-46 ], [ 903.4375 , 2.7898882373920722e-46 ], 
                    [ 903.5 , 2.7902131680040297e-46 ], [ 903.5625 , 2.7905381364597816e-46 ], [ 903.625 , 2.7908631427636958e-46 ], [ 903.6875 , 2.791188186920221e-46 ], 
                    [ 903.75 , 2.7915132689337253e-46 ], [ 903.8125 , 2.7918383888086578e-46 ], [ 903.875 , 2.7921635465494076e-46 ], [ 903.9375 , 2.7924887421603667e-46 ], 
                    [ 904.0 , 2.7928139756459834e-46 ], [ 904.0625 , 2.793139178305642e-46 ], [ 904.125 , 2.7934639893366227e-46 ], [ 904.1875 , 2.7937888381394954e-46 ], 
                    [ 904.25 , 2.7941137247186713e-46 ], [ 904.3125 , 2.794438649078545e-46 ], [ 904.375 , 2.794763611223489e-46 ], [ 904.4375 , 2.7950886111579568e-46 ], 
                    [ 904.5 , 2.7954136488862837e-46 ], [ 904.5625 , 2.7957387244129045e-46 ], [ 904.625 , 2.796063837742155e-46 ], [ 904.6875 , 2.7963889888785104e-46 ], 
                    [ 904.75 , 2.796714177826308e-46 ], [ 904.8125 , 2.7970394045899446e-46 ], [ 904.875 , 2.797364669173878e-46 ], [ 904.9375 , 2.797689971582446e-46 ], 
                    [ 905.0 , 2.7980153118200667e-46 ], [ 905.0625 , 2.7983406898911402e-46 ], [ 905.125 , 2.7986661058000855e-46 ], [ 905.1875 , 2.7989915595512633e-46 ], 
                    [ 905.25 , 2.7993170511490935e-46 ], [ 905.3125 , 2.7996425805979575e-46 ], [ 905.375 , 2.799968147902297e-46 ], [ 905.4375 , 2.8002937530665144e-46 ], 
                    [ 905.5 , 2.800619396094992e-46 ], [ 905.5625 , 2.8009450769920734e-46 ], [ 905.625 , 2.801270795762262e-46 ], [ 905.6875 , 2.8015965524099416e-46 ], 
                    [ 905.75 , 2.8019223469394777e-46 ], [ 905.8125 , 2.8022481793552953e-46 ], [ 905.875 , 2.8025740496617802e-46 ], [ 905.9375 , 2.8028999578633984e-46 ], 
                    [ 906.0 , 2.803225903964497e-46 ], [ 906.0625 , 2.803551887969523e-46 ], [ 906.125 , 2.80387773577542e-46 ], [ 906.1875 , 2.8042032997736014e-46 ], 
                    [ 906.25 , 2.804528901573667e-46 ], [ 906.3125 , 2.804854541180066e-46 ], [ 906.375 , 2.8051802185971077e-46 ], [ 906.4375 , 2.8055059338292234e-46 ], 
                    [ 906.5 , 2.8058316868808423e-46 ], [ 906.5625 , 2.806157477756277e-46 ], [ 906.625 , 2.8064833064599588e-46 ], [ 906.6875 , 2.8068091729963e-46 ], 
                    [ 906.75 , 2.8071350773696533e-46 ], [ 906.8125 , 2.8074610195844726e-46 ], [ 906.875 , 2.8077869996450512e-46 ], [ 906.9375 , 2.808113017555884e-46 ], 
                    [ 907.0 , 2.8084390733213248e-46 ], [ 907.0625 , 2.80876516694577e-46 ], [ 907.125 , 2.809091298433575e-46 ], [ 907.1875 , 2.8094174677892564e-46 ], 
                    [ 907.25 , 2.8097436750170714e-46 ], [ 907.3125 , 2.810069920121477e-46 ], [ 907.375 , 2.810396203106931e-46 ], [ 907.4375 , 2.810722523977692e-46 ], 
                    [ 907.5 , 2.8110488827382593e-46 ], [ 907.5625 , 2.8113752793930316e-46 ], [ 907.625 , 2.81170171394635e-46 ], [ 907.6875 , 2.812028186402714e-46 ], 
                    [ 907.75 , 2.812354696766425e-46 ], [ 907.8125 , 2.812681245041944e-46 ], [ 907.875 , 2.8130078312336537e-46 ], [ 907.9375 , 2.8133344553459364e-46 ], 
                    [ 908.0 , 2.813661117383235e-46 ], [ 908.0625 , 2.8139878173499133e-46 ], [ 908.125 , 2.814314555250435e-46 ], [ 908.1875 , 2.8146410809044746e-46 ], 
                    [ 908.25 , 2.814967384266585e-46 ], [ 908.3125 , 2.815293725457299e-46 ], [ 908.375 , 2.8156201044809614e-46 ], [ 908.4375 , 2.8159465213419787e-46 ], 
                    [ 908.5 , 2.816272976044757e-46 ], [ 908.5625 , 2.8165994685936434e-46 ], [ 908.625 , 2.8169259989930455e-46 ], [ 908.6875 , 2.8172525672473715e-46 ], 
                    [ 908.75 , 2.8175791733609694e-46 ], [ 908.8125 , 2.817905817338249e-46 ], [ 908.875 , 2.8182324991835992e-46 ], [ 908.9375 , 2.8185592189014106e-46 ], 
                    [ 909.0 , 2.8188859764960735e-46 ], [ 909.0625 , 2.8192127719719993e-46 ], [ 909.125 , 2.8195396053335386e-46 ], [ 909.1875 , 2.819866476585105e-46 ], 
                    [ 909.25 , 2.8201933857310897e-46 ], [ 909.3125 , 2.8205203327758872e-46 ], [ 909.375 , 2.82084731772391e-46 ], [ 909.4375 , 2.821174340579533e-46 ], 
                    [ 909.5 , 2.82150140134713e-46 ], [ 909.5625 , 2.8218285000311173e-46 ], [ 909.625 , 2.8221556366359097e-46 ], [ 909.6875 , 2.8224828111658435e-46 ], 
                    [ 909.75 , 2.822810023625396e-46 ], [ 909.8125 , 2.8231372740189034e-46 ], [ 909.875 , 2.823464562350764e-46 ], [ 909.9375 , 2.8237918886254365e-46 ], 
                    [ 910.0 , 2.824119252847259e-46 ], [ 910.0625 , 2.8244466550206704e-46 ], [ 910.125 , 2.8247740951500712e-46 ], [ 910.1875 , 2.825101573239741e-46 ], 
                    [ 910.25 , 2.8254288101122267e-46 ], [ 910.3125 , 2.8257558607606875e-46 ], [ 910.375 , 2.826082949266062e-46 ], [ 910.4375 , 2.8264100756328116e-46 ], 
                    [ 910.5 , 2.8267372398652594e-46 ], [ 910.5625 , 2.827064441967808e-46 ], [ 910.625 , 2.8273916819448415e-46 ], [ 910.6875 , 2.827718959800743e-46 ], 
                    [ 910.75 , 2.8280462755398983e-46 ], [ 910.8125 , 2.8283736291666917e-46 ], [ 910.875 , 2.8287010206854887e-46 ], [ 910.9375 , 2.829028450100736e-46 ], 
                    [ 911.0 , 2.8293559174167595e-46 ], [ 911.0625 , 2.829683422637947e-46 ], [ 911.125 , 2.8300109657687456e-46 ], [ 911.1875 , 2.8303385468134637e-46 ], 
                    [ 911.25 , 2.8306661657765502e-46 ], [ 911.3125 , 2.8309938226623737e-46 ], [ 911.375 , 2.8313215174753236e-46 ], [ 911.4375 , 2.8316492502197912e-46 ], 
                    [ 911.5 , 2.8319770209001664e-46 ], [ 911.5625 , 2.8323048295208402e-46 ], [ 911.625 , 2.8326326760862047e-46 ], [ 911.6875 , 2.8329605606006524e-46 ], 
                    [ 911.75 , 2.833288483068595e-46 ], [ 911.8125 , 2.8336164434943666e-46 ], [ 911.875 , 2.8339444418824407e-46 ], [ 911.9375 , 2.834272478237131e-46 ], 
                    [ 912.0 , 2.8346005525629134e-46 ], [ 912.0625 , 2.834928664864102e-46 ], [ 912.125 , 2.8352568151451327e-46 ], [ 912.1875 , 2.8355850034104223e-46 ], 
                    [ 912.25 , 2.8359132296643474e-46 ], [ 912.3125 , 2.8362412174406313e-46 ], [ 912.375 , 2.8365690119733367e-46 ], [ 912.4375 , 2.836896844390387e-46 ], 
                    [ 912.5 , 2.837224714696201e-46 ], [ 912.5625 , 2.8375526228951785e-46 ], [ 912.625 , 2.837880568991698e-46 ], [ 912.6875 , 2.8382085529901206e-46 ], 
                    [ 912.75 , 2.838536574894805e-46 ], [ 912.8125 , 2.8388646347101535e-46 ], [ 912.875 , 2.8391927324405466e-46 ], [ 912.9375 , 2.8395208680903676e-46 ], 
                    [ 913.0 , 2.839849041663998e-46 ], [ 913.0625 , 2.8401772531658207e-46 ], [ 913.125 , 2.84050550260022e-46 ], [ 913.1875 , 2.8408337899715787e-46 ], 
                    [ 913.25 , 2.8411621152842824e-46 ], [ 913.3125 , 2.841490478542716e-46 ], [ 913.375 , 2.8418188797512643e-46 ], [ 913.4375 , 2.8421473189143347e-46 ], 
                    [ 913.5 , 2.8424757960362516e-46 ], [ 913.5625 , 2.8428043111215047e-46 ], [ 913.625 , 2.8431328641743386e-46 ], [ 913.6875 , 2.843461455199284e-46 ], 
                    [ 913.75 , 2.843790084200668e-46 ], [ 913.8125 , 2.84411875118286e-46 ], [ 913.875 , 2.8444474561502697e-46 ], [ 913.9375 , 2.844776199107287e-46 ], 
                    [ 914.0 , 2.8451049800583023e-46 ], [ 914.0625 , 2.8454337990077075e-46 ], [ 914.125 , 2.8457626559598936e-46 ], [ 914.1875 , 2.846091550919253e-46 ], 
                    [ 914.25 , 2.8464204838901978e-46 ], [ 914.3125 , 2.8467494548770813e-46 ], [ 914.375 , 2.8470782330295815e-46 ], [ 914.4375 , 2.84740677776948e-46 ], 
                    [ 914.5 , 2.847735360422488e-46 ], [ 914.5625 , 2.848063980993041e-46 ], [ 914.625 , 2.8483926394854335e-46 ], [ 914.6875 , 2.848721335904103e-46 ], 
                    [ 914.75 , 2.8490500702533853e-46 ], [ 914.8125 , 2.8493788425376976e-46 ], [ 914.875 , 2.8497076527613977e-46 ], [ 914.9375 , 2.850036500928843e-46 ], 
                    [ 915.0 , 2.8503653870444328e-46 ], [ 915.0625 , 2.8506943111125865e-46 ], [ 915.125 , 2.8510232731376023e-46 ], [ 915.1875 , 2.851352273123922e-46 ], 
                    [ 915.25 , 2.851681311075905e-46 ], [ 915.3125 , 2.852010386997913e-46 ], [ 915.375 , 2.852339500894368e-46 ], [ 915.4375 , 2.8526686527696313e-46 ], 
                    [ 915.5 , 2.852997842628066e-46 ], [ 915.5625 , 2.8533270704740747e-46 ], [ 915.625 , 2.8536563363121026e-46 ], [ 915.6875 , 2.853985640146392e-46 ], 
                    [ 915.75 , 2.854314981981428e-46 ], [ 915.8125 , 2.854644361821597e-46 ], [ 915.875 , 2.854973779671264e-46 ], [ 915.9375 , 2.855303235534794e-46 ], 
                    [ 916.0 , 2.8556327294166153e-46 ], [ 916.0625 , 2.8559622613210744e-46 ], [ 916.125 , 2.8562918312526196e-46 ], [ 916.1875 , 2.8566214392155583e-46 ], 
                    [ 916.25 , 2.856951085214339e-46 ], [ 916.3125 , 2.857280769253312e-46 ], [ 916.375 , 2.8576104913369064e-46 ], [ 916.4375 , 2.8579400970742665e-46 ], 
                    [ 916.5 , 2.85826937998272e-46 ], [ 916.5625 , 2.858598700830119e-46 ], [ 916.625 , 2.8589280596208555e-46 ], [ 916.6875 , 2.8592574563592805e-46 ], 
                    [ 916.75 , 2.8595868910497657e-46 ], [ 916.8125 , 2.8599163636966443e-46 ], [ 916.875 , 2.8602458743043494e-46 ], [ 916.9375 , 2.8605754228772553e-46 ], 
                    [ 917.0 , 2.8609050094197158e-46 ], [ 917.0625 , 2.8612346339360653e-46 ], [ 917.125 , 2.8615642964307595e-46 ], [ 917.1875 , 2.861893996908114e-46 ], 
                    [ 917.25 , 2.8622237353725253e-46 ], [ 917.3125 , 2.8625535118283893e-46 ], [ 917.375 , 2.8628833262800445e-46 ], [ 917.4375 , 2.863213178731908e-46 ], 
                    [ 917.5 , 2.8635430691883384e-46 ], [ 917.5625 , 2.8638729976536527e-46 ], [ 917.625 , 2.864202964132373e-46 ], [ 917.6875 , 2.864532968628737e-46 ], 
                    [ 917.75 , 2.864863011147246e-46 ], [ 917.8125 , 2.865193091692201e-46 ], [ 917.875 , 2.8655232102679614e-46 ], [ 917.9375 , 2.865853366878971e-46 ], 
                    [ 918.0 , 2.8661835615295917e-46 ], [ 918.0625 , 2.8665137942242257e-46 ], [ 918.125 , 2.866844064967217e-46 ], [ 918.1875 , 2.8671743737630095e-46 ], 
                    [ 918.25 , 2.867504720615886e-46 ], [ 918.3125 , 2.8678351055303333e-46 ], [ 918.375 , 2.8681655285106553e-46 ], [ 918.4375 , 2.8684959895613187e-46 ], 
                    [ 918.5 , 2.868826455556962e-46 ], [ 918.5625 , 2.8691564789004298e-46 ], [ 918.625 , 2.869486540209041e-46 ], [ 918.6875 , 2.869816639487162e-46 ], 
                    [ 918.75 , 2.8701467767391616e-46 ], [ 918.8125 , 2.870476951969409e-46 ], [ 918.875 , 2.8708071651822514e-46 ], [ 918.9375 , 2.87113741638212e-46 ], 
                    [ 919.0 , 2.8714677055733233e-46 ], [ 919.0625 , 2.8717980327602123e-46 ], [ 919.125 , 2.8721283979472786e-46 ], [ 919.1875 , 2.8724588011387726e-46 ], 
                    [ 919.25 , 2.8727892423391067e-46 ], [ 919.3125 , 2.8731197215526126e-46 ], [ 919.375 , 2.8734502387837856e-46 ], [ 919.4375 , 2.8737807940368768e-46 ], 
                    [ 919.5 , 2.8741113873162804e-46 ], [ 919.5625 , 2.874442018626391e-46 ], [ 919.625 , 2.8747726879716247e-46 ], [ 919.6875 , 2.8751033953563166e-46 ], 
                    [ 919.75 , 2.8754341407848017e-46 ], [ 919.8125 , 2.8757649242615786e-46 ], [ 919.875 , 2.87609574579082e-46 ], [ 919.9375 , 2.876426605377108e-46 ], 
                    [ 920.0 , 2.8767575030246975e-46 ], [ 920.0625 , 2.8770884387380496e-46 ], [ 920.125 , 2.8774194125214603e-46 ], [ 920.1875 , 2.8777504243793923e-46 ], 
                    [ 920.25 , 2.878081474316122e-46 ], [ 920.3125 , 2.878412562336174e-46 ], [ 920.375 , 2.8787436884437657e-46 ], [ 920.4375 , 2.8790748526434003e-46 ], 
                    [ 920.5 , 2.879406054939461e-46 ], [ 920.5625 , 2.8797372953362482e-46 ], [ 920.625 , 2.8800681873257428e-46 ], [ 920.6875 , 2.8803989912722673e-46 ], 
                    [ 920.75 , 2.8807298332148954e-46 ], [ 920.8125 , 2.88106071315793e-46 ], [ 920.875 , 2.8813916311057557e-46 ], [ 920.9375 , 2.8817225870627382e-46 ], 
                    [ 921.0 , 2.8820535810332437e-46 ], [ 921.0625 , 2.8823846130216785e-46 ], [ 921.125 , 2.8827156830323068e-46 ], [ 921.1875 , 2.883046791069579e-46 ], 
                    [ 921.25 , 2.883377937137882e-46 ], [ 921.3125 , 2.883709121241482e-46 ], [ 921.375 , 2.8840403433848305e-46 ], [ 921.4375 , 2.884371603572275e-46 ], 
                    [ 921.5 , 2.884702901808186e-46 ], [ 921.5625 , 2.885034238096954e-46 ], [ 921.625 , 2.8853656124429292e-46 ], [ 921.6875 , 2.8856970248504623e-46 ], 
                    [ 921.75 , 2.8860284753239667e-46 ], [ 921.8125 , 2.8863599638677725e-46 ], [ 921.875 , 2.8866914904862944e-46 ], [ 921.9375 , 2.8870230551838642e-46 ], 
                    [ 922.0 , 2.8873546579648765e-46 ], [ 922.0625 , 2.887686298833684e-46 ], [ 922.125 , 2.8880179777947246e-46 ], [ 922.1875 , 2.8883496948523105e-46 ], 
                    [ 922.25 , 2.88868145001086e-46 ], [ 922.3125 , 2.8890132432746863e-46 ], [ 922.375 , 2.8893450746482495e-46 ], [ 922.4375 , 2.8896769441358444e-46 ], 
                    [ 922.5 , 2.8900088517419302e-46 ], [ 922.5625 , 2.890340797470824e-46 ], [ 922.625 , 2.8906727813269466e-46 ], [ 922.6875 , 2.891004612838908e-46 ], 
                    [ 922.75 , 2.891336155509173e-46 ], [ 922.8125 , 2.8916677362010308e-46 ], [ 922.875 , 2.8919993549188013e-46 ], [ 922.9375 , 2.8923310116668863e-46 ], 
                    [ 923.0 , 2.892662706449606e-46 ], [ 923.0625 , 2.8929944392713427e-46 ], [ 923.125 , 2.8933262101364592e-46 ], [ 923.1875 , 2.8936580190493588e-46 ], 
                    [ 923.25 , 2.893989866014323e-46 ], [ 923.3125 , 2.8943217510357365e-46 ], [ 923.375 , 2.894653674118025e-46 ], [ 923.4375 , 2.8949856352655117e-46 ], 
                    [ 923.5 , 2.895317634482542e-46 ], [ 923.5625 , 2.8956496717735022e-46 ], [ 923.625 , 2.8959817471427796e-46 ], [ 923.6875 , 2.8963138605946993e-46 ], 
                    [ 923.75 , 2.896646012133649e-46 ], [ 923.8125 , 2.8969782017640183e-46 ], [ 923.875 , 2.8973104294901127e-46 ], [ 923.9375 , 2.8976426953163836e-46 ], 
                    [ 924.0 , 2.8979749992471595e-46 ], [ 924.0625 , 2.8983073412867896e-46 ], [ 924.125 , 2.8986397214397057e-46 ], [ 924.1875 , 2.8989721397101764e-46 ], 
                    [ 924.25 , 2.899304596102654e-46 ], [ 924.3125 , 2.899637090621492e-46 ], [ 924.375 , 2.8999696232711017e-46 ], [ 924.4375 , 2.9003021940557544e-46 ], 
                    [ 924.5 , 2.900634802979926e-46 ], [ 924.5625 , 2.9009674500479288e-46 ], [ 924.625 , 2.9013001352641366e-46 ], [ 924.6875 , 2.9016328586329864e-46 ], 
                    [ 924.75 , 2.901965620158812e-46 ], [ 924.8125 , 2.9022979524356416e-46 ], [ 924.875 , 2.9026302745891444e-46 ], [ 924.9375 , 2.9029626347945067e-46 ], 
                    [ 925.0 , 2.9032950330562098e-46 ], [ 925.0625 , 2.9036274693785277e-46 ], [ 925.125 , 2.903959943765882e-46 ], [ 925.1875 , 2.9042924562225676e-46 ], 
                    [ 925.25 , 2.904625006752944e-46 ], [ 925.3125 , 2.9049575953613923e-46 ], [ 925.375 , 2.905290222052251e-46 ], [ 925.4375 , 2.905622886829943e-46 ], 
                    [ 925.5 , 2.905955589698788e-46 ], [ 925.5625 , 2.9062883306631482e-46 ], [ 925.625 , 2.906621109727343e-46 ], [ 925.6875 , 2.9069539268957973e-46 ], 
                    [ 925.75 , 2.9072867821728958e-46 ], [ 925.8125 , 2.9076196755629182e-46 ], [ 925.875 , 2.9079526070702915e-46 ], [ 925.9375 , 2.9082855766992967e-46 ], 
                    [ 926.0 , 2.908618584454403e-46 ], [ 926.0625 , 2.908951630339955e-46 ], [ 926.125 , 2.909284714360298e-46 ], [ 926.1875 , 2.9096178365197774e-46 ], 
                    [ 926.25 , 2.909950996822782e-46 ], [ 926.3125 , 2.9102841952737198e-46 ], [ 926.375 , 2.9106174318768767e-46 ], [ 926.4375 , 2.9109507066366622e-46 ], 
                    [ 926.5 , 2.9112840195574457e-46 ], [ 926.5625 , 2.911617370643617e-46 ], [ 926.625 , 2.9119507598995055e-46 ], [ 926.6875 , 2.912284187329502e-46 ], 
                    [ 926.75 , 2.912617652937998e-46 ], [ 926.8125 , 2.9129511567293235e-46 ], [ 926.875 , 2.9132845064109937e-46 ], [ 926.9375 , 2.9136175655392024e-46 ], 
                    [ 927.0 , 2.9139506627441326e-46 ], [ 927.0625 , 2.914283798030179e-46 ], [ 927.125 , 2.914616971401674e-46 ], [ 927.1875 , 2.914950182862972e-46 ], 
                    [ 927.25 , 2.915283432418407e-46 ], [ 927.3125 , 2.9156167200723957e-46 ], [ 927.375 , 2.915950045829211e-46 ], [ 927.4375 , 2.9162834096932716e-46 ], 
                    [ 927.5 , 2.9166168116688925e-46 ], [ 927.5625 , 2.916950251760451e-46 ], [ 927.625 , 2.917283729972367e-46 ], [ 927.6875 , 2.9176172463088743e-46 ], 
                    [ 927.75 , 2.917950800774394e-46 ], [ 927.8125 , 2.9182843933733053e-46 ], [ 927.875 , 2.918618024109948e-46 ], [ 927.9375 , 2.9189516929886606e-46 ], 
                    [ 928.0 , 2.919285400013803e-46 ], [ 928.0625 , 2.9196191451898624e-46 ], [ 928.125 , 2.919952928521013e-46 ], [ 928.1875 , 2.920286750011679e-46 ], 
                    [ 928.25 , 2.9206206096662637e-46 ], [ 928.3125 , 2.9209545074891523e-46 ], [ 928.375 , 2.9212884434846864e-46 ], [ 928.4375 , 2.9216224176571272e-46 ], 
                    [ 928.5 , 2.9219564300109424e-46 ], [ 928.5625 , 2.922290480550539e-46 ], [ 928.625 , 2.9226245692801372e-46 ], [ 928.6875 , 2.9229586962042477e-46 ], 
                    [ 928.75 , 2.9232928613271555e-46 ], [ 928.8125 , 2.923627064653205e-46 ], [ 928.875 , 2.923961306186869e-46 ], [ 928.9375 , 2.9242955859324108e-46 ], 
                    [ 929.0 , 2.9246295055658245e-46 ], [ 929.0625 , 2.924963334418402e-46 ], [ 929.125 , 2.9252972013755125e-46 ], [ 929.1875 , 2.9256311064415678e-46 ], 
                    [ 929.25 , 2.9259650496208764e-46 ], [ 929.3125 , 2.926299030917726e-46 ], [ 929.375 , 2.926633050336593e-46 ], [ 929.4375 , 2.926967107881724e-46 ], 
                    [ 929.5 , 2.927301203557555e-46 ], [ 929.5625 , 2.927635337368354e-46 ], [ 929.625 , 2.9279695093185166e-46 ], [ 929.6875 , 2.9283037194124367e-46 ], 
                    [ 929.75 , 2.928637967654386e-46 ], [ 929.8125 , 2.9289722540487805e-46 ], [ 929.875 , 2.929306578599955e-46 ], [ 929.9375 , 2.9296409413122854e-46 ], 
                    [ 930.0 , 2.9299753421900856e-46 ], [ 930.0625 , 2.9303097812377327e-46 ], [ 930.125 , 2.9306442584595842e-46 ], [ 930.1875 , 2.930978773859976e-46 ], 
                    [ 930.25 , 2.9313133274432874e-46 ], [ 930.3125 , 2.931647919213918e-46 ], [ 930.375 , 2.9319825491761434e-46 ], [ 930.4375 , 2.9323172173343854e-46 ], 
                    [ 930.5 , 2.932651923692983e-46 ], [ 930.5625 , 2.9329866682562965e-46 ], [ 930.625 , 2.933321451028687e-46 ], [ 930.6875 , 2.9336562720144947e-46 ], 
                    [ 930.75 , 2.933991131218123e-46 ], [ 930.8125 , 2.934326028643935e-46 ], [ 930.875 , 2.9346609642962303e-46 ], [ 930.9375 , 2.9349959381794348e-46 ], 
                    [ 931.0 , 2.935330950297871e-46 ], [ 931.0625 , 2.9356659664868027e-46 ], [ 931.125 , 2.9360005366829388e-46 ], [ 931.1875 , 2.9363351450092097e-46 ], 
                    [ 931.25 , 2.9366697914698986e-46 ], [ 931.3125 , 2.9370044760693925e-46 ], [ 931.375 , 2.9373391988119757e-46 ], [ 931.4375 , 2.9376739597020793e-46 ], 
                    [ 931.5 , 2.938008758743946e-46 ], [ 931.5625 , 2.9383435959420494e-46 ], [ 931.625 , 2.938678471300675e-46 ], [ 931.6875 , 2.9390133848241517e-46 ], 
                    [ 931.75 , 2.939348336516828e-46 ], [ 931.8125 , 2.9396833263831176e-46 ], [ 931.875 , 2.940018354427287e-46 ], [ 931.9375 , 2.9403534206537704e-46 ], 
                    [ 932.0 , 2.940688525066858e-46 ], [ 932.0625 , 2.941023667670942e-46 ], [ 932.125 , 2.941358848470292e-46 ], [ 932.1875 , 2.9416940674693666e-46 ], 
                    [ 932.25 , 2.9420293246724965e-46 ], [ 932.3125 , 2.942364620084016e-46 ], [ 932.375 , 2.9426999537082374e-46 ], [ 932.4375 , 2.943035325549579e-46 ], 
                    [ 932.5 , 2.9433707356123747e-46 ], [ 932.5625 , 2.9437061839010013e-46 ], [ 932.625 , 2.944041670419753e-46 ], [ 932.6875 , 2.9443771951730708e-46 ], 
                    [ 932.75 , 2.9447127581652698e-46 ], [ 932.8125 , 2.9450483594006455e-46 ], [ 932.875 , 2.9453839988837023e-46 ], [ 932.9375 , 2.9457196766186745e-46 ], 
                    [ 933.0 , 2.9460553926099206e-46 ], [ 933.0625 , 2.9463911468618852e-46 ], [ 933.125 , 2.9467269393788865e-46 ], [ 933.1875 , 2.9470626110088517e-46 ], 
                    [ 933.25 , 2.947397949185764e-46 ], [ 933.3125 , 2.9477333255198684e-46 ], [ 933.375 , 2.9480687400155676e-46 ], [ 933.4375 , 2.9484041926771004e-46 ], 
                    [ 933.5 , 2.948739683508935e-46 ], [ 933.5625 , 2.949075212515331e-46 ], [ 933.625 , 2.9494107797006524e-46 ], [ 933.6875 , 2.9497463850692445e-46 ], 
                    [ 933.75 , 2.9500820286254514e-46 ], [ 933.8125 , 2.95041771037364e-46 ], [ 933.875 , 2.9507534303181136e-46 ], [ 933.9375 , 2.9510891884632394e-46 ], 
                    [ 934.0 , 2.9514249848133428e-46 ], [ 934.0625 , 2.9517608193727934e-46 ], [ 934.125 , 2.9520966921459586e-46 ], [ 934.1875 , 2.952432603137208e-46 ], 
                    [ 934.25 , 2.9527685523507435e-46 ], [ 934.3125 , 2.953104539791103e-46 ], [ 934.375 , 2.9534405654625316e-46 ], [ 934.4375 , 2.9537766293693998e-46 ], 
                    [ 934.5 , 2.9541127315160383e-46 ], [ 934.5625 , 2.954448871906903e-46 ], [ 934.625 , 2.9547850505461985e-46 ], [ 934.6875 , 2.9551212674382982e-46 ], 
                    [ 934.75 , 2.95545752258768e-46 ], [ 934.8125 , 2.9557938159985507e-46 ], [ 934.875 , 2.9561301476753686e-46 ], [ 934.9375 , 2.9564665176224045e-46 ], 
                    [ 935.0 , 2.9568029258440544e-46 ], [ 935.0625 , 2.9571393723446948e-46 ], [ 935.125 , 2.9574758571286395e-46 ], [ 935.1875 , 2.9578123802003076e-46 ], 
                    [ 935.25 , 2.958148941563993e-46 ], [ 935.3125 , 2.9584853027807752e-46 ], [ 935.375 , 2.9588214192169484e-46 ], [ 935.4375 , 2.9591575738395993e-46 ], 
                    [ 935.5 , 2.95949376665313e-46 ], [ 935.5625 , 2.959829997661879e-46 ], [ 935.625 , 2.960166266870123e-46 ], [ 935.6875 , 2.9605025742822855e-46 ], 
                    [ 935.75 , 2.9608389199026442e-46 ], [ 935.8125 , 2.9611753037355605e-46 ], [ 935.875 , 2.961511725785356e-46 ], [ 935.9375 , 2.9618481860564132e-46 ], 
                    [ 936.0 , 2.962184684553075e-46 ], [ 936.0625 , 2.9625212212796633e-46 ], [ 936.125 , 2.9628577962405006e-46 ], [ 936.1875 , 2.9631944094399515e-46 ], 
                    [ 936.25 , 2.9635310608823603e-46 ], [ 936.3125 , 2.9638677505720926e-46 ], [ 936.375 , 2.9642044785134097e-46 ], [ 936.4375 , 2.964541244710784e-46 ], 
                    [ 936.5 , 2.9648780491684986e-46 ], [ 936.5625 , 2.9652148918908367e-46 ], [ 936.625 , 2.9655517728822727e-46 ], [ 936.6875 , 2.965888692147112e-46 ], 
                    [ 936.75 , 2.966225649689597e-46 ], [ 936.8125 , 2.966562645514203e-46 ], [ 936.875 , 2.9668996796252584e-46 ], [ 936.9375 , 2.967236752027092e-46 ], 
                    [ 937.0 , 2.967573862724012e-46 ], [ 937.0625 , 2.9679110117204322e-46 ], [ 937.125 , 2.9682481990206825e-46 ], [ 937.1875 , 2.968585424629136e-46 ], 
                    [ 937.25 , 2.968922688550103e-46 ], [ 937.3125 , 2.9692599907879565e-46 ], [ 937.375 , 2.9695973313470506e-46 ], [ 937.4375 , 2.9699344230374017e-46 ], 
                    [ 937.5 , 2.970271306260997e-46 ], [ 937.5625 , 2.9706082276975974e-46 ], [ 937.625 , 2.9709451873516633e-46 ], [ 937.6875 , 2.971282185227446e-46 ], 
                    [ 937.75 , 2.9716192213292812e-46 ], [ 937.8125 , 2.9719562956615253e-46 ], [ 937.875 , 2.9722934082285156e-46 ], [ 937.9375 , 2.9726305590346508e-46 ], 
                    [ 938.0 , 2.972967748084123e-46 ], [ 938.0625 , 2.973304975381395e-46 ], [ 938.125 , 2.9736422409307434e-46 ], [ 938.1875 , 2.9739795447365487e-46 ], 
                    [ 938.25 , 2.9743168868031283e-46 ], [ 938.3125 , 2.9746542671348238e-46 ], [ 938.375 , 2.9749916857359747e-46 ], [ 938.4375 , 2.975329142610901e-46 ], 
                    [ 938.5 , 2.9756666377639654e-46 ], [ 938.5625 , 2.9760041711995526e-46 ], [ 938.625 , 2.976341742921898e-46 ], [ 938.6875 , 2.976679352935494e-46 ], 
                    [ 938.75 , 2.977017001244535e-46 ], [ 938.8125 , 2.9773546878534496e-46 ], [ 938.875 , 2.9776924127665822e-46 ], [ 938.9375 , 2.978030175988173e-46 ], 
                    [ 939.0 , 2.9783679775227354e-46 ], [ 939.0625 , 2.978705817374488e-46 ], [ 939.125 , 2.979043695547842e-46 ], [ 939.1875 , 2.979381612047101e-46 ], 
                    [ 939.25 , 2.9797195668766343e-46 ], [ 939.3125 , 2.9800575600408103e-46 ], [ 939.375 , 2.980395591543977e-46 ], [ 939.4375 , 2.9807336613904202e-46 ], 
                    [ 939.5 , 2.981071769584532e-46 ], [ 939.5625 , 2.981409638875025e-46 ], [ 939.625 , 2.9817473119196725e-46 ], [ 939.6875 , 2.982085023209031e-46 ], 
                    [ 939.75 , 2.9824227727473887e-46 ], [ 939.8125 , 2.9827605605391414e-46 ], [ 939.875 , 2.9830983865885585e-46 ], [ 939.9375 , 2.983436250899952e-46 ], 
                    [ 940.0 , 2.9837741534777606e-46 ], [ 940.0625 , 2.9841120943261917e-46 ], [ 940.125 , 2.984450073449686e-46 ], [ 940.1875 , 2.984788090852536e-46 ], 
                    [ 940.25 , 2.9851261465390554e-46 ], [ 940.3125 , 2.985464240513623e-46 ], [ 940.375 , 2.9858023727805545e-46 ], [ 940.4375 , 2.9861405433441655e-46 ], 
                    [ 940.5 , 2.986478752208835e-46 ], [ 940.5625 , 2.986816999378881e-46 ], [ 940.625 , 2.9871552848586202e-46 ], [ 940.6875 , 2.9874936086524126e-46 ], 
                    [ 940.75 , 2.9878319707646193e-46 ], [ 940.8125 , 2.9881703711995584e-46 ], [ 940.875 , 2.988508809961549e-46 ], [ 940.9375 , 2.988847287054954e-46 ], 
                    [ 941.0 , 2.9891858024841147e-46 ], [ 941.0625 , 2.989524356253415e-46 ], [ 941.125 , 2.989862948367112e-46 ], [ 941.1875 , 2.990201578829592e-46 ], 
                    [ 941.25 , 2.9905402476452402e-46 ], [ 941.3125 , 2.9908789548182735e-46 ], [ 941.375 , 2.9912177003532265e-46 ], [ 941.4375 , 2.991556484254254e-46 ], 
                    [ 941.5 , 2.991895306525807e-46 ], [ 941.5625 , 2.9922341671722316e-46 ], [ 941.625 , 2.992573066197852e-46 ], [ 941.6875 , 2.9929117509047726e-46 ], 
                    [ 941.75 , 2.9932501873875717e-46 ], [ 941.8125 , 2.9935886621405666e-46 ], [ 941.875 , 2.9939271751680195e-46 ], [ 941.9375 , 2.994265726474345e-46 ], 
                    [ 942.0 , 2.994604316063807e-46 ], [ 942.0625 , 2.9949429439407553e-46 ], [ 942.125 , 2.995281610109521e-46 ], [ 942.1875 , 2.9956203145744544e-46 ], 
                    [ 942.25 , 2.995959057339822e-46 ], [ 942.3125 , 2.9962978384100405e-46 ], [ 942.375 , 2.996636657789377e-46 ], [ 942.4375 , 2.9969755154821638e-46 ], 
                    [ 942.5 , 2.9973144114927968e-46 ], [ 942.5625 , 2.997653345825546e-46 ], [ 942.625 , 2.997992318484765e-46 ], [ 942.6875 , 2.998331329474789e-46 ], 
                    [ 942.75 , 2.9986703787999086e-46 ], [ 942.8125 , 2.9990094664646083e-46 ], [ 942.875 , 2.999348592473032e-46 ], [ 942.9375 , 2.9996877568296218e-46 ], 
                    [ 943.0 , 3.000026959538735e-46 ], [ 943.0625 , 3.0003662006045817e-46 ], [ 943.125 , 3.000705480031648e-46 ], [ 943.1875 , 3.001044797824165e-46 ], 
                    [ 943.25 , 3.001384153986556e-46 ], [ 943.3125 , 3.0017235485230967e-46 ], [ 943.375 , 3.0020629814381684e-46 ], [ 943.4375 , 3.0024024527360255e-46 ], 
                    [ 943.5 , 3.0027419624211374e-46 ], [ 943.5625 , 3.0030815104977368e-46 ], [ 943.625 , 3.0034210969702723e-46 ], [ 943.6875 , 3.003760721842958e-46 ], 
                    [ 943.75 , 3.004100385120242e-46 ], [ 943.8125 , 3.004439926238486e-46 ], [ 943.875 , 3.004779161459168e-46 ], [ 943.9375 , 3.0051184349833615e-46 ], 
                    [ 944.0 , 3.00545774681537e-46 ], [ 944.0625 , 3.005797096959498e-46 ], [ 944.125 , 3.006136485420092e-46 ], [ 944.1875 , 3.0064759122015e-46 ], 
                    [ 944.25 , 3.006815377308006e-46 ], [ 944.3125 , 3.0071548807440016e-46 ], [ 944.375 , 3.007494422513729e-46 ], [ 944.4375 , 3.0078340026215595e-46 ], 
                    [ 944.5 , 3.008173621071843e-46 ], [ 944.5625 , 3.008513277868866e-46 ], [ 944.625 , 3.008852973017002e-46 ], [ 944.6875 , 3.009192706520558e-46 ], 
                    [ 944.75 , 3.0095324783838238e-46 ], [ 944.8125 , 3.0098722886112153e-46 ], [ 944.875 , 3.010212137207065e-46 ], [ 944.9375 , 3.010552024175575e-46 ], 
                    [ 945.0 , 3.010891949521187e-46 ], [ 945.0625 , 3.0112319132482335e-46 ], [ 945.125 , 3.0115719153609834e-46 ], [ 945.1875 , 3.0119119558638786e-46 ], 
                    [ 945.25 , 3.0122520347611682e-46 ], [ 945.3125 , 3.0125921520572084e-46 ], [ 945.375 , 3.012932307756292e-46 ], [ 945.4375 , 3.013272501862841e-46 ], 
                    [ 945.5 , 3.0136127343811492e-46 ], [ 945.5625 , 3.0139530053155322e-46 ], [ 945.625 , 3.0142933146703487e-46 ], [ 945.6875 , 3.0146336624499805e-46 ], 
                    [ 945.75 , 3.014974048658637e-46 ], [ 945.8125 , 3.015314473300829e-46 ], [ 945.875 , 3.01565493638081e-46 ], [ 945.9375 , 3.015995391935193e-46 ], 
                    [ 946.0 , 3.016335422447755e-46 ], [ 946.0625 , 3.0166754912961464e-46 ], [ 946.125 , 3.017015598484733e-46 ], [ 946.1875 , 3.0173557440178146e-46 ], 
                    [ 946.25 , 3.017695927899693e-46 ], [ 946.3125 , 3.0180361501347352e-46 ], [ 946.375 , 3.018376410727222e-46 ], [ 946.4375 , 3.0187167096814992e-46 ], 
                    [ 946.5 , 3.0190570470018922e-46 ], [ 946.5625 , 3.0193974226927692e-46 ], [ 946.625 , 3.019737836758392e-46 ], [ 946.6875 , 3.0200782892030865e-46 ], 
                    [ 946.75 , 3.0204187800312016e-46 ], [ 946.8125 , 3.0207593092470645e-46 ], [ 946.875 , 3.021099876855025e-46 ], [ 946.9375 , 3.0214404828593683e-46 ], 
                    [ 947.0 , 3.021781127264487e-46 ], [ 947.0625 , 3.0221218100746047e-46 ], [ 947.125 , 3.022462531294157e-46 ], [ 947.1875 , 3.022803290927433e-46 ], 
                    [ 947.25 , 3.0231440889787406e-46 ], [ 947.3125 , 3.023484925452455e-46 ], [ 947.375 , 3.0238258003528424e-46 ], [ 947.4375 , 3.024166713684323e-46 ], 
                    [ 947.5 , 3.024507665451206e-46 ], [ 947.5625 , 3.024848655657762e-46 ], [ 947.625 , 3.0251896843083666e-46 ], [ 947.6875 , 3.0255307514073763e-46 ], 
                    [ 947.75 , 3.025871856959061e-46 ], [ 947.8125 , 3.026213000967885e-46 ], [ 947.875 , 3.0265541834380334e-46 ], [ 947.9375 , 3.0268954043738858e-46 ], 
                    [ 948.0 , 3.0272366637798647e-46 ], [ 948.0625 , 3.0275779616602e-46 ], [ 948.125 , 3.0279189225413715e-46 ], [ 948.1875 , 3.028259818849613e-46 ], 
                    [ 948.25 , 3.028600753537425e-46 ], [ 948.3125 , 3.0289417266092147e-46 ], [ 948.375 , 3.029282738069217e-46 ], [ 948.4375 , 3.0296237879217545e-46 ], 
                    [ 948.5 , 3.029964876171192e-46 ], [ 948.5625 , 3.030306002821809e-46 ], [ 948.625 , 3.0306471678779523e-46 ], [ 948.6875 , 3.0309883713439867e-46 ], 
                    [ 948.75 , 3.031329613224151e-46 ], [ 948.8125 , 3.0316708935228136e-46 ], [ 948.875 , 3.0320122122443214e-46 ], [ 948.9375 , 3.0323535693929563e-46 ], 
                    [ 949.0 , 3.0326949649730885e-46 ], [ 949.0625 , 3.033036398989023e-46 ], [ 949.125 , 3.0333778714450656e-46 ], [ 949.1875 , 3.0337193823455865e-46 ], 
                    [ 949.25 , 3.0340609316948934e-46 ], [ 949.3125 , 3.034402519497293e-46 ], [ 949.375 , 3.0347441457571354e-46 ], [ 949.4375 , 3.0350858104787734e-46 ], 
                    [ 949.5 , 3.035427513666514e-46 ], [ 949.5625 , 3.0357692553246675e-46 ], [ 949.625 , 3.0361110354575865e-46 ], [ 949.6875 , 3.0364528540695803e-46 ], 
                    [ 949.75 , 3.0367947111651114e-46 ], [ 949.8125 , 3.037136606748274e-46 ], [ 949.875 , 3.0374785408235752e-46 ], [ 949.9375 , 3.0378205133953264e-46 ], 
                    [ 950.0 , 3.0381625244678396e-46 ], [ 950.0625 , 3.0385045740454287e-46 ], [ 950.125 , 3.038846662132471e-46 ], [ 950.1875 , 3.03918878873326e-46 ], 
                    [ 950.25 , 3.0395307971109155e-46 ], [ 950.3125 , 3.0398726103987877e-46 ], [ 950.375 , 3.040214462125529e-46 ], [ 950.4375 , 3.0405563522955707e-46 ], 
                    [ 950.5 , 3.0408982809131918e-46 ], [ 950.5625 , 3.0412402479826734e-46 ], [ 950.625 , 3.0415822535084692e-46 ], [ 950.6875 , 3.0419242974947304e-46 ], 
                    [ 950.75 , 3.042266379945935e-46 ], [ 950.8125 , 3.042608500866298e-46 ], [ 950.875 , 3.042950660260169e-46 ], [ 950.9375 , 3.043292858131961e-46 ], 
                    [ 951.0 , 3.0436350944859137e-46 ], [ 951.0625 , 3.0439773693263556e-46 ], [ 951.125 , 3.0443196826576786e-46 ], [ 951.1875 , 3.0446620344841257e-46 ], 
                    [ 951.25 , 3.045004424810089e-46 ], [ 951.3125 , 3.045346853639836e-46 ], [ 951.375 , 3.0456893209778028e-46 ], [ 951.4375 , 3.0460318268282333e-46 ], 
                    [ 951.5 , 3.0463743711954806e-46 ], [ 951.5625 , 3.0467169540838766e-46 ], [ 951.625 , 3.0470595754976875e-46 ], [ 951.6875 , 3.04740223544142e-46 ], 
                    [ 951.75 , 3.0477449339192106e-46 ], [ 951.8125 , 3.048087670935545e-46 ], [ 951.875 , 3.0484304464946926e-46 ], [ 951.9375 , 3.048773260600922e-46 ], 
                    [ 952.0 , 3.0491161132586985e-46 ], [ 952.0625 , 3.0494590044722275e-46 ], [ 952.125 , 3.0498019342459527e-46 ], [ 952.1875 , 3.050144902584146e-46 ], 
                    [ 952.25 , 3.0504879094911874e-46 ], [ 952.3125 , 3.050830954971392e-46 ], [ 952.375 , 3.051174039029098e-46 ], [ 952.4375 , 3.0515170109267553e-46 ], 
                    [ 952.5 , 3.051860011835656e-46 ], [ 952.5625 , 3.0522030512990275e-46 ], [ 952.625 , 3.052546129321204e-46 ], [ 952.6875 , 3.0528892459064974e-46 ], 
                    [ 952.75 , 3.053232401059307e-46 ], [ 952.8125 , 3.0535755947839034e-46 ], [ 952.875 , 3.0539188270846437e-46 ], [ 952.9375 , 3.0542620979658423e-46 ], 
                    [ 953.0 , 3.0546054074318795e-46 ], [ 953.0625 , 3.0549487554870923e-46 ], [ 953.125 , 3.0552921421357738e-46 ], [ 953.1875 , 3.0556355673822634e-46 ], 
                    [ 953.25 , 3.0559790312309642e-46 ], [ 953.3125 , 3.05632253368615e-46 ], [ 953.375 , 3.056666074752182e-46 ], [ 953.4375 , 3.0570096544334003e-46 ], 
                    [ 953.5 , 3.0573532727341674e-46 ], [ 953.5625 , 3.057696929658737e-46 ], [ 953.625 , 3.0580406252115594e-46 ], [ 953.6875 , 3.058384359396954e-46 ], 
                    [ 953.75 , 3.0587281322192207e-46 ], [ 953.8125 , 3.0590719436827025e-46 ], [ 953.875 , 3.0594157937917863e-46 ], [ 953.9375 , 3.059759682550793e-46 ], 
                    [ 954.0 , 3.0601036099641117e-46 ], [ 954.0625 , 3.0604475760359566e-46 ], [ 954.125 , 3.0607915807707813e-46 ], [ 954.1875 , 3.061135624172976e-46 ], 
                    [ 954.25 , 3.0614797062467995e-46 ], [ 954.3125 , 3.0618238269965765e-46 ], [ 954.375 , 3.0621679864266986e-46 ], [ 954.4375 , 3.062512184541557e-46 ], 
                    [ 954.5 , 3.0628564213454124e-46 ], [ 954.5625 , 3.0632010377859145e-46 ], [ 954.625 , 3.0635462020930733e-46 ], [ 954.6875 , 3.0638914052936202e-46 ], 
                    [ 954.75 , 3.064236647392024e-46 ], [ 954.8125 , 3.064581928392603e-46 ], [ 954.875 , 3.0649272482997623e-46 ], [ 954.9375 , 3.0652726071179077e-46 ], 
                    [ 955.0 , 3.0656180048514018e-46 ], [ 955.0625 , 3.0659634415046085e-46 ], [ 955.125 , 3.0663089170819345e-46 ], [ 955.1875 , 3.066654431587766e-46 ], 
                    [ 955.25 , 3.0669999850264894e-46 ], [ 955.3125 , 3.0673455774025134e-46 ], [ 955.375 , 3.0676912087201604e-46 ], [ 955.4375 , 3.068036878983927e-46 ], 
                    [ 955.5 , 3.068382588198071e-46 ], [ 955.5625 , 3.068728336367091e-46 ], [ 955.625 , 3.06907412349531e-46 ], [ 955.6875 , 3.0694199495871623e-46 ], 
                    [ 955.75 , 3.069765814646995e-46 ], [ 955.8125 , 3.0701117186792196e-46 ], [ 955.875 , 3.070457661688272e-46 ], [ 955.9375 , 3.0708036436784354e-46 ], 
                    [ 956.0 , 3.071149664654211e-46 ], [ 956.0625 , 3.0714957246199266e-46 ], [ 956.125 , 3.071841823580018e-46 ], [ 956.1875 , 3.0721879615388364e-46 ], 
                    [ 956.25 , 3.0725341385007987e-46 ], [ 956.3125 , 3.0728803544702768e-46 ], [ 956.375 , 3.073226609451754e-46 ], [ 956.4375 , 3.0735729034494956e-46 ], 
                    [ 956.5 , 3.0739192364679842e-46 ], [ 956.5625 , 3.074265608511661e-46 ], [ 956.625 , 3.074612019584771e-46 ], [ 956.6875 , 3.0749584696918206e-46 ], 
                    [ 956.75 , 3.0753131665241756e-46 ], [ 956.8125 , 3.07567017011235e-46 ], [ 956.875 , 3.0760272151439882e-46 ], [ 956.9375 , 3.0763843016239015e-46 ], 
                    [ 957.0 , 3.0767414295568126e-46 ], [ 957.0625 , 3.0770985989476445e-46 ], [ 957.125 , 3.077455809801187e-46 ], [ 957.1875 , 3.0778130621222107e-46 ], 
                    [ 957.25 , 3.078170355915507e-46 ], [ 957.3125 , 3.078527691185977e-46 ], [ 957.375 , 3.078885067938372e-46 ], [ 957.4375 , 3.079242486177571e-46 ], 
                    [ 957.5 , 3.079599945908304e-46 ], [ 957.5625 , 3.0799574471354745e-46 ], [ 957.625 , 3.0803149898638125e-46 ], [ 957.6875 , 3.080672574098201e-46 ], 
                    [ 957.75 , 3.0810301998434367e-46 ], [ 957.8125 , 3.0813878671042948e-46 ], [ 957.875 , 3.0817455758857035e-46 ], [ 957.9375 , 3.082103326192396e-46 ], 
                    [ 958.0 , 3.0824611180292365e-46 ], [ 958.0625 , 3.08281895140098e-46 ], [ 958.125 , 3.0831768263124487e-46 ], [ 958.1875 , 3.083534742768596e-46 ], 
                    [ 958.25 , 3.083892700774135e-46 ], [ 958.3125 , 3.0842507003339325e-46 ], [ 958.375 , 3.084608741452726e-46 ], [ 958.4375 , 3.0849668241354486e-46 ], 
                    [ 958.5 , 3.0853249483868816e-46 ], [ 958.5625 , 3.0856831142118285e-46 ], [ 958.625 , 3.0860413216151815e-46 ], [ 958.6875 , 3.0863995706016794e-46 ], 
                    [ 958.75 , 3.0867578611761718e-46 ], [ 958.8125 , 3.087116193343574e-46 ], [ 958.875 , 3.0875451239954157e-46 ], [ 958.9375 , 3.0887084640745024e-46 ], 
                    [ 959.0 , 3.089872242482472e-46 ], [ 959.0625 , 3.0910364593844154e-46 ], [ 959.125 , 3.092201114945638e-46 ], [ 959.1875 , 3.0933662093312876e-46 ], 
                    [ 959.25 , 3.094531742706883e-46 ], [ 959.3125 , 3.095697715237719e-46 ], [ 959.375 , 3.096864127089285e-46 ], [ 959.4375 , 3.0980309784271092e-46 ], 
                    [ 959.5 , 3.0991982694167402e-46 ], [ 959.5625 , 3.1003660002239423e-46 ], [ 959.625 , 3.1015341710143655e-46 ], [ 959.6875 , 3.1027027819537446e-46 ], 
                    [ 959.75 , 3.1038718332079877e-46 ], [ 959.8125 , 3.1050413249429765e-46 ], [ 959.875 , 3.1062112573246552e-46 ], [ 959.9375 , 3.1073816305190974e-46 ], 
                    [ 960.0 , 3.1085524446923723e-46 ], [ 960.0625 , 3.109723700010612e-46 ], [ 960.125 , 3.110895396640121e-46 ], [ 960.1875 , 3.1120675347470466e-46 ], 
                    [ 960.25 , 3.113240114497775e-46 ], [ 960.3125 , 3.1144131360587776e-46 ], [ 960.375 , 3.115586599596499e-46 ], [ 960.4375 , 3.116760505277403e-46 ], 
                    [ 960.5 , 3.117934853268149e-46 ], [ 960.5625 , 3.1191096437353707e-46 ], [ 960.625 , 3.1202848768457865e-46 ], [ 960.6875 , 3.1214605527661772e-46 ], 
                    [ 960.75 , 3.1226366716634094e-46 ], [ 960.8125 , 3.123813233704301e-46 ], [ 960.875 , 3.124990239055911e-46 ], [ 960.9375 , 3.1261676878853165e-46 ], 
                    [ 961.0 , 3.127345580359479e-46 ], [ 961.0625 , 3.1280030827412073e-46 ], [ 961.125 , 3.1275174022655453e-46 ], [ 961.1875 , 3.127031797200805e-46 ], 
                    [ 961.25 , 3.126546267535232e-46 ], [ 961.3125 , 3.126060813257142e-46 ], [ 961.375 , 3.1255754343547855e-46 ], [ 961.4375 , 3.1250901308165693e-46 ], 
                    [ 961.5 , 3.1246049026307036e-46 ], [ 961.5625 , 3.1241197497855105e-46 ], [ 961.625 , 3.12363467226927e-46 ], [ 961.6875 , 3.1231496700703293e-46 ], 
                    [ 961.75 , 3.122664743176951e-46 ], [ 961.8125 , 3.12217989157753e-46 ], [ 961.875 , 3.1216951152601994e-46 ], [ 961.9375 , 3.1212104142134687e-46 ], 
                    [ 962.0 , 3.1207257884255403e-46 ], [ 962.0625 , 3.120241237884729e-46 ], [ 962.125 , 3.1197567625793953e-46 ], [ 962.1875 , 3.119272362497858e-46 ], 
                    [ 962.25 , 3.1187880376283703e-46 ], [ 962.3125 , 3.118303787959299e-46 ], [ 962.375 , 3.1178196134789886e-46 ], [ 962.4375 , 3.117335514175766e-46 ], 
                    [ 962.5 , 3.1168514900379142e-46 ], [ 962.5625 , 3.1163675410537175e-46 ], [ 962.625 , 3.115883667211662e-46 ], [ 962.6875 , 3.1153998684999927e-46 ], 
                    [ 962.75 , 3.114916144907021e-46 ], [ 962.8125 , 3.114432496421084e-46 ], [ 962.875 , 3.1139489230305198e-46 ], [ 962.9375 , 3.113465424723779e-46 ], 
                    [ 963.0 , 3.112982001489093e-46 ], [ 963.0625 , 3.112498653314783e-46 ], [ 963.125 , 3.1120153801892386e-46 ], [ 963.1875 , 3.111532182100808e-46 ], 
                    [ 963.25 , 3.111426799820254e-46 ], [ 963.3125 , 3.111765757363003e-46 ], [ 963.375 , 3.1121047518316686e-46 ], [ 963.4375 , 3.1124437832302504e-46 ], 
                    [ 963.5 , 3.1127828515628164e-46 ], [ 963.5625 , 3.1131219568332568e-46 ], [ 963.625 , 3.1134610990457515e-46 ], [ 963.6875 , 3.113800278204235e-46 ], 
                    [ 963.75 , 3.114139494312755e-46 ], [ 963.8125 , 3.11447874737536e-46 ], [ 963.875 , 3.1148180373960076e-46 ], [ 963.9375 , 3.1151573643787917e-46 ], 
                    [ 964.0 , 3.115496728327694e-46 ], [ 964.0625 , 3.115836129246808e-46 ], [ 964.125 , 3.116175567140095e-46 ], [ 964.1875 , 3.1165150420116274e-46 ], 
                    [ 964.25 , 3.1168545538653667e-46 ], [ 964.3125 , 3.117194102705409e-46 ], [ 964.375 , 3.1175336885357602e-46 ], [ 964.4375 , 3.117873311360451e-46 ], 
                    [ 964.5 , 3.1182129711835335e-46 ], [ 964.5625 , 3.1185526680090156e-46 ], [ 964.625 , 3.1188924018408405e-46 ], [ 964.6875 , 3.119232172683238e-46 ], 
                    [ 964.75 , 3.1195719805400638e-46 ], [ 964.8125 , 3.119911825415417e-46 ], [ 964.875 , 3.1202517073133955e-46 ], [ 964.9375 , 3.1205916262379e-46 ], 
                    [ 965.0 , 3.1209315821930533e-46 ], [ 965.0625 , 3.1212715751828453e-46 ], [ 965.125 , 3.121611605211354e-46 ], [ 965.1875 , 3.1219516722826153e-46 ], 
                    [ 965.25 , 3.12229177640062e-46 ], [ 965.3125 , 3.1226319175694255e-46 ], [ 965.375 , 3.1229720957930912e-46 ], [ 965.4375 , 3.1233150605256506e-46 ], 
                    [ 965.5 , 3.1236605305223947e-46 ], [ 965.5625 , 3.1240060387316333e-46 ], [ 965.625 , 3.1243515851575032e-46 ], [ 965.6875 , 3.1246971698042544e-46 ], 
                    [ 965.75 , 3.1250427926761807e-46 ], [ 965.8125 , 3.1253884537774655e-46 ], [ 965.875 , 3.125734153112338e-46 ], [ 965.9375 , 3.126079890685005e-46 ], 
                    [ 966.0 , 3.126425666499718e-46 ], [ 966.0625 , 3.1267714805607067e-46 ], [ 966.125 , 3.127117332872246e-46 ], [ 966.1875 , 3.1274632234384564e-46 ], 
                    [ 966.25 , 3.1278091522636574e-46 ], [ 966.3125 , 3.1281551193520806e-46 ], [ 966.375 , 3.128501124707937e-46 ], [ 966.4375 , 3.1288471683354366e-46 ], 
                    [ 966.5 , 3.129193250238857e-46 ], [ 966.5625 , 3.1295393704224097e-46 ], [ 966.625 , 3.1298855288902846e-46 ], [ 966.6875 , 3.130231725646805e-46 ], 
                    [ 966.75 , 3.1305779606961177e-46 ], [ 966.8125 , 3.1309242340425687e-46 ], [ 966.875 , 3.1312705456903054e-46 ], [ 966.9375 , 3.1316168956435645e-46 ], 
                    [ 967.0 , 3.131963283906605e-46 ], [ 967.0625 , 3.132309710483664e-46 ], [ 967.125 , 3.132656175379002e-46 ], [ 967.1875 , 3.1330026785968352e-46 ], 
                    [ 967.25 , 3.1333492201413806e-46 ], [ 967.3125 , 3.133695800016899e-46 ], [ 967.375 , 3.1340424182276527e-46 ], [ 967.4375 , 3.13438907477786e-46 ], 
                    [ 967.5 , 3.1347357696716944e-46 ], [ 967.5625 , 3.135082502913553e-46 ], [ 967.625 , 3.1354294805627113e-46 ], [ 967.6875 , 3.135776690407669e-46 ], 
                    [ 967.75 , 3.136123938701824e-46 ], [ 967.8125 , 3.136471225449434e-46 ], [ 967.875 , 3.136818550654735e-46 ], [ 967.9375 , 3.137165914322008e-46 ], 
                    [ 968.0 , 3.137513316455422e-46 ], [ 968.0625 , 3.137860757059372e-46 ], [ 968.125 , 3.138208236138006e-46 ], [ 968.1875 , 3.138555753695696e-46 ], 
                    [ 968.25 , 3.138903309736525e-46 ], [ 968.3125 , 3.13925090426491e-46 ], [ 968.375 , 3.1395985372851135e-46 ], [ 968.4375 , 3.1399462088013313e-46 ], 
                    [ 968.5 , 3.1402939188178702e-46 ], [ 968.5625 , 3.1406416673389276e-46 ], [ 968.625 , 3.1409894543688335e-46 ], [ 968.6875 , 3.1413372799118305e-46 ], 
                    [ 968.75 , 3.1416851439721832e-46 ], [ 968.8125 , 3.1420330465541125e-46 ], [ 968.875 , 3.1423809876619733e-46 ], [ 968.9375 , 3.142728967299987e-46 ], 
                    [ 969.0 , 3.143076985472421e-46 ], [ 969.0625 , 3.143425042183542e-46 ], [ 969.125 , 3.143773137437618e-46 ], [ 969.1875 , 3.1441212712389166e-46 ], 
                    [ 969.25 , 3.144469443591707e-46 ], [ 969.3125 , 3.144817654500258e-46 ], [ 969.375 , 3.1451659039688166e-46 ], [ 969.4375 , 3.145514192001743e-46 ], 
                    [ 969.5 , 3.1458625186031505e-46 ], [ 969.5625 , 3.146210883777422e-46 ], [ 969.625 , 3.1465592875288293e-46 ], [ 969.6875 , 3.1469077298616216e-46 ], 
                    [ 969.75 , 3.147256210780094e-46 ], [ 969.8125 , 3.1476046189079345e-46 ], [ 969.875 , 3.1479529144773438e-46 ], [ 969.9375 , 3.148301248587109e-46 ], 
                    [ 970.0 , 3.1486496212415177e-46 ], [ 970.0625 , 3.1489980324447905e-46 ], [ 970.125 , 3.1493464822012145e-46 ], [ 970.1875 , 3.149694970515079e-46 ], 
                    [ 970.25 , 3.1500434973906055e-46 ], [ 970.3125 , 3.1503920628320828e-46 ], [ 970.375 , 3.1507406668438025e-46 ], [ 970.4375 , 3.1510893094299855e-46 ], 
                    [ 970.5 , 3.151437990594924e-46 ], [ 970.5625 , 3.1517867103429092e-46 ], [ 970.625 , 3.152135468678143e-46 ], [ 970.6875 , 3.152484265604985e-46 ], 
                    [ 970.75 , 3.1528331011276605e-46 ], [ 970.8125 , 3.1531819752503956e-46 ], [ 970.875 , 3.153530887977551e-46 ], [ 970.9375 , 3.153879839313332e-46 ], 
                    [ 971.0 , 3.154228829262032e-46 ], [ 971.0625 , 3.154577857827947e-46 ], [ 971.125 , 3.154926925015305e-46 ], [ 971.1875 , 3.155276030828402e-46 ], 
                    [ 971.25 , 3.155625175271535e-46 ], [ 971.3125 , 3.155974358348911e-46 ], [ 971.375 , 3.1563235800648942e-46 ], [ 971.4375 , 3.1566728404236936e-46 ], 
                    [ 971.5 , 3.157022139429607e-46 ], [ 971.5625 , 3.157371477086934e-46 ], [ 971.625 , 3.1577208533999288e-46 ], [ 971.6875 , 3.1580702683728234e-46 ], 
                    [ 971.75 , 3.158419722009986e-46 ], [ 971.8125 , 3.1587692143155825e-46 ], [ 971.875 , 3.159118745294049e-46 ], [ 971.9375 , 3.159468314949553e-46 ], 
                    [ 972.0 , 3.1598178071964013e-46 ], [ 972.0625 , 3.1601669969644354e-46 ], [ 972.125 , 3.1605162253211986e-46 ], [ 972.1875 , 3.1608654922710218e-46 ], 
                    [ 972.25 , 3.1612147978181478e-46 ], [ 972.3125 , 3.1615641419668647e-46 ], [ 972.375 , 3.161913524721371e-46 ], [ 972.4375 , 3.1622629460860002e-46 ], 
                    [ 972.5 , 3.162612406064997e-46 ], [ 972.5625 , 3.1629619046626502e-46 ], [ 972.625 , 3.163311441883161e-46 ], [ 972.6875 , 3.163661017730887e-46 ], 
                    [ 972.75 , 3.164010632209985e-46 ], [ 972.8125 , 3.1643602853248582e-46 ], [ 972.875 , 3.1647099770797092e-46 ], [ 972.9375 , 3.165059707478786e-46 ], 
                    [ 973.0 , 3.1654094765263814e-46 ], [ 973.0625 , 3.1657592842267437e-46 ], [ 973.125 , 3.166109130584189e-46 ], [ 973.1875 , 3.1664590156029452e-46 ], 
                    [ 973.25 , 3.1668089392873746e-46 ], [ 973.3125 , 3.1671589016416143e-46 ], [ 973.375 , 3.1675089026700512e-46 ], [ 973.4375 , 3.1678589423768457e-46 ], 
                    [ 973.5 , 3.1682090207663855e-46 ], [ 973.5625 , 3.1685591378429e-46 ], [ 973.625 , 3.1689092936106644e-46 ], [ 973.6875 , 3.1692594880739542e-46 ], 
                    [ 973.75 , 3.1696097212370463e-46 ], [ 973.8125 , 3.1699599931042623e-46 ], [ 973.875 , 3.170310303679744e-46 ], [ 973.9375 , 3.1706606529679263e-46 ], 
                    [ 974.0 , 3.1710110409730206e-46 ], [ 974.0625 , 3.171361467699237e-46 ], [ 974.125 , 3.171711933150968e-46 ], [ 974.1875 , 3.1720624355458375e-46 ], 
                    [ 974.25 , 3.1724124708363677e-46 ], [ 974.3125 , 3.172762544753138e-46 ], [ 974.375 , 3.173112657300344e-46 ], [ 974.4375 , 3.1734628084822714e-46 ], 
                    [ 974.5 , 3.173812998303205e-46 ], [ 974.5625 , 3.1741632267673197e-46 ], [ 974.625 , 3.1745134938790144e-46 ], [ 974.6875 , 3.174863799642486e-46 ], 
                    [ 974.75 , 3.175214144061977e-46 ], [ 974.8125 , 3.1755645271417763e-46 ], [ 974.875 , 3.1759149488861498e-46 ], [ 974.9375 , 3.1762654092993866e-46 ], 
                    [ 975.0 , 3.1766159083857085e-46 ], [ 975.0625 , 3.1769664461494056e-46 ], [ 975.125 , 3.177317022594746e-46 ], [ 975.1875 , 3.1776676377259994e-46 ], 
                    [ 975.25 , 3.1780182915474325e-46 ], [ 975.3125 , 3.1783689840633393e-46 ], [ 975.375 , 3.1787197152779432e-46 ], [ 975.4375 , 3.1790704851955375e-46 ], 
                    [ 975.5 , 3.1794212938204166e-46 ], [ 975.5625 , 3.1797721411567602e-46 ], [ 975.625 , 3.1801230272089987e-46 ], [ 975.6875 , 3.180473951981269e-46 ], 
                    [ 975.75 , 3.180824915477866e-46 ], [ 975.8125 , 3.1811759177031537e-46 ], [ 975.875 , 3.1815269586613154e-46 ], [ 975.9375 , 3.1818780383566025e-46 ], 
                    [ 976.0 , 3.182229156793357e-46 ], [ 976.0625 , 3.1825803139757878e-46 ], [ 976.125 , 3.182931509908304e-46 ], [ 976.1875 , 3.1832827445950244e-46 ], 
                    [ 976.25 , 3.183634018040316e-46 ], [ 976.3125 , 3.1839853302484783e-46 ], [ 976.375 , 3.1843366812237206e-46 ], [ 976.4375 , 3.1846877030117865e-46 ], 
                    [ 976.5 , 3.18503858504671e-46 ], [ 976.5625 , 3.1853895057411067e-46 ], [ 976.625 , 3.1857404650991672e-46 ], [ 976.6875 , 3.1860914631251965e-46 ], 
                    [ 976.75 , 3.1864424998234102e-46 ], [ 976.8125 , 3.186793575198092e-46 ], [ 976.875 , 3.1871446892534796e-46 ], [ 976.9375 , 3.1874958419939257e-46 ], 
                    [ 977.0 , 3.1878470334235576e-46 ], [ 977.0625 , 3.1881982635467725e-46 ], [ 977.125 , 3.1885495323676985e-46 ], [ 977.1875 , 3.188900839890667e-46 ], 
                    [ 977.25 , 3.189252186119987e-46 ], [ 977.3125 , 3.189603571059879e-46 ], [ 977.375 , 3.1899549947145836e-46 ], [ 977.4375 , 3.19030645708839e-46 ], 
                    [ 977.5 , 3.1906579581855642e-46 ], [ 977.5625 , 3.191009498010373e-46 ], [ 977.625 , 3.191361076567128e-46 ], [ 977.6875 , 3.191712693860006e-46 ], 
                    [ 977.75 , 3.192064349893298e-46 ], [ 977.8125 , 3.1924160446713396e-46 ], [ 977.875 , 3.192767778198287e-46 ], [ 977.9375 , 3.1931195504785676e-46 ], 
                    [ 978.0 , 3.1934713615163152e-46 ], [ 978.0625 , 3.1938232113158688e-46 ], [ 978.125 , 3.19417509988143e-46 ], [ 978.1875 , 3.1945270272173386e-46 ], 
                    [ 978.25 , 3.194878993327889e-46 ], [ 978.3125 , 3.195230998217217e-46 ], [ 978.375 , 3.1955830418897765e-46 ], [ 978.4375 , 3.1959351243497044e-46 ], 
                    [ 978.5 , 3.1962872456013425e-46 ], [ 978.5625 , 3.196639405648874e-46 ], [ 978.625 , 3.1969914596329956e-46 ], [ 978.6875 , 3.1973431468146466e-46 ], 
                    [ 978.75 , 3.1976948726839263e-46 ], [ 978.8125 , 3.198046637245023e-46 ], [ 978.875 , 3.1983984405021924e-46 ], [ 978.9375 , 3.1987502824597366e-46 ], 
                    [ 979.0 , 3.199102163121868e-46 ], [ 979.0625 , 3.1994540824928893e-46 ], [ 979.125 , 3.1998060405770594e-46 ], [ 979.1875 , 3.2001580373785674e-46 ], 
                    [ 979.25 , 3.200510072901742e-46 ], [ 979.3125 , 3.2008621471508194e-46 ], [ 979.375 , 3.2012142601300376e-46 ], [ 979.4375 , 3.2015664118437015e-46 ], 
                    [ 979.5 , 3.2019186022960497e-46 ], [ 979.5625 , 3.2022708314913446e-46 ], [ 979.625 , 3.20262309943387e-46 ], [ 979.6875 , 3.2029754061278205e-46 ], 
                    [ 979.75 , 3.2033277515775267e-46 ], [ 979.8125 , 3.2036801357872294e-46 ], [ 979.875 , 3.2040325587612157e-46 ], [ 979.9375 , 3.2043850205036583e-46 ], 
                    [ 980.0 , 3.204737521018913e-46 ], [ 980.0625 , 3.2050900603112683e-46 ], [ 980.125 , 3.2054426383849664e-46 ], [ 980.1875 , 3.2057952552441827e-46 ], 
                    [ 980.25 , 3.2061479108932296e-46 ], [ 980.3125 , 3.2065006053364654e-46 ], [ 980.375 , 3.2068533385780432e-46 ], [ 980.4375 , 3.2072061106222998e-46 ], 
                    [ 980.5 , 3.207558921473458e-46 ], [ 980.5625 , 3.2079117711358783e-46 ], [ 980.625 , 3.2082646596136927e-46 ], [ 980.6875 , 3.208617586911263e-46 ], 
                    [ 980.75 , 3.2089705530328366e-46 ], [ 980.8125 , 3.209323557982661e-46 ], [ 980.875 , 3.2096761667925933e-46 ], [ 980.9375 , 3.210028686333003e-46 ], 
                    [ 981.0 , 3.2103812445907e-46 ], [ 981.0625 , 3.2107338415700043e-46 ], [ 981.125 , 3.2110864772750783e-46 ], [ 981.1875 , 3.211439151710243e-46 ], 
                    [ 981.25 , 3.2117918648797298e-46 ], [ 981.3125 , 3.2121446167877694e-46 ], [ 981.375 , 3.21249740743864e-46 ], [ 981.4375 , 3.2128502368366196e-46 ], 
                    [ 981.5 , 3.2132031049859403e-46 ], [ 981.5625 , 3.213556011890836e-46 ], [ 981.625 , 3.2139089575556312e-46 ], [ 981.6875 , 3.214261941984492e-46 ], 
                    [ 981.75 , 3.2146149651817443e-46 ], [ 981.8125 , 3.2149680271516232e-46 ], [ 981.875 , 3.215321127898364e-46 ], [ 981.9375 , 3.215674267426271e-46 ], 
                    [ 982.0 , 3.216027445739559e-46 ], [ 982.0625 , 3.2163806628425324e-46 ], [ 982.125 , 3.2167339187394293e-46 ], [ 982.1875 , 3.217087213434487e-46 ], 
                    [ 982.25 , 3.21744054693199e-46 ], [ 982.3125 , 3.2177939192362685e-46 ], [ 982.375 , 3.218147330351425e-46 ], [ 982.4375 , 3.218500780281881e-46 ], 
                    [ 982.5 , 3.2188542690317635e-46 ], [ 982.5625 , 3.219207796605427e-46 ], [ 982.625 , 3.2195613630071586e-46 ], [ 982.6875 , 3.2199149682411085e-46 ], 
                    [ 982.75 , 3.2202686123117015e-46 ], [ 982.8125 , 3.220622295223043e-46 ], [ 982.875 , 3.2209760169794448e-46 ], [ 982.9375 , 3.221329777585288e-46 ], 
                    [ 983.0 , 3.221683577044678e-46 ], [ 983.0625 , 3.22203729830831e-46 ], [ 983.125 , 3.2223905952484553e-46 ], [ 983.1875 , 3.222743930927648e-46 ], 
                    [ 983.25 , 3.2230973053501346e-46 ], [ 983.3125 , 3.223450718520256e-46 ], [ 983.375 , 3.223804170442146e-46 ], [ 983.4375 , 3.2241576611201225e-46 ], 
                    [ 983.5 , 3.2245111905584125e-46 ], [ 983.5625 , 3.2248647587612198e-46 ], [ 983.625 , 3.225218365732864e-46 ], [ 983.6875 , 3.2255720114775957e-46 ], 
                    [ 983.75 , 3.2259256959996433e-46 ], [ 983.8125 , 3.2262794193032144e-46 ], [ 983.875 , 3.226633181392651e-46 ], [ 983.9375 , 3.226986982272185e-46 ], 
                    [ 984.0 , 3.2273408219459763e-46 ], [ 984.0625 , 3.2276947004183946e-46 ], [ 984.125 , 3.2280486176936016e-46 ], [ 984.1875 , 3.2284025737759445e-46 ], 
                    [ 984.25 , 3.228756568669632e-46 ], [ 984.3125 , 3.2291106023788976e-46 ], [ 984.375 , 3.229464674908019e-46 ], [ 984.4375 , 3.2298187862612546e-46 ], 
                    [ 984.5 , 3.230172936442906e-46 ], [ 984.5625 , 3.2305271254571626e-46 ], [ 984.625 , 3.2308813533082595e-46 ], [ 984.6875 , 3.231235620000523e-46 ], 
                    [ 984.75 , 3.2315899255381904e-46 ], [ 984.8125 , 3.231944269925543e-46 ], [ 984.875 , 3.2322986531668185e-46 ], [ 984.9375 , 3.2326530752662538e-46 ], 
                    [ 985.0 , 3.233007536228156e-46 ], [ 985.0625 , 3.233362036056717e-46 ], [ 985.125 , 3.2337165747563142e-46 ], [ 985.1875 , 3.234071152331117e-46 ], 
                    [ 985.25 , 3.2344257687854347e-46 ], [ 985.3125 , 3.234780112643406e-46 ], [ 985.375 , 3.2351342405283395e-46 ], [ 985.4375 , 3.2354884071814824e-46 ], 
                    [ 985.5 , 3.235842612607056e-46 ], [ 985.5625 , 3.236196856809282e-46 ], [ 985.625 , 3.2365511397924283e-46 ], [ 985.6875 , 3.236905461560718e-46 ], 
                    [ 985.75 , 3.2372598221184885e-46 ], [ 985.8125 , 3.237614221469848e-46 ], [ 985.875 , 3.237968659619159e-46 ], [ 985.9375 , 3.2383231365706228e-46 ], 
                    [ 986.0 , 3.2386776523284644e-46 ], [ 986.0625 , 3.2390322068969552e-46 ], [ 986.125 , 3.2393868002803662e-46 ], [ 986.1875 , 3.2397414324829018e-46 ], 
                    [ 986.25 , 3.240096103508834e-46 ], [ 986.3125 , 3.240450813362436e-46 ], [ 986.375 , 3.24080556204789e-46 ], [ 986.4375 , 3.241160349569515e-46 ], 
                    [ 986.5 , 3.2415151759315874e-46 ], [ 986.5625 , 3.2418700411382656e-46 ], [ 986.625 , 3.242224945193849e-46 ], [ 986.6875 , 3.242579888102613e-46 ], 
                    [ 986.75 , 3.2429348698687657e-46 ], [ 986.8125 , 3.2432898904965833e-46 ], [ 986.875 , 3.243644949990344e-46 ], [ 986.9375 , 3.244000048354256e-46 ], 
                    [ 987.0 , 3.244355185592621e-46 ], [ 987.0625 , 3.2447103617096253e-46 ], [ 987.125 , 3.245065576709618e-46 ], [ 987.1875 , 3.245420830596809e-46 ], 
                    [ 987.25 , 3.2457761233754098e-46 ], [ 987.3125 , 3.24613145504977e-46 ], [ 987.375 , 3.246486825624102e-46 ], [ 987.4375 , 3.2468422351026404e-46 ], 
                    [ 987.5 , 3.2471976834896683e-46 ], [ 987.5625 , 3.247552689915258e-46 ], [ 987.625 , 3.2479076279768353e-46 ], [ 987.6875 , 3.2482626048310176e-46 ], 
                    [ 987.75 , 3.248617620482067e-46 ], [ 987.8125 , 3.2489726749341784e-46 ], [ 987.875 , 3.2493277681916148e-46 ], [ 987.9375 , 3.2496829002586413e-46 ], 
                    [ 988.0 , 3.2500380711394526e-46 ], [ 988.0625 , 3.2503932808383147e-46 ], [ 988.125 , 3.2507485293594693e-46 ], [ 988.1875 , 3.2511038167071604e-46 ], 
                    [ 988.25 , 3.2514591428856306e-46 ], [ 988.3125 , 3.251814507899125e-46 ], [ 988.375 , 3.2521699117519098e-46 ], [ 988.4375 , 3.2525253544482072e-46 ], 
                    [ 988.5 , 3.25288083599224e-46 ], [ 988.5625 , 3.253236356388277e-46 ], [ 988.625 , 3.2535919156405643e-46 ], [ 988.6875 , 3.2539475137533483e-46 ], 
                    [ 988.75 , 3.2543031507309004e-46 ], [ 988.8125 , 3.2546588265774207e-46 ], [ 988.875 , 3.2550145412972046e-46 ], [ 988.9375 , 3.2553702948944306e-46 ], 
                    [ 989.0 , 3.255726087373418e-46 ], [ 989.0625 , 3.256081918738416e-46 ], [ 989.125 , 3.2564377889936277e-46 ], [ 989.1875 , 3.2567936981433042e-46 ], 
                    [ 989.25 , 3.2571496461917657e-46 ], [ 989.3125 , 3.257505633143171e-46 ], [ 989.375 , 3.257861659001888e-46 ], [ 989.4375 , 3.2582177237720534e-46 ], 
                    [ 989.5 , 3.258573827458012e-46 ], [ 989.5625 , 3.258929970063901e-46 ], [ 989.625 , 3.259286151594114e-46 ], [ 989.6875 , 3.259642372052743e-46 ], 
                    [ 989.75 , 3.2599986289814875e-46 ], [ 989.8125 , 3.2603543571766837e-46 ], [ 989.875 , 3.26071012418863e-46 ], [ 989.9375 , 3.2610659300215373e-46 ], 
                    [ 990.0 , 3.2614217746796884e-46 ], [ 990.0625 , 3.2617776581673204e-46 ], [ 990.125 , 3.2621335804885773e-46 ], [ 990.1875 , 3.2624895416478122e-46 ], 
                    [ 990.25 , 3.262845541649241e-46 ], [ 990.3125 , 3.2632015804970543e-46 ], [ 990.375 , 3.263557658195492e-46 ], [ 990.4375 , 3.2639137747488625e-46 ], 
                    [ 990.5 , 3.2642699301613364e-46 ], [ 990.5625 , 3.2646261244371534e-46 ], [ 990.625 , 3.2649823575806013e-46 ], [ 990.6875 , 3.265338629595898e-46 ], 
                    [ 990.75 , 3.2656949404872847e-46 ], [ 990.8125 , 3.2660512902590036e-46 ], [ 990.875 , 3.266407678915298e-46 ], [ 990.9375 , 3.26676410646041e-46 ], 
                    [ 991.0 , 3.267120572898584e-46 ], [ 991.0625 , 3.26747707823404e-46 ], [ 991.125 , 3.2678336224710924e-46 ], [ 991.1875 , 3.268190205613917e-46 ], 
                    [ 991.25 , 3.268546827666781e-46 ], [ 991.3125 , 3.2689034886339542e-46 ], [ 991.375 , 3.2692601885196143e-46 ], [ 991.4375 , 3.269616927328076e-46 ], 
                    [ 991.5 , 3.269973705063564e-46 ], [ 991.5625 , 3.2703305217303033e-46 ], [ 991.625 , 3.270687377332588e-46 ], [ 991.6875 , 3.271044271874666e-46 ], 
                    [ 991.75 , 3.271401205360718e-46 ], [ 991.8125 , 3.2717581777950626e-46 ], [ 991.875 , 3.2721151891819496e-46 ], [ 991.9375 , 3.2724722395255366e-46 ], 
                    [ 992.0 , 3.2728292641044067e-46 ], [ 992.0625 , 3.273185802600289e-46 ], [ 992.125 , 3.2735423799370676e-46 ], [ 992.1875 , 3.2738989961189973e-46 ], 
                    [ 992.25 , 3.2742556511503328e-46 ], [ 992.3125 , 3.2746123450353073e-46 ], [ 992.375 , 3.2749690777781057e-46 ], [ 992.4375 , 3.2753258493829387e-46 ], 
                    [ 992.5 , 3.275682659854133e-46 ], [ 992.5625 , 3.2760395091958755e-46 ], [ 992.625 , 3.2763963974123542e-46 ], [ 992.6875 , 3.276753324507898e-46 ], 
                    [ 992.75 , 3.277110290486672e-46 ], [ 992.8125 , 3.2774672953529815e-46 ], [ 992.875 , 3.2778243391109703e-46 ], [ 992.9375 , 3.2781814217649217e-46 ], 
                    [ 993.0 , 3.278538543319073e-46 ], [ 993.0625 , 3.278895703777708e-46 ], [ 993.125 , 3.279252903144973e-46 ], [ 993.1875 , 3.2796101414251517e-46 ], 
                    [ 993.25 , 3.2799674186225072e-46 ], [ 993.3125 , 3.280324734741233e-46 ], [ 993.375 , 3.280682089785592e-46 ], [ 993.4375 , 3.2810394837598478e-46 ], 
                    [ 993.5 , 3.281396916668172e-46 ], [ 993.5625 , 3.2817543885149217e-46 ], [ 993.625 , 3.2821118993042e-46 ], [ 993.6875 , 3.282469449040389e-46 ], 
                    [ 993.75 , 3.2828270377276605e-46 ], [ 993.8125 , 3.2831846653701893e-46 ], [ 993.875 , 3.283542331972358e-46 ], [ 993.9375 , 3.2839000375382946e-46 ], 
                    [ 994.0 , 3.2842577820722668e-46 ], [ 994.0625 , 3.2846155655785898e-46 ], [ 994.125 , 3.2849733880614633e-46 ], [ 994.1875 , 3.2853312495251092e-46 ], 
                    [ 994.25 , 3.2856890708005843e-46 ], [ 994.3125 , 3.2860464189673558e-46 ], [ 994.375 , 3.286403805998993e-46 ], [ 994.4375 , 3.2867612318996535e-46 ], 
                    [ 994.5 , 3.2871186966736104e-46 ], [ 994.5625 , 3.2874762003250695e-46 ], [ 994.625 , 3.2878337428582113e-46 ], [ 994.6875 , 3.2881913242773817e-46 ], 
                    [ 994.75 , 3.288548944586693e-46 ], [ 994.8125 , 3.288906603790445e-46 ], [ 994.875 , 3.289264301892868e-46 ], [ 994.9375 , 3.2896220388980752e-46 ], 
                    [ 995.0 , 3.2899798148104853e-46 ], [ 995.0625 , 3.2903376296341887e-46 ], [ 995.125 , 3.2906954833735122e-46 ], [ 995.1875 , 3.291053376032593e-46 ], 
                    [ 995.25 , 3.2914113076156887e-46 ], [ 995.3125 , 3.291769278127102e-46 ], [ 995.375 , 3.2921272875710208e-46 ], [ 995.4375 , 3.292485335951655e-46 ], 
                    [ 995.5 , 3.2928434232733333e-46 ], [ 995.5625 , 3.293201549540151e-46 ], [ 995.625 , 3.293559714756413e-46 ], [ 995.6875 , 3.2939179189263794e-46 ], 
                    [ 995.75 , 3.2942761620542635e-46 ], [ 995.8125 , 3.294634444144278e-46 ], [ 995.875 , 3.2949927652006618e-46 ], [ 995.9375 , 3.2953511252277214e-46 ], 
                    [ 996.0 , 3.2957095242296497e-46 ], [ 996.0625 , 3.2960679622106855e-46 ], [ 996.125 , 3.2964264391750436e-46 ], [ 996.1875 , 3.296784955127012e-46 ], 
                    [ 996.25 , 3.297143510070783e-46 ], [ 996.3125 , 3.2975021040106216e-46 ], [ 996.375 , 3.2978607369508387e-46 ], [ 996.4375 , 3.2982194088955123e-46 ], 
                    [ 996.5 , 3.2985780750202346e-46 ], [ 996.5625 , 3.298936237538697e-46 ], [ 996.625 , 3.299294438946785e-46 ], [ 996.6875 , 3.2996526792486983e-46 ], 
                    [ 996.75 , 3.30001095844866e-46 ], [ 996.8125 , 3.3003692765508697e-46 ], [ 996.875 , 3.300727633559598e-46 ], [ 996.9375 , 3.3010860294790476e-46 ], 
                    [ 997.0 , 3.3014444643134417e-46 ], [ 997.0625 , 3.301802938067007e-46 ], [ 997.125 , 3.302161450743945e-46 ], [ 997.1875 , 3.3025200023485057e-46 ], 
                    [ 997.25 , 3.3028785928849633e-46 ], [ 997.3125 , 3.303237222357474e-46 ], [ 997.375 , 3.3035958907702897e-46 ], [ 997.4375 , 3.3039545981276148e-46 ], 
                    [ 997.5 , 3.3043133444337015e-46 ], [ 997.5625 , 3.3046721296928018e-46 ], [ 997.625 , 3.3050309539091452e-46 ], [ 997.6875 , 3.305389817086893e-46 ], 
                    [ 997.75 , 3.30574871923032e-46 ], [ 997.8125 , 3.306107660343706e-46 ], [ 997.875 , 3.3064666404311644e-46 ], [ 997.9375 , 3.3068256594970448e-46 ], 
                    [ 998.0 , 3.307184717545532e-46 ], [ 998.0625 , 3.307543814580882e-46 ], [ 998.125 , 3.307902950607259e-46 ], [ 998.1875 , 3.3082621256289886e-46 ], 
                    [ 998.25 , 3.3086213396502364e-46 ], [ 998.3125 , 3.3089805926752355e-46 ], [ 998.375 , 3.3093398847082457e-46 ], [ 998.4375 , 3.309699215753525e-46 ], 
                    [ 998.5 , 3.310058585815263e-46 ], [ 998.5625 , 3.310417994897743e-46 ], [ 998.625 , 3.3107774430051783e-46 ], [ 998.6875 , 3.311136930141806e-46 ], 
                    [ 998.75 , 3.3114964563118416e-46 ], [ 998.8125 , 3.3118554732906836e-46 ], [ 998.875 , 3.312214492402745e-46 ], [ 998.9375 , 3.312573550434017e-46 ], 
                    [ 999.0 , 3.312932647388719e-46 ], [ 999.0625 , 3.3132917832709756e-46 ], [ 999.125 , 3.3136509580851486e-46 ], [ 999.1875 , 3.3140101718353635e-46 ], 
                    [ 999.25 , 3.314369424525818e-46 ], [ 999.3125 , 3.314728716160852e-46 ], [ 999.375 , 3.3150880467445678e-46 ], [ 999.4375 , 3.3154474162812594e-46 ], 
                    [ 999.5 , 3.315806824775149e-46 ], [ 999.5625 , 3.316166272230366e-46 ], [ 999.625 , 3.316525758651251e-46 ], [ 999.6875 , 3.3168852840419576e-46 ], 
                    [ 999.75 , 3.3172448484068054e-46 ], [ 999.8125 , 3.3176044517498536e-46 ], [ 999.875 , 3.3179640940754688e-46 ], [ 999.9375 , 3.318323775387783e-46 ], 
                    [ 1000.0 , 3.3186834956911643e-46 ], [ 1000.0625 , 3.319043254989674e-46 ], [ 1000.125 , 3.3194030532876335e-46 ], [ 1000.1875 , 3.3197628905892707e-46 ], 
                    [ 1000.25 , 3.3201227668987194e-46 ], [ 1000.3125 , 3.3204826822203263e-46 ], [ 1000.375 , 3.3208426365582734e-46 ], [ 1000.4375 , 3.3212026299167427e-46 ], 
                    [ 1000.5 , 3.3215626623000824e-46 ], [ 1000.5625 , 3.321922733712428e-46 ], [ 1000.625 , 3.3222828441580114e-46 ], [ 1000.6875 , 3.3226429936411107e-46 ], 
                    [ 1000.75 , 3.3230031821659105e-46 ], [ 1000.8125 , 3.3233634097367135e-46 ], [ 1000.875 , 3.323723676357659e-46 ], [ 1000.9375 , 3.3240839820330735e-46 ], 
                    [ 1001.0 , 3.3244443267670973e-46 ], [ 1001.0625 , 3.3248042753366188e-46 ], [ 1001.125 , 3.325164090906689e-46 ], [ 1001.1875 , 3.325523945416545e-46 ], 
                    [ 1001.25 , 3.3258838388704007e-46 ], [ 1001.3125 , 3.326243771272518e-46 ], [ 1001.375 , 3.3266037426271357e-46 ], [ 1001.4375 , 3.3269637529383748e-46 ], 
                    [ 1001.5 , 3.327323802210498e-46 ], [ 1001.5625 , 3.3276838904476988e-46 ], [ 1001.625 , 3.328044017654241e-46 ], [ 1001.6875 , 3.328404183834342e-46 ], 
                    [ 1001.75 , 3.3287643889921725e-46 ], [ 1001.8125 , 3.3291246331319734e-46 ], [ 1001.875 , 3.3294849162579644e-46 ], [ 1001.9375 , 3.329845238374388e-46 ], 
                    [ 1002.0 , 3.33020559948544e-46 ], [ 1002.0625 , 3.330565999595294e-46 ], [ 1002.125 , 3.3309264387082636e-46 ], [ 1002.1875 , 3.3312869168285236e-46 ], 
                    [ 1002.25 , 3.331647433960201e-46 ], [ 1002.3125 , 3.3320079901077057e-46 ], [ 1002.375 , 3.3323685852751665e-46 ], [ 1002.4375 , 3.3327292194667824e-46 ], 
                    [ 1002.5 , 3.3330898926867994e-46 ], [ 1002.5625 , 3.333450604939466e-46 ], [ 1002.625 , 3.333811356228959e-46 ], [ 1002.6875 , 3.334172146559526e-46 ], 
                    [ 1002.75 , 3.3345329759354164e-46 ], [ 1002.8125 , 3.3348938443608315e-46 ], [ 1002.875 , 3.3352547518399737e-46 ], [ 1002.9375 , 3.33561569837707e-46 ], 
                    [ 1003.0 , 3.335976683976418e-46 ], [ 1003.0625 , 3.3363377086421738e-46 ], [ 1003.125 , 3.3366987723785654e-46 ], [ 1003.1875 , 3.3370598751898926e-46 ], 
                    [ 1003.25 , 3.3374210170803125e-46 ], [ 1003.3125 , 3.3377819526937122e-46 ], [ 1003.375 , 3.338142578283566e-46 ], [ 1003.4375 , 3.338503242836665e-46 ], 
                    [ 1003.5 , 3.3388639463572186e-46 ], [ 1003.5625 , 3.3392246888494135e-46 ], [ 1003.625 , 3.3395854703175082e-46 ], [ 1003.6875 , 3.3399462907656415e-46 ], 
                    [ 1003.75 , 3.3403071501981207e-46 ], [ 1003.8125 , 3.34066804861911e-46 ], [ 1003.875 , 3.3410289860328218e-46 ], [ 1003.9375 , 3.3413899624434455e-46 ], 
                    [ 1004.0 , 3.341750977855242e-46 ], [ 1004.0625 , 3.3421120322723534e-46 ], [ 1004.125 , 3.3424731256990893e-46 ], [ 1004.1875 , 3.3428342581396174e-46 ], 
                    [ 1004.25 , 3.3431954295981284e-46 ], [ 1004.3125 , 3.3435566400788622e-46 ], [ 1004.375 , 3.343917889586058e-46 ], [ 1004.4375 , 3.3442791781239093e-46 ], 
                    [ 1004.5 , 3.344640505696609e-46 ], [ 1004.5625 , 3.3450018723084223e-46 ], [ 1004.625 , 3.3453632779634955e-46 ], [ 1004.6875 , 3.3457247226661426e-46 ], 
                    [ 1004.75 , 3.3460862064205102e-46 ], [ 1004.8125 , 3.346447729230866e-46 ], [ 1004.875 , 3.3468092911014058e-46 ], [ 1004.9375 , 3.347170892036349e-46 ], 
                    [ 1005.0 , 3.3475325320398937e-46 ], [ 1005.0625 , 3.3478942111162842e-46 ], [ 1005.125 , 3.348255929269766e-46 ], [ 1005.1875 , 3.348617686504537e-46 ], 
                    [ 1005.25 , 3.348979482824772e-46 ], [ 1005.3125 , 3.3493413182347897e-46 ], [ 1005.375 , 3.3497031927387656e-46 ], [ 1005.4375 , 3.3500651063408995e-46 ], 
                    [ 1005.5 , 3.3504270590454638e-46 ], [ 1005.5625 , 3.3507890347774895e-46 ], [ 1005.625 , 3.3511504744888967e-46 ], [ 1005.6875 , 3.351511953187738e-46 ], 
                    [ 1005.75 , 3.351873470878218e-46 ], [ 1005.8125 , 3.3522350275645906e-46 ], [ 1005.875 , 3.3525966232509437e-46 ], [ 1005.9375 , 3.3529582579416023e-46 ], 
                    [ 1006.0 , 3.353319931640702e-46 ], [ 1006.0625 , 3.353681644352451e-46 ], [ 1006.125 , 3.3540433960811297e-46 ], [ 1006.1875 , 3.3544051868308506e-46 ], 
                    [ 1006.25 , 3.3547670166059416e-46 ], [ 1006.3125 , 3.3551288854104702e-46 ], [ 1006.375 , 3.3554907932487645e-46 ], [ 1006.4375 , 3.355852740124941e-46 ], 
                    [ 1006.5 , 3.356214726043304e-46 ], [ 1006.5625 , 3.3565767510079953e-46 ], [ 1006.625 , 3.3569388150232497e-46 ], [ 1006.6875 , 3.3573009180932794e-46 ], 
                    [ 1006.75 , 3.357663060222321e-46 ], [ 1006.8125 , 3.35802524141454e-46 ], [ 1006.875 , 3.3583874616741736e-46 ], [ 1006.9375 , 3.358749721005461e-46 ], 
                    [ 1007.0 , 3.3591120194125676e-46 ], [ 1007.0625 , 3.3594743568997327e-46 ], [ 1007.125 , 3.359836733471172e-46 ], [ 1007.1875 , 3.360199149131077e-46 ], 
                    [ 1007.25 , 3.3605616038837605e-46 ], [ 1007.3125 , 3.3609240977333187e-46 ], [ 1007.375 , 3.3612866306839935e-46 ], [ 1007.4375 , 3.3616492027400983e-46 ], 
                    [ 1007.5 , 3.362011813905684e-46 ], [ 1007.5625 , 3.3623744641851362e-46 ], [ 1007.625 , 3.362737153582578e-46 ], [ 1007.6875 , 3.3630998821023015e-46 ], 
                    [ 1007.75 , 3.3634626497484307e-46 ], [ 1007.8125 , 3.36382545652521e-46 ], [ 1007.875 , 3.364187954825835e-46 ], [ 1007.9375 , 3.3645502190382844e-46 ], 
                    [ 1008.0 , 3.3649125222602307e-46 ], [ 1008.0625 , 3.3652748644959225e-46 ], [ 1008.125 , 3.365637245749489e-46 ], [ 1008.1875 , 3.3659996660252028e-46 ], 
                    [ 1008.25 , 3.366362125327267e-46 ], [ 1008.3125 , 3.366724623659836e-46 ], [ 1008.375 , 3.367087161027136e-46 ], [ 1008.4375 , 3.367449737433346e-46 ], 
                    [ 1008.5 , 3.3678123528827673e-46 ], [ 1008.5625 , 3.368175007379483e-46 ], [ 1008.625 , 3.3685377009277226e-46 ], [ 1008.6875 , 3.3689004335317385e-46 ], 
                    [ 1008.75 , 3.3692632051957364e-46 ], [ 1008.8125 , 3.369626015923875e-46 ], [ 1008.875 , 3.3699888657203844e-46 ], [ 1008.9375 , 3.370351754589448e-46 ], 
                    [ 1009.0 , 3.3707146825353446e-46 ], [ 1009.0625 , 3.3710776495622102e-46 ], [ 1009.125 , 3.3714406556743016e-46 ], [ 1009.1875 , 3.371803700875755e-46 ], 
                    [ 1009.25 , 3.3721667851708758e-46 ], [ 1009.3125 , 3.3725299085637547e-46 ], [ 1009.375 , 3.3728930710587918e-46 ], [ 1009.4375 , 3.373256272659984e-46 ], 
                    [ 1009.5 , 3.373619513371661e-46 ], [ 1009.5625 , 3.37398279319801e-46 ], [ 1009.625 , 3.374346112143268e-46 ], [ 1009.6875 , 3.3747094702115752e-46 ], 
                    [ 1009.75 , 3.3750728674072395e-46 ], [ 1009.8125 , 3.3754363037343553e-46 ], [ 1009.875 , 3.3757997791972803e-46 ], [ 1009.9375 , 3.3761632938001803e-46 ], 
                    [ 1010.0 , 3.3765268475471984e-46 ], [ 1010.0625 , 3.376890440442549e-46 ], [ 1010.125 , 3.377254067843078e-46 ], [ 1010.1875 , 3.3776171580494017e-46 ], 
                    [ 1010.25 , 3.377980287291728e-46 ], [ 1010.3125 , 3.3783434555742528e-46 ], [ 1010.375 , 3.3787066629011505e-46 ], [ 1010.4375 , 3.3790699092766423e-46 ], 
                    [ 1010.5 , 3.3794331947049503e-46 ], [ 1010.5625 , 3.379796519190225e-46 ], [ 1010.625 , 3.3801598827367136e-46 ], [ 1010.6875 , 3.3805232853485917e-46 ], 
                    [ 1010.75 , 3.380886727030035e-46 ], [ 1010.8125 , 3.381250207785316e-46 ], [ 1010.875 , 3.3816137276185634e-46 ], [ 1010.9375 , 3.3819772865339547e-46 ], 
                    [ 1011.0 , 3.3823408845357873e-46 ], [ 1011.0625 , 3.382704521628216e-46 ], [ 1011.125 , 3.383068197815443e-46 ], [ 1011.1875 , 3.3834319131016714e-46 ], 
                    [ 1011.25 , 3.383795667491105e-46 ], [ 1011.3125 , 3.3841594609879472e-46 ], [ 1011.375 , 3.3845232935963554e-46 ], [ 1011.4375 , 3.384887165320606e-46 ], 
                    [ 1011.5 , 3.385251076164952e-46 ], [ 1011.5625 , 3.385615026133456e-46 ], [ 1011.625 , 3.3859790152303956e-46 ], [ 1011.6875 , 3.3863430434600257e-46 ], 
                    [ 1011.75 , 3.386707110826434e-46 ], [ 1011.8125 , 3.387071217333995e-46 ], [ 1011.875 , 3.387435362986774e-46 ], [ 1011.9375 , 3.3877995477889546e-46 ], 
                    [ 1012.0 , 3.3881637717448904e-46 ], [ 1012.0625 , 3.3885280348587185e-46 ], [ 1012.125 , 3.3888923371346488e-46 ], [ 1012.1875 , 3.389256678576892e-46 ], 
                    [ 1012.25 , 3.389621059189658e-46 ], [ 1012.3125 , 3.389985478977135e-46 ], [ 1012.375 , 3.3903499379435583e-46 ], [ 1012.4375 , 3.390714203020327e-46 ], 
                    [ 1012.5 , 3.391078120391157e-46 ], [ 1012.5625 , 3.3914420768203572e-46 ], [ 1012.625 , 3.3918060723121442e-46 ], [ 1012.6875 , 3.3921701068707096e-46 ], 
                    [ 1012.75 , 3.3925341805002232e-46 ], [ 1012.8125 , 3.392898293204926e-46 ], [ 1012.875 , 3.3932624449889883e-46 ], [ 1012.9375 , 3.3936266358566042e-46 ], 
                    [ 1013.0 , 3.3939908658119678e-46 ], [ 1013.0625 , 3.3943551348592508e-46 ], [ 1013.125 , 3.394719443002697e-46 ], [ 1013.1875 , 3.3950837902464786e-46 ], 
                    [ 1013.25 , 3.3954481765948155e-46 ], [ 1013.3125 , 3.3958126020518094e-46 ], [ 1013.375 , 3.3961770666218017e-46 ], [ 1013.4375 , 3.3965415703089177e-46 ], 
                    [ 1013.5 , 3.3969061131172837e-46 ], [ 1013.5625 , 3.3972706950512434e-46 ], [ 1013.625 , 3.397635316114899e-46 ], [ 1013.6875 , 3.3979999763124753e-46 ], 
                    [ 1013.75 , 3.3983646756481715e-46 ], [ 1013.8125 , 3.398729414126189e-46 ], [ 1013.875 , 3.3990941917507284e-46 ], [ 1013.9375 , 3.3994590085259667e-46 ], 
                    [ 1014.0 , 3.399823864456179e-46 ], [ 1014.0625 , 3.4001887595454706e-46 ], [ 1014.125 , 3.4005536937981173e-46 ], [ 1014.1875 , 3.4009186672182977e-46 ], 
                    [ 1014.25 , 3.401283679810216e-46 ], [ 1014.3125 , 3.4016487315780766e-46 ], [ 1014.375 , 3.402013822526083e-46 ], [ 1014.4375 , 3.402378952658393e-46 ], 
                    [ 1014.5 , 3.4027441219793092e-46 ], [ 1014.5625 , 3.4031093304929638e-46 ], [ 1014.625 , 3.403474578203613e-46 ], [ 1014.6875 , 3.403839865115438e-46 ], 
                    [ 1014.75 , 3.404204791535615e-46 ], [ 1014.8125 , 3.404569550418012e-46 ], [ 1014.875 , 3.404934348384072e-46 ], [ 1014.9375 , 3.405299185438129e-46 ], 
                    [ 1015.0 , 3.405664061584274e-46 ], [ 1015.0625 , 3.406028976826671e-46 ], [ 1015.125 , 3.406393931169631e-46 ], [ 1015.1875 , 3.406758924617223e-46 ], 
                    [ 1015.25 , 3.4071239571737087e-46 ], [ 1015.3125 , 3.407489028843279e-46 ], [ 1015.375 , 3.4078541396300763e-46 ], [ 1015.4375 , 3.408219289538292e-46 ], 
                    [ 1015.5 , 3.408584478572166e-46 ], [ 1015.5625 , 3.4089497067358673e-46 ], [ 1015.625 , 3.4093149740335637e-46 ], [ 1015.6875 , 3.4096802804694977e-46 ], 
                    [ 1015.75 , 3.4100456260478127e-46 ], [ 1015.8125 , 3.4104110107727774e-46 ], [ 1015.875 , 3.410776434648537e-46 ], [ 1015.9375 , 3.411141897679189e-46 ], 
                    [ 1016.0 , 3.4115073998691002e-46 ], [ 1016.0625 , 3.411872941222392e-46 ], [ 1016.125 , 3.4122385217432613e-46 ], [ 1016.1875 , 3.41260414143593e-46 ], 
                    [ 1016.25 , 3.412969800304497e-46 ], [ 1016.3125 , 3.4133354983532823e-46 ], [ 1016.375 , 3.413701235586483e-46 ], [ 1016.4375 , 3.414067012008153e-46 ], 
                    [ 1016.5 , 3.4144328276226125e-46 ], [ 1016.5625 , 3.414798682434061e-46 ], [ 1016.625 , 3.4151645764466495e-46 ], [ 1016.6875 , 3.4155305096646753e-46 ], 
                    [ 1016.75 , 3.415896482092171e-46 ], [ 1016.8125 , 3.4162624937335057e-46 ], [ 1016.875 , 3.4166285445927603e-46 ], [ 1016.9375 , 3.416994634674211e-46 ], 
                    [ 1017.0 , 3.417360763982059e-46 ], [ 1017.0625 , 3.4177264047594775e-46 ], [ 1017.125 , 3.418091998163455e-46 ], [ 1017.1875 , 3.4184576306749226e-46 ], 
                    [ 1017.25 , 3.418823302297967e-46 ], [ 1017.3125 , 3.419189013036845e-46 ], [ 1017.375 , 3.4195547628956918e-46 ], [ 1017.4375 , 3.4199205518787405e-46 ], 
                    [ 1017.5 , 3.420286379990153e-46 ], [ 1017.5625 , 3.4206522472341134e-46 ], [ 1017.625 , 3.4210181536148086e-46 ], [ 1017.6875 , 3.421384099136376e-46 ], 
                    [ 1017.75 , 3.4217500838031244e-46 ], [ 1017.8125 , 3.422116107619119e-46 ], [ 1017.875 , 3.4224821705886217e-46 ], [ 1017.9375 , 3.4228482727157953e-46 ], 
                    [ 1018.0 , 3.423214414004829e-46 ], [ 1018.0625 , 3.4235805944598876e-46 ], [ 1018.125 , 3.423946814085209e-46 ], [ 1018.1875 , 3.424313072884959e-46 ], 
                    [ 1018.25 , 3.4246793708633526e-46 ], [ 1018.3125 , 3.4250457080244837e-46 ], [ 1018.375 , 3.4254120843726892e-46 ], [ 1018.4375 , 3.425778499912063e-46 ], 
                    [ 1018.5 , 3.42614495464675e-46 ], [ 1018.5625 , 3.4265114485811128e-46 ], [ 1018.625 , 3.426877981719223e-46 ], [ 1018.6875 , 3.4272445540652975e-46 ], 
                    [ 1018.75 , 3.427611165623556e-46 ], [ 1018.8125 , 3.4279778163981195e-46 ], [ 1018.875 , 3.4283445063932313e-46 ], [ 1018.9375 , 3.428711235613136e-46 ], 
                    [ 1019.0 , 3.4290780040619076e-46 ], [ 1019.0625 , 3.4294448117438887e-46 ], [ 1019.125 , 3.429811658663154e-46 ], [ 1019.1875 , 3.430178544823974e-46 ], 
                    [ 1019.25 , 3.4305454702305216e-46 ], [ 1019.3125 , 3.430912434886971e-46 ], [ 1019.375 , 3.4312788414281747e-46 ], [ 1019.4375 , 3.4316452724090387e-46 ], 
                    [ 1019.5 , 3.432011742521546e-46 ], [ 1019.5625 , 3.4323782517699247e-46 ], [ 1019.625 , 3.4327448001582812e-46 ], [ 1019.6875 , 3.433111387690867e-46 ], 
                    [ 1019.75 , 3.4334780143718636e-46 ], [ 1019.8125 , 3.4338446802053785e-46 ], [ 1019.875 , 3.4342113851956904e-46 ], [ 1019.9375 , 3.4345781293469073e-46 ], 
                    [ 1020.0 , 3.4349449126632602e-46 ], [ 1020.0625 , 3.435311735148883e-46 ], [ 1020.125 , 3.4356785968080073e-46 ], [ 1020.1875 , 3.4360454976447676e-46 ], 
                    [ 1020.25 , 3.4364124376633723e-46 ], [ 1020.3125 , 3.4367794168680303e-46 ], [ 1020.375 , 3.437146435262877e-46 ], [ 1020.4375 , 3.437513492852098e-46 ], 
                    [ 1020.5 , 3.4378805896399275e-46 ], [ 1020.5625 , 3.4382477256305516e-46 ], [ 1020.625 , 3.438614900828132e-46 ], [ 1020.6875 , 3.438982115236808e-46 ], 
                    [ 1020.75 , 3.4393493688608636e-46 ], [ 1020.8125 , 3.4397166617044625e-46 ], [ 1020.875 , 3.440083993771647e-46 ], [ 1020.9375 , 3.4404513650668742e-46 ], 
                    [ 1021.0 , 3.4408187755940885e-46 ], [ 1021.0625 , 3.441186225357627e-46 ], [ 1021.125 , 3.4415537143615813e-46 ], [ 1021.1875 , 3.44192124261024e-46 ], 
                    [ 1021.25 , 3.4422888101077207e-46 ], [ 1021.3125 , 3.4426564168582395e-46 ], [ 1021.375 , 3.443024062865988e-46 ], [ 1021.4375 , 3.443391748135134e-46 ], 
                    [ 1021.5 , 3.4437594726699932e-46 ], [ 1021.5625 , 3.4441272364745384e-46 ], [ 1021.625 , 3.4444950336527714e-46 ], [ 1021.6875 , 3.4448622613192686e-46 ], 
                    [ 1021.75 , 3.445229528136979e-46 ], [ 1021.8125 , 3.445596834110052e-46 ], [ 1021.875 , 3.445964179242663e-46 ], [ 1021.9375 , 3.446331563539083e-46 ], 
                    [ 1022.0 , 3.4466989870033664e-46 ], [ 1022.0625 , 3.4470664496397618e-46 ], [ 1022.125 , 3.447433951452421e-46 ], [ 1022.1875 , 3.447801492445472e-46 ], 
                    [ 1022.25 , 3.448169072623189e-46 ], [ 1022.3125 , 3.448536691989702e-46 ], [ 1022.375 , 3.4489043505491887e-46 ], [ 1022.4375 , 3.449272048305802e-46 ], 
                    [ 1022.5 , 3.4496397852637462e-46 ], [ 1022.5625 , 3.4500075614272253e-46 ], [ 1022.625 , 3.4503753768003454e-46 ], [ 1022.6875 , 3.4507432313873844e-46 ], 
                    [ 1022.75 , 3.4511111251924744e-46 ], [ 1022.8125 , 3.4514790582197723e-46 ], [ 1022.875 , 3.4518470304734833e-46 ], [ 1022.9375 , 3.45221504195784e-46 ], 
                    [ 1023.0 , 3.4525830926769493e-46 ], [ 1023.0625 , 3.452951182634971e-46 ], [ 1023.125 , 3.4533193118361865e-46 ], [ 1023.1875 , 3.453687480284681e-46 ], 
                    [ 1023.25 , 3.454055687984713e-46 ], [ 1023.3125 , 3.454423934940491e-46 ], [ 1023.375 , 3.4547922211561027e-46 ], [ 1023.4375 , 3.4551605466358315e-46 ], 
                    [ 1023.5 , 3.4555289113837654e-46 ], [ 1023.5625 , 3.4558973154041393e-46 ], [ 1023.625 , 3.456265758701117e-46 ], [ 1023.6875 , 3.4566342412789833e-46 ], 
                    [ 1023.75 , 3.4570027631418034e-46 ], [ 1023.8125 , 3.4573713242938145e-46 ], [ 1023.875 , 3.45773992473923e-46 ], [ 1023.9375 , 3.4581085644822143e-46 ], 
                    [ 1024.0 , 0.0 ] ])

    f1 = data[:,0]
    S1 = data[:,1]
    fstop = [f1[0],f1[-1]]

    #% ---- Interpolate to desired frequencies.  Set noise to infinity outside
    #%      band of frequencies covered by f1.
    ##S = inf(size(f)); 
    ##k = find(f>=f1(1) & f<=f1(end));
    S = interp1d(f1,S1)

    return S, fstop